/* editcen.c -- XGEN functionality for editing a CENTROIDS file

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions; reorganize use of "{"
	27 Nov 2006: makeq21 only takes xp as an argument
	 2 Jun 2006: extern declarations for g_centroid,p_centroid
	19 Mar 2006: bugfixes: always calc hkl; display initial refl correctly;
		replace %7.2f with %7.1f formats in basic reflection display
	 7 Mar 2006: rename "frame" element of REFREF to "framem"
	 3 Oct 2005: housekeeping
	11 May 2005: check return status from xgjournal
	23 Apr 2005: refl_proc renamed refl_edit to reduce confusion
	12 Sep 2004: correctly handle case where parameters aren't available
	11 Sep 2004: use scattv() and refgonio() to get diffraction vectors
	10 Sep 2004: rescale (rfp->rfr_xp)->xrec by lambda in edcen_setup
	 6 Sep 2004: bugfix: edcen_setup uses rfp->rfr_xp explicitly; cosmetics
	28 Aug 2004: routines used only here are declared as static
	19 Aug 2004: static globals replaced by elements of EDICENRUN structure
	18 Aug 2004: replace sgsp with rfp->rfr_spl in edcen_setup
	24 Jan 2004: omit opening UPARAMS, since that's done in refparin()
	21 Jan 2004: license message inserted
	 3 Jan 2004: replace external globals with elements of REFPARS
		structures from function arguments and a few static globals;
		deal with UPARAMS info using refparin()
	 7 Dec 2003: correct 3rd arg of edcen_setup(): it's &rfs, not &rms
	24 Nov 2003: replace globals associated with curses with elements
		of the REFMRUN structures
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	 5 Jan 2003: reorganized source-file header
	15 Nov 2002: extern declarations made here rather than in refiext.h
	25 Mar 2002: removed MSI attachments
	 5 Oct 1994: use xgread and xgwrite for REFPARS and TROID data
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN;
		mwp declared as a global WINDOW * for VMS
	16 Mar 1994: some meaningful bug fixes;
		global variables reorganized a little
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version
	xx xxx 1990: add "-c" functionality
			. . . various undocumented revisions . . .
	15 Oct 1985: original code
	
   Purpose: editing a CENTROIDS file to eliminate spots that violate
	the UPARAMS error limits in hkl, X, Y, or omega
   Cross references:

 This reads a "centroids" file and writes out another one,
 from which some of the observations have been deleted.
 Normally only observations which have errors in h, k, l, X, Y, or phi greater
 than a certain value are printed and presented to the user for deletion.
 If the "-a" option (for "automatic") is given, the deletions are performed
 without asking.
 If the "-e" option (for "every") is used, every reflection rather than only
 the ones violating the error criteria are presented for deletion.
 The "-c" (for "complement") option does automatic deletions, but only if
 the reflections which DO satisfy the criteria for acceptance are deleted.
 This provides a method of separating the reflections
 belonging to one lattice from the ones belonging to another lattice.
 The uninteresting conclusion from this is that using both "-a" and "-e" will
 cause all reflections to be deleted, so no output file needs to get written.
*************************************************************************/

#define		NEED_CURSES	1
#define		NEED_TIME	1
#include	"refidefs.h"
#define		ASK_DELETION	63
#define		END_DELETIONS	64

extern int g_centroid(FILE *fp, REFREF *rp);
extern int p_centroid(EDITCENRUN *ecp);

static EDITCENRUN *manageecp(int opno)
{ /*	handles memory for a static global pointer ecp */
	static EDITCENRUN	*ecp;
	static char		ecpmsg[] = "ecp in manageecp";

	if (!opno)
	  {
		if (NULL != ecp)
		  {
			if (NULL == ecp->ec_rfp) managerfp(0, 0);
			memzap((AVOID **)&ecp, ecpmsg);
		  }
		return NULL;
	  }
	if (opno == 1) ecp = manageecp(0);
	if (NULL == ecp) ecp = (EDITCENRUN *)callop(1,
		sizeof (EDITCENRUN), ecpmsg);
	return ecp;
}

static void sumedcenrun(FILE *fp, EDITCENRUN *ecp)
{ /*	This summarizes run-time conditions to the stream `fp'. */
	static char	*flist[] =
	 {"centroids", "uparams", "uspline", "amask", "cenout", "xlog", NULL};
	static char	ftype[] = {1, 1, 1, 1, 2, 2, 0 };

	header(fp, PROCESSING, flist, ftype); /* announce what we're doing */
	if (ecp->ec_flags & EC_AUTODEL) /* describe type of deletions */
	  {
		fprintf(fp," Deletions will be performed automatically\n");
		fprintf(fp,
	 " ... keeping reflections which fall %s the error limits\n",
		 (ecp->ec_flags & EC_COMPLEMENT) ? "OUTSIDE" : "INSIDE" );
	  }
	if (ecp->ec_flags & EC_EVERYREFL) fprintf(fp,
		" All reflections will be presented for examination\n");
	else if (!(ecp->ec_flags & EC_AUTODEL)) fprintf(fp,
" Only reflections with excessive errors will be presented for examination\n");
}

static int edcen_setup(int argc, char *argv[], EDITCENRUN *ecp)
{ /*	Argument and file processor for the editcen program.
	Files used are given conventional logical or environment names:
		centroids	input centroids file
		uparams		input sample & detector parameters
		uspline		input pixel-to-centimeter mapping for detector
		amask		input active-pixel bitmask
		cenout		output centroids file
		xlog		printable logfile (appended)
	Flags:
		-a		automatic deletion of violators
		-c		automatically KEEP the violators and
				 throw out the good refls.
		-d<val>		maximum relative plane spacing for closely
				 spaced reflections
		-e		test every reflection, not just violators
		-h		check for (X,Y,Omega,h) errors
		-i<val>		minimum intensity to be written out
		-l<val>		low-resolution cutoff in Angstroms
		-m<val>		max non-monotonic fraction allowed
		-n		do NOT check for (X,Y,Omega,h) errors
		-u<val>		high-resolution cutoff in Angstroms	*/
	int		nread, i, j, quality;
	char		**flags, **dcp, *val;
	double		fval;
	FRAMEINDEX	*lft;
	REFINERUN	*rfp;
	REFREF		rs;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	ecp->ec_nrin = ecp->ec_nrout = ecp->ec_nrdel = 0; /* zero counters */
	if (NULL == (lft = ftables(3))) /* frame-number limits if no FRAMES */
	  {
		ecp->ec_fmmin = -10000;	ecp->ec_fmmax = 10000;
	  }
	else {
		ecp->ec_fmmin = lft->fi_firstfm - 1;
		ecp->ec_fmmax = lft->fi_lastfm + 1;
	     }
	ecp->ec_mininten = 0;	ecp->ec_maxnonmono = 1;
	ecp->ec_minsecpeak = -1000;
	ecp->ec_resmin = 10000.;	ecp->ec_resmax = 0.;
	ecp->ec_maxreld = 10000.;
	ecp->ec_flags = EC_CHECKHXYO | EC_USECURSES;
	flags = flagproc(argc, argv, "ACDEHILMNU"); /* handle arguments */
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	   {
	     val = (*dcp) + 1;
	     fval = ((val) && ('\0' != *val)) ? atof(val) : 0.;
	     switch (**dcp)
	      {
		case	'A':	ecp->ec_flags |= EC_AUTODEL;	break;
		case	'C':	ecp->ec_flags |= (EC_AUTODEL | EC_COMPLEMENT);
			break;
		case	'D':	ecp->ec_maxreld = fval;	break;
		case	'E':	ecp->ec_flags |= EC_EVERYREFL;	break;
		case	'H':	ecp->ec_flags |= EC_CHECKHXYO;	break;
		case	'I':	ecp->ec_mininten = fval;	break;
		case	'L':	ecp->ec_resmin = fval;	break;
		case	'M':	ecp->ec_maxnonmono = fval;	break;
		case	'N':	ecp->ec_minsecpeak = fval;	break;
		case	'S':	ecp->ec_flags &= ~EC_CHECKHXYO;	break;
		case	'U':	ecp->ec_resmax = fval;	break;
		default:	break;
	      }
	   }
	if ((ecp->ec_flags & EC_AUTODEL) && (ecp->ec_flags & EC_EVERYREFL))
	 {
	  fprintf(stderr," You specified autodeletion of everything\n");
	  fprintf(stderr," You might as well not run the program at all\n");
	  return EBADVALUE;
	 }
	if (qerror((AVOID *)(ecp->ec_fpi = xgopen("centroids", "r")),
		"open input centroids file")) return EOPENERR;
	/* open the input CENTROIDS file and count the refls in it */
	for (ecp->ec_nrin = 0; (0 < xgread((void *)&rs, XG_TROID,
	 ecp->ec_fpi)); ) (ecp->ec_nrin)++;
	(void)fseek(ecp->ec_fpi, 0, 0); /* rewind centroids file */
	if (qerror((AVOID *)(ecp->ec_fpo = xgopen("cenout", "w")),
		"open output centroids file")) return EOPENERR;
	if (NULL == (rfp = managerfp(3, (size_t)(ecp->ec_nrin))))
		return EALLOC;
	ecp->ec_rfp = rfp;	/* remember that pointer */
	rfp->rfr_tl0 = NOW;
	if (EOK != (i = refparin(rfp->rfr_xp, &nread)))
	  {
		memzap((AVOID **)&(rfp->rfr_xp), "rfp->rfr_xp in filtercen");
		ecp->ec_flags &= ~EC_CHECKHXYO;
	  }
	else {	/* rotations about tilt and swing */
		makeq21(rfp->rfr_xp);
		/* for internal purposes we need the reciprocal-space matrix
		to be rescaled by lambda to make it unitless */
		for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
			(rfp->rfr_xp)->xrec[i][j] *= (rfp->rfr_xp)->xlambda;
	     }
	if (qerror((AVOID *)(ecp->ec_fpl = propen("xlog")),
	 "open output log file")) return EOPENERR;
	printf("Reading and processing centroids: ");
	/* now read all the reflections in */
	for (rfp->rfr_rpp = rfp->rfr_rpp0, i = 0; i < ecp->ec_nrin;
	 i++, (rfp->rfr_rpp)++)
	   {
		if (0 == g_centroid(ecp->ec_fpi, rfp->rfr_rpp)) break;
		if (ecp->ec_flags & EC_CHECKHXYO) quality = reflinfo(rfp);
		else quality = 0;
		if (NULL == rfp->rfr_xp) /* flag missing params */
			(rfp->rfr_rpp)->rflags |= CHANGEDINDEX;
		(rfp->rfr_rpp)->rflags |= (quality << 8);
	   }
	ecp->ec_nrin = i;	/* in case the count is different */
	printf(" %6d reflections read and processed\n", ecp->ec_nrin);
	(void)fclose(ecp->ec_fpi); /* we'll do the rest in memory */
	ecp->ec_fpi = NULL;	rfp->rfr_tl0 = NOW;
	strcpy(rfp->rfr_rundate, ctime(&(rfp->rfr_tl0)));
	sumedcenrun(stdout, ecp);	sumedcenrun(ecp->ec_fpl, ecp);
	/* set up the headers for the logfile */
	filterhead(ecp->ec_fpl, (rfp->rfr_xp != NULL));
	return EOK;
}

/* *** NO CURSES STUFF ABOVE HERE *** */
#include	"menu.h"

static WINDOW *managelistw(int opno)
{ /*	Manages a static pointer to *listwin */
	static WINDOW	*listwin;

	if (!opno)
	  {
		delwin(listwin);	listwin = NULL; return listwin;
	  }
	if (opno == 1) listwin = managelistw(0); /* force recreation */
	if (NULL == listwin) listwin = newwin(3, 80, 19, 0);
	return listwin;
}

static void smoothend(void)
{ /* routine to exit cleanly from this mess */
	EDITCENRUN	*ecp;
	WINDOW		*listw;

	signal(SIGINT,SIG_DFL);		/* ignore interrupts at this point */
	if (NULL == (ecp = manageecp(3))) return;
	if (ecp->ec_flags & EC_DONE) return; /* in case we call this twice */
	listw = managelistw(3);
	if ((!(ecp->ec_flags & EC_AUTODEL)) && (NULL != listw))
	  {
		delwin(listw);	standend();	clear();	refresh();
		endwin();	listw = NULL;
	  }
	if (NULL != ecp->ec_fpo)
	  {
		(void)fclose(ecp->ec_fpo);	ecp->ec_fpo = NULL;
	  }
	if (NULL != ecp->ec_fpl)
	  {
		sum_edcenresults(stdout, ecp);
		sum_edcenresults(ecp->ec_fpl, ecp);
	  }
	ecp->ec_flags |= EC_DONE; /* flag*/
	ecp = manageecp(0); setmzap();	/* free memory allocated in `setmore'*/
}

static int user_answer(EDITCENRUN *ecp, REFMRUN *rmp)
{ /*	This returns ASK_DELETION if the user requests a deletion;
	it returns 0 otherwise.   It performs a few other functions
	when the user makes other requests.  If it gets a "quit"
	command or an "exit" command, it returns END_DELETIONS so the
	calling routine knows that it should shuffle off to Buffalo.
	*/
	char		c;
	REFREF		*rp;
	
	move(3, 69);		/* move to query location */
	refresh();		/* update screen now */
	c = wget1ch(stdscr);	/* read it in */
	addch((xctype)(c));	/* display it */
	refresh();
	rp = (ecp->ec_rfp)->rfr_rpp;
	switch (c)
	   {
		case	'd':	/* delete this observation */
		case	'y':	return ASK_DELETION; /* ask for deletion */
		case	'q':	/* quit immediately */
			wclear(rmp->rm_mwp);	wrefresh(rmp->rm_mwp);
			smoothend();	return END_DELETIONS;
		case	'e':	/* copy rest of input out and quit */
			wclear(rmp->rm_mwp);	wrefresh(rmp->rm_mwp);
			if (NULL == rp)
			  {
				endwin(); return END_DELETIONS;
			  }
			if (EOK != p_centroid(ecp)) return EWRITERR;
			while (0 != g_centroid(ecp->ec_fpi, rp))
				if (EOK != p_centroid(ecp)) return EWRITERR;
			smoothend();	return END_DELETIONS;
		default: return 0;	/* anything else: go on to next ob */
	   }
}

static void cmdlist(void)
{ /*	This sets up and displays a summary of the available commands,
	and then restores the old windows if they have anything in them. */
	WINDOW	*listw;

	if (NULL == (listw = managelistw(3))) return;
	wclear(listw);	wrefresh(listw);
	mvwaddstr(listw, 0, 6,
		"Commands Available in the Centroids Database Editor");
	wstandout(listw);	/* make header and funcs stand out */
	mvwaddstr(listw, 1, 1,
"d,y  : mark ob for Deletion;           e : quit checking; write rest");
	mvwaddstr(listw, 2, 1,
"q,^C : quit immediately;   anything else : go on to next observation");
	wstandend(listw);	/* turn of boldface */
	wrefresh(listw);	/* put up what we just asked for */
}

static int refl_display(REFREF *rp, REFMRUN *rmp)
{ /*	This displays a reflection.  We create the CURSES baggage
	here if it isn't already present. */
	register int	i, fmno, j, nstars;
	char		intenstr[54];
	static WINDOW	*listwin;

	if (NULL == listwin)	/* first time: create the CURSES stuff. */
	  {
#ifndef VMS	/* define some terminal-dependent keys */
		if (verror(setmore(0),
			"set up keystroke equivalences for this terminal"))
				return EOPENERR;
#endif
		initscr();	/* initialize the screen-manipulation stuff */
		signal(SIGINT, (sig_t)smoothend); /* call exit routine on ^C*/
		crmode();		/* set terminal to char-by-char mode */
		noecho();		/* don't echo input back */
		nonl();			/* turn off newline mapping */
		clear();
		standout();	/* set up header for display in boldface */
		mvaddstr(1, 0,
 "    h    k    l    X (pixels)    Y (pixels)  phi (frames)   inten-");
		mvaddstr(2, 0,
 "                 obsvd calctd  obsvd calctd  obsvd calctd     sity");
		standend();	/* turn off boldface for everything else */
		/* reserve memory for and set up the window structures */
		rmp->rm_mwp = newwin(10, 80, 5, 0);	/* data window */
		listwin = managelistw(3);	/* help window */
		cmdlist();	/* display the commands one can perform */
	  }
	standend();	move(3, 0);	printw(
	 "%5.1f%5.1f%5.1f%7.1f%7.1f%7.1f%7.1f%7.1f%7.1f%9.1f ?: ",
	 rp->rh[0], rp->rh[1], rp->rh[2], rp->xto, rp->xtc, 
	 rp->yto, rp->ytc, rp->framem, rp->framc, rp->totinten);
	standend();	refresh();
	if (NULL != rmp->rm_mwp)
#ifdef VMS
				delwin(rmp->rm_mwp);
#else
	  {
		wclear(rmp->rm_mwp);	wrefresh(rmp->rm_mwp);
		delwin(rmp->rm_mwp);
	  }
#endif
	rmp->rm_mwp = newwin(rp->nobk + 3, 79, 4, 0); /* create the window */
	box(rmp->rm_mwp, (xctype)('|'), (xctype)('-'));
	wmove(rmp->rm_mwp, 1, 4); wprintw(rmp->rm_mwp,"frac. inten. frm");
	wstandout(rmp->rm_mwp);
	for (i = 0; i < rp->nobk; i++)
	   {
		fmno = rp->obkmin + i;
		nstars = (int)(50. * fabs(rp->obparts[i]));
		if (rp->obparts[i] < 0)
		  {
			if (nstars > 15) nstars = 15;
			for (j = 0; j < nstars; j++) intenstr[j] = '#';
			intenstr[j] = '\0';
			wmove(rmp->rm_mwp, 2+i, 15 - nstars);
			wprintw(rmp->rm_mwp, "%s %4d%6.3f",
				intenstr, fmno, rp->obparts[i]);
		  }
		else
		  {
			if (nstars > 50) nstars = 50;
			for (j = 0; j < nstars; j++) intenstr[j] = '*';
			intenstr[j] = '\0';
			wmove(rmp->rm_mwp, 2+i, 10);
			wprintw(rmp->rm_mwp, "%6.4f%4d %s",
				rp->obparts[i], fmno, intenstr);
		  }
	   }
	wstandend(rmp->rm_mwp);	wrefresh(rmp->rm_mwp);
	return EOK;
}

static int curse_process(EDITCENRUN *ecp, REFMRUN *rmp, int reason)
{ /* This throws up a CURSES menu to allow the user to check a reflection.
 It returns 1 if the reflection gets deleted, 0 otherwise */
	int		i;
	REFREF		*rp;

	reason = -reason;	rp = (ecp->ec_rfp)->rfr_rpp;
	if (EOK != (i = refl_display(rp, rmp))) return i; /* display the refl*/
	if (ASK_DELETION == (i = user_answer(ecp, rmp))) /* make query */
	  {
		log_ecendel(ecp->ec_fpl, rp, reason); /* log deletion */
		(ecp->ec_nrdel)++;	return 1;
	  }
	else if (END_DELETIONS == i) return i;
	else {
		if (EOK != (i = p_centroid(ecp))) return i;
		return EOK;
	     }
}

int edcenmain(int argc, char *argv[])
{
	int		i, j;
	REFMRUN		rms;
	REFINERUN	*rfp;
	EDITCENRUN	*ecp;

		/* decode arguments for files and flags */
	if (NULL == (ecp = manageecp(3))) return EALLOC;
	if (EOK != (i = edcen_setup(argc, argv, ecp))) return i;
	rfp = ecp->ec_rfp;
		/* loop through the data, processing each observation */
	for (rfp->rfr_rpp = rfp->rfr_rpp0, j = 0;
	 j < ecp->ec_nrin; j++, (rfp->rfr_rpp)++)
	   {
		if (EWRITERR == (i = refl_edit(ecp))) break;
		if (0 > i) i = curse_process(ecp, &rms, i);
		if ((i) && (1 != i))
		  {
			if (i == END_DELETIONS) i = EOK;
			break;
		  }
		if (ecp->ec_flags & EC_DONE)
		  {
			i = EOK;	break;
		  }
	   }
	if (NULL != ecp) smoothend();	if (1 == i) i = EOK;	return i;
}

/* end of editcen.c */
