/* ewmask.c -- X-GEN's CURSES-based model profile editor

	Copyright (c) 2003, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions; reorganize use of "{"
	22 Jun 2006: include xgmodel.h because that's where the structures are
	 2 Jun 2006: norm99 moved here, because it's the only place it's used
	 3 Oct 2005: housekeeping and extra comments
	22 Jun 2005: in proc_char(): "command" -> "comd" to reduce confusion
	11 May 2005: check return status from xgjournal
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	28 Aug 2004: routines used only here are declared as static; cosmetics
	20 Aug 2004: static globals replaced by elements of EWMASKRUN structure
	18 Aug 2004: eliminate static global currdisp
	25 Feb 2004: flush stdout before exiting: helps if we're calling
		ewmask from the outside world
	24 Jan 2004: fpwi and fpwo made static local
	21 Jan 2004: license message inserted
	 3 Jan 2004: replace external globals with static globals
		and use of progname-name conventions found elsewhere;
		replace shcalloc() & shfree() calls with callop() & memzap()
	13 Dec 2003: eliminate unneeded vbls xe,ye,menu_y in zero_areas();
		lz2 in curspread(); menu_y in copy_mask();
		mz2, mxy, mxyz in ewmmain()
	19 Oct 2003: merge ewmain.c,ewellipse.c,ewcmd.c together
 Changes to ewmain.c, ewellipse.c, ewcmd.c prior to that date:
	19 Sep 2003: more careful initialization of ix in start_of_frame();
		default cutoff for ellipsoid fitting is 0.8% of peak, not 2%;
		explicit default in func_ellipsoids()
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	 7 Mar 2003: corrected typecast for currdisp
	 5 Jan 2003: reorganized source-file header
	15 Nov 2002: extern declarations made here rather than in xgenext.h
	12 Nov 2002: contents of ewdisp.c incorporated into this source file
	 1 Nov 2002: greater flexibility: m3_* components of REGIONPROFILE
		structures rather than fixed sizes
	25 Mar 2002: removed MSI attachments
	 5 Oct 1994: xgread and xgwrite on MASK_3D structures
	 5 May 1994: source-file headers cleaned up, POLYDET->XGEN;
		more function-type declarations; wmptovec() redeclared
	16 Mar 1994: significant rewrite: Curses redone;
		functionality moved to ../xglib/xgenmask
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version

   Purpose: an interactive editor for three-dimensional model profiles

   Cross references: ../xglib/xgenmask.c

*************************************************************************/
#define		NEED_CURSES	1
#include	"xgmodel.h"
#include	"menu.h"
#define	ESCAPE	033
#define		LOW_VALUES	1
#define		NONCONTIGUOUS	2
#define		BORDER		4
#define		ENTIRE_FRAME	010
#ifdef CTRL
#undef CTRL
#endif
#define		CTRL(x)		(x & 077)

 typedef struct {
		int		ew_curfram, ew_framex, ew_framey;
		MASK_3D		*ew_mask;
		REGIONPROFILE	*ew_regb, *ew_rgp;
		char		ew_shifted;
		} EWMASKRUN;

static sichar distonum(char dc, EWMASKRUN *ewp)
{ /*This converts a printable representation of an intensity
	(0-9, a-z, A-Z, *) into a one-byte integer between 0 and 99.
	We lose some accuracy in the dots and stars, but ordinarily
	that won't matter unless the user is foolhardy enough to
	enter either a dot or a star by hand. */
	sichar	numval;

	if (dc == ' ' || dc == '.') numval = 0;
	else if ('0' <= dc && dc <= '9') numval = dc - '0';
	else if ('a' <= dc && dc <= 'z') numval = (dc - 'a') + 10;
	else if ('A' <= dc && dc <= 'Z') numval = (dc - 'A') + 36;
	else numval = (ewp->ew_shifted) ? 62 : 81;
	if (ewp->ew_shifted)	/* if we're in ewp->ew_shifted mode, add 38 */
	  {
		if (numval) numval += 38; /* ... but not if it's a blank */
		else if (dc == '.') numval = 19; /* dots are weird */
	  }
	return numval;
}

static char start_of_frame(EWMASKRUN *ewp)
{ /*	This positions the cursor and the index values at the first significant
	value in the current frame.  It returns the character contained in
	that element if it's non-blank; it returns '\0' if there are no
	non-blank characters in the current frame. */
	register int	iy, ix;
	int		foundit, lx, ly, lxy;
	char		currchar, *csp;

	lx = (ewp->ew_rgp)->m3_bx;	ly = (ewp->ew_rgp)->m3_by;
	lxy = lx * ly;
	csp = (char *)((ewp->ew_rgp)->m3_xyoip + lxy * ewp->ew_curfram);
	ix = lx / 2;	currchar = '\0';
	for (iy = foundit = 0; iy < ly; iy++) /* find nonblank val */
	   {
		for (ix = 0; ix < lx; ix++, csp++)
		   if ((currchar = *csp) != ' ')
		     {
			foundit = currchar;	break;
		     }
		if (foundit) break;
	   }
	if (foundit)
	  {
		ewp->ew_framex = ix;	ewp->ew_framey = iy;
	  }
	else {
		ewp->ew_framex = lx / 2; ewp->ew_framey = ly / 2;
		currchar = '\0'; /* no nonblank: stick in middle */
	     }
	return currchar;
}

static int rgptodisp(EWMASKRUN *ewp)
{ /*	This converts from floating-point format to a display format.
	The routine returns 0 unless the input mask is blank,
	in which case it returns 1.  */
	char		*csp;
	sichar		cval, nval, dval;
	int		fmno, ix, iy, lx, ly, lz, offset, retval;
	float		*flp;

	offset = (ewp->ew_shifted) ? 38 : 0;
	retval = 1;	/* flag: no nonzero values found */
	flp = (ewp->ew_rgp)->m3_xyop;	lx = (ewp->ew_rgp)->m3_bx;
	ly = (ewp->ew_rgp)->m3_by;	lz = (ewp->ew_rgp)->m3_bz;
	csp = (char *)(ewp->ew_rgp)->m3_xyoip;
	for (fmno = 0; fmno < lz; fmno++)
	   for (iy = 0; iy < ly; iy++)
		for (ix = 0; ix < lx; ix++, flp++, csp++)
		   {
			cval = *flp;	if (cval > 0) retval = 0;
			nval = cval - offset;
			if (nval < 0) dval = (cval > 0 ? '.' : ' ');
			else if (nval == 0)
				dval = (ewp->ew_shifted ? '0' : ' ');
			else if (nval < 10) dval = '0' + nval;
			else if (nval < 36) dval = 'a' + (nval - 10);
			else if (nval < 62) dval = 'A' + (nval - 36);
			else dval = '*';
			*csp = dval;
		   }
	return retval;
}

static void draw_menu(EWMASKRUN *ewp)
{
	register int	fmno, rowno, colno;
	int		lx, ly, lz, lxy, masknum;
	char		*csp, *cspy, *cspz;

	masknum = (ewp->ew_rgp)->m3_imodel;
	lx = (ewp->ew_rgp)->m3_bx;	ly = (ewp->ew_rgp)->m3_by;
	lz = (ewp->ew_rgp)->m3_bz;	lxy = lx * ly;
	move(0, 24);
	addstr("XGEN Model-Profile Editing Program: Model Profile number");
	printw(" %2d", masknum);	move(1, 1);
	for (fmno = 0; fmno < lz; fmno++)
	   {
		if (ewp->ew_shifted)	printw("   SHIFTEDFM %1d____", fmno);
		else		printw("   ____Frame %1d____", fmno);
	   }
	for (rowno = 0, cspy = (char *)((ewp->ew_rgp)->m3_xyoip); rowno < ly;
	 rowno++, cspy += lx)
	   {
		printw("\n%2d |", rowno);
		for (fmno = 0, cspz = cspy; fmno < lz; fmno++, cspz += lxy)
		   {
			if (fmno) addstr("|-|");
			for (colno = 0,csp = cspz; colno <lx; colno++,csp++)
			   printw("%c", *csp);
			if (fmno == lz-1) addstr("|");
		   }
	   }
	addstr("\n  ");
	for (fmno = 0; fmno < lz; fmno++)
		addstr(" \\012345678901234/");
	addstr("\nArrow Keys move as expected. ");
	addstr("Type a character (0-9,a-z,A-Z) to set current pixel ");
	if (ewp->ew_shifted)
		addstr("to the corresponding value (38-47,48-73,74-99).\n");
	else	addstr("to the corresponding value ( 0- 9,10-35,36-61).\n"); 
	addstr(
"^H back 1 frame; ^B back 1 mask; ^K contiguity;    ^E exit;");
	addstr(
"          ^G Gaussian boundaries;       ^V Lorentzian boundaries;\n");
	addstr(
"^I next frame;   ^L next mask;   ^X recenter mask; ^W redraw;");
	addstr("        ^D replace mask;        ");
	if (ewp->ew_shifted) addstr("<PF3>,^T toggle to  0-61 mode;\n");
	else addstr("<PF3>,^T toggle to 38-99 mode;\n");
	addstr( "^R undo changes; ^F zero border; ^A zero frame;");
	addstr( "    ^U zero pix<floor; ^P zero frame\n");
	refresh();
}

static void go_to_start(EWMASKRUN *ewp)
{
/* This positions the cursor and the index values at the first significant
	value in the array */
	int	lz, menu_x, menu_y;

	lz = (ewp->ew_rgp)->m3_bz;
	for (ewp->ew_curfram = 0; ewp->ew_curfram < lz; (ewp->ew_curfram)++)
	   if ('\0' != start_of_frame(ewp)) break;
	if (ewp->ew_curfram >= lz) ewp->ew_curfram = lz / 2;
	menu_x = ewp->ew_framex + 4 + 18 * ewp->ew_curfram;
	menu_y = ewp->ew_framey + 2;	move(menu_y, menu_x);	refresh();
}

static void undo_changes(int region, EWMASKRUN *ewp)
{ /* This restores values in a given region to what was there to start with*/
	int		lxyz, nmodels;
	REGIONPROFILE	*rg1, *rgc;

	nmodels = (ewp->ew_regb)->m3_nmodels;
	lxyz = (ewp->ew_regb)->m3_bx *
		(ewp->ew_regb)->m3_by * (ewp->ew_regb)->m3_bz;
	rg1 = (ewp->ew_regb) + nmodels + nmodels + region;
	rgc = (ewp->ew_regb) + region;
	cpymem((void *)rgc->m3_xyop, (void *)rg1->m3_xyop,
	 lxyz * sizeof (float));	ewp->ew_rgp = rgc;
	(void)rgptodisp(ewp);	/* create character-display version */
	draw_menu(ewp);	/* let the user see the new values */
}

static void query(char *qans, char *qprompt)
{ /* This is a (possibly!) system-independent way of asking a one-line
	question and getting a one-line answer after a carriage return.
	It uppercaseifies the reply as it goes. */
	char	tc, *anp;
	int	endq, iny, inx, xpos;

	endq = strlen(qprompt);
	getyx(stdscr, iny, inx);	clrtoeol();
	addstr(qprompt);
	move(iny, inx + endq);	refresh();
	anp = qans;	xpos = inx + endq;
	while (1)
	     {
		tc = getch();
		if (('\0' == tc) || ('\n' == tc) || ('\r' == tc))
			break;
		if ((anp > qans) && ((010 == tc) || (0177 == tc)))
		  {
			anp--;	move(iny, xpos + anp - qans);
			addch((xctype)(' '));	move(iny, xpos + anp - qans);
			refresh();
		  }
		else {
			if ('a' <= tc && tc <= 'z') tc -= 'a' - 'A';
			if (' ' <= tc && tc <= '~')
			  {
				move(iny, xpos + anp - qans);
				addch((xctype)(tc));	refresh();
				*anp++ = tc;	*anp = '\0';
			  }
		     }
	     }
	move(iny, xpos + strlen(anp));	clrtoeol();
	move(iny, xpos);	refresh();
}

static void zero_areas(int region, int reason, EWMASKRUN *ewp)
{ /* This blanks out some pixels--either an entire frame or a smaller area*/
	char		first_char, floor_string[20];
	int		curr_frame, curr_region, frame_min, frame_max;
	int		lx, ly, lz, lxy, mask_max, mask_min, nmodels;
	double		floor_value;

	ewp->ew_rgp = (ewp->ew_regb) + region;
	nmodels = (ewp->ew_rgp)->m3_nmodels;	lx = (ewp->ew_rgp)->m3_bx;
	ly = (ewp->ew_rgp)->m3_by;	lz = (ewp->ew_rgp)->m3_bz;
	lxy = lx * ly;		floor_value = 0.8;
	if (reason == ENTIRE_FRAME) /* user has asked to blank a whole frame */
	  {
		zeromem((void *)((ewp->ew_rgp)->m3_xyop +
			lxy * ewp->ew_curfram), lxy * sizeof (float));
		(void)rgptodisp(ewp);	/* move to start */
		(void)start_of_frame(ewp);	draw_menu(ewp); return;
	  }
	if (reason == LOW_VALUES) /* user wants a floor-type cutoff */
	  {
		move(21, 0);	query(floor_string,
	 " Enter floor value below which all pixels will be set to zero: ");
		floor_value = atof(floor_string);	/* 0-99 */
		if (floor_value < 0 || floor_value > MASKRESCALE)
			floor_value = 0.8;		/* default */
		move(21, 63); printw("%6.1f", floor_value);
	  }
	else if (reason == BORDER)
	  {
		move(21, 0);
		addstr("This will zero pixels along the (X,Y) boundaries.");
		clrtoeol();	refresh();
	  }
	else if (reason == NONCONTIGUOUS)
	  {
		move(21, 0);	addstr("This will zero isolated pixels.");
		clrtoeol();	refresh();
	  }
	while (1)
	  {
		move(22, 0);	query(floor_string,
			"This frame, this mask, or all masks (F, M, A) ? ");
		first_char = floor_string[0];
		if (first_char == 'F' || first_char == 'M'
			|| first_char == 'A') break;
	  }
	curr_region = mask_min = mask_max = region;
	frame_min = frame_max = curr_frame = ewp->ew_curfram;
	switch (first_char)
	 {
		case	'A':	mask_min = 0;	mask_max = nmodels - 1;
		case	'M':	frame_min = 0;	frame_max = lz - 1; break;
		default :	break;
	 }
	if (reason == LOW_VALUES)
	  for (region = mask_min, ewp->ew_rgp = ewp->ew_regb + mask_min;
	   region <= mask_max; region++, (ewp->ew_rgp)++)
	   {
		zeromem((void *)((ewp->ew_rgp)->m3_xyop + frame_min * lxy),
			(frame_max + 1 - frame_min) * lxy * sizeof (float));
	   }
	else if (reason == NONCONTIGUOUS)
	  {
		for (region = mask_min, ewp->ew_rgp = ewp->ew_regb + mask_min;
		 region <= mask_max; region++, (ewp->ew_rgp)++)
			(void)contig(ewp->ew_rgp,
				(first_char == 'F') ? ewp->ew_curfram: -1);
	  }
	else if (reason == BORDER)	/* user wants all border pixels = 0 */
	  {
		for (region = mask_min, ewp->ew_rgp = ewp->ew_regb + mask_min;
		 region <= mask_max; region++, (ewp->ew_rgp)++)
			zero_boundary(ewp->ew_rgp, CLF_ZCBORDER);
	  }
	region = curr_region;	ewp->ew_curfram = curr_frame;
	ewp->ew_rgp = ewp->ew_regb + region;
	(void)rgptodisp(ewp);	(void)start_of_frame(ewp);
	move(21, 0);	clrtoeol();	move(22, 0);	clrtoeol();
	draw_menu(ewp);
}

static void curspread(int maskno, EWMASKRUN *ewp)
{ /*	This examines a model or all the models to see if the profile
	is limited to one image. If it is, a miniature version of the
	profile is copied to the two images adjoining the one that
	indeed has some data in it. */
	char		first_char, floor_string[20];
	int		lz, maskmin, maskmax, nmodels, regi, nup[2];
	double		floorfrac, sums[3];

	lz = (ewp->ew_regb)->m3_bz;	if (lz < 2) return;
	nmodels = (ewp->ew_regb)->m3_nmodels;
	move(21, 0);	addstr("This feature converts a 2-D profile to 3-D\n");
	clrtoeol();	refresh();
	while (1) /* ask user: just the current mask or all of 'em ? */
	  {
		move(22, 0);
		query(floor_string, "This mask or all masks (M, A) ? ");
		first_char = floor_string[0];
		if (first_char == 'M' || first_char == 'A') break;
	  }
	if (first_char == 'M')	/* ... answer is "just this one" */
	  {
		maskmin = maskmax = maskno;	move(20, 0);
	  }
	else {
		maskmin = 0;	maskmax = nmodels - 1;	move(0, 0); /* all! */
	     }
	wclrtobot(stdscr);
	addstr( " Model Inten(2) Inten(3) Inten(4) Added(2) Added(4)\n");
	floorfrac = 0.035;
	for (regi =maskmin, ewp->ew_rgp = ewp->ew_regb + maskmin;
	 regi <= maskmax; regi++, (ewp->ew_rgp)++)
	   {
		(void)dospread(ewp->ew_rgp, 0, sums, nup, floorfrac);
		printw(" %7d %8.2lf %8.2lf %8.2lf %8d %8d\n",
		 regi, sums[0], sums[1], sums[2], nup[0], nup[1]);
	   }
	refresh();	sleep(3);
	if (maskmin != maskmax)
	  {
		move(0, 0);	clrtoeol();
	  }
	regi = maskno;	ewp->ew_rgp = ewp->ew_regb + maskno;
	(void)rgptodisp(ewp);	(void)start_of_frame(ewp);
	move(21, 0);	clrtoeol();	move(22, 0);	clrtoeol();
	draw_menu(ewp);
}

static void copy_mask(int region, EWMASKRUN *ewp)
{ /*	This copies one mask into another.  It prompts the user for the
	input mask and assumes that it's being copied into mask `region'. */
	char		sourcestr[20];
	int		lxyz, source_region;
	REGIONPROFILE	*rgs;

	for (source_region = -1; ; ) /*ask whence the new one should come*/
	   {
		move(21, 0); query(sourcestr,
		 " Enter region (0-11) from which model will be copied: ");
		source_region = atoi(sourcestr);
		if (region != source_region &&
		  (0 <= source_region && source_region < 12)) break;
		printw(
 "\nIllegal value: source must be 0-11 and not the current region");
		refresh();
	   }
		/* ... now copy that one into the current one */
	rgs = ewp->ew_regb + source_region;
	ewp->ew_rgp = ewp->ew_regb + region;
	lxyz = (ewp->ew_rgp)->m3_bx * (ewp->ew_rgp)->m3_by *
		(ewp->ew_rgp)->m3_bz;
	cpymem((void *)((ewp->ew_rgp)->m3_xyop), (void *)rgs->m3_xyop,
	 lxyz * sizeof (float));
	(ewp->ew_rgp)->m3_nobs = rgs->m3_nobs;
	(ewp->ew_rgp)->m3_isqsum = rgs->m3_isqsum;
	(void)rgptodisp(ewp);
	ewp->ew_curfram = 0;	(void)start_of_frame(ewp);	draw_menu(ewp);
}

static void recenter_mask(int maskno, EWMASKRUN *ewp)
{	/* this guarantees that the mean value of the mask is at the
	physical center of the mask */
	char		first_char, floor_string[20];
	register int	z;
	int		lx, ly, lz, lxyz, maskmax, maskmin, nmodels, region;
	double		delv[3], idel[3], sumf;
	static float	*flp;
	REGIONPROFILE	*rop;

	move(21, 0);	addstr("This feature re-centers the model.\n");
	clrtoeol();	refresh();
	while (1) /* ask user: just the current mask or all of 'em ? */
	  {
		move(22, 0);
		query(floor_string, "This mask or all masks (M, A) ? ");
		first_char = floor_string[0];
		if (first_char == 'M' || first_char == 'A') break;
	  }
	lx = (ewp->ew_regb)->m3_bx;	ly = (ewp->ew_regb)->m3_by;
	lz = (ewp->ew_regb)->m3_bz;	lxyz = lx * ly * lz;
	nmodels = (ewp->ew_regb)->m3_nmodels;
	if (first_char == 'M')	/* ... answer is "just this one" */
	  {
		maskmin = maskmax = maskno;	move(20, 0);
	  }
	else {
		maskmin = 0;	maskmax = nmodels - 1;	move(0, 0); /* all! */
	     }
	wclrtobot(stdscr);
	addstr(
	 " Recentering Model # Preshift center:<xc>     <yc>     <zc>");
	addstr("  Postshift:<xc>     <yc>     <zc>\n");
	/* find the centroids and recenter them */
	for (region = maskmin, ewp->ew_rgp = ewp->ew_regb + maskmin,
	 rop = ewp->ew_regb + nmodels + maskmin; region <= maskmax;
	  region++, (ewp->ew_rgp)++, rop++)
	   {
		idel[0] = idel[1] = idel[2] = 0.;
		sumf = gen_centroid((ewp->ew_rgp)->m3_xyop,
			NULL, lx, ly, lz, idel, delv);
		if (sumf > FUZZ)
		  {
			printw(" %19d %20.3f %8.3f %8.3f",
				region, delv[0], delv[1], delv[2]);
			recen_vec((ewp->ew_rgp)->m3_xyop,
				delv, rop->m3_xyop, lx, ly, lz);
			/* renormalize the output */
			for (sumf = z = 0, flp = rop->m3_xyop;
			 z < lxyz; z++, flp++) if (*flp > sumf) sumf = *flp;
			if (sumf > FUZZ)
			  {
				sumf = MASKRESCALE / sumf;
				for (z = 0, flp = rop->m3_xyop; z < lxyz;
				 z++, flp++) *flp *= sumf;
			  }
			sumf = gen_centroid(rop->m3_xyop, NULL, lx, ly, lz,
				idel, delv);
			printw("   %13.3f %8.3f %8.3f\n",
				delv[0], delv[1], delv[2]);
			cpymem((void *)(ewp->ew_rgp)->m3_xyop,
				(void *)rop->m3_xyop, lxyz * sizeof (float));
		  } /* end of "if (sumf > FUZZ)" clause */
		else printw("\n");
	  } /* end of "for (region = mask_min" loop */
	refresh();	sleep(3);
	if (maskmin != maskmax)
	  {
		move(0, 0);	clrtoeol();
	  }
	region = maskno;	ewp->ew_rgp = ewp->ew_regb + maskno;
	(void)rgptodisp(ewp);	(void)start_of_frame(ewp);
	move(21, 0);	clrtoeol();	move(22, 0);	clrtoeol();
	draw_menu(ewp);
}

static void ell_header(FILE *fp)
{ /*	This prints a suitable header on the table of ellipsoid results */
	if (fp == stderr || fp == stdout)
	  {
		printw("Results of Ellipsoidal fitting:\n");
		printw(" Region  # of pixels     rms    Peak  Computed ");
		printw("Centroid             Ellipsoid Parameters\n");
		printw(
		 " number zeroed upped     fit    i000    x0    y0    z0");
		printw("     a00     a01     a02     a11     a12     a22\n");
		refresh();
	  }
	else {
		fprintf(fp, "Results of Ellipsoidal fitting:\n");
	 fprintf(fp, " Region # of pixels     rms    Peak  Computed ");
	 fprintf(fp, "Centroid             Ellipsoid Parameters\n ");
	 fprintf(fp,"number  zeroed upped     fit    i000    x0    y0     z0");
	 fprintf(fp," lambda0 lambda1 lambda2  euler0  euler1  euler2\n");
	     }
}

static void sum_ellipse(FILE *fp, int region, double *ellpars)
{ /*	This summarizes the results of the ellipsoidal fitting
	to the stream fp. */
	register int	i;

	if (fp == stderr || fp == stdout)
	  {
		printw(" %6d %6d %5d %7.3lf %7.4lf %5.1lf %5.1lf %5.1lf",
		 region, (int)ellpars[10], (int)ellpars[12], ellpars[11],
		 ellpars[0], ellpars[7], ellpars[8], ellpars[9]);
		for (i = 1; i < 7; i++) printw(" %7.4lf", ellpars[i]);
		printw("\n");
		refresh();
	  }
	else {
		fprintf(fp, " %6d %6d %5d %7.3f %7.4f %4.1f %4.1f %4.1f",
		 region, (int)ellpars[10], (int)ellpars[12], ellpars[11],
		  ellpars[0], ellpars[7], ellpars[8], ellpars[9]);
		for (i = 1; i < 7; i++) fprintf(fp, " %7.4f", ellpars[i]);
		fprintf(fp, "\n");
	     }
}

static void func_ellipsoids(int region, int ellipsetype, EWMASKRUN *ewp)
{ /*	This sets all the values in a model profile to zero if they lie outside
	a user-set contour level on the best-fit ellipsoid describing the
	current model. */
	char		first_char, action_char;
	char		floor_string[20], action_string[20];
	int		current_region, ellflags, iy, lxyz;
	int		mask_max, mask_min, nmodels;
	float		*flp, *flpe;
	double		floor_value, peak, peake, thesepars[14];
	REGIONPROFILE	*rg0;

	move(20, 0);	wclrtobot(stdscr);	refresh();
	query(floor_string,
	 " Enter cutoff contour value (0.01 -> 99) on ellipsoid: ");
	floor_value = atof(floor_string);
	if (floor_value < 0.01 || floor_value > MASKRESCALE)
		floor_value = 1.5;
	while (1)
	  {
		move(21, 0);
		query(floor_string, "This mask, or all masks (M, A) ? ");
		first_char = floor_string[0];
		if ('a' <= first_char && first_char <= 'z')
			first_char -= ('a' - 'A');
		if (first_char == 'M' || first_char == 'A') break;
	  }
	while (1)
	  {
		move(22, 0);
		query(action_string,
	 "Flatten boundary, extrap edges, both, or replace (F,E,B,R) ? ");
		action_char = action_string[0];
		if ('a' <= action_char && action_char <= 'z')
			action_char -= ('a' - 'A');
		if (action_char == 'F' || action_char == 'E' ||
			action_char == 'B' || action_char == 'R') break;
	  }
	ellflags = CLF_FLATTEN | CLF_EXTRAP;	/* default is to do both */
	if (action_char == 'F')		ellflags = CLF_FLATTEN;
	else if (action_char == 'E')	ellflags = CLF_EXTRAP;
	else if (action_char == 'B')	ellflags = CLF_FLATTEN | CLF_EXTRAP;
	else if (action_char == 'R')	ellflags = CLF_OUTELLIPSE;
	current_region = region;
	nmodels = (ewp->ew_regb)->m3_nmodels;
	if (first_char == 'M')
	  {
		mask_min = mask_max = region;
		for (iy = 20; iy < 23; iy++)
		   {
			move(iy, 0); clrtoeol();
		   }
		move(19, 0);
	  }
	else {
		mask_min = 0;	mask_max = nmodels - 1;	clear();
	     }
	refresh();		ell_header(stdout);
	floor_value *= 0.01;	/* convert from percent to fraction */
	for (region = mask_min, ewp->ew_rgp = ewp->ew_regb + mask_min;
	 region <= mask_max; region++, (ewp->ew_rgp)++)
	   {
		if (EOK != ellsoid(ewp->ew_rgp, ellipsetype, ellflags,
			floor_value, 4.,  thesepars))
		  {
			thesepars[0] = 1.;
			for (iy = 1; iy < 13; iy++) thesepars[iy] = 0;
		  }
		else	/* normalize back to peak == 99 */
		  {
			rg0 = ewp->ew_rgp + 2 * nmodels;
			lxyz = (ewp->ew_rgp)->m3_bx *
				(ewp->ew_rgp)->m3_by * (ewp->ew_rgp)->m3_bz;
			for (peake = peak = iy = 0,
			 flp = (ewp->ew_rgp)->m3_xyop,
			 flpe = rg0->m3_xyop; iy < lxyz; iy++, flp++, flpe++)
			   {
				if (*flp > peak) peak = *flp;
				if (*flpe > peake) peake = *flpe;
			   }
			if (peak > FUZZ)
			  {
				peak = 99. / peak;
				for (iy = 0, flp = (ewp->ew_rgp)->m3_xyop;
				 iy < lxyz; iy++, flp++) *flp *= peak;
			  }
			if (peake > FUZZ)
			  {
				peake = 99. / peake;
				for (iy = 0, flpe = rg0->m3_xyop;
				 iy < lxyz; iy++, flpe++) *flpe *= peake;
			  }
		  }
		sum_ellipse(stdout, region, thesepars);
	   }
	if (first_char == 'M')
	  {
		sleep(2);
		for (iy = 20; iy < 23; iy++)
		   {
			move(iy, 0); clrtoeol();
		   }
	  }
	else {
		clear();	refresh();
	     }
	region = current_region;	ewp->ew_rgp = ewp->ew_regb + region;
	iy = rgptodisp(ewp); draw_menu(ewp);
}

static void proc_char(sichar comd, int region, EWMASKRUN *ewp)
{ /* This processes a user command. */
	char		currchar, *csp;
	int		lx, lxy, ly, lz, menu_x, menu_y, off;
	float		*flp;

	lx = (ewp->ew_rgp)->m3_bx;	ly = (ewp->ew_rgp)->m3_by;
	lz = (ewp->ew_rgp)->m3_bz;	lxy = lx * ly;
	off = ewp->ew_curfram * lxy + ewp->ew_framey * lx + ewp->ew_framex;
	csp = (char *)((ewp->ew_rgp)->m3_xyoip + off);
	flp = (ewp->ew_rgp)->m3_xyop + off;
	switch (comd)
	  {
		case	'\n': case	'\r': ewp->ew_framey =
			 (ewp->ew_framey < ly - 1) ?  ewp->ew_framey + 1 : 0;
			ewp->ew_framex = 0;	break;
		case	CTRL('A') :	case	CTRL('P') :
			zero_areas(region, ENTIRE_FRAME, ewp); break;
		case	CTRL('D') : copy_mask(region, ewp);	break;
		case	CTRL('E') : return;
		case	CTRL('F') : zero_areas(region, BORDER, ewp);	break;
		case	CTRL('G') : func_ellipsoids(region, GAUSSIAN, ewp);
					break;
		case	CTRL('H') : if (ewp->ew_curfram) (ewp->ew_curfram)--;
			menu_y = start_of_frame(ewp);		break;
		case	CTRL('I') :
			if (ewp->ew_curfram < lz - 1) (ewp->ew_curfram)++;
			menu_y = start_of_frame(ewp);		break;
		case	CTRL('K') : zero_areas(region, NONCONTIGUOUS, ewp);
			break;
		case	CTRL('L') : return;
		case	CTRL('O') : curspread(region, ewp);	break;
		case	CTRL('R') : undo_changes(region, ewp); break;
		case	CTRL('U') :
			zero_areas(region, LOW_VALUES, ewp);	break;
		case	CTRL('V') :
			func_ellipsoids(region, LORENZIAN, ewp); break;
		case	CTRL('W') :
			clear(); refresh(); draw_menu(ewp);	break;
		case	CTRL('X') :	case	K_PF4:
			recenter_mask(region, ewp);	break;
		case	K_LEFT :	if (ewp->ew_framex) (ewp->ew_framex)--;
			else ewp->ew_framex = lx-1; break;
		case	K_RIGHT : if (ewp->ew_framex < lx-1) (ewp->ew_framex)++;
			else ewp->ew_framex = 0;	break;
		case	K_UP : ewp->ew_framey = ewp->ew_framey ?
				ewp->ew_framey - 1 : ly - 1; break;
		case	K_DOWN : ewp->ew_framey = (ewp->ew_framey < ly - 1) ?
			ewp->ew_framey + 1 : 0;	break;
		case	K_PF3 :	case	CTRL('T') :
				/* toggle ewp->ew_shifted mode */
			ewp->ew_shifted = 1 - ewp->ew_shifted;
			(void)rgptodisp(ewp);	draw_menu(ewp);	break;
		case	0177:
			if (ewp->ew_framex)
			  {
				(ewp->ew_framex)--;	csp--;
			  }
			*csp = ' ';	break;
		default : currchar = *csp;
			if ((!ewp->ew_shifted) &&
			 (comd == '0' || comd == '.')) comd = ' ';
			if (('A' <= comd && comd <= 'Z') ||
				('a' <= comd && comd <= 'z') ||
				('0' <= comd && comd <= '9') ||
				('.' == comd) || ' ' == comd)
			    {
				*csp = comd;
				if (comd != currchar)
					*flp = distonum((char)comd, ewp);
			    }
			  menu_x = ewp->ew_framex + 4 + 18 * ewp->ew_curfram;
			  menu_y = ewp->ew_framey + 2;	move(menu_y, menu_x);
			  addch((xctype)(*csp));
			  if (ewp->ew_framex < lx) (ewp->ew_framex)++;
			  else {
				ewp->ew_framex = 0;
				if ((ewp->ew_curfram)< lz) (ewp->ew_curfram)++;
			       }
	   }
	menu_x = ewp->ew_framex + 4 + 18 * ewp->ew_curfram;
	menu_y = ewp->ew_framey + 2;
	move(menu_y, menu_x);
	refresh();
}

static int endew(int irv, EWMASKRUN *ewp)
{	/* This exits gracefully from `ewmask' */
	REGIONPROFILE	*lrb;

	setmzap();	/* kill allocations done in `setmore' */
	memzap((AVOID **)&(ewp->ew_mask), "ewp->ew_mask in ewmain");
	if (NULL != (lrb = ewp->ew_regb))
	  {
		memzap((AVOID **)&(lrb->m3_xyop), "lrb->m3_xyop in ewmain");
		memzap((AVOID **)&(lrb->m3_xyoip), "lrb->m3_xyoip in ewmain");
		memzap((AVOID **)&lrb, "lrb in ewmain");
		ewp->ew_regb = NULL;
	  }
	(void)xgprogname(NULL, 1);	(void)xeflush(stdout);	return irv;
}

static void norm99(REGIONPROFILE *rgp)
{ /* This renormalizes *rgp so that its peak is 99. */
	int	iz, lxyz;
	float	maxmax, *flp;

	lxyz = rgp->m3_bx * rgp->m3_by * rgp->m3_bz;
	for (maxmax = iz = 0, flp = rgp->m3_xyop; iz < lxyz; iz++, flp++)
		if (*flp > maxmax) maxmax = *flp;
	if (maxmax < FUZZ) return;	/* nothing to normalize */
	maxmax = 99. / maxmax;
	for (iz = 0, flp = rgp->m3_xyop; iz < lxyz; iz++, flp++)
		*flp *= maxmax;
}

/*ARGSUSED*/
int ewmmain(int argc, char *argv[])
{ /*	This is the mainline for "ewmask".
	It deals with each editable three-dimensional model profile
	independent of the other eleven except when certain commands
	are invoked, viz. ^B, ^G.  */
	char			mline[5];
	sichar			c;
	int			jreg, lx, lxy, lxyz, ly, lz, lz2, mx, my, mz;
	int			nmodels, region, result, region_occ[NREGIONS];
	MASK_3D			*wmp;
	float			*flp, *flp1;
	REGIONPROFILE		*rg0, *rg1, *rgp;
	static FILE		*fpwi, *fpwo;
	static EWMASKRUN	ewr;
	static char		*fnames[] = {"wmask", NULL };
	static char		ftypes[] =  { 3, 0 };

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	if (qerror((AVOID *)(fpwi = xgopen("wmask", "r")),
		"open input model-profiles file")) return EOPENERR;
	header(stdout, PROCESSING, fnames, ftypes);	/* summarize inputs */
#ifndef VMS
	if (verror(setmore(0),
	 "set up keystroke equivalences for this terminal"))
		return endew(EOPENERR, &ewr);
#endif
	sleep(2);				/* so we can see the header */
	/* establish operational variables and reserve some memory */
	lx = BIGSPOTWID;	ly = BIGSPOTHT;	lz = PFRAMES;
	nmodels = NREGIONS;	lz2 = lz / 2;
	lxy = lx * ly;		lxyz = lxy * lz;
	mx = 13;	my = 13;	mz = 9;
	if (NULL == (AVOID *)(ewr.ew_mask = (MASK_3D *)callop(nmodels,
	 sizeof (MASK_3D),"ewr.ew_mask in ewmain"))) return endew(EALLOC,&ewr);
	if (NULL == (ewr.ew_regb = (REGIONPROFILE *)callop(nmodels * 3,
	 sizeof (REGIONPROFILE),"ewr.ew_regb in cleanmask"))) return EALLOC;
	if (NULL == (flp = (float *)callop(nmodels * 3 * lxyz,
	 sizeof (float), "flp in ewmask"))) return EALLOC;
	if (NULL == ((ewr.ew_regb)->m3_xyoip = (sichar *)mallop(lxyz,
	 "(ewr.ew_regb)->m3_xyoip in ewmask"))) return EALLOC;
/* read in the twelve regional model profiles and normalize them. */
	for (region = 0, wmp = ewr.ew_mask, ewr.ew_rgp = ewr.ew_regb;
	 region < nmodels; region++, wmp++, flp += lxyz, (ewr.ew_rgp)++)
	   {
		if (0 >= xgread((void *)wmp, XG_MASK_3D, fpwi))
		 return endew(EREADERR, &ewr);
		if ((EOK != (result = wmptovec(wmp, flp))) &&
		 (EBADVALUE != result)) return endew(result, &ewr);
		rgp = ewr.ew_rgp;
		rgp->m3_bx = lx;	rgp->m3_by = ly;
		rgp->m3_bz = lz;	rgp->m3_qx = mx;
		rgp->m3_qy = my;	rgp->m3_qz = mz;
		rgp->m3_imodel = region;
		rgp->m3_nmodels = nmodels;
		rgp->m3_nobs = wmp->phielt[lz2].mnobs;
		rgp->m3_isqsum = wmp->phielt[lz2].isqsum;
		if (region) rgp->m3_xyoip = (ewr.ew_regb)->m3_xyoip;
		rgp->m3_xyop = flp;	norm99(rgp);
	   }
	(void)fclose(fpwi);	fpwi = NULL;	/* no more inputs! */
	initscr();	crmode();	nonl();	noecho(); /* set up "curses" */
	/* for any blank profile, determine a weighted average of the others.
	 In all cases, copy the profile into a saved-profile buffers */
	rg0 = ewr.ew_regb + nmodels;	rg1 = ewr.ew_regb + nmodels + nmodels;
	flp1 = flp + nmodels * lxyz;
	for (region = 0, ewr.ew_rgp = ewr.ew_regb; region < nmodels;
	 region++, (ewr.ew_rgp)++, rg0++, flp += lxyz, rg1++, flp1 += lxyz)
	   {
		if (blankf((ewr.ew_rgp)->m3_xyop, lz))
			meanprofile(ewr.ew_regb, region);
		rg0->m3_bx = lx; rg0->m3_by = ly; rg0->m3_bz = lz;
		rg1->m3_bx = lx; rg1->m3_by = ly; rg1->m3_bz = lz;
		rg0->m3_nobs = (ewr.ew_rgp)->m3_nobs;
		rg0->m3_isqsum = (ewr.ew_rgp)->m3_isqsum;
		rg1->m3_nobs = (ewr.ew_rgp)->m3_nobs;
		rg1->m3_isqsum = (ewr.ew_rgp)->m3_isqsum;
		rg0->m3_xyop = flp;		rg1->m3_xyop = flp1;
		cpymem((void *)rg0->m3_xyop, (void *)(ewr.ew_rgp)->m3_xyop,
			lxyz * sizeof (float));
		cpymem((void *)rg1->m3_xyop, (void *)(ewr.ew_rgp)->m3_xyop,
			lxyz * sizeof (float));
	   }
/* now let the user do his thing. */
	for (region = 0; region < nmodels; )	/* go through the regions */
	   {
		ewr.ew_rgp = ewr.ew_regb + region;
	     if (!rgptodisp(&ewr))	/* format mask for display */
	       {
		draw_menu(&ewr);	/* display it */
		go_to_start(&ewr);		/* find starting position */
		/* accept character input from the terminal.  When the
		user types ^B, ^E, or ^L we drop out of this loop */
		while ('\0' != (c = wget1ch(stdscr)) &&
		  CTRL('B') != c && CTRL('E') != c && CTRL('L') != c)
		  {
#ifdef VMS
			if (K_KP0 <= c && c <= K_PERIOD) c = ordinary_key(c);
#else
			if (ESCAPE == c) c = getescseq(stdscr);
#endif
			proc_char(c, region, &ewr);
		  } /* close on while loop over characters */
		/* we're done (one way or another) with this region */
		if (c == CTRL('B')) /* ^B: go back one region */
			region = (region ? region - 1: nmodels - 1);
		else if (c == CTRL('E')) break; /* ^E: exit */
		else region++;		/* ^L: go forward one region */
	     } /* close of "if (!m992_disp" clause */
	    else /* no valid refls in this region: borrow others!! */
	     {
		draw_menu(&ewr);	/* display it */
		move(20, 0);	clrtoeol();	addstr(
		 "This profile is blank.  Nonblank profiles exist in regions");
		for (jreg = 0; jreg < nmodels; jreg++)
		   if ((jreg != region) && (region_occ[jreg]))
			printw(" %2d ", jreg);
		copy_mask(region, &ewr);
	     } /* close of "else" to "if (!rgptodisp" clause */
	  } /* close of loop over regions */
	clear(); refresh();	endwin();	/* get rid of window stuff */
	fprintf(stderr,
		"\tDo you want to write the revised model profiles out? ");
	while (NULL == fgets(mline, 4, stdin)) ;
	if (mline[0] == 'y' || mline[0] == 'Y')
	  {
		printf( "\t\tWriting model profiles out to %s\n",
			trname("wmask"));
		if (qerror((AVOID *)(fpwo = xgopen("wmask", "w")),
		 "open output model-profiles file"))
			return endew(EOPENERR, &ewr);
		for (region = 0, wmp = ewr.ew_mask, ewr.ew_rgp = ewr.ew_regb;
		  region < nmodels; region++, wmp++, (ewr.ew_rgp)++)
		   {
			(void)vectowmp((ewr.ew_rgp)->m3_xyop, wmp,
			 (ewr.ew_rgp)->m3_nobs, (ewr.ew_rgp)->m3_isqsum, lz);
			if (0 >= xgwrite((void *)wmp, XG_MASK_3D, fpwo))
				return endew(EWRITERR, &ewr);
		   }
		(void)fclose(fpwo);	fpwo = NULL;
	  }
	else fprintf(stderr, "Okay; no output models being written\n");
	return endew(EOK, &ewr);
}

/* end of ewmain.c */
