/* menu.c -- routines for dealing with CURSES menus

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

 Several routines in this package are specific either to the
 VMS implementation or the non-VMS implementation.  There are a few
 MIPS-specific chunks dealing with the fact that a "char" is by
 default unsigned on a MIPS and signed for (nearly) everyone else.
 It's ugly, but not a big deal.

Change record:
	19 Apr 2007: reorganize use of "{"
	15 Jan 2006: calls to actual_char corrected
	 3 Oct 2005: housekeeping
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	19 Oct 2004: add ^F as a "HELP" signal (^Y is often INTR)
	27 Aug 2004: routines used only here are declared as static; cosmetics
	23 Mar 2004: args to isupper(), etc. are now ints, not chars
	31 Jan 2004: memory allocations are made conditional on the variable
		being NULL
	30 Jan 2004: rename variable xstart to xvstar to eliminate confusion
	21 Jan 2004: license message inserted
	15 Nov 2003: initialize a few variables (ltank, wp, oldmenu)
		to keep -Wall happy
	20 Sep 2003: menu_setup() moved earlier in the file so its type
		is explicitly known within getans()
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	 9 Jul 2003: embolden program names more carefully:
		don't embolden if the application isn't really available.
	 5 Jan 2003: reorganized source-file header
	10 Nov 2002: cosmetic changes
	25 Mar 2002: removed MSI attachments
	15 May 1999: in getans(), control-X is now "eXecute" (PF1);
		control-R is now "pRevious" (like control-X used to be).
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version
	15 Nov 1988: added generalized menu-handling routines--
		menu_link and all routines below it.
	xx xxx 1986: initial version

   Purpose: utilities used with the "curses" user-interface library.

   Cross references: ../include/menu.h

*************************************************************************/

#define		NEED_CURSES	1
#define		NEED_CTYPE	1
#include	"xgen.h"
#include	"menu.h"
#define		SEQLENGTH	80
#define		TSYMBOLWID	32

extern char	*tgetstr(char *id, char **area);

 typedef struct {
		char	*xm_answer, *xm_blanks, *xm_symb, **xm_escseq;
		size_t	xm_maxlength;
		} XGCMENU;

static XGCMENU *managexcp(int opno)
{ /*	This handles memory for the menus */
	int		par;
	char		*cp;
	static XGCMENU	*xcp;
	static char	xcpmsg[] = "xcp in managexcp";
	static char	escmsg[] = "xcp->xm_escseq in managexcp";
	static char	symmsg[] = "xcp->xm_symb in managexcp";
	static char	blamsg[] = "xcp->xm_blanks in managexcp";
	static char	ansmsg[] = "xcp->xm_answer in managexcp";
	static size_t	sizb[] = {
	 NSYMBOLS * 2, NSYMBOLS * TSYMBOLWID * 2, 90, 80 };

	if (!opno) /* eliminate memory for these strings */
	  {
		if (NULL != xcp)
		  {
			memzap((AVOID **)&(xcp->xm_escseq), escmsg);
			memzap((AVOID **)&(xcp->xm_symb), symmsg);
			memzap((AVOID **)&(xcp->xm_blanks), blamsg);
			memzap((AVOID **)&(xcp->xm_answer), ansmsg);
			memzap((AVOID **)&xcp, xcpmsg);
		  }
		return NULL;
	  }
	if (!(opno & 01))	/* even numbers: kill specific strings */
	  {
		if (NULL == xcp) return NULL;
		if (opno & 04) memzap((AVOID **)&(xcp->xm_escseq), escmsg);
		if (opno & 010) memzap((AVOID **)&(xcp->xm_symb), symmsg);
		if (opno & 020) memzap((AVOID **)&(xcp->xm_blanks), blamsg);
		if (opno & 040) memzap((AVOID **)&(xcp->xm_answer), ansmsg);
		return xcp;
	  }
	if (opno == 1) xcp = managexcp(0); /* force a new one */
	if (NULL == xcp) xcp = (XGCMENU *)callop(1, sizeof (XGCMENU), xcpmsg);
	if (opno & 044)	/* neither 01 nor 03 */
	  {
		if (NULL == xcp->xm_escseq)
		  {
			if (NULL == (xcp->xm_escseq = (char **)callop(sizb[0],
			 sizeof (char *), escmsg))) return managexcp(0);
		  }
		if (NULL == xcp->xm_symb)
		  {
			if (NULL == (xcp->xm_symb = (char *)mallop(
			 sizb[1], symmsg))) return managexcp(0);
			zeromem((AVOID *)(xcp->xm_symb), sizb[1]);
		  }
		if (NULL == xcp->xm_blanks)
		  {
			if (NULL == (xcp->xm_blanks = (char *)mallop(
			 sizb[2], symmsg))) return managexcp(0);
			*(xcp->xm_blanks + 89) = '\0';
			for (par = 89, cp = xcp->xm_blanks; par--; cp++)
				*cp = ' ';
		  }
		if (NULL == xcp->xm_answer)
		  {
			if (NULL == (xcp->xm_answer= (char *)mallop(
			 sizb[3], symmsg))) return managexcp(0);
			zeromem((AVOID *)(xcp->xm_answer), sizb[3]);
		  }
	  }
	return xcp;
}

static void twoin(char *cp, char c0, char c1)
{ /* This inserts c0 into *cp and c1 into *(cp+1) */
	*cp = c0;	*(cp+1) = c1;
}

void setmzap(void)
{ /* This frees memory that is allocated in `setmore' */
	(void)managexcp(0);
}

static char *careful_print(sichar *str)
{ /*	This returns a string containing 2 characters for each input character.
	For each character in the input it puts a pair of characters in the
	output in the following way:
		range of		1st output	2nd output
		input char		character	character
		0 <= c <= 037		^		(c | 0100)
		040 <= c < 0176		(blank)		c
		0177			^		?
		0200 <= c <= 0276	~		((c & 077) | 0100)
		0277 <= c <= 0376	_		((c & 077) | 0100)
		0377			_		?
	The string is then null-terminated.  */
	sichar		*sc;
	char		*cpr;
	unsigned char	c;
	static char	*lbuf;

	if (NULL == lbuf) lbuf = (char *)mallop(100, "lbuf in careful_print");
	lbuf[2 * strlen((char *)str)] = '\0';
	for (sc = str, cpr = lbuf; '\0' != *sc; sc++)
	   {
		if (*sc > 0) c = (char)*sc;	else c = (0377 + *sc) + 1;
		if (' ' == c || '\t' == c || (040 <= c && c <= 0177))
			twoin(cpr, ' ', c);
		else if (c <= 037) twoin(cpr, '^', c | 0100);
		else if (0177 == c) twoin(cpr, '^', '?');
		else if (0200 < c && c <= 0276)
			twoin(cpr, '~', (c & 077) | 0100);
		else if (0377 == c) twoin(cpr, '_', '?');
		else twoin(cpr, '_', (c & 077) | 0100);
		cpr += 2;
	   }
	return lbuf;
}

#ifdef VMS

char wget1ch(WINDOW *wp)
{
/* this returns one character at a time to the terminal.
 In the VMS version we check the output of smg$read_keystroke to figure
 out what to do.  We pass values between 0 and 127 unchallenged.
 Values between 128 and 255 are ill-defined, as far as I'm concerned.
 Values between 257 and 384 (aha! 3 * 128 !) have 384 subtracted from them
 before they're returned, so they come back between -128 and 0.  Values between
 384 and 511 are always errors (see SYS$LIBRARY:SMGDEF.H) so they're returned
 as 0.  I use 0 to signify end of input, so that makes sense.
*/
	unsigned 	smg_return, yposn, xposn;
	unsigned short	smg_code;

	yposn = wp->_cur_y + wp->_beg_y + 1;
	xposn = wp->_cur_x + wp->_beg_x + 1;
	smg_return = smg$set_physical_cursor(&(stdpb->_id), &yposn, &xposn);
	if (SS$_NORMAL != smg_return) return '\0';
	smg_return = smg$read_keystroke(&(stdkb->_id), &smg_code);
	if (SS$_NORMAL != smg_return || smg_code >= 384) return '\0';
	if (smg_code >= 256) return (char)(smg_code - 384);
	else return (char)smg_code;
}

int winsline(WINDOW *wp)
{	/* this replaces `winsertln' in the VMS version, since
		`winsertln' doesn't work. */
	unsigned int	dispid, direction, retval;
	int		line_number;

	line_number = wp->_cur_y + 1;	dispid = wp->_id;
	direction = SMG$M_DOWN;
	if (SS$_NORMAL != smg$insert_line(
		&dispid, &line_number, 0, &direction)) return ERR;
	return OK;
}
	
#else
	/* non-VMS routines */
char wget1ch(WINDOW *wp)
{
	return wgetch(wp);
}

int winsline(WINDOW *wp)
{
	winsertln(wp);	return OK;
}


int setmore(int diagnostic)
{ /* This defines the escape sequences using the termcap or terminfo entry.
	It prints the results of the user's call if `diagnostic' is 1.  */
	int		i;
	char		*asp, *bp, *name, *result, *where;
	char		**adp, **dcp, **escseq;
	unsigned	esclen, maxlength;
	XGCMENU		*xcp;
	static char	*codes[NSYMBOLS] =
		{"k0","k1","k2","k3","k4","k5","k6","k7","k8","k9",
		 "K0","K1","K2","K3","K4","K5","K6","K7","K8","K9",
		 "kb","ke","kh","ks","md","me","kd","kl","kr","ku",
		 "bc","vd","vl","vr","vu","vb" } ;
#ifdef sun
	int	hostlen;
	char	*hostdispl, *local, *thishost;
#endif

	/* get the name of the terminal and save it. */
	if ((NULL == (result = getenv("TERM"))) || ('\0' == *result)) return 0;
	/* special case on a Sun workstation: if we're on the Sun's own
	console and it reads out as an "xterm", we substitute the escape
	sequences for the Sun itself.  Ugly, but apparently necessary */
#ifdef sun
	thishost = getenv("HOSTNAME");
	if ((NULL == thishost) || ('\0' == *thishost))
	  {
		hostdispl = "";	hostlen = 1;
	  }
	else {
		hostlen = strlen(thishost) + 2;
		hostdispl = mallop(hostlen + 2, "hostdispl in setmore");
		strcpy(hostdispl, thishost);	strcat(hostdispl, ":0");
	     }
	if (((!strcmp(result, "xterm")) || (!strcmp(result, "XTERM"))) &&
	 (NULL == (local = getenv("DISPLAY")) || ('\0' == *local) ||
	 (!strncmp(local, "unix:0", 6)) || (*local == '0') ||
	 (!strncmp(local, hostdispl, hostlen)))) result = "sun";
#endif
	if (NULL == (name = (char *)mallop(strlen(result) + 2,
	 "name in setmore"))) return 0;
#ifdef linux
/* This may have more general applicability, but I know I want this here*/
	if (!strcmp(result, "xterm")) strcpy(name, "vt220");
	else strcpy(name, result);
#else
	strcpy(name,result);
#endif
	/* save some memory for the `tgetent' operation. */
	if (NULL == (bp = (char *)mallop(1024, "bp in setmore"))) return 0;
	/* store up the terminfo or termcap information */
	if (1 != tgetent(bp, name)) return 0;
	if (NULL == (xcp = managexcp(7))) return 0; escseq = xcp->xm_escseq;
	adp = escseq + NSYMBOLS; /* used in a related exercise */
	asp = xcp->xm_symb + NSYMBOLS * TSYMBOLWID; /* ditto */
	/* now use `tgetstr' to save up all the terminal characteristics
	that we need to understand in order to use this terminal.  Meanwhile,
	we remember the longest string we find. */
	for (i = 0, where = asp; i < NSYMBOLS; i++, adp++, where += TSYMBOLWID)
		*adp = asp;
	adp -= NSYMBOLS;	where = xcp->xm_symb;
	/* go through the symbols the user wants.  The ones with names
	beginning with "v" aren't checked because they're peculiar to vtxxx's*/
	for (i = maxlength = 0, dcp = escseq; i < NSYMBOLS - 5; i++, dcp++)
	   {
#ifdef sgi
		if (NULL != (result = tgetstr(codes[i], (char **)NULL)))
#else
		if (NULL != (result = tgetstr(codes[i], adp)))
#endif
		  {
			esclen = strlen(result);
			*dcp = where;	strcpy(*dcp, result);
			where += esclen + 2;
			if (esclen > maxlength) maxlength = esclen;
			if (i >= NSYMBOLS - 10)
			  {
				*(dcp + 5) = where;
				strcpy(*(dcp + 5), result);
			  }
		  }
		else {
			*dcp = NULL; if (i >= NSYMBOLS - 10) *(dcp + 5) = NULL;
		     }
	   }
	xcp->xm_maxlength = maxlength;
	/* some special cases: the K1 through K9 sequences are set equal to the
	k1 through k9 sequences if the former don't exist and the latter do. */
	for (i = 10, dcp = escseq + 10; i < 20; i++, dcp++)
		if (!(*dcp)) *dcp = *(dcp - 10);
	/* on VT100's and cousins ku,kd,kl,kr,and bc are kludged thusly: */
	if ((!strncmp(name, "vt", 2)) || (!strncmp(name, "VT", 2)) ||
	    (!strncmp(name, "cit", 3)) || (!strncmp(name, "CIT", 3)))
		for (dcp = escseq + NSYMBOLS-5; dcp < escseq + NSYMBOLS; dcp++)
			if (*dcp) *((*dcp) + 1) = '[';
	/* further kludge: we suspect that specifications that look like \EO are
	 really read out as \E[. To prevent overuse of this feature, we limit it
	 to sequences that have only three characters in them */
	for (dcp = escseq; dcp < escseq + NSYMBOLS; dcp++)
	   {
		if (NULL != *dcp)
		  {
			asp = *dcp;
			if ((033 == *asp) && ('O' == *(asp + 1)) &&
				('\0' == *(asp + 3))) *(asp + 1) = '[';
		  }
	   }
	/* list them out if desired. */
	if (diagnostic)
	  {
		printf("Number Code   Escape Sequence\n");
		for (i = 0, dcp = escseq; i < NSYMBOLS; i++, dcp++)
		   if (NULL != *dcp) printf( "%6d %4s   %s\n",
			i, codes[i], careful_print((sichar *)*dcp));
	  }
	/* release the memory we no longer need */
	memzap((AVOID **)&name, "name in setmore");
	memzap((AVOID **)&bp, "bp in setmore");	return 1;
}

sichar getescseq(WINDOW *mwp)
{ /*	This returns the code value associated with an escape sequence
	produced by a numbered function key, an arrow key, or some similar
	object. It assumes that all the functions are the same length, e.g.,
	that the DOWNARROW and F1 keys produce the same number of characters.
	It returns -1 if no correspondence is found.
	Note: this code has been cleaned up for XENGEN V.1.3. */
	register int	i, j, m;
	char		estr[SEQLENGTH+1], *lcp;
	XGCMENU		*xcp;
	static sichar	vcodes[NSYMBOLS] = {
	-124,-128,-127,-126,-125,-119, -98, -97, -96, -95,
	-124,-128,-127,-126,-125,-119, -98, -97, -96, -95,
	   8,   8,  26,  40,-106,-105,-109,-108,-107,-110,
	-108,-109,-108,-107,-110,-108 } ;

	estr[0] = ESCAPE;	/* first char of sequence is an ESCAPE */
	/* now we go through the list of legal escape sequences for
	the current terminal type.  If we get a total match, we pronounce it
	finished.  If we get a partial match, we figure we're on to something
	and read another character and check that.  If we miss altogether
	we return -1. */
	if (NULL == (xcp = managexcp(3))) return 'a';
	for (i = 1; i < (int)(xcp->xm_maxlength); i++)
	   {
		estr[i] = wgetch(mwp);
		estr[i+1] = '\0';
		for (j = 0; j < NSYMBOLS; j++)
		   if ((lcp = xcp->xm_escseq[j]))
		     {
			for (m = 0; m <= i; m++) if (estr[m] != lcp[m]) break;
			if (m > i) break;
		     }
		if (j >= NSYMBOLS) return -1;
		/* if we've found the end of this escseq, we've found the
		right answer.  If not, we've found the middle of one, and
		we'll need to look further by incrementing i and continuing.*/
		else if (lcp[i+1] == '\0') return vcodes[j];
	   }
	return (-1);	/* if we got this far, nothing matched. */
}
#endif

/* routines general to VMS and non-VMS versions */
sichar ordinary_key(sichar c)
{	/* this returns the ordinary (non-function) key associated with the
	keypad characters */
	static sichar equiv_char[] = {'1','2','3','4',
			'0','1','2','3','4','5','6','7','8','9','0',
			'\n','-',',','.' };

	return equiv_char[(c - K_PF1)];
}

int effecx(WINDOW_ENTRY *wp)
{ /*	This returns the effective x position of a window entry. This is
	basically the point on which typing would start within the entry. */
	if ((wp->we_type == WST) || (wp->we_type == WPT)) return wp->we_x0;
	return wp->we_x0 + (wp->we_strlen - 1);
}

static void menu_link(MENUDATA *mp, int parno)
{ /*	this determines the indices of the parameters which are neighbors to
	the parameter numbered `parno'.  It sticks them in the slots of the
	structure pointed to by mp->wen[parno].we_link. */
	int		diffx, effx, mindiffx, nmin, npar, nrsy;
	WINDOW_ENTRY	*wp, *wpp, *wp0;

	wp = mp->wen + parno;
	/* don't even try to find them if the parameter isn't active. */
	if (wp->we_flags & WSKP) return;
	effx = effecx(wp);
	wp0 = mp->wen + mp->wfirstactive;
	if (parno == mp->wlastactive)
	  {
		npar = mp->wfirstactive;	wpp = mp->wen + npar;
	  }
	else for (wpp = wp + 1, npar = parno + 1; wpp->we_flags & WSKP;
	 wpp++, npar++) /* find the next element in the list */
	  if (npar > mp->wlastactive)
	    {
		npar = mp->wfirstactive;	wpp = wp0;	break;	}
	wp->we_nextelt = npar;	/* this is the next active parameter */
	for ( ; (wpp->we_flags & WSKP) || (wpp->we_y == wp->we_y);
	  npar++, wpp++) /* find the first elt. in the row after this one */
	   if (npar >= mp->wlastactive)
	     {
		npar = mp->wfirstactive;	wpp = wp0;	break;	}
	wp->we_nrstart = npar;		/*first active elt in next row*/
	if (npar >= mp->wlastactive) wp->we_nextrow = npar = mp->wlastactive;
	else {
		nrsy = wpp->we_y;	/* remember this row number */
		diffx = abs(effecx(wpp)- effx);	mindiffx = diffx; nmin = npar;
		for (wpp++, npar++ ; ((npar <= mp->wlastactive) &&
		 (wpp->we_y == nrsy)); wpp++, npar++)
		   if (!(wpp->we_flags & WSKP))
		     {
			diffx = abs(effecx(wpp) - effx);
			if (diffx < mindiffx)
			  {
				mindiffx = diffx; nmin = npar;
			  }
		     }
		wp->we_nextrow = nmin; /* element below the current one */
	     }
	if (parno <= mp->wfirstactive) npar = mp->wlastactive;
	else for (wpp = wp - 1, npar = parno - 1; wpp->we_flags & WSKP;
	 wpp--, npar--) /* look for the one before ours */
	   if (npar < mp->wfirstactive)
	     {
		npar = mp->wlastactive;	wpp = mp->wen + npar;	break;
	     }
	wp->we_prevelt = npar;	/* active element previous to current one*/
	/* to find the first element in the current row
	we work backward until we find the first one that isn't in this row. */
	for (nmin = npar = parno, wpp = wp; (npar >= mp->wfirstactive) &&
	 (wpp->we_y == wp->we_y); wpp--, npar--)
		if (!(wpp->we_flags & WSKP)) nmin = npar;
	wp->we_trstart = npar = nmin;
	/* now we find the last active element in a row above the current one.
	 We'll work backward from there to find the element
	 directly above the current one */
	if (npar > mp->wfirstactive)
	  {
		for (npar--; npar >= mp->wfirstactive; npar--)
			if ((mp->wen + npar)->we_y != wp->we_y) break;
		while (((mp->wen + npar)->we_flags) & WSKP) npar--;
	  }
	else npar = mp->wlastactive;
	if (npar == mp->wfirstactive) wp->we_prevrow = npar;
	else {
		wpp = mp->wen + npar; nrsy = wpp->we_y;	/* remember row # */
		diffx = abs(effecx(wpp)- effx);	mindiffx = diffx; nmin = npar;
		for (wpp--, npar--; ((npar >= mp->wfirstactive) &&
		 (wpp->we_y == nrsy)); wpp--, npar--)
		   if (!(wpp->we_flags & WSKP))
		     {
			diffx = abs(effecx(wpp) - effx);
			if (diffx < mindiffx)
			  {
				mindiffx = diffx; nmin = npar;
			  }
		     }
		wp->we_prevrow = nmin; /* element below the current one */
	     }
}
	 
void construct_links(MENUDATA *mp)
{ /* "construct_links":
	This constructs the linkages between elements in the MENUDATA
	structure pointed to by mp.  Currently 6 geometrical relationships
	are recognized:
	Motion	Keys Enabling		Function
	------	---- --------		--------
	up	uparrow, ^K		move up more or less in column
	down	downarrow, <LF>		move down more or less in column
	left	leftarrow, ^H		move one element earlier in list
	right	righarrow, ^L		move one element later in list
	newline	newline, ^M		move to 1st element in next line
	return	^B			move to 1st element in current line.
 We assume that the user has already constructed the geometry of the
 MENUDATA structure and that the initial values for we_y and we_x0 for each
 entry in the structure are in place.  */
	int		parno;
	WINDOW_ENTRY	*wp;

	/* first, find the first & last active elements in the table */
	for (wp = &(mp->wen[0]), parno = 0; parno <= mp->wnparams - 1;
	 wp++, parno++) if (!(wp->we_flags & WSKP)) break;
	mp->wfirstactive = parno;
	for (wp = mp->wen + mp->wnparams - 1, parno = mp->wnparams - 1; parno;
	 wp--, parno--) if (!(wp->we_flags & WSKP)) break;
	mp->wlastactive = parno;
	/* for each active parameter, fill up the link table */
	for (wp = &(mp->wen[mp->wfirstactive]), parno = mp->wfirstactive;
	 parno <= mp->wlastactive; parno++, wp++)
	   if (!(wp->we_flags & WSKP)) menu_link(mp, parno);
}

void stuff_value(WINDOW_ENTRY *wep)
{ /* this stuffs an actual value into we_wvlong or we_wvfloat. */
	register int	ic;
	register char	*cp;
	char		vformat[12];

	if (wep->we_type == WST || wep->we_type == WCH || wep->we_type == WPT)
		return;	/* if it's a string, do nothing */
		/* the string "DEF" (meaning "default") is ignored */
	if ((!strncmp(wep->we_string, "DEF", 3)) || 
		(!strncmp(wep->we_string, "def", 3))) return;
	/* otherwise we need to record the format information */
	for (ic = 12, cp = &(vformat[0]); ic--; cp++) *cp = '\0';
	vformat[0] = '%';
	sprintf(&(vformat[1]), "%d", wep->we_strlen);
	if (wep->we_type == WFL)
	  {
		wep->we_wvfloat = atof(wep->we_string);
		sprintf(&(vformat[strlen(vformat)]), ".%df", wep->we_afterdot);
		sprintf(wep->we_string, vformat, wep->we_wvfloat);
	  }
	else {
		wep->we_wvlong = atoi(wep->we_string); strcat(vformat,"ld");
		sprintf(wep->we_string, vformat, wep->we_wvlong);
	     }
}

void valtostr(WINDOW_ENTRY *wep)
{	/* this takes a floating or integer value and puts it into the
	appropriate string */

	register int	ic;
	register char	*cp;
	char		vformat[12];
	if (wep->we_type == WST || wep->we_type == WCH || wep->we_type == WPT)
		return;	/* if it's a string, do nothing */
	/* otherwise we need to record the format information */
	for (ic = 12, cp = &(vformat[0]); ic--; cp++) *cp = '\0';
	vformat[0] = '%';
	sprintf(&(vformat[1]), "%d", wep->we_strlen);
	if (wep->we_type == WFL)
	  {
		sprintf(&(vformat[strlen(vformat)]), ".%df", wep->we_afterdot);
		sprintf(wep->we_string, vformat, wep->we_wvfloat);
	  }
	else {
		strcat(vformat,"ld");
		sprintf(wep->we_string, vformat, wep->we_wvlong);
	     }
	wep->we_string[wep->we_strlen] = '\0'; /* unnecessary, we hope! */
}

static int legalchar(char ci, WINDOW_ENTRY *wp, int pos)
{ /*	This returns true iff the character ci is legal within the string
	associated with this window element.  Alphabetics are not checked
	for case here; that's done elsewhere.  `pos' is the position of
	the character within the string.  In all cases we disallow the
	vertical bar or absolute-value character (|) since it's used as
	a delimiter between entries on the diskfile version of the menu. */
	char	c, *cp;
	int	i, ic;

	if (wp->we_type == WPT) return 0;
	if ('|' == ci) return 0;
	if ('-' == ci) /* minus sign is treated specially */
	  {
		if (wp->we_type == WCH) return 0; /* illegal for char vbl */
		else if (wp->we_type == WST) return 1; /* ok in string */
		else if (!pos) return 1;	/* ok in numeric at start */
		else /* ok in numeric if only blanks precede it */
		  {
			for (i = 0, cp = &(wp->we_string[0]);
			 '\0' != *cp && i < pos; i++, cp++)
			   {
				if (*cp != ' ') return 0;
			   }
			return 1;
		  }
	  }
	c = ('a' <= ci && ci <= 'z' ? ci - ('a' - 'A') : ci);
	ic = c;	/* to keep the picky compilers happy */
	switch (wp->we_type) /* note '-' is already taken care of! */
	  {
		case	WCH: return isupper(ic) || isdigit(ic) ||
			(c == '@');
		case	WSH: return isdigit(ic);
		case	WFL: return isdigit(ic) || (c == '.');
		case	WST:	default:	return isprint((int)ic);
	  }
}

static void embolden(WINDOW *win)
{	/* This emboldens the subsequent entries in the WINDOW *win.
	It looks good on VMS and Irix, and lousy elsewhere currently */
#ifdef VMS
	wsetattr(win, _BOLD);
#else
#ifdef A_BOLD
	wattrset(win, (xctype)(A_BOLD));
#else
	wstandout(win);
#endif
#endif
}

static void unembolden(WINDOW *win)
{	/* This turns off what `embolden' does. */
#ifdef VMS
	wclrattr(win, _BOLD);
#else
	wstandend(win);
#endif
}

void disp_param(MENUDATA *mp, WINDOW  *mwp, int  parno)
{ /*	This updates the display of the `parno'th parameter of menu `mp' */
	char		*cp;
	int		do_embold, needblank;
	WINDOW_ENTRY	*wp;
	XGCMENU		*xcp;

	if ((NULL == mwp) || (NULL == (xcp = managexcp(3)))) return;
	do_embold = 1;
	if (parno < 0)	/* code meaning "don't embolden" */
	  {
		do_embold = 0;	parno = -parno;
		wp = &(mp->wen[parno]);
	  }
	wp = &(mp->wen[parno]);
	if ((wp->we_flags & WNPG) || (!wp->we_strlen) ||
		wp->we_string[0] == '\0') return;
	wmove(mwp, (int)wp->we_y, (int)((wp->we_x0 < 1 ? 0 : wp->we_x0 - 1)));
	if (wp->we_flags & WFLG) waddch(mwp, (xctype)(':'));
	else if (wp->we_delim == ' ') wmove(mwp, (int)wp->we_y,(int)wp->we_x0);
	else waddch(mwp, (xctype)(wp->we_delim));
	if (do_embold) embolden(mwp);
	needblank = ((int)wp->we_strlen) - ((int)strlen(wp->we_string));
	if (needblank < 0) *(wp->we_string + wp->we_strlen) = '\0';
	if ((wp->we_string) && ('\0' != *(wp->we_string)))
		waddstr(mwp, wp->we_string);
	if (needblank > 0)
	  {
		cp = xcp->xm_blanks + needblank;	*cp = '\0';
		waddstr(mwp, xcp->xm_blanks);		*cp = ' ';
	  }
	if (do_embold) unembolden(mwp);
}

static void disp_label(MENUDATA *mp, WINDOW *mwp, int parno)
{ /* This updates the display of the label on the `parno'th parameter */
	char		*cp;
	int		needblank, obs_lablen, postmove, xvstar;
	WINDOW_ENTRY	*wp;
	XGCMENU		*xcp;

	if ((NULL == mwp) || (NULL == (xcp = managexcp(3)))) return;
	wp = mp->wen + parno;
	unembolden(mwp);
	if (wp->we_flags & WNPG) return;
	if ((NULL != wp->we_label) && ('\0' != wp->we_label[0]) &&
	  wp->we_lablen != 0)
	  {
		obs_lablen = strlen(wp->we_label);
		needblank = ((int)wp->we_lablen) - obs_lablen;
		if (needblank < 0)
			*(wp->we_label + wp->we_lablen) = '\0';
		if ((!parno) || (wp->we_y != (wp-1)->we_y) ||
		 wp->we_flags & WTYT)
		  {
			xvstar = wp->we_x0 - wp->we_lablen - 1;
			if (xvstar < 1) xvstar = 0;
			wmove(mwp, (int)wp->we_y, xvstar);
			waddstr(mwp, wp->we_label);
		  }
		else {
			wmove(mwp, (int)wp->we_y,
				(int)(wp->we_x0 - wp->we_lablen - 3));
			wprintw(mwp, "| %s", wp->we_label);
		     }
		if (needblank > 0)
		  {
			cp = xcp->xm_blanks + needblank;	*cp = '\0';
			wmove(mwp, (int)wp->we_y,(int)(wp->we_x0 - needblank));
			waddstr(mwp, xcp->xm_blanks);		*cp = ' ';
		  }
	  }
	if (NULL != wp->we_post && '\0' != wp->we_post[0] &&
	  strlen(wp->we_post) > 0)
	  {
		postmove = wp->we_x0 + wp->we_strlen +
			((wp->we_flags & WFIL) != 0);
		wmove(mwp, (int)wp->we_y, postmove);
		waddstr(mwp, wp->we_post);
	  }
}

void restore_value(MENUDATA *curmp, MENUDATA *orp, int parno)
{ /*	This copies an element of one MENUDATA structure into the
	corresponding element of another MENUDATA structure.
	The assumption is that only the actual contents of the WINDOW_ENTRY
	string (we_string), the corresponding value,
	and the pre- and post-labels need to be copied. */
	register WINDOW_ENTRY	*wnp, *wop;

	wnp = curmp->wen + parno;	wop = orp->wen + parno;
	if (wop->we_strlen != 0)
	  {
		cpymem((AVOID *)wnp->we_string, (AVOID *)wop->we_string,
			(unsigned)wop->we_strlen);
		wnp->we_string[wop->we_strlen] = '\0';
	  }
	if (wop->we_lablen != 0)
	  {
		cpymem((AVOID *)wnp->we_label, (AVOID *)wop->we_label,
			(unsigned)wop->we_lablen);
		wnp->we_label[wop->we_lablen] = '\0';
	  }
	if ((wop->we_post) && ('\0' != wop->we_post[0]))
	  {
		cpymem((AVOID *)wnp->we_post, (AVOID *)wop->we_post,
			strlen(wop->we_post));
		wnp->we_post[strlen(wop->we_post)] = '\0';
	  }
	if (wnp->we_type == WFL) wnp->we_wvfloat = wop->we_wvfloat;
	else if (wnp->we_type != WST && wnp->we_type != WCH)
		wnp->we_wvlong = wop->we_wvlong;
}

void menufree(MENUDATA *oldmenu)
{ /* This frees up memory allocated within `newmenu' for an old menu */
	char		*ltank;
	WINDOW_ENTRY	*wp;

	if (NULL != oldmenu)
	  {
		if (NULL != (wp = oldmenu->wen))
		  {
			if (NULL != (ltank = wp->we_string))
			  {
				memzap((AVOID **)&ltank, "ltank in menufree");
				ltank = (char *)0;
			  }
			memzap((AVOID **)&wp, "wp in menufree");
			wp = (WINDOW_ENTRY *)0;
		  }
		memzap((AVOID **)&oldmenu, "oldmenu in menufree");
		oldmenu = (MENUDATA *)0;
	  }
}

MENUDATA *newmenu(unsigned nparams, WINDOW_TEMPLATE *window_init)
{	/* This saves memory for a MENUDATA structure and its contents. */
	unsigned	i, nchars, psp, thislen;
	MENUDATA	*mp;
	WINDOW_ENTRY	*wp;
	WINDOW_TEMPLATE	*wtp;
	char		*ltank, *wcp;

	if (NULL == (mp = (MENUDATA *)callop(1, sizeof (MENUDATA),
		"mp in newmenu"))) return NULL;
	if (NULL == (mp->wen = (WINDOW_ENTRY *)callop(nparams,
		sizeof (WINDOW_ENTRY), "mp->wen in newmenu"))) return NULL;
	mp->wnparams = nparams;
	for (nchars = i = 0, wtp = window_init; i < nparams; i++, wtp++)
	   {
		psp = (wtp->wt_post && '\0' != *(wtp->wt_post) ?
			strlen(wtp->wt_post) + 4 : 6);
		nchars += wtp->wt_strlen + wtp->wt_lablen + psp + 8;
	   }
	if (NULL == (ltank = (char *)mallop(nchars, "ltank in newmenu")))
		return NULL;
	wp = &(mp->wen[0]);	wtp = window_init;
	for (i = 0, wcp = ltank; i < nparams; i++, wp++, wtp++)
	   {
		wp->we_x0 = wtp->wt_x0;		wp->we_y = wtp->wt_y;
		wp->we_strlen = wtp->wt_strlen;	wp->we_lablen = wtp->wt_lablen;
		wp->we_type = wtp->wt_type;	wp->we_delim = wtp->wt_delim;
		wp->we_flags = wtp->wt_flags;
		wp->we_afterdot = wtp->wt_afterdot;	wp->we_string = wcp;
		wcp = wp->we_label = wcp + wtp->wt_strlen + 4;
		wcp = wp->we_post = wcp + wtp->wt_lablen + 4; 
		if (NULL != wtp->wt_string && ('\0' != *(wtp->wt_string)))
		  {
			strncpy(wp->we_string, wtp->wt_string, wp->we_strlen);
			thislen = strlen(wtp->wt_string);
			if (thislen > wp->we_strlen)
				wp->we_string[wp->we_strlen] = '\0';
			else	wp->we_string[thislen] = '\0';
		  }
		else	*wp->we_string = '\0';
		if (NULL != wtp->wt_label && ('\0' != *(wtp->wt_label)))
		  {
			strncpy(wp->we_label, wtp->wt_label, wp->we_lablen);
			thislen = strlen(wtp->wt_label);
			if (thislen > wp->we_lablen)
				wp->we_label[wp->we_lablen] = '\0';
			else	wp->we_label[thislen] = '\0';
		  }
		else	*wp->we_label = '\0';
		if (NULL != wtp->wt_post && '\0' != *(wtp->wt_post))
		  {
			psp = strlen(wtp->wt_post) + 4;
			strcpy(wp->we_post, wtp->wt_post);
		  }
		else {
			*(wp->we_post) = '\0';	psp = 6; }
		wcp += psp;
	   }
	construct_links(mp);	return mp;
}

void put_pars(FILE *fp, MENUDATA *mp, int where)
{ /* This writes out the menu to a diskfile in approximately the
	same format as the screen menu itself. The file is rewound
	before starting if `where' is zero; otherwise we write where we are.*/
	register int	i, j;
	int		endpos, labpos, pstart;
	char		*cp, *icp;
	char		lbuf[255];
	WINDOW_ENTRY	*wp, *wn;

	wp = mp->wen;	wn = wp + 1;
	if (where == 0) (void)fseek(fp, 0L, 0);	/* rewind */
	for (j = 0; j < 254; j++) lbuf[j] = ' ';	lbuf[254] = '\0';
	for (i = 0; i < mp->wnparams; i++, wp++, wn++)
	   {
		if ((wp->we_label) && ('\0' != *wp->we_label))
		  {
			labpos = wp->we_x0 - strlen(wp->we_label) - 1;
			if (labpos < 0) labpos = 0;
			for (cp = &(lbuf[labpos]), icp = wp->we_label, j = 0;
			  (j < (int)wp->we_lablen) && ('\0' != *icp);
			  icp++, cp++, j++) *cp = *icp;
		  }
		if (('\0' != wp->we_delim) && (' ' != wp->we_delim) &&
		  (wp->we_x0)) lbuf[wp->we_x0 - 1] = wp->we_delim;
		if ((wp->we_string) && ('\0' != *wp->we_string))
		  for (cp = &(lbuf[wp->we_x0]), icp = wp->we_string, j = 0;
		    (j < (int)wp->we_strlen) && ('\0' != *icp); icp++,cp++,j++)
			*cp = *icp;
		pstart = wp->we_x0 + wp->we_strlen;
		if ((wp->we_post) && ('\0' != *wp->we_post))
		  for (cp = &(lbuf[pstart]), icp = wp->we_post;
		    '\0' != *icp; icp++, cp++) *cp = *icp;
		if ((i == mp->wnparams - 1) || (wp->we_y != wn->we_y))
		  {
			endpos = pstart;
			if ((wp->we_post) && (*wp->we_post))
				endpos += strlen(wp->we_post);
			lbuf[endpos] = '\n';	lbuf[endpos+1] = '\0';
			fputs(lbuf, fp);
			for (j = 0; j < endpos+2; j++) lbuf[j] = ' ';
		  }
	   }
}

void update_par(MENUDATA *mp, WINDOW *mwp, int par)
{	/* This updates the contents of the appropriate parameter
	if anything has been stuck into its slot.  */
	int		nch;
	char		*icp, *ocp;
	WINDOW_ENTRY	*wp;
	XGCMENU		*xcp;

	if (NULL == (xcp = managexcp(3))) return;
	if (*(xcp->xm_answer) == '\0') return;
	wp = mp->wen + par;	if (wp->we_flags & WSKP) return;
	if (WCH == wp->we_type)
	  {
		*(xcp->xm_answer + 1) = '\0';
		wp->we_string[0] = *(xcp->xm_answer);
	  }
	else
	  {
		if (strcmp(wp->we_string, xcp->xm_answer))
		  {
			for (nch = wp->we_strlen, icp = xcp->xm_answer,
			 ocp = wp->we_string; (nch > 0) &&('\0'!= *icp);
				nch--) *ocp++ = *icp++;
			*ocp = '\0';
		  }
	  }
	if (wp->we_type != WST) stuff_value(wp);
	disp_param(mp, mwp, par);	/* put the new value into the menu */
	wmove(mwp, (int)wp->we_y, (int)wp->we_x0);
	wrefresh(mwp);	/* display it */	return;
}

static int actual_char(MENUDATA *mp, WINDOW *mwp,
 int  par, int pos, int insm, sichar  c)
{ /*	This processes an ordinary displayable character.  It usually returns
	the position within the entry of the next character to be processed.
	If the entry is of type WPT it returns the (-next-1), where `next'
	is the parameter number of the next parameter after the current one. */
	int		thisx, thisy;
	char		*answer, *blanks, *cp;
	WINDOW_ENTRY	*wp;
	XGCMENU		*xcp;

	wp = mp->wen + par;
	if (wp->we_type == WPT) return (-wp->we_nextelt - 1);
	if (wp->we_flags & WCAP)
	  {
		if ('a' <= c && c <= 'z') c -= ('a' - 'A');
	  }
	else if ((wp->we_flags & WLCA) && 'A' <= c && c <= 'Z')
		c += ('a' - 'A');
	/* the variable `xcp->xm_answer' should contain whatever was in the
	string before we added this character */
	if (NULL == (xcp = managexcp(3))) return 0;
	answer = xcp->xm_answer;	blanks = xcp->xm_blanks;
	if ('\0' == *answer)
	 strncpy(answer, wp->we_string, (unsigned)wp->we_strlen);
	if (legalchar((char) c, wp, pos))
	  { /* Change for XGEN 3.0 and XENGEN 2.1:
		don't clean up the window here for multi-character fields */
		/* record answer right away if 1 char */
		thisy = wp->we_y;
		if (wp->we_strlen <= 1)
		  {
			twoin(answer, c, '\0'); update_par(mp, mwp, par);
		  }
		else if (insm) /* push remainder of msg to right */
		  {
			cp = answer + wp->we_strlen - 2;
			for ( ; (cp >= answer + pos) && ('\0' == *cp) ; ) cp--;
			for ( ; cp >= answer + pos; cp--) *(cp + 1) = *cp;
			wmove(mwp, (int)wp->we_y, (int)wp->we_x0);
			cp = blanks + wp->we_strlen;
			*cp = '\0'; waddstr(mwp, blanks); *cp = ' ';
			if (pos < (int)wp->we_strlen)
			  {
				*(answer + pos) = c;	pos++;	/* increment */
			  }
			else if ((wp->we_type != WST) && (' ' == *answer))
			  {
				for (cp = answer; cp < answer + pos - 1; cp++)
					*cp = *(cp + 1);
				*cp = c;	*(cp + 1) = '\0';
			  }
			if (wp->we_type == WST) thisx = wp->we_x0;
			else thisx = wp->we_x0 +wp->we_strlen -strlen(answer);
			mvwaddstr(mwp, thisy, thisx, answer);
			wmove(mwp, thisy, (int)wp->we_x0 + pos);
		  }
		else if ((wp->we_type != WST) && (pos == wp->we_strlen) &&
		 ' ' == *answer)
		  {
			for (cp = answer; cp < answer + pos - 1; cp++)
				*cp = *(cp + 1);	
			*cp = c;	*(cp + 1) = '\0';
			thisx = wp->we_x0 + wp->we_strlen - strlen(answer);
			mvwaddstr(mwp, thisy, thisx, answer);
			wmove(mwp, thisy, (int)wp->we_x0 + pos);
		  }
		else /* overstrike the current character */
		  {
			thisx = wp->we_x0 + pos;
			mvwaddch(mwp, thisy, thisx, (xctype)(c));
			*(answer + pos) = c;
			/* null-terminate just past here so that if we leave
			this field right after this it'll end there*/
			*(answer + pos + 1) = '\0';
			if (pos < (int)wp->we_strlen)
			  {
				pos++; thisx++;
			  }
			wmove(mwp, thisy,  thisx);
		  }
		wrefresh(mwp);
	  } /* close of "if (legalchar...) */
	return pos; /* if we get here we're still in the middle of the entry */
}

static int erasure(WINDOW_ENTRY *wp, WINDOW *mwp, int pos)
{ /* This code handles an "erase character" command */
	char	*answer, *cp, *cpe;
	int	thisx, thisy;
	XGCMENU	*xcp;

	if (NULL == (xcp = managexcp(3))) return 0; answer = xcp->xm_answer;
	for (cpe = answer + wp->we_strlen - 1;
	 cpe > answer + pos && (*cpe == '\0' || *cpe == ' '); ) cpe--;
	for (cp = answer + pos; cp < cpe; cp++) *cp = *(cp + 1);
	*cp = '\0';
	thisx = wp->we_x0 + pos;	thisy = wp->we_y;
	mvwaddch(mwp, (int)wp->we_y,
		(int)wp->we_x0 + (cp - answer), (xctype)(' '));
	mvwaddstr(mwp, thisy, thisx, answer + pos);
	wmove(mwp, thisy, thisx);	wrefresh(mwp);	return pos;
}

void menu_setup(MENUDATA *mp, WINDOW *mwp, char *headerstr, int menuflags)
{ /* This sets up the entire menu with values found in the MENUDATA
	structure `mp'.  The window on which all of them are written
	is `mwp'.  The string `headerstr' is displayed in boldface on the
	top row. The flag `menuflags' governs whether the results being
	stuck into this menu are actually displayed as they're constructed
	and whether the emboldening of the title should be conditional */
	int		isready, par;
	WINDOW_ENTRY	*wp;
	XGCMENU		*xcp;

	if (NULL == (xcp = managexcp(7))) return;
	if (menuflags & 01)
	  {
		wmove(mwp, 0, 0);	wclrtobot(mwp); wrefresh(mwp);
	  }
	if (NULL != headerstr && '\0' != *headerstr)
	  {
		embolden(mwp);	waddstr(mwp, headerstr); unembolden(mwp);
	  }
	for (par = 0; par < mp->wnparams; par++)
	   {
		disp_label(mp, mwp, par);
		if (menuflags & 02) /* make emboldening conditional */
		  {
			wp = &(mp->wen[par]); isready = ready(wp->we_string);
			if (!isready) disp_param(mp, mwp, -par);
			else disp_param(mp, mwp, par);
		  }
		else disp_param(mp, mwp, par);
	   }
	par = mp->wfirstactive;
	/* special case of a menu with no active elements--only passives:
	 In this case we move the cursor to the first element */
	if (par >= mp->wnparams) par = 0;
	if (menuflags & 01)
	  {
		wmove(mwp, (int)mp->wen[par].we_y, (int)mp->wen[par].we_x0);
		wrefresh(mwp);
	  }
}

int getans(MENUDATA *mp, MENUDATA *oldp, WINDOW *mwp, int par, char *headerst)
{ /* this reads the user inputs in a known parameter location.
	If the user is typing ordinary characters, it displays them
	if they're legal and ignores them if not.  If the user types
	a special character it performs some function, like moving around
	on the menu screen.  The routine returns the slot number of
	the parameter that is to be examined next (because of the
	geometrical commands entered by the user), or else a code value,
	viz. K_PF1 if the user types PF1, K_PF2 if the user types PF2,
	or K_HELP if the user types a question mark in the 1st col or HELP
	on the VT2xx keyboard.
	The caller has to figure out what to do with these. */
	register int	i, j;
	int		thisx, thisy, tlen;
	sichar		c;
	char		*answer, *cp, *icp;
	WINDOW_ENTRY	*wp;
	XGCMENU		*xcp;
	static char	insmode = 0; /* controls whether we insert or not*/

	wp = mp->wen + par;	tlen = wp->we_strlen;
	/* if this is a parameter we're supposed to skip, skip it! */
	if (wp->we_flags & WSKP) return mp->wfirstactive;
	if (NULL == (xcp = managexcp(3))) return mp->wfirstactive;
	answer = xcp->xm_answer;
	for (i = 0, cp = answer, icp = wp->we_string;
	 i < tlen; i++, cp++, icp++) *cp = *icp;
	*cp = '\0';
	for (--cp; cp >= answer; cp--)
	   {
		if (*cp == ' ') *cp = '\0'; else break;
	   }
	thisy = wp->we_y;
	if (wp->we_type == WST || wp->we_type == WPT) i = 0;
	else for (i = 0; i < tlen - 1; i++)
		if ((*(cp = wp->we_string + i) != '\0') && (*cp != ' ')) break;
	thisx = wp->we_x0 + i;
	wmove(mwp, thisy, thisx);	wrefresh(mwp);
	while ('\0' != (c = wget1ch(mwp)))
	     {
#ifdef VMS
	     if (K_KP0 <= c && c <= K_PERIOD) c = ordinary_key(c);
#else
	     if (ESCAPE == c || 01 == c) c = getescseq(mwp);
#endif
	     switch (c)
	      {
		case	K_PF1:		/* PF1: perform command */
		case	K_DO:	case	030:		/* control-X */
			return	K_PF1;	/* flag return value meaning "do it" */
		case	K_PF2:		/* PF2: get help */
		case	031:		/* control-Y */
		case	006:		/* control-F--^Y==INTR some systems*/
			return	K_PF2;	/* flag meaning "get help" */
		/* Rightarrow or ^L gives either "next field" or else simply
		pushes the cursor 1 character to the right, depending on
		whether or not we're at the end of the field.  If the field is
		only 1 character long, it always goes to the next field */
		case	K_RIGHT:	case 014:
		if ((wp->we_type != WPT) && (i < tlen) &&
		 	!(wp->we_flags & WSKP) && (tlen > 1))
			  {
				getyx(mwp, thisy, thisx);
				wmove(mwp, thisy, thisx + 1);
				wrefresh(mwp);	i++;	break;
			  }
			else return	wp->we_nextelt;
		/* spacebar is like a tab on a pointer input;
		it's a destructive right-erase on a numeric input;
		it's an ordinary character on a string input */
		case	' ':
			if (wp->we_type == WPT) return wp->we_nextelt;
			else if ((wp->we_type != WST) && (wp->we_type != WCH))
			  {
				if (i < tlen)
				  {
					i = erasure(wp, mwp, i);	break;
				  }
				else	return wp->we_nextelt;
			  }
			else {
				if (0 > (i = actual_char(mp, mwp, par, i,
					insmode, c))) return (-i - 1);
				break;
			  }
		case	'\t':	/* tabs always go to next element */
			return wp->we_nextelt;
		case	K_UP:	case 013:	/* up 1 row in column */
			return	wp->we_prevrow;
		/* leftarrow gives either "previous field" (with wraparound) or
		pushes the cursor one char to the left, depending on whether or
		not we're at the beginning of the field. We're always at the
		start of 1-char fields, so it always means "prev field" then */
		case	K_LEFT:
			if ((wp->we_type != WPT) && (i) &&
			 !(wp->we_flags & WSKP))
			  {
				getyx(mwp, thisy, thisx);
				wmove(mwp, thisy, thisx - 1);
				wrefresh(mwp);	i--; break;
			  }
			else return wp->we_prevelt;
		case	022:	/* go to previous field */
			return wp->we_prevelt;
		case	K_DOWN:	/* go down one row; stay in the column */
		case	K_NEXT_SCREEN:	return wp->we_nextrow;
		case	K_PF3:	/* get previous value for this param */
			restore_value(mp, oldp, par);
			disp_param(mp, mwp, par);
			wrefresh(mwp);	return par;
		case	K_PF4:	/* restore all previous values */
			for (j = 0; j < mp->wnparams; j++)
			   {
				restore_value(mp, oldp, j);
				disp_param(mp, mwp, j);
			   }
			wrefresh(mwp);	return par;
		case	027:			/* redraw */
			menu_setup(mp, mwp, headerst, 1);
			strncpy(answer, wp->we_string, (unsigned)tlen);
			wmove(mwp, thisy, (int)wp->we_x0 + i);
			wrefresh(mwp);	break;
		case	025:	/* start over on this param */
		case	K_REMOVE:
			for (i = tlen; i--; )
			   {
				wp->we_string[i] = ' ';	answer[i] = '\0';
			   }
			mvwaddstr(mwp, (int)wp->we_y, (int)wp->we_x0,
			 wp->we_string); wrefresh(mwp);	return par;
		case	020:	case	01: /*toggle insert mode*/
			insmode = 1 - insmode;	break;
		case	05:	/* go to end of text in current field*/
			if ((wp->we_type != WPT) && !(wp->we_flags & WSKP) &&
			 ((i = tlen - 1) > 0))
			  {
				for (cp = answer + tlen - 1; (cp >= answer) &&
				 (('\0' == *cp) || (' ' == *cp)); cp--) i--;
				wmove(mwp, (int)wp->we_y,
				 (int)wp->we_x0 + (++i));	wrefresh(mwp);
			  }
			break;
		case	02:	/* beginning of current row */
			return	wp->we_trstart;
		case	'\n':	case	'\r':	/* return: go to next row */
			return	wp->we_nrstart;
		case	'\b':	case	0177:	/* Backspace or delete:*/
			if (wp->we_type == WPT) return wp->we_prevelt;
			if (i)	/* erase last char */
			  {
				i = erasure(wp, mwp, i-1);	break;
			  }
			else if (c == '\b') return wp->we_prevelt;
			else	break;
		case	'?':	/* sometimes special, sometimes not */
			if (i && legalchar((char)c, wp, i)) /* ordinary char */
			  {
				if (0 > (i = actual_char(mp, mwp, par, i,
				 insmode, c))) return (-i - 1);
				break;
			  }
			/* ... otherwise we activate a HELP menu of some sort*/
		case	K_HELP:	return	K_HELP;
		case	'^':	/* sometimes move to the top of the menu */
			if (i && legalchar((char)c, wp, i))
			  {
				if (0 > (i = actual_char(mp, mwp, par, i,
				 insmode, c))) return (-i - 1);
				break;
			  }
			else return mp->wfirstactive; /* move to top of menu */
		case	K_INSERT_HERE:	return K_INSERT_HERE;
		case	K_PREV_SCREEN:	return K_PREV_SCREEN;
		case	'@':	/* sometimes special, sometimes not */
		case	'#':
			if (legalchar((char)c, wp, i))
			  {
				i = actual_char(mp, mwp, par, i,
				 insmode, c); break;
			  }
			else return (c == '@' ? K_INSERT_HERE : K_PREV_SCREEN);
		case	'$':	/* sometimes special, sometimes not */
			if ((wp->we_type != WST) || (tlen <= 1))
				return mp->wlastactive;
			/* ... otherwise this case falls thru to `default' */
		default:	/* this is an actual character! hot doggie! */
			if (0 > (i = actual_char(mp, mwp, par, i, insmode, c)))
				return (-i - 1);
	      }	/* close switch on character type */
	     }	/* close loop over input characters */
	update_par(mp, mwp, par);	return par;
}

/* end of menu.c */
