/* rejmain.c -- XGEN's reject functionality, with Curses capabilities

	Copyright (c) 2003, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: reorganize use of "{"
	23 Jun 2006: old rejdefs.h amalgamated into xgpostint.h, so we use it
	 2 Jun 2006: reorder routines so smoothexi can be static
	31 Dec 2005: handle qsort with < 2 elements correctly
	19 Dec 2005: replace static globals to calls to managerjw
	15 Oct 2005: eliminate many static globals
	 4 Oct 2005: housekeeping
	28 Jun 2005: smoothexi(): typecast pointer differences on printing
	22 Jun 2005: final_matters() is now static
	11 Apr 2005: name changed to rejmain.c again (what goes around ...)
	27 Aug 2004: routines used only here are declared as static; cosmetics
	18 Mar 2004: save_refl has only one arg;
	 slight simplifications in munpack() and xgread() calls
	21 Jan 2004: license message inserted and initial comments reorganized
	 9 Jan 2004: replace globals with elements of REJECTRUN structs
	13 Dec 2003: unneeded vbls origflags, prevflags  elim. from rejmain()
	18 Nov 2003: bugfix in smoothexi(): apply deletions in non-batch
		mode even if no user-selections were done.
	15 Nov 2003: replace globals with REJECTRUN structure elements
	19 Oct 2003: merge rejmenu.c and rejmain.c together to create this file
 Changes to rejmenu.c and rejmain.c prior to that merger:
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	18 Aug 2003: replace global uphkl[3] with static lphkl[3] and
		populate it as needed
	 5 Jan 2003: reorganized source-file header
	15 Nov 2002: extern declarations made here rather than in rejext.h
	25 Mar 2002: removed MSI attachments
	 5 Oct 1994: use xgread to input MERGEDREF data
	23 Sep 1994: use `rejgref' rather than `g_mref' to read reflections in
	 2 Sep 1994: correctly handle case where no deletions occur
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
		additional extern function declarations
	16 Mar 1994: a few bug fixes and some reorganization
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version
	17 May 1993: reorganized so all the `Curses' stuff is in one place.
	15 Jun 1989: incorporate spherical-harmonic scaling conventions.
	29 May 1987: establish "-b" options, wherein automatic deletions
		are performed without resorting to setting up Curses menus.
	15 May 1987: enlarge reflection tank so we won't have to use
		multiple passes through data except in really perverse cases.
	22 Feb 1986: display observations in increasing order of `gofp'
		rather than in an arbitrary order; also, take care of
		situations where very large numbers of observations appear.
		Also: split program up into a bunch of separate files.
	27 Dec 1985: incorporate errors in the scale factors into the
		computations of the intensity variances.
	18 Oct 1985: various minor revisions,
		including computing some statistics as we go along.
	 1 Oct 1985: incorporate the interactive menu into this program
		(previously it had been a program on its own)
	30 Sep 1985: incorporate conventions of the Nicolet system.
	18 Sep 1984: initial coding based on `outliers' written at UCLA.

   Purpose: CURSES menu displays of potentially deletable reflections
	as part of the XGEN "reject" capability.
   Cross references: ../postprocess/rej*.c

 This program flags suspicious or faulty observations from XGEN datasets.
 The algorithm works as follows:
	(1)	Each reflection in a multiref is read in and scaled, and a
		weighted residual, (I-Y)^2/(sigma(I)^2) is computed;
		here Y = <I> is the weighted mean of the scaled intensity
		observations and "^2" means "squared".
	(2)	The residual is compared with a standard to see if the
		observation is potentially suspicious.  The standard value Q is
		a function of the total number of observations in the dataset.
	(3)	If none of the observations for a specific reflection has been
		flagged, we move on to the next reflection.
	(4)	If one or more are flagged, the goodness-of-fit of the
		observations of the reflection to the mean is examined with each
		observation deleted from the population.  If deletion causes a
		significant improvement in the chi-square goodness-of-fit value,
		the observation is flagged for deletion.
	(5)	Suspect reflections with precisely two observations are
		displayed on an interactive menu to allow the user to decide
		whether one observation, both, or none should be deleted.
	(6)	Suspect reflections in which one or more observations had
		bad deviation values but which do not show satisfactory
		chi-squared improvement upon deletion are displayed for
		interactive deletion as well.
	N.B.: If the "-b" (for "batch-mode") flag is invoked, no windows
		are set up.  Instead, the "recommended" deletions are performed
		and the confusing cases are left for another pass.
*************************************************************************/
#define		NEED_TIME	1
#define		NEED_SIGNAL	1
#define		NEED_CURSES	1
#include	"xgpostint.h"
#include	"menu.h"
#include	<setjmp.h>
#define		MAXSHOW		17	/* max #obs displayable @ 1 time*/

 typedef struct {
		jmp_buf		rjw_jbuf;
		WINDOW		*rjw_listw, *rjw_mwp, *rjw_rwin;
		REJECTRUN	*rjw_rjp;
		} REJCWORK;

static REJCWORK *managerjw(int opno)
{ /*	Manage a static pointer to a REJCWORK structure */
	jmp_buf		tjb;
	static REJCWORK	*rjwp;
	static char	rjwmsg[] = "rjwp in managerjw";

	if (NULL != rjwp) cpymem(tjb, rjwp->rjw_jbuf, sizeof (jmp_buf));
	else zeromem((AVOID *)tjb, sizeof (jmp_buf));
	if (!opno)
	  {
		if (NULL != rjwp)
		  {
			rjwp = managerjw(2);	/* Curses menus */
			(void)managerjp(0);	/* rejection formalisms */
			memzap((AVOID **)&rjwp, rjwmsg);
		  }
		return rjwp;
	  }
	if (opno == 2) /* just get rid of the menus */
	  {
		if (NULL != rjwp)
		  {
			if (NULL != rjwp->rjw_listw) delwin(rjwp->rjw_listw);
			rjwp->rjw_listw = NULL;
			if (NULL != rjwp->rjw_mwp) delwin(rjwp->rjw_mwp);
			rjwp->rjw_mwp = NULL;
			if (NULL != rjwp->rjw_rwin) delwin(rjwp->rjw_rwin);
			rjwp->rjw_rwin = NULL;
		  }
		return rjwp;
	  }
	if (1 == opno) /* force clearance of old memory, keeping jbuf */
		rjwp = managerjw(0);
	if ((NULL == rjwp) && (NULL == (rjwp = (REJCWORK *)callop(1,
	 sizeof (REJCWORK), rjwmsg)))) return rjwp;
	if (NULL == (rjwp->rjw_rjp = managerjp(07))) return managerjw(0);
	cpymem(rjwp->rjw_jbuf, tjb, sizeof (jmp_buf));
	return rjwp;
}

static void curse_deletions(void)
{ /* This does some CURSES announcements just before the deletion step begins*/
	REJCWORK	*rjwp;
	REJECTRUN	*rjp;

	if (NULL == (rjwp = managerjw(3))) return;	rjp = rjwp->rjw_rjp;
	if (rjp->rj_drefp == rjp->rj_drefs) /* nothing got flagged! */
	  {
		move(5, 0);	printw(
		 " No deletions in the list: No output multiref written");
		standend();	/* in case it's accidentally still emboldened*/
		refresh();	endwin();	longjmp(rjwp->rjw_jbuf, 1);
	  }
	else {
		move(5, 1);	printw(
		 " Modified multiref %s is being written", trname("mrefout"));
		refresh();
	     }
}

static void curse_prints(void)
{ /*	This splats a few things up on the Curses menu when we are writing
	the flagged reflections to the DELETIONS file instead of writing out
	a modified multiref--just before beginning to write that. */
	REJCWORK	*rjwp;
	REJECTRUN	*rjp;

	if (NULL == (rjwp = managerjw(3))) return;	rjp = rjwp->rjw_rjp;
	if (rjp->rj_drefp == rjp->rj_drefs)
	  {
		mvaddstr(5,0,
	 " No observations flagged: no deletions need to be done!\n");
		addstr(" You must be living right...\n");
		standend();	/* in case it's accidentally still emboldened*/
		refresh();	endwin();	longjmp(rjwp->rjw_jbuf, 1);
	  }
	else
	  {
		move(5, 0);	printw(
	 "Deletable reflections are being printed in %s", trname("deletions"));
		refresh();
	  }
}

static void final_matters(void)
{ /*	This neatly eliminates the curses stuff and then does the
	end-of-program calculations. */
	REJCWORK	*rjwp;
	REJECTRUN	*rjp;

	signal(SIGINT, SIG_DFL);	/* now we change the error catcher */
	signal(SIGBUS, SIG_DFL);
	signal(SIGSEGV, SIG_DFL);
	if (NULL == (rjwp = managerjw(3))) return;	rjp = rjwp->rjw_rjp;
	if ((!(rjp->rj_flags & RJ_BATCH)) && rjp->rj_somedecisions)
	  {
		standend();	/* in case it's accidentally still emboldened*/
		/* no more curses stuff */
		clear();	refresh();	endwin();	setmzap();
	  }
	last_calcs();
}

static void smoothexi(void)
{ /* routine to exit cleanly from this mess */
	char		c;
	int		retval;
	FILE		*lfpl;
	REJCWORK	*rjwp;
	REJECTRUN	*rjp;

	signal(SIGINT, (sig_t)final_matters);	/* change the error catcher */
	signal(SIGBUS, (sig_t)final_matters);
	signal(SIGSEGV, (sig_t)final_matters);	retval = EOK; /*hope it works*/
	if (NULL == (rjwp = managerjw(3))) return;	rjp = rjwp->rjw_rjp;
	 /* always APPLY the deletions if either we're in batch mode or
		if we didn't ask any questions as we went through the data */
	lfpl = rjp->rj_fpl;
	if ((rjp->rj_flags & RJ_BATCH) || !(rjp->rj_somedecisions))
	  {
		if (rjp->rj_drefp != rjp->rj_drefs)
		  {
			if (NULL != lfpl) fprintf(lfpl,
			 "%6d observations have been flagged automatically.\n",
				(int)(rjp->rj_drefp - rjp->rj_drefs));
		  }
		else {
			if (lfpl) fprintf(lfpl,
			 "No reflections were flagged; no output performed\n");
			printf(
			 "No reflections were flagged; no output performed\n");
		     }
		retval = ((NULL == rjp->rj_drefs) || (rjp->rj_drefp ==
		 rjp->rj_drefs)) ?  EOK : apply_deletions();
		if (retval == EOK) retval = 1;
		last_calcs();
		longjmp(rjwp->rjw_jbuf, retval);
	  }
	if (rjp->rj_somedecisions) /* otherwise curses stuff never was made */
	  {
		/* get rid of the separate windows and just use stdscr */
		rjwp = managerjw(2);
		clear();	refresh();
		standend(); 		/* ...so final msg isn't boldface */
		move(1, 1);	printw(
	 "%6d observations have been flagged either manually or automatically.",
			rjp->rj_drefp - rjp->rj_drefs);
		if (lfpl) fprintf(lfpl,
 "%6d observations have been flagged either manually or automatically.\n",
			(int)(rjp->rj_drefp - rjp->rj_drefs));
		mvaddstr(2,1," Do you want to APPLY the deletions     (a,A),");
		mvaddstr(3,1,"                PRINT the deletion list (p,P),");
		mvaddstr(4,1,
		 "            or  BIT-BUCKET the list     (b,B) ? ");
		refresh();
		while (1)
		     {
			c = wget1ch(stdscr);
			if ('a' <= c && c <= 'z') c -= 'a' - 'A';
			if ('A' == c || 'P' == c || 'B' == c) break;
			move(4, 49);
		     }
		addch((xctype)(c));	refresh();
		switch (c)
		      {
			case	'A': curse_deletions();
				retval = apply_deletions();	break;
			case	'P': curse_prints();
				retval = print_deletions();	break;
			case	'B': mvaddstr(4,1,
					"You chose to throw away the list\n");
				retval = EOK;	sleep(2);	break;
		      }
	  }
	final_matters();	if (retval == EOK) retval = 1;
	longjmp(rjwp->rjw_jbuf, retval);
}

static void intexit(void)
{ /* warns that we're quitting early, then it does it */
	FILE		*lfp;
	static REJCWORK	*rjwp;

	if ((NULL != (rjwp = managerjw(3))) && (NULL !=
	 (lfp = (rjwp->rjw_rjp)->rj_fpl))) fprintf(lfp,
	 "premature termination of `reject' due to user interrupt\n");
	smoothexi();
}

static void refdisplay(void)
{ /*	This displays the reflection on the window. */
	REJCWORK	*rjwp;
	REJECTRUN	*rjp;
	WINDOW		*rwin;
	MERGEDREF	*mrp;

	if (NULL == (rjwp = managerjw(3))) return;
	rjp = rjwp->rjw_rjp;	rwin = rjwp->rjw_rwin;	mrp = rjp->rj_mrp;
	(void)get_deleted_sums(); /* get current g.o.f. values */
	if (rjp->rj_flags & RJ_BATCH) return;
		/* wipe out the silly "good book" message if it's there */
	wmove(rwin, 1, 0);	wclrtoeol(rwin);
	wstandend(rwin);	/* no boldface for headers for categories */
	mvwaddstr(rwin, 0, 0,
"   h  k  l   n shift omega   x   y         I    sigma    gof worst wtd.dev.\n");
	wstandout(rwin);	/* print reflection info in boldface */
	munpack(mrp->mr_hkl, rjp->rj_lhkl, (rjp->rj_xhdp)->xt_intl);
  	wprintw(rwin,
	 " %3d%3d%3d: sth/l =%6.4f,%3d obs: %9.2f %8.2f%7.2f %14.2f\n",
	 rjp->rj_lhkl[0], rjp->rj_lhkl[1], rjp->rj_lhkl[2], mrp->mr_stl,
	 rjp->rj_nobused, rjp->rj_ibar, rjp->rj_sig, rjp->rj_gof,
	 sqrt(rjp->rj_worst));
	wprintw(rwin,
	 "   after making deletions: %2d obs: %9.2f %8.2f%7.2f %14.2f",
		rjp->rj_nobaf, rjp->rj_ibaraf, rjp->rj_sigaf,
		rjp->rj_gofaf, sqrt(rjp->rj_worstaf));
	wrefresh(rwin);	/* put it on the dern screen */
}

static void obdisplay(void)
{ /*	This displays the observations.	*/
	register int		i, n2dis;
	register WORKINGOB	*wobp;
	REJCWORK		*rjwp;
	REJECTRUN		*rjp;
	WINDOW			*mwp;

	if (NULL == (rjwp = managerjw(3))) return;
	rjp = rjwp->rjw_rjp;	mwp = rjwp->rjw_mwp;
	if (rjp->rj_flags & RJ_BATCH) return;
	wstandend(mwp);	/* explicitly set it at no-standout */
	n2dis = (rjp->rj_nobused < MAXSHOW) ? rjp->rj_nobused : MAXSHOW;
	for (wobp = rjp->rj_wobs, i = 0;
	 (wobp < rjp->rj_lastdis) && (i < n2dis); wobp++)
	   {
		wmove(mwp, i+1, 1);	wprintw(mwp,
		 "%11s%2d %5d %5d %3d %3d %9.2f %8.2f%7.2f ?: %11s",
		 anyob_display(wobp->w_flag), i, wobp->w_shift,
		 wobp->w_omob, wobp->w_xob, wobp->w_yob, wobp->w_isc,
		 wobp->w_sigsc, wobp->w_gofp,marker_display(wobp->w_flag));
		i++;
	   }
	box(mwp, (xctype)('|'), (xctype)('-')); /* box around obs. window */
}

static void command_list(int oldwin)
{ /*	This sets up and displays a summary of the available commands,
	and then restores the old windows if they have anything in them.
	The argument "oldwin" is 1 if there's anything in the old windows;
	it's 0 otherwise. */
	REJCWORK		*rjwp;
	REJECTRUN		*rjp;
	WINDOW			*listw, *mwp;

	if (NULL == (rjwp = managerjw(3))) return;
	rjp = rjwp->rjw_rjp;
	listw = rjwp->rjw_listw;	mwp = rjwp->rjw_mwp;
	if (rjp->rj_flags & RJ_BATCH)
	  {
		refdisplay();	return;
	  }
	wstandout(listw);	/* make header and funcs stand out */
	mvwaddstr(listw, 0, 6,
		"Commands Available in the Multiref Deletion Program");
 	mvwaddstr(listw,1, 1,"d,y");	mvwaddstr(listw,1,36,"t,[PF4]");
	mvwaddstr(listw,2, 1,"u,n");	mvwaddstr(listw,2,36,"h,r,[PF1],[<-]");
	mvwaddstr(listw,3, 1,"e,q,^C");	mvwaddstr(listw,3,36,"k,[UP]");
	mvwaddstr(listw,4, 1,"l,n,[->]");mvwaddstr(listw,4,36,"i,[PF3]");
	mvwaddstr(listw,5, 1,"c,[PF2]");
	mvwaddstr(listw,5,36,"anything else");
	wstandend(listw);
	mvwaddstr(listw,1,10,": mark ob for deletion  ");
	mvwaddstr(listw,1,51,": Toggle delete stat of ob    ");
	mvwaddstr(listw,2,10,": Undelete observation  ");
	mvwaddstr(listw,2,51,": Redo this reflection        ");
	mvwaddstr(listw,3,10,": Exit the program      ");
	mvwaddstr(listw,3,51,": go up to the previous ob");
	mvwaddstr(listw,4,10,": go on to the next refl");
	mvwaddstr(listw,4,51,": IMMEDIATELY to next refl    ");
	mvwaddstr(listw,5,10,": get this Command list ");
	mvwaddstr(listw,5,51,": go on to the next ob");
	wrefresh(listw);	/* put up what we just asked for */
	if (oldwin)
	  {
		refdisplay();
		obdisplay();	wmove(mwp,1,64);	wrefresh(mwp);
	  }
}

static WORKINGOB *ask_for_review(void)
{
	char		answer;
	REJCWORK	*rjwp;
	REJECTRUN	*rjp;
	WINDOW		*mwp;

	if (NULL == (rjwp = managerjw(3))) return NULL;
	rjp = rjwp->rjw_rjp;	mwp = rjwp->rjw_mwp;
	wstandout(mwp); /* give this message in boldface */
	mvwaddstr(mwp, rjp->rj_nobused + 1, 1, "review? ");
	wstandend(mwp);	/* turn off boldface */
	wrefresh(mwp);	/* update ye olde screene */
	if (((answer = wget1ch(mwp)) == 'y') || (answer == 'Y'))
	  { /* user has said 'y' or 'Y': review it. */
		box(mwp, (xctype)('|'), (xctype)('-'));	/*restore munged box*/
		return rjp->rj_wobs;	/* go back to top of observation */
	  }
	return	(rjp->rj_lastdis + 1);
}

static WORKINGOB *act_on_command(sichar answer, WORKINGOB *wobin)
{ /*	This acts on the user's command.  It returns a pointer to
	the NEXT observation to be processed. */
	u_char		obstate;
	REJCWORK	*rjwp;
	REJECTRUN	*rjp;
	WINDOW		*mwp, *rwin;
	MERGEDREF	*mrp;

	if (NULL == (rjwp = managerjw(3))) return NULL;
	rjp = rjwp->rjw_rjp;	mwp = rjwp->rjw_mwp;	rwin = rjwp->rjw_rwin;
	mrp = rjp->rj_mrp;
	if (rjp->rj_flags & RJ_BATCH) return rjp->rj_lastdis + 1;
	obstate = wobin->w_flag & ~W_ISMARKED;
	switch (answer)
	      {
		case	K_PF2:
		case	'c':	/* command list summary */
			command_list(1);	return wobin;
		case	'd':	/* delete this observation */
		case	'y':
			/* take this observation out of the summations */
			if (!(wobin->w_flag & W_ISMARKED))
			  {
				(rjp->rj_nobaf)--; use_deleted_sums(wobin);
				wobin->w_flag = obstate | W_HANDMARKED;
			  }
			refdisplay();	obdisplay();	break;
		case	'n':	/* undelete this observation */
		case	'u':
			/* put this observation back into the summations */
			if (wobin->w_flag & W_ISMARKED) redo_ob(wobin);
			refdisplay();	obdisplay();	break;
		case	K_PF4:
		case	't':	/* toggle deletion state of refl */
			if (wobin->w_flag & W_ISMARKED) redo_ob(wobin);
			else
			  {
				(rjp->rj_nobaf)--; use_deleted_sums(wobin);
				wobin->w_flag = obstate | W_HANDMARKED;
			  }
			refdisplay();	obdisplay();	break;
		case	K_UP:
		case	'k':	/* go up to previous usable observation */
				/* this command DOES wrap around now... */
			return	(wobin == rjp->rj_wobs ?
				rjp->rj_lastdis - 1 : wobin - 1);
		case	K_PF1:
		case	K_RIGHT:
		case	K_KP5:
		case	's':	/* skip rest of reflection; */
		case	'l':	/* ... but give the user another chance*/
			return (ask_for_review());
		case	K_PF3:
		case	'i':	/* go immediately to the next refl */
			return	(rjp->rj_lastdis + 1);
		case	K_LEFT:
		case	'r':	/* redo this reflection */
		case	'h':
			return rjp->rj_wobs;
		case	'q':	/* update deletions file and */
		case	'e':	/* quit immediately */
#ifndef VMS
			wclear(mwp);	wrefresh(mwp);
			wclear(rwin);	wrefresh(rwin);
#endif
			(void)save_refl();
			while (0 < xgread((void *)mrp, XG_MERGEDREF,
			 rjp->rj_fpmi))
			     {
				input_means(mrp);
				if (rjp->rj_e0 > FUZZ && rjp->rj_ibar > FUZZ)
				  {
					rjp->rj_rfacout += rjp->rj_e2 -
					 rjp->rj_e1 * (rjp->rj_e1 / rjp->rj_e0);
					rjp->rj_rfdenout += rjp->rj_e2;
					rjp->rj_onobout += rjp->rj_nobused;
				  }
			     }
			rjp->rj_flags |= RJ_ALLCALC;
			smoothexi();	return NULL;
		default:	/* anything else: go on to next ob */
			break;
	     }
/* anytime we get here, we go on to the next observation unless
 we're already on the last one, in which case we go to the first one. */
	return (wobin == rjp->rj_lastdis - 1 ? rjp->rj_wobs : wobin + 1);
}

static int decisions(MERGEDREF *mrp)
{ /*	This is where the reflection is displayed and the user gets a chance
	to decide what to do with it.  The routine returns the output from
	save_refl.  Note that the setup for CURSES occurs in here if it hasn't
	already happened.  */
	register int		i, n2dis;
	register WORKINGOB	*wobp, *wobnew;
	sichar			c;
	size_t			nelts;
	REJCWORK		*rjwp;
	REJECTRUN		*rjp;
	WINDOW			*mwp, *rwin;

	if (NULL == (rjwp = managerjw(3))) return 0;
	rjp = rjwp->rjw_rjp;
	rwin = rjwp->rjw_rwin;
	if (rjp->rj_flags & RJ_BATCH) return 0; /* shouldn't occur inside here*/
	 /*sort the observations in decreasing order of the gofp.*/
	if (NULL == rwin)	/* set up CURSES stuff if needed */
	  {
#ifndef	VMS	/* define some terminal-dependent keys if needed */
		verror(setmore(0),
			"set up keystroke equivalences for this terminal");
#endif
		sleep(1);	/* so diagnostic messages are visible */
		/* reserve memory for and set up the window structures */
		initscr();	/* initialize screen-manipulation stuff */
		crmode();	/* set terminal to char-by-char mode */
		noecho();	/* don't echo input back */
		nonl();		/* turn off newline mapping for optimization */
		rjwp->rjw_rwin = rwin = newwin(3,79,0,0); /* window for refl */
		rjwp->rjw_listw = newwin(6,80,15,0);	/* window for obs */
		command_list(0);	/* display commands one can perform */
	  }
	rjp->rj_somedecisions = 1;	/* flag: we've made some decisions */
	nelts = rjp->rj_wobend - rjp->rj_wobs;
	if (nelts > 1) qsort((AVOID *)rjp->rj_wobs, nelts,
		sizeof (WORKINGOB), (QPE)wobcmp);
	n2dis = (rjp->rj_nobused < MAXSHOW ? rjp->rj_nobused : MAXSHOW);
	rjp->rj_lastdis = rjp->rj_wobs + n2dis;
	refdisplay();	/* display the reflection */
	i = n2dis + 2;
	rjp->rj_mrp = mrp;
	if (NULL == (mwp = newwin(i, 79, 3, 0))) /* create the window */
	  {
		munpack(mrp->mr_hkl, rjp->rj_lhkl, (rjp->rj_xhdp)->xt_intl);
		fprintf(rjp->rj_fpl,
		 "Premature exit due to null mwp with hkl=%3d,%3d,%3d\n",
		 rjp->rj_lhkl[0], rjp->rj_lhkl[1], rjp->rj_lhkl[2]);
		smoothexi();	return 020;
	  }
	rjwp->rjw_mwp = mwp;
	obdisplay();		/* display the observations */
	/* Go back to the top of the screen and ask the user what he
	wants to do with each of these observations. */
	for (i = 1, wobp = rjp->rj_wobs; i < MAXSHOW; wobp = wobnew)
	   {
		wmove(mwp,i,64); /* position the cursor */
		wrefresh(mwp);	/* update screen now */
		c = wget1ch(mwp);	/* go get a character of response */
#ifndef VMS
		if (ESCAPE == c) c = getescseq(mwp);
#endif
		if (NULL == (wobnew = act_on_command(c, wobp))) /* do it! */
			return 040;
		mvwaddstr(mwp, i, 64, marker_display(wobp->w_flag));
		if (wobnew > rjp->rj_lastdis) break;
		/* compute the cursor position for the next pass */
		i = (wobnew - rjp->rj_wobs) + 1;
	   }
/* erase and delete the observation WINDOW so we can make a different-sized one
	for the next reflection.  To save time, we erase only the lines
	printed on if it's a short one; if it's a long one we clear it.*/
	if (n2dis > 8) wclear(mwp);
	else {
		for (i = 0; i < n2dis + 4; i++)
		   {
			wmove(mwp, i, 0);	wclrtoeol(mwp);
		   }
	     }
	for (i = 0; i < 3; i++)
	   {
		wmove(rwin, i, 0);	wclrtoeol(rwin);
	   }
	wrefresh(mwp);	wrefresh(rwin);
	delwin(mwp);	rjwp->rjw_mwp = NULL;
	rjp->rj_tl1 = NOW;	/* record time when window was erased */
	return save_refl();
}

static void overload_quit(void)
{ /* this routine is called when we've used up all the memory
	reserved for the deletable observations */
	REJCWORK	*rjwp;
	REJECTRUN	*rjp;
	WINDOW		*mwp;

	if (NULL == (rjwp = managerjw(3))) return;
	rjp = rjwp->rjw_rjp;	mwp = rjwp->rjw_mwp;
	if (rjp->rj_flags & RJ_BATCH)
	  {
		smoothexi();	return;
	  }
	if (NULL != mwp)
	  {
		delwin(mwp);	rjwp->rjw_mwp = NULL;
	  }
#ifndef VMS
	clear();
#endif
	move(6, 0);
	addstr("There's no more room to put deletable observations in\n");
	printw(
	 "The last reflection you were processing was [%3d,%3d,%3d]\n",
		rjp->rj_lhkl[0], rjp->rj_lhkl[1], rjp->rj_lhkl[2]);
	addstr(
	 "You can resume at that point in a subsequent `reject' run\n");
	addstr("Hit any key to go on to the final processing pass\n");
	refresh();
	(void)wget1ch(stdscr);
	smoothexi();
}

static void show_working(void)
{ /*	This blaps a message onto the screen telling the user that the
	program is still chugging away */
	REJCWORK	*rjwp;
	WINDOW		*rwin;
	
	if (NULL == (rjwp = managerjw(3))) return;	rwin = rjwp->rjw_rwin;
	wmove(rwin, 0, 0);	wclrtoeol(rwin);
	wmove(rwin, 1, 0);	wclrtoeol(rwin);
	mvwaddstr(rwin, 1, 0,
" This will take a little while.  Find a good book or something             ");
	wrefresh(rwin);
}

int rejmain(int argc, char *argv[])
{ /* The main line for curses-based outlier rejection */
	int		i, nrexam, nfiddled, ndecisions;
	int		procres, procres0, procres1;
	MERGEDREF	mrpri, mrbij, *mrq;
	REJCWORK	*rjwp;
	REJECTRUN	*rjp;

	if (NULL == (rjwp = managerjw(3))) return EALLOC; rjp = rjwp->rjw_rjp;
	/* to begin with, check for a jmp_buf return */
	if ((i = setjmp(rjwp->rjw_jbuf)))
	  {
		if (i == 1)
		  {
			i = rejdie(EOK);	managerjw(0);	return i;
		  }
		else {
			managerjw(0);	return i;
		     }
	  }
	/* decode arguments for files and flags */
	if (EOK != (i = rejsetup(argc, argv))) return rejdie(i);
	signal(SIGINT, (sig_t)intexit); /* smooth-exit routine on interrupt */
	if (rjp->rj_flags & RJ_RESUME) /* skip early refls if in "resume" mode*/
	  while ((0 != rejgref()) &&
	    (hklcmp((rjp->rj_mrp)->mr_hkl, rjp->rj_shkl) < 0))
	       {  /* ...but do R-factor stuff anyway */
		input_means(rjp->rj_mrp);
		if (rjp->rj_ibar > FUZZ && rjp->rj_e0 > FUZZ)
		  {
			rjp->rj_rfacout += rjp->rj_e2 -
				rjp->rj_e1 * rjp->rj_e1 / rjp->rj_e0;
			rjp->rj_rfdenout += rjp->rj_e2;
			rjp->rj_onobout += rjp->rj_nobused;
		  }
	       }
	nrexam = nfiddled = ndecisions = 0;
	while (0 != rejgref())	/* read data that the user wants */
	     { /* we set rjp->rj_fpl to NULL when we get to `final_matters',
		 so that's an indication that we interrupted things with ^C. */
		if (NULL == rjp->rj_fpl) return rejdie(EOK);
		mrq = rjp->rj_mrp;
		if ((rjp->rj_flags & RJ_SEPBJ) &&
		 (mrq->mr_nobpri) && (mrq->mr_nobbij))
		  {
			splitmr(mrq, &mrpri, &mrbij);
			if (04 & (procres0 = refl_process(&mrpri)))
			  {
				ndecisions++; procres0 = decisions(&mrpri);
			  }
			procres1 = 0;
			if ((!(010 & procres0)) &&
			  (04 & (procres1 = refl_process(&mrbij))))
			  {
				ndecisions++; procres1 = decisions(&mrpri);
			  }
			procres = procres0 | procres1;
		  }
		else if (04 & (procres = refl_process(mrq)))
		 {
			ndecisions++; procres = decisions(mrq);
		  }
			/* user asked for exit */
		if (procres & 040)
		  {
			i = rejdie(EOK); (void)managerjw(0); return i;
		  }
			/*problem with mwp*/
		if (procres & 020)
		  {
			i = rejdie(EOPENERR);	(void)managerjw(0); return i;
		  }
		if (procres & 010)
		  {
			overload_quit(); i = rejdie(EOK);
			(void)managerjw(0);	return i;
		  }
		else if (procres & 04) show_working();
		nrexam++;	if (procres) nfiddled++;
	     }
		/* flag: we did R-factor calcs on ALL refls */
	rjp->rj_flags |= RJ_ALLCALC;	smoothexi();
	(void)rejdie(EOK);	(void)managerjw(0);	return EOK;
}

/* end of rejmain.c */
