/* set_up.c -- Curses-based menu system for program operation in X-GEN

	Copyright (c) 2003, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: reorganize use of "{"
	10 Jun 2006: housekeeping, e.g. replacing some tabs with spaces
	16 Oct 2005: collapse static globals into SETUPWORK structure
	16 Aug 2005: DO_EXIT, etc. moved from ../include/xgen.h
	18 Jun 2005: xgaccess() has verbosity arg; get_fmlimits has fromcur arg
	11 May 2005: check return status from xgjournal
	11 Apr 2005: name of this file changed to set_up.c to avoid overlap
	22 Oct 2004: all labels are <= 18 chars, simplifying formatting;
	 help text files have names that end in .txt now
	15 Oct 2004: use internal Curses formatting for help displays;
		tighten up toplevel displays so everything fits
	27 Aug 2004: routines used only here are declared as static; cosmetics
	27 Mar 2004: in stuf2menu(), use xmm_miscb[1024] as start of scratch
	29 Feb 2004: getproject() call supersedes a lot of code in setupmain
	28 Feb 2004: get_xgdoc() now returns the directory it seeks,
		so we call it in a different way
	31 Jan 2004: environment variables are not chingered by the
		shortening of their names; hitting "E" after a command works
	24 Jan 2004: timestamp variables made local and namb structure used
		to hold some other text buffers; dangerous redefinition
		of ptp in progargs() eliminated
	21 Jan 2004: license message inserted and initial comments reorganized
	 7 Jan 2004: replace globals with XGNAME & FRAMEINDEX struct elements
	13 Dec 2003: elim. unneeded vbls accv from canuse(),
		isset, c from chg_filemenu(); add one more test to canuse()
	 7 Dec 2003: eliminate references to labels[] and defaults[]
	25 Nov 2003: load the PMENUSTUFF structures with call to pmall()
		rather than in this file
	20 Oct 2003: merge setupglob.c,setupprog.c,setupcur.c,setupmain.c 
 Changes prior to that date for the individual files:
	19 Sep 2003: initialize arrays according to rules for ansi C braces
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h;
		eliminate declaration of an unused routine
	16 Sep 2003: enable gunzipping by replacing access() calls with
		xgaccess() calls
	22 Aug 2003: modify use of glistname() and the listname variable
		according to the changes in xgenviron.c;
		stick the setup-menu labels in downstream of listname,
		rather than downstream of scrtname.
	13 Aug 2003: remove `fped' (now only static in one routine)
	10 Jul 2003: search for X-GEN documentation by call to get_xgdoc()
	 5 Jan 2003: reorganized source-file header
	25 Mar 2002: removed MSI attachments
	 5 Jan 2003: reorganized source-file header;
		update displayed copyright notice
	15 Nov 2002: extern declarations made here rather than in setupext.h
	10 Nov 2002: replaced line[] with nabuf[] to prevent name overlap
	25 Mar 2002: removed MSI attachments
	xx xxx 2000: (?) add canuse() as a way of checking file availability
	15 Oct 1998: in setupmain, override the old names for the
		individual files in WORK, DATA, and SCRT in this call.
	 1 Dec 1994: some globals moved to xgenglob
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: a few global variables moved around
	14 Jan 1994: initial X-GEN version 

   Purpose: initializing an X-GEN project and invoking individual applications
   Cross references: progglob.c

 "setup" is a general-purpose bookkeeping program running under Curses.
 The program allows the user to set logical names, define a frame index,
 manipulate those setup lists, and fork to the other XGEN programs.

 Routines that actually invoke CURSES are:
	endsetmenu	closes down the Curses-specific stuff and exits
	setupvals	gets reasonable starting values for the
			 entries in the filename-and-frame-index menu
	start_curses	creates the curses structures and calls the
			 appropriate initialization routines
	setuptop	displays the top-level (program-choice) menu
	look_topcmd	processes user input in the top-level menu;
			 decides which program the user has asked to jump to
	setup_prog	displays one of the second-level menus
	cvt_fixed	converts a fixed WINDOW_TEMPLATE to a WINDOW_ENTRY
	make_cmd	generates a command line from the associated
			 WINDOW_ENTRY and PMENUSTUFF structures
	stuf2menu	converts a PMENUSTUFF structure to a WINDOW_ENTRY
	do_filecmd	processes a command to the filename-and-index menu
	do_progmenu	processes user input in the lower-level menu.
			 Depending on its argument, it may manipulate a help
			 screen, fiddle with a names-and-files menu, or fiddle
			 with a menu specific to a particular XGEN program.
			 In the last case, if the user has "pushed" one of the
			 buttons associated with starting the program up,
			 it calls routines to fork to that program.
*************************************************************************/
#define	NEED_DESCRIP	1
#define	NEED_CURSES	1
#define	NEED_SIGNAL	1
#define	NEED_ERRNO	1
#define	NEED_STAT	1
#define	NEED_TIME	1
#include	"xgen.h"
#include	"menu.h"
#define		MAXPMENUS	28
#define		FILEWID		25
#define		SWWID		 9
#define		PARAMWID	11
#define		DO_FILES	0
#define		CM		26
#define		RM		48
#define		DOCSUFFIX	"txt"
#define		DO_EXIT		(-1000)	/* code: exit from GUI */
#define		DO_TOPMENU	(-999)	/* code: return to toplevel of GUI */
#define		DO_HELP		(-998)	/* code: produce help information */

 typedef struct {
		PMENUSTUFF	*se_ptp, **se_pmlist;
		int		se_cmdpar, se_exbutt;
		int		se_nentsf, se_npgms, se_topbutt;
		FRAMEINDEX	*se_fmi;
		WINDOW		*se_mwp;
		MENUDATA	*se_mprog, *se_mprevprog;
		MENUDATA	*se_mtopp, *se_mtprevp;
		} SETUPWORK;

#ifdef VMS
extern int	system(char *str);
#endif

static SETUPWORK *managesep(int opno)
{ /*	Management of a static pointer to a SETUPWORK structure */
	static SETUPWORK	*sep;
	static char		sepmsg[] = "sep in managesep";
	static char		pmlmsg[] = "sep->se_pmlist in managesep";
	static char		fmimsg[] = "sep->se_fmi in managesep";
	static char		mtopwmsg[] = "mtopp->wen in managesep";
	static char		mtopmsg[] = "mtopp in managesep";
	static char		mtpwmsg[] = "mtprevp->wen in managesep";
	static char		mtpmsg[] = "mtprevp in managesep";

	if (opno == 0)
	  {
		if (NULL != sep)
		  {
			memzap((AVOID **)&(sep->se_pmlist), pmlmsg);
			memzap((AVOID **)&(sep->se_fmi), fmimsg);
			memzap((AVOID **)&((sep->se_mtopp)->wen), mtopwmsg);
			memzap((AVOID **)&(sep->se_mtopp), mtopmsg);
			memzap((AVOID **)&((sep->se_mtprevp)->wen), mtpwmsg);
			memzap((AVOID **)&(sep->se_mtprevp), mtpmsg);
			memzap((AVOID **)&sep, sepmsg);
		  }
		return NULL;
	  }
	if (opno == 1) sep = managesep(0);
	if ((NULL == sep) && (NULL == (sep = (SETUPWORK *)callop(1,
	 sizeof (SETUPWORK), sepmsg)))) return NULL;
	return sep;	/* more later */
}

/*	The next few routines are associated with disemboweling a command like
		reduce -exd4 -f-3 -w1 -d4 1 0 99 3.4
	such that the outcome is a set of specific Booleans, switches, and
	parameters that are stored in the mep->wen structures. */
static int endcmp(char *stra, char *strb)
{ /*	This compares the two arguments as strings, but if one is
	longer than the other, it compares the end of the longer one
	with all of the shorter one. */
	int	lena, lenb;

	lena = strlen(stra);	lenb = strlen(strb);
	if (lena <= lenb)	return strcmp(stra, strb + (lenb - lena));
	else			return strcmp(stra + (lena - lenb), strb);
}

static int progargs(MENUDATA *mep, int azc, char *azv[], PMENUSTUFF **pdp)
{ /*	This converts the args `azv' into the curses menu entries needed */
	int		i, j;
	char		*valstr, **dcp, put;
	WINDOW_ENTRY	*wcp;
	XMSHOW		*pxp;

	/* get to the starting point of the WINDOW_ENTRY structures */
	wcp = &(mep->wen[1]);
	if ((*pdp)->pr_nfile) wcp += (*pdp)->pr_nfile + 1;
	if ((*pdp)->pr_nbool) wcp++;
	pxp = &((*pdp)->pr_show[(*pdp)->pr_nfile]);
	for (i = 0; i < (*pdp)->pr_nbool; i++, wcp++, pxp++)
	   {
		if (findbool(azc, azv, pxp->xs_defau[0])) /* Boolean found */
			put = (pxp->xs_defau[1] == 'Y') ? 'Y' : 'y';
		else	put = (pxp->xs_defau[1] == 'N') ? 'N' : 'n';
		*(wcp->we_string) = put; wcp->we_string[1] = '\0';
	   }
	if ((*pdp)->pr_nsw) wcp++;
	for (i = 0; i < (*pdp)->pr_nsw; i++, wcp++, pxp++)
	   {
		if (NULL != (valstr = findsw(azc, azv, pxp->xs_defau[0])))
			strcpy(wcp->we_string, valstr);
		else *(wcp->we_string) = '\0';
	   }
	for (dcp = azv + 1, j = 1; j < azc; j++, dcp++)
	   if (**dcp != '-') break;
	if ((j < azc) && ((*pdp)->pr_nparam))
	   {
		wcp++;
		for (i = 0; i < (*pdp)->pr_nparam; i++, wcp++, pxp++)
		   {
			if ((j < azc) && (NULL != *dcp)  && ('\0' != **dcp))
			  {
				strcpy(wcp->we_string, *dcp); j++; dcp++;
			  }
			else *(wcp->we_string) = '\0';
		   }
	  }
	return EOK;
}

static int cmd2strs(MENUDATA *mep, char  *cmd)
{ /*	This converts a command string into individual words */
	int		azc, i;
	char		*verb, **azv;
	SETUPWORK	*sep;
	PMENUSTUFF	**pdp;

	if ((NULL == cmd) && ('\0' == *cmd)) return EBADVALUE;
	(void)xgprogname(NULL, 2); /* hide old program name */
	/* now convert the command string to an argc-argv combination. */
	if (NULL == (azv = str2acav(cmd, &azc))) return EALLOC;
	/* now go through the PMENUSTUFF structures and see
	 if the command matches any of those we know about. */
	verb = azv[0];
	if (NULL == (sep = managesep(3))) return EALLOC;
	for (pdp = sep->se_pmlist, i = 0; i < sep->se_npgms; i++, pdp++)
	   {
		if (!endcmp(verb, (*(pdp))->pr_name)) /* match found */
			return progargs(mep, azc, azv, pdp);
	   }
	return EBAD;	/* we only get here on an unknown command */
}

static void endsetmenu(void)
{ /*	This is called just before exiting.  */
	
	SETUPWORK	*sep;
	WINDOW		*mwp;
	
	if (NULL == (sep = managesep(3))) return;
	signal(SIGINT,SIG_DFL);	mwp = sep->se_mwp;
	if (NULL != mwp)
	  {
		wmove(mwp, 21, 0);	/* move to the bottom */
		wclrtoeol(mwp);	/* clean up that line */
#ifndef VMS
		waddstr(mwp, " ");	/* to make that actually take effect */
		wrefresh(mwp);	wclear(mwp);
#endif
		wrefresh(mwp);
	  }
	endwin();		/* turn off the whole menu thing */
	setmzap();	/* get rid of memory obtained in `setmore' */
	memzap((AVOID **)&(sep->se_pmlist), "pmlist in setup");
	if (NULL != sep->se_mtopp)
	  {
		memzap((AVOID **)&((sep->se_mtopp)->wen),"mtopp->wen in setup");
		memzap((AVOID **)&(sep->se_mtopp), "mtopp in setup");
	  }
	if (NULL != sep->se_mtprevp)
	  {
		memzap((AVOID **)&((sep->se_mtprevp)->wen),
		 "mtprevp->wen in setup");
		memzap((AVOID **)&(sep->se_mtprevp), "mtprevp in setup");
	  }
	menufree(sep->se_mprog);	menufree(sep->se_mprevprog);
	globalkill();	/* get rid of various other chunks of memory */
}

static void setupvals(MENUDATA *mpr)
{ /*	This sets the starting values of the variables in the
	names-and-frame index menu. */
	char		*cp;
	int		iva;
	FRAMEINDEX	*lfmi;
	SETUPWORK	*sep;

	if ((NULL == (cp = getenv("RUN_ID"))) || ('\0' == *cp)) return;
	strcpy(mpr->wen[2].we_string, cp);
	if ((NULL != (cp = getenv("WORK"))) && ('\0' != *cp))
		strcpy(mpr->wen[5].we_string, cp);
	if ((NULL != (cp = getenv("DATA"))) && ('\0' != *cp))
		strcpy(mpr->wen[6].we_string, cp);
	if ((NULL != (cp = getenv("SCRT"))) && ('\0' != *cp))
		strcpy(mpr->wen[7].we_string, cp);
	if (NULL != (cp = get_xgdoc(1))) strcpy(mpr->wen[8].we_string, cp);
	/* grab the filename template and related info from where they're
	 hidden */
	if (NULL != (cp = grab_template(1))) strcpy(mpr->wen[10].we_string,cp);
	if (0 < (iva = grab_base()))
		sprintf(mpr->wen[11].we_string, "%3d", iva);
	if (0 < (iva = grab_interval()))
		sprintf(mpr->wen[12].we_string, "%3d", iva);
	 /* try to open the frame index to estimate the frame limits */
	if (NULL == (sep = managesep(3))) return;
	sep->se_fmi = lfmi = ftables(15);
	sprintf(mpr->wen[13].we_string, "%7d", (int)(lfmi->fi_firstfm));
	sprintf(mpr->wen[14].we_string, "%7d", (int)(lfmi->fi_lastfm));
}

static void cvt_fixed(WINDOW_ENTRY *wep, WINDOW_TEMPLATE *wtp, int cmdtoo)
{ /*	This transfers information for a fixed variable from
	 a WINDOW_TEMPLATE structure into a WINDOW_ENTRY structure.
	 WARNING: don't use this on a non-fixed string! Note that
	 the strlen and lablen fields are measured rather than estimated */
	wep->we_x0 = wtp->wt_x0;		wep->we_y = wtp->wt_y;
	if (cmdtoo) wep->we_strlen = strlen(wtp->wt_string);
	wep->we_lablen = strlen(wtp->wt_label);
	wep->we_type = wtp->wt_type;	wep->we_delim = wtp->wt_delim;
	wep->we_flags = wtp->wt_flags;
	wep->we_afterdot = wtp->wt_afterdot;
	/* ALL HELL WILL BREAK LOOSE IF THESE STRINGS ARE EVER ALTERED!!! */
	if (cmdtoo) wep->we_string = wtp->wt_string;
	wep->we_label = wtp->wt_label;	wep->we_post = wtp->wt_post;
}

static int start_curses(void)
{ /* This gets the CURSES package going.  It returns 0 if everything's ok */
	time_t		tl0;
	static char	setdate[26];
	SETUPWORK	*sep;

	if (NULL == (sep = managesep(3))) return EALLOC;
#ifndef VMS	/* set up symbols for menus */
	if (verror(setmore(0),
	 "set up keystroke equivalences for this terminal")) return EOPENERR;
#endif
	tl0 = NOW;
	strcpy(setdate, ctime(&tl0));	/* record start of menu work */
	/* smooth handling of the common interrupting signals */
	signal(SIGINT, (sig_t)endsetmenu);
	initscr();			/* initialize screen stuff */
	crmode();			/* char-by-char mode */
	noecho();			/* don't echo input back */
	nonl();				/* turn off newline mapping */
	if (qerror((AVOID *)(sep->se_mwp = newwin(22, 80, 0, 0)),
		"create CURSES start-up window")) return EOPENERR;
	clear();	wrefresh(stdscr);	/* clear the screen */
	return EOK;
}

static MENUDATA *maketop(int *nentstp)
{ /*	This creates a top-level menu structure and returns it.
	It gets its text strings from the PMENUSTUFF structures. */
	int		i, curx, cury;
	MENUDATA	*mep;
	WINDOW_ENTRY	*wp;
	PMENUSTUFF	**pdp;
	SETUPWORK	*sep;
	static WINDOW_TEMPLATE	topfixed[] = {
 { "", "setup", " logicals & frame index  ", ' ', CM, 1, 32, 0, 0, 0, WPT},
 { "", "help",  " with the `setup' program", ' ',  1,20, 32, 0, 0, 0, WPT},
 { "", "exit",  " from the `setup' program", ' ', RM,20, 32, 0, 0, 0, WPT},
 {"(P)F1,^X gives menu for chosen program; (P)F2,^Y gives documentation on it",
				"", "", ' ',	76,21,  0,74,WSKP, 0, WST } } ;

	if (NULL == (sep = managesep(3))) return NULL;
	*nentstp = 0;
	if (NULL == (mep = (MENUDATA *)callop(1, sizeof (MENUDATA),
	 "mep in maketop"))) return NULL;
	*nentstp = mep->wnparams = sep->se_npgms + 4;
	mep->wfirstactive = 0;	mep->wlastactive = mep->wnparams - 2;
	if (NULL == (mep->wen = (WINDOW_ENTRY *)callop((unsigned)mep->wnparams,
		sizeof (WINDOW_ENTRY), "mep->wen in maketop"))) return NULL;
	wp = mep->wen;	cvt_fixed(wp++, &(topfixed[0]), 1);
	pdp = sep->se_pmlist;
	for (i = 0, cury = 2; i < sep->se_npgms; i++, pdp++, wp++, cury++)
	   {
		if ((*pdp)->pr_flags & 020) curx = CM;
		else if ((*pdp)->pr_flags & 040)
		  {
			curx = RM;	cury--;
		  }
		else curx = 1;
		if ((*pdp)->pr_flags & 0100) cury++;
		wp->we_y = cury;	wp->we_x0 = curx;
		wp->we_label = "";	wp->we_string = (*pdp)->pr_name;
		wp->we_strlen = strlen((*pdp)->pr_name);
		wp->we_post = (*pdp)->pr_label;	wp->we_delim = ' ';
		wp->we_lablen = wp->we_afterdot = wp->we_flags = 0;
		wp->we_type = WPT;
	   }
	cvt_fixed(wp++, &(topfixed[1]), 1); cvt_fixed(wp++,&(topfixed[2]), 1);
	cvt_fixed(wp, &(topfixed[3]), 1);	construct_links(mep);
	return mep;
}

static int fcode(char *str)
{ /*	This searches through the list of PMENUSTUFF structures to and returns
	the index within the list of the one for which the program name is
	`str'.  Because of the way the list is set up, we actually return a
	number one larger than the program's actual position on the list.
	`fcode' returns 0 if `str' isn't on the list. */
	int		i;
	SETUPWORK	*sep;
	PMENUSTUFF	**pdp;

	if (NULL == (sep = managesep(3))) return 0;
	for (pdp = sep->se_pmlist, i = 0; i < sep->se_npgms; i++, pdp++)
		if (!strcmp((*pdp)->pr_name, str)) return i+1;
	return 0;
}

static int canuse(char *str)
{ /*	This returns 1 iff the file or environment variable `str' exists,
	is readable, and has nonzero length */

	char	*tname;
	struct stat	stbu;

	if ((NULL == str) || ('\0' == *str)) return 0;
	tname = trname(str);	if (0 != xgaccess(tname, 4, 0)) return 0;
	if ((-1 == stat(tname, &stbu)) || (stbu.st_size < 1)) return 0;
	return 1;
}

static int next_program(void)
{ /*	This returns the `filecode' associated with the program that we believe
	to be the logical next step in the progression through the package.
	This is accomplished by looking through a list of logical names and
	deciding, based on which files are present and absent,
	what to do next. */
	char		*result;
	XGNAME		*lnamb;

	if ((NULL == (result = getenv("WORK"))) || ('\0' == *result))
		return fcode("setup");
	if (NULL == (lnamb = sharename(0x4f))) return EALLOC;
	strcpy(lnamb->xmm_workna, result); /* stuff this value in xmm_workna */
	if (0 != dir_access(result, 6)) return fcode("setup");
	if (0 != xgaccess(trname("XGENCMD"), 4, 0)) return fcode("setup");
	 /* what we check next depends on whether we're fiddling with
	 the early stages of processing (up through "reduce") or the
	 later stages.  In the early stages, we need USPLINE, AMASK, etc.
	 In the later stages, we mostly need SCALEIN. */
	if (!canuse("multiref"))
	  {
		if (!canuse("uspline")) return fcode("calibrate");
		else if (!canuse("amask")) return fcode("border");
		else if (!canuse("centroids")) return fcode("spots");
		else if (!canuse("uparams")) return fcode("refine");
		else if (!canuse("urefls")) return fcode("integrate");
		else return fcode("reduce");
	  }
	else {
		if (!canuse("scalein")) return fcode("scalek");
		else if (!canuse("mulist")) return fcode("makemu");
		else return fcode("stats");
	     }
	/*NOTREACHED*/
}

static int setuptop(int *diagtp)
{ /* this runs the top-level menu.  It returns the position on the menu
	 of the next program that ought to be run */
	int		nentstop;
	time_t		tltop;
	SETUPWORK	*sep;
	XGNAME		*lxgp;
	static char	datetop[26];

	if (NULL == (sep = managesep(3))) return 0;
	if (NULL == (lxgp = sharename(0x80f))) return 0;
	if (NULL == sep->se_mtopp)	/* create the structures if needed */
	  {
		sep->se_mtopp = maketop(&nentstop);
		sep->se_mtprevp = maketop(&nentstop);
		*diagtp = (sep->se_mtopp)->wen[nentstop - 1].we_y + 1;
	  }
	tltop = NOW;	strcpy(datetop, ctime(&tltop));	datetop[24] = '\0';
	sprintf(lxgp->xmm_ptitle,
	 "XGEN %s: Copyright IIT, 2002. Toplevel menu:%s",
		XGEN_VERSION, datetop);
	menu_setup(sep->se_mtopp, sep->se_mwp, lxgp->xmm_ptitle, 3);
	wmove(sep->se_mwp, *diagtp, 1);
	wrefresh(sep->se_mwp);		return next_program();
}

static int look_topcmd(int inpar)
{ /*	This processes keystrokes for the top-level menu. The value it
	returns is the "parameter number" of the command the user asks for. */
	int		par, newpar;
	SETUPWORK	*sep;
	XGNAME		*lxgp;
	WINDOW_ENTRY	*wp;

	if (NULL == (sep = managesep(3))) return inpar;
	if (NULL == (lxgp = sharename(3))) return inpar;
	par = inpar;	wp = &((sep->se_mtopp)->wen[par]);
	wmove(sep->se_mwp, (int)wp->we_y, (int)wp->we_x0);
	/* ask user for keystroke. If it was a positional command (^L,etc.)
	 just move to the appropriate place and ask for another. */
	while ((K_PF1 != (newpar = getans(sep->se_mtopp, sep->se_mtprevp,
	 sep->se_mwp, par, lxgp->xmm_ptitle))) && (K_PF2 != newpar))
	  {
		par = newpar;	wp = &((sep->se_mtopp)->wen[par]);
		wmove(sep->se_mwp, (int)wp->we_y, (int)wp->we_x0);
		wrefresh(sep->se_mwp);
	  }
	/* if the user hits <PF1> we return the position value.
	If he hits <PF2> we return a code that generates a help menu. */
	return (newpar == K_PF2 ? DO_HELP :
		(par == (sep->se_mtopp)->wnparams - 3 ? DO_HELP :
		(par == (sep->se_mtopp)->wnparams - 2 ? DO_EXIT : par)));
}

static int nprintable(char *str, int maxlen)
{ /*	This counts the number of characters in `str' that are printable,
	i.e. alphabetic, numeric, or punctuation.  We search up to the first
	'\0', or up through `maxlen' characters, whichever comes first. */
	int	i, nprint;
	char	*cp;

	for (cp = str, nprint = i = 0; i < maxlen; i++, cp++)
	   {
		if ('\0' == *cp) break;	nprint += (041 <= *cp && *cp <= 0176);
	   }
	return nprint;
}

static void make_cmd(MENUDATA *mep)
{ /*	This constructs a command from the contents of the WINDOW_ENTRY
	and PMENUSTUFF structures. */
	char		*cmd, *cp, *icp, ccwork[82];
	int		i, adddash, isinverted, ison;
	size_t		ntocopy;
	WINDOW_ENTRY	*wcp;
	XMSHOW		*pxp;
	PMENUSTUFF	*ptp;
	SETUPWORK	*sep;

	if (NULL == (sep = managesep(3))) return;
	cmd = &(ccwork[0]);
	zeromem((AVOID *)cmd, 80);
	/* command begins with the name of the program*/
	ptp = sep->se_ptp;
	for (cp = cmd, icp = ptp->pr_name;'\0' != *icp; cp++,icp++)
		*cp = *icp;
	wcp = &(mep->wen[1]);	pxp = &(ptp->pr_show[0]);	adddash = 1;
	/* set up the logicals for the files */
	if (ptp->pr_nfile)
	  {
		wcp++;	/* skip over the WINDOW_ENTRY for the file label */
		for (i = 0; i < ptp->pr_nfile; i++, wcp++, pxp++)
			(void)set_logical(pxp->xs_defau, wcp->we_string);
	  }
	/* add in the Boolean flags */
	if (ptp->pr_nbool)
	  {
		wcp++;	/* skip over the WINDOW_ENTRY for Boolean label */
		for (i = 0, adddash = 1;
		 i < ptp->pr_nbool; i++, wcp++, pxp++)
		   {
			isinverted = wcp->we_flags & WNBO;
			ison = (*(wcp->we_string) == 'y');
			if ((!isinverted) == ison)
			  {
				if (adddash)
				  {
					*cp++ = ' '; *cp++ = '-'; adddash = 0;
				  }
				*cp++ = *(pxp->xs_defau);	*cp = '\0';
			  }
		   }
	  }
	/* add in the switches, i.e. the flags beginning with "-"
	 that take numeric parameters */
	if (ptp->pr_nsw)
	  {
		wcp++;	/* skip over WINDOW_ENTRY for switch label */
		for (i = 0; i < ptp->pr_nsw; i++, wcp++, pxp++)
		   {
			if (('\0' != *wcp->we_string) &&
			 (nprintable(wcp->we_string, (int)wcp->we_strlen) > 0))
			  {
				if (adddash)
				  {
					*cp++ = ' ';	*cp++ = '-';
				  }
				else	adddash = 1;
				*cp++ = *(pxp->xs_defau);
				for (icp = wcp->we_string;' ' == *icp; ) icp++;
				for ( ; '\0' != *icp; icp++, cp++) *cp = *icp;
				*cp = '\0';
			  }
		   }
	  }
	/* finally, add in the non-switch-type parameters */
	if (ptp->pr_nparam)
	  {
		wcp++;	/* skip over WINDOW_ENTRY for parameter label */
		for (i = 0; i < ptp->pr_nparam; i++, wcp++)
		   {
			if (('\0' != *wcp->we_string) &&
			 (nprintable(wcp->we_string, (int)wcp->we_strlen) > 0))
			  {
				*cp++ = ' ';
				for (icp = wcp->we_string;' ' == *icp; ) icp++;
				for ( ; '\0' != *icp; icp++, cp++) *cp = *icp;
				*cp = '\0';
			  }
		  }
	  }
	wcp = &(mep->wen[sep->se_cmdpar]); /* go to the actual command slot */
	ntocopy = strlen(cmd);
	if (ntocopy > wcp->we_strlen) ntocopy = wcp->we_strlen;
	for (cp = wcp->we_string, icp = cmd, i = 0;
	 (i < (int)ntocopy) && ('\0' != *icp); i++, cp++, icp++) *cp = *icp;
	*cp = '\0';
	disp_param(mep, sep->se_mwp, sep->se_cmdpar);	wrefresh(sep->se_mwp);
}

static MENUDATA *stuf2menu(int pnum)
{ /*	This converts a PMENUSTUFF structure to a displayable MENUDATA
	structure.  It returns a pointer to the structure thus obtained. */
	char		*bpool, *pool, *result, *cp, *cpp;
	int		i, canbatch, cansubmit, curry, stacking, nrows;
	int		widfile, widparam, maxlabel, currtype;
	int		nrowsu, nrows13, nrows023, nrows6, nrows6p;
	MENUDATA	*mep;
	WINDOW_ENTRY	*wp;
	XMSHOW		*pxp;
	PMENUSTUFF	*ptp;
	FRAMEINDEX	*lfmi;
	SETUPWORK	*sep;
	static char	boolpost[] = " ( -e)";
	static char	locbuf[80], templogical[512];
	static WINDOW_TEMPLATE	wfixed[] = {
 { "Menu for",  "no program",		"", ' ',10, 1,12, 8,WSKP, 0, WST},
 {"FILES       ","file name",		"", ' ',21, 2,16,10,WSKP, 0, WST},
 {"BOOLEANS    ","hit y or n to change","", ':',59, 2,20,10,WSKP, 0, WST},
 {"SWITCHES    ","type in changes",	"", ':',21, 9,20,10,WSKP, 0, WST},
 {"PARAMETERS  ","type in changes",	"", ':',60, 9,20,10,WSKP, 0, WST},
 {"OVERALL COMMAND", "hit <PF1> or ctrl-X with cursor on action key when ready",
					"", ':',19,18,57,16,WSKP, 0, WST},
 {"", "                                                                     ",
					"", ' ', 1,19,70, 0,   0, 0, WST},
 {"Action",	"Run&Wait",		"", ':',11,20, 8, 6,WTYT, 0, WPT},
 {"",		"Run&Don't Wait",	"", ' ',22,20,14, 0,WTYT, 0, WPT},
 {"",		"Submit to Queue",	"", ' ',39,20,15, 0,WTYT, 0, WPT},
 {"",		"Toplevel Menu",	"", ' ',58,20,13, 0,WTYT, 0, WPT},
 {"",		"Exit",			"", ' ',74,20, 4, 0,WTYT, 0, WPT } };

	if (NULL == (sep = managesep(3))) return NULL;
	if (pnum <= 0) return NULL;	/* shouldn't happen */
	sep->se_ptp = ptp = *(sep->se_pmlist + pnum - 1);
	/* can the requested program be run offline? */
	cansubmit = canbatch = !(ptp->pr_flags & 010);
#ifndef VMS
	cansubmit = 0;	/* Some Unix systems have queues, but I'm lazy */
#endif
	if (NULL == (mep = (MENUDATA *)callop(1, sizeof (MENUDATA),
	 "mep in stuf2menu"))) return NULL;
	/* how we stack the sub-menus depends on how many entries we make.
	The variable `stacking' will have the following meanings.  In
	general the more crowded the menu, the larger the value of `stacking'.
	value	result
	0	FILES, BOOLEANS, SWITCHES, PARAMETERS in a single column;
		 command-title line above the overall command line
	1	like 0 except with the command-title line omitted
	2	FILES, SWITCHES, PARAMETERS along left column;
		 BOOLEANS on the right, starting near the middle;
		 command-title line above the overall command line
	3	like 2 except with the command-title line omitted
	4	FILES, SWITCHES in left column; BOOLEANS, PARAMETERS in right.
		 command-title line above overall command line
	5	like 4 except with the command-title line omitted.
	6	FILES, BOOLEANS, and PARAMETERS along left;
		SWITCHES along the right. Command-title included.
	7	like 6 except with command-title line omitted.
	*/
	nrows023 = ptp->pr_nfile + ptp->pr_nsw + ptp->pr_nparam +
	 (ptp->pr_nfile > 0) + (ptp->pr_nsw > 0) + (ptp->pr_nparam > 0);
	nrows = nrows023 + ptp->pr_nbool + (ptp->pr_nbool > 0);
	stacking = (nrows <= 16 ? 0 : (nrows <= 17 ? 1 :
	  (nrows023 <= 16 ? 2 : (nrows023 <= 17 ? 3 : 4))));
	/* in the most crowded configuration (stacking==4),
	 we check to see if there's room to put the command-line label.
	 Also, see if the switches-alone-along-the-right config is better*/
	if (stacking == 4)
	  {
		nrowsu = ptp->pr_nfile + ptp->pr_nsw + (ptp->pr_nfile > 0)
			+ (ptp->pr_nsw > 0);
		nrows13 = ptp->pr_nbool + ptp->pr_nparam + (ptp->pr_nbool > 0)
			+ (ptp->pr_nsw > 0);
		if (nrows13 > nrowsu) nrowsu = nrows13;
		nrows6 = ptp->pr_nfile + ptp->pr_nbool + ptp->pr_nparam +
		 (ptp->pr_nfile > 0) + (ptp->pr_nbool > 0) +
		 (ptp->pr_nparam > 0);
		nrows6p = ptp->pr_nsw + (ptp->pr_nsw > 0);
		if (nrows6p > nrows6) nrows6 = nrows6p;
		if (nrows6 < nrowsu)
		  {
			nrowsu = nrows6;	stacking = 6;
		  }
		if (nrowsu > 16) stacking++;
	  }
	mep->wnparams = nrows + 5 + canbatch + cansubmit;
	if (!(stacking & 01)) (mep->wnparams++);
	mep->wfirstactive = 3;	mep->wlastactive = mep->wnparams - 1;
	if (NULL == (mep->wen = (WINDOW_ENTRY *)callop((unsigned)mep->wnparams,
		sizeof (WINDOW_ENTRY), "mep->wen in stuf2menu"))) return NULL;
	if (strlen(ptp->pr_title) > 56) ptp->pr_title[56] = '\0';
	/*establish an entry for the program name & summary */
	if (NULL == (bpool = (char *)mallop(4096, "bpool in stuf2menu")))
		return NULL;
	pool = bpool;
	wp = &(mep->wen[0]);	cvt_fixed(wp, &(wfixed[0]), 1);
	wp->we_string = pool;	strcpy(wp->we_string, ptp->pr_name);
	wp->we_post = pool + 32;	strcpy(wp->we_post, ptp->pr_title);
	wp++;	pool += 96;
	pxp = &(ptp->pr_show[0]);
		/* establish window entries for the "FILE" lines */
	curry = 2;
	zeromem((void *)templogical, 512);
	if (ptp->pr_nfile) /* files always live in the upper left */
	  {
		cvt_fixed(wp++, &(wfixed[1]), 1); /* label for FILE lines */
		widfile = (stacking > 3 ? FILEWID : 2 * FILEWID);
		for (i = 0, curry = 3; i < ptp->pr_nfile;
		  i++, wp++, pxp++, pool += widfile + 2, curry++)
		   {
			if (strlen(pxp->xs_label) > 19)
				pxp->xs_label[19] = '\0';
			wp->we_label = pxp->xs_label;
			wp->we_string = pool;
			zeromem((AVOID *)wp->we_string, (unsigned)widfile);
			result = getenv(pxp->xs_defau);
			if ((NULL != result) && ('\0' != *result))
			  { /* shorten the name if possible */
				strcpy(templogical, result);
				if (!shorten(templogical, "SCRT"))
				  {
					if (!shorten(templogical, "DATA"))
					  (void)shorten(templogical,"WORK");
				  }
				strcpy(wp->we_string, templogical);
				zeromem((void *)templogical, 512);
			  }
			if (strlen(wp->we_string) > (unsigned)widfile)
				*(wp->we_string + widfile) = '\0';
			if (stacking > 3) wp->we_post = "";
			else wp->we_post = pxp->xs_defau;
			wp->we_delim = ' ';
			wp->we_x0 = 21; wp->we_y = curry;
			wp->we_strlen = widfile;
			wp->we_lablen = 19;	wp->we_flags = WTYT;
			wp->we_afterdot = 0;	wp->we_type = WST;
		   }
	  }
	if ((stacking == 4) || (stacking == 5)) curry = 2;
	else if (stacking > 1) curry = 2+ ptp->pr_nfile +(ptp->pr_nfile > 0);
	if (ptp->pr_nbool)
	  {
		cvt_fixed(wp, &(wfixed[2]), 1);	/* label for Boolean lines */
		if ((2 <= stacking) && (stacking <= 5)) wp->we_x0 = 60;
		else wp->we_x0 = 21;
		wp->we_y = curry++;	wp++;
		for (i = 0; i < ptp->pr_nbool;
			i++, pxp++, wp++, curry++, pool += 12)
		   {
			if (strlen(pxp->xs_label) > 24) pxp->xs_label[24]='\0';
			wp->we_label = pxp->xs_label;
			cp = wp->we_string = pool;
			cpp = wp->we_post = pool + 2;
			strcpy(cpp, boolpost);	*(cpp + 4) = *pxp->xs_defau;
			wp->we_string[0] = pxp->xs_defau[1];
			if ('A' <= *cp && *cp <= 'Z')
			  {
				*(cpp + 2) = '!';	*cp += 'a' - 'A';
				wp->we_flags = WFNB;
			  }
			else	wp->we_flags = WFBO;
			wp->we_string[1] = '\0';	wp->we_delim = ' ';
			if ((2 <= stacking) && (stacking <= 5)) wp->we_x0 = 68;
			else wp->we_x0 = 27;
			wp->we_y = curry;	 wp->we_strlen = 1;
			wp->we_lablen = 24;	wp->we_afterdot = 0;
			wp->we_type = WCH;
		   }
	  }
	if ((2 <= stacking) && (stacking <= 5))
	  {
		curry = ptp->pr_nfile + 3;	if (curry < 5) curry = 5;
	  }
	else if (stacking >= 6) curry = 2;
	lfmi = sep->se_fmi;
	if (ptp->pr_nsw)
	  {
		cvt_fixed(wp, &(wfixed[3]), 1);	/* label for switches */
		if (6 <= stacking) wp->we_x0 = 57;
		wp->we_y = curry++;	wp++;
		for (i = 0; i < ptp->pr_nsw;
		 i++, wp++, pxp++, pool += SWWID + 12, curry++)
		   {
			maxlabel = 18;
			if (strlen(pxp->xs_label) > (size_t)maxlabel)
				pxp->xs_label[maxlabel] = '\0';
			wp->we_label = pxp->xs_label;
			wp->we_string = pool;
			cpp = wp->we_post = pool + SWWID + 2;
			strcpy(wp->we_post, boolpost);
			*(cpp + 4) = *(pxp->xs_defau);
			zeromem((AVOID *)wp->we_string, SWWID);
			if ((ptp->pr_flags & 04) &&
			 (*(pxp->xs_defau) == 'a')) sprintf(wp->we_string,
				"%d,%d", (int)(lfmi->fi_firstfm),
				(int)(lfmi->fi_lastfm));
			else if (strlen(pxp->xs_defau) > 1)
			     strncpy(wp->we_string, pxp->xs_defau + 1, SWWID);
			else strcpy(wp->we_string, "        ");	
			wp->we_delim = ' ';
			wp->we_x0 = (stacking >= 6) ? 63: 27;
			wp->we_y = curry;	wp->we_strlen = SWWID;
			wp->we_lablen = maxlabel; wp->we_flags = WTYT;
			wp->we_afterdot = 0;	wp->we_type = WST;
		   }
	  }
	if ((3 <= stacking) && (stacking <= 5))
		curry = ptp->pr_nbool + (ptp->pr_nbool > 0) + 3;
	else if (stacking >= 6)
		curry = nrows023 + 2 - ptp->pr_nsw - (ptp->pr_nsw > 0);
	if (ptp->pr_nparam)
	  {
		cvt_fixed(wp, &(wfixed[4]), 1);	/* label for parameters */
		if (stacking < 4)
		  {
			wp->we_x0 = 21; widparam = 4 * PARAMWID;
		  }
		else if (stacking >= 6)
		  {
			wp->we_x0 = 21;	widparam = PARAMWID;
		  }
		else {
			widparam = PARAMWID;	wp->we_x0 = 60;
		     }
		wp->we_y = curry++;	wp++;
		if ((ptp->pr_flags & 01) &&
		 ((NULL == lfmi) || (lfmi->fi_lastfm == 999)))
			lfmi = ftables(15);
		for (i = 0; i < ptp->pr_nparam;
		 i++, wp++, pxp++, pool += widparam + 2, curry++)
		   {
			if (strlen(pxp->xs_label) > 21) pxp->xs_label[21]='\0';
			wp->we_label = pxp->xs_label;
			wp->we_string = pool;
			zeromem((AVOID *)wp->we_string, (unsigned)widparam);
			if ((!i) && (!strcmp(ptp->pr_name, "pmotif")) &&
			 (NULL != (cp = getenv("RUN_ID"))) && ('\0' != *cp))
				strncpy(wp->we_string, cp, (size_t)widparam);
			else if ('\0' != *(wp->we_string)) strncpy(
			 wp->we_string, pxp->xs_defau, (size_t)widparam);
			wp->we_post = "";	wp->we_delim = ' ';
			if ((stacking == 4) || (stacking == 5)) wp->we_x0 = 68;
			else wp->we_x0 = 24;
			wp->we_y = curry;
			wp->we_strlen = widparam;
			wp->we_lablen = 21;	wp->we_flags = WTYT;
			wp->we_afterdot = 0;	wp->we_type = WST;
			if ((!i) && (ptp->pr_flags & 01) &&
			 (0 <= lfmi->fi_firstfm && lfmi->fi_firstfm <= 9999))
			  sprintf(wp->we_string, "%d", (int)(lfmi->fi_firstfm));
			else if ((i == 1) && (ptp->pr_flags & 02) &&
			 (0 <= lfmi->fi_lastfm && lfmi->fi_lastfm <= 9999))
			  sprintf(wp->we_string, "%d", (int)(lfmi->fi_lastfm));
		   }
	  }
	/* set up a line for the label of the overall command, unless we're
	squeezed for space */
	if (!(stacking & 01)) cvt_fixed(wp++, &(wfixed[5]), 1);
	/* ... and a line for the overall command.  The command string's
	length varies, so we set it externally after calling cvt_fixed */
	sep->se_cmdpar = wp - &(mep->wen[0]); /* remember this entry number */
	cvt_fixed(wp, &(wfixed[6]), 0);	wp->we_strlen = 69;
	wp->we_string = &(locbuf[0]);
	for (cp = &(wp->we_string[0]), i = 0; i < 69; i++, cp++) *cp = ' ';
	*cp = '\0';
	if (stacking & 01)	/* if we're squeezed, we make a mini-label */
	  {
		wp->we_label = "CMD: "; wp->we_x0 = 5;	wp->we_lablen = 5;
	  }
	/* smartdef(mep, pnum); */
	/* make_cmd(mep); */
	strcpy(wp->we_string, smartdefa(ptp->pr_name, 0, &currtype));
	i = cmd2strs(mep, wp->we_string);
	wp++;	/* advance WINDOW_ENTRY pointer to the buttons on the bottom*/
	sep->se_exbutt = wp - &(mep->wen[0]);	/* remember button number */
	cvt_fixed(wp++, &(wfixed[7]), 1);			/* Run&Wait */
	if (canbatch) cvt_fixed(wp++, &(wfixed[8]), 1); /* Run & Don't Wait */
	if (canbatch && cansubmit) cvt_fixed(wp++, &(wfixed[9]), 1); /*Queue*/
	sep->se_topbutt = wp - &(mep->wen[0]);	/*remember this button # too*/
	cvt_fixed(wp++, &(wfixed[10]), 1);		/* top-level */
	cvt_fixed(wp, &(wfixed[11]), 1);		/* exit */
	construct_links(mep);	/* set up keystroke linkages */
	return mep;
}

static void do_cmd(int runtype)
{ /* This calls `system' to invoke a program.  Then it prints
	a response error message and asks for a reaction from the user.
	if `runtype' is nonzero and we're on VMS, we call lib$spawn
	to invoke the program. */
	int		result;
	WINDOW_ENTRY	*wp;
	SETUPWORK	*sep;
	XGNAME		*lxgp;

	if (NULL == (sep = managesep(3))) return;
	if (NULL == (lxgp = sharename(3))) return;
#ifdef VMS
	char	*cp;
	static $DESCRIPTOR(cmdexec,  " ");
	static $DESCRIPTOR(logexec,  " ");
	static char	flogname[] = "WORK:XGENSPAWN.LOG";
	unsigned	spawnflags;
	int		retval, quotes;
	FILE		*fpcmd;

	if (runtype == 1)
	  { /* spawn it appropriately */
		for (cp = lxgp->xmm_miscb, quotes = 0;
		 '\0' != *cp; cp++)
		   {
			if (*cp == '"') quotes = 1 - quotes;
			if ((!quotes) && ('a' <= *cp) && (*cp <= 'z'))
				*cp -= ('a' - 'A');
		   }
		cmdexec.dsc$a_pointer = lxgp->xmm_miscb;
		cmdexec.dsc$w_length = strlen(lxgp->xmm_miscb);
		logexec.dsc$a_pointer = flogname;
		logexec.dsc$w_length = strlen(flogname);
		spawnflags = (020 | 01); /* spawn flags: NOTIFY, NOWAIT */
		if (SS$_NORMAL != (retval =
		 lib$spawn(&cmdexec, 0, &logexec, &spawnflags)))
		  fprintf(stderr, "Error %d in spawning subprocess\n", retval);
		else printf("Command (%s) is executing in a subprocess\n",
		 lxgp->xmm_miscb);
	  }
	else if (runtype == 2) /* submit to batch queue SYS$BATCH */
	  {
		if (NULL == (fpcmd = fopen("WORK:XGENBAT.COM", "w")))
			fprintf(stderr,
			 "Error opening %s as XGEN batch command;sorry!\n");
		else {
			fputs("$ @xgmain:[coms]xgenset\n", fpcmd);
			fprintf(fpcmd, "$ @%s%s\n",
			 lxgp->xmm_workna, lxgp->xmm_idna);
			fprintf(fpcmd, "$ %s\n", lxgp->xmm_miscb);
			fclose(fpcmd);
			result =
			 system("SUBMIT/NOPRINT/NOTIFY/LOG=WORK: WORK:XGENBAT");
		      }
	  }
	else 	result = system(lxgp->xmm_miscb); /* `system' call */
#else	/* we have to turn off CURSES for a while in this case */
	endwin();	result = locsysc(lxgp->xmm_miscb, !runtype);
#endif
	(void)xeflush(stdout);
	printf(" Run of (%s) resulted in status %d\n",
	 lxgp->xmm_miscb, result);
	printf("Hit E or e to exit from `setup', anything else to return: ");
	while (NULL == fgets(lxgp->xmm_miscb, 2, stdin)) ;
	if ('e' == *(lxgp->xmm_miscb) || 'E' == *(lxgp->xmm_miscb)) return;
#ifndef VMS	/* now we have to turn CURSES back on and blank the screen */
#ifdef linux
	initscr();
#endif
#ifdef	sgi
	initscr();
#endif
	(void)setmore(0);	crmode();	noecho();	nonl();
	for (result = 0; result < 22; result++)
	   {
		wmove(stdscr, result, 0);	wclrtoeol(stdscr);
		wmove(sep->se_mwp, result, 0);		wclrtoeol(sep->se_mwp);
	   }
	result = next_program();	wp = &((sep->se_mtopp)->wen[result]);
	wmove(sep->se_mwp, (int)wp->we_y, (int)wp->we_x0);
	wrefresh(stdscr);	wrefresh(sep->se_mwp);
#endif
}

static void show_help(int filecode)
{ /*	Print out some help information a la "more" */
	int		lineno;
	char		*cp, incha;
	SETUPWORK	*sep;
	XGNAME		*lxgp;
	static FILE	*locfp;
	
	if (NULL == (sep = managesep(3))) return;
	if (NULL == (lxgp = sharename(0x100f))) return;
	if ((0 <= filecode) && (filecode <= sep->se_npgms + 2) &&
	  (NULL != (cp = getenv("XGDOC"))) && ('\0' != *cp) &&
	  (0 == dir_access(cp, 4))) /* see if help is available */
	  { /* help filename */
		sprintf(lxgp->xmm_miscb, "%s/%s.%s",
		 cp, (filecode == sep->se_npgms + 1 ? "setup" :
		 (filecode == sep->se_npgms + 2 ? "exit" :
		 (filecode ?
		 (*(sep->se_pmlist + (filecode-1)))->pr_name : "setup"))),
		 DOCSUFFIX);
		if (NULL == (locfp = fopen(lxgp->xmm_miscb, "r")))
		  {
			printw(
 "Sorry: help file (%s) unavailable;\nHit Enter to continue: ",
			 lxgp->xmm_miscb);
			incha = wget1ch(stdscr);
			move(0, 0); wclrtobot(stdscr); wrefresh(stdscr);
			wmove(sep->se_mwp, 0, 0); wclrtobot(sep->se_mwp);
			wrefresh(sep->se_mwp); return;
		  }
		/* use a facility like "more" to print the help */
		for (lineno = 0; NULL != fgets(lxgp->xmm_miscb,
		 100, locfp); lineno++)
		   {
			addstr(lxgp->xmm_miscb); if (!((lineno + 1) % 22))
			  {
				addstr("__More__");
				incha = wget1ch(stdscr);
				while (('\n' == incha) || ('\r' == incha))
				  {
					if (NULL == fgets(
					 lxgp->xmm_miscb,100, locfp))
					  {
						incha = ' ';	break;
					  }
					addstr(lxgp->xmm_miscb);
					addstr("__More__");
					incha = wget1ch(stdscr);
				  }
				if ((' ' != incha) && ('\t' != incha)) break;
				move(0, 0);	wclrtobot(stdscr);
				wrefresh(stdscr);
			  }
		   }
		(void)fclose(locfp);	locfp = NULL;
		addstr("Hit Enter to continue: ");
		incha = wget1ch(stdscr);
		/* clear entire screen and go back */
		move(0, 0);	wclrtobot(stdscr);	wrefresh(stdscr);
		wmove(sep->se_mwp, 0, 0);	wclrtobot(sep->se_mwp);
		wrefresh(sep->se_mwp);
	 }
}

static int setup_prog(int prognum, int diagno)
{ /* this sets up a menu for a specific program.  There are three
	kinds: the filename-and-index menu, a help menu, and an ordinary
	program menu. */
	int		locy, locx;
	time_t		tlprog;
	static char	dateprog[26];
	WINDOW_ENTRY	*wp, *wp0, *wpe;
	MENUDATA	*mtopp;
	SETUPWORK	*sep;
	XGNAME		*lxgp;
	static WINDOW_TEMPLATE	wfiles[] = {
 { "", "file-setup", "", ' ', 21, 0, 10, 1, WNPR, 0, WST},
 {"Software Copyright 2002 IIT, Chicago IL USA.", "","",
	' ',65, 1, 1, 63,WSKP, 0, WST},
 {"Run ID Prefix  ", "MY_XTAL",   "RUN_ID",' ',32,3,36,30,WFIL,0,WST},
 {"  Directories to be Used in the Run:", "","",' ',65, 4, 1, 63,WSKP, 0, WST},
 {"Directory Type   ","Directory Name","Logical Name",
	' ', 22,5,46,20,WSKP,0,WST},
 {"Processed Results  ", "/home/xgen/proj",  "WORK",' ',22,6,46,20,WFIL,0,WST},
 {"Raw Frame Data     ", "/home/xgen/data",  "DATA",' ',22,7,46,20,WFIL,0,WST},
 {"Scratch Results    ", "/home/xgen/proj",  "SCRT",' ',22,8,46,20,WFIL,0,WST},
 {"Documentation      ", "/home/xgen/docs",   "XGDOC",
	' ',22,9,46,20,WFIL,0,WST},
 {"Characteristics of the frame index:", "", "",
	' ', 39, 10, 1, 36, WSKP, 0, WST},
 {"Filename template", "george01_###.img", "",
	' ', 22, 11, 40, 18, WFIL, 0, WST},
 {"Base", " 10", "", ' ', 5, 12, 3, 4, WTYT, 0, WSH},
 {"Interval", "  1", "", ' ', 22, 12, 4, 8, WTYT, 0, WSH},
 {"Start Frame",    "      1", "", ' ',40, 12, 7,11, WTYT, 0, WSH},
 {"End Frame",      "    999", "", ' ',60, 12, 7, 9, WTYT, 0, WSH},
 {"This allows for creation, modification, and use of files for", "","",
	' ',70,14,1,66,WSKP,0,WST},
 {"logical names and the frame index in the XGEN package.", "","",
	' ',70,15,1,65,WSKP,0,WST},
 {"", "Create Logical-Name File", "", ' ', 1,16,24, 0, 0, 0, WPT},
 {"", "Define Logical Names",     "", ' ',28,16,20, 0, 0, 0, WPT},
 {"", "Create Frame Index",       "", ' ',51,16,18, 0, 0, 0, WPT},
 {"", "Help!",                    "", ' ',72,16, 5, 0, 0, 0, WPT},
 {"", "Edit Logical-Name File",   "", ' ', 1,17,22, 0, 0, 0, WPT},
 {"", "Edit Frame Index",         "", ' ',28,17,16, 0, 0, 0, WPT},
 {"", "Return to Main Menu",      "", ' ',51,17,19, 0, 0, 0, WPT},
 {"", "Exit!",                    "", ' ',72,17, 5, 0, 0, 0, WPT}
} ;

	if (NULL == (sep = managesep(3))) return 0;
	if (NULL == (lxgp = sharename(3))) return 0;
	mtopp = sep->se_mtopp;	lxgp->xmm_miscc[0] = '\0';
	if (prognum == DO_HELP) /* get help based on button position */
	  {
		if (mtopp)	/* figure out where we are */
		  {
			getyx(sep->se_mwp, locy, locx);
			wp0 = &(mtopp->wen[(int)mtopp->wfirstactive]);
			wpe = &(mtopp->wen[(int)mtopp->wlastactive]);
			for (wp = wp0 ; wp <= wpe; wp++)
			   {
				if ((wp->we_y == locy) && (wp->we_x0 == locx))
					break;
			   }
			prognum = wp - wp0;
		  }
		else	prognum = 0;
		/* clear the bottom of the screen in case of error msgs */
		wmove(sep->se_mwp, diagno, 1);	wclrtobot(sep->se_mwp);
		wrefresh(sep->se_mwp);
		show_help(prognum);	return DO_HELP;
	  }
	sep->se_nentsf = sizeof (wfiles) / sizeof (WINDOW_TEMPLATE);
	tlprog = NOW;
	strcpy(dateprog, ctime(&tlprog));	dateprog[24] = '\0';
	menufree(sep->se_mprog);
	menufree(sep->se_mprevprog); /* free up memory*/
	wmove(stdscr, 0, 0);	wclrtobot(stdscr);	wrefresh(stdscr);
	if (DO_FILES == prognum)
	  {
		sprintf(lxgp->xmm_miscc,
		 "XGEN menu for name and index management on %s", dateprog);
		sep->se_mprog = newmenu((unsigned)(sep->se_nentsf), wfiles);
		sep->se_mprevprog = newmenu((unsigned)(sep->se_nentsf), wfiles);
		setupvals(sep->se_mprog);
		menu_setup(sep->se_mprog, sep->se_mwp, lxgp->xmm_miscc, 1);
	  }
	else {
		sprintf(lxgp->xmm_miscc,
		 "XGEN %s: Copyright IIT, 2002. Run on %s",
			XGEN_VERSION, dateprog);
		sep->se_mprog = stuf2menu(prognum);
		sep->se_mprevprog = stuf2menu(prognum);
		menu_setup(sep->se_mprog, sep->se_mwp, lxgp->xmm_miscc, 1);
	     }
	return 0;
}

static void setedit(char *fname)
{ /*	This allows the user to edit, using a commonly-available editor,
	the file named in its argument */
	XGNAME		*lxgp;
#ifdef VMS
	static char	editcmd[] = "EDIT ";
#else
	static char	editcmd[] = "vi ";
#endif
	if (NULL == (lxgp = sharename(0x100f))) return;
	strcpy(lxgp->xmm_miscb, editcmd);
	strcat(lxgp->xmm_miscb, trname(fname));	do_cmd(0);
}

static int do_filecmd(int currpar, int diag)
{ /* This interprets a command to the filename-and-index menu. */
	SETUPWORK	*sep;

	if (NULL == (sep = managesep(3))) return currpar;
	/* Figure out which command has been invoked.   This is determined
	from where the cursor is on the menu.  If it's up above somewhere,
	the default (which is a execlognames call) is performed.  If we're
	down below among the command buttons, we perform the command cited. */
	if (currpar == (sep->se_nentsf - 1)) return DO_EXIT;
	else if (currpar == (sep->se_nentsf - 2)) return DO_TOPMENU;
	else if (currpar == (sep->se_nentsf - 3))
	  {
		setedit("FRAMES");	(void)setup_prog(DO_FILES, diag);
	  }
	else if (currpar == (sep->se_nentsf - 4))
	  {
		setedit("XGENCMD");	(void)setup_prog(DO_FILES, diag);
	  }
	else if (currpar == (sep->se_nentsf - 5))
	  {
		show_help(DO_FILES);	(void)setup_prog(DO_FILES, diag);
	  }
	else if (currpar == (sep->se_nentsf - 6))
	  {
		if (NULL == sep->se_fmi) sep->se_fmi = ftables(15);
		writeindex(sep->se_fmi, 1);
	  }
	else if (currpar == (sep->se_nentsf - 8)) (void)writelognames(1);
	else (void)execlognames(1);
	return currpar;	/* i.e., return to caller at the current position */
}

static void chg_filemenu(WINDOW_ENTRY *wtp)
{ /* This reacts to change in the window_entry pointed to by wtp in
  the filename-and-frame-index menu. */
	int		pnum;
	char		*str;
	FRAMEINDEX	*lfmi;
	SETUPWORK	*sep;

	/* only the window_entry structures we can change are relevant */
	if (NULL == (sep = managesep(3))) return;
	if (wtp->we_flags == WSKP || wtp->we_flags == WNPR) return;
	pnum = wtp - &(sep->se_mprog->wen[0]);	str = wtp->we_string;
	lfmi = sep->se_fmi;
	switch (pnum)
	      {
		case	2: (void)set_logical("RUN_ID", str);	return;
		case	5: (void)set_logical("WORK", str);	return;
		case	6: (void)set_logical("DATA", str);	return;
		case	7: (void)set_logical("SCRT", str);	return;
		case	8: (void)set_logical("XGDOC", str);	return;
		case	10:	transfer_template(str);		return;
		case	11:	transfer_base(str);		return;
		case	12:	transfer_interval(str);		return;
		case	13:	lfmi->fi_firstfm = atoi(str);	return;
		case	14:	lfmi->fi_lastfm = atoi(str);	return;
	     }
}

static int do_progmenu(int prognum, int inpar, int diag)
{ /*	This manipulates material in the program-specific menu.  If we're in
	the filename-and-index menu, it returns the position value on the
	current menu where the cursor ended up.  In the ordinary programs it
	(eventually) returns a code indicating that we should go back to the
	toplevel menu. */
	int		par, newpar, boty, result, ifile;
	WINDOW_ENTRY	*tp, *tbot;
	XMSHOW		*pxp;
	PMENUSTUFF	*ptp;
	XGNAME		*lxgp;
	SETUPWORK	*sep;

	if (NULL == (sep = managesep(3))) return DO_EXIT;
	if (NULL == (lxgp = sharename(0x100f))) return DO_EXIT;
	par = inpar;	tp = &(sep->se_mprog->wen[par]);
	ptp = sep->se_ptp;
	tbot = &((sep->se_mprog)->wen[sep->se_mprog->wnparams - 1]);
	wmove(sep->se_mwp, (int)tp->we_y, (int)tp->we_x0);
	/* ask user for keystroke. If it was a positional command (^L,etc.)
	 we modify the system's idea of the parameter in question.
	 Then we move to the new position and ask for another keystroke. */
	while (K_PF1 != (newpar = getans(sep->se_mprog, sep->se_mprevprog,
	 sep->se_mwp, par, lxgp->xmm_ptitle)))
	  { /* if we ever detect that mwp is NULL, it's a sign that the
		user has hit ^C to kill the program; so we exit gracefully */
		if (NULL == sep->se_mwp) return DO_EXIT;
		if (newpar == K_PF2) /* show a help screen, then return here*/
		  {
			show_help(prognum); (void)setup_prog(prognum, diag);
		  }
		else {
			update_par(sep->se_mprog, sep->se_mwp, par);
			if (prognum == DO_FILES) chg_filemenu(tp);
			else if (prognum != DO_HELP) make_cmd(sep->se_mprog);
			par = newpar;	tp = &((sep->se_mprog)->wen[par]);
		  }
		wmove(sep->se_mwp, (int)tp->we_y, (int)tp->we_x0);
		wrefresh(sep->se_mwp);
	  }
	/* again, if we notice that mwp is NULL, we go away */
	if (NULL == sep->se_mwp) return DO_EXIT;
	/* if the user hits <PF1> we're being asked to
	 actually perform a command.  We figure out what the command is
	 and then carry it out. */
	boty = tbot->we_y + 1;	if (boty > 21) boty = 21;
	wmove(sep->se_mwp, boty, 0);	wclrtobot(sep->se_mwp);
	wrefresh(sep->se_mwp);	/* so we can see diagnostics*/
	if (prognum == DO_FILES)
	  {
		result = do_filecmd(par, diag);
		wmove(sep->se_mwp, (int)tp->we_y, (int)tp->we_x0);
		wrefresh(sep->se_mwp);	return result;
	  }
	else if (prognum == DO_HELP) return DO_TOPMENU;
	else if (par < sep->se_topbutt)
	  {
#ifndef VMS	/* un-abbreviate file names just before processing */
		for (tp = &(sep->se_mprog->wen[2]), ifile= 0,
		 pxp = &(ptp->pr_show[0]);
		 ifile < ptp->pr_nfile; ifile++, tp++, pxp++)
		   {
			if (1 == envtoname(tp->we_string, lxgp->xmm_miscb))
			  (void)set_logical(pxp->xs_defau, lxgp->xmm_miscb);
		   }
#endif
		strcpy(lxgp->xmm_miscb,
		 (sep->se_mprog)->wen[sep->se_cmdpar].we_string);
		do_cmd((par < sep->se_exbutt) ? 0 : par - sep->se_exbutt);
		return ((*(lxgp->xmm_miscb) == 'e') ||
		 (*(lxgp->xmm_miscb) == 'E')) ?  DO_EXIT : DO_TOPMENU;
	  }
	else return (par == sep->se_topbutt ? DO_TOPMENU : DO_EXIT);
}

int setupmain(int argc, char *argv[])
{ /* The mainline for curses-based control-flow for X-GEN */
	int		diagt, npsetup, par, progresult, result;
	SETUPWORK	*sep;
	XGNAME		*lxgp;

	if (EHELP == xgjournal(argc, argv)) return EOK;	/* journal the effort */
	if (NULL == (sep = managesep(3))) return EALLOC;
	/* define the initial project name and some environment variables */
	if (NULL == (lxgp = getproject(((argc < 2) ?
		NULL : argv[1]), 1))) return EALLOC;
	sep->se_fmi = ftables(15);	/* read 1st & last frames in index */
	(void)get_xgdoc(1); /* try to find the X-GEN documentation in XGDOC */
	/* start up the Curses-specific stuff */
	if (EOK != (par = start_curses())) return par;
	/* establish the PMENUSTUFF structures */
	if (NULL == (sep->se_pmlist = pmall(&npsetup))) return EALLOC;
	sep->se_npgms = npsetup;
		/* display top-level menu */
	if (0 > (par = setuptop(&diagt))) endsetmenu();
	/* keep track of user's keyboard entries.  If he asks to exit, do so */
	while (DO_EXIT != (result = look_topcmd(par))) /*get user's command */
	     { /* figure out what command the user has invoked.  The user
		 may have asked for the logical-names menu, a help listing, or
		 a specific program.  In any case we process commands thus:*/
		if (0 == setup_prog(result, diagt)) /* set up 2nd-level menu */
		  { /* accept and act upon commands at second level. We'll
			 stay within this loop till the user asks either
			 to exit or to move back up to the top-level menu. */
			progresult = 0;	/* start at top of menu */
			while ((DO_EXIT != (progresult = do_progmenu(result,
			 progresult, diagt)))&& (DO_TOPMENU != progresult)) ;
			if (DO_EXIT == progresult) break;
		  }
		/* any time we get here we re-draw the top-level menu */
		if (0 > (par = setuptop(&diagt))) endsetmenu();
	     }
	endsetmenu();	/* close Curses stuff and get out */
	return EOK;
}

/* end of set_up.c */
