/* dispcalc.c -- calculational engine for displays of XGEN images

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	28 Mar 2007: recognize environment variable EQUALSCALE
	 as meaning that we want the display to have the same magnification
	 in both directions
	14 Feb 2006: reflstats bugfix: don't confuse width with height
	27 Nov 2006: explicitly use xp->xq21 in diffraction-vector calcs
	22 Jun 2006: expand_sdisplay: rimda, magda part of DISP_ATTRS structure
	17 Jun 2006: replace act_pixel calls with activmask(x,y,041);
		replace inappropriate tabs with spaces
	 7 Mar 2006: disallow 16-bit overflows in imhisto
	 5 Oct 2005: housekeeping
	10 Sep 2005: add imhisto(), histogram-binning to calc_ranges();
		reorder other routines; rearrange DISP_ATTRS to handle those
	28 Jun 2005: mkborderolay() and rowcolpts:
		typecast pointer differences on printout
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	 3 Oct 2004: DEFAULT_MULT set to 1.2 and used between color gradations;
		number of colors now externally derived rather than set here;
		size of display determined externally -- may not be 800 or less.
	 2 Sep 2004: routines used only here are declared as static; cosmetics;
		calib_contours() moved here from dispcalc.c
	28 Aug 2004: routines used only here are declared as static; cosmetics;
		calib_contours() to dispcen.c; imedian() to ../xglib/xgenmath.c
	18 Aug 2004: dispatt->di_(lar,lpcor) used internally
	16 Aug 2004: grab_image: sometimes suppress "already loaded" message
	27 Jul 2004: use dimscale() to get the right scale for (X,Y) positions
		in grab_image()
	26 Mar 2004: output fewer diagnostics if di_flags & DISPVERBOSE_ANY==0;
	 more informative diagnostics with detector dimensions are wrong
	24 Jan 2004: ctbuf made a static local in reflstats()
	21 Jan 2004: license message inserted
	 7 Jan 2004: static globals and REFPARS structure elements replace
		global variables; careful handling of array of MAIL structures
	13 Dec 2003: eliminate unused variables maprz from calc_ranges(),
		lhi from mkborderolay(), locct in glocave()
	 8 Nov 2003: move CTROID stuff to dispcen.c; cosmetic changes
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	17 Sep 2003: in grab_image() the error message contains the filename;
		we save dsp[0] and delsp[0] earlier in grespts
	16 Sep 2003: remove xgaccess calls for X-GEN inputs because
		they're latent in the xgopen calls now anyway
	13 Sep 2003: estimate backgrounds using medians, not means;
		routine imedian() written to facilitate that
	10 Sep 2003: expand_sdisplay() gets the box boundaries as
		an array rather than as four separate arguments;
		use xgaccess() on images so that we can uncompress them
	08 Sep 2003: grab_image() no longer takes image name as an arg;
		it's included in the DISP_ATTRS structure.
	05 Sep 2003: enable uncompression of compressed processing files
	01 Sep 2003: cleanups of routines ported from ../motif/dispx.c
	23 Aug 2003: add several routines previously latent in ../motif/dispx.c
	18 Aug 2003: eliminate all global variables
		(convert to static globals or eliminate the need for them)
		and add a few arguments to some routines
	12 Aug 2003: add quickstl() routine
	28 Mar 2003: add `count65k' routine
	 5 Jan 2003: reorganized source-file header
	25 Mar 2002: removed MSI attachments
	12 May 1995: move ddetec_limit(),dresval(),gpars() to disppred.c;
		remove un-needed globals
	12 Apr 1995: xytoh() now returns sin(theta)/lambda
	 2 Mar 1995: remove mainline and arg handler; diagnostics to stdout;
		add frame-number argument to pix2cm
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 5 Oct 1994: use xgread on REFPARS data
	 7 Jul 1994: more informative diagnostics when doing (X,Y) statistics
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: minor bugfixes
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version

   Purpose: calculational engines for the display capability in XGEN.

   Cross references: dispx.c, other routines in this directory.

*************************************************************************/
#include	"displays.h"
#define		MIN_MEDSIZE	3200	/* max elements for median calcs */
#define		MINELTS		100	/* min # of elements for refl stats */
#define		DEFAULT_MULT	1.2	/*default multiplier, 1 color to next*/

static int uspcmp(u_short *usp1, u_short *usp2)
{ /* Compare routine for unsigned shorts */
	return (*usp1 < *usp2 ? -1 : (*usp1 == *usp2 ? 0 : 1));
}

int reflstats(DISP_ATTRS *dispatt, double *istat) 
{ /*	This computes the mean and median value of the data in the region near
	(xc, yc) and uses that in an estimate of the background in the region.
	It then uses that background estimate to determine the (X,Y) centroid
	and the intensity around the point.  The routine returns 0
	if successful, various negative integers on errors. */ 
	int		cpass, dx, dy, lhi, lwi, nelts, neltsbg, neltsc;
	int		no2, no21, x, xc, xmn, xmx, xysq;
	int		y, yc, ymn, ymx, ysq, zy;
	u_short		*usp, *uspe;
	double		bgsum, thisv, sum, xsum, ysum, *isp;
	MAIL		*lhd;
	static u_short	*ctbuf;
	static char	ctbufmsg[] = "ctbuf in reflstats";

	if (NULL == dispatt)
	  {
		memzap((AVOID **)&ctbuf, ctbufmsg);	return 0;
	  }
	lhi = dispatt->di_imht;	lwi = dispatt->di_imwid;
	xc = dispatt->di_currx;	yc = lhi - 1 - dispatt->di_curry;
	lhd = dispatt->di_head + dispatt->di_fmno % dispatt->di_nmbox;
	if ((NULL == ctbuf) && (NULL == (ctbuf = (u_short *)callop(729,
		  sizeof (u_short), ctbufmsg)))) return -1;
	xmn = xc - 13; if (xmn < 0) xmn = 0;
	xmx = xc + 13; if (xmx >= lwi) xmx = lwi - 1;
	ymn = yc - 13; if (ymn < 0) ymn = 0;
	ymx = yc + 13; if (ymx >= lhi) ymx = lhi - 1;
	for (isp = istat; isp < istat + 7; isp++) *isp = 0.;
	isp = istat;
	 /* compute the mean and make a list of count values. */
	for (y = ymn, usp = ctbuf, zy = lhi - ymn - 1,
	 *isp = nelts = 0; y <= ymx; y++, zy--)
		for (x = xmn; x <= xmx; x++, usp++, nelts++)
		   {
			*usp = imvalue(lhd, x, zy); *isp += (double)*usp;
		   }
	if (nelts < MINELTS) return -2;	*isp++ /= (double)nelts;
	no2 = nelts / 2;
	/* sort the count-value list on count value */
	qsort((AVOID *)ctbuf, (unsigned)nelts, sizeof (u_short), (QPE)uspcmp);
	/* the median is the midpoint of this list */
	no21 = no2 - 1;
	if (nelts % 2)	*isp = *(ctbuf + no2);
	else *isp = 0.5 *(((double)(*(ctbuf +no21))) +
		((double)(*(ctbuf +no2))));
	isp++;
	 /* use only the pixels that are below the sum of the
	 mean and the median */
	bgsum = *(isp-2) + *(isp-1);
	for (uspe = ctbuf + nelts - 1; uspe > ctbuf; uspe--)
		if (((double)*uspe) <= bgsum) break;
	if ((neltsbg = uspe - ctbuf) < 50) return -3;
	for (*isp = 0, usp = ctbuf; usp < uspe; usp++) *isp += (double)(*usp);
	*isp /= (double)neltsbg;
	 /* now go through a circle of radius 7 centered on the spot
	   and compute the centroid.  We do this twice--once starting the
	   calculation at the user's starting point and then starting
	   at the centroid calculated on the first pass.  If the intensity
	   is less than 3*sigma on the first pass we don't iterate. */
	for (cpass = 0; cpass < 2; cpass++)
	   {
		xmn = xc - 7;	if (xmn < 0) xmn = 0;
		xmx = xc + 7;	if (xmx >= lwi) xmx = lwi - 1;
		ymn = yc - 7; if (ymn < 0) ymn = 0;
		ymx = yc + 7;	if (ymx >= lhi) ymx = lhi - 1;
		for (y = ymn, bgsum = sum = xsum = ysum = neltsc = 0,
		 zy = lhi - ymn - 1; y <= ymx; y++, zy--)
		   {
			dy = y - yc;	ysq = dy * dy;
			for (x = xmn; x <= xmx; x++)
			   {
				dx = x - xc;	xysq = ysq + dx * dx;
				if (xysq < 49)
				  {
					thisv = imvalue(lhd, x, zy) - *isp;
					neltsc++;	sum += thisv;
					xsum += thisv * dx;
					ysum += thisv * dy;
					bgsum += *isp;
				  }
			   }
		   }
		if (0 >= neltsc) return -4;
		if (0 >= sum)
		  {
			xsum = xc;	ysum = yc;
		  }
		else {
			xsum = xc + xsum / sum;	ysum = yc + ysum / sum;
		     }
		xc = xsum + 0.49999;	yc = ysum + 0.49999;
		bgsum = sqrt(sum + 1.2 * bgsum);
		if (DFUZZ < dispatt->di_lpcor) /* Lorentz/polariz. corr. */
		  {
			sum *= dispatt->di_lpcor; bgsum *= dispatt->di_lpcor;
		  }
		if (sum < 3. * bgsum) break;
	   }
	isp++;	*isp++ = sum;	*isp++ = bgsum;	*isp++ = xsum;
	*isp++ = ysum;	*isp++ = dispatt->di_lpcor;	return 0;
}

#define	FRAMEDELTA	3.5
void rot2l(DISP_ATTRS *dispatt)
{	/* This sets up the local realspace unit-cell matrix used
	in converting from diffraction vectors to hkl's */
	int	fmnu, i, j;
	double	dphi, q3[3][3], inr[3][3], outr[3][3];
	REFPARS	*xp;

	fmnu = dispatt->di_fmno;	xp = dispatt->di_xp;
	dphi = ((double)fmnu + 0.5) * xp->xfrmsize;
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		inr[i][j] = xp->xareal[i][j];
	q3[0][0] = q3[1][1] = cos(dphi);	q3[0][1] = -sin(dphi);
	q3[1][0] = -q3[0][1];			q3[2][2] = 1.;
	q3[0][2] = q3[1][2] = q3[2][0] = q3[2][1] = 0.;
	dmm3(outr, inr, q3);
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		dispatt->di_lar[i][j] = outr[i][j];
	dispatt->di_disflags |= DISPHKL_CONT;
}

static int xytoh(DISP_ATTRS *dispatt, double zhkl[])
{ /*	This takes an integer pixel position (mx, my) and computes an hkl
	from it. If we can't do the conversion it returns 0; it returns 1
	otherwise. It also computes the Lorentz/polarization factor for the
	spot, since that's connected to calculating its diffraction
	vector components. */
	int	fmnu, mx, myy;
	float	xcm, ycm;
	double	rni, rni2, xin, yin, sl[3], ss[3];
	REFPARS	*xp;

	xp = dispatt->di_xp;
	mx = dispatt->di_currx;	myy = dispatt->di_curry;
	fmnu = dispatt->di_fmno;	zhkl[3] = 0.;
	if (1 > pix2cm((float)mx, (float)myy, &xcm, &ycm, fmnu)) return 0;
	xin = xcm - xp->xx0;	yin = ycm - xp->xy0; /* correct main beam */
	rni2 = xin * xin + yin * yin + xp->xpltd2;	rni = sqrt(rni2);
	ss[0] = xp->xplated / rni;
	ss[1] = xin / rni;	ss[2] = yin / rni;
/* rotate this by the detector tilt and swing-angle matrices */
	for (mx = 0; mx < 3; mx++)
	   for (sl[mx] = myy = 0; myy < 3; myy++)
		sl[mx] += xp->xq21[mx][myy] * ss[myy];
	dispatt->di_lpcor = 2. * fabs(sl[1]) /
		(2. - sl[1] * sl[1] - sl[2] * sl[2]);
	sl[0] += 1.;	/* correct the beam-direction component */
	zhkl[3] = sqrt(sl[0] * sl[0] + sl[1] * sl[1] + sl[2] * sl[2]) /
			(2. * xp->xlambda);
	for (mx = 0; mx < 3; mx++)
		for (zhkl[mx] = myy = 0; myy < 3; myy++)
			zhkl[mx] += dispatt->di_lar[mx][myy] * sl[myy];
	return 1;
}

int dget_hkl(DISP_ATTRS *dispatt, double zhkl[4])
{ /*	This obtains the hkl and sin(theta)/lambda value for the reflection
	referenced at dispatt. */
	int	iret;

	dispatt->di_lpcor = 0.;	/* flag so we'll know it failed if it fails ! */
	if (!((dispatt->di_disflags) & DISPHKL_IN))
	  {
		if (!gpars(dispatt)) return 0;
	  }
	if (!((dispatt->di_disflags) & DISPHKL_CONT)) rot2l(dispatt);
	if (((dispatt->di_disflags) & DISPCALIB_CALC) &&
	 ((dispatt->di_disflags) & DISPHKL_CALC))
	  {
		iret = xytoh(dispatt, zhkl);	return iret;
	  }
	return 0;
}

double quickstl(DISP_ATTRS *dispatt, int mx, int myy, int fmnu)
{ /*	This takes an integer pixel position (mx, my) and computes a stl
	from it.  It returns 0 in case of problems. */
	int	i, j;
	float	xcm, ycm;
	double	xin, yin, rni2, rni, rev, sl[3], ss[3];
	REFPARS	*xp;

	rev = 0.;	xp = dispatt->di_xp;
	if (1 > pix2cm((float)mx, (float)myy, &xcm, &ycm, fmnu))
		return rev;
	xin = xcm - xp->xx0;	yin = ycm - xp->xy0; /* correct main beam */
	rni2 = xin * xin + yin * yin + xp->xpltd2;	rni = sqrt(rni2);
	ss[0] = xp->xplated / rni;
	ss[1] = xin / rni;	ss[2] = yin / rni;
/* rotate this by the detector tilt and swing-angle matrices */
	for (i = 0; i < 3; i++)
	   for (sl[i] = j = 0; j < 3; j++) sl[i] += xp->xq21[i][j] * ss[j];
	sl[0] += 1.;	/* correct the beam-direction component */
	rev = sqrt(sl[0] * sl[0] + sl[1] * sl[1] + sl[2] * sl[2]) /
			(2. * xp->xlambda);
	return rev;
}

int glocave(DISP_ATTRS *dispatt, int  *nregvp)
{ /* This finds local <intensities> in each (roughly) 32x32-pixel region*/
	int	*locq, *locqp, *regp, nrx, nry, nrxy, iloc, thisv;
	int	ujy, iregy, ujyma, ujx, iregx, ujxma, jy, jx, lhi, lwi;
	MAIL	*lhd;

	if (0 >= (jx = match_inode(&(dispatt->di_fmind), dispatt->di_imaname)))
		jx = 0;
	lhd = dispatt->di_head + jx % dispatt->di_nmbox;
	lwi = dispatt->di_imwid;	lhi = dispatt->di_imht;
	nrx = lwi / 32;	if (nrx * 32 < lwi) nrx++;
	nry = lhi / 32;	if (nry * 32 < lhi) nry++;
	*nregvp = nrx;	*(nregvp+1) = nry;
	/* if the correct values are already stored, quit */
	if ((dispatt->di_curino != (ino_t)0) &&
	 (dispatt->di_curino == (ino_t)(lhd->quadrant))) return EOK;
	nrxy = nrx * nry;
	if (NULL == dispatt->di_regb)
	  {
		if (NULL == (dispatt->di_regb = (int *)callop(nrxy,
		 sizeof (int), "dispatt->di_regb in glocave"))) return EALLOC;
		if (dispatt->di_disflags & DISPVERBOSEANY) printf(
		 "Created %4d regions, roughly 32x32 pixels in size\n", nrxy);
	  }
	else {
		zeromem((AVOID *)(dispatt->di_regb), nrxy * sizeof (int));
	     }
	if (NULL == (locq = (int *)callop(1500, sizeof (int),
	 "locq in glocave")))
	  {
		fprintf(stderr, "can't create scratch array locq\n");
		return EALLOC;
	  }
	lhd = dispatt->di_head + dispatt->di_fmno % dispatt->di_nmbox;
	for (ujy = iregy = 0, regp = dispatt->di_regb;
	 iregy < nry; iregy++, ujy += 32)
	   {
		ujyma = ujy + 32;
		if (ujyma > dispatt->di_imht) ujyma = dispatt->di_imht;
		for (ujx = iregx = 0; iregx < nrx; iregx++, regp++, ujx += 32)
		   {
			ujxma = ujx + 32;
			if (ujxma > dispatt->di_imwid)
				ujxma = dispatt->di_imwid;
			iloc = 0;
			zeromem((AVOID *)locq, 1500 * sizeof (int));
			for (jy = ujy, locqp = locq; jy < ujyma; jy++)
			   for (jx = ujx; jx < ujxma; jx++)
			      {
				thisv = imvalue(lhd, jx, jy);
				if (thisv > 0)
				  {
					iloc++;	*locqp = thisv;	locqp++;
				  }
			      }
			if (iloc)	*regp = (int)
				(0.7 * (double)imedian(locq, iloc));
			else		*regp = 0;
			if (dispatt->di_disflags & DISPVERBOSEANY)
			  {
				if (!((regp - dispatt->di_regb) % 57))
				  {
					printf(
		" Background in region [%4d,%4d] = %6d\r", ujx, ujy, *regp);
					xeflush(stdout);
				  }
			  }
		   }
	   }
	if (dispatt->di_disflags & DISPVERBOSEANY) printf("\n");
	return EOK;
}

int locba(xglong *locof, int yce, int xce, int nrey, int nrex)
{ /*	this returns an estimate of the background near (xce,yce) as
	a weighted average of four neighboring background values.
	Innovation: we don't include identically zero values in the average
	unless all four are identically zero. */
	int	ba, off00, off01, off10, off11, sumwt, this, thisw;
	int	wtpx, wtpy, wtx, wty, xof, xof0, yof, yof0;

	/* get the indices of the lower-right-hand box nearest us */
	xof = xce + 16;	xof /= 32;	yof = yce + 16;	yof /= 32;
	/* get the indices above and to the left of that */
	xof0 = xof ? xof - 1 : xof;	yof0 = yof ? yof - 1 : yof;
	/* special cases: far right and bottom */
	if (xof >= nrex) xof = xof0;	if (yof >= nrey) yof = yof0;
	/* the overall offsets for the four boxes: */
	off11 = yof * nrex + xof;	off10 = yof * nrex + xof0;
	off01 = yof0 * nrex + xof;	off00 = yof0 * nrex + xof0;
	/* weight factors go like the distances from the opposite centers */
	wtpx = xce % 32 - 16;	if (wtpx < 0) wtpx += 32; wtx = 32 - wtpx;
	wtpy = yce % 32 - 16;	if (wtpy < 0) wtpy += 32; wty = 32 - wtpy;
	if ((wtpx < 0) || (wtpy < 0) || (wtx < 0) || (wty < 0))
		printf(" xce, yce = %4d,%4d: wtpx,py,x,y = %3d,%3d,%3d,%3d\n",
			xce, yce, wtpx, wtpy, wtx, wty);
	/* now the overall background value */
	ba = sumwt = 0;
	if ((this = *(locof + off00)) > 0)
	  {
		thisw = wtx * wty;	sumwt += thisw;	ba += thisw * this;
	  }
	if ((this = *(locof + off10)) > 0)
	  {
		thisw = wtx * wtpy;	sumwt += thisw;	ba += thisw * this;
	  }
	if ((this = *(locof + off01)) > 0)
	  {
		thisw = wtpx * wty;	sumwt += thisw;	ba += thisw * this;
	  }
	if ((this = *(locof + off11)) > 0)
	  {
		thisw = wtpx * wtpy;	sumwt += thisw;	ba += thisw * this;
	  }
	return (sumwt > 0) ? ba / sumwt : 0;
}

int imhisto(DISP_ATTRS *dispatt, int force)
{ /*	This returns EOK if it creates a histogram successfully,
	EALLOC if not */
	int		dih, diw, htup, htdo, jx, jy, jxe, jye, lhi, locbg, lwi;
	int		mav, mx, my, thisval, tval, ujx, ujy, wiup, wido, x, y;
	static int	nregvp[2];
	double		bgmul;
	MAIL		*lhd;
	xglong		*ihv;

	/* compute a simple histogram of background-corrected intensities */
	if (NULL == dispatt->di_hist)
	  {
		dispatt->di_hist = (int *)callop(65536, sizeof (int),
		 "dispatt->di_hist in imhisto");
		if (NULL == (dispatt->di_hist)) return EALLOC;
	  }
	if (0 >= (x = match_inode(&(dispatt->di_fmind), dispatt->di_imaname)))
		x = 0;
	lhd = dispatt->di_head + x % dispatt->di_nmbox;
	/* quit if "force" is false & correct values are already recorded */
	if ((!force) && (dispatt->di_curino != (ino_t)0) &&
	 (dispatt->di_curino == (ino_t)(lhd->quadrant))) return EOK;
	mav = lhd->zmax; lwi = lhd->xdimen; lhi = lhd->ydimen;
	zeromem((AVOID *)(dispatt->di_hist), 65536 * sizeof (int));
	if (mav > 65535) mav = 65535;
	if (EOK != (x = glocave(dispatt, nregvp))) return x;
	wiup = dispatt->di_wiup;	wido = dispatt->di_widown;
	htup = dispatt->di_htup;	htdo = dispatt->di_htdown;
	diw = dispatt->di_wid;		dih = dispatt->di_ht;
	for (ihv = dispatt->di_hist, y = 0; y < 65536; y++, ihv++) *ihv = 0;
	ihv = dispatt->di_hist;	bgmul = dispatt->di_valmul;
	if ((0. > bgmul) || (bgmul > 1.)) bgmul = 0.8;
	for (y = 0; y < dih; y++)
	   {
		my = y / htup;	ujy = my * htdo; jye = (my + 1) * htdo;
		for (x = 0; x < diw; x++)
		   {
			mx = x / wiup;	ujx = mx * wido; jxe = (mx + 1) * wido;
			locbg = locba(dispatt->di_regb, ujy + htdo / 2,
				ujx + wido / 2, nregvp[1], nregvp[0]);
			for (jy = ujy, tval = 0.; jy < jye; jy++)
			   for (jx = ujx; jx < jxe; jx++)
			      {
				thisval = imvalue(lhd, jx, jy) - bgmul * locbg;
				if (thisval < 0) thisval = 0;
				tval += thisval;
			      }
			/* rescale tval by the area
			(in detector pixels squared) of a display pixel */
			tval /= (htdo * wido);	if (tval > 65535) tval = 65535;
			(*(ihv + tval))++;
		   }
	   }
	/* having done both the background calculation and the
	 histogram determination, we remember the inode number of the current
	 image so that we don't needlessly do this multiple times. */
	dispatt->di_curino = (ino_t)lhd->quadrant;
	return EOK;
}

void calc_ranges(DISP_ATTRS *dispatt)
{ /*	This determines the ranges of the color values.  Unlike earlier
	versions, it by default takes the multiplier to be 2 and sets up the
	ranges so that the peak is twice the threshold of the topmost range.
	The depicted intensities are scaled by their merged values, not the
	individual-pixel values. So if the total count in the strongest 2x2
	group of pixels is 160000,
	we use that as the top of the uppermost range. */
  
	int		interval, irange, jnterval, lhi, lwi;
	int		mav, noran, nsofar, x, y;
	xglong		iz, *ihv;
	double		z, prz;
	MAIL		*lhd;

	/* if (1. > dispatt->di_valmul) dispatt->di_valmul = 2.; */
	noran = dispatt->di_nranges;
	prz = dispatt->di_range[noran - 1] = iz = dispatt->di_maxv;
	/* traditional range calculations unless di_valmul < 1 */
	if (dispatt->di_valmul >= 1.)
	  {
		for (irange = noran - 2; irange >= 0; irange--)
		   {
			z = prz / dispatt->di_valmul;
			iz = (z < 0.5 ? 0 : z + 0.499999);
			if (((double)iz) >= prz) iz--;
			if (((int)iz) < irange) iz = irange;
			prz = dispatt->di_range[irange] = iz;
		   }
		return;
	  }
	/* on the other hand, if di_valmul is < 1,
	 we base our color scheme on a histogram */
	if (NULL == dispatt->di_hist) /* grab it if needed */
	  {
		if (EOK != imhisto(dispatt, 0)) return;
	  }
	if (0 >= (x = match_inode(&(dispatt->di_fmind), dispatt->di_imaname)))
		x = 0; /* figure out which image we're using */
	lhd = dispatt->di_head + x % dispatt->di_nmbox;
	mav = lhd->zmax;
	lwi = dispatt->di_wid;		lhi = dispatt->di_ht;
		/* estimate interval between colors */
	jnterval = interval = (lhi * lwi) / ((int)dispatt->di_nranges);
	for (nsofar = y = x = 0, ihv = dispatt->di_hist; x <= mav; x++)
	   {
		nsofar += ihv[x];
		if (nsofar > jnterval)
		  {
			dispatt->di_range[y] = x;
			interval = ((double)(lhi * lwi - nsofar)) /
			 ((double)(dispatt->di_nranges - y));
			jnterval = nsofar + interval;
			y++;	if (y >= dispatt->di_nranges) break;
		  }
	  }
	for ( ; y < dispatt->di_nranges; y++) dispatt->di_range[y] = mav;
}

void disrot(int rotco, int direc, int imlim, int *imco)
{ /*	This rotates the coordinates of a point.
	rotco is the rotation convention (0-7).
	The transformation is from display to detector coords if direc==0.
	it's the other way if direc == 1.  This won't work if width!= height.
	imlim is the upper limit of the coordinates.
	imco is the vector of coordinates:
	imco[0] == x in , imco[1] == y in,
	imco[2] == x out, imco[1] == y out.	*/
	int rotcop;
  
	if ((direc == 1) && ((rotco == 3) || (rotco == 5))) rotcop = 8 - rotco;
	else rotcop = rotco;
	switch (rotcop)
	  {
		case	0:	default:
		 imco[2] = imco[0];	imco[3] = imco[1]; return;
		case	1:
		 imco[2] = imco[1];	imco[3] = imco[0]; return;
		case	2:
		 imco[2] = imlim - 1 - imco[0];	imco[3] = imco[1];	return;
		case	3:
		 imco[2] = imco[1];	imco[3] = imlim - 1 - imco[0];	return;
		case	4:
		 imco[2] = imco[0];	imco[3] = imlim - 1 - imco[1];	return;
		case	5:
		 imco[2] = imlim - 1 - imco[1];	imco[3] = imco[0];	return;
		case	6:
		 imco[2] = imlim - 1 - imco[0];
		 imco[3] = imlim - 1 - imco[1]; return;
		case	7:
		 imco[2] = imlim - 1 - imco[1];	imco[3] = imlim - 1 - imco[0];
		 return;
	  }

}

void fdisrot(int rotco, int direc, int imlim, double *imco)
{ /*	This rotates the coordinates of a point.
	It differs from `disrot' (above) only in that *imco is a double *. */
	int rotcop;
  
	if ((direc == 1) && ((rotco == 3) || (rotco == 5))) rotcop = 8 - rotco;
	else rotcop = rotco;
	switch (rotcop)
	  {
		case	0:	default:
		 imco[2] = imco[0];	imco[3] = imco[1]; return;
		case	1:
		 imco[2] = imco[1];	imco[3] = imco[0]; return;
		case	2:
		 imco[2] = imlim - 1 - imco[0];	imco[3] = imco[1];	return;
		case	3:
		 imco[2] = imco[1];	imco[3] = imlim - 1 - imco[0];	return;
		case	4:
		 imco[2] = imco[0];	imco[3] = imlim - 1 - imco[1];	return;
		case	5:
		 imco[2] = imlim - 1 - imco[1];	imco[3] = imco[0];	return;
		case	6:
		 imco[2] = imlim - 1 - imco[0];
		 imco[3] = imlim - 1 - imco[1]; return;
		case	7:
		 imco[2] = imlim - 1 - imco[1];	imco[3] = imlim - 1 - imco[0];
		 return;
	  }

}

void itoggle(int *ival, int thismask)
{ /* This flips the `thismask' bits in the integer `ival' */
	if ((*ival) & thismask) *ival &= ~thismask;
	else *ival |= thismask;
}

void dorotate(DISP_ATTRS *dispatt, char *im00, char *im0r)
{ /* This rotates the image according to rotation convention roco.
   The input is in im00; the output is in im0r. */
	int	dih, diw, iac, ido, ix, ixx0, iy, iyy0;
	int	nbperp, roco, startoff;
	char	*icp, *ocp, *ocpc;
	short	*isp, *osp, *ospc;
	xglong	*iip, *oip, *oipc;
	PAU	drui, druo;

	if (NULL == im0r)
	  {
		fprintf(stderr, "Sorry. Memory failure in dorotate\n");
		return;
	  }
	roco = dispatt->di_roco;	nbperp = dispatt->di_nbperp;
	if (dispatt->di_wintype)
	  {
		diw = dispatt->di_magw;	dih = dispatt->di_magh;
	  }
	else {
		diw = dispatt->di_wid;	dih = dispatt->di_ht;
	     }
	switch (roco)
	      {
		case	0:	default:	ixx0 = iyy0 = 0;
				iac = 1;	ido = diw;	break;
		case	1:	ixx0 = iyy0 = 0;
				iac = dih;	ido = 1;	break;
		case	2:	ixx0 = 0;	iyy0 = dih - 1;
				iac = -1;	ido = diw;	break;
		case	3:	ixx0 = 0;	iyy0 = dih - 1;
				iac = dih;	ido = -1;	break;
		case	4:	ixx0 = diw - 1;	iyy0 = 0;
				iac = 1;	ido = -diw;	break;
		case	5:	ixx0 = diw - 1;	iyy0 = 0;
				iac = -dih;	ido = 1;	break;
		case	6:	ixx0 = diw - 1;	iyy0 = dih - 1;
				iac = -1;	ido = -diw;	break;
		case	7:	ixx0 = diw - 1;	iyy0 = dih - 1;
				iac = -dih;	ido = -1;	break;
	      }
	startoff = ixx0 * dih + iyy0;
	switch (nbperp)
	  {
		case	1:
			for (icp = im00, iy = 0; iy < dih; iy++)
			   {
				ocpc = im0r + startoff + iy * ido;
				for (ocp = ocpc, ix = 0; ix < diw;
					ix++, icp++, ocp += iac) *ocp = *icp;
			   }
			break;
		case	2:
			drui.ucpt = (u_char *)im00; isp = (short *)(drui.uspt);
			for (iy = 0; iy < dih; iy++)
			   {
				druo.ucpt = (u_char *)im0r;
				ospc = (short *)druo.uspt + startoff + iy *ido;
				for (osp = ospc, ix = 0; ix < diw;
					ix++, isp++, osp += iac) *osp = *isp;
			   }
			break;
		case	4:	default:
			drui.ucpt = (u_char *)im00; iip = (xglong *)(drui.vpt);
			for (iy = 0; iy < dih; iy++)
			   {
				druo.ucpt = (u_char *)im0r;
				oipc = (xglong *)druo.vpt + startoff + iy *ido;
				for (oip = oipc, ix = 0; ix < diw;
					ix++, iip++, oip += iac) *oip = *iip;
			   }
			break;
	     }
	if (roco) printf("data rotated according to scheme %d\n", roco);
}

int grab_image(int darkfla, DISP_ATTRS *dispatt)
{ /*	This reads in the detector image named in dispatt->di_imaname and
	populates the `dispatt' structure therewith. The routine
	returns EOK if the read goes successfully, various errors if not. */
	char		*cp, *inname;
	xglong		thisval, lmaxv;
	int		darkflag, dihi, diwi, lhi, lwi;
	int		maxx, maxy, x, y, readresult;
	double		rowcount, tcount;
	static int	oldwidth, oldheight;
	static char	*shortname;
	MAIL		*lhd;

	inname = dispatt->di_imaname;
	shortname = filepart(inname, 0);	darkflag = darkfla;
	if (dispatt->di_disflags & DISPVERBOSEANY)
		fprintf(stdout, "Opening %s...\n", shortname);
	dispatt->di_imflags = GET_RUNHEADER | darkflag;
	if (0 >= (x = match_inode(&(dispatt->di_fmind), inname))) x = 0;
	lhd = dispatt->di_head + x % dispatt->di_nmbox;
	lhd->frameno = dispatt->di_fmno = x;	/* remember the frame number!*/
	/* we don't actually have to read the image if it's already present */
	/* use xgaccess to see if the image is compressed */
	if (0 != xgaccess(inname, 4, 1))
	  {
		printf("Image %s cannot be opened for reading\n", inname);
		return EOPENERR;
	  }
	if (record_inode(inname) != lhd->quadrant)
	  {
		if (FRAME_OK != (readresult = read_image(inname,
			(int)dispatt->di_imflags, lhd)))
		  {
			(void)image_error(stderr, inname, readresult, 1);
			return EREADERR;
		  }
		if (dispatt->di_imflags & DISPVERBOSEANY)
			printf("Image %s read in\n", inname);
	  }
	else if (dispatt->di_imflags & DISPVERBOSEANY)
		printf("Image %s is already loaded\n", inname);
	/* Crucial point: we need to know the detector width & height
	 at this point, and we need to remember if those values have changed.
	 In particular, if they were nonzero before and they've changed to
	 different nonzero values, we die here. */
	oldwidth = dispatt->di_imwid;	oldheight = dispatt->di_imht;
	dispatt->di_imwid = lwi = lhd->xdimen;
	dispatt->di_imht = lhi = lhd->ydimen;
	if (((oldwidth) && (oldwidth != lwi)) ||
	 ((oldheight) && (oldheight != lhi)))
	  {
		fprintf(stderr,
 "Can't change detector dimensions from [%4d,%4d] to [%4d,%4d] midway\n",
			oldwidth, oldheight, lwi, lhi);
		fprintf(stderr, "Cannot manage data from image %s\n", inname);
		return EREADERR;
	  }
	dispatt->di_wiup = dispatt->di_widown =
		dispatt->di_htup = dispatt->di_htdown = 1;
	dispatt->di_wid = diwi = lwi;	dispatt->di_ht = dihi = lhi;
	if (lwi < 300)
	  {
		for (diwi = lwi * 2; diwi < 300; ) diwi *= 2;
		dispatt->di_wiup = diwi / lwi;	dispatt->di_wid = diwi;
	  }
	else if (lwi > dispatt->di_maxw)
	  {
		for (diwi = lwi / 2; diwi > dispatt->di_maxw; ) diwi /= 2;
		dispatt->di_widown = lwi / diwi;	dispatt->di_wid = diwi;
	  }
	if (lhi < 300)
	  {
		for (dihi = lhi *2; dihi < 300; ) dihi *=2;
		dispatt->di_htup = dihi / lhi;	dispatt->di_ht = dihi;
	  }
	else if (lhi > dispatt->di_maxh)
	  {
		for (dihi = lhi / 2; dihi > dispatt->di_maxh; ) dihi /=2;
		dispatt->di_htdown = lhi / dihi;	dispatt->di_ht = dihi;
	  }
	/* special case: if the environment variable EQUALSCALE is set and
	 is nonzero, then force htdown = widdown and htup = widup */
	if ((NULL != (cp = getenv("EQUALSCALE"))) && ('\0' != *cp))
	  {
		if (dispatt->di_htdown < dispatt->di_widown)
		  {
			dispatt->di_ht *= dispatt->di_htdown;
			dispatt->di_ht /= dispatt->di_widown;
			dispatt->di_htdown = dispatt->di_widown;
		  }
		else if (dispatt->di_widown < dispatt->di_htdown)
		  {
			dispatt->di_wid *= dispatt->di_widown;
			dispatt->di_wid /= dispatt->di_htdown;
			dispatt->di_widown = dispatt->di_htdown;
		  }
		else if (dispatt->di_htup < dispatt->di_wiup)
		  {
			dispatt->di_wid *= dispatt->di_htup;
			dispatt->di_wid /= dispatt->di_wiup;
			dispatt->di_wiup = dispatt->di_htup;
		  }
		else if (dispatt->di_wiup < dispatt->di_htup)
		  {
			dispatt->di_ht *= dispatt->di_wiup;
			dispatt->di_ht /= dispatt->di_htup;
			dispatt->di_htup = dispatt->di_wiup;
		  }
		
	  }
	dispatt->di_xsca = dimscale((short)lwi);
	dispatt->di_ysca = dimscale((short)lhi); 
	dispatt->di_minw = (diwi < 512 ? 512 : diwi);
	dispatt->di_minh = (dihi < 512 ? 512 : dihi);
	/* check the range of values;
	 At the same time estimate the initial base of the topmost color
	 to be the mean count times a fudge factor */
	for (lmaxv = tcount = maxx = maxy = y = 0; y < lhi; y++)
	   {
		for (rowcount = x = 0; x < lwi; x++)
		   {
			if ((thisval = imvalue(lhd, x, y)) > lmaxv)
			  {
				maxx = x; maxy = y; lmaxv = thisval;
			  }
			rowcount += thisval;
		   }
		tcount += rowcount;
	   }
	/* change as of Aug 03: since we're now subtracting a sliding
	 background, we need a smaller value of the maxval */
	if (dispatt->di_maxv < 0.) /* compute di_maxv only to begin with */
	  {
		if ((lhi > 0) && (lwi > 0))
		  {
			dispatt->di_maxv =
			 (xglong)(tcount * 12. / (double)(lhi * lwi));
			dispatt->di_valmul = DEFAULT_MULT;
			if (dispatt->di_disflags & DISPVERBOSEANY)
			 printf("Mean count on detector face = %8.2f\n",
			 tcount / (double)(lhi * lwi));
		  }
		else	dispatt->di_maxv = lmaxv / 2.; /* bizarre! */
	  }
	lhd->zmax = lmaxv;	lhd->zmaxx = maxx;	lhd->zmaxy = maxy;
	/* compute color table */
	if (NULL == dispatt->di_range)
	  dispatt->di_range = (xglong *)callop(2 * dispatt->di_nranges,
		 sizeof (xglong), "dispatt->di_range in grab_image");
	if (EOK != (x = imhisto(dispatt, 0))) return x;
	calc_ranges(dispatt);	return EOK;
}

XG_Point *mkborderolay(DISP_ATTRS *dispatt, XG_Point *xgxp, int  *nptp)
{ /*	This creates the border overlay and returns it.
	If space for it is already available in `xgxp',
	we return it in there. */
	int		ixma, ixmi, iyma, iymi, lroc, lwi, x, y, imco[4];
	static int	maxpts;
	double		magh, magho, magw, magwo;
	short		sx, sy;
	XG_Point	*uxp, *tpt;

	lroc = dispatt->di_roco;	lwi = dispatt->di_imwid;
	if (dispatt->di_wintype)
	  {
		iymi = dispatt->di_xylim[1];	iyma = dispatt->di_xylim[3];
		ixmi = dispatt->di_xylim[0];	ixma = dispatt->di_xylim[2];
		if (ixmi < 1) ixmi = 1;	if (ixma >= lwi) ixma = lwi - 1;
		magh = 8. * ((double)(dispatt->di_htup)) /
			(double)(dispatt->di_htdown);
		magw = 8. * ((double)(dispatt->di_wiup)) /
			(double)(dispatt->di_widown);
		magho = 80. - 8. * ((double)(dispatt->di_disy));
		magwo = 80. - 8. * ((double)(dispatt->di_disx));
	  }
	else {
		iymi = 0;	iyma = dispatt->di_imht;
		ixmi = 1;	ixma = dispatt->di_imwid - 1;
		magh = ((double)(dispatt->di_htup)) /
			(double)(dispatt->di_htdown);
		magw = ((double)(dispatt->di_wiup)) /
			(double)(dispatt->di_widown);
		magho = magwo = 0.;
	     }
	maxpts = 40 * dispatt->di_imwid;
	if (dispatt->di_disflags & DISPVERBOSEANY) printf(
	 "Displaying border from [X,Y] = [%4d,%4d] to [%4d,%4d]: ",
		ixmi, iymi, ixma, iyma);
	if (NULL == xgxp)
	  {
		if (NULL == (uxp = (XG_Point *)callop(maxpts,
		 sizeof (XG_Point), "bordpt in superpose_border")))
		  {
			printf("Can't reserve memory for border XPoints\n");
			return NULL;
		  }
	  }
	else {
		uxp = xgxp; zeromem((AVOID *)uxp, maxpts * sizeof (XG_Point));
	     }
	for (imco[1] = y = iymi, tpt = uxp; y < iyma; y++, (imco[1])++)
	 {
	   for (imco[0] = x = ixmi; x < ixma; x++, (imco[0])++)
	      {
		if (NULL == activmask(x, y, 041))
		  {
			if (NULL != activmask(x + 1, y, 041))
			  {
				disrot(lroc, 1, lwi, imco);
				sx = 0.49999 + magwo +
					((double)(imco[2])) * magw;
				sy = 0.49999 + magho +
					((double)(imco[3])) * magh;
				tpt->xgx = sx - 1;	tpt->xgy = sy; tpt++;
				tpt->xgx = sx;		tpt->xgy = sy; tpt++;
				if (tpt > uxp + (maxpts - 10))
				  {
					fprintf(stderr,
 "Warning: large number of border points. Not all are displayed\n");
					break;
				  }
			  }
			else if (NULL != activmask(x - 1, y, 041))
			  {
				disrot(lroc, 1, lwi, imco);
				sx = 0.49999 + magwo +
					((double)(imco[2])) * magw;
				sy = 0.49999 + magho +
					((double)(imco[3])) * magh;
				tpt->xgx = sx;		tpt->xgy = sy; tpt++;
				tpt->xgx = sx + 1;	tpt->xgy = sy; tpt++;
			  }
		  }
	      }
	    if (tpt > uxp + (maxpts - 10)) break;
	  }
	if (dispatt->di_disflags & DISPVERBOSEANY)
	  {
		if (!(dispatt->di_wintype))
			printf(" %6d border points found\n", (int)(tpt - uxp));
		else 	printf(" %6d border points found in blowup window\n",
			(int)(tpt - uxp));
	  }
	*nptp = tpt - uxp;	return uxp;
}

XG_Point *calibdisp(int *calptp, DISP_ATTRS *dispatt, int ncpts, int *nptp)
{ /*	Sets up the display of calibration contours */
	int		dih, diw, ixma, ixmi, iy, iyma, iymi, x, y, *cap;
	static XG_Point	*lxp, *tpt;
	static int	ncres;
	double		magh, magho, magw, magwo;

	if (dispatt->di_wintype)
	  {
		diw = dispatt->di_magw;	dih = dispatt->di_magh;
		ixmi = dispatt->di_xylim[0];
		iymi = dispatt->di_xylim[1];
		ixma = dispatt->di_xylim[2];
		iyma = dispatt->di_xylim[3];
		magh = 8. * ((double)(dispatt->di_htup)) /
			(double)(dispatt->di_htdown);
		magw = 8. * ((double)(dispatt->di_wiup)) /
			(double)(dispatt->di_widown);
		magho = 80. - 8. * (double)(dispatt->di_disy);
		magwo = 80. - 8. * (double)(dispatt->di_disx);
		dispatt->di_boxh = 8;
	  }
	else {
		diw = dispatt->di_wid;	dih = dispatt->di_ht;
		dispatt->di_xylim[0] = dispatt->di_xylim[1] = ixmi = iymi = 0;
		dispatt->di_xylim[2] = ixma = dispatt->di_imwid;
		dispatt->di_xylim[3] = iyma = dispatt->di_imht;
		magh = ((double)(dispatt->di_htup)) /
			(double)(dispatt->di_htdown);
		magw = ((double)(dispatt->di_wiup)) /
			(double)(dispatt->di_widown);
		magho = magwo = 0.;	dispatt->di_boxh = 4.;
	     }
	if ((NULL != lxp) && (ncres < ncpts))
	  {
		memzap((AVOID **)&lxp, "lxp in calibdisp");
		printf("eliminating old lxp memory and creating new\n");
		ncres = 0;
	  }
	if (NULL == lxp)
	  {
		ncres = 4 * ncpts;
		if (NULL == (lxp = (XG_Point *)callop(ncres,
		 sizeof (XG_Point), "lxp in calibdisp")))
		  {
			printf("Memory alloc problem in calibdisp\n");
			return 0;
		  }
	  }
	for (iy = 0, cap = calptp, tpt = lxp; iy < ncpts; iy++, cap++)
	   {
		y = (*cap) >> 16;	x = (*cap) & 65535;
		if ((iymi <= y) && (y < iyma) && (ixmi <= x) && (x <= ixma))
		  {
			y = ((double)y) * magh + magho;
			x = ((double)x) * magw + magwo;
			if ((0 < x) && (x < diw-1) && (0 < y) && (y < dih-1))
			  {
				tpt->xgx = (short)x;
				tpt->xgy = (short)y;	tpt++;
			  }
		  }
	   }
	*nptp = tpt - lxp;
	printf(" %6d contour points set up in the %s window\n",
		*nptp, dispatt->di_wintype ? "magnified" : "image");
	return lxp;
}

XG_Point *grespts(DISP_ATTRS *dispatt, int *resinfo)
{ /*	This finds the points the lie close to particular resolution bins.
	It returns a pointer to the list of drawable points.
	It also returns the array resinfo, which contains:
	resinfo[0] == number of drawable (X,Y) points (npts)
	resinfo[1] == number of resolution circles to draw (ngps)
	resinfo[2, . . . (1+ngps)] == minxc array (see usage below)
	resinfo[(2 + ngps) . . . (1 +2*ngps)] == minyxc array (see usage below)
	resinfo[(2+2*ngps) . . . (1 +3*ngps)] == dspacing for circles * 100 */
	int		dihi, diwi, igp, ispt, ix, ixd, ixma, ixmi;
	int		iy, iyd, iyma, iymi, jx, jy, lfm, lhi;
	int		lix, liy, lwi, maxjx, maxjy, maxpt, minjx, minjy;
	int		*ipt, *jpt, imcoo[4];
	double		dspac, dstl, fleft, magh, magho, magw, magwo;
	double		savedel, savedsp;
	static XG_Point *lxp, *tpt;
	static int	ngps, npts, isgp[3], minxc[3], minyxc[3];
	static int	lngps = 3;
	static double	dsp[] = { 6., 3., 2.};
	static double	delsp[] = { 0.02, 0.006, 0.004 };

	lhi = dispatt->di_imht;	lwi = dispatt->di_imwid; /* detector dimen.*/
	if (dispatt->di_wintype)
	  {
		dihi = dispatt->di_magh;	diwi = dispatt->di_magw;
	  }
	else {
		dihi = dispatt->di_ht;	diwi = dispatt->di_wid;
	     }
	lfm = dispatt->di_fmno;
	maxpt = (dispatt->di_wintype) ? 1000 : 100 * lwi; /* memory required */
	if (NULL == lxp) /* reserve memory for point-list if needed */
	  {
		lxp = (XG_Point *)callop(maxpt, sizeof (XG_Point),
		 "respts in grespts");
	  }
	else zeromem((AVOID *)lxp, maxpt * sizeof (XG_Point)); /* zero it out */
	/* keep track of minimum & maximum pixel (X,Y) values marked */
	minjx = minjy = 32763;
	maxjx = maxjy = -32763;
	/* save the permanent values of dsp[0] and delsp[0] */
	ngps = 1;	savedsp = dsp[0];	savedel = delsp[0];
	/* if we're dealing with the magnification window, we just do one
	 arc--the one corresponding to the point in the center of the display */
	if (dispatt->di_wintype)
	  {
		iymi = dispatt->di_xylim[1];	iyma = dispatt->di_xylim[3];
		ixmi = dispatt->di_xylim[0];	ixma = dispatt->di_xylim[2];
		if (ixmi < 1) ixmi = 1;	if (ixma >= lwi) ixma = lwi - 1;
		/* establish scale factors that define the (X,Y) positions */
		magh = 8. * ((double)(dispatt->di_htup)) /
			(double)(dispatt->di_htdown);
		magw = 8. * ((double)(dispatt->di_wiup)) /
			(double)(dispatt->di_widown);
		magho = 80. - 8. * ((double)(dispatt->di_disy));
		magwo = 80. - 8. * ((double)(dispatt->di_disx));
		iyd = ixd = 1;
		lix = dispatt->di_currx;	liy = dispatt->di_curry;
		/* find the sin(theta)/lambda value of the blowup's center */
		dstl = quickstl(dispatt, lix, liy, lfm);
		if (0 > dstl)
		  {
			fprintf(stderr,
 "Apparent sin(theta)/lambda = %11.4f; No resolution arc plotted\n", dstl);
			return NULL;
		  }
		dsp[0] = (0 < dstl) ? 0.5 / dstl : 4.;	/* D spacing for that */
		/* we plot just a few points on either side of that one. */
		if (abs(lix - lwi / 2) > abs(liy - lhi / 2))
			fleft = quickstl(dispatt, lix - 1, liy, lfm);
		else	fleft = quickstl(dispatt, lix, liy - 1, lfm);
		if (0 > fleft) delsp[0] = 0.004;
		else if (fleft < dstl) delsp[0] = 0.5 / fleft - dsp[0];
		else delsp[0] = dsp[0] - 0.5 / fleft;
	  }
	else {
		iymi = 0;	iyma = dispatt->di_imht;
		ixmi = 1;	ixma = dispatt->di_imwid - 1;
		magh = ((double)(dispatt->di_htup)) /
			(double)(dispatt->di_htdown);
		magw = ((double)(dispatt->di_wiup)) /
			(double)(dispatt->di_widown);
		magho = magwo = 0.;
		ngps = lngps;	ixmi = iymi = 0; iyma = lhi; ixma = lwi;
		/* don't look at every point: step among points */
		iyd = (lhi < 500) ? 1 : (lhi < 2000) ? 2 : 4;
		ixd = (lwi < 500) ? 1 : (lwi < 2000) ? 2 : 4;
	     }
	/* step through possible (X,Y) values and
	 mark any given pixel as overlaid _if_ the D spacing
	 corresponds to one of the values desired */
	for (igp = 0; igp < ngps; igp++) minxc[igp] = 10000;
	for (npts = 0, iy = iymi, tpt = lxp; iy < iyma; iy += iyd)
	 {
	   for (ix = ixmi; ix < ixma; ix += ixd)
	      { /* calculate the sin(theta)/lambda value */
		if (0 < (dstl = quickstl(dispatt, ix, iy, lfm)))
		  {
			dspac = 0.5 / dstl; /* convert to D spacing */
			/* if D is almost exactly 6, 3, or 2, mark */
			ispt = 0;
			for (igp = 0; igp < ngps; igp++) isgp[igp] = 0;
			for (igp = 0; igp < ngps; igp++)
			   {
				if (fabs(dspac -dsp[igp]) < delsp[igp])
				  {
					ispt = isgp[igp] = 1;	break;
				  }
			   }
			if (ispt)
			  { /* do the rotation into this image */
				imcoo[0] = ix;	imcoo[1] = iy;
				disrot(dispatt->di_roco, 1, lwi, imcoo);
				/* convert to display coords */
				jx = imcoo[2] * magw + magwo;
				jy = imcoo[3] * magh + magho;
				/* point has to be visible */
				if ((0 < jx) && (jx < diwi) &&
				 (0 < jy) && (jy < dihi))
				  {	/* don't overrun array */
					if (npts >= maxpt - 1) break;
					/* save X,Y in XG_Point struct*/
					tpt->xgx = (short)jx;
					tpt->xgy = (short)jy;
					/* establish (X,Y) extrema */
					if (jx < minjx) minjx = jx;
					if (jy < minjy) minjy = jy;
					if (jx > maxjx) maxjx = jx;
					if (jy > maxjy) maxjy = jy;
					for (igp = 0;igp < ngps;igp++)
					   if ((isgp[igp]) &&
					    (jx < minxc[igp]))
					     {
						minxc[igp] = jx;
						minyxc[igp] = jy;
						break;
					     }
					/* increment pointer&counter*/
					npts++;	tpt++;
				  }
			  } /* end if "if ((fabs..." clause */
		  } /* end of "if (0 < (dstl..." clause */
	      } /* end of loop through ix values */
	  if (npts >= maxpt - 1) break;
	 } /* end of loop through iy values */
	if (dispatt->di_wintype) printf(
		 "%6d pixels contained in the arc near D = %8.2f Angstroms\n",
			npts, dsp[0]);
	else printf("%6d pixels are contained in the (6,3,2) Angstrom rings\n",
		npts);
	/* populate the "resinfo" array */
	ipt = resinfo;	*ipt++ = npts;	*ipt++ = ngps;
	for (jpt = minxc, igp = 0; igp < ngps; igp++) *ipt++ = *jpt++;
	for (jpt = minyxc, igp = 0; igp < ngps; igp++) *ipt++ = *jpt++;
	for (igp = 0; igp < ngps; igp++, ipt++)
		*ipt = (int)(dsp[igp] * 100. + 0.4999);
	if (dispatt->di_wintype)
	  {
		dsp[0] = savedsp;	delsp[0] = savedel;
	  }
	return lxp;
}

XG_Point *rowcolpts(DISP_ATTRS *dispatt, int *nrcptp)
{ /*	This creates (X,Y) points indicative of the sum of the intensities
	along each row and column on the detector face. */
	int		dihi, diwi, ix, iy, jx, jy, lhi, lwi;
	double		maxcol, maxpt, maxrow, scalev, tsum;
	float		*tsu, *tsv;
	XG_Point	*lxp, *tpt, *tpq;
	MAIL		*lhd;

	if (dispatt->di_wintype)
	  {
		dihi = dispatt->di_magh;	diwi = dispatt->di_magw;
	  }
	else {
		dihi = dispatt->di_ht;		diwi = dispatt->di_wid;
	     }
	lhi = dispatt->di_imht;	lwi = dispatt->di_imwid;
	lhd = dispatt->di_head + dispatt->di_fmno % dispatt->di_nmbox;
	maxpt = 8 * (dihi + diwi);
	lxp = (XG_Point *)callop(maxpt, sizeof (XG_Point),
		"lxp in rowcolpts");
	tsu = (float *)callop(dihi + diwi, sizeof (float),
		"tsu in rowcolpts");
	tsv = tsu + dihi;
	printf("Performing row and column sums:\n");
	for (iy = 0; iy < lhi; iy++)
	   {
		for (tsum = ix = 0; ix < lwi; ix++)
			tsum += (double)(imvalue(lhd, ix, iy));
		jy = iy * dispatt->di_htup / dispatt->di_htdown;
		if ((jy < 0) || (jy > dihi)) printf(
" Attempt to sum outside y range: iy, jy = %5d, %5d\n", iy, jy);
		else tsu[jy] += tsum;
	   }
	for (maxrow = jy = 0; jy < dihi; jy++)
		if (tsu[jy] > maxrow) maxrow = tsu[jy];
	printf("Maximum row = %13.4e; ", maxrow);
	scalev = 0.95 * ((double)dihi) / maxrow;
	for (jy = 0, tpt = lxp; jy < dihi; jy++)
	   {
		tsu[jy] *= scalev;	ix = tsu[jy];
		if ((0 <= ix) && (ix <= diwi))
		  {
			tpt->xgy = dihi - 1 - jy;
			tpt->xgx = ix;	tpt++;
		  }
		else printf(
		 "Warning: can't display value tsu[%4d] = %6.1f\n",
			jy, tsu[jy]);
	   }
	printf(" %4d row values displayed\n", (int)(tpt - lxp));
	zeromem((AVOID *)tsv, diwi * sizeof (float));
	for (ix = 0; ix < lwi; ix++)
	   {
		for (tsum = iy = 0; iy < lhi; iy++)
			tsum += (double)(imvalue(lhd, ix, iy));
		jx = ix * dispatt->di_wiup / dispatt->di_widown;
		if ((jx < 0) || (jx > diwi)) printf(
 " Attempt to sum outside x range: ix, jx = %5d, %5d\n", ix, jx);
		else tsv[jx] += tsum;
	   }
	for (maxcol = jx = 0; jx < diwi; jx++)
		if (tsv[jx] > maxcol) maxcol = tsv[jx];
	printf(" Maximum column = %13.5e; ", maxcol);
	scalev = 0.95 * ((double)diwi) / maxcol;
	tpq = tpt;
	for (jx = 0; jx < diwi; jx++)
	   {
		tsv[jx] *= scalev;	iy = tsv[jx];
		if ((0 <= iy) && (iy <= dihi))
		  {
			tpt->xgy = dihi - 1 - iy;
			tpt->xgx = jx;	tpt++;
		  }
		else printf(
		 "Warning: can't display value tsv[%4d] = %6.1f\n",
			jx, tsv[jx]);
	   }
	printf(" %4d column values displayed\n", (int)(tpt - tpq));
	*nrcptp = tpt - lxp;
	memzap((AVOID **)&tsu, "tsu in rowcolpts");
	return lxp;
}

void expand_sdisplay(DISP_ATTRS *dispatt, int bound[])
{ /* This blows up the already-extant display data from the small size to
	the larger one approprirate to the blowup window. We have to do
	this differently for different numbers of bytes per pixel */
	int	diw, dx, dy, ix, ixma, ixmi, iy, iymi, iyma, magw;
	int	nbperp, x, xmn, xmx, y, ymn, ymx, zv[8];
	char	*icp, *magda, *ocp, *ocpy, *rimda;
	short	*isp, *osp, *ospy;
	xglong	*iip, *olp, *olpy;
	PAU	xsui, xsuo;

	nbperp = dispatt->di_nbperp;
	diw = dispatt->di_wid;	magw = dispatt->di_magw;
	xmn = bound[0];	ymn = bound[1];	xmx = bound[2];	ymx = bound[3];
	zv[0] = xmn * dispatt->di_widown / dispatt->di_wiup;
	zv[1] = ymn * dispatt->di_htdown / dispatt->di_htup;
	zv[4] = xmx * dispatt->di_widown / dispatt->di_wiup;
	zv[5] = ymx * dispatt->di_htdown / dispatt->di_htup;
	disrot((int)(dispatt->di_roco), 0, (int)(dispatt->di_imwid), zv);
	disrot((int)(dispatt->di_roco), 0, (int)(dispatt->di_imwid), &(zv[4]));
	if (dispatt->di_disflags & DISPVERBOSEANY) printf(
	 "Blowup in detector region [%4d,%4d] to [%4d,%4d]\n",
	 zv[2], zv[3], zv[6], zv[7]);
	magda = dispatt->di_magm;	rimda = dispatt->di_rotm;
	if (nbperp == 1)
	  {
	    for (dy = 0, y = ymn; y <= ymx; y++, dy++)
	       {
		icp = rimda + diw * y + xmn;
		iymi = dy * 8;	iyma = iymi + 7;
		for (x = xmn, dx = 0; x <= xmx; x++, dx++, icp++)
		   {
			ixmi = dx * 8;	ixma = ixmi + 7;
			ocpy = magda + iymi * magw + ixmi;
			for (iy = iymi; iy <= iyma; iy++, ocpy += magw)
			   {
				for (ix = ixmi, ocp = ocpy; ix <= ixma; ix++)
					*ocp++ = *icp;
			   } /* close of "for (iy" loop */
		   } /* close of "for (x" loop */
	       } /* close of "for (dy " loop */
	    } /* close of "if (nbperp == 1)" clause */
	   else if (nbperp == 2)
	    {
	     for (dy = 0, y = ymn; y <= ymx; y++, dy++)
	      {
		xsui.ucpt = (u_char *)rimda;
		isp = (short *)xsui.uspt + diw * y + xmn;
		iymi = dy * 8;	iyma = iymi + 7;
		for (x = xmn, dx = 0; x <= xmx; x++, dx++, isp++)
		   {
			ixmi = dx * 8;	ixma = ixmi + 7;
			xsuo.ucpt = (u_char *)magda;
			ospy = (short *)xsuo.uspt + iymi * magw + ixmi;
			for (iy = iymi; iy <= iyma; iy++, ospy += magw)
			   {
				for (ix = ixmi, osp = ospy; ix <= ixma; ix++)
					*osp++ = *isp;
			   } /* close of "for (iy" loop */
		   } /* close of "for (x" loop */
	       } /* close of "for (dy " loop */
	    } /* close of "else if (nbperp == 2)" clause */
	   else
	    {
	     for (dy = 0, y = ymn; y <= ymx; y++, dy++)
	      {
		xsui.ucpt = (u_char *)rimda;
		iip = (xglong *)xsui.vpt + diw * y + xmn;
		iymi = dy * 8;	iyma = iymi + 7;
		for (x = xmn, dx = 0; x <= xmx; x++, dx++, iip++)
		   {
			ixmi = dx * 8;	ixma = ixmi + 7;
			xsuo.ucpt = (u_char *)magda;
			olpy = (xglong *)xsuo.vpt + iymi * magw + ixmi;
			for (iy = iymi; iy <= iyma; iy++, olpy += magw)
			   {
				for (ix = ixmi, olp = olpy; ix <= ixma; ix++)
					*olp++ = *iip;
			   } /* close of "for (iy" loop */
		   } /* close of "for (x" loop */
	       } /* close of "for (dy " loop */
	    } /* close of "else " clause */
}

/* end of dispcalc.c */
