/* dispcen.c -- calculations for CENTROIDS & CALCENTROIDS overlaid on images

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions; reorganize use of "{"
	31 Dec 2005: handle qsort with < 2 elements correctly
	 5 Oct 2005: housekeeping
	 4 Sep 2004: calib_contours() moved here from dispcalc.c; cosmetics
	21 Jan 2004: license message inserted
	 9 Jan 2004: framrad replaced by REFPARS structure element from caller
	10 Dec 2003: eliminate framdeg as a variable
	 8 Nov 2003: move the CTROID stuff from dispcalc.c
	19 Sep 2003: in gettroids(), with ntroids > LOTS_OF_CENTROIDS,
		we initialize trp (as trb) properly
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Sep 2003: remove xgaccess() calls for X-GEN processed data
		because they're latent in the xgopen() calls
	 5 Sep 2003: enable uncompression of compressed CENTROIDS files
	27 Aug 2003: change how gettroids() returns centroids to the caller;
	 increase LOTS_OF_CENTROIDS to 5000.
	18 Aug 2003: eliminate references to external variables framenum, trb;
		add a few args to routines accordingly
	 5 Jan 2003: reorganized source-file header
	25 Mar 2002: removed MSI attachments
	 8 Nov 1996: centroids displays made more useful;
		minor bugfixes here and there
		cosmetic cleanups
		(remove unnecessary declarations, unused #defines)
	 2 Mar 1995: diagnostics to stdout;
		add frame-number argument to pix2cm
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 5 Oct 1994: use xgread on TROID data
	 5 May 1994: source-file header cleaned up
		names changed from POLYDET to XGEN
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version

   Purpose: calculations enabling display of the contents of a CENTROIDS file
	as an overlay on a detector-image display in XGEN.
   Cross references: dispcalc.c

*************************************************************************/
#include	"displays.h"
#define	LOTS_OF_CENTROIDS	(5000)

static int framecmp(TROID *tr1, TROID *tr2)
{ /* compare routine for frame numbers in a TROID array */
	return (tr1->frame0 < tr2->frame0 ? -1 : 1);
}

TROID *gettroids(DISP_ATTRS *dispatt, int fmnu, int *ncenp)
{ /*	This returns a pointer to a list of centroids in the vicinity of the
	current image. The # of centroids in the list is returned in *ncenp. */
	int		whither;
	REFPARS		*xp;
	TROID		*tr0, *tre, *trl, *trp;
	unsigned	nin;
	double		astep, frame_min, frame_max, framedel;
	static char	*cename;	/* name of centroids file */
	static FILE	*fpc;	/* centroids file pointer */
	static TROID	*trb, tr;
	static unsigned	ntroids, ntosave;

	if (NULL == dispatt->di_xp) dispatt->di_xp = safepars(&whither);
	xp = dispatt->di_xp;
	if (!ntroids) /* we haven't counted # of  centroids in the file yet */
	  {
		cename = trname("CENTROIDS");
		if (NULL == (fpc = xgopen(cename, "r")))
		  {
			fprintf(stderr,
			 "Can't open %s as input centroids file\n", cename);
			return NULL;
		  }
		/* see how big the centroids file is.  If it's huge,
		we just "dip" into it to get the centroids we need.
		Otherwise, we read all the centroids into memory */
		for (nin = 0; SIZREF <= xgread((void *)&tr,XG_TROID,fpc); )
		 nin++;
		printf("File %s has %6d centroids in it\n", cename, (int)nin);
		if (nin < 1)
		  {
			fprintf(stderr, "The centroids file is empty!\n");
			return NULL;
		  }
		if (-1 == fseek(fpc, 0L, 0)) /* try to rewind */
		  {
			fprintf(stderr, "Can't rewind CENTROIDS file\n");
			return NULL;
		  }
		ntroids = nin;
		ntosave = LOTS_OF_CENTROIDS;
		if (ntosave > ntroids) ntosave = ntroids;
		if (NULL == (trb = (TROID *)callop(ntosave, sizeof (TROID),
		 "trb in gettroids")))
		  {
			fprintf(stderr, "Can't reserve memory for troids\n");
			return NULL;
		  }
		if (nin < LOTS_OF_CENTROIDS) /* a few centroids: read now */
		  {
			for (trp = trb; SIZREF <= xgread((void *)trp,
			 XG_TROID, fpc); ) trp++;
			(void)fclose(fpc);	/* done with the file */
			/* this list is mostly sorted on frame #, but not
			exactly.  Sort it properly */
			if (nin > 1) qsort((AVOID *)trb, nin,
				sizeof (TROID), (QPE)framecmp);
		  }
	  }
	if (-1 == fmnu)
	  {
		fprintf(stderr,
		 "Can't deal with CENTROIDS: no frame index set up yet\n");
		return NULL;
	  }
	/* if we know the stepsize between images, we use it to determine
	 how many images' worth of centroids to display. If we don't, we
	 assume the user knows what he's about and just use about 5.
	 Specifically, if the frame width is < 0.5 degrees, we display 5 frames;
	 We reduce that down gradually to 1 as the stepsize goes up. */
	if (((astep = fabs(xp->xfrmsize)) > 0.5) && (astep < 10.))
		framedel = (-8. * astep + 19.) / 6.;
	else	framedel = 2.5;
	frame_min = (double)fmnu + 0.5 - framedel;
	frame_max = fmnu + framedel + framedel;
	if (ntroids < LOTS_OF_CENTROIDS) /* all troids in memory now */
	  { /* in this case we just find 1st & last relevant refls.
		If I were ambitious this would be done with a tree search.
		As it is I'll just walk through the list */
		trl = trb + ntroids;
		for (trp = trb; trp < trl; trp++)
			if (trp->frame0 > frame_min) break;
		if (trp >= trl)
		  {
			printf("No centroids close to current image\n");
			return NULL;
		  }
		/* at this point trp points to the first TROID that
		 is within range of the current frame. Remember it. */
		tr0 = trp;	trp = trl - 1;
		for (trp = trl - 1; trp > tr0; trp--)
			if (trp->frame0 < frame_max) break;
		/* now trp points at the last TROID within range. */
		tre = trp + 1;
	  }
	else	/* otherwise we have to read 'em in from the file */
	  {
		if (-1 == fseek(fpc, 0L, 0))
		  {
			fprintf(stderr, "Can't rewind CENTROIDS file\n");
			return NULL;
		  }
		for (trp = trb; SIZREF < xgread((void *)&tr, XG_TROID, fpc); )
		   if (frame_min < tr.frame0)
		     {
			if (tr.frame0 < frame_max)
			  {
				cpymem((AVOID *)trp, (AVOID *)&tr,
				 sizeof (TROID));
				trp++;
				if (trp - trb > (int)ntosave)
				  {
					printf(
	 "Warning: ran out of memory. Not all centroids displayed\n");
					break;
				  }
			  } /* end of "if (tr.frame0" clause */
			else break;
		     } /* end of loop through reads */
		tr0 = trb;	tre = trp;
	  }
	*ncenp = tre - tr0;	return tr0;
}

CTROID *get_calcen(int *ncalcep)
{ /*	This reads in the calibration centroids from a file and returns a
	pointer to the list of centroids thus obtained.
	The pointer ncalcep contains the number of centroids */
	char		calin[40];
	int		iux, iuy, ncalce, nitems;
	FILE		*fctr;
	CTROID		*lcb, *catr;

	*ncalcep = 0;
	if (NULL == (fctr = xgopen("calcentroids", "r")))
	  {
		fprintf(stderr, "Cannot open CALCENTROIDS\n");	return NULL;
	  }
	for (ncalce = 0; NULL != fgets(calin, 36, fctr); ) ncalce++;
	if (ncalce < 1)
	  {
		fprintf(stderr, "no calibration centroids found in file\n");
		return NULL;
	  }
	if (NULL == (lcb = (CTROID *)callop(ncalce + 2,
		sizeof (CTROID), "lcb in get_calcen")))
	  {
		fprintf(stderr, "Cannot reserve memory for CALCENTROIDS\n");
		return NULL;
	  }
	if (-1 == fseek(fctr, 0L, 0))
	  {
		fprintf(stderr, "Cannot rewind CALCENTROIDS\n"); return NULL;
	  }
	for (catr = lcb; NULL != fgets(calin, 36, fctr); )
		if (calin[0] != '#')
		  {
			nitems =- sscanf(calin, "%lf %lf %d %d",
			 &catr->cxcen, &catr->cycen, &iux, &iuy);
			catr->ux = iux;	catr->uy = iuy;
			if (nitems < 4) catr->uy = 0;
			if (nitems < 3) catr->ux = 0;
			catr++;
		  }
	ncalce = catr - lcb;	(void)fclose(fctr);
	fprintf(stderr, "%4d calibration centroids read in\n", ncalce);
	*ncalcep = ncalce;	return lcb;
}

int *calib_contours(DISP_ATTRS *dispatt, int *ncaptp)
{ /*	This creates an overlay containing the calibration data contained
	in the USPLINE file. Unlike the earlier version, it returns the array
	of calibration values, and returns the count of points as *ncaptp.
	We stick some code values into *ncaptp on error */
 
	int		ixs, ixs2, ixso, iys, iys2, iyso;
	int		lhi, lwi, ntosave, x, y, *cap;
	static int	*calpts, ncpts;
	float		xcm, xoff, ycm, yoff;
	VSPLINE		*lsp;

	*ncaptp = 0;	lwi = dispatt->di_imwid;	lhi = dispatt->di_imht;
	if (!((dispatt->di_disflags) & DISPCALIB_CALC))
	  {
		printf("Obtaining calibration information: ");
		xeflush(stdout);
		if (NULL == (dispatt->di_spl = gspline(5, &x)))
		  {
			fprintf(stderr,
 "Unable to read pix-to-cm spline information: error %d\n", x);
			*ncaptp = -2;	return NULL;
		  }
		lsp = dispatt->di_spl;
		dispatt->di_disflags |= DISPCALIB_IN;
		printf("done\n");
	  }
	/* create contours of constant centimeter-values.  We use the
	inverse fiducial positions (0.508 cm apart) as the separations,
	but we offset them by half a point so we don't obscure the pattern */
	/* to save time, we set up an array to hold these contours and
	 only display them when we need 'em */
	lsp = dispatt->di_spl;
	if (!((dispatt->di_disflags) & DISPCALIB_CONT))
	  {
		printf("Contouring calibration...");
		xeflush(stdout);
		ntosave = 4 * (lsp->v_ncol * lhi + lsp->v_nrow * lwi);
		if (NULL == (calpts = (int *)callop((unsigned)ntosave,
			sizeof (int), "calpts in calib_contours")))
		  {
			*ncaptp = -3;	return NULL;
		  }
		xoff = lsp->v_x0cm - lsp->v_xrcm / 2.;
		yoff = lsp->v_y0cm - lsp->v_yrcm / 2.;
		for (ncpts = x = 0, cap = calpts; x < lwi; x++)
		   {
			(void)pix2cm((float)x, 0., &xcm, &ycm, 0);
			iyso = (ycm - yoff) / lsp->v_yrcm;
			if (iyso < 0) iyso = 0;	iys2 = iyso;
			for (y = 1; y < lhi; y++)
			   {
				(void)pix2cm((float)x, (float)y, &xcm, &ycm, 0);
				iys = (ycm - yoff) / lsp->v_yrcm;
				if (iys < 0) iys = 0;
				if (iys != iyso)
				  {
					if (iys != iys2)
					  {
						*cap++ = ((y-1) << 16) | x;
						*cap++ = (y << 16) | x;
						ncpts += 2;
					  }
					 iys2 = iyso;	iyso = iys;
				  }
			   }
		   }
		printf("%5d contour points from rows\n", ncpts);
		for (y = 0; y < lhi; y++)
		   {
			(void)pix2cm(0., (float)y, &xcm, &ycm, 0);
			ixso = (xcm - xoff) / lsp->v_xrcm;
			if (ixso < 0) ixso = 0;	ixs2 = ixso;
			for (x = 1; x < lwi; x++)
			   {
				(void)pix2cm((float)x, (float)y, &xcm, &ycm, 0);
				ixs = (xcm - xoff) / lsp->v_xrcm;
				if (ixs < 0) ixs = 0;
				if ((ixs != ixso))
				  {
					if (ixs != ixs2)
					  {
						*cap++ = (y << 16) | (x-1);
						*cap++ = (y << 16) | x;
						ncpts += 2;
					  }
					ixs2 = ixso;	ixso = ixs;
				  }
			   }
		   }
		printf("%5d total contour points\n", ncpts);
		dispatt->di_disflags |= DISPCALIB_CONT;
	  }
	*ncaptp = ncpts;	return	calpts;
}

/* end of dispcen.c */
