/* dispmref.c -- calculations for displaying contents of XGEN multirefs

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	22 Jun 2006: exam_mref: idata, limits part of DISP_ATTRS structure
	 5 Oct 2005: housekeeping
	30 Jun 2005: exam_mref(): read exactly 256 bytes in for each DATARUN
	 3 Oct 2004: 32 I/sigma ranges (ioslim[]), not 8
	28 Aug 2004: replace zapscales() with repeated explicit calls;
		routines used only here are declared as static; cosmetics
	 4 Mar 2004: replace gspline() call with cleanspline() call--safer!
	24 Jan 2004: several static globals made local;
		sgsp replaced with dispatt->di_spl
	21 Jan 2004: license message inserted
	 9 Jan 2004: globals replaced by elements of DISP_ATTRS structures
		from function arguments and some static globals;
		changes to args to find_shift(), gspline(), etc.;
		replace shcalloc() & shfree() calls with callop() & memzap()
	13 Dec 2003: eliminate unused vbl diwi in exact_mrxy()
	19 Sep 2003: careful management of default values in exam_mref()
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Sep 2003: remove explicit xgaccess() calls because
		they're latent in the xgopen() calls anyway
	14 Sep 2003: handle overflows and underflows more carefully
	 4 Sep 2003: enable uncompression of compressed X-GEN output files
	25 Aug 2003: moved from ../motif after removing Motif-specific stuff
	17 Aug 2003: gauge size of display using locally-calculated variable
		called `disparea' rather than `imarea'
	16 Aug 2003: get rid of global variables afmin and afmax
		at the price of two additional elements of the limits[] array
	 6 Apr 2003: major rewrite to enable useful, efficent textual output
		and better image blowups
	27 Mar 2003: exam_mref() takes extra argument to handle image rotation;
		rotation performed from detector coords to display coords
	 5 Jan 2003: reorganized source-file header
	15 Oct 2002: (?) rename bx to bdx
	15 Apr 2002: (?) in zapscales(), exam_mref(),
		modify calls related to shlp[] since it's handled differently
	25 Mar 2002: removed MSI attachments
	 2 Mar 1995: shared-memory allocation and freeing
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 5 Oct 1994: use xgread on various MULTIREF structures
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: minor bugfixes
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version.
	
   Purpose: obtaining a displayable mapping from the contents of a MULTIREF,
	which is the merged-reflection file format within X-GEN.
   Cross references: dispcalc.c, ../xglib/mrutil.c

*************************************************************************/
#include	"xgpostint.h"
#include	"displays.h"
#define		MINMATCH	4

MROBJECT *managemrv(int opno)
{ /*	Routine to handle a static pointer to MROBJECTs */
	static MROBJECT	*mrv;
	static char	mrvmsg[] = "mrv in managemrv";

	if (!opno)
	  {
		if (NULL != mrv) memzap((AVOID **)&mrv, mrvmsg);
		return NULL;
	  }
	if (opno == 1) mrv = managemrv(0);
	if ((opno > 3) && (NULL == mrv))
	  {
		if (NULL == (mrv = (MROBJECT *)callop((size_t)opno,
		 sizeof (MROBJECT), mrvmsg))) return managemrv(0);
	  }
	return mrv;
}

int exam_mref(char *mname, int whichstat, DISP_ATTRS *dispatt)
{ /* This routine stuffs a map of values between 1 and 7 inclusive into
 `idata' containing the values of one of 3 relevant multiref statistics:
 I, I/sigma, or (I-<I>)/sigma.  The statistic desired is specified in the
 variable `whichstat', which takes on the values 8, 9, and 10 for I, I/sigma,
 and (I-<I>)/sigma respectively.  The first two give rise to logarithmic
 scales, whereas the scale for (I-<I>)/sigma is preset in the program.
 Pixels that contain more than one spot are output with their most extreme
 value.  It returns 0 if successful, negative integers for failures.  */
	char		*cp, *cpy, *cpp, *idata;
	int		dewi, dex, dey, dihi, disparea, diwi, dx, dy;
	int		htdo, htup, irange, j, nscales, newval, nmruniq;
	int		nmrvs, nobs, nrun, nsc, nsca, numranges, ob;
	int		offset, roco, runno, shind, wido, wiup, x, y;
	size_t		totshifts;
	short		losh;
	xgulong		ulpre, ulsh, uplsh;
	float		*fp1, *limits;
	double		dios, fdel, fmx, fmn, res, sig, wi, yden, ski, wg, ynu;
	double		avec[MAXVECPARS], thisscale[MAXOB], *fp2;
	CRYSTAL		*xhp;
	DATARUN		*rhp;
	ABBROB		*aobp;
	SHIFT		*tsh;
	VECPAR		*vshp, *tvp;
	RUNWORK		*rup;
	MROBJECT	*mrv, *mrvp, *mrvq;
	static u_short	gotone;
	static int	bdx, nvpars, vecflags, icoo[4], nv[68];
	static FILE	*fpmi;
	static RUNWORK	*rub;
	static MERGEDREF	mr;
	static SHIFT	*shlp;
	static float	ioslim[] = { -999., -12., -9., -8., -7., -6., -5.,
	 -4.0, -3.5, -3.0, -2.5, -2.0, -1.6, -1.2, -0.8, -0.4, 0.,
	  0.4,  0.8,  1.2,  1.6,  2.0,  2.5,  3.0,  3.5,  4.0,
	  5.0,  6.0,  7.0,  8.0, 9.0, 12.0 } ;
	static size_t	sizdatarun = 256;

	if (NULL == (fpmi = fopen(mname, "r"))) return -1;
	numranges = dispatt->di_nranges;	roco = dispatt->di_roco;
	if (dispatt->di_wintype)
	  {
		diwi = dispatt->di_magw;	dihi = dispatt->di_magh;
	  }
	else {
		diwi = dispatt->di_wid;		dihi = dispatt->di_ht;
	     }
	wiup = dispatt->di_wiup;	wido = dispatt->di_widown;
	htup = dispatt->di_htup;	htdo = dispatt->di_htdown;
	dewi = dispatt->di_imwid;
		/* read the multiref headers */
	if (EOK != (x = headmem(&dispatt->di_xhdp, &dispatt->di_rhdp, 1)))
		return -1;
	xhp = dispatt->di_xhdp;	rhp = dispatt->di_rhdp;
	if (sizeof (CRYSTAL) > (size_t)xgread((void *)xhp, XG_CRYSTAL,fpmi))
		return -5;
	if (NULL == (rub = (RUNWORK *)callop((unsigned)xhp->xt_nruns,
	 sizeof (RUNWORK) + sizeof (VECPAR), "rub in exam_mref"))) return -6;
	rup = rub + xhp->xt_nruns; /* pointer beyond end of RUNWORK list */
	vshp = (VECPAR *)rup;	/* start vshp just beyond there */
	totshifts = xhp->xt_nshifts + 10;
	for (j = 0; j < MAXOB; j++) thisscale[j] = 0.;
	for (nrun = 0, rup = rub; nrun < xhp->xt_nruns; nrun++, rup++)
	   {
		if (sizdatarun > (size_t)xgread((void *)rhp, XG_DATARUN, fpmi))
		  {
			memzap((AVOID **)&rub, "rub in exam_mref");
			return -7;
		  }
		keep_dimen(rhp);	workrun(rhp, rup, xhp, nrun * 200);
	   }
	/* change as of April 2003: try to read scale factors regardless */
	if (NULL == (shlp = get_shifts(NULL, vshp, &nsca)))
	  {
		fprintf(stderr, "Error setting up scale-factor elements\n");
		memzap((AVOID **)&rub, "rub in exam_mref");	return -8;
	  }
	xhp->xt_nshifts = nscales = nsca;
	if (nscales) /* if we have scale factors, then nscales > 0 */
	  { /* check to make sure we actually have vector parameters */
		gotone = 0;
		for (tvp = vshp, runno = 0;
		 runno < xhp->xt_nruns; runno++,tvp++)
		   {
			if (fabs(tvp->v_coef[1]) > FUZZ)
			  {
				gotone = 1;	break;
			  }
		   }
		if (!gotone)
		  {
			vecflags = nvpars = 0;
		  }
		else {
			if (vshp->v_ptype == 'd')
			  {
				vecflags = SC_DIFFRACTION; nvpars = 10;
			  }
			else {
				nvpars = 9; vecflags = SC_EXITBEAM;
			     }
/* need spline information in this case to do the xy conversions */
			if (NULL == (dispatt->di_spl =
			 cleanspline(dispatt->di_spl, &x)))
			  {
				fprintf(stderr,
		 "Cannot find spline information: error code %d\n", x);
				memzap((AVOID **)&rub, "rub in exam_mref");
				return -12;
			  }
		     } /* end of else for "if (!gotone)" clause */
		nsc = (int)nscales;
		get_omcen(fpmi, shlp, nsc, 0); /* <omega> of scale grp */
	  } /* end if if (nscales) clause */
	else nsc = nscales = 0; /* we don't have scale factors yet */
	limits = dispatt->di_limi;
	if (whichstat == 10)
	  {
		for (newval = 1; newval < numranges - 1; newval++)
			limits[newval] = ioslim[newval];
	  }
	/* now go through the data */
	bdx = 32;
	if (xhp->xt_ninteg > 512L)
	  {
		for (y = 11; y <= 17; y += 2)
			if (xhp->xt_ninteg < (1 << y)) break;
		bdx >>= (y - 9) / 2;	if (bdx < 2) bdx = 2;
	  }
	printf("Multiref contains %6d observations of %6d reflections\n",
		(int)(xhp->xt_ninteg), (int)(xhp->xt_nmerged));
	printf("%d x %d boxes will be averaged\n", bdx, bdx);
	fmn = 1.e30;	fmx = -fmn;
	disparea = diwi * dihi;
	if (NULL == (mrv = managemrv(disparea))) return -9;
	nmruniq = nmrvs = 0;
	if (whichstat < 10)	/* establish limits of function */
	  {
		while (0 < xgread((void *)&mr, XG_MERGEDREF, fpmi))
		     {
			for (ob = mr.mr_nob,aobp = &(mr.mr_ob[0]); ob--;aobp++)
			   {
				ski = aobp->abo_ii;
				if (whichstat == 9) ski /= aobp->abo_sigma;
				if (ski < fmn) fmn = ski;
				if (ski > fmx) fmx = ski;
			   }
		     }
		limits[numranges] = fmn;	limits[numranges+1] = fmx;
		printf("observed range: from %8.2f to %8.2f\n",
			limits[numranges], limits[numranges+1]);
		fmx = mycbrt((double)fmx);
		fmn = mycbrt((double)fmn);
		 /* determine upper limit of each range of data */
		fdel = (fmx - fmn) / (double)(numranges - 1);
		if (fdel < DFUZZ)
		  {
			memzap((AVOID **)&rub, "rub in exam_mref");
			return -5;
		  }
		 /* we stick the min observed value in limits[0] */
		limits[0] = fmn;
		for (irange = 1; irange < numranges; irange++)
			limits[irange] = fmn + (double)irange * fdel;
		(void)fseek(fpmi, 0L, 0);	/* rewind and skip headers */
		if (sizeof (CRYSTAL) > (size_t)xgread((void *)xhp,
		 XG_CRYSTAL, fpmi))
		  {
			memzap((AVOID **)&rub, "rub in exam_mref"); return -6;
		  }
		for (nrun = 0, rup = rub; nrun < xhp->xt_nruns; nrun++, rup++)
		   {
			if (sizdatarun > (size_t)xgread((void *)rhp,
			 XG_DATARUN, fpmi))
			  {
				memzap((AVOID **)&rub, "rub in exam_mref");
				return -7;
			  }
		   }
	  }
	else	fdel = 1.;
		/* zero out the image data array */
	for (cp = idata = dispatt->di_rotm, y = disparea; y--; cp++) *cp = 0;
	while (0 < xgread((void *)&mr, XG_MERGEDREF, fpmi)) /* read refls */
	     {
		nobs = mr.mr_nob; ynu = yden = 0; res = mr.mr_stl;
		for (ob = 0, aobp= &(mr.mr_ob[0]); ob < nobs; ob++, aobp++)
		 if (nsc)
		   {
			ulpre = LSH(aobp->abo_runno, 24) |
				LSH(aobp->abo_chamno, 16);
			losh = aobp->abo_shift;	ulsh = ulpre | losh;
			thisscale[ob] = 0.;
			shind = find_shift(shlp, ulsh, nsc);
			if (shind < 0)
			  {
				if (losh < 0)
				  {
					uplsh = ulpre | (-losh);
					shind = find_shift(shlp, uplsh, nsc);
				  }
			  }
			if (shind >= 0)
			 {
				tsh = shlp + shind;
				ski = terpscale(aobp->abo_omob, tsh, res);
				if (gotone)
				  {
					runno = aobp->abo_shift / 200;
					refvec(aobp, rub + runno,
					 vecflags, nvpars, avec);
					fp1 = &((vshp + runno)->v_coef[0]);
					for (j = nvpars, fp2 = &(avec[0]),
					 wg = 0.; j--; fp1++, fp2++)
						wg += *fp1 * *fp2;
					ski *= wg;
				  }
				thisscale[ob] = ski;
				/* only let the observation contribute to
				 <I> if it has not been deleted */
				if (losh >= 0)
				  {
					wi = aobp->abo_sigma;
					if (wi < 0.01) wi = 0.01;
					wi = 1. / (wi * wi);
					ynu += wi * ski * aobp->abo_ii;
					yden += wi * ski * ski;
				  }
			  } /* end of "if (shinc >= 0)" clause */
		   } /* end of "if (nsc) clause of for (ob = 0 ... " loop */
		if (yden < DFUZZ) continue;
		ynu /= yden;
		for (ob = 0, aobp = &(mr.mr_ob[0]); ob < nobs; ob++, aobp++)
		   {
			if ((ski = thisscale[ob]) > DFUZZ)
			  {
				sig = aobp->abo_sigma;
				if (sig < 0.01) sig = 0.01;
				dios = (aobp->abo_ii - ynu * ski) / sig;
				if (whichstat == 10)
				  {
					if (dios < fmn) fmn = dios;
					if (dios > fmx) fmx = dios;
					for (newval = 1; newval <
						 numranges-1; newval++)
					  if (dios < limits[newval]) break;
				  }
				else newval = 1;
			  }
			else newval = 1;
			ski = aobp->abo_ii;
			if (whichstat == 9) ski /= aobp->abo_sigma;
			ski = mycbrt(ski);
			if (whichstat != 10) newval = 1 + (ski - fmn)/ fdel;
			if (newval < 1)
			  {
				if (ski - fmn < -0.001) printf(
 "[%4d,%4d,%4d,%2d]: statistic %2d is %9.3f, below lower limit of %9.3f\n",
				 (int)(mr.mr_hkl[0]), (int)(mr.mr_hkl[1]),
				 (int)(mr.mr_hkl[2]), (int)(mr.mr_hkl[3]),
				 whichstat, ski * ski * ski,
				 fmn * fmn * fmn);
				newval = 1;
			  }
			else if (newval >= numranges)
			  {
				if (fmx - ski > 0.001) printf(
 "[%4d,%4d,%4d,%2d]: statistic %2d is %9.3f, above upper limit of %9.3f\n",
				 (int)(mr.mr_hkl[0]), (int)(mr.mr_hkl[1]),
				 (int)(mr.mr_hkl[2]), (int)(mr.mr_hkl[3]),
				 whichstat, ski * ski * ski,
				 fmx * fmx * fmx);
				newval = numranges-1;
			  }
			/* now, generate the actual x and y values */
			dex = abptox(aobp);	dey = abptoy(aobp);
			/* round down to nearest 2, 4, 8, ... */
			dex -= (dex % bdx);	dey -= (dey % bdx);
			/* rotate into the rotated-detector coordinates */
			icoo[0] = dex;	icoo[1] = dey;
			disrot(roco, 1, dewi, icoo);
			/* rotated-detector coords to display pixels */
			x = icoo[2] * wiup;	x /= wido;
			y = icoo[3] * htup;	y /= htdo;
			offset = y * diwi + x; cp = idata + offset;
			/* update the display within a bdx * bdx box. */
			if (!(*cp) ||(abs((int)newval - 4) > abs((int)*cp -4)))
			  {	/* remember the contents of this object */
				mrvp = mrv + offset;
				munpack(mr.mr_hkl, mrvp->mro_hkl, xhp->xt_intl);
				mrvp->mro_ii = aobp->abo_ii;
				mrvp->mro_sig = aobp->abo_sigma;
				mrvp->mro_xob = dex;	mrvp->mro_yob = dey;
				mrvp->mro_omob = 0.01 * (double)aobp->abo_omob;
				mrvp->mro_stl = mr.mr_stl;
				/* flag the deleted obs with stl < 0 */
				if (aobp->abo_shift < 0) mrvp->mro_stl *= -1;
				if (fabs(ski = thisscale[ob]) > DFUZZ)
				  {
					mrvp->mro_iscal = aobp->abo_ii / ski;
					mrvp->mro_ibar = ynu;
				  }
				else {
					mrvp->mro_iscal = -HUGE;
					mrvp->mro_ibar = 0;
				     }
				nmruniq++;
				for (cpy = cp, dy = 0; dy < bdx;
				  dy++, cpy += diwi)
				   {
					mrvq = mrvp + dy * diwi;
					for (cpp = cpy, dx = 0; dx < bdx;
					 dx++, mrvq++)
					   {
						*cpp++ = newval; nmrvs++;
						if (mrvp != mrvq) cpymem(
						 (AVOID *)mrvq, (AVOID *)mrvp,
						   sizeof (MROBJECT));
					   } /* close of "for (cpp" loop */
				   } /* close of "for (cpy" loop */
			  } /* end of "if (!(*cp" clause */
		   } /* end of "for (ob = 0 ..." loop */
	     } /* end of "while (0 < xgread(" loop */
	(void)fclose(fpmi);	fpmi = NULL;
	zeromem((void *)&(nv[0]), 68 * sizeof (int));
	if (whichstat == 10)
	  {
		limits[0] = fmn;	limits[numranges-1] = fmx;
	  }
	printf(
 " Statistics have been remembered for %6d (X,Y) locations, %6d unique\n",
		nmrvs, nmruniq);
	for (cp = idata, y = disparea; y--; cp++) (nv[(int)(*cp)])++;
	printf("Histogram of occupied pixels:\n");
	printf("Value  # of pixels min value max value\n");
	for (y = 0; y < numranges; y++)
	   {
		if (whichstat < 10) limits[y] =
			limits[y] * limits[y] * limits[y];
		if (nv[y])
		  {
			if (!y) printf(" %4d %12d   unoccupied\n",
				y, nv[y]);
			else printf(" %4d %12d %9.2f %9.2f\n",
				y, nv[y], limits[y-1], limits[y]);
		  }
	   }
	if (nv[y]) printf(" >=%2d %12d %9.2f\n", y, nv[y], limits[y-1]);
	if (whichstat == 10)
	  {
		limits[0] = limits[numranges] = fmn;
		limits[numranges + 1] = fmx;
	  }
	limits[numranges-1] = limits[numranges + 1];
	memzap((AVOID **)&rub, "rub in exam_mref");	return 0;
}

int exact_mrxy(DISP_ATTRS *dispatt, short *cuh, double *iv)
{ /*	This looks for a spot found at the (X,Y) position (currx, curry)
	in the current MULTIREF file. (currx, curry) are in detector coords.
	If it finds it, it remembers its hkl in cuh[3] and
	various floating-point numbers in iv[].
	We may be doing this while MULTIREF is open elsewhere,
	so we have to open a separate stream to do this task.  */
	int		disx, disy, imwi, j, offset, roco, icoo[4];
	MROBJECT	*mrv, *mrvp;

	if (NULL == (mrv = managemrv(3))) return -1;
	roco = dispatt->di_roco;	imwi = dispatt->di_imwid;
	icoo[0] = dispatt->di_currx;	icoo[1] = dispatt->di_curry;
	disrot(roco, 1, imwi, icoo);
	disx = icoo[2] * dispatt->di_wiup;	disx /= dispatt->di_widown;
	disy = icoo[3] * dispatt->di_htup;	disy /= dispatt->di_htdown;
	if ((0 > disx) || (disx >= dispatt->di_wid) ||
	 (0 > disy) || (disy >= dispatt->di_ht))
	  {
		printf("(X,Y) = (%4d,%4d) -> displayed (%4d,%4d) illegal\n",
			icoo[0], icoo[1], disx, disy);
		return 0;
	  }
	/* offset needs to be calculated based on the main window's dimensions,
	 even if we're looking at the blowup window's dimensions */
	offset = disy * dispatt->di_wid + disx;
	mrvp = mrv + offset;
	if ((mrvp->mro_hkl[0] == 0) && (mrvp->mro_hkl[1] == 0) &&
	 (mrvp->mro_hkl[2] == 0))
	  {
		printf("(X,Y) = (%4d,%4d) unoccupied\n", icoo[0], icoo[1]);
		return 0;
	  }
	for (j = 0; j < 3; j++) cuh[j] = mrvp->mro_hkl[j];
	iv[0] = mrvp->mro_ii;	iv[1] = mrvp->mro_sig;
	iv[2] = mrvp->mro_xob;	iv[3] = mrvp->mro_yob;
	iv[4] = mrvp->mro_omob;	iv[5] = mrvp->mro_iscal;
	iv[6] = mrvp->mro_ibar;	iv[7] = mrvp->mro_stl;
	return 1;
}

/* end of dispmref.c */
