/* disppred.c -- calculations for predicted spots overlaid on XGEN images

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	27 Nov 2006: eliminate dispatt->di_q21 because xp->xq21 is available
	31 Dec 2005: handle qsort with < 2 elements correctly
	 5 Oct 2005: housekeeping
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	20 Oct 2004: more informative diagnostics in gpars()
	 2 Sep 2004: routines in ../xglib/xgpredict.c instead of locals
	26 Mar 2004: output fewer diagnostics if di_flags & DISPVERBOSE_ANY==0
	24 Jan 2004: some static global variables converted to locals;
		gspline returns into element of dispatt
	21 Jan 2004: license message inserted
	 7 Jan 2004: globals replaced by DISP_ATTRS elements from args;
		replace shcalloc() & shfree() with callop() & memzap()
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	25 Aug 2003: eliminate references to variables obtained from pdispmain.c
		using constructions involving DISP_ATTRS structures
	18 Aug 2003: reorganized to reduce dependence on external variables
	12 Aug 2003: handle reflection widths differently in dfindref():
		reflections with large gamma values appear in more images.
	 6 Apr 2003: only warn the user once about the absence of predictions
	 5 Jan 2003: reorganized source-file header
	25 Mar 2002: removed MSI attachments
	 2 Mar 1995: major bugfix, more flexibility
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
		invoke proper "rec[][]" matrix in obtaining prediction
	 1 Dec 1994: better diagnostics when display-predictions fail
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	 7 Dec 1993: initial X-GEN version
	
   Purpose: calculatings enabling displays of spot predictions
	(assuming UPARAMS information is present).
   Cross references: dispcalc.c

*************************************************************************/
#include	"displays.h"
#include	"xgenpred.h"
#define		MAXPERFM	(2000)	/* max # predictable in a frame */

extern PREDREF *pred_range(DISP_ATTRS *dispatt, int  *npredp);

PREDREF *pred_range(DISP_ATTRS *dispatt, int *npredp)
{ /*	This is a wrapper for "prloop" in the context of displays.
	*npredp contains the count of predicted reflections,
	and the routine returns a pointer to the list thereof. */
	register int	j;
	int		lhi, lwi, nfound, ntopred, sizr;
	double		fwid, mudelta;
	REFPARS		*xp;
	PREDWORK	*pwp;
	static int	fmnum;

	/* pick a range for the display. For narrow frames (< 0.3 deg)
	 we make it 3.5 frames wide; it tapers to 1 as framesize goes up.
	 The total width is 2*mudelta+1, so that's 8 tapering down to 3.
	 This has changed as of Aug 2003, because we're choosing to display
	 reflections with larger gamma values, so we need wider ranges. */
	xp = dispatt->di_xp;
	fwid = fabs(xp->xfrmsize) / RADC;
	fmnum = dispatt->di_fmno;
	mudelta = (fwid < 0.35) ? 3.5 : 1. + 0.225 / (fwid * fwid);
	/* save memory for the predicted spots. */
	lwi = dispatt->di_imwid;	lhi = dispatt->di_imht;
	sizr = ((double)lwi / 512.) * ((double)lhi / 512.);
	if (sizr < 1) sizr = 1;
	ntopred = MAXPERFM * sizr * (2. * mudelta + 1.);
	if (ntopred < 100) ntopred = 100;
	/* make sure the prediction structure exists, is properly populated,
	 and has enough memory in it to do the needed predictions */
	*npredp = 0;
	if (NULL == (pwp = manageprw(ntopred))) return NULL;
	pwp->po_xp = xp;
	if (((NULL == pwp->po_spl) || (fabs(pwp->po_maxxcm) < DFUZZ)) &&
	 (EOK != (j = prpar(pwp))))
	  {
		fprintf(stderr, "Error %d in setting up prediction\n", j);
		return pwp->po_prp0;
	  }
	/* rotate the reciprocal-space matrix to the center of the scanning
	range.  This is done mostly to assure proper choice of slowest-varying
	axis in the index-ordering operation which follows.  */
	pwp->po_mumin = (((double)fmnum) - mudelta) * xp->xfrmsize;
	pwp->po_mumax = (((double)fmnum) + mudelta + 1.) * xp->xfrmsize;
	if (pwp->po_mumax < pwp->po_mumin) /* neg. stepsize treated specially*/
	  {
		pwp->po_rangecen = pwp->po_mumax;
		pwp->po_mumax = pwp->po_mumin;
		pwp->po_mumin = pwp->po_rangecen;
	  }
	if (0 >= (j = prloop(pwp, 0)))
	  {
		fprintf(stdout, "Error %d in prloop\n", j);
		return pwp->po_prp0;
	  }
	nfound = pwp->po_prp - pwp->po_prp0;	/* count found refls */
	if (nfound > 1) qsort((AVOID *)pwp->po_prp0, (unsigned)nfound,
		sizeof (PREDREF), (QPE)dphicmp);
	if (dispatt->di_disflags & DISPHKL_CALC)
	  {
		if ((!(dispatt->di_wintype)) &&
		 (dispatt->di_disflags & DISPVERBOSEANY))
		  {
			if (nfound < 1) fprintf(stderr,
		"No reflections found in an otherwise valid range; ");
			else fprintf(stderr,
		 " %6d reflections predicted from omega = %7.2f to %7.2f\n",
			 nfound, (pwp->po_prp0)->pphi,
			 (pwp->po_prp0 + (nfound - 1))->pphi);
		  }
		dispatt->di_disflags |= DISPHKL_CALC;
	  }
	*npredp = nfound;	return pwp->po_prp0;
}

int gpars(DISP_ATTRS *dispatt)
{ /*	This reads the current UPARAMS file and saves the appropriate
	parameters therefrom.  It returns 1 if successful, 0 if not */
	int		i;
	static int	complaints;
	PREDWORK	*pwp;

	if (NULL == (pwp = manageprw(3))) return 0;
	pwp->po_xp = dispatt->di_xp; /* establish xp if we have it now */
	if (EOK != (i = prpar(pwp))) /* set up for doing predictions */
	  {
		if (!complaints)
		  {
			if (EOPENERR == i) fprintf(stderr,
			 "Note: parameter file has not been written yet\n");
			else if (EREADERR == i) fprintf(stderr,
			 "Warning: parameter file unreadable\n");
			else fprintf(stderr,
		 "Error %d in obtaining parameters\n", i);
		  }
		complaints++;	return 0;
	  }
	dispatt->di_xp = pwp->po_xp; /* reconnect dispatt->di_xp */
	dispatt->di_disflags |= DISPHKL_IN; /* flag our pred. knowledge */
	if (!((dispatt->di_disflags) & DISPBORDER_CALC))
		dispatt->di_disflags |= DISPBORDER_IN; /* we know the border */
	dispatt->di_spl = pwp->po_spl; /* we also know the spline */
	dispatt->di_disflags |= DISPCALIB_IN;	return 1;
}

/* end of disppred.c */
