/* dispuref.c -- calculations for displaying UREFLS statistics

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	22 Jun 2006: exam_urefl,scatt_uh,scatt_uref:
		idata,limits now part of DISP_ATTRS structure
	15 Oct 2005: eliminate most of the static globals
	 5 Oct 2005: housekeeping
	 8 Oct 2004: bugfix, scatt_uref() in memory alloc for *histo
	 3 Oct 2004: histo[], cenofrange[] alloc'd & freed; they're
		nranges long rather than 8 in exam_urefl & scatt_uref();
		error checking for empty UREFLS file performed
	28 Aug 2004: routines used only here are declared as static; cosmetics
	24 Jan 2004: a few static globals converted to locals
	21 Jan 2004: license message inserted
	 7 Jan 2004: wavelength derived from safelambda()
	19 Sep 2003: default cases defined in scatt_uref()
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Sep 2003: eliminate explicit calls to xgaccess() because
		they're implicit in the xgopen() calls
	10 Sep 2003: handle wavelength more agressively
		and, in exact_xy in blowup windows, allow limited
		searches around the picked point.
	 4 Sep 2003: enable uncompression of compressed X-GEN output files
	30 Aug 2003: moved from ../motif to ../displays;
		contents of dispuh.c and dispuscat.c included here
	17 Aug 2003: gauge size of display using locally-calculated variable
		called `disparea' rather than `imarea'
	16 Aug 2003: remove reliance on global variables afmin and afmax
		at the cost of two extra elements in the limits[] array;
		check_uh(), scatt_uh(), exam_urefl(), scatt_uref()
		  return the number of refls in UREFLS file to their callers
		in scatt_uref(), return variable-limits in limits[] array
	12 Jun 2003: fixed bug in exact_inten so it returns a value
	 4 Apr 2003: major rewrite to provide for better textual output,
		management of mouse-clicks, and better image blowups
	29 Mar 2003: add I and I/sigma as dependent variables in scatt_uref()
	27 Mar 2003: exam_urefl() has extra argument to handle image rotation;
		rotation performed from detector coords to display coords
	 5 Jan 2003: reorganized source-file header
	15 Nov 2002: trivial reformatting
	10 Nov 2002: mycbrt is declared as extern double, not double
	25 Mar 2002: removed MSI attachments
	 2 Mar 1995: diagnostics to stdout
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 5 Oct 1994: use xgread on OBSREF structures
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	 7 Dec 1993: initial X-GEN version

   Purpose: calculations enabling displays of UREFLS statistics,
	i.e. the file output by the integration capability in XGEN
   Cross references: dispx.c, ../integrate/int*.c, ../displays/dispcalc.c
 For the scatter-plots,
	possible ordinates are sin(theta)/lambda, X, Y, and omega;
	possible abscissas are err(X), err(Y), err(omega), I, and I/sigma.
	The color coding is by I/sigma.
*************************************************************************/

#define		NEED_ERRNO	1
#include	"displays.h"
#define		MINMATCH	2

URWORK *manageuwp(int opno, int nobj)
{ /*	Manage a static variable containing a URWORK */
	static URWORK	*uwp;
	static char	uwpmsg[] = "uwp in manageuwp";
	static char	vamsg[] = "uwp->urw_val in manageuwp";
	static char	hvmsg[] = "uwp->urw_hv in manageuwp";

	if (!opno) /* eliminate all local memory */
	  {
		if (NULL != uwp)
		  {
			(void)manageuwp(014, 0);
			memzap((AVOID **)&uwp, uwpmsg);
		  }
		return NULL;
	  }
	if (!(opno & 01)) /* eliminate specific allocations */
	  {
		if (NULL != uwp)
		  {
			if (opno & 4) memzap((AVOID **)&(uwp->urw_val), vamsg);
			if (opno & 010) memzap((AVOID **)&(uwp->urw_hv),hvmsg);
		  }
		return uwp;
	  }
	if (1 == opno) uwp = manageuwp(0, 0); /* force reload */
	if (NULL == uwp) uwp = (URWORK *)callop(1, sizeof (URWORK), uwpmsg);
	if ((014 & opno) && (NULL != uwp)) /* memory for urw_val and urw_hv*/
	  {
		if (020 & opno) /* force release of old memory */
		  {
			if (opno & 4) memzap((AVOID **)&(uwp->urw_val), vamsg);
			if (opno & 010) memzap((AVOID **)&(uwp->urw_hv),hvmsg);
		  }
		if ((opno & 4) && (NULL == uwp->urw_val))
		  {
			if (NULL == (uwp->urw_val = (UROBJECT *)callop(
			 (size_t)nobj, sizeof (UROBJECT), vamsg)))
				return manageuwp(0, 0);
		  }
		if ((opno & 010) && (NULL == uwp->urw_hv))
		  {
			if (NULL == (uwp->urw_hv = (UROBJECT *)callop(
			 (size_t)nobj, sizeof (UROBJECT), hvmsg)))
				return manageuwp(0, 0);
		  }
	  }
	return uwp;
}

static double safelambda(DISP_ATTRS *dispatt)
{ /* This cautiously returns a wavelength value based on a DISP_ATTRS struct */
	MAIL	*lhd;
	DATARUN	*lru;
	REFPARS	*xp;
	double	lla;

	xp = dispatt->di_xp;
	if (FUZZ < xp->xlambda) return xp->xlambda;
	lhd = dispatt->di_head + dispatt->di_fmno % dispatt->di_nmbox;
	lru = lhd->runheader;
	lla = 0.;	if (NULL != lru) lla = lru->run_lambda;
	if ((lla < DFUZZ) || (lla > 1000.)) lla = 1.5418;
	return lla;
}

int exam_urefl(char *uname, int whichstat, DISP_ATTRS *dispatt)
{ /*	This routine does statistics on a urefls file. It returns various
	negative integers upon failure, and the number of located reflections
	on a successful return. */
	char		*cp, *idata;
	int		allh, detx, dety, dih, disparea, diw, hoffset;
	int		htu, htdo, hwid, idh, idk, idl, imw, irange, kwid;
	int		lwid, numranges, offset, reffound, roco, thisx, thisy;
	int		useit, wido, wiup, *histo, *hlim, icoo[4];
	double		fdel, fmn, fmx, llam, lxsc, lysc, middle;
	double		oldval, thisf, thiss, *cenofrange;
	float		*limits, *lip;
	FILE		*fplu;
	OBSREF		rf;
	UROBJECT	*urvp, *urvq;
	URWORK		*uwp;

	numranges = dispatt->di_nranges;	roco = dispatt->di_roco;
	if (dispatt->di_wintype)
	  {
		diw = dispatt->di_magw;	dih = dispatt->di_magh;
	  }
	else  {
		diw = dispatt->di_wid;	dih = dispatt->di_ht;
	      }
	htu = dispatt->di_htup;		htdo = dispatt->di_htdown;
	wiup = dispatt->di_wiup;	wido = dispatt->di_widown;
	lxsc = dispatt->di_xsca;	lysc = dispatt->di_ysca;
	llam = safelambda(dispatt);	imw = dispatt->di_imwid;
	if (whichstat < 1 || whichstat > 5) return -6;
	if (NULL == (fplu = xgopen(uname, "r"))) return -1; /* open UREFLS */
	/* find min and max of the relevant function within the UREFLS file*/
	fmn = 1.e30;	fmx = -fmn;	disparea = diw * dih;
	idata = dispatt->di_rotm; zeromem((AVOID *)idata, (size_t)disparea);
	/* get memory for the structures where we keep the pixel-specific
	 UREFLS statistics (if needed) and zero out those structures */
	if (NULL == (uwp = manageuwp(07, disparea))) return -8;
	zeromem((AVOID *)(uwp->urw_val), disparea * sizeof (UROBJECT));
	/* now set the uro_sig element of each of these pixel-specific
	 structures to a nonsense value as a flag */
	for (urvp = uwp->urw_val, thisx = 0; thisx < disparea; thisx++, urvp++)
		urvp->uro_sig = -1000.;
	hlim = uwp->urw_hlim;
	hlim[2] = hlim[1] = hlim[0] = 10000; /* silly values for extrema */
	hlim[5] = hlim[4] = hlim[3] = -10000;
	fprintf(stderr, "Examining %s to prepare statistics:\n", uname);
	for (reffound = 0; 0 < xgread((void *)&rf, XG_OBSREF,fplu); reffound++)
	   {
		switch (whichstat)
		      {
			case	1:	default : /* Intensity */
				thisf = rf.obs_ii;	break;
			case	2:	/* I/sigma */
				if ((thiss = rf.obs_sigma) < FUZZ)
					thiss = 0.1;
				thisf = rf.obs_ii / sqrt(thiss);	break;
			case	3:	/* err(omega) */
				thisf = rf.obs_omob - rf.obs_ompred;	break;
			case	4:	/* err(X) */
				thisf = ((double)(rf.obs_xob - rf.obs_xpred))
					/ lxsc ;	break;
			case	5:	/* err(Y) */
				thisf = ((double)(rf.obs_yob - rf.obs_ypred))
					/ lysc;	break;
		      }
		if (thisf < fmn) fmn = thisf;
		if (thisf > fmx) fmx = thisf;
		if (rf.obs_hkl[0] < hlim[0]) hlim[0] = rf.obs_hkl[0];
		if (rf.obs_hkl[1] < hlim[1]) hlim[1] = rf.obs_hkl[1];
		if (rf.obs_hkl[2] < hlim[2]) hlim[2] = rf.obs_hkl[2];
		if (rf.obs_hkl[0] > hlim[3]) hlim[3] = rf.obs_hkl[0];
		if (rf.obs_hkl[1] > hlim[4]) hlim[4] = rf.obs_hkl[1];
		if (rf.obs_hkl[2] > hlim[5]) hlim[5] = rf.obs_hkl[2];
		if ((reffound) && !(reffound % 8000)) fprintf(stderr,
		 "Examining reflection %8d\r", reffound);
	   }
	if (reffound >= 8000) fprintf(stderr, "\n");
	hwid = 1 + hlim[3] - hlim[0];
	kwid = 1 + hlim[4] - hlim[1];
	lwid = 1 + hlim[5] - hlim[2];
	if ((hwid < 0) || (kwid < 0) || (lwid < 0))
	  {
		fprintf(stderr, "UREFLS file contains no interpretable data\n");
		(void)fclose(fplu);	return 0;
	  }
	printf("HKL limits: [%3d,%3d],[%3d,%3d],[%3d,%3d]\n",
	 hlim[0], hlim[3], hlim[1], hlim[4], hlim[2], hlim[5]);
	printf("intensity range: %8.1f to %8.1f\n", fmn, fmx);
	allh = hwid * kwid * lwid;
	/* reserve memory (if needed) for the urw_hv structures */
	if (NULL == (uwp = manageuwp(013, allh))) return -7;
	zeromem((AVOID *)(uwp->urw_hv), allh * sizeof (UROBJECT));
	if (0 >= reffound) return -2;	/* if no refls, return error code */
	if (-1 == fseek(fplu, 0L, 0)) return -3;	/* rewind file */
	if (fmx - fmn < DFUZZ)
	  {
		printf("insufficient contrast for useful displays\n");
		return -4;	/* not enough contrast */
	  }
	limits = dispatt->di_limi;
	limits[numranges] = fmn;	limits[numranges + 1] = fmx;
	/* launder the range values.  All output ranges will be rescaled
	 by the cube-root of the raw value. This puts more attention
	 on the values close to zero while preserving a sign dependence.  */
	fmn = mycbrt(fmn);	fmx = mycbrt(fmx);
	 /* determine upper limit of each range of data */
	fdel = (fmx - fmn) / (double)(numranges - 1);
	if (fdel < DFUZZ) return -5;
	 /* limits[0] is just a place to stick the minimum observed value */
	limits[0] = fmn;
	if (NULL == (cenofrange = (double *)callop((size_t)(numranges + 1),
	 sizeof (double), "cenofrange in exam_urefl"))) return -6;
	for (irange = 1; irange < numranges; irange++)
	   {
		limits[irange] = fmn + (double)irange * fdel;
		cenofrange[irange] = limits[irange] - fdel / 2.;
	   } /* "middle" holds an average value of our function */
	if (numranges % 2) middle = limits[numranges / 2];
	else middle = 0.5 * (limits[numranges / 2 - 1] + limits[numranges / 2]);
	fprintf(stderr, "Assembling statistics on data from %s\n", uname);
	for (reffound = 0; 0 < xgread((void *)&rf, XG_OBSREF, fplu); reffound++)
	   {
		thiss = rf.obs_sigma;	if (thiss < FUZZ) thiss = 0.1;
		thiss = sqrt(thiss);
		switch (whichstat)
		      {
			case	1:	default:	/* Intensity */
				thisf = rf.obs_ii;	break;
			case	2:	/* I/sigma */
				thisf = rf.obs_ii / thiss;	break;
			case	3:	/* err(omega) */
				thisf = rf.obs_omob - rf.obs_ompred;	break;
			case	4:	/* err(X) */
				thisf = ((double)(rf.obs_xob - rf.obs_xpred))
					/ lxsc;	break;
			case	5:	/* err(Y) */
				thisf = ((double)(rf.obs_yob - rf.obs_ypred))
					/ lysc;	break;
		      }
		thisf = mycbrt(thisf);
		irange = 1 + (thisf - fmn) / fdel;
		if (irange == numranges) irange = numranges - 1;
		if (irange < 1 || irange > numranges) continue;
		 /* get the X-GEN internal coordinates of the spot */
		detx = ((double)rf.obs_xob) / lxsc + 0.49999;
		dety = ((double)rf.obs_yob) / lysc + 0.49999;
		 /* determine the rotated coordinates */
		icoo[0] = detx;	icoo[1] = dety;
		disrot(roco, 1, imw, icoo);
		thisx = icoo[2] * wiup;	thisx /= wido;
		if (0 > thisx) thisx = 0;
		else if (thisx >= diw) thisx = diw - 1;
		thisy = icoo[3] * htu;	thisy /= htdo;
		if (0 > thisy) thisy = 0;
		else if (thisy >= dih) thisy = dih - 1;
		offset = diw * thisy + thisx;
		if ((offset < 0) || (offset > disparea))
		  {
			fprintf(stderr,
 "Reflection [%4d,%4d,%4d] can't be visualized:(%4d,%4d) outside view area\n",
			 rf.obs_hkl[0], rf.obs_hkl[1], rf.obs_hkl[2],
			 thisx, thisy); continue;
		  }
		cp = idata + offset;
		/* if a point is already occupied, save the more extreme
		value, i.e. the value furthest from `middle' */
		useit = 1;	/* bugfix: this used to be set to zero */
		if (*cp)
		  {
			oldval = cenofrange[(int)(*cp)];
			if (fabs(thisf - middle) <= fabs(oldval - middle))
				useit = 0;
		  }
		if (useit)
		  {
			*cp = irange; urvp = uwp->urw_val + offset;
			idh = rf.obs_hkl[0] - hlim[0];
			idk = rf.obs_hkl[1] - hlim[1];
			idl = rf.obs_hkl[2] - hlim[2];
			hoffset = idh * kwid * lwid + idk * lwid + idl;
			if ((hoffset < 0) || (hoffset >= allh))
			  {
				fprintf(stderr,
 "Can't assign [%4d,%4d,%4d] to an hkl bin: offset is out of range\n",
				 rf.obs_hkl[0], rf.obs_hkl[1], rf.obs_hkl[2]);
				continue;
			  }
			urvq = uwp->urw_hv + hoffset;
			for (useit = 0; useit < 3; useit++)
				urvp->uro_hkl[useit] = rf.obs_hkl[useit];
			urvp->uro_ii = rf.obs_ii;
			urvp->uro_sig = thiss;
			urvp->uro_xob = detx;	urvp->uro_yob = dety;
			urvp->uro_omob = rf.obs_omob;
			urvp->uro_stl = 100. * llam /
				sqrt((double)(rf.obs_ssq));
			urvp->uro_delx = ((double)(rf.obs_xob - rf.obs_xpred))/
				lxsc;
			urvp->uro_dely = ((double)(rf.obs_yob - rf.obs_ypred))/
				lysc;
			urvp->uro_delz = rf.obs_omob - rf.obs_ompred;
			cpymem((AVOID *)urvq,(AVOID *)urvp,sizeof (UROBJECT));
		  }
		if ((reffound) && !(reffound % 8000)) fprintf(stderr,
		 "Examining reflection %8d\r", reffound);
	   }
	if (reffound >= 8000) fprintf(stderr, "\n");
	/* Convert the limits back to the original metric by cubing them */
	for (irange = numranges, lip = &(limits[0]); irange--; lip++)
			*lip = *lip * *lip * *lip;
	if (NULL == (histo = (int *)callop((size_t)(numranges + 1),
	 sizeof (int), "histo in exam_urefl")))
	  {
		memzap((AVOID **)&cenofrange, "cenofrange in exam_urefl");
		return -7;
	  }
	printf("Histogram: color #  # points\n");
	for (thisy = disparea, cp = idata; thisy--; cp++) (histo[(int)(*cp)])++;
	for (irange = 0; irange < 9; irange++)
	   if (histo[irange]) printf("          %8d  %8d\n",
		irange, histo[irange]);
	for (urvq = uwp->urw_hv, thisx = useit = 0; useit < allh;
	 useit++, urvq++) if (urvq->uro_sig > FUZZ) thisx++;
	printf(" %8d hkl structures out of %8d are occupied\n", thisx, allh);
	for (urvp = uwp->urw_val, thisx = useit = 0; useit < disparea;
	 useit++, urvp++) if (urvp->uro_sig > FUZZ) thisx++;
	printf(" %8d (X,Y) structures out of %8d are occupied\n", thisx, allh);
	memzap((AVOID **)&cenofrange, "cenofrange in exam_urefl");
	memzap((AVOID **)&histo, "histo in exam_urefl");
	(void)fclose(fplu);	return reffound;
}

int exact_xy(DISP_ATTRS *dispatt, short *cuh, double iv[])
{ /*	This looks for a spot found at the (X,Y) position
	(dispatt->di_currx, dispatt->di_curry) in the current UREFLS file.
	If it finds it, it remembers its hkl in cuh[3] and
	various floating-point numbers in iv[].
	We may be doing this while UREFLS is open elsewhere, so we
	have to open a separate stream to do this task.  */
	int		closest, dih, disparea, disx, disy, diw, dtx, dty;
	int		dux, duy, j, kux, kuy, offset, radi, roco;
	int		ux, uxma, uxmi, uy, uyma, uymi, icoo[4];
	URWORK		*uwp;
	UROBJECT	*urvp;

	dtx = dispatt->di_currx;	dty = dispatt->di_curry;
	roco = dispatt->di_roco;
	diw = dispatt->di_wid;		dih = dispatt->di_ht;
	disparea = diw * dih;
	if (NULL == (uwp = manageuwp(7, disparea))) return -1;
	icoo[0] = dtx;	icoo[1] = dty;
	disrot(roco, 1, dispatt->di_imwid, icoo);
	disx = icoo[2] * dispatt->di_wiup;	disx /= dispatt->di_widown;
	disy = icoo[3] * dispatt->di_htup;	disy /= dispatt->di_htdown;
	offset = disy * diw + disx;
	urvp = uwp->urw_val + offset;
	if (urvp->uro_sig < -998.) /* no point in this pixel */
	  { /* Sep 03: if we're in the main window, look at a few neighbors */
		closest = 30;	kux = disx;	kuy = disy;
		if (!(dispatt->di_wintype))
		  {
			uymi = disy - 3;	if (uymi < 0) uymi = 0;
			uyma = disy + 4;	if (uyma > dih) uyma = dih;
			uxmi = disx - 3;	if (uxmi < 0) uxmi = 0;
			uxma = disx + 4;	if (uxma > diw) uxma = diw;
			for (uy = uymi; uy < uyma; uy++)
			 for (ux = uxmi; ux < uxma; ux++)
			    {
				offset = uy * diw + ux;
				if ((offset < 0) || (offset > disparea))
					continue;
				urvp = uwp->urw_val + offset;
				if (urvp->uro_sig > -998.)
				  {
					duy = uy - disy; dux = ux - disx;
					radi = duy * duy + dux * dux;
					if (radi < closest)
					  {
					   kuy = uy; kux = ux; closest = radi;
					  }
				  }
			     }
		  }
		if (closest > 27)
		  {
			printf("(X,Y) = (%4d,%4d) unoccupied\n", dtx, dty);
			return 0;
		  }
		offset = kuy * diw + kux;	urvp = uwp->urw_val + offset;
	  }
	for (j = 0; j < 3; j++) cuh[j] = urvp->uro_hkl[j];
	iv[0] = urvp->uro_ii;	iv[1] = urvp->uro_sig;
	iv[2] = urvp->uro_xob;	iv[3] = urvp->uro_yob;
	iv[4] = urvp->uro_omob;	iv[5] = urvp->uro_stl;
	iv[6] = urvp->uro_delx;	iv[7] = urvp->uro_dely;
	iv[8] = urvp->uro_delz;	return 1;
}

int exact_inten(short cuhkl[], double *iv)
{ /*	This looks for the hkl vector `cuhkl' in the current list of
	UROBJECTs. If it finds it, it puts some information in `iv'
	and returns a nonzero value; otherwise, it returns zero. */
	int		kwid, lwid, loff, *hlim;
	UROBJECT	*urvq;
	URWORK		*uwp;

	if (NULL == (uwp = manageuwp(3, 0))) return 0;
	hlim = uwp->urw_hlim;
	kwid = 1 + hlim[4] - hlim[1];
	lwid = 1 + hlim[5] - hlim[2];
	loff = kwid * lwid * (cuhkl[0] - hlim[0])
		+ lwid * (cuhkl[1] - hlim[1]) + cuhkl[2] - hlim[2];
	urvq = uwp->urw_hv + loff;
	if (urvq->uro_sig < -998.)
	  {
		printf(" Reflection [%3d,%3d,%3d] absent\n",
			cuhkl[0], cuhkl[1], cuhkl[2]); return 0;
	  }
	iv[0] = urvq->uro_ii;	iv[1] = urvq->uro_sig;
	iv[2] = urvq->uro_xob;	iv[3] = urvq->uro_yob;
	iv[4] = urvq->uro_omob;	iv[5] = urvq->uro_stl;
	iv[6] = urvq->uro_delx;	iv[7] = urvq->uro_dely;
	iv[8] = urvq->uro_delz;	return loff;
}

static double getinten(OBSREF *rp, int dtype)
{ /*	This returns the appropriate statistic (intensity or I/sigma) for
	a given reflection */
	double	locf, locs;

	switch (dtype % 4)	/* see what kind of data to use */
	      {
		case	0:	default: /*Profile-fitted,(X,Y,omega) answers*/
			locf = rp->obs_ii; locs = rp->obs_sigma;	break;
		case	1:	/* summation (X,Y,omega) */
			locf = rp->obs_iunfit; locs = rp->obs_sunfit;	break;
		case	2:	/* Profile fitted local diffr.vecs. */
			locf = rp->obs_ii + ((double)rp->obs_shift) /16.;
			locs = rp->obs_sigma + ((double)rp->obs_widob) /8.;
			break;
		case	3:	/* summation local diffr. vectors */
			locf = rp->obs_ii + ((double)rp->obs_bgunfit) / 16.;
			locs = rp->obs_sunfit + ((double)rp->obs_reflno) / 8.;
			break;
	      }
	locs = (locs < 0.01 ? 0.1 : sqrt(locs)); /* massage the sigma */
	if (dtype & 04) locf /= locs;	/* return the correct stat */
	return locf;
}

int check_uh(char *uname, short *minh, short *maxh)
{ /*	this determines the extreme - and + values of the most wide-ranging
	index a given set of data. It returns the number of observations read
	if successful, various negative integers if it fails. */
	int		te, reffound;
	short		hmax[3], hmin[3];
	FILE		*fplu;
	OBSREF		rf;

	if (NULL == (fplu = xgopen(uname, "r"))) return -1; /* open UREFLS */
	for (te = 3; te--; )
	   {
		hmin[te] = 32765;	hmax[te] = -hmin[te];
	   }
	for (reffound = 0; 0 < xgread((void *)&rf, XG_OBSREF,fplu); reffound++)
	   { /* establish ranges of hkl */
		for (te = 3; te--; )
		   {
			if (rf.obs_hkl[te] <hmin[te]) hmin[te] =rf.obs_hkl[te];
			if (rf.obs_hkl[te] >hmax[te]) hmax[te] =rf.obs_hkl[te];
		   }
	   }
	(void)fclose(fplu);	/* close it; we'll reopen it later */
	if (0 >= reffound) return -2;	/* if no refls, return error code */
	printf("Extrema of the three indices:\n");
	printf(" hmin hmax  kmin kmax  lmin lmax\n");
	printf(" %4d %4d  %4d %4d  %4d %4d\n",
	 hmin[0], hmax[0], hmin[1], hmax[1], hmin[2], hmax[2]);
	*minh = hmin[0]; if ((te = hmin[1]) < *minh) *minh = te;
	if ((te = hmin[2]) < *minh) *minh = te;
	*maxh = hmax[0]; if ((te = hmax[1]) > *maxh) *maxh = te;
	if ((te = hmax[2]) > *maxh) *maxh = te;
	return reffound;
}

int scatt_uh(char *uname, int uhva[], DISP_ATTRS *dispatt)
{ /*	This routine does hkl plots from a urefls file.
	It returns various negative integers upon failure, and
	it returns the number of reflections in the file if it succeeds. */
	char		*cp, *idata;
	int		datatype, dih, disparea, diw, iho, io, irange;
	int		ive, noor, numranges, offset, reffound;
	int		thisx, thisy, valconst, whichconstant, histo[33];
	float		*limitv, *lip;
	double		fdel, fmx, fmn, thisf;
	FILE		*fplu;
	OBSREF		rf;
	static int 	poff[9];
	static char chkl[] = "hkl";

	whichconstant = uhva[0]; valconst = uhva[1]; datatype = uhva[2];
	if (dispatt->di_wintype)
	  {
		diw = dispatt->di_magw;	dih = dispatt->di_magh;
	  }
	else {
		diw = dispatt->di_wid;	dih = dispatt->di_ht;
	     }
	numranges = dispatt->di_nranges;
	for (irange = 0; irange <= numranges; irange++) histo[irange] = 0;
	if (whichconstant < 0 || whichconstant > 2) return -6;
	if (NULL == (fplu = xgopen(uname,"r"))) return -1; /* open UREFLS */
	/* find min and max of the relevant function within the UREFLS file*/
	fmn = 1.e30;	fmx = -fmn;	disparea = diw * dih;
	idata = dispatt->di_rotm; zeromem((AVOID *)idata, (size_t)disparea);
	fprintf(stderr, "Establishing color ranges for hkl data from %s\n",
		uname);
	for (reffound = 0; 0 < xgread((void *)&rf, XG_OBSREF,fplu); reffound++)
	   if (valconst == rf.obs_hkl[whichconstant])
	     { /* establish ranges of the color code value */
		thisf = getinten(&rf, datatype);
		if (thisf < fmn) fmn = thisf; if (thisf > fmx) fmx = thisf;
	     }
	if (0 >= reffound) return -2;	/* if no refls, return error code */
	if (fmn > fmx)
	  {
		fprintf(stderr, "No refls at %c = %d\n",
			chkl[whichconstant], valconst);
		return -2;	/* no refls in this range */
	  }
	printf("Range of %s for %c = %d: %8.3f to %8.3f\n",
		(datatype & 04 ? "I/sigma" : "Intensity"),
		chkl[whichconstant], valconst, fmn, fmx);
	if (-1 == fseek(fplu, 0L, 0)) return -3;	/* rewind file */
	if (fmx - fmn < DFUZZ) return -4;	/* not enough contrast */
	limitv = dispatt->di_limi;
	limitv[numranges] = fmn;	limitv[numranges + 1] = fmx;
	/* launder the I or I/sigma range values using
	 the cube root of f as the scaling function */
	fmx = mycbrt(fmx);	fmn = mycbrt(fmn);
	 /* determine upper limit of each range of data */
	fdel = (fmx - fmn) / (double)(numranges - 1);
	if (fdel < DFUZZ) return -5;
	 /* limitv[0] is just a place to stick the minimum observed value */
	limitv[0] = fmn;
	for (irange = 1; irange < numranges; irange++)
		limitv[irange] = fmn + (double)irange * fdel;
	/* go back through the data and bin things.  We only do this calc.
	if we're on the right section, e.g. if we're displaying h==2,
	we skip all the spots for which h != 2 */
	zeromem((AVOID *)idata, (size_t)disparea);
	/* define the offsets */
	poff[0] = 0;		poff[1] = 1;		poff[2] = 2;
	poff[3] = diw;		poff[4] = diw + 1;	poff[5] = diw + 2;
	poff[6] = diw * 2;	poff[7] = diw * 2 + 1;	poff[8] = diw * 2 + 2;
	fprintf(stderr, "Setting up display of index statistics from %s\n",
		uname);
	for ( ; 0 < xgread((void *)&rf, XG_OBSREF, fplu); )
	   if (valconst == rf.obs_hkl[whichconstant])
	     {
		thisf = mycbrt(getinten(&rf, datatype));
		irange = 1 + (thisf - fmn) / fdel;
		if (irange == numranges) irange = numranges - 1;
		if (irange < 1 || irange > numranges) continue;
		iho = (whichconstant + 1) % 3;	ive = (iho + 1) % 3;
		thisx = diw / 2 + rf.obs_hkl[iho] * 4;
		if (thisx < 0 || thisx > diw - 1) continue;
		thisy = dih / 2 - rf.obs_hkl[ive] * 4;
		if (thisy < 0 || thisy > dih - 1) continue;
		/* determine the position of the point to plot */
		offset = diw * thisy + thisx;
		if ((offset < 0) || (offset >= disparea))
		  {
			printf("Attempt to plot point at [%4d,%4d] failed\n",
				thisx, thisy);
			continue;
		  }
		cp = idata + diw * thisy + thisx;
		/* if a point is already occupied, save the brightest value */
		if ((*cp) < irange)
		  { /* duplicate this value over a 3x3 region and leave
		 	the fourth row and column blank */
			for (io = 0; io < 9; io++) *(cp + poff[io]) = irange;
		  }
	   }
	/* Convert the limits back to the original metric */
	for (irange = numranges, lip = &(limitv[0]); irange--; lip++)
		*lip = *lip * *lip * *lip;
	printf("Histogram: color #  # points\n");
	for (noor = thisy = 0, cp = idata; thisy < dih; thisy += 4)
	   {
		for (thisx = 0; thisx < diw; thisx += 4, cp += 4)
		   {
			if ((0 <= (io = *cp)) && (io < 9))
				(histo[(int)(*cp)])++;
			else noor++;
		   }
		cp += 3 * diw; /* skip duplicated & blank rows */
	   }
	for (irange = 0; irange < 9; irange++)
	   if (histo[irange]) printf("          %8d  %8d\n",
		irange, histo[irange]);
	if (noor) printf(
	 " %6d pixels appeared to have out-of-range histogram values\n", noor);
	(void)fclose(fplu);
	return reffound;
}

double nudge(double fv)
{ /*	this alters the range value of a variable v. */
	int	j;
	double	rounder, u, v, w;

	v = (fv < 0. ? -fv : fv);	/* get absolute value */
	if (v < DFUZZ) return 0.;
		/* figure out which decade we're in */
	if (v < 1.)
	  {
		for (u = 1.; u > DFUZZ; u /= 10.) if (v > u) break;
	  }
	else {
		if (v >= 1.e8) return fv;	/* don't try on huge ones */
		for (u = 1.; u < 1.e8; u *= 10.) if (v < u) break;
		u /= 10.;
	     }
	w = v / u;	/* always between 1 and 10 */
	if (w < 1. + DFUZZ) return fv;	/* not worth changing */
	/* round down to the nearest 10th from 1 to 1.5;
	 to the nearest 5th from 1.5 to 3; to the nearest half from 3 to 6;
	 and to the nearest integer from 6 to 10. */
	rounder = (w < 1.5 ? 10. : (w < 3. ? 5. : (w < 6. ? 2. : 1. )));
	j = rounder * w + 1.;	w = u * ((double)j) / rounder;
	return (fv < 0. ? -w : w);
}

int scatt_uref(char *uname, int whichdep, int whichind, DISP_ATTRS *dispatt)
{ /*	This routine does scatter-plot statistics on a urefls file.
	It returns various negative integers upon failure,
	and it returns the number of reflections found if successful. */
	char		*cp, *idata;
	int		dih, disparea, diw, irange, numranges;
	int		reffound, thisx, thisy, *histo;
	double		dv, dvdel, fdel, fmx, fmn, iv, ivdel;
	double		llam, lscx, lscy, thisf, thiss;
	double		dvmax, dvmin, ivmax, ivmin;
	float		*limitv, *lip;
	FILE		*fplu;
	OBSREF		rf;

	numranges = dispatt->di_nranges;
	if (dispatt->di_wintype)
	  {
		diw = dispatt->di_magw;	dih = dispatt->di_magh;
	  }
	else {
		diw = dispatt->di_wid;	dih = dispatt->di_ht;
	     }
	lscx = dispatt->di_xsca;	lscy = dispatt->di_ysca;
	llam = safelambda(dispatt);
	if (whichind < 0 || whichind >= 4 || whichdep < 0 || whichdep > 5)
		return -6;
	if (NULL == (fplu = xgopen(uname, "r"))) return -1; /* open UREFLS */
	/* find min and max of the relevant function within the UREFLS file*/
	ivmin = dvmin = fmn = 1.e30;	ivmax = dvmax = fmx = -fmn;
	disparea = diw * dih;
	idata = dispatt->di_rotm; zeromem((AVOID *)idata, (size_t)disparea);
	for (reffound = 0; 0 < xgread((void *)&rf, XG_OBSREF,fplu); reffound++)
	   { /* establish ranges of the color code (I/sigma) value */
		if ((thiss = rf.obs_sigma) < FUZZ) thiss = 0.1;
		thisf = rf.obs_ii / sqrt(thiss);
		if (thisf < fmn) fmn = thisf;	if (thisf > fmx) fmx = thisf;
		switch (whichdep)	/* switch on dependent variable */
		      {
			case	0:	/* err(omega) */
				dv = rf.obs_omob - rf.obs_ompred;	break;
			case	1:	/* err(X) */
				dv = ((double)(rf.obs_xob - rf.obs_xpred))
					/ lscx;	break;
			case	2:	/* err(Y) */
				dv = ((double)(rf.obs_yob - rf.obs_ypred))
					/ lscy;	break;
			case	3:	/* <I> */
				dv = rf.obs_ii;	break;
			case	4:	default:	/* <I/sigma> */
				if (rf.obs_sigma > FUZZ)
					dv = rf.obs_ii / sqrt(rf.obs_sigma);
				else	dv = 0.;
				break;
		      }
		if (dv < dvmin) dvmin = dv;	if (dv > dvmax) dvmax = dv;
		switch (whichind)	/* switch on independent variable */
		      {
			case	0:	default: /* sin(theta) / lambda */
				iv = ((double)rf.obs_ssq) / 16380.;
				iv = sqrt(iv) / (2. * llam);	break;
			case	1:	/* omega */
				iv = rf.obs_omob;	break;
			case	2:	/* X */
				iv = ((double)rf.obs_xob) / lscx;	break;
			case	3:	/* Y */
				iv = ((double)rf.obs_yob) / lscy;	break;
		      }
		if (iv < ivmin) ivmin = iv;	if (iv > ivmax) ivmax = iv;
	   }
	fprintf(stdout, "Range of I/sigma,independent,dependent variables:\n");
	fprintf(stdout, " I/sig min I/sig max indmin indmax depmin depmax\n");
	fprintf(stdout, " %9.2f %9.2f %6.1f %6.1f %6.2f %6.2f\n",
		fmn, fmx, ivmin, ivmax, dvmin, dvmax);
	if (0 >= reffound) return -2;	/* if no refls, return error code */
	if (-1 == fseek(fplu, 0L, 0)) return -3;	/* rewind file */
	if (fmx - fmn < DFUZZ) return -4;	/* not enough contrast */
	if (dvmax - dvmin < DFUZZ) return -4;	/* "    "      " */
	if (ivmax - ivmin < DFUZZ) return -4;	/* "    "      " */
	limitv = dispatt->di_limi;
	limitv[numranges] = fmn;	limitv[numranges + 1] = fmx;
		/* adjust the ranges of the (X,Y) variables a bit */
	ivmax = nudge(ivmax);	ivmin = nudge(ivmin);
	dvmax = nudge(dvmax);	dvmin = nudge(dvmin);
	if (ivmin > 0. && ivmax > 4. * ivmin) ivmin = 0.;
	if (dvmin > 0. && dvmax > 4. * dvmin) dvmin = 0.;
	/* launder the I/sigma range values using cube root of f as scale */
	fmx = mycbrt(fmx);	fmn = mycbrt(fmn);
	 /* determine upper limit of each range of data */
	fdel = (fmx - fmn) / (double)(numranges - 1);
	if (fdel < DFUZZ) return -5;
	 /* limitv[0] is just a place to stick the minimum observed value */
	limitv[0] = fmn;
	for (irange = 1; irange < numranges; irange++)
		limitv[irange] = fmn + (float)irange * fdel;
	/* determine range widths for the independent and dependent vbls */
	ivdel = (ivmax - ivmin) / (double)(diw - 1);
	dvdel = (dvmax - dvmin) / (double)(dih - 1);
	/* go back through the data and bin things */
	for ( ; 0 < xgread((void *)&rf, XG_OBSREF, fplu); )
	   {
		if ((thiss = rf.obs_sigma) < FUZZ) thiss = 0.1;
		thisf = mycbrt(rf.obs_ii / sqrt(thiss));
		irange = 1 + (thisf - fmn) / fdel;
		if (irange == numranges) irange = numranges - 1;
		if (irange < 1 || irange > numranges) continue;
		switch (whichdep)	/* switch on dependent variable */
		      {
			case	0:	/* err(omega) */
				dv = rf.obs_omob - rf.obs_ompred;	break;
			case	1:	/* err(X) */
				dv = ((double)(rf.obs_xob - rf.obs_xpred))
					/ lscx;	break;
			case	2:	/* err(Y) */
				dv = ((double)(rf.obs_yob - rf.obs_ypred))
					/ lscy;	break;
			case	3:	/* <I> */
				dv = rf.obs_ii;	break;
			case	4:	default:	/* <I/sigma> */
				if (rf.obs_sigma > FUZZ)
					dv = rf.obs_ii / sqrt(rf.obs_sigma);
				else	dv = 0.;
				break;
		      }
		/* we store the y values upside down so < 0 is below > 0 */
		thisy = dih - 1 - ((dv - dvmin) / dvdel + 0.49999);
		if (thisy < 0 || thisy > dih) continue;
		switch (whichind)	/* switch on independent variable */
		      {
			case	0:	default: /* sin(theta) / lambda */
				iv = ((double)rf.obs_ssq) / 16380.;
				iv = sqrt(iv) / (2. * llam);	break;
			case	1:	/* omega */
				iv = rf.obs_omob;	break;
			case	2:	/* err(X) */
				iv = ((double)rf.obs_xob) / lscx;	break;
			case	3:	/* err(Y) */
				iv = ((double)rf.obs_yob) / lscy;	break;
		      }
		thisx = (iv - ivmin) / ivdel + 0.49999;
		if (thisx < 0 || thisx > diw) continue;
		/* determine the position of the point to plot */
		cp = idata + diw * thisy + thisx;
		/* if a point is already occupied, save the brightest value */
		if (irange > *cp) *cp = irange;
	   }
	/* Convert the limits back to the original metric */
	for (irange = numranges, lip = limitv; irange--; lip++)
		*lip = *lip * *lip * *lip;
	if (NULL == (histo = (int *)callop((size_t)(numranges + 1),
	 sizeof (int), "histo in scatt_uref"))) return -6;
	fprintf(stdout, "Histogram: color #  # points\n");
	for (thisy = disparea, cp = idata; thisy--; cp++) (histo[(int)(*cp)])++;
	for (irange = 0; irange < 9; irange++)
	   if (histo[irange]) fprintf(stdout, "          %8d  %8d\n",
		irange, histo[irange]);
	(void)fclose(fplu);
	/* we want to return some of these intermediate calculations to
	 the caller as elements of the `limitv' array */
	limitv[numranges+2] = ivmin;	limitv[numranges+3] = ivmax;
	limitv[numranges+4] = ivdel;	limitv[numranges+5] = dvmin;
	limitv[numranges+6] = dvmax;	limitv[numranges+7] = dvdel;
	memzap((AVOID **)&histo, "histo in scat_uref");	return reffound;
}

/* end of dispuref.c */
