/* xgimage.h -- defines, typedefs, prototypes for image management in X-GEN

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions
	10 Aug 2006: eliminate separate defines for byteswapped formats
	 9 Aug 2006: redefine Mar IPs so they all have format=0x4B0=1200
	 2 Aug 2006: add 2300-pixel Mar IPs
	23 Jun 2006: create xgimage.h from image-specific stuff in xgen.h
 Relevant changes to xgen.h:
	 3 Jun 2006: 5.7.5: many routines now static: not declared here now
	 7 Mar 2006: 5.7.3: Mar IP filetypes augmented and systematized;
		Mar headers improved; plor element added to PEAK
	 5 Dec 2005: 5.7.0: inline function declarations & function prototypes;
		numerous bugfixes; effective framenumbers available
	14 Jan 2005: maybe_image -> imagename and imagelike: marccd 300mm added
	08 Feb 2004: ANNOUNCE_BOX flag for detector reading added;
	 numerous #defines reordered so they're a bit easier to find
	21 Jan 2004: license message inserted
	18 Jan 2004: cosmetic changes to several structure types
	15 Jan 2004: DUMPRUN image-dumping structure type
	 7 Jan 2003: recognize RAXIS6 format and its byteswapped equivalent
	 5 Jan 2003: reorganized source-file header
	25 Mar 2002: removed MSI attachments
	24 Aug 1995: add IPR_HAS16BFLOWS flag for images;
		other flags now in hex, not octal
	 6 Jul 1995: add FMT_DIPX000 and its byteswapped equivalent
		new flags for image-reading;
	12 May 1995: add #defines for FMT_DIP100 and DIP320
		(MACScience image plates)
	 2 Mar 1995: add FRAMEOFF structure type
		IPR_??? flags for images; FMT_FUJI and byteswapped equiv added
	 4 Nov 1994: add FMT_CORNELL and byteswapped equiv detector formats
		add structure elements for detector type and X-GEN version #
		to REFPARS and CRYSTAL structures
	12 Aug 1994: add 2000x2000 MAR-detector formats
	21 Apr 1994: initial X-GEN version
	24 Nov 1986: incorporate 1024-byte image header as an option.
	xx xxx 1983: created in the dim past.

   Purpose: #defines, typedefs, function prototypes for images in X-GEN
   Cross References: xgen.h, unixxdefs.h, vmsxdefs.h

********************************************************************/
#ifndef XGIMAGE_H
#define XGIMAGE_H	1
#include	"xgen.h"
	/* miscellaneous parameters */
#define		DEFBLOCK	04000	/* # bytes/record in image */
#define		SDMS_RANGE	(4096)	/* SDMS frame format = 4096 + height */
#define		SIZMAIL		128	/* read&write portion of MAIL struct */
	/* detector format values */
#define		FMT_CAD		0xa		/* Cadmus format */
#define		FMT_CCD		(131328L)
#define		FMT_CCD_VMS	(66048L)
#define		FMT_CCD_ALT	(256)
#define		FMT_CCD_VMSALT	(512)
#define		FMT_PCS_16	0x10		/* PCS 16-bit format */
#define		FMT_PCS_8	0x8		/* PCS  8-bit format */
#define		FMT_SAXII	(0x464F524DL)	/* SAXII  8-bit unswapped fmt*/
#define		FMT_SAXII_16	(0x464F524EL)	/* SAXII 16-bit "    " */
#define		FMT_FAST	0x45480a7b	/* Ascii FAST/CCD, X&Y normal */
#define		FMT_FASTALT	0x45480a7a	/* internal:FAST/CCD X&Y rev.*/
#define		FMT_RAXIS	0x58412D52	/* R-axis */
#define		FMT_RAXIS6	0x49584152	/* R-axisVI, byteswapped */
#define		FMT_MAR		0x04b0		/* Mar image plate, unswapped*/
#define		FMT_MARCCD	0x4D4D2A00	/* mar CCD tif format (MM*\0)*/
#define		FMT_MARCCD_L	0x49492A00	/* mar CCD tif,Littlend(II*\0)*/
#define		FMT_QTEX	0x40003FF	/* Quantex fmt, unswapped */
#define		FMT_CORNELL	0x9c44944	/* Cornell Fuji image plate */
#define		FMT_DIP100	0x1357986	/* MACScience 1600x1600 */
#define		FMT_DIP320	0x1357990	/* MACScience 1600x3200 */
#define		FMT_DIPX000	0x44495030	/* MACScience DIP2000&3000 */
#define		FMT_VMS		0xa	/* VMS fmt,after conversion*/
#define		FMT_UNREAD	0x7ff	/* "format" val before reading it */
#define		FMT_FUJI	0x35798642	/* headerless Fuji */
 /* return values from routines that read detector images */
#define		FRAME_OK	0	/* successful operation */
#define		FRAME_BADOPEN	1	/* error opening frame file */
#define		FRAME_BADMAIL	3	/* error reading mailbox of frm*/
#define		FRAME_BADFLOWS	5	/* error reading overflows */
#define		FRAME_BAD8BIT	011	/* error reading 8-bit data */
#define		FRAME_BADCLOSE	021	/* error closing frame file */
#define		FRAME_ALREADY_CONVERTED	041	/* frame has been converted */
#define		FRAME_FMT_UNCODED	0101	/*no code avail for this fmt*/
#define		FRAME_FMT_UNKNOWN	0201	/* unrecognized filetype */
#define		FRAME_NOT_COMPRESSED	0401	/* compression failed */
#define		FRAME_NO_MEMORY	01001	/* memory alloc failed in image work*/
#define		FRAME_BAD16BIT	02001	/* error reading 16-bit data */
#define		FRAME_BADLENGTH	04001	/* frame is too short */
#define		FRAME_TIMEDOUT	010001	/* frame didn't appear within timeout*/
	 /* constitutive properties of detector images (MAIL.improps) */
#define		IPR_XFASTEST	0x1	/* X dimension reads fastest */
#define		IPR_STARTBOTTOM	0x2	/* read starts from bottom corner */
#define		IPR_STARTRIGHT	0x4	/* read starts from right corner */
#define		IPR_BYTESWAPPED	0x8	/* image's byteorder != computer's */
#define		IPR_IS8BIT	0x10	/* 8-bit data, not 16 */
#define		IPR_ISCOMPRESS	0x20	/* Lempel-Ziv compressed data*/
#define		IPR_ISDARK	0x40	/* it's a dark-current image */
#define		IPR_HAS16BFLOWS	0x80	/* image has 16-bit overflows */
#define		IPR_PAD		0x100	/* R-Axis II 1900x1900 padded images*/
#define		IPR_MIRRORX	0x200	/* swap left for right in image */
#define		IPR_MIRRORY	0x400	/* swap top for bottom */
#define		IPR_INVERTSWAP	0x800	/* toggle whether to byteswap*/
#define		IPR_NEEDCOUNT	0x1000	/* eventcount not recorded in header*/
#define		IPR_LOGSCALE	0x2000	/* logarithmic scale for image */
#define		IPR_RAXISCOMPRESS	0x4000	/* R-Axis Compression type */
/* #define		IPR_BADCOLUMN0	0x4000	column 0 contains bad data */
#define		IPR_HASPEDESTAL	0x8000	/* image with uncompensated pedestal*/
#define		MAXMODIFIERS	48

/* structure type definitions */

/* 	Overflow-table element, as read in from disk */
 typedef struct {
		u_short	value, hiword, loword;
		} OFLOW;

	/* header data structures for specific detector types */
 typedef struct {
		xgulong	pformat;	/*16=PCS 16-bit; 8=PCS 8-bit; 10=Cad*/
		xgulong	pstatus;	/* bit0=read lock; bit1=write lock */
		xglong	pmailbox;	/* offset to mailbox */
		xglong	pcorrx;	/* offset x correction array */
		xglong	pcorry;	/* offset y correction array */
		xglong	pthedata;	/* offset to data space */
		xglong	ptopdata;	/* offset to end of data space */
		xgulong	pdimen;		/* one side of data space */
		xgulong	pcellsize;	/* size of pixel in 1024 bins */
		xgulong	pzoom;		/* magnification of image */
		xgulong	pquadrant;	/* 0=center, 1=ulhc,2=ur,3=ll,4=lr */
		xgulong	peventcount;	/* total number of counts */
		xgulong	p8bitoflow;	/* # of pixels with count > 255 */
		xgulong	p16bitoflow;	/* # of pixels with count > 65535*/
		xgulong	pzmax;		/* maximum count in a pixel */
		xgulong	pzmaxx;		/* x address of pzmax */
		xgulong	pzmaxy;		/* y address of pzmax */
		xgulong	pzdisp;		/* default display value */
		xgulong	pontime;	/* # of ontime counts when polled */
		xgulong	plate;		/* # of counts present when polled */
		xgulong	pdatetime[2];	/* date and time */
		xgulong	paccumsecs;	/* # of seconds since start of RUN */
		xgulong	pelap100ths;	/* accum 100ths of secs, this frame */
		xgulong	pontime_inc;	/* ontime from last add */
		xgulong	pnonzpix;	/* # of pixels with >= 1 count */
		xgulong	prffu;		/* reserved for future use */
		xglong	pframeno;	/* frame number */
		xglong	pdummy[4];	/* extra variable space */
		xgulong	poverflow[96];	/* offsets of the 16-bit overflows */
		/* pointers and variables take up 512 bytes of space ||| */
		char	pfilename[16];	/* name of the frame */
		char	pfile_info[240]; /* ascii space for frame info */
		char	pdataset_info[256];	/*ascii comment space,dataset*/
		} PMAIL;

 typedef struct { /* contents of Mar Image-plate detector image header */
		xgulong	ma_width, ma_height, ma_recordlen, ma_numrecords;
		xgulong	ma_n16flows, ma_n16records, ma_ncountstart;
		xgulong	ma_ncountend, ma_expostime, ma_dose;
		float	ma_xrayctsum, ma_maxradius, ma_minradius;
		float	ma_xpcm, ma_ypcm, ma_xpcm2, ma_ypcm2; 
		float	ma_x0, ma_y0, ma_lambda;
		float	ma_plated, ma_phi0, ma_phie, ma_omega, ma_highmult;
		char	ma_datetim[28];
		} MARHEAD;

 typedef struct	{	xglong mxyoff, minten;
		} MARFLOW; /* mar image plate overflows */

 typedef struct {						   /* #bytes*/
		char	raxis0[10], ver0[10], cid0[20], csys0[12]; /*	  52*/
		float	cell0[6];				/*	  24*/
		char	spg0[12];				/*	  12*/
		float	emos0;					/*	   4*/
		char	memo0[80], res1[84];			/*       164*/
		char	date0[12], user0[20], targ0[4];		/*	  36*/
		float	wave0;					/*	   4*/
		char	monoc0[20];				/*	  20*/
		float	twoth0;					/*	   4*/
		char	coli0[20], filt0[4];			/*	  24*/
		float	dist0, xkv0, xma0;			/*	  12*/
		char	focus[12], xraymemo[80];		/*	  92*/
		xglong	ncylinder;				/*	   4*/
		float	weissenberg;				/*	   4*/
		char	res2[56];				/*	  56*/
		char	maxis0[4], baxis0[4];			/*	   8*/
		float	fi0, fis0, fie0;			/*	  12*/
		xglong	nosc0;					/*	   4*/
		float	expos0, beam0, zbeam0, duma0[3], ramu;	/*	  28*/
		char	res3[204];				/*	 204*/
		xglong	nxpix0, nzpix0;				/*	   8*/
		float	ysize0, zsize0;				/*	   8*/
		xglong	lrecl0, nrec0, nst0, ipno0;		/*	  16*/
		float	hilo, fad1, fad2;			/*	  12*/
		char	cpu[10], ip[10];			/*	  20*/
		xglong	drxz[3];				/*	  12*/
		char	res4[180];				/*	 180*/
		char	xtal[16];				/*	  16*/
		float	cell1[6], fixyz1[3];			/*	  36*/
		float	dist1, xbeam1, zbeam1, what1[3], backg1; /*	  28*/
		xglong	nbox1[5];				/*	  20*/
		float	amat1[9];				/*	  36*/
		xglong	istat1[6];				/*	  24*/
		char	res5[96], resu[768];		/* 96 + 768=	 864*/
	  } RAXISHEAD;		/* total bytes= (5+3)*256=8*256 =	2048*/

 typedef struct { /* working structures for image manipulation */
		int		imw_fd;		/* image file descriptor */
		u_int		imw_filetype;	/* image file type */
		size_t		imw_bsize;	/* size of working image buf*/
		size_t		imw_headlen;	/* size of image header */
		u_char		*imw_buf;	/* working image buffer */
		struct stat	*imw_stabu;	/* ptr to file stat struct */
		MAIL		*imw_lhd;	/* working MAIL structure ptr*/
		MAIL		*imw_darkp;	/* dark-current image ptr */
		DATARUN		*imw_drp;	/* working DATARUN pointer */
		} IMAGEWORK;

	/* function prototypes */
extern int cvt_8_to_16(MAIL *mp, int ftype);
extern int head2type(MAIL *headp);
extern int mchread(MAIL *mp, DATARUN *rhdp, int verboseread);
extern int mksaxflows(MAIL *mp, int imfd, u_char *obuf);
extern int mksaxihead(MAIL *mp, int otype, int imfd);
extern int open_image(char *imname, int rflags);
extern int output_frame(char *name, MAIL *mailp, xgulong outtype, int ruflags);
extern int read_header(MAIL *mpo, int ifd, int ruflags);
extern int saxiflows(MAIL *mp, int imfd);
extern int saxii_ascii(MAIL *mp, DATARUN *dp, char *inbu, int imfd);
extern void getcounts(MAIL *mp, double *nctp, xglong maxctp[]);
extern void makedrun(DATARUN *pcsdr, int iscvtsaxii);
extern void raxis_swap(RAXISHEAD *rhp, int direction);
extern void vmtoduf(u_char *ufp);
extern void xgensteps(DATARUN *datr, double flval);
extern char *imtypestr(int styp);
extern size_t imread(AVOID *base, size_t nb);
extern size_t imwrite(AVOID *base, unsigned nb);
extern IMAGEWORK *iminfo(int opno);
 
#endif /* XGIMAGE_H */

/* end of xgimage.h */
