/* xgmodel.h -- defines, typedefs, prototypes for 3-D model profiles in X-GEN

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions
	24 Jun 2006: created from xgen.h for 3-D model stuff
 Relevant changes to xgen.h
	08 Feb 2004: numerous #defines reordered so they're easier to find
	 1 Nov 2002: total rework of REGIONPROFILE structure
	20 Sep 2002: reordering of some defines and comments;
		incorporate CLF_* bits in cleanflags variable from cleanmask
	25 Mar 2002: removed MSI attachments
	21 Apr 1994: initial X-GEN version of xgen.h

   Purpose: Providing #defines, typedefs, and function declarations
	for handling 3-D model profiles
   Cross References: xgen.h, ../xglib/xgenmask.c

********************************************************************/
#ifndef XGMODEL_H
#define XGMODEL_H	1
#include	"xgen.h"
/* parameters */
	/* bits in the cleanflags variable */
#define	CLF_SEMIVERBOSE	01	/* verbose output */
#define	CLF_VERBOSE	02	/* even more verbose output */
#define	CLF_FLATTEN	04	/* zero pixels outside cutoff */
#define	CLF_EXTRAP	010	/* bring up pixels inside cutoff */
#define	CLF_BOTH	(CLF_FLATTEN | CLF_EXTRAP)
#define	CLF_OUTELLIPSE	020	/* replace profile with ellipsoid */
#define	CLF_WHATTODO	(CLF_OUTELLIPSE | CLF_BOTH)
#define	CLF_KEEPBORDER	040	/* don't set edge values to zero */
#define	CLF_ZCBORDER	0100	/* set edges to 0 even in center frm */
#define	CLF_INTERACT	0200	/* set edges to 0 even in center frm */
#define	CLF_DIFFVECTORS	0400	/* working diffrac-vector profiles */

	/* typedefs for structures */
 typedef struct { /* structure describing a detector-space 2-D profile */
		float	isqsum, maxf; /* <sum(Ii^2) and max intensity */
		xglong	mnobs;	/* number of contributing refls */
		sichar	xyint[BIGSPOTWID][BIGSPOTHT]; /* profile */
	       } PHIMASK;

 typedef struct { /* 3-D profile: array of 2-D profile elements */
		PHIMASK	phielt[PFRAMES];
	       } MASK_3D;

 typedef struct { /* 3-D profile for a detector region */
		short	m3_bx, m3_by, m3_bz;	/* (X,Y,omega)  dimensions */
		short	m3_nobs;		/* # of observations found */
		u_char	m3_imodel;		/* which model this is */
		u_char	m3_nmodels;		/* total # of models */
		short	m3_qx, m3_qy, m3_qz;	/* diffr.vector dimensions */
		short	m3_xm, m3_ym, m3_zm;	/* max of entire profile */
		short	m3_xmz, m3_ymz;		/* (XY) max of z element */
		short	m3_dx_dz, m3_dy_dz;	/* derivs of x & y wrt. z */
		double	m3_q1, m3_slopex, m3_slopey; /* deriv parameters */
		/* the following 9 parameters describe the scaling between
		 (X,Y,omega) and diffraction-vector dimensions */ 
		double	m3_lim[9];	/* (min, max, delta)(X,Y,Z) */
		double	m3_isqsum;	/* sum (I^2) */
		double	m3_xyoi, m3_dvi;	/* intensity values both ways*/
		float	*m3_xyop;	/* (X,Y,omega)  3D profile array */
		float	*m3_dvp;	/* diffr.vector 3D profile array */
		sichar	*m3_xyoip;	/* (X,Y,omega)   char 3-D profiles */
		sichar	*m3_dvip;	/* diffr. vector char 3-D profiles */
		} REGIONPROFILE;

 typedef struct { /* structure for dumping images to stdout or a file */
		int	dui_nmbox;		/* number of headers */
		int	dui_xce, dui_yce;	/* center of region in pix */
		int	dui_wid, dui_ht;	/* width & height of region */
		int	dui_imflags;		/* image flags in use */
		int	dui_maxlines;		/* max # lines to stdout */
		char	*dui_imname;		/* name of image */
		char	*dui_lcname;		/* lowercase file type */
		FILE	*dui_fpd;		/* file to dump info to */
		MAIL	*dui_heads;		/* pointer to header array */
		MASK_3D	*dui_wmask;		/* 3-D profile to display */
		FRAMEINDEX	*dui_fmi;	/* frame index */
		} DUMPRUN;

extern int blankf(float *ve, int cz);
extern int contig(REGIONPROFILE *rgp, int thisfmonly);
extern int dospread(REGIONPROFILE *rep, int wh, double *suf, int *nuf, double flr);
extern int ellip(REGIONPROFILE *rgp, int elty, int cfl, double flv, double lr, double *elp);
extern int ellsoid(REGIONPROFILE *rgp, int elty, int cfl, double flv, double lr, double *elp);
extern int force_populate(MAIL *mailp);
extern int rgptowmp(REGIONPROFILE *rgp, MASK_3D *wmp);
extern int vectowmp(float *ve, MASK_3D *wmp, int inobs, float misq, int cz);
extern int wmptovec(MASK_3D *wmp, float *ve);
extern int ikabc(double fx, double fy, double ikout[]);
extern int ixy2ev(int ix, int iy, double *ev);
extern void bstoloc(float *fve, int xyzl[]);
extern void cleanmodels(REGIONPROFILE *reb, int which, int recordsum);
extern void dispmodel(FILE *fp, REGIONPROFILE *m3p, int which);
extern void dumpimage(DUMPRUN *dumpp, int dumptype);
extern void kablim(MAIL *mailp, REGIONPROFILE *lrgb, int currbz);
extern void meanprofile(REGIONPROFILE *rgb, int iprof);
extern void model_ftoc(REGIONPROFILE *rgp);
extern void prmask(MASK_3D *wmp, int region, FILE *fp);
extern void show_masks(FILE *fp, MASK_3D *wp, BOUNDARY *bdp);
extern void spreadwmp(MASK_3D *wmp);
extern void tell_dvcoef(FILE *fp, REGIONPROFILE *rgp);
extern void zero_boundary(REGIONPROFILE *rgp, int cflags);
 
#endif /* XGMODEL_H */

/* end of xgmodel.h */
