/* analu.c -- Get statistics for a UREFLS file as a function of frame number

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	 5 Oct 2005: housekeeping
	11 May 2005: check return status from xgjournal
	12 Apr 2005: change name from anal_u.c to analu.c to eliminate confusion
	28 Aug 2004: routines used only here are declared as static; cosmetics
	18 Aug 2004: ANALURUN type replaces static globals
	24 Jan 2004: some static globals replaced with locals
	21 Jan 2004: license message inserted
	 2 Jan 2004: globals replaced by static globals
	19 Sep 2003: bright2 initialized in add_to_stats() no matter what
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	 5 Jan 2003: reorganized source-file header
	25 Mar 2002: removed MSI attachments
	15 Jun 1995: add estimate of fraction of reflections that are fully
		recorded within the brightest image.
	31 May 1995: replace long u_nob, etc. with ints; %d replaces %ld twice
	 2 Mar 1995: diagnostics to stdout
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 5 Oct 1994: use xgread to input OBSREF data
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: a few I/O changes and global variables moved around
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version

   Purpose: analyzing a UREFLS file.  For groups of frames and for the
	entire file it obtains various statistics and prints them.
   Cross references: ../utilities/redurefl.c
 For each N frames and for the entire UREFLS file this routine prints
 out various indicia. N by default is 20, but that can be changed.
*************************************************************************/

#include	"xgen.h"
#define		DEF_DELFM	(20.)	/* default # fms per analyzed range*/
#define		NIOSRANGES	8	/* # of bins in I/sigma */
#define		DEF_FULLCUTOFF	0.05	/* def. I(neighbor)/I(bright) cutoff*/
#define		DEF_IOSFULLCUT	3.	/* partiality anal. I/sigma cutoff */
#define		USE_SUMMED	1	/* do stats on non-profile-fitted I's*/
#define		USE_KABSCH	2	/* do stats on Kabsch-geometry I's*/

 typedef struct {
		int	u_nob;	/* # of observations */
		int	u_nios[NIOSRANGES];	/* histogram on I/sigma */
		int	u_nallneg;	/* # refls with all frames < -<bgnd> */
		int	u_fullden;	/* all obs avail.for profile analysis*/
		double	u_fmmin, u_fmmax;	/* min, max frame # */
		double	u_imean;	/* mean I, profile-fitted data */
		double	u_iosmean;	/* mean I/sigma, profile-fitted */
		double	u_smean;	/* mean sigma, profile-fitted */
		double	u_bgmean;	/* mean background */
		double	u_full;		/* # of fully-recorded observations */
		} UR_STAT;

 typedef struct {
		xglong	au_flags;
		double	au_delfm, au_fullcut, au_iosfull;
		} ANALURUN;

extern int analumain(int argc, char *argv[]);

static void sumanalurun(FILE *fp, ANALURUN *aup)
{ /*	This summarizes the run conditions to the stream fp */
	static char	*fnames[] = {"urefls", "xlog", NULL };
	static char	ftypes[] = { 1, 2, 0 };

	header(fp, PROCESSING, fnames, ftypes);
	fprintf(fp, "Data will be grouped into ranges %6.2f frames wide\n",
		aup->au_delfm);
	fprintf(fp, "%s %s data will be analyzed\n",
	 ((aup->au_flags & USE_KABSCH) ? "Kabsch" : "X,Y,omega"),
	 ((aup->au_flags & USE_SUMMED) ? "Summation" : "Profile-fitted"));
	fprintf(fp,
	 "Partiality analysis performed on data with I/sigma > %6.2f\n",
		aup->au_iosfull);
	fprintf(fp,
	 "Reflections are considered fully-recorded if I(2nd brightest) < ");
	fprintf(fp, "%6.2f * I(brightest frame)\n", aup->au_fullcut);
}

static int get_iosrange(double ival)
{ /*	This returns an integer between 0 and NIOSRANGES-1 inclusive
	indicating which I/sigma range an observation falls in.
	It's stupid at the moment; I'll replace it with a tree analysis later*/
	register int	range;
	register double	*zp;
	static double	iosmax[] = {0., 1., 2., 5., 10., 20., 40. };

	for (range = 0, zp = &(iosmax[0]); range < NIOSRANGES-1; range++, zp++)
		if (ival < *zp) return range;
	return (NIOSRANGES - 1);
}

static void add_to_stats(OBSREF *urp, UR_STAT *usp, ANALURUN *aup)
{ /*	This adds to the sums used in the statistical analysis of the data. */
	register int	allneg, fm, ithis, ithis2;
	double		bright2, inten, ios, pebright, sig, this;

	usp->u_nob++;
	if (aup->au_flags & USE_SUMMED)
	  {
		if (aup->au_flags & USE_KABSCH)
		  {
			inten = urp->obs_iunfit +
				((double)(urp->obs_bgunfit))/ 16.;
			sig = urp->obs_sunfit + ((double)(urp->obs_reflno)) /8.;
		  }
		else {
			inten = urp->obs_iunfit; sig = urp->obs_sunfit;
		     }
	  }
	else {
		if (aup->au_flags & USE_KABSCH)
		  {
			inten = urp->obs_ii +
				((double)(urp->obs_shift)) / 16.;
			sig = urp->obs_sigma + ((double)(urp->obs_widob)) / 8.;
		  }
		else {
			inten = urp->obs_ii;	sig = urp->obs_sigma;
		     }
	  }
	if (sig < FUZZ)
	  {
		sig = 0.; ios = 1.;
	  }
	else {
		sig = sqrt(sig);	ios = inten / sig;
	     }
	usp->u_imean += inten; usp->u_smean += sig;
	usp->u_bgmean += (double)urp->obs_bgnd; usp->u_iosmean += ios;
	(usp->u_nios[get_iosrange(ios)])++;
	for (fm = 0, allneg = 1; fm < PFRAMES; fm++)
	   if (urp->obs_frmint[fm] > -(urp->obs_bgnd))
	     {
		allneg = 0; break;
	     }
	if (allneg) (usp->u_nallneg)++;
	/* see if the observation's profile is sharp enough to
	 qualify as "fully recorded" in the brightest profile element */
	if (ios > aup->au_iosfull)
	  {
		bright2 = urp->obs_frmint[PFRAMES/2];
		for (ithis = ithis2= -1, pebright = fm = 0; fm < PFRAMES; fm++)
		   if ((this = urp->obs_frmint[fm]) > pebright)
		     {
			pebright = this;	ithis = fm;
		     }
		if (ithis > -1)
		  {
			for (bright2 = fm = 0; fm < PFRAMES; fm++)
			   if ((fm != ithis) && ((this = urp->obs_frmint[fm]) >
				bright2))
			     {
				bright2 = this;	ithis2 = fm;
			     }
		  }
		if ((ithis2 > -1) && (pebright > DFUZZ))
		  {
			if (bright2 < pebright * aup->au_fullcut)
				usp->u_full += 1.;
			usp->u_fullden++;
		  }
	  }
}

static void mean_stats(UR_STAT *usp)
{ /*	This computes mean values based on the sums found in *usp */
	double	rnob;

	if ((rnob = usp->u_nob) < DFUZZ) return;
	usp->u_imean /= rnob;	usp->u_iosmean /= rnob;
	usp->u_bgmean /= rnob;	usp->u_smean /= rnob;
	if ((rnob = usp->u_fullden) >= DFUZZ) usp->u_full /= rnob;
	else usp->u_full = 0.;
}

static void out_range(FILE *fp, UR_STAT *usp)
{ /*	This prints the contents of a UR_STAT structure on the stream fp */
	register int	nios;

	fprintf(fp, " %7.2f %7.2f %7d %9.3f %9.4f %9.4f %9.4f %6d %7.4f",
		usp->u_fmmin, usp->u_fmmax, usp->u_nob, usp->u_imean,
		usp->u_iosmean, usp->u_smean, usp->u_bgmean, usp->u_nallneg,
		usp->u_full);
	for (nios = 0; nios < NIOSRANGES; nios++)
	   fprintf(fp, " %5d", usp->u_nios[nios]);
	fprintf(fp, "\n");
}

static void ustat_head(FILE *fp)
{ /*	This prints a header on stream fp for the statistics */
	fprintf(fp,
 " Minimum Maximum    # of      Mean      Mean      Mean      Mean  # obs  ");
	fprintf(fp, " frac. # of obs in I/sigma ranges\n");
	fprintf(fp,
 " Frame # Frame #   refls     Inten     I/sig     sigma      bgnd allneg  ");
	fprintf(fp," fulls    <0    <1    <2    <5   <10   <20   <40   >40\n");
}

int analumain(int argc, char *argv[])
{ /*	Toplevel routine for anal_u functionality */
	char		**flags, **dcp, *car, **origv;
	double		curr_frame, prev_omega;
	double		spec_delfm, spec_full, spec_iosfull;
	OBSREF		urf;
	FILE		*fpi, *fplog;
	UR_STAT		urall, urrange;
	ANALURUN	aur;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	if (qerror((AVOID *)(fpi = xgopen("urefls", "r")),
		"open input UREFLS file")) return EOPENERR;
	aur.au_flags = 0;	/* default: profile-fitted, non-Kabsch data */
	flags = flagproc(argc, argv, "FIKS");	/* flag arguments */
	/* reflection is considered fully-recorded if the intensity in the
	 second-brightest frame is less than aur.au_fullcut * the brightest. */
	aur.au_fullcut = DEF_FULLCUTOFF;
	/* in analyses of fully-recorded vs. partials, we only consider
	 reflections with I/sigma > aur.au_iosfull. */
	aur.au_iosfull = DEF_IOSFULLCUT;
	spec_full = -1.;	spec_iosfull = 0.;
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	   {
		car = (*dcp) + 1;
		if (**dcp == 'F') spec_full = atof(car);
		else if (**dcp == 'I') spec_iosfull = atof(car);
		else if (**dcp == 'K') aur.au_flags |= USE_KABSCH;
		else if (**dcp == 'S') aur.au_flags |= USE_SUMMED;
	   }
	if (spec_full >= 0.) aur.au_fullcut = spec_full;
	if (spec_iosfull > DFUZZ) aur.au_iosfull = spec_iosfull;
	origv = argv;
	if (argc)
	  {
		for (--argc, ++argv; argc; argc--,argv++)
			if (allnumeric(*argv)) break;
	  }
	if (argc && (0 < (spec_delfm = atof(*argv)))) /*command level default*/
		aur.au_delfm = spec_delfm;
	else if (NULL == origv) aur.au_delfm = 1; /* called from integmain */
	else aur.au_delfm = DEF_DELFM;
	if (verror(0 < xgread((void *)&urf, XG_OBSREF, fpi),
	 "read a reflection from the input UREFLS file")) return EREADERR;
	if (NULL == (fplog = propen("xlog"))) fplog = stdout;
	if (NULL == origv) /* give the functionality a name if needed */
		(void)xgprogname("anal_u", 4);
	sumanalurun(stdout, &aur);
	if (fplog != stdout) sumanalurun(fplog, &aur);
	ustat_head(fplog);
	zeromem((AVOID *)&urrange, sizeof (UR_STAT));
	zeromem((AVOID *)&urall, sizeof (UR_STAT));
	prev_omega = urrange.u_fmmin = urall.u_fmmin = urf.obs_omob;
	curr_frame = prev_omega + aur.au_delfm;
	add_to_stats(&urf, &urrange, &aur);
	add_to_stats(&urf, &urall, &aur);
	while (0 < xgread((void *)&urf, XG_OBSREF, fpi))
	     {
		if (curr_frame < urf.obs_omob)
		  {
			urrange.u_fmmax = prev_omega;
			mean_stats(&urrange);
			out_range(fplog, &urrange);
			zeromem((AVOID *)&urrange, sizeof (UR_STAT));
			urrange.u_fmmin = urf.obs_omob;
			curr_frame = urf.obs_omob + aur.au_delfm;
		  }
		add_to_stats(&urf, &urrange, &aur);
		add_to_stats(&urf, &urall, &aur);
		prev_omega = urf.obs_omob;
	    }
	(void)fclose(fpi);	fpi = NULL;
	mean_stats(&urrange);	mean_stats(&urall);
	urall.u_fmmax = urrange.u_fmmax = prev_omega;
	out_range(fplog, &urrange);	out_range(fplog, &urall);
	if (stdout != fplog)
	  {
		ustat_head(stdout);	out_range(stdout, &urall);
	  }
	/* we leave memory open if we're calling this from
	 inside another program */
	if (NULL != origv)
	  {
		if (stdout != fplog) (void)fclose(fplog);
		globalkill();	fplog = NULL;
	  }
	return EOK;
}

/* end of analu.c */
