/* integbg.c -- XGEN's routines for manipulating background values

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	23 Mar 2007: major rewrite of planebg so it scans a parallelogram-
		shaped region that roughly follows sin(theta) contours
	11 Aug 2006: put_bgimage: reorganize management of padding
	23 Jun 2006: include xgimage.h so we can read & write backgrounds
	17 Jun 2006: replace act_pixel with activmask
	 6 Jun 2006: housekeeping
	 4 Mar 2006: put_bgimage handles Mar IP formats better 
	31 Dec 2005: handle qsort with fewer than 2 elements correctly
	 5 Oct 2005: housekeeping
	 3 Sep 2005: planebg bugfix: gradient calc fixed
	15 Aug 2005: simplify bgp->bgp_p1 and bgp_p2 calculations
	28 Jun 2005: put_bgimage: zero all of bgoutbuf, not just the end
	14 Jun 2005: major rewrite: double pointers eliminated, bugs fixed;
	 in upd_bg(), scale the stored background properly in comparisons
	11 Jun 2005: housekeeping in planebg
	 5 Jun 2005: saveplane() bugfix: bgrf->bgy_sc = newsc.
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	30 Aug 2004: routines used only here are declared as static; cosmetics
	18 Aug 2004: eliminate all remaining static global variables (hooray);
		endputbg() has an explicit arg: pointer to reserved memory
	15 Jun 2004: imageflag bit for gzipping is now correctly abbreviated
	26 Mar 2004: build header of output background image more fully
	07 Feb 2004: in init_bg(), correct y limit in check for
		nonsensical values
	24 Jan 2004: several static globals replaced with local variables
	21 Jan 2004: license message inserted
	 2 Jan 2004: globals replaced by static globals and elements of
		INTEGRUN (and REFPARS) structures, and calls to getpact
	13 Dec 2003: elim unused vbls vofe from put_bgimage()
		and ly2 from planebg
	10 Dec 2003: eliminate pactp and outbuf as variables,
		use pactive and bgoutbuf instead
	22 Nov 2003: replace globals with elements of INTEGRUN structure
	21 Oct 2003: initialize oversc and vofp properly in output
	14 Oct 2003: rescale background values before output more aggressively
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	 6 Aug 2003: fix horrible misfeature in upd_bg() that caused
		8-bit overflows to be flat-topped at 255 counts
		from detectors for which the data are originally 16-bit
	10 May 2003: change which element of igp->igr_heads is accessed for
		reading during background construction
	17 Apr 2003: rename `frameno' to `fmno' to eliminate confusion
		with MAIL structure element
	20 Mar 2003: handle case where exposure time is unspecified;
		extra diagnostics for pathological output in init_bg()
	 5 Jan 2003: reorganized source-file header
	15 Nov 2002: extern declarations made here rather than in integext.h
	 1 Nov 2002: altered REGIONATT structures for region attributes;
		trivial reformatting for code readability
	25 Mar 2002: removed MSI attachments
	15 Jun 2001: in put_bgimage, get rid of memory for
		16-bit or 8-bit data and overflows when we're done with it.
	 1 Jul 1997: in upd_bg, only do the plane-background calcs
	24 Aug 1995: change shoebox storage conventions
		for bright pixels in Mar data
	31 May 1995: "register" -> "register int" throughout
	20 May 1995: return EWRITERR from put_bgimage if width > 512
	10 Mar 1995: truncate floating-point frame number more carefully
	 2 Mar 1995: get rid of references to BLOTREF structures
	 2 Mar 1995: major rewrite: PLANE backgrounds; new bookkeeping system
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
		bgnd image is 32-bits per pixel;
		bgnd initialization is back to front
	 2 Dec 1994: fix typo in fprintf that prevented compilation
	 1 Dec 1994: diagnostic output allows for detector (X,Y) > 999
	 4 Nov 1994: memory handling for list of pixels contained in refls
	14 Jun 1994: memory alloc for PIXENTs reworked; stripe-smoothing fixed
	 5 May 1994: source-file header cleaned up
		names changed from POLYDET to XGEN
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version
	18 Jan 1988: in upd_bg(), use a paralyzable model for deadtime loss
	
   Purpose: manipulations of the 16-bit, updatable background image
	created in and used in integration.
   Cross references:

*************************************************************************/
#define		NEED_TIME	1
#include	"xgimage.h"
#include	"integdefs.h"
#define		VMSBLOCK	2048
#define		NSTD_DEVS	6.	/* cutoff for averaging in stripes */
#define		COL_STD_DEVS	2.5	/* cutoff for averaging in columns */
#define		MAX_STRIPE_HEIGHT	70
#define		NSIGNEWBG	3.
#define		MAXFLIND	50
#define		PAD_TO_2TO_N	0x4000	/* pad each row so it's a power of 2 */
#define		NOTFLATENOUGH	0.005
#define		BIGDEVIAT	0.8
#define		FRACDEVIAT	0.3
#define		MINDIM		6	/* in planebg: smallest half-width */
#define		DIMCOEF		6	/*cphi-dependent half-width coeff */

static VOFLOW *do16flow(MAIL *schead, VOFLOW *vopi, int *premaxp)
{ /*	This handles a 16-bit overflow. It returns a pointer to the VOFLOW */
	int	offf;
	VOFLOW	*outv, *voob;
	static char do16fmsg[] = "overflows in do16flow";

	outv = NULL;	schead->oflowcount++;
	if (NULL == schead->overflow)
	  {
		if (NULL == (schead->overflow = (VOFLOW *)callop(*premaxp,
			sizeof (VOFLOW), do16fmsg))) return NULL;
		outv = schead->overflow;
	  }
	else if ((offf = vopi - schead->overflow) >= *premaxp)
	  {
		if (NULL == (voob = (VOFLOW *)callop((*premaxp) * 2,
		 sizeof (VOFLOW), "overflow reorg in do16flow")))
			return NULL;
		fprintf(stderr, "doubling the output image memory store\n");
		cpymem((AVOID *)voob, (AVOID *)schead->overflow,
		 offf * sizeof (VOFLOW));
		memzap((AVOID **)&(schead->overflow), do16fmsg);
		schead->overflow = voob;	outv = voob + offf;
		*(premaxp) *= 2;
	  }
	return outv;
}

int put_bgimage(MAIL *scrhead, INTEGRUN *igp)
{ /*	This writes out the background image as an ordinary frame file */
	register int	x, y;
	u_char		*cvend, *cvstart, *cp;
	u_short		bzmaxx, bzmaxy, *usp;
	int		hit, irv, maxbx, maxby;
	int		over28, over29, over2a, over2b, over2c, over2d;
	int		overscale, ruflags, wid;
	double		fval;
	size_t		imsiz, prevmax;
	xgulong		bzmax, maxinim, outtype, save_fmt, scalemax, *ip;
	MAIL		*thd;
	REFPARS		*xp;
	VOFLOW		*vofs, *vofp;
	PAU		ofu;
	static char	*shortname, *currname;
	static u_char	*bgoutbuf;
	static char	bgoutmsg[] = "bgoutbuf in put_bgimage";

	/* special call: eliminate bgoutbuf memory */
	if (NULL == scrhead)
	  {
		memzap((AVOID **)&bgoutbuf, bgoutmsg);	return EOK;
	  }
	/* We use an old algorithm if the height and width are <= 512.
	 Otherwise we get daring and use the actual output_frame routine. */
	xp = igp->igr_xp;
	scrhead->xdimen = wid = xp->xdetwid;
	scrhead->ydimen = hit = xp->xdetht;
	currname = trname("bgimage");	shortname = filepart(currname, 2);
	strncpy(scrhead->filename, shortname, 14);
	scrhead->filename[15] = '\0';
	scrhead->late = 0;	/* so no deadtime correction done */
	igp->igr_oversc = 1.;
	if (wid > 512 || hit > 512)
	  { /* all of these formats are inherently 16-bit. Simple that way */
		ruflags = 0;
		if ((wid == 1024) || (wid == 2048)) outtype = FMT_SAXII_16;
		else if (wid == 1900)
		  {
			outtype = FMT_RAXIS;	ruflags = PAD_TO_2TO_N;
		  }
		else outtype = FMT_MAR;	/* Mar IP formats */
#ifndef LITTLEND
		outtype = swawb(outtype);
#endif
		/* Oct 2003: before trying to assemble this output, we scale
		 it so that there aren't any 16-bit overflows. */
		for (maxinim = maxbx = maxby =  x = 0; x < wid; x++)
		   {
			ip = igp->igr_upwp + x * hit;
			for (y = 0; y < hit; y++, ip++)
			 if (*ip > maxinim)
			  {
				maxinim = *ip; maxbx = x;maxby = hit - y - 1;
			  }
		   }
		printf(
		 " Brightest pixel in background image is %8d @ [%4d,%4d]\n",
			maxinim, maxbx, maxby);
		if (maxinim >= 60000)
		  {
			igp->igr_oversc = ((double)maxinim) / 60000.;
			printf("Background scaled down by %8.4f\n",
				igp->igr_oversc);
			for (x = 0; x < wid; x++)
			   {
				ip = igp->igr_upwp + x * hit;
				for (y = 0; y < hit; y++, ip++)
				   {
					fval = *ip; fval /= igp->igr_oversc;
					*ip = fval + 0.49999;
				   }
			   }
		  }
		scrhead->count0 = igp->igr_oversc;
		scrhead->elap100ths = BGDSCALE * igp->igr_avexp100 /
			igp->igr_oversc;
		/* construct a header for the background image */
		/* pre-screen the overflow count. If it's excessive, rescale */
		scrhead->improps &= ~(IPR_IS8BIT);
		imsiz = (size_t)(wid * hit);
		/* June 2001: eliminate memory for 16-bit or 8-bit data and
		 overflows here */
		memzap((AVOID **)&(scrhead->overflow),
			"scrhead->overflow in put_bgimage");
		if (NULL == scrhead->data16)
		  {
			memzap((AVOID **)&(scrhead->data8),
				"scrhead->data8 in put_bgimage");
			/* construct the 16-bit data for it. Note that
			 we only perform this step if the elements of
			 data16 were previously zero; now they're not */
			if (NULL == (scrhead->data16 = (u_short *)callop(imsiz,
			 sizeof (u_short), "data16 in put_bgimage")))
				return EALLOC;
		  }
		else zeromem((AVOID *)(scrhead->data16),
			imsiz * sizeof (u_short));
		scrhead->zmax = scrhead->oflowcount = 0;
		scrhead->zmaxx = wid / 2;
		scrhead->zmaxy = hit / 2;
		scrhead->overflow = (VOFLOW *)0;
		scrhead->oflow16 = (VOFL16 *)0;
		prevmax = 10000;
		/* pre-screen the overflow count. If it's excessive, rescale */
		for (x = 0, usp = scrhead->data16,
		  vofp = scrhead->overflow; x < wid; x++)
		   {
			ip = igp->igr_upwp + x * hit;
			for (y = 0; y < hit; y++, ip++, usp++)
			   {
				if (scrhead->zmax < *ip)
				  {
					scrhead->zmax = (xgulong )(*ip);
					scrhead->zmaxx = (u_short)x;
					scrhead->zmaxy = (u_short)y;
				  }
				if (*ip >= 65535)
				  {
					vofp = do16flow(scrhead, vofp,
					 (int *)(&prevmax));
					if (NULL == vofp) return EALLOC;
					vofp->value = (*ip << 4) - 65535;
					vofp->xcoord = x;
					vofp->ycoord = hit - y - 1;
					vofp++;	*usp = 65535;
				  }
				else *usp = (u_short)(*ip);
			   }
		   }
		scrhead->format = outtype;
#ifndef LITTLEND	/* byte-swap ! */
		y = wid * hit;
		for (usp = scrhead->data16, x = 0; x < y; x++, usp++)
			*usp = uswab(*usp);
#endif
		irv = output_frame(currname, scrhead, outtype, ruflags);
		if (scrhead->overflow) memzap((AVOID **)&(scrhead->overflow),
			"overflows in put_bgimage");
		memzap((AVOID **)&(scrhead->data16),
			"16-bit data in put_bgimage");
		return irv;
	  }
#ifdef VMS
	if (NULL == (igp->igr_fpbgnd = fopen("bgimage", "w",
		"rfm=fix", "mrs=2048")))
#else
	if (NULL == (igp->igr_fpbgnd = xgopen("bgimage", "w")))
#endif
	  {
		fprintf(stderr,
		 "Warning: unable to open output background image file\n");
		fprintf(igp->igr_fpl,
		 "Warning: unable to open output background image file\n");
	  }
	if ((NULL == bgoutbuf) && (NULL == (bgoutbuf =
	 (u_char *)mallop(VMSBLOCK, bgoutmsg)))) return EALLOC;
	zeromem((AVOID *)bgoutbuf, VMSBLOCK);
/* stuff some useful things into the header for the background image */
	save_fmt = scrhead->format;	/* so we can restore it later */
	scrhead->format = FMT_VMS;
/* see how large the values in the background image are.  We grok this by
	counting how many pixels are greater than 2^n, n = 8,...,13. */
	over28 = over29 = over2a = over2b = over2c = over2d = 0;
	for (x = 0; x < wid; x++)
	   {
		ip = igp->igr_upwp + x * hit;
		for (y = 0; y < hit; y++, ip++)
		   if (*ip > 256)
		     {
			over28++;
			if (*ip > 512)
			  {
				over29++;
				if (*ip > 1024)
				  {
					over2a++;
					if (*ip > 2048)
					  {
						over2b++;
						if (*ip > 4096)
						  {
							over2c++;
							if (*ip > 8192)
								over2d++;
						  } /* 4096 */
					  } /* 2048 */
				  } /* 1024 */
			  } /* 512 */
		     } /* loop over y and 256 */
	   } /* loop over x */
/* pick an overall scale factor based on these numbers. */
	if (over2d > 1000) /* too many overflows! */
	  {
		fprintf(igp->igr_fpl,
		 "Too many very high values in background: no output\n");
		return EWRITERR;
	  }
	if (over28 < 343) overscale = 1;
	else if (over29 < 343) overscale = 2;
	else if (over2a < 343) overscale = 4;
	else if (over2b < 343) overscale = 8;
	else if (over2c < 343) overscale = 16;
	else if (over2d < 343) overscale = 32;
	else overscale = 64;
	scalemax = 255 * overscale;
	scrhead->count0 = overscale;
	scrhead->elap100ths = BGDSCALE * igp->igr_avexp100 / (double)overscale;
/* now go through the image, recording overflows and dividing all values by
	`overscale'.  We're going to use the memory ordinarily reserved for
	one of the 8-bit frame data arrays, so we'll be sure we have enough. */
	thd = igp->igr_heads;
	if (thd->data8)	ofu.ucpt = thd->data8;	else ofu.uspt = thd->data16;
	vofp = vofs = (VOFLOW *)(ofu.vpt);
	for (bzmax = x = bzmaxx = bzmaxy = 0; x < wid; x++)
	   {
		ip = igp->igr_upwp + x * hit;
		for (y = 0; y < hit; y++, ip++)
		   {
			if (*ip > bzmax)
			  {
				bzmax = *ip; bzmaxx = x; bzmaxy = y;
			  }
			if (*ip > scalemax)
			  {
				vofp->value = (*ip) / overscale;
				vofp->xcoord = x;
				(vofp++)->ycoord = hit - y - 1;
			  }
		   }
	  }
	scrhead->oflowcount = vofp - vofs;
	if ((scrhead->zmax = bzmax / overscale))
	  {
		scrhead->zmaxx = bzmaxx; scrhead->zmaxy = bzmaxy;
	  }
	else {
		scrhead->zmaxx = scrhead->zmaxy = 255;
	     }
	/* overflows should be sorted primarily on Y, second on X */
	if (scrhead->oflowcount > 1) qsort((AVOID *)vofs,
	 (unsigned)scrhead->oflowcount, sizeof (VOFLOW), (QPE)ofycmp);
/* now write out the header and the overflows */
	cpymem((AVOID *)bgoutbuf, (AVOID *)scrhead, sizeof (MAIL));
	if (verror((int)fwrite((void *)bgoutbuf, VMSBLOCK, 1, igp->igr_fpbgnd),
		"write header of background file to disk")) return EWRITERR;
	if (scrhead->oflowcount)
	  {
		cvstart = (u_char *)vofs;	cvend = (u_char *)vofp;
		for (x = 0; x < VMSBLOCK; x++) bgoutbuf[x] = 0;
		if (scrhead->oflowcount < VMSBLOCK / sizeof (VOFLOW))
		  {
			cpymem((AVOID *)bgoutbuf, (AVOID *)cvstart,
				(unsigned)(cvend - cvstart));
			if (verror((int)fwrite((void *)bgoutbuf, VMSBLOCK, 1,
			 igp->igr_fpbgnd),
			 "write background overflows to disk")) return EWRITERR;
		  }
		else
		  {
			for (cp= cvstart; cp< cvend-VMSBLOCK; cp += VMSBLOCK)
			   if (verror((int)fwrite((void *)cp, VMSBLOCK, 1,
				igp->igr_fpbgnd),
				 "write background overflows to disk"))
					return EWRITERR;
			cpymem((AVOID *)bgoutbuf, (AVOID *)cp,
				(unsigned)(cvend - cp));
			if (verror((int)fwrite((void *)bgoutbuf, VMSBLOCK, 1,
			 igp->igr_fpbgnd),
			  "write final background overflows to disk"))
				return EWRITERR;
		  }
	  }
/* now write out the 8-bit data */
	for (x = 0, cp = bgoutbuf; x < wid; x++)
	   {
		ip = igp->igr_upwp + x * hit;
		for (y = 0; y < hit; y++, ip++)
			*cp++ = (*ip <= scalemax ? ((*ip) / overscale) : 255);
		if (!((x+1)%4)) /* write 'em out 4 rows at a time */
		  {
			if (verror((int)fwrite((void *)bgoutbuf, VMSBLOCK, 1,
			 igp->igr_fpbgnd), "write 8-bit data to disk"))
				return EWRITERR;
			cp = bgoutbuf;
		  }
	   }
	(void)fclose(igp->igr_fpbgnd);
	scrhead->format = save_fmt;	/* restore format value */
	return EOK;
}

int bg_from_disk(INTEGRUN *igp)
{ /*	This reads in a background image from disk */
	register int	x, y;
	int		bg_flags, hit, iht, overscale, wid;
	u_short		*uisp;
	u_char		*ucp;
	xgulong		*ip;
	MAIL		*thd;
	REFPARS		*xp;
	VOFLOW		*vofp, *vend;

	bg_flags = igp->igr_imfl & ~(GZIP_AFTER_CLOSE);
	xp = igp->igr_xp;	wid = xp->xdetwid;	hit = xp->xdetht;
/* first, read the 8bit data and the overflows from disk */
	thd = igp->igr_heads;
	if (FRAME_OK != read_image(trname("bgimage"), bg_flags,igp->igr_heads))
	  {
		fprintf(igp->igr_fpl,
	 "Can't read background from disk; new image created instead\n");
		fprintf(stderr,
	 "Can't read background from disk; new image created instead\n");
		return 0;
	  }
/* now put the 8-bit data in the 16-bit slot for it.  We rescale the
	data by a factor we determine from a variable in the header
	of the background file. */
	overscale = thd->count0;	if (!overscale) overscale = 2;
	if ((ucp = thd->data8))
	  {
		for (x = 0; x < wid; x++)
		   {
			ip = igp->igr_upwp + x * hit;
			for (y = 0; y < hit; y++, ip++, ucp++)
				*ip = ((int)(*ucp)) * overscale;
		   }
		if (thd->oflowcount)
		  { /* now reconstitute the overflows */
			vend = thd->overflow + thd->oflowcount;
	        	for (vofp = thd->overflow; vofp < vend; vofp++)
			   {
				iht = hit - vofp->ycoord - 1;
				ip = igp->igr_upwp + vofp->xcoord * hit + iht;
				*ip = vofp->value * overscale;
			   }
		  }
	  }
	else {
		uisp = thd->data16;
		for (x = 0; x < wid; x++)
		   {
			ip = igp->igr_upwp + x * hit;
			for (y = 0; y < hit; y++, ip++, uisp++)
				*ip = ((int)(*uisp)) * overscale;
		   }
	     }
	return 1;
}

static int partition(int fmno, INTEGRUN *igp)
{ /*	This establishes the list of pixels contained within reflections
	and the pixels not contained in reflections in a given image.
	It returns the number of pixels contained in reflections. */
	register int	x, y, fb, jx, dz;
	int		iws, ix, ixs, iy, iys, lx, lx2, ly, ly2, lz, lz2;
	int		nbg, nbigc, nok, sizact, xe, xs, ye, ys;
	u_char		*ac, *acy0, *tbmp, *tby;
	double		uz;
	FILE		*fpbm;
	REFPARS		*xp;
	OBSREF		*lrp;
	REGIONATT	*lbm0, *bmp;
	static u_char	cbo[] = { 254, 253, 251, 247, 239, 223, 191, 127 } ;
	static char	bgpnam[80] = "bgpix000.ams";

	xp = igp->igr_xp;
	/* the active-pixel list was originally read into the background list.
	 We'll chip away at this afterward. */
	ixs = xp->xdetwid;	iys = xp->xdetht;
	acy0 = *activmask(ixs, iys, 1);
	/* count how many active pixels there are */
	nbigc = ((unsigned)ixs) / 8;	if (ixs % 8) nbigc++;
	sizact = xp->xdetht * nbigc;	cpymem(igp->igr_curinbg, acy0, sizact);
	acy0 = igp->igr_curinbg + sizact;
	for (ac = igp->igr_curinbg, nok = 0; ac < acy0; ac++)
	   {
		for (x = 0; x < 8; x++) if ((*ac) & (1 << x)) nok++;
	   }
	/* establish starting and ending reflections to consider. */
	lbm0 = igp->igr_bmask;
	lx = lbm0->ra_bx;	ly = lbm0->ra_by;	lz = lbm0->ra_bz;
	lx2 = lx / 2;	ly2 = ly / 2;	lz2 = lz / 2;
	igp->igr_rfmin = rforder((double)(fmno - igp->igr_zhalfw - 2), igp);
	igp->igr_rfmax = rforder((double)(fmno + igp->igr_zhalfw + 2), igp);
	/* go through those reflections. */
	for (lrp = igp->igr_rfmin; lrp <= igp->igr_rfmax; lrp++)
	   { /* set up the binary model-profile array and use it to
		 zero out some pieces of those reflections.*/
		igp->igr_reflp = lrp;
		bmp = lbm0 + lrp->obs_region;
		get_limits(lrp, PREDICTED, igp);
		ys = igp->igr_ycen - ly2;	if (ys < 0) ys = 0;
		ye = igp->igr_ycen + ly2;
		if (ye >= xp->xdetht) ye = xp->xdetht - 1;
		iys = ys + ly2 - igp->igr_ycen;	iws = iys * lx;
		xs = igp->igr_xcen - lx2;	if (xs < 0) xs = 0;
		ixs = xs + lx2 - igp->igr_xcen;
		xe = igp->igr_xcen + lx2;
		if (xe >= xp->xdetwid) xe = xp->xdetwid - 1;
		acy0 = igp->igr_curinbg + ys * nbigc + xs / 8;
		uz = ((double)(fmno + lz2)) +0.49999 -igp->igr_fphipr; dz = uz;
		if ((0 <= dz) && (dz <= lz))
		 {
		   fb = 1 << dz;
		   if (dz) fb |= (1 << (dz - 1));
		   if (dz < lz) fb |= (1 << (dz + 1));
		   for (iy = iys, y = ys, tby = &(bmp->ra_incl[iws]);
		    y < ye; iy++, y++, tby += lx, acy0 += nbigc)
		     {
			for (ix = ixs, x = xs, ac = acy0, tbmp = tby + ixs,
			 jx = xs % 8; x <= xe; ix++, x++, tbmp++, jx++)
			   {
				if ((*tbmp) & fb) *ac &= cbo[jx];
				if (!((x + 1) % 8))
				  {
					ac++;	jx = -1;
				  }
			   }
		     } /* end of "for (iy" clause */
		 } /* end of "if ((0 <= dz)" clause */
	   } /* end of "for (lrp = igp->igr_rfmin" clause */
	/* count how many background pixels there are */
	for (ac = igp->igr_curinbg, acy0 = igp->igr_curinbg + sizact, nbg = 0;
	 ac < acy0; ac++)
	   {
		for (x = 0; x < 8; x++) if ((*ac) & (1 << x)) nbg++;
	   }
	/* write the mask out if requested */
	if (igp->igr_bgout < 32767)
	  {
		bgpnam[5] = '0' + (fmno / 100);
		bgpnam[6] = '0' + ((fmno / 10) % 10);
		bgpnam[7] = '0' + (fmno % 10);
		if (NULL == (fpbm = xgopen(bgpnam, "w")))
		  {
			fprintf(stderr,
			 "Cannot open %s as output bg mask\n", bgpnam);
			return nok - nbg;
		  }
		for (y = 0, tby = igp->igr_curinbg; y < xp->xdetht;
		 y++, tby += nbigc)
		   if (0 == fwrite((void *)tby, (unsigned)nbigc, 1, fpbm))
		     {
			fprintf(stderr, "Error writing amask row %3d\n", y);
			break;
		     }
		(void)fclose(fpbm);
	  }
	return nok - nbg;
}

#define		XYFLOOR		250
static int planebg(int fmno, double frameco, double bgpa[], INTEGRUN *igp)
{ /*	This uses pixels near a reflection to estimate the background
	in its neighborhood using a least-squares plane
	(Rossman, J.Appl. Crystallogr. (1979) 12:225).
	Input: frame number
	Output: bgpa[0][1][2], the floating-point least-squares parameters
		describing the background-value plane, i.e.
	B(X,Y) = bgpa[0] + (X-Xc) * bgpa[1] + (Y-Yc) * bgpa[2]
	where (Xc,Yc) is the integerized reflection centroid position.
	The routine returns the following values:
 0	plane-background calculation was validly performed.
 1	scanning-angle centroid is too far from `fmno' to be useful
	 in plane-fitting
 2	insufficient background pixels in neighborhood of reflection
	 to allow plane equation to succeed
 3	insufficient background pixels in neighborhood of reflection,
	 once outliers are removed, to allow plane equation to succeed
 4	least-squares normal matrix is singular.
	In all nonzero returns the routine returns reasonably plausible
	values for bgpa[0][1][2] anyway.
 5	background plane isn't flat enough
 6	observed background values too far from plane.  */
	register int	ntot, x, y;
	int		dz, fb, hit, lx, lx2, ly, lz, lz2, nbigc, ndiff;
	int		wid, xe, xnb, xs, ye, ynb, ys;
	u_char		*acy0, *ac;
	xgulong		*ip;
	double		ccp, cphi, delu, deviat, dex, dey;
	double		maxb, meanb, minb, scp, sigb, slope;
	double		usep, ux, uy, uz, w, wb, wx, wy, xce;
	double		bm[3][3], bmi[3][3], lbgpa[3], lqe[10];
	float		lcx, lcy;
	MAIL		*thd;
	REFPARS		*xp;
	OBSREF		*lrp;
	REGIONATT	*lbm0;

	lbm0 = igp->igr_bmask;	lrp = igp->igr_reflp;	xp = igp->igr_xp;
	lx = lbm0->ra_bx;	ly = lbm0->ra_by;	lz = lbm0->ra_bz;
	lx2 = lx / 2;	lz2 = lz / 2;
	/* determine predicted (X,Y,omega) centroid */
	get_limits(lrp, PREDICTED, igp);
	wid = xp->xdetwid;	hit = xp->xdetht;
	x = igp->igr_xcen;
	if (x < lx2) x = lx2; else if (x > wid - lx2 - 1) x = wid - lx2 -1;
	y = igp->igr_ycen;
	if (y < igp->igr_ytopb[x]) y = igp->igr_ytopb[x];
	else if (y > igp->igr_ybotb[x]) y = igp->igr_ybotb[x];
	ip = igp->igr_upwp + x * hit + y;
	bgpa[0] = ((double)(*ip)) * igp->igr_oversc / BGDSCALE;
	bgpa[1] = bgpa[2] = 0.; /* default values. */
	uz = ((double)(fmno + lz2)) + 0.49999 - igp->igr_fphipr; dz = uz;
	if ((0 > dz) || (dz > lz)) return 1;
	/* get limits of summation for background plane calculations */
	xnb = (lx * 2) / 3;	ynb = (ly * 2) / 3;
	ys = igp->igr_ycen - ynb;	if (ys < 1) ys = 1;
	ye = igp->igr_ycen + ynb;
	if (ye > hit - 2) ye = hit - 2;
	xs = igp->igr_xcen - xnb;	if (xs < 1) xs = 1;
	xe = igp->igr_xcen + xnb;
	if (xe > wid - 2) xe = wid - 2;
	if (1 <= pix2cm(igp->igr_xcen, igp->igr_ycen, &lcx, &lcy, 0))
	  {
		dex = lcx;	dey = lcy;
		cphi = atan2(dey, dex);
		if (cphi > PIO2) cphi -= PI;
		else if (cphi < -PIO2) cphi += PI;
		scp = sin(cphi);	ccp = cos(cphi);
		dey = MINDIM + DIMCOEF * ccp;
		dex = MINDIM + DIMCOEF * fabs(scp);
		slope = flimits(tan(cphi), -DIMCOEF, DIMCOEF);
	  }
	else {
		cphi = 0.;	ccp = 1.;	scp = 0.;
		dex = dey = MINDIM + DIMCOEF / 2;	slope = 0.;
	     }
	ys = igp->igr_ycen + 0.49999 - dey;
	ye = ys + 1 + dey + dey;
	if (ys < 1) ys = 1;
	if (ye > hit - 2) ye = hit - 2;
	/* figure out starting value of pointer that tells us whether
	 or not any given pixel is a background pixel. */
	nbigc = wid / 8;	if (wid % 8) nbigc++;
	/* acy0 = igp->igr_curinbg + ys * nbigc + xs / 8; */
	/* zero out the counters and sums */
	zeromem((void *)lqe, 10 * sizeof (double)); usep = 0.;
	/* accumulate sums for the least-squares values */
	thd = igp->igr_heads + (fmno % (igp->igr_nmbox - 1));
	if (igp->igr_flags & IR_VERBOSEANY)
	  {
		fprintf(igp->igr_fpl,
 "Planebg@[%4d,%4d]: cphi,ys,ye,dex,slope = %7.2f,%4d,%4d,%6.2f,%6.3f\n",
	 igp->igr_xcen, igp->igr_ycen, cphi / RADC, ys, ye, dex, slope);
		fprintf(igp->igr_fpl, "Y,xs,xe values:\n");
	  }
	for (y = ys, uy = ys - igp->igr_ycen; y <= ye; y++, uy += 1.)
	   {
		wy = uy * uy;
		if (fabs(slope) < DIMCOEF)
			xce = 0.49999 + (double)igp->igr_xcen + uy * slope;
		else xce = igp->igr_xcen;
		xs = xce - dex;	xe = xs + 1 + dex + dex;
		if (xs < 1) xs = 1;
		if (xe > wid - 2) xe = wid - 2;
		acy0 = igp->igr_curinbg + y * nbigc + xs / 8;
		for (x = xs, ac = acy0, ux = xs - igp->igr_xcen;
		 x <= xe; x++, ux += 1.)
		   {
			fb = 1 << (x % 8);
			if ((*ac) & fb) /* is this a bgnd pixel? */
			  {
				wx = ux * ux;
				w = 1.;
				usep += 1.;
				wb = w * frameco * (double)(imvalue(thd,x,y));
				lqe[0] += w;		lqe[1] += w * ux;
				lqe[2] += w * uy;	lqe[3] += w * wx;
				lqe[4] += w * ux * uy;	lqe[5] += w * wy;
				lqe[6] += wb;		lqe[7] += wb * ux;
				lqe[8] += wb * uy;	lqe[9] += wb * wb / w;
			  }
			if (!((x + 1) % 8)) ac++; /* advance bg-pix pointer */
		   }
		if (igp->igr_flags & IR_VERBOSEANY)
		  {
			fprintf(igp->igr_fpl, "[%4d:%4d->%4d]", y, xs, xe);
			if (!((y+1-ys) % 6)) fprintf(igp->igr_fpl, "\n");
			else fprintf(igp->igr_fpl, " ");
		  }
	   }
	if (igp->igr_flags & IR_VERBOSEANY) fprintf(igp->igr_fpl, "\n");
	/* make sure we have enough points to make this work */
	if ((usep < 6.) || (lqe[0] < DFUZZ)) return 2;
	/* compute mean and standard deviation of count */
	meanb = lqe[6] / lqe[0];
	sigb = sqrt((lqe[9] *lqe[0] - lqe[6] * lqe[6]) / (lqe[0] * usep));
	/* use those to establish limits on plausible bgnd values */
	minb = meanb - 2. * sigb;	maxb = meanb + 2. * sigb;
	/* subtract (from the sums) pixels that deviate too much */
	/* acy0 = igp->igr_curinbg + ys * nbigc + xs / 8; */
	for (y = ys, uy = ys - igp->igr_ycen; y <= ye; y++, uy += 1.)
	   {
		wy = uy * uy;
		if (fabs(slope) < DIMCOEF)
			xce = 0.49999 + (double)igp->igr_xcen + uy * slope;
		else xce = igp->igr_xcen;
		xs = xce - dex;	xe = xs + 1 + dex + dex;
		if (xs < 1) xs = 1;
		if (xe > wid - 2) xe = wid - 2;
		acy0 = igp->igr_curinbg + y * nbigc + xs / 8;
		for (x = xs, ac = acy0, ux = xs - igp->igr_xcen;
		 x <= xe; x++, ux += 1.)
		      {
			fb = 1 << (x % 8);
			wb = frameco * (double)imvalue(thd, x, y);
			if (((*ac) & fb) && ((wb < minb) || (wb > maxb)))
			  {
				wx = ux * ux;
				w = 1.;
				wb *= w;	usep -= 1.;
				lqe[0] -= w;		lqe[1] -= w * ux;
				lqe[2] -= w * uy;	lqe[3] -= w * wx;
				lqe[4] -= w * ux * uy;	lqe[5] -= w * wy;
				lqe[6] -= wb;		lqe[7] -= wb * ux;
				lqe[8] -= wb * uy;	lqe[9] -= wb * wb / w;
			  }
			if (!((x + 1) % 8)) ac++;
		     }
	   }
	/* make sure we still have enough points to make this work */
	if ((usep < 6.) || (lqe[0] < DFUZZ)) return 3;
	/* now compute the least-squares plane parameters */
	bm[0][0] = lqe[0];		bm[0][1] = bm[1][0] = lqe[1];
	bm[0][2] = bm[2][0] = lqe[2];	bm[1][2] = bm[2][1] = lqe[4];
	bm[1][1] = lqe[3]; bm[2][2] = lqe[5];	meanb = lqe[6] / lqe[0];
	sigb = sqrt((lqe[9] *lqe[0] - lqe[6] * lqe[6]) / (lqe[0] * usep));
	if (0 != dinv3(bm, bmi)) return 4;	dmm31(lbgpa, bmi, &(lqe[6]));
	for (x = 0; x < 3; x++) bgpa[x] = lbgpa[x];
	if (bgpa[0] < DFUZZ) return 0;
	/* Analyze the parameters to see if they reflect a reasonably flat
	 background. Until 3 September 05 this calculation was in error. */
	deviat = lbgpa[1] * lbgpa[1] + lbgpa[2] * lbgpa[2];
	if (deviat > lbgpa[0] * lbgpa[0] * NOTFLATENOUGH)
	  {
		if (igp->igr_flags & IR_VERBOSEANY) fprintf(igp->igr_fpl,
	 " [%3d,%3d,%3d]@[%4.0f %4.0f %4.0f]: bg = %12.4f %12.4f %12.4f\n",
		 lrp->obs_hkl[0], lrp->obs_hkl[1], lrp->obs_hkl[2],
		 ((double)lrp->obs_xpred) / igp->igr_xscal,
		 ((double)lrp->obs_ypred) / igp->igr_yscal,
		 lrp->obs_ompred, bgpa[0], bgpa[1], bgpa[2]);
		return 5;
	  }
	/* go back to make sure the local version isn't terribly far away */
	/* subtract (from the sums) pixels that deviate too much */
	/* acy0 = igp->igr_curinbg + ys * nbigc + xs / 8; */
	ndiff = ntot = 0;
	for (y = ys, uy = ys - igp->igr_ycen; y <= ye; y++, uy += 1.)
	   {
		wy = uy * uy;
		acy0 = igp->igr_curinbg + y * nbigc + xs / 8;
		for (x = xs, ac = acy0, ux = xs - igp->igr_xcen;
		 x <= xe; x++, ux += 1.)
		    {
			fb = 1 << (x % 8);
			wb = frameco * (double)imvalue(thd, x, y);
			if (((*ac) & fb) && ((wb < minb) || (wb > maxb)))
			  {
				w = lbgpa[0] + lbgpa[1] * ux + lbgpa[2] * uy;
				delu = w - wb;	ntot++;
				if ((delu > BIGDEVIAT * lbgpa[0]) &&
				 (delu > BIGDEVIAT * wb)) ndiff++;
			  }
			if (!((x + 1) % 8)) ac++;
		     }
	   }
	if (ndiff > ntot * FRACDEVIAT)
	  {
		if (igp->igr_flags & IR_VERBOSEANY)
		 fprintf(igp->igr_fpl,
 " [%3d,%3d,%3d]@[%4.0f %4.0f %4.0f]:bg=%8.1f %8.1f %8.1f;nd,nt=%4d,%4d\n",
		 lrp->obs_hkl[0], lrp->obs_hkl[1], lrp->obs_hkl[2],
		 ((double)lrp->obs_xpred) / igp->igr_xscal,
		 ((double)lrp->obs_ypred) / igp->igr_yscal,
		 lrp->obs_ompred, bgpa[0], bgpa[1], bgpa[2], ndiff, ntot);
		return 6;
	  }
	return 0;
}

static void saveplane(int fmno, double *bpl, INTEGRUN *igp)
{ /*	This stores the result of the least-squares background plane
	calc in a structure contained within the `igp->igr_reflp' structure */
	register int	dz, oz;
	int		lz, lz2;
	double		ab0, ab1, ab2, lb0, lb1, lb2, m0sofar, m1sofar;
	double		m2sofar, thism0, thism1, thism2, newsc, uz;
	BG_PLANE	*bgpl, *bgpq;
	BGREFL		*bgrf;
	OBSREF		*lrp;
	REGIONATT	*lbm0;

	lbm0 = igp->igr_bmask;	lz = lbm0->ra_bz;	lz2 = lz / 2;
	uz = ((double)(fmno + lz2)) + 0.49999 - igp->igr_fphipr; dz = uz;
	if ((dz < 0) || (dz > lz)) return;
	lrp = igp->igr_reflp;	bgrf = (BGREFL *)&(lrp->obs_omob);
	bgpl = (BG_PLANE *)&(bgrf->bg_pl[dz]);
	ab0 = fabs(bpl[0]);	ab1 = fabs(bpl[1]);	ab2 = fabs(bpl[2]);
	/* if we haven't ever gotten a scale factor for these background
	 plane values, set one up */
	if (bgrf->bg_sc < DFUZZ)
	  {
		bgrf->bg_sc = (ab0 < 1.) ? 1.e4 : 1.e4 / ab0;
		bgrf->bgx_sc = (ab1 < 0.005) ? 20000 : 100./ ab1;
		bgrf->bgy_sc = (ab2 < 0.005) ? 20000 : 100./ ab2;
		lb0 = bpl[0] * bgrf->bg_sc + 0.49999;
		lb1 = bpl[1] * ((double)bgrf->bgx_sc) + 0.49999;
		lb2 = bpl[2] * ((double)bgrf->bgy_sc) + 0.49999;
		/* set all values throughout the array (not just <= dz)! */
		for (bgpq = &(bgrf->bg_pl[0]), oz = 0; oz < lz; oz++, bgpq++)
		   {
			bgpq->bgp_p0 = lb0;	bgpq->bgp_p1 = lb1;
			bgpq->bgp_p2 = lb2;
		   }
		return;
	  }
	/* otherwise we see whether the background scale factors already
	 present for this reflection are big enough to accommodate the current
	 image. If not, we rescale everything */
	m0sofar = m1sofar = m2sofar = 0.;
	for (bgpq = &(bgrf->bg_pl[0]), oz = 0; oz < dz; oz++, bgpq++)
	   {
		thism0 = fabs((double)(bgpq->bgp_p0)) /	bgrf->bg_sc;
		if (thism0 > m0sofar) m0sofar = thism0;
		thism1 = fabs((double)(bgpq->bgp_p1)) / ((double)bgrf->bgx_sc);
		if (thism1 > m1sofar) m1sofar = thism1;
		thism2 = fabs((double)(bgpq->bgp_p2)) / ((double)bgrf->bgy_sc);
		if (thism2 > m2sofar) m2sofar = thism2;
	   }
	if (ab0 > m0sofar)
	  {
		newsc = 10000. / ab0;
		for (bgpq = &(bgrf->bg_pl[0]), oz = 0; oz < lz; oz++, bgpq++)
		   {
			thism0 = ((double)(bgpq->bgp_p0)) / bgrf->bg_sc;
			bgpq->bgp_p0 = thism0 * newsc + 0.49999;
		   }
		bgrf->bg_sc = newsc;
	  }
	if (ab1 > m1sofar)
	  {
		newsc = 100. / ab1;
		for (bgpq = &(bgrf->bg_pl[0]), oz = 0; oz < lz; oz++, bgpq++)
		   {
			thism1 = ((double)(bgpq->bgp_p1)) /
				((double)bgrf->bg_sc);
			bgpq->bgp_p1 = thism1 * newsc;
		   }
		bgrf->bgx_sc = newsc;
	  }
	if (ab2 > m2sofar)
	  {
		newsc = 100. / ab2;
		for (bgpq = &(bgrf->bg_pl[0]), oz = 0; oz < lz; oz++, bgpq++)
		   {
			thism2 = ((double)(bgpq->bgp_p2)) /
				((double)bgrf->bg_sc);
			bgpq->bgp_p2 = thism2 * newsc;
		   }
		bgrf->bgy_sc = newsc;
	  }
	thism0 = bpl[0] * bgrf->bg_sc  + 0.49999;
	if ((thism0 < -32764) || (thism0 > 32764))
	  {
		fprintf(igp->igr_fpl,
		 "Warning: [%4d,%4d,%4d] background meanval =  %8.2f\n",
			lrp->obs_hkl[0], lrp->obs_hkl[1], lrp->obs_hkl[2],
			thism0);
		bgpl->bgp_p0 = (thism0 < 0.) ? -32764: 32764;
	  }
	else bgpl->bgp_p0 = thism0;
	thism1 = flimits(bpl[1] * bgrf->bgx_sc + 0.49999, -128., 127.);
	bgpl->bgp_p1 = thism1;
	thism2 = flimits(bpl[2] * bgrf->bgy_sc + 0.49999, -128., 127.);
	bgpl->bgp_p2 = thism2;
}

static void bg_head(FILE *fp, INTEGRUN *igp, int int_interval, int lastbgf)
{ /* header for background operations */
	fprintf(fp, "\tSummary of Background Initialization Effort\n");
	fprintf(fp,
"\tBackground images will be taken from every%s frame between %d and %d\n",
	(abs(int_interval) == 2 ? " other" : " "), igp->igr_zmi, lastbgf);
	fprintf(fp, " frame           frame time reqd for:\n");
	fprintf(fp, "   no.            name     read calcs\n");
}

static void smooth_stripe(u_short xbegin, u_short xend,
	u_short ys, u_short ye, u_short *nchp, INTEGRUN *igp)
{ /*	This computes the mean of all the background values in a region.
	It then finds the mean of all the data below NSTD_DEVS * the mean.
	Finally, it replaces all elements above NSTD_DEVS * the new mean
	with the local average.
	If the detector is of the type in which strong correlations among
	effective pixel sizes exist within columns of data (expressed by
	the "-y" option), we perform stricter checking for flatness within
	a column than over the entire stripe. */
	
	register int	x, y, npts;
	int		colpts, hit;
	int		nfixed, yend, ystart;
	xgulong		bigelt, colmean, colok, imean, maxincol, maxok, thisbg;
	xgulong		*ip, *ip0;
	REFPARS		*xp;

	if (ye <= ys) return;
	nfixed = 0;
	xp = igp->igr_xp;	hit = xp->xdetht;
	/* we apply the columnar check before the box check (see above). */
	if (igp->igr_flags & IR_YCORREL)
	 for (x = xbegin; x <= (int)xend; x++)
	   {
		ystart = ys;	yend = ye;
		if (igp->igr_ytopb[x] > ystart) ystart = igp->igr_ytopb[x];
		if (igp->igr_ybotb[x] < yend) yend = igp->igr_ybotb[x];
		if ((ystart <= (int)ye) && (yend >= (int)ys))
		  { /* find the mean within the column */
			ip = ip0 = igp->igr_upwp + x * hit + ystart;
			for (colmean = colpts = maxincol = 0, y = ystart;
			 y <= yend; y++, ip++)
			   if (NULL != activmask(x, y, 041))
			     {
				colmean += (thisbg = *ip);	colpts++;
				if (thisbg > maxincol) maxincol = thisbg;
			     }
			/* flatten values > <bg> + COL_STD_DEVS * sig<bg> */
			if (colpts > 3) /* don't average unless >= 3 pts */
			  { /* throw out the highest value */
				colmean -= maxincol;	colpts--;
				colmean = 0.49999 +
				 ((double)colmean) / ((double)colpts);
				colok = 0.49999 + (double)colmean +
				 sqrt((double)colmean) * COL_STD_DEVS;
				ip = ip0;
				for (y = ystart; y <= yend; y++, ip++)
				   {
					if (*ip > colok)
					  {
						*ip = colmean;	nfixed++;
					  }
				   }
			  }
		  }
	   }
	/* find the mean and the largest element */
	for (x = xbegin, imean = npts = bigelt = 0; x <= (int)xend; x++)
	   {
		ystart = ys;	yend = ye;
		if (igp->igr_ytopb[x] > ystart) ystart = igp->igr_ytopb[x];
		if (igp->igr_ybotb[x] < yend) yend = igp->igr_ybotb[x];
		ip = igp->igr_upwp + x * hit + ystart;
		if ((ystart <= (int)ye) && (yend >= (int)ys))
		  {
			for (y = ystart; y <= yend; y++, ip++)
			  if (NULL != activmask(x, y, 041))
			    {
				npts++;	imean += (thisbg = *ip);
				if (thisbg > bigelt) bigelt = thisbg;
			    }
		  }
	   }
	imean -= bigelt;	npts--; /* exclude the biggest element */
	if (npts < 2 || imean < 2) return;
	imean = 0.49999 + (double)imean / (double)npts;
	maxok = 0.49999 + imean + sqrt((double)imean) * NSTD_DEVS;
	if (bigelt < maxok) return;	/* if max <= 2 * mean, quit */
	/* now exclude elements above maxok and recompute the mean */
	for (x = xbegin, imean = npts = 0; x <= (int)xend; x++)
	   {
		ystart = ys;	yend = ye;
		if (igp->igr_ytopb[x] > ystart) ystart = igp->igr_ytopb[x];
		if (igp->igr_ybotb[x] < yend) yend = igp->igr_ybotb[x];
		ip = igp->igr_upwp + x * hit + ystart;
		if ((ystart <= (int)ye) && (yend >= (int)ys))
		  {
			for (y = ystart; y <= yend; y++, ip++)
			   {
				if ((NULL != activmask(x, y, 041)) &&
				 ((thisbg = *ip) < maxok))
				  {
					imean += thisbg;	npts++;
				  }
			   }
		  }
	   }
	if (!npts) return;
	imean = 0.49999 + (double)imean / (double)npts;
	maxok = 0.49999 + imean + sqrt((double)imean) * NSTD_DEVS;
	for (x = xbegin; x <= (int)xend; x++)
	   {
		ystart = ys;	yend = ye;
		if (igp->igr_ytopb[x] > ystart) ystart = igp->igr_ytopb[x];
		if (igp->igr_ybotb[x] < yend) yend = igp->igr_ybotb[x];
		ip = igp->igr_upwp + x * hit + ystart;
		if ((ystart <= (int)ye) && (yend >= (int)ys))
		  {
			for (y = ystart; y <= yend; y++, ip++)
			   {
				if ((*ip) > maxok)
				  {
					*ip = imean; nfixed++;
				  }
			   }
		  }
	   }
	if (igp->igr_flags & IR_VERBOSEANY) fprintf(igp->igr_fpl,
	 " %4d %4d %4d %4d %7d %5d%s", (int)xbegin, (int)xend, (int)ys,
	 (int)ye, nfixed, imean, (++(*nchp)% 3) ? "   " : "\n");
}

static int treeind(int *ar, int val)
{ /*	This returns the index (0 to MAXFLIND) of the element of the sorted
	array "ar" just below `val'. */
	register int	j, bot, top;

		/* special cases: before initial element or after last one */
	if (val <= ar[0]) return 0;
	if (val >= ar[MAXFLIND]) return MAXFLIND;
	/* otherwise, step through the list and look for a match */
	for (top = MAXFLIND, bot = 0; bot != (j = (top + bot)/2); )
	   {
		if (val == ar[j]) return j;
		if (val < ar[j]) top = j;
		else		bot = j;
	   }
	if (ar[j] > val || ((j < MAXFLIND) && (val > ar[j+1])))
	  {
		fprintf(stderr, "Cannot fit %d into list\n", val);
		return 0;
	  }
	return j;
}

double losscorr(MAIL *mp, INTEGRUN *igp)
{ /*	Estimate the coincidence (deadtime) losses for an image */
	int		flind, flux, use100ths;
	double		revent, rlate, latefrac, frac, cf;
	static int	flux0[] = {
	      0,   8168,  16013,  23544,  30771,  37702,  44346,  50713,
	  56810,  62645,  68228,  73564,  78663,  83531,  88175,  92602,
	  96820, 100834, 104651, 108278, 111720, 114983, 118073, 120996,
	 123757, 126361, 128813, 131118, 133282, 135309, 137203, 138969,
	 140611, 142134, 143541, 144837, 146026, 147110, 148094, 148982,
	 149776, 150481, 151099, 151633, 152087, 152464, 152766, 152996,
	 153157, 153252, 153283 };
	static int	flux1[] = {
	      0,   3770,   7391,  10867,  14202,  17401,  20467,  23406,
	  26220,  28913,  31490,  33953,  36306,  38553,  40696,  42740,
	  44686,  46539,  48301,  49974,  51563,  53069,  54495,  55844,
	  57118,  58320,  59452,  60516,  61515,  62450,  63324,  64140,
	  64898,  65600,  66250,  66848,  67396,  67897,  68351,  68761,
	  69128,  69453,  69738,  69985,  70194,  70368,  70507,  70613,
	  70688,  70732,  70746 } ;
	static float	fracv[] = {
	0.00000,0.01980, 0.03921, 0.05824, 0.07688, 0.09516, 0.11308, 0.13064,
	0.14786,0.16473, 0.18127, 0.19748, 0.21337, 0.22895, 0.24422, 0.25918,
	0.27385,0.28823, 0.30232, 0.31614, 0.32968, 0.34295, 0.35596, 0.36872,
	0.38122,0.39347, 0.40548, 0.41725, 0.42879, 0.44010, 0.45119, 0.46206,
	0.47271,0.48315, 0.49338, 0.50341, 0.51325, 0.52289, 0.53233, 0.54159,
	0.55067,0.55957, 0.56829, 0.57684, 0.58522, 0.59343, 0.60148, 0.60937,
	0.61711,0.62469, 0.63212 } ;

	if ((!(igp->igr_flags & IR_LOSSCORR)) || !(mp->eventcount)) return 0;
	/* figure out which kind of data we have by looking at the
	"late" fraction.  This number is very small for Argonne and PC data,
	and it's sizeable for PCS data.  Occasionally PCS data will give a
	small value of this number, but when it does the deadtime value will
	be so small that it's not worth worrying about anyway! */
	revent = mp->eventcount;	if (revent < 0.) revent *= -64.;
	rlate = mp->late;		if (rlate < 0.) rlate *= -64.;
	latefrac = rlate / revent;
	if (mp->elap100ths < 1) mp->elap100ths = 1;
	use100ths = (igp->igr_flags & IR_IGNOREEXPOS) ? 100 : mp->elap100ths;
	flux = revent / (0.01 *(double)use100ths)+ 0.4999;
	if (latefrac < 0.02)
	  {
		flind = treeind(flux0, flux);
		if (flind < MAXFLIND)
		  {
			cf = ((double)(flux - flux0[flind])) /
				((double)(flux0[flind + 1] - flux0[flind]));
			frac = cf * fracv[flind + 1] + (1. -cf) * fracv[flind];
		  }
		else	frac = fracv[MAXFLIND];
	  }
	else
	  {
		flind = treeind(flux1, flux);
		if (flind < MAXFLIND)
		  {
			cf = ((double)(flux - flux1[flind])) /
				((double)(flux1[flind + 1] - flux1[flind]));
			frac = cf * fracv[flind + 1] + (1. -cf) * fracv[flind];
		  }
		else	frac = fracv[MAXFLIND];
	  }
	return frac;
}

static void bg_near_refl(double *bpl, u_short *nbg0, INTEGRUN *igp)
{ /* This increments the backgrounds for pixels near a reflection. */
	register int	x, y;
	int		hit, wid, xmi, ymi, xma, yma, xyof;
	double		bp0, bp1, bp2, bgx, bgxy, bgnew;
	xgulong		*bgp;
	u_short		*nbp;
	REFPARS		*xp;

	xp = igp->igr_xp;	hit = xp->xdetwid;	wid = xp->xdetht;
	bp0 = bpl[0];	bp1 = bpl[1];	bp2 = bpl[2];
	xmi = igp->igr_xmi; if (xmi < 1) xmi = 1;
	xma = igp->igr_xma; if (xma > wid - 2) xma = wid - 2;
	ymi = igp->igr_ymi; if (ymi < 1) ymi = 1;
	yma = igp->igr_yma; if (yma > hit - 2) yma = hit - 2;
	bgx = bp0 + bp1 * ((double)(xmi - igp->igr_xcen));
	for (x = xmi; x <= xma; x++, bgx += bp1)
	   {
		xyof = x * hit + ymi;
		bgp = igp->igr_upwp + xyof;	nbp = nbg0 + xyof;
		bgxy = bgx + bp2 * ((double)(igp->igr_ymi - igp->igr_ycen));
		for (y = ymi; y <= yma; y++, bgp++, nbp++, bgxy += bp2)
		   if (NULL != activmask(x, y, 041))
		     {
			if (*bgp > 100000000) printf(
 "Warning: nearing overflows in bg_near_refl at [%4d,%4d]: bgxy = %9.1f\n",
			 x, y, bgxy);
			if (*nbp > 60000) printf(
 "Warning: nearing denominator oflow in bg_near_refl: [%4d,%4d], %d\n",
			 x, y, *nbp);
			bgnew = ((double)(*bgp)) + bgxy + 0.499999;
			*bgp = bgnew;	(*nbp)++;
		     }
	   }
}

int init_bg(INTEGRUN *igp)
{ /* This gets starting values for the 16-bit background image from the
	first range of data.  It does some smoothing of data in vertical
	stripes unless the user invokes "jagged mode" with "integrate -j". */
	register int	fmno, x, y, xbit;
	int		lx, lx2, nbigc, need_repredict, lastbgf;
	xgulong		*bgp;
	unsigned	imsize, xyof;
	u_short		nchanged;
	double		bpla[3], framecorr, usum;
	MAIL		*thd;
	REFPARS		*xp;
	REGIONATT	*lbm0;
	static int	int_interval, nstripes, newbg, istart, iend, dt0, dt1;
	static int	ncounted, nprint, read_result, bg_flags, use100ths;
	static char	*bgname, *shortname;
	static unsigned	stripe_width;
	static u_short	ystart, yend, xend, ys, yeg;
	static u_short	*nbgp, *nbgq;
	static u_char	*ac;
	static double	framescale, thisbg;

	xp = igp->igr_xp;
	lbm0 = igp->igr_bmask;	lx = lbm0->ra_bx;	lx2 = lx / 2;
	int_interval = (igp->igr_zma - igp->igr_zmi > 2 * igp->igr_nbginit) ?
		2 : 1;
	lastbgf = igp->igr_zmi + int_interval * igp->igr_nbginit;
	need_repredict = 0;
	bg_head(igp->igr_fpl, igp, int_interval, lastbgf);
	bg_head(stdout, igp, int_interval, lastbgf);
/* memory for the background COUNTERS, i.e. the array that keeps track
 of how many frames we've updated, is allocated here */
	imsize = xp->xdetwid * xp->xdetht;
	if (NULL == (AVOID *)(nbgp = (u_short *)callop(imsize,
	 sizeof (u_short), "nbgp in init_bg"))) return EALLOC;
/* zero out the memory for the background values */
	zeromem((AVOID *)(igp->igr_upwp), imsize * sizeof (int));
/* in older versions of this code we could swipe some memory from the
 frame-data store to keep the assignment array.  In this version we have
 to glom on to some memory just for this purpose */
	bg_flags = igp->igr_imfl & ~(GZIP_AFTER_CLOSE);
	/* work from back to front */
	istart = lastbgf;	iend = igp->igr_zmi - int_interval;
	int_interval = -int_interval;
/* read in the background-image frames.  Increment the mean background sum
 for all pixels not assigned to reflections in the relevant frame. */
	nbigc = xp->xdetwid / 8;	if (xp->xdetwid % 8) nbigc++;
	igp->igr_oversc = 1.;
	for (fmno = istart; fmno != iend; fmno += int_interval)
	   if (NULL != (bgname = indexname(fmno)))
	     {
		igp->igr_tl0 = NOW;
		(void)partition(fmno, igp);
		/* we read the data image into one of the standard buffers */
		thd = igp->igr_heads + (fmno % (igp->igr_nmbox - 1));
		if (FRAME_OK == (read_result =
		  read_image(bgname, bg_flags, thd)))
		  {
			igp->igr_tl1 = NOW;
			use100ths = (igp->igr_flags & IR_IGNOREEXPOS) ||
			 (thd->elap100ths <= 0) ? 100 : thd->elap100ths;
			framecorr = (igp->igr_avexp100 /(double)use100ths) /
				(1. - losscorr(thd, igp));
			(igp->igr_logst + 1)->fcorr = framecorr;
			framescale = 8. * framecorr;
			/* update background near reflections with the planar
			 background approximations */
			for (igp->igr_reflp = igp->igr_rfmin;
			 igp->igr_reflp <= igp->igr_rfmax; igp->igr_reflp++)
			  if (0 == planebg(fmno, framecorr, bpla, igp))
				bg_near_refl(bpla, nbgp, igp);
			/* everywhere else, add it in directly */
			for (x = 0; x < xp->xdetwid; x++)
			  {
				ystart = igp->igr_ytopb[x];
				yend = igp->igr_ybotb[x];
				ac = igp->igr_curinbg + x / 8 + ystart * nbigc;
				xbit = 1 << (x % 8);
				xyof = x * xp->xdetht + ystart;
				bgp = igp->igr_upwp + xyof;
				nbgq = nbgp + xyof;
				for (y = ystart; y <= (int)yend;
				 y++, bgp++, nbgq++, ac += nbigc)
				   if ((*ac) & xbit)
				     {
					thisbg = imvalue(thd, x,y);
					thisbg *= framescale;
					usum = ((double)(*bgp)) + thisbg;
					if (usum > 1.e9) printf(
 "Warning: nearing overflow at %4d,%4d: usum = %12.5e\n", x, y, usum);
					if ((*nbgq) > 65000) printf(
 "Warning: nearing denominator overflow at %4d,%4d: *nbgq = %d\n",
						x, y, *nbgq);
					newbg = usum;
					*bgp = newbg;	(*nbgq) += 8;
				     }
			  } /* end of loop over x */
			shortname = filepart(bgname, 0);
			dt0 = igp->igr_tl1 - igp->igr_tl0;
			dt1 = NOW - igp->igr_tl1;
			fprintf(stdout, " %5d %15s %7d %5d\n",
				fmno, shortname, dt0, dt1);
			fprintf(igp->igr_fpl , " %5d %15s %7d %5d\n",
				fmno, shortname, dt0, dt1);
		  }
		else {
			(void)image_error(igp->igr_fpl, bgname,read_result, 0);
			(void)image_error(stderr, bgname, read_result, 0);
		     }
	     }
	if (igp->igr_flags & IR_VERBOSEANY)
	  {
		fprintf(igp->igr_fpl, "\tRepresentative Background Pixels%s\n",
		 igp->igr_flags & IR_JAGGED ? ":" : "Before Smoothing:\n");
		for (x = 0; x < 4; x++)
			fprintf(igp->igr_fpl, "    X    Y #pts   <bg>");
		fprintf(igp->igr_fpl, "\n");
	  }
	nprint = 0;
	for (x = 0; x < xp->xdetwid; x++)
	   {
		bgp = igp->igr_upwp + x * xp->xdetht;
		for (y = 0; y < xp->xdetht; y++, bgp++)
		   {
			if (*bgp > 100000000) printf(
			  " Background at [%4d, %4d] = %d\n", x, y, *bgp);
		   }
	   }
	for (x = ncounted = 0; x < xp->xdetwid; x++)
	  {
	    ystart = igp->igr_ytopb[x]; yend = igp->igr_ybotb[x] - 1;
	    xyof = x * xp->xdetht + ystart;
	    bgp = igp->igr_upwp + xyof;	nbgq = nbgp + xyof;
	    for (y = ystart; y <= (int)yend; y++, bgp++, nbgq++)
	      if (*nbgq)
		{
			framecorr = (double)*bgp * BGDSCALE / (double)(*nbgq);
			*bgp = framecorr + 0.49999;
			ncounted++;
			if ((igp->igr_flags & IR_VERBOSEANY) &&
			 (!(x%47)) && (!(y%44)))
			 {
			  fprintf(igp->igr_fpl, " %4d %4d %4d %6.2f",
				x, y, (int)*nbgq, ((double)*bgp) / BGDSCALE);
			  if (!((++nprint)%4)) fprintf(igp->igr_fpl, "\n");
			 }
		}
	   }
/* now go through stripes of data lx pixels wide by a variable height
	and smooth the data within stripes unless we've set the jagged flag*/
	if (!(igp->igr_flags & IR_JAGGED))
	  {
		if (igp->igr_flags & IR_VERBOSEANY)
		  {
			fprintf(igp->igr_fpl,
			 "\n Number of Pixels Smoothed in Each Region:\n");
			for (x = 0; x < 3; x++) fprintf(igp->igr_fpl,
			 " Xmin Xmax Ymin Ymax Npixels Value   ");
			fprintf(igp->igr_fpl,"\n");
		  }
		for (x = nchanged = 0; x < xp->xdetwid - lx2; x += lx)
		   {
			xend = x + lx - 1;
			if (xend > (u_short)(xp->xdetwid-1))
				xend = xp->xdetwid - 1;
			ystart = igp->igr_ytopb[x];
			yend = igp->igr_ybotb[x];
			if ((y = igp->igr_ytopb[xend]) < (int)ystart)
				ystart = y;
			if ((y = igp->igr_ybotb[xend]) > (int)yend) yend = y;
			nstripes = 1 + ((int)(yend + 1 - ystart)) /
					MAX_STRIPE_HEIGHT;
			stripe_width = 1 + ((int)(yend -ystart +1)) / nstripes;
			for (ys = ystart; ys <= yend; ys += stripe_width)
			   {
				yeg = ys + stripe_width;
				if (yeg > (u_short)(xp->xdetht -1))
					yeg = xp->xdetht -1;
				smooth_stripe((u_short)x, xend, ys, yeg,
				 &nchanged, igp);
			   }
		   }
	  }
	if (igp->igr_flags & IR_VERBOSEANY) fprintf(igp->igr_fpl,"\n");
	if (igp->igr_flags & IR_VERBOSE2)
	  {
		fprintf(igp->igr_fpl,
			"\tRepresentative Background Pixels after rescaling\n");
		for (x = 0; x < 4; x++)
			fprintf(igp->igr_fpl, "    X    Y      <bg>");
		fprintf(igp->igr_fpl, "\n");
		nprint = 0;
		for (x = 0; x < xp->xdetwid; x += 47)
		   {
			bgp = igp->igr_upwp + x * xp->xdetht;
			for (y = 0; y <= xp->xdetht; y += 44, bgp += 44)
			  if (NULL != activmask(x, y, 041))
			    {
				fprintf(igp->igr_fpl," %4d %4d %10.2f",
					x, y, ((double)*bgp) / BGDSCALE);
				if (!((++nprint) % 4))
					fprintf(igp->igr_fpl, "\n");
			    }
		   }
	  }
	memzap((AVOID **)&nbgp, "nbgp in init_bg");
	fprintf(igp->igr_fpl,
	 "%5d pixels had at least one background contributor\n", ncounted);
	/* Let's use the extra image for this: */
	thd = igp->igr_heads + (igp->igr_nmbox - 1);
	(void)put_bgimage(thd, igp);	/* write resulting bgnd image out */
		/* if we had to predict on later chunks of the range,
		we now repredict on the initial range */
	if ((need_repredict) && (0 > pred_bunch(igp->igr_zmi - igp->igr_zhalfw,
	 igp->igr_bunch + igp->igr_zwidth, 0, igp))) return EBADVALUE;
	return EOK;
}

void upd_bg(int fmno, INTEGRUN *igp)
{ /*	This performs the update of background values for the current frame.
	Pixels belonging to reflections do not have their backgrounds updated.
	A special calculation is done if fmno == startfm; in that case the
	ratio of the updated background to the current frame is ill-determined,
	and we need to get it before we can use it. */
	register int	x, y;
	int		hit, nbigc, nupdated, use100ths, wid;
	int		xbbin, xbit, ys, ye, xbw, ybw;
	xgulong		*ip;
	u_char		*ac;
	double		currb, diff, diffsq, framecorr, maxdevr, meandeltab;
	double		newbg, newspread, oldbg, origbg, totrat, zcorr16;
	double		bpla[3];
	MAIL		*thd;
	OBSREF		*lrp;
	REFPARS		*xp;
	REGION_BG	*regx, *rgp, *rgend;
	FRAMELOG	*loclogp;
	REGION_BG	entire;

	igp->igr_tl0 = NOW;	/* pointer to file header: */
	xp = igp->igr_xp;	wid = xp->xdetwid;	hit = xp->xdetht;
	thd = igp->igr_heads + (fmno % (igp->igr_nmbox - 1));
		/* ptr to log-info */
	igp->igr_thislog = loclogp = igp->igr_logst +
		(fmno % (igp->igr_nmbox - 1));
	xbw = igp->igr_xbwidth;	ybw = igp->igr_ybwidth;
	/* change as of XENGEN 2.1: deadtime loss correction based on a
	 paralyzable model (cf. G.F. Knoll, _Radiation Detection &
	 Measurement_ (New York: Wiley & Sons, 19..) Ch. 3, pp. 97-98.
	 We use a deadtime of 5.2*10^-6 for PCS data and 2.4*10^-6 for PC and
	 Argonne data, corresponding to maximum count values of 70 and 150 KHz,
	 respectively. The tables above are the lookup tables corresponding to
	 solving for n in the equation n * exp(-nt) = m
	 for given values of m and the 2 `t' values of 2.4 and 5.2*10^-6
	 respectively. */
	loclogp->lossfrac = losscorr(thd, igp);
	use100ths = (igp->igr_flags & IR_IGNOREEXPOS) ?  100 : thd->elap100ths;
	framecorr = (igp->igr_avexp100 / (double)use100ths) /
		(1. - loclogp->lossfrac);
	loclogp->fcorr = framecorr;	/* correction for expos time,deadtime*/
	zcorr16 = framecorr * BGDSCALE;	/* 16 * frame correction */
	newspread = zcorr16 * NSIGNEWBG;
	rgend = igp->igr_region0 + NXBBINS * NYBBINS;
	/* figure out where refls occur in this frame & count pixels belonging
	to refls */
	igp->igr_rfmin = rforder((double)(fmno - igp->igr_zhalfw - 2), igp);
	igp->igr_rfmax = rforder((double)(fmno + igp->igr_zhalfw + 2), igp);
	meandeltab = 0.; /* null zero delb calculation for updated pixels*/
	nupdated = 0;
	for (x = 0, rgp = igp->igr_region0; x < NXBBINS * NYBBINS; x++, rgp++)
		rgp->r_newrat = rgp->r_npix = 0; /* zero region accumulators */
	/* assign each pixel of the image as background or in-a-reflection */
	loclogp->pinrefls = partition(fmno, igp);
	nbigc = wid / 8;	if (wid % 8) nbigc++;
	maxdevr = MAXDEV_RATIO * igp->igr_aduperph * igp->igr_aduperph;
	if (fmno == igp->igr_zmi) /* initialize the "r_ratio" calculation */
	  { /* go through the pixels and find the ratio of the
		 new background to the old background in each region. */
		for (x = 1; x < wid - 1; x++)
		   {
			xbbin = x / xbw;
			regx = igp->igr_region0 + xbbin * NYBBINS;
			ys = igp->igr_ytopb[x];	ye = igp->igr_ybotb[x];
			ac = igp->igr_curinbg + x / 8 + ys * nbigc;
			xbit = 1 << (x % 8);
			ip = igp->igr_upwp + x * hit + ys;
			for (y = ys; y <= ye; y++, ac += nbigc, ip++)
			   {
				if (((*ac) & xbit) && ((oldbg =
				 igp->igr_oversc * (double)(*ip)) > DFUZZ))
				  {
					rgp = regx + y / ybw;
					newbg = zcorr16 *
					 ((double)imvalue(thd, x, y));
					diff = newbg - oldbg;
					diffsq = diff * diff;
					if (diffsq < maxdevr * oldbg ||
					 diffsq < newspread * newbg)
					  {
						(rgp->r_npix)++;
						rgp->r_ratio += newbg / oldbg;
					  } /* end of "if (diffsq <" clause */
				  }	/* end of "if (((*ac)" clause */
			   } /* end of loop over y */
		   } /* end of loop over x */
	/* Now calculate the r_ratio values. We require at least xbw points in
	 a bin or else we don't trust the value, and we set it to one. */
		if (igp->igr_flags & IR_VERBOSEANY)
		  {
			fprintf(igp->igr_fpl,
 "\tRegional ratios of bgnd in first frame to bgnd over starting range:\n");
			for (x = 0; x < 4; x++) fprintf(igp->igr_fpl,
			 "  <x>  <y> <bg0/<bg>>");
			fprintf(igp->igr_fpl, "\n");
		  }
		for (x = 0, rgp = igp->igr_region0; x < NXBBINS; x++)
		 for (y = 0; y < NYBBINS; y++, rgp++)
		   {
			if ((oldbg = rgp->r_npix) > xbw)
			  rgp->r_ratio = flimits(rgp->r_ratio/oldbg, 0.8,1.25);
			else	rgp->r_ratio = 1.;
			rgp->r_npix = 0;	/* to get ready for next step*/
			if (igp->igr_flags & IR_VERBOSEANY)
			  {
				fprintf(igp->igr_fpl, " %4d %4d %10.4f",
				 x * xbw + xbw / 2, y * ybw + ybw / 2,
				 rgp->r_ratio);
				if (!((y + 1) % 4)) fprintf(igp->igr_fpl,"\n");
			  }
		   }
	  } /* end of "if (fmno == igp->igr_zmi)" clause */
	/* now go through the whole array. Pixels that are believed to be
	 in the background region are run through the updating process.
	 All other pixels in the active area are believed to be in reflections
	 and are therefore saved to the shoebox arrays.
	 Aug 95: we handle Mar data with 16-bit overflows specially,
	 by jamming them into the same kind of 3-bit shift as is used for
	 R-Axis data. The old system, under which we would alert the system
	 to the fact that we were doing this, was flawed at best, and it's
	 completely inapplicable to the current implementation in imvalue() */
	for (x = 1; x < wid - 1; x++)
	   {
		xbbin = x / igp->igr_xbwidth; /* defines horizontal bin */
		regx = igp->igr_region0 + xbbin * NYBBINS;
		/* establish active-area limits of y for this x */
		ys = igp->igr_ytopb[x];	ye = igp->igr_ybotb[x];
		ac = igp->igr_curinbg + x / 8 + ys * nbigc;
		xbit = 1 << (x % 8);
		ip = igp->igr_upwp + x * hit + ys;
		for (y = ys; y <= ye; y++, ac += nbigc, ip++)
		   {
			rgp = regx + y / ybw;
			origbg = igp->igr_oversc * (double)(*ip);
			oldbg = origbg * rgp->r_ratio;
			if ((*ac) & xbit) /* pixel should be in background */
			  {
				newbg = ((double)imvalue(thd,x,y)) *zcorr16;
				diff = newbg - oldbg;	diffsq = diff * diff;
				if (diffsq < maxdevr * oldbg
				 || diffsq < newspread * newbg)
				  {
					meandeltab += diff;
					if (oldbg > DFUZZ)
					  {
						(rgp->r_npix)++;
						rgp->r_newrat += newbg / oldbg;
					  }
					else	nupdated++;
					currb = 0.49999 + oldbg +
						diff * igp->igr_newinflu;
				  }
				else	currb = 0.4999 + oldbg;
				*ip = (int)(currb / igp->igr_oversc + 0.49999);
			  }
			else if (NULL != activmask(x, y, 041))
			  { /* rescale background if inside a refl */
				*ip = (int)(oldbg / igp->igr_oversc + 0.49999);
			  }
		   }
	   }
	/* go back through reflections and obtain planar backgrounds.
	 only do these if the reflection has not already been output */
	if (igp->igr_flags & IR_PLANEBGND)
	  for (lrp = igp->igr_rfmin; lrp <= igp->igr_rfmax; lrp++)
	     {
		if (!(lrp->obs_flags & I_ALREADYOUT))
		  {
			igp->igr_reflp = lrp;
			if (planebg(fmno, framecorr, bpla, igp))
				lrp->obs_flags |= I_BADPLANEBG;
			else {
				saveplane(fmno, bpla, igp);
				lrp->obs_flags &= ~I_BADPLANEBG;
			     }
		  }
	     }
	/* now get the frame-specific ratios right. */
	for (totrat = 0., rgp = igp->igr_region0; rgp < rgend; rgp++)
	   {
		nupdated += rgp->r_npix; totrat += rgp->r_newrat;
		if (rgp->r_npix > 0)
		  {
			rgp->r_newrat /= (double)rgp->r_npix;
			rgp->r_ratio = rgp->r_newrat;
		  }
	   }
		/*remember last chosen pixel*/
	entire.r_npix = nupdated;
	if (nupdated)
		entire.r_ratio = entire.r_newrat = totrat / (double)nupdated;
	else	entire.r_ratio = entire.r_newrat = 1.;
	/* if a region didn't register enough points, use the fit
	from the entire detector */
	for (rgp = igp->igr_region0; rgp < rgend; rgp++)
	   if (rgp->r_npix < 1) rgp->r_ratio = entire.r_ratio;
			/* record <delta B> */
	loclogp->meandelb = (nupdated ?
		meandeltab / (BGDSCALE * (double)nupdated) : 0.);
	loclogp->pupdated = nupdated; /* record # of updated pixels */
	igp->igr_tl1 = NOW;	/* record # secs required */
	loclogp->bgsecs = igp->igr_tl1 - igp->igr_tl0;
}

/* end of integbg.c */
