/* integmain.c -- top-level routines for XGEN's integration function

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions; reorganize use of "{"
	 8 Jan 2007: process_image: curino must be nonzero to reuse image
	27 Dec 2006: decision-making for remap_intdistort localized in it
	27 Nov 2006: delay 1st refinement till 2nd group of images.
	23 Jun 2006: include xgimage.h so we can read detector images here
	 5 Jun 2006: intr_integ moved to intsetup so it can be static
	18 Dec 2005: replace static global igs with calls to manageigp
	 5 Oct 2005: housekeeping
	16 Aug 2005: FRAME_NOT_DELETED replaced with FRAME_NOT_COMPRESSED
	 6 Aug 2005: pre_image() takes "preflags" arg;
		delf now local to pre_image() and process_image().
	28 Jun 2005: pre_image(): enable remapping (X,Y) distortions if needed
	13 Jun 2005: orderly_exit(): handle interruptions correctly
	10 May 2005: handle --help correctly at run-time
	21 Apr 2005: orderly_exit(): timed-out images get a special exit code
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	18 Dec 2004: reorder final diagnostic and call to anal_u for neatness
	30 Aug 2004: routines used only here are declared as static; cosmetics;
	 llog->outrefl and llog->outstrong set to zero in process_image()
	30 Jun 2004: add calls to framestats() and plotting routines;
		this enables on-the-fly monitoring of results
	15 Jun 2004: imageflag bit for gzipping is now correctly abbreviated
	25 Mar 2004: routine getigs() returns pointer to INTEGRUN structure;
	 longjmp skipped if runflags&IR_LONGJUMP is zero; most access to
	 igp structures is done with pointer rather than global
	18 Feb 2004: change kill_ftab call so that it doesn't have args
	24 Jan 2004: several static globals replaced by locals
	21 Jan 2004: license message inserted and initial comments reorganized
	 3 Jan 2004: globals replaced by elements of INTEGRUN structures;
		replace shcalloc() & shfree() calls with callop() and memzap()
	13 Dec 2003: return the `retv' value from orderly_exit();
		eliminate unused vbl `result' from integmain()
	22 Nov 2003: replace global variables with members of INTEGRUN struct;
		move intr_integ here so &igs works.
	 8 Nov 2003: move orderly_exit() here; thus jmp_buf stuff is local now
	19 Sep 2003: nout is initialized in out_wmask() no matter what
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	18 Aug 2003: make the jmp_buf an external variable
		(defined in integglob.c)
	16 Aug 2003: add call to elim_rings() inside orderly_exit()
		to get rid of memory allocated for ice-ring calculations
	10 May 2003: read images directly from disk rather than through
		the PIXENT formalism
	17 Apr 2003: rename `frameno' to `fmno' to eliminate confusion
		with MAIL structure element
	 5 Jan 2003: reorganized source-file header;
		in integmain(), don't delete or compress files too early
	15 Nov 2002: extern declarations made here rather than in integext.h
	 1 Nov 2002: use of REGIONPROFILE structures for reflection-box sizes
	25 Mar 2002: removed MSI attachments
	15 Sep 1997: in process_image, don't free memory >= 1 time
		for thishd->overflow, thishd->runheader
	28 Aug 1995: in out_wmask: use xgwrite, not fwrite,
		so it'll work on AXP's
	20 May 1995: reorganize exit from main; return from orderly_exit()
	12 Apr 1995: break integmain() into smaller functions
	 2 Mar 1995: diagnostics to stdout
	 5 May 1994: source file headers, POLYDET->XGEN;
		more external function declarations
	16 Mar 1994: a few variables moved around
	 7 Dec 1993: initial X-GEN version

   Purpose: controlling the Bragg-reflection integration operations:
		reading frames, updating background, calculating intensities,
		updating model profiles, refining on-the-fly.
   Cross references: int*.c

 This is the mainline for the "integrate" program.  It reads in the
 specified frames and calls `upd_bg' to update the background
 image.  Then it calls `getrefls' to get initial estimates of <X>,
 <Y>, <phi>, and Itot for all the refls expected to have completed
 their transit through the Ewald sphere in this frame.  It then calls
 `getrefls' to compute final intensities by shape-fit and simple
 integration and to add to the sums used in updating the crystal and
 detector parameters and the masks.   Once every forty frames it does
 a parameter refinement and obtains a new set of reflections to check.
*************************************************************************/
#define		NEED_TIME	1
#include	"integdefs.h"
#include	"xgenpred.h"
#include	"xgimage.h"
#define		MIN_REFLS	5000

INTEGRUN *manageigp(int opno, size_t nelts)
{ /*	This manages memory for the static global pointer igp.
	For the moment we don't have a memory-freeing option because
	igp has too many allocated pointers to keep track of */
	int		ibox;
	short		itry, **trlp;
	size_t		dearea, deh, desq, dew, dew2;
	REFPARS		*xp;
	REGIONPROFILE	*rgb;
	float		beamx;
	static INTEGRUN	*igp;
	static size_t	oldnref, newnr;
	static char	igpmsg[] = "igp in manageigp";
	static char	rfatemsg[] = "igp->igr_reflfates in manageigp";
	static char	headmsg[] = "igp->igr_heads in manageigp";
	static char	logstmsg[] = "igp->igr_logst in manageigp";
	static char	reg0msg[] = "igp->igr_region0 in manageigp";
	static char	rgbmsg[] = "igp->igr_rgb in manageigp";
	static char	xyopmsg[] = "(igp->igr_rgb)->xyop in manageigp";
	static char	wintvmsg[] = "igp->igr_wintv in manageigp";
	static char	bmaskmsg[] = "igp->igr_bmask in manageigp";
	static char	ytopmsg[] = "igp->igr_ytopb in manageigp";
	static char	curimsg[] = "igp->igr_curinbg in manageigp";
	static char	upwpmsg[] = "igp->igr_upwp in manageigp";
	static char	rfbmsg[] = "igp->igr_rfbase in manageigp";
	static char	trlbmsg[] = "igp->igr_trlb in manageigp";
	static char	trwpmsg[] = "igp->igr_trwp in manageigp";
	static char	boumsg[] = "igp->igr_boundp in manageigp";
	static char	wmaskmsg[] = "igp->igr_wmask in manageigp";
	static char	inclmsg[] = "(igp->igr_bmask)->ra_incl in manageigp";
	static char	xpmsg[] = "igp->igr_xp in manageigp";

	if (!opno)
	  {
		igp = manageigp(076, 0);	memzap((AVOID **)&igp, igpmsg);
		return NULL;
	  }
	if ((!(opno & 01)) && (NULL != igp)) /* even numbers: selective kills*/
	  {
		if (opno & 02) /* memory with completely fixed sizes */
		  {
			memzap((AVOID **)&(igp->igr_logst), logstmsg);
			kill_mail(&(igp->igr_heads), &(igp->igr_nmbox));
			memzap((AVOID **)&(igp->igr_xp), xpmsg);
			manageprw(0);	/* get rid of prediction memory */
			memzap((AVOID **)&(igp->igr_region0), reg0msg);
			memzap((AVOID **)&(igp->igr_reflfates), rfatemsg);
		  }
		if (opno & 04) /* sizes only depend on refl box sizes */
		  {
			if (NULL != igp->igr_bmask) memzap((AVOID **)
			 &((igp->igr_bmask)->ra_incl), inclmsg);
			memzap((AVOID **)&(igp->igr_bmask), bmaskmsg);
			memzap((AVOID **)&(igp->igr_wintv), wintvmsg);
			memzap((AVOID **)&(igp->igr_wmask), wmaskmsg);
			if (NULL != (rgb = igp->igr_rgb))
			  {
				memzap((AVOID **)&(rgb->m3_xyop), xyopmsg);
				memzap((AVOID **)&rgb, rgbmsg);
			  }
		  }
		if (opno & 010) /* memory depends on detector dimensions */
		  {
			memzap((AVOID **)&(igp->igr_upwp), upwpmsg);
			memzap((AVOID **)&(igp->igr_curinbg), curimsg);
			memzap((AVOID **)&(igp->igr_ytopb), ytopmsg);
		  }
		if (opno & 020) /* memory depends on # of reflections */
			memzap((AVOID **)&(igp->igr_rfbase), rfbmsg);
		if (opno & 040) /* memory for ice-rings, if any */
		  {
			memzap((AVOID **)&(igp->igr_boundp), boumsg);
			memzap((AVOID **)&(igp->igr_trwp), trwpmsg);
		  }
		return igp;
	  }
	if (opno == 1) igp = manageigp(0, 0);
	if (opno == 3) return igp;	/* don't get tricky in this case */
	if (NULL == igp)
	  {
		if (NULL == (igp = (INTEGRUN *)callop(1, sizeof (INTEGRUN),
		 igpmsg))) return igp;
	  }
	if (NULL == igp->igr_reflfates)
	  {
		if (NULL == (igp->igr_reflfates = (xglong *)callop(NFATES,
		 sizeof (xglong), rfatemsg))) return manageigp(0, 0);
	  }
	if (NULL == igp->igr_region0)
	  {
		if (NULL == (igp->igr_region0 = (REGION_BG *)callop(NXBBINS *
		 NYBBINS,sizeof (REGION_BG), reg0msg))) return manageigp(0, 0);
	  }
	igp->igr_nmbox = ibox = PFRAMES + 2;
	if (NULL == igp->igr_heads)
	  {
		if (NULL == (AVOID *)(igp->igr_heads = (MAIL *)callop(ibox,
		 sizeof (MAIL), headmsg))) return manageigp(0, 0);
	  }
	if (NULL == igp->igr_logst)
	  {
		if (NULL == (igp->igr_logst = (FRAMELOG *)callop(ibox,
		 sizeof (FRAMELOG), logstmsg))) return manageigp(0, 0);
	  }
	if (NULL == igp->igr_rgb)
	  {
		if (EOK != regionmem(igp)) return manageigp(0, 0);
	  }
	if (opno & 010)	/* detector-size-dependent allocations */
	  {
		dew = nelts & 0xFFFF;	deh = nelts >> 16;
		dew2 = dew * 2;	dearea = dew * deh;
		desq = dew / 8;	if (dew % 8) desq++;	desq *= deh;
		if ((NULL == igp->igr_ytopb) && (NULL == (igp->igr_ytopb =
		 (int *)callop(dew2, sizeof (int), ytopmsg))))
			return manageigp(0, 0);
		if ((NULL == igp->igr_curinbg) && (NULL == (igp->igr_curinbg =
		 (u_char *)mallop(desq, curimsg)))) return manageigp(0, 0);
		if ((NULL == igp->igr_upwp) && (NULL == (igp->igr_upwp =
		 (xgulong *)callop(dearea, sizeof (xgulong), upwpmsg))))
			return manageigp(0, 0);
	  }
	else if (opno & 020) /* refl allocations depend on # of reflections */
	  {
		newnr = nelts; if (newnr < MIN_REFLS) newnr = MIN_REFLS;
		if ((NULL == igp->igr_rfbase) || (newnr > oldnref))
		  {
			igp = manageigp(020, 0); /* get rid of the old one */
			oldnref = newnr;
			if (NULL == (igp->igr_rfbase = (OBSREF *)callop(newnr,
			 sizeof (OBSREF), rfbmsg))) return manageigp(0, 0);
		  }
		else zeromem((AVOID *)(igp->igr_rfbase),
		 newnr * sizeof (OBSREF));
	  }
	else if (opno & 040) /* ice-ring allocations */
	  {
		igp = manageigp(040, 0);
		if (!nelts) return NULL;
		if (NULL == (igp->igr_trlb = (short **)callop(nelts,
		 sizeof (short *), trlbmsg))) return NULL;
		if (NULL == (igp->igr_trwp = (short *)callop(nelts * 3,
		 sizeof (short), trwpmsg))) return NULL;
		for (itry = nelts, trlp = igp->igr_trlb;
		 itry--; trlp++, igp->igr_trwp += 3) *trlp = igp->igr_trwp;
		if ((NULL == igp->igr_boundp) && (NULL != (xp = igp->igr_xp)))
		  {
			beamx = (xp->xdetwid - 1) / 2 +
			 (int)(MEANPIXPERCM * xp->xplated * tan(xp->xswing));
			igp->igr_boundp = make_boundaries(beamx,
				xp->xdetwid, xp->xdetht);
		  }
	  }
	return igp;
}

static int out_wmask(INTEGRUN *igp)
{ /*	This writes out the model profiles in an orderly way. */
	int		maskno, nmodels, nout;
	MASK_3D		*wmp;
	REGIONPROFILE	*rgb, *rgp;

	rgb = igp->igr_rgb;	nmodels = rgb->m3_nmodels;
	for (maskno = 0, wmp = igp->igr_wmask, rgp = rgb;
	 maskno < nmodels; maskno++, wmp++, rgp++) (void)rgptowmp(rgp, wmp);
	if (NULL == (igp->igr_fpmask = xgopen("wmask", "w")))
	  {
		fprintf(stderr,
		 "Warning: cannot open output model-profile file\n");
		fprintf(igp->igr_fpl,
		 "Warning: cannot open output model-profile file\n");
		return EOPENERR;
	  }
	for (maskno = 0, wmp = igp->igr_wmask, nout = 1680;
	 maskno < nmodels; maskno++, wmp++)
	   if (1680 != (nout = xgwrite((void *)wmp, XG_MASK_3D,
		igp->igr_fpmask)))
	     {
		fprintf(stderr,
		 "Warning: 3-D profile output failed in model %2d\n", maskno);
		fprintf(igp->igr_fpl,
		 "Warning: 3-D profile output failed in model %2d\n", maskno);
		break;
	     }
	(void)fclose(igp->igr_fpmask);	igp->igr_fpmask = NULL;
	return (nout == 1680) ? EOK : EWRITERR;
}

int pre_image(int framen, int preflags, INTEGRUN *igp)
{ /*	These are the steps performed before integrating an image.
	Change as of Nov 2006: no refinement on the first delf%bunch. */
	int	delf, result;

	/* do nothing unless delf is an integer multiple of igp->igr_bunch */
	if ((delf = framen - igp->igr_zmi) % igp->igr_bunch) return EOK;
	if ((delf) && (delf > igp->igr_bunch))
	  {
		upd_xtal(framen, igp);	/* refine xtal and detector params */
		/* do a detector-distortion remapping if needed. If it is needed
		 and is successful, redo the xtal and detector parameters */
		if (EOK == remap_intdistort(igp, preflags))
			upd_xtal(framen, igp);
			/* modify masks */
		if (EOK != (result = upd_masks(igp))) return result;
	  }
	if (0 > pred_bunch(framen - igp->igr_zhalfw,
	 igp->igr_bunch + igp->igr_zwidth, 0, igp)) return EBADVALUE;
	if ((!(preflags & 01)) && (!delf))
	  {
		if (NULL != (igp->igr_fpbgnd))
		  {
			if (!bg_from_disk(igp)) igp->igr_fpbgnd = NULL;
		  }
		if ((NULL == (igp->igr_fpbgnd)) &&
		 (EOK != (result = init_bg(igp)))) return result;
	  }
	pr_header_summing(igp->igr_fpl); pr_header_summing(stdout);
	return EOK;
}

int orderly_exit(char *exitstring)
{ /*	This is called at the end of the program under all circumstances.
	The argument is a string printed at exit.  We release a lot of
	dynamic memory in this routine. */
	int		locflags, retv;
	char		rundate[26];
	INTEGRUN	*igp;
	REGIONPROFILE	*rgb;
	static jmp_buf	ljbu;
	static MAIL	*rawhead;
	static char	interruptmsg[] = "due to user-generated interrupt";

	/* restore ordinary system trapping for interrupts and errors */
	signal(SIGINT, SIG_DFL);	signal(SIGFPE, SIG_DFL);
	signal(SIGBUS, SIG_DFL);	signal(SIGSEGV, SIG_DFL);
		/* calling this with the string "help" means that we
		 are really just running integrate --help. */
	if ((NULL != exitstring) && ('\0' != *exitstring) &&
	 !strcmp(exitstring, "help")) return EOK;
	if (NULL == (igp = manageigp(3, 0))) return EALLOC;
	if (igp->igr_flags & IR_ICERINGS) (void)manageice(0);
	rgb = igp->igr_rgb;
	rawhead = igp->igr_heads + igp->igr_nmbox - 1;
	put_bgimage(rawhead, igp);	/* write background image */
	/* final messages */
	igp->igr_tl0 = NOW;	strcpy(rundate, ctime(&(igp->igr_tl0)));
	rundate[24] = '\0';	/* get rid of newline at end */
	if (exitstring && *exitstring)
	  {
		if (igp->igr_fpl != stdout) fprintf(igp->igr_fpl,
		 "integrate exiting %s; %s\n", rundate, exitstring);
		/* treat timeouts differently */
		if (!strcmp(exitstring, "timeout")) retv = ETIMEOUT;
		else retv = EBAD;
	  }
	else
	  {
		if (igp->igr_fpl != stdout)	/* announce completion */
		  {
			fprintf(igp->igr_fpl,
		 "Normal exit from integrate on %s\n", rundate);
			(void)xeflush(igp->igr_fpl); /* ... before anal_u */
		  }
		printf("Normal exit from integrate on %s\n", rundate);
		/* add `anal_u' functionality at this final stage of things */
		(void)analumain(0, NULL);	retv = EOK;
	  }
	if ((NULL != igp->igr_fpl) && (stdout != igp->igr_fpl))
		(void)fclose(igp->igr_fpl);
	igp->igr_fpl = NULL;	locflags = igp->igr_flags;
	cpymem((AVOID *)ljbu, (AVOID *)(igp->igr_jbuf), sizeof (jmp_buf));
	(void)put_bgimage(NULL, NULL);	/* get rid of put_bgimage memory */
	igp = manageigp(0, 0);	/* get rid of most memory */
	/* longjmp allows orderly return to integmain, the mother of this app*/
	if (locflags & IR_LONGJUMP)
	  {
		globalkill();	longjmp(ljbu, 1);
	  }
	(void)integplot(2, 0, stdout, NULL);	/* terminate the plots */
	/* If we got here via a signal catch, we have to exit,
	 because we don't really know how we got here. */
	if ((NULL != exitstring) && ('\0' != *exitstring) &&
	 (!strcmp(exitstring, interruptmsg))) exit(EOK);
	return retv;
}

int process_image(int framen, INTEGRUN *igp)
{ /*	Here we actually manipulate the image with number `framen'. */
	int		delf, headoff, result, workframe;
	char		*framename;
	ino_t		curino;
	MAIL		*thishd, *nexthd;
	FRAMELOG	*llog;

	igp->igr_tl0 = NOW; /* remember what time it is */
		/* get frm name*/
	if (NULL == (framename = indexname(framen)))
		return orderly_exit("requested frame not on list");
	headoff = framen % (igp->igr_nmbox - 1);
	thishd = igp->igr_heads + headoff;
	nexthd = (headoff == igp->igr_nmbox - 1) ?  igp->igr_heads : thishd +1;
	igp->igr_thislog = llog = igp->igr_logst + headoff;
	/* we now read the image in, unless it's already located in
	 the right buffer, at which point we can skip it */
	curino = record_inode(framename);
	if (((int)curino != 0) && (thishd->quadrant == curino))
	  { /* if the right image is already in the right place,
		 we look to see whether we're supposed to compress it.
		 If we are, we do so at this point */
		if ((igp->igr_imfl & GZIP_AFTER_CLOSE) &&
		 (0 != (result = shrinkit(framename))))
		  {
			(void)image_error(igp->igr_fpl , framename, result, 0);
			(void)image_error(stderr, framename, result, 0);
		  }
	  }
	else if (FRAME_OK != (result =
	  read_image(framename, igp->igr_imfl, thishd))) /* read it in */
	  {
		(void)image_error(igp->igr_fpl, framename, result, 0);
		(void)image_error(stderr, framename, result, 0);
		if (result == FRAME_TIMEDOUT)
			return orderly_exit("timeout");
		else if (result != FRAME_NOT_COMPRESSED)
			 return orderly_exit("requested frame unreadable");
	  }
	igp->igr_tl1 = NOW;	llog->readsecs = igp->igr_tl1 - igp->igr_tl0;
	upd_bg(framen, igp);
	if (framen >= igp->igr_zmi + igp->igr_zwidth)
	  { /* go through refls peaking in this frame */
		workframe = framen - igp->igr_zhalfw - 1;
		/* do intensity calcs */
		if (EOK != (result = getrefls(workframe, igp)))
		 return orderly_exit("error in writing reflections");
		framestats(workframe, igp);	/* statistics for this image */
		(void)integplot(1, workframe, igp->igr_fpl, igp);
		(void)integplot(1, workframe, stdout, igp);
	  }
	else if (framen < igp->igr_zmi + igp->igr_zhalfw)
	  {
		llog->sumsecs = llog->outrefl = llog->outstrong = 0;
		llog->predrefl = rforder((double)framen+1.0001, igp) -
			rforder((double)framen - 0.0001, igp) + 1;
		(void)integplot(1, framen, igp->igr_fpl, igp);
		(void)integplot(1, framen, stdout, igp);
	  }
	delf = framen - igp->igr_zmi;
	if ((delf) && (!(delf % igp->igr_bgout)) && framen != igp->igr_zma)
		(void)put_bgimage(nexthd, igp);
	return EOK;
}

int after_images(INTEGRUN *igp)
{ /*	This closes up a run of integration */
	int	fmno, result;

	for (fmno = igp->igr_zma - igp->igr_zhalfw;
	 fmno <= igp->igr_zma; fmno++)
	   {
		(void)integplot(1, fmno, igp->igr_fpl, igp);
		(void)integplot(1, fmno, stdout, igp);
	   }
	if (NULL != igp->igr_fpo)
	  {
		(void)fclose(igp->igr_fpo);	/* close the UREFLS file */
		igp->igr_fpo = NULL;
	  }
		/* summarize final results to logfile and stdout */
	summ_integ(igp->igr_fpl, igp);	summ_integ(stdout, igp);
		/* print 3-D model counts */
	show_masks(igp->igr_fpl, &(igp->igr_wmask[0]), igp->igr_boundp);
	show_masks(stdout, &(igp->igr_wmask[0]), igp->igr_boundp);
	if (EOK != (result = out_wmask(igp))) return result; 
	return orderly_exit((char *)NULL);
}

int integmain(int argc, char *argv[])
{ /*	Mainline for integrating a set of images */
	int		fmno, preflags, origflags, resu;
	static jmp_buf	ljbu;
	INTEGRUN	*igp;

	/* set up for an orderly exit after an interrupt */
	if (setjmp(ljbu)) return EOK;
	if (NULL == (igp = manageigp(7, 0))) return EALLOC;
	cpymem((AVOID *)igp->igr_jbuf, (AVOID *)ljbu, sizeof (jmp_buf));
	igp->igr_flags = IR_LONGJUMP;	/* flag to jump back to this routine */
	/* process args and set up program run */
	if (EOK != (resu = integ_setup(argc, argv, igp)))
	  {
		if (EHELP == resu) return orderly_exit("help");
		return orderly_exit("integration setup failed");
	  }
	/* do a preliminary integration unless forbidden */
	/* don't delete frames on this first pass */
	origflags = igp->igr_imfl;	igp->igr_imfl &= ~GZIP_AFTER_CLOSE;
	for (fmno = igp->igr_zmi; fmno <= igp->igr_initend; fmno++)
	   { /* once per "igp->igr_bunch" frames, re-refine & re-predict */
		if (EOK != pre_image(fmno, 0, igp))
		  return orderly_exit("Problem in setting up image");
		if ((resu = process_image(fmno, igp))) return resu;
	   }
	if (igp->igr_initend >= igp->igr_zmi)
	  {
		if (NULL != igp->igr_fpo)
			(void)fseek(igp->igr_fpo, 0L, 0); /* to restart*/
		preflags = 3;
	  }
	else preflags = 2;
	igp->igr_imfl = origflags;
	/* start plotting this time */
	if (EOK != (fmno = integplot(0, igp->igr_zmi, igp->igr_fpo, igp)))
		return orderly_exit("Plotting problem");
	for (fmno = igp->igr_zmi; fmno <= igp->igr_zma; fmno++)
	   { /* once per "igp->igr_bunch" frames, re-refine & re-predict */
		if (EOK != pre_image(fmno, preflags, igp))
		  return orderly_exit("Problem in setting up image");
		if ((resu = process_image(fmno, igp))) return resu;
	   }
	/* print stats and timings for later frames; final output; shut down */
	return after_images(igp);
}

/* end of integmain.c */
