/* integpred.c -- routines for predicting spot positions during integration

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	28 Nov 2006: update call to ucell in pred_bunch
	 3 Apr 2006: housekeeping in mean_gamma()
	 3 Jan 2006: enforce resolution limits during integration if asked;
		requires existence of po_dminspec, po_dmaxspec elements
	18 Dec 2005: replace prs structure with calls to manageprw
	 5 Oct 2005: housekeeping
	 7 Sep 2005: report Lorentz & close-to-edge states for REFLIST refls
	 8 Aug 2005: regional adjustments (ra_offx, etc.) only if called for
	22 Jun 2005: hkl2d() now declared static
	17 Jan 2005: disp_meangam() rewritten for compactness and cleanliness;
	 it and mean_gamma() use new REGIONATT elements
	 9 Oct 2004: pred_bunch() bugfix: initialize po_rec, xp->xast,xp->xanst
	 6 Sep 2004: major reorg: use ../xglib/xgpredict.c routines and
	 PREDWORK structures; routines intpredini() and pred2ob() written;
	 meangamma() limits are <= igr_rflast; cosmetics
	30 Aug 2004: routines used only here are declared as static; cosmetics
	30 Jun 2004: minor reformmating in disp_meangam to handle X,Y > 1000
	24 Jan 2004: numerous static globals replaced with local variables
	21 Jan 2004: license message inserted and initial comments reorganized
	 5 Jan 2004: gamval() moved higher up in the file
	 3 Jan 2004: globals replaced by elements of INTEGRUN structures
	 7 Dec 2003: re-set lrp to igp->igr_reflp in findref() after
		call to redo_pmem so we don't lose our place in the array;
		explicit declaration of arg igp in llimits()
	22 Nov 2003: replace global variables with members of INTEGRUN struct
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	17 Aug 2003: change management of icerings in findref();
		move isinring() to ../xglib/xgenmisc.c.
	16 Aug 2003: change management of exclusion codes for keephkl
	 5 Jan 2003: reorganized source-file header
	15 Nov 2002: extern declarations made here rather than in integext.h
	 1 Nov 2002: use of REGIONPROFILE structures for reflection-box sizes
		and altered REGIONATT structures for region attributes
	25 Mar 2002: removed MSI attachments
	15 Aug 1997: in pred_bunch on a test run (testonly == 1),
		we must have npts[i] = 0 for all i.
	31 May 1995: "register" -> "register int" throughout
	12 Apr 1995: replace s_char with sichar
	 2 Mar 1995: diagnostics to stdout; prediction memory-check added
	14 Jun 1994: allow for < 7 frames width;
		check reflp memory; Lp calc redone
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version

   Purpose: predicting which reflections will appear within one range
	of data frames.
   Cross references: ../displays/disppred.c

*************************************************************************/
#define		NEED_TIME	1
#include	"integdefs.h"
#include	"xgenpred.h"
#define		SLOP		(0.1 * RADC)	/* omega slop, in radians */
#define		SLOPCONST	PIO2	/* h,k,l slop scale factor */

static void disp_meangam(FILE *fp, INTEGRUN *igp)
{	/* This prints the mean gamma values within the 12 regions to the
	stream `fp' */
	int		nmodels, region;
	REGIONATT	*bmp;

	bmp = igp->igr_bmask;	nmodels = bmp->ra_nmodels;
	if (bmp->ra_nrref)
	  {
		fprintf(fp, "Linear Offsets, RMS Errors, and 3-D Profile");
		fprintf(fp, " Properties in the %2d Regions\n", nmodels);
		fprintf(fp, " Reg #reflecs. -- Offsets in: -- --RMS Errors--");
		fprintf(fp, "  mean  -location- -Gauss.widths-\n");
		fprintf(fp, " ion pred.ref.     X     Y   phi    X    Y  phi");
		fprintf(fp, " gamma   <X>   <Y> sigX sigY sigZ\n");
		for (region = 0; region < nmodels; region++, bmp++)
		   {
			fprintf(fp, "%3d%7d%4d%6.2f%6.2f%6.2f%5.2f%5.2f%5.2f",
			 region, (int)(bmp->ra_nrpred), (int)(bmp->ra_nrref),
			 bmp->ra_offx, bmp->ra_offy, bmp->ra_offphi,
			 bmp->ra_rmx, bmp->ra_rmy, bmp->ra_rmz);
			fprintf(fp, "%6.3f%6.0f%6.0f%5.2f%5.2f%5.2f\n",
			 bmp->ra_meangam, bmp->ra_xc, bmp->ra_yc,
			 bmp->ra_sigx / 1.414,
			 bmp->ra_sigy / 1.414, bmp->ra_sigz / 1.414);
		   }
	  }
	else {
		fprintf(fp, "\tMean predicted reflection widths and (x,y,z)");
		fprintf(fp, " width sigmas in the %2d detector regions:\n",
			NREGIONS);
		for (region = 0; region < 2; region++)
		 fprintf(fp,
		   " Region #refls <gamma>    <X>    <Y>  sigx  sigy  sigz ");
		fprintf(fp,"\n");
		for (region = 0; region < nmodels; region++, bmp++)
		   fprintf(fp," %6d %6d %7.4f %6.1f %6.1f %5.3f %5.3f %5.3f%s",
			region, (int)(bmp->ra_nrpred), bmp->ra_meangam,
			bmp->ra_xc, bmp->ra_yc, bmp->ra_sigx / 1.414,
			bmp->ra_sigy / 1.414, bmp->ra_sigz / 1.414,
			((region+1) % 2 ? " " : "\n"));
	     }
}

static void mean_gamma(INTEGRUN *igp)
{ /*	This computes <gamma> within each of the (currently)
	 12 regions of the detector face */
	register int	nnzr, reg;
	double		meanmean, rnpts;
	REGIONATT	*bmp;
	OBSREF		*lrp;
	static short	defxc[NREGIONS] = { 72, 208, 366,
		32, 96, 168, 248, 327, 439, 72, 208, 366};
	static short	defyc[NREGIONS] = { 90, 80, 90,
		256, 256, 256, 256, 256, 256, 420, 432, 420 };

	/* current system is to refresh the bmp buffers here */
	for (reg = 0, bmp = igp->igr_bmask; reg < NREGIONS; bmp++, reg++)
		bmp->ra_meangam = bmp->ra_xc = bmp->ra_yc = bmp->ra_nrpred = 0;
	/* now add in the new reflections */
	for (lrp = igp->igr_rfbase; lrp <= igp->igr_rflast; lrp++)
	   if (!(lrp->obs_flags & (I_BADPHISHAPE | I_XYEDGE)))
	     {
		reg = lrp->obs_region;	bmp = igp->igr_bmask + reg;
		bmp->ra_meangam += ((double)lrp->obs_gama) * 1.e-3;
		bmp->ra_xc += ((double)lrp->obs_xpred) / igp->igr_xscal;
		bmp->ra_yc += ((double)lrp->obs_ypred) / igp->igr_yscal;
		(bmp->ra_nrpred)++;
	     }
	igp->igr_reflp = lrp;	/* remember the last one we went through */
	for (reg = meanmean = nnzr = 0, bmp = igp->igr_bmask;
	 reg < NREGIONS; bmp++, reg++)
	   if ((rnpts = bmp->ra_nrpred))
	     {
		bmp->ra_meangam /= rnpts;	bmp->ra_xc /= rnpts;
		bmp->ra_yc /= rnpts; meanmean += bmp->ra_meangam; nnzr++;
	     }
	meanmean /= (double)nnzr;
	for (reg = 0, bmp = igp->igr_bmask; reg < NREGIONS; reg++, bmp++)
	   if (!(bmp->ra_nrpred))
	     {
		bmp->ra_meangam = meanmean;
		bmp->ra_xc = defxc[reg];	bmp->ra_yc = defyc[reg];
	     }
	disp_meangam(igp->igr_fpl, igp);
	if (igp->igr_flags & IR_VERBOSEANY) disp_meangam(stdout, igp);
}

static double hkl2d(short shkl[], REFPARS *xp)
{ /*	Return the D spacing associated with the hkl reflection shkl[] */
	int	i;
	double	dout, elt;
	float	*flp;

	for (dout = i = 0, flp = &(xp->xrec[0][0]); i < 3; i++)
	   {
		elt = *flp * ((double)(shkl[0])) + *(flp + 1) *
		((double)(shkl[1])) + *(flp + 2) * ((double)(shkl[2]));
		dout += elt * elt;
	   }
	if (dout < DFUZZ) return 999.;	else return 1. / sqrt(dout);
}

static void pred2ob(OBSREF *lrp, PREDREF *prp, INTEGRUN *igp)
{ /*	This converts a PREDREF structure to an OBSREF structure */
	int		lz;
	REGIONATT	*bmp;

	lrp->obs_ii = lrp->obs_sigma = lrp->obs_iunfit = lrp->obs_sunfit = 0;
	lrp->obs_gof = lrp->obs_xob = lrp->obs_yob = 0;
	lrp->obs_omob = lrp->obs_bgnd = 0;
	hunpack(prp->pindx, lrp->obs_hkl);
	lrp->obs_lpa = 1.e3 * prp->pcor;
	lz = (igp->igr_rgb)->m3_bz;
	lrp->obs_frmint[lz - 1] = lrp->obs_gama = fabs(prp->pgama) * 1.e3;
	lrp->obs_ompred = prp->pphi;
	lrp->obs_region = getregion((int)(prp->pxt), (int)(prp->pyt),
	 /* prp->pphi, */ igp->igr_boundp);
	bmp = igp->igr_bmask + lrp->obs_region;
	if (igp->igr_flags & IR_REGIONADJUST) /* regional adjustments */
	  {
		lrp->obs_ompred += bmp->ra_offphi;
		lrp->obs_xpred = igp->igr_xscal * (prp->pxt + bmp->ra_offx) +
			0.4999;
		lrp->obs_ypred = igp->igr_yscal * (prp->pyt + bmp->ra_offy) +
			0.4999;
	  }
	else {
		lrp->obs_xpred = igp->igr_xscal * prp->pxt + 0.4999;
		lrp->obs_ypred = igp->igr_yscal * prp->pyt + 0.4999;
	     }
	/* negative pcor and pgama values indicate reflections with
	 bad Lorentz factor or coordinates close to the detetor edge */
	lrp->obs_flags = 0;
	if (prp->pgama < 0.) lrp->obs_flags |= I_XYEDGE;
	if (prp->pcor < 0.) lrp->obs_flags |= (I_BADPHISHAPE | I_MOVETOOFAR);
	/* check for ice-rings if the user has requested it */
	if ((igp->igr_flags & IR_ICERINGS) && (isinring(hkl2d(
	 lrp->obs_hkl, igp->igr_xp)))) lrp->obs_flags |= I_XYEDGE;
}

int intpredini(INTEGRUN *igp)
{ /*	A wrapper for `prpar' for the integration context. */
	int		i;
	PREDWORK	*prw;

	if (NULL == (prw = manageprw(3))) return EALLOC;
	prw->po_polfrac = igp->igr_polfrac;
	prw->po_polnormrot = igp->igr_polnormrot;
	prw->po_miny2sq = igp->igr_miny2sq;
	prw->po_dminspec = igp->igr_dminspec;
	prw->po_dmaxspec = igp->igr_dmaxspec;
	if (EOK != (i = prpar(prw))) return i;
	igp->igr_xp = prw->po_xp;	igp->igr_spl = prw->po_spl;
	return EOK;
}

static void predbad(INTEGRUN *igp, char *msg)
{ /*	Report on a predicted reflection in an appropriate way */
	OBSREF	*rfl;

	rfl = igp->igr_reflp;
	fprintf(igp->igr_fpl, "[%3d,%3d,%3d] @ [%6.1f,%6.1f,%6.1f]: %s\n",
	 rfl->obs_hkl[0], rfl->obs_hkl[1], rfl->obs_hkl[2],
	 ((double)rfl->obs_xpred) / igp->igr_xscal,
	 ((double)rfl->obs_ypred) / igp->igr_yscal, rfl->obs_ompred, msg);
}

int pred_bunch(int framemin, int nframes, int testonly, INTEGRUN *igp)
{ /*	Predict reflections over a range beginning at framemin
	and extending over nframes images. Most of the complexity
	is bound up in prloop() and the routines it calls.
	If the argument `testonly' is true on input, the reflections
	predicted aren't sorted or fiddled with after they're found;
	instead, the routine just returns right after the prediction is done.
	This routine returns the number of reflections it predicts. */
	register int	i, j;
	unsigned	nfound;
	PREDREF		*prp;
	REFPARS		*xp;
	OBSREF		*rf0, *rfl;
	PREDWORK	*prw;

	if (NULL == (prw = manageprw(3))) return 0;
	igp->igr_tl0 = NOW;	/* record start of prediction effort */
	if (NULL == (xp = prw->po_xp)) return 0;
	/* first rotate the reciprocal-space matrix to the center of the
	 scanning range.  This is done mostly to assure proper choice of
	 slowest-varying axis in the index-ordering operation which follows. */
	prw->po_mumin = (framemin - igp->igr_zhalfw) * xp->xfrmsize;
	prw->po_mumax = (framemin + nframes + igp->igr_zhalfw + 1) *
		xp->xfrmsize;
	if (prw->po_mumax < prw->po_mumin) /* neg. stepsize treated specially */
	  {
		prw->po_rangecen = prw->po_mumax;
		prw->po_mumax = prw->po_mumin; prw->po_mumin = prw->po_rangecen;
	  }
	/* The prediction needs po_orec */
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		prw->po_orec[i][j] = xp->xrec[i][j];
	(void)ucell(xp->xabc, xp->xang, xp->xast, xp->xanst, xp->xisys);
	/* do the prediction using the PREDREF structures and then sort 'em */
	if (0 >= prloop(prw, 0)) return 0;
	igp->igr_totalrefls = nfound = prw->po_nfound;
	if (nfound > 1) qsort((AVOID *)(prw->po_prp0), nfound,
	 sizeof (PREDREF), (QPE)dphicmp);
	if (NULL == (igp = manageigp(021, nfound))) return 0;
	for (igp->igr_reflp = igp->igr_rfbase, prp = prw->po_prp0, i = 0;
	 i < (int)nfound; i++, (igp->igr_reflp)++, prp++)
	   {
		pred2ob(igp->igr_reflp, prp, igp);
		if (istestrefl(igp))
		  {
			if (prp->pcor < 0) predbad(igp,
			 "Lorentz factor too close to zero");
			else if (prp->pgama < 0) predbad(igp,
			 "Reflection lies outside detector active area");
		  }
	   }
	/* save addr of last predicted refl */
	igp->igr_rfend = igp->igr_rfbase + nfound;
	igp->igr_rflast = rfl = igp->igr_reflp - 1;
	/* revert to defining xp->xrec as at the run start */
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		xp->xrec[i][j] = prw->po_orec[i][j];
	if (testonly) /* change size of prediction tank */
	  {
		prw = manageprw(nfound * (igp->igr_bunch +
			igp->igr_zwidth + 1));	return nfound;
	  }
	igp->igr_tl1 = NOW;	/* record time now */
	rf0 = igp->igr_rfbase;	rfl = igp->igr_rfbase + (nfound - 1);
	fprintf(igp->igr_fpl,
 "Predicting %6d reflections from frame %7.1f to %7.1f required %4d sec\n",
	 (int)(igp->igr_totalrefls), rf0->obs_ompred,
	 rfl->obs_ompred, (int)(igp->igr_tl1 - igp->igr_tl0));
	fprintf(stdout,
 "Predicting %6d reflections from frame %7.1f to %7.1f required %4d sec\n",
	 (int)(igp->igr_totalrefls), rf0->obs_ompred,
	 rfl->obs_ompred, (int)(igp->igr_tl1 - igp->igr_tl0));
	mean_gamma(igp);	return nfound;
}

/* end of integpred.c */
