/* integutil.c -- miscellaneous utilities used in integration in XGEN

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	21 Apr 2007: ellmom: allow for negative-definite wma matrix,
		and fix bug in use of MAXWID parameter
	19 Apr 2007: float->double conversions; reorganize use of "{"
	28 Nov 2006: use condpr to do neat printouts in integplot
	 5 Jun 2006: update documentation
	29 Nov 2005: redo init3dvec so it really works
	 5 Oct 2005: housekeeping
	16 Aug 2005: MISSING_PIXEL concept is gone; eliminate references to it
	15 Aug 2005: pr_header_summing() and sum_ifrm() always are verbose
	12 Aug 2005: elim. old formattings in pr_header_summing and sum_ifrm;
		print <delta bgnd> as %6.2f,not %6.3f
	22 Jun 2005: adj_model(), ellmom() now static
	13 Jun 2005: extend temporary-verbose flag to include more printing;
		background done with single pointer
	 7 Jun 2005: in ev2lp(), calculate evzz and return it;
		in varlp, calculate lx0 and lx1 as floats, not ints.
	 7 Sep 2004: bugfix in ellmom(): if (*vvp < FUZZ) thisv = 0; cosmetics
	 1 Sep 2004: routines used only here are declared as static
	18 Aug 2004: count of verbosely-printed refls is now igp->igr_nrverbose
	12 Aug 2004: 6 routines moved here from intrefl.c
	 2 Jul 2004: only 3 RMS error plots, not 4
	29 Jun 2004: plotting capabilities added: startiplots(), finishiplots()
	21 Jan 2004: license message inserted and initial comments reorganized
	 3 Jan 2004: globals replaced by elements of INTEGRUN structures;
		some routines require additional arguments as a result
	13 Dec 2003: eliminate unused variable lz from disp_bglp()
	22 Nov 2003: replace global variables with members of INTEGRUN struct
	 8 Nov 2003: move orderly_exit() to integmain.c to localize use
		of jmp_buf stuff; remove declarations used only there.
	29 Oct 2003: move some routines from intrefl.c here for load-balancing
	20 Sep 2003: wnbr 2-D array initialized using double braces (ansi C)
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Sep 2003: replace access() and fopen() call for REFLIST with an
		xgopen() call; that's cleaner and allows for gunzipping files
	12 Jun 2003: bugfix sum_ifrm() fixed nonsense with respect to displays
	17 Apr 2003: rename `frameno' to `fmno'
	 5 Jan 2003: reorganized source-file header
	15 Nov 2002: extern declarations made here rather than in integext.h
	 1 Nov 2002: use of REGIONPROFILE structures for reflection-box sizes
		and altered REGIONATT structures for region attributes
	25 Mar 2002: removed MSI attachments
	 8 Aug 1995: in ellmom(), use product of error functions to calc tails.
	31 May 1995: "register" -> "register int" throughout
	12 Apr 1995: replace s_char with sichar
	 2 Mar 1995: get rid of references to BLOTREF structures
	 2 Mar 1995: diagnostics to stdout; BLOTREF stuff is unused now
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 1 Dec 1994: diagnostic output allows for detector (X,Y) > 999
	 2 Sep 1994: changed frame number by 0.5 frames in pre-centering
	14 Jun 1994: no longer deallocating big arrays
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: a few variables moved around
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version

   Purpose: utilities for diagnostic I/O, exit flow, etc.
	for Bragg-data integration in XGEN
   Cross references:

*************************************************************************/
#include	"integdefs.h"
#define		RESCALER	62.	/* rescale printouts of profiles */
#define		MAXERFARG	(4.)	/* erf(x) for x > MAXERFARG is 1 */
#define		NNEIGHBORS	5	/* max # of models to be averaged */
#define		ALPHAX		(0.05)
#define		ALPHAY		(0.025)
#define		LN2		0.69314718	/* natural log of two */
#define		SIGRAT		(0.5)	/* -SIGRAT*sig=min count for moments */
#define		PICUBED		(41.00627668)	/* Pi cubed */
#define		MAXWID		(5.)	/* nominal width */
#define		NSTD_SUM	(0.8)	/* 0.8sigma slop in profile redefine */
#define		NPLOTS		3	/* number of plots created */

void summ_integ(FILE *fp, INTEGRUN *igp)
{ /*	This summarizes the results of the run of `integrate'
	to the file pointer fp.  The "fates" counts exclude reflections
	near the beginnings and ends of prediction ranges (I_RANGEEDGE == 8)
	because those are evanescent errors. */
	int	i, multiple_prob;

	if ((fp != stderr) && (fp != stdout)) fprintf(fp,"\f");
	fprintf(fp,"\t\tFates of located reflections:\n");
	multiple_prob =
	 igp->igr_reflfates[3] + igp->igr_reflfates[5] + igp->igr_reflfates[7];
	for (i = 17; i < NFATES; i++)
	   if (!(i & I_RANGEEDGE) && (i != 32 && i != 64 && i != 128))
		multiple_prob += igp->igr_reflfates[i];
	if (igp->igr_reflfates[I_BADMODEL])
	  fprintf(fp, "Refls with pathological model profiles:   %6d\n",
		(int)(igp->igr_reflfates[I_BADMODEL]));
	if (igp->igr_reflfates[I_XYEDGE])
	  fprintf(fp, "Refls close to (X,Y) borders of chamber:  %6d\n",
		(int)(igp->igr_reflfates[I_XYEDGE]));
	if (igp->igr_reflfates[I_BADPHISHAPE])
	  fprintf(fp, "Refls too broad or twinned in phi:        %6d\n",
		(int)(igp->igr_reflfates[I_BADPHISHAPE]));
	if (igp->igr_reflfates[I_OVERLAP])
	  fprintf(fp, "Refls significantly overlapping others:   %6d\n",
		(int)(igp->igr_reflfates[I_OVERLAP]));
	if (igp->igr_reflfates[I_PEAKONEDGE])
	  fprintf(fp, "Refls with their peaks on the range edge: %6d\n",
		(int)(igp->igr_reflfates[I_PEAKONEDGE]));
	if (igp->igr_reflfates[I_MOVETOOFAR])
	  fprintf(fp, "Refls whose centroids move too far:       %6d\n",
		(int)(igp->igr_reflfates[I_MOVETOOFAR]));
	if (igp->igr_reflfates[0400])
	  fprintf(fp, "Refls with >= %3d overflowed pixels:      %6d\n",
		TOO_MANY_UNRELIABLE, (int)(igp->igr_reflfates[0400]));
	if (multiple_prob)
	  fprintf(fp, "Refls with more than one problem:         %6d\n",
		multiple_prob);
	if (igp->igr_reflfates[I_USABLE])
	  fprintf(fp, "Well-behaved refls with I > %5.1f sigma:  %6d\n",
		igp->igr_iosmin, (int)(igp->igr_reflfates[I_USABLE]));
	if (igp->igr_reflfates[I_TOOWEAK])
	  fprintf(fp, "Well-behaved refls with I <=%5.1f sigma:  %6d\n",
		igp->igr_iosmin, (int)(igp->igr_reflfates[I_TOOWEAK]));
	return;
}

int integplot(int opno, int fmno, FILE *fp, INTEGRUN *igp)
{ /* Handler for plotting from integrate */
	int		errt, headoff;
	double		revent;
	float		plodep[4];
	MAIL		*hp;
	FRAMELOG	*frp;
	static double	bsig0, d20, int0, ios0, iosz0;
	static FILE	*fpiplo[NPLOTS];	/* file pointers for plots */
	static int	latercall = 0;
	static char	*nrefleg[] =
	 { "predicted", "output", "strong", "weak" };
	static char	*qualleg[] =
	 { "Intensity", "I/sigma", "Bsig,A^2", "D2,A" };
	static char	*rmseleg[] =
	 { "Err(X),pixels", "Err(Y),pixels", "Err(z),images"};

	if (!opno)	/* initialize the plots */
	  {
		if (NULL == (fpiplo[0] = plotsetup(4, 0, &errt, "inrefls.out",
		 "Number of reflections", "Image number", "# of reflections",
		 nrefleg))) return errt;
		if (NULL == (fpiplo[1] = plotsetup(4, 0, &errt, "iquality.out",
		 "Normalized values", "Image number", "Normalized property",
		 qualleg))) return errt;
		if (NULL == (fpiplo[2] = plotsetup(3, 0, &errt, "irmserrs.out",
		 "Root-Mean-Squared Errors", "Image number", "RMS Errors",
		 rmseleg))) return errt;
		return EOK;
	  }
	else if (opno == 1) /* summarize results of image fmno to fp and plots*/
	  {
		headoff = fmno % (igp->igr_nmbox - 1);
		hp = igp->igr_heads + headoff; 	frp = igp->igr_logst + headoff;
		revent = hp->eventcount;	if (revent < 0.) revent *= -64.;
		if (revent < 9.99e8)
		  fprintf(fp, "%4d %16s %9.0f %5.2f %8d %8d %5d %5d %5d",
			fmno, hp->filename, revent, frp->lossfrac * 100.,
			frp->pinrefls, frp->pupdated, (int)frp->predrefl,
			(int)frp->outrefl, (int)frp->outstrong);
		else fprintf(fp, "%4d %16s %9.3e %5.2f %8d %8d %5d %5d %5d",
			fmno, hp->filename, revent, frp->lossfrac * 100.,
			frp->pinrefls, frp->pupdated, (int)frp->predrefl,
			(int)frp->outrefl, (int)frp->outstrong);
		if (fp != stdout)
		  {
			plodep[0] = frp->predrefl; plodep[1] = frp->outrefl;
			plodep[2] = frp->outstrong;
			plodep[3] = frp->outrefl - frp->outstrong;
			(void)plotadd(fpiplo[0], 4, (float)fmno, plodep);
		  }
		fprintf(fp, " %5d %6.2f", (int)(frp->maskupd), frp->meandelb);
		if (frp->outrefl > 0) /* <I>, <I/s>... output if outrefl > 0 */
		  {
			fprintf(fp, " %7.1f %6.2f", frp->iis_in, frp->iis_ios);
			condpr(fp, frp->iis_i000, 4, 1, 1);
			condpr(fp, frp->iis_b, 4, 1, 1);
			condpr(fp, frp->iis_d2, 4, 2, -2);
			if (!latercall)	/* first time: renormalization */
			  {
				latercall = 1;		d20 = frp->iis_d2;
				int0 = frp->iis_in;	ios0 = frp->iis_ios;
				iosz0 = frp->iis_i000;	bsig0 = frp->iis_b;
			  }
			else if ((d20 < DFUZZ) && (frp->iis_d2 > DFUZZ))
				d20 = frp->iis_d2;
			if (fp != stdout)
			  {
				plodep[0] = (int0 > DFUZZ) ?
					frp->iis_in  / int0 : 1.;
				plodep[1] = (ios0 > DFUZZ) ?
					frp->iis_ios / ios0 : 1.;
				plodep[2] = (bsig0 > DFUZZ) ?
					frp->iis_b / bsig0 : 1.;
				plodep[3] = (frp->iis_d2 > DFUZZ) &&
				 (d20 > FUZZ) ?  frp->iis_d2 / d20 : 1.;
				(void)plotadd(fpiplo[1], 4,
					(float)fmno, plodep);
			  }
		  }
		/* summarize RMS errors to fp */
		if (frp->outstrong)
		  {
			condpr(fp, frp->iis_rmx, 4, 1, 1);
			condpr(fp, frp->iis_rmy, 4, 1, 1);
			condpr(fp, frp->iis_rmz, 4, 1, 1);
		  }
		fprintf(fp, "\n");
		if (fp != stdout)
		  {
			plodep[0] = frp->iis_rmx; plodep[1] = frp->iis_rmy;
			plodep[2] = frp->iis_rmz;
			(void)plotadd(fpiplo[2], 3, (float)fmno, plodep);
		  }
		return EOK;
	  }
	/* if we get this far, we're closing out the plots from `integrate' */
	(void)plotfinish(fpiplo[0], -1000.);
	(void)plotfinish(fpiplo[1], -1000.);
	(void)plotfinish(fpiplo[2], -1000.);
	return EOK;
}

void pr_header_summing(FILE *fp)
{ /*	This prints a suitable header for the summing steps of the program */
	fprintf(fp, " frm            frame       counts ");
	fprintf(fp, "    number of pixels  # of reflections");
	fprintf(fp,
	 " masks <delta  <Inten  <I/s> I/sg Bsig  D2,  RMS Errors in\n");
	fprintf(fp, " no.             name     found %%lost");
	fprintf(fp, " in refls bgupdatd  pred outpt   big");
	fprintf(fp,
	 " updat  bgnd>   sity>        @s=0  A^2 Ang.    X    Y    Z\n");
}

OBSREF *rforder(double zphi, INTEGRUN *igp)
{ /*	This returns a pointer to the OBSREF structure whose phi value
	is closest to the argument.  It returns `igp->igr_rfbase' if the
	first phi value on the list is later than the specified value;
	it returns `rflast' at the other end. */
	register int	bot, j, top;
	OBSREF		*rf0, *rfe;

	rf0 = igp->igr_rfbase;	if (zphi < rf0->obs_ompred) return rf0;
	top = igp->igr_totalrefls - 1;	rfe = rf0 + top;
	if (zphi > rfe->obs_ompred) return rfe;
	for (bot = 0; bot != (j = (top + bot) / 2); )
	   {
		if (zphi < (rf0 + j)->obs_ompred) top = j;
		else bot = j;
	   }
	return rf0 + j;
}

double ev2lp(double evp[], INTEGRUN *igp)
{ /*	This calculates the Lorenz-polarization factor */
	double	ev1, ev12, ev2, ev22;
	double	evzz;

	ev1 = evp[1]; ev2 = evp[2]; ev12 = ev1 * ev1; ev22 = ev2 * ev2;
	evzz = fabs(ev1) / (1. - ev12 * igp->igr_e1c - ev22 * igp->igr_e2c -
		fabs(ev1 * ev2 * igp->igr_exc));
	return evzz;
}

void note_badmodel(FILE *fp, double *devn, INTEGRUN *igp)
{ /*	This prints a message when we're trying to update a model
	in a way that would leave it miscentered. */
	OBSREF	*lrp;

	lrp = igp->igr_reflp;
	fprintf(fp, "Note: Miscentered model %2d: centroid off ",
		lrp->obs_region);
	fprintf(fp, "in [X,Y,omega] by [%5.1lf,%5.1lf,%5.1lf]\n",
		devn[0], devn[1], devn[2]);
	fprintf(fp, "after [%3d,%3d,%3d]@[%6.1lf,%6.1lf,%5.1lf],",
		lrp->obs_hkl[0], lrp->obs_hkl[1], lrp->obs_hkl[2],
		((double)lrp->obs_xob) / igp->igr_xscal,
		((double)lrp->obs_yob) / igp->igr_yscal, lrp->obs_omob);
	fprintf(fp, " I = %7.1lf,s=%7.2lf\n",
		lrp->obs_ii, sqrt(lrp->obs_sigma));
}

void dispprofile(int datacode, int jzmin, double scaleval,
		 float	*vec, INTEGRUN *igp)
{ /*	Displays a 3-D profile efficiently */
	register int	ix, iy, iz;
	int		lx, ly, lz, mx, my, mz, nx, nxy, ny, nz, ux, wvec;
	sichar		cval;
	char		dval, *cp;
	double		uscale, thisf;
	float		*zv, *zvp, *zvy, *zvz;
	char		xbu[17];
	static char	*labl[3] = { "background-subtracted counts",
			"model values", "variance values" } ;
	REGIONPROFILE	*rgb;

	if (fabs(scaleval) < DFUZZ) return;
	uscale = RESCALER / scaleval;
	rgb = igp->igr_rgb;
	lx = rgb->m3_bx;	ly = rgb->m3_by;	lz = rgb->m3_bz;
	mx = rgb->m3_qx;	my = rgb->m3_qy;	mz = rgb->m3_qz;
	zv = vec;	wvec = datacode >> 1;
	if (datacode & 01)
	  {
		nx = mx;	ny = my;	nz = mz;
		fprintf(igp->igr_fpl,
		 "Kabsch-transformed profile of %s; max value = %7.3f;",
			labl[wvec], uscale);
		fprintf(igp->igr_fpl, " alpha across; beta down\n   ");
		for (iz = 0; iz < nz; iz++)
			fprintf(igp->igr_fpl, ".___gamma=%1d___", iz);
	  }
	else {
		nx = lx;	ny = ly;	nz = lz;
		fprintf(igp->igr_fpl,
		 "(X,Y,omega) profile of %s; max value = %7.3f;",
		 labl[wvec], uscale);
		fprintf(igp->igr_fpl, " X across; Y down\n     ");
		for (iz = 0; iz < nz; iz++)
			fprintf(igp->igr_fpl, ".__frame %3d____", iz + jzmin);
	  }
	fputs(".\n", igp->igr_fpl);
	nxy = nx * ny;
	for (iy = 0, zvy = zv; iy < ny; iy++, zvy += nx)
	   {
		fputc((int)' ', igp->igr_fpl);
		if (datacode & 01)
		  {
			fputc((int)' ', igp->igr_fpl);
			if (!(iy % 2)) fprintf(igp->igr_fpl, "%1x", iy);
			else fputc((int)' ', igp->igr_fpl);
		  }
		else if (!((iy + igp->igr_ymi) % 5))
			fprintf(igp->igr_fpl, "%4d", iy + igp->igr_ymi);
		else fputs("    ", igp->igr_fpl);
		for (iz = 0, zvz = zvy; iz < nz; iz++, zvz += nxy)
		   {
			fputc((int)'|', igp->igr_fpl);
			for (ix = 0, zvp = zvz; ix < nx; ix++, zvp++)
			   {
				thisf = *zvp * scaleval;
				cval = thisf + 0.49999;
				if (thisf < -20.) dval = '<';
				else if (thisf < -5.) dval = '!';
				else if (thisf < -0.5) dval = '-';
				else if (thisf <= 0.4) dval = ' ';
				else if (thisf < 1.) dval = '.';
				else if (thisf >= RESCALER) dval = '*';
				else if (1 <= cval && cval <= 9)
					dval = cval + '0';
				else if (10 <= cval && cval <= 35)
					dval = (cval - 10) + 'a';
				else if (36 <= cval && cval <= 61)
					dval = (cval - 36) + 'A';
				else dval = '*';
				fputc((int)dval, igp->igr_fpl);
			   }
		   }
		fputs("|\n", igp->igr_fpl);
	   }
	fputs("   ", igp->igr_fpl);
	if (datacode & 01)
		 for (iz = 0; iz < nz; iz++)
			fputs(".0_2_4_6_8_a_c", igp->igr_fpl);
	else {
		fputs("  ", igp->igr_fpl);
		for (iz = 0; iz < nz; iz++)
		   {
			zeromem((void *)xbu, 17);	xbu[0] = '.';
			ux = igp->igr_xcen % 5;
			sprintf(&(xbu[5 - ux]), "%4d %4d",
				(int)(igp->igr_xcen - ux),
				(int)(igp->igr_xcen + 5 - ux));
			if (ux > 2) sprintf(&(xbu[15 - ux]), "%4d",
				(int)(igp->igr_xcen + 10 - ux));
			for (ix = 1, cp = &(xbu[1]); ix < 16; cp++, ix++)
				if (('0' > *cp) || (*cp > '9')) *cp = '_';
			fputs(xbu, igp->igr_fpl);
		   }
	     }
	fputs(".\n", igp->igr_fpl);
}

void disp_bglp(INTEGRUN *igp)
{ /*	Print a display of the background and Lorentz/polarization values */
	register int	ix, iy, x, y;
	int		dellp, lx, ly;
	xgulong		*ip;
	double		lpcen, lpmax, lpmin, mean_bgnd, thisf, thlp, ev[3];
	REFPARS		*xp;
	REGIONPROFILE	*rgb;

	if (EOK != ixy2ev(igp->igr_xcen, igp->igr_ycen, ev)) return;
	lpcen = ev2lp(ev, igp);	xp = igp->igr_xp;
	rgb = igp->igr_rgb;	lx = rgb->m3_bx;	ly = rgb->m3_by;
	for (x = igp->igr_xmi, mean_bgnd = ix = 0; ix < lx; ix++, x++)
	   {
		ip = igp->igr_upwp + x * xp->xdetht + igp->igr_ymi;
		for (y = igp->igr_ymi, iy = 0;  iy < ly; iy++, y++, ip++)
			mean_bgnd += (double)(*ip);
	   }
	mean_bgnd /= (double)(lx * ly);
	fprintf(igp->igr_fpl,
	 "Background values in  neighborhood: mean = %7.2f\t", mean_bgnd);
	lpmin = lpmax = lpcen;
	for (iy = 0, y = igp->igr_ymi; iy < ly; iy++, y++)
	   {
		for (ix = 0, x = igp->igr_xmi; ix < lx; ix++, x++)
		   {
			if (EOK == ixy2ev(x, y, ev)) thlp = ev2lp(ev, igp);
			else thlp = lpcen;
			if (thlp < lpmin) lpmin = thlp;
			if (thlp > lpmax) lpmax = thlp;
		   }
	   }
	fprintf(igp->igr_fpl, "\t\tDelta LP, with LPmin,max = %6.4f,%6.4f\n",
		lpmin, lpmax);
	fprintf(igp->igr_fpl,
	 "Each displayed value represents delta B = %6.1f\t\t",
		mean_bgnd / 20.);
	fprintf(igp->igr_fpl, "Displayed val is (lp-lpcen) * 100/lpcen");
	fprintf(igp->igr_fpl, " for lpcen = %6.4f\n   ", lpcen);
	for (ix = 0, x = igp->igr_xmi; ix < lx; ix++, x++)
	   {
		if (x % 5) fprintf(igp->igr_fpl, "   |");
		else fprintf(igp->igr_fpl, "%4d", x);
	   }
	fprintf(igp->igr_fpl, "     ");
	for (ix = 0, x = igp->igr_xmi; ix < lx; ix++, x++)
	   {
		if (x % 5) fprintf(igp->igr_fpl, "   |");
		else fprintf(igp->igr_fpl, "%4d", x);
	   }
	fprintf(igp->igr_fpl,"\n");
	for (iy = 0, y = igp->igr_ymi; iy < ly; iy++, y++)
	   {
		ip = igp->igr_upwp + igp->igr_xmi * xp->xdetht + y;
		if (iy % 5) fprintf(igp->igr_fpl, "_  _");
		else fprintf(igp->igr_fpl, "%4d", y);
		if (mean_bgnd > DFUZZ)
		  {
		   for (ix = 0, x = igp->igr_xmi; ix < lx;
		     ix++, x++, ip += xp->xdetht)
		     {
			thisf = ((((double)(*ip)) / mean_bgnd) - 1.) * 20.;
			if (thisf < -99.) thisf = -99.;
			if (fabs(thisf) < 1.) fprintf(igp->igr_fpl, "    ");
			else fprintf(igp->igr_fpl, "%4d", (int)thisf);
		     }
		  }
		else
		  {
			for (ix = 0; ix < lx; ix++)
			fprintf(igp->igr_fpl, "    ");
		  }
		fprintf(igp->igr_fpl, "   _");
		if (lpcen > DFUZZ)
		  {
		   for (ix = 0, x = igp->igr_xmi; ix < lx; ix++, x++)
		     {
			thlp = (EOK == ixy2ev(x, y, ev)) ?
				ev2lp(ev, igp) : lpcen;
			dellp = ((thlp - lpcen) * 100.) / lpcen;
			if (dellp < -99) dellp = -99;
			if (abs(dellp) < 1) fprintf(igp->igr_fpl, "    ");
			else fprintf(igp->igr_fpl, "%4d", dellp);
		     }
		  }
		else
		  {
			for (ix = 0; ix < lx; ix++)
				fprintf(igp->igr_fpl, "    ");
		  }
		fprintf(igp->igr_fpl, " _\n");
	   }
	fprintf(igp->igr_fpl, "     ");
	for (ix = 0, x = igp->igr_xmi; ix < lx; ix++, x++)
	   {
		if (x % 5) fprintf(igp->igr_fpl, "   |");
		else fprintf(igp->igr_fpl, "%4d", x);
	   }
	fprintf(igp->igr_fpl, "    ");
	for (ix = 0, x = igp->igr_xmi; ix < lx; ix++, x++)
	   {
		if (x % 5) fprintf(igp->igr_fpl, "   |");
		else fprintf(igp->igr_fpl, "%4d", x);
	   }
	fprintf(igp->igr_fpl, "\n");
		
}

void get_limits(OBSREF *rp, int whichpart, INTEGRUN *igp)
{ /*	This computes the center values (igp->igr_xcen, igp->igr_ycen)
	and the corners--
	igp->igr_xmi, igp->igr_xma, igp->igr_ymi, and igp->igr_yma--
	of the box surrounding the reflection `rp'.
	The index `whichpart' is PREDICTED if we're to
	use the predicted centroid; it's OBSERVED otherwise. */
	int		lx2, ly2;
	REGIONPROFILE	*rgb;

	if (whichpart == PREDICTED)
	  {
		igp->igr_fphipr = rp->obs_ompred;
		igp->igr_fxpr = ((double)rp->obs_xpred) / igp->igr_xscal;
		igp->igr_fypr = ((double)rp->obs_ypred) / igp->igr_yscal;
	  }
	else
	  {
		igp->igr_fphipr = rp->obs_omob;
		igp->igr_fxpr = ((double)rp->obs_xob) / igp->igr_xscal;
		igp->igr_fypr = ((double)rp->obs_yob) / igp->igr_yscal;
	  }
	rgb = igp->igr_rgb;	lx2 = rgb->m3_bx / 2;	ly2 = rgb->m3_by / 2;
	igp->igr_xcen = (int)(igp->igr_fxpr + 0.49999);
	igp->igr_ycen = (int)(igp->igr_fypr + 0.49999);
	igp->igr_xmi = igp->igr_xcen - lx2; igp->igr_xma = igp->igr_xcen + lx2;
	igp->igr_ymi = igp->igr_ycen - ly2; igp->igr_yma = igp->igr_ycen + ly2;
}

int istestrefl(INTEGRUN *igp)
{ /*	This returns true if (igp->igr_reflp)->obs_hkl
	is one of the test reflections */
	int	ix, iy, iz, ntry;
	short	**trlp, *trh;
	OBSREF	*lrp;

	lrp = igp->igr_reflp;	ntry = igp->igr_nrverbose;
	for (trlp = igp->igr_trlb, iz = ntry; iz--; trlp++) /*detailed print?*/
	   {
		for (iy = ix = 0, trh = *trlp; iy < 3; iy++, trh++)
		   if (lrp->obs_hkl[iy] != *trh)
		     {
			ix = 1; break;
		     }
		if (!ix) return 1;	/* yes: refl is on "details" list */
	   }
	return 0;
}

static void orderaxes(double mm[3][3], double ev[])
{ /*	This fiddles with a rotation matrix `mm'.  It first guarantees that
	its rows (which are eigenvalues) are set up with their largest
	element > 0.  It then determines which axis of the rotation matrix mm
	is closest to z, and which of the remaining two is closest to y.
	It then reorders the axes accordingly.  If ev is non-NULL, the same
	reordering is performed on it (presumably it's the vector of
	eigenvalues). It also forces the determinant of mm to be positive by
	negating the first row if necessary. */
	register int	i, j;
	int		imax, imid, imin;
	double		a2, a1, rowmin, rowsum;

	/* in each row, make the element with the largest absolute value > 0.*/
	for (j = 3; j--; )
	   {
		imax = 2;	a2 = fabs(mm[j][2]);
		if ((a1 = fabs(mm[j][1])) > a2)
		  {
			a2 = a1;	imax = 1;
		  }
		if ((a1 = fabs(mm[j][0])) > a2)
		  {
			a2 = a1;	imax = 0;
		  }
		if (mm[j][imax] < 0)
		  {
			mm[j][0] = -mm[j][0]; mm[j][1] = -mm[j][1];
			mm[j][2] = -mm[j][2];
		  }
	   }
	/* now find the row with the largest element along z. */
	imax = 2;	a2 = fabs(mm[imax][2]);
	if ((a1 = fabs(mm[1][2])) > a2)
	  {
		a2 = a1; imax = 1;
	  }
	if ((a1 = fabs(mm[0][2])) > a2)
	  {
		a2 = a1; imax = 0;
	  }
	/* now `imax' is the index of the row closest to z */
	imid = (imax + 1) % 3;	imin = 3 - imax - imid;
	a2 = fabs(mm[imid][1]);
	if ((a1 = fabs(mm[imin][1])) > a2)
	  {
		imid = imin;	imin = 3 - imax - imid;
	  }
	/* now `imid' is closest to y and `imin' is the one left over */
	for (j = 3; j--; )
	   {
		a2 = mm[imax][j]; a1 = mm[imid][j];
		mm[0][j] = mm[imin][j];	mm[1][j] = a1; mm[2][j] = a2;
	   }
	if (NULL != ev) /* now take care of the eigenvalues, if any */
	  {
		a2 = ev[imax];	a1 = ev[imid];
		ev[0] = ev[imin];	ev[1] = a1;	ev[2] = a2;
	  }
	/* flip a row of mm if the determinant is less than zero.
	We choose to flip the one with the smallest signed sum. */
	if (ddeter(mm) < 0.)
	   {
		rowmin = 1.e30;
		for (imin = -1, i = 0; i < 3; i++)
		   {
			for (rowsum = j = 0; j < 3; j++) rowsum += mm[i][j];
			if (rowsum < rowmin)
			  {
				imin = i; rowmin = rowsum;
			  }
		   }
		for (j = 3; j--; ) mm[imin][j] = -mm[imin][j];
	   }
}

double zerf(double zx)
{	/* This returns the error function of the argument.  It returns
	identically 1 for zx > MAXERFARG.  This is formula 7.1.28 in
	Abromowitz & Stegun, 1964. */
	double		x, zt;
	static double	az[6] = {	.0705230784, .0422820123, .0092705272,
					.0001520143, .0002765672, .0000430638};

	if (zx < 0.) zx = -zx;	if (zx > MAXERFARG) return 1.;
	x = zx;
	zt = 1. + x * (az[0] + x * (az[1] + x * (az[2] + x * (az[3] + x *
		(az[4] + x * az[5])))));
	zt *= (zt * zt * zt);	zt *= (zt * zt * zt);
	return 1. - 1. / zt;
}

double varlp(INTEGRUN *igp)
{ /*	This returns an estimate of the variance of the Lorentz-polarization
	factor in the neighborhood of the reflection currently active.  We
	assume that if C = 1/LP then sigma(C) = (C(x1) - C(x0)) * 3 / (x1 - x0)
	where x1 and x0 are the x coordinates in the neighborhood of the
	reflection.  The spread in a spot is partly due to the size and
	crossfire of the direct beam itself, in addition to the width
	associated with the crystal's own finite size.  The "3" comes from a
	guess that the direct-beam ambiguity is roughly 0.6mm or 3 pixels'
	worth. The routine returns varlp = sigma(C) * sigma(C). */
	int	ix0, ix1;
	double	lx0, lx1, sigc, ev[3];

	for (ix0 = igp->igr_xmi; ix0 <= igp->igr_xcen; ix0++)
		if (EOK == ixy2ev(ix0, igp->igr_ycen, ev)) break;
	lx0 = ev2lp(ev, igp);
	for (ix1 = igp->igr_xma; ix1 > igp->igr_xcen; ix1--)
		if (EOK == ixy2ev(ix1, igp->igr_ycen, ev)) break;
	lx1 = ev2lp(ev, igp);
	if (ix1 <= ix0) return 0;	/* we hope not!!! */
	sigc = 3. * ((double)(lx1 - lx0)) / (double)((ix1 - ix0));
	return sigc * sigc;
}

int mkmodel(int regi, INTEGRUN *igp)
{ /*	This constructs the model */
	register int	iz, iy, ix, iw;
	int		jreg, lx, lxyz, ly, lz, lz2, neighbor;
	double		alpha, dz, fmx, imascale, meansharp, mgam, mweight;
	double		sumgam, sumwt, sumwts, thisrat, tgam;
	double		devob[3], regwt[NNEIGHBORS], *wtp;
	float		*bip, *bmp;
	OBSREF		*lrp;
	REGIONATT	*wbmp;
	REGIONPROFILE	*rgb, *rgp;
	static int	wnbr[NREGIONS][NNEIGHBORS] = {
		{ 0, 1, 3, 4,-1 }, { 1, 0, 2, 5, 6 }, { 2, 1, 7, 8,-1 },
		{ 3, 0, 4, 9,-1 }, { 4, 0, 3, 5, 9 }, { 5, 1, 4, 6,10 },
		{ 6, 1, 5, 7,10 }, { 7, 2, 6, 8,11 }, { 8, 2, 7,11,-1 },
		{ 9, 3, 4,10,-1 }, {10, 5, 6, 9,11 }, {11, 7, 8,10,-1 } } ;

	/* Compute the weights associated with the various model profiles that
	will be used in determining the overall model profile for the refl */
	rgb = igp->igr_rgb;
	lx = rgb->m3_bx;	ly = rgb->m3_by;	lz = rgb->m3_bz;
	lz2 = lz / 2;	lxyz = lx * ly * lz;
	for (iw = fmx = 0, wtp = &(regwt[0]); iw < NNEIGHBORS; iw++, wtp++)
	   {
		if ((neighbor = wnbr[regi][iw]) > -1)
		  {
			wbmp = igp->igr_bmask + neighbor;
			mweight = exp(-ALPHAX * fabs(igp->igr_fxpr -
			 wbmp->ra_xc) -
			 ALPHAY * fabs(igp->igr_fypr - wbmp->ra_yc));
			if (mweight > fmx) fmx = mweight;
			*wtp = mweight;
		  }
		else	*wtp = 0.;
	   }
	for (sumwts = iw = 0, wtp = &(regwt[0]); iw < NNEIGHBORS; iw++, wtp++)
	   {
		if (*wtp < 0.1 * fmx) *wtp = 0;
		else sumwts += *wtp;
	   }
	for (sumgam = sumwt = iw = 0, wtp = &(regwt[0]);
	 iw < NNEIGHBORS; iw++, wtp++)
	   {
		*wtp /= sumwts;
		if (0 <= (jreg = wnbr[regi][iw]))
		  {
			wbmp = igp->igr_bmask + jreg;
			sumgam += wbmp->ra_meangam * *wtp;	sumwt += *wtp;
		  }
	   }
	wbmp = igp->igr_bmask + regi;
	mgam = (sumwt < DFUZZ) ? wbmp->ra_meangam : sumgam / sumwt;
	/* determine scale factors for the models themselves and
	 then add 'em all up */
	zeromem((void *)igp->igr_bm, (unsigned)(lxyz * sizeof (float)));
	for (iw = 0, wtp = &(regwt[0]); iw < NNEIGHBORS; iw++, wtp++)
	   {
		if (*wtp > DFUZZ)
		  {
			jreg = wnbr[regi][iw];	rgp = rgb + jreg;
			for (bip = rgp->m3_xyop, imascale = iz = 0;
			 iz < lxyz; iz++, bip++) imascale += *bip;
			if (imascale > DFUZZ)
			  {
				imascale = 1000. * (*wtp) / imascale;
				bmp = igp->igr_bm;	bip = rgp->m3_xyop;
				for (iz = 0; iz < lxyz; iz++, bmp++, bip++)
					*bmp += imascale * *bip;
			  }
		  }
	   }
	lrp = igp->igr_reflp;
	tgam = ((double)(lrp->obs_gama)) * 1.e-3;
	if ((fabs(tgam) > DFUZZ) && (fabs(mgam) > DFUZZ))
		alpha = 0.5 * (1. / (tgam * tgam) - 1. / (mgam * mgam));
	else	alpha = 0.;
	meansharp = exp(-alpha);
	if (igp->igr_flags & (IR_VERBOSE2 | IR_VERBOSETEMP))
	  {
		wtp = igp->igr_indlim;
		fprintf(igp->igr_fpl,
	 "tgam,mgam,alpha,meansharp: %7.3f,%7.3f,%7.4f,%7.3f\n",
			tgam, mgam, alpha, meansharp);
		fprintf(igp->igr_fpl,
	 "Indices range from [%6.2f,%6.2f,%6.2f] to [%6.2f,%6.2f,%6.2f]\n",
		 *wtp, *(wtp + 1), *(wtp + 2),
		 *(wtp + 3), *(wtp + 4), *(wtp + 5));
		fprintf(igp->igr_fpl, "Models used: ");
		for (iw = 0, wtp = &(regwt[0]); iw < NNEIGHBORS; iw++, wtp++)
		   if (*wtp > DFUZZ) fprintf(igp->igr_fpl,
			"wt[%d]=%5.3f; ", wnbr[regi][iw], *wtp);
		fprintf(igp->igr_fpl, "\n");
	  }
	/* adjust the model for the difference between the predicted spot
	width and the mean of the predicted widths in the regions that
	were used to make up the profile */
	iz = -lz2;	dz = iz;
	for (sumgam = iz = 0, bmp = igp->igr_bm; iz < lz; iz++, dz += 1)
	   {
		thisrat = flimits(exp(dz * dz * alpha) * meansharp, 0.2, 5.);
		for (iy = 0; iy < ly; iy++) for (ix = 0; ix < lx; ix++, bmp++)
		   {
			*bmp *= thisrat;	sumgam += *bmp;
		   }
	   }
	if (sumgam < DFUZZ)
	  {
		fprintf(stderr,
		 "[%3d,%3d,%3d]: effective gamma value = %12.5e\n",
		 lrp->obs_hkl[0], lrp->obs_hkl[1], lrp->obs_hkl[2], sumgam);
		return I_BADMODEL;
	  }
	for (iz = 0, bmp = igp->igr_bm; iz < lxyz; iz++, bmp++) *bmp /= sumgam;
	/* now we need to deliberately mis-center the profile by the deviation
	 between the actual profile's centroid and the nearest integer to it.*/
	devob[0] = ((int)igp->igr_fxpr) - igp->igr_fxpr;
	if (devob[0] < -0.5) devob[0] += 1.;
	devob[1] = ((int)igp->igr_fypr) - igp->igr_fypr;
	if (devob[1] < -0.5) devob[1] += 1.;
	devob[2] = ((int)igp->igr_fphipr) - igp->igr_fphipr;
	if (devob[2] < -0.5) devob[2] += 1.;
	recen_vec(igp->igr_bm, devob, igp->igr_tq, lx, ly, lz);
	for (sumgam = iz = 0, bmp = igp->igr_tq; iz < lxyz; iz++, bmp++)
		sumgam += *bmp;
	if (sumgam < DFUZZ)
	  {
		fprintf(stderr,
		 "[%3d,%3d,%3d]: profile is nonpositive after recentering\n",
		 lrp->obs_hkl[0], lrp->obs_hkl[1], lrp->obs_hkl[2]);
		return I_BADMODEL;
	  }
	for (iz = 0, bmp = igp->igr_bm, bip = igp->igr_tq; iz < lxyz;
		iz++, bmp++, bip++) *bmp = *bip / sumgam;
	return I_USABLE;
}

void framestats(int fmno, INTEGRUN *igp)
{ /*	This prepares the statistics associated with a specific image.
	The computation of the D2 value is based on a least-squares fit
	if ln(I/sigma) to ln(Is0) + B * (-s*s) where s = sin(theta)/lambda;
	D2 = lambda*sqrt(ln(Is0)/(B-ln2)) */
	int		headoff;
	double		discr, loclam, numa, numb, rout;
	FRAMELOG	*frp;

	headoff = fmno % (igp->igr_nmbox - 1);
	frp = igp->igr_logst + headoff;	loclam = (igp->igr_xp)->xlambda;
	 /* only the strong reflections contribute to the rms errors */
	if ((rout = frp->outstrong) > 0.99)
	  {
		frp->iis_rmx = sqrt(frp->iis_rmx / rout);
		frp->iis_rmy = sqrt(frp->iis_rmy / rout);
		frp->iis_rmz = sqrt(frp->iis_rmz / rout);
	  }
	if ((rout = frp->outrefl) > 0.99) /* get <I/sigma>, I000, D2, etc. */
	  {
		frp->iis_in /= rout;	frp->iis_ios /= rout;
		discr = frp->iis_b * frp->iis_d2 - frp->iis_ssq * frp->iis_ssq;
		if (discr > DFUZZ)
		  {
			numa = frp->iis_b * frp->iis_lni -
				frp->iis_ssq * frp->iis_i000;
			numb = frp->iis_d2 * frp->iis_i000 -
				frp->iis_ssq * frp->iis_lni;
			frp->iis_i000 = numa / discr;
			frp->iis_b = -numb / discr;
			if ((frp->iis_b > DFUZZ) &&
			 (0 < (numa = frp->iis_i000 - LN2)))
			  frp->iis_d2 = 0.5 * sqrt(frp->iis_b / numa);
			else	frp->iis_d2 = -1.;	/* flag value */
			frp->iis_i000 = exp(frp->iis_i000);
		  }
	  }
	else {
		frp->iis_in = frp->iis_ios = frp->iis_rmx = frp->iis_rmy =
		 frp->iis_rmz = frp->iis_b = frp->iis_i000 = frp->iis_d2 = 0.;
	     }
}

static int adj_model(int datatype, int direction, INTEGRUN *igp)
{ /*	This adjusts the 3-D model profile to make it resemble the signal
	profile. The adjustment involves summing all the elements in a
	given plane, determined by `direction' to be (XY), (YZ), or (ZX),
	of the model profile, the signal profile, and the variance profile.
	The model is then adjusted so that the sum within each plane is
	proportional to the corresponding signal sum. The adjustment is only
	partial: the variance sum is used to `temper' the degree to which the
	adjustment gets made. In order for this routine to work with sums
	perpendicular to Z, Y, or X, the incrementing is all done in a rather
	abstract way.  The indices "o", "c", and "i" refer to the outer,
	center, and inner summations respectively.  The routine returns 0 if
	successful, 1 if the signal is too weak to do an adjustment, and 2
	if there's something wrong with the model */
	int		ii, im, io, ji, jm, jo, lx, ly, lz, mx, my, mz;
	int		ni, nm, no, nx, nxy, nxyz, ny, nz, no_nl;
	double		mall, nmall, nu, sall, sus[16], sum[16], suv[16];
	double		*mup, *sup, *vup;
	REGIONPROFILE	*rgb;
	static float	*vcm, *vcs, *vcv, *vim, *vis, *viv;
	static float	*vm, *vom, *vos, *vov, *vs, *vv;
	static char	adir[] = "xyz";
	static char	*atyp[] = {"Dir.", "Kab." };

	rgb = igp->igr_rgb;
	lx = rgb->m3_bx;	ly = rgb->m3_by;	lz = rgb->m3_bz;
	mx = rgb->m3_qx;	my = rgb->m3_qy;	mz = rgb->m3_qz;
	if (datatype)
	  {
		vs = igp->igr_ks; vv = igp->igr_kv; vm = igp->igr_km;
		nx = mx;	ny = my;	nz = mz;
	  }
	else {
		vs = igp->igr_bs; vv = igp->igr_bv; vm = igp->igr_bm;
		nx = lx;	ny = ly;	nz = lz;
	     }
	nxy = nx * ny;	nxyz = nxy * nz;
	if (direction == 2)	/* sums perpendicular to Z */
	  {
		no = nz;	nm = ny;	ni = nx;
		io = nxy;	im = nx;	ii = 1;
	  }
	else if (direction == 1)	/* sums perpendicular to Y */
	  {
		no = ny;	nm = nx;	ni = nz;
		io = nx;	im = 1;		ii = nxy;
	  }
	else {
		no = nx;	nm = nz;	ni = ny; /* perp to X */
		io = 1;		im = nxy;	ii = nx;
	     }
	sup = &(sus[0]);	mup = &(sum[0]);	vup = &(suv[0]);
	/* determine the signal, model, and variance sums within each plane */
	for (sall = mall = jo = 0, vos =vs, vom = vm, vov = vv; jo < no;
	   jo++, vos += io, vom += io, vov += io, sup++, mup++, vup++)
	   {
		*sup = *mup = *vup = 0.;
		for (jm = 0, vcs = vos, vcm = vom, vcv = vov; jm < nm;
		   jm++, vcs += im, vcm += im, vcv += im)
		   {
			for (ji = 0, vis = vcs, vim = vcm, viv = vcv; ji < ni;
			   ji++, vis += ii, vim += ii, viv += ii)
			   if ((*viv > 0.) && (*vim > igp->igr_minmod))
			     {
				*sup += *vis;	*mup += *vim;	*vup += *viv;
			     }
		   }
		sall += *sup;	mall += *mup;
	   }
	if (mall < DFUZZ) return 2;	/* unstable model: eech! */
	if (sall < DFUZZ) return 1;	/* weak intensity: more palatable */
	sall /= mall;			/* renormalize */
	sup -= no;	mup -= no;	vup -= no; /* rewind sum pointers */
	/* adjust the models so that each plane has the same fraction of
	the total intensity as was true in the signal.  The adjustment is
	only partial: the adjustments are themselves adjusted toward one
	by a term that is proportional to the variance. */
	for (no_nl = nmall = jo = 0, vom = vm; jo < no;
	 jo++, vom += io, sup++, mup++, vup++)
	   {
		if (*mup > DFUZZ && *vup > DFUZZ)
		  {
			*mup *= sall;
			*sup /= *mup;	/* nominal rescaler */
			nu = NSTD_SUM * sqrt(*vup) / *mup; /* std. dev. */
			if (*sup < 1. - nu) *sup += nu; /* move up by 1sigma*/
			else if (*sup > 1. + nu) *sup -= nu; /*down by 1sigma*/
			else *sup = 1.;	/* close to one: don't adjust */
			*sup = flimits(*sup, 0.25, 4.); /* avoid extremes */
			if (igp->igr_flags & IR_VERBOSEANY)
			  {
				if (!no_nl) fprintf(igp->igr_fpl,
					"%s",atyp[datatype]);
				else if (!(no_nl % 9))
					fprintf(igp->igr_fpl, "    ");
				fprintf(igp->igr_fpl, "%ch[%2d]=%5.3f; ",
				 adir[direction], jo, *sup);
				if (!((++no_nl) % 9))
					fprintf(igp->igr_fpl, "\n");
			  }
			/* having computed the rescaling value, apply it */
			for (jm = 0, vcm = vom; jm < nm; jm++, vcm += im)
			   {
				for (ji = 0, vim = vcm; ji < ni; ji++,vim +=ii)
				   {
					*vim *= *sup;	nmall += *vim;
				   }
			   }
		  }
	   }
	if ((igp->igr_flags & IR_VERBOSEANY) && (no_nl % 9))
		fprintf(igp->igr_fpl, "\n");
	if (nmall < DFUZZ) return 2;
	for (jo = 0, vom = vm; jo < nxyz; jo++, vom++) /* now renormalize */
		*vom /= nmall;
	return 0;
}

static int ellmom(int datatype, int whichv, double *emom, INTEGRUN *igp)
{ /*	This computes the parameters of a Gaussian ellipsoid on the model or
	signal implicitly, by moment calculations, rather than explicitly.
	Then it obtains the offsets, eigenvalues, and eigenvectors of the
	resulting parameter set. The routine delivers a routine value thus:
	return value	offsets		eigenvalues	eigenvectors
		0	reliable	reliable	reliable
		1	reliable	reliable	unreliable
		3	reliable	unreliable	unreliable
		7	unreliable	unreliable	unreliable */
	register int	ix, iy, iz;
	int		lx, ly, lz, mx, my, mz, nx, nx2, ny, ny2, nz, nz2, retv;
	double		detw, discrim, dx, dy, dy2, dyz2, dz, dz2;
	double		m2dy, m2dz, minors, thiss, thisv, trw;
	double		locsol[3], momsu[10], qfi[3][3], wma[3][3], *emp;
	REGIONPROFILE	*rgb;
	static float	*vm, *vmp, *vu, *vup, *vv, *vvp;

	rgb = igp->igr_rgb;
	lx = rgb->m3_bx;	ly = rgb->m3_by;	lz = rgb->m3_bz;
	mx = rgb->m3_qx;	my = rgb->m3_qy;	mz = rgb->m3_qz;
	if (datatype)
	  {
		nx = mx; ny = my; nz = mz;
		vu = (whichv)? igp->igr_km : igp->igr_ks;
		vm = igp->igr_km;	vv = igp->igr_kv;
	  }
	else
	  {
		nx = lx; ny = ly; nz = lz;
		vu = (whichv)? igp->igr_bm : igp->igr_bs;
		vm = igp->igr_bm;	vv = igp->igr_bv;
	  }
	nx2 = nx / 2;	ny2 = ny / 2;	nz2 = nz / 2;
	for (iz = 0; iz < 10; iz++) momsu[iz] = 0.;
		/* accumulate the moment sums */
	for (iz = 0, dz = -nz2, vup = vu, vmp = vm, vvp = vv;
	  iz < nz; iz++, dz += 1.)
	   {
		dz2 = dz * dz;	m2dz = -2. * dz;
		for (iy = 0, dy = -ny2; iy < ny; iy++, dy += 1.)
		   {
			dy2 = dy * dy;	dyz2 = 2. * dy * dz; m2dy = -2. * dy;
			for (ix = 0, dx = -nx2; ix < nx;
			 ix++, dx += 1., vup++, vmp++, vvp++)
			  if (*vmp > igp->igr_minmod)
			    {
				thisv = *vup;
				if (*vup < 0.)	/*negatives treated guardedly*/
				  {
					if (*vvp < DFUZZ) thisv = 0.;
					else  {
						thiss = SIGRAT * (-sqrt(*vvp));
						if (thisv < thiss)
							thisv = thiss;
					      }
				  }
				momsu[0] += thisv;
				momsu[1] += dx * dx * thisv;
				momsu[2] += 2. * dx * dy * thisv;
				momsu[3] += 2. * dx * dz * thisv;
				momsu[4] += dy2 * thisv;
				momsu[5] += dyz2 * thisv;
				momsu[6] += dz2 * thisv;
				momsu[7] += -2. * dx * thisv;
				momsu[8] += m2dy * thisv;
				momsu[9] += m2dz * thisv;
			    }
		   }
	   }
	if (momsu[0] < 0.01) /* very weak */
	  {
		if (igp->igr_flags & IR_VERBOSEANY) fprintf(igp->igr_fpl,
 "Moments unreliable for %s %s data; Sum of elements = %13.5e\n",
		 (datatype ? "Kabsch" : "direct"),
		 (whichv ? "model" : "signal"), momsu[0]);
		return 7;
	  }
	for (iz = 1; iz < 10; iz++) momsu[iz] /= momsu[0];
	for (iz = 7; iz < 10; iz++) momsu[iz] *= -0.5;
	for (iz = 7; iz < 10; iz++) emom[iz + 3] = momsu[iz]; /* translations*/
	wma[0][0] = 2. * (momsu[1] - momsu[7] * momsu[7]);
	wma[0][1] = wma[1][0] = momsu[2] - 2. * momsu[7] * momsu[8];
	wma[0][2] = wma[2][0] = momsu[3] - 2. * momsu[7] * momsu[9];
	wma[1][1] = 2. * (momsu[4] - momsu[8] * momsu[8]);
	wma[1][2] = wma[2][1] = momsu[5] - 2. * momsu[8] * momsu[9];
	wma[2][2] = 2. * (momsu[6] - momsu[9] * momsu[9]);
	if (0 > (detw = ddeter(wma)))
	  {
		for(iz = 0; iz < 3; iz++) for (iy = 0; iy < 3; iy++)
			wma[iz][iy] = -wma[iz][iy];
		detw = -detw;
	  }
	/* construct the parameters of the characteristic polynomial */
	trw = wma[0][0] + wma[1][1] + wma[2][2];	/* the trace */
	/* get the sum of the minors of the diagonal elements */
	minors = wma[0][0] * (wma[1][1] + wma[2][2]) + wma[1][1] * wma[2][2] -
	 wma[0][2] * wma[0][2] - wma[0][1] * wma[0][1] - wma[1][2] * wma[1][2];
	/* the discriminant of the characteristic polynomial */
	discrim = trw * trw - 3. * minors;
	/* this matrix must be well-behaved, i.e. have positive determinant
	and trace, since all its eigenvalues are supposed to be positive */
	if ((DFUZZ > detw) || (trw < 0.) || (discrim < 0.))
	  {
		if (igp->igr_flags & IR_VERBOSEANY)
		  {
			fprintf(igp->igr_fpl,
			 "Ill-behaved moment matrix for %s %s data; ",
				(datatype ? "Kabsch" : "direct"),
				(whichv ? "model" : "signal"));
			fprintf(igp->igr_fpl,
			 "determinant,trace,discriminant= %8.4f,%8.4f,%8.4f\n",
				detw, trw, discrim);
		  }
		/* in this case we do the best we can... */
		emom[0] = log(sqrt(PICUBED) / momsu[0]); /* "intensity" */
		/* if we assume a diagonal model then wma[i][i] contain
		the reciprocals of the widths along x,y,z. */
		for (iz = iy = 0, emp = emom + 13; iy < 3; iy++, emp++)
		   {
			if ((thisv = wma[iy][iy]) < DFUZZ) *emp = MAXWID;
			else {
				iz++;	*emp = 1. / thisv;
			     }
		   }
		for (iy = 0, emp = emom + 16; iy < 3; iy++) /* identity mtrx */
			for (ix = 0; ix < 3; ix++, emp++) *emp = (iy == ix);
		return (iz < 3 ? 3 : 1);
	  }
	/* save the lsq parameters */
	if (dinv3(wma, qfi))
	  {
		if (igp->igr_flags & IR_VERBOSEANY)
			fprintf(igp->igr_fpl, "Weird singularity!!\n");
		return 3;
	  }
	emom[1] = qfi[0][0];	emom[2] = qfi[0][1];	emom[3] = qfi[0][2];
	emom[4] = qfi[1][1];	emom[5] = qfi[1][2];	emom[6] = qfi[2][2];
	dmm31(locsol, qfi, &(momsu[7]));
	for (iy = 0; iy < 3; iy++) emom[7+iy] = locsol[iy];
	/* well-behaved case */
	emom[0] = log(sqrt(PICUBED * detw / momsu[0])); /* Peak */
	discrim = sqrt(discrim);
	/* Use these to calculate, by Newton's method, two roots.
	They should be positive in order for the surface to be an ellipsoid
	and not a hyperboloid.
	For each eigenvalue we determine the associated eigenvector. */
	retv = 0;	/* we hope it's okay now...*/
	if (DFUZZ > (emom[14] = newtcube(0., -trw, minors, -detw))) retv = 3;
	if (eigvec(wma, emom[14], qfi[1])) retv |= 1;
	emom[13] = trw - emom[14];
	if (DFUZZ > (emom[13] = newtcube(emom[13], -trw, minors, -detw)))
		retv = 3;
	if (eigvec(wma, emom[13], qfi[0])) retv |= 1;
	/* the last eigenvalue we get from the fact that the trace of a
	matrix is invariant to its representation, i.e.
	 lambda0 + lambda1 + lambda2 = trace(em)  */
	if (DFUZZ > (emom[15] = trw - emom[13] - emom[14])) retv = 3;
	/* the last eigenvector is orthogonal to the first 2, and since they're
	 normalized, we can compute it as the cross-product of the first 2.*/
	qfi[2][0] = qfi[0][1] * qfi[1][2] - qfi[0][2] * qfi[1][1];
	qfi[2][1] = qfi[0][2] * qfi[1][0] - qfi[0][0] * qfi[1][2];
	qfi[2][2] = qfi[0][0] * qfi[1][1] - qfi[0][1] * qfi[1][0];
	 /* reorder the eigenvectors so the one closest to z is last */
	orderaxes(qfi, emom + 13);
	/* keep the reciprocals of the eigenvalues.  These are
	 scaled the way we actually want them. */
	for (iy = 13, emp = emom + 13; iy < 16; iy++, emp++)
		*emp = (*emp > DFUZZ ? 1. / *emp : MAXWID);
	/* make the eigenvector matrix a bit more accurate */
	if (dinv3(qfi, wma))	/* if this fails we're in deep doo-doo! */
	  {
		retv |= 1;
		for (iy = 3; iy--; ) for (ix = 3; ix--; )
			qfi[iy][ix] = (iy == ix);
	  }
	else for (iy = 3; iy--; ) for (ix = 3; ix--; )
		{
			if (fabs(qfi[iy][ix] - wma[ix][iy]) > 0.005) retv |= 1;
			qfi[iy][ix] = 0.5 * (qfi[iy][ix] + wma[ix][iy]);
		}
	/* keep the matrix elements */
	for (emp = emom + 16, iy = 0; iy < 3; iy++)
		for (ix = 0; ix < 3; ix++, emp++) *emp = qfi[ix][iy];
	if (retv && (igp->igr_flags & IR_VERBOSEANY)) fprintf(igp->igr_fpl,
	 "Partial failure in moments for %s %s data: retv = %d\n",
	 (datatype ? "Kabsch" : "direct"), (whichv ? "model":"signal"), retv);
	/* estimate the fraction of the model profile that's missing. */
	if (whichv == 1)
	  {
		thiss = zerf(sqrt(emom[13]) * (((double)nx2) + 0.5)) *
		 zerf(sqrt(emom[14]) * (((double)ny2) + 0.5)) *
		 zerf(sqrt(emom[15]) * (((double)nz2) + 0.5));
		igp->igr_tailc[datatype] = 1. / flimits(thiss, 0.91, 1.);
	  }
	return retv;
}

int sharpmodel(int datatype, double riv, INTEGRUN *igp)
{ /*	This sharpens or broadens a model profile using a Gaussian ellipsoid
	approximation to the real data and the model data.  It actually does
	a recentering at the same time, one hopes.  The routine returns 0
	if successful, nonzero otherwise.  */
	register int	iz, nx, ny, nz;
	char		sigmombad;
	int		lx, ly, lz, mx, my, mz, nxyz;
	double		dz, xcom[25], xcos[25];
	OBSREF		*lrp;
	REGIONPROFILE	*rgb;
	static float	*vm, *vmp;

	if (riv < 0.5) return 1;	/* too weak to adjust */
	rgb = igp->igr_rgb;
	lx = rgb->m3_bx;	ly = rgb->m3_by;	lz = rgb->m3_bz;
	mx = rgb->m3_qx;	my = rgb->m3_qy;	mz = rgb->m3_qz;
	if (datatype)
	  {
		nx = mx;	ny = my;	nz = mz; vm = igp->igr_km;
	  }
	else {
		nx = lx;	ny = ly;	nz = lz; vm = igp->igr_bm;
	     }
	/* correct the model based on equating the fractional contributions
	 of the profile within planes */
	lrp = igp->igr_reflp;
	if (!lrp->obs_flags)
	  {
		nxyz = nx * ny * nz;
		for (dz = iz = 0, vmp = vm; iz < nxyz; iz++, vmp++)
			dz += *vmp;	/* renormalize model */
		if (dz < DFUZZ) return 2;	/* bizarre model: no dice */
		for (iz = 0, vmp = vm; iz < nxyz; iz++, vmp++) *vmp /= dz;
		/* adjust in Z first; then (if it's okay) in Y and X */
		if (1 < adj_model(datatype, 2, igp)) return 9; /* adjust in z*/
		if (1 < adj_model(datatype, 1, igp)) return 10; /*adjust in y*/
		if (1 < adj_model(datatype, 0, igp)) return 11; /*adjust in x*/
	  }
		/* approximate the  model data as a Gaussian ellipsoid */
	if ((ellmom(datatype, 1, xcom, igp))) return 2;
		/* approximate the signal data as a Gaussian ellipsoid */
	if ((sigmombad = ellmom(datatype, 0, xcos, igp)))
	  {
		if (sigmombad & 04)	/* very weak */
			for (iz = 10; iz < 13; iz++) xcos[iz] = xcom[iz];
		if (sigmombad & 02)	/* unreliable eigenvalues */
			for (iz = 13; iz < 16; iz++) xcos[iz] = xcom[iz];
		/* we never trust the eigenvectors here, so we set em to xcom*/
		for (iz = 16; iz < 25; iz++) xcos[iz] = xcom[iz];
	  }
	if (igp->igr_flags & IR_VERBOSEANY)
	  {
		fprintf(igp->igr_fpl,
		 "%s     deltaX  deltaY  deltaZ EigvalX EigvalY EigvalZ",
		 (datatype ? "Kabsch" : "direct"));
		for (iz = 0; iz < 3; iz++) fprintf(igp->igr_fpl,
			" Eigve%1dX Eigve%1dY Eigve%1dZ", iz, iz, iz);
		fprintf(igp->igr_fpl,"\n   Model:");
		for (iz = 10; iz < 25; iz++)
			fprintf(igp->igr_fpl, " %7.3f",xcom[iz]);
		fprintf(igp->igr_fpl, "\n");
		if (!(sigmombad & 06))
		  {
			fprintf(igp->igr_fpl,"  Signal:");
			for (iz = 10; iz < 25; iz++)
				fprintf(igp->igr_fpl, " %7.3f",xcos[iz]);
			fprintf(igp->igr_fpl, "\n");
		  }
	  }
	return 0;
}

void init3dvec(INTEGRUN *igp)
{ /*	Reserve memory for working 3D vectors if needed, or else zero them */
	int		lxyz, mxy, mxyz;
	REGIONPROFILE	*rgb;
	static size_t	wrefsize;

	/* establish where the working vectors are */
	rgb = igp->igr_rgb;
	lxyz = rgb->m3_bx * rgb->m3_by * rgb->m3_bz;
	mxy = rgb->m3_qx * rgb->m3_qy;	mxyz = mxy * rgb->m3_qz;
	wrefsize = (sizeof (float)) * (4 * lxyz + 3 * mxyz + 2 * mxy) +
		sizeof (u_char) * mxyz + 16000;
	if (NULL == igp->igr_wintv)
	  {
		igp->igr_wintv = (char *)mallop(wrefsize,"wintv in init3dvec");
	  }
	zeromem(((AVOID *)(igp->igr_wintv)), wrefsize);
	igp->igr_bs = (float *)(igp->igr_wintv);
	igp->igr_bm = igp->igr_bs + lxyz; igp->igr_bv = igp->igr_bm + lxyz;
	igp->igr_tq = igp->igr_bv + lxyz; igp->igr_ks = igp->igr_tq + lxyz;
	igp->igr_km = igp->igr_ks + mxyz; igp->igr_kv = igp->igr_km + mxyz;
	igp->igr_kbg = igp->igr_kv + mxyz; igp->igr_kbgd = igp->igr_kbg + mxy;
	igp->igr_binc = (u_char *)(void *)(igp->igr_kbgd + mxy);
}

short ssqcalc(INTEGRUN *igp)
{ /* This determines the squared norm of the diffraction vector
	and returns 10000 * that value */
	double	lv0, lv1, lv2, exc[3];

	if (EOK != ixy2ev(igp->igr_xcen, igp->igr_ycen, exc)) return 0;
	lv0 = 1. + exc[0];	lv1 = exc[1];	lv2 = exc[2];
	return (short)((lv0 * lv0 + lv1 * lv1 + lv2 * lv2) * 1.e4 + 0.49999);
}

/* end of integutil.c */
