/* intrefine.c -- routines for on-the-fly refinements in integration

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	27 Dec 2006: major rewrite of remap_intdistort: decision-making now
	 internal to the routine; polynomial as well as regional remaps
	28 Nov 2006: update makeq21 call
	10 Jun 2006: eliminate unused variable bsf in upd_xtal; housekeeping
	 3 Jan 2006: handle set_logical correctly; flags to remap_process fixed
		unnecessary zeroing avoided in upd_xtal
	15 Oct 2005: eliminate some static globals
	 5 Oct 2005: housekeeping
	13 Aug 2005: correct weighting factors in overall residual in resid()
	 8 Aug 2005: increase max refinables per region from 30 to 90
	 6 Aug 2005: preflags arg for remap_intdistort; distinguishes cases
	 5 Aug 2005: remap_process() now takes "dowrite" argument
	28 Jun 2005: remap_intdistort() and related functions added
	 7 Mar 2005: spread out formatting of (X,Y,phi) RMSerr values
	20 Jan 2005: clean up use of closed-form refinement and its diagnostics
	17 Jan 2005: increase CYCLESPERPRINT&CYCLESPERCHECK from 160 to 256;
	 use ra_{nrref,rmx,rmy,rmz} elements of REGIONATT structures;
	 tighten up diagnostic outputs after refinements
	 9 Sep 2004: eliminate unnecessary operations in endintref();
		in make_svecs(), the makeq21() call uses a local q21 matrix
	26 Aug 2004: routines used only here are declared as static; cosmetics
	 4 Apr 2004: bugfix in make_svecs(): subtract xp->xx0, not xp->xy0;
		add closed-form least-squares refinement as experiment
	24 Jan 2004: a few static globals converted to locals
	23 Jan 2004: correct management of output UPARAMS files
	21 Jan 2004: license message inserted and initial comments reorganized
	 5 Jan 2004: while writing out on-the-fly refinement UPARAMS files,
		get the transpositions, etc. of the matrices right
	 3 Jan 2004: globals replaced by elements of INTEGRUN structures;
		some routines require additional arguments as a result
	13 Dec 2003: eliminate unused static global variable `nused'
	22 Nov 2003: replace global variables with members of INTEGRUN struct;
		correct bug in resid() in calculating rphi[]
	 8 Nov 2003: don't calc ar00,etc. here; that's done in getrefls();
		calculate rphi[] instead of ome, chi, phi.
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	14 Aug 2003: eliminate unused storage of qq00, qq01, etc.
	17 Apr 2003: rename `frameno' to `fmno'
	 5 Jan 2003: reorganized source-file header
	15 Nov 2002: extern declarations made here rather than in integext.h
	 1 Nov 2002: altered REGIONATT structures for region attributes
	25 Mar 2002: removed MSI attachments
	15 Mar 1998: in endintref(), flush the logfile output after refinement
	 5 Dec 1995: enlarge `newname' vector to handle long directory names
	 2 Jun 1995: declare some residual-related variables as doubles
	31 May 1995: "register" -> "register int" throughout
	12 Apr 1995: replace s_char with sichar
	 2 Mar 1995: diagnostics to stdout; add frame # arg to pix2cm, cm2pix
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 1 Dec 1994: diagnostic output allows for detector (X,Y) > 999
	 5 Oct 1994: use xgwrite to output REFPARS data
	14 Jun 1994: ar(ij), qq(ij) scalars set after refinement
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	 7 Dec 1993: initial X-GEN version
	18 Jan 1988: in endintref(), rotate areal[][] back to the
		data start position, since xareal is at (0,0,0)

   Purpose: on-the-fly refinement of crystal and detector parameters
	during integration of Bragg data.
   Cross references: ../xgrefine/nonlinear.c

*************************************************************************/
#define		NEED_TIME	1
#include	"integdefs.h"
#define		ANGLEDEL	0.1	/* starting deviation of angles */
#define		LENGTHDEL	(0.001) /* "          " ratio for lengths*/
#define		LIMITRAT	5	/* see its use */
#define		DIVR		(0.2)	/* how much to shrink delta by */
#define		ITER		3	/* # of iterations to perform */
#define		NXTALPARS	9	/* number of crystal parameters */
#define		NPARS		(NXTALPARS + 5)
#define		CYCLESPERPRINT	256	/* print summary this often */
#define		CYCLESPERCHECK	256	/* give up if flat for 256 cycles */
#define		BASICALLY_FLAT	1.003	/* old/new <= this means we quit*/
#define		MINSIGMA	1	/* smallest legal value of sigma */
#define		MIN_IN_REGION	5	/* smallest # obs to adjust with */
#define		PHIWEIGHT	0.2	/* weight factor assoc. with phi resid*/
#define		HWEIGHT		0.5	/* weight assoc. with integerness */
#define		XYWEIGHT	(1. - PHIWEIGHT - HWEIGHT)
#define CIR(x)	((sichar)(xp->xisref[x]))

 typedef struct {
		int	irf_quitnl, irf_fmstart, irf_fmend, irf_navailable;
		int	irf_ndetec_ref, irf_nrescalc, irf_nusable;
		double	irf_resx0, irf_pnew0[NPARS];
		double	irf_pnow0[NPARS], irf_bestsofar[NPARS];
		double	irf_bestrs, irf_lastbest, irf_besthres;
		double	irf_bestphires, irf_bestxres, irf_bestyres;
		} INTREFWORK;

static int calc_xyphi(double *outx, INTEGRUN *igp)
{ /*	This computes the calculated (X,Y,phi) position of the reflection
	pointed to by `igp->igr_reflp'.
	It returns 0 if and only if the reflection has a calculable (X,Y,phi).
	The vector outx on output will contain the DIFFERENCES
	between observed and calculated (X,Y,phi) values. */
	int		i, j, jfr, quadr;
	double		beta, cxcm, cycm, cphi0, cphi1, diff0, diff1;
	double		fac, fxpr, fypr, refphi, rh, rsz, wden;
	double		xisq, xw, y2, yw, zw, zetasq, zz, rs[3];
	float		opxx, opxy;
	REFPARS		*xp;
	OBSREF		*lrp;

/* first get the reference-position scattering vector */
	lrp = igp->igr_reflp;	xp = igp->igr_xp;
	for (rs[0] = rs[1] = rs[2] = i = 0; i < 3; i++)
		for (rh = lrp->obs_hkl[i], j = 0; j < 3; j++)
			rs[j] += xp->xlambda * xp->xrec[j][i] * rh;
/* get the two possible intersections of that vector with the Ewald sphere */
	if (DFUZZ >= (xisq = rs[0] * rs[0] + rs[1] * rs[1])) return 2;
	if (1 < (zetasq = rs[2] * rs[2])) return 3;
	rsz = 0.5 * (xisq + zetasq);
	if (igp->igr_miny2sq > (y2 = xisq - rsz * rsz)) return 4;
	y2 = sqrt(y2); beta = atan2(y2,rsz);
	refphi = atan2(-rs[0],rs[1]); /* get phi value */
	fac = 1. / sqrt(1.- zetasq); zw = (1. - rsz) * fac; zz = y2 * fac;
	fac *= (rs[2] * xp->xplated);
	cphi0 = refphi + beta + PIO2; if (cphi0 > PI) cphi0 -= PIX2;
	cphi1 = refphi - beta + PIO2; /* this value already in range */
		/* omega centroid: */
	refphi = pirange((lrp->obs_omob + 0.5) * xp->xfrmsize);
	if ((diff0 = fabs(cphi0 - refphi)) > PI) diff0 = PIX2 - diff0;
	if ((diff1 = fabs(cphi1 - refphi)) > PI) diff1 = PIX2 - diff1;
	if (diff0 < diff1)	/* first intersection is closer: use it */
	  {
		refphi -= cphi0;	wden = zw * xp->xcsw - zz * xp->xssw;	
		xw = xp->xplated * (zz * xp->xcsw + zw * xp->xssw) / wden;
	  }
	else	/* second intersection is closer: use it */ 
	  {
		refphi -= cphi1;	wden = zw * xp->xcsw + zz * xp->xssw;
		xw = xp->xplated * (-zz * xp->xcsw + zw * xp->xssw) / wden;
	  }
	yw = fac / wden;
	outx[3] = cxcm = xw * xp->xctilt - yw * xp->xstilt + xp->xx0;
	outx[4] = cycm = xw * xp->xstilt + yw * xp->xctilt + xp->xy0;
	if (refphi > PI) refphi = PIX2 - refphi;
	else if (refphi < -PI) refphi = PIX2 + refphi;
	outx[2] = refphi / xp->xfrmsize;	jfr = outx[2] + 0.49999;
	igp->igr_fxpr = fxpr = ((double)lrp->obs_xob) / igp->igr_xscal;
	igp->igr_fypr = fypr = ((double)lrp->obs_yob) / igp->igr_yscal;
	quadr = detregion(fxpr, fypr, igp->igr_spl);
	if (1 != cm2pixq(cxcm, cycm, quadr, &opxx, &opxy, jfr))
		return 1;	/* can't compute output centroid */
	outx[0] = (double)(lrp->obs_xob) / igp->igr_xscal - opxx;
	outx[1] = (double)(lrp->obs_yob) / igp->igr_yscal - opxy;
	if (fabs(outx[2]) > xp->xerrphi || fabs(outx[1]) > xp->xerry ||
		fabs(outx[0]) > xp->xerrx) return -1;
	return 0;
}

static void dev_regions(INTEGRUN *igp)
{ /*	This computes the mean deviations of (Xob,Yob,Phiob) from
	(Xc,Yc,Phic) for the data from each mask.  These values will be
	reapplied to the predicted (Xc,Yc,Phic) on the next prediction.
	We also calculate the overall RMS error in (X,Y) and return that
	as igp->igr_overxy */

	register int	region, nchanged;
	double		swt, sxy, wt, offden[NREGIONS], xyp[5];
	OBSREF		*lrp;
	REGIONATT	*bmp;

	for (region = 0, bmp = igp->igr_bmask;
	 region < NREGIONS; region++, bmp++)
	   {
		offden[region] = bmp->ra_offx = bmp->ra_offy =
			bmp->ra_offphi = bmp->ra_nrref = 0.;
		bmp->ra_rmx = bmp->ra_rmy = bmp->ra_rmz = 0;
	   }
	swt = sxy = 0.;
	for (lrp = igp->igr_rfbase, nchanged = 0; lrp < igp->igr_rflast; lrp++)
	   {
		igp->igr_reflp = lrp;
	   	if (lrp->obs_flags == I_ALREADYOUT && (!calc_xyphi(xyp, igp)))
		  {
			wt = lrp->obs_gof;	region = lrp->obs_region;
			bmp = igp->igr_bmask + region;
			(bmp->ra_nrref)++;
			offden[region] += wt;	swt += wt;
			bmp->ra_offx += wt * xyp[0];
			bmp->ra_offy += wt * xyp[1];
			bmp->ra_offphi += wt * xyp[2];
			bmp->ra_rmx += wt * (xyp[0] * xyp[0]);
			bmp->ra_rmy += wt * (xyp[1] * xyp[1]);
			sxy += wt * (xyp[0] * xyp[0] + xyp[1] * xyp[1]);
			bmp->ra_rmz += wt * (xyp[2] * xyp[2]);
			nchanged++;
		  }
	     }
	if (swt > DFUZZ) igp->igr_overxy = sqrt(sxy / swt);
	else igp->igr_overxy = 0.;
	for (region = 0, bmp = igp->igr_bmask;
	 region < NREGIONS; region++, bmp++)
	   {
		if (((wt = offden[region]) > DFUZZ) &&
			(bmp->ra_nrref > MIN_IN_REGION))
		  {
			bmp->ra_offx /= wt;	bmp->ra_offy /= wt;
			bmp->ra_offphi /= wt;
			bmp->ra_rmx = sqrt(flimits((bmp->ra_rmx / wt) -
				bmp->ra_offx * bmp->ra_offx, 0., 9999.));
			bmp->ra_rmy = sqrt(flimits((bmp->ra_rmy / wt) -
				bmp->ra_offy * bmp->ra_offy, 0., 9999.));
			bmp->ra_rmz = sqrt(flimits((bmp->ra_rmz / wt) -
				bmp->ra_offphi * bmp->ra_offphi, 0., 9999.));
		  }
	      else {
			bmp->ra_offx = bmp->ra_offy = bmp->ra_offphi = 0;
			bmp->ra_rmx = bmp->ra_rmy = bmp->ra_rmz = 0;
		   }
	   }
}

static void record_cycle(FILE *fp, INTREFWORK *irp)
{ /*	This prints a one-line summary of the results of a cycle of
	refinement to the stream `fp'. */
	double	*ffp;
	if (irp->irf_nrescalc == 1) fprintf(fp, "C.F.");
	else fprintf(fp, "%4d", irp->irf_nrescalc);
	ffp = irp->irf_bestsofar;
	fprintf(fp, " %6.2f %6.2f %6.2f %6.2f %6.2f %6.2f%7.2f%7.2f%7.2f ",
	 *ffp, *(ffp + 1), *(ffp + 2), *(ffp + 3), *(ffp + 4),
	 *(ffp + 5), *(ffp + 6), *(ffp + 7), *(ffp + 8));
	fprintf(fp, "%5.2f%6.3f%6.3f%6.3f%6.2f %5.0f%5.0f %5.0f %5.0f %5.0f\n",
	 *(ffp + 9), *(ffp + 10), *(ffp + 11), *(ffp + 12),
	 *(ffp + 13), sqrt(irp->irf_bestrs) * 1.e4,
	 sqrt(irp->irf_besthres) * 1.e4, sqrt(irp->irf_bestxres) * 1.e4,
	 sqrt(irp->irf_bestyres) * 1.e4, sqrt(irp->irf_bestphires) *1.e4);
}

static int endintref(INTEGRUN *igp, INTREFWORK *irp)
{ /*	This neatly finishes off a run of `hminpar' */
	register int	i, j;
	char		*cp, *oldupname, newname[256];
	static char	*shortname;
	float		lrot[3];
	double		*ffp;
	REFPARS		*xp;
	static REFPARS	locr;

/* first, keep the best values of the refinable parameters */
	xp = igp->igr_xp;
	for (i = 0, ffp = irp->irf_pnow0; i < 3; i++) xp->xabc[i] = *ffp++;
	for (i = 0; i < 3; i++) xp->xang[i] = *ffp++;
	xp->xdelom = *ffp++;	xp->xdelchi = *ffp++;	xp->xdelphi = *ffp++;
	xp->xplated = *ffp++;	xp->xx0 = *ffp++;	xp->xy0 = *ffp++;
	xp->xtilt = RADC * (*ffp++);	xp->xswing = RADC * (*ffp++);
	xp->xpltd2 = xp->xplated * xp->xplated;
	if (vecmat(xp->xabc, xp->xang, xp->xareal)) return EBADVALUE;
	lrot[0] = -RADC * xp->xdelom;	lrot[1] = -RADC * xp->xdelchi;
	lrot[2] = -RADC * xp->xdelphi;
	rot9r(xp->xareal, lrot);	/* rotate to (0,0,0) of goniostat */
	for (i = 3; i--; ) for (j = 3; j--; ) xp->xrec[i][j]= xp->xareal[i][j];
	mm3(xp->xareal, xp->xrec, igp->igr_r0tos); /* rot to run start */
	if (inv3(xp->xareal, xp->xrec)) return ESINGULAR;
	/* scale the real-space version by the X-ray wavelength */
	for (i = 3; i--; ) for (j = 3; j--; ) xp->xareal[i][j] /= xp->xlambda;
	record_cycle(igp->igr_fpl, irp);	record_cycle(stdout, irp);
	/* compute the mean deviations of the observed X, Y, and phi from the
	new predicted values thereof--separately for each mask.
	These mean deviations will be reapplied to the data from same mask
	in the next prediction. */
	dev_regions(igp);
/* construct the name of the new parameters file. */
	shortname = filepart(trname("uparams"), 2);
#ifdef VMS	 /* VMS: WORK:namenum.upr */
	if (irp->irf_fmend >= 0)
	 sprintf(newname, "WORK:%6s%d.upr", shortname, irp->irf_fmend);
	else sprintf(newname, "WORK:%6sM%d.upr", shortname, -(irp->irf_fmend));
#else		/* Unix: $PARM/namenum.upr */
	if (NULL == (cp = getenv("PARM")) || '\0' == *cp)
	  {
		if (NULL == (cp = getenv("WORK")) || '\0' == *cp) cp = "./";
	  }
	if (irp->irf_fmend >= 0)
	 sprintf(newname, "%s/%.6s%d.upr", cp, shortname, irp->irf_fmend);
	else sprintf(newname, "%s/%.6sm%d.upr",
		cp, shortname, -(irp->irf_fmend));
#endif
	/* write out a new version of the parameters file */
	/* ... first, save the current value of the environment variable */
	oldupname = getenv("UPARAMS");
	/* then write the current parameters to the file with the funny name.
	 We make a copy of the parameters so the mangling that happens
	 to refparout() doesn't affect our operational params */
	cpymem((void *)&locr, (void *)xp, sizeof (REFPARS));
	if (1 <= set_logical("UPARAMS", newname)) (void)refparout(&locr);
	/* then restore the parameter filename, assuming it looks okay */
	if ((NULL != oldupname) && ('\0' != *oldupname))
		(void)set_logical("UPARAMS", oldupname);
	igp->igr_tl1 = NOW;
	/* users like to see the output at this point */
	xeflush(igp->igr_fpl);	if (igp->igr_fpl != stdout) xeflush(stdout);
	return EOK;
}

static int vfy_refl(double rx, double ry, INTEGRUN *igp)
{ /*	This checks to see whether `igp->igr_reflp' satisfies the criteria
	for a refinable reflection.  It must be reasonably bright,
	not too close to the edge of the detector, have a legal sigma,
	and (as of V. 0.6) satisfy the error criteria for
	refinement given in the `refine' menu.  If it passes all these
	tests, the routine returns 1; otherwise, an appropriate flag is
	set in the `igp->igr_reflp' structure and the routine returns 0. */
	register int	j;
	int		nbadob;
	float		*sl2;
	double		iovsig, thisdh, calc_hkl[3], xyp[5];
	REFPARS		*xp;
	OBSREF		*lrp;

	lrp = igp->igr_reflp;	xp = igp->igr_xp;
	sl2 = (float *)(void *)&(lrp->obs_lowfrm);
	if (rx < DETECTOR_EDGE || rx > xp->xdetwid - DETECTOR_EDGE ||
	  ry < DETECTOR_EDGE || ry > xp->xdetht - DETECTOR_EDGE)
		lrp->obs_flags |= I_PEAKONEDGE;
	else if (lrp->obs_sigma < MINSIGMA) lrp->obs_flags |= I_TINYSIGMA;
	else if ((iovsig = lrp->obs_ii / sqrt(lrp->obs_sigma)) <
	 igp->igr_iosmin) lrp->obs_flags |= I_TOOWEAK;
	else /* make sure the indices and x,y,phi values are reasonable */
	  {
		lrp->obs_gof = 1000. * log(iovsig);
		for (j = 3, nbadob = 0; j--; )
		   {
			calc_hkl[j] = xp->xareal[j][0] * lrp->obs_iunfit +
			 xp->xareal[j][1] * lrp->obs_sunfit +
			 xp->xareal[j][2] * (*sl2);
			thisdh = calc_hkl[j] - lrp->obs_hkl[j];
			if (fabs(thisdh) > xp->xerrind) nbadob++;
		   }
		if(calc_xyphi(xyp, igp) || nbadob)
			lrp->obs_flags |= I_MOVETOOFAR;
	    }
	return (lrp->obs_flags == I_ALREADYOUT);
}

static int accum_irclosed(INTEGRUN *igp, float cl0[3][3], float cl1[3][3])
{ /*	This accumulates the relevant sums for closed-form refinement */
	int		i, j;
	float		*sl2;
	double		wdi, lls[3];
	static REFPARS	*xp;
	static OBSREF	*lrp;

	lrp = igp->igr_reflp;	xp = igp->igr_xp;
	if (fabs(lrp->obs_xob - lrp->obs_xpred) > xp->xerrx * igp->igr_xscal)
		return 0;
	if (fabs(lrp->obs_yob - lrp->obs_ypred) > xp->xerry * igp->igr_yscal)
		return 0;
	if ((wdi = lrp->obs_ii / (sqrt(lrp->obs_sigma) *igp->igr_iosmin)) <=1.)
		return 0;
	lls[0] = lrp->obs_iunfit;	lls[1] = lrp->obs_sunfit;
	sl2 = (float *)(void *)&(lrp->obs_lowfrm);	lls[2] = *sl2;
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
	   {
		cl0[i][j] += wdi * lls[j] * (float)lrp->obs_hkl[i];
		cl1[i][j] += wdi * lls[j] * lls[i];
	   }
	return 1;
}

static int cllsq(INTEGRUN *igp)
{ /*	This does a closed-form least-squares refinement based on the sums
	found in igp->igr_plbg[i][j]. It returns 1 if successful, 0 if not */
	int	cff, i, j, miref;
	double	cl0[3][3], cl1[3][3], cl2[3][3];
	REFPARS	*xp, tempr;

	miref = igp->igr_plbg[6][0];
	if (miref <= 9)
	  {
		fprintf(igp->igr_fpl,
 "Closed-form refinement not attempted: only %d reflections available\n",
			miref);	return 0;
	  }
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
	   {
		cl0[i][j] = igp->igr_plbg[i][j];
		cl1[i][j] = igp->igr_plbg[i+3][j];
	   }
	if (0 != (cff = dinv3(cl1, cl2)))
	  {
		fprintf(igp->igr_fpl,
 "Closed-form refinement failed: normal matrix not invertible, code %d\n",cff);
		return 0;
	  }
	dmm3(cl1, cl0, cl2);	/* solve least-squares problem */
	xp = igp->igr_xp;
	cpymem((void *)&tempr, (void *)xp, sizeof (REFPARS));
	/* copy resulting real-space cell into temporary REFPARS */
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		tempr.xareal[i][j] = cl1[i][j];
	/* clean that up according to standard rules and copy back */
	if (EOK != (i = rp2vecs(&tempr)))
	  {
		fprintf(igp->igr_fpl,
		 "Closed-form refinement failed: rp2vecs returned %d\n", i);
		return 0;
	  }
	cpymem((void *)xp, (void *)&tempr, sizeof (REFPARS));
	igp->igr_plbg[6][2] = 1;	/* signal that this worked */
	return 1;
}

static int make_svecs(int first_call, INTEGRUN *igp, INTREFWORK *irp)
{ /*	This defines the scattering vectors s[3] for a each refl to be refined
	and sticks the components in slots `obs_iunfit', `obs_sunfit', and
	`obs_lowfrm'.  These values depend only on the detector parameters and
	and the observed (x,y,phi) centroid, and they will be used in the
	course of refinement.  If the argument `first_call' is true, the
	routine also does some pre-screening of	the data by putting a flag
	up if a reflection is too weak or has a	faulty sigma.
	Also, if first_call is true, any region (one of the twelve)
	contributing more than 90 refls has some contributors removed. */
	register int	i, j;
	int		jfr;
	double		aphi, cphi, rni, rni2, rx, ry, sphi;
	float		rxcm, rycm, *sl2;
	float		clsq0[3][3], clsq1[3][3], sl[3], sr[3];
	REFPARS		*xp;
	OBSREF		*lrp;
	static int	minref, nureg[NREGIONS];

	xp = igp->igr_xp;
	if (vecmat(xp->xabc, xp->xang, xp->xareal)) return 10007; /*unit cell*/
	sl[0] = (-RADC) * xp->xdelom;	sl[1] = (-RADC) * xp->xdelchi;
	sl[2] = (-RADC) * xp->xdelphi;
	rot9r(xp->xareal, sl);
	for (i = 3; i--; ) for (j = 3; j--; ) xp->xrec[i][j]= xp->xareal[i][j];
	if (first_call) ztos(xp->xstart, igp->igr_r0tos);
		/* ... and rotate to start position */
	mm3(xp->xareal, xp->xrec, igp->igr_r0tos);
	if (inv3(xp->xareal, xp->xrec)) /* invert to get realspace matrix */
		return 10002;
		/* scale the real-space version by the X-ray wavelength */
	for (i = 3; i--; ) for (j = 3; j--; ) xp->xareal[i][j] /= xp->xlambda;
	/* if this isn't the first call, AND we aren't refining any
	detector parameters, we're done */
	if ((!first_call) && (!(irp->irf_ndetec_ref))) return 0;
	xp->xpltd2 = xp->xplated * xp->xplated;
	makeq21(xp);
	if (first_call)
	  {
		for (i = NREGIONS; i--; ) nureg[i] = 0;
		for (i = 0; i < 3; i++)
		  {
			for (j = 0; j < 3; j++)
			   {
				clsq0[i][j] = clsq1[i][j] = 0.;
				igp->igr_plbg[i][j]= igp->igr_plbg[i+3][j] = 0;
			   }
			igp->igr_plbg[6][i] = 0;
		  }
		minref = 0;
	  }
	for (lrp = igp->igr_rfbase, irp->irf_navailable = irp->irf_nusable = 0;
	 lrp <= igp->igr_rflast; lrp++)
	   if (lrp->obs_flags == I_ALREADYOUT)
	     {
		igp->igr_reflp = lrp;
		aphi = pirange(xp->xfrmsize * (lrp->obs_omob + 0.5));
		rx = (double)lrp->obs_xob / igp->igr_xscal; /*X obs in pixels*/
		ry = (double)lrp->obs_yob / igp->igr_yscal; /*Y obs in pixels*/
		jfr = lrp->obs_omob + 0.49999;
		if (rx < 1 || ry < 1 || rx > xp->xdetwid - 2 ||
		 ry > xp->xdetht - 2 || 1 != pix2cm(rx, ry, &rxcm, &rycm, jfr))
		  {
			lrp->obs_flags |= I_NOPIXTOCM; continue;
		  }
		rxcm -= xp->xx0;	rycm -= xp->xy0; /* mainbeam offset */
		rni2 = rxcm * rxcm + rycm * rycm + xp->xpltd2; rni = sqrt(rni2);
		sr[0] = xp->xplated / rni;
		sr[1] = rxcm / rni; sr[2] = rycm / rni;
	/* rotate this by the detector tilt and swing-angle matrices */
		mm31(sl, xp->xq21, sr); /* local scattering vector */
		sl[0] += 1.;	/* correct beam-direction component */
			/* finally, rotate by the observed phi centroid */
		cphi = cos(aphi);	sphi = sin(aphi);
		lrp->obs_iunfit = cphi * sl[0] - sphi * sl[1];
		lrp->obs_sunfit = sphi * sl[0] + cphi * sl[1];
		sl2 = (float *)(void *)&(lrp->obs_lowfrm);
		*sl2 = sl[2];
		if ((!first_call) || vfy_refl(rx, ry, igp))
		  {
			(irp->irf_nusable)++;	(irp->irf_navailable)++;
			if ((first_call) && (fabs(lrp->obs_omob -
			 lrp->obs_ompred) < xp->xerrphi))
				minref += accum_irclosed(igp, clsq0, clsq1);
		  }
	     }
	if (!first_call) return (irp->irf_nusable < 1);
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++) /* store lsq sums */
	   {
		igp->igr_plbg[i][j] = clsq0[i][j];
		igp->igr_plbg[i+3][j] = clsq1[i][j];
	   }
	igp->igr_plbg[6][0] = minref;
	igp->igr_plbg[6][1] = irp->irf_nusable;
	/* don't even try to refine unless we have at least 2M spots,
	where M is the number of refinable parameters.  We require
	12 spots even if M < 6. */
	for (minref = i = 0; i < 14; i++)
	   if ((xp->xisref[i] == 'Y') || (xp->xisref[i] == 'y')) minref += 2;
	if (minref < 12) minref = 12;
	if (irp->irf_nusable < minref)
	  {
		fprintf(igp->igr_fpl,
		 "refinement failed: only %d refinable reflections, need %d\n",
			irp->irf_nusable, minref);
		return 2;
	  }
	/* get rid of refls we don't need in refinement.  Note that we
	keep the last ones collected, since they'll be most relevant
	in subsequent portions of the data */
	if (irp->irf_nusable > 500)
	  {
		for (lrp = igp->igr_rflast; lrp >= igp->igr_rfbase; lrp--)
		  if ((lrp->obs_flags == I_ALREADYOUT) &&
			(++(nureg[(int)lrp->obs_region])) > 90)
		    {
			lrp->obs_flags |= I_TOOWEAK;	(irp->irf_nusable)--;
		    }
		igp->igr_reflp = lrp;	/* re-set this pointer after that */
	  }
	return (irp->irf_nusable < minref);
}

static double resid(INTEGRUN *igp, INTREFWORK *irp)
{ /*	This returns an rms residual for a set of refinable reflections.
	For this application (inside `integrate') we compute the equivalent
	of an "A" refinement, i.e. the returned residual is
		R = Wp * Rp + Wh * Rh + (Rx + Ry) * (1 - Wp - Wh)
	where Rp, Rh, Rx, and Ry are rms residuals on phi, integerness,
	X, and Y, respectively, and Wp and Wh are weight factors associated
	with the phi and integerness residuals.  Each of the residuals is
		SUM (wi^2)*Di
		 i
	R   =   -----------------
		SUM (wi^2)
		 i
	where Di = (Mi-mi)^2 in one case, (hi-[hi])^2 in another, and
	(pi(Mi-mi)^2 + qi(hi-[hi])^2 + ri(Xi-xi)^2 + ri(Yi-yi)^2, with
	Mi,Xi,Yi the observed phi,X,Y values, and mi,xi,yi the computed ones;
	[hi] is the integerized version of hi and pi,qi,ri are weights.
	The weight factors wi are currently 0 for illegal refls and 1 for
	legals.  `nok' is the number of observations found to be legal.
	Note that we don't compute the hkl using `geth' here; the integerized
	version of hkl is assumed to be correct before we start cycling.
	See the "defines" at the top of this file for values of Wp, Wh. */
	register int	i, j, nok;
	register double	*zp;
	float		*sl2;
	double		dhok, difh, dok, resh, resphi, resx, resy;
	double		retres, rwt, thisdh, *zp0, xyp[5];
	REFPARS		*xp;
	OBSREF		*lrp;
	static double	resh0, resphi0, resy0;

	/* Impose equivalences between parameters of various sorts */
	xp = igp->igr_xp;	zp0 = irp->irf_pnew0;
	if (xp->xisys > 3)
	  {
		*(zp0 + 1) = *zp0;
		if (xp->xisys == 5) *(zp0 + 2) = *zp0;
		else if (xp->xisys == 7)
		  {
			*(zp0 + 2) = *zp0; *(zp0 + 3) = *(zp0 + 4) = *(zp0 + 5);
		  }
	  }
	/* now turn the calling values into crystal and detector parameters */
	for (i = 0, zp = zp0; i < 3; i++, zp++) xp->xabc[i] = *zp;
	for (i = 0; i < 3; i++, zp++) xp->xang[i] = *zp;
	xp->xdelom = *zp++;	xp->xdelchi = *zp++;	xp->xdelphi = *zp++;
	xp->xplated = *zp++;	xp->xx0 = *zp++;	xp->xy0 = *zp++;
	xp->xtilt = *(zp++) * RADC;	xp->xswing = *(zp++) * RADC;
		/* scattering vectors of new refls */
	if (make_svecs(0, igp, irp)) return 10000;
	zp = zp0;
	for (resphi = resx = resy = resh = dok = dhok = nok = 0,
	 lrp = igp->igr_rfbase; lrp <= igp->igr_rflast; lrp++)
	  if (lrp->obs_flags == I_ALREADYOUT)
	   {
		igp->igr_reflp = lrp;	rwt = lrp->obs_gof;
		sl2 = (float *)(void *)&(lrp->obs_lowfrm);
		for (j = 3, difh = 0; j--; )
		   {
			thisdh = xp->xareal[j][0] * lrp->obs_iunfit +
			 xp->xareal[j][1] * lrp->obs_sunfit +
			 xp->xareal[j][2] * (*sl2) - lrp->obs_hkl[j];
			difh += thisdh * thisdh;
		   }
		resh += difh * rwt; /*increment numerator for h resid*/
		dhok += 3. * rwt;	/* denominator for h resid */
		if (!calc_xyphi(xyp, igp)) /* get computed x,y,phi */
		  {
			resx += xyp[0] * xyp[0] * rwt;
			resy += xyp[1] * xyp[1] * rwt;	dok += rwt;
			resphi += xyp[2] * xyp[2] * rwt;	nok++;
		  }
	   }
	if (!nok) return 10000.;	/* awful residual if no refls */
	resh /= dhok;	resphi /= dok;	resx /= dok;	resy /= dok;
	if (irp->irf_resx0 < -98.)	/* record initial residuals */
	  {
		resh0 = resh; resphi0 = resphi;
		irp->irf_resx0 = resx; resy0 = resy;
	  }
	/* rescale the x, y, phi residuals appropriately */
	resphi *= resh0 / resphi0; resx *= resh0 / irp->irf_resx0;
	resy *= resh0 /resy0;
	/* correction and alteration 13 Aug 05:
	 highest weighting goes to integerness residual, not phi residual */
	retres = HWEIGHT * resh + PHIWEIGHT * resphi + XYWEIGHT * (resx + resy);
	/* retres = PHIWT * resh + HWT * resphi + XYWT * (resx + resy); */
	if (retres < irp->irf_bestrs) /*record best residuals if this is best*/
	  {
		for (i = NPARS; i--; ) irp->irf_bestsofar[i] = zp0[i];
		irp->irf_bestrs = retres;	irp->irf_besthres = resh;
		irp->irf_bestphires = resphi * resphi0 / resh0;
		irp->irf_bestxres = resx * irp->irf_resx0 / resh0;
		irp->irf_bestyres = resy * resy0 / resh0;
	  }
	if (((irp->irf_nrescalc == 1) && (igp->igr_plbg[6][2] > 0))
	 || !(irp->irf_nrescalc % CYCLESPERPRINT)) /* print results*/
	  {
		record_cycle(igp->igr_fpl, irp); record_cycle(stdout, irp);
		if (!((irp->irf_nrescalc) % CYCLESPERCHECK))
		  {
			if (irp->irf_lastbest <=
			 irp->irf_bestrs * BASICALLY_FLAT)
			  {
				for (i = 0, zp = irp->irf_pnow0; i < NPARS;
				   i++, zp++) *zp = irp->irf_bestsofar[i];
				irp->irf_quitnl = 1;
				if (EOK != endintref(igp, irp)) return 10000.;
			  }
			irp->irf_lastbest = irp->irf_bestrs;
		  }
	  }
	(irp->irf_nrescalc)++;	return retres;
}

static void refdisp_hdr(FILE *fp, INTEGRUN *igp, INTREFWORK *irp)
{ /*	Set up a header for the display of the results of refinement cycles */
	REFPARS	*xp;

	xp = igp->igr_xp;
	fprintf(fp,
	 "On-the-Fly Refinement over frames %4d to %4d;",
		irp->irf_fmstart, irp->irf_fmend);
	fprintf(fp, "%7d reflections examined,%7d available,%7d used\n",
	 1 + (int)(igp->igr_rflast - igp->igr_rfbase),
		irp->irf_navailable, irp->irf_nusable);
	fprintf(fp," cyc    Unit Cell Lengths     Unit Cell Angles");
	fprintf(fp,"   Orientation Angles detec");
	fprintf(fp,"  Geometrical Parameters   RMS residuals * 10000\n");
	fprintf(fp," le#      a      b      c  alpha   beta  gamma");
	fprintf(fp,"  omega    chi    phi  dist");
	fprintf(fp,"  Xcen  Ycen  tilt swing total    h     X     Y     Z\n");
	fprintf(fp,"Ref?      %c      %c      %c      %c      %c      %c    ",
		CIR(0), CIR(1), CIR(2), CIR(3), CIR(4), CIR(5));
	fprintf(fp,"  %c      %c      %c     %c     %c     %c     %c     %c\n",
	 CIR(6), CIR(7), CIR(8), CIR(9), CIR(10), CIR(11), CIR(12), CIR(13));
}

static void xtaltovec(REFPARS *lxp, double *lve)
{ /*	Convert the crystal parameters in a REFPARS structure to a vector */
	double	*bsf;

	bsf = lve;
	*bsf++ = lxp->xabc[0];	*bsf++ = lxp->xabc[1];	*bsf++ = lxp->xabc[2];
	*bsf++ = lxp->xang[0];	*bsf++ = lxp->xang[1];	*bsf++ = lxp->xang[2];
	*bsf++ = lxp->xdelom;	*bsf++ = lxp->xdelchi;	*bsf++ = lxp->xdelphi;
	*bsf++ = lxp->xplated;	*bsf++ = lxp->xx0;	*bsf++ = lxp->xy0;
	*bsf++ = lxp->xtilt / RADC;	*bsf++ = lxp->xswing / RADC;
}

int remap_intdistort(INTEGRUN *igp, int preflags)
{ /* This decides whether a distortion remapping is needed.
	If it is, it remaps the detector face in response to errors in X and Y.
	If igp->igr_remaplim is negative, we do this by polynomials:
	dX = a+b*X+c*Y+d*X*X+e*X*Y+f*Y*Y
	dY = g+h*X+i*Y+j*X*X+k*X*Y+l*Y*Y
	... computing the values of a through l by least-squares fits.
	If igp->igr_remaplim is positive, we do it by regional mapping.
	In both cases we invert the mapping and render everything
	self-consistent. We keep the modified mapping in memory,
	and if this is the first call, we write the modified VSPLINE out.
	The routine's return values are:
	0	success
	-1	failure to reserve memory for remapping
	1	failure to open new USPLINE file
	2	failure to write out new USPLINE file header
	3	failure to write main body of USPLINE file
	4	failure to determine main-beam position
	33	didn't need to do a remapping in the first place
	The `preflags' argument is less than 2 if this is the initial pass
	through the data, so that the remapping DOES get written out to
	disk and described in detail in the logfile. If `preflags' is >= 2,
	then this is a later pass through the data, for which the
	rebuilt spline is NOT written to disk. */
	int		ier, nusable, reflags;
	double		delp, delpden, scx, scy, xyp[11];
	double		*adel;
	static VSPLINE	*lsp;
	OBSREF		*lrp;

	/* if we don't need a remapping, return a code. There isn't really
	 a seek-error involved here: it just needs something other than EOK. */
	if ((igp->igr_remaplim > 9999999.) ||
	 (fabs(igp->igr_remaplim) > 10000. * igp->igr_overxy)) return ESEEKERR;
	/* before each call after the first one,
	 we re-read the version of the spline file that is on disk
	 so we don't get too far from the initial values */
	if (preflags > 2)
	  {
		if (NULL == (lsp = gspline(5, &ier))) lsp = igp->igr_spl;
		else igp->igr_spl = lsp;
	  }
	else	lsp = igp->igr_spl;
	reflags = (preflags & 02) ? 0 : 2;
	scx = igp->igr_xscal;	scy = igp->igr_yscal;
	if (igp->igr_remaplim > DFUZZ) /* regional remapping */
	  {
		if (NULL == (adel = remap_setup(lsp, &(xyp[8])))) return -1;
		/* go through the observational data, computing the shifts and
		weights for each reflection relative to each fiducial point */
		for (lrp = igp->igr_rfbase; lrp < igp->igr_rflast; lrp++)
		   {
			igp->igr_reflp = lrp;
		   	if (lrp->obs_flags == I_ALREADYOUT &&
			 (!calc_xyphi(xyp, igp)))
			  {
				xyp[5] = lrp->obs_gof;
				xyp[6] = ((double)(lrp->obs_xob)) / scx;
				xyp[7] = ((double)(lrp->obs_yob)) / scy;
				if (EOK != remap_sums(lsp, xyp, adel))
					return -1;
			  }
		   }
		igp->igr_reflp = igp->igr_rfbase;	/* so pointer is safe */
		return remap_process(lsp, adel, igp->igr_fpl, reflags);
	  }
	/* otherwise we're doing a polynomial remapping */
	inv_init(lsp->v_lax, lsp->v_ldx, 6); /* zero the vectors */
	inv_init(lsp->v_lay, lsp->v_ldy, 6);
	if ((delpden = (igp->igr_xp)->xerrphi) < 0.1) delpden = 0.1;
	nusable = 0;
	for (lrp = igp->igr_rfbase; lrp < igp->igr_rflast; lrp++)
	   { /* go through the refls and add to the polynomial sums */
		igp->igr_reflp = lrp;
		if (lrp->obs_flags == I_ALREADYOUT && (!calc_xyphi(xyp, igp)))
		  {
			delp = xyp[2] / delpden;
			if (DFUZZ < (xyp[10] = (fabs(delp) < 1.) ?
				cos(PIO2 * delp) : 0.))
			  {
				xyp[6] = ((double)(lrp->obs_xob)) / scx;
				xyp[7] = ((double)(lrp->obs_yob)) / scy;
				xyp[8] = xyp[6] - xyp[0];
				xyp[9] = xyp[7] - xyp[1];
				polyadd(lsp, &(xyp[6])); nusable++;
			  }
		  }
	   }
	if (nusable < 12) return EBAD; /* need >= 12 refls */
	/* now do the actual polynomial remapping and recalc errors */
	igp->igr_reflp = igp->igr_rfbase;	/* so pointer is safe */
	return polymap(lsp, igp->igr_fpl, reflags);
}

void upd_xtal(int fmno, INTEGRUN *igp)
{ /*	Optimization routine for refinement of parameters based on a
	minimization of the deviations from integerness of the reflections. */
	register int		i, nk;
	double			fnew, fnow, fold, t;
	double			*delta, *pmax, *pmin, *pnew, *pnow;
	double			delta0[NPARS], pmax0[NPARS], pmin0[NPARS];
	REFPARS			*xp;
	INTREFWORK		irs;
	static float		del[] = { 0.05,0.02,0.02,0.1,0.1 };

	igp->igr_tl0 = NOW;	/* save starting time of refinement operation*/
	xp = igp->igr_xp;
	zeromem((AVOID *)&irs, sizeof (INTREFWORK));
	irs.irf_fmstart = ((fmno - igp->igr_bunch) - igp->igr_zhalfw) - 1;
	if (irs.irf_fmstart < igp->igr_zmi)
		irs.irf_fmstart = igp->igr_zmi;
	irs.irf_fmend = fmno - igp->igr_zhalfw - 1;
	/* initialize some local variables out of the refinement globals */
	for (i = 9, irs.irf_ndetec_ref = 0; i < 14; i++)
		if (xp->xisref[i] == 'Y') (irs.irf_ndetec_ref)++;
		/* get scatt vectors of reference refls */
	if ((i = make_svecs(1, igp, &irs)))
	  {
		fprintf(igp->igr_fpl, "make_svecs returned %d\n", i);
		return;
	  }
	refdisp_hdr(igp->igr_fpl, igp, &irs);
	refdisp_hdr(stdout, igp, &irs);	/* set up headers */
	irs.irf_resx0 = -100.;	/* flag saying we haven't done any cycles yet */
	nk = ITER;	/* get the number of iterations to do */
	irs.irf_lastbest = irs.irf_bestrs = 20000.; /* ridiculous residual */
	xtaltovec(xp, irs.irf_pnow0);	/* initial parameters */
	pnow = &(irs.irf_pnow0[0]);	pnew = &(irs.irf_pnew0[0]);
	delta = &(delta0[0]);	pmax = &(pmax0[0]);	pmin = &(pmin0[0]);
	for (i = 0; i < NPARS; i++, pnew++, pnow++, delta++)
	   {
		*pnew = *pnow;
		*delta = (i < 3 ? LENGTHDEL * (*pnow) : (i < NXTALPARS ?
			ANGLEDEL : del[i-NXTALPARS]));
		*pmax++ = *pnow + LIMITRAT * (*delta);  /* max to move up */
		*pmin++ = *pnow - LIMITRAT * (*delta);	/*max to move down*/
	   }
	pnow = &(irs.irf_pnow0[0]);	pnew = &(irs.irf_pnew0[0]);
	delta = &(delta0[0]);	pmax = &(pmax0[0]);	pmin = &(pmin0[0]);
	fold = fnow = resid(igp, &irs); /* get initial residuals */
	if (cllsq(igp))	/* do closed-form least squares if possible */
	  {
		xp = igp->igr_xp; xtaltovec(xp, irs.irf_pnow0);
		xtaltovec(xp, irs.irf_pnew0);
		fold = fnow = resid(igp, &irs);
	  }
/* adjust the parameters, finding the minimum difference between predicted
    and observed values.  The overall loop is entered nk times.  */
	while (nk)
	     { /* move params one at a time up by delta and down by delta */
		for (i = 0; i < NPARS; i++, pnew++, pnow++,
			pmax++, pmin++, delta++) 
		 if (xp->xisref[i] == 'Y')
		   {
			*pnew += *delta;
			if (*pnew > *pmax) *pnew = *pmax;
			fnew = resid(igp, &irs);
			if (irs.irf_quitnl) return;
			if (fnew < fnow) fnow = fnew;
			else
			  {
				*pnew = *pnow - *delta;
				if (*pnew < *pmin) *pnew = *pmin;
				fnew = resid(igp, &irs);
				if (irs.irf_quitnl) return;
				if (fnew < fnow) fnow = fnew;
				else *pnew = *pnow;
			  }
		   }
		pnow = &(irs.irf_pnow0[0]);	pnew = &(irs.irf_pnew0[0]);
		delta = &(delta0[0]);	pmax = &(pmax0[0]); pmin = &(pmin0[0]);
/* if the current residual is better than the old one, we continue
   the refinement by checking some other values of x...  */
		if (fnow < fold)
		  {
		   while(1) /* move params to (twice current val - old val)*/
		      {
			for (i = 0; i < NPARS; i++,pnow++,pnew++,pmin++,pmax++)
			 if (xp->xisref[i] == 'Y')
			   {
				t = *pnow;	*pnow = *pnew;
				*pnew = *pnew + *pnew - t;
				if (*pnew > *pmax) *pnew = *pmax;
				if (*pnew < *pmin) *pnew = *pmin;
			   }
			pnow = &(irs.irf_pnow0[0]); pnew = &(irs.irf_pnew0[0]);
			pmax = &(pmax0[0]);	pmin = &(pmin0[0]);
			fnew = resid(igp, &irs);
			if (irs.irf_quitnl)	return;
			if (fnew >= fnow) break;
			fnow = fnew;
		      }
/* when that particular trick has exhausted its usefulness, store current
   param values in pnew and evaluate refinement progress */
		    for (i = 0; i < NPARS; i++, pnew++, pnow++)
			if (xp->xisref[i] == 'Y') *pnew = *pnow;
		    pnow = &(irs.irf_pnow0[0]);	pnew = &(irs.irf_pnew0[0]);
		    fold = fnow;
		  }
/* the residual is as good as we can get it for the current values of deltas.
   So we shrink them and loop through again.  */
		else {
			nk--;	for (i = 0; i < NPARS; i++, delta++)
				 *delta *= DIVR;
			delta = &(delta0[0]);
		     }
	     }
/* we've gone through the shrink loop as often as permitted, so
   we're finished.  Reconvert the "pnow" vbls to globals and finish off. */
	endintref(igp, &irs);	return;
}

/* end of intrefine.c */
