/* intrefl.c -- Routines that compute intensities and update 3D profiles

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	17 Jun 2006: replace act_refl and act_pixel calls with activmask calls
	 5 Jun 2006: xy2h and precenter now static; housekeeping
	 6 Oct 2005: housekeeping
	 2 Sep 2005: in intcalc(), turn off TEMP_VERBOSE as needed
		and report upper index limits correctly in moli[3]-[5]
	16 Aug 2005: MISSING_PIXEL concept is gone; eliminate references to it
	13 Jun 2005: background handled with single-level pointer, not double
	10 Jun 2005: full verbose info for chosen reflections in sum_refl()
	27 Aug 2004: routines used only here are declared as static; cosmetics
	12 Aug 2004: move 6 routines to integutil.c; eliminate static globals
	29 Jun 2004: statistics per frame (<I>, <I/sigma>, etc) in intcalc()
	24 Jan 2004: replace a few static globals with local variables
	21 Jan 2004: license message inserted and initial comments reorganized
	 3 Jan 2004: more globals replaced by elements of INTEGRUN structures;
		some routines require additional arguments as a result
	13 Dec 2003: elim unused vbls datacode from sum_refl()
		and lx, ly, mx, my, mz, mxy, mxyz from intcalc()
	22 Nov 2003: replace global variables with members of INTEGRUN struct
	 8 Nov 2003: ar00, etc. calculated in getrefls so we can ignore
		them elsewhere.
	 2 Nov 2003; several routines moved to integutil.c so this file
		is not so awkwardly long; fxpr, fypr made global to allow that
	27 Oct 2003: cosmetic changes
	19 Sep 2003: proper initializations of pb1, pb2 in precenter()
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	10 May 2003: use igp->igr_heads structure for reading in `precenter'
	 5 Jan 2003: reorganized source-file header
	15 Nov 2002: extern declarations made here rather than in integext.h
	 1 Nov 2002: use of REGIONPROFILE structures for reflection box sizes
		and altered REGIONATT structures for region attributes
	30 Sep 2002: xy2ev(), xy2hab(), kab_coord() calls replaced by calls to
		ixy2ev(), xy2h(), ikabc(), with concomitant reorganization.
	25 Mar 2002: removed MSI attachments
	15 Nov 1997: in bg_model, we allow bg up to 40000, rather than 4000.
	30 Sep 1995: rename photperadu to aduperphot
	24 Aug 1995: revised handling of missing pixels and reflection tails
		bugfix for sigmas of Kabsch-geometry data
	15 Aug 1995: in do_sums(), change the coefficient of the error
		correction from 0.1 to 0.04 in the simple-summation intensity
	10 Aug 1995: in do_sums(), revise the goodness-of-fit formula
	31 May 1995: "register" -> "register int" throughout
	10 Mar 1995: zero memory correctly at the end of `precenter';
		fix bug in "precenter": comparisons are <, not <=;
		preliminary summation box is 5x5x3 for small detectors,
		 7x7x3 for medium-sized detectors, 9x9x3 for big ones.
		precenter() uses variable-sized (X,Y) box:
		allow some slop in binary profiles when near half-integer
		frame number; precenter the model with respect to the data
	 2 Mar 1995: use plane backgrounds; add frame # arg to pix2cm, cm2pix
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
		reformat verbose output of background & Lp plots
	 1 Dec 1994: diagnostic output allows for detector (X,Y) > 999
	 1 Dec 1994: modify pixel-by-pixel index checking
	 4 Nov 1994: kga = fabs(sin(3.5*framrad))*gamco,
		not kga = sin(3.5*framrad)*gamco;
		otherwise Kabsch profile-fitting bungles if stepsize < 0
	 5 Oct 1994: use xgwrite to output OBSREF data
	14 Jun 1994: replaced lookup tables with local calcs for exit-beam & Lp
		reworked min & max hkl values;
		reorganized test reflection printout
	 5 May 1994: added profile-flattening and extrapolation to clean_wmask;
		source-file header cleaned up; names changed from POLYDET to
		XGEN; "note_badmodel" routine added and used; some routines
		re-ordered; recentering of updated model profiles reordered
	16 Mar 1994: semi-major reorg: many routines moved to xgenmask.c
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version
	 5 Aug 1993: in precenter(), make sure the currently-considered
		pixel really belongs to the current reflection.
	28 Oct 1992: in bg_model(), calculate <Kabsch background> per pixel

   Purpose: calculating integrated intensities of individual reflections.
   Cross references: ../spots/spotshoes.c, ./integmain.c, etc.

 In each case intensities are calculated by direction summation and
 by profile fitting using both (X,Y,omega) space and using the geometrical
 re-binning described by Kabsch (J.Appl.Cryst. (1988) 21:916).
*************************************************************************/
#define		NEED_TIME	1
#include	"integdefs.h"
#define		MAX_DELIND	(0.45)	/* max error in index to be counted */
#define		MAX_MINMODEL	(1.E-3)	/* ignore model pixels < this*/
#define		MINMSUM		(0.1)	/* need >= this fraction of the model*/
#define		RESCALER	(62.)
#define		VERBOSE_DIFF	(225.)	/* (delI/sig)^2 for verbose printout */
#define		VERBOSE_DELI	(100.)	/* ((Ifit-Isum)/sig)^2 for verbose */
#define		WIDTH_RESCALE	(1.e-4)	/* rescaler for widths */
#define		NEWT_CONVERGED	(0.1e-5)	/* convergence for zero test */
#define		MAXTRIES	(100)	/* max # cycles of Newton's method*/
#define		XYOM_IOSMAX	20.
#define		XYOM_IOSMIN	1.
#define		KAB_IOSMAX	64.
#define		KAB_IOSMIN	2.
#define		BGWTFLOOR	(0.1)
#define		MIN_VAR		(0.05) 	/* min value for rough variance*/
#define		MISSVARIANCE	0.04
#define		MAX_BGPERPIX	40000.
#define		PIO6		(0.5236) /* volume of sphere inside cube */

static int xy2h(int xpx, int ypx, INTEGRUN *igp)
{ /*	This computes the local (h,k,l) based on an integer pixel position.
	This gets compared with the current reflection's hkl value;
	if it's close we return 0.
	If it isn't, or if the pixel is outside the active area,
	we return various integers expressive of the error type. */

	double		lv0, lv1, lv2, rv0, rv1, *dhp;
	REFPARS		*xp;

	if (EOK != ixy2ev(xpx, ypx, &(igp->igr_kaf[2]))) return -1;
	lv0 = igp->igr_kaf[2];	lv1 = igp->igr_kaf[3];	lv2 = igp->igr_kaf[4];
	if (NULL == activmask(xpx, ypx, 041)) return -2; /*not in active area*/
	lv0 += 1;	/* move to reciprocal-space origin */
		/* rotate lv0 and lv1 by the omega angle */
	rv0 = igp->igr_cost * lv0 - igp->igr_sint * lv1;
	rv1 = igp->igr_sint * lv0 + igp->igr_cost * lv1;
	/* multiply by the unit-cell matrix and see if the
	 resulting index vector is close to the current reflection.
	 If it isn't, return an error value immediately. */
	dhp = &(igp->igr_loch[0]);	xp = igp->igr_xp;
	*dhp = xp->xareal[0][0] * rv0 + xp->xareal[0][1] * rv1 +
		xp->xareal[0][2] * lv2;
	if ((*dhp < igp->igr_indlim[0]) || (*dhp > igp->igr_indlim[3]))
		return -3;
	*++dhp = xp->xareal[1][0] * rv0 + xp->xareal[1][1] * rv1 +
		xp->xareal[1][2] * lv2;
	if ((*dhp < igp->igr_indlim[1]) || (*dhp > igp->igr_indlim[4]))
		return -4;
	*++dhp = xp->xareal[2][0] * rv0 + xp->xareal[2][1] * rv1 +
		xp->xareal[2][2] * lv2;
	if ((*dhp < igp->igr_indlim[2]) || (*dhp > igp->igr_indlim[5]))
		return -5;
	return 0; /* passed the tests */
}

static void precenter(INTEGRUN *igp)
{ /*	This does a preliminary estimate of the X,Y,omega centroid and the
	intensity of the spot.  The fraction of the intensity missed in this
	deliberately incomplete summation is estimated from the model profile.
	We cover five frames and a 7x7 region within each frame to do this. 
	We ignore errors in LP factors within pixels this time 'round. */
	register int	dz, x, y, z;
	int		ikept, iz, lz, lz2, mz, nnopix, nunreliable, xmn, xmx;
	int		xyf, xyh, ymn, ymx, zmn, nouts[5], xyl[6];
	xgulong		*ip;
	double		bg, bgxy, bgy, ct, ffl, ffu, frmcorr, ioc;
	double		pb0, pb1, pb2, rzu, sv, tlp, uz, xymid, zwo2;
	double		cob[3], tc[3];
	static float	*bsp, *bvp;
	MAIL		*thishd;
	REFPARS		*xp;
	BGREFL		*rebg;
	BG_PLANE	*repla;
	REGIONATT	*wbmp;
	REGIONPROFILE	*rgb;
	OBSREF		*lrp;

	xp = igp->igr_xp;
	if (igp->igr_zwidth >= 5)
	  {
		zmn = ((int)(igp->igr_fphipr + 0.49999)) - 2;	mz = 5;
		zwo2 = 2.5;
	  }
	else {
		zmn = ((int)(igp->igr_fphipr + 0.49999)) - 1;	mz = 3;
		zwo2 = 1.5;
	     }
	/* 10 March 1995: preliminary summation box is 5x5x3 for little
	 detectors, 7x7x3 for medium-sized detectors, and 9x9x3 for big ones */
	rgb = igp->igr_rgb;	lz = rgb->m3_bz;	lz2 = lz / 2;
	xyh = 2 + (xp->xdetwid >= 500) + (xp->xdetwid >= 1000);
	xyf = xyh + xyh + 1;
	ymn = ((int)(igp->igr_fypr + 0.49999)) - xyh;	ymx = ymn + xyf;
	xmn = ((int)(igp->igr_fxpr + 0.49999)) - xyh;	xmx = xmn + xyf;
	xymid = 0.5 * ((double)xyf);
	ffl = igp->igr_fphipr - 0.49999;
	ffu = igp->igr_fphipr + 0.49999;
	/* define the coefficients of the plane-backgrounding */
	lrp = igp->igr_reflp;
	rebg = (BGREFL *)&(lrp->obs_omob);
	repla = (BG_PLANE *)&(rebg->bg_pl[0]);
	if (igp->igr_flags & IR_PLANEBGND)
	 {
	   for (dz = 0; dz <= lz; dz++, repla++)
	      {
		if (repla->bgp_p0)
		  {
			pb0 = ((double)(repla->bgp_p0)) /  rebg->bg_sc;
			if (rebg->bgx_sc) pb1 =
			 ((double)(repla->bgp_p1)) / ((double)(rebg->bgx_sc));
			else pb1 = 0;
			if (rebg->bgy_sc) pb2 =
			 ((double)(repla->bgp_p2)) / ((double)(rebg->bgy_sc));
			else pb2 = 0;
		  }
		else {
			ip = igp->igr_upwp + igp->igr_xcen * xp->xdetht +
				igp->igr_ycen;
			pb0 = ((double)(*ip)) * igp->igr_oversc / BGDSCALE;
			pb1 = pb2 = 0.;
		     }
		igp->igr_plbg[dz][0] = pb0;	igp->igr_plbg[dz][1] = pb1;
		igp->igr_plbg[dz][2] = pb2;
	      } 
	  }
	zeromem((void *)nouts, mz * sizeof (int));	nunreliable = 0;
	for (z = zmn, bsp = igp->igr_bs, bvp = igp->igr_bv,
	 sv = iz = nnopix = 0; iz < mz; iz++, z++)
	   {
		thishd = igp->igr_heads + (z % (igp->igr_nmbox - 1));
		frmcorr = (igp->igr_logst + (z % (igp->igr_nmbox - 1)))->fcorr;
		rzu = z;
		rzu = (rzu < ffl) ? rzu + 1.: ((rzu < ffu) ? ffu : rzu);
		rzu *= xp->xfrmsize;
		igp->igr_cost = cos(rzu);	igp->igr_sint = sin(rzu);
		uz = ((double)(z + lz2)) + 0.49999 - igp->igr_fphipr;
		dz = uz;
		pb0 = igp->igr_plbg[dz][0]; pb1 = igp->igr_plbg[dz][1];
		pb2 = igp->igr_plbg[dz][2];
		bgy = pb0 - ((double)xyh) * pb1 - ((double)xyh) * pb2;
		for (y = ymn; y < ymx; y++, bgy += pb2)
		 {
		   ip = igp->igr_upwp + xmn * xp->xdetht + y;
		   for (x = xmn, bgxy = bgy; x < xmx;
		    x++, bsp++, bvp++, bgxy += pb1, ip += xp->xdetht)
		      {
			if (igp->igr_flags & IR_PLANEBGND) bg = bgxy;
			else bg = ((double)(*ip)) / BGDSCALE;
			/* does pixel really belong to current reflection? */
			if (!(ikept = xy2h(x, y, igp)))
			  {
				ikept = imvalue(thishd, x, y);
				/* increment count of saturations here */
				ct = ((double)ikept) * frmcorr;
				if (ikept >= igp->igr_minunrely) nunreliable++;
			  }
			else {
				ct = bg;
				if (ikept > -3) nnopix++; else (nouts[iz])++;
			     }
			tlp = ev2lp(igp->igr_kaf + 2, igp);
			*bsp = (ct - bg) * tlp;
			*bvp = tlp * tlp + (ct + BGVARSCALE * bg);
			sv += *bvp;
		      }
		 }
	   }
	if (nunreliable >= TOO_MANY_UNRELIABLE)
	  {
		if (igp->igr_flags & IR_VERBOSEANY)
		  {
			fprintf(igp->igr_fpl,
			 "[%3d,%3d,%3d] @ [%6.1f,%6.1f,%6.1f]",
		 	 lrp->obs_hkl[0], lrp->obs_hkl[1], lrp->obs_hkl[2],
			 igp->igr_fxpr, igp->igr_fypr, igp->igr_fphipr);
			fprintf(igp->igr_fpl,
			 " Unreliable: %3d pixels had >= %6d adu's\n",
				nunreliable, (int)(igp->igr_minunrely));
		  }
		lrp->obs_flags |= I_NUNRELIABLE;
		return;
	  }
	wbmp = igp->igr_bmask + lrp->obs_region;
	tc[0] = igp->igr_fxpr - igp->igr_xcen;
	tc[1] = igp->igr_fypr - igp->igr_ycen;
	tc[2] = igp->igr_fphipr - (double)((int)(igp->igr_fphipr));
	if (tc[2] > 0.5) tc[2] -= 1.;
	/* print the bs vector at this stage*/
	if ((igp->igr_flags & (IR_VERBOSE2 | IR_VERBOSETEMP)))
	  {
		xyl[0] = xmn;	xyl[1] = xmx - 1;
		xyl[2] = ymn;	xyl[3] = ymx - 1;
		xyl[4] = zmn;	xyl[5] = zmn + mz - 1;
		fprintf(igp->igr_fpl,
 "Preliminary centering for [%3d,%3d,%3d] @ [%6.1f,%6.1f,%6.1f];",
		 lrp->obs_hkl[0], lrp->obs_hkl[1], lrp->obs_hkl[2],
		 igp->igr_fxpr, igp->igr_fypr, igp->igr_fphipr);
		fprintf(igp->igr_fpl, " #pixels outside active area = %5d\n",
			nnopix);
		fprintf(igp->igr_fpl,
 " #pixels/frm outside hkl limits =");
		for (iz = 0; iz < mz; iz++) fprintf(igp->igr_fpl,
			" %3d", nouts[iz]);
		fprintf(igp->igr_fpl, "\n");
		bstoloc(igp->igr_bs, xyl);	disp_3dprof(igp->igr_fpl, xyl);
	  }
	lrp->obs_ii = gen_centroid(igp->igr_bs, igp->igr_bv, xyf, xyf,
	 mz, tc, cob);
	/* this correction factor takes into account the portion of the
	profile omitted in this 5x7x7 box; it knows about miscentering */
	ioc =	(zerf((xymid - tc[0]) / wbmp->ra_sigx) +
		 zerf((xymid + tc[0]) / wbmp->ra_sigx)) *
		(zerf((xymid - tc[1]) / wbmp->ra_sigy) +
	 	 zerf((xymid + tc[1]) / wbmp->ra_sigy)) *
		(zerf((zwo2  - tc[2]) / wbmp->ra_sigz) +
		 zerf((zwo2  + tc[2]) / wbmp->ra_sigz))  /  8.;
	ioc = (ioc > DFUZZ ? flimits(1. / ioc, 1., 5.) : 1.);
	sv *= PIO6;	/* eliminate corners of cube from variance estimate */
	lrp->obs_ii *= ioc;
	lrp->obs_sigma = ioc * ioc * sv * igp->igr_aduperph;
	/* on this pass, we record the weighted centroid (the one we're
	going to use for centering), not the observed centroid `cob' */
	lrp->obs_xob = ((double)(igp->igr_xcen) + tc[0]) * igp->igr_xscal +
		0.49999;
	lrp->obs_yob = ((double)(igp->igr_ycen) + tc[1]) * igp->igr_yscal +
		0.49999;
	lrp->obs_omob = (int)(igp->igr_fphipr);
	if (igp->igr_fphipr > lrp->obs_omob + 0.5) lrp->obs_omob += 1.;
	lrp->obs_omob += tc[2];
	if (igp->igr_flags & IR_VERBOSEANY)
	  {
		fprintf(igp->igr_fpl,
		 "[%3d,%3d,%3d] pred.@[%6.1f,%6.1f,%7.2f]",
		 lrp->obs_hkl[0], lrp->obs_hkl[1], lrp->obs_hkl[2],
		 igp->igr_fxpr, igp->igr_fypr, igp->igr_fphipr);
		fprintf(igp->igr_fpl, ",obs.@[%6.1f,%6.1f,%7.2f]; ",
		 ((double)lrp->obs_xob) / igp->igr_xscal,
		 ((double)lrp->obs_yob) / igp->igr_yscal, lrp->obs_omob);
	  }
		/* zero out arrays so they can be reused */
	iz = mz * xyf * xyf * sizeof (float);
	zeromem((AVOID *)(igp->igr_bs), (size_t)iz);
	zeromem((AVOID *)(igp->igr_bv), (size_t)iz);
}

static int vec3d(int iz, int z, INTEGRUN *igp)
{ /*	This accumulates the six vector sums within a single frame. It returns
	the number of pixels whose intensity is above the reliability cutoff */

	register int	dz, ix, iy, x, y;
	u_char		*bip, *cbp, *cbp0;
	int		ika, ikb, ikept, ikg, jka, jkb, jkg, lreg;
	int		lx, lxy, ly, lz, lz2, mx, mxy, my, mz;
	int		nunreliable_fm, shz, skipframe, useit;
	xgulong		*ip;
	double		bg, bgxy, bgy, ct, deka, dekb, dekg, ffl, ffu, frmcorr;
	double		kalpha, kbeta, kgamma, mika, mikb, mikg, pb0, pb1, pb2;
	double		rka, rkb, rkg, thisf, thislp, thism, thisv, uz;
	double		wahi, walo, wbhi, wblo, wghi, wglo, rzu;
	static float	*bmp, *bsp, *bvp, *kbg, *kbgd;
	static float	*kmig, *kmjg, *ksig, *ksjg, *kvig, *kvjg;
	MAIL		*thishd;
	REFPARS		*xp;
	REGIONPROFILE	*rgb, *rgp;
	REGIONATT	*wbmp;

	/* first, see if the binary mask contains any relevant pixels at all */
	xp = igp->igr_xp;	rgb = igp->igr_rgb;
	lreg = (igp->igr_reflp)->obs_region;	rgp = rgb + lreg;
	mx = rgp->m3_qx;	my = rgp->m3_qy;	mz = rgp->m3_qz;
	mxy = mx * my;
	lx = rgp->m3_bx;	ly = rgp->m3_by;	lz = rgp->m3_bz;
	lxy = lx * ly;		lz2 = lz / 2;
	mika = rgp->m3_lim[0]; deka = rgp->m3_lim[2];
	mikb = rgp->m3_lim[3]; dekb = rgp->m3_lim[5];
	mikg = rgp->m3_lim[6]; dekg = rgp->m3_lim[8];
	wbmp = igp->igr_bmask + lreg;	shz = 1 << iz;
	uz = igp->igr_fphipr - (int)(igp->igr_fphipr);
	/* if we're close to the midpoint between two frames, we allow
	 some influence from the neighboring piece of model profile */
	if ((iz < lz2) && (uz > 0.35) && (uz < 0.5)) shz |= (1 << (iz + 1));
	else if ((iz > lz2) && (uz > 0.5) && (uz < 0.65))
		shz |= (1 << (iz - 1));
	for (skipframe = 1, iy = 0, cbp = cbp0 = &(wbmp->ra_incl[0]);
	  iy < lxy; iy++, cbp++)
	    {
		if ((*cbp) & shz)
		  {
			skipframe = 0; break;
		  }
	    }
	if (skipframe) return 0;	/* no useful pixels: give up now */
	cbp = cbp0;	/* back to beginning of model */
	bsp = igp->igr_bs + iz * lxy;	bmp = igp->igr_bm + iz * lxy;
	bvp = igp->igr_bv + iz * lxy;	bip = igp->igr_binc + iz * lxy;
	thishd = igp->igr_heads + (z % (igp->igr_nmbox - 1));
	if (igp->igr_zmin <= z && z <= igp->igr_zmax)
	  {
		frmcorr = (igp->igr_logst + (z % (igp->igr_nmbox - 1)))->fcorr;
		skipframe = 0;	/* we have data available for this frame */
	  }
	else	frmcorr = skipframe = 1; /* no data available for this frame */
	kgamma = sin(((double)z - igp->igr_fphipr) * xp->xfrmsize) *
		igp->igr_gamco;
	rkg = flimits((kgamma - mikg) / dekg, 0., 8.);
	ikg = rkg;
	wglo = 1. + (double)ikg - rkg;	wghi = 1. - wglo;
	jkg = (ikg < mz - 1 ? ikg + 1 : ikg);
	/* the omega value we use takes into account the fact that
	 the frame itself covers a range of omega values */
	ffl = igp->igr_fphipr - 0.49999;
	ffu = igp->igr_fphipr + 0.49999;
	rzu = z;
	rzu = (rzu < ffl) ? rzu + 1.: ((rzu < ffu) ? ffu : rzu);
	rzu *= xp->xfrmsize;
	igp->igr_cost = cos(rzu);	igp->igr_sint = sin(rzu);
	ksig = igp->igr_ks + ikg * mxy;	ksjg = igp->igr_ks + jkg * mxy;
	kvig = igp->igr_kv + ikg * mxy;	kvjg = igp->igr_kv + jkg * mxy;
	kmig = igp->igr_km + ikg * mxy;	kmjg = igp->igr_km + jkg * mxy;
	kbg = igp->igr_kbg;	kbgd = igp->igr_kbgd;
	uz = ((double)(z + lz2)) + 0.49999 - igp->igr_fphipr;	dz = uz;
	pb0 = igp->igr_plbg[dz][0]; pb1 = igp->igr_plbg[dz][1];
	pb2 = igp->igr_plbg[dz][2];
	bgy = pb0 + ((double)(igp->igr_ymi - igp->igr_ycen)) * pb1 +
		((double)(igp->igr_xmi - igp->igr_xcen)) * pb2;
	nunreliable_fm = 0;
	for (iy = 0, y = igp->igr_ymi; iy < ly; iy++, y++, bgy += pb2)
	   {
		for (x = igp->igr_xmi, ix = 0, bgxy = bgy; ix < lx;
		 ix++, x++, bsp++, bmp++, bvp++, cbp++, bip++, bgxy += pb1)
		   if ((*cbp) & shz)	/* is the pixel in the binary mask?*/
		     { /* yes: use it in the reflection */
			thism = *bmp;
			useit = !reg2hab(x, y, igp->igr_kaf);
			if (useit) useit = !xy2h(x, y, igp);
			kalpha = igp->igr_kaf[0]; kbeta = igp->igr_kaf[1];
			rka = flimits((kalpha - mika) /deka, 0., 12.);
			ika = rka;
			walo = 1. + (double)ika - rka; wahi = 1. - walo;
			jka = (ika < 12 ? ika + 1 : ika);
			rkb = flimits((kbeta - mikb) / dekb, 0., 12.);
			ikb = rkb;
			wblo = 1. + (double)ikb - rkb; wbhi = 1. - wblo;
			jkb = (ikb < 12 ? ikb + 1 : ikb);
			ikb *= mx;	jkb *= mx;
			if (thism > DFUZZ)
			  {
				*(kmig+ ikb+ ika) += thism *walo *wblo * wglo;
				*(kmig+ ikb+ jka) += thism *wahi *wblo * wglo;
				*(kmig+ jkb+ ika) += thism *walo *wbhi * wglo;
				*(kmig+ jkb+ jka) += thism *wahi *wbhi * wglo;
				*(kmjg+ ikb+ ika) += thism *walo *wblo * wghi;
				*(kmjg+ ikb+ jka) += thism *wahi *wblo * wghi;
				*(kmjg+ jkb+ ika) += thism *walo *wbhi * wghi;
				*(kmjg+ jkb+ jka) += thism *wahi *wbhi * wghi;
			  }
		   	/* if pixel is legal & in this refl, use it */
			if (useit && !skipframe)
			  {
			    ikept = imvalue(thishd, x, y);
			     /* pixels with high (unreliable) counts are
				treated as unavailable for integration */
			    if (ikept >= igp->igr_minunrely)
			      {
				nunreliable_fm++;	*bip = 0;
			      }
			    else
			      {
				thislp = ev2lp(igp->igr_kaf + 2, igp);
				if (igp->igr_flags & IR_PLANEBGND)
					bg = bgxy;
				else {
					ip = igp->igr_upwp + x * xp->xdetht +y;
					bg = ((double)(*ip)) / BGDSCALE;
				     }
				ct = ((double)ikept) * frmcorr;
				*bsp = thisf = thislp * (ct - bg);
				*bvp = thisv = thislp * thislp * (ct +
				 BGVARSCALE * bg) + (ct - bg) * (ct - bg) *
					igp->igr_vlp;
				*(ksig+ ikb +ika) += thisf *walo *wblo * wglo;
				*(ksig+ ikb +jka) += thisf *wahi *wblo * wglo;
				*(ksig+ jkb +ika) += thisf *walo *wbhi * wglo;
				*(ksig+ jkb +jka) += thisf *wahi *wbhi * wglo;
				*(ksjg+ ikb +ika) += thisf *walo *wblo * wghi;
				*(ksjg+ ikb +jka) += thisf *wahi *wblo * wghi;
				*(ksjg+ jkb +ika) += thisf *walo *wbhi * wghi;
				*(ksjg+ jkb +jka) += thisf *wahi *wbhi * wghi;
				*(kvig+ ikb +ika) += thisv *walo *wblo * wglo;
				*(kvig+ ikb +jka) += thisv *wahi *wblo * wglo;
				*(kvig+ jkb +ika) += thisv *walo *wbhi * wglo;
				*(kvig+ jkb +jka) += thisv *wahi *wbhi * wglo;
				*(kvjg+ ikb +ika) += thisv *walo *wblo * wghi;
				*(kvjg+ ikb +jka) += thisv *wahi *wblo * wghi;
				*(kvjg+ jkb +ika) += thisv *walo *wbhi * wghi;
				*(kvjg+ jkb +jka) += thisv *wahi *wbhi * wghi;
				*(kbg + ikb +ika) += bg * walo * wblo;
				*(kbg + ikb +jka) += bg * wahi * wblo;
				*(kbg + jkb +ika) += bg * walo * wbhi;
				*(kbg + jkb +jka) += bg * wahi * wbhi;
				*(kbgd + ikb +ika) += walo * wblo;
				*(kbgd + ikb +jka) += wahi * wblo;
				*(kbgd + jkb +ika) += walo * wbhi;
				*(kbgd + jkb +jka) += wahi * wbhi;
				igp->igr_roughi += thisf;
				igp->igr_roughv += thisv;
				*bip = 1;
			      } /* close of "else" of "if ((!ikept" clause */
			  } /* close of "if (!useit" clause */
			else	*bip = 0;
		   } /* close "for (x" loop */
	   } /* close "for (y" loop */
	return nunreliable_fm;
}

static void do_sums(int datatype, double *resul, INTEGRUN *igp)
{ /*	This composes the actual summations associated with
	the intensity calculations.  `datatype' == 0 for directly-determined
	data; it's 1 for Kabsch-formulation data */
	register int	ix, iy, iz;
	int		lx, ly, lz, mx, my, mz, npix, nx, ny, nz;
	double		fitsc, iscale, mscale, msum, qdeli, tailco, vscale;
	double		sumz[9], susz[9];
	double		bgwt, gsub, gwt, ifd, ifit, igof, isum, isvar, ivar;
	double		iwt, pwt, pwt2, sfud, sumcor, td0, thisf, thisf2;
	double		thism, thism2, thisv, tinyf;
	OBSREF		*lrp;
	REGIONPROFILE	*rgb;
	static float	*vm, *vmp, *vs, *vsp, *vv, *vvp;

	isum = ifit = ifd = isvar = ivar = msum = igof = gwt = 0.;
	rgb = igp->igr_rgb;
	lx = rgb->m3_bx;	ly = rgb->m3_by;	lz = rgb->m3_bz;
	mx = rgb->m3_qx;	my = rgb->m3_by;	mz = rgb->m3_bz;
	if (datatype)
	  {
		nx = mx;	ny = my;	nz = mz;
		vs = igp->igr_ks; vm = igp->igr_km; vv = igp->igr_kv;
	  }
	else {
		nx = lx;	ny = ly;	nz = lz;
		vs = igp->igr_bs; vm = igp->igr_bm; vv = igp->igr_bv;
	     }
	for (iz = 0; iz < nz; iz++) susz[iz] = sumz[iz] = 0.;
	vsp = vs;	vmp = vm;	vvp = vv;
	bgwt = (1. + BGVARSCALE) * igp->igr_meanbg[datatype];
	if (bgwt < BGWTFLOOR) bgwt = BGWTFLOOR;
	iwt = (igp->igr_roughi < DFUZZ) ? 0.01 : igp->igr_roughi;
	for (tinyf = iscale = mscale = vscale = npix = iz = 0; iz < nz; iz++)
	  for (iy = 0; iy < ny; iy++)
	    for (ix = 0; ix < nx; ix++, vsp++, vmp++, vvp++)
	      {
		if ((thisf = *vsp) > iscale) iscale = thisf;
		if ((thism = *vmp) > mscale) mscale = thism;
		if ((thisv = *vvp) > vscale) vscale = thisv;
		msum += thism;	/* TEMPORARY */
		if (thism > igp->igr_minmod)
		  {
		    if (thisv > DFUZZ)
		      {
			thism2 = thism * thism;	thisf2 = thisf * thisf;
			isum += thisf;	/* msum += thism; TEMPORARY */
			pwt = 1. / (thism * iwt + bgwt); pwt2 = pwt * pwt;
			ifit += thisf * thism * pwt;	ifd += thism2 * pwt;
			susz[iz] += thisf; sumz[iz] += thism;
			ivar += thism2 * pwt2 * thisv;	gwt += pwt;
			isvar += thisv;	igof += thisf2 * pwt;	npix++;
		      }
		  }
		else tinyf += thism; /* only increment tinyf if thism small */
	      }
	/* correct for excluded pixels, excluded model regions, & weak pixels*/
	tailco = igp->igr_tailc[datatype];
	tinyf = 1. / flimits(1. - tinyf, 0.945, 1.0);
	sumcor = igp->igr_missf * tailco * tinyf;
	lrp = igp->igr_reflp;
	if (!datatype)	/* remember the frame-by-frame intensity */
	  {
		lrp->obs_lowfrm = lrp->obs_hifrm = 99;
		for (iz = 0; iz < nz; iz++)
		   {
			lrp->obs_frmint[iz] = susz[iz];
			if ((sumz[iz] > 0) && (lrp->obs_lowfrm == 99))
				lrp->obs_lowfrm = iz;
		   }
		for (iz = nz; iz--; )
		   {
			if ((sumz[iz] > 0) && (lrp->obs_hifrm == 99))
			  {
				lrp->obs_hifrm = iz;	break;
			  }
		   }
		if (lrp->obs_lowfrm == 99)
			lrp->obs_lowfrm = lrp->obs_hifrm = nz / 2;
		lrp->obs_ncells = npix;
	  }
	/* correct the simple-summation intensity for missing pixels.
	The error thus introduced must be modeled too.
	As of Aug 1995 the coefficient of the error-correction
	has been lowered from 0.1 to 0.04 */
	isvar = isvar * sumcor * sumcor +
		MISSVARIANCE * isum * isum * (sumcor - 1.) * (sumcor - 1.);
	if (isvar < MIN_VAR) isvar = MIN_VAR;
	isum *= sumcor;
	npix = (npix > 2 ? npix - 1 : 1);
	if ((ifd > DFUZZ) && (msum > DFUZZ) && (npix > 0))
	  {	/* the profile-fitted variance has a term in it
		 associated with the goodness-of-fit between the observed
		 and calculated profile.  We put it in at this point.
		 The G.O.F. formula was revised in Aug 1995. */
		gsub = (ifit / ifd) * ifit;
		ifit *= msum / ifd;
		igof = (igof - gsub) * ((double)npix) / gwt;
		ivar *= (msum / ifd) * (msum / ifd);
	  }
	else {
		ifit = 0.;	igof = 1.;	ivar = 2.;
	     }
	/* when we apply the breadth fudge, we tweak the sigma too: */
	fitsc = tailco;
	sfud = (1. - fitsc) / 8.;
	ivar = ivar * fitsc * fitsc + (ifit * sfud) * (ifit * sfud);
	ifit *= fitsc;
	igof *= fitsc * fitsc;	if (igof < FUZZ) igof = FUZZ;
	/* for bright reflections we set up an upper bound on the goodness
	of fit, based on the fact that compensating errors may occur between
	pixels of the observed profile. This upper bound is in fact the
	square of the difference between the profile-fitted and summed I. */
	if ((isum > 1.) && (isum * isum > isvar * 4.))
	  {
		qdeli = ifit - isum;	qdeli *= qdeli;
		if (igof > qdeli) igof = qdeli;
	  }
	ivar += igof;	if (ivar < MIN_VAR) ivar = MIN_VAR;
	/* turn on verbose if this intensity disagrees with
	the previously-calculated value by a lot */
	fitsc = ifit - igp->igr_roughi;
	fitsc = (fitsc / igp->igr_roughv) * fitsc;
	/* also turn it on if Ifit-Isum is large */
	td0 = ifit - isum;	td0 = (td0 / isvar) * td0;
	if (igp->igr_flags & IR_VERBOSEANY)
	  {
		if (!datatype)
		  {
			fprintf(igp->igr_fpl, "   h   k   l  Iprelim sprelim");
			fprintf(igp->igr_fpl,
			 "  Dir.Ifit   sfit    Isum   ssum tails tinyf ");
			fprintf(igp->igr_fpl,
		 "  Kab.Ifit   sfit    Isum   ssum tails tinyf  missf\n");
			fprintf(igp->igr_fpl, "%4d%4d%4d %8.1f %7.2f",
			 lrp->obs_hkl[0], lrp->obs_hkl[1],
			 lrp->obs_hkl[2], lrp->obs_ii, sqrt(lrp->obs_sigma));
		  }
		fprintf(igp->igr_fpl,
		 " %9.1f %6.2f %7.1f %6.2f %5.3f %5.3f",
		 ifit, sqrt(ivar), isum, sqrt(isvar), tailco, tinyf);
		if (datatype) fprintf(igp->igr_fpl, "  %5.3f\n",
			igp->igr_missf);
		else fprintf(igp->igr_fpl, " ");
	  }
	iscale = (iscale > 0 ? RESCALER / iscale : 0.);
	mscale = (mscale > 0 ? RESCALER / mscale : 0.);
	vscale = (vscale > 0 ? RESCALER / vscale : 0.);
	resul[0] = ifit;	resul[1] = ivar;
	resul[2] = isum;	resul[3] = isvar;
	resul[4] = iscale;	resul[5] = mscale;	resul[6] = vscale;
	resul[7] = fitsc;	resul[8] = sumcor;	resul[18] = igof;
	for (iz = 0; iz < nz; iz++) resul[iz+9] = susz[iz];
}

static void f_after_refl(INTEGRUN *igp)
{ /*	This creates an updated version of the 3D shape function in
	the appropriate region of the detector. */
	register int	ix, iy, iz;
	int		ljz, lx, lxyz, ly, lz;
	double		bsum, nmult, ofmax, omult, thisf, idv[3], dev[3];
	REGIONPROFILE	*rgb, *rgp;
	OBSREF		*lrp;
	static float	*bmp, *bsp, *tqp;
	static double	outfloor = 0.001;

/* Clean up, normalize and recenter the `bs' array */
	rgb = igp->igr_rgb;	lx = rgb->m3_bx;
	ly = rgb->m3_by;	lz = rgb->m3_bz;
	lxyz = lx * ly * lz;	bsp = igp->igr_bs;	bmp = igp->igr_bm;
	for (bsum = iz = 0; iz < lz; iz++) /* add up elements of bs vector */
	   for (iy = 0; iy < ly; iy++)
	      for (ix = 0; ix < lx; ix++, bsp++, bmp++)
		 {
			if (*bmp > igp->igr_minmod) bsum += *bsp;
			else *bsp = 0.;
		 }
	if (bsum < DFUZZ) return;
	for (bsp = igp->igr_bs, iz = 0; iz < lz; iz++) /* renormalize */
	   for (iy = 0; iy < ly; iy++)
	      for (ix = 0; ix < lx; ix++, bsp++) *bsp /= bsum;
	idv[0] = idv[1] = idv[2] = 0.;
	if (DFUZZ > (bsum = gen_centroid(igp->igr_bs, NULL,
	 lx, ly, lz, idv, dev))) return;
	recen_vec(igp->igr_bs, dev, igp->igr_tq, lx, ly, lz);
/* add it cleanly into the appropriate model.  We then check that
 the resulting model is reasonably well-centered; if it is, we update
 the model for this region.  If it isn't, we try one more time to center
 it and then give up. */
	lrp = igp->igr_reflp;	rgp = rgb + lrp->obs_region;
	nmult = lrp->obs_ii * (lrp->obs_ii / lrp->obs_sigma);
	for (iz = 0, bsp = igp->igr_bs, bmp = rgp->m3_xyop, tqp = igp->igr_tq;
	 iz < lxyz; iz++, bsp++, bmp++, tqp++)
	   {
		thisf = *bmp + *tqp * nmult;
		*bsp = (thisf < outfloor) ? 0. : thisf;
	   }
	ljz = igp->igr_jzmin;
	if (igp->igr_flags & (IR_VERBOSE2 | IR_VERBOSETEMP))
	  {
		bmp = igp->igr_bm;	bsp = igp->igr_bs;
		for (tqp = rgp->m3_xyop, omult = ofmax = 0., iz = 0;
		 iz < lxyz; iz++, bmp++, bsp++, tqp++)
		  {
			*bmp = *tqp;
			if (*bmp > omult) omult = *bmp;
			if (*bsp > ofmax) ofmax = *bsp;
		  }
		if (omult > DFUZZ)
		  {
			fprintf(igp->igr_fpl, " input model: ");
			dispprofile(2, ljz, RESCALER / omult, igp->igr_bm,igp);
		  }
		if (ofmax > DFUZZ)
		  {
			fprintf(igp->igr_fpl, "output model: ");
			dispprofile(0, ljz, RESCALER / ofmax, igp->igr_bs,igp);
		  }
	  }
	idv[0] = idv[1] = idv[2] = 0.;
	if (DFUZZ > gen_centroid(igp->igr_bs, NULL, lx, ly, lz, idv, dev))
		return;
	if ((fabs(dev[0]) > 0.6) || (fabs(dev[1]) > 0.6) ||
		(fabs(dev[2]) > 0.3))
	  {
		cpymem((void *)(igp->igr_tq), (void *)(igp->igr_bs),
		 lxyz * sizeof (float));
		recen_vec(igp->igr_tq, dev, igp->igr_bs, lx, ly, lz);
		idv[0] = idv[1] = idv[2] = 0.;
		if (DFUZZ > gen_centroid(igp->igr_bs, NULL, lx, ly, lz,
		 idv, dev)) return;
		if ((fabs(dev[0]) > 0.6) || (fabs(dev[1]) > 0.6) ||
			(fabs(dev[2]) > 0.3))
		  {
			if (igp->igr_flags & IR_VERBOSEANY)
				note_badmodel(igp->igr_fpl, dev, igp);
			return;
		  }
	  }
	rgp->m3_isqsum += nmult;	(rgp->m3_nobs)++;
	for (bsp = igp->igr_bs, bmp = rgp->m3_xyop, iz = 0; iz < lxyz;
		iz++, bsp++, bmp++) *bmp = *bsp;
}

static void sum_refl(int datatype, int tver, double *ar, INTEGRUN *igp)
{ /*	This prints useful info about a reflection to the logfile */
	register int	iz, nz;
	int		ljz;
	REGIONPROFILE	*rgb;

	rgb = igp->igr_rgb;
	nz = (datatype) ? rgb->m3_qz : rgb->m3_bz;
	fprintf(igp->igr_fpl, "Profile in %s: ",
		(datatype ? "gamma": "omega"));
	for (iz = 0; iz < nz; iz++)
		fprintf(igp->igr_fpl, " %6.0f", ar[iz + 9]);
	fprintf(igp->igr_fpl, "\n");
	ljz = igp->igr_jzmin;
	if ((igp->igr_flags & (IR_VERBOSE2 | IR_VERBOSETEMP)) && !(tver & 04))
	  {
		if (datatype)
		  {
			dispprofile(1, ljz, ar[4], igp->igr_ks, igp);
			dispprofile(3, ljz, ar[5], igp->igr_km, igp);
			dispprofile(5, ljz, ar[6], igp->igr_kv, igp);
		  }
		else {
			dispprofile(0, ljz, ar[4], igp->igr_bs, igp);
			dispprofile(2, ljz, ar[5], igp->igr_bm, igp);
			dispprofile(4, ljz, ar[6], igp->igr_bv, igp);
			disp_bglp(igp);
		     }
	  }
}

static void bg_model(INTEGRUN *igp)
{ /*	This modifies the model profile `bm' based on the notion that the
	area of a pixel is proportional to the background value found therein.
	As a bonus, we calculate the mean background per pixel.
	We do the same thing for the Kabsch background.
	We do this cautiously to avoid overrunning detector boundaries.
	We hope that making 40000 the upper limit for bgnd is plausible.  */
	register int	ix, iy, iz, x, y;
	int		lx, ly, lz, lxy, lxyz;
	int		mx, my, mz, mxy, mxyz, npused;
	xgulong		*ip;
	double		bgv, bgxy, bgy, mbg, mosum, pl0, pl1, pl2;
	float		*bmp, *bmxy, *bgp;
	REFPARS		*xp;
	OBSREF		*lrp;
	REGIONPROFILE	*rgb;

	lrp = igp->igr_reflp;	rgb = igp->igr_rgb;	xp = igp->igr_xp;
	pl0 = igp->igr_plbg[3][0];	pl1 = igp->igr_plbg[3][1];
	pl2 = igp->igr_plbg[3][2];
	bgy = pl0 + ((double)(igp->igr_xmi - igp->igr_xcen)) * pl1 +
		((double)(igp->igr_ymi - igp->igr_ycen)) * pl2;
	lx = rgb->m3_bx;	ly = rgb->m3_by;	lxy = lx * ly;
	lz = rgb->m3_bz;	lxyz = lxy * lz;
	mx = rgb->m3_qx;	my = rgb->m3_qy;
	mz = rgb->m3_qz;	mxy = mx * my;	mxyz = mxy * mz;
	for (npused = mosum = mbg = iy = 0, bmxy = igp->igr_bm,
	 y = igp->igr_ymi; iy < ly; iy++, y++, bgy += pl2)
	   for (ix = 0, x = igp->igr_xmi, bgxy = bgy; ix < lx;
	    ix++, x++, bmxy++, bgxy += pl1)
	    if (NULL != activmask(x, y, 041))
	      {
		if (igp->igr_flags & IR_PLANEBGND) bgv = bgxy;
		else {
			ip = igp->igr_upwp + x * xp->xdetht + y;
			bgv = ((double)(*ip)) / BGDSCALE;
		     }
		bgv = flimits(bgv, 0.01, MAX_BGPERPIX);
		mbg += bgv;	npused++;
		for (bmp = bmxy, iz = 0; iz < lz; iz++, bmp += lxy)
		   {
			*bmp *= bgv;	mosum += *bmp;
		   }
	      }
	if ((npused) && (mosum > DFUZZ))
	  {
		for (bmp = igp->igr_bm, iz = 0; iz < lxyz; iz++, bmp++)
			*bmp /= mosum;
		mbg /= (double)npused;
		lrp->obs_bgnd = igp->igr_meanbg[0] = mbg;
	  }
	else lrp->obs_bgnd = igp->igr_meanbg[0] = 0.;
	/* now do the same correction for the Kabsch data */
	for (mosum = mbg = iy = 0, bmxy = igp->igr_km, bgp = igp->igr_kbg;
	 iy < my; iy++)
	 for (ix = 0; ix < mx; ix++, bmxy++, bgp++)
	    {
		bgv = flimits(*bgp, 0.01, 4000.);	mbg += bgv;
		for (bmp = bmxy, iz = 0; iz < mz; iz++, bmp += mxy)
		   {
			*bmp *= bgv;	mosum += *bmp;
		   }
	    }
	if (mosum > DFUZZ)
	  {
		for (bmp = igp->igr_km, iz = 0; iz < mxyz; iz++, bmp++)
			*bmp /= mosum;
	  }
	igp->igr_meanbg[1] = mbg / (double)mxy;
}

static void kab_bg(INTEGRUN *igp)
{ /*	This renormalizes the Kabsch background values and sets the values
	currently blank to the mean value. */
	int		iy, ix, mx, my, nknz;
	float		*flp, *fldp;
	double		mkb;
	REGIONPROFILE	*rgb;

	rgb = igp->igr_rgb;	mx = rgb->m3_qx;	my = rgb->m3_qy;
	for (mkb = nknz = iy = 0, flp = igp->igr_kbg, fldp = igp->igr_kbgd;
	 iy < my; iy++)
	   for (ix = 0; ix < mx; ix++, flp++, fldp++)
	      {
		if (*fldp > DFUZZ)
		  {
			*flp /= *fldp;	mkb += *flp; nknz++;
		  }
	      }
	if (nknz) /* can we compute <Bgrnd> for Kabsch data? */
	  {
		mkb /= (double)nknz;	/* yes; here it is! */
		for (iy = 0, flp = igp->igr_kbg, fldp = igp->igr_kbgd;
		 iy < my; iy++)
		   for (ix = 0; ix < mx; ix++, flp++, fldp++)
			if (*fldp < FUZZ) *flp = mkb; /* fill in blanks */
	  }
}

static double getmissf(INTEGRUN *igp)
{ /*	This recomputes the fraction missing from the simple summation based
	on the readjusted real-space model.  It counts the fraction of the
	model's intensity that is NOT contained within pixels that really
	belong to this reflection, i.e. that fall within the (X,Y) limits
	of the active area of the detector and whose (hkl) indices are close
	enough to the correct integer values to count. */
	int		iz, lxyz;
	u_char		*bip;
	double		lmsum, outm;
	float		*flp;
	REGIONPROFILE	*rgb;

	rgb = igp->igr_rgb;
	lxyz = rgb->m3_bx * rgb->m3_by * rgb->m3_bz;	
	for (bip = igp->igr_binc, flp = igp->igr_bm,
	 lmsum = outm = iz = 0; iz < lxyz; iz++, bip++, flp++)
	   {
		lmsum += *flp;	if (!(*bip)) outm += *flp;
	   }
	outm = (lmsum > outm ?  flimits(lmsum / (lmsum - outm), 1., 1.1) : 1.);
	return outm;
}

static int intcalc(u_short rinr[], INTEGRUN *igp)
{ /*	The actual intensity of a reflection is calculated here.
	We do a coarse centering of the reflection so we can impose the model
	profile on the right pixels.  Then we construct six 3D boxes: the data,
	the model, and the variance of the reflection in ordinary (X,Y,omega)
	space, and the data, the model, and the variance of the reflection in
	the Kabsch-transformed (alpha,beta,gamma) space.  Only pixels that are
	calculated to be within 0.4 indices of the reflection's (hkl) value
	are included in these sums.  The overlap of the model with the data
	with the model is improved both by recentering and by adjustment
	of the model.  The intensity is then calculated by simple summation
	in both spaces and corrected for pixels that are missing,
	either because of overlap with another refl or because they lie too
	close to the excluded regions of the detector.  The intensity is
	also calculated by profile-fitting and those answers are corrected
	further for differences in overall breadth between the real reflection
	and the model.  Then either the direct-space answers or the
	Kabsch-space answers are written out to disc.  */
	register int	iz;
	int		jzcen, jzdel, lreg, lz, lz2, nunreliable, tv, xyer, z;
	double		delin, delpa, lamco, lthis, roughiv, rz, thisf, tios;
	double		iko[7], kabg[19], xyom[19];
	REFPARS		*xp;
	OBSREF		*lrp;
	FRAMELOG	*llp;
	REGIONPROFILE	*rgb;

	xp = igp->igr_xp;	lamco = 0.25e-4 / (xp->xlambda * xp->xlambda);
	rgb = igp->igr_rgb;	lz = rgb->m3_bz;	lz2 = lz / 2;
	lrp = igp->igr_reflp;	get_limits(lrp, PREDICTED, igp);	tv = 0;
	if (istestrefl(igp))
	  {
		tv = 19;	igp->igr_flags |= IR_VERBOSETEMP;
	  }
	else {
		igp->igr_flags &= ~IR_VERBOSETEMP;
	     }
	if (NULL == activmask((int)(igp->igr_xcen), (int)(igp->igr_ycen),0101))
	  {
		if (igp->igr_flags & IR_VERBOSEANY) fprintf(igp->igr_fpl,
		 "pred. centroid @ [%4d,%4d] not in active area\n",
		 (int)(igp->igr_xcen), (int)(igp->igr_ycen));
		lrp->obs_flags |= I_XYEDGE; return EOK;
	  }
	init3dvec(igp);	/* reserve memory for 3D vectors or zero them */
	/* define the bounds around the current reflection indices. */
	for (iz = 0; iz < 3; iz++)
	   {
		igp->igr_indlim[iz] = (double)(lrp->obs_hkl[iz]) - MAX_DELIND;
		igp->igr_indlim[iz + 3] = igp->igr_indlim[iz] +
			2. * MAX_DELIND;
	   }
	precenter(igp);	/* preliminary centroid & intensity calculation */
	if (lrp->obs_flags & I_NUNRELIABLE) return EOK;
	if ((igp->igr_flags & IR_PLANEBGND) &&
	 (igp->igr_flags & (IR_VERBOSE2 | IR_VERBOSETEMP)))
	  {
		fprintf(igp->igr_fpl, "Plane-model backgrounds:\n");
		for (iz = 0; iz < 4; iz++) fprintf(igp->igr_fpl,
			"Frame  Scalar   Xmult   Ymult  ");
		fprintf(igp->igr_fpl, "\n");
		for (iz = 0; iz <= lz; iz++)
			fprintf(igp->igr_fpl, " %4d %7.2f %7.4f %7.4f%s",
			 iz, igp->igr_plbg[iz][0], igp->igr_plbg[iz][1],
			 igp->igr_plbg[iz][2], (iz + 1) % 4 ? "  " : "\n");
	  }
	get_limits(lrp, OBSERVED, igp);
	if (NULL == activmask((int)(igp->igr_xcen), (int)(igp->igr_ycen),0101))
	  {
		if (igp->igr_flags & IR_VERBOSEANY) fprintf(igp->igr_fpl,
		 " obs. centroid @ [%4d,%4d] not in active area\n",
			(int)(igp->igr_xcen), (int)(igp->igr_ycen));
		lrp->obs_flags |= I_XYEDGE; return EOK;
	  }
	if (igp->igr_xmi < 1 || igp->igr_ymi < 1 ||
	 igp->igr_xma >= xp->xdetwid || igp->igr_yma >= xp->xdetht)
	  {
		if (igp->igr_flags & IR_VERBOSEANY) fprintf(igp->igr_fpl,
			"Reflection on detector edge\n");
		lrp->obs_flags |= I_XYEDGE; return EOK;
	  }
		/* compute the Kabsch coefficients of the reflection */
	if (1 != ikabc(igp->igr_fxpr, igp->igr_fypr, iko))
	  {
		if (igp->igr_flags & IR_VERBOSEANY) fprintf(igp->igr_fpl,
		 "Reflection rejected: can't calculate Kabsch coordinates\n");
		lrp->obs_flags |= I_BADPHISHAPE; return EOK;
	  }
	igp->igr_gamco = iko[6];	/* use this in vec3d calculations */
	lreg = lrp->obs_region;
	/* If we've had to move away from the predicted centroid, the
	 bounds around the indices may have been offset a bit so they can
	 be symmetric about the hkl computed for the observed centroid */
	rz = (igp->igr_fphipr + 0.5) * xp->xfrmsize;
	igp->igr_cost = cos(rz); igp->igr_sint = sin(rz);
	if ((xyer = xy2h(igp->igr_xcen, igp->igr_ycen, igp))) /*too far?*/
	  {
		if (igp->igr_flags & IR_VERBOSEANY) fprintf(igp->igr_fpl,
		 "too large a centroid movement: xy2h error %d\n", xyer);
		lrp->obs_flags |= I_MOVETOOFAR;	return EOK;
	  }
	for (iz = 0; iz < 3; iz++) /* adjust the index limits a bit */
	   {
		delin = igp->igr_loch[iz] - lrp->obs_hkl[iz];
		if (delin > 0.)
		  {
			igp->igr_indlim[iz + 3] += delin * 0.8;
			igp->igr_indlim[iz] += delin * 0.4;
		  }
	   }
	/* we no longer have to account for slop due to finite frame
	 width in this routine because it's being handled in `vec3d' */
	igp->igr_vlp = varlp(igp);	/* approximate Lp factor for refl */
	igp->igr_roughi = igp->igr_roughv = 0.;
	jzcen = igp->igr_fphipr;
	jzdel = (igp->igr_fphipr > (double)jzcen + 0.5);
	igp->igr_jzmin = jzcen + jzdel - lz2;
	/* we may have scribbled on lrp->obs_gama when we did the
	 plane-background calculations; in that case we need to reclaim it
	 before we run `mkmodel', which needs to know that number. */
	lrp->obs_gama = lrp->obs_frmint[lz - 1];
	/* construct the 3D model profile in ordinary (X,Y,omega) space.
	Then widen or narrow the result according to the ratio of the observed
	width to the predicted width as estimated by the gamma values */
	if (I_USABLE != mkmodel(lreg, igp))
	  {
		if (igp->igr_flags & IR_VERBOSEANY)
			fprintf(igp->igr_fpl, "Model profile faulty\n");
		lrp->obs_flags |= I_BADMODEL; return EOK;
	  }
	/* loop through the relevant pixels of the relevant frames,
	 incrementing the appropriate elements of the six vectors */
	for (nunreliable = iz = 0, z = igp->igr_jzmin; iz < lz; iz++, z++)
		nunreliable += vec3d(iz, z, igp);
	if (nunreliable >= TOO_MANY_UNRELIABLE)
	  {
		if (igp->igr_flags & IR_VERBOSEANY) fprintf(igp->igr_fpl,
		 " %3d pixels have > %6d adu's; reflection rejected\n",
			nunreliable, (int)(igp->igr_minunrely));
		lrp->obs_flags |= I_NUNRELIABLE;	return EOK;
	  }
	/* determine the cutoff model value */
	if (igp->igr_roughv < MIN_VAR) igp->igr_roughv = MIN_VAR;
	 /* (I/sigma) squared */
	roughiv = (igp->igr_roughi / igp->igr_roughv) * igp->igr_roughi;
	if (roughiv < 4.) igp->igr_minmod = MAX_MINMODEL;	/* weak refls*/
	else if (roughiv > 36.) igp->igr_minmod = 0.3 * MAX_MINMODEL;
	else igp->igr_minmod = (1. - 0.7 * (roughiv - 4.) / 32.) * MAX_MINMODEL;
	if (igp->igr_roughv < DFUZZ)
	  {
		if (igp->igr_flags & IR_VERBOSEANY) fprintf(igp->igr_fpl,
		 "Failure in summation: roughv = %7.2f\n", igp->igr_roughv);
		lrp->obs_flags |= I_OVERLAP; return EOK;
	  }
	if ((igp->igr_roughi > 1.) && (roughiv > 4.))
		lrp->obs_flags &= ~(I_TOOWEAK);
	else	lrp->obs_flags |= I_TOOWEAK;
	kab_bg(igp);	/* correct the Kabsch background values */
	bg_model(igp);	/* modify (XYOm) model for bgnd values */
	(void)sharpmodel(0, roughiv, igp); /* adjust direct space model*/
	(void)sharpmodel(1, roughiv, igp); /* adjust Kabsch model */
	igp->igr_missf = getmissf(igp); /* recompute `missf' in model */
	lrp->obs_ssq = ssqcalc(igp); /* record squared norm of diffrac vec*/
	/* determine the intensity and variance in
	(X,Y,omega) and (alpha,beta,gamma) space */
	do_sums(0, xyom, igp); /* summations for the X,Y,omega data */
	do_sums(1, kabg, igp); /* summations for the Kabsch data */
	/* if we're using frms 1-8 rather than 0-7, record that: */
	lrp->obs_lowfrm += jzdel;	lrp->obs_hifrm += jzdel;
	if (igp->igr_flags & (IR_VERBOSE2 | IR_VERBOSETEMP))
	  {
		sum_refl(0, tv, xyom, igp); sum_refl(1, tv, kabg, igp);
	  }
	/* store the appropriate values in `lrp->obs_...
	We use obs_ii, iunfit, etc. for the direct-space answers;
	we use obs_shift, obs_widob, obs_bgunfit, and obs_reflno as
	repositories for the Kabsch values.  Specifically
	lrp->obs_shift = 16 * (Ifit(Kabsch) - Ifit(direct-space))
	lrp->obs_widob = 8 * (sigfit(Kabsch) - sigfit(direct-space))
	lrp->obs_bgunfit = 16 * (Isum(Kabsch) - Isum(direct-space))
	lrp->obs_reflno = 8 * (sigsum(Kabsch) - sigsum(direct-space)) */
	lrp->obs_ii = xyom[0];
	lrp->obs_sigma = xyom[1] * igp->igr_aduperph;
	lrp->obs_iunfit = xyom[2];
	lrp->obs_sunfit = xyom[3] * igp->igr_aduperph;
	lrp->obs_gof = sqrt(xyom[18]);
	lrp->obs_shift = flimits(16. * (kabg[0] - xyom[0]), -32766., 32766.);
	lrp->obs_widob = flimits(8. * igp->igr_aduperph * (kabg[1] - xyom[1]),
		-32766., 32766.);
	lrp->obs_bgunfit = flimits(16. * (kabg[2] - xyom[2]), -32766.,32766.);
	lrp->obs_reflno = flimits(8. * igp->igr_aduperph * (kabg[3] - xyom[3]),
		-32766., 32766.);
	if (tv)
	  {
		igp->igr_flags &= ~IR_VERBOSEANY;	xeflush(igp->igr_fpl);
	  }
	tv = 0;
	if (((thisf = lrp->obs_ii) > 1.) && (thisf * thisf > (igp->igr_iosmin *
	 igp->igr_iosmin * lrp->obs_sigma))) lrp->obs_flags &= ~(I_TOOWEAK);
	else	lrp->obs_flags |= I_TOOWEAK;
	llp = igp->igr_thislog;
	if (!lrp->obs_flags)
	  {
		(llp->outstrong)++;
		if (igp->igr_flags & IR_UPD_MODEL) f_after_refl(igp);
	  }
	if (verror(0 < xgwrite((void *)lrp, XG_OBSREF, igp->igr_fpo),
		"write an observed reflection to the output")) return EWRITERR;
	/* record properties of this reflection for image-by-image stats */
	(llp->outrefl)++;	/* increment count of total obs in frame */
	llp->iis_in += thisf;	/* increment average-intensity calc */
	if (lrp->obs_sigma > FUZZ)
		llp->iis_ios += (tios = thisf / sqrt(lrp->obs_sigma));
	else	tios = 0.;
	if (tios > DFUZZ) /* increment sums used in D2, I000, B calculations */
	  {
		llp->iis_d2 += 1.; delpa = lamco * (double)(lrp->obs_ssq);
		llp->iis_ssq += delpa;
		llp->iis_b += delpa * delpa;
		llp->iis_lni += (lthis = log(tios));
		llp->iis_i000 += delpa * lthis;
	  }
	if (!(lrp->obs_flags & I_TOOWEAK))
	  {
		delpa = ((double)(lrp->obs_xob - lrp->obs_xpred)) /
			igp->igr_xscal;	llp->iis_rmx += delpa * delpa;
		delpa = ((double)(lrp->obs_yob - lrp->obs_ypred)) /
			igp->igr_yscal;	llp->iis_rmy += delpa * delpa;
		delpa = lrp->obs_omob - lrp->obs_ompred;
		llp->iis_rmz += delpa * delpa;
	  }
	(rinr[lreg])++; /*we've now seen this one*/
	if (tv)
	  {
		igp->igr_flags &= ~IR_VERBOSEANY;	xeflush(igp->igr_fpl);
	  }
	tv = 0;
	if (igp->igr_flags & IR_VERBOSEANY) xeflush(igp->igr_fpl);
	return EOK;
}

int getrefls(int workframe, INTEGRUN *igp)
{ /*	This is used to analyze all the refls found in frame `workframe',
	whether they peak in that frame or if they peak in neighboring frames*/
	int		lreg, result;
	double		cpsp, cpsq, spsq;
	static u_short	refls_in_region[NREGIONS];
	REFPARS		*xp;
	OBSREF		*lrp;
	FRAMELOG	*llp;
	
	igp->igr_tl0 = NOW;	/* get the first and last peaks in the frame: */
	xp = igp->igr_xp;
	igp->igr_rfmin = rforder(((double)workframe) - 0.001, igp);
	igp->igr_rfmax = rforder(((double)(workframe + 1)) + 0.001, igp);
	igp->igr_thislog = llp = igp->igr_logst +
		(workframe % (igp->igr_nmbox - 1));
	llp->predrefl = (igp->igr_rfmax - igp->igr_rfmin) + 1;
	llp->outrefl = 0;	llp->iis_in = llp->iis_ios = 0.;
	llp->iis_rmx = llp->iis_rmy = llp->iis_rmz = 0.;
	llp->iis_d2 = llp->iis_ssq = llp->iis_b = 0.;
	llp->iis_lni = llp->iis_i000 = 0.;
	for (lreg = 0; lreg < NREGIONS; lreg++) refls_in_region[lreg] = 0;
	/* go through the reflections associated with the current frame.
	Find an initial estimate of each reflection's integrated intensity
	and its <X>, <Y>, and <phi> values.  Obtain accurate intensity and
	centroid values by profile-fitting and summation */
	igp->igr_zmin = workframe - igp->igr_zhalfw;
	igp->igr_zmax = igp->igr_zmin + igp->igr_zwidth;
	/* parameters used in the polarization calculations */
	cpsq = cos(igp->igr_polnormrot * RADC);	cpsq *= cpsq;
	spsq = 1. - cpsq;	cpsp = sqrt(cpsq * spsq);
	igp->igr_e1c = igp->igr_polfrac + (1. - 2. * igp->igr_polfrac) * spsq;
	igp->igr_e2c = igp->igr_polfrac + (1. - 2. * igp->igr_polfrac) * cpsq;
	igp->igr_exc = 2. * (1. - 2. * igp->igr_polfrac) * cpsp;
	for (lrp = igp->igr_rfmin, llp->outstrong = 0;
	 lrp <= igp->igr_rfmax; lrp++)
	   {
		igp->igr_reflp = lrp;
		if ((!lrp->obs_flags) && (EOK != (result = intcalc(
		 refls_in_region, igp)))) return result;
		if (!(lrp->obs_flags & I_ALREADYOUT)) /* record fate */
		  {
			if (lrp->obs_flags & I_NUNRELIABLE)
				(igp->igr_reflfates[0400])++;
			else igp->igr_reflfates[(lrp->obs_flags & 0377)]++;
		  }
		lrp->obs_flags |= I_ALREADYOUT;
	   }
	/* count how many masks we've contributed to in this run */
	for (lreg = NREGIONS, llp->maskupd = 0; lreg--; )
		if (refls_in_region[lreg]) (llp->maskupd)++;
	/* flush the output at this point */
	xeflush(stdout);
	if (igp->igr_fpl != stdout) xeflush(igp->igr_fpl);
	igp->igr_tl1 = NOW;	/* record time req'd for this step */
	llp->sumsecs = igp->igr_tl1 - igp->igr_tl0;
	return EOK;
}

/* end of intrefl.c */
