/* intsetup.c -- routines for setting up a run of `integrate' in XGEN

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	23 Mar 2007: bugfix in manageice call: it _should_ return non-null
	14 Feb 2007: intdefaults: allow for Aviex detector types
	27 Dec 2006: -j: allow for polynomial _and_ regional remappings
	28 Nov 2006: pre-integrate over twice one refinement range
	17 Jun 2006: replace act_pixel with activmask
	 5 Jun 2006: intr_integ moved here from integmain as a static
	29 Nov 2005: makemask: define sizes of rgb->m3_bx, etc. before
		calling init3dvec
	 5 Nov 2005: merge integmask.c into here
	 5 Oct 2005: housekeeping
	16 Aug 2005: MISSING_PIXEL concept is gone; eliminate references to it
	 8 Aug 2005: -m changes meaning for value <= 0.1, not 1;
		specify regional adjustments with -g12 or -z12 flags
	28 Jun 2005: -j option for remapping (X,Y) distortions if requested
	22 Jun 2005: intdefaults() now declared as static
	15 Jun 2005: background kept as single-level pointer, not double
	11 May 2005: check return status from xgjournal
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	 8 Sep 2004: make IR_IGNOREEXPOS the default -- only invoke the
		scaling for exposure if the -z8 flag is set;
		Correct the -z option so it turns on IR_YCORREL as it should.
	 5 Sep 2004: use ../xglib/xgpredict.c routines; polarization handled
	 with envtopol(), etc.; eliminate prpars(), resval(), detec_limit()
	29 Aug 2004: routines used only here are declared as static; cosmetics
	18 Aug 2004: replace static globals with INTEGRUN structure elements
	 2 Aug 2004: cosmetic update
	27 Jul 2004: in mkintbord(), get the XY scale factors via dimscale()
	15 Jun 2004: imageflag bit for gzipping is now correctly abbreviated;
		take full advantage of specf and speci in intswitch();
		add in catchall flags "-g" and "-z" in intswitch();
		set up detector-type-sensitive defaults in integ_defaults();
		polarization done using environment variable POLARIZATION_TYPE
	25 Mar 2004: runflags sometimes starts with IR_LONGJUMP;
	 set most file poitners to NULL after closing them; cosmetic changes
	21 Mar 2004: call force_populate() in integ_setup so that the
		kablim() calls, etc. are guaranteed to work right
	24 Jan 2004: eliminate fpp, since refparin() doesn't need it
	21 Jan 2004: license message inserted and initial comments reorganized
	 3 Jan 2004: globals replaced by elements of INTEGRUN structures
	13 Dec 2003: elim 8 unused variables from makemask()
		and `nicer' from integ_setup()
	22 Nov 2003: replace global variables with members of INTEGRUN struct
	 8 Nov 2003: eliminate ar00, etc. calculations;
		use rphi[] instead of ome, chi, phi
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Aug 2003: alter management of ice-rings so we call an external
		routine to read the list of ice-ring resolution ranges
	13 Aug 2003: move getregion to ../xglib/xgenmask.c
	 5 Aug 2003: if the user specifies a frame range that is wider
		than what appears in the FRAMES index, impose
		the limits of the FRAMES index.
	10 May 2003: dimension of igp->igr_heads is igp->igr_nmbox,
		not zwidth + 1
	17 Apr 2003: rename `frameno' to `fmno'
	20 Mar 2003: set exposure time to 1 sec if unspecified
	14 Jan 2003: announce adu/photon value even when we're using
		the value that is standard for the current detector type
	 5 Jan 2003: reorganized source-file header
	 4 Jan 2003: images aren't gzipped with -D until final pass through.
	 3 Jan 2003: fixed bug in flagproc() call: 'X' arg had been missing
	28 Dec 2002: 'D' option gzips the images rather than deleting them.
	15 Nov 2002: extern declarations made here rather than in integext.h
	 1 Nov 2002: use of REGIONPROFILE structures for reflection box sizes
		and altered REGIONATT structures for region attributes
	20 Sep 2002: tell_dvcoef() reports results of REGIONPROFILE calcs;
		move CLF_* defines to xgen.h
	25 Mar 2002: removed MSI attachments
	15 Dec 2000: in detec_limit(), correct for the translational offset
	15 Oct 1999: in integ_setup(), pre-integrate over the same range
		of frames as the backgrounding effort.
	15 Dec 1997: in integ_setup(), allow for ranges of plane spacings
		to be excluded, typically because they're in ice-rings.
	 8 Nov 1996: minor changes for X-GEN checkin at MSI
	30 Sep 1995: user can specify #adu/photon during integration
	15 Sep 1995: in integ_setup(), define the mininum number of
		images that can be validly processed
	24 Aug 1995: fix display of exposure time
	 6 Jul 1995: cleanmask: extrapolate masks as well as flattening them
	31 May 1995: "register" -> "register int" throughout
	12 May 1995: make_ftab takes no argument; totalframes is a global
	10 Mar 1995: correct bug in deciding when to spread the binary profile
	 2 Mar 1995: provide for plane backgrounds; diagnostics to stdout
		frame # arg in pix2cm, cm2pix; shared-memory alloc and free;
	 	cross-reference added
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
		background image is 32-bits per pixel
	 4 Nov 1994: record END of pixels-in-reflections list (thishd->datae)
	 5 Oct 1994: use xgread to input REFPARS data
	 2 Sep 1994: allow as few as 8 frames to be integrated
	14 Jun 1994: removed big lookup tables -> no `mkxytables' routine
		reworked memory for PIXENT structures
	 5 May 1994: source-file header cleaned up; POLYDET -> XGEN;
		names changed from POLYDET to XGEN;
		added profile-flattening and extrapolation to clean_wmask;
		fixed bug in call to recen_vec in clean_wmask
	17 Mar 1994: eliminate duplicate declaration of `bz2'
	16 Mar 1994: several changes related to 3D model profiles;
		Major rewrite: many routines moved to ../xglib/xgenmask.c
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version
	 3 May 1988: in integ_setup(), estimate how much space we'll need
		for reflections by predicting a tiny range.
	18 Jan 1988: in prpars(), rotate the orientation matrix from
		(0,0,0) to the the position of the beginning of the run;
		also: use paralyzable model for deadtime loss
	 6 May 1987: in mkintbord(), disallow the default active-pixel mask.

   Purpose: initial file I/O, memory allocations, model-profiles, and set-up of
	permanent data over a run of the Bragg integration functionality.
   Cross references: integmain.c, ../xglib/xgenmask.c

*************************************************************************/
#define		NEED_TIME	1
#define		NEED_ERRNO	1
#include	"integdefs.h"
#define		DEF_REFLS_USED	16000	/*alloc space for >= this many refls*/
#define		ZSLOP		0.05
#define		DEFAULT_MINY2SQ	(0.9e-3)	/* minimum allowed value of
 the square of the y component of rotated scatt vector == 1/Lorenz factor */
#define		DEF_NEWINFLU	(0.0625)	/* frac of bgnd from curr.frm*/
#define		DEF_IOVSIGMIN	10	/* default I/sigma used in refinement*/
#define		DEFAULT_BUNCH	40	/* def. # frames before refinement*/
#define		MODELFMAX	(400.)
#define		MODELISQSUM	(1000.)
#define		MINSLOPPY	(MODELFMAX / 4.)

static int clean_wmask(int region, INTEGRUN *igp)
{ /*	This rationalizes the form of the three-dimensional profile as follows:
	(1) if the model profile is seriously flat or very badly miscentered,
		we substituted a weighted mean of the other model profiles.
	(2) we recenter the model profile.
	(3) we apply a contiguity check that (among other things) sets the
		weakest pixels to zero.
	(4) we renormalize the profile.
	The routine returns the # of pixels still included after all this. */
	int		ipass, lx, lxy, lxyz, ly, lz, nonzpix, z;
	float		*flb, *flp;
	double		isum, maxf, delv[3], epars[14], idv[3];
	REGIONPROFILE	*rgb, *rgp;

	rgb = igp->igr_rgb;
	rgp = rgb + region;	/* get pointer to current model */
	lx = rgp->m3_bx;	ly = rgp->m3_by;	lz = rgp->m3_bz;
	lxy = lx * ly;	lxyz = lxy * lz;
	/* see if it's blank; if it is, substitute the weighted mean
	 of all the other profiles, with the weighting strongest for
	 the nearest neighbors. */
	if (blankf(rgp->m3_xyop, lz)) meanprofile(rgb, region);
	/* determine the centroid and "intensity" of the profile */
	idv[0] = idv[1] = idv[2] = 0.;
	isum = gen_centroid(rgp->m3_xyop, NULL, lx, ly, lz, idv, delv);
	/* if the profile is degenerate (no intensity) do the averaging
	 bit again. If it still looks bad, give up.*/
	if (isum < FUZZ)
	  {
		meanprofile(rgb, region);
		isum = gen_centroid(rgp->m3_xyop, NULL,lx,ly,lz,idv, delv);
		if (isum <= FUZZ)
		  {
			printf("failed to find valid 3D profile in region %d\n",
				region);	return 0;
		  }
	  }
	/* recenter the profile based on the centroid we just found */
	flb = (float *)(igp->igr_wintv);	/* some convenient memory */
	for (ipass = 0; ipass < 4; ipass++)
	   {
		recen_vec(rgp->m3_xyop, delv, flb, lx, ly, lz);
		idv[0] = idv[1] = idv[2] = 0.;
		isum = gen_centroid(flb, NULL, lx, ly, lz, idv, delv);
		if (isum <= FUZZ)
		  {
			printf(
 "unable to obtain valid 3D profile in recentering pass %d, region %d\n",
				ipass, region);	return 0;
		  }
		if ((delv[0] < 0.4) && (delv[1] < 0.4) && (delv[2] < 0.2))
			break;
		cpymem((void *)rgp->m3_xyop, (void *)flb,
			lxyz * sizeof (float));
	   }
	if (verror((ipass < 4), "adequately recenter the profile"))
		return 0;
	cpymem((void *)rgp->m3_xyop, (void *)flb, lxyz * sizeof (float));
	/* apply the contiguity check */
	if (verror(contig(rgp, -1), "apply contiguity check to profile"))
		return 0;
	/* get rid of pixels on the periphery by asking for a best-fit
	 Gaussian ellipsoid boundary correction */
	(void)ellsoid(rgp, GAUSSIAN, CLF_BOTH, 0.009, 4., epars);
	/* renormalize the profile in such a way that the
	 new reflections will influence it fairly strongly. */
	for (flp = rgp->m3_xyop, maxf = z = 0; z < lxyz; z++, flp++)
		if (*flp > maxf) maxf = *flp;
	if (verror(maxf > FUZZ, "renormalize model profile")) return 0;
	maxf = MODELFMAX / maxf;	rgp->m3_isqsum = MODELISQSUM;
	if (rgp->m3_nobs < 10) rgp->m3_nobs = 10;
	if (rgp->m3_nobs > 1000) rgp->m3_nobs = 1000;
	for (flp = rgp->m3_xyop, nonzpix = z = 0; z < lxyz; z++, flp++)
	   {
		*flp *= maxf;	if (*flp > FUZZ) nonzpix++;
	   }
	return nonzpix;
}

static int makebmask(REGIONPROFILE *rgp, REGIONATT *bmp)
{ /*	This constructs a 2-D binary mask from a 3-D mask.
	It also determines the fraction of the 3D mask's intensity that is
	contained within the centermost 5x7x7 pixels of it. */
	register int	x, y, z;
	int		lx, lx2, lxy, ly, ly2, lz, lz2, nmodels, shift0, shift1;
	float		*flp;
	double		allf, allfx, allfy, allfz, dx, dy, dz;
	u_char		*ip;

	allf = allfx = allfy = allfz = 0;
	nmodels = rgp->m3_nmodels;	lx = rgp->m3_bx;
	ly = rgp->m3_by;	lz = rgp->m3_bz;	lxy = lx * ly;
	lx2 = lx / 2;	ly2 = ly / 2;	lz2 = lz / 2;
	bmp->ra_nmodels = nmodels;	bmp->ra_bx = lx;
	bmp->ra_by = ly;	bmp->ra_bz = lz;
	for (z = 0, dz = -lz2, flp = rgp->m3_xyop; z < lz; z++, dz += 1.)
	   {
		shift1 = shift0 = 1 << z; /* bit-shift for this frame */
		if (z) shift1 |= 1 << (z-1); /* bit shift for prev.fm */
		if (z < lz - 1) shift1 |= 1 << (z+1); /* ..and next fm */
		for (y = 0, dy = -ly2, ip = &(bmp->ra_incl[0]);
		 y < ly; y++, dy += 1.)
		 {
		   for (x = 0, dx = -lx2;
		    x < lx; x++, ip++, dx += 1., flp++)
		   {
			allf += *flp;	/* look at each pixel : */
			allfx += *flp * dx * dx; allfy += *flp * dy * dy;
			allfz += *flp * dz * dz;
			if (*flp > MINSLOPPY)
			  { /* bright pixel: turn it and its neighbors on*/
				*ip |= shift1;
				if (x) *(ip-1) |= shift0;
				if (x < lx - 1) *(ip+1) |= shift0;
				if (y) *(ip - lx) |= shift0;
				if (y < ly - 1) *(ip + lx) |= shift0;
			  }
			/* weak, nonzero pixel: turn only it on */
			else if (*flp) *ip |= shift0;
		    }
		   }
	   }
		/* always set the four corners of the 15x15 to zero */
	ip = &(bmp->ra_incl[0]);
	*ip = *(ip + (lx - 1)) = *(ip + (lxy - lx)) = *(ip + (lxy - 1)) = 0;
	/* Compute approximate width sigmas in x,y,and z for the region.
	Note that we don't let these values fall about 0.6 because then the
	error-function correction is 1 anyway, and we don't let them rise above
	a correction factor of 4, for which we get about 15 for sigma.  We've
	found through bitter experience that the corrections we need to get
	out of these numbers are bigger than the models give us, so
	we make these sigmas bigger so the corrections will be bigger.
	Hence the factors of 1.2 in these equations */
	if (allf < DFUZZ) bmp->ra_sigx = bmp->ra_sigy = bmp->ra_sigz =
		3. * sqrt(2.);
	else {
		bmp->ra_sigx = flimits(1.2 * sqrt(2. *allfx / allf),
			0.619, 15.22);
		bmp->ra_sigy = flimits(1.2 * sqrt(2. *allfy / allf),
			0.619, 15.22);
		bmp->ra_sigz = flimits(1.2 * sqrt(2. *allfz / allf),
			0.442, 10.87);
	     }
	return EOK;
}

static int getzwidth(REGIONPROFILE *rgp, double frac)
{ /*	This returns the number of frames (<=7) outside of which no more
	than `frac' of the total intensity in the mask cwmp is found.
	Note that `maxinmask' is already known when we run this. */
	register int	x, y, z;
	int		lx, ly, lz, lz2;
	float		*flp;
	double		absheight, left, lfrac[PFRAMES];

	lx = rgp->m3_bx;	ly = rgp->m3_by;	lz = rgp->m3_bz;
	lz2 = rgp->m3_bz / 2;	left = 1. - frac;
	for (z = 0, flp = rgp->m3_xyop; z < lz; z++)
	 for (lfrac[z] = y = 0; y < ly; y++)
	  for (x = 0; x < lx; x++, flp++) lfrac[z] += *flp;
	for (absheight = z = 0; z < lz; z++) absheight += lfrac[z];
		/* ultra-weak profiles: return -1 as an error return */
	if (absheight < DFUZZ) return -1;
	for (z = 0; z < lz; z++) lfrac[z] /= absheight;
	if ((absheight = lfrac[lz2]) > left) return 1;
	for (z = 1; z < lz2; z++)
	   {
		absheight += lfrac[lz2 - z] + lfrac[lz2 + z];
		if (absheight > left) return z + z + 1;
	   }
	return lz;
}

int upd_masks(INTEGRUN *igp)
{ /*	Set up cleaned-up weighted masks and recreate the binary masks */
	register int	region, result;
	int		nmodels;
	REGIONPROFILE	*rgb, *rgp;

	rgb = igp->igr_rgb;	nmodels = rgb->m3_nmodels;
	for (region = 0, rgp = rgb; region < nmodels; region++, rgp++)
	   {
		(void)clean_wmask(region, igp); /* rationalize the mask */
		if (EOK != (result = makebmask(rgp, igp->igr_bmask + region)))
			return result;
	   }
	return EOK;
}

int regionmem(INTEGRUN *igp)
{	/* This sets up memory allocations for 3-D binary and floating-point
	model profiles */
	int		lm, lx, lxy, lxyz, ly, lz, mx, my, mz;
	float		*flp;
	static u_char	*bmb;
	REGIONPROFILE	*rgb;

	/* for now we set profile dimensions as constants. We'll trust
	 the ones we read off the REGIONPROFILE structures later */
	lx = BIGSPOTWID;	ly = BIGSPOTHT; lz = PFRAMES;
	lxy = lx * ly;		lxyz = lxy * lz;	lm = NREGIONS;
	mx = 13;		my = 13;		mz = 9;
	/* reserve memory for the various types of profiles */
	if (NULL == igp->igr_wmask)
	  {
		if (NULL == (AVOID *)(igp->igr_wmask = (MASK_3D *)callop(lm,
		 sizeof (MASK_3D),"wmask in intsetup"))) return EALLOC;
	  }
	rgb = igp->igr_rgb;
	if (NULL == rgb)
	  {
		if (NULL == (AVOID *)(rgb = (REGIONPROFILE *)callop(lm,
		 sizeof (REGIONPROFILE), "rgb in intsetup")))
			return EALLOC;
		if (NULL == (AVOID *)(flp = (float *)callop(lm * lxyz,
		 sizeof (float), "rgb->m3_xyop in intsetup"))) return EALLOC;
		rgb->m3_xyop = flp;	igp->igr_rgb = rgb;
	  }
	rgb->m3_bx = lx;	rgb->m3_by = ly;	rgb->m3_bz = lz;
	rgb->m3_qx = mx;	rgb->m3_qy = my;	rgb->m3_qz = mz;
	rgb->m3_nmodels = lm;
	if (NULL == (AVOID *)(igp->igr_bmask = (REGIONATT *)callop(lm,
	 sizeof (REGIONATT), "bmask in intsetup"))) return EALLOC;
	if (NULL == (AVOID *)(bmb = (u_char *)callop(lm * lxyz,
	 sizeof (u_char), "bmb in intsetup"))) return EALLOC;
	(igp->igr_bmask)->ra_incl = bmb;
	/* this creates space to use for recentering the profiles */
	init3dvec(igp);
	return EOK;	/* everything worked if we got to here */
}

static int makemasks(INTEGRUN *igp)
{	/* This reads in the 3-D weighted masks and constructs the
	floating-point versions thereof and the the 3-D binary masks
	from them.  It also determines `zwidth', the max number of frames
	actually summed to integrate the data */
	register int	region, result;
	int		lm, lx, lxyz, ly, lz, mx, my, mz, zframes;
	float		*flp;
	static u_char	*bmb, *bmq;
	MASK_3D		*wmp;
	REGIONPROFILE	*rgb, *rgp;
	REGIONATT	*bmp;

	if (NULL == (rgb = igp->igr_rgb)) /* get memory if needed */
	  {
		if (EOK != (result = regionmem(igp))) return result;
	  }
	/* open the 3D profile file, read 'em in, and convert them to
	 the floating-point form. */
	if (qerror((AVOID *)(igp->igr_fpmask = xgopen("wmask","r")),
	 "open input 3-D weighted mask file")) return EOPENERR;
	bmb = (igp->igr_bmask)->ra_incl;
	lx = rgb->m3_bx;	ly = rgb->m3_by;	lz = rgb->m3_bz;
	mx = rgb->m3_qx;	my = rgb->m3_qy;	mz = rgb->m3_qz;
	lm = rgb->m3_nmodels;	lxyz = lx * ly * lz;	flp = rgb->m3_xyop;
	for (region = 0, rgp = rgb, wmp = igp->igr_wmask;
	 region < lm; region++, rgp++, wmp++, flp += lxyz)
	   {
		if (0 >= xgread((void *)wmp, XG_MASK_3D, igp->igr_fpmask))
			return EREADERR;
		rgp->m3_xyop = flp;
		rgp->m3_bx = lx; rgp->m3_by = ly; rgp->m3_bz = lz;
		rgp->m3_qx = mx; rgp->m3_qy = my; rgp->m3_qz = mz;
		rgp->m3_nmodels = lm;	rgp->m3_imodel = region;
		rgp->m3_isqsum = rgp->m3_nobs = 0;
		if (EOK == wmptovec(wmp, rgp->m3_xyop))
		  {
			rgp->m3_isqsum = wmp->phielt[3].isqsum;
			rgp->m3_nobs = wmp->phielt[3].mnobs;
		  }
	   }
	/* if any of the 3D profiles started out blank, we average the others
	 to get a proper value in that profile.  Then we clean up the
	 resulting profile and convert it to a binary model. */
	for (igp->igr_zwidth = region = 0, rgp = rgb, bmp = igp->igr_bmask,
	 bmq = bmb; region < lm; region++, rgp++, bmp++, bmq += lxyz)
	   {
		bmp->ra_incl = bmq;
		if (!clean_wmask(region, igp)) return EBADVALUE;
		if (EOK != (result = makebmask(rgp, igp->igr_bmask + region)))
			return result;
		if ((zframes = getzwidth(rgp, igp->igr_fracout)) >
		 igp->igr_zwidth) igp->igr_zwidth = zframes;
	   }
	igp->igr_zhalfw = igp->igr_zwidth / 2;
	return EOK;
}

int mkintbord(INTEGRUN *igp)
{ /*	This constructs a lookup table for determining whether various points
	are actually in the data array.  We use the active-mask pixel list to
	build this border; for each x we find the largest and smallest active
	y value.  We define the minimum and maximum to be 256 for x == 0,1,
	(width-1),(width-2).  We insist that the user supply an AMASK file.
	N.B. This can't be static because it's called from t_ref.c */
	register int	x, y;
	int		detword;
	REFPARS		*xp;
	static u_char	**locamask;

	if (verror(!access(trname("amask"), 4),"open input active-pixel file"))
		return EOPENERR;
	xp = igp->igr_xp;	xp->xdetwid = (igp->igr_spl)->v_detw;
	xp->xdetht = (igp->igr_spl)->v_deth;
	igp->igr_xscal = dimscale(xp->xdetwid);
	igp->igr_yscal = dimscale(xp->xdetht);
	detword = ((xp->xdetht) << 16) | xp->xdetwid;
	/* get memory for a bunch of detector-size-dependent stuff */
	if (NULL == (igp = manageigp(011, detword))) return EALLOC;
	igp->igr_ybotb = igp->igr_ytopb + xp->xdetwid;
	/* grab the local active-pixel mask and remember the pointer to it.
	 We don't have to read it in here: that's been done in prpars */
	if (NULL == (locamask = activmask((int)(xp->xdetwid),
	 (int)(xp->xdetht), 1))) return (errno == ENOMEM) ? EALLOC : EREADERR;
	x = xp->xdetwid / 8;	if (xp->xdetwid % 8) x++; x *= xp->xdetht;
	cpymem((AVOID *)igp->igr_curinbg, (AVOID *)(*locamask), (size_t)x);
	for (x = 2; x < xp->xdetwid - 2; x++)
	   {
		for (y = 1; y < xp->xdetht / 2; y++)
			if (NULL != activmask(x, y, 041)) break;
		igp->igr_ytopb[x] = y;
		for (y = xp->xdetht - 2; y > xp->xdetht / 2; y--)
			if (NULL != activmask(x, y, 041)) break;
		igp->igr_ybotb[x] = y;
	   }
	igp->igr_ytopb[0] = igp->igr_ytopb[1] = igp->igr_ytopb[xp->xdetwid-1]
		= igp->igr_ytopb[xp->xdetwid - 2] = xp->xdetht / 2;
	igp->igr_ybotb[0] = igp->igr_ybotb[1] = igp->igr_ybotb[xp->xdetwid-1]
		= igp->igr_ybotb[xp->xdetwid - 2] = xp->xdetht / 2;
	return EOK;
}

static void sumintinput(FILE *fp, INTEGRUN *igp)
{	/* this writes a summary of the run-time inputs to stream fp */

	static char	*flist[] = {"uparams", "uspline", "frames", "bgimage",
		"wmask", "urefls", "xlog", NULL};
	static char	ftypes[] = {1, 1, 1, 2, 3, 2, 2, 0 };
	REFPARS		*xp;

	xp = igp->igr_xp;
	ftypes[3] = (NULL == igp->igr_fpbgnd) ? 2 : 3;
	ftypes[6] = (igp->igr_fpl == stdout)  ? 0 : 2;
	header(fp, PROCESSING, flist, ftypes);	/* summarize files */
	fprintf(fp, "Integration range: frame %4d through frame %4d\n",
		igp->igr_zmi, igp->igr_zma);
	fprintf(fp, "Resolution limits: %6.2f to %6.2f Angstroms\n",
		xp->xresmin, xp->xresmax);
	fprintf(fp, "Reflections with |sy| > %6.4f will be integrated\n",
		sqrt(igp->igr_miny2sq));
	fprintf(fp,
 "Unit cell and orientational parameters will be refined once per %4d frames\n",
		(int)(igp->igr_bunch));
	fprintf(fp,
 "Reflections with I/sigma > %7.2f will be used in refinement\n",
		igp->igr_iosmin);
	if (igp->igr_minsoi > DFUZZ) fprintf(fp,
 "Machine error term = ((I * %6.2f)squared) will be added to variances\n",
		igp->igr_minsoi);
	fprintf(fp,
 "Fractional influence of each new frame on background will be %6.4f\n",
		igp->igr_newinflu);
	fprintf(fp, "The initial background image %s be smoothed before use\n",
		(igp->igr_flags & IR_JAGGED) ? "WILL NOT" : "WILL");
	if (igp->igr_imfl & DARK_CURRENT) fprintf(fp,
 " %6.4lf of the darkcurrent image %s will be subtracted from each input\n",
		((double)igp->igr_idarkfrac) / 8000., trname("DARKIMAGE"));
	fprintf(fp, "Pixels with >= %6d ADUs are considered unreliable\n",
		(int)(igp->igr_minunrely));
	fprintf(fp,
	 "Angle between polarization-plane normal and omega axis = %6.3f\n",
		igp->igr_polnormrot);
	fprintf(fp, "Fraction of incident radiation polarized = %6.3f\n",
		igp->igr_polfrac);
	if (fabs(igp->igr_polfrac - 0.5) < DFUZZ) fprintf(fp,
	 "i.e. incident radiation is completely unpolarized\n");
	else if (fabs(igp->igr_polfrac - 1.) < DFUZZ) fprintf(fp,
	 "i.e. incident radiation is fully polarized\n");
	if (igp->igr_flags & IR_VERBOSE2) fprintf(fp,
		"Extremely detailed diagnostics will be printed\n");
	if (igp->igr_flags & IR_VERBOSE1) fprintf(fp,
		"Moderately detailed diagnostics will be printed\n");
	fprintf(fp,
	 "(X,Y,omega) predictions %s adjusted according to regional offsets\n",
	 (igp->igr_flags & IR_REGIONADJUST) ? "will be" : "will not be");
	fprintf(fp,
	 "Model profiles will%sbe updated as data processing proceeds\n",
		(igp->igr_flags & IR_UPD_MODEL ? " " : " NOT "));
	if (igp->igr_imfl & GZIP_AFTER_CLOSE) fprintf(fp,
	 "Each frame will be compressed after processing\n");
	if ((!(igp->igr_flags & IR_USEOLDBG)) || (NULL == igp->igr_fpbgnd))
	  {
		fprintf(fp,
 "A new background image will be calculated over the first %3d frames\n",
			(int)(igp->igr_nbginit));
		fprintf(fp,
 "Background-initialization images will be read from latest to earliest\n");
	  }
	else fprintf(fp,
 "Background will be taken from existing data in %s\n", trname("BGIMAGE"));
	if (igp->igr_flags & IR_YCORREL) fprintf(fp,
 "We assume correlation among background values along vertical stripes\n");
	else fprintf(fp,
 "assume NO correlation among background values along vertical stripes\n");
	fprintf(fp,
 "at most %1d frames of data will be examined for each reflection\n",
	 (int)(igp->igr_zwidth));
	fprintf(fp,
 "\tbased on allowing %8.4f of intensity outside those %d frames\n",
		igp->igr_fracout, (int)(igp->igr_zwidth));
	fprintf(fp, "Count values %s be corrected for deadtime loss\n",
		igp->igr_flags & IR_LOSSCORR ? "WILL" : "WILL NOT");
	fprintf(fp,"\tStarting Unit Cell and Orientational Parameters:\n");
	fprintf(fp,
"       a       b       c    alpha   beta  gamma     omega     chi     phi\n");
	fprintf(fp,
	 " %7.2f %7.2f %7.2f   %6.2f %6.2f %6.2f   %7.2f %7.2f %7.2f\n",
		xp->xabc[0], xp->xabc[1], xp->xabc[2],
		xp->xang[0], xp->xang[1], xp->xang[2],
		xp->xdelom, xp->xdelchi, xp->xdelphi);
	if (igp->igr_bgout < 32767) fprintf(fp,
	 "Background images will be written to disk once per %d frames\n",
		(int)(igp->igr_bgout));
	if (igp->igr_flags & IR_PLANEBGND) fprintf(fp,
 "Backgrounds estimated from least-squares plane fits within images\n");
	if (igp->igr_overphot > DFUZZ) fprintf(fp,
 " Variances will be calculated assuming %6.3f adu per photon\n",
		igp->igr_overphot);
	if (igp->igr_flags & IR_ICERINGS) describe_rings(fp);
	if (igp->igr_flags & IR_IGNOREEXPOS) fprintf(fp,
 "No rescaling for exposure time will be performed\n");
	if (igp->igr_remaplim < 9999999.) fprintf(fp,
 "Detector will be remapped if XY residual > %8.4f using %s fits\n",
		0.0001 * fabs(igp->igr_remaplim),
		(igp->igr_remaplim > DFUZZ) ? "regional" : "polynomial");
	if (fp != stderr) xeflush(fp);
}

static void intswitch(int ac, char *av[], INTEGRUN *igp)
{ /*	process flags, if any.  Currently defined flags are:
	"-a<val>" sets the angle between the normal to the polarization
		plane and the rotation axis to <val> in degrees.
		This value is 90 for a synchrotron with horizontal omega.
	"-b<val>" averages the background over <val> frames instead of
		the default of 16 frames.
	"-c<val>" causes the minimum value of I/sigma used in refinement
		to change from the default value (10) to <val>.
	"-d" causes the program to delete each frame after it is read.
		This is suitable for processing data after they've already
		been backed up or in situations where no backup is needed.
	"-e<val>" obliges `integrate' to define the number of adu
		per photon to be <val>. If this isn't specified, then
		the number of adu per photon is set based on the
		detector type (1 for Poisson counters).
	"-f<val>" alters the fraction of the intensity allowed to be found
		outside the frames integrated.  Thus "integrate -f0.01"
		causes at most 5 frames to be integrated if the widest
		model profile has less than 0.01 of its total intensity
		outside the centermost 5 frames.
	"-g" tells the program NOT to rescale intensities based on the exposure
		time, presumably because the exposure times are unreliable.
		If "-g" is followed by a numeric argument, it takes on other
		meanings, as specified below.
	"-h" reads a file with environment variable name ICERING as a paired
		list of resolution ranges to exclude because they are considered
		to be invalid. Thus if $ICERING contains the lines
	3.67 3.63
	2.24 2.20
	1.93 1.89
	1.48 1.44
	1.31 1.27
		Then data from 3.67A to 3.63A, etc. would not be integrated.
	"-i<val>" changes the background-updating fraction from 1/16 to <val>.
	"-j<val>" tells the program to remap the detector nonlinearities
		as part of each refinement procedure, provided the (X,Y)
		residual is greater than <val>, where <val> is in 10000ths of
		pixels. Thus -j1000 means remap if the RMS error is > 0.1 pix.
	"-k<val>" tells the program that the upper resolution limit, i.e.,
		the minimum D-spacing, of the data should be <val> Angstroms.
	"-l"	turns off the deadtime loss corrections.
	"-m<val>" Defines the minimum unreliable #counts/pixel or #adus/pixel
		to be <val>. If a pixel has <val> or more counts in it,
		it is assumed to be an unreliable measurement. If
		three or more pixels are unreliable in this sense, the
		reflection is not integrated. If only one or two are
		unreliable, the profile-fitted value is substituted
		for the observed value and the reflection is output.
		To preserve compatibility with old versions, if <val>
		is < 1, it has an entirely different meaning:
		it causes the variance of each reflection to be inflated
		by the square of a machine error term (<val>*I).
		By default <val> = 0.
	"-n" forces collection of a new background image rather
		than using the image in file "bgimage".
	"-o<val> sets the polarization fraction to <val>. This value
		is 0.5 for unpolarized radiation, 1.0 for fully polarized
		radiation, and 0.4722 for a graphite monochromator.
	"-p" causes backgrounds to be estimated from least-squares plane
		fits to the data rather than by pixel-by-pixel updating.
	"-q" sets up the polarization formula to fully polarized input
		radiation with the rotation axis perpendicular to
		the normal to the polarization plane.
	"-r<val>" causes the program to refine the crystal parameters
		once every <val> frames instead of every 40 frames.
	"-s" causes more diagnostics than normal to be written to the
		logfile, but not as many as with "-v".
	"-t" causes subtraction of a dark-current image from each data
		frame at the moment in which it's read in.
	"-u" sets up the polarization correction appropriate for
		unpolarized input radiation.
	"-v" causes some diagnostics to be written to stdout
		as well as to the logfile, and to produce more detailed
		diagnostics IN the logfile.
	"-w<val>" directs `integrate' to write the background image (BGIMAGE)
		out every <val>'th frame. By default the BGIMAGE is only
		written out at the end of background initialization and
		at the end of the data run.
	"-x" sets up the polarization formulate to fully polarized input
		radiation with the rotation axis parallel to the normal
		to the polarization plane.
	"-y<val>" sets the minimum value of |sy| to <val>.  Here |sy| is the
		component of the diffraction vector perpendicular to the
		plane containing the direct beam and the rotation axis.
		By default this value is 0.03.  The purpose of this flag is to
		allow users to adjust how stringent the cutoff for reflections
		close to the rotation axis will be.
	"-z" high-correlation-along-Y (default with XENGEN): tells the program
		that the detector's geometric and electronic characteristics
		are such that a high degree of correlation among effective
		pixel sizes within a COLUMN is to be expected. */
	char	**dcp, **flags;
	int	speci;
	double	specf;

	flags = flagproc(ac, av, "ABCDEFGHIJKLMNOPQRSTUVWXYZ");
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	 {
	   if ('\0' != *((*dcp) + 1))
	     {
		specf = atof((*dcp) + 1);	speci = atoi((*dcp) + 1);
	     }
	   else	{
		specf = 0.;	speci = 0;
		}
	   switch (**dcp)
	   {
		case 'A': /* specify angle between normal to polarization
				plane and the crystal rotation axis */
			igp->igr_polnormrot = specf;	break;
		case 'B':	/* #fms averaged for bgnd initialization */
			if ((1 <= speci) && (speci <= 999))
				igp->igr_nbginit = speci;	break;
		case 'C':	/* minimum I/sigma in refinement */
			if (1. < specf) igp->igr_iosmin = specf;	break;
		/* 28 Dec 2002: 'D' gzips images rather than deleting them */
		case 'D':	/* gzip frames after reading them */
			igp->igr_imfl |= GZIP_AFTER_CLOSE;		break;
		case 'E':	/* set # of adu per photon */
			if ((0.01 <= specf) && (specf <= 1000.))
				igp->igr_overphot = specf;	break;
		case 'F': /* change fraction for finding max # of fms */
			if ((0. < specf) && (specf < 0.2))
				igp->igr_fracout = specf;
			break;
		/* June '04: -g<val> turns on various flags.
		 -g by itself turns on IR_IGNOREEXPOS for backward compat */
		case 'G':
		 if (speci > 0)
		  {
		   switch (speci)
		     {
			case  1: igp->igr_flags |= IR_JAGGED;	break;
		  	case  2: igp->igr_flags |= IR_UPD_MODEL;	break;
		  	case  3: igp->igr_flags |= IR_LOSSCORR;	break;
		  	case  4: igp->igr_flags |= IR_YCORREL;	break;
		  	case  5: igp->igr_flags |= IR_LONGJUMP;	break;
		  	case  6: igp->igr_flags |= IR_USEOLDBG;	break;
		  	case  7: igp->igr_flags |= IR_PLANEBGND; break;
			case  8: igp->igr_flags |= IR_IGNOREEXPOS; break;
			case  9: igp->igr_flags |= IR_ICERINGS; break;
			case 10: igp->igr_flags |= IR_VERBOSE1; break;
			case 11: igp->igr_flags |= IR_VERBOSE2; break;
			case 12: igp->igr_flags |= IR_REGIONADJUST; break;
		      } /* close the switch (speci) */
		    } /* close the "if (speci > 0" clause */
		   else igp->igr_flags |= IR_IGNOREEXPOS;	break;
		case 'H': igp->igr_flags |= IR_ICERINGS;	break;
		case 'I':	/* change background-updating parameter */
			if (DFUZZ < specf && specf < 1.)
				igp->igr_newinflu = specf;	break;
		/* -j<val> tells the program to re-map detector (X,Y) if the
		 RMS (X,Y) residual > igp->igr_remaplim. The units are
		 pixels * 10000.  If <val> is missing or zero,
		 then -j means to activate jagged background mode.
		 If <val> is negative, we use polynomial remapping rather than
		 regional remapping */
		case 'J': if (fabs(specf) > DFUZZ) igp->igr_remaplim = specf;
			  else igp->igr_flags |= IR_JAGGED;	break;
			/* don't update model profiles: */
		 /* use 'k<val>' to specify an upper resolution limit;
		 without a value it freezes the model profiles */
		case 'K': if (fabs(specf) > FUZZ) igp->igr_dmaxspec = specf;
			  else igp->igr_flags &= ~IR_UPD_MODEL;	break;
			/* turn off deadtime losses: */
		case 'L': igp->igr_flags &= ~IR_LOSSCORR;	break;
		/* the M flag means different things depending on the value
		 associated with it. If the value is less than 0.1, it's
		 a machine-error addend for variances; if it's greater,
		 it's the minimum I/sigma for reliable data */
		case 'M':	/* add machine error value to each variance*/
			if (specf <= 0.1)
			  {
				if (FUZZ < specf) igp->igr_minsoi = specf;
			  }
			else igp->igr_minunrely = specf;
			break;
			/* ignore the old BGIMAGE file (if present): */
		case 'N': igp->igr_flags &= ~IR_USEOLDBG;	break;
		case 'O': /* polarization fraction. This value is 0.5 for
			   unpolarized radiation, 1 for fully polarized,
			   and about 0.4722 for a graphite monochromator */
			if ((DFUZZ <= specf) && (specf <= 1.))
				igp->igr_polfrac = specf;
			else if (specf > 1.) igp->igr_polfrac = 1.;	break;
		case 'P': igp->igr_flags |= IR_PLANEBGND;	break;
		case 'Q': /* fully-polarized synchrotron, horiz omega */
			igp->igr_polfrac = 1.;	igp->igr_polnormrot = 90.;
			break;
		case 'R':	/* number of frames between refinements */
			if (1 <= specf) igp->igr_bunch = specf;	break;
			/* semi-verbose output: */
		case 'S': igp->igr_flags |= IR_VERBOSE1;	break;
		case 'T':
			igp->igr_imfl |= DARK_CURRENT;
			if (specf > DFUZZ) igp->igr_idarkfrac = 8000. * specf;
			break;
		case 'U': igp->igr_polfrac = 0.5; break; /* unpolarized input*/
			/* very verbose output: */
		case 'V': igp->igr_flags |= IR_VERBOSE2;	break;
		case 'W': igp->igr_bgout = ((0. < specf) && (specf < 32767.)) ?
				specf : 32767;	break;
		case 'X': /* fully-polarized synchrotron, vertical omega */
			igp->igr_polfrac = 1.;	igp->igr_polnormrot = 0.;
			break;
		case 'Y': if ((DFUZZ <= specf) && (specf <= 2.))
				igp->igr_miny2sq = specf * specf;	break;
		/* June '04: -z<val> turns off various flags.
		 -z by itself turns on IR_YCORREL for backward compat */
		case 'Z':
		 if (speci > 0)
		  {
		    switch (speci)
		     {
			case  1: igp->igr_flags &= ~IR_JAGGED;	break;
		  	case  2: igp->igr_flags &= ~IR_UPD_MODEL;	break;
		  	case  3: igp->igr_flags &= ~IR_LOSSCORR;	break;
		  	case  4: igp->igr_flags &= ~IR_YCORREL;	break;
		  	case  5: igp->igr_flags &= ~IR_LONGJUMP;	break;
		  	case  6: igp->igr_flags &= ~IR_USEOLDBG;	break;
		  	case  7: igp->igr_flags &= ~IR_PLANEBGND; break;
			case  8: igp->igr_flags &= ~IR_IGNOREEXPOS; break;
			case  9: igp->igr_flags &= ~IR_ICERINGS; break;
			case 10: igp->igr_flags &= ~IR_VERBOSE1; break;
			case 11: igp->igr_flags &= ~IR_VERBOSE2; break;
			case 12: igp->igr_flags &= ~IR_REGIONADJUST; break;
		      } /* close the switch (speci) */
		    } /* close the "if (speci > 0)" clause */
			/* high correlation along Y: */
		   else igp->igr_flags |= IR_YCORREL;	break;
	    }
	 }
	if (igp->igr_idarkfrac > 16000) igp->igr_idarkfrac = 16000;
	if (igp->igr_imfl & DARK_CURRENT)
		igp->igr_imfl |= (igp->igr_idarkfrac << 16);
	return;
}

static int intdefaults(INTEGRUN *igp)
{ /*	This establishes run-time default parameters for a run of `integrate'
	based on the kind of detector we're using and related stuff. */
	int		detype, ssize, stype, detdim[2];
	static int	defbg[] = { 16, 16, 16, 12, 16 } ;  /*#fms initbg*/
	static double	defiosmin[] = { 8., 8., 8., 11., 8.}; /* I/s refine*/
	static double	defadu[] = /* # of adu per pixel */
	 { 1.,1.,1.,4.,4.,5.,4.,4.,5.,5.,3.8,1.8,1.4,1.,1.8,1.8,4.5,4.5 };
	static int	defunrely[] =
	 { 65000, 65000, 65000,  4094, 65000,240000,240000, 65000, 65000,
	   65000, 60000, 60000, 60000, 60000, 60000, 60000, 60000 } ;

	if (qerror((AVOID *)(igp->igr_fmi = ftables(3)),
	 "read frame-index file")) return EOPENERR;
	igp->igr_flags |= IR_UPD_MODEL | IR_IGNOREEXPOS;
	igp->igr_bgout = 32767;	igp->igr_fracout = 0.015;
	igp->igr_idarkfrac = 6400;	igp->igr_minsoi = 0.;
	igp->igr_remaplim = 10000000.;
	/* wait till the images arrive if currently absent */
	igp->igr_imfl = WAIT_IF_ABSENT;	igp->igr_miny2sq = DEFAULT_MINY2SQ;
	/* default settings for polarization parameters:
	 angle between omega and polarization direction = 0;
	 fraction polarized (typical for graphite): 0.4722 */
	if (0 >= (detype = dettype(detdim))) return EBADVALUE;
	stype = detype & 0377;	ssize = detype >> 8;
	if (stype < 5) /* SDMS, Xentronics, old Argonne CCD, Fast */
	  {
		igp->igr_nbginit = defbg[stype];
		igp->igr_iosmin = defiosmin[stype];
		igp->igr_bunch = 24;
		igp->igr_flags |= IR_LOSSCORR | IR_USEOLDBG;
		igp->igr_newinflu = 0.0625;
		if (stype < 3) igp->igr_flags &= ~IR_PLANEBGND;
		else igp->igr_flags |= IR_PLANEBGND;
	  }
	else {
		igp->igr_nbginit = 8; /* everything else (IPs and CCDs) */
		igp->igr_iosmin = 11.;
		igp->igr_bunch = 12;
		igp->igr_newinflu = 0.08;
		igp->igr_flags &= ~(IR_LOSSCORR | IR_USEOLDBG);
		igp->igr_flags |= IR_PLANEBGND;
	     }
	if (stype <= 18) /* specify # adu/photon and saturation count */
	  {
		igp->igr_overphot = defadu[stype];
		igp->igr_minunrely = defunrely[stype];
	  }
	else {
		igp->igr_overphot = 4.5;
		igp->igr_minunrely = 60000;
	     }
	return EOK;
}

static void verbose_search(INTEGRUN *igp)
{ /*	Set up a search for reflections to be printed out verbosely */
	short		itry, ntry, **trlp;
	char		*rlname, rlbuf[40];
	static FILE	*fpref;

	rlname = trname("reflist");	ntry = 0;
	if (NULL != (fpref = xgopen(rlname, "r")))
	  {
		for ( ; NULL != fgets(rlbuf, 36, fpref); ) ntry++;
		(void)fseek(fpref, 0L, 0);
		if (NULL == manageigp(041, (size_t)ntry)) return;
		for (trlp = igp->igr_trlb; NULL != fgets(rlbuf, 36, fpref);
		 trlp++) (void)sscanf(rlbuf, "%hd %hd %hd",
			*trlp, (*trlp) + 1, (*trlp) + 2);
		(void)fclose(fpref);	fpref = NULL;
		fprintf(igp->igr_fpl, "%d Refls to be printed:\n", ntry);
		for (trlp = igp->igr_trlb, itry = ntry; itry--; trlp++)
			fprintf(igp->igr_fpl, " %3d %3d %3d\n",
			 **trlp, *((*trlp) + 1), *((*trlp) + 2));
	  }
	igp->igr_nrverbose = ntry;
}

static void intr_integ(void)
{	static char	interruptmsg[] = "due to user-generated interrupt";

	orderly_exit(interruptmsg);
}

int integ_setup(int argc, char *argv[], INTEGRUN *igp)
{ /*	This sets up for a run of `integrate' by decoding runtime inputs,
	setting up the frame index, opening the log file,
	and scribbling therein. It also allocates memory for various arrays. */
	register int	fate, fmno;
	int		beamx, ffound, headoff, inifm, ir, ix, iy, lastfr;
	int		lm, lx, ly, lz, naveraged, nhere, ref_per_frame;
	int		region, result, survey_flags, use100ths;
	unsigned	needed;
	double		lfrac, mcorr, testfound;
	char		*locpname, *testname;
	u_char		*tbmp;
	MAIL		*thishd;
	REFPARS		*xp;
	DATARUN		*drup;
	FRAMEINDEX	*fmind;
	REGIONATT	*bmp;
	REGIONPROFILE	*rgb;
	OBSREF		*lrp;
	short		nperreg[NREGIONS];

	if (EOK != (ix = xgjournal(argc, argv))) return ix; /* journal it */
	if (EOK != (ix = intdefaults(igp))) return ix;
	fmind = igp->igr_fmi;	igp->igr_dmaxspec = 0.;
	inifm = fmind->fi_firstfm;	lastfr = fmind->fi_lastfm;
	/* get default values for polarization parameters from environment */
	(void)envtopol(&(igp->igr_polfrac), &(igp->igr_polnormrot));
	/* process switch-style arguments. see `intswitch' for details */
	intswitch(argc, argv, igp);
	for (argc--, argv++; (argc) && !allnumeric(*argv); argc--) argv++;
	/* process remaining arguments: frame numbers of start and end,
	and, if present, a high-resolution minimum D spacing and
	a low-resolution maximum D spacing. */
	if (argc < 2)
	  {
		igp->igr_zmi = inifm;	igp->igr_zma = lastfr;
	  }
	else {
		igp->igr_zmi = atoi(*argv++);
		igp->igr_zma = atoi(*argv++); argc -= 2;
		if (igp->igr_zmi < inifm) igp->igr_zmi = inifm;
		if (igp->igr_zma > lastfr) igp->igr_zma = lastfr;
	     }
	if (igp->igr_nbginit > (igp->igr_zma - igp->igr_zmi))
		igp->igr_nbginit = igp->igr_zma - igp->igr_zmi;
	if (argc)
	  {
		argc--;	igp->igr_dminspec = atof(*argv++);
	  }
	else igp->igr_dminspec = 0.;
	if (argc)
	  {
		argc--;	igp->igr_dmaxspec = atof(*argv++);
	  }
	igp->igr_rfend = NULL;
	/* initialize the reflection-prediction stuff
	 and insert some of the output of that process into igp */
	if (EOK != (ix = intpredini(igp))) return ix;
	/* construct the elliptical border of the system */
	if (EOK != (result = mkintbord(igp))) return result;
	igp->igr_reflp = igp->igr_rfbase; /* start `refls' pointer properly */
	/* create or append to the crystallographic logfile */
	if (NULL == (igp->igr_fpl = propen("xlog"))) igp->igr_fpl = stdout;
	igp->igr_tl0 = NOW;
	if (qerror((AVOID *)(igp->igr_fpo = xgopen("urefls", "w")),
		"create output reflections file")) return EOPENERR;
	for (fate = 0; fate < NFATES; fate++) igp->igr_reflfates[fate] = 0;
	igp->igr_fpbgnd = (igp->igr_flags & IR_USEOLDBG ?
		xgopen("bgimage", "r") : NULL);
	if (EOK != (result = makemasks(igp))) return result;
	/* Dec 97: allow for ranges of plane spacings to be excluded,
	 typically because they're in ice-rings */
	if ((igp->igr_flags & IR_ICERINGS) && (NULL == manageice(7)))
		igp->igr_flags &= ~IR_ICERINGS;
	/* Sep 95: we check the minimum number of images to process now.
	 We could get useful results with as few as igp->igr_zwidth images. */
	if (igp->igr_zma < igp->igr_zmi + igp->igr_zwidth - 1)
	  {
		locpname = xgprogname(NULL, 0);
		fprintf(igp->igr_fpl,
		 "%s: can't integrate fewer than %d images\n",
			locpname, (int)(igp->igr_zwidth));
		printf("%s: can't integrate fewer than %d images\n",
			locpname, (int)(igp->igr_zwidth));
		return EBADVALUE;
	  }
	xp = igp->igr_xp;
/* Determine where the boundaries of the twelve detector regions are. */
	beamx = (xp->xdetwid - 1) / 2 +
		(int)(MEANPIXPERCM * xp->xplated * tan(xp->xswing));
	igp->igr_boundp = make_boundaries(beamx, xp->xdetwid, xp->xdetht);
	/* Check to make sure that we've reserved enough space for predicted
	reflections by predicting a tiny range and counting how many arrived */
	testfound = pred_bunch((int)(igp->igr_zmi), 1, 1, igp);
/* XGEN 3.1: we reserve memory for the active pixels in the frames.  We
 get this estimate from the number of active pixels in the predicted
 reflections for this test frame. We underestimate here and get a
 better value when we start using this memory */
	rgb = igp->igr_rgb;
	lx = rgb->m3_bx;	ly = rgb->m3_by;	lz = rgb->m3_bz;
	lm = rgb->m3_nmodels;
	for (region = 0, bmp = igp->igr_bmask; region < lm; region++, bmp++)
	   {
		nhere = 0;
		for (iy = 0, tbmp = &(bmp->ra_incl[0]); iy < ly; iy++)
		   {
			for (ix = 0; ix < lx; ix++, tbmp++)
			   if (*tbmp) nhere +=
				 ((*tbmp >> 7) & 01) + ((*tbmp >> 6) & 01) +
				 ((*tbmp >> 5) & 01) + ((*tbmp >> 4) & 01) +
				 ((*tbmp >> 3) & 01) + ((*tbmp >> 2) & 01) +
				 ((*tbmp >> 1) & 01) + (*tbmp & 01);
		   }
		nperreg[region] = ((double)nhere) * 0.2 + 1;
	   }
	for (lrp = igp->igr_rfbase, igp->igr_nactive = ir = 0; ir < testfound;
	 ir++, lrp++) igp->igr_nactive += nperreg[lrp->obs_region];
	if (igp->igr_nactive < DEF_REFLS_USED * 5)
		igp->igr_nactive = DEF_REFLS_USED * 5;
	ref_per_frame = ((int)(testfound /
	 ((double)(igp->igr_zwidth + 1))) * 4) / 3;
	/* we're going to need oodles of memory for other purposes,
	 so we get rid of the big chunk we allocated before for reflections
	 and replace it with one that's the right size for the data we're
	 going to actually deal with */
	needed = ref_per_frame * (igp->igr_bunch + igp->igr_zwidth + 1);
	if (needed < 5000) needed = 5000;
	if (NULL == (igp = manageigp(021, needed))) return EALLOC;
	igp->igr_reflp = igp->igr_rfbase;
	igp->igr_rfend = igp->igr_rfbase + needed;
	/* make sure we know how many adu/photon we'll be using
	 even if the user doesn't override the relevant value */
	sumintinput(igp->igr_fpl, igp);	sumintinput(stdout, igp);
	ffound = testfound / ((double)(igp->igr_zwidth + 1));
	fprintf(igp->igr_fpl,
		"Roughly %d reflections per frame will be predicted\n",ffound);
	fprintf(stdout,"Roughly %d reflections per frame will be predicted\n",
		ffound);
	fprintf(igp->igr_fpl , " %6d pixels / frame expected to be active\n",
		(int)(igp->igr_nactive));
	fprintf(stdout, " %6d pixels / frame expected to be active\n",
		(int)(igp->igr_nactive));
/* survey the first forty frames' headers to arrive at an
	average value of the exposure time.  This will be used in
	the renormalization of the counts from various frames.  Meanwhile,
	we set up the memory for the frames we're going to look at */
	mcorr = naveraged = 0.;
	lastfr = (igp->igr_zma < igp->igr_zmi + igp->igr_bunch ?
		igp->igr_zma : igp->igr_zmi + igp->igr_bunch);
/* pre-integrate over twice the range of frames as the backgrounding effort--
 except that if the backgrounding covers a very short range, we expand some.
 If the environment variable INTEG_NORESTART is set we skip all this */
	if (NULL != getenv("INTEG_NORESTART"))
		igp->igr_initend = igp->igr_zmi - 1;
	else {
		igp->igr_initend = lastfr + igp->igr_bunch;
		if (igp->igr_initend > igp->igr_zma)
			igp->igr_initend = igp->igr_zma;
		if ((ffound < 200) && (lastfr < igp->igr_zmi + 15))
		  {
			igp->igr_initend = igp->igr_zmi + 3 * igp->igr_bunch;
			if (igp->igr_initend > igp->igr_zma)
				igp->igr_initend = igp->igr_zma;
		  }
	     }
	survey_flags = (igp->igr_imfl | JUST_HEADER) & ~(GZIP_AFTER_CLOSE);
	thishd = igp->igr_heads;
	for (fmno = igp->igr_zmi; fmno <= lastfr; fmno++)
	  {
		headoff = fmno % (igp->igr_nmbox - 1);
		thishd = igp->igr_heads + headoff;
		if ((NULL != (testname= indexname(fmno))) &&
		  (FRAME_OK == read_image(testname, survey_flags, thishd)) &&
		  (thishd->eventcount != 0))
		  {
/* modified for XENGEN 2.1 to incorporate the paralyzable model */
			lfrac = losscorr(thishd, igp);
			use100ths = (igp->igr_flags & IR_IGNOREEXPOS) ?
				100 : thishd->elap100ths;
			mcorr += 1. / (((double)use100ths) * (1. - lfrac));
			naveraged++;
		  }
	   }
	/* having read through those, we can find the number of adu
	 per photon if the user hasn't already specified it */
	if (DFUZZ < igp->igr_overphot) igp->igr_aduperph = igp->igr_overphot;
	else {
		drup = thishd->runheader;
		igp->igr_aduperph = drup->run_aduperph;
		fprintf(igp->igr_fpl,
	 " Variances will be calculated assuming %6.3f adu per photon\n",
			igp->igr_aduperph);
		fprintf(stdout,
	 " Variances will be calculated assuming %6.3f adu per photon\n",
			igp->igr_aduperph);
	     }
	if (verror(naveraged, "read headers from any of the first few images"))
		return EREADERR;
	igp->igr_avexp100 = naveraged ? ((double)naveraged) / mcorr : 0;
	if (igp->igr_avexp100 <= DFUZZ) igp->igr_avexp100 = 100.;
/* catch user interrupts -- don't catch bus errors, segmentation violations,
 or floating-point exceptions. */
	signal(SIGINT, (sig_t)intr_integ);
	igp->igr_xbwidth = xp->xdetwid / NXBBINS;
	igp->igr_ybwidth = xp->xdetht / NYBBINS;
	fprintf(stdout,
	 " mean%sexposure time over the first %2d frames = %8.2f sec\n",
		((igp->igr_flags & IR_LOSSCORR) ? " loss-corrected " : " "),
		naveraged, igp->igr_avexp100 / 100.);
	fprintf(igp->igr_fpl,
	 " mean%sexposure time over the first %2d frames = %8.2f sec\n",
		((igp->igr_flags & IR_LOSSCORR) ? " loss-corrected " : " "),
		naveraged, igp->igr_avexp100 / 100.);
	/* populate the REFPARS structure internal to xgenmask.c */
	if (EOK != (ix = force_populate(NULL)))
	  {
		fprintf(stderr, "Unable to read run parameters: error %d\n",
			ix);
		return ix;
	  }
		/* get upper,lower bounds of diffraction-vector params: */
	kablim(igp->igr_heads, rgb, lz);
	tell_dvcoef(igp->igr_fpl, rgb);	/* announce those results */
	tell_dvcoef(stdout, rgb);
	verbose_search(igp);	/* set up verbose-printed reflections */
	return EOK;
}

/* end of intsetup.c */
