/* t_ref.c -- self-contained UREFLS refinement program
	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	 6 Oct 2005: housekeeping
	19 Apr 2007: reorganize use of "{"
	17 Jun 2005: in trefmain(): mkintbord called with &igs argument
	11 May 2005: check return status from xgjournal
	28 Aug 2004: call to prpars() replaced with call to intpredini()
	25 Jan 2004: all static globals converted to local variables
	21 Jan 2004: license message inserted and initial comments reordered
	 8 Jan 2004: elements of INTEGRUN struct replace extern globals
	23 Nov 2003: replace global variables with references to INTEGRUN
		structures
	19 Sep 2003: posfirst explicitly initialized in loop
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	 5 Jan 2003: reorganized source-file header
	15 Nov 2002: extern declarations made here rather than in integext.h
	 1 Nov 2002: eliminate FRAMEHALFWID definition
	25 Mar 2002: removed MSI attachments
	 5 Oct 1994: use xgread to input OBSREF data
	 5 May 1994: source-file header cleaned up;
		and names changed from POLYDET to XGEN
	16 Mar 1994: a few variables moved around
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version

   Purpose: reading refls from a UREFLS file and refining crystal and
	detector parameters with them using the same code as is used
	in the on-the-fly refinements.
   Cross references: ../integrate/intrefine.c
*************************************************************************/
#include	"integdefs.h"
#define		MIN_REFLS	15

int trefmain(int argc, char *argv[])
{ /* This does parameter refinements directly on UREFLS data */
	int		result;
	size_t		posfirst;
	unsigned	ninrefs;
	FILE		*fpi;
	OBSREF		urf;
	INTEGRUN	*igp;
	static char	*fnames[] = {"xlog", "uparams", "urefls", NULL};
	static char	ftypes[] = {2, 1, 1, 0 };

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	if (NULL == (igp = manageigp(7, 0))) return EALLOC;
	if (NULL == (igp->igr_fpl = propen("xlog"))) igp->igr_fpl = stdout;
	ftypes[0] = (igp->igr_fpl == stdout) ? 0 : 2;
	header(stdout, PROCESSING, fnames, ftypes);
	if (igp->igr_fpl != stdout)
		header(igp->igr_fpl, PROCESSING, fnames, ftypes);
	igp->igr_zmi = 1;	igp->igr_zma = 40;
	if (argc > 1)
	  {
		igp->igr_zmi = atoi(argv[1]);
		if (argc > 1) igp->igr_zma = atoi(argv[2]);
	  }
	printf(" Refinement will be performed over fms %3d to %3d\n",
		igp->igr_zmi, igp->igr_zma);
	fprintf(igp->igr_fpl,
	 " Refinement will be performed over fms %3d to %3d\n",
		igp->igr_zmi,igp->igr_zma);
	if (qerror((AVOID *)(fpi = xgopen("urefls", "r")),
		"open input urefls file")) return EOPENERR;
	/* go through the urefls file to see how many usable refls there are.*/
	for (posfirst= ninrefs = 0; 0 < xgread((void *)&urf, XG_OBSREF, fpi); )
	   if (igp->igr_zmi < urf.obs_omob)
	     {
		if (urf.obs_omob >= igp->igr_zma) break;
		else if (!urf.obs_flags)
		  {
			if (!ninrefs) posfirst = ftell(fpi);
			ninrefs++;
		  }
	     }
	if (ninrefs < MIN_REFLS)
	  {
		fprintf(stderr,
			"Run failed: only %2d usable reflection%sfound\n",
				(int)ninrefs, (ninrefs == 1 ? " " : "s "));
		return EBADVALUE;
	  }
	if (NULL == (igp = manageigp(021, (size_t)(ninrefs + 20))))
		return EALLOC;
	if (verror(1 + fseek(fpi, posfirst, 0), "rewind input urefls file"))
		return ESEEKERR;
	igp->igr_reflp = igp->igr_rfbase;
	/* determine border of active area */
	if (EOK != (result = mkintbord(igp))) return result;
	/* get the crystal and detector parameters set up */
	if (EOK != (result = intpredini(igp))) return result;
	for (ninrefs = 0; 0 < xgread((void *)&urf, XG_OBSREF, fpi); )
	   if (igp->igr_zmi < urf.obs_omob)
	     {
		if (urf.obs_omob > igp->igr_zma) break;
		else if (!urf.obs_flags)
		  {
			urf.obs_flags |= I_ALREADYOUT;	ninrefs++;
			cpymem((AVOID *)igp->igr_reflp, (AVOID *)(&urf),
				sizeof (OBSREF));
			(igp->igr_reflp)++;
		  }
	   }
	igp->igr_rflast = igp->igr_reflp;
	printf("%3d usable reflections found between frames %3d and %3d\n",
		(int)ninrefs, igp->igr_zmi, igp->igr_zma);
	igp->igr_bunch = igp->igr_zma - igp->igr_zmi + 1;
	upd_xtal(igp->igr_zma + (PFRAMES / 2) + 1, igp);
	if (NULL != fpi) (void)fclose(fpi);	fpi = NULL;
	(void)manageigp(0, 0);	globalkill();	return EOK;
}

/* end of t_ref.c */
