/* dispover.c -- utilities for displaying overlays onto `pdisplay' windows

	Copyright (c) 2003, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions; reorganize use of "{"
	23 Jun 2006: include xgenpred.h so the structure types are available
	22 Jun 2006: eliminate extra superpose_* calls in do_superpositions
	 5 Jun 2006: eliminate extra superpose_calib call in do_superpositions
	21 Jan 2005: write fixbutton() to replace toggleon().
	 5 Oct 2005: housekeeping
	17 Oct 2004: toggleon() used to keep the overlay buttons up-to-date
	 3 Oct 2004: color specifications are relative to di_nranges
	27 Aug 2004: routines used only here are declared as static;cosmetics
	 3 Apr 2004: superpose_border(): revive active-pixel list if needed
	27 Mar 2004: some diagnostics squelched if flags & DISPVERBOSEANY is 0
	21 Jan 2004: license message inserted
	 8 Jan 2004: make sure all diamonds get drawn in superpose_calib()
	 5 Jan 2004: replace global splp with element of DISP_ATTR structure;
		requires an extra arg to gettroids()
	21 Sep 2003: correct coordinates of blowup in overl_label()
	19 Sep 2003: superpose_predict returns the # refls predicted
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Sep 2003: remove explicit calls to xgaccess() because
		they're latent in the xgopen() calls
	15 Sep 2003: Initial version, created from pieces of dispx.c
 Relevant changes made to dispx.c:
	12 Aug 2003: add new overlays for resolution circles &
		row+column sums; change displays of predictions to
		reflections with large gamma values cover more images
	 6 Aug 2003: change Motif global variables to static globals
	 5 Jan 2003: reorganized source-file header
	25 Mar 2002: removed MSI attachments
	 5 May 1994: source-file header, POLYDET->XGEN
	 7 Dec 1993: initial X-GEN version

   Purpose: utilities for displays of overlays showing the detector border,
	reflection predictions, centroids, calibrations, calibration centroids,
	row-and-column sums, and resolution circles
	directly on an X window display governed under Motif.
	Unlike ../displays/dispcalc.c, these routines ARE Motif-specific,
		but they do not depend very much on X-GEN itself.
   Cross references: ./disp*.c, ../displays/disp*.c
*************************************************************************/
#define		NEED_ERRNO	1
#define		NEED_MOTIF	1
#include	"xgmdefs.h"
#include	"xgenpred.h"
extern PREDREF *pred_range(DISP_ATTRS *dispatt, int  *npredp);

static Widget fixbutton(XG_X11PROPS *xgp, int whichb)
{ /*	This checks for viability of the requested Widget.
	If it's absent, it calls make_buttons to try to create it and friends.
	If it's still absent, it returns NULL.
	If it's okay, then we toggle it on and return it. */
	int	iset, na;
	Widget	inwid;
	Arg	lda[1];

	if (NULL == (inwid = xgp->gp_dpush[whichb]))
	  {
		make_buttons();
		if (NULL == (inwid = xgp->gp_dpush[whichb]))
		  {
			fprintf(stderr, "Button heirarchy faulty : ");
			fprintf(stderr, "can't toggle button %d\n", whichb);
			return NULL;
		  }
	  }
	na = 0;	iset = 1;	XtSetArg(lda[na], XmNset, iset);	na++;
	XtSetValues(inwid, lda, na);	XtManageChild(inwid);	return inwid;
}

static int superpose_calib(XG_X11PROPS *olayp, int ismag)
{ /*	This overlays contours of the calibration data in the USPLINE
	file onto the rest of the display */
	int		ix, iy, ndiamonds, *cap;
	double		fimc[4];
	CALPOINT	*calp;
	DISP_ATTRS	*lattr, *maina;
	VSPLINE		*lsp;
	Window		windo;
	static int	*calpts, ncpts;
	static XPoint	*calxpt;
	static int	ncalxpts;

	/* create calibration contours */
	maina = olayp->gp_at;
	if (!((maina->di_disflags) & DISPCALIB_NEED)) return 0;
	if (NULL == fixbutton(olayp, FIRSTOVERL)) return 0;
	if (ismag)
	  {
		lattr = olayp->gp_bl;	windo = olayp->gp_maw;
	  }
	else {
		lattr = olayp->gp_at;	windo = olayp->gp_wi;
	     }
	if (ncpts <= 0) memzap((AVOID **)&calpts, "calpts in superpose_calib");
	if (NULL == calpts)
	  {
		if (NULL == (calpts = calib_contours(maina, &ncpts)))
		  {
			fprintf(stderr,
			 "Unable to obtain calibration contours: error %d\n",
				ncpts);
			return 0;
		  }
	  }
	/* special case: we want di_spl visible from both windows */
	else if (ismag) lattr->di_spl = (olayp->gp_at)->di_spl;
	cap = calpts + (ncpts - 1);
	if (!(lattr->di_wintype)) printf(
	 "Found %6d calibration contour points from [%4d,%4d] to [%4d,%4d]\n",
	 ncpts, (*calpts) >> 16, (*calpts & 0xFFFF),
	 (*cap) >> 16, (*cap & 0xFFFF));
	/* bugfix 8 Jan 04: we used to quit after discovering that there were
	 no contours to draw; this is wrong if there are diamonds to draw */
	if ((NULL == (calxpt = (XPoint *)calibdisp(calpts, lattr,
	 ncpts, &ncalxpts))) || (ncalxpts < 1))
	  {
		fprintf(stderr, "no contours drawn\n");
	  }
	else {
		printf(" %6d calibration-display structures produced\n",
			ncalxpts);
		xeflush(stdout);
		XDrawPoints(olayp->gp_di, windo, olayp->gp_ogc,
			calxpt, ncalxpts, CoordModeOrigin);
	     }
	/* draw little diamond-shaped boxes around the fiducial points */
	lattr->di_boxh = (lattr->di_wintype) ? 8 : 4;
	lsp = lattr->di_spl;
	for (ndiamonds = iy = 0, calp = lsp->v_cpt; iy < lsp->v_nrow; iy++)
	  for (ix = 0; ix < lsp->v_ncol; ix++, calp++)
	     {
		fimc[0] = calp->c_ptocx;	fimc[1] = calp->c_ptocy;
		if ((lattr->di_xylim[1] <= fimc[1]) &&
		 (fimc[1] < lattr->di_xylim[3]) &&
		 (lattr->di_xylim[0] <= fimc[0]) &&
		 (fimc[0] < lattr->di_xylim[2]))
		  {
			fdisrot((int)(lattr->di_roco), 1,
			 (int)(lattr->di_imwid), fimc);
			maina->di_currx = lattr->di_currx = fimc[2];
			maina->di_curry = lattr->di_curry = fimc[3];
			ndiamonds += diamond(olayp, ismag);
		  }
	     }
	if (ndiamonds > 0) printf(" %4d %s drawn in the %s window\n",
		ndiamonds, (ndiamonds == 1) ? "diamond" : "diamonds",
		(lattr->di_wintype) ? "magnified" : "image");
	return	ncpts;
}

static void superpose_border(XG_X11PROPS *olayp, int ismag)
{ /*	This displays the border of the active area of the detector
	as defined by the contents of the AMASK file */
	int		lwi, lhi, maxpt;
	DISP_ATTRS	*lattr, *maina;
	Window		windo;
	static XPoint	*bordpt = (XPoint *)0;

	maina = olayp->gp_at;
	if (!((maina->di_disflags) & DISPBORDER_NEED)) return;
	if (NULL == fixbutton(olayp, FIRSTOVERL + 2)) return;
	lwi = maina->di_imwid;	lhi = maina->di_imht;
	if (ismag)
	  {
		lattr = olayp->gp_bl;	windo = olayp->gp_maw;
	  }
	else {
		lattr = maina;	windo = olayp->gp_wi;
	     }
	/* sometimes we get here having killed off the border. Reactivate:*/
	if ((maina->di_disflags) & DISPBORDER_IN)
	  {
		if (NULL == activmask(lwi, lhi, 1))
			maina->di_disflags &= ~DISPBORDER_IN;
	  }
	if (!((maina->di_disflags) & DISPBORDER_IN)) /* read border now */
	  {
		if (NULL == activmask(lwi, lhi, 5))
		  {
			if (errno == ENOMEM) fprintf(stderr,
	"Memory-allocation failed in active-area determination\n");
			else fprintf(stderr,
	"Unable to read active-area mask from disk\n");
			return;
		  }
		maina->di_disflags |= DISPBORDER_IN;
	  }
	if (NULL != (bordpt = (XPoint *)mkborderolay(lattr,
	 (XG_Point *)bordpt, &maxpt))) (void)XDrawPoints(olayp->gp_di,
	 windo, olayp->gp_ogc, bordpt, maxpt, CoordModeOrigin);
	else printf("Unable to draw the border\n");
}

static void show_centroids(XG_X11PROPS *olayp, TROID *tr0, int ntr, int ismag)
{ /*	This prints a little diamond around each centroid obtained from `spots'
	or `integrate' followed by `reduce -c'.  It's color-coded,
	with the centroids previous to (fmno - 0.5) colored blue, the
	ones after (fmno+0.5) colored red, and the ones right on top
	of fmno colored white.  Since they're sorted on phi, we won't
	have to change GC's very often to do this.  We make the size of
	the diamond dependent on the intensity. */
	int		lfm, lhi, lwi, nranges, ntrdis;
	double		fimc[4];
	Pixel		prevcolor, thiscolor;
	TROID		*trl, *trp;
	DISP_ATTRS	*lattr, *maina;
	
	maina = olayp->gp_at;
	if (!((maina->di_disflags) & DISPCEN_NEED)) return;
	if (NULL == fixbutton(olayp, FIRSTOVERL + 3)) return;
	lattr = ismag ? olayp->gp_bl : olayp->gp_at;
	nranges = lattr->di_nranges;
	prevcolor = olayp->gp_dbcol[nranges];	trl = tr0 + ntr;
	lfm = maina->di_fmno;
	lwi = maina->di_imwid;	lhi = maina->di_imht;
	thiscolor = olayp->gp_col;
	for (trp = tr0, ntrdis = 0; trp < trl; trp++)
	   if ((2 < (fimc[0] = trp->xto)) && (2 < (fimc[1] = trp->yto)) &&
	       (fimc[0] < lwi - 2) && (fimc[0] < lhi - 2))
	     {
		thiscolor = (trp->frame0 < lfm - 0.5 ?
		olayp->gp_dbcol[nranges + 4]:
		 (trp->frame0 > lfm + 1. ?
		 olayp->gp_dbcol[nranges - 1] : olayp->gp_dbcol[nranges]));
		if (thiscolor != prevcolor)
			XSetForeground(olayp->gp_di, olayp->gp_ogc, thiscolor);
		prevcolor = thiscolor;	/* remember this color for next time*/
		lattr->di_boxh =
			flimits(log(trp->totinten / 100.) + 2., 2., 6.);
		if (lattr->di_wintype) lattr->di_boxh += 3;
		fdisrot((int)(lattr->di_roco), 1, (int)(lattr->di_imwid),fimc);
		maina->di_currx = lattr->di_currx = fimc[2];
		maina->di_curry = lattr->di_curry = fimc[3];
		ntrdis += diamond(olayp, ismag);
	   }
	if (maina->di_disflags & DISPVERBOSEANY)
	   printf(" %4d %s displayed in %s window\n",
		ntrdis, (ntrdis == 1) ? "centroid" : "centroids",
		(lattr->di_wintype) ? "blowup" : "main");
	if (thiscolor != olayp->gp_col)
		XSetForeground(olayp->gp_di, olayp->gp_ogc, olayp->gp_col);
}

static void super_calcen(XG_X11PROPS *olayp, int ismag)
{ /*	This prints a little diamond around each calibration centroid
	obtained from a run of `calibrate -o'.  All diamonds are white. */
	int		imwi, ldiflags, ncalce, ncshow, roco;
	double		fimc[4];
	CTROID		*catr, *catre;
	DISP_ATTRS	*lattr, *maina;
	Pixel		curcol;
	static CTROID	*catr0;

	maina = olayp->gp_at;
	if (!((ldiflags = maina->di_disflags) & DISPCALCEN_NEED)) return;
	if (NULL == fixbutton(olayp, FIRSTOVERL + 4)) return;
	lattr = ismag ? olayp->gp_bl : olayp->gp_at;
	roco = maina->di_roco;	imwi = maina->di_imwid;
	curcol = olayp->gp_col;
	if ((!(ldiflags & DISPCALCEN_IN)) &&
	 (NULL == (catr0 = get_calcen(&ncalce)))) return;
	catre = catr0 + ncalce;
	maina->di_disflags |= DISPCALCEN_IN;
	if (curcol != olayp->gp_col)
		XSetForeground(olayp->gp_di, olayp->gp_ogc, olayp->gp_col);
	lattr->di_boxh = (lattr->di_wintype) ? 7 : 4;
	for (ncshow = 0, catr = catr0; catr < catre; catr++)
	   {
		fimc[0] = catr->cxcen;	fimc[1] = catr->cycen;
		fdisrot(roco, 1, imwi, fimc);
		maina->di_currx = lattr->di_currx = fimc[2];
		maina->di_curry = lattr->di_curry = fimc[3];
		ncshow += diamond(olayp, ismag);
	   }
	printf(" %6d calibration centroids displayed\n", ncshow);
}

static void overl_label(int zy, int zx, int ismag,
	char *labstr, XG_X11PROPS *ovp)
{ /*	This puts a small label near (zy, zx) containing the string `labstr'
	into the main window */
	int		lx, nch, yu;
	Window		windo;
	DISP_ATTRS	*lattr;

	if (ismag)
	  {
		lattr = ovp->gp_bl;	windo = ovp->gp_maw;
		yu = zy;
	  }
	else {
		lattr = ovp->gp_at;	windo = ovp->gp_wi;
		yu = lattr->di_ht - 1 - zy;
	     }
	nch = strlen(labstr);	if (nch > 6) nch = 6;
	lx = zx - 5 * (nch / 2);	/* attempt to center it */
	XDrawString(ovp->gp_di, windo, ovp->gp_ogc, lx, yu, labstr, nch);
}

static void super_resolcircles(XG_X11PROPS *olayp, int ismag)
{ /*	This prints a circle for each of several resolution ranges.
	These calculations are time-consuming so we try not to do them
	too often. */
	int		igp, ngps, npts, onleft;
	int		dispxv, dispyv;
	double		tdval;
	Window		windo;
	XPoint		*currpts;
	DISP_ATTRS	*lattr, *maina;
	static char	labstri[9];
	static int	resinfo[28];
	static XPoint	*respts, *magpts;

	maina = olayp->gp_at;
	if (!((maina->di_disflags) & DISPRESOL_NEED)) return;
	if (NULL == fixbutton(olayp, FIRSTOVERL + 5)) return;
	if (ismag)
	  {
		lattr = olayp->gp_bl;	windo = olayp->gp_maw;
	  }
	else {
		lattr = olayp->gp_at;	windo = olayp->gp_wi;
	     }
	XSetForeground(olayp->gp_di, olayp->gp_ogc, olayp->gp_col);
	currpts = (lattr->di_wintype) ? magpts : respts;
	if (NULL == currpts) /* Create the resolution circles */
	  {
		if (NULL == (currpts = (XPoint *)grespts(lattr, resinfo)))
		  {
			fprintf(stderr, "Sorry. No resolution circles made\n");
			return;
		  }
	  }
	npts = resinfo[0];	ngps = resinfo[1];
	if (npts <= 0) return;
	/* now actually put the overlay onto the screen */
	XDrawPoints(olayp->gp_di, windo, olayp->gp_ogc, currpts, npts,
		CoordModeOrigin);
		/* label the circles. For overlays the label is placed either
		 below or to the left of center, depending on whether the
		 contour is running mostly horizontally or vertically. */
	onleft = (abs(lattr->di_currx - lattr->di_imwid / 2) <
		abs(lattr->di_curry - lattr->di_imht / 2));
	for (igp = 0; igp < ngps; igp++)
	   if (resinfo[igp + 2] < 10000)
	     {
		zeromem((AVOID *)labstri, 8);
		tdval = ((double)resinfo[2 + 2 * ngps + igp]) * 0.01;
		if (lattr->di_wintype)
		  {
			sprintf(labstri, "%5.2fA", tdval);
			if (onleft)
			  {
				dispyv = lattr->di_magh / 2;
				dispxv = lattr->di_magw / 2 - 26;
			  }
			else {
				dispyv = lattr->di_magh / 2 - 26;
				dispxv = lattr->di_magw / 2;
			     }
		  }
		else {
			sprintf(labstri, "%4.1fA", tdval);
			dispyv = resinfo[2 + ngps + igp];
			if (dispyv < 8) dispyv = 8;
			if (dispyv > lattr->di_wid - 8)
				dispyv = lattr->di_wid - 8;
			dispxv = resinfo[2 + igp];
			if (dispxv < 8) dispxv = 8;
			if (dispxv > lattr->di_ht - 8)
				dispxv = lattr->di_ht - 8;
		     }
		overl_label(dispyv, dispxv, ismag, labstri, olayp);
	     }
}

static void super_rowcolsums(XG_X11PROPS *olayp, int ismag)
{ /*	This puts out an overlay showing the summed intensity as a function 
	of X along columns and of Y along rows.
	For the moment they are unlabeled. */
	MAIL		*lhd;
	DISP_ATTRS	*lattr, *maina;
	Window		windo;
	static int	nrcpts;
	static ino_t	curino;
	static XPoint	*rcpts;

	maina = olayp->gp_at;
	if (!((maina->di_disflags) & DISPROWCOL_NEED)) return;
	if (NULL == fixbutton(olayp, FIRSTOVERL + 6)) return;
	if (ismag)
	  {
		lattr = olayp->gp_bl;	windo = olayp->gp_maw;
	  }
	else {
		lattr = olayp->gp_at;	windo = olayp->gp_wi;
	     }
	if (lattr->di_wintype) return; /* only allowed in main window*/
	XSetForeground(olayp->gp_di, olayp->gp_ogc, olayp->gp_col);
	lhd = lattr->di_head + maina->di_fmno % maina->di_nmbox;
	if ((!curino) || (curino != lhd->quadrant))
		memzap((AVOID **)&rcpts, "rcpts in super_rowcolsums");
	curino = lhd->quadrant;
	if (NULL == rcpts) /* Create row & column sums */
	  {
		if (NULL == (rcpts = (XPoint *)rowcolpts(lattr, &nrcpts)))
		  {
			printf("No row/column overlay created\n");
			return;
		  }
	  }
		/* now actually put the overlay onto the screen */
	if (nrcpts) XDrawPoints(olayp->gp_di, windo, olayp->gp_ogc,
		rcpts, nrcpts, CoordModeOrigin);
	else printf(" No points available for row & column sums\n");
}

static int superpose_predict(XG_X11PROPS *olayp, int ismag)
{ /*	This sets up a prediction of the spots peaking near the current frame
	and enables a display of an overlay of that prediction.
	It returns the number of displayed spots */
	int		disht, diswid, fmno, htdo, htup;
	int		imwi, needhead, nomxdel, nomydel, npredi;
	int		nranges, ntodisplay, roco, wido, wiup;
	short		bheight, bwidth, xmip, xmxp, ymip, ymxp, luh[3];
	double		maghoff, maghrat, magwoff, magwrat;
	double		thisx, thisy;
	double		dspac, fimc[4];
	PREDREF		*prp, *pre;
	DISP_ATTRS	*lattr, *maina;
	Pixel		prevcolor, thiscolor;
	Window		windo;
	static PREDREF	*locprb;

	maina = olayp->gp_at;	nranges = maina->di_nranges;
	if (!((maina->di_disflags) & DISPHKL_NEED)) return 0;
	if (NULL == fixbutton(olayp, FIRSTOVERL + 1)) return 0;
	if (ismag)
	  {
		lattr = olayp->gp_bl;	windo = olayp->gp_maw;
	  }
	else {
		lattr = olayp->gp_at;	windo = olayp->gp_wi;
	     }
	roco = lattr->di_roco;	imwi = lattr->di_imwid;	fmno = lattr->di_fmno;
	if (!((maina->di_disflags) & DISPHKL_IN))
	  {
		if (!gpars(lattr))
		  {
			fprintf(stderr, "Error reading parameters!\n");
			return 0;
		  }
	  }
	maina->di_disflags |= DISPHKL_IN;
	if (!(maina->di_disflags & DISPHKL_CONT)) rot2l(maina);
	if (NULL == (locprb = pred_range(maina, &npredi))) return 0;
	/* Now print a little box around each predicted spot. It's color-coded,
	with the spots previous to (fmno - 0.5) colored blue, the
	ones after (fmno+0.5) colored red, and the ones right on top
	of fmno colored white.  Since they're sorted on phi, we won't
	have to change GC's very often to do this */
	prevcolor = olayp->gp_col;
	htup = lattr->di_htup;	htdo = lattr->di_htdown;
	wiup = lattr->di_wiup;	wido = lattr->di_widown;
	if (lattr->di_wintype)
	  {
		maghrat = 8. * (double)htup / (double)htdo;
		magwrat = 8. * (double)wiup / (double)wido;
		maghoff = 80. - 8. * ((double)(lattr->di_disy));
		magwoff = 80. - 8. * ((double)(lattr->di_disx));
		nomxdel = nomydel = 3;
		disht = lattr->di_magh;	diswid = lattr->di_magw;
		if ((disht < 5) || (diswid < 5)) printf(
		 "Warning: disht, diswid = %d, %d\n", disht, diswid); 
	  }
	else { /* July 2002: make the box size in the main window
		 sensitive to the density of the pattern */
		ntodisplay = npredi;
		maghrat = (double)htup / (double)htdo;
		magwrat = (double)wiup / (double)wido;
		if (ntodisplay > 5000) nomxdel = nomydel = 1;
		else if (ntodisplay > 1250) nomxdel = nomydel = 2;
		else nomxdel = nomydel = 1;
		maghoff = magwoff = 0.;
		disht = lattr->di_ht;	diswid = lattr->di_wid;
		if ((disht < 5) || (diswid < 5)) printf(
		 "Warning: disht, diswid = %d, %d\n", disht, diswid); 
	     }
	needhead = 1;	pre = locprb + npredi;
	thiscolor = olayp->gp_dbcol[nranges + 4];
	for (prp = locprb, npredi = 0; prp < pre; prp++)
	   {
		thiscolor = (prp->pphi < fmno - 0.5 ?
		 olayp->gp_dbcol[nranges+4]:
		 (prp->pphi > fmno + 1. ?
		 olayp->gp_dbcol[nranges-1]: olayp->gp_dbcol[nranges]));
		if (thiscolor != prevcolor)
			XSetForeground(olayp->gp_di, olayp->gp_ogc, thiscolor);
		prevcolor = thiscolor;
		fimc[0] = prp->pxt;	fimc[1] = prp->pyt;
		fdisrot(roco, 1, imwi, fimc);
		thisy = fimc[3] * maghrat + maghoff;
		thisx = fimc[2] * magwrat + magwoff;
		if ((0 <= thisy) && (thisy < disht) &&
		 (0 <= thisx) && (thisx < diswid))
		  {
			ymip = thisy - nomydel;	ymxp = ymip + 1 + 2 * nomydel;
			if (ymip < 0) ymip = 0;
			if (ymxp >= disht) ymxp = disht - 1;
			xmip = thisx - nomxdel;	xmxp = xmip + 1 + 2 * nomxdel;
			if (xmip < 0) xmip = 0;
			if (xmxp >= diswid) xmxp = diswid - 1;
			bwidth = xmxp - xmip + 1; bheight = ymxp - ymip + 1;
			npredi++;	/* count the # of points drawn */
			XDrawRectangle(olayp->gp_di, windo, olayp->gp_ogc,
			 xmip, ymip, bwidth, bheight);
			if ((lattr->di_wintype) &&
			 (lattr->di_disflags & DISPVERBOSEANY))
			  {
				hunpack(prp->pindx, luh);
				if (needhead)
				  {
					printf(
		" Reflections in magnification window:\n");
					printf(
		"   h   k   l Xpre Ypre Phipre   Lpcor   gamma Dspacing\n");
					needhead = 0;
				  }
				dspac = 0.5 / quickstl(lattr, (int)(prp->pxt),
					(int)(prp->pyt), (int)(prp->pphi));
				printf(
			 "%4d%4d%4d %4.0f %4.0f %6.1f %7.1f %7.1f %8.3lf\n",
				 luh[0], luh[1], luh[2], prp->pxt, prp->pyt,
				 prp->pphi, prp->pcor, prp->pgama, dspac);
			  }
		  }
	   }
	if (thiscolor != olayp->gp_col)
		XSetForeground(olayp->gp_di, olayp->gp_ogc, olayp->gp_col);
	return npredi;
}

static int superpose_troids(XG_X11PROPS *olayp, int ismag)
{ /*	This sets up a determination of the centroids from a CENTROIDS file
	peaking near the current frame and enables a display of them */
	int		lfmno, ntri;
	TROID		*trin, *trl;
	DISP_ATTRS	*lattr, *maina;

	maina = olayp->gp_at;
	if (!((maina->di_disflags) & DISPCEN_NEED)) return 0;
	if (NULL == fixbutton(olayp, FIRSTOVERL + 3)) return 0;
	lattr = ismag ? olayp->gp_bl : olayp->gp_at;
	lfmno = maina->di_fmno;
	if (NULL == (trin = gettroids(maina, lfmno, &ntri)))
	  {
		fprintf(stderr, "Failed to find nearby centroids\n");
		return 0;
	  }
	trl = trin + (ntri - 1);
	if (!lattr->di_wintype)
	  {
		printf("Overlaying %6d centroids from [%6.1f,%6.1f,%6.1f]",
			ntri, trin->xto, trin->yto, trin->frame0);
		printf(" to [%6.1f,%6.1f,%6.1f]\n",
			trl->xto, trl->yto, trl->frame0);
	  }
	show_centroids(olayp, trin, ntri, ismag);
	return ntri;
}


void do_superpositions(XG_X11PROPS *xxg, int ismag)
{ /*	this computes and displays the overlays onto the data */
	if (ismag) attr2locals();
	(void)superpose_predict(xxg, ismag);
	(void)superpose_border(xxg, ismag);
	(void)superpose_troids(xxg, ismag);
	(void)superpose_calib(xxg, ismag);
	(void)super_calcen(xxg, ismag);
	(void)super_resolcircles(xxg, ismag);
	(void)super_rowcolsums(xxg, ismag);
}

/* end of dispover.c */
