/* dispxutil.c -- utilities for X-windows displays of XGEN results

	Copyright (c) 2003, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions; reorganize use of "{"
	23 Jun 2006: image memory moved from the XG_X11PROPS structure to
		the DISP_ATTRS structure so things are better localized
	12 Jun 2006: show_actual,show_region,showrefmr moved to dispxmag;
		gdisnam,varlabel moved here
	23 Jan 2006: rename uref->showuref, mref->showmref;
	 eliminate pointless color definitions; store_image loses args;
	 recreate gp_frsca only if needed; toggleon rewritten and moved out
	 6 Jan 2006: function prototypes moved to xgmdefs.h
	 6 Oct 2005: housekeeping
	10 Sep 2005: make_disimage() gets background from di_regb
	22 Jun 2005: dcircle() omitted: nothing calls it anyway
	20 Oct 2004: setupfrmscale always calls make_ftab; toggleon() added
	13 Oct 2004: show_table(): color-labels more carefully named
	 9 Oct 2004: get_colorscheme() explicitly recalcs multiref and urefls
	 output if the color scheme changes; valtocolors() is more careful
	 about out-of-bounds values; reorganized;
	 make_texports() only puts out 4, not 5, textports
	 3 Oct 2004: all color specs use XAllocColor with explicit one-byte
	 hex specifications for 32 colors; DEFAULT_MULT=1.2; colorstripes
	 reshaped so we can fit 32 of them onto the screen 
	28 Aug 2004: routines used only here are declared as static;
	 store_image moved down in file to reduce declarations; cosmetics
	18 Aug 2004: convert some static globals to local variables;
	 htick(),vtick() takes dvmin,dvdel and ivmin,ivdel as args
	25 Feb 2004: progcolors() deleted because we didn't call it anywhere
	21 Jan 2004: license message inserted
	 6 Jan 2004: globals replaced with elements of various structures;
		update calls to make_ftab appropriately
	13 Dec 2003: eliminate unused variables retv from dextract_first()
		and lhd from show_table()
	19 Sep 2003: fna0 explicitly initialized to NULL in get_progcolors();
		one-byte and two-byte pointers nulled in make_disimage()
		in the nbperp==4 case, etc.
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	15 Sep 2003: reorganize to take advantage of additional
		elements in XG_X11PROPS structures
	14 Sep 2003: move some routines into dispxmag.c and dispover.c
	 4 Sep 2003: enable reading of compressed X-GEN output files
	29 Aug 2003: created from relatively undependent pieces of dispx.c
 Relevant changes made to dispx.c:
	16 Aug 2003: major workover to reduce reliance on global variables,
		at the cost of some extra elements in the limits[] array
		and a few extra arguments to some functions
	12 Aug 2003: add new overlays for resolution circles &
		row+column sums; change displays of predictions to
		reflections with large gamma values cover more images
	11 Aug 2003: subtract local sliding-averaged background from
		displayed intensity values
	 6 Aug 2003: change Motif global variables to static globals
	 1 May 2003: add additional data-processing functions;
		improve handling of data in blowups
	25 Mar 2003: major rewrite to enable input from stdin
		and to allow eight different image orientations
	14 Mar 2003: corrected bug in one call to strstufstr()
	 5 Jan 2003: reorganized source-file header
	25 Aug 2002: (?) redo color schemes: add black-and-white, inverses.
		add routine get_colorscheme() to accomplish this
	25 Mar 2002: removed MSI attachments
	 5 May 1994: source-file header, POLYDET->XGEN
	 7 Dec 1993: initial X-GEN version

   Purpose: utilities for displays of detector images and other objects
	directly on an X window display governed under Motif.
	Unlike ../displays/dispcalc.c, these routines ARE Motif-specific,
		but they do not depend very much on X-GEN itself.
   Cross references: ./disp*.c, ../displays/disp*.c
*************************************************************************/
#define		NEED_STAT	1
#define		NEED_ERRNO	1
#include	"xgmdefs.h"
#define		TICK_WIDTH	10
#define		GAP		5
#define		DGAP		(TICK_WIDTH + GAP)
#define		DEFAULT_MULT	1.2	/*default multiplier, 1 color to next*/

extern void abort(void);

char *dextract_first(XmString cs)
{ /*	Get normal string from XmString */
	XmStringContext context;
	XmStringCharSet charset;
	XmStringDirection direction;
	Boolean separator;
	char *primitive_string;

	if (True != XmStringInitContext(&context, cs)) return NULL;
	XmStringGetNextSegment(context, &primitive_string,
	   &charset,&direction,&separator);
	XmStringFreeContext(context);
	return(primitive_string);
}

XmString strstufset(char *str, char *chaset)
{ /* Convenience routine */
	return XmStringCreateLtoR(str, chaset);
}

XmString strstufstr(char *str, char *chaset_str)
{ /* wrapper */
	return XmStringCreateLtoR(str, (XmStringCharSet)chaset_str);
}

void replace_string(Widget wid, char *newstr, XmStringCharSet chaset)
{ /*	This sticks the string `newstr' into an XmNlabelString associated
	with the widget `wid' and the charset `chaset'.
	It manages memory for the string using XmStringFree.*/
	static Arg	rar[2];
	static XmString	locstr;

	locstr = strstufset(newstr, chaset);
	XtSetArg(rar[0], XmNlabelString, locstr);
	XtSetValues(wid, rar, 1);
	XtManageChild(wid);	XmStringFree(locstr);
}

void getbitmapv(XG_X11PROPS *xxg)
{ /*	This establishes the bitmap pad and the number of bytes per pixel
	in an image. It does so by creating a meaningless image, reading out
	the appropriate structure elements, and then destroying the image. */
	int		bitmappad, bperline, imformat, imoffset, screen;
	unsigned	depth;
	XImage		*timage;
	Visual		*dispv;
	DISP_ATTRS	*lattr;
	static char	*nulldat;

	bitmappad = 8;	bperline = imoffset = 0;	imformat = ZPixmap;
	screen = DefaultScreen(xxg->gp_di);
	dispv = DefaultVisual(xxg->gp_di, screen);
	depth = DefaultDepth(xxg->gp_di, screen);
	bitmappad = ((_XPrivDisplay)(xxg->gp_di))->bitmap_pad;
	lattr = xxg->gp_at;
	if (NULL == (nulldat = (char *)mallop(20,
	 "nulldat in getbitmapv")))
	  {
		lattr->di_nbperp = bitmappad / 8; return;
	  }
	zeromem((AVOID *)nulldat, 20);
	timage = XCreateImage(xxg->gp_di, dispv, depth, imformat, imoffset,
	 nulldat, 2, 2, bitmappad, bperline);
	if (NULL != timage)
	  {
		bitmappad = timage->bitmap_pad;
		lattr->di_nbperp = timage->bits_per_pixel / 8;
		(void)XDestroyImage(timage);
	  }
	else {
		(void)free((AVOID *)nulldat);
		lattr->di_nbperp = bitmappad /8;	/* the best we can do */
	     }
	xxg->gp_disv = dispv;	xxg->gp_dep = depth; xxg->gp_bmp = bitmappad;
	xxg->gp_bpl = bperline;	xxg->gp_imo = imoffset;	xxg->gp_imf = imformat;
}

void get_colorscheme(XG_X11PROPS *xxg)
{ /*	This determines what the user has set the color scheme to and
	arranges the `xxg->gp_coi[i]' array accordingly */
	int		i, j, lditype, licn, ndcol, nra, previcn, pushoff;
	Widget		*wp;
	Boolean		setnow;
	Arg		ldag[1];
	DISP_ATTRS	*lattr;

	previcn = licn = xxg->gp_icn; /* remember previous color scheme */
	lattr = xxg->gp_at;	nra = lattr->di_nranges;
	ndcol = xxg->gp_ndcolors;
	for (licn = 0, wp= &(xxg->gp_color[0]); licn < NSCHEMES; licn++, wp++)
	   {
		XtSetArg(ldag[0], XmNset, &setnow);
		XtGetValues(*wp, ldag, 1);
		if (setnow) break; /* read current color scheme */
	   }
	if ((licn < 0) || (licn >= NSCHEMES)) licn = 0; /* disallow nonsense */
	xxg->gp_icn = licn;
	switch (licn)
	  {
		case	0:	default:	/* black -> white */
			for (i = 0, j = ndcol; i < nra; i++, j++)
				xxg->gp_coi[i] = j;
			break;
		case	1:	/* spectral */
			for (i = 0; i < nra; i++) xxg->gp_coi[i] = i;
			break;
		case	2:	/* reverse (white -> black) */
			for (i = 0; i < nra; i++)
				xxg->gp_coi[i] = ndcol + nra - 1 - i;
			break;
		case	3:	/* spectral, reversed (white -> red) */
			for (i = 0, j = nra; i < nra; i++, j--)
				xxg->gp_coi[i] = j;
			break;
	  }
	if (previcn != licn) /* redo the calculations if the scheme changed */
	  {
		if ((lditype = lattr->di_distype)) /* not a detector image */
		  {	
			if (lditype < NUREFLBS) /* UREFLS calc of some sort */
			  {
				pushoff = lditype + NFRAMEBS;
				lattr->di_distype = 0; /* force it to recalc */
				showuref(xxg->gp_dpush[pushoff]);
			  }
			else if (lditype < NUREFLBS + NMREFBS - 1) /*multiref*/
			  {
				pushoff = lditype + NFRAMEBS + 1;
				lattr->di_distype = 0; /* force it to recalc */
				showmref(xxg->gp_dpush[pushoff]);
			  }
		  }
		chgdisp(); /* display new color scheme */
	  }
}

void valtocolors(XG_X11PROPS *xxg)
{ /*	This converts values to colors. */
	int		dispac, nbad, tcoi, y;
	char		*cp;
	short		*osp;
	xglong		*olp;
	DISP_ATTRS	*lattr;
	PAU		vcu;
	static char	*imtbuf = (char *)0;

	get_colorscheme(xxg);	/* understand color scheme */
	lattr = xxg->gp_at;	nbad = 0;
	dispac = ((int)lattr->di_wid) * ((int)lattr->di_ht);
	if (lattr->di_nbperp == 1)
	  {
		for (y = dispac, cp = lattr->di_rotm; y--; cp++)
		   {
			tcoi = xxg->gp_coi[(int)(*cp)];
			if ((tcoi < 0) || (tcoi > NALLOCCOL))
			  {
				nbad++;	tcoi = xxg->gp_coi[0];
			  }
			*cp = (char)(xxg->gp_dbcol[tcoi]);
		   }
	  }
	else
	  {
		if (NULL == imtbuf)
		  {
			if (NULL == (imtbuf = (char *)mallop(dispac,
			 "imtbuf in valtocolors")))
			  {
				fprintf(stderr,
	 "Error allocating temporary memory for image buffer\n");
				return;
			  }
		  }
		cpymem((void *)imtbuf, (void *)(lattr->di_rotm),
			(unsigned)dispac);
		if (lattr->di_nbperp == 2)
		  {
			vcu.ucpt = (u_char *)(lattr->di_rotm);
			for (y = dispac , osp = (short *)(vcu.uspt),
			 cp = imtbuf; y--; osp++,cp++)
			   {	
				tcoi = xxg->gp_coi[(int)(*cp)];
				if ((tcoi < 0) || (tcoi > NALLOCCOL))
				  {
					nbad++;	tcoi = xxg->gp_coi[0];
				  }
				*osp = (short)(xxg->gp_dbcol[tcoi]);
			   }
		  }
		else {
			vcu.ucpt = (u_char *)(lattr->di_rotm);
			for (y = dispac, olp = (xglong *)(vcu.vpt),
			 cp = imtbuf; y--; olp++, cp++)
			   {
				tcoi = xxg->gp_coi[(int)(*cp)];
				if ((tcoi < 0) || (tcoi > NALLOCCOL))
				  {
					nbad++;	tcoi = xxg->gp_coi[0];
				  }
				*olp = (long)(xxg->gp_dbcol[tcoi]);
			   }
		  }
	  }
	if (nbad) printf(" %8d pixels had nonsense stored values\n", nbad);
}

int diamond(XG_X11PROPS *olayp, int ismag)
{ /*	This creates a hollow diamond of half-width `wo2' centered at (yc,xc).
	For now it's just 1 pixel wide; we may widen it later.*/
	int		ix, ixma, iy, iyma, wo2, xma, xmi, yma, ymi;
	DISP_ATTRS	*lattr, *maina;
	Display		*tdispla;
	Window		windo;
	GC		olaygc;

	maina = olayp->gp_at;
	if (ismag)
	  {
		lattr = olayp->gp_bl;	windo = olayp->gp_maw;
	  }
	else {
		lattr = olayp->gp_at;	windo = olayp->gp_wi;
	     }
	wo2 = lattr->di_boxh;
	if (wo2 < 1)
	  {
		printf("Cannot draw diamond: width is non-positive\n");
		return 0;
	  }
	ix = maina->di_currx * lattr->di_wiup / lattr->di_widown;
	iy = maina->di_curry * lattr->di_htup / lattr->di_htdown;
	if (lattr->di_wintype)
	  {
		ix = 80 - 8 * maina->di_disx + ix * 8;
		iy = 80 - 8 * maina->di_disy + iy * 8;
		ixma = lattr->di_magw - wo2;	iyma = lattr->di_magh - wo2;
	  }
	else {
		ixma = lattr->di_wid - wo2;	iyma = lattr->di_ht - wo2;
	     }
	if ((ix < wo2) || (ix >= ixma) || (iy < wo2) || (iy >= iyma))
		return 0;
	ymi = iy - wo2;	yma = iy + wo2;
	xmi = ix - wo2;	xma = ix + wo2;
	tdispla = olayp->gp_di;	olaygc = olayp->gp_ogc;
	XDrawLine(tdispla, windo, olaygc, xmi,  iy,  ix, yma);
	XDrawLine(tdispla, windo, olaygc,  ix, yma, xma,  iy);
	XDrawLine(tdispla, windo, olaygc, xma,  iy,  ix, ymi);
	XDrawLine(tdispla, windo, olaygc,  ix, ymi, xmi,  iy);
	return 1;
}

void start_dcolors(XG_X11PROPS *xxg, Colormap *tma)
{ /*	This sets up the colors used in the images, plus a few others.  */
	int		grayr, gscale, i, nranges, status, totcolors;
	XColor		*scrp;
	Pixel		*bgp;
	Arg		ldag[1];
	static XColor	*xcreen;
	static int	n = 1;
	static u_char	exactred[] =
	 {	0x00, 0x08, 0x10, 0x19, 0x12, 0x0a, 0x05, 0x00,	/*  0- 7 */
		0x10, 0x20, 0x30, 0x40, 0x50, 0x3c, 0x28, 0x14,	/*  8-15 */
		0x00, 0x33, 0x66, 0x99, 0xcc, 0xff, 0xff, 0xff,	/* 16-23 */
		0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,	/* 24-31 */
		0xff, 0x31, 0xff, 0x6b, 0x84, 0xef };		/* 32-36 */
	static u_char	exactgreen[] =
	 {	0x00, 0x08, 0x10, 0x19, 0x12, 0x0a, 0x05, 0x00,	/*  0- 7 */
		0x20, 0x40, 0x5f, 0x7e, 0x9d, 0xb6, 0xcf, 0xe8,	/*  8-15 */
		0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xe6, 0xcd,	/* 16-23 */
		0xb3, 0x9a, 0x80, 0x66, 0x4d, 0x33, 0x1a, 0x00,	/* 24-31 */
		0xff, 0xcb, 0xbe, 0x69, 0xcf, 0xe7 };		/* 32-36 */
	static u_char	exactblue[] =
	 {	0x00, 0x25, 0x4a, 0x69, 0x94, 0xb8, 0xdc, 0xff,	/*  0- 7 */
		0xeb, 0xd7, 0xc4, 0xb0, 0x9d, 0x75, 0x4e, 0x27,	/*  8-15 */
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 16-23 */
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 24-31 */
		0xff, 0x31, 0xce, 0x6b, 0xe7, 0x8c } ;		/* 32-36 */

	nranges = (xxg->gp_at)->di_nranges;
	totcolors = 2 * nranges + NEXTRAS;
	XtSetArg(ldag[0], XmNcolormap, tma);	n = 1;
	XtGetValues(xxg->gp_dimain, ldag, n);
	if (NULL == xcreen)
	  {
		if (NULL == (xcreen = (XColor *)callop(totcolors,
		 sizeof (XColor), "xcreen in start_dcolors"))) return;
	  }
	/* first, the colorful colors, including the signaling ones */
	for (i = 0, scrp = &(xcreen[0]), bgp = xxg->gp_dbcol;
	  i < nranges + NEXTRAS; i++, scrp++, bgp++)
	   {
		scrp->red = (exactred[i] << 8) | (exactred[i]);
		scrp->green = (exactgreen[i] << 8) | (exactgreen[i]);
		scrp->blue = (exactblue[i] << 8) | (exactblue[i]);
		scrp->flags = scrp->pad = 0;
		status = XAllocColor(xxg->gp_di, *tma, scrp);
		*bgp = scrp->pixel;
	   }
	/* then, the grayscales */
	for ( ; i < totcolors; i++, scrp++, bgp++)
	   {
		grayr = (i - (nranges + NEXTRAS)) * 8;
		gscale = (grayr << 8) | grayr;
		scrp->red = scrp->green = scrp->blue = gscale;
		scrp->flags = scrp->pad = 0;
		status = XAllocColor(xxg->gp_di, *tma, scrp);
		*bgp = scrp->pixel;
	   }
}

static void htick(int zy, int zx, int ntor, int ismag,
	double dvmin, double dvdel, XG_X11PROPS *olayp)
{ /*	This puts a horizontal tick-mark at y = zy.
	It then puts a label up next to that tick. */
	int		diw, dih, lx, nchars, xmi, xma, yu;
	char		tstr[40];
	static char	tfmt[] = "%7.2f";
	double		val;
	DISP_ATTRS	*lattr;
	Window		windo;

	if (ismag)
	  {
		lattr = olayp->gp_bl;	windo = olayp->gp_maw;
	  }
	else {
		lattr = olayp->gp_at;	windo = olayp->gp_wi;
	     }
	if (lattr->di_wintype)
	  {	
		diw = lattr->di_magw;	dih = lattr->di_magh;
	  }
	else {
		diw = lattr->di_wid;	dih = lattr->di_ht;
	     }
	xmi = zx - TICK_WIDTH;	if (xmi < 0) xmi = 0;
	xma = zx + TICK_WIDTH;	if (xma > diw) xma = diw;
	yu = dih - 1 - zy;
	XDrawLine(olayp->gp_di, windo, olayp->gp_ogc, xmi, yu, xma, yu);
	lx = (zx < diw / 2 ? zx + DGAP : zx - DGAP);
	val = dvmin + zy * dvdel;
	tfmt[3] = '0' + ntor;
	sprintf(tstr, tfmt, val);	nchars = strlen(tstr);
	XDrawString(olayp->gp_di, windo, olayp->gp_ogc, lx,yu+3, tstr, nchars);
}

static void vtick(int zx, int zy, int ntor, int ismag,
	double ivmin, double ivdel, XG_X11PROPS *olayp)
{ /*	This puts a vertical tick-mark at x = zx.
	It then puts a label up just below or above that tick. */
	int		dih, diw, lx, ly, nchars, yma, ymi, yu;
	char		tstr[40];
	double		val;
	Window		windo;
	DISP_ATTRS	*lattr;
	static char	tfmt[] = "%7.2f";

	lattr = ismag ? olayp->gp_bl : olayp->gp_at;
	if (lattr->di_wintype)
	  {
		diw = lattr->di_magw;	dih = lattr->di_magh;
		windo = olayp->gp_maw;
	  }
	else {
		diw = lattr->di_wid;	dih = lattr->di_ht;
		windo = olayp->gp_wi;
	     }
	yu = zy;
	ymi = yu - TICK_WIDTH;	if (ymi < 0) ymi = 0;
	yma = yu + TICK_WIDTH;	if (yma >= dih) yma = dih - 1;
	XDrawLine(olayp->gp_di, windo, olayp->gp_ogc, zx, ymi, zx,yma);
	ly = (yu < dih / 2 ? yu + DGAP : yu - DGAP);
	val = ivmin + zx * ivdel;
	tfmt[3] = '0' + ntor;
	sprintf(tstr, tfmt, val);	nchars = strlen(tstr);
	lx = zx - 3 * nchars;
	if (lx < 0) return;
	if (lx > diw - 5 * nchars) lx = diw - 5 * nchars;
	XDrawString(olayp->gp_di, windo, olayp->gp_ogc, lx, ly, tstr, nchars);
}

void draw_axes(XG_X11PROPS *olayp, int ismag)
{ /*	This overlays a set of axes with tick marks on 'em over a scatterplot.
	It also puts some labels up */
	int		dih, diw, interv, nra, ntori, xe, xu, xz, ye, yu, yz;
	double		extrem, ivdel, ivmax, ivmin, oth;
	float		*flp;
	double		dvdel, dvmax, dvmin;
	Window		windo;
	DISP_ATTRS	*lattr, *maina;

	lattr = ismag ? olayp->gp_bl : olayp->gp_at;	maina = olayp->gp_at;
	if (lattr->di_wintype)
	  {
		diw = lattr->di_magw;	dih = lattr->di_magh;
		windo = olayp->gp_maw;
	  }
	else {
		diw = lattr->di_wid;	dih = lattr->di_ht;
		windo = olayp->gp_wi;
	     }
	nra = lattr->di_nranges;
	xe = diw - 1;	ye = dih - 1;
	XSetForeground(olayp->gp_di, olayp->gp_ogc, olayp->gp_col);
	flp = maina->di_limi + (nra + 2);
	ivmin = *flp++;	ivmax = *flp++;	ivdel = *flp++;
	dvmin = *flp++;	dvmax = *flp++;	dvdel = *flp++;
	if (ivmin <= 0. && ivmax >= 0.) /* draw y axis if it's on screen */
		xz = (-ivmin * (double)(diw - 1)) / (ivmax - ivmin);
	else	xz = 0;
	XDrawLine(olayp->gp_di, windo, olayp->gp_ogc, xz, 1, xz, ye);
	interv = nudge((dvmax - dvmin) / 10.) / dvdel;
	extrem = fabs(dvmax);	if ((oth = fabs(dvmin)) > extrem) extrem = oth;
	ntori = (extrem < 0.01 ? 5 : (extrem < 0.1 ? 4 :
		(extrem < 1 ? 3 : (extrem < 10 ? 2 : (extrem < 100 ? 1 :0)))));
	for (yu = 0; yu < dih; yu += interv)
		htick(yu, xz, ntori, ismag, dvmin, dvdel, olayp);
	if (dvmin <= 0. && dvmax >= 0.) /* draw x axis if it's on screen */
		yz = dih - 1 - (-dvmin * (double)(dih-1)) / (dvmax - dvmin);
	else	yz = dih - 1;
	XDrawLine(olayp->gp_di, windo, olayp->gp_ogc, 1, yz, xe, yz);
	interv = nudge((ivmax - ivmin) / 10.) / ivdel;
	extrem = fabs(ivmax);	if ((oth = fabs(ivmin)) > extrem) extrem = oth;
	ntori = (extrem < 0.01 ? 5 : (extrem < 0.1 ? 4 :
		(extrem < 1 ? 3 : (extrem < 10 ? 2 : (extrem < 100 ? 1 :0)))));
	for (xu = 0; xu < diw; xu += interv)
		vtick(xu, yz, ntori, ismag, ivmin, ivdel, olayp);
}

void hkl_axes(XG_X11PROPS *olayp, int ismag)
{ /*	This displays axes & labels on a urefls hkl plot */
	char		nc, oc, tstr[40];
	int		dih, diw, nchars, nv, uhco, uhty, uhva, wend, xz, yz;
	DISP_ATTRS	*lattr;
	Window		windo;
	static char	hl[] = "hkl";

	lattr = ismag ? olayp->gp_bl : olayp->gp_at;
	if (lattr->di_wintype)
	  {
		diw = lattr->di_magw;	dih = lattr->di_magh;
		windo = olayp->gp_maw;
	  }
	else {
		diw = lattr->di_wid;	dih = lattr->di_ht;
		windo = olayp->gp_wi;
	     }
	uhco = olayp->gp_uhco;	uhva = olayp->gp_uhva;	uhty = olayp->gp_uhty;
	XSetForeground(olayp->gp_di, olayp->gp_ogc, olayp->gp_col);
	nv = (uhco + 1) % 3;	nc = hl[nv];	oc = hl[3 - uhco - nv];
	wend = diw - 1;	yz = dih / 2 + 1;
	if (!(lattr->di_wintype))
		XDrawLine(olayp->gp_di, windo, olayp->gp_ogc, 0, yz, wend, yz);
	tstr[0] = nc;	tstr[1] = '\0';
	XDrawString(olayp->gp_di, windo, olayp->gp_ogc, wend -9, yz+12,tstr, 1);
	wend = dih - 1;	xz = diw / 2 + 1;
	if (!(lattr->di_wintype))
		XDrawLine(olayp->gp_di, windo, olayp->gp_ogc, xz, 0, xz, wend);
	tstr[0] = oc;
	XDrawString(olayp->gp_di, windo, olayp->gp_ogc, xz + 12, 10, tstr, 1);
	sprintf(tstr, "%s @ %c =%4d",
	 (uhty & 04 ? "Int/sig" : "Intensities"), hl[uhco], uhva);
	nchars = strlen(tstr);
	XDrawString(olayp->gp_di, windo, olayp->gp_ogc, 10,dih-15,tstr,nchars);
}

static int make_disimage(void)
{ /*	This creates the real display image. We display it differently
	depending on the number of bytes per displayed pixel */
	int		dih, diw, htdo, htup, irange, ival, jx, jxe, jy, jye;
	int		lnrx, lnry, locbg, loval, maxinr, mx, my, nbperp;
	int		newsz, nranges, tcoi, tval, ujx, ujy, wido, wiup, x, y;
	static int	nincol[40];
	char		*cp, *rawdat, *rotdat;
	short		*sp;
	xglong		*lp, *locof;
	DISP_ATTRS	*dispatt;
	XG_X11PROPS	*xxg;
	MAIL		*lhd;
	PAU		imu;

	if (NULL == (xxg = managex11p(3))) return EALLOC;
	dispatt = xxg->gp_at;	locof = dispatt->di_regb;
	nbperp = dispatt->di_nbperp;
	diw = dispatt->di_wid;	dih = dispatt->di_ht;
	wiup = dispatt->di_wiup;	wido = dispatt->di_widown;
	htup = dispatt->di_htup;	htdo = dispatt->di_htdown;
	lhd = dispatt->di_head + dispatt->di_fmno % dispatt->di_nmbox;
	nranges = dispatt->di_nranges;
	lnrx = xxg->gp_nregx;		lnry = xxg->gp_nregy;
	newsz = diw * dih * nbperp;
	if (xxg->gp_imdasize < 2 * newsz) xxg = managex11p(2 * newsz);
	rawdat = dispatt->di_imm;
	rotdat = dispatt->di_rotm = dispatt->di_imm + newsz;
	if (nbperp == 1)
	  {
		cp = rawdat;	sp = (short *)0;	lp = (xglong *)0;
	  }
	else if (nbperp == 2)
	  {
		imu.ucpt = (u_char *)rawdat;	sp = (short *)(imu.uspt);
		cp = (char *)0;			lp = (xglong *)0;
	  }
	else {
		imu.ucpt = (u_char *)rawdat;	lp = (xglong *)(imu.vpt);
		sp = (short *)0;		cp = (char *)0;
	     }
	for (x = 0; x < 40 ; x++) nincol[x] = 0;
	for (y = 0; y < dih; y++)
	   {
		my = y / htup;	ujy = my * htdo;
		jye = (my + 1) * htdo;
		for (x = 0; x < diw; x++)
		   {
			mx = x / wiup;	ujx = mx * wido;
			jxe = (mx + 1) * wido;
			locbg = locba(locof, ujy + htdo / 2,
				ujx + wido / 2, lnry, lnrx);
			maxinr = 0;
			for (jy = ujy, tval = 0.; jy < jye; jy++)
			   for (jx = ujx; jx < jxe; jx++)
			      {
				loval = imvalue(lhd, jx, jy) - locbg / 0.85;
				if (loval > maxinr) maxinr = loval;
				if (loval < 0) loval = 0;
				tval += loval;
			      }
			/* rescale tval by the area
			(in detector pixels squared) of a display pixel */
			tval /= (htdo * wido);
			/* innovation: if a single pixel in the array has
			 a higher background-subtracted intensity than the mean,
			 save it instead */
			/* if (tval < maxinr) tval = maxinr; */
			if (tval < 1) irange = 0;
			else {
				for (irange = 0; irange < nranges -1; irange++)
				   if (tval < (dispatt->di_range[irange]))
					break;
			     }
			(nincol[irange])++;
			tcoi = xxg->gp_coi[irange];
			ival = xxg->gp_dbcol[tcoi];
			if (nbperp == 1) *cp++ = ival;
			else if (nbperp == 2) *sp++ = ival;
			else if (nbperp == 4) *lp++ = ival;
		   }
	   }
	/* rotate the image according to user's preference */
	dorotate(dispatt, rawdat, rotdat);
	return EOK;
}

int store_image(void)
{ /*	Convert the actual detector image array into a color-range array.
	Most of the complexity in this comes from the fact that we may
	be either magnifying or demagnifying the image. */

	int		jy, x;
	DISP_ATTRS	*lattr, *mattr;
	MAIL		*lhd;
	static int	nregv[2];
	static ino_t	curino = 0;
	XG_X11PROPS	*xxg;

	if (NULL == (xxg = managex11p(3))) return EALLOC;
	lattr = xxg->gp_at;	if (0 >= lattr->di_nbperp) getbitmapv(xxg);
	get_colorscheme(xxg);	/* understand color schemes */
	jy = ((int)lattr->di_wid) * ((int)lattr->di_ht) *
		((int)lattr->di_nbperp);
	if (jy < 16384) jy = 16384;
	if (NULL == lattr->di_imm)
	  {
		xxg = managex11p(2 * jy);	lattr = xxg->gp_at;
	  }
	lattr->di_rotm = lattr->di_imm + jy;
	/* we determine local averages of the intensities. We only need
	 to re-do this if we have a new image */
	lhd = lattr->di_head + lattr->di_fmno % lattr->di_nmbox;
	if (curino != lhd->quadrant)
	  {
		if (EOK != glocave(lattr, nregv))
		  {
			fprintf(stderr, "Unable to obtain local averages\n");
			return EALLOC;
		  }
		curino = lhd->quadrant;
	  }
	lattr->di_distype = 0; /* flag: display frame data, not delI/sigma */
	for (x = jy = 0; jy < lattr->di_nranges; jy++)
		if (xxg->gp_dbcol[jy] > 0) x++;
	if (x < 2) return EOK;	/* color table must be defined by this point */
	mattr = xxg->gp_bl;
	mattr->di_nbperp = lattr->di_nbperp;
	xxg->gp_nregx = nregv[0];	xxg->gp_nregy = nregv[1];
	/* now create the actual display image and rotate it appropriately */
	return make_disimage();
}

static char *tablestr(int irange, XG_X11PROPS *xxg)
{ /*	This returns the string that belongs in one of the color stripes */
	int		nranges;
	double		high;
	DISP_ATTRS	*dispatt;
	MAIL		*lhd;
	static char	cmin[40];
	static xglong	rmaxv;
	static char	cfmt[] = "%6.0f";

	dispatt = xxg->gp_at;
	lhd = dispatt->di_head + dispatt->di_fmno % dispatt->di_nmbox;
	nranges = dispatt->di_nranges;	zeromem((AVOID *)&(cmin[0]), 20);
	if (dispatt->di_distype) /* urefls or multiref statistics */
	  {
		if (!irange) strcpy(cmin, "no pts");
		else {
			high = dispatt->di_limi[irange];
			if ((high < -99.9) || (high > 999.)) cfmt[3] = '0';
			else if ((high < -9.99) ||(high > 99.9)) cfmt[3] = '1';
			else cfmt[3] = '2';
			sprintf(cmin, cfmt, high);
		     }
	  }
	else {
		if ((NULL != lhd) && (0 < (lhd->zmax))) rmaxv = lhd->zmax;
		else	rmaxv = 65535;
		sprintf(cmin, "%6d", (irange == nranges - 1) ? (int)rmaxv :
		 (int)(dispatt->di_range[irange]) - 1);
	     }
	cmin[6] = cmin[19] = '\0'; return cmin; /* force null-termination */
}

char *gdisnam(int dispno)
{ /*	This returns the appropriate element of the `dispnam' array */
	static char		*dispnam[] = {
	 "Image data", "Intensity", "Inten/sig", "Error(omega)",
	 "Error(X)", "Error(Y)", "Intensity", "Inten/sig",
	 "Intensity", "Inten/sig", "(I-<I>)/sig" } ;

	return dispnam[dispno];
}

char *varlabel(int varno, int isdepend)
{ /*	This returns either istr[varno] or dstr[varno],
	depdending on isdepend */
	static char	*dstr[] = {
	 "Err(Omega)", "Err(X)", "Err(Y)", "Inten", "I/sig" } ;
	static char	*istr[] = { "sin(theta)/lam", "Omega", "X", "Y" } ;

	return (isdepend) ? dstr[varno] : istr[varno];
}

static char *file_info(XG_X11PROPS *xxprop)
{ /*	This provides a summary of what's in a file. It returns a lengthy
	string that will get stuck in one particular place. */

	char		u0, u1, u2;
	int		ditype, hundredths, ldvn, livn;
	int		n65k, nobi, nra, seconds, uhco;
	double		revent;
	DISP_ATTRS	*dispatt;
	CRYSTAL		*xhp;
	MAIL		*lhd;
	static char	*localname, *scshort, *shorter, fbuf[256];
	static char	fb0[] = "Multiref: ";
	static char	fb1[] = " observations of ";
	static char	fb2[] = " Scale input: ";
	static char	fb3[] = " scaling groups";
	static char	fb4[] = " Intensity min, max";
	static char	fb5[] = " Intensity/sigma min, max";
	static char	uhkl[] = "hkl";

	dispatt = xxprop->gp_at;	nobi = xxprop->gp_nobinu;
	ditype = dispatt->di_distype;	nra = dispatt->di_nranges;
	xhp = dispatt->di_xhdp;
	switch (ditype)
	      {	/* Multiref Intensity and I/sigma */
		case	NUREFLBS:	case	NUREFLBS + 1:
			localname = trname("MULTIREF");
			shorter = filepart(localname, 0);
			sprintf(fbuf,
			 "%s%s\n %6d%s\n%6d reflections\n%s =\n%9.3f, %9.3f\n",
			 fb0, shorter, (int)(xhp->xt_ninteg), fb1,
			 (int)xhp->xt_nmerged, (ditype == NUREFLBS ? fb4: fb5),
			 dispatt->di_limi[nra], dispatt->di_limi[nra+1]);
			break;
		case	NUREFLBS + 2:	/* multiref I/sigma */
			localname = trname("MULTIREF");
			shorter = filepart(localname, 0);
			strcpy(fbuf, fb0);	strcat(fbuf, shorter);
			scshort = filepart(trname("SCALEIN"), 0);
			sprintf(fbuf + strlen(fbuf),
			 "\n %6d%s\n%6d reflections\n%s%s\n %6d%s\n",
			 (int)(xhp->xt_ninteg), fb1, (int)(xhp->xt_nmerged),
			 fb2, scshort, (int)(xhp->xt_nshifts), fb3);
			break;
		case	0:	/* data frame */
			localname = dispatt->di_imaname;
			shorter = filepart(localname, 0);
			if (dispatt->di_nmbox != PFRAMES + 1)
			  {
				fprintf(stderr,
				 "nmbox is %d\n", dispatt->di_nmbox);
				abort();
			  }
			lhd = dispatt->di_head +
				dispatt->di_fmno % dispatt->di_nmbox;
			hundredths = lhd->elap100ths % 100;
			seconds = lhd->elap100ths / 100;
			revent = lhd->eventcount;
			n65k = count65k(lhd);
			if (revent < 0.) revent *= -64.;
			sprintf(fbuf,
			 "File: %s\n%8.0f Cts, %5d o'%s\nExposure time:",
			 shorter, revent,(int)(lhd->oflowcount),
			 (lhd->oflowcount == 1 ? "flow" : "flows"));
			if (hundredths < 10) sprintf(&(fbuf[strlen(fbuf)]),
			"%6d.0%1d sec\nMax Ct:%6d at [%4d %4d]",
			 seconds, hundredths, (int)lhd->zmax,
			 (int)lhd->zmaxx, (int)lhd->zmaxy);
			else	sprintf(&(fbuf[strlen(fbuf)]),
				"%6d.%2d sec\nMax Ct:%6d at [%4d %4d]",
				 seconds, hundredths, (int)(lhd->zmax),
				 (int)lhd->zmaxx, (int)lhd->zmaxy);
			if (n65k) sprintf(&fbuf[strlen(fbuf)],
				"\n%5d pixels with 65K counts", n65k);
			else sprintf(&fbuf[strlen(fbuf)],
			 "\nNo pixels with 65K counts");
			break;
		case	(NUREFLBS-1):	case	(NUREFLBS-2): /* urefls hkls */
			localname = trname("UREFLS");
			shorter = filepart(localname, 0);
			uhco = xxprop->gp_uhco; u0 = uhkl[uhco];
			u1 = uhkl[(uhco + 1) % 3]; u2 = uhkl[(uhco + 2) % 3];
			sprintf(fbuf,
 "Urefls %s\n%7d observ.\nPlot of %c vs %c at %c=%3d\nColor-coded on %s\n",
			  shorter, nobi, u1, u2, u0, xxprop->gp_uhva,
			  gdisnam(ditype));	break;
		default :	/* urefls displays */
			localname = trname("UREFLS");
			shorter = filepart(localname, 0);
			livn = xxprop->gp_ivn;	ldvn = xxprop->gp_dvn;
			if (ditype < 0) sprintf(fbuf,
 "Urefls %s\n%6d observations\nScatter plot of %s vs %s\nColor-coded on %s\n",
			 shorter, nobi, varlabel(ldvn, 1),
			 varlabel(livn, 0), gdisnam(2));
			else sprintf(fbuf,
 "Urefls %s\n%6d observations\nDisplay of %s\nMin=%8.3f, Max=%8.3f\n",
			 shorter, nobi, gdisnam(ditype),
			 dispatt->di_limi[nra], dispatt->di_limi[nra+1]);
	     }
	return fbuf;
}

void filereport(XG_X11PROPS *xxg)
{ /*	This reports back information describing a file's contents */
	char	*cp;

	cp = file_info(xxg);
	replace_string(xxg->gp_tport[3], cp, xxg->gp_chpol);
}

void show_table(XG_X11PROPS *xxp)
{ /*	This displays the current color table as a set of labels */

	int		irange, icy, narg, cx, icx, jrange;
	int		nranges, tcoi, tcol, licn;
	Widget		*ctp;
	Pixel		tcb, tcf;
	Arg		ldag[10];
	static char	*cmi;
	static int	cy = 180;
	static int	cw = 44;
	static int	ch = 19;
	static int	*cfc;
	DISP_ATTRS	*dispatt;
	static XmString	locstr;
	static Widget	*ctable;

	dispatt = xxp->gp_at;	licn = xxp->gp_icn;
	cx = dispatt->di_minw + 184;	nranges = dispatt->di_nranges;
	if (NULL == cfc) cfc = (int *)callop((size_t)nranges,
	 sizeof (int), "cfc in show_table");
	if (NULL == cfc) return;
	for (icy = 0; icy < nranges / 2; icy++) cfc[icy] = nranges;
	for ( ; icy < nranges; icy++) cfc[icy] = 0;
	if (NULL == ctable) ctable = (Widget *)callop((size_t)nranges,
	 sizeof (Widget), "ctable in show_table");
	if (NULL == ctable) return;
	for (irange = 0, ctp = ctable; irange < nranges; irange++, ctp++)
	   {
		icx = cx + cw * (irange / (nranges / 4));
		cmi = tablestr(irange, xxp);
		tcol = (licn >= 2) ? cfc[nranges - (1 + irange)] : cfc[irange];
		tcoi = xxp->gp_coi[irange];
		if ((tcoi < 0) || (tcoi >= (2 * nranges + NEXTRAS)))
		  {
			fprintf(stderr,
 "Warning: irange = %2d, colorindex = %d reset to zero\n", irange, tcoi);
			tcoi = 0;
		  }
		tcb = xxp->gp_dbcol[tcoi];	tcf = xxp->gp_dbcol[tcol];
		if (NULL != *ctp)
		  {
			narg = 0;
			XtSetArg(ldag[narg], XmNbackground, tcb);	narg++;
			XtSetArg(ldag[narg], XmNforeground, tcf);	narg++;
			XtSetValues(*ctp, ldag, narg);
			replace_string(*ctp, cmi, xxp->gp_chpol);
		  }
		else {
			narg = 0;
			locstr = strstufstr(cmi, "");
			XtSetArg(ldag[narg], XmNlabelString, locstr); narg++;
			XtSetArg(ldag[narg], XmNx, icx);	narg++;
			jrange = irange % (nranges / 4);
			icy = cy - 20 * jrange;
			XtSetArg(ldag[narg], XmNy, icy);	narg++;
			XtSetArg(ldag[narg], XmNwidth, cw);	narg++;
			XtSetArg(ldag[narg], XmNheight, ch);	narg++;
			XtSetArg(ldag[narg], XmNbackground,tcb); narg++;
			XtSetArg(ldag[narg], XmNforeground,tcf); narg++;
			XtSetArg(ldag[narg], XmNrecomputeSize, False); narg++;
			XtSetArg(ldag[narg], XmNfontList, xxp->gp_dflis);
				narg++;
			if (irange < 10) sprintf(cmi, "Color%1d_label", irange);
			else	sprintf(cmi, "Color%c_label",
					(char)('A' + (irange - 10)));
			*ctp = XmCreateLabel(xxp->gp_dimain, cmi, ldag, narg);
			XtManageChild(*ctp);	XmStringFree(locstr);
		  }
	   }
}

void find_indices(XG_X11PROPS *xxp, int xi, int yi, short *ohkl)
{ /*	This pulls out the hkl indices from a position on the display */
	int		uhco, dih, diw, indxx, indxy;
	DISP_ATTRS	*lattr;

	uhco = xxp->gp_uhco;	lattr = xxp->gp_at;
	ohkl[uhco] = xxp->gp_uhva;
	indxx = (uhco + 1) % 3;	indxy = 3 - uhco - indxx;
	if (lattr->di_wintype)
	  {
		diw = lattr->di_magw;	dih = lattr->di_magh;
	  }
	else {
		diw = lattr->di_wid;	dih = lattr->di_ht;
	     }
	ohkl[indxx] = (xi - (diw / 2)) / 4;
	ohkl[indxy] = -((yi - (dih / 2)) / 4);
}

void init_inten(XG_X11PROPS *xxg)
{ /* Announce the brightest pixel in the image */
	char		itext[40];
	DISP_ATTRS	*lattr;
	MAIL		*lhd;

	lattr = xxg->gp_at;
	lhd = lattr->di_head + lattr->di_fmno % lattr->di_nmbox;
	sprintf(itext, "Frame peak: I[%4d,%4d]=%5d",
		(int)lhd->zmaxx, (int)lhd->zmaxy,
		(int)imvalue(lhd, (int)lhd->zmaxx, (int)lhd->zmaxy));
	replace_string(xxg->gp_tport[2], itext, xxg->gp_chpol);
}

Widget setupfrmscale(XG_X11PROPS *xxg)
{ /*	This displays the frame-index scale widget, creating it if necessary.
	It also (creates and) displays an OK button and a CANCEL button to go
	with it */
	int		fmma, fmmi, lfm, ng, nranges, scwid;
	DISP_ATTRS	*dispatt;
	static XmString	locstr;
	Arg		ldag[12];
	static Widget	lfra = (Widget)0;
	FRAMEINDEX	*lfmi;

	dispatt = xxg->gp_at;	nranges = dispatt->di_nranges;
	lfmi = dispatt->di_fmind = ftables(3);
	if (NULL != lfmi)
	  {
		scwid = lfmi->fi_totfms + 19;	/* width of scale */
		fmmi = lfmi->fi_firstfm;	fmma = lfmi->fi_lastfm;
	  }
	else {
		scwid = 480;	fmmi = 1;	fmma = 1;
	     }
	while (scwid > 800) scwid /= 2;
	if (scwid < 256) scwid = 256;	/* so we can fit the buttons in */
	/* create, if needed  the form the scale will live inside of */
	if (0 == lfra)
	  {
		ng = 0;	XtSetArg(ldag[ng], XmNx, 10);	ng++;
		XtSetArg(ldag[ng], XmNy, 10);		ng++;
		XtSetArg(ldag[ng], XmNwidth, scwid + 40);	ng++;
		XtSetArg(ldag[ng], XmNheight, 120);	ng++;
		XtSetArg(ldag[ng], XmNbackground, xxg->gp_dbcol[nranges + 3]);
		ng++;
		XtSetArg(ldag[ng], XmNforeground, xxg->gp_dbcol[nranges - 1]);
		ng++;
		lfra = XmCreateForm(xxg->gp_dimain, "frmsc_form", ldag, ng);
	  }
		/* create the scale itself */
	if (NULL == xxg->gp_frsca[0])
	  {
		ng = 0; XtSetArg(ldag[ng], XmNminimum, fmmi); ng++;
		XtSetArg(ldag[ng], XmNmaximum, fmma);	ng++;
		XtSetArg(ldag[ng], XmNscaleHeight, 20);		ng++;
		XtSetArg(ldag[ng], XmNscaleWidth, scwid);		ng++;
		XtSetArg(ldag[ng], XmNorientation, XmHORIZONTAL);	ng++;
		XtSetArg(ldag[ng], XmNbackground, xxg->gp_dbcol[0]);	ng++;
		XtSetArg(ldag[ng], XmNforeground, xxg->gp_dbcol[nranges]); ng++;
		XtSetArg(ldag[ng], XmNshowValue, True);		ng++;
		locstr = strstufstr("Frame index value in use", "");
		XtSetArg(ldag[ng], XmNtitleString, locstr);	ng++;
		lfm = dispatt->di_fmno;
		XtSetArg(ldag[ng], XmNvalue, lfm);		ng++;
		XtSetArg(ldag[ng], XmNfontList, xxg->gp_dflis);		ng++;
		xxg->gp_frsca[0] = XmCreateScale(lfra, "frame_scale", ldag, ng);
		XmStringFree(locstr);
	  }
		/* create an "OK" button to go with it */
	if (NULL == xxg->gp_frsca[1])
	  {
		ng = 0;	XtSetArg(ldag[ng], XmNx, 50);	ng++;
		XtSetArg(ldag[ng], XmNy, 80);		ng++;
		locstr = strstufstr("OK", "");
		XtSetArg(ldag[ng], XmNlabelString, locstr);	ng++;
		xxg->gp_frsca[1] = XmCreatePushButtonGadget(lfra,
		 "frame_ok", ldag,ng);
		XtAddCallback(xxg->gp_frsca[1], XmNactivateCallback,
			(XtCallbackProc)chgframe, XpNU);
		XmStringFree(locstr);
	  }
		/* create a "CANCEL" button to go with it */
	if (NULL == xxg->gp_frsca[2])
	  {
		ng = 0;	XtSetArg(ldag[ng], XmNx, 100);	ng++;
		XtSetArg(ldag[ng], XmNy, 80);		ng++;
		locstr = strstufstr("Cancel", "");
		XtSetArg(ldag[ng], XmNlabelString, locstr);	ng++;
		xxg->gp_frsca[2]= XmCreatePushButtonGadget(lfra,
		 "frame_can", ldag,ng);
		XtAddCallback(xxg->gp_frsca[2], XmNactivateCallback,
			(XtCallbackProc)chgframe, XpNU);
	  }
	XtManageChildren(xxg->gp_frsca, 3);
	XtManageChild(lfra);
	return lfra;
}

void make_textports(XG_X11PROPS *xxg)
{ /* Create the text boxes for the Motif display */
	char		tlab[50];
	int		it, miw, ng, nranges, toic, toif;
	static int	tx[] = { 400, 400, 400, 400 } ;
	static int	ty[] = { 210, 360, 386, 412 } ;
	static int	tw[] = { 180, 210, 210, 210 } ;
	static int	th[] = { 148,  24,  24,  98 } ;
	static int	tc[] = {   4,  16,  24,  12 } ;
	static int	tf[NTPORTS];
	static char	*tn[] = { "Inten","hkl","Pixel","frame" } ;
	Widget		*wp;
	Arg		ldag[10];
	DISP_ATTRS	*lattr;
	static XmString	locstr;

	lattr = xxg->gp_at;	miw = lattr->di_minw;
	nranges = lattr->di_nranges;
	for (it = 0; it < NTPORTS; it++)
	   {
		tx[it] = miw + 8;
		tf[it] = ((it == 1) || (it == 2)) ? 0 : nranges;
	   }
	for (it = 0, wp = &(xxg->gp_tport[0]); it < NTPORTS; it++, wp++)
	   if (NULL == *wp)
	     {
		toif = tf[it];	toic = tc[it];
		strncpy(tlab, tn[it], 6);	tlab[6] = '\0';
		strcat(tlab, "_text");
		ng = 0;	XtSetArg(ldag[ng], XmNx, tx[it]);		ng++;
		XtSetArg(ldag[ng], XmNy, ty[it]);			ng++;
		XtSetArg(ldag[ng], XmNwidth, tw[it]);		ng++;
		locstr = strstufstr(tlab, "");
		XtSetArg(ldag[ng], XmNlabelString, locstr);	ng++;
		XtSetArg(ldag[ng], XmNrecomputeSize, False);	ng++;
		XtSetArg(ldag[ng], XmNbackground, xxg->gp_dbcol[toic]); ng++;
		XtSetArg(ldag[ng], XmNforeground, xxg->gp_dbcol[toif]); ng++;
		XtSetArg(ldag[ng], XmNheight, th[it]);		ng++;
		XtSetArg(ldag[ng], XmNfontList, xxg->gp_dflis);		ng++;
		*wp = XmCreateLabel(xxg->gp_dimain, tlab, ldag, ng);
		XmStringFree(locstr);
	     }
	XtManageChildren(xxg->gp_tport, NTPORTS);
}

void show_iscales(XG_X11PROPS *xxg)
{ /*	This sets up and displays the two scales associated with the display
	of frame data: one that sets the maximum value to be displayed and
	the other, the fractional multiplier between colors. It creates these
	scales if they don't already exist; if they do it recalculates the
	values for them based on the contents of the current frame. */
	int		imax, imult, mih, mymax, myval, ng, nranges, thx, thy;
	double		mymult, mymmax;
	static XmString	locstr;
	static Widget	colorform;
	Arg		ldag[15];
	MAIL		*lhd;
	DISP_ATTRS	*lattr;

	lattr = xxg->gp_at;	nranges = lattr->di_nranges;
	lhd = lattr->di_head + lattr->di_fmno % lattr->di_nmbox;
	mymax = (lhd->zmax < 65535 ? lhd->zmax : 65535);
	if (lattr->di_maxv > 0) myval = ((double)(lattr->di_maxv)/DEFAULT_MULT);
	else myval = mymax / 2;
	if (lattr->di_valmul >= 1.) mymult = lattr->di_valmul;
	else mymult = DEFAULT_MULT;
	mymmax = mymult * 4.;
	imult = mymult * 1000.;	imax = mymmax * 1000.;
	thx = 4;	thy = 8;	mih = lattr->di_minh;
		/* create the scales */
	if (NULL == colorform)
	  {
		ng = 0;	XtSetArg(ldag[ng], XmNx, lattr->di_minw + 188);	ng++;
		XtSetArg(ldag[ng], XmNy, 210);	ng++;
		XtSetArg(ldag[ng], XmNwidth, 180);	ng++;
		XtSetArg(ldag[ng], XmNheight,124);	ng++;
		XtSetArg(ldag[ng], XmNbackground, xxg->gp_dbcol[nranges + 3]);
		ng++;
		XtSetArg(ldag[ng], XmNforeground, xxg->gp_dbcol[nranges - 1]);
		ng++;
		colorform = XmCreateForm(xxg->gp_dimain,
			"color_form", ldag, ng);
	  }
	if (NULL == xxg->gp_iscal[0])
	  {
		ng = 0; XtSetArg(ldag[ng], XmNminimum, 1); ng++;
		XtSetArg(ldag[ng], XmNmaximum, mymax);	ng++;
		XtSetArg(ldag[ng], XmNx, thx);	ng++;
		XtSetArg(ldag[ng], XmNy, thy);	ng++;
		XtSetArg(ldag[ng], XmNscaleWidth, 172);	ng++;
		XtSetArg(ldag[ng], XmNscaleHeight, 14);		ng++;
		XtSetArg(ldag[ng], XmNorientation, XmHORIZONTAL);	ng++;
		XtSetArg(ldag[ng], XmNbackground, xxg->gp_dbcol[0]);	ng++;
		XtSetArg(ldag[ng], XmNforeground,xxg->gp_dbcol[nranges]); ng++;
		XtSetArg(ldag[ng], XmNshowValue, True);		ng++;
		locstr = strstufstr("Top-color threshold", "");
		XtSetArg(ldag[ng], XmNtitleString, locstr);	ng++;
		XtSetArg(ldag[ng], XmNvalue, myval);		ng++;
		XtSetArg(ldag[ng], XmNfontList, xxg->gp_dflis);		ng++;
		xxg->gp_iscal[0] = XmCreateScale(colorform, "disp_max",
			ldag, ng);
		XtAddCallback(xxg->gp_iscal[0], XmNvalueChangedCallback,
			(XtCallbackProc)chgdisp,XpNU);
		XmStringFree(locstr);
	  }
	else {
		ng = 0;	XtSetArg(ldag[ng], XmNmaximum, mymax);	ng++;
		XtSetArg(ldag[ng], XmNvalue, myval);	ng++;
		XtSetValues(xxg->gp_iscal[0], ldag, ng);
	     }
	XtManageChild(xxg->gp_iscal[0]);
	thy += 60;
	if (NULL == xxg->gp_iscal[1])
	  {
		ng = 0; XtSetArg(ldag[ng], XmNminimum, 100); ng++;
		XtSetArg(ldag[ng], XmNmaximum, imax);	ng++;
		XtSetArg(ldag[ng], XmNdecimalPoints, 3);	ng++;
		XtSetArg(ldag[ng], XmNx, thx);	ng++;
		XtSetArg(ldag[ng], XmNy, thy);	ng++;
		XtSetArg(ldag[ng], XmNscaleWidth, 172);	ng++;
		XtSetArg(ldag[ng], XmNscaleHeight, 16);	ng++;
		XtSetArg(ldag[ng], XmNorientation, XmHORIZONTAL);	ng++;
		XtSetArg(ldag[ng], XmNbackground, xxg->gp_dbcol[0]);	ng++;
		XtSetArg(ldag[ng], XmNforeground,xxg->gp_dbcol[nranges]); ng++;
		XtSetArg(ldag[ng], XmNshowValue, True);		ng++;
		locstr = strstufstr("Multiplier between colors", "");
		XtSetArg(ldag[ng], XmNtitleString, locstr);	ng++;
		XtSetArg(ldag[ng], XmNvalue, imult);		ng++;
		XtSetArg(ldag[ng], XmNfontList, xxg->gp_dflis);		ng++;
		xxg->gp_iscal[1] = XmCreateScale(colorform, "disp_mult",
			ldag, ng);
		XtAddCallback(xxg->gp_iscal[1], XmNvalueChangedCallback,
			(XtCallbackProc)chgdisp,XpNU);
		XmStringFree(locstr);
	  }
	else {
		ng = 0;	XtSetArg(ldag[ng], XmNmaximum, imax);	ng++;
		XtSetArg(ldag[ng], XmNvalue, imult);	ng++;
		XtSetValues(xxg->gp_iscal[1], ldag, ng);
	     }
	XtManageChild(xxg->gp_iscal[1]);
	XtManageChild(colorform);
}

/* end of dispxutil.c */
