/* cmpurf.c -- compare two UREFLS files.

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	31 Dec 2005: avoid qsorting if fewer than 2 elements
	 6 Oct 2005: housekeeping
	22 Jun 2005: cmpuargs(),mm3ii(),uhcmp(),inputobs(),effmatch(),
		cmppair(),cmpaccum(),scaleurf all declared as static
	11 May 2005: check return status from xgjournal
	27 Aug 2004: routines used only here are declared as static;cosmetics;
		routine cmpurun called rather than just existing(!)
	 2 Aug 2004: add new arg to find_dimension()
	27 Jul 2004: use dimscale() in cmpuargs() to get the (X,Y) scales
	24 Jan 2004: intendiffer is now a local variable
	21 Jan 2004: license message inserted and initial comments reordered
	 7 Jan 2004: external globals replaced with static globals or locals
		computed from find_dimension()
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	 5 Jan 2003: reorganized source-file header
	15 Oct 2002: reformat logfile output statements
	25 Mar 2002: removed MSI attachments
	 1 May 1998: initial version

   Purpose: read in two UREFLS files and compare them in various ways.
   Cross references: ../integrate/int*.c

 This reads in two UREFLS files and counts how many observations
 there are in each. It sorts them (if necessary) on omega-predicted.
 If necessary it reindexes the ones in the second set according to
 to the reindexing matrix found in the file REINDEX.
 It looks for reflections common to the two files, i.e. where the
 (hkl) values are the same and the observed (X,Y,omega) positions
 are similar. For common reflections it then looks for large deviations
 in X,Y, or omega. It also calculates a relative scale and R-factor
 between the two data sets and notes the reflections for which
 |I0 - k*I1| is large. All of these large deviations are noted in
 a logfile.
*************************************************************************/
#include	"xgen.h"
#define		CMPU_REINDU	1
#define		CMPU_USEFIT	2
#define		CMPU_USEKABSCH	4
#define		CMPU_USEFITK	(CMPU_USEFIT | CMPU_USEKABSCH)
#define	SAMEOMEGA	18
#define	SAMEXY		9
#define	OMDIFFERENT	3
#define	XYDIFFERENT	3
#define	INTENDIFFERENT	8

 typedef struct {
		char	*cr_urn0, *cr_urn1;
		int	cr_nin0, cr_nin1, cr_nmatch, cr_nprinted;
		int	cr_xydiffer, cr_xysame, cr_flags;
		double	cr_e0, cr_e1, cr_e2, cr_ibar0, cr_ibar1;
		double	cr_idsq, cr_ios0, cr_ios1;
		double	cr_le0, cr_le1, cr_le2, cr_lf0, cr_lf1, cr_lf2;
		double	cr_omdiffer, cr_rrel1, cr_scale1, cr_scl[3];
		float	cr_rein[3][3];
		FILE	*cr_fpi, *cr_fpi1, *cr_fplog;
		OBSREF	*cr_ob0, *cr_ob1;
		} CMPURUN;
		
extern int cmpumain(int argc, char *argv[]);

static void sum_cmpurun(FILE *fp, CMPURUN *crp)
{ /*	This summarizes a run of `cmpurf'. */
	fprintf(fp,
 "File name             # obs      <I>    <I/s> scalefactor Rrelative\n");
	fprintf(fp, "%20s %7d %8.1f %8.3f\n",
	 crp->cr_urn0, crp->cr_nin0, crp->cr_ibar0, crp->cr_ios0);
	fprintf(fp, "%20s %7d %8.1f %8.3f %10.5f %9.5f\n",
	 crp->cr_urn0, crp->cr_nin1, crp->cr_ibar1, crp->cr_ios1,
	 crp->cr_scale1, crp->cr_rrel1);
}

static int cmpuargs(int ac, char *av[], CMPURUN *crp)
{ /* argument processor for cmpurf */
	int		i, locwid, locht;
	char		reil[257];
	static char	*fnames[] = {
	 "urefls", "urefls1", "reindex", "xlog", NULL } ;
	static char	ftypes[] = { 1, 1, 1, 2, 0 };
	static char	**flags, **dcp, *car;
	double		el0, el1, el2, intendiffer, rxydiff, spev;

	if (EHELP == xgjournal(ac, av)) return EOK; /* journal the effort */
	crp->cr_flags = 0;
	crp->cr_omdiffer = OMDIFFERENT;
	intendiffer = INTENDIFFERENT;
	rxydiff = XYDIFFERENT;
	flags = flagproc(ac, av, "FIKORX");
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	   {
		car = (*dcp) + 1;	spev  = atof(car);
		if (**dcp == 'F') crp->cr_flags |= CMPU_USEFIT;
		else if (**dcp == 'K') crp->cr_flags |= CMPU_USEKABSCH;
		else if (**dcp == 'R') crp->cr_flags |= CMPU_REINDU;
		else if (**dcp == 'I') /* set intensity-difference level */
		  {
			if ((0. < spev) && (spev < 1.e5)) intendiffer = spev;
		  }
		else if (**dcp == 'O') /* set omega-difference level */
		  {
			if ((0. < spev) && (spev < 1.e4))
				crp->cr_omdiffer = spev;
		  }
		else if (**dcp == 'X') /* set XY-difference level */
		  {
			if ((0. < spev) && (spev < 1.e4)) rxydiff = spev;
		  }
	   }
	find_dimension(&locwid, &locht, 0, (char *)NULL);
	crp->cr_scl[0] = dimscale((short)locwid);
	crp->cr_scl[1] = dimscale((short)locht);
	crp->cr_xysame = SAMEXY * crp->cr_scl[0];
	crp->cr_xydiffer = rxydiff * crp->cr_scl[0];
	crp->cr_urn0 = trname("urefls");
	crp->cr_urn1 = trname("urefls1");
	if (crp->cr_flags & CMPU_REINDU)
	  {
		if (NULL == (AVOID *)(crp->cr_fpi = xgopen("reindex", "r")))
			crp->cr_flags &= ~(CMPU_REINDU);
		else {
			for (i = 0; i < 3; i++)
			   {
				if (NULL == fgets(reil, 80, crp->cr_fpi))
				  {
					crp->cr_flags &= ~(CMPU_REINDU);
					break;
				  }
				if (3 > sscanf(reil, "%lf%lf%lf",
				 &el0,&el1,&el2))
				  {
					crp->cr_flags &= ~(CMPU_REINDU);
					break;
				  }
				crp->cr_rein[i][0] = el0;
				crp->cr_rein[i][1] = el1;
				crp->cr_rein[i][2] = el2;
			   }
			(void)fclose(crp->cr_fpi);	crp->cr_fpi = NULL;
		     }
	  }
	crp->cr_idsq = intendiffer * intendiffer;
	if (qerror((AVOID *)(crp->cr_fpi = xgopen(crp->cr_urn0, "r")),
		"open  first input UREFLS file")) return EOPENERR;
	if (qerror((AVOID *)(crp->cr_fpi1 = xgopen(crp->cr_urn1, "r")),
		"open second input UREFLS file")) return EOPENERR;
	if (NULL == (crp->cr_fplog = propen("xlog"))) crp->cr_fplog = stdout;
	ftypes[2] = (crp->cr_flags & CMPU_REINDU) ? 1 : 0;
	ftypes[3] = (crp->cr_fplog == stdout) ? 0 : 2;
	header(stdout, REDUCTION, fnames, ftypes);
	if (crp->cr_flags & CMPU_REINDU)
	  {
		sprintf(reil,
			" Reflections in %s will be reindexed according to",
			crp->cr_urn1);
		pr3m(stdout, crp->cr_rein, reil);
	  }
	printf(" %s %s intensities will be examined\n",
	 ((crp->cr_flags & CMPU_USEKABSCH) ? "Kabsch" : "XYOmega"),
	 ((crp->cr_flags & CMPU_USEFIT) ? "Profile-fitted" : "Summation"));
	printf( " Reflections will be printed if deltaXY > %7.1f pixels, ",
		rxydiff);
	printf(" deltaOm > %7.1f frames, or deltaI/sigma > %6.1f\n",
		crp->cr_omdiffer, intendiffer);
	if (crp->cr_fplog != stdout)
	  {
		header(crp->cr_fplog, REDUCTION, fnames, ftypes);
		if (crp->cr_flags & CMPU_REINDU)
			pr3m(crp->cr_fplog, crp->cr_rein, reil);
		fprintf(crp->cr_fplog, " %s %s intensities will be examined\n",
		 ((crp->cr_flags & CMPU_USEKABSCH) ? "Kabsch" : "XYOmega"),
		 ((crp->cr_flags & CMPU_USEFIT) ?
		 "Profile-fitted" : "Summation"));
		fprintf(crp->cr_fplog,
		 " Reflections will be printed if deltaXY > %7.1f pixels, ",
		 rxydiff);
		fprintf(crp->cr_fplog,
		 " deltaOm > %7.1f frames, or deltaI/sigma > %6.1f\n",
			crp->cr_omdiffer, intendiffer);
	  }
	return EOK;
}

static void mm3ii(short ih[], CMPURUN *crp)
{ /* This reindexes a reflection with indices ih in place */
	int	i, j;
	short	iht[3];
	double	zt;

	for (i = 0; i < 3; i++)
	   {
		for (zt = j = 0; j < 3; j++) zt += crp->cr_rein[i][j] * ih[j];
		iht[i] = zt;
	   }
	for (i = 0; i < 3; i++) ih[i] = iht[i];
}

static int uhcmp(OBSREF *urp0, OBSREF *urp1)
{ /* comparison routine used in sorting on hkl&omega */
	int	di, i;
	for (i = 0; i < 3; i++)
	   {
		if ((di = urp0->obs_hkl[i] - urp1->obs_hkl[i]))
			return (di < 0) ? -1 : 1;
	   }
	return 0;	
}

static int inputobs(CMPURUN *crp)
{ /*	This counts the observations in the 2 groups of data and resorts them.
	It sticks them all in appropriate-sized buffers for sorts and
	comparisons. It then sorts each one on hkl and reindexes the second
	one if requested. */
	int	issort0, issort1, isinit;
	double	var;
	OBSREF	obt, *obq, obcp;

	issort0 = issort1 = isinit = 1;
	crp->cr_ibar0 = crp->cr_ios0 = crp->cr_ibar1 = crp->cr_ios1 = 0.;
	for (crp->cr_nin0 = 0; 0 < xgread((void *)&obt, XG_OBSREF, crp->cr_fpi);
	 (crp->cr_nin0)++)
	   {
		if (isinit != 0)
		  {
			isinit = 0;
			cpymem((AVOID *)&obcp, (AVOID *)&obt, sizeof (OBSREF));
		  }
		else if (issort0)
		  {
			if (1 <= uhcmp(&obcp, &obt)) issort0 = 0;
			cpymem((AVOID *)&obcp, (AVOID *)&obt, sizeof (OBSREF));
		  }
		var = obt.obs_sigma;	if (var < 0.01) var = 0.01;
		crp->cr_ibar0 += obt.obs_ii;
		crp->cr_ios0 += obt.obs_ii / sqrt(var);
	   }
	if (crp->cr_nin0 > 0)
	  {
		crp->cr_ibar0 /= (double)(crp->cr_nin0);
		crp->cr_ios0 /= (double)(crp->cr_nin0);
	  }
	if (-1 == fseek(crp->cr_fpi, 0L, 0))
	  {
		fprintf(stderr, "Failure to rewind %s\n", crp->cr_urn0);
		return ESEEKERR;
	  }
	if (NULL == (crp->cr_ob0 = (OBSREF *)callop(crp->cr_nin0,
	 sizeof (OBSREF), "ob0 in inputobs")))
	  {
		fprintf(stderr,
		 "Failure to allocate memory for %7d observations from %s\n", 
			crp->cr_nin0, crp->cr_urn0);
		return EALLOC;
	  }
	printf(
 " %7d observations found in %s;\nthis file %s already sorted on hkl\n",
	 crp->cr_nin0, crp->cr_urn0, (issort0 ? "is": "is not"));
	obq = crp->cr_ob0;
	for (crp->cr_nin0 = 0; 0 < xgread((void *)obq, XG_OBSREF, crp->cr_fpi);
	 crp->cr_nin0++) obq++;
	(void)fclose(crp->cr_fpi);	crp->cr_fpi = NULL;
	if ((!issort0) && (crp->cr_nin0 > 1))
	  {
		qsort((AVOID *)crp->cr_ob0, crp->cr_nin0, sizeof (OBSREF),
		 (QPE)uhcmp);
		printf("...Observations found in %s now sorted\n",
			crp->cr_urn0);
	  }
	else printf("\n");
	isinit = 1;
	for (crp->cr_nin1 = 0; 0 < xgread((void *)&obt, XG_OBSREF,
	 crp->cr_fpi1); crp->cr_nin1++)
	   {
		if (isinit != 0)
		  {
			isinit = 0;
			cpymem((AVOID *)&obcp, (AVOID *)&obt, sizeof (OBSREF));
		  }
		else if (issort1)
		  {
			if (1 <= uhcmp(&obcp, &obt)) issort1 = 0;
			cpymem((AVOID *)&obcp, (AVOID *)&obt, sizeof (OBSREF));
		  }
		var = obt.obs_sigma;	if (var < 0.01) var = 0.01;
		crp->cr_ibar1 += obt.obs_ii;
		crp->cr_ios1 += obt.obs_ii / sqrt(var);
	   }
	if (crp->cr_nin1 > 0)
	  {
		crp->cr_ibar1 /= (double)(crp->cr_nin1);
		crp->cr_ios1 /= (double)(crp->cr_nin1);
	  }
	if (-1 == fseek(crp->cr_fpi1, 0L, 0))
	  {
		fprintf(stderr, "Failure to rewind %s\n", crp->cr_urn1);
		return ESEEKERR;
	  }
	if (NULL == (crp->cr_ob1 = (OBSREF *)callop(crp->cr_nin1,
	 sizeof (OBSREF), "ob1 in inputobs")))
	  {
		fprintf(stderr,
		 "Failure to allocate memory for %7d observations from %s\n", 
			crp->cr_nin1, crp->cr_urn1);
		return EALLOC;
	  }
	printf(
 " %7d observations found in %s;\nthis file %s already sorted on hkl\n",
	 crp->cr_nin1, crp->cr_urn1, ((issort1) ? "is" : "is not"));
	obq = crp->cr_ob1;
	if (crp->cr_flags & CMPU_REINDU)
	  {
		for (crp->cr_nin1 = 0; 0 < xgread((void *)obq, XG_OBSREF,
		 crp->cr_fpi1); (crp->cr_nin1)++)
		   {
			mm3ii(obq->obs_hkl, crp);	obq++;
		   }
		printf("...Observations found in %s now reindexed\n",
		 crp->cr_urn1);
	  }
	else for (crp->cr_nin1 = 0; 0 < xgread((void *)obq, XG_OBSREF,
	 crp->cr_fpi1); (crp->cr_nin1)++) obq++;
	(void)fclose(crp->cr_fpi1);
	if ((!issort1) && (crp->cr_nin1 > 1))
	  {
		qsort((AVOID *)(crp->cr_ob1), crp->cr_nin1, sizeof (OBSREF),
			(QPE)uhcmp);
		printf("...Observations found in %s now sorted\n",
			crp->cr_urn1);
	  }
	else printf("\n");
	return EOK;
}

static int effmatch(OBSREF *obq1, OBSREF *obq0, CMPURUN *crp)
{ /*	This returns 0 if the two reflections are essentially identical.
	It returns +/-2 if they actually have different reflection indices;
	it returns 1 if they differ substantially in X,Y, or omega. */
	int	i, di;

	for (i = 0; i < 3; i++)
		if (0 != (di = obq0->obs_hkl[i] - obq1->obs_hkl[i]))
			return (di < 0) ? -2 : 2;
	if (fabs(obq0->obs_omob - obq1->obs_omob) > SAMEOMEGA) return 1;
	if (abs(obq0->obs_xob - obq1->obs_xob) > crp->cr_xysame) return 1;
	if (abs(obq0->obs_yob - obq1->obs_yob) > crp->cr_xysame) return 1;
	return 0;
}

static void cmppair(OBSREF *obq0, OBSREF *obq1, CMPURUN *crp)
{ /*	This compares two reflections already established to be similar,
	when we've already calculated an overall scale factor between the
	two datasets. It returns 1 iff they are noticeably different. */
	int	doprint;
	double	usesig0, usesig1, weight, int0, int1, wdiff;

	switch (crp->cr_flags & CMPU_USEFITK)
	      {
		case	0:	case	4:	default:
			usesig0 = obq0->obs_sunfit;
			usesig1 = obq1->obs_sunfit;
			int0 = crp->cr_scale1 * obq0->obs_iunfit;
			int1 = obq1->obs_iunfit; break;
		case	2:	case	6:
			usesig0 = obq0->obs_sigma;
			usesig1 = obq1->obs_sigma;
			int0 = crp->cr_scale1 * obq0->obs_ii;
			int1 = obq1->obs_ii; break;
	     }
	weight = crp->cr_scale1 * crp->cr_scale1 * usesig0 + usesig1;
	wdiff = int0 - int1;	wdiff *= wdiff / weight;
	doprint = 0;
	if (0 != effmatch(obq0, obq1, crp)) doprint = 1; /* shouldn't happen */
	if (fabs(obq0->obs_omob - obq1->obs_omob) > crp->cr_omdiffer)
		doprint |= 2;
	if (abs(obq0->obs_xob - obq1->obs_xob) > crp->cr_xydiffer)
		doprint |= 4;
	if (abs(obq0->obs_yob - obq1->obs_yob) > crp->cr_xydiffer)
		doprint |= 8;
	if (wdiff > crp->cr_idsq) doprint |= 16;
	if (doprint)
	  {
		if (!crp->cr_nprinted)
		  {
			fprintf(crp->cr_fplog,
	 "    indices            Reflection from dataset 0");
			fprintf(crp->cr_fplog,
	 "                         Reflection from dataset 1\n");
			fprintf(crp->cr_fplog,
	 "   h   k   l     Xob     Yob    Omob    Inten  Iscaled sigscale");
			fprintf(crp->cr_fplog,
	 "     Xob     Yob    Omob    Inten    sigma flag\n");
		  }
		(crp->cr_nprinted)++;
		fprintf(crp->cr_fplog,
		 " %3d %3d %3d %7.2lf %7.2lf %7.2lf %8.1lf %8.1lf %8.2lf",
		 obq0->obs_hkl[0], obq0->obs_hkl[1], obq0->obs_hkl[2],
		 ((double)obq0->obs_xob) / crp->cr_scl[0], 
		 ((double)obq0->obs_yob) / crp->cr_scl[1], obq0->obs_omob,
		 int0 / crp->cr_scale1, int0, crp->cr_scale1 * sqrt(usesig0));
		fprintf(crp->cr_fplog,
		 " %7.2lf %7.2lf %7.2lf %8.1f %8.2lf %4d\n",
		 ((double)obq1->obs_xob) / crp->cr_scl[0], 
		 ((double)obq1->obs_yob) / crp->cr_scl[1], obq1->obs_omob,
		 int1, sqrt(usesig1), doprint);
		xeflush(crp->cr_fplog);
	  }
}

static void cmpaccum(OBSREF *obq0, OBSREF *obq1, CMPURUN *crp)
{ /* accumulate scale and R-factor sums for cmpurf */
	double	weight, int0, int1;

	switch (crp->cr_flags & CMPU_USEFITK)
	      {
		case	0:	case	4:	default:
		 weight = obq0->obs_sunfit + obq1->obs_sunfit;
		 int0 = obq0->obs_iunfit; int1 = obq1->obs_iunfit; break;
		case	2:	case	6:
		 weight = obq0->obs_sigma + obq1->obs_sigma;
		 int0 = obq0->obs_ii; int1 = obq1->obs_ii; break;
	     }
	(crp->cr_nmatch)++;
	crp->cr_e0 += int0 * int0 * weight;
	crp->cr_e1 += int1 * int1 * weight;
	crp->cr_e2 += int0 * int1 * weight;
	/* augment sums associated with Lorentz/polarization
	 comparison */
	int0 = ((double)obq0->obs_lpa) * 1.e-4;
	int1 = ((double)obq1->obs_lpa) * 1.e-4;
	crp->cr_le0 += 1.;		crp->cr_lf0 += int1;
	crp->cr_le1 += int0;		crp->cr_lf1 += int0 * int1;
	crp->cr_le2 += int0 * int0;	crp->cr_lf2 += int1 * int1;
}

static int scaleurf(CMPURUN *crp)
{ /*	This goes through the two sorted lists of OBSREF reflections and
	finds the ones that match. In order to match the hkl values must be
	the same and the omega, X, and Y values must be within SAMEOMEGA
	frames and SAMEXY pixels. For each reflection so found we:
	 (a) increment the count of matched reflections
	 (b) increment sums that will calculate a scale factor between the
	     intensities as measured in set 0 and those measured in set 1.
	 We calculate and print the scale and R-factors so derived.
	 We then go back through and examine the matched reflections again.
	 This time we flag those for which:
	  |omobs(1) - omobs(0)| > omdiffer;
	  | Xobs(1) -  Xobs(0)| > xydiffer;
	  | Yobs(1) -  Yobs(0)| > xydiffer;
	  | I(1) -  g * I(0)| > intendiffer; (g == scale factor) */
	int	em, ji, j00;
	double	den, linter, lrfac, lslope;
	OBSREF	*obt0, *obt0i, *obt1, *obt10, *ob1end, *obt1i;
	static char	disjerr[] = " Disjoint omega ranges";

	crp->cr_e0 = crp->cr_e1 = crp->cr_e2 = crp->cr_nmatch = 0;
	crp->cr_le0 = crp->cr_le1 = crp->cr_le2 = 0.;
	crp->cr_lf0 = crp->cr_lf1 = crp->cr_lf2 = 0;
	if (1 <= uhcmp(crp->cr_ob0, crp->cr_ob1))
	  {
		for (obt1 = crp->cr_ob1, ji = 0; ji < crp->cr_nin1;
		 ji++, obt1++) if (0 >= uhcmp(crp->cr_ob0, obt1)) break;
		obt1i = obt1;
		if (ji >= crp->cr_nin1)
		  {
			printf(disjerr);	return EBAD;
		  }
		obt0i = obt0 = crp->cr_ob0;	ji = 0;
	  }
	else {
		for (obt0 = crp->cr_ob0, ji = 0; ji < crp->cr_nin0;
		 ji++, obt0++) if (0 >= uhcmp(crp->cr_ob1, obt0)) break;
		obt0i = obt0;
		if (ji >= crp->cr_nin0)
		  {
			printf(disjerr);	return EBAD;
		  }
		obt1i = obt1 = crp->cr_ob1;
	     }
	ob1end = crp->cr_ob1 + (crp->cr_nin1 - 1); /* note end of 2nd list */
		/* advance through 1st list */
	for (j00 = ji; ji < crp->cr_nin0; ji++, obt0++)
	   {
		obt10 = obt1;
		if ((em = uhcmp(obt1, obt0)) < 0) /* let hkl catch up */
		  {
			while ((em = uhcmp(obt1, obt0)) < 0)
			     {
				obt1++;	if (obt1 >= ob1end) break;
			     }
		  }
		if (obt1 > ob1end) break;
		if (!em)
		  {
			if (0 == (em = effmatch(obt1, obt0, crp)))
				cmpaccum(obt0, obt1, crp);
			else {
				while (abs(em) < 2)
				  {
					obt1++;	if (obt1 > ob1end) break;
					if (0 == (em = effmatch(obt1, obt0,
					 crp))) cmpaccum(obt0, obt1, crp);
					else if (2 == em) break;
				  }
				obt1 = obt10;
			  }
		  }
	   }
	if ((crp->cr_nmatch < 3) || (crp->cr_e0 < DFUZZ))
	  {
		printf("Not enough matches to be useful: only %d\n",
		 crp->cr_nmatch);
		if (crp->cr_fplog != stdout) fprintf(crp->cr_fplog,
		 "Not enough matches to be useful: only %d\n", crp->cr_nmatch);
		return EBAD;
	  }
	crp->cr_scale1 = crp->cr_e2 / crp->cr_e0;
	printf("Scale factor between datasets for %7d reflections = %12.5e\n",
		crp->cr_nmatch, crp->cr_scale1);
	fprintf(crp->cr_fplog,
	 "Scale factor between datasets for %7d reflections = %12.5e\n",
		crp->cr_nmatch, crp->cr_scale1);
	if (fabs(crp->cr_e2) > DFUZZ)
	  {
		printf("e0, e1, e2 = %14.7e, %14.7e, %14.7e\n",
			crp->cr_e0, crp->cr_e1, crp->cr_e2);
		crp->cr_rrel1 = (crp->cr_e1 / crp->cr_e2) *
			(crp->cr_e0 / crp->cr_e2) - 1.;
		if (crp->cr_rrel1 > DFUZZ)
		  {
			crp->cr_rrel1 = sqrt(crp->cr_rrel1);
			printf("R-factor between datasets = %12.5e\n",
			 crp->cr_rrel1);
			if (crp->cr_fplog != stdout) fprintf(crp->cr_fplog,
			"R-factor between datasets = %12.5e\n", crp->cr_rrel1);
		  }
		else printf("R-factor unstable: R^2 = %12.5e\n",
			crp->cr_rrel1);
	  }
	else printf("R-factor nonsense: crossproduct = %12.5e\n", crp->cr_e2);
	crp->cr_nprinted = 0;	obt1 = obt1i;	obt0 = obt0i;
	den = crp->cr_le0 * crp->cr_le2 - crp->cr_le1 * crp->cr_le1;
	if (fabs(den) < DFUZZ) printf(
	 "Lorentz comparison impossible: determinant = %14.7e\n", den);
	else {
		lslope = (crp->cr_le0 * crp->cr_lf1 -
			crp->cr_le1 * crp->cr_lf0) / den;
		linter = (crp->cr_le2 * crp->cr_lf0 -
			crp->cr_le1 * crp->cr_lf1) / den;
		printf("Lorentz out = %8.4f * Lorentz in + %8.5f\n",
		 lslope, linter);
		if (crp->cr_lf2 > DFUZZ)
		  {
			lrfac = 1. + (lslope * crp->cr_le2 + 2. * lslope *
			 (linter * crp->cr_le1 - crp->cr_lf1) + linter *
			 (linter * crp->cr_le0 - 2. * crp->cr_lf0)) /
			 crp->cr_lf2;
			if (lrfac >=0.) printf("Lorentz R-factor = %14.7e\n",
			 sqrt(lrfac));
			else printf("Lorentz R-factor imaginary!\n");
		  }
		else printf("Lorentz R-factor calc singular\n");
	     }
		/* advance through 1st list */
	for (ji = j00; ji < crp->cr_nin0; ji++, obt0++)
	   {
		obt10 = obt1;
		if ((em = uhcmp(obt1, obt0)) < 0) /* let hkl catch up */
		  {
			while ((em = uhcmp(obt1, obt0)) < 0)
			     {
				obt1++;	if (obt1 >= ob1end) break;
			     }
		  }
		if (obt1 > ob1end) break;
		if (!em)
		  {
			if (0 == (em = effmatch(obt1, obt0, crp)))
				cmppair(obt0, obt1, crp);
			else {
				while (abs(em) < 2)
				  {
					obt1++;	if (obt1 > ob1end) break;
					if (0 == (em = effmatch(obt1, obt0,	
					 crp))) cmppair(obt0, obt1, crp);
					else if (2 == em) break;
				  }
				obt1 = obt10;
			  }
		  }
	   }
	printf(
	 " %6d reflections differed significantly between the two sets\n",
		crp->cr_nprinted);
	return EOK;
}
	
int cmpumain(int argc, char *argv[])
{ /* Mainline for the cmpurf functionality */
	int		resu;
	static CMPURUN	crs;

	if (EOK != (resu = cmpuargs(argc, argv, &crs))) return resu;
	if (EOK != (resu = inputobs(&crs))) return resu;
	if (EOK != (resu = scaleurf(&crs))) return resu;
	if (crs.cr_fplog != stdout) sum_cmpurun(crs.cr_fplog, &crs);
	sum_cmpurun(stdout, &crs);
	globalkill();	return EOK;
}

/* end of cmpurf.c */
