/* editimage.c -- database editor for detector images

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions; reorganize use of "{"
	11 Aug 2006: manage byteswapped formats without recourse to
		defines for those formats
	23 Jun 2006: include xgimage.h so we can read and write images
	 6 Mar 2006: handle Mar IP output formats properly
	16 Oct 2005: major rewrite to eliminate static globals and use
		a generally useful typedef
	 6 Oct 2005: housekeeping
	22 Jun 2005: editimagemain(): "command"-> "comd" to reduce confusion
	11 May 2005: check return status from xgjournal
	28 Aug 2004: routines used only here are declared as static;cosmetics
	24 Jan 2004: some static globals replaced by local variables
	21 Jan 2004: license message inserted
	 7 Jan 2004: static globals and function args replace extern globals
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	11 Jun 2003: correct handling of values > 65535
		in `modcircle' and `modring'
	 5 Jan 2003: reorganized source-file header
	 1 Nov 2002: (?) rename defines for action variable
		(GAUSSIAN becomes EI_GAUSSIAN, etc.); replace framename
		with lfname so it doesn't collide with global;
		trivial source reformatting
	25 Mar 2002: removed MSI attachments
	 1 Dec 1994: diagnostic output allows for detector (X,Y) > 999
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: minor I/O reorganization
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version

   Purpose: first stab at a database editor for detector images.
	Currently this simply takes arguments from the terminal.
   Cross references: ../xglib/image.c, ../xglib/imhead.c, ../border/editbord.c
*************************************************************************/
#include	"xgimage.h"
/* values of the `actval' variable */
#define	EI_ZEROOUT	0
#define	EI_LOCALMEAN	1
#define	EI_USERSPECVAL	2
#define	EI_GAUSSIAN	3
/* Values of the "shape" variable: */
#define		COLUMN		2
#define		ROW		4
#define		BOX		010
#define		OBOX		020
#define		CIRCLE		040
#define		OCIR		0100
#define		RING		0200

#define		NO_ACTION	4
#define		MAX_ARGS	5
#define		CHARS_PER_ARG	20
#define		NSUPPORTED	5	/* # of supported output formats */
#define		MAX_DIMENSION	9000	/* max # rows or columns */

 typedef struct {
		char	**ei_cargs, *ei_inshape, *ei_lfname;
		int	ei_ncargs, ei_imflags, ei_nargs, ei_npts;
		int	ei_dwid, ei_dht, ei_xmn, ei_xmx, ei_ymn, ei_ymx;
		int	ei_xcent, ei_ycent, ei_radius;
		int	ei_shape, ei_msgwid, ei_action;
		int	*ei_iargs;
		xgulong	ei_outtype;
		double	ei_meancts, ei_specval;
		u_short	ei_newval;
		FILE	*ei_fpl;
		MAIL	*ei_heads;
		} EIMWORK;

extern int editimagemain(int argc, char *argv[]);

static void pixelmod(int x, int y, EIMWORK *eip)
{ /*	This alters the value at the point (x, y) to the value `newval'.
	The tricky part is getting the overflow table right. */
	u_short	oldval;
	u_char	*ucp;
	MAIL	*lhd;
	VOFLOW	*vofp, *vofn, *vofe;

	lhd = eip->ei_heads;
	oldval = imvalue(lhd, x, y);
	/* set the new 8-bit value properly */
	ucp = lhd->data8 + x * eip->ei_dht + y;
	*ucp = (eip->ei_newval < 256 ? eip->ei_newval : 255);
	if (eip->ei_newval < oldval)
		lhd->eventcount -= (oldval - eip->ei_newval);
	else lhd->eventcount += eip->ei_newval - oldval;
	if (eip->ei_newval < 256 && oldval < 256) return;
	/* Fix up the overflow table if necessary */
	if (lhd->oflowcount)
	  {
		vofe = lhd->overflow + lhd->oflowcount;
		for (vofp = lhd->overflow; vofp < vofe; vofp++)
		   {
			if ((vofp->xcoord == x) &&
			 (vofp->ycoord == (eip->ei_dht - y - 1)))
			  {
				if (eip->ei_newval > 255) /* change overflow */
				  {
					vofp->value = eip->ei_newval;	return;
				  }
				else	/* get rid of this overflow */
				  {
					if (lhd->oflowcount == 1)
					  {
						lhd->oflowcount = 0;
						return;
					  }
					for (vofn = vofp + 1;
					 vofn < vofe; vofn++,vofp++)
					   cpymem((AVOID *)vofp, (AVOID *)vofn,
						sizeof (VOFLOW));
					(lhd->oflowcount)--;	return;
				  } /* end of else (newval < 255) */
			  } /* end of if overflow match */
		   } /* end of loop through current overflows */
		/* if we get here, the current pixel wasn't overflowed before*/
		if (eip->ei_newval < 256) return; /* no sweat */
		(lhd->oflowcount)++;	vofp = vofe;
		vofp->xcoord = x;	vofp->ycoord = eip->ei_dht - y - 1;
		vofp->value = eip->ei_newval;	vofe++;	return;
	  } /* end of if clause */
	/* if we didn't have any overflows before and the new value
	is > 255, we will have to create an overflow table.  I'm assuming
	we've created room for at least 800 overflows (see image.c) so that
	we're just defining the space now */
	if (eip->ei_newval < 256) return;	/* phew */
	lhd->oflowcount = 1;	vofe = lhd->overflow + 1;
}

static void modbox(EIMWORK *eip)
{ /*	This changes pixels in a rectangular box from [xmn,ymn]
	to [xmx, ymx]. */
	register int	x, y;
	double		rsq, rval, rx, rxsq, ry;

	if (eip->ei_xmx < eip->ei_xmn)
	  {
		x = eip->ei_xmn;	eip->ei_xmn = eip->ei_xmx;
		eip->ei_xmx = x;
	  }
	if (eip->ei_ymx < eip->ei_ymn)
	  {
		y = eip->ei_ymn;	eip->ei_ymn = eip->ei_ymx;
		eip->ei_ymx = y;
	  }
	eip->ei_npts = (eip->ei_xmx - eip->ei_xmn + 1) *
		(eip->ei_ymx - eip->ei_ymn + 1);
	if (eip->ei_action == EI_LOCALMEAN)
	  {
		for (x = eip->ei_xmn, eip->ei_meancts = 0;
		 x <= eip->ei_xmx; x++)
		      for (y = eip->ei_ymn; y <= eip->ei_ymx; y++)
			eip->ei_meancts += imvalue(eip->ei_heads, x, y);
		if (eip->ei_npts > 0)
		  {
			eip->ei_meancts /= (double)(eip->ei_npts);
			eip->ei_newval = (eip->ei_meancts < 0.4999 ?
				0 : eip->ei_meancts);
		  }
		else	eip->ei_newval = 0;
	  }
	if (eip->ei_action == EI_GAUSSIAN)
	  {
		rx = (eip->ei_xmn - eip->ei_xmx) / 2.;
		for (x = eip->ei_xmn; x <= eip->ei_xmx; x++, rx += 1.)
		   {
			rxsq = rx * rx;	ry = (eip->ei_ymn - eip->ei_ymx) / 2.; 
			for (y = eip->ei_ymn; y <= eip->ei_ymx; y++, ry += 1.)
			   {
				rsq = rxsq + ry * ry;
				rval = eip->ei_specval * exp(-rsq / 4.);
				eip->ei_newval = (rval < 0.49999 ? 0 : rval);
				pixelmod(x, y, eip);
			   }
		   }
	  }
	else for (x = eip->ei_xmn; x <= eip->ei_xmx; x++)
		for (y = eip->ei_ymn; y <= eip->ei_ymx; y++)
			pixelmod(x, y, eip);
}

static void modobox(EIMWORK *eip)
{ /*	This sets all the pixels outside the box bounded by
	xmn, ymn, xmx, and ymx to the value `newval'. */
	register int	x, y;

	if (eip->ei_xmx < eip->ei_xmn)
	  {
		x = eip->ei_xmn;	eip->ei_xmn = eip->ei_xmx;
		eip->ei_xmx = x;
	  }
	if (eip->ei_xmx >= eip->ei_dwid) eip->ei_xmx = eip->ei_dwid - 1;
	if (eip->ei_xmn < 0) eip->ei_xmn = 0;
	if (eip->ei_ymx < eip->ei_ymn)
	  {
		y = eip->ei_ymn;	eip->ei_ymn = eip->ei_ymx;
		eip->ei_ymx = y;
	  }
	if (eip->ei_ymx >= eip->ei_dht) eip->ei_ymx = eip->ei_dht - 1;
	if (eip->ei_ymn < 0) eip->ei_ymn = 0;
	if (eip->ei_action == EI_LOCALMEAN)
	  {
		eip->ei_meancts = eip->ei_npts = 0;
		for (x = 0; x < eip->ei_xmn; x++)
		   for (y = 0; y < eip->ei_dht; y++)
		      {
			eip->ei_meancts += imvalue(eip->ei_heads, x, y);
			eip->ei_npts++;
		      }
		for (x = eip->ei_xmx + 1; x < eip->ei_dwid; x++)
		   for (y = 0; y < eip->ei_dht; y++)
		      {
			eip->ei_meancts += imvalue(eip->ei_heads, x, y);
			eip->ei_npts++;
		      }
		for (x = eip->ei_xmn; x <= eip->ei_xmx; x++)
		   {
			for (y = 0; y < eip->ei_ymn; y++)
			   {
				eip->ei_meancts += imvalue(eip->ei_heads,x,y);
				eip->ei_npts++;
		           }
			for (y = eip->ei_ymx + 1; y < eip->ei_dht; y++)
			   {
				eip->ei_meancts += imvalue(eip->ei_heads,x,y);
				eip->ei_npts++;
			   }
		   }
		if (eip->ei_npts > 0)
		  {
			eip->ei_meancts /= (double)(eip->ei_npts);
			eip->ei_newval = (eip->ei_meancts < 0.4999 ?
				0 : eip->ei_meancts);
		  }
		else	eip->ei_newval = 0;
	  }
	for (x = 0; x < eip->ei_xmn; x++)
	   for (y = 0; y < eip->ei_dht; y++) pixelmod(x, y, eip);
	for (x = eip->ei_xmx + 1; x < eip->ei_dwid; x++)
	   for (y = 0; y < eip->ei_dht; y++) pixelmod(x, y, eip);
	for (x = eip->ei_xmn; x <= eip->ei_xmx; x++)
	   {
		for (y = 0; y < eip->ei_ymn; y++) pixelmod(x, y, eip);
		for (y = eip->ei_ymx + 1; y < eip->ei_dht; y++)
			pixelmod(x, y, eip);
	   }
}

static void modcircle(EIMWORK *eip)
{ /*	This sets all the pixels within a circle to the value `newval'.
	The circle has center (xcent, ycent) and radius `radius'. */
	register int	x, y;
	int		dx, dxsq, dy, maxx, maxy, minx, miny, radsq;
	double		alpha, rsq, rval;

	minx = eip->ei_xcent - eip->ei_radius;	if (minx < 1) minx = 1;
	maxx = eip->ei_xcent + eip->ei_radius;
	radsq = eip->ei_radius * eip->ei_radius;
	if (maxx >= eip->ei_dht-1) maxx = eip->ei_dht-2;
	miny = eip->ei_ycent - eip->ei_radius;	if (miny < 1) miny = 1;
	maxy = eip->ei_ycent + eip->ei_radius;
	if (maxy >= eip->ei_dht-1) maxy = eip->ei_dht-2;
	if (eip->ei_action == EI_LOCALMEAN)
	  {
		eip->ei_meancts = eip->ei_npts = 0;
		for (x = minx, dx = minx - eip->ei_xcent; x <= maxx; x++, dx++)
		   {
			dxsq = dx * dx;
			for (y = miny, dy = miny - eip->ei_ycent;
			 y <= maxy; y++, dy++)
			   if (dxsq + dy * dy < radsq)
			     {
				(eip->ei_npts)++;
				eip->ei_meancts += imvalue(eip->ei_heads, x,y);
			     }
		   }
		if (eip->ei_npts > 0)
		  {
			eip->ei_meancts /= (double)(eip->ei_npts);
			eip->ei_newval = (eip->ei_meancts < 0.4999) ?
				0 : eip->ei_meancts;
		  }
		else	eip->ei_newval = 0;
	  }
	if (eip->ei_action == EI_GAUSSIAN)
	  {
		alpha = (radsq >= 1 ? 8. / radsq : 8);
		for (x = minx, dx = minx - eip->ei_xcent; x <= maxx; x++, dx++)
		   {
			dxsq = dx * dx;
			for (y = miny, dy = miny - eip->ei_ycent;
			 y <= maxy; y++, dy++)
			   if ((rsq = dxsq + dy * dy) < radsq)
			     {
				rval = eip->ei_specval * exp(-alpha * rsq);
				eip->ei_newval = (rval < 0.49999) ? 0 :
				 (rval > 65535 ? 65535 : rval);
				pixelmod(x, y, eip);
			     }
		   }
	  }
	else for (x = minx, dx = minx - eip->ei_xcent; x <= maxx; x++, dx++)
	   {
		dxsq = dx * dx;
		for (y = miny, dy = miny - eip->ei_ycent; y <= maxy; y++, dy++)
		   if (dxsq + dy * dy < radsq)
			pixelmod(x, y, eip);
	   }
}

static void modring(EIMWORK *eip)
{ /*	This sets all the pixels within a ring to the value `newval'.
	The ring has center (xcent, ycent) and radius `radius'. */
	register int	x, y;
	int		dx, dxsq, dy, maxx, maxy, minx, miny, radsq;
	double		alpha, rsq, rsqp, rval;
	char		firstenc;

	minx = eip->ei_xcent - eip->ei_radius;	if (minx < 1) minx = 1;
	maxx = eip->ei_xcent + eip->ei_radius;
	radsq = eip->ei_radius * eip->ei_radius;
	if (maxx >= eip->ei_dht-1) maxx = eip->ei_dht-2;
	miny = eip->ei_ycent - eip->ei_radius;	if (miny < 1) miny = 1;
	maxy = eip->ei_ycent + eip->ei_radius;
	if (maxy >= eip->ei_dht-1) maxy = eip->ei_dht-2;
	if (eip->ei_action == EI_LOCALMEAN)
	  {
		eip->ei_meancts = eip->ei_npts = 0;
		for (x = minx, dx = minx - eip->ei_xcent; x <= maxx; x++, dx++)
		   {
			dxsq = dx * dx;
			for (y = miny, dy = miny - eip->ei_ycent;
			 y <= maxy; y++, dy++)
			   if (dxsq + dy * dy < radsq)
			     {
				(eip->ei_npts)++;
				eip->ei_meancts += imvalue(eip->ei_heads, x,y);
			     }
		   }
		if (eip->ei_npts > 0)
		  {
			eip->ei_meancts /= (double)(eip->ei_npts);
			eip->ei_newval = (eip->ei_meancts < 0.4999) ?
				0 : eip->ei_meancts;
		  }
		else	eip->ei_newval = 0;
	  }
	if (eip->ei_action == EI_GAUSSIAN)
	  {
		alpha = (radsq >= 1 ? 8. / radsq : 8);
		for (x = minx, dx = minx - eip->ei_xcent; x <= maxx; x++, dx++)
		   {
			dxsq = dx * dx;
			firstenc = 1; /* flag: 1st good point in column */
			for (y = miny, dy = miny - eip->ei_ycent;
			 y <= maxy; y++, dy++)
			   if ((rsq = dxsq + dy * dy) < radsq)
			     {
				if (firstenc)
				  {
					rval = eip->ei_specval *
						exp(-alpha * rsq);
					eip->ei_newval = (rval < 0.49999) ? 0 :
					 (rval > 65535 ? 65535 : rval);
					pixelmod(x, y, eip);
					firstenc = 0;/* not 1st in column*/
				  }
				else
				  {
					rsqp = dxsq + (dy + 1) * (dy + 1);
				    if (rsqp >= radsq)
				      {
					rval = eip->ei_specval *
						exp(-alpha * rsq);
					eip->ei_newval = (rval < 0.49999) ? 0 :
					 (rval > 65535 ? 65535 : rval);
					pixelmod(x, y, eip);
				      } /* end if (rsqp >= radsq) clause */
				  } /* end of else for "if (firstenc)"*/
			      } /* end of if ((rsq ... < radsq) */
				      
		   } /* end of loop through x */
	  } /* end of "if (eip->ei_action == EI_GAUSSIAN)" clause */
	else for (x = minx, dx = minx - eip->ei_xcent; x <= maxx; x++, dx++)
	   {
		dxsq = dx * dx;
		firstenc = 1; /* flag: 1st good point in column */
		for (y = miny, dy = miny - eip->ei_ycent; y <= maxy; y++, dy++)
		   if (dxsq + dy * dy < radsq)
		     {
		       if (firstenc)
			 {
				pixelmod(x, y, eip); firstenc = 0; }
			else {
				rsqp = dxsq + (dy + 1) * (dy + 1);
				if (rsqp >= radsq) pixelmod(x, y, eip);
			     } /* end of else for "if (firstenc)"*/
		     } /* end of "if (dxsq .. " clause */
	   } /* end of loop through x values and else for EI_GAUSSIAN */
}

static void modocircle(EIMWORK *eip)
{ /*	This sets all the pixels outside a circle to the value `newval'.
	The circle has center (xcent, ycent) and radius `radius'. */
	register int	x, y;
	int		dx, dxsq, dy, minx, miny, maxx, maxy;
	int		qmaxx, qmaxy, radsq;

	minx = eip->ei_xcent - eip->ei_radius;	if (minx < 1) minx = 1;
	maxx = eip->ei_xcent + eip->ei_radius;
	radsq = eip->ei_radius * eip->ei_radius;
	if (maxx >= eip->ei_dht-1) maxx = eip->ei_dht-2;
	miny = eip->ei_ycent - eip->ei_radius;	if (miny < 1) miny = 1;
	maxy = eip->ei_ycent + eip->ei_radius;
	if (maxy >= eip->ei_dht-1) maxy = eip->ei_dht-2;
	if (eip->ei_action == EI_LOCALMEAN)
	  {
		eip->ei_meancts = eip->ei_npts = 0;	qmaxx = eip->ei_dht - 2;
		qmaxy = eip->ei_dwid - 2;
		for (x = 0, dx = -eip->ei_xcent; x <= qmaxx; x++, dx++)
		   {
			dxsq = dx * dx;
			for (y = 0, dy = -eip->ei_ycent; y <= qmaxy; y++, dy++)
			   if (dxsq + dy * dy > radsq)
			     {
				(eip->ei_npts)++;
				eip->ei_meancts += imvalue(eip->ei_heads, x,y);
			     }
		   }
		if (eip->ei_npts > 0)
		  {
			eip->ei_meancts /= (double)(eip->ei_npts);
			eip->ei_newval = (eip->ei_meancts < 0.4999 ?
				0 : eip->ei_meancts);
		  }
		else	eip->ei_newval = 0;
		for (x = 0, dx = -eip->ei_xcent; x <= qmaxx; x++, dx++)
		   {
			dxsq = dx * dx;
			for (y = 0, dy = -eip->ei_ycent; y <= qmaxy; y++, dy++)
			   if (dxsq + dy * dy > radsq)
				pixelmod(x, y, eip);
		   }
	  }
	else {	/* first, wipe out the box outside the circle */
		modobox(eip);
		/* Then get the nail-parings between box and circle */
		for (x = minx, dx = minx - eip->ei_xcent; x <= maxx; x++, dx++)
		   {
			dxsq = dx * dx;
			for (y = miny, dy = miny - eip->ei_ycent;
			 y <= maxy; y++, dy++)
			   if (dxsq + dy * dy > radsq) pixelmod(x, y, eip);
		   }
	     }
}

static void modrows(EIMWORK *eip)
{ /*	This sets the pixels in `msgwid' rows centered on y = ycent
	to the value `newval' */
	register int	x, y;
	int		ymn, ymx;

	ymn = eip->ei_ycent - eip->ei_msgwid / 2;
	ymx = eip->ei_ycent + eip->ei_msgwid / 2;
	if (ymn < 1) ymn = 1;
	if (ymx > eip->ei_dht-2) ymx = eip->ei_dht-2;
	if (eip->ei_action == EI_LOCALMEAN)
	  {
		eip->ei_meancts = eip->ei_npts = 0;
		for (y = ymn; y <= ymx; y++)
		   for (x = 1; x < eip->ei_msgwid - 1; x++)
		      {
			eip->ei_npts++;
			eip->ei_meancts += imvalue(eip->ei_heads, x, y);
		      }
		if (eip->ei_npts > 0)
		  {
			eip->ei_meancts /= (double)eip->ei_npts;
			eip->ei_newval = (eip->ei_meancts < 0.4999 ?
				0 : eip->ei_meancts);
		  }
		else	eip->ei_newval = 0;
	  }
	for (y = ymn; y <= ymx; y++)
	   for (x = 1; x < eip->ei_msgwid - 1; x++) pixelmod(x, y, eip);
}

static void modcolumns(EIMWORK *eip)
{ /*	This changes the activation status of `sgwid' columns centered
	on column `xcent'. */
	register int	x, y;
	int		xmn, xmx;

	xmn = eip->ei_xcent - eip->ei_msgwid / 2;
	xmx = eip->ei_xcent + eip->ei_msgwid / 2;
	if (xmn < 1) xmn = 1;
	if (xmx > eip->ei_msgwid-2) xmx = eip->ei_msgwid - 2;
	if (eip->ei_action == EI_LOCALMEAN)
	  {
		eip->ei_meancts = eip->ei_npts = 0;
		for (x = xmn; x <= xmx; x++)
		   for (y = 1; y < eip->ei_dht - 1; y++)
		      {
			eip->ei_npts++;
			eip->ei_meancts += imvalue(eip->ei_heads, x, y);
		      }
		if (eip->ei_npts > 0)
		  {
			eip->ei_meancts /= (double)eip->ei_npts;
			eip->ei_newval = (eip->ei_meancts < 0.4999 ?
				0 : eip->ei_meancts);
		  }
		else	eip->ei_newval = 0;
	  }
	for (x = xmn; x <= xmx; x++)
	   for (y = 1; y < eip->ei_dht - 1; y++) pixelmod(x, y, eip);
}

static char *get_comd(EIMWORK *eip)
{ /*	This prompts the user for a command and returns a lowercase version of
	the command given by the user.  It also stuffs any other strings
	given by the user into eip->ei_cargs. */
	register int	i, j;
	static char	cmd[CHARS_PER_ARG], lolin[80];

	fprintf(stderr, "> ");	for (j = CHARS_PER_ARG; j--; ) cmd[j] = '\0';
	for (i = MAX_ARGS; i--; ) for (j = CHARS_PER_ARG; j--; )
		eip->ei_cargs[i][j] = '\0';
	if (NULL == fgets(lolin, 78, stdin))
	  {
		strcpy(cmd, "ex");	return &(cmd[0]);	}
	(void)sscanf(lolin, "%s %s %s %s %s %s",
		cmd, eip->ei_cargs[0], eip->ei_cargs[1],
		eip->ei_cargs[2], eip->ei_cargs[3], eip->ei_cargs[4]);
	lowercaseify(cmd);
	for (eip->ei_ncargs = 0; eip->ei_ncargs < MAX_ARGS &&
	 eip->ei_cargs[eip->ei_ncargs][0] != '\0'; (eip->ei_ncargs)++)
		lowercaseify(eip->ei_cargs[eip->ei_ncargs]);
	return &(cmd[0]);
}

static void logchange(EIMWORK *eip, FILE *fp)
{ /*	This summarizes a change in the mask to the stream fp. */
	switch (eip->ei_shape)
	      {
		case	BOX:	fprintf(fp,
		 " box from [x,y]=[%4d,%4d] to [%4d,%4d] ", eip->ei_iargs[0],
		  eip->ei_iargs[1], eip->ei_iargs[2], eip->ei_iargs[3]); break;
		case	CIRCLE:	fprintf(fp,
		 "Circle of radius %4d centered at [%4d,%4d] ",
		 eip->ei_iargs[2], eip->ei_iargs[0], eip->ei_iargs[1]);	break;
		case	COLUMN:	fprintf(fp,
		 " %3d columns centered on column %4d ",
			eip->ei_iargs[1], eip->ei_iargs[0]);	break;
		case	ROW:	fprintf(fp,
		 " %3d rows centered on row %4d",
			eip->ei_iargs[1], eip->ei_iargs[0]);	break;
		case	OBOX:	fprintf(fp,
 " Points outside of box from [x,y]=[%4d,%4d] to [%4d,%4d] ", eip->ei_iargs[0],
	eip->ei_iargs[1], eip->ei_iargs[2], eip->ei_iargs[3]);	break;
		case	OCIR:	fprintf(fp,
 "Outside of circle of radius %4d centered at [%4d,%3d] ", eip->ei_iargs[2],
			eip->ei_iargs[0], eip->ei_iargs[1]);	break;
		case	RING:	fprintf(fp,
		 "Ring of radius %4d centered at [%4d,%4d] ",
		 eip->ei_iargs[2], eip->ei_iargs[0], eip->ei_iargs[1]); break;
	     }
	if (eip->ei_action != EI_GAUSSIAN)
		fprintf(fp, "set to %d\n", (int)(eip->ei_newval));
	else fprintf(fp, "treated as a gaussian with peak %6.0f\n",
		eip->ei_specval);
}

static void proc_cmd(char *icmd, EIMWORK *eip)
{ /*	This calls the appropriate routine to change a particular part of
	the image.  It then scribbles meaningful stuff to the
	standard error and the logfile. */
	int		shape, j, len;
	char		cmd[50];
	static int	shapeval[] =
		{ BOX, CIRCLE, COLUMN, ROW, OBOX, OCIR, RING } ;
	static char	*cmds[] =
		{ "b", "ci", "co", "ro", "ob", "oc", "ri" } ;

	eip->ei_inshape = eip->ei_cargs[0];	/* for now */
	for (j = 0, shape = -1; j < 7; j++)
	   {
		len = strlen(cmds[j]);
		if (!strncmp(eip->ei_inshape, cmds[j], (size_t)len))
		  {
			eip->ei_shape = shapeval[j];	break;
		  }
	   }
	if (0 > eip->ei_shape)
	  {
		fprintf(stderr, "Unrecognized shape: %s\n", eip->ei_inshape);
		return;
	  }
	strcpy(cmd, icmd);
	/* The value we're going to stick in all the modifiable pixels
	is already known by now unless we're doing a local average. */
	if (eip->ei_action == EI_USERSPECVAL)
	  {
		strcat(cmd, " <specval>");
		eip->ei_newval = eip->ei_specval;
	  }
	else if (eip->ei_action == EI_GAUSSIAN)
	  {
		strcat(cmd, " <peakval>");
		eip->ei_newval = eip->ei_specval; /* fancier for box & circle*/
	  }
	else if (eip->ei_action == EI_ZEROOUT) eip->ei_newval = 0;
	switch (shape)
	      {
		case	BOX:
			if (eip->ei_ncargs < 4) fprintf(stderr,
				 "Syntax: %s BO xmin ymin xmax ymax\n", cmd);
			else {
				eip->ei_xmn = eip->ei_iargs[0];
				eip->ei_ymn = eip->ei_iargs[1];
				eip->ei_xmx = eip->ei_iargs[2];
				eip->ei_ymx = eip->ei_iargs[3];
				modbox(eip);
			     }
			break;
		case	CIRCLE:
			if (eip->ei_ncargs < 3) fprintf(stderr,
			     "Syntax: %s CI xcent ycent radius\n", cmd);
			else {
				eip->ei_xcent = eip->ei_iargs[0];
				eip->ei_ycent = eip->ei_iargs[1];
				eip->ei_radius = eip->ei_iargs[2];
				modcircle(eip);
			     }
			break;
		case	COLUMN:
			if (eip->ei_ncargs < 2) fprintf(stderr,
			 "Syntax: %s CO xcent msgwid\n", cmd);
			else {
				eip->ei_xcent = eip->ei_iargs[0];
				eip->ei_msgwid = eip->ei_iargs[1];
				modcolumns(eip);
			     }
			break;
		case	OBOX:
			if (eip->ei_ncargs < 4) fprintf(stderr,
			 "Syntax: %s OB xmin ymin xmax ymax\n", cmd);
			else {
				eip->ei_xmn = eip->ei_iargs[0];
				eip->ei_ymn = eip->ei_iargs[1];
				eip->ei_xmx = eip->ei_iargs[2];
				eip->ei_ymx = eip->ei_iargs[3];
				modobox(eip);
			     }
			break;
		case	OCIR:
			if (eip->ei_ncargs < 3) fprintf(stderr,
			 "Syntax: %s OC xcent ycent radius\n", cmd);
			else {
				eip->ei_xcent = eip->ei_iargs[0];
				eip->ei_ycent = eip->ei_iargs[1];
				eip->ei_radius = eip->ei_iargs[2];
				modocircle(eip);
			     }
			break;
		case	ROW:
			if (eip->ei_ncargs < 2) fprintf(stderr,
			 "Syntax: %s RO ycent eip->ei_dht\n", cmd);
			else {
				eip->ei_ycent = eip->ei_iargs[0];
				eip->ei_msgwid = eip->ei_iargs[1];
				modrows(eip);
			     }
			break;
		case	RING:
			if (eip->ei_ncargs < 3) fprintf(stderr,
			     "Syntax: %s RI xcent ycent radius\n", cmd);
			else {
				eip->ei_xcent = eip->ei_iargs[0];
				eip->ei_ycent = eip->ei_iargs[1];
				eip->ei_radius = eip->ei_iargs[2];
				modring(eip);
			     }
			break;
	     }
	if (eip->ei_fpl != stdout) logchange(eip, eip->ei_fpl);
	logchange(eip, stdout);
}

static void fixup_hdr(EIMWORK *eip)
{ /*	This updates headers preparatory to output */
	int	x, y, infmt, inswap;
	u_char	*ucp;
	u_short	*usp;
	VOFLOW	*vofp, *vofe;
	MAIL	*mp;

/* we may have changed the zmax value, so check it */
	mp = eip->ei_heads;	infmt = mp->format;	inswap = swawb(infmt);
	mp->zmax = 0;	mp->zmaxx = mp->zmaxy = 255;
	/* get the right MAR output format if needed */
	if (infmt == FMT_MAR)
		eip->ei_outtype = mp->format = mp->xdimen;
	else if (inswap == FMT_MAR)
		eip->ei_outtype = mp->format = swawb(mp->xdimen);
	if (mp->oflowcount)
	  {
		vofe = mp->overflow + mp->oflowcount;
		for (vofp = (mp->overflow); vofp < vofe; vofp++)
		   {
			if (vofp->value > mp->zmax)
			  {
				mp->zmax = vofp->value;
				mp->zmaxx = vofp->xcoord;
				mp->zmaxy = vofp->ycoord;
			   }
		   }
	  } /* end of "if (mp->oflowcount" clause */
	else {
		if (mp->data8)
		  {
			for (x = 0, ucp = mp->data8; x < eip->ei_dwid; x++)
			   for (y = 0; y < eip->ei_dht; y++, ucp++)
				if (*ucp > mp->zmax)
				  {
					mp->zmax = *ucp;
					mp->zmaxx = (u_short)x;
					mp->zmaxy = (u_short)
						(eip->ei_dht -y -1);
				  }
		  } /* end of "if (mp->data8" clause */
		else
		  {
			for (x = 0, usp = mp->data16; x < eip->ei_dwid; x++)
			   for (y = 0; y < eip->ei_dht; y++, usp++)
				if (*usp > mp->zmax)
				  {
					mp->zmax = *usp; mp->zmaxx = x;
					mp->zmaxy = eip->ei_dht - y - 1;
				  }
		  } /* end of else of "if (mp->data8" clause */
	     } /* end of else of "if (mp->oflowcount" clause */
}

static int eimargs(int ac, char *av[], EIMWORK *eip)
{ /*	Argument processor for `editimage' */
	register int	i;
	register char	*cp, **dcp;
	int		imerr;
	static char	*eitank;
	static char	*fnames[] = { "xlog", NULL};
	static char	ftypes[] = { 2, 0};

	if (ac <= 1)
	  {
		fprintf(stderr, "Usage: %s [-m] filename\n",
			xgprogname(NULL, 0));
		return EOK;
	  }
	eip->ei_outtype = FMT_PCS_8;
	if ((ac > 2) && (av[1][0] == '-'))
	  {
		eip->ei_outtype = FMT_MAR;
		eip->ei_lfname = trname(av[2]);
		printf("output will be in MAR format\n");
	  }
	else eip->ei_lfname = trname(av[1]);
#ifdef LITTLEND
	eip->ei_outtype = swawb(eip->ei_outtype);
#endif
	eip->ei_imflags = WAIT_IF_ABSENT;
	if (FRAME_OK != (imerr = open_image(eip->ei_lfname, eip->ei_imflags)))
		return imerr;
	if (NULL == eip->ei_heads)
	  {
		if (NULL == (AVOID *)(eip->ei_heads = (MAIL *)callop(1,
		 sizeof (MAIL), "heads in editimage"))) return EALLOC;
	  }
	/* open logfile and put a summary of the files there and on stdout */
	if (NULL == (eip->ei_fpl = propen("xlog"))) eip->ei_fpl = stdout;
	ftypes[0] = (eip->ei_fpl == stdout ? 0 : 2);
	header(stdout, PROCESSING, fnames, ftypes);
	if (eip->ei_fpl != stdout)
		header(eip->ei_fpl, PROCESSING, fnames, ftypes);
	if (NULL != (eitank = (char *)mallop(MAX_ARGS * CHARS_PER_ARG,
	 "eitank in eimargs"))) return EALLOC;
	if (NULL != (eip->ei_cargs = (char **)callop(MAX_ARGS, sizeof (char *),
	 "cargs in eimargs"))) return EALLOC;
	for (cp = eitank, dcp = eip->ei_cargs, i = 0; i < MAX_ARGS;
	 i++, dcp++, cp += CHARS_PER_ARG) *dcp = cp;
	if (NULL != (eip->ei_iargs = (int *)callop(MAX_ARGS, sizeof (int),
	 "iargs in eimargs"))) return EALLOC;
	return EOK;
}

static void tell_edicoms(FILE *fp, char *ofmt[])
{ /* This describes the options for editimage to the stream fp */
	int	jarg;

	fprintf(fp,
" Commands are Z[ero], A[verage] over area, SE[t] to user-specified value]\n");
	fprintf(fp,
 " F[ormat], U[nwarp], W[rite], X[dimension], Y[dimension], E[xit], Q[uit]\n");
	fprintf(fp," Legal subcommands for Z, A, or S are ");
	fprintf(fp,
	 "CI[rcle], R[ow], CO[lumn], B[ox], OC[ircle], OB[ox]\n");
	fprintf(fp," CIrcle requires 3 numeric args; Row and COlumn 2\n");
	fprintf(fp," Box requires 4 numeric args: xmn ymn xmx ymx\n");
	fprintf(fp," OC and OB affect the area outside a circle or box\n");
	fprintf(fp," Thus typical commands are:\n");
	fprintf(fp,"\tZero circle 255 260 4\n\tAverage row 214 2\n");
	fprintf(fp,
	 "\tSET command syntax: SE[t] CI[rcle] setval xcent ycent rad, etc\n");
	fprintf(fp,
	 "\tF[ormat] arg should be");
	for (jarg = 0; jarg < NSUPPORTED; jarg++)
	   fprintf(fp, " %d for %s%c", jarg, ofmt[jarg],
		(jarg == NSUPPORTED - 1) ? '\n' : ',');
	fprintf(fp,
	 "\tFormat command with no arguments returns current format\n");
	fprintf(fp,
 "\tX[dimension], Y[dimension] commands only relevant when unwarping\n");
	fprintf(fp, "U, W, E, and Q commands require no arguments\n");
}

static int eiproc(char *comd, char *ofmt[], EIMWORK *eip)
{ /*	Process a single command `comd' for editimage */
	int		argno, curtype, jarg, outindex, parg;
	static xgulong	fotype[] =
	 { FMT_PCS_8, FMT_MAR, FMT_SAXII, FMT_SAXII_16, FMT_RAXIS } ;

	if (!strncmp(comd, "st", 2) || *comd == 'q' ||
	   *comd == 'e') return EOK;
	else if (*comd == 'w') /* write without exiting */
	  {
		fixup_hdr(eip);
		output_frame(eip->ei_lfname, eip->ei_heads,
			eip->ei_outtype, 0x200);
		printf("edited image written to %s\n", eip->ei_lfname);
		return EOK;
	  }
	else if (*comd == 'u') /* un-warp the image */
		spatial(eip->ei_dwid, eip->ei_dht, 0x3000, eip->ei_heads);
	else if (*comd == 'x') /* change X dimension */
	  {
		if (strlen(eip->ei_cargs[0]) <= 0) /* query */
			printf(" Current X dimension = %d\n", eip->ei_dwid);
		else if (allnumeric(eip->ei_cargs[0]))
		  {
			jarg = atoi(eip->ei_cargs[0]);
			if ((jarg <= 0) || (jarg > MAX_DIMENSION))
			  {
				fprintf(stderr,
			 "X dimension must be positive and <= %4d\n",
				 MAX_DIMENSION);
			  }
			else {
				printf( "X dimension will be %4d\n", jarg);
				eip->ei_dwid = jarg;
			     }
		  }
	  }
	else if (*comd == 'y') /* change Y dimension */
	  {
		if (strlen(eip->ei_cargs[0]) <= 0) /* query */
			printf(" Current Y dimension = %d\n", eip->ei_dht);	
		else if (allnumeric(eip->ei_cargs[0]))
		  {
			jarg = atoi(eip->ei_cargs[0]);
			if ((jarg <= 0) || (jarg > MAX_DIMENSION))
			  {
				fprintf(stderr,
			 "Y dimension must be positive and <= %4d\n",
				 MAX_DIMENSION);
			  }
			else {
				printf(
			 "Y dimension will be %4d\n", jarg);
				eip->ei_dht = jarg;
			     }
		  }
	  }
	else if (*comd == 'f') /* query or modify format */
	  {
		outindex = 0;	/* this is the default */
#ifdef LITTLEND
		curtype = swawb(eip->ei_outtype);
#else
		curtype = eip->ei_outtype;
#endif
		if (strlen(eip->ei_cargs[0]) <= 0) /* query */
		  {
			for (jarg = 0; jarg < NSUPPORTED; jarg++)
			   if (curtype == fotype[jarg])
			     {
				outindex = jarg;	break;
			     }
			if (jarg >= NSUPPORTED) printf(
			 "Current output format is undefined\n");
			else printf("Current output format is %s\n",
				ofmt[outindex]);
		  }
		else if (allnumeric(eip->ei_cargs[0])) /* specify by # */
		  {
			jarg = atoi(eip->ei_cargs[0]);
			if ((jarg < 0) || (jarg >= NSUPPORTED))
			  {
				printf("Format %d not yet supported\n",
				 jarg);
			  }
			else {
				outindex = jarg;
				eip->ei_outtype = fotype[outindex];
				printf("Output format will be %s\n",
					ofmt[outindex]);
			     }
		  }
		else	/* specify by name */
		  {
			for (jarg = 0; jarg < NSUPPORTED; jarg++)
			   if (!strncmp(eip->ei_cargs[0], ofmt[jarg],
				strlen(ofmt[jarg])))
			     {
				outindex = jarg;
				eip->ei_outtype = fotype[outindex];
				printf("Output format will be %s\n",
					ofmt[outindex]);
				break;
			     }
			if (jarg >= NSUPPORTED)
			  printf("Format %s not yet supported\n",
				eip->ei_cargs[0]);
		  }
	  }
	eip->ei_action = NO_ACTION;
	parg = 1;
	if (*comd == 'z') eip->ei_action = EI_ZEROOUT;
	else if (*comd == 'a') eip->ei_action = EI_LOCALMEAN;
	else if (!strncmp(comd, "se", 2))
	  {
		parg = 2;	eip->ei_action = EI_USERSPECVAL;
		eip->ei_specval = atof(eip->ei_cargs[1]);
		if ((0. > eip->ei_specval) || (eip->ei_specval > 65534.))
			eip->ei_specval = 1.;
	  }
	else if (*comd == 'g')
	  {
		parg = 2;	eip->ei_action = EI_GAUSSIAN;
		eip->ei_specval = atoi(eip->ei_cargs[1]);
		if ((0. > eip->ei_specval) || (eip->ei_specval > 65534.))
			eip->ei_specval = 1.;
	  }
	else if ((*comd != 'u') && (*comd != 'w') &&
	 (*comd != 'f') && (*comd != 'x') && (*comd != 'y'))
		fprintf(stderr, "Unrecognized command: %s\n", comd);
	if (eip->ei_action != NO_ACTION)
	  {
		for (argno = parg, jarg = 0; argno < eip->ei_ncargs;
		 argno++, jarg++)
		   {
			eip->ei_iargs[jarg] = atoi(eip->ei_cargs[argno]);
			if (eip->ei_iargs[jarg] < 1 ||
			  eip->ei_iargs[jarg] > eip->ei_dwid - 1)
			  {
				fprintf(stderr,
				 "All numeric args must be 1 < n < %d\n",
				eip->ei_dwid-1);
			    fprintf(stderr,
				"Here iargs[%d] = %d from %s\n", jarg,
				eip->ei_iargs[jarg], eip->ei_cargs[argno]);
			  /* invoke garbage cmd so nothing happens */
			    strcpy(eip->ei_cargs[0], "pp");
			    break;
			  }
		  }
		if ((*comd == 'g') || !strcmp(comd, "se"))
			(eip->ei_ncargs)--;
		proc_cmd(comd, eip);
	  }
	return EOK;
}

int editimagemain(int argc, char *argv[])
{ /* Mainline for the imageediting functionality */
	int		imerr, jarg, rev;
	char		*comd;
	EIMWORK		eis;
	static char	*oformat[] =
		{ "pcs", "mar", "saxii8", "saxii16", "R-Axis" } ;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort*/
	if (EOK != (jarg = eimargs(argc, argv, &eis))) return jarg;
	if (FRAME_OK != (imerr = read_image(eis.ei_lfname, eis.ei_imflags,
	 eis.ei_heads)))
	  {
		(void)image_error(stdout, eis.ei_lfname, imerr, 0);
		if (eis.ei_fpl != stdout)
		  {
			(void)image_error(eis.ei_fpl, eis.ei_lfname, imerr, 0);
			(void)fclose(eis.ei_fpl);
		  }
		globalkill();	return EBAD;
	  }
	tell_edicoms(stderr, oformat); /* explain what can be done */
	eis.ei_dwid = (eis.ei_heads)->xdimen;
	eis.ei_dht = (eis.ei_heads)->ydimen;
	while (*(comd = get_comd(&eis)) != 'e') /* process commands */
		if (EOK != (rev = eiproc(comd, oformat, &eis))) return rev;
	/* the `q' command is seen as meaning `quit without writing' */
	if (*comd == 'q')
	  {
		if (stdout != eis.ei_fpl) (void)fclose(eis.ei_fpl);
		eis.ei_fpl = NULL;	globalkill();	return EOK;
	  }
	fixup_hdr(&eis);	/* modify header as needed */
	output_frame(eis.ei_lfname, eis.ei_heads, eis.ei_outtype, 0x200);
	if (stdout != eis.ei_fpl) (void)fclose(eis.ei_fpl);
	eis.ei_fpl = NULL;
	jarg = 1;	kill_mail(&(eis.ei_heads), &jarg);
	memzap((AVOID **)&(eis.ei_cargs[0]), "eis.ei_cargs[0] in editimage");
	memzap((AVOID **)&(eis.ei_cargs), "eis.ei_cargs in editimage");
	memzap((AVOID **)&(eis.ei_iargs), "eis.ei_iargs in editimage");
	globalkill();	return EOK;
}

/* end of editimage.c */
