/* imtools.c -- tools for manipulating and changing detector images

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	27 Jun 2007: bruhead returns on error rather than calling exit
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	11 Aug 2006: manage byteswapped formats without using
		swap-specific defines
	23 Jun 2006: include xgimage.h so we can read and write images
	17 May 2006: merge together {correlate,cvt_cad,dark_sub,
	 his_to,multimage,spa_cor,fliprot}.c
	15 Oct 2005: multimage: MIMWORK structure elts replace static globals
	 6 Oct 2005: housekeeping
	22 Jun 2005: flipraxis() declared as static
	13 May 2005: flipraxis() moved from ../xglib/imageout.c
	12 May 2005: spa_cor converted to callable routine
	11 May 2005: check return status from xgjournal
	 7 May 2005: convert histomain to callable routine
	 6 May 2005: use actual X-GEN library calls in flipraxis
	11 Apr 2005: change name from darksub.c to dark_sub.c to prevent overlap
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	 6 Sep 2004: eliminate unused routine spatialadj2(); cosmetics
	28 Aug 2004: routines used only here are declared as static;cosmetics
	18 Aug 2004: CVTCADRUN structure obviates globals
	18 Feb 2004: change kill_ftab calls so that they don't have args
	24 Jan 2004: static globals are now static locals
	21 Jan 2004: license message inserted
	 8 Jan 2004: multimage: static globals replace extern globals;
		call kill_ftab and kill_mail() at end
	 7 Jan 2004: external globals replaced with static globals or locals
	 computed from find_dimension(); kill_ftab called; callop & memzap used
	 7 Jan 2004: dark_sub: a few variables renamed
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Sep 2003: dark_sub: call xgaccess() rather than access();
		this allows us to gunzip the file on the way through
	 5 Jan 2003: reorganized source-file header
	20 Nov 2002: multimage: framename -> frname to avoid colliding with a
		global; trivial formatting of source assoc. with global vbls
	 1 Nov 2002: trivial reformatting of static global declarations
	25 Mar 2002: removed MSI attachments
	15 Jun 2000: multimage: remove globalkill call
	10 Mar 1999: initial version of multimage
	 6 Nov 1997: initial version of spa_cor.
	 1 Dec 1994: diagnostic output allows for detector (X,Y) > 999
	 5 May 1994: source file cleanup, POLYDET->XGEN;
		"xgulong" construction added
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version of bruhead
	30 Sep 1994: use floating-point stepsize
	 6 May 1994: initial MSI version of his_to.c
	 7 Dec 1993: initial X-GEN version of correlate, cvtcad, darksub
	13 Feb 1990: initial code for dark_sub.c,
		based on chunks of ../lib/image.c and ../integrate/integbg.c.
	15 Feb 1987: have this program call read_image() to do its job. 
	 4 Dec 1985: incororate parameter-file conventions into the XENGEN
		general release package and allow for PCS-format inputs.
	14 Nov 1985: initial version of cvtcad, based on `readcad'

   Purpose: manipulate detector image files. Specifically:
	correlate obtains a correlation coefficient between two detector
		images and prints the resulting value.
	cvtcad enables interconversions among detector image formats
	dark_sub allows us to subtract a dark-image from a regular image
	his_to computes a histogram of a detector image
	multimage multiplies an image by a flood-field image
	spa_cor applies a spatial correction to an image
	bruhead alters the line-termination conventions in a Bruker image
	fliprot reverses the step-sizes in an R-Axis header.
   Cross references: ../xglib/im*.c, ../utilities/{show_head,spatial}.c
*************************************************************************/
#define		NEED_UNISTD	1
#define		NEED_ERRNO	1
#include	"xgimage.h"
#define		MAXSPECIFIED	13	/* max # of files specified */
#define		UNDEFINED	(-17834)	/* garbage value */
#define		DEFAULT_DARKFRAC	(0.8)
#define		MAX_CORR	(1.6)
#define		MIN_CORR	(0.625)
#define		SPJMIN	(0.4)
#define		SPJMAX	(2.5)
	/* bits in the ccp->cc_flags variable */
#define		CVT_VERBOSE	1
#define		CVT_SUBDARK	2
#define		CVT_DELETEIN	4
/* bits in the mimflags variable */
#define	INVERT_SENSE	1
#define	SPATIAL_ADJUST	2

 typedef struct { /* structure type for passing stuff around  */
		int	cc_flags, cc_nconvert, cc_nzp, cc_over254;
		int	cc_sumi, cc_sumisq, cc_nunmatch;
		char	*cc_logn;
		xgulong	cc_lzmax, cc_otype;
		FILE	*cc_fplo;
		MAIL	*cc_heads;
		} CVTCADRUN;

 typedef struct {
		char	*mi_imname, *mi_fcname;
		int	mi_flags, mi_imwid, mi_imht;
		int	mi_newval, mi_nmbox, mi_oldval;
		double	mi_meanar;
		FILE	*mi_fpl;
		MAIL	*mi_heads;
		VSPLINE	*mi_spl;
		} MIMWORK;

extern int correlmain(int argc, char *argv[]);
extern int cvtcadmain(int argc, char *argv[]);
extern int darksubmain(int argc, char *argv[]);
extern int histomain(int argc, char *argv[]);
extern int multimagemain(int argc, char *argv[]);
extern int spacormain(int argc, char *argv[]);
extern int bruheadmain(int argc, char *argv[]);
extern int fliprmain(int argc, char *argv[]);

/************************
	This first routine computes a correlation coefficient
	between two detector images
************************/
int correlmain(int argc, char *argv[])
{ /* Mainline for calculating correlation coefficient between 2 images */
	char			*cp, fn;
	int			ix, iy, idarkfrac, locht, locwid, nmiddle;
	int			spechigh, speclow, this1, this2, totlen;
	double			corco, corcop, den1, den1p, den2, den2p, num;
	double			nump, rescale, scale, scalep, this1s, this2s;
	static int		imflags;
	static char		*framename, *framename2;
	static MAIL		head1, head2;
	static FRAMEINDEX	*lfmi;
	static int		cutlow  = 2;
	static int		cuthigh = 255;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	idarkfrac = imflags = 0;	argc--;	argv++;
	speclow = spechigh = -1;
	while ((argc) && (**argv == '-'))
	  {
		fn = argv[0][1];
		if ('a' <= fn && fn <= 'z') fn -= ('a' - 'A');
		cp = &(argv[0][2]);
		if (fn == 'S')
		  {
			imflags |= DARK_CURRENT;
			idarkfrac = 8000;
			if ('\0' != *cp) idarkfrac = 8000. * atof(cp);
		  }
		else if ((fn == 'L') && ('\0' != *cp)) speclow = atoi(cp);
		else if ((fn == 'H') && ('\0' != *cp)) spechigh = atoi(cp);
		argc--;	argv++;
	  }
	if (idarkfrac > 16000) idarkfrac = 16000;
	if (imflags & DARK_CURRENT) imflags |= (idarkfrac << 16);
	if ((0 <= speclow) && (speclow <= 32767)) cutlow = speclow;
	if ((0 <= spechigh) && (spechigh <= 32767)) cuthigh = spechigh;
	if (argc < 2)
	  {
		fprintf(stderr,
		 "Usage:\tcorrelate [-s<val>] framename framename2\n");
		fprintf(stderr,
		 "  or:\tcorrelate [-s<val>] framenumber1 framenumber2\n");
	  }
	if ((allnumeric(argv[0]) || allnumeric(argv[1])) &&
		qerror((AVOID *)(lfmi = ftables(3)),
			"read frame-index file")) return EOPENERR;
	if ((allnumeric(argv[0])) &&
	  qerror((AVOID *)(framename = indexname(atoi(argv[0]))),
		"find first requested frame file")) return EOPENERR;
	else framename = argv[0];
	if (verror(!read_image(framename, imflags, &head1),
		 "read first frame requested")) return EREADERR;
	if (allnumeric(argv[1]) &&
	  qerror((AVOID *)(framename2 = indexname(atoi(argv[1]))),
		"find second requested frame file")) return EOPENERR;
	else framename2 = argv[1];
	if (verror(!read_image(framename2, imflags, &head2),
		 "read second frame requested")) return EREADERR;
	totlen = strlen(framename) + strlen(framename2);
	printf("Determining correlation between frames%s%s and%s%s\n",
		(totlen < 50 ? " " : "\n\t"), framename,
		(totlen < 70 ? " " : "\n\t"), framename2);
	locwid = head1.xdimen;	locht = head1.ydimen;
	rescale = sqrt((double)(locwid * locht));
	if (verror(((locwid == (int)head2.xdimen) &&
	 (locht == (int)(head2.ydimen))),
	 "correlate the two frames: their dimensions aren't equal\n"))
		return EBADVALUE;
	printf("Image locwid and locht: %d,%d\n", locwid, locht);
	den1p = den2p = nump = nmiddle = 0;
	for (num = den1 = den2 = iy = 0; iy < locht; iy++)
	   {
		for (ix = 0; ix < locwid; ix++)
		   {
			this1 = imvalue(&head1, ix, iy);
			this2 = imvalue(&head2, ix, iy);
			this1s = ((double)this1) / rescale;
			this2s = ((double)this2) / rescale;
			if ((this1 >= cutlow) && (this2 >= cutlow) &&
			 (this1 <= cuthigh) && (this2 <= cuthigh))
			  {
				nump += this1s * this2s;
				den1p += this1s * this1s;
				den2p += this2s * this2s;
				nmiddle++;
			  }
			num += this1s * this2s;
			den1 += this1s * this1s; den2 += this2s * this2s;
		   }
	   }
	if (verror(((den1 > DFUZZ) && (den2 > DFUZZ)),
	 "determine correlation: one or both images have no intensity"))
		return EBADVALUE;
	corco = num / sqrt(den1 * den2);
	printf("Correlation coefficient = %9.6f\n", corco);
	if (den2 < den1)
	  {
		scale = num / den2; printf(
 "Scale factor minimizing SUM (f2 - k*f1)^2 is k = %9.4f\n", scale);
	  }
	else {
		scale = num / den1; printf(
 "Scale factor minimizing SUM (f1 - k*f2)^2 is k = %9.4f\n", scale);
	     }
	if ((nmiddle > 1) && (den1p > DFUZZ) && (den2p > DFUZZ))
	  {
		corcop = nump / sqrt(den1p * den2p); printf(
 "Correlation coefficient for %6d pixels with %5d <= count <= %5d = %8.5f\n",
			nmiddle, cutlow, cuthigh, corcop);
		printf("Scale factor for those pixels, ");
		if (den2 < den1)
		  {
			scalep = nump / den2p; printf(
		 "minimizing SUM (f2 - k*f1)^2 is k = %9.4f\n", scalep);
		  }
		else {
			scalep = nump / den1p; printf(
		 "minimizing SUM (f1 - k*f2)^2 is k = %9.4f\n", scalep);
		     }
	  }
	(void)ftables(0);	globalkill();	return EOK;
}

/* end of correlate functionality */

/*********************
	The next few routines implement interconversions among
	various specific image file formats.
	Input filenames come from either the command line or the FRAMES file.
**********************/
static int alldigit(char *s)
{
	register char	*cp;

	for (cp = s; '\0' != *cp; cp++) if (*cp < '0' || *cp > '9') return 0;
	return 1; /* returns true if and only if s is all digits */
}

static void vfy_frame(CVTCADRUN *ccp)
{ /*	This verifies the overflows and mailbox values
	by going through the data.  */
	register int	x, y;
	int		hit, wid;
	u_short		currval, lzmaxx, lzmaxy;
	double		imagesize, xbar, xvar;
	MAIL		*lhd;

	lhd = ccp->cc_heads;	hit = lhd->ydimen;	wid = lhd->xdimen;
	lzmaxx = wid / 2;	lzmaxy = hit / 2;
	for (ccp->cc_lzmax = y = 0; y < hit; y++)
	   for (x = 0; x < wid; x++)
	      {
		if ((currval = imvalue(lhd, x, y)) > 254) ccp->cc_over254++;
		ccp->cc_sumi += currval;
		ccp->cc_sumisq += (currval * currval);
		if (currval)
		  {
			ccp->cc_nzp++; if (currval > ccp->cc_lzmax)
			  {
				ccp->cc_lzmax = currval;
				lzmaxx = x; lzmaxy = y;
			  }
		  }
	     }
	imagesize = (double)(hit * wid);
	xbar = (double)ccp->cc_sumi / imagesize;
	xvar = (double)ccp->cc_sumisq / imagesize;
	fprintf(ccp->cc_fplo, "%9d%10d%15d%7d%6d%4d%4d%8.4f%11.4f\n", 
	 (int)(lhd->frameno), ccp->cc_nunmatch,
	 (int)lhd->over64k, (int)lhd->nonzpix,
	 (int)lhd->zmax, (int)lhd->zmaxx,
	 (int)lhd->zmaxy, xbar, sqrt(xvar - xbar * xbar));
/* Correct the mailbox based on what we've just found */
	lhd->nonzpix = ccp->cc_nzp;
	lhd->zmax = ccp->cc_lzmax;
	lhd->zmaxx = lzmaxx;
	lhd->zmaxy = lzmaxy;
	if (lhd->over64k) fprintf(ccp->cc_fplo,
	 "\t\tFrame suspect: %d %s had > 64k counts\n", (int)lhd->over64k,
	 (lhd->over64k == 1 ? "pixel" : "pixels"));
}

static void setup_log(CVTCADRUN *ccp)
{	/* this creates and writes a header to the logfile */
	MAIL	*lhd;

	lhd = ccp->cc_heads;
	if (NULL == (ccp->cc_logn = (char *)callop(32, sizeof (char),
	 "ccp->cc_logn in setup_log")))
	  {
		fprintf(stderr, "Log information will go to stdout\n");
		ccp->cc_fplo = stdout;
	  }
	else
	  {
		strncpy(ccp->cc_logn,lhd->filename,
			(strlen(lhd->filename)-2));
		strcat(ccp->cc_logn,"log");
		if (NULL == (ccp->cc_fplo = fopen(ccp->cc_logn,"w")))
		  {
			fprintf(stderr,
 			 "Error opening %s as logfile\n",ccp->cc_logn);
			fprintf(stderr,
			 "Log information will come to the standard output\n");
			ccp->cc_fplo = stdout;
		  }
	   }
	/* put a header on logfile and stdout */
	if (ccp->cc_flags & CVT_VERBOSE)
	  {
		fprintf(ccp->cc_fplo,
"  Frame # lostoflos Mbox: Pix >64k    > 0  Zmax Zmx Zmy cts/pix rms(ct/px)\n");
		fprintf(ccp->cc_fplo,
"    Secs  Ontime    Late Over  Pix Pixels    Max Count:    Flux Output File\n");
		fprintf(ccp->cc_fplo,
"   Req'd  Counts  Counts flow >254     >0   Val   X   Y     cps Size   Name\n");
	  }
	else {
		fprintf(ccp->cc_fplo,
"    Secs  Ontime    Late Over Pixels    Max Count:    Flux Output File\n");
		fprintf(ccp->cc_fplo,
"   Req'd  Counts  Counts flow     >0   Val   X   Y     cps Size   Name\n");
	  }
	fprintf(stdout,
"    Secs  Ontime    Late Over Pixels    Max Count:    Flux Output  File\n");
	fprintf(stdout,
"   Req'd  Counts  Counts flow     >0   Val   X   Y     cps  Size   Name\n");
}

static void log_frame(FILE *fp, char *name, CVTCADRUN *ccp)
{	/* this summarizes frame `name' to the stream `fp'. */
	int		nblocks;
	double		flux, realsec, revent;
	MAIL		*lhd;
	static char	*shortname;

	lhd = ccp->cc_heads;
	revent = lhd->eventcount;	if (revent < 0.) revent *= -64.;
	realsec = 0.01 * (double)(lhd->elap100ths);
	flux = (realsec > DFUZZ) ? revent / realsec : 0;
	nblocks = 516 + 4 * (1 + (lhd->oflowcount * 3) / 1024);
	fprintf(fp, " %4.0f%s%d%8d%8d%5d", realsec,
	 ((lhd->elap100ths % 100) < 10 ? ".0" : "."),
	 (int)lhd->elap100ths % 100, (int)lhd->ontime,
	 (int)lhd->late, (int)lhd->oflowcount);
	if ((fp == ccp->cc_fplo) && (ccp->cc_flags & CVT_VERBOSE))
		fprintf(fp, "%5d", ccp->cc_over254);
	shortname = filepart(name, 0);
	fprintf(fp, "%7d%6d%4d%4d %7.1f %4d %s\n",
	 (int)lhd->nonzpix, (int)lhd->zmax, (int)lhd->zmaxx, (int)lhd->zmaxy,
	 flux, nblocks, shortname);
}

static void proc_frame(char *name, int flags, CVTCADRUN *ccp)
{ /*	This reads in the header to the image `name' and (usually) the
	remainder of the data.  If it's an already-converted image or there's
	an error in reading the image, it prints a message and continues;
	otherwise it converts the file to the FMT_VMS format and writes it out.
	Each uccp->cc_nconvert frame is deleted before the converted frame
	 is written,
	so that the converted file will have the same version number. */
	int	image_status, outstat;
	MAIL	*lhd;

	ccp->cc_nzp = ccp->cc_over254 = ccp->cc_lzmax = 0L;	/*zero stats */
	ccp->cc_sumi = ccp->cc_sumisq = ccp->cc_nunmatch = 0;
	lhd = ccp->cc_heads;
	if (FRAME_OK == (image_status = read_image(name, flags, lhd)))
	  {
		(ccp->cc_nconvert)++; /* increment count of converted frames */
		if (NULL == ccp->cc_fplo) setup_log(ccp); /* set up log file */
			/* if the user wants the image verified, do so */
		if (ccp->cc_flags & CVT_VERBOSE) vfy_frame(ccp);
			/* delete input unless user said not to */
		if (ccp->cc_flags & CVT_DELETEIN)
		  if (-1 == xgdelete(name))
		    {
			fprintf(stderr,
			 "\tSystem was unable to delete input file %s\n",name);
			fprintf(ccp->cc_fplo,
			 "\tSystem was unable to delete input file %s\n",name);
		    }
		lhd->format = ccp->cc_otype;
		lhd->runheader = NULL;	/* to play it safe */
		if (FRAME_OK == (outstat = output_frame(name, lhd,
		 ccp->cc_otype, 0x200)))
		  {
			log_frame(ccp->cc_fplo, name, ccp);
			log_frame(stdout, name, ccp);
		  }
		else {
			(void)image_error(ccp->cc_fplo, name, outstat, 2);
			(void)image_error(stdout,name, outstat, 2);
		     }
	  }
	else	/* write appropriate error messages if no conversion */
	  {
		(void)image_error(ccp->cc_fplo, name, image_status, 0);
		(void)image_error(stdout, name, image_status, 0);
		/* if we haven't done any frames at all, zero the pointer
		to the overflows so that image.c doesn't get confused. */
		lhd->overflow = NULL;
	  }
}

int cvtcadmain(int argc, char *argv[])
{ /* mainline for converting into and out of CADMUS format */
	register int		frame;
	int			flags, framemax, framemin, nnames;
	unsigned		nfiles;
	char			**namelist, **rflags, **dcp;
	static int		nmbox;
	static FRAMEINDEX	*lfmi;
	static CVTCADRUN	ccr;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	ccr.cc_flags = CVT_DELETEIN;	ccr.cc_otype = FMT_VMS;	
	rflags = flagproc(argc, argv, "DPSV");
	for (dcp = rflags; NULL != dcp && NULL != *dcp; dcp++)
	   switch (**dcp)
	     {
		case	'D':	/* don't delete input files */
			ccr.cc_flags &= ~CVT_DELETEIN;	break;
		case	'P':	/* PCS-format output */
#ifdef LITTLEND
			ccr.cc_otype = swawb(FMT_PCS_8);	break;
#else
			ccr.cc_otype = FMT_PCS_8;	break;
#endif
		case	'S':	/* subtract dark-current image */
			ccr.cc_flags |= CVT_SUBDARK;	break;
		case	'V':	/* ccp->cc_flags logfile */
			ccr.cc_flags |= CVT_VERBOSE;	break;
	     }
	for (argc--, argv++; (argc) && **argv == '-'; argc--) argv++;
	framemin = framemax = UNDEFINED;
	if (NULL == ccr.cc_heads)
	  {
		nmbox = 1;
		if (NULL == (AVOID *)(ccr.cc_heads = (MAIL *)callop(nmbox,
		 sizeof (MAIL), "ccr.cc_heads in cvtcadmain"))) return EALLOC;
	  }
	if (NULL == (AVOID *)(namelist = (char **)callop(MAXSPECIFIED,
	 sizeof (char *), "namelist in cvtcadmain")))
	  {
		kill_mail(&ccr.cc_heads, &nmbox);	return EALLOC;
	  }
	for (nnames = 0; argc--; argv++)
	   {
		if (alldigit(*argv))
		  {
			if (framemin != UNDEFINED) framemax = atoi(*argv);
			else	framemin = atoi(*argv);
		  }
		else	namelist[nnames++] = *argv;
	   }
	if (nnames)
	  {
		fprintf(stdout," %3d frame%sspecified:\n",
		 nnames, (nnames == 1 ? " " : "s "));
		for (nfiles = 0; nfiles < (unsigned)nnames; nfiles++)
		   {
			fputs(namelist[nfiles], stdout);
			fputs(" ", stdout);
		   }
		fputs("\n", stdout);
	  }
	lfmi = ftables(3);
	if ((NULL == lfmi) && !nnames)
	  {
		fprintf(stderr,
	"You must either specify file names or create a frame table. Sorry\n");
		return EBAD;
	  }
	if (!nnames && framemin == UNDEFINED) framemin = lfmi->fi_firstfm;
	if (!nnames && framemax == UNDEFINED) framemax = lfmi->fi_lastfm;
	ccr.cc_nconvert = 0;
	if (ccr.cc_otype == FMT_VMS)
	  {
		flags = RETURN_IF_CONVERTED;
		fprintf(stdout,
	   "Files will be output in XGEN standard (FMT_VMS) format\n");
	  }
	else {
		fprintf(stdout, "Files will be output in PCS format\n");
		flags = 0;
	     }
	if (ccr.cc_flags & CVT_SUBDARK) flags |= DARK_CURRENT;
	if (framemin != UNDEFINED)
	  for (frame = framemin; frame <= framemax; frame++)
		proc_frame(indexname(frame), flags, &ccr);
	for (nfiles = 0; nfiles < (unsigned)nnames; nfiles++)
		proc_frame(namelist[nfiles], flags, &ccr);
	fprintf(ccr.cc_fplo, "\t\t%6d file%sread and converted\n",
		ccr.cc_nconvert, (ccr.cc_nconvert == 1 ? " " : "s "));
	fprintf(stdout,"\t\t%6d file%sread and converted\n", ccr.cc_nconvert,
		(ccr.cc_nconvert == 1 ? " " : "s "));
	if (stdout != ccr.cc_fplo) (void)fclose(ccr.cc_fplo);	globalkill();
	memzap((AVOID **)&ccr.cc_logn, "ccr.cc_logn in cvtcad");
	memzap((AVOID **)&namelist, "namelist in cvtcad");
	kill_mail(&ccr.cc_heads, &nmbox);	globalkill();	return EOK;
}

/* end of cvt_cad functionality */

/***********************
	These routines enable subtraction of a dark-current
	image from a normal image. We keep all images
	in 16-bit format throughout (to save on bookeeping).

	The purpose of this program is to deal with the nasty case of a frame
	in which most of the pixels are overflowed, but for which the dark
	image also has most of the pixels overflowed at almost the same level
	so that the difference image is almost perfectly zero.
	Syntax:
 		darksub [-p] [-d] [-s<val>]
	The image read in and written out has logical name IMAGE;
	the image subtracted from it has logical name DARKIMAGE.
	The output file is written out in normal FMT_PCS_8 (or its swapped
	equivalent) format, depending on the user's desire.  Under VMS,
	the frame is output in the swapped version of FMT_PCS_8 format
	unless the "-p" flag is invoked, in which case FMT_PCS_8 format is used.
	Under Big-end Unix, the frame is output in FMT_PCS unless the
	"-v" flag is invoked, in which case FMT_VMS is used.
	We actually subtract 0.8 times the value in DARKIMAGE
	unless the user asks for a different fraction with -d<val>.
	Normally `darksub' writes a new version of the file `primaryname',
	but if it is invoked with "-d" the old version is overwritten.
	Thus
 		darksub -ps1
	will subtract 100% (-s1) of the values in DARKIMAGE from
	the corresponding values in image.0001.
	As in image.c, the difference is set to zero if it comes out negative.
	The inputs MUST be in FMT_CCD, FMT_CCD_ALT, FMT_CCD_VMS, or
	FMT_CCD_VMSALT format.
*************************************************************************/
#ifndef VMS
static void altname(char *tname)
{ /*	This comes up with an alternate name for the file whose current name
	is `tname'.  It does nothing if `tname' doesn't currently exist. */
	int		accreturn, bad, try;
	char		*augname, *lastc, *prevc;
	unsigned	namelen;

	if (0 == xgaccess(tname, 0, 1)) /* does this file exist already? */
	  {
		namelen = strlen(tname); /* yes.  See how long the name is. */
			/* save some memory for the augmented name */
		if (NULL == (augname = (char *)mallop(namelen + 4,
			"augname in altname"))) return;
		strcpy(augname, tname);	/* copy name to augmented-name */
		strcat(augname, ",0"); /* append ",0" to the name */
		augname[namelen+3] = '\0'; /* extra null at the end */
		 /* see if the augmented name exists.  If it does,
		 try another version... if "name,0" exists, try "name,1";
		  if "name,1" exists, try "name,2" ... ; if "name,9" exists,
		  try "name,10" through "name,99".  If "name,99" already
		  exists, just overwrite it; you have too many versions anyway.
		 */
		lastc = augname + (namelen + 1); /* position of last char */
		accreturn = 0;
		for (try = 0; (try < 100) &&
		 (0 == (accreturn = xgaccess(augname, 0, 1))); try++)
		   {
			if ('9' == *lastc) /* is the last char a 9? */
			  {
				prevc = lastc - 1; /*yes. Note previous char*/
				if ('9' == *prevc) break; /* give up if "99"*/
				else if (',' == *prevc) /* blah,9 -> blah,10*/
				  {
					*lastc++ = '1';	*lastc = '0';
				  }
				/* blah,39 -> blah,40 */
				else {
					(*prevc)++;	*lastc = '0';
				     }
			  }
			else	(*lastc)++; /* blah,6 -> blah,7 */
		   }
		bad = 0;
		/* if we get here and the variable "accreturn" is still true,
		we unlink the current augmented name. */
		if ((0 == accreturn) && (-1 == (unlink(augname)))) bad = 1;
		/* now we link the input filename to the augmented filename */
		if (-1 == link(tname, augname)) bad = 1;
		if (-1 == unlink(tname)) bad = 1; /* unlink input filename */
			/* free up memory we saved */
		memzap((AVOID **)&augname, "augname in altname");
		if (bad) return;	/* don't try to open if error */
	  }
	return;
}
#endif

int darksubmain(int argc, char *argv[])
{ /* In this mainline, we handle the arguments and then process the images.*/
	int		deleteold, idfrac, imflags, ix;
	xgulong		outformat;
	char		*dname, *iname, *oname, **dcp, **pflags;
	double		dfrac, revent, rlate;
	MAIL		imhead;
	PAU		prd8, provf, prov16;
	static DATARUN	*runhdp;
	static char	*fnames[] = {"IMAGE", "DARKIMAGE", "IMAGEOUT", NULL } ;
	static char	ftypes[] = { 1, 1, 2, 0 } ;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	deleteold = 0;
	/* establish default output format */
	outformat = FMT_PCS_8;
#ifdef LITTLEND
	outformat = swawb(outformat);
#endif
	dfrac = DEFAULT_DARKFRAC;
	pflags = flagproc(argc, argv, "DPSV");
	for (dcp = pflags; NULL != dcp && NULL != *dcp; dcp++)
	   switch (**dcp)
	     {
		case	'D':	deleteold = 1;			break;
		case	'P':	outformat = FMT_PCS_8;		break;
		case	'S':	dfrac = atof((*dcp) + 1);	break;
		case	'V':	outformat = FMT_VMS;		break;
		default:	break;
	     }
	header(stdout, PROCESSING, fnames, ftypes);
	if (deleteold) fprintf(stdout,
		"Input image will be deleted during conversion\n");
	fprintf(stdout, "Output image will be in %s format\n",
		(outformat == FMT_VMS ? "VMS" : "PCS"));
	fprintf(stdout, "Resultant values = input - %5.3f * darkimage\n",
		dfrac);
/* make sure both images are readable; otherwise we're wasting our time*/
	dname = trname("DARKIMAGE");	iname = trname("IMAGE");
	if (verror(!xgaccess(dname, 4, 1), "open dark image")) return EOPENERR;
	if (verror(!xgaccess(iname, 4, 1), "open image")) return EOPENERR;
/* read in dark and regular images and subtract the dark current. */
	zeromem((AVOID *)&imhead, sizeof (MAIL));
	idfrac = (int)(dfrac * 8000. + 0.49999);
	imflags = DARK_CURRENT | GRIPE_IF_SHORT |
		GET_RUNHEADER | (idfrac << 16);
	if (FRAME_OK != (ix = read_image(iname, imflags, &imhead)))
	  {
		image_error(stdout, iname, ix, 0); return EREADERR;
	  }
	if (deleteold) (void)xgdelete(iname);
#ifdef VMS
	oname = "IMAGEOUT";
#else
	oname = trname("IMAGEOUT");
	altname(oname);
#endif
	prd8.ucpt = imhead.data8;
	provf.vpt = imhead.overflow;	prov16.vpt = imhead.oflow16;
	printf("data begin at 0x%x\n", prd8.ui);
	printf("8-bit and 16-bit overflows begin at 0x%8x, 0x%8x\n",
	 provf.ui, prov16.ui);
	printf(
	 "format, xdimension, ydimension, quadrant = %6d, %6d, %6d, %6d\n",
		(int)imhead.format, (int)imhead.xdimen,
		(int)imhead.ydimen, (int)imhead.quadrant);
	revent = imhead.eventcount;	if (revent < 0.) revent *= -64.;
	rlate = imhead.late;	if (rlate < 0.) rlate *= -64.;
	printf(
	 "events, #overflows = %9.0f, %4d; max = %5d at [%4d,%4d]\n",
		revent, (int)imhead.oflowcount,
		(int)imhead.zmax, (int)imhead.zmaxx, (int)imhead.zmaxy);
	printf(
 "zdisp, ontime, late, accumsecs, elap100ths = %4d,%8d,%7.0f,%6d,%6d\n",
		(int)imhead.zdisp, (int)imhead.ontime, rlate,
		(int)imhead.accumsecs, (int)imhead.elap100ths);
	printf("datetime[0], datetime[1] = 0x%x, 0x%x\n",
		(int)imhead.datetime[0], (int)imhead.datetime[1]);
	printf("no. of pixels > 0, >254, > 64k = %8d, %5d, %2d\n",
	 (int)imhead.nonzpix, (int)imhead.oflowcount,(int)imhead.over64k);
	printf("frameno = %5d; filename = %s\n",
		(int)(imhead.frameno), imhead.filename);
	if ((NULL != (runhdp = imhead.runheader)) &&
		(FUZZ < runhdp->run_d0pos[0]))
	  {
		printf("Comments on run: %s\n", runhdp->run_notes);
		printf("Generator at %3d kV, %3d ma\n",
		runhdp->run_volxray, runhdp->run_curxray);
		if (fabs(runhdp->run_stepreal) > FUZZ)
		 printf("Stepsize: %7.3f degrees\n", runhdp->run_stepreal);
		else printf("Stepsize: %3d / 96 degrees\n",
		runhdp->run_stepsize);
		printf("Number of frames in run: %3d\n",
		runhdp->run_nframes);
		printf("Approx. # of scaling groups: %3d\n",
		runhdp->run_nshifts);
		printf("Start omega, chi, phi: %7.2f, %7.2f, %7.2f\n",
		 runhdp->run_startpos[0], runhdp->run_startpos[1],
		 runhdp->run_startpos[2]);
		printf("Exposure time: %7.2f seconds\n",
		 runhdp->run_expos);
		printf(
		 "Crystal-to-detector distance, angle = %9.4f, %7.2f\n",
		 runhdp->run_d0pos[0], runhdp->run_d0ang[0]);
		printf("Resolution limit, ssqmax = %8.3f, %9.5f\n",
		 runhdp->run_reslim, runhdp->run_ssqmax);
	  }
	output_frame(oname, &imhead, outformat, 0x200);
	globalkill();	return EOK;
}

/* end of dark_sub functionality */
/******************
	This routine computes a histogram of the raw counts in an image.
*******************/
int histomain(int argc, char *argv[])
{ /* This is the mainline for generating a raw-count histogram for an image */
	register int	i, ix;
	int		iu, maxc, minc, readflags, readret, thisv, x, y;
	static int	hist[20];
	static MAIL	*heads;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	if (argc < 2)
	  {
		fprintf(stderr,"usage: histo filename\n"); return EOK;
	  }
	readflags = GET_RUNHEADER;
	if (NULL == heads)
	  {
		if (NULL == (AVOID *)(heads = (MAIL *)callop(1, sizeof (MAIL),
		 "heads in histo"))) return EALLOC;
	  }
	for (i = 1; i < argc; i++)
	   {
		if (FRAME_OK != (readret =
		  read_image(argv[i], readflags, heads)))
		  {
			(void)image_error(stdout, argv[i], readret, 0);
			continue;
		  }
		for (y = 0; y < (int)(heads->ydimen); y++)
		   {
			for (x = 0; x < (int)(heads->xdimen); x++)
			   {
				thisv = imvalue(heads, x, y);
				if (thisv < 0) ix = 0;
				else if (thisv < 3) ix = thisv;
				else {
					for (iu = 4, ix = 3; ix <= 19;
					 ix++, iu *= 2) if (thisv <=iu) break;
					if (ix > 18) ix = 19;
				     }
				(hist[ix])++;
			   }
		   }
		printf("Number of pixels with observed counts:\n");
		printf("Min count Max count  # pixels\n");
		for (ix = 0, iu = 1; ix < 20; ix++, iu *= 2)
		   {
			if (ix < 3) minc = maxc = ix;
			else {
				maxc = iu / 2;	minc = (maxc / 2) + 1; }
			printf("%9d %9d  %8d\n", minc, maxc,hist[ix]);
			hist[ix] = 0;
		   }
	   }
	x = 1;	kill_mail(&heads, &x);
	globalkill();	return EOK;
}

/* end of his_to functionality */

/*******************************
	This multiplies an image by a FLOOD image to flatten local
	fluctuations in the detector response.
	The result is normalized so that the total count is unchanged.
*********************************/
static void pixelmod8(int x, int y, MIMWORK *mip)
{ /*	This alters the value at the point (x, y) to the value
	 `mip->mi_newval' in an 8-bit image.
	 The tricky part is getting the overflow table right. */
	u_char	*ucp;
	VOFLOW	*vofe, *vofn, *vofp;

	/* set the new 8-bit value properly */
	ucp = mip->mi_heads->data8 + x * mip->mi_imht + y;
	*ucp = (mip->mi_newval < 256) ? mip->mi_newval : 255;
	if (mip->mi_newval < mip->mi_oldval)
	  mip->mi_heads->eventcount -= (mip->mi_oldval - mip->mi_newval);
	else mip->mi_heads->eventcount += mip->mi_newval - mip->mi_oldval;
	if (mip->mi_newval < 256 && mip->mi_oldval < 256) return;
	/* Fix up the overflow table if necessary */
	if (mip->mi_heads->oflowcount)
	  {
		vofe = mip->mi_heads->overflow + mip->mi_heads->oflowcount;
		for (vofp = mip->mi_heads->overflow; vofp < vofe; vofp++)
		   {
			if ((vofp->xcoord == x) &&
			 (vofp->ycoord == (mip->mi_imht - y - 1)))
			  {
				if (mip->mi_newval > 255) /* change overflow */
				  {
					vofp->value = mip->mi_newval;	return;
				  }
				else	/* get rid of this overflow */
				  {
					if (mip->mi_heads->oflowcount == 1)
					  {
						mip->mi_heads->oflowcount = 0;
						return;
					  }
					for (vofn = vofp + 1;
					 vofn < vofe; vofn++,vofp++)
					   cpymem((AVOID *)vofp, (AVOID *)vofn,
						sizeof (VOFLOW));
					mip->mi_heads->oflowcount--;	return;
				  } /* end of else (mip->mi_newval < 255) */
			  } /* end of if overflow match */
		   } /* end of loop through current overflows */
		/* if we get here, the current pixel wasn't overflowed before*/
		if (mip->mi_newval < 256) return; /* no sweat */
		(mip->mi_heads->oflowcount)++;	vofp = vofe;
		vofp->xcoord = x;	vofp->ycoord = mip->mi_imht - y - 1;
		vofp->value = mip->mi_newval;	vofe++;	return;
	  } /* end of if clause */
	/* if we didn't have any overflows before and the new value
	is > 255, we will have to create an overflow table.  I'm assuming
	we've created room for at least 800 overflows (see image.c) so that
	we're just defining the space now */
	if (mip->mi_newval < 256) return;	/* phew */
	vofp = mip->mi_heads->overflow;
	vofp->value = mip->mi_newval;
	vofp->xcoord = x;	vofp->ycoord = mip->mi_imht - y - 1;
	mip->mi_heads->oflowcount = 1;	vofe = mip->mi_heads->overflow + 1;
}

static void pixelmod16(int x, int y, MIMWORK *mip)
{ /*	This alters the value at the point (x, y) to the value
	 `mip->mi_newval' in a 16-bit image.
	 The tricky part is getting the overflow table right. */
	u_short	keptval, *usp;
	VOFLOW	*vofe, *vofn, *vofp;

	/* set the new 16-bit value properly */
	usp = mip->mi_heads->data16 + x * mip->mi_imht + y;
#ifdef LITTLEND
	*usp = (mip->mi_newval < 65535) ? uswab(mip->mi_newval) : 65535;
#else
	*usp = (mip->mi_newval < 65535) ? mip->mi_newval : 65535;
#endif
	if (mip->mi_newval < mip->mi_oldval) mip->mi_heads->eventcount -= (mip->mi_oldval - mip->mi_newval);
	else mip->mi_heads->eventcount += mip->mi_newval - mip->mi_oldval;
	if (mip->mi_newval < 65535 && mip->mi_oldval < 65535) return;
	/* Fix up the overflow table if necessary */
	if (mip->mi_heads->oflowcount)
	  {
		vofe = mip->mi_heads->overflow + mip->mi_heads->oflowcount;
		for (vofp = mip->mi_heads->overflow; vofp < vofe; vofp++)
		   {
			if ((vofp->xcoord == x) &&
			 (vofp->ycoord == (mip->mi_imht - y - 1)))
			  {
				if (mip->mi_newval >= 65535) /*alter overflow*/
				  {
					keptval = (mip->mi_newval - 65535) / 4;
					vofp->value = keptval;
					return;
				  }
				else	/* get rid of this overflow */
				  {
					if (mip->mi_heads->oflowcount == 1)
					  {
						mip->mi_heads->oflowcount = 0;
						return;
					  }
					for (vofn = vofp + 1; vofn < vofe;
					 vofn++,vofp++) cpymem((AVOID *)vofp,
					 (AVOID *)vofn, sizeof (VOFLOW));
					mip->mi_heads->oflowcount--;	return;
				  } /* end of else (mip->mi_newval <= 65535) */
			  } /* end of if overflow match */
		   } /* end of loop through current overflows */
		/* if we get here, the current pixel wasn't overflowed before*/
		if (mip->mi_newval <= 65535) return; /* no sweat */
		vofp = vofe;
		vofp->xcoord = x;	vofp->ycoord = mip->mi_imht - y - 1;
		keptval = (mip->mi_newval - 65535) / 4;	vofp->value = keptval;
		(mip->mi_heads->oflowcount)++;	vofe++;	return;
	  } /* end of if clause */
	/* if we didn't have any overflows before and the new value
	is > 65535, we will have to create an overflow table.  I'm assuming
	we've created room for at least 800 overflows (see image.c) so that
	we're just defining the space now */
	if (mip->mi_newval <= 65535) return;	/* phew */
	vofp = mip->mi_heads->overflow;
	keptval = (mip->mi_newval - 65535) / 4;	vofp->value = keptval;
	vofp->xcoord = x;	vofp->ycoord = mip->mi_imht - y - 1;
	mip->mi_heads->oflowcount = 1;	vofe = mip->mi_heads->overflow + 1;
}

static void mfix_hdr(MIMWORK *mip)
{ /*	This updates headers preparatory to output */
	int	x, y;
	xgulong	realval;
	u_char	*ucp;
	u_short	*usp;
	MAIL	*mp;
	VOFLOW	*vofp, *vofe;

/* we may have changed the zmax value, so check it */
	mp = mip->mi_heads;
	mp->zmax = 0;	mp->zmaxx = mip->mi_imwid / 2;
	mp->zmaxy = mip->mi_imht / 2;
	if (mp->oflowcount)
	  {
		vofe = mp->overflow + mp->oflowcount;
		for (vofp = (mp->overflow); vofp < vofe; vofp++)
		   {
			realval = (mp->data16) ?
				4 * ((int)(vofp->value)) + 65535 : vofp->value;
			if (realval > mp->zmax)
			  {
				mp->zmax = realval; mp->zmaxx = vofp->xcoord;
				mp->zmaxy = vofp->ycoord;
			   }
		   }
	  } /* end of "if (mp->oflowcount" clause */
	else {
		if (mp->data8)
		  {
			for (x = 0, ucp = mp->data8; x < mip->mi_imwid; x++)
			   for (y = 0; y < mip->mi_imht; y++, ucp++)
				if (*ucp > mp->zmax)
				  {
					mp->zmaxx = (u_short)x;
					mp->zmaxy = (u_short)y;
				  }
		  } /* end of "if (mp->data8" clause */
		else
		  {
			for (x = 0, usp = mp->data16; x < mip->mi_imwid; x++)
			   for (y = 0; y < mip->mi_imht; y++, usp++)
			      {
				if ((*usp) > mp->zmax)
				  {
					mp->zmax = (*usp);
					mp->zmaxx = (u_short)x;
					mp->zmaxy = (u_short)y;
				  }
			     }
		  } /* end of else of "if (mp->data8" clause */
	     } /* end of else of "if (mp->oflowcount" clause */
}

static void log_header(FILE *fp, double *cop, MIMWORK *mip)
{ /*	Print a header to the log output */
	fprintf(fp, "Summary of flood-image %s:\n", mip->mi_fcname);
	if (mip->mi_flags & SPATIAL_ADJUST)
	  {
		fprintf(fp,
	 "spatial correction minimum and maximum values: %9.4f, %9.4f\n",
		cop[6], cop[7]);
		fprintf(fp,
 "spatial corrections outside (%8.4f,%8.4f) set to appropriate extrema\n",
		 SPJMIN, SPJMAX);
	  }
	fprintf(fp,
 "   # of pixels          nonzero values           scaled values\n");
	fprintf(fp,
 " identically 0   minimum   maximum      mean   minimum   maximum\n");
	if (mip->mi_flags & INVERT_SENSE)
		fprintf(fp, " %13.0f %9.2f %9.2f %9.2f %9.2f %9.2f\n",
	 		cop[0], 1. / cop[2], 1. / cop[1], 1. / cop[3],
			1. / cop[5], 1. / cop[4]);
	else	fprintf(fp, " %13.0f %9.2f %9.2f %9.2f %9.2f %9.2f\n",
		 cop[0], cop[1], cop[2], cop[3], cop[4], cop[5]);
	fprintf(fp,
 " Scaled corrections outside [%8.4f,%8.4f] set to appropriate extrema\n",
			MIN_CORR, MAX_CORR);
	fprintf(fp, "Images processed:\n");
	fprintf(fp, "     Counts   Zmax @  X    Y Filename\n");
}

static xgulong out16(xgulong informat)
{ /*	This forces 16-bit output on return */
	xgulong	out, outswap;

	out = informat;	outswap = swawb(out);
	if ((out == FMT_SAXII) || (outswap == FMT_SAXII)) out += 1;
	else if ((out == FMT_PCS_8) || (outswap == FMT_PCS_8)) out *= 2;
	return out;
}

static double spatialadj(int x, int y, MIMWORK *mip)
{ /*	This computes the area surrounding a single pixel and uses it
	to estimate an area correction */
	int		u0, u1, u2, u3;
	double		ux, vx, uy, vy;
	double		ar0, ar1, ar;
	float		xc0, xc1, xc2, xc3, yc0, yc1, yc2, yc3;

	ux = ((double)x) - 0.5; uy = ((double)y) - 0.5;
	vx = ux + 1.; vy = uy + 1.;
	u0 = pix2cm(ux, uy, &xc0, &yc0, 0); u1 = pix2cm(ux, vy, &xc1, &yc1, 0);
	u2 = pix2cm(vx, uy, &xc2, &yc2, 0); u3 = pix2cm(vx, vy, &xc3, &yc3, 0);
	if ((u0 <= 0) || (u1 <= 0) || (u2 <= 0) || (u3 <= 0)) return 1.;
	ar0 = 0.5 *fabs((xc1 - xc2) * (yc1 - yc0) - (xc1 - xc0) * (yc1 - yc2));
	ar1 = 0.5 *fabs((xc2 - xc1) * (yc2 - yc3) - (xc2 - xc3) * (yc2 - yc1));
	ar = ar0 + ar1;	if (ar < DFUZZ) return 1.;
	else return mip->mi_meanar / ar;
}

static int floodcorr_image(MIMWORK *mip)
{ /*	This corrects an image using a flood image */
	int		fimsiz, flood_flags, fr, hit, nmax, nmin, npts, nrows;
	int		wid, x, y, zcorr;
	xglong		rowcount, totcount;
	static double	rowsum, oasum, oasmin, oasmax;
	double		rcorr, this, mincorr, maxcorr, minspj, maxspj, spj;
	double		corrp[8];
	double		meanw, meanh;
	MAIL		*fcm, *lhd;
	VSPLINE		*lspl;
	static float	*corr_arr, *fcp;

	lhd = mip->mi_heads;	fcm = lhd + 1;	lspl = mip->mi_spl;
	if (NULL == fcm) /* compute correction values if needed */
	  {
		mip->mi_fcname = trname("flood"); /* get name of FLOOD image */
		printf(" Now processing flood-image information in %s...\n",
			mip->mi_fcname);
		flood_flags = GET_RUNHEADER; /* image-reading flag */
			/* read in the FLOOD image */
		if (FRAME_OK != (fr = read_image(mip->mi_fcname,
		 flood_flags, fcm)))
		  {
			(void)image_error(stdout, mip->mi_fcname, fr, 0);
			if (mip->mi_fpl != stdout)
			  (void)image_error(stdout, mip->mi_fcname, fr, 0);
			return EBAD;
		  }
		wid = fcm->xdimen;	hit = fcm->xdimen;
		/* make sure the correction dimensions are okay */
		if ((wid < 5) || (hit < 5))
		  {
			fprintf(stderr, "Bad dimensions: [%d,%d]\n", wid, hit);
			return EBAD;
		  }
		mincorr = 65535;	maxcorr = 0;	fimsiz = wid * hit;
		if (NULL == (corr_arr = (float *)callop(fimsiz,sizeof (float),
		 "correction array in floodcorr_image"))) return EALLOC;
		/* get an estimate of the mean area occupied by a pixel */
		if (mip->mi_flags & SPATIAL_ADJUST)
		  {
			meanw = lspl->v_xrcm * ((double)lspl->v_ncol) /
				((double)lspl->v_detw);
			meanh = lspl->v_yrcm * ((double)lspl->v_nrow) /
				((double)lspl->v_deth);
			mip->mi_meanar = fabs(meanw * meanh);
		  }
		else mip->mi_meanar = 1.;
		zcorr = nrows = oasum = 0;
		minspj = 1.e34;	maxspj = -1.e34;
		for (fcp = corr_arr, y = 0; y < hit; y++)
		   {
			npts = rowsum = 0;
			for (x = 0; x < wid; x++, fcp++)
			   {
				this = imvalue(fcm, x, y);
				if (this > FUZZ)
				  {
					if (mip->mi_flags & INVERT_SENSE)
						this = 1. / this;
					if (mip->mi_flags & SPATIAL_ADJUST)
					  {
						spj = spatialadj(x, y, mip);
						if (spj < minspj) minspj = spj;
						if (spj > maxspj) maxspj = spj;
						if (spj < SPJMIN) spj = SPJMIN;
						if (spj > SPJMAX) spj = SPJMAX;
						this *= spj;
					  }
					npts++;	rowsum += this;
					if (this < mincorr) mincorr = this;
					if (this > maxcorr) maxcorr = this;
					*fcp = this;
				  }
				else {
					zcorr++;	*fcp = 0.;
				     }
			   }
			if (npts) oasum += ((double)rowsum) / ((double)npts);
			if (rowsum) nrows++;
		   }
		if (nrows) oasum /= ((double)nrows);
		oasmin = mincorr / oasum;	oasmax = maxcorr / oasum;
		corrp[0] = zcorr;	corrp[1] = mincorr;
		corrp[2] = maxcorr;	corrp[3] = oasum;
		corrp[4] = oasmin;	corrp[5] = oasmax;
		corrp[6] = minspj;	corrp[7] = maxspj;
		log_header(mip->mi_fpl, corrp, mip);
		if (mip->mi_fpl != stdout) log_header(stdout, corrp, mip);
		/* now go back and compute the normalized correction vals */
		nmin = nmax = 0;
		for (fcp = corr_arr, y = 0; y < hit; y++)
		   {
			for (x = 0; x < wid; x++, fcp++)
			   {
				*fcp /= oasum;
				if (*fcp <= 0.) *fcp = 1.;
				if (*fcp < MIN_CORR)
				  {
					*fcp = MIN_CORR;	nmin++;
				  }
				else if (*fcp > MAX_CORR)
				  {
					*fcp = MAX_CORR;	nmax++;
				  }
			   }
		   }
	  }
	/* make sure the flood-field is the right size */
	if (((int)fcm->xdimen != mip->mi_imwid) ||
	 ((int)fcm->ydimen != mip->mi_imht))
	  {
		fprintf(stderr,
		 "flood-field is %dx%d; image is %dx%d\n", (int)(fcm->xdimen),
		 (int)(fcm->ydimen), mip->mi_imwid, mip->mi_imht);
		return EBAD;
	  }
	for (fcp = corr_arr, totcount = y = 0; y < mip->mi_imht; y++)
	   {
		for (rowcount = x = 0; x < mip->mi_imwid; x++, fcp++)
		   {
			rcorr = *fcp;
			if (rcorr > FUZZ)
			  {
				mip->mi_oldval = imvalue(lhd, x, y);
				mip->mi_newval = ((double)mip->mi_oldval) *
					rcorr + 0.49999;
			  }
			else mip->mi_newval = mip->mi_oldval =
				imvalue(lhd, x, y);
			if (lhd->improps & IPR_IS8BIT) pixelmod8(x, y, mip);
			else pixelmod16(x, y, mip);
			rowcount += mip->mi_newval;
		   }
		if (totcount >= 0) totcount += rowcount;
		else {
			rowcount /= 64;	totcount -= rowcount;
		     }
	   }
	lhd->eventcount = totcount;
	mfix_hdr(mip);
	fprintf(mip->mi_fpl, " %10d %6d %4d %4d %s\n", (int)(lhd->eventcount),
	 (int)(lhd->zmax), lhd->zmaxx, lhd->zmaxy, mip->mi_imname);
	fprintf(stdout, " %10d %6d %4d %4d %s\n", (int)(lhd->eventcount),
	 (int)(lhd->zmax), lhd->zmaxx, lhd->zmaxy, mip->mi_imname);
	return EOK;
}

static void summ_mimrun(FILE *fp, MIMWORK *mip)
{ /*	Print summary of what we're preparing to do on the stream fp */
	static char	*fnames[] =
	 { "frames", "flood", "uspline", "xlog", NULL};
	static char	ftypes[] = { 1, 1, 1, 2, 0};

	ftypes[0] = (mip->mi_fpl == stdout) ? 0 : 2;
	ftypes[2] = (mip->mi_flags & SPATIAL_ADJUST) ? 1 : 0;
	header(fp, PROCESSING, fnames, ftypes);
	fprintf(fp,
	 "Each output image will be input image %s by correction value\n",
		(mip->mi_flags & INVERT_SENSE) ? "divided" : "multiplied");
	if (mip->mi_flags & SPATIAL_ADJUST) fprintf(fp,
	 "Output image values will be corrected for pixel area\n");
}

int multimagemain(int argc, char *argv[])
{ /* Mainline for multiplicative corrections to an image */
	int		ifm, imerr, image_flags, ival1, ival2;
	xgulong		outtype;
	char		multiframe;
	MAIL		*lhd;
	FRAMEINDEX	*lfmi;
	FRAMEENTRY	*ftlast;
	static MIMWORK	mis;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	mis.mi_flags = 0;	lfmi = NULL;
	while ((argc > 1) && (argv[1][0] == '-'))
	  {
		if ((argv[1][1] == 'i') || (argv[1][1] == 'I'))
			mis.mi_flags |= INVERT_SENSE;
		else if ((argv[1][1] == 'a') || (argv[1][1] == 'A'))
			mis.mi_flags |= SPATIAL_ADJUST;
		argc--;	argv++;
	  }
	if (argc < 1)
	  {
		multiframe = 1; if (NULL == (lfmi = ftables(3)))
		  {
			fprintf(stderr, "Can't open frame index\n");
			return EBAD;
		  }
		ival1 = lfmi->fi_firstfm;
		ftlast = lfmi->fi_fme;
		ival2 = lfmi->fi_lastfm;
	  }
	else if ((argc > 2) && allnumeric(argv[1]) && allnumeric(argv[2]))
	  {
		multiframe = 1;
		ival1 = atoi(argv[1]);	ival2 = atoi(argv[2]);
		if (NULL == (lfmi = ftables(3)))
		  {
			fprintf(stderr, "Can't open frame index\n");
			return EBAD;
		  }
		ftlast = lfmi->fi_fme;
		if (ival1 < lfmi->fi_firstfm) ival1 = lfmi->fi_firstfm;
		if (ival1 > ftlast->fmno) ival1 = ival2 = lfmi->fi_firstfm;
		if (ival2 < lfmi->fi_firstfm) ival1 = ival2 = lfmi->fi_firstfm;
		if (ival2 > ftlast->fmno) ival2 = ftlast->fmno;
	  }
	else {
		multiframe = 0;
		if (allnumeric(argv[1]))
		  {
			ival1 = atoi(argv[1]);
			if (NULL == (lfmi = ftables(3)))
			  {
				fprintf(stderr, "Can't open frame index\n");
				return EBAD;
			  }
			if ((0 > ival1) || (lfmi->fi_totfms < ival1))
			  {
				fprintf(stderr, "Frame %d is out of range\n",
				 ival1);
				return EBADVALUE;
			  }
			mis.mi_imname = indexname(ival1);
			ival2 = 1;
		  }
		else	{
				mis.mi_imname = trname(argv[1]);
				ival1 = ival2 = 1;
		        }
	  }
	mis.mi_nmbox = 2;
	if (NULL == mis.mi_heads)
	  {
		if (NULL == (mis.mi_heads = (MAIL *)callop(mis.mi_nmbox,
		 sizeof (MAIL), "mis.mi_heads in multimage")))
		  {
			fprintf(stderr, "Can't allocate mailbox memory\n");
			return EALLOC;
		  }
	  }
	if (NULL == (mis.mi_fpl = propen("xlog"))) mis.mi_fpl = stdout;
	image_flags = GET_RUNHEADER;
	if ((mis.mi_flags & SPATIAL_ADJUST) &&
	 (NULL == (mis.mi_spl = gspline(5, &imerr))))
	  {
		fprintf(stderr, "Spatial corrections unavailable.\n");
		mis.mi_flags &= ~(SPATIAL_ADJUST);
	  }
	/* summarize what we're preparing to do */
	summ_mimrun(mis.mi_fpl, &mis);
	if (mis.mi_fpl != stdout) summ_mimrun(stdout, &mis);
	lhd = mis.mi_heads;
	if (multiframe)
	  {
		for (ifm = ival1; ifm <= ival2; ifm++)
		   {
			if (NULL != (mis.mi_imname = indexname(ifm)))
			  {
				if (FRAME_OK != (imerr = read_image(
				 mis.mi_imname, image_flags, lhd)))
				  {
					(void)image_error(stdout,
					 mis.mi_imname, imerr, 0);
					if (mis.mi_fpl != stdout)
					 (void)image_error(
					  stdout, mis.mi_imname, imerr, 0);
				  }
				else {
					outtype = out16(lhd->format);
					mis.mi_imwid = lhd->xdimen;
					mis.mi_imht = lhd->ydimen;
					if (EOK == floodcorr_image(&mis))
					  {
						if (outtype != lhd->format)
						  cvt_8_to_16(lhd,(int)outtype);
						output_frame(mis.mi_imname,
						 lhd, outtype, 0x200);
					  }
					else break;
				     }
			  }
		   }
	  }
	else {
		if (FRAME_OK != ( imerr = read_image(mis.mi_imname,
		 image_flags, lhd)))
		  {
			(void)image_error(stdout, mis.mi_imname, imerr, 0);
			if (mis.mi_fpl != stdout) (void)image_error(
				mis.mi_fpl, mis.mi_imname, imerr, 0);
			if (stdout != mis.mi_fpl) (void)fclose(mis.mi_fpl);
			mis.mi_fpl = NULL;
			globalkill();	return EBAD;
		  }
		outtype = out16(lhd->format);
		mis.mi_imwid = lhd->xdimen;
		mis.mi_imht = lhd->ydimen;
		if (EOK == floodcorr_image(&mis))
		  {
			if (outtype != lhd->format) cvt_8_to_16(lhd, outtype);
			output_frame(mis.mi_imname, lhd, outtype, 0x200);
		  }
	     }
	if (stdout != mis.mi_fpl) (void)fclose(mis.mi_fpl);
	mis.mi_fpl = NULL;
		/* temporary: so "globalkill" doesn't fail */
	/* temporarily delete "globalkill" since it's not working.
	mis.mi_heads->overflow = (mis.mi_heads+1)->overflow = NULL;
	(void)ftables(0);	kill_mail(&lhd, &nmbox);
	globalkill();	*/	return EOK;
}

/* end of multimage functionality */

/***********************************
	This converts 1 or more images to a spatially-corrected format.
************************************/

int spacormain(int argc, char *argv[])
{ /* Mainline for performing a spatial correction on one or more images */
	register int	i;
	int		readflags, readret;
	MAIL		*thishd;
	static MAIL	*heads;
	static int	nmbox;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	if (argc < 2)
	  {
		fprintf(stderr,"usage: spacor filename [...]\n");
		return EOK;
	  }
	readflags = GET_RUNHEADER;
	nmbox = (argc > PFRAMES) ? PFRAMES + 1 : argc;
	if (NULL == heads)
	  {
		if (NULL == (AVOID *)(heads = (MAIL *)callop(nmbox,
		 sizeof (MAIL), "heads in spacor"))) return EALLOC;
	  }
	for (i = 1; i < argc; i++)
	   {
		thishd = heads + ((i-1) % nmbox);
		if (FRAME_OK != (readret =
		  read_image(argv[i], readflags, thishd)))
		  {
			(void)image_error(stdout, argv[i], readret, 0);
			continue;
		  }
		spatial(0, 0, 0x3000, thishd);
	   }
	kill_mail(&heads, &nmbox);
	globalkill();	return EOK;
}

/* end of spa_cor functionality */

/*********************************
	This converts the contents of a Bruker header to newline-terminated
	ascii, making it easier to print the header.
	It reads stdin as a Bruker image or its header and writes out
	a modified version of the header in which newlines are appended to
	the end of each nonblank text line.
**********************************/

int bruheadmain(int argc, char *argv[])
{ /* Mainline for printing Bruker header information */
	int		i, nchars;
	char		*cp;
	static char	iline[83];

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	for (i = 0; i < 96; i++)
	   {
		if (1 == fread(iline, 80, 1, stdin))
		  {
			iline[80] = '\n';	iline[81] = '\0';
			nchars = 80;
			for (cp = &(iline[79]); cp >= &(iline[0]); cp--)
			   {
				if (' ' != *cp)
				  {
					*(cp+1) = '\n'; *(cp+2) = '\0';
					break;
				  }
				nchars = (cp + 1) - &(iline[0]);
			   }
			if (1 != fwrite(iline, nchars, 1, stdout))
			  {
				fprintf(stderr,
				 "Error converting line %d\n", i);
				return EBAD;
			  }
		  }
		else fprintf(stderr, "Error reading line %d\n", i);
	   }
	fprintf(stderr, " %5d lines of header written out\n", i);
	globalkill();	return EOK;
}

/* end of bruhead functionality */

/***********************************
	This reads in an R-Axis header, flips the sign of the stepsize,
	and writes the header back out.
*************************************/

static int flipraxis(char *finame)
{ /*	This reads in an R-Axis header, flips the sign of the
	stepsize, and writes the file back out. */
	size_t		siz;
	double		oldstep;
	RAXISHEAD	trax;
	static FILE	*fpi;

	if (NULL == (fpi = fopen(finame, "r+"))) return EOPENERR;
	siz = sizeof (RAXISHEAD);
	if (siz != fread(&trax, 1, siz, fpi)) return EREADERR;
	oldstep = trax.fie0 - trax.fis0;
	trax.fie0 = trax.fis0 - oldstep;
	if (-1 == fseek(fpi, 0L, 0)) return ESEEKERR;
	if (siz != fwrite(&trax, 1, siz, fpi)) return EWRITERR;
	(void)fclose(fpi);	fpi = NULL;	return EOK;
}

int fliprmain(int argc, char *argv[])
{ /* Mainline for reversing the sign of the stepsize in an R-Axis header */
	int ia, ir;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	for (ia = 0; ia < argc; ia++)
	   {
		if (!allnumeric(argv[ia]))	
		  {
			if (EOK != (ir = flipraxis(argv[ia]))) return ir;
		  }
	   }
	return EOK;
}

/* end of fliprot.c */
