/* mreftools.c -- conversion, modification, & printing tools for MULTIREFs

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	27 Jun 2007: rescmrmain returns rather than calling exit
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	24 Jun 2006: include xgpostint.h rather than rejdefs.h
	16 May 2006: merge bigsigma, expunge, mergeshift, multiobs, rescmr.c
	31 Dec 2005: expunge: handle qsort with fewer than 2 elements correctly
	15 Oct 2005: mergeshift: eliminate static globals by
		reference to MSHWORK structs
	15 Oct 2005: expunge: replace static globals with REJECTRUN struct elts
	10 Oct 2005: housekeeping
	 6 Sep 2004: bigsigma: cosmetic change
	28 Jun 2005: adjust_refl(): handle return val from xgwrite correctly
	28 Jun 2005: delread(): typecast pointer difference on printout
	15 May 2005: multiobs: moved from utilities to nogui
	11 May 2005: check return status from xgjournal
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	29 Aug 2004: routines used only here are declared as static;cosmetics
	26 Aug 2004: endmult, used only here, is declared as static; cosmetics
	 1 Feb 2004: bigsigma & multiobs: don't add 100 to the spacegroup for
		rhombohedrals if it's already been done.
	24 Jan 2004: a few static globals converted to locals
	21 Jan 2004: license message inserted and initial comments reordered
	 8 Jan 2004: static globals replace extern globals
	 8 Jan 2004: mergeshift: update call to init_shift()
	19 Sep 2003: rescmr: explicitly typecast rsp->rbi_lowlim on printout
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	17 Aug 2003: expunge and mergeshift: replace global vbl uphkl[] with
	 static lphkl[]; call munpack a few extra times to populate lphkl[].
	 5 Jan 2003: reorganized source-file header
	15 Nov 2002: removed declaration of an unused array in multiobs
	 1 Nov 2002: expunge: trivial reorganization of logfile output
	15 Apr 2002: (?) in mergesmain(),
		modify calls related to shlp[] since it's handled differently
	25 Mar 2002: removed MSI attachments
	15 Oct 1998: initial version of multiobs
	27 Jul 1997: initial version of rescmr
	24 Aug 1995: minor modification of bigsima
	 4 Aug 1995: initial version of bigsigma
	 5 Oct 1994: use xgread and xgwrite on various MULTIREF structure types
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: a few I/O changes and global variables moved around
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN versions of expunge,
	18 Jun 1986: mergeshift: ignore shifts with shift value < 0,
		i.e. deleted observations.
	22 Oct 1985: expunge: rename the program to "expunge"
	29 Sep 1985: expunge: write different output file and incorporate XENGEN
		data collection/reduction programming conventions.
	20 Sep 1984: expunge: initial program, based on `delmref' from La Jolla.

   Purpose: alterations, reorganizations, and printouts of MULTIREF files
	specifically: 
	 "bigsigma" prints reflections in a multiref file for which
		at least one observation has sigma > Q * meansigma
	 "multiobs" prints reflections in a MULTIREF for which at least
		two observations are present.
	 "rescmr" rescales the sigma values in a MULTIREF.
	 "expunge" marks MULTIREF observations found in a printable list
		so they will be ignored (excluded) in subsequent manipulations
		of the MULTIREF (scaling, output, etc.)
	 "mergeshift" recalculates scaling-group numbers for the observations
		in a MULTIREF
   Cross references: ../postprocess/{sc*,rej*}.c, ../utilities/redurefl.c,
		../utilities/x_dump.c, ../xglib/ascout.c
*************************************************************************/
#define		NEED_TIME	1
#include	"xgpostint.h"
#define		DEF_BIGSIGCUT	4
#define		MAXDEL	2000
#define		DEFAULT_MIN_OK	20	/* minimum # of ok obs per shift */

 typedef struct {
		xglong	rbi_lowlim;
		double	rbi_gof100;
		} RESCALEIT;

 typedef struct {
		unsigned	ms_nobin, ms_nobout, ms_nobmoved, ms_nscales;
		unsigned	ms_nrefin, ms_nrefout, ms_nreflost;
		unsigned	ms_nshin, ms_nshout;
		short		*ms_shinrun, *ms_nobinsh;
		short		*ms_newshval;
		FILE		*ms_fpl, *ms_fpmo;
		CRYSTAL		*ms_xhp;
		SHIFT		*ms_shlp;
		} MSHWORK;

extern int bigsigmain(int argc, char *argv[]);
extern int rescmrmain(int argc, char *argv[]);
extern int multmain(int argc, char *argv[]);
extern int expungemain(int argc, char *argv[]);
extern int mergesmain(int argc, char *argv[]);

int bigsigmain(int argc, char *argv[])
{ /* Mainline for printing out refls with big sigmas */
	register int	nbigsig, iob;
	register ABBROB	*aobp;
	double		bigsigcut, signum, sigden, specbig;
	FILE		*fpmi, *fplog;
	CRYSTAL		xhd;
	DATARUN		rhd;
	MERGEDREF	mr;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	if (qerror((AVOID *)(fpmi = xgopen("multiref","r")),
		"open multiref input file")) return EOPENERR;
	if (NULL == (AVOID *)(fplog = propen("xlog"))) fplog = stdout;
	if (verror(0 < xgread((void *)&xhd, XG_CRYSTAL, fpmi),
		"read crystal header from input multiref"))
	  {
		if (stdout != fplog) (void)fclose(fplog);
		return EREADERR;
	  }
	if ((xhd.xt_isys == 7) && (xhd.xt_intl < 200)) xhd.xt_intl +=  100;
	for (iob = xhd.xt_nruns; iob--; )
	   if (verror(0 <= xgread((void *)&rhd, XG_DATARUN, fpmi),
	    "read datarun header from input multiref"))
	     {
		if (stdout != fplog) (void)fclose(fplog);
		return EREADERR;
	     }
	if ((argc > 1) && (allnumeric(argv[1])) &&
	 (DFUZZ < (specbig = atof(argv[1])))) bigsigcut = specbig;
	else bigsigcut = DEF_BIGSIGCUT;
	fprintf(fplog,
	 "Program will search for sigma > %6.1f * <sigma>\n",
		bigsigcut);
	(void)keep_dimen(&rhd);
	nbigsig = 0;
	while (0 < xgread((void *)&mr, XG_MERGEDREF, fpmi))
	  if ((mr.mr_ibar > 1) && (mr.mr_ibar > mr.mr_isig) &&
	      (mr.mr_nob > 1))
	     {
		signum = sigden = 0.;
		for (aobp = &(mr.mr_ob[0]), iob = 0;
		 iob < mr.mr_nob; iob++, aobp++)
		   {
			sigden += 1.;	signum += aobp->abo_sigma;
		   }
		aobp -= mr.mr_nob;
		if (sigden > DFUZZ)
		  {
			signum /= sigden;
			for (iob = 0; iob < mr.mr_nob; iob++, aobp++)
			   if (aobp->abo_sigma > bigsigcut * signum)
			     {
				pr_merref(fplog, &mr, xhd.xt_intl);
				nbigsig++;	break;
			     }
		  } /* end of "if (sigden" clause */
	     } /* end of "while (0 < xgread" and accompanying "if" */
	(void)fclose(fpmi);	fpmi = NULL;
	fprintf(fplog,
	 "\t%6d reflections found with sigma > %6.1f * <sigma>\n",
		nbigsig, bigsigcut);
	if (stdout != fplog) printf(
	 "\t%6d reflections found with sigma > %6.1f * <sigma>\n",
		nbigsig, bigsigcut);
	(void)fclose(fplog);	fplog = NULL;	globalkill();	return EOK;
}

/* end of bigsigma functionality */

/***************************
	The next two routines enable us to print the multiply-observed
	reflections in a MULTIREF.
****************************/

static int endmult(int reval, FILE *fpmi)
{ /*	Clean ending of this function */
	if (NULL != fpmi) (void)fclose(fpmi);	fpmi = NULL;
	globalkill();	return reval;
}

int multmain(int argc, char *argv[])
{ /* Mainline for printing multiply-observed reflections from a MULTIREF */
	register int	minmult, nmult;
	short		ispgp;
	int		nruns, specminmult;
	FILE		*fplog, *fpmi;
	CRYSTAL		xhd;
	DATARUN		rhd;
	MERGEDREF	mr;
	static char	*fnames[] =	{ "multiref", "xlog", NULL };
	static char	ftypes[] =	{ 1, 3, 0 };
	static char	countmsg[] =
	 "\t%6d reflections with at least %2d observations found\n";

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	if (qerror((AVOID *)(fpmi = xgopen("multiref","r")),
		"open multiref input file")) return endmult(EOPENERR, fpmi);
	if (NULL == (AVOID *)(fplog = propen("xlog"))) fplog = stdout;
	minmult = 2;
	if ((argc > 1) && (argv[1][0] == '-') && ((argv[1][1] == 'm') ||
	 (argv[1][1] == 'M')))
	  {
		specminmult = atoi(&(argv[1][2]));
		if (specminmult > 0) minmult = specminmult;
	  }
	header(fplog , REDUCTION, fnames, ftypes);
	header(stdout, REDUCTION, fnames, ftypes);
	nmult = 0;
	if (verror(0 < xgread((void *)&xhd, XG_CRYSTAL, fpmi),
		"read crystal header from input multiref"))
	  {
		if (stdout != fplog) (void)fclose(fplog);
		return endmult(EREADERR, fpmi);
	  }
	ispgp = xhd.xt_intl;
	if ((ispgp < 200) && (xhd.xt_isys == 7)) ispgp += 100;
	for (nruns = xhd.xt_nruns; nruns--; )
		if (verror(0 <= xgread((void *)&rhd, XG_DATARUN, fpmi),
		  "read datarun header from input multiref"))
		  {
			if (stdout != fplog) (void)fclose(fplog);
			return endmult(EREADERR, fpmi);
		  }
	(void)keep_dimen(&rhd);
	while (0 < xgread((void *)&mr, XG_MERGEDREF, fpmi))
	     {
		if (mr.mr_nob >= minmult)
		  {
			pr_merref(fplog, &mr, ispgp); nmult++;
		  }
	     }
	fprintf(fplog, countmsg, nmult, minmult);
	if (stdout != fplog)
	  {
		fprintf(stdout, countmsg, nmult, minmult);
		(void)fclose(fplog);
	  }
	return endmult(EOK, fpmi);
}

/* end of multiobs functionality */

/***************************
	The next two routines enable us to rescale the sigma values
	in a MULTIREF.

   Purpose: reading an input MULTIREF and writing a modified MULTIREF
	for which the sigmas have been adjusted.
   Cross references: ../utilities/redurefl.c, ../postprocess/stat*.c

	It reads in a multiref and writes it back out with the sigmas modified
	according to the goodness-of-fit values specified (for now) through
	the text file GOFVALUES.  GOFVALUES looks something like this:
      0     0.00
     40    87.10
     80    96.55
    120   103.28
    160   105.99
    320   116.30
    640   143.95
   2560   195.63
  20480   233.23
 where each line contains the lower limit of an intensity bin and
 the goodness-of-fit value for that bin.
****************************/

static double get_gof(double inten, int nline, RESCALEIT *rsb0)
{ /*	This computes the correction factor on sigma that should be applied
	to a particular reflection's sigma, based on its intensity. */
	int		ibi;
	double		hiwt, lowt;
	RESCALEIT	*rsp, *rspp;

	/* very low intensity: use zeroth bin to get gof100 value */
	if (nline < 2) return 0.01 * rsb0->rbi_gof100;
	/* very high: use (n-1)st bin */
	else if (inten < rsb0->rbi_lowlim) return 0.01 * rsb0->rbi_gof100;
	else {
		rsp = rsb0 + (nline - 1);
		/* otherwise, interpolate between two of them */
		if (inten >= rsp->rbi_lowlim) return 0.01 * rsp->rbi_gof100;
		for (rsp = rsb0 + 1, ibi = 1; ibi < nline; ibi++, rsp++)
		   if (rsp->rbi_lowlim > inten)
		     {
			rspp = rsp - 1;
			lowt = flimits((rsp->rbi_lowlim - inten) /
			 (rsp->rbi_lowlim - rspp->rbi_lowlim), 0., 1.);
			hiwt = 1. - lowt;
			return 0.01 * (lowt * rspp->rbi_gof100 +
			 hiwt * rsp->rbi_gof100);
		     }
	     }
	return 1.;	/* rare! */
}

static void sum_rescmr(FILE *fp, int haslog)
{ /*	This summarizes a run of `rescmr' to the stream `fp'. If the
	argument `haslog' is nonzero, we report on the logfile */
	static char	*flist[] = {"multiref", "mrefout", "xlog", NULL};
	static char	ftype[] = {1, 2, 2, 0 };

	ftype[2] = (haslog) ? 0 : 2;	header(fp, REDUCTION, flist, ftype);
}

int rescmrmain(int argc, char *argv[])
{ /*	This reads an input multiref and writes out an output multiref
	identical to it except that the shift values for all reflections
	with sin(theta)/lambda outside a user-specified range are set to
	the negative of their input absolute value.
	This has the effect of deleting those observations from the data set.*/
	register int	i, iwe;
	int		binlim, nitems, nlines;
	char		gofline[80], gofname[80], outname[80];
	double		gofv;
	FILE		*fplog, *fpmi, *fpmo, *fpp;
	ABBROB		*abp;
	RESCALEIT	*rsb, *rsp;
	CRYSTAL		xhd;
	DATARUN		rhd;
	MERGEDREF	mr;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	strcpy(gofname, trname("gofvalues"));
	if (qerror((AVOID *)(fpp = xgopen(gofname, "r")),
	 "open input goodness-of-fit file")) return EOPENERR;
		/* count how many userfile lines in the file */
	for (nlines = 0; NULL != fgets(gofline, 80, fpp); )
	     {
		nitems = sscanf(gofline, "%d %lf", &binlim, &gofv);
		if ((nitems >= 2) && (gofv > DFUZZ)) nlines++;
	     }
	fprintf(stderr, "After first pass: nlines = %d\n", nlines);
	if (NULL == (rsb = (RESCALEIT *)callop(nlines, sizeof (RESCALEIT),
	 "goodness-of-fit stats for data reduction")))
	  {
		fprintf(stderr,
		 "Error allocating memory for G.O.F. structures\n");
		return EALLOC;
	  }
	if (-1 == fseek(fpp, (long)0L, 0))
	  {
		fprintf(stderr, "Cannot rewind GOF file\n"); return EREADERR;
	  }
	for (rsp = rsb, i = 0; i < nlines; i++)
	   {
		if (NULL != fgets(gofline, 80, fpp))
		  {
			nitems = sscanf(gofline, "%d %lf", &binlim, &gofv);
			if ((nitems >= 2) && (gofv > DFUZZ))
			  {
				rsp->rbi_lowlim = binlim;
				rsp->rbi_gof100 = gofv;
				fprintf(stderr, "Valid line: %6d %9.3f\n",
				 (int)(rsp->rbi_lowlim), rsp->rbi_gof100);
				rsp++;
			  }
		  }
		else	break;
	   }
	(void)fclose(fpp);	fpp = NULL;	nlines = rsp - rsb;
	if (nlines < 1)
	  {
		fprintf(stderr,
		 "No valid sigma-rescaler lines in the gof file\n");
		return EBAD;
	  }
	if (qerror((AVOID *)(fpmi = xgopen("multiref","r")),
		"open input multiref")) return EOPENERR;
	if (verror(0 < xgread((void *)&xhd, XG_CRYSTAL, fpmi),
		"read crystal header from input multiref")) return EREADERR;
	strcpy(outname, trname("mrefout"));
	if (qerror((AVOID *)(fpmo = xgopen(outname, "w")),
		"open input GOF file")) return EOPENERR;
	if (verror(0 < xgwrite((void *)&xhd, XG_CRYSTAL, fpmo),
		"write crystal header to output multiref")) return EWRITERR;
	for (i = xhd.xt_nruns; i--; )
	   {
		if (verror(0 < xgread((void *)&rhd, XG_DATARUN, fpmi),
		 "read datarun header from input multiref")) return EREADERR;
		if (verror(0 < xgwrite((void *)&rhd, XG_DATARUN, fpmo),
		 "write datarun header to output multiref")) return EWRITERR;
	   }
	if (NULL == (fplog = propen("xlog"))) fplog = stdout;
	sum_rescmr(stdout, (fplog != stdout));
	if (fplog != stdout) sum_rescmr(fplog, 1);
	nitems = 0;
	while (0 < xgread((void *)&mr, XG_MERGEDREF, fpmi))
	     {
		for (i = mr.mr_nob, abp = &(mr.mr_ob[0]); i--; abp++)
		   {
			abp->abo_sigma *= (gofv =
				get_gof(abp->abo_ii, nlines, rsb));
			nitems++;
			if (!(nitems % 2000))
			  {
				fprintf(stderr,
				 "I,sig,resc = %9.2f,%9.2f,%9.4f\n",
					abp->abo_ii, abp->abo_sigma, gofv);
			  }
		   }
		iwe = xgwrite((void *)&mr, XG_MERGEDREF, fpmo);
		if (0 >= iwe)
		  {
			fprintf(fplog,
			 "Error in writing [%d,%d,%d,%d] to %s\n",
				mr.mr_hkl[0], mr.mr_hkl[1],
				mr.mr_hkl[2], mr.mr_hkl[3], outname);
			if (stdout != fplog) (void)fclose(fplog);
			fplog = NULL;	return iwe;
		  }
	     }
	(void)fclose(fpmo);	fpmo = NULL;
	if (stdout != fplog) (void)fclose(fplog);
	fplog = NULL;	globalkill();	return EOK;
}

/* end of rescmr functionality */

/***************************
	The next several routines allow us to flag observations in a MULTIREF
	based on the contents of a textfile.
	This marks observations in a multiref so that they will be ignored
	during scaling and averaging.  The observations to be marked are those
	found on a printable list with environment variable DELETIONS.
	The observations on the list must match obs found in the MULTIREF
	rather closely.  "Marking" consists of setting the shift value of
	the observation to its negative.  Thus (but don't tell anybody!)
	already-deleted observations can be undeleted with this program.
	The functionality reads one multiref and writes another.
****************************/

static int delsetup(REJECTRUN *erp)
{ /* File processing for expunge */
	static char	*flist[] = {
		"deletions", "multiref", "mrefout", "xlog", NULL };
	static char	ftype[] = { 1, 1, 2, 2, 0 };

	erp->rj_tl0 = NOW;
	if (qerror((AVOID *)(erp->rj_fpmi = xgopen("multiref", "r")),
		"open input multiref")) return EOPENERR;
	if (qerror((AVOID *)(erp->rj_fpmo = xgopen("mrefout", "w")),
	 "open output multiref")) return EOPENERR;
	if (NULL == (erp->rj_fpl = propen("xlog"))) erp->rj_fpl = stdout;
	ftype[3] = (erp->rj_fpl == stdout ? 0 : 2);
	if (erp->rj_fpl != stdout) header(erp->rj_fpl, REDUCTION, flist,ftype);
	header(stdout, REDUCTION, flist, ftype);
	return EOK;
}

static int delread(REJECTRUN *erp)
{ /*	This reads in the deletion list and sorts it by packed hkl */
	register int	i;
	char		line[100];
	short		qh[3];
	int		rdcham, rdh, rdk, rdl, rdom, rdrun, rdsh, rdx, rdy;
	unsigned	nobsdel;
	static FILE	*fpdel;

/* first we see how many lines we have so we can allocate enough memory */
	if (qerror((AVOID *)(fpdel = xgopen("deletions", "r")),
		"open deletion list for input")) return EOPENERR;
	for (nobsdel = 0; (NULL != fgets(line, 80, fpdel)); ) nobsdel++;
	(void)fseek(fpdel, 0L, 0);	/* rewind the deletion list */
	if (nobsdel > MAXDEL)
	  {
		fprintf(stderr,
		 "Warning: only the first %d deletions will be performed\n",
			MAXDEL);
		nobsdel = MAXDEL;
	  }
	printf("%6d deletions attempted\n", (int)nobsdel);
			/* get memory for the binary deletion list */
	if (NULL == (AVOID *)(erp->rj_drefs = (DELETEDOB *)callop(nobsdel + 10,
	 sizeof (DELETEDOB), "erp->rj_drefs in expunge"))) return EALLOC;
	erp->rj_drefend = erp->rj_drefs + nobsdel;
	/* read and write the headers */
	if (verror(0 < xgread((void *)erp->rj_xhdp, XG_CRYSTAL, erp->rj_fpmi),
		"read crystal header from input multiref")) return EREADERR;
	if (verror(0 < xgwrite((void *)erp->rj_xhdp, XG_CRYSTAL,
	 erp->rj_fpmo), "write crystal header to output multiref"))
		return EWRITERR;
	for (i = 0; i < (erp->rj_xhdp)->xt_nruns; i++)
	   {
		if (verror(0 < xgread((void *)erp->rj_rhdp, XG_DATARUN,
		 erp->rj_fpmi), "read datarun header from input multiref"))
			return EREADERR;
		if (verror(0 < xgwrite((void *)erp->rj_rhdp, XG_DATARUN,
		 erp->rj_fpmo), "write datarun header to output multiref"))
			return EWRITERR;
	   }
	/* print a header for the found refls on the logfile */
	fprintf(erp->rj_fpl, "\tDeletions requested:\n");
	fprintf(erp->rj_fpl,
		"   h   k   l shift omega   x   y    h   k   l shift");
	fprintf(erp->rj_fpl,
		" omega   x   y    h   k   l shift omega   x   y\n");
	for (erp->rj_drefp = erp->rj_drefs;
	 NULL != fgets(line, 80, fpdel); (erp->rj_drefp)++)
	   { /* read each line into a FASTOB structure */
		rdh = rdk = rdl = rdsh = rdom = rdx = rdy = rdcham = rdrun = 0;
		(void)sscanf(line,"%d%d%d %d%d%d %d%d%d",
		 &rdh, &rdk, &rdl, &rdom, &rdx, &rdy, &rdsh, &rdcham, &rdrun);
		qh[0] = rdh;	qh[1] = rdk;	qh[2] = rdl;
		rdx /= 2;	if (rdy >= 256) rdy -= 256;
		/* get packed versions of (h,k,l) and (omega,x,y) */
		mpack(qh, (erp->rj_drefp)->del_hkl, (erp->rj_xhdp)->xt_intl);
		(erp->rj_drefp)->del_omxy = LSH(rdy, 24) | LSH(rdx, 16) | rdom;
		(erp->rj_drefp)->del_shift = rdsh;
		(erp->rj_drefp)->del_chamno = rdcham;
		(erp->rj_drefp)->del_runno = rdrun;
		fprintf(erp->rj_fpl, "%4d%4d%4d%6d%6d%4d%4d%c",
			qh[0], qh[1], qh[2], rdsh,  rdom,  rdx,  rdy, 
			((erp->rj_drefp - erp->rj_drefs) + 1)%3 ? ' ' : '\n');
	   }
	(void)fclose(fpdel);	fpdel = NULL;
	fprintf(erp->rj_fpl,
	 "\n\nThe program will attempt to delete %6d observations\n",
		(int)(erp->rj_drefp - erp->rj_drefs));
		/* sort reflections on hkl with homemade routine or qsort */
	if (nobsdel > 1) qsort((AVOID *)(erp->rj_drefs), nobsdel,
	 sizeof (DELETEDOB), (QPE)hklcmp);
	/* print a header for the found refls on the logfile */
	fprintf(erp->rj_fpl,"\n\tReflections processed:\n");
	fprintf(erp->rj_fpl,
 "   h   k   l\t omega   x   y\tintensity    sigma shift chamber run\n");
	return EOK;
}

static void expu_logdel(MERGEDREF *mrp, int obnum, REJECTRUN *erp)
{ /*	This spits to the logfile a message appropriate to the attempt
	to delete observation 'obnum' from reflection 'mrp'. 'obnum'
	is -1 if the attempt was unsuccessful. */
	register ABBROB	*aobp;
	short		thisom, thisx, thisy;
	static short	lphkl[3];

	munpack((erp->rj_drefp)->del_hkl, lphkl, (erp->rj_xhdp)->xt_intl);
	thisy = (short)((erp->rj_drefp)->del_omxy >> 24)  |
			(((erp->rj_drefp)->del_shift % 2) << 8);
	thisx = (short)(((erp->rj_drefp)->del_omxy >> 16) & 0xff) * 2 + 1;
	thisom = (short)((erp->rj_drefp)->del_omxy & 0xffff);
	if (obnum > -1)
	  {
		aobp = &(mrp->mr_ob[obnum]);
		fprintf(erp->rj_fpl,
	 "%4d%4d%4d\t%6d%4d%4d\t %8.2f %8.3f %6d %7d %3d ACTUALLY DELETED\n",
			lphkl[0],lphkl[1],lphkl[2], thisom, thisx, thisy,
			aobp->abo_ii, aobp->abo_sigma,
			(erp->rj_drefp)->del_shift,
			(erp->rj_drefp)->del_chamno,
			(erp->rj_drefp)->del_runno);
	  }
	else 	fprintf(erp->rj_fpl,
 "%4d%4d%4d\t%6d%4d%4d\t                  %6d %2d %2d        NOT DELETED\n",
		lphkl[0], lphkl[1], lphkl[2], thisom, thisx, thisy,
		(erp->rj_drefp)->del_shift,
		(erp->rj_drefp)->del_chamno, (erp->rj_drefp)->del_runno);
	return;
}

static int expun_final(REJECTRUN *erp)
{ /*	This does the last few cleanup operations
	AND reports some statistics to the screen and to the log file */
	time_t	tl1;

	if (!(erp->rj_nobmarked)) return EOK;
	fprintf(erp->rj_fpl,
 "%6d observations were actually flagged with shift < 0 in output multiref\n",
		(int)(erp->rj_nobmarked));
	printf(
 "%6d observations were actually flagged with shift < 0 in output multiref\n",
		(int)(erp->rj_nobmarked));
	tl1 = NOW;
	fprintf(erp->rj_fpl ,"Deletions required %6d seconds of run time\n",
		(int)(tl1 - erp->rj_tl0));
	printf("Deletions required %6d seconds of run time\n",
		(int)(tl1 - erp->rj_tl0));
	return EOK;
}

static int expucopy(MERGEDREF *mrp, REJECTRUN *erp)
{ /*	This routine is called when we've exhausted the deletion list.
	At that point we need to write out the current reflection and
	copy all the remaining reflections.*/
	if (NULL != mrp)
	  {
		if (0 >= xgwrite(mrp, XG_MERGEDREF, erp->rj_fpmo))
			return expun_final(erp);
		while (0 < xgread(mrp, XG_MERGEDREF, erp->rj_fpmi))
			if (0 >= xgwrite(mrp, XG_MERGEDREF, erp->rj_fpmo))
				return expun_final(erp);
	  }
	return expun_final(erp);
}

static void flag_rest(MERGEDREF *mrp, REJECTRUN *erp)
{ /*	This routine is called when we run out of multiref reflections
	before we run out of deletable observations */
	while (++(erp->rj_drefp) < erp->rj_drefend) expu_logdel(mrp, -1, erp);
	(void)expun_final(erp);
}

static int expuapply(REJECTRUN *erp)
{ /*	This takes the DELETEDOB list and uses it as a guide to the output.
	Any reflection absent from the list is written to the output unchanged.
	Any reflection present in the list has its observations checked,
	and the flagged observations are written out with the negative shift
	values. The list of deleted observations is in the same order as the
	multiref itself.*/
	register int	diffind, i;
	register ABBROB	*aobp;
	xgulong		pomxy;
	MERGEDREF	mr;

	if (verror(1 + fseek(erp->rj_fpmi, 0L, 0), "rewind input multiref"))
		return ESEEKERR;
	if (qerror((AVOID *)(erp->rj_fpmo = xgopen("mrefout", "w")),
		"open output multiref")) return ESEEKERR;
	if (verror(0 < xgread((void *)erp->rj_xhdp, XG_CRYSTAL, erp->rj_fpmi),
		"reread input crystal header")) return EREADERR;
	if (verror(0 < xgwrite((void *)erp->rj_xhdp, XG_CRYSTAL,
	 erp->rj_fpmo), "write output crystal header")) return EWRITERR;
	for (i = (erp->rj_xhdp)->xt_nruns; i--; )
	   {
		if (verror(0 < xgread((void *)erp->rj_rhdp, XG_DATARUN,
		 erp->rj_fpmi), "reread input datarun header"))
			return EREADERR;
		if (verror(0 < xgwrite((void *)erp->rj_rhdp, XG_DATARUN,
		 erp->rj_fpmo), "write output datarun header"))
			return EWRITERR;
	   }
	if (verror(0 < xgwrite((void *)&mr, XG_MERGEDREF, erp->rj_fpmi),
	 "reread first reflection")) return EREADERR;
	erp->rj_nobmarked = 0;
	erp->rj_drefend = erp->rj_drefp;
	erp->rj_drefp = erp->rj_drefs;	/* rewind delet. list*/
		/* compare hkl indices */
	diffind = hklcmp(mr.mr_hkl, erp->rj_drefp->del_hkl);
	while (erp->rj_drefp <= erp->rj_drefend)
	     {
		 while (!diffind)/* multiref hkl and list hkl are equal */
		     { /* ... look thru obs and modify them */
			for (i = 0; i < (int)mr.mr_nob; i++)
			   {
				aobp = &(mr.mr_ob[i]);
			/* see if the omega, x, y values are the same */
				pomxy = LSH(aobp->abo_yob, 24) |
				  LSH(aobp->abo_xob, 16)| aobp->abo_omob;
			/* they are: flag the ob by negating its shift # */
				if (pomxy == (erp->rj_drefp)->del_omxy)
				  {
					aobp->abo_shift = -abs(aobp->abo_shift);
					(erp->rj_nobmarked)++;
					expu_logdel(&mr, i, erp);
				  }
			   }
			/* advance the pointer in the deletion list.  If that
			runs us out the end, copy out the rest of the input */
			if ((++(erp->rj_drefp)) == erp->rj_drefend)
				return expucopy(&mr, erp);
			/* see if we have another ob of the same refl */
			diffind = hklcmp(mr.mr_hkl, erp->rj_drefp->del_hkl);
		     }
		while (diffind < 0) /* index in multiref is earlier */
		     { /* ... write refl out and read the next one */
			(void)xgwrite((void *)&mr, XG_MERGEDREF,
				erp->rj_fpmo);
			/* if we run out of reflections at this point,
			we note in the logfile the fact that some other
			deletions were requested and weren't acted upon */
			if (0 >= xgread((void *)&mr, XG_MERGEDREF,
			 erp->rj_fpmi)) flag_rest(&mr, erp);
			/* otherwise, we just compare hkls again. */
			diffind = hklcmp(mr.mr_hkl, erp->rj_drefp->del_hkl);
		     }
		while (diffind > 0)
		     { /* index in list is earlier.  This happens when
			we've jumped past the hkl of the list refl. */
			expu_logdel(&mr, -1, erp);
			if (++(erp->rj_drefp) == erp->rj_drefend)
				return expucopy(&mr, erp);
			diffind = hklcmp(mr.mr_hkl, erp->rj_drefp->del_hkl);
		     }
	     }
	return EOK;
}

static void expun_finish(REJECTRUN *erp)
{ /* graceful shutdown of expunge */
	if (NULL != erp->rj_fpmi) (void)fclose(erp->rj_fpmi);
	erp->rj_fpmi = NULL;
	if (NULL != erp->rj_fpmo) (void)fclose(erp->rj_fpmo);
	erp->rj_fpmo = NULL;
	if ((NULL != erp->rj_fpl) && (stdout != erp->rj_fpl))
		(void)fclose(erp->rj_fpl);
	memzap((AVOID **)&(erp->rj_drefs), "erp->rj_drefs in expunge");
	erp->rj_fpl = NULL;	globalkill();
}

/*ARGSUSED*/
int expungemain(int argc, char *argv[])
{ /* mainline for expunge functionality */
	int		result;
	REJECTRUN	ers;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	if (EOK == (result = delsetup(&ers))) /* set up the program run */
	  {
		if (EOK == (result = delread(&ers))) /* read in deletion list*/
		  { /* perform the deletions on the multiref */
			if (EOK == (result = expuapply(&ers)))
				result = expun_final(&ers);
		  }
	  }
	expun_finish(&ers);	return EOK;
}

/* end of expunge functionality */

/***************************
	The next few routines enable us to re-bin the data in a MULTIREF.
	It goes through a multiref and determines how many observations are
	present in each input shift.  It then takes all shifts with fewer than
	N observations in it and merges them into the shift with shift value (2)
	smaller than the current one. If the shift-value 2-smaller than the
	current one doesn't exist, the observations in the suspect shift are
	simply deleted.  The "minimum ok observations" value N defaults to 20
	but may be set by the user to some other value.
****************************/

static void exam_refl(MERGEDREF *mrp, MSHWORK *msp)
{ /*	This examines a single reflection and
	does the shift-binning on the first pass */
	register int	i, ishift, shind;
	register ABBROB	*aobp;
	int		tscale;
	SHIFT		*sqp;
	xgulong		ulsh;

	for (i = 0; i < (int)mrp->mr_nob; i++)
	   {
		aobp = &(mrp->mr_ob[i]);
		if (aobp->abo_shift < 0) continue;
		ulsh = LSH(aobp->abo_runno, 24) |
			LSH(aobp->abo_chamno, 16) | aobp->abo_shift;
		shind = (msp->ms_nscales ? find_shift(msp->ms_shlp, ulsh,
			(int)msp->ms_nscales) : -1);
		if (shind < 0) /* flag: indicates shift index not yet found*/
		  {
			tscale = ++msp->ms_nscales;
			shind = ((shind == -tscale) ? 0 : -shind);
			add_shift(msp->ms_shlp, shind, (int)msp->ms_nscales);
			sqp = msp->ms_shlp + shind;
			sqp->runno = aobp->abo_runno;
			sqp->chamno = aobp->abo_chamno;
			sqp->shval = aobp->abo_shift;
			if (shind < (int)(msp->ms_nscales-1))
			  for (ishift = msp->ms_nscales-2; ishift >= shind;
			   ishift--) msp->ms_nobinsh[ishift+1] =
				msp->ms_nobinsh[ishift];
			msp->ms_nobinsh[shind] = 1;
		  }
		else (msp->ms_nobinsh[shind])++;
	  }
}

static int proc_shifts(int minokobs, MSHWORK *msp)
{
	register int	ishift;
	short		oldcham, oldrun, oldshval;
	xgulong		ulsh;
	register SHIFT	*sqp;

	msp->ms_nshin = msp->ms_nshout = msp->ms_nscales;
	fprintf(msp->ms_fpl,
"  shift run chmbr  # obs  shift run chmbr  # obs  shift run chmbr  # obs\n");
	for (ishift = 0; ishift < (int)(msp->ms_nscales); ishift++)
	   {
		sqp = msp->ms_shlp + ishift;
		fprintf(msp->ms_fpl," %6d %3d %5d %6d",
		 sqp->shval, sqp->runno, sqp->chamno, msp->ms_nobinsh[ishift]);
		if (!((ishift + 1) % 3)) fprintf(msp->ms_fpl,"\n");
	   }
	for (ishift = 0; ishift < (int)(msp->ms_nscales); ishift++)
	   {
		sqp = msp->ms_shlp + ishift;	oldshval = sqp->shval;
		oldrun = sqp->runno;		oldcham = sqp->chamno;
		if (msp->ms_nobinsh[ishift] < minokobs) /* are there enough? */
		  { /* ... no: see if we can merge it into something else */
			(msp->ms_nshout)--;
			ulsh = LSH(oldrun, 24) | LSH(oldcham, 16) |
					(oldshval - 2);
			if (0 <= find_shift(msp->ms_shlp, ulsh,
			 (int)msp->ms_nscales))
				msp->ms_newshval[ishift] = oldshval - 2;
			else
			  {
				ulsh = LSH(oldrun, 24) | LSH(oldcham, 16) |
					(oldshval + 2);
				if (0 <= find_shift(msp->ms_shlp, ulsh,
				 (int)msp->ms_nscales))
				 msp->ms_newshval[ishift] = oldshval + 2;
				else	msp->ms_newshval[ishift] = -1;
			  }
			if (msp->ms_newshval[ishift] >= 0)
			 {
				fprintf(msp->ms_fpl,
"Reflections in shift %5d, run %2d, chmbr %1d merged into %5d, %2d, %1d\n",
				 oldshval, oldrun, oldcham,
				 msp->ms_newshval[ishift], oldrun, oldcham);
			  }
			else fprintf(msp->ms_fpl,
"Reflections in shift %5d, run %2d, chmbr %1d will be deleted\n",
				oldshval, oldrun, oldcham);
		  }
		else	msp->ms_newshval[ishift] = oldshval;
		msp->ms_nobinsh[ishift] = 0; /* clear it for recount */
	   }
	if (msp->ms_nshout <= 0)
	  {
		fprintf(stderr,"Tilt! no observations will be written out!\n");
		return 0;
	  }
	return 1;
}

static void adjust_refl(MERGEDREF *mrp, MSHWORK *msp)
{ /*	Alter the reflection appropriately */
	register int	i, j, shind;
	register ABBROB	*aobp;
	int		nobpristart, nobstart;
	xgulong		ulsh;
	char		haslost;
	short		oldshval, lphkl[3];

	nobstart = mrp->mr_nob;	nobpristart = mrp->mr_nobpri;
	msp->ms_nrefin++;	msp->ms_nobin += nobstart;	haslost = 0;
	for (i = 0; i < (int)mrp->mr_nob; i++)
	   {
		aobp = &(mrp->mr_ob[i]);
		oldshval = aobp->abo_shift;
		if (oldshval < 0) continue;
		ulsh = LSH(aobp->abo_runno, 24) |
			LSH(aobp->abo_chamno, 16) | aobp->abo_shift;
		if (0 > (shind = find_shift(msp->ms_shlp, ulsh,
		 (int)msp->ms_nscales)))
		  {
			fprintf(stderr,
	"Warning! no entry on 2nd pass for shift %4d, run %2d, chamber %1d\n",
			aobp->abo_shift, aobp->abo_runno, aobp->abo_chamno);
			aobp->abo_shift = -1;
		  }
		else	aobp->abo_shift = msp->ms_newshval[shind];
		munpack(mrp->mr_hkl, lphkl, (msp->ms_xhp)->xt_intl);
		if (aobp->abo_shift < 0)
		  {
			mrp->mr_nob--;
			if (i < nobpristart) mrp->mr_nobpri--;
			else	mrp->mr_nobbij--;
			for (j = i; j < (int)mrp->mr_nob; j++)
				cpymem((AVOID *)&(mrp->mr_ob[j]),
				 (AVOID *)&(mrp->mr_ob[j+1]), sizeof (ABBROB));
			fprintf(msp->ms_fpl,
 "[%3d,%3d,%3d] in shift %4d, run %2d, chamber %1d deleted\n",
			 lphkl[0], lphkl[1], lphkl[2],
			 oldshval, aobp->abo_runno, aobp->abo_chamno);
			if (!haslost) msp->ms_nreflost++;
			haslost = 1;
			((msp->ms_xhp)->xt_ninteg)--;
		  }
		else if (aobp->abo_shift != oldshval)
		  {
			fprintf(msp->ms_fpl,
 "[%3d,%3d,%3d] in shift %4d, run %2d, chamber %1d moved to %4d, %2d, %1d\n",
			 lphkl[0], lphkl[1], lphkl[2],
			 oldshval, aobp->abo_runno, aobp->abo_chamno,
			 aobp->abo_shift, aobp->abo_runno, aobp->abo_chamno);
			(msp->ms_nobmoved)++;
			ulsh = LSH(aobp->abo_runno, 24) |
				LSH(aobp->abo_chamno, 16) | aobp->abo_shift;
			shind = find_shift(msp->ms_shlp, ulsh,
				(int)msp->ms_nscales);
		  }
		if (shind >= 0) msp->ms_nobinsh[shind]++;
	  }
	msp->ms_nobout += mrp->mr_nob;
	if (mrp->mr_nob)
	  {
		(void)xgwrite((void *)mrp, XG_MERGEDREF, msp->ms_fpmo);
		msp->ms_nrefout++;
	  }
}

static void sum_mrshres(FILE *fp, MSHWORK *msp)
{
	register int	ishift;

	fprintf(fp, "%7d observations of %7d reflections read in\n",
		msp->ms_nobin,  msp->ms_nrefin);
	fprintf(fp, "%7d observations of %7d reflections written out\n",
		msp->ms_nobout, msp->ms_nrefout);
	fprintf(fp,
	 "%7d observations of %7d reflections were deleted\n",
		msp->ms_nobin - msp->ms_nobout, msp->ms_nreflost);
	fprintf(fp, "%7d observations had shift values redefined\n",
		msp->ms_nobmoved);
	fprintf(fp, "%7d entire reflections were deleted altogether\n",
		msp->ms_nrefin - msp->ms_nrefout);
	fprintf(fp, "%7d shifts were read in; %7d shifts were written out\n",
		msp->ms_nshin, msp->ms_nshout);
	if (fp == msp->ms_fpl)
	 {
		fprintf(msp->ms_fpl,
"  shift run chmbr  # obs  shift run chmbr  # obs  shift run chmbr  # obs\n");
		for (ishift = 0; ishift < (int)(msp->ms_nscales); ishift++)
		   {
			fprintf(msp->ms_fpl," %6d %3d %5d %6d",
			 (msp->ms_shlp + ishift)->shval,
			 (msp->ms_shlp + ishift)->runno,
			 (msp->ms_shlp + ishift)->chamno,
			 msp->ms_nobinsh[ishift]);
			if (!((ishift+1)%3)) fprintf(msp->ms_fpl,"\n");
		   }
	 }
}

static int endmerge(int ina, MSHWORK *msp)
{ /*	This frees up some resources before exit */
	if ((stdout != msp->ms_fpl) && (NULL != msp->ms_fpl))
	  {
		(void)fclose(msp->ms_fpl);	msp->ms_fpl = NULL;
	  }
	if (NULL != msp->ms_fpmo)
	  {
		(void)fclose(msp->ms_fpmo);	msp->ms_fpmo = NULL;
	  }
	memzap((AVOID **)&(msp->ms_xhp), "xhp in mergeshift");
	memzap((AVOID **)&(msp->ms_shinrun), "shinrun in mergeshift");
	memzap((AVOID **)&(msp->ms_newshval), "newshval in mergeshift");
	memzap((AVOID **)&(msp->ms_shlp), "msp->ms_shlp in mergeshift");
	globalkill();	return ina;
}

int mergesmain(int argc, char *argv[])
{ /*	Mainline for the shift-merging activity */
	register int	nrun;
	int		minok;
	FILE		*fpmi;
	DATARUN		rhd;
	MERGEDREF	mr;
	time_t		tl0;
	static char	rundate[26];
	static MSHWORK	mss;
	static char	*flist[] = {"multiref", "mrefout", "xlog", NULL};
	static char	ftypes[] = {1, 2, 2, 0 };

	tl0 = NOW;	strcpy(rundate, ctime(&tl0));
	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	if (qerror((AVOID *)(fpmi = xgopen("multiref", "r")),
		"open input multiref")) return EOPENERR;
	if (argc < 2 || 0 >= (minok = atoi(argv[1]))) minok = DEFAULT_MIN_OK;
		/* read the multiref headers */
	if (NULL == (AVOID *)(mss.ms_xhp = (CRYSTAL *)callop(1,
	 sizeof (CRYSTAL), "xhp in mergeshift"))) return EALLOC;
	if (verror(0 < xgread((void *)(mss.ms_xhp), XG_CRYSTAL, fpmi),
		"read crystal header from multiref")) return EREADERR;
	if (NULL == (AVOID *)(mss.ms_shinrun = (short *)callop((unsigned)
	 (mss.ms_xhp)->xt_nruns, sizeof (short), "shinrun in mergeshift")))
		return EALLOC;
	if (NULL == (AVOID *)(mss.ms_newshval = (short *)callop(
	 (unsigned)(mss.ms_xhp)->xt_nshifts, sizeof (short),
		"newshval in mergeshift"))) return EALLOC;
	for (nrun = 0; nrun < (mss.ms_xhp)->xt_nruns; nrun++)
	   {
		if (verror(0 < xgread((void *)&rhd, XG_DATARUN, fpmi),
			"read datarun header from multiref"))
				return endmerge(EREADERR, &mss);
		mss.ms_shinrun[nrun] = rhd.run_nshifts;
	   }
	if (NULL == (mss.ms_shlp = init_shift((int)
	 ((mss.ms_xhp)->xt_nshifts)))) return endmerge(EALLOC, &mss);
	if (NULL == (mss.ms_fpl = propen(trname("xlog"))))
		mss.ms_fpl = stdout;
	ftypes[2] = (mss.ms_fpl == stdout ? 2 : 0);
	header(stdout, REDUCTION, flist, ftypes);
	printf(
	 "Shifts with fewer than %d obs will be merged into their neighbors\n",
		minok);
	if (stdout != mss.ms_fpl)
	  {
		header(mss.ms_fpl, REDUCTION, flist, ftypes);
		fprintf(mss.ms_fpl,
	 "Shifts with fewer than %d obs will be merged into their neighbors\n",
			minok);
	  }
/* now go through the data, noting the shift assignments */
	while (0 < xgread((void *)&mr, XG_MERGEDREF, fpmi))
		exam_refl(&mr, &mss);
/* now go through the list of shifts and figure out how many have
	too few observations in them */
	if (0 >= proc_shifts(minok, &mss)) endmerge(EOK, &mss);
/* if we didn't do any reassignments, we don't write an output */
	if (mss.ms_nshin == mss.ms_nshout)
	  {
		printf("No reassignments performed: no output written\n");
		fprintf(mss.ms_fpl,
		 "No reassignments performed: no output written\n");
		return endmerge(EOK, &mss);
	  }
/* now go back through the input multiref and write out a version
	identical to it except with the shift numbers revised */
	if (verror(1 + fseek(fpmi, 0L, 0), "rewind input multiref"))
		return endmerge(ESEEKERR, &mss);
	if (verror(0 < xgread((void *)(mss.ms_xhp), XG_CRYSTAL, fpmi),
		"read crystal header from multiref"))
	  {
		if (stdout != mss.ms_fpl) (void)fclose(mss.ms_fpl);
		mss.ms_fpl = NULL;	return EREADERR;
	  }
	(mss.ms_xhp)->xt_nshifts = mss.ms_nshout;
	if (qerror((AVOID *)(mss.ms_fpmo = xgopen("mrefout", "w")),
	 "open output multiref"))
	  {
		if (stdout != mss.ms_fpl) (void)fclose(mss.ms_fpl);
		mss.ms_fpl = NULL;	return EOPENERR;
	  }
	if (verror(0 < xgwrite((void *)(mss.ms_xhp), XG_CRYSTAL, mss.ms_fpmo),
		"write crystal header to multiref"))
	  {
		if (stdout != mss.ms_fpl) (void)fclose(mss.ms_fpl);
		mss.ms_fpl = NULL;	return EWRITERR;
	  }
	for (nrun = 0; nrun < (mss.ms_xhp)->xt_nruns; nrun++)
	   {
		if (verror(0 < xgread((void *)&rhd, XG_DATARUN, fpmi),
			"read datarun header from multiref"))
		  {
			if (stdout != mss.ms_fpl) (void)fclose(mss.ms_fpl);
			mss.ms_fpl = NULL;	return EREADERR;
		  }
		rhd.run_nshifts = mss.ms_shinrun[nrun];
		if (verror(0 < xgwrite((void *)&rhd, XG_DATARUN, mss.ms_fpmo),
			"write datarun header to multiref"))
		  {
			if (stdout != mss.ms_fpl) (void)fclose(mss.ms_fpl);
			mss.ms_fpl = NULL;	return EWRITERR;
		  }
	   }
	mss.ms_nobin = mss.ms_nobout = mss.ms_nobmoved = 0;
	mss.ms_nrefin = mss.ms_nrefout = mss.ms_nreflost = 0;
	while (0 < xgread((void *)&mr, XG_MERGEDREF, fpmi)) /* read refls */
		adjust_refl(&mr, &mss);	/* fix 'em and write 'em out */
	(void)fclose(fpmi);	fpmi = NULL;
	if (mss.ms_nrefout != mss.ms_nrefin)
	  {
		if (verror(1 + fseek(mss.ms_fpmo, 0L, 0),
		 "rewind output multiref")) return endmerge(ESEEKERR, &mss);
		(mss.ms_xhp)->xt_nmerged = mss.ms_nrefout;
		if (verror(0 < xgwrite((void *)(mss.ms_xhp), XG_CRYSTAL,
		 mss.ms_fpmo), "write crystal header to multiref"))
			return endmerge(EWRITERR, &mss);
	  }
	sum_mrshres(stdout, &mss);
	if (stdout != mss.ms_fpl)
	  {
		sum_mrshres(mss.ms_fpl, &mss);	(void)fclose(mss.ms_fpl);
	  }
	return endmerge(EOK, &mss);
}

/* end of mergeshift functionality */

/* end of mreftools.c */
