/* peaktools.c -- filter various things out of CENTROIDS files

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	28 Nov 2006: tr_to_scatt: replace cc_q21[3][3] with xp->xq21
	17 May 2006: merge cut_{cen,close,weak}.c into peaktools.c
	 6 Oct 2005: housekeeping
	11 May 2005: check return status from xgjournal
	11 Apr 2005: change name from cutcen.c to cut_cen.c to prevent overlap
	 6 Sep 2004: cosmetic change in cutcen.c
	28 Jun 2005: cutclosemain(): typecast pointer difference on printout
	17 Jun 2005: tr_to_scat(): pix2cm() takes frame-number argument
	11 Apr 2005: rename cutclose.c to cut_close.c to prevent overlap
	11 Apr 2005: rename cutweak.c to cut_weak.c to prevent overlap
	 6 Sep 2004: cutweak: cosmetic change
	18 Aug 2004: CUTCLOSERUN struct type: eliminates static globals
	24 Jan 2004: cr is now local in 2 routines rather than a static global
	24 Jan 2004: cutclose: most static global variables converted
		to locals; ignore VSPLINE return value from gspline()
	21 Jan 2004: license message inserted
	 7 Jan 2004: cutclose: external globals replaced with static globals,
	 including elements of a REFPARS structure called from the mainline
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	 5 Jan 2003: reorganized source-file header
	 1 Nov 2002: cutclose: reformatting of stdout summary of operations
	25 Mar 2002: removed MSI attachments
	 5 Oct 1994: use xgread and xgwrite for I/O on TROID data
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: initial X-GEN version of cutclose
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version
	 8 Jun 1991: initial code for cutweak, based on `cutcen.c'
	 5 Jan 1990: (?) initial code for cutcen

   Purpose: various functionalities that read a CENTROIDS file and write out
	CENOUT containing only spots that satisfy certain specific criteria:
	cut_cen		retains spots within a specified frame range.
	cut_close	retains spots that are distant from one another
	cut_weak	retain spots with intensity above specified cutoff
   Cross references: ../prerefine/spot*.c
***************************************************************************/
#include	"xgen.h"
#define	BADSPACING	9999.

extern int cutcenmain(int argc, char *argv[]);
extern int cutclosemain(int argc, char *argv[]);
extern int cutweakmain(int argc, char *argv[]);

/*****************************************
	cut_cen: read a CENTROIDS and write CENOUT,
	retaining reflections within a specified range of images.
	The first argument to this program is the minimum frame number output;
	the second is the maximum frame number output, if you specify it.  Thus
 		cutcen 12 48
	will write out only the data between frames 12 and 48 inclusive, and
 		cutcen 30
	will write out all data from frame 30 onwards.
******************************************/
int cutcenmain(int argc, char *argv[])
{ /* The mainline for cuttinf off a CENTROIDS file by frame number */
	int		incount, maxfm, minfm, outcount, sizr;
	FILE		*fpi, *fpo;
	TROID		cr;
	static char	*fnames[] = {"centroids", "cenout", NULL };
	static char	ftypes[] = {1, 2, 0 };

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	header(stdout, PROCESSING, fnames, ftypes);
	minfm = 0;	maxfm = 9999;
	if (argc > 1)
	  {
		minfm = atoi(argv[1]); if (argc > 2) maxfm = atoi(argv[2]);
	  }
	if (minfm > maxfm)
	  {
		incount = minfm; minfm = maxfm; maxfm = incount;
	  }
	printf(
 "Reflections will be output if the frame number is between %d and %d\n",
		minfm, maxfm);
	if (qerror((AVOID *)(fpi = xgopen("centroids", "r")),
	 "open input centroids file")) return EOPENERR;
	if (qerror((AVOID *)(fpo = xgopen("cenout", "w")),
	 "open output centroids file"))
	  {
		(void)fclose(fpi);	return EOPENERR;
	  }
	incount = outcount = 0;
	while (0 < xgread((void *)&cr, XG_TROID, fpi))
	     {
		incount++;
		if (minfm <= cr.frame0 && cr.frame0 <= maxfm)
		  {
			sizr = SIZREF + ((unsigned)cr.nobk) * sizeof (float);
			if (verror(sizr == xgwrite((void *)&cr, XG_TROID, fpo),
			 "write refl to output centroids file"))
			  {
				(void)fclose(fpi);	(void)fclose(fpo);
				return EWRITERR;
			  }
			outcount++;
		  }
	     }
	printf(" %3d reflections read in; %3d written out\n",
		incount, outcount);
	(void)fclose(fpi);	(void)fclose(fpo);
	globalkill();	return EOK;
}

/* end of cut_cen functionality */

/*********************************
	cut_close retains only reflections that are distant from
	one another in reciprocal space. That is, we omit spots that lie
	closer together in reciprocal space than a specified plane spacing.
	The user specifies the cutoff plane spacing as a numerical argument:
		cutclose 130
	will write out only reflections with no neighbors implying a plane
	spacing of 130A or greater. The default plane spacing is 100 A.
***********************************/

static int tr_to_scat(TROID *tr, double *sv, REFPARS *xp)
{ /*	This returns the diffraction vector associated with
	the TROID `tr'. Any problems that arise in the computation
	result in a nonzero return value; a good computation returns 0. */
	int	i, j, jfr;
	float	xcm, ycm;
	double	rfi, rni, rni2, xin, yin, q3[3][3], sl[3], ss[3];

	jfr = tr->frame0 + 0.49999;
	if (1 != pix2cm(tr->xto, tr->yto, &xcm, &ycm, jfr)) return EBADVALUE;
	xin = xcm - xp->xx0; yin = ycm - xp->xy0; /*mainbeam*/
	rni2 = xin * xin + yin * yin + xp->xpltd2;	rni = sqrt(rni2);
	ss[0] = xp->xplated / rni; ss[1] = xin / rni; ss[2] = yin / rni;
	/* rotate this by the detector tilt and swing-angle matrices */
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		q3[i][j] = xp->xq21[i][j];
	dmm31(sl, q3, ss);	/* gifs the local scattering vector */
	sl[0] += 1.;	/* correct the beam-direction component */
	rfi = pirange(xp->xfrmsize * (tr->frame0 + 0.5));
	q3[0][0] = q3[1][1] = cos(rfi);	q3[0][1] = -sin(rfi);
	q3[1][0] = -q3[0][1];		q3[2][2] = 1.;
	q3[0][2] = q3[1][2] = q3[2][0] = q3[2][1] = 0.;
	dmm31(sv, q3, sl);	return EOK;
}

static double rdspace(TROID *c0, TROID *c1, REFPARS *xp)
{ /*	This computes the D spacing associated with a pair of TROIDs.
	Any errors that arise from attempting to compute it result in
	a return of 9999. Angstroms, which should result in the refls
	being barred from the output list */
	int	i;
	double	rspa, tspa, s0[3], s1[3];

	if ((EOK != tr_to_scat(c0, s0, xp)) ||
	 (EOK != tr_to_scat(c1, s1, xp))) return BADSPACING;
	for (rspa = i = 0; i < 3; i++)
	   {
		tspa = s0[i] - s1[i];	rspa += tspa * tspa;
	   }
	return (rspa < DFUZZ) ? BADSPACING : xp->xlambda / sqrt(rspa);
}

static REFPARS *ccurrpars(void)
{ /* reserve memory for working vectors and then read in parameters.
  We return a pointer to the REFPARS structure unless we fail,
  in which case we return NULL */
	int		nbr, retv;
	REFPARS		*lxp;
	static char	lxpmsg[] = "lxp in ccurrpars";

	if (NULL == (lxp = (REFPARS *)callop(1, sizeof (REFPARS), lxpmsg)))
		return NULL;
	if (EOK != (retv = refparin(lxp, &nbr)))
	  {
		memzap((AVOID **)&lxp, lxpmsg);	return NULL;
	  }
	makeq21(lxp); (void)gspline(5, &retv);	return lxp;
}

int cutclosemain(int argc, char *argv[])
{ /* Mainline for eliminating closely spaced refls in a CENTROIDS file */
	int		clpair, incount, outcount;
	double		minspacing, space, spec_space, ll[3];
	FILE		*fpi, *fpo;
	TROID		cr, *crb, *cre, *crp, *crq;
	REFPARS		*cxp;
	static char	*fnames[] = {"centroids", "uparams", "cenout", NULL };
	static char	ftypes[] = {1, 1, 2, 0 };

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	if (NULL == (cxp = ccurrpars())) return EREADERR;
	header(stdout, PROCESSING, fnames, ftypes);
	minspacing = 100.;
	if ((argc > 1) && (0. <= (spec_space = atof(argv[1]))))
		minspacing = spec_space;
	printf(
 "Reflections will be output only if no neighbors approach within %7.2f A\n",
		minspacing);
	if (qerror((AVOID *)(fpi = xgopen("centroids", "r")),
	 "open input centroids file")) return EOPENERR;
	incount = outcount = 0;
	/* read 'em in to see how many there are */
	while (0 < xgread((void *)&cr, XG_TROID, fpi)) incount++;
	printf(" %6d reflections read in\n", incount);
	/* reserve the memory for the list of inputs */
	if (NULL == (crb = (TROID *)callop(incount+4, sizeof (TROID),
	 "crb in cutclose")))
	  {
		(void)fclose(fpi);	fpi = NULL;	return EALLOC;
	  }
	if (-1 == fseek(fpi, 0L, 0)) return ESEEKERR;
	/* now read 'em in for real.  Only keep the good ones. */
	for (crp = crb; 0 < xgread((void *)&cr, XG_TROID, fpi); )
	   if (EOK == tr_to_scat(&cr, ll, cxp))
	     {
		cpymem((void *)crp, (void *)&cr, sizeof (TROID)); crp++;
	     }
	(void)fclose(fpi);	fpi = NULL;
	printf(" %6d of the reflections are good\n", (int)(crp - crb));
	printf(
	 "starting & ending refls: [%5.1f,%5.1f,%5.1f], [%5.1f,%5.1f,%5.1f]\n",
		crb->xto, crb->yto, crb->frame0, (crp-1)->xto,
		(crp-1)->yto, (crp-1)->frame0);
	cre = crp;
	/* open the output file */
	if (qerror((AVOID *)(fpo = xgopen("cenout", "w")),
	 "open output centroids file")) return EOPENERR;
	/* go through the input list and compute the D spacing associated
	 with each pair of spots.  Any spot with an excessively-near
	 neighbor is eliminated. */
	for (crp = crb, clpair = 0; crp < cre; crp++)
	   {
		for (crq = crp + 1; crq < cre; crq++)
		   {
			if ((space = rdspace(crp, crq, cxp)) > minspacing)
			  {
				crp->totinten = -32767.;
				crq->totinten = -32767.;
				if (!(clpair % 10)) printf(
	"Spacing = %8.2f between [%5.1f,%5.1f,%5.1f],[%5.1f,%5.1f,%5.1f]\n",
				 space, crp->xto, crp->yto, crp->frame0,
				 crq->xto, crq->yto, crq->frame0);
				clpair++;
			  }
		   }
		if (crp->totinten > -32766.)
		  {
			if (0 >= xgwrite((void *)crp, XG_TROID, fpo))
			  {
				(void)fclose(fpo);	return EWRITERR;
			  }
			outcount++;
		  }
	   }
	(void)fclose(fpo);	fpo = NULL;
	printf(" %3d reflections read in; %3d written out\n",
		incount, outcount);
	memzap((AVOID **)&cxp, "cxp in cutclose");
	globalkill();	return EOK;
}

/* end of cut_close functionality */

/***********************************
	cut_weak: Read CENTROIDS, write CENOUT, omitting weak reflections
	If `cut_weak' is invoked with no arguments, only the spots with
	 intensity > 5000 will be output; if it's invoked as
		cutweak 275
	 then spots with I>275 will be written out.
*************************************************************************/

int cutweakmain(int argc, char *argv[])
{ /* The mainline for weeding out weak reflections from a CENTROIDS file */
	int		incount, outcount, sizr;
	double		mininten, spec_inten;
	FILE		*fpi, *fpo;
	TROID		cr;
	static char	*fnames[] = {"centroids", "cenout", NULL };
	static char	ftypes[] = {1, 2, 0 };

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	header(stdout, PROCESSING, fnames, ftypes);
	mininten = 5000;
	if ((argc > 1) && (0. <= (spec_inten = atof(argv[1]))))
		mininten = spec_inten;
	printf("Reflections will be output if their intensity is > %8.2f\n",
		mininten);
	if (qerror((AVOID *)(fpi = xgopen("centroids", "r")),
	 "open input centroids file")) return EOPENERR;
	if (qerror((AVOID *)(fpo = xgopen("cenout", "w")),
	 "open output centroids file"))
	  {
		(void)fclose(fpi);	return EOPENERR;
	  }
	incount = outcount = 0;
	while (0 < xgread((void *)&cr, XG_TROID, fpi))
	     {
		incount++;
		if (mininten < cr.totinten)
		  {
			sizr = SIZREF + ((unsigned)cr.nobk) * sizeof (float);
			if (verror(sizr == xgwrite((void *)&cr, XG_TROID, fpo),
			 "write refl to output centroids file"))
			  {
				(void)fclose(fpi);	(void)fclose(fpo);
				return EWRITERR;
			  }
			outcount++;
		  }
	     }
	printf(" %3d reflections read in; %3d written out\n",
		incount, outcount);
	(void)fclose(fpi);	(void)fclose(fpo);
	globalkill();	return EOK;
}

/* end of cut_weak functionality */

/* end of peaktools.c */
