/* ureftools.c -- tools for manipulating and printing UREFLS files.

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	28 Nov 2006: replace cup->cu_q21 with xp->xq21 in xyomtoh
	24 Jun 2006: xgenpred.h incorporates xgen.h, so we just need one.
	16 May 2006: merge cross_u, ifilter, predtourf, raw2urf, urf2d
	15 Mar 2006: predtourfmain: 2 lurking bugs fixed, using xgread & xgwrite
	31 Dec 2005: inputobs: avoid qsorting if fewer than 2 elements
	10 Oct 2005: raw2urfmain: eliminate static globals;
		create RAW2WORK structures
	 7 Oct 2005: predtourfmain: I/sigma is partly a function of sin(theta)
	 6 Oct 2005: housekeeping in several functions
	29 Jun 2005: summ_ifilrun(): formats are %d, not %ld now
	29 Jun 2005: raw2urfmain(): sscanf formats are %d, not %ld now
	22 Jun 2005: many routines declared static
	17 Jun 2005: in xyomtoh(): pix2cm() takes frame-number argument
	11 May 2005: check return status from xgjournal
	11 May 2005: urf2dmain: correct out_xh call
	11 Apr 2005: change name from crossu.c to cross_u.c to prevent overlap
	13 Sep 2004: remove rot9rt() call after xeflush() in crossumain();
		correct OBSREF structure references in crossumain
	 6 Sep 2004: predtourfmain: cosmetic change
	28 Aug 2004: routines used only here are declared as static;cosmetics
	18 Aug 2004: urf2dmain:
		replace static globals with U2DRUN structure elements
	18 Aug 2004: CROSSURUN structure type: allows elimination of globals
	 2 Aug 2004: add new arg to find_dimension()
	27 Jul 2004: raw2urfmain, predtourfmain, urf2dmain: etc.:
		use dimscale() to get the (X,Y) scales
	21 Jun 2004: initial version of predtourfmain,
		based vaguely on some other codes
	22 Apr 2004: urf2dmain created, based vaguely on utoshel.c
	25 Jan 2004: raw2urfmain():
		a few static globals converted into local variables
	24 Jan 2004: crossuset(): xgopen(fpp) omitted since it's internal to
		refparin(); numerous static globals converted to local variables
	21 Jan 2004: license message inserted and initial comments reordered
	 7 Jan 2004: static globals replace extern globals some from
		find_dimension; calls to make_boundaries() updated
	19 Sep 2003: summ_ifilrun() & raw2urfmain:
		use %ld formats for xglong variables
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	13 Aug 2003: correct spelling of globalkill()
	 5 Jan 2003: reorganized source-file header
	 1 Nov 2002: ifilter: trivial reformatting of stdout and logfile output
	15 Oct 2002: reformat logfile output statements
	25 Mar 2002: removed MSI attachments
	17 Jun 1998: initial version of ifilter
	 5 Oct 1994: use xgread and xgwrite on OBSREF data
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: a few I/O changes and global variables moved around
	 7 Dec 1993: initial X-GEN versions of raw2urfmain and cross_u
	25 Jun 1991: initial code for cross_u,
		partly borrowed from parts of `refine'

   Purpose: conversions and printouts for UREFLS files.
	Specifically,
	cross_u identifes reflections in a UREFLS file that overlap with
	 spots that actually came from a different lattice, as specified
	 in a separate UPARAMS orientation matrix.
	ifilter reads in a UREFLS file and writes out another in which
	 reflections have been eliminated in shells if I < Imin or I > Imax
	predtourf creates a fake UREFLS file from the output of `predict'
	raw2urf converts from SAINT's ".raw" format to a UREFLS format
	urf2d converts from XGEN's UREFLS format to D*TREK format
   Cross references: ../integrate/int*.c, ../xglib/xgpredict.c
*************************************************************************/
#include	"xgenpred.h"

#define		ENOUGH_ITEMS		38
#define		UNSPEC_MAX	(1.e7)

 typedef struct { /* Structure type for managing the conversion process */
		int		cu_flags, cu_nrin, cu_nrout;
		OBSREF		*cu_urp;
		REFPARS		*cu_xp;
		PREDWORK	*cu_pwp;
		FILE		*cu_fpi, *cu_fpo, *cu_fpl;
		double		cu_oh[3], cu_xsca, cu_ysca;
		} CROSSURUN;

 typedef struct {
		double		pu_f000, pu_ios000, pu_xscal, pu_yscal;
		PREDREF		*pu_prp;
		REFPARS		*pu_xp;
		BOUNDARY	*pu_bdp;
		OBSREF		*pu_urp;
		} P2UWORK;

 typedef struct {
		double	fsh_resma, fsh_imi, fsh_ima, fsh_pkrima;
		xglong	fsh_nok, fsh_nlti, fsh_ngti;
		xglong	fsh_ngtfm, fsh_nboth, fsh_nobounds;
		} URSHEL;

 typedef struct {
		xglong	ra_hkl[3];	/* indices of reflection */
		float	ra_inten, ra_sigma; /* intensity, sigma */
		xglong	ra_batch;	/* batch number */
		float	ra_sxi, ra_sxo;	/* direction cosines for x*/
		float	ra_syi, ra_syo;	/* direction cosines for y*/
		float	ra_szi, ra_szo;	/* direction cosines for z*/
		xglong	ra_status;	/* status word */
		float	ra_xyzob[3], ra_xyzpred[3]; /* centroids */
		float	ra_lpetc, ra_corr, ra_hrsaccum; /* LP, etc.*/
		float	ra_2theta, ra_scanabs;	/* 2theta, abs scanangle*/
		xgulong	ra_scanaxis, ra_stl10k; /* axis#,sintheta*/
		xglong	ra_rawctpbg, ra_rawbg;	/* rawcount, bgnd */
		float	ra_scanrel;		/* relative scanangle*/
		xglong	ra_xtalno;		/* crystal number */
		float	ra_fracfill;		/* fraction profile*/
		xglong	ra_sortkey, ra_multiple; /* sort key,multiplicity*/
		float	ra_lortz, ra_xpixcor, ra_ypixcor; /* various */
		float	ra_chi, ra_nonmove;	/* chi, nonmoving motor*/
	       } SAINTRAW;

 typedef struct {
		int		r2_ht, r2_wid;
		double		r2_scl[4];
		BOUNDARY	*r2_bdp;
		} RAW2WORK;

 typedef struct {
		int	xd_hkl[3], xd_detno, xd_nonunf;
		int	xd_packh, xd_batch, xd_erflag, xd_scanid;
		float	xd_inten, xd_sigma, xd_xcalc, xd_ycalc;
		float	xd_crotmid, xd_crotstart, xd_crotend, xd_crotwid;
		float	xd_polarz, xd_lor, xd_oblique, xd_resol;
		float	xd_xobs, xd_yobs, xd_orotmid, xd_orotwid, xd_orotsig;
		float	xd_bgnd, xd_bgsigma, xd_svec[3], xd_svec0[3];
		int	xd_sbatch;
		} XTAL_DTREK;

 typedef struct {
		double	u2_mos, u2_omoff, u2_xsca, u2_ysca;
		FILE	*u2_fpo;
		OBSREF	u2_urf;
		REFPARS	*u2_xp;
		} U2DRUN;

extern long random(void);

extern int crossumain(int argc, char *argv[]);
extern int ifiltermain(int argc, char *argv[]);
extern int predtourfmain(int argc, char *argv[]);
extern int raw2urfmain(int argc, char *argv[]);
extern int urf2dmain(int argc, char *argv[]);

/***********************
	The next few routines define the "cross_u" functionality,
	which eliminates overlapped refls in severely twinned crystals.
	It reads a UREFLS file and a UPARAMS file that is NOT associated
	with the UREFLS file.  For each reflection in the UREFLS file the
	program calculates the indices associated with the (X,Y,omega)
	according to the "foreign" UPARAMS file.  If these indices are close
	to integers the reflection is NOT written to the output;
	if they aren't, the reflection IS written to the output.
	The purpose of this algorithm is to deal with the situation where two
	crystals (or crystallites) are diffracting in the same experiment.
	The reflections in the UREFLS file for the first crystallite are then
	being compared with the lattice associated with the second crystallite.
	When this second lattice happens to overlap with the first, we exclude
	the reflections from the output because they may actually be coming
	from the other lattice, or from both.
	 The cutoff in integerness is determined from the integerness error
	value in the "foreign" UPARAMS file.
	Typically one would run this with the UREFLS file for one lattice
	 and the UPARAMS file for the other lattice, and then reverse that.
**********************************/

static void xyomtoh(CROSSURUN *cup)
{ /*	This reads the (x,y,omega) of a reflection and uses the orientation
	matrix in the UPARAMS file to compute its floating-point indices
	according to that UPARAMS file. */
	int	i, j, jfr;
	float	xcm, ycm;
	double	crsp, rni, rni2, rsp, srsp, xin, yin;
	double	rs[3], rsl[3], ss[3], tq[3][3];
	REFPARS	*xp;
	OBSREF	*urp;

	urp = cup->cu_urp;	/* identify the reflection */
	xin = ((double)urp->obs_xob + 0.49999) / cup->cu_xsca; /* observed X */
	yin = ((double)urp->obs_yob + 0.49999) / cup->cu_ysca; /* observed Y */
	jfr = urp->obs_omob + 0.49999;
	/* if the pix2cm stuff fails, the refl is untrustworthy: */
	if (1 != pix2cm((float)xin, (float)yin, &xcm, &ycm, jfr))
	  {
		cup->cu_oh[0] = cup->cu_oh[1] = cup->cu_oh[2] = 127;
		return;
	  }
	xp = cup->cu_xp; /* parameters for alternate lattice */
	xcm -= xp->xx0;	ycm -= xp->xy0;	/* correct main beam */
	rni2 = xcm * xcm + ycm * ycm + xp->xpltd2;	rni = sqrt(rni2);
	ss[0] = xp->xplated / rni;
	ss[1] = xcm / rni;	ss[2] = ycm / rni;
/* rotate this by the detector tilt and swing-angle matrices */
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		tq[i][j] = xp->xq21[i][j];
	dmm31(rsl, tq, ss);	/* gifs the local scattering vector */
	rsl[0] += 1.;	/* correct the beam-direction component */
/* We rotate the definition of the scattering vector so that all the scattering
   vectors are indexed to the reference position of the spindle.  */
	rsp = pirange(xp->xfrmsize * (urp->obs_omob + 0.5)); crsp = cos(rsp);
	srsp = sin(rsp);	rs[0] = crsp * rsl[0] - srsp * rsl[1];
	rs[1] = srsp * rsl[0] + crsp * rsl[1]; rs[2] = rsl[2];
/* determine the non-integer indices of the reflection by left-multiplying
 the diffraction vector by the real-space unit cell matrix */
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		tq[i][j] = xp->xareal[i][j];
	dmm31(cup->cu_oh, tq, rs);
}

static void culogbad(CROSSURUN *cup, PREDREF *clospred)
{ /*	This prints a (potentially overlapped) reflection to the logfile */
	fprintf(cup->cu_fpl,
	 " %3d %3d %3d%7.1f%7.1f%7.1f%7.1f %5.1lf %5.1lf %5.1lf",
	 (cup->cu_urp)->obs_hkl[0], (cup->cu_urp)->obs_hkl[1],
	 (cup->cu_urp)->obs_hkl[2],
	 ((double)(cup->cu_urp)->obs_xob) / cup->cu_xsca,
	 ((double)(cup->cu_urp)->obs_yob) / cup->cu_ysca,
	 (cup->cu_urp)->obs_omob, (cup->cu_urp)->obs_ii,
	 cup->cu_oh[0], cup->cu_oh[1], cup->cu_oh[2]);
	if (!(cup->cu_flags)) fprintf(cup->cu_fpl, "%7.1f%7.1f%7.1f",
	 clospred->pxt, clospred->pyt, clospred->pphi);
	fprintf(cup->cu_fpl, "\n");
}

static int ihtoxyo(int loh[3], CROSSURUN *cup)
{ /* This uses the integerized (hkl) value in loh[] to determine the predicted
	(X,Y,omega) for the reflection according to the alternate lattice.
	It then compares that (X,Y,omega) with the observed value. If
	it's within the (X,Y,omega) error limits for either intersection,
	it returns a nonzero value; if they're both far away, it returns 0. */
	int		far0, far1, i;
	double		xin, yin, zin;
	REFPARS		*xp;
	PREDWORK	*pwp;
	PREDREF		*pwc;

	/* establish pointers to the 2 possible predicted reflections */
	pwp= cup->cu_pwp;	pwp->po_prp = pwp->po_prp0;
	xp = cup->cu_xp;
	/* populate the hkl vector */
	for (i = 0; i < 3; i++) pwp->po_ph[i] = loh[i];
	/* compute the 2 possible predicted reflections */
	if ((0 >= (i = pfindref(pwp))) || (i > 3)) return 0;
	/* remember where we originally observed the reflection */
	xin = ((double)((cup->cu_urp)->obs_xob) + 0.49999) / cup->cu_xsca;
	yin = ((double)((cup->cu_urp)->obs_yob) + 0.49999) / cup->cu_ysca;
	zin = (cup->cu_urp)->obs_omob + 0.5;
	far0 = far1 = 3; /* first we assume both intersections are far away */
	if (i & 1) /* test the first intersection if it's viable */
	  {
		pwc = pwp->po_prp0;
		if (xp->xerrx > fabs(xin - pwc->pxt)) far0--;
		if (xp->xerry > fabs(yin - pwc->pyt)) far0--;
		if (xp->xerrphi > fabs(zin - pwc->pphi)) far0--;
		if (!far0) culogbad(cup, pwc);
	  }
	if (i & 2) /* test the second intersection if it's viable */
	  {
		pwc = (i == 2) ? pwp->po_prp0 : pwp->po_prp0 + 1;
		if (xp->xerrx > fabs(xin - pwc->pxt)) far1--;
		if (xp->xerry > fabs(yin - pwc->pyt)) far1--;
		if (xp->xerrphi > fabs(zin - pwc->pphi)) far1--;
		if (!far1) culogbad(cup, pwc);
	  }
	/* at this point far0 will be zero if the zeroth predicted
	 position is very close to the observed position;
	 far1 will be zero if the other one is */
	i = 0; if (!far0) i = 1; if (!far1) i |= 2; return i;
}

static int startcupreds(CROSSURUN *cup)
{ /*	This populates the prediction-work structure for this procedure. */
	int		i, j;
	PREDWORK	*pwp;

	pwp = cup->cu_pwp;
	pwp->po_xp = cup->cu_xp;	/* identify UPARAMS */
	pwp->po_miny2sq = DFUZZ;	/* smallest |sy|^2 value */
	for (i = 0; i < 3; i++) /* initialize pwp->po_p[0] and po_ih */
	   {
		for (j = 0; j < 3; j++)
		  pwp->po_p[0][i][j] = (cup->cu_xp)->xrec[i][j];
		pwp->po_ih[i] = i;
	   }
	/* starting and ending mu values */
	pwp->po_mumin = -PI;	pwp->po_mumax = PI;	pwp->po_rangecen = 0.;
	return prpar(pwp);	/* other initializations in this routine*/
}

static int crossuset(int ac, char *av[], CROSSURUN *cup)
{ /*	This sets up for a run of `crossu' */
	int		i, j, locht, locwid, nbr;
	char		**dcp, **flags;
	static char	*fnames[] = {"uparams", "urefls", "uspline",
			 "xlog", "urefout", NULL } ;
	static char	ftypes[] = { 1, 1, 1, 2, 2, 0 } ;
	VSPLINE		*lsp;

	if (EHELP == xgjournal(ac, av)) return EOK; /* journal the effort */
	cup->cu_flags = 0;	/* default is the pickier algorithm */
	flags = flagproc(ac, av, "NO"); /* choose the algorithm */
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	 switch (**dcp)
	   {
		case	'O':	cup->cu_flags = 1; break; /* old method */
		case	'N':	cup->cu_flags = 0; break; /* new method */
		default:	break;
	   }
	if (qerror((AVOID *)(cup->cu_fpi = xgopen("urefls", "r")),
		"open input UREFLS file")) return EOPENERR;
	if (qerror((AVOID *)(cup->cu_fpo = xgopen("urefout", "w")),
		"open output UREFLS file")) return EOPENERR;
	if (NULL == (cup->cu_xp = (REFPARS *)callop(1, sizeof (REFPARS),
	 "cup->cu_xp in crossumain"))) return EALLOC;
	if (NULL == (cup->cu_urp = (OBSREF *)callop(1, sizeof (OBSREF),
	 "cup->cu_urp in crossumain"))) return EALLOC;
	if (NULL == (cup->cu_pwp = (PREDWORK *)callop(1, sizeof (PREDWORK),
	 "cup->cu_pwp in crossumain"))) return EALLOC;
	if (EOK != (i = refparin(cup->cu_xp, &nbr))) return i;
		/* read in input spline information */
	if (NULL == (cup->cu_fpl = propen("XLOG"))) cup->cu_fpl = stdout;
	ftypes[3] = (cup->cu_fpl == stdout ? 0 : 2);
	header(stdout, PROCESSING, fnames, ftypes);
	if (cup->cu_fpl != stdout)
		header(cup->cu_fpl , PROCESSING, fnames, ftypes);
	fprintf(cup->cu_fpl,
	 " Reflections rejected because they lie close to lattice points\n");
	fprintf(cup->cu_fpl,
	 " for the orientation defined in this UPARAMS file:\n");
	fprintf(cup->cu_fpl,
	 "     Input                                           Output\n");
	fprintf(cup->cu_fpl,
	 "   h   k   l X(pix) Y(pix)  frame Intens     h     k     l");
	if (!(cup->cu_flags)) fprintf(cup->cu_fpl, " X(pix) Y(pix)  frame");
	fprintf(cup->cu_fpl, "\n");
	xeflush(cup->cu_fpl);
	/* we used to have a rot9rt step here; it's already
	 been done in refparin() now. */
	/* rotations for detector tilt and swing; rescale recip-sp unit cell */
	makeq21(cup->cu_xp);
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		(cup->cu_xp)->xrec[i][j] *= (cup->cu_xp)->xlambda;
	if (!(cup->cu_flags)) /* set up prediction parameters */
	  {
		if (EOK != (i = startcupreds(cup))) return i;
	  }
	lsp = (cup->cu_pwp)->po_spl; /* we opened the spline in startcupreds */
	locwid = lsp->v_detw;	locht = lsp->v_deth; /* detector dimensions */
	cup->cu_xsca = dimscale((short)locwid);
	cup->cu_ysca = dimscale((short)locht);
	return EOK;
}

static int endcu(CROSSURUN *cup, int rev)
{ /*	cleanly closes off a run of crossu */
	printf(
	 " %6d reflections read in; %6d written out, %6d not written out\n",
		cup->cu_nrin, cup->cu_nrout, cup->cu_nrin - cup->cu_nrout);
	if (stdout != cup->cu_fpl)
	  {
		fprintf(cup->cu_fpl,
	 " %6d reflections read in; %6d written out, %6d not written out\n",
		 cup->cu_nrin, cup->cu_nrout, cup->cu_nrin - cup->cu_nrout);
		(void)fclose(cup->cu_fpl);	cup->cu_fpl = NULL;
	  }
	if (NULL != cup->cu_fpi) (void)fclose(cup->cu_fpi); cup->cu_fpi = NULL;
	if (NULL != cup->cu_fpo) (void)fclose(cup->cu_fpo); cup->cu_fpo = NULL;
	memzap((AVOID **)&(cup->cu_xp), "cup->cu_xp in crossu");
	memzap((AVOID **)&(cup->cu_urp), "cup->cu_urp in crossu");
	if (NULL != cup->cu_pwp)
	  {
		memzap((AVOID **)&((cup->cu_pwp)->po_prp0),
			"(cup->cu_pwp)->po_prp0 in crossu");
		memzap((AVOID **)&(cup->cu_pwp), "cup->cu_pwp in crossu");
	  }
	(void)ftables(0);	globalkill();	return rev;
}

static int crossproc(CROSSURUN *cup)
{ /*	Process the actual reflections in the input UREFLS file */
	int	i, ih, nearby, loh[3];
	double	zh;

	for (cup->cu_nrin = cup->cu_nrout = 0; /* loop through input refls */
	 0 < xgread((void *)cup->cu_urp, XG_OBSREF,cup->cu_fpi);
	 (cup->cu_nrin)++)
	   {
		xyomtoh(cup);	/* calculate hkl in the alternate unit cell */
		for (nearby = i = 0; i < 3; i++) /* are h,k,and l integers? */
		   {
			zh = cup->cu_oh[i]; ih = (cup->cu_oh[i] < 0 ?
			 cup->cu_oh[i] - 0.49999 : cup->cu_oh[i] + 0.49999);
			if (cup->cu_flags)
			  {
				if (fabs(zh - (double)ih)< cup->cu_xp->xerrind)
					nearby++;
			  }
			else {
				if (fabs(zh - (double)ih) < 0.3) nearby++;
				loh[i] = ih;
			     }
		   }
		/* in the old (cup->cu_flags==1) case, we just compare
		 the integerness with the integerness cutoff to decide
		whether this one really constitutes an overlap. In the new
		 (cup->cu_flags==0) case, see if the (X,Y,omega) lines up.
		 We use ihtoxyo() to check that */
		if ((nearby == 3) && !(cup->cu_flags))
		  {
			if (!ihtoxyo(loh, cup)) nearby = 0;
		  }
		/* if we don't see integerness and or neighborliness,
		 the reflection doesn't belong to the alternate lattice,
		 so we figure it's okay and we write it out */
		if (nearby < 3)
		  {
			if (verror(0 < xgwrite((void *)(cup->cu_urp),
			XG_OBSREF, cup->cu_fpo),
			"write a reflection to the output")) return EWRITERR;
			(cup->cu_nrout)++; /* increment count of output obs */
		  }
		/* if it _is_ integer-like or neighborly, we log that fact */
		else if (cup->cu_flags) culogbad(cup, NULL);
	   }
	return EOK;
}

/*ARGSUSED*/
int crossumain(int argc, char *argv[])
{ /* Mainline for this effort to accomplish a cross-comparison of UREFLS */
	int			i;
	static CROSSURUN	cur;

	if (EOK != (i = crossuset(argc, argv, &cur))) return endcu(&cur, i);
	return endcu(&cur, crossproc(&cur));
}

/* end of cross_u functionality */

/*************************
	The next few routines define the ifilter functionality,
	which filters out weak and strong refls by shells from a UREFLS.
	This gets rid of reflections from the input if they are
	found to have nonsense intensities or sigmas.
*************************************************************************/

static void summ_ifilrun(FILE *fp, int numin, int nshell, URSHEL *urs0)
{ /*	This summarizes a run of `ifilter'. */
	int	ishel;
	URSHEL	*ursp, *ursall;

	ursall = urs0 + nshell + 1;
	fprintf(fp, "%6d reflections read; %6d reflections written\n",
		numin, ursall->fsh_nok);
	fprintf(fp, " Results by shell:\n");
	fprintf(fp, " Dmax,A      N    Imin     N(I<     Imax");
	fprintf(fp, "  N(I>   Iraw-      N(    Nboth    out of\n");
	fprintf(fp, "     ,A     ok            Imin)     Imax");
	fprintf(fp, " Imax)   pkmax    peak)            bounds\n");
	for (ishel = 0, ursp = urs0; ishel < nshell; ishel++, ursp++)
	   {
		fprintf(fp,
	 	 " %6.2f %7d %9.1f %5d %9.1f %5d %9.1f %5d %5d\n",
		  ursp->fsh_resma, ursp->fsh_nok, ursp->fsh_imi,
		  ursp->fsh_nlti, ursp->fsh_ima, ursp->fsh_ngti,
		  ursp->fsh_pkrima, ursp->fsh_ngtfm, ursp->fsh_nboth);
	   }
	fprintf(fp,
	 " TOTALS %7d    ---    %5d    ---   %5d    ---   %5d %5d %5d\n",
	  (int)(ursp->fsh_nok), (int)(ursp->fsh_nlti), (int)(ursp->fsh_ngti),
	  (int)(ursp->fsh_ngtfm), (int)(ursp->fsh_nboth),
	  (int)(ursp->fsh_nobounds));
}

static int rescmp(URSHEL *ursp, URSHEL *ursq)
{ /* compare routine for a pair of URSHEL structures based on resolution */
	if (ursp->fsh_resma > ursq->fsh_resma) return -1;
	return 1;
}

static int shellstuff(int *nshellp, URSHEL **urs0p)
{ /*	This reads the shell-specific information from the file with
	environment variable "shells".
	It returns EOK if successful, various error codes if not. */
	int	insort, ishel, nitems; 
	char	*shnam, sline[128];
	double	shima, shimi, shpkrima, shrem;
	URSHEL	*match, *ursp, *ursq;
	FILE	*fpshe;

	shnam = trname("shells"); /* get the name of the SHELLS file */
	/* attempt to open it */
	if (NULL == (fpshe = xgopen(shnam, "r")))
	  {
		fprintf(stderr, "shell-file %s can't be opened\n", shnam);
		return EOPENERR;
	  }
	/* count how many usable lines there are in it */
	for (ishel = 0; NULL != fgets(sline, 120, fpshe); ishel++) ;
	if (ishel < 1)
	  {
		fprintf(stderr, "no readable lines in the shell-file %s\n",
			shnam);
		return EBAD;
	  }
	/* rewind so we can read the file for real */
	if (-1 == fseek(fpshe, 0L, 0))
	  {
		fprintf(stderr, "Error rewinding shell-file %s\n", shnam);
		return ESEEKERR;
	  }
	*nshellp = ishel;	/* record how many lines in the file */
	memzap((AVOID **)urs0p, "urs0 in ifilter"); /* zero memory */
	/* reserve memory for URSHEL structures */
	if (NULL == (*urs0p = (URSHEL *)callop((*nshellp) + 2, sizeof (URSHEL),
	 "urs0 in ifilter"))) return EALLOC;
	/* read the URSHEL information for real */
	for (ishel = 0, ursp = *urs0p; NULL != fgets(sline, 120, fpshe); )
	   {
		nitems = sscanf(sline, "%lf %lf %lf %lf",
		 &shrem, &shimi, &shima, &shpkrima);
		if (nitems < 4)
		  {
			shpkrima = UNSPEC_MAX;
			if (nitems < 3)
			  {
				shima = UNSPEC_MAX;
				if (nitems < 2) shimi = -UNSPEC_MAX;
			  }
		  }
		/* make sure this shell isn't already defined */
		if (nitems > 0)
		 {
		   for (match = NULL, ursq = *urs0p; ursq < ursp; ursq++)
		    {
			if (fabs(ursq->fsh_resma - shrem) < DFUZZ)
			  {
				match = ursq; break;
			  }
		    }
		   /* if this shell is a new one, save its information */
		   if (NULL == match)
		    {
			ursp->fsh_resma = shrem;	ursp->fsh_imi = shimi;
			ursp->fsh_ima = shima;	ursp->fsh_pkrima = shpkrima;
			ursp++;	ishel++;
		    }
		  }
	   }
	*nshellp = ishel;
	if ((*nshellp) > 2) /* see if we need to sort the shells by Dmax */
	  {
		for (insort = 1, ursp = *urs0p + 1, ishel = 1;
		 ishel < *nshellp; ishel++, ursp++)
		   {
			if (ursp->fsh_resma > (ursp-1)->fsh_resma)
			  {
				insort = 0; break;
			  }
		   }
		/* sort by increasing resolution */
		if ((!insort) && (*nshellp > 1)) qsort((void *)(*urs0p),
		 (size_t)(*nshellp), sizeof (URSHEL), (QPE)rescmp);
	  }
	(void)fclose(fpshe);	fpshe = NULL;
	return EOK;
}

int ifiltermain(int argc, char *argv[])
{ /* Mainline for this low-level filtering functionality for UREFLS files */
	int		etype, ifm, ishel, nin, nshel, retv, ufm;
	short		maxsh, *fmshp;
	double		obd, sclam, unco;
	FILE		*fpui, *fpuo, *fplog;
	OBSREF		ob;
	URSHEL		*ursall, *ursp, *ursq, *choice;
	static URSHEL	*urs0;
	static char *fnames[] = {"urefls", "shells", "urefout", "xlog", NULL};
	static char ftypes[] = {1, 1, 2, 2, 0 };

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	nin = 0;
	if (EOK != (retv = shellstuff(&nshel, &urs0))) return retv;
	if (qerror((AVOID *)(fpui = xgopen("urefls", "r")),
		"open input UREFLS file")) return EOPENERR;
	if (qerror((AVOID *)(fpuo = xgopen("urefout", "w")),
		"open output UREFLS file")) return EOPENERR;
	if (NULL == (fplog = propen("xlog"))) fplog = stdout;
	ftypes[2] = (fplog == stdout ? 0 : 2);
	/* this scale factor gets us from |s| to plane-spacing */
	sclam = 100. * needlambda();
	header(stdout, REDUCTION, fnames, ftypes);
	if (fplog != stdout) header(fplog, REDUCTION, fnames, ftypes);
	ursall = urs0 + nshel + 1;
	for ( ; 0 < xgread((void *)&ob, XG_OBSREF, fpui); nin++)
	   {
		choice = NULL;
		obd = sclam / sqrt((double)ob.obs_ssq);
		for (ursp = urs0, ishel = 0; ishel < nshel; ishel++, ursp++)
		   {
			ursq = ursp + 1;
			if ((ishel == nshel-1) || ((ursp->fsh_resma > obd) &&
				(obd >= ursq->fsh_resma)))
			  {
				choice = ursp;	break;
			  }
		   }
		if (NULL == choice) (ursall->fsh_nobounds)++;
		else {
			etype = 0;
			if (ob.obs_ii < choice->fsh_imi) etype = 1;
			if (ob.obs_ii > choice->fsh_ima) etype = 2;
			for (maxsh = ifm = 0, fmshp = &(ob.obs_frmint[0]);
			 ifm < PFRAMES; ifm++, fmshp++)
			   {
				ufm = *fmshp; if (ufm < -10000) ufm += 65535;
				if ((ufm) > maxsh) maxsh = ufm;
			   }
			if (maxsh > 1)
			  {
				unco = ((double)maxsh) *
					(1.E4 / ((double)(ob.obs_lpa)));
				if (unco > choice->fsh_pkrima) etype |= 4;
			  }
			if (!etype)
			  {
				if (verror(0 < xgwrite((void *)&ob, XG_OBSREF,
				 fpuo), "cannot write reflection to output"))
					return EWRITERR;
				(choice->fsh_nok)++; (ursall->fsh_nok)++;
			  }
			else if (etype == 1)
			  {
				(choice->fsh_nlti)++; (ursall->fsh_nlti)++;
			  }
			else if (etype == 2)
			  {
				(choice->fsh_ngti)++; (ursall->fsh_ngti)++;
			  }
			else if (etype == 4)
			  {
				(choice->fsh_ngtfm)++; (ursall->fsh_ngtfm)++;
			  }
			else if (etype == 6)
			  {
				(choice->fsh_nboth)++; (ursall->fsh_nboth)++;
			  }
			else fprintf(stderr, "Unknown etype = %d\n", etype);
		     }
	   }
	(void)fclose(fpui);	(void)fclose(fpuo);
	summ_ifilrun(fplog, nin, nshel, urs0);
	if (fplog != stdout)
	  {
		summ_ifilrun(stdout, nin, nshel, urs0); (void)fclose(fplog);
	  }
	globalkill();	return EOK;
}

/* end of ifilter functionality */

/*******************************
	The next few routines convert a USPOTLIST to a fake UREFLS file.
********************************/
static void p2u(P2UWORK *pup)
{ /* converts a PREDREF structure to an OBSREF */
	register int	i, j;
	short		shx, shy, lsh[3];
	long		ira;
	double		ios, ssei, ssq;
	double		ra;
	OBSREF		*urp;
	REFPARS		*xp;
	PREDREF		*prp;

	xp = pup->pu_xp;	urp = pup->pu_urp;	prp = pup->pu_prp;
	zeromem((void *)urp, sizeof (OBSREF));
	hunpack(prp->pindx, lsh);	/* get integer hkl as shorts */
	for (ssq = 0, i = 0; i < 3; i++)
	   {
		urp->obs_hkl[i] = lsh[i];
		for (ssei = j = 0; j < 3; j++)
			ssei += xp->xrec[i][j] * lsh[j];
		ssei *= xp->xlambda;	ssq += ssei * ssei;
	   }
	urp->obs_ssq = 1.e4 * ssq;
	ira = random();
	ra = ((double)ira) / (2.147483647E9 * 20.);
	urp->obs_ii = urp->obs_iunfit = pup->pu_f000 *
		exp(-12. * ssq) * (1. + ra);
	ios = flimits(pup->pu_ios000 * exp(-7. * ssq), 0.3, 20.);
	urp->obs_sigma = urp->obs_sunfit = urp->obs_ii * urp->obs_ii / ios;
	urp->obs_flags = (ios < 6.) ? 32 : 0;
	if (fabs(prp->pcor) < 32.75)
		urp->obs_lpa = 0.49999 + 1000. * prp->pcor;
	else	urp->obs_lpa = 65535;
	urp->obs_xpred = urp->obs_xob = prp->pxt * pup->pu_xscal + 0.49999;
	urp->obs_ypred = urp->obs_yob = prp->pyt * pup->pu_yscal + 0.49999;
	shx = prp->pxt + 0.49999;	shy = prp->pyt + 0.49999;
	urp->obs_region = getregion(shx, shy, /* prp->pphi, */ pup->pu_bdp);
	urp->obs_ompred = urp->obs_omob = prp->pphi;
	urp->obs_gof = 50;
	urp->obs_shift = 0; /* means Kabsch & non-Kabsch are equal */
	urp->obs_gama = 0.49999 + prp->pgama * 1000.;
	urp->obs_widob = 2;
	urp->obs_bgnd = urp->obs_bgunfit = 10;
	urp->obs_ncells = 150;
	urp->obs_reflno = urp->obs_chamno = 0;
	urp->obs_lowfrm = 2;	urp->obs_hifrm = 4; urp->obs_runno = 0;
	urp->obs_frmint[0] = urp->obs_frmint[1] = urp->obs_frmint[5] =
		urp->obs_frmint[6] = 0;
	urp->obs_frmint[2] = urp->obs_frmint[4] = urp->obs_ii * 0.1;
	urp->obs_frmint[3] = urp->obs_ii * 0.8;
}

int predtourfmain(int argc, char *argv[])
{ /* mainline for converting a prediction file into a fake UREFLS file */
	register int	nin;
	int		beamx, sizref;
	static FILE	*fppr, *fpur;
	static PREDREF	pri;
	static REFPARS	xr;
	static OBSREF	ur;
	static P2UWORK	pus;
	static char	*flist[] = {"uparams", "uspotlist", "urefls", NULL};
	static char	ftype[] = {1, 1, 2, 0};

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	header(stdout, PROCESSING, flist, ftype);
		/* open the input USPOTLIST file */
	if (NULL == (fppr = xgopen("uspotlist", "r"))) return EOPENERR;
	if (NULL == (fpur = xgopen("urefls", "w"))) return EOPENERR;
		/* implicitly read in the contents of the UPARAMS file */
	if (EOK != refparin(&xr, &sizref)) return EREADERR;
	pus.pu_f000 = 10000.;	pus.pu_ios000 = 150.;
		/* remember the scale factors for (X,Y) */
	pus.pu_xscal = dimscale(xr.xdetwid);
	pus.pu_yscal = dimscale(xr.xdetht);
	beamx = (xr.xdetwid - 1) / 2 +
		(int)(MEANPIXPERCM * xr.xplated * tan(xr.xswing));
	pus.pu_bdp = make_boundaries(beamx,
		(int)(xr.xdetwid), (int)(xr.xdetht));
	printf("Converting prediction to fake integration output;\n");
	pus.pu_urp = &ur;	pus.pu_xp = &xr;	pus.pu_prp = &pri;
	for (nin = 0; 0 < xgread(pus.pu_prp, XG_PREDREF, fppr); nin++)
	   {
		p2u(&pus);
		if (0 >= xgwrite(pus.pu_urp, XG_OBSREF, fpur)) return EWRITERR;
	   }
	fclose(fppr);	fclose(fpur);	fppr = fpur = NULL;
	memzap((AVOID **)&(pus.pu_bdp), "bdp in predtourfmain");
	printf(" %6d reflections read and written\n", nin);
	globalkill();	return EOK;
}

/* end of predtourf functionality */

/********************************
	The next several routines allow conversion of a
	SAINT ".raw" file to a UREFLS file.

	The input format for this activity contains the following:
	%4d%4d%4d	h, k, l
	%f%f            Intensity, sigma
	%d		batch number
	%f%f		input, output direction cosines for
			 diffraction vector with respect to x
	%f%f		input, output direction cosines for
			 diffraction vector with respect to y
	%f%f		input, output direction cosines for
			 diffraction vector with respect to z
	%d		status (not used, currently)
	%f%f%f		X,Y,scanning angle observed, in
			 pixels, pixels, and frames
	%f%f%f		X,Y,scanning angle predicted, in
			 pixels, pixels, and frames
	%f		product of Lorentz, polarization, air-absorption,
			 and faceplate-transmittance corrections
	%f		correlation coefficient between observed
			 and model profile
	%f		accumulated exposure time in hours up to
			 the point where this reflection is observed
	%f		detector two-theta-center value for images
	%f		observed scanning-angle value in degrees
	%d		which motor is moving (2=omega,3=phi)
	%d		sin(theta)/lambda * 10000
	%d		raw counts for reflection (peak+background)
	%d		raw background counts for reflection
	%f		scanning angle for reflection in degrees
			 relative to start of scan
	%d		crystal number
	%f		fraction of profile directly observed
	%d		unique sorting key for observation
	%d		multiplicity value (may be negative)
	%f		Lorentz correction
	%f%f		detector X and Y, corrected for spatial distortion,
			 relative to beam center (so may be < 0)
	%f		goniostat chi value where reflection was measured
	%f		goniostat phi or omega value (phi if omega is moving,
			 omega if phi is moving)

	Syntax:
		raw2urf
	Flags:
	 none so far.
	The program reads a ".raw" file with logical name or environment
	 variable SAINT_RAW; it writes to the logical name or env. vbl. UREFLS.
*************************************************************************/

static void rawproc(RAW2WORK *r2w, SAINTRAW *srp, OBSREF *obp)
{ /*	This converts a SAINTRAW structure into an OBSREF structure.
	It fakes the model profile */
	int	i;
	short	thisx, thisy;
	double	rthisx, rthisy, useinten;

	for (i = 0; i < 3; i++) obp->obs_hkl[i] = srp->ra_hkl[i];
	obp->obs_lpa = (short)(srp->ra_lpetc * 1.0e4 + 0.49999);
	useinten = ((double)srp->ra_stl10k) * 1.0e-4;
	useinten = useinten * useinten * 4.e4;
	obp->obs_ssq = useinten;
	rthisx = ((double)(r2w->r2_wid)) - 1. -
		srp->ra_xyzpred[0] * ((double)(r2w->r2_wid)/ 512.);
	rthisy = ((double)(r2w->r2_ht)) - 1. -
		srp->ra_xyzpred[1] * ((double)(r2w->r2_ht)/ 512.);
	thisx = rthisx + 0.49999;	thisy = rthisy + 0.49999;
	obp->obs_xpred = (short)(rthisx * r2w->r2_scl[0] + 0.49999);
	obp->obs_ypred = (short)(rthisy * r2w->r2_scl[1] + 0.49999);
	obp->obs_region = getregion(thisx, thisy, /* srp->ra_xyzob[2], */
		r2w->r2_bdp);
	obp->obs_ii = obp->obs_iunfit = srp->ra_inten;
	obp->obs_sigma = obp->obs_sunfit = srp->ra_sigma * srp->ra_sigma;
	obp->obs_ompred = srp->ra_xyzpred[2] + 1.;
	obp->obs_omob = srp->ra_xyzob[2] + 1.;
	obp->obs_gof = (short)(srp->ra_corr * obp->obs_ii + 0.49999);
	/* the following 4 params are really used for storing the
	 deltas that yields the Kabsch intensity and sigma values.
	 We therefore want them to be zero, since that's really what
	 SAINT measures anyway: */
	obp->obs_shift = obp->obs_widob = obp->obs_bgunfit =
		obp->obs_reflno = 0;
	obp->obs_gama = 1.;	/* close enough */
	rthisx = ((double)(r2w->r2_wid)) - 1. -
		srp->ra_xyzob[0] * ((double)(r2w->r2_wid) / 512.);
	rthisy = ((double)(r2w->r2_ht)) - 1. -
		srp->ra_xyzob[1] * ((double)(r2w->r2_ht)/ 512.);
	obp->obs_xob = (short)(rthisx * r2w->r2_scl[0] + 0.49999);
	obp->obs_yob = (short)(rthisy * r2w->r2_scl[1] + 0.49999);
	/* we make an arbitrary assumption that the reflection
	 covers thirty pixels */
	obp->obs_bgnd = srp->ra_rawbg / 30;
	obp->obs_ncells = 30;
	obp->obs_lowfrm = 1;
	obp->obs_hifrm = 5;
	obp->obs_chamno = 0;
	obp->obs_runno = (srp->ra_batch < 1) ? 0 : srp->ra_batch - 1;
	obp->obs_frmint[0] = obp->obs_frmint[6] = 0;
	useinten = srp->ra_inten * 0.08;
	if (useinten > 65535.) useinten = 65535.;
	obp->obs_frmint[1] = obp->obs_frmint[5] = useinten;
	useinten = srp->ra_inten * 0.20;
	if (useinten > 65535.) useinten = 65535.;
	obp->obs_frmint[2] = obp->obs_frmint[4] = useinten;
	useinten = srp->ra_inten * 0.44;
	if (useinten > 65535.) useinten = 65535.;
	obp->obs_frmint[3] = useinten;
	if (srp->ra_sigma > 0.001) useinten = srp->ra_inten / srp->ra_sigma;
	else useinten = 1;
	if (useinten < 6.) obp->obs_flags = 32;
	else obp->obs_flags = 0;
}

int raw2urfmain(int argc, char *argv[])
{ /*	This is the mainline for "raw2urf".  It reads in the reflections once
	 to find the minimum omega and maximum intensity; then it goes
	 through for real for processing. */
	int		beamx, nitems, nrefls, proper;
	char		*cpi, *cpo, hbuf[16], r2wl[514];
	FILE		*fpi, *fpo;
	static RAW2WORK	r2wo;
	static SAINTRAW	sr;
	static OBSREF	rf;
	static char	*fnames[] = { "SAINT_RAW", "UREFLS", NULL } ;
	static char	ftypes[] = { 1, 2, 0 } ;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
		/* open the input SAINT_RAWS file */
	if (qerror((AVOID *)(fpi = xgopen("SAINT_RAW", "r")),
	 "open input SAINT_RAW file")) return EOPENERR;
	if (qerror((AVOID *)(fpo = xgopen("UREFLS", "w")),
	 "open output UREFLS file")) return EOPENERR;
	header(stdout, REDUCTION, fnames, ftypes);
	if (EOK != find_dimension(&(r2wo.r2_wid), &(r2wo.r2_ht), 0, NULL))
		r2wo.r2_wid = r2wo.r2_ht = 1024;
	r2wo.r2_scl[0] = dimscale((short)(r2wo.r2_wid));
	r2wo.r2_scl[1] = dimscale((short)(r2wo.r2_ht));
	beamx = r2wo.r2_wid / 2;
	r2wo.r2_bdp = make_boundaries(beamx, r2wo.r2_wid, r2wo.r2_ht);
	for (nrefls = proper = 0; NULL != fgets(r2wl, 511, fpi); nrefls++)
	   {
		for (cpi = r2wl, cpo = hbuf; cpi < r2wl + 12; )
			*cpo++ = *cpi++;
		hbuf[12] = '\0';
		nitems = sscanf(hbuf, "%d%d%d",
		 &(sr.ra_hkl[0]), &(sr.ra_hkl[1]), &(sr.ra_hkl[2]));
		nitems += sscanf(&(r2wl[12]),
 "%f%f%d%f%f%f%f%f%f%d%f%f%f%f%f%f%f%f%f%f%f%d%d%d%d%f%d%f%d%d%f%f%f%f%f",
  &(sr.ra_inten), &(sr.ra_sigma), &(sr.ra_batch), &(sr.ra_sxi),
  &(sr.ra_sxo), &(sr.ra_syi), &(sr.ra_syo), &(sr.ra_szi), &(sr.ra_szo),
  &(sr.ra_status), &(sr.ra_xyzob[0]), &(sr.ra_xyzob[1]), &(sr.ra_xyzob[2]),
  &(sr.ra_xyzpred[0]), &(sr.ra_xyzpred[1]), &(sr.ra_xyzpred[2]),
  &(sr.ra_lpetc), &(sr.ra_corr), &(sr.ra_hrsaccum), 
  &(sr.ra_2theta), &(sr.ra_scanabs), &(sr.ra_scanaxis),
  &(sr.ra_stl10k), &(sr.ra_rawctpbg), &(sr.ra_rawbg),
  &(sr.ra_scanrel), &(sr.ra_xtalno), &(sr.ra_fracfill),
  &(sr.ra_sortkey), &(sr.ra_multiple),
  &(sr.ra_lortz), &(sr.ra_xpixcor), &(sr.ra_ypixcor), 
  &(sr.ra_chi), &(sr.ra_nonmove));
		if (nitems >= ENOUGH_ITEMS) proper++;
		else fprintf(stderr, "Line %d: only %3d items found\n",
		 nrefls, nitems);
		if (!(nrefls % 4000)) printf(
		 " %3d %3d %3d %10.2f %8.2f %7.1f %7.1f %7.1f %7.1f\n",
		 sr.ra_hkl[0], sr.ra_hkl[1], sr.ra_hkl[2],
		 sr.ra_inten, sr.ra_sigma,
		 sr.ra_xyzob[0], sr.ra_xyzob[1],
		 sr.ra_xyzpred[0], sr.ra_xyzpred[1]);
		rawproc(&r2wo, &sr, &rf);
		if (sizeof (OBSREF) != (nitems = xgwrite(&rf, XG_OBSREF, fpo)))
			printf(
 "Warning: refl %5d written wrong: attempted to write %3d bytes, got %3d\n",
			 nrefls, nitems, (int)(sizeof (OBSREF)));
	   }
	(void)fclose(fpi);	(void)fclose(fpo);	fpi = fpo = NULL;
	printf("%6d reflections processed: %6d valid, %6d invalid\n",
		nrefls, proper, nrefls - proper);
	globalkill();	return EOK;
}

/* end of raw2urf functionality */

/****************************
	The next few routines enable conversion of
	a UREFLS file to the D*TREK format found at University of Georgia.
	The output format for each refl is similar to Jim Plugrath's.
 Syntax:
	u2tod [-flags] [runnumber]
 Flags:
	 -s	use the summation data rather than the profile fitted data.
 The program reads a UREFLS file with environment variable UREFLS
 and the UPARAMS file; it writes to the env. vbl. XDTREK.
*************************************************************************/

static void headstuff(U2DRUN *u2p)
{ /*	This writes out the header info to the D*Trek header */
	fprintf(u2p->u2_fpo, "%1d %2d %1d %1d %4d %4d\n",
	 9, 25, 1, 4, (u2p->u2_xp)->xdetwid, (u2p->u2_xp)->xdetht);
	fprintf(u2p->u2_fpo, "CRYSTAL_MOSAICITY = %6.4f %6.4f %6.4f;\n",
	 2. * (u2p->u2_xp)->xgam[0] / RADC, 2. * (u2p->u2_xp)->xgam[1] / RADC,
	 2. * (u2p->u2_xp)->xgam[2] / RADC);
	fprintf(u2p->u2_fpo, "CRYSTAL_ORIENT_ANGLES = %8.4f %8.4f %8.4f;\n",
	 (u2p->u2_xp)->xdelom, (u2p->u2_xp)->xdelchi, (u2p->u2_xp)->xdelphi);
	fprintf(u2p->u2_fpo, "CRYSTAL_SPACEGROUP=%d;\n", (u2p->u2_xp)->xispgp);
	fprintf(u2p->u2_fpo,
	 "CRYSTAL_UNIT_CELL= %8.3f %8.3f %8.3f %8.3f %8.3f %8.3f\n",
	 (u2p->u2_xp)->xabc[0], (u2p->u2_xp)->xabc[1], (u2p->u2_xp)->xabc[2],
	 (u2p->u2_xp)->xang[0], (u2p->u2_xp)->xang[1], (u2p->u2_xp)->xang[2]);
	fprintf(u2p->u2_fpo, "nH\nnK\nnL\nnDetector_number\nnNonunf_flag\n");
	fprintf(u2p->u2_fpo,"nPackedHKL\nnRefineBatch\nnErrorFlag\nnScanID\n");
	fprintf(u2p->u2_fpo,
		"fIntensity\nfSigmaI\nfCalc_pixel0\nfCalc_pixel1\n");
	fprintf(u2p->u2_fpo,
		"fCalc_rot_mid\nfCalc_rot_start\nfCalc_rot_end\n");
	fprintf(u2p->u2_fpo, "fCalc_polarz\nfCalc_lorentz\nfCalc_oblique\n");
	fprintf(u2p->u2_fpo, "fResolution\nfObs_pixel0fObs_pixel1\n");
	fprintf(u2p->u2_fpo, "fObs_rot_mid\nfObs_rot_width\nfObs_rot_sigma\n");
	fprintf(u2p->u2_fpo, "fBackground\nfBackgroundSiggma\n");
	fprintf(u2p->u2_fpo, "fSvec\nfSvec1\nfSvec2\n");
	fprintf(u2p->u2_fpo, "fSvec0\nfSvec1\nfSvec2\n");
	fprintf(u2p->u2_fpo, "fSbatch\n");	xeflush(u2p->u2_fpo);
}

static void get_dtsvec(float *sv, U2DRUN *u2p)
{ /*	Compute the diffraction vector associated with a particular spot */
	int	res;
	double	wrad, xip, yip;
	float	xoc, yoc;

	xip = ((double)(u2p->u2_urf).obs_xob) / u2p->u2_xsca;
	yip = ((double)(u2p->u2_urf).obs_yob) / u2p->u2_ysca;
	res = pix2cm(xip, yip, &xoc, &yoc, 0);
	if (0 == res)
	  {
		sv[0] = sv[1] = sv[2] = 0.;	return;
	  }
	wrad = sqrt((u2p->u2_xp)->xpltd2 + xoc * xoc + yoc * yoc);
	sv[0] = (u2p->u2_xp)->xplated / wrad;
	sv[1] = xoc / wrad;	sv[2] = yoc / wrad;
}

static char *ffmt(float fl)
{ /*	This provides a conditional floating-point format */
	static char	forma[] = " %8.4f";

	if (fl < 0)
	  {
		if (fl > -10.) forma[4] = '4';
		else if (fl > -100.) forma[4] = '3';
		else if (fl > -1000.) forma[4] = '2';
		else forma[4] = '1';
	  }
	else {
		if (fl < 100.) forma[4] = '4';
		else if (fl < 1000.) forma[4] = '3';
		else if (fl < 10000.) forma[4] = '2';
		else forma[4] = '1';
	     }
	return forma;
}

static void out_xh(XTAL_DTREK *xhp, U2DRUN *u2p)
{ /*	Provides a printout of an XTAL_DTREK reflection */
	fprintf(u2p->u2_fpo, ffmt(xhp->xd_inten), xhp->xd_inten);
	fprintf(u2p->u2_fpo, ffmt(xhp->xd_sigma), xhp->xd_sigma);
	fprintf(u2p->u2_fpo, ffmt(xhp->xd_xcalc), xhp->xd_xcalc);
	fprintf(u2p->u2_fpo, ffmt(xhp->xd_ycalc), xhp->xd_ycalc);
	fprintf(u2p->u2_fpo, ffmt(xhp->xd_crotmid), xhp->xd_crotmid);
	fprintf(u2p->u2_fpo, ffmt(xhp->xd_crotstart), xhp->xd_crotstart);
	fprintf(u2p->u2_fpo, ffmt(xhp->xd_crotend), xhp->xd_crotend);
	fprintf(u2p->u2_fpo, ffmt(xhp->xd_crotwid), xhp->xd_crotwid);
	fprintf(u2p->u2_fpo, ffmt(xhp->xd_polarz), xhp->xd_polarz);
	fprintf(u2p->u2_fpo, ffmt(xhp->xd_lor), xhp->xd_lor);
	fprintf(u2p->u2_fpo, ffmt(xhp->xd_oblique), xhp->xd_oblique);
	fprintf(u2p->u2_fpo, ffmt(xhp->xd_resol), xhp->xd_resol);
	fprintf(u2p->u2_fpo, ffmt(xhp->xd_xobs), xhp->xd_xobs);
	fprintf(u2p->u2_fpo, ffmt(xhp->xd_yobs), xhp->xd_yobs);
	fprintf(u2p->u2_fpo, ffmt(xhp->xd_orotmid), xhp->xd_orotmid);
	fprintf(u2p->u2_fpo, ffmt(xhp->xd_orotwid), xhp->xd_orotwid);
	fprintf(u2p->u2_fpo, ffmt(xhp->xd_orotsig), xhp->xd_orotsig);
	fprintf(u2p->u2_fpo, ffmt(xhp->xd_bgnd), xhp->xd_bgnd);
	fprintf(u2p->u2_fpo, ffmt(xhp->xd_bgsigma), xhp->xd_bgsigma);
	fprintf(u2p->u2_fpo, ffmt(xhp->xd_svec[0]), xhp->xd_svec[0]);
	fprintf(u2p->u2_fpo, ffmt(xhp->xd_svec[1]), xhp->xd_svec[1]);
	fprintf(u2p->u2_fpo, ffmt(xhp->xd_svec[2]), xhp->xd_svec[2]);
	fprintf(u2p->u2_fpo, ffmt(xhp->xd_svec0[0]), xhp->xd_svec0[0]);
	fprintf(u2p->u2_fpo, ffmt(xhp->xd_svec0[1]), xhp->xd_svec0[1]);
	fprintf(u2p->u2_fpo, ffmt(xhp->xd_svec0[2]), xhp->xd_svec0[2]);
	fprintf(u2p->u2_fpo, " %5d\n", xhp->xd_sbatch);
}

static void cvtu2d(U2DRUN *u2p)
{ /* Performs most of the conversion from UREFLS format to D*Trek format */
	int		i;
	short		thkl[3];
	double		reals;
	XTAL_DTREK	xh;

	for (i = 3; i--; ) thkl[i] = xh.xd_hkl[i] = (u2p->u2_urf).obs_hkl[i];
	xh.xd_detno = 0;	/* chamno has junk in it */
	xh.xd_nonunf = 0;	/* temporary, I hope */
	xh.xd_packh = (int)hpack(thkl);
	xh.xd_batch = (int)((u2p->u2_urf).obs_omob / 2.);
	xh.xd_erflag = 0;	/* this might be wrong! */
	xh.xd_scanid = 17;	/* ho hum */
	xh.xd_inten = (u2p->u2_urf).obs_ii;
	xh.xd_sigma = sqrt((u2p->u2_urf).obs_sigma);
	xh.xd_xcalc = (u2p->u2_xp)->xdetwid -
		((double)((u2p->u2_urf).obs_xpred)) / u2p->u2_xsca;
	xh.xd_ycalc = (u2p->u2_xp)->xdetht -
		((double)((u2p->u2_urf).obs_ypred)) / u2p->u2_ysca;
	xh.xd_crotmid = (u2p->u2_urf).obs_ompred + u2p->u2_omoff;
	xh.xd_crotstart = xh.xd_crotmid - u2p->u2_mos / 2.;
	xh.xd_crotend = xh.xd_crotmid + u2p->u2_mos / 2;
	xh.xd_crotwid = u2p->u2_mos;	/* we'll make this better later */
	 /* Lp factor is not really right, but here's a start */
	xh.xd_polarz = 0.67;	/* correct this later */
	xh.xd_lor = ((double)((u2p->u2_urf).obs_lpa)) * 0.0001 / xh.xd_polarz;
	xh.xd_oblique = 2.45;	/* obliquity? huh? */
	reals = sqrt(((double)((u2p->u2_urf).obs_ssq)) * 0.0001);
	xh.xd_resol = (u2p->u2_xp)->xlambda / reals;
	xh.xd_xobs = (u2p->u2_xp)->xdetwid -
		((double)((u2p->u2_urf).obs_xob)) / u2p->u2_xsca;
	xh.xd_yobs = (u2p->u2_xp)->xdetht -
		((double)((u2p->u2_urf).obs_yob)) / u2p->u2_ysca;
	xh.xd_orotmid = (u2p->u2_urf).obs_omob + u2p->u2_omoff;
	xh.xd_orotwid = u2p->u2_mos;	/* we'll make this better later */
	xh.xd_orotsig = sqrt(fabs(u2p->u2_mos));
	xh.xd_bgnd = (u2p->u2_urf).obs_bgnd;
	xh.xd_bgsigma = sqrt(xh.xd_sigma * xh.xd_sigma * 0.2 +
		xh.xd_bgnd * xh.xd_bgnd); /* wild guess */
	get_dtsvec(xh.xd_svec, u2p);
	/* highly temporary */
	for (i = 0; i < 3; i++) xh.xd_svec0[i] = xh.xd_svec[i];
	xh.xd_sbatch = 10000 + (int)(u2p->u2_urf).obs_omob;
	fprintf(u2p->u2_fpo, " %5d %5d %5d %1d %1d %10d %2d %6d %2d",
	 xh.xd_hkl[0], xh.xd_hkl[1], xh.xd_hkl[2], xh.xd_detno,
	 xh.xd_nonunf, xh.xd_packh, xh.xd_batch, xh.xd_erflag, xh.xd_scanid);
	out_xh(&xh, u2p);
}

int urf2dmain(int argc, char *argv[])
{ /*	This is the mainline for "urf2d". */
	int		j, nfound, nrefls;
	U2DRUN		u2r;
	static FILE	*fpi;
	static VSPLINE	*splp;
	static char	*fnames[] =
	 { "UREFLS", "UPARAMS", "USPLINE", "XDTREK", NULL } ;
	static char	ftypes[] = { 1, 1, 1, 2, 0 } ;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
		/* open the input UREFLS file */
	if (qerror((AVOID *)(fpi = xgopen("UREFLS", "r")),
	 "open input UREFLS file")) return EOPENERR;
	if (NULL == (u2r.u2_xp = (REFPARS *)callop(1, sizeof (REFPARS),
	 "u2r.u2_xp in urf2d"))) return EALLOC;
	if (EOK != refparin(u2r.u2_xp, &nfound)) return EOPENERR;
	if ((NULL == (splp = gspline(5, &j))) || (EOK != j)) return EOPENERR;
	header(stdout, REDUCTION, fnames, ftypes);
		/* open the output XDTREK file */
	if (qerror((AVOID *)(u2r.u2_fpo = xgopen("XDTREK", "w")),
		"open output XDTREK file")) return EOPENERR;
		/* do the real work */
	u2r.u2_xsca = dimscale((u2r.u2_xp)->xdetwid);
	u2r.u2_ysca = dimscale((u2r.u2_xp)->xdetht);
	u2r.u2_omoff = (u2r.u2_xp)->xstart[2];
	u2r.u2_mos = 2. * (u2r.u2_xp)->xgam[0] / RADC;
	headstuff(&u2r);
	for (nrefls = 0;
		0 < xgread((void *)&(u2r.u2_urf), XG_OBSREF, fpi); nrefls++)
			cvtu2d(&u2r);
	(void)fclose(fpi); (void)fclose(u2r.u2_fpo); fpi = u2r.u2_fpo = NULL;
	printf("%6d reflections processed\n", nrefls);
	globalkill();	return EOK;
}

/* end of urf2d functionality */

/* end of ureftools.c */
