/* dihist.c -- create a histogram of (I-<I>)/sigma

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	24 Jun 2006: include xgpostint.h instead of rejdefs.h
	20 Nov 2005: inline argument declarations and strict prototyping
	10 Oct 2005: replace static globals with elements of DIHWORK structure
	17 Jun 2005: call to gspline() in dihsetup() returns correctly
	11 May 2005: check return status from xgjournal
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	27 Aug 2004: routines used only here are declared as static;cosmetics
	25 Jan 2004: convert most static globals into locals or elements
		of REJECTRUN structures
	21 Jan 2004: license message inserted and initial comments reordered
	 7 Jan 2004: static globals and elements of REJECTRUN structures
		replace extern globals; calls to get_shifts, etc. updated;
		callop() & memzap() replace shcalloc() & shfree()
	25 Nov 2003: correct memzap call diagnostic in dihsetup()
	27 Oct 2003: EXPECT_CUSHION is multiplier of fraction expected
	19 Sep 2003: gotone always initialized to zero
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	 5 Jan 2003: reorganized source-file header
	 1 Nov 2002: more informative diagnostics when files can't be opened
	15 Apr 2002: (?) in dihcalc(),
		modify calls related to shlp[] since it's handled differently
	25 Mar 2002: removed MSI attachments
	21 Mar 1998: initial version

   Purpose: go through a scaled MULTIREF and determine how many
	observations have u = (I-<I>)/sigma in ranges of
	u < -20, -20 < u < -19.75, . . . 19.75 < u < 20, u > 20.
   Cross references: ../postprocess/rej*.c
*************************************************************************/
#include	"xgpostint.h"
/*
	This reads through the headers in a MULTIREF and then looks at
	the data. It counts how many observations fall in various shells
	of (I-<I>)/sigma.
*/
#define			RANGEWIDTH	(0.4)
#define			NDIOSRANGES	102	/* used in dihist */
#define			MIDRANGE	(NDIOSRANGES / 2)

 typedef struct {
		int	*dh_ndelios;
		double	dh_minconsider, dh_maxconsider, dh_orfd, dh_orfn;
		double	dh_tellfrac, dh_totobs;
		double	*dh_xrfd, *dh_xrfn;
		} DIHWORK;

static DIHWORK *managedih(int opno)
{ /*	Manage a static global structure pointed to by diw */
	static DIHWORK	*diw;
	static char	diwmsg[] = "diw in managedih";
	static char	ndelmsg[] = "diw->dh_ndelios in managedih";
	static char	xrfmsg[] = "diw->dh_xrfd in managedih";

	if (0 == opno)
	  {
		(void)managedih(030);	memzap((AVOID **)&diw, diwmsg);
		return NULL;
	  }
	if (!(opno & 01))
	  {
		if (NULL != diw)
		  {
			if (opno & 010)
				memzap((AVOID **)&(diw->dh_ndelios), ndelmsg);
			if (opno & 020)
				memzap((AVOID **)&(diw->dh_xrfd), xrfmsg);
		  }
		return diw;
	  }
	if (opno == 3) return diw; /* unconditional return */
	if (NULL == diw) /* odd and > 3: check for existence of diw first */
	  {
		if (NULL == (diw = (DIHWORK *)callop(1, sizeof (DIHWORK),
			diwmsg))) return NULL;
	  }
	if (opno & 010) /* check for existence of ndelios */
	  {
		if ((NULL == diw->dh_ndelios) && (NULL == (diw->dh_ndelios =
		 (int *)callop(NDIOSRANGES, sizeof (int), ndelmsg))))
			return managedih(0);
	  }
	if (opno & 020) /* check for existence of xrfd */
	  {
		if ((NULL == diw->dh_xrfd) && (NULL == (diw->dh_xrfd =
		 (double *)callop(NDIOSRANGES, sizeof (double), xrfmsg))))
			return managedih(0);
		diw->dh_xrfn = diw->dh_xrfd + MIDRANGE;
	  }
	return diw;
}

static int dihsetup(int ac, char *av[])
{ /*	Set up defaults and deal with arguments for the dih functionality */
	char		**flags, **dcp;
	int		gotone, i;
	CRYSTAL		*xhp;
	DATARUN		*rhp;
	RUNWORK		*rup;
	VECPAR		*vpp, *vpe;
	DIHWORK		*diw;
	REJECTRUN	*rjr;
	static char	*flist[] = {
		"multiref", "scalein", "xlog", NULL };
	static char	ftype[] = { 1, 1, 2, 0 };

	if (NULL == (diw = managedih(03))) return EALLOC;
	if (NULL == (rjr = managerjp(07))) return EALLOC;
	diw->dh_tellfrac = gotone = 0;
	for (dcp = av + 1, i = 1; i < ac; i++, dcp++) /* interpret arg*/
		if (allnumeric(*dcp)) diw->dh_tellfrac = atof(*dcp);
	flags = flagproc(ac, av, "NS");
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	   switch (**dcp)
	     {
		case	'N':	rjr->rj_flags |= RJ_SEPBJ;		break;
		case	'S':	rjr->rj_flags |= RJ_PRESERVEOLD;	break;
		default:	break;
	     }
	if (NULL == (rjr->rj_fpl = propen("xlog"))) rjr->rj_fpl = stdout;
	if (NULL == (AVOID *)(rjr->rj_fpmi = xgopen("multiref", "r")))
	  {
		fprintf(stderr, "Cannot open input multiref\n");
		return EOPENERR;
	  }
	if (EOK != (i = headmem(&(rjr->rj_xhdp), &(rjr->rj_rhdp), 1))) return i;
	xhp = rjr->rj_xhdp;	rhp = rjr->rj_rhdp;
	if (0 > xgread((void *)xhp, XG_CRYSTAL, rjr->rj_fpmi))
	  {
		fprintf(stderr, "Cannot read crystal header from multiref\n");
		return EREADERR;
	  }
	/* reserve memory for diffraction-vector and run-header stuff,
	 plus the actual scaling input information */
	if (NULL == (rjr = managerjp(0307))) return EALLOC;
	for (i = 0, rup = rjr->rj_rub; i < xhp->xt_nruns; i++, rup++)
	   {
		if (verror(0 < xgread((void *)rhp, XG_DATARUN, rjr->rj_fpmi),
		 "read data run header in input multiref")) return EREADERR;
		(void)keep_dimen(rhp);
		workrun(rhp, rup, xhp, (int)(i * 200));
	   }
	/* examine the scale-factor output to see if we actually got any
	 geometry- dependent coefficients. */
	vpp = rjr->rj_vshp;	vpe = vpp + xhp->xt_nruns;
	for ( ; vpp < vpe; vpp++)
	   {
		for (i = gotone = 0; i < MAXVECPARS; i++)
		   if (fabs(vpp->v_coef[i]) > FUZZ)
		     {
			gotone = 1; break;
		     }
		if (gotone) break;
	   }
	rjr->rj_scflags &= ~(SC_DIFFRACTION | SC_EXITBEAM);
	/* eliminate memory for diffraction-vector stuff if it's unneeded */
	if (!gotone)
	  {
		rjr = managerjp(04000);	rjr->rj_nvpars = 0 ;
	  }
	else {
		vpp = rjr->rj_vshp;
		if (vpp->v_ptype == 'd')
		  {
			rjr->rj_scflags |= SC_DIFFRACTION; rjr->rj_nvpars = 10;
		  }
		else {
			rjr->rj_nvpars = 9;	rjr->rj_scflags |= SC_EXITBEAM;
		     }
		/* we need some spline information in this case
		to do the xy conversions */
		if (NULL == gspline(5, &i)) return i;
	     }
	ftype[3] = (rjr->rj_fpl == stdout) ? 0 : 2;
	fprintf(stderr, "Ready to print header\n");
	header(stdout, REDUCTION, flist, ftype);
	printf(" Already-deleted observations will%sbe included\n",
		(rjr->rj_flags & RJ_PRESERVEOLD) ? " NOT " : " ");
	printf(" Bijvoets %s be merged\n",
	 (rjr->rj_flags & RJ_SEPBJ) ? "WILL NOT" : "WILL");
	if (rjr->rj_fpl != stdout)
	  {
		header(rjr->rj_fpl, REDUCTION, flist, ftype);
		fprintf(rjr->rj_fpl,
		 " Already-deleted observations will%sbe included\n",
			(rjr->rj_flags & RJ_PRESERVEOLD) ? " NOT " : " ");
		fprintf(rjr->rj_fpl, " Bijvoets %s be merged\n",
		 (rjr->rj_flags & RJ_SEPBJ) ? "WILL NOT" : "WILL");
	  }
	return EOK;
}

void dihinit(void)
{ /*	Initialize the counters used in this process */
	int	i;
	DIHWORK	*diw;

	if (NULL == (diw = managedih(037))) return;
	diw->dh_maxconsider = RANGEWIDTH * (MIDRANGE - 1);
	diw->dh_minconsider = -(diw->dh_maxconsider);
	diw->dh_orfn = diw->dh_orfd = 0.;
	for (i = 0; i < MIDRANGE; i++) diw->dh_xrfn[i] = diw->dh_xrfd[i] = 0.;
	for (i = 0; i < NDIOSRANGES; i++) diw->dh_ndelios[i] = 0;
}

static int ibrange(double devval)
{ /*	This returns the range associated with a given (I-<I>)/sigma value. */
	int	range;
	DIHWORK	*diw;

	if (NULL == (diw = managedih(03))) return MIDRANGE;
	if (devval <= diw->dh_minconsider) return 0;
	else if (devval >= diw->dh_maxconsider) return NDIOSRANGES - 1;
	else range = (devval - diw->dh_minconsider) / RANGEWIDTH;
	return (range < 0) ? 0 : ((range >= NDIOSRANGES) ?
		NDIOSRANGES - 1 : range);
}

static int incrhis(ABBROB *abp, double ibar, double scfac)
{ /*	This increments the appropriate sums for an observation abp */
	int	ira, irq;
	double	thiss, thisden, thisdev;
	DIHWORK	*diw;

	if (NULL == (diw = managedih(03))) return 0;
	if (scfac > DFUZZ)
	  {
		thiss = abp->abo_sigma;
		if (thiss < FUZZ) thiss = 0.01;
		thisden = abp->abo_ii / (scfac * thiss);
		thisdev = thisden - ibar / thiss;
		ira = ibrange(thisdev);
		(diw->dh_ndelios[ira])++;
		thisdev *= thisdev;	thisden *= thisden;
		diw->dh_orfn += thisdev;	diw->dh_orfd += thisden;
		if (ira >= MIDRANGE) ira = NDIOSRANGES - ira - 1;
		for (irq = 0; irq < ira; irq++)
		   {
			diw->dh_xrfn[irq] += thisdev;
			diw->dh_xrfd[irq] += thisden;
		   }
		return 1;
	  }
	else return 0;
}

int dihcalc(MERGEDREF *mrp)
{ /*	This does the analysis of a reflection.
	It computes the sums that go into the <I> and (I-<I>)/sigma
	calculations, and increments the appropriate histogram. */
	register int		i, k;
	register ABBROB		*abp;
	SHIFT			*shlp, *tsh;
	int			shind, nob, run_num, nobp, nobn, retv;
	float			*fp1;
	double			ifactor, inten, ipbar, ibbar, ipden, ibden;
	double			qv, res, scfac, varob, wi;
	double			avec[MAXVECPARS], scai[MAXOB], *fp2;
	xgulong			ulsh;
	REJECTRUN		*rjr;

	if (NULL == (rjr = managerjp(03))) return EALLOC;
	retv = 0; if ((nob = mrp->mr_nob) < 2) return retv; /* can't process */
	ipbar = ibbar = ipden = ibden = nobp = nobn = 0; /* clear sums */
	res = mrp->mr_stl;
	shlp = rjr->rj_shlp;
	for (i = 0; i < nob; i++) /* get the means, etc. */
	   {
		abp = &(mrp->mr_ob[i]);	/* pointer to the observation */
		scai[i] = 0.;
		if (abp->abo_shift < 0)
		  {
			if (rjr->rj_flags & RJ_PRESERVEOLD) continue;
			else	abp->abo_shift = -(abp->abo_shift);
		  }
		ulsh = LSH(abp->abo_runno, 24) | LSH(abp->abo_chamno, 16) |
			abp->abo_shift;
		if (0 <= (shind = find_shift(shlp, ulsh, rjr->rj_nscales)))
		  { /* observation is scaled: use it */
			tsh = shlp + shind;
			scfac = terpscale(abp->abo_omob, tsh, res);
			if (NULL != rjr->rj_vshp)
			  {
				run_num = abp->abo_shift / 200;
				refvec(abp, rjr->rj_rub + run_num,
				 rjr->rj_scflags, rjr->rj_nvpars, avec);
				fp1 = &((rjr->rj_vshp + run_num)->v_coef[0]);
				for (k = rjr->rj_nvpars, qv = 0.,
				 fp2 = &(avec[0]); k--; fp1++, fp2++)
					qv += ((double)(*fp1)) * *fp2;
				scfac *= qv;
			  }
			if (scfac < FUZZ) scfac = 100. * FUZZ;
			scai[i] = scfac;
			/* compute the scaled intensity and sigma */
			inten = abp->abo_ii / scfac;
			varob = abp->abo_sigma;
			ifactor = tsh->shvar;
			wi = varob * varob + inten * inten * ifactor;
			if (wi < MINVAR) wi = MINVAR;
			wi = sqrt(wi) / scfac;
			wi = 1. / (wi * wi); /* least-squares weighting */
			/* increment the sums for <I> */
			if (i < mrp->mr_nobpri)
			  {
				ipbar += wi * inten; ipden += wi; nobp++;
			  }
			else {
				ibbar += wi * inten; ibden += wi; nobn++;
			     }
		  }
	   }
	if (rjr->rj_flags & RJ_SEPBJ)
	  {
		if ((nobp > 1) && (ipden > DFUZZ))
		  {
			i = 0;	ipbar /= ipden;
			for (abp = &(mrp->mr_ob[i]); i < mrp->mr_nobpri;
			 i++,abp++) retv += incrhis(abp, ipbar, scai[i]);

		  }
		if ((nobn > 1) && (ibden > DFUZZ))
		  {
			i = mrp->mr_nobpri;	ibbar /= ibden;
			for (abp = &(mrp->mr_ob[i]); i <mrp->mr_nob;i++, abp++)
				retv += incrhis(abp, ibbar, scai[i]);
		  }
	  }
	else {
		ipden += ibden;	ipbar += ibbar;	nobp += nobn;
		if ((nobp > 1) && (ipden > DFUZZ))
		  {
			i = 0;	ipbar /= ipden;
			for (abp = &(mrp->mr_ob[i]); i <mrp->mr_nob; i++,abp++)
				retv += incrhis(abp, ipbar, scai[i]);
		  }
	     }
	return retv;
}

void dihfinal(void)
{ /*	This does the calculations that we'll print afterward */
	int	idr;
	double	trfn, trfd;
	DIHWORK	*diw;

	if (NULL == (diw = managedih(03))) return;
	diw->dh_totobs = (double)(diw->dh_ndelios[0] +
		diw->dh_ndelios[NDIOSRANGES-1]);
	for (idr = 1; idr < NDIOSRANGES - 1; idr++)
		diw->dh_totobs += (double)(diw->dh_ndelios[idr]);
	diw->dh_orfn = ((diw->dh_orfn > DFUZZ) && (diw->dh_orfd > DFUZZ)) ?
	 sqrt(diw->dh_orfn / diw->dh_orfd) : 1.;
	for (idr = 0; idr < MIDRANGE; idr++)
	   {
		trfn = diw->dh_xrfn[idr];	trfd = diw->dh_xrfd[idr];
		diw->dh_xrfn[idr] = (trfn > DFUZZ) && (trfd > DFUZZ) ?
			sqrt(trfn / trfd) : 0.;
	   }
}

void dihshow(FILE *fp, int nru)
{ /*	This summarizes what we've discovered to the stream fp. */
	int	idr, idp;
	double	rangemin, rangemax, fracout, trfn;
	DIHWORK	*diw;

	if (NULL == (diw = managedih(03))) return;
	fprintf(fp, " %6d reflections examined\n", nru);
	fprintf(fp,
		" Number of observations with I-<I> < %7.3f = %6d\n",
			diw->dh_minconsider, diw->dh_ndelios[0]);
	fprintf(fp,
		" Number of observations with I-<I> > %7.3f = %6d\n",
			diw->dh_maxconsider, diw->dh_ndelios[NDIOSRANGES-1]);
	fprintf(fp, "Remaining ranges:\n");
	for (idr = 0; idr < 4; idr++) fprintf(fp,
	 " (DI/s)min (DI/s)max   Nobs  ");
	fprintf(fp, "\n");
	for (idr = 1, rangemin = diw->dh_minconsider; idr < NDIOSRANGES - 1;
	 idr++, rangemin += RANGEWIDTH)
	   {
		rangemax = rangemin + RANGEWIDTH;
		fprintf(fp, " %9.3f %9.3f %6d  ",
			rangemin, rangemax, diw->dh_ndelios[idr]);
		if (!((idr+1) % 4)) fprintf(fp, "\n");
	   }
	fprintf(fp, "\n");
	fprintf(fp,
	 "number of observations of multiply-observed reflections: %9.0f\n",
		diw->dh_totobs);
	fprintf(fp, "Overall Weighted RMS R on I = %9.4f\n", diw->dh_orfn);
	fprintf(fp,
	 "Fraction of data and excluded-R-factor outside each sigma cutoff\n");
	for (idr = 0; idr < 4; idr++)
		fprintf(fp, "  delI/sig Fraction Rexclude");
	fprintf(fp, "\n");
	for (fracout = idr = 0, idp = NDIOSRANGES - 1,
	 rangemin = diw->dh_minconsider;
	 idr < MIDRANGE; idr++, idp--, rangemin += RANGEWIDTH)
	   {
		fracout += ((double)(diw->dh_ndelios[idr] +
			diw->dh_ndelios[idp])) / diw->dh_totobs;
		if (fracout > 1.) break;
		if ((trfn = diw->dh_xrfn[idr]) < 1. )
			fprintf(fp, " %9.2f %8.4f %8.4f",
		 fabs(rangemin), fracout, trfn);
		else fprintf(fp, " %9.2f %8.4f         ",
		 fabs(rangemin), fracout);
		if (!((idr + 1) % 4)) fprintf(fp, "\n");
	   }
	fprintf(fp, "\n");
}

double qfromfrac(double delfrac)
{ /*	This returns an estimated |I-<I>|/sigma cutoff value that will yield
	roughly the user's requested number of deletions. Thus if delfrac=0.01,
	then this will return the |I-<I>|/sigma cutoff for which the total
	number of deletions will be about 1% of the multiply-observed data. */
	int	idr, idp;
	double	fracout, prevfrac, rangemin, prevrange, cross;
	DIHWORK	*diw;

	if (delfrac > 1.) delfrac = 1.;	prevfrac = 0.;
	if (NULL == (diw = managedih(03))) return 4;
	for (fracout = idr = 0, idp = NDIOSRANGES - 1,
	 rangemin = diw->dh_minconsider;
	 idr < MIDRANGE; idr++, idp--, rangemin += RANGEWIDTH)
	   {
		fracout += ((double)(diw->dh_ndelios[idr] +
			diw->dh_ndelios[idp])) / diw->dh_totobs;
		if (fracout >= delfrac)
		  {
			if (fracout > prevfrac)
			  {
				prevrange = rangemin - RANGEWIDTH;
				cross = prevfrac * rangemin -
					fracout * prevrange;
				return fabs((delfrac * RANGEWIDTH - cross) /
					(fracout - prevfrac));
			  }
			else return fabs(rangemin);
		  }
		prevfrac = fracout;
	   }
	return 0.;	/* shouldn't ever get here */
}

double qfromrval(double rtarget)
{ /*	This returns an estimated |I-<I>|/sigma cutoff value that will yield
	roughly the user-specified weighted RMS R-factor. This if rtarget =
	0.05, then this will return the |I-<I>|sigma cutoff that would result
	in deleting enough of the data to achieve Rweighted = 0.05. */
	int	idr, idp;
	double	prevrval, prevrange, cross, trfn, fracout, rangemin;
	DIHWORK	*diw;

	if (NULL == (diw = managedih(03))) return 4.;
	if (rtarget > diw->dh_orfn)
	  {
		fprintf(stderr, "No deletions required!\n");
		return 999.; /* handle special case */
	  }
	prevrval = rtarget + 0.001;
	for (fracout = idr = 0, idp = NDIOSRANGES - 1,
	 rangemin = diw->dh_minconsider;
	 idr < MIDRANGE; idr++, idp--, rangemin += RANGEWIDTH)
	   {
		fracout += ((double)(diw->dh_ndelios[idr] +
		 diw->dh_ndelios[idp])) / diw->dh_totobs;
		if ((trfn = diw->dh_xrfn[idr]) < rtarget)
		  {
			if (trfn < prevrval)
			  {
				prevrange = rangemin - RANGEWIDTH;
				cross = prevrval * rangemin - trfn * prevrange;
				return fabs((rtarget * RANGEWIDTH - cross) /
					(trfn - prevrval));
			  }
			else return fabs(rangemin);
		  }
		prevrval = trfn;
	   }
	return 0.;
}

/*ARGSUSED*/
int dihistmain(int argc, char *argv[])
{ /* Mainline for the histogram analysis */
	int		result, nrused;
	MERGEDREF	mr;
	REJECTRUN	*rjr;
	DIHWORK		*diw;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	nrused = 0;
		/*set up the program run*/
	if (NULL == (diw = managedih(037))) return EALLOC;
	if (NULL == (rjr = managerjp(07))) return EALLOC;
	if (EOK != (result = dihsetup(argc, argv))) return result;
	dihinit();	/* initialize some values */
	while (0 < xgread((AVOID *)&mr, XG_MERGEDREF, rjr->rj_fpmi))
		nrused += dihcalc(&mr);
	if (NULL != rjr->rj_fpmi) (void)fclose(rjr->rj_fpmi);
	rjr->rj_fpmi = NULL;
	dihfinal();	/* perform the final calculations */
	dihshow(stdout, nrused);
	if ((DFUZZ <= diw->dh_tellfrac) && (diw->dh_tellfrac <= 1.))
	  {
		printf(
		 " |I-<I>|/sigma cutoff = %12.5e to delete %7.4f of the data\n",
		 fabs(qfromfrac(diw->dh_tellfrac)), diw->dh_tellfrac);
	  }
	if ((NULL != rjr->rj_fpl) && (rjr->rj_fpl != stdout))
	  {
		dihshow(rjr->rj_fpl, nrused);
		(void)fclose(rjr->rj_fpl);
	  }
	rjr = managerjp(0);
	diw = managedih(0);	globalkill();	return result;
}

/* end of dihist.c */
