/* make_mu.c -- create a merged, scaled ascii list of hkls

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	 6 Dec 2006: handle missing primary or Bijvoet in anomalous scalepack;
		output MUREFLS with outmu(); use %hd for outputting shorts.
	24 Jun 2006: include xgpostint.h rather than rejdefs.h
	31 Dec 2005: don't qsort with fewer than 2 elements
	20 Nov 2005: inline argument declarations and strict prototyping
	10 Oct 2005: some housekeeping (should we get rid of static globals?)
	 9 Sep 2005: reorder_out() bugfix for XCNS and XXPLOR: missing ! fixed
	29 Jun 2005: shl_inten(): skip over correct number of bytes
	11 May 2005: check return status from xgjournal
	11 Apr 2005: name of this file changed from makemu.c to make_mu.c
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	 6 Sep 2004: cosmetic changes
	28 Aug 2004: routines used only here are declared as static; cosmetics;
	 reindexing done with try_reindex(), using mm_matrei and mm_vecrei elts
	 8 Aug 2004: add SCALEPACK output formats
	 2 Jun 2004: zero the MAKEMURUN structure before populating it;
		explicitly set mm_specres to zero if user gives no value
	28 May 2004: move structure types to ../include/rejdefs.h and
		include that; use MAKEMURUN structure type;
		add SHELX output functionality; clean up other output types
	28 Mar 2004: XCNS added to header bookkeeping
	19 Mar 2004: add CNS output capability;
	 routine countboth() written to facilitate that
	 1 Feb 2004: don't add 100 to the spacegroup for rhombohedrals
		if it's already been done.
	25 Jan 2004: convert some static globals into local variables
	21 Jan 2004: license message inserted and initial comments reordered
	 7 Jan 2004: static globals replace extern globals;
		calls to get_shifts, etc. updated;
		callop() & memzap() replace shcalloc() & shfree()
	19 Sep 2003: gotone always initialized to zero
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	15 Aug 2003: replace global uphkl[3] with static global lphkl[3];
		a couple of extra calls to hunpack to populate it
	 5 Jan 2003: reorganized source-file header
	 3 Jan 2003: always output X-PLOR format file.
		-x and -y options simply govern whether Friedel mates
		are separated.
	15 Aug 2002: in calc_refl(), construct MUREFLs only;
		write reflections out in reorder_out() (the old absorder_out())
	 5 Aug 2002: in calc_refl(), for acentrics: calculate values that will
		enable determination of relative k & B between F+ and F-
	15 Apr 2002: (?) in calc_refl() and shl_calcs(),
		modify calls related to shlp[] since it's handled differently
	25 Mar 2002: removed MSI attachments
	19 Sep 1995: close the output file before we're done in `makemu'.
	31 May 1995: "register" -> "register int" throughout;
		replace several longs with ints;
		replace %ld with %d accordingly; ulsh is an xgulong
	 2 Mar 1995: diagnostics to stdout; shared memory alloc and free
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 5 Oct 1994: use xgread and xgwrite on MULTIREF structure types
	30 Sep 1994: add X-PLOR output options "-x" and "-y"
	 5 May 1994: source file headers; POLYDET->XGEN;
		bugfix in calls to "memzap(&x,str)"
	16 Mar 1994: a few I/O changes and global variables moved around
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1994: initial X-GEN version
	20 Apr 1998: add "-m" option.
	26 Feb 1998: incorporate functionality of "absorder";
		allow for reindexing of reflections;
		allow for sorting on sin(theta); redefine sigma(<I>).
	18 Jan 1988: in makemumain(), the specified resolution limit
		takes effect if the plane spacing is LARGER
		(i.e., more restrictive) than that contained in the multiref.
	21 Sep 1987: use only obs. for which 0.2 < I(Scale)/I(raw) < 5.
	24 Aug 1987: in shl_inten(), ignore observations with I < -100.
	 1 Jul 1987: modifications in two-theta correction.
	xx xxx 1987: in shl_inten(), we compute <goodness-of-fit> in each shell
	11 Dec 1986: fixed bug in calculation of stscl in calc_refl().
	28 Jun 1986: use smoothly varying correction for sin(theta)/lambda
		instead of an indexed lookup.
	27 Dec 1985: incorporate sigmas of the scale factors into mean calcs.
	 1 Nov 1985: original code

   Purpose: creating an ascii list of merged hkls based on a MULTIREF.
   Cross references: mukb.c

	The "mulist" is an ascii file with the following format:
   h   k   l   stl       I sigma(I)      I+ sigma(I+)     I- sigma(I-) flag
		or
   h   k   l   stl       F sigma(F)      F+ sigma(F+)     F- sigma(F-) flag
	and has the name or logical name or environment variable "mulist".
	The program reads the scale file "scalein" to get scale factors
	for the individual intensity measurements and to figure out which
	observations the user wants to skip.
	The flag value ranges from 0 to 63 and is the ORing of the following
	conditions:
		bit no.		condition for off	condition for on
		------		-------------------	-----------------
		0		acentric reflection	centric reflection
		1		I+ measurement made	no I+ measurement
		2		I- measurement made	no I- measurement
		3		mean intensity > 0	mean I <= 0
		4		mean I+ > 0		mean I- <= 0
		5		mean I- > 0		mean I- <= 0
*************************************************************************/
#define		NEED_TIME	1
#include	"xgpostint.h"
#ifdef VMS
#include	"vsort.h"
#endif
#define		SIGMA_ZERO_F	(9999.)	/* sigma for F's from I <= 0 */
#define		SMALL		0.01	/* detectable value for Ai or Bi */
#define		MIN_USABLE	(-200)	/* minimum intensity used */
#define		SQ10OV3		(1.82574185)	/* sqrt(10/3) */
#define		DEF_SCUT	2	/* default I/sig cutoff for k,B calc */
	/* bits in the reflection flag value: */
#define		FL_CENTRIC	01	/* flag bit for centric refls */
#define		FL_NOPRIMARY	02	/* flag bit for no I+ obs */
#define		FL_NOBIJVOET	04	/* flag bit for no I- obs */
#define		FL_IBAR_LE_0	010	/* flag bit for <I>  <= 0 */
#define		FL_IPLUS_LE_0	020	/* flag bit for <I+> <= 0 */
#define		FL_IMINUS_LE_0	040	/* flag bit for <I-> <= 0 */

 typedef struct {
		int		mm_flagval, mm_npused, mm_nmused, mm_nshells;
		int		mm_nmiss, mm_vecflags, mm_nvpars;
		int		mm_ncenallbad, mm_nacenallbad;
		int		mm_ncensingly, mm_nacensingly, mm_ncentric;
		int		mm_nboth, mm_nprionly, mm_nbijonly, mm_nrefin;
		int		mm_nrnmin, mm_nrout, mm_nrfc, mm_nrfcb;
		int		mm_nrfbadi, mm_nrfbadfl, mm_kbn;
		unsigned	mm_nobin, mm_nobused, mm_nnoshift;
		unsigned	mm_nrtoohi, mm_nobtoohi;
		double		mm_e2, mm_e2b, mm_e2p;
		double		mm_imden, mm_iminus, mm_ipden, mm_iplus;
		double		mm_kbe0, mm_kbe1, mm_kbe2, mm_kbe3, mm_kbe4;
		double		mm_r1, mm_r3, mm_r5, mm_r7, mm_relb, mm_relk;
		double		mm_rsqcoef, mm_stl3, mm_stlmax, mm_yden, mm_yy;
		double		mm_avec[MAXVECPARS], mm_resmean[MAX_RUN_INDEX];
		double		mm_resqmean[MAX_RUN_INDEX];
		} MMUWORK; 

static int chk_absorder(MAKEMURUN *mmup)
{ /*	This returns 1 if and only if the current crystal has its reflections
	already listed in absolute increasing hkl order within the multiref.
	As such it returns 0 if crystal system 2 (monoclinic) and in crystal
	system 6 if the point group is 321 (trigonal with a twofold 60 degrees
	off axis).  For any other situation it returns 1 because the
	reflections are already in absolute hkl order. */
	CRYSTAL	*lhed;

	lhed = &(mmup->mm_xhed);
	if ((mmup->mm_flags & MUF_REINDEX) || (lhed->xt_isys == 2)) return 0;
	else if (lhed->xt_isys == 6)
		return !(lhed->xt_intl == 152 || lhed->xt_intl == 154 ||
			lhed->xt_intl == 156);
	return 1;
}

static void headsca(FILE *fp, CRYSTAL *xtp)
{ /*	This prints the first three lines to a SCALEPACK output file */
	fprintf(fp, "%5d\n%5d\n%10.3f%10.3f%10.3f%10.3f%10.3f%10.3f %s\n",
	 1, -985, xtp->xt_ar[0], xtp->xt_ar[1], xtp->xt_ar[2],
	 xtp->xt_ang[0], xtp->xt_ang[1], xtp->xt_ang[2],
	 intltolc(xtp->xt_intl));
}

int mufiles(MAKEMURUN *mmup)
{ /*	This opens and minimally processes the input files for makemu */
	int		nsca;
	CRYSTAL		*lhed;

	mmup->mm_mrname = trname("multiref");
	mmup->mm_cnsname = trname("xcns");
	mmup->mm_shelname = trname("xshel");
	if (qerror((AVOID *)(mmup->mm_fpmi = xgopen(mmup->mm_mrname, "r")),
		"open input multiref")) return EOPENERR;
	if (qerror((AVOID *)(mmup->mm_fpo = xgopen("mulist", "w")),
		"open output mulist")) return EOPENERR;
	if (qerror((AVOID *)(mmup->mm_fpx = xgopen("xxplor", "w")),
		"open XPLOR input")) return EOPENERR;
	if (qerror((AVOID *)(mmup->mm_fpc = xgopen(mmup->mm_cnsname, "w")),
		"open CNS input")) return EOPENERR;
	if (qerror((AVOID *)(mmup->mm_fpshel = xgopen(mmup->mm_shelname, "w")),
		"open SHELX input")) return EOPENERR;
	if (qerror((AVOID *)(mmup->mm_fpsca = xgopen("xsca", "w")),
		"open merged SCALEPACK output")) return EOPENERR;
	if (qerror((AVOID *)(mmup->mm_fpasca = xgopen("xasca", "w")),
		"open unmerged SCALEPACK output")) return EOPENERR;
	/* read the multiref headers */
	if (verror(0 < xgread((void *)&(mmup->mm_xhed), XG_CRYSTAL,
	 mmup->mm_fpmi), "read crystal header from multiref")) return EREADERR;
	lhed = &(mmup->mm_xhed);
	if (NULL == (AVOID *)(mmup->mm_vshp = (VECPAR *)callop(
	 (unsigned)lhed->xt_nruns, sizeof (VECPAR),"mmup->mm_vshp in makemu")))
		return EALLOC;
		/* read in the scale factors */
	if (NULL == (mmup->mm_shlp = get_shifts(NULL, mmup->mm_vshp, &nsca)))
	  {
		fprintf(stderr, "Cannot read scale-factor input file\n");
		return EOPENERR;
	  }
	mmup->mm_nscales = nsca;
	mmup->mm_flags = MUF_XPLORSEP;
	mmup->mm_oversc = 1.;	mmup->mm_scut = DEF_SCUT;
	if (NULL == (AVOID *)(mmup->mm_rub = (RUNWORK *)callop((unsigned)
	 (lhed->xt_nruns), sizeof (RUNWORK), "rub in makemu"))) return EALLOC;
	/* save mem for reordering */
	if (NULL == (AVOID *)(mmup->mm_mur0 = (MUREFL *)callop((unsigned)
	 lhed->xt_nmerged, sizeof (MUREFL), "mmup->mm_mur0 in makemu")))
		return EALLOC;
	mmup->mm_mup = mmup->mm_mur0;
	if (mmup->mm_flags & MUF_ABSORDER)
	  {
		if (chk_absorder(mmup))	mmup->mm_flags &= ~(MUF_ABSORDER);
	  }
	if (NULL == (mmup->mm_fpl = propen("xlog"))) mmup->mm_fpl = stdout;
	return EOK;
}

static int setup_mu(int argc, char *argv[], MAKEMURUN *mmup)
{ /*	This decodes the runtime arguments and opens the files */
	int	retv;
	char	**dcp, **flags;

	zeromem((void *)mmup, sizeof (MAKEMURUN));
	mmup->mm_tl0 = NOW;
	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	if (EOK != (retv = mufiles(mmup))) return retv;
	flags = flagproc(argc, argv, "ACFIMNPSTVXY");
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	   switch (**dcp)
	     {
		case	'A': /* output in absolute hkl order */
		 mmup->mm_flags |= MUF_ABSORDER;	break;
		case	'C':	/* specify I/sigma cutoff for k,B calcs */
		 mmup->mm_scut = atof((*dcp) + 1);	break;
		case	'F': /* output F's, not I's */
		 mmup->mm_flags |= MUF_OUTPUT_F;	break;
		case	'I': /* reindex reflections */
		 mmup->mm_flags |= MUF_REINDEX;	break;
		case	'M': /* only mult-obs.refls*/
		 mmup->mm_flags |= MUF_MULOBSERVED;	break;
		case	'N': /* set I<0 values to 0*/
		 mmup->mm_flags |= MUF_NONEGI;	break;
		case	'P': /* preserve sin(theta) dependence of input
			 scale factors rather than correcting for it */
		 mmup->mm_flags |= MUF_PRESERVE;	break;
		case	'S': /*overall scale factor*/
		 (void)sscanf((*dcp) + 1,"%f", &(mmup->mm_oversc));	break;
		case	'T': /* sin(theta) order*/
		 mmup->mm_flags |= MUF_STLORDER;	break;
		case	'V': /* alternate variance */
		 mmup->mm_flags |= MUF_VAROFMEAN;	break;
		case	'X': /* Separate Friedel mates in XHKL file */
		 mmup->mm_flags |= MUF_XPLORSEP;	break;
		case	'Y': /*   merged Friedel mates in XHKL file */
		 mmup->mm_flags &= ~MUF_XPLORSEP;
		 mmup->mm_flags |=  MUF_XPLORMER; break;
		default	:	break;
	   }
	for (argc--, argv++; argc; argc--,argv++) if (allnumeric(*argv)) break;
	if (argc)
	  {
		mmup->mm_specres = atof(*argv);
		if ((0 > mmup->mm_specres) || (mmup->mm_specres > 999.))
			mmup->mm_specres = 0.;
	  }
	else mmup->mm_specres = 0.;
	/* sin(theta) order takes precedence over absolute hkl order */
	if ((mmup->mm_flags & MUF_ABSORDER) && (mmup->mm_flags & MUF_STLORDER))
		mmup->mm_flags &= ~(MUF_ABSORDER);
	if ((mmup->mm_flags & MUF_REINDEX) && !try_reindex(mmup->mm_vecrei,
	 mmup->mm_matrei)) mmup->mm_flags -= MUF_REINDEX;
	if (mmup->mm_oversc < FUZZ || mmup->mm_oversc > 10000.)
		mmup->mm_oversc = 1.;
	mmup->mm_oversq = mmup->mm_oversc * mmup->mm_oversc;
	return EOK;
}

static void announce_reorder(FILE *fp, MAKEMURUN *mmup)
{ /*	This tells the world that we're going to output reflections in
	nonstandard order */
	if (mmup->mm_flags & MUF_ABSORDER)
	  {
	   fprintf(fp, "Reflections will be output in absolute hkl order\n");
	   fprintf(fp, "\trather than in the order in which they're stored\n");
	   if ((mmup->mm_xhed).xt_isys == 2) fprintf(fp,
	 "Reflections input as (h, k, -l) will be output as (-h, k, l)\n");
	  }
	else fprintf(fp,
	 "Reflections will be output sorted on sin(theta)\n");
}

static void sum_muset(FILE *fp, MAKEMURUN *mmup)
{ /* Summarize the run conditions for a run of makemu */
	static char	*flist[] = {
		"multiref", "scalein", "mulist", "xxplor", "xcns", "xshel",
		"xsca", "xasca", "xlog", NULL };
	static char	ftype[] = { 1, 1, 2, 2, 2, 2, 2, 2, 2, 0 };

	ftype[6] = (mmup->mm_fpl == stdout) ? 0 : 2;
	header(fp, REDUCTION, flist, ftype);
	fprintf(fp,"\t%s will be output\n", mmup->mm_flags & MUF_OUTPUT_F ?
	 "Structure amplitudes" : "Intensities");
 	if (mmup->mm_oversc != 1.) fprintf(fp,
	 "\tAll mean %c values will be multiplied by %9.5f before output\n",
		(mmup->mm_flags & MUF_OUTPUT_F ? 'F' : 'I'), mmup->mm_oversc);
	if (mmup->mm_flags & MUF_NONEGI) fprintf(fp,
	  "\tAll negative <I>, <I+>, and <I-> values < 0 will be set to 0\n");
	if (mmup->mm_flags & MUF_REINDEX)
	  {
		fprintf(fp,
"Reflections will be reindexed according to h' = Mh + v\n");
		fprintf(fp,"\t\t  %6.2f %6.2f %6.2f\t\t%3d\n",
		 mmup->mm_matrei[0][0], mmup->mm_matrei[0][1],
		 mmup->mm_matrei[0][2], (int)(mmup->mm_vecrei[0]));
		fprintf(fp,"\twhere M = %6.2f %6.2f %6.2f\t\t%3d\n",
		 mmup->mm_matrei[1][0], mmup->mm_matrei[1][1],
		 mmup->mm_matrei[1][2], (int)(mmup->mm_vecrei[1]));
		fprintf(fp,"\t\t  %6.2f %6.2f %6.2f\t\t%3d\n",
		 mmup->mm_matrei[2][0], mmup->mm_matrei[2][1],
		 mmup->mm_matrei[2][2], (int)(mmup->mm_vecrei[2]));
	  }
	fprintf(fp, "\ttheta dependence in scale factors will be %s\n",
		(mmup->mm_flags & MUF_PRESERVE) ? "preserved" : "recovered");
	if (mmup->mm_flags & MUF_MULOBSERVED) fprintf(fp,
 "Only reflections with at least 2 usable observations will be output\n");
	if (DFUZZ < mmup->mm_specres) fprintf(fp,
 "Only reflections with plane spacings > %6.3f Angstroms will be output\n",
		mmup->mm_specres);
	if (mmup->mm_flags & MUF_REORDER) /* explain the reordering */
		announce_reorder(fp, mmup);
	fprintf(fp, "X-PLOR output will have %s %c+ and %c- values\n",
		(mmup->mm_flags & MUF_XPLORSEP) ? "Separated" : "Merged",
		(mmup->mm_flags & MUF_OUTPUT_F) ? 'F' : 'I',
		(mmup->mm_flags & MUF_OUTPUT_F) ? 'F' : 'I');
	fprintf(fp,
 "Reflections in the MULIST will be output in standard MULIST format\n");
	fprintf(fp, "Bijvoet R-factor calcs: I/sigma cutoff = %7.3f\n",
		mmup->mm_scut);
}

static void get_fmeans(MAKEMURUN *mmup, MMUWORK *mmuw)
{ /*	this computes <F>, sigma<F>, etc. */

	if (mmuw->mm_yy > DFUZZ)
	  {
		mmuw->mm_yy = mmup->mm_oversc * sqrt(mmuw->mm_yy /
			mmuw->mm_yden);
		mmuw->mm_yden = mmup->mm_oversq * sqrt((1. + mmuw->mm_e2) /
			(8. * mmuw->mm_yden)) / mmuw->mm_yy;
	  }
	else {
		mmuw->mm_yy = 0.; mmuw->mm_yden = SIGMA_ZERO_F;
		mmuw->mm_flagval |= FL_IBAR_LE_0;
	     }
	if (mmuw->mm_flagval & FL_CENTRIC) /* centrics: set +,- equal to mean*/
	  {
		mmuw->mm_iminus = mmuw->mm_iplus = mmuw->mm_yy;
		mmuw->mm_ipden = mmuw->mm_imden = mmuw->mm_yden;
		return;
	  }
	if (mmuw->mm_npused && mmuw->mm_ipden > DFUZZ)	/* any primaries? */
	  {
		if (mmuw->mm_iplus > DFUZZ)
		  {
			mmuw->mm_iplus = mmup->mm_oversc *
				sqrt(mmuw->mm_iplus / mmuw->mm_ipden);
			mmuw->mm_ipden = mmup->mm_oversq * sqrt((1. +
			 mmuw->mm_e2p) / (8. *mmuw->mm_ipden))/ mmuw->mm_iplus;
		  }
		else {
			mmuw->mm_iplus = 0.;	mmuw->mm_ipden = SIGMA_ZERO_F;
			mmuw->mm_flagval |= FL_IPLUS_LE_0;
		     }
	  }
	else /* no primary obs: use the overall sum and flag the result */
	   {
		mmuw->mm_iplus = mmuw->mm_iminus = mmuw->mm_yy;
		mmuw->mm_ipden = mmuw->mm_imden = mmuw->mm_yden;
		if (mmuw->mm_yy < DFUZZ) mmuw->mm_flagval |= FL_IMINUS_LE_0;
		mmuw->mm_flagval |= FL_NOPRIMARY;	return;
	   }
	if (mmuw->mm_nmused && mmuw->mm_imden > DFUZZ) /* any Bijvoets? */
	  {
		if (mmuw->mm_iminus > DFUZZ)
		  {
			mmuw->mm_iminus = mmup->mm_oversc *
				sqrt(mmuw->mm_iminus/mmuw->mm_imden);
			mmuw->mm_imden = mmup->mm_oversq * sqrt((1. +
			 mmuw->mm_e2b)/(8. * mmuw->mm_imden))/ mmuw->mm_iminus;
		  }
		else
		  {
			mmuw->mm_iminus = 0.;	mmuw->mm_imden = SIGMA_ZERO_F;
			mmuw->mm_flagval |= FL_IMINUS_LE_0;
		  }
	  }
	else	/* no Bijvoet obs: use the primaries and flag it */
	  {
		mmuw->mm_iminus = mmuw->mm_iplus = mmuw->mm_yy;
		mmuw->mm_imden = mmuw->mm_ipden = mmuw->mm_yden;
		if (mmuw->mm_yy < DFUZZ) mmuw->mm_flagval |= FL_IPLUS_LE_0;
		mmuw->mm_flagval |= FL_NOBIJVOET;
	  }
	return;
}

static void get_imeans(MAKEMURUN *mmup, MMUWORK *mmuw)
{ /*	This computes <I>, sigma(<I>), etc. */

	mmuw->mm_yy = mmup->mm_oversc * mmuw->mm_yy / mmuw->mm_yden;
	mmuw->mm_yden = mmup->mm_oversc * sqrt(0.5 * (1. + mmuw->mm_e2) /
		mmuw->mm_yden);
	if (mmuw->mm_yy <= 0.)
	  {
		mmuw->mm_flagval |= FL_IBAR_LE_0;
		if (mmup->mm_flags & MUF_NONEGI) mmuw->mm_yy = 0.;
	  }
	else	mmuw->mm_flagval &= ~(FL_IBAR_LE_0);
	if (mmuw->mm_flagval & FL_CENTRIC)
	  {
		mmuw->mm_iminus = mmuw->mm_iplus = mmuw->mm_yy;
		mmuw->mm_imden = mmuw->mm_ipden = mmuw->mm_yden;
		return;
	  }
	if (mmuw->mm_npused && mmuw->mm_ipden > DFUZZ)	/* any primaries? */
	  {
		mmuw->mm_iplus *= mmup->mm_oversc /mmuw->mm_ipden;
		mmuw->mm_ipden = mmup->mm_oversc * sqrt(0.5 * (1. +
			mmuw->mm_e2p) / mmuw->mm_ipden);
		if (mmuw->mm_iplus <= 0.)
		  {
			mmuw->mm_flagval |= FL_IPLUS_LE_0;
			if (mmup->mm_flags & MUF_NONEGI) mmuw->mm_iplus = 0.;
		  }
	  }
	else	/* no primary obs: use Bijvoets and flag result */
	   {
		mmuw->mm_iplus = mmuw->mm_iminus = mmuw->mm_yy;
		mmuw->mm_ipden = mmuw->mm_imden = mmuw->mm_yden;
		if (mmuw->mm_yy < DFUZZ) mmuw->mm_flagval |= FL_IMINUS_LE_0;
		mmuw->mm_flagval |= FL_NOPRIMARY;	return;
	   }
	if (mmuw->mm_nmused && mmuw->mm_imden > DFUZZ)
	  {
		mmuw->mm_iminus *= mmup->mm_oversc / mmuw->mm_imden;
		mmuw->mm_imden = mmup->mm_oversc * sqrt(0.5 * (1. +
			mmuw->mm_e2b) / mmuw->mm_imden);
		if (mmuw->mm_iminus <= 0.)
	  	  {
			mmuw->mm_flagval |= FL_IMINUS_LE_0;
			if (mmup->mm_flags & MUF_NONEGI) mmuw->mm_iminus = 0.;
		  }
	  }
	else { /* no Bijvoet obs: use primaries and flag result */
		mmuw->mm_iminus = mmuw->mm_iplus = mmuw->mm_yy;
		mmuw->mm_imden = mmuw->mm_ipden = mmuw->mm_yden;
		if (mmuw->mm_yy < DFUZZ) mmuw->mm_flagval |= FL_IPLUS_LE_0;
		mmuw->mm_flagval |= FL_NOBIJVOET;
	     }
	return;
}

static void printsca(FILE *fp, short *lhkl, double *vals, int bothpresent)
{ /*	This prints a reflection to the SCALEPACK-style output file */

	if (bothpresent) fprintf(fp, "%4hd%4hd%4hd%8.1lf%8.1lf%8.1lf%8.1lf\n",
	 lhkl[0], lhkl[1], lhkl[2], vals[2], vals[3], vals[4], vals[5]);
	else fprintf(fp, "%4hd%4hd%4hd%8.1lf%8.1lf\n",
	 lhkl[0], lhkl[1], lhkl[2], vals[0], vals[1]);
}

static void shelmeans(MAKEMURUN *mmup, MMUWORK *mmuw, short lhkl[])
{ /*	This computes and outputs Sheldrick's and SCALEPACK's <I>, sigma(<I>),
	etc.  Since these programs don't care how the data are sorted,
	we can output these as they're calculated. */
	int		bothhere, i;
	double		fsca[6];
	SHEL_HKL	shh;

	for (i = 0; i < 3; i++) shh.x_hkl[i] = lhkl[i]; /* indices */
	shh.x_batch = 1;	/* batch number (could be changed later) */
	fsca[0] = mmup->mm_oversc * mmuw->mm_yy / mmuw->mm_yden;
	fsca[1] = mmup->mm_oversc * sqrt(0.5 * (1. + mmuw->mm_e2) /
		mmuw->mm_yden);
	if (mmuw->mm_flagval & FL_CENTRIC) /* Centric reflection: we're done */
	  {
		shh.x_inten = fsca[0];	shh.x_sigma = fsca[1];
		printshel(mmup->mm_fpshel, &shh);
		printsca(mmup->mm_fpsca, lhkl, fsca, 0);
		printsca(mmup->mm_fpasca, lhkl, fsca, 0);
		return;
	  }
	bothhere = 0;
	if (mmuw->mm_npused && mmuw->mm_ipden > DFUZZ)	/* any primaries? */
	  {
		fsca[2] = shh.x_inten = mmuw->mm_iplus * mmup->mm_oversc /
			mmuw->mm_ipden;
		fsca[3] = shh.x_sigma = mmup->mm_oversc *
		 sqrt(0.5 * (1. + mmuw->mm_e2p) / mmuw->mm_ipden);
		printshel(mmup->mm_fpshel, &shh);
		bothhere++;
	  }
	if (mmuw->mm_nmused && mmuw->mm_imden > DFUZZ) /* any Bijvoets? */
	  {
		fsca[4] = shh.x_inten = mmuw->mm_iminus * mmup->mm_oversc /
			mmuw->mm_imden;
		fsca[5] = shh.x_sigma = mmup->mm_oversc *
		 sqrt(0.5 * (1. + mmuw->mm_e2b) / mmuw->mm_imden);
		for (i = 0; i < 3; i++) shh.x_hkl[i] = -lhkl[i];
		printshel(mmup->mm_fpshel, &shh);
		bothhere++;
	  }
	printsca(mmup->mm_fpsca, lhkl, fsca, 0);
	/* Dec.06: special case for anomalous SCALEPACK file.
	 If the primary is missing, set fsca[2] to 0 and fsca[3] to -1.
	 If the Bijvoet is missing, set fsca[4] to 0 and fsca[4] to -1. */
	if (!(mmuw->mm_npused))
	  {
		fsca[2] = 0.;	fsca[3] = -1.;
	  }
	if (!(mmuw->mm_nmused))
	  {
		fsca[4] = 0.;	fsca[5] = -1.;
	  }
	printsca(mmup->mm_fpasca, lhkl, fsca, (bothhere > 1));
}

static int mm31ir(short *osv, float mat[3][3], short *isv)
{ /*	This left-multiplies the 3x3 matrix `mat' by the vector `isv'
	to give osv.  If the result is non-integral it returns 0;
	otherwise, 1 */
	register int	i, j, mm;
	double		tv;

	for (i = 3; i--; )
	   {
		for (tv = j = 0; j < 3; j++)
			tv += mat[i][j] * (double)(isv[j]);
		mm = tv;	if (fabs(tv - ((double)mm)) > 0.05) return 0;
		osv[i] = mm;
	   }
	return 1;
}

static int gshell(float stl, MMUWORK *mmuw)
{ /*	This returns the shell number associated with the sin(theta)/lambda
	value */
	register int	n;
	double		ts;

	ts = stl / mmuw->mm_stlmax;	ts *= ts * ts;
	mmuw->mm_stl3 = ts;
	n = ts * (double)(mmuw->mm_nshells);
 	return (n < mmuw->mm_nshells) ? n : mmuw->mm_nshells - 1;
}

static void calc_refl(MERGEDREF *mrp, MAKEMURUN *mmup, MMUWORK *mmuw)
{ /*	This performs the calculations that enable us to calculate the mean
	intensity and sigma for the whole reflection, the primary, and the
	Bijvoet. Currently, we don't attempt to write reflections out in this
	routine; instead, we content ourselves with constructing the MUREFLs.*/
	register int	i, j;
	register ABBROB	*aobp;
	register SHIFT	*tsh;
	register SHELL	*shp, *shp1;
	int		nused, run_num, shellno, shind;
	double		dellog, e0, e0p, e1, e1e, e1p, e0b, e1b, e2e;
	double		idenelt, ielt, mst2, res, ski, stscl, stwt;
	double		wg, wi, wt, wtn, wtp, *fp2;
	float		*fp1;
	short		lphkl[3], newhkl[3];
	xgulong		ulsh;
	char		iscen;
	CRYSTAL		*lhed;
	MUREFL		*tmu;

	lhed = &(mmup->mm_xhed);
	if ((res = mrp->mr_stl) > mmuw->mm_stlmax)
	  {
		mmuw->mm_nrtoohi++; mmuw->mm_nobtoohi += mrp->mr_nob; return;
	  }
	mmuw->mm_yy = mmuw->mm_yden = nused = 0;	/* zero out the sums */
	mmuw->mm_iplus = mmuw->mm_iminus = mmuw->mm_ipden =
		mmuw->mm_imden = mmuw->mm_npused = mmuw->mm_nmused = 0;
	mmuw->mm_e2 = mmuw->mm_e2p = mmuw->mm_e2b = 0.;
	e0 = e1 = e0p = e1p = e0b = e1b = 0.;
	(mmuw->mm_nrefin)++;	(mmuw->mm_nrnmin)++;
	mmuw->mm_nobin += mrp->mr_nob;
	if (mrp->mr_nobpri && mrp->mr_nobbij) mmuw->mm_nrnmin++;
	iscen = iscentric(mrp->mr_hkl, lhed->xt_intl);
	shp = mmup->mm_shlb + (shellno = gshell(mrp->mr_stl, mmuw));
	shp1 = (shellno < mmuw->mm_nshells - 1 ? shp + 1 : shp);
	if (shp->sh_iratio < FUZZ || shp1->sh_iratio < FUZZ)
	  {
		if (iscen) (mmuw->mm_ncenallbad)++;
		else (mmuw->mm_nacenallbad)++; return;
	  }
	if (shellno == mmuw->mm_nshells - 1) stscl = shp->sh_iratio;
	else	stscl = (shp1->sh_iratio - shp->sh_iratio) * mmuw->mm_stl3 *
		 (double)mmuw->mm_nshells + shp->sh_iratio *
		 (double)(shellno + 1) - shp1->sh_iratio * (double)shellno;
	for (i = 0; i < (int)mrp->mr_nob; i++) /* compute mean intensity */
	   {
		aobp = &(mrp->mr_ob[i]);
		ulsh = LSH(aobp->abo_runno, 24) |
			LSH(aobp->abo_chamno, 16) | aobp->abo_shift;
		if (0 > (shind = find_shift(mmup->mm_shlp, ulsh,
		 (int)(mmup->mm_nscales))))
		  {
			mmuw->mm_nnoshift++;	continue;
		  }
		tsh = mmup->mm_shlp + shind;
		ski = terpscale(aobp->abo_omob, tsh, res) / stscl;
		if (ski < DFUZZ) continue;
		if (mmup->mm_vshp)
		  {
			run_num = aobp->abo_shift / 200;
			refvec(aobp, mmup->mm_rub + run_num,
			 mmuw->mm_vecflags, mmuw->mm_nvpars, mmuw->mm_avec);
			fp1 = &((mmup->mm_vshp + run_num)->v_coef[0]);
			for (j = mmuw->mm_nvpars, fp2 = &(mmuw->mm_avec[0]),
			 wg = 0.; j--; fp1++, fp2++)
				wg += ((double)(*fp1)) * *fp2;
			ski *= wg;
		  }
		if (ski < DFUZZ) continue;
		wi = aobp->abo_sigma;	ielt = aobp->abo_ii / ski;
		wi = 1./(wi * wi + ielt * (ielt * tsh->shvar));
		e1e = ielt * wi;	e2e = ielt * ielt * wi;
		e0 += wi;	e1 += e1e;	mmuw->mm_e2 += e2e;
		idenelt = ski * ski * wi;	ielt *= idenelt;
		mmuw->mm_yy += ielt;
		mmuw->mm_yden += idenelt;	nused++;
		if (i < (int)mrp->mr_nobpri)
		  {
			mmuw->mm_iplus += ielt;	mmuw->mm_ipden += idenelt;
			mmuw->mm_npused++;	e0p += wi;	e1p += e1e;
			mmuw->mm_e2p += e2e;
		  }
		else {
			mmuw->mm_iminus += ielt;
			mmuw->mm_imden += idenelt;	mmuw->mm_nmused++;
			e0b += wi;	e1b += e1e;	mmuw->mm_e2b += e2e;
		     }
	   }
	/* we output the reflection only if the statistics are okay. */
	if (nused < 1 || mmuw->mm_yden < DFUZZ)
	  {
		if (iscen) mmuw->mm_ncenallbad++;
		else mmuw->mm_nacenallbad++;	return;
	  }
	if ((mmup->mm_flags & MUF_MULOBSERVED) && nused < 2)
	  {
		if (iscen) mmuw->mm_ncensingly++;
		else mmuw->mm_nacensingly++;	return;
	  }
	mmuw->mm_flagval = (iscen ? FL_CENTRIC : 0);
	mmuw->mm_nobused += nused; mmuw->mm_nrout++;
	if (mmup->mm_flags & MUF_VAROFMEAN)
		mmuw->mm_e2 = mmuw->mm_e2p = mmuw->mm_e2b = 1.;
	else {
		mmuw->mm_e2 = (nused < 2) ? shp->sh_gof : ((e0 <= DFUZZ) ? 1. :
		 (mmuw->mm_e2 - e1 * (e1 / e0)) /
		 (double)((nused - 1) * nused));
		mmuw->mm_e2p = (mmuw->mm_npused < 2 ?
		 shp->sh_gof : (e0p <= DFUZZ) ? 1. :
		 (mmuw->mm_e2p - e1p * (e1p / e0p)) /
		 (double)((mmuw->mm_npused - 1) * mmuw->mm_npused));
		mmuw->mm_e2b = (mmuw->mm_nmused < 2 ? (mmup->mm_shlb)->sh_gof :
		 (e0b <= DFUZZ) ? 1. : (mmuw->mm_e2b - e1b * (e1b / e0b)) /
		 (double)((mmuw->mm_nmused - 1) * mmuw->mm_nmused));
	     }
	munpack(mrp->mr_hkl, lphkl, lhed->xt_intl);
	if (mmup->mm_flags & MUF_REINDEX) /* reindex the refl if requested */
	  {
		if (mm31ir(newhkl, mmup->mm_matrei, lphkl))
		   {
			for (i = 3; i--; )
				lphkl[i] = newhkl[i] + mmup->mm_vecrei[i];
		   }
		else {
			mmuw->mm_nmiss++; return;
		     }
	  }
	/* output Sheldrick-style reflections immediately */
	shelmeans(mmup, mmuw, lphkl);
	/* for MULISTS, stuff the relevant data into a structure for now. */
	if (lhed->xt_isys == 2 && lphkl[2] < 0)
	  {
		lphkl[2] = -lphkl[2];	lphkl[0] = -lphkl[0];
	  }
		/* take some sqrts if we're doing F's */
	if (mmup->mm_flags & MUF_OUTPUT_F) get_fmeans(mmup, mmuw);
	else	get_imeans(mmup, mmuw);
	tmu = mmup->mm_mup;
	for (i = 3; i--; ) tmu->mu_hkl[i] = lphkl[i];
	tmu->mu_stl = mrp->mr_stl;
	tmu->mu_f = mmuw->mm_yy;
	tmu->mu_sig = mmuw->mm_yden;
	tmu->mu_fplus = mmuw->mm_iplus;
	tmu->mu_sigplus = mmuw->mm_ipden;
	tmu->mu_fminus = mmuw->mm_iminus;
	tmu->mu_sigminus = mmuw->mm_imden;
	tmu->mu_flag = mmuw->mm_flagval;
	/* (Aug 2002): for acentrics, calculate values that will
	 enable determination of the relative k and B between F+ and F- */
	if ((!iscen) && (mmuw->mm_nmused) && (mmuw->mm_npused) &&
	 (mmuw->mm_ipden > DFUZZ) && (mmuw->mm_iplus > mmup->mm_scut *
	 mmuw->mm_ipden) && (mmuw->mm_imden > DFUZZ) &&
	 (mmuw->mm_iminus > mmup->mm_scut * mmuw->mm_imden))
	  {
		wtp = mmuw->mm_ipden / mmuw->mm_iplus;
		wtn = mmuw->mm_imden / mmuw->mm_iminus;
		wt = 1. / (wtp * wtp + wtn * wtn);
		mst2 = mrp->mr_stl;	mst2 *= -mst2;
		stwt = mst2 * wt;
		dellog = (log(mmuw->mm_iplus) - log(mmuw->mm_iminus)) * wt;
		mmuw->mm_kbe0 += wt;
		mmuw->mm_kbe1 += stwt;	mmuw->mm_kbe2 += mst2 * stwt;
		mmuw->mm_kbe3 += dellog;
		mmuw->mm_kbe4 += dellog * mst2;	(mmuw->mm_kbn)++;
	  }
	(mmup->mm_mup)++;
	if (iscen)		(mmuw->mm_ncentric)++;
	else if (mmuw->mm_nmused)
	  {
		if (mrp->mr_nobpri)	(mmuw->mm_nboth)++;
		else			(mmuw->mm_nbijonly)++;
	  }
	else	if (mmuw->mm_npused)	(mmuw->mm_nprionly)++;
	return;
}

static void summkmurun(FILE *fp, MAKEMURUN *mmup, MMUWORK *mmuw, time_t tl1)
{ /* Summarize the run of makemu */
	CRYSTAL	*lhed;

	lhed = &(mmup->mm_xhed);
	fprintf(fp,
 "Multiref contains %6d obs of %6d merged, %6d unmerged reflections\n",
	 (int)(lhed->xt_ninteg), (int)(lhed->xt_nmerged),
	 (int)(lhed->xt_nrefls));
	fprintf(fp,
 "We actually count %6d obs of %6d merged, %6d unmerged reflections\n",
		(int)mmuw->mm_nobin, (int)mmuw->mm_nrefin,
		(int)mmuw->mm_nrnmin);
	fprintf(fp,
 " Mulist contains  %6d reflections:          %6d centrics, %6d acentrics\n",
		(int)mmuw->mm_nrout, (int)mmuw->mm_ncentric,
		(int)(mmuw->mm_nboth + mmuw->mm_nbijonly + mmuw->mm_nprionly));
	fprintf(fp,
 " %6d observations were used in determining the mean intensities\n",
		(int)mmuw->mm_nobused);
	fprintf(fp,
 " %6d observations were ignored because no scale factor was input for them\n",
		(int)mmuw->mm_nnoshift);
	fprintf(fp,
 " %6d acentrics were measured + and -;       %6d only +,   %6d only -\n",
		(int)mmuw->mm_nboth, (int)mmuw->mm_nprionly,
		(int)mmuw->mm_nbijonly);
	if (mmuw->mm_nacenallbad || mmuw->mm_ncenallbad)
	  {
		fprintf(fp, " %6d acentrics and %6d",
			(int)mmuw->mm_nacenallbad, (int)mmuw->mm_ncenallbad);
		fprintf(fp,
		 " centrics lacked usable observations and were skipped\n");
	  }
	if (mmuw->mm_nacensingly || mmuw->mm_ncensingly)
	  {
		fprintf(fp, " %6d acentrics and %6d",
			(int)mmuw->mm_nacensingly, (int)mmuw->mm_ncensingly);
		fprintf(fp,
		 " centrics had exactly 1 observation and were skipped\n");
	  }
	if (mmuw->mm_nrtoohi) fprintf(fp,
	 " %6d obs of %6d refls were beyond the resolution limit\n",
			(int)mmuw->mm_nobtoohi, (int)mmuw->mm_nrtoohi);
	if (mmuw->mm_nmiss) fprintf(fp,
	 " %6d refls had non-integer output indices and were dropped\n",
		mmuw->mm_nmiss);
	if (mmuw->mm_r1 > DFUZZ)
	  {
		fprintf(fp, "Anomalous-data k, B, and R-factor analysis:\n");
		fprintf(fp, " #Matched #Matched    Relative_kb: ");
		fprintf(fp, "Rval,all data  Rval,F/s>%4.1f\n", mmup->mm_scut);
		fprintf(fp, "    Refls F/s>%4.1f        k        B ",
			mmup->mm_scut);
		fprintf(fp, "     Rwtd    Runwtd      Rwtd    Runwtd\n");
		fprintf(fp, " %8d %8d %8.3f %8.4f %9.4f %9.4f %9.4f %9.4f\n",
		 mmuw->mm_nrfc, mmuw->mm_nrfcb, mmuw->mm_relk, mmuw->mm_relb,
		 mmuw->mm_r1, mmuw->mm_r3, mmuw->mm_r5, mmuw->mm_r7);
		if (mmuw->mm_nrfbadi) fprintf(fp,
		 " %5d reflections had bad sigmas\n", mmuw->mm_nrfbadi);
		if (mmuw->mm_nrfbadfl) fprintf(fp,
		 " %5d reflections were centric or had F+ or F- unmeasured\n",
			mmuw->mm_nrfbadfl);
	  }
	else fprintf(fp, "Anomalous data not calculated: determinant zero\n");
	fprintf(fp, "Program required  %6d seconds of runtime\n",
		(int)(tl1 - mmup->mm_tl0));
}

static void shl_calcs(MERGEDREF *mrp, MAKEMURUN *mmup, MMUWORK *mmuw)
{ /*	This does the calculations appropriate to the reflection "mrp" allowing
	for updating of the shell intensity and goodness-of-fit sums.*/

	register int	j, n, shellval, shind;
	register SHIFT	*tsh;
	register SHELL	*shp;
	register ABBROB	*aobp;
	int		irun, nusable, run_num;
	double		e0, e1, ielt, locscale, rawi, rawscale, res, rm;
	double		rsqm, scaling_ratio, scrawi, wg, wi, *fp2;
	float		*fp1;
	xgulong		ulsh;

		/* exclude high-resolution values from this calc */
	if (mrp->mr_stl > mmuw->mm_stlmax) return;
	shellval = gshell(mrp->mr_stl, mmuw);
	res = mrp->mr_stl;	shp = mmup->mm_shlb + shellval;
	e0 = e1 = mmuw->mm_e2 = nusable = 0;
	for (n = 0, aobp = &(mrp->mr_ob[0]); n < (int)mrp->mr_nob; n++, aobp++)
	   {
		ulsh = LSH(aobp->abo_runno, 24) | LSH(aobp->abo_chamno, 16) |
			aobp->abo_shift;
		if (0 <= (shind = find_shift(mmup->mm_shlp, ulsh,
		 (int)(mmup->mm_nscales))))
		  {
			irun = aobp->abo_runno;
			rm = mmuw->mm_resmean[irun];
			rsqm = mmuw->mm_resqmean[irun];
			tsh = mmup->mm_shlp + shind;
			rawscale = tsh->shg + tsh->sha * rm + tsh->shb * rsqm;
			locscale = terpscale(aobp->abo_omob, tsh, res);
			/* the values associated with the geometry-dependent
			parameters are determined by recognizing that the
			mean value of the multiplier of each of these
			parameters, except for the 0th and 9th ones, is zero.
			The zeroth parameter is 1 and the 9th is
			sqrt(10/3)*<s^2>, so the calcs are easy... */
			if (mmup->mm_vshp)
			  {
				run_num = aobp->abo_shift / 200;
				refvec(aobp, mmup->mm_rub + run_num,
					mmuw->mm_vecflags,
					mmuw->mm_nvpars, mmuw->mm_avec);
				fp1 = &((mmup->mm_vshp + run_num)->v_coef[0]);
				for (j = mmuw->mm_nvpars,
				 fp2 = &(mmuw->mm_avec[0]), wg = 0.;
				 j--; fp1++, fp2++) wg += *fp1 * *fp2;
				locscale *= wg;	fp1 -= mmuw->mm_nvpars;
				rawscale *= fp1[0] + fp1[9] * rsqm *
				 mmuw->mm_rsqcoef;
			  }
			if (rawscale > DFUZZ && locscale > DFUZZ)
			  {
				rawi = aobp->abo_ii;
				scrawi = rawi / rawscale;
				ielt = rawi / locscale;
				scaling_ratio = (fabs(ielt) < DFUZZ ?
					0 : scrawi / ielt);
				if (scrawi > MIN_USABLE && ielt > MIN_USABLE
				 && 0.2 < scaling_ratio && scaling_ratio < 5.)
				  {
					(shp->sh_nob)++;
					(shp->sh_int) += scrawi;
					(shp->sh_scint) += ielt;
					if ((wi = aobp->abo_sigma) > 0.1)
					  {
						wi = 1. / (wi * wi + ielt *
							(ielt * tsh->shvar));
						e0 += wi;
						e1 += wi * ielt;
						mmuw->mm_e2 += wi * ielt * ielt;
						nusable++;
					  } /* end of "if wi" clause */
				  } /* end of "if scrawi" clause */
			  } /* end of "if rawscale " clause */
		  } /* end of "if 0 <= shind" clause */
	   } /* end of "for (n " loop */
	if ((nusable > 1) && e0 > DFUZZ)
	  {
		shp->sh_gof += (mmuw->mm_e2 - e1 * (e1 / e0)) /
			(double)((nusable-1) * nusable);
		shp->sh_nref++;
	  }
}

static void shl_header(FILE *fp)
{ /* 	This prints a header for the shell statistics on the stream `fp' */
	fprintf(fp,"\t\tSummary of Input Intensities by Shells:\n");
	fprintf(fp,
 "Shell No. of  sin(theta)/lambda       Mean Intensity   # multiply\n");
	fprintf(fp,
 "    # observ      min      max     raw  scaled    ratio obs.refls  g.o.f.\n");
}

static void disp_shell(FILE *fp, SHELL *shp)
{ /* This displays information about shell `shp' on the stream fp. */

	fprintf(fp,"%5d %6d %8.5f %8.5f %7.2f %7.2f %8.5f %9d %7.3f\n",
	 shp->sh_index, shp->sh_nob, shp->sh_resmin, shp->sh_resmax,
	 shp->sh_int, shp->sh_scint, shp->sh_iratio, shp->sh_nref, shp->sh_gof);
}

static int shl_inten(MAKEMURUN *mmup, MMUWORK *mmuw)
{ /*	In this routine we compute the mean one-parameter-per-shift-scaled
	intensity in every resolution shell and save those normalized means
	to reapply to the data.
	XENGEN V.1.3: this routine also computes the mean value of the
	goodness-of-fit in each of the shells. */
	register int	shellval;
	off_t		toskip;
	double		rn;
	MERGEDREF	mr;
	SHELL		*shp;
	CRYSTAL		*lhed;
	static size_t	sizdatarun = 256;

/* see how many shells we will have.  We set up between 10 and 50 shells,
	depending on how many input data there are. */
	lhed = &(mmup->mm_xhed);
	mmuw->mm_nshells = (lhed->xt_nmerged < 10000 ? 10 :
	 (lhed->xt_nmerged > 50000 ?  50 : (lhed->xt_nmerged / 1000)));
	if (NULL == (AVOID *)(mmup->mm_shlb = (SHELL *)callop(mmuw->mm_nshells,
	 sizeof (SHELL), "mmup->mm_shlb in makemu"))) return EALLOC;
/* Now read through the input data to accumulate the mean intensities. */
	/* ignore observations with I < -100 and 0.2 < I(scale)/I(raw) < 5. */
	while (0 < xgread((void *)&mr, XG_MERGEDREF, mmup->mm_fpmi))
		shl_calcs(&mr, mmup, mmuw);
	shl_header(mmup->mm_fpl);	shl_header(stdout);
	(mmup->mm_shlb)->sh_resmin = 0.;
	for (shellval = 0, shp = mmup->mm_shlb;
	 shellval < mmuw->mm_nshells; shellval++, shp++)
	   {
		shp->sh_index = shellval;
		if (shellval) shp->sh_resmin = (shp-1)->sh_resmax;
		shp->sh_resmax = mmuw->mm_stlmax * mycbrt((double)(shellval+1)
		 / (double)mmuw->mm_nshells);
		   if ((rn = shp->sh_nob))
		     {
			shp->sh_int /= rn;	shp->sh_scint /= rn;
			shp->sh_iratio = (shp->sh_scint > FUZZ ?
				shp->sh_int / shp->sh_scint : 1.);
			if (shp->sh_nref) shp->sh_gof /= (float)shp->sh_nref; 
			else	shp->sh_gof = 1.;
			disp_shell(mmup->mm_fpl, shp);
			disp_shell(stdout, shp);
			if (mmup->mm_flags & MUF_PRESERVE) shp->sh_iratio = 1.;
		     }
	   }
	toskip = sizeof (CRYSTAL) + lhed->xt_nruns * sizdatarun;
	if (-1 == fseek(mmup->mm_fpmi, (long)toskip, 0))
	  {
		fprintf(stderr, "makemu: failure to rewind multiref\n");
		return ESEEKERR;
	  }
	xeflush(mmup->mm_fpl);
	return EOK;
}

static int countboth(MAKEMURUN *mmup)
{ /*	This returns the number of reflections that will be written out
	in the circumstance where both primaries and Bijvoets exist */
	int	totct;
	MUREFL	*tmu, *muend;

	muend = mmup->mm_mup;
	for (totct = 0, tmu = mmup->mm_mur0; tmu < muend; tmu++)
	   {
		if (!(tmu->mu_flag & 02)) totct++;
		if (!(tmu->mu_flag & 05)) totct++;
	   }
	mmup->mm_mup = tmu;	return totct;
}

static void reorder_out(MAKEMURUN *mmup, MMUWORK *mmuw)
{ /*	This re-sorts the reflections (if necessary) and then
	outputs the reflections. It also calculates the k and B values
	between the primaries and the Bijvoets in the dataset */
	char		dtchar, *shname;
	static char	*cnsfmt;
	size_t		nmus;
	double		determ, df, dfw, dfw2, f1, f2, f2p, fw, fw2;
	double		r0, r2, r4, r6, s1, s2, sigcns, st, wt;
	MUREFL		*muend, *tmu;
	static char	xpfmt[] =
		"INDEx %4hd %4hd %4hd %cOBS %8.2f SIGMa %8.2f\n";
	static char	cnfmtf[] =
		"INDEx= %4hd %4hd %4hd FOBS= %8.2f SIGMA= %8.2f\n";
	static char	cnfmti[] =
		"INDEx= %4hd %4hd %4hd IOBS=%9.2f SIGI=%9.2f\n";
#ifdef VMS
	static u_short	hklbuf[] = {3,
			DSC$K_DTYPE_B, SO_ASCENDING, 0, 1,
			DSC$K_DTYPE_B, SO_ASCENDING, 1, 1,
			DSC$K_DTYPE_B, SO_ASCENDING, 2, 1 };
	static u_short	stlbuf[] = {1,
			DSC$K_DTYPE_F, SO_ASCENDING, 4, 4 };
#endif

	muend = mmup->mm_mup;	nmus = muend - mmup->mm_mur0;
	if ((mmup->mm_flags & MUF_REORDER) && (nmus > 1))
	  {
		printf("About to sort the reflections: ");
#ifdef VMS
		vsort((AVOID *)mmup->mm_mur0, nmus, sizeof (MUREFL),
		 (SORT_KEY *)((mmup->mm_flags & MUF_ABSORDER) ?
		 hklbuf : stlbuf), NULL);
#else
		if (mmup->mm_flags & MUF_ABSORDER) qsort(
		 (AVOID *)mmup->mm_mur0, nmus, sizeof (MUREFL), (QPE)murcmp);
		else if (mmup->mm_flags & MUF_STLORDER) qsort(
		 (AVOID *)mmup->mm_mur0, nmus, sizeof (MUREFL), (QPE)stlcmp);
#endif
		printf(" Sort completed\n");
	  }
	shname = filepart(mmup->mm_cnsname, 0);
	fprintf(mmup->mm_fpc, "{ * CNS file %s converted from ", shname);
	shname = filepart(mmup->mm_mrname, 0);
	fprintf(mmup->mm_fpc, "X-GEN file %s\n", shname);
	fprintf(mmup->mm_fpc, "    Converted by program makemu *}\n");
	if (mmup->mm_flags & MUF_XPLORSEP)
	  {
		fprintf(mmup->mm_fpc, "NREFLections= %d\n", countboth(mmup));
		fprintf(mmup->mm_fpc, "ANOMalous= TRUE\n");
	  }
	else {
		fprintf(mmup->mm_fpc, "NREFLections= %d\n", (int)nmus);
		fprintf(mmup->mm_fpc, "ANOMalous= FALSe\n");
	     }
	if (mmup->mm_flags & MUF_OUTPUT_F)
	  {
		fprintf(mmup->mm_fpc,
		 "DECLare NAME=FOBS  DOMAin=RECIprocal TYPE=REAL END\n");
		fprintf(mmup->mm_fpc,
		 "DECLare NAME=SIGMA DOMAin=RECIprocal TYPE=REAL END\n");
	  }
	else {
		fprintf(mmup->mm_fpc,
		 "DECLare NAME=IOBS  DOMAin=RECIprocal TYPE=REAL END\n");
		fprintf(mmup->mm_fpc,
		 "DECLare NAME=SIGI  DOMAin=RECIprocal TYPE=REAL END\n");
	     }
	if (mmup->mm_flags & MUF_XPLORSEP)
	  {
		dtchar = (mmup->mm_flags & MUF_OUTPUT_F) ? 'F' : 'I';
		cnsfmt = (mmup->mm_flags & MUF_OUTPUT_F) ? cnfmtf : cnfmti;
		for (tmu = mmup->mm_mur0; tmu < muend; tmu++)
		   {
			if (!(tmu->mu_flag & 02))
			  {
				sigcns = tmu->mu_sigplus;
				if (sigcns > 9998.) sigcns = 0.;
				fprintf(mmup->mm_fpx, xpfmt,
			 	 tmu->mu_hkl[0], tmu->mu_hkl[1],tmu->mu_hkl[2],
			 	 dtchar, tmu->mu_fplus, tmu->mu_sigplus);
				fprintf(mmup->mm_fpc, cnsfmt,
				 tmu->mu_hkl[0], tmu->mu_hkl[1],tmu->mu_hkl[2],
				 tmu->mu_fplus, sigcns);
			  }
			if (!(tmu->mu_flag & 05))
			  {
				sigcns = tmu->mu_sigminus;
				if (sigcns > 9998.) sigcns = 0.;
				fprintf(mmup->mm_fpx, xpfmt,
				 -(tmu->mu_hkl[0]), -(tmu->mu_hkl[1]),
				 -(tmu->mu_hkl[2]), dtchar,
				 tmu->mu_fminus, tmu->mu_sigminus);
				fprintf(mmup->mm_fpc, cnsfmt,
				 -(tmu->mu_hkl[0]), -(tmu->mu_hkl[1]),
				 -(tmu->mu_hkl[2]), tmu->mu_fminus, sigcns);
			  }
		   }
	  }
	else if (mmup->mm_flags & MUF_XPLORMER)
	  {
		dtchar = (mmup->mm_flags & MUF_OUTPUT_F) ? 'F' : 'I';
		cnsfmt = (mmup->mm_flags & MUF_OUTPUT_F) ? cnfmtf : cnfmti;
		for (tmu = mmup->mm_mur0; tmu < muend; tmu++)
		 {
			sigcns = tmu->mu_sig;
			if (sigcns > 9998.) sigcns = 0.;
			fprintf(mmup->mm_fpx, xpfmt,
			 tmu->mu_hkl[0], tmu->mu_hkl[1], tmu->mu_hkl[2],
			 dtchar, tmu->mu_f, tmu->mu_sig);
			fprintf(mmup->mm_fpc, cnsfmt,
			 tmu->mu_hkl[0], tmu->mu_hkl[1], tmu->mu_hkl[2],
			 tmu->mu_f, sigcns);
		   }
	  }
	for (tmu = mmup->mm_mur0; tmu < muend; tmu++)
		(void)outmu(mmup->mm_fpo, tmu);
	printf("%d sorted reflections written out\n",
	 (int)(muend - mmup->mm_mur0));
	/* Now calculate the actual relative k and relative B. */
	determ = mmuw->mm_kbe0 * mmuw->mm_e2 - mmuw->mm_kbe1 * mmuw->mm_kbe1;
	if (fabs(determ) < DFUZZ) return;	/* need positive determinant */
	mmuw->mm_relk = exp((mmuw->mm_e2 * mmuw->mm_kbe3 -
	 mmuw->mm_kbe1 * mmuw->mm_kbe4) / determ);
	mmuw->mm_relb = (mmuw->mm_kbe0 * mmuw->mm_kbe4 -
	 mmuw->mm_kbe1 * mmuw->mm_kbe3) / determ;
	/* get RMS and absolute-value R-factors between the F+ and F- data*/
	mmuw->mm_r1 = mmuw->mm_r3 = mmuw->mm_r5 = mmuw->mm_r7  = 0.;
	r0 = r2 = r4 = r6 = 0.;
	mmuw->mm_nrfc = mmuw->mm_nrfcb = 0;
	mmuw->mm_nrfbadi = mmuw->mm_nrfbadfl = 0;
	for (tmu = mmup->mm_mur0; tmu < muend; tmu++)
	 {
	   if (!(tmu->mu_flag & 07))
	     {
		f1 = tmu->mu_fplus;	s1 = tmu->mu_sigplus;
		f2 = tmu->mu_fminus;	s2 = tmu->mu_sigminus;
		/* flag-value sigmas re-set to something modest */
		if (s1 > 1000.) s1 = 2.;	if (s2 > 1000.) s2 = 2.;
		st = tmu->mu_stl;
		f2p = f2 * mmuw->mm_relk * exp(-mmuw->mm_relb * st * st);
		if ((s1 > DFUZZ) && (s2 > DFUZZ))
		  {
			wt = 1. / (s1 * s1 + s2 * s2);	fw = f1 * wt;
			fw2 = fw * f1;
			df = f1 - f2p;	dfw = df * wt;	dfw2 = dfw * df;
			r0 += fw2;	mmuw->mm_r1 += dfw2;
			r2 += fabs(f1);	mmuw->mm_r3 += fabs(df);
			if (f1 > mmup->mm_scut * s1 && f2 > mmup->mm_scut * s2)
			  {
				r4 += fw2;	mmuw->mm_r5 += dfw2;
				r6 += fabs(f1);	mmuw->mm_r7 += fabs(df);
				mmuw->mm_nrfcb++;
			  }
			mmuw->mm_nrfc++;
		  } /* end if "if ((s1 > DFUZZ) && (s2 > DFUZZ))" clause */
		else	mmuw->mm_nrfbadi++;
	      } /* end of "if (!(tmu->mu_flag & 07)" clause */
	    else mmuw->mm_nrfbadfl++;
	 } /* end of loop through input reflections */
	mmuw->mm_r1 = (r0 > DFUZZ && mmuw->mm_r1 > DFUZZ) ?
	 sqrt(mmuw->mm_r1 / r0) : 1.;
	mmuw->mm_r3 = (r2 > DFUZZ && mmuw->mm_r3 > DFUZZ) ?
	 mmuw->mm_r3 / r2 : 1.;
	mmuw->mm_r5 = (r4 > DFUZZ && mmuw->mm_r5 > DFUZZ) ?
	 sqrt(mmuw->mm_r5 / r4) : 1.;
	mmuw->mm_r7 = (r6 > DFUZZ && mmuw->mm_r7 > DFUZZ) ?
	 mmuw->mm_r7 / r6 : 1.;
}

int endmu(int ina, MAKEMURUN *mmup)
{ /* cleanly terminate a run of makemu */
	if ((stdout != mmup->mm_fpl) && (NULL != mmup->mm_fpl))
		(void)fclose(mmup->mm_fpl);
	memzap((AVOID **)&(mmup->mm_shlp), "mmup->mm_shlp in makemu");
	memzap((AVOID **)&(mmup->mm_vshp), "mmup->mm_vshp in makemu");
	memzap((AVOID **)&mmup->mm_mur0, "mmup->mm_mur0 in makemu");
	memzap((AVOID **)&(mmup->mm_shlb), "mmup->mm_shlb in makemu");
	memzap((AVOID **)&(mmup->mm_rub), "mmup->mm_rub in makemu");
	if (NULL != mmup->mm_fpo)
	  {
		(void)fclose(mmup->mm_fpo);	mmup->mm_fpo = NULL;
	  }
	if (NULL != mmup->mm_fpx)
	  {
		(void)fclose(mmup->mm_fpx);	mmup->mm_fpx = NULL;
	  }
	if (NULL != mmup->mm_fpc)
	  {
		(void)fclose(mmup->mm_fpc);	mmup->mm_fpc = NULL;
	  }
	mmup->mm_fpl = NULL;	globalkill();	return ina;
}

int makemuops(MAKEMURUN *mmup)
{	/* This performs most of the real work in makemu */
	register int	nrun, irun;
	int		gotone, i, result;
	time_t		tl1;
	DATARUN		rhd;
	MERGEDREF	mr;
	RUNWORK		*rup;
	VECPAR		*vpp;
	CRYSTAL		*lhed;
	static MMUWORK	*mmuw;
	static char	mmuwmsg[] = "mmuw in makemuops";

	if (NULL != mmuw) zeromem((AVOID *)mmuw, sizeof (MMUWORK));
	else if (NULL == (mmuw = (MMUWORK *)callop(1,
	 sizeof (MMUWORK), mmuwmsg))) return EALLOC;
	lhed = &(mmup->mm_xhed);
	if ((lhed->xt_isys == 7) && (lhed->xt_intl < 200))
		lhed->xt_intl += 100;
	mmuw->mm_rsqcoef = SQ10OV3 * 4. * lhed->xt_lambda * lhed->xt_lambda;
	for (nrun = 0, rup = mmup->mm_rub; nrun < lhed->xt_nruns; nrun++, rup++)
	   {
		if (verror(0 < xgread((void *)&rhd, XG_DATARUN, mmup->mm_fpmi),
		 "read datarun header from multiref"))
		  {
			memzap((AVOID **)&mmuw, mmuwmsg);
			return endmu(EREADERR, mmup);
		  }
		(void)keep_dimen(&rhd);
		workrun(&rhd, rup, lhed, nrun * 200);
		irun = rhd.run_id;
		mmuw->mm_resqmean[irun] = rhd.run_ssqave /
				(4. * lhed->xt_lambda * lhed->xt_lambda);
		mmuw->mm_resmean[irun] = sqrt(mmuw->mm_resqmean[irun]);
	   }
	/* specified resolution limit takes effect if the plane spacing is
	LARGER (i.e. more restrictive) than that contained in the multiref */
	if (mmup->mm_specres > lhed->xt_reslim)
		lhed->xt_reslim = mmup->mm_specres;
	mmuw->mm_stlmax = 0.5 / lhed->xt_reslim;
	/* examine the scale-factor output to see if we actually got any
	 geometry-dependent coefficients. */
	gotone = 0;
	for (vpp = mmup->mm_vshp; vpp < mmup->mm_vshp + lhed->xt_nruns; vpp++)
	   {
		for (i = gotone = 0; i < MAXVECPARS; i++)
		   if (fabs(vpp->v_coef[i]) > FUZZ)
		     {
			gotone = 1; break;
		     }
		if (gotone) break;
	   }
	if (!gotone)
	  {
		memzap((AVOID **)&(mmup->mm_vshp), "vshp in makemumain");
		mmuw->mm_vecflags = mmuw->mm_nvpars = 0 ;
	  }
	else {
		if ((mmup->mm_vshp)->v_ptype == 'd')
		  {
			mmuw->mm_vecflags = SC_DIFFRACTION;
			mmuw->mm_nvpars = 10;
		  }
		else {
			mmuw->mm_nvpars = 9;
			mmuw->mm_vecflags = SC_EXITBEAM;
		     }
	/* need spline information in this case to do the xy conversions */
		if (NULL == gspline(5, &i))
		  {
			memzap((AVOID **)&mmuw, mmuwmsg);
			return endmu(i, mmup);
		  }
	     }
	/* If there is more than one param per shift, compute mean
	 one-param-scaled intensities in each resolution shell so that
	 we can reapply the resolution dependence of the initial data
	 onto the final data.*/
	if (EOK != (result = shl_inten(mmup, mmuw)))
	  {
		memzap((AVOID **)&mmuw, mmuwmsg); return endmu(result, mmup);
	  }
	/* read the reflections.  Get the mean intensity and sigma
	for the reflection and (individually) for the primary and Bijvoet.
	If the reflection is centric, nothing is computed for the Bijvoet,
	and we write an intensity of 0 and a sigma of 9999 for the Bijvoet.
	If the reflection is acentric and either the primary or the Bijvoet
	is missing, we write an intensity of 0 and a sigma of 9999 in the
	appropriate slot.  */
	headsca(mmup->mm_fpsca, &mmup->mm_xhed);
	headsca(mmup->mm_fpasca, &mmup->mm_xhed);
	while (0 < xgread((void *)&mr, XG_MERGEDREF, mmup->mm_fpmi))
	  calc_refl(&mr, mmup, mmuw); /* do calcs to assemble reflections */
	(void)fclose(mmup->mm_fpmi);	mmup->mm_fpmi = NULL;
	/* we're done with Sheldrick and SCALEPACK formats */
	(void)fclose(mmup->mm_fpshel);	mmup->mm_fpshel = NULL;
	(void)fclose(mmup->mm_fpsca);	mmup->mm_fpsca = NULL;
	(void)fclose(mmup->mm_fpasca);	mmup->mm_fpasca = NULL;
	reorder_out(mmup, mmuw);	/* sort reflections and output them */
	tl1 = NOW;
	summkmurun(stdout, mmup, mmuw, tl1); /* summarize results */
	if (stdout != mmup->mm_fpl) summkmurun(mmup->mm_fpl, mmup, mmuw, tl1);
	memzap((AVOID **)&mmuw, mmuwmsg);	return endmu(EOK, mmup);
}

int makemumain(int argc, char *argv[])
{
	MAKEMURUN	mmur;

	/* get runtime args and open files */
	if (EOK != setup_mu(argc, argv, &mmur))
		return endmu(EOPENERR, &mmur);
	sum_muset(stdout, &mmur);
	if (stdout != mmur.mm_fpl) sum_muset(mmur.mm_fpl, &mmur);
	return makemuops(&mmur);
}

/* end of make_mu.c */
