/* mr_merge.c -- merge MULTIREFs from multiple data runs into one MULTIREF

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	24 Jun 2006: include xgpostint.h: relevant structures are there
	20 Nov 2005: inline argument declarations and strict prototyping
	17 Oct 2005: replace static globals with elements of MRMERWORK struct
	10 Oct 2005: housekeeping
	11 May 2005: check return status from xgjournal
	11 Apr 2005: name of this file changed from mrmerge.c to mr_merge.c
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	27 Aug 2004: routines used only here are declared as static;cosmetics
	21 Jan 2004: license message inserted and initial comments reordered
	 7 Jan 2004: static globals replace extern globals
	19 Sep 2003: origisys, origspaceg initialized before use
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Aug 2003: replace global uphkl[3] with static lphkl[3];
		requires a few extra calls to hunpack()
	 5 Jan 2003: reorganized source-file header
	18 Apr 2002: eliminate_dups() moved to ../xglib/mrutil.c
		so that "reduce" can take advantage of it too
	25 Mar 2002: removed MSI attachments
	13 Oct 1995: xt_ninteg typecast in printout
	31 Aug 1995: fix bug that caused miscalc. of cell dimensions on output
		when functionality is called >= once per X-GEN invocation
	31 May 1995: "register" -> "register int" throughout;
		replace some longs with ints;
		eliminate_dups and putmref are declared `void';
		replace %ld with %d
	 2 Mar 1995: diagnostics to stdout
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	27 Oct 1994: mrp->mr_isig set to 1 if isig < 0
	11 Oct 1994: bug fix: mr_nob set to zero when we run out of input spots
	 5 Oct 1994: use xgread and xgwrite on various MULTIREF structure types
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: a few I/O changes; global variables moved around;
		bug fix re averaging cell dimensions (?)
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version
	12 Feb 1988: fix -i option and its complement
	 4 May 1987: multiple observations which come from runs which are
		actually just overlaps of the SAME run
		(e.g., frames 1-110 overlapped with frames 90-200)
		are pared out, unless the "-k" (for "keep") option is invoked.
	18 Dec 1986: handle deleted observations properly;
		reorganize logfile header.
	18 Sep 1985: incorporate XENGEN package conventions, simplify input
	 8 Sep 1984: original code based on imerge.c from La Jolla
	
   Purpose: merging together 1 or more input MULTIREFs to create an
	output MULTIREF.
   Cross references: redurefl.c

 This merges one or more multiref files together
 and writes the resulting multiref on the file "mrefout".
 The multirefs can have been created either with previous
 runs of "mrmerge" or with a run of "reduce".
*************************************************************************/
#define		NEED_TIME	1
#define		NEED_ERRNO	1
#include	"xgpostint.h"
#define		MAXFILES	14	/* max # of input files */
#define		SHIFTSPERRUN	200
#define		MIN_SIGMA	(0.05)

 typedef struct {
		int		mrm_keepdups, mrm_nruns, mrm_nshifts;
		int		mrm_mfordel, mrm_ndupstoss, mrm_nobfound;
		int		mrm_nreffound, mrm_nmerfound, mrm_ninfiles;
		int		mrm_runsb4[MAXFILES];
		time_t		mrm_tl0, mrm_tl1;
		double		mrm_maxstl, mrm_hcoeff[6];
		FILE		*mrm_fpl, *mrm_fpo, **mrm_fpdi;
		CRYSTAL		*mrm_xhdp;
		DATARUN		*mrm_rhdp;
		MERGEDREF	*mrm_tank;
		} MRMERWORK;

extern int mrmmain(int argc, char *argv[]);

static void sum_inrun(FILE *fp, int shadd, char *name, MRMERWORK *mmrp)
{ /*	This summarizes the contents of input file "name", with
	shift addend "shadd", in a line of output on stream "fp".  If the
	argument "name" is null, only the first four fields are filled.*/
	static char	*shortname;
	CRYSTAL		*lxp;
	DATARUN		*lrp;

	lxp = mmrp->mrm_xhdp;	lrp = mmrp->mrm_rhdp;
	if (name) /* construct "main body" of filename */
	  {
		shortname = filepart(name, 2);
		fprintf(fp,
 "%3d %7s %6d %6d %6d %6d %3.0f %3.0f %3.0f %4.0f %4.0f %4.0f %s\n",
		 (int)lrp->run_id, lrp->run_code, shadd,
		 (int)lrp->run_ninteg, (int)lxp->xt_nmerged,
		 (int)lxp->xt_nrefls,
		 lxp->xt_ar[0], lxp->xt_ar[1], lxp->xt_ar[2],
		 lxp->xt_ang[0], lxp->xt_ang[1], lxp->xt_ang[2], shortname);
	  }
	else	fprintf(fp, "%3d %7s %6d %6d\n", (int)lrp->run_id,
		 lrp->run_code, shadd, (int)lrp->run_ninteg);
	return;
}

static void sum_mrmres(FILE *fp, MRMERWORK *mmrp)
{ /* Summarize the findings of this effort to the stream fp */
	static char	*tname, *shortname;
	CRYSTAL		*lxp;

	lxp = mmrp->mrm_xhdp;
	fprintf(fp,"\t\tFinal multiref output:\n");
	fprintf(fp,
" FINAL OUTPUT:       # of # of reflect.   Unit cell  al-      gam- File\n");
	fprintf(fp,
"             observations  merg. unmer.   a   b   c  pha beta   ma name\n");
	tname = trname("mrefout");	shortname = filepart(tname, 0);
	fprintf(fp,
	 "%25d %6d %6d %3.0f %3.0f %3.0f %4.0f %4.0f %4.0f %s\n",
		mmrp->mrm_nobfound, (int)lxp->xt_nmerged, (int)lxp->xt_nrefls,
		lxp->xt_ar[0], lxp->xt_ar[1], lxp->xt_ar[2], lxp->xt_ang[0],
		lxp->xt_ang[1], lxp->xt_ang[2], shortname);
	if (mmrp->mrm_mfordel)
	  {
		fprintf(fp,
 "%6d observations were marked for deletion in the input multirefs\n",
		 mmrp->mrm_mfordel);
		fprintf(fp,"\tand remain marked in the output\n");
	  }
	if ((!mmrp->mrm_keepdups) && (mmrp->mrm_ndupstoss))
	  {
		fprintf(fp,
 "%6d observations were identified as duplicates from overlapped runs\n",
			mmrp->mrm_ndupstoss);
		fprintf(fp,"\tand were suppressed from output\n");
	  }
	fprintf(fp, " Total Run Time for Merging Reflections: %6d seconds\n",
		(int)(mmrp->mrm_tl1 - mmrp->mrm_tl0));
	return;
}

static void sum_mrmset(FILE *fp, MRMERWORK *mmrp)
{ /*	This puts a header for the input summaries onto stream fp */
	static char	*flist[] = {"mrefout", "xlog", NULL};
	static char	ftypes[] = { 2, 2, 0};

	header(fp, REDUCTION, flist, ftypes);
	if (mmrp->mrm_keepdups) fprintf(fp,
	 "Duplicate observations from overlapping runs will be kept\n");
	fprintf(fp,"\t\tSummary of input data:\n");
	fprintf(fp,
 "run     run  shift  # of  # of reflect.   Unit cell  al-      gam- File\n");
	fprintf(fp,
 "no.    code addend observ  merg. unmer.   a   b   c  pha beta   ma name\n");
}

static int merheads(int argc, char *argv[], MRMERWORK *mmrp)
{ /*	Open and go thru each input file.  Keep the input crystal header from
	the first run and modify it as more information comes in.  Average the
	unit cell parameters and write out the average into the output file.
	Accumulate a vector "mmrp->mrm_runsb4" containing the CUMULATIVE NUMBER
	of runs prior to the current file--i.e., mmrp->mrm_runsb4[0] = 0,
	mmrp->mrm_runsb4[1] = # of runs in file 0,
	mmrp->mrm_runsb4[2] = # of runs in file 1 + mmrp->mrm_runsb4[1], etc.
	Summarize what's going on to the standard error and the logfile.
	Xengen V.1.3: stick the recomputed cell dimensions back into the
	 final "xhd" structure. */
	char		**dcp, **flags;
	register int	i, j, k;
	short		origisys, origspaceg, shadd;
	int		ninteg, thisn;
	double		reslim;
	float		suma[3], sumang[3], rabc[3], rang[3];
	static CRYSTAL	xhd;		/* structure for crystal header info */
	static DATARUN	rhd;		/* structure for datarun header info */

	if (1 >= argc)
	  {
		fprintf(stderr,
		 "Usage: mrmerge [-i] input_mref1 [input_mref2 [...]]\n");
		return EOK;
	  }
	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	zeromem((AVOID *)mmrp, sizeof (MRMERWORK));
	mmrp->mrm_tl0 = NOW;
	flags = flagproc(argc, argv, "IK");
	ninteg = mmrp->mrm_keepdups = 0;
	reslim = 10000.;
	for (i = 3; i--; ) suma[i] = sumang[i] = 0.;
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
		if ('I' == **dcp || 'K' == **dcp) mmrp->mrm_keepdups = 1;
	for (argc--, argv++; argc; argc--, argv++) if (**argv != '-') break;
	if (NULL == (AVOID *)(mmrp->mrm_fpdi = (FILE **)callop((unsigned)argc,
	 sizeof (FILE *), "mmrp->mrm_fpdi in merheads"))) return EALLOC;
	if (NULL == (mmrp->mrm_fpl = propen("xlog"))) mmrp->mrm_fpl = stdout;
	for (mmrp->mrm_ninfiles = 0; mmrp->mrm_ninfiles < argc;
	 (mmrp->mrm_ninfiles)++)
		if (qerror((AVOID *)(mmrp->mrm_fpdi[mmrp->mrm_ninfiles] =
		 xgopen(argv[mmrp->mrm_ninfiles], "r")),
		 "open an input multiref file")) return EOPENERR;
	sum_mrmset(mmrp->mrm_fpl, mmrp);	sum_mrmset(stdout, mmrp);
	if (qerror((AVOID *)(mmrp->mrm_fpo = xgopen("mrefout", "w+")),
		"open output multiref")) return EOPENERR;
	mmrp->mrm_nruns = mmrp->mrm_nshifts = origisys = origspaceg = 0;
	mmrp->mrm_xhdp = &xhd;	mmrp->mrm_rhdp = &rhd;
	for (i = 0; i < mmrp->mrm_ninfiles; i++) /* read through input files */
	   {
		mmrp->mrm_runsb4[i] = i ?
			mmrp->mrm_runsb4[i-1] + xhd.xt_nruns : 0;
		if (verror(0 < xgread((void *)&xhd, XG_CRYSTAL,
		 mmrp->mrm_fpdi[i]),
		 "read crystal header from input multiref")) return EREADERR;
		if (!i)
		  {
			if (verror(0 < xgwrite((void *)&xhd, XG_CRYSTAL,
			 mmrp->mrm_fpo),
			 "write crystal header to output multiref"))
					return EWRITERR;
			origisys = xhd.xt_isys;
			origspaceg = xhd.xt_intl;
		  }
		else if (xhd.xt_isys != origisys)
			fprintf(stderr,
"\n\nInconsistency: system No. = %1d in file 0, %1d in file %2d\n",
			origisys, xhd.xt_isys, i);
		else if (xhd.xt_intl != origspaceg)
			fprintf(stderr,
"\n\nInconsistency: spacegroup No. = %3d in file 0, %3d in file %2d\n",
			origspaceg, xhd.xt_intl, i);
		for (k = 0; k < xhd.xt_nruns; k++) /* get the run headers */
		   {
			if (verror(0 < xgread((void *)&rhd, XG_DATARUN,
			 mmrp->mrm_fpdi[i]),
			 "read datarun header from input multiref"))
				return EREADERR;
			for (j = strlen(rhd.run_code); j < 7; j++)
			   rhd.run_code[j] = ' '; /* blank-pad for neatness */
			if (verror(0 < xgwrite((void *)&rhd, XG_DATARUN,
			 mmrp->mrm_fpo),
			 "write datarun header to output multiref"))
				return EWRITERR;
			thisn = rhd.run_ninteg;	/* number of obs in the run */
			mmrp->mrm_nshifts += rhd.run_nshifts; /* # of shifts */
			for (j = 0; j < 3; j++) /* sums for ave. cell params */
			   {
				suma[j] += xhd.xt_ar[j] * thisn;
				sumang[j] += xhd.xt_ang[j] * thisn;
			   }
			ninteg += thisn;	/* sum of observations */
			if (reslim > xhd.xt_reslim) reslim = xhd.xt_reslim;
			shadd = SHIFTSPERRUN * (mmrp->mrm_runsb4[i] + k);
			if (k) /* summarize input to logfile */
			  {
				sum_inrun(stdout, shadd, NULL, mmrp);
				sum_inrun(mmrp->mrm_fpl, shadd, NULL, mmrp);
			  }
			else {
				sum_inrun(stdout, shadd, argv[i], mmrp);
				sum_inrun(mmrp->mrm_fpl, shadd, argv[i], mmrp);
			     }
			(mmrp->mrm_nruns)++;
		   }
	   }
	for (j = 0; j < 3; j++)	/* average the unit cell parameters */
	   {
		suma[j] /= ninteg;	sumang[j] /= ninteg;
		xhd.xt_ar[j] = suma[j];	xhd.xt_ang[j] = sumang[j];
	   }
	/* use the averaged parameters to compute new reciprocal-space
	unit cell parameters */
	if (verror(!ucell(suma, sumang, rabc, rang, xhd.xt_isys),
	 "create averaged reciprocal-space unit cell")) return EBADVALUE;
	mmrp->mrm_hcoeff[0] = 0.25 * rabc[0] * rabc[0];
	mmrp->mrm_hcoeff[1] = 0.25 * rabc[1] * rabc[1];
	mmrp->mrm_hcoeff[2] = 0.25 * rabc[2] * rabc[2];
	mmrp->mrm_hcoeff[3] = 0.5 * rabc[0] * rabc[1] * cos(rang[2]);
	mmrp->mrm_hcoeff[4] = 0.5 * rabc[0] * rabc[2] * cos(rang[1]);
	mmrp->mrm_hcoeff[5] = 0.5 * rabc[1] * rabc[2] * cos(rang[0]);
	if (NULL == (AVOID *)(mmrp->mrm_tank = (MERGEDREF *)callop(
	 mmrp->mrm_ninfiles, sizeof (MERGEDREF), "mrm_tank in merheads")))
		return EALLOC;
	fprintf(mmrp->mrm_fpl,
 "Mean unit cell parameters: a = %9.3f,    b = %9.3f,     c = %9.3f\n",
		suma[0], suma[1], suma[2]);
	fprintf(mmrp->mrm_fpl,
 "                       alpha = %9.3f, beta = %9.3f, gamma = %9.3f\n",
		sumang[0], sumang[1], sumang[2]);
	mmrp->mrm_mfordel = mmrp->mrm_ndupstoss = mmrp->mrm_nobfound =
		mmrp->mrm_nreffound = mmrp->mrm_nmerfound = 0;
	mmrp->mrm_maxstl = 0.;		/*set max sin(th)/lam to zero */
	return EOK;
}

static void getmref(int filenum, MRMERWORK *mmrp)
{ /*	This returns a pointer to a multiref reflection
	read from file mmrp->mrm_fpdi[filenum].  If the routine encounters
	the end of the input file, mr.mr_nob will come back 0.
	The processing of the shifts is done here.  */
	register int	i, soffset;
	MERGEDREF	*lmrp;
	ABBROB		*aobp;

	lmrp = mmrp->mrm_tank + filenum;
	if (0 >= xgread((void *)lmrp, XG_MERGEDREF, mmrp->mrm_fpdi[filenum]))
	  {
		lmrp->mr_nob = 0;
		lmrp->mr_hkl[0] = lmrp->mr_hkl[1] =
			lmrp->mr_hkl[2] = lmrp->mr_hkl[3] = 255;
		return;
	  }
	soffset = SHIFTSPERRUN * mmrp->mrm_runsb4[filenum];
	for (i = 0, aobp = &(lmrp->mr_ob[0]); i <(int)lmrp->mr_nob; i++,aobp++)
	   {
		if (aobp->abo_shift < 0)
		  {
			aobp->abo_shift -= soffset;	(mmrp->mrm_mfordel)++;
		  }
		else	aobp->abo_shift += soffset;
	   }
}

static int getmin(MERGEDREF *tank0, MRMERWORK *mmrp)
{ /*	This returns the index of the tank containing the reflection with
	the lowest-order hkl.  Tanks which have already been emptied have a
	fake hkl == [255,255,255,255] so they never come back as the lowest. */
	register int	i, j;
	int		itank;
	u_char		hkl0[4];
	MERGEDREF	*mrtank;

	itank = 0;	/* as in "itank you very much"...*/
	mrtank = tank0;	/* rewind tanks to first one */
	for (j = 0; j < 4; j++) hkl0[j] = mrtank->mr_hkl[j];
	for (i = 1; i < mmrp->mrm_ninfiles; i++)
	   {
		if (hklcmp((++mrtank)->mr_hkl,hkl0) < 0)
		  {
			for (j = 0; j < 4; j++) hkl0[j] = mrtank->mr_hkl[j];
			itank = i;
		  }
	   }
	mmrp->mrm_tank = tank0;	/* rewind again */
	return itank;
}

static void loadmr(MERGEDREF *holdp, MERGEDREF *tankp,
 int filenum, MRMERWORK *mmrp)
{ /*	This loads the appropriate tanked reflection pointed to by tankp
	into the MERGEDREF structure pointed to by holdp.
	The sin theta/lambda element of holdp is checked first; if it's (-1),
	then we simply stuff mrm_tank[indx] into holdp.  If it's not (-1),
	then we add the observations in mrm_tank[indx] to holdp. The indexing
	gets complicated if there are both primaries and bijvoets to contend
	with. That won't usally be true.  After the tanked reflection gets
	loaded or added, we refill the tank with a new reflection from the
	appropriate file. */
	register int	i;
	int		holdpri, tankpri, tothold;

	if (holdp->mr_stl > -0.9) /* we're adding to an existing refl */
	  {
		holdpri = holdp->mr_nobpri;	tankpri = tankp->mr_nobpri;
		tothold = holdp->mr_nob;
		/* first slide the Bijvoets up to make room for primaries*/
		for (i = tothold - 1; i >= holdpri; i--)
			cpymem((AVOID *)&(holdp->mr_ob[i+tankpri]),
				(AVOID *)&(holdp->mr_ob[i]), sizeof (ABBROB));
		/* now stick in the primaries */
		cpymem((AVOID *)&(holdp->mr_ob[holdpri]),
		 (AVOID *)&(tankp->mr_ob[0]), tankpri * sizeof (ABBROB));
		/* now stick the Bijvoets onto the end */
		cpymem((AVOID *)&(holdp->mr_ob[tothold + tankpri]),
		  (AVOID *)&(tankp->mr_ob[tankpri]),
			(unsigned)(tankp->mr_nobbij * sizeof (ABBROB)));
		holdp->mr_nobpri += tankpri;
		holdp->mr_nobbij += tankp->mr_nobbij;
		holdp->mr_nob += tankp->mr_nob;
	  }
	else	/* we're loading a new reflection */
	  cpymem((AVOID *)holdp, (AVOID *)tankp,
		(unsigned)(REFBYTES + tankp->mr_nob * sizeof (ABBROB)));
	getmref(filenum, mmrp);		/* refill the tank */
}

static float err_sigma(ABBROB *abp, short lhkl[], MRMERWORK *mmrp)
{ /*	Print out a message describing an erroneous sigma to the stream fp. */
	fprintf(mmrp->mrm_fpl,
	 "[%3d,%3d,%3d], run %2d, chamber %1d, shift %5d",
		lhkl[0], lhkl[1], lhkl[2], abp->abo_runno,
		abp->abo_chamno, abp->abo_shift);
	fprintf(mmrp->mrm_fpl," has I = %8.2f, sigma = %8.3f\n",
		abp->abo_ii, abp->abo_sigma);
	abp->abo_sigma = (abp->abo_ii < 40 ? 40 : abp->abo_ii);
	fprintf(mmrp->mrm_fpl," sigma reset to %8.3f\n", abp->abo_sigma);
	return abp->abo_sigma;
}

static void putmref(FILE *fp, MERGEDREF *mrp, MRMERWORK *mmrp)
{ /*	This does the output of each refl to disk and cleans up a little.
	The mean intensity and sigma are recomputed here if more than one
	observation is present.
	In every case the sin(theta)/lambda value is recomputed.*/
	register int	i;
	register ABBROB	*abp;
	short		lphkl[3];	/* un-packed hkl of current refl */
	double		stlsq, weight, thisi, wtsum, isum, i2sum;

	(mmrp->mrm_nmerfound)++; /* increment count of Bijvoet-merged refls */
	(mmrp->mrm_nreffound)++; /* increment count of unbijvoet-merged refls*/
		/* get unpacked indices */
	munpack(mrp->mr_hkl, lphkl, (mmrp->mrm_xhdp)->xt_intl);
	stlsq =	lphkl[0] * lphkl[0] * mmrp->mrm_hcoeff[0] +
		lphkl[1] * lphkl[1] * mmrp->mrm_hcoeff[1] +
		lphkl[2] * lphkl[2] * mmrp->mrm_hcoeff[2] +
		lphkl[0] * lphkl[1] * mmrp->mrm_hcoeff[3] +
		lphkl[0] * lphkl[2] * mmrp->mrm_hcoeff[4] +
		lphkl[1] * lphkl[2] * mmrp->mrm_hcoeff[5];
	mrp->mr_stl = (stlsq > 0. ? sqrt(stlsq) : 0.);
	if ((!mmrp->mrm_keepdups) && ((mrp->mr_nobpri > 1) ||
	 (mrp->mr_nobbij > 1))) mmrp->mrm_ndupstoss += eliminate_dups(mrp);
	if (mrp->mr_nob > 1)
	  {
		wtsum = isum = i2sum = 0;
		for (i = 0; i < (int)mrp->mr_nob; i++)
		   {
			abp = &(mrp->mr_ob[i]); weight = abp->abo_sigma;
			if (weight < MIN_SIGMA)
				weight = err_sigma(abp, lphkl, mmrp);
			weight = 1./(weight * weight);
			thisi = abp->abo_ii;	wtsum += weight;
			isum += thisi * weight;
			i2sum += thisi * (thisi * weight);
		   }
		if (wtsum > DFUZZ)
		  {
			isum /= wtsum;
			i2sum = (i2sum / wtsum - isum * isum) /
				 ((double)(((int)mrp->mr_nob) - 1));
			if (isum < 0.) isum = 0;
			else if (isum > 65534.) isum = 65534.;
			mrp->mr_ibar = isum + 0.4999;
			mrp->mr_isig = (i2sum >= 0.) ? sqrt(i2sum) +0.499: 1.;
		  }
	  }
	 /* if primary&bijvoet are both present, increment that count again */
	if (mrp->mr_nobpri && mrp->mr_nobbij) (mmrp->mrm_nreffound)++;
	if (mrp->mr_stl > mmrp->mrm_maxstl) mmrp->mrm_maxstl = mrp->mr_stl;
	if (0 >= xgwrite((void *)mrp, XG_MERGEDREF, fp)) return; /* write out*/
	mrp->mr_stl = -1;	/*reset the output-empty flag*/
	mmrp->mrm_nobfound += mrp->mr_nob;	/* incr.count of observations */
}

static void fix_mhd(MRMERWORK *mmrp)
{ /*	This cleans up the multiref header and writes it back out. */
	CRYSTAL	*lxp;

	lxp = mmrp->mrm_xhdp;
	lxp->xt_ninteg = mmrp->mrm_nobfound;
	lxp->xt_nrefls = mmrp->mrm_nreffound;
	lxp->xt_nmerged = mmrp->mrm_nmerfound;
	lxp->xt_reslim = 0.5 / mmrp->mrm_maxstl;
	lxp->xt_nruns = mmrp->mrm_nruns;
	lxp->xt_nshifts = mmrp->mrm_nshifts;
	if (EOF == fseek(mmrp->mrm_fpo, 0L, 0))
		fprintf(stderr, "Warning: header may be slightly faulty\n");
	else {
		if (sizeof (CRYSTAL) > (size_t)xgwrite((void *)lxp, XG_CRYSTAL,
	 mmrp->mrm_fpo))
		  {
			fprintf(stderr,
	 "Cannot rewrite the crystal header to the output multiref; ");
			fprintf(stderr,"errno = %d", errno);
#ifdef VMS
			if (errno == EVMSERR)
			  fprintf(stderr,"; vaxc$errno = %d = 0%o = 0x%x\n",
				vaxc$errno,vaxc$errno,vaxc$errno);
			else fprintf(stderr, "\n");
#else
			fprintf(stderr, "\n");
#endif
		  }
	     }
	(void)fclose(mmrp->mrm_fpo);	mmrp->mrm_fpo = NULL;
	mmrp->mrm_tl1 = NOW;
	sum_mrmres(mmrp->mrm_fpl, mmrp);	sum_mrmres(stdout, mmrp);
}

static int endmer(int irv, MRMERWORK *mmrp)
{ /*	Get rid of open files and allocated memory for `mrmerge' */
	if ((stdout != mmrp->mrm_fpl) && (NULL != mmrp->mrm_fpl))
		(void)fclose(mmrp->mrm_fpl);
	memzap((AVOID **)&(mmrp->mrm_fpdi), "mmrp->mrm_fpdi in mrmerge");
	memzap((AVOID **)&(mmrp->mrm_tank), "mmrp->mrm_tank in mrmerge");
	mmrp->mrm_fpl = NULL;	globalkill();	return irv;
}

int mrmmain(int argc, char *argv[])
{ /* Mainline for merging MULTIREF files together */
	register int	i, mintank;
	MERGEDREF	mrhold, *mtank0;
	MRMERWORK	mmrs;

	/* read in headers and set up memory and files */
	if (EOK != (i = merheads(argc, argv, &mmrs))) return endmer(i, &mmrs);
	mtank0 = mmrs.mrm_tank;			/* start of tanks */
	for(i = 0; i < mmrs.mrm_ninfiles; i++)	/* fill all the refl tanks */
		getmref(i, &mmrs);
	mintank	= getmin(mtank0, &mmrs); /* find smallest hkl among tanks*/
	mrhold.mr_stl = -1;		/* flag: output buffer is empty*/
/* Main reflection-reading loop.  As long at least one of the tanks has
   a real reflection in it, we continue loading the lowest among the tanks
   into the buffer, reading a new reflection into the appropriate
   tank, and  finding the smallest hkl among the tanked reflections.
   If this smallest one has the same hkl as the one that's already in the
   buffer, we add the observations from it to the buffer and continue.
   If not, we output the buffer reflection to the output diskfile and clear
   the output buffer. */
	while ((mmrs.mrm_tank + mintank)->mr_nob)
	     {
		loadmr(&mrhold, mmrs.mrm_tank + mintank, mintank, &mmrs);
		mintank = getmin(mtank0, &mmrs);
		if (hklcmp(mrhold.mr_hkl, (mmrs.mrm_tank + mintank)->mr_hkl))
			putmref(mmrs.mrm_fpo, &mrhold, &mmrs);
	     }
	/* Now all the tanks have flags set.  Output what's in the holding
		buffer if it's a legitimate reflection itself. */
	if ((mrhold.mr_nob) && (mrhold.mr_stl > -1))
		putmref(mmrs.mrm_fpo, &mrhold, &mmrs);
	fix_mhd(&mmrs);  /* Rewind the output file and correct the header */
	return endmer(EOK, &mmrs);
}

/* end of mr_merge.c */
