/* mrefcor.c -- set up and apply spatial corrections for MULTIREF files

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	24 Jun 2006: include xgpostint.h; relevant structures are there
	20 Nov 2005: inline argument declarations & strict prototyping
	18 Oct 2005: qecor: QCWORK struct elements replace static globals
	10 Oct 2005: qecor: housekeeping
	10 Oct 2005: analmr: AMWORK struct elements replace static globals
	10 Oct 2005: apply_cor: eliminate static globals
	11 May 2005: check return status from xgjournal
	27 Aug 2004: routines used only here are declared as static;cosmetics
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	28 Aug 2004: routines used only here are declared as static;cosmetics;
		routine region() renamed amregion() to eliminate confusion
	25 Jan 2004: convert a few static globals into local variables
	21 Jan 2004: license message inserted and initial comments reordered
	 7 Jan 2004: static globals, function args replace external globals;
		calls to gspline, get_shifts updated
	19 Sep 2003: qecor: explicit initialization of ymn, ymx, ymaxu, xmaxu
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Sep 2003: qecor: replace access() call with xgaccess();
		this enables gunzipping of the dark-current image on the fly
	 5 Jan 2003: reorganized source-file header
	15 Apr 2002: (?) in mr_geom(),
		modify calls related to shlp[] since it's handled differently
	25 Mar 2002: removed MSI attachments
	31 May 1995: "register" -> "register int" throughout;
		replace several longs with ints
	 2 Mar 1995: diagnostics to stdout; shared memory alloc and free
	 2 Mar 1995: qecor: frame # arg added to cm2pix and pix2cm;
		facility for handling > 2^31 counts per image
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	18 Oct 1994: apply_cor: relerrcorr--variable variance-inflation factor
	18 Oct 1994: qecor: correction characterized as counts/microsteradian;
		area per pixel calculated more carefully;
		extrapolations handled more sophisticatedly.
	11 Oct 1994: apply_cor: add complement option to remove annuli of data
	11 Oct 1994: analmr: bugfix for grouped (X,Y) positions;
		require 0.4 < I/<I> < 2.5;
		bugfix for spots near edge and for small # of spots per region
	 5 Oct 1994: use xgread & xgwrite for MULTIREF and VSPLINE structures
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: a few I/O changes and global variables moved around
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version of qecor, apply_cor, analmr
	25 Aug 1993: qecor: add a bunch of additional flags.
	29 Apr 1993: qecor: incorporate new spline conventions.
	20 Mar 1989: initial code for analmr, based loosely on apply_cor.
	23 Feb 1989: qecor: add -p flag.
	15 Aug 1988: original code for qecor

   Purpose: create and apply spatial corrections for MULTIREFs
	qecor		compute a detector-position map of instrument
	 sensitivity based on a flood-field image
	analmr		maps (I/<I>) as a function of detector position;
	 optionally, it can use that map to estimate detector's q.e. response.
	apply_cor	apply correction from analmr or qecor to MULTIREF
   Cross references: ../utilities/redurefl.c, ../xglib/spline.c

****************************************************************************/
#include	"xgpostint.h"
#define		DEF_MINSIG	(1.)
#define		MIN_GEOMOBS	2
#define		DEF_RELERR	(0.1)
#define		DEF_RADIUS	24.	/* radius of curvature of detec.*/
#define		DEF_PLATED	16.	/* default xtal-to-detec dist. */
#define		DEF_POWER	1.	/* default exponent in correction*/
#define		DEF_RELABSCOR	(0.01)	/* (Fe55 - CuKa) absorp.coeff */
#define		FRAC_NEEDED	(0.25)	/* Min frac of possible pixels needed*/
#define		MIN_PIXELS	40	/* min # pixels required in diamond */
	/* bits in the runflags variable */
#define		QE_VERBOSE	1	/* more details to logfile */
#define		QE_AREACORR	2	/* correct for area of pixels */

 typedef struct {
		char		*qc_darkname, *qc_floodname;
		int		qc_darkfrac, qc_imflags, qc_imht, qc_imwid;
		int		qc_ixc, qc_iyc, qc_nusable, qc_runflags;
		double		qc_areaperpx, qc_ctspermisr, qc_ctsperpix;
		double		qc_dcur, qc_maxzeta, qc_minzeta, qc_nsigbelow; 
		double		qc_plated, qc_pltd2, qc_power, qc_relabscor;
		double		qc_rx0, qc_ry0, qc_sigrat;
		float		*qc_areapp, *qc_tfp, **qc_zetav;
		FILE		*qc_fpl;
		MAIL		*qc_heads;
		VSPLINE		*qc_spl;
		} QCWORK; 

 typedef struct {
		int	g_nob;
		double	g_e0, g_e1, g_e2;
		} GEOMSTAT;

 typedef struct {
		unsigned	am_nscales;
		double		am_meancor, am_meanden, am_mincor, am_minsigy;
		double		am_maxcor, am_xdel, am_ydel;
		FILE		*am_fpl;
		VSPLINE		*am_spl;
		SHIFT		*am_shlp;
		VECPAR		*am_vshp;
		GEOMSTAT	*am_geoms;
	 	} AMWORK;

extern int analmrmain(int argc, char *argv[]);
extern int applycmain(int argc, char *argv[]);
extern int qecormain(int argc, char *argv[]);

/************************************
	qecor: compute sensitivity correction based on a floodfield.
	This sets up a database for constructing a quantum-efficiency
	correction for a set of area detector data.
	It does this by reading in a floodfield image (typically made
	by exposing the detector to a naked Fe-55 source for about 1 hour)
	and manipulating the count values found therein.  The algorithm
	used is described in subsequent comments.  The results of the
	calculations are written out to the USPLINE file.
		"-s" causes the floodfield to be analyzed after subtracting a
		dark-current image from each it.
	The dark-current image is found at environment variable DARKIMAGE.
**************************************/

static void summ_region(FILE *fp, QCWORK *qp)
{ /*	This summarizes the results for a single detector region */
	fprintf(fp, " %4d %4d %7d", qp->qc_ixc, qp->qc_iyc, qp->qc_nusable);
	if (qp->qc_nusable)
	  {
		fprintf(fp, " %11.3f", qp->qc_ctsperpix);
		if (qp->qc_areaperpx > DFUZZ) fprintf(fp, " %10.6f %14.4f",
			qp->qc_areaperpx, qp->qc_ctspermisr);
	  }
	fprintf(fp, "\n");
}

static double get_zeta(int iy, int ix, QCWORK *qp)
{ /*	This computes the value
	zeta(ix, iy) = # counts per microsteradian in region (ix, iy).
			  SUM Ci * [ D^2 + (X^2 + Y^2) * (1 - D/U)]
	We define this as -----------------------------------------
			           N * A * 10^6
	where Ci = count in a single legitimate pixel within the region,
	with the sum being over all okay pixels; N = # of pixels summed;
	A = area of a single pixel on the detector face;
	D = crystal-to-detector distance; X,Y = dist. in X,Y in cm from the
	region to the detector's normal to the direct beam; and
	U = radius of curvature of the detector.  Note that the value
	in brackets reduces to D^2 + X^2 + Y^2 if U=infinity and it reduces
	to D^2 if U=1.
	The tricky part of this is the computation of Axy. We do it as
	A(x,y) = 0.25 * (x(X+1,Y) - x(X-1,Y) * (y(X,Y+1) - y(X,Y-1))
	Actually the equation above is fudged with an additional term that
	looks like exp(v * (X^2 + Y^2)), where v is a differential
	absorption correction; in the most typical case this is used to
	correct for the fact that the flood-field intensity is falling
	off due to air absorption more strongly than the actual
	intensities used in data collection, since the floodfield is
	often collected with 55Fe radiation and a higher energy is used
	for real data collection. */
	int	maxu, minct, nneighb, npossible, pstd, x, xc, xdel, xmn, xmx;
	int	xmaxu, y, yc, ydel, ymn, ymx, ymaxu;
	double	a2, acorr, b2, c2, correc, ct, d2, delct, detpsq;
	double	mx2, rad0, rad1, revent, rx, ry, u2, xm1, xp1, ym1, yp1;
	float	xcm1, xcm2, xcm3, xcm4, ycm1, ycm2, ycm3, ycm4;
	VSPLINE	*lspl;

	lspl = qp->qc_spl;
	xc = lspl->v_x0px + ((double)ix) * lspl->v_xrpx;
	yc = lspl->v_y0px + ((double)iy) * lspl->v_yrpx;
	/* find boundaries in x of the diamond-shaped region we'll examine. */
	xmn = xc - lspl->v_xrpx / 2.;
	xdel = xc - xmn;	xmx = xc + xdel;	ydel = 0;
	if (xmx > qp->qc_imwid - 2) xmx = qp->qc_imwid - 2;
	pstd = (qp->qc_runflags & QE_VERBOSE) ? (!(ix % 3)) && (!(iy % 4)) : 0;
	/* Now we make a quick pass through the points in the diamond to
	find the mean value and the standard deviation.  All points lying
	more than two standard deviations away from the mean are excluded
	from the "real" calculations.*/
	qp->qc_ctsperpix = qp->qc_areaperpx = mx2 =
		qp->qc_nusable = npossible = 0;
	acorr = 1.;
	revent = (qp->qc_heads)->eventcount; if (revent < 0.) revent *= -64.;
	qp->qc_ctspermisr = revent / (double)(qp->qc_imwid * qp->qc_imht);
	minct = (int)(qp->qc_ctspermisr -
	 sqrt(qp->qc_ctspermisr) * qp->qc_nsigbelow + 0.49999);
	if (minct < 1) minct = 1;	ymn = ymx = ymaxu = yc;	xmaxu = xc;
	for (x = xmn; x <= xmx; x++)
	   {
		if (x > 0)
		  {
			ymn = yc - ydel;	ymx = yc + ydel;
			if (ymn < 1) ymn = 1;
			if (ymx > qp->qc_imht -2) ymx = qp->qc_imht -2;
			for (y = ymn; y <= ymx; y++)
			   {
				npossible++; ct = imvalue(qp->qc_heads, x, y);
				if (ct > minct)
				  {
					qp->qc_nusable++;
					qp->qc_ctsperpix += ct;
					mx2 += ct * ct;
				  }
			   }
		  }
		if (x < xc) ydel++;	else if (ydel > 0) ydel--;
	   }
	qp->qc_ixc = (xmn + xmx) / 2;	qp->qc_iyc = (ymn + ymx) / 2;
	/* we must look at enough pixels to make it a reliable effort */
	if (qp->qc_nusable > 0) qp->qc_ctspermisr = qp->qc_ctsperpix =
		qp->qc_ctsperpix / (double)qp->qc_nusable;
	if ((qp->qc_nusable < MIN_PIXELS) ||
	  ((double)qp->qc_nusable < (double)npossible * FRAC_NEEDED))
	  {
		if (qp->qc_runflags & QE_VERBOSE)
		  {
			summ_region(qp->qc_fpl, qp);
			if (pstd) summ_region(stdout, qp);
		  }
		return 0.;
	  }
	mx2 = qp->qc_sigrat * (mx2 / (double)qp->qc_nusable -
		qp->qc_ctspermisr * qp->qc_ctspermisr);
	qp->qc_ctsperpix = qp->qc_nusable = 0;	maxu = 0;
	for (x = xmn; x <= xmx; x++)
	   {
		if (x > 0)
		  {
			ymn = yc - ydel;	ymx = yc + ydel;
			if (ymn < 1) ymn = 1;
			if (ymx > qp->qc_imht -2) ymx = qp->qc_imht -2;
			for (y = ymn; y <= ymx; y++)
			   {
				ct = imvalue(qp->qc_heads, x, y);
				delct = ct - qp->qc_ctspermisr;	delct *= delct;
				if ((ct > qp->qc_ctspermisr) || delct < mx2)
				  {
					qp->qc_nusable++;
					ct = imvalue(qp->qc_heads, x, y);
					qp->qc_ctsperpix += ct;
					if (ct > maxu)
					  {
						maxu = ct; xmaxu = x; ymaxu = y;
					  }
				  }
			   }
		  }
		if (x < xc) ydel++;	else if (ydel > 0) ydel--;
	   }
	if (qp->qc_nusable > 0) qp->qc_ctsperpix /= (double)qp->qc_nusable;
	if ((qp->qc_nusable < MIN_PIXELS) || ((double)qp->qc_nusable <
	 (double)npossible * FRAC_NEEDED) || (maxu < qp->qc_ctspermisr))
	  {
		if (qp->qc_runflags & QE_VERBOSE)
		  {
			summ_region(qp->qc_fpl, qp);
			if (pstd) summ_region(stdout, qp);
		  }
		return 0.;
	  }
	if (qp->qc_runflags & QE_AREACORR)
	  { /* compute the area of a pixel in the immediate neighborhood
		of the brightest point.  We'll use this as the mean
		area for all the points in the diamond. I'm assuming that
		1 < xc < qp->qc_imwid - 2 and 1 < yc < qp->qc_imht - 2. */
		rx = xmaxu;	xp1 = rx + 1;	xm1 = rx - 1;
		ry = ymaxu;	yp1 = ry + 1;	ym1 = ry - 1;
		nneighb = (1 <= pix2cm(xp1, ry, &xcm1, &ycm1, 0)) +
			 (1 <= pix2cm(rx, yp1, &xcm2, &ycm2, 0)) +
			 (1 <= pix2cm(xm1, ry, &xcm3, &ycm3, 0)) +
			 (1 <= pix2cm(rx, ym1, &xcm4, &ycm4, 0));
		/* all four neighbors have to have positions we can determine.
		The area per pixel, "qp->qc_areaperpx", is 0.25 * area of the
		quadrilateral bounded by the points around the pixel.
		We use 6.25 rather than 0.0625 so the value is in mm
		squared rather than cm squared. */
		if (nneighb >= 4)
		  {
			b2 = (xcm3-xcm2)*(xcm3-xcm2)+(ycm3-ycm2)*(ycm3-ycm2);
			u2 = (xcm4-xcm2)*(xcm4-xcm2)+(ycm4-ycm2)*(ycm4-ycm2);
			c2 = (xcm4-xcm3)*(xcm4-xcm3)+(ycm4-ycm3)*(ycm4-ycm3);
			d2 = (xcm4-xcm1)*(xcm4-xcm1)+(ycm4-ycm1)*(ycm4-ycm1);
			a2 = (xcm1-xcm2)*(xcm1-xcm2)+(ycm1-ycm2)*(ycm1-ycm2);
			rad0 = 4.*(b2*u2 +b2*c2 +u2*c2) -b2*b2-u2*u2-c2*c2;
			rad1 = 4.*(a2*u2 +a2*d2 +u2*d2) -a2*a2-u2*u2-d2*d2;
			if ((rad0 < DFUZZ) || (rad1 < DFUZZ)) return 0.;
			qp->qc_areaperpx = 6.25 * (sqrt(rad0) + sqrt(rad1));
			*(qp->qc_areapp + ix) = qp->qc_areaperpx;
			xcm2 = 0.5 * (xcm2 + xcm4);
			ycm1 = 0.5 * (ycm1 + ycm3);
			detpsq = (xcm2 - qp->qc_rx0) * (xcm2 - qp->qc_rx0) +
				(ycm1 - qp->qc_ry0) * (ycm1 - qp->qc_ry0);
			/* "correc" is the obliquity component. It's in
			 square centimeters */
			correc = qp->qc_pltd2 + detpsq * (1. - qp->qc_plated /
				qp->qc_dcur);
			/* "acorr" is the obliquity correction multiplied by
			the relative air absorption, divided by the area
			per pixel. We multiply by 100 since `correc' is
			in cm^2 and `qp->qc_areaperpx' is in mm^2. */
			acorr = 100. * correc * exp(qp->qc_relabscor * detpsq)
			 / qp->qc_areaperpx;
		  }
	  }
	qp->qc_ctspermisr = qp->qc_ctsperpix * (acorr * 1.e-6);
	if (qp->qc_runflags & QE_VERBOSE)
	  {
		summ_region(qp->qc_fpl, qp);
		if (pstd) summ_region(stdout, qp);
	  }
	return qp->qc_ctspermisr;
}

static void disp_zeta(FILE *fp, QCWORK *qp)
{ /*	This displays a 19x19 array of zeta values on the stream fp. */
	int	ix, iy;
	float	**zp;
	VSPLINE	*lspl;

	lspl = qp->qc_spl;	zp = qp->qc_zetav;
	fprintf(fp, "     |");
	for (ix = 0; ix < lspl->v_ncol; ix++)
	   {
		fprintf(fp, " %5d", ix);
		if (!((ix + 1) % 19)) fprintf(fp, "\n     _");
	   }
	fprintf(fp, " | \n      ");
	for (ix = 0; ix < lspl->v_ncol; ix++)
	   {
		fprintf(fp, "------");
		if (!((ix + 1) % 19)) fprintf(fp, "\n     _");
	   }
	if (ix % 19) fprintf(fp, "\n");
	for (iy = 0; iy < lspl->v_nrow; iy++)
	   {
		fprintf(fp, "%4d |", iy);
		for (ix = 0; ix < lspl->v_ncol; ix++)
		   {
			fprintf(fp, " %5.3f", zp[iy][ix]);
			if (!((ix+1) % 19))
				fprintf(fp, "\n     _");
		   }
		if (ix % 19) fprintf(fp, "\n");
	   }
	fprintf(fp, "      ");
	fprintf(fp, " | \n      ");
	for (ix = 0; ix < lspl->v_ncol; ix++)
	   {
		fprintf(fp, "------");
		if (!((ix + 1) % 19)) fprintf(fp, "\n     _");
	   }
	if (ix % 19) fprintf(fp, "\n");
}

static int memzeta(QCWORK *qp)
{ /*	This organizes memory allocations for this program. */
	int	iy;
	float	**dfp, *fp;
	VSPLINE	*lspl;

	lspl = qp->qc_spl;
	if (NULL == (AVOID *)(qp->qc_zetav = (float **)callop((unsigned)
	 lspl->v_nrow, sizeof (float *), "qp->qc_zetav in memzeta")))
		 return EALLOC;
	if (NULL == (AVOID *)(qp->qc_tfp = (float *)callop(
	 (unsigned)(lspl->v_ncol * lspl->v_nrow),
	 sizeof (float), "qp->qc_tfp in memzeta"))) return EALLOC;
	for (dfp = qp->qc_zetav, iy = 0, fp = qp->qc_tfp; iy < lspl->v_nrow;
		iy++, dfp++, fp += lspl->v_ncol) *dfp = fp;
	if (NULL == (float *)(qp->qc_areapp = (float *)callop(lspl->v_ncol,
		sizeof (float), "qp->qc_areapp in memzeta"))) return EALLOC;
	return EOK;
}

static double best_extrap(int iy, int ix, QCWORK *qp)
{ /*	This computes an extrapolated value for zeta at the
	index position (iy, ix). */
	int		nincol, ninrow, uy, ux;
	double		e0, e1, e2, e3, e4, f0, f1, f2, v, v2, wt, zv;
	double		aex[3][3], aexi[3][3], bex[3], pex[3];
	float		**zev;
	VSPLINE		*lspl;

	e0 = e1 = e2 = e3 = e4 = f0 = f1 = f2 = 0.;
	zev = qp->qc_zetav;	lspl = qp->qc_spl;
	for (ninrow = ux = 0; ux < lspl->v_ncol; ux++)
		if (zev[iy][ux] > FUZZ) ninrow++;
	for (nincol = uy = 0; uy < lspl->v_nrow; uy++)
		if (zev[uy][ix] > FUZZ) nincol++;
	if (nincol > ninrow)
	  {
		if (nincol < 3) return 0.;
		else
		 {
			for (uy = 0; uy < lspl->v_nrow; uy++)
			 if ((zv = zev[uy][ix]) > FUZZ)
			   {
				wt = ((double)(uy -iy)) /
					(double)(lspl->v_nrow);
				wt = wt ? 1. / (wt * wt) : 4.;
				v = ((double)uy) / (double)(lspl->v_nrow);
				zv = log(zv);
				v2 = v * v;	e0 += wt;	e1 += wt * v;
				e2 += wt * v2;	e3 += wt * v2 * v;
				e4 += wt * v2 * v2;	f0 += wt * zv;
				f1 += wt * zv * v; f2 += wt * zv * v2;
			   }
		  }
	  }
	else
	  {
		if (ninrow < 3) return 0.;
		else {
			for (ux = 0; ux < lspl->v_ncol; ux++)
			  if ((zv = zev[iy][ux]) > FUZZ)
			   {
				wt = ((double)(ux -ix)) /
					(double)(lspl->v_ncol);
				wt = wt ? 1. / (wt * wt) : 4.;
				v = ((double)ux) / (double)(lspl->v_ncol);
				zv = log(zv);
				v2 = v * v;	e0 += wt;	e1 += wt * v;
				e2 += wt * v2;	e3 += wt * v2 * v;
				e4 += wt * v2 * v2;	f0 += wt * zv;
				f1 += wt * zv * v; f2 += wt * zv * v2;
			   }
		     }
	  }
 	aex[0][0] = e0;		aex[2][2] = e4;	aex[0][1] = aex[1][0] = e1;
 	aex[0][2] = aex[1][1] = aex[2][0] = e2;	aex[1][2] = aex[2][1] = e3;
	if (dinv3(aex, aexi)) return 0.;
	bex[0] = f0;	bex[1] = f1;	bex[2] = f2;
	dmm31(pex, aexi, bex);
	v = (nincol > ninrow) ?	((double)iy) / (double)(lspl->v_nrow) :
	 ((double)ix) / (double)(lspl->v_ncol); v2 = v * v;
	e0 = pex[0] + pex[1] * v + pex[2] * v2;	return exp(e0);
}

static void extrap_zeta(QCWORK *qp)
{ /*	This extrapolates the zeta values out to where we haven't got
	anything yet.  For a point (ix,iy) with no corresponding
	zeta value, it determines whether there are more points
	already defined in the column or in the row. If there are
	more points defined in the column, it does a least-squares
	calculation of z(ix) = a*ix*ix + b*ix + c and plugs in
	that value for z. If there are more in the row, it finds
	z(iy) = a*iy*iy + b*iy + c by least-squares and uses that value.
	The least-squares calculations are weighted by distance.  */
	int		dx, dy, ix, ixu, iy, iyu;
	int		ndone, todo, ipass, npass, nprev;
	VSPLINE		*lspl;
	CALPOINT	*calp;

	lspl = qp->qc_spl;
	for (ndone = iy = 0; iy < lspl->v_nrow; iy++)
		for (ix = 0; ix < lspl->v_ncol; ix++)
			if (FUZZ < qp->qc_zetav[iy][ix]) ndone++;
	todo = lspl->v_nrow * lspl->v_ncol;
	if (ndone >= todo)
	  {
		fprintf(stdout, "No extrapolations necessary\n"); return;
	  }
	for (nprev = ndone, npass = 0; npass < 9; npass++)
	   {
		fprintf(stdout, "before pass %2d: %4d of %4d points done\n",
			npass, ndone, todo);
	   ipass = (npass < 3 ? 2 - npass : 0);
	   ixu = lspl->v_ncol - ipass;	iyu = lspl->v_nrow - ipass;
	   for (iy = lspl->v_midr, dy = 0; ; )
	     {
		for (ix = lspl->v_midc, dx = 0; ; )
		   {
			if ((FUZZ > qp->qc_zetav[iy][ix]) && (FUZZ <
			 (qp->qc_zetav[iy][ix] = best_extrap(iy, ix, qp))))
			  {
				calp = lspl->v_cpt + iy * lspl->v_ncol + ix;
				calp->c_qec = qp->qc_zetav[iy][ix];
				ndone++;	if (ndone >= todo) break;
			  }
			dx = (dx < 0 ? -dx : -dx - 1);
			ix = lspl->v_midc + dx;
			if ((ix < ipass) || (ix >= ixu)) break;
		   }
		if (ndone >= todo) break;
		dy = (dy < 0 ? -dy : -dy - 1);
		iy = lspl->v_midr + dy;
		if ((iy < ipass) || (iy >= iyu)) break;
	     }
	   if ((ndone >= todo) || (ndone <= nprev)) break;
	   nprev = ndone;
	 }
	if (ndone >= todo) printf("all points found\n");
	else printf("No extrapolation obtained for %3d of %3d points\n",
			(todo - ndone), todo);
}

static void sumqecorrun(FILE *fp, QCWORK *qp)
{ /*	This summarizes what will happen in `qecor' to the stream fp. */
	static char	*fnames[] = {"flood", "uspline", "xlog", NULL } ;
	static char	ftypes[] = { 1, 3, 2, 0 };
	
	header(fp, PROCESSING, fnames, ftypes);
	if (qp->qc_imflags & DARK_CURRENT) fprintf(fp,
 "%6.4lf of dark-current values contained in %s will be subtracted out\n",
		((double)qp->qc_darkfrac) / 8000., qp->qc_darkname);
	if (fabs(qp->qc_power - 1.) > DFUZZ) fprintf(fp,
		 "Each correction value will be raised to the %7.3f power\n",
			qp->qc_power);
	if (qp->qc_runflags & QE_AREACORR)
	  {
		fprintf(fp,
		 "Q.E. estimates will be corrected for geometric effects\n");
		fprintf(fp,
		 "Crystal-to-detector distance = %7.2f cm\n", qp->qc_plated);
		fprintf(fp, "Radius of curvature of the detector = %8.2f cm\n",
			qp->qc_dcur);
		if (fabs(qp->qc_relabscor) > DFUZZ) fprintf(fp,
		 "Values will be absorption-adjusted by a factor = %7.4f\n",
			qp->qc_relabscor);
	  }
	else fprintf(fp, "Q.E. estimates will not be corrected for geometry\n");
	fprintf(fp,
	 "Pixels with count < <count> -%6.2f * sqrt(<count>) will be ignored\n",
		qp->qc_nsigbelow);
	fprintf(fp,
	 "Pixels with |ct-local<ct>| <%6.2f * sig(local<ct>) will be ignored\n",
		qp->qc_sigrat);
	if (qp->qc_runflags & QE_VERBOSE) fprintf(fp,
	 " Xcen Ycen nusable counts/pixel area/pixel counts/microsteradian\n");
}

static void area_fix(int iy, QCWORK *qp)
{ /*	This goes back through the area-per-pixel values in a given row of
	fiducial points and finds those that deviate strongly from the
	mean.  In those cases the zeta values are corrected by the ratio. */
	int	ix, nu;
	double	e0, e1, e2, e3, e4, f0, f1, f2, meanzy, wt, zx, zx2, zy;
	double	eb[3], ep[3], ea[3][3], eai[3][3];
	VSPLINE	*lspl;

	lspl = qp->qc_spl;
	for (meanzy = nu = ix = 0; ix < lspl->v_ncol; ix++)
	   if (fabs(qp->qc_zetav[iy][ix]) > FUZZ)
	     {
		zy = *(qp->qc_areapp + ix);	meanzy += zy;	nu++;
	     }
	if (!nu) return;	meanzy = (double)nu / meanzy;
	e0 = e1 = e2 = e3 = e4 = f0 = f1 = f2 = 0.;
	for (nu = ix = 0; ix < lspl->v_ncol; ix++)
	   if (fabs(qp->qc_zetav[iy][ix]) > FUZZ)
	     {
		zx = ix - lspl->v_midc;	zx2 = zx * zx;	wt = 1.;
		zy = meanzy * *(qp->qc_areapp + ix);	nu++;	e0 += wt;
		e1 += wt * zx;	e2 += wt * zx2;	e3 += wt * zx * zx2;
		e4 += wt * zx2 * zx2;
		f0 += wt * zy;	f1 += wt * zx * zy;	f2 += wt * zx2 * zy;
	     }
	if ((!nu) || (e0 < DFUZZ)) return;
	ea[0][0] = e0;	ea[2][2] = e4;	ea[1][0] = ea[0][1] = e1;
	ea[0][2] = ea[1][1] = ea[2][0] = e2;	ea[1][2] = ea[2][1] = e3;
	if (dinv3(ea, eai)) return;
	eb[0] = f0;	eb[1] = f1;	eb[2] = f2;	dmm31(ep, eai, eb);
	for (ix = 0; ix < lspl->v_ncol; ix++)
	   if (fabs(e1 = qp->qc_zetav[iy][ix]) > FUZZ)
	     {
		zx = ix - lspl->v_midc;	zx2 = zx * zx;
		zy = meanzy * *(qp->qc_areapp + ix);
		e0 = ep[0] + ep[1] * zx + ep[2] * zx2;
		if ((e0 > DFUZZ) && fabs(zy - e0) > e0 * 0.3)
		  {
			if (qp->qc_runflags & QE_VERBOSE) fprintf(stderr,
 "Discrepancy @ [%2d,%2d]: obs,calc area = %9.5f,%9.5f: zeta=%9.5f -> %9.5f\n",
			 iy, ix, zy / meanzy, e0 / meanzy, e1, e1 * zy / e0);
			*(qp->qc_areapp + ix) = e0 / meanzy;
			qp->qc_zetav[iy][ix] = e1 * zy / e0;
		  }
	     }
}

static int qcargs(int ac, char *av[], QCWORK *qp)
{ /*	Argument processor for qecor */
	int		ix, readresult;
	char		**dcp, **flags;
	double		spec_plated;
	float		lxcm, lycm;
	static int	idarkfrac = 6400;
	VSPLINE		*lspl;

	qp->qc_power = DEF_POWER;
	qp->qc_relabscor = DEF_RELABSCOR; qp->qc_dcur = DEF_RADIUS;
	qp->qc_darkfrac = qp->qc_imflags = 0;
	qp->qc_runflags = QE_AREACORR;	qp->qc_nsigbelow = 6.;
	flags = flagproc(ac, av, "ACNPRSV");
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	   switch (**dcp)
	     {
		case	'A':	qp->qc_runflags &= ~QE_AREACORR;	break;
		case	'C':	qp->qc_dcur = atof((*dcp) + 1);	break;
		case	'N':	qp->qc_nsigbelow = atof((*dcp) + 1);	break;
		case	'P':	qp->qc_power = atof((*dcp) + 1);	break;
		case	'R':	qp->qc_relabscor = atof((*dcp) + 1);	break;
		case	'S':	qp->qc_darkfrac = 8000. * atof((*dcp) + 1);
				qp->qc_imflags |= DARK_CURRENT;	break;
		case	'V':	qp->qc_runflags &= QE_VERBOSE;		break;
	     }
	if (qp->qc_imflags & DARK_CURRENT)
	  {
		qp->qc_darkname = trname("darkimage");
		if (0 != xgaccess(qp->qc_darkname, 4, 1))
		  {
			qp->qc_imflags &= ~DARK_CURRENT;
			qp->qc_darkname = NULL;
		  }
		qp->qc_darkfrac = idarkfrac;
	  }
	else	qp->qc_darkname = NULL;
	if (qp->qc_power < 0.1) qp->qc_power = 0.1;
	else if (qp->qc_power > 10.) qp->qc_power = 10.;
	qp->qc_plated = DEF_PLATED;
	for (dcp = av + 1, ix = 1; ix < ac; ix++, dcp++)
		if ((**dcp != '-') && (allnumeric(*dcp)) &&
		 (0. < (spec_plated = atof(*dcp))))
			qp->qc_plated = spec_plated;
	qp->qc_pltd2 = qp->qc_plated * qp->qc_plated;
	qp->qc_sigrat = qp->qc_nsigbelow / 3.;
	if (qp->qc_darkfrac > 16000) qp->qc_darkfrac = 16000;
	if (qp->qc_imflags & DARK_CURRENT)
		qp->qc_imflags |= (qp->qc_darkfrac << 16);
	qp->qc_floodname = trname("flood");
	if (NULL == (qp->qc_spl = gspline(5, &readresult))) return readresult;
	if (NULL == (qp->qc_heads = (MAIL *)callop(1, sizeof (MAIL),
	 "qp->qc_heads in qcargs"))) return EALLOC;
	if (FRAME_OK != (readresult = read_image(qp->qc_floodname,
	 qp->qc_imflags, qp->qc_heads)))
		(void)image_error(stderr, qp->qc_floodname, readresult, 1); 
	qp->qc_imwid = (qp->qc_heads)->xdimen;
	qp->qc_imht  = (qp->qc_heads)->ydimen;
	lspl = qp->qc_spl;
	if (verror(pix2cm(lspl->v_axc, lspl->v_ayc, &lxcm, &lycm, 0),
	 "find main-beam position: nonsense spline value")) return EBADVALUE;
	qp->qc_rx0 = lxcm;	qp->qc_ry0 = lycm;
	return EOK;
}

int qecormain(int argc, char *argv[])
{ /* Mainline for the quantum-efficiency calculations */
	int		ix, iy, nnonz;
	double		m2zeta, mzeta, sumzeta, zeta;
	VSPLINE		*lspl;
	CALPOINT	*calp;
	static QCWORK	qs;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal effort */
	if (EOK != (ix = qcargs(argc, argv, &qs))) return ix;
	/* reserve memory for various arrays used here */
	if (EOK != memzeta(&qs)) return EALLOC;
	sumqecorrun(stdout, &qs);
	if (NULL == (qs.qc_fpl = propen("xlog"))) qs.qc_fpl = stdout;
	else sumqecorrun(qs.qc_fpl, &qs);
	lspl = qs.qc_spl;
	for (iy = nnonz = sumzeta = 0; iy < lspl->v_nrow; iy++)
	   { /* within this row, find the zeta values at each point */
		for (ix = 0; ix < lspl->v_ncol; ix++)
		   {
			qs.qc_zetav[iy][ix] = zeta = get_zeta(iy, ix, &qs);
			if (zeta > DFUZZ)
			  {
				nnonz++; sumzeta += zeta;
			  }
		   }
		/* now go back and review the area corrections. If they
		 deviate considerably from a lsq parabola through them,
		 replace 'em with the parabola value */
		if (qs.qc_runflags & QE_AREACORR) area_fix(iy, &qs);
	   }
	if (!nnonz)
	  {
		fprintf(stderr, "Sorry! no nonzero zeta values found\n");
		if (stdout != qs.qc_fpl)
		  {
			fprintf(qs.qc_fpl,
			 "Sorry! no nonzero zeta values found\n");
			(void)fclose(qs.qc_fpl);
		  }
		qs.qc_fpl = NULL;	return EBADVALUE;
	  }
	sumzeta /= (double)nnonz;
	fprintf(qs.qc_fpl,
	 "%6d usable points; mean zeta value before renormalization = %9.4f\n",
		nnonz, sumzeta);
	fprintf(stdout,
	 "%6d usable points; mean zeta value before renormalization = %9.4f\n",
		nnonz, sumzeta);
	/* at this stage we raise each value to the user-selected exponent
	value if he's asked for it.  This value must be between 0.1 and 10.*/
	for (mzeta = m2zeta = iy = 0; iy < lspl->v_nrow; iy++)
	   {
		for (ix = 0; ix < lspl->v_ncol; ix++)
		   {
			qs.qc_zetav[iy][ix] /= sumzeta;
			if (fabs(qs.qc_power - 1.) > DFUZZ)
				qs.qc_zetav[iy][ix] =
				 pow(qs.qc_zetav[iy][ix], qs.qc_power);
			zeta = qs.qc_zetav[iy][ix];
			mzeta += zeta;	m2zeta += zeta * zeta;
		   }
	   }
	if (qs.qc_runflags & QE_VERBOSE) fprintf(qs.qc_fpl, "\f");
	fprintf(qs.qc_fpl,
	 "flux values corrected for obliquity and absorption:\n");
	disp_zeta(qs.qc_fpl, &qs);
	mzeta /= (double)nnonz;
	m2zeta = m2zeta / ((double)nnonz) - mzeta * mzeta;
	m2zeta = (m2zeta > DFUZZ) ? sqrt(m2zeta) : 0.1;
	if (m2zeta > 0.33) m2zeta = 0.33;
	qs.qc_minzeta = 2.;	qs.qc_maxzeta = -1.;
	for (iy = 0, calp = lspl->v_cpt; iy < lspl->v_nrow; iy++)
		for (ix = 0; ix < lspl->v_ncol; ix++, calp++)
		 {
		   if ((zeta = qs.qc_zetav[iy][ix]) > 1. - 3. * m2zeta)
		     {
			calp->c_qec = qs.qc_zetav[iy][ix] = zeta;
			if (zeta < qs.qc_minzeta) qs.qc_minzeta = zeta;
			if (zeta > qs.qc_maxzeta) qs.qc_maxzeta = zeta;
		     }
		   else qs.qc_zetav[iy][ix] = calp->c_qec = 0.;
		 }
	fprintf(qs.qc_fpl,
	 "Min, max, std dev of normalized zeta values = %8.4f, %8.4f, %8.4f\n",
		qs.qc_minzeta, qs.qc_maxzeta, m2zeta);
	fprintf(stdout,
	 "Min, max, std dev of normalized zeta values = %8.4f, %8.4f, %8.4f\n",
		qs.qc_minzeta, qs.qc_maxzeta, m2zeta);
	/* now extrapolate out from the ones we've gotten values for */
	extrap_zeta(&qs);
	fprintf(qs.qc_fpl, "flux values actually saved:\n");
	disp_zeta(qs.qc_fpl, &qs);
	if (qs.qc_runflags & QE_VERBOSE)
	  {
		fprintf(stdout, "flux values actually saved:\n");
		disp_zeta(stdout, &qs);
	  }
	memzap((AVOID **)&(qs.qc_areapp), "qs.qc_areapp in qecor");
	memzap((AVOID **)&(qs.qc_zetav), "qs.qc_zetav in qecor");
	memzap((AVOID **)&(qs.qc_tfp), "qs.qc_tfp in qecor");
	if (stdout != qs.qc_fpl) (void)fclose(qs.qc_fpl);
	qs.qc_spl = gspline(9, &ix);	(void)gspline(0, &iy);
	qs.qc_fpl = NULL;	globalkill();	return ix;
}

/* end of qecor functionality */

/*************************************
	analmr: Print a detector-position distribution chart for I/<I>.
	That is, for all observations with I/sigma > a cutoff and
	<I>/sigma > the same cutoff, it computes I/<I> and takes the
	mean of those values within a small geometrical region on the
	detector face.  It then prints out the chart of those regions.
	If the "-s" option is specified, it writes out a modified USPLINE file
	containing the mean values to be used as sensitivity-correction values.
*************************************/

static GEOMSTAT *amregion(AMWORK *amw, ABBROB *aobp)
{ /*	This returns a pointer to the relevant region associated with
	observation aobp.  If we just got the value associated with the
	current reflection, the wrong index would be picked half the time;
	so we shift the X,Y value up by half an interval. */
	register int	ix, iy;
	double		xp, yp;
	GEOMSTAT	*geomp;

	/* first, generate the actual x and y values */
	xp = ((double)abptox(aobp)) + amw->am_xdel;
	yp = ((double)abptoy(aobp)) + amw->am_ydel;
	ix = (xp - (amw->am_spl)->v_x0px) / (amw->am_spl)->v_xrpx;
	if ((0 > ix) || (ix >= (amw->am_spl)->v_ncol)) return NULL;
	iy = (yp - (amw->am_spl)->v_y0px) / (amw->am_spl)->v_yrpx;
	if ((0 > iy) || (iy >= (amw->am_spl)->v_nrow)) return NULL;
	geomp = amw->am_geoms + (iy * (amw->am_spl)->v_ncol + ix);
	return geomp;
}

static void mr_geom(AMWORK *amw, MERGEDREF *mrp)
{ /*	This takes the reflection pointed to by mrp and determines its
	mean intensity and variance.  If <I>/sigma(<I>) is small it
	returns; otherwise the individual I/<I> values are computed.
	Each observation's I/<I> is then stuffed into the appropriate
	geometrical statistical structure. */

	register int	iob, j, nobs;
	double		iosig, iyrat, res, ressq, sk, sigy; 
	double		thisi, thiss, wt, y, yden, ys, ski[MAXOB];
	xgulong		ulsh;
	ABBROB		*aobp;
	GEOMSTAT	*geomp;
	SHIFT		*tsp;

	y = yden = 0.;	nobs = mrp->mr_nob;
	/* First compute the scaled mean intensity */
	for (aobp = &(mrp->mr_ob[0]), iob = 0; iob < nobs; iob++, aobp++)
	   {
		ulsh = LSH(aobp->abo_runno, 24) |
			LSH(aobp->abo_chamno, 16) | aobp->abo_shift;
		if (0 <= (j = find_shift(amw->am_shlp, ulsh,
		 (int)amw->am_nscales)))
		  {
			res = mrp->mr_stl;	ressq = res * res;
			tsp = amw->am_shlp + j;
			ski[iob] = sk = tsp->shg +
			 tsp->sha * res + tsp->shb * ressq;
			wt = aobp->abo_sigma;	wt = sk / (wt * wt);
			y += wt * aobp->abo_ii;	yden += wt * sk;
		  }
		else ski[iob] = -47.;
	  }
/* if it's too weak, don't use it */
	if (yden < DFUZZ || y < DFUZZ) return;
	y /= yden;	sigy = sqrt(yden);
/* If <I>/sigma<I> is too small, don't use it */
	if (y < amw->am_minsigy * sigy) return;
/* Now compute I/<I> and figure out which geometric bin to deal with.
	Accumulate the appropriate sums */
	for (aobp = &(mrp->mr_ob[0]), iob = 0; iob < nobs; iob++, aobp++)
	   if (ski[iob] > -46.)
	     {
		ys = y * ski[iob];	thiss = aobp->abo_sigma;
		thisi = aobp->abo_ii; iosig = thisi / thiss; wt = ys / thiss;
		iyrat = thisi / ys;
		if ((0.4 < iyrat) && (iyrat < 2.5) &&
		 (NULL != (geomp = amregion(amw, aobp))))
		  {
			(geomp->g_nob)++;	geomp->g_e0 += wt;
			geomp->g_e1 += iosig; geomp->g_e2 += iosig * iyrat;
		  }
	     }
}

static void stat_region(AMWORK *amw, GEOMSTAT *geomp)
{ /*	This does the statistical manipulations to get <I/<I>> and
	the variance of I/<I> */
	double	e2, e1, e0, enm;

	if (geomp->g_nob <= MIN_GEOMOBS || geomp->g_e0 <= DFUZZ)
	  {
		geomp->g_e1 = geomp->g_e2 = 1.;	return; }
	e2 = geomp->g_e2;	e1 = geomp->g_e1;	e0 = geomp->g_e0;
	enm = geomp->g_nob - 1;
	geomp->g_e1 /= geomp->g_e0;
	geomp->g_e2 = sqrt((e2 * e0 - e1 * e1) / (enm * e0 * e0));
	if (geomp->g_e1 > amw->am_maxcor) amw->am_maxcor = geomp->g_e1;
	if (geomp->g_e1 < amw->am_mincor) amw->am_mincor = geomp->g_e1;
	amw->am_meancor += geomp->g_e1 * geomp->g_e0;
	amw->am_meanden += geomp->g_e0;
	return;
} 

static void geom_stats(AMWORK *amw)
{ /*	This does the statistics on the I/<I> values in the various regions
	and for the data as a whole */

	register int	ix, iy;
	GEOMSTAT	allregion, *geomp, *geomrow;

	fprintf(amw->am_fpl, "regional statistics on I/<I>:\n  ix:     ");
	for (ix = 0; ix < (amw->am_spl)->v_ncol; ix++)
	   {
		fprintf(amw->am_fpl, " %5d", ix);
		if (!((ix + 1) % 19)) fprintf(amw->am_fpl, "\n          ");
	   }
	fprintf(amw->am_fpl, "\n          ");
	for (ix = 0; ix < (amw->am_spl)->v_ncol; ix++)
	   {
		fprintf(amw->am_fpl, "------");
		if (!((ix + 1) % 19)) fprintf(amw->am_fpl, "\n          ");
	   }
	fprintf(amw->am_fpl, "\n");
	allregion.g_nob = allregion.g_e0 = allregion.g_e1 =
		allregion.g_e2 = 0.;
	amw->am_meancor = amw->am_meanden = 0.;
	amw->am_maxcor = -1.e6;	amw->am_mincor = -(amw->am_maxcor);
	for (iy = 0, geomp = amw->am_geoms; iy < (amw->am_spl)->v_nrow; iy++)
	   {
		fprintf(amw->am_fpl, "iy=%2d;nob=", iy);
		geomrow = geomp;
		for (ix = 0; ix < (amw->am_spl)->v_ncol; ix++, geomp++)
		   {
			allregion.g_nob += geomp->g_nob;
			allregion.g_e0 += geomp->g_e0;
			allregion.g_e1 += geomp->g_e1;
			allregion.g_e2 += geomp->g_e2;
			stat_region(amw, geomp);
			fprintf(amw->am_fpl, " %5d", geomp->g_nob);
			if (!((ix + 1) % 19))
				fprintf(amw->am_fpl, "\n          ");
		   }
		fprintf(amw->am_fpl, "\n   <I/<I>>");
		for (ix = 0, geomp = geomrow;
		 ix < (amw->am_spl)->v_ncol; ix++, geomp++)
		  {
			if (geomp->g_nob > MIN_GEOMOBS && geomp->g_e0 > DFUZZ)
				fprintf(amw->am_fpl, " %5.3f", geomp->g_e1);
			else	fprintf(amw->am_fpl, "      ");
			if (!((ix + 1) % 19))
				fprintf(amw->am_fpl, "\n          ");
		  }
		fprintf(amw->am_fpl, "\nsig(I/<I>>");
		for (ix = 0, geomp = geomrow;
		 ix < (amw->am_spl)->v_ncol; ix++, geomp++)
		  {
			if (geomp->g_nob > MIN_GEOMOBS && geomp->g_e0 > DFUZZ)
				fprintf(amw->am_fpl, " %5.3f", geomp->g_e2);
			else	fprintf(amw->am_fpl, "      ");
			if (!((ix + 1) % 19))
				fprintf(amw->am_fpl, "\n          ");
		  }
		fprintf(amw->am_fpl, "\n\n");
	   }
	fprintf(amw->am_fpl, "          ");
	for (ix = 0; ix < (amw->am_spl)->v_ncol; ix++)
	   {
		fprintf(amw->am_fpl, "------");
		if (!((ix + 1) % 19)) fprintf(amw->am_fpl, "\n          ");
	   }
	fprintf(amw->am_fpl, "\n");
	if (amw->am_meanden > DFUZZ)
	  {
		amw->am_meancor /= amw->am_meanden;
		printf("Mean     correction value = %12.7f\n",
			amw->am_meancor);
		fprintf(amw->am_fpl, "Mean     correction value = %12.7f\n",
			amw->am_meancor);
	  }
	if (amw->am_mincor < 100.)
	  {
		printf("Smallest correction value = %12.7f\n",
			amw->am_mincor);
		fprintf(amw->am_fpl, "Smallest correction value = %12.7f\n",
			amw->am_mincor);
	  }
	if (amw->am_maxcor > 0.)
	  {
		printf("Largest  correction value = %12.7f\n",
			amw->am_maxcor);
		fprintf(amw->am_fpl, "Largest  correction value = %12.7f\n",
			amw->am_maxcor);
	  }
	if (allregion.g_nob)
	  {
		stat_region(amw, &allregion);
		printf(
 "ALL DATA: %6d observations; < I / <I> > = %12.7f; sigma = %12.7f\n",
			allregion.g_nob, allregion.g_e1, allregion.g_e2);
		fprintf(amw->am_fpl,
 "ALL DATA: %6d observations; < I / <I> > = %12.7f; sigma = %12.7f\n",
			allregion.g_nob, allregion.g_e1, allregion.g_e2);
	  }
}

static int make_newspline(AMWORK *amw)
{ /*	This writes out a new VSPLINE structure with the quantum efficiency
	correction slots calculated based on the values determined in
	this run. */
	int		ix, iy;
	GEOMSTAT	*geomp;
	FILE		*fpcal;
	CALPOINT	*calp;

	for (iy = 0, calp = (amw->am_spl)->v_cpt, geomp = amw->am_geoms;
	 iy < (amw->am_spl)->v_nrow;iy++)
	   for (ix = 0; ix < (amw->am_spl)->v_ncol; ix++, calp++, geomp++)
		calp->c_qec = geomp->g_e1;
	if (qerror((AVOID *)(fpcal = xgopen("uspline", "w")),
		"open output spline file")) return EOPENERR;
	if (verror(0 < xgwrite((void *)(amw->am_spl), XG_VSPLINE, fpcal),
	 "write calibration data to disk")) return EWRITERR;
	(void)fclose(fpcal);	fpcal = NULL;	return EOK;
}

static int endanalmr(AMWORK *amw, int ina)
{ /* Cleanly finishes off a run of analmr */
	if ((stdout != amw->am_fpl) && (NULL != amw->am_fpl))
		(void)fclose(amw->am_fpl);
	memzap((AVOID **)&(amw->am_geoms), "geoms in anal_mr");
	memzap((AVOID **)&(amw->am_shlp), "shlp in anal_mr");
	memzap((AVOID **)&(amw->am_vshp), "vshp in anal_mr");
	amw->am_fpl = NULL;	globalkill();	return ina;
}

int analmrmain(int argc, char *argv[])
{ /*	This is the mainline for `anal_mr.'  In it we set up the I/O, read in
	the scale-input file and the multiref, and compute I/<I> for all
	the relevant observations.  Stats are then performed on these values
	and then they're output.*/

	int		i, nsca;
	char		new_spline, **flags;
	double		spec_minsigma;
	FILE		*fpmi;
	CRYSTAL		xhd;
	DATARUN		rhd;
	MERGEDREF	mr;
	AMWORK		ams;
	static char	*fnames[] =
		{ "MULTIREF", "USPLINE", "SCALEIN", "XLOG", NULL };
	static char	ftypes[] = { 1, 1, 1, 2, 0 };

	/* open the files and set up the scaling procedures */
	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	ams.am_minsigy = DEF_MINSIG;	new_spline = 0;
	if (qerror((AVOID *)(fpmi = xgopen("multiref", "r")),
		"open input multiref")) return endanalmr(&ams, EOPENERR);
	if (verror(0 < xgread((void *)&xhd, XG_CRYSTAL, fpmi),
	 "read crystal header from multiref")) return endanalmr(&ams, EREADERR);
	if (NULL == ((ams.am_spl) = gspline(5, &i))) return endanalmr(&ams, i);
	i = xhd.xt_nshifts + 10;
	printf("number of shifts for which space is reserved = %d\n", i);
	if (NULL == (AVOID *)(ams.am_vshp =
	 (VECPAR *)callop((unsigned) xhd.xt_nruns, sizeof (VECPAR),
	 "vshp in anal_mr"))) return endanalmr(&ams, EALLOC);
	if (NULL == (ams.am_shlp = get_shifts(NULL, ams.am_vshp, &nsca)))
		return endanalmr(&ams, EREADERR); /* read scfacs */
	ams.am_nscales = nsca;
	if (NULL == (AVOID *)(ams.am_geoms = (GEOMSTAT *)callop((unsigned)
	 ((ams.am_spl)->v_ncol * (ams.am_spl)->v_nrow), sizeof (GEOMSTAT),
	 "geoms in anal_mr"))) return endanalmr(&ams, EALLOC);
	if (NULL == (ams.am_fpl = propen("xlog"))) ams.am_fpl = stdout;
	flags = flagproc(argc, argv, "S");
	if ((flags) && (*flags) && (**flags == 'S')) new_spline = 1;
	for (argc--, argv++; argc; argc--,argv++) if (allnumeric(*argv)) break;
	if ((argc) && (0. < (spec_minsigma = atof(*argv))))
		ams.am_minsigy = spec_minsigma;
	ftypes[1] = (new_spline ? 3 : 1);
	ftypes[3] = (ams.am_fpl == stdout ? 0 : 2);
	header(stdout, REDUCTION, fnames, ftypes);
	if (ams.am_fpl != stdout)
		header(ams.am_fpl,  REDUCTION, fnames, ftypes);
	printf("Minimum <I>/sigma used = %12.7f\n", ams.am_minsigy);
	fprintf(ams.am_fpl , "Minimum <I>/sigma used = %12.7f\n",
		ams.am_minsigy);
	if (new_spline)
	  {
		printf(
 "New spline file with resulting QE correction will be written to %s\n",
			trname("USPLINE"));
		fprintf(ams.am_fpl,
 "New spline file with resulting QE correction will be written to %s\n",
			trname("USPLINE"));
	  }
	ams.am_xdel = (ams.am_spl)->v_xrpx / 2.;
	ams.am_ydel = (ams.am_spl)->v_yrpx / 2.;
	for (i = xhd.xt_nruns; i--; ) /* Read in the datarun headers */
	   if (verror(0 < xgread((void *)&rhd, XG_DATARUN, fpmi),
	    "read datarun header from input multiref"))
			return endanalmr(&ams, EWRITERR);
	(void)keep_dimen(&rhd);
/* Read in the reflections and analyze them. */
	while (0 < xgread((void *)&mr, XG_MERGEDREF, fpmi)) mr_geom(&ams, &mr);
	(void)fclose(fpmi);	fpmi = NULL;
/* Compute the statistics and output them. */
	geom_stats(&ams);	i = (new_spline ? make_newspline(&ams) : EOK);
	return endanalmr(&ams, i);
}

/* end of anal_mr functionality */

/**************************************
	apply_cor:	apply a sensitivity correction from qecor or anal_mr
	This takes the quantum efficiency correction generated in "qecor"
	or "analmr" and applies it to the intensities in a multiref.
	It then writes out a new multiref with the corrected I's and sigmas.
**************************************/

static double chk_spline(VSPLINE *splp)
{ /*	This checks the spline corrections to make sure there's actually
	something there to spline with.  If there isn't, the program exits
	immediately.  The routine also computes <correction> and returns it.*/
	int		ix, iy, npos;
	double		lmc, sumsqcor, thiscor;
	CALPOINT	*calp;

	for (sumsqcor = lmc = npos = iy = 0, calp = splp->v_cpt;
	 iy < splp->v_nrow; iy++)
	   for (ix = 0; ix < splp->v_ncol; ix++, calp++)
	     {
		thiscor = calp->c_qec;
		sumsqcor += thiscor * thiscor;
		if (thiscor > DFUZZ)
		  {
			npos++;	lmc += thiscor;
		  }
	     }
	if (verror((sumsqcor > DFUZZ),
	 "perform correction: no correction tables found")) return -1.e10;
	if (verror(npos, "perform correction: no correction tables found"))
		return -2.E10;
	return lmc / (double)npos;
}

static int pix2cor(VSPLINE *splp, double xp, double yp, double *corv)
{ /*	This returns the correction value
	associated with the pixel position (xp,yp).
	Return values:
		 1	: (x,y) inside active area
		 2	: (x,y) outside active area
				but still close to detector
		-1	: (x,y) outside active area
				and a long way from the detector
		 0	: cm position couldn't be determined.	*/
	register int	ix, iy;
	double		dx, dy, p, q, rx, ry;
	CALPOINT	*ca00, *ca01, *ca10, *ca11;

	dx = xp - splp->v_x0px;		rx = dx / splp->v_xrpx;
	ix = (rx < 0. ? 0 : (rx < splp->v_ncol - 1 ? rx : splp->v_ncol - 2));
	dy = yp - splp->v_y0px;		ry = dy / splp->v_yrpx;
	iy = (ry < 0. ? 0 : (ry < splp->v_nrow - 1 ? ry : splp->v_nrow - 2));
	ca00 = splp->v_cpt + iy * splp->v_ncol + ix;
	ca01 = ca00 + 1; ca10 = ca00 + splp->v_ncol; ca11 = ca10 + 1;
	p = rx - (double)ix;
	q = ry - (double)iy;
	*corv = 1. /((1. -p) * (1.-q) * ca00->c_qec + (1.-p)* q * ca10->c_qec +
			  p  * (1.-q) * ca01->c_qec +	  p * q * ca11->c_qec);
	return position_code(xp, yp);
}

static void rescale_ob(VSPLINE *splp, ABBROB *aobp, double meanc,double relerc)
{ /*	This rescales I and sigma for an observation aobp. Unlike earlier
	versions, I'm not assuming a strict fractional variance for the
	correction value. But we do increase the variance beyond the simple
	rescaling value by a term that is proportional to the intensity and
	(correction - <correction>) squared. */
	double	corval, csigma, realx, realy, varcor;

	realx = abptox(aobp);	realy = abptoy(aobp);	/* get pixel coords */
	if (1 != pix2cor(splp, realx, realy, &corval)) return;
	varcor = relerc * (corval - meanc) * (corval - meanc);
	csigma = corval * aobp->abo_sigma;
	aobp->abo_sigma = sqrt(csigma * csigma + aobp->abo_ii * varcor);
	aobp->abo_ii *= corval;
}

/*ARGSUSED*/
int applycmain(int argc, char *argv[])
{ /* This is the mainline for applying an absorption correction */
	int		i, nobs, nref, nsep, obno;
	double		meancor, relerrcorr, spec_relerr;
	FILE		*fplog, *fpmi, *fpmo;
	CRYSTAL		xhd;
	DATARUN		rhd;
	MERGEDREF	mr;
	VSPLINE		*splp;
	static char	*fnames[] =
		{ "multiref", "uspline", "mrefout", "xlog", NULL };
	static char	ftypes[] = { 1, 1, 2, 2, 0 };

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	relerrcorr = DEF_RELERR;
	if ((argc > 1) && (argv[1][0] == '-') && (spec_relerr =
	 atof(&(argv[1][2]))) > -DFUZZ) relerrcorr = spec_relerr;
	if (NULL == (splp = gspline(5, &i))) return i; /* pix-to-cm converter*/
	/* check that we have a correction and find the mean correction val.*/
	meancor = chk_spline(splp);	/* check for existence of correction */
	if (meancor < -1.E9) return EBADVALUE;	/* error codes */
	if (qerror((AVOID *)(fpmi = xgopen("multiref", "r")),
		"open input multiref")) return EOPENERR;
	if (verror(0 < xgread((void *)&xhd, XG_CRYSTAL, fpmi),
		"read crystal header from input multiref")) return EREADERR;
	if (qerror((AVOID *)(fpmo = xgopen("mrefout", "w")),
		"open output multiref")) return EOPENERR;
	if (verror(0 < xgwrite((void *)&xhd, XG_CRYSTAL, fpmo),
		"write crystal header to output multiref")) return EWRITERR;
	for (i = xhd.xt_nruns; i--; )
	   {
		if (verror(0 < xgread((void *)&rhd, XG_DATARUN, fpmi),
		 "read datarun header from input multiref")) return EREADERR;
		(void)keep_dimen(&rhd);
		if (verror(0 < xgwrite((void *)&rhd, XG_DATARUN, fpmo),
		 "write datarun header to output multiref")) return EWRITERR;
	   }
	if (NULL == (fplog = propen("xlog"))) fplog = stdout;
	ftypes[3] = (fplog == stdout ? 2 : 0);
	header(stdout, REDUCTION, fnames, ftypes);
	if (stdout != fplog) header(fplog,  REDUCTION, fnames, ftypes);
	if (relerrcorr > DFUZZ)
	  {
		printf(
		 "Variances will be inflated by (%7.3f * (C-1) * I)^2\n",
			relerrcorr);
		if (stdout != fplog) fprintf(fplog,
		 "Variances will be inflated by (%7.3f * (C-1) * I)^2\n",
			relerrcorr);
	  }
	for (nsep = nref = nobs = 0;
	 0 < xgread((void *)&mr, XG_MERGEDREF, fpmi); nref++)
	   {
		for (obno = 0; obno < (int)mr.mr_nob; obno++) rescale_ob(splp,
		 &(mr.mr_ob[obno]), meancor, relerrcorr);
		if (0 >= xgwrite((void *)&mr, XG_MERGEDREF, fpmo))
		  {
			if (stdout != fplog) (void)fclose(fplog);
			fplog = NULL;	return i;
		  }
		nobs += mr.mr_nob;
		nsep++;	if (mr.mr_nobpri && mr.mr_nobbij) nsep++;
	   }
	(void)fclose(fpmi);	fpmi = NULL;
	(void)fclose(fpmo);	fpmo = NULL;
	printf(
 "%6d observations of %6d merged, %6d unmerged reflections were converted\n",
		nobs, nref, nsep);
	if (stdout != fplog)
	  {
		fprintf(fplog,
 "%6d observations of %6d merged, %6d unmerged reflections were converted\n",
		nobs, nref, nsep);
		(void)fclose(fplog);
	  }
	fplog = NULL;	globalkill();	return EOK;
}

/* end of apply_cor functionality */

/* end of mrefcor.c */
