/* mu_kb.c -- merge, scale, and output data from multiple MULISTs

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	 6 Dec 2006: output MUREFLS with outmu; enable 127 < |h| < 255;
	  'C' option: enable printout of |Inat-k*Ider| > sigma; tidy up output
	24 Jun 2006: include xgpostint.h; relevant structures are there
	31 Dec 2005: don't qsort with fewer than 2 elements
	20 Nov 2005: inline argument declarations and strict prototyping
	18 Oct 2005: replace static globals with MUKBWORK structure elements
	10 Oct 2005: housekeeping
	11 May 2005: check return status from xgjournal
	11 Apr 2005: name of this file changed from mukb.c to mu_kb.c
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	28 Aug 2004: routines used only here are declared as static;cosmetics
	25 Jan 2004: deadrefl is now static local in mrg_mu; cosmetic changes
	21 Jan 2004: license message inserted and initial comments reordered
	 7 Jan 2004: static globals replace extern globals;
		program-name conventions updated
	19 Sep 2003: deadrefl array initialized with double braces a la ansi C
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Sep 2003: replace access() with xgaccess() calls;
		that enables gunzipping on the fly
	 5 Jan 2003: reorganized source-file header
	 1 Nov 2002: in incr_rfacs(),
		calculational formulae slightly more efficient
	25 Mar 2002: removed MSI attachments
	31 May 1995: "register" -> "register int" throughout
	20 May 1995: return EOK from setup routine
	12 Apr 1995: replace s_char with sichar
	 2 Mar 1995: diagnostics to stdout
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	27 Oct 1994: remove unused flags "-h", "-l" and associated variables
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version
	17 Dec 1991: bug in Patterson option fixed so if F(nat) is missing,
		nothing at all is output.
	23 Feb 1989: original code.
   Purpose: to calculate k, B, and R between a native dataset and one
	or more derivatives; optionally, to output the scaled results.
   Cross references:

 This reads one or more mulists whose names are specified as arguments
 and computes a scale factor and a relative B value between the first
 one and each of the others.  If only one mulist is specified,
 a scale factor and a relative B are obtained between the F+'s and
 the F-'s; i.e. the F+ are treated as native data and F- as derivative
 data.  Otherwise, the "parent" is the first mulist given;
 the "derivatives" is the second and subsequent ones, and the k and B are
 those which make the derivative resemble the parent.
 It also computes the weighted RMS R-factor between the
 native mulists and each of the others, and the unweighted absolute
 value R-factor between them.  If the "-o" option is specified,
 a mulist with logical name or environmental variable MULIST written out.
 It contains the data from the derivative corrected by the k and B
 so that it's on the same scale as the parent.
 If the "-p" option is specified, a Patterson file is output,
 containing h, k, l, Fsquared, and a zero.  If the "-d" (for "DORGBIN")
 option is specified, an ascii DORGBIN file is created, with format
 	fprintf(fpo, " %3d%3d%3d%10.4f%10.4f%10.4f%10.4f . . .\n",
 	h, k, l, resol, fpar, sigpar, dfpar, fder1, sigder1, dfder1,
 	fder2, sigder2, dfder2, . . .);
 This can be read by Richard Fisher's "FTOB" as long as no more
 than two derivatives are being specified.  Convention for missing data
 with DORGBN output: sigma = -32000 if reflection is missing from any
 one input file.
*************************************************************************/
#define		NEED_TIME	1
#include	"xgpostint.h"
#define		DEF_SCUT	2
#define		LINELENGTH	77
#define		DEF_HICUT	(1.e6)
#define		DEF_LOWCUT	(-1.)
#define		MISSING_1	1
#define		MISSING_2	2
#define		BOTH_PRESENT	0
#define		READ_ACCESS	4
	/* bits in the mukbflags variable: */
#define		MK_RELMU	1	/* write out a relative MULIST */
#define		MK_PATTERSON	2	/* write out an ascii Patterson */
#define		MK_BIJONLY	4	/* calculations on Bijvoet diffs only */
#define		MK_FORCEKB	010	/* force k & B calculations */
#define		MK_DORGBIN	020	/* write out an ascii DORGBIN file */
#ifdef VMS
#include	"vsort.h"
#endif

 typedef struct {
		int	d_nrefl, d_nmatch, d_used;
		double	d_dmin, d_dmax, d_relk, d_relb, d_rfac[4];
		char	*d_fname;
		MUREFL	*d_ms, *d_me, *d_mu;
		} REFSET;

 typedef struct {
		double	ds_f, ds_sig, ds_delf;
		} DERIV;

 typedef struct {
		short	me_hkl[3];
		double	me_stl;
		DERIV	*me_ds;
		} MERGEDMU;

 typedef struct {
		int		mk_kbn, mk_flags, mk_nmu, mk_nrf, mk_nrfs;
		short		mk_spgrp;
		double		mk_bigsig, mk_minstl, mk_maxstl;
		double		mk_relk, mk_relb, mk_scut;
		double		mk_e[5], mk_r[8];
		FILE		*mk_fpdo, *mk_fpl, *mk_fpo, *mk_fppa;
		REFSET		*mk_rset, *mk_rend;
		MUREFL		*mk_mub;
		MERGEDMU	*mk_mmu;
		} MUKBWORK;

extern int mukbmain(int argc, char *argv[]);

static int ltomu(char *mline, MUREFL *mup, MUKBWORK *mkp)
{ /*	This reads the data from a line of a mulist and stuffs it in the
	structure pointed to by `mup'.  If `mup' isn't at the beginning
	of its list we examine its hkl to see if it is out of order or not;
	We return 1 if and only if it is in order.
	If we're just looking for Bijvoets, return 1 if the refl is acentric
	and has distinct measurements for F+ and F-; return 0 otherwise. */
	int	i, tfl, th, tk, tl;
	short	ihkl[3], ohkl[3];
	u_char	ih[4], oh[4];
	double	tf, tfm, tfms, tfp, tfps, tmp, ts, tst;

	(void)sscanf(mline, "%d %d %d %lf %lf %lf %lf %lf %lf %lf %d",
	 &th, &tk, &tl, &tst, &tf, &ts, &tfp, &tfps, &tfm, &tfms, &tfl);
	if (mkp->mk_spgrp) /* convert the indexing if requested. This code was
			created and contributed by Barry Finzel. Thanks! */
	  {
		ihkl[0] = th;	ihkl[1] = tk;	ihkl[2] = tl;
		mpack(ihkl, ih, mkp->mk_spgrp); /* convert to 4-byte format */
		/* now symmetry-transform that according to `mkp->mk_spgrp' */
		symtr(oh, mkp->mk_spgrp);
		/* fiddle the Bijvoet bit if necessary */
		if ((!iscentric(ih, mkp->mk_spgrp)) && (oh[3] % 2))
		  {
			munpack(oh, ohkl, mkp->mk_spgrp);
			for (i = 3; i--; ) ihkl[i] = -ohkl[i];
			mpack(ihkl, ih, mkp->mk_spgrp);
			for (i = 4; i--; ) oh[i] = ih[i];
			symtr(oh, mkp->mk_spgrp);
			munpack(oh, ohkl, mkp->mk_spgrp);
			if (mkp->mk_flags & MK_BIJONLY)
			  {
				tmp = tfp; tfp = tfm;	tfm = tmp;
				tmp = tfp; tfps = tfms;	tfms = tmp;
			  }
		  }
		/* . . . otherwise, just convert back to 3-integer format */
		else munpack(oh, ohkl, mkp->mk_spgrp);
		th = ohkl[0];	tk = ohkl[1];	tl = ohkl[2];
	  }
	if ((mkp->mk_flags & MK_BIJONLY) && (tfl & 07)) return 0;
	if (ts > 1000.) /* flag-value sigma: set it to a reasonable one */
		ts = 2.;
	if (tfps > 1000.) tfps = 2.;	if (tfms > 1000.) tfms = 2.;
	mup->mu_hkl[0] = th;	mup->mu_hkl[1] = tk;	mup->mu_hkl[2] = tl;
	mup->mu_stl = tst;	mup->mu_f = tf;		mup->mu_sig = ts;
	mup->mu_fplus = tfp;	mup->mu_sigplus = tfps;
	mup->mu_fminus = tfm;	mup->mu_sigminus = tfms;
	mup->mu_flag = tfl;
	if (mup == mkp->mk_mub) return 1;
	return ((murcmp(mup, (mup-1))) > 0);
}

static int get_murefls(REFSET *rder, MUKBWORK *mkp)
{ /*	This counts how many reflections are in the file `rder->d_fname'.
	It then reads the reflections in and checks to see that they are
	in the proper sorting order.  If they aren't, it resorts them.
	The sorting check is disabled if we're analyzing Bijvoets from
	one data set, since then all the data are present.  Instead,
	only the acentrics for which distinct answers are available are output.
	This routine returns 0 if the memory allocations and reads fail,
	1 otherwise. */
	char		line[254];
	static char	*shortname;
	int		approx_refls, insort, length, nrefls, sortok;
	unsigned	reserved;
	double		founddmax, founddmin, foundsmax, foundsmin;
	MUREFL		*mup;
	FILE		*fpmu;
#ifdef VMS
	static u_short	hklbuf[] = {3,
			DSC$K_DTYPE_B, SO_ASCENDING, 0, 1,
			DSC$K_DTYPE_B, SO_ASCENDING, 1, 1,
			DSC$K_DTYPE_B, SO_ASCENDING, 2, 1};
#endif

	if (NULL == (fpmu = xgopen(rder->d_fname, "r"))) return 0;
	(void)fseek(fpmu, 0L, 2);	length = ftell(fpmu);
	approx_refls = length / LINELENGTH;
	reserved = approx_refls + 200;
	if (NULL == (mkp->mk_mub = (MUREFL *)callop(reserved, sizeof (MUREFL),
	 "mkp->mk_mub in mukb"))) return 0;
	/* Now read the data in "for real" and store it appropriately.
	 We note whether or not it's sorted as we go along. */
	if (-1 == fseek(fpmu, 0L, 0)) return 0;
	foundsmin = 1.e6;	foundsmax = -foundsmin;
	for (mup = mkp->mk_mub, insort = 0; NULL != fgets(line, 202, fpmu); )
	   {
		sortok = ltomu(line, mup, mkp);
		if ((mkp->mk_minstl <= mup->mu_stl) &&
		 (mup->mu_stl <= mkp->mk_maxstl))
		  {
			if (mup->mu_stl < foundsmin) foundsmin = mup->mu_stl;
			if (mup->mu_stl > foundsmax) foundsmax = mup->mu_stl;
			mup++;	insort += sortok;
		  }
	   }
	founddmax = (foundsmin < DFUZZ) ? 9999. : 1. / (2. * foundsmin);
	founddmin = (foundsmax < DFUZZ) ? 9999. : 1. / (2. * foundsmax);
	(void)fclose(fpmu);	fpmu = NULL;	/* finished with it */
	rder->d_nrefl = nrefls = mup - mkp->mk_mub;
/* re-sort the mulist if needed */
	if ((insort < nrefls) && (nrefls > 1))
	  {
		printf("Now sorting reflections from %s; ", rder->d_fname);
	  	qsort((AVOID *)mkp->mk_mub, (size_t)nrefls,
			sizeof (MUREFL), (QPE)murcmp);
		printf("sort finished\n");
	  }
	rder->d_ms = mkp->mk_mub;	rder->d_me = mkp->mk_mub + nrefls;
	shortname = filepart(rder->d_fname, 0);
	printf(" %14s %12d %6.2f %6.2f %16d\n", shortname,
	 approx_refls, founddmax, founddmin, rder->d_nrefl);
	if (mkp->mk_fpl != stdout) fprintf(mkp->mk_fpl,
	 " %14s %12d %6.2f %6.2f %16d\n", shortname,
	 approx_refls, founddmax, founddmin, rder->d_nrefl);
	return 1;
}

static int incr_sums(MUREFL *mup1, MUREFL *mup2, MUKBWORK *mkp)
{ /*	This increments the least-squares sums for the k, B calculation. */
	double	dellog, f1, f2, mst2, s1, s2, stwt, wt, wtn, wtp;

	if (mup1 == mup2)	/* this happens only if we're doing Bijvoets*/
	  { /* don't include reflections that are centric or for
		 which F+ or F- wasn't observed */
		if (mup1->mu_flag & 07) return 0;
		f1 = mup1->mu_fplus;	f2 = mup1->mu_fminus;
		s1 = mup1->mu_sigplus;	s2 = mup1->mu_sigminus;
	  }
	else {
		f1 = mup1->mu_f;	f2 = mup2->mu_f;
		s1 = mup1->mu_sig;	s2 = mup2->mu_sig;
	     }
	if ((s1 > FUZZ) && (f1 > mkp->mk_scut * s1) &&
	 (s2 > FUZZ) && (f2 > mkp->mk_scut * s2))
	  {
		wtp = s1 / f1;	wtn = s2 / f2;
		wt = 1. / (wtp * wtp + wtn * wtn);
		mst2 = mup2->mu_stl;	mst2 *= -mst2;	stwt = mst2 * wt;
		dellog = (log(f1) - log(f2)) * wt;
		mkp->mk_e[0] += wt;	mkp->mk_e[1] += stwt;
		mkp->mk_e[2] += mst2 * stwt;	mkp->mk_e[3] += dellog;
		mkp->mk_e[4] += dellog * mst2;	(mkp->mk_kbn)++;
		return 1;
	  }
	return 0;
}

static int incr_rfacs(MUREFL *mup1, MUREFL *mup2, MUKBWORK *mkp)
{ /*	This increments the sums used in the R-factor calculations. */
	double	df, dfw, dfw2, f1, f2, f2p, fw, fw2, s1, s2, st, wt;

	if (mup1 == mup2)	/* this happens when we're doing Bijvoets */
	  { /* skip centrics and reflections for which F+ or F-
		 wasn't observed */
		if (mup1->mu_flag & 07) return 0;
		f1 = mup1->mu_fplus;	s1 = mup1->mu_sigplus;
		f2 = mup1->mu_fminus;	s2 = mup1->mu_sigminus;
	  }
	else {
		f1 = mup1->mu_f;	s1 = mup1->mu_sig;
		f2 = mup2->mu_f;	s2 = mup2->mu_sig;
	     }
	st = mup2->mu_stl;
	f2p = f2 * mkp->mk_relk * exp(-(mkp->mk_relb) * st * st);
	if ((s1 > FUZZ) && (s2 > FUZZ))
	  {
		wt = 1. / (s1 * s1 + s2 * s2);	fw = f1 * wt;
		fw2 = fw * f1;
		df = f1 - f2p;	dfw = df * wt;	dfw2 = dfw * df;
		mkp->mk_r[0] += fw2;	mkp->mk_r[1] += dfw2;
		mkp->mk_r[2] += fabs(f1);	mkp->mk_r[3] += fabs(df);
		if (f1 > mkp->mk_scut * s1 && f2 > mkp->mk_scut * s2)
		  {
			mkp->mk_r[4] += fw2;	mkp->mk_r[5] += dfw2;
			mkp->mk_r[6] += fabs(f1);
			mkp->mk_r[7] += fabs(df);	(mkp->mk_nrfs)++;
		  }
		(mkp->mk_nrf)++;
		return 1;
	  }
	return 0;
}

static void diffhead(MUKBWORK *mkp)
{
	   fprintf(mkp->mk_fpl,
		" Reflections with scaled differences > %8.2f * sigma\n",
		mkp->mk_bigsig);
	   fprintf(mkp->mk_fpl, "   h   k   l ParInten ParSigma");
	   fprintf(mkp->mk_fpl, " DerInten Dersigma  ScInten  Scsigma\n");
}

static void mucolumn_heads(FILE *fp, MUKBWORK *mkp)
{ /* prints a header for the output columns to the stream fp */
	fprintf(fp, "k, B, and R-factor analysis:\n");
	fprintf(fp, "     Derivative #Matched #Matched    Relative_kb:");
	fprintf(fp, " Rval,all data  Rval,F/s>%4.1f\n", mkp->mk_scut);
	fprintf(fp, "       Filename    Refls F/s>%4.1f       k       B",
		mkp->mk_scut);
	fprintf(fp, "   Rwtd Runwtd  Rwtd   Runwtd\n");
}

static int kbr_calc(REFSET *rsp, REFSET *rder, MUKBWORK *mkp)
{ /*	This does the k and B calculations for a pair of data sets, namely
	`rsp' (taken as the "native") set and `rder' (taken as the
	`derivative').  It then calculates various R-factors relevant to each
	one.  If the user has asked for relative MULISTS it puts those out. */ 
	int		hdif, nbig, nmatch, nusable;
	double		determ, fixi, fixs, pari, pars, sc;
	MUREFL		*m1end, *m2end, *mup1, *mup2;
	static char	*shortname;

	m1end = mkp->mk_rset->d_me;	m2end = rder->d_me;
	nmatch = nusable = 0;
	for (mup1 = rsp->d_ms, mup2 = rder->d_ms;
	  mup1 < m1end && mup2 < m2end; )
	   {
		hdif = murcmp(mup1, mup2);
		if (hdif < 0)	/* earlier in list 1 than list 2 */
			while ((hdif = murcmp(mup1, mup2)) < 0)
				if (++mup1 >= m1end) break;
		if (mup1 >= m1end || mup2 >= m2end) break;
		if (hdif > 0) /* we're earlier in list 2 than list 1 */
			while ((hdif = murcmp(mup1, mup2)) > 0)
				if (++mup2 >= m2end) break;
		if (mup1 >= m1end || mup2 >= m2end) break;
		if (!hdif)
		  {
			nmatch++; /* exact match: increment count */
			nusable += incr_sums(mup1, mup2, mkp);
			mup1++;	mup2++;
		  }
	   }
	rder->d_nmatch = nmatch;	rder->d_used = nusable;
/* Now calculate the actual relative k and relative B. */
	determ = mkp->mk_e[0] * mkp->mk_e[2] - mkp->mk_e[1] * mkp->mk_e[1];
	if (verror((fabs(determ) > DFUZZ),
		"compute k and B: normal matrix singular")) return EBADVALUE;
	rder->d_relk = mkp->mk_relk = exp((mkp->mk_e[2] * mkp->mk_e[3] -
	 mkp->mk_e[1] * mkp->mk_e[4]) / determ);
	rder->d_relb = mkp->mk_relb = (mkp->mk_e[0] * mkp->mk_e[4] -
	 mkp->mk_e[1] * mkp->mk_e[3]) / determ;
	if (mkp->mk_flags & MK_FORCEKB)
	  {
		mkp->mk_relk = 1.;	mkp->mk_relb = 0.;
	  }
/* Now go back through the data and compute both a weighted rms R-factor
	and an unweighted absolute-value R-factor between the adjusted
	derivative data and the native data. */
	nmatch = nusable = mkp->mk_nrf = mkp->mk_nrfs = 0;
	for (mup1 = rsp->d_ms, mup2 = rder->d_ms;
	  mup1 < m1end && mup2 < m2end; )
	   {
		hdif = murcmp(mup1, mup2);
		if (hdif < 0)	/* earlier in list 1 than list 2 */
			while ((hdif = murcmp(mup1, mup2)) < 0)
				if (++mup1 >= m1end) break;
		if (mup1 >= m1end || mup2 >= m2end) break;
		if (hdif > 0) /* we're earlier in list 2 than list 1 */
			while ((hdif = murcmp(mup1, mup2)) > 0)
				if (++mup2 >= m2end) break;
		if (mup1 >= m1end || mup2 >= m2end) break;
		if (!hdif)
		  {
			nmatch++; /* exact match: increment count */
			nusable += incr_rfacs(mup1, mup2, mkp);
			mup1++;	mup2++;
		  }
	   }
	rder->d_rfac[0] = (mkp->mk_r[0] > DFUZZ && mkp->mk_r[1] > DFUZZ) ?
		sqrt(mkp->mk_r[1] / mkp->mk_r[0]) : 1.;
	rder->d_rfac[1] = (mkp->mk_r[2] > DFUZZ && mkp->mk_r[3] > DFUZZ) ?
		mkp->mk_r[3] / mkp->mk_r[2] : 1.;
	rder->d_rfac[2] = (mkp->mk_r[4] > DFUZZ && mkp->mk_r[5] > DFUZZ) ?
		sqrt(mkp->mk_r[5] / mkp->mk_r[4]) : 1.;
	rder->d_rfac[3] = (mkp->mk_r[6] > DFUZZ && mkp->mk_r[7] > DFUZZ) ?
		mkp->mk_r[7] / mkp->mk_r[6] : 1.;
	/* if the user asked to see big differences, output them here */
	if (mkp->mk_bigsig < 5000.)
	  {
	   for (mup1 = rsp->d_ms, mup2 = rder->d_ms, nbig = 0;
	   mup1 < m1end && mup2 < m2end; )
	    {
		hdif = murcmp(mup1, mup2);
		if (hdif < 0)	/* earlier in list 1 than list 2 */
			while ((hdif = murcmp(mup1, mup2)) < 0)
				if (++mup1 >= m1end) break;
		if (mup1 >= m1end || mup2 >= m2end) break;
		if (hdif > 0) /* we're earlier in list 2 than list 1 */
			while ((hdif = murcmp(mup1, mup2)) > 0)
				if (++mup2 >= m2end) break;
		if (mup1 >= m1end || mup2 >= m2end) break;
		if (!hdif)
		  {
			sc = mkp->mk_relk *
			  exp(-mkp->mk_relb * mup2->mu_stl * mup2->mu_stl);
			fixi = mup2->mu_f * sc;	fixs = mup2->mu_sig * sc;
			pari = mup1->mu_f;	pars = mup1->mu_sig;
			if (fabs(fixi - pari) > mkp->mk_bigsig * pars)
			  {
				if (!nbig) diffhead(mkp);
				nbig++;
				fprintf(mkp->mk_fpl,
	 " %3hd %3hd %3hd %8.2f %8.2f %8.2f %8.2f %8.2f %8.2f\n",
				 mup1->mu_hkl[0], mup1->mu_hkl[1],
				 mup1->mu_hkl[2], mup1->mu_f, mup1->mu_sig,
				 mup2->mu_f, mup2->mu_sig, fixi, fixs);
			  }
			mup1++;	mup2++;
		  } /* end of "if (!hdif)" clause */
	    } /* end of loop through both lists */
	   fprintf(mkp->mk_fpl,
		" %6d reflections deviated by more than %8.2f sigma\n",
			nbig, mkp->mk_bigsig);
	  } /* end of "if (mkp->mk_bigsig < 5000" clause */
	mucolumn_heads(stdout, mkp);
	if (mkp->mk_fpl != stdout) mucolumn_heads(mkp->mk_fpl, mkp);
/* if the user asked for a relative mulist, put it out here.  Note that
 under Unix these relative mulists will be overwritten since the same
 name keeps being used.  So under Unix, don't use this option on more
 than one derivative! */
	if (mkp->mk_flags & MK_RELMU)
	  {
		if (qerror((AVOID *)(mkp->mk_fpo = xgopen("MULIST", "w")),
			"open output corrected mulist")) return EOPENERR;
		nmatch = nusable = 0;
		/* this chunk of code corrected by B.C.Finzel. Thanks Barry */
		for (mup2 = rder->d_ms; mup2 < m2end; )
		   {
			sc = mkp->mk_relk *
			  exp(-mkp->mk_relb * mup2->mu_stl * mup2->mu_stl);
			mup2->mu_f *= sc;	 mup2->mu_sig *= sc;
			mup2->mu_fplus *= sc; mup2->mu_sigplus *= sc;
			mup2->mu_fminus *= sc; mup2->mu_sigminus *= sc;
			(void)outmu(mkp->mk_fpo, mup2);
			mup2++;
		   }
		(void)fclose(mkp->mk_fpo);	mkp->mk_fpo = NULL;
	  }
/* Now print out a line summarizing this derivative. */
	shortname = filepart(rder->d_fname, 0);
	printf(" %14s %8d %8d %7.3f %7.4f %6.4f %6.4f %6.4f  %6.4f\n",
	 shortname, rder->d_nmatch, rder->d_used, rder->d_relk, rder->d_relb,
	 rder->d_rfac[0], rder->d_rfac[1], rder->d_rfac[2], rder->d_rfac[3]);
	if (mkp->mk_fpl != stdout) fprintf(mkp->mk_fpl,
	 " %14s %8d %8d %7.3f %7.4f %6.4f %6.4f %6.4f  %6.4f\n",
	 shortname, rder->d_nmatch, rder->d_used, rder->d_relk, rder->d_relb,
	 rder->d_rfac[0], rder->d_rfac[1], rder->d_rfac[2], rder->d_rfac[3]);
	return EOK;
}

static REFSET *getmumin(MUKBWORK *mkp)
{ /*	This returns a pointer to the tank containing the
	reflection with the lowest-order hkl.  */
	register int	j;
	REFSET		*rp, *rlow;
	MUREFL		*mup, mur0;

	mup = mkp->mk_rset->d_mu;
	for (j = 3; j--; ) mur0.mu_hkl[j] = mup->mu_hkl[j];
	rlow = mkp->mk_rset;
	for (rp = mkp->mk_rset + 1; rp < mkp->mk_rend; rp++)
	   {
		mup = rp->d_mu;
		if (murcmp(mup, &mur0) < 0)
		  {
			for (j = 3; j--; ) mur0.mu_hkl[j] = mup->mu_hkl[j];
			rlow = rp;
		  }
	   }
	return rlow;
}

static REFSET *mrg_mu(REFSET *minset, MUKBWORK *mkp)
{ /*	This forms a merged reflection from the individual data sets.  It
	returns a pointer to the REFSET containing the hkl that is lowest
	among the remaining reflections.  */
	int		j;
	double		scale;
	REFSET		*rder;
	DERIV		*dp;
	MUREFL		*wm, mur0;
	MERGEDMU	*lmemu;
	static MUREFL	deadrefl = { { 127, 127, 127}, 0,
			1.999, -1., 5., -1., 5., -1., -5.,
			0, {0, 0, 0}, {127, 127, 127, 0} } ;

	wm = minset->d_mu;	lmemu = mkp->mk_mmu;
	for (j = 3; j--; ) lmemu->me_hkl[j] = mur0.mu_hkl[j] = wm->mu_hkl[j];
	lmemu->me_stl = wm->mu_stl;
	for (j = 0, rder = mkp->mk_rset, dp = &(lmemu->me_ds[0]);
	  j < mkp->mk_nmu; j++, rder++, dp++)
	   {
		wm = (rder->d_mu < rder->d_me ? rder->d_mu : &(deadrefl));
		if (0 == murcmp(wm, &mur0))
		  {
			if (rder == mkp->mk_rset) scale = 1.;	
			else {
				scale = wm->mu_stl;
				scale = rder->d_relk *
				 exp(-(rder->d_relb) * scale * scale);
			     }
			dp->ds_f = wm->mu_f * scale;
			dp->ds_sig = wm->mu_sig * scale;
			dp->ds_delf = (wm->mu_fplus - wm->mu_fminus) * scale;
			(rder->d_mu)++;
		  }
		else {
			dp->ds_f = -1.;	dp->ds_sig = 5.; dp->ds_delf = 0.;
		     }
	   }
	return getmumin(mkp);
}

static int start_dorgbin(MUKBWORK *mkp)
{ /* This creates and writes a header on a DORGBN file */
	int		i;
	char		rundate[26];
	time_t		tl0;
	REFSET		*rder;
	static char	*shortname;

	if (qerror((AVOID *)(mkp->mk_fpdo = xgopen("DORGBIN", "w")),
	 "open output DORGBIN file")) return EOPENERR;
		/* # of DORGBN columns after the resolution column */
	fprintf(mkp->mk_fpdo, " %5d\n", mkp->mk_nmu * 3);
	tl0 = NOW;	strcpy(rundate, ctime(&tl0));
	fprintf(mkp->mk_fpdo, "XGEN native and derivative data assembled %s",
		rundate);
	shortname = filepart((mkp->mk_rset)->d_fname, 2);
	fprintf(mkp->mk_fpdo,
	 "Structure factors   in     native data taken from %s\n", shortname);
	fprintf(mkp->mk_fpdo,
	 "Standard deviations in     native data taken from %s\n", shortname);
	fprintf(mkp->mk_fpdo,
	 "Bijvoet differences in     native data taken from %s\n", shortname);
	for (i = 1, rder = mkp->mk_rset + 1;
	 i < mkp->mk_nmu && rder < mkp->mk_rend; i++, rder++)
	   {
		shortname = filepart(rder->d_fname, 2);
		fprintf(mkp->mk_fpdo,
		 "Structure factors   in derivative data taken from %s\n",
			shortname);
		fprintf(mkp->mk_fpdo,
		 "Standard deviations in derivative data taken from %s\n",
			shortname);
		fprintf(mkp->mk_fpdo,
		 "Bijvoet differences in derivative data taken from %s\n",
			shortname);
	   }
	return EOK;
}

static void out_patterson(MUKBWORK *mkp)
{ /*	This writes one reflection in "Patterson" format */
	int		i;
	double		delf;
	DERIV		*dp, *dp0;
	MERGEDMU	*lmemu;
	static double	zero = 0.;

	lmemu = mkp->mk_mmu;	dp0 = lmemu->me_ds;
	if (mkp->mk_flags & MK_BIJONLY)
	  {
		delf = dp0->ds_delf;	delf *= delf;
		fprintf(mkp->mk_fppa, " %3hd %3hd %3hd %8.3f %8.3f\n",
		 lmemu->me_hkl[0], lmemu->me_hkl[1], lmemu->me_hkl[2],
		 delf, zero);
	  }
	else {
		fprintf(mkp->mk_fppa, " %3hd %3hd %3hd",
			lmemu->me_hkl[0], lmemu->me_hkl[1], lmemu->me_hkl[2]);
		if (dp0->ds_f >= 0.)
		  {
			for (dp = &(lmemu->me_ds[1]), i = 1;
			 i < mkp->mk_nmu; i++, dp++)
			   {
				if (dp->ds_f >= 0.)
				  {
					delf = dp->ds_f - dp0->ds_f;
					delf *= delf;
					fprintf(mkp->mk_fppa, " %8.3f %8.3f",
						delf, zero);
				  }
				else {
					delf = -1.;
					fprintf(mkp->mk_fppa, " %8.3f %8.3f",
						delf, delf);
				     }
			   } /* end of loop through derivatives */
		  } /* end of "if (dp0->ds_f >= 0.)" clause */
		fprintf(mkp->mk_fppa, "\n");
	     } /* end of else of Bijvoet-only if statement */
}

static void out_dorg(MUKBWORK *mkp)
{ /*	This outputs a merged reflection in DORGBN format. */
	int		i;
	DERIV		*dp;
	MERGEDMU	*lmemu;

	lmemu = mkp->mk_mmu;
	fprintf(mkp->mk_fpdo, " %4hd%4hd%4hd %9.4f",
		lmemu->me_hkl[0], lmemu->me_hkl[1], lmemu->me_hkl[2],
		1. / (2. * lmemu->me_stl));
	for (dp = lmemu->me_ds, i = 0; i < mkp->mk_nmu; i++, dp++)
		fprintf(mkp->mk_fpdo, " %9.4f %9.4f %9.4f",
			dp->ds_f, dp->ds_sig, dp->ds_delf);
	fprintf(mkp->mk_fpdo, "\n");
}

static int endmukb(int irv, MUKBWORK *mkp)
{ /*	Deallocates memory at the end of `mukb' */
	REFSET	*rder;

	if (NULL != mkp->mk_rset)
	  {
		for (rder = mkp->mk_rset; rder < mkp->mk_rend; rder++)
		   {
			memzap((AVOID **)&(rder->d_ms),
			 "rder->d_ms in mukb");
			memzap((AVOID **)&(rder->d_fname),
			 "rder->d_fname in mukb");
		   }
		memzap((AVOID **)&mkp->mk_rset, "mkp->mk_rset in mukb");
	  }
	if (NULL != mkp->mk_mmu)
	  {
		memzap((AVOID **)&(mkp->mk_mmu->me_ds),
			"mkp->mk_mmu->me_ds in mukb");
		memzap((AVOID **)&mkp->mk_mmu, "mkp->mk_mmu in mukb");
	  }
	if ((stdout != mkp->mk_fpl) && (NULL != mkp->mk_fpl))
		(void)fclose(mkp->mk_fpl);	mkp->mk_fpl = NULL;
	globalkill();	return irv;
}

static void head_sum(FILE *fp)
{
	fprintf(fp, "Summary of input data:\n");
	fprintf(fp,
 "       Filename  Total Refls Dmax,A Dmin,A  #Refls in Range\n");
}

static int mukbargs(int ac, char *av[], MUKBWORK *mkp)
{ /* argument processor for mukb */
	char		*cpname, **dcp, **flags, **muname, *tmu;
	short		iflag;
	int		imu;
	double		dmax, dmin, fflag, swapit;
	REFSET		*rder;
	static char	*flist[] = {"xlog", NULL };
	static char	ftypes[] = {2, 0};

	dmax = 9999.;	dmin = 0;
	zeromem((AVOID *)mkp, sizeof (MUKBWORK));
	mkp->mk_scut = DEF_SCUT;
	mkp->mk_spgrp = 0;	mkp->mk_bigsig = 100000.;
	flags = flagproc(ac, av, "CDIOPRS");	mkp->mk_flags = 0;
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	 {
	   fflag = atof((*dcp) + 1);	iflag = atoi((*dcp) + 1);
	   switch (**dcp)
	     {
		case	'C':	mkp->mk_bigsig = fflag;		break;
		case	'D':	mkp->mk_flags |= MK_DORGBIN;	break;
		case	'I':	mkp->mk_flags |= MK_FORCEKB;	break;
		case	'O':	mkp->mk_flags |= MK_RELMU;	break;
		case	'P':	mkp->mk_flags |= MK_PATTERSON;	break;
		case	'R':	mkp->mk_spgrp = iflag;		break;
		case	'S':	mkp->mk_scut = fflag;		break;
	     }
	 }
	for (ac--, av++ ; ac; ac--, av++)
		if ('-' != **av) break;
	if (ac && allnumeric(*av))
	  {
		dmax = atof(*av); ac--; av++;
	  }
	if (ac && allnumeric(*av))
	  {
		dmin = atof(*av); ac--; av++;
	  }
	if (NULL == (mkp->mk_fpl = propen("xlog"))) mkp->mk_fpl = stdout;
	ftypes[0] = (mkp->mk_fpl == stdout ? 0 : 2);
	header(stdout, PROCESSING, flist, ftypes);
	if (mkp->mk_fpl != stdout)
		header(mkp->mk_fpl, PROCESSING, flist, ftypes);
	cpname = xgprogname(NULL, 0);
	if (ac <= 1)
	  {
		mkp->mk_flags |= MK_BIJONLY;
		if (mkp->mk_flags & MK_RELMU)
		  {
			fprintf(stderr,
			 "Relative mulist can't be written for Bijvoet data\n");
			return EOK;
		  }
		else {
			printf("Outputs will be for Bijvoets only\n");
			if (mkp->mk_fpl != stdout) fprintf(mkp->mk_fpl,
				"Outputs will be for Bijvoets only\n");
		     }
	  }
	if (dmax < dmin) /* swap resolution limits if needed */
	  {
		swapit = dmax;	dmax = dmin;	dmin = swapit;
	  }
	mkp->mk_minstl = (dmax < DFUZZ) ? 9999. : 1. / (2. * dmax);
	mkp->mk_maxstl = (dmin < DFUZZ) ? 9999. : 1. / (2. * dmin);
	printf(
	 "Data with sin(theta)/lambda between %7.5f and %7.5f will be used\n",
		mkp->mk_minstl, mkp->mk_maxstl);
	if (mkp->mk_fpl != stdout) fprintf(mkp->mk_fpl,
	 "Data with sin(theta)/lambda between %7.5f and %7.5f will be used\n",
		mkp->mk_minstl, mkp->mk_maxstl);
	if (mkp->mk_spgrp) /* make sure spacegroup was properly identified */
	  {
		if ((1 > mkp->mk_spgrp) || (mkp->mk_spgrp > 255))
		  {
			fprintf(stderr,
 "%s: illegal spacegroup number %d; no spacegroup conversion performed\n",
				cpname, (int)(mkp->mk_spgrp));
			mkp->mk_spgrp = 0;
		  }
		else fprintf(stderr,
 "Reflection data will be forced into the asymmetric unit of spacegroup %d\n",
			(int)(mkp->mk_spgrp));
	  }
/* See how many MULISTs have been specified and verify that they all exist
	and are readable. */
	for (muname = av, mkp->mk_nmu = 0;
	 mkp->mk_nmu < ac; muname++, (mkp->mk_nmu)++)
	   {
		if (0 != xgaccess(*muname, READ_ACCESS, 1))
		  {
			fprintf(stderr, "%s: Mulist %s is not readable\n",
				cpname, *muname);
			return EOPENERR;
		  }
	   }
	muname -= mkp->mk_nmu;	/* rewind this pointer */
/* If we've run out of arguments, the user hasn't specified any mulists.
 In that case it's assumed that we're doing Bijvoet calculations on
 the current mulist. Make sure it's readable. */
	if (!ac)
	  {
		tmu = trname("mulist");
		if (0 != xgaccess(tmu, 4, 1))
		  {
			if ((NULL != tmu) && ('\0' != *tmu)) fprintf(stderr,
		 "%s: can't open mulist (%s) for bijvoet calculations\n",
				cpname, tmu);
			else fprintf(stderr,
		 "%s: can't find a mulist to use for bijvoet calculations\n",
				cpname);
			return EOPENERR;
		  }
		*muname = tmu; /* there's only one, so this will work */
		mkp->mk_nmu = 1;	
	  }
	/* reserve memory for the data sets themselves */
	if (NULL == (AVOID *)(mkp->mk_rset = (REFSET *)callop(
	 (size_t)(mkp->mk_nmu), sizeof (REFSET),
	 "mkp->mk_rset in mukbmain"))) return EALLOC;
	mkp->mk_rend = mkp->mk_rset + mkp->mk_nmu;
	for (imu = 0, rder = mkp->mk_rset; imu < mkp->mk_nmu;
	 muname++, imu++, rder++)
	   {
		rder->d_fname = (char *)mallop(strlen(*muname) + 2,
		 "rder->d_fname in mukbmain");
		strcpy(rder->d_fname, *muname);
		fprintf(stderr, "Getting reflections from %s\n", rder->d_fname);
		if (verror(get_murefls(rder, mkp),
		 "get reflections from a mulist")) return EREADERR;
	   }
	return EOK;
}

int mukbmain(int argc, char *argv[])
{ /* Mainline for comparing MULISTs */
	int		imu, result;
	REFSET		*rder;
	static MUKBWORK	mks;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	/* process flags, resolution limits, and filename arguments */
	if (EOK != (imu = mukbargs(argc, argv, &mks))) return imu;
	/* Measure how long the mulists are, reserve space for the
	 reflections in them, and then read in the data. */
	head_sum(stdout);
	if (mks.mk_fpl != stdout) head_sum(mks.mk_fpl);
	/* Every list is in the same order and ready to go.  If we only have
	one list, we just do the k & B calculation on it and an R-factor.
	Otherwise we calculate k & B values for each derivative set relative
	to the native (assumed to be the first one) and the appropriate
	R-factors.  Also, if the user has asked for relative MULISTS,
	we output them here.  */
	rder = ((mks.mk_flags & MK_BIJONLY) ?
		mks.mk_rset : mks.mk_rset + 1);
	for ( ; rder < mks.mk_rend; rder++)
	   if (EOK != (result = kbr_calc(mks.mk_rset, rder, &mks)))
		return endmukb(result, &mks);
	/* If we're not writing out Pattersons or DORGBN files, we quit. */
	if (!(mks.mk_flags & (MK_PATTERSON | MK_DORGBIN)))
		return endmukb(EOK, &mks);
	/* Set up memory for handling merged reflections. */
	if (NULL == (AVOID *)(mks.mk_mmu = (MERGEDMU *)callop(1,
	 sizeof (MERGEDMU), "mmu in mukbmain"))) return endmukb(EALLOC, &mks);
	if (NULL == (AVOID *)((mks.mk_mmu)->me_ds = (DERIV *)callop(
	 (size_t)(mks.mk_nmu), sizeof (DERIV), "mmu->me_ds in mukbmain")))
		return endmukb(EALLOC, &mks);
/* Within each array of mulists we initialize a ptr to individual mulists */
	for (rder = mks.mk_rset; rder < mks.mk_rend; rder++)
		rder->d_mu = rder->d_ms;
/* Figure out which mulist has the "earliest" first reflection */
	rder = getmumin(&mks);
/* Open the output file(s). */
	if ((mks.mk_flags & MK_PATTERSON) &&
	 qerror((AVOID *)(mks.mk_fppa = xgopen("PATTERSON", "w")),
	 "open output patterson file")) return endmukb(EOPENERR, &mks);
	if ((mks.mk_flags & MK_DORGBIN) &&
	  (EOK != (result = start_dorgbin(&mks)))) return endmukb(result, &mks);
/* Now go through the lists in concert with one another.  For every reflection
 we write out the merged refl as a Patterson or DORGBN file (or both)
 and then increment the pointer within the individual murefl structure */
	while (rder->d_mu < rder->d_me)
	     {
		rder = mrg_mu(rder, &mks);
		if (mks.mk_flags & MK_PATTERSON) out_patterson(&mks);
		if (mks.mk_flags & MK_DORGBIN) out_dorg(&mks);
	     }
/* We've now exhausted all the lists. */
	if (mks.mk_flags & MK_PATTERSON)
	  {
		(void)fclose(mks.mk_fppa);	mks.mk_fppa = NULL;
	  }
	if (mks.mk_flags & MK_DORGBIN)
	  {
		(void)fclose(mks.mk_fpdo);	mks.mk_fpdo = NULL;
	  }
	return endmukb(EOK, &mks);
}

/* end of mu_kb.c */
