/* rejcalc.c -- calculational and I/O engines for outlier rejection

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	20 Oct 2006: move autoreject(); used to be scalemm.c/scmreject() ;
		rejswitch now global; rejstart now static
	24 Jun 2006: include xgpostint.h rather than rejdefs.h
	 5 Jun 2006: end_stats moved earlier in the file so it can be static
	15 Dec 2005: getq moved from ../xglib/mrutil.c
	 9 Dec 2005: created from rejend.c, rejio.c, rejproc.c, rejsetup.c
	20 Nov 2005: inline argument declarations and strict prototyping
	10 Oct 2005: lphkl[3] made local; housekeeping;
		dihist calls depend on DIHIST structure; elim static globals
	28 Aug 2004: routines used only here are declared as static; cosmetics
	28 Jun 2005: exhaustreturn() and save_refl():
			typecast pointer differences upon printout
	11 May 2005: check return status from xgjournal
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	26 Jan 2005: if we run out of DELETEDOB memory, we allocate more
	 6 Sep 2004: cosmetic changes
	28 Aug 2004: routines used only here are declared as static; cosmetics
	19 Jun 2004: bugfix in input_means(): rj_onobused incremented properly;
		new flags added (-o,u,w,y);
		these invert the sense of existing flags -n,m,p,v;
		get rid of superfluous blanking-out of REJECTRUN structure
	27 May 2004: use rj_qfrac and rj_qrval structure elements in calls;
		much of the functionality of rejsetup() moved to rejstart();
		rejsetup() has rjr argument
	18 Mar 2004: save_refl has only one arg now;
		slight simplifications in log_deletion
	 1 Feb 2004: don't add 100 to the spacegroup for rhombohedrals
		if it's already been done.
	26 Jan 2004: a few static globals converted to local variables
	25 Jan 2004: a few static globals converted to local variables
	21 Jan 2004: license message inserted
	 9 Jan 2004: delheader() written for cleanup purposes
	 7 Jan 2004: REJECTRUN structure elements in function args
		replace extern globals; args updated accordingly
	15 Nov 2003: replace some globals with elements of REJECTRUN structure;
		cleanquit routine deleted because it's no longer in use
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Aug 2003: replace global uphkl[3] with static lphkl[3]
		requires a few extra calls to hunpack()
	15 Aug 2003: replace global uphkl[3] with static global lphkl[3];
		a couple of extra calls to hunpack to populate it
	 5 Jan 2003: reorganized source-file header
	15 Nov 2002: extern declarations made here rather than in rejext.h
	15 Apr 2002: (?) in shift_stats, save_refl, input_means, rejsetup:
		modify call related to shlp[] since it's handled differently
	25 Mar 2002: removed MSI attachments
	15 Jul 2001: make batch mode (-b),
		rather than curses-menu mode, the default
	19 Sep 1995: don't write out a bodyless multiref header when the
		rejection effort shows that there's nothing
		that needs to be deleted.
	31 May 1995: "register" -> "register int" throughout;
		u_longs replaced with xgulongs; tl2 is a time_t variable
	 2 Mar 1995: diagnostics to stdout; "cleanquit" routine added;
		return the correct value from "rejgref" on success;
		add "rejdie" routine; shared-memory alloc and free
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 1 Dec 1994: diagnostic output allows for detector (X,Y) > 999
	27 Oct 1994: call "use_deleted_sums()" even with only 2 observations
	 5 Oct 1994: use xgread and xgwrite for various MULTIREF structures
	30 Sep 1994: fix list of input files
	23 Sep 1994: use `rejgref' rather than `g_mref';
		change call to `copy_rest'; msg to logfile when buf overflows;
		fix bug in `get_deleted_sums';
		change returns from refl_process so that the routine handles
		RJ_HUMANALL, RJ_HUMANBAD, RJ_HUMANAMBIG ok;
		add RJ_PRESERVEOLD flag;
		announce how much storage we're allocating
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: a few bug fixes and some reorganization
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version

   Purpose: calculations, startup, and I/O in the "reject" functionality.
   Cross references: dihist.c, scalemm.c, ../curses/rejmain.c

*************************************************************************/

#define		NEED_TIME	1
#define		NEED_SIGNAL	1
#include	"xgpostint.h"
#define		MIN_IMPROVEMENT		0.9	/* GOF must improve by this */
#define		TWO_OB_RATIO	2
#define		NXYBINS		(NXBINS * NYBINS)
#define		TOTBINS		(2 * NXYBINS)

REJECTRUN *managerjp(int opno)
{ /*	This takes care of memory for a static REJECTRUN structure */
	int			nsca;
	size_t			maxdel, newnruns, newscales;
	static size_t		oldnruns, oldscales;
	static REJECTRUN	*rjr;
	static char		rjrmsg[] = "rjr in managerjp";
	static char		ndsmsg[] = "rjr->rj_ndshift in managerjp";
	static char		ncpmsg[] = "rjr->rj_ncp0 in managerjp";
	static char		vshmsg[] = "rjr->rj_vshp in managerjp";
	static char		shlmsg[] = "rjr->rj_shlp in managerjp";
	static char		rubmsg[] = "rjr->rj_rub in managerjp";
	static char		wobmsg[] = "rjr->rj_wobp in managerjp";
	static char		dremsg[] = "rjr->rj_drefs in managerjp";
	static char		mrpmsg[] = "rjr->rj_mrp in managerjp";

	if (!opno)
	  {
		(void)managerjp(07776);	memzap((AVOID **)&rjr, rjrmsg);
		return NULL;
	  }
	if (!(opno & 01))	/* selectively free memory */
	  {
		if (NULL != rjr) /* meaningless otherwise */
		  {
			if (opno & 010)
				memzap((AVOID **)&(rjr->rj_ndshift), ndsmsg);
			if (opno & 020)
				memzap((AVOID **)&(rjr->rj_ncp0), ncpmsg);
			if (opno & 040) headmem(NULL, NULL, 0);
			if (opno & 0100)
			  {
				memzap((AVOID **)&(rjr->rj_vshp), vshmsg);
				memzap((AVOID **)&(rjr->rj_rub), rubmsg);
			  }
			if (opno & 0200)
				memzap((AVOID **)&(rjr->rj_shlp), shlmsg);
			if (opno & 0400)
				memzap((AVOID **)&(rjr->rj_wobs), wobmsg);
			if (opno & 01000)
				memzap((AVOID **)&(rjr->rj_drefs), dremsg);
			if (opno & 02000)
				memzap((AVOID **)&(rjr->rj_mrp), mrpmsg);
			if (opno & 04000) /* vshp all by itself */
				memzap((AVOID **)&(rjr->rj_vshp), vshmsg);
		  }
		return rjr;
	  }
	if (opno == 3) return rjr;	/* unconditional return */
	/* any odd value > 3 we check for existence of the rjr structure */
	if ((NULL == rjr) && (NULL == (rjr = (REJECTRUN *)callop(1,
		sizeof (REJECTRUN), rjrmsg)))) return NULL;
	if (opno & 010) /* deal with list of scaling groups */
	  {
		if (0 >= rjr->rj_nscales) (void)managerjp(0200);
		if ((newscales = rjr->rj_nscales) < 1) newscales = 1;
		if ((NULL == rjr->rj_ndshift) || (newscales > oldscales))
		  {
			memzap((AVOID **)&(rjr->rj_ndshift), ndsmsg);
			if (NULL == (rjr->rj_ndshift = (short *)callop(
			 newscales, sizeof (short), ndsmsg)))
				return managerjp(0);
		  }
		oldscales = newscales;
	  }
	if (opno & 020) /* deal with detector-region counters for rejections */
	  {
		if ((NULL == rjr->rj_ncp0) && (NULL == (rjr->rj_ncp0 =
		 (u_char *)callop(TOTBINS, sizeof (u_char), ncpmsg))))
			return managerjp(0);
	  }
	if (opno & 040) /* deal with CRYSTAL and DATARUN structures */
	  {
		if ((NULL == rjr->rj_xhdp) && (EOK != headmem(
		 &(rjr->rj_xhdp), &(rjr->rj_rhdp), 1))) return managerjp(0);
	  }
	if (opno & 0100) /* spherical-harmonic params and RUNWORK structures */
	  {
		newnruns = (NULL != rjr->rj_xhdp) ?
		 (rjr->rj_xhdp)->xt_nruns : 0;
		if (newnruns < 1)
		  {
			fprintf(stderr, "Bad number of runs = %d\n",
			 (int)newnruns);	return managerjp(0);
		  }
		if ((NULL == rjr->rj_vshp) || (newnruns > oldnruns))
		  {
			memzap((AVOID **)&(rjr->rj_vshp), vshmsg);
			if (NULL == (AVOID *)(rjr->rj_vshp = (VECPAR *)callop(
			 newnruns, sizeof (VECPAR), vshmsg)))
				return managerjp(0);
		  }
		if ((NULL == rjr->rj_rub) || (newnruns > oldnruns))
		  {
			memzap((AVOID **)&(rjr->rj_rub), rubmsg);
			if (NULL == (AVOID *)(rjr->rj_rub = (RUNWORK *)callop(
			 newnruns, sizeof (RUNWORK), rubmsg)))
				return managerjp(0);
		  }
		oldnruns = newnruns;
	  }
	if (opno & 0200) /* deal with scaling groups */
	  {
		if (NULL == rjr->rj_vshp) (void)managerjp(0107);
		if (NULL == (rjr->rj_shlp = get_shifts(NULL, rjr->rj_vshp,
		 &nsca)))
		  {
			fprintf(stderr,
			 "Cannot read input scale-factor file\n");
			return managerjp(0);
		   }
		rjr->rj_nscales = nsca;
	  }
	if (opno & 0400) /* deal with wobs structures */
	  {
		if ((NULL == rjr->rj_wobs) && (NULL == (rjr->rj_wobs =
		 (WORKINGOB *)callop(MAXOB, sizeof (WORKINGOB), wobmsg))))
			return managerjp(0);
	  }
	if (opno & 01000) /* deal with drefs structures */
	  {
		if (NULL == rjr->rj_mrp) maxdel = 1000000;
		else maxdel = (rjr->rj_xhdp)->xt_ninteg;
		if ((NULL == rjr->rj_drefs) && (NULL == (rjr->rj_drefs =
		 callop(maxdel, sizeof (DELETEDOB), dremsg))))
			return managerjp(0);
	  }
	if (opno & 02000) /* deal with mrp structure */
	  {
		if ((NULL == rjr->rj_mrp) && (NULL == (rjr->rj_mrp =
		 (MERGEDREF *)callop(1, sizeof (MERGEDREF), mrpmsg))))
			return managerjp(0);
	  }
	return rjr;
}

static void xy_stats(FILE *fp, u_char *xyp, char *str)
{ /*	This displays either the (x,y) distribution of refls marked
	at I < <I> or those at I > <I>. */
	char		dl[NXBINS+2];
	u_char		*ucp, *ucpy;
	int		ix, iy, iymid, jx, jy, lastnonzero, maxx, maxy;
	u_char		z;
	REJECTRUN	*rjr;
	DATARUN		*rhp;

	if (NULL == (rjr = managerjp(03))) return;
	if ((fp != stderr) && (fp != stdout)) fprintf(fp,"\f");
	fprintf(fp,
	 "\t(X,Y) distribution of marked observations with %s\n", str);
	fprintf(fp, "   | ");
	rhp = rjr->rj_rhdp;
	maxx = rhp->run_width / rjr->rj_xbinwid;
	if (maxx > NXBINS) maxx = NXBINS;
	maxy = rhp->run_height / rjr->rj_ybinwid;
	if (maxy > NYBINS) maxy = NYBINS;
	for (ix = 1; ix < maxx / 4; ix++)
	   {
		jx = ix * 4 * rjr->rj_xbinwid; fprintf(fp,"%4d", jx);
	   }
	fprintf(fp, "\n");
	dl[NXBINS] = '\n';	dl[NXBINS+1] = '\0';
	iymid = rjr->rj_hto2 / rjr->rj_ybinwid;
	for (iy = 0, ucpy = xyp; iy < maxy; iy++, ucpy++)
	   {
		jy = iy * rjr->rj_ybinwid;
		if (!(iy % 4)) fprintf(fp, "%3d| ", jy);
		else fprintf(fp, "   | ");
		ucp = ucpy;
		for (ix = lastnonzero = 0; ix < NXBINS; ix++, ucp += maxy)
		   {
			z = *ucp;
			dl[ix] = (!z ? ' ' : (z < 10 ? ('0' + z) :
			 (z < 36 ? ('a' + (z - 10)) :
			 (z < 62 ? ('A' + (z - 36)) : '*' ))));
			if (z) lastnonzero = ix;
		   }
		if (lastnonzero < NXBINS)
		  {
			dl[lastnonzero+1] = '\n'; dl[lastnonzero+2] = '\0'; }
		if (iy == iymid)
		  {
			for (ix = lastnonzero; ix--; )
			   if (dl[ix] == ' ') dl[ix] = '-';
			for (ix = lastnonzero; ix < NXBINS; ix++) dl[ix] = '-';
		  }
		fputs(dl, fp);
	   }
}

static void shift_stats(FILE *fp)
{ /*	This prints the distribution of deleted observations by shift number */
	register int	ndisplayed, shind;
	short		*ndelp;
	SHIFT		*tsp;
	REJECTRUN	*rjr;

	if (NULL == (rjr = managerjp(03))) return;
	fprintf(fp,"Distribution of deleted observations by scaling-group\n");
	for (shind = 0; shind < 4; shind++)
		fprintf(fp, " shift chmbr run # flagged  ");
	fprintf(fp, "\n");
	for (shind = ndisplayed = 0, ndelp = rjr->rj_ndshift,
	 tsp = rjr->rj_shlp; shind < rjr->rj_nscales; shind++, ndelp++, tsp++)
	   if (*ndelp > 0) fprintf(fp, " %5d %5d %3d %9d%s",
		tsp->shval, tsp->runno, tsp->chamno,
		*ndelp,	((++ndisplayed) % 4) ? "  " : "\n");
	fprintf(fp, "\n");
}

static void end_stats(FILE *fp, double gofout)
{ /* this prints what happened to the stats for the multiref */
	REJECTRUN	*rjr;

	if (NULL == (rjr = managerjp(03))) return;
	if (rjr->rj_nobmarked) fprintf(fp,
 "%6d observations were actually flagged with shift < 0 in output multiref\n",
		(int)rjr->rj_nobmarked);
/* None of the R-factor calcs are done unless we have completed the multiref.
 We will find mr.mr_nob == 0 under that circumstance */
	if ((rjr->rj_rfdenin > FUZZ) && (rjr->rj_flags & RJ_ALLCALC))
	  {
		fprintf(fp, "Overall R-factor in  input multiref = %12.7f\n",
			rjr->rj_rfacin);
		if (rjr->rj_gof > 0) fprintf(fp,
	 "Overall goodness-of-fit value for  input multiref = %12.7f\n",
			rjr->rj_gof);
	  }
	if ((rjr->rj_rfdenout > FUZZ) && (rjr->rj_flags & RJ_ALLCALC))
	  {
		fprintf(fp, "Overall R-factor in output multiref = %12.7f\n",
			rjr->rj_rfacout);
		if (gofout > 0) fprintf(fp,
	 "Overall goodness-of-fit value for output multiref = %12.7f\n",
			gofout);
	  }
}

void last_calcs(void)
{ /*	This does the last few cleanup operations AND reports some statistics
	to the screen and to the log file.  Then comes the moment you've
	all been waiting for: the end. */
	double		gof1d, gof2d, gofout;
	REJECTRUN	*rjr;

	if (NULL == (rjr = managerjp(03))) return;
	rjr->rj_gof = gofout = 0;
	gof1d = rjr->rj_onobused - rjr->rj_nscales;
	gof2d = rjr->rj_onobout  - rjr->rj_nscales;
	if ((rjr->rj_rfdenin > FUZZ) && (rjr->rj_flags & RJ_ALLCALC))
	  {
		rjr->rj_gof = (gof1d > 0) ? rjr->rj_rfacin / gof1d : -1;
		rjr->rj_rfacin = sqrt(rjr->rj_rfacin / rjr->rj_rfdenin);
	  }
	if (rjr->rj_rfdenout > FUZZ && (rjr->rj_flags & RJ_ALLCALC))
	  {
		gofout = (gof2d > 0) ? rjr->rj_rfacout / gof2d : -1;
		rjr->rj_rfacout = sqrt(rjr->rj_rfacout / rjr->rj_rfdenout);
	  }
	if (0 != rjr->rj_nobmarked)
	  {
		shift_stats(rjr->rj_fpl);
		xy_stats(rjr->rj_fpl, rjr->rj_ncp0, "I < <I>");
		xy_stats(rjr->rj_fpl, rjr->rj_ncp0 + NXYBINS, "I > <I>");
	  }
	end_stats(stdout, gofout);
	if (NULL != rjr->rj_fpl)
	  {
		end_stats(rjr->rj_fpl, gofout);
		(void)fclose(rjr->rj_fpl);	rjr->rj_fpl = NULL;
	  }
	if (NULL != rjr->rj_fpmi) (void)fclose(rjr->rj_fpmi);
	rjr->rj_fpmi = NULL;
	if (rjr->rj_fpcheck)
	  {
		end_stats(rjr->rj_fpcheck, gofout);
		(void)fclose(rjr->rj_fpcheck);	rjr->rj_fpcheck = NULL;
	  }
	if (NULL != rjr->rj_fpmo) (void)fclose(rjr->rj_fpmo);
	rjr->rj_fpmo = NULL;	return;
}

static int copy_rest(int irv)
{ /*	This routine is called when we've exhausted the deletion list.
	At that point we need to write out the current reflection and
	copy all the remaining reflections.*/
	REJECTRUN	*rjr;

	if (NULL == (rjr = managerjp(03))) return EALLOC;
	if (0 >= xgwrite((void *)(rjr->rj_mrp), XG_MERGEDREF, rjr->rj_fpmo))
		return EWRITERR;
	while (0 != rejgref()) if (0 >= xgwrite((void *)(rjr->rj_mrp),
	  XG_MERGEDREF, rjr->rj_fpmo)) return EWRITERR;
	return irv;
}

static void log_deletion(MERGEDREF *mrp, int obnum)
{ /* this spits to the logfile a message appropriate to the attempt
	to delete observation 'obnum' from reflection 'mrp'. 'obnum'
	is -1 if the attempt was unsuccessful. */
	short		thisom, thisx, thisy, lphkl[3];
	int		posom;
	register ABBROB	*aobp;
	DELETEDOB	*drp;
	REJECTRUN	*rjr;

	if (NULL == (rjr = managerjp(03))) return;
	drp = rjr->rj_drefp;
	munpack(drp->del_hkl, lphkl, (rjr->rj_xhdp)->xt_intl);
	thisx = (short)((drp->del_omxy >> 16) & 0xff) * rjr->rj_widrat;
	thisy = ((short)(drp->del_omxy >> 24)) * rjr->rj_htrat +
		rjr->rj_hto2 * (drp->del_shift % 2);
	if (thisy < 1) thisy = rjr->rj_hto2;
	posom = drp->del_omxy & 0xffff;
	thisom = (posom > 18000 ? (short)(posom - 36000L) : posom);
	if (obnum > -1)
	  {
		aobp = &(mrp->mr_ob[obnum]);
		if (rjr->rj_fpcheck) fprintf(rjr->rj_fpcheck,
 "%4d%4d%4d\t%6d%6d %4d %4d\t %7.2f %7.3f %3d %3d ACTUALLY DELETED\n",
			lphkl[0],lphkl[1],lphkl[2], drp->del_shift,
			thisom, thisx, thisy, aobp->abo_ii, aobp->abo_sigma,
			drp->del_chamno, drp->del_runno);
	  }
	else if (rjr->rj_fpcheck) fprintf(rjr->rj_fpcheck,
 "%4d%4d%4d\t%6d%6d %4d %4d\t                 %3d %3d        NOT DELETED\n",
		lphkl[0], lphkl[1], lphkl[2], drp->del_shift, thisom,
		thisx, thisy, drp->del_chamno, drp->del_runno);
	return;
}

int apply_deletions(void)
{ /*	This takes the DELETEDOB list and uses it as a guide to the output.
	Any reflection absent from the list is written to the output unchanged.
	Any reflection present in the list has its observations checked, and the
	flagged observations are written out with the negative shift values.
	The list of deleted observations has the same order as the multiref */
	register int	diffind, i;
	register ABBROB	*aobp;
	xgulong		pomxy;
	int		posom;
	CRYSTAL		*xhp;
	DATARUN		*rhp;
	DELETEDOB	*drp;
	MERGEDREF	*lmp;
	REJECTRUN	*rjr;

	if (NULL == (rjr = managerjp(03))) return EALLOC;
	/* Sep 95: don't open the output multiref if there are no deletions
	 to apply. Thus we will be left with just an input multiref! */
	if ((NULL == rjr->rj_drefs) || (rjr->rj_drefp == rjr->rj_drefs))
	  {
		if (NULL != rjr->rj_fpl)
		  {
			fprintf(rjr->rj_fpl,
	 "No observations were marked for deletion in the input multiref;\n");
			fprintf(rjr->rj_fpl,
				"No new multiref will be output\n");
			xeflush(rjr->rj_fpl);
		  }
		printf(
	 "No observations were marked for deletion in the input multiref;\n");
		printf("No new multiref will be output\n");
		return EOK;
	  }
	if (-1 == fseek(rjr->rj_fpmi, 0L, 0))
	  {
		fprintf(stderr, "Reject: cannot rewind input multiref\n");
		return ESEEKERR;
	  }
	if (qerror((AVOID *)(rjr->rj_fpmo = xgopen("mrefout", "w")),
	 "open output multiref")) return EOPENERR;
	if (NULL == (rjr = managerjp(047))) return EALLOC;
	xhp = rjr->rj_xhdp;	rhp = rjr->rj_rhdp;
	if (verror(0 < xgread((void *)xhp, XG_CRYSTAL, rjr->rj_fpmi),
		"reread input crystal header")) return EREADERR;
	if (verror(0 < xgwrite((void *)xhp, XG_CRYSTAL, rjr->rj_fpmo),
		"write output crystal header")) return EWRITERR;
	for (i = xhp->xt_nruns; i--; )
	   {
		if (verror(0 < xgread((void *)rhp, XG_DATARUN, rjr->rj_fpmi),
			"reread input datarun header")) return EREADERR;
		if (verror(0 < xgwrite((void *)rhp, XG_DATARUN, rjr->rj_fpmo),
			"write output datarun header")) return EWRITERR;
	   }
	if (verror(rejgref(), "reread first reflection")) return EREADERR;
	rjr->rj_drefend = rjr->rj_drefp;
	drp = rjr->rj_drefp = rjr->rj_drefs;	/* rewind delet. list */
	lmp = rjr->rj_mrp;
	diffind = hklcmp(lmp->mr_hkl, drp->del_hkl); /* compare hkl indices */
	if (rjr->rj_fpcheck) fprintf(rjr->rj_fpcheck,
 "    h    k    l\t shift omega    X    Y\t  Inten.   sigma chm run\n");
	while (rjr->rj_drefp <= rjr->rj_drefend)
	     {
		while (!diffind)/* multiref hkl and list hkl are equal */
		     { /* ... look thru obs and modify them */
			lmp = rjr->rj_mrp;
			for (i = 0, aobp = lmp->mr_ob;
			 i < (int)(lmp->mr_nob); i++, aobp++)
			   { /* see if the omega, x, y values are the same */
				posom = (aobp->abo_omob < 0 ?
				 36000L + (int)aobp->abo_omob : aobp->abo_omob);
				pomxy = LSH(aobp->abo_yob, 24) |
					LSH(aobp->abo_xob, 16) |(xgulong)posom;
			/* they are: flag the ob by negating its shift # */
				if (pomxy == rjr->rj_drefp->del_omxy)
				  {
					aobp->abo_shift= -abs(aobp->abo_shift);
					(rjr->rj_nobmarked)++;
					log_deletion(lmp, i);
				  }
			   }
			/* advance the pointer in the deletion list.  If that
			runs us out the end, copy out the rest of the input */
			if ((++(rjr->rj_drefp)) == rjr->rj_drefend)
				return copy_rest(EOK);
			/* see if we have another ob of the same refl */
			lmp = rjr->rj_mrp;	drp = rjr->rj_drefp;
			diffind = hklcmp(lmp->mr_hkl, drp->del_hkl);
		     }
		while (diffind < 0) /* index in multiref is earlier */
		     { /* ... write refl out and read the next one */
			lmp = rjr->rj_mrp;
			(void)xgwrite((void *)lmp, XG_MERGEDREF, rjr->rj_fpmo);
			/* if we run out of reflections at this point,
			we note in the logfile the fact that some other
			deletions were requested and weren't acted upon */
			if (0 == rejgref())
			  {
				while (++(rjr->rj_drefp) < rjr->rj_drefend)
					log_deletion(rjr->rj_mrp, -1);
				return EOK; /* we're done, except for stats */
			  }
			/* otherwise, we just compare hkls again. */
			drp = rjr->rj_drefp;	lmp = rjr->rj_mrp;
			diffind = hklcmp(lmp->mr_hkl, drp->del_hkl);
		     }
		while (diffind > 0)
		     { /* index in list is earlier.  This happens when
			we've jumped past the hkl of the list refl. */
			log_deletion(rjr->rj_mrp, -1);
			if ((++(rjr->rj_drefp)) == rjr->rj_drefend)
				return copy_rest(EOK);
			drp = rjr->rj_drefp;	lmp = rjr->rj_mrp;
			diffind = hklcmp(lmp->mr_hkl, drp->del_hkl);
		     }
	     }
	return EOK;
}

int print_deletions(void)
{ /* this prints the deletable refls to a file whence
	the multiref can be actually modified at a later date */
	int		thisx, thisy;
	short		lphkl[3];
	CRYSTAL		*xhp;
	DELETEDOB	*drp;
	REJECTRUN	*rjr;
	static FILE	*fpdel;

	if (NULL == (rjr = managerjp(03))) return EALLOC;
	rjr->rj_drefend = rjr->rj_drefp;
	if (qerror((AVOID *)(fpdel = propen("deletions")),
		"open output deletions file")) return EOPENERR;
	xhp = rjr->rj_xhdp;
	for (drp = rjr->rj_drefs; drp < rjr->rj_drefend; drp++)
	   {
		rjr->rj_drefp = drp;
		thisx = ((((xgulong)(drp->del_omxy)) >> 16) & 0xff) *
			rjr->rj_widrat;
		thisy = ((((xgulong)(drp->del_omxy)) >> 24) * rjr->rj_htrat +
			rjr->rj_hto2 * drp->del_shift % 2);
		munpack(drp->del_hkl, lphkl, xhp->xt_intl);
		fprintf(fpdel,
		 "%4d%4d%4d\t%6d%4d%4d\t                  %6d %2d %2d\n",
		 lphkl[0], lphkl[1], lphkl[2],
		 (int)(drp->del_omxy & 0xffff), thisx, thisy,
		 (int)drp->del_shift, (int)drp->del_chamno,
		 (int)drp->del_runno);
	   }
	(void)fclose(fpdel);	fpdel = NULL;
	return EOK;
}

static char *first_display(u_char obtype)
{ /* this returns a string appropriate to the value of "obtype" */
	if (obtype & W_1ST_OF_TYPE)
	 return	(obtype & W_ISBIJVOET ?	"Bijvoets  :" :
		(obtype & W_ISCENTRIC ?	"centrics  :" : "Primaries :"));
	else
	 return	"           ";
}

char *anyob_display(u_char obtype)
{ /* this returns a string depending on the nature of `obtype' */
	return	(obtype & W_ISBIJVOET ? " Bijvoet:  ":
		(obtype & W_ISCENTRIC ? " centric:  ": " Primary:  "));
}

char *marker_display(u_char marker)
{ /* this returns a string appropriate to the value of "marker" */
	return	(marker & W_HANDMARKED ? "hand-marked":
		(marker & W_AUTOMARKED ? "auto-marked" :
		(marker & W_PREMARKED  ? " pre-marked" : "           ")));
}

int rejgref(void)
{ /* a simple wrapper for xgread in which the reflection is read in and,
 unless RJ_PRESERVEOLD is set, all the observations previously marked for
 deletion are unmarked. */
	int		i, nb;
	MERGEDREF	*mrq;
	ABBROB		*aobp;
	REJECTRUN	*rjr;

	if (NULL == (rjr = managerjp(03))) return 0;
	mrq = rjr->rj_mrp;
	if (rjr->rj_flags & RJ_PRESERVEOLD)
		return xgread((void *)mrq, XG_MERGEDREF, rjr->rj_fpmi);
	if (0 >= (nb = xgread((void *)mrq, XG_MERGEDREF, rjr->rj_fpmi)))
		return 0;
	for (i = 0, aobp = &(mrq->mr_ob[0]); i < mrq->mr_nob; i++, aobp++)
		if (aobp->abo_shift < 2) aobp->abo_shift = -aobp->abo_shift;
	return nb;
}

static void delheader(FILE *fp)
{ /* This prints a suitable header for the deletion results to stream fp */
	fprintf(fp, "   h   k   l    X    Y   omega shift      I sigma");
	fprintf(fp,
"    Dev.\tTotals before deletions\t\tTotals after all deletions\n");
	fprintf(fp, "\t\t\t\t\t\t\t\t<I> <sig> g.o.f. worst dev #obs      ");
	fprintf(fp, "<I> <sig> g.o.f. worst dev #obs\n");
	xeflush(fp);
}

int save_refl(void)
{ /*	This prints the results of an "reject" calculation on a single
	reflection.  It sticks flagged observations in the appropriate
	"DELETEDOB" structure.  It also marks the location in x,y where
	the reflection was found if it was more than two sigma from the mean.
	Return values from this routine and their meanings:
   0	we aren't deleting anything
   1	we've done at least one deletion.
   2	it's been a long time since the screen has changed: send a flag
	 telling the caller to remind the user that the program is alive.
 010	we've overloaded the memory for the list of deletable observations
 020	memory error in setting this call up */
	register int		first_shown, i;
	register WORKINGOB	*wobp;
	xgulong			ulsh;
	int			retval, shind, xbin, xybin, ybin;
	short			lphkl[3];
	time_t			tl2;
	u_char			*ucp;
	MERGEDREF		*mrq;
	DELETEDOB		*drp;
	REJECTRUN		*rjr;

	if (NULL == (rjr = managerjp(03))) return 020;
	mrq = rjr->rj_mrp;
	munpack(mrq->mr_hkl, lphkl, (rjr->rj_xhdp)->xt_intl);
	if (rjr->rj_fpcheck)
	  {
		fprintf(rjr->rj_fpcheck,
		 " %4d%4d%4d: sth/l =%6.4f,%3d obs %9.2f %8.2f%7.2f\n",
		 lphkl[0], lphkl[1], lphkl[2], mrq->mr_stl,
		 rjr->rj_nobused, rjr->rj_ibar, rjr->rj_sig, rjr->rj_gof);
		fprintf(rjr->rj_fpcheck,
		 "  After deleting marked obs: %3d obs %9.2f %8.2f%7.2f\n",
		 rjr->rj_nobaf, rjr->rj_ibaraf, rjr->rj_sigaf, rjr->rj_gofaf);
	  }
	retval = 0;
	for (wobp = rjr->rj_wobs, first_shown = 1;
	 wobp < rjr->rj_wobend; wobp++)
	   {
		if (rjr->rj_fpcheck) fprintf(rjr->rj_fpcheck,
 "%11s   %2d %5d %5d %4d %4d %9.2f %8.2f%7.2f  %11s  %7.2f %7d %5d\n",
		 first_display(wobp->w_flag), (int)(wobp - rjr->rj_wobs),
		 wobp->w_shift,
		 wobp->w_omob, wobp->w_xob, wobp->w_yob, wobp->w_isc,
		 wobp->w_sigsc, wobp->w_gofp, marker_display(wobp->w_flag),
		 wobp->w_dev, wobp->w_chamno, wobp->w_runno);
		if (wobp->w_flag & W_ISMARKED)
		  {
			retval = 1;	drp = rjr->rj_drefp;
			for (i = 4; i--; ) drp->del_hkl[i] = mrq->mr_hkl[i];
			drp->del_omxy = wobp->w_omxy;
			drp->del_shift = wobp->w_shift;
			drp->del_chamno = wobp->w_chamno;
			drp->del_runno = wobp->w_runno;
			/* put header on logfile if needed & if verbose */
			if ((drp == rjr->rj_drefs) &&
			 (rjr->rj_flags & RJ_VERBOSE)) delheader(rjr->rj_fpl);
			ulsh = LSH(wobp->w_runno,24) | LSH(wobp->w_chamno,16) |
				wobp->w_shift;
			if (0 <= (shind = find_shift(rjr->rj_shlp, ulsh,
			 rjr->rj_nscales)) && (shind < rjr->rj_nscales))
				(rjr->rj_ndshift[shind])++;
			if (rjr->rj_flags & RJ_VERBOSE)
			  {
			    if (first_shown) fprintf(rjr->rj_fpl,
			 "%4d%4d%4d", lphkl[0], lphkl[1], lphkl[2]);
			    else fprintf(rjr->rj_fpl,"\t    ");
			    fprintf(rjr->rj_fpl,
				" %4d %4d %7.2f %5d %6.0f %5.0f %7.1f",
				wobp->w_xob, wobp->w_yob,
				((double)wobp->w_omob) / 100., wobp->w_shift,
				wobp->w_isc, wobp->w_sigsc, sqrt(wobp->w_dev));
			    if (first_shown) fprintf(rjr->rj_fpl,
"    %6.0f %5.0f %6.1f %9.1f %4d   %6.0f %5.0f %6.1f %9.1f %4d\n",
				rjr->rj_ibar, rjr->rj_sig, rjr->rj_gof,
				sqrt(rjr->rj_worst), rjr->rj_nobused,
				rjr->rj_ibaraf, rjr->rj_sigaf, rjr->rj_gofaf,
				sqrt(rjr->rj_worstaf), rjr->rj_nobaf);
			    else fprintf(rjr->rj_fpl,"\n");
			  }
			first_shown = 0;
			xbin = wobp->w_xob / rjr->rj_xbinwid;
			if (xbin < 0) xbin = 0;
			else if (xbin >= NXBINS) xbin = NXBINS - 1;
			ybin = wobp->w_yob / rjr->rj_ybinwid;
			if (ybin < 0) ybin = 0;
			else if (ybin >= NYBINS) ybin = NYBINS - 1;
			(rjr->rj_drefp)++;	/* increment pointer */
			drp++;
			xybin = xbin * NYBINS + ybin;
			ucp = rjr->rj_ncp0 + xybin;
			if (wobp->w_isc >= rjr->rj_ibar) ucp += NXYBINS;
			if (*ucp < 255) (*ucp)++;
		  }
	     }
	if (rjr->rj_flags & RJ_VERBOSE) xeflush(rjr->rj_fpl);
	if (rjr->rj_fpcheck) xeflush(rjr->rj_fpcheck);
	/* accumulate the sums for R-factors and goodness-of-fit */
	if (rjr->rj_nobaf > 1 && rjr->rj_ibaraf > FUZZ)
	  {
		rjr->rj_rfacout += rjr->rj_e2 -
			rjr->rj_e1 * rjr->rj_e1 / rjr->rj_e0;
		rjr->rj_rfdenout += rjr->rj_e2;
		rjr->rj_onobout += rjr->rj_nobaf;
	  }
	/* prepare to print a message every once in a while to show the
	user we haven't gone to sleep */
	if ((!(rjr->rj_flags & RJ_BATCH)) && ((tl2 = NOW) - rjr->rj_tl1) >
	 MESSAGETIME)
	  {
		rjr->rj_tl1 = tl2;	retval |= 2;
	  }
	return retval;
}

int rejdie(int irv)
{ /* This closes off a run of `reject' and returns its first argument */
	REJECTRUN	*rjr;

	if (NULL == (rjr = managerjp(03))) return irv;
	if (NULL != rjr->rj_fpl) (void)fclose(rjr->rj_fpl);
	if (NULL != rjr->rj_fpcheck) (void)fclose(rjr->rj_fpcheck);
	rjr->rj_fpl = NULL;	rjr->rj_fpcheck = NULL;
	rjr = managerjp(0);	globalkill();	return irv;
}

int wobcmp(WORKINGOB *wp0, WORKINGOB *wp1)
{
	return	(fabs(wp0->w_gofp) < fabs(wp1->w_gofp) ? -1 : 1);
}

void redo_ob(WORKINGOB *wobp)
{ /* this recomputes the totals upon restoring an observation */
	register WORKINGOB	*wp;
	double			devdel, inten, wt;
	REJECTRUN		*rjr;

	if (NULL == (rjr = managerjp(03))) return;
	if (!(wobp->w_flag & W_ISMARKED)) return;
	wobp->w_flag &= ~W_ISMARKED;	/* unmark the observation */
	if ((wt = wobp->w_sigsc) < FUZZ) wt = 1.;
	wt = 1. / (wt * wt);	inten = wobp->w_isc;
	(rjr->rj_nobaf)++;	/* increment the count of usable observations */
	/* increment the sums that lead to the usable values */
	rjr->rj_e0 += wt;	rjr->rj_e1 += wt * inten;
	rjr->rj_e2 += wt * inten * inten;
	rjr->rj_sigaf = 1. / sqrt(rjr->rj_e0);
	rjr->rj_ibaraf = rjr->rj_e1 / rjr->rj_e0;
	if (rjr->rj_nobaf == 1)
	  {
		rjr->rj_gofaf = wobp->w_sigsc;
		rjr->rj_worstaf = wobp->w_dev = 0;	return;
	  }
	for (wp = rjr->rj_wobs, rjr->rj_worstaf = rjr->rj_gofaf = 0;
	 wp < rjr->rj_wobend; wp++)
	   {
		inten = wp->w_isc;	devdel = inten - rjr->rj_ibaraf;
		devdel *= (devdel * rjr->rj_e0 / (double)rjr->rj_nobaf);
		wobp->w_dev = devdel;
		if (!(wobp->w_flag & W_ISMARKED))
		  {
			if (devdel > rjr->rj_worstaf) rjr->rj_worstaf = devdel;
			rjr->rj_gofaf += devdel;
		  }
	   }
	rjr->rj_gofaf /= (double)(rjr->rj_nobaf - 1);
}

int input_means(MERGEDREF *mrp)
{ /*	This does the first part of the analysis of a reflection.
	It computes the sums that go into the mean and goodness-of-fit
	calculations.  It also does the setup for the "WORKINGOB"
	structure associated with the reflection.
	It returns the index of the first primary observation (firstpri). */
	register int		i, j, k;
	register ABBROB		*abp;
	register WORKINGOB	*wobp;
	int			firstpri, nob, run_num, posom, shind;
	char			firstbij;
	double			ifactor, inten, qv, res, scfac, thisdev;
	double			varob, wi, *fp2, avec[MAXVECPARS];
	float			*fp1;
	xgulong			ulsh;
	CRYSTAL			*xhp;
	SHIFT			*tsh;
	REJECTRUN		*rjr;

	if (NULL == (rjr = managerjp(03))) return 0;
	nob = mrp->mr_nob;	xhp = rjr->rj_xhdp;
	rjr->rj_nobused = 0;		/* clear count of used observations*/
	if (nob < 2) return 0;	/* Man kann nichts tun damit ... */
		/* clear the averaging sums */
	rjr->rj_e1 = rjr->rj_e0 = rjr->rj_e2 =rjr->rj_gof = rjr->rj_worst = 0.;
	res = mrp->mr_stl;
	firstpri = mrp->mr_nobpri;	firstbij = mrp->mr_nobbij;
	j = xhp->xt_intl;	if ((j < 200) && (xhp->xt_isys == 7)) j += 100;
	j = iscentric(mrp->mr_hkl, j);
	for (i = 0, wobp = rjr->rj_wobs; i < nob; i++) /* get the means, etc. */
	   {
		abp = &(mrp->mr_ob[i]);	/* pointer to the observation */
		wobp->w_flag = (abp->abo_shift < 0);
		ulsh = LSH(abp->abo_runno, 24) | LSH(abp->abo_chamno, 16) |
			abp->abo_shift;
		if (0 <= (shind = find_shift(rjr->rj_shlp, ulsh,
		 rjr->rj_nscales)))
		  { /* observation is scaled: set the flag properly */
			if (firstpri) /* first primary observation */
			  {
				wobp->w_flag |= W_1ST_OF_TYPE; firstpri = 0;
			  }
			else if (firstbij && i >= (int)mrp->mr_nobpri)
			  {	
				wobp->w_flag |= W_1ST_OF_TYPE; firstbij = 0;
			  }
			else wobp->w_flag |= W_ORDINARY_OB;
			if (i >= (int)mrp->mr_nobpri)
				wobp->w_flag |= W_ISBIJVOET;
			else if (j) wobp->w_flag |= W_ISCENTRIC;
			wobp->w_shift = abp->abo_shift;
			wobp->w_runno = abp->abo_runno;
			wobp->w_chamno = abp->abo_chamno;
			wobp->w_omob = abp->abo_omob;
			wobp->w_xob = abptox(abp);
			wobp->w_yob = abptoy(abp);
			posom = (abp->abo_omob < 0 ?
				36000L + (int)abp->abo_omob : abp->abo_omob);
			wobp->w_omxy = LSH(abp->abo_yob, 24) |
				LSH(abp->abo_xob, 16) | (xgulong)posom;
			tsh = rjr->rj_shlp + shind;
			scfac = terpscale(abp->abo_omob, tsh, res);
			if (rjr->rj_vshp)
			  {
				run_num = abp->abo_shift / 200;
				refvec(abp, rjr->rj_rub + run_num,
				 rjr->rj_scflags, rjr->rj_nvpars, avec);
				fp1 = &((rjr->rj_vshp + run_num)->v_coef[0]);
				for (k = rjr->rj_nvpars, qv = 0.,
				 fp2 = &(avec[0]); k--; fp1++, fp2++)
					qv += *fp1 * *fp2;
				scfac *= qv;
			  }
			if (scfac < DFUZZ) scfac = 100. * DFUZZ;
			/* compute the scaled intensity and sigma */
			wobp->w_isc = inten = abp->abo_ii / scfac;
			varob = abp->abo_sigma;
			ifactor = tsh->shvar;
			if (rjr->rj_flags & RJ_MACHINESIG)
				ifactor += MACHINE_FAC;
			wi = varob * varob + inten * inten * ifactor;
			if (wi < MINVAR) wi = MINVAR;
			wobp->w_sigsc = wi = sqrt(wi) / scfac;
			wi = 1. / (wi * wi); /* least-squares weigthing */
			/* increment the sums for <I> and goodness of fit */
			rjr->rj_e1 += wi * inten;	/* numerator of <I> */
			rjr->rj_e0 += wi;		/* denominator of <I> */
			rjr->rj_e2 += wi * inten * inten; /* 2nd moment ... */
			rjr->rj_nobused++;	wobp++;
		  }
	   }
	rjr->rj_wobend = wobp;	/* record address of the end of the list */
	if (rjr->rj_nobused < 2) return firstpri; /* < 2 scaled observations */
	if (rjr->rj_e0 < FUZZ) return firstpri;	/* sigmas too big */
	rjr->rj_ibar = rjr->rj_e1 / rjr->rj_e0;		/* Y or <I> */
	rjr->rj_sigysq = 1. / rjr->rj_e0;		/* sigmasquared ofY */
	rjr->rj_sig = sqrt(rjr->rj_sigysq);	/* get the sigma of y */
	if (rjr->rj_ibar > FUZZ)
	  {
		rjr->rj_rfacin += rjr->rj_e2 - rjr->rj_e1 * rjr->rj_e1 /
			rjr->rj_e0;	rjr->rj_rfdenin += rjr->rj_e2;
		rjr->rj_onobused += rjr->rj_nobused; 
	  }
	for (wobp = rjr->rj_wobs; wobp < rjr->rj_wobend; wobp++)
	   {
		thisdev = wobp->w_isc - rjr->rj_ibar;
		thisdev *= thisdev * rjr->rj_e0 / (double)(rjr->rj_nobused);
		wobp->w_dev = thisdev;
		if (thisdev > rjr->rj_worst) rjr->rj_worst = thisdev;
		rjr->rj_gof += thisdev;
	   }
	rjr->rj_gof /= (double)(rjr->rj_nobused - 1);
	rjr->rj_ibaraf = rjr->rj_ibar;	rjr->rj_sigaf = rjr->rj_sig;
	rjr->rj_gofaf = rjr->rj_gof; /*nothing deleted so far*/
	rjr->rj_worstaf = rjr->rj_worst;		return firstpri;
}

void use_deleted_sums(WORKINGOB *wobmis)
{ /* this sets the statistical sums for the whole reflection equal to
	the corresponding sums associated with deleting observation "wobmis"*/
	register WORKINGOB	*wobp;
	double			thisdev;
	REJECTRUN		*rjr;

	if (NULL == (rjr = managerjp(03))) return;
	rjr->rj_e0 = wobmis->w_e0;	rjr->rj_e1 = wobmis->w_e1;
	rjr->rj_e2 = wobmis->w_e2;
	rjr->rj_gofaf = wobmis->w_gofp;	rjr->rj_worstaf = wobmis->w_worst;
	if (rjr->rj_e0 < FUZZ)
	  {
		rjr->rj_nobaf = 0;	return;
	  }
	rjr->rj_ibaraf = rjr->rj_e1 / rjr->rj_e0;
	rjr->rj_sigaf = sqrt(1. / rjr->rj_e0);
	if (rjr->rj_nobaf)
	  for (wobp = rjr->rj_wobs; wobp < rjr->rj_wobend; wobp++)
	   {
		thisdev = wobp->w_isc - rjr->rj_ibaraf;
		wobp->w_dev = thisdev * thisdev * rjr->rj_e0 /
			(double)rjr->rj_nobaf;
	   }
	else for (wobp = rjr->rj_wobs; wobp < rjr->rj_wobend; wobp++)
		wobp->w_dev = 0.;
}

WORKINGOB *get_deleted_sums(void)
{ /* this obtains the statistical sums analogous to e0, e1, and e2 for
	the reflection minus each individual observation.  It returns
	a pointer to the observation which has the smallest value of gofp,
	i.e. the observation whose deletion would help the most.
	Change as of Sept 94: the "deleted sums" (w_e0, etc.) are computed
	bearing in mind that some observations have been deleted.! */

	register WORKINGOB	*wobp, *wp, *wobbest;
	double			devdel, e0d, e1d, e2d, inten, wdel, wt, ydel;
	REJECTRUN		*rjr;

	if (NULL == (rjr = managerjp(03))) return NULL;
	rjr->rj_bestgofp = rjr->rj_gofaf * 10000;	wobbest = rjr->rj_wobs;
	for (e0d = e1d = e2d = 0, wobp = rjr->rj_wobs;
	 wobp < rjr->rj_wobend; wobp++)
	   if (!(wobp->w_flag & W_ISMARKED))
	     {
		wt = wobp->w_sigsc;
		if (wt < FUZZ || wt == 0.)
		  {
			rjr->rj_nobaf = 0;	return rjr->rj_wobs;
		  }
		wt = 1. / (wt * wt);	inten = wobp->w_isc;
		e0d += wt;	e1d += wt * inten;
		e2d += wt * inten * inten;
	     }
	for (wobp = rjr->rj_wobs; wobp < rjr->rj_wobend; wobp++)
	   if (!(wobp->w_flag & W_ISMARKED))
	     {
		wobp->w_gofp = wdel = 0.;
		wt = wobp->w_sigsc;
		if (wt < FUZZ || wt == 0.)
		  {
			rjr->rj_nobaf = 0;	return rjr->rj_wobs;	}
		wt = 1. / (wt * wt);	inten = wobp->w_isc;
		wobp->w_e0 = e0d - wt;	wobp->w_e1 = e1d - inten * wt;
		wobp->w_e2 = e2d - inten * inten * wt;
		if (rjr->rj_nobaf > 1 && wobp->w_e0 > FUZZ)
		 {
			ydel = wobp->w_e1 / wobp->w_e0;
		   for (wp = rjr->rj_wobs; wp < rjr->rj_wobend; wp++)
		    if ((wp != wobp) && !(wp->w_flag & W_ISMARKED))
		      {
			inten = wp->w_isc;	devdel = inten - ydel;
			devdel *= (devdel * wobp->w_e0) / (rjr->rj_nobaf - 1);
			if (devdel > wdel) wdel = devdel;
			wobp->w_gofp += devdel;
		      }
		    if (rjr->rj_nobaf > 2)
			wobp->w_gofp /= (double)(rjr->rj_nobaf - 2);
		    else wobp->w_gofp = 1. / sqrt(wobp->w_e0);
		  }
		else {
			wdel = 0;	wobp->w_gofp = wobp->w_sigsc;
		     }
		wobp->w_worst = wdel;
		if (wobp->w_gofp < rjr->rj_bestgofp)
		  {
			rjr->rj_bestgofp = wobp->w_gofp; wobbest = wobp;
		  }
	     }
	return wobbest;
}

int refl_process(MERGEDREF *mrp)
{ /*	This routine is called for every reflection in the multiref.
	It computes the mean intensity and a standard error for it.
	If fewer than 2 obs are present, it returns.  Deviation values for
	every observation are computed; if the worst deviation is
	less than the cutoff value `qval', we're done.
	Otherwise we loop through the observations,
	deleting one at a time until:
	(1) the worst deviation is less than the cutoff;
	(2) the goodness-of-fit parameter has stopped improving
		with the deletions; OR
	(3) we have stripped away all but two observations.
	Then we look to see if the worst deviation is greater than the cutoff;
	if it is, we display the reflection for the user to look at;
	if not, we quit.
	Return value is an or-ing of these values, some of which
	 come from a call to `save_refl':
	  0	no deletions performed.
	  1	one or more automatic deletion flags set.
	  2	there has been a long enough delay that we should remind			 the user that we're still alive.
	  4	We can't decide what to do: we'll have to let the user decide.
	010	we've overloaded the buffer for deletable observations.
	020	memory failure for rjr. */
	register int		i, retnow;
	register WORKINGOB	*lws, *wobdel;
	int			prim1;
	short			lphkl[3];
	CRYSTAL			*xhp;
	REJECTRUN		*rjr;

	if (NULL == (rjr = managerjp(03))) return 020;
	xhp = rjr->rj_xhdp;
	prim1 = input_means(mrp);	/* compute <I>, S.E.M.<I> */
	if ((rjr->rj_nobaf = rjr->rj_nobused) < 2) return 0;
		/* all deviations small: increment rfactor sums and quit*/
	if (rjr->rj_worst < rjr->rj_qval)
	  {
		rjr->rj_rfacout += rjr->rj_e2 - rjr->rj_e1 * rjr->rj_e1 /
			rjr->rj_e0;
		rjr->rj_rfdenout += rjr->rj_e2;
		rjr->rj_onobout += rjr->rj_nobaf;
		if (rjr->rj_flags & RJ_HUMANALL) retnow = 1;
		return 0;
	  }
	else retnow = 0;
	munpack(mrp->mr_hkl, lphkl, xhp->xt_intl); /* get signed hkl */
	/* if no primaries were used, we print the Bijvoet's indices */
	if ((!(mrp->mr_nobpri)) || (prim1))
		for (i = 3; i--; ) lphkl[i] = -lphkl[i];
	if (retnow) return 4;
	if (rjr->rj_nobaf == 2)	/* we treat doubly-observed refls specially */
	 {
		if ((rjr->rj_worst < rjr->rj_qval2) ||
		 (rjr->rj_flags & RJ_2_SKIP))
		  {
			rjr->rj_rfacout += rjr->rj_e2 - rjr->rj_e1 *
			rjr->rj_e1 / rjr->rj_e0;
			rjr->rj_rfdenout += rjr->rj_e2;
			rjr->rj_onobout += rjr->rj_nobaf;	return 0;
		  }
			/* let user decide */
		if ((rjr->rj_flags & RJ_2_DISPLAY) &&
		 !(rjr->rj_flags & RJ_BATCH)) return 4;
		/* this call returns the observation with the LARGER sigma.
		thus if we have invoked reject with the default or -m option,
		this observation will be deleted and the other kept. */
		wobdel = get_deleted_sums();	lws = rjr->rj_wobs;
		if (rjr->rj_flags & RJ_2_MAXI) /* keep ob with bigger I */
		  wobdel = (lws->w_isc < (lws+1)->w_isc ? lws : lws+1);
		else if (rjr->rj_flags & RJ_2_KILL) /* kill the whole thing */
		  {
			lws->w_flag |= W_AUTOMARKED;
			(lws+1)->w_flag |= W_AUTOMARKED;
		  }
		else if (rjr->rj_flags & RJ_2_MAXIOS) /* ob with max I/sig */
		  wobdel = ((lws->w_isc / lws->w_sigsc) <
			((lws+1)->w_isc / (lws+1)->w_sigsc) ? lws : lws + 1);
		/*note that in case 'm', we just go on to the flag operation*/
		wobdel->w_flag |= W_AUTOMARKED;
		if (rjr->rj_flags & RJ_HUMANBAD) return 4;
		use_deleted_sums(wobdel); /* recompute <I>, etc.*/
		return save_refl();	/* going to delete something */
	  }
	for ( ; rjr->rj_nobaf > 2 && rjr->rj_worstaf > rjr->rj_qval;
	 (rjr->rj_nobaf--))
	   { /* there are some bad obs's. Compute the mean
			intensity and the g.o.f. with each ob missing in turn*/
		wobdel = get_deleted_sums(); /*pick an ob to try to delete*/
			/* did it help? */
		if (rjr->rj_bestgofp > MIN_IMPROVEMENT * rjr->rj_gofaf) break;
		wobdel->w_flag |= W_AUTOMARKED; /* mark it for deletion*/
		use_deleted_sums(wobdel); /* recompute <I>, etc.*/
	   }
	/* if there are still problems and we're in menu mode, display it */
	if ((rjr->rj_flags & RJ_HUMANANY) && (rjr->rj_worstaf > rjr->rj_qval))
		return 4;
	return save_refl();	/* record deletions we have set up */
}

int rejswitch(int argc, char *argv[])
{ /*	This decodes the switch-style arguments to "reject".
	The following switch arguments are recognized:
		-a	Interactive ("ask for menus") mode.
			In this mode some CURSES menus are set up,
			and the user is prompted for decisions about
			each ambiguous reflection. This used to be the
			default, and is no longer.
		-b	batch mode.  The interactive menu isn't set up at all;
			instead, the flagging the program wants to do is
			performed, and nothing else. This is now the default.
		-c	checked-refls list written to "checkrefl"
		-f<val>	Find the q (=|I-<I>|/sigma) value for which
			the estimated fraction deleted rises above <val>.
			Use this value instead of the externally determined q.
		-m	Inflate the sigmas by a machine-error term, i.e.
			sigma' = sqrt(sigma*sigma + I * I * MACHINE_FAC)
		-n	Do not merge Bijvoets (default == merge them)
		-o	Do merge Bijvoets (usually the default)
		-p	Preserve old deletions, i.e. if an observation was
			marked for deletion previously, leave it marked
			(this always happened till Sept. 1994)
		-q<val>	Use <val> as the deviation cutoff instead of
			the program-supplied value
		-r h k l	Resume a run by skipping through all the
				refls up through (h,k,l);(probably useless now)
		-t<val>	Set the q (=|I-<I>|/sigma) so that the
			 after-deletion R-value will be <val>.
		-u	Don't inflate the sigmas by a machine-error term
		-v	verbose logfile: each deleted observation has
			its hkl, X,Y,omega,I,sigma, and the deletion
			statistics printed to XLOG in a moderately
			sensible format.  This used to be default, but
			I'm killing it now unless you use the -v.
		-w	Don't preserve the old deletions.
		-y	Don't produce a verbose logfile.
		The following Booleans regulate how the program deals with
		reflections with precisely two observations:
		-g	keep the reflection with the smaller sigma
		-i	ignore the reflection, i.e. keep both observations
		-k	flag both observations for deletion
		-l	keep the reflection with the maximum intensity
		-s	display the reflection for the user to see
			 (reverts to -x if -b is on)
		-x	keep the reflection with maximum I/sigma (default) */
	char		**dcp, **flags;
	REJECTRUN	*rjr;

	if (NULL == (rjr = managerjp(07))) return EALLOC;
	rjr->rj_flags = RJ_2_MAXIOS | RJ_BATCH; /* default is all flag bits
	 off, batch mode turned on; two-ob refls: keep larger I/sigma */
	flags = flagproc(argc, argv, "ABCFGIKLMNOPQSTUVWXY");
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	   switch (**dcp)
	     {
		case	'A':	/* non-batch mode: menu stuff */
			rjr->rj_flags &= ~(RJ_BATCH);
			rjr->rj_flags |= RJ_HUMANAMBIG;	break;
		case	'B':	/* batch mode: no menu stuff */
			rjr->rj_flags |= RJ_BATCH;
			rjr->rj_flags &= ~(RJ_HUMANANY);	break;
		case	'C':	/* make detailed check-refls list */
			if (qerror((AVOID *)(rjr->rj_fpcheck =
			 xgopen("checkrefl","w")),
			 "open output checked-refl list")) return EOPENERR;
			break;
		case	'F':	/* specify fraction of data to delete */
			rjr->rj_qfrac = atof((*dcp) + 1);	break;
		case	'G':	rjr->rj_flags &= ~(RJ_2_CHOICES);
			rjr->rj_flags |= RJ_2_MINSIG;	break; /*keep min sig*/
		case	'I':	rjr->rj_flags &= ~(RJ_2_CHOICES); /*keep both*/
			rjr->rj_flags |= RJ_2_SKIP;	break;
		case	'K':	rjr->rj_flags &= ~(RJ_2_CHOICES);
			rjr->rj_flags |= RJ_2_KILL; break; /*kill both obs */
		case	'L':	rjr->rj_flags &= ~(RJ_2_CHOICES);
			rjr->rj_flags |= RJ_2_MAXI;	break; /* keep max I */
		case	'M': /* inflate sigmas with "machine error" term */
			rjr->rj_flags |= RJ_MACHINESIG;	break;
		case	'N': /* don't merge Bijvoets*/
			rjr->rj_flags |= RJ_SEPBJ;	break;
		case	'O':	/* do merge Bijvoets */
			rjr->rj_flags &= ~RJ_SEPBJ;	break;
		case	'P': /* preserve the originally deleted observations*/
			rjr->rj_flags |= RJ_PRESERVEOLD;	break;
		case	'Q': /* set quality value */
			rjr->rj_qval = atof((*dcp) + 1);	break;
		case	'S':	rjr->rj_flags &= ~(RJ_2_CHOICES);
			rjr->rj_flags |= RJ_2_DISPLAY;	break; /*display 2-ob*/
		case	'T': /* specify R-value to arrive at */
			rjr->rj_qrval = atof((*dcp) + 1);	break;
		case	'U': /* don't inflate sigmas with machine errors */
			rjr->rj_flags &= ~RJ_MACHINESIG;	break;
		case	'V':	/* verbose logfile */
			rjr->rj_flags |= RJ_VERBOSE;	break;
		case	'X':	rjr->rj_flags &= ~(RJ_2_CHOICES);
			rjr->rj_flags |= RJ_2_MAXIOS;	break; /*keep max I/s*/
		case	'W': /* don't preserve originally deleted observations*/
			rjr->rj_flags &= ~RJ_PRESERVEOLD;	break;
		case	'Y':	/* no verbose logfile */
			rjr->rj_flags &= ~RJ_VERBOSE;	break;
		default:	break;
	   }
	return EOK;
}

static void sum_rejset(FILE *fp)
{ /*	This summarizes the setup of a "reject" run
	and prints the summary on stream "fp". */
	short		lphkl[3];
	CRYSTAL		*xhp;
	REJECTRUN	*rjr;
	static char	*flist[] = {"multiref", "scalein", "xlog",
		"deletions", "mrefout", "checkrefl", NULL};
	static char	ftype[] = {1, 1, 2, 2, 2, 2, 0 };

	if (NULL == (rjr = managerjp(03))) return;
	ftype[5] = (NULL != rjr->rj_fpcheck) ? 2 : 0;
	header(fp, REDUCTION, flist, ftype);
	fprintf(fp,"Bijvoets will%sbe merged\n",
		(rjr->rj_flags & RJ_SEPBJ ? " NOT " : " "));
	if (rjr->rj_qval > 0.)
	  fprintf(fp, "|I-<I>|/sigma criterion value = %8.4f\n",rjr->rj_qval);
	if (rjr->rj_flags & RJ_RESUME)
	  {
		xhp = rjr->rj_xhdp;
		munpack(rjr->rj_shkl, lphkl, xhp->xt_intl);
		fprintf(fp,
	 "Only reflections starting at [%4d,%4d,%4d] will be checked\n",
		 lphkl[0], lphkl[1], lphkl[2]);
	  }
	else fprintf(fp,"All reflections will be checked\n");
	if (rjr->rj_flags & RJ_MACHINESIG)
	  {
		fprintf(fp,
	 "Sigma values will be inflated by a machine-error term:\n");
		fprintf(fp,
	 "Sigma' = sqrt(sigma * sigma + I * I * %9.7f)\n", MACHINE_FAC);
	  }
	fprintf(fp, "bad reflections with exactly 2 observations");
	switch (rjr->rj_flags & RJ_2_CHOICES)
	  {
		case	RJ_2_SKIP:	fprintf(fp,
		 " will not be scrutinized\n");	break;
		case	RJ_2_DISPLAY:	fprintf(fp,
		 " will be displayed for user choice\n"); break;
		case	RJ_2_MINSIG:	fprintf(fp,
		 ": ob with smaller sigma will be kept\n"); break;
		case	RJ_2_KILL:	fprintf(fp,
		 " will be deleted altogether\n"); break;
		case	RJ_2_MAXI:	fprintf(fp,
		 ": with larger intensity will be kept\n"); break;
		case	RJ_2_MAXIOS:	fprintf(fp,
		 ": ob with larger I/sigma will be kept\n"); break;
	  }
	if (rjr->rj_flags & RJ_BATCH) fprintf(fp,
 "No interactive menus will be created; automatic deletions will be done\n");
	else fprintf(fp,
 "Ambiguous reflections will be displayed interactively for user choice\n");
	if (rjr->rj_flags & RJ_VERBOSE) fprintf(fp,
 "A list of deleted observations will be written to the logfile\n");
	if (rjr->rj_flags & RJ_PRESERVEOLD) fprintf(fp,
 "Previously-deleted observations will remain deleted\n");
	else fprintf(fp,
 "Previously-deleted observations will be re-examined and perhaps restored\n");
}

static double getq(xglong nobs)
{ /*	Returns a residual cutoff value appropriate to the total number of
	observations in the multiref.  The numbers come from a computation
	of the number of observations which on the average would generate
	exactly one deviant with the returned residual value. */
	double cutoff;
	static float	onekval[100] = {
	 3.290,3.480,3.580,3.660,3.710,3.760,3.800,3.830,3.860,3.890,
	 3.920,3.940,3.950,3.970,3.980,4.000,4.010,4.030,4.040,4.050,
	 4.060,4.070,4.080,4.090,4.100,4.110,4.120,4.130,4.140,4.150,
	 4.155,4.170,4.175,4.180,4.190,4.195,4.200,4.205,4.210,4.220,
	 4.225,4.230,4.235,4.240,4.245,4.250,4.255,4.260,4.263,4.267,
	 4.270,4.275,4.280,4.286,4.288,4.293,4.296,4.300,4.304,4.307,
	 4.310,4.313,4.316,4.320,4.323,4.326,4.330,4.333,4.336,4.340,
	 4.343,4.346,4.350,4.353,4.356,4.360,4.363,4.366,4.370,4.373,
	 4.375,4.378,4.380,4.383,4.385,4.388,4.390,4.392,4.394,4.396,
	 4.398,4.400,4.402,4.404,4.406,4.408,4.410,4.413,4.415,4.417};
	static float	tenkval[100] = {
	 3.290,3.920,4.060,4.155,4.225,4.270,4.310,4.343,4.375,4.398,
	 4.430,4.440,4.460,4.480,4.510,4.520,4.540,4.550,4.560,4.570,
	 4.580,4.590,4.600,4.610,4.620,4.625,4.630,4.640,4.645,4.650,
	 4.660,4.665,4.670,4.675,4.680,4.685,4.690,4.700,4.705,4.710,
	 4.715,4.720,4.725,4.730,4.735,4.740,4.743,4.747,4.750,4.754,
	 4.760,4.763,4.767,4.770,4.773,4.776,4.779,4.783,4.788,4.792,
	 4.795,4.798,4.801,4.804,4.808,4.810,4.813,4.816,4.820,4.823,
	 4.825,4.827,4.829,4.831,4.833,4.836,4.840,4.843,4.846,4.848,
	 4.852,4.854,4.856,4.858,4.860,4.862,4.864,4.867,4.870,4.872,
	 4.874,4.876,4.878,4.880,4.882,4.884,4.886,4.888,4.890,4.892};

	if (nobs < 100000L)
		cutoff = onekval[((short)((int)nobs/1000))];
	else if (nobs < 1000000L)
		cutoff = tenkval[((short)((int)nobs/10000))];
	else	cutoff = 4.900;
	return cutoff;
}

static int rejstart(void)
{ /*	This does most of the setup operations for a run of `reject'. */
	int		gotone, i, nshifts, shadd, xystuf;
	off_t		mrpos;
	VSPLINE		*sgsp;
	CRYSTAL		*xhp;
	DATARUN		*rhp;
	RUNWORK		*rup;
	VECPAR		*vpp;
	MERGEDREF	*mrq;
	REJECTRUN	*rjr;

	if (NULL == (rjr = managerjp(03))) return EALLOC;
	if (qerror((AVOID *)(rjr->rj_fpmi = xgopen("multiref", "r")),
		"open input multiref")) return EOPENERR;
	if (EOK != (i = headmem(&(rjr->rj_xhdp), &(rjr->rj_rhdp),1))) return i;
	xhp = rjr->rj_xhdp;	rhp = rjr->rj_rhdp;
	if (verror(0 < xgread((void *)xhp, XG_CRYSTAL, rjr->rj_fpmi),
		"read crystal header from input multiref")) return EREADERR;
	/* if user didn't specify a q value, calculate it from redundancy */
	if ((rjr->rj_qval < DFUZZ) && (rjr->rj_qfrac < DFUZZ) &&
	 (rjr->rj_qrval < DFUZZ)) rjr->rj_qval = getq(xhp->xt_ninteg);
	if (qerror((AVOID *)(rjr->rj_fpl = propen("xlog")),
		"open output logfile")) return EOPENERR;
	sum_rejset(stdout);	sum_rejset(rjr->rj_fpl);
	if (rjr->rj_fpcheck) sum_rejset(rjr->rj_fpcheck);
	if (rjr->rj_fpcheck)
	  {
		fprintf(rjr->rj_fpcheck,
		 "\n    h   k   l ob shift omega    X    Y    inten.");
		fprintf(rjr->rj_fpcheck,
		 "    sigma    gof  obs. status  deviat. chmbr # run #\n");
	  }
	if (rjr->rj_qval > -DFUZZ) /* compare with SQUARE of criterion */
	  {
		rjr->rj_qval *= rjr->rj_qval;
			/* more lenient with 2 observations */
		rjr->rj_qval2 = rjr->rj_qval * TWO_OB_RATIO;
	  }
	/* reserve memory for working obs, multiref struct, RUNWORK, and vshp*/
	rjr = managerjp(02507); /* reserve memory for working-observations */
	for (i = nshifts = shadd = 0, rup = rjr->rj_rub; i < xhp->xt_nruns;
	 i++, shadd += 200, rup++)
	   {
		if (verror(0 < xgread((void *)rhp, XG_DATARUN, rjr->rj_fpmi),
		 "read datarun header from input multiref")) return EREADERR;
		(void)keep_dimen(rhp);
		workrun(rhp, rup, xhp, shadd);
		nshifts += rhp->run_nshifts;
	   }
	if (NULL == (rjr = managerjp(0207))) /* read in the scale factors */
	  {
		fprintf(stderr, "Cannot read input scale-factor file\n");
		return EREADERR;
	  }
	/* examine the scale-factor output to see if we actually got any
	 geometry-dependent coefficients. */
	gotone = 0;
	for (vpp = rjr->rj_vshp; vpp < rjr->rj_vshp + xhp->xt_nruns; vpp++)
	   {
		for (i = gotone = 0; i < MAXVECPARS; i++)
		   if (fabs(vpp->v_coef[i]) > FUZZ)
		     {
			gotone = 1; break;
		     }
		if (gotone) break;
	   }
	rjr->rj_scflags &= ~(SC_DIFFRACTION | SC_EXITBEAM);
	if (!gotone)
	  {
		memzap((AVOID **)&rjr->rj_vshp, "rjr->rj_vshp in rejstart");
		rjr->rj_nvpars = 0 ;
	  }
	else {
		if (rjr->rj_vshp->v_ptype == 'd')
		  {
			rjr->rj_nvpars = 10; rjr->rj_scflags |= SC_DIFFRACTION;
		  }
		else {
			rjr->rj_nvpars = 9; rjr->rj_scflags |= SC_EXITBEAM;
		     }
	/* we need some spline information to do the xy conversions */
		if (NULL == (sgsp = gspline(5, &i))) return i;
	     }
	if ((rjr->rj_qfrac > DFUZZ) || (rjr->rj_qrval > DFUZZ))
	  {
		mrpos = ftell(rjr->rj_fpmi); /* remember where we were */
		dihinit();	/* initialize some values */
		xystuf = 0;	mrq = rjr->rj_mrp;
		while (0 < xgread((AVOID *)mrq, XG_MERGEDREF, rjr->rj_fpmi))
			xystuf += dihcalc(mrq);
		if (-1 == fseek(rjr->rj_fpmi, (long)mrpos, 0)) /* rewind */
		  {
			fprintf(stderr, "Error seeking to end of headers\n");
			return ESEEKERR;
		  }
		dihfinal(); /* final calculations */
		dihshow(rjr->rj_fpl, xystuf); /* output results */
		if (rjr->rj_qfrac > DFUZZ)
		  rjr->rj_qval = qfromfrac(rjr->rj_qfrac);
		else if (rjr->rj_qrval)
		  rjr->rj_qval = qfromrval(rjr->rj_qrval);
		printf( "|I-<I>|/sigma criterion value = %8.4f", rjr->rj_qval);
		fprintf(rjr->rj_fpl,
		  "|I-<I>|/sigma criterion value = %8.4f", rjr->rj_qval);
		if (rjr->rj_qfrac > DFUZZ)
		  {
			printf(" based on deleting %7.4f of the data\n",
				rjr->rj_qfrac);
			fprintf(rjr->rj_fpl,
			 " based on deleting %7.4f of the data\n",
				rjr->rj_qfrac);
		  }
		else if (rjr->rj_qrval > DFUZZ)
		  {
			printf(
			 " based on seeking Rmerge <= %7.4f\n", rjr->rj_qrval);
			fprintf(rjr->rj_fpl,
			 " based on seeking Rmerge <= %7.4f\n", rjr->rj_qrval);
		  }
		rjr->rj_qval *= rjr->rj_qval; /* use it as its square, again */
		rjr->rj_qval2 = rjr->rj_qval * TWO_OB_RATIO;
	  }
	/* reserve memory for deletable refls, counts of deleted obs per shift,
	 geometry-dependent bins, and the working MERGEDREF structure */
	rjr = managerjp(03037);
	rjr->rj_drefp = rjr->rj_drefs;
	rjr->rj_drefend = rjr->rj_drefs + (rjr->rj_xhdp)->xt_ninteg;
	if (!(rhp->run_width)) rhp->run_width = 512;
	if (!(rhp->run_height)) rhp->run_height = 512;
	rjr->rj_xbinwid = rhp->run_width / NXBINS;
	if (rhp->run_width - rjr->rj_xbinwid * NXBINS > 16) (rjr->rj_xbinwid)++;
	rjr->rj_ybinwid = rhp->run_height / NYBINS;
	if (rhp->run_height - rjr->rj_ybinwid * NYBINS> 16) (rjr->rj_ybinwid)++;
	xystuf = keep_dimen(rhp);	rjr->rj_widrat = xystuf >> 24;
	rjr->rj_htrat = (xystuf >> 16) & 0xFF;	rjr->rj_hto2 = xystuf & 0xFFFF;
		/* determine <omega> in each scaling group */
	get_omcen(rjr->rj_fpmi, rjr->rj_shlp, (int)(rjr->rj_nscales), 0);
	return EOK;
}

int rejsetup(int argc, char *argv[])
{ /*	Argument and file processor for the reject program.
	Files used are given conventional logical or environment names:
		multiref	input multiref
		mrefout		output multiref if used
		scalein		input scale shift file
		deletions	printable output deletion-list file
		xlog		printable logfile
		checkrefl	detailed checked-refls list	*/
	int		i;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	/* get the run-time arguments as "switches" */
	if (EOK != (i = rejswitch(argc, argv))) return i;
	return rejstart();	/* do the rest of the setup operations */
}

int autoreject(void)
{ /* Handles the rejection portion of the scalem procedure */
	int		procres, procres0, procres1, retv;
	MERGEDREF	mrbij, mrpri, *mrq;
	FILE		*lfpl;
	REJECTRUN	*rjp;

	(void)xgprogname("reject", 4);
	rjp = managerjp(03);
	if (EOK != (retv = rejstart())) return retv;
	while (0 != rejgref())	/* read data that the user wants */
	     { /* we set rjp->rj_fpl to NULL later, so
		 that's an indication that we've interrupted things with ^C. */
		if (NULL == rjp->rj_fpl) return rejdie(EOK);
		mrq = rjp->rj_mrp;
		if ((rjp->rj_flags & RJ_SEPBJ) &&
		 (mrq->mr_nobpri) && (mrq->mr_nobbij))
		  {
			splitmr(mrq, &mrpri, &mrbij);
			procres0 = refl_process(&mrpri);
			if (!(010 & procres0))
				procres1 = refl_process(&mrbij);
			else procres1 = 0;
			procres = procres0 | procres1;
		  }
		else procres = refl_process(mrq);
	     }
		/* flag: we did R-factor calcs on ALL refls */
	rjp->rj_flags |= RJ_ALLCALC;
	/* write out the output multiref */
	lfpl = rjp->rj_fpl;
	if ((NULL != rjp->rj_drefs) && (rjp->rj_drefp != rjp->rj_drefs))
	  {
		fprintf(lfpl, " %d reflections flagged for deletion\n",
			(1 + (int)(rjp->rj_drefp - rjp->rj_drefs)));
		if (EOK != (retv = apply_deletions())) return retv;
		last_calcs();
	  }
	else fprintf(lfpl, "No deletions requested: no multiref output\n");
	if (NULL != rjp->rj_fpl) xeflush(rjp->rj_fpl);
	(void)managerjp(0);
	return EOK;
}

/* end of rejcalc.c */
