/* scalemm.c -- toplevel routines for XGEN's post-processing function

	Copyright (c) 2004, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions; reorganize use of "{"
	20 Oct 2006: scmreject() moved to rejcalc.c and renamed to autoreject
	24 Jun 2006: include xgpostint.h rather than separate include files
	 5 Jun 2006: sumscalemrun eliminated, sccycles splits off precycle
	 3 Apr 2006: housekeeping in scalemmain
	20 Nov 2005: inline argument declarations and strict prototyping
	10 Oct 2005: housekeeping and elimination of all static globals
	28 Jun 2005: scmreject(): typecast pointer difference on printout
	11 May 2005: check return status from xgjournal
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	17 Sep 2004: -b option implemented; calls to do_cycle updated
		since the cycle # is now an element of SCALERUN
	28 Aug 2004: routines used only here are declared as static; cosmetics;
	 try_reindex() and mm_matrei,mm_vecrei elements MAKEMURUN elts used
	 4 Aug 2004: separate Friedel mates in `stats' if appropriate
	26 Jul 2004: minimum redundancy for Bijvoet-unmerged scaling
		implemented properly
	 2 Jun 2004: set STATRUN flag _after_ zeroing it (duh);
		zero out the MAKEMURUN structure
	23 May 2004: new application, built from existing applications
	 scalek, reject, makemu, stats, and spaceg

   Purpose: toplevel calling routines for `scalem'
   Cross references: sc*.c, rej*.c, spaceg.c, stat*.c, makemu.c

 "scalem" performs a linear scaling of crystallographic data from area
 detector data, iteratively computing averaged intensities and scale and
 temperature factors for "shifts", which are batches of collected reflections
 covering a limited range of scanning angle and (sometimes) detector area.
 It then looks through the data and flags certain observations as being
 outliers and excludes them from scaling. Then it scales the data again
 with a more sophisticated scaling algorithm, goes through another round
 of rejections (potentially restoring some observations to the usable list
 as well as potentially rejecting new ones), and then does another round
 of scaling with the more sophisticated algorithm. If the redundancy is
 high enough it will then scale the data a fourth time with the Friedel
 mates separated. At that point it will compute a set of statistics
 characterizing the data and  revalidate the space-group choice for the data.
 At that point it will write out ascii merged data files in several formats.
 Thus a run of "scalem -r -g0.01" with a reasonably high redundancy
 is equivalent to
	scalek -ktp1
	reject -bf0.01
	scalek -ktgp1
	reject -bf0.01
	scalek -ktghp1
	scalek -ktghnp1
	stats -on
	spaceg
	makemu -f
	mutos
 Details of the scaling and R-factor algorithms are given in the comments
 at the top of "sccalc.c".
 Runtime options are as follows:
	-a<val>	Minimum redundancy below which the Bijvoet-unmerged scaling
		will not be performed. Default value is 4.
	-c	Create detailed list of deleted (flagged, unused) observations
		during the outlier-rejection step.
	-f	In outputting data to the ascii formats at the end,
		write out structure amplitudes (F's) rather than intensities.
		The default is to write out intensities.
	-g<val>	During outlier rejection, exclude a fraction of the data
		 roughly equal to <val>. Thus -g0.01 will exclude roughly
		 one percent of the data.
	-i	In ascii outputs, write reflections out after converting
		the reflection indices according to a matrix multiplier,
		h' = Mh.
	-j<val>	Set minimum <I> value below which a reflection will not
		 contribute to scaling to <val>. Default is 0.
	-k	See "outlier rejection: 2 observations," below.
	-m	In ascii outputs, only write the reflection out if it has
		two or more observations.
	-n	In ascii outputs, set negative intensities to zero.
		This is default if -f is set.
	-o	In the statistical analyses, suppress output of the list
		of missing reflections0.
	-p	In ascii output, preserve the sin(theta) dependence of
		 intensities found in the computed scale factors.
		 If this flag isn't set, we do a recalculation of the
		 overall sin(theta) dependence of intensity.
	-r	Scale bottom and top halves of the detector together.
	-s<val>	In ascii output, set an overall scale factor for the
		 observed data to <val>. Thus -s2 will double the size
		 of all the output intensities or amplitudes above
		 what they would otherwise be.
	-t	In ascii output, write the reflections out in increasing
		 order of sin(theta)/lambda rather than sorted on (h,k,l).
	-u	In ascii output, use a different variance formula for the
		 merged reflections. See makemu.c
	-v	Verbose output from each step in the process.
		 By default, summaries of the scale factors and R-factors
		 are printed only on the 0th and last cycles.
		 Specifying -v puts the summaries out more often.
	-w	Wilson-statistics. The program computes initial
		 scale factors based on Wilson statistics in the
		 input data. If no SCALEIN file is specified,
		 these initial scale factors are used in the
		 scaling calculations; if SCALEIN is specified,
		 they're provided only as logged data.
	-z	Inflate the sigmas during rejection
		 by a machine-error term.
*************************************************************************/
#define		NEED_TIME	1
#define		NEED_STAT	1
#include	"xgpostint.h"
 /* default minimum redundancy so that we can do Friedel-separated scaling */
#define		DEF_MINREDUN	4

 typedef struct {
		SCALERUN	*sm_spi;
		REJECTRUN	*sm_rjp;
		STATRUN		*sm_star;
		MAKEMURUN	*sm_mmup;
		double		sm_minredun;
		} SCMWORK;

static int scmswitch(int argc, char *argv[], SCMWORK *scp)
{ /*	Processes switch-type arguments for scalem. */
	char		**dcp, **flags;
	int		ispec;
	double		fspec, specredu;
	SCALERUN	*spi;
	REJECTRUN	*rjp;
	STATRUN		*star;
	MAKEMURUN	*mmup;

	flags = flagproc(argc, argv, "ABCFGIJKMNOPQRSTUVWZ");
	specredu = 0;	scp->sm_minredun = DEF_MINREDUN;
	spi = scp->sm_spi;	rjp = scp->sm_rjp;	star = scp->sm_star;
	mmup = scp->sm_mmup;
	rjp->rj_qfrac = 0.015;
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	 {
	   fspec = atof(*(dcp) + 1);	ispec = atoi(*(dcp) + 1);
	   switch (**dcp)
	     {
		case	'A':	/* min redundancy for inclusion in scaling */
		 specredu = fspec;	break;
		case	'B':	/* # of resolution bins during stats output */
		 if ((ispec > 0) && (ispec <= MAXRSHELLS + 1))
			star->sta_nrshells = ispec - 1;	break;
		case	'C': /* try to open CHECKREFL for output */
		 rjp->rj_fpcheck = xgopen("checkrefl", "w"); break;
		case	'F': /* output F's, not I's */
		 mmup->mm_flags |= MUF_OUTPUT_F;	break;
		case	'G':	/* specify fraction of data to delete */
		 rjp->rj_qfrac = fspec;	break;
		case	'I': /* reindex reflections */
		 mmup->mm_flags |= MUF_REINDEX;	break;
		case	'J':	/* minimum <I> to use in scaling */
		 (void)sscanf((*dcp) + 1, "%hd", &(spi->sc_imin)); break;
		case	'K':	rjp->rj_flags &= ~(RJ_2_CHOICES);
		 rjp->rj_flags |= RJ_2_KILL; break; /*kill both obs */
		case	'M': /* only mult-obs.refls output */
		 mmup->mm_flags |= MUF_MULOBSERVED;	break;
		case	'N': /* set I<0 values to 0*/
		 mmup->mm_flags |= MUF_NONEGI;	break;
		case	'O':	/* don't write out a "missing" list */
		 star->sta_flags &= ~(ST_OUTMISS);	break;
		case	'P': /* preserve sin(theta) dependence of input
		  scale factors rather than correcting for it */
		 mmup->mm_flags |= MUF_PRESERVE;	break;
		case	'Q': /* specify |I-<I>|/sigma cutoff for rejection */
		 rjp->rj_qval = fspec;	break;
		case	'R':	/* scale top & bottom together */
		 spi->sc_flags |= SC_ALLOFDET;	break;
		case	'S': /*overall scale factor*/
		 (void)sscanf((*dcp) + 1,"%f", &(mmup->mm_oversc));	break;
		case	'T': /* sin(theta) order*/
		 mmup->mm_flags |= MUF_STLORDER;	break;
		case	'U': /* alternate variance */
		 mmup->mm_flags |= MUF_VAROFMEAN;	break;
		case	'V':	/* verbose output */
		 spi->sc_flags |= SC_VERBOSE; rjp->rj_flags |= RJ_VERBOSE;
		 star->sta_flags |= SP_VERBOSE;	break;
		case	'W':	/* initial scales by Wilson analysis*/
		 spi->sc_flags |= SC_WILSON;	break;
		case	'Z': /* inflate sigmas with "machine error" term */
		 rjp->rj_flags |= RJ_MACHINESIG;	break;
		default:	break;
	   } /* end of switch */
	 } /* end of for loop through switches */
	/* nonlinear and switch-to-nonlinear are incompatible */
	if ((spi->sc_flags & SC_NONLINEAR) && (spi->sc_flags & SC_SWITCH))
		spi->sc_flags &= ~SC_NONLINEAR;
	if (1 < specredu) scp->sm_minredun = specredu;
	/* if we've specified both a deletion q value and a desired fraction,
	 we let the deletion q value take precedence */
	if (rjp->rj_qval > 0) rjp->rj_qfrac = 0;
	return EOK;
}

static int scmargs(int argc, char *argv[], SCMWORK *scp)
{ /*	scmargs() does all setup necessary to begin processing scaling data.
	In this version, the names of the multiref, the input shift file,
	the output shift file, and the logfile are conventional:
	"multiref", "scalein", "scaleout", and "xlog". Therefore the only
	arguments to the program in this version are the switches.
	First, we establish default values of global variables;
	see "scdefs.h" for numerical or logical values of these defaults. */
	int		retv;
	SCALERUN	*spi;
	REJECTRUN	*rjp;
	STATRUN		*star;
	MAKEMURUN	*mmup;

	scp->sm_minredun = DEF_MINREDUN;	/* default minimum redundancy*/
	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort*/
	/* initialize the SCALERUN structure */
	spi = scp->sm_spi;	zeromem((void *)spi, sizeof (SCALERUN));
	spi->sc_tl0 = NOW;
	spi->sc_flags = SC_DEFAULT;	spi->sc_imin = SC_IMIN;
	spi->sc_nrepmin = SC_NOBSMIN;	spi->sc_maxcycles = SC_MAXCYCLES;
	spi->sc_convcrit = SC_CONVCRIT; spi->sc_nparpshift = SC_NPARAMS;
	spi->sc_appshift = 1.;		spi->sc_nparpshift = 1;
	/* initialize the REJECTRUN structure */
	rjp = scp->sm_rjp = managerjp(02467);
	rjp->rj_flags = RJ_2_MAXIOS | RJ_BATCH; /* default is all flag bits
	 off, batch mode turned on; two-ob refls: keep larger I/sigma */
	/* initialize the STATRUN structure */
	star = scp->sm_star;	zeromem((void *)star, sizeof (STATRUN));
	star->sta_flags = ST_OUTMISS;	star->sta_nrshells = -1;
	/* initialize the MAKEMURUN structure */
	mmup = scp->sm_mmup;	zeromem((void *)mmup, sizeof (MAKEMURUN));
	/* Examine arguments.  Those beginning with a '-' are interpreted
	 as switches and processed with 'scmswitch'.
	Any other arguments are ignored for the moment.  */
	if (EOK != (retv = scmswitch(argc, argv, scp))) return retv;
	for (argc--, argv++; argc; argc--, argv++) if ('-' != **argv) break;
	if (NULL == (rjp->rj_mrp = (MERGEDREF *)callop(1, sizeof (MERGEDREF),
		"rjp->rj_mrp in scmargs"))) return EALLOC;
	return retv;
}

static int scmend(int irv, SCALERUN *spi)
{ /*	This ends a run of `scalem' properly */
	QOB	*qow;

	if ((NULL != spi->sc_fpl) && (stdout != spi->sc_fpl))
		(void)fclose(spi->sc_fpl);
	if (NULL != spi->sc_fpmi) (void)fclose(spi->sc_fpmi);
	spi->sc_fpl = NULL;	spi->sc_fpmi = NULL;
	if (NULL != spi)
	  {
		memzap((AVOID **)&(spi->sc_refbase),
			"spi->sc_refbase in scalem");
		qow = spi->sc_obbase;
		if ((NULL != qow) && (spi->sc_flags & SC_VECREFINED))
		  memzap((AVOID **)&(qow->qo_sc), "qow->qo_sc in scalem");
		memzap((AVOID **)&(spi->sc_obbase), "qow in scalem");
		memzap((AVOID **)&(spi->sc_shwl), "spi->sc_shwl in scalem");
		memzap((AVOID **)&(spi->sc_vshp), "spi->sc_vshp in scalem");
		memzap((AVOID **)&(spi->sc_rub), "spi->sc_rub in scalem");
		memzap((AVOID **)&(spi->sc_shlp), "spi->sc_shlp in scalem");
	  }
	sccalc_kill(spi); globalkill();	return irv;
}

static int precycle(SCALERUN *spi)
{ /* This prepares for a series of scaling cycles within scalemm */
	char		*siname;
	struct stat	sibuf;

	(void)xgprogname("scalek", 4);	/* get the right label into logfiles */
	/* Now open the files according to the "trname" conventions. */
	if (qerror((AVOID *)(spi->sc_fpmi = xgopen("multiref","r")),
		"open input multiref")) return EOPENERR;
	/* see whether the scale input file exists,
	 is readable, and has positive length */
	siname = trname("scalein");
	if ((0 == access(siname, 4)) && (0 == stat(siname, &sibuf)) &&
		(sibuf.st_size > 10))
	  {
		spi->sc_flags |= SC_INPUTSCALE; spi->sc_flags &= ~SC_ALLOBS;
	  }
	else {
		spi->sc_flags |= SC_ALLOBS; spi->sc_flags &= ~SC_INPUTSCALE;
	     }
	/* no Wilson stats if input .si file */
	if (spi->sc_flags & SC_INPUTSCALE) spi->sc_flags &= ~SC_WILSON;
	return EOK;
}

static int do_makemu(MAKEMURUN *mmup)
{ /* This does the equivalent of makemu within scalem */
	int	retv;

	mmup->mm_tl0 = NOW;
	if (EOK != (retv = mufiles(mmup))) return retv;
	if ((mmup->mm_flags & MUF_ABSORDER) && (mmup->mm_flags & MUF_STLORDER))
		mmup->mm_flags &= ~(MUF_ABSORDER);
	if ((mmup->mm_flags & MUF_REINDEX) && !try_reindex(mmup->mm_vecrei,
		mmup->mm_matrei)) mmup->mm_flags -= MUF_REINDEX;
	if (mmup->mm_oversc < FUZZ || mmup->mm_oversc > 10000.)
		mmup->mm_oversc = 1.;
	mmup->mm_oversq = mmup->mm_oversc * mmup->mm_oversc;
	return makemuops(mmup);
}

static SCMWORK *managescm(int opno)
{
	static SCMWORK	*scm;
	static char	scmsg[] = "scm in managescm";
	static char	spmsg[] = "scm->sm_spi in managescm";
	static char	stmsg[] = "scm->sm_star in managescm";
	static char	mmumsg[] = "scm->sm_mmup in managescm";

	if (!opno) /* delete everything */
	  {
		if (NULL != scm)
		  {
			scm = managescm(074); memzap((AVOID **)&scm, scmsg);
		  }
		return NULL;
	  }
	if ((!(opno & 01)) && (NULL != scm)) /* selective deletions */
	  {
		if (opno & 04) memzap((AVOID **)&(scm->sm_spi), spmsg);
		if (opno & 020) memzap((AVOID **)&(scm->sm_star), stmsg);
		if (opno & 040) memzap((AVOID **)&(scm->sm_mmup), mmumsg);
		return scm;
	  }
	if (opno == 1) scm = managescm(0);
	if (opno == 3) return scm;
	if ((NULL == scm) && (NULL == (scm = callop(1, sizeof (SCMWORK),
	 scmsg)))) return scm;
	if ((NULL == scm->sm_spi) && (NULL == (scm->sm_spi = callop(1,
	 sizeof (SCALERUN), spmsg)))) return managescm(0);
	if ((NULL == scm->sm_star) && (NULL == (scm->sm_star = callop(1,
	 sizeof (STATRUN), stmsg)))) return managescm(0);
	if ((NULL == scm->sm_mmup) && (NULL == (scm->sm_mmup = callop(1,
	 sizeof (MAKEMURUN), mmumsg)))) return managescm(0);
	return scm;
}

static int smend(int statu)
{ /* smooth ending of a run of scalem */
	int	lostat;
	SCMWORK	*scm;

	if (NULL == (scm = managescm(3))) return statu;
	lostat = 0;
	if (NULL != scm->sm_spi) lostat = scend(lostat, scm->sm_spi);
	if (NULL != scm->sm_rjp) lostat = rejdie(lostat);
	if (NULL != scm->sm_star) lostat = endstat(lostat, scm->sm_star);
	if (NULL != scm->sm_mmup) lostat = endmu(lostat, scm->sm_mmup);
	scm = managescm(0);	return statu;
}

int scalemmain(int argc, char *argv[])
{ /*	Mainline for the all-in-1 scaling, rejection, output functionality */
	int			i, iv, jq;
	double			locredun;
	static SCMWORK		*scm;
	static SCALERUN		*spi;
	static REJECTRUN	*rjp;
	static STATRUN		*star;
	static MAKEMURUN	*mmup;

	/*process args,defaults*/
	if (NULL == (scm = managescm(7))) return EALLOC;
	spi = scm->sm_spi;	rjp = scm->sm_rjp;
	star = scm->sm_star;	mmup = scm->sm_mmup;
	if (EOK != (i = scmargs(argc, argv, scm))) return smend(i);
	for (jq = 0; jq < 4; jq++)
	   {
		if (EOK != (iv = precycle(spi))) /* get ready to scale */
		  {
			fprintf(stderr, "Error opening multiref on cycle %d\n",
			 jq);
			break;
		  }
		if (EOK != (iv = sccycles(spi))) /* do the scaling */
		  {
			fprintf(stderr, "sccycles returned %d on cycle %d\n",
				iv, jq);
			break;
		  }
		if (EOK != (iv = scmend(i, spi))) /* compute <I>,sigma,R */
		  {
			fprintf(stderr, "scmend returned %d on cycle %d\n",
				iv, jq);
			break;
		  }
		/* FIXME: if we ever decide to try to reconstruct the scaling
		 database, we'd do it here. */
			/* spherical harmonics */
		spi->sc_flags |= SC_DIFFRACTION | SC_DIFF_NOS2;
		if (jq >= 1) spi->sc_flags |= SC_SWITCH;
		if (jq >= 2)
		  {
			locredun = ((double)(spi->sc_nobput)) /
				((double)(spi->sc_nrefput));
			if (locredun < scm->sm_minredun)
			  {
				printf(
 " No final scaling with Bijvoets unmerged: redundancy = %8.2f < %8.2f\n",
				 locredun, scm->sm_minredun);
				break;
			  }
			spi->sc_flags &= ~SC_MERGEBJ;
		  }
		if (jq < 2) (void)autoreject();
	   }
	(void)xgprogname("stats", 4);
	/* we separate Bijvoets in `stats' iff. we did so in scalek */
	if (jq >= 3) star->sta_flags |= ST_SEPBJ;
	if (EOK != (iv = statstart(star)))
	  {
		fprintf(stderr, "statistics setup failed: error %d\n", iv);
		return smend(iv);
	  }
	if (EOK != (iv = statops(star)))
	  {
		fprintf(stderr, "statistics output failed: error %d\n", iv);
		return smend(iv);
	  }
	(void)xgprogname("spaceg", 4);
	zeromem((void *)star, sizeof (STATRUN));
	if (EOK != (iv = spacegstart(star)))
	  {
		fprintf(stderr, "spacegroup setup failed: error %d\n", iv);
		return smend(iv);
	  }
	if (EOK != (iv = spacegops(star)))
	  {
		fprintf(stderr, "spacegroup calc failed: error %d\n", iv);
		return smend(iv);
	  }
	(void)xgprogname("makemu", 4);
	if (EOK != (iv = do_makemu(mmup)))
	  {
		fprintf(stderr, "makemu calculation failed: error %d\n", iv);
		return smend(iv);
	  }
	return smend(EOK);
}

/* end of scalemm.c */
