/* sccalc.c -- calculational routines for XGEN's scaling functionality

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	21 Apr 2007: get_postr sped up by keeping unchanged info
	19 Apr 2007: float->double conversions; reorganize use of "{"
	24 Jun 2006: include xgpostint.h rather than scdefs.h
	 5 Jun 2006: statsh,sum_wilson,wilscale moved to scrio
	 9 Dec 2005: insert cycle_summary and cycle_quicksum from scsum.c
		and stat_sh, sum_wilson, and wilscale from scwilson.c
	20 Nov 2005: inline argument declarations and strict prototyping
	19 Oct 2005: replace static globals with more SCALERUN structure elts
	10 Oct 2005: housekeeping
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	17 Sep 2004: use additional SCALERUN elements for cycle summaries;
	 output RMSdiffs and max diffs even for nonlinear scaling;
	 call cycle_quicksum to give one-line summaries of scaling cycles;
	 cycle_summary() doesn't need separate arg for cycle #;
	 add cycle_quicksum() to give a one-line summary of a cycle
	28 Aug 2004: routines used only here are declared as static; cosmetics
	25 Jan 2004: a few static globals converted to locals (not enough!)
	21 Jan 2004: license message inserted and initial comment reordered
	 7 Jan 2004: SCALERUN structure elements in function args
		replace extern globals; args updated accordingly
	13 Dec 2003: eliminate unused variables t1q, t2q from harmonicmem()
	14 Nov 2003: replaced all globals from scmain.c with components of
		the SCALERUN structure
	10 Nov 2003; fix bug in calculation of sbar, sbarsq
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	 5 Jan 2003: reorganized source-file header
	15 Nov 2002: extern declarations made here rather than in scext.h;
		ndiffv_groups renamed to ndv_groups
	10 Nov 2002: compute and report RMS deviations between scale factors
		in any cycle relative to those in the previous cycle
	 1 Nov 2002: wilscale: deal properly with merged detector top&bottom
	20 May 2002: (?) in each cycle, remember the best R-merge among the
		cycles so far; the parameters found at that point
		will be the ones that we output
	15 Apr 2002: (?) in several routines,
		modify calls related to shlp[] since it's handled differently
	25 Mar 2002: removed MSI attachments
	31 May 1995: "register" -> "register int" throughout
	 2 Mar 1995: diagnostics to stdout
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	14 Nov 1994: add Simplex scaling and conditional-shift code;
		provide for better diagnostics during linear & Simplex scaling
	 4 Nov 1994: add SC_FLATVARIANCE option so sigmas independ of R-factor
		of scalegroup; otherwise variances augmented by (I*sigma(g))^2
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version
	16 Mar 1992: in y_and_gsums, in calculating the scale factors,
		perform linear interpolation between groups when we can.
	 9 Jan 1991: in diffv_ref, we upweight high-resolution data
		for the difference-vector style of scaling.
	
   Purpose: calculations for scaling. Each cycle's least-squares calculations
	(sometimes including spherical-harmonic stuff) are done in here.
   Cross references: scrio.c, scmain.c
	The scaling values obtained from the Wilson calculations wind up
	being related in some less-than-obvious ways to the parameters that
	DO get used in symmetry scaling.  This stuff needs to be re-examined!!

*************************************************************************/
#define		NEED_TIME	1
#include	"xgpostint.h"
#define		MINRELY		4	/* min reliable # obs/shift */
#define		LIMITRAT	7	/* see its use */
#define		INITDEL		0.3	/* see its use */
#define		DIVR		(0.2)	/* how much to shrink delta by */

static void badshift(CYCLESHIFT *tg, double detval, SCALERUN *spi)
{ /*	Take note of a faulty scaling group */
	register int	shiftno;
	register SHIFT	*tsh;

	shiftno = tg - spi->sc_g;
	tsh = spi->sc_shlp + shiftno;
	fprintf(spi->sc_fpl,
 "Shift,Chmbr,Run %5d,%2d,%2d: %6d obs,sumisq,sum(y*s)sq = %10.3e, %10.3e\n",
	 tsh->shval, tsh->chamno, tsh->runno, tg->nobsh, tg->i2, tg->y2s2);
	if (detval > -10000.) fprintf(spi->sc_fpl, "%10.3e\n", detval);
	else	fprintf(spi->sc_fpl,"\n");
	tg->newg = 1;	tg->rfac = tg->newa = tg->newb = 0.;
	return;
}

static void clear_shift(CYCLESHIFT *tg, SCALERUN *spi)
{ /*	This clears all the sums in the CYCLESHIFT structure */
	tg->i2 = tg->y2 = tg->iy = tg->nobsh = 0;
	if (spi->sc_nparpshift > 1)
	  {
		tg->iys2 = tg->y2s2 = tg->y2s4 = 0.;
		if (spi->sc_nparpshift > 2)
		  {
			tg->y2s = tg->y2s3 = tg->iys = 0.;
		  }
	  }
}

static void y_and_gsums(QREF *rfl, QOB *obl, SCALERUN *spi)
{ /*	This increments the appropriate sums for the <I> and
	scale factor calculations */
	register int		i, nob;
	register QOB		*obp, *obend;
	register CYCLESHIFT	*tg;
	SHIFT			*tsh;
	short			*dvp;
	double			gisq, gw, iij, iij2, res, rescube, resquad;
	double			ressq, thisg, thisii, wij, wiij2, wimix;
	double			wiyj2, yij, yij2, yj, yjden;
	float			*fp1;
	VECPAR			*vpp;

	res = rfl->qr_stl;	ressq = res * res;
	/* Compute the mean intensity, Yj = <Iij> = (SUM Iij * wi) / (SUM wij).
	Note that the weight associated with each observation in computing
	Yj is not just the inverse of the variance of the intensity; instead,
	it's a function the error in gi as well as the error in Iij. */
	yj = yjden = 0.;	/* clear the <I> & denominator sums */
	obend = obl + rfl->qr_nob;
	for (obp = obl, nob = 0; obp < obend; obp++, nob++)
	   {  /* do linear interpolation when we can*/
		vpp = spi->sc_vshp + obp->qo_dvgrp;
		tsh = spi->sc_shlp + obp->qo_shift;
		obp->qo_lscal = thisg = terpscale(obp->qo_omob, tsh, res);
	/* When performing vector scaling, multiply by the previous cycle's
	 calculation of the diffraction-vector stuff */
		if (spi->sc_flags & SC_VECREFINED)
		  {
			for (i = spi->sc_noutpar, gw = 0.,
			 fp1 = &(vpp->v_coef[0]), dvp = obp->qo_sc; i--;
			 fp1++, dvp++) gw += *fp1 *(DVCSCALE *((double)(*dvp)));
			thisg *= (obp->qo_lharm = gw);
			if (gw < spi->sc_minqf)
			  {
				spi->sc_minqf = gw;
				for (i = 0; i < 10; i++)
				   {
					spi->sc_minco[i] = vpp->v_coef[i];
					spi->sc_minp[i] = obp->qo_sc[i];
				   }
			  }
			if (gw > spi->sc_maxqf)
			  {
				spi->sc_maxqf = gw;
				for (i = 0; i < 10; i++)
				   {
					spi->sc_maxco[i] = vpp->v_coef[i];
					spi->sc_maxp[i] = obp->qo_sc[i];
				   }
			  }
		  }
		else	obp->qo_lharm = gw = 1.;
		gisq = thisg * thisg;
		thisii = obp->qo_inten;
		wij = (double)obp->qo_sigma;	wij *= wij;
		if ((!(spi->sc_flags & SC_FLATVARIANCE)) && (gisq > DFUZZ))
			wij += tsh->shvar * thisii * thisii / gisq;
		obp->qo_iwt = wij = 1. / wij;
		yj += wij * thisii * thisg;
		yjden += wij * gisq;
		if (spi->sc_flags & SC_VECREFINED)
		  {
			spi->sc_meanqf += gw * wij;
			spi->sc_qfden += wij;
		  }
	   }
	if (yjden < FUZZ || yj < FUZZ) return;
	(spi->sc_nrefput)++;
	/* get new g values and new r factor sums */
	yj /= yjden;	/* this is average intensity*/
	if (yj < spi->sc_imin) return; /* skip super-weak <I> */
	rescube = res * ressq;	resquad = ressq * ressq;
	for (obp = obl, nob = 0; obp < obend; obp++, nob++)
	   {
		tg = spi->sc_g + obp->qo_shift;
		wij = obp->qo_iwt;	iij = obp->qo_inten;
		iij2 = iij * iij;	wiij2 = wij * iij2;
		yij = yj * obp->qo_lharm;
		yij2 = yij * yij;	wiyj2 = wij * yij2;
		wimix = wij * iij * yij;
		tg->nobsh++;	(spi->sc_nobput)++;
		tg->y2		+= wiyj2;
		tg->iy		+= wimix;
		tg->i2		+= wiij2;
		if (spi->sc_nparpshift > 1)
		  {
			tg->y2s2	+= wiyj2 * ressq;
			tg->y2s4	+= wiyj2 * resquad;
			tg->iys2	+= wimix * ressq;
			if (spi->sc_nparpshift > 2)
			  {
				tg->y2s	+= wiyj2 * res;
				tg->y2s3+= wiyj2 * rescube;
				tg->iys += wimix * res;
			  } /* end of innermost "if " clause*/
		  } /* end of if (spi->sc_nparpshift > 1 clause*/
	   } /* end of for (obp ... loop */
}

static void diffv_ref(QREF *rfl, QOB *obl, SCALERUN *spi)
{ /*	Once we're through with the conventional scaling, we pass through
	another sub-cycle in which we calculate new values of the
	diffraction-vector-dependent components of the mess.  */
	int		nob, j;
	register QOB	*obp, *obend;
	SHIFT		*tsh;
	short		*dvp;
	double		bval, gisq, gw, nweight, res, ressq;
	double		thisg, thisii, wij, yj, yjden, **rdp, **ydp;
	float		*fp1;
	VECPAR		*vpp;

	/* First step: compute the mean intensity, as calculated
	 with the current scale factors.  */
	obend = obl + rfl->qr_nob;	yj = yjden = 0.;
	res = rfl->qr_stl;	ressq = res * res;
	for (obp = obl, nob = 0; obp < obend; obp++, nob++)
	   {
		tsh = spi->sc_shlp + obp->qo_shift;
		vpp = spi->sc_vshp + obp->qo_dvgrp;
		obp->qo_lscal = thisg = terpscale(obp->qo_omob, tsh, res);
		/* old version
		obp->qo_lscal = thisg =
			tsh->shg + tsh->sha * res + tsh->shb * ressq; */
		for (j = spi->sc_noutpar, gw = 0, fp1 = &(vpp->v_coef[0]),
		 dvp = obp->qo_sc; j--; fp1++, dvp++)
			gw += *fp1 * (DVCSCALE * ((double)(*dvp)));
		obp->qo_lharm = gw;
		thisg *= gw;
		gisq = thisg * thisg;	thisii = obp->qo_inten;
		wij = (double)obp->qo_sigma;	wij *= wij;
		if ((!(spi->sc_flags & SC_FLATVARIANCE)) && (gisq > FUZZ))
			wij += tsh->shvar * thisii * thisii / gisq;
		wij = 1. / wij;
		/* for this type of scaling, upweight high-resolution data. */
		if (spi->sc_noutpar == 10)
			wij *= 1. + DVCSCALE * 0.5 * (double)(obp->qo_sc[9]);
		obp->qo_iwt = wij;
		yj += wij * thisii * thisg;	yjden += wij * gisq;
	   }
	if (yjden < FUZZ) return;
	yj /= yjden;
/* Now use those means in the calculations leading to the new parameters */
	for (obp = obl, nob = 0; obp < obend; obp++, nob++)
	   {
		tsh = spi->sc_shlp + obp->qo_shift;
		obp->qo_lscal = thisg = terpscale(obp->qo_omob, tsh, res);
	/*	obp->qo_lscal = thisg =
			tsh->shg + tsh->sha * res + tsh->shb * ressq; */
		rdp = spi->sc_rb + obp->qo_dvgrp;
		ydp = spi->sc_yb + obp->qo_dvgrp;
		nweight = obp->qo_iwt * yj * thisg;
		for (j = 0, dvp = obp->qo_sc; j < (int)spi->sc_noutpar;
		 j++, dvp++)
		   spi->sc_avec[j] = nweight * (DVCSCALE * ((double)(*dvp)));
		bval = obp->qo_iwt * obp->qo_inten;
		inv_add(*rdp, *ydp,
			spi->sc_avec, bval, (int)(spi->sc_noutpar));
	   }
}

static void calc1par(CYCLESHIFT *tg, SCALERUN *spi)
{ /*	This does the shift calculations for the 1-parameter case */

	if (tg->y2 > FUZZ && tg->i2 > FUZZ && tg->nobsh > 1)
	  {
		tg->newg = tg->iy / tg->y2;
		tg->rfac = tg->i2 - tg->iy * tg->iy / tg->y2;
		spi->sc_ovrf += tg->rfac;	spi->sc_ovrfden += tg->i2;
		tg->mvar[0] = tg->rfac / ((tg->nobsh - 1) * tg->y2);
		tg->rfac /= tg->i2;	/* R-factor */
		spi->sc_gsum += tg->nobsh * tg->newg;
		spi->sc_gsumden += tg->nobsh;
	  }
	else {
		tg->newg = tg->mvar[0] = 0.; tg->rfac = 1.; }
}

static void calc2par(CYCLESHIFT *tg, SCALERUN *spi)
{ /*	This does the shift calcs for the 2-parameter case */
	double	detnor, gof;

	detnor = tg->y2 * tg->y2s4 - tg->y2s2 * tg->y2s2;
	if ((detnor <= FUZZ) || (tg->i2 <= FUZZ) || (tg->nobsh <= 2))
	  {
		tg->newg = tg->newb = 0.;	tg->rfac = 1.;
		tg->mvar[0] = tg->mvar[1] = tg->mvar[2] = 0.;	return;
	  }
	tg->newg = (tg->y2s4 * tg->iy - tg->y2s2 * tg->iys2) / detnor;
	tg->newb = (tg->y2 * tg->iys2 - tg->y2s2* tg->iy ) / detnor;
	tg->rfac = tg->i2 - 2. * tg->newg * tg->iy - 2 * tg->newb *
	 tg->iys2 + tg->newg * tg->newg * tg->y2 + tg->newb *
	 tg->newb * tg->y2s4 + 2. * tg->newg * tg->newb * tg->y2s2;
	spi->sc_ovrf += tg->rfac;	spi->sc_ovrfden += tg->i2;
	gof = tg->rfac / ((tg->nobsh - 2) * detnor);
	spi->sc_gsum += tg->nobsh * (tg->newg + tg->newb * spi->sc_barsq);
	spi->sc_gsumden += tg->nobsh;	tg->rfac /= tg->i2;
	tg->mvar[0] = gof * tg->y2s4; tg->mvar[1] = -gof * tg->y2s2;
	tg->mvar[2] = gof * tg->y2;
}

static void calc3par(CYCLESHIFT *tg, SCALERUN *spi)
{ /*	This takes a pointer to a CYCLESHIFT structure and constructs
	the scale factor vector (tg->shg, tg->sha, tg->shb) and the error
	matrix for it in the case where all 3 parameters are being used.*/
	register int	i,j;
	double		a00, a01, a12, a22; /* uninverted normal matrix elts*/
	double		detrescale, detval, gof, renorm, emat[3][3];

	/* in inverting the normal matrix, we renormalize such that the upper
	righthand element is 1 ... it will be roughly the geometric mean of the
	other four elements, since they all are related to sin(theta)/lambda.
	Thus the determinant of the renormalized matrix will be close to one.*/
	if (tg->nobsh <= 3 || ((renorm = tg->y2s2) < FUZZ) || tg->i2 < FUZZ)
	  {
		badshift(tg, -11000., spi);	return;
	  }
	a00 = tg->y2 / renorm;		a01 = tg->y2s / renorm;
	a12 = tg->y2s3 / renorm;	a22 = tg->y2s4 / renorm;
	/* The formulas for the inverse matrix reflect the fact that the
	input upper righthand element is one. */
	emat[0][0] = a22 - a12 * a12;
	emat[1][0] = emat[0][1] = a12 - a01 * a22;
	emat[2][0] = emat[0][2] = a01 * a12 - 1.;
	emat[1][1] = a00 * a22 - 1.;
	emat[2][1] = emat[1][2] = a01 - a12 * a00;
	emat[2][2] = a00 - a01 * a01;
	detval = a00 * emat[0][0] + a01 * emat[1][0] + emat[2][0];
	if (detval < FUZZ)	/* pathological case */
	  {
		badshift(tg, detval, spi);	return;
	  }
	detrescale = detval * renorm;	/* rescaled determinant */
	/* Construct the actual inverse of the normal matrix */
	for (i = 3; i--; ) for (j = 3; j--; ) emat[i][j] /= detrescale;
	/* Get the updated values of the scale coefficients */
	tg->newg = emat[0][0] * tg->iy + emat[0][1] * tg->iys +
		emat[0][2] * tg->iys2;
	tg->newa = emat[1][0] * tg->iy + emat[1][1] * tg->iys +
		emat[1][2] * tg->iys2;
	tg->newb = emat[2][0] * tg->iy + emat[2][1] * tg->iys +
		emat[2][2] * tg->iys2;
	/* compute the numerator of the new R-factor for the shift */
	tg->rfac = tg->i2 - 2. * tg->newg * tg->iy -
		2. * tg->newa * tg->iys - 2. * tg->newb * tg->iys2 +
		tg->newg * tg->newg * tg->y2 + tg->newa * tg->newa * tg->y2s2 +
		tg->newb * tg->newb * tg->y2s4 +
		2. * tg->newg * tg->newa * tg->y2s +
		2. * tg->newg * tg->newb * tg->y2s2 +
		2. * tg->newa * tg->newb * tg->y2s3;
	/* increment sums for overall R-factors and scale factor renormalizers*/
	spi->sc_ovrf += tg->rfac;	spi->sc_ovrfden += tg->i2;
	spi->sc_gsum += tg->nobsh * (tg->newg + tg->newa * spi->sc_bar +
		tg->newb * spi->sc_barsq);
	spi->sc_gsumden += tg->nobsh;
	gof = tg->rfac / (tg->nobsh - 3); /* to be used for variances */
	/* Compute the actual R-factor for the shift */
	tg->rfac /= tg->i2;
	/* Compute the variances and covariances among in g, a, and b */
	tg->mvar[0] = gof * emat[0][0];	tg->mvar[1] = gof * emat[0][1];
	tg->mvar[2] = gof * emat[0][2];	tg->mvar[3] = gof * emat[1][1];
	tg->mvar[4] = gof * emat[1][2];	tg->mvar[5] = gof * emat[2][2];
}

static int altered(int changed, SCALERUN *spi, QOB *obp)
{ /* tells us which parts of the observation need recalculation.
	It returns 0 if nothing needs recalculation;
	It returns 1 if the linear scale factor(s) need recalculation;
	It returns 2 if the spherical-harmonic stuff needs redoing;
	It returns 3 if they both do. */
	short	qshift, tom;
	int	idel;
	SHIFT	*tsh;

	if (changed < 0) return (spi->sc_flags & SC_VECREFINED) ? 3 : 1;
	/* we could get more careful here, but it's probably unimportant. */
	if (changed >= spi->sc_nscales) return 2;
	if (changed == (qshift = obp->qo_shift)) return 1;
	tom = obp->qo_omob;	tsh = spi->sc_shlp + obp->qo_shift;
	return (changed == qshift + (idel = terpalt(tom, tsh)));
}

static void get_postr(double *pvec, SCALERUN *spi, int changed)
{ /*	This calculates the overall R-value for the entire dataset
	 after the calculation of the new scaling functions on any cycle.
	 It also returns an unweighted rvalue in *(pvec+1) */

	register int	i, mcha, nob;
	register QOB	*obp, *obend;
	short		*dvp;
	double		gden, gisq, gnum, gw, iij, iij2;
	double		re0, re1, re2, re3, re4, res, thisg, thisii, wij;
	double		wiij2, yj, yjden, yscj;
	float		*fp1;
	QOB		*obl;
	QREF		*rfl;
	SHIFT		*tsh;
	VECPAR		*vpp;

	re0 = re1 = re2 = re3 = re4 = gnum = gden = 0.;
	spi->sc_nobput = spi->sc_nrefput = 0;
	for (rfl = spi->sc_refbase, obl = spi->sc_obbase; (nob =rfl->qr_nob);
	 rfl++, obl +=nob)
	   { /* clear the <I> & denominator sums */
		yj = yjden = 0.;	res = rfl->qr_stl;
		obend = obl + rfl->qr_nob;
		for (obp = obl, nob = 0; obp < obend; obp++, nob++)
		   {
			tsh = spi->sc_shlp + obp->qo_shift;
			if (1 & (mcha = altered(changed, spi, obp)))
				obp->qo_lscal = thisg =
					terpscale(obp->qo_omob, tsh, res);
			else	thisg = obp->qo_lscal;
			if (!(spi->sc_flags & SC_VECREFINED))
				obp->qo_lharm = gw = 1.;
			if (2 & mcha)
			  {
				/* When performing vector scaling, multiply by
			 	previous cycle's diffraction-vector stuff */
				vpp = spi->sc_vshp + obp->qo_dvgrp;
				for (i = spi->sc_noutpar, gw = 0.,
				 fp1 = &(vpp->v_coef[0]), dvp = obp->qo_sc;
				 i--; fp1++, dvp++)
				   gw += *fp1 * (DVCSCALE * ((double)(*dvp)));
				obp->qo_lharm = gw;
			  }
			else gw = obp->qo_lharm;
			thisg *= gw;
			gisq = thisg * thisg; thisii = obp->qo_inten;
			wij = (double)obp->qo_sigma;	wij *= wij;
			if ((!(spi->sc_flags & SC_FLATVARIANCE)) &&
			 (gisq > FUZZ))
				wij += tsh->shvar * thisii * thisii / gisq;
			obp->qo_iwt = wij = 1. / wij;
			gnum += thisg * wij;	gden += wij;
			yj += wij * thisii * thisg;
			yjden += wij * gisq;
		   }
		if (yjden > FUZZ)
		  {
			yj /= yjden;	/* this is average intensity*/
			(spi->sc_nrefput)++;
			for (obp = obl, nob = 0; obp < obend; obp++, nob++)
			   {
				wij = obp->qo_iwt;
				iij = obp->qo_inten;
				iij2 = iij * iij;	wiij2 = wij * iij2;
				yscj = obp->qo_lscal * obp->qo_lharm * yj;
				re0 += wiij2;	re1 += wij * iij * yscj;
				re2 += wij * yscj * yscj;
				re3 += fabs(iij - yscj); re4 += fabs(iij);
				(spi->sc_nobput)++;
			   }
		  }
	   }
	/* put results into the four-element vector `pvec'. */
	if (re0 < FUZZ) *pvec = 1.;
	else {
		re1 = (re0 + re2 - 2. * re1) / re0;
		*pvec = (re1 < FUZZ) ? 0. : re1;	
	     }
	*(pvec+1) = (re4 < FUZZ) ? 1. : re3 / re4;
	*(pvec+2) = gnum;	*(pvec + 3) = gden;
	return;
}

static void nlrescale(SCALERUN *spi)
{ /*	This divides all the scaling-group parameters through by
	the mean scale factor of the entire dataset, effectively normalizing */
	register int	i;
	SHIFT		*shp;

	if ((spi->sc_gsca < FUZZ) ||(fabs(spi->sc_gsca - 1.) < FUZZ)) return;
	for (shp = spi->sc_shlp, i = 0; i < spi->sc_nscales; i++, shp++)
		shp->shg /= spi->sc_gsca;
	if (spi->sc_nparpshift > 1)
	  {
		for (shp = spi->sc_shlp, i = 0; i < spi->sc_nscales;
		 i++, shp++) shp->shb /= spi->sc_gsca;
		if (spi->sc_nparpshift > 2)
		  {
			for (shp = spi->sc_shlp, i = 0; i < spi->sc_nscales;
			 i++, shp++) shp->sha /= spi->sc_gsca;
		  }
	  }
}

static void partoscale(double *xp0, SCALERUN *spi)
{ /*	This converts from an amorphous parameter list to a bunch of
	actual scaling parameters */
	int	i, j;
	double	*xp;
	SHIFT	*shp;
	VECPAR	*vpp;
	
	for (i = 0, shp = spi->sc_shlp, xp = xp0;
	 i < spi->sc_nscales; i++, shp++)
	   {
		shp->shg = *xp++;
		if (spi->sc_nparpshift > 1)
		  {
			if (spi->sc_nparpshift > 2) shp->sha = *xp++;
			shp->shb = *xp++;
		  }
	   }
	if (spi->sc_flags & SC_VECREFINED)
	  {
		for (i = 0, vpp = spi->sc_vshp; i < (int)(spi->sc_ndvgroups);
		 i++, vpp++)
		   {
			for (j = 0; j < (int)(spi->sc_noutpar); j++)
				vpp->v_coef[j] = *xp++;
		   }
	  }
	nlrescale(spi);	/* renormalize the scale factors */
}

static double scresid(SCALERUN *spi, int changed)
{ /*	This returns the weighted RMS R-value for the current set
	of scaling parameters. There's not much to it: we convert from the
	amorphous spi->sc_new0 array back to the actual scaling parameters,
	do a residual calculation using standard methods, and return
	the scaling residual we get.  */
	register int	i;
	register double	*bp, *xp;
	double		*xp0, lr3, retv;

	xp0 = spi->sc_new0;
	partoscale(xp0, spi); /* convert from param list to scaling params */
	(spi->sc_cycno)++;	/* increment count of cycles */
	get_postr(spi->sc_pout, spi, changed);	/* get residual */
	retv = spi->sc_pout[0];	lr3 = spi->sc_pout[1];
	if (spi->sc_cycno == 1)	/* record initial residuals */
	  {
		spi->sc_r3 = lr3;
		for (i = 0, xp = xp0, bp = spi->sc_s0;
		 i < (int)(spi->sc_totpar); i++) *bp++ = *xp++;
	  }
	/* if this is the best so far, record it */
	if (retv < spi->sc_r1)
	  {
		for (i = 0, xp = xp0, bp = spi->sc_best0;
		 i < (int)(spi->sc_totpar); i++) *bp++ = *xp++;
	    	spi->sc_r1 = retv;	spi->sc_r3 = lr3;
		if (spi->sc_pout[3] > FUZZ)
			spi->sc_gsca = spi->sc_pout[2] / spi->sc_pout[3];
			/* remember what cycle we got the "best" in */
		spi->sc_nkbest = spi->sc_nk;
	  }
	return retv;
}

static void allrvals(double *pvec, SCALERUN *spi)
{ /*	This computes the R-factors by scaling group within a dataset
	for the current scaling functions. It's essentially equivalent to
	`get_postr()', except that it does some additional calculations and
	is therefore a bit slower. */

	register int	i, nob;
	register QOB	*obp, *obend;
	QOB		*obl;
	QREF		*rfl;
	SHIFT		*tsh;
	short		*dvp;
	double		gden, gisq, gnum, gw, iij, iij2;
	double		re0, re1, re2, re3, re4, res, thisg, thisii, wij;
	double		wiij2, yj, yjden, yscj;
	float		*fp1;
	VECPAR		*vpp;
	CYCLESHIFT	*tg;

	if ((NULL == spi->sc_g) && (NULL == (spi->sc_g =
	 (CYCLESHIFT *)callop((size_t)(spi->sc_nscales),
	 sizeof (CYCLESHIFT), "spi->sc_g in allrvals")))) return;
	for (i = 0, tg = spi->sc_g, tsh = spi->sc_shlp;
	 i < spi->sc_nscales; i++, tg++, tsh++)
	   {
		zeromem((void *)tg, sizeof (CYCLESHIFT)); tg->newg = tsh->shg;
		tg->newa = tsh->sha;	tg->newb = tsh->shb;
	   }
	re0 = re1 = re2 = re3 = re4 = gnum = gden = 0.;
	spi->sc_nobput = spi->sc_nrefput = 0;
	for (rfl = spi->sc_refbase, obl = spi->sc_obbase;
	 (nob =rfl->qr_nob); rfl++, obl +=nob)
	   { /* clear the <I> & denominator sums */
		yj = yjden = 0.;	res = rfl->qr_stl;
		obend = obl + rfl->qr_nob;
		for (obp = obl, nob = 0; obp < obend; obp++, nob++)
		   {
			vpp = spi->sc_vshp + obp->qo_dvgrp;
			tsh = spi->sc_shlp + obp->qo_shift;
			obp->qo_lscal = thisg =
				terpscale(obp->qo_omob, tsh, res);
			/* When performing vector scaling, multiply by the
			 previous cycle's calc. of diffraction-vector stuff */
			if (spi->sc_flags & SC_VECREFINED)
			  {
				for (i = spi->sc_noutpar, gw = 0.,
				 fp1 = &(vpp->v_coef[0]), dvp = obp->qo_sc;
				 i--; fp1++, dvp++)
				   gw += *fp1 * DVCSCALE * (double)(*dvp);
				obp->qo_lharm = gw;
				thisg *= gw;
			  }
			else	obp->qo_lharm = gw = 1.;
			gisq = thisg * thisg;
			thisii = obp->qo_inten;
			wij = (double)obp->qo_sigma;	wij *= wij;
			if ((!(spi->sc_flags & SC_FLATVARIANCE)) &&
			 (gisq > FUZZ))
				wij += tsh->shvar * thisii * thisii / gisq;
			obp->qo_iwt = wij = 1. / wij;
			gnum += thisg * wij;	gden += wij;
			yj += wij * thisii * thisg;
			yjden += wij * gisq;
		   }
		if (yjden > DFUZZ)
		  {
			yj /= yjden;	/* this is average intensity*/
			spi->sc_nrefput++;
			for (obp = obl, nob = 0; obp < obend; obp++, nob++)
			   {
				wij = obp->qo_iwt;
				iij = obp->qo_inten;
				iij2 = iij * iij;	wiij2 = wij * iij2;
				yscj = obp->qo_lscal * obp->qo_lharm * yj;
				re0 += wiij2;	re1 += wij * iij * yscj;
				re2 += wij * yscj * yscj;
				re3 += fabs(iij - yscj); re4 += fabs(iij);
				tg = spi->sc_g + obp->qo_shift;
				(tg->nobsh)++;
				tg->i2 += wiij2; tg->iy += wij * iij * yscj;
				tg->y2 += wij * yscj * yscj; (spi->sc_nobput)++;
			   }
		  }
	   }
	for (tg = spi->sc_g, tsh = spi->sc_shlp, i = 0; i < spi->sc_nscales;
	 i++, tg++,tsh++)
	   {
		if ((tg->nobsh) && (DFUZZ < tg->i2))
			tg->rfac = (tg->i2 + tg->y2 - 2. * tg->iy) / tg->i2;
		else	tg->rfac = 1.;
		if (tg->nobsh > 1) tg->mvar[0] = tsh->shvar =
			tg->rfac / ((double)(tg->nobsh - 1));
	   }
	/* put results into the four-element vector `pvec'. */
	if (re0 < DFUZZ) *pvec = 1.;
	else {
		re1 = (re0 + re2 - 2. * re1) / re0;
		*pvec = (re1 < DFUZZ) ? 0. : re1;	
	     }
	*(pvec+1) = (re4 < DFUZZ) ? 1. : re3 / re4;
	*(pvec+2) = gnum;	*(pvec + 3) = gden;
	return;
}

static void compare_nl(SCALERUN *spi)
{ /*	This compares the current parameters with the ones we found on the 
	previous batch of nonlinear cycles. It then sets parameters for the
	"previous" grouping to the current values so we can do this again. */
	size_t	i;
	double	lodn, lomax, tmax, *xpc, *xpp;

	spi->sc_maxdi = spi->sc_rmsde = lodn = lomax = 0.;
	if (spi->sc_totpar < 1) return;
	for (i = 0, xpp = spi->sc_prev0, xpc = spi->sc_best0;
	 i < spi->sc_totpar; i++, xpp++, xpc++)
	   {
		if ((tmax = fabs(*xpp - *xpc)) > lomax) lomax = tmax;
		lodn += tmax * tmax;	*xpp = *xpc;
	   }
	lodn = sqrt(((double)lodn) / ((double)(spi->sc_totpar)));
	spi->sc_rmsde = lodn;	spi->sc_maxdi = lomax;
}

static void cycle_quicksum(FILE *fp, SCALERUN *spi)
{ /*	Quickly summarize a cycle of scaling to the stream fp */
	fprintf(fp,
	 " %4d %7d %7d %4d %8.5f %8.5f %7.5f %7.5f %8.4f %9.3e\n",
		(int)(spi->sc_cycno), (int)spi->sc_nobput,
		(int)spi->sc_nrefput, (int)(spi->sc_tl1 - spi->sc_tl0),
		spi->sc_gsca, spi->sc_minshiftv, sqrt(spi->sc_r1),
		spi->sc_r3, spi->sc_maxdi, spi->sc_rmsde);
	if (fp != stdout) (void)xeflush(fp);
}

static void cycle_summary(CYCLESHIFT *g0, SCALERUN *spi)
{ /* Summarize a single scaling cycle to spi->sc_fpl */
	register int		i;
	register CYCLESHIFT	*tg;
	FILE			*lfp;
	SHIFT			*tsh;
	VECPAR			*vpe, *vpp;

	lfp = spi->sc_fpl;
	fprintf(lfp, "\nScale and R factors on cycle %d\n",
		(int)(spi->sc_cycno));
	switch (spi->sc_nparpshift)
	  {
	   case	1:
		fprintf(lfp,
		 " Shift    # of   New g   Old g   RMS R-    variance\n");
		fprintf(lfp,
		 "Number     Obs                   factor      in (g)\n");
		for (tg = g0, i = 0, tsh = spi->sc_shlp; i < spi->sc_nscales;
		 i++, tsh++, tg++)
		   if ((!(tsh->shval % 2)) || !(spi->sc_flags & SC_ALLOFDET))
			fprintf(lfp, "%6d%8d%8.4f%8.4f%9.5f%12.5e\n",
		  		tsh->shval, tg->nobsh, tg->newg,
				tsh->shg, sqrt(tg->rfac), tg->mvar[0]);
	  	break;
	    case	2:
		fprintf(lfp, " Shift # of  New g  Old g  New b  Old b");
		fprintf(lfp, "  RMS R-  variance  variance     covar\n");
		fprintf(lfp, "Number Obs                               ");
		fprintf(lfp, "factor    in (g)    in (b)      (gb)\n");
		for (tg = g0, i = 0, tsh = spi->sc_shlp; i < spi->sc_nscales;
		 i++, tsh++, tg++)
			fprintf(lfp,
			 "%6d%5d%7.4f%7.4f%7.3f%7.3f%8.5f%10.3e%10.3e%10.3e\n",
		  	 tsh->shval, tg->nobsh, tg->newg, tsh->shg,
		  	 tg->newb, tsh->shb, sqrt(tg->rfac),
			 tg->mvar[0], tg->mvar[2], tg->mvar[1]);
	  	break;
	    case	3:
		fprintf(lfp, " Shift  # of  New g  Old g  New a  Old a");
		fprintf(lfp, "  New b  Old b   RMS R-  variance  variance");
		fprintf(lfp, "  variance     covar     covar     covar\n");
		fprintf(lfp, "Number   Obs                            ");
		fprintf(lfp, "                 factor    in (g)    in (a)");
		fprintf(lfp, "       (b)      (ga)      (gb)      (ab)\n");
		for (tg = g0, i = 0, tsh = spi->sc_shlp; i < spi->sc_nscales;
		 i++, tsh++, tg++)
			fprintf(lfp,
			 "%6d%6d%7.4f%7.4f%7.2f%7.2f%7.2f%7.2f%10.6f",
		  	 tsh->shval, tg->nobsh, tg->newg, tsh->shg, tg->newa,
			 tsh->sha, tg->newb, tsh->shb, sqrt(tg->rfac));
			fprintf(lfp, "%10.3e%10.3e%10.3e%10.3e%10.3e%10.3e\n",
			 tg->mvar[0], tg->mvar[3], tg->mvar[5],
			 tg->mvar[1], tg->mvar[2], tg->mvar[4]);
	  	break;
	  }
	if (spi->sc_flags & SC_VECREFINED)
	  {
		vpe = spi->sc_vshp + spi->sc_ndvgroups;
		fprintf(lfp, "Spherical-harmonic scaling parameters:\n");
		fprintf(lfp,
		 "Param scaling   A0,0  A1,0    A1,1  A1,-1");
		if (spi->sc_noutpar > 4)
		  {
			fprintf(lfp, "   A2,0   A2,1   A2,2  A2,-1  A2,-2");
			if (spi->sc_noutpar == 10) fprintf(lfp, "   A2,r");
			else if (spi->sc_noutpar > 10) fprintf(lfp,
		 "   A3,0   A3,1   A3,-1  A3,2  A3,-2   A3,3  A3,-3");
		  }
		fprintf(lfp, "\n type   group\n");
		for (vpp = spi->sc_vshp; vpp < vpe; vpp++)
		   {
			fprintf(lfp, " %c %10d",
				(sichar)vpp->v_ptype, vpp->v_ms);
			for (i = 0; i < (int)(spi->sc_noutpar); i++)
				fprintf(lfp, "%7.3f", vpp->v_coef[i]);
			fprintf(lfp, "\n");
		   }
	  }
}

static void bestscnl(SCALERUN *spi)
{ /*	This is called when we finish with "scminim". */
	register double	*bp, *xp;
	size_t		i;

	for (i = 0, xp = spi->sc_now0, bp = spi->sc_best0;
	 i < spi->sc_totpar; i++) *xp++ = *bp++;
	spi->sc_scrnow = spi->sc_r1;
	partoscale(spi->sc_now0, spi);
	/* do a better scaling of the final results */
	allrvals(spi->sc_pout, spi);
	if (spi->sc_pout[3] > DFUZZ)
	  {
		spi->sc_gsca = spi->sc_pout[2] / spi->sc_pout[3];
		nlrescale(spi);
		compare_nl(spi);
		cycle_quicksum(stdout, spi);
		cycle_quicksum(spi->sc_fpl, spi);
	  }
	cycle_summary(spi->sc_g, spi);
	memzap((AVOID **)&spi->sc_new0, "new0 in bestscnl");
	memzap((AVOID **)&spi->sc_g, "spi->sc_g in bestscnl");
}

static double initscmin(int state, SCALERUN *spi)
{ /*	This starts up a run of `scminim' and gets the initial residual. */
	register int	i, j;
	int		firstharmonic;
	double		currg, startdel, *sc, *scbest, *scdel;
	double		*scmax, *scmin, *scnew, *scnow, *scprev;
	SHIFT		*shp;
	VECPAR		*vpp;

	spi->sc_cycno = 0;	/* clear count of # of times "resid" called */
	spi->sc_minshiftv = 1.;	/* always use entire shift */
	/* The number of parameters is N * p + M * r, where:
		N is the number of scaling groups,
		p is the # of scaling parameters per scaling group, 
		M is the number of data runs,
		r is the number of spherical-harmonic coefficients per run.
	Thus p = 1, 2, or 3, and r = 0, 9, or 10, depending on the
	user's runtime preferences. */
	firstharmonic = spi->sc_nscales * spi->sc_nparpshift;
	spi->sc_totpar = (size_t)firstharmonic +
		spi->sc_ndvgroups * spi->sc_noutpar;
	if (spi->sc_totpar < 1)
	  {
		fprintf(stderr, "Nothing to scale on!\n");	return 1.;
	  }
	if (NULL == (spi->sc_new0 = (double *)callop(spi->sc_totpar * 8,
	 sizeof (double), "new0 in initscmin"))) return 1.;
	scnew = spi->sc_new0;
	scnow = spi->sc_now0 = spi->sc_new0 + spi->sc_totpar;
	scmin = spi->sc_min0 = spi->sc_now0 + spi->sc_totpar;
	scmax = spi->sc_max0 = spi->sc_min0 + spi->sc_totpar;
	scdel = spi->sc_del0 = spi->sc_max0 + spi->sc_totpar;
	scbest = spi->sc_best0 = spi->sc_del0 + spi->sc_totpar;
	sc = spi->sc_s0 = spi->sc_best0 + spi->sc_totpar;
	scprev = spi->sc_prev0 = spi->sc_s0 + spi->sc_totpar;
	spi->sc_r1 = 20000.; /* ridiculous residual */
	for (i = 0, shp = spi->sc_shlp; i < spi->sc_nscales; i++, shp++)
	   {
		if ((currg = shp->shg) < FUZZ) currg = 1.;
		*scnow++ = currg;
		if (spi->sc_nparpshift > 1)
		  {
			if (spi->sc_nparpshift > 2) *scnow++ = shp->sha;
			*scnow++ = shp->shb;
		  }
	   }
	if (spi->sc_flags & SC_VECREFINED)
	  {
		for (i = 0, vpp = spi->sc_vshp; i < (int)(spi->sc_ndvgroups);
		 i++, vpp++)
			for (j = 0; j < (int)(spi->sc_noutpar); j++)
				*scnow++ = vpp->v_coef[j];
	  }
	startdel = INITDEL * ((double)(state + 8)) / ((double)(10 * state + 8));
	for (i = 0, scnow = spi->sc_now0;
	 i < (int)(spi->sc_totpar); i++, scnow++, scdel++)
	   {
		*scprev++ = *scnew++ = *scbest++ = *sc++ = *scnow;
		*scdel = startdel;
		*scmax = *scnow + LIMITRAT * (*scdel); /* max to move up */
		*scmin = *scnow - LIMITRAT * (*scdel); /*max to move down*/
		if (i < firstharmonic)
		  {
			if (!(i % spi->sc_nparpshift) && (*scmin < 0.05))
				*scmin = 0.05;
		  }
		else if (!((i - firstharmonic) % (spi->sc_noutpar)))
		  {
			*scdel = 0.;	*scmax = *scmin = *scnew;
		  }
		scmax++;	scmin++;
	   }
	/* fix the first scale parameter to its original value */
	*spi->sc_del0 = 0.;
	*spi->sc_min0 = *spi->sc_max0 = *spi->sc_best0 =
		*spi->sc_s0 = *spi->sc_new0;
	spi->sc_gsca = 1.;
	allrvals(spi->sc_pout, spi); /* residuals for scaling groups */
	spi->sc_r1 = spi->sc_pout[0];	spi->sc_r3 = spi->sc_pout[1];
	if (spi->sc_pout[3] > DFUZZ) spi->sc_gsca =
		spi->sc_pout[2] / spi->sc_pout[3];
	spi->sc_cycno = 1;	/* we've now done the first R-value calc. */
	spi->sc_nkbest = state; /* remember which cycle we're in */
	if (spi->sc_flags & SC_VERBOSE) cycle_summary(spi->sc_g, spi);
	return spi->sc_pout[0];	/* return weighted RMS residual */
}

static void scaleph1(int doshrink, SCALERUN *spi)
{ /*	Phase I of the minimization process. */
	size_t	i;
	double	scrnew, *scdel, *scmax, *scmin, *scnew, *scnow;

	scnew = spi->sc_new0;	scnow = spi->sc_now0;	scdel = spi->sc_del0;
	scmin = spi->sc_min0;	scmax = spi->sc_max0;
	/* if appropriate, we shrink all the scdel values at this point */
	if (doshrink)
	  {
		for (i = 0; i < spi->sc_totpar; i++, scdel++) *scdel *= DIVR;
		scdel -= spi->sc_totpar;
	  }
	/* move params one at a time up by scdel and down by scdel */
	for (i = 0; i < spi->sc_totpar;
	 i++, scnew++, scnow++, scmax++,scmin++,scdel++) 
	   {
		*scnew += *scdel;
		if (*scnew > *scmax) *scnew = *scmax;
		scrnew = scresid(spi, i);
		if (scrnew < spi->sc_scrnow) spi->sc_scrnow = scrnew;
		else {
			*scnew = *scnow - *scdel;
			if (*scnew < *scmin) *scnew = *scmin;
			scrnew = scresid(spi, i);
			if (scrnew < spi->sc_scrnow)
				spi->sc_scrnow = scrnew;
			else			*scnew = *scnow;
		     }
	   }
}

static void scaleph2(SCALERUN *spi)
{ /*	Phase II of the minimization process. */
	size_t	i;
	double	scrnew, *scmax, *scmin, *scnew, *scnow, t;

	scnew = spi->sc_new0;	scnow = spi->sc_now0;
	scmin = spi->sc_min0;	scmax = spi->sc_max0;
	/* if the current residual is better than the old one,
	 we continue the refinement by checking some other values of x */
	while (1) /* move params to (twice current val - old val)*/
	     {
		for (i = 0; i < spi->sc_totpar; i++, scnow++, scnew++,
			scmin++, scmax++)
		   {
			t = *scnow;	*scnow = *scnew;
			*scnew = *scnew + *scnew - t;
			if (*scnew > *scmax) *scnew = *scmax;
			if (*scnew < *scmin) *scnew = *scmin;
		   }
		scnew = spi->sc_new0;	scnow = spi->sc_now0;
		scmax = spi->sc_max0;	scmin = spi->sc_min0;
		scrnew = scresid(spi, -1);
		if (scrnew >= spi->sc_scrnow) break;
		spi->sc_scrnow = scrnew;
	     }
	/* when that particular trick has exhausted its usefulness,
	 store current param values in scnew & evaluate refinement progress*/
	for (i = 0; i < spi->sc_totpar; i++, scnew++, scnow++)
		*scnew = *scnow;
	spi->sc_scrold = spi->sc_scrnow;
}

void scminim(int state, SCALERUN *spi)
{ /* 	Optimization routine for nonlinear scaling of diffraction data.
	At the moment we will simply minimize the weighted RMS residual;
	maybe later we'll minimize on some sort of weighted combination of
	the weighted RMS residual with the unweighted absolute-value residual.
	`state' is the number of cycles of linear refinement that have
	already been performed before beginning this operation; that affects
	the delta values used in the nonlinear refinement. */
	int		doshrink;

	/* reserve memory, set up minimization, and get initial residual */
	spi->sc_scrold = spi->sc_scrnow = initscmin(state, spi);
	if (spi->sc_scrnow > 0.999)
	  {
		fprintf(stderr, "Initial residual = %9.5f: quit!\n",
			spi->sc_scrnow);	return;
	  }
/* adjust the parameters, finding the minimum scaling residual.
    The overall loop is entered spi->sc_nk times.  */
	doshrink = 0;	spi->sc_tl0 = NOW;
	/* force-reset all our summations in get_postr */
	(void)scresid(spi, -1);
	for (spi->sc_nk = state; spi->sc_nk < spi->sc_maxcycles; )
	   { /* have we been stuck for 2 cycles? */
		if (spi->sc_nk > (spi->sc_nkbest + 2)) break;
		scaleph1(doshrink, spi);
		if (spi->sc_scrnow < spi->sc_scrold)
		  {
			scaleph2(spi);	doshrink = 0;
		  }
		/* residual is as good as we can get it for current values
		 of deltas.  Shrink them and loop through again. */
		else {
			spi->sc_tl1 = NOW;	compare_nl(spi);
			cycle_quicksum(spi->sc_fpl, spi);
			cycle_quicksum(stdout, spi);
			(spi->sc_nk)++;	doshrink = 1;
			spi->sc_tl0 = spi->sc_tl1;
		     }
	   }
	/* we've gone through the shrink loop as often as permitted, so
	 we're finished.  Reconvert the `scnow' variables to globals. */
	bestscnl(spi);
}

void sccalc_kill(SCALERUN *spi)
{ /*	This frees the memory obtained in `harmonicmem' and one other place */

	if ((spi->sc_ndvgroups > 0) && (NULL != spi->sc_rb)
	 && (NULL != spi->sc_yb))
	  {
		memzap((AVOID **)&(*(spi->sc_rb)),
			"*spi->sc_rb in sccalc_kill");
		memzap((AVOID **)&(*(spi->sc_yb)),
			"*spi->sc_yb in sccalc_kill");
		memzap((AVOID **)&spi->sc_rb, "spi->sc_rb in sccalc_kill");
		memzap((AVOID **)&spi->sc_yb, "spi->sc_yb in sccalc_kill");
	  }
	memzap((AVOID **)&(spi->sc_avec), "spi->sc_avec in sccalc_kill");
	memzap((AVOID **)&(spi->sc_saveg), "spi->sc_saveg in sccalc_kill");
}

static void harmonicmem(SCALERUN *spi)
{ /*	This reserves memory for the vectors used calculating the
	spherical-harmonic parameters by least squares. */
	int	uptriang;
	size_t	i;
	double	**rdp, *t1p, *t2p, **ydp;

	spi->sc_rb = (double **)callop(spi->sc_ndvgroups, sizeof (double *),
	 "spi->sc_rb in harmonicmem");
	spi->sc_yb = (double **)callop(spi->sc_ndvgroups, sizeof (double *),
	 "spi->sc_yb in harmonicmem");
	spi->sc_avec = (double *)callop(spi->sc_noutpar, sizeof (double),
	 "spi->sc_avec in harmonicmem");
	uptriang = (spi->sc_noutpar * (spi->sc_noutpar + 1)) / 2;
	t1p = (double *)callop(uptriang * spi->sc_ndvgroups,
		sizeof (double), "t1p  in harmonicmem");
	t2p = (double *)callop(spi->sc_noutpar * spi->sc_ndvgroups,
		sizeof (double), "t2p  in harmonicmem");
	rdp = spi->sc_rb;	ydp = spi->sc_yb;
	for (i = 0; i < spi->sc_ndvgroups; i++)
	   {
		*rdp = t1p;	*ydp = t2p;
		rdp++;	ydp++;	t1p += uptriang; t2p += spi->sc_noutpar;
	   }
}

static void report_geom(FILE *fp, SCALERUN *spi)
{ /*	This summarizes the geometric results of a linear scaling cycle */
	xgulong		i, k;
	float		*tfp;
	VECPAR		*vpp;

	fprintf(fp,
 "Min, Max, Mean Geometric Correction Values: %7.3f, %7.3f, %7.3f\n",
		spi->sc_minqf, spi->sc_maxqf, spi->sc_meanqf);
	fprintf(fp, " Run Number   Geometric Parameters\n");
	  for (vpp = spi->sc_vshp, i = 0; i < spi->sc_ndvgroups; vpp++, i++)
	     {
		fprintf(fp, "    %7d", i);
		for (k = 0, tfp = &(vpp->v_coef[0]); k < spi->sc_noutpar;
		 k++, tfp++) fprintf(fp , "%6.3f", *tfp);
		fprintf(fp , "\n");
	     }
	fprintf(fp," Minval coef");
	for (k = 0; k < spi->sc_noutpar; k++)
		fprintf(fp, "%6.3f", spi->sc_minco[k]);
	fprintf(fp,"\n Minval pars, X, Y");
	for (k = 0; k < spi->sc_noutpar; k++)
		fprintf(fp, "%6.3f", DVCSCALE * spi->sc_minp[k]);
	fprintf(fp,"\n Maxval coef");
	for (k = 0; k < spi->sc_noutpar; k++)
		fprintf(fp, "%6.3f", spi->sc_maxco[k]);
	fprintf(fp,"\n Maxval pars, X, Y");
	for (k = 0; k < spi->sc_noutpar; k++)
		fprintf(fp, "%6.3f", DVCSCALE * spi->sc_maxp[k]);
	fprintf(fp, "\n");
}

static int preserve_current(int nuse, SCALERUN *spi)
{ /*	This remembers what the current values are */
	int			i;
	register SHIFT		*tsh;

	if ((spi->sc_cycno < 1) || (NULL == spi->sc_saveg))
	  {
		memzap((AVOID **)&spi->sc_saveg,
			"spi->sc_saveg in preserve_current");
		if (NULL == (spi->sc_saveg = (double *)callop(6 * nuse,
		 sizeof (double), "spi->sc_saveg in preserve_current")))
			return EALLOC;
	  }
	spi->sc_savesh = spi->sc_saveg + nuse;
	spi->sc_saveb = spi->sc_savesh + nuse;
	spi->sc_savea = spi->sc_saveb + nuse;
	zeromem((AVOID *)spi->sc_saveg, 4 * nuse * sizeof (double));
	for (i = 0, tsh = spi->sc_shlp; i < nuse; tsh++, i++)
	   {
		spi->sc_saveg[i] = tsh->shg;
		spi->sc_savesh[i] = tsh->shvar;
	   }
	if (spi->sc_nparpshift > 1)
	  {
		for (i = 0, tsh = spi->sc_shlp; i < nuse; tsh++, i++)
			spi->sc_saveb[i] = tsh->shb;
		if (spi->sc_nparpshift > 2)
		  {
			for (i = 0, tsh = spi->sc_shlp; i < nuse;
			 tsh++, i++) spi->sc_savea[i] = tsh->sha;
		  }
	  }
	return EOK;
}

static void install_current(int nuse, double tshift, SCALERUN *spi)
{ /*	This sets the scale parameters to those required for this trial. */
	register int	i;
	double		oldwt;
	SHIFT		*tsh;
	CYCLESHIFT	*tg;

	oldwt = 1. - tshift;
	for (tsh = spi->sc_shlp, tg = spi->sc_g, i = 0; i < nuse;
	 tsh++, i++, tg++)
		tsh->shg = oldwt * spi->sc_saveg[i] + tshift * tg->newg;
	if (spi->sc_nparpshift > 1)	 /* update scale factors */
	  {
		for (tsh = spi->sc_shlp, tg = spi->sc_g, i = 0;
		 i < nuse; tsh++, i++, tg++)
		  tsh->shb = oldwt * spi->sc_saveb[i] + tshift * tg->newb;
		if (spi->sc_nparpshift > 2)
		  for (tsh = spi->sc_shlp, tg = spi->sc_g, i = 0; i < nuse;
		   tsh++, i++, tg++)
		    tsh->sha = oldwt * spi->sc_savea[i] + tshift * tg->newa;
	  }
}

static void prcycle_header(FILE *fp)
{ /* Print a header for the cycle-specific summaries */
	fprintf(fp, "Cycle   # obs # refls time  Overall Fraction");
	fprintf(fp, " ---R-Values----Max Shift    RMSDev\n");
	fprintf(fp, "  No.    used    used  sec Scalefac  Applied");
	fprintf(fp, "    Rwtd  Runwtd   from previous\n");
}

int do_cycle(SCALERUN *spi)
{ /*	Perform one linear scale cycle, including preliminaries & end stuff.
		Notations used here:
	"^2" means "squared" in these equations.
	i indexes the shifts and j indexes the reflections, so
	Iij is the intensity of the refl j in shift i, and 
	wij is the unsquared weight factor, wij = 1/sigma(Iij).
	Yj = <Iij>, taken as the weighted mean over the shifts.
	Gij is the scale factor for reflection j in shift i,
	such that the minimization we seek is min((Yj*Gi - Iij)^2),
	and we get Yj by least-squares:
		Yj =	SUM GijIijwij / SUM (Gij)^2wij
			 i		 i
	Gij is currently taken as a one, two, or three-term Taylor
	expansion in sin(theta)/lambda:
		Gij =  gi + Ai * sj + Bi * sj * sj, sj = sin(thetaj)/lambda.
	The R-factor we currently compute is as follows:
			  SUM ((Iij - GijYj)wij)^2
			   ij
		R = sqrt ------------------------ = weighted rms R-factor on I
			      SUM (Iij*wij)^2)
			       ij
	This routine returns 1 if convergence has occurred and we're
	supposed to exit after this cycle; it returns -1 if a fatal error
	has occurred; it returns 0 otherwise. */
	register int	k;		/* indices */
	register QOB	*obl;
	register QREF	*rfl;
	int		ngoodruns, nob, nused;
	size_t		i;
	char		doprint, quitnow;	/* flags: printing, quitting */
	double		gdiff, newwt, nsc, oldwt, shift_bottom, shift_top;
	double		shiftv, sum0, **rdp, **ydp, lpo[4];
	float		*tsp;
	CRYSTAL		*xhp;
	VECPAR		*vpp, *vpq;
	CYCLESHIFT	*tg, *tgend;
	SHIFT		*tsh, *tshq;
	static int	bestcyc;
	static double	bsofar, prevrf;

	/*1st pass:initialize vbls*/
	xhp = spi->sc_xhdp;
	if (!(spi->sc_cycno))
	  { /* Get memory for array of structures used in shift scaling.  */
		if (spi->sc_flags & SC_VECREFINED) harmonicmem(spi);
		if (NULL ==(AVOID *)(spi->sc_g =
		 (CYCLESHIFT *)callop(xhp->xt_nshifts, sizeof (CYCLESHIFT),
		 "spi->sc_g in do_cycle"))) return EALLOC;
		spi->sc_ovrf = 1.;	bsofar = 1.e12;	bestcyc = 65535;
		spi->sc_maxdi = spi->sc_rmsde = 0.;
	  }
	doprint = ((spi->sc_cycno == spi->sc_maxcycles) ||
	 ((spi->sc_flags & SC_VERBOSE) && !(spi->sc_flags & SC_INPUTSCALE)));
	spi->sc_tl0 = NOW;
	for (k = 0, tg = spi->sc_g; k < spi->sc_nscales; k++, tg++)
		clear_shift(tg, spi);	/* zero summations for shifts */
	/* Now go through the reflections in the memory tank,
	 compute Yj = <Iij> for each reflection, and increment the
	 appropriate sums for getting the scale and R-factors. */
	spi->sc_nrefput = spi->sc_nobput = 0;
	if (spi->sc_flags & SC_VECREFINED)
	  {
		spi->sc_minqf = 1.e10;	spi->sc_maxqf = -spi->sc_minqf;
		spi->sc_meanqf = spi->sc_qfden = 0.;
	  }
	for (rfl = spi->sc_refbase, obl = spi->sc_obbase; (nob = rfl->qr_nob);
		rfl++, obl += nob)	y_and_gsums(rfl, obl, spi);
	/* Finished looping through the reflections.  Compute the new round of
	scale constants and the resulting R factors.  The details of this
	calculation depend on how many parameters per shift we're computing.*/
	if (!(spi->sc_cycno)) spi->sc_r1 = 1.;
	prevrf = spi->sc_r1;
	spi->sc_ovrf = spi->sc_ovrfden =
		spi->sc_gsum = spi->sc_gsumden = 0.;
	tgend = spi->sc_g + spi->sc_nscales;
	switch (spi->sc_nparpshift)
	 {
		case	1:	/* linear scaling */
			for (tg = spi->sc_g; tg < tgend; tg++)
				calc1par(tg, spi);
			break;
		case	2:	/* scale using g + B*s*s */
			for (tg = spi->sc_g; tg < tgend; tg++)
				calc2par(tg, spi);
			break;
		case	3:	/* scale using g + A*s + B*s*s */
			for (tg = spi->sc_g; tg < tgend; tg++)
				calc3par(tg, spi);
			break;
	  } /* end of R-factor and g-coefficient calculational loop */
	if (verror((spi->sc_gsumden > DFUZZ),
	 "scale data: no <I> > 0 values in tank")) return -1;
	spi->sc_maxdi = spi->sc_rmsde = 0.;
	spi->sc_gsum /= spi->sc_gsumden;
/* renormalize the scale factors */
	for (tg = spi->sc_g, tsh = spi->sc_shlp; tg < tgend; tg++, tsh++)
	   {
		tg->newg /= spi->sc_gsum;
		if (spi->sc_nparpshift > 1)
		  {
			tg->newb /= spi->sc_gsum;
			if (spi->sc_nparpshift > 2) tg->newa /= spi->sc_gsum;
		  }
		for (k = 0; k < 6; k++) tg->mvar[k] /= spi->sc_gsum;
		gdiff = fabs(tg->newg - tsh->shg);
		spi->sc_rmsde += gdiff * gdiff;
		if (tg->nobsh > MINRELY && gdiff > spi->sc_maxdi)
			spi->sc_maxdi = gdiff;
	  }
	if (spi->sc_rmsde > DFUZZ) spi->sc_rmsde =
		sqrt(spi->sc_rmsde / ((double)(tgend - spi->sc_g)));
	else spi->sc_rmsde = 0.;
	if ((!(spi->sc_flags & SC_KEEPCYC)) &&
	 (spi->sc_maxdi < CONV_RESC * (double)(spi->sc_convcrit)))
	  {
		quitnow = 1;	doprint = 1;
		fprintf(stdout,
		 "\nAll scales converged to within %8.5f on cycle %d\n",
			spi->sc_maxdi, (int)(spi->sc_cycno));
	  }
	else quitnow = 0;
	oldwt = 1. - spi->sc_appshift;	newwt = spi->sc_appshift;
	if (doprint) cycle_summary(spi->sc_g, spi); /* results to logfile */
	nused = tgend - spi->sc_g; /* save unmodified values */
	/* in unstable cases, we may want the shift vector to
	 be scaled by a constant ranging between -0.2 and +1. */
	if (EOK != preserve_current(nused, spi)) return quitnow;
	spi->sc_minshiftv = shiftv = 1.;
	install_current(nused, shiftv, spi);
	get_postr(lpo, spi, -1);
	spi->sc_r1 = lpo[0];	spi->sc_r3 = lpo[1];
	for (shiftv = -0.2; shiftv <= 1.; shiftv += 0.4)
	   {
		install_current(nused, shiftv, spi); get_postr(lpo, spi, -1);
		if (lpo[0] < spi->sc_r1)
		  {
			spi->sc_minshiftv = shiftv;
			spi->sc_r1 = lpo[0];	spi->sc_r3 = lpo[1];
		  }
	   }
	shift_bottom = flimits(spi->sc_minshiftv - 0.2, -0.2, 0.55);
	shift_top = shift_bottom + 0.4;
	for (shiftv = shift_bottom; shiftv <= shift_top; shiftv += 0.05)
	   {
		install_current(nused, shiftv, spi);
		get_postr(lpo, spi, -1);
		if (lpo[0] < spi->sc_r1)
		  {
			spi->sc_minshiftv = shiftv;
			spi->sc_r1 = lpo[0];	spi->sc_r3 = lpo[1];
		  }
	   }
	/* if, among the cycles we've seen so far, this is the best
	 we've seen, remember the scale factors and diffraction-vector stuff */
	if (bsofar > spi->sc_r1)
	  {
		bsofar = spi->sc_r1;	bestcyc = spi->sc_cycno;
		tsh = spi->sc_shlp;	tshq = tsh + spi->sc_nscales;
		for (k = 0; k < spi->sc_nscales; k++, tsh++, tshq++)
			cpymem((AVOID *)tshq, (AVOID *)tsh, sizeof (SHIFT));
		if (spi->sc_flags & SC_VECREFINED)
		 {
		   for (i = 0, vpp = spi->sc_vshp,
		    vpq = spi->sc_vshp + spi->sc_ndvgroups;
			i < spi->sc_ndvgroups; i++, vpp++, vpq++)
			      for (k = 0; k < MAXVECPARS; k++)
				vpq->v_coef[k] = vpp->v_coef[k];
		 }
	  }
	install_current(nused, spi->sc_minshiftv, spi);
	spi->sc_maxdi *= fabs(spi->sc_minshiftv);
	for (tg = spi->sc_g, tsh = spi->sc_shlp; tg < tgend; tg++, tsh++)
	   {
		tsh->shvar = (tg->nobsh > spi->sc_nparpshift ?
			tg->rfac / (tg->nobsh - spi->sc_nparpshift) : 0.);
	   }
		/* overall RMS R-factor*/
	if (spi->sc_ovrfden > DFUZZ)	spi->sc_ovrf /= spi->sc_ovrfden;
	else			spi->sc_ovrf = 1.;
	if (spi->sc_r1 >= prevrf)
	  {
		if (spi->sc_flags & SC_SWITCH)
		  {
			fprintf(stdout,
			 "R-factor going up: switching to nonlinear\n");
			fprintf(spi->sc_fpl,
			 "R-factor going up: switching to nonlinear\n");
			k = spi->sc_maxcycles - 5;	if (k < 0) k = 0;
			scminim(k, spi);	return 1;
		  }
		else if (fabs(spi->sc_minshiftv) < DFUZZ)
		  {
			fprintf(stdout,
			 "R-factor appears to have bottomed out; quitting\n");
			fprintf(spi->sc_fpl,
			 "R-factor appears to have bottomed out; quitting\n");
			quitnow = 1;	doprint = 1;
			cycle_summary(spi->sc_g, spi); /* results to logfile */
		  }
	  }
	if (spi->sc_flags & SC_VECREFINED)
	  { /* do the computations associated with the
		diffraction-vector-dependent scaling */
		for (i = 0, ydp = spi->sc_yb, rdp = spi->sc_rb;
		 i < spi->sc_ndvgroups; i++, ydp++, rdp++)
			inv_init(*rdp, *ydp, (int)(spi->sc_noutpar));
		/* on highly redundant datasets we only run diffv_ref
		on a fraction of the reflections to save time. */
		if (spi->sc_freq > 1)
		  {
			for (rfl = spi->sc_refbase, obl = spi->sc_obbase;
			 (nob = rfl->qr_nob); rfl++, obl += nob)
				if (!((rfl - spi->sc_refbase) %spi->sc_freq))
					diffv_ref(rfl, obl, spi);
		  }
		else for (rfl = spi->sc_refbase, obl = spi->sc_obbase;
		 (nob = rfl->qr_nob); rfl++, obl += nob)
			diffv_ref(rfl, obl, spi);
		for (ydp = spi->sc_yb, rdp = spi->sc_rb, vpp = spi->sc_vshp,
		 sum0 = i = ngoodruns = 0; i < spi->sc_ndvgroups;
		  i++, ydp++, rdp++, vpp++)
		   {
			inv_solve(*rdp, *ydp, spi->sc_avec,
				(int)(spi->sc_noutpar));
			for (k = 0, tsp = &(vpp->v_coef[0]);
			 k < (int)(spi->sc_noutpar); k++, tsp++)
			   {
				nsc = oldwt * (*tsp) + newwt * spi->sc_avec[k];
				*tsp = nsc;
			   }
			sum0 += vpp->v_coef[0];
			if (fabs(vpp->v_coef[0]) > FUZZ) ngoodruns++;
		   }
		if (spi->sc_qfden > DFUZZ && spi->sc_meanqf > DFUZZ)
		  {
			spi->sc_meanqf /= spi->sc_qfden;
			sum0 *= spi->sc_meanqf;
		  }
		sum0 /= (double)ngoodruns;
		if ((fabs(sum0) > DFUZZ) && (spi->sc_flags & SC_VERBOSE))
			report_geom(spi->sc_fpl, spi);
	  }
	spi->sc_tl1 = NOW;
	if (!(spi->sc_cycno))
	  {
		prcycle_header(stdout);	prcycle_header(spi->sc_fpl);
	  }
	else if (doprint) prcycle_header(spi->sc_fpl);
	spi->sc_gsca = spi->sc_gsum;
	cycle_quicksum(stdout, spi);	cycle_quicksum(spi->sc_fpl, spi);
	if ((quitnow) || (spi->sc_cycno == spi->sc_maxcycles))
	  { /* release local memory */
		rdp = spi->sc_rb;
		memzap((AVOID **)&(*rdp), "*rdp in do_cycle");
		ydp = spi->sc_yb;
		memzap((AVOID **)&(*ydp), "*ydp in do_cycle");
		memzap((AVOID **)&spi->sc_rb, "spi->sc_rb in do_cycle");
		memzap((AVOID **)&spi->sc_yb, "spi->sc_yb in do_cycle");
		memzap((AVOID **)&(spi->sc_avec),"avec in do_cycle");
		memzap((AVOID **)&(spi->sc_g), "spi->sc_g in do_cycle");
		memzap((AVOID **)&(spi->sc_saveg),
			"spi->sc_saveg in do_cycle");
		spi->sc_savea = spi->sc_saveb = spi->sc_savesh = NULL;
		/* also, at this point we look back through what we've done
		 and keep the scaling parameters that gave the lowest Rmerge */
		if (bsofar < spi->sc_r1) /* ... an earlier instance worked */
		  {
			fprintf(spi->sc_fpl ,
			 "Using parameters from cycle %d\n", bestcyc);
			fprintf(stdout, "Using parameters from cycle %d\n",
				bestcyc);
			tsh = spi->sc_shlp;	tshq = tsh + spi->sc_nscales;
			for (k = 0; k < spi->sc_nscales; k++, tsh++, tshq++)
			 cpymem((AVOID *)tsh, (AVOID *)tshq, sizeof (SHIFT));
			if (spi->sc_flags & SC_VECREFINED)
			  {
			    for (i = 0, vpp = spi->sc_vshp,
			     vpq = spi->sc_vshp + spi->sc_ndvgroups;
			      i < spi->sc_ndvgroups; i++, vpp++, vpq++)
				for (k = 0; k < MAXVECPARS; k++)
					vpp->v_coef[k] = vpq->v_coef[k];
			  }
		  }
	  }
	return quitnow;
}

/* end of sccalc.c */
