/* scmain.c -- toplevel routines for XGEN's scaling function

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions; reorganize use of "{"
	24 Jun 2006: include xgpostint.h rather than scdefs.h
	 6 Jun 2006: use sccycles to walk through scaling;
			sumscalekrun moved to scrio
	 9 Dec 2005: insert sumscalekrun(),scdefaults(),scargs(),scswitch()
		from various now-defunct files
	20 Nov 2005: inline argument declarations and strict prototyping
	18 Oct 2005: fpso can now be local to scmain()
	11 Oct 2005: housekeeping
	11 May 2005: check return status from xgjournal
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	17 Sep 2004: call to do_cycle() updated; cycle # is now in SCALERUN
	 6 Sep 2004: cosmetic changes
	27 Aug 2004: routines used only here are declared as static; cosmetics
	27 Jul 2004: bug fix in memzap() call in scdefaults()
	19 Jun 2004: new flags -j,l,r,u,x invert sense of -k,g,d,t,h;
		scdefaults() created: sets defaults based on current
		status of $SCALEIN file; SC_DEFAULT now includes -t option
	25 Jan 2004: eliminate VSPLINE *sgsp, since it isn't being used
	21 Jan 2004: license message inserted and initial comment reordered
	 9 Jan 2004: in scargs, adjust the way we measure
		whether SCALEIN is too short to have anything useful in it
	 7 Jan 2004: SCALERUN structure elements in function args
		replace extern globals; args updated accordingly
	14 Nov 2003: replaced all globals with components of SCALERUN structure
	 9 Nov 2003: norders converted to a local variable
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	 9 Jan 2003: redundant extern declarations removed
	 5 Jan 2003: reorganized source-file header
	15 Nov 2002: contents of scglob.c moved here;
		extern declarations made here rather than in scext.h;
		ndiffv_groups renamed to ndv_groups
	15 Apr 2002: (?) in scmain(), modify uses of shlp[]
	25 Mar 2002: removed MSI attachments
	31 May 1995: "register" -> "register int" throughout
	 2 Mar 1995: diagnostics to stdout; eliminate "rwb" variable
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 1 Dec 1994: scminim declared before use
	14 Nov 1994: allow for nonlinear scaling (-b option);
		add Booleans -b,-h,-u for Simplex scaling & conditional shifts
	 4 Nov 1994: add SC_FLATVARIANCE option so sigmas don't always depend
		on (I*sigma(g))^2; echo this appropriately
	30 Sep 1994: fix list of input files in header() call
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: global variables reorganized a little
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version
	18 Jan 1988: in scmain(), call del_omegas() to establish smoothing
		values to allow linearly splined scaling parameters
	15 Jan 1988: in scmain(), only open SCALEOUT (not SCALEIN) initially.
	21 Sep 1985: incorporate conventions of XENGEN.

   Purpose: toplevel calling routines for `scalek'
   Cross references: sc*.c

 "scalek" performs a linear scaling of crystallographic data from area
 detector data, iteratively computing averaged intensities and scale and
 temperature factors for "shifts", which are batches of collected reflections
 covering a limited range of scanning angle and (sometimes) detector area.
 Details of the scaling and R-factor algorithms are given in the comments
 at the top of "sccalc.c".
 Runtime options are as follows:
	-a	All shifts used, i.e. include data from
		 shifts which don't appear in the scale file.
		 If no scale file is specified, this is default.
	-b	Perform scaling by a nonlinear, Simplex algorithm
		 instead of closed-form linear least-squares.
	-c#	Cycle for # cycles.  Default = 9.  Cycling will
		 terminate earlier on convergence unless -k
		 is specified (see below).
	-d	Compute second-order spherical-harmonic parameters
		 in the diffraction vectors (sx, sy, sz)
		 for each datarun in the dataset.
	-e	Compute second-order spherical-harmonic parameters
		 in the exit-beam vectors (kx', ky', kz')
		 for each datarun in the dataset.
	-f	Compute the variances associated with each observation
		 without regard to the variance in the scale factors.
	-g	Like "-d" except that the multiplier of the
		 (sin(theta) squared) term is forced to be zero.
	-h	Begin scaling using linear least-squares, but if
		 the overall R-value begins to increase, switch to
		 the Simplex algorithm.
	-i#	Minimum average intensity used in scaling = #.
		 Default is 0.
	-k	Keep cycling even if refinement has converged.
		 By default, program exits if scale factors
		 have not changed significantly between the
		 last run and the current run. (see -q)
	-m,-n	No merging of Bijvoets.
		 By default, Bijvoets are merged.
	-o#	At least # observations must be present to
		 include a reflection in the scaling. Default=2.
	-p#	Use # scaling parameters per shift.  Legal
		 values are 1,2,3; default is 1.
	-q#	Convergence criterion = #.  If |g(new)-g(old)|
		 < #*10^-4 for all shifts, we have converged.
		 Default = 10, corresponding to a shift of 0.001.
	-s#	Apply only partial shifts, up to a fraction #.
		 Thus if the change in a scale factor might be
		 0.3, then if #=0.4, we apply a change of 0.3*0.4=0.12.
	-t	Scale the top and bottom halves of each detector
		 together, rather than leaving them independent.
	-v	Verbose output.  By default, summaries of the
		 scale factors and R-factors are printed only
		 on the 0th and last cycles.  Specifying -v
		 puts the summaries out more often.
	-w	Wilson-statistics. The program computes initial
		 scale factors based on Wilson statistics in the
		 input data. If no SCALEIN file is specified,
		 these initial scale factors are used in the
		 scaling calculations; if SCALEIN is specified,
		 they're provided only as logged data.
*************************************************************************/
#define		NEED_TIME	1
#define		NEED_STAT	1
#include	"xgpostint.h"

static int scdefaults(SCALERUN *spi)
{ /*	This establishes defaults for a run of `scalek' using its
	knowledge of the contents of various files */
	char		*locbu, *siname;
	FILE		*fpte;
	struct stat	sibuf;

	zeromem((void *)spi, sizeof (SCALERUN));
	spi->sc_tl0 = NOW;
	spi->sc_flags = SC_DEFAULT;	spi->sc_imin = SC_IMIN;
	spi->sc_nrepmin = SC_NOBSMIN;	spi->sc_maxcycles = SC_MAXCYCLES;
	spi->sc_convcrit = SC_CONVCRIT; spi->sc_nparpshift = SC_NPARAMS;
	spi->sc_appshift = 1.;
	if (qerror((AVOID *)(spi->sc_fpmi = xgopen("multiref","r")),
		"open input multiref")) return EOPENERR;
	/* see whether the scale input file exists,
	 is readable, and has positive length */
	siname = trname("scalein");
	/* if it exists, then we're probably ready for spherical harmonics */
	if ((0 == access(siname, 4)) && (0 == stat(siname, &sibuf)) &&
		(sibuf.st_size > 10))
	  {
		spi->sc_flags |=
			(SC_INPUTSCALE | SC_DIFFRACTION | SC_DIFF_NOS2);
		/* furthermore, if there are already spherical harmonics
		 present, we're probably ready for the -h flag */
		if (NULL == (locbu = mallop(258, "locbu in scdefaults")))
			return EALLOC;
		if (NULL == (fpte = xgopen(siname, "r"))) return EOPENERR;
		while (NULL != fgets(locbu, 140, fpte))
		     {
			if ((('A' <= *locbu) && (*locbu <= 'Z')) ||
			 (('a' <= *locbu) && (*locbu <= 'z')))
			  {
				spi->sc_flags |= SC_SWITCH; break;
			  }
		     }
		(void)fclose(fpte);
		memzap((AVOID **)&locbu, "locbu in scdefaults");
	  }
	else {
		spi->sc_flags |= SC_ALLOBS;
		spi->sc_flags &= ~(SC_INPUTSCALE | SC_DIFFRACTION);
	     }
	return EOK;
}

static void scswitch(int argc, char *argv[], SCALERUN *spi)
{ /*	Process switch-type arguments for scalek.  */
	char		**flags, **dcp;
	double		spec_shift;

	spec_shift = -1.;
	flags = flagproc(argc, argv, "ABCDEFGHIJKLMNOPQRSTUVWX");
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	   switch (**dcp)
	     {
		case	'A':	/* use all observations */
			spi->sc_flags |= SC_ALLOBS; break;
		case	'B':	/* nonlinear scaling */
			spi->sc_flags |= SC_NONLINEAR;	break;
		case	'C':	/* max # of cycles */
			(void)sscanf((*dcp) + 1, "%hd", &(spi->sc_maxcycles));
			break;
		case	'D':	/* diffraction-vector sph. harm. */
			spi->sc_flags |= SC_DIFFRACTION;	break;
		case	'E':	/* exit-beam spherical harmonics */
			spi->sc_flags |= SC_EXITBEAM;	break;
		case	'F':	/* calculate weights without regard
				 to the variance in the scale factor */
			spi->sc_flags |= SC_FLATVARIANCE;	break;
		case	'G':	/* diff.vec.sph.harm.--no s^2 term */
			spi->sc_flags |= (SC_DIFFRACTION | SC_DIFF_NOS2);
			break;
		case	'H':	/* switch from linear to nonlinear */
			spi->sc_flags |= SC_SWITCH;	break;
		case	'I':	/* min <I> to use in scaling */
			(void)sscanf((*dcp) + 1, "%hd", &(spi->sc_imin));
			break;
		case	'J':	/* don't keep cycling after convergence */
			spi->sc_flags &= ~SC_KEEPCYC;	break;
		case	'K':	/* keep cycling after convergence */
			spi->sc_flags |= SC_KEEPCYC;	break;
		case	'L':	/* don't do diffvec sph.harm */
			spi->sc_flags &= ~(SC_DIFFRACTION | SC_DIFF_NOS2);
			break;
		case	'M':
		case	'N':	/* no merging of Bijvoets */
			spi->sc_flags &= ~SC_MERGEBJ;	break;
		case	'O':	/* min # of observations/refl req'd */
			(void)sscanf((*dcp) + 1, "%hd", &(spi->sc_nrepmin));
			if (spi->sc_nrepmin < 2) spi->sc_nrepmin = 2;	break;
		case	'P':	/* # of params calculated per shift */
			(void)sscanf((*dcp) +1, "%hd", &(spi->sc_nparpshift));
			if (spi->sc_nparpshift < 1 || spi->sc_nparpshift > 3)
				spi->sc_nparpshift = 1;
			break;
		case	'Q':	/* value of convergence criterion */
			(void)sscanf((*dcp) + 1, "%hd", &(spi->sc_convcrit));
			break;
		case	'R':	/* don't do ordinary spherical harmonics */
			spi->sc_flags &= ~SC_DIFFRACTION;	break;
		case	'S':	/* apply partial shifts */
			(void)sscanf((*dcp) + 1, "%lf", &spec_shift); break;
		case	'T':	/* scale top & bottom together */
			spi->sc_flags |= SC_ALLOFDET;	break;
		case	'U':	/* don't scale top & bottom together */
			spi->sc_flags &= ~SC_ALLOFDET;	break;
		case	'V':	/* verbose output */
			spi->sc_flags |= SC_VERBOSE;	break;
		case	'W':	/* initial scales by Wilson analysis*/
			spi->sc_flags |= SC_WILSON;	break;
		case	'X':	/* don't switch from linear to nonlinear */
			spi->sc_flags &= ~SC_SWITCH;	break;
		default:	break;
	   }
	if (0. < spec_shift && spec_shift <= 1.) spi->sc_appshift =spec_shift;
	/* nonlinear and switch-to-nonlinear are incompatible */
	if ((spi->sc_flags & SC_NONLINEAR) && (spi->sc_flags & SC_SWITCH))
		spi->sc_flags &= ~SC_NONLINEAR;
	return;
}

static int scargs(int argc, char *argv[], SCALERUN *spi)
{ /*	scargs does all setup necessary to begin processing scaling
	data. In this version, the names of the multiref, the input shift file,
	the output shift file, and the logfile are conventional: "multiref",
	"scalein", "scaleout", and "xlog".  Therefore the only arguments to
	the program in this version are the switches.  First, we establish
	default values of global variables; see "scdefs.h" for the numerical
	or logical values of these defaults.  */
	int		norders;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	if (EOK != (norders = scdefaults(spi))) return norders;
	/* Examine arguments.  Those beginning with a '-' are interpreted as
	switches and processed with 'scswitch'.
	Other arguments are ignored for the moment.	*/
	scswitch(argc, argv, spi);
	for (argc--, argv++; argc; argc--, argv++) if ('-' != **argv) break;
	/* Now open the files according to the "trname" conventions.
	 no Wilson stats if input .si file */
	if (spi->sc_flags & SC_INPUTSCALE) spi->sc_flags &= ~SC_WILSON;
	norders = 3;	/* 3 orders (0, 1, 2) for now */
	spi->sc_noutpar = (spi->sc_flags & SC_DIFFRACTION ? 
		(norders * (norders+1) * (norders + 2) / 6) :
		(spi->sc_flags & SC_EXITBEAM ? (norders * norders) : 0));
	if (spi->sc_flags & SC_DIFF_NOS2) (spi->sc_noutpar)--;
	return EOK;
}

int scend(int irv, SCALERUN *spi)
{ /*	This ends a run of `scalek' properly */
	QOB	*qow;

	if ((NULL != spi->sc_fpl) && (stdout != spi->sc_fpl))
		(void)fclose(spi->sc_fpl);
	if (NULL != spi->sc_fpmi) (void)fclose(spi->sc_fpmi);
	spi->sc_fpl = NULL;	spi->sc_fpmi = NULL;
	if (NULL != spi)
	  {
		sccalc_kill(spi);	headmem(NULL, NULL, 0);
		memzap((AVOID **)&(spi->sc_rub), "spi->sc_rub in scalek");
		memzap((AVOID **)&(spi->sc_shlp), "sip->sc_shlp in scalek");
		memzap((AVOID **)&(spi->sc_refbase),
			"spi->sc_refbase in scalek");
		qow = spi->sc_obbase;
		if ((NULL != qow) && (spi->sc_flags & SC_VECREFINED))
		  memzap((AVOID **)&(qow->qo_sc), "qow->qo_sc in scalek");
		memzap((AVOID **)&qow, "qow in scalek");
		memzap((AVOID **)&(spi->sc_shwl), "spi->sc_shwl in scalek");
		memzap((AVOID **)&(spi->sc_vshp), "spi->sc_vshp in scalek");
	  }
	globalkill();	return irv;
}

int scmain(int argc, char *argv[])
{ /* Mainline for the scaling functionality */
	int		i;
	SCALERUN	sps;

	/*process args,defaults*/
	if (EOK != (i = scargs(argc, argv, &sps))) return scend(i, &sps);
	return scend(sccycles(&sps), &sps);
}

/* end of scmain.c */
