/* "space_g.c":

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	20 Dec 2006: SPACEGSTAT typedef moved here from ../include/xgpostint.h
	28 Nov 2006: update ucell call in spaceghead;
	 replace mat2vecs calls with rp2vecs calls in spaceghead and elsewhere
	24 Jun 2006: include xgpostint.h rather than statdefs.h
	31 Dec 2005: don't qsort with fewer than 2 elements
	20 Nov 2005: inline argument declarations and strict prototyping
	18 Oct 2005: replace static globals with elements of SPGWORK structures
	11 Oct 2005: housekeeping
	22 Jun 2005: out_omissions() declared as global
	11 May 2005: check return status from xgjournal
	11 Apr 2005: name of this file changed from spaceg.c to space_g.c
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	27 Aug 2004: routines used only here are declared as static; cosmetics
	 7 Jun 2004: mrtofast(): star->sta_fpmi = NULL after closing it;
		correct memzap calls and set fpmo to NULL in endspace()
	27 May 2004: structure types, #defines moved to ../include/statdefs.h;
		much of functionality of spaceargs() moved to spacegstart();
		much of what was in spacegmain() is now in spacegops()
	 3 Apr 2004: bugfix re ignored observations in p1tocurr(); cosmetics
	19 Mar 2004: ignore deleted observations; adjustment in FOM weights
	08 Feb 2004: fix bug in screws() that made it blow up if
		the spacegroup actually turned out to be tetrahedral 422
	31 Jan 2004: screw-axis calcs omitted for rhombohedrals;
		print out header for screw-axis calc more efficiently
	27 Jan 2004: enlarged qk array so it doesn't step on other globals;
	 elminate some unneeded static globals; fix spacegroup bug in handling
	 reflections from REFLIST; mushed routine stat_allowed into its caller
	25 Jan 2004: some static globals converted to locals;
		VSPLINE *sgsp deleted because we don't use it
	21 Jan 2004: license message inserted and initial comment reordered
	18 Jan 2004: major bug fixed: orientation matrix for each run is
		already available in run_recip[][], so it doesn't get rotated;
		minor cleanup of logfile output
	 7 Jan 2004: STATRUN structure elements in function args
		replace extern globals; args updated accordingly;
		callop() & memzap() replace shcalloc() & shfree()
	13 Dec 2003: eliminate unused variable jor from sp_rstat();
		elim unused static global variables hklcoef[], fout, fcurr
	22 Nov 2003: replace deltaom, etc. with rphi[]
	20 Nov 2003: replace globals found in stats program with elements
		of STATRUN structure
	 8 Nov 2003: lca converted from a vector to a pointer
	 4 Nov 2003: LATTICESCORE structure type declared in ../include/xgen.h;
		bugs associated with rhombohedral spacegroups corrected;
		logic in walking through the 44 lattice characters cleaned up.
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Sep 2003: fix bug in use of centercode() routine in p1tocurr();
		call routines glatfi4() and gcenfi() where needed
	24 Aug 2003: manipulate and display lattice type & centering
		using routines lattcode, centercode, glatt4, gcentering
	16 Aug 2003: replace global uphkl[3] with static lphkl[3]
		and handle exclusions differently for keephkl()
	 5 Jan 2003: reorganized source-file header
	15 Nov 2002: extern declarations made here rather than in statext.h
	27 Apr 2002: Initial revision

   Purpose: determining the proper spacegroup.
   Cross references: statcalc.c

 This calculates the symmetry R factors for a MULTIREF using the current
 SCALEIN file to determine scaling coefficients for it.
 If the spacegroup is subject to a possible higher symmetry, it then calculates
 the symmetry R in the higher-symmetry spacegroup and reports the results.
 This program borrows a lot of its functionality from `stats', unsurprisingly.
*************************************************************************/
#include	"xgpostint.h"
#define		DELETED_FLAG	(-1000.)

 typedef struct { /* structure type for determination of spacegroup */
		short	sp_latc, sp_intl;
		int	sp_unreliable;
		double	sp_fom, sp_score, sp_abc[3], sp_ang[3];
		SHLSTAT	sp_st;
		} SPACEGSTAT;

 typedef struct {
		int		sgw_nunreliable, sgw_nrscaled;
		int		sgw_nobscaled, sgw_tindex, sgw_iqtop[3][3];
		float		sgw_qk[46][3][3], sgw_qkz[3][3];
		FILE		*sgw_fpmo;
		LATTICESCORE	*sgw_lca0, *sgw_lcap;
		FASTOB		*sgw_fobt, *sgw_fend, *sgw_fnext, *sgw_foe;
		} SPGWORK;

static short indtosp(int spind, short *sysp)
{ /*	This returns the spacegroup number associated with the index `spind'.
	It also puts the crystal system # associated with that index in *sysp*/
	static short		uspg[] =
	{ 209, 211, 207, 196, 197, 195, 177, 168, 155, 255, 146, 246, 151,
	  150, 143,  97,  89,  79,  75,  22,  23,  20,  16,   5,   3,   1, 0 };
	static short		ullco[] =
	{   5,   5,   5,   5,   5,   5,   6,   6,   7,   7,   7,   7,   6,
	    6,   6,   4,   4,   4,   4,   3,   3,   3,   3,   2,   2,   1, 0 };
	*sysp = 0;
	if ((spind < 0) || (spind > 26)) return 0;
	*sysp = ullco[spind];	return uspg[spind];
}

static void sumspacerun(FILE *fp, STATRUN *star)
{ /*	This summarizes the runtime conditions to the stream fp */
	static char	*flist[] = {"multiref", "xlog", "scalein", NULL};
	static char	ftypes[] = {1, 2, 1, 0 };

	ftypes[1] = (star->sta_fpl == stdout) ? 0 : 2;
	ftypes[2] = (0 == star->sta_nscales) ? 0 : 2;
	header(fp, REDUCTION, flist, ftypes);
	if (star->sta_flags & SP_VERBOSE) fprintf(fp,
 "\tReflections with ambiguous indices will be printed to the logfile\n");
	if (star->sta_flags & SP_OUTMREF) fprintf(fp,
 "\tAn output multiref will be written to the file %s\n", trname("mrefout"));
	return;
}

int spacegstart(STATRUN *star)
{ /* start up a run of the `spaceg' functionality */
	int		gotone, i, iline, nlines, nsca;
	short		uhkl[3];
	char		line[258];
	u_char		*cp, **checks, **chp;
	CRYSTAL		*lxhp, *xhp;
	DATARUN		*rhp;
	VECPAR		*vpp, *vpe;
	FILE		*fpil;
	static u_char	*cp0;

	if (NULL == (star->sta_xp = (REFPARS *)callop(1, sizeof (REFPARS),
	 "star->spa_xp in spaceg"))) return EALLOC;
	star->sta_spgrp = 1;
	if (qerror((AVOID *)(star->sta_fpmi = xgopen("multiref", "r")),
	 "open input multiref")) return EOPENERR; /* open the input multiref */
	if (NULL == (lxhp = (CRYSTAL *)callop(1, sizeof (CRYSTAL),
		"lxhp in spacestart"))) return EALLOC;
	if (verror(0 < xgread((void *)lxhp, XG_CRYSTAL, star->sta_fpmi),
	 "read crystal header in input multiref")) return EREADERR;
	if (EOK != (i = headmem(&(star->sta_xhdp), &(star->sta_rhdp),
		(int)(lxhp->xt_nruns)))) return i;
	xhp = star->sta_xhdp;	rhp = star->sta_rhdp;
	cpymem((AVOID *)xhp, (AVOID *)lxhp, sizeof (CRYSTAL));
	memzap((AVOID **)&lxhp, "lxhp in spacegstart");
	if (NULL == (AVOID *)(star->sta_vshp = (VECPAR *)callop(
	 (unsigned)xhp->xt_nruns, sizeof (VECPAR),
	 "star->sta_vshp in spacegstart"))) return EALLOC;
	if (NULL == (AVOID *)(star->sta_rub = (RUNWORK *)callop((unsigned)
	 (xhp->xt_nruns), sizeof (RUNWORK), "star->sta_rub in spacegstart")))
		return EALLOC;
	if (NULL == (star->sta_shlp = get_shifts(NULL, star->sta_vshp,
	 &nsca))) return EREADERR; /* read scaling inputs if available */
	star->sta_nscales = nsca;
	star->sta_spgrp = xhp->xt_intl;
	if ((xhp->xt_isys == 7) && (star->sta_spgrp < 200))
		star->sta_spgrp += 100;
	if (NULL == gspline(5, &i)) return i; /* for detector maps */
/* examine the scale-factor output to see if we actually got any geometry-
 dependent coefficients. */
	gotone = 0;	vpp = star->sta_vshp;	vpe = vpp + xhp->xt_nruns;
	for ( ; vpp < vpe; vpp++)
	   {
		for (i = gotone = 0; i < MAXVECPARS; i++)
			if (fabs(vpp->v_coef[i]) > FUZZ)
		   {
			gotone = 1; break;
		   }
		if (gotone) break;
	   }
	star->sta_flags &= ~(SC_DIFFRACTION | SC_EXITBEAM);
	if (!gotone)
	  {
		memzap((AVOID **)&(star->sta_vshp), "vshp in spacegstart");
		star->sta_nvpars = 0 ;
	  }
	else {
		vpp = star->sta_vshp;
		if (vpp->v_ptype == 'd')
		  {
			star->sta_flags |= SC_DIFFRACTION;
			star->sta_nvpars = 10;
		  }
		else {
			star->sta_nvpars = 9;
			star->sta_flags |= SC_EXITBEAM;
		     }
	     }
	if (NULL == (star->sta_fpl = propen("xlog"))) star->sta_fpl = stdout;
	if (NULL != (fpil = xgopen("reflist", "r")))
	  {
		for (nlines = 0; NULL != fgets(line, 80, fpil); ) nlines++;
		(void)fseek(fpil, 0L, 0);	/* rewind input refl list */
		if (NULL == (AVOID *)(checks = (u_char **)callop(
		 (unsigned)nlines, sizeof (u_char *), "checks in space_g")))
			return EALLOC;
		if (NULL == (AVOID *)(cp0 =(u_char *)callop(4 * nlines,
		 sizeof (u_char), "cp0 in pickuref"))) return EALLOC;
		cp = cp0;	chp = checks;
		for (iline = 0; iline < nlines; iline++, chp++, cp += 4)
		   {
			*chp = cp;
			if (NULL == fgets(line, 80, fpil))
			  {
				fprintf(stderr,
 "Error reading line %d of list file: remaining reflections ignored\n", iline);
				break;
			  }
			(void)sscanf(line, "%hd %hd %hd",
				&uhkl[0], &uhkl[1],&uhkl[2]);
			mpack(uhkl, *chp, star->sta_spgrp);
			symtr(*chp, star->sta_spgrp);
			(*chp)[3] &= 0376;	/* turn off bijvoet bit */
		   }
		(void)fclose(fpil);
		star->sta_checks = checks;	star->sta_nlines = nlines;
	  }
	else {
		star->sta_checks = NULL;	star->sta_nlines = 0;
	     }
	if (star->sta_fpl != stdout) sumspacerun(star->sta_fpl, star);
	sumspacerun(stdout, star);	return EOK;
}

static int spaceargs(int argc, char *argv[], STATRUN *star)
{ /*	"spaceargs" processes the arguments and sets up files for "spaceg".
	It reads a multiref with logical or environment name "multiref",
	along with similarly-set-up files "scalein" as input scale file
	and "xlog" as the output log file.
	If "-n" or "-m" is an input argument, Bijovets will not be merged;
	otherwise, they will be merged.
	If "-o" is an input argument, we omit the missing-reflection list.
	A purely numeric input is assumed to be the resolution limit
	of the program run (default=max.resol. of data) */
	int	i;
	char	**dcp, **flags;

	zeromem((void *)star, sizeof (STATRUN));
/*
	loop through args, setting file io, resolution limits,
	and bijvoet merges according to args.
	Options:
	-o Output new multiref if calculated spacegroup is different from input
	-v Verbose output
	-c Conditional output: output a new multiref if calculated spgrp
		is different from input, but point group has NOT changed.
		This is useful for situations where (for example) the input
		spacegroup is P3 and the output is calculated to be P622.
*/
	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	flags = flagproc(argc, argv, "COV");
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	   switch (**dcp)
	     {
		case	'C':	star->sta_flags |= SP_CONDOUTMREF;	break;
		case	'O':	star->sta_flags |= SP_OUTMREF;	break;
		case	'V':	star->sta_flags |= SP_VERBOSE;	break;
	     }
	for (dcp = argv + 1, i = argc - 1; (i--); dcp++)
		if (allnumeric(*dcp)) star->sta_resstate = atof(*dcp);
	return spacegstart(star);
}

static int spaceghead(FILE *fp, STATRUN *star, SPGWORK *sgp)
{ /*	Processes the headers in the MULTIREF pointed to by fp for `spaceg'. */
	int		curc, curch, curl, fch, i, j, nls, runno;
	short		excluc;
	double		newiq, zq, tdar[3][3];
	float		kstart[3], qtop[3][3], rm00[3][3];
	REFPARS		*xp;
	RUNWORK		*rup;
	DATARUN		*rhp;
	CRYSTAL		*xlp;
	LATTICESCORE	*lcap;
	static int	ex2c[] = { 1, 2, 4, 3, 5, 4 };

	star->sta_resmax = 10000.;	/* set to ridiculous value */
	/* Read in the crystal header and then the run headers,
	storing some useful information from them in the "orstat" structures
	and accumulating totals. */
	xlp = star->sta_xhdp;	rhp = star->sta_rhdp;	xp = star->sta_xp;
	star->sta_spgrp = xp->xispgp = xlp->xt_intl;
	xp->xisys = xlp->xt_isys;
	if ((xlp->xt_isys == 7) && (star->sta_spgrp < 200))
		star->sta_spgrp += 100;
	excluc = exclucode(star->sta_spgrp, xlp->xt_isys);
	if (verror(!(ucell(xlp->xt_ar, xlp->xt_ang, xp->xast, xp->xanst,
	 xlp->xt_isys)), "compute reciprocal-space unit cell parameters"))
		return EBADVALUE;
	for (runno = 0; runno < MAXRUNS; runno++)
		star->sta_runindex[runno] = -1;
	for (runno = 0, rup = star->sta_rub; runno < xlp->xt_nruns;
	 runno++, rup++, rhp++)
	   {
		if (verror(0 < xgread((void *)rhp, XG_DATARUN, fp),
		 "read data run header in input multiref")) return EREADERR;
		(void)keep_dimen(rhp);
		workrun(rhp, rup, xlp, (int)(runno * 200));
		for (j = 0; j < 3; j++) xp->xstart[j] = rhp->run_startpos[j];
		/* n.b.: We don't have to rotate to the (omega,chi,phi)start
		position because, in fact, that's where rup->rp_areal is. */
		if (!runno) for (j = 0; j < 3; j++) kstart[j] = xp->xstart[j];
		if (star->sta_flags & SP_VERBOSE)
		  {
			pr3m(star->sta_fpl, rup->rp_areal,
			 "multiref areal at run-start position");
		fprintf(star->sta_fpl,"xp->xstart = %8.2f %8.2f %8.2f\n",
			 xp->xstart[0], xp->xstart[1], xp->xstart[2]);
		  }
		if (rhp->run_reslim < star->sta_resmax)
			star->sta_resmax = rhp->run_reslim;
		star->sta_runindex[rhp->run_id] = runno;
		/* if the cell is not primitive, convert to primitive */
		for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
			xp->xareal[i][j] = rup->rp_areal[i][j];
		xp->xlambda = rhp->run_lambda;
		xp->xdetwid = rhp->run_width;	xp->xdetht = rhp->run_height;
		if (EOK != (i = rp2vecs(xp)))
		  {
			fprintf(stderr,
 "run %d: Error %3d in rp2vecs before cellconvert\n", runno, i);
			return EBAD;
		  }
		if (star->sta_flags & SP_VERBOSE) fprintf(star->sta_fpl,
		 "abc,ang: %8.2f %8.2f %8.2f %8.2f %8.2f %8.2f\n",
		 xlp->xt_ar[0], xlp->xt_ar[1], xlp->xt_ar[2],
		 xlp->xt_ang[0], xlp->xt_ang[1], xlp->xt_ang[2]);
		cellconvert(xp, 0);
		if (star->sta_flags & SP_VERBOSE) pr3m(star->sta_fpl,
			xp->xareal, "areal after cellconvert");
		/* define the cell lengths and angles in the primitive cell */
		if (EOK != (i = rp2vecs(xp)))
		  {
			fprintf(stderr,
 "run %d: Error %3d in rp2vecs after cellconvert\n", runno, i);
			return EBAD;
		  }
		if (star->sta_flags & SP_VERBOSE) fprintf(star->sta_fpl,
		 "abc,ang: %8.2f %8.2f %8.2f %8.2f %8.2f %8.2f\n",
		 xp->xabc[0], xp->xabc[1], xp->xabc[2],
		 xp->xang[0], xp->xang[1], xp->xang[2]);
		/* record the current and zero-level cells */
		for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
			rup->rp_ar0[i][j] = xp->xareal[i][j];
		/* calculate the transformation that gets us from the
		 current cell back to the Bravais cell */
		if ((i = inv3(rup->rp_areal, rm00)))
		  {
			fprintf(stderr,
				"Failure %d in inv3 on original areal\n", i);
			return EBAD;
		  }
		mm3(sgp->sgw_qk[0], xp->xareal, rm00);
		(void)inv3(sgp->sgw_qk[0], qtop); /* we'll use this later */
		for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		   {
			if ((zq = qtop[i][j]) < -FUZZ)
				sgp->sgw_iqtop[i][j] = zq - 0.49999;
			else	sgp->sgw_iqtop[i][j] = zq + 0.49999;
		   }
		/* we're going to need the inverse of sgp->sgw_iqtop, too: */
		for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		   {
			newiq = sgp->sgw_iqtop[i][j];
			if (fabs(newiq - qtop[i][j]) > 0.1) fprintf(stderr,
 "Warning: qtop[%1d][%1d] is changing from %6.1f to %6.1f\n",
				i, j, qtop[i][j], newiq);
			qtop[i][j] = newiq;
		   }
		(void)inv3(qtop, sgp->sgw_qkz);	/* so modified one is saved */
		if ((i = inv3(rup->rp_ar0, rm00)))
		  {
			fprintf(stderr,
				"Failure %d in inv3 on rotated areal\n", i);
			return EBAD;
		  }
		if (!runno) /* run the Bravais calculator for 1st run */
		  {
			for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
				tdar[i][j] = rup->rp_ar0[i][j];
			if ((NULL == (sgp->sgw_lca0 =
			 allbravais(tdar, &nls))) || (16 > nls))
			  {
				fprintf(stderr,
			 "run %d: only %d lattices returned\n", runno, nls);
				return EBAD;
			  }
			/* use this calculation to determine the 46
			 transformation matrices required for all data runs */
			for (sgp->sgw_lcap = sgp->sgw_lca0, i = 0;
			 i <= 45; i++, sgp->sgw_lcap++)
			  mm3(sgp->sgw_qk[i], (sgp->sgw_lcap)->lc_cel, rm00);
		  }
	   }
	if (star->sta_flags & SP_VERBOSE)
	  {
		fprintf(star->sta_fpl,
		 "transformation matrices for the lattice characters:\n");
		for (sgp->sgw_lcap = sgp->sgw_lca0, i = 0;
		 i < 45; i++, (sgp->sgw_lcap)++)
		   {
			lcap = sgp->sgw_lcap;	j = lcap->lc_latchar;	
			fprintf(star->sta_fpl,
 " %2d %2d %s %s %5.1f %5.1f %5.1f %5.1f %5.1f %5.1f %5.1f %5.1f %5.1f\n",
			 i, j, glatt4(j), gcentering(j),
			 sgp->sgw_qk[i][0][0], sgp->sgw_qk[i][0][1],
			 sgp->sgw_qk[i][0][2], sgp->sgw_qk[i][1][0],
			 sgp->sgw_qk[i][1][1], sgp->sgw_qk[i][1][2],
			 sgp->sgw_qk[i][2][0], sgp->sgw_qk[i][2][1],
			 sgp->sgw_qk[i][2][2]);
		   }
	  }
	/* (gulp) reserve space for the FASTOB observations that will
	 be used for reindexing */
	if (NULL == (sgp->sgw_fobt = (FASTOB *)callop(2 * (xlp->xt_ninteg +
	 100), sizeof (FASTOB), "sgp->sgw_fobt in spaceghead")))
	  {
		fprintf(stderr,
		 "Failure reserving memory for FASTOBs in spaceghead\n");
		return EALLOC;
	  }
		/* second batch */
	sgp->sgw_fnext = sgp->sgw_fobt + xlp->xt_ninteg + 50;
	/* resmax *= 0.995; */
	if (star->sta_resstate > star->sta_resmax)
		star->sta_resmax = star->sta_resstate;
	/* convert "resmax" into sin(theta)/lambda terms */
	star->sta_resmax = 0.5 / star->sta_resmax;
	/* determine <omega> for each scaling group */
	(void)get_omcen(fp, star->sta_shlp, (int)(star->sta_nscales), 0);
	if ((star->sta_flags & SP_VERBOSE) || (NULL != (star->sta_checks)))
	  {
		fprintf(star->sta_fpl,
		 "    X    Y   omega  diffvx  diffvy  diffvz");
		fprintf(star->sta_fpl,
	 "  localh  localk  locall    observed hkl    multiref hkl\n");
	  }
	sgp->sgw_nunreliable = sgp->sgw_nrscaled = sgp->sgw_nobscaled = 0;
	/* figure out which lattice is associated with this spacegroup */
	curl = xlp->xt_isys;
		/* kludge for rhombohedral indexed hexagonally */
	if ((curl == 6) && ((star->sta_spgrp == 155) ||
		(star->sta_spgrp == 146))) curl = 7;
	curc = ex2c[(int)(excluc)];
	if ((star->sta_spgrp == 255) || (star->sta_spgrp == 246)) curc = 6;
	for (fch = -1, i = 0, sgp->sgw_lcap = sgp->sgw_lca0;
	 i <= 45; i++, sgp->sgw_lcap++)
	   {
		curch = sgp->sgw_lcap->lc_latchar;
		if ((curl == lattcode(curch)) && (curc == centercode(curch)))
		  {
			fch = curch;	break;
		  }
	   }
	if (fch == -1)
	  {
		fprintf(stderr,
		 "Can't identify lattice: curl, excluc, curc = %2d, %2d, %d\n",
			curl, (int)excluc, curc);
		return EBADVALUE;
	  }
	if (i > 45)
	  {
		fprintf(stderr, "lattice code %d greater than 44\n", i);
		return EBADVALUE;
	  }
	/* convert to the appropriate symmetry of this cell.
	 Remember the unit cell for the first run. */
	for (rup = star->sta_rub, runno = 0; runno < xlp->xt_nruns;
	 runno++, rup++)
	  {
		mm3(xp->xareal, sgp->sgw_qk[i], rup->rp_ar0);
		if (!runno)
		  {
			for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
				xp->xareal[i][j] = rup->rp_areal[i][j];
			for (i = 0; i < 3; i++) xp->xstart[i] = kstart[i];
			if (EOK != (i = rp2vecs(xp)))
			  {
				fprintf(stderr,
				 "Run %d: Failure %d in rp2vecs\n", runno, i);
				return EBADVALUE;
			  }
		  }
	   }
	return EOK;
}

static void log_diffv(FILE *fp, ABBROB *abp, double *ave,
	double loch[], u_char uliv[], u_char mri[])
{ /* summarizes differences between input and computed reflection indices */
	fprintf(fp,
 "%5d%5d%8.2f%8.4f%8.4lf%8.4lf%8.3lf%8.3lf%8.3lf%4d%4d%4d%4d%4d%4d%4d%4d\n",
	 abptox(abp), abptoy(abp), 0.01 * (double)(abp->abo_omob),
	 ave[0], ave[1], ave[2],
	 loch[0], loch[1], loch[2],
	 (int)(uliv[0]), (int)(uliv[1]), (int)(uliv[2]), (int)(uliv[3]),
	 (int)(mri[0]), (int)(mri[1]), (int)(mri[2]), (int)(mri[3]));
}

static int onlist(u_char mri[], STATRUN *star)
{ /*	This returns 1 if the reflection mri is on the list of test refls */
	int	iline;
	u_char	**chp;

	if (star->sta_nlines < 1) return 0;
	for (iline = 0, chp = star->sta_checks; iline < star->sta_nlines;
	 iline++, chp++) if (!hklcmp(*chp, mri)) return 1;
	return 0;
}

static int indexchk(u_char *mrind, ABBROB *aobp, float amat[3][3],
		double ave[], double loch[], STATRUN *star, SPGWORK *sgp)
{ /*	This looks at `ave' in the event that the integerized version
	didn't originally agree with the multiref hkl. If we were close
	to a half-integer value, so that we might slip over a boundary,
	we use this routine to fudge `ave' a little to make things fit.
	The routine returns 1 if we successfully get a fit with the
	multiref value; it returns 2 if we had to fudge it to get it to fit;
	it returns 0 if it fails even after fudging */
	int	i, i1, i2, j, m, mm;
	short	ilv[3], tilv[3];
	u_char	uliv[4];
	double	ad, maxdel, nextdel, delh[3], xfl[3];
	double	dmat[3][3], mati[3][3];

	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		dmat[i][j] = amat[i][j];
	dmm31(delh, dmat, ave);
	for (j = 0; j < 3; j++) loch[j] = delh[j];
	/* multiply by scaleup matrix */
	for (j = 0; j < 3; j++)
	   {
		xfl[j] = ((double)sgp->sgw_iqtop[j][0]) * loch[0] +
			((double)sgp->sgw_iqtop[j][1]) * loch[1] +
			((double)sgp->sgw_iqtop[j][2]) * loch[2];
		ilv[j] = (xfl[j] < 0.) ? xfl[j] - 0.49999 : xfl[j] + 0.49999;
	   }
	/* create unsigned-char version thereof */
	mpack(ilv, uliv, star->sta_spgrp);
	/* symmetry-transform those */
	symtr(uliv, star->sta_spgrp);
	if (!hklcmp(mrind, uliv))
	  {
		if (onlist(mrind, star))
		  log_diffv(star->sta_fpl, aobp, ave, delh, uliv, mrind);
		return 1;	/* oh good. It matches */
	  }
	/* if we didn't get agreement, we have to work a little harder */
	/* figure out which index is furthest from an integer */
	for (maxdel = j = 0, i = -1; j < 3; j++)
	   {
		delh[j] = ((double)(ilv[j])) - xfl[j];
		if ((ad = fabs(delh[j])) > maxdel)
		  {
			maxdel = ad;	i = j;
		  }
	   }
	i1 = i + 1;	if (i1 > 2) i1 = 0;	i2 = 3 - i - i1;
	nextdel = fabs(delh[i1]);
	if ((ad = fabs(delh[i2])) > nextdel)
	  {
		i1 = i2;	nextdel = ad;	i2 = 3 - i - i1;
	  }
	if ((i < 0) || (i > 2))
	  {
		fprintf(stderr, "Tilt! i = %d\n", i);	return 1;
	  }
	if ((i1 < 0) || (i1 > 2))
	  {
		fprintf(stderr, "Tilt! i1 = %d\n", i1);	return 1;
	  }
	if ((i2 < 0) || (i2 > 2))
	  {
		fprintf(stderr, "Tilt! i2 = %d\n", i2);	return 1;
	  }
	/* try all eight combinations, in order */
	for (j = 1; j < 8; j++)
	   {
		for (m = 0; m < 3; m++) tilv[m] = ilv[m];
		/* we adjust the worst, the next-worst, and / or the best
		 depending on the value of j */
		if (j & 01) tilv[i] += (delh[i] < 0.) ? 1 : -1;
		if (j & 02) tilv[i1] += (delh[i1] < 0.) ? 1 : -1;
		if (j & 04) tilv[i2] += (delh[i2] < 0.) ? 1 : -1;
		/* having thus mangled ilv[], we pack and symm-transform it */
		mpack(tilv, uliv, star->sta_spgrp);
		symtr(uliv, star->sta_spgrp);
		/* compare the mangled version to the multiref. If it
		 matches, we found a hit! */
		if (!hklcmp(mrind, uliv))
		  {
			if (j & 01) xfl[i] += (delh[i] < 0.) ? 0.5 : -0.5;
			if (j & 02) xfl[i1] += (delh[i1] < 0.) ? 0.5 : -0.5;
			if (j & 04) xfl[i2] += (delh[i2] < 0.) ? 0.5 : -0.5;
			/* transform xfl back to the original (high) symm */
			for (m = 0; m < 3; m++)
			   {
				for (ad = mm = 0; mm < 3; mm++)
					ad += sgp->sgw_qkz[m][mm] * xfl[mm];
				delh[m] = ad;
			   }
			/* remember inverse of the unit-cell matrix */
			if (dinv3(dmat, mati)) return 0; /* shouldn't happen! */
			dmm31(ave, mati, delh); /* modified version of ave */
			for (m = 0; m < 3; m++) loch[m] = delh[m];
			if (onlist(mrind, star)) log_diffv(star->sta_fpl,
				aobp, ave, delh, uliv, mrind);
			return j+1;
		  }
	   }
	/* if we get here, we blew it */
	if (star->sta_flags & SP_VERBOSE) log_diffv(star->sta_fpl,
		aobp, ave, delh, uliv, mrind);
	return 0;
}

static void mrtofast(STATRUN *star, SPGWORK *sgp)
{ /*	This reads in reflections from a multiref file.
	It verifies that they are indexed correctly as it converts the
	observations into FASTOB structures in the current spacegroup.
	It returns EOK if successful, various codes if not */
	register int	i, j;
	register FASTOB	*fobp;
	int		icout, nin, shadd, nrely[8];
	u_char		lohkl[4];
	double		avec[3], lochv[3];
	CRYSTAL		*xlp;
	RUNWORK		*rup;
	ABBROB		*aobp, *aobq;
	MERGEDREF	mr;

	xlp = star->sta_xhdp;
	fobp = sgp->sgw_fobt; /* initialize pointer to FASTOB structures */
	for (i = 0; i < 8; i++) nrely[i] = 0;
	for (nin = 0; 0 < xgread((void *)&mr, XG_MERGEDREF, star->sta_fpmi);
	 nin++)
	   { /* for each reflection, examine all the observations */
		for (j = 0; j < 4; j++) lohkl[j] = mr.mr_hkl[j];
		for (i = 0, aobp = &(mr.mr_ob[0]);
		 i < mr.mr_nob; i++, aobp++, fobp++)
		  if (aobp->abo_shift >= 0)
		   {
			if (i < mr.mr_nobpri) lohkl[3] &= 0376;
			else lohkl[3] |= 01;
			for (j = 0; j < 4; j++) /* remember hkl */
				fobp->fo_hkl[j] = mr.mr_hkl[j];
			/* record whether it's Bijvoet or not.
			 If it is, set nobbij to 1 and set the Bijvoet
			 bit in the hkl sign byte. If not,
			 set nobpri to 1 and clear the Bijvoet bit */
			if (i >= mr.mr_nobpri)
			  {
				fobp->fo_hkl[3] |= 01;
				fobp->fo_nobpri = 0; fobp->fo_nobbij = 1;
			  }
			else {
				fobp->fo_nobpri = 1; fobp->fo_nobbij = 0;
				fobp->fo_hkl[3] &= 0376;
			     }
			/* remember how many observations there
			 are for the refl as a whole in fo_nob
			 of the first observation */
			if (!i) fobp->fo_nob = mr.mr_nob;
			else fobp->fo_nob = 0; /* others: set to 0 */
			/*remember <inten>,<sigma>,stl(mostly unused)*/
			fobp->fo_ibar = mr.mr_ibar;
			fobp->fo_isig = mr.mr_isig;
			fobp->fo_stl = mr.mr_stl;
			/*copy observation info in */
			cpymem((AVOID *)&(fobp->fo_ob), aobp,
				sizeof (ABBROB));
			if (aobp->abo_sigma < 0) continue;
			/* remainder of manipulations should be on the
			 `fobp' version */
			aobq = &(fobp->fo_ob);
			/* figure out which run this is. We don't try
			 to use deleted observations here at all */
			shadd = aobq->abo_shift;
			shadd = 200 * (int)(shadd / 200);
			for (rup = star->sta_rub, j = 0; j < xlp->xt_nruns;
			 rup++, j++) if (rup->rp_shadd == shadd) break;
			if (j >= xlp->xt_nruns)
			  {
				fprintf(star->sta_fpl,
 "Obs %2d of [%3d,%3d,%3d,%3d] has shiftno = %4d--outside runlist--set to 0\n",
   i, mr.mr_hkl[0], mr.mr_hkl[1], mr.mr_hkl[2], mr.mr_hkl[3], aobq->abo_shift);
				rup = star->sta_rub;
			  }
			/* now see whether the recalculated hkl is
			 the same as the original. If not, flag it*/
			if (1 == refvec(aobq, rup,
			 SC_DIFFRACTION | SC_VECONLY, 0, avec))
			  {
				if (!(icout = indexchk(lohkl, aobq,
				 rup->rp_ar0, avec, lochv, star, sgp)))
		 		  { /* flag this as unreliable */
					aobq->abo_chamno |= 0100;
					(sgp->sgw_nunreliable)++;
				 }
				else if ((1 <= icout) && (icout <= 8))
					(nrely[icout-1])++;
			  } /* end of "if (1 == refvec" clause */
		   } /* end of loop through observations */
	     } /* end of loop through entire reflections */
	i = nrely[0] + nrely[1] + nrely[2] + nrely[3] + nrely[4] + nrely[5]
		+ nrely[6] + nrely[7];
	printf(
	 " %6d reflections input: %6d reliable, %6d unreliable observations\n",
	 	nin, i, sgp->sgw_nunreliable);
	/* report on reflections for which we had to jog the indices around
		a little--but only if there were some to which we did that. */
	if (i > nrely[0])
	  {
		printf(
	 " Number of reliable observations that required transformations:\n");
		printf("       0     1     2     3     4     5     6     7\n");
		printf(" %7d%6d%6d%6d%6d%6d%6d%6d\n",
		 nrely[0], nrely[1], nrely[2], nrely[3],
		 nrely[4], nrely[5], nrely[6], nrely[7]);
	  }
	sgp->sgw_fend = fobp;	/* remember where the end of the list is */
	/* let's flag 'sgp->sgw_fend' in a distinctive way */
	zeromem((AVOID *)sgp->sgw_fend, sizeof (FASTOB));
		/* not a value that would normally be legal: */
	(sgp->sgw_fend)->fo_hkl[3] = 254;
	(void)fclose(star->sta_fpmi);	star->sta_fpmi = NULL;
}

static void sp_rstat(FASTOB *fbs, STATRUN *star, SPGWORK *sgp)
{ /*	This computes the statistics for a reflection.
	The R-factor and reflection-redundancy calculations
	apply to merged reflections */
	register int	i,j,n;
	register ABBROB	*aobp;
	register FASTOB	*fbp, *fbq;
	int		indx, io, nobs, run_num, seenit;
	short		ip, obsmr, lphkl[3];
	double		fbar, fdiff, fij, gw, idiff, iij, ipscale, iscale, sc;
	double		sigij, sigmay, sk, wij, wijsq, yden, yovsigy, yy, *fp2;
	double		avec[MAXVECPARS], r[6], s[6], skf[MAXOB], ski[MAXOB];
	float		*fp1;
	xgulong		ulsh;
	RUNWORK		*rup;
	SHIFT		*tsh;
	SHLSTAT		*ashp;
	static sichar	runix[MAXOB];
	static double	ysmin[NYSRANGE+1] = {0.,2.,5.,10.,20.,40.,60.,0.};

	munpack(fbs->fo_hkl, lphkl, star->sta_spgrp);	/* unpack the hkl */
	star->sta_res = fbs->fo_stl;
	/* Get the R-factor sums for the individual observations */
	nobs = fbs->fo_nob;
	for (fbp = fbs, yy = yden = obsmr = indx = 0; indx < nobs;indx++,fbp++)
	   { /* Statistics computed independent of shift status */
		ski[indx] = DELETED_FLAG - 1; runix[indx] = -1;
		aobp = &(fbp->fo_ob);
		io = star->sta_runindex[aobp->abo_runno];
		if ((io >= 0) && (aobp->abo_shift >= 0)) /*ignore deleted obs*/
		  {
			if (aobp->abo_sigma > 0) /* respectable sigma */
			  {
				runix[indx] = io;
				ulsh = LSH(aobp->abo_runno, 24) |
				 LSH(aobp->abo_chamno, 16) | aobp->abo_shift;
				if (0 <= (j = find_shift(star->sta_shlp,
				 ulsh, (int)(star->sta_nscales))))
				  {
					tsh = star->sta_shlp + j;
					sk = terpscale(aobp->abo_omob, tsh,
						(double)(star->sta_res));
					run_num = aobp->abo_shift / 200;
					rup = star->sta_rub + run_num;
					if (star->sta_vshp)
					  {
						(void)refvec(aobp, rup,
				star->sta_flags, star->sta_nvpars, avec);
						fp1 = &((star->sta_vshp +
						 run_num)->v_coef[0]);
						for (j = star->sta_nvpars,
						 fp2 = &(avec[0]), gw = 0;
					  	 j--; fp1++, fp2++)
							gw += *fp1 * *fp2;
						sk *= gw;
					  } /* end of "if (vshp" clause */
					ski[indx] = sk;
					skf[indx] = (sk < 0 ? 0. : sqrt(sk));
					wijsq = aobp->abo_sigma;
					wijsq = sk / (wijsq * wijsq);
					yy += wijsq * aobp->abo_ii;
					yden += wijsq * sk;	obsmr++;
				  } /* end of "if (0 <= (j" clause */
			  } /* end of "if (aobp->abo_sigma ..." clause */
		  } /* end if "if ((io >= 0)..." clause */
	   } /* end of loop through the observations */
	if (!obsmr) /* all observations were marked for deletion */
	  {
		(star->sta_allmark)++;	return;
	  }
	/* Increment some counters */
	ashp = &(star->sta_allshl);
	ashp->sh_resol += star->sta_res;	(ashp->sh_ncoll)++;
	ashp->sh_nob += obsmr;			j = min2(obsmr, NOBMAX+1);
	(ashp->sh_nquan[j])++;
	/* if we get here with yden == 0, it means all the observations
	were legitimate but came from shifts that have no scale factor
	associated with them: the shifts don't appear in the scalein file*/
	if (yden < DFUZZ) return;
	yy /= yden;	sigmay = sqrt(1. / yden);
	fbar = (yy < 0. ? 0. : sqrt(yy)); /* ave.struct.amplit.*/
	ashp->sh_ybar += yy;	yovsigy = (sigmay <= 0. ? 0. : yy / sigmay);
	ashp->sh_yovsig += yovsigy;
	for (ip = 0; ip < NYSRANGE; ip++) if (yovsigy <= ysmin[ip]) break;
	ashp->sh_nyosig[ip]++;
	if (yy <= 0. || obsmr < 2) return; /* must have <I> > 0, n > 1 */
	for (seenit = i = 0, fbp = fbs; i < nobs; i++, fbp++)
	 {
	   aobp = &(fbp->fo_ob);
	   if (((sc = ski[i]) > DFUZZ) && (aobp->abo_sigma > FUZZ) &&
	    (runix[i] >= 0))
	     {
		iij = aobp->abo_ii;		fij = (iij < 0? 0 : sqrt(iij));
		sigij = aobp->abo_sigma;	wij = 1./(sigij*sigij);
		idiff = yy * sc - iij;		fdiff = fbar * skf[i] - fij;
		iscale = iij / sc;
		r[1] = wij * idiff * idiff;	s[1] = wij * iij * iij;
		r[2] = idiff * idiff;		s[2] = iij * iij;
		r[3] = fabs(idiff);		s[3] = fabs(iij);
		r[4] = fabs(fdiff);		s[4] = fij;
		r[5] = 0.;			s[5] = 0.;
		for (ip = 0, fbq = fbs; ip < nobs; ip++, fbq++)
		   if ((ip != i) && (ski[ip] > DFUZZ) &&
			(fbq->fo_ob.abo_sigma > FUZZ))
		     {
			ipscale = fbq->fo_ob.abo_ii / ski[ip];
			r[5] += fabs(iscale - ipscale);
			s[5] += 0.5 * (iscale + ipscale);
		     }
		r[0] = r[1];			s[0] = 0;
		for (n = 0; n <= NSRFACS; n++)
		   {
			ashp->sh_rnum[n] += r[n]; ashp->sh_rden[n] += s[n];
		   }
		if (!seenit) (sgp->sgw_nrscaled)++;
		seenit = 1;	(sgp->sgw_nobscaled)++;
	      }
	 } /* end of "for (i = 0; ..." loop */
	return;
}

static void space_header(FILE *fp)
{ /* Print a header for the spacegroup statistics to the stream fp */
	fprintf(fp,
	 "\t\tSummary of agreement statistics in various spacegroups\n");
	fprintf(fp, "Xtal   Space  Mean  Mean");
	fprintf(fp, "\t\t\t\t\tAgreement Statistics\t\tUnit Cell\n");
	fprintf(fp,
 "Sys-   Group  Ref.  Y/sig      Total           Scaled\n");
	fprintf(fp,
 "tem          Inten.        # Refs   # Obs  # Refs   # Obs  FOM");
	fprintf(fp,
 "  G.O.F.     R1     R3      a      b      c alpha  beta gamma resid\n");
}

static void out_sgstat(FILE *fp, SPACEGSTAT *spsp)
{ /*	Summarizes the results for a given spacegroup */
	int	chara;
	SHLSTAT	*shp;
	static char	fmt1[] = 
	 "%s %s%5d%8.1f%6.1f%8d%8d%8d%8d%5.0f%8.1f%7.2f%7.2f";
	static char	fmt2[] = "%7.1f%7.1f%7.1f%6.1f%6.1f%6.1f%6.2f\n";

	chara = spsp->sp_latc;	shp = &(spsp->sp_st);
	fmt1[34] = (spsp->sp_fom < 1.E3) ? 'f' : 'e'; /* conditional format */
	fprintf(fp, fmt1,
	 glatt4(chara), gcentering(chara), (int)spsp->sp_intl,
	 shp->sh_ybar, shp->sh_yovsig, shp->sh_ncoll, shp->sh_nob,
	 shp->sh_nrscaled, shp->sh_nobscaled, 100. * spsp->sp_fom,
	 100.* shp->sh_rnum[0], 100.* shp->sh_rnum[1], 100.* shp->sh_rnum[3]);
        if (spsp->sp_score < 10.)
	  {
		fmt2[33] = '2';	fmt2[34] = 'f';
	  }
	else {
		fmt2[33] = '0';	fmt2[34] = 'e';
	     }
	fprintf(fp, fmt2,
	 spsp->sp_abc[0], spsp->sp_abc[1], spsp->sp_abc[2], spsp->sp_ang[0],
	 spsp->sp_ang[1], spsp->sp_ang[2], 100. * spsp->sp_score);
	return;
}

static void fastrfacs(SPACEGSTAT *spsp, STATRUN *star,
	double symfr, SPGWORK *sgp)
{ /*	This computes the R-factors for the entire list */
	int		j;
	register FASTOB	*fobp;
	FASTOB		*fobs;
	SHLSTAT		*ashp;

	/* walk through the entire list of observations */
	sgp->sgw_nobscaled = sgp->sgw_nrscaled = 0;
	ashp = &(star->sta_allshl);
	zeromem((AVOID *)ashp, sizeof (SHLSTAT));
	for (fobp = fobs = sgp->sgw_fobt; fobp < sgp->sgw_fend; fobp++)
	   {
		if ((fobp > sgp->sgw_fobt) && (fobp->fo_nob > 0))
		  {
			sp_rstat(fobs, star, sgp);	fobs = fobp;
		  }
	   }
	if (fobp->fo_nob > 0) sp_rstat(fobs, star, sgp); /* final reflection */
	ashp->sh_nobscaled = sgp->sgw_nobscaled;
	ashp->sh_nrscaled = sgp->sgw_nrscaled;
	calc_sh(ashp, star);
	spsp->sp_latc = (sgp->sgw_lcap)->lc_latchar;
	spsp->sp_intl = star->sta_spgrp;
	spsp->sp_unreliable = sgp->sgw_nunreliable;
	for (j = 0; j < 3; j++)
	   {
		spsp->sp_abc[j] = (star->sta_xp)->xabc[j];
		spsp->sp_ang[j] = (star->sta_xp)->xang[j];
	   }
	spsp->sp_fom = ashp->sh_rnum[0] * symfr +
	 200. * (sgp->sgw_lcap)->lc_score;
	spsp->sp_score = (sgp->sgw_lcap)->lc_score;
	cpymem((AVOID *)&(spsp->sp_st), (AVOID *)ashp, sizeof (SHLSTAT));
	return;
}

static void out_omissions(FILE *fp, int nun)
{ /*	This announces how many observations would be omitted */
	fprintf(fp, " %6d observations will be excluded from calculations",
		nun);
	fprintf(fp,
 " because their original indices could not be reliably determined\n");
}

static void currtop1(STATRUN *star, SPGWORK *sgp)
{ /*	This converts the unit cell from the current spacegroup to a standard
	P1 spacegroup and alters the multiref indices accordingly. */
	int	j, oldunrely, shadd;
	u_char	lociv4[4];
	short	ilv[3];
	double	avec[3], lochv[3];
	size_t	fonobs;
	CRYSTAL	*xlp;
	ABBROB	*aobp;
	RUNWORK	*rup;
	FASTOB	*foi, *foo;

	xlp = star->sta_xhdp;	oldunrely = sgp->sgw_nunreliable;
	for (foi = sgp->sgw_fobt, foo = sgp->sgw_fnext;
	 foi < sgp->sgw_fend; foi++,foo++)
	   { /* copy over everything in the input FASTOB to the output */
		cpymem((AVOID *)foo, (AVOID *)foi, sizeof (FASTOB));
		/* if necessary, fiddle with the indices */
		aobp = &(foo->fo_ob);
		/* skip observations that have been flagged as non-
		 contributing to the current dataset; also, skip observations
		 for which the indexes cannot be unambiguously determined */
		if ((aobp->abo_chamno & 0100) || (aobp->abo_shift < 0))
			continue;
		shadd = aobp->abo_shift; shadd = 200 * (int)(shadd / 200);
		for (rup = star->sta_rub, j = 0; j < xlp->xt_nruns;
		 rup++, j++) if (rup->rp_shadd == shadd) break;
		if (j >= xlp->xt_nruns) rup = star->sta_rub;
		if (1 == refvec(aobp, rup,
		 SC_DIFFRACTION | SC_VECONLY, 0, avec)) /*get diffrac.vector*/
		  {
			if (!indexchk(foo->fo_hkl, aobp,
			 rup->rp_ar0, avec, lochv, star, sgp))
	 		  { /* we record the differential */
				aobp->abo_chamno |= 0100;
				(sgp->sgw_nunreliable)++;
			  }
			/* if the above worked, then we have the correct
			 diffraction vector and nominal indices. Get the
			 actual primitive-cell P1 indices */
			for (j = 0; j < 3; j++)
			   {
				ilv[j] = (lochv[j] < 0.) ?
				 lochv[j] - 0.49999 : lochv[j] + 0.49999;
			   }
			mpack(ilv, lociv4, (short)1);
			for (j = 0; j < 4; j++) foo->fo_hkl[j] = lociv4[j];
			/* record whether it's primary or Bijvoet */
			if (foo->fo_hkl[3] & 01)
			  {
				foo->fo_nobbij = 1;	foo->fo_nobpri = 0;
			  }
			else {
				foo->fo_nobbij = 0;	foo->fo_nobpri = 1;
			     }
		  }
		else { /* if we can't calc the diffraction vector, flag it */
			aobp->abo_chamno |= 040;
			for (j = 0; j < 4; j++) foo->fo_hkl[j] = 255;
		     }
	   }
	if (sgp->sgw_nunreliable > oldunrely)
	  printf(" %6d additional unreliable reflections on 2nd pass\n",
		sgp->sgw_nunreliable - oldunrely);
	fonobs = foi - sgp->sgw_fobt;	sgp->sgw_foe = foo;
	/* sort the data in hkl order */
	if (fonobs > 1) qsort((AVOID *)sgp->sgw_fnext, (size_t)fonobs,
	 sizeof (FASTOB), (QPE)hklcmp);
	if (sgp->sgw_nunreliable)
	  {
		out_omissions(star->sta_fpl,  sgp->sgw_nunreliable);
		out_omissions(stdout, sgp->sgw_nunreliable);
	  }
}

static void p1tocurr(int spindex, STATRUN *star, SPGWORK *sgp)
{ /*	This transforms the indices from their P1 version to the ones we
	would get in the spacegroup associated with spindex */
	int		i, iscen, j, tchari, tuc;
	u_char		uliv[4], uoiv[4];
	size_t		fonobs;
	short		tspace, tul, ilv[3];
	double		tv, hvo[3], lma[3][3];
	FASTOB		*foi, *foo, *foq;
	static short	ulcco[] =
	{   4,   3,   1,   4,   3,   1,   1,   1,   5,   6,   5,   6,   1,
	    1,   1,   3,   1,   3,   1,   4,   3,   2,   1,   2,   1,   1, 0 };

	/* figure out what spacegroup and lattice character we have */
	star->sta_spgrp = tspace = indtosp(spindex, &tul);
	tuc = ulcco[spindex];
	for (sgp->sgw_tindex = -1, sgp->sgw_lcap = sgp->sgw_lca0, j = 0;
	 j <= 45; j++, sgp->sgw_lcap++)
	   {
		tchari = (sgp->sgw_lcap)->lc_latchar;
		if ((tul == lattcode(tchari)) && (tuc == centercode(tchari)))
		  {
			sgp->sgw_tindex = j;	break;
		  }
	   }
	if (sgp->sgw_tindex == -1)
	  {
		fprintf(stderr,
		 "Can't find lattice character for spacegroup[%2d]=%3d, ",
			spindex, (int)tspace);
		fprintf(stderr, " tul,tuc=%2d,%2d = %s %s\n",
		 (int)tul, tuc, glatfi4(tul), gcenfi(tuc));
		return;
	  }
	tuc = sgp->sgw_tindex;
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		lma[i][j] = sgp->sgw_qk[tuc][i][j]; /* transformation matrix */
	for (foi = sgp->sgw_fnext, foo = sgp->sgw_fobt;
	 foi < sgp->sgw_foe; foi++)
	   if (!(0140 & ((foi->fo_ob).abo_chamno)))
	     { /* copy over everything in the input FASTOB to the output */
		cpymem((AVOID *)foo, (AVOID *)foi, sizeof (FASTOB));
		/* if necessary, fiddle with the indices. In this direction,
		 we expect that the indices themselves can be
		 matrix-transformed instead of having to calculate the
		 diffraction vectors */
		munpack(foo->fo_hkl, ilv, (short)1);
		/* matrix-transform this */
		for (i = 0; i < 3; i++)
		   {
			for (tv = j = 0; j < 3; j++)
				tv += lma[i][j] * (double)(ilv[j]);
			hvo[i] = tv;
		   }
		for (j = 0; j < 3; j++)
		   {
			tv = hvo[j];
			ilv[j] = (tv < 0.) ?  tv - 0.49999 : tv + 0.49999;
		   }
		/* create unsigned-char version thereof */
		mpack(ilv, uliv, tspace);
		for (j = 0; j < 4; j++) foo->fo_hkl[j] = uliv[j];
		/* symmetry-transform those */
		symtr(foo->fo_hkl, tspace);
		iscen = iscentric(foo->fo_hkl, tspace);
		/* record whether it's primary or Bijvoet */
		if (foo->fo_hkl[3] & 01) /* Bijvoet bit is on */
		  {
			foo->fo_nobbij = 1;	foo->fo_nobpri = 0;
			if (iscen) fprintf(star->sta_fpl,
 "spg %3d:ilv,xform,final=[%4d,%4d,%4d],[%4d,%4d,%4d,%1d],[%4d,%4d,%4d,%1d]\n",
			 tspace, ilv[0], ilv[1], ilv[2], 
			 uliv[0], uliv[1], uliv[2], uliv[3],
			 foo->fo_hkl[0], foo->fo_hkl[1], foo->fo_hkl[2],
			 foo->fo_hkl[3]);
		  }
		else {
			foo->fo_nobbij = 0;	foo->fo_nobpri = 1;
		     }
		foo->fo_nob = 1; /* at this stage all refls are separate */
		foo++;	/* advance the output pointer */
	     }
	fonobs = foo - sgp->sgw_fobt;
	sgp->sgw_fend = foo;	/* remember end of list */
	/* sort the data in hkl order */
	if (fonobs > 1) qsort((AVOID *)sgp->sgw_fobt, fonobs,
		sizeof (FASTOB), (QPE)hklcmp);
	/* reorganize so that we recognize where reflections begin */
	foo = sgp->sgw_fobt;
	/* remember the unsigned char indices of the 1st reflection */
	for (j = 0; j < 4; j++) uliv[j] = foo->fo_hkl[j];
	iscen = iscentric(uliv, tspace);
	if (iscen && (uliv[3] & 01)) fprintf(star->sta_fpl,
	 "spg %3d:uliv out=[%4d,%4d,%4d,%1d]\n",
			 tspace, uliv[0], uliv[1], uliv[2], uliv[3]);
	uliv[3] &= 0376;
	/* go through all the other reflections, marking the number of
	 observations in the reflection appropriately */
	for (foi = foo + 1; foi < sgp->sgw_fend; foi++)
	   {
		for (j = 0; j < 4; j++) uoiv[j] = foi->fo_hkl[j];
		if ((uoiv[0] != 255) && (uoiv[1] != 255) && (uoiv[2] != 255))
		  {
			iscen = iscentric(uoiv, tspace);
			if (iscen && (uoiv[3] & 01)) fprintf(star->sta_fpl,
 "spg %3d:uliv,uoiv=[%4d,%4d,%4d,%1d],[%4d,%4d,%4d,%1d]\n",
			 tspace, uliv[0], uliv[1], uliv[2], uliv[3],
			 uoiv[0], uoiv[1], uoiv[2], uoiv[3]);
		  }
		uoiv[3] &= 0376;
		if (hklcmp(uliv, uoiv))
		  {
			foo->fo_nob = foi - foo;
			for (foq = foo + 1; foq < foi; foq++) foq->fo_nob = 0;
			foo = foi;	foo->fo_nob = 1;
			for (j = 0; j < 4; j++) uliv[j] = foo->fo_hkl[j];
			if ((uliv[0] != 255) && (uliv[1] != 255) &&
			 (uliv[2] != 255))
			  {
				iscen = iscentric(uliv, tspace);
				if (iscen && (uliv[3] & 01))
				 fprintf(star->sta_fpl,
 "spg %3d:uoiv,uliv =[%4d,%4d,%4d,%1d],[%4d,%4d,%4d,%1d]\n",
				 tspace, uoiv[0], uoiv[1], uoiv[2], uoiv[3],
				 uliv[0], uliv[1], uliv[2], uliv[3]);
			  }
			uliv[3] &= 0376;
		  }
	   }
	/* define the unit cell parameters */
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		(star->sta_xp)->xareal[i][j] = (sgp->sgw_lcap)->lc_cel[i][j];
	if (0 != (i = rp2vecs(star->sta_xp))) fprintf(stderr,
		 "Warning: unit cell produced error %3d in rp2vecs\n", i);
}

static int endspace(int irv, STATRUN *star, SPGWORK *sgp)
{ /*	This closes off a run of space_g cleanly */
	memzap((AVOID **)&(sgp->sgw_fobt), "sgp->sgw_fobt in endspace");
	if (NULL != star->sta_checks)
	  {
		memzap((AVOID **)star->sta_checks,
			"*star->sta_checks in endspace");
		memzap((AVOID **)&star->sta_checks,
			"star->sta_checks in endspace");
	  }
	memzap((AVOID **)&(star->sta_vshp), "star->sta_vshp in endspace");
	memzap((AVOID **)&(star->sta_rub), "star->sta_rub in endspace");
	memzap((AVOID **)&(star->sta_shlp), "star->sta_shlp in endspace");
	headmem(NULL, NULL, 0);
	memzap((AVOID **)&(star->sta_xp), "star->sta_xp in endspace");
	if (NULL != star->sta_fpmi) (void)fclose(star->sta_fpmi);
	if ((NULL != star->sta_fpl) && (star->sta_fpl != stdout))
		(void)fclose(star->sta_fpl);
	star->sta_fpmi = star->sta_fpl = NULL;
	if (NULL != sgp)
	  {
		if (NULL != sgp->sgw_fpmo) (void)fclose(sgp->sgw_fpmo);
		sgp->sgw_fpmo = NULL;
	  }
	globalkill();	return irv;
}

static int spcmp_fom(SPACEGSTAT *sp1, SPACEGSTAT *sp2)
{ /*	Compare routine based on figure-of-merit; this is
	the goodness-of-fit multiplied by a factor expressive of symmetry,
	plus a multiple of the Bravais-lattice residual. */
	return (sp1->sp_fom < sp2->sp_fom) ? -1 : 1;
}

static void head_screw(FILE *fp)
{ /* Prints a header explaining the screw-axis calculations to stream fp */

	fprintf(fp,
 "Screw-axis calculations for point group with lowest figure of merit:\n");
	fprintf(fp,
 " axis multi-    allowed observations disallowed observations\n");
	fprintf(fp,
 "      plicity # obs N(I>4s) <I/sigma> #obs N(I>4s) <I/sigma>\n");
}

static void anal_axial(int whichax, int multiple, STATRUN *star, SPGWORK *sgp)
{ /*	This goes through the data in the current form and
	analyzes the probability of a screw axis along `whichax' (0==h, 1==k,
	 2==l) where we contrast q=kn, q=h,k,or l, and k=`multiple'. */
	int	disallow, n4sallow, n4sdisal, nallow, ndisal, oi, oj, ok, voi;
	double	ios, iosallow, iosdisal;
	char	axc[] = "hkl";
	FASTOB	*fobp;
	ABBROB	*aobp;

	if (!(star->sta_nlines))
	  {
		head_screw(stdout);	head_screw(star->sta_fpl);
		star->sta_nlines = -1;
	  }
	oi = whichax;	oj = whichax + 1;	if (oj > 2) oj = 0;
	ok = 3 - oi - oj;
	iosallow = n4sallow = nallow = 0;
	iosdisal = n4sdisal = ndisal = 0;
	for (fobp = sgp->sgw_fobt; fobp < sgp->sgw_fend; fobp++)
	   {
		if ((fobp->fo_hkl[oj] == 0) && (fobp->fo_hkl[ok] == 0))
		  {
			voi = fobp->fo_hkl[oi];
			aobp = &(fobp->fo_ob);
			disallow = voi % multiple;
			ios = (aobp->abo_sigma <= FUZZ) ? aobp->abo_ii :
				aobp->abo_ii / aobp->abo_sigma;
			if (disallow)
			  {
				ndisal++;	iosdisal += ios;
				if (ios > 4.) n4sdisal++;
			  }
			else {
				nallow++;	iosallow += ios;
				if (ios > 4.) n4sallow++;
			     }
		  }
	   }
	if (nallow) iosallow /= (double)nallow;
	if (ndisal) iosdisal /= (double)ndisal;
	fprintf(stdout, "    %c %6dn %5d %7d %9.3f %4d %7d %9.3f\n",
	 axc[whichax], multiple,
	 nallow, n4sallow, iosallow, ndisal, n4sdisal, iosdisal);
	fprintf(star->sta_fpl, "    %c %6dn %5d %7d %9.3f %4d %7d %9.3f\n",
	 axc[whichax], multiple,
	 nallow, n4sallow, iosallow, ndisal, n4sdisal, iosdisal);
}

static void screws(SPACEGSTAT *spp, STATRUN *star, SPGWORK *sgp)
{ /*	This analyzes the possibility of a screw axis in a given spacegroup.
	We aren't dealing with centrosymmetric cases yet.
	Jan 2004: rhombohedrals are ugly and don't merit our attention */
	int	i;

	i = spp->sp_intl;
	star->sta_nlines = 0;	/* flag: we need a header in the output */
	/* cubic, not  P23, P2(1)3: check (h00), h = 4n */
	if ((196 <= i) && (i <= 230) && (i != 198)) anal_axial(0, 4, star,sgp);
	/* cubic with some exceptions: check (h00), h = 2n */
	if ((195 <= i && i <= 230) && ((i != 197) && (i != 199) && (i != 196)
		&& (i != 209) && (i != 210) && (i != 211) && (i != 214)))
			anal_axial(0, 2, star, sgp);
	/* hexagonal: check (00l), l=6n, 3n, and 2n */
	if ((168 <= i) && (i <= 194))
	  {
		anal_axial(2, 6, star, sgp); /* always check l=6n */
		anal_axial(2, 3, star, sgp); /* and l=3n */
		anal_axial(2, 2, star, sgp); /* and l=2n */
	  }
	/* trigonal: check (00l), l=3n */
	if ((143 <= i) && (i <= 167)) anal_axial(2, 3, star, sgp);
	/* tetragonal: check (00l), l=2n or 4n, and sometimes 0k0, k=2n */
	if ((75 <= i) && (i <= 142))
	  {
		anal_axial(2, 4, star, sgp);
		if ((i != 79) && (i != 80) && (i != 97) && (i != 98))
			anal_axial(2, 2, star, sgp);
		if ((89 <= i) && ((i != 97) && (i != 98)))
			anal_axial(1, 2, star, sgp);
	  }
	/* orthorhombic: always check (h00),(0k0),(00l), l=2n, etc. */
	if (((16 <= i) && (i <= 74)) && ((22 > i) || (i > 24)))
	  {
		anal_axial(2, 2, star, sgp);
		if ((i != 20) && (i != 21))
		  {
			anal_axial(1, 2, star, sgp);
			anal_axial(0, 2, star, sgp);
		  }
	  }
	/* monoclinic: check 0k0, k=2n except in C-centered */
	if (((2 <= i) && (i <= 15)) && (i != 5)) anal_axial(1, 2, star, sgp);
}

static int modxtal(STATRUN *star, SPACEGSTAT *spsp,
 short new_sp, short new_sys, SPGWORK *sgp)
{ /*	Makes appropriate modifications to the CRYSTAL header and then
	writes it out. This returns EOK if successful, various I/O-related
	error codes if not */
	int	i;
	CRYSTAL	*xlp;
	SHLSTAT	*shp;

	shp = &(spsp->sp_st);	xlp = star->sta_xhdp;
	xlp->xt_intl = new_sp;
	xlp->xt_stdpos[0] = (star->sta_xp)->xdelom;
	xlp->xt_stdpos[1] = (star->sta_xp)->xdelchi;
	xlp->xt_stdpos[2] = (star->sta_xp)->xdelphi;
	strcpy(xlp->xt_spaceg, spacegname(new_sp));
	xlp->xt_isys = new_sys;	xlp->xt_ninteg = sgp->sgw_fend - sgp->sgw_fobt;
	xlp->xt_nmerged = shp->sh_ncoll;
	if (0 != (i = rp2vecs(star->sta_xp))) return i;
	/* we'll have to fix xt_nrefls later */
	for (i = 0; i < 3; i++)
	   {
		xlp->xt_ar[i] = (star->sta_xp)->xabc[i];
		xlp->xt_ang[i] = (star->sta_xp)->xang[i];
	   }
	/* shazam! It's time to create and open the output multiref */
	if (NULL == (sgp->sgw_fpmo = xgopen("mrefout", "w"))) return EOPENERR;
	if (verror(0 < xgwrite((void *)xlp, XG_CRYSTAL, sgp->sgw_fpmo),
		"write crystal header to output")) return EWRITERR;
	return EOK;
}

static int moddrun(STATRUN *star, SPGWORK *sgp)
{ /*	This modifies and outputs the DATARUN structures.
	We have to keep track of the reference position associated
	with the unit-cell matrix as we do it */
	int	i, j, m;
	float	radstart[3];
	CRYSTAL	*xlp;
	DATARUN	*rhp;
	RUNWORK	*rup;

	/* for each datarun, we regenerate the orientation matrix */
	xlp = star->sta_xhdp;	rhp = star->sta_rhdp;
	for (j = 0, rup = star->sta_rub; j < xlp->xt_nruns; j++, rhp++, rup++)
	   { /* apply the spacegroup-specific transform */
		mm3(rup->rp_areal, sgp->sgw_qk[sgp->sgw_tindex], rup->rp_ar0);
		for (i = 0; i < 3; i++) for (m = 0; m < 3; m++)
			(star->sta_xp)->xareal[i][m] = rup->rp_areal[i][m];
		/* transform back by the local rotation */
		for (i = 0; i < 3; i++)
		   {
			(star->sta_xp)->xstart[i] = rhp->run_startpos[i];
			radstart[i] = (star->sta_xp)->xstart[i] * RADC;
		   }
		rot9r((star->sta_xp)->xareal, radstart);
		/* then get the recip-space matrix from that */
		if ((i = rp2vecs(star->sta_xp)))
		  {
			fprintf(stderr, "Error %d in rp2vecs in moddrun\n", i);
			return EBADVALUE;
		  }
		if (!i)
		  {
			for (i = 0; i < 3; i++) for (m = 0; m < 3; m++)
			   rhp->run_recip[i][m] = (star->sta_xp)->xrec[i][m];
		  }
		if (verror(0 < xgwrite((void *)rhp, XG_DATARUN, sgp->sgw_fpmo),
		 "write datarun header to output")) return EWRITERR;
	   }
	return EOK;
}

static int fobtomref(FASTOB *fbp, FASTOB *fbq, STATRUN *star, SPGWORK *sgp)
{ /*	We construct from the observations *fbp, *(fbp+1), ... *(fbq-1)
	 a MULTIREF structure and then write it out */
	int		iscen, j;
	double		meani, meanvar, wt;
	ABBROB		*aobp, *aobq;
	FASTOB		*fbz;
	MERGEDREF	mr;

	/* Get the reflection indices */
	for (j = 0; j < 4; j++) mr.mr_hkl[j] = fbp->fo_hkl[j];
	mr.mr_nob = fbq - fbp; /* count the observations */
	iscen = iscentric(fbp->fo_hkl, star->sta_spgrp);
	/* figure out how many of the observations are primaries */
	if (fbp->fo_hkl[3] & 01)
	  {
		mr.mr_nobpri = 0;	fbz = fbp;
	  }
	else {
		for (fbz = fbp; fbz < fbq; fbz++)
		   if (fbz->fo_hkl[3] != fbp->fo_hkl[3])
		     {
			mr.mr_nobpri = fbz - fbp;	break;
		     }
		if (fbz == fbq) mr.mr_nobpri = fbq - fbp;
	    }
	mr.mr_nobbij = mr.mr_nob - mr.mr_nobpri; /* ...the rest are Bijvoets */
	if ((iscen) && (mr.mr_nobpri < mr.mr_nob))
	  {
		fbz = fbp + mr.mr_nobpri;
		fprintf(star->sta_fpl,
 " spg %3d:[%3d,%3d,%3d,%3d] centric,#,pri,bij=%3d,%3d,%3d;",
		(int)(star->sta_spgrp),
		fbp->fo_hkl[0], fbp->fo_hkl[1], fbp->fo_hkl[2], fbp->fo_hkl[3],
		mr.mr_nob, mr.mr_nobpri, mr.mr_nobbij);
		fprintf(star->sta_fpl, "first bad = [%3d,%3d,%3d,%3d]\n",
		fbz->fo_hkl[0], fbz->fo_hkl[1], fbz->fo_hkl[2],fbz->fo_hkl[3]);
	  }
	mr.mr_stl = fbp->fo_stl; /* keep track of sin(theta)/lambda */
	/* accumulate sums to get the <I> and sigma(<I>) values,
	 and copy over the information on the individual observations */
	for (meani = meanvar = 0., fbz = fbp, aobq = &(mr.mr_ob[0]);
	 fbz < fbq; fbz++, aobq++)
	   {
		aobp = &(fbz->fo_ob);
		wt = aobp->abo_sigma;	wt = (wt < DFUZZ) ? 0. : 1./ (wt * wt);
		meanvar += wt;		meani += wt * aobp->abo_ii;
		cpymem((AVOID *)aobq, (AVOID *)aobp, sizeof (ABBROB));
	   }
	/* now compute <I> and sigma<I> */
	if (meanvar < DFUZZ)
	  {
		mr.mr_ibar = 0;	mr.mr_isig = 100;
	  }
	else {
		mr.mr_ibar = (meani <= 0.) ? 0 :
			(int)(meani / meanvar + 0.49999);
		mr.mr_isig = 1. / sqrt(meanvar);
	     }
	/* write the reflection to the output file */
	if (verror(0 < xgwrite((void *)&mr, XG_MERGEDREF, sgp->sgw_fpmo),
	 "write a reflection to multiref")) return EWRITERR;
	return EOK;
}

static int out_modmref(int spindx, STATRUN *star,
	SPACEGSTAT *spsp, SPGWORK *sgp)
{ /*	This writes out a multiref in the best spacegroup,
	provided that the original spacegroup is inconsistent with that one. */
	register FASTOB	*fobp;
	int		j, ret;
	short		new_spgrp, new_sys;
	u_char		currhkl[4], newhkl[4];
	FASTOB		*fobq;

	new_spgrp = indtosp(spindx, &new_sys);
	if (new_spgrp == (star->sta_xp)->xispgp)
	  {
		fprintf(star->sta_fpl,
		 "Final spacegroup %3d is consistent with initial", new_spgrp);
		fprintf(star->sta_fpl, " spacegroup; no multiref output\n");
		return EBADVALUE;
	  }
	if ((new_spgrp < 200) && (new_sys == 7)) new_sys = 6;
	/* modify & output CRYSTAL header structure */
	if (EOK != (ret = modxtal(star, spsp, new_spgrp, new_sys, sgp)))
		return ret;
	/* modify & output DATARUN headers */
	if (EOK != (ret = moddrun(star, sgp))) return ret;
	/* loop through observations. When we encounter one with a
	 flag set, we refuse to write it out */
	for (fobp = sgp->sgw_fobt; fobp < sgp->sgw_fend; )
	 {
	   if (fobp->fo_hkl[0] != 255)
	     {
	       for (j = 0; j < 4; j++) currhkl[j] = fobp->fo_hkl[j];
		currhkl[3] &= 0376;
		for (fobq = fobp + 1; fobq < sgp->sgw_fend; fobq++)
		   {
			if (fobq->fo_hkl[0] == 255) /* bad FASTOB: skip it */
			  {
				if (fobp->fo_hkl[0] != 255)
				  {
					if (EOK != (ret = fobtomref(fobp,
					 fobq, star, sgp))) return ret;
				  }
				fobp = fobq + 1;	break;
			  }
			for (j = 0; j < 4; j++) newhkl[j] = fobq->fo_hkl[j];
			newhkl[3] &= 0376;
			if (hklcmp(currhkl, newhkl))
			  {
				if (EOK != (ret = fobtomref(fobp,
				 fobq, star, sgp))) return ret;
				fobp = fobq; break;
			  }
		   } /* end of "for (fobq " loop */
		/* if we get to here, it means that we've run off the
		 end of the list in the inner loop. */
		fobp = fobq;
	     } /* end of "if (fobp[0] != 255" clause */
	  else fobp++;
	 } /* end of "for (fobp" loop */
	return EOK;
}

int spacegops(STATRUN *star)
{ /*	Performs the actual calculations for `spaceg' */
	register int		i, j;
	int			tchari, tuc, tul;
	short			lspaceg, lsys;
	LATTICESCORE		*lcap;
	SPACEGSTAT		*spsp;
	static SPACEGSTAT	sps[28];
	static SPGWORK		sgs;
	static double		symfac[] =
	{ .30, .34, .38, .34, .38, .42, .50, .60, .60, .60, .70, .70, .60,
	  .52, .70, .44, .54, .54, .58, .59, .59, .59, .60, .77, .80, 1.0, 0.};

	/* process multiref header info */
	if (EOK != (i = spaceghead(star->sta_fpmi, star, &sgs)))
		return endspace(i, star, &sgs);
	/* read MULTIREF in and convert to FASTOB structures*/
	mrtofast(star, &sgs);
	currtop1(star, &sgs);	/* convert FASTOBs to P1 */
	if (star->sta_flags & SP_VERBOSE)
	  {
		fprintf(star->sta_fpl,
		 "Codes and unit-cell matrices for valid choices:\n");
		fprintf(star->sta_fpl," ind cha cho tul tuc latt cen score");
		fprintf(star->sta_fpl,"    real-space unit cell matrix\n");
		for (j = 0, sgs.sgw_lcap = sgs.sgw_lca0;
		 j < 45; j++, (sgs.sgw_lcap)++)
		   if ((sgs.sgw_lcap)->lc_choice > -1)
		     {
			lcap = sgs.sgw_lcap;	tchari = lcap->lc_latchar;
			tul = lattcode(tchari);	tuc = centercode(tchari);
			fprintf(star->sta_fpl,
			 " %3d %3d %3d %3d %3d %4s %1s %8.4f",
			 j, tchari, (int)lcap->lc_choice, tul, tuc,
			 glatfi4(tul), gcenfi(tuc), lcap->lc_score);
			fprintf(star->sta_fpl,
		 " %6.1f %6.1f %6.1f %6.1f %6.1f %6.1f %6.1f %6.1f %6.1f\n",
			 lcap->lc_cel[0][0], lcap->lc_cel[0][1],
			 lcap->lc_cel[0][2], lcap->lc_cel[1][0],
			 lcap->lc_cel[1][1], lcap->lc_cel[1][2], 
			 lcap->lc_cel[2][0], lcap->lc_cel[2][1],
			 lcap->lc_cel[2][2]);
		      }
	  }
	/* for each space group, compute the various R-factors */
	for (j = 0; j < 26; j++)
	   {
		fprintf(stderr, "Processing data from spacegroup %3d\r",
			indtosp(j, &lsys));
		p1tocurr(j, star, &sgs); /* convert P1 -> current spacegroup */
		fastrfacs(&(sps[j]), star, symfac[j], &sgs);
	   }
	fprintf(stderr, "\n");
	/* sort the results on the figure of merit */
	if (j > 1) qsort((AVOID *)&(sps[0]), (size_t)j,
		sizeof (SPACEGSTAT), (QPE)spcmp_fom);
	/* output the sorted results */
	space_header(star->sta_fpl);	space_header(stdout);
	for (spsp = &(sps[0]); j--; spsp++)
	   {
		out_sgstat(star->sta_fpl, spsp); out_sgstat(stdout, spsp);
	   }
	/* now, assuming that the best FOM is the right point group,
	 we attempt to analyze the screw axis differences */
	spsp = &(sps[0]);
	for (j = 0; j < 26; j++)
	   {
		if ((lspaceg = indtosp(j, &lsys)) == sps[0].sp_intl)
		  {
			star->sta_spgrp = lspaceg; p1tocurr(j, star, &sgs);
			screws(spsp, star, &sgs);	break;
		  }
	   }
	/* if the user has requested that we write out a modified multiref,
	 of if the user has requested a conditional output (change in
	 spacegroup but not a change in point group), we do that now */
	if ((star->sta_flags & SP_OUTMREF) ||
	 ((star->sta_flags & SP_CONDOUTMREF) &&
	 (lsys == (star->sta_xp)->xisys)))
	  {
		if (EOK != (i = out_modmref(j, star, spsp, &sgs)))
			return endspace(i, star, &sgs);
	  }
	return endspace(EOK, star, &sgs); /* finished */
}

int spacegmain(int argc, char *argv[])
{ /*	This is the mainline for the effort to determine the spacegroup by
	looking at the R-merge in various spacegroups and picking the
	highest-symmetry spacegroup that delivers a low R-merge value */
	int		i;
	STATRUN		spas;
  
	if (EOK != (i = spaceargs(argc, argv, &spas))) /* process arguments */
		return endspace(i, &spas, NULL);
	return spacegops(&spas); /* perform actual spacegroup calculations */
}

/* end of space_g.c */
