/* statistic.c -- compute R-factors, completeness, and other statistics

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	28 Nov 2006: sta_xp memory needed in statstart;
	 ucell call updated; sta_rabc,sta_rang replaced with xp->xast,xp->xanst
	 in many places; calls to artocel updated
	24 Jun 2006: include xgpostint.h rather than statdefs.h
	29 Apr 2006: outshell: more flexible output
	14 Apr 2006: spacepname called to print spacegroup name
	 9 Dec 2005: merge stat*.c together into this file.
	20 Nov 2005: inline argument declarations and strict prototyping
	18 Oct 2005: move lphkl[] inside goodref() where it belongs;
		replace static globals with elements of STATCWORK struct
	 5 Jan 2003: reorganized source-file header
	11 Oct 2005: housekeeping
	22 Jun 2005: strcpnb() declared as static
	11 May 2005: check return status from xgjournal
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	16 Jan 2005: eliminate redundant reflection-counting in statops()
	17 Sep 2004: in stathead(), make star->sta_nrshells flexible--between
		4 and 10 depending on how many observations are present;
		use -b flag to override default for that;
		in out_redun(), output fraction complete;
		modify summ_mr() to clean up the formatting of that.
	 6 Sep 2004: cosmetic changes
	27 Aug 2004: routines used only here are declared as static; cosmetics
	 1 Jul 2004: add plotting capabilities for <I/s> vs. sin(theta), etc.
	 7 Jun 2004: close sta_fpmis and set it to NULL in endstat()
	28 May 2004: most of statargs() moved to statstart()
	27 May 2004: replace NRFACS with NSRFACS throughout;
		move most of statmain() functionality to statops();
		call misstat() with hkl argument
	26 May 2004: bugfix: misstat() takes hklmi argument;
	27 May 2004: replace NRFACS with NSRFACS throughout
	25 Jan 2004: convert a few static globals to local variables
	25 Jan 2004: convert the only static global here into a local variable
	21 Jan 2004: license message inserted and initial comment reordered
	21 Jan 2004: license message inserted
	 7 Jan 2004: STATRUN structure elements in function args
		replace extern globals; args updated accordingly
	14 Nov 2003: merge old statargs.c and statheads.c into this file;
		external globals from old statglob.c replaced
		with elements of STATRUN structure
	19 Sep 2003: in out_run(), handle logic for ST_ISSCALED more carefully
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Aug 2003: remove call to exclucode() that is no longer needed;
		replace global uphkl[3] with static lphkl[3];
		modify handler for space-group exclusions in keephkl;
		hkl[] and hklmr[] are now local, not global
	 5 Jan 2003: reorganized source-file header
	15 Nov 2002: extern declarations made here rather than in statext.h
	 1 Nov 2002: minor reformatting of diagnostic output
	 1 Oct 2002: bugfix: don't increment rup->r_nscaled unless
		the number of observations used is >= 2;
		(?) allshl.sh_n* calculated here rather than elsewhere
	15 Apr 2002: modify call to init_shift() because shlp is altered;
		modify use of shlp[] in stathead(), obstat()
	25 Mar 2002: removed MSI attachments
	15 Jul 2001: modify the way `resmax' is managed.
	 8 Nov 1996: minor changes for X-GEN checkin at MSI
	31 May 1995: long variables now declared as ints; %ld becomes %d;
		"register" -> "register int" throughout
	20 May 1995: return EOK from stathead();
		make runix[] a signed char array
	 2 Mar 1995: add cross-reference;
			add "end of routine" comment to source
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 5 Oct 1994: use xgread to input MERGEDREF data
	 5 Oct 1994: use xgread to input CRYSTAL and DATARUN data
	 2 Sep 1994: create >= 6 intensity-statistics shells
		even with small input files
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: reorganized global variables a bit
	14 Mar 1994: new memory management and user interface tools
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version
	 5 May 1987: slow it down a little in exchange for greater
		reliability in not skipping high-resolution reflections.
	15 Sep 1984: initial code, based on "mrstat" from La Jolla
	
   Purpose: calculate R-factor, completeness, and other MULTIREF statistics.
	Routines here are also used in scalem and spaceg.
   Cross references: scalemm.c, space_g.c

	This offers the mainline of a program that provides the user with
	a variety of statistical assessments of his or her multiref.  The most
	time-consuming calculation is of the reflections missing from the
	multiref, either merging Bijvoets or not.  The list of missing
	reflections can be output to a file or not as the user
	chooses.  Other statistics obtained in this program are
	several different R-factors, both shift-by-shift and
	over the entire multiref, and a breakdown of the intensity
	data by sin(theta)/lambda and by intensity.
*************************************************************************/
#include	"xgpostint.h"
#define		DELETED_FLAG	(-1000.)
#define		FLOOR		(0.5)	/* resol-weight floor */
#define		OBSLP		8	/* obs per line on printer */

 typedef struct {
		double	stc_yy, stc_yd;
		double	stc_ski[MAXOB], stc_skf[MAXOB];
		sichar	stc_runix[MAXOB];
		} STATCWORK;

static void sumstatrun(FILE *fp, STATRUN *star)
{ /*	This summarizes the runtime conditions to the stream fp */
	static char	*flist[] =
		{"multiref", "xlog", "scalein", "missing", NULL};
	static char	ftypes[] = {1, 2, 1, 2, 0 };

	ftypes[1] = (star->sta_fpl == stdout) ? 0 : 2;
	ftypes[2] = (0 == star->sta_nscales) ? 0 : 2;
	ftypes[3] = (star->sta_flags & ST_OUTMISS) ? 2 : 0;
	header(fp, REDUCTION, flist, ftypes);
	fprintf(fp,"\nBijvoets WILL%sbe merged\n",
		(star->sta_flags & ST_SEPBJ) ? " NOT " : " ");
	if (0 == star->sta_nscales)
		fprintf(fp,"All observations will be included\n");
	else {
		fprintf(fp,
	 "Only observations appearing in scaled shifts will be included\n");
		fprintf(fp, "\t\t%6d shifts are in use\n",
		 (int)(star->sta_nscales));
	     }
}

int statstart(STATRUN *star)
{ /* define default values and initializations for a run of stats.
	This can't be static because it's called from scalemm */
	int	gotone, i, nsca;
	CRYSTAL	*xhp;
	DATARUN	*rhp;
	VECPAR	*vpp, *vpe;
	static VSPLINE	*sgsp;

	if (qerror((AVOID *)(star->sta_fpmi = xgopen("multiref", "r")),
	 "open input multiref")) return EOPENERR; /* open the input multiref */
	if (EOK != (i = headmem(&(star->sta_xhdp), &(star->sta_rhdp), 1)))
		return i;
	xhp = star->sta_xhdp;	rhp = star->sta_rhdp;
	if (verror(0 < xgread((void *)xhp, XG_CRYSTAL, star->sta_fpmi),
	 "read crystal header in input multiref")) return EREADERR;
	if (NULL == (star->sta_xp = (REFPARS *)callop(1, sizeof (REFPARS),
	 "star->spa_xp in spaceg"))) return EALLOC;
	if (NULL == (AVOID *)(star->sta_vshp = (VECPAR *)callop(
	 (unsigned)xhp->xt_nruns, sizeof (VECPAR),
	 "star->sta_vshp in statstart"))) return EALLOC;
	if (NULL != (star->sta_shlp = get_shifts(NULL, star->sta_vshp,
	 &nsca))) star->sta_flags |= ST_ISSCALED;
	star->sta_nscales = nsca;
/* examine the scale-factor output to see if we actually got any geometry-
 dependent coefficients. */
	gotone = 0;	vpp = star->sta_vshp;	vpe = vpp + xhp->xt_nruns;
	for ( ; vpp < vpe; vpp++)
	   {
		for (i = gotone = 0; i < MAXVECPARS; i++)
			if (fabs(vpp->v_coef[i]) > FUZZ)
			  {
				gotone = 1; break;
			  }
		if (gotone) break;
	   }
	star->sta_flags &= ~(SC_DIFFRACTION & SC_EXITBEAM);
	if (!gotone)
	  {
		memzap((AVOID **)&(star->sta_vshp), "vshp in statstart");
		star->sta_nvpars = 0 ;
	  }
	else {
		vpp = star->sta_vshp;
		if (vpp->v_ptype == 'd')
		  {
			star->sta_flags |= SC_DIFFRACTION;
			star->sta_nvpars = 10;
		  }
		else {
			star->sta_nvpars = 9;
			star->sta_flags |= SC_EXITBEAM;
		     }
/* we need some spline information in this case to do the xy conversions */
		if (NULL == (sgsp = gspline(5, &i))) return i;
	     }
	if (NULL == (star->sta_fpl = propen("xlog")))
		star->sta_fpl = stdout;
	if (star->sta_flags & ST_OUTMISS)
	  {
		if (NULL == (star->sta_fpmis = xgopen("missing", "w")))
			star->sta_fpmis = stdout;
	  }
	sumstatrun(stdout, star);
	if (star->sta_fpl != stdout) sumstatrun(star->sta_fpl, star);
	return EOK;
}

static int statargs(int argc, char *argv[], STATRUN *star)
{ /*	"statargs" processes the arguments and sets up files for "stats".
	It reads a multiref with logical or environment name "multiref",
	along with similarly-set-up files "scalein" as input scale file
	and "xlog" as the output log file.
	If "-n" or "-m" is an input argument, Bijovets will not be merged;
	otherwise, they will be merged.
	If "-o" is an input argument, we omit the missing-reflection list.
	A purely numeric input is assumed to be the resolution limit
	of the program run (default=max.resol. of data)	*/
	int	i;
	char	**dcp, **flags;
	
	/* default: flags off except missing list */
	zeromem((void *)star, sizeof (STATRUN)); /* initialize run struct */
	star->sta_flags = ST_OUTMISS;	star->sta_nrshells = -1;
	/* loop through args, setting file io, resolution limits,
	and bijvoet merges according to args. */
	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	flags = flagproc(argc, argv, "BMNO");
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	   switch (**dcp)
	     {
		case 'B':	/* specify # of resolution bins */
			if ((0 < (i = atoi(*(dcp) + 1))) &&
			 (i <= MAXRSHELLS + 1)) star->sta_nrshells = i - 1;
		case 'M':	case 'N':	/* don't merge Bijvoets */
			star->sta_flags |= ST_SEPBJ;	break;
		case	'O':	/* don't write out a "missing" list */
			star->sta_flags &= ~(ST_OUTMISS);	break;
	     }
	for (dcp = argv + 1, i = argc - 1; (i--); dcp++) if (allnumeric(*dcp))
		star->sta_resstate = atof(*dcp);
	return statstart(star);
}

static void strcpnb(char *s1, char *s2)
{ /* Copies string s2 to string s1, except for blanks.  */
	register int	i, j;
	char		c;

	for (i = j = 0; ('\0' != (c = s2[i])); i++) if (' ' != c) s1[j++] = c;
	s1[j] = '\0';
}

static void show_hlimits(FILE *fp, STATRUN *star)
{ /* display the hkl limits of the search to the stream fp */
	CRYSTAL	*xhp;

	xhp = star->sta_xhdp;
	fprintf(fp,
	 "Reflections will be examined in a range from infinity to %9.4f A\n",
		star->sta_resmax);
	if (star->sta_hklmax[3] == 1) fprintf(fp,
		"Limits on h, k, l: [%3d, %3d, %3d] to [%3d, %3d, %3d]\n",
		0, 0, 0, star->sta_hklmax[0],
		star->sta_hklmax[1], star->sta_hklmax[2]);
	else if (star->sta_hklmax[3] == 3)
	  {
		if (xhp->xt_isys == 2) fprintf(fp,
		 "Limits on h, k, l: [%3d, %3d, %3d] to [%3d, %3d, %3d]\n",
		 0, 0, -star->sta_hklmax[2], star->sta_hklmax[1],
		 star->sta_hklmax[0], star->sta_hklmax[2]);
		else fprintf(fp,
		 "Limits on h, k, l: [%3d, %3d, %3d] to [%3d, %3d, %3d]\n",
		 0, 0, -(star->sta_hklmax[2]), star->sta_hklmax[0],
		 star->sta_hklmax[1], star->sta_hklmax[2]);
	  }
	else fprintf(fp,
		"Limits on h, k, l: [%3d, %3d, %3d] to [%3d, %3d, %3d]\n",
		0, -(star->sta_hklmax[1]), -(star->sta_hklmax[2]),
		star->sta_hklmax[0], star->sta_hklmax[1],
		star->sta_hklmax[2]);
}

static int stathead(STATRUN *star)
{ /*	Processes headers in multiref file pointed to by star->sta_fpmi */
	register int	i;
	short		upmax[3];
	float		*flpi, *flpo;
	double		thismin;
	CRYSTAL		*xhp;
	DATARUN		*rhp;
	REFPARS		*xp;
	RUNSTAT		*runb, *runp;
	RUNWORK		*rup;
	SHLSTAT		*shlrp;
	SHIFT		*tsh;
	static float	lismin[NISRANGE+1] = {0.,1.,2., 5.,10.,20.,40.,0.};
	static float	lysmin[NYSRANGE+1] = {0.,2.,5.,10.,20.,40.,60.,0.};
	static float	lyymin[NYRANGE+1] = {0., 40., 80., 120., 160., 320.,
				640., 2560., 20480., 0.};

	star->sta_resmax = 10000.;/* set max resolution to ridiculous value */
	/* Read in the crystal header and then the run headers,
	storing some useful information from them in the "orstat" structures
	and accumulating totals. */
	xhp = star->sta_xhdp;	rhp = star->sta_rhdp;	xp = star->sta_xp;
	if (verror(!(ucell(xhp->xt_ar, xhp->xt_ang, xp->xast, xp->xanst,
	 xhp->xt_isys)), "compute reciprocal-space unit cell parameters"))
		return EBADVALUE;
	if (NULL == (AVOID *)(star->sta_runs =
	 (RUNSTAT *)callop((unsigned)xhp->xt_nruns,
	 sizeof (RUNSTAT), "runs in stathead"))) return EALLOC;
	if (NULL == (AVOID *)(star->sta_rub = (RUNWORK *)callop(
	 (unsigned)xhp->xt_nruns, sizeof (RUNWORK),
	 "star->sta_rub in stathead"))) return EALLOC;
	star->sta_spgrp = xhp->xt_intl;
	if ((xhp->xt_isys == 7) && (star->sta_spgrp < 200))
		star->sta_spgrp += 100;
	runb = &(star->sta_allruns);
	for (i = 0; i < MAXRUNS; i++) star->sta_runindex[i] = -1;
	for (i = 0, runp = star->sta_runs, rup = star->sta_rub;
	 i < xhp->xt_nruns; i++, runp++, rup++)
	   {
		if (verror(0 < xgread((void *)rhp, XG_DATARUN,
		 star->sta_fpmi), "read data run header in input multiref"))
			return EREADERR;
		(void)keep_dimen(rhp);
		workrun(rhp, rup, xhp, (int)(i * 200));
		runp->r_nexam = rhp->run_nexam;
		runp->r_ninteg = rhp->run_ninteg;
		runp->r_id = rhp->run_id;
		strcpnb(runp->r_code, rhp->run_code);
		runb->r_ninteg += rhp->run_ninteg; /*# of obs integrated*/
		runb->r_nexam += rhp->run_nexam; /* # of obs examined */
		if (rhp->run_reslim < star->sta_resmax)
			star->sta_resmax = rhp->run_reslim;
		star->sta_runindex[rhp->run_id] = i;
	   }
	/* Unlike earlier versions, we no longer artificially reduce 'resmax'
	 slightly to be sure to bring in all the reasonable reflections.
	 We still save the un-reduced value in sin(theta)/lambda terms as
	 "star->sta_origlim".  */
	star->sta_origlim = 0.5 / star->sta_resmax;
	/* resmax *= 0.995; */
	if (star->sta_resstate > star->sta_resmax)
		star->sta_resmax = star->sta_resstate;
/* determine the limits on hkl to be checked, based on the resolution limit. */
	for (i = 0; i < 3; i++) upmax[i] =
		(short)(xhp->xt_ar[i] / star->sta_resmax);
	mpack(upmax, star->sta_hklmax, star->sta_spgrp);
	/* in monoclinics and rhombohedrals, plus trigonals other than
	point group 312, the sign byte can be 0 or 2.  In triclinic it
	can be 0,2,4, or 6. Everywhere else it can only be zero.  Whee. */
	star->sta_hklmax[3] = ((2 < xhp->xt_isys && xhp->xt_isys < 7) ? 1:
		((xhp->xt_isys > 1) ? 3: 7));
	if ((143 <= xhp->xt_intl && xhp->xt_intl <= 167) &&
	  (xhp->xt_intl != 152 && xhp->xt_intl != 154 && xhp->xt_intl != 156))
		star->sta_hklmax[3] = 3;
	show_hlimits(star->sta_fpl, star);	show_hlimits(stdout, star);
		/* decide how many R-factor bins to create */
	if ((0 > star->sta_nrshells) || (star->sta_nrshells > MAXRSHELLS))
	  {
		if (runb->r_ninteg < 1000) star->sta_nrshells = 3;
		else if (runb->r_ninteg < 5000) star->sta_nrshells = 4;
		else if (runb->r_ninteg < 9000) star->sta_nrshells = 5;
		else if (runb->r_ninteg < 14000) star->sta_nrshells = 6;
		else if (runb->r_ninteg < 20000) star->sta_nrshells = 7;
		else if (runb->r_ninteg < 27000) star->sta_nrshells = 8;
		else star->sta_nrshells = MAXRSHELLS - 1;
	  }
	/* Determine how many Wilson-plot shells to divide the data into,
	and determine the lower limits of the Wilson plot shells.
	Compute the corresponding spacing limits in Angstroms, too. */
	star->sta_nwshells = min2(MAXWSHELLS,
		(short)(runb->r_ninteg / MINOSHELL));
	/* for X-GEN 5.2: number of Wilson bins >= number of R-factor bins */
	if (star->sta_nwshells < star->sta_nrshells)
		star->sta_nwshells = star->sta_nrshells;
	for (i = 0; i <= star->sta_nwshells; i++)
	   {
		thismin = star->sta_resmax * pow((double)(star->sta_nwshells +
		 1) / (double)(i+1), 1. / 3.);
		runb->r_wil[i].w_dmin = thismin;
		star->sta_wresb[i] = 0.5 / thismin;
	   }
	/* Determine the lower limits of the R-factor shells */
	for (i = 0, shlrp = &(star->sta_shl[0]);
	 i <= star->sta_nrshells; i++, shlrp++)
	   {
		thismin = star->sta_resmax * pow((double)(star->sta_nrshells +
			1) / (double)(i+1), 1. / 3.);
		shlrp->sh_dmin = thismin; star->sta_rresb[i] = 0.5 / thismin;
	   }
	if (!(star->sta_flags & ST_ISSCALED)) return EOK;
	for (i = 0, tsh = star->sta_shlp; i < star->sta_nscales; i++, tsh++)
	   {
		runp = star->sta_runs + star->sta_runindex[tsh->runno];
		(runp->r_nshifts)++;
	   }
	runb->r_nshifts = star->sta_nscales;
	/* get some coefficients to be used in computing sin(theta)/lambda
	for the missing reflections */
	star->sta_hklco[0] = 0.25 * xp->xast[0] * xp->xast[0];
	star->sta_hklco[1] = 0.25 * xp->xast[1] * xp->xast[1];
	star->sta_hklco[2] = 0.25 * xp->xast[2] * xp->xast[2];
	star->sta_hklco[3] = 0.5 * xp->xast[0] * xp->xast[1] *
		cos(xp->xanst[2]);
	star->sta_hklco[4] = 0.5 * xp->xast[0] * xp->xast[2] *
		cos(xp->xanst[1]);
	star->sta_hklco[5] = 0.5 * xp->xast[1] * xp->xast[2] *
		cos(xp->xanst[0]);
	/* establish ranges of sta_ismin,  sta_ysmin, and sta_yymin */
	for (i = 0, flpo = star->sta_ismin, flpi = lismin;
	 i <= NISRANGE; i++, flpi++, flpo++) *flpo = *flpi;
	for (i = 0, flpo = star->sta_ysmin, flpi = lysmin;
	 i <= NYSRANGE; i++, flpi++, flpo++) *flpo = *flpi;
	for (i = 0, flpo = star->sta_yymin, flpi = lyymin;
	 i <= NYRANGE; i++, flpi++, flpo++) *flpo = *flpi;
	/* express star->sta_resmax in sin(theta)/lambda terms */
	star->sta_resmax = 0.5 / star->sta_resmax;
		/* get <omega> for each scaling group */
	get_omcen(star->sta_fpmi, star->sta_shlp, star->sta_nscales, 0);
	return EOK;
}

static void out_run(FILE *fp, RUNSTAT *rup, STATRUN *star)
{ /* 	Prints out essential information about the datarun.
	pointed to by rup onto the file pointed to by fp. */
	register int	i;
	if (rup->r_i000 < 1.e5)
	  fprintf(fp,
	   " %6d %6d %6d %5d %5d %5.3f %7.1f %8.2f %7.1f %7.3f",
	 rup->r_ninteg,rup->r_nfound,rup->r_nscaled,rup->r_nabsent,
	 rup->r_nflagged,rup->r_allw.w_resol,rup->r_i000,rup->r_b,
	 rup->r_allw.w_ibar,rup->r_allw.w_iovsig);
	  else fprintf(fp,
	   " %6d %6d %6d %5d %5d %5.3f %7.0f %8.2f %7.1f %7.3f",
	 rup->r_ninteg,rup->r_nfound,rup->r_nscaled,rup->r_nabsent,
	 rup->r_nflagged,rup->r_allw.w_resol,rup->r_i000,rup->r_b,
	 rup->r_allw.w_ibar,rup->r_allw.w_iovsig);
	if (star->sta_flags & ST_ISSCALED)
	  {
		for (i = 0; i <= NSRFACS; i++)
		   {
			if (rup->r_nscaled > 0 && rup->r_rden[i] > 0.)
				fprintf(fp," %6.2f",rup->r_rnum[i] * 100.);
		  	else	fprintf(fp,"       ");
		   }
	  }
	fprintf(fp,"\n");
}

static void out_redun(FILE *fp, FILE *fppl, SHLSTAT *shp)
{ /*	Prints observation redundancies to filepointer fp
	for the resolution shell-pointer shp. It also adds to plots. */
	register int	i;
	float		lov[3];

	lov[0] = (shp->sh_nposs <= 0) ? 0 :
	 ((float)(shp->sh_ncoll)) / ((float)(shp->sh_nposs));
	lov[1] = (shp->sh_ncoll <= 0) ? 0 :
	 ((float)shp->sh_nob) / ((float)shp->sh_ncoll);
	lov[2] = (shp->sh_ncoll <= 0) ? 0 :
	 ((float)shp->sh_nob) / ((float)shp->sh_nposs);
	fprintf(fp, "%7.3f%9d%10d%9d%9.3f",  shp->sh_resol,
		shp->sh_nposs, shp->sh_ncoll, shp->sh_nquan[0], lov[0]);
	for (i = 1; i <= (NOBMAX+1); i++) fprintf(fp, "%7d", shp->sh_nquan[i]);
	fprintf(fp,"%9d %6.2f %6.2f", shp->sh_nob, lov[1], lov[2]);
	(void)plotadd(fppl, 3, 0.5 / shp->sh_resol, lov);
}

static void out_wilstat(FILE *fp, FILE *fppl, WILSON *wp)
{ /*	Summarizes the Wilson-plot results */
	register int	i;
	fprintf(fp,"%9.3f%9.2f%8.3f  ",
		wp->w_resol,wp->w_ibar,wp->w_iovsig);
	for (i = 0; i <= NISRANGE; i++)
	   fprintf(fp, "%6d", (int)wp->w_niosig[i]);
	fprintf(fp, "%7d", wp->w_n);
	(void)plotadd(fppl, 1, 0.5 / wp->w_resol, &(wp->w_iovsig));
}

static void out_shell(FILE *fp, SHLSTAT *shp)
{ /*	Outputs results for a resolution shell */
	register int	i;
	double		rlo;

	fprintf(fp,"%9.2f%8.3f",shp->sh_ybar,shp->sh_yovsig);
	for (i = 0; i <= NYSRANGE; i++)
		fprintf(fp,"%5d",shp->sh_nyosig[i]);
	fprintf(fp,"%7d%7d",shp->sh_ncoll,shp->sh_nob);
	for (i = 0; i <= NSRFACS; i++)
	   {
		if (shp->sh_rden[i] > 0.)
		  {
			if ((rlo = shp->sh_rnum[i] * 100.) > 9999.)
				fprintf(fp, "%8.1f", rlo);
			else	fprintf(fp, "%8.2f", rlo);
		  } 
		else	fprintf(fp,"        ");
	   }
}

static void out_header(FILE *fp, STATRUN *star)
{ /*	Prints out summary of information contained in the multiref header(s) */
	register int	i;
	CRYSTAL		*xhp;
	REFPARS		*xp;
	double		c[3], corr, volume;

	xhp = star->sta_xhdp;	xp = star->sta_xp;
	fprintf(fp,
 "\n\n\t\t\tStatistical Treatment of Crystallographic Data for %s\n",
		xhp->xt_code);
	fprintf(fp,"\n\nCrystal Description:\t%s\n",xhp->xt_desc);
	fprintf(fp,
"Crystal Spacegroup: %s, International Tables No. %d, System No. %d\n",
		spacepname(xhp->xt_intl), xhp->xt_intl,xhp->xt_isys);
	for (i = 0; i < 3; i++) c[i] = cos(xp->xanst[i]);
	corr = 1. + 2. * c[0] * c[1] * c[2] -
		c[0] * c[0] - c[1] * c[1] - c[2] * c[2];
	volume = xp->xast[0] * xp->xast[1] * xp->xast[2] * sqrt(corr);
	fprintf(fp,
	 "Real-Space Parameters:\t\t    a        b        c");
	fprintf(fp,
	 "    alpha     beta    gamma        volume\n");
	fprintf(fp,"\t\t\t%13.4f%9.4f%9.4f%9.4f%9.4f%9.4f%14.4f\n",
		xhp->xt_ar[0], xhp->xt_ar[1], xhp->xt_ar[2],
		xhp->xt_ang[0], xhp->xt_ang[1], xhp->xt_ang[2], 1./volume);
	fprintf(fp, "Reciprocal-Space Parameters:\t   ");
	fprintf(fp, "a*       b*       c*");
	fprintf(fp, "   alpha*    beta*   gamma*        volume\n");
	fprintf(fp, "\t\t\t%13.6f%9.6f%9.6f%9.4f%9.4f%9.4f%14.7e\n",
		xp->xast[0], xp->xast[1], xp->xast[2], xp->xanst[0]/RADC,
		xp->xanst[1]/RADC, xp->xanst[2]/RADC, volume);
	fprintf(fp, "General Notes on the Crystal: %s\n \n",xhp->xt_notes);
	fprintf(fp,
	 "Total Number of Reflections, Merging Bijvoets       = %7d\n",
		(int)(star->sta_nrmerge));
	fprintf(fp,
	 "Total Number of Reflections, Not Merging Bijvoets   = %7d\n",
		(int)(star->sta_nrnomerge));
	fprintf(fp,
	 "Max. No. of Observations/Refl, Merging Bijvoets     = %7d\n",
		(int)(star->sta_mobmerge));
	fprintf(fp,
	 "Max. No. of Observations/Refl, Not Merging Bijvoets = %7d\n",
		(int)(star->sta_mobnomerge));
	if (star->sta_nrill) fprintf(fp,
	 "\nNumber of illegal reflections & observations = %d, %d\n",
		(int)(star->sta_nrill), (int)(star->sta_nobill));
	if (star->sta_nrobd) fprintf(fp,
	 "Number of out-of-resolution-bounds refls & obs = %d, %d\n",
		(int)(star->sta_nrobd), (int)(star->sta_nobobd));
	if (star->sta_nobnohead) fprintf(fp,
	   "%d observations appear with NO corresponding header\n",
		(int)(star->sta_nobnohead));
	if (star->sta_nobnegi) fprintf(fp,
	 "%d observations appear with negative I or non-positive sigma\n",
		(int)(star->sta_nobnegi));
	if (star->sta_allmark)
	 {
	  fprintf(fp,
 "%d reflections have had ALL observations marked for deletion\n",
		(int)(star->sta_allmark));
	  fprintf(fp," or had no scale constant input for them\n");
	 }
	if (star->sta_nroutorder)
	  fprintf(fp,
 "%d reflections appear to be out of order in the input file\n",
		(int)(star->sta_nroutorder));
	fprintf(fp,
"Bijvoets have%sbeen merged in computing the following statistics\n",
		(star->sta_flags & ST_SEPBJ) ? " NOT " : " ");
}

static void out_legend(FILE *fp, STATRUN *star)
{ /*	This prints out the definitions of the quantities displayed
	in the output.  The definitions not actually used are not defined. */
	fprintf(fp,"\n\nDEFINITIONS:\n\n");
	fprintf(fp,"I:	Intensity of a single observation.\n");
	fprintf(fp,"I000:	The intensity of the (000) reflection");
	fprintf(fp," as determined by Wilson statistics.\n");
	fprintf(fp,"B:	The Debye-Waller (`temperature') factor");
	fprintf(fp," obtained from Wilson statistics, in Angstroms squared.\n");
	fprintf(fp,"<I>:	Average intensity for all observations in a");
	fprintf(fp," data run or a multiref.\n");
	fprintf(fp,"G.O.F.:	Goodness-of-fit Parameter * 100.\n");
	if (star->sta_flags & ST_ISSCALED)
	  {
	   fprintf(fp,
 "R1:	Weighted-squared R-factor on intensity * 100.\n");
	   fprintf(fp,
 "R2:	Unweighted-squared R-factor on intensity * 100.\n");
	   fprintf(fp,
 "R3:	Unweighted absolute-value R-factor on intensity * 100.\n");
	   fprintf(fp,
 "R4:	Unweighted absolute-value R-factor on F * 100.\n");
	   fprintf(fp,
 "R5:	Pairwise absolute-value R-factor on intensity * 100.\n");
	   fprintf(fp,
 "Y:	Average intensity for all scaled observations of a reflection.\n");
	  }
	fprintf(fp,"Deleted Observations:	Measurements which were made");
	fprintf(fp," but which are actually absent from the multiref.\n");
	fprintf(fp,
 "Flagged Observations:	Measurements which are flagged");
	fprintf(fp,
 " with negative shift-values, causing them to be ignored.\n");
	fprintf(fp,"Missing Reflections:	Reflections for which no");
	fprintf(fp," observations appear in the multiref.\n");
}

static void summ_mr(FILE *fp, STATRUN *star)
{ /*	This prints the results of the statistical survey
	of the multiref to the filepointer fp */
	int		i, plerr;
	static char	*rfacleg[] = { "R1", "R2", "R3", "R4", "R5" };
	float		*ypp;
	double		td;
	CRYSTAL		*xhp;
	RUNSTAT		*rsb, *rsp;
	SHLSTAT		*shlb, *shlrp;
	WILSON		*ruwp;
	static FILE	*fpps;
	static char	*redunleg[] =
	 { "Fraction_Complete", "Obs/Collected", "Obs/Possible" };

/* General information in the header */
	out_header(fp, star);
	fprintf(fp,
 "\n\t\t\tSummary of Measured Reflections by Data Run:\n");
	fprintf(fp,"  Data Run   Number of Observations\t     Ave.");
	fprintf(fp," Wilson Statistics    <I> <I/sig>");
	if (star->sta_flags & ST_ISSCALED)
		fprintf(fp,"\t\tAgreement Statistics: \n");
	fprintf(fp," #    Code   Col-  Found Scaled   Ab- Flag-");
	fprintf(fp," Resol    I000        B");
	if (star->sta_flags & ST_ISSCALED) fprintf(fp,
 	 "\t\t   G.O.F.     R1     R2     R3     R4     R5\n");
	fprintf(fp,
		   "           lected                sent   ged\n");
	xhp = star->sta_xhdp;
	for (i = 0, rsp = &(star->sta_runs[0]); i < xhp->xt_nruns; i++, rsp++)
	   {
		fprintf(star->sta_fpl, "%2d%8s", rsp->r_id, rsp->r_code);
		out_run(fp, rsp, star);
	   }
	fprintf(fp,"Totals:   ");
	rsb = &(star->sta_allruns);	out_run(fp, rsb, star);
/* Observation Redundancies */
	fprintf(fp,
	 "\n\t\t\tSummary of Observation Redundancies by Shells:\n");
	fprintf(fp, "Lower     Ave. Number of Bragg Reflections Fraction   ");
	fprintf(fp, "  No. of Reflections with Given No. of Observations  ");
	fprintf(fp, "    Total  Multiplicity\n");
	fprintf(fp, "Limit,A  Resol Possible Collected  Missing Complete   ");
	fprintf(fp, "   1      2      3      4      5      6      7    > 7");
	fprintf(fp, "  Observ. ob/col ob/pos");
	fpps = plotsetup(3, 0, &plerr, "statredun.out",
	 "Completeness and Redundancy", "<sin(theta)/lambda>",
	 "Completeness and Redundancy", redunleg);
	for (i = 0, shlrp = &(star->sta_shl[0]);
	 i <= star->sta_nrshells; i++, shlrp++)
	   {
		if ((td = shlrp->sh_dmin) > 99.9) fprintf(fp, "\n%7.3f", td);
		else fprintf(fp,"\n%7.4f", td);
		out_redun(fp, fpps, shlrp);
	   }
	fprintf(fp,"\nTotals:");
	shlb = &(star->sta_allshl);	out_redun(fp, NULL, shlb);
	(void)plotfinish(fpps, shlb->sh_resol);
/* Intensity Distributions */
	fpps = plotsetup(1, 0, &plerr, "statios.out", "I/sigma vs Resolution",
	 "<sin(theta)/lambda>", "<I/sigma>", NULL);
	fprintf(fp,
 "\n\n\t\t\tSummary of Observation Intensities by Shells:\n");
	fprintf(fp,"  Shell  Average  Average Average    ");
	fprintf(fp,
 "    Number of Observations with");
	fprintf(fp,"\n  Lower   Resol.    Obs.  I/sig(I)        ");
	fprintf(fp,"I/sigma(I) in Given Range     ");
	fprintf(fp, "\n  Limit,\t   Inten-             <=0    <1    <2    <5");
	fprintf(fp, "   <10   <20   <40   >40  Total\n");
	fprintf(fp," Angst.\t\t     sity\t\t\t\t\t\t\t      Obs.");
	for (i = 0, ruwp = &(rsb->r_wil[0]);
	 i <= star->sta_nwshells; i++, ruwp++)
	   {
		fprintf(fp,"\n%7.3f", ruwp->w_dmin);
		out_wilstat(fp, fpps, ruwp);
	   }
	fprintf(fp,"\nTotals:");
	out_wilstat(fp, NULL, &(rsb->r_allw));
	(void)plotfinish(fpps, (rsb->r_allw).w_resol);
	fprintf(fp,"\n");
	if (star->sta_nrshells != star->sta_nwshells)
	  {
		fprintf(fp,
	 "\n\t\t\tSummary of Observation Intensities by %s Shells:\n",
			(star->sta_nrshells < star->sta_nwshells) ?
			"wider": "narrower");
		fprintf(fp,"  Shell  Average  Average Average    ");
		fprintf(fp,
	 "    Number of Observations with");
		fprintf(fp,"\n  Lower   Resol.    Obs.  I/sig(I)        ");
		fprintf(fp,"I/sigma(I) in Given Range     ");
		fprintf(fp,
		 "\n  Limit,\t   Inten-             <=0    <1    <2    <5");
		fprintf(fp, "   <10   <20   <40   >40  Total\n");
		fprintf(fp," Angst.\t\t     sity\t\t\t\t\t\t\t      Obs.");
		for (i = 0, ruwp = &(star->sta_runrw[0]);
		 i <= star->sta_nrshells; i++, ruwp++)
		   {
			fprintf(fp,"\n%7.3f", ruwp->w_dmin);
			out_wilstat(fp, NULL, ruwp);
		   }
		fprintf(fp,"\nTotals:");
		out_wilstat(fp, NULL, &(rsb->r_allw)); fprintf(fp,"\n");
	  }
	if (!(star->sta_flags & ST_ISSCALED)) return;
/* Reflection intensities and R-factors by R-factor shells */
	fpps = plotsetup(NSRFACS, 0, &plerr, "rfactorbyd.out",
	 "R-factors vs. sin(theta)/lambda", "<sin(theta)/lambda>",
	 "Various R-factors", rfacleg);
	fprintf(fp,
 "\n\t\t\tSummary of Reflection Intensities and R-factors by Shells:\n");
	fprintf(fp,
	 "  Shell  Average Average");
	fprintf(fp,
 "    Number of Reflections with\t\t\t\t\tAgreement Statistics");
	fprintf(fp,"\n  Lower    Ref.  Y/sig(Y)      ");
	fprintf(fp,"Y/sigma(Y) in Given Range     ");
	fprintf(fp,"\n  Limit,  Inten-\t    ");
	fprintf(fp,
 		"0   <2   <5  <10  <20  <40  <60  >60");
	fprintf(fp,
 " # Refs  # Obs  G.O.F.      R1      R2      R3      R4      R5");
	fprintf(fp,"\n Angst.    sity");
	for (i = 0, shlrp = &(star->sta_shl[0]);
	 i <= star->sta_nrshells; i++, shlrp++)
	   {
		fprintf(fp,"\n%7.3f", shlrp->sh_dmin);	out_shell(fp, shlrp);
		(void)plotadd(fpps, NSRFACS, 0.5 / shlrp->sh_resol,
			&(shlrp->sh_rnum[1]));
	   }
	fprintf(fp, "\nTotals:");	out_shell(fp, shlb);
	(void)plotfinish(fpps, 0.5 / shlb->sh_resol);
	fpps = plotsetup(NSRFACS, 0, &plerr, "rfactorbyi.out",
	 "G.O.F. and R-factors vs. I", "Intensity range",
	 "R-factor", rfacleg);
	fprintf(fp, "\n\n\t\t\tSummary of R-factors by Intensity Range:\n");
	fprintf(fp, " Inten.  Average Average");
	fprintf(fp,
 "    Number of Reflections with\t\t\t\t\tAgreement Statistics\n");
	fprintf(fp,"  Upper    Ref.  Y/sig(Y)      ");
	fprintf(fp,"Y/sigma(Y) in Given Range     ");
	fprintf(fp,"\n  Limit   Inten-\t    ");
	fprintf(fp, "0   <2   <5  <10  <20  <40  <60  >60");
	fprintf(fp,
 " # Refs  # Obs  G.O.F.      R1      R2      R3      R4      R5\n");
	fprintf(fp,"           sity");
	for (i = 0, ypp = &(star->sta_yymin[0]),
	 shlrp = &(star->sta_yrange[0]); i < NYRANGE; i++, ypp++, shlrp++)
	   {
		fprintf(fp,"\n%7.0f", *ypp);	out_shell(fp, shlrp);
		if (i > 0)
		  (void)plotadd(fpps, NSRFACS, *ypp, &(shlrp->sh_rnum[1]));
	   }
	fprintf(fp,"\nLargest");	out_shell(fp, shlrp);
	fprintf(fp,"\nTotals:");	out_shell(fp, shlb);
	(void)plotfinish(fpps, shlb->sh_ybar);
	out_legend(fp, star);		/* print definitions to logfile */
}

static void calc_wil(WILSON *wp)
{ /*	This does some simple calculations for a "wilson" structure. */
	if (wp->w_n <= 0) wp->w_ibar = wp->w_iovsig = 0.;
	else {
		wp->w_resol /= wp->w_n;	wp->w_ibar /= wp->w_n;
		wp->w_iovsig /= wp->w_n;
	     }
}

static void accum_run(RUNSTAT *rup, STATRUN *star)
{ /*	Accumulates the sums for an "run" structure and does some of the
	pertinent division operations.  Then it a weighted least-squares
	calculation of the B and I000 values for the data set or run. */
	register int	i, j;
	double		c, d, den, disc, e, f, g, logave, q0, q1, wi;
	WILSON		*wp;

	e = f = g = c = d = 0.;
	for (i = 0; i <= star->sta_nwshells; i++)
	   {
		wp = &(rup->r_wil[i]); rup->r_allw.w_resol += wp->w_resol;
		rup->r_allw.w_n += wp->w_n; rup->r_allw.w_ibar += wp->w_ibar;
		rup->r_allw.w_iovsig += wp->w_iovsig;
		for (j = 0; j <= NISRANGE; j++)
			rup->r_allw.w_niosig[j] += wp->w_niosig[j];
		calc_wil(wp);
		if (wp->w_n > 0)
		  {
			if (wp->w_ibar > 1.)
			  {
				wi = wp->w_n * (FLOOR + wp->w_resol);
				logave = log(wp->w_ibar);
			  }
			else {
				wi = logave = 0.;
			     }
			q0 = logave * wi;	q1 = -0.5 * wp->w_resol;
			e += q1 * q1 * wi;	f += q1 * wi;
			g += wi;	c += q0 * q1;	d += q0;
			wp->w_resol = (wp->w_resol > FUZZ ?
				0.5 / wp->w_resol : star->sta_wresb[i]);
		  }
		  else {
			logave = wp->w_ibar = 0.;
			wp->w_resol = star->sta_wresb[i];
		       }
	   }
	calc_wil(&(rup->r_allw));
	if (rup->r_allw.w_resol > FUZZ)
		rup->r_allw.w_resol = 0.5 / rup->r_allw.w_resol;
	disc = f * f - e * g;
	if (fabs(disc) < DFUZZ)
	  {
		rup->r_i000 = 1.e7;	rup->r_b = 99.;
	  }
	else {
		q0 = (f * c - e * d)/disc;
		rup->r_i000 = (q0 < 16. ? exp(q0) : exp(16.));
		rup->r_b = (f * d - g * c) / disc;
	     }
	rup->r_nabsent = rup->r_ninteg - (rup->r_nfound + rup->r_nflagged);
	rup->r_rden[0] = rup->r_nscaled - rup->r_nshifts;
	if (rup->r_rden[0] < 1) rup->r_rden[0] = 1;
	for (i = 0; i <= NSRFACS; i++)
	   if (fabs(den = rup->r_rden[i]) > FUZZ)
	     {
		rup->r_rnum[i] /= den;
		if (rup->r_rnum[i] < 0.) rup->r_rnum[i] = 0.;
		if (i <= 2) rup->r_rnum[i] = sqrt(rup->r_rnum[i]);
	     }
}

static void statlast(STATRUN *star)
{ /*	This performs the final calculations which lead to the
	computation of the outputted quantities. */
	register int	i, j, k;
	CRYSTAL		*xhp;
	RUNSTAT		*rubb, *rup;
	SHLSTAT		*shb, *shp;
	WILSON		*wp,*wpo;

	/* Accumulate the sums for the run totals */
	rubb = &(star->sta_allruns);	xhp = star->sta_xhdp;
	for (i = 0, rup = star->sta_runs; i < xhp->xt_nruns; i++, rup++)
	   {
		rubb->r_nfound += rup->r_nfound;
		rubb->r_nscaled += rup->r_nscaled;
		rubb->r_nflagged += rup->r_nflagged;
		for (j = 0; j <= NSRFACS; j++)
		   {
			rubb->r_rnum[j] += rup->r_rnum[j];
			rubb->r_rden[j] += rup->r_rden[j];
		   }
		for (j = 0; j <= star->sta_nwshells; j++)
		   {
			wp = &(rubb->r_wil[j]);
			wpo = &(rup->r_wil[j]);
			wp->w_resol += wpo->w_resol;
			wp->w_n += wpo->w_n;
			wp->w_ibar += wpo->w_ibar;
			wp->w_iovsig += wpo->w_iovsig;
			for (k = 0; k <= NISRANGE; k++)
			   wp->w_niosig[k] += wpo->w_niosig[k];
		   }
		accum_run(rup, star);
	   }
	accum_run(rubb, star);
	shb = &(star->sta_allshl);
	shb->sh_nquan[0] = 0;	/* zero this so we can recount it */
	for (i = 0, wp = &(star->sta_runrw[0]);
	 i <= star->sta_nrshells; i++, wp++)
	   {
		calc_wil(wp);
		wp->w_dmin = (star->sta_rresb[i] > FUZZ) ?
			0.5 / star->sta_rresb[i] : 0.;
		wp->w_resol = (wp->w_resol > FUZZ) ? 
			0.5 / wp->w_resol : wp->w_dmin;
		shp = &(star->sta_shl[i]);
		shb->sh_resol += shp->sh_resol;
		shb->sh_nposs += shp->sh_nposs;
		shb->sh_ncoll += shp->sh_ncoll;
		shb->sh_nob += shp->sh_nob;
		shb->sh_nrscaled += shp->sh_nrscaled;
		shb->sh_nobscaled += shp->sh_nobscaled;
		for (j = 0; j <= (NOBMAX+1); j++)
			shb->sh_nquan[j] += shp->sh_nquan[j];
		for (j = 0; j <= NSRFACS; j++)
		   {
			shb->sh_rnum[j] += shp->sh_rnum[j];
			shb->sh_rden[j] += shp->sh_rden[j];
		   }
		shb->sh_ybar += shp->sh_ybar;
		shb->sh_yovsig += shp->sh_yovsig;
		for (j = 0; j <= NYSRANGE; j++)
			shb->sh_nyosig[j] += shp->sh_nyosig[j];
		calc_sh(shp, star);
	   }
	for (i = 0, shp = &(star->sta_yrange[0]); i <= NYRANGE; i++, shp++)
		calc_sh(shp, star);
	calc_sh(shb, star);
	if (star->sta_mobnomerge > star->sta_mobmerge)
		star->sta_mobmerge = star->sta_mobnomerge;
	summ_mr(star->sta_fpl, star);	/* Print results to output file(s) */
}

static void findstl(short *uhkl, STATRUN *star)
{ /* this computes sin(theta)/lambda for the unpacked hkl `uhkl' */
	float	*hcop;
	double	ressq;

	hcop = &(star->sta_hklco[0]);
	ressq =
	 uhkl[0] * uhkl[0] * *(hcop)     + uhkl[1] * uhkl[1] * *(hcop + 1) +
	 uhkl[2] * uhkl[2] * *(hcop + 2) + uhkl[0] * uhkl[1] * *(hcop + 3) +
	 uhkl[0] * uhkl[2] * *(hcop + 4) + uhkl[1] * uhkl[2] * *(hcop + 5);
	star->sta_res = sqrt(ressq);
}

static short getir(double resol, STATRUN *star)
{ /* This returns the resolution shell index associated with "resol" */
	register int	i;

	for (i = 0; i < star->sta_nrshells; i++)
		if (resol < star->sta_rresb[i]) break;
	return	i;
}

static short getiw(double resol, STATRUN *star)
{ /*	This returns the Wilson-shell index */
	register int	i;

	for (i = 0; i < star->sta_nwshells; i++)
		if (resol < star->sta_wresb[i]) break;
	return	i;
}

static short obstat(int indx, ABBROB *aobp, STATRUN *star, STATCWORK *stcw)
{ /*	Statistical accumulations for a single observation.
	This routine has NOT received a `nihil obstat'. */
	register int	iuse, j;
	int		io, run_num;
	double		fsqovsig, gw, lres, sk, wijsq, *fp2, avec[MAXVECPARS];
	float		*fp1;
	SHIFT		*tsh;
	RUNSTAT		*rup;
	WILSON		*opw, *opr;
	xgulong		ulsh;
	/* Statistics computed independent of shift status */
	stcw->stc_ski[indx] = DELETED_FLAG - 1; stcw->stc_runix[indx] = -1;
	if (aobp->abo_runno >= MAX_RUN_INDEX) return 0;
	io = star->sta_runindex[aobp->abo_runno];
	if (0 > io) return 0; stcw->stc_runix[indx] = io;
	rup = star->sta_runs + io;
	if (aobp->abo_shift < 0) /* shift < 0 implies deleted observation */
	  {
		if (star->sta_flags & ST_ISSCALED)
		  { /* see if the observation was scalable originally */
			ulsh = LSH(aobp->abo_runno, 24) |
			  LSH(aobp->abo_chamno, 16) | (-aobp->abo_shift);
				/* shift not scaled--don't count obs*/
			if (0 > find_shift(star->sta_shlp, ulsh,
			 (int)(star->sta_nscales))) return 0;
		  }
		(rup->r_nflagged)++;
		return 0;		/* observation ignored */
	  }
	if (aobp->abo_sigma <= 0) /* nonsense I or sigma */
	  {
		(star->sta_nobnegi)++; return 0; /* observation ignored */
	  }
	fsqovsig = ((double)aobp->abo_ii) / ((double)aobp->abo_sigma);
	if (fsqovsig < 0) fsqovsig = 0.;
	for (iuse = 0; iuse < NISRANGE; iuse++)
		if (fsqovsig <= star->sta_ismin[iuse]) break;
	if (verror((iuse >= 0 && iuse <= NISRANGE), "use range index"))
		return EBADVALUE;
	opw = &(rup->r_wil[star->sta_iw]);
	opr = &(star->sta_runrw[star->sta_ir]);
	(rup->r_nfound)++;
	lres = star->sta_res;	opw->w_resol += lres;
	opw->w_ibar += aobp->abo_ii;	opw->w_n++;
	opw->w_iovsig += fsqovsig;	opw->w_niosig[iuse]++;
	opr->w_resol += lres;		opr->w_ibar += aobp->abo_ii;
	opr->w_n++;			opr->w_iovsig += fsqovsig;
	opr->w_niosig[iuse]++;
	/* Statistics for the scaled intensity */
	if (star->sta_flags & ST_ISSCALED)
	  {
		ulsh = LSH(aobp->abo_runno, 24) |
			LSH(aobp->abo_chamno, 16) | aobp->abo_shift;
		if (0 <= (j = find_shift(star->sta_shlp, ulsh,
		 (int)(star->sta_nscales))))
		  {
			tsh = star->sta_shlp + j;
			sk = terpscale(aobp->abo_omob, tsh, lres);
			if (star->sta_vshp)
			  {
				run_num = aobp->abo_shift / 200;
				refvec(aobp, star->sta_rub + run_num,
				 star->sta_flags, star->sta_nvpars, avec);
				fp1 = &((star->sta_vshp + run_num)->v_coef[0]);
				for (j = star->sta_nvpars, fp2 = &(avec[0]),
				 gw = 0; j--; fp1++, fp2++) gw += *fp1 * *fp2;
				sk *= gw;
			  }
			stcw->stc_ski[indx] = sk;
			stcw->stc_skf[indx] = (sk < 0 ? 0. : sqrt(sk));
			wijsq = aobp->abo_sigma;
			wijsq = sk / (wijsq * wijsq);
			stcw->stc_yy += wijsq * aobp->abo_ii;
			stcw->stc_yd += wijsq * sk;	return 1;
		  }
		else return 0;	/* count the observation only if it's scaled*/
	  }
	else return 1;	/* if we're not scaling, we count the observation*/
}

static int refstat(MERGEDREF *mp, STATRUN *star)
{ /*	This computes the statistics for a reflection.  The r-factor and
	reflection-redundancy calculations apply to merged reflections
	unless "-n" has been specified.  */
	register int		i, j, n;
	register ABBROB		*aobp, *aobpp;
	int			ir, iw, nobs, seenit;
	short			ip, jor, obsmr, thisnob;
	double			fbar, fdiff, fij, idiff, iij, ipscale, iscale;
	double			lres, sc, sigij, sigmay, wij, yovsigy, yy;
	double			r[6], s[6];
	CRYSTAL			*xhp;
	SHLSTAT			*shp,*yp;
	RUNSTAT			*rup;
	char			rfinor[MAXRUNS+1]; /* flags for obs per run */
	static short		lphkl[3];
	static STATCWORK	stcs;

	munpack(mp->mr_hkl, lphkl, star->sta_spgrp); /* unpack the hkl */
	star->sta_res = lres = mp->mr_stl;
	/* Get the resolution-shell and wilson-shell values */
	if (verror((((ir = star->sta_ir = getir(lres, star)) >= 0) &&
	 (ir <= star->sta_nrshells)), "use Rfactor shell number"))
		return EBADVALUE;
	if (verror((((iw = star->sta_iw = getiw(lres, star)) >= 0) &&
	 (iw <= star->sta_nwshells)), "use Wilson shell number"))
		return EBADVALUE;
		/* zero the cumulative intensity & weight sums */
	stcs.stc_yy = stcs.stc_yd = 0;	xhp = star->sta_xhdp;
	for (i = 0; i <= xhp->xt_nruns; i++)
		rfinor[i] = 0;	/* clear the "ref-found" flags */
	obsmr = 0;	/* zero the count of valid observations */
	nobs = mp->mr_nob;
	for (i = 0; i < nobs; i++)
	   {
		thisnob = obstat(i, &(mp->mr_ob[i]), star, &stcs);
		if (EBADVALUE == thisnob) return EBADVALUE;
		obsmr += thisnob;
	  }
	if (!obsmr)
	 { /* all obs were marked for deletion */
		(star->sta_allmark)++;	return EOK;
	 }
	/* Increment some counters for the resolution shell */
	shp = &(star->sta_shl[ir]);
	shp->sh_resol += lres;
	shp->sh_ncoll++;
	shp->sh_nob += obsmr;
	j = min2(obsmr,NOBMAX+1);
	shp->sh_nquan[j]++;
		/* rest of calcs done only if scaled */
	if (!(star->sta_flags & ST_ISSCALED)) return EOK;
	/* if we get here with yd == 0, it means all the observations
	were legitimate but came from shifts that have no scale factor
	associated with them: the shifts don't appear in the scalein file*/
	if (stcs.stc_yd < DFUZZ) return EOK;
	yy = stcs.stc_yy / stcs.stc_yd;
	sigmay = sqrt(1. / stcs.stc_yd);
		/* average structure amplitude: */
	fbar = (yy < 0. ? 0. : sqrt(yy));
	shp->sh_ybar += yy;
	yovsigy = (sigmay <= 0. ? 0. : yy / sigmay);
	shp->sh_yovsig += yovsigy;
	for (ip = 0; ip < NYSRANGE; ip++)
		if (yovsigy <= star->sta_ysmin[ip]) break;
	shp->sh_nyosig[ip]++;
	/*	Determine the Y-value bin to use	*/
	for (i = 0; i < NYRANGE; i++)
		if (yy <= star->sta_yymin[i]) break;
	yp = &(star->sta_yrange[i]);
	yp->sh_ncoll++;		yp->sh_nob += obsmr;
	yp->sh_nquan[j]++;	yp->sh_nyosig[ip]++;
	yp->sh_resol += lres;	yp->sh_ybar += yy;
	yp->sh_yovsig += yovsigy;
		/* must have <I> > 0, n > 1 */
	if (yy <= 0. || obsmr < 2) return EOK;
	for (seenit = i = 0, aobp = &(mp->mr_ob[0]); i < nobs; i++, aobp++)
	   if ((sc = stcs.stc_ski[i]) > DFUZZ && (aobp->abo_sigma > FUZZ) &&
	    ((jor = stcs.stc_runix[i]) >= 0))
	     {
		rup = star->sta_runs + jor;
		/* we increment this ONLY when obsmr >= 2, so we do it here */
		(rup->r_nscaled)++;
		if (!rfinor[jor])
		  {
			rfinor[jor] = 1; rup->r_nrefs++;
		  }
		iij = aobp->abo_ii;		fij = (iij < 0? 0 : sqrt(iij));
		sigij = aobp->abo_sigma;	wij = 1./(sigij*sigij);
		idiff = yy * sc - iij;
		fdiff = fbar * stcs.stc_skf[i] - fij;
		iscale = iij / sc;
		r[1] = wij * idiff * idiff;	s[1] = wij * iij * iij;
		r[2] = idiff * idiff;		s[2] = iij * iij;
		r[3] = fabs(idiff);		s[3] = fabs(iij);
		r[4] = fabs(fdiff);		s[4] = fij;
		r[5] = 0.;			s[5] = 0.;
		for (ip = 0, aobpp = &(mp->mr_ob[0]); ip < nobs; ip++, aobpp++)
		   if ((ip != i) && (stcs.stc_ski[ip] > DFUZZ) &&
		    (aobpp->abo_sigma > FUZZ))
		     {
			ipscale = aobpp->abo_ii / stcs.stc_ski[ip];
			r[5] += fabs(iscale - ipscale);
			s[5] += 0.5 * (iscale + ipscale);
		     }
		r[0] = r[1];			s[0] = 0;
		for (n = 0; n <= NSRFACS; n++)
		   {
			rup->r_rnum[n] += r[n];	 rup->r_rden[n] += s[n];
			shp->sh_rnum[n] += r[n]; shp->sh_rden[n] += s[n];
			yp->sh_rnum[n] += r[n];	 yp->sh_rden[n] += s[n];
		   }
		shp->sh_nobscaled++;	yp->sh_nobscaled++;
		if (!seenit)
		  {
			shp->sh_nrscaled++; yp->sh_nrscaled++; seenit = 1;
		  }
	 } /* end of "for (i = 0; ..." loop */
	return EOK;
}

static int goodref(MERGEDREF *mp, short excluc, u_char *hklmre, STATRUN *star)
{ /*	"goodref" reads in multiref reflections until it finds
	one which is legal (satisfies the spacegroup restrictions)
	and is in bounds (resolution value less than max).  It
	returns the (HKL) of the final reflection as 'hklmre'.
	The routine returns EOK if it encounters the last good reflection;
	it returns various error values if `reflstat' barfs; it returns
	-1 otherwise.  The normal return is -1.  */
	int		i;
	MERGEDREF	mrpri, mrbij;
	static short	lphkl[3];

	while (1) /* read each reflection; if we get to the end, flag it */
	     {
		if (0 >= xgread((void *)mp, XG_MERGEDREF, star->sta_fpmi))
		  {
			star->sta_flags |= ST_DONE; /* flagged:end of input */
			return -1;
		  }
		if (0 < hklcmp(mp->mr_hkl, star->sta_hklmax))
		  {
			statlast(star);	/* last good refl: quit */
			return EOK;
		  }
		munpack(mp->mr_hkl, lphkl, star->sta_spgrp);
		if (!keephkl(lphkl, excluc) ||
		 !screw_ok(lphkl, star->sta_spgrp))
		  {
			(star->sta_nrill)++;	/* count of illegal refs */
			star->sta_nobill += mp->mr_nob; /* illegal obs */
		  }
		  else	/* we recompute sin(theta)/lambda so that the
			input reflections and the search list are using
			exactly the same resolution computation */
		  {
			findstl(lphkl, star);
			if (star->sta_res > star->sta_resmax)
			  { /* count high-resolution reflections, obs */
				(star->sta_nrobd)++;
				star->sta_nobobd += mp->mr_nob;
			  }
			  else
			  { /* recomputed sin(theta)/la*/
				mp->mr_stl = star->sta_res;
				break; /* we've got a good one; quit */
			  }
		  }
	     }
	cpymem((AVOID *)hklmre,(AVOID *)mp->mr_hkl,4);
	(star->sta_nrmerge)++;	/* increment count of merged reflections */
	star->sta_nrnomerge++;	/* increment count of unmerged reflections */
	/* if both primary & bijvoet are there, increment unmerged again*/
	if (mp->mr_nobpri && mp->mr_nobbij) (star->sta_nrnomerge)++;
	if (((short)mp->mr_nob) > star->sta_mobmerge)
		star->sta_mobmerge = mp->mr_nob;
	if (((short)mp->mr_nobpri) > star->sta_mobnomerge)
		star->sta_mobnomerge = mp->mr_nobpri;
	if (((short)mp->mr_nobbij) > star->sta_mobnomerge)
		star->sta_mobnomerge = mp->mr_nobbij;
		/* unmerged: split it into primary & bijvoet*/
	if (star->sta_flags & ST_SEPBJ)
	  {
		splitmr(mp, &mrpri, &mrbij);
		if (EOK != (i = refstat(&mrpri, star))) return i;
		if (EOK != (i = refstat(&mrbij, star))) return i;
	  }
	else /* merged case: examine the entire reflection */
	  { /* do statistics on the reflection */
		if (EOK != (i = refstat(mp, star))) return i;
	  }
	return -1;
}

static void misstat(STATRUN *star, u_char *hklmi)
{ /*	Does the bookkeeping for missing reflections.  If a reflection is
	missing but actually has a larger sin(theta)/lambda than the original
	multiref's cutoff, we just ignore it at this point.  */
	int	lir;
	short	lochkl[3];
	SHLSTAT	*shb, *shlrp;

	if (star->sta_res > star->sta_origlim) return;
	shb = &(star->sta_allshl);
	lir = star->sta_ir;	shlrp = &(star->sta_shl[lir]);
	if (star->sta_flags & ST_OUTMISS)
	  {
		if (shb->sh_nquan[0] % OBSLP) fprintf(star->sta_fpmis,"   ");
		else	fprintf(star->sta_fpmis,"\n");
		munpack(hklmi, lochkl, star->sta_spgrp);
		fprintf(star->sta_fpmis, "%4d%4d%4d",
			(int)(lochkl[0]), (int)(lochkl[1]), (int)(lochkl[2]));
	   }
	(shb->sh_nquan[0])++;	(shlrp->sh_nquan[0])++;	return;
}

static int islegal(u_char *inuchkl, short excluc, STATRUN *star)
{ /* This checks to see that the refl has legal indices, is inside
	the legal resolution limits and is already in lowest terms.  */
	register int	i;
	u_char		shkl[4];
	static short	lphkl[3];

	munpack(inuchkl, lphkl, star->sta_spgrp);	/* unpack indices*/
	/* first see if it's an illegal reflection due to a screw-axis
	exclusion or a Laue-group exclusion */
	if (!keephkl(lphkl, excluc) || !screw_ok(lphkl, star->sta_spgrp))
		return 0;
	/* now see if it's in lowest terms */
	cpymem((AVOID *)shkl, (AVOID *)inuchkl, 4);
	symtr(shkl, star->sta_spgrp);
	for (i = 0; i < 4; i++) if (inuchkl[i] != shkl[i]) return 0;
	findstl(lphkl, star);	/* get the sin(theta)/lambda value */
	return (star->sta_res <= star->sta_resmax);
}

void calc_sh(SHLSTAT *shp, STATRUN *star)
{ /*	This does some final calculations for the "shellstat" structure
	pointed to by shp. This isn't static because space_g uses it too. */
	register int	i;
	int		nref;
	double		den;

	if ((nref = shp->sh_ncoll) <= 0) shp->sh_ybar = shp->sh_yovsig = 0.;
	else {
		shp->sh_resol /= nref;	shp->sh_ybar /= nref;
		shp->sh_yovsig /= nref;
		if (shp->sh_resol > FUZZ) shp->sh_resol = 0.5 / shp->sh_resol;
	     }
		/* goodness-of-fit: */
	shp->sh_rden[0] = shp->sh_nobscaled - (int)(star->sta_nscales);
	if (shp->sh_rden[0] < 1) shp->sh_rden[0] = 1;
	for (i = 0; i <= NSRFACS; i++)
	   if (fabs(den = shp->sh_rden[i]) > FUZZ)
	     {
		shp->sh_rnum[i] /= den;
		if (shp->sh_rnum[i] < 0.) shp->sh_rnum[i] = 0.;
		if (i <= 2) shp->sh_rnum[i] = sqrt(shp->sh_rnum[i]);
	     }
}

int endstat(int irv, STATRUN *star)
{ /* finishes off a run of the `stats' functionality */
	memzap((AVOID **)&(star->sta_rub), "star->sta_rub in stats");
	memzap((AVOID **)&(star->sta_shlp), "star->sta_shlp in stats");
	memzap((AVOID **)&(star->sta_vshp), "star->sta_vshp in stats");
	memzap((AVOID **)&(star->sta_runs), "star->sta_runs in stats");
	memzap((AVOID **)&(star->sta_xp), "star->sta_xp in stats");
	headmem(NULL, NULL, 0);
	if (NULL != star->sta_fpmi) (void)fclose(star->sta_fpmi);
	if (NULL != star->sta_fpmis) (void)fclose(star->sta_fpmis);
	if ((NULL != star->sta_fpl) && (star->sta_fpl != stdout))
		(void)fclose(star->sta_fpl);
	star->sta_fpmi = star->sta_fpl = star->sta_fpmis = NULL;
	globalkill();	return irv;
}

int statops(STATRUN *star)
{ /*	This does most of the work in a stats run. It's used in scalem too */
	int		i, ir, lslim, sint, z[3];
	short		cmpval, excluco, hlim, klim, llim, hklrun[3];
	u_char		hkl[4], hklmr[4];
	CRYSTAL		*xhp;
	SHLSTAT		*shlrp;
	MERGEDREF	mr;

	/* process multiref header info */
	if (EOK != (i = stathead(star))) return endstat(i, star);
	xhp = star->sta_xhdp; excluco = exclucode(xhp->xt_intl, xhp->xt_isys);
	/*   Read the first reflection in from the multiref	*/
	if (-1 != (i = goodref(&mr, excluco, hklmr, star)))
		return endstat(i, star);
	/* loop through the possible hkl values.  If the current
	possible hkl value is the same as the current value in
	the multiref, get the next multiref value and increment
	the poss-hkl value.  If the possible hkl value is "earlier"
	than the current-multiref value, print the missing refl
	to the log file and increment the poss-hkl value.  */
	z[0] = (xhp->xt_isys == 2);	z[1] = 1 - z[0];	z[2] = 2;
	hlim = xhp->xt_ar[z[0]] / xhp->xt_reslim + 1.;
	for (hklrun[0] = 0; hklrun[0] <= hlim; (hklrun[0])++)
	 {
	  klim = xhp->xt_ar[z[1]] / xhp->xt_reslim + 1.;
	  for (hklrun[1] = 0; hklrun[1] <= klim; (hklrun[1])++)
	     {
	      llim = xhp->xt_ar[2] / xhp->xt_reslim + 1.;
	      if (3 < (lslim = star->sta_hklmax[3])) /* special case triclinic*/
	       {
		if (hklrun[0] == 0)
		  {
			lslim = (hklrun[1] == 0) ? 1 : 3;
		  }
		else if (hklrun[1] == 1) lslim = 3;
	       }
	      if ((xhp->xt_isys == 2) && (hklrun[1] == 0)) lslim = 1;
	      for (hklrun[2] = 0; hklrun[2] <= llim; (hklrun[2])++)
	       { /*note that we always unpack as if we were in P1,
		  since the monoclinic case has h and k ALREADY reversed */
		mpack(hklrun,hkl,1);	/* get indexed hkl(see above comment) */
		sint = (hklrun[2] == 0) ? 4 : 2;
		for (hkl[3] = 0; hkl[3] <= lslim; hkl[3] += sint)
		   if (islegal(hkl, excluco, star))
		     {
			star->sta_ir = ir =
				getir((double)(star->sta_res), star);
			shlrp = &(star->sta_shl[ir]);
			(shlrp->sh_nposs)++;
			if ((star->sta_flags & ST_SEPBJ) && (!iscentric(hkl,
				star->sta_spgrp))) (shlrp->sh_nposs)++;
			cmpval = hklcmp(hkl,hklmr);
			if (star->sta_flags & ST_DONE) misstat(star, hkl);
			  else
			      {
				if (cmpval < 0) misstat(star, hkl);
				else {
					if (cmpval > 0)
						(star->sta_nroutorder)++;
					if (-1 != (i = goodref(&mr, excluco,
					 hklmr, star))) return endstat(i, star);
				     }
			      }
		     }
	       }
	     }
	 }
	statlast(star);	return endstat(EOK, star);
}

int statmain(int argc, char *argv[])
{ /* mainline for stats functionality */
	int		i;
	STATRUN		sts;
  
	/* process arguments */
	if (EOK != (i = statargs(argc, argv, &sts))) return endstat(i, &sts);
	return statops(&sts);	/* carry out the statistical calculations */
}

/* end of statistic.c */
