/* bordmain.c -- generates XGEN's list of active pixels on the detector face

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions; reorganize use of "{"
	14 Feb 2007: bord_defaults: allow for Aviex detector types
	20 Nov 2005: inline function declarations & function prototypes
	29 Oct 2005: move merge_frames() and process_total() to
	 ../xglib/actmask.c; eliminate static globals
	11 Oct 2005: housekeeping
	11 May 2005: check return status from xgjournal
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	28 Aug 2004: routines used only here are declared as static; cosmetics
	19 Jun 2004: bord_defaults() created: defaults based on detector types
	21 May 2004: a few extra variables added to routines so that we
		can call the border functionality from within `spots'
	28 Feb 2004: be honest in comments about the non-existence of -k flag
	18 Feb 2004: change kill_ftab call so that it doesn't have args
	26 Jan 2004: replace a few static globals with local variables
	21 Jan 2004: license message inserted and initial comment reordered
	 6 Jan 2004: static globals replace extern globals;
		calls to kill_ftab and kill_mail near the end
	12 Dec 2003: eliminate unnecessary variable loctype in detshrink()
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	17 Apr 2003: rename `frameno' to `fmno'
	 5 Jan 2003: reorganized source-file header
	 2 Jan 2003: add -q flag (facilitates script-writing).
	15 Nov 2002: -s flag modified so that shrinkval is set to
		reasonable defaults for particular detectors
	15 Aug 2002: add -a flag (facilitates script-writing).
	25 Mar 2002: removed MSI attachments
	31 May 1995: "register" -> "register int" throughout;
		replace longs with ints
	12 May 1995: make_ftab takes no argument; totalframes is a global
	12 Apr 1995: *framename is now an extern global
	 2 Mar 1995: diagnostics to stdout handle > 2^31 counts correctly
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 1 Dec 1994: diagnostic output allows for detector (X,Y) > 999
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: minor bugfixes, reorganized libraries
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: Initial X-GEN version
	xx xxx 1993: in process_total(), dbmask is now of variable length
	21 Jan 1988: in process_total(), smoothing is done to
		prevent the appearance of isolated DEAD points

   Purpose: generating the file defining the active area of the detector.
   Cross references: ../utilities/editbord.c

	The active area is approximately a circle truncated by a rectangle
	at the top, bottom and sides, and may also be missing a few pixels
	along the interior where either a beamstop or a beamstop hanger
	produces a sufficiently dense shadow that it does not allow
	much of the intensity through.
*************************************************************************/
#include	"xgen.h"
#define		DEFAULT_QVAL		(0.35)
#define		DEFAULT_NFRAMES		25
#define		MIN_IN_IMAGES	(65533)	/*flag value for "-d" option */
	/* bits in the bwp->bw_runflags variable */
#define		BORD_VERBOSE		1

 typedef struct	{
		int		bw_endfm, bw_firstrange, bw_imflags;
		int		bw_runflags, bw_startfm, bw_valshrink;
		double		bw_qvalu;
		FILE		*bw_fpl;
		FRAMEINDEX	*bw_fmi;
		} BORDWORK;

extern int bordmain(int argc, char *argv[]);

static void sum_brun(FILE *fp, BORDWORK *bwp)
{	/* summarizes a run of `border' to the stream `fp' */
	static char	*flist[] = {"frames", "xlog", "amask", NULL };
	static char	ftypes[] = {1, 2, 2, 0};

	ftypes[1] = (bwp->bw_fpl == stdout ? 0 : 2);
	header(fp, PROCESSING, flist, ftypes);
	fprintf(fp,"\tFrames from %d to %d will be averaged to get the mask\n",
		bwp->bw_startfm, bwp->bw_endfm);
	fprintf(fp,
	 "\tA pixel will be defined as dead if count/<count> < %6.3lf\n",
	 bwp->bw_qvalu);
	fprintf(fp,
	 "\t%s rows of the binary mask will be listed in the log file\n",
		(bwp->bw_runflags & BORD_VERBOSE) ? "All" : "Selected");
	if (bwp->bw_valshrink) fprintf(fp,
 "\tThe edges of the active region will be pulled inward by %3d pixels\n",
		bwp->bw_valshrink);
	if (bwp->bw_imflags & DARK_CURRENT) fprintf(fp,
 "\tDark current from %s will be subtracted from all pixels before stats\n",
		trname("DARKIMAGE"));
	fprintf(fp,"\tFrames read in the effort to compute the dead region:\n");
	fprintf(fp,
 " Frame name                 Total    Max      at   Max Bgnd      # Bgnd\n");
	fprintf(fp,
 "                           Counts  Count    X    Y    count      counts\n");
}

static int detshrink(int detwi, BORDWORK *bwp)
{ /*	Returns shrinkage value characteristic of the current detector type. */
	int	detdim[2];

	if (detwi < 1)
	  {
		(void)dettype(detdim);	detwi = detdim[0];
	  }
	if (detwi < 256) return 3;
	else if (detwi < 384) return 4;
	else if (detwi < 512) return 5;
	else if (detwi < 768) return 6;
	else return detwi / 128;
}

static int bord_defaults(BORDWORK *bwp)
{ /*	This establishes some defaults, partly based on detector type */
	int		detype, stype, dedim[2];
	static int	nbtype[] = /* # of frames to search for spots */
	 { 30, 30, 30, 20, 30, 6, 6, 6, 5, 6, 8, 6, 6, 6, 6, 6, 6, 6, 6} ;
	static double	qvu[] = /* Count/<Count> below which pixel is dead*/
	 { 0.38, 0.36, 0.37, 0.38, 0.4, 0.37, 0.37, 0.4, 0.4, 0.37, 0.42,
	   0.38, 0.38, 0.41, 0.42, 0.42, 0.42, 0.42, 0.42 } ;

	if (NULL == (bwp->bw_fmi = ftables(3))) return EOPENERR;
	if (NULL == (bwp->bw_fpl = propen("xlog"))) bwp->bw_fpl = stdout;
	bwp->bw_startfm = 32761;
	bwp->bw_endfm = -32761;	bwp->bw_qvalu = 9.999e12;
		/* images treaded normally */
	bwp->bw_imflags = bwp->bw_runflags = bwp->bw_firstrange = 0;
	/* determine what kind of detector we have */
	if (0 >= (detype = dettype(dedim))) return EREADERR;
	stype = detype & 0377;
	/* shrink border in by a number of pixels related to detector size */
	if (dedim[0] <= 512) bwp->bw_valshrink = 6;
	else if (dedim[0] <= 1024) bwp->bw_valshrink = 9;
	else if (dedim[0] <= 2048) bwp->bw_valshrink = 16;
	else bwp->bw_valshrink = 20;
	/* by default we begin with the first frame */
	bwp->bw_startfm = (bwp->bw_fmi)->fi_firstfm;
	/* decide which frame we'll end with and set the count/<count> val */
	if (stype <= 18)
	  {
		bwp->bw_endfm = bwp->bw_startfm + nbtype[stype];
		bwp->bw_qvalu = qvu[stype];
	  }
	else {
		bwp->bw_endfm = bwp->bw_startfm + 6;	bwp->bw_qvalu = 0.38;
	     }
	if (bwp->bw_endfm > (bwp->bw_fmi)->fi_lastfm)
		bwp->bw_endfm = (bwp->bw_fmi)->fi_lastfm;
	return EOK;
}

static int bordargs(int argc, char *argv[], BORDWORK *bwp)
{ /*	argument handler for `border'.
	Syntax:		border [-flags] [fm_min fm_max [Q] ]
		where fm_min and fm_max are the frames examined
			(default: 0 through 18)
		Q is the value such that if (count/<count>) < Q we
		set the mask value to 0 and set it to 1 otherwise
			(default: Q = 0.35)
		Currently defined flags:
			-a<val>	Use the first <val> images in the FRAMES
				file to do the calculation. This will
				supersede the fm_min and fm_max values
				if those are specified.
			-d	subtract dark current before doing statistics;
				by default this causes the program to look at
				(raw val - 0.8 * dark current val), but
				if a numeric argument follows the "d" it is
				taken to be the dark-current subtraction value.
				Thus "border -d1" will cause the full value of
				the dark-current to be subtracted.
			-k	keep the global variables alive after the run.
				This is suitable if `bordmain' is being
				called from within something else.
			-q<val>	Like the Q value described above, but
				expressed as a switch rather than a parameter.
			-s<val>	shrink the active area inward by <val> pixels
				on all sides.
			-v	verbose output:
				print every column of the result
				rather than every 16th column
		N.B. the -k flag described above isn't actually implemented yet
*/
	char		**dcp, **flags;
	int		i;
	double		thisq;
	static int	idarkfrac =  6400;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	if (EOK != (i = bord_defaults(bwp))) return i;
	flags = flagproc(argc, argv, "ADQSV");
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	   {
		if (**dcp == 'D')
		  {	bwp->bw_imflags |= DARK_CURRENT;
			if ('\0' != *((*dcp) + 1))
				idarkfrac = 8000. * atof((*dcp) + 1);
		  }
		else if (**dcp == 'A') bwp->bw_firstrange = atoi((*dcp)+1);
		else if (**dcp == 'Q')
		  {
			if ((0 <= (thisq = atof((*dcp) + 1))) && (thisq <= 1))
				bwp->bw_qvalu = thisq;
		  }
		else if (**dcp == 'S') bwp->bw_valshrink = atoi((*dcp)+1);
		else if (**dcp == 'V') bwp->bw_runflags |= BORD_VERBOSE;
	   }
	if (idarkfrac > 16000) idarkfrac = 16000;
	if (bwp->bw_imflags & DARK_CURRENT)
		bwp->bw_imflags |= (idarkfrac << 16);
	for (dcp = argv+1, argc--; argc--; dcp++)
	   if (allnumeric(*dcp))
	     {
		if (bwp->bw_startfm > 32760) bwp->bw_startfm = atoi(*dcp);
		else if (bwp->bw_endfm < -32760) bwp->bw_endfm = atoi(*dcp);
		else if (bwp->bw_qvalu > 1.e12) bwp->bw_qvalu = atof(*dcp);
	     }
	/* flag set up Aug 02: -a<val> specifies that
	 the first <val> images will be used. We require that the
	 <val> be no larger than the number of images in the FRAMES file */
	if (bwp->bw_firstrange > 0)
	  {
		if (bwp->bw_firstrange > (bwp->bw_fmi)->fi_totfms)
			bwp->bw_firstrange = bwp->bw_fmi->fi_totfms;
		bwp->bw_startfm = bwp->bw_fmi->fi_firstfm;
		bwp->bw_endfm = ((bwp->bw_fmi)->fi_fmi +
			(bwp->bw_firstrange-1))->fmno;
	  }
	/* -s flag modified Nov 02: unless user has specifically said -s0 or
	 -s<val>, with <val> greater than zero, we set bwp->bw_valshrink
	 to a value associated with the detector dimensions */
	if (bwp->bw_valshrink < 0) bwp->bw_valshrink = detshrink(0, bwp);
	if (bwp->bw_startfm > 32760)
		bwp->bw_startfm = (bwp->bw_fmi)->fi_firstfm;
	if (bwp->bw_endfm < -32760)
		bwp->bw_endfm = bwp->bw_startfm + DEFAULT_NFRAMES - 1;
	if (bwp->bw_qvalu > 1.e12) bwp->bw_qvalu = DEFAULT_QVAL;
	sum_brun(bwp->bw_fpl, bwp);
	if (bwp->bw_fpl != stdout) sum_brun(stdout, bwp);
	return EOK;
}

static void sum_bfrm(FILE *fp, MAIL *headp, double goodc, int nom_bgnd)
{ /*	This summarizes one frame's worth of input to the stream fp */
	double	revent;

	revent = headp->eventcount;	if (revent < 0.) revent *= -64.;
	fprintf(fp, " %16s %15.0f %6d %4d %4d %8d %15.0f\n",
		headp->filename, revent, (int)headp->zmax,
		(int)headp->zmaxx, (int)headp->zmaxy, nom_bgnd, goodc);
}

int bordmain(int argc, char *argv[])
{ /* The mainline for the standalone border-determination functionality */
	register int	fmno;
	int		nom_bgnd, read_result, totincl;
	double		totalcounts;
	static int	nmbox;
	static char	*framename;
	static MAIL	*heads;
	static BORDWORK	*bwp;

	if (NULL == bwp)
	  {
		if (NULL == (bwp = (BORDWORK *)callop(1, sizeof (BORDWORK),
		 "bwp in bordmain"))) return EALLOC;
	  }
	if (EOK != (read_result = bordargs(argc, argv, bwp)))
		return read_result;
	if (NULL == heads)
	  {
		nmbox = 1;
		if (NULL == (AVOID *)(heads = (MAIL *)callop(nmbox,
		 sizeof (MAIL), "heads in bordmain"))) return EALLOC;
	  }
	for (fmno = bwp->bw_startfm, totalcounts = 0.;
	 fmno <= bwp->bw_endfm; fmno++)
	   {
		framename = indexname(fmno);
		 /* read in each frame and deal with it */
		if (FRAME_OK == (read_result = 
		  read_image(framename, bwp->bw_imflags, heads)))
		  {
			totalcounts += merge_frames(heads, &nom_bgnd);
			sum_bfrm(bwp->bw_fpl, heads, totalcounts, nom_bgnd);
			if (bwp->bw_fpl != stdout)
				sum_bfrm(stdout, heads, totalcounts, nom_bgnd);
		  }
		else  {
			(void)image_error(bwp->bw_fpl, framename,
				read_result, 0);
			if (bwp->bw_fpl != stdout)
			 (void)image_error(stderr,framename, read_result, 0);
		      }
	   }
	head_results(bwp->bw_fpl);
	totincl = process_total(bwp->bw_fpl, totalcounts, bwp->bw_qvalu,
		(int)heads->xdimen, (int)heads->ydimen, bwp->bw_runflags);
	fprintf(bwp->bw_fpl,
	 "\n\t%6d pixels will be included in the active-pixel-mask output\n",
		totincl);
	if (bwp->bw_fpl != stdout) fprintf(stdout,
	 "\t%6d pixels will be included in the active-pixel-mask output\n",
		totincl);
	/* if the user requests it, shrink the mask by
	 `bwp->bw_valshrink' pixels on all sides */
	if (bwp->bw_valshrink) shrink(bwp->bw_valshrink,
	 (int)(heads->xdimen), (int)(heads->ydimen));
	if ((NULL != bwp->bw_fpl) && (stdout != bwp->bw_fpl))
		(void)fclose(bwp->bw_fpl);
	bwp->bw_fpl = NULL;
	if (NULL == activmask((int)(heads->xdimen), (int)(heads->ydimen), 011))
	  {
		fprintf(stderr, "border: Cannot write out active--pixel mask");
		read_result = EWRITERR;
	  }
	else	read_result = EOK;
	(void)ftables(0);	kill_mail(&heads, &nmbox);
	globalkill();	/* free mem in run */
	return read_result ? EWRITERR : EOK;
}

/* end of bordmain.c */
