/* calcetool.c -- utilities for handling CALCENTROIDS (.cci) files

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	30 May 2007: splitc handles 2-element systems as well as 4;
	 splitcerr used for error msgs
	19 Apr 2007: float->double conversions; reorganize use of "{"
	 5 Jun 2006: luycmp is now static here
	12 May 2006: define ecalmain, missimain here
		rather than ../include/calibdefs.h;
		merge {editcal,missind,uca_2c,splitcen}.c -> calcetool.c
	 3 Jan 2006: don't qsort with fewer than 2 elements
	 8 Dec 2005: remove memzap of VSPLINE stuff -- globalkill does it
	20 Nov 2005: inline function declarations & function prototypes
	28 Oct 2005: eliminate static globals and do some housekeeping
	28 Jun 2005: writecalce(): typecast pointer differences upon output
	22 Jun 2005: ecalmain(): "command" -> "comd" to reduce confusion;
	17 Jun 2005: cci2ucamain(): putspline() takes VSPLINE argument
	11 May 2005: check return status from xgjournal
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	 6 Sep 2004: cosmetic change in missind and splitcmain
	28 Aug 2004: routines used only here declared as static; cosmetics
	 2 Aug 2004: add new arg to find_dimension()
	26 Jan 2004: convert most of the static globals into local variables
	21 Jan 2004: license message inserted and initial comment reordered
	 6 Jan 2004: CALIBRUN structure elements in function calls
		& static globals replace extern globals
	20 Nov 2003: replace global variables with elements of CALIBRUN struct
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	 5 Jan 2003: reorganized source-file header
	20 Sep 2002: CF_* flags moved to ../include/xgen.h
	25 Mar 2002: removed MSI attachments
	 9 Oct 1998: initial version of splitcen
	25 Nov 1997: initial version of cci2uca.c, based on `mergeuca'
	15 Nov 1997: initial version of editcal
	11 Nov 1997: initial version of missind
   Purpose: various specific functions for manipulating CALCENTROIDS:
	cci2uca	reads a CALCENTROIDS and writes a USPLINE,
	editcal	provides a text-based editor for CALCENTROIDS files
	missind	determines which CALCENTROIDS points are missing or dup'd
	uca_2c	reads a USPLINE and writes out a CALCENTROIDS
	splitcen reads a CALCENTROIDS and splits it into 4 separate ones.
   Cross references: ../prerefine/cal*.c
*************************************************************************/
#include	"calibdefs.h"
#define	MAXSPOTS	14000	/* max # of recordable spots in image*/
#define	MAX_ARGS	9	/* maximum number of editing arguments */
#define	CHARS_PER_ARG	20	/* max # of characters per argument */

/* Values of the "verb" variable in editcal. We'll need more later. */
#define	ECCOLUMN	1	/* ix */
#define	ECROW		2	/* iy */
#define	ECQUIT		4	/* leave program */
#define	ECCOLUMNS	010	/* several ix's */
#define	ECROWS		020	/* several iy's */
#define	ECWRITE		040	/* write out CALCENTROIDS */
#define	ECEXIT		044	/* ECQUIT | ECWRITE */
#define	ECCOLINTERP	0100	/* interpolate in a column */
#define	ECROWINTERP	0200	/* interpolate in a row */
#define	ECNEWPOINT	0400	/* put in a new point */
#define	ECCOLSINTERP	01000	/* interpolate in columns */
#define	ECROWSINTERP	02000	/* interpolate in rows */
#define	INSUFFARGS	010000	/* not enough args for cmd */
#define	UNKNOWNVERB	020000	/* not enough args for cmd */

 typedef struct {
		int		cc_maxux, cc_maxwx, cc_maxuy, cc_maxwy;
		int		cc_minux, cc_minuy, cc_keepinregion;
		} CCIWORK;

 typedef struct {
		char		**eca_cargs;
		unsigned	eca_ncargs;
		int		eca_nind, eca_iargs[8];
		} ECAWORK;

extern int outindcen(VSPLINE *lsp);

extern int ecalmain(int argc, char *argv[]);
extern int missimain(int argc, char *argv[]);
extern int uca2cmain(int argc, char *argv[]);
extern int splitcmain(int argc, char *argv[]);

/*****************************
	editcal: These routines enable editing of a CALCENTROIDS file.
	For the moment it will just take arguments from the terminal;
	we'll make it a window-based software package eventually.
*****************************/
static char *get_eccom(ECAWORK *ecw)
{ /*	This prompts the user for a command and returns a lowercase version
	of the command given by the user.  It also stuffs any other strings
	given by the user into ecw->eca_cargs. */
	register int	i, j, lnca, nitems;
	char		**dcp;
	static char	cmd[CHARS_PER_ARG], ecline[80];

	fprintf(stderr, "> ");	for (j = CHARS_PER_ARG; j--; ) cmd[j] = '\0';
	for (i = MAX_ARGS; i--; ) for (j = CHARS_PER_ARG; j--; )
		ecw->eca_cargs[i][j] = '\0';
	if (NULL == fgets(ecline, 78, stdin))
	  {
		strcpy(cmd, "ex");	return &(cmd[0]);
	  }
	dcp = ecw->eca_cargs;
	nitems = sscanf(ecline, "%s %s %s %s %s %s %s %s %s %s",
		cmd, *dcp, *(dcp + 1), *(dcp + 2), *(dcp + 3), *(dcp + 4),
		*(dcp + 5), *(dcp + 6), *(dcp + 7), *(dcp + 8));
	lowercaseify(cmd);
	for (lnca = 0; lnca < MAX_ARGS && ecw->eca_cargs[lnca][0] != '\0';
	  lnca++) lowercaseify(ecw->eca_cargs[lnca]);
	if (lnca != nitems - 1) fprintf(stderr,
	 "Discrepancy in argument number: lnca = %d, nitems - 1 = %d\n",
		lnca, nitems - 1);
	ecw->eca_ncargs = lnca;	return &(cmd[0]);
}

static void log_ecchange(int verb, FILE *fp, ECAWORK *ecw)
{ /* summarize a change in the calcentroids file to the stream fp. */
	char	*eccommand;
	int	*icp;
	
	icp = &(ecw->eca_iargs[0]);	eccommand = *(ecw->eca_cargs);
	switch (verb)
	      {
		case		ECCOLUMN: fprintf(fp,
		 " %5d points in box from [%5d,%5d] to [%5d,%5d]",
		 *(icp + 5), *(icp + 1), *(icp + 2), *(icp + 3), *(icp + 4));
			fprintf(fp," assigned to column number %3d\n", *icp);
			break;
		case		ECROW:	fprintf(fp,
		 " %5d points in box from [%5d,%5d] to [%5d,%5d]",
		 *(icp + 5), *(icp + 1), *(icp + 2), *(icp + 3), *(icp + 4));
			fprintf(fp, "assigned to  row   number %3d\n", *icp);
			break;
		case		ECCOLUMNS:	fprintf(fp,
		 " %5d points in box from [%5d,%5d] to [%5d,%5d]\n",
		 *(icp + 7), *(icp + 1), *(icp + 2), *(icp + 3), *(icp + 4));
			fprintf(fp, "assigned to column numbers %3d to %3d\n",
				 *icp, *(icp + 6));	break;
		case		ECROWS:	fprintf(fp,
		 " %5d points in box from [%5d,%5d] to [%5d,%5d]", *(icp + 7),
		 *(icp + 1), *(icp + 2), *(icp + 3), *(icp + 4));
			fprintf(fp, "assigned to   row  numbers %3d to %3d\n",
			 *icp, *(icp + 6));	break;
		case		ECCOLINTERP:	fprintf(fp,
		 " %5d points interpolated into column %d\n",
			*(icp + 1), *icp); break;
		case		ECROWINTERP:	fprintf(fp,
		 " %5d points interpolated into row %d\n",
			*(icp + 1), *icp); break;
		case		ECCOLSINTERP:	fprintf(fp,
		 " %5d points interpolated into columns %d to %d\n",
			*(icp + 2), *icp, *(icp + 1));	break;
		case		ECROWSINTERP:	fprintf(fp,
		 " %5d points interpolated into rows %d to %d\n",
			*(icp + 2), *icp, *(icp + 1));	break;
		case		ECNEWPOINT:	fprintf(fp,
		 " Point at [%5d,%5d] indexed as [%2d,%2d] has been added\n",
		 *icp, *(icp + 1), *(icp + 2), *(icp + 3));
			break;
		case		ECQUIT:	fprintf(fp,
		 " quitting from applicption");	break;
		case		ECWRITE:	fprintf(fp,
		 " writing %6d CALCENTROIDS to output; %6d are indexed\n",
		 *icp, *(icp + 1));	break;
		case		ECEXIT:	fprintf(fp,
		 " writing %6d CALCENTROIDS to output; %6d are indexed\n",
		 *icp, *(icp + 1));
			fprintf(fp, "exiting from applicption\n"); break;
		case		INSUFFARGS:	fprintf(fp,
		 " command %s requires at least %3d arguments; got only %2d\n",
			eccommand, *icp, *(icp + 1));	break;
		case		UNKNOWNVERB:	fprintf(fp,
		 " command %s is an unknown verb\n", eccommand); break;
	     }
}

static int find_col(CALIBRUN *cap, ECAWORK *ecw)
{ /*	Find all the points in a certain column and set their column value */
	int	i, *icp;
	char	**dcp;
	double	xmi, xma, ymi, yma;
	CTROID	*lct;

	icp = &(ecw->eca_iargs[0]);	dcp = ecw->eca_cargs;
	for (i = 0; i < 5; i++, icp++, dcp++) *icp = atoi(*dcp);
	icp -= 4; xmi = *icp++; ymi = *icp++; xma = *icp++; yma = *icp;
	for (i = 0, lct = cap->car_ctr0; lct < cap->car_ctend; lct++)
	   {
		if ((xmi <= lct->cxcen) && (lct->cxcen <= xma) &&
		  (ymi <= lct->cycen) && (lct->cycen <= yma))
		  {
			i++;
			if ((lct->uy > 0) && (lct->ux <= 0)) (ecw->eca_nind)++;
			lct->ux = ecw->eca_iargs[0];
			fprintf(stderr, "Setting ux = %d for [%6.1f,%6.1f]\n",
				lct->ux, lct->cxcen, lct->cycen);
		  }
	   }
	cap->car_ctr = lct;	return i;
}

static int find_row(CALIBRUN *cap, ECAWORK *ecw)
{ /*	This finds all the points in a certain row and sets their row value */
	int	i, *icp;
	char	**dcp;
	double	xmi, xma, ymi, yma;
	CTROID	*lct;

	icp = &(ecw->eca_iargs[0]);	dcp = ecw->eca_cargs;
	for (i = 0; i < 5; i++, icp++, dcp++) *icp = atoi(*dcp);
	icp -= 4; xmi = *icp++; ymi = *icp++; xma = *icp++; yma = *icp++;
	for (i = 0, lct = cap->car_ctr0; lct < cap->car_ctend; lct++)
	   {
		if ((xmi <= lct->cxcen) && (lct->cxcen <= xma) &&
		  (ymi <= lct->cycen) && (lct->cycen <= yma))
		  {
			i++;	lct->uy = ecw->eca_iargs[0];
			if ((lct->ux > 0) && (lct->uy <= 0)) (ecw->eca_nind)++;
			fprintf(stderr, "Setting uy = %d for [%6.1f,%6.1f]\n",
				lct->uy, lct->cxcen, lct->cycen);
		  }
	   }
	cap->car_ctr = lct;	return i;
}

static int find_cols(CALIBRUN *cap, ECAWORK *ecw)
{ /*	This finds all points in a range of columns
	 and sets their column values */
	int	i, *icp;
	char	**dcp;
	double	xmi, xma, ymi, yma, widcol;
	CTROID	*lct;

	icp = &(ecw->eca_iargs[0]);	dcp = ecw->eca_cargs;
	for (i = 0; i < 6; i++, icp++, dcp++) *icp = atoi(*dcp);
	icp -= 4; xmi = *icp++; ymi = *icp++; xma = *icp++; yma = *icp++;
	widcol = *icp;	icp = &(ecw->eca_iargs[0]);
	for (i = 0, lct = cap->car_ctr0; lct < cap->car_ctend; lct++)
	   {
		if ((xmi <= lct->cxcen) && (lct->cxcen <= xma) &&
		  (ymi <= lct->cycen) && (lct->cycen <= yma))
		  {
			i++;	lct->ux = *icp + (lct->cxcen - xmi) / widcol;
			if ((lct->uy > 0) && (lct->ux <= 0)) (ecw->eca_nind)++;
		  }
	   }
	ecw->eca_iargs[6] = ecw->eca_iargs[0] + (xma - xmi) / widcol;
	cap->car_ctr = lct;	return i;
}

static int find_rows(CALIBRUN *cap, ECAWORK *ecw)
{ /*	This finds all the points in a certain row and sets their row value */
	int	i, *icp;
	char	**dcp;
	double	xmi, xma, ymi, yma, widrow;
	CTROID	*lct;

	icp = &(ecw->eca_iargs[0]);	dcp = ecw->eca_cargs;
	for (i = 0; i < 6; i++, icp++, dcp++) *icp = atoi(*dcp);
	icp -= 4; xmi = *icp++; ymi = *icp++; xma = *icp++; yma = *icp++;
	widrow = *icp;	icp = &(ecw->eca_iargs[0]);
	for (i = 0, lct = cap->car_ctr0; lct < cap->car_ctend; lct++)
	   {
		if ((xmi <= lct->cxcen) && (lct->cxcen <= xma) &&
		  (ymi <= lct->cycen) && (lct->cycen <= yma))
		  {
			i++;	lct->uy = *icp + (lct->cycen - ymi) / widrow;
			if ((lct->ux > 0) && (lct->uy <= 0)) (ecw->eca_nind)++;
		  }
	   }
	*(icp + 6) = *icp + (yma - ymi) / widrow;
	cap->car_ctr = lct;	return i;
}

static int luxcmp(CTROID *ctq0, CTROID *ctq1)
{ /*	This returns a value used in sorting on x */
	if (ctq0->ux < ctq1->ux) return -1;
	else if (ctq0->ux > ctq1->ux) return 1;
	else if (ctq0->uy < ctq1->uy) return -1;
	else if (ctq0->uy > ctq1->uy) return 1;
	else return 0;
}

static int luycmp(CTROID *ctq0, CTROID *ctq1)
{ /*	This returns a value used in sorting on y */
	if (ctq0->uy < ctq1->uy) return -1;
	else if (ctq0->uy > ctq1->uy) return 1;
	else if (ctq0->ux < ctq1->ux) return -1;
	else if (ctq0->ux > ctq1->ux) return 1;
	else return 0;
}

static int writecalce(CALIBRUN *cap)
{ /*	This writes indexed centroids to a file rather than doing a USPLINE */
	char		*calnam;
	int		ntr;
	FILE		*fpet;
	CTROID		*lct;

	calnam = trname("CALCENTROIDS");
	if (NULL == (fpet = xgopen(calnam, "w")))
	  {
		fprintf(stderr, "Unable to open %s for output\n",
			calnam);
		return EOPENERR;
	  }
	/* sort before output */
	ntr = cap->car_ctend - cap->car_ctr0;
	if (ntr > 1) qsort((void *)(cap->car_ctr0), (size_t)ntr,
	 sizeof (CTROID), (QPE)luxcmp);
	for (lct = cap->car_ctr0; lct < cap->car_ctend; lct++)
	   {
		fprintf(fpet, " %7.2f %7.2f %3d %3d\n",
			lct->cxcen, lct->cycen, lct->ux, lct->uy);
	   }
	printf(" %6d reflections written to %s\n",
		(int)(cap->car_ctend - cap->car_ctr0), calnam);
	cap->car_ctr = lct;	(void)fclose(fpet);	return EOK;
}

static int ptinterp(int type, int newind,
 CTROID *closest, CTROID *nextclosest, CALIBRUN *cap)
{ /*	This adds a new point to the list */
	double	cx, cy, delta;
	int	oldcl, oldnext, other;
	CTROID	*lct;

	if (type == ECCOLINTERP)
	  {
		oldcl = closest->uy;	oldnext = nextclosest->uy;
		other = closest->ux;
	  }
	else {
		oldcl = closest->ux;	oldnext = nextclosest->ux;
		other = closest->uy;
	     }
	if (oldcl == oldnext)
	  {
		fprintf(stderr,
		 "Interpolation failed: no observed points in col/row\n");
		return 0;	/* error */
	  }
	delta = ((double)(newind - oldcl)) / ((double)(oldcl - oldnext));
	cx = closest->cxcen + delta *
		(closest->cxcen - nextclosest->cxcen);
	cy = closest->cycen + delta *
		(closest->cycen - nextclosest->cycen);
	lct = cap->car_ctend;	lct->cxcen = cx;	lct->cycen = cy;
	if (type == ECCOLINTERP)
	  {
		lct->ux = other;	lct->uy = newind;
	  }
	else {
		lct->uy = other;	lct->ux = newind;
	     }
	fprintf(stderr, "Interp [%6.1f,%6.1f,%3d,%3d] near ",
		cx, cy, lct->ux, lct->uy);
	fprintf(stderr, "[%6.1f,%6.1f,%3d,%3d],[%6.1f,%6.1f,%3d,%3d]\n",
	 closest->cxcen, closest->cycen, closest->ux, closest->uy,
	 nextclosest->cxcen, nextclosest->cycen,
	 nextclosest->ux, nextclosest->uy);
	(cap->car_ctend)++;	return 1;
}

static int interpcol(int colno, CALIBRUN *cap)
{ /*	this inserts additional points into column `colno' by interpolation
	and (gulp) extrapolation. It returns the number of points inserted. */
	int	foundit, guyma, guymi, icl, mi2, mindi, ret, tuyma, tuymi, ty;
	CTROID	*lct, *ctq, *ctsc, *ctec, *ctcl, *ct2;

	/* find the extrema of the row indices for the entire data set
	 and for this column */
	tuymi = guymi = 10000;	tuyma = guyma = -guymi;
	for (lct = cap->car_ctr0, icl = 0; lct < cap->car_ctend; lct++)
	   {
		if ((ty = lct->uy) > 0)
		  {
			if (ty < guymi) guymi = ty;
			if (ty > guyma) guyma = ty;
			if (lct->ux == colno)
			  {
				if (ty < tuymi) tuymi = ty;
				if (ty > tuyma) tuyma = ty;
				icl++;
			  }
		  }
	   }
	printf(" %5d spots found with ux = %d; tuymi, ma = %d, %d\n",
		icl, colno, tuymi, tuyma);
	/* sort the list on ux */
	ty = cap->car_ctend - cap->car_ctr0;
	if (ty > 1) qsort((void *)(cap->car_ctr0), (size_t)ty,
	 sizeof (CTROID), (QPE)luxcmp);
	ctq = cap->car_ctend - 1;	lct = cap->car_ctr0;
	printf(
	 "sorted: from [%6.1f,%6.1f,%3d,%3d] to [%6.1f,%6.1f,%3d,%3d]\n",
	 lct->cxcen, lct->cycen, lct->ux, lct->uy,
	 ctq->cxcen, ctq->cycen, ctq->ux, ctq->uy);
	/* find the beginning of the ux values that equal colno */
	for (lct = cap->car_ctr0, ctsc = NULL; lct < cap->car_ctend; lct++)
	   {
		if (lct->ux >= colno)
		  {
			if (lct->ux == colno) ctsc = lct; break;
		  }
	   }
	if (ctsc == NULL)
	  { /*no pts with this column index!*/
		printf("no points in column %2d\n", colno);
		return 0;
	  }
	printf("first point in column: [%6.1f,%6.1f,%3d,%3d]\n",
		ctsc->cxcen, ctsc->cycen, ctsc->ux, ctsc->uy);
	for (lct = cap->car_ctend - 1, ctec = NULL; lct >= ctsc; lct--)
	   {
		if (lct->ux <= colno)
		  {
			if (lct->ux == colno) ctec = lct; break;
		  }
	   }
	if (ctec == NULL)
	  {
		printf("ending-point error with [%6.1f,%6.1f,%3d,%3d]\n",
		 lct->cxcen, lct->cycen, lct->ux, lct->uy);
		return 0;	/* shouldn't happen */
	  }
	printf(" last point in column: [%6.1f,%6.1f,%3d,%3d]\n",
		ctec->cxcen, ctec->cycen, ctec->ux, ctec->uy);
	for (ret = 0, ty = guymi; ty <= guyma; ty++)
	   {
		for (foundit = 0, lct = ctsc; lct <= ctec; lct++)
		   {
			if ((lct->ux == colno) && (lct->uy == ty))
			  {
				foundit = 1;	break;
			  }
		   }
		if (!foundit) /* missing point */
		  {
			fprintf(stderr, "column search for [%3d,%3d]: ",
			 colno, ty);
			mindi = 10000;	ctcl = NULL;
			for (ctq = ctsc; ctq <= ctec; ctq++)
			   {
				icl = abs(ctq->uy - ty);
				if (icl < mindi)
				  {
					mindi = icl;	ctcl = ctq;
				  }
			   }
			if (NULL != ctcl)
			  {
				mi2 = 10000;	ct2 = NULL;
				for (ctq = ctsc; ctq <= ctec; ctq++)
				   {
					icl = abs(ctq->uy - ty);
					if ((ctq != ctcl) && (icl < mi2))
					  {
						mi2 = icl; ct2 = ctq;
					  }
				   }
				if ((mi2 < 1000) && (NULL != ct2)) /* success*/
				  ret += ptinterp(ECCOLINTERP, ty,
					ctcl, ct2, cap);
				else	fprintf(stderr,
					"failed to interpolate\n");
			  }
			else	fprintf(stderr, "failed to interpolate\n");
		  }
		else printf(" [%6.1f,%6.1f] found at [%3d,%3d]\n",
		 lct->cxcen, lct->cycen, lct->ux, lct->uy);
	   }
	return ret;
}

static int interprow(int rowno, CALIBRUN *cap)
{ /*	this inserts additional points into row `rowno' by interpolation
	and (gulp) extrapolation. It returns the number of points inserted. */
	int	foundit, guxma, guxmi, icl, mi2, mindi, ret, tuxma, tuxmi, tx;
	CTROID	*lct, *ctq, *ctsr, *cter, *ctcl, *ct2;

	/* find the extrema of the column indices for the entire data set
	 and for this row */
	tuxmi = guxmi = 10000;	tuxma = guxma = -guxmi;
	for (lct = cap->car_ctr0, icl = 0; lct < cap->car_ctend; lct++)
	   {
		if ((tx = lct->ux) > 0)
		  {
			if (tx < guxmi) guxmi = tx;
			if (tx > guxma) guxma = tx;
			if (lct->uy == rowno)
			  {
				if (tx < tuxmi) tuxmi = tx;
				if (tx > tuxma) tuxma = tx;
				icl++;
			  }
		  }
	   }
	printf(" %5d spots found with uy = %d; tuxmi, ma = %d, %d\n",
		icl, rowno, tuxmi, tuxma);
	/* sort the list on uy */
	tx = cap->car_ctend - cap->car_ctr0;
	if (tx > 1) qsort((void *)(cap->car_ctr0), (size_t)tx,
	 sizeof (CTROID), (QPE)luycmp);
	/* find the beginning of the uy values that equal rowno */
	for (lct = cap->car_ctr0, ctsr = NULL; lct < cap->car_ctend; lct++)
	   {
		if (lct->uy >= rowno)
		  {
			if (lct->uy == rowno) ctsr = lct; break;
		  }
	   }
	if (ctsr == NULL)
	  {
		printf("no points in    row %2d\n", rowno);
		return 0; /*no pts with this row index!*/
	  }
	printf("first point in    row: [%6.1f,%6.1f,%3d,%3d]\n",
		ctsr->cxcen, ctsr->cycen, ctsr->ux, ctsr->uy);
	for (ret = 0, lct = cap->car_ctend - 1, cter = NULL; lct >= ctsr; lct--)
	   {
		if (lct->uy <= rowno)
		  {
			if (lct->uy == rowno) cter = lct; break;
		  }
	   }
	if (cter == NULL)
	  {
		printf("ending-point error with [%6.1f,%6.1f,%3d,%3d]\n",
		 lct->cxcen, lct->cycen, lct->ux, lct->uy);
		return 0;	/* shouldn't happen */
	  }
	printf(" last point in    row: [%6.1f,%6.1f,%3d,%3d]\n",
		cter->cxcen, cter->cycen, cter->ux, cter->uy);
	for (tx = guxmi; tx <= guxma; tx++)
	   {
		for (foundit = 0, lct = ctsr; lct <= cter; lct++)
		   {
			if ((lct->uy == rowno) && (lct->ux == tx))
			  {
				foundit = 1;	break;
			  }
		   }
		if (!foundit) /* missing point */
		  {
			fprintf(stderr, "   row search for [%2d,%2d]: ",
			 tx, rowno);
			mindi = 10000;	ctcl = NULL;
			for (ctq = ctsr; ctq <= cter; ctq++)
			   {
				icl = abs(ctq->ux - tx);
				if (icl < mindi)
				  {
					mindi = icl;	ctcl = ctq;
				  }
			   }
			if (NULL != ctcl) /* success */
			  {
				mi2 = 10000;	ct2 = NULL;
				for (ctq = ctsr; ctq <= cter; ctq++)
				   {
					icl = abs(ctq->ux - tx);
					if ((ctq != ctcl) && (icl < mi2))
					  {
						mi2 = icl; ct2 = ctq;
					  }
				   }
				if ((mi2 < 1000) && (NULL != ct2)) /* success*/
				  ret += ptinterp(ECROWINTERP, tx,
					ctcl, ct2,cap);
				else	fprintf(stderr,
					 "failed to interpolate\n");
			  }
			else	fprintf(stderr, "failed to interpolate\n");
		  }
		else printf(" [%6.1f,%6.1f] found at [%3d,%3d]\n",
		 lct->cxcen, lct->cycen, lct->ux, lct->uy);
	   }
	return ret;
}

static int mknewpt(CALIBRUN *cap, ECAWORK *ecw)
{ /*	This creates a new point at (x,y) with indices (ux, uy) */
	int	*icp;
	CTROID	*cte;

	cte = cap->car_ctend;	icp = &(ecw->eca_iargs[0]);
	cte->cxcen = *icp++;	cte->cycen = *icp++;
	cte->ux = *icp++;	cte->uy = *icp;
	(cap->car_ctend)++;	return 1;
}

static void ecoperate(char *str, CALIBRUN *cap, ECAWORK *ecw)
{ /*	This operates upon a single `editcal' command.  It checks it for
	validity and passes off the appropriate action to `proc_command' */
	char	**dcp;
	int	action, ic, ir, nar, *icp;

	nar = ecw->eca_ncargs;	icp = &(ecw->eca_iargs[0]);
	dcp = ecw->eca_cargs;
	printf(" str = (%s); ncargs = %d\n", str, nar);
	if (*str == 'c')
	  {
		if (nar < 5)
		  {
			action = INSUFFARGS; ecw->eca_iargs[0] = 5;
			*(icp + 1) = nar;
		  }
		else if (nar < 6)
		  {
			action = ECCOLUMN; *(icp + 5) = find_col(cap, ecw);	
		  }
		else
		  {
			action = ECCOLUMNS; *(icp + 7) = find_cols(cap, ecw);	
		  }
	  }
	else if (*str == 'r')
	  {
		if (nar < 5)
		  {
			action = INSUFFARGS;	*icp = 6;
			*(icp + 1) = ecw->eca_ncargs;
		  }
		else if (nar < 6)
		  {
			action = ECROW; *(icp + 5) = find_row(cap, ecw);
		  }
		else
		  {
			action = ECCOLUMNS; *(icp + 7) = find_rows(cap, ecw);	
		  }
	  }
	else if (*str == 'w')
	  {
		action = ECWRITE;	*(icp + 1) = cap->car_ntroids;
		*icp = ecw->eca_nind;	(void)writecalce(cap);
	  }
	else if (*str == 'q') action = ECQUIT;
	else if (*str == 'e')
	  {
		action = ECEXIT;	*(icp + 1) = cap->car_ntroids;
		*icp = ecw->eca_nind;	(void)writecalce(cap);
	  }
	else if (!strncmp(str, "intc", 4))
	  {
		for (ic = 0; ic < nar; ic++)
			*(icp + ic) = atoi(*(dcp + ic));
		if (nar < 1)
		  {
			*icp = 1;	*(icp + 1) = nar; action = INSUFFARGS;
		  }
		else if (nar == 1)
		  {
			action = ECCOLINTERP;
			printf("interpolating in column %d\n", *icp);
			*(icp + 1) = interpcol(*icp, cap);
		  }
		else {
			action = ECCOLSINTERP;	*(icp + 2) = 0;
			printf("interpolating in cols %d to %d\n",
				*icp, *(icp + 1));
			for (ic = *icp; ic <= *(icp + 1); ic++)
				*(icp + 2) += interpcol(ic, cap);
		     }
	  }
	else if (!strncmp(str, "intr", 4))
	  {
		for (ic = 0; ic < nar; ic++) *(icp + ic) = atoi(*(dcp + ic));
		if (nar < 1)
		  {
			*icp = 1;	*(icp + 1) = nar; action = INSUFFARGS;
		  }
		else if (nar == 1)
		  {
			action = ECROWINTERP;
			printf("interpolating in row %d\n", *icp);
			*(icp + 1) = interprow(*icp, cap);
		  }
		else {
			action = ECROWSINTERP;	*(icp + 2) = 0;
			printf("interpolating in rows %d to %d\n",
				*icp, *(icp + 1));
			for (ir = *icp; ir < *(icp + 1); ir++)
				*(icp + 2) += interprow(ir, cap);
		     }
	  }
	else if (*str == 'n')
	  {
		if (nar < 4)
		  {
			*icp = 4;	*(icp + 1) = nar; action = INSUFFARGS;
		  }
		else {
			action = ECNEWPOINT;	*(icp + 4) = mknewpt(cap, ecw);
		     }
	  }
	else action = UNKNOWNVERB;
	log_ecchange(action, cap->car_fpl, ecw);
	log_ecchange(action, stdout, ecw);
}

/*ARGSUSED*/
int ecalmain(int argc, char *argv[])
{
	int		i, res;
	char		*comd, *cp, **dcp;
	CALIBRUN	cas;
	ECAWORK		ecws;
	static char	*cwork;
	static char	*fnames[] = {"calcentroids", "xlog", NULL};
	static char	ftypes[] = {3, 2, 0};

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	/* open logfile and put a summary of the files there and on stdout */
	if (NULL == (cas.car_fpl = propen("xlog"))) cas.car_fpl = stdout;
	ftypes[1] = (cas.car_fpl == stdout ? 0 : 2);
	header(stdout, PROCESSING, fnames, ftypes);
	if (cas.car_fpl != stdout)
		header(cas.car_fpl, PROCESSING, fnames, ftypes);
	cas.car_flags = CF_ASCIICEN;
	if (NULL == (AVOID *)(cas.car_ctr0 = (CTROID *)callop(MAXSPOTS,
	 sizeof (CTROID), "ctr0 in ecalmain")))
	  {
		if (stdout != cas.car_fpl) (void)fclose(cas.car_fpl);
		return EALLOC;
	  }
	cas.car_ctend = cas.car_ctr0 + MAXSPOTS;
	if (EOK != (res = ext_troids(&(ecws.eca_nind), &cas)))
	  {
		fprintf(stderr, "Error opening CALCENTROIDS file\n");
		return res;
	  }
	printf(
 " Legal commands are C[olumn[s]], R[ow[s]], Q[uit], W[rite], E[xit]\n");
	printf(" Column[s] and Row[s] can take either 5 or 6 arguments\n");
	printf(" namely Col colno Xmin Ymin Xmax Ymax\n");
	printf("     or Cols colmin Xmin Ymin Xmax Ymax intervalx\n");
	printf(" e.g. Col 13 126 5 128 510\n");
	printf("  or  Cols 13 126 5 448 510 26\n");
	printf(" number of fully-indexed spots to begin with = %5d\n",
	 ecws.eca_nind);
	if (NULL == (cwork = (char *)mallop(MAX_ARGS * CHARS_PER_ARG,
	 "cwork in ecalmain"))) return EALLOC;
	if (NULL == (ecws.eca_cargs = (char **)callop(MAX_ARGS,
	 sizeof (char *), "ecws.eca_cargs in ecalmain"))) return EALLOC;
	for (dcp = ecws.eca_cargs, cp = cwork, i = 0; i < MAX_ARGS;
	 i++, dcp++, cp += CHARS_PER_ARG) *dcp = cp;
	while (*(comd = get_eccom(&ecws)) != 'q')
	     {
		if (*comd == 's') break;	ecoperate(comd, &cas, &ecws);
		if (*comd == 'e') break;
	     }
	if ((NULL != cas.car_fpl) && (stdout != cas.car_fpl))
		(void)fclose(cas.car_fpl);
	memzap((AVOID **)&(ecws.eca_cargs), "ecws.eca_cargs in ecalmain");
	memzap((AVOID **)&cwork, "cwork in ecalmain");
	globalkill();	cas.car_fpl = NULL;	return (res ? EWRITERR : EOK);
}

/* end of editcal functionality */

/*****************************
	"missind": These routines enable you to determine which
	CALCENTROIDS points are either missing or duplicated.
*****************************/

/*ARGSUSED*/
int missimain(int argc, char *argv[])
{ /* Mainline for checking to see which indices are bad in CALCENTROIDS file */
	int		lux, luy, maxux, maxuy, minux, minuy, nind, res;
	static char	*fnames[] = {"calcentroids", "xlog", NULL};
	static char	ftypes[] = {3, 2, 0};
	CTROID		*lct, *ctprev;
	CALIBRUN	cas;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	/* open logfile and put a summary of the files there and on stdout */
	if (NULL == (cas.car_fpl = propen("xlog"))) cas.car_fpl = stdout;
	ftypes[1] = (cas.car_fpl == stdout ? 0 : 2);
	header(stdout, PROCESSING, fnames, ftypes);
	if (cas.car_fpl != stdout)
		header(cas.car_fpl, PROCESSING, fnames, ftypes);
	cas.car_flags = CF_ASCIICEN;
	if (NULL == (AVOID *)(cas.car_ctr0 = (CTROID *)callop(MAXSPOTS,
	 sizeof (CTROID), "ctr0 in missimain")))
	  {
		if (stdout != cas.car_fpl) (void)fclose(cas.car_fpl);
		return EALLOC;
	  }
	cas.car_ctend = cas.car_ctr0 + MAXSPOTS;
	if (EOK != (res = ext_troids(&nind, &cas)))
	  {
		fprintf(stderr, "Error opening CALCENTROIDS file\n");
		return res;
	  }
	minuy = minux = 10000;	maxuy = maxux = -minuy;
	for (lct = cas.car_ctr0; lct < cas.car_ctend; lct++)
	   {
		if (lct->ux < minux) minux = lct->ux;
		if (lct->uy < minuy) minuy = lct->uy;
		if (lct->ux > maxux) maxux = lct->ux;
		if (lct->uy > maxuy) maxuy = lct->uy;
	   }
	printf("X indices range from %3d to %3d\n", minux, maxux);
	printf("Y indices range from %3d to %3d\n", minuy, maxuy);
	lux = cas.car_ctend - cas.car_ctr0;
	if (lux > 1) qsort((void *)(cas.car_ctr0), (size_t)lux,
	 sizeof (CTROID), (QPE)luycmp);
	lct = cas.car_ctr = cas.car_ctr0;	ctprev = NULL;
	for (luy = minuy; luy <= maxuy; luy++)
	   {
		for (lux = minux; lux <= maxux; lux++)
		   {
			if ((lct->ux > lux) || (lct->uy > luy))
			  {
				printf(" No point with indices [%3d,%3d]\n",
					lux, luy);
			  }
			else if ((lct->ux == lux) && (lct->uy == luy))
			 {
				ctprev = lct; lct++;
			 }
			if (((ctprev->ux) == (lct->ux)) &&
				((ctprev->uy) == (lct->uy)))
			  {
				printf(
				 " Point [%3d,%3d] appears more than once\n",
					lct->ux, lct->uy);
				printf(
				 "\tcoordinates [%6.1f,%6.1f],[%6.1f,%6.1f]\n",
					lct->cxcen, lct->cycen,
					ctprev->cxcen, ctprev->cycen);
				lux--;	/* don't want to advance lux here */
			  }
		   }
	   }
	globalkill();	cas.car_fpl = NULL;	return (res ? EWRITERR : EOK);
}

/* end of missind functionalities */

/*****************************
	uca_2c: This reads a USPLINE and creates a CALCENTROIDS from it.
*****************************/

int uca2cmain(int argc, char *argv[])
{ /* Create a CALCENTROIDS output from an existing VSPLINE structure */
	int	ret;
	VSPLINE	*sgsp;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	(void)flagproc(argc, argv, (char *)NULL);
	if (NULL == (sgsp = gspline(5, &ret))) return ret;
	if (EOK != (ret = outindcen(sgsp))) return ret;
	globalkill();	return EOK;
}

/* end of uca_2c functionality */

/*****************************
	splitcen: Split a CALCENTROIDS file into four quadrants--
	one for each region of a four-element detector system, as follows:
	file#	X range		Yrange
	0      0 <= x < W/2	0 <= y < H/2
	1	W/2 <= x < W	0 <= y < H/2 
	2      0 <= x < W/2	H/2 <= y < H
	3	W/2 <= x < W	H/2 <= y < H
 Additional functionality May 2007:
	If "-t" is invoked, then only these two files are output:
	0      0 <= x < W/2	0 <= y < H
	1	W/2 <= x < W	0 <= y < H
*****************************/

static int splitcerr(FILE *fp, int etype, char *progn, char *str)
{ /* Error-message handler for splitcen. It returns an error flag
	appropriate to the error type: sometimes it's different from etype */
	switch (etype)
	      {
		case	EREADERR:	fprintf(fp,
		 "%s: Cannot open %s as CALCENTROIDS input\n", progn, str);
			return EOPENERR;
		case	EWRITERR:	fprintf(fp,
		 "%s: Cannot open %s as CALCENTROIDS output\n", progn, str);
			return EOPENERR;
		case	EBADVALUE:	fprintf(fp,
		 "%s: no translation available for `CALCENTROIDS'\n", progn);
			return EBADVALUE;
		case	-2:		fprintf(fp,
		 "%s: No values of X >= 0 present in CALCENTROIDS file\n",
		 progn);
			return EBADVALUE;
		case	-3:		fprintf(fp,
		 "%s: No values of Y >= 0 present in CALCENTROIDS file\n",
		 progn);
			return EBADVALUE;
		case	EOK:	return EOK;
		default:		fprintf(fp,
		 "%s: Unknown error %d\n", progn, etype);
			return EBADVALUE;
	    }
	/*NOTREACHED*/
	return EBADVALUE;
}

int splitcmain(int argc, char *argv[])
{ /* The mainline for splitting up a CALCENTROIDS into four subfiles */
	char		*calin, **flags, **dcp;
	int		inomx, inomy, justx;
	int		nin, nitems, nou0, nou1, nou2, nou3, nval;
	double		tx, ty, xma, xma2, yma, yma2;
	FILE		*fpcc, *fplo, *fpo0, *fpo1, *fpo2, *fpo3;
	static char	inl[120], inn[320];
	static char	fn0[320], fn1[320], fn2[320], fn3[320];

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	flags = flagproc(argc, argv, "TXY");
	justx = 0;	inomx = inomy = -1;
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	   {
		nval = atoi((*dcp) + 1);
		switch (**dcp)
		      {
			case	'T':	justx = 1;	break;
			case	'X':	if ((0 < nval) && (nval < 100000))
				inomx = nval;	break;
			case	'Y':	if ((0 < nval) && (nval < 100000))
				inomy = nval;	break;
		     }
	   }
	calin = trname("CALCENTROIDS");
	if ((NULL == calin) || ('\0' == *calin))
		return splitcerr(stderr, EBADVALUE, argv[0], NULL);
	strcpy(inn, calin);
	if (NULL == (fpcc = xgopen(inn, "r")))
		return splitcerr(stderr, EREADERR, argv[0], inn);
	xma = -1.e9;	yma = -1.e9;	nin = 0;
	while (NULL != fgets(inl, 60, fpcc))
	     {
		nitems = sscanf(inl, "%lf %lf", &tx, &ty);	nin++;
		if (tx > xma) xma = tx;	if (ty > yma) yma = ty;
	     }
	if (xma < 0.) return splitcerr(stderr, -2, argv[0], NULL);
	if (yma < 0.) return splitcerr(stderr, -3, argv[0], NULL);
	if (inomx < 0)
	  {
		if ((1920. < xma) && (xma < 2060.)) xma = 2048.;
		if ((960. < xma) && (xma < 1030.)) xma = 1024.;
		if ((480. < xma) && (xma < 515.)) xma = 512.;
	  }
	else	xma = inomx;
	if (inomy < 0)
	  {
		if ((1920. < yma) && (yma < 2060.)) yma = 2048.;
		if ((960. < yma) && (yma < 1030.)) yma = 1024.;
		if ((480. < yma) && (yma < 515.)) yma = 512.;
	  }
	else	yma = inomy;
	printf("nominal detector dimensions: %5d by %5d\n",
		(int)xma, (int)yma);
	xma2 = 0.5 * xma;
	yma2 = (justx) ? yma : 0.5 * yma;
	(void)fseek(fpcc, 0L, 0);	calin = filepart(inn, 2);
	strcpy(fn0, calin);	strcat(fn0, "0.cci");
	if (NULL == (fpo0 = xgopen(fn0, "w")))
		return splitcerr(stderr, EWRITERR, argv[0], fn0);
	strcpy(fn1, calin);	strcat(fn1, "1.cci");
	if (NULL == (fpo1 = xgopen(fn1, "w")))
		return splitcerr(stderr, EWRITERR, argv[0], fn1);
	if (justx) fpo2 = fpo3 = NULL;
	else {
		strcpy(fn2, calin);	strcat(fn2, "2.cci");
		if (NULL == (fpo2 = xgopen(fn2, "w")))
			return splitcerr(stderr, EWRITERR, argv[0], fn2);
		strcpy(fn3, calin);	strcat(fn3, "3.cci");
		if (NULL == (fpo3 = xgopen(fn3, "w")))
			return splitcerr(stderr, EWRITERR, argv[0], fn3);
	    }
	nin = nou0 = nou1 = nou2 = nou3 = 0;
	while (NULL != fgets(inl, 118, fpcc))
	     {
		nitems = sscanf(inl, "%lf %lf", &tx, &ty);
		if (nitems >= 2)
		  {
			nin++;
			if (tx < xma2)
			  {
				if ((justx) || (ty < yma2))
				  {
					nou0++;	fplo = fpo0;
				  }
				else {
					nou1++;	fplo = fpo1; ty -= yma2;
				     }
			  }
			else {
				tx -= xma2;
				if (justx)
				  {
					nou1++;	fplo = fpo1;
				  }
				else if (ty < yma2)
				  {
					nou2++;	fplo = fpo2;
				  }
				else {
					nou3++;	fplo = fpo3; ty -= yma2;
				     }
			     }
			fprintf(fplo, " %7.2f %7.2f\n", tx, ty);
		  }
		else fprintf(stderr, "nitems = %d on line %d\n",
			nitems, nin);
	     }
	if (justx) fprintf(stdout,
	 "%6d calcentroids read in: %5d to left half, %5d to right half\n",
		nin, nou0, nou1);
	else {
		fprintf(stdout,
 " %6d calcentroids read in; %4d to upper-left, %4d to lower-left,",
			nin, nou0, nou1);
		fprintf(stdout,
 " %4d to upper-right, %4d to lower-right\n", nou2, nou3);
	     }
	(void)fclose(fpcc);
	(void)fclose(fpo0);	(void)fclose(fpo1);
	if (!justx)
	  {
		(void)fclose(fpo2);	(void)fclose(fpo3);
	  }
	return EOK;
}

/* end of splitcen functionality */

/* end of calcetool.c */
