/* calfill.c -- Routines for filling in missing data in fiducial image

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	20 Nov 2005: inline function declarations & function prototypes
	28 Oct 2005: replace static globals with CALIBRUN structure elements
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	28 Aug 2004: routines used only here are declared as static; cosmetics
	26 Jan 2004: a few static globals converted to local variables
	21 Jan 2004: license message inserted and initial comment reordered
	 6 Jan 2004: CALIBRUN structure elements in function calls
		replace extern globals
	23 Nov 2003: replace global variables with CALIBRUN structure elements
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	 5 Jan 2003: reorganized source-file header
	20 Sep 2002: CF_* flags moved to ../include/xgen.h
	25 Mar 2002: removed MSI attachments
	31 May 1995: use %d format in tl0-tl1 calculations
	 2 Mar 1995: diagnostics to stdout
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: major rewrite of `calspots'; bugfixes elsewhere
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version

   Purpose: filling in the missing points in the calibration spot list
	after the initial centroids have been found and indexed but
	before the mapping inversions are invoked.
   Cross references: calibmain.c, calindex.c, calspots.c, ../xglib/calinv.c

	These routines operate by finding least-squares parabolas containing
	the indexed points, using those parabolas to extrapolate to missing
	points AND to index some leftover observed points, and leaving the
	extrapolations in the remaining points.
*************************************************************************/
#define		NEED_TIME	1
#include	"calibdefs.h"
#define		ROW		2
#define		COLUMN		4
#define		OKRE		(0.1e-3)
#define		RADIAL_MULTIPLIER	3.5	/* see its use */
#define		ISLOP		4

static void solve_parab(int indx, char direction, int doprint, CALIBRUN *cap)
{ /*	this determines the equation of the parabola going through the row
	or column indexed with `indx'.  It computes both the best line and
	the best arc and sets up the one with the smaller residual. */
	int		j, iy, ix, ninrow, jmax, nobsd;
	double		discrim, e0, e1, e2, e3, e4, f0, f1, f2, parerr;
	double		tx, tx2, ty, ty2, wt;
	double		ata[3][3], atai[3][3], atb[3], result[3];
	CALPOINT	*calp;
	PARABOLA	*thispara;
	VSPLINE		*lsp;

	e0 = e1 = e2 = e3 = e4 = f0 = f1 = f2 = ninrow = iy = ix = 0;
	if (direction == ROW)
	  {
		iy = indx;	jmax = cap->car_maxcols;
		thispara = cap->car_rowpara + indx;
	  }
	else {
		ix = indx;	jmax = cap->car_maxrows;
		thispara = cap->car_colpara + indx;
	     }
	thispara->ar_known = 0;
	lsp = cap->car_spl;
	for (nobsd = ninrow = j = 0; j < jmax; j++)
	   {
		if (direction == ROW)
			calp = lsp->v_cpt + iy * cap->car_maxcols + j;
		else	calp = lsp->v_cpt +  j * cap->car_maxcols + ix;
		if (calp->c_set)
		  {
			ninrow++;
			if (calp->c_ift)
			  {
				nobsd++;	wt = 2.5;
			  }
			else wt = 0.4;
			tx = calp->c_ptocx / cap->car_scalxy - 1.;
			tx2 = tx * tx;
			ty = calp->c_ptocy / cap->car_scalxy - 1.;
			ty2 = ty * ty;
			e0 += wt;	f1 += tx * ty * wt;
			if (direction == ROW)
			  {
				e1 += tx * wt;		f0 += ty * wt;
				e2 += tx2 * wt;		f2 += tx2 * ty * wt;
				e3 += tx2 * tx * wt;	e4 += tx2 * tx2 * wt;
			  }
			else {
				e1 += ty * wt;		f0 += tx * wt;
				e2 += ty2 * wt;		f2 += ty2 * tx * wt;
				e3 += ty2 * ty * wt;	e4 += ty2 * ty2 * wt;
			     }
		  }
	     }
	discrim = e0 * e2 - e1 * e1;
	if ((e0 < DFUZZ) || (ninrow < 1)) return; /*give up: no points at all*/
	else if ((ninrow  < 3) || fabs(discrim) < DFUZZ) /*1 or 2 pts or flat*/
	  {
		thispara->ar_pe2 = thispara->ar_pe1 = 0.; /* zero slope */
		thispara->ar_pe0 = f0 / e0;
	  }
	else {
		ata[0][0] = e0;	ata[0][1] = ata[1][0] = e1;
		ata[0][2] = ata[1][1] = ata[2][0] = e2;
		ata[1][2] = ata[2][1] = e3;	ata[2][2] = e4;
		if ((ninrow < 3) || dinv3(ata, atai)) /* 2 pts or singular */
		  {
			thispara->ar_pe2 = 0.;	/* must be a line now */
			thispara->ar_pe1 = (e0 * f1 - e1 * f0) / discrim;
			thispara->ar_pe0 = (e2 * f0 - e1 * f1) / discrim;
		  }
		else {
			atb[0] = f0;	atb[1] = f1;	atb[2] = f2;
			dmm31(result, atai, atb);
			thispara->ar_pe2 = result[2];
			thispara->ar_pe1 = result[1];
			thispara->ar_pe0 = result[0];
		     }
	      }
/*now calculate the deviation of the parabola from the actual values.*/
	for (parerr = j = 0; j < jmax; j++)
	   {
		if (direction == ROW)
		  {
			calp = lsp->v_cpt + iy * cap->car_maxcols + j;
			e1 = calp->c_ptocx;	e2 = calp->c_ptocy;
		  }
		else {
			calp = lsp->v_cpt + j * cap->car_maxcols + ix;
			e1 = calp->c_ptocy;	e2 = calp->c_ptocx;
		     }
		if (calp->c_set)
		  {
			wt = (calp->c_ift ? 2.5 : 0.4);
			e1 = e1 / cap->car_scalxy - 1.;
			e2 = e2 / cap->car_scalxy - 1.;
			e3 = thispara->ar_pe2 * e1 * e1 +
				thispara->ar_pe1 * e1 + thispara->ar_pe0;
			parerr += (e2 - e3) * (e2 - e3) * wt;
		  }
	   }
	parerr /= (e0 * (double)ninrow);
	thispara->ar_known = 1;		thispara->ar_npts = ninrow;
	thispara->ar_nobs = nobsd;	thispara->ar_resid = parerr;
	if (doprint) fprintf(cap->car_fpl,
	 " %6d %6d %6d   %7.2lf %7.3lf %7.3lf %10.4le\n",
	 indx, thispara->ar_npts, thispara->ar_nobs, thispara->ar_pe0,
	 thispara->ar_pe1, thispara->ar_pe2, sqrt(thispara->ar_resid));
}

static void get_parabs(int doprint, CALIBRUN *cap)
{ /*	This computes the equations of the arcs associated with each row
	and column in the list.  In order for an arc equation to be
	computed, we must have at least 4 points defined on the row
	or column.   The argument `doprint' is 1 when we are going
	to summarize the results to a logfile; it's zero otherwise */
	static int	iy, ix;

	if (doprint)
	  {
		fprintf(cap->car_fpl,
		 "Equations of the parabolas through the rows:\n");
		fprintf(cap->car_fpl,
 "    Row   # of # obsd        x-   coeff   coeff  RMSErr\n");
		fprintf(cap->car_fpl,
 " number points points intercept     (x)   (x*x)\n");
	  }
	for (iy = cap->car_maxrows; iy--; ) solve_parab(iy, ROW, doprint, cap);
	if (doprint)
	  {
		fprintf(cap->car_fpl,
		 "Equations of the parabolas through the columns:\n");
		fprintf(cap->car_fpl,
 " Column   # of # obsd        y-   coeff   coeff  RMSErr\n");
		fprintf(cap->car_fpl,
 " number points points intercept     (y)   (y*y)\n");
	  }
	for (ix = cap->car_maxcols; ix--; )
		solve_parab(ix, COLUMN, doprint, cap);
}

static double extrap_indx(int iy, int ix, int which, int direction,
	CALIBRUN *cap)
{ /*	This returns either the x or the y value at the grid point iy,ix
	based a least squares fit to the INDICES along the appropriate
	row or column.  If `which' is 0, we look at x values; if `which'
	is 1, we look at y values.  If "direction" == ROW, then the
	four nearest points on the row are used to compute the leastsquares
	fit to the equation (x or y) = ai*i + b*i + c, where i is the index.
	Then the x or y value is returned.  If "direction"==COLUMN, the
	four nearest points on the column are used and x or y is returned.
	If NO points are defined on the row or column in question,
	a "flag" value of -200 is returned. */
	int		j, js, je, jmid, jincr, jj, npts;
	double		di, di2, discrim, e0, e1, e2, e3, e4, f0, f1, f2;
	double		thisv, wt, ata[3][3], atai[3][3];
	static CALPOINT	*calp;
	VSPLINE		*lsp;

	e0 = e1 = e2 = e3 = e4 = f0 = f1 = f2 = 0.;
	if (direction == ROW)
	  {
		jj = ix; jmid = cap->car_cencol; je = cap->car_maxcols;
	  }
	else {
		jj = iy; jmid = cap->car_cenrow; je = cap->car_maxrows;
	     }
	if (jj < jmid)
	  {
		js = jj + 1;	jincr = 1;
	  }
	else {
		js = jj - 1;	jincr = -1;	je = -1;
	     }
	lsp = cap->car_spl;
	for (j = js, npts = 0; j != je; j += jincr)
	   {
		calp = (direction == ROW) ?
		 lsp->v_cpt + iy * cap->car_maxcols + j :
		 lsp->v_cpt +  j * cap->car_maxcols+ ix;
		if (calp->c_set)
		  {
			thisv = (which ? calp->c_ptocy : calp->c_ptocx);
			wt = (calp->c_ift ? 2.5 : 0.4);	npts++;
			di = j - jj;	di2 = di * di;
			thisv = thisv / cap->car_scalxy - 1.;
			e0 += wt / di2;		e1 += wt / di;
			e2 += wt;		e3 += wt * di;
			e4 += wt * di2;		f0 += wt * thisv / di2;
			f1 += wt * thisv / di;	f2 += wt * thisv;
			if (npts > 6) break;	/* don't need more */
		  }
	   }
	ata[0][0] = e0; ata[0][1] = ata[1][0] = e1;
	ata[0][2] = ata[1][1] = ata[2][0] = e2;
	ata[1][2] = ata[2][1] = e3; ata[2][2] = e4;
	if (npts < 2) return -200.;
	if ((npts < 3) || dinv3(ata, atai)) /* 2 pts or lsq failed */
	  {
		if (fabs(discrim = e2 * e0 - e1 * e1) < DFUZZ) return -200.;
		return (e2 * f0 - e1 * f1) / discrim;
	  }
	return atai[0][0] * f0 + atai[0][1] * f1 + atai[0][2] * f2;
}

static int parab_fit(int iy, int ix, int fitflag, CALIBRUN *cap)
{ /*	This computes the x, y values at the index point iy, ix.
	For points at the intersection of a row arc and a column arc,
	we estimate c_ptocx, c_ptocy as the simultaneous solution to
	the equations of the row arc and the column arc.  For points on a
	row arc for which we don't have a column equation, get the x value
	by a local linear approximation to the x values on the row and get
	the y value from the equation of the arc.  Points on a column arc
	are handled similarly. */

	register int	jxf, jxl, jyf, jyl;
	int		insane, nrdsq, pass, pol;
	char		dubious, flag;
	double		dex, dey, deriv, e0, e1, e2, e3, e4, f0, f1, f2;
	double		prevres, rdsq, resid, strapx, strapy;
	double		thisx, thisy, trd, ty0, ty1;
	CTROID		*lct;
	CALPOINT	*calp, *co;
	PARABOLA	*rop, *cop, *tp;
	VSPLINE		*lsp;
	static double	checkx, checky;

	lsp = cap->car_spl;	dubious = 0;
	rop = cap->car_rowpara + iy;	cop = cap->car_colpara + ix;
	flag = 2 * rop->ar_known + cop->ar_known;
	if (-190. > (strapx = extrap_indx(iy, ix, 0, COLUMN, cap)))
	  {
		if (-190. > (strapx = extrap_indx(iy, ix, 0, ROW, cap)))
		  {
			dubious |= 1;
			for (tp = cap->car_colpara, jxf = 0; jxf <
			 cap->car_maxcols; jxf++, tp++) if (tp->ar_known) break;
			jxl = jxf;
			if (jxf < cap->car_maxcols - 1)
			  {
				jxl = cap->car_maxcols - 1;
				for (tp = cap->car_colpara + jxl;
				 jxl > jxf; jxl--) if (tp->ar_known) break;
			  }
			if (jxf < jxl) strapx = -1. +
			 ((double)(2 * (ix - jxf))) / ((double)(jxl - jxf));
			else strapx = -1. + ((double)ix) /
				(double)cap->car_cencol;
		  }
	  }
	if (-190. > (strapy = extrap_indx(iy, ix, 1, ROW, cap)))
	  {
		if (-190. > (strapy = extrap_indx(iy, ix, 1, COLUMN, cap)))
		  {
			dubious |= 2;
			for (tp = cap->car_rowpara, jyf = 0;
			 jyf < cap->car_maxrows; jyf++, tp++)
				if (tp->ar_known) break;
			jyl = jyf;
			if (jyf < cap->car_maxrows - 1)
			  {
				jyl = cap->car_maxrows - 1;
				for (tp = cap->car_rowpara + jyl;
				 jyl > jyf; jyl--) if (tp->ar_known) break;
			  }
			if (jyf < jyl) strapx = -1. +
			 ((double)(2 * (iy - jyf))) / ((double)(jyl - jyf));
			else strapx = -1. + ((double)iy) /
				(double)cap->car_cenrow;
		  }
	  }
	thisx = strapx;	thisy = strapy;
	switch (flag)
	      {
		case	0:	/* neither eqn. known: need trusty x, y */
			if (dubious) return 0;	else break;
		case	1:	/* column eqn known: need trusty y */
			if (dubious & 02) return 0;
			thisx = cop->ar_pe2 * thisy * thisy +
				cop->ar_pe1 * thisy + cop->ar_pe0;	break;
		case	2:	/* row eqn is known: need trusty x */
			if (dubious & 01) return 0;
			thisy = rop->ar_pe2 * thisx * thisx +
				rop->ar_pe1 * thisx + rop->ar_pe0;	break;
		case	3:	/* both eqns known.  To solve them
			simultaneously involves a quartic equation.  Rather
			than solving it, I'll start at the extrapolated
			position and use Newton's method to solve it */
			if (dubious & 02) ty0 = rop->ar_pe2 * strapx *strapx
			 + rop->ar_pe1 * strapx + rop->ar_pe0;
			else ty0 = strapy;	ty1 = strapy * strapy;
			e0 = rop->ar_pe2 * cop->ar_pe0 * cop->ar_pe0 +
				rop->ar_pe1 * cop->ar_pe0 + rop->ar_pe0;
			e1 = 2. * rop->ar_pe2 * cop->ar_pe1 * cop->ar_pe0 +
				rop->ar_pe1 * cop->ar_pe1 - 1.;
			e2 = rop->ar_pe2 * cop->ar_pe1 * cop->ar_pe1 +
				2. * rop->ar_pe2 * cop->ar_pe2 * cop->ar_pe0 +
				rop->ar_pe1 * cop->ar_pe2;
			e3 = 2. * rop->ar_pe2 * cop->ar_pe2 * cop->ar_pe1;
			e4 = rop->ar_pe2 * cop->ar_pe2 * cop->ar_pe2;
			f0 = 2. * e2;	f1 = 3. * e3;	f2 = 4. * e4;
			resid = e4 * ty1 * ty1 + e3 * ty1 * ty0 +
				e2 * ty1 + e1 * ty0 + e0;
			prevres = fabs(resid);
			for (pass = 0; (pass < 9) && (prevres > OKRE); pass++)
			   {
				deriv = f2 * ty1 * ty0 + f1 * ty1 +
					f0 * ty0 + e1;
				if (fabs(deriv) < DFUZZ) break;
				deriv = resid / deriv;
				ty0 -= deriv;	ty1 = ty0 * ty0;
				resid = e4 * ty1 * ty1 + e3 * ty1 * ty0 +
					e2 * ty1 + e1 * ty0 + e0;
				if (prevres < fabs(resid))
				  {
					ty0 += 0.7 * deriv; ty1 = ty0 * ty0;
					resid = e4 * ty1 *ty1 + e3 *ty1 *ty0
					 + e2 * ty1 + e1 * ty0 + e0;
					if (prevres *1.2 < fabs(resid)) break;
				  }
				prevres = fabs(resid);
			   }
			if (prevres > OKRE)
			  {
				if (dubious) return 0;
				thisx = strapx;	thisy = strapy;
			  }
			else {
				thisy = ty0;
				thisx = cop->ar_pe2 * ty0 * ty0 +
				 cop->ar_pe1 * ty0 + cop->ar_pe0;
			     }
			break;
	     }
	thisx = (thisx + 1.) * cap->car_scalxy;
	thisy = (thisy + 1.) * cap->car_scalxy;
	if (fitflag & 01)
	  {
		checkx = thisx;	checky = thisy;
		return (thisx > -40 && thisx < (lsp->v_detw + 40) &&
			thisy > -40 && thisy < (lsp->v_deth + 40));
	  }
	/* sanity checks on the value! */
	if (iy < cap->car_cenrow)
	  {
		calp = lsp->v_cpt + (iy + 1) * cap->car_maxcols + ix;
		if ((calp->c_set) && (thisy > calp->c_ptocy))
		  thisy = calp->c_ptocy - 0.6 * ((double)lsp->v_deth) /
						((double)cap->car_maxrows);
	  }
	else
	  {
		calp = lsp->v_cpt + (iy - 1) * cap->car_maxcols + ix;
		if ((calp->c_set) && (thisy < calp->c_ptocy))
		  thisy = calp->c_ptocy + 0.6 * ((double)lsp->v_deth) /
					((double)cap->car_maxrows);
	  }
	if (ix < cap->car_cencol)
	  {
		calp = lsp->v_cpt + iy * cap->car_maxcols + ix + 1;
		if ((calp->c_set) && (thisx > calp->c_ptocx))
		  thisx = calp->c_ptocx - 0.6 * ((double)lsp->v_detw) /
					((double)cap->car_maxcols);
	  }
	else
	  {
		calp = lsp->v_cpt + iy * cap->car_maxcols + ix - 1;
		if ((calp->c_set) && (thisx < calp->c_ptocx))
		  thisx = calp->c_ptocx +  0.6 * ((double)lsp->v_detw) /
					((double)cap->car_maxcols);
	  }
	insane = 0;
	if (thisy < -2000. || thisy > lsp->v_deth + 2000.) insane = 1;
	if (thisx < -2000. || thisx > lsp->v_detw + 2000.) insane = 1;
	if ((iy < cap->car_maxrows / 3) &&
	 (thisy > 0.7 * ((double)lsp->v_deth))) insane = 1;
	else if ((iy >= cap->car_maxrows / 3) && (iy < 2*cap->car_maxrows/3)
	 && (thisy < 0. || thisy > lsp->v_deth)) insane = 1;
	else if ((iy > 2 * cap->car_maxrows / 3) &&
	 (thisy < 0.3 * ((double)lsp->v_deth))) insane = 1;
	if ((ix < cap->car_maxcols / 3) &&
	 (thisx > 0.7 * ((double)lsp->v_detw))) insane = 1;
	else if ((ix >= cap->car_maxcols / 3) && (ix < 2*cap->car_maxcols/3) &&
	(thisx < 0. || thisx > lsp->v_detw)) insane = 1;
	else if ((ix > 2 * cap->car_maxcols / 3) &&
	 (thisx < 0.3 * ((double)lsp->v_detw))) insane = 1;
	if (insane) return 0;
	if ((!(fitflag & 02)) && (thisx < -40 || thisx > (lsp->v_detw + 40) ||
		thisy < -40 || thisy > lsp->v_deth + 40)) return 0;
	calp = lsp->v_cpt + iy * cap->car_maxcols + ix;
	calp->c_set = 1;
	calp->c_ptocx = thisx;	calp->c_ptocy = thisy;
	if (thisx < 0. || thisx > lsp->v_detw ||
	 thisy < 0. ||thisy > lsp->v_deth)
	 return 1; /* don't search for observed point if offscreen */
	/* don't look for an observed point if they're all indexed */
	if ((short)(cap->car_ntroids) <= lsp->v_ntrue) return 1;
	/* having defined the point, we try to match it to
	 an observed point.  Before doing so we must see what
	 the characteristic distance between points in this
	 neighborhood is */
	rdsq = nrdsq = 0;
	if (ix)	/* measure dist. to point one to its left */
	  {
		co = calp - 1;	trd = 0;
		if (co->c_set)
		  {
			dex = co->c_ptocx - calp->c_ptocx;
			dey = co->c_ptocy - calp->c_ptocy;
			trd = dex * dex + dey * dey;
		  }
		else if (parab_fit(iy, ix - 1, 1, cap))
		  {
			dex = checkx - calp->c_ptocx;
			dey = checky - calp->c_ptocy;
			trd = dex * dex + dey * dey;
		  }
		if (trd > DFUZZ)
		  {
			rdsq += trd;	nrdsq++;
		  }
	  }
	if (ix < cap->car_maxcols - 1) /* ... point one to its right */
	  {
		co = calp + 1;	trd = 0;
		if (co->c_set)
		  {
			dex = co->c_ptocx - calp->c_ptocx;
			dey = co->c_ptocy - calp->c_ptocy;
			trd = dex * dex + dey * dey;
		  }
		else if (parab_fit(iy, ix + 1, 1, cap))
		  {
			dex = checkx - calp->c_ptocx;
			dey = checky - calp->c_ptocy;
			trd = dex * dex + dey * dey;
		  }
		if (trd > DFUZZ)
		  {
			rdsq += trd;	nrdsq++;
		  }
	  }
	if (iy)	/* distance to point just above this one */
	  {
		co = calp - cap->car_maxcols;	trd = 0;
		if (co->c_set)
		  {
			dex = co->c_ptocx - calp->c_ptocx;
			dey = co->c_ptocy - calp->c_ptocy;
			trd = dex * dex + dey * dey;
		  }
		else if (parab_fit(iy - 1, ix, 1, cap))
		  {
			dex = checkx - calp->c_ptocx;
			dey = checky - calp->c_ptocy;
			trd = dex * dex + dey * dey;
		  }
		if (trd > DFUZZ)
		  {
			rdsq += trd;	nrdsq++;
		  }
	  }
	if (iy < cap->car_maxrows + 1) /* dist.to point just below */
	  {
		co = calp + cap->car_maxcols;	trd = 0;
		if (co->c_set)
		  {
			dex = co->c_ptocx - calp->c_ptocx;
			dey = co->c_ptocy - calp->c_ptocy;
			trd = dex * dex + dey * dey;
		  }
		else if (parab_fit(iy + 1, ix, 1, cap))
		  {
			dex = checkx - calp->c_ptocx;
			dey = checky - calp->c_ptocy;
			trd = dex * dex + dey * dey;
		  }
		if (trd > DFUZZ)
		  {
			rdsq += trd;	nrdsq++;
		  }

	  }
	if (nrdsq) rdsq /= (double)nrdsq;
	/* if we have no neighbors, we use the overall density of
	 points to estimate the spacing with a fivefold fudge factor*/
	else	rdsq = 0.2 * (((double)lsp->v_detw) /
		 (double)(cap->car_ntroids)) * (double)lsp->v_deth;
	/* having estimated the characteristic distance, we look for
	 an un-indexed centroid that lies close enough to use */
	for (lct = cap->car_unindex0; lct < cap->car_ctend; lct++)
	   {
		dex = lct->cxcen - calp->c_ptocx;
		dey = lct->cycen - calp->c_ptocy;
		if (dex * dex + dey * dey < 0.3 * rdsq) break;
	   }
	if (lct < cap->car_ctend)	/* we got one: use it */
	  {
		if (cap->car_flags & CF_HEXAGGRID)
		  {
			pol = (ix + iy) % 2;
			if (cap->car_flags & CF_POLARODD) pol = 1 -pol;
			if (lct->cinten < 0.) pol = 1 - pol;
		  }
		else pol = 0;
		if (!pol)
		  {
			calp->c_ptocx = lct->cxcen;
			calp->c_ptocy = lct->cycen;
			if (!calp->c_ift) (lsp->v_ntrue)++;
			calp->c_ift = 1; lct->ux = ix; lct->uy = iy;
		/* reorganize the centroids list so the just-indexed
		 point appears before the un-indexed ones. */
			if (lct != cap->car_unindex0) swap((AVOID *)lct,
			 (AVOID *)(cap->car_unindex0), sizeof (CTROID));
			(cap->car_unindex0)++;
		  }
	  }
	cap->car_ctr = lct; /* in case we need it outside */
	return 1;
}

static void disp_calib(int iy, int ix, char axis, FILE *fp, CALIBRUN *cap)
{ /*	This prints a six-character representation of calp->c_ptocx
	or calp->c_ptocy with underscores in front of the numbers if
	the value was an extrapolation.  */
	VSPLINE		*lsp;
	CALPOINT	*calp;
	double		val;

	lsp = cap->car_spl; calp = lsp->v_cpt + iy * cap->car_maxcols + ix;
	val = (axis == 'x' ? calp->c_ptocx : calp->c_ptocy);
	if (calp->c_ift) fprintf(fp, "%6.1f", val);
	else if (val <= -1000.) fprintf(fp, "<-999.");
	else if (val <= -100.) fprintf(fp,"%6.1f", val);
	else if (val <= -10.) fprintf(fp,"_%5.1f", val);
	else if (val < 0.) fprintf(fp,"__%4.1f", val);
	else if (val < 10.) fprintf(fp,"___%3.1f", val);
	else if (val < 99.) fprintf(fp,"__%4.1f", val);
	else fprintf(fp,"_%5.1f", val);
}

static void summ_blank(FILE *fp, CALIBRUN *cap, int limi[])
{  /*	This summarizes the blanking-out operation */
	fprintf(fp, "Rows %2d to %2d, Columns %2d to %2d are populated\n",
		limi[2], limi[3], limi[0], limi[1]);
	if (limi[2] > 2) fprintf(fp, "Blanking rows 0 through %2d\n",
		limi[2] - 2);
	else if (limi[2] == 2) fprintf(fp, "Blanking row 0\n");
	if (limi[3] < cap->car_maxrows - 3) fprintf(fp,
	 "blanking rows %2d through %2d\n",
		limi[3] + 2, (int)(cap->car_maxrows) - 1);
	else if (limi[3] == cap->car_maxrows - 3) fprintf(fp,
	 "blanking row %2d\n", (int)(cap->car_maxrows) - 1);
	if (limi[0] > 2) fprintf(fp, "Blanking cols 0 through %2d\n",
		limi[0] - 2);
	else if (limi[0] == 2) fprintf(fp, "Blanking col 0\n");
	if (limi[1] < cap->car_maxcols - 3) fprintf(fp,
	 "blanking cols %2d through %2d\n",
		limi[1] + 2, (int)(cap->car_maxcols) - 1);
	else if (limi[1] == cap->car_maxcols - 3) fprintf(fp,
		"blanking col %2d\n", (int)(cap->car_maxcols) - 1);
}

static void known_first(CALIBRUN *cap)
{ /*	This goes through the list of centroids and puts all those with
	 known indices at the top */
	CTROID		*lct;
	static CTROID	*nc;

	for (lct = cap->car_ctr0; lct < cap->car_ctend - 1; lct++)
	   {
		if ((lct->ux < 0) || (lct->uy < 0))
		  {
			for (nc = cap->car_ctend - 1; nc > lct; nc--)
			   if (nc->ux >= 0 && nc->uy >= 0)
			     {
				swap((AVOID *)lct,(AVOID *)nc,sizeof (CTROID));
				break;
			     }
			if (nc == lct)	/* no more known indices below:quit*/
				break;
		  }
	   }
	for ( ; lct < cap->car_ctend; lct++)
		if (lct->ux < 0 || lct->uy < 0) break;
	/* remember the first unindexed one */
	cap->car_ctr = cap->car_unindex0 = lct;
}

static void kill_wayout(CALIBRUN *cap, int limi[])
{ /*	This gets rid of entire rows and columns that are more than one
	row or column away from the nearest observed point.  That is, if the
	1st row containing an actual observed point is row 12 and the last one
	is row 43, then rows 0-10 and 45-cap->car_maxrows are blanked out. */
	static int	iy, ix;
	static CALPOINT	*calp, *car;
	static double	maxox, maxoy, minox, minoy;
	VSPLINE		*lsp;

	/* first identify the first & last row and column with observed spots*/
	lsp = cap->car_spl;
	minox = minoy = 10000.;	maxox = maxoy = -10000.;
	limi[0] = cap->car_maxcols;	limi[1] = 0;
	limi[2] = cap->car_maxrows;	limi[3] = 0;
	for (iy = 0, calp = lsp->v_cpt; iy < cap->car_maxrows; iy++)
	  for (ix = 0; ix < cap->car_maxcols; ix++, calp++)
	   if (calp->c_ift)
	     {
		if (ix < limi[0]) limi[0] = ix; if (ix > limi[1]) limi[1] = ix;
		if (iy < limi[2]) limi[2] = iy; if (iy > limi[3]) limi[3] = iy;
		if (calp->c_ptocx < minox) minox = calp->c_ptocx;
		if (calp->c_ptocx > maxox) maxox = calp->c_ptocx;
		if (calp->c_ptocy < minoy) minoy = calp->c_ptocy;
		if (calp->c_ptocy > maxoy) maxoy = calp->c_ptocy;
	     }
	if (cap->car_flags & CF_VERBOSE) summ_blank(stdout, cap, limi);
	summ_blank(cap->car_fpl, cap, limi);
	/* we allow one row and one column on either side of these if there
	 are more than five unindexed spots at this stage or if there's lots
	 of empty space around these */
	if (lsp->v_ntrue + 5 < (short)(cap->car_ntroids))
	  {
		if (limi[2] > 1) limi[2]--;
		if (limi[3] < (int)(cap->car_maxrows - 1)) (limi[3])++;
		if (limi[0] > 1) limi[0]--;
		if (limi[1] < (int)(cap->car_maxcols - 1)) (limi[1])++;
	  }
	else {
		if ((limi[3] < cap->car_maxrows - 1) &&
		 (maxox < 0.9 * (double)lsp->v_detw)) limi[3]++;
		if ((limi[2] > 1) && (minox > 0.1 * (double)lsp->v_detw))
			limi[2]--;
		if ((limi[1] < cap->car_maxcols - 1) &&
		 (maxox < 0.9 * (double)lsp->v_deth)) limi[1]++;
		if ((limi[0] > 1) &&
			(minox > 0.1 * (double)lsp->v_deth)) limi[0]--;
	     }
	/* blank out the rows up to one before the first observed */
	for (iy = 0, calp = lsp->v_cpt; iy < limi[2]; iy++)
		for (ix = 0; ix < cap->car_maxcols; ix++, calp++)
			zeromem((AVOID *)calp, sizeof (CALPOINT));
	/* blank out rows starting two after the first observed */
	calp = lsp->v_cpt + (limi[3] + 1) * cap->car_maxcols;
	for (iy = limi[3] + 1; iy < cap->car_maxrows; iy++)
	   for (ix = 0; ix < cap->car_maxcols; ix++, calp++)
		zeromem((AVOID *)calp, sizeof (CALPOINT));
	/* blank out columns up to one before the first observed &
	 and starting two after the last */
	for (iy = 0, car = lsp->v_cpt; iy < cap->car_maxrows;
	 iy++, car += cap->car_maxcols)
	   {
		for (ix = 0, calp = car; ix < limi[0]; ix++, calp++)
			zeromem((AVOID *)calp, sizeof (CALPOINT));
		ix = limi[1] + 1;
		for (calp = car + ix; ix < cap->car_maxcols; ix++, calp++)
			zeromem((AVOID *)calp, sizeof (CALPOINT));
	   }
}

static void slide_indices(CALIBRUN *cap, int limi[])
{ /*	This moves the indexings down according to the just-previously
	determined definition of the first and last rows and columns.
	The CALPOINT list itself needs to be adjusted; so do the CTROID
	structures.  We'll recalculate the PARABOLA structures, so they
	don't need to be re-done. */
	int		iy, ix, morefilled, oldobs, pass, wasthere;
	CTROID		*lct;
	CALPOINT	*calp, *car, *ci;
	VSPLINE		*lsp;

	lsp = cap->car_spl;
	lsp->v_midc = cap->car_cencol = cap->car_maxcols / 2;
	lsp->v_midr = cap->car_cenrow = cap->car_maxrows / 2;
	lsp->v_x0cm = -(cap->car_usedrx) * cap->car_cencol;
	lsp->v_y0cm = cap->car_usedry * cap->car_cenrow;
	if (limi[0] == 0 && limi[2] == 0 && limi[1] >= cap->car_maxcols-1 &&
		limi[3] >= cap->car_maxrows-1) return;
	/* ... first, the CALPOINT structures */
	calp = lsp->v_cpt;	car = lsp->v_cpt + limi[2] * cap->car_maxcols;
	for (iy = limi[2]; iy <= limi[3]; iy++, car += cap->car_maxcols)
	 for (ix = limi[0], ci = car + limi[0]; ix <= limi[1];
	  ix++, ci++, calp++)
		cpymem((AVOID *)calp, (AVOID *)ci, sizeof (CALPOINT));
	ci = lsp->v_cpt + cap->car_maxcols * cap->car_maxrows;
	for ( ; calp < ci; calp++) zeromem((AVOID *)calp, sizeof (CALPOINT));
	/* ... and finally the centroids.  This is easy */
	for (lct = cap->car_ctr0; lct < cap->car_ctend; lct++)
	   {
		if (lct->ux >= 0) lct->ux -= limi[0];
		if (lct->uy >= 0) lct->uy -= limi[2];
	   }
	cap->car_ctr = lct;
	/* with a hexagonal grid we must see if we've changed polarity */
	if ((cap->car_flags & CF_HEXAGGRID) && ((limi[0] + limi[2]) % 2))
	  {
		if (cap->car_flags & CF_POLARODD)
			cap->car_flags &= ~CF_POLARODD;
		else	cap->car_flags |= CF_POLARODD;
	  }
	/* update the controlling values */
	lsp->v_ncol = cap->car_maxcols = limi[1] + 1 - limi[0];
	lsp->v_nrow = cap->car_maxrows = limi[3] + 1 - limi[2];
	lsp->v_midc = cap->car_cencol = cap->car_maxcols / 2;
	lsp->v_midr = cap->car_cenrow = cap->car_maxrows / 2;
	lsp->v_x0cm = -(cap->car_usedrx) * cap->car_cencol;
	lsp->v_y0cm = cap->car_usedry * cap->car_cenrow;
	lsp->v_xrcm = cap->car_usedrx; lsp->v_yrcm = -(cap->car_usedry);
	lsp->v_x0px = lsp->v_y0px = -5.;
	lsp->v_xrpx = ((double)lsp->v_detw + 10.) /
		((double)(cap->car_maxcols - 1));
	lsp->v_yrpx = ((double)(lsp->v_deth) + 10.) /
	 ((double)(cap->car_maxrows - 1));
	/* having done that, we try three more times to index things. Before
	 starting, we get rid of all extrapolated points before
	 getting started on the arc-fits. */
	oldobs = lsp->v_ntrue;
	car = lsp->v_cpt + cap->car_cenrow * cap->car_maxcols;
	for (iy = 0, calp = lsp->v_cpt; iy < cap->car_maxrows; iy++)
	   {
		for (ix = 0; ix < cap->car_maxcols; ix++, calp++)
		   if (!(calp->c_ift))
			zeromem((AVOID *)calp, sizeof (CALPOINT));
	   }
	for (morefilled = pass = 0; pass < 3; pass++)
	   {
		get_parabs(0, cap);	/* determine equations of arcs */
		car = lsp->v_cpt + cap->car_cenrow * cap->car_maxcols;
		for (iy = cap->car_cenrow; iy < cap->car_maxrows;
		 iy++, car += cap->car_maxcols)
		   {
			calp = car + cap->car_cencol;
			for (ix = cap->car_cencol; ix < cap->car_maxcols;
			 ix++, calp++)
			   if (!(calp->c_ift))
			     {
				wasthere = calp->c_set;
				if (parab_fit(iy, ix, 2, cap) && !wasthere)
					morefilled++;
			     }
			calp = car + cap->car_cencol - 1;
			for (ix = cap->car_cencol - 1; ix >= 0; ix--, calp--)
			   if (!(calp->c_ift))
			     {
				wasthere = calp->c_set;
				if (parab_fit(iy, ix, 2, cap) && !wasthere)
					morefilled++;
			     }
		   }
		car = lsp->v_cpt + (cap->car_cenrow - 1) * cap->car_maxcols;
		for (iy = cap->car_cenrow - 1; iy >= 0;
		 iy--, car -= cap->car_maxcols)
		   {
			calp = car + cap->car_cencol;
			for (ix = cap->car_cencol; ix < cap->car_maxcols;
			 ix++, calp++)
			   if (!(calp->c_ift))
			     {
				wasthere = calp->c_set;
				if (parab_fit(iy, ix, 2, cap) && !wasthere)
					morefilled++;
			     }
			calp = car + cap->car_cencol - 1;
			for (ix = cap->car_cencol - 1; ix >= 0; ix--, calp--)
		   	   if (!(calp->c_ift))
			     {
				wasthere = calp->c_set;
				if (parab_fit(iy, ix, 2, cap) && !wasthere)
					morefilled++;
			     }
		   }
		if (morefilled + oldobs >=
			cap->car_maxcols * cap->car_maxrows) break;
	   }
	if (morefilled)
	  {
		fprintf(stdout,
		 "%6d more points added in slide_indices\n", morefilled);
		fprintf(cap->car_fpl,
		 "%6d more points added in slide_indices\n", morefilled);
	  }
	if (lsp->v_ntrue > oldobs)
	  {
		fprintf(stdout, "%6d more observed points added there\n",
			lsp->v_ntrue - oldobs);
		fprintf(cap->car_fpl, "%6d more observed points added there\n",
			lsp->v_ntrue - oldobs);
	  }
	get_parabs(1, cap);	/* so we can see the final arcs */
}

static void summ_parab(FILE *fp, CALIBRUN *cap, int nfil)
{ /*	This summarizes the results of the parabolic fittings.
	We list the unindexed points now if we're outputting
	to cap->car_fpl OR if there are fewer than 14. */ 
	char	uxs[5], uys[5];
	int	ix, nnotyet, maxonrow, maxlabel;
	VSPLINE	*lsp;
	CTROID	*lct;

	lsp = cap->car_spl;
	fprintf(fp, "Number of Points added by parabola Fits:  %5d\n",
		nfil);
	fprintf(fp, "Number of Spots Successfully Indexed:     %5d\n",
		lsp->v_ntrue);
	if ((nnotyet = cap->car_ntroids - lsp->v_ntrue))
	  {
		fprintf(fp,
		 "Number of Spots Remaining Unindexed:         %5d\n",
			nnotyet);
		if ((fp == cap->car_fpl) || (nnotyet < 16))
		  {
			maxonrow = (fp == cap->car_fpl) ? 7 : 4;
			maxlabel = maxonrow;
			if (nnotyet < maxlabel) maxlabel = nnotyet;
			for (ix = 0; ix < maxlabel; ix++)
				fprintf(fp,"     X     Y ix iy");
			fprintf(fp, "\n");
			for (lct = cap->car_unindex0, nnotyet = 0;
			 lct < cap->car_ctend; lct++)
			   if (lct->ux < 0 || lct->uy < 0)
			     {
				if (lct->ux < 0) strcpy(uxs, "??");
				else sprintf(uxs, "%2d", lct->ux);
				if (lct->uy < 0) strcpy(uys, "??");
				else sprintf(uys, "%2d", lct->uy);
				fprintf(fp, " %5.1f %5.1f %s %s",
				 lct->cxcen, lct->cycen, uxs, uys);
				if (!((++nnotyet) % maxonrow))
					fprintf(fp, "\n");
			     }
			cap->car_ctr = lct;
			fprintf(fp,"\n");
		  }
	  }
	else fprintf(fp,
		"\t\tAll remaining points were indexed that time\n");
}

int calfill(CALIBRUN *cap)
{ /*	Now we've indexed all the central points, i.e., points for which
	both x and y are within about 140 of their central values.  Now we
	need to index all the other points.  For points whose row value is
	known we do this by finding the column value which fits the row-arc
	equation best; for points whose column value is known we do it by
	finding the row value which fits the column-arc equation best.
	For any point lying outside the range of both row and column
	arcs, we try to catch it on the next pass.  On each pass we widen
	the search space somewhat until we get all the way out to the edge. */
	int		iy, ix, iymax, iymin, ixmax, ixmin, nfill2, nfilled;
	int		pass, nok, wasthere, farok, prevtrue, rclimits[4];
	unsigned	nxy;
	CTROID		*lct;
	VSPLINE		*lsp;
	static CALPOINT	*calp, *car;
	static PARABOLA	*tp;

	/* reserve memory for the parabolas and CALPOINT structures */
	lsp = cap->car_spl;
	nxy = cap->car_maxrows * cap->car_maxcols;
	cap->car_colpara = (PARABOLA *)callop((unsigned)cap->car_maxcols,
	 sizeof (PARABOLA), "cap->car_colpara in calfill");
	cap->car_rowpara = (PARABOLA *)callop((unsigned)cap->car_maxrows,
	 sizeof (PARABOLA), "cap->car_rowpara in calfill");
	lsp->v_cpt = (CALPOINT *)callop((unsigned)nxy, sizeof (CALPOINT),
		"lsp->v_cpt in calfill");
	if ((NULL == cap->car_colpara) || (NULL == cap->car_rowpara) ||
	 (NULL == lsp->v_cpt)) return EALLOC;
	known_first(cap); /* reorganize centroids so indexed ones appear 1st */
	for (lct = cap->car_ctr0, lsp->v_ntrue = 0;
	 lct < cap->car_ctend; lct++)
	   {
	     if ((ix = lct->ux) >= 0 && (iy = lct->uy) >= 0)
	      {
		calp = lsp->v_cpt + cap->car_maxcols * iy + ix;
		if (!(calp->c_ift))		
		  {
			calp->c_ptocx = lct->cxcen; calp->c_ptocy = lct->cycen;
			(lsp->v_ntrue)++;	calp->c_set = calp->c_ift = 1;
		  }
		else
		  {
			fprintf(stderr,
 "Warning! ctr=[%5.1f,%5.1f,%2d,%2d] overlaps [%2d,%2d] @ [%5.1f,%5.1f]\n",
			 lct->cxcen, lct->cycen, lct->ux, lct->uy, ix, iy,
			 calp->c_ptocx, calp->c_ptocy);
		  }
	      }
	     else if (NULL == cap->car_unindex0) cap->car_unindex0 = lct;
	   }
	fprintf(cap->car_fpl,
	 "Number of Spots Indexed, Unindexed by Arcs:  %5d, %5d\n",
		lsp->v_ntrue, (int)(cap->car_ntroids) - lsp->v_ntrue);
	fprintf(stdout,
	 "Number of Spots Indexed, Unindexed by Arcs:  %5d, %5d\n",
		lsp->v_ntrue, (int)(cap->car_ntroids) - lsp->v_ntrue);
	/* stick in some modeled points and match them to observed centroids
	 when we can.  We expand our search gradually.  The last three times
	 through we allow points far away from the detector face. */
		/* get parabola equations and print them */
	get_parabs(cap->car_flags & CF_VERBOSE, cap);
	prevtrue = 1;
	fprintf(cap->car_fpl, "Gradual expansion of observed data region:\n");
	fprintf(cap->car_fpl,
 "Pass iymin iymax ixmin ixmax #new #new obs'd #total #obs'd\n");
	fprintf(cap->car_fpl,
 "                              pts     points points points\n");
	for (pass = 0, nfilled = lsp->v_ntrue; pass < 100; pass++)
	   {
		nok = lsp->v_ntrue;	nfill2 = nfilled;
		car = lsp->v_cpt + cap->car_cenrow * cap->car_maxcols;
		for (iymax = cap->car_maxrows - 1, tp = cap->car_rowpara +
		 cap->car_maxrows - 1; iymax > cap->car_cenrow; iymax--, tp--)
			if (tp->ar_known) break;
		if (iymax < cap->car_maxrows - 1) iymax++;
		for (iymin = 0, tp = cap->car_rowpara; iymin < cap->car_cenrow;
		 iymin++, tp++) if (tp->ar_known) break;
		if (iymin > 0) iymin--;
		for (ixmax = cap->car_maxcols - 1, tp = cap->car_colpara +
		 cap->car_maxcols - 1; ixmax > cap->car_cencol; ixmax--, tp--)
			if (tp->ar_known) break;
		if (ixmax < cap->car_maxcols - 1) ixmax++;
		for (ixmin = 0, tp = cap->car_colpara; ixmin < cap->car_cencol;
		 ixmin++, tp++) if (tp->ar_known) break;
		if (ixmin > 0) ixmin--;
		farok = (pass <= 12 ? 0 : 2);
		for (iy = cap->car_cenrow; iy <= iymax;
		 iy++,car += cap->car_maxcols)
		   {
			calp = car + cap->car_cencol;
			for (ix = cap->car_cencol; ix <= ixmax; ix++, calp++)
			   if (!(calp->c_ift))
			     {
				wasthere = calp->c_set;
				if (parab_fit(iy, ix, farok, cap) &&
				 (!wasthere)) nfilled++;
			     }
			calp = car + cap->car_cencol - 1;
			for (ix = cap->car_cencol - 1; ix >= ixmin;
			 ix--, calp--)
			   if (!(calp->c_ift))
			     {
				wasthere = calp->c_set;
				if (parab_fit(iy, ix, farok, cap) &&
				 (!wasthere)) nfilled++;
			     }
		   }
		car = lsp->v_cpt + (cap->car_cenrow - 1) * cap->car_maxcols;
		for (iy = cap->car_cenrow - 1; iy >= iymin; iy--,
		 car -= cap->car_maxcols)
		   {
			calp = car + cap->car_cencol;
			for (ix = cap->car_cencol; ix <= ixmax; ix++, calp++)
			   if (!(calp->c_ift))
			     {
				wasthere = calp->c_set;
				if (parab_fit(iy, ix, farok, cap) &&
				 (!wasthere)) nfilled++;
			     }
			calp = car + cap->car_cencol - 1;
			for (ix = cap->car_cencol - 1; ix >= ixmin;
			 ix--, calp--)
			   if (!(calp->c_ift))
			     {
				wasthere = calp->c_set;
				if (parab_fit(iy, ix, farok, cap) &&
				 (!wasthere)) nfilled++;
			     }
		   }
		/* re-define the parabola equations */
		get_parabs(0, cap);
		fprintf(cap->car_fpl, " %3d %5d %5d %5d %5d %4d %9d %6d %6d\n",
		 pass, iymin, iymax, ixmin, ixmax, nfilled - nfill2,
		 lsp->v_ntrue - nok, nfilled, lsp->v_ntrue);
		/* give up unless we added an observed point either on
		 this pass or the previous one */
		if ((!prevtrue) && (lsp->v_ntrue == nok)) break;
		prevtrue = lsp->v_ntrue - nok;
	   }
	summ_parab(cap->car_fpl, cap, nfilled);
	summ_parab(stdout, cap, nfilled);
	/* elim rows&columns far from observed vals */
	kill_wayout(cap, rclimits);
	slide_indices(cap, rclimits); /* reindex so useful ix,iy start @ 0*/
/* if the user has specified that he doesn't trust the arc-extrapolations,
	we get rid of the extrapolated points at this stage of things. */
	if (cap->car_flags & CF_NOEXTRAP)
	  {
		for (iy = 0, calp = lsp->v_cpt; iy < cap->car_maxrows; iy++)
		  for (ix = 0; ix < cap->car_maxcols; ix++, calp++)
		    if (!(calp->c_ift))
		     {
			calp->c_ptocx = calp->c_ptocy = calp->c_set = 0;
		     }
	  }
	cap->car_tl0 = NOW;
	fprintf(cap->car_fpl,
	 " %4d seconds were required to assign the peaks\n",
		(int)(cap->car_tl0 - cap->car_tl1));
	if (cap->car_flags & CF_VERBOSE)
	  {
		fprintf(cap->car_fpl,
		  " Locations of observed centroids in fiducial image:\n");
		for (iy = 0; iy < cap->car_maxrows; iy++)
		   {
			for (ix = 0; ix < cap->car_maxcols; ix++)
			   {
				disp_calib(iy, ix, 'y', cap->car_fpl, cap);
				if (!((ix+1) % 21)) fputs("\n", cap->car_fpl);
			   }
			if (ix % 21) fputs("\n", cap->car_fpl);
			for (ix = 0; ix < cap->car_maxcols; ix++)
			   {
				disp_calib(iy, ix, 'x', cap->car_fpl, cap);
				if (!((ix+1) % 21)) fputs("\n", cap->car_fpl);
			   }
			if (ix % 21) fputs("\n", cap->car_fpl);
			fputs("\n", cap->car_fpl);
		   }
	   }
	memzap((AVOID **)&cap->car_colpara, "cap->car_colpara in calfill");
	memzap((AVOID **)&cap->car_rowpara, "cap->car_rowpara in calfill");
	return EOK;
}

/* end of calfill.c */
