/* calibmain.c -- Toplevel routines for `calibrate' functionality in XGEN

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	14 Feb 2007: calib_defaults: we don't need ssize and stype
	13 Dec 2006: bugfix in spot overflow in do_cenmass;
	 handle LINEAR correctly for CALCENTROIDS output case
	24 Jun 2006: include xgimage.h; allow FMT_FASTALT in getsigmult
	17 Jun 2006: replace act_pixel with activmask(x,y,041)
	 5 Jun 2006: housekeeping
	12 May 2006: merge calspots.c and calnoim.c into here;
		xcmp moved from calindex because it doesn't get used there
	 3 Jan 2006: elim record_detxy calls by remembering structure elts;
		don't qsort with fewer than 2 elements
	20 Nov 2005: inline function declarations & function prototypes
	28 Oct 2005: rename centroid() -> calibcen() (this is more descriptive);
		replace static globals with elements of CALSPOTWORK structure
	11 Oct 2005: housekeeping
	12 Aug 2005: calverbose in calinv() is now 3, not 1
	 6 Aug 2005: calinv() called with third (calverbose) argument
	28 Jun 2005: do_cenmass(): typecast pointer difference upon printout
	11 May 2005: check return status from xgjournal
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	19 Nov 2004: in header_linear(), make search limits <= fmmax;
		fix silly bug in loop -- fra++, not lfmi++ (duh)
	28 Aug 2004: routines used only here are declared as static; cosmetics
	 2 Aug 2004: modify cm2pixq() call in recenter() to keep Linux happy;
		add new arg to find_dimension() and record_detxy
	19 Jun 2004: calib_defaults() created: sets defaults based on detector
		type and size using get_calibparams();
		add additional flags -e,p,z negate sense of -b,n,v.
	18 Feb 2004: change kill_ftab call so that it doesn't have args
	26 Jan 2004: convert some static globals to local variables;
		cdname made an allocatable variable
	21 Jan 2004: license message inserted and initial comment reordered
	 6 Jan 2004: CALIBRUN & VSPLINE structure elements in function calls
		replace extern globals; kill_ftab & kill_mail called at end
	20 Nov 2003: replace global variables with elements of CALIBRUN struct
	21 Sep 2003: set the detector width & height based on the test image
		for all linear detector images, even if there's already
		a USPLINE present with the wrong width and height present
	19 Sep 2003: explicitly set locn and dimname to NULL before using them
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Sep 2003: replace access() with xgaccess() calls;
		enables gunzipping on the fly; eliminate pre-check
		for amask file before calling get_actmask, since
		there's an xgaccess call within get_actmask anyway
	13 Aug 2003: putspline() moved to ../xglib/spline.c;
		and now it takes an argument.
	 5 Jan 2003: reorganized source-file header
	 2 Jan 2003: remove call to access() for CALDATA before reading it
		so that gzipped versions can be read
	 1 Nov 2002: access(locdet, 2) calls replaced with access(locdet, 4);
		trivial reformatting elsewhere
	20 Sep 2002: CF_* flags moved to ../include/xgen.h
	25 Mar 2002: removed MSI attachments
	 6 Dec 1995: enlarge `cdname' array to handle long names
	31 Aug 1995: call record_detxy before output
		so spline.c doesn't get confused
	25 Aug 1995: output VSPLINE is 512 bytes even on 64-bit machines
	31 May 1995: "register" -> "register int" throughout;
		use %d format in tl0-tl1 calculations
	20 May 1995: acknowledge return code from ext_troids()
	12 Apr 1995: MAXSPOTS=14000: handle fiducial plates with lots of holes;
		boxxy=3.5*sqrt(minlim) in all cases, not just minlim>10
	 2 Mar 1995: diagnostics to stdout; framenumber arg in cm2pix
		shared-memory alloc & free; smaller buffer to read cal troids;
		spot-too-broad redined if width>512
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 9 Dec 1994: prevent divide-by-zero
		when called from GUI in `getsigmult'
	20 Oct 1994: set upper limit of Y in centroid searches--
		avoid memory errors
	 5 Oct 1994: use xgwrite to output VSPLINE structure
	14 Jun 1994: fix "linear" option to get detector dimensions right
	 5 May 1994: source file headers, POLYDET->XGEN,
		"if (EOK != function())"
	16 Mar 1994: major rewrite of `calspots'; bugfixes elsewhere
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: Initial X-GEN version

   Purpose: startup routines and top-level processor for calibration, i.e.,
	the mapping from pixels to cm on the detector face and vice versa;
	also includes routines for processing real & fake calibration spots.
   Cross references: calindex.c, calfill.c, ../xglib/calinv.c

*************************************************************************/
#define		NEED_TIME	1
#include	"calibdefs.h"
#include	"xgimage.h"
#define	DEFMINLIM	(5)	/* min #pixels above thresh. per spot */
#define	MAXSPOTS	14000	/* max # of recordable spots in image*/
#define	SIGMULT		4.	/* pixels >localmean +4.*sqrt(<count>) */
#define	MAX_MCOUNT	25.	/* maximum default sigma multiplier */
#define	BORDEROFF	1	/* # of points in from edge to start looking */
#define	BOXSEARCH	10	/* dimension of search box */
#define	WARPFACTOR	(0.04)	/* fractional warping in X */
#define	MINSPOTS	(40)	/* must get >= MINSPOTS centroids */
#define	MEANHALFW	31
#define	MEANHALFH	31
#define	MEANW	(MEANHALFW + MEANHALFW + 1)
#define	MEANH	(MEANHALFH + MEANHALFH + 1)

 typedef struct { /* structure to simplify calls to do_cenmass */
		int		cs_boxxy, cs_qboxxy, cs_xn, cs_y00, cs_yn0;
		short		*cs_xend;
		double		cs_cuto2;
		} CALSPOTWORK;

static void do_cenmass(int ix, int iy, CALIBRUN *cap, CALSPOTWORK *csp)
{ /*	This performs the calculations for a single center of mass.
	If (cap->car_flags & CF_USEBORDER) is true it won't process pixels
	that lie outside the active area. */
	int		ixb, ixn, iy0, iyb, iyn, ninspot;
	double		inten, peakspot, total, totaly, xc, xcy, yc, ycy;
	MAIL		*lhd;
	CTROID		*lct;

	total = ninspot = peakspot = xc = yc = 0;
	iy0 = iy - csp->cs_qboxxy; if (iy0 < csp->cs_y00) iy0 = csp->cs_y00;
	ixn = ix + csp->cs_boxxy - 1; if (ixn > csp->cs_xn) ixn = csp->cs_xn;
	iyn = iy0 + csp->cs_boxxy; if (iyn > csp->cs_yn0) iyn = csp->cs_yn0;
	lhd = cap->car_heads;
	for (ixb = ix - 1; ixb <= ixn; ixb++)
	   {
		xcy = ycy = totaly = 0;
		for (iyb = iy0; iyb <= iyn; iyb++)
		   if (((csp->cs_xend)[iyb] < ixb) && ((inten = ((double)
		    (imvalue(lhd, ixb, iyb)) - csp->cs_cuto2)) > 0.) &&
		    ((!(cap->car_flags & CF_USEBORDER)) ||
			(NULL != activmask(ixb, iyb, 041))))
		     {
			if (inten > peakspot) peakspot = inten;
			ninspot++;		xcy += (ixb - ix) * inten;
			totaly += inten;	ycy += (iyb - iy) * inten;
		     }
		total += totaly;	xc += xcy;	yc += ycy;
	   } /* end of "for (ixb " loop */
	if ((total <= csp->cs_cuto2) || (ninspot < cap->car_minlim / 2))
	  {
		if (cap->car_flags & CF_VERBOSE) fprintf(cap->car_fpl,
			" %4d %4d %3d  too few pts: %8.1f\n",
			ix, iy, ninspot, total);
		return;
	  }
	/* remember that we've covered this range: */
	for (iyb = iy0; iyb < iyn; iyb++) (csp->cs_xend)[iyb] = ixn;
	lct = cap->car_ctr;
	lct->cxcen = xc / total + (double)ix;
	lct->cycen = yc / total + (double)iy;
	lct->cinten = total;	lct->ux = ninspot;
	lct->ccutoff = csp->cs_cuto2 * 2.;
	if (cap->car_flags & CF_VERBOSE) fprintf(cap->car_fpl,
		" %4d %4d %3d %6.1f %6.1f %8.1f\n",
		ix, iy, ninspot, lct->cxcen, lct->cycen, total);
	if (lct >= cap->car_ctend - 1)
	  {
		fprintf(stderr,
		 "Warning: %4d spots found; all others ignored\n",
			(int)(lct - cap->car_ctr0));
		fprintf(cap->car_fpl ,
		 "Warning: %4d spots found; all others ignored\n",
			(int)(lct - cap->car_ctr0));
	  }
	else {
		lct++;	cap->car_ctr = lct;
	     }
}

static double cradsq(CTROID *ct0, CTROID *ct1)
{ /*	This returns the radial distance squared between two centroids */

	return (ct0->cxcen - ct1->cxcen) * (ct0->cxcen - ct1->cxcen) +
		(ct0->cycen - ct1->cycen) * (ct0->cycen - ct1->cycen);
}

static void combine_ctr(CTROID *ct0, CTROID *ct1, CALIBRUN *cap)
{ /*	This merges centroid `ct1' into `ct0' and pops the stack.*/
	double		intsum, wt0, wt1;
	CTROID		*ctp, *ctq;

	if (ct0->cinten < DFUZZ || ct1->cinten < DFUZZ) return;
	intsum = ct0->cinten + ct1->cinten;
	wt0 = ct0->cinten / intsum;	wt1 = 1. - wt0;
	if (cap->car_flags & CF_VERBOSE) fprintf(cap->car_fpl,
	 "Merging [%6.1f,%6.1f] into [%6.1f,%6.1f]; ",
		ct1->cxcen, ct1->cycen, ct0->cxcen, ct0->cycen);
	ct0->cxcen = ct0->cxcen * wt0 + ct1->cxcen * wt1;
	ct0->cycen = ct0->cycen * wt0 + ct1->cycen * wt1;
	ct0->cinten = intsum;	/* ... in case we need to merge still others*/
	if (cap->car_flags & CF_VERBOSE) fprintf(cap->car_fpl,
	 "Merged position [%6.1f,%6.1f]\n", ct0->cxcen, ct0->cycen);
	for (ctp = ct1 + 1, ctq = ct1; ctp < cap->car_ctend; ctp++, ctq++)
		cpymem((AVOID *)ctq, (AVOID *)ctp, sizeof (CTROID));
	(cap->car_ctend)--;	(cap->car_ntroids)--;
}

static int err_ctroid(int irv, CALIBRUN *cap)
{ /*	This prints a diagnostic and returns its called value */
	CTROID		*lct;
	static int	needcr = 0;

	lct = cap->car_ctr;
	if (cap->car_flags & CF_VERBOSE)
	  {
		fprintf(cap->car_fpl, " %7.2f %7.2f %7d %9.1f ",
			lct->cxcen, lct->cycen, lct->ux, lct->cinten);
		     if (irv == 1) fprintf(cap->car_fpl, "X out of range ");
		else if (irv == 2) fprintf(cap->car_fpl, "Y out of range ");
		else if (irv == 3) fprintf(cap->car_fpl, "too few points ");
		else if (irv == 4) fprintf(cap->car_fpl, "inten. too low ");
		else if (irv == 5) fprintf(cap->car_fpl, "center too weak");
		else if (irv == 0) fprintf(cap->car_fpl, "usable centroid");
		if (needcr) fprintf(cap->car_fpl, "\n");
		needcr = 1 - needcr;
	  }
	return irv;
}

static int badctroid(CALIBRUN *cap)
{ /*	This looks at the current calibration centroid and judges whether
	it's valid or not.  The criteria for validity are:
	(1) X must fall between 1 and  width - 1.
	(2) Y must fall between 1 and height - 1.
	(3) it must contain at least `cap->car_minlim' pixels;
	(4) total intensity >= 2 * `cap->car_minlim' * the local cutoff;
	(5) intensity in the centermost bunch of pixels must be >= 4 * cutoff.
		A "bunch" is four pixels for a 512x512 detector;
		it's eight pixels for a 1024x1024 detector.
	Each failure returns a different code; success is 0.	*/
	int	locht, locwid, lx0, lx1, lx2, ly0, ly1, ly2;
	double	totc;
	MAIL	*lhd;
	VSPLINE	*lspl;
	CTROID	*lct;

	lct = cap->car_ctr;	lhd = cap->car_heads;	lspl = cap->car_spl;
	locwid = lspl->v_detw;	locht = lspl->v_deth;
	lx0 = (int)(lct->cxcen);	ly0 = (int)(lct->cycen);
	if (lx0 < 1 || lx0 >=  locwid - 2) return err_ctroid(1, cap);
	if (ly0 < 1 || ly0 >= locht - 2) return err_ctroid(2, cap);
	if (lct->ux < cap->car_minlim) return err_ctroid(3, cap);
	if (lct->cinten < 2. * lct->ccutoff * (double)cap->car_minlim)
		return err_ctroid(4, cap);
	lx1 = lx0 + 1;	ly1 = ly0 + 1;
	if (locwid > 512)
	  {
		if (lct->cxcen > 0.5 + (double)lx0) lx2 = lx1 + 1;
		else lx2 = lx0 - 1;
		if (lct->cycen > 0.5 + (double)ly0) ly2 = ly1 + 1;
		else ly2 = ly0 - 1;
		totc = imvalue(lhd, lx0, ly0) + imvalue(lhd, lx0, ly1) +
		 imvalue(lhd, lx1, ly0) + imvalue(lhd, lx1, ly1) +
		 imvalue(lhd, lx2, ly0) + imvalue(lhd, lx2, ly1) +
		 imvalue(lhd, lx0, ly2) + imvalue(lhd, lx1, ly2);
	  }
	else totc = imvalue(lhd, lx0, ly0) + imvalue(lhd, lx0, ly1) +
		imvalue(lhd, lx1, ly0) + imvalue(lhd, lx1, ly1);
	if (totc < 4. * lct->ccutoff) return err_ctroid(5, cap);
	return err_ctroid(0, cap);
}

static void mrg_points(CALIBRUN *cap)
{ /*	This goes through the CTROID list and merges together points which
	are within a squared radial distance of `cap->car_minradsq' of
	one another.  It then goes through the merged list and eliminates
	reflections that are illegal for various reasons, viz. too close
	to the detector edge, not enough points above the cutoff, total
	intensity too low, or intensity of center four pixels too low. */

	int	i;
	double	locd;
	CTROID	*ctp, *lct;

	cap->car_ctend = cap->car_ctr0 + cap->car_ntroids;
	locd = 2. * cap->car_mincordist;
	for (lct = cap->car_ctr0; lct < cap->car_ctend; lct++)
	  for (ctp = lct + 1; ctp < cap->car_ctend; ctp++)
	     {
		if (ctp->cxcen - lct->cxcen > locd) break;
		if (cradsq(lct, ctp) < cap->car_minradsq)
		   {
			combine_ctr(lct, ctp, cap);	if (ctp > lct) ctp--;
		   }
	     }
	fprintf(cap->car_fpl,
	 "Number of Spots after Merging Close Contacts:%5d\n",
		(int)(cap->car_ntroids));
	printf("Number of Spots after Merging Close Contacts:%5d\n",
		(int)(cap->car_ntroids));
	if (cap->car_flags & CF_VERBOSE)
	  {
		fprintf(cap->car_fpl, "Analysis of merged centroids:\n");
		for (i = 0; i < 2; i++) fprintf(cap->car_fpl,
 " Xcenter Ycenter #points Intensity result         ");
		fprintf(cap->car_fpl, "\n");
	  }
	for (lct = cap->car_ctr0; lct < cap->car_ctend; )
	   {
		cap->car_ctr = lct;
		if (badctroid(cap))
		  {
			(cap->car_ctend)--;	(cap->car_ntroids)--;
			if (lct == cap->car_ctend) break;
			swap((void *)lct,
				(void *)(cap->car_ctend), sizeof (CTROID));
		  }
		else lct++;
	   }
	cap->car_ctr = lct;
	fprintf(cap->car_fpl,
	 "Number of Spots after eliminating bad spots:%5d\n",
		(int)(cap->car_ntroids));
	printf("Number of Spots after eliminating bad spots:%5d\n",
		(int)(cap->car_ntroids));
	if (cap->car_flags & CF_VERBOSE) list_troids(cap->car_fpl,
	 cap->car_ctr0, cap->car_ntroids, "after close contacts");
}

static int outcen(CALIBRUN *cap)
{ /*	This writes the merged centroids to a file
	rather than doing a USPLINE */
	FILE	*fpet;
	CTROID	*lct;

	if (qerror((AVOID *)(fpet = xgopen("CALCENTROIDS", "w")),
	 "open output calibration-centroids file")) return EOPENERR;
	for (lct = cap->car_ctr0; lct < cap->car_ctend; lct++)
		fprintf(fpet, " %7.2f %7.2f\n", lct->cxcen, lct->cycen);
	cap->car_ctr = lct;
	(void)fclose(fpet);
	if (stdout != cap->car_fpl) (void)fclose(cap->car_fpl);
	cap->car_fpl = NULL;	return EOK;
}

static int xcmp(CTROID *ct0, CTROID *ct1)
{ /*	Comparison routine used in sorting on xcen */
	return (ct0->cxcen < ct1->cxcen) ? -1 : 1;
}

static int calibcen(CALIBRUN *cap)
{ /*	routine "calibcen":
	This computes the centroids of all the peaks found within the data
	in a single frame.  A peak is a collection of pixels with intensity
	values greater than `localcut' inside a search box which is
	BOXSEARCHxBOXSEARCH in size. The resulting centroids are stuffed
	in an array; any peak with at least `cap->car_minlim' pixels over the
	cutoff in it are written to disk.
	This routine passes through the (width*height) array and looks for
	the spots. It calls do_cenmass to calculate the center of mass of each.
	If (cap->car_flags & CF_USEBORDER) is true, we ignore pixels that
	are illegal according to the AMASK file. */
	int		ix, iy, lmean, lmct, locht, locwid, nregion;
	int		wx, wxe, wy, wye;
	short		iyn8, x00;
	double		localcut, meanc, sigc;
	VSPLINE		*lspl;
	MAIL		*lhd;
	CALSPOTWORK	*csp;

	lspl = cap->car_spl;	locwid = lspl->v_detw;	locht = lspl->v_deth;
	/* get memory for a working structure and
	 an array of end-of-column counters*/
	if (NULL == (csp = (CALSPOTWORK *)callop(1, sizeof (CALSPOTWORK),
	 "csp in calibcen"))) return EALLOC;
	if (NULL == (csp->cs_xend = (short *)callop(locht, sizeof (short),
	 "xend in calibcen"))) return EALLOC;
	x00 = BORDEROFF;	csp->cs_xn = locwid - BORDEROFF - 1;
	csp->cs_y00 = BORDEROFF;	csp->cs_yn0 = locht - BORDEROFF - 1;
	csp->cs_boxxy = BOXSEARCH;
	/* previously we required minlin > 10 to reset this */
	csp->cs_boxxy = 3.5 * sqrt((double)(cap->car_minlim));
	csp->cs_qboxxy = csp->cs_boxxy / 4;
	for (iy = csp->cs_y00; iy < csp->cs_yn0; iy++) csp->cs_xend[iy] = 0;
	if (cap->car_flags & CF_VERBOSE) fprintf(cap->car_fpl,
	 "   ix   iy npt   xcen   ycen    inten\n");
	iyn8 = csp->cs_yn0 - 4;	cap->car_ctr = cap->car_ctr0;
	lhd = cap->car_heads;
	for (nregion = 0, wx = x00 + 1; wx <= csp->cs_xn - 4; wx += MEANW)
	   {
		wxe = wx + MEANW;	if (wxe >= locwid) wxe = locwid - 1;
		for (wy = csp->cs_y00 + csp->cs_qboxxy; wy <= (int)iyn8;
		 wy += MEANH, nregion++)
		 {
			wye = wy + MEANH; if (wye >= locht) wye = locht - 1;
			if (cap->car_sigmult > DFUZZ)
			  { /* if the user has asked to compare the intensity
				with the local mean count, we determine the
				local mean count and its sigma here.  We then
				use that for defining local cutoff intensity,
				overriding the overall cutoff if the latter
				is lower. */
				for (ix = wx, lmean = lmct = 0; ix <= wxe; ix++)
				   for (iy = wy; iy < wye; iy++)
				      {
					lmean += imvalue(lhd, ix, iy); lmct++;
				      }
				if (lmct < 1) break;	/* not enough points */
				meanc = (double)lmean / (double)lmct;
				sigc = sqrt(meanc);
				localcut = meanc + cap->car_sigmult * sigc;
				if (cap->car_overcut > localcut)
					localcut = cap->car_overcut;
				if (cap->car_flags & CF_VERBOSE)
				  {
					fprintf(cap->car_fpl,
				  "Region bounded by [%4d,%4d],[%4d,%4d]: ",
						wx, wy, wxe, wye);
					fprintf(cap->car_fpl,
				  "<count>,sigma,cutoff = %7.2f,%7.3f,%8.1f\n",
						meanc, sigc, localcut);
				  } /* end of "if (cap->car_flags" clause */
			  } /* end if "if (cap->car_sigmult" clause */
		   	else localcut = cap->car_overcut;
			csp->cs_cuto2 = localcut / 2.;
			for (iy = wy; iy < wye; iy++) csp->cs_xend[iy] = 0;
			for (ix = wx; ix < wxe; ix++)
			   {
				for (iy = wy; iy < wye; iy++)
				   if (ix >= (int)(csp->cs_xend)[iy])
				     {
					for ( ; (csp->cs_xend)[iy] &&
					  iy < (int)iyn8; iy++)
						(csp->cs_xend)[iy] = 0;
					if ((iy < iyn8) &&
					 (imvalue(lhd, ix, iy) > localcut) &&
					 (imvalue(lhd, ix, iy + 1) > localcut))
					  {
						do_cenmass(ix, iy, cap, csp);
						if (cap->car_ctr >=
						 cap->car_ctend -1) break;
					  } /* end of "if ((iy" clause */
				      } /* end of "if (ix" & loop thru iy*/
				if (cap->car_ctr >= cap->car_ctend - 1) break;
			    } /* end of loop through ix */
		 } /* end of loop through wy */
	   } /* end of loop through wx */
	memzap((AVOID **)&(csp->cs_xend), "xend in calibcen");
	memzap((AVOID **)&csp, "csp in calibcen");
	cap->car_ntroids = cap->car_ctr - cap->car_ctr0;
	if (cap->car_ntroids < MINSPOTS)
	  {
		fprintf(stderr,
 "Only %3d centroid%sfound; we cannot construct the calibration with them\n",
		 (int)(cap->car_ntroids), (cap->car_ntroids == 1 ? " " : "s "));
		fprintf(cap->car_fpl,
 "Only %3d centroid%sfound; we cannot construct the calibration with them\n",
		 (int)(cap->car_ntroids),(cap->car_ntroids == 1 ? " " : "s "));
		if (stdout != cap->car_fpl) (void)fclose(cap->car_fpl);
		cap->car_fpl = NULL;	return EBAD;
	  }
	printf(
	 "Number of Spots found in Fiducial Image:     %5d\n",
		(int)(cap->car_ntroids));
	fprintf(cap->car_fpl,
	 "Number of Spots found in Fiducial Image:     %5d\n",
		(int)(cap->car_ntroids));
	cap->car_tl1 = NOW;
	fprintf(cap->car_fpl,
	 " %4d seconds required to find the centroids\n",
		(int)(cap->car_tl1 - cap->car_tl0));
	/* a safe estimate of the upper limit of the number of rows and columns
	is 8+sqrt(2*ncentroids); thus for a 300-spot set it's 8+24.5=33. */
	cap->car_maxrows = cap->car_maxcols =
		8 + (int)(sqrt(2. * (double)(cap->car_ntroids)));
	/* special case: for hexagonal grids we need to put in as many fake
	spots as real ones, so we must double the apparent number of rows */
	if (cap->car_flags & CF_HEXAGGRID) cap->car_maxrows *= 2;
	/* minimum distance between points (I hope) */
	cap->car_minradsq = 0.390625 * ((double)(locwid * locht)) /
		((double)(cap->car_maxrows * cap->car_maxcols));
	cap->car_mincordist = 2.5 * sqrt(cap->car_minradsq);
	/* sort the resulting centroids on x */
	if (cap->car_ntroids > 1) qsort((AVOID *)(cap->car_ctr0),
	 (size_t)(cap->car_ntroids), sizeof (CTROID), (QPE)xcmp);
	/* merge neighbors together and eliminate those that are weak
	 or don't have enough points; then sort again. */
	mrg_points(cap);
	if (cap->car_ntroids > 1) qsort((AVOID *)(cap->car_ctr0),
	 (size_t)(cap->car_ntroids), sizeof (CTROID), (QPE)xcmp);
	/* recompute maxrows, maxcols, etc. based on smaller troid list */
	cap->car_maxrows = cap->car_maxcols =
		8 + (int)(sqrt(2. * (double)(cap->car_ntroids)));
	if (cap->car_flags & CF_HEXAGGRID) cap->car_maxrows *= 2;
	cap->car_minradsq = 0.390625 * ((double)(locwid * locht)) /
		((double)(cap->car_maxrows * cap->car_maxcols));
	cap->car_mincordist = 2.5 * sqrt(cap->car_minradsq);
	/* index of centermost row & column */
	cap->car_cenrow = (cap->car_maxrows + 1) / 2;
	cap->car_cencol = (cap->car_maxcols + 1) / 2;
	return EOK;
}

static void lin_troids(int nlin, CALIBRUN *cap)
{ /* This manages the creation of a list of fake CTROIDs for the linear map */
	register int	ix, iy;
	int		locht, locwid;
	double		rpx, rpx0, rpxe, rpy, rpy0, rpye, rxd, ryd;
	CTROID		*lct;
	VSPLINE		*lsp;

	cap->car_tl1 = NOW;	/* so timings in calfill, etc. work properly */
		/* user specifies how many columns */
	lsp = cap->car_spl;	lsp->v_ncol = cap->car_maxcols = nlin;
	locwid = lsp->v_detw;	locht = lsp->v_deth;
	/* number of rows is proportional to `nlin', but it goes up or down
	from that based on the ratio of height to width  */
	lsp->v_nrow = cap->car_maxrows =
	 ((double)nlin) * ((double)locht) / ((double)locwid) + 0.49999;
	cap->car_cenrow = cap->car_maxrows / 2;
	cap->car_cencol = cap->car_maxcols / 2;
	lsp->v_x0px = lsp->v_y0px = rpx0 = rpy0 = -5.;
	rpxe = ((double)locwid) + 5.;
	rpye = ((double)locht) + 5.;
	lsp->v_xrpx = rxd = (rpxe - rpx0) / ((double)(lsp->v_ncol - 1));
	lsp->v_yrpx = ryd = (rpye - rpy0) / ((double)(lsp->v_nrow - 1));
	/* construct a list of faked centroids as a simple linear map */
	for (iy = 0, rpy = rpy0, lct = cap->car_ctr0; iy < cap->car_maxrows;
	 iy++, rpy += ryd)
	   {
		for (ix = 0, rpx = rpx0; ix < cap->car_maxcols;
		 ix++, rpx += rxd, lct++)
		   {
			lct->cxcen = rpx;	lct->cycen = rpy;
			lct->ux = ix;	lct->uy = iy;	lct->cinten = 1000.;
		   }
	   }
	cap->car_ctend = cap->car_ctr = lct;	/* count # of points: */
	cap->car_ntroids = cap->car_maxrows * cap->car_maxcols;
	/* make sure we don't end up merging anything later */
	cap->car_minradsq = lsp->v_xrpx;
	if (lsp->v_yrpx > cap->car_minradsq) cap->car_minradsq = lsp->v_yrpx;
	cap->car_minradsq *= 1.3 * cap->car_minradsq;
	cap->car_mincordist = 2.5 * sqrt(cap->car_minradsq);
}

static void sum_calcenin(FILE *fp, int nind, CALIBRUN *cap)
{ /*	Summarize the results of reading CALCENTROIDS */
	fprintf(fp,
	 " %6d centroids read in from CALCENTROIDS file\n",
	 (int)(cap->car_ntroids));
	if (nind) fprintf(fp, " %6d were already indexed\n", nind);
}

int ext_troids(int *nindexed, CALIBRUN *cap)
{ /*	This reads in the file with logical name CALCENTROIDS and generates a
	list of centroids from it. Each row of the file may have up to seven
	centroids listed on it in the format
	X Y ix iy
	where X and Y are the coordinates of the centroid in pixels and
	ix and iy are the indices in the horizontal and vertical directions.
	Any centroid listed with one or two negative indices is ignored. */
	char		cline[64];
	int		lnumc, lnumr, locht, locwid, maxux, maxuy;
	int		minux, minuy, nitems, uxil, uyil;
	double		xil, yil;
	FILE		*fpet;
	VSPLINE		*lspl;
	CTROID		*lct;

	if (qerror((AVOID *)(fpet = xgopen("CALCENTROIDS", "r")),
	 "open external calibration centroids file")) return EOPENERR;
	*nindexed = 0;	maxux = maxuy = -10000;	minux = minuy = 10000;
	lspl = cap->car_spl;	locwid = lspl->v_detw;	locht = lspl->v_deth;
	for (lct = cap->car_ctr0, cap->car_ntroids = 0;
	 NULL != fgets(cline, 62, fpet); )
	   {
		nitems = sscanf(cline, "%lf %lf %d %d", &xil,&yil,&uxil,&uyil);
		lct->cxcen = xil;	lct->cycen = yil;
		if ((nitems > 2) && (cap->car_flags & CF_INDEXED))
		  {
			lct->ux = uxil;	lct->uy = uyil; (*nindexed)++;
			if (uxil < minux) minux = uxil;
			if (uyil < minuy) minuy = uyil;
			if (uxil > maxux) maxux = uxil;
			if (uyil > maxuy) maxuy = uyil;
		  }
		else {
			lct->ux = lct->uy = 0;
		     }
		lct->cinten = 1000.;	lct++;	(cap->car_ntroids)++;
	   }
	(void)fclose(fpet);	cap->car_ctend = cap->car_ctr = lct;
	/* a safe estimate of the upper limit of the number of rows and columns
	is 8+sqrt(2*ncentroids); thus for a 300-spot set it's 8+24.5=33.
	For fully-indexed sets, we know how many rows and columns there are. */
	cap->car_maxrows = cap->car_maxcols =
	 8 + (int)sqrt(2. * (double)(cap->car_ntroids));
	if (*nindexed)
	  {
		lnumr = maxuy + 1 - minuy;	lnumc = maxux + 1 - minux;
		if (*nindexed >= (int)(cap->car_ntroids))
		  {
			cap->car_maxrows = lnumr; cap->car_maxcols = lnumc;
		  }
		else {
			if (lnumr > cap->car_maxrows) cap->car_maxrows = lnumr;
			if (lnumc > cap->car_maxcols) cap->car_maxcols = lnumc;
		     }
	  }
	/* special case: for hexagonal grids we need to put in as many fake
	spots as real ones, so we must double the apparent number of rows */
	if (cap->car_flags & CF_HEXAGGRID) cap->car_maxrows *= 2;
	sum_calcenin(cap->car_fpl,  (*nindexed), cap);
	sum_calcenin(stdout, (*nindexed), cap);
	/* index of centermost row & column */
	cap->car_cenrow = (cap->car_maxrows + 1) / 2;
	cap->car_cencol = (cap->car_maxcols + 1) / 2;
	/* minimum distance between points (I hope) */
	cap->car_minradsq = 0.390625 * ((double)(locwid * locht)) /
		((double)(cap->car_maxrows * cap->car_maxcols));
	cap->car_mincordist = 2.5 * sqrt(cap->car_minradsq);
	/* there's no guarantee that the centroids have been input
	 in any particular sorting order, so we'll sort 'em */
	if (cap->car_ntroids > 1) qsort((AVOID *)(cap->car_ctr0),
	 (size_t)(cap->car_ntroids), sizeof (CTROID), (QPE)xcmp);
	cap->car_tl1 = NOW; /* so the timing in `calindex' + `calfill' works */
	return EOK;
}

static void sum_calset(FILE *fp, CALIBRUN *cap)
{ /*	This summarizes the set-up of a run of "CALIBRATE" to the stream fp.*/
	static char	*fused[] = {
	 "caldata", "uspline", "xlog", "calcentroids",
		"amask", "darkimage", NULL };
	static char	ftypes[] = { 1, 2, 2, 3, 1, 1, 0 };

	ftypes[0] = (cap->car_linear || (cap->car_flags & CF_READCEN)) ? 0 : 1;
	ftypes[1] = (cap->car_flags & CF_WRITECEN) ? 0 : 2;
	ftypes[2] = (cap->car_fpl == stdout ? 0 : 2);
	ftypes[3] = (cap->car_flags & CF_READCEN) ? 1 :
		((cap->car_flags & CF_WRITECEN) ? 2 : 0);
	ftypes[4] = (cap->car_flags & CF_USEBORDER) ? 1 : 0;
	ftypes[5] = (cap->car_imflags & DARK_CURRENT) ? 1 : 0;
	header(fp, PROCESSING, fused, ftypes);
	if (cap->car_flags & CF_WRITECEN) fprintf(fp,
 "An ascii CALCENTROIDS list, not a USPLINE file, will be output\n");
	else {
		fprintf(fp,
	 "Direct Beam Position at Swing = 0:   [%5.1f,%5.1f]\n",
			cap->car_beamx, cap->car_beamy);
		fprintf(fp,
 "Spacing between fiducial points = %7.3f cm in X, %7.3f cm in Y\n",
			cap->car_usedrx, cap->car_usedry);
		if (cap->car_flags & CF_NOEXTRAP) fprintf(fp,
	 "\tNo arc-extrapolated points will be input to the inverse map\n");
	      }
	fprintf(fp, "\tLogfile will receive %s diagnostics\n",
		(cap->car_flags & CF_VERBOSE) ? "detailed" : "brief");
	if (cap->car_flags & CF_HEXAGGRID) fprintf(fp,
	 "Hexagonal grid rather than rectangular grid assumed\n");
	if (cap->car_flags & CF_USEBORDER) fprintf(fp,
	  "Only fiducial spots in permitted regions will be used\n");
	if (cap->car_linear)
	  {
		fprintf(fp,
 "Linear mapping with %d rows and %d columns of artificial centroids used\n",
			(int)(cap->car_maxrows), cap->car_linear);
		return;
	  }
	if (cap->car_flags & CF_READCEN) fprintf(fp,
 "An ascii CALCENTROIDS list will be read instead of a brass-plate image\n");
	else {
		fprintf(fp, "Spots must be > local mean + %7.2f * sigma\n",
			cap->car_sigmult);
		fprintf(fp,
 "Minimum # of pixels above threshold per fiducial point: %5d\n",
			(int)(cap->car_minlim));
		if (DFUZZ < cap->car_overcut) fprintf(fp,
		 "Absolute minimum count = %6.1f\n", cap->car_overcut);
		if (cap->car_imflags & DARK_CURRENT) fprintf(fp,
 "%6.4f of dark-current values will be subtracted out\n",
			((double)(cap->car_idarkfrac)) / 8000.);
	      }
	if (cap->car_imposespacing > DFUZZ) fprintf(fp,
 "Spacing between fiducial points assumed to be ~ %6.1f pixels near center\n",
		cap->car_imposespacing);
}

static double getsigmult(CALIBRUN *cap)
{ /*	This returns a sigma-over-the-local mean cutoff value for intensity.
	If the image is cruddy or has a faulty eventcount value,
	we set the sigma cutoff as 4. Otherwise we set it to
	4 * sqrt(mean count)-- unless it's a FAST image for which we set
	it to 0.5 * sqrt(mean count). */
	int	bformat, ims;
	double	mcount;
	double	revent;
	MAIL	*lhd;

	lhd = cap->car_heads;
	if ((NULL == lhd)  || (0 == lhd->eventcount)) return SIGMULT;
	revent = lhd->eventcount;	if (revent < 0.) revent *= -64.;
	bformat = lhd->format | 01;	/* set dimensions, just in case ... */
	cap->car_imwid = lhd->xdimen;	cap->car_imht = lhd->ydimen;
	ims = lhd->xdimen * lhd->ydimen;	if (ims < 1) return SIGMULT;
	mcount = SIGMULT * sqrt(revent / (double)ims);
	if (((bformat == FMT_FAST) || (bformat == FMT_FASTALT)) &&
		ims == 262144) mcount *= 0.125;
	return (mcount < MAX_MCOUNT ? mcount : MAX_MCOUNT);
}

static int header_linear(CALIBRUN *cap, double *locrxp)
{ /*	This returns the number of lines across the detector face
	in its return value and the raster size in x in *locrxp.
	It bases these calculations on the contents of a detector image
	header. We calculate nintervals ~ sqrt(width),
	and set nlines = nintervals +1, raster = (width+10)/nintervals.
	There are some special cases built in.
	If it fails it returns -1. */
	int			effsize, fra, nintervals;
	char			locdet[81], *locn;
	double			locpixel;
	MAIL			*lhd;
	VSPLINE			*lsp;
	DATARUN			*thisdp;
	static FRAMEINDEX	*lfmi;
	static int 		ispecial[] =
 {  522, 1034, 1210, 1610, 1910, 2010, 2058, 3010, 3082, 3210, 6154, 0 };
	static int 		nspecial[] =
 {   24,   47,   44,   46,   40,   68,   49,   43,   46,   60,   68, 0 };

	*locrxp = 0.254; /* nominal starting raster value in cm */
	strcpy(locdet, trname("CALDATA"));	locn = NULL;
	if (0 != xgaccess(locdet, 4, 1))
	  {
		if (NULL == (lfmi = ftables(3))) return -1;
		for (fra = lfmi->fi_firstfm; fra <= lfmi->fi_lastfm; fra++)
			if (NULL != (locn = indexname(fra)))
			  {
				strcpy(locdet, locn);
				if (0 == xgaccess(locdet, 4, 1)) break;
			  }
		if (NULL == locn) return -1;
		(void)ftables(0);
	   }
	lhd = cap->car_heads;
	if (FRAME_OK != read_image(locdet, cap->car_imflags, lhd)) return -1;
	/* we override the current width and height values if they're
		inconsistent with the ones found in the detector image. */
	if ((lhd->xdimen != (xgulong)(cap->car_imwid)) ||
	 (lhd->ydimen != (xgulong)(cap->car_imht)))
	  {
		cap->car_imwid = lhd->xdimen;	cap->car_imht = lhd->ydimen;
	  }
	if (NULL == cap->car_spl)
	  {
		if (NULL == (cap->car_spl = gspline(1, &fra))) return -1;
	  }
	zeromem((AVOID *)cap->car_spl, sizeof (VSPLINE)); lsp = cap->car_spl;
	lsp->v_deth = cap->car_imwid;	lsp->v_detw = cap->car_imht;
	/* currently it'll be considered an error if the user has specified
	 a detector dimension that is inconsistent with the header */
	if (((cap->car_specwid > 0) && (cap->car_specwid != cap->car_imwid)) ||
	 ((cap->car_specht > 0) && (cap->car_specht != cap->car_imht)))
	  {
		fprintf(stderr,
 "Specified dimensions [%4d,%4d] disagree with image dimensions [%4d,%4d]\n",
		 cap->car_specwid, cap->car_specht, cap->car_imwid,
		 cap->car_imht);	return -1;
	  }
	if (NULL != (thisdp = lhd->runheader))
		locpixel = thisdp->run_d1pos[1] * 0.1;
	else	locpixel = ((double)lhd->cellsize) / 10000.;
	effsize  = lhd->xdimen + 10; nintervals = -1;
	for (fra = 0; (ispecial[fra] != 0); fra++)
	   {
		if (effsize == ispecial[fra])
		  {
			nintervals = nspecial[fra]; break;
		  }
	  }
	if (nintervals < 0) nintervals = sqrt((double)effsize);
	if (nintervals > 68) nintervals = 50;
	*locrxp = locpixel * ((double)effsize) / (double)nintervals;
	return nintervals + 1;
}

static int calib_defaults(CALIBRUN *cap)
{ /*	This establishes detector-type-specific defaults */
	char		*capr;
	int		detype, lincol, dedim[2];
	double		xspa, yspa;
	static int	idarkfrac = 6400;

	zeromem((void *)cap, sizeof (CALIBRUN)); /* start structure blank */
	cap->car_tl0 = NOW;
	cap->car_nmbox = 1;	/* only one image header required */
	cap->car_idarkfrac = idarkfrac;
	if (NULL == (AVOID *)(cap->car_heads = (MAIL *)callop(cap->car_nmbox,
	 sizeof (MAIL), "heads in calib_defaults"))) return EALLOC;
	if (NULL == (cap->car_fpl = propen("xlog"))) cap->car_fpl = stdout;
	if ((NULL == (capr = trname("caldata"))) || ('\0' == *capr))
		return EBADVALUE;
	if (NULL == (cap->car_cdname = (char *)mallop(strlen(capr) + 2,
	 "cap->car_cdname in calib_defaults"))) return EALLOC;
	strcpy(cap->car_cdname, capr);	/*keep brass-plate image name*/
	cap->car_usedrx = cap->car_usedry = DEFRASTER;
	cap->car_minlim = DEFMINLIM;
	cap->car_flags = cap->car_linear = cap->car_overcut = 0;
	cap->car_imflags = GET_RUNHEADER;
	cap->car_imposespacing = cap->car_specwid = cap->car_specht = -1;
	if (0 >= (detype = dettype(dedim))) return EREADERR;
	/* now that we know the detector type, we determine
	 the pixel spacing and the number of linear columns from it.
	 Detectors for which we expect an actual brass-plate image
	 will return a zero in `lincol'. */
	get_calibparams(detype, &xspa, &yspa, &lincol);
	cap->car_linear = lincol;
	cap->car_usedrx = xspa;	cap->car_usedry = yspa;
	cap->car_beamx = dedim[0] / 2;	cap->car_beamy = dedim[1] / 2;
	return EOK;
}

static int calibargs(int argc, char *argv[], CALIBRUN *cap)
{ /*	This decodes the flags and command-line args, if any, and,
	if appropriate, opens the CALDATA image and reads it in.*/
	int		i, idarkfrac;
	char		*capr, **dcp, *dimname, **flags;
	double		locrx;
	MAIL		*lhd;
	VSPLINE		*lsp;
	DATARUN		*drp;
	static char	*amname;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	if (EOK != (i = calib_defaults(cap))) return i;
	flags = flagproc(argc, argv, "ABCDEFGHIJKLMNOPRSVWXYZ"); /*flag args*/
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	   {
		capr = (*dcp) + 1;
		if (**dcp == 'A') cap->car_sigmult = atof(capr);
		else if (**dcp == 'B') cap->car_flags |= CF_USEBORDER;
		else if (**dcp == 'C') cap->car_flags |= CF_READCEN;
		else if (**dcp == 'D') cap->car_specht = atoi(capr);
		else if (**dcp == 'E') cap->car_flags &= ~CF_USEBORDER;
		else if (**dcp == 'F') cap->car_overcut = atof(capr);
		else if (**dcp == 'G') cap->car_linear = 0; /* force brass */
		else if (**dcp == 'H') cap->car_flags |= CF_HEXAGGRID;
		else if (**dcp == 'I') cap->car_flags |= CF_INDEXED;
		else if (**dcp == 'J') cap->car_linear = -1;
		else if (**dcp == 'K') cap->car_imposespacing = atof(capr);
		else if (**dcp == 'L')	/* linear mapping */
		  {
			if ('\0' != *capr) cap->car_linear = atoi(capr);
			else cap->car_linear = 19;
		  }
		else if (**dcp == 'M') cap->car_minlim = atoi(capr);
		else if (**dcp == 'N') cap->car_flags |= CF_NOEXTRAP;
		else if (**dcp == 'O') cap->car_flags |= CF_WRITECEN;
		else if (**dcp == 'P') cap->car_flags &= ~CF_NOEXTRAP;
		else if (**dcp == 'R')
			cap->car_usedrx = cap->car_usedry = atof(capr);
		else if (**dcp == 'S')
		  {
			cap->car_imflags |= DARK_CURRENT;
			if ('\0' != *capr)
				cap->car_idarkfrac = 8000. * atof(capr);
		  }
		else if (**dcp == 'V') cap->car_flags |= CF_VERBOSE;
		else if (**dcp == 'W') cap->car_specwid = atoi(capr);
		else if (**dcp == 'X') cap->car_usedrx = atof(capr);
		else if (**dcp == 'Y') cap->car_usedry = atof(capr);
		else if (**dcp == 'Z') cap->car_flags &= ~CF_VERBOSE;
	   }
	idarkfrac = cap->car_idarkfrac;
	if (cap->car_specwid > -1) printf("specified width = %d\n",
	 cap->car_specwid);
	if (cap->car_specht > -1)  printf("specified height= %d\n",
	 cap->car_specht);
	/* the `indexed' flag is meaningless unless either
	 CF_READCEN or CF_WRITECEN is turned on */
	if (!(cap->car_flags & (CF_READCEN | CF_WRITECEN)))
		cap->car_flags &= ~(CF_INDEXED);
	if (cap->car_idarkfrac > 16000) cap->car_idarkfrac = 16000;
	idarkfrac = cap->car_idarkfrac;
	if (cap->car_imflags & DARK_CURRENT)
		cap->car_imflags |= (idarkfrac << 16);
	/* we determine the detector dimensions from an ordinary frame if we're
	doing a linear calibration.  If we're reading a CALDATA image, we use
	it to tell us what the dimensions are.  If we're reading a CALCENTROIDS
	file, we (a) specify a detector dimension, (b) use CALDATA, or (c) use
	or an ordinary frame, depending on what is out there and available */
	lsp = cap->car_spl;	lhd = cap->car_heads;	dimname = NULL;
	if ((cap->car_linear) || (cap->car_flags & CF_READCEN))
	  {
		if ((cap->car_specwid > 0) && (cap->car_specht > 0))
		  {
			cap->car_imwid = cap->car_specwid;
			cap->car_imht = cap->car_specht;
			/* memory for lsp is required in this case but not the
			other case, since find_dimension does it for us there*/
			if (NULL == (lsp = gspline(1, &i))) return EALLOC;
			lsp->v_detw = cap->car_imwid;
			lsp->v_deth = cap->car_imht;
			lsp->v_maxx = lsp->v_detw + 10;
			lsp->v_maxy = lsp->v_deth + 10;
		  }
		else
		  {
			if (cap->car_linear) dimname = NULL;
			else if (cap->car_flags & CF_READCEN) dimname =
			 (0 == xgaccess(cap->car_cdname, 4, 1)) ?
			 cap->car_cdname : NULL;
			if (EOK != (i = find_dimension(&cap->car_imwid,
			 &cap->car_imht, 1, dimname))) return i;
		  }
		/* linear < 0 is a flag meaning calculate
		 the raster size and the number of lines from
		 information in the image header */
		if (cap->car_linear < 0)
		  {
			cap->car_linear = header_linear(cap, &locrx);
			cap->car_usedrx = locrx;
			cap->car_usedry = locrx * ((double)cap->car_imht) /
						(double)cap->car_imwid;
		  }
		/* if it's still negative, we failed. */
		if (cap->car_linear < 0)
		  {
			globalkill();	cap->car_fpl = NULL; return EBADVALUE;
		  }
		if (cap->car_linear) cap->car_maxrows =
		 ((double)cap->car_linear) * ((double)cap->car_imht) /
			((double)cap->car_imwid) + 0.49999;
	  }
	else /* read the image from disk if needed. */
	  {
		if (FRAME_OK != (i = read_image(cap->car_cdname,
		 cap->car_imflags,lhd)))
		  {
			(void)image_error(stderr, cap->car_cdname, i, 1); 
			if (stdout != cap->car_fpl) (void)fclose(cap->car_fpl);
			globalkill();	cap->car_fpl = NULL; return EREADERR;
		  }
		cap->car_imwid = lhd->xdimen;	cap->car_imht = lhd->ydimen;
		/* memory for lsp is required in this case but not the
		other case, since find_dimension does it for us there */
		if (NULL == (lsp = gspline(1, &i))) return EALLOC;
		lsp->v_detw = cap->car_imwid;	lsp->v_deth = cap->car_imht;
		cap->car_spl = lsp;
		/* if the user didn't specify a sigma cutoff, calculate it*/
		if (cap->car_sigmult < 1.) cap->car_sigmult = getsigmult(cap);
	  }
	/* there are some weird circumstances where we will have lost lsp
	 at this point */
	if (NULL == (lsp = cap->car_spl)) lsp = cap->car_spl = gspline(1, &i);
	lsp->v_detw = cap->car_imwid;	lsp->v_deth = cap->car_imht;
	cap->car_dhwidth = cap->car_imwid / 2;
	cap->car_dhheight = cap->car_imht / 2;
	cap->car_scalxy = 0.5 * (double)(cap->car_dhwidth+ cap->car_dhheight);
	/* if the user asks to use the amask file and there isn't one there,
	we ignore the request */
	amname = trname("AMASK");
	if (cap->car_flags & CF_USEBORDER)
	  {
		if (NULL == activmask((int)(lsp->v_detw),(int)(lsp->v_deth),5))
		  {
			fprintf(stderr,
	 "AMASK file %s missing or faulty; no border checking done\n", amname);
			cap->car_flags &= ~CF_USEBORDER;
		  }
	  }
	/* skip past the non-numeric arguments */
	for (argc--, argv++; (argc) && !allnumeric(*argv); argc--) argv++;
	if (argc > 1)
	  {
		cap->car_beamx = atof(argv[0]);	cap->car_beamy = atof(argv[1]);
	  }
	else if ((NULL != (drp = lhd->runheader)) &&
	 ((fabs(drp->run_d0pos[1]) > FUZZ) || (fabs(drp->run_d0pos[2]) >FUZZ)))
	  {
		cap->car_beamx = cap->car_imwid / 2 + drp->run_d0pos[1];
		cap->car_beamy = cap->car_imwid / 2 + drp->run_d0pos[2];
	  }
	else {
		cap->car_beamx = cap->car_imwid / 2;
		cap->car_beamy = cap->car_imht / 2;
	     }
	lsp->v_axc = lsp->v_rxc = cap->car_beamx;
	lsp->v_ayc = lsp->v_ryc = cap->car_beamy;
	memzap((AVOID **)&(cap->car_cdname), "cap->car_cdname in calibargs");
	return EOK;
}

int recenter(VSPLINE *lsp, double bex, double bey)
{ /*	This redefines the centering of the splining based on the
	user-supplied coordinates of the direct beam */
	register int	ix, iy;
	int		quadr;
	float		obex, obey, xscen, yscen;
	CALPOINT	*calp;

	lsp->v_axc = bex;	lsp->v_ayc = bey;
	if (0 == pix2cm((float)bex, (float)bey, &xscen, &yscen, 0))
	  {
		fprintf(stderr, "Warning: cannot locate beam center yet.\n");
		xscen = yscen = 0.;
	  }
	if ((fabs(xscen) > 60.) || (fabs(yscen) > 60.))
	  {
		fprintf(stderr, "Nonsense beam center value %13.5e, %13.5e\n",
			xscen, yscen);
		xscen = yscen = 0.;
	  }
	printf("calculated beam center is %9.3f %9.3f\n", xscen, yscen);
	if (xscen != 0. || yscen != 0.)	/* modify center based on these data */
	  {
		lsp->v_x0cm -= xscen;	lsp->v_y0cm -= yscen;
		for (iy = lsp->v_nrow, calp = lsp->v_cpt; iy--; )
		   for (ix = lsp->v_ncol; ix--; calp++)
		      {
			calp->c_ctopx -= xscen;	calp->c_ctopy -= yscen;
		      }
		quadr = (bey >= (lsp->v_deth / 2)) +
			2 * (bex >= lsp->v_detw / 2);
		if (-1 == cm2pixq(0., 0., quadr, &obex, &obey, 0))
		  {
			lsp->v_rxc = bex;	lsp->v_ryc = bey;
		  }
		else {
			lsp->v_rxc = obex;	lsp->v_ryc = obey;
		     }
		if ((fabs(lsp->v_rxc) < lsp->v_detw / 4.) ||
		 (fabs(lsp->v_rxc) < lsp->v_deth / 4.))
		  {
			lsp->v_rxc = bex;	lsp->v_ryc = bey;
		  }
	  }
	return EOK;
}

int outindcen(VSPLINE *lsp)
{ /*	This writes indexed centroids to a file rather than doing a USPLINE */
	int		ix, iy;
	CALPOINT	*cap;
	static FILE	*fpet;

	if (qerror((AVOID *)(fpet = xgopen("CALCENTROIDS", "w")),
	 "open output calibration-centroids file")) return EOPENERR;
	for (cap = lsp->v_cpt, iy = 0; iy < lsp->v_nrow; iy++)
	   for (ix = 0; ix < lsp->v_ncol; ix++, cap++)
	      {
		fprintf(fpet, " %7.2f %7.2f %3d %3d\n",
			cap->c_ptocx, cap->c_ptocy, ix, iy);
	      }
	(void)fclose(fpet);	fpet = NULL;	return EOK;
}

int calibmain(int argc, char *argv[])
{ /*	This sets up the USPLINE file by reading in the calibration
	data frame `caldata', indexing and sorting it,
	and inverting it with simple interpolations. */
	int		i, nind, read_result;
	CALIBRUN	cas;
	VSPLINE		*lsp;

	/* get arguments and beam (x,y) position */
	if (EOK != (i = calibargs(argc, argv, &cas))) return i;
		/* summarize run params */
	sum_calset(cas.car_fpl, &cas);	sum_calset(stdout, &cas);
	if (NULL == (AVOID *)(cas.car_ctr0 = (CTROID *)callop(MAXSPOTS,
	 sizeof (CTROID), "ctr0 in calibmain")))
	  {
		if (stdout != cas.car_fpl) (void)fclose(cas.car_fpl);
		return EALLOC;
	  }
	cas.car_ctend = cas.car_ctr0 + MAXSPOTS;
	lsp = cas.car_spl;
	if ((!(cas.car_linear)) && (cas.car_flags & CF_WRITECEN))
	  { /* find centroids and index them */
		if (EOK != calibcen(&cas)) return EBAD;
		if (cas.car_flags & CF_INDEXED) /* index them before output */
		  {
			calindex(&cas);	/* index the located spots */
			calfill(&cas);	/* expand the indexing */
				/* write the indexed centroids to a file */
			i = outindcen(cas.car_spl);
			if (cas.car_fpl != stdout)
			  {
				(void)fclose(cas.car_fpl); cas.car_fpl = NULL;
			  }
			return i;
		  }
		return outcen(&cas);	/* write the centroids to a file */
	  }
	else if (cas.car_linear)
	  {
		lsp->v_xrcm = cas.car_usedrx; lsp->v_yrcm = -(cas.car_usedry);
		lin_troids(cas.car_linear, &cas);
	  }
	else if (cas.car_flags & CF_READCEN)
	  {
		(void)ext_troids(&nind, &cas);	calindex(&cas);
	  }
	else {
		if (EOK != calibcen(&cas)) return EBAD;
		calindex(&cas);	/* index the located spots */
	     }
		/* extrapolate to the indexings not found */
	if (EOK != (read_result = calfill(&cas))) i = read_result;
		/* invert mapping */
	else if (EOK != calinv(cas.car_fpl, cas.car_spl, 2)) i = EALLOC;
		/* find center of pattern*/
	else if (EOK != recenter(cas.car_spl, cas.car_beamx, cas.car_beamy))
		i = EBADVALUE;
	if ((EOK == i) && (cas.car_flags & CF_WRITECEN))
	  {
		i = outindcen(cas.car_spl);
		if (cas.car_fpl != stdout)
		  {
			(void)fclose(cas.car_fpl);
			cas.car_fpl = NULL;
		  }
		return i;
	  }
	else if (EOK == i) cas.car_spl = gspline(9, &i); /* output VSPLINE */
	if (stdout != cas.car_fpl) (void)fclose(cas.car_fpl);
	memzap((AVOID **)&(cas.car_ctr0), "ctr0 in calibmain");
	kill_mail(&(cas.car_heads), &(cas.car_nmbox));
	globalkill();	return i;
}

/* end of calibmain.c */
