/* calindex.c -- routines for indexing spots on the fiducial image

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	27 May 2007: MAXDLINE increased from 0.2 to 0.32
	19 Apr 2007: float->double conversions; reorganize use of "{"
	 7 Jun 2006: housekeeping
	12 May 2006: xcmp moved to calspots because it doesn't get used here
	 3 Jan 2006: don't qsort with fewer than 2 elements
	20 Nov 2005: inline function declarations & function prototypes
	11 Oct 2005: housekeeping
	22 Jun 2005: ycmp(), uxcmp(), uycmp() declared as static
	28 Aug 2004: routines used only here are declared as static; cosmetics
	21 Jan 2004: license message inserted and initial comment reordered
	 6 Jan 2004: CALIBRUN structure elements in function calls
		& static globals replace extern globals
	20 Nov 2003: replace global variables with elements of CALIBRUN struct
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	 5 Jan 2003: reorganized source-file header
	20 Sep 2002: CF_* flags moved to ../include/xgen.h
	25 Mar 2002: removed MSI attachments
	15 Apr 2000: user can specify starting spacing value in pixels
	 2 Mar 1995: diagnostics to stdout
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 5 Oct 1994: remove unimportant diagnostics in indx_x, indx_y
	23 Sep 1994: add fit to quadratic during initial indexing of spots
	 2 Sep 1994: revised indexing scheme when spots are closely spaced
	 5 May 1994: source file headers, POLYDET->XGEN,
		calculate of # of flat rows & columns redone
	16 Mar 1994: major rewrite of `calspots'; bugfixes elsewhere
	 7 Dec 1993: initial X-GEN version
	
   Purpose: finding a pair of integer indices for each of the fiducial-hole
	spots found in `calspots'. The indexings work outward from the
	center along lines and arcs.
   Cross references: calibmain.c, calspots.c, calfill.c

*************************************************************************/
#include	"calibdefs.h"
#define		MIN_INDEXED	3	/* min # indexed to continue row */
#define		SPACING_FACTOR	(280.)	/* used to calc. nominal spacing */
#define		SPACING_RATIO	0.7	/* ratio of one spacing to next */
#define		OFFSET_FACTOR	(0.88)	/*14/16, more or less; see its use*/
#define		MAX_EXCURSION	(2.)	/* hexag. max excursion betw. points*/
#define		FRAC_TRUST	(0.06)	/* don't trust index if <6% of edge */
#define		FRAC_CEN	(0.225)	/* within here we're "in the middle"*/
#define		NEAR_INTEGER	(0.25)	/* disbelieve non-integer intervals*/
#define		MAX_OKINTERVAL	4	/* maximum extrap. we believe */

static int ycmp(CTROID *ct0, CTROID *ct1)
{ /*	Comparison routine used in sorting on ycen */
	return (ct0->cycen < ct1->cycen) ? -1 : 1;
}

static int uxcmp(CTROID *ct0, CTROID *ct1)
{ /*	This compare routine allows for sorting the centroids
	primarily on the X index and secondarily on the Y value. */
	int	idiff;

	if ((idiff = ct0->ux - ct1->ux)) return (idiff < 0 ? -1 : 1);
	return (ct0->cycen < ct1->cycen ? -1 : 1);
}

static int uycmp(CTROID *ct0, CTROID *ct1)
{ /* This compare routine allows for sorting the centroids
	primarily on the Y index and secondarily on the X value. */
	int	idiff;

	if ((idiff = ct0->uy - ct1->uy)) return idiff;
	return (ct0->cxcen < ct1->cxcen ? -1 : 1);
}

void list_troids(FILE *fp, CTROID *ctb, unsigned ntr, char *str)
{ /* 	This writes a summary of the list of centroids to the stream fp.
	The string `str' is included in the head of the list */
	unsigned	i;
	CTROID		*ctp;

	fprintf(fp, "\tList of observed centroids %s\n",
		(str == NULL || '\0' == *str) ? "" : str);
	for (i = 0; i < 7; i++) fprintf(fp, "  Xobs  Yobs ix iy");
	fprintf(fp, "\n");
	for (i = 0, ctp = ctb; i < ntr; i++, ctp++)
	   {
		fprintf(fp, " %5.1f %5.1f %2d %2d",
			ctp->cxcen, ctp->cycen, ctp->ux, ctp->uy);
		if (!((i+1) % 7)) fprintf(fp, "\n");
	   }
	fprintf(fp, "\n");
}

static void colhdr(FILE *fp, double spacing)
{ /* prints a header to stream fp for the column-indexing information */
	fprintf(fp,
 "Axis Initial In- # of  Spa- In- # of  Spa- In- # of  Spa- In- # of  Spa-\n");
	fprintf(fp,
 "type Spacing dex  pts  cing dex  pts  cing dex  pts  cing dex  pts  cing\n");
	fprintf(fp, " col %7.1lf", spacing);
}

static int fakepoint(CTROID *c0p, CTROID *c1p, CALIBRUN *cap)
{ /*	This creates a new CTROID point at the end of the list.
	It will lie midway between the two points c0p, c1p that are named.
	It refuses to create the point unless either c0p->cxcen is very
	close to c1p->cxcen or c0p->cycen is very close to c1p->cycen. It also
	requires that the other distance be reasonably small.  Finally, if
	both points whose mean is being examined are fully indexed, i.e.
	c0p->ux >= 0, c0p->uy >= 0, c1p->ux >= 0, and c1p->uy >= 0, we require
	that either (c0p->ux == c1p->ux and |c0p->uy - c1p->uy| == 2)
	OR (c0p->uy == c1p->uy and |c1p->ux - c0p->ux| == 2). */
	int	dux, duy;
	double	delg, desm, dex, dey;

	dex = fabs(c0p->cxcen - c1p->cxcen);
	dey = fabs(c0p->cycen - c1p->cycen);
	if (dey < dex)
	  {
		desm = dey;	delg = dex;
	  }
	else {
		desm = dex;	delg = dey;
	     }
	if (desm > MAX_EXCURSION) return 0;
	if (((delg == dey) && (delg > cap->car_mincordist * 3.)) ||
	 ((delg == dex) && (delg > cap->car_mincordist * 6.))) return 0;
	dux = duy = 0;
	if (c0p->ux >= 0 && c1p->ux >= 0) dux = c0p->ux - c1p->ux;
	if (c0p->uy >= 0 && c1p->uy >= 0) duy = c0p->uy - c1p->uy;
	if (dux != 0 && duy != 0) return 0;
	if ((dux == 0) && (duy != 0) && (duy != -2) && (duy != 2))
		return 0;
	if ((duy == 0) && (dux != 0) && (dux != -2) && (dux != 2))
		return 0;
	/* having passed all these tests, we can create the new point */
	(cap->car_ctend)->cxcen = 0.5 * (c0p->cxcen + c1p->cxcen);
	(cap->car_ctend)->cycen = 0.5 * (c0p->cycen + c1p->cycen);
	/* we set the intensity to minus the mean.  This allows us to
	 distinguish between observed and interpolated points */
	(cap->car_ctend)->cinten = -0.5 * (c0p->cinten + c1p->cinten);
	/* find the indices of the new point.  If we can, we also get
	 an indexing for the point beyond the new one. */
	if (c0p->ux >= 0)
	  {
		if (c1p->ux >= 0)
			(cap->car_ctend)->ux = (c0p->ux + c1p->ux) / 2;
		else if (c0p->cxcen > c1p->cxcen)
		  {
			if (c0p->ux > 0)
			  {
				(cap->car_ctend)->ux = c0p->ux - 1;
				if ((c1p->ux < 0) && (c0p->ux > 1))
					c1p->ux = c0p->ux - 2;
			  }
		  }
		else {
			if (c0p->ux < cap->car_maxcols - 1)
			  {
				(cap->car_ctend)->ux = c0p->ux + 1;
				if ((c1p->ux < 0) &&
				  (c0p->ux < cap->car_maxcols - 2))
					c1p->ux = c0p->ux + 2;
			  }
		     }
	  }
	else if (c1p->ux >= 0)
	  {
		if (c1p->cxcen > c0p->cxcen)
		  {
			if (c1p->ux > 0)
			  {
				(cap->car_ctend)->ux = c1p->ux - 1;
				if (c1p->ux > 1) c0p->ux = c1p->ux -2;
			  }
			else (cap->car_ctend)->ux = -2;
		  }
		else {
			if (c1p->ux < cap->car_maxcols - 1)
			  {
				(cap->car_ctend)->ux = c1p->ux + 1;
				if (c1p->ux < cap->car_maxcols - 2)
					c0p->ux = c1p->ux + 2;
			  }
			else (cap->car_ctend)->ux = -2;
		     }
	  }
	else (cap->car_ctend)->ux = -2;
	if (c0p->uy >= 0)
	  {
		if (c1p->uy >= 0)
			(cap->car_ctend)->uy = (c0p->uy + c1p->uy) / 2;
		else if (c0p->cycen > c1p->cycen)
		  {
			if (c0p->uy > 0)
			  {
				(cap->car_ctend)->uy = c0p->uy - 1;
				if ((c1p->uy < 0) && (c0p->uy > 1))
					c1p->uy = c0p->uy - 2;
			  }
			else	(cap->car_ctend)->uy = -2;
		  }
		else {
			if (c0p->uy < cap->car_maxrows - 1)
			  {
				(cap->car_ctend)->uy = c0p->uy + 1;
				if ((c1p->uy < 0) &&
				 ((cap->car_ctend)->uy < cap->car_maxrows - 1))
					c1p->uy = (cap->car_ctend)->uy + 1;
			  }
		     }
	  }
	else if (c1p->uy >= 0)
	  {
		if (c1p->cycen > c0p->cycen)
		  {
			if (c1p->uy > 0)
			  {
				(cap->car_ctend)->uy = c1p->uy - 1;
				if (c1p->uy > 1) c0p->uy = c1p->uy -2;
			  }
			else (cap->car_ctend)->uy = -2;
		  }
		else {
			if (c1p->uy < cap->car_maxrows - 1)
			  {
				(cap->car_ctend)->uy = c1p->uy + 1;
				if (c1p->uy < cap->car_maxrows - 2)
					c0p->uy = c1p->uy + 2;
			  }
			else (cap->car_ctend)->uy = -2;
		     }
	  }
	else (cap->car_ctend)->uy = -2;
	cap->car_ctend++;	return 1;
}

static void hexag_fakes(CTROID *ctrp, int direction, CALIBRUN *cap)
{ /*	This invents a set of points midway between each of the points
	that appear in the column of points we've just finished indexing. */
	double	closest, dely;
	CTROID	*c0p, *c2p, *cep, *clp, *cwp;

	/* figure out what the first & last points within the column are. */	
	if (direction) /* direction == 1: we're @ 1st point in column */
	  {
		c0p = ctrp;
		for (cep = c0p + 1;
		 (cep < cap->car_ctend && cep->ux ==c0p->ux); cep++) ;
		cep--;
	  }
	else {	/* direction == 0: we're at the last point in the column */
		cep = ctrp;
		for (c0p = ctrp - 1;
		 (c0p >= cap->car_ctr0) && (c0p->ux == ctrp->ux); c0p--) ;
		c0p++;
	     }
/* for each point in the column, we find the
 point closest to it but above it.  When we've found that, we create
 a fake point midway between 'em. */
	for (cwp = c0p; cwp <= cep; cwp++)
	   {
		closest = 4. * (double)(cap->car_imht);	clp = NULL;
		for (c2p = c0p; c2p <= cep; c2p++)
		   if (((dely = c2p->cycen - cwp->cycen) > DFUZZ) &&
		     (dely < closest))
			{
				closest = dely; clp = c2p;
			}
		/* the topmost point won't have found anything, naturally.
		 In that case clp will be NULL, so we won't do anything */
		if (NULL != clp) cap->car_ntroids += fakepoint(cwp, clp, cap);
	   }
}

static void showpt(FILE *fp, int wind, int wn, double spaci, int npri)
{ /* prints out a summary of one column's worth of indexing efforts */
	fprintf(fp,  " %3d %4d%6.1f", wind, wn, spaci);
	if (!(npri % 4)) fprintf(fp, "\n            ");
}

#define	MAXDLINE	0.32

static int vfy_u(int zind, int which, double spaci, double *sump,CALIBRUN *cap)
{ /*	This goes through the centroids list and looks at all the points ctr
	for which (ctr->ux ==zind if `which' ==0) or (ctr->uy ==zind if
	`which' ==1). Each point is included in summations leading to a
	calculation of a best-fit parabola of the form
	X = aY^2 + bY + c (which == 0)	or	Y = aX^2 + bX + c (which == 1)
	It then goes back through the list and computes
	q = |X - (aY^2 + bY + C)|	or	q = |Y - (aX^2 + bX + c)|
	for each of the points in question. Any point deviating seriously,
	i.e.  q > MAXDLINE * spacing , is un-indexed in X. The routine returns
	the number of points that still have (ctr->ux == zind) or
	(ctr->uy == zind), and pointer to the sum of the X or Y values in the
	column is returned in `sump'.*/
	int		ncy, nlost, nlost0, nrecover, nz;
	short		*ipz;
	double		a0, a1, a2, cv, det, maxdev, that, that2, thatsc;
	double		this, thissc, ue0, ue1, ue2, ue3, ue4, uf0, uf1, uf2;
	double		eans[3], ev[3], em[3][3], emi[3][3];
	CTROID		*ctf, *ctl, *ctp;

	if (zind < 0) return 0;
	ue0 = ue1 = ue2 = ue3 = ue4 = uf0 = uf1 = uf2 = 0.;
	ctf = cap->car_ctr0;	ctl = cap->car_ctend - 1;
	if (which)
	  {
		thissc = cap->car_imht;	 thatsc = cap->car_imwid;
	  }
	else {
		thissc = cap->car_imwid; thatsc = cap->car_imht;
	     }
	maxdev = MAXDLINE * spaci / thissc;
	for (ctp = ctf, nz = 0; ctp < cap->car_ctend; ctp++)
	   {
		ipz = which ? &(ctp->uy) : &(ctp->ux);
		if (*ipz == zind) /* does the point have the current index ? */
		  {
			/* is X the independent or the dependent variable? */
			if (which)
			  {
				this = ctp->cycen; that = ctp->cxcen;
			  }
			else {
				this = ctp->cxcen; that = ctp->cycen;
			     }
			this /= thissc;	that /= thatsc;
			that2 = that * that;	nz++;	ue0 += 1.;
			ue1 += that;		ue2 += that2;
			ue3 += that * that2;	ue4 += that2 * that2;
			uf0 += this; uf1 += this * that; uf2 += this * that2;
			ctl = ctp;	if (ctf == cap->car_ctr0) ctf = ctp;
		  }
	   }
	*sump = uf0 * thissc;	if (nz < 3) return nz;
	if (nz < 4)
	  {
		if (fabs(det = ue0 * ue2 - ue1 * ue1) < DFUZZ) return nz;
		a0 = (ue2 * ue3 - ue1 * ue4) / det;
		a1 = (ue0 * ue4 - ue1 * ue3) / det; a2 = 0.;
	  }
	else {
		em[0][0] = ue0;	em[0][1] = em[1][0] = ue1;
		em[0][2] = em[1][1] = em[2][0] = ue2;
		em[1][2] = em[2][1] = ue3;	em[2][2] = ue4;
		if (dinv3(em, emi)) return nz;
		ev[0] = uf0;	ev[1] = uf1;	ev[2] = uf2;
		eans[0] = eans[1] = eans[2] = 0.;
		dmm31(eans, emi, ev);
		a0 = eans[0];	a1 = eans[1];	a2 = eans[2];
	     }
	for (ctp = ctf, nlost = 0; ctp <= ctl; ctp++)
	   {
		ipz = which ? &(ctp->uy) : &(ctp->ux);
		if (*ipz == zind) /* does the point have the current index ? */
		  {
			/* is X the independent or the dependent variable? */
			if (which)
			  {
				this = ctp->cycen; that = ctp->cxcen;
			  }
			else {
				this = ctp->cxcen; that = ctp->cycen;
				
			  }
			this /= thissc;	that /= thatsc;
			that2 = that * that; cv = a0 + a1 * that + a2 * that2;
			if (fabs(cv - this) > maxdev)
			  {
				*ipz = -5; /* temp flag */
				ue0 -= 1.; ue1 -= that; ue2 -= that2;
				ue3 -= that * that2;	ue4 -= that2 * that2;
				uf0 -= this; uf1 -= this * that;
				uf2 -= this * that2;	nlost++;
				if (--nz < 1)
				  {
					for (ctp = ctf; ctp < ctl; ctp++)
					   { /* make temp flags permanent */
						ipz = which ?
						 &(ctp->uy) : &(ctp->ux);
						if (*ipz == -5) *ipz = -2;
					   }
					*sump = uf0 * thissc;
					return 0;
				  }
			  }
		  }
	   }
	if (!nlost) return nz;
	/* try to restore some of the points we just deleted. If we actually
	had one outlier that was screwing up the indexing of a few others,
	we can get 'em back here. */
	for (ncy = 0, nlost0 = nlost; ncy < nlost0; ncy++)
	   {
		nrecover = 0;	/* clear count of how many we fixed */
		if (nlost > 1)
		  {
		    if (nz < 4)
		     {
			if (fabs(det = ue0 * ue2 - ue1 * ue1) < DFUZZ)
				return nz;
			a0 = (ue2 * ue3 - ue1 * ue4) / det;
			a1 = (ue0 * ue4 - ue1 * ue3) / det; a2 = 0.;
		     }
		    else
		     {
			em[0][0] = ue0;	em[0][1] = em[1][0] = ue1;
			em[0][2] = em[1][1] = em[2][0] = ue2;
			em[1][2] = em[2][1] = ue3;	em[2][2] = ue4;
			if (dinv3(em, emi)) return 0;
			ev[0] = uf0;	ev[1] = uf1;	ev[2] = uf2;
			eans[0] = eans[1] = eans[2] = 0.;
			dmm31(eans, emi, ev);
			a0 = eans[0];	a1 = eans[1];	a2 = eans[2];
		     }
		   for (ctp = ctf, nrecover = 0; ctp <= ctl; ctp++)
		    {
			ipz = which ? &(ctp->uy) : &(ctp->ux);
			if (-5 == *ipz) /* is this a flagged point? */
			  { /* is X independent or dependent variable? */
				if (which)
				  {
					this = ctp->cycen; that = ctp->cxcen;
				  }
				else {
					this = ctp->cxcen; that = ctp->cycen;
				     }
				this /= thissc;	that /= thatsc;
				that2 = that * that;
				cv =  a0 + a1 * that + a2 * that2;
				if (fabs(cv - this) < maxdev)
				  {
					*ipz = zind;	/* un-flag it */
					ue0 += 1.; ue1 += that; ue2 += that2;
					ue3 += that * that2;
					ue4 += that2 * that2;
					uf0 += this; uf1 += this * that;
					uf2 += this * that2;		nz++;
					nrecover++;	nlost--;
				  }
			  }
		     }
		 }
		if (!nrecover) /* if we didn't improve things, quit */
		  { /* permanently mark the temp-marked points */
			for (ctp = ctf; ctp <= ctl; ctp++)
			   {
				ipz = which ? &(ctp->uy) : &(ctp->ux);
				if (-5 == *ipz) *ipz = -2;
			   }
			break;
		  }
	   }
	*sump = uf0 * thissc;
	return nz;
}

static int indx_x(CALIBRUN *cap, double ylimi[])
{ /* This routine determines the x index of all the points whose y
	value is within a moderate tolerance of the beam center position.
	The routine returns the number of columns so determined.
	The spots enter this routine sorted on X, and we start the indexing
	outward from the spot closest to X=width/2.  With large numbers
	of fiducial points we don't try to index all the way out to the
	end of the row or column; that way we aren't influenced as much
	by the warping close to the edges. */

	int		flat_cols, midct, newc, nprint, nthisind, prevx, shift;
	double		boundary, colcen, colsum, delcol, midcen, spacing;
	CTROID		*ccol0, *cmaxind, *cminind, *ctrmid, *ctrp, *lct;

	for (lct = cap->car_ctr0, colsum = nthisind = 0;
	 lct < cap->car_ctend; lct++)
	   if ((ylimi[4] < lct->cycen) && (lct->cycen < ylimi[5]))
		{
			colsum += lct->cxcen;	nthisind++;
		}
	if (nthisind) colcen = colsum / (double)nthisind;
	else colcen = cap->car_dhwidth;
	fprintf(stdout,
	 "Mean X for observed spots in center rows:    %5.0f\n", colcen);
	fprintf(cap->car_fpl,
	 "Mean X for observed spots in center rows:    %5.0f\n", colcen);
	for (lct = cap->car_ctr0;
	 lct < cap->car_ctend && lct->cxcen < colcen; ) lct++;
	/* go a little farther to make sure we get to a point whose y value
	 is reasonably close to the middle */
	while ((lct < cap->car_ctend) &&
		((lct->cycen < ylimi[0]) || (lct->cycen > ylimi[1]))) lct++;
	if (verror(cap->car_ctend - lct,
	 "find well-behaved points in 2nd half of list")) return 0;
	lct->ux = prevx = cap->car_cencol;
	ctrmid = lct;	midct = nthisind = 1;
	midcen = colcen = colsum = lct->cxcen;
	spacing = cap->car_startspacing;
	if (cap->car_flags & CF_VERBOSE)
	  {
		colhdr(stdout, spacing);	colhdr(cap->car_fpl, spacing);
	  }
	flat_cols = cap->car_maxcols / 8 + 1;
	if (flat_cols < 5) flat_cols = 5;
	cmaxind = cminind = ccol0 = cap->car_ctr = lct;
	for (nprint = 0, lct++; lct < cap->car_ctend; lct++)
	   {
		cap->car_prev = lct - 1;
		/* indexing very close to the edge in Y is untrustworthy: */
		while ((lct < cap->car_ctend-1) &&
		  ((lct->cycen < ylimi[0]) || (lct->cycen > ylimi[1]))) lct++;
		delcol = lct->cxcen - colcen;
		newc = delcol > spacing;
		/* if we are getting close to hitting a new column, and the
		 y value is actually very close to one we've already seen,
		 then it's time to bump to the next column */
		if ((!newc) && (delcol > spacing * 0.7))
		  {
			for (ctrp = cap->car_prev; ctrp >= ccol0; ctrp--)
			   if ((ctrp->ux == prevx) &&
			     fabs(ctrp->cycen - lct->cycen) < spacing * 0.3)
			     {
				newc = 1;	break;
			     }
		  }
		if (newc)
		  { /* we've found a new column to the right.  */
			boundary = colcen + spacing * 0.8;
			/* If the column we just completed
			is >= flat_cols past the center, we quit looking */
			if (prevx - cap->car_cencol >= flat_cols)
			  {
				for (ctrp = cap->car_prev; ctrp->cxcen >
				 boundary; ctrp--) ctrp->ux = -2;
				break;
			  }
			/* verify points in previous column*/
			nthisind = vfy_u(prevx, 0, spacing, &colsum, cap);
			if (nthisind > 0) colcen = colsum / (double)nthisind;
			/* new columns require more careful scrutiny, so
			 we skip points close to the top and bottom */
			while ((lct < cap->car_ctend) && ((lct->cycen <
			 ylimi[2]) || (lct->cycen > ylimi[3]))) lct++;
			if (lct >= cap->car_ctend) break;
			if (prevx == cap->car_cencol)
			  {
				midct = nthisind;	midcen = colcen; }
			/* with a hexagonal grid we now add in the fake
			 points in the column we just finished */
			if (cap->car_flags & CF_HEXAGGRID)
				hexag_fakes(cap->car_prev, 0, cap);
			lct->ux = prevx = prevx + 1;	ccol0 = lct;
			spacing = SPACING_RATIO * (lct->cxcen - colcen);
			if (cap->car_flags & CF_VERBOSE)
			  {
				nprint++;	showpt(cap->car_fpl,
				 prevx, nthisind, spacing,nprint);
				showpt(stdout,prevx, nthisind, spacing,nprint);
			  }
			nthisind = 1;	colcen = colsum = lct->cxcen;
		  }
		else if (++nthisind > cap->car_maxcols)
		  {
			fprintf(stdout,
			 " Too many in col %2d: skipping rest\n", prevx);
			fprintf(cap->car_fpl,
			 " Too many in col %2d: skipping rest\n", prevx);
			break;
		  }
		else
		  {
			lct->ux = prevx;	colsum += lct->cxcen;
			colcen = colsum / (double)nthisind;
		  }
	   }
	/* verify the points in the last column */
	nthisind = vfy_u(prevx, 0, spacing, &colsum, cap);
	if (cap->car_flags & CF_VERBOSE)
	  {
		fprintf(cap->car_fpl , "\n col %7.1f", spacing);
		fprintf(stdout, "\n col %7.1f", spacing);
	  }
	nthisind = midct;	prevx = cap->car_cencol;
	colcen = midcen;	colsum = midcen * (double)midct;
	spacing = cap->car_startspacing;	midcen += spacing;
	for (nprint = 0, ccol0 = lct = ctrmid - 1; lct >= cap->car_ctr0; lct--)
	   {
		cap->car_prev = lct + 1;
		while ((lct > cap->car_ctr0) && ((lct->cycen < ylimi[0]) ||
				(lct->cycen > ylimi[1]))) lct--;
		delcol = colcen - lct->cxcen;
		newc = delcol > spacing;
		/* if we are getting close to hitting a new column, and the
		 y value is actually very close to one we've already seen,
		 then it's time to bump to the next column */
		if ((!newc) && (delcol > spacing * 0.7))
		  {
			for (ctrp = cap->car_prev; ctrp <= ccol0; ctrp++)
			   if ((ctrp->ux == prevx) &&
			     fabs(ctrp->cycen - lct->cycen) < spacing * 0.3)
			     {
				newc = 1;	break;
			     }
		  }
		if (newc)
		  { /* we've found a new column to the left.  */
			boundary = colcen - spacing * 0.8;
			/* If we already have >= flat_cols up there,quit */
			if (cap->car_cencol - prevx >= flat_cols)
			  {
				for (ctrp = cap->car_prev; ctrp->cxcen <
				 boundary; ctrp++) ctrp->ux = -2;
				break;
			  }
			/* verify points in previous column*/
			nthisind = vfy_u(prevx, 0, spacing, &colsum, cap);
			if (nthisind > 0) colcen = colsum / (double)nthisind;
			while ((lct > cap->car_ctr0) && ((lct->cycen <
			 ylimi[2]) || (lct->cycen > ylimi[3]))) lct--;
			if (lct <= cap->car_ctr0) break;
			/* with a hexagonal grid we now add in the faked
			 points in the column we just finished */
			if (cap->car_flags & CF_HEXAGGRID)
				hexag_fakes(cap->car_prev,1,cap);
			lct->ux = prevx = prevx - 1;	ccol0 = lct;
			spacing = SPACING_RATIO * (colcen - lct->cxcen);
			if (cap->car_flags & CF_VERBOSE)
			  {
				nprint++;
				showpt(cap->car_fpl,
					prevx, nthisind,spacing,nprint);
				showpt(stdout,prevx, nthisind,spacing,nprint);
			  }
			nthisind = 1;	colcen = colsum = lct->cxcen;
		  }
		else if (++nthisind > cap->car_maxcols)
		  {
			fprintf(stdout,
			 " Too many in col %2d: skipping rest\n", prevx);
			fprintf(cap->car_fpl,
			 " Too many in col %2d: skipping rest\n", prevx);
			break;
		  }
		else	/* we're in the same column as the previous point */
		  {
			lct->ux = prevx;	colsum += lct->cxcen;
			colcen = colsum / (double)nthisind;
		  }
		cap->car_ctr = lct;
	   }
	nthisind = vfy_u(prevx, 0, spacing, &colsum, cap);
	if (cap->car_flags & CF_VERBOSE)
	  {
		fprintf(cap->car_fpl , "\n");	fprintf(stdout, "\n");
	  }
	/* remember the first and last successfully half-indexed point */
	cmaxind = cap->car_ctend - 1;
	/* for a hexagonal grid, there's a bunch of half-indexed points among
	 the interpolated values that need to be ignored for this purpose */
	if (cap->car_flags & CF_HEXAGGRID)
	  for ( ; cmaxind > cap->car_ctr0; cmaxind--)
		if (cmaxind->cinten > 0.) break;
	for ( ; cmaxind->ux < 0; ) cmaxind--;
	if (lct < cap->car_ctr0) lct = cap->car_ctr0;
	for (cminind = cap->car_ctr0; cminind->ux < 0; ) cminind++;
	if (verror(cminind < cmaxind, "index enough points to work with"))
		return 0;
	shift = cap->car_cencol - (cmaxind->ux + cminind->ux) / 2;
	if (shift) for (lct = cminind; lct <= cmaxind; lct++)
			if (lct->ux > -2) lct->ux += shift;
	cap->car_ctr = lct;
	return (cmaxind->ux - cminind->ux) + 1;
}

static int indx_y(int *nboth, CALIBRUN *cap, double xlimi[])
{ /*	This routine determines the y index of all the points on the
	centermost nine rows (roughly) of the array of fiducial spots.
	The routine returns the number of rows so determined.
	The spots enter this routine sorted on Y, and we start the
	indexing outward from the spot closest to Y=height/2.
	We also count the number of points successfully indexed both ways
	and report it */

	int		flat_rows, midct, newr, nprint, nthisind, prevy, shift;
	double		boundary, delrow, midcen, rowcen, rowsum, spacing;
	CTROID		*cmaxind, *cminind, *crow0, *ctrmid, *ctrp, *lct;

	for (lct = cap->car_ctr0, rowsum = nthisind = 0;
	 lct < cap->car_ctend; lct++)
	   if ((xlimi[4] < lct->cxcen) && (lct->cxcen < xlimi[5]))
		{
			rowsum += lct->cycen;	nthisind++;
		}
	rowcen = (nthisind) ?  rowsum / (double)(nthisind) : cap->car_dhheight;
	fprintf(stdout,
	 "Mean Y for observed spots in center columns: %5.0f\n", rowcen);
	fprintf(cap->car_fpl,
	 "Mean Y for observed spots in center columns: %5.0f\n", rowcen);
	for (lct = cap->car_ctr0; lct < cap->car_ctend && lct->cycen < rowcen; )
		lct++;
	/* go a little farther to make sure we get to a point whose x value
	 is reasonably close to the middle */
	while ((lct < cap->car_ctend) && ((lct->cxcen < xlimi[0]) ||
				(lct->cxcen > xlimi[1]))) lct++;
	if (verror(cap->car_ctend - lct,
		"find well-behaved points in 2nd half of list")) return 0;
	/* hexagonal image case:
	 if the starting point we're working with is a real (not interpolated)
	 point, then the sum of its indices should be even.  If it's an
	 extrapolated point, the sum of its indices should be odd */
	if ((cap->car_flags & CF_HEXAGGRID) && (lct->ux >= 0))
	  {
		if (lct->cinten > 0.)
		  {
			if ((cap->car_cenrow +lct->ux) %2) (cap->car_cenrow)++;
		  }
		else if (!((cap->car_cenrow +lct->ux) %2)) (cap->car_cenrow)++;
	  }
	lct->uy = prevy = cap->car_cenrow;	ctrmid = lct;
	midct = nthisind = 1;
	midcen = rowcen = rowsum = lct->cycen;
	spacing = cap->car_startspacing;
	if (cap->car_flags & CF_VERBOSE)
	  {
		fprintf(stdout, " row %7.1f", spacing);
		fprintf(cap->car_fpl , " row %7.1f", spacing);
	  }
	flat_rows = cap->car_maxrows / 8;
	if (flat_rows < 5) flat_rows = 5;
	cminind = cmaxind = crow0 = cap->car_ctr = lct;
	for (*nboth = nprint = 0, lct++; lct < cap->car_ctend; lct++)
	   {
		cap->car_prev = lct - 1;
		while ((lct < cap->car_ctend) && ((lct->cxcen < xlimi[0]) ||
			(lct->cxcen > xlimi[1]))) lct++;
		delrow = lct->cycen - rowcen;
		newr = delrow > spacing;
		/* if we are getting close to hitting a new row, and the
		 x value is actually very close to one we've already seen,
		 then it's time to bump to the next row */
		if ((!newr) && (delrow > spacing * 0.7))
		  {
			for (ctrp = cap->car_prev; ctrp >= crow0; ctrp--)
			   if ((ctrp->uy == prevy) &&
			     fabs(ctrp->cycen - lct->cycen) < spacing * 0.3)
			     {
				newr = 1;	break;
			     }
		  }
		if (newr)
		  { /* we've found a new row below the previous one.  */
			boundary = rowcen + spacing * 0.8;
			/* if the new row is >= flat_rows after center, quit*/
			if (prevy - cap->car_cenrow >= flat_rows)
			  {
				for (ctrp = cap->car_prev;
				 ctrp->cycen > boundary; ctrp--) ctrp->uy = -2;
				break;
			  }
			/* verify points in previous row */
			nthisind = vfy_u(prevy, 1, spacing, &rowsum, cap);
			if (nthisind > 0) rowcen = rowsum / (double)nthisind;
			while ((lct < cap->car_ctend) && ((lct->cxcen <
			 xlimi[2]) || (lct->cxcen > xlimi[3]))) lct++;
			if (lct >= cap->car_ctend) break;
			if (prevy == cap->car_cenrow)
			  {
				midct = nthisind;	midcen = rowcen;
			  }
			lct->uy = prevy = prevy + 1;	crow0 = lct;
			if ((cap->car_flags & CF_HEXAGGRID) && (lct->ux >= 0)
			 && ((lct->ux + lct->uy + (lct->cinten < 0.)) % 2))
			  {
				fprintf(stdout,
 "Disallowing assignment: [%5.1f,%5.1f], %s, ux=%2d; uy wants to be %d\n",
				 lct->cxcen, lct->cycen, (lct->cinten < 0. ?
 "observed" : "interpolated"), lct->ux, lct->uy);
				lct->uy = -2;
			  }
			spacing = SPACING_RATIO * (lct->cycen - rowcen);
			if (cap->car_flags & CF_VERBOSE)
			  {
				nprint++; showpt(cap->car_fpl,
					prevy, nthisind, spacing,nprint);
				showpt(stdout,prevy, nthisind, spacing,nprint);
			  }
			nthisind = 1;	rowcen = rowsum = lct->cycen;
		  }
		else if (++nthisind > cap->car_maxrows)
		  {
			fprintf(stdout,
			 " Too many in row %2d: skipping rest\n", prevy);
			fprintf(cap->car_fpl,
			 " Too many in row %2d: skipping rest\n", prevy);
			break;
		  }
		else
		  {	
			lct->uy = prevy;	rowsum += lct->cycen;
			rowcen = rowsum / (double)nthisind;
		  }
		if (lct->ux >= 0) (*nboth)++;
	   }
	/* verify points in previous row */
	nthisind = vfy_u(prevy, 1, spacing, &rowsum, cap);
	if (cap->car_flags & CF_VERBOSE)
	  {
		fprintf(cap->car_fpl , "\n row %7.1f", spacing);
		fprintf(stdout, "\n row %7.1f", spacing);
	  }
	if (lct >= cap->car_ctend) lct--;
	for (cmaxind = lct; cmaxind->uy < 0; ) cmaxind--;
	nthisind = midct;	prevy = cap->car_cenrow;
	rowcen = midcen;	rowsum = midcen * (double)midct;
	spacing = cap->car_startspacing;
	for (crow0 = lct = ctrmid - 1; lct >= cap->car_ctr0; lct--)
	   {
		cap->car_prev = lct + 1;
		while ((lct > cap->car_ctr0) && ((lct->cxcen < xlimi[0]) ||
			(lct->cxcen > xlimi[1]))) lct--;
		delrow = rowcen - lct->cycen;
		newr = delrow > spacing;
		/* if we are getting close to hitting a new column, and the
		 y value is actually very close to one we've already seen,
		 then it's time to bump to the next column */
		if ((!newr) && (delrow > spacing * 0.7))
		  {
			for (ctrp = cap->car_prev; ctrp <= crow0; ctrp++)
			   if ((ctrp->uy == prevy) &&
			     fabs(ctrp->cxcen - lct->cxcen) < spacing * 0.3)
			     {
				newr = 1;	break;
			     }
		  }
		if (newr)
		  { /* we've found a new row above the previous.  */
			boundary = rowcen - spacing * 0.8;
			if (cap->car_cenrow - prevy >= flat_rows)
			  {
				for (ctrp = cap->car_prev;
				 ctrp->cycen < boundary; ctrp++) ctrp->uy = -2;
				break;
			  }
			/* verify points in previous row */
			nthisind = vfy_u(prevy, 1, spacing, &rowsum, cap);
			if (nthisind > 0) rowcen = rowsum / (double)nthisind;
			while ((lct > cap->car_ctr0) && ((lct->cxcen <
			 xlimi[2]) || (lct->cxcen > xlimi[3]))) lct--;
			if (lct <= cap->car_ctr0) break;
			lct->uy = prevy = prevy - 1;	crow0 = lct;
			if ((cap->car_flags & CF_HEXAGGRID) && (lct->ux >= 0)
			 && ((lct->ux + lct->uy + (lct->cinten < 0.)) % 2))
			  {
				fprintf(stdout,
 "Disallowing assignment: [%5.1f,%5.1f], %s, ux=%2d; uy wants to be %d\n",
				 lct->cxcen, lct->cycen, (lct->cinten < 0. ?
 "observed" : "interpolated"), lct->ux, lct->uy);
				lct->uy = -2;
			  }
			spacing = SPACING_RATIO * (rowcen - lct->cycen);
			if (cap->car_flags & CF_VERBOSE)
			  {
				nprint++; showpt(cap->car_fpl,
					prevy, nthisind,spacing,nprint);
				showpt(stdout,prevy, nthisind,spacing,nprint);
			  }
			nthisind = 1;	rowcen = rowsum = lct->cycen;
		  }
		else if (++nthisind > cap->car_maxrows)
		  {
			fprintf(stdout," Too many in row %2d: skipping rest\n",
			 (cap->car_prev)->uy);
			fprintf(cap->car_fpl,
			 " Too many in row %2d: skipping rest\n",
			 (cap->car_prev)->uy);
			break;
		  }
		else
		  {
			lct->uy = (cap->car_prev)->uy;	rowsum += lct->cycen;
			rowcen = rowsum / (double)nthisind;
		  }
		if (lct->ux >= 0) (*nboth)++;
	   }
	nthisind = vfy_u(prevy, 1, spacing, &rowsum, cap);
	if (cap->car_flags & CF_VERBOSE)
	  {
		fprintf(cap->car_fpl, "\n");	fprintf(stdout, "\n");
	  }
	if (lct < cap->car_ctr0) lct = cap->car_ctr0;
	for (cminind = lct; cminind->uy < 0; ) cminind++;
	shift = cap->car_cenrow - (cmaxind->uy + cminind->uy) / 2;
	if (shift) for (lct = cminind; lct <= cmaxind; lct++)
			if (lct->uy > -2) lct->uy += shift;
	cap->car_ctr = lct;
	return (cmaxind->uy - cminind->uy) + 1;
}

static void skipnote(FILE *fp, char direction,
	CTROID *cp0, CTROID *cp1, int interv)
{ /*	This notes the fact that there's a non-unity interval between 2 pts */
	int	iz;

	iz = (direction == 'x' ? cp0->ux : cp0->uy);
	fprintf(fp,
 "Note: interval, %c = %2d, %2d between [%5.1f,%5.1f] and [%5.1f,%5.1f]\n",
	 direction, interv, iz, cp1->cxcen, cp1->cycen, cp0->cxcen,cp0->cycen);
}

static int get_meand(char which, double *sumdp, CALIBRUN *cap)
{ /*	This finds the mean value of delta Y or delta X between points
	that we've already double-indexed in the current row or column.
	It returns the number of intervals tested and returns the sum of
	the deltas between 'em in *sumdp.  It determines the first and
	last CTROIDs in the current row or column as well as the one in
	the middle of the row or column. */
	int	ninterv, prevu, prevv, thisu, thisv;
	double	delv;
	CTROID	*lct;

	*sumdp = ninterv = 0;	cap->car_prev = cap->car_g0;
	for (lct = cap->car_g0 + 1;
	 (cap->car_ctr0 <= lct && lct < cap->car_ctend); lct++)
	   {
		if (which == 'x')
		  {
			thisv = lct->uy;	thisu = lct->ux;
			prevv = (cap->car_prev)->uy;
			prevu = (cap->car_prev)->ux;
			delv = lct->cycen - (cap->car_prev)->cycen;
		  }
		else  {
			thisv = lct->ux;	thisu = lct->uy;
			prevv = (cap->car_prev)->ux;
			prevu = (cap->car_prev)->uy;
			delv = lct->cxcen - (cap->car_prev)->cxcen;
		      }
		if (thisu != prevu) break;
		if (thisv >= 0)
		  {
			cap->car_prev = lct;
			if ((prevv >= 0) && (prevv != thisv))
			  {
				*sumdp += delv / ((double)(thisv - prevv));
				ninterv++;
			  }
		  }
	   }
	cap->car_next = lct; thisu = lct - cap->car_g0;
	cap->car_mid = cap->car_g0 + (thisu / 2);	return ninterv;
}
	
static int fillone(int *ninterp, double *sumdp, CALIBRUN *cap)
{ /*	This establishes the second of two indexings for a half-indexed point.
	The new point is `cap->car_ctr'; the old one is `cap->car_prev'.
	This routine returns 0 if it succeeds in indexing `ctr',
	various nonzero integers if it fails.
	It also updates the values fed into it of `*ninterp', which is the
	number of intervals encountered in the current row or column,
	and `*sumdp', which is the sum of the distance intervals along the
	current row or column */
	char	which;
	int	absi, fakestat, interval, need_warning, newcode, other, trialv;
	double	avedel, rinterv, zdiff;
	CTROID	*lct;

	lct = cap->car_ctr;
	newcode = (2 * (lct->ux >= 0)) + (lct->uy >= 0);
	/* if the new point is fully indexed, or if it isn't indexed at
	 all, we don't (or can't) do anything */
	if (newcode <= 0) return 1;	if (newcode >= 3) return 2;
	which = (newcode == 2) ? 'x': 'y';
	/* define the "previous point"; this is an indexed point within
	 the same row or column that is close to our current one. We
	 look back into the list if we're past the halfway point;
	 we look forward if we're short of the halfway point. */
	if (lct >= cap->car_mid)
	  {
		for (cap->car_prev = lct - 1;
		 cap->car_prev >= cap->car_g0; (cap->car_prev)--)
		 if (((cap->car_prev)->ux >= 0) && ((cap->car_prev)->uy >= 0))
			break;
		if (cap->car_prev < cap->car_g0) return 3; /* no ref pt found*/
	  }
	else {
		for (cap->car_prev = lct + 1;
		 cap->car_prev < cap->car_next; cap->car_prev++)
		 if (((cap->car_prev)->ux >= 0) && ((cap->car_prev)->uy >= 0))
			break;
		if (cap->car_prev >= cap->car_next) return 4;
	     }
	avedel = fabs((*sumdp) / ((double)(*ninterp)));
	zdiff = (which == 'x') ? lct->cycen - (cap->car_prev)->cycen :
		lct->cxcen - (cap->car_prev)->cxcen;
	rinterv = zdiff / avedel;
	interval = (rinterv < 0.) ? rinterv - 0.49999 : rinterv + 0.49999;
	/* an interval of zero is a non-sequitur */
	if (!interval) return 5;
	absi = abs(interval);
	/* intervals greater than about 4 are untrustworthy */
	if (absi > MAX_OKINTERVAL) return 6;
	/* the interval must be close to an integer or we don't trust it */
	if (fabs(rinterv - (double)interval) > NEAR_INTEGER) return 7;
	trialv = (which =='x') ?
	 (cap->car_prev)->uy +interval : (cap->car_prev)->ux +interval;
	need_warning = 0;
	if (cap->car_flags & CF_HEXAGGRID)
	  { /* for hexagonal indexings we need to check the polarity */
		other = (which == 'x') ? lct->ux : lct->uy;
		if ((trialv + other + (lct->cinten < 0.)) % 2) return 8;
		if (absi == 2)
		  {
			if ((fakestat = fakepoint(lct, cap->car_prev, cap)))
				cap->car_ntroids += fakestat;
			else	need_warning = 1;
		  }
		else if (absi > 2) need_warning = 1;
	  }
	else need_warning = (absi > 1);
	if (need_warning)
	  {
		skipnote(cap->car_fpl , which, lct, cap->car_prev, interval);
		if (cap->car_flags & CF_VERBOSE)
			skipnote(stdout, which, lct, cap->car_prev, interval);
	  }
	if (which == 'x') lct->uy = trialv;	else lct->ux = trialv;
	(*ninterp)++;	*sumdp += fabs(zdiff) / (double)absi;
	return 0;
}

static int fill_half_index(CALIBRUN *cap)
{ /*	This takes spots which have one of their indices determined
	and determines the other one.  It returns the number of
	points thus successfully indexed.*/
	int	nadded, nintervals;
	double	sumdel;
	CTROID	*lct;

	/* sort the spots primarily on the X index and secondarily on
	the Y value.  The spots which haven't been indexed in X come
	at the beginning of the list and can be ignored. */
	if (cap->car_ntroids > 1) qsort((AVOID *)(cap->car_ctr0),
	 (size_t)(cap->car_ntroids), sizeof (CTROID), (QPE)uxcmp);
	for (lct = cap->car_ctr0; lct->ux < 0; ) lct++;
	if (lct >= cap->car_ctend)
	  {
		cap->car_ctr = cap->car_ctend;	return 0;
	  }
	/* for each value of ux we look through the Y values.  Somewhere
	within the sub-list for the current ux there will (usually) be a spot
	whose uy value is known.  We index on y backward from that point.*/
	for (nadded = 0; lct < cap->car_ctend; )
	   {
		cap->car_g0 = lct;
		/* find the mean distance between pairs of points in
		 this column.  We need to have found at least 3 pairs
		 in order to trust this calculation. */
		if (MIN_INDEXED-1 <= (nintervals =
		 get_meand('x', &sumdel, cap)))
		  { /* find the uy values for elements in the current
			 column for which we haven't got an indexing. We
			 start in the middle and work outward.   */
			for (lct = cap->car_mid; (cap->car_g0 <= lct &&
			 lct < cap->car_next); )
			   {
				if (!fillone(&nintervals, &sumdel, cap))
					nadded++;
				if (lct < cap->car_mid) lct =
					cap->car_mid + (cap->car_mid -lct);
				else lct = cap->car_mid -
					(lct + 1 - cap->car_mid);
			   }
		   }
		/* advance to the next column */
		lct = cap->car_next;
	   }
	/* sort the spots primarily on the Y index and secondarily on
	the X value.  The spots which haven't been indexed in Y come
	at the beginning of the list and can be ignored. */
	if (cap->car_ntroids > 1) qsort((AVOID *)(cap->car_ctr0),
	 (size_t)(cap->car_ntroids), sizeof (CTROID), (QPE)uycmp);
	for (lct = cap->car_ctr0; lct->uy < 0; ) lct++;
	/* for each value of uy we look through the X values.  Somewhere
	within the sub-list for the current uy there will (usually) be a spot
	whose ux value is known.  We index on x backward from that point.*/
	for ( ; lct < cap->car_ctend; )
	   {
		cap->car_g0 = lct;
		/* find the mean distance between pairs of points in
		 this column.  We need to have found at least 3 pairs
		 in order to trust this calculation. */
		if (MIN_INDEXED-1 <= (nintervals =
		 get_meand('y', &sumdel, cap)))
		  { /* find the ux values for elements in the current
			 row for which we haven't got an indexing. We
			 start in the middle and work outward.   */
			for (lct = cap->car_mid; (cap->car_g0 <= lct &&
			 lct < cap->car_next); )
			   {
				if (!fillone(&nintervals, &sumdel, cap))
					nadded++;
				if (lct < cap->car_mid) lct = cap->car_mid +
				 (cap->car_mid -lct);
				else lct = cap->car_mid -
					(lct + 1 - cap->car_mid);
			   }
		   }
		/* advance to the next column */
		lct = cap->car_next;
	   }
	return nadded;
}

void calindex(CALIBRUN *cap)
{ /*	This reads the centroids from the calibration run and assigns
	the observed points to (iy,ix) in the matrix.  It starts
	by sorting on X, merging points which are too close
	to each other, and then getting X indices for the points
	close to the center in X (it's easy here: the lines aren't bent).
	We also	need to decide whether there are really 19 columns in the
	input; if there aren't, the program decides whether to call the
	first column column 0 or column 1.  The list is then resorted
	on y and the y indices are assigned similarly.
	Then the points which have been indexed in X but not in Y
	are indexed in Y and vice versa. */
	int	badpoints, nadded, nboth, ncolumns, ninteven, nintodd;
	int	nrealeven, nrealodd, nrows, polarsum;
	double	spacemin, xlimi[6], ylimi[6];
	VSPLINE	*lsp;
	CTROID	*lct;

	lsp = cap->car_spl;
	cap->car_imwid = lsp->v_detw;	cap->car_imht = lsp->v_deth;
	cap->car_ctr = cap->car_ctend;
	for (lct = cap->car_ctr0; lct < cap->car_ctend; lct++)
		lct->ux = lct->uy = -2;
	/* start with nominal spacing between centroids inversely
	proportional to the sqrt of the number of centroids since the area
	occupied by a spot is inversely proportional to the number of spots.
	We now allow user to specify this start spacing externally */
	if (cap->car_imposespacing <= 0.)
	  {
		cap->car_startspacing  = (cap->car_ntroids > 1) ?
		 SPACING_FACTOR / sqrt((double)(cap->car_ntroids)) : 200;
		cap->car_startspacing *= ((double)(cap->car_imwid)) / 512.;
	  }
	else	cap->car_startspacing = cap->car_imposespacing;
	if (cap->car_flags & CF_VERBOSE) fprintf(cap->car_fpl,
	 "starting spacing, cenrow, cencol = %7.2f pixels, %d, %d\n",
	 cap->car_startspacing, (int)(cap->car_cenrow), (int)(cap->car_cencol));
	/* establish boundaries of trustworthy indexing in X and Y.
	 We make them more stringent for newly indexed rows or columns. */
	xlimi[0] = FRAC_TRUST * (double)(cap->car_imwid);
	spacemin = ((double)cap->car_dhwidth) - 20. * cap->car_startspacing;
	if (xlimi[0] < spacemin)
	  {
		xlimi[0] = spacemin;
		xlimi[2] = 0.5 * (((double)cap->car_dhwidth) + xlimi[0]);
	  }
	else xlimi[2] = 2 * xlimi[0];
	xlimi[4] = FRAC_CEN * (double)(cap->car_imwid);
	if (xlimi[4] < xlimi[2]) xlimi[4] = xlimi[2];
	if (xlimi[0] < cap->car_dhwidth - 320)
	  {
		xlimi[0] = cap->car_dhwidth - 320;
		xlimi[2] = cap->car_dhwidth - 160;
		xlimi[4] = cap->car_dhwidth - 80;
	  }
	xlimi[1] = cap->car_imwid - xlimi[0];
	xlimi[3] = cap->car_imwid - xlimi[2];
	xlimi[5] = cap->car_imwid - xlimi[4];
	ylimi[0] = FRAC_TRUST * (double)cap->car_imht;
	spacemin = ((double)cap->car_dhheight) - 20. * cap->car_startspacing;
	if (ylimi[0] < spacemin)
	  {
		ylimi[0] = spacemin;
		ylimi[2] = 0.5 * (((double)cap->car_dhheight) + ylimi[0]);
	  }
	else	ylimi[2] = 2. * ylimi[0];
	ylimi[4] = FRAC_CEN * (double)cap->car_imht;
	if (ylimi[4] < ylimi[2]) ylimi[4] = ylimi[2];
	if (ylimi[0] < cap->car_dhheight - 320)
	  {
		ylimi[0] = cap->car_dhheight - 320;
		ylimi[2] = cap->car_dhheight - 160;
		ylimi[4] = cap->car_dhheight - 80;
	  }
	ylimi[1] = cap->car_imht - ylimi[0];
	ylimi[3] = cap->car_imht - ylimi[2];
	ylimi[5] = cap->car_imht - ylimi[4];
	/* get the x indices of the centroids and then sort on Y */
	ncolumns = indx_x(cap, ylimi);
	if (cap->car_ntroids > 1) qsort((AVOID *)(cap->car_ctr0),
	 (size_t)(cap->car_ntroids), sizeof (CTROID), (QPE)ycmp);
	/* determine the y indices of the points */
	cap->car_startspacing *= ((double)(cap->car_imht)) /
		((double)(cap->car_imwid));
	/* with hexagonal grids we've now created a bunch of new points,
	 and the effective spacing in Y has been halved. */
	if (cap->car_flags & CF_HEXAGGRID) cap->car_startspacing *= 0.5;
	nrows = indx_y(&nboth, cap, xlimi);
	fprintf(stdout,
 "Number of Spots Indexed in Center of Pattern:%5d in %2d rows, %2d columns\n",
		nboth, nrows, ncolumns);
	fprintf(cap->car_fpl,
 "Number of Spots Indexed in Center of Pattern:%5d in %2d rows, %2d columns\n",
		nboth, nrows, ncolumns);
	/* now take the half-indexed points and determine their indices */
	nadded = fill_half_index(cap);
	if (cap->car_flags & CF_VERBOSE) list_troids(cap->car_fpl,
		cap->car_ctr0, cap->car_ntroids, "after initial indexing");
	/* hexagonal grids: all the truly observed (non-interpolated)
	points should have the same `polarity'.  All the interpolated
	points should have the opposite polarity */
	if (cap->car_flags & CF_HEXAGGRID)
	  {
		nrealeven = nrealodd = ninteven = nintodd = 0;
		for (lct = cap->car_ctr0; lct < cap->car_ctend; lct++)
		   {
			if (lct->ux >= 0 && lct->uy >= 0)
			  {
				polarsum = lct->ux + lct->uy;
				if (polarsum % 2)
				  {
					if (lct->cinten > 0.) nrealodd++;
					else nintodd++;
				  }
				else {
					if (lct->cinten > 0.) nrealeven++;
					else nrealodd++;
				     }
			  }
		   }
		if (nrealeven > nrealodd) badpoints = nrealodd + ninteven;
		else {
			badpoints = nrealeven + nintodd;
			cap->car_flags |= CF_POLARODD;
		     }
		if (badpoints)
		  {
			fprintf(stdout,
				"Polarity error: # even pts # odd pts\n");
			fprintf(stdout, "Observed pts  : %9d %9d\n",
				nrealeven, nrealodd);
			fprintf(stdout, "Interpol. pts : %9d %9d\n",
				ninteven, nintodd);
			for (lct = cap->car_ctr0; lct < cap->car_ctend; lct++)
			   if (lct->ux >= 0 && lct->uy >= 0)
			     {
				polarsum = (lct->ux + lct->uy) % 2;
				if (cap->car_flags & CF_POLARODD)
					polarsum = 1 - polarsum;
				if (lct->cinten < 0.) polarsum = 1 - polarsum;
				if (polarsum)
				  {
					fprintf(stdout,
					 "Unindex [%5.1f,%5.1f], %2d, %2d\n",
					 lct->cxcen, lct->cycen,
					 lct->ux, lct->uy);
					lct->ux = lct->uy = -2;
				  }
			     }
		  }
	  }
	/* adjust indexing using still-half-indexed points if necessary */
	fprintf(stdout,
 "Number of Spots Indexed from Flat Lines     :%5d\n", nadded);
	fprintf(cap->car_fpl,
 "Number of Spots Indexed from Flat Lines     :%5d\n", nadded);
}

/* end of calindex.c */
