/* caltools.c -- tools for manipulating USPLINE files and related stuff

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	27 Jun 2007: return rather than exit in scaleumain
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	20 Jun 2007: return rather than exit in splcvtmain
	19 Apr 2007: float->double conversions; reorganize use of "{"
	15 May 2006: merge {compuca,p4pconvert,splcvt,cvtuca,scaleuca}.c here
	 3 Jan 2006: set v_maxx,v_maxy elements instead of calling record_detxy
	 9 Dec 2005: memzap(spline stuff) removed from p4pconvert
	20 Nov 2005: inline function declarations & function prototypes
	12 Oct 2005: eliminate static globals from p4pconvert; housekeeping
	11 Oct 2005: replace static globals with ECENWORK structure elements;
		housekeeping
	30 Jun 2005: get_spxy(): sscanf %d, not %ld now
	22 Jun 2005: merge ix_2uca.c and sie_2c.c to create p4pconvert.c
	11 May 2005: check return status from xgjournal
	 4 May 2005: convert to a callable routine
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	28 Aug 2004: routines used only here are declared as static; cosmetics
	 2 Aug 2004: arg for record_detxy
	26 Jan 2004: replace sgsp and other static globals with local vbls
	21 Jan 2004: license message inserted and initial comment reordered
	 7 Jan 2004: cvtucam: handle program name through currpname, etc.
	 7 Jan 2004: calls to gspline(), putspline() updated
	 6 Jan 2004: CALIBRUN structure elements in function calls
		& static globals replace extern globals
	20 Nov 2003: replace global variables with elements of CALIBRUN struct
	20 Sep 2003: two ampersands in comparison in makesaxi()
	19 Sep 2003: initialize printworst and worstcut before using them;
		printouts in get_spxy() of xglongs use explicit %ld fmts;
		variables in invertcm() initialized better
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Sep 2003: cvtucam: call xgaccess() rather than access()
		to allow us to gunzip the USPLINE file as we pass through
	 5 Jan 2003: reorganized source-file header
	 1 Nov 2002: various minor reformattings
	25 Mar 2002: removed MSI attachments
	28 May 1998: initial version of scaleuca
	28 Nov 1997: initial version of compuca
	 5 Jul 1997: initial version of splcvt
	 5 May 1997: initial version of what eventually became p4pconvert
	 5 Oct 1994: cvtucam: use xgread to input VSPLINE data
	23 Sep 1994: initial version of cvtucam

   Purpose: various functions related to USPLINE files and related stuff
   Cross references: ../prerefine/cal*.c, ../xglib/spline.c

 Individual capabilities:
	compuca looks at a series of points on a detector image space and
	 compares the splined values to the inputs.  Thus:
	 input (X,Y) -(pix2cm)-> intermediate (x,y) -(cm2pix)-> output (X,Y)
	p4pconvert converts from Siemens/Bruker ".p4p" format
		to either a calibration centroids format or VSPLINE
	splcvt enables rescaling of the contents of a USPLINE file
	cvtucam converts between XENGEN 1.0-2.0 "USPLINE" files and
		the type found in XENGEN 2.1 and X-GEN.
	scaleuca: re-scale a USPLINE file. Typical application
	 is converting a binned .uca file to an unbinned.
*************************************************************************/
#define		NEED_ERRNO	1
#define		NEED_STAT	1
#include	"xgen.h"
#define		DEFAULT_WORSTCUT	2
#define		NCALX	19
#define		NCALY	19

 typedef struct {
		short	sax_dimen, sax_nholes;
		double	sax_xcen, sax_ycen;
		xglong	*sax_holex, *sax_holey, *sax_invhx, *sax_invhy;
		} SAXISPATIAL;

 typedef struct {
		float	xc[NCALY][NCALX], yc[NCALY][NCALX];
		} CALVEC;

 typedef struct {
		float	xv[NCALX], yv[NCALY];
		} COORD;

 typedef struct {						/* bytes:*/
		float		absxcen,absycen,relxcen,relycen;/*    16 */
		CALVEC		calib, qecor;			/*  5776 */
		float		splrffu[5054];			/* 20216 */
		CALVEC		fttos;				/*  2888 */
		COORD		stot,ttos;			/*   304 */
		float		raster;				/*     4 */
		u_char		ift[NCALY][NCALX];		/*   364 */
		short		ntrue;				/*     4 */
		} SPLINE;			/* total:	   29572 bytes*/

extern int compucamain(int argc, char *argv[]);
extern int p4pconvertmain(int argc, char *argv[]);
extern int splcvtmain(int argc, char *argv[]);
extern int cvtucamain(int argc, char *argv[]);
extern int scaleumain(int argc, char *argv[]);

/*****************
	compuca: Compare the pixel (X,Y) values of some USPLINE points with
	the values computed by converting those to centimeters and back.
*****************/
static void log_head(FILE *fp)
{ /* provide a header for the useful logfile output */
	fprintf(fp, " Xin(pix) Yin(pix)      xcm      ycm  Xout(pix)");
	fprintf(fp, "  Yout(pix)  error(X)  error(Y)\n");
}

static void head_ucafin(FILE *fp)
{ /* Header for final uca reportage */
	fprintf(fp, " Results of pixtocm and cmtopix calculations:\n");
	fprintf(fp,
 "Pixel-to-centimeter calculation          centimeter-to-pixel calculation\n");
	fprintf(fp,
 "  Good  Outside     Far from Illegal   Good Outside     Far from Illegal\n");
	fprintf(fp,
 " pixels Active Area Detector  Pixels Pixels Active Area Detector  Pixels\n");
}

int compucamain(int argc, char *argv[])
{ /*	Mainline for comparing splined points to observed points */
	int		locht, locwid, printworst, quadr, quady, retv, rownum;
	int		x, xma, xmi, y, yma, ymi, rctop[4], rptoc[4]; 
	double		dex, dey, rmsx, rmsy, rnpts, sigx, sigy, worstcut;
	float		xcm, ycm, xpxo, ypxo;
	VSPLINE		*sgsp;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	/* read in the input spline file */
	if (NULL == (sgsp = gspline(5, &retv))) return retv;
	/* see if the "-w" flag (meaning "print worst") is set. */
	printworst = 0;	worstcut = DEFAULT_WORSTCUT;
	if ((argc > 1) && (argv[1][0] == '-'))
	  {
		argc--;	argv++;
		if ((argv[0][1] == 'w') || (argv[0][1] == 'W'))
		  {
			printworst = 1; worstcut = DEFAULT_WORSTCUT;
			if (allnumeric(&(argv[0][2])))
			  {
				xcm = atof(&(argv[0][2]));
				if ((xcm > 0.) && (xcm < 1000.))
					worstcut = xcm;
			  }
		  }
	  }
	/* override the detector dimensions if requested */
	if ((argc > 1) && allnumeric(argv[1]))
	  {
		sgsp->v_detw = atoi(argv[1]);
		if ((argc > 2) && allnumeric(argv[2]))
			sgsp->v_deth = atoi(argv[2]);
		locwid = sgsp->v_detw;	locht = sgsp->v_deth;
		sgsp->v_maxx = sgsp->v_detw + 10;
		sgsp->v_maxy = sgsp->v_deth + 10;
	  }
	else {
		locwid = sgsp->v_detw;	locht = sgsp->v_deth;
	     }
	if (NULL == activmask(locwid, locht, 5))
		return (errno == ENOMEM) ? EALLOC : EREADERR;
	for (y = 0; y < 4; y++) rptoc[y] = rctop[y] = 0;
	rmsx = rmsy = sigx = sigy = 0;
	xmi = sgsp->v_x0px; xma = xmi + sgsp->v_xrpx * (double)sgsp->v_ncol;
	ymi = sgsp->v_y0px; yma = ymi + sgsp->v_yrpx * (double)sgsp->v_nrow;
	fprintf(stdout, "X limits: %6d, %6d; Y limits: %6d, %6d\n",
	 xmi, xma, ymi, yma);
	log_head(stdout);
	rownum = 0;
	for (y = ymi; y <= yma; y++)
	   {
		quady = (y >= sgsp->v_deth / 2);
		for (x = xmi; x < xma; x++)
		   {
			quadr = quady + 2 * (x >= sgsp->v_detw / 2);
			retv = pix2cm((float)x, (float)y,
			 &xcm, &ycm, 0);
			if ((-1 <= retv) && (retv <= 2))
				(rptoc[retv + 1])++;
			if (retv == 1)
			  {
				if (sgsp->v_nreg == 4) retv =
				 cm2pixq(xcm, ycm, quadr, &xpxo, &ypxo, 0);
				else retv = cm2pix(xcm, ycm, &xpxo, &ypxo, 0);
				if ((-1 <= retv) && (retv <= 2))
					(rctop[retv + 1])++;
				/* we used to only want retv == 1;
				 now we accept anything. */
				dex = xpxo - (double)x;	dey = ypxo - (double)y;
				rmsx += dex * dex;	rmsy += dey * dey;
				sigx += dex;		sigy += dey;
				if (((printworst) && (((fabs(dex) > worstcut))
				 || (fabs(dey) > worstcut))) ||
				 ((!(y%17)) && (!(x%13)))) fprintf(stdout,
			 " %8d %8d %8.5f %8.5f %10.2f %10.2f %9.3f %9.3f\n",
				 x, y, xcm, ycm, xpxo, ypxo,
				 xpxo - (float)x, ypxo - (float)y);
			  }
		   }
		rownum++;
	   }
	rnpts = rctop[2];
	rmsx = sqrt(rmsx/rnpts);	rmsy = sqrt(rmsy/rnpts);
	sigx /= rnpts;	sigy /= rnpts;
	head_ucafin(stdout);
	fprintf(stdout, " %6d %11d %8d %7d %6d %11d %8d %7d\n",
	 rptoc[2], rptoc[3], rptoc[1], rptoc[0],
	 rctop[2], rctop[3], rctop[1], rctop[0]);
	fprintf(stdout, " RMS errors in X,Y = %13.5e %13.5e\n", rmsx, rmsy);
	fprintf(stdout, " Mean diff in X, Y = %13.5e %13.5e\n", sigx, sigy);
	(void)gspline(0, &retv);	globalkill();	return EOK;
}

/* end of compuca functionalities */

/*****************************
	p4pconvert: these routines enable you to create a USPLINE from
	information contained in a Siemens / Bruker .p4p format file
*****************************/

static int get_spxy(FILE *fp, char *ib, xglong *holep)
{ /*	This reads the actual lines of the spatial correction. It
	returns the number of spatial elements actually found */
	int	nread, ii, lineno;
	xglong	iv[11], *ip, *iq;

	ip = holep;
	for (lineno = 0; NULL != fgets(ib, 254, fp); lineno++)
	   {
		nread = sscanf(ib,
		 " %d %d %d %d %d %d %d %d %d %d",
		 &(iv[0]), &(iv[1]), &(iv[2]), &(iv[3]),
		 &(iv[4]), &(iv[5]), &(iv[6]), &(iv[7]),
		 &(iv[8]), &(iv[9]));
		iq = &(iv[0]);
		for (ii = 0; ii < nread; ii++) *ip++ = *iq++;
		if (nread < 10) break;
	   }
	return (int)(ip - holep);
}

static int makesaxi(SAXISPATIAL *sspl)
{ /*	This creates and fills a SAXISPATIAL structure by reading a p4p file */
	char		inl[258], *cp;
	int		nitems, indimen, inholes, retval, nfound, intsq;
	double		inx, iny, indist, uf0, uf1, uf2, fdm;
	xglong		*mbase;
	static FILE	*fppi;

	if (qerror((AVOID *)(fppi = xgopen("p4p", "r")),
	 "open input p4p file")) return EOPENERR;
	retval = EREADERR; /* ... until we've read what we need */
	while (NULL != fgets(inl, 254, fppi))
	     {
		for (cp = inl; ('\0' != *cp && '\n' != *cp); cp++)
		   if (('a' <= *cp) && (*cp <= 'z')) *cp -= 'a' - 'A';
		if (!strncmp(inl, "ADPAR", 5))
		  {
			nitems = sscanf(&(inl[5]), " %lf %lf %lf %d",
			  &inx, &iny, &indist, &indimen);
			printf( "beam center at [%7.2f,%7.2f]\n", inx, iny);
			printf( "detector dimension = %d\n", indimen);
			printf( "detector distance = %7.2f cm\n", indist);
			if (nitems >= 4)
			  {
				sspl->sax_dimen = indimen;
				sspl->sax_xcen = ((double)indimen) - inx;
				sspl->sax_ycen = ((double)indimen) - iny;
			  }
		  }
		else if (!strncmp(inl, "DATA  SPATIAL", 13))
		  {
			nitems = sscanf(&(inl[38]), " %lf %lf %lf %lf %d %d",
			 &uf0, &uf1, &fdm, &uf2, &inholes, &indimen);
			printf(" number of fiducial points = %d\n", inholes);
			intsq = (int)sqrt((double)inholes);
			if (intsq * intsq < inholes) fprintf(stderr,
			 "Warning: # holes = %d is not a perfect square\n",
				inholes);
			if (nitems >= 5)
			  {
				sspl->sax_nholes = inholes;
				if (sspl->sax_nholes < 1) break;
				sspl->sax_dimen = indimen;
				if (NULL != sspl->sax_holex)
					memzap((AVOID **)&(sspl->sax_holex),
					 "sspl->sax_holex in makesaxi");
				mbase = (xglong *)callop((unsigned)(4*inholes),
				 sizeof (xglong), "spline-holes");
				if (NULL == mbase) return EALLOC;
				sspl->sax_holex = mbase;
				sspl->sax_holey = mbase + inholes;
				sspl->sax_invhx = mbase + 2 * inholes;
				sspl->sax_invhy = mbase + 3 * inholes;
				nfound = get_spxy(fppi, inl, sspl->sax_holex);
				printf(
				 "Number of hole values actually found = %d\n",
				 nfound);
				if (nfound >= 4 * inholes) retval = EOK;
				break;
			  }
		  }
	     }
	(void)fclose(fppi);	fppi = NULL;
	return retval;
}

static int invertcm(int iy, int ix, VSPLINE *nspl)
{ /*	This assumes that the "ctopx" array is correctly filled and creates a
	point in the "ptocx" array from it. It returns 0 iff. successful. */
	int		jx, jy, poff, ix0, iy0, interp_needed;
	int		ux[4], uy[4];
	double		dmi0, dx, dy, xci, yci, xcc, ycc;
	double		tdi2, wi, pvx, pvy, denom;
	CALPOINT	*icp, *ocp;

	/* remember pointer to current position */
	ocp = nspl->v_cpt + iy * nspl->v_ncol + ix;
	xci = ocp->c_ctopx;	yci = ocp->c_ctopx;
	/* find point in cockeyed grid closest to this one */
	iy0 = nspl->v_nrow + 12;	ix0 = nspl->v_ncol + 12;
	dmi0 = 1.e12;	icp = nspl->v_cpt;
	for (jy = 0; jy < nspl->v_nrow; jy++)
	 for (jx = 0; jx < nspl->v_ncol; jx++, icp++)
	    {
		xcc = icp->c_ctopx;	ycc = icp->c_ctopy;
		dx = xci - xcc;	dy = yci - ycc;
		tdi2 = dx * dx + dy * dy;
		if (tdi2 < dmi0)
		  {
			dmi0 = tdi2; ix0 = jx; iy0 = jy;
		  }
	   }
	if (iy0 > nspl->v_nrow)
	  { /* no good points */
		fprintf(stderr, "No good points!\n"); return 1;
	  }
	/* see which direction to move in order to get the
	 other points in the list */
	poff = nspl->v_ncol * iy0 + ix0;
	icp = nspl->v_cpt + poff;
	xcc = icp->c_ctopx;	ycc = icp->c_ctopy;
	ux[2] = ux[0] = ix0;	uy[1] = uy[0] = iy0;
	if (iy0 <= 0) uy[2] = 1;
	else if (iy0 >= nspl->v_nrow - 1) uy[2] = nspl->v_nrow - 2;
	else if (nspl->v_yrcm > 0.) uy[2] = (yci > ycc) ? iy0 + 1 : iy0 - 1;
	else uy[2] = (yci > ycc) ? iy0 - 1 : iy0 + 1;
	if (ix0 <= 0) ux[1] = 1;
	else if (ix0 >= nspl->v_ncol - 1) ux[1] = nspl->v_nrow - 2;
	else if (nspl->v_xrcm > 0.) ux[1] = (xci > xcc) ? ix0 + 1 : ix0 - 1;
	else ux[1] = (xci > xcc) ? ix0 - 1 : ix0 + 1;
	uy[3] = uy[2];	ux[3] = ux[1];
	/* phew. We've now found the four closest points. Now do
	 an interpolation with respect to those. */
	interp_needed = 1;	pvx = pvy = denom = 0.;
	for (jx = 0; jx < 4; jx++)
	   {
		icp = nspl->v_cpt + uy[jx] * nspl->v_ncol + ux[jx];
		xcc = icp->c_ctopx;	ycc = icp->c_ctopy;
		dx = xci - xcc;	dy = yci - ycc;
		tdi2 = dx * dx + dy * dy;
		if (tdi2 < DFUZZ)
		  {
			pvx = nspl->v_x0px + ux[jx] * nspl->v_xrpx;
			pvy = nspl->v_y0px + uy[jx] * nspl->v_yrpx;
			interp_needed = 0;	break;
		  }
		else {
			wi = 1. / sqrt(tdi2);	denom += wi;
			pvx += (nspl->v_x0px + ux[jx] * nspl->v_xrpx) * wi;
			pvy += (nspl->v_y0px + uy[jx] * nspl->v_yrpx) * wi;
		     }
	   }
	if (interp_needed)
	  {
		if (denom > DFUZZ)
		  {
			pvx /= denom;	pvy /= denom;
		  }
		else fprintf(stderr,
			"Warning: denom = %12.5e for ix,iy=%3d,%3d\n",
			denom, ix, iy);
		return 2;
	  }
	ocp = nspl->v_cpt + iy * nspl->v_nrow + ix;
	ocp->c_ptocx = pvx;	ocp->c_ptocy = pvy;
	return 0;
}

static int saxitouca(SAXISPATIAL *sspl, VSPLINE *nspl, int ismosaic)
{ /*	This converts a Siemens spatial information structure (as contained
	in a ".p4p" file) into an XENGEN 2.1 and X-GEN VSPLINE structure.
	The routine returns EOK if successful, EALLOC otherwise. */
	int		ix, iy, scols, srows, nholes, retv;
	CALPOINT	*calp;
	xglong		*rpxh, *rpyh, *rcxh, *rcyh;
	double		rescalex, rescaley, offsetx, offsety;

	nspl->v_detw = nspl->v_deth = sspl->sax_dimen; 
	nholes = sspl->sax_nholes;
	nspl->v_ncol = nspl->v_nrow = scols = srows = sqrt((double)nholes);
	nspl->v_midc = nspl->v_midr = scols / 2;
	printf( "detector is %d-by-%d pixels\n", nspl->v_detw, nspl->v_deth);
	printf( "correction is fiducialized on a %d-by-%d hole grid\n",
		nspl->v_ncol, nspl->v_nrow);
	nspl->v_nreg = ismosaic;	nspl->v_ntrue = nholes;
	nspl->v_axc = sspl->sax_xcen;	nspl->v_ayc = sspl->sax_ycen;
	/* kludge: I don't understand the format of p4p altogether... */
	if (sspl->sax_nholes < 1000) nspl->v_xrcm = 0.508;
	else	nspl->v_xrcm = 0.254;
	nspl->v_yrcm = -nspl->v_xrcm;
	nspl->v_xrpx = ((double)(nspl->v_detw + 10)) / (double)(nspl->v_ncol);
	nspl->v_yrpx = ((double)(nspl->v_deth + 10)) / (double)(nspl->v_nrow);
	nspl->v_x0cm = -((double)nspl->v_midc) * nspl->v_xrcm;
	nspl->v_y0cm = -((double)nspl->v_midr) * nspl->v_yrcm;
	nspl->v_x0px = -5.;	nspl->v_y0px = -5.;
	printf(" input beam center is at [%8.2f,%8.2f]\n",
		nspl->v_axc, nspl->v_ayc);
	printf(
	 " fiducial points are %8.3f cm apart in X, %8.3f cm apart in Y\n",
		nspl->v_xrcm, nspl->v_yrcm);
	printf(
 	 " upper lefthand corner of detector is at coordinates [%8.4f,%8.4f]\n",
		nspl->v_x0cm, nspl->v_y0cm);
	printf(
 " pixel indexing points are nominally %8.3f pixels apart in X, %8.3f in Y\n",
		nspl->v_xrpx, nspl->v_yrpx);
	printf(
 "starting pixel positions are [%8.3f, %8.3f]\n", nspl->v_x0px, nspl->v_y0px);
	if (NULL == (nspl->v_cpt))
	  {
		if (NULL == (nspl->v_cpt = (CALPOINT *)callop((unsigned)nholes,
		 sizeof (CALPOINT),"nspl->v_cpt in saxitouca"))) return EALLOC;
	  }
	rpxh = sspl->sax_holex;	rpyh = sspl->sax_holey;
	rcxh = sspl->sax_invhx;	rcyh = sspl->sax_invhy;
	rescalex = (nspl->v_xrcm * (double)nspl->v_ncol)/
		(1000. * (nspl->v_xrpx * (double)nspl->v_ncol));
	rescaley = (nspl->v_yrcm * (double)nspl->v_nrow) /
		(1000. * (nspl->v_yrpx * (double)nspl->v_nrow));
	offsetx = 500. * (nspl->v_xrpx * (double)nspl->v_ncol);
	offsety = 500. * (nspl->v_yrpx * (double)nspl->v_nrow);
	printf(
 " Siemens integer X coordinates rescaled and offset by %13.5e, %9.2f\n",
		rescalex, offsetx);
	printf(
 " Siemens integer Y coordinates rescaled and offset by %13.5e, %9.2f\n",
		rescaley, offsety);
	/* I believe the Siemens stuff in the cm direction */
	for (calp = nspl->v_cpt, iy = 0; iy < srows; iy++)
	   for (ix = 0; ix < scols; ix++, calp++)
	      {
		calp->c_ptocx = ((double)(*rpxh)) / 1000.;
		calp->c_ptocy = ((double)(*rpyh)) / 1000.;
		calp->c_ctopx = (((double)(*rcxh)) - offsetx) * rescalex;
		calp->c_ctopy = (((double)(*rcyh)) - offsety) * rescaley;
		calp->c_qec = 0.;	calp->c_ift = 1;
		calp->c_set = calp->c_cset = 1;
		rpxh++;	rpyh++;	rcxh++;	rcyh++;
	      }
	/* do the pixel-to-cm side by inversion */
	for (iy = 0; iy < srows; iy++)
	   for (ix = 0; ix < scols; ix++)
		if (0 != (retv = invertcm(iy, ix, nspl))) return retv;
	printf( "%d holes processed\n", srows * scols);
	/* see if we can re-do the center correctly */
	if (-1 == cm2pix(0.,0., &(nspl->v_rxc), &(nspl->v_ryc), 0))
	  {
		nspl->v_rxc = nspl->v_axc;	nspl->v_ryc = nspl->v_ayc;
		fprintf(stderr, "Warning: cm2pix failed\n");
	  }
	else if ((fabs(nspl->v_rxc - nspl->v_axc) > 0.2) ||
	 (fabs(nspl->v_rxc - nspl->v_axc) > 0.2))
	 printf(
 " Beam center recalculated from [%8.2f,%8.2f] to [%8.2f,%8.2f]\n",
		nspl->v_axc, nspl->v_ayc, nspl->v_rxc, nspl->v_ryc);
	return EOK;
}

static int p4pconvert(SAXISPATIAL *sspl, int whichdir)
{ /*	This converts a Siemens spatial information structure (as contained
	in a ".p4p" file) into an XENGEN 2.1 and X-GEN CALCENTROIDS file.
	The routine returns EOK if successful, EALLOC otherwise. */
	int		ix, iy, scols, srows, nholes;
	xglong		*rcxh, *rcyh;
	double		ux, uy, offx, offy, scx, scy;
	FILE		*fpo;

	if (NULL == (fpo = xgopen("calcentroids", "w"))) return EOPENERR;
	nholes = sspl->sax_nholes;
	scols = srows = sqrt((double)nholes);
	rcxh = sspl->sax_invhx;	rcyh = sspl->sax_invhy;
	/* use either the first or the second set */
	if (whichdir)
	  {
		rcxh = sspl->sax_invhx; rcyh = sspl->sax_invhy;
	  }
	else
	  {
		rcxh = sspl->sax_holex; rcyh = sspl->sax_holey;
	  }
	offy = offx = ((double)(sspl->sax_dimen)) - 1.;
	scy = scx = 1. / 1000.;
	for (iy = 0; iy < srows; iy++)
		for (ix = 0; ix < scols; ix++, rcxh++, rcyh++)
		   {
			uy = offx - scx * (double)(*rcxh);
			ux = offy - scy * (double)(*rcyh);
		   	fprintf(fpo, " %8.3f %8.3f\n", ux, uy);
		   }
	(void)fclose(fpo);	fpo = NULL;
	printf(" %6d calibration centroids output\n", srows * scols);
	return EOK;
}

int p4pconvertmain(int argc, char *argv[])
{ /*	In this mainline, we handle the arguments and then process each file.
	the flag value "p4flags" has three active bits:
	bit	meaning of "off"		meaning of "on"
	  0	select 1st set of inputs	select 2nd set of inputs
	  1	create VSPLINE output		create CALCENTROIDS output
	  2	single-element detector		4-element detector
	The run-time flags that handle this are:
	-c	Create CALCENTROIDS rather than VSPLINE
	-m	Four-element detector, not one-element detector
	-r	Use the second set of inputs for each spot in p4p file
	-s	Create VSPLINE rather than CALCENTROIDS (default)	*/
	int			omit, p4flags, retval;
	char			**dcp, **flags;
	static VSPLINE		*sgsp;
	static SAXISPATIAL	ssp;
	static char		*flist[] =
		{"p4p", "calcentroids", "uspline", NULL};
	static char		ftypes[] = { 1, 2, 2, 0};

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	p4flags = 0;
	flags = flagproc(argc, argv, "CMRV");
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	   {
		switch (**dcp)
		     {
			case	'C':	p4flags |= 02;	break;
			case	'M':	p4flags |= 04;	break;
			case	'R':	p4flags |= 01;	break;
			case	'V':	p4flags &= ~02;	break;
		     }
	   }
	/* report only on the files that we're going to work with */
	omit = (p4flags & 02) ? 2 : 1; ftypes[omit] = 0; ftypes[3-omit] = 2;
	header(stdout, PROCESSING, flist, ftypes);
	printf(" %s pair of arrays will be output\n",
	 (p4flags & 01) ? "Second" : "First");
	/* read in the p4p file's calibration values */
	if (EOK != (retval = makesaxi(&ssp)))
	  {
		fprintf(stderr, "Failure %d in makesaxi\n", retval);
		return retval;
	  }
	if (!(p4flags & 02)) /* if user asked for a VSPLINE, get it */
	  {
		if (NULL == sgsp)
		  {
			if (NULL == (sgsp = (VSPLINE *)callop(1,
			 sizeof (VSPLINE), "sgsp in p4pconvert")))
				return EALLOC;
		  }
		else /* get rid of old memory for sgsp->v_cpt if any */
		  memzap((AVOID **)&(sgsp->v_cpt), "v_cpt in p4pconvert");
		if (EOK != (retval = saxitouca(&ssp, sgsp, p4flags & 04)))
		  {
			fprintf(stderr, "Failure %d in saxitouca\n", retval);
			return retval;
		  } /* convert p4p stuff to VSPLINE */
		if (NULL == (sgsp = gspline(9, &retval))) /* output it */
		  {
			fprintf(stderr, "Failure %d in outputting spline\n",
			 retval); return retval;
		  }
	  }
	else { /* in this case, we just write out a CALCENTROIDS file */
		if (EOK != (retval = p4pconvert(&ssp, p4flags & 01)))
		  {
			fprintf(stderr, "Failure %d in p4pconvert\n", retval);
			return retval;
		  }
	     }
	globalkill();	return EOK;
}

/* end of p4pconvert functionalities */

/*****************************
 	splcvt: interactive conversion of VSPLINE files.
	This rescales the pixel-to-cm conversion.
*****************************/

int splcvtmain(int argc, char *argv[])
{ /*	In this mainline, we handle the arguments and then process each file */
	int		retv;
	float		cmx, cmy, iq0, iq1;
	double		midx, midy, scx, scy, ucmx, ucmy;
	char		inl[120];
	VSPLINE		*sgsp;
	static char	*flist[] = {"uspline", NULL};
	static char	ftypes[] = { 1, 0};

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	header(stdout, PROCESSING, flist, ftypes);
	if (NULL == (sgsp = gspline(5, &retv)))
	  {
		fprintf(stderr,
			"Failure in reading spline file: error %d\n", retv);
		return retv;
	  }
	midx = 0.5 * (double)sgsp->v_detw; midy = 0.5 * (double)sgsp->v_deth;
	scx = sgsp->v_xrpx / sgsp->v_xrcm; scy = sgsp->v_yrpx / sgsp->v_yrcm;
	while (1)
	     {
		printf("Enter (X,Y) in pixels; <CR> to quit: ");
		if (NULL == fgets(inl, 118, stdin)) break;
		if (('\0' == inl[0]) || ('\n' == inl[0])) break;
		sscanf(inl, " %f %f", &iq0, &iq1);
		if ((0. <= iq0) && (iq0 <= (float)(sgsp->v_detw)) &&
		 (0. <= iq1) && (iq1 <= (float)(sgsp->v_deth)))
		  {
			retv = pix2cm(iq0, iq1, &cmx, &cmy, 0);
			ucmx = midx + cmx * scx; ucmy = midy + cmy * scy;
			if (0 == retv) printf(" %7.2f %7.2f: no cm position\n",
				iq0, iq1);
			else printf(
			 " %7.2f %7.2f %8.5f %8.5f %8.2f %8.2f %3d\n",
			 iq0, iq1, cmx, cmy, ucmx, ucmy, retv);
		  }
		else printf(" %7.2f %7.2f: outside detector bounds\n",
			iq0, iq1);
	     }
	globalkill();	return EOK;
}

/* end of splcvt functionality */

/********************************
	cvtucam: convert between old and new USPLINE formats.
*********************************/
static int oldtonew(SPLINE *ospl, VSPLINE *nspl, int swapflag)
{ /*	This converts a XENGEN 1.0-2.0 "SPLINE" structure to
	a XENGEN 2.1 and X-GEN VSPLINE structure. This is the easy direction.
	The routine returns EOK if successful, EALLOC otherwise. */
	int		ix, iy;
	CALPOINT	*calp;
	float		*fl;

	nspl->v_detw = nspl->v_deth = 512;
	nspl->v_ncol = nspl->v_nrow = 19;
	nspl->v_midc = nspl->v_midr = 9;
	nspl->v_nreg = 0;
	printf("number true before swap = %d\n", ospl->ntrue);
	switch (swapflag)
	      {
		case	1:	/* byte-and-word swap */
			ospl->ntrue = sswab(ospl->ntrue);
			for (fl = &(ospl->absxcen); fl <= &(ospl->raster);
			 fl++) float_cvt((void *)fl);
			break;
		case	2:	/* VMS swaps */
			ospl->ntrue = sswab(ospl->ntrue);
			for (fl = &(ospl->absxcen); fl <= &(ospl->raster);
			 fl++) flcvt((void *)fl, 0);
			break;
		default:	break;
	       }
	printf("number true = %d\n", ospl->ntrue);
	printf(
	 "absxcen, absycen = %12.4e, %12.4e\n", ospl->absxcen, ospl->absycen);
	nspl->v_ntrue = ospl->ntrue;
	nspl->v_axc = ospl->absxcen; nspl->v_ayc = ospl->absycen;
	nspl->v_rxc = ospl->relxcen; nspl->v_ryc = ospl->relycen;
	nspl->v_xrcm = ospl->raster; nspl->v_yrcm = -(ospl->raster);
	nspl->v_xrpx = nspl->v_yrpx = 29.;
	nspl->v_x0cm = ospl->stot.xv[0]; nspl->v_y0cm = ospl->stot.yv[0];
	nspl->v_x0px = ospl->ttos.xv[0]; nspl->v_y0px = ospl->ttos.yv[0];
	if (NULL == (nspl->v_cpt))
	  {
		if (NULL == (nspl->v_cpt = (CALPOINT *)callop(361,
		 sizeof (CALPOINT), "nspl->v_cpt in oldtonew")))
			return EALLOC;
	  }
	for (calp = nspl->v_cpt, iy = 0; iy < NCALY; iy++)
	   for (ix = 0; ix < NCALX; ix++, calp++)
	      {
		calp->c_ptocx = ospl->calib.xc[iy][ix];
		calp->c_ptocy = ospl->calib.yc[iy][ix];
		calp->c_ctopx = ospl->fttos.xc[iy][ix];
		calp->c_ctopy = ospl->fttos.yc[iy][ix];
		calp->c_qec = ospl->qecor.xc[iy][ix];
		calp->c_ift = ospl->ift[iy][ix];
		calp->c_set = calp->c_cset = 1;
	      }
	return EOK;
}

static int newtoold(VSPLINE *nspl, SPLINE *ospl /* , int swapflag */)
{ /*	This converts a XENGEN 2.1 and X-GEN VSPLINE structure
	to a XENGEN 1.0-2.0 "SPLINE" structure. This is the hard direction.
	The routine returns EOK if successful, EREADERR otherwise.
	For the moment we don't use the swapflag. */
	int		ix, ix0, ixm, iy, iy0, iym, jx0, jy0;
	CALPOINT	*calp, *calp0, *calpy;
	double		ux, ux0, uy, uy0;

	ospl->absxcen = nspl->v_axc; ospl->absycen = nspl->v_ayc;
	ospl->relxcen = nspl->v_rxc; ospl->relycen = nspl->v_ryc;
	ospl->raster = nspl->v_xrcm; ospl->ntrue = nspl->v_ntrue;
	if (ospl->ntrue > 361) ospl->ntrue = 361;
	ix0 = iy0 = jx0 = jy0 = 0;
	if (nspl->v_ncol > NCALX) ix0 = (nspl->v_ncol - NCALX) / 2;
	else if (nspl->v_ncol < NCALX)
		jx0 = (NCALX - nspl->v_ncol) / 2;
	if (nspl->v_nrow > NCALY) iy0 = (nspl->v_nrow - NCALY) / 2;
	else if (nspl->v_nrow < NCALY)
		jy0 = (NCALY - nspl->v_nrow) / 2;
	uy0 = nspl->v_y0cm + nspl->v_yrcm * (double)(iy0 - jy0);
	for (iy = 0, uy = uy0; iy < NCALY;
	 iy++, uy += nspl->v_yrcm) ospl->stot.yv[iy] = uy;
	ux0 = nspl->v_x0cm + nspl->v_xrcm * (double)(ix0 - jx0);
	for (ix = 0, ux = nspl->v_x0cm; ix < NCALX;
	 ix++, ux += nspl->v_xrcm) ospl->stot.xv[ix] = ux;
	uy0 = nspl->v_y0px + nspl->v_yrpx * (double)(iy0 - jy0);
	for (iy = 0, uy = uy0; iy < NCALY;
	 iy++, uy += nspl->v_yrpx) ospl->ttos.yv[iy] = uy;
	ux0 = nspl->v_x0px + nspl->v_xrpx * (double)(ix0 - jx0);
	for (ix = 0, ux = nspl->v_x0px; ix < NCALX;
	 ix++, ux += nspl->v_xrpx) ospl->ttos.xv[ix] = ux;
	if (NULL == (nspl->v_cpt))
	  {
		fprintf(stderr, "Calibration-point structures not present\n");
		return EREADERR;
	  }
	calp0 = nspl->v_cpt + nspl->v_ncol * iy0;
	printf("Input has %2d rows, %2d columns; ",
		nspl->v_nrow, nspl->v_ncol);
	ixm = ix0;	if (jx0 > ixm) ixm = jx0;
	if (ixm) printf(
	 "Leftmost and rightmost %d columns of the %s will be %s\n",
		ixm, (ix0 ? "input" : "output"),
		(ix0 ? "ignored" : "extrapolated"));
	iym = iy0;	if (jy0 > iym) iym = jy0;
	if (iym) printf("Topmost and bottommost %d %s rows will be %s\n",
		iy0, (ix0 ? "input" : "output"),
		(iy0 ? "ignored" : "extrapolated"));
	for (calpy = calp0, iy = jy0; iy < NCALY - jy0;
	 iy++, calpy += nspl->v_ncol)
	   for (ix = jx0, calp = calpy + ix0; ix < NCALX - jx0;
	    ix++, calp++)
	      {
		ospl->calib.xc[iy][ix] = calp->c_ptocx;
		ospl->calib.yc[iy][ix] = calp->c_ptocy;
		ospl->fttos.xc[iy][ix] = calp->c_ctopx;
		ospl->fttos.yc[iy][ix] = calp->c_ctopy;
		ospl->qecor.xc[iy][ix] = calp->c_qec;
		ospl->ift[iy][ix] = calp->c_ift;
	      }
	if (nspl->v_nrow < NCALX) /* fill in the holes */
	  {
		for (ix = 0; ix < NCALX; ix++)
		   {
			ux0 = ospl->calib.xc[jy0][ix];
			uy0 = ospl->calib.yc[jy0][ix];
			ux  = ospl->fttos.xc[jy0][ix];
			uy  = ospl->fttos.yc[jy0][ix];
			for (iy = 0; iy < jy0; iy++)
			   {
				ospl->calib.xc[iy][ix] = ux0;
				ospl->calib.yc[iy][ix] =
				 uy0 - 15. * (double)(jy0 - iy);
				ospl->fttos.xc[iy][ix] = ux;
				ospl->fttos.yc[iy][ix] =
				 uy + 0.4 * (double)(jy0 - iy);
			   }
			iy0 = NCALY - jy0 - 1;
			ux0 = ospl->calib.xc[iy0][ix];
			uy0 = ospl->calib.yc[iy0][ix];
			ux  = ospl->fttos.xc[iy0][ix];
			uy  = ospl->fttos.yc[iy0][ix];
			for (iy = iy0 + 1; iy < NCALY; iy++)
			   {
				ospl->calib.xc[iy][ix] = ux0;
				ospl->calib.yc[iy][ix] =
				 uy0 + 15. * (double)(iy - iy0);
				ospl->fttos.xc[iy][ix] = ux;
				ospl->fttos.yc[iy][ix] =
				 uy - 0.4 * (double)(iy - iy0);
			   }
		   }
	  }
	if (nspl->v_ncol < NCALX)
	  {
		for (iy = 0; iy < NCALY; iy++)
		   {
			ux0 = ospl->calib.xc[iy][jx0];
			uy0 = ospl->calib.yc[iy][jx0];
			ux = ospl->fttos.xc[iy][jx0];
			uy = ospl->fttos.yc[iy][jx0];
			for (ix = 0; ix < jx0; ix++)
			   {
				ospl->calib.xc[iy][ix] =
				 ux0 - 26. * (double)(jx0 - ix);
				ospl->calib.yc[iy][ix] = uy0;
				ospl->fttos.xc[iy][ix] =
				 ux - 0.55 * (double)(jx0 - ix); 
				ospl->fttos.yc[iy][ix] = uy0;
			   }
			ix0 = NCALX - jx0;
			ux0 = ospl->calib.xc[iy][ix0];
			uy0 = ospl->calib.yc[iy][ix0];
			ux  = ospl->fttos.xc[iy][ix0];
			uy  = ospl->fttos.yc[iy][ix0];
			for (ix = ix0 + 1; ix < NCALX; ix++)
			   {
				ospl->calib.xc[iy][ix] =
				 ux0 + 26. * (double)(ix - ix0);
				ospl->calib.yc[iy][ix] = uy0;
				ospl->fttos.xc[iy][ix] =
				 ux + 0.65 * (double)(ix - ix0);
				ospl->fttos.yc[iy][ix] = uy0;
			   }
		   }
	  }
	return EOK;
}

int cvtucamain(int argc, char *argv[])
{ /* In this mainline, we handle the arguments and then process each file */
	char		*uname, *cpname, c;
	int		intype, irv, swapflag;
	size_t		insize;
	struct stat	inb;
	FILE		*fpi, *fpo;
	static VSPLINE	*slsp;
	static SPLINE	spl;
	static char	openmsg0[] = "for reading or writing";
	static char	openmsg1[] = "for writing";

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	swapflag = 0;
	if ((argc > 1) && (argv[1][0] == '-'))
	  {
		if (('s' == (c = argv[1][1])) || ('S' == c)) swapflag = 1;
		else if (('v' == c) || ('V' == c)) swapflag = 2;
	  }
	uname = trname("uspline");
	if (0 != xgaccess(uname, 6, 1))
	  {
		cpname = xgprogname(NULL, 0);
		fprintf(stderr, "%s: Cannot open %s for %s\n",
		 cpname, uname, (0 != access(uname, 4)) ? openmsg0 : openmsg1);
		return EOPENERR;
	  }
	if (0 != stat(uname, &inb))
	  {
		fprintf(stderr, "Cannot stat %s: nothing done!\n", uname);
		return EREADERR;
	  }
	if ((29560 <= (insize = inb.st_size)) && (insize <= 29780)) intype = 0;
	else	intype = 1;
	printf("Insize = %d: Input is a%s-style%sUSPLINE\n",
	 (int)insize, (intype ? " new" : "n old"),
	 ((swapflag == 2) ? " VMS " :
	 ((swapflag == 1) ? " byte/word swapped " : " ")));
	if (intype)	/* intype==1 means a new-format input */
	  {
		if (NULL == (slsp = gspline(5, &irv)))
		  {
			fprintf(stderr, "Failure %d in gspline\n", irv);
			return irv;
		  }
		if (EOK != newtoold(slsp, &spl /*, swapflag */))
		  {
			fprintf(stderr, "Failure %d in newtoold\n", irv);
			return irv;
		  }
		if (NULL == (fpo = xgopen("uspline", "w")))
		  {
			fprintf(stderr, "Cannot open output spline file\n");
			return EOPENERR;
		  }
		if (1 != fwrite((void *)&spl, sizeof (SPLINE) - 4, 1, fpo))
		  {
			fprintf(stderr, "Error writing output spline\n");
			return EWRITERR;
		  }
		(void)fclose(fpo);	fpo = NULL;
	  }
	else {
		if (NULL == (fpi = xgopen("uspline", "r")))
		  {
			fprintf(stderr, "Error opening input uspline\n");
			return EOPENERR;
		  }
		if (insize > (size_t)xgread((void *)&spl, 14, fpi))
		  {
			fprintf(stderr, "Error reading input uspline\n");
			return EREADERR;
		  }
		(void)fclose(fpi);	fpi = NULL;
		if (NULL == (slsp = (VSPLINE *)callop(1, sizeof (VSPLINE),
		 "slsp in cvtucamain")))
		  {
			fprintf(stderr, "error allocating VSPLINE memory\n");
			return EALLOC;
		  }
		if (EOK != (irv = oldtonew(&spl, slsp, swapflag)))
		  {
			fprintf(stderr, "Error %d in oldtonew\n", irv);
			return irv;
		  }
		if ((NULL == (slsp = gspline(9, &irv))) || (EOK != irv))
		  {
			fprintf(stderr, "Error %d in outputting spline\n", irv);
			return irv;
		  }
	     }
	globalkill();	return EOK;
}

/* end of cvtucam functionality */

/*********************************
	scaleuca: rescale a USPLINE file, typically to convert an unbinned
	.uca file to a binned one or vice versa
**********************************/

static void intoout(double scalefac, VSPLINE *sgp)
{ /*	This accepts an X-GEN VSPLINE structure and rescales the relevant
	parameters by the scalefactor. */
	int		ix, iy;
	CALPOINT	*calp;

	sgp->v_detw *= scalefac;	sgp->v_deth *= scalefac;
	sgp->v_axc *= scalefac;		sgp->v_ayc *= scalefac;
	sgp->v_rxc *= scalefac;		sgp->v_ryc *= scalefac;
	sgp->v_xrpx *= scalefac;	sgp->v_yrpx *= scalefac;
	for (calp = sgp->v_cpt, iy = 0; iy < sgp->v_nrow; iy++)
	   for (ix = 0; ix < sgp->v_ncol; ix++, calp++)
	      {
		calp->c_ptocx *= scalefac; calp->c_ptocy *= scalefac;
	      }
}

int scaleumain(int argc, char *argv[])
{ /* In this mainline, we handle the arguments and then process each file */
	int		retv;
	double		scalefactor;
	static char	*fnames[] = { "USPLINE", NULL } ;
	static char	ftypes[] = { 3, 0 };
	static VSPLINE	*sgsp;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	if ((argc < 2) || !allnumeric(argv[1]))
	  {
		fprintf(stderr, "Usage: %s scalefactor\n", xgprogname(NULL,0));
		return EOK;
	  }
	scalefactor = atof(argv[1]);
	header(stdout, REDUCTION, fnames, ftypes);
	printf(" scale factor to be applied is %6.3f\n", scalefactor);
	if (NULL == (sgsp = gspline(5, &retv)))
	  {
		fprintf(stderr, "Error %d in reading spline\n", retv);
		return EBAD;
	  }
	intoout(scalefactor, sgsp);	/* convert it */
	if ((NULL == gspline(9, &retv)) || (EOK != retv))
	  {
		fprintf(stderr, "Error %d in writing spline\n", retv);
		return EBAD;
	  }
	globalkill();	return EOK;
}

/* end of scaleuca.c */

/* end of caltools.c */
