/* peaks.c -- find and process bright peaks within a single frame of XGEN data

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	18 Oct 2006: eftable: ignore code value -6000 for *efvals,*(efvals+1)
	20 Jul 2006: respond to P_FORCEPHI and P_FORCEOMEGA attributes
	 8 Jun 2006: spindle(),prpeak,logfm,anal_peaks,sift_peaks moved here;
		PA_* flags moved to ../include/spotdefs.h;
		assign_pixels moved to spotmain; other routines reordered
	16 May 2006: merge peakproc.c into peaks.c
	 3 Apr 2006: housekeeping in centroid_peak()
	 1 Apr 2006: eftable(5, and 11): fix firstfound != lfmi->fi_firstfm
	15 Mar 2006: eftable(5...) handles spots midway between 2 images
	 5 Mar 2006: eftable() much more all-encompassing and has more options
	 4 Mar 2006: major revision: managed by spindle values as well as
		frame numbere; Lorentz factors and merging handled better
	 1 Mar 2006: replace effecfram with spindle(); it returns actual phi
	 3 Jan 2006: eftable written: uses effective frame numbers
	 2 Jan 2006: call eftable in cleanup_peaks() to get effective fm #s
	 5 Dec 2005: effecfram() returns effective frame # based on gonio
	17 Jun 2005: in sift_peaks: mrg_peaks() called with spp as 3rd arg
	26 Mar 2003: modify radialdist() so it's sensitive to Lorentz
		correction-- peak2lor() routine written.
	20 Nov 2005: inline function declarations & function prototypes
	28 Oct 2005: replace static globals with elements of SPOTRUN structure
	12 Oct 2005: more rational output from mrg_peaks in verbose mode;
		housekeeping
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	28 Aug 2004: routines used only here are declared as static; cosmetics
	26 Jan 2004: make framename local to get_peaks() (need more!)
	21 Jan 2004: license message inserted and initial comment reordered
	17 Jun 2004: in mrg_peaks, don't overrun array boundaries
	 6 Jan 2004: SPOTRUN structure elements referenced in function args
		& static globals replace extern globals
	13 Dec 2003: remove unused variable res from peak2lor()
	23 Nov 2003: replace globals with elements of SPOTRUN structure;
		change the way we look for headers in peak2lor()
	20 Sep 2003: redo calls for dump_spot(): use the lhd pointer in them;
		fix calls to dump_spot(): proper image header argument
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	 2 May 2003: set `lochp' pointer more explicitly in get_peaks()
	 4 Apr 2003: fix bug in cleanup_peaks() that otherwise ran risk
		of core dumps in for (pkcmp... loop
	 5 Jan 2003: reorganized source-file header
	 1 Nov 2002: reworking of loop limits in dump_spot(), regional_peaks();
		eliminate call to peakskill from within get_peaks on last image
	25 Mar 2002: removed MSI attachments
	30 Sep 1995: user can specify #adu/photon during integration
		therefore photperadu has been renamed to aduperphot
	31 May 1995: "register" -> "register int" throughout;
		ints replace longs
	 2 Mar 1995: allows for frame-specific (X,Y) offsets
		of detector center
	14 Jun 1994: extra diagnostics when reflections are being rejected
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: new capability for eliminating close-together cenTROIDs;
		moved some routines and variables around
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version

   Purpose: finding some bright two-D peaks within a frame of data as part of
	the `spots' functionality and then finding primitive 3-D centroids
	from those; some quality control done in eliminating bad 2-D peaks.
   Cross references: spotmain.c, spotshoes.c

*************************************************************************/
#define		NEED_TIME	1
#include	"spotdefs.h"
#define		CENTERRATIO	(0.7)	/* used in finding dimples */
#define		INCLUDE_IN_HISTOGRAM	254	/* number of histogram bins */
#define		MINSIGMA	3	/* smallest background sigma value */
#define		MINVAR		0.6	/* floor for variance / count */
#define		MINSTAT		(MINSIGMA * MINSIGMA)
#define		MINNEWFRAME	0.1	/* minimum influence of new frame */
#define		BOXSIZE		(BOXHEIGHT * BOXWIDTH)

static void unassign_pixels(SPOTRUN *spp)
{ /*	This sets the peak assignment value to PA_UNASSIGNED for all points
	not permanently excluded by elliptical or rectangular borders. */
	register int	x, y;
	u_short		*pka;

	for (x = 0, pka = spp->spr_peakass; x < spp->spr_detw; x++)
		for (y = 0; y < spp->spr_deth; y++, pka++)
		   if (*pka < PA_PERM_FLAG) *pka = PA_UNASSIGNED;
}

static void kill_corners(SPOTRUN *spp, int xmn, int ymn, int xmx, int ymx)
{ /*	this sets the spp->spr_peakass values at the corners of the box
	bounded by the arguments to PA_REFLCORNER and saves the original
	values so they can be restored later.*/
	int	idely;
	u_short	*pk00, *pk01, *pk10, *pk11;

	idely = ymx - ymn;
	pk00 = spp->spr_peakass + xmn * spp->spr_deth + ymn;
	pk01 = pk00 + idely;
	pk10 = pk00 + (xmx - xmn) * spp->spr_deth;	pk11 = pk10 + idely;
	spp->spr_cor[0] = *pk00;	spp->spr_cor[1] = *pk01;
	spp->spr_cor[2] = *pk10;	spp->spr_cor[3] = *pk11;
	*pk00 = *pk01 = *pk10 = *pk11 = PA_REFLCORNER;
}

static void rest_corners(SPOTRUN *spp, int xmn, int ymn, int xmx, int ymx)
{ /*	This reverses the action of `kill_corners'. */
	int	idely;
	u_short	*pk00, *pk01, *pk10, *pk11;

	idely = ymx - ymn;
	pk00 = spp->spr_peakass + xmn * spp->spr_deth + ymn;
	pk01 = pk00 + idely;
	pk10 = pk00 + (xmx - xmn) * spp->spr_deth;	pk11 = pk10 + idely;
	*pk00 = spp->spr_cor[0];	*pk01 = spp->spr_cor[1];
	*pk10 = spp->spr_cor[2];	*pk11 = spp->spr_cor[3];
}

static void dump_spot(SPOTRUN *spp, PEAK *pkp, FILE *fp, MAIL *lhp, char *rea)
{ /* routine for dumping out the contents of a background-corrected spot */
	int	ix, iy, lx, ly, x, xmn, y, ymn;
	double	lbg, scale, thisf, umax, *tp, tbox[225];
	char	cval, dval;

	x = pkp->pxcen + 0.49999;	y = pkp->pycen + 0.49999;
	xmn = (x < 7 ? 1 : (x < spp->spr_detw - 15 ? x - 7 :
		spp->spr_detw - 15));
	ymn = (y < 7 ? 1 : (y < spp->spr_deth - 15? y - 7 :
		spp->spr_deth - 15));
	umax = -9999.;	lbg = pkp->pbgnd;
	lx = BIGSPOTWID;	ly = BIGSPOTHT;
	for (iy = 0, y = ymn, tp = &(tbox[0]); iy < ly; iy++, y++)
	   {
		for (ix = 0, x = xmn; ix < lx; ix++, x++, tp++)
		   {
			*tp = ((double)imvalue(lhp, x, y)) - lbg;
			if (*tp > umax) umax = *tp;
		   }
	   }
	fprintf(fp,
	 "Refl @[%5.1f,%5.1f,%5.1f],peak=%5.0f,bgnd=%5.1f rejected: %s\n",
	 pkp->pxcen, pkp->pycen, pkp->pphicen, umax, lbg, rea);
	scale = (umax < 1. ? 1. : 62. / umax);
	for (iy = 0, y = ymn, tp = &(tbox[0]); iy < ly; iy++, y++)
	   {
		if (!(y % 3)) fprintf(fp, "%3d", y);
		else fputs("   ", fp);
		fputc((int)'|', fp);
		for (ix = 0, x = xmn; ix < lx; x++, ix++, tp++)
		   {
			thisf = *tp * scale;	cval = thisf + 0.49999;
			if (thisf < -20.) dval = '<';
			else if (thisf < -5.) dval = '!';
			else if (thisf < -0.5) dval = '-';
			else if (thisf <= 0.4) dval = ' ';
			else if (thisf < 1.) dval = '.';
			else if (thisf >= 61.) dval = '*';
			else if (1 <= cval && cval <= 9)
				dval = cval + '0';
			else if (10 <= cval && cval <= 35)
				dval = (cval - 10) + 'a';
			else if (36 <= cval && cval <= 61)
				dval = (cval - 36) + 'A';
			else dval = '*';
			fputc((int)dval, fp);
		   }
		fputs("|\n", fp);
	   }
	fprintf(fp, "   |%4d%4d%4d...|\n", xmn + 3, xmn + 6, xmn + 9);
}

static void headtospp(SPOTRUN *spp)
{ /* This sets some run-dependent parameters based on image headers.
   We don't know when we call this routine which image headers will
   have sensible values in them, so we loop through them till we find one */
	int	i;
	double	cossw, costi, sinsw, sinti;
	MAIL	*lhp;
	DATARUN	*ruhd;

	for (lhp = spp->spr_heads + 1, i = 0; i < spp->spr_nmbox; i++, lhp++)
	   if ((NULL != (ruhd = lhp->runheader)) && (0.1 < ruhd->run_d0pos[0]))
	     {
		spp->spr_plated = ruhd->run_d0pos[0];
		spp->spr_ax0 = ruhd->run_d0pos[1];
		spp->spr_ay0 = ruhd->run_d0pos[2];
		spp->spr_swing = -ruhd->run_d0ang[0] * RADC;
		spp->spr_tilt = -ruhd->run_d0ang[1] * RADC;
		if (!(spp->spr_flags & P_EXTERNSTEP))
			spp->spr_framrad = ruhd->run_stepreal * RADC;
		break;
	    }
	if (i >= spp->spr_nmbox) /* failed to find it: use safe vals */
	  {
		spp->spr_plated = 12.;
		if (!(spp->spr_flags & P_EXTERNSTEP)) spp->spr_framrad = RADC;
		spp->spr_swing = spp->spr_tilt =
			spp->spr_ax0 = spp->spr_ay0 = 0.;
	  }
	cossw = cos(spp->spr_swing * RADC);
	sinsw = sin(spp->spr_swing * RADC);
	costi = cos(spp->spr_tilt * RADC);
	sinti = sin(spp->spr_tilt * RADC);
	spp->spr_pltd2 = spp->spr_plated * spp->spr_plated;
	spp->spr_q21[0][0] = -cossw;
	spp->spr_q21[0][1] = -sinsw *  costi;
	spp->spr_q21[0][2] = -sinsw * sinti;
	spp->spr_q21[1][0] =  sinsw;
	spp->spr_q21[1][1] = -cossw *  costi;
	spp->spr_q21[1][2] = -cossw * sinti;
	spp->spr_q21[2][0] =  0;
	spp->spr_q21[2][1] = -sinti;
	spp->spr_q21[2][2] = costi;
}

int eftable(int opno, SPOTRUN *spp)
{ /*	Maintains a table of spindle values and effective frame numbers.
	opno == 0: delete table.
	opno == 1: create table based on FRAMEINDEX size
	opno == 3: add to table: input a frame index number spp->spr_fmno,
		store the associated spindle value spp->spr_spind.
	opno == 5: query table: input spindle angle value as spp->spr_spind,
		output index in spp->spr_ifrm and spp->spr_rfrm.
		The closest integer to the index value is stored
		in spp->spr_ifrm; the real value is stored in spp->spr_rfrm.
	opno == 011: query table: input index as spp->spr_rfrm,
		output spindle spp->spr_spind and effective fmno spp->spr_efrm.
	opno == 021: query table: input spp->spr_efrm,
		output spp->spr_ifrm, spp->spr_rfrm, spp->spr_spind.
	opno == 041: print current contents of table to logfile.
	Return EOK if successful, EBADVALUE or EALLOC on failures */
	int			curr, doprint, i, iflp, ifn, ioff, next, npr;
	double			delfm, efmax, efmin, fdel2, roff, thisdel;
	static size_t		nelts, oldnelts;
	FRAMEINDEX		*lfmi;
	static int		firstfound, lastfound;
	static double		firstphi, lastphi, *efvals, *flp;
	static size_t		nefvals;
	static char	efmsg[] = "effective frame-number list in eftable";

	if (!opno) /* destroy list */
	  {
		if (NULL != efvals)
		  {
			memzap((AVOID **)&efvals, efmsg);
			nefvals = 0;	firstfound = 9999; lastfound = -9999;
		  }
		return EOK;
	  }
	if (NULL == (lfmi = ftables(3))) return EALLOC; /* need fmindex */
	nelts = lfmi->fi_totfms;	/* number of images */
	if ((NULL == efvals) || (nelts > oldnelts))
	  {
		(void)eftable(0, NULL);
		if (NULL == (efvals = (double *)callop(nelts, sizeof (double),
			efmsg)))
		  {
			(void)eftable(0, NULL);	return EALLOC;
		  }
		for (flp = efvals, i = 0; i < nelts; i++, flp++) *flp = -6000.;
		oldnelts = nelts;	firstfound = 9999; lastfound = -9999;
		firstphi = -9999;	lastphi = 9999;
	  }
	if (opno == 041) /* print out table unless it's boring */
	  {
		if (NULL == spp->spr_fpl) return EOK;
		for (doprint = 0, flp = efvals, ifn = lfmi->fi_firstfm, i = 0;
		 i < nelts; i++, flp++, ifn++)
		   if (*flp > -6000.)
		     {
			iflp = (int)(*flp + 0.4999);
			if (iflp != ifn)
			  {
				doprint = 1;	break;
			  }
		     }
		if (!doprint) return EOK;
		fprintf(spp->spr_fpl, " Table of effective frame numbers:\n");
		fprintf(spp->spr_fpl, "  Index Effective\n Number     Frame\n");
		for (flp = efvals, ifn = lfmi->fi_firstfm, i = npr = 0;
		 i < nelts; i++, flp++, ifn++) if (*flp > -6000.)
		   {
			npr++;
			fprintf(spp->spr_fpl, " %4d %7.2f", ifn, *flp);
			if (npr % 5) fprintf(spp->spr_fpl, "  ");
			else fprintf(spp->spr_fpl, "\n");
		   }
		fprintf(spp->spr_fpl, "\n");
		return EOK;
	  }
	/* all subsequent cases require a reliable stepsize value */
	if (spp->spr_plated < DFUZZ) headtospp(spp);
	delfm = -spp->spr_framrad / RADC;
	if (fabs(delfm) < DFUZZ) delfm = 0.25;
	if ((firstphi > lastphi) && (delfm > 0.)) delfm *= -1.;
	else if (delfm < 0.) delfm *= -1.;
	fdel2 = fabs(delfm) * 0.505;	 /* allow a little slack */
	if (opno == 3) /* use spp->spr_ifrm, spr_spind to add to efvals table*/
	  {
		if ((spp->spr_ifrm < lfmi->fi_firstfm) ||
			(spp->spr_ifrm > lfmi->fi_lastfm)) return EBADVALUE;
		ioff = spp->spr_ifrm - lfmi->fi_firstfm;
		*(efvals + ioff) = spp->spr_spind;
		if (spp->spr_ifrm < firstfound)
		  {
			firstfound = spp->spr_ifrm; firstphi = spp->spr_spind;
		  }
		if (firstfound < 9998) spp->spr_efrm = ((double)firstfound) +
			fabs((spp->spr_spind - firstphi) / delfm);
		if (spp->spr_ifrm > lastfound)
		  {
			lastfound = spp->spr_ifrm; lastphi = spp->spr_spind;
		  }
		return EOK;
	  }
	/* input spp->spr_spind, output spp->spr_rfrm, _ifrm, _efrm */
	if (opno == 5)
	  {
		if (firstphi < lastphi)
		  {
			efmin = firstphi - 0.01; efmax = lastphi + fabs(delfm);
		  }
		else {
			efmin = lastphi - 0.01;	efmax = firstphi + fabs(delfm);
		     }
		spp->spr_efrm = ((double)firstfound) +
		 fabs((spp->spr_spind - firstphi) / delfm);
		if (spp->spr_spind < efmin)
		  {
			spp->spr_rfrm = spp->spr_ifrm =
				firstfound;	return EBADVALUE;
		  }
		if (spp->spr_spind > efmax)
		  {
			spp->spr_rfrm = spp->spr_ifrm = lastfound;
			return EBADVALUE;
		  }
		roff = 1.e30;	iflp = nelts;
		for (flp = efvals, i = 0; i < nelts; i++, flp++)
		 if ((*flp > -5999.) &&
			fabs((thisdel = spp->spr_spind - *flp)) < roff)
		   {
			roff = fabs(thisdel);	iflp = i;
			if (roff < 0.001) break;
		   }
		if ((iflp < nelts) && (roff < fdel2))
		  {
			thisdel = spp->spr_spind - *(efvals + iflp);
			spp->spr_rfrm = ((double)(lfmi->fi_firstfm + iflp)) +
				thisdel / delfm;
			spp->spr_ifrm = spp->spr_rfrm + 0.499999; return EOK;
		  }
		/* we should only get here under weird circumstances */
		spp->spr_ifrm = spp->spr_rfrm = firstfound; return EBADVALUE;
	  }
	if (opno == 011) /* input spp->spr_rfrm, output spp->spr_spind */
	  {
		spp->spr_ifrm = spp->spr_rfrm + 0.49999; /* round integer */
		if (spp->spr_ifrm < firstfound)
		  {
			spp->spr_spind = *efvals;	return EBADVALUE;
		  }
		if (spp->spr_ifrm > lastfound)
		  {
			i = lastfound;	if (i < 1) i = 1;
			spp->spr_spind = *(efvals + i - 1); return EBADVALUE;
		  }
		roff = spp->spr_rfrm - lfmi->fi_firstfm;
		ioff = roff;
		roff -= ioff;
		if ((*(flp = efvals + ioff)) <= -5999.)
		  {
			roff = spp->spr_rfrm - firstfound;
			spp->spr_spind = *efvals + roff * delfm;
			return EBADVALUE;
		  }
		spp->spr_spind = *flp + roff * delfm;
		spp->spr_efrm = ((double)firstfound) +
			fabs((spp->spr_spind - firstphi) / delfm);
		return EOK;
	  }
	if (opno == 021) /* input effective frame #, output all others */
	  {
		spp->spr_spind = delfm * (spp->spr_efrm - (double)firstfound) +
			*efvals;
		return eftable(05, spp);
	  }
	if (opno == 0101)	/* look for contiguity in 1st & last images */
	  {
		spp->spr_flags &= ~P_CONTIG;
		curr = *efvals;	next = *(efvals + 1);
		if ((curr > -5800) && (next > -5800) && (fabs(next - curr)
		 <= fabs(delfm))) spp->spr_flags |= P_CONTIGLEFT;
		if (nelts < 2) return EOK;
		i = nelts - 2;	curr = *(efvals + i); next = *(efvals + i + 1);
		if ((curr > -5800) && (next > -5800) && (fabs(next - curr)
		 <= fabs(delfm))) spp->spr_flags |= P_CONTIGRIGHT;
		return EOK;
	  }
	return EBADVALUE;	/* nonsense argument */
}

static double spindle(int nomframe, MAIL *lhp)
{ /*	This returns the actual spindle angle in degrees that's associated with
	the nominal frame number `nomframe'. The first time we call it within
	an application, it remembers omega, chi, and phi values and the
	frame number of the starting frame; thenceforth it compares the
	remaining frames to that one to calculate the effective frame number.*/
	int			fno, forcemotor, i, limflags;
	double			rfm;
	char			*lonam;
	FRAMEINDEX		*lfmi;
	static double		inig[3], stepsiz;
	static int		inifm;
	static int		moving;
	static DATARUN		*drp;
	static MAIL		lhd;

	/* if the frame index isn't available, give up */
	rfm = nomframe;
	if (NULL == (lfmi = ftables(1))) return rfm;
	limflags = JUST_HEADER;
	if (inifm < 0) return rfm; /* we've gotten a signal not to try */
	forcemotor = 0;
	if ((NULL != lhp) && (lhp->ptrspace1))
	  {
		if (lhp->ptrspace1 & P_FORCEOMEGA) forcemotor = 1;
		else if (lhp->ptrspace1 & P_FORCEPHI) forcemotor = 2;
	  }
	if (!inifm)	/* remember goniostat values & frameno from 1st call*/
	  {
		for (fno = lfmi->fi_firstfm; fno <= lfmi->fi_lastfm; fno++)
		   {
			if ((NULL != (lonam =indexname(fno))) &&
			  (FRAME_OK == read_image(lonam, limflags, &lhd)))
			  {
				if (NULL != (drp = lhd.runheader))
				  {
					for (i = 0; i < 3; i++)
					 inig[i] = drp->run_startpos[i];
					stepsiz = drp->run_stepreal;
					if (forcemotor) moving = forcemotor;
					else moving = drp->run_motor & 02;
					memzap((AVOID **)&drp,
					 "drp in spindle");
					inifm = fno + 1000;	break;
				  }
			  }
		   }
		if (!inifm)
		  {
			inifm = -1;	/* signal not to try this again */
			return rfm; /* give up at this point */
		  }
	  }
	if ((nomframe <= lfmi->fi_firstfm) || (nomframe > lfmi->fi_lastfm))
		return inig[moving];
	 /* this should only happen on firstfm */
	/* we can call this routine either with or without a local header */
	if (NULL == lhp)
	  {
		lhp = &lhd;	drp = lhp->runheader;
		zeromem((AVOID *)lhp, sizeof (MAIL));	lhp->runheader = drp;
		if (NULL == (lonam = indexname(nomframe))) return rfm;
		if (FRAME_OK != read_image(lonam, limflags, lhp)) return rfm;
	  }
	if (fabs(stepsiz) < DFUZZ) stepsiz = 0.25;
	if (NULL == (drp = lhp->runheader)) return rfm / stepsiz;
	return drp->run_startpos[moving];
}

static int centroid_peak(SPOTRUN *spp, MAIL *lhp,
	int x, int y, double bgreg, double sigbg)
{ /*	This computes the centroid of a peak which has its first above-cutoff
	value at (x,y).  It does this in two passes: first, it finds the
	largest value in a small box with (x,y) halfway
	up its left edge.  Then it searches a slightly larger box CENTERED
	on that first estimate to determine a centroid of all the
	background-corrected data.  As it does so it blots out the pixels
	above background within the larger box.  If the total intensity of
	that peak is above a cutoff, the peak is `saved', i.e. the PEAKS
	pointer is advanced.  The routine returns the largest value of y
	examined within this peak; the calling routine can use this
	information to advance its y pointer. */
	register int	ix, iy;
	int		corrinten, icurr, iwork, lx, ly, nused, pknum;
	int		xmn, xmx, xwork, ymn, ymx, ywork;
	float		locax, locay;
	double		avebox, cenint, center5, xcn, ycn;
	u_short		*pka;
	PEAK		*lpk;

	/* We do the initial search over a 3x3 to 7x7 box depending on
	the value of "samepeaksq".  The box is placed such that
	the point we've input is halfway down the left edge. */
	xmx = x + spp->spr_smallwidth;
	if (xmx > spp->spr_detw - 2) xmx = spp->spr_detw - 2;
	ymn = y - spp->spr_smlhalf;	ymx = y + spp->spr_smallheight;
	if (ymn < 1) ymn = 1;
	if (ymx > spp->spr_deth - 2) ymx = spp->spr_deth - 2;
	iwork = imvalue(lhp, x, y);
	for (ix = x; ix <= xmx; ix++)
	   {
		pka = spp->spr_peakass + ix * spp->spr_deth + ymn;
		for (iy = ymn; iy <= ymx; iy++, pka++)
		   if ((*pka == PA_UNASSIGNED) &&
			(icurr = imvalue(lhp, ix, iy)) > iwork)
		     {
			x = ix;	y = iy;	iwork = icurr;
		     }
	   }
	/* better centroid estimate taken from all pixels around first one ...
	first, find the limits of the box we'll sum over. */
	xmn = x - spp->spr_smlhalf;	ymn = y - spp->spr_smlhalf;
	xmx = xmn + spp->spr_smallwidth;
	ymx = ymn + spp->spr_smallheight;
	if (xmx > spp->spr_detw - 2) xmx = spp->spr_detw - 2;
	else if (xmn < 1) xmn = 1;
	if (ymx > spp->spr_deth - 2) ymx = spp->spr_deth - 2;
	else if (ymn < 1) ymn = 1;
	kill_corners(spp, xmn, ymn, xmx, ymx); /* flag the corners of the box */
	/* if the center 5 elements of the peak have an average intensity
	 smaller than CENTERRATIO * whole-box average,
	 we flag the refl as 2-humped */
	if (x < 1) x = 1;
	else if (x > spp->spr_detw - 2) x = spp->spr_detw - 2;
	if (y < 1) y = 1;
	else if (y > spp->spr_deth - 2) y = spp->spr_deth - 2;
	center5 = (imvalue(lhp, x-1, y) + imvalue(lhp, x, y) +
		imvalue(lhp, x+1, y) + imvalue(lhp, x, y-1) +
		imvalue(lhp, x, y+1)) - 5. * bgreg;
	/* now we sum the pixels that haven't been assigned to other refls &
	 assign the relatively bright pixels in the peak to the peak */
	pknum = spp->spr_peakp - spp->spr_pk0infm;	/* the # of the peak */
	xwork = ywork = iwork = nused = 0;
	for (lx = 0, ix = xmn; ix <= xmx; lx++, ix++)
	   {
		pka = spp->spr_peakass + ix * spp->spr_deth + ymn;
		for (ly = 0, iy = ymn; iy <= ymx; ly++, iy++, pka++)
		   if (*pka == PA_UNASSIGNED)
		     {
			corrinten = imvalue(lhp, ix, iy) - bgreg;
			iwork += corrinten;		nused++;
			xwork += corrinten * lx;	ywork += corrinten * ly;
	/*if pixel is at least sigma above background, assign it to this refl*/
			if (corrinten > sigbg) *pka = pknum;
		     }
	   }
	rest_corners(spp, xmn, ymn, xmx, ymx);
	if (!nused || iwork < 1) return ymx; /* too weak */
	lpk = spp->spr_peakp;	lpk->pncells = nused;
	xcn = (double)xwork;	ycn = (double)ywork;
	lpk->pinten = cenint = (double)iwork; avebox = cenint / (double)nused;
	lpk->pxcen = (double)xmn + xcn / cenint;
	lpk->pycen = (double)ymn + ycn / cenint;
	lpk->pbgnd = bgreg + 0.49999;
	lpk->pflags = P_USABLE;
	lpk->ptroi.nobk = 1;	lpk->ptroi.obparts[0] = 1.;
	if (spp->spr_plated < FUZZ) headtospp(spp);
	if (spp->spr_flags & P_FORCEPHI) lhp->ptrspace1 = P_FORCEPHI;
	else if (spp->spr_flags & P_FORCEOMEGA) lhp->ptrspace1 = P_FORCEOMEGA;
	spp->spr_spind = lpk->pphicen = spindle(lhp->frameno, lhp);
	spp->spr_rfrm = spp->spr_ifrm = lpk->ptroi.obkmin = lhp->frameno;
	if (EOK != eftable(3, spp))
	  {
		fprintf(stderr,
		 "centroid_peak: eftable failure at [%7.2f,%7.2f,%7.2f]\n",
		 lpk->pxcen, lpk->pycen, lpk->pphicen);
	  }
	if ((center5 / 5.) < CENTERRATIO * avebox)
	  {
		if (spp->spr_flags & P_VERBOSE2) dump_spot(spp, spp->spr_peakp,
			spp->spr_fpl, lhp, "P_XYDIMPLED in peaksspp, ");
		lpk->pflags |= P_XYDIMPLED;
	  }
	/* correct for the frame-specific detector offsets. We do this
	 AFTER dumping the spot so the dump happens in the right place. */
	if (spp->spr_flags & P_FRAMEOFFSETS)
	  {
		locax = lpk->pxcen;	locay = lpk->pycen;
		adjust_to_fm(&locax, &locay, (int)(lhp->frameno), 1);
		lpk->pxcen = locax;	lpk->pycen = locay;
	  }
	(spp->spr_peakp)++;	/* advance pointer for the peak */
	return ymx;
}

static double excludemean(int *bp, int np, int nthrow)
{ /*	This does a calculation of the mean of a set of integer values,
	but it deliberately throws out the `nthrow' largest values in the set.
	If bp has < nthrow elements then it calculates an ordinary mean.
	Oct 2005: special circumstance: if called with bp == NULL, that's
	a signal that we should free the memory for the hv vector */
	int		sumu, *vp, *vpe, **hve, **hvq, **tp, **tpn;
	static int	nthprev = 0;
	static int	**hv;

	if (NULL == bp)
	  {
		if (NULL != hv) memzap((AVOID **)&hv, "hv in spots");
		return 0;
	  }
	if (np <= 0) return 0.;	/* shouldn't ever happen anyway */
	/* calculate the ordinary mean of these values */
	for (sumu = 0, vp = bp, vpe = vp + np; vp < vpe; vp++) sumu += *vp;
	if (nthrow >= np) return ((double)sumu) / (double)np;
	/* reserve memory for the list of largest values in the list */
	if ((NULL == hv) || (nthrow > nthprev))
	  {
		if (NULL != hv) memzap((AVOID **)&hv, "hv in excludemean");
		if (NULL == (hv = (int **)callop(nthrow, sizeof (int *),
		 "hv in excludemean"))) return ((double)sumu) / (double)np;
		nthprev = nthrow;
	  }
	/* load pointers to the 1st `nthrow' points into the `hv' list in
	 increasing order,
	 i.e. they'll end up sorted from smallest to largest. */
	for (vp = bp, hvq = hv + nthrow, hve = hv; hve < hvq; vp++, hve++)
	   {
		for (tp = hve - 1, tpn = hve; (tpn > hv) && (**tp > *vp);
		 tp--, tpn--) *tpn = *tp;
		*tpn = vp;
	   }
	hve = hvq - 1;
	/* now look at all the others.  For each value we check to see if
	 it's smaller than the largest value in the large-value list. If it
	 is, we insert it in its proper place in the large-value list. */
	for ( ; vp < vpe; vp++)
	   {
		if (*vp > **hv)
		  {
			for (tp = hv, tpn = hv + 1; (tp < hve) &&
				(**tpn < *vp); tp++, tpn++) *tp = *tpn;
			*tp = vp;
		  }
	   }
	/* now the hv vector contains the `nthrow' largest values in the
	 overall list.  We subtract the sum of the elements of that vector
	 from the overall sum and return the excluded mean based on that.*/
	for (tp = hv; tp <= hve; tp++) sumu -= (**tp);
	return ((double)sumu) / ((double)(np - nthrow));
}

static int regional_peaks(SPOTRUN *spp, MAIL *lhp, int xleft, int ytop)
{ /*	This finds the peaks in a region whose upper lefthand corner is at
	(xleft,ytop) with width BOXWIDTH and height BOXHEIGHT.
	The background is taken as the average background over the
	already-processed frames if we've processed at least two;
	otherwise it's taken as the minimum of the mean and the modal
	intensity over a somewhat larger box.
	The cutoff intensity for counting a pixel as being part of a
	possible peak is set to `spp->spr_nsigma' times the standard deviation
	of the background value. */
	register int	i, x, y, usedcells;
	int		bg2, bht, coff, ngood, ninbox, ureg_ind, uri2;
	int		xmx, xreg_ind, ymx, yreg_ind, *bgbox, *bgp, *bp, *bpy;
	unsigned	nboxes;
	double		bgnd_delta, bgnd_region, boxmean, boxunblot, boxvar;
	double		newmean, oldmed, oldvar, sig_bgnd, wtcur, wtsum, xmean;
	register PEAK	*localpk, *pk0inbox;
	u_short		*pka, *pkac, *pkac0;

	xreg_ind = xleft / BOXWIDTH;	yreg_ind = ytop / BOXHEIGHT;
	ureg_ind = xreg_ind * spp->spr_nboxony + yreg_ind;
	nboxes = spp->spr_nboxonx * spp->spr_nboxony;
	if ((unsigned)ureg_ind >= nboxes) return 0;
	/* There aren't any data in the three boxes closest to each corner */
	if (xreg_ind == 0 || xreg_ind == (spp->spr_nboxonx - 1))
	  {
		if (yreg_ind < 2 || yreg_ind >= (spp->spr_nboxony - 2))
		  {
			if (spp->spr_flags & P_VERBOSE2) fprintf(spp->spr_fpl,
			 " %5d %4d: close to a corner\n", xleft, ytop);
			return 0;
		  }
	  }
	else if (xreg_ind == 1 || xreg_ind == (spp->spr_nboxonx - 2))
	  {
		if (yreg_ind == 0 || yreg_ind == (spp->spr_nboxony - 1))
		  {
			if (spp->spr_flags & P_VERBOSE2) fprintf(spp->spr_fpl,
			 " %5d %4d: close to a corner\n", xleft, ytop);
		  }
		return 0;	/* need at least 200 points */
	  }
	xmx = xleft + BOXWIDTH;	if (xmx > spp->spr_detw) xmx = spp->spr_detw;
	ymx = ytop + BOXHEIGHT;	if (ymx > spp->spr_deth) ymx = spp->spr_deth;
	bht = ymx - ytop;		ninbox = (xmx - xleft) * bht;
	if (ninbox < 200)
	  {
		if (spp->spr_flags & P_VERBOSE2) fprintf(spp->spr_fpl,
		 " %5d %4d: ninbox = %d\n", xleft, ytop, ninbox);
		return 0;	/* need at least 200 points */
	  }
	/* load in the data from the current box in two places: an array that
	knows about position (for the later determination of the peaks) and
	another that doesn't care about position (for background estimates)*/
	if (NULL == spp->spr_curbox)
	  {
		if (NULL == (spp->spr_curbox = (int *)callop(2 * BOXSIZE,
		 sizeof (int), "spp->spr_curbox in regional_peaks")))
		  {
			fprintf(stderr,
			 "Cannot allocate memory for regional-peak box array");
			return 0;
		  }
	  }
	bgbox = spp->spr_curbox + BOXSIZE;
	pkac0 = spp->spr_peakass + xleft * spp->spr_deth + ytop;
	for (x = xleft, bp = spp->spr_curbox, bgp = bgbox, pkac = pkac0;
	 x < xmx; x++, pkac += spp->spr_deth)
	   {
		for (y = ytop, pka = pkac; y < ymx; y++, bp++, pka++)
		   if (*pka == PA_UNASSIGNED)
			*bgp++ = *bp = imvalue(lhp, x, y);
	   }
	/* we need at least 60 good points in this region to use it */
	if ((ngood = bgp - bgbox) < 60)
	  {
		if (spp->spr_flags & P_VERBOSE2) fprintf(spp->spr_fpl,
			" %5d %4d: ngood = %d\n", xleft, ytop, ngood);
		return 0;
	  }
	/* get the initial bgnd estimate from medians or old data */
	oldmed = spp->spr_regbg[ureg_ind];
	if (0. >= (xmean = excludemean(bgbox, ngood, 20)))
	  {
		if (spp->spr_flags & P_VERBOSE2) fprintf(spp->spr_fpl,
		 " %5d %4d: excludemean returns %9.4f\n", xleft, ytop, xmean);
		return 0;
	  }
	wtcur = spp->spr_nfmsofar < 8 ?
		1. - 0.1 * ((double)spp->spr_nfmsofar) : 0.2;
	(spp->spr_regbg)[ureg_ind] = bgnd_region =
		xmean * wtcur + oldmed * (1. - wtcur);
	sig_bgnd = sqrt(MINVAR * bgnd_region);
	if (sig_bgnd < 1.) sig_bgnd = 1.;
	if (spp->spr_flags & P_VERBOSE2) fprintf(spp->spr_fpl,
	 " %5d %4d      %5.1f        %5.1f",
		xleft, ytop, bgnd_region, sig_bgnd);
	/* compute the cutoff value from this */
	coff = bgnd_region + (double)spp->spr_nsigma * sig_bgnd *
		sqrt(spp->spr_aduperph) + 0.49999;
	if (spp->spr_flags & P_VERBOSE2) fprintf(spp->spr_fpl, " %6d", coff);
	pk0inbox = spp->spr_peakp;
	/* Examine the `spp->spr_curbox' array to find the peaks in the box */
	for (x = xleft, bpy = spp->spr_curbox, pkac = pkac0; x < xmx;
	 x++, bpy += bht, pkac += spp->spr_deth)
	   {
		for (y = ytop, bp = bpy, pka = pkac; y < ymx; y++,bp++,pka++)
		 if ((*bp > coff) && (*pka == PA_UNASSIGNED))
		   {
			i = centroid_peak(spp, lhp, x, y,
				bgnd_region, sig_bgnd);
			bp += (i - y);
			if ((spp->spr_curbox > bp) ||
			 (bp >= (spp->spr_curbox + 2 * BOXSIZE)))
			  {
				fprintf(spp->spr_fpl,
	 "Warning! peak search box out of bounds for peak [%4d,%4d]\n", x, y);
			  }
			pka += (i - y);			y = i;
		   } /* end of "for (y = ytop" loop */
	   } /* end of "for (x = xleft" loop */
	/* Now that we have found the peaks, we can improve the box's
	background estimates.  We first compute a background value equal
	to the mean of the unassigned pixels in this frame and a
	sigma equal to the standard deviation of that mean.  Then if we
	aren't in the initial frame we average that background with what's
	already in spp->spr_regbg[].  The resulting background value is
	used to correct the intensities and estimate the standard deviations
	of the peaks in the region.  */
	bg2 = 2. * bgnd_region + 0.49999;	bpy = bgp;
	for (boxmean = usedcells = 0, bgp = bgbox; bgp < bpy; bgp++)
		if (*bgp < bg2)
		  {
			boxmean += *bgp; usedcells++;
		  }
	if (usedcells > 50)
	  {
		boxunblot = usedcells; boxmean /= boxunblot;
		/* combine this estimate with the old values if there are any */
		if (spp->spr_nfmsofar)
		  {
			oldvar = (double)(sig_bgnd * sig_bgnd) /
				(double)spp->spr_nfmsofar;
			boxvar = (boxmean > 0 ? MINVAR * boxmean : 1);
			if (boxvar > oldvar / MINNEWFRAME)
				oldvar = boxvar * MINNEWFRAME;
			wtsum = 1. / boxvar + 1. / oldvar;
			newmean = (boxmean / boxvar + oldmed / oldvar) /wtsum;
			boxvar = (1. + boxmean * boxmean / boxvar + oldmed *
			  oldmed / oldvar) /wtsum - newmean * newmean;
			if (boxvar < MINVAR * newmean)
				boxvar = MINVAR * newmean;
			if (spp->spr_flags & P_VERBOSE2) fprintf(spp->spr_fpl,
			 " %7.1f %7.1f %6.1f\n", boxmean, newmean, boxvar);
			boxmean = newmean;
		  }
		else {
			boxvar = (boxmean > 0 ? MINVAR * boxmean : 1);
			if (spp->spr_flags & P_VERBOSE2) fprintf(spp->spr_fpl,
			 " %7.1f %7.1f %6.1f\n", boxmean, boxmean, boxvar);
		     }
		bgnd_delta = boxmean - bgnd_region;
		/* store the updated background estimate */
		uri2 = xreg_ind * spp->spr_nboxony + yreg_ind;
		if ((uri2 < 0) || (uri2 >= (int)nboxes)) fprintf(spp->spr_fpl,
 "Can't do updating in region (%4d,%4d): box index %d >= nboxes = %d\n",
			 xreg_ind, yreg_ind, uri2, nboxes);
		else (spp->spr_regbg)[uri2] = boxmean;
	  }
	else
	 {
		bgnd_delta = 0; boxmean = bgnd_region;
		boxvar = sig_bgnd * sig_bgnd;
		if (spp->spr_flags & P_VERBOSE2) fprintf(spp->spr_fpl,
		 " %7.1f %7.1f %6.1f\n", boxmean, boxmean, boxvar);
	 }
	/*correct intensity measurements for peaks found in this region
	and compute the background and variance values */
	for (localpk = pk0inbox; localpk < spp->spr_peakp; localpk++)
	   {
		localpk->pinten -= bgnd_delta * localpk->pncells;
		localpk->pbgnd = boxmean; /* note: this is PER PIXEL!! */
		localpk->psigma = sqrt(localpk->pinten +
			localpk->pncells * (localpk->pbgnd + boxvar) +
			(localpk->pinten/40.)*(localpk->pinten/40.));
		if (localpk->pinten < spp->spr_iosmin * localpk->psigma)
		  { /* flag weak refls and set their phi values 'way up */
			localpk->pflags |= P_TOOWEAK;
			if (spp->spr_flags & P_VERBOSE2)
			 dump_spot(spp, localpk, spp->spr_fpl, lhp,
				"P_TOOWEAK in peaks");
			localpk->pphicen = HUGEPHI; localpk->pxcen = HUGEX;
		  }
	   }
	return (int)(spp->spr_peakp - pk0inbox); /* return # of found peaks */
}

int peakskill(int irv, SPOTRUN *spp)
{ /*	This frees memory allocated specifically within `peaks.c'.
	 It's not a geographical feature in New York state.
	 It returns its argument */
	(void)excludemean(NULL, 0, 0); /* elim memory for meanval pointers */
	memzap((AVOID **)&(spp->spr_peakass), "spp->spr_peakass in spots");
	memzap((AVOID **)&(spp->spr_regbg), "spp->spr_regbg in spots");
	memzap((AVOID **)&(spp->spr_curbox), "spp->spr_curbox in spots");
	return irv;
}

static int phicmp(PEAK *pk0, PEAK *pk1) /* compare routine for image number */
{
	return ((pk0->ptroi).frame0 <= (pk1->ptroi).frame0 ? -1 : 1);
}

static int peak2lor(SPOTRUN *spp, PEAK *pkp)
{ /*	This obtains the square of the local Lorentz correction based on a
	PEAK structure.  If we mess up because the pixel-to-cm calibration
	fails, we just take a wild guess. Upper limit is spp->spr_maxlor.
	Change as of Nov 03: we don't know which of the headers is already
	populated, so we search through all of them to try to find one with
	a DATARUN structure that's nonzero. 
	We stuff this value eventually in pkp->plor.
	The routine itself returns EOK if successful, various errors if not. */
	double	lsv1, rni, rni2, xin, yin, *flp, slo[3];
	float	lxcm, lycm;

	pkp->plor = spp->spr_maxlor + 100;
	if (NULL == spp->spr_spl) return 1; /* we expect to get spline early */
	/* we attempt to read plateD, swing, etc. if necessary */
	if (FUZZ > spp->spr_plated) headtospp(spp);
	if (0 >= pix2cm(pkp->pxcen, pkp->pycen, &lxcm, &lycm, 0)) return 2;
	xin = lxcm - spp->spr_ax0;
	yin = lycm - spp->spr_ay0; /* correct main beam */
	rni2 = xin * xin + yin * yin + spp->spr_pltd2;	rni = sqrt(rni2);
	slo[0] = spp->spr_plated / rni;	slo[1] = xin / rni;
	slo[2] = yin / rni;
	flp = &(spp->spr_q21[1][0]);
	lsv1 = *flp * slo[0] + *(flp + 1) * slo[1] + *(flp + 2) * slo[2];
	lsv1 *= lsv1;
	/* if we want to normalize this to, i.e. return (|sy| / |s|)^2
	instead of |sy|^2, then add this:
	flp = &(spp->spr_q21[0][0]);
	xin = 1. + *flp * slo[0] + *(flp + 1) * slo[1] + *(flp + 2) * slo[2];
	flp = &(spp->spr_q21[2][0]);
	yin = *flp * slo[0] + *(flp + 1) * slo[2] + *(flp + 2) * slo[2];
	lsv1 /= (xin * xin + lsv1 + yin * yin);
	*/
	pkp->plor = (lsv1 < 1. /spp->spr_maxlor) ? spp->spr_maxlor : 1. / lsv1;
	return EOK;
}

static double fracbig(TROID *tk)
{ /*	This returns the fraction of the intensity in the TROID *tk that
	is contained in its brightest frame */
	int	ifm;
	double	tmax, tsum, tthis;

	for (ifm = 1, tsum = tmax = tk->obparts[0]; ifm < tk->nobk; ifm++)
	   {
		if ((tthis = tk->obparts[ifm]) > tmax) tmax = tthis;
		tsum += tthis;
	   }
	return ((tsum > DFUZZ) && (tsum > tmax)) ? tmax / tsum : 1.;
}

double radialdist(PEAK *pk0, PEAK *pk1, SPOTRUN *spp)
{ /*	This returns the radial distance between peaks pk0 and pk1. A scale
	factor is incorporated to relate phi to x and y.  In the special case
	in which pk0 and pk1 are almost exactly one frame apart, their (X,Y)
	centroids are nearby, and almost all the intensity in each unmerged
	peak is contained in the principal frame, we assume that the real
	centroid lies between the two and therefore the radial distance in
	omega between 'em is therefore much smaller than we'd otherwise think*/
	double	delx, delxy, dely, delz, rawz;

	delx = pk0->pxcen - pk1->pxcen;	dely = pk0->pycen - pk1->pycen;
	delxy = delx * delx + dely * dely;
	rawz = pk0->pphicen - pk1->pphicen;	rawz *= rawz;
	/* March 2003: apply Lorentz correction, using the average
	 of the Lorentz values for the two peaks */
	delz = rawz * 2. / (pk0->plor + pk1->plor);
	if ((delxy < 2. * spp->spr_samepeaksq) && (rawz < 1.1) &&
	 (fracbig(&(pk0->ptroi)) > 0.92) && (fracbig(&(pk1->ptroi)) > 0.92))
	 	delz *= 0.01;
	return spp->spr_phitoxscale * delz + delxy;
}

static void prpeak(FILE *fp, PEAK *pkp)
{ /*	This does a quickie printout of a PEAK structure.
	If called with NULL as the file argument, it zeros alreadyhead */
	int		ifm;
	float		*op;
	TROID		*tp;
	static int	alreadyhead;

	if (NULL == fp)
	  {
		alreadyhead = 0;	return;
	  }
	tp = &(pkp->ptroi);
	if (!alreadyhead)
	  {
		fprintf(fp,
 " Xcenter Ycenter Zcenter  Intensity    Sigma Backgnd #pixels");
		fprintf(fp, " flags #frames frame-fractional intensities\n");
		alreadyhead = 1;
	  }
	fprintf(fp, " %7.1f %7.1f %7.1f %10.1f %8.1f %7.2f %7d %5o %7d",
	 pkp->pxcen, pkp->pycen, pkp->pphicen, pkp->pinten, pkp->psigma,
	 pkp->pbgnd, pkp->pncells, pkp->pflags, tp->nobk);
	for (op = &(tp->obparts[0]), ifm = 0; ifm < tp->nobk; ifm++, op++)
		fprintf(fp, "%5.2f", *op);
	fprintf(fp, "\n");
}

static void logfm(FILE *fp, int cfm, int ninfm, int nstr, int nrep, int stfm)
{ /* This prints a message to the stream fp after a frame's processing*/
	char		termc;

	termc = ((stfm < 0) || ((cfm + 1 - stfm) % 3 == 0)) ? '\n' : ' ';
	fprintf(fp, "%4d %6d %6d %7d%c", cfm, ninfm, nstr, nrep, termc);
}

void anal_peaks(SPOTRUN *spp, int fail[])
{ /*	This analyzes the peaks and provides a frame-by-frame summary
	of the output reflections */
	int	curfm, newfm, ninfm, nrep, nstrong, region, rf, rfla, stfm;
	int	ninr[NREGIONS];
	double	iosd, iosn;
	PEAK	*curpk, *first;
	static char	summsg[] =
		"Frame-by-frame summary of output reflections:\n";
	static char	headmsg[] = " Frm # Reflections #Models%c";
	static char	fmsg[] = " No. Output >%3.0f*s  Helped%c";

	for (curpk = spp->spr_pkbase; curpk < spp->spr_pklast; curpk++)
	   {
		if (!(rfla = curpk->pflags)) break;
		for (rf = 0; rf < 14; rf++) if (rfla == (1 << rf)) break;
		if (rf < 14) (fail[rf+1])++;
	   }
	if (curpk >= spp->spr_pklast) return;
	first = curpk;
	if (spp->spr_flags & P_VERBOSEANY)
	  {
		fprintf(spp->spr_fpl, "Valid peaks after final revisions:\n");
		prpeak(NULL, NULL);
		for ( ; curpk < spp->spr_pklast; curpk++)
		   if (!curpk->pflags) prpeak(spp->spr_fpl, curpk);
	  }
	/* establish the mean I/sigma value for diagnostics */
	for (iosn = iosd = 0., curpk = first; curpk < spp->spr_pklast; curpk++)
	   if (!curpk->pflags) 
	     {
		iosd += 1.;
		if (curpk->psigma > DFUZZ)
			iosn += curpk->pinten / curpk->psigma;
	     }
	if (iosd > DFUZZ) iosn /= iosd;	else return;
	ninfm = nstrong = 0;	zeromem((void *)ninr, NREGIONS * sizeof (int));
	fprintf(spp->spr_fpl, summsg);	printf(summsg);
	for (stfm = 0; stfm < 3; stfm++)
	   {
		fprintf(spp->spr_fpl, headmsg, (stfm < 2) ? ' ' : '\n');
		printf(headmsg, (stfm < 2) ? ' ': '\n');
	   }
	for (stfm = 0; stfm < 3; stfm++)
	   {
		fprintf(spp->spr_fpl, fmsg, iosn, stfm < 2 ? ' ': '\n');
		printf(fmsg, iosn, (stfm < 2) ? ' ': '\n');
	   }
	stfm = curfm = (first->ptroi).frame0;
	for (curpk = first ; curpk < spp->spr_pklast; curpk++)
	 {
	   if (!(rfla = curpk->pflags))
	     {
		newfm = (curpk->ptroi).frame0;	(fail[0])++;
		if (newfm > curfm) /* add additional frame to the list */
		  {
			for (nrep = region = 0; region < NREGIONS; region++)
				if (ninr[region]) nrep++;
			logfm(spp->spr_fpl, curfm, ninfm, nstrong, nrep, stfm);
			logfm(stdout      , curfm, ninfm, nstrong, nrep, stfm);
			curfm = newfm;	ninfm = nstrong = 0;
			zeromem((void *)ninr, NREGIONS * sizeof (int));
		  }
		ninfm++;
		if (curpk->pinten > 3. * curpk->psigma)
		  {
			region = getregion((short)(curpk->pxcen), (short)
			 (curpk->pycen), /* curpk->pphicen, */ spp->spr_bdr);
			(ninr[region])++;
			if (curpk->pinten > iosn * curpk->psigma) nstrong++;
		  }
	     }
	 else {
		for (rf = 0; rf < 14; rf++) if (rfla == (1 << rf)) break;
		if (rf < 14) (fail[rf+1])++;
	      }
	 }
	for (nrep = region = 0; region < NREGIONS; region++)
		if (ninr[region]) nrep++;
	logfm(stdout, curfm, ninfm, nstrong, nrep, -1); /* finish listing */
	logfm(spp->spr_fpl, curfm, ninfm, nstrong, nrep, -1);
}

static void mrg_peaks(PEAK *pk0, PEAK *pk1, SPOTRUN *spp, int whichmerge)
{ /*	This merges the contents of peak 1 into peak 0.
	New arg as of 12 Oct 05: `whichmerge' determines which kind of merge
	we're doing (0 from cleanup_peaks(), 1 from sift_peaks()) */
	int		dfm, fmcen, fmmax, fmmin;
	int		ifm, ifmmin, jfm, maxi, maxl, tfm;
	double		inte, iq0, iq1, minten, outside;
	TROID		*pte, *ptl, *ptr0, *ptr1;
	float		lprof[2*TROIDWIDTH];
	static int	maxxok, maxyok, minxok, minyok;
	static MAIL	*lhp;

	ptr0 = &(pk0->ptroi);	ptr1 = &(pk1->ptroi);
	iq0 = pk0->pinten;	iq1 = pk1->pinten;
	minten = iq0 + iq1;
	if (spp->spr_flags & P_VERBOSEANY) fprintf(spp->spr_fpl,
 "%s [%6.1f,%6.1f,%7.2f], I=%7.0f into [%6.1f,%6.1f,%7.2f], I=%7.0f\n",
	 (whichmerge) ? "Sifting" : "Merging",
	  pk1->pxcen, pk1->pycen, pk1->pphicen, pk1->pinten,
	  pk0->pxcen, pk0->pycen, pk0->pphicen, pk0->pinten);
	pk0->pxcen = (pk0->pxcen * iq0 + pk1->pxcen * iq1) / minten;
	pk0->pycen = (pk0->pycen * iq0 + pk1->pycen * iq1) / minten;
	pk0->pphicen = (pk0->pphicen * iq0 + pk1->pphicen * iq1) / minten;
	ptr0->frame0 = (ptr0->frame0 * iq0 + ptr1->frame0 * iq1) / minten;
	pk0->pinten = minten;	/*total intensity = sum of frame intensities*/
	pk0->pbgnd = pk0->pbgnd * pk0->pncells + pk1->pbgnd * pk1->pncells;
	pk0->pncells += pk1->pncells;	/* total # of cells */
	pk0->pbgnd /= (double)pk0->pncells;	/* background per frame */
	pk1->pphicen = HUGEPHI;	/* so that the garbage peak doesn't get used*/
	ptr1->frame0 = HUGEPHI;	/* so that the garbage peak doesn't get used*/
	pk0->psigma = sqrt(pk0->psigma * pk0->psigma +
		pk1->psigma * pk1->psigma);
	/* merge the intensities within the frames carefully.  We put
	 the real intensities results into the array `lprof' and,
	 if necessary, pick the strongest segment of it.  If the whole
	 thing is too broad already we reject the merged reflection. */
	if (ptr1->obkmin < ptr0->obkmin)
	  {
		pte = ptr1; ptl = ptr0; inte = iq0; iq0 = iq1; iq1 = inte;
	  }
	else {
		pte = ptr0;	ptl = ptr1;
	     }
	maxi = pte->nobk;	maxl = 2 * TROIDWIDTH;
	if (maxi > maxl) maxi = maxl;
	dfm = ptl->obkmin - pte->obkmin;
	if (dfm > maxl - 1) dfm = maxl - 1;
	tfm = dfm + ptl->nobk;
	 /* special case: earlier peak is already so big that the later
	 one is entirely contained within it: */
	if (tfm < pte->nobk) tfm = pte->nobk;
	for (ifm = 0; ifm < maxl; ifm++) lprof[ifm] = 0.;
	for (ifm = 0; ifm < maxi; ifm++) lprof[ifm] = pte->obparts[ifm] * iq0;
	maxi = ptl->nobk;
	if (maxi > maxl) maxi = maxl;
	for (ifm = 0, jfm = dfm; ifm < maxi; ifm++, jfm++)
		lprof[jfm] += ptl->obparts[ifm] * iq1;
	spp->spr_spind = pk0->pphicen;
	if (EOK != eftable(5, spp))
	  {
		fprintf(stderr,
 "mrg_peaks failed to convert spindle value at [%7.2f,%7.2f,%7.2f]\n",
			pk0->pxcen, pk0->pycen, pk0->pphicen);
		return;
	  }
	fmcen = spp->spr_ifrm;
	/* if we have more than TROIDWIDTH frames, we take the TROIDWIDTH
	 frames closest to the centroid as the relevant ones.  If the
	 intensity outside that group is too large a percentage of the
	 total, we reject the reflection */
	if (tfm > TROIDWIDTH)
	  { /* establish real-frame range of output centroids */
		fmmax = fmcen + 5;	fmmin = fmcen - 5;
		/* if we're beyond the end, back up */
		if ((dfm = fmmax - (ptl->obkmin +ptl->nobk)) > 0) fmmin -= dfm;
		/* if instead we're before the beginning, move forward */
		else if ((dfm = pte->obkmin - fmmin) > 0) fmmin += dfm;
		/* now establish index range */
		ifmmin = fmmin - pte->obkmin;
		/* check the part of the profile outside the center group */
		for (outside = ifm = 0; ifm < ifmmin; ifm++)
			outside += lprof[ifm];
		for (ifm = ifmmin + TROIDWIDTH; ifm < tfm; ifm++)
			outside += lprof[ifm];
		/* if the outer elements are > 10% of the total, we reject */
		if (outside > 0.1 * minten)
		  {
			pk0->pflags |= P_BADPHISHAPE;
			pk1->pflags |= P_BADPHISHAPE;
		  }
		else pk0->pflags &= ~P_BADPHISHAPE;
		/* put the re-normalized profile back in place */
		for (ifm = 0, dfm = ifmmin; ifm < TROIDWIDTH; ifm++, dfm++)
			pk0->ptroi.obparts[ifm] = lprof[dfm] / minten;
		pk0->ptroi.obkmin = fmmin;
		pk0->ptroi.nobk = TROIDWIDTH;
	  }
	else {
		for (ifm = 0; ifm < tfm; ifm++)
			pk0->ptroi.obparts[ifm] = lprof[ifm] / minten;
		pk0->ptroi.obkmin = pte->obkmin;
		pk0->ptroi.nobk = tfm;
	     }
	if (0 == maxxok)	/* make sure boundaries are defined */
	  {
		minxok = minyok = DETECTOR_EDGE;
		maxxok = spp->spr_detw - minxok;
		maxyok = spp->spr_deth - minyok;
	  }
	if ((pk0->pxcen < minxok) || (pk0->pxcen >= maxxok) ||
	  (pk0->pycen < minyok) || (pk0->pycen >= maxyok))
	  {
		if (spp->spr_flags & P_VERBOSE2)
		  {
			tfm = pk0->pphicen;
			lhp = spp->spr_heads + 1 +
				(tfm % (spp->spr_nmbox - 1));
			dump_spot(spp, pk0, spp->spr_fpl, lhp,
				"P_XYEDGE in peakproc");
		  }
		pk0->pflags |= P_XYEDGE;
	  }
	else	pk0->pflags &= ~P_XYEDGE;
}

static int cleanup_peaks(int fmno, int npeaks, SPOTRUN *spp)
{ /*	This cleans up the list of peaks.  It defines the frame
	(or phi) value for all the new peaks as `fmno' and then
	looks through the new list to see if anything needs to get
	merged into other peaks.  It merges from the end of the list
	forward.  For each peak merged it swaps the nullified peak
	with the last good one on the list so that the end of the
	list has all the garbage in it.  At the end of all of this
	the peaks prior to those in the current frame will be somewhat
	out of order, so they're resorted on phi.  Also, at that point
	`spp->spr_peakp', the marker of the end of the peak list,
	is set just past the last good (unmerged) peak.  */

	register PEAK	*pkcmp, *pkp;
	int		tfm;
	unsigned	ntosort;
	double		raddist, realfm;
	PEAK		*first_reorder, *pk0;
	TROID		*pt0;
	static MAIL	*lhp;

	if (npeaks < 1) return 0;
	first_reorder = spp->spr_peakp;	spp->spr_pklast = spp->spr_peakp - 1;
	spp->spr_totalpeaks = spp->spr_peakp - spp->spr_pkbase;
	pk0 = spp->spr_peakp - npeaks;
	lhp = spp->spr_heads + 1 + (fmno % (spp->spr_nmbox - 1));
	if (spp->spr_plated < FUZZ) headtospp(spp);
	if (spp->spr_flags & P_FORCEPHI) lhp->ptrspace1 = P_FORCEPHI;
	else if (spp->spr_flags & P_FORCEOMEGA) lhp->ptrspace1 = P_FORCEOMEGA;
	realfm = spp->spr_spind = spindle(lhp->frameno, lhp);
	if (EOK != eftable(3, spp))
	  {
		fprintf(stderr,
		 "cleanup_peaks: eftable failure at spindle = %7.2f\n",
		 spp->spr_spind);
	  }
	for (pkp = spp->spr_pklast; pkp >= pk0; pkp--)
	   {
		pkp->pphicen = realfm;	/* frame # of this peak */
		pt0 = &(pkp->ptroi);
		pt0->nobk = 1;	/* one frame so far */
		pt0->frame0 = spp->spr_efrm;	pt0->obkmin = fmno;
		pt0->obparts[0] = 1.;	/* I / I(0) is identically 1 */
		/* stick it at the end if it's too weak to use */
		if (pkp->pflags & P_TOOWEAK) swap((AVOID *)pkp,
			(AVOID *)(spp->spr_pklast--), sizeof (PEAK));
	   }
	/* compute the Lorentz factor for all our reflections */
	for (pkp = pk0; pkp <= spp->spr_pklast; pkp++)
	   {
		if (EOK != peak2lor(spp, pkp)) pkp->pflags |= P_NOLORENTZ;
		else if (pkp->plor >= spp->spr_maxlor)
			pkp->pflags |= P_BIGLORENTZ;
	   }
	for (pkp = spp->spr_pklast; pkp >= pk0; pkp--)
	 if (pkp->pphicen < HUGEPHI - 1.) /* don't use eliminated peaks */
	  {
		for (pkcmp = pkp - 1; (pkcmp >= spp->spr_pkbase) &&
	     (fmno - (pkcmp->ptroi).frame0 < spp->spr_nearphi); pkcmp--)
		if (((pkcmp->ptroi).frame0 < HUGEPHI - 1.) &&
		 (fabs(pkp->pxcen - pkcmp->pxcen) < spp->spr_nearxy ||
		    fabs(pkp->pycen - pkcmp->pycen) < spp->spr_nearxy))
		     {
			raddist = radialdist(pkp, pkcmp, spp);
			if (raddist < spp->spr_samepeaksq)
			  {
				mrg_peaks(pkcmp, pkp, spp, 0); /*merge them */
	/* swap the current peak with the last unmerged peak at this phi value.
		Then decrement the pointer to the last unmerged peak */
				swap((AVOID *)pkp, (AVOID *)spp->spr_pklast--,
					sizeof (PEAK));
				if (spp->spr_pklast <= pk0) pkp = pk0;
			/* save a pointer to the merged PEAK so that we know
			how far back we have to go in resorting the list */
				if (first_reorder > pkcmp)
					first_reorder = pkcmp;
				break;	/* stop looking */
			  } /* end of `if (raddist < spp...)' clause */
			else if (raddist < spp->spr_partoverlap)
			  {
				pkp->pflags |= P_OVERLAP;
				if (spp->spr_flags & P_VERBOSE2)
				  {
					tfm = (pkp->ptroi).frame0;
					lhp = spp->spr_heads + 1 +
					 (tfm % (spp->spr_nmbox -1));
					dump_spot(spp, pkp, spp->spr_fpl, lhp,
						"P_OVERLAP in peakproc");
				  }
				pkcmp->pflags |= P_OVERLAP;
				break;
			  } /* end of `else if (raddist...' clause */
		     } /* end of `if (fabs...' clause and `for' loops */
	  }
	/* set spr_peakp pointer to start the next frame at the right place */
	spp->spr_peakp = spp->spr_pklast + 1;
		/* count the number of peaks under scrutiny */
	spp->spr_totalpeaks = spp->spr_peakp - spp->spr_pkbase;
	/* now a segment of the list is out of order in phi.  Fix that. */
	if ((fmno > spp->spr_zmi) && (pk0 > first_reorder))
	  {
		ntosort = pk0 - first_reorder;
		qsort((AVOID *)first_reorder, ntosort, sizeof (PEAK),
		 (QPE)phicmp);
	  }
	return (spp->spr_peakp - pk0);	/* return # of merged, good peaks */
}

int get_peaks(int fmno, SPOTRUN *spp)
{ /*	This routine reads in a frame of data and finds the peaks in it.
	The detector is divided into 64 regions of 64x64 pixels each
	and a background estimate is obtained inside each one.
	The cutoff within each of the 64 regions is then set to
	`spp->spr_nsigma' * sigma over the background (`spp->spr_nsigma'
	defaults to 10 but may be modified by the user).
	Anyway...pixels above the cutoff are merged into groups,
	and centroids are determined for each peak with
	total background-corrected intensity above another cutoff.
	These centroids (along with the integrated intensities)
	are stuck in an array of PEAK structures.
	The mailbox variable "lochp->count0" is used to contain
	the count of peaks in the frame.  */
	register int	npeaks, xleft, ytop;
	unsigned	nboxes;
	char		*framename;
	static MAIL	*lochp;

	/* First figure out the filename and read in the frame */
	spp->spr_tl0 = NOW;
	if (fmno == spp->spr_zmi) spp->spr_nfmsofar = 0;
	if (qerror((AVOID *)(framename = indexname(fmno)),
		"find requested frame file in get_peaks"))
	  {
		if (getenv("ALLOC_VERBOSE")) printf(
		 "calling peakskill because indexname failed\n");
		return peakskill(EREADERR, spp);
	  }
	spp->spr_nboxonx = (spp->spr_detw + (BOXWIDTH - 1))/ BOXWIDTH;
	spp->spr_nboxony = (spp->spr_deth + (BOXHEIGHT - 1))/ BOXHEIGHT;
	nboxes = spp->spr_nboxonx * spp->spr_nboxony;
	if (NULL == spp->spr_regbg)
	  {
		if (NULL == (AVOID *)(spp->spr_regbg = (float *)callop(nboxes,
		 sizeof (float), "spp->spr_regbg in get_peaks")))
		  {
			if (getenv("ALLOC_VERBOSE")) printf(
		 "calling peakskill because alloc for spp->spr_regbg failed\n");
			return peakskill(EALLOC, spp);
		  }
	  }
	lochp = spp->spr_heads + 1 + (fmno % (spp->spr_nmbox - 1));
	if (FRAME_OK != (npeaks = read_image(framename, spp->spr_imfl, lochp)))
	  {
		(void)image_error(stderr,        framename, npeaks, 0);
		(void)image_error(spp->spr_fpl,  framename, npeaks, 0);
	  }
	lochp->frameno = fmno;	/* so we can get at it elsewhere */
	spp->spr_tl1 = NOW;
	lochp->datetime[0] = spp->spr_tl1 - spp->spr_tl0;
	spp->spr_pk0infm = spp->spr_peakp = (spp->spr_pklast ?
		spp->spr_pklast + 1 : spp->spr_pkbase);
	unassign_pixels(spp); /* elts of spp->spr_peakass = PA_UNASSIGNED */
	if (spp->spr_flags & P_VERBOSE2)
	  {
		fprintf(spp->spr_fpl,
		 " Regional Backgrounds and Variances in Frame %d\n", fmno);
		fprintf(spp->spr_fpl,
	 " Xleft Ytop Used Boxmean Mode Sigma Cutoff Boxmean Newmean Boxvar\n");
	  }
	for (npeaks = xleft = 0; (xleft < spp->spr_detw) &&
	 (spp->spr_peakp <= spp->spr_pkend - 2); xleft += BOXWIDTH)
	   for (ytop = 0; (ytop < spp->spr_deth) &&
	    (spp->spr_peakp <= spp->spr_pkend - 2); ytop += BOXHEIGHT)
		npeaks += regional_peaks(spp, lochp, xleft, ytop);
	(spp->spr_nfmsofar)++;	/* increment count of # of fms seen so far */
	lochp->count0 = npeaks;	/* record # of peaks in fm */
	spp->spr_tl0 = NOW;
	lochp->datetime[1] = spp->spr_tl0 - spp->spr_tl1;
	/* define the phi value for each of the new peaks and if necessary
		merge it in with one or more that fall earlier in the list.*/
	lochp->count3 = cleanup_peaks(fmno, npeaks, spp);
	lochp->elap100ths = 0;	/*so printout can be ok on 1st 3 fms*/
	spp->spr_tl1 = NOW;
	lochp->accumsecs = spp->spr_tl1 - spp->spr_tl0;	return EOK;
}

int sift_peaks(SPOTRUN *spp)
{ /*	This goes through the peak list and merges any peaks whose proximity
	to one another was missed earlier */
	int	peaksnow;
	PEAK	*lpk, *pkp, *start_resort;

	if (spp->spr_flags & P_VERBOSEANY)
	  {
		fprintf(spp->spr_fpl, " Peaks found on the first pass:\n");
		prpeak(NULL, NULL);
		for (lpk = spp->spr_pkbase; lpk < spp->spr_pklast; lpk++)
			prpeak(spp->spr_fpl, lpk);
	  }
	for (lpk = spp->spr_pkbase; lpk < spp->spr_pklast; lpk++)
	   {
		spp->spr_spind = lpk->pphicen;
		if (EOK != eftable(5, spp)) fprintf(stderr,
		 "sift_peaks: eftable failure at [%7.2f,%7.2f,%7.2f]\n",
			 lpk->pxcen, lpk->pycen, lpk->pphicen);
		(lpk->ptroi).frame0 = spp->spr_efrm;
	   }
	for (lpk = spp->spr_pkbase, start_resort = NULL;
	 lpk < spp->spr_pklast; lpk++)
	   {
		for (pkp = lpk + 1; pkp <= spp->spr_pklast; pkp++)
		   { /* see if pkp is close to peakp */
			if (radialdist(lpk, pkp, spp) <
			 2. * spp->spr_samepeaksq)
			  { /* if it is, merge pkp into lpk */
				mrg_peaks(lpk, pkp, spp, 1);
				/* then swap pkp for last peak in list and
				 decrement count of which peak is last */
				swap((AVOID *)pkp,
				 (AVOID *)((spp->spr_pklast)--),sizeof (PEAK));
				if (NULL == start_resort) start_resort = lpk;
				/* decrement pointer `pkp' so the
				 newly-swapped peak can be checked too */
				pkp--;
			  }
		   }
	   }
	/* if the peaks are out of order in omega, re-sort them from the
	 point where the first out-of-order peak appears */
	peaksnow = 1 + (spp->spr_pklast - spp->spr_pkbase);
	if ((NULL != start_resort) && (start_resort < spp->spr_pklast))
		qsort((void *)spp->spr_pkbase, (unsigned)peaksnow,
			sizeof (PEAK), (QPE)phicmp);
	/* back up to find the last non-merged peak */
	for (lpk = spp->spr_pklast - 1; lpk >= spp->spr_pkbase; lpk--)
	   if (lpk->pphicen < HUGEPHI - 1.) break;
	spp->spr_pklast = lpk + 1;
	if (spp->spr_flags & P_VERBOSEANY)
	  {
		fprintf(spp->spr_fpl, " Peaks found after sifting:\n");
		prpeak(NULL, NULL);
		for (lpk = spp->spr_pkbase; lpk < spp->spr_pklast; lpk++)
			prpeak(spp->spr_fpl, lpk);
	  }
	/* this is also a convenient point to get memory set up for
	 the peak-finding operations */
	return peaksnow;
}

/* end of peaks.c */
