/* spotmain.c -- mainline and final-processing routines for `spots'

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	 8 Jun 2007: enable external specification of stepsize between images
	19 Apr 2007: float->double conversions; reorganize use of "{"
	22 Mar 2007: additional xgvariable output: Dmin
	 4 Mar 2007: emit some output information via "xgvariable" mechanism
	26 Feb 2007: fudge the B-values somewhat in Wilson stats
	24 Feb 2007: out_lsq, calc_lsq used to produce Wilson stats
	14 Feb 2007: spot_defaults: handle Aviex detector types
	22 Jul 2006: P_FORCEPHI and P_FORCEOMEGA flags implemented
	17 Jun 2006: replace act_pixel with activmask(x, y, 041)
	 6 Jun 2006: prpeak,logfm,anal_peaks,sift_peaks moved to peaks.c;
		assign_pixels,out_diffvmod moved here
	15 May 2006: DEF_PHISCALE #define'd here since it's only used here
	23 Apr 2006: minor rewording of diagnostic in sumspotinp
	 3 Apr 2006: spotmain: if first requested image isn't at beginning of
		the list of images, read the header for initial image anyway
	 5 Mar 2006: major revision: spindle values and effective frame # used;
		Lorentz, RANGEEDGE handled better; more diagnostics
	 2 Jan 2006: eftable() calls for cleanup; better memzap diagnostics
	20 Nov 2005: inline function declarations & function prototypes
	28 Oct 2005: replace static globals with elements of SPOTRUN structure
	12 Oct 2005: more efficient use of mrg_peaks call; housekeeping
	28 Sep 2005: add -j flag: specifies quitting if none of the specified
	 images is present, and foreshortening the run if some are missing
	16 Aug 2005: #define for UNSPEC_BEAMX moved here
	11 May 2005: check return status from xgjournal;
		handle helpapp and versionapp correctly in spotkill()
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	18 Oct 2004: getfoff no longer takes an argument
	26 Aug 2004: routines used only here are declared as static; cosmetics
	 2 Aug 2004: cosmetic change
	26 Jul 2004: more efficient handling of numerical run-time inputs
	28 Jun 2004: eliminate dangerous redefinition of lfmi in spot_defaults()
	15 Jun 2004: spot_defaults() written: defaults based on detector
		type and size; beamx now static global
	28 May 2004: clean up diagnostics for border functionality
	21 May 2004: add `border' functionality to `spots'. By default
		the `border' stuff only gets called if there's no AMASK file,
		but we can force to never do it or to always do it.
	18 Feb 2004: change kill_ftab call so that it doesn't have args
	26 Jan 2004: alert user to use of border;
		eliminate ytopbord, ybotbord, because they're not being used
	21 Jan 2004: license message inserted and initial comment reordered
	 6 Jan 2004: SPOTRUN structure elements
		& static globals replace extern globals;
		callop() & memzap() replace shcalloc() & shfree()
	13 Dec 2003: removed unused variables rpm,biggest from out_diffvmod()
	23 Nov 2003: replace globals with elements of SPOTRUN structure
	20 Sep 2003: always initialize `added' in outspotmask()
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	 2 May 2003: set `lhp' pointer more explicitly in sum_peaks()
	25 Mar 2003: call gspline() early on to allow radialdist()
		to handle Lorentz corrections
	14 Jan 2003: announce the adu/photon value even when we're using
		the default for the current detector type
	 5 Jan 2003: reorganized source-file header
	 2 Jan 2003: add "-o" flag; facilitates scripting
	 1 Nov 2002: trivial reformatting of globals;
		modify management of REGIONPROFILE structures;
		add # of images argument to sumspotinp();
		create Kabsch-space model capabilities as part of
		binary model profile functions in WMASK files
	30 Aug 2002: add "-c" flag: facilitates scripting
	15 Aug 2002: add "-a" flag: facilitates scripting
	25 Mar 2002: removed MSI attachments
	15 Nov 1998: in spot_setup, for very short frame ranges,
		we may have to restrict sgz and lz2 still further.
	13 Oct 1995: set spec_phix to zero in spotmain so defaults work.
	30 Sep 1995: allow the user to specify #adu/photon during spot-finding
	25 Aug 1995: reorganize final model profile output;
		diagnostics clearer, and it might help with bug 95230201001
	24 Aug 1995: 16-bit Mar overflows denoted right in frame summary
		copy images from other regions before analysis
		 so displays are neater;
		flush printout during passes 2 and 3 for clarity
	21 Jul 1995: major rewrite: 2nd pass thru data replaced by 2 passes,
	 each containing an examination of locally-read 3-D profiles.
	31 May 1995: "register" -> "register int" throughout;
		simplify call to outspotmask
	12 May 1995: make_ftab takes no argument; totalframes is a global
	12 Apr 1995: break spotmain() into several functions
	 2 Mar 1995: diagnostics to stdout; shared-memory alloc and free
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 1 Dec 1994: adjustable cutoff of # of low-count pixels
		examine elements of PEAK structure, not TROID
	 5 Oct 1994: use xgwrite to output TROID information
	30 Sep 1994: use floating-point stepsize in estimating pixel/cm ratio
	 7 Jul 1994: add diagnostic to memzap(&rgb,...) call
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN;
		more extern functions declared
	16 Mar 1994: model profile variables reorganized
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: Initial X-GEN version
	
   Purpose: toplevel routines for `spots' functionality, which finds
	centroids and model profiles for reference reflections
	obtained from area detector data.
   Cross references:
	./peaks.c, ./peakproc.c, ./spotshoes.c
*************************************************************************/
#define		NEED_TIME	1
#define		NEED_ERRNO	1
#include	"spotdefs.h"
#define		DEF_MAXMOVE	2.	/* max # pixels of movement
					allowed on 2nd pass, by default*/
#define		DEF_MAXBELOW	(0.3)	/* frac pixels << background */
#define		SMALL_RATIO	(1.0600)
#define		PART_RATIO	(1.10)	/* ratio of overlap to radius */
#define		MIN_MAXPEAKS	20000	/* at least this many peaks */
#define		PKSAT1SIGMA	40000	/* max # of peaks @ 1sigma*/
#define		UNSPEC_BEAMX	(-3173) /* code for unspecified beam X value */
#define		MAX_LCOR	400.	/* default max Lorentz correction */
/* default value sometimes overriden by user flag */
#define		DEF_PHISCALE	3.16	/* default (phi-to-x scale factor) */

static int spotkill(int rval, SPOTRUN *spp)
{ /*	This closes the logfile, releases any dynamically-allocated memory,
	and returns its argument */

	if (EHELP == rval) return EOK; /* help- or version- menu */
	if ((spp->spr_fpl != stdout) && (NULL != spp->spr_fpl))
		(void)fclose(spp->spr_fpl);
	memzap((AVOID **)&(spp->spr_pkbase), "spp->spr_pkbase in spots");
	memzap((AVOID **)&(spp->spr_shoe), "spp->spr_shoe in spots");
	memzap((AVOID **)&(spp->spr_thisi0), "spp->spr_thisi0 in spots");
	memzap((AVOID **)&(spp->spr_wmask), "spp->spr_wmask in spots");
	kill_mail(&(spp->spr_heads), &(spp->spr_nmbox));
	memzap((AVOID **)&(spp->spr_rgb), "rgb in spots");
	memzap((AVOID **)&(spp->spr_bdr), "bdp in spots");
	(void)ftables(0); globalkill();	spp->spr_fpl = NULL;	return rval;
}

static void sumspotinp(FILE *fp, int lz, int idark, SPOTRUN *spp)
{ /*	This writes a summary of the run-time inputs to stream fp */
	static char	*flist[] = {"frames", "amask", "darkimage",
		"centroids", "wmask", "xlog", NULL};
	static char	ftypes[] = { 1, 1, 1, 2, 2, 3, 0 };
	FRAMEINDEX	*lfmi;

	ftypes[1] = (spp->spr_flags & P_USEBORDER) ? 1 : 0;
	if (spp->spr_flags & P_DOBORDER) ftypes[1] |= 2;
	ftypes[2] = (spp->spr_imfl & DARK_CURRENT) ? 1 : 0;
	ftypes[5] = (spp->spr_fpl == stdout) ? 0 : 2;
	header(fp, PROCESSING, flist, ftypes);
	lfmi = spp->spr_fmind;
	fprintf(fp,
 "Search for bright spots conducted over frames %4d to %4d\n",
		spp->spr_zmi, spp->spr_zma);
	fprintf(fp,
 "Minimum count/sigma(background) for recording peaks:  %8.3f\n",
		spp->spr_nsigma);
	fprintf(fp,
 "Minimum %% of intensity in center 3 frames:            %8.3f%%\n",
		100.*spp->spr_mincenter);
	fprintf(fp,
 "Characteristic reflection radius in pixels:           %8.3f\n",
		sqrt(spp->spr_samepeaksq));
	fprintf(fp,
 "Maximum # of pixels a spot can move during 2nd pass:  %8.3f\n",
		spp->spr_maxmove);
	fprintf(fp,
 "Minimum summed intensity allowed for a reflection:    %8.3f\n",
		spp->spr_minint);
	fprintf(fp,
 "Number of spots for which memory will be available:   %8d\n", spp->spr_maxpeaks);
	fprintf(fp,
 "Scale factor between frame and pixel distances:       %8.3f pix/fm\n",
		sqrt(spp->spr_phitoxscale));
	fprintf(fp,
 "Number of frames examined in centroid calculations:	%8d\n", lz);
	fprintf(fp, "Maximum Lorentz factor 1/|sy|^2 = %8.1f\n",
		spp->spr_maxlor);
	if (spp->spr_imfl & DARK_CURRENT) fprintf(fp,
 "%6.4f of dark-current values contained in %s will be subtracted out\n",
		((double)idark) / 8000., trname("DARKIMAGE"));
	if (spp->spr_continue_refls) fprintf(fp,
 "Search will continue till %5d spots are found or frame %3d is examined\n",
		spp->spr_continue_refls, (int)(lfmi->fi_lastfm));
	if (spp->spr_mnbright) fprintf(fp,
 "At least %2d pixels in a spot must contain at least 10%% of the intensity\n",
		spp->spr_mnbright);
	fprintf(fp, "At most %5.3f of a spot's pixels may be well below bgnd\n",
		spp->spr_maxbelow);
	fprintf(fp,
	 " Variances will be calculated assuming %6.3f adu per photon\n",
		spp->spr_aduperph);
	if (0 <= spp->spr_beamx && spp->spr_beamx < spp->spr_detw)
	  fprintf(fp,
		"Nominal position of main beam in pixels in X:         %8.3f\n",
		(double)spp->spr_beamx);
	else if (spp->spr_beamx < 0) fprintf(fp,
		"The mainbeam is to the left of the detector\n");
	else fprintf(fp, "The mainbeam is to the right of the detector\n");
	if (spp->spr_flags & P_DOBORDER)
	  {
		fprintf(fp,
 "border calculation will set pixels inactive if count < %6.3f * <count>\n",
		 spp->spr_qvalu);
		if (spp->spr_valshrink > 0) fprintf(fp,
		 "border will be shrunk inward by %3d pixels\n",
		 spp->spr_valshrink);
	  }
	fprintf(fp,
	  "frame           frame    total over    max      at ");
	fprintf(fp, "  No. of Spots");
	if (spp->spr_flags & P_DOBORDER) fprintf(fp, " Cumulat. Mean");
	if ((fp != stdout) || (spp->spr_flags & P_VERBOSE2))
		fprintf(fp, "  time required for"); 
	fprintf(fp,
	 "\nnumb.            name   counts flow  count    X    Y found merg.");
	if (spp->spr_flags & P_DOBORDER) fprintf(fp, "   bgcount bgnd");
	if (((fp != stdout) && (fp != stderr)) ||(spp->spr_flags & P_VERBOSE2))
		fprintf(fp,"  input peaks merge");
	fprintf(fp,"\n");	if (fp != stderr) xeflush(fp);
}

static void sum_peaks(FILE *fp, int fmno, SPOTRUN *spp, double totc, int nombg)
{ /*	This summarizes the peak-finding pass through the data */
	char		*use;
	double		tev;
	MAIL		*lhp;
	static char	mar16[] = " %4d %15s %7.2e *%3d %6d %4d %4d %5d %5d ";
	static char	usual[] = " %4d %15s %7.2e %4d %6d %4d %4d %5d %5d ";

	lhp = spp->spr_heads + 1 + (fmno % (spp->spr_nmbox - 1));
	tev = lhp->eventcount;	if (tev < 0.) tev *= -64.;
	if ((lhp->data16) && (lhp->oflowcount > 0)) use = mar16;
	else use = usual;
	fprintf(fp, use, fmno, lhp->filename, tev,
	 (int)lhp->oflowcount, (int)lhp->zmax, (int)lhp->zmaxx,
	 (int)lhp->zmaxy, (int)lhp->count0, (int)lhp->count3);
	if (spp->spr_flags & P_DOBORDER) fprintf(fp, " %7.2e %4d",
		totc, nombg);
	if ((fp != stdout) || (spp->spr_flags & P_VERBOSE2))
		fprintf(fp, " %6d %5d %5d", (int)lhp->datetime[0],
			(int)lhp->datetime[1], (int)lhp->accumsecs);
	fprintf(fp,"\n");	if (fp != stderr) xeflush(fp);
}

static void summ_spots(FILE *fp, int *fap)
{ /*	Summary of the output results from `spots' */
	fprintf(fp, "Number of properly summed spots   = %5d\n", fap[0]);
	if (fap[1])
	  fprintf(fp, "Number of spots with XY dimples   = %5d\n", fap[1]);
	if (fap[2])
	  fprintf(fp, "Number of spots close to XY edge  = %5d\n", fap[2]);
	if (fap[3])
	  fprintf(fp, "Number of spots w. bad phi shape  = %5d\n", fap[3]);
	if (fap[4])
	  fprintf(fp, "Number of spots near 1st/last fm  = %5d\n", fap[4]);
	if (fap[5])
	  fprintf(fp, "Number of spots w.severe overlap  = %5d\n", fap[5]);
	if (fap[6])
	  fprintf(fp, "Number of spots with I < Imin     = %5d\n", fap[6]);
	if (fap[7])
	  fprintf(fp, "Number of spots that move too far = %5d\n", fap[7]);
	if (fap[8])
	  fprintf(fp, "Number of spots merged into others= %5d\n", fap[8]);
	if (fap[9])
	  fprintf(fp, "Number of spots with read errors  = %5d\n", fap[9]);
	if (fap[10])
	  fprintf(fp, "Number of spots too sharp in XYZ  = %5d\n", fap[10]);
	if (fap[11])
	  fprintf(fp, "Number of spots with many Ct<Bgnd = %5d\n", fap[11]);
	if (fap[12])
	  fprintf(fp, "Number of spots with no  Lor. Fac = %5d\n", fap[12]);
	if (fap[13])
	  fprintf(fp, "Number of spots with large LorFac = %5d\n", fap[13]);
}

static int calc_beamx(SPOTRUN *spp, char *beamname)
{ /*	This does a rough estimate of the X value of the direct beam by reading
	the header of a frame and computing it from the distance and twotheta
	value.  It returns UNSPEC_BEAMX if no useful information comes of
	this attempt.*/
	int	temp_flags;
	MAIL	*lhd;
	DATARUN	*runhdp;
	double	xtocen;
	static char readmsg[] =
		"Unable to read test-frame %s to get beam-center value\n";

	if (NULL == beamname || '\0' == *beamname) return UNSPEC_BEAMX;
	temp_flags = GET_RUNHEADER | JUST_HEADER;
	lhd = spp->spr_heads;
	if (FRAME_OK != read_image(beamname, temp_flags, lhd))
	  {
		fprintf(stderr, readmsg, beamname);
		fprintf(spp->spr_fpl, readmsg, beamname); return UNSPEC_BEAMX;
	  }
	if (NULL == (runhdp = lhd->runheader)) return UNSPEC_BEAMX;
	if (runhdp->run_d0pos[0] < FUZZ) return UNSPEC_BEAMX;
	xtocen = runhdp->run_d0pos[0] * tan(-runhdp->run_d0ang[0] * RADC);
	return (int)((double)((lhd->xdimen) / 2) + xtocen * MEANPIXPERCM);
}

static int spot_defaults(SPOTRUN *spp)
{ /*	This sets up appropriate default values for run-time parameters to
	the `spots' functionality based on the detector type. */

	int		detype, dedim[2], stype, ssize;
	double		samep;
	FRAMEINDEX	*lfmi;
	static int	nftype[] = /* # of frames to search for spots */
	 { 65, 65, 65, 40, 65, 6, 6, 6, 5, 6, 8, 6, 6, 6, 6, 6, 6, 6} ;
	static double	aduperp[] = /* # of adu per pixel */
	 { 1.,1.,1.,4.,4.,5.,4.,4.,5.,5.,3.8,1.8,1.4,1.,1.8,1.8,4.5,4.5,4.5 };
	static double	qvalu[] = /* Count/<Count> below which pixel is dead*/
	 { 0.38, 0.36, 0.37, 0.38, 0.4, 0.37, 0.37, 0.4, 0.4, 0.37, 0.42,
	   0.38, 0.38, 0.41, 0.42, 0.42, 0.42, 0.42, 0.42 } ;

	zeromem((void *)spp, sizeof (SPOTRUN));
	if (qerror((AVOID *)(spp->spr_fmind = ftables(3)),
		"read frame-index file")) return EREADERR;
	lfmi = spp->spr_fmind;
	spp->spr_flags = P_MAYBEBORDER;
	spp->spr_continue_refls = spp->spr_firstrange = 0;
	if (EOK == foffset(1)) spp->spr_flags |= P_FRAMEOFFSETS;
	spp->spr_beamx = UNSPEC_BEAMX;
	/* wait for a frame to appear if needed */
	spp->spr_maxmove = DEF_MAXMOVE;	spp->spr_minint = MININTEN;
	spp->spr_maxbelow = DEF_MAXBELOW; spp->spr_imfl = WAIT_IF_ABSENT;
	spp->spr_valshrink = 16;
	spp->spr_maxlor = MAX_LCOR;	/* maximum Lorentz correction */
	spp->spr_lz2 = 0;	/* place-holder: real value set later */
	/* determine what kind of detector we have */
	if (0 >= (detype = dettype(dedim))) return EREADERR;
	spp->spr_detw = dedim[0];	spp->spr_deth = dedim[1];
	stype = detype & 0377;	ssize = detype >> 8;
	/* shrink border in by a number of pixels related to detector size */
	if (dedim[0] <= 512) spp->spr_valshrink = 6;
	else if (dedim[0] <= 1024) spp->spr_valshrink = 9;
	else if (dedim[0] <= 2048) spp->spr_valshrink = 16;
	else spp->spr_valshrink = 20;
	/* by default we begin with the first frame */
	spp->spr_zmi = lfmi->fi_firstfm;
	/* decide which frame we'll end with and set the adu/photon value */
	if (stype <= 18)
	  {
		spp->spr_zma = spp->spr_zmi + nftype[stype];
		spp->spr_overphot = aduperp[stype];
		spp->spr_qvalu = qvalu[stype];
	  }
	else {
		spp->spr_zma = spp->spr_zmi + 8;	spp->spr_overphot = 4;
		spp->spr_qvalu = 0.38;
	     }
	if (spp->spr_zma > lfmi->fi_lastfm) spp->spr_zma = lfmi->fi_lastfm;
	/* pick the default distance between peaks */
	switch (ssize)
	  {
		case	1:	case	2:	samep = 4.5;	break;
		case	3:			samep = 5.7;	break;
		case	4:	case	5:	samep = 6;	break;
		default:			samep = 6.;	break;
	  }
	spp->spr_samepeaksq = samep * samep;
	if (stype < 5)
	  {
		spp->spr_lz2 = 3;	spp->spr_maxbelow = 0.3;
		spp->spr_nsigma = 20;	spp->spr_mincenter = 0.6;
	  }
	else if (stype == 10)
	  {
		spp->spr_lz2 = 2;	spp->spr_maxbelow = 0.75;
		spp->spr_nsigma = 80;	spp->spr_mincenter = 0.65;
	  }
	else {
		spp->spr_lz2 = 2;	spp->spr_maxbelow = 0.5;
		spp->spr_nsigma = 80;	spp->spr_mincenter = 0.65;
	     }
	return EOK;
}

static int assign_pixels(SPOTRUN *spp)
{ /*	set up the peak assignment array and give all the pixels
	legitimately inside the detector array the value PA_UNASSIGNED.
	All the ones close to the rectangular borders are given the value
	PA_RECTBORDER.  All pixels outside the active area of the detector
	are given the value PA_ELLBORDER */
	register int		x, y;
	register u_short	*pka;
	int			xmx, ymx;
	unsigned		imsiz;

	imsiz = spp->spr_deth * spp->spr_detw;
	if (NULL == (AVOID *)(spp->spr_peakass = (u_short *)callop(imsiz,
	 sizeof (u_short), "spp->spr_peakass in assign_pixels"))) return EALLOC;
	pka = spp->spr_peakass;	xmx = spp->spr_detw - USELESS;
	ymx = spp->spr_deth - USELESS;
	for (x = 0, pka = spp->spr_peakass; x < spp->spr_detw; x++)
	   {
		if ((x < USELESS) || (x >= xmx))
			for (y = 0; y < spp->spr_deth; y++, pka++)
				*pka = PA_RECTBORDER;
		else {
			for (y = 0; y < USELESS; y++, pka++)
				*pka = PA_RECTBORDER;
			for ( ; y < ymx; y++, pka++) *pka =
			 (NULL != activmask(x, y, 041)) ?
			 PA_UNASSIGNED : PA_ELLBORDER;
			for ( ; y < spp->spr_deth; y++,pka++)
				*pka = PA_RECTBORDER;
		     }
	   }
	return EOK;
}

static int spot_setup(int argc, char *argv[], SPOTRUN *spp)
{ /*	This sets up for a run of `spots' by decoding the runtime inputs,
	setting up the frame index, opening the log file,
	and scribbling therein.  */
	int		firstgood, fmno, lastgood, readerr, specshrink, tfno;
	double		framrad, nval, samepk;
	double		spec_low, spec_phix, spec_sigma, specstep;
	char		*amname, *framename, **dcp, **flags;
	MAIL		*lhd;
	FRAMEINDEX	*lfmi;
	FRAMEENTRY	*fee, *ind0, *indentry;
	DATARUN		*drp;
	static int	idarkfrac = 6400;

	if (EOK != (fmno = xgjournal(argc, argv))) return fmno;
	spot_defaults(spp);	/* get detector-type-specific defaults */
	lfmi = spp->spr_fmind;
	spec_phix = spec_sigma = spec_low = 0.;	specshrink = -1;
	/* process flags, if any.  Currently defined flags are:
	"-a<val>" specifies that the first <val> images in the FRAMES
		file be used for the spot-finding effort. <val> must
		be no larger than the number of images in the FRAMES file.
	"-b"	causes "spots" to look only at pixels within the active
		region of the detector as defined by the AMASK file.
	"-c<val>" specifies that, if insufficient reflections have been
		obtained from the specified frame range, the search
		for spots should continue until <val> reflections have
		been obtained.
	"-d"	means that the code should definitely _not_ calculate
		a detector border. By default the border is calculated
		if the AMASK file is absent but not otherwise.
	"-e<val>" obliges `spots' to define the number of adu
		per photon to be <val>. If this isn't specified, then
		the number of adu per photon is set based on the
		detector type (1 for Poisson counters).
	"-f<val>" sets the expected half-width in frames of the reflections
		to <val>.  <val> must be between 1 and 5 inclusive.
		The reflections will be examined over a profile box of
		(2 * <val> + 1) frames, i.e. if <val> = 3 we will examine
		a 7-frame box.  The default value of <val> is 3 unless the
		scale factor between frames and pixels is greater than 16,
		in which case the default is 2.
	"-g<val>" sets the Q fraction for the border calculation to <val>.
		Thus -g0.38 means that pixels for which count/<count> < 0.38
		will be considered to be outside the active area.
	"-h<val>" sets the minimum number of bright pixels per reflection
		to <val>.  The default is 0.
	"-i<val>" prevents "spots" from writing out a reflection with
		total intensity less than <val>.  Note that this intensity
		has not been Lorentz-corrected, so this cutoff can actually
		be pretty large and still let in lots of reflections.
	"-j<val>" checks to see that the images requested are all
		present at the beginning of the `spots' run. If all are
		missing, exit quietly without complaining. If some are missing,
		skip over them without waiting for them. This option
		is useful in scripts that should behave in this way if
		certain images are unavailable.
	"-k<val>" defines the scale factor relating distances between points
		in scanning angle to distances between points on the detector
		face.  This value is sqrt(10) by default if the program
		is unable to determine the detector distance OR the detector's
		properties; it's determined from the latter if possible;
		and if this -k flag is used these calculations are overridden
		by the value represented by the -k flag.
	"-l<val>" defines the fraction of pixels within a reflection's shoebox
		that will be allowed to be substantially below background.
		Default is 0.3; it should be higher (0.6?) for image plates.
	"-m<val>" allows a centroid to move by <val> pixels between the
		first & 2nd passes of the program.
	"-n<val>" imposes an extra requirement on the profile, namely that
		at least one element besides the brightest be at least <val>
		as large as the brightest.  Thus "spots -n0.1" requires
		that the second-largest element be at least 10% as large
		as the largest element.
	"-o<val>" sets the I/sigma(Background) value to <val>.
		Thus it's like the third non-switch parameter.
	"-p<val>" causes the profile sharpness parameter
		"spp->spr_mincenter" to be set to <val>.
	"-q"	sets up a quasi-verbose output--less verbose than "-v" but
		 more verbose than default.
	"-s" causes the frames to be analyzed after subtracting a dark-current
		image from each one.  The dark-current image has environment
		variable or logical name DARKIMAGE.
	"-t<val>" specifies that, for the border calculation, the border should
		be shrunk inward by <val> pixels
	"-u"	means that the code unconditionally _should_ calculate
		a detector border. By default the border is calculated
		if the AMASK file is absent but not otherwise.
	"-v" causes diagnostics to be written to stdout as well as
		to the logfile and enormously increases the logfile output.
	"-w<val>" causes the nominal radius of a reflection to be defined as
		<val> pixels instead of the default, which is about 5.5.
	"-x<val>" specifies that the direct beam is found at X = <val>.
		This governs the choice of where the boundaries between model
		profiles will be placed.  By default, "spots" determines this
		value from the header of one of the frames, so this flag is
		only required when you need to overrule what the header says.
	"-y<val>" specifies the maximum value of 1/|sy} that will be
		allowed; this enables removal of reflections with
		nearly infinite Lorentz corrections
*/
	flags = flagproc(argc, argv, "ABCDEFGHIJKLMNOPQRSTUVWXYZ");
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	 {
	   nval = atof((*dcp) + 1);
	   switch (**dcp)
	     {
		case	'A':	spp->spr_firstrange = nval;	break;
		case	'B':	spp->spr_flags |= P_USEBORDER;	break;
		case	'C':	spp->spr_continue_refls = nval;	break;
		case	'D':	spp->spr_flags &= ~P_DOBORDER;	break;
		case	'E':	if ((0.01 < nval) &&
		 (nval < 1000.)) spp->spr_overphot = nval;	break;
		case	'F':	spp->spr_lz2 = nval;		break;
		case	'G':	if ((0. <= nval) && (nval <= 1.))
			spp->spr_qvalu = nval;	break;
		case	'H':	spp->spr_mnbright = nval;	break;
		case	'I':	spp->spr_minint = nval;	break;
		case	'J':	spp->spr_flags |= P_QUITQUIETLY;	break;
		case	'K':	spec_phix = nval;	break;
		case	'L':	spec_low = nval;	break;
		case	'M':	spp->spr_maxmove = nval;		break;
		case	'N':	spp->spr_flags |= P_FORCEPHI;	break;
		case	'O':	spec_sigma = nval;	break;
		case	'P':	spp->spr_mincenter = nval;	break;
		case	'Q':	spp->spr_flags |= P_VERBOSE1;	break;
		case	'R':	spp->spr_flags |= P_FORCEOMEGA;	break;
		case	'S':	spp->spr_imfl |= DARK_CURRENT;
			if (0. < nval) idarkfrac = 8000. * nval;	break;
		case	'T':	specshrink = nval;	break;
		case	'U':	spp->spr_flags |= P_MUSTDOBORDER; break;
		case	'V':	spp->spr_flags |= P_VERBOSEANY;	break;
		case	'W':	spp->spr_samepeaksq = nval;
		 	spp->spr_samepeaksq *= spp->spr_samepeaksq;	break;
		case	'X':	spp->spr_beamx = nval;		break;
		case	'Y':	spp->spr_maxlor = nval;		break;
		case	'Z':	specstep = nval;
			if ((specstep < -DFUZZ) || (specstep > DFUZZ))
			  {
				spp->spr_flags |= P_EXTERNSTEP;
				spp->spr_framrad = specstep * RADC;
			  }
			break;
	     }
	 }
	amname = trname("amask");
	if ((0 == xgaccess(amname, 4, 0)) && (spp->spr_flags & P_MAYBEBORDER)
	 && (0 == (spp->spr_flags & P_MUSTDOBORDER)))
		spp->spr_flags &= ~P_DOBORDER;
	if (idarkfrac > 16000) idarkfrac = 16000;
	if (spp->spr_imfl & DARK_CURRENT) spp->spr_imfl |= (idarkfrac << 16);
	if (spp->spr_minint < MININTEN) spp->spr_minint = MININTEN;
	if ((DFUZZ < spec_low) && (spec_low <= 1.))
		spp->spr_maxbelow = spec_low;
	spp->spr_samepeaksq = flimits(spp->spr_samepeaksq, 9.01, 60.);
	spp->spr_partoverlap = spp->spr_samepeaksq * PART_RATIO;
	samepk = sqrt(spp->spr_samepeaksq);
	spp->spr_nearxy = sqrt(spp->spr_samepeaksq * PARTIAL_OVERLAP_RATIO) +1;
	spp->spr_smallwidth = (int)(samepk * SMALL_RATIO + 0.4999);
	spp->spr_smallheight = spp->spr_smallwidth;
	spp->spr_smlhalf = spp->spr_smallwidth / 2;
	if (spp->spr_maxlor  < 1.) spp->spr_maxlor = MAX_LCOR;
/* process remaining arguments: frame numbers of start and end,
	and, if present, an I/sigma cutoff for use of spots */
	for (argc--, argc++; (argc) && !allnumeric(*argv); argc--) argv++;
	if (argc >= 2)
	  {
		spp->spr_zmi = atoi(*argv++);
		spp->spr_zma = atoi(*argv++); argc -= 2;
	  }
	if (spp->spr_zma < spp->spr_zmi) spp->spr_zma = spp->spr_zmi;
	/* flag set up Aug 02: -a<val> specifies that
	 the first <val> images will be used. We require that the
	 <val> be no larger than the number of images in the FRAMES file */
	if (spp->spr_firstrange > 0)
	  {
		if (spp->spr_firstrange > lfmi->fi_totfms)
			spp->spr_firstrange = lfmi->fi_totfms;
		spp->spr_zmi = lfmi->fi_firstfm;
		fee = lfmi->fi_fmi + (spp->spr_firstrange - 1);
		spp->spr_zma = fee->fmno;
	  }
	/* if the "quit quietly" flag is set, make sure the images are here */
	if (spp->spr_flags & P_QUITQUIETLY)
	  {
		ind0 = lfmi->fi_fmi + (spp->spr_zmi - lfmi->fi_firstfm);
		firstgood = lastgood = -1;
		for (indentry = ind0, tfno = spp->spr_zmi;
		 tfno <= spp->spr_zma; tfno++, indentry++)
		   {
			if (0 != xgaccess(indentry->fmname, 04, 0)) break;
			else {
				lastgood = tfno;
				if (firstgood == -1) firstgood = tfno;
			     }
		   }
		if (lastgood < 0)
		  {
			if (spp->spr_zma <= spp->spr_zmi)
				printf("Image %3d was absent.\n",spp->spr_zmi);
			else	printf("Images %3d through %3d were absent.\n",
					spp->spr_zmi, spp->spr_zma);
			return EOPENERR;
		  }
		else {
			spp->spr_zma = lastgood;
			spp->spr_zmi = firstgood;
		     }
	  }
	/* flag as of Jan 03: 'O' is like the third arg. */
	if ((spec_sigma < 1.) && (argc)) spec_sigma = atof(*argv);
	if (spec_sigma >= 1) spp->spr_nsigma = spec_sigma;
	spp->spr_iosmin = (spp->spr_nsigma < 18 ? 1.5 : spp->spr_nsigma / 12);
		/* measure the height and width of the detector */
	if (specshrink >= 0) spp->spr_valshrink = specshrink;
	spp->spr_maxpeaks = PKSAT1SIGMA * (spp->spr_detw / 512) *
		(spp->spr_deth / 512) / spp->spr_nsigma;
	if (spp->spr_maxpeaks < MIN_MAXPEAKS) spp->spr_maxpeaks = MIN_MAXPEAKS;
	if (NULL == (AVOID *)(spp->spr_pkbase =
	 (PEAK *)callop((unsigned)spp->spr_maxpeaks, sizeof (PEAK),
		"spp->spr_pkbase in spot_setup"))) return EALLOC;
	spp->spr_pklast = NULL;	/* otherwise multiple SPOTS runs get confused */
	spp->spr_pkend = spp->spr_pkbase + spp->spr_maxpeaks;
	/* reserve memory for the image header structures */
	spp->spr_nmbox = PFRAMES + 2; /* max # of headers we'll need */
	if (NULL == (AVOID *)(spp->spr_heads = (MAIL *)callop(spp->spr_nmbox,
	 sizeof (MAIL), "heads in spot_setup"))) return EALLOC;
	lhd = spp->spr_heads;
	/* obtain either the real active-area mask or a nominal one */
	if (NULL == activmask(spp->spr_detw, spp->spr_deth, 5))
		return (errno == ENOMEM) ? EALLOC : EREADERR;
	/* if the user hasn't specified `beamx', calculate it */
	for (fmno = spp->spr_zmi;
	 fmno <= spp->spr_zmi + 5 && spp->spr_beamx == UNSPEC_BEAMX; fmno++)
		spp->spr_beamx = calc_beamx(spp, indexname(fmno));
	if (UNSPEC_BEAMX == spp->spr_beamx) spp->spr_beamx = spp->spr_detw / 2;
	spp->spr_bdr = make_boundaries(spp->spr_beamx,
		spp->spr_detw, spp->spr_deth);
	if (spec_phix > DFUZZ) spp->spr_phitoxscale = spec_phix;
	/* change as of Dec 2003: always determine default phixtoscale value;
	 that way we always have a valid drp value if the images permit it */
	if (qerror((AVOID *)(framename = indexname(spp->spr_zmi)),
	 "find requested frame file in spot_setup")) return EREADERR;
	if (FRAME_OK != (readerr = read_image(framename, spp->spr_imfl, lhd)))
		(void)image_error(stdout, framename, readerr, 0);
	/* try to determine the scale factor relating omega to (X,Y) */
	if ((NULL != (drp = lhd->runheader)) && (drp->run_d0pos[0] > FUZZ)
	  && (lhd->cellsize > 0))
	    {
		if (spp->spr_flags & P_EXTERNSTEP)
			framrad = spp->spr_framrad;
		else if (fabs(drp->run_stepreal) > FUZZ)
			 framrad = RADC * fabs(drp->run_stepreal);
		else if (abs(drp->run_stepsize) > 0)
			framrad = RADC *fabs((double)(drp->run_stepsize)) /96.;
		else framrad = RADC * 0.25;	/* oh well... */
		spp->spr_phitoxscale = drp->run_d0pos[0] *
			RADC * 10000. / (double)(lhd->cellsize);
	    }
	else spp->spr_phitoxscale = DEF_PHISCALE;
	if (spec_phix > DFUZZ) spp->spr_phitoxscale = spec_phix;
	spp->spr_phitoxscale *= spp->spr_phitoxscale;	/* square it for use */
	if ((0 >= spp->spr_lz2) || (6 <= spp->spr_lz2))
		spp->spr_lz2 = (spp->spr_phitoxscale < 121.) ?
		3 : (spp->spr_phitoxscale < 256.) ? 2 : 1;
	spp->spr_sgz = spp->spr_lz2 + spp->spr_lz2 + 1;
	/* for very short frame ranges, restrict sgz and lz2 further */
	if ((nval = spp->spr_zma + 1 - spp->spr_zmi) < spp->spr_sgz)
		spp->spr_lz2 = (nval - 1) / 2;
	spp->spr_sgz = spp->spr_lz2 + spp->spr_lz2 + 1;
	spp->spr_nearphi = sqrt(spp->spr_samepeaksq * PARTIAL_OVERLAP_RATIO *
		spp->spr_phitoxscale) + 1;
	/* set up the peak assignment matrix */
	if (EOK != (readerr = assign_pixels(spp))) return readerr;
	spp->spr_peakp = spp->spr_pkbase; /* start pointer in right place */
	spp->spr_tl0 = NOW;
	if (qerror((AVOID *)(spp->spr_fpo = xgopen("centroids", "w")),
		"create output centroids file")) return EOPENERR;
	spp->spr_aduperph = (spp->spr_overphot > DFUZZ) ? spp->spr_overphot :
		(((NULL != drp) && (drp->run_aduperph > FUZZ)) ?
		drp->run_aduperph : 1.);
	/* create or append to the crystallographic logfile */
	if (NULL == (spp->spr_fpl = propen("xlog"))) spp->spr_fpl = stdout;
	else {
		sumspotinp(spp->spr_fpl, spp->spr_sgz, idarkfrac, spp);
		xeflush(spp->spr_fpl);
	     }
	/* March 2003: try to create a spline early on */
	if (NULL == (spp->spr_spl = gspline(5, &readerr))) return readerr;
	sumspotinp(stdout, spp->spr_sgz, idarkfrac, spp);	return EOK;
}

static int out_diffvmod(SPOTRUN *spp)
{ /*	This recenters the diffraction vector profiles, constructs the
	structure in which each vector model profile will be kept,
	and writes the structures out. */
	register int	i, im;
	int		lx, ly, lz, lxy, lxyz, lm, mx, my, mz, mxy, mxyz;
	float		*kpm, *kpmo, *rpmo;
	double		minten, delv[3], didv[3];
	REGIONPROFILE	*rgb, *rgp;

	fprintf(spp->spr_fpl,
 "Summary of diffraction-vector model miscenterings:\n");
	fprintf(spp->spr_fpl,
 "  Model       before adjustment           after adjustment\n");
	fprintf(spp->spr_fpl,
 " Number      X      Y      Z  inten      X      Y      Z  inten\n");
	/* center the diffraction-vector models and define the values
	 in the REGIONPROFILE structure */
	rgb = spp->spr_rgb;
	lx = rgb->m3_bx;	ly = rgb->m3_by;	lz = rgb->m3_bz;
	lxy = lx * ly;	lxyz = lxy * lz;	lm = rgb->m3_nmodels;
	mx = rgb->m3_qx;	my = rgb->m3_qy;	mz = rgb->m3_qz;
	mxy = mx * my;	mxyz = mxy * mz;
	rpmo = spp->spr_rprop + lm * lxyz;	kpm = rpmo + lm * lxyz;
	kpmo = kpm + mxyz * lm;
	for (rgp = rgb, im = 0;
	 im < lm; im++, kpm += mxyz, kpmo += mxyz, rpmo += lxyz, rgp++)
	   {
		for (i = 0; i < 3; i++) didv[i] = delv[i] = 0.;
		minten = gen_centroid(kpm, NULL, mx, my, mz, didv, delv);
		fprintf(spp->spr_fpl, " %6d %6.3f %6.3f %6.3f %6.1f",
		 im, delv[0], delv[1], delv[2], minten);
		recen_vec(kpm, delv, kpmo, mx, my, mz);
		for (i = 0; i < 3; i++) didv[i] = delv[i] = 0.;
		minten = gen_centroid(kpmo, NULL, mx, my, mz, didv, delv);
		fprintf(spp->spr_fpl, " %6.3f %6.3f %6.3f %6.1f\n",
		 delv[0], delv[1], delv[2], minten);
		rgp->m3_bx = lx;	rgp->m3_by = ly;
		rgp->m3_bz = lz;	rgp->m3_imodel = im;
		rgp->m3_nmodels = lm;
		rgp->m3_xyop = rpmo;
		rgp->m3_qx = mx;	rgp->m3_qy = my;
		rgp->m3_qz = mz;
		rgp->m3_dvp = kpmo;
		rgp->m3_xyoip = spp->spr_sic;
		rgp->m3_dvip = spp->spr_sic + lxyz;
		model_ftoc(rgp);
		if (0 >= fwrite((AVOID *)rgp, sizeof (REGIONPROFILE),
			1, spp->spr_fpmask)) return EWRITERR;
		if (0 >= fwrite((AVOID *)spp->spr_sic, sizeof (sichar),
			lxyz + mxyz, spp->spr_fpmask)) return EWRITERR;
	   }
	memzap((AVOID **)&spp->spr_sic, "sic in out_diffvmod");
	memzap((AVOID **)&(spp->spr_rprop), "rprop in out_diffvmod");
	return EOK;
}

static int outspotmask(SPOTRUN *spp)
{ /*	This checks each floating-point 3D profile to make sure it has
	something in it.  If it doesn't, it averages all the others to
	produce something moderately reasonable.
	It then converts the 3D profiles from the rational format used
	internally into the weird format used for input and output.
	Then it writes 'em out. It returns EOK if everything is okay,
	various booboos if not. */
	int		added, boff, iz, lm, lx, lxyz, ly, lz;
	int		maskno, retv, nupf[2];
	double		floorfrac, minten, delv[3], idv[3], lellp[14], sumf[3];
	float		*zvi, *zvp;
	REGIONPROFILE	*rgb, *rgp, *rgq;
	MASK_3D		*wmp;
	static char	modmsg[] = "Can't convert model %2d to output format\n";
	static char	openmsg[] = "Can't open output model profile file!\n";
	static char	writmsg[] = "failure writing model %d\n";

	/* clean up the (X,Y,omega)-space models */
	rgb = spp->spr_rgb;	cleanmodels(rgb, 0, 1);
	xeflush(spp->spr_fpl);
	fprintf(spp->spr_fpl, "Summary of model mis-centerings:\n");
	fprintf(spp->spr_fpl, "  Model  # of    before adjustment");
	fprintf(spp->spr_fpl, "     after adjustment    integerized model\n");
	fprintf(spp->spr_fpl, " Number Refls      X      Y      Z");
	fprintf(spp->spr_fpl, "      X      Y      Z      X      Y      Z\n");
	floorfrac = 0.035;
	lm = rgb->m3_nmodels;
	if (NULL == spp->spr_thisi0)
	  {
		if (NULL == (AVOID *)(spp->spr_thisi0 =
		 (float *)callop((unsigned)8000, sizeof (float),
		  "spp->spr_thisi0 in outspotmask"))) return EALLOC;
	  }
	for (added = maskno = 0, rgp = rgb, rgq = rgb + lm,
	 wmp = spp->spr_wmask; maskno < lm; maskno++, rgp++, wmp++, rgq++)
	   {
		lx = rgp->m3_bx; ly = rgp->m3_by; lz = rgp->m3_bz;
		lxyz = lx * ly * lz;	spp->spr_lz2 = lz / 2;
		/* enforce contiguity on the model */
		if (0 >= contig(rgp, -1))
		  {
			fprintf(spp->spr_fpl,
			 "Contiguity correction failed in region %2d; ",
			 maskno);
			fprintf(spp->spr_fpl, "default model used\n");
			default_model(rgp->m3_xyop, spp);
		  }
		/* recenter the model */
		for (iz = 0; iz < 3; iz++) idv[iz] = delv[iz] = 0.;
		minten = gen_centroid(rgp->m3_xyop, NULL, lx, ly,lz,idv, delv);
		if (DFUZZ > minten)
		  {
			fprintf(spp->spr_fpl,
		 " Warning: no intensity in region %2d: default model used\n",
				maskno);
			default_model(rgp->m3_xyop, spp);
		  }
		rgq->m3_nobs = rgp->m3_nobs; rgq->m3_isqsum = rgp->m3_isqsum;
		recen_vec(rgp->m3_xyop, delv, rgq->m3_xyop, lx, ly, lz);
		fprintf(spp->spr_fpl, " %6d %5d %6.3f %6.3f %6.3f",
			maskno, rgp->m3_nobs, delv[0], delv[1], delv[2]);
		for (iz = 0; iz < 3; iz++) idv[iz] = delv[iz] = 0.;
		minten = gen_centroid(rgq->m3_xyop, NULL,
			lx, ly, lz, idv, delv);
		/* add intensity on the wings if needed */
		if (lz > 1) added = dospread(rgq, 0, sumf, nupf, floorfrac);
		/* Gaussian cleanup */
		/* first copy the corrected profile into the original */
		for (iz = 0, zvp = rgp->m3_xyop, zvi = rgq->m3_xyop;
		 iz < lxyz; iz++, zvp++, zvi++) *zvp = *zvi;
		if (EOK != (retv = ellsoid(rgp, 1, 014, 0.008, 4., lellp)))
		  {
			fprintf(spp->spr_fpl,
 " Warning: Gaussian cleanup returned %d in region %2d: default model used\n",
				retv, maskno);
			default_model(rgp->m3_xyop, spp);
		  }
	 	fprintf(spp->spr_fpl,
			" %6.3f %6.3f %6.3f", delv[0], delv[1],delv[2]);
		if ((EOK != (retv = vectowmp(rgq->m3_xyop, wmp, rgq->m3_nobs,
			rgq->m3_isqsum, lz))) && (EBADVALUE !=retv))
		  {
			fprintf(stderr, modmsg, maskno);
			fprintf(spp->spr_fpl, modmsg, maskno);	return retv;
		  }
		/* we don't want singularities in the model profile ellipsoid
		 calcs, so we require that a little intensity slop over into
		 the neighboring images if it is originally all in the
		 centermost image. */
		spreadwmp(wmp);
		if (EOK == wmptovec(wmp, spp->spr_thisi0))
		  {
			for (iz = 0; iz < 3; iz++) idv[iz] = delv[iz] = 0.;
			boff = (3 - spp->spr_lz2) * lx * ly;
			zvi = spp->spr_thisi0 + boff;
			for (iz = 0, zvp = rgp->m3_xyop; iz < lxyz; iz++)
				*zvp++ = *zvi++;
			minten = gen_centroid(rgp->m3_xyop, NULL,
				lx, ly, lz, idv, delv);
	 		fprintf(spp->spr_fpl, " %6.3f %6.3f %6.3f",
				delv[0], delv[1], delv[2]);
			if (added) fprintf(spp->spr_fpl, " spread %s",
			 (added == 1) ? "left" :
			 ((added == 2) ? "right" : "both"));
			fprintf(spp->spr_fpl, "\n");
		  }
		else fprintf(spp->spr_fpl, " * * * reconversion failed\n");
	   }
	/* clean up diffraction-vector profile too */
	cleanmodels(rgb, 1, 1);
	for (maskno = 0, rgp = rgb, rgq = rgb + lm;
	 maskno < lm; maskno++, rgp++, rgq++)
	   {
		lx = rgp->m3_qx; ly = rgp->m3_qy; lz = rgp->m3_qz;
		lxyz = lx * ly * lz;
		for (iz = 0; iz < 3; iz++) idv[iz] = delv[iz] = 0.;
		minten = gen_centroid(rgp->m3_dvp, NULL, lx, ly, lz,idv,delv);
		for (iz = 0; iz < 3; iz++) delv[iz] = delv[iz];
		recen_vec(rgp->m3_dvp, delv, rgq->m3_dvp, lx, ly, lz);
		for (iz = 0; iz < 3; iz++) idv[iz] = delv[iz] = 0.;
		minten = gen_centroid(rgq->m3_dvp, NULL,
			lx, ly, lz, idv, delv);
		/* add intensity on the wings if needed */
		if (lz > 1) added = dospread(rgq, 1, sumf, nupf, floorfrac);
		/* copy the corrected profile into the original */
		for (iz = 0, zvp = rgp->m3_dvp, zvi = rgq->m3_dvp;
		 iz < lxyz; iz++, zvp++, zvi++) *zvp = *zvi;
	   }
	show_masks(stdout, spp->spr_wmask, spp->spr_bdr);
	if (stdout != spp->spr_fpl)
		show_masks(spp->spr_fpl, spp->spr_wmask, spp->spr_bdr);
	/* write models out after making them so diagnostics are clearer */
	if (NULL == (spp->spr_fpmask = xgopen("wmask", "w")))
	  {
		fprintf(stderr, openmsg);	fprintf(spp->spr_fpl, openmsg);
		return EBADVALUE;
	  }
	for (maskno = 0, wmp = spp->spr_wmask; maskno < lm; maskno++, wmp++)
	   {
		if (0 >= xgwrite((void *)wmp, XG_MASK_3D, spp->spr_fpmask))
		  {
			fprintf(stderr, writmsg, maskno);
			fprintf(spp->spr_fpl, writmsg, maskno); return EWRITERR;
		  }
	   }
	retv = out_diffvmod(spp);
	(void)fclose(spp->spr_fpmask);	spp->spr_fpmask = NULL;
	if (stderr != spp->spr_fpl) xeflush(spp->spr_fpl);
	return retv;
}

static void out_lsq(double *flp, FILE *fp)
{
	if (*(flp + 4) > DFUZZ) fprintf(fp,
	 "I/sigma at stol = 0 is %7.3lf; Bsigma = %7.3lf; D2 = %8.3lf A\n",
	 *(flp + 1), *(flp + 2), *(flp + 4));
	else if (*(flp + 4) == 0.) fprintf(fp,
	 "I/sigma at stol = 0 is %7.3lf; Bsigma = %7.3lf; D2 undeterminable\n",
		*(flp + 1), *(flp + 2));
	else fprintf(fp, "Least-squares calculation failed\n");

}

static int spot_finish(SPOTRUN *spp)
{ /*	This cleans up and finishes a spot-finding effort. */
	int	i, spfail[15];
	float	flsq[8];
	double	dmin, loclam;

	(void)fclose(spp->spr_fpo);	spp->spr_fpo = NULL;
	zeromem((AVOID *)spfail, 15 * sizeof (int));
	(void)eftable(0, NULL);	/* zero out the memory for eftable */
	anal_peaks(spp, spfail);
	summ_spots(stdout, spfail);
	out_lsq(spp->spr_lsq, stdout);
	loclam = ((spp->spr_heads)->runheader == NULL) ? 1.5418 :
		((spp->spr_heads)->runheader)->run_lambda;
	if ((dmin = spp->spr_maxs) > DFUZZ) dmin = loclam / sqrt(dmin);
	fprintf(stdout,
	 "Highest-resolution reflection at s^2 = %9.5f, Dmin = %7.2lf A\n",
	 spp->spr_maxs, dmin);
	if (stdout != spp->spr_fpl)
	  {
		summ_spots(spp->spr_fpl, spfail);
		out_lsq(spp->spr_lsq, spp->spr_fpl);
		fprintf(spp->spr_fpl,
	 "Highest-resolution reflection at s^2 = %9.5f, Dmin = %7.2lf A\n",
		 spp->spr_maxs, dmin);
	  }
	for (i = 0; i < 6; i++) flsq[i] = spp->spr_lsq[i];
	flsq[5] = spp->spr_maxs;	flsq[6] = dmin; flsq[7] = spfail[0];
	(void)xgvariable("nspots", &(flsq[7]), 1);
	(void)xgvariable("ios0", &(flsq[1]), 1);
	(void)xgvariable("bsig", &(flsq[2]), 1);
	(void)xgvariable("d2", &(flsq[4]), 1);
	(void)xgvariable("maxs", &(flsq[5]), 1);
	(void)xgvariable("dmin", &(flsq[6]), 1);
	(void)xgvariable(NULL, NULL, 0);
	/* create the output 3D models and write them out */
	return spotkill(outspotmask(spp), spp);
}

static void alert_ending(FILE *fp, SPOTRUN *spp)
{ /*	This offers an alert regarding insufficient incoming reflections */
	fprintf(fp, " Only %5d reflections were found by frame %4d; ",
		(int)(spp->spr_peakp - spp->spr_pkbase), spp->spr_zma);
	fprintf(fp, " more frames will be searched\n");
	if (fp == spp->spr_fpl) xeflush(fp);
}

static int bordhere(SPOTRUN *spp, double totcounts)
{ /*	This does the border calculations in the `spots' environment */
	int		totincl;
	MAIL		*thd;
	static char	pixmsg[] =
 "\n\t%6d pixels will be included in the active-pixel-mask output\n";

	head_results(spp->spr_fpl);	thd = spp->spr_heads;
	totincl = process_total(spp->spr_fpl, totcounts, spp->spr_qvalu,
		thd->xdimen, thd->ydimen, 0);
	fprintf(spp->spr_fpl, pixmsg, totincl);
	if (spp->spr_fpl != stdout) fprintf(stdout, pixmsg, totincl);
	/* shrink mask by spp->spr_valshrink pixels on all sides if requested*/
	if (spp->spr_valshrink)
		shrink(spp->spr_valshrink, thd->xdimen, thd->ydimen);
	return (NULL == activmask((int)(thd->xdimen), (int)(thd->ydimen),
	 011)) ? EREADERR : EOK; /* output resulting structure to AMASK */
}

static void calc_lsq(SPOTRUN *spp)
{ /* This completes the least-squares analysis used to calculate I000, B,
	and D2 for the data. The D2 value is fudged to take into account
	the fact that we have perforce excluded weak reflections. */
	double	*flp, n0, n1, de, lam;

	flp = spp->spr_lsq;
	de = *flp * *(flp + 2) - *(flp + 1) * *(flp + 1);
	if (de < DFUZZ)
	  {
		*(flp + 4) = -1.;	return;
	  }
	n0 = (*(flp + 2) * *(flp + 3) - *(flp + 1) * *(flp + 4)) / de;
	n1 = (*flp * *(flp + 4) - *(flp + 1) * *(flp + 3)) / de;
	lam = ((spp->spr_heads)->runheader)->run_lambda;
	if (lam < FUZZ) lam = 1.5418;
	/* heuristic: fudge the B-value somewhat in all cases,
	 and even more so if we just have one image */
	n1 -= 2.5;
	if (spp->spr_zmi >= spp->spr_zma) n1 -= 0.6;
	*(flp + 1) = exp(n0);	*(flp + 2) = -4. * lam * lam * n1;
	*(flp + 3) = (log(2.)  - n0) / n1;
	*(flp + 4) = *(flp + 3) > DFUZZ ?  lam / sqrt(*(flp + 3)) : 0.;
}


int spotmain(int argc, char *argv[])
{ /*	This is the mainline for the "spots" program.  It reads in the
	specified frames and calls `peaks' once per frame to generate the
	initial list of peaks.  Each frame's peaks are merged with those
	from previous frames as soon as they're available.  Once we have
	enough frames in memory to construct complete (X,Y,phi) centroids
	and 3-dimensional, weighted model profiles, we start making them.
	We output the centroids as we go; the 3D profiles are output at
	the end of the effort. */
	int		cret, fmno, inform, nombg, origend, peakno, region, rf;
	double		totalcounts;
	MAIL		*lochp;
	PEAK		*curpk;
	REGIONPROFILE	*rgb;
	static SPOTRUN	sps;
	FRAMEINDEX	*lfmi;

	/* process args and set up program run */
	if (EOK != (cret = spot_setup(argc, argv, &sps)))
		return spotkill(cret, &sps);
	/* go through the frames in order.  On this pass, we simply look for
	 peaks within each frame and determine the (X,Y) centroid of each
	 peak within that frame. */
	origend = sps.spr_zma;	nombg = totalcounts = 0;
	sps.spr_passno = 1;
	/* experiment: */
	lfmi = sps.spr_fmind;
	if (sps.spr_zmi != lfmi->fi_firstfm)
	  {
		sps.spr_rfrm = sps.spr_ifrm = lfmi->fi_firstfm;
		if (EOK != eftable(3, &sps)) fprintf(stderr,
			 "Warning: eftable failed on first image\n");
	  }
	for (fmno = sps.spr_zmi; fmno <= sps.spr_zma; fmno++)
	   {
		if (EOK != (cret = get_peaks(fmno, &sps)))
			return spotkill(cret, &sps);
		lochp = sps.spr_heads + 1 + (fmno % (sps.spr_nmbox - 1));
		if (sps.spr_flags & P_DOBORDER)
			totalcounts += merge_frames(lochp, &nombg);
		/* summarize what we found */
		if (stdout != sps.spr_fpl)
		  sum_peaks(sps.spr_fpl, fmno, &sps, totalcounts, nombg);
		sum_peaks(stdout, fmno, &sps, totalcounts, nombg);
		/* extend the frame limit if we don't have enough refls */
		if ((fmno >= sps.spr_zma) && ((sps.spr_peakp - sps.spr_pkbase)
		 < sps.spr_continue_refls) &&
		 (sps.spr_zma < (sps.spr_fmind)->fi_lastfm))
		  {
			if (origend == sps.spr_zma)
			  {
				if (stdout != sps.spr_fpl)
					alert_ending(sps.spr_fpl, &sps);
				alert_ending(stdout, &sps);
			  }
			(sps.spr_zma)++;
		  }
		if ((fmno >= sps.spr_zma) && (EOK != (cret =
		 peakskill(EOK, &sps)))) return spotkill(cret, &sps);
	   }
	if (sps.spr_flags & P_DOBORDER) (void)bordhere(&sps, totalcounts);
	sps.spr_totalpeaks = sift_peaks(&sps);	/* re-sort and winnow list */
	/* second pass: improve centroid and intensity estimates for
	 merged peaks and devise some 3-D profiles */
	printf(" %5d spots were obtained from the first pass\n",
		sps.spr_totalpeaks);
	inform = (sps.spr_totalpeaks <= 9999) ?
		sps.spr_totalpeaks / 10 : sps.spr_totalpeaks / 5;
	if (EOK != (rf = init_models(&sps))) return spotkill(rf, &sps);
	(void)eftable(041, &sps); /*print list of spindle values */
	(void)eftable(0101, &sps);	/* see if spindles are contiguous */
	fprintf(sps.spr_fpl,
	 "Range of images is%scontiguous on left,%scontiguous on right\n",
		sps.spr_flags & P_CONTIGLEFT ? " " : " not ",
		sps.spr_flags & P_CONTIGRIGHT ? " " : " not ");
	rgb = sps.spr_rgb;	sps.spr_sgz = rgb->m3_bz;
	sps.spr_passno = 2;
	printf("pass %d: processing peak ", sps.spr_passno);
	for (curpk = sps.spr_pkbase, rf = peakno = 0;
	 curpk < sps.spr_pklast; curpk++, rf++)
	  if (curpk->pphicen < HUGEPHI - 1.)
	   {
		if ((EOK == (cret = fill_shoe(curpk, &sps))) && (EOK == (cret =
		 mkcentroid(curpk, &region, &sps))) &&
		 (curpk->pinten > 3. * curpk->psigma))
		  {
			(void)add2model(curpk, region, &sps);	peakno++;
		  }
		else if (cret == EALLOC) return spotkill(cret, &sps);
		if (!(rf % inform))
		  {
			printf("%4d ", rf); xeflush(stdout);
		  }
	   }
	printf("\n");
	/* clean up the learned profiles, copy them to the input slots,
	  and zero the output profiles */
	if (EOK != (rf = init_models(&sps))) return spotkill(rf, &sps);
	/* third pass: get even better centroids and intensities, since
	 the 3-D profiles will now be beautiful */
	sps.spr_passno = 3;
	zeromem((AVOID *)sps.spr_lsq, 5 * sizeof (double));
	printf("pass %d: processing peak ", sps.spr_passno);
	for (curpk = sps.spr_pkbase, rf = 0; curpk < sps.spr_pklast;
	 curpk++, rf++)
	  if (curpk->pphicen < HUGEPHI - 1.)
	   {
		if ((EOK == (cret = fill_shoe(curpk, &sps))) &&
		 (EOK == (cret = mkcentroid(curpk, &region, &sps))))
		  {
			curpk->pflags = P_USABLE;
			if (curpk->pinten > 3. * curpk->psigma)
				(void)add2model(curpk, region, &sps);
		  }
		else if ((cret == EALLOC) || (cret == EWRITERR))
			return spotkill(cret, &sps);
		else	curpk->pflags = cret;
		if (!(rf % inform))
		  {
			printf("%4d ", rf); xeflush(stdout);
		  }
	   }
	printf("\n");
	if (sps.spr_lsq[0] > 0.) calc_lsq(&sps);
	return spot_finish(&sps);
}

/* end of spotmain.c */
