/* spotshoes.c -- manipulates `shoeboxes' of data for `spots'

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	21 Mar 2007: record maximum |s|^2 value among encountered reflections
	24 Feb 2007: accum_lsq for I000,B calcs; called on last pass
	23 Feb 2007: mkcentroid: #pixels is #>bgnd, not #>2*bgnd.
	24 Jun 2006: mkcentroid: use israxis to set minlint value
	17 Jun 2006: replace act_pixel calls with activmask(x,y,041) call
	 6 Jun 2006: out_diffvmod moved to spotmain
	 3 Apr 2006: housekeeping in fill_shoe
	 5 Mar 2006: major revision: uses spindle and effective frame values;
		handles RANGEEDGE and Lorentz better; better diagnostics
	 2 Jan 2006: replace effecfram calls with eftable manipulations
	20 Nov 2005: inline function declarations & function prototypes
	28 Oct 2005: replace static globals with SPOTRUN elements, etc.
	12 Oct 2005: housekeeping
	28 Aug 2004: routines used only here are declared as static; cosmetics;
		move mymedian() to ../xglib/xgenmath.c
	28 May 2004: handle one-frame runs correctly in default_model()
	26 Jan 2004: some static globals converted to local variables
	21 Jan 2004: license message inserted and initial comment reordered
	 6 Jan 2004: SPOTRUN structure elements referenced in function calls
		& static globals replace extern globals;
		callop() & memzap() replace shcalloc() & shfree()
	13 Dec 2003: remove unused variables mx2,my2,mz2 from kmodel()
	23 Nov 2003: replace globals with elements of SPOTRUN structure;
		call kablim() even if we already know plated in init_models()
	20 Sep 2003: initialize wni to NULL in init_models()
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Sep 2003: eliminate access() call for WMASK file--it's
		implied in the xgopen() call.
	 2 May 2003: set lochp pointer more explicitly in fill_shoe()
	15 Apr 2003: use different header for each image examined in fill_shoe
	25 Mar 2003: correct bug in overlap-checking in mkcentroid()
	 5 Jan 2003: reorganized source-file header
	 1 Nov 2002: greater flexibility in processing;
		altered management of the altered REGIONPROFILE structures;
		functionality associated with making and
		outputting binary 3-D models, including Kabsch 3-D profiles
	25 Mar 2002: removed MSI attachments
	 8 Nov 1996: minor changes for X-GEN checkin at MSI
	10 Oct 1995: passes 2&3: preserve image-by-image memory allocation
		for overflow list and runheaders to save on allocations
		and prevent memory leaks
	30 Sep 1995: rename photperadu to aduperphot
	25 Aug 1995: `default_model' routine available
		for creating fake profiles
	24 Aug 1995: read shoebox values as ints, not u_shorts
		don't allow made-up profiles to be excessively flat
		cleaner printout during model-profile cleanups
	21 Jul 1995: major rewrite: 2nd pass thru data replaced
		by a pair of passes,
		each containing an examination of locally-read 3-D profiles.
	12 Apr 1995: replace s_char with sichar
	 2 Mar 1995: frame-specific detector image offsets;
		shared-memory alloc and free
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
		very large backgrounds handled correctly
	 1 Dec 1994: adjustable cutoff of # of low-count pixels
		examine elements of PEAK structure, not TROID
	 5 Oct 1994: use xgwrite to output TROID information
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN;
		more extern functions declared
	16 Mar 1994: significant rewrite both for `peaks' and `spotshoes'
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version
	
   Purpose: manipulating 3-D blocks of data ("shoeboxes") used in dealing
	with reflection centroids and in building up model profiles
	as part of the "spots" functionality.
   Cross references:
	spotmain.c
*************************************************************************/
#include	"spotdefs.h"

#define			SIZOUTREF	(20)
#define			MXEDGE	0.02	/* shrink bz2 if 1st&last fms < this */
#define	MIN_EFFSCALE	0.1
#define	MAX_EFFSCALE	64.
#define	MIN_WINGS	0.03333
#define	WING_RATIO	0.03
#define	CELLS_NEEDED	8000

extern void abort(void);
extern int israxis(int fomt);

void default_model(float *te0, SPOTRUN *spp)
{ /*	This generates a default profile in the vector te0. It's a
	Gaussian sphere truncated where the coefficient of the
	exponential is less than -7. It also checks that the elements
	in z to the left and right of center are really a bit nonzero;
	if they're not, it pushes a little intensity into them. */
	register int	x, y, z;
	int		lx, lx2, lxy, ly, ly2, lz, lz2;
	double		coef, dyz2, dz2, effscale, phixscale, wingsum;
	float		*flp, *flpe;
	REGIONPROFILE	*rgb;

	phixscale = spp->spr_phitoxscale;
	effscale = flimits(phixscale, MIN_EFFSCALE, MAX_EFFSCALE);
	rgb = spp->spr_rgb;
	lx = rgb->m3_bx;	ly = rgb->m3_by;	lz = rgb->m3_bz;
	lx2 = lx / 2;	ly2 = ly / 2;	lz2 = lz / 2;	lxy = lx * ly;
	for (z = -lz2, flp = te0; z <= lz2; z++)
	   {
		dz2 = effscale * (double)(z * z);
		for (y = -ly2; y <= ly2; y++)
		   {
			dyz2 = dz2 + (double)(y * y);
			for (x = -lx2; x <= lx2; x++, flp++)
			   {
				coef = -0.0625 * (dyz2 + (double)(x * x));
				if (coef > -7.) *flp = exp(coef);
				else *flp = 0.;
			   }
		   }
	   }
	/* check to see if any z values to the left and right
	 of the centermost one are populated.  If not, stick
	 something modest in the adjoining ones. If lz2 <= 0
	 we don't even try this, because we know they're too narrow */
	if (lz2 < 1) return;
	flp = te0 + lxy * (lz2 - 1);	flpe = flp + lxy;
	for (wingsum = 0.; flp < flpe; flp++) wingsum += *flp;
	flp += lxy;	flpe = flp + lxy;
	for ( ; flp < flpe; flp++) wingsum += *flp;
	if (wingsum < FUZZ)
	  {
		flp = te0 + lxy * lz2;	flpe = flp + lxy;
		for ( ; flp < flpe; flp++)
		   if ((coef = *flp) > MIN_WINGS)
			*(flp - lxy) = *(flp + lxy) = coef * WING_RATIO;
	  }
}

static int cenerr(int ert, PEAK *tpk, SPOTRUN *spp)
{ /* This returns its `ert' argument after printing diagnostics if needed.
	In the special case in which the *tpk argument is NULL,
	it is used to reset `alreadhead' to zero */
	int		ison, lowbiton, etype;
	static int	alreadyhead;
	static char	*ertstr[] = {
	 "    good", " dimpled", " XY edge", "phishape", "  Zrange",
	 " overlap", "too weak", "move2far", "wasmergd", " readerr",
	 "toosharp", " belowBG", "noLorenz", "badLoren" } ;

	if (NULL == tpk)
	  {
		alreadyhead = 0;	return EOK;
	  }
	etype = spp->spr_etype;
	if ((spp->spr_flags & P_VERBOSEANY) && (tpk->pphicen < HUGEPHI - 1.))
	  {
		if (!alreadyhead)
		  {
			fprintf(spp->spr_fpl,
	" pass    error stage    Xcen    Ycen   Zcen meanbg");
			fprintf(spp->spr_fpl,
			 " medianbg  #pts outpx lowpx Intensity Winginten\n");
			alreadyhead = 1;
		  }
		for (ison = 1, lowbiton = 1;
		 lowbiton < 14; lowbiton++, ison *= 2) if (ert & ison) break;
		if (lowbiton == 14) lowbiton = 0;
		fprintf(spp->spr_fpl, " %4d %8s %5d %7.1f %7.1f %6.1f %6.1f",
		 spp->spr_passno, ertstr[lowbiton], etype,
		 tpk->pxcen, tpk->pycen, tpk->pphicen, spp->spr_fbgnd);
		if (!etype)
		  {
			fprintf(spp->spr_fpl, "\n");	return ert;
		  }
		fprintf(spp->spr_fpl, " %8.2f", spp->spr_medbg);
		if (etype == 1)
		  {
			fprintf(spp->spr_fpl, "\n");	return ert;
		  }
		else if (etype == 12)
		  {
			fprintf(spp->spr_fpl,
			 " peakfm = %2d -- too close to start of run\n",
				spp->spr_badi); return ert;
		  }
		else if (etype == 13)
		  {
			fprintf(spp->spr_fpl,
			 " peakfm = %2d -- too close to end of run\n",
			 spp->spr_badi);	return ert;
		  }
		else if (etype == 14)
		  {
			fprintf(spp->spr_fpl, " inten=%10.3e is too weak\n",
				spp->spr_badf);	return ert;
		  }
		else if (etype == 15)
		  {
			fprintf(spp->spr_fpl, " Xcentroid moved %6.1f pix\n",
				spp->spr_badf);	return ert;
		  }
		else if (etype == 16)
		  {
			fprintf(spp->spr_fpl, " Ycentroid moved %6.1f pix\n",
				spp->spr_badf);	return ert;
		  }
		else if (etype == 17)
		  {
			fprintf(spp->spr_fpl, " Zcentroid moved %6.1f fms\n",
				spp->spr_badf);	return ert;
		  }
		else if (etype == 18)
		  {
			fprintf(spp->spr_fpl, " inten=%10.3e is too weak\n",
				spp->spr_badf);	return ert;
		  }
		else if (etype == 19)
		  {
			fprintf(spp->spr_fpl, " delXcen=%6.2f after 4 tries\n",
				spp->spr_badf);	return ert;
		  }
		else if (etype == 20)
		  {
			fprintf(spp->spr_fpl, " delYcen=%6.2f after 4 tries\n",
				spp->spr_badf);	return ert;
		  }
		else if (etype == 21)
		  {
			fprintf(spp->spr_fpl, " delZcen=%6.2f after 4 tries\n",
				spp->spr_badf);	return ert;
		  }
		fprintf(spp->spr_fpl, " %5d %5d %5d %9.2f %9.2f",
		 spp->spr_npts, spp->spr_outpx, spp->spr_lowpx,
		 spp->spr_locint, spp->spr_outom);
		if (etype == 2)
			fprintf(spp->spr_fpl, " zcen = %d\n", spp->spr_badi);
		else if (etype == 3) fprintf(spp->spr_fpl, "\n");
		else if (etype == 4)
		 fprintf(spp->spr_fpl, " doz = %d\n", spp->spr_badi);
		else if (etype == 5) fprintf(spp->spr_fpl,
		 " censum = %8.2f < %8.2f\n",
		 spp->spr_badf, spp->spr_mincenter);
		else if (etype == 6)
		 fprintf(spp->spr_fpl, " doz0 = %d\n", spp->spr_badi);
		else if (etype == 7)
		 fprintf(spp->spr_fpl, " doze = %d\n", spp->spr_badi);
		else if (etype == 8) fprintf(spp->spr_fpl,
		 " dimpled:dnext,badf=%d,%6.3f\n",
		 spp->spr_badi, spp->spr_badf);
		else if (etype == 9) fprintf(spp->spr_fpl,
		 " tx=%7.1f>%7.1f\n", spp->spr_badf, spp->spr_maxmove);
		else if (etype == 10) fprintf(spp->spr_fpl,
		 " ty=%7.1f>%7.1f\n", spp->spr_badf, spp->spr_maxmove);
		else if (etype == 11) fprintf(spp->spr_fpl,
		 " tz=%7.1f>%7.1f\n", spp->spr_badf, spp->spr_maxmove);
		else fprintf(spp->spr_fpl, "\n");
	  }
	return ert;
}

int init_models(SPOTRUN *spp)
{ /*	This creates and initializes the 3-D models used in the second
	pass through the data--both the working integer vector `spp->spr_shoe'
	and the REGIONPROFILE structures in the various detector regions.
	In this version we set up input model profiles too and stick them in
	as regions 12-23. If we have some models out there already we use 'em;
	if not, we invent Gaussian models. */
	register int	z;
	char		*wni;
	int		boxyz, cannarrow, lm, lx, lxy, lxyz, ly, lz, lz2;
	int		mx, mxy, mxyz, my, mz, obxyz, pz2, reg, y, nupf[2];
	float		*flp, *flpi;
	double		mhigh, minten, mlow, msum, sumedge;
	double		delv[3], idv[3], lellp[14], sumf[3];
	FILE		*fpml;
	REGIONPROFILE	*rgb, *rgi, *rgp;
	MASK_3D		*wmp;
	static float	*te0, *tes, *tprop;
	static char	chmsg[] =
	 "Changing model-profile width from %2d to %2d frames\n";

	/* don't WAIT_IF_ABSENT on the model-profile pass */
	spp->spr_imfl = 0;
	pz2 = PFRAMES / 2;
	/* for now we fix these values */
	if (spp->spr_passno == 1)
	  {
		rgb = NULL;	lx = BIGSPOTWID;	ly = BIGSPOTHT;
		lz = spp->spr_sgz; mx = 13;	my = 13;	mz = 9;
	  }
	else {
		rgb = spp->spr_rgb; lx = rgb->m3_bx; ly = rgb->m3_by;
		lz = rgb->m3_bz;
		mx = rgb->m3_qx; my = rgb->m3_qy; mz = rgb->m3_qz;
	     }
	lm = NREGIONS;	lxy = lx * ly;	lxyz = lxy * lz;
	lz2 = lz / 2;	wni = NULL;
	mxy = mx * my;	mxyz = mxy * mz;	boxyz = lxyz + mxyz;
	if (spp->spr_passno < 2)
	  {
		if (NULL == (AVOID *)(rgb = (REGIONPROFILE *)callop(
		 lm * 2, sizeof (REGIONPROFILE),
		 "rgb in init_models"))) return EALLOC;
		spp->spr_rgb = rgb;
		if (NULL == (AVOID *)(spp->spr_rprop = (float *)callop(
		 lm * 2 * boxyz, sizeof (float), "rprop in init_models")))
			return EALLOC;
		/* set up memory for signed-char versions of 3D profiles*/
		if (NULL == (spp->spr_sic = (sichar *)mallop(boxyz,
			"sic in init_models"))) return EALLOC;
		if (NULL == (AVOID *)(spp->spr_shoe = (int *)callop(lxyz,
		 sizeof (int), "spp->spr_shoe in init_models"))) return EALLOC;
		/* get memory for wmask now in case we can use it for input */
		if (NULL == (AVOID *)(spp->spr_wmask = (MASK_3D *)callop(lm,
		 sizeof (MASK_3D), "spp->spr_wmask in init_models")))
			return EALLOC;
		/* set up the working and input model profiles */
		flp = spp->spr_rprop;
		flpi = spp->spr_rprop + lm * 2 * lxyz;
		for (rgp = rgb, reg = 0; reg < 2 * lm;
		 reg++, rgp++, flp += lxyz, flpi += mxyz)
		   {
			rgp->m3_bx= lx;		rgp->m3_by = ly;
			rgp->m3_bz = lz;	rgp->m3_qx = mx;
			rgp->m3_qy = my;	rgp->m3_qz = mz;
			rgp->m3_xyop= flp;	rgp->m3_dvp = flpi;
			rgp->m3_xyoip = spp->spr_sic;
			rgp->m3_dvip = spp->spr_sic + lxyz;
			rgp->m3_imodel = reg % lm;
			rgp->m3_nmodels = lm;
			rgp->m3_isqsum = rgp->m3_nobs = 0;
		   }
		/* see if we can read in some models from the outside world */
		wni = trname("WMASK");
		if (NULL != (fpml = xgopen(wni, "r")))
		  {
			te0 = rgb->m3_xyop + lxyz * lm; /* scratch mem*/
			for (reg = 0, rgp = rgb, wmp = spp->spr_wmask;
			 reg < lm; reg++, rgp++, wmp++)
			   {
				if (0 >= xgread((void *)wmp,XG_MASK_3D,fpml))
				  {
					wni = NULL;	break;
				  }
				else if (EOK == wmptovec(wmp, te0))
				  {
					tes = te0 + (pz2 - lz2) * lxy;
					cpymem(rgp->m3_xyop, te0,
						lxyz * sizeof (float));
				  }
				else {
					wni = NULL;	break;
				     }
			   }
			(void)fclose(fpml);	fpml = NULL;
		  }
		else wni = NULL;
		/* if that failed in any way, make up some profiles */
		if (NULL == wni)
		  {
			te0 = rgb->m3_xyop;	default_model(te0, spp);
			/* stick those in all the other slots too */
			for (tes = te0 + lxyz, reg = 1; reg < lm;
			 reg++, tes += lxyz)
			   cpymem((void *)tes, te0, lxyz * sizeof (float));
		  }
		if (spp->spr_plated < FUZZ)
		  {
			if (NULL == spp->spr_spl) return EREADERR;
			kablim(spp->spr_heads, rgb, lz);
			tell_dvcoef(spp->spr_fpl, rgb);
		  }
		else if (fabs(rgb->m3_lim[2]) < DFUZZ)
		  {
			if (NULL == spp->spr_spl) return EREADERR;
			kablim(spp->spr_heads, rgb, lz);
			tell_dvcoef(spp->spr_fpl, rgb);
		  }
	  }
	/* now we have something semi-reasonable in rgb. Center and
	 renormalize that and stick it in (rgb + lm), and zero
	 out the working profiles rgb. */
	if (spp->spr_passno > 1)
	  {
		fprintf(spp->spr_fpl,
		 "Three-dimensional models found on pass %d:\n",
		 spp->spr_passno);
		fprintf(spp->spr_fpl,
		 "region    fraction     number of pixels\n");
		fprintf(spp->spr_fpl,
		 "number  in edge frames  zeroed unzeroed\n");
	  }
	for (rgp = rgb, cannarrow = reg = 0, rgi = rgb + lm;
	 reg < lm; reg++, rgp++, rgi++)
	   {
		if (rgp->m3_bz > 1)
		  dospread(rgp, 0, sumf, nupf, 0.015);	/* spread it out */
		if (blankf(rgp->m3_xyop, lz)) meanprofile(rgb, reg);
		for (y = 0; y < 3; y++) idv[y] = delv[y] = 0;
		/* find the centroid of the adjusted profile */
		minten = gen_centroid(rgp->m3_xyop, NULL,lx, ly, lz, idv,delv);
		if (DFUZZ < minten) /* ... and re-center it */
			recen_vec(rgp->m3_xyop, delv, rgi->m3_xyop, lx,ly,lz);
		else	cpymem((void *)(rgi->m3_xyop), (void *)(rgp->m3_xyop),
				lxyz * sizeof (float));
		for (minten = 0., z = lxyz, flp = rgi->m3_xyop; z--; flp++)
			if (*flp > minten) minten = *flp;
		if (minten < DFUZZ)
		  {
			if (spp->spr_passno > 1) fprintf(spp->spr_fpl,
			  " %5d intensity(profile) <= 0\n",reg);
			continue;
		  }
		for (z = lxyz, flp = rgi->m3_xyop; z--; flp++) *flp /= minten;
		/* do a quick gaussian cleanup on this profile:
		 flatten and extrapolate with cutoff = 0.005 */
		if (lz < 2) y = ellip(rgi, 1, 014, 0.005, 4., lellp);
		else y = ellsoid(rgi, 1, 014, 0.005, 4., lellp);
		minten = msum = sumedge = 0.;
		for (z = 0, flp = rgi->m3_xyop; z < lxyz; z++, flp++)
		   {
			msum += *flp;	if (*flp > minten) minten = *flp;
			if ((lz > 1) && ((z < lxy) || (z >= (lxyz - lxy))))
				sumedge += *flp;
		   }
		if (msum <= FUZZ) 
		  {
			if (spp->spr_passno > 1) fprintf(spp->spr_fpl,
				" %5d profile is flat\n", reg);
			continue;
		  }
		if (minten > FUZZ) for (z = lxyz, flp = rgi->m3_xyop;
		 z--; flp++) *flp /= minten;
		/* see if the resulting profile could be narrowed */
		if ((lz2 > 1) && (msum > FUZZ) && (sumedge < MXEDGE * msum))
			cannarrow++;
		if (spp->spr_passno > 1)
		  {
		    fprintf(spp->spr_fpl, " %5d %15.5f ", reg, sumedge / msum);
		    switch (y)	/* describe ellipsoid's properties briefly */
		     {
			case	0:	fprintf(spp->spr_fpl,
			 "%7.0lf %8.0lf\n", lellp[10],lellp[12]);	break;
			case	EBADVALUE: fprintf(spp->spr_fpl,
			 "no reflections found in region\n");	break;
			case	ESINGULAR:	case -1: fprintf(spp->spr_fpl,
			 "ellipsoid-matrix is singular\n");	break;
			case	EALLOC:	fprintf(spp->spr_fpl,
			 "ellipsoid memory allocation err\n");	break;
			case	-2:	fprintf(spp->spr_fpl,
			 "trace(ellipsoid matrix) <= 0\n");	break;
			case	-3:	fprintf(spp->spr_fpl,
			 "determ.(ellipsoid matrix) <= 0\n");	break;
			case	-4:	fprintf(spp->spr_fpl,
			 "discrim.(ellipsoid matrix) <= 0\n");	break;
			case	-5:	case	-7:	fprintf(spp->spr_fpl,
			 "ellipsoid eigenvalues unknown\n");	break;
			case	-6:	case	-8:	fprintf(spp->spr_fpl,
			 "ellipsoid eigenvectors unknown\n");	break;
			case	-9:	fprintf(spp->spr_fpl,
			 "ellipsoid eigenvalues faulty\n");	break;
			default:	fprintf(spp->spr_fpl,
			 "error %d finding ellipsoid params\n", y);	break;
		     }
		  }
		rgp->m3_isqsum = rgp->m3_nobs = 0;
		zeromem((void *)(rgp->m3_xyop), lxyz * sizeof (float));
	   }
	/* the remaining stuff is irrelevant if bz == 1 */
	if (lz < 2) return EOK;
	/* don't try to narrow it if we're starting from made-up profiles */
	if ((spp->spr_passno < 2) && (NULL == wni)) cannarrow = 0;
	if (cannarrow == lm)	/* narrow profile if possible */
	  {
		lz2--;	lz -= 2;	obxyz = lxyz;	lxyz = lxy * lz;
		boxyz = lxyz + mxyz;
		if (NULL == (AVOID *)(tprop = (float *)callop(
		 lm * 2 * boxyz, sizeof (float),
		 "tprop in init_models"))) return EALLOC;
		printf(chmsg, lz + 2, lz);
		if ((spp->spr_fpl != stderr) && (spp->spr_fpl != stdout))
			fprintf(spp->spr_fpl, chmsg, lz + 2, lz);
		for (flp = tprop, reg = 0, rgi = rgb;
		 reg < 2 * lm; reg++, rgi++)
		   {
			flpi = rgi->m3_xyop + lxy;
			for (z = lxyz; z--; flpi++, flp++) *flp = *flpi;
		   }
		/* move the diffraction-vector-space profiles */
		flp = tprop + 2 * lxyz * lm;
		flpi = spp->spr_rprop + 2 * obxyz * lm;
		for (reg = 0; reg < 2 * lm;
		 reg++, flp += mxyz, flpi += mxyz)
		   {
			cpymem((AVOID *)flp, (AVOID *)flpi,
			 (mxyz * sizeof (float)));
		   }
		/* obliterate the old memory and re-set it */
		memzap((AVOID **)&(spp->spr_rprop), "rprop in init_models");
		flp = spp->spr_rprop = tprop;
		flpi = spp->spr_rprop + lm * 2 * lxyz;
		for (rgp = rgb, reg = 0; reg < 2 * lm;
		 reg++, rgp++, flp += lxyz, flpi += mxyz)
		   {
			rgp->m3_bx= lx;		rgp->m3_by = ly;
			rgp->m3_bz = lz;	rgp->m3_qx = mx;
			rgp->m3_qy = my;	rgp->m3_qz = mz;
			rgp->m3_xyop= flp;	rgp->m3_dvp = flpi;
			rgp->m3_isqsum = rgp->m3_nobs = 0;
			rgp->m3_nmodels = lm;	rgp->m3_imodel = reg;
		   }
		cannarrow = 0;
			/* we have to recalc the gamma values too */
		kablim(spp->spr_heads, rgb, lz);
	  }
	/* record whether or not the zeroth and final frames of any given
	 model are low enough that they can be ignored for reflections on
	 the edges of the available range. Thus if the zeroth frame only has
	 1% of the total intensity, then a reflection with centroid at
	 (spp->spr_zmi + 2) can be considered usable */
	for (reg = 0, rgi = rgb + lm; reg < lm; reg++, rgi++)
	   {
		flp = rgi->m3_xyop;
		for (mlow = z = 0; z < lxy; z++, flp++) mlow += *flp;
		for (msum = mlow; z < lxyz - lxy; z++, flp++) msum += *flp;
		for (mhigh = 0.; z < lxyz; z++, flp++) mhigh += *flp;
		msum += mhigh;
		spp->spr_ignorelow[reg] =
			(msum > FUZZ) && (mlow < msum * MXEDGE);
		spp->spr_ignorehigh[reg] =
			(msum > FUZZ) && (mhigh < msum * MXEDGE);
	   }
	(void)cenerr(0, NULL, NULL);	/*signal new diagnostic header needed*/
	return EOK;
}

static void show_shoe(FILE *fp, PEAK *curpk, char *str, SPOTRUN *spp)
{ /*	This does a printed representation of a reflection shoebox. Pixels
	close to background are printed as blanks; above background we get
	0-9, a-z, A-Z, and *; below background we get '_', '!', and '<'.
	Apr 03: show real frame number, not just the offset from the start */
	register int	ix, iy, iz;
	int		bz2, loz, lx, lx2, lxy, ly, ly2, lz, lz2, lz2l, lz2r;
	int		lbgnd, maxv, norm, zoff, *shv, *shvy, *shvz, zcen;
	sichar		cval;
	char		dval;
	double		fbgnd, scale, thisf;
	REGIONPROFILE	*rgb;

	/* first, calculate a normalization factor */
	maxv = *(spp->spr_shoe);	lbgnd = fbgnd = curpk->pbgnd;	
	rgb = spp->spr_rgb;
	lx = rgb->m3_bx;	ly = rgb->m3_by;
	lx2 = lx / 2;	ly2 = ly / 2;	lxy = lx * ly;
	zcen = spp->spr_ifrm;
	lz2l = zcen - (curpk->ptroi).obkmin;	lz = (curpk->ptroi).nobk;
	lz2r = lz - 1 - lz2l;	lz2 = (lz2l + lz2r) / 2;
	shv = spp->spr_shoe;	bz2 = rgb->m3_bz / 2;
	if (bz2 > lz2l) shv += lxy * (bz2 - lz2l);
	for (iz = -lz2l, maxv = *(spp->spr_shoe); iz <= lz2r; iz++)
	  for (iy = -ly2; iy <= ly2; iy++)
	   for (ix = -lx2; ix <= lx2; ix++, shv++)
	    if (*shv > maxv) maxv = *shv;
	norm = maxv - lbgnd;	if (norm < 1) return;
	scale = 61. / (double)norm;
	fprintf(fp,
 " I = %8.1f+/-%7.2f @ [%6.1f,%6.1f,%7.2f]; maxval,bg=%7d,%6.1f; %s\n   ",
		 curpk->pinten, curpk->psigma, curpk->pxcen, curpk->pycen,
		 curpk->pphicen, (int)maxv, curpk->pbgnd, str);
	(void)cenerr(0, NULL, NULL);	/* to reset headers for cenerr */
	spp->spr_spind = curpk->pphicen;
	if (EOK != eftable(5, spp))
	  {
		fprintf(fp, "show_shoe: eftable error at [%7.2f,%7.2f,%7.2f]\n",
		 curpk->pxcen, curpk->pycen, curpk->pphicen);
		return;
	  }
	loz = spp->spr_ifrm;	zoff = loz - lz2l;
	for (iz = 0; iz < lz; iz++) fprintf(fp, ".__frame %3d____", iz + zoff);
	fputs(".\n", fp);
	shvy = spp->spr_shoe;	if (bz2 > lz2l) shvy += lxy * (bz2 - lz2l);
	for (iy = 0; iy < ly; iy++, shvy += lx)
	   {
		fputs("  ", fp);
		if (!(iy % 2)) fprintf(fp, "%1x", iy); else fputc((int)' ', fp);
		for (iz = 0, shvz = shvy; iz < lz; iz++, shvz += lxy)
		   {
			fputc((int)'|', fp);
			for (ix = 0, shv = shvz; ix < lx; ix++, shv++)
			   {
				thisf = (((double)(*shv)) - fbgnd) * scale;
				cval = thisf + 0.49999;
				if (thisf < -20.) dval = '<';
				else if (thisf < -5.) dval = '!';
				else if (thisf < -0.5) dval = '-';
				else if (thisf <= 0.4) dval = ' ';
				else if (thisf < 1.) dval = '.';
				else if (thisf >= 61.) dval = '*';
				else if (1 <= cval && cval <= 9)
					dval = cval + '0';
				else if (10 <= cval && cval <= 35)
					dval = (cval - 10) + 'a';
				else if (36 <= cval && cval <= 61)
					dval = (cval - 36) + 'A';
				else dval = '*';
				fputc((int)dval, spp->spr_fpl);
			   }
		   }
		if (iy == 5) fputs("| near\n", spp->spr_fpl);
		else if (iy == 6) fprintf(spp->spr_fpl, "| [%7.2f,\n",
			curpk->pxcen);
		else if (iy == 7) fprintf(spp->spr_fpl, "|  %7.2f,\n",
			curpk->pycen);
		else if (iy == 8) fprintf(spp->spr_fpl, "|  %7.2f]\n",
			curpk->pphicen);
		else fputs("|\n", spp->spr_fpl);
	   }
	fputs("   ", spp->spr_fpl);
	for (iz = 0; iz < lz; iz++) fputs(".0_2_4_6_8_a_c_e", spp->spr_fpl);
	fprintf(spp->spr_fpl, ".\n");
}

static int fmbefore(int zcen, PEAK *tpk, SPOTRUN *spp)
{ /* This returns the number of frames prior to the current one that
	should actually be considered in summations. Ordinarily this
	will just be lz2, but if the previous or next images are
	actually non-contiguous with the current one, it may be smaller. */
	int	lz, lz2, thisz;
	double	rangemax;

	lz = (spp->spr_rgb)->m3_bz;
	lz2 = lz / 2;	if (lz2 < 1) return 0;
	rangemax = 1.99 * fabs(spp->spr_framrad) / RADC;
	for (thisz = zcen - 1; thisz >= zcen - lz2; thisz--)
	   {
		spp->spr_rfrm = spp->spr_ifrm = thisz;
		if ((thisz < spp->spr_zmi) ||
		 (EOK != eftable(011, spp)) || fabs(tpk->pphicen -
		 spp->spr_spind) > rangemax * ((double)(zcen - thisz)))
			return zcen - (thisz + 1);
	   }
	return lz2;
}

static int fmafter(int zcen, PEAK *tpk, SPOTRUN *spp)
{ /* This defines how many frames after the current one we should examine */
	int	lz, lz2, thisz;
	double	rangemax;

	lz = (spp->spr_rgb)->m3_bz;
	lz2 = lz / 2;	if (lz2 < 1) return 0;
	rangemax = 1.99 * fabs(spp->spr_framrad) / RADC;
	for (thisz = zcen + 1; thisz <= zcen + lz2; thisz++)
	   {
		spp->spr_rfrm = spp->spr_ifrm = thisz;
		if ((thisz > spp->spr_zma) ||
		 (EOK != eftable(011, spp)) || fabs(tpk->pphicen -
		 spp->spr_spind) > rangemax * ((double)(thisz - zcen)))
			return thisz - (zcen + 1);
	   }
	return lz2;
}

int fill_shoe(PEAK *tpk, SPOTRUN *spp)
{ /* Populate the shoebox from data in a particular peak */
	register int	x, y, z, *shv;
	int		bz2, lbgnd, lx, lx2, lxy, lxyz, ly, ly2, lz;
	int		lz2l, lz2r, modz, region, xcen, ycen, zcen;
	int		*shvx, *shvy, *shvz, xlim[6];
	char		*framename;
	MAIL		*lochp;
	REGIONPROFILE	*rgb;

	lbgnd = tpk->pbgnd + 0.49999;
	region = getregion((short)(tpk->pxcen), (short)(tpk->pycen),
		/* tpk->pphicen, */ spp->spr_bdr);
	rgb = spp->spr_rgb;
	lx = rgb->m3_bx;	ly = rgb->m3_by;
	lxy = lx * ly;	lx2 = lx / 2;	ly2 = ly / 2;
	/* check the range of X, Y, and Z. Do Z first for speed.
	 Simultaneously establish the limits in Z of the box */
	spp->spr_spind = tpk->pphicen;
	if (EOK != eftable(5, spp))
	  {
		spp->spr_etype = 12;	spp->spr_badi = (tpk->ptroi).frame0;
		return cenerr(P_RANGEEDGE, tpk, spp);
	  }
	spp->spr_fbgnd = spp->spr_medbg = tpk->pbgnd;
	shvz = spp->spr_shoe;
	zcen = spp->spr_rfrm + 0.49999;
	lz2l = fmbefore(zcen, tpk, spp);	lz2r = fmafter(zcen, tpk, spp);
	lz = 1 + lz2l + lz2r;	lxyz = lxy * lz;
	xlim[4] = zcen - lz2l;	xlim[5] = zcen + lz2r;
	(tpk->ptroi).obkmin = xlim[4];
	(tpk->ptroi).nobk = 1 + xlim[5] - xlim[4];
	/* stick background in missing segments of profile */
	bz2 = rgb->m3_bz / 2;	shvz = spp->spr_shoe;
	if (lz2l < bz2)
	  {
		y = lxy * (bz2 - lz2l);
		for (shv = spp->spr_shoe, z = 0; z < y; z++, shv++)
			*shv = lbgnd;
		shvz = spp->spr_shoe + y;
	  }
	if (lz2r < bz2)
	  {
		x = lxy * (lz + bz2 - lz2l);	y = lxy * (bz2 - lz2r);
		for (shv = spp->spr_shoe + x, z = 0; z < y; z++, shv++)
			*shv = lbgnd;
	  }
	ycen = tpk->pycen + 0.49999;
	if (ycen < (ly2 + 2)) return P_XYEDGE;
	else if (ycen >= (spp->spr_deth - ly2 - 3)) return P_XYEDGE;
	xcen = tpk->pxcen + 0.49999;
	if (xcen < (lx2 + 2)) return P_XYEDGE;
	else if (xcen >= (spp->spr_detw - lx2 - 3)) return P_XYEDGE;
	/* establish the limits in (X,Y) of the (X,Y,Z) box */
	xlim[0] = xcen - lx2;	xlim[1] = xcen + lx2;
	xlim[2] = ycen - ly2;	xlim[3] = ycen + ly2;
	/* read in each slice of the box into the right part of the shoebox */
        for (z = xlim[4]; z <= xlim[5]; z++, shvz += lxy)
           { /* as of Apr 03, for each frame we choose a different header */
		modz = spp->spr_nmbox - 1;
		if ((modz < 0) || (modz >= spp->spr_nmbox))
		  {
			fprintf(stderr,
 "Bad header designation: modz = %d, should be between %d and %d inclusive\n",
			 modz, 0, spp->spr_nmbox - 1);
			return P_LOCALREADERR;
		  }
		lochp = spp->spr_heads + 1 + (z % modz);
		if ((NULL == (framename = indexname(z)))
		 || ('\0' == *framename))
		  {
			if (spp->spr_flags& P_VERBOSEANY) fprintf(spp->spr_fpl,
				"indexname failed on number %d\n", z);
			fprintf(stderr, "We were trying to get frame %d\n", z);
			return P_RANGEEDGE;
		  }
		if (FRAME_OK != (y = read_box(framename, spp->spr_imfl,
			lochp, xlim, shvz)))
		  {
			printf(
 "failure %d from read_box in image %s near [%3d,%3d,%3d]\n",
			 y, framename, xlim[0] +7, xlim[2] +7, z);
			return P_LOCALREADERR;
		  }
	   }
	/* if we are excluding stuff outside the border, set those vals to bg*/
	if (spp->spr_flags & P_USEBORDER)
	  {
		for (y = xlim[2], shvy = spp->spr_shoe;
		 y <= xlim[3]; y++, shvy += lx)
		  for (x = xlim[0], shvx = shvy; x <= xlim[1]; x++, shvx++)
		    if (NULL == activmask(x, y, 041))
		      {
			for (shv = shvx, z = xlim[4]; z <= xlim[5];
			  z++, shv += lxy) *shv = lbgnd;
		      }
	  }
	return EOK;
}

static void accum_lsq(PEAK *tpk, SPOTRUN *spp)
{ /* This auguments the sums used in the i000, B, and D2 calculations.
	It also checks to see if the s^2 value is the maximum one recorded */
	int	iux, iuy;
	double	delpa, tios, ldv[3];

	if (((tios = tpk->psigma) < 0.01) ||
	 ((tios = tpk->pinten / tios) < 1.1)) return;
	iux = tpk->pxcen + 0.49999;	iuy = tpk->pycen + 0.49999;
	if (EOK != ixy2ev(iux, iuy, ldv)) return;
	ldv[0] += 1.;
	for (delpa = iux = 0; iux < 3; iux++) delpa += ldv[iux] * ldv[iux];
	tios = log(tios);
	spp->spr_lsq[0] += 1.;
	spp->spr_lsq[1] += delpa;
	spp->spr_lsq[2] += delpa * delpa;
	spp->spr_lsq[3] += tios;
	spp->spr_lsq[4] += delpa * tios;
	if (delpa > spp->spr_maxs) spp->spr_maxs = delpa;
}

int mkcentroid(PEAK *tpk, int *regionp, SPOTRUN *spp)
{ /*	This computes the centroid of the spot with completed
	(or nearly-completed) spot `tpk'. It returns 0 if the spot
	has intensity >= `spp->spr_minint' and is well-behaved in other senses;
	it returns various codes otherwise. */
	int		bz2, bgnd_th, dnext, dox, doxe, dox0, doxz0, doy, doy0;
	int		doye, doyz0, doz, doze, doz0, dx, dy, dz;
	int		itz, lbgnd, loffx, loffy, loffz, lowpix;
	int		lx, lx2, lxy, lxyz, ly, ly2, lz, lz2, lz2l, lz2r;
	int		maxinshoe, nbright, nover0, noverl, npts, nrover;
	int		orx, ory, orz, outpix, thiso, voffx, voffy, voffz;
	int		xcen, ycen, zcen, *shv;
	unsigned	bx4;
	double		alfa, alfa1, censum, curri, dr2, dyz2, dz2, fbgnd;
	double		fmi, fmmax, fmnext, fmsum, mbg, medbg, minbr, minlint;
	double		mlo2, outom, ti, toz, trz, tsig, tx, ty, tz, zlint;
	float		locax, locay, *fmp, *fmp0, *mpx, *sumi, *thisi;
	MAIL		*lhd;
	PEAK		*lpk;
	TROID		*ttr, *ltr;
	REGIONPROFILE	*rgb, *rgp;
	static float	*sumi0;

	
		*regionp = getregion((short)(tpk->pxcen), (short)(tpk->pycen),
		/* tpk->pphicen, */ spp->spr_bdr);
	ttr = &(tpk->ptroi);	spp->spr_medbg = 0;	spp->spr_etype = 0;
	if ((spp->spr_fbgnd = fbgnd = tpk->pbgnd) < 0.)
		return cenerr(P_XYDIMPLED, tpk, spp);
	if (fbgnd > 1.e5) return cenerr(P_TOOWEAK, tpk, spp);
	/* refuse to write reflections out if the Lp is bad or large */
	if (tpk->pflags & P_NOLORENTZ) return cenerr(P_NOLORENTZ, tpk, spp);
	if (tpk->pflags & P_BIGLORENTZ) return cenerr(P_BIGLORENTZ, tpk, spp);
	xcen = tpk->pxcen + 0.49999;	ycen = tpk->pycen + 0.49999;
	spp->spr_spind = tpk->pphicen;
	if (EOK != eftable(5, spp))
	  {
		spp->spr_etype = 13;	spp->spr_badi = ttr->frame0;
		return cenerr(P_RANGEEDGE, tpk, spp);
	  }
	zcen = spp->spr_ifrm;
	lz2l = zcen - ttr->obkmin;	lz = ttr->nobk;
	lz2r = lz - 1 - lz2l;	lz2 = (lz2l + lz2r) / 2;
	lbgnd = fbgnd + 0.49999;
	rgb = spp->spr_rgb;	lx = rgb->m3_bx;	ly = rgb->m3_by;
	lx2 = lx / 2;	ly2 = ly / 2;
	lxy = lx * ly;	lz = 1 + lz2l + lz2r;
	lxyz = lxy * lz;
	/* if needed, reserve memory for 4 floating-point working vectors */
	bx4 = CELLS_NEEDED;
	if (NULL == spp->spr_thisi0)
	  {
		if (NULL == (AVOID *)(spp->spr_thisi0 =
		 (float *)callop((unsigned)bx4, sizeof (float),
		  "spp->spr_thisi0 in mkcentroid"))) return EALLOC;
		sumi0 = spp->spr_thisi0 + rgb->m3_bz * lxy;
	  }
	/* in any case, zero 'em out here */
	zeromem((AVOID *)spp->spr_thisi0, bx4 * sizeof (float));
	if (spp->spr_flags & P_VERBOSE2) show_shoe(spp->spr_fpl, tpk,
		" before recomputing background", spp);
	/* now we improve our estimate of the background value. We do this
	 by looking at all the pixels outside the centermost few points that
	 are less than thrice the current background estimate.  We determine
	 the mean of those values and use it. */
	bgnd_th = (lbgnd < 21845 ? 3 * lbgnd : 65534);
	bz2 = rgb->m3_bz / 2;	shv = spp->spr_shoe;
	if (bz2 > lz2l) shv += (bz2 - lz2l) * lxy;
	for (thisi = spp->spr_thisi0, mbg = npts = 0, dz = -lz2l;
	 dz <= lz2r; dz++)
	   {
		dz2 = spp->spr_phitoxscale * dz * dz;
		for (dy = -ly2; dy <= ly2; dy++)
		   {
			dyz2 = dz2 + dy * dy;
			for (dx = -lx2; dx <= lx2; dx++, shv++)
			   {
				dr2 = dyz2 + dx * dx;
				if ((dr2 > 121.) && (*shv < bgnd_th))
				  {
					*thisi++ = *shv; npts++; mbg += *shv;
				  }
			   }
		   }
	   }
	/* use the larger of the mean and the median */
	if (npts > 10)
	  {
		fbgnd = mbg / (double)npts;
		medbg = mymedian(spp->spr_thisi0, npts);
	  }
	else medbg = fbgnd;
	spp->spr_etype = 1;	spp->spr_fbgnd = fbgnd;	spp->spr_medbg = medbg;
	if ((fbgnd < 0.) || (medbg < 0.)) return cenerr(P_XYDIMPLED, tpk, spp);
	if ((fbgnd > 1.e5) || (medbg > 1.e5))
		return cenerr(P_TOOWEAK, tpk, spp);
	if (spp->spr_flags & P_VERBOSE2) (void)cenerr(0, tpk, spp);
	if (medbg > fbgnd) fbgnd = medbg;
	tpk->pbgnd = fbgnd;
	lbgnd = tpk->pbgnd + 0.49999;
	/* we check BEFORE removing overlaps to see if the spot is
	 too flat in omega or has an unexpectedly large number of pixels
	 either outside the active region or substantially below background */
	doz0 = zcen - lz2l;	doy0 = ycen - ly2;	dox0 = xcen - lx2;
	fmp0 = &(ttr->obparts[0]);
	ti = outom = npts = lowpix = outpix = 0; fmp = fmp0;
	minlint = -sqrt(fbgnd * spp->spr_aduperph);
	lhd = spp->spr_heads;
	if (israxis(lhd->format)) minlint *= 1.71;
	else if (ismarip(lhd->format)) minlint *= 1.414;
	mlo2 = minlint * 0.5;
	shv = spp->spr_shoe;	if (bz2 > lz2l) shv += lxy * (bz2 - lz2l);
	for (dz = -lz2l, doz = doz0; dz <= lz2r; dz++, fmp++, doz++)
	   {
		*fmp = fmsum = 0.;	doxz0 = dox0;	doyz0 = doy0;
		if (spp->spr_flags & P_FRAMEOFFSETS)
		  {
			locax = doxz0;	locay = doyz0;
			adjust_to_fm(&locax, &locay, zcen + dz, 0);
			doxz0 = locax + 0.49999; doyz0 = locay + 0.49999;
		  }
		for (fmsum = 0., dy = -ly2, doy = doyz0; dy <= ly2; dy++,doy++)
		   for (dx = -lx2, dox = doxz0; dx <= lx2; dx++, shv++, dox++)
		      {
			curri = *shv;	curri -= fbgnd;
			if (curri < minlint) lowpix++;
			if (curri < mlo2) curri = mlo2;
			fmsum += curri;
			if ((dox < 2) || (dox >= spp->spr_detw - 2) ||
			  (doy < 2) || (doy >= spp->spr_deth - 2)) outpix++;
			else if ((spp->spr_flags & P_USEBORDER) &&
			 (NULL == activmask(dox, doy, 041))) outpix++;
			npts++;
		      }
		*fmp = fmsum;	ti += *fmp;
		if (dz < -1 || dz > 1) outom += *fmp;
	   }
	/* n.b. we used to require that ti be > 0 at this point: not now */
	spp->spr_locint = ti;	spp->spr_outom = outom;
	spp->spr_npts = npts; spp->spr_outpx = outpix; spp->spr_lowpx = lowpix;
	spp->spr_etype = 2;
	if (npts < 10) return cenerr(P_BADPHISHAPE, tpk, spp);
	if ((ti > DFUZZ) && (outom / ti > 1. - spp->spr_mincenter))
	  {
		spp->spr_badi = zcen;
		return cenerr(P_BADPHISHAPE, tpk, spp);
	  }
	if ((((double)outpix) / (double)npts) > 0.3)
		return cenerr(P_XYEDGE, tpk, spp);
	if ((((double)lowpix) / (double)npts) > spp->spr_maxbelow)
		return cenerr(P_BELOWBG, tpk, spp);
	/* now go through the shoebox and get rid of the portions of the
	 profile that actually belong to other reflections.  The overlapped
	 spots are dealt with according to
		C(xyz) = B(xyz) + SUM (Iixyz)
				   i
	 where the estimates of Iixyz are taken either from an existing
	 model profile or from a Gaussian sphere model.
	 For now the Gaussian spheres are taken to have radius 6.  We may
	 go to adjustable spheres or even (gasp!) ellipsoids at a later date.
	 We need the `spp->spr_thisi0' array zeroed first. */
	zeromem((AVOID *)(spp->spr_thisi0), lxy * rgb->m3_bz * sizeof (float));
	zeromem((AVOID *)sumi0, lxy * rgb->m3_bz * sizeof (float));
	rgp = rgb + NREGIONS + (*regionp);
	for (lpk = spp->spr_pkbase, nrover = nover0 = 0;
	 lpk <= spp->spr_pklast; lpk++)
	   if ((lpk->pinten > 1.) &&
		 (radialdist(lpk, tpk, spp) < 500.)) /*strong&nearby*/
	     {
		thiso = 0;	ltr = &(lpk->ptroi);
		doz0 = -lz2l + (ttr->frame0 - ltr->frame0);
		doy0 = -ly2 + (tpk->pycen - lpk->pycen);
		dox0 = -lx2 + (tpk->pxcen - lpk->pxcen);
		doze = doz0 + lz; doye = doy0 + ly; doxe = dox0 + lx;
		for (doz = doz0, orz = 0; doz < doze; doz++, orz++)
		 {
		   voffz = (-lz2l <= doz) && (doz <= lz2r) ?
				lxy * (doz + lz2l) : -1;
		   loffz = lxy * orz;
		   for (doy = doy0, ory = 0; doy < doye; doy++, ory++)
		    {
		      voffy = (voffz >= 0) && (-ly2 <= doy) && (doy <= ly2) ?
				voffz + lx * (doy + ly2) : -1;
		      loffy = loffz + ory * lx;
		      for (dox = dox0, orx = 0; dox < doxe; dox++, orx++)
		      {
			voffx = (voffy >= 0) && (-lx2 <= dox) && (dox <= lx2) ?
				voffy + dox + lx2 : -1;
			loffx = loffy + orx;
			if ((0 > loffx) || (loffx > lxyz))
			  {
				fprintf(stderr, "Tilt! loffx = %d\n", loffx);
				abort();
			  }
			if ((voffx >= 0) &&
			 (*(mpx = rgp->m3_xyop + voffx) > FUZZ))
			  {
				curri = lpk->pinten * (*mpx);
				*(sumi0 + loffx) += curri;
				if (lpk == tpk)
					*(spp->spr_thisi0 + loffx) = curri;
				else if (curri > FUZZ) thiso++;
			  }
		      } /* close of loop over dox */
		    } /* close of loop over doy */
		 } /* close of loop over doz */
		if (thiso)
		  {
			nrover++;	nover0 += thiso;
		  }
	   } /* close of if "peak is relevant?" clause and loop over lpk */
	spp->spr_peakp = lpk;	/* so we know where we are externally */
	/* having determined the relative intensities at each point,
	 we correct the shoebox values accordingly. We remember the peak
	 value at this stage. */
	shv = spp->spr_shoe;	if (bz2 > lz2l) shv += (bz2 - lz2l) * lxy;
	for (dz = -lz2l, sumi = sumi0, thisi = spp->spr_thisi0, ti = 0,
	 fmp = fmp0, maxinshoe = noverl = 0; dz <= lz2r; dz++, fmp++)
	   {
		for (dy = -ly2, *fmp = 0.; dy <= ly2; dy++)
		 for (dx = -lx2; dx <= lx2; dx++, thisi++, sumi++, shv++)
		    {
		        if ((*shv > lbgnd) && (*sumi > FUZZ))
			  {
				alfa = flimits(*thisi / *sumi, 0., 1.);
				alfa1 = 1. - alfa; if (alfa1 > DFUZZ) noverl++;
				*shv = alfa * ((double)(*shv)) + alfa1 * fbgnd;
				if (*shv > maxinshoe) maxinshoe = *shv;
			  }
			curri = ((double)(*shv)) - fbgnd;
			if (curri <  mlo2) curri = mlo2;
			*fmp += curri;
		    }
		ti += *fmp;
	   }
	spp->spr_locint = ti;	spp->spr_outom = outom;	spp->spr_etype = 3;
	spp->spr_npts = npts; spp->spr_outpx = outpix; spp->spr_lowpx = lowpix;
	if (ti < DFUZZ) return cenerr(P_XYEDGE, tpk, spp);
	for (fmp = fmp0, dz = -lz2l; dz <= lz2; dz++, fmp++) *fmp /= ti;
	if (spp->spr_flags & P_VERBOSE2)
	  {
		show_shoe(spp->spr_fpl, tpk, " after removing overlaps", spp);
		xeflush(spp->spr_fpl);
	  }
	/* now we have a good shoebox.  Get the centroid and a one-dimensional
	profile. We exclude pixels more than 15 radial units from the center
	within a given profile element, and we take take any point whose
	background-subtracted intensity is < -sqrt(fbgnd) and set it to
	that floor value. Also, if we know in advance that an entire frame
	of the omega profile is distinctly below zero, we exclude the entire
	frame from the centroid-and-intensity calculation. Finally, if the
	user has invoked `spots' with the "-b<val>" (for breadth) flag, we
	demand that at least <val> different pixels contain at least
	10% of the peak value. This is to prevent hyper-sharp spots (probably
	coming from hot dots on the detector face) from being counted. */
	minbr = ((double)maxinshoe) * 0.10;	nbright = 0;
	ti = tx = ty = tz = 0.;
	ttr->obkmin = zcen - lz2l;	fmp = fmp0;
	shv = spp->spr_shoe;	if (bz2 > lz2l) shv += lxy * (bz2 - lz2l);
	for (dz = -lz2l, tpk->pncells = 0; dz <= lz2r; dz++, fmp++)
	   {
	    if (*fmp > -0.05)
	      {
		for (fmi = 0., dy = -ly2; dy <= ly2; dy++)
		 {
		   dyz2 = dy * dy;
		   for (dx = -lx2; dx <= lx2; dx++, shv++)
		     {
			if (dx * dx + dyz2 < 225.)
			  {
				zlint = ((double)(*shv)) - fbgnd;
				if (zlint < minlint) zlint = minlint;
				if (zlint > minbr) nbright++;
				fmi += zlint;	ti += zlint;
				tx += ((double)dx) * zlint;
				ty += ((double)dy) * zlint;
				tz += ((double)dz) * zlint;
				if (zlint > minlint) tpk->pncells++;
			  } /* close of "if (dx " clause */
		     } /* close of loop over dx */
		 } /* close of loop over dy */
		*fmp = fmi;
	       } /* close of "if (*fmp >" clause */
	     else {
			*fmp = 0.;	shv += lxy;
		  }
	   } /* close of loop over dz */
	spp->spr_locint = ti;
	if (ti < spp->spr_minint) return cenerr(P_TOOWEAK, tpk, spp);
	if (nbright < spp->spr_mnbright) return cenerr(P_TOOSHARP, tpk, spp);
	/* examine the one-D profile. If it's offcenter or  flat, reject it.
	 Also, reject it if it's severely dimpled in omega, i.e. if
	 (a) the peak frame and the second-highest frame are not adjacent
	 AND (b) the intensity in the second-highest frame > 0.3 * the peak.
	 Meanwhile, note the first and last positive elements of it */
	doz0 = doze = -99;
	/* get the sum and the max frame */
	for (fmmax = censum = doz = 0, dz = -lz2l, fmp = fmp0;
	 dz <= lz2r; dz++, fmp++)
	   {
		*fmp /= ti;
		if (*fmp > DFUZZ)
		  {
			doze = dz;	if (doz0 == -99) doz0 = dz;
		  }
		if (*fmp > fmmax)
		  {
			fmmax = *fmp; doz = dz;
		  }
		if (-1 <= dz && dz <= 1) censum += *fmp;
	   }
	if (doz0 < -97) return cenerr(P_TOOWEAK, tpk, spp);
	if (doz < -1 || doz > 1)
	  {
		spp->spr_etype = 4;	spp->spr_badi = doz;
		return cenerr(P_RANGEEDGE, tpk, spp);
	  }
	if (censum < spp->spr_mincenter)
	  {
		spp->spr_etype = 5;	spp->spr_badf = censum;
		return cenerr(P_BADPHISHAPE, tpk, spp);
	  }
	for (fmnext = 0., dz = -lz2l, fmp = fmp0, dnext = -lz;
	 dz <= lz2r; dz++, fmp++) if ((dz != doz) && (*fmp > fmnext))
	   {
		fmnext = *fmp;	dnext = dz;
	   }
	if ((dnext > -lz) && (abs(doz - dnext) > 1) &&
	 (fmmax > 0.1) && (fmnext > 0.5 * fmmax))
	  {
		spp->spr_etype = 8;	spp->spr_badi = dnext;
		spp->spr_badf = fmnext;
		return cenerr(P_BADPHISHAPE, tpk, spp);
	  }
	 /* if the first positive element isn't element 0,
	 slide the profile down */
	doz0 += lz2l;	doze += lz2l;
	if (doz0 > 0)
	  {
		for (doz = doz0, fmp = fmp0; doz <= doze; doz++, fmp++)
			*fmp = *(fmp + doz0);
		for ( ; doz < lz; doz++, fmp++) *fmp = 0.;
		ttr->obkmin += doz0;
	  }
	ttr->nobk = doze + 1 - doz0;
	if ((ttr->nobk < 0) || (ttr->obkmin > spp->spr_zma) ||
	 (ttr->obkmin < spp->spr_zmi))
	  {
		spp->spr_etype = 8;	spp->spr_badi = ttr->nobk;
		return cenerr(P_BADPHISHAPE, tpk, spp);
	  }
	tx = ((double)xcen) + tx / ti;
	if (fabs(tx - tpk->pxcen) > 1.5 * spp->spr_maxmove)
	  {
		spp->spr_etype = 9;	spp->spr_badf = tx;
		return cenerr(P_MOVETOOFAR | P_XYDIMPLED, tpk, spp);
	  }
	ttr->xto = tpk->pxcen = tx;
	ty = ((double)ycen) + ty / ti;
	if (fabs(ty - tpk->pycen) > 1.5 * spp->spr_maxmove)
	  {
		spp->spr_etype = 10;	spp->spr_badf = ty;
		return cenerr(P_MOVETOOFAR | P_XYDIMPLED, tpk, spp);
	  }
	ttr->yto = tpk->pycen = ty;
	spp->spr_rfrm = toz = ((double)zcen) + tz / ti;
	spp->spr_ifrm = itz = toz + 0.49999;
	if (EOK != eftable(011, spp))
	  {
		spp->spr_etype = 12;	spp->spr_badi = itz;
		return cenerr(P_RANGEEDGE, tpk, spp);
	  }
	else	trz = spp->spr_spind;
	if (fabs(trz - tpk->pphicen) > spp->spr_maxmove)
	  {
		spp->spr_etype = 11;	spp->spr_badf = tz;
		return cenerr(P_MOVETOOFAR | P_XYDIMPLED, tpk, spp);
	  }
	tpk->pphicen = trz;	ttr->frame0 = spp->spr_efrm;
	/* if necessary, adjust the centroid so it reflects what's
	 happening in the peak frame of this reflection. Note that
	 we do that only with the ttr values, not the tpk values. */
	if (spp->spr_flags & P_FRAMEOFFSETS)
	  {
		locax = ttr->xto;	locay = ttr->yto;
		adjust_to_fm(&locax, &locay, ((int)(toz + 0.49999)), 0);
		ttr->xto = locax;	ttr->yto = locay;
	  }
	ttr->totinten = tpk->pinten = ti;
	/* the sigma is taken as having a 20% error in the background estimate
	(a bit higher than in `integrate') and 1% "machine error */
	tsig = ti + 1.2 * fbgnd * spp->spr_aduperph * ((double)tpk->pncells) +
		(0.01 * ti) * (0.01 * ti);
	tpk->psigma = (tsig < 1. ? 1. : sqrt(tsig));
	/* do one last renormalization of the profile */
	for (ti = dz = 0; dz < ttr->nobk; dz++) ti += ttr->obparts[dz];	
	if (ti > DFUZZ)
	  {
		for (dz = 0; dz < ttr->nobk; dz++) ttr->obparts[dz] /= ti;
	  }
	/* see if the peak is on the edge of the range and,
	 if appropriate, reject it */
	for ( ; dz < TROIDWIDTH; dz++) ttr->obparts[dz] = 0.;
	if (spp->spr_flags & P_CONTIG)
	  {
		for (doz0 = 0, ti = ttr->obparts[0], dz = 1;
		 dz < ttr->nobk; dz++) if (ttr->obparts[dz] > ti)
		   {
			ti = ttr->obparts[dz];	doz0 = dz;
		   }
		if ((doz0 <= 0) && (ttr->obkmin <= spp->spr_zmi) &&
		 (spp->spr_flags & P_CONTIGLEFT))
			return cenerr(P_RANGEEDGE, tpk, spp);
		doze = ttr->obkmin + ttr->nobk - 1;
		if ((doz0 >= (ttr->nobk - 1)) && (doze >= spp->spr_zma) &&
		 (spp->spr_flags & P_CONTIGRIGHT))
			return cenerr(P_RANGEEDGE, tpk, spp);
	  }
	spp->spr_etype = 0;
	/* If this is the final pass, then accumulate least-squares sums for
	 i000, B, and D2 calcs, and write
	 the preservable portion of the reflection out. */
	if (spp->spr_passno > 2)
	  {
		accum_lsq(tpk, spp);
		if (verror(0 < xgwrite((void *)ttr, XG_TROID, spp->spr_fpo),
		 "write a TROID reflection to the output")) return EWRITERR;
	  }
	return EOK;
}

static void kmodel(PEAK *tpeak, int region, SPOTRUN *spp)
{ /*	This adds a reflection to the diffraction-vector model profiles */

	register int	ix, iy, iz;
	int		ika, ikb, ikg, irx, iry;
	int		jka, jkb, jkg, lx, lx2, lxyz;
	int		ly, ly2, lz, lz2, mx, mxy, mxyz, my, mz, xce, yce;
	double		ah, al, bh, bl, deka, dekb, dekg, gh, gl, kalpha;
	double		kbeta, kgamma, lgamco, mika, mikb, mikg, rka;
	double		rkb, rkg, rxu, ryu, rz, rzu, thism, iko[7], kaf[6];
	float		*bmp, *kmig, *kmjg, *kprop;
	REGIONPROFILE	*rgb, *rgp;

	/* get the approximate diffraction vector coords of the spot */
	if (!ikabc(tpeak->pxcen, tpeak->pycen, iko)) return;
	spp->spr_spind = tpeak->pphicen;
	if (EOK != eftable(5, spp)) return; /* phi probably bad */
	lgamco = iko[6];	rgb = spp->spr_rgb;	rgp = rgb + region;
	lx = rgp->m3_bx;	ly = rgp->m3_by;	lz = rgp->m3_bz;
	lx2 = lx / 2;		ly2 = ly / 2;		lz2 = lz / 2;
	lxyz = lx * ly * lz;
	mx = rgp->m3_qx;	my = rgp->m3_qy;	mz = rgp->m3_qz;
	mxy = mx * my;	mxyz = mxy * mz;
	mika = rgp->m3_lim[0];	deka = rgp->m3_lim[2];
	mikb = rgp->m3_lim[3];	dekb = rgp->m3_lim[5];
	mikg = rgp->m3_lim[6];	dekg = rgp->m3_lim[8];
	rxu = mx - 1;	ryu = my - 1;	rzu = mz - 1;
	rz = (spp->spr_rfrm - (double)spp->spr_ifrm) - lz2;
	xce = tpeak->pxcen + 0.49999;	yce = tpeak->pycen + 0.49999;
	kprop = spp->spr_rprop + rgp->m3_nmodels * 2 * lxyz;
	for (bmp = spp->spr_thisi0 + 3 * lxyz, iz = 0; iz < lz; iz++, rz += 1.)
	   {
		kgamma = sin(rz * spp->spr_framrad) * lgamco;
		rkg = flimits((kgamma - mikg) / dekg, 0., rzu);
		ikg = rkg;
		gl = 1. + (double)ikg - rkg;	gh = 1. - gl;
		jkg = (ikg < mz - 1 ? ikg + 1 : ikg);
		/* the omega value we use takes into account the fact that
		 the frame itself covers a range of omega values */
		kmig = kprop + mxyz * region + ikg * mxy;
		kmjg = kprop + mxyz * region + jkg * mxy;
		for (iy = 0, iry = yce - ly2; iy < ly; iy++, iry++)
		   {
		     for (ix = 0, irx = xce - lx2; ix < lx; ix++, bmp++,irx++)
		      {
			thism = *bmp;
			if (0 == reg2hab(irx, iry, kaf))
			  {
				kalpha = kaf[0];	kbeta = kaf[1];
				rka = flimits((kalpha - mika) /deka,
						0., rxu); ika = rka;
				al = 1. + (double)ika - rka; ah = 1. - al;
				jka = (ika < 12 ? ika + 1 : ika);
				rkb = flimits((kbeta - mikb) / dekb,
						0., ryu); ikb = rkb;
				bl = 1. + (double)ikb - rkb; bh = 1. - bl;
				jkb = (ikb < 12 ? ikb + 1 : ikb);
				ikb *= mx;	jkb *= mx;
				if (thism > DFUZZ)
				  {
					*(kmig+ ikb+ ika) += thism *al *bl *gl;
					*(kmig+ ikb+ jka) += thism *ah *bl *gl;
					*(kmig+ jkb+ ika) += thism *al *bh *gl;
					*(kmig+ jkb+ jka) += thism *ah *bh *gl;
					*(kmjg+ ikb+ ika) += thism *al *bl *gh;
					*(kmjg+ ikb+ jka) += thism *ah *bl *gh;
					*(kmjg+ jkb+ ika) += thism *al *bh *gh;
					*(kmjg+ jkb+ jka) += thism *ah *bh *gh;
				  } /* close "if (thism > FUZZ)" clause */
			  } /* close "if (0 == reg2hab" clause */
		      } /* close "for (ix" loop */
		   } /* close "for (iy" loop */
	   } /* close of "for iz" loop */
}

int add2model(PEAK *tpk, int region, SPOTRUN *spp)
{ /*	This converts the shoebox to a floating-point,
	background-corrected, centered object.  The centering is done
	by a combination of numbskull reindexing and careful (maximum-entropy)
	redistribution of data. The routine then adds it to the model profile
	associated with the appropriate region of the detector face. The routine
	returns 0 if successful, various errors otherwise. */
	register int	ix, iy, iz;
	int		bxyz, bz2, lx, lx2, lxy, lxyz, ly, ly2;
	int		lz, lz2, lz2l, lz2r, maxfm, uz, zcen;
	double		fbgnd, sumax, minlint, thissig, totest, totxwt;
	double		totsqxwt, tsi, zlint, delv[3], idv[3], sumz[11];
	float		*bv, *tq, *bvp, *rprf;
	int		*shv;
	TROID		*ttr;
	REGIONPROFILE	*rgb, *rgp;

	fbgnd = tpk->pbgnd;	minlint = -sqrt(fbgnd * spp->spr_aduperph);
	rgb = spp->spr_rgb;
	lx = rgb->m3_bx;	ly = rgb->m3_by; lx2 = lx / 2;	ly2 = ly / 2;
	lxy = lx * ly;	bxyz = lxy * rgb->m3_bz;
	zcen = spp->spr_ifrm;	ttr = &(tpk->ptroi);
	lz2l = zcen - ttr->obkmin;	lz = ttr->nobk;
	lz2r = lz - 1 - lz2l;	lz2 = (lz2l + lz2r) / 2; lxyz = lxy * lz;
	bv = spp->spr_thisi0 + bxyz + bxyz;	tq = bv + bxyz;
	bz2 = rgb->m3_bz / 2;	shv = spp->spr_shoe;
	if (bz2 > lz2l) shv += (bz2 - lz2l) * lxy;
	for (uz = 0; uz < lz; uz++) sumz[uz] = 0.;
	for (uz = 0, iz = -lz2l, bvp = bv; iz <= lz2r; iz++, uz++)
	 for (iy = -ly2; iy <= ly2; iy++)
	  for (ix = -lx2; ix <= lx2; ix++, shv++, bvp++)
	     {
		zlint = ((double)(*shv)) - fbgnd;
		if (zlint < minlint) zlint = minlint;	*bvp = zlint;
		sumz[uz] += zlint;
	     }
	for (maxfm = 0, sumax = sumz[0], uz = 1; uz < lz; uz++)
	 if (sumz[uz] > sumax)
	   {
		maxfm = uz;	sumax = sumz[uz];
	   }
	idv[0] = idv[1] = idv[2] = 0.;
	if ((tsi = gen_centroid(bv, NULL, lx, ly, lz, idv, delv)) < 1.)
	  {
		spp->spr_badf = tsi;	spp->spr_etype = 14;
		return cenerr(P_TOOWEAK, tpk, spp);
	  }
	/* really big changes are disallowed. */
	if (fabs(delv[0]) > 2.)
	  {
		spp->spr_badf = delv[0];	spp->spr_etype = 15;
		return cenerr(P_MOVETOOFAR, tpk, spp);
	  }
	if (fabs(delv[1]) > 2.)
	  {
		spp->spr_badf = delv[1];	spp->spr_etype = 16;
		return cenerr(P_MOVETOOFAR, tpk, spp);
	  }
	if (fabs(delv[2]) > 2.)
	  {
		spp->spr_badf = delv[2];	spp->spr_etype = 17;
		return cenerr(P_MOVETOOFAR, tpk, spp);
	  }
	for (iy = 0; iy < 4; iy++)
	   {
		recen_vec(bv, delv, tq, lx, ly, lz);
		idv[0] = idv[1] = idv[2] = 0.;
		if ((tsi = gen_centroid(tq, NULL, lx, ly, lz, idv, delv)) <1.)
		  {
			spp->spr_badf = tsi;	spp->spr_etype = 18;
			return P_TOOWEAK;
		  }
		if ((fabs(delv[0]) < 0.2) && (fabs(delv[1]) < 0.2) &&
		 (fabs(delv[2]) < 0.2)) break;
		cpymem((void *)bv, (void *)tq, bxyz * sizeof (float));
	   }
	if (iy >= 4) /* couldn't center it in 4 tries */
	  {
		if (fabs(delv[0]) > 0.2)
		  {
			spp->spr_badf = delv[0];	spp->spr_etype = 19;
			return cenerr(P_MOVETOOFAR, tpk, spp);
		  }
		if (fabs(delv[1]) > 0.2)
		  {
			spp->spr_badf = delv[1];	spp->spr_etype = 20;
			return cenerr(P_MOVETOOFAR, tpk, spp);
		  }
		/* often this failure means we're at the start or end in phi */
		if ((ttr->obkmin == spp->spr_zmi) && (0 >= maxfm))
		  {
			spp->spr_badi = spp->spr_zmi;	spp->spr_etype = 12;
			return cenerr(P_RANGEEDGE, tpk, spp);
		  }
		else if (((ttr->obkmin + lz - 1) == spp->spr_zma) &&
		 (lz - 1) <= maxfm)
		  {
			spp->spr_badi = spp->spr_zma;	spp->spr_etype = 13;
			return cenerr(P_RANGEEDGE, tpk, spp);
		  }
		spp->spr_badf = delv[2];	spp->spr_etype = 21;
		return cenerr(P_BADPHISHAPE, tpk, spp);
	  }
	/* just for displays, we re-create the int version */
	for (iz = 0, shv = spp->spr_shoe, bvp = tq; iz < bxyz;
	 iz++, shv++, bvp++) *shv = *bvp + fbgnd + 0.499999;
	if (spp->spr_flags & P_VERBOSE2)
	  {
		show_shoe(spp->spr_fpl, tpk, " after recentering", spp);
		xeflush(spp->spr_fpl);
	  }
/* now add the current 3-D profile to the appropriate model. */
	totest = tpk->pinten;	thissig = tpk->psigma;
	totxwt = totest / (thissig * thissig);
	totsqxwt = totest * totxwt;
	if (totsqxwt > 2500.)
	  {
		totxwt *= 2500. / totsqxwt;	totsqxwt = 2500.;
	  }
	rgp = rgb + region;	rprf = rgp->m3_xyop;
	(rgp->m3_nobs)++;		rgp->m3_isqsum += totsqxwt;
	if (rgp->m3_isqsum > 1.e8) fprintf(stderr,
	 "Warning: in region %2d m3_isqsum = %11.4e\n", region, rgp->m3_isqsum);
	for (bvp = tq, rprf = rgp->m3_xyop, iz = 0;
	 iz < lxyz; iz++, rprf++,bvp++) *rprf += *bvp * totxwt;
	/* also, transform this 3-D profile into diffraction-vector
	coordinates and add it to the appropriate model */
	if (spp->spr_plated > FUZZ) kmodel(tpk, region, spp);
	return EOK;
}

/* end of spotshoes.c */
