/* ucautils.c -- utilities for handling USPLINE (.uca) files

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	29 May 2007: mergeumain handles 2-segment as well as 4-segment systems;
	 comments and code updated accordingly; validerr used for error msgs
	19 Apr 2007: float->double conversions; reorganize use of "{"
	12 May 2006: define cci2ucamain here, not in ../include/calibdefs.h;
		merge cci2uca.c, mergeuca.c, moduca.c as ucautils.c
	 3 Jan 2006: mergeumain: set_logical can return >= 1
	 2 Jan 2006: replace record_detxy call with setting v_maxx, v_maxy
	 8 Dec 2005: remove memzap of VSPLINE stuff -- globalkill does it
	20 Nov 2005: inline function declarations & function prototypes
	28 Oct 2005: eliminate static globals and do some housekeeping
	17 Jun 2005: cci2ucamain(), moducamain(), mergeumain():
		putspline() takes VSPLINE argument
	11 May 2005: check return status from xgjournal
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	28 Aug 2004: routines used only in moduca declared as static; cosmetics
	 2 Aug 2004: add new arg to find_dimension() and record_detxy()
	26 Jan 2004: convert most of the static globals into local variables
	21 Jan 2004: license message inserted and initial comment reordered
	 6 Jan 2004: CALIBRUN structure elements in function calls
		& static globals replace extern globals
	20 Nov 2003: replace global variables with elements of CALIBRUN struct
	19 Sep 2003: carefully manage the outwid, outht values in moduca...
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	13 Aug 2003: correct declarations of return types in mergeuca
	 5 Jan 2003: reorganized source-file header
	 1 Nov 2002: trivial reformatting in mergeuca
	20 Sep 2002: CF_* flags moved to ../include/xgen.h
	25 Mar 2002: removed MSI attachments
	25 Nov 1997: initial version of cci2uca.c, based on `mergeuca';
		initial version of moduca.c
	15 Nov 1997: initial version of mergeuca
   Purpose: various specific functions for manipulating USPLINES, specifically:
	cci2uca, which reads a CALCENTROIDS and writes a USPLINE,
	mergeuca, which merges together a set of USPLINEs, and
	moduca, which enables editing of a USPLINE
   Cross references: ../prerefine/cal*.c
*************************************************************************/
#define		NEED_ERRNO	1
#include	"calibdefs.h"
#define	MOD_START	1
#define	MOD_END		2
#define	MAXSPOTS	14000	/* max # of recordable spots in image*/

 typedef struct {
		int		cc_maxux, cc_maxwx, cc_maxuy, cc_maxwy;
		int		cc_minux, cc_minuy, cc_keepinregion;
		} CCIWORK;

 typedef struct {
		short	muc_ixc, muc_iyc, muc_mx, muc_my, muc_nx, muc_ny;
		double	muc_dx, muc_dy, muc_ox, muc_oy;
		} MODIFYUCA;

extern int cci2ucamain(int argc, char *argv[]);
extern int moducamain(int argc, char *argv[]);
extern int mergeumain(int argc, char *argv[]);

/****************************
	Routines used for two or more of these functionalities
****************************/

static void modxy0(int outw, int outh, VSPLINE *lsp)
{ /*	This goes through the array and calculates the mean signed offset
	for all the data and reapplies it */
	int	ux, uy, ux0, uy0, uxe, uye, dex, dey, wid0, hit0;
	int	badcm, badpx, quadr;
	float	xin, yin, xcm, ycm, xo, yo;
	double	rct, sumx, sumy;

	ux0 = lsp->v_x0px;	uy0 = lsp->v_y0px;
	uxe = ux0 + ((double)lsp->v_ncol) * lsp->v_xrpx;
	uye = uy0 + ((double)lsp->v_nrow) * lsp->v_yrpx;
	wid0 = lsp->v_detw;	hit0 = lsp->v_deth;
	if (outw != 0) lsp->v_detw = outw;
	if (outh != 0) lsp->v_deth = outh;
	if ((outw != 0) || (outh != 0))
	  {
		lsp->v_maxx = lsp->v_detw + 10;
		lsp->v_maxy = lsp->v_deth + 10;
	  }
	dex = 5;	dey = 5;
	badcm = badpx = 0;
	for (rct = sumx = sumy = 0, uy = uy0; uy <= uye; uy += dey)
	   {
		for (ux = ux0; ux <= uxe; ux += dex)
		   {
			xin = ux;	yin = uy;
			if (1 <= pix2cm(xin, yin, &xcm, &ycm, 0))
			  {
				quadr = detregion(xin, yin, lsp);	
				if (1 <= cm2pixq(xcm, ycm, quadr, &xo, &yo, 0))
				  {
					sumx += xo - xin; sumy += yo - yin;
					rct += 1.;
				  }
				else badcm++;
			  }
			else badpx++;
		   }
	   }
	if (rct < 1.)
	  {
		printf("no legal points: no recentering done\n"); return;
	  }
	sumx /= rct;	sumy /= rct;
	fprintf(stdout,
 "failures in pix2cm = %7d; failures in cm2pix = %7d; good pixels=%7d\n",
		badpx, badcm, (int)rct);
	fprintf(stdout, "Recentering in (X,Y) by (%8.3f,%8.3f) pixels\n",
		sumx, sumy);
	lsp->v_x0px += sumx;	lsp->v_y0px += sumy;
	if ((lsp->v_deth != hit0) || (lsp->v_detw != wid0))
	  {
		lsp->v_detw = wid0;	lsp->v_maxx = wid0 + 10;
		lsp->v_deth = hit0;	lsp->v_maxy = hit0 + 10;
	  }
}

static CALPOINT *closest_cpt(VSPLINE *lspl, double xpx,
	double ypx, int *uxp, int *uyp, int keepin)
{ /*	This returns in pointers *uxp, *uyp the indices of the point
	in the array that is closest to the pixel value xpx, ypx */
	int		regi, vx, vy;
	double		dex, dey, rmi, thr;
	CALPOINT	*capt;

	rmi = 1.e12;	*uxp = -1;	*uyp = -1;
	regi = detregion(xpx, ypx, lspl);
	for (capt = lspl->v_cpt, vy = 0; vy < lspl->v_nrow; vy++)
	 for (vx = 0; vx < lspl->v_ncol; vx++, capt++)
	   if ((!keepin) || (regi == detregion(
	    capt->c_ptocx, capt->c_ptocy, lspl)))
	     {
		dex = xpx - capt->c_ptocx;	dey = ypx - capt->c_ptocy;
		thr = dex * dex + dey * dey;
		if (thr < rmi)
		  {
			rmi = thr; *uxp = vx; *uyp = vy;
		  }
	     }
	/* at this point we've found the point closest to our reference
	 position. We require, however, that the point be above and to
	 the left of our reference position, if possible. */
	capt = lspl->v_cpt + (*uyp) * lspl->v_ncol + (*uxp);
	if (((*uxp) > 0) && (xpx < capt->c_ptocx))
	  {
		(*uxp)--;	capt--;
	  }
	if (((*uyp) > 0) && (ypx < capt->c_ptocy)) (*uyp)--;
	/* don't allow the right or bottom to be used */
	if ((*uxp) >= lspl->v_ncol -1 ) (*uxp) = lspl->v_ncol - 2;
	if ((*uyp) >= lspl->v_nrow -1 ) (*uyp) = lspl->v_nrow - 2;
	/* We make some further adjustments in a multi-element detector
	 to require that the point stay within the quadrant. */
	if (keepin)
	  {
		if (xpx < lspl->v_detw / 2)
		  {
			if ((*uxp) >= lspl->v_midc) *uxp = lspl->v_midc - 1;
		  }
		else if ((*uxp) < lspl->v_midc) *uxp = lspl->v_midc;
		if (ypx < lspl->v_deth / 2)
		  {
			if ((*uyp) >= lspl->v_midr) *uyp = lspl->v_midr - 1;
		  }
		else if ((*uyp) < lspl->v_midr) *uyp = lspl->v_midr;
	  	if ((*uxp) == lspl->v_midc - 1) (*uxp)--;
		if ((*uyp) == lspl->v_midr - 1) (*uyp)--;
	  }
	capt = lspl->v_cpt + (*uyp) * lspl->v_ncol + (*uxp); return capt;
}

static void interp_ctop(VSPLINE *lspl, CALPOINT *cpt, int keepin)
{ /*	This interpolates in the neighborhood of a point.  If `keepin' is
	nonzero, it requires that the points used for the interpolation be
	contained within the detector quadrant that our point is in */
	int		jx, jy, offset, ux0, uy0;
	double		cx0, cx1, cy0, cy1, pwx, pwy, xp, yp;
	CALPOINT	*cbp0, *cbp1, *cbp2, *cbp3;

	offset = cpt - lspl->v_cpt;
	jy = offset / lspl->v_ncol; jx = offset % lspl->v_ncol;
	xp = lspl->v_x0px + ((double)jx) * lspl->v_xrpx;
	yp = lspl->v_y0px + ((double)jy) * lspl->v_yrpx;
	cbp0 = closest_cpt(lspl, xp, yp, &ux0, &uy0, keepin); cbp1 = cbp0 + 1;
	cbp2 = cbp0 + lspl->v_ncol;	cbp3 = cbp2 + 1;
	/* dubeity check */
	if (fabs(cbp0->c_ptocx - cbp2->c_ptocx) > 4.) fprintf(stderr,
	 "Warning! x(0,2) = %9.2f,%9.2f for ix,iy,X,Y=%3d,%3d,%8.2f,%8.2f\n",
		 cbp0->c_ptocx, cbp2->c_ptocx, ux0, uy0, xp, yp);
	if (fabs(cbp1->c_ptocx - (cbp3)->c_ptocx) > 4.) fprintf(stderr,
	 "Warning! x(1,3) = %9.2f,%9.2f for ix,iy,X,Y=%3d,%3d,%8.2f,%8.2f\n",
		 cbp1->c_ptocx, cbp3->c_ptocx, ux0, uy0, xp, yp);
	if (fabs(cbp0->c_ptocy - cbp1->c_ptocy) > 4.) fprintf(stderr,
	 "Warning! y(0,1) = %9.2f,%9.2f for ix,iy,X,Y=%3d,%3d,%8.2f,%8.2f\n",
		 cbp0->c_ptocy, cbp1->c_ptocy, ux0, uy0, xp, yp);
	if (fabs(cbp2->c_ptocy - cbp3->c_ptocy) > 4.) fprintf(stderr,
	 "Warning! y(2,3) = %9.2f,%9.2f for ix,iy,X,Y=%3d,%3d,%8.2f,%8.2f\n",
		 cbp2->c_ptocy, cbp3->c_ptocy, ux0, uy0, xp, yp);
	pwx = (2. * xp - (cbp0->c_ptocx + cbp2->c_ptocx)) /
		(cbp1->c_ptocx + cbp3->c_ptocx -
		cbp0->c_ptocx - cbp2->c_ptocx);
	pwy = (2. * yp - (cbp0->c_ptocy + cbp1->c_ptocy)) /
		(cbp2->c_ptocy + cbp3->c_ptocy - cbp0->c_ptocy -
		cbp1->c_ptocy); 
	cx0 = lspl->v_x0cm + ((double)ux0) * lspl->v_xrcm;
	cy0 = lspl->v_y0cm + ((double)uy0) * lspl->v_yrcm;
	cx1 = cx0 + lspl->v_xrcm;	cy1 = cy0 + lspl->v_yrcm;
	cpt->c_ctopx = (1. - pwx) * cx0 + pwx * cx1;
	cpt->c_ctopy = (1. - pwy) * cy0 + pwy * cy1;
}

/*****************************
	Next few routines here are associated with reading a CALCENTROIDS file
	and writing a USPLINE.
	We use a simple centering algorithm found originally in `mergeuca'.
*****************************/
static void summ_cci2u(FILE *fp, CALIBRUN *cap, CCIWORK *ccp)
{ /*	This summarizes the run to the stream fp */
	static char	*fnames[] = {"calcentroids", "uspline", "xlog", NULL};
	static char	ftypes[] = {3, 2, 2, 0};

	ftypes[2] = (cap->car_fpl == stdout ? 0 : 2);
	header(fp, PROCESSING, fnames, ftypes);
	fprintf(fp, "Beam center is at [%6.1f,%6.1f]\n",
		cap->car_beamx, cap->car_beamy);
	fprintf(fp,
	 "Raster between fiducial points is %7.3f cm in x, %7.3f cm in y\n",
		cap->car_usedrx, cap->car_usedry);
	if (ccp->cc_keepinregion) fprintf(fp,
	 "interpolation will be kept within a detector region\n");
}

static int cciargs(int ac, char *av[], CALIBRUN **cadp, CCIWORK **ccdp)
{ /*	Decodes the flags and command-line args, if any */
	int		locht, locwid, ret;
	char		*car, **dcp, **flags;
	CALIBRUN	*cap;
	CCIWORK		*ccp;
	static VSPLINE	*lspl;

	if (EHELP == xgjournal(ac, av)) return EOK; /* journal the effort */
	if (NULL == (*cadp = (CALIBRUN *)callop(1, sizeof (CALIBRUN),
	 "cadp in cci2uca"))) return EALLOC;
	if (NULL == (*ccdp = (CCIWORK *)callop(1, sizeof (CCIWORK),
	 "ccdp in cci2uca"))) return EALLOC;
	/* open logfile and put a summary of the files there and on stdout */
	ccp = *ccdp;	cap = *cadp;
	if (NULL == (cap->car_fpl = propen("xlog"))) cap->car_fpl = stdout;
	flags = flagproc(ac, av, "CKRXY"); /*flag arguments*/
	cap->car_usedrx = cap->car_usedry = DEFRASTER;
	ccp->cc_keepinregion = 0;
	cap->car_flags = CF_ASCIICEN;
	cap->car_ntroids = ccp->cc_maxwx = ccp->cc_maxwy = 0;
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	   {
		car = (*dcp) + 1;
		if (**dcp == 'C') ccp->cc_maxwx = atoi(car);
		else if (**dcp == 'K') ccp->cc_keepinregion = 1;
		else if (**dcp == 'R') ccp->cc_maxwy = atoi(car);
		else if (**dcp == 'X') cap->car_usedrx = atof(car);
		else if (**dcp == 'Y') cap->car_usedry = atof(car);
	   }
	/* measure the detector dimension by looking at one of
	 the images in the current list of images */
	if (EOK != (ret = find_dimension(&locwid, &locht, 1,NULL))) return ret;
	/* skip past the non-numeric arguments */
	for (ac--, av++; (ac) && !allnumeric(*av); ac--) av++;
	/* if the user has specified a center, record it.
	 Otherwise we assume it's the midrange of the detector */
	if (ac > 0)
	  {
		cap->car_beamx = atof(av[0]);
		if (ac > 1) cap->car_beamy = atof(av[1]);
		else cap->car_beamy = 0.5 * (double)locht;
	  }
	else {
		cap->car_beamx = 0.5 * ((double)locwid);
		cap->car_beamy = 0.5 * (double)locht;
	     }
	/* reserve memory for the VSPLINE structure we're going to build */
	if (NULL == lspl)
	  {
		if (qerror((AVOID *)(lspl = (VSPLINE *)callop(1,
		 sizeof (VSPLINE), "lspl in cci2uargs")),
			"reserve memory for spline structure")) return EALLOC;
	  }
	else /* get rid of old memory for lspl->v_cpt if there is any */
		memzap((AVOID **)&(lspl->v_cpt), "lspl->v_cpt in cci2args");
	lspl->v_detw = locwid;	lspl->v_deth = locht;
	cap->car_spl = lspl;
	if (ccp->cc_keepinregion) lspl->v_nreg = 4;	else lspl->v_nreg = 0;
	if (NULL == (AVOID *)(cap->car_ctr0 = (CTROID *)callop(MAXSPOTS,
	 sizeof (CTROID), "ctr0 in cciargs")))
	  {
		if (stdout != cap->car_fpl) (void)fclose(cap->car_fpl);
		return EALLOC;
	  }
	cap->car_ctend = cap->car_ctr0 + MAXSPOTS;
	summ_cci2u(stdout, cap, ccp);
	if (cap->car_fpl != stdout) summ_cci2u(cap->car_fpl, cap, ccp);
	return EOK;
}

static void extrema_calcen(CALIBRUN *cap, CCIWORK *ccp)
{ /*	This finds the minimum and maximum indices and positions of the
	CALCENTROIDS file */
	CTROID	*lct;

	ccp->cc_minuy = ccp->cc_minux = 10000;
	ccp->cc_maxuy = ccp->cc_maxux = -ccp->cc_minuy;
	for (lct = cap->car_ctr0; lct < cap->car_ctend; lct++)
	   {
		if (lct->ux < ccp->cc_minux) ccp->cc_minux = lct->ux;
		if (lct->uy < ccp->cc_minuy) ccp->cc_minuy = lct->uy;
		if (lct->ux > ccp->cc_maxux) ccp->cc_maxux = lct->ux;
		if (lct->uy > ccp->cc_maxuy) ccp->cc_maxuy = lct->uy;
	   }
	cap->car_ctr = lct;
}

static void extrapux(CTROID *cout, CTROID *pre, CTROID *ant)
{ /*	This extrapolates for points in the x direction */
	double	mult;

	if (pre->ux == ant->ux) return;	/* won't work! */
	mult = ((double)(cout->ux - pre->ux)) / ((double)(pre->ux - ant->ux));
	cout->cxcen = pre->cxcen + mult * (pre->cxcen - ant->cxcen);
	cout->cycen = pre->cycen + mult * (pre->cycen - ant->cycen);
	printf( "[%6.1f,%6.1f,%3d,%3d] <- ",
		cout->cxcen, cout->cycen, cout->ux, cout->uy);
	printf("[%4.0f,%4.0f,%3d,%3d],[%4.0f,%4.0f,%3d,%3d]\n",
		pre->cxcen, pre->cycen, pre->ux, pre->uy,
		ant->cxcen, ant->cycen, ant->ux, ant->uy);
}

static void extrapuy(CTROID *cout, CTROID *pre, CTROID *ant)
{ /*	This extrapolates for points in the y direction */
	double	mult;

	if (pre->uy == ant->uy) return;	/* won't work! */
	mult = ((double)(cout->uy - pre->uy)) / ((double)(pre->uy - ant->uy));
	cout->cxcen = pre->cxcen + mult * (pre->cxcen - ant->cxcen);
	cout->cycen = pre->cycen + mult * (pre->cycen - ant->cycen);
	printf( "[%6.1f,%6.1f,%3d,%3d] <- ",
		cout->cxcen, cout->cycen, cout->ux, cout->uy);
	printf("[%4.0f,%4.0f,%3d,%3d],[%4.0f,%4.0f,%3d,%3d]\n",
		pre->cxcen, pre->cycen, pre->ux, pre->uy,
		ant->cxcen, ant->cycen, ant->ux, ant->uy);
}

static void strap(CALIBRUN *cap, CCIWORK *ccp)
{ /*	This extends the indexing toward smaller or larger ux and uy*/
	int	ide, lux, luy, maxi, midux, miduy, mini;
	CTROID	*anted, *anteu, *cte, *lct, *prevd, *prevu;

	miduy = (ccp->cc_minuy + ccp->cc_maxuy) / 2;
	for (ide = 1, luy = miduy, lct = cap->car_ctr0; luy >= 0; ide++)
	 {
	  if (luy <= ccp->cc_maxuy)
	     {
		mini = 10000;	maxi = -mini;
		prevd = anted = prevu = anteu = NULL;
		for (lct = cap->car_ctr0; lct < cap->car_ctend; lct++)
		   {
			if (lct->uy == luy)
			  {
				if (lct->ux < mini)
				  {
					prevd = lct; mini = lct->ux;
				  }
				if (lct->ux > maxi) 
				  {
					prevu = lct; maxi = lct->ux;
				  }
			  }
		   }
		if ((NULL != prevd) && (prevd->ux > 0))
		  {
			mini = 10000;	cte = cap->car_ctend;
			for (lct = cap->car_ctr0; lct < cap->car_ctend; lct++)
			   {
				if ((lct != prevd) && (lct->uy == luy) &&
				 (lct->ux < mini))
				  {
					anted = lct;	mini = lct->ux;
				  }
			   }
			if (NULL != anted)
			  {
				for (lux = prevd->ux - 1; lux >= 0; lux--)
				   {
					cte->ux = lux; cte->uy = luy;
					extrapux(cte, prevd, anted);
					cte++;	cap->car_ctend = cte;
				   }
			  }
		  }
		if ((NULL != prevu) && (prevu->ux < ccp->cc_maxwx))
		  {
			maxi = -10000;
			for (lct = cap->car_ctr0; lct < cap->car_ctend; lct++)
			   {
				if ((lct != prevu) && (lct->uy == luy) &&
				 (lct->ux > maxi))
				  {
					anteu = lct;	maxi = lct->ux;
				  }
			   }
			if (NULL != anteu)
			  {
				cte = cap->car_ctend;
				for (lux = prevu->ux + 1;
				 lux <= ccp->cc_maxwx; lux++)
				   {
					cte->ux = lux; cte->uy = luy;
					extrapux(cte, prevu, anteu);
					cte++;	cap->car_ctend = cte;
				   }
			  }
		  }
	     }
	  if (luy <= miduy) luy += ide;	else luy -= ide;
	 }
	cap->car_ctr = lct;
	extrema_calcen(cap, ccp);
	midux = (ccp->cc_minux + ccp->cc_maxux) / 2;
	for (ide = 1, lux = midux; lux >= 0; ide++)
	 {
	   if (lux <= ccp->cc_maxux)
	     {
		mini = 10000;	maxi = -mini;
		prevd = anted = prevu = anteu = NULL;
		for (lct = cap->car_ctr0; lct < cap->car_ctend; lct++)
		   {
			if (lct->ux == lux)
			  {
				if (lct->uy < mini)
				  {
					prevd = lct; mini = lct->uy;
				  }
				if (lct->uy > maxi) 
				  {
					prevu = lct; maxi = lct->uy;
				  }
			  }
		   }
		if ((NULL != prevd) && (prevd->uy > 0))
		  {
			mini = 10000;
			for (lct = cap->car_ctr0; lct < cap->car_ctend; lct++)
			   {
				if ((lct != prevd) && (lct->ux == lux) &&
				 (lct->uy < mini))
				  {
					anted = lct;	mini = lct->uy;
				  }
			   }
			if (NULL != anted)
			  {
				cte = cap->car_ctend;
				for (luy = prevd->uy - 1; luy >= 0; luy--)
				   {
					cte->uy = luy; cte->ux = lux;
					extrapuy(cte, prevd, anted);
					cte++;	cap->car_ctend = cte;
				   }
			  }
		  }
		if ((NULL != prevu) && (prevu->uy < ccp->cc_maxwy))
		  {
			maxi = -10000;
			for (lct = cap->car_ctr0; lct < cap->car_ctend; lct++)
			   {
				if ((lct != prevu) && (lct->ux == lux) &&
				 (lct->uy > maxi))
				  {
					anteu = lct;	maxi = lct->uy;
				  }
			   }
			if (NULL != anteu)
			  {
				cte = cap->car_ctend;
				for (luy = prevu->uy + 1;
				 luy <= ccp->cc_maxwy; luy++)
				   {
					cte->ux = lux; cte->uy = luy;
					extrapuy(cte, prevu, anteu);
					cte++;	cap->car_ctend = cte;
				   }
			  }
		  }
	     }
	  if (lux <= midux) lux += ide;	else lux -= ide;
	 }
	cap->car_ctr = lct;	extrema_calcen(cap, ccp);
}

static int tr2vspl(int nind, CALIBRUN *cap, CCIWORK *ccp)
{ /*	This sets up the VSPLINE structure and populates it out to
	the point of getting the c_ptocx and c_ptocy values */
	int	npts;
	VSPLINE	*lspl;

	/* number of columns and rows is measured from input list */
	lspl = cap->car_spl;
	lspl->v_ncol = ccp->cc_maxux + 1 - ccp->cc_minux;
	lspl->v_nrow = ccp->cc_maxuy + 1 - ccp->cc_minuy;
	lspl->v_ntrue = nind;	lspl->v_midc = lspl->v_ncol / 2;
	lspl->v_midr = lspl->v_nrow / 2;
	lspl->v_axc = cap->car_beamx;	lspl->v_ayc = cap->car_beamy;
	/* for now we'll assume the recomputed values of the center
	 are equal to the input values */
	lspl->v_rxc = lspl->v_axc;	lspl->v_ryc = lspl->v_ayc;
	/* spacing between reference holes has been read in.
	 By convention we make the spacing in y negative */
	lspl->v_xrcm = cap->car_usedrx;	lspl->v_yrcm = -(cap->car_usedry);
	/* by convention, the reference positions of the spots
	 is calculated from (width+10)/(#spaces) and (height+10)/(#spaces).*/
	lspl->v_xrpx = (((double)(lspl->v_detw)) + 10.) /
		((double)(lspl->v_ncol - 1));
	lspl->v_yrpx = (((double)(lspl->v_deth)) + 10.) /
		((double)(lspl->v_nrow - 1));
	/* `left edge' in centimeters and pixels is set as shown */
	lspl->v_x0cm = -(lspl->v_xrcm * (double)lspl->v_midc);
	lspl->v_y0cm = -(lspl->v_yrcm * (double)lspl->v_midr);
	lspl->v_x0px = -5;	lspl->v_y0px = -5;
	npts = lspl->v_ncol * lspl->v_nrow;
	if (NULL == (lspl->v_cpt = (CALPOINT *)callop((size_t)npts,
	 sizeof (CALPOINT), "lspl->v_cpt in tr2vspl"))) return EALLOC;
	return EOK;
}

static int populatespl(CALIBRUN *cap, CCIWORK *ccp)
{ /*	This puts actual values into the CALPOINTs */
	int		edge, lux, luy, yedge;
	VSPLINE		*lspl;
	CALPOINT	*calp;
	CTROID		*lct;

	lspl = cap->car_spl;	lct = cap->car_ctr0;
	for (calp = lspl->v_cpt, luy = 0; luy < lspl->v_nrow; luy++)
	   {
		yedge = (1 <= luy) && (luy < lspl->v_nrow - 1);
		for (lux = 0; lux < lspl->v_ncol; lux++, calp++)
		   {
			edge = yedge & ((1 <= lux) && (lux < lspl->v_ncol - 1));
			for (lct = cap->car_ctr0; lct < cap->car_ctend; lct++)
			   {
				if ((lct->ux == lux) && (lct->uy == luy))
				  {
					calp->c_ptocx = lct->cxcen;
					calp->c_ptocy = lct->cycen;
					calp->c_set = calp->c_cset = 1;
					calp->c_ift = edge;
				  }
			   }
		   }
	   }
	cap->car_ctr = lct;	luy = lspl->v_ncol * lspl->v_nrow;
	for (calp = lspl->v_cpt; luy--; calp++)
		interp_ctop(lspl, calp, ccp->cc_keepinregion);
	return EOK;	/* any failure modes? I don't think so */
}

static int endcci(int resu, CALIBRUN **cadp, CCIWORK **ccdp)
{ /*	This blots out memory for two structures and returns its argument */
	memzap((AVOID **)cadp, "cap in cci2uca");
	memzap((AVOID **)ccdp, "ccp in cci2uca");
	return resu;
}

/*ARGSUSED*/
int cci2ucamain(int argc, char *argv[])
{ /* Mainline for the conversion of CALCENTROIDS to USPLINE */
	int		nind, res;
	static CALIBRUN	*cap;
	static CCIWORK	*ccp;

	if (EOK != (res = cciargs(argc, argv, &cap, &ccp)))
		return endcci(res, &cap, &ccp);
	if (EOK != (res = ext_troids(&nind, cap)))
	  {
		fprintf(stderr, "Error opening CALCENTROIDS file\n");
		return endcci(res, &cap, &ccp);
	  }
	extrema_calcen(cap, ccp); /* find extreme values among CALCENTROIDS*/
	if (ccp->cc_maxux > ccp->cc_maxwx) ccp->cc_maxwx = ccp->cc_maxux;
	if (ccp->cc_maxuy > ccp->cc_maxwy) ccp->cc_maxwy = ccp->cc_maxuy;
	strap(cap, ccp);	/* extrapolate the indexing to wider ranges */
	/* create the actual spline */
	if (EOK != (res = tr2vspl(nind, cap, ccp)))
		return endcci(res, &cap, &ccp);
	/* populate the individual elements thereof */
	if (EOK != (res = populatespl(cap, ccp)))
		return endcci(res, &cap, &ccp);
	/* recenter the entire spline if requested */
	if (EOK != (res = recenter(cap->car_spl, cap->car_beamx,
	 cap->car_beamy))) return endcci(res, &cap, &ccp);
	if (NULL == (cap->car_spl = gspline(9, &res))) /*write to output file*/
		return endcci(res, &cap, &ccp);
	if ((NULL != cap->car_fpl) && (stdout != cap->car_fpl))
		(void)fclose(cap->car_fpl);
	cap->car_fpl = NULL;
	memzap((AVOID **)&(cap->car_ctr0), "ctr0 in cci2uca");
	globalkill();	return endcci(res, &cap, &ccp);
}

/* end of cci2uca functionality */

/*********************************************
	Routines below handle merging 2 or 4 USPLINE files to
	form a single USPLINE.
	Later versions will handle more than four.
	They should be ordered as follows:
Two-segment case:
    -----------
   |  0  |  1  | Thus if (mk,nk) are the number of columns and rows,
    -----------  respectively, of input file k (k = 0, 1), then
		 we require n0 = n1.
Four-segment case:
    -----------
   |  0  |  2  | Thus if (mk,nk) are the number of columns and rows,
   |     |     | respectively, of input file k (k = 0, 1, 2, 3), then
   |  1  |  3  | we require n0 = n2, n1 = n3, m0 = m1, m2 = m3,
    -----------
		  and m0 + m2 = m1 + m3 == total number of columns
		  and n0 + n1 = n2 + n3 == total number of rows.
  Any other circumstance will result in an error.
*************************************************************************/
static int cpyspline(VSPLINE *osp, VSPLINE *isp)
{ /*	this copies the VSPLINE in *isp to *osp.
	The comment that used to be here about rescaling the splines
	was incorrect and has been removed! */
	int		npts;
	CALPOINT	*ipt, *opt;

	if (NULL != osp->v_cpt) memzap((AVOID **)&osp->v_cpt,
	 "osp->v_cpt in cpyspline");
	cpymem((void *)osp, (void *)isp, sizeof (VSPLINE));
	osp->v_cpt = NULL;
	npts = isp->v_ncol * isp->v_nrow;
	if (NULL == (osp->v_cpt = (CALPOINT *)callop(npts,
	 sizeof (VSPLINE), "osp->v_cpt in cpyspline"))) return EALLOC;
	ipt = isp->v_cpt;	opt = osp->v_cpt;
	for ( ; npts--; ipt++, opt++)
		cpymem((void *)opt, (void *)ipt, sizeof (CALPOINT));
	return EOK;
}

static int validerr(FILE *fp, char *rva, char *rvb, int vla, int vlb, int rtyp)
{
 /* This outputs a particular message to fp
	describing an error in valid_regions */

	static char	*rtc[2] = {"columns", "rows"} ;

	fprintf(fp, "Region %s has %4d %s; region %s has %4d\n",
		rva, vla, rtc[rtyp], rvb, vlb);
	return EBADVALUE;
}

static int valid_regions(int nregions, VSPLINE lspl[])
{ /*	This checks that the conditions listed in the opening comment
	are actually met by these data. */
	int	n0, n1, n2, n3, nc, nr, m0, m1, m2, m3;

	n0 = lspl[0].v_nrow;	m0 = lspl[0].v_ncol;
	n1 = lspl[1].v_nrow;	m1 = lspl[1].v_ncol;
	if (nregions < 4)
	  {
		if (n0 != n1) return validerr(stderr, "0", "1", n0, n1, 2);
		return EOK;
	  }
	n2 = lspl[2].v_nrow;	m2 = lspl[2].v_ncol;
	n3 = lspl[3].v_nrow;	m3 = lspl[3].v_ncol;
	if (n0 != n2) return validerr(stderr, "0", "2", n0, n2, 1);
	if (n1 != n3) return validerr(stderr, "1", "3", n1, n3, 1);
	if (m0 != m1) return validerr(stderr, "0", "1", m0, m1, 0);
	if (m2 != m3) return validerr(stderr, "2", "3", m2, m3, 0);
	if ((nr = n0 + n1) != n2 + n3)
		return validerr(stderr, "0+1", "2+3", nr, n2 + n3, 1);
	if ((nc = m0 + m2) != m1 + m3)
		return validerr(stderr, "0+2", "1+3", nc, m1 + m3, 0);
	return EOK;
}

static int merge_regions(int nregions, VSPLINE *osp, VSPLINE lspl[])
{ /*	This actually shoves the four splines together.
	It returns EOK unless memory-allocation errors arise. */
	int		col, mi, npts, row;
	CALPOINT	*caia, *caib, *cao;

	mi = (nregions > 2) ? 2 : 1;
	osp->v_detw = lspl[0].v_detw + lspl[mi].v_detw;
	osp->v_ncol = lspl[0].v_ncol + lspl[mi].v_ncol;
	if (nregions <= 2)
	  {
		osp->v_deth = lspl[0].v_deth;
		osp->v_nrow = lspl[0].v_nrow;
		osp->v_ntrue = lspl[0].v_ntrue + lspl[1].v_ntrue;
		osp->v_axc = 0.5 * (lspl[0].v_axc +
			lspl[1].v_axc + lspl[0].v_detw);
		osp->v_ayc = 0.5 * (lspl[0].v_ayc + lspl[1].v_ayc);
		osp->v_rxc = 0.5 * (lspl[0].v_rxc +
			lspl[1].v_rxc + lspl[0].v_detw);
		osp->v_ryc = 0.5 * (lspl[0].v_ryc + lspl[1].v_ryc);
		osp->v_xrcm = 0.5 * (lspl[0].v_xrcm + lspl[1].v_xrcm);
		osp->v_yrcm = 0.5 * (lspl[0].v_yrcm + lspl[1].v_yrcm);
	  }
	else {
		osp->v_deth = lspl[0].v_deth + lspl[1].v_deth;
		osp->v_nrow = lspl[0].v_nrow + lspl[1].v_nrow;
		osp->v_ntrue = lspl[0].v_ntrue + lspl[1].v_ntrue +
			lspl[2].v_ntrue + lspl[3].v_ntrue;
		/* center values taken as averages of internal centers */
		osp->v_axc = (lspl[0].v_axc + lspl[1].v_axc +
		 lspl[2].v_axc + lspl[3].v_axc + 2 * lspl[0].v_detw) / 4;
		osp->v_ayc = (lspl[0].v_ayc + lspl[1].v_ayc +
		 lspl[2].v_ayc + lspl[3].v_ayc + 2 * lspl[0].v_deth) / 4;
		osp->v_rxc = (lspl[0].v_rxc + lspl[1].v_rxc +
		 lspl[2].v_rxc + lspl[3].v_rxc + 2 * lspl[0].v_deth) / 4;
		osp->v_ryc = (lspl[0].v_ryc + lspl[1].v_ryc +
		 lspl[2].v_ryc + lspl[3].v_ryc + 2 * lspl[0].v_deth) / 4;
		/* raster values in cm are taken as averages, but
		 we hope they're identical anyway */
		osp->v_xrcm = (lspl[0].v_xrcm + lspl[1].v_xrcm +
		 lspl[2].v_xrcm + lspl[3].v_xrcm) / 4;
		osp->v_yrcm = (lspl[0].v_yrcm + lspl[1].v_yrcm +
		 lspl[2].v_yrcm + lspl[3].v_yrcm) / 4;
		/* start values in pixels are taken from the upper left */
		/* DISABLED: SEE BELOW
		osp->v_x0px = lspl[0].v_x0px;	osp->v_y0px = lspl[0].v_y0px;*/
	  }
	/* other values are calculated from the above */
	osp->v_midc = osp->v_ncol / 2; osp->v_midr = osp->v_nrow / 2;
	osp->v_x0cm = -(osp->v_xrcm * (double)osp->v_midc);
	osp->v_y0cm = -(osp->v_yrcm * (double)osp->v_midr);
	osp->v_xrpx = (((double)(osp->v_detw)) + 10.) /
			((double)(osp->v_ncol - 1));
	osp->v_yrpx = (((double)(osp->v_deth)) + 10.) /
			((double)(osp->v_nrow - 1));
	osp->v_x0px = ((double)osp->v_detw) / 2. -
		osp->v_xrpx * (double)osp->v_midc;
	osp->v_y0px = ((double)osp->v_deth) / 2. -
		osp->v_yrpx * (double)osp->v_midr;
	osp->v_maxx = osp->v_detw + 10;
	osp->v_maxy = osp->v_deth + 10;
	/* total number of points in spline is product of rows*columns */
	npts = osp->v_nrow * osp->v_ncol;
	printf("Detector is [%4d,%4d] pixels, [%3d,%3d] fiducial points\n",
	 osp->v_detw, osp->v_deth, osp->v_ncol, osp->v_nrow);
	printf(
	 "cm array begins at [%8.4f,%8.4f]; px array at [%8.2f,%8.2f]\n",
		osp->v_x0cm, osp->v_y0cm, osp->v_x0px, osp->v_y0px);
	printf(
	 "cm array intervals:[%8.4f,%8.4f]; px array:   [%8.2f,%8.2f]\n",
		osp->v_xrcm, osp->v_yrcm, osp->v_xrpx, osp->v_yrpx);
	if (NULL == (osp->v_cpt = (CALPOINT *)callop(npts,
	 sizeof (CALPOINT), "osp->v_cpt in merge_regions"))) return EALLOC;
	/* read in regions zero and mi */
	cao = osp->v_cpt;
	caia = lspl[0].v_cpt;	caib = lspl[mi].v_cpt;
	for (row = 0; row < lspl[0].v_nrow; row++)
	   {
		for (col = 0; col < lspl[0].v_ncol; col++, caia++, cao++)
			cpymem((void *)cao, (void *)caia, sizeof (CALPOINT));
		for (col = 0; col < lspl[mi].v_ncol; col++, caib++, cao++)
		   {
			cpymem((void *)cao, (void *)caib, sizeof (CALPOINT));
			cao->c_ptocx += lspl[0].v_detw;
		   }
	   }
	/* release the memory associated with the lists in regions 0 and mi */
	memzap((AVOID **)&(lspl[0].v_cpt), "lspl[0].v_cpt in merge_regions");
	memzap((AVOID **)&(lspl[mi].v_cpt), "lspl[mi].v_cpt in merge_regions");
	/* read in regions one and three if needed */
	if (nregions > 2)
	  {
		caia = lspl[1].v_cpt;	caib = lspl[3].v_cpt;
		for (row = 0; row < lspl[1].v_nrow; row++)
		   {
			for (col = 0; col < lspl[1].v_ncol;
			 col++, caia++, cao++)
			   {
				cpymem((void *)cao, (void *)caia,
					sizeof (CALPOINT));
				cao->c_ptocy += lspl[0].v_deth;
			   }
			for (col = 0; col < lspl[3].v_ncol;
			 col++, caib++, cao++)
			   {
				cpymem((void *)cao, (void *)caib,
					sizeof (CALPOINT));
				cao->c_ptocx += lspl[1].v_detw;
				cao->c_ptocy += lspl[2].v_deth;
			   }
		   }
		/* release memory associated with lists in regions 1 and 3 */
		memzap((AVOID **)&(lspl[1].v_cpt),
			"lspl[1].v_cpt in merge_regions");
		memzap((AVOID **)&(lspl[3].v_cpt),
			"lspl[3].v_cpt in merge_regions");
	  }
	/* explicitly get the AMASK */
	if (NULL == (activmask((int)(osp->v_detw), (int)(osp->v_deth), 5)))
		return (errno == ENOMEM) ? EALLOC : EREADERR;
	return EOK;
}

static void inspl_header(FILE *fp)
{ /* Provide a header for the spline outputs to the stream fp */
	fprintf(fp, "Input spline files:\n");
	fprintf(fp,
	 "# Columns # Rows Width Height # obs.centroids Name\n");
}

static void summar_spl(FILE *fp, VSPLINE *lspl, char *fnam)
{ /*	Summarize the output dimensions to the stream fp */
	fprintf(fp, " %8d %6d %5d %5d %15d %s\n",
	 lspl->v_ncol, lspl->v_nrow, lspl->v_detw, lspl->v_deth,
	 lspl->v_ntrue, fnam);
}

int mergeumain(int argc, char *argv[])
{ /* Mainline for merging 2 or 4 VSPLINEs for individual segments */
	char		*cp, *cpname, **lav, saveenv[256];
	int		lac, npt, nspli, sret;
	double		specx0px, specy0px;
	CALPOINT	*vp;
	static VSPLINE	*sgsp, lspl[4];

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	specx0px = specy0px = -1.e9;	cpname = xgprogname(NULL, 0);
	if (argc < 3) /* two or four filenames must be specified */
	  {
		printf("Usage: %s [x0 y0] file0 file1 [file2 file3]\n",
			cpname); return EOK;
	  }
	/* remember the original spline logical name */
	if (NULL != (cp = getenv("USPLINE"))) strcpy(saveenv, cp);
	else strcpy(saveenv, "new.uca");
	/* read in the 2 or 4 input splines and copy them to buffers */
	nspli = 0;
	inspl_header(stdout);
	for (lav = argv + 1, lac = 1; lac < argc; lac++, lav++)
	   {
		if (allnumeric(*lav))
		  {
			if (specx0px < -0.9e9) specx0px = atof(*lav);
			else specy0px = atof(*lav);
		  }
		else {
			if (1 > (sret = set_logical("USPLINE", *lav)))
			  {
				fprintf(stderr,
				 "Error setting logical name USPLINE as %s\n",
					*lav); return sret;
			  }
			if (NULL == (sgsp = gspline(5, &sret)))
			  {
				fprintf(stderr,
				 "Error %d in reading spline\n", sret);
				return sret;
			  }
			if (EOK != (sret = cpyspline(&(lspl[nspli]), sgsp)))
			  {
				fprintf(stderr,
				 "Error %d copying spline to buffer\n", sret);
				return sret;
			  }
			summar_spl(stdout, sgsp, *lav);
			nspli++;
		     }
	   }
	if (nspli < 2)
	  {
		fprintf(stderr, "%s: specified %d files, expected 2 or 4\n",
			cpname, nspli); return EBAD;
	  }
	/* make sure they're compatible in width and height */
	if (EOK != (sret = valid_regions(nspli, lspl)))
	  {
		fprintf(stderr, "Error %d comparing the regions\n", sret);
		return sret;
	  }
	sgsp->v_nreg = nspli;	/* by default here! */
	/* merge them together */
	if (EOK != (sret = merge_regions(nspli, sgsp, lspl)))
	  {
		fprintf(stderr,
		 "Error %d merging the data together\n", sret); return sret;
	  }
	for (vp = sgsp->v_cpt, npt = sgsp->v_ncol * sgsp->v_nrow; npt--; vp++)
		interp_ctop(sgsp, vp, 1);
	modxy0(0, 0, sgsp); /* compare splined values with inputs & recenter */
	printf("just after      modxy0: width,height = %4d,%4d\n",
		sgsp->v_detw, sgsp->v_deth);
	/* modify beam-center coordinates if specified */
	if ((specx0px > -100.) && (specy0px > -100.))
	  {
		if (EOK != (sret = recenter(sgsp, specx0px, specy0px)))
			return sret;
	  }
	printf("just after   recenter: width,height = %4d,%4d\n",
		sgsp->v_detw, sgsp->v_deth);
	/* restore the USPLINE logical name */
	if (1 > (sret = set_logical("USPLINE", saveenv)))
	  {
		fprintf(stderr, "Error %d setting output USPLINE name to %s\n",
			sret, saveenv); return sret;
	  }
	fprintf(stdout, "Output spline:\n");
	summar_spl(stdout, sgsp, saveenv);
	sgsp = gspline(9, &sret);	/* write out the spline */
	globalkill();	return sret;
}

/* end of mergeuca functionality */

/*******************************************************
	Routines below enable modifying a USPLINE based on text instructions.
	They modify a VSPLINE structure by adding & subtracting rows & columns
	from existing version according to instructions in a data file.
	They then modify the values in the rows and columns according to other
	instructions in the data file with environment variable name UCAMOD.
 Contents of a typical data file:
   15      15		# current indices of reference point (ixc,iyc)
   16      15		# desired indices of reference point (mx, my)
   32      32		# number of total columns and rows (nx, ny)
    0.      0.		# pixel offset in X and Y (dx, dy)
   -4.064   4.064	# cm offset in x and y (ox, oy)
	So the algorithm is as follows, assuming nc, nr are the
	current number of rows and columns in the file:
 while (mx < ixc) { (add a new column on the left); nc++; }
 while (mx > ixc) { (delete a column from the left); nc--; }
 while (nc < nx) { (add a new column on the right); nc++; }
 while (nc > nx) { (delete a column from the right); nc--; }
 while (my < iyc) { (add a new row on the top); nc++; }
 while (my > iyc) { (delete a row from the top); nc--; }
 while (nc < ny) { (add a new row on the bottom); nc++; }
 while (nc > ny) { (delete a row from the bottom); nc--; }
 for (all points in the surviving array)
    {	X(pix) += dx;	Y(pix) += dy;
	x(cm) += ox;	y(cm) += oy;
    }
 b-b-b-dat's all, folks!
*************************************************************************/
static int addcols(int delc, int whichmod, VSPLINE *lsp)
{ /*	This adds additional columns to the right of the last
	of the existing columns or to the left of the initial one. */
	int		colno, newpt, npts, rowno;
	static VSPLINE	lv;
	CALPOINT	*ante, *lb, *lpi, *lpo, *prev;

	if (delc <= 0) return EOK;	/* do nothing unless we must */
	/* define how many points there are in the output spline */
	npts = (lsp->v_ncol + delc) * lsp->v_nrow;
	/* reserve memory for the CALPOINT structures in the output spline */
	if (NULL == (lb = (CALPOINT *)callop(npts, sizeof (CALPOINT),
		"lb in addcols"))) return EALLOC;
	printf(
 "array previously had %d columns; adding %d additional columns on the %s\n",
		lsp->v_ncol, delc, (whichmod == MOD_START) ? "left" : "right");
	/* copy the general information about the old spline into the new one */
	cpymem((void *)&lv, (void *)lsp, sizeof (VSPLINE));
	/* increment the number of columns (that's what we're doing!) */
	lv.v_ncol += delc;	lv.v_midc = lv.v_ncol / 2;
	lv.v_x0cm = -(lsp->v_xrcm * (double)lv.v_midc);
	lv.v_xrpx = (((double)(lsp->v_detw)) + 10.) / ((double)(lv.v_ncol -1));
	lpi = lsp->v_cpt;
	if (whichmod == MOD_START) lpo = lb + delc;
	else lpo = lb;
	for (rowno = 0; rowno < lsp->v_nrow; rowno++)
	   { /* for each row, copy over the existing values */
		for (colno = 0; colno < lsp->v_ncol; colno++, lpi++, lpo++)
			cpymem((void *)lpo, (void *)lpi, sizeof (CALPOINT));
		/* if we're adding new points at the beginning of the row,
		 we have to rewind back to the first unknown value */
		if (whichmod == MOD_START) lpo -= (lsp->v_ncol) + 1;
		/* figure out what the new values should be.
		 This is currently a simple linear extrapolation */
		for (newpt = 0; newpt < delc; newpt++)
		   {
			if (whichmod == MOD_START)
			  {
				prev = lpo + 1;	ante = lpo + 2;
			  }
			else {
				prev = lpo - 1;	ante = lpo - 2;
			     }
			lpo->c_ptocx = 2 * prev->c_ptocx - ante->c_ptocx;
			lpo->c_ptocy = 2 * prev->c_ptocy - ante->c_ptocy;
			lpo->c_ctopx = 2 * prev->c_ctopx - ante->c_ctopx;
			lpo->c_ctopy = 2 * prev->c_ctopy - ante->c_ctopy;
			lpo->c_ift = 0;	/* it's extrapolated */
			lpo->c_qec = prev->c_qec; /* no other info avail */
			lpo->c_set = lpo->c_cset = 1;
			if (whichmod == MOD_START) lpo--;	else lpo++;
		   }
		/* set the pointer at the appropriate output point for
		 the next row */
		if (whichmod == MOD_START) lpo += 1 + lv.v_ncol + delc;
	   }
	/* release mem for old CALPOINTS*/
	memzap((AVOID **)&(lsp->v_cpt), "lsp->v_cpt in addcols");
	/* copy the working structure into the permanent one */
	cpymem((void *)lsp, (void *)&lv, sizeof (VSPLINE));
	lsp->v_cpt = lb;	/* set the CALPOINT structure appropriately */
	return EOK;
}

static int delcols(int delc, int whichmod, VSPLINE *lsp)
{ /*	This deletes columns on the right or left side of the array */
	int		colno, npts, rowno;
	CALPOINT	*lb, *lpi, *lpo;
	static VSPLINE	lv;

	if (delc <= 0) return EOK;	/* do nothing unless we must */
	/* make sure we will have at least one column left */
	if (delc >= lsp->v_ncol) return EBADVALUE;
	/* define how many points there are in the output spline */
	npts = (lsp->v_ncol - delc) * lsp->v_nrow;
	/* reserve memory for the CALPOINT structures in the output spline */
	if (NULL == (lb = (CALPOINT *)callop(npts, sizeof (CALPOINT),
		"lb in delcols"))) return EALLOC;
	printf(
 "array previously had %d columns; deleting %d columns from the %s\n",
		lsp->v_ncol, delc, (whichmod == MOD_START) ? "left" : "right");
	/* copy the general information about the old spline into the new one */
	cpymem((void *)&lv, (void *)lsp, sizeof (VSPLINE));
	/* decrement the number of columns (that's what we're doing!) */
	lv.v_ncol -= delc;	lv.v_midc = lv.v_ncol / 2;
	lv.v_x0cm = -(lsp->v_xrcm * (double)lv.v_midc);
	lv.v_xrpx = (((double)(lsp->v_detw)) + 10.) / ((double)(lv.v_ncol -1));
	if (whichmod == MOD_START) lpi = lsp->v_cpt + delc;
	else lpi = lsp->v_cpt;
	lpo = lb;
	for (rowno = 0; rowno < lsp->v_nrow; rowno++)
	   { /* for each row, copy over the existing values */
		for (colno = 0; colno < lv.v_ncol; colno++, lpi++, lpo++)
			cpymem((void *)lpo, (void *)lpi, sizeof (CALPOINT));
		/* increment the input pointer to skip over the ones
		 we aren't going to use. This works at either the beginning
		 or the end of the row. Isn't that nice! */
		lpi += delc;
	   }
	/* release mem for old CALPOINTS*/
	memzap((AVOID **)&(lsp->v_cpt), "lsp->v_cpt in delcols");
	/* copy the working structure into the permanent one */
	cpymem((void *)lsp, (void *)&lv, sizeof (VSPLINE));
	lsp->v_cpt = lb;	/* set the CALPOINT structure appropriately */
	return EOK;
}

static int addrows(int delr, int whichmod, VSPLINE *lsp)
{ /*	This adds additional rows on the bottom of the last
	of the existing rows or to the top of the initial one. */
	int		npts, rowno, colno, newpt, nc;
	static VSPLINE	lv;
	CALPOINT	*lb, *lpi, *lpo, *prev, *ante;

	if (delr <= 0) return EOK;	/* do nothing unless we must */
	nc = lsp->v_ncol;	/* useful shorthand */
	/* define how many points there are in the output spline */
	npts = (lsp->v_nrow + delr) * nc;
	/* reserve memory for the CALPOINT structures in the output spline */
	if (NULL == (lb = (CALPOINT *)callop(npts, sizeof (CALPOINT),
		"lb in addrows"))) return EALLOC;
	printf(
 "array previously had %d rows; adding %d additional rows on the %s\n",
		lsp->v_nrow, delr, (whichmod == MOD_START) ? "top" : "bottom");
	/* copy the general information about the old spline into the new one */
	cpymem((void *)&lv, (void *)lsp, sizeof (VSPLINE));
	/* increment the number of rows (that's what we're doing!) */
	lv.v_nrow += delr;	lv.v_midr = lv.v_nrow / 2;
	lv.v_y0cm = -(lsp->v_yrcm * (double)lv.v_midr);
	lv.v_yrpx = (((double)(lsp->v_deth)) + 10.) / ((double)(lv.v_nrow -1));
	if (whichmod == MOD_START) lpo = lb + delr * nc;
	else lpo = lb;
	for (colno = 0; colno < nc; colno++)
	   { /* set the input pointer for the current column */
		lpi = lsp->v_cpt + colno;
		/* . . . and the output pointer */
		if (whichmod == MOD_START) lpo = lb + delr * nc + colno;
		else lpo = lb + colno;
		/* for each column, copy over the existing values */
		for (rowno = 0; rowno < lsp->v_nrow; rowno++)
		   {
			cpymem((void *)lpo, (void *)lpi, sizeof (CALPOINT));
			lpi += nc;	lpo += nc;
		   }
		/* if we're adding new points at the beginning of the column,
		 we have to rewind back to the first unknown value */
		if (whichmod == MOD_START) lpo -= ((lsp->v_nrow) + 1) * nc;
		/* figure out what the new values should be.
		 This is currently a simple linear extrapolation */
		for (newpt = 0; newpt < delr; newpt++)
		   {
			if (whichmod == MOD_START)
			  {
				prev = lpo + nc;	ante = lpo + 2 * nc;
			  }
			else {
				prev = lpo - nc;	ante = lpo - 2 * nc;
			     }
			lpo->c_ptocx = 2 * prev->c_ptocx - ante->c_ptocx;
			lpo->c_ptocy = 2 * prev->c_ptocy - ante->c_ptocy;
			lpo->c_ctopx = 2 * prev->c_ctopx - ante->c_ctopx;
			lpo->c_ctopy = 2 * prev->c_ctopy - ante->c_ctopy;
			lpo->c_ift = 0;	/* it's extrapolated */
			lpo->c_qec = prev->c_qec; /* no other info avail */
			lpo->c_set = lpo->c_cset = 1;
			if (whichmod == MOD_START) lpo -= nc; else lpo += nc;
		   }
	   }
	/* release mem for old CALPOINTS*/
	memzap((AVOID **)&(lsp->v_cpt), "lsp->v_cpt in addrows");
	/* copy the working structure into the permanent one */
	cpymem((void *)lsp, (void *)&lv, sizeof (VSPLINE));
	lsp->v_cpt = lb;	/* set the CALPOINT structure appropriately */
	return EOK;
}

static int delrows(int delr, int whichmod, VSPLINE *lsp)
{ /*	This deletes rows on the top or bottom of the array */
	int		npts, rowno, colno, nc;
	static VSPLINE	lv;
	CALPOINT	*lb, *lpi, *lpo;

	if (delr <= 0) return EOK;	/* do nothing unless we must */
	/* make sure we will have at least one row left */
	if (delr >= lsp->v_nrow) return EBADVALUE;
	nc = lsp->v_ncol;	/* useful shorthand */
	/* define how many points there are in the output spline */
	npts = (lsp->v_nrow - delr) * nc;
	/* reserve memory for the CALPOINT structures in the output spline */
	if (NULL == (lb = (CALPOINT *)callop(npts, sizeof (CALPOINT),
		"lb in delrows"))) return EALLOC;
	printf(
 "array previously had %d rows; deleting %d additional rows from the %s\n",
		lsp->v_nrow, delr, (whichmod == MOD_START) ? "top" : "bottom");
	/* copy the general information about the old spline into the new one */
	cpymem((void *)&lv, (void *)lsp, sizeof (VSPLINE));
	/* decrement the number of rows (that's what we're doing!) */
	lv.v_nrow -= delr;	lv.v_midr = lv.v_nrow / 2;
	lv.v_y0cm = -(lsp->v_yrcm * (double)lv.v_midr);
	lv.v_yrpx = (((double)(lsp->v_deth)) + 10.) / ((double)(lv.v_nrow -1));
	for (colno = 0; colno < nc; colno++)
	   { /* set the input and output pointers appropriately */
		if (whichmod == MOD_START)
			lpi = lsp->v_cpt + delr * nc + colno;
		else	lpi = lsp->v_cpt + colno;
		lpo = lb + colno;
		/* for each row, copy over the existing values */
		for (rowno = 0; rowno < lv.v_nrow; rowno++)
		   {
			cpymem((void *)lpo, (void *)lpi, sizeof (CALPOINT));
			lpi += nc;	lpo += nc;
		   }
		/* increment the input pointer to skip over the ones
		 we aren't going to use. This works at either the beginning
		 or the end of the row. Isn't that nice! */
		lpi += delr * nc;
	   }
	/* release mem for old CALPOINTS*/
	memzap((AVOID **)&(lsp->v_cpt), "lsp->v_cpt in delrows");
	/* copy the working structure into the permanent one */
	cpymem((void *)lsp, (void *)&lv, sizeof (VSPLINE));
	lsp->v_cpt = lb;	/* set the CALPOINT structure appropriately */
	return EOK;
}

int moducamain(int argc, char *argv[])
{ /* Mainline for the moduca functionality */
	int		delc, delr, lineno, nitems, npt;
	int		orignc, orignr, outht, outwid, retv, tpts;
	static char	inl[120];
	double		fi0, fi1, newovox, newovoy;
	FILE		*fpi;
	CALPOINT	*vp;
	MODIFYUCA	cmu;
	static VSPLINE	*sgsp;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	outwid = outht = -10000;
	if ((argc > 1) && (allnumeric(argv[1])))
	  {
		outwid = atoi(argv[1]);
		if ((argc > 2) && (allnumeric(argv[2])))
			outht = atoi(argv[2]);
	  }
	if (NULL == (fpi = xgopen("UCAMOD", "r"))) return -1;
	zeromem((void *)&cmu, sizeof (MODIFYUCA));
	for (lineno = 0; (lineno < 5) && (NULL != fgets(inl, 116, fpi)); )
	   {
		if (inl[0] != '#') nitems = sscanf(inl, "%lf %lf", &fi0, &fi1);
		else	nitems = 0;
		if (nitems >= 2)
		  {
			switch (lineno)
			      {
				case	0: cmu.muc_ixc = fi0; cmu.muc_iyc = fi1;
					break;
				case	1: cmu.muc_mx = fi0; cmu.muc_my = fi1;
					break;
				case	2: cmu.muc_nx = fi0; cmu.muc_ny = fi1;
					break;
				case	3: cmu.muc_dx = fi0; cmu.muc_dy = fi1;
					break;
				case	4: cmu.muc_ox = fi0; cmu.muc_oy = fi1;
					break;
			     }
			lineno++;
		  }
	   }
	(void)fclose(fpi);	fpi = NULL;
	/* minimal check for data completeness */
	if ((abs(cmu.muc_nx) == 0) || (abs(cmu.muc_ny) == 0)) return -2;
	/* read the VSPLINE structure in */
	if (NULL == (sgsp = gspline(5, &retv))) return retv;
	if (outwid < -9000)
	  {
		outwid = (int)(sgsp->v_detw / 2);
		outht = (int)(sgsp->v_deth / 2);
	  }
	orignc = sgsp->v_ncol;	orignr = sgsp->v_nrow;
	delc = cmu.muc_mx - cmu.muc_ixc;
	/* add new columns on left if center-point's X index is too small */
	if ((delc > 0) && (EOK != (retv = addcols(delc, MOD_START, sgsp))))
		return retv;
	/* delete columns from left if center-point's X index is too large */
	else if ((delc < 0) && (EOK != (retv = delcols(
	 -delc, MOD_START, sgsp)))) return retv;
	/* add new columns on the right if the total number of columns
	 is smaller than we wanted it to be. */
	delc = cmu.muc_nx - sgsp->v_ncol;
	if ((delc > 0) && (EOK != (retv = addcols(delc, MOD_END, sgsp))))
		return retv;
	/* delete some columns from the right if the total number of columns
	 is larger than we wanted it to be. */
	else if ((delc < 0) && (EOK != (retv = delcols(
	 -delc, MOD_END, sgsp)))) return retv;
	delr = cmu.muc_my - cmu.muc_iyc;
	/* add new rows on the top if the center-point's Y index
	 is smaller than we want it to be */
	if ((delr > 0) && (EOK != (retv = addrows(delr, MOD_START, sgsp))))
		return retv;
	/* delete some rows from the top if the center-point's Y index
	 is larger than we want it to be */
	else if ((delr < 0) && (EOK != (retv = delrows(
	 -delr, MOD_START, sgsp)))) return retv;
	/* add new rows on the bottom if the total number of rows
	 is smaller than we wanted it to be. */
	delr = cmu.muc_ny - sgsp->v_nrow;
	if ((delr > 0) && (EOK != (retv = addrows(delr, MOD_END, sgsp))))
		return retv;
	/* delete some rows from the bottom if the total number of rows
	 is larger than we wanted it to be. */
	else if ((delr < 0) && (EOK != (retv = delrows(-delr, MOD_END, sgsp))))
		return retv;
	/* rescale the (X,Y) values in the cm mapping by the ratio
	 of the old # of cols and rows to the new #s.
	Also, apply the additive offsets to the resulting structures */
	newovox = ((double)(sgsp->v_ncol - 1)) / ((double)(orignc - 1));
	if (newovox != 1.) printf("Rescaling x values by %8.5f\n", newovox);
	else printf("No rescaling necessary in x\n");
	newovoy = ((double)(sgsp->v_nrow - 1)) / ((double)(orignr - 1));
	if (newovoy != 1.) printf("Rescaling y values by %8.5f\n", newovoy);
	else printf("No rescaling necessary in y\n");
	tpts = sgsp->v_nrow * sgsp->v_ncol;
	vp = sgsp->v_cpt + sgsp->v_ncol * cmu.muc_my + cmu.muc_mx;
	fi0 = vp->c_ctopx;	fi1 = vp->c_ctopy;
	printf("Reference point coordinates: %7.2f %7.2f\n", fi0, fi1);
	for (vp = sgsp->v_cpt, npt = 0; npt < tpts; vp++, npt++)
	   {
		vp->c_ctopx /= newovox;		vp->c_ctopy /= newovoy;
		vp->c_ptocx += cmu.muc_dx;	vp->c_ptocy += cmu.muc_dy;
		vp->c_ctopx += cmu.muc_ox - fi0;
		vp->c_ctopy += cmu.muc_oy - fi1;
	   }
	sgsp->v_x0px += cmu.muc_dx;	sgsp->v_y0px += cmu.muc_dy;
	sgsp->v_x0cm += cmu.muc_ox;	sgsp->v_y0cm += cmu.muc_oy;
	/* at this point we have to correct the x0px and y0px values for
	 some ill-understood recentering issues. We do this by calculating
	 the mean signed offset for a defined set of points */
	modxy0(outht, outwid, sgsp);
	sgsp = gspline(9, &retv);	globalkill();	return retv;
}

/* end of moduca functionality */
/* end of ucautils.c */
