/* cenutils.c -- utilities associated with CENTROIDS files

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	27 Dec 2006: filtercen: allow ice-ring editing
	11 Dec 2006: addcen written: appends ascii refls to CENTROIDS file
	28 Nov 2006: makeq21 call updated in filc_setup
	 9 Jun 2006: filterhead and reflinfo moved earlier in the file
	19 May 2006: merge cenmerge.c, filtercen.c into this file;
		eliminate rp->... and si2 args to getinthkl
	19 Mar 2006: sum_edcenresults does linefeeds correctly in Curses
	 4 Mar 2006: REFREF element frame is now called framem
	12 Oct 2005: housekeeping
	11 May 2005: check return status from xgjournal
	23 Apr 2005: replace refl_proc with refl_edit to reduce confusions
	12 Sep 2004: replace old filtercen.c with a version that looks
		like the old editcen.c, but without the CURSES calls
	11 Sep 2004: editcen & filtercen use routines here: refl_edit(), etc.
	 6 Sep 2004: cenmerge: cosmetic change
	21 Jan 2004: license message inserted
	 8 Jan 2004: replace global variable progname with call to currpname
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	 5 Jan 2003: reorganized source-file header
	25 Mar 2002: removed MSI attachments
	31 May 1995: "register" -> "register int" throughout
	 2 Mar 1995: add "end of sourcefile" comment
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 5 Oct 1994: use xgread and xgwrite for I/O on TROID data
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version of cenmerge

 Previous version of filtercen undocumented here since I'm not using that code;
	refer to ../curses/editcen.c for history.
	
   Purpose: manipulations of contents of CENTROIDS-style files
	cenmerge	merges 2 or more CENTROIDS-style files to form a
			new one, with environment-variable name CENTROIDS.
	filtercen	edit a CENTROIDS file to eliminate spots that violate
			the UPARAMS error limits in hkl, X, Y, or omega
   Cross references: ../prerefine/spot*.c, ../xgrefine/refresid.c

*************************************************************************/
#define		NEED_TIME	1
#include	"refidefs.h"
#define		MAXFILES	1000
#define		SIZOUTREF	20
#define		ASK_DELETION	63	/* code indicating that we should ask*/
#define		END_DELETIONS	64	/* code indicating it's time to quit*/
#define		BADSPACING	9999.	/* error value in relative D calcs */
#define		MIN_POSSIBLE_XY	(-49.)	/* detector size limit in cm */

extern int g_centroid(FILE *fp, REFREF *rp);
extern int p_centroid(EDITCENRUN *ecp);

extern int addcmain(int argc, char *argv[]);
extern int cenmmain(int argc, char *argv[]);
extern int filtercmain(int argc, char *argv[]);

/*************************************
	addcen: incorporate reflections into a CENTROIDS file
	based on ascii input. The input file should contain lines with
	the following format:
 " %f %f %f %d %d %f %f %f %f %f %f %f %f %f %f %f %f",
 &xobs, &yobs, &phiobs, &startframe, &nobs, &toti,
 &prof0, &prof1, &prof2, &prof3, &prof4, &prof5,
 &prof6, &prof7, &prof8, &prof9, &profa, &profb, &profc
 where the prof[0-c] are the fractional profile elements.
 If the profile elements aren't normalized, we do that in here.
 The end result is the incorporation into a CENTROIDS file of the
 reflections mentioned here.
*************************************/

static int inasccen(FILE *fp, TROID *tri)
{
 /* This reads an ascii CENTROID from the stream fp and puts it in
  the TROID structure pointed to be tri. It returns the number of
  profile elements found within the reflection if successful;
  it returns 0 if it reads a line in but there's something wrong with it;
  it returns -1 if it can't read a line. */
	int	j, nitems;
	char	instr[256];
	double	psum;

	zeromem((AVOID *)instr, 256);
	zeromem((AVOID *)tri, sizeof (TROID));
	if (NULL == fgets(instr, 240, fp)) return -1;
	nitems = sscanf(instr,
	 " %f %f %f %hd %hd %f %f %f %f %f %f %f %f %f %f %f %f",
	 &(tri->xto), &(tri->yto), &(tri->frame0), &(tri->obkmin),
	 &(tri->nobk), &(tri->totinten),
	 &(tri->obparts[0]), &(tri->obparts[1]), &(tri->obparts[2]),
	 &(tri->obparts[3]), &(tri->obparts[4]), &(tri->obparts[5]),
	 &(tri->obparts[6]), &(tri->obparts[7]), &(tri->obparts[8]),
	 &(tri->obparts[9]), &(tri->obparts[10]));
	if (nitems < 7)
	  { /* make sure we have at least one profile element */
		zeromem((AVOID *)instr, sizeof (TROID));
		return 0;
	  }
	 /* not enough elements in the listed structure: reduce tr_nobk */
	if (nitems < tri->nobk + 6) tri->nobk = nitems - 6;
	else if (nitems > tri->nobk + 6)
	  {
		for (j = tri->nobk; j < TROIDWIDTH; j++)
			tri->obparts[j] = 0.;
	  }
	/* renormalize the profile if necessary */
	for (psum = j = 0; j < tri->nobk; j++)
	   {
		psum += tri->obparts[j];
	   }
	if (psum < FUZZ)
	  { /* return empty reflection if sum of profiles isn't positive */
		zeromem((AVOID *)instr, sizeof (TROID)); return 0;
	  }
	for (j = 0; j < tri->nobk; j++) tri->obparts[j] /= psum;
	return tri->nobk;
}

int addcmain(int argc, char *argv[])
{ /* Mainline for reading ascii CENTROIDS in. We append to an existing
   CENTROIDS file */
	int		ntroids, nhere;
	TROID		utref;
	FILE		*fpi, *fpo;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	if (qerror((AVOID *)(fpo = xgopen("centroids", "a")),
		"open centroids output file"))
	  {
		globalkill();	return EOPENERR;
	  }
	if (qerror((AVOID *)(fpi = xgopen("udump", "r")),
		"open input dump-file"))
	  {
		globalkill();	return EOPENERR;
	  }
	for (ntroids = 0; 0 <= (nhere = inasccen(fpi, &utref)); )
	   if (nhere > 0)
	     {
		ntroids += nhere;
		if (verror(0 < xgwrite((void *)&utref, XG_TROID, fpo),
		 "write a reflection to the output"))
		  {
			(void)fclose(fpi);	(void)fclose(fpo);
			globalkill();	return EWRITERR;
		  }
	     }
	fprintf(stderr,
	  "%6d reflections read from dumpfile and written to CENTROIDS file\n",
		ntroids);
	(void)fclose(fpi); (void)fclose(fpo);	globalkill();	return EOK;
}

/* end of "addcen" functionality */

/*************************************
	cenmerge: merges two or more CENTROIDS-style files together.
	The names of the input files are specified in the command line,
	and the output is written to environment variable CENTROIDS.
	Under Unix one can should be able to get away with simply "cat"'ing
	the inputs together; under VMS it ought to require only a VMS
	"append", but somehow that sometimes fails.
**************************************/
int cenmmain(int argc, char *argv[])
{ /* The mainline for merging two or more CENTROIDS files together */
	register int	inputno, ntfile, ntroids;
	TROID		utref;
	FILE		*fpi, *fpo;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	if (argc < 3)
	  {
		fprintf(stderr,"Usage: %s input1 input2 ...\n",
		 xgprogname(NULL, 0));
		globalkill();	return EOK;
	  }
	if (qerror((AVOID *)(fpo = xgopen("centroids", "w")),
		"open centroids output file"))
	  {
		globalkill();	return EOPENERR;
	  }
	for (ntroids = inputno = 0; --argc; inputno++)
	   {
		if (qerror((AVOID *)(fpi = xgopen(*++argv, "r")),
			"open an input file"))
		  {
			(void)fclose(fpo); globalkill(); return EOPENERR;
		  }
		for (ntfile = 0; 0 < xgread((void *)&utref, XG_TROID, fpi);
		 ntfile++)
		   {
			if (verror(0 < xgwrite((void *)&utref, XG_TROID, fpo),
			 "write a reflection to the output"))
			  {
				(void)fclose(fpi);	(void)fclose(fpo);
				globalkill();	return EWRITERR;
			  }
		   }
		fprintf(stderr,
		  "%6d reflections read from %s and written to %s\n",
			ntfile, *argv, trname("centroids"));
		ntroids += ntfile;
		(void)fclose(fpi);
	    }
	fprintf(stderr,"%6d files read, %6d reflections collected\n",
		inputno, ntroids);
	(void)fclose(fpo);	globalkill();	return EOK;
}

/* end of cenmerge functionality */

/*************************************
	filterc:	non-interactive editing of a CENTROIDS file.
	This reads a "centroids" file and writes out another one,
	from which some of the observations have been deleted for various
	reasons. Normally only observations which have errors in
	h, k, l, X, Y, or phi greater than a certain value are deleted.
	Other criteria for potential deletions include excessive sharpness and
	nonmonotonic profiles.
	The "-c" (for "complement") option does deletions, but only if the
	reflections which DO satisfy the (hklXYo) criteria for acceptance are
	deleted. This provides a method of separating the reflections
	belonging to one lattice from the ones belonging to another lattice.
*************************************************************************/

int g_centroid(FILE *fp, REFREF *rp)
{ /*	This reads data for a single peak into the REFREF reflection pointed
	to by `rp' from the unix-style "centroids" file.
	It returns 1 if it's successful; it returns 0 if the first
	part of the reflection is absent. Flags are set when
	the mapping from pixels to centimeters fails. */
	int	jfr, pixres;

	if (0 >= xgread((void *)rp, XG_TROID, fp)) return 0;
	rp->framem = rp->frame0;	/* save the unrefined frame number */
	jfr = rp->frame0 + 0.49999;
	if (1 != (pixres = pix2cm(rp->xto,rp->yto,&rp->xtcm,&rp->ytcm, jfr)))
	  {
		rp->rflags |= NOCALIB; if (!pixres) rp->xtcm = rp->ytcm = -100.;
	  }
	else rp->rflags &= ~NOCALIB;
	return 1;
}

void filterhead(FILE *fp, int withparams)
{ /*	Header for output to reflection information for filtercen or editcen */
	fprintf(fp,"\t\t\tDeleted Reflections:\n");
	if (withparams) fprintf(fp,
	 "     h     k     l   X in pixels   Y in pixels phi in frames");
	else fprintf(fp, " X,pixels Y,pixels Phi,frames");
	fprintf(fp,
	 "   inten- init   Reason for Normalized Profile Elements\n");
	if (withparams) fprintf(fp,
	 "                    obsvd calctd  obsvd calctd obsvd  calctd");
	else fprintf(fp, "                             ");
	fprintf(fp, "     sity  frm    rejection\n");
}

int reflinfo(REFINERUN *rfp)
{ /*	This populates the reflection structure rfp->rfr_rpp and
	returns a value between 0 and 5 describing its quality */
	int	i, qual;
	double	fmdif;
	REFREF	*rp;

	qual = 0;	rp = rfp->rfr_rpp;
	if ((rp->xtcm < MIN_POSSIBLE_XY) || (rfp->rfr_yin < MIN_POSSIBLE_XY))
		qual = EE_BADCM2PIX;
	else {
		scattv(rfp);	refgonio(rfp); /* get coordinates of refl */
		if (!legald(rfp, rp)) qual = EE_INICERING;
		(void)getinthkl(rfp);
		if (0 != (i = s2xyp(rp, rfp))) qual = i;
		else fmdif = gcloser(rp, rfp);
	     }
	if (qual)
	  {
		rp->xtc = rp->ytc = -10.; rp->framc = -100.;
	  }
	return qual;
}

static int filc_setup(int argc, char *argv[], EDITCENRUN *ecp)
{ /*	Argument and file processor for the filtercen program.
	Files used are given conventional logical or environment names:
		centroids	input centroids file
		cenout		output centroids file
		xlog		printable logfile
	Flags:
		-c		automatically KEEP the (hklXYO) violators and
				 throw out the good refls.
		-d<val>		maximum relative plane spacing for closely
				 spaced reflections
		-h		check for (X,Y,Omega,h) errors
		-i<val>		minimum intensity to be written out
		-l<val>		low-resolution cutoff in Angstroms
		-m<val>		max non-monotonic fraction allowed
		-n		min value for secondary peak alllowed
		-p		check that refl doesn't fall in an icering
		-s		do NOT check for (X,Y,omega,h) errors
		-u<val>		high-resolution cutoff in Angstroms	*/
	register int	j;
	int		i, nread, quality;
	char		*val, **dcp, **flags;
	double		fval;
	FRAMEINDEX	*lft;
	REFINERUN	*rfp;
	REFREF		rs;
	static char	*flist[] =
	 {"centroids", "uparams", "uspline", "amask", "cenout", "xlog", NULL};
	static char	ftype[] = {1, 1, 1, 1, 2, 2, 0 };

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	ecp->ec_flags = 0; /* default: all flags off */
	ecp->ec_nrin = ecp->ec_nrout = ecp->ec_nrdel = 0;
	if (NULL == (lft = ftables(3)))
	  {
		ecp->ec_fmmin = -10000;	ecp->ec_fmmax = 10000;
	  }
	else {
		ecp->ec_fmmin = lft->fi_firstfm - 1;
		ecp->ec_fmmax = lft->fi_lastfm + 1;
	     }
	ecp->ec_mininten = 0;	ecp->ec_maxnonmono = 1;
	ecp->ec_minsecpeak = -1000;
	ecp->ec_resmin = 10000.;	ecp->ec_resmax = 0.;
	ecp->ec_maxreld = 10000.;
	ecp->ec_flags = EC_CHECKHXYO | EC_AUTODEL;
	flags = flagproc(argc, argv, "CDHILMNPU");
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	   {
		val = (*dcp) + 1;
		fval = ((val) && ('\0' != *val)) ? atof(val) : 0.;
		switch (**dcp)
		 {
			case	'C':	ecp->ec_flags |=
				(EC_AUTODEL | EC_COMPLEMENT); break;
			case	'D':	ecp->ec_maxreld = fval;	break;
			case	'H':	ecp->ec_flags |= EC_CHECKHXYO;	break;
			case	'I':	ecp->ec_mininten = fval;	break;
			case	'L':	ecp->ec_resmin = fval;	break;
			case	'M':	ecp->ec_maxnonmono = fval;	break;
			case	'N':	ecp->ec_minsecpeak = fval;	break;
			case	'P':	ecp->ec_flags |= EC_ICERINGS;	break;
			case	'S':	ecp->ec_flags &= ~EC_CHECKHXYO;	break;
			case	'U':	ecp->ec_resmax = fval;	break;
			default:	break;
	      }
	   }
	if (qerror((AVOID *)(ecp->ec_fpi = xgopen("centroids", "r")),
		"open input centroids file")) return EOPENERR;
	/* open the input CENTROIDS file and count the refls in it */
	for (ecp->ec_nrin = 0; (0 < xgread((void *)&rs, XG_TROID,
	 ecp->ec_fpi)); ) (ecp->ec_nrin)++;
	(void)fseek(ecp->ec_fpi, 0, 0); /* rewind centroids file */
	if (qerror((AVOID *)(ecp->ec_fpo = xgopen("cenout", "w")),
		"open output centroids file")) return EOPENERR;
	if (NULL == (rfp = managerfp(7, (size_t)ecp->ec_nrin))) return EALLOC;
	rfp->rfr_tl0 = NOW;	ecp->ec_rfp = rfp;
	if (NULL == (rfp->rfr_spl = gspline(5, &i))) return i; /* get spline */
	if (NULL == (rfp->rfr_xp = (REFPARS *)callop(1, sizeof (REFPARS),
	 "rfp->rfr_xp in filtercen"))) return EALLOC;
	if (EOK != (i = refparin(rfp->rfr_xp, &nread)))
	  {
		memzap((AVOID **)&(rfp->rfr_xp), "rfp->rfr_xp in filtercen");
		ecp->ec_flags &= ~EC_CHECKHXYO;
	  }
	else { /* rotations about tilt and swing */
		makeq21(rfp->rfr_xp);
		/* for internal purposes we need the reciprocal-space matrix
		to be rescaled by lambda to make it unitless */
		for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
			(rfp->rfr_xp)->xrec[i][j] *= (rfp->rfr_xp)->xlambda;
	     }
	if (qerror((AVOID *)(ecp->ec_fpl = propen("xlog")),
	 "open output log file")) return EOPENERR;
	if (NULL == (rfp->rfr_rpp0 = (REFREF *)callop(ecp->ec_nrin,
	 sizeof (REFREF), "rfp->rfr_rpp0 in filc_setup"))) return EALLOC;
	if (ecp->ec_flags & EC_ICERINGS) rfp->rfr_flags |= RR_EXCLUDEICE;
	printf("Reading and processing centroids: ");
	/* now read all the reflections in */
	for (rfp->rfr_rpp = rfp->rfr_rpp0, i = 0; i < ecp->ec_nrin;
	 i++, (rfp->rfr_rpp)++)
	   {
		if (0 == g_centroid(ecp->ec_fpi, rfp->rfr_rpp)) break;
		quality = 0;
		if (ecp->ec_flags & EC_CHECKHXYO) quality = reflinfo(rfp);
		else if (ecp->ec_flags & EC_ICERINGS)
		  { /* need the diffraction vector to calc sin(theta)/lambda */
			scattv(rfp);
			if (!legald(rfp, rfp->rfr_rpp)) quality = EE_INICERING;
		  }
		if (NULL == rfp->rfr_xp) /* flag missing params */
			(rfp->rfr_rpp)->rflags |= CHANGEDINDEX;
		(rfp->rfr_rpp)->rflags |= (quality << 8);
	   }
	ecp->ec_nrin = i;	/* in case the count is different */
	printf(" %6d reflections read and processed\n", ecp->ec_nrin);
	(void)fclose(ecp->ec_fpi); /* we'll do the rest in memory */
	rfp->rfr_tl0 = NOW;
	strcpy(rfp->rfr_rundate, ctime(&(rfp->rfr_tl0)));
	header(stdout, PROCESSING, flist, ftype);
	header(ecp->ec_fpl, PROCESSING, flist, ftype);
	/* set up the headers for the logfile */
	filterhead(ecp->ec_fpl, (rfp->rfr_xp != NULL));
	return EOK;
}

static double udspace(REFREF *c0, REFREF *c1, double clam)
{ /* This computes the D spacing associated with a pair of TROIDs.
	Any errors that arise from attempting to compute it result in
	a return of 9999. Angstroms, which should result in the refls
	being barred from the output list */
	int	i;
	double	dif, rspa;

	for (rspa = i = 0; i < 3; i++)
	   {
		dif = c0->rs[i] - c1->rs[i];	rspa += dif * dif;
	   }
	return (rspa < DFUZZ) ? BADSPACING : clam / sqrt(rspa);
}

void log_ecendel(FILE *fp, REFREF *rp, int reason)
{ /* this spits to the stream `fp' a message appropriate to the
	deletion of  the reflection pointed to by `rp'. */
	register int	i;
	static char *desc[] = {		 "Good reflect",
	 "cm->pix err ", "sx^2+sy^2 ~0", "sz^2 ~ 0    ", "Lorentz err ",
	 "dh too large", "dk too large", "dl too large", "dX too large",
	 "dY too large", "dZ too large", "Dspace >Dmax", "Dspace <Dmin",
	 "outside det.", "inact. area ", "Inten < Imin", "Frame# Error",
	 "bad profile ", "Z too sharp ", "close neighb", "in icering  " } ;

	if (rp->rflags & CHANGEDINDEX) fprintf(fp, " %8.2f %8.2f %10.3f",
		rp->xto, rp->yto, rp->framem);
	else fprintf(fp,
	 "%6.1f%6.1f%6.1f%7.1f%7.1f%7.1f%7.1f%7.1f%7.1f",
	 rp->rh[0], rp->rh[1], rp->rh[2], rp->xto, rp->xtc, rp->yto,
	 rp->ytc, rp->framem, rp->framc);
	fprintf(fp, "%9.1f%5d %s", rp->totinten, rp->obkmin, desc[reason]);
	for (i = 0; i < rp->nobk; i++) fprintf(fp, "%6.3f", rp->obparts[i]);
	fprintf(fp, "\n");
	return;
}

int p_centroid(EDITCENRUN *ecp)
{ /* write a TROID structure to an output file */
	if (verror(0 < xgwrite((void *)((ecp->ec_rfp)->rfr_rpp), XG_TROID,
	 (ecp->ec_fpo)), "write centroid to output file")) return EWRITERR;
	(ecp->ec_nrout)++;	return EOK;
}

static void getprofvals(REFREF *tr, double pva[])
{ /*	Returns	pva[0] as the peak of the profile;
		pva[1] as the strongest nonmonotonic element
		pva[2] as the maximum secondary peak in the profile. */
	int	i, ima;
	double	maxover, maxsec, pmax, prev, this;

	ima = (tr->nobk) / 2;	/* find the peak: */
	for (pmax = i = 0; i < tr->nobk; i++)
		if ((this = tr->obparts[i]) > pmax)
		  {
			pmax = this; ima = i;
		  }
	/* find the second-highest element and check to see that the
	 profile is monotonic (i.e., it goes up, then down) */
	maxover = maxsec = 0.;
	if (ima)
	  {
		for (i = ima - 1, prev = pmax; i >= 0; i--)
		   {
			if ((this = tr->obparts[i]) > prev) maxover = this;
			if (this > maxsec) maxsec = this;	prev = this;
		   }
	  }
	if (ima < tr->nobk - 1)
	  {
		for (i = ima + 1, prev = pmax; i < tr->nobk; i++)
		   {
			if ((this = tr->obparts[i]) > prev) maxover = this;
			if (this > maxsec) maxsec = this;	prev = this;
		   }
	  }
	pva[0] = pmax;	pva[1] = maxover;	pva[2] = maxsec;
}

static double scclosest(REFREF *rp, REFINERUN *rfp, int totrefs)
{ /*	This finds the reflection closest to the current one other than
	the reflection itself. It returns the D spacing between the two. */
	int	i;
	double	maxrel, reld;
	REFREF	*rq;

	for (maxrel = i = 0, rq = rfp->rfr_rpp0; i < totrefs; i++, rq++)
	   if (rq != rp)
	     {
		reld = udspace(rp, rq, (double)((rfp->rfr_xp)->xlambda));
		if (reld > maxrel) maxrel = reld;
	     }
	return maxrel;
}

int refl_edit(EDITCENRUN *ecp)
{ /*	This does most of the hard work involved in filtering data.
	The h,k,l of the reflection is computed, along with predicted
	phi, X, and Y values.  If all of these are within the error
	values of the observed values, we skip the refl, i.e. we return 0.
	If there is an error and we are doing automatic deletions,
	we delete the observation and return 1.
	If there is an error and we aren't doing automatic deletions,
	we return -1 so another the calling routine can put up a display
	and allow the user to decide what to do with the reflection.
	All of this activity is recorded in a log file. */
	int		i, checkit, chuckit;
	double		diffh, fmdif, worstdh, profvals[3];
	REFINERUN	*rfp;
	REFPARS		*xp;
	REFREF		*rp;

	fmdif = checkit = 0;	rfp = ecp->ec_rfp;
	rp = rfp->rfr_rpp;	xp = rfp->rfr_xp;
	if (rp->rflags & NOCALIB) checkit = EE_XYOFFDETEC;
	else checkit = (rp->rflags & ~0377) >> 8;
	getprofvals(rp, profvals);
	/* if the refl is erroneous or the user has asked for queries
	on all reflections, put the refl up for analysis.
	The characteristics that actually count as errors are checked
	prior to looking to see whether it agrees with the prediction. */
	if (!checkit)
	  {
		if (rp->totinten < ecp->ec_mininten) checkit = EE_TOOWEAK;
		else if ((rp->frame0 < ecp->ec_fmmin) ||
			(rp->frame0 > ecp->ec_fmmax)) checkit = EE_OUTOFRANGE;
		if (profvals[1] > ecp->ec_maxnonmono * profvals[0])
			checkit = EE_NONMONOTONE;
		else if (profvals[2] < ecp->ec_minsecpeak * profvals[0])
			checkit = EE_TOOSHARP;
		else if (NULL != rfp->rfr_xp) /* These tests require params */
		  {
			if (rp->dspac > ecp->ec_resmin) checkit = EE_TOOLOWRES;
			else if (rp->dspac < ecp->ec_resmax)
				checkit = EE_TOOHIGHRES;
			else if (scclosest(rp, ecp->ec_rfp, ecp->ec_nrin) >
				ecp->ec_maxreld) checkit = EE_CLOSESPACE;
		  }
	  }
	if ((!checkit) && (NULL != rfp->rfr_xp)) /*make sure hkl are integers*/
	  {
		for (worstdh = i = 0; i < 3; i++)
		   if ((diffh = fabs(rp->rh[i] - rp->rhi[i])) > xp->xerrind)
		     {
			if (diffh > worstdh)
			  {
				checkit = i + 5;	worstdh = diffh;
			  }
		     }
	  }
	/* make sure X,Y, and phi(Z) are close to predictions */
	if ((!checkit) && (NULL != rfp->rfr_xp))
	  {
		if (fabs(rp->xto - rp->xtc) > xp->xerrx) checkit = EE_XERROR;
		else if (fabs(rp->yto - rp->ytc) > xp->xerry)
			checkit = EE_YERROR;
		else if (fmdif > xp->xerrphi) checkit = EE_ZERROR;
	  }
	/* the concept of complementarity only applies to hkl or XYZ errors */
	if ((0 == checkit) ||
	 ((EE_HERROR <= checkit) && (checkit <= EE_ZERROR)))
	  {
		chuckit = (ecp->ec_flags & EC_COMPLEMENT) ? !checkit : checkit;
	  }
	else	chuckit = checkit;	/* otherwise we definitely flag it */
	if (ecp->ec_flags & EC_EVERYREFL) chuckit = 1;
	if (chuckit)
	  {
		if (ecp->ec_flags & EC_AUTODEL) /* do the deletion and quit */
		  {
			log_ecendel(ecp->ec_fpl, rp, (int)checkit);
			(ecp->ec_nrdel)++;	return 1;
		  }
		else return -checkit; /* flag: "put it on the screen" */
	  }
	else return p_centroid(ecp);
}

void sum_edcenresults(FILE *fp, EDITCENRUN *ecp)
{ /* This summarizes a deletion effort to the stream `fp' */
	char	*endst;
	REFREF	*rp0, *rpe;

	endst = lineend(fp, (0 != (ecp->ec_flags & EC_USECURSES)));
	if (ecp->ec_nrdel) fprintf(fp,
	 "%6d observation%sdeleted from input centroids file%s",
	 ecp->ec_nrdel, ecp->ec_nrdel != 1 ? "s " : " ", endst);
	fprintf(fp,
	 "%6d observation%swritten to the output file%s",
	 ecp->ec_nrout, ecp->ec_nrout != 1 ? "s " : " ", endst);
	rp0 = (ecp->ec_rfp)->rfr_rpp0;	rpe = (ecp->ec_rfp)->rfr_rpp - 1;
	if ((NULL != rp0) && (NULL != rpe))
	  {
		fprintf(fp, "starting and ending reflections: ");
		fprintf(fp, "[%7.1f,%7.1f,%7.1f],[%7.1f,%7.1f,%7.1f]%s",
		 rp0->xto, rp0->yto, rp0->frame0,
		 rpe->xto, rpe->yto, rpe->frame0, endst);
	  }
}

int filtercmain(int argc, char *argv[])
{ /* Mainline for noninteractive editing of a CENTROIDS file */
	int			i, j;
	REFINERUN		*rfp;
	static EDITCENRUN	ecr;

		/* decode arguments for files and flags */
	if (EOK != (i = filc_setup(argc, argv, &ecr))) return i;
		/* loop through the data, processing each observation */
	rfp = ecr.ec_rfp;
	for (rfp->rfr_rpp = rfp->rfr_rpp0, j = 0;
	 j < ecr.ec_nrin; j++, (rfp->rfr_rpp)++)
	   {
		if (EWRITERR == (i = refl_edit(&ecr))) break;
		if ((i) && (1 != i))
		  {
			if (i == END_DELETIONS) i = EOK; break;
		  }
	   }
	if (NULL != ecr.ec_fpo)
	  {
		(void)fclose(ecr.ec_fpo);	ecr.ec_fpo = NULL;
	  }
	sum_edcenresults(stdout, &ecr);	sum_edcenresults(ecr.ec_fpl, &ecr);
	memzap((AVOID **)&(ecr.ec_rfp), "ecr.ec_rfp in endfilter");
	if (1 == i) i = EOK;	return i;
}

/* end of filtercen functionality */

/* end of cenutils.c */
