/* imutils.c -- utilities for manipulating and printing contents of images

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	12 Jun 2007: imformatmain does dark and flood corrections
	 1 Jun 2007: imformatmain: support for output in SMV format,
	 including forcing rescaling at 65535 if requested;
	 delete final diagnostic; reorder runtime options and comments
	19 Apr 2007: float->double conversions; reorganize use of "{"
	11 Aug 2006: manage byteswapped formats without specific defines
	24 Jun 2006: include xgimage.h so we can manipulate detector images
	18 May 2006: merge {edit_mar,marccd_h,show_head,merge_fm,imformat}.c
	22 Apr 2006: new code for edit_mar, based loosely on marccd_h.c
	 7 Mar 2006: show_head: handle Mar IP formats with "ismarip";
		handle file-format name correctly
	31 Dec 2005: merge_fm: don't qsort if fewer than 2 elements
	 6 Nov 2005: eliminate static globals
	12 Oct 2005: housekeeping in show_head
	13 May 2005: move merge_fm from nogui to utilities
	12 May 2005: rename show_head & slightly reorganize from show_over.c
		to incorporate functionalities of show_over and show_mail
	11 May 2005: check return status from xgjournal
	 6 May 2005: marccd_h: convert this to a callable routine
	11 Apr 2005: rename mergefm.c to merge_fm.c to prevent overlap
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	 6 Sep 2004: cosmetic change to merge_fm
	28 Aug 2004: routines used only here are declared as static; cosmetics
	21 May 2004: marccd_h: set ANNOUNCE flag so a useful diagnostic emerges
	18 Feb 2004: imformat: change kill_ftab call so it doesn't have args
	26 Jan 2004: marccd_h: static MAIL *heads made local to main()
	26 Jan 2004: some static globals replaced by local variables
	21 Jan 2004: license message inserted and initial comments reordered
	 8 Jan 2004: replace extern global variables with static globals;
		callop() & memzap() replace shcalloc() & shfree();
		call kill_mail(), kill_ftabl near the end of each functionality
	 8 Jan 2004: merge_fm: replace extern globals with static globals;
		call kill_mail() at the end
	23 Sep 2003: remove declaration of strcpy()--sometimes it's a macro
	20 Sep 2003: force16 is initialized to zero in showheadmain
	19 Sep 2003: in informatmain, initialize resu = 0.
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Sep 2003: imformat: replace access() calls with xgaccess();
		this allows us to gunzip files as we pass through
	18 Aug 2003: replace local code for identifying detector type
		with a call to head2type() and imtypestr()
	17 Aug 2003: imformat: trivial renaming of some arguments
	 5 Jan 2003: reorganized source-file header
	15 Nov 2002: correct (we hope) a leap-year bug
	 1 Nov 2002: merge_fm: replace framename[] with lfname[]
		so we don't collide with a global
	25 Mar 2002: removed MSI attachments
	31 Jan 1998: initial version of imformat
	24 Aug 1995: show_over: 16-bit Mar overflows handled correctly
	 7 Jul 1995: use pointer-equivalent integers for printouts
	31 May 1995: "register" -> "register int" throughout
	 2 Mar 1995: diagnostics to stdout; shared-memory alloc and free;
		late-count value treated as float
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 1 Dec 1994: diagnostic output allows detector (X,Y) > 999
	30 Sep 1994: show_over: use floating-point stepsize;
		print wavelength value if nonzero
	 5 May 1994: show_over: file header cleanup, POLYDET->XGEN;
		replace %lx,%ld formats with %x, %d
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version of show_over and merge_fm
	22 Dec 1992: merge_fm: add dark-current subtraction
	16 Apr 1990: merge_fm: fixed bug for >= 3 frames in index-# mode
	14 Oct 1987: merge_fm: original code, written at Roche in Basel.
   Purpose: various utilities for changing or printing the contents of images
	edit_mar	edits a MarCCD header in place
	marccd_h	dump out contents of a MarCCD header
	show_head	ascii dump of contents of an image header and
			 its overflow table
	merge_fm	merges two or more X-GEN detector images into one;
			 The tricky part is fixing mailbox & overflow tables
	imformat	converts readable images into one of a select group
			of formats; normally, it does spatial correction
			and deletes the original image.
   Cross references: ../xglib/{marccd,imhead,image,imageout}.c,
		../utilities/spatial.c
*************************************************************************/
#define		NEED_TIME	1
#include	"xgimage.h"
#include	<dirent.h>
#define		MAXCOLS		7
#define		SHF_FORCE16	4
#define		SHF_OVERFLOWS	8
/* bits in the `imf_flags' variable */
#define	TO_PCS	1	/* convert to PCS format */
#define	TO_MAR	2	/* convert to Mar format */
#define	TO_BRUKER	3	/* convert to 8-bit Bruker/Siemens format */
#define	TO_BRUKER16	4	/* convert to 16-bit Bruker/Siemens format */
#define	TO_RAXIS	5	/* convert to R-Axis format */
#define	TO_SMV		6	/* convert to SMV format */
#define	RESCALE65K	0x80	/* rescale < 65K during interpolation */
#define	DO_SPATIAL	0x100	/* perform spatial correction */
#define	DELETE_ORIGINAL	0x200	/* delete original file */
#define	REDIMENSION	0x400	/* redimension detector image */
#define	OVERRIDE_COORDS	0x800	/* override the coordinate conventions */
#define	PARTIAL_VERBOSE	0x1000	/* only a moderate number of comments */
#define	VERY_VERBOSE	0x2000	/* more comments output */
#define	PAD_TO_2TO_N	0x4000	/* pad each row to make it a power of 2 */
#define	DARKSUB		0x8000	/* subtract dark-current */
#define	FLOODCORR	0x10000	/* divide by flood-field image */
#define	OVERMIN	0.0001
#define	OVERMAX	10

 typedef struct {
		int		imf_flags, imf_nmbox;
		int		imf_outwid, imf_outht;
		xgulong		imf_outtype;
		double		imf_outdist, imf_out2th, imf_darkscale;
		FILE		*imf_fpl;
		MAIL		*imf_heads;
		FRAMEINDEX	*imf_lfmi;
		} IMWORK;

extern int edmarccd(char *imname, char *edname);

extern int editmarmain(int argc, char *argv[]);
extern int marccd_hmain(int argc, char *argv[]);
extern int showheadmain(int argc, char *argv[]);
extern int mergefmmain(int argc, char *argv[]);
extern int imformatmain(int argc, char *argv[]);

/***********************************************
	editmar: read a Mar image and alter its header contents
		according to the contents of a text file
************************************************/
int editmarmain(int argc, char *argv[])
{ /* Mainline for editing the contents of a Mar CCD header */
	int		iret;
	char		*imname, *editname;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	(void)flagproc(argc, argv, (char *)NULL);
	if (argc < 1) /* must specify at least the file name */
	  {
		fprintf(stderr, "Usage: editmar imagefile [editfile]\n");
		return EOK;
	  }
	imname = argv[1];
	if (argc < 2) editname = "marccd.ed";
	else editname = argv[2];
	if (EOK != (iret = edmarccd(imname, editname)))
	  {
		fprintf(stderr,
		 "Editing of image %s with editfile %s returned %d\n",
			imname, editname, iret);
	  }
	else fprintf(stdout,
		"Editing of image %s with editfile %s returned %d\n",
			imname, editname, iret);
	return EOK;
}

/* end of edit_mar functionality */

/**********************************
	marccd_h:	dump of contents of Mar CCD header
***********************************/

int marccd_hmain(int argc, char *argv[])
{ /* Mainline for reporting the contents of a Mar CCD header */
	int		iflags, iret;
	static MAIL	*heads;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	(void)flagproc(argc, argv, (char *)NULL);
	iflags = RETURN_IF_CONVERTED | ANNOUNCE;
	if (NULL == (heads = (MAIL *)callop(1, sizeof (MAIL),
		"heads in marccd main")))
	  {
		fprintf(stderr, "Error allocating mailbox memory\n");
		return EALLOC;
	  }
	if ((argc > 1) && (argv[1][0] != '\0'))
	  {
		if (EOK != (iret = read_image(argv[1], iflags, heads)))
		  {
			fprintf(stderr, "Error %d from read_image\n", iret);
			return EREADERR;
		  }
	  }
	else fprintf(stderr, "Usage: %s imagename\n", argv[0]);
	iflags = 1;	kill_mail(&heads, &iflags);
	return EOK;
}

/* end of marccd_h functionality */

/**********************************
	show_head: prints a summary of an image's header (and overflows)
***********************************/

static int disp_overlist(MAIL *headp, int is16)
{ /*	This displays the list of overflowed pixels without regard to
	whether they appear in the same form in the data themselves. */
	register int	i;
	int		ival, ncols, novers;
	VOFLOW		*vofp;

	ncols = novers = headp->oflowcount;
	if (ncols > MAXCOLS) ncols = MAXCOLS;
	printf(" Noted Overflows:\n");
	for (i = 0; i < ncols; i++) printf("  Value    X    Y ");
	printf("\n");
	for (vofp = headp->overflow, i = 0; i < novers; i++, vofp++)
	   {
		ival = (is16) ? (4 * (int)vofp->value) + 65535 : vofp->value;
		printf(" %6d %4d %4d ", ival,
		 (int)vofp->xcoord, (int)(headp->ydimen - vofp->ycoord -1));
		if (!((i+1) % MAXCOLS)) printf("\n");
	   }
	printf("\n");
	return ncols;
}

static void disp_top(MAIL *headp, int is16, int ncols)
{ /*	This lists all pixels over 255 or 65535 */

	register int	i, ix, iy;
	int		hit, jlim, wid;
	unsigned short	j, *usp;
	unsigned char	*ucp;

	jlim = is16 ? 65535 : 255; wid = headp->xdimen; hit = headp->ydimen;
	printf(" Positions with %d's in them:\n", jlim);
	for (i = 0; i < ncols; i++) printf("           X    Y ");
	printf("\n");
	if (NULL != headp->data8) /* 8-bit data case */
	  {
		for (ix = i = 0, ucp = headp->data8; ix < wid; ix++)
			for (iy = 0; iy < hit; iy++, ucp++)
				if ((j = *ucp) >= jlim)
				  {
					printf("        %4d %4d ", ix, iy);
					if (!((++i) % MAXCOLS)) printf("\n");
				  }
		printf("\n");
	  }
	else if (NULL != headp->data16)	/* 16-bit data case */
	  {
		for (ix = i = 0, usp = headp->data16; ix < wid; ix++)
			for (iy = 0; iy < hit; iy++, usp++)
				if ((j = *usp) >= jlim)
				  {
					printf("        %4d %4d ", ix, iy);
					if (!((++i) % MAXCOLS)) printf("\n");
				  }
		printf("\n");
	  }
}

int showheadmain(int argc, char *argv[])
{ /* This is the mainline for the functionality to display image headers */
	register int	i;
	int		day, detyp, hour, is16bit, min, ncols;
	int		readflags, readret, showhflags, tzoff, year;
	xgulong		ift, swaft;
	time_t		sec;
	char		isleap, *tzstr, **dcp, **flags, rundate[26];
	double		tev;
	PAU		prd8, prd16, provf, prov16;
	static MAIL	*heads;
	static int	nmbox;
	static DATARUN	*runhdp;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	showhflags = 0;	/* defaults for run-time flags */
	if (argc < 2)
	  {
		fprintf(stderr,"usage: showhead filename\n");	return EOK;
	  }
	flags = flagproc(argc, argv, "FO");
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	 {
	   switch (**dcp)
	     {
		case	'F':	showhflags |= SHF_FORCE16; break;
		case	'O':	showhflags |= SHF_OVERFLOWS;	break;
	     }
	  }
	readflags = GET_RUNHEADER | GRIPE_IF_SHORT;
	if (NULL == heads)
	  {
		nmbox = 1;
		if (NULL == (AVOID *)(heads = (MAIL *)callop(1, sizeof (MAIL),
		 "heads in showhead"))) return EALLOC;
	  }
	for (i = 1; i < argc; i++)
	   {
	    if (argv[i][0] != '-')
	      {
		if (FRAME_OK != (readret =
		  read_image(argv[i], readflags, heads)))
		  {
			(void)image_error(stdout, argv[i], readret, 0);
			continue;
		  }
		ift = heads->format;
		is16bit = ismarip(ift) && (heads->data16);
		if (showhflags & SHF_FORCE16) is16bit = 1;
		printf("\t\tFrame %s:\n", argv[i]);
		prd8.ucpt = heads->data8;	prd16.uspt = heads->data16;
		printf("8bit data at 0x%8x; 16bit data at 0x%8x\n",
		 prd8.ui, prd16.ui);
		provf.vpt = heads->overflow;	prov16.vpt = heads->oflow16;
		printf("8,16-bit overflows begin at 0x%8x, 0x%8x\n",
		 provf.ui, prov16.ui);
		detyp = head2type(heads);
		swaft = lswawb(ift, 0);
		if (ift < swaft)
		  printf(" format = %d (%s)\n", (int)ift,
			imtypestr(detyp & 0xff));
		else printf(" format = %d byteswapped (%s)\n",
			(int)swaft, imtypestr(detyp & 0xff));
		printf(
		 "xdimension, ydimension, quadrant = %6d, %6d, %6d\n",
		 (int)heads->xdimen,(int)heads->ydimen,(int)heads->quadrant);
		tev = heads->eventcount;	if (tev < 0.) tev *= -64.;
		printf(
		 "events, #overflows = %11.0f, %4d; max = %5d at [%4d,%4d]\n",
			tev, (int)heads->oflowcount,
			(int)heads->zmax,(int)heads->zmaxx,(int)heads->zmaxy);
		tev = heads->late;	if (tev < 0.) tev *= -64.;
		printf(
 "zdisp, ontime, late, accumsecs, elap100ths = %4d,%8d,%11.0f,%6d,%6d\n",
			(int)heads->zdisp, (int)heads->ontime,
			tev, (int)heads->accumsecs,
			(int)heads->elap100ths);
		year = (heads->datetime[0]) >> 16;
		if (year < 200)
		  {
			if (year < 70) year += 2000;
			else year += 1900;
		  }
		day = ((unsigned short)(heads->datetime[0] & 0777)) - 1;
		hour = heads->datetime[1] >> 24;
		min = ((heads->datetime[1] >> 16) & 0377);
		sec = (heads->datetime[1] & 0377);
		printf("datetime[0], datetime[1] = 0x%x, 0x%x\n",
			(int)heads->datetime[0], (int)heads->datetime[1]);
		printf(
 "Data collection began %2.2d:%2.2d:%2.2d, day %3d of %4d, i.e. ",
		 hour, min, (int)sec, day, year);
		isleap = !((year + 2) % 4);
		if (NULL != (tzstr = getenv("TZ"))) tzoff = atoi(&tzstr[3]);
		else	tzoff = 5;
		if ((90 + isleap) <= day && day <= (340+isleap)) tzoff--;
		hour += tzoff;	year -= 1970;
		day += (year + 2) / 4 - 1; if ((isleap) && (day <= 59)) day--;
		/* I hope this is right... */
		if (year >= 30)
		  {
			day++;	hour++;
		  }
		sec += 60L * min + 3600L * hour + 86400L * day +
			(86400L * 365L) * year;
#ifdef VMS	/* VMS doesn't understand about daylight savings */
		if (91 <= day && day <= 273) sec += 3600;
#endif
		strcpy(rundate, ctime(&sec));
		printf("%s", rundate);
		printf("no. of pixels > 0, >254, > 64k = %8d, %5d, %2d\n",
		 (int)heads->nonzpix, (int)heads->oflowcount,
		 (int)heads->over64k);
		printf("frameno = %5d; filename = %s\n",
			(int)heads->frameno, heads->filename);
		if (NULL != (runhdp = heads->runheader) &&
		 (FUZZ < runhdp->run_d0pos[0]))
		  {
			printf("Crystal/run codename: %s\n", runhdp->run_code);
			printf("Comments on run: %s\n", runhdp->run_notes);
			printf("Generator at %3d kV, %3d ma\n",
				runhdp->run_volxray, runhdp->run_curxray);
			if (fabs(runhdp->run_stepreal) > FUZZ) printf(
			 "Stepsize: %7.3f degrees\n", runhdp->run_stepreal);
			else printf("Stepsize: %3d / 96 degrees\n",
				runhdp->run_stepsize);
			printf("Number of frames in run: %3d\n",
				runhdp->run_nframes);
			printf("Approx. # of scaling groups: %3d\n",
				runhdp->run_nshifts);
			printf("Start omega, chi, phi: %7.2f, %7.2f, %7.2f",
			 runhdp->run_startpos[0], runhdp->run_startpos[1],
			 runhdp->run_startpos[2]);
			printf(";\tMotor moving is %d\n", runhdp->run_motor);
			printf("Exposure time: %7.2f seconds",
			 runhdp->run_expos);
			if (runhdp->run_lambda > FUZZ)
			 printf(";\tWavelength = %7.4f Angstroms\n",
				runhdp->run_lambda);
			else printf("\n");
			printf(
			 "Crystal-to-detector distance, angle = %9.4f, %7.2f\n",
			 runhdp->run_d0pos[0], runhdp->run_d0ang[0]);
			if (runhdp->run_d1pos[1] > FUZZ) printf(
			 "detector pixels: %7.4f mm in X, %7.4f mm in Y\n",
			 runhdp->run_d1pos[1], runhdp->run_d1pos[2]);
			printf("Resolution limit, ssqmax = %8.3f, %9.5f\n",
			 runhdp->run_reslim, runhdp->run_ssqmax);
		   }
		if (heads->ydimen < 100) heads->ydimen = heads->xdimen;
		if ((showhflags && SHF_OVERFLOWS) && (heads->oflowcount))
		  {
			ncols = disp_overlist(heads, is16bit);
			disp_top(heads, is16bit, ncols);
		  }
	     } 
	   }
	kill_mail(&heads, &nmbox);	globalkill();	return EOK;
}

/* end of show_head functionality */

/***********************************
	merge_fm: merge two or more XGEN images together
 syntax:
	mergefm inname1 inname2 [	] output_name
		or
	mergefm startfm endfm output_name
		so that to merge frames 13 through 15 together
		and call the output test.13_15 we would say
	mergefm 13 15 test.13_15
		Needless to say, this second syntax requires that the
		FRAMES index must be present.
*************************************/

int mergefmmain(int argc, char *argv[])
{ /* The mainline for merging 2 or more images together */
	register int	nin, off, x, y;
	int		frame_max, frame_min, hit, hm1, imerr, imflags;
	int		n16, nmbox, numeric_format, spectype;
	int		tempflags, temprhead, typeno, wid;
	unsigned	imsiz, novers;
	u_char		*ucp;
	u_short		curval, *sp;
	xglong		*xlp;
	xgulong		filetype, maxospace;
	char		**dcp, **flags, lfname[81];
	double		oevent, outlate, revent, rlate, totstep;
	MAIL		*heads;
	FRAMEINDEX	*lfmi;
	VOFLOW		*vofp;
	DATARUN		*drp;
	static MAIL	*outmail;
	static char	*shortname;
	static xglong	*xgdata;
	static int	idarkfrac = 6400;
	static char	*fnames[] = {"frames", NULL};
	static char	ftype[] = {1, 0};
	static xgulong	ptype[] =
	 { FMT_PCS_8, FMT_MAR, FMT_SAXII, FMT_VMS, 0 } ;
	static char	*ctype[] = { "PCS", "MAR", "SAXII","VMS","UNLABELED" };

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	imflags = 0;	spectype = typeno = 0;
	flags = flagproc(argc, argv, "MPSVX");
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	   switch (**dcp)
	     {
		case	'M':	typeno = 1;	break;
		case	'P':	spectype = 1;	typeno = 0;	break;
		case	'V':	typeno = 3;	break;
		case	'X':	typeno = 2;	break;
		case	'S':	imflags |= DARK_CURRENT;
			if ('\0' != *((*dcp) + 1))
				idarkfrac = 8000. * atof((*dcp) + 1);
			break;
	     }
	filetype = ptype[typeno];
#ifdef LITTLEND
	filetype = swawb(filetype);
#endif
	if (typeno) spectype = 1;	/* clearly we've specified it */
	argc--;	argv++;
	if (argc > 0)
	  {
		while (**argv == '-')
		  {
			argc--;	argv++;
		  }
	  }
	if (argc < 3)
	  {
		fprintf(stderr,
		 "Usage: %s [-psv] infile1 infile2 [	] outfile\n",
			xgprogname(NULL, 0));	return EOK;
	  }
	if (idarkfrac > 16000) idarkfrac = 16000;
	if (imflags & DARK_CURRENT) imflags |= (idarkfrac << 16);
	nmbox = 2;
	heads = (MAIL *)callop(nmbox, sizeof (MAIL), "heads in mergefm");
	frame_min = atoi(*argv);	frame_max = atoi(*(argv + 1));
	numeric_format = ((0 < frame_min) && (0 < frame_max));
	lfmi = NULL;
	if (numeric_format)
	  {
		header(stdout, PROCESSING, fnames, ftype);
		lfmi = ftables(3);
		printf("Input frames %3d through %3d will be merged\n",
			atoi(*argv), atoi(*(argv + 1)));
		if (!spectype) strcpy(lfname, indexname(frame_min));
	  }
	else
	  {
		header(stdout, PROCESSING, (char **)NULL, (char *)NULL);
		printf("Input frame files:\n");
		frame_min = 0;	frame_max = argc - 2;
		for (nin = frame_min; nin <= frame_max; nin++)
			printf("\t%s\n", argv[nin]);
		if (!spectype) strcpy(lfname, argv[frame_min]);
	  }
	if (!spectype) /* determine output file type from input */
	  {
		tempflags = imflags | JUST_HEADER;
		if (FRAME_OK == read_image(lfname, tempflags, heads))
			filetype = heads->format;
		for (off = 0; ptype[off] != 0; off++)
		   if ((filetype == ptype[off]) || (filetype-1 == ptype[off]))
		     {
			typeno = off;	break;
		     }
		if ((ptype[off] == 0) && ((filetype == FMT_SAXII_16) ||
			 (swawb(filetype) == FMT_SAXII_16))) typeno = 2;
	  }
	if (imflags & DARK_CURRENT) printf(
 " %6.4f of the darkcurrent image %s will be subtracted from each input\n",
		((double)idarkfrac) / 8000., trname("DARKIMAGE"));
	printf("Output frame file:\t%s; written in %s format\n",
		argv[argc-1], ctype[typeno]);
	outmail = heads + 1;
/* read in the individual images one at a time and accumulate the
	results appropriately */
	oevent = outlate = imsiz = 0;
	for (totstep = 0., nin = frame_min; nin <= frame_max; nin++)
	   {
		if (numeric_format)	strcpy(lfname, indexname(nin));
		else			strcpy(lfname, argv[nin]);
		if (FRAME_OK != (imerr =
		 read_image(lfname, imflags, heads)))
			(void)image_error(stdout, lfname, imerr, 0);
		else {
			revent = heads->eventcount;
			if (revent < 0.) revent *= -64.;
			rlate = heads->late; if (rlate < 0.) rlate *= -64.;
			printf("Frame %s: %11.0f counts, %3d overflows\n",
			 heads->filename, revent, (int)heads->oflowcount);
			wid = heads->xdimen;	hit = heads->ydimen;
			if (nin != frame_min)
			  {
				oevent += revent;	outlate += rlate;
				outmail->elap100ths += heads->elap100ths;
			  }
			else {
				cpymem((AVOID *)outmail, (AVOID *)heads,
					sizeof (MAIL));
				oevent = revent;	outlate = rlate;
			     }
			imsiz = wid * hit;
			if (NULL == xgdata)
			  {
				if (NULL == (AVOID *)(xgdata = (xglong *)
				 callop(imsiz, sizeof (xglong),
				 "xgdata in mergefm"))) return EALLOC;
			  }
			for (x = 0, xlp = xgdata; x < wid; x++)
			   for (y = 0; y < hit; y++, xlp++)
			      {
				curval = imvalue(heads, x, y);
				*xlp += curval;
			      }
			if ((drp = heads->runheader))
			  {
				totstep += drp->run_stepreal;
			  }
		     }
	   }
	if (imsiz <= 0) return EBADVALUE;
	/* see whether it makes sense to output in 8-bit format */
	maxospace = imsiz / 2;
	for (x = imsiz, xlp = xgdata, n16 = novers = 0; x--; xlp++)
		if (*xlp > 255)
		  {
			novers++;	if (*xlp > 65535) n16++;
		  }
	memzap((AVOID **)&(outmail->overflow), "overflows in mergefm");
	printf(" %8d pixels have more than 255 counts; ", novers);
	printf(" %6d have more than 65535\n", n16);
	wid = outmail->xdimen;	hit = outmail->ydimen;
	hm1 = hit - 1;
	outmail->zmax = outmail->zmaxx = outmail->zmaxy = 0;
	/* write out in 16-bit format */
	if ((typeno == 1) || (novers * sizeof (VOFLOW) > maxospace) || n16)
	  {
		if (typeno == 2)
		  {
			if (filetype == FMT_SAXII) filetype = FMT_SAXII_16; 
			else if (filetype == swawb(FMT_SAXII))
				filetype = swawb(FMT_SAXII_16);
		  }
		memzap((AVOID **)&(outmail->data8), "data8 in mergefm");
		if (NULL == outmail->data16)
		  {
			if (NULL == (outmail->data16 = (u_short *)
			 callop((size_t)imsiz, sizeof (u_short),
			 "outmail->data16 in mergefm"))) return EALLOC;
		  }
		if (n16)
		  {
			if (NULL == (outmail->overflow = (VOFLOW *)
			 callop((size_t)n16, sizeof (VOFLOW),
			 "outmail->overflow in mergefm"))) return EALLOC;
			vofp = outmail->overflow;
			for (off = 0, xlp = xgdata, sp = outmail->data16;
			 off < (int)imsiz; off++, xlp++, sp++)
			   {
				if (*xlp < 65535) *sp = *xlp;
				else {
					*sp = 65535;
					x = off / wid;
					y = hm1 - off % wid;
					if (*xlp > (xglong)(outmail->zmax))
					  {
						outmail->zmax = *xlp;
						outmail->zmaxx = x;
						outmail->zmaxy = y;
					  }
					if (*xlp < 327679L) vofp->value =
					 (*xlp - 65535) / 4;
					else vofp->value = 65535;
					vofp->xcoord = x;
					vofp->ycoord = y;	vofp++;
				     }
			   }
			novers = n16;
			vofp = outmail->overflow;
		  }
		else {
			for (off = 0, xlp = xgdata, sp = outmail->data16;
			 off < (int)imsiz; off++, xlp++, sp++)
			  {
				*sp = *xlp;
				if (*xlp > (xglong)(outmail->zmax))
				  {
					x = off / wid;
					y = hm1 - off % wid;
					outmail->zmax = *xlp;
					outmail->zmaxx = x;
					outmail->zmaxy = y;
				  }
			   }
			outmail->overflow = NULL;	novers = 0;
		     }
		printf(" Output: %6d 16-bit overflows; ", novers);
		printf(" zmax = %7d at [%4d,%4d]\n",
		 (int)(outmail->zmax), outmail->zmaxx, outmail->zmaxy);
		outmail->oflowcount = n16;
	  }
	else {	/* write out in 8-bit format */
		memzap((AVOID **)&(outmail->data16), "data16 in mergefm");
		if (NULL == outmail->data8)
		  {
			if (NULL == (outmail->data8 = (u_char *)
			 callop((size_t)imsiz, sizeof (u_char),
			 "outmail->data8 in mergefm"))) return EALLOC;
		  }
		if (novers)
		  {
			if (NULL == (outmail->overflow = (VOFLOW *)
			 callop((size_t)novers, sizeof (VOFLOW),
			 "outmail->overflow in mergefm"))) return EALLOC;
			vofp = outmail->overflow;
			for (off = 0, xlp = xgdata, ucp = outmail->data8;
			 off < (int)imsiz; off++, xlp++, ucp++)
			   {
				if (*xlp < 255) *ucp = *xlp;
				else {
					*ucp = 255;
					x = off / hit; y = hm1 - off % wid;
					if (*xlp > (xglong)(outmail->zmax))
					  {
						outmail->zmax = *xlp;
						outmail->zmaxx = x;
						outmail->zmaxy = y;
					  }
					vofp->value = *xlp;
					vofp->xcoord = x; vofp->ycoord = y;
					vofp++;
				     }
			   }
		  }
		else {
			for (off = 0, xlp = xgdata, ucp = outmail->data8;
			 off < (int)imsiz; off++, xlp++, ucp++)
			   {
				*ucp = *xlp;
				if (*xlp > (xglong)(outmail->zmax))
				  {
					x = off % hit;
					y = hm1 - off % wid;
					outmail->zmax = *xlp;
					outmail->zmaxx = x;
					outmail->zmaxy = y;
				  }
			   }
		     }
	     }
	if (novers > 1) qsort((AVOID *)outmail->overflow, novers,
	 sizeof (VOFLOW), (QPE)ofycmp); /* resort the mailbox */
	else	(outmail->zmaxy)++;
	outmail->oflowcount = novers;
	outmail->eventcount = (oevent > 2.147483E9) ? -oevent / 64. : oevent;
	outmail->late = (outlate > 2.147483E9) ? -outlate / 64. : outlate;
	outmail->ontime = outmail->eventcount - outmail->late;
	shortname = filepart(argv[argc - 1], 0);
	strncpy(outmail->filename, shortname, 15);
	if (NULL == outmail->runheader)
	  {
		outmail->runheader = (DATARUN *)callop(1, sizeof (DATARUN),
		 "outmail->runheader in mergefm");
		temprhead = 1;
	  }
	temprhead = 0;
	if (NULL != (drp = outmail->runheader))
	  {
		drp->run_stepreal = totstep;
		drp->run_stepsize = (int)(totstep * 96.);
	  }
	output_frame(argv[argc-1], outmail, filetype, 0x200);
	printf(" Frame %s:%3d overflows,%9.0f counts for %d seconds\n",
		outmail->filename, (int)(outmail->oflowcount), oevent,
		(int)(outmail->elap100ths / 100));
	printf("Brightest pixel: %6d at [%4d,%4d]\n",
		(int)outmail->zmax, (int)outmail->zmaxx,
		(int)(hit - outmail->zmaxy - 1));
	if (temprhead) memzap((AVOID **)&drp, "outmail->runheader in mergefm");
	kill_mail(&heads, &nmbox);
	globalkill();	return EOK;
}

/* end of merge_fm functionality */

/**********************************
	imformat: convert images in any format to selected format
		see comments within "imformatmain" for details
***********************************/

static void sumimformin(FILE *fp, int fmtno, IMWORK *imp)
{ /*	This summarizes the run of 'imformat' */
	static char	*flist[] = {"frames", "uspline", "xlog", NULL } ;
	static char	ftypes[] = { 1, 1, 2, 0 };
	static char	*oformat[] =
	{ "pcs", "mar", "saxii8", "saxii16", "R-Axis", "smv" } ;

	ftypes[0] = (imp->imf_fpl == stdout ? 0 : 2);
	header(fp, PROCESSING, flist, ftypes);
	fprintf(fp, "Images will be output in %s format\n", oformat[fmtno]);
	if (imp->imf_flags & OVERRIDE_COORDS) fprintf(fp,
 "Image coordinates will be output according to geometric convention %1d\n",
		(imp->imf_flags & 0x80) >> 5);
	fprintf(fp, "Images will%sbe spatially corrected before output\n",
		imp->imf_flags & DO_SPATIAL ? " " : " NOT ");
	fprintf(fp, "Input images will%s be deleted before output\n",
		imp->imf_flags & DELETE_ORIGINAL? " " : " NOT ");
	fprintf(fp, "Results will be presented with %s verbosity\n",
		((imp->imf_flags & VERY_VERBOSE) ? "substantial" :
		(imp->imf_flags & PARTIAL_VERBOSE) ? "moderate" : "no"));
	if (imp->imf_outdist > -1.e7) fprintf(fp,
	 "Sample-to-detector distance will be set to %9.3f cm on output\n",
		imp->imf_outdist);
	if (imp->imf_out2th > -1.e7) fprintf(fp,
	 "detector 2theta will be set to %9.3f degrees on output\n",
		imp->imf_out2th);
	if ((imp->imf_outwid > 0) || (imp->imf_outht > 0))
	  fprintf(fp, "Images will be redimensioned to %d-by-%d\n",
		imp->imf_outwid, imp->imf_outht);
	if (imp->imf_flags & PAD_TO_2TO_N) fprintf(fp,
	 "Each row of the output will be padded to the next power-of-2\n");
	if (imp->imf_flags & RESCALE65K) fprintf(fp,
	 "If Imax > 65K, we will rescale so the brightest is 65K\n");
}

static int imcvterr(FILE *fp, IMWORK *imp, int retv, char *source)
{ /* This returns an appropriate error value after emitting a
	relevant error message after failures in do_conversion */
	fprintf(fp, "Failure %d in %s\n", retv, source);
	if (EALLOC == retv) return FRAME_NO_MEMORY;
	return (imp->imf_heads)->data8 ? FRAME_BAD8BIT : FRAME_BAD16BIT;
}

static int imf_floodcorr(IMWORK *imp)
{ /* This performs a flood-field correction on the image referenced
   by the IMWORK structure *imp. Unlike the flood-field correction
   performed in `multimage', this one assumes that the scale of the
   flood-image and the output data image to which the correction
   will be applied are identical */
	u_short		*usp;
	int		flood_flags, fr, hit, locmx, locmy, locv;
	int		maxlv, minlv, wid, x, y;
	float		*flcp;
	double		dlocv, meanall, meancol, overscale, sumcol;
	static char	*fcname;
	static size_t	sizf = 0;
	static float	*flc;
	static double	maxover, minover;
	static MAIL	*fhd, *lhd;
	static char	flcmsg[] = "flc in imf_floodcorr";
	static char	fhdmsg[] = "fhd in imf_floodcorr";

	if (NULL == imp)
	  { /* special call to flush memory */
		memzap((AVOID **)&flc, flcmsg);
		memzap((AVOID *)&fhd, fhdmsg);	return EOK;
	  }
	if (!(imp->imf_flags & FLOODCORR)) return EOK;
	lhd = imp->imf_heads;
	if (!(lhd->data16))
	  {
		fprintf(stderr,
		 "Sorry: flood not supported in 8-bit formats\n");
		return EBADVALUE;
	  }
	wid = lhd->xdimen;	hit = lhd->ydimen;
	/* in some cases we're calling this routine after we've
	 already read in the correction itself. */
	if (NULL == flc)
	  {
		sizf = wid * hit;
		if (NULL == (flc = (float *)callop(sizf,
		 sizeof (float), flcmsg))) return EALLOC;
		if (NULL == (fhd = (MAIL *)callop(1,
		 sizeof (MAIL), fhdmsg))) return EALLOC;
		fcname = trname("flood");
		flood_flags = GET_RUNHEADER;
		fprintf(stdout, "Reading %s as flood-field:\n", fcname);
		if (FRAME_OK != (fr = read_image(fcname,
		 flood_flags, fhd)))
		  {
			(void)image_error(stdout, fcname, fr, 0);
			return EREADERR;
		  }
		if ((wid != fhd->xdimen) || (hit != fhd->ydimen))
		  {
			fprintf(stderr,
 "Image size inconsistency: data are [%4d,%4d], flood [%4d,%4d]\n",
				wid, hit, fhd->xdimen, fhd->ydimen);
			return EBADVALUE;
		  }
		maxlv = -1073741824;	minlv = -maxlv;
		for (flcp = flc, meanall = x = 0; x < wid; x++)
		   {
			 for (sumcol = y = 0; y < hit; y++, flcp++)
			   {
				locv = imvalue(fhd, x, y);
				if (locv > maxlv) maxlv = locv;
				if (locv < minlv) minlv = locv;
				*flcp = locv;
				sumcol += *flcp;
			   }
			meancol = sumcol / (double)hit;
			meanall += meancol;
		   }
		meanall /= (double)wid;
		fprintf(stdout, "Mean, min, max = %12.4e, %7d, %7d\n",
			meanall, minlv, maxlv);
		if ((maxlv < 1) || (meanall < DFUZZ))
		  {
			fprintf(stderr,
			 "Insufficient positive values in flood\n");
			return EBADVALUE;
		  }
		minover = 1.e9;
		for (flcp = flc, maxover = x = 0; x < wid; x++)
		 for (y = 0; y < hit; y++, flcp++)
		    {
			*flcp = (*flcp < 0.1) ? 0 :
				flimits(meanall / *flcp, OVERMIN, OVERMAX);
			/* experiment: if *flcp > OVERMAX, then the point
			 is probably in a dead region anyway */
			if (*flcp > OVERMAX - 1.) *flcp = 1.;
			if (*flcp > maxover) maxover = *flcp;
			if (*flcp < minover) minover = *flcp;
		   }
		fprintf(stdout, "Min, Max correction value = %12.4e, %12.4e\n",
			minover, maxover);
	  }
	/* now we apply these corrections to the actual image */
	maxover = 0.;	minover = 1.e9;
	locmx = locmy = maxlv = -1;
	for (x = 0, usp = lhd->data16, flcp = flc; x < wid; x++)
	   {
		for (y = 0; y < hit; y++, usp++, flcp++)
		   {
			dlocv = ((double)*usp) * *flcp;
			if (dlocv > maxover)
			  {
				maxover = dlocv;
				locmx = x;	locmy = y;
			  }
			if (dlocv < minover) minover = dlocv;
		   }
	   }
	fprintf(stdout, "Apparent min corrected value = %12.4e\n",
		minover);
	fprintf(stdout,
	 "Apparent max corrected value = %12.4e @ [%4d,%4d]\n",
		maxover, locmx, locmy);
	if (maxover > 65534.)
	  {
		overscale = 65534. / maxover;
		fprintf(stdout, "All values rescaled by %12.4e\n",
			overscale);
	  }
	else	overscale = 1.;
	locmx = locmy = maxlv = -1;
	for (x = 0, usp = lhd->data16, flcp = flc; x < wid; x++)
	   {
		for (y = 0; y < hit; y++, usp++, flcp++)
		   {
			dlocv = *usp;
			dlocv *= overscale * *flcp;
			*usp = (dlocv + 0.49999);
			if (*usp > maxlv)
			  {
				maxlv = dlocv; locmx = x;	locmy = y;
			  }
		   }
	   }
	/* correct some header elements */
	lhd->zmax = maxlv;	lhd->zmaxx = locmx;	lhd->zmaxy = locmy;
	fprintf(stdout, "Zmax = %6d at [%4d,%4d]\n", maxlv, locmx, locmy);
	return EOK;
}

static int do_conversion(char *finame, IMWORK *imp)
{ /*	This does the conversion on a single image */
	int	idfrac, imflags, imerr, retv;
	DATARUN	*drp;

	if (NULL == imp->imf_heads)
	  {
		imp->imf_nmbox = 1;
		if (NULL == (AVOID *)(imp->imf_heads =
		 (MAIL *)callop(imp->imf_nmbox, sizeof (MAIL),
		 "heads in imformat"))) return EALLOC;
	  }
	/* open logfile and put a summary of the files there and on stdout */
	if ((NULL == imp->imf_fpl) || (imp->imf_fpl == stdout))
	  {
		if (NULL == (imp->imf_fpl = propen("xlog")))
			imp->imf_fpl = stdout;
	  }
	imflags = WAIT_IF_ABSENT | GET_RUNHEADER;
	if (imp->imf_flags & VERY_VERBOSE)
		printf(" Examining image %s:\n", finame);
	if (imp->imf_flags & DARKSUB)
	  {
		idfrac = (int)(imp->imf_darkscale * 8000. + 0.49999);
		imflags |= DARK_CURRENT | GRIPE_IF_SHORT | (idfrac << 16);
	  }
	if (FRAME_OK != ( imerr = read_image(finame, imflags, imp->imf_heads)))
	  {
		(void)image_error(stdout, finame, imerr, 0);
		if (imp->imf_fpl != stdout)
			(void)image_error(imp->imf_fpl,  finame, imerr, 0);
		return imerr;
	  }
	if (NULL != (drp = ((imp->imf_heads)->runheader)))
	  {
		if (imp->imf_outdist > -1.e7)
			drp->run_d0pos[0] = imp->imf_outdist;
		if (imp->imf_out2th > -1.e7)
			drp->run_d0ang[0] = imp->imf_out2th;
	  }
	if ((imp->imf_flags & DO_SPATIAL) &&
	 (EOK != (retv = spatial(imp->imf_outwid, imp->imf_outht,
	 imp->imf_flags & 0x3080, imp->imf_heads))))
		return imcvterr(stderr, imp, retv, "spatial conversion");
	if ((imp->imf_flags & FLOODCORR) &&
	 (EOK != (retv = imf_floodcorr(imp))))
		return imcvterr(stderr, imp, retv, "flood correction");
	output_frame(finame, imp->imf_heads,
		imp->imf_outtype, imp->imf_flags);
	return FRAME_OK;
}

static void summar_image(FILE *fp, char *finame, IMWORK *imp)
{ /*	This summarizes the contents of the converted image to stream fp */
	double	realcounts;
	MAIL	*lhd;

	lhd = imp->imf_heads;
	realcounts = (lhd->eventcount > 0) ?  lhd->eventcount :
		64. * ((double)(-(lhd->eventcount)));
	fprintf(fp,
		 " %s: %12.0f counts, %5d overflows; zmax = %7d at [%4d,%4d]\n",
			finame, realcounts, (int)(lhd->oflowcount),
			(int)(lhd->zmax), lhd->zmaxx,
			(int)(lhd->ydimen - 1 - lhd->zmaxy));
}

static int im_extension(char *fnam)
{ /*	This returns true if fnam's file extension (the stuff beyond the last
	dot) is either allnumeric, "img", "image", "osc", "ipf", "stl",
	"smv", or "_xy" where x and y are any letters */
	int	i;
	char	*cp, *cend, *af, *af1, *af2;
	static char *legal[] = {
		"img", "image", "osc", "stl", "ipf", "smv", NULL };

	for (cp = fnam; '\0' != *cp; cp++);	cend = cp - 1;
	if (cend <= fnam) return 0;
	for (af = NULL, cp = cend; cp >= fnam; cp--)
		if (*cp == '.')
		  {
			af = cp + 1; break;
		  }
	if ((NULL == af) || ('\0' == *af)) return 0;
	if (allnumeric(af)) return 1;
	for (i = 0; (NULL != legal[i]) && ('\0' != *legal[i]); i++)
		if (!strcmp(af, legal[i])) return 1;
	if (strlen(af) != 3) return 0;	af1 = af + 1;	af2 = af + 2;
	if ((*af == '_') && ('A' <= *af1) && (*af1 <= 'z') &&
	 ('A' <= *af2) && (*af2 <= 'z')) return 1;
	return 0;
}

static int dir_search(IMWORK *imp)
{ /*	This goes through the current directory and picks out files that have
	all-numeric file extensions or extensions "img", "image", "osc",
	"stl", "ipf", "smv", or "_xy" where x and y are any letters.
	It attempts conversions on these and writes the files out. */
	int			retv;
	DIR			*drp;
	char			lbu[256];
	static struct dirent	*mdire;

	fprintf(stderr,
 "All recognizably-named images in the current directory will be converted\n");
	if (NULL == (drp = opendir("."))) return EOPENERR;
	retv = 0;
	while (NULL != (mdire = readdir(drp)))
	     {
		if (0 != mdire->d_ino)
		  {
			strncpy(lbu, mdire->d_name, mdire->d_reclen);
			if (im_extension(lbu))
			  {
				retv = do_conversion(lbu, imp);
				if (imp->imf_flags & PARTIAL_VERBOSE)
				  {
					if (imp->imf_fpl != stdout)
					  summar_image(imp->imf_fpl, lbu, imp);
					summar_image(stdout, lbu, imp);
				  }
			  }
		  }
	     }
	return retv;
}

static int ft_search(IMWORK *imp)
{ /*	This goes through the current directory and picks out files that have
	all-numeric file extensions or extensions "img", "image", "osc",
	"stl", "ipf", "smv", or "_xy" where x and y are any characters.
	It attempts conversions on these and writes the files out.
	The routine returns the success code associated with the last
	conversion it performs */
	int		retv;
	char		*tname;
	FRAMEINDEX	*lfmi;
	FRAMEENTRY	*lft;

	lfmi = imp->imf_lfmi;
	for (retv = 0, lft = lfmi->fi_fmi; lft <= lfmi->fi_fme; lft++)
	   {
		if ((NULL != (tname = lft->fmname)) && ('\0' != *tname) &&
		 (0 == xgaccess(tname, 4, 1)))
		  {
			retv = do_conversion(tname, imp);
			if (imp->imf_flags & PARTIAL_VERBOSE)
			  {
				if (imp->imf_fpl != stdout)
					summar_image(imp->imf_fpl, tname, imp);
				summar_image(stdout, tname, imp);
			  }
		  }
	   }
	return retv;
}

int imformatmain(int argc, char *argv[])
{ /* Mainline for converting detector images from one format to another */
	int		foundsome, fmtnum, specoutw, specouth, resu;
	char		*cpname, **flags, **dcp;
	double		nval;
	static IMWORK	imfs;
	static xgulong	fotype[] =
	 { FMT_PCS_8, FMT_MAR, FMT_SAXII, FMT_SAXII_16, FMT_RAXIS, FMT_FAST } ;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	cpname = xgprogname(NULL, 0);
	if (argc <  1)
	  {
		fprintf(stderr,
		 "Usage: %s [-pmbsrud] [-x<val>] [-y<val>] [filename]\n",
			cpname);
		return EOK;
	  }
	imfs.imf_flags = DO_SPATIAL | DELETE_ORIGINAL | PARTIAL_VERBOSE;
	if (!strcmp(cpname, "topcs")) imfs.imf_flags |= TO_PCS;
	else if (!strcmp(cpname, "tomar")) imfs.imf_flags |= TO_MAR;
	else if (!strcmp(cpname, "tobruker")) imfs.imf_flags |= TO_BRUKER;
	else if (!strcmp(cpname, "tobruker16")) imfs.imf_flags |= TO_BRUKER16;
	else if (!strcmp(cpname, "toraxis")) imfs.imf_flags |= TO_RAXIS;
	else if (!strcmp(cpname, "tosmv")) imfs.imf_flags |= TO_SMV;
	else imfs.imf_flags |= TO_MAR;
/* process flags, if any.  Currently defined flags are:
	-a	pad out each row of the array with
		enough zeroes to make the row a power of two.
		Thus if the row dimension is 1900, this will
		pad each row with zeroes from 1900 through 2047.
	-b	output should be in Bruker/Siemens 8-bit format
	-c<val> set the sample-to-detector distance value in header to <val>
	-d	do NOT delete the original before writing out converted file.	
	-e<val>	alter the coordinate conventions of the image.
		Ordinarily the coordinate conventions are chosen in
		the context of the type of output desired; this flag overrides
		those considerations. The <val> value is 0-7 as follows;
 <val>	coordinate convention			example
 	dimension varying	read begins
 	fastest			from
 0	Y			top left	PCS (little-endian)
 1	X			top left	Mar, SDMS
 2	Y			bottom left
 3	X			bottom left	MacScience, SAXII
 4	Y			top right
 5	X			top right	PCS (big-endian)
 6	Y			bottom right
 7	X			bottom right	R-Axis, Fuji
	In each case the definition of left and right assumes that the
	view is OF the sample FROM the detector, not the other 'way around.
	In the alternative view (from the sample toward the detector)
	<val> = 0 corresponds to Y fastest, reading from top right,
	<val> = 5 corresponds to X fastest, reading from top left, etc.
	-f	output should be in SMV format
		(note that "f" has little to do with SMV --
		none of the logical letters was still available)
	-g	during the spatial correction, the intensities should be
		rescaled so that the brightest one is less than 65535
	-h<val> Subtract a dark image before conversion. If <val>
		is specified, the dark image values are scaled by <val>
		before subtracting them. If not, the scale factor is 1.
	-j<val>	Divide the values in the image by values in a flood-field
		image. If <val> is specified, the flood-field image values
		are rescaled by <val>. If not, the scale factor is 1.
	-k	keep the spatial properties of the original. If this
		option is NOT set, the program will perform a spatial
		correction on the image before writing it out.
	-m	output should be in Mar Image-Plate format
	-p	output should be in PCS format
	-q	no verbosity at all
	-r	output should be in R-Axis format
	-s	output should be in Bruker/Siemens 16-bit format
	-t<val> set the two-theta value in the header to <val>
	-v	verbose mode: additional diagnostics describing the conversion
	-x<val>	re-set the X detector dimension to <val>
	-y<val>	re-set the Y detector dimension to <val>
	If -h, -j, and !-k are specified, then the order of application
	of the various filters is:
	 (1) dark subtraction
	 (2) spatial correction
	 (3) flood correction
	Accordingly, the dark image should be raw;
	the image used to create the spatial correction should be dark-corrected
	but not flood-corrected (and, obviously, not spatially corrected);
	and the flood correction should be both dark-corrected and spatially
	corrected.
*/
	flags = flagproc(argc, argv, "ABCDEFGHJKMPQRSTVXY");
	specouth = specoutw = -1; imfs.imf_out2th = imfs.imf_outdist = -1.e9;
	imfs.imf_darkscale = 1.;
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	 {
	   nval = atof((*dcp) + 1);
	   switch (**dcp)
	     {
		case	'A': imfs.imf_flags |= PAD_TO_2TO_N;	break;
		case	'B': imfs.imf_flags &= ~0x7F;
			imfs.imf_flags |= TO_BRUKER; break;
		case	'C': imfs.imf_outdist = nval;	break;
		case	'D': imfs.imf_flags &= ~(DELETE_ORIGINAL); break;
		case	'E': imfs.imf_flags |= (OVERRIDE_COORDS |
				(((int)nval) << 5));	break;
		case	'F': imfs.imf_flags &= ~0x7F;
			imfs.imf_flags |= TO_SMV;	break;
		case	'G': imfs.imf_flags |= RESCALE65K;	break;
		case	'H': imfs.imf_flags |= DARKSUB;
			if ((0. < nval) && (nval < 1.e10))
				imfs.imf_darkscale = nval;	break;
		case	'J': imfs.imf_flags |= FLOODCORR;	break;
		case	'K': imfs.imf_flags &= ~(DO_SPATIAL);	break;
		case	'M': imfs.imf_flags &= ~0x7F;
			imfs.imf_flags |= TO_MAR; break;
		case	'P': imfs.imf_flags &= ~0x7F;
			imfs.imf_flags |= TO_PCS; break;
		case	'Q': imfs.imf_flags &= ~(PARTIAL_VERBOSE |
		 VERY_VERBOSE); break;
		case	'R': imfs.imf_flags &= ~0x7F;
			imfs.imf_flags |= TO_RAXIS; break;
		case	'S': imfs.imf_flags &= ~0x7F;
			imfs.imf_flags |= TO_BRUKER16; break;
		case	'T': imfs.imf_out2th = nval;	break;
		case	'V': imfs.imf_flags |= PARTIAL_VERBOSE | VERY_VERBOSE;
			break;
		case	'X': specoutw = nval;	break;
		case	'Y': specouth = nval;	break;
	     }
	 }
	if (NULL == (imfs.imf_fpl = propen("xlog"))) imfs.imf_fpl = stdout;
	/* the lowest-order 5 bits in the "imfs.imf_flags" variable constitute
	 the identifier for the output data type. If we try to do more
	 than 31 output formats, we're in deep yoghurt. */
	fmtnum = (imfs.imf_flags & 0x1F) - 1;
	if (fmtnum >= 6) fmtnum = 4; /* this will change as more fmts occur */
	if ((specoutw > 0) || (specouth > 0))
	  {
		imfs.imf_outwid = specoutw; imfs.imf_outht  = specouth;
		if (specoutw <= 0) imfs.imf_outwid = imfs.imf_outht;
		else if (specouth <= 0) imfs.imf_outht = imfs.imf_outwid;
	  }
	else	imfs.imf_outwid = imfs.imf_outht = 0;
	sumimformin(stdout, fmtnum, &imfs);
	if (imfs.imf_fpl != stdout) sumimformin(imfs.imf_fpl, fmtnum, &imfs);
	imfs.imf_outtype = fotype[fmtnum];
#ifdef LITTLEND
	imfs.imf_outtype = swawb(imfs.imf_outtype);
#endif
	/* the choice of which images to convert is made in the following way:
	 If filenames are specified, then only those files will be converted.
	 If the environment variable FRAMES is defined, then only the images
	 named in the file pointed to by $FRAMES will be converted.
	 If $FRAMES is undefined, all files found in the current directory with
	 file extensions that are either numeric or consist of "img", "image",
	 "osc", "stl", "ipf", or "smv" are converted.
	*/
	for (resu = foundsome = 0, argc--, argv++; argc > 0; argc--, argv++)
	   if ((**argv != '-') && (0 == xgaccess(*argv, 4, 1)))
	     {
		foundsome = 1;
		resu = do_conversion(*argv, &imfs);
		if (imfs.imf_flags & PARTIAL_VERBOSE)
		  {
			if (imfs.imf_fpl != stdout)
				summar_image(imfs.imf_fpl, *argv, &imfs);
			summar_image(stdout, *argv, &imfs);
		  }
	     }
	if (!foundsome)
	  {
		if (NULL == (imfs.imf_lfmi = ftables(3)))
			resu = dir_search(&imfs);
		else resu = ft_search(&imfs);
	  }
	if (stdout != imfs.imf_fpl) (void)fclose(imfs.imf_fpl);
	imfs.imf_fpl = NULL;
	resu = 1;	kill_mail(&(imfs.imf_heads), &resu);
	(void)ftables(0); kill_mail(&(imfs.imf_heads), &(imfs.imf_nmbox));
	globalkill();	return resu;
}

/* end of imformat.c */
