/* maskutils.c -- utilities to merge or edit model profiles

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	24 Jun 2006: include xgmodel.h; that's where the structures live
	19 May 2006: merge mergemask.c, cleanmask.c into this file
	26 Oct 2005: xgbuf() calls replace str2xgbuf() and flush_xgbuf() calls
	12 Oct 2005: eliminate global variables and maxpop[]; housekeeping
	11 May 2005: check return status from xgjournal
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	28 Aug 2004: routines used only here are declared as static; cosmetics
	26 Jan 2004: convert some static global variables into locals
	21 Jan 2004: license message inserted and initial comments reordered
	 8 Jan 2004: static globals and call to assign_pname replace extern
		globals; callop() and memzap() replace shcalloc() and shfree()
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	17 Apr 2003: change "frameno" to "fmno" to distinguish it from
		MAIL structure element
	 5 Jan 2003: reorganized source-file header
	 1 Nov 2002: use of REGIONPROFILE structures for
	 reflection-box sizes; enables more general formats to be handled
	20 Sep 2002: cleanmask: CLF_* definitions moved to ../include/xgen.h;
		ditto LORENZIAN, GAUSSIAN, MASKRESCALE;
		trivial reformatting of diagnostic output
	25 Mar 2002: removed MSI attachments
	31 May 1995: "register" -> "register int" throughout
	 2 Mar 1995: diagnostics to stdout; shared-memory alloc and free
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 5 Oct 1994: use xgread and xgwrite on MASK_3D data
	 5 May 1994: source-file headers; POLYDET->XGEN;
		more external function declarations
	16 Mar 1994: mergemask: accommodate new "REGIONPROFILE" structures
	16 Mar 1994: cleanmask: major rewrite: facilities to ../xglib/xgenmask;
		algorithms redesigned
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version of mergemask and cleanmask
	26 Sep 1988: cleanmask: permit Lorentzian to have variable exponent
	15 Sep 1988: cleanmask: generalize ellipsoid and clean it all up.
	26 Nov 1987: cleanmask: add in adjustable centroid.
	25 Nov 1987: original code for cleanmask
   Purpose:	manipulate 3-D profiles (WMASK) files
	mergemask	produces a weighted average of the model profiles
		contained in two or more files of type MASK_3D.
	cleanmask	automated smoothing & recentring of 3-D model profiles
   Cross references: ../xglib/xgenmask.c,
	../integrate/integmask.c, ../displays/xgmodel.c

*************************************************************************/
#define		NEED_UNISTD	1
#include	"xgmodel.h"

#define		DEF_CONTOUR	0.010	/*1% contour level by default*/
#define		HYPERSECANT	2	/* sech(x) ellipsoids */
#define		ARCCOT		3	/* arccot(x) ellipsoids */
#define		DEF_LORENZ_POW	(2.5)	/* default lorenzian power */
#define		NEWT_CONVERGED	(0.1e-5) /* convergence test, Newton's method*/
#define		MAX_TRIES	(100)	/* max cycles of Newton's method */
#define		TWOOPI		(2. / PI)	/* used in arctan calcs. */

 typedef struct {
		FILE		*mrm_fmaski, *mrm_fmasko;
		MASK_3D		*mrm_wmask;
		REGIONPROFILE	*mrm_regb;
		} MERMWORK;

 typedef struct {
		int		cm_flags;
		char		cm_elltype, cm_locjb[82];
		double		cm_mincontour, cm_lorpow;
		double		cm_maxmo[NREGIONS];
		double		cm_ellipspars[NREGIONS][13];
		FILE		*cm_fpi, *cm_fpl;
		MASK_3D		*cm_wmin;
		REGIONPROFILE	*cm_regb;
		pid_t		cm_currpid;
		} CLEANMWORK;

extern int cleanmmain(int argc, char *argv[]);
extern int mergemmain(int argc, char *argv[]);

/****************************************
	mergemask:	merge together two or more 3-D profiles files
	to form a single weighted average of the inputs
****************************************/
static int endmerm(int irv, MERMWORK *mmp)
{ /*	finishing-up operations in `mergemask' */
	if (NULL != (mmp->mrm_fmaski)) (void)fclose(mmp->mrm_fmaski);
	if (NULL != (mmp->mrm_fmasko)) (void)fclose(mmp->mrm_fmasko);
	memzap((AVOID **)&(mmp->mrm_wmask), "mmp->mrm_wmask in mergemask");
	if (NULL != (mmp->mrm_regb))
	  {
		memzap((AVOID **)&((mmp->mrm_regb)->m3_xyop),
			"regb->m3_xyop in mergemask");
		memzap((AVOID **)&(mmp->mrm_regb), "regb in mergemask");
	  }
	mmp->mrm_fmaski = NULL;	mmp->mrm_fmaski = NULL;
	globalkill();	return irv;
}

static int numpos(float *flb, int abxy, int abz)
{ /*	This counts the number of frames that are nonzero among the seven
	frames in the REGIONPROFILE *rgp.  A frame is considered to
	be nonzero if the brightest element in it is at least 0.5% of the
	brightest element in the brightest frame */
	int	frame, nfmbr, y;
	double	overmax, tmax;
	float	*flp, fmmax[PFRAMES];

	for (overmax = frame = 0, flp = flb; frame < abz; frame++)
	   {
		for (tmax = y = 0; y < abxy; y++, flp++)
		   {
			if (tmax < *flp) tmax = *flp;
		   }
		fmmax[frame] = tmax;	if (overmax < tmax) overmax = tmax;
	   }
	if (overmax < FUZZ) return 0;
	nfmbr = 0;
	for (frame = 0, flp = &(fmmax[0]); frame < abz; frame++, flp++)
		if (*flp > 0.005 * overmax) nfmbr++;
	return nfmbr;
}

int mergemmain(int argc, char *argv[])
{ /* Mainline for merging multiple WMASK files together */
	register int	iz;
	int		bx, bxy, bxyz, by, bz;
	int		nnonzero, ntnz, region, totalboxes;
	float		*flp, *flpo;
	MASK_3D		*wmp;
	REGIONPROFILE	*rgp;
	static float	newf[2000];
	static MERMWORK	mms;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	if ((--argc) < 2)
	  {
		fprintf(stderr,
		 "Usage: mergemask maskfile0 maskfile1 [maskfile2]...\n");
		return EOK;
	  }
	bx = BIGSPOTWID;	by = BIGSPOTHT;	bz = PFRAMES;
	totalboxes = PFRAMES * NREGIONS;
	bxy = bx * by;	bxyz = bxy * bz;
	if (qerror((AVOID *)(mms.mrm_fmaski = xgopen(*(++argv), "r")),
		"open first input maskfile")) return EOPENERR;
	if (NULL == (AVOID *)(mms.mrm_wmask = (MASK_3D *)callop(NREGIONS,
	 sizeof (MASK_3D), "wmask in mergemask"))) return endmerm(EALLOC, &mms);
	if (NULL == (AVOID *)(mms.mrm_regb = (REGIONPROFILE *)callop(NREGIONS,
	 sizeof (REGIONPROFILE), "regb in mergemask")))
		return endmerm(EALLOC, &mms);
	if (NULL == (AVOID *)(flp = (float *)callop(NREGIONS * bxyz,
	 sizeof (float), "flp in mergemask")))
		return endmerm(EALLOC, &mms);
	for (region = nnonzero = 0, rgp = mms.mrm_regb,
	 wmp = mms.mrm_wmask; region < NREGIONS;
	 region++, wmp++, rgp++, flp += bxyz)
	   {
		if (0 >= xgread((void *)wmp, XG_MASK_3D, mms.mrm_fmaski))
			return endmerm(EREADERR, &mms);
		rgp->m3_bx = bx; rgp->m3_by = by; rgp->m3_bz = bz;
		rgp->m3_nobs = wmp->phielt[3].mnobs;
		rgp->m3_isqsum = wmp->phielt[3].isqsum;
		rgp->m3_xyop = flp;
		(void)wmptovec(wmp, rgp->m3_xyop);
		nnonzero += numpos(rgp->m3_xyop, bxy, bz);
	   }
	(void)fclose(mms.mrm_fmaski);	mms.mrm_fmaski = NULL;
	printf("%3d frames out of %3d in mask file %s are nonzero\n",
		nnonzero, totalboxes, trname(*argv));
	fprintf(stderr, "about to loop: argc = %d\n", argc);
	for (argv++; --argc; argv++)
	   {
		if (qerror((AVOID *)(mms.mrm_fmaski = xgopen(*argv, "r")),
			"open input maskfile")) return endmerm(EOPENERR, &mms);
		fprintf(stderr, "just opened %s\n", *argv);
		rgp = mms.mrm_regb;	wmp = mms.mrm_wmask;
		for (ntnz = region = 0; region < NREGIONS;
		 region++, rgp++, wmp++)
		   {
			fprintf(stderr, "about to read mask %2d\n", region);
			if (0 >= xgread((void *)wmp, XG_MASK_3D,
			 mms.mrm_fmaski)) return endmerm(EREADERR, &mms);
			fprintf(stderr, "about to call wmptovec\n");
			(void)wmptovec(wmp, newf);
			rgp->m3_nobs += wmp->phielt[3].mnobs;
			rgp->m3_isqsum += wmp->phielt[3].isqsum;
			fprintf(stderr, "about to call numpos\n");
			ntnz += numpos(newf, bxy, bz);
			fprintf(stderr, "ntnz = %d\n", ntnz);
			for (flp = newf, flpo = rgp->m3_xyop, iz = 0;
			 iz < bxyz; iz++, flp++, flpo++) *flpo += *flp;
			fprintf(stderr, "output flpos calculated\n");
			nnonzero += numpos(rgp->m3_xyop, bxy, bz);
			fprintf(stderr, "region = %2d: nnonzero = %8d\n",
			 region, nnonzero);
		   }
		(void)fclose(mms.mrm_fmaski);	mms.mrm_fmaski = NULL;
		fprintf(stderr, "just closed %s\n", *argv);
		printf("%3d frames out of %3d are nonzero in %s\n",
			ntnz, totalboxes, trname(*argv));
		printf(
	 "%3d frames out of %3d are now nonzero in the merged model in %s\n",
			nnonzero, totalboxes, trname(*argv));
	   }
	if (qerror((AVOID *)(mms.mrm_fmasko = xgopen("wmask", "w")),
		"open output maskfile")) return endmerm(EOPENERR, &mms);
	for (region = 0, rgp = mms.mrm_regb, wmp = mms.mrm_wmask;
	 region < NREGIONS; region++, rgp++, wmp++)
	   {
		(void)rgptowmp(rgp, wmp);
		if (0 >= xgwrite((void *)wmp, XG_MASK_3D, mms.mrm_fmasko))
			return endmerm(EWRITERR, &mms);
	   }
	return endmerm(EOK, &mms);
}

/* end of mergemask functionality */

/****************************************
	cleanmask:	automated smoothing and recentering of 3-D profiles.
	This performs a smoothing on the 3-D masks.  At the moment it does
	this by calculating the Lorenzian or Gaussian or hyperbolic secant
	ellipsoid most similar in a least-squares sense to the current mask
	and setting to zero all pixels outside a certain contour level of
	the ellipsoid.  It computes the deviation between the ellipsoid mask
	and the true mask.  The contour level is by default 1% ;
	this may be altered by a run-time input. 
*****************************************/

static void ellheader(FILE *fp)
{ /*	This prints a suitable header on the table of ellipsoid results */
	fprintf(fp, "Results of Ellipsoidal fitting:\n");
	fprintf(fp, " Region # pixels # pixels     rms     Peak  Computed ");
	fprintf(fp, "Centroid             Ellipsoid Parameters\n ");
	fprintf(fp,
	 "number   zeroed    upped     fit    i000     x0     y0     z0");
	fprintf(fp," lambda0 lambda1 lambda2   phi2z   phi1y   phi2x\n");
}

static void sumellipse(FILE *fp, int region, double *ellpars)
{ /*	Summarize the results of the ellipsoidal fitting to the stream fp.*/
	register int	i;

	fprintf(fp, " %6d %8d %8d %7.3lf %7.4lf %6.2lf %6.2lf %6.2lf",
		 region, (int)ellpars[10], (int)ellpars[12], ellpars[11],
		 ellpars[0], ellpars[7], ellpars[8], ellpars[9]);
	for (i = 1; i < 4; i++) fprintf(fp, " %7.4lf", ellpars[i]);
	for ( ; i < 7; i++) fprintf(fp, " %7.2lf", ellpars[i] / RADC);
	fprintf(fp, "\n");
}

static void sum_clean(FILE *fp, CLEANMWORK *cmp)
{ /*	This summarizes the run-time conditions for a run of "cleanmask". */
	static char	*fnames[] = {"wmask", "xlog", NULL };
	static char	ftypes[] = {3, 2, 0 };
	static char	*ellname[] =
	{ "Lorenzian", "Gaussian", "hyperbolic secant", "arc-cotangent" };

	ftypes[1] = (cmp->cm_fpl == stdout ? 0 : 2);
	header(fp, PROCESSING, fnames, ftypes);
	fprintf(fp, " Best-fit %s ellipsoids will be calculated\n",
		ellname[(int)(cmp->cm_elltype)]);
	if (cmp->cm_elltype == LORENZIAN) fprintf(fp,
	  "Exponent for Lorenzian fitting will be %8.3f\n", cmp->cm_lorpow);
	if (cmp->cm_flags & CLF_FLATTEN)
	  {
		fprintf(fp,
		 "Pixels in the output models will be set to zero if the\n");
		fprintf(fp,
		 "\tcorresponding ellipsoidal value is < %7.4f * the peak\n",
				cmp->cm_mincontour);
	  }
	if (cmp->cm_flags & CLF_EXTRAP)
	  {
		fprintf(fp,
		 "Weak pixels in output models inside %7.4f contour will\n",
			cmp->cm_mincontour);
		fprintf(fp,
		 "\thave their intensity increased to the ellipsoid value\n");
	  }
	if (cmp->cm_flags & CLF_OUTELLIPSE) fprintf(fp,
	 "Ellipsoidal models will be substituted for the input data\n");
	if (cmp->cm_flags & CLF_VERBOSE) fprintf(fp,
	  "Input and ellipsoid masks will be written to the logfile\n");
	if (cmp->cm_flags & CLF_SEMIVERBOSE) fprintf(fp,
	  "             Output masks will be written to the logfile\n");
	if (cmp->cm_flags & CLF_KEEPBORDER) fprintf(fp,
 "Nonzero values along each frame's outside border will be preserved\n");
	else {
		fprintf(fp,
 "Nonzero values along each frame's outside border will be set to zero\n");
		fprintf(fp, "\t%s in the centermost frame\n",
			(cmp->cm_flags & CLF_ZCBORDER) ? "even" : "except");
	     }
}

static void disp_model(FILE *fp, int masknum, int isshift, CLEANMWORK *cmp)
{ /*	This does a display of an image a la ewmask */
	register int	colno, fmno, rowno;
	int		ip, iv, lx, lxy, lxyz, ly, lz;
	char		*csp, *cspy, *cspz, *locdisp;
	float		*flp, *flp0;
	double		fv, sc;
	REGIONPROFILE	*rgp;

	if ((sc = cmp->cm_maxmo[masknum]) < DFUZZ) return;
	fprintf(fp,
	 "XGEN Model-Profile Editing Program: Model Profile number %2d\n",
		masknum);
	rgp = cmp->cm_regb + masknum;	flp0 = rgp->m3_xyop;
	lx = rgp->m3_bx;	ly = rgp->m3_by;	lz = rgp->m3_bz;
	lxy = lx *ly;	lxyz = lxy * lz;
	for (fmno = 0; fmno < lz; fmno++)
	   {
		if (isshift)	fprintf(fp, "   SHIFTEDFM %1d____", fmno);
		else		fprintf(fp, "   ____Frame %1d____", fmno);
	   }
	if (NULL == (locdisp = (char *)mallop(lxyz, "locdisp in disp_model")))
		return;
	for (ip = 0, csp = locdisp, flp = flp0; ip < lxyz; ip++, csp++, flp++)
	   {
		fv = *flp * 99. / sc;	iv = fv + 0.49999;
		if (isshift)
		  {
			if (iv < 1) *csp = ' ';
			else if (iv < 10) *csp = '.';
			else if (iv < 20) *csp = ',';
			else if (iv < 38) *csp = ';';
			else if (iv < 48) *csp = '0' + (iv - 38);
			else if (iv < 74) *csp = 'a' + (iv - 48);
			else if (iv < 100) *csp = 'A' + (iv - 74);
			else iv = '*';
		  }
		else {
			if (iv < 1) *csp = ' ';
			else if (iv < 10) *csp = '0' + iv;
			else if (iv < 36) *csp = 'a' + (iv - 10);
			else if (iv < 62) *csp = 'A' + (iv - 36);
			else *csp = '*';
		     }
	   }
	for (rowno = 0, cspy = locdisp; rowno < ly; rowno++, cspy += lx)
	   {
		fprintf(fp, "\n%2d |", rowno);
		for (fmno = 0, cspz = cspy;
		 fmno < lz; fmno++, cspz += lxy)
		   {
			if (fmno) fprintf(fp, "|-|");
			for (colno = 0,csp = cspz; colno < lx; colno++,csp++)
			   fprintf(fp, "%c", *csp);
			if (fmno == lz-1) fprintf(fp, "|");
		   }
	   }
	fprintf(fp, "\n  ");
	for (fmno = 0; fmno < lz; fmno++) fprintf(fp, " \\012345678901234/");
	fprintf(fp, "\n");
	fprintf(fp,
" H back 1 frame;  B back 1 mask;  K contiguity;     E exit;");
	fprintf(fp,
"           G Gaussian boundaries;  V Lorentzian boundaries;\n");
	fprintf(fp,
" I next frame;    L next mask;    X recenter mask;  W redraw;");
	fprintf(fp, "         D replace mask;        ");
	if (isshift) fprintf(fp, " T toggle to  0-61 mode;\n");
	else fprintf(fp, " T toggle to 38-99 mode;\n");
	fprintf(fp,
" R undo changes;  F zero border;  A zero frame;     U zero pix<floor;");
	fprintf(fp, " P zero frame;           J replace value\n");
	fprintf(fp,
" M down 1 pixel;  N left 1 pixel; Y up 1 pixel;     Z right 1 pixel;");
	fprintf(fp, "  S redisplay\n");
	memzap((AVOID **)&locdisp, "locdisp in disp_model");
}

static int endclean(int irv, CLEANMWORK *cmp)
{	/* This cleans up and exits a run of `cleanmask'. */
	static char clmsg[] = "Abnormal exit from `cleanmask': exit code %d\n";

	if (irv != EOK)
	  {
		if ((stdout != cmp->cm_fpl) && (NULL != cmp->cm_fpl))
			fprintf(cmp->cm_fpl, clmsg, irv);
		fprintf(stderr, clmsg, irv);
	  }
	if (NULL != cmp->cm_fpi) (void)fclose(cmp->cm_fpi);
	cmp->cm_fpi = NULL;
	memzap((AVOID **)&(cmp->cm_wmin), "cmp->cm_wmin in cleanmask");
	if (NULL == cmp->cm_regb)
	  {
		memzap((AVOID **)&((cmp->cm_regb)->m3_xyop),
			"cmp->cm_regb->m3_xyop in cleanmask");
		memzap((AVOID **)&(cmp->cm_regb), "cmp->cm_regb in cleanmask");
	  }
	if (cmp->cm_flags & CLF_INTERACT)
	  {
		if (0 == (int)cmp->cm_currpid) cmp->cm_currpid = getpid();
		sprintf(cmp->cm_locjb, "EOF_%d\n", (int)cmp->cm_currpid);
		if (EOK != xgbuf(cmp->cm_locjb, 7)) return EBAD;
	  }
	cmp->cm_fpl = NULL;	globalkill();	return irv;
}

static int clean_setup(int argc, char *argv[], CLEANMWORK *cmp)
{ /*	Argument and default setup for `cleanmask. */
	char	*pna, **dcp, **flags;
	int	lm, lxyz;
	double	stated;

	if (NULL == (cmp->cm_fpl = propen("xlog"))) cmp->cm_fpl = stdout;
	cmp->cm_elltype = HYPERSECANT;		cmp->cm_flags = CLF_BOTH;
	cmp->cm_lorpow = DEF_LORENZ_POW;
	cmp->cm_mincontour = stated = DEF_CONTOUR;
	flags = flagproc(argc, argv, "ABEFGHLNOPRSVZ");
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	   switch (**dcp)
	     {
		case	'A':	cmp->cm_elltype = ARCCOT;		break;
		case	'B':	cmp->cm_flags |= CLF_BOTH;		break;
		case	'E':	cmp->cm_flags &= ~CLF_FLATTEN;
				cmp->cm_flags |= CLF_EXTRAP;	break;
		case	'F':	cmp->cm_flags &= ~CLF_EXTRAP;
				cmp->cm_flags |= CLF_FLATTEN;	break;
		case	'G':	cmp->cm_elltype = GAUSSIAN;		break;
		case	'H':	cmp->cm_elltype = HYPERSECANT;		break;
		case	'L':	cmp->cm_elltype = LORENZIAN;		break;
		case	'N':	cmp->cm_flags |= CLF_INTERACT;	break;
		case	'O':	cmp->cm_flags |= CLF_KEEPBORDER;	break;
		case	'P':	cmp->cm_lorpow = flimits(fabs(
			atof((*dcp) + 1)), 1., 10.);		break;
		case	'R':	cmp->cm_flags |= CLF_OUTELLIPSE;	break;
		case	'S':	cmp->cm_flags |= CLF_SEMIVERBOSE;	break;
		case	'V':	cmp->cm_flags |= CLF_VERBOSE;	break;
		case	'Z':	cmp->cm_flags |= CLF_ZCBORDER;	break;
	     }
	if ((!(cmp->cm_flags & CLF_INTERACT)) &&
	 (EHELP == xgjournal(argc, argv))) return EOK; /* journal the effort */
	else {
		pna = filepart(*argv, 2);	(void)xgprogname(pna, 4);
		if (0 == (int)cmp->cm_currpid) cmp->cm_currpid = getpid();
		sprintf(cmp->cm_locjb, "%s -n << EOF_%d\n",
		 *argv, (int)(cmp->cm_currpid));
		if (EOK != xgbuf(cmp->cm_locjb, 1)) return EBAD;
		fprintf(stderr, "journal buffer begun\n");
	     }
	for (--argc, ++argv; argc--; argv++)
	   if ((**argv != '-')&& allnumeric(*argv)) stated = atof(*argv);
	if ((0. < stated) && (stated < 1.)) cmp->cm_mincontour = stated;
	if (!(cmp->cm_flags & CLF_WHATTODO)) cmp->cm_flags |= CLF_FLATTEN;
	if (cmp->cm_flags & CLF_OUTELLIPSE) cmp->cm_flags &= ~CLF_BOTH;
	if (cmp->cm_flags & CLF_KEEPBORDER) cmp->cm_flags &= ~CLF_ZCBORDER;
	if (qerror((AVOID *)(cmp->cm_fpi = xgopen("wmask", "r")),
	 "open input weighted-mask file")) return EOPENERR;
	sum_clean(stderr, cmp);	sum_clean(cmp->cm_fpl, cmp);
	lm = NREGIONS;	lxyz = BIGSPOTWID * BIGSPOTHT * PFRAMES;
	if (NULL == (cmp->cm_regb = (REGIONPROFILE *)callop(lm * 2,
	 sizeof (REGIONPROFILE),"cmp->cm_regb in cleanmask"))) return EALLOC;
	if (NULL == (cmp->cm_regb->m3_xyop = (float *)callop(lm * 2 * lxyz,
	 sizeof (float), "cmp->cm_regb->m3_xyop in cleanmask"))) return EALLOC;
	if (NULL == (cmp->cm_wmin = (MASK_3D *)callop(lm, sizeof (MASK_3D),
	 "wmpi in cleanmask"))) return EALLOC;
	return EOK;
}

static void get_rescaler(int modelno, int doprint, CLEANMWORK *cmp)
{ /*	This computes the scale factor and the brightest pixel in a model */
	int		ip, ipm, lx, lxy, lxyz;
	float		*flp0;
	double		maxinm;
	REGIONPROFILE	*rgp;

	maxinm = -HUGE;	ipm = -1;	rgp = cmp->cm_regb + modelno;
	lx = rgp->m3_bx; lxy = lx * rgp->m3_by; lxyz = lxy * rgp->m3_bz;
	for (ip = 0, flp0 = rgp->m3_xyop; ip < lxyz; ip++, flp0++)
		if (*flp0 > maxinm)
		  {
			maxinm = *flp0; ipm = ip;
		  }
	if (maxinm > 0)
	  {
		cmp->cm_maxmo[modelno] = maxinm;
		if (doprint) fprintf(stdout,
		 "Region %2d: max = %8.2f at [%2d,%2d,%1d]\n", modelno,
		 cmp->cm_maxmo[modelno], ipm / lxy, (ipm % lxy) / lx, ipm % lx);
	  }
}

static void zero_image(int ac, char *av[], int curv[], CLEANMWORK *cmp)
{ /*	This sets all the pixels within a single image to zero. */
	int		ix, iy, lx, lxy, ly, lz, we, ze;
	float		*flp;
	REGIONPROFILE	*rgp;

	lz = cmp->cm_regb->m3_bz;
	if (ac > 3)
	  {
		we = atoi(av[2]);	ze = atoi(av[3]);
		curv[3] = ((0 <= we) && (we < NREGIONS)) ? we : 0;
		curv[2] = ((0 <= ze) && (ze < lz)) ? ze : 0;
	  }
	rgp = cmp->cm_regb + curv[3];
	lx = rgp->m3_bx; ly = rgp->m3_by; lz = rgp->m3_bz;
	lxy = lx * ly;
	for (iy = 0, flp = rgp->m3_xyop + lxy * curv[2]; iy < ly; iy++)
		   for (ix = 0; ix < lx; ix++, flp++) *flp = 0.;
	fprintf(stdout, "Set all pixels in image %1d of model %2d to zero\n",
		curv[2], curv[3]);
}

static void zero_bounds(int ac, char *av[], int curv[], CLEANMWORK *cmp)
{ /*	Zero the boundaries in a model */
	char		thisco;
	register int	ix, iy, iz, region;
	int		frame_min, frame_max, incl_cen;
	int		lx, lxy, ly, lz, lz2, mask_max, mask_min, nzer;
	float		*flp;
	REGIONPROFILE	*rgp;

	lx = (cmp->cm_regb)->m3_bx; ly = (cmp->cm_regb)->m3_by;
	lz = (cmp->cm_regb)->m3_bz;
	lxy = lx * ly;	lz2 = lz / 2;
	mask_min = 0;	mask_max = NREGIONS - 1;
	frame_min = 0;	frame_max = lz - 1;
	incl_cen = 0;
	if (ac > 2)
	  {
		if (('a' <= (thisco = av[2][0])) && (thisco <= 'z'))
			thisco -= 'a' - 'A';
		if ('M' == thisco) mask_min = mask_max = curv[3];
		else if ('F' == thisco)
		  {
			mask_min = mask_max = curv[3];
			frame_min = frame_max = curv[2];
		  }
		if ((ac > 3) && ((av[3][0] == 'C') || (av[3][0] == 'c')))
			incl_cen = 1;
	  }
	for (region = mask_min, rgp = cmp->cm_regb + mask_min;
	   region <= mask_max; region++, rgp++)
	   {
		flp = rgp->m3_xyop + frame_min * lxy; nzer = 0;
		for (iz = frame_min; iz <= frame_max; iz++)
		   {
			if ((incl_cen) || (iz != lz2))
			  {
			     for (iy = 0; iy < ly; iy++)
			       for (ix = 0; ix < lx; ix++, flp++)
				  {
					if ((ix == 0) || (ix == lx - 1)
					 || (iy == 0) || (iy == ly - 1))
					  {
						if (*flp > FUZZ) nzer++;
						*flp = 0;
					  }
				  }
			     if (frame_min == frame_max) fprintf(stdout,
	"%4d nonzero perimeter pixels of frame %1d of model %2d zeroed\n",
				nzer, iz, region);
			  }
			else flp += lxy;
		  }
		if (frame_min != frame_max) fprintf(stdout,
		 "%4d nonzero perimeter pixels of model %2d zeroed %s\n",
		 nzer, region, incl_cen ?
		 "in all images" : "in all images except the center");
	   }
}

static void declare_position(FILE *fp, int *curv)
{ /*	This declares where the imaginary cursor is currently sitting */
	fprintf(fp,
	 "Current position: model %2d, image %1d, (X,Y) = %2d, %2d\n",
		curv[3], curv[2], curv[0], curv[1]);
}

static void zero_weak(int ac, char *av[], int curv[], CLEANMWORK *cmp)
{ /*	Zero out the weak pixels */
	char		thisco;
	register int	ix, iy, iz, region;
	int		frame_min, frame_max, mask_min, mask_max;
	int		lx, lxy, ly, nzer;
	float		*flp;
	double		floor_value, spec;
	REGIONPROFILE	*rgp;

	floor_value = 0.8;	thisco = 'A';
	mask_min = 0;	mask_max = NREGIONS - 1;
	frame_min = 0;	frame_max = (cmp->cm_regb)->m3_bz - 1;
	lx = (cmp->cm_regb)->m3_bx; ly = (cmp->cm_regb)->m3_by;
	lxy = lx * ly;
	if (ac > 2)
	  {
		if ((0. <= (spec = atof(av[2]))) && (spec < 99.))
			floor_value = spec;
		if (ac > 3)
		  {
			if (('a' <= (thisco = av[3][0])) && (thisco <= 'z'))
				thisco -= 'a' - 'A';
			if (thisco == 'M') mask_min = mask_max = curv[3];
			else if (thisco == 'F')
			  {
				mask_min = mask_max = curv[3];
				frame_min = frame_max = curv[2];
			  }
		  }
	  }
	for (region = mask_min, rgp = cmp->cm_regb + mask_min;
	   region <= mask_max; region++, rgp++)
	   {
		flp = rgp->m3_xyop + frame_min * lxy;	nzer = 0;
		for (iz = frame_min; iz <= frame_max; iz++)
		   {
			for (iy = 0; iy < ly; iy++)
			   for (ix = 0; ix < lx; ix++, flp++)
				if (*flp < floor_value)
				  {
					if (*flp > 0) nzer++;	*flp = 0.;
				  }
			if (frame_min == frame_max) fprintf(stdout,
		 "%4d Pixels below %14.5e in image %1d of model %2d zeroed\n",
				nzer, floor_value, iz, region);
		   }
		if (frame_min != frame_max) fprintf(stdout,
		  "%4d Pixels below %14.5e in model %2d zeroed\n",
			nzer, floor_value, region);
	   }
}

static void inter_contig(int ac, char *av[], int curv[], CLEANMWORK *cmp)
{ /* Enforce contiguity in a model */
	register int	iw;
	char		thisco;
	int		mask_max, mask_min, nfail, nsuc, ret, thisfmonly;
	int		suc[NREGIONS];
	REGIONPROFILE	*rgp;

	mask_min = 0; mask_max = NREGIONS - 1;	thisfmonly = -1;
	if (ac > 2) 
	  {
		if (('a' <= (thisco = av[1][0])) && (thisco <= 'z'))
			thisco -= 'a' - 'A';
		if ('M' == thisco) mask_min = mask_max = curv[3];
		else if ('F' == thisco)
		  {
			mask_min = mask_max = curv[3]; thisfmonly = curv[2];
		  }
	  }
	for (iw = 0; iw < NREGIONS; iw++) suc[iw] = -9;
	for (rgp = cmp->cm_regb + mask_min, iw = mask_min;
	 iw <= mask_max; iw++, rgp++)
	   {
		ret = contig(rgp, thisfmonly);
		if (thisfmonly == -1) suc[iw] = ret;
		else fprintf(stdout,
 "Contiguity enforced%ssuccessfully in image %1d of region %2d\n",
			 (0 < ret) ? " " : " un", thisfmonly, iw);
	   }
	/* count and then display the successes and failures */
	for (nsuc = nfail = iw = 0; iw < NREGIONS; iw++)
	   {
		if (suc[iw] > 0) nsuc++;	else if (suc[iw] == 0) nfail++;
	   }
	if (nsuc)
	  {
		fprintf(stdout, "Contiguity enforced successfully in %s",
		 (nsuc == 1) ? "model" : "models");
		for (iw = 0; iw < NREGIONS; iw++)
		   if (suc[iw] > 0) fprintf(stdout, " %2d", iw);
		fprintf(stdout, "\n");
	  }
	if (nfail)
	  {
		fprintf(stdout, "Contiguity enforced unsuccessfully in %s",
		 (nfail == 1) ? "model" : "models");
		for (iw = 0; iw < NREGIONS; iw++)
		   if (suc[iw] == 0) fprintf(stdout, " %2d", iw);
		fprintf(stdout, "\n");
	  }
}

static void cpymodel(REGIONPROFILE *rop, REGIONPROFILE *rip)
{ /* This copies a 3-D model from rip to rop */
	int		lxyz;
	size_t		modelsize, worksize;

	lxyz = rip->m3_bx * rip->m3_by * rip->m3_bz;
	worksize = sizeof (REGIONPROFILE);
	modelsize = lxyz * sizeof (float);
	cpymem((AVOID *)rop, (AVOID *)rip, worksize);
	cpymem((AVOID *)rop->m3_xyop, (AVOID *)rip->m3_xyop, modelsize);
}

static int irecen_vec(int ac, char *av[], int curv[], CLEANMWORK *cmp)
{ /*	recenters one or more more models */
	register int	region;
	int		lx, ly, lz, lxyz, mask_max, mask_min;
	float		*cenbu;
	double		idv[3], delv[3];
	REGIONPROFILE	*ogp, *rgp;

	mask_min = 0;	mask_max = NREGIONS - 1;
	if ((ac > 2) && ((av[2][0] == 'm') || (av[2][0] == 'M')))
		mask_min = mask_max = curv[3];
	rgp = cmp->cm_regb + mask_min;	ogp = rgp + NREGIONS;
	lx = (cmp->cm_regb)->m3_bx;	ly = (cmp->cm_regb)->m3_by;
	lz = (cmp->cm_regb)->m3_bz;
	lxyz = lx * ly * lz;
	if (NULL == (cenbu = (float *)callop(lxyz, sizeof (float),
	 "cenbu in irecen_vec"))) return EALLOC;
	for (region = mask_min; region <= mask_max; region++, rgp++, ogp++)
	   {
		idv[0] = idv[1] = idv[2] = delv[0] = delv[1] = delv[2] = 0.;
		if (FUZZ < gen_centroid(rgp->m3_xyop, NULL, lx, ly, lz,
			idv, delv))
		  {
			if ((fabs(delv[0]) > 0.0001) ||
			 (fabs(delv[1]) > 0.001) || (fabs(delv[2]) > 0.0001))
			  {
				recen_vec(rgp->m3_xyop, delv, cenbu, lx,ly,lz);
				cpymem((void *)(rgp->m3_xyop), (void *)cenbu,
				 lxyz * sizeof (float));
				fprintf(stdout,
				 "Model %2d recentered:%8.4f %8.4f %8.4f\n",
				 region, delv[0], delv[1], delv[2]);
				get_rescaler(region, 1, cmp);
			  }
			else fprintf(stdout,
			 "model %2d was centered to begin with\n", region);
		  }
		else fprintf(stdout, "Unable to recenter model %2d\n", region);
	   }
	memzap((AVOID **)&cenbu, "cenbu in irecen_vec");
	return EOK;
}

static void ellipseclean(int ac, char *av[], int curv[],
	int ety, CLEANMWORK *cmp)
{ /*	This sets up and performs a Gaussian or Lorenzian cleanup */
	char		ccfl;
	int		cfla, iw, mask_max, mask_min, ret;
	double		lorp, mincon, spec;
	REGIONPROFILE	*rgp;

	mincon = 0.008;
	mask_min = 0;	mask_max = NREGIONS - 1;	lorp = 4.;
	cfla = CLF_EXTRAP | CLF_FLATTEN;
	if ((ac > 2) && (0. < (spec = atof(av[2]))) && (spec < 1.))
		mincon = spec;
	if ((ac > 3) && ((av[3][0] == 'M') || (av[3][0] == 'm')))
		mask_min = mask_max = curv[3];
	if (ac > 4)
	  {
		ccfl = av[4][0];
		if (('a' <= ccfl) && (ccfl <= 'z')) ccfl -= 'a' - 'A';
		if (ccfl == 'F') cfla = CLF_FLATTEN;
		if (ccfl == 'E') cfla = CLF_EXTRAP;
		if (ccfl == 'R') cfla = CLF_OUTELLIPSE;
	  }
	if ((ac > 5) && (1.5 < (spec = atof(av[5]))) && (spec < 99.))
		lorp = spec;
	ellheader(stdout);
	for (iw = mask_min, rgp = cmp->cm_regb + mask_min;
	 iw <= mask_max; iw++, rgp++)
	   {
		cmp->cm_ellipspars[iw][0] = 1.;
		for (ret = 1; ret < 13; ret++) cmp->cm_ellipspars[iw][ret] = 0.;
		ret = ellsoid(rgp, ety, cfla, mincon,
		 lorp, cmp->cm_ellipspars[iw]);
		if (EOK == ret)
		  {
			sumellipse(stdout, iw, cmp->cm_ellipspars[iw]); 
			get_rescaler(iw, 0, cmp);
		  }
		else fprintf(stdout,
 "Ellipsoidal modifications on model %2d failed: err = %d\n", iw, ret);
	   }
}

static void change_model(int ac, char *av[], int curv[],
	int isback, CLEANMWORK *cmp)
{ /*	This moves us from one model to another */
	int	ispec;

	if ((ac > 2) && (0 <= (ispec = atoi(av[2]))) && (ispec <= NREGIONS-1))
		curv[3] = ispec;
	else {
		if (isback)
		  {
			if (curv[3] > 0) (curv[3])--;
			else curv[3] = NREGIONS - 1;
		  }
		else {
			if (curv[3] < NREGIONS - 1) (curv[3])++;
			else curv[3] = 0;
		     }
	     }
	curv[0] = (cmp->cm_regb)->m3_bx / 2;
	curv[1] = (cmp->cm_regb)->m3_by / 2;
	curv[2] = (cmp->cm_regb)->m3_bz / 2;
}

static void change_image(int ac, char *av[], int curv[], int isback,
	CLEANMWORK *cmp)
{ /*	This moves us from one image within a model to another */
	int	ispec, lz;

	lz = (cmp->cm_regb)->m3_bz;
	if ((ac > 2) && (0 <= (ispec = atoi(av[2]))) && (ispec <= NREGIONS-1))
		curv[2] = ispec;
	else {
		if (isback)
		  {
			if (curv[2] > 0) (curv[2])--;
			else curv[2] = lz - 1;	return;
		  }
		else {
			if (curv[2] < lz - 1) (curv[2])++;
			else curv[2] = 0;	return;
		     }
	     }
}

static void replaceval(int ac, char *av[], int curv[], CLEANMWORK *cmp)
{ /*	This replaces the value at the current position with a new one */
	int		ip, lx, lxy;
	double		newval, newvalc, spec;
	float		*flp0;
	REGIONPROFILE	*rgp;

	if (cmp->cm_maxmo[curv[3]] < DFUZZ)
	  {
		fprintf(stdout, "Model %2d unscalable\n", curv[3]); return;
	  }
	if (ac < 3)
	  {
		fprintf(stdout, "No new value specified\n"); return;
	  }
	if ((0 <= (spec = atof(av[2]))) && (spec <= 99.))
	  {
		newval = spec; newvalc = newval * cmp->cm_maxmo[curv[3]] / 99.;
		rgp = cmp->cm_regb + curv[3];	flp0 = rgp->m3_xyop;
		lx = rgp->m3_bx;	lxy = lx * rgp->m3_by;
		ip = curv[2] * lxy + curv[1] * lx + curv[0];
		spec = *(flp0 + ip);
		if (fabs(spec - newvalc) < DFUZZ)
		  {
			fprintf(stdout,
 "No replacement: new value = %7.2f @ [%2d,%2d,%1d,%2d] =  old value\n",
			 newvalc, curv[0], curv[1], curv[2], curv[3]); return;
		  }
		*(flp0 + ip) = newvalc;
		fprintf(stderr,
		 "New value of %8.2f at [%2d,%2d,%1d,%2d] replaces %8.2f\n",
			newvalc, curv[0], curv[1], curv[2], curv[3], spec);
		return;
	  }
	else {
		fprintf(stdout, "Specified value %13.4e out of range\n", spec);
		return;
	     }
}

static void replace_image(int ac, char *av[], int curv[], CLEANMWORK *cmp)
{ /*	This replaces one model with another. It's forbidden to
	copy a model into itself */
	int		dest, source;
	REGIONPROFILE	*rgp, *ogp;

	if ((ac < 2) || (0 > (source = atoi(av[2]))) || (source > NREGIONS - 1)
	 || (source == curv[3])) return;
	dest = curv[3];
	rgp = cmp->cm_regb + dest;	ogp = cmp->cm_regb + source;
	cpymodel(rgp, ogp);
	fprintf(stdout, "Installed contents of model %2d into model %2d\n",
		source, dest);
	cmp->cm_maxmo[dest] = cmp->cm_maxmo[source];
}

static void addtowings(int ac, char *av[], int curv[], CLEANMWORK *cmp)
{ /*	This takes one or more models for which all the intensity is
	concentrated in the central image and feeds a little to the neighboring
	images. */
	register int	iw, ip;
	char		thisco;
	int		lxy, lz2, mask_max, mask_min, newpix;
	double		fll, flm, flr, imratio, newv, oldv, spec;
	float		*flp, *flp0, *flpl, *flq;
	REGIONPROFILE	*rgp;

	thisco = 'A';	imratio = 0.01;
	mask_min = 0;	mask_max = NREGIONS - 1;
	lz2 = (cmp->cm_regb)->m3_bz / 2;
	if (ac > 2)
	  {
		if (('a' <= (thisco = av[2][0])) && (thisco <= 'z'))
			thisco -= 'a' - 'A';
		if (thisco == 'M') mask_min = mask_max = curv[3];
		if (ac > 3)
		  {
			if ((0. <= (spec = atof(av[2]))) && (spec < 0.1))
				imratio = spec;
		  }
	  }
	for (iw = mask_min, rgp = cmp->cm_regb + mask_min;
	 iw <= mask_max; iw++, rgp++)
	   { /* count up intensity in images 2,3,4 */
		lxy = rgp->m3_bx * rgp->m3_by;
		flp0 = rgp->m3_xyop;	flpl = flp0 + lxy * (lz2 - 1);
		for (fll = ip = 0, flp = flpl; ip < lxy; ip++, flp++)
			fll += *flp;
		for (flm = ip = 0; ip < lxy; ip++, flp++) flm += *flp;
		for (flr = ip = 0; ip < lxy; ip++, flp++) flr += *flp;
		if (flm > DFUZZ)
		  {
			if (fll < imratio * flm) /* add inten on left */
			  {
				flp = flpl;	flq = flp + lxy; newpix = 0;
				for (ip = 0; ip < lxy; ip++, flp++, flq++)
				   {
					oldv = *flp;	newv = *flq * imratio;
					if (newv > oldv)
					  {
						*flp = newv; newpix++;
					  }
				   }
				if (newpix) fprintf(stdout,
 " image 2,model %2d: %4d values replaced with %7.3f * val(image 3)\n",
	iw, newpix, imratio);
			  }
			if (flr < imratio * flm) /* add inten on right */
			  {
				flq = flpl + lxy; flp = flq + lxy; newpix = 0;
				for (ip = 0; ip < lxy; ip++, flp++, flq++)
				   {
					oldv = *flp;	newv = *flq * imratio;
					if (newv > oldv)
					  {
						*flp = newv;	newpix++;
					  }
				   }
				if (newpix) fprintf(stdout,
 " image 4,model %2d: %4d values replaced with %7.3f * val(image 3)\n",
	iw, newpix, imratio);
			  }
		  }
	  }
}

static void change_xy(int ac, char *av[], int curv[], int code, CLEANMWORK *cmp)
{ /*	This repositions the imaginary pointer within a specific image */
	int	ispec, lx, ly;

	lx = (cmp->cm_regb)->m3_bx;	ly = (cmp->cm_regb)->m3_by;
	if ((ac > 2) && (0 <= (ispec = atoi(av[2]))) && (ispec < lx))
	  {
		if (code & 01) curv[0] = ispec;	/* move left or right */
		else curv[1] = ispec;		/* move up or down */
		return;
	  }
	else switch (code)
	  {
		case	0:	/* move down */
			if (curv[1] < ly - 1) (curv[1])++;
			else curv[1] = 0;	return;
		case	1:	/* move left */
			if (curv[0] > 0) (curv[0])--;
			else curv[0] = lx - 1;	return;
		case	2:	/* move up */
			if (curv[1] > 0) (curv[1])--;
			else curv[1] = ly - 1;	return;
		case	3:	/* move right */
			if (curv[0] < lx - 1) (curv[0])++;
			else curv[0] = 0;	return;
	  }
}

static int cleanm_interact(CLEANMWORK *cmp)
{ /*	This enables the interactive mode of `cleanmask' */
	int		ac, i, lx, ly, lz, shift, currv[4];
	REGIONPROFILE	*rgp, *rsp;
	char		**av, *cp;
	static char	cll[82];

	 /* default values indicating where we are */
	lx = (cmp->cm_regb)->m3_bx;
	ly = (cmp->cm_regb)->m3_by;	lz = (cmp->cm_regb)->m3_bz;
	currv[0] = lx / 2; currv[1] = ly / 2; currv[2] = lz / 2;
	currv[3] = 0;
	declare_position(stdout, currv);
	/* we need rescalers for the 'J' command and others */
	for (i = 0; i < NREGIONS; i++) get_rescaler(i, 1, cmp);
	shift = 0;
	disp_model(stdout, currv[3], shift, cmp);
	while (1)
	     {
		printf("M%2d,[%2d,%2d,%1d]: ",
		 currv[3], currv[0], currv[1], currv[2]);
		zeromem((AVOID *)cll, 80);
		if (NULL != fgets(cll, 80, stdin))
		  {
			for (cp = cll; ('\0' != *cp) && ('\n' != *cp); cp++)
			 if ((('A' <= *cp) && (*cp <= 'Z')) ||
			 	(('a' <= *cp) && (*cp <= 'z'))) break;
			if (('a' <=  *cp) && (*cp <= 'z')) *cp -= 'a' - 'A';
			if (('A' <=  *cp) && (*cp <= 'Z'))
			  {
			    if (NULL == (av = str2acav(cp, &ac))) break;
			    if (('S' != *cp) && ('T' != *cp))
			     {
				strcpy(cmp->cm_locjb, cp);
				if (EOK != xgbuf(cmp->cm_locjb,1)) return EBAD;
			     }
			    switch (*cp)
			     {
				case	'A':	case	'P':
				 zero_image(ac, av, currv, cmp); break;
				case	'B': /* previous model */
				 change_model(ac, av, currv, 1, cmp); break;
				case	'D': /*copy model to another*/
				 replace_image(ac, av, currv, cmp); break;
				case	'E': /* exit */
				 fprintf(stdout, "Exit requested!\n");
				 return EOK;
				case	'F': /* zero boundaries */
				 zero_bounds(ac, av, currv, cmp); break;
				case	'G': /* Gaussian rebuild */
				 ellipseclean(ac, av, currv, GAUSSIAN, cmp);
				 break;
				case	'H': /* previous image */
				 change_image(ac, av, currv, 1, cmp); break;
				case	'I': /* next image */
				 change_image(ac, av, currv, 0, cmp); break;
				case	'J': /* replace value */
				 replaceval(ac, av, currv, cmp);	break;
				case	'K': /* contiguity */
				 inter_contig(ac, av, currv, cmp); break;
				case	'L': /* next model */
				 change_model(ac, av, currv, 0, cmp); break;
				case	'M': /* move down */
				 change_xy(ac, av, currv, 0, cmp); break;
				case	'N': /* move left */
				 change_xy(ac, av, currv, 1, cmp); break;
				case	'O': /* add inten to wings */
				 addtowings(ac, av, currv, cmp); break;
				case	'Q': /* quit without output */
				 return EBAD;
				case	'R': /* restore orig model */
				 rgp = cmp->cm_regb + currv[3];
				 rsp = cmp->cm_regb + NREGIONS + currv[3];
				 cpymodel(rgp, rsp);	break;
				case	'S': /* show model */
				 disp_model(stdout, currv[3], shift, cmp);
				 break;
				case	'T': /* toggle shift status */
				 shift = 1 - shift;
				 disp_model(stdout, currv[3], shift, cmp);
				 break;
				case	'U': /* zero weak pixels */
				 zero_weak(ac, av, currv, cmp); break;
				case	'V':
				 ellipseclean(ac, av, currv, LORENZIAN, cmp);
				 break;
				case	'X':
				 (void)irecen_vec(ac, av, currv, cmp); break;
				case	'Y':	/* move up */
				 change_xy(ac, av, currv, 2, cmp); break;
				case	'Z':	/* move right */
				 change_xy(ac, av, currv, 3, cmp); break;
			     } /* end of switch on *cp */
			    memzap((AVOID **)&(*av),
				"*av in cleanm_interact");
			    memzap((AVOID **)&av,
				"av in cleanm_interact");
			    ac = 0;
			  }	/* end of "if ('A' <= ..." clause */
		  }
	     }
	return EOK;
}

int cleanmmain(int argc, char *argv[])
{ /*	This is the mainline for "cleanmask".  It reads in the WMASK file,
	computes the ellipsoid form, prints out the latter,
	applies it at the 1.5% contour level to the input mask, and
	prints the latter. */
	int			cleanresult, lm, lx, ly, lz, lxy, lxyz;
	int			region, result;
	double			delv[3], idv[3];
	FILE			*fpo;
	MASK_3D			*wmpi;
	REGIONPROFILE		*ogp, *rgp;
	static float		*flp, *flpo;
	static CLEANMWORK	cms;

	if (EOK != (result = clean_setup(argc, argv, &cms)))
		return endclean(result, &cms);
	/* temporarily we will set these values: soon we'll read them. */
	lx = BIGSPOTWID;	ly = BIGSPOTHT;	lz = PFRAMES;
	lxy = lx * ly;	lxyz = lxy * lz;	lm = NREGIONS;
	ogp = cms.cm_regb + lm;
	flpo = cms.cm_regb->m3_xyop + lm * lxyz;
	/* read models in and convert to floating-point arrays. */
	if (cms.cm_flags & CLF_VERBOSE)
		fprintf(cms.cm_fpl, "\f\t\tInput masks:\n");
	for (region = 0, wmpi = cms.cm_wmin, rgp = cms.cm_regb,
	 flp = cms.cm_regb->m3_xyop; region < lm;
	 region++, wmpi++, rgp++, flp += lxyz, ogp++, flpo+= lxyz)
	   {
		if (0 >= xgread((void *)wmpi, XG_MASK_3D, cms.cm_fpi))
		 return endclean(EREADERR, &cms);
		if (cms.cm_flags & CLF_VERBOSE)
			prmask(wmpi, region, cms.cm_fpl);
		if ((EOK != (result = wmptovec(wmpi, flp))) &&
		 (EBADVALUE != result)) return endclean(result, &cms);
		ogp->m3_bx = rgp->m3_bx = lx;
		ogp->m3_by = rgp->m3_by = ly;
		ogp->m3_bz = rgp->m3_bz = lz;
		ogp->m3_nmodels = lm;	ogp->m3_imodel = region;
		ogp->m3_nobs = rgp->m3_nobs = wmpi->phielt[3].mnobs;
		ogp->m3_isqsum = rgp->m3_isqsum = wmpi->phielt[3].isqsum;
		rgp->m3_xyop = flp;	ogp->m3_xyop = flpo;
	   }
	(void)fclose(cms.cm_fpi);	cms.cm_fpi = NULL;
	if (cms.cm_flags & CLF_VERBOSE) fprintf(cms.cm_fpl, "\f");
	if (cms.cm_flags & CLF_INTERACT)
	  {
		cleanresult = cleanm_interact(&cms);
		if (EOK != cleanresult)
		  {
			fprintf(stdout,
			 "okay; you've chosen not to write this out\n");
			return EOK;
		  }
	  }
	else {
		ellheader(stderr);	ellheader(cms.cm_fpl);
		fprintf(cms.cm_fpl, " region fm iy ix obs.val predval\n");
		for (region = 0, rgp = cms.cm_regb,
		 ogp = cms.cm_regb + NREGIONS;
		 region < NREGIONS; region++, rgp++, ogp++)
		   {
			if (blankf(rgp->m3_xyop, lz))
				meanprofile(cms.cm_regb, region);
			idv[0] = idv[1] = idv[2] = 0.;
			if (DFUZZ > gen_centroid(rgp->m3_xyop, NULL, lx, ly,
			 lz, idv, delv)) return endclean(EBADVALUE, &cms);
			if ((fabs(delv[0]) > 0.1) || (fabs(delv[1]) > 0.1) ||
			 (fabs(delv[2]) > 0.1))
			  {
				recen_vec(rgp->m3_xyop, delv, ogp->m3_xyop,
					lx, ly, lz);
				cpymem((void *)(rgp->m3_xyop),
				 (void *)(ogp->m3_xyop), lxyz * sizeof (float));
			  }
			if (EOK != (result = ellsoid(rgp, cms.cm_elltype,
			 cms.cm_flags, cms.cm_mincontour, cms.cm_lorpow,
			 cms.cm_ellipspars[region])))
				return endclean(EALLOC, &cms);
			sumellipse(stderr, region, cms.cm_ellipspars[region]); 
			sumellipse(cms.cm_fpl, region,
				cms.cm_ellipspars[region]);
		   }
		xeflush(cms.cm_fpl);
		if (cms.cm_flags & CLF_VERBOSE)
		  {
			fprintf(cms.cm_fpl,"\f\t\t\tEllipsoidal Masks:\n");
			for (region = 0, rgp = cms.cm_regb + NREGIONS;
			 region < NREGIONS; region++, rgp++)
			   if (EOK == rgptowmp(rgp, cms.cm_wmin))
				prmask(cms.cm_wmin, region, cms.cm_fpl);
		  }
	    }
	if (qerror((AVOID *)(fpo = xgopen("wmask", "w")),
	 "open output weighted-mask file")) return endclean(EOPENERR, &cms);
	if (cms.cm_flags & CLF_SEMIVERBOSE)
		fprintf(cms.cm_fpl,"\f\t\t\tFlattened Masks:\n");
	for (region = 0, wmpi = cms.cm_wmin, rgp = cms.cm_regb;
	 region < NREGIONS; region++, wmpi++, rgp++)
	   {
		if (cms.cm_flags & CLF_OUTELLIPSE)
		 cpymem((void *)(rgp->m3_xyop),
		 (void *)((rgp + NREGIONS)->m3_xyop), lxyz * sizeof (float));
		if (!(cms.cm_flags & CLF_KEEPBORDER))
			zero_boundary(rgp, cms.cm_flags);
		if (EOK != rgptowmp(rgp, wmpi))
		  {
			fprintf(stderr,
			 "Note! profile %2d required substitution\n", region);
			fprintf(stderr,
			 "Note! profile %2d required substitution\n", region);
			meanprofile(cms.cm_regb, region);
			(void)rgptowmp(rgp, wmpi);
		  }
		if (cms.cm_flags & CLF_SEMIVERBOSE)
			prmask(wmpi, region, cms.cm_fpl);
		if (0 >= xgwrite((void *)wmpi, XG_MASK_3D, fpo))
		 return endclean(EWRITERR, &cms);
	   }
	return endclean(EOK, &cms);
}

/* end of cleanmask functionality */

/* end of maskutils.c */
