/* mrefutils.c -- utilities associated with MULTIREF files

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	26 Jun 2006: include xgenpred.h and xgpostint.h: need that stuff
	19 May 2006: merge praxial.c, cut_off.c to form this file
	12 Oct 2005: eliminate static globals; housekeeping
	22 Jun 2005: cut_results() declared as static
	11 May 2005: check return status from xgjournal
	28 Aug 2004: routines used only here are declared as static; cosmetics
	 1 Feb 2004: praxial: don't add 100 to the spacegroup for rhombohedrals
		if it's already been done
	26 Jan 2004: some static globals replaced with local variables
	21 Jan 2004: license message inserted and initial comments reordered
	 8 Jan 2004: replace extern global variables with static globals
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Aug 2003: praxial: replace global uphkl[3] with static lphkl[3]
	 5 Jan 2003: reorganized source-file header
	25 Mar 2002: removed MSI attachments
	 8 Nov 1996: minor changes for X-GEN checkin at MSI
	31 May 1995: "register" -> "register int" throughout
	20 May 1995: praxial: return EOK at end of main
	 2 Mar 1995: diagnostics to stdout
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	11 Oct 1994: cut_off: add "complement" option to remove annuli of data
	 5 Oct 1994: praxial: use xgread for PREDREF and MULTIREF structures
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: a few I/O changes and global variables moved around
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version of praxial and cut_off
	22 Oct 1986: make actual deletions the default and define "-f" flag
	 1 Jun 1985: initial version of cut_off

   Purpose: manipulating and printing segments of MULTIREF files
	praxial	prints reflections in a multiref or prediction file for
	 which any two indices are zero, i.e. the h00, 0k0, and 00l refls.
	cut_off reads a MULTIREF and writes another containing only the
	 reflections within a specified resolution range
   Cross references: xdump.c, redurefl.c, ../xglib/ascout.c

*************************************************************************/
#include	"xgpostint.h"
#include	"xgenpred.h"
#define		PREDICTION_TYPE	47
#define		MULTIREF_TYPE	29
/* run-time flag bits for cut_off: */
#define		CUT_COMPLEMENT	01
#define		CUT_FLAGONLY	02

 typedef struct {	/* convenience structure type for cut_off */
		int	cu_nmer, cu_nnomer, cu_nobs;
		} CUTCOUNT;	

extern int praxmain(int argc, char *argv[]);
extern int cutoffmain(int argc, char *argv[]);

/*******************************************
	praxial:	print reflections in a MULTIREF or USPOTLIST file for
	which any 2 indices are zero, i.e. the h00, 0k0, and 00l reflections.
*******************************************/

static int endprax(int reval, FILE *fpi)
{ /*	clean ending of this function */
	if (NULL != fpi) (void)fclose(fpi);
	globalkill();	return reval;
}

int praxmain(int argc, char *argv[])
{ /* Mainline for function to print axial reflections (h00),(0k0),(00l) */
	register int	naxial, nzeroes;
	int		file_type, omit;
	short		ispgp, lphkl[3];
	FILE		*fpi, *fplog;
	PREDREF		prrefl;
	CRYSTAL		xhd;
	DATARUN		rhd;
	MERGEDREF	mr;
	static char	*fnames[] =
		{ "uspotlist", "multiref", "xlog", NULL };
	static char	ftypes[] = { 0, 1, 2, 0 };

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	if (argc > 1) file_type = PREDICTION_TYPE;
	else file_type = MULTIREF_TYPE;
	if ((file_type == PREDICTION_TYPE) && (qerror((AVOID *)(fpi =
		xgopen("uspotlist","r")), "open prediction input file")))
			return endprax(EOPENERR, NULL);
	else if (qerror((AVOID *)(fpi = xgopen("multiref","r")),
		"open multiref input file")) return endprax(EOPENERR, NULL);
	if (NULL == (AVOID *)(fplog = propen("xlog"))) fplog = stdout;
	omit = (file_type == PREDICTION_TYPE) ? 1 : 0;
	ftypes[omit] = 0;	ftypes[1-omit] = 1;
	header(fplog , REDUCTION, fnames, ftypes);
	header(stdout, REDUCTION, fnames, ftypes);
	naxial = 0;
	if (file_type == PREDICTION_TYPE)
	  while (0 < xgread((void *)&prrefl, XG_PREDREF, fpi))
	     {
		hunpack(prrefl.pindx, lphkl);
		nzeroes = (!lphkl[0]) + (!lphkl[1]) + (!lphkl[2]);
		if (nzeroes > 1)
		  {
			naxial++;	fprintf(fplog,
			 "%4d%4d%4d %5d %8.3f %8.3f %9.3f %11.4f %8.5f\n",
				lphkl[0], lphkl[1], lphkl[2], prrefl.pfmno,
				prrefl.pxt, prrefl.pyt, prrefl.pphi,
				prrefl.pcor,prrefl.pgama);
		  }
	     }
	else
	 {
	  if (verror(0 < xgread((void *)&xhd, XG_CRYSTAL, fpi),
		"read crystal header from input multiref"))
	     {
		if (stdout != fplog) (void)fclose(fplog);
		return endprax(EREADERR, fpi);
	     }
	   ispgp = xhd.xt_intl;
	   if ((ispgp < 200) && (xhd.xt_isys == 7)) ispgp += 100;
	   for (nzeroes = xhd.xt_nruns; nzeroes--; )
		if (verror(0 <= xgread((void *)&rhd, XG_DATARUN, fpi),
		  "read datarun header from input multiref"))
		  {
			if (stdout != fplog) (void)fclose(fplog);
			return endprax(EREADERR, fpi);
		  }
	   (void)keep_dimen(&rhd);
	   while (0 < xgread((void *)&mr, XG_MERGEDREF, fpi))
	     {
		munpack(mr.mr_hkl, lphkl, ispgp);
		nzeroes = (!lphkl[0]) + (!lphkl[1]) + (!lphkl[2]);
		if (nzeroes > 1)
		  {
			pr_merref(fplog, &mr, ispgp); naxial++;
		  }
	     }
	  }
	fprintf(fplog, "\t%6d axial reflections found\n", naxial);
	if (stdout != fplog)
	  {
		fprintf(stdout,"\t%6d axial reflections found\n", naxial);
		(void)fclose(fplog);
	  }
	return endprax(EOK, fpi);
}

/* end of praxial functionality */

/***************************************
	cut_off: eliminate reflections from a MULTIREF based on sin(theta)
	By default, the output will contain only the reflections
	within a resolution range, viz. Dmin < D < Dmax.
	If the user specifies a "-f" (for FLAG) option, it will write out all
	the input reflections, but with the reflections outside Dmin < D < Dmax
	flagged by having their "shift" values set to
	minus their absolute value.
*************************************************************************/

static float maxn(float *flp, int n)
{	/* This returns the max of the first n elements of the vector flp */
	register int	i;
	register float	*xu;
	float		retval;

	retval = *flp;
	for (i = n, xu = flp; i--; xu++) if (*xu > retval) retval = *xu;
	return retval;
}

static void sum_cutoff(FILE *fp, FILE *fpl, int cuflags,
 double resma, double resmi)
{ /* This summarizes a run of `cutoff' to the stream `fp' */
	static char	*flist[] = {"multiref", "mrefout", "xlog", NULL};
	static char	ftype[] = {1, 2, 2, 0 };

	ftype[2] = (fpl == stdout) ? 0 : 2;
	header(fp, REDUCTION, flist, ftype);
	if (cuflags & CUT_COMPLEMENT) fprintf(fp,
 "\tReflections with D >%8.3f A or D <%8.3f A will be written out unmodified\n",
		resma, resmi);
	else fprintf(fp,
 "\tReflections with %8.3f A < D < %8.3f A will be written out unmodified\n",
		resma, resmi);
	if (cuflags & CUT_FLAGONLY) fprintf(fp,
 "\tReflections violating the limits will have their shift-values flagged\n");
	else fprintf(fp,
 "\tReflections violating the limits will be deleted altogether\n");
}

static void incr_cutcount(MERGEDREF *mrp, CUTCOUNT *ccp)
{ /*	This increments the appropriate elements in the structure `ccp'. */
	(ccp->cu_nmer)++;
	ccp->cu_nobs += mrp->mr_nob;
	ccp->cu_nnomer += 1 + (mrp->mr_nobpri && mrp->mr_nobbij);
}

static void cut_results(FILE *fp, int cuflags, int reerr, CUTCOUNT *cutw)
{ /*	This summarizes the results of a run of `cutoff' to the stream `fp' */
	fprintf(fp,"\t\tResults of the run of `cutoff':\n");
	fprintf(fp,
 "Category        # unique    # Bijvoet-unmerged               #\n");
	fprintf(fp,
 "             reflections           reflections    observations\n");
	fprintf(fp,
 "input             %6d                %6d          %6d\n",
	 cutw->cu_nmer, cutw->cu_nnomer, cutw->cu_nobs);
	if (cuflags & CUT_COMPLEMENT) fprintf(fp,
 "stl  illegal      %6d                %6d          %6d\n",
	 (cutw + 1)->cu_nmer, (cutw + 1)->cu_nnomer, (cutw + 1)->cu_nobs);
	else {
		fprintf(fp,
 "stl too high      %6d                %6d          %6d\n",
	 (cutw + 2)->cu_nmer, (cutw + 2)->cu_nnomer, (cutw + 2)->cu_nobs);
		fprintf(fp,
 "stl too low       %6d                %6d          %6d\n",
	 (cutw + 3)->cu_nmer, (cutw + 3)->cu_nnomer, (cutw + 3)->cu_nobs);
	      }
	fprintf(fp,
 "output            %6d                %6d          %6d\n",
	 (cutw + 1)->cu_nmer, (cutw + 1)->cu_nnomer, (cutw + 1)->cu_nobs);
	if (reerr)
	  {
		fprintf(fp,
 "\tWARNING!! The header of the output multiref contains the uncorrected\n");
		fprintf(fp,
 "\t\t\tcounts of reflections and observations\n");
	  }
}

int cutoffmain(int argc, char *argv[])
{ /*	This reads an input multiref and writes out an output multiref
	identical to it except that the shift values for all reflections
	with sin(theta)/lambda outside a user-specified range are set to
	the negative of their input absolute value.  This has the effect
	of deleting those observations from the data set. */
	register int	i, iwe;
	int		cutoflags, rewriterr, skipit;
	char		**flags, **dcp, outname[80];
	double		maxlen, maxssq, resmax, resmin, stlbot, stltop;
	FILE		*fplog, *fpmi, *fpmo;
	CRYSTAL		xhd;
	DATARUN		rhd;
	MERGEDREF	mr;
	ABBROB		*abp;
	CUTCOUNT	*cutw;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	resmax = 0.;	cutoflags = 0;
	flags = flagproc(argc, argv, "CDF");
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	   switch (**dcp)
	     {
		case	'C':	cutoflags |= CUT_COMPLEMENT;	break;
		case	'D':	cutoflags |= CUT_FLAGONLY;	break;
		case	'F':	cutoflags |= CUT_FLAGONLY;	break;
	     }
	for (argc--, argv++; (argc) && !allnumeric(*argv); argc--) argv++;
	if ((!argc--) || (0. > (resmax = atof(*argv++))))
	  {
		fprintf(stderr,"Usage: cutoff [-cd] resmax [resmin]\n");
		return EOK;
	  }
	resmin = 10000.;
	if (argc) resmin = atof(*argv);
	if (resmax > resmin)	/* swap resmax, resmin if necessary */
	  {
		maxlen = resmin; resmin = resmax; resmax = maxlen;
	  }
	if (qerror((AVOID *)(fpmi = xgopen("multiref","r")),
		"open input multiref")) return EOPENERR;
	if (verror(0 < xgread((void *)&xhd, XG_CRYSTAL, fpmi),
		"read crystal header from input multiref")) return EREADERR;
	maxlen = maxn(xhd.xt_ar,3);
	if (resmax < xhd.xt_reslim && resmin > maxlen)
	  {
		fprintf(stderr,
	 "No output created: those limits wouldn't exclude anything\n");
		return EOK;
	  }
	stltop = (resmax < xhd.xt_reslim ? 0.5 / xhd.xt_reslim :
		0.5 / resmax);
	stlbot = (resmin < maxlen ? 0.5 / resmin : 0.5 / maxlen);
	maxssq = 4. * stltop * stltop;
	strcpy(outname, trname("mrefout"));
	if (qerror((AVOID *)(fpmo = xgopen(outname, "w")),
	 "open output multiref")) return EOPENERR;
	if (resmax > xhd.xt_reslim) xhd.xt_reslim = resmax;
	if (verror(0 < xgwrite((void *)&xhd, XG_CRYSTAL, fpmo),
		"write crystal header to output multiref")) return EWRITERR;
	if (NULL == (cutw = (CUTCOUNT *)callop(4, sizeof (CUTCOUNT),
	 "cutw in cutoffmain"))) return EALLOC;
	for (i = xhd.xt_nruns; i--; )
	   {
		if (verror(0 < xgread((void *)&rhd, XG_DATARUN, fpmi),
		 "read datarun header from input multiref")) return EREADERR;
		if (resmax > rhd.run_reslim)
		  {
			rhd.run_reslim = resmax; rhd.run_ssqmax = maxssq;
		  }
		if (verror(0 < xgwrite((void *)&rhd, XG_DATARUN, fpmo),
		 "write datarun header to output multiref")) return EWRITERR;
	   }
	if (NULL == (fplog = propen("xlog"))) fplog = stdout;
	sum_cutoff(stdout, fplog, cutoflags, resmax, resmin);
	if (fplog != stdout)
		sum_cutoff(fplog, fplog, cutoflags, resmax, resmin);
	while (0 < xgread((void *)&mr, XG_MERGEDREF, fpmi))
	     {
		incr_cutcount(&mr, cutw);	iwe = 0;
		if (cutoflags & CUT_COMPLEMENT)
			skipit = (stlbot < mr.mr_stl && mr.mr_stl < stltop);
		else if (stlbot > mr.mr_stl) skipit = 1;
		else if (mr.mr_stl > stltop) skipit = 2;
		else skipit = 0;
		if (!skipit)
		  {
			incr_cutcount(&mr, cutw + 1);
			iwe = xgwrite((void *)&mr, XG_MERGEDREF, fpmo);
		  }
		else {
			if (skipit == 1) incr_cutcount(&mr, cutw + 3);
			else incr_cutcount(&mr, cutw + 2);
			if (cutoflags & CUT_FLAGONLY)
			  {
				for (i = mr.mr_nob, abp = &(mr.mr_ob[0]); i--;
				 abp++) abp->abo_shift = -abs(abp->abo_shift);
				iwe = xgwrite((void *)&mr, XG_MERGEDREF, fpmo);
			  }
			else	iwe = 1;
		     }
		if (0 >= iwe)
		  {
			fprintf(fplog,"Error in writing [%d,%d,%d,%d] to %s\n",
				mr.mr_hkl[0], mr.mr_hkl[1],
				mr.mr_hkl[2], mr.mr_hkl[3], outname);
			if (stdout != fplog) (void)fclose(fplog);
			fplog = NULL;	return iwe;
		  }
	     }
	(void)fclose(fpmi);	fpmi = NULL;
	/* If we've actually been refraining from writing out the
	out-of-bounds reflections, rather than merely flagging, them,
	we clean up the multiref header and write it back out. */
	if (!(cutoflags & CUT_FLAGONLY))
	  {
		rewriterr = 0;
		if (EOF != fseek(fpmo, 0L, 0))
		  {
			xhd.xt_ninteg = (cutw + 1)->cu_nobs;
			xhd.xt_nrefls = (cutw + 1)->cu_nnomer;
			xhd.xt_nmerged = (cutw + 1)->cu_nmer;
			xhd.xt_reslim = 0.5 / stltop;
			if (0 >= xgwrite((void *)&xhd, XG_CRYSTAL, fpmo))
				rewriterr = 1;
		  }
		else rewriterr = 1;
	  }
	else rewriterr = 0;
	cut_results(stdout, cutoflags, rewriterr, cutw);
	if (fplog != stdout) cut_results(fplog, cutoflags, rewriterr, cutw);
	if (stdout != fplog) (void)fclose(fplog);
	if (NULL != fpmo) (void)fclose(fpmo);	fpmo = NULL; fplog = NULL;
	memzap((AVOID **)&cutw, "cutw in cutoffmain");
	globalkill();	return EOK;
}

/* end of cut_off functionality */

/* end of mrefutils.c */
