/* muutils.c -- utilities for manipulating MULISTs

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	27 Jun 2007: return rather than exiting from error conditions
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	28 Dec 2006: untwinmain: initialize lhq in all cases
	 6 Dec 2006: use outmu and muinterp in MUREFL I/O;
	  all hkl I/O uses shorts so |h,k,l| > 127 will work.
	 4 Dec 2006: add scatomu functionality and accompanying I/O routines
	24 Jun 2006: include xgpostint.h; that's where the structures live
	18 May 2006: merge {mutoshel,untwin,reindexmu}.c to create this
	31 Dec 2005: untwin, reindexmu: don't qsort with fewer than 2 elements
	12 Oct 2005: eliminate static globals; housekeeping
	28 Jun 2005: reindmumain(): typecast pointer difference upon printout
	24 May 2005: untwin: trivial reformatting
	13 May 2005: reindexmu and untwin: move from nogui to utilities
	28 Aug 2004: untwin: static globals replaced by UNTWINRUN struct elts
	27 Aug 2004: routine muproc() used only here declared static; cosmetics
	18 Aug 2004: reindexmu: globals converted to function args;
		call try_reindex()
	29 May 2004: move structure types to xgen.h;
		move output functionality to ../xglib/ascout.c;
		define Bijvoet indices a little more cleanly
	20 Apr 2004: mutoshel: add ANOMALOUS (-a) option,
		which outputs both F+ and F- or I+ and I-
	25 Jan 2004: convert most static globals to local variables
	21 Jan 2004: license message inserted and initial comments reordered
	 8 Jan 2004: untwin: static globals fpi, fplog replace extern globals
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	 5 Jan 2003: reorganized source-file header
	25 Mar 2002: removed MSI attachments
	 5 Oct 1994: mutoshel: use xgread to input OBSREF data
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	 5 May 1994: "mm31iru" rather than "mm31ir" called
	16 Mar 1994: untwin: a few I/O changes, global variables moved around
	 7 Dec 1993: initial X-GEN version of mutoshel, untwin, and reindexmu
	10 Sep 1990: (?) check for sign of transformation and tweak accordingly
	 5 Mar 1987: (?) initial version of reindexmu

   Purpose: various manipulations of MULIST files
	mutoshel:	convert from XGEN's MULIST format to SHELX input format
	untwin:		examine extent of merohedral twinning;
			currently it works only in point group 321.
	reindexmu:	read a MULIST and rewrite it with different hkl values,
			based on the contents of an ascii 3x4 matrix
   Cross references: ../postprocess/{makemu,mukb}.c, ../nogui/mutools.c,
	Acta Cryst. 36: 755 (1980)
*************************************************************************/
#include	"xgpostint.h"
#define		SQUAREFIRST	1
#define		BIJPATTERSON	2
#define		ANOMALOUS	4
#define		MU_DONT_USE	63
#define		MAX_FRAC	0.5
#define		DIFFREFL	(0.025)	/* different refl sin(theta)/lambda*/
#define		TWIN_INTERVAL	(0.005)
#define		MAX_INTERVALS	1000
#ifdef VMS
#include	"vsort.h"
#endif

 typedef struct {
		int	mus_batchno, mus_flags;
		short	mus_hkl[3];
		double	mus_iscale;
		FILE	*mus_fpo;
		} MUTOSHELWORK;

 typedef struct {
		int		*ut_nneg;
		unsigned	ut_nrefls, ut_nfound, ut_nnotfound;
		double		ut_interval, ut_fraction;
		} UNTWINRUN;

extern int mutoshelmain(int argc, char *argv[]);
extern int untwinmain(int argc, char *argv[]);
extern int reindmumain(int argc, char *argv[]);
extern int sca2mumain(int argc, char *argv[]);

/************************************************
	scatomu: read a SCALEPACK output file and write a MULIST.
************************************************/

static short spstrtoi(char *spstr)
{
/* This returns the International Tables number associated with one of
 the lower-case spacegroup strings */
	char	*cp;
	short	si;

	for (si = 1; si < 256; si++)
	   {
		if (NULL != (cp = intltolc(si)))
		  {
			if (!strcmp(cp, spstr)) return si;
		  }
	  }
	return 0;
}

static void calc_flag(MUREFL *inmu)
{
	/* Return the flag value associated with a reflection */
	int	lflag;

	lflag = inmu->mu_flag;
	if (inmu->mu_f <= FUZZ) lflag |= 010;
	if (!(lflag & 01))
	  {
		if (inmu->mu_fplus < FUZZ) lflag |= 020;
		if (inmu->mu_fminus < FUZZ) lflag |= 040;
	  }
	inmu->mu_flag = lflag;
}

int sca2mumain(int argc, char *argv[])
{
	int		i, iscen, maxnelts, nacen, ncen, nelts, nrefls;
	u_char		hp[4];
	short		intlno, hhkl[3];
	double		imi, iplu, lstl, sigmi, sigplu, sm2, sp2, swt;
	float		iabc[3], iang[3], inmat[3][3], inrec[3][3];
	float		locs[3], rh[3];
	FILE		*fplog, *fpi, *fpmuo;
	static char	scaline[140], spgrpc[40];
	static char	*fnames[] = { "XSCA", "MULIST", "XLOG", NULL } ;
	static char	ftypes[] = { 1, 2, 2, 0 } ;
	static MUREFL	mui;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
		/* open the input SCALEPACK input file */
	if (qerror((AVOID *)(fpi = xgopen("XSCA", "r")),
	 "open input SCALEPACK file")) return EOPENERR;
	if (qerror((AVOID *)(fpmuo = xgopen("MULIST", "w")),
	 "open output MULIST")) return EOPENERR;
	if (NULL == (fplog = propen("xlog"))) fplog = stdout;
	header(stdout, REDUCTION, fnames, ftypes);
	header(fplog , REDUCTION, fnames, ftypes);
		/* process the flag arguments */
	/* decipher header of .sca or .asc file, e.g.
    1
 -985
   111.283   111.283   176.547    90.000    90.000   120.000 p3212
	*/
	for (i = 0; i < 3; i++)
	   {
		if (NULL == fgets(scaline, 120, fpi))
		  {
			fprintf(stderr, "headerline %d unavailable\n", i);
			return EREADERR;
		  }
	   }
	/* interpret line #3 as a,b,c,alpha,beta,gamma, spacegroup. */
	zeromem((AVOID *)spgrpc, 40);
	nelts = sscanf(scaline, "%f %f %f %f %f %f %s", 
	 &iabc[0], &iabc[1], &iabc[2],
	 &iang[0], &iang[1], &iang[2], (char *)spgrpc);
	if (nelts < 7)
	  {
		fprintf(stderr, "Unit cell line of .sca file unreadable\n");
		return EREADERR;
	  }
	intlno = spstrtoi(spgrpc);
	if (EOK != (i = vecmat(iabc, iang, inmat)))
	  {
		fprintf(stderr, "Failure %d in vecmat\n", i);
		return EBAD;
	  }
	if (EOK != (i = inv3(inmat, inrec)))
	  {
		fprintf(stderr, "Failure %d in inv3\n", i);
		return EBAD;
	  }
	/* go through reflection data to see if there are
	 anomalous data */
	maxnelts = 0;
	for (nrefls = 0; NULL != fgets(scaline, 120, fpi); nrefls++)
	   {
		nelts = sscanf(scaline, "%hd %hd %hd %lf %lf %lf %lf",
			 &hhkl[0], &hhkl[1], &hhkl[2],
			 &iplu, &sigplu, &imi, &sigmi);
		if (nelts > maxnelts) maxnelts = nelts;
		if (maxnelts > 5) break;
	   }
	if (-1 == fseek(fpi, 0L, 0))
	  {
		fprintf(stderr, "Unable to rewind input file\n");
		return EREADERR;
	  }
	for (i = 0; i < 3; i++)
	   {
		if (NULL == fgets(scaline, 120, fpi))
		  {
			fprintf(stderr, "headerline %d unavailable\n", i);
			return EREADERR;
		  }
	   }
		/* read actual reflection data and convert them:
   1   1   7  4062.8   162.6  4062.0   121.3 */
	ncen = nacen = 0;
	if (maxnelts > 5)
	  {
		for (nrefls = 0; NULL != fgets(scaline, 120, fpi); nrefls++)
		   {
			nelts = sscanf(scaline, "%hd %hd %hd %lf %lf %lf %lf",
			 &hhkl[0], &hhkl[1], &hhkl[2],
			 &iplu, &sigplu, &imi, &sigmi);
			if (nelts < 5)
			  {
				fprintf(stderr, "Only %d elements in line %d\n",
					nelts, nrefls);
				return EREADERR;
			  }
			for (i = 0; i < 3; i++) rh[i] = mui.mu_hkl[i] = hhkl[i];
			mpack(hhkl, hp, intlno);
			mui.mu_flag = iscen = iscentric(hp, intlno);
			if (iscen) ncen++;	else nacen++;
			if (sigplu < -0.01) /* I+ not measured */
			  {
				mui.mu_flag |= 02;
				iplu = imi;	sigplu = sigmi;
			  }
			if (sigmi  < -0.01) /* I- not measured */
			  {
				mui.mu_flag |= 04;
				imi = iplu;	sigmi = sigplu;
			  }
			if (sigplu < 0.002) sigplu = 0.002;
			if (sigmi < 0.002) sigmi = 0.002;
			/* any bits in mu_flag imply a need to collapse them*/
			if ((nelts < 7) || (mui.mu_flag))
			  { /* only I,sigma given: collapse it */
				mui.mu_f = mui.mu_fplus = mui.mu_fminus = iplu;
				mui.mu_sig = mui.mu_sigplus = mui.mu_sigminus =
					sigplu;
			  }
			else { /* both I+, I- given */
				mui.mu_fplus = iplu;
				mui.mu_sigplus = sigplu;
				mui.mu_fminus = imi;
				mui.mu_sigminus = sigmi;
				sp2 = sigplu * sigplu;	sm2 = sigmi * sigmi;
				swt = 1. / sp2 + 1. / sm2;
				mui.mu_f = ((iplu / sp2) + (imi / sm2)) / swt;
				mui.mu_sig = 1. / sqrt(swt);
			     }
			mm31(locs, inrec, rh);
			lstl = locs[0] * locs[0] + locs[1] * locs[1] +
				locs[2] * locs[2];
			mui.mu_stl = (lstl < DFUZZ) ? 0. : 0.5 * sqrt(lstl);
			calc_flag(&mui);
			(void)outmu(fpmuo, &mui);
		   } /* end of loop through reflections */
	   } /* end of "if (maxnelts > 5) ... " clause */
	else {
		for (nrefls = 0; NULL != fgets(scaline, 120, fpi); nrefls++)
		   {
			nelts = sscanf(scaline, "%hd %hd %hd %lf %lf",
			 &hhkl[0], &hhkl[1], &hhkl[2], &iplu, &sigplu);
			if (nelts < 5)
			  {
				fprintf(stderr,
				 "Only %d elements in line %d\n",
					nelts, nrefls);
				return EREADERR;
			  }
			for (i = 0; i < 3; i++) rh[i] = mui.mu_hkl[i] = hhkl[i];
			mpack(hhkl, hp, intlno);
			mui.mu_flag = iscen = iscentric(hp, intlno);
			if (iscen) ncen++;	else nacen++;
			if (sigplu < 0.002) sigplu = 0.002;
			mui.mu_f = mui.mu_fplus = mui.mu_fminus = iplu;
			mui.mu_sig = mui.mu_sigplus = mui.mu_sigminus = sigplu;
			mm31(locs, inrec, rh);
			lstl = locs[0] * locs[0] + locs[1] * locs[1] +
				locs[2] * locs[2];
			mui.mu_stl = (lstl < DFUZZ) ? 0. : sqrt(lstl);
			calc_flag(&mui);
			(void)outmu(fpmuo, &mui);
		   } /* end of loop through reflections */
	     } /* end of else of "if (maxnelts > 5" ... stuff */
	fprintf(stdout, "%6d refls converted: %6d acentric, %6d centric\n",
	 nrefls, nacen, ncen);
	fprintf(fplog, "%6d refls converted: %6d acentric, %6d centric\n",
	 nrefls, nacen, ncen);
	(void)fclose(fpi);	(void)fclose(fpmuo);	(void)fclose(fplog);
	fpi = fpmuo = fplog = NULL;
	globalkill();	return EOK;
}

/* end of scatomu functionality */

/************************************************
	mutoshel: read MULIST, output Sheldrick's "hkl" format
	The output format for each reflection is given in
	printshel() in ../xglib/ascout.c
	Syntax:
		mutoshel [-flags] [batchno]
	Flags:
	 -a	Output both F+ and F- (or I+ and I-)
	 -b	Bijvoet Patterson output
	 -f	square the input F's before output.
		(Sheldrick expects intensities, so if the mulist
		has intensities, you _don't_ need to square them;
		If the mulist has F's, you do.)
	`batchno' is set to 1 if the user doesn't specify it.
	The program reads a MULIST file with logical name or environment
	 variable MULIST; it writes to the logical name or env. vbl. XHKL.
	The data are scanned quickly before the real calculations begin to
	determine the largest intensity value in the data; if it's > 9999,
	the whole set is rescaled so that the maximum intensity is in fact 9999.
************************************************/

static void muproc(MUTOSHELWORK *msw, MUREFL *mup)
{ /* Processes a MULIST reflection to convert it to SHELX format */
	int		i;
	static SHEL_HKL	xh, xhm;

	zeromem((void *)&xh, sizeof (SHEL_HKL));
	for (i = 3; i--; )
	   {
		xh.x_hkl[i] = mup->mu_hkl[i]; xhm.x_hkl[i] = -(xh.x_hkl[i]);
	   }
	xhm.x_batch = xh.x_batch = msw->mus_batchno;
	if (msw->mus_flags & SQUAREFIRST)
	  {
		mup->mu_sig *= 2. * mup->mu_f; mup->mu_f *= mup->mu_f;
		mup->mu_sigplus *= 2. * mup->mu_fplus;
		mup->mu_fplus *= mup->mu_fplus;
		mup->mu_sigminus *= 2. * mup->mu_fminus;
		mup->mu_fminus *= mup->mu_fminus;
	  }
	if (msw->mus_flags & BIJPATTERSON)
	  {
		if (mup->mu_flag & MU_DONT_USE) return;
		xh.x_inten = mup->mu_fplus * msw->mus_iscale;
		xh.x_sigma = mup->mu_sigplus * msw->mus_iscale;
		printshel(msw->mus_fpo, &xh);
		xhm.x_inten = mup->mu_fminus * msw->mus_iscale;
		xhm.x_sigma = mup->mu_sigminus * msw->mus_iscale;
		printshel(msw->mus_fpo, &xhm);
	  }
	else if ((msw->mus_flags & ANOMALOUS) && !(mup->mu_flag))
	  {
		xh.x_inten = mup->mu_fplus * msw->mus_iscale;
		xh.x_sigma = mup->mu_sigplus * msw->mus_iscale;
		printshel(msw->mus_fpo, &xh);
		xhm.x_inten = mup->mu_fminus * msw->mus_iscale;
		xhm.x_sigma = mup->mu_sigminus * msw->mus_iscale;
		printshel(msw->mus_fpo, &xhm);
	  }
	else {
		xh.x_inten = mup->mu_f * msw->mus_iscale;
		xh.x_sigma = mup->mu_sig * msw->mus_iscale;
		printshel(msw->mus_fpo, &xh);
	     }
}

int mutoshelmain(int argc, char *argv[])
{ /*	This is the mainline for "mutoshel".  It reads in the reflections
	once to find the minimum omega and maximum intensity; then it goes
	through for real for processing.	*/
	char			**dcp, **flags, muline[122];
	int			j, nrefls, thisf;
	double			maxinten;
	FILE			*fpi;
	static MUREFL		mui;
 	static MUTOSHELWORK	msws;
	static char	*fnames[] = { "MULIST", "XSHEL", NULL } ;
	static char	ftypes[] = { 1, 2, 0 } ;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
		/* open the input MULIST file */
	if (qerror((AVOID *)(fpi = xgopen("MULIST", "r")),
	 "open input MULIST file")) return EOPENERR;
	header(stdout, REDUCTION, fnames, ftypes);
		/* process the flag arguments */
	flags = flagproc(argc, argv, "ABFP");
	msws.mus_flags = 0;
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	   switch (**dcp)
	     {
		case	'F':	msws.mus_flags |= SQUAREFIRST;	break;
		case	'B':	case	'P':
			msws.mus_flags |= BIJPATTERSON;	break;
		case	'A':	msws.mus_flags |= ANOMALOUS;	break;	
		default:	break;
	     }
		/* skip the flags to get to the numeric argument, if any */
	for (argc--, argc++; (argc) && !allnumeric(*argv); argc--) argv++;
	if (!argc ) msws.mus_batchno = 0;
	else msws.mus_batchno = atoi(*argv);
	printf(" batch number = %d\n", msws.mus_batchno);
	printf(" Input values will%sbe squared before output\n",
		(msws.mus_flags & SQUAREFIRST) ? " " : " NOT ");
		/* set silly value for the max intensity */
	maxinten = -1.e30;
		/* go through the data and set proper values for the above */
	for (nrefls = 0; NULL != fgets(muline, 120, fpi); nrefls++)
	   {
		(void)muinterp(muline, &mui);
		if (msws.mus_flags & SQUAREFIRST)
		  {
			mui.mu_f *= mui.mu_f; mui.mu_fplus *= mui.mu_fplus;
			mui.mu_fminus *= mui.mu_fminus;
		  }
		thisf = mui.mu_f;
		if (msws.mus_flags & BIJPATTERSON)
		  {
			thisf = mui.mu_fplus;
			if (mui.mu_fminus > thisf) thisf = mui.mu_fminus;
		  }
		if (thisf > maxinten)
		  {
			maxinten = thisf;
			for (j = 3; j--; ) msws.mus_hkl[j] = mui.mu_hkl[j];
		  }
	   }
	/* if max intensity > 9999, use it as a rescaler; otherwise ignore it */
	msws.mus_iscale = (maxinten > 9999.) ? 9999. / maxinten : 1.;
		/* rewind the input file */
	if (verror (!fseek(fpi, 0L, 0), "rewind input MULIST file"))
		return ESEEKERR;
		/* open the output XHKL file */
	if (qerror((AVOID *)(msws.mus_fpo = xgopen("XHKL", "w")),
		"open output XHKL file")) return EOPENERR;
	printf(
	 "Maximum intensity = %11.2f for [%3d,%3d,%3d]\n",
		maxinten, msws.mus_hkl[0], msws.mus_hkl[1], msws.mus_hkl[2]);
	printf("Overall scale factor = %8.4f\n", msws.mus_iscale);
	printf("%6d reflections read in\n", nrefls);
		/* go back through the data and do the real work */
	for (nrefls = 0; NULL != fgets(muline, 120, fpi); nrefls++)
	   {
		(void)muinterp(muline, &mui);
		muproc(&msws, &mui);
	   }
	(void)fclose(fpi);	(void)fclose(msws.mus_fpo);
	fpi = msws.mus_fpo = NULL;
	printf("%6d reflections processed\n", nrefls);
	globalkill();	return EOK;
}

/* end of mutoshel functionality */

/************************************
	untwin:	determine extent of merohedral twinning in a MULIST
	A program to calculate the twinning percentage on an allegedly
	merohedrally twinned dataset.  It currently works in point group 321,
	where the twins are (hkl) and (khl).
	Thus we're calculating the percentage of negative reflections for
	which J(hkl) = [ I(hkl) * (1-x) + I(khl) * x ] / (1 - 2 * x)
	is negative and looking for the point where the curve takes off.
	See Francis Tsui's dissertation and R.G.Fisher & R.M.Sweet,
	Acta Cryst. 36: 755 (1980).
	The mulist may come in in any order but it's assumed to
	have only h >=0, k>=0, l >=0.  We sort it on sin(theta) so we never
	have to look very far around a reflection to find its mate.
*************************************************************************/

static void sum_frac(FILE *fp, UNTWINRUN *utp)
{ /* Summarize, to the stream fp, the fraction twinning found */
	register int	itw;
	double		ftw;

	fprintf(fp,
	 " %6d reflections found: %6d successfully paired, %6d not paired\n",
		(int)(utp->ut_nrefls), (int)(utp->ut_nfound),
		(int)(utp->ut_nnotfound));
	fprintf(fp,
		"Fraction Twinned   Fraction Negative Refls\n");
	for (ftw = itw = 0; ftw < utp->ut_fraction;
	 itw++, ftw += utp->ut_interval)
	   fprintf(fp, "         %7.4f                %9.5f\n",
		ftw, ((double)(utp->ut_nneg[itw])) / (double)(utp->ut_nfound));
}

int untwinmain(int argc, char *argv[])
{ /* The mainline for the "untwin" operation */
	register MUREFL	*muend, *mup, *muq;
	int		itw;
	unsigned	nintervals;
	char		found, uline[90];
	short		kh, kk, kl, *lhp, *lhq;
	double		fh, fj, fsum, ftw, ip, iq, kstl, maxfj, minfj;
	double		sdif, spec_fraction, spec_interval;
	FILE		*fpi, *fplog;
	MUREFL		*mudata;
	UNTWINRUN	*utp;
	static char	*fnames[] = {"mulist", "xlog", NULL};
	static char	ftypes[] = {1, 2, 0 };
#ifdef VMS
	static u_short	stlbuf[] = {1,
			DSC$K_DTYPE_F, SO_ASCENDING, 6, sizeof (float) };
#endif

	/* This resorts on sin(theta) in point group 321.  It then calculates
	 the fraction of negative intensities obtained from
		J(hkl) = (I(hkl) * (1-x) + I(khl) * x) / (1 - 2 * x)
	as a function of x. */
	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort*/
	if (NULL == (utp = (UNTWINRUN *)callop(1, sizeof (UNTWINRUN),
	 "utp in untwinmain"))) return EALLOC;
	utp->ut_interval = TWIN_INTERVAL;	utp->ut_fraction = MAX_FRAC;
	if (qerror((AVOID *)(fpi = xgopen("mulist", "r")),
		"open input mulist")) return EOPENERR;
	if (argc > 1 && (DFUZZ < (spec_interval = atof(argv[1]))))
	  {
		utp->ut_interval = spec_interval;
		if (argc > 2 && (utp->ut_interval < (spec_fraction =
			atof(argv[2])))) utp->ut_fraction = spec_fraction;
	  }
	nintervals = (unsigned)(utp->ut_fraction / utp->ut_interval + 2);
	if (verror((nintervals < MAX_INTERVALS),
	 "analyze that many intervals at once")) return EBADVALUE;
	if (NULL == (AVOID *)(utp->ut_nneg = (int *)callop(nintervals,
		sizeof (int), "utp->ut_nneg in untwin"))) return EALLOC;
	if (NULL == (fplog = propen("xlog"))) fplog = stdout;
	ftypes[1] = (fplog == stdout ? 0 : 2);
	header(stdout, REDUCTION, fnames, ftypes);
	printf(
 "Fraction with I<0 calculated at intervals of %8.5f from %8.3f to %8.3f\n",
		utp->ut_interval, 0., utp->ut_fraction);
	if (fplog != stdout)
	  {
		header(fplog,  REDUCTION, fnames, ftypes);
		fprintf(fplog,
 "Fraction with I<0 calculated at intervals of %8.5f from %8.3f to %8.3f\n",
			utp->ut_interval, 0., utp->ut_fraction);
	  }
	utp->ut_nfound = utp->ut_nnotfound = 0;
	for (utp->ut_nrefls = 0; NULL != fgets(uline, 82, fpi); )
		(utp->ut_nrefls)++;
	printf("%d lines detected in the input mulist\n",
		(int)(utp->ut_nrefls));
	rewind(fpi);
	if (NULL == (AVOID *)(mudata = (MUREFL *)callop(utp->ut_nrefls + 2,
	 sizeof (MUREFL), "mudata in untwin")))
	  {
		memzap((AVOID **)&(utp->ut_nneg), "utp->ut_nneg in untwin");
		return EALLOC;
	  }
	for (mup = mudata; NULL != fgets(uline, 82, fpi); mup++)
		if (11 > muinterp(uline, mup)) break;	
	(void)fclose(fpi);	fpi = NULL;
	muend = mup;
	printf("%d reflections read in and fixed up\n", (int)(mup - mudata));
	if (mup > mudata) qsort((AVOID *)mudata,(unsigned)(mup - mudata),
	 sizeof (MUREFL), (QPE)stlcmp);
	printf(" Sort completed\n");
	minfj = 10000.;	maxfj = -minfj;
	for (mup = mudata, utp->ut_nfound = 0; mup < muend; mup++)
	   {
		lhp = mup->mu_hkl;
		if ((*lhp > *(lhp+1)) && (*(lhp+2) > 0))
		  {
			kh = *(lhp + 1);	kk = *lhp;
			kl = *(lhp + 2);	kstl = mup->mu_stl;
			for (muq = mup - 1, found = 0; muq >= mudata; muq--)
			   {
				lhq = muq->mu_hkl;
				if ((kh == *lhq) && (kk == *(lhq + 1)) &&
				  (kl == *(lhq + 2)))
				  {
					found = 1; break;
				  }
				sdif = kstl - muq->mu_stl;
				if (sdif > DIFFREFL) break;
			   }
			if (!found)
			  for (muq = mup + 1; muq < muend; muq++)
			   {
				lhq = muq->mu_hkl;
				if ((kh == *lhq) && (kk == *(lhq + 1)) &&
				 (kl == *lhq + 2))
				  {
					found = 1;	break;
				  }
				sdif = muq->mu_stl - kstl;
				if (sdif > DIFFREFL) break;
			   }
			if (found)
			  {
				(utp->ut_nfound)++;
				ip = mup->mu_f * mup->mu_f;
				iq = muq->mu_f * muq->mu_f;
				lhq = muq->mu_hkl;
				fsum = iq + ip;
				fh = (ip > iq ? iq : ip);
				for (ftw = itw = 0; ftw < utp->ut_fraction;
				  itw++, ftw += utp->ut_interval)
				   {
					fj = (fh - ftw * fsum) /(1. - 2 * ftw);
					if (fj < 0.) (utp->ut_nneg[itw])++;
					if (fj < minfj) minfj = fj;
					if (fj > maxfj) maxfj = fj;
				   }
				if (!((utp->ut_nfound) % 800)) printf(
 " %hd %hd %hd %6.4f %7.2f +/- %7.2f; %hd %hd %hd %6.4f %7.2f +/- %7.2f\n",
				 *lhp, *(lhp + 1), *(lhp + 2),
				 mup->mu_stl, mup->mu_f, mup->mu_sig,
				 *lhq, *(lhq + 1), *(lhq + 2),
				 muq->mu_stl, muq->mu_f, muq->mu_sig);
			  }	/* end of "if (found)" clause */
			else
			  {
				if (!((utp->ut_nnotfound) % 100)) printf(
				 "Mate to [%3hd %3hd %3hd] not found\n",
					*lhp, *(lhp + 1), *(lhp + 2));
				(utp->ut_nnotfound)++;
			  }
		  } /* end of "if (... > 0" clause */
	   } /* end of loop through reflections */
	printf("min, max corrected structure amplitudes = %9.3f, %9.3f\n",
		minfj, maxfj);
	sum_frac(stdout, utp);
	if (stdout != fplog)
	  {
		sum_frac(fplog, utp);	(void)fclose(fplog);
	  }
	memzap((AVOID **)&mudata, "mudata in untwin");
	memzap((AVOID **)&(utp->ut_nneg), "utp->ut_nneg in untwin");
	memzap((AVOID **)&utp, "utp in untwin");
	globalkill();	fplog = NULL; return EOK;
}

/* end of untwin functionality */

/****************************************
	reindexmu: reindex a MULIST with a rational 3x4 transformation matrix
	A program to read in, reindex, and resort a mulist according
	to a rotation matrix.
****************************************/

static int muccmp(MUREFL *mup0, MUREFL *mup1)
{ /*	Compare routine for qsort into multiref order */
	register int	diff, i;
	u_char		*phkl0, *phkl1;

	phkl0 = (u_char *)(&(mup0->mu_hkl[0]));
	phkl1 = (u_char *)(&(mup1->mu_hkl[0]));
	for (i = 0; i < 4; i++, phkl0++, phkl1++)
	   {
		if ((diff = ((int)(*phkl0) - (int)(*phkl1))) < 0) return -1;
		else if (diff > 0) return 1;
	   }
	return 0;
}

static void mm31iru(short *osv, float mat[3][3], short *isv)
{
	/*	This left-multiplies the 3x3 matrix
	`mat' by the vector `isv' to give osv. */
	register int	i, j;

	for (i = 3; i--; )
		for (j = 3, osv[i] = 0; j--; )
			osv[i] += (short)(mat[i][j] * (float)(isv[j]));
}

static void sumreimurun(FILE *fp, char reflorder,
 short vecre[], float matre[3][3], double detma)
{ /*	This summarizes the run conditions for `reindexmu' to the stream fp. */
	fprintf(fp, "Reflections will be reindexed according to h' = Mh + v\n");
	fprintf(fp,"\t\t  %6.2lf %6.2lf %6.2lf\t\t%3d\n",
		matre[0][0], matre[0][1], matre[0][2], vecre[0]);
	fprintf(fp,"\twhere M = %6.2lf %6.2lf %6.2lf\t\t%3d\n",
		matre[1][0], matre[1][1], matre[1][2], vecre[1]);
	fprintf(fp,"\t\t  %6.2lf %6.2lf %6.2lf\t\t%3d\n",
		matre[2][0], matre[2][1], matre[2][2], vecre[2]);
	fprintf(fp, "Reflections will be output in ");
	switch (reflorder)
	      {
		case	0: fprintf(fp, "standard multiref order\n");	break;
		case	1: fprintf(fp, "absolute hkl order\n");		break;
		case	2: fprintf(fp, "sin(theta) order\n");		break;
	      }
	if (detma < -DFUZZ)
	  {
		fprintf(fp, "Transformation has determinant %7.2f;\n", detma);
		fprintf(fp, "Primary and Bijvoet columns will be swapped\n");
	  }
}

int reindmumain(int argc, char *argv[])
{ /*	Mainline for reindexing a MULIST */
	register int	j;
	register MUREFL	*mup, *muend;
	unsigned	nrefls;
	char		c, refl_order, tfl, rline[90];
	double		detmat;
	float		matrei[3][3];
	short		shkl[3], vecrei[3];
	FILE		*fpi, *fpo;
	static MUREFL	*mudata;
	static char	*fnames[] = {"mulist", "reindex", "muout", NULL};
	static char	ftypes[] = {1, 1, 2, 0};
#ifdef VMS
	static u_short	hklbuf[] = {3,
			DSC$K_DTYPE_B, SO_ASCENDING, 0, 1,
			DSC$K_DTYPE_B, SO_ASCENDING, 1, 1,
			DSC$K_DTYPE_B, SO_ASCENDING, 2, 1 };
	static u_short	stlbuf[] = {1,
			DSC$K_DTYPE_F, SO_ASCENDING, 4, 4 };
#endif

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	header(stdout, REDUCTION, fnames, ftypes);
	refl_order = 0;
	if (argc > 1)
	  {
		c = argv[1][1];	if ('a' <= c && c <= 'z') c -= ('a' - 'A');
		if ('A' == c) refl_order = 1;
		else if ('S' == c) refl_order = 2;
	  }
	if (qerror((AVOID *)(fpi = xgopen("mulist", "r")),
		"open input mulist")) return EOPENERR;
	if (1 != try_reindex(vecrei, matrei)) return EBADVALUE;
	if (fabs((detmat = deter(matrei))) < DFUZZ)
	  {
		fprintf(stderr,
		 "Warning: determinant of transformation matrix = %13.7e\n",
			detmat);
		fprintf(stderr, "This transformation is singular!\n");	
	  }
	sumreimurun(stdout, refl_order, vecrei, matrei, detmat);
	for (nrefls = 0; NULL != fgets(rline, 82, fpi); ) nrefls++;
	fprintf(stdout,"%d lines detected in the input mulist\n", (int)nrefls);
	rewind(fpi);
	if (NULL == (AVOID *)(mudata = (MUREFL *)callop(nrefls + 2,
	 sizeof (MUREFL), "mudata in reindexmu"))) return EALLOC;
	for (mup = mudata; NULL != fgets(rline, 82, fpi); mup++)
	   {
		if (11 > muinterp(rline, mup)) break;
		mm31iru(shkl, matrei, mup->mu_hkl);
		for (j = 3; j--; ) mup->mu_hkl[j] = shkl[j] + vecrei[j];
	   }
	(void)fclose(fpi);
	muend = mup;
	nrefls = mup - mudata;
	fprintf(stdout,"%d reflections read in and fixed up\n", (int)nrefls);
	if (nrefls > 1)
	  {
		if (refl_order == 0)
		  { /* sorted in multiref order: calculate 4-byte hkls */
			for (mup = mudata; mup < muend; mup++)
				mpack(mup->mu_hkl, mup->mu_uhkl, 1);
		 	qsort((AVOID *)mudata, nrefls, sizeof (MUREFL),
				(QPE)muccmp);
			for (mup = mudata; mup < muend; mup++)
				munpack(mup->mu_uhkl, mup->mu_hkl, 1);
		  }
		else if (refl_order == 1)	/* absolute hkl order */
		 qsort((AVOID *)mudata, nrefls, sizeof (MUREFL), (QPE)murcmp);
		else if (refl_order == 2)	/* sin(theta)/lambda order */
		 qsort((AVOID *)mudata, nrefls, sizeof (MUREFL), (QPE)stlcmp);
	  }
	fprintf(stdout," Sort completed\n");
	if (qerror((AVOID *)(fpo = xgopen("muout", "w")),
		"open output mulist"))
	  {
		memzap((AVOID **)&mudata, "mudata in reindexmu");
		return EOPENERR;
	  }
/* if the transformation is left handed, reverse F+ and F-.  We also need to
	tweak the flag value if F+ <= 0 and F- > 0 or if F+ > 0 and F- <= 0.*/
	if (detmat < -DFUZZ)
	  {
		for (mup = mudata; mup < muend; mup++)
		   {
			tfl = mup->mu_flag;
			if ((16 <= tfl) && (tfl < 32))
			  {
				tfl += 16;
				if (tfl & 04)
				  {
					tfl -= 4; tfl |= 02;
				  }
			  }
			else if ((32 <= tfl) && (tfl < 48))
			  {
				tfl -= 16;
				if (tfl & 02)
				  {
					tfl -= 2; tfl |= 04;
				  }
			  }
			mup->mu_flag = tfl;	(void)outmu(fpo, mup);
		   }
	  }
	else
	  {
		for (mup = mudata; mup < muend; mup++)
			(void)outmu(fpo, mup);
	  }
	(void)fclose(fpo);
	fprintf(stdout,"%d sorted reflections written out\n",
	 (int)(muend - mudata));
	memzap((AVOID **)&mudata, "mudata in reindexmu");
	globalkill();	return EOK;
}

/* end of reindexmu functionality */

/* end of muutils.c */
