/* parutils.c -- utilities associated with parameters in a UPARAMS file

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	14 Feb 2007: bestg, matrix functionalities added
	28 Nov 2006: artocel call updated in newrunmain;
	 ucell call in reindumain updated
	16 Aug 2006: bugfix in predmain so resolution minima work
	19 May 2006: merge {newrun,reindexu,predictit}.c here
	31 Mar 2006: reindexu: bugfix in handling tm[][]
	31 Dec 2005: predictit: don't qsort if fewer than 2 elements
	12 Oct 2005: eliminate static global from predictit; housekeeping
	30 Aug 2005: predmain: don't report sorting times
	13 May 2005: move reindexu and predictit from nogui to utilities
	11 May 2005: check return status from xgjournal
	11 Apr 2005: namechange: predict.c -> predictit.c to prevent overlap
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	16 Jan 2005: predict: reverse order of resolution limits if necessary
	15 Oct 2004: newrun: source moved from ../nogui to ../utilities
	23 Sep 2004: predict: initialize mu2 to keep the compiler happy
	 9 Sep 2004: merge predsetup() into predmain(); elim po_y2co, etc.;
		local variables for mu1, mu2, deltamu, nranges
	 6 Sep 2004: most functions for prediction in ../xglib/xgpredict.c;
		predsetup() wrapper for prpar() written
	 6 Sep 2004: newrun: cosmetic change
	 2 Aug 2004: predict: cosmetic change
	27 Jul 2004: predict: correct references to frame size
	21 Jun 2004: predict: increase REFLSPERRANGE up to 480000;
		move pdetec_limit() and presval() to ../xglib/xgenmisc.c
	 1 Apr 2004: predict: set file-pointer fpo to NULL after closing it
	30 Jan 2004: newrun: old and new run start values are in deg, not rad;
		fix inconsistencies that had made this program useless
	25 Jan 2004: convert static globals to locals; eliminate
	 xgopen of UPARAMS because that's latent in refparin(); cosmetics
	21 Jan 2004: license message inserted and initial comments reordered
	 9 Jan 2004: static globals (including static REFPARS struct) replace
		extern globals
	 6 Jan 2004: static REFPARS structure elements replace extern globals;
		refparin() and makeq21() used
	10 Dec 2003: predict: eliminate references to framdeg
	27 Nov 2003: replace deltaom, deltachi, deltaphi with rphi[] elements
	25 Nov 2003: predict: get rid of tansw as a global
	20 Sep 2003: correct spelling of "transform" in diagnostic message
	19 Sep 2003: hm, rm arrays initializations with double braces (ansi C)
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h;
		initializations of arrays performed according to ansi standard
	16 Sep 2003: predict: access call for amask file is eliminated,
		because it's subsumed into the get_actmask call
	16 Aug 2003: predict: handle exclusion codes differently so we don't
		have to carry around a global variable for them
	 5 Jan 2003: reorganized source-file header
	15 Nov 2002: extern declarations made here rather than in refiext.h
	15 Nov 2002: predict: trivial reformatting of diagnostic-output code
	25 Mar 2002: removed MSI attachments
	15 Dec 2000: in pdetec_limit(), correct for translational effects.
	 5 Oct 1994: use xgread and xgwrite for REFPARS and PREDREF data
	30 Sep 1994: fix list of input files
	 2 Sep 1994: reindexu: count bytes in UPARAMS file upon reading it in
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: a few I/O changes and global variables moved around
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version
	18 Jan 1988: in predsetup(), rotate the orientation matrix from (0,0,0)
		to the actual starting position of the run.

   Purpose: manipulations of UPARAMS files and the REFPARS structures in them.
	newrun	reads a UPARAMS and writing out one with altered
	 physical and pseudo-goniostat omega, chi, phi values.
	reindexu	converts a UPARAMS file to another using a 3x4
	 re-indexing matrix.  Unit cell and pseudo-goniostat angles are
	 updated along with the the reciprocal- and real-space unit cell
	 matrices and parameters.
	predictit	produce a list of predicted reflections over a
	 specified range of images
   Cross references: ../xgrefine/ref*.c, ../utilities/redurefl.c,
	../postprocess/make_mu.c

*************************************************************************/
#define		NEED_TIME	1
#include	"refidefs.h"
#include	"xgenpred.h"
#ifdef VMS
#include	"vsort.h"
#endif
#define		HEXAGONAL	6
#define		RHOMBOHEDRAL	7
#define		MAXRANGE	(30 * RADC)	/* longest uninterruped range*/
#define		DEFAULT_MINY2SQ	(0.9e-3)	/* minimum allowed value of
 the square of the y component of rotated scatt vector == 1/Lorenz factor */
#define		UNPOL_FRAC	0.5
#define		RING_FRAC	1.0

extern int newrunmain(int argc, char *argv[]);
extern int reindumain(int argc, char *argv[]);
extern int predmain(int argc, char *argv[]);
extern int matrmain(int argc, char *argv[]);
extern int bestgmain(int argc, char *argv[]);

/*********************************************
	bestg: Find pseudo-goniostat omega,chi,phi
	based on a,b,c,alpha,beta,gamma, etc.
*********************************************/

static int lgettroid(FILE *fp, REFPARS *xp, REFREF *rp)
{ /*	version of scattv for this purpose */
	int	i;
	double	jfr, lxi, lyi, lrni, lrni2, rsq;
	float	*rs, lq3[3][3], lss[3];

	zeromem((AVOID *)rp, sizeof (REFREF));
	if (0 >= xgread((void *)rp, XG_TROID, fp)) return EREADERR;
	jfr = rp->frame0 + 0.49999;
	if (1 > pix2cm(rp->xto, rp->yto, &(rp->xtcm), &(rp->ytcm), jfr))
		return EBADVALUE;
	rp->framem = rp->frame0;
	/* the log of the intensity is used in weighting the observations */
	rp->totinten = (rp->totinten < 3.) ?  1 : log(rp->totinten);
	lxi = rp->xtcm - xp->xx0;
	lyi = rp->ytcm - xp->xy0; /* correct main beam */
	lrni2 = lxi * lxi + lyi * lyi + xp->xpltd2;
	lrni = sqrt(lrni2);
	lss[0] = xp->xplated / lrni;
	lss[1] = lxi / lrni;
	lss[2] = lyi / lrni;
	/* rotate this by the detector tilt and swing-angle matrices */
	mm31(rp->rsl, xp->xq21, lss);
	rp->rsl[0] += 1.;	/* correct the beam-direction component */
	if (FUZZ > xp->xlambda) xp->xlambda = 1.5418;
	for (i = 3, rsq = 0., rs = rp->rsl; i--; rs++) rsq += *rs * *rs;
	if (rsq < DFUZZ) return EBADVALUE;
	rp->dspac = xp->xlambda / sqrt(rsq);
	rp->rsp = pirange(xp->xfrmsize * (rp->framem + 0.5));
	lq3[0][0] = lq3[1][1] = cos(rp->rsp);
	lq3[0][1] = -sin(rp->rsp);
	lq3[1][0] = -lq3[0][1];
	lq3[2][2] = 1.;
	lq3[0][2] = lq3[1][2] = lq3[2][0] = lq3[2][1] = 0.;
	rp->wt = 1.;	rp->maxdevh = 0.;	rp->rflags = OKREFL;
	mm31(rp->rs, lq3, rp->rsl);
	return EOK;
}

static int rottohere(REFPARS *xp)
{
 /* This does the necessary rotation to get the real-space matrix right. */
	int	i, j, k;
	double	lel;
	float	lrp[3], lare[3][3], lzs[3][3];

	if ((i = ucell(xp->xabc, xp->xang, xp->xast, xp->xanst, xp->xisys)))
		return i;
	if ((i = vecmat(xp->xabc, xp->xang, lare))) return i + 200;
	lrp[0] = -xp->xdelom  * RADC;	lrp[1] = -xp->xdelchi * RADC;
	lrp[2] = -xp->xdelphi * RADC;
	rot9r(lare, lrp);	/* rotate to goniostat 0,0,0 */
	ztos(xp->xstart, lzs);
	for (i = 0; i < 3; i++) /* rotate to start of run and go unitless */
	 for (j = 0; j < 3; j++)
	    {
		for (lel = k = 0; k < 3; k++) lel += lare[i][k] * lzs[k][j];
		xp->xareal[i][j] = lel / xp->xlambda;
	    }
	return 0;
}

static int lscmp(REFREF *rp0, REFREF *rp1)
{ /* compare routine used in bgettroid */
	return (rp0->dspac < rp1->dspac ? 1 : -1);
}

static REFREF	*bgettroid(int *ngot, REFPARS *xp)
{
 /* This returns a pointer to a list of REFREF structures.
  ngot is the number that have been found */
	int		i, nloc;
	static FILE	*fpce;
	REFREF		*bgbase, *bgp, ltr;

	*ngot = 0;
	if (NULL == (fpce = xgopen("centroids", "r")))
	  {
		fprintf(stderr, "Error opening centroids input\n");
		return NULL;
	  }
	for (nloc = 0; 0 < xgread((void *)&ltr, XG_TROID, fpce); ) nloc++;
	fprintf(stdout, " %6d reflections found in CENTROIDS file\n", nloc);
	if (NULL == (bgbase = (REFREF *)callop(nloc + 20, sizeof (REFREF),
	 "bgbase in bgettroid")))
	  {
		fprintf(stderr, "Can't reserve space for bgbase\n");
		return NULL;
	  }
	if (-1 == fseek(fpce, 0L, 0))
	  {
		fprintf(stderr, "Seek error in centroids file\n");
		return NULL;
	  }
	for (bgp = bgbase, i = 0; i < nloc; i++)
		if (EOK == lgettroid(fpce, xp, &ltr))
		  {
			cpymem((AVOID *)bgp, (AVOID *)&ltr, sizeof (REFREF));
			bgp++;
		  }
	*ngot = bgp - bgbase;
	if (*ngot > 1) qsort((AVOID *)bgbase,
		(size_t)(*ngot), sizeof (REFREF), (QPE)lscmp);
	fprintf(stdout, "%6d reflections actually used\n", *ngot);
	return bgbase;
}

static double errh(REFPARS *xp, REFREF *rfc)
{ /* Determines the error in h for the current reflection */
	int	j;
	double	loh, dev, errv, hr;

	mm31(rfc->rh, xp->xareal, rfc->rs);	/* compute the index vector */
	for (j = 3, errv = 0.; j--; )	/* integerize the h vector */
	   {
		loh = ((hr = rfc->rh[j]) < 0. ? (int)(hr - 0.4999) :
			(int)(hr + 0.4999));
		dev = loh - hr;	errv += dev * dev;
	   }
	return errv;
}

static double curres(REFPARS *xp, REFREF *rfc0, int ncen)
{ /* This returns the current residual */
	int	i;
	double	tres;
	REFREF	*rfc;

	if (EOK != (i = rottohere(xp)))
	  {
		fprintf(stderr, "Error %d in rottohere\n", i);
		return 10000.;
	  }
	for (rfc = rfc0, tres = i = 0; i < ncen; i++, rfc++)
		tres += errh(xp, rfc);
	return sqrt(tres / (3. * (double)ncen));
}

static void anglelims(float *cellang, float *lim)
{ /*	This determines the limits on omega, chi, and phi
	based on the unit cell angles. We could get fancy and include
	unit cell lengths too, but I'll get to that later.
	(h,k,l)->  ( h,k,l)  for (omega,chi,phi) -> (omega+180,-chi,180+phi)
	So we can limit omega to (-90 <= om <= 90)
	(h,k,l) -> (-h,k,-l) for (omega,chi,phi) -> (-omega,180-chi,180+phi)
	So if alpha=gamma=90 we can limit phi to (-90 <= phi<= 90)
	(h,k,l) -> (-h,-k,l) for (omega,chi,phi) -> ( omega,   -chi,180+phi)
	So if alpha=beta =90 we can limit chi to (0 <= chi <= 180)
	(h,k,l) -> (h,-k,-l) for (omega,chi,phi) -> (180-om,180-chi,180+phi)
	So if beta=gamma= 90 we can limit chi to (-90 <= chi <= 90)
	*/
	int	i, is120, is90[3];

	lim[0] = -90.;	lim[1] = 90.;	lim[2] = -180.;
	lim[3] = 180.;	lim[4] = -180.;	lim[5] = 180.;
	for (i = 0; i < 3; i++) is90[i] = (fabs(cellang[i] - 90.) < 0.01);
	is120 = (fabs(cellang[2] - 120.) < 0.01);
	if (is90[0] && is90[2])
	  {
		lim[4] = -90.;	lim[5] = 90.;
	  }
	if (is90[0] && is90[1] && is120)
	  {
		lim[0] = -30.;	lim[1] = 30.;	lim[2] = 0.;
		lim[4] = -90.;	lim[5] = 90.;
	  }
	if (is90[1] && is90[2])
	  {
		lim[3] = 90.;	if (lim[2] < 0.) lim[2] = -90.;
	  }
}

int bestgmain(int argc, char *argv[])
{	/* main line for `bestg' calculation.  */
	int	i, nbr, ncen;
	double	best, bestom, bestchi, chi, delom, delchi, delphi, om, phi;
	double	specdel, tchi, tchiom, tom, tphi, tphichi, tphiom, tres;
	float	lim[6], perma[3][3];
	REFREF	*rfc0;
	REFPARS	*xp;
	VSPLINE	*lspl;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	if ((argc > 1) && (specdel = atof(argv[1])))
		delom = specdel;
	else	delom = 3.6;
	delchi = delphi = delom;
		/* open the spline */
	if ((NULL == (lspl = gspline(5, &i))) || (i != EOK))
	  {
		fprintf(stderr, "Failure %d opening USPLINE\n", i);
		return EREADERR;
	  }
	if (NULL == (xp = (REFPARS *)callop(1, sizeof (REFPARS),
	 "xp in newrun"))) return EALLOC;
		/* read in old parameters */
	if (EOK != (i = refparin(xp, &nbr))) return i;
	makeq21(xp);	/* get the matrix related to tilt and swing */
	printf( "recorded Euler angles: %7.2f, %7.2f, %7.2f\n",
	 xp->xdelom, xp->xdelchi, xp->xdelphi);
	best = HUGE;
	if ((i = ucell(xp->xabc, xp->xang, xp->xast, xp->xanst, xp->xisys)))
	  {
		fprintf(stderr, "Error %d in ucell\n", i); return EREADERR;
	  }
	if ((i = vecmat(xp->xabc, xp->xang, perma)))
	  {
		fprintf(stderr, "Error %d in vecmat\n", i); return EREADERR;
	  }
	anglelims(xp->xang, lim);
	fprintf(stdout, "limits: [%7.2f,%7.2f],[%7.2f,%7.2f],[%7.2f,%7.2f]\n",
	 lim[0], lim[1], lim[2], lim[3], lim[4], lim[5]);
	if ((NULL == (rfc0 = bgettroid(&ncen, xp))) || (ncen < 2))
	  {
		fprintf(stderr, "Only %d troids found\n", ncen); return EBAD;
	  }
	tres = curres(xp, rfc0, ncen);
	fprintf(stdout, "tres for current matrix is %9.5f\n", tres);
	tom = tchi = tphi = tphichi = tphiom = tchiom = -PI;
	/* cycle through relevant values of omega, chi, phi.
	 Because we get the same matrix with (omega+PI,-chi,phi+PI),
	 we can restrict our attention to omega between -PI/2 and PI/2 */
	for (om = lim[0]; om < lim[1]; om += delom)
	   {
		bestom = HUGE;	xp->xdelom = om;
		for (chi = lim[2]; chi < lim[3]; chi += delchi)
		   {
			bestchi = HUGE;	xp->xdelchi = chi;
			for (phi = lim[4]; phi < lim[5]; phi += delphi)
			   {
				xp->xdelphi = phi;
				if ((tres = curres(xp, rfc0, ncen)) < bestchi)
				  {
					tphichi = phi;	bestchi = tres;
				  }
			   }
			for (phi = tphichi - delphi;
			 phi <= tphichi + delphi; phi += delphi * 0.2)
			   {
				xp->xdelphi = phi;
				if ((tres = curres(xp, rfc0, ncen)) < bestchi)
				  {
					tphichi = phi;	bestchi = tres;
				  }
			   }
			if (bestchi < bestom)
			  {
				tphiom = tphichi;
				tchiom = chi;	bestom = bestchi;
			  }
		   }
		for (chi = tchiom - delchi;
		 chi <= tchiom + delchi; chi += delchi * 0.2)
		   {
			bestchi = HUGE;	xp->xdelchi = chi;
			for (phi = tphiom - delphi;
			 phi <= tphiom + delphi; phi += delphi * 0.1)
			   {
				xp->xdelphi = phi;
				if ((tres = curres(xp, rfc0, ncen)) < bestchi)
				  {
					tphichi = phi;	bestchi = tres;
				  }
			   }
			if (bestchi < bestom)
			  {
				tphiom = tphichi;
				tchiom = chi;	bestom = bestchi;
			  }
		   }
		printf("best resid @om=%8.2f = %8.4f @chi,phi= %8.2f,%8.2f\n",
			om, bestom, tchiom, tphiom);
		if (bestom < best)
		  {
			tphi = tphiom;	tchi = tchiom;
			tom = om;	best = bestom;
		  }
	   }
	/* having gotten close, we hone in on the right answer */
	for (om = tom - delom; om <= tom + delom; om += delom * 0.1)
	   {
		bestom = HUGE;	xp->xdelom = om;
		for (chi = tchi - delchi; chi <= tchi + delchi;
		 chi += delchi * 0.1)
		   {
			bestchi = HUGE;	xp->xdelchi = chi;
			for (phi = tphi - delphi;
			 phi <= tphi + delphi; phi += delphi * 0.1)
			   {
				xp->xdelphi = phi;
				if ((tres = curres(xp, rfc0, ncen)) < bestchi)
				  {
					tphichi = phi;	bestchi = tres;
				  }
			   }
			if (bestchi < bestom)
			  {
				tphiom = tphichi;
				tchiom = chi;	bestom = bestchi;
			  }
		   }
		if (bestom < best)
		  {
			tphi = tphiom;	tchi = tchiom;
			tom = om;	best = bestom;
		  }
	   }
	/* having gotten close, we hone in on the right answer */
	for (om = tom - delom * 0.1; om <= tom + delom * 0.1;
	 om += delom * 0.02)
	   {
		bestom = HUGE;	xp->xdelom = om;
		for (chi = tchi - delchi * 0.1; chi <= tchi + delchi * 0.1;
		 chi += delchi * 0.02)
		   {
			bestchi = HUGE;	xp->xdelchi = chi;
			for (phi = tphi - delphi * 0.1;
			 phi <= tphi + delphi * 0.1; phi += delphi * 0.02)
			   {
				xp->xdelphi = phi;
				if ((tres = curres(xp, rfc0, ncen)) < bestchi)
				  {
					tphichi = phi;	bestchi = tres;
				  }
			   }
			if (bestchi < bestom)
			  {
				tphiom = tphichi;
				tchiom = chi;	bestom = bestchi;
			  }
		   }
		if (bestom < best)
		  {
			tphi = tphiom;	tchi = tchiom;
			tom = om;	best = bestom;
		  }
	   }
	printf("best resid =%8.4f @ [%8.2f,%8.2f,%8.2f]\n",
		best, tom, tchi, tphi);
	return EOK;
}

/* end of bestg functionality */

/*********************************************
	matrix:	input a b c alpha beta gamma ang1 ang2 ang3
	and then output two real-space unit cell matrices:
	The first assumes ang1,ang2,ang3 are omega,chi,phi
	as defined in X-GEN
	The second assumes ang1,ang2,ang3 are rotx,roty,rotz
	 as defined in a fixed coordinate space.
*********************************************/

int matrmain(int argc, char *argv[])
{ /* mainline for converting UPARAMS file from 1 starting position to another*/
	register int	i, j;
	double		c0, c1, c2, lam, s0, s1, s2;
	float		cel[9], basem[3][3];
	float		mat1[3][3], mat2[3][3], mat3[3][3], mat4[3][3];

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	if (argc <= 10)
	  {
		fprintf(stderr,
 "Usage: matrix a b c alpha beta gamma ang1 ang2 ang3 lambda; argc = %d\n",
		 argc);
		return EREADERR;
	  }
	for (i = 0; i < 9; i++)
	   {
		cel[i] = atof(argv[i+1]);
		if (i < 3)
		  {
			if (cel[i] < DFUZZ)
			  {
				fprintf(stderr,
				 "Realspace length cel[%d] = %11.4e\n",
					i, cel[i]);
				return EREADERR;
			  }
		  }
		else if (i < 6)
		  {
			if ((cel[i] < 10.) || (cel[i] > 170.))
			  {
				fprintf(stderr,
				 "Realspace angle cel[%d] = %11.4e\n",
					i - 3, cel[i]);
				return EREADERR;
			  }
		  }
	   }
	lam = atof(argv[10]);
	if (EOK != (i = vecmat(cel, &(cel[3]), basem)))
	  {
		fprintf(stderr, "Error %d in vecmat call\n", i);
		return EREADERR;
	  }
	fprintf(stdout, "Unit cell has a,b,c,alpha,beta,gamma =");
	fprintf(stdout, "%8.2f %8.2f %8.2f %8.2f %8.2f %8.2f\n",
	 cel[0], cel[1], cel[2], cel[3], cel[4], cel[5]);
	fprintf(stdout, " angles are %8.2f %8.2f %8.2f\n",
	 cel[6], cel[7], cel[8]);
	pr3m(stdout, basem, "base unit cell matrix");
	c0 = cos(cel[6] * RADC);	s0 = sin(-cel[6] * RADC);
	c1 = cos(cel[7] * RADC);	s1 = sin(-cel[7] * RADC);
	c2 = cos(cel[8] * RADC);	s2 = sin(-cel[8] * RADC);
	mat1[0][0] = s0 * c1 * s2 + c0 * c2;
	mat1[1][0] = c0 * s2 - s0 * c1 * c2;
	mat1[2][0] = s0 * s1;
	mat1[0][1] = s0 * c2 - c0 * c1 * s2;
	mat1[1][1] = c0 * c1 * c2 + s0 * s2;
	mat1[2][1] = -c0 * s1;
	mat1[0][2] = -s1 * s2;
	mat1[1][2] = s1 * c2;
	mat1[2][2] = c1;
	pr3m(stdout, mat1, "omega-chi-phi rotation");
	mm3(mat3, mat1, basem);
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++) mat3[i][j] /= lam;
	pr3m(stdout, mat3, "omega-chi-phi matrix");
	mat2[0][0] = c1 * c2;
	mat2[0][1] = s0 * s1 * c2 + c0 * s2;
	mat2[0][2] = s0 * s2 - c0 * s1 * c2;
	mat2[1][0] = -c1 * s2;
	mat2[1][1] = c0 * c2 - s0 * s1 * s2;
	mat2[1][2] = c0 * s1 * s2 + s0 * c2;
	mat2[2][0] = s1;
	mat2[2][1] = -s0 * c1;
	mat2[2][2] = c0 * c1;
	mm3(mat4, mat2, basem);
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++) mat4[i][j] /= lam;
	pr3m(stdout, mat4, "rotxyz matrix");
	return EOK;
}

/* end of matrix functionality */

/*********************************************
	newrun:	read a UPARAMS file and output another with different
		pseudo-goniostat values.
	This takes an input UPARAMS file and rewrites it with a new set of
	pseudo-goniostat "omega", "chi", and "phi" values.  The user must input
	the old starting true goniostat position and the new starting true
	goniostat position; the program does the rest, silently and subtly.
********************************************/

int newrunmain(int argc, char *argv[])
{ /* mainline for converting UPARAMS file from 1 starting position to another*/
	register int	i, j;
	int		nbr;
	double		newchi, newom, newphi, oldchi, oldom, oldphi;
	float		cel[9], lsave[3][3];
	char		nline[82];
	static float	zs[3][3];
	static REFPARS	*xp;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	/* see how big the UPARAMS file is to decide how to manage this */
	if (NULL == (xp = (REFPARS *)callop(1, sizeof (REFPARS),
	 "xp in newrun"))) return EALLOC;
		/* read in old parameters */
	if (EOK != (i = refparin(xp, &nbr))) return i;
	printf( "recorded Euler angles: %7.2f, %7.2f, %7.2f\n",
	 xp->xdelom, xp->xdelchi, xp->xdelphi);
	if (argc < 4) /* fewer than 3 arguments: prompt for answers */
	  {
	    if ((size_t)nbr < sizeof (REFPARS)) /* old file: ask for old ones*/
	      {
		while (1)
		    {
			printf(
		 "Enter old omega, chi, and phi, separated by spaces: ");
			if ((NULL != fgets(nline, 80, stdin)) &&
			  (3 <= sscanf(nline, "%lf %lf %lf",
				&oldom, &oldchi, &oldphi))) break;
		    }
	      } /* end if "nbr < sizeof (REFPARS)" clause */
	     else { /* newer format: we know the old omega,chi,phi */
			oldom = xp->xstart[0];
			oldchi = xp->xstart[1];
			oldphi = xp->xstart[2];
		  }
	     while (1) /* in either case, prompt for the new values */
		{
			printf(
		 "Enter new omega, chi, and phi, separated by spaces: ");
			if ((NULL != fgets(nline, 80, stdin)) &&
			  (3 <= sscanf(nline, "%lf %lf %lf",
				&newom, &newchi, &newphi))) break;
		}
	  } /* end if (if (argc < 4) clause */
	else if (argc < 7) /* three values specified: assume they're new vals*/
	  {
	    if ((size_t)nbr < sizeof (REFPARS))
	      {
		while (1)
		    {
			printf(
		 "Enter old omega, chi, and phi, separated by spaces: ");
			if ((NULL != fgets(nline, 80, stdin)) &&
			  (3 <= sscanf(nline, "%lf %lf %lf",
				&oldom, &oldchi, &oldphi))) break;
		    }
	      } /* end if "nbr < sizeof (REFPARS)" clause */
		newom = atof(argv[1]);	newchi = atof(argv[2]);
		newphi = atof(argv[3]);
	  }
	else {
		oldom  = atof(argv[1]);	oldchi = atof(argv[2]);
		oldphi = atof(argv[3]);	newom  = atof(argv[4]);
		newchi = atof(argv[5]);	newphi = atof(argv[6]);
	     }
	/* in the case where we have a newer file, the unit cell matrix
	 is stored at the (0,0,0) position rather than the start position.
	 So, perversely, we'll rotate to (0,0,0) and then go right back... */
	if ((size_t)nbr >= sizeof (REFPARS))
	  {
		xp->xstart[0] = oldom;	xp->xstart[1] = oldchi;
		xp->xstart[2] = oldphi;	ztos(xp->xstart, zs);
		mm3(lsave, xp->xareal, zs);
		for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
			xp->xareal[i][j] = lsave[i][j];
	  }
	/* in all cases we now rotate from the data position to (0,0,0) */
	xp->xstart[0] = oldom;	xp->xstart[1] = oldchi; xp->xstart[2] = oldphi;
	ztos(xp->xstart, zs);
	mm3t(lsave, xp->xareal, zs);
	/* and now we rotate up to the new position */
	xp->xstart[0] = newom;	xp->xstart[1] = newchi;	xp->xstart[2] = newphi;
	ztos(xp->xstart, zs);
	mm3(xp->xareal, lsave, zs);
	/* scale that by lambda to enable use of `artocel' */
	for (i = 3; i--; ) for (j = 3; j--; )
		lsave[i][j] = xp->xareal[i][j] * xp->xlambda;
	artocel(xp->xstart, lsave, cel); /* to get pseudo-goniostat angles */
	printf("Resulting psuedo-goniostat angles: [%8.2f %8.2f %8.2f]\n",
		cel[6], cel[7], cel[8]);
	if (EOK != (i = refparout(xp))) return i;
	memzap((AVOID **)&xp, "xp in newrun");
	globalkill();	return EOK;
}

/* end of newrun functionality */

/****************************************
	reindexu: use rational 3x4 matrix to re-index UPARAMS file for XGEN
		syntax:	reindexu [sysno]
	where the optional argument `sysno' is the spacegroup-system number
	of the resultant unit cell (1==triclinic, 2==monoclinic,
	3==orthorhombic, 4==tetragonal, 5==cubic, 6==trigonal or hexagonal,
	and 7==rhombohedral indexed rhombohedrally).
	This allows for switches between rhombohedral and hexagonal indexing of
	rhombohedral cells; to go from rhombohedral to hexagonal, we say
		reindexu 6
	and to go the other way we say
		reindexu 7
	Input files: UPARAMS (input parameters), REINDEX (transformation file)
	Incidentally, the appropriate transformation matrix for the
	rhomohedral-to-hexagonal transformation is
	 1 -1  0
	 0  1 -1
	 1  1  1
	and the appropriate transformation for the reverse operation is
		 (  2  1  1 )
	0.3333 * ( -1  1  1 )
		 ( -1 -2  1 )
	In the absence of a REINDEX file the user may invoke `reindexu' with
	either 6 or 7; it then will do the hexagonal-to-rhombohedral or
	rhombohedral-to hexagonal transformation even in the absence of an
	explicit file specification.
	This program is almost alone among the programs that manipulate or
	read the "areal" matrix in that it works entirely with the version
	of said matrix that applies at goniostat (0,0,0) rather than at
	the datarun's start position.
****************************************/

int reindumain(int argc, char *argv[])
{ /* The mainline for reindexing a UPARAMS file */
	register int	i, j, k;
	int		convert, cval, nbr, nonztrans, tv[3];
	float		rphi[3], rsave[3][3];
	double		dotp, ov, drphi[3];
	double		dq[3][3], drec[3][3], drsave[3][3], tm[3][3];
	char		reline[60];
	FILE		*fpi;
	REFPARS		*xp;
	static float	hm[3][3] = {	{ 1., -1.,  0.}, { 0.,  1., -1.},
					{ 1.,  1.,  1. } } ;
	static float	rm[3][3] = {	{ 2.,  1.,  1.}, {-1.,  1.,  1.},
					{ -1., -2., 1. } } ;
	static char	*fnames[] = { "uparams", "reindex", NULL } ;
	static char	ftypes[] = { 3, 1, 0 } ;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	/* open and read the input files */
	convert = cval = 0;
	if (argc > 1)
	  {
		for (argv++, argc--; argc--; argv++)
		  if (1 <= (cval = atoi(*argv)))
		    {
			if (6 == cval) convert = HEXAGONAL;
				else if (7 == cval) convert = RHOMBOHEDRAL;
			else convert = cval;
		    }
	  }
	if (NULL == (xp = (REFPARS *)callop(1, sizeof (REFPARS),
		"xp in reindexu"))) return EALLOC;
	if (EOK != (i = refparin(xp, &nbr))) return i;
	if (verror((int)(xp->xlambda > FUZZ),
	 "transform parameters file: wavelength is non-positive"))
		return EBADVALUE;
	nonztrans = 0;
	if (NULL == (fpi = xgopen("reindex", "r")))
	  {
		if (verror(convert,
		 "transform UPARAMS: no transformation file supplied"))
			return EOPENERR;
		if (convert == HEXAGONAL)
		  {
			for (i = 3; i--; ) for (j = 3; j--; )
				tm[i][j] = hm[i][j];
		  }
		else if (convert == RHOMBOHEDRAL)
		  {
			for (i = 3; i--; ) for (j = 3; j--; )
				tm[i][j] = rm[i][j] / 3.;
		  }
		fnames[1] = NULL;	ftypes[1] = 0;
	  }
	else
	  {
		for (i = 0; i < 3; i++)
		   {
			if (qerror((AVOID *)fgets(reline, 58, fpi),
				"read line from tranformation-matrix file"))
					return EREADERR;
			(void)sscanf(reline, "%lf %lf %lf %d",
				&tm[i][0], &tm[i][1], &tm[i][2], &tv[i]);
			if (tv[i]) nonztrans = 1;
		   }
		(void)fclose(fpi);	fpi = NULL;
		fnames[1] = "reindex";	ftypes[1] = 1;
	  }
			/* summarize what we're doing to stdout */
	header(stdout, PROCESSING, fnames, ftypes);
	if (nonztrans) fprintf(stdout,
	 "Note: nonzero tranformation vector [%d, %d, %d] will be ignored\n",
		tv[0], tv[1], tv[2]);
	fprintf(stdout, "Matrix will be reindexed as R' = MR, with M=:\n");
	fprintf(stdout,"\t\t  %6.2f %6.2f %6.2f\n",
		tm[0][0], tm[0][1], tm[0][2]);
	fprintf(stdout,"\twhere M = %6.2f %6.2f %6.2f\n",
		tm[1][0], tm[1][1], tm[1][2]);
	fprintf(stdout,"\t\t  %6.2f %6.2f %6.2f\n",
		tm[2][0], tm[2][1], tm[2][2]);
		/* convert the structure `rp' into a set of global vbls */
	if (convert != xp->xisys) fprintf(stdout,
		 "Spacegroup system number changed from %d to %d\n",
			xp->xisys, convert);
	if (deter(xp->xareal) < -DFUZZ)	/* negative determinant: */
	   for (i = 3; i--; ) xp->xareal[i][2] *= -1.;
	/* now perform the matrix tranformation on the real-space
	unit cell matrix.  Note that it's stored in transposed form. */
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
	   {
		for (ov = k = 0; k < 3; k++) ov += tm[k][i] * xp->xareal[k][j];
		drsave[i][j] = ov;
	   }
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		xp->xareal[i][j] = drsave[i][j];
	if (convert) xp->xisys = convert;
	xp->xispgp = check_sys(xp->xispgp, xp->xisys);	/* verify spacegroup */
	if (xp->xisys < 1 || xp->xisys > 7) xp->xispgp = xp->xisys = 1;
	/* now update all the appropriate matrices and parameters */
	for (i = 3; i--; )	/* unit cell lengths: */
	   { /* length = norm of one row */
		for (xp->xabc[i] = 0, j = 3; j--; )
			xp->xabc[i] += xp->xareal[i][j] * xp->xareal[i][j];
			/* put lambda back in*/
		xp->xabc[i] = sqrt(xp->xabc[i]) * xp->xlambda;
	   }
	for (i = 0; i < 3; i++)	/* unit cell angles: */
	   { /* determined from dot products, viz.
				gamma = acos((a.b) / |a||b|) */
		j = (i == 2 ? 0 : (i+1));	k = 3 - i - j;
		dotp = (xp->xareal[j][0] * xp->xareal[k][0] +
			xp->xareal[j][1] * xp->xareal[k][1] +
			xp->xareal[j][2] * xp->xareal[k][2]) *
			xp->xlambda * xp->xlambda;
		xp->xang[i] = (fabs(dotp) > DFUZZ ?
		  acos(dotp / (xp->xabc[j] * xp->xabc[k])) / RADC : 90.);
	   }
	/* Next, we parameterize the rotation from the "standard" position
	obtained by vecmat to the current position by obtaining the orthogonal
	matrix required and then pulling the angles out of that matrix.
	We do this by noting that if A is the recip-space unit cell matrix
	at the "vecmat" position and B is the corresponding position at the
	reference position of data collection, then
		B = QA for some orthogonal transformation Q.
	Thus Q = BA~, or Q' = (BA~)~ = AB~ = AR * lambda.  Pretty neat, eh? */
	for (i = 3; i--; ) for (j = 3; j--; )
		rsave[i][j] = drsave[i][j] = xp->xareal[i][j];
	if (verror(!ucell(xp->xabc, xp->xang, xp->xast, xp->xanst, xp->xisys),
	 "determine recip-space cell dimensions from realspace ones"))
		return EBADVALUE;
	if (verror(!vecmat(xp->xabc, xp->xang, xp->xareal),
	 "recreate unit cell matrix from cell parameters")) return EBADVALUE;
	for (i = 3; i--; ) rphi[i] = 0.;
	rot9r(xp->xareal, rphi); /* rotate to goniostat (0,0,0) position */
	if (verror(!inv3(xp->xareal, xp->xrec),
	 "invert intermediate realspace matrix")) return ESINGULAR;
		/* scale the real-space version by the X-ray wavelength */
	for (i = 3; i--; ) for (j = 3; j--; )
	   {
		xp->xareal[i][j] /= xp->xlambda;
		drec[i][j] = xp->xrec[i][j];
	   }
	dmm3(dq, drec, drsave);	/* get nearly orthogonal matrix Q */
	for (i = 3; i--; ) for (j = 3; j--;) dq[i][j] *= xp->xlambda;
	cleanorthog(dq);	/* make the orthogonal matrix more precise */
	qtoangs(dq, drphi); /* get the rotation angles from the q matrix */
	for (i = 0; i < 3; i++) rphi[i] = drphi[i];
	rot9r(xp->xareal, rphi); /* reconstitute the matrices */
	if (EOK != (i = refparout(xp))) return i;
	fprintf(stdout, "Transformed unit cell and orientation angles:\n");
	fprintf(stdout, "          a:      b:      c:  alpha:   beta:  \n");
	fprintf(stdout, "gamma:  omega:    chi:     phi:\n");
	fprintf(stdout,
		"Value:%8.3f%8.3f%8.3f%8.3f%8.3f%8.3f%8.3f%8.3f%8.3f\n",
		xp->xabc[0], xp->xabc[1], xp->xabc[2],
		xp->xang[0], xp->xang[1], xp->xang[2], 
		xp->xdelom, xp->xdelchi, xp->xdelphi);
	memzap((AVOID **)&xp, "xp in reindexu");
	globalkill();	return EOK;
}

/* end of reindexu functionality */

/*****************************************
	predictit:	Predict reflections over a range based on UPARAMS

   Purpose: obtain list of reflections expected to occur over a defined
	frame range, given a known set of crytal and detector parameters.
   Cross references: ../integrate/integpred.c
*************************************************************************/

static void sum_predin(FILE *fp, FILE *fpl, PREDWORK *pwp,
	double lmu1, double lmu2)
{	/* This summarizes the run-time inputs to prediction. */
	REFPARS		*xp;
	static char	*flist[] =
		{"uparams", "uspline", "amask", "uspotlist", "xlog", NULL};
	static char	ftype[] = {1, 1, 1, 2, 2, 0};

	ftype[4] = (fpl == stdout ? 0 : 2);
	header(fp, PROCESSING, flist, ftype);
	xp = pwp->po_xp;
	fprintf(fp,
	"Range: Frames %6.1f to %6.1f in steps of %7.3f degrees\n",
	 lmu1 / xp->xfrmsize, lmu2 / xp->xfrmsize, xp->xfrmsize / RADC);
	fprintf(fp,"Resolution range for prediction: %9.3f A to %9.3f A\n",
		xp->xresmin, xp->xresmax);
	if (pwp->po_miny2sq > (DEFAULT_MINY2SQ / 2.)) fprintf(fp,
	   "Only reflections with s(y) > %9.5f will be predicted\n",
		sqrt(pwp->po_miny2sq));
	else fprintf(fp,
	   "No Lorenz-factor restrictions on predictions will be made\n");
	fprintf(fp, "Polarization: fraction polarized = %7.4f; ",
		pwp->po_polfrac);
	fprintf(fp, "angle wrt plane = %7.2f degrees\n", pwp->po_polnormrot);
/* set up logfile header for the ranges */
	fprintf(fp,"\t\tSummary of Prediction Ranges\n");
	fprintf(fp, "  Start    End    Indices Varied:    Slow Index");
	fprintf(fp, "  Number of HKLs  Time Required\n");
	fprintf(fp, "  Frame  Frame Slowest Middle Fastest  Limits");
	fprintf(fp, "  Checked Predicted  Predict\n");
}

static int pcmp(PREDREF *pp0, PREDREF *pp1)
{ /* Compare routines for the frame number and X in PREDWORK structures */
	register int	dframe;

	if ((dframe = pp0->pfmno - pp1->pfmno)) return dframe;
	return (pp0->pxt < pp1->pxt ? -1 : 1);
}

static int endpred(int irv, PREDWORK *pwp, FILE *fpl)
{	/* last few items in `predict' before terminating */
	if ((stdout != fpl) && (NULL != fpl)) (void)fclose(fpl);
	memzap((AVOID **)&(pwp->po_prp0), "pwp->po_prp0 in predict");
	globalkill();	return irv;
}

static void predfates(FILE *fp, PREDWORK *pwp)
{ /*	This prints a summary of a prediction effort to the stream fp */
	fprintf(fp,"Fate of attempts at predicting hkls:\n");
	fprintf(fp, " Fate                            Number of hkls\n");
	fprintf(fp," First phi value used:                 %8d\n",
		pwp->po_nresult[PR_FIRSTPHIOK]);
	fprintf(fp," Second phi value used:                %8d\n",
		pwp->po_nresult[PR_SECPHIOK]);
	fprintf(fp," Both phi values used:                 %8d\n",
		pwp->po_nresult[PR_BOTHPHIOK]);
	fprintf(fp," X or Y value out of range:            %8d\n",
		pwp->po_nresult[PR_XYRANGE]);
	fprintf(fp," Both phi values out of range:         %8d\n",
		pwp->po_nresult[PR_PHIRANGE]);
	fprintf(fp," Reflection too close to rotation axis:%8d\n",
		pwp->po_nresult[PR_BADY2]);
	fprintf(fp," Scattering vector along main beam:    %8d\n",
		pwp->po_nresult[PR_ALONGBEAM]);
	fprintf(fp," Scattering vector too small:          %8d\n",
		pwp->po_nresult[PR_TOOLOWRES]);
	fprintf(fp," Scattering vector too large:          %8d\n",
		pwp->po_nresult[PR_TOOHIRES]);
	fprintf(fp," hkl indices illegal in spacegroup:    %8d\n",
		pwp->po_nresult[PR_BADHKL]);
	fprintf(fp," Z component of scatt vec out of range:%8d\n",
		pwp->po_nresult[PR_SZRANGE]);
	fprintf(fp,
 "%d reflections have been checked; %d have been written to the output file\n",
		(int)pwp->po_totchecked, (int)pwp->po_totfound);
	return;
}

int predmain(int argc,char *argv[])
{ /* "predmain": the calling routine used in predicting Bragg reflections.
	The main doesn't do much. */
	register int	i, range, writeok;
	int		ltfound, npred, nranges;
	double		deltamu, maxgiven, mingiven, mu0, mu1, mu2, nval;
	char		**flags,  **dcp, pline[120];
	FILE		*fpframe, *fplog, *fpo;
	REFPARS		*xp;
	PREDREF		*prp, *prpend;
	static PREDWORK	pws;
#ifdef VMS
	static short	sortspec[] = { 2,
			DSC$K_DTYPE_W, SO_ASCENDING, 0, sizeof (short),
			DSC$K_DTYPE_F, SO_ASCENDING, 2, sizeof (float) };
#endif

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
		/* populate pws structure */
	if (EOK != (i = prpar(&pws))) return endpred(i, &pws, NULL);
	/* handle run-time inputs */
	maxgiven = mingiven = -30000.;
	flags = flagproc(argc, argv, "AMNPRSUZ");
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	   {
		nval = atof((*dcp) + 1);
		switch (**dcp)
		  {
			case	'A':	pws.po_miny2sq = DFUZZ;		break;
			case	'M':	case 'U':
				pws.po_polfrac = UNPOL_FRAC;	break;
			case	'N':	pws.po_polnormrot = 90.;	break;
			case	'P':	pws.po_polfrac = nval;		break;
			case	'R':	pws.po_polfrac = RING_FRAC;	break;
			case	'S':	pws.po_polnormrot = nval;	break;
			case	'Z':	pws.po_polnormrot = 0.;	break;
		  }
	   }
	for (argc--, argv++; (argc) && !allnumeric(*argv); argc--) argv++;
	if (argc < 2) /* < 2 args: get the frame range from the frames file*/
	  {
		if (qerror((AVOID *)(fpframe = xgopen("frames", "r")),
		 "open frame index file"))
			return endpred(EOPENERR, &pws, NULL);
		if (qerror((AVOID *)fgets(pline, 80, fpframe),
		 "read line of frame index"))
			return endpred(EREADERR, &pws, NULL);
		mu2 = mu1 = atof(pline) - 5; /* (1st frame in index) - 5 */
		while (NULL != fgets(pline, 80, fpframe)) mu2 = atof(pline) +5;
		(void)fclose(fpframe);
	  }
	else	/* read range values (& resolution ? ) from command line */
	  {
		mu1 = atof(argv[0]);	mu2 = atof(argv[1]);
		if (argc > 2) maxgiven = atof(argv[2]);
		if (argc > 3) mingiven = atof(argv[3]);
	  }
	/* reverse the order of these if needed */
	if ((maxgiven > 0.) && (mingiven > 0.) && (maxgiven > mingiven))
	  {
		nval = maxgiven; maxgiven = mingiven; mingiven = nval;
	  }
	xp = pws.po_xp;	mu1 *= xp->xfrmsize;
	mu2 = (mu2 + 1) * xp->xfrmsize;
	if (maxgiven > DFUZZ)
	  {
		xp->xresmax = maxgiven;
		pws.po_smasq = xp->xlambda / xp->xresmax;
		pws.po_smasq *= pws.po_smasq;
	  }
	if (mingiven > DFUZZ)
	  {
		xp->xresmin = mingiven;
		pws.po_smisq = xp->xlambda / xp->xresmin;
		pws.po_smisq *= pws.po_smisq;
	  }
	/* now impose a minimum resolution limit if the user wants it */
	if (mingiven > 0.) xp->xresmin = mingiven;
	/* Determine the width of the ranges of scanning angle to be used.
	If the requested total range is less than MAXRANGE, only one range
	will be used, and its width will be the total width.  Otherwise
	the range is broken up into pieces about MAXRANGE in size.  */
	if (fabs(mu2 - mu1) > MAXRANGE)
		nranges = 1 + (int)(fabs((mu2 - mu1)) / MAXRANGE);
	else	nranges = 1;
	deltamu= (mu2 - mu1) / (double)(nranges);
	if (NULL == (fplog = propen("xlog"))) fplog = stdout;
	sum_predin(fplog, fplog, &pws, mu1, mu2);
	if (fplog != stdout) sum_predin(stdout, fplog, &pws, mu1, mu2);
	xeflush(fplog);	pws.po_tl0 = NOW;
	/* create & open output */
	if (qerror((AVOID *)(fpo = xgopen("uspotlist", "w")),
	 "open output predicted-spot list"))
		return endpred(EOPENERR, &pws, fplog);
	/* go through ranges of scanning angle, predicting over each range;
		sort the reflections and output them once per range */
	writeok = 1;
	for (mu0 = mu1, range = 0; range < nranges; range++, mu0 += deltamu)
	   {
		if ((pws.po_xp)->xfrmsize >= 0)
		  {
			pws.po_mumin = mu0; pws.po_mumax = mu0 + deltamu;
		  }
		else {
			pws.po_mumax = mu0; pws.po_mumin = mu0 + deltamu;
		     }
		/* predict over this range */
		if (0 > (npred = prloop(&pws, 1)))
			return endpred(EBADVALUE, &pws, fplog);
		if (npred > 1) qsort((AVOID *)(pws.po_prp0), (size_t)npred,
		 sizeof (PREDREF), (QPE)pcmp); /* sort on frame # and X */
		prpend = pws.po_prp0 + npred;
		/* pcor < 0 means Lorentz correction is extreme;
		 pgama < 0 means reflection is on the edge of the detetor */
		/* don't write out the reflections with pcor < 0; that's a
		 flag indicating that the Lorentz correction is too extreme */
		for (ltfound = 0, prp = pws.po_prp0; prp < prpend; prp++)
		   if ((prp->pcor >= 0.) && (prp->pgama >= 0.))
		     {
			if (verror(0 < xgwrite((void *)prp, XG_PREDREF, fpo),
			 "output refl to disk"))
			  {
				writeok = 0; break;
			  }
			ltfound++;
		     }
		pws.po_tl0 = NOW;	pws.po_totfound = ltfound;
		if (!writeok) break;
	   }
	(void)fclose(fpo);	fpo = NULL;
	predfates(fplog, &pws);	/* report prediction statistics to logfile */
	printf(
 "%d reflections have been checked; %d have been written to the output file\n",
		(int)(pws.po_totchecked), (int)(pws.po_totfound));
	return endpred(EOK, &pws, fplog);
}

/* end of predictit functionality */

/* end of parutils.c */
