/* redurefl.c -- convert UREFLS to MULTIREF and filter

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	28 Nov 2006: os_q21 element of OUT_OBSSTYLE replaced with xp->xq21;
	 xp = rwp->ru_rpp referenced for simplicity in many places in convert()
	24 Jun 2006: include xgpostint.h rather than reddefs.h
	13 Jun 2006: merged old redurefl.c and ../xglib/xgenmisc.c to
	 create this file and redutils.c; total reordering of routines.
	16 Jan 2006: fix SUMSTRONG_ABG construct
	31 Dec 2005: don't qsort if fewer than 2 elements
	20 Nov 2005: inline argument declarations and strict prototyping
	12 Oct 2005: housekeeping
	24 Apr 2005: move source from ../integrate to ../utilities
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	28 Aug 2004: routines used only here are declared as static;cosmetics
	13 Aug 2004: use OUT_OBSSTYLE structure to pass args around
	 7 Aug 2004: correct bug in red_setup so non-flag args are okay
	15 Jul 2004: run-time options handled directly in red_setup() using
		flagproc() and other standard tools
	29 Mar 2004: set filepointers to NULL after closing the files
	28 Jan 2004: many static globals removed
	21 Jan 2004: license message inserted and initial comment reordered
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	17 Aug 2003: handle ice-rings through calls to external routines
	16 Aug 2003: handle exclusion codes for keephkl() using an
		argument to the routine rather than a global variable
	 5 Jan 2003: reorganized source-file header
	15 Oct 2002: announce that that the omega width of the scaling groups
		has been enlarged when that happens in order to keep the
		number of scaling groups below 200.
	18 Apr 2002: get rid of duplicate observations before output
		unless the user specifies otherwise
	15 Apr 2002: in unique_refls, modify #frm/deg so #scaling groups < 200
	25 Mar 2002: removed MSI attachments
	15 Feb 1998: in `convert', fix implementation to allow for possibility
		that the intensity measurements may be Kabsch values
	31 May 1995: "register" -> "register int" throughout;
		%d replaces %ld; typecast tl1-tl0
	12 Apr 1995: replace s_char with sichar; add option to exclude ice-ring
	 2 Mar 1995: provide for plane backgrounds; diagnostics to stdout;
		frame # arg in pix2cm, cm2pix; shared-memory alloc and free
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 5 Oct 1994: use xgread for OBSREF, REFPARS;
		use xgwrite for TROID, MERGEDREF
	 5 May 1994: source-file headers; POLYDET->XGEN;
		"vsort.h" included as absolute reference
	16 Mar 1994: a few I/O changes and global variables moved around
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version
	18 Jan 1988: compute the scaling group numbers after
		all the data have been read in so that they can be of
		almost precisely equal size and we won't end up with
		almost unpopulated scaling groups on the edges.
	30 Apr 1987: make sure the reflection is within the
		detector boundaries defined by the AMASK file.
	 1 Dec 1986: incorporate improvements to cvtrefl here.
	 1 Jun 1986: original code

   Purpose: to convert data from UREFLS format to MULTIREF format,
	optionally filtering out reflections by various criteria.
   Cross references: redutils.c, ../postprocess/mrmerge.c, ../include/reddefs.h

 `reduce' is a program for converting the UREFLS format emitted by
 the XGEN `integrate' program
 or the "REFLECTIONS" format emitted by the XENSYS `SCAN' program
 into the MULTIREF format used by the XGEN data reduction programs.
 As a secondary capability it can write a CENTROIDS file (suitable
 as an input to the XGEN `refine' program) as an output.

		Run synopsis:
	reduce [-flags] [runno [minsigoi [Dmin [Dmax]]]]

	runno is an integer run-number identifier, minsigoi is
	the minimum value of sigma/I to be output, and Dmin and Dmax are the
	high-resolution cutoff and the low-resolution cutoff in Angstroms.
	Runtime flags are documented in redutils.c
*/
#include	"xgpostint.h"
#define		MINSIGMA	1	/* all sigmas > this */
#define		MAXOBINT	40000.	/* largest allowed intensity */
#define		CORCMON		18.945	/* 10 * (1+MONOC2) */
#define		MONOC2		0.8945	/* monochromator (cos2theta)squared */

static void refl_to_uref(OBSREF *rp, float xscale, float yscale)
{ /*	This converts a reflection in SCANREF format to the corresponding
	OBSREF.  The resulting intensity is NOT Lorentz-corrected;
	that's done later.*/
	double	pixratio;
	SCANREF	srf;

	/* first, copy the data into a working buffer */
	cpymem((AVOID *)&srf, (AVOID *)rp, sizeof (SCANREF));
	hunpack(srf.rhkl, rp->obs_hkl);	/* unpack the hkl value */
	rp->obs_xob = rp->obs_xpred = srf.rxcen * xscale + 0.49999;/*get X, Y*/
	rp->obs_yob = rp->obs_ypred = srf.rycen * yscale + 0.49999;
	rp->obs_omob = srf.rzcen;	/* get omega value */
	rp->obs_ompred = (rp->obs_omob > 0. ? srf.rzcen + 0.49999 :
						srf.rzcen - 0.49999);
	pixratio = (double)srf.rnspot / (double)srf.rnback;
	rp->obs_ii = rp->obs_iunfit =
		srf.rsumspot - pixratio * (double)srf.rsumback;
	rp->obs_sigma = rp->obs_sunfit =
		srf.rvar_spot + srf.rvar_back * pixratio * pixratio;
}

static double get_iratio(FILE *fpin, FILE *fplo)
{ /* This routine finds <I(fit) / I(sum)> within a UREFLS run.
	 We assume `fpin' is already open and positioned at the start */
	int	nlocok;
	double	thisrat, sumrat;
	OBSREF	lqrf;

	for (sumrat = nlocok = 0; /* accmumulate sums for Ifit/Isum ratio */
	 0 < xgread((void *)&lqrf, XG_OBSREF, fpin); )
		if ((lqrf.obs_ii > sqrt(lqrf.obs_sigma)) &&
		 (lqrf.obs_iunfit > sqrt(lqrf.obs_sunfit)))
		  {
			thisrat = lqrf.obs_ii / lqrf.obs_iunfit;
			if ((0.01 < thisrat) && (thisrat < 100.))
			  {
				sumrat += thisrat;	nlocok++;
			  }
		  }
	(void)fseek(fpin, (size_t)0, 0); /* rewind UREFLS file */
	if ((nlocok < 2) || (sumrat < DFUZZ))
	  {
		fprintf(stderr, "only %d okay observations, sumrat = %9.5f\n",
			nlocok, sumrat);
		fprintf(fplo, "only %d okay observations, sumrat = %9.5f\n",
			nlocok, sumrat);
		return 1.;
	  }
	sumrat /= (double)nlocok;
	fprintf(stdout, "<I(fit)/I(sum)> for %d observations = %9.5f\n",
		nlocok, sumrat);
	fprintf(fplo, "<I(fit)/I(sum)> for %d observations = %9.5f\n",
		nlocok, sumrat);
	return sumrat;
}

static int proc_heads(FILE *fpout, int npair, CRYSTAL *xhp, DATARUN *rhp)
{ /*	This finishes the statistical operations which result in outputs
	to the headers. It's called from redurefl.c */

	if (rhp->run_wtsum > FUZZ)
	  {
		rhp->run_intave /= rhp->run_wtsum;
		rhp->run_sigave /= rhp->run_wtsum;
		rhp->run_iosave /= rhp->run_wtsum;
	  }
	rhp->run_ssqave /= rhp->run_ninteg;
	rhp->run_reslim = 0.5 / rhp->run_ssqmax; /* plane spacing */
	rhp->run_ssqmax *= 2. * rhp->run_lambda; /* rescale for lambda/2 */
	rhp->run_ssqmax *= rhp->run_ssqmax; /* max |s|^2 */
	rhp->run_nframes = (rhp->run_fmmax - rhp->run_fmmin) + 0.49999;
	rhp->run_nshifts = 2 * npair;
	xhp->xt_ninteg = rhp->run_ninteg; xhp->xt_reslim = rhp->run_reslim;
	xhp->xt_reslim = rhp->run_reslim; xhp->xt_ssqmax = rhp->run_ssqmax;
	xhp->xt_nshifts = rhp->run_nshifts;
	if (verror(0 < xgwrite((void *)xhp, XG_CRYSTAL, fpout),
		"write crystal header to output")) return EWRITERR;
	if (verror(0 < xgwrite((void *)rhp, XG_DATARUN, fpout),
		"write run header to output")) return EWRITERR;
	return EOK;
}

static void statrefl(DATARUN *rhp, FASTOB *op, double omfrm)
{ /*	This accumulates some statistics kept in the "DATARUN" header */
	double	thisi, thissig, weight;

	(rhp->run_ninteg)++;
	thisi = (op->fo_ob).abo_ii;	thissig = (op->fo_ob).abo_sigma;
	weight = 1. / (thissig * thissig); rhp->run_wtsum += weight;
	rhp->run_intave += thisi * weight; rhp->run_sigave += thissig * weight;
	rhp->run_iosave += (thisi / thissig) * weight;
	thisi = 2. * rhp->run_lambda * op->fo_stl;
	rhp->run_ssqave += thisi * thisi;
	if (omfrm < rhp->run_fmmin) rhp->run_fmmin = omfrm;
	if (omfrm > rhp->run_fmmax) rhp->run_fmmax = omfrm;
	if (op->fo_stl > rhp->run_ssqmax) rhp->run_ssqmax = op->fo_stl;
}

static int centroid_summary(OUT_OBSSTYLE *oob, FILE *fplo)
{ /*	This summarizes the results of a centroid-creation effort */
	short		*cshp;
	int		xbin, ybin;
	double		xrcen, yrcen, zr;

	fprintf(fplo,
	 " %6d observations written to %2d bins in the centroids file\n", 
		(int)oob->os_outobs, (int)oob->os_ncbins);
	fprintf(stdout,
	 " %6d observations written to %2hd bins in the centroids file\n", 
		(int)(oob->os_outobs), oob->os_ncbins);
	fprintf(fplo,
	 "Regions in (X,Y) and the number of reflections saved in each:\n");
	fprintf(fplo, "       X-> ");
	for (xbin = 0; xbin < oob->os_ncbx; xbin++)
	   {
		zr = (xbin < (oob->os_ncbx - 1)) ? 2 * xbin + 1 : 2 * xbin;
		xrcen = (oob->os_spl)->v_x0px + zr * (oob->os_spl)->v_xrpx;
		fprintf(fplo, " %5.1f", xrcen);
	   }
	fprintf(fplo, "\n     Y.    ");
	for (xbin = 0; xbin < oob->os_ncbx; xbin++) fprintf(fplo, " _____");
	fprintf(fplo, "\n");
	for (ybin = 0, cshp = oob->os_cenpbin; ybin < oob->os_ncby; ybin++)
	   {
		zr = (ybin < (oob->os_ncby - 1)) ? 2 * ybin + 1 : 2 * ybin;
		yrcen = (oob->os_spl)->v_y0px + zr * (oob->os_spl)->v_yrpx;
		fprintf(fplo, "  %5.1f    ", yrcen);
		for (xbin = 0; xbin < oob->os_ncbx; xbin++, cshp++)
			fprintf(fplo, " %5d", *cshp);
		fprintf(fplo, "\n");
	   }
	return -1;
}

static int out_centroid(OBSREF *rp, OUT_OBSSTYLE *oob, FILE *fpo, FILE *fplo)
{ /* this makes and outputs a "centroid" from the refl `rp'.
	As of V.2.0, the roughly 100 regions are more rationally
	determined: the spline file is consulted, and the region
	is chosen based on the fiducial points closest to the spot.
	There are 19 x 19 fiducial points, and we round somewhat so
	that we end up with about 100 usable regions.
	*/
	int		i, j, done, m, jfr;
	short		xybin, xbin, ybin;
	float		xrcen, yrcen, xrcm, yrcm;
	double		fix, locomc, sumnomi, xq, yq, zr;
	static TROID	thiscen;

	thiscen.nobk = (rp->obs_hifrm - rp->obs_lowfrm) + 1;
	for (sumnomi = 0, j = rp->obs_lowfrm; j <= rp->obs_hifrm; j++)
		sumnomi += rp->obs_frmint[j];
	if (sumnomi < 10) return 0;
	thiscen.xto = (double)rp->obs_xob / oob->os_xscal;
	thiscen.yto = (double)rp->obs_yob / oob->os_yscal;
	xq = thiscen.xto + oob->os_xdel; yq = thiscen.yto + oob->os_ydel;
	xbin = xq < (oob->os_spl)->v_x0px ? 0 :
	 (xq - (oob->os_spl)->v_x0px) / (2. * (oob->os_spl)->v_xrpx) + 0.49999;
	if (xbin >= oob->os_ncbx) xbin = oob->os_ncbx - 1;
	ybin = yq < (oob->os_spl)->v_y0px ? 0 :
	 (yq - (oob->os_spl)->v_y0px) / (2. * (oob->os_spl)->v_yrpx) + 0.49999;
	if (ybin >= oob->os_ncby) ybin = oob->os_ncby - 1;
	xybin =  ybin * oob->os_ncbx + xbin;
	if (oob->os_cenpbin[xybin] >= oob->os_ncperbin) return 0;
	(oob->os_cenpbin[xybin])++;	(oob->os_outobs)++;
	thiscen.frame0 = rp->obs_omob;
	thiscen.obkmin = ((int)(rp->obs_ompred)) - 4 + rp->obs_lowfrm; 
	thiscen.totinten = rp->obs_ii;
	locomc = 0.;	m = thiscen.obkmin;
	for (i = 0, j = rp->obs_lowfrm; j <=rp->obs_hifrm; i++,j++,m++)
	   {
		thiscen.obparts[i] = (double)rp->obs_frmint[j] /sumnomi;
		locomc += thiscen.obparts[i] * (double)m;
	   }
	/* Hack! sometimes the omega centroid is off by 1 in one direction
	 or the other.  We'll have to adjust obkmin up or down */
	if (locomc > (rp->obs_omob + 0.4999))
		fix = -(locomc + 0.4999 - rp->obs_omob);
	else if (locomc < (rp->obs_omob - 0.4999))
		fix = (rp->obs_omob + 0.4999 - locomc);
	else	fix = 0.;
	j = fix;	thiscen.obkmin += j;
	if (verror(0 < xgwrite((void *)&thiscen, XG_TROID, fpo),
		 "write a TROID reflection to the output")) return -2;
	/* if we've gotten enough centroids in this bin, see if we've gotten
	enough in all the bins located inside the active area */
	if (oob->os_cenpbin[xybin] >= oob->os_ncperbin)
	  {
		for (ybin = done = 0; ybin < oob->os_ncby; ybin++)
		   {
			zr = (ybin < oob->os_ncby - 1) ?
				2 * ybin + 1 : 2 * ybin;
			yrcen = (oob->os_spl)->v_y0px +
				zr * (oob->os_spl)->v_yrpx;
			for (xbin = 0; xbin < oob->os_ncbx; xbin++)
			   {
				zr = xbin < oob->os_ncbx - 1 ?
				 2 * xbin + 1 : 2 * xbin;
				xrcen = (oob->os_spl)->v_x0px +
					zr * (oob->os_spl)->v_xrpx;
				i = ybin * oob->os_ncbx + xbin;
				if (oob->os_cenpbin[i] >= oob->os_ncperbin)
					done++;
				else /*see if bin is really in active area*/
				  {
					jfr = rp->obs_omob + 0.49999;
					if (1 == pix2cm(xrcen, yrcen,
						&xrcm,&yrcm, jfr)) done++;
					else return 0;
				  }
			   }
		   }
		if (done >= oob->os_ncbins - 2)
			return centroid_summary(oob, fplo);
	  }
	return 1;
}

static int log_bad(ABBROB *aobp, short lhkl[], int reason, REDUWORK *rwp)
{ /*	This logs illegal reflections found in a "urefls" file */
	OUT_OBSSTYLE	*loo;

	loo = rwp->ru_oob;	(rwp->ru_nbad[reason])++;
	if ((reason == SCREW_BAD) || (loo->os_rflags & RD_PRINTALL) ||
		(rwp->ru_nbad[reason] <= rwp->ru_representative))
	  {
		if (!(loo->os_rflags & RD_BADFOUND))
		  {
			loo->os_rflags |= RD_BADFOUND;
			fprintf(rwp->ru_fpl,
			 "\tIllegal Reflections Not Written to Output:\n");
			fprintf(rwp->ru_fpl,
			 "   h   k   l  I(fit) sigma  I(sum)");
			fprintf(rwp->ru_fpl,
			 " sigma     phi   X   Y reason for exclusion\n");
		  }

		fprintf(rwp->ru_fpl,
		 "%4d%4d%4d %7.1f %5.1f %7.1f %5.1f %7.2f %4d %4d ",
		 (int)lhkl[0], (int)lhkl[1], (int)lhkl[2], rwp->ru_ifit,
		 rwp->ru_sigfit, rwp->ru_isum, rwp->ru_sigsum,
		 (double)aobp->abo_omob / 100., abptox(aobp), abptoy(aobp));
		switch (reason)
		      {
			case	STL_TOO_HIGH:
			 fputs("sin(theta) too high\n", rwp->ru_fpl);	break;
			case	STL_TOO_LOW:
			 fputs("sin(theta) too low\n", rwp->ru_fpl);	break;
			case	ICERING:
			 fputs("refln. in ice-ring\n", rwp->ru_fpl);	break;
			case	LAUE_BAD:
			 fputs("hkl illegal:pt. group\n", rwp->ru_fpl);	break;
			case	SCREW_BAD:
			 fputs("hkl illegal:screw ax.\n", rwp->ru_fpl);	break;
			case	IOSIGMA_BAD:
			 fputs("I/sigma < the cutoff\n", rwp->ru_fpl);	break;
			case	NOT_INTEGER:
			 fputs("err(h) > the cutoff\n", rwp->ru_fpl);	break;
			case	BAD_X_Y:
			 fputs("XY not in active area\n", rwp->ru_fpl);	break;
			case	FIT_NE_SUM:
			 fputs("|Ifit - Isum| too big\n", rwp->ru_fpl); break;
			case	INDEX_CHANGED:
			 fputs("recalc.hkl different\n", rwp->ru_fpl); break;
			case	NONSENSE_HKL:
			 fputs("nonsense hkl:|h| >255\n", rwp->ru_fpl); break;
			case	DEL_XYZ:
			 fputs("|dx|,|dy|,|dz|too big\n", rwp->ru_fpl);	break;
		      }
	     }
	else if (rwp->ru_nbad[reason] == rwp->ru_representative + 1)
		fputs("...and others...\n", rwp->ru_fpl);
	return 0;	/* value will be passed up the stack */
}

static int convert(OBSREF *rp, FASTOB *op, REDUWORK *rwp)
{ /*	This converts the input reflection into the FASTOB format
	and checks for various errors. */
	register int	i;
	register ABBROB	*aobp;
	int		jfr, pix_x, pix_y, pixresult;
	short		nbadi, ndiff, oldh, thkl, ohkl[3];
	double		deli, gpol, lor_ratio, ndel, nopol, pol_ratio;
	double		rni, rni2, spol, tios, thish, thislp, thisphi, trigv;
	double		tssq, vfit, vsum, xin, yin;
	float		xcm, ycm, newhkl[3], rehkl[3], ss[3], q3[3][3];
	double		locald;
	REFPARS		*xp;
	OUT_OBSSTYLE	*loo;

	loo = rwp->ru_oob;	xp = rwp->ru_rpp;
	if (rwp->ru_inttype == SCAN_FORMAT)
		refl_to_uref(rp, loo->os_xscal, loo->os_yscal);
	for (i = 3; i--; ) ohkl[i] = rp->obs_hkl[i];
	/* get rid of completely unlikely hkls before we try to bit-pack them.
	These will generally arise only from truly nonsensical data. */
	for (i = 3; i--; )
	   {
		if ((oldh = ohkl[i]) < -255 || oldh > 255)
		  {
			aobp = &(op->fo_ob);
			aobp->abo_xob = aobp->abo_yob = aobp->abo_omob = 0;
			return log_bad(aobp, ohkl, NONSENSE_HKL, rwp);
		  }
	   }
	nbadi = ndiff = 0;
	xin = ((double)rp->obs_xob) / loo->os_xscal;
	yin = ((double)rp->obs_yob) / loo->os_yscal;
	if (xin <= 3. || xin >= (xp->xdetwid -4.) ||
	 yin < 3. || yin >=(xp->xdetht -4.))
	  {
		xcm = ycm = -10.;	pixresult = -1;
	  }
	jfr = rp->obs_omob + 0.49999;
	pixresult = pix2cm(xin, yin, &xcm, &ycm, jfr);
	if (((rwp->ru_inttype == SCAN_FORMAT) ||
	 (loo->os_rflags & RD_RECOMPUTE) ||
	 (!(loo->os_rflags & RD_POLARIZED)) || (loo->os_rflags & RD_SYNCHRO) ||
	 (rwp->ru_lordel > FUZZ)) && (1 == pixresult))
	  { /* ... first, get and rotate the local scattering vector */
		xin = xcm - xp->xx0;
		yin = ycm - xp->xy0; /* mainbeam corr.*/
		rni2 = xin * xin + yin * yin + xp->xpltd2;
		rni = sqrt(rni2);
		ss[0] = xp->xplated / rni;
		ss[1] = xin / rni; ss[2] = yin / rni;
		/* rotate by the detector tilt and swing-angle matrices */
		mm31(rwp->ru_sloc, xp->xq21, ss);	/* local scattvector */
		rwp->ru_sloc[0] += 1.;	/* correct beam-direction component */
		thisphi = pirange(xp->xfrmsize * (rp->obs_omob + 0.5));
		if (loo->os_rflags & RD_POLARIZED) thislp = CORCMON *
		 fabs(rwp->ru_sloc[1]) / (MONOC2 * (1. - rwp->ru_sloc[1] *
		  rwp->ru_sloc[1]) + 1. - rwp->ru_sloc[2] * rwp->ru_sloc[2]);
		else	thislp = 20. * fabs(rwp->ru_sloc[1]) / (2. -
		 rwp->ru_sloc[1] * rwp->ru_sloc[1] -
		 rwp->ru_sloc[2] * rwp->ru_sloc[2]);
		/* apply Lorentz correction for SCAN-format data */
		if (rwp->ru_inttype == SCAN_FORMAT)
		  {
			rp->obs_ii *= thislp;	rp->obs_iunfit *= thislp;
			rp->obs_sigma *= thislp * thislp;
			rp->obs_sunfit *= thislp * thislp;
		  }
		if (loo->os_rflags & RD_RECOMPUTE)
		  {
			q3[0][0] = q3[1][1] = cos(thisphi);
			q3[0][1] = -sin(thisphi);	q3[1][0] = -q3[0][1];
			q3[2][2] = 1.;
			q3[0][2] = q3[1][2] = q3[2][0] = q3[2][1] = 0.;
				/*reference-pos scatt vector */
			mm31(rehkl, q3, rwp->ru_sloc);
				/*mult by unit cell matrix*/
			mm31(newhkl, xp->xareal, rehkl);
			/* now get the integer hkl and compare it to the
			non-integer form and to the unconverted form */
			for (i = 3; i--; )
			   {
				thkl = ((thish = newhkl[i]) < 0. ?
					thish - 0.49999 : thish + 0.49999);
				if (fabs(thkl - newhkl[i]) > xp->xerrind)
					nbadi++;
				if (thkl != ohkl[i]) ndiff++;
				ohkl[i] = thkl;	/* save the new hkl */
			   }
		  }
		if ((rwp->ru_inttype != SCAN_FORMAT) && ((loo->os_rflags &
		 RD_SYNCHRO) || !(loo->os_rflags & RD_POLARIZED)))
		  { /* polarization tweaks. Eliminate a graphite
			monochromator correction if requested, and add
			in a synchrotron correction if requested. */
		    gpol = (20. * (MONOC2 * (1. - rwp->ru_sloc[1] *
			rwp->ru_sloc[1]) + 1. - rwp->ru_sloc[2] *
			rwp->ru_sloc[2])) / (CORCMON * 2.);
		    nopol = (2. - rwp->ru_sloc[1] * rwp->ru_sloc[1] -
			rwp->ru_sloc[2]*rwp->ru_sloc[2]) / 2.;
		    if (loo->os_rflags & RD_SYNCHRO)
		      {
			if (loo->os_rflags & RD_SYNCHRO_V) /* vert.goniostat */
				spol = 1. - rwp->ru_sloc[1] * rwp->ru_sloc[1];
					/* horizontal: */
			else	spol = 1. - rwp->ru_sloc[2] * rwp->ru_sloc[2];
			pol_ratio = (spol < DFUZZ) ? 1. :
			 (loo->os_rflags & RD_POLARIZED) ?
				nopol / spol : gpol / spol;
			if (pol_ratio > rwp->ru_maxpr)
				rwp->ru_maxpr = pol_ratio;
			if (pol_ratio < rwp->ru_minpr)
				rwp->ru_minpr = pol_ratio;
		      }
		    else	pol_ratio = gpol / nopol;
		    rp->obs_ii *= pol_ratio; rp->obs_iunfit *= pol_ratio;
		    rp->obs_sigma *= pol_ratio * pol_ratio;
		    rp->obs_sunfit *= pol_ratio * pol_ratio;
		    ndel = ((double)rp->obs_shift) * pol_ratio;
		    rp->obs_shift = (ndel < 0.) ?
			ndel - 0.49999 : ndel + 0.49999;
		    ndel = ((double)rp->obs_bgunfit) * pol_ratio;
		    rp->obs_bgunfit = (ndel < 0.) ?
			ndel - 0.49999 : ndel + 0.49999;
		    ndel = ((double)rp->obs_widob) * pol_ratio * pol_ratio;
		    rp->obs_widob = (ndel < 0.) ?
			ndel - 0.49999 : ndel + 0.49999;
		    ndel = ((double)rp->obs_reflno) * pol_ratio;
		    rp->obs_reflno = (ndel < 0.) ?
			ndel - 0.49999 : ndel + 0.49999;
		  }
		if ((rwp->ru_lordel > FUZZ) &&
		 (lor_ratio = fabs(rwp->ru_sloc[1])) > FUZZ)
		  {
			lor_ratio =  1. + rwp->ru_lordel / lor_ratio;
			rp->obs_ii *= lor_ratio;
			rp->obs_iunfit *= lor_ratio;
			rp->obs_sigma *= lor_ratio * lor_ratio;
			rp->obs_sunfit *= lor_ratio * lor_ratio;
		  }
	  }
	if (rwp->ru_trigc > FUZZ)
	  {
		trigv = 1. - rwp->ru_trigc * cos(PIX2 * rp->obs_omob);
		rp->obs_ii *= trigv;	rp->obs_iunfit *= trigv;
		rp->obs_shift = ((double)rp->obs_shift) * trigv;
		rp->obs_bgunfit = ((double)rp->obs_bgunfit) * trigv;
		trigv *= trigv;
		rp->obs_sigma *= trigv;	rp->obs_sunfit *= trigv;
		rp->obs_widob = ((double)rp->obs_widob) * trigv;
		rp->obs_reflno = ((double)rp->obs_reflno) * trigv;
	  }
	/* we calculate sin(theta) before checking for reindexing so
	 we don't end up with wrong answers after reindexing... */
	op->fo_stl = stlval(ohkl, xp);
	for (i = 3; i--; ) ohkl[i] = rp->obs_hkl[i];
	/* get rid of completely unlikely hkls before we try to bit-pack them.
	These will generally arise only from truly nonsensical data. */
	for (i = 3; i--; )
	   {
		if ((oldh = ohkl[i]) < -255 || oldh > 255)
		  {
			aobp = &(op->fo_ob);
			aobp->abo_xob = aobp->abo_yob = aobp->abo_omob = 0;
			return log_bad(aobp, ohkl, NONSENSE_HKL, rwp);
		  }
	   }
	if (loo->os_rflags & RD_REINDEX)
	  {
		for (i = 3; i--; ) rehkl[i] = ohkl[i];
		mm31(newhkl, rwp->ru_matrei, rehkl);
		for (i = 3; i--; )
		   {
			newhkl[i] += rwp->ru_vecrei[i];
			if (newhkl[i] < 0) newhkl[i] -= 0.0001;
			else newhkl[i] += 0.0001;
			ohkl[i] = newhkl[i];
		   }
	  }
	mpack(ohkl,op->fo_hkl, xp->xispgp); /* pack as multiref */
	symtr(op->fo_hkl, xp->xispgp); /*symmetry-transform result*/
	if (op->fo_hkl[3] % 2)		/* record the Bijvoet bit */
	  {
		op->fo_nobbij = 1;	op->fo_nobpri = 0;
	  }
	else {
		op->fo_nobpri = 1;	op->fo_nobbij = 0;
	     }
	op->fo_nob = 1;
	/* We calculate the scaling group ("shift") number of the
	 observation after we've read all the data in.  All we're going to
	 stick in that slot now is a marker to remember whether the
	 observation was made in the top half of the detector or the bottom. */
	aobp = &(op->fo_ob);
	pix_x = rp->obs_xob / (int)loo->os_xscal; /* observed X in pixels */
	pix_y = rp->obs_yob / (int)loo->os_yscal; /* observed Y in pixels */
	if (pix_y < rwp->ru_hto2) aobp->abo_shift = 0;	/*detector's top half*/
	else { /* bottom half */
		aobp->abo_shift = 1; pix_y -= rwp->ru_hto2;
	     }
	aobp->abo_xob = (u_char)(pix_x / rwp->ru_widrat);
	aobp->abo_yob = (u_char)(pix_y / rwp->ru_htrat);
	thisphi = to360(rp->obs_omob * xp->xfrmsize / RADC);
	aobp->abo_omob = (short)(100. * thisphi);
	rwp->ru_ifit = rp->obs_ii;	vfit = rp->obs_sigma;
	rwp->ru_isum = rp->obs_iunfit;	vsum = rp->obs_sunfit;
	if (rwp->ru_inttype & ABG_EITHER)
	  {
		rwp->ru_ifit += ((double)rp->obs_shift) / 16.;
		rwp->ru_isum += ((double)rp->obs_bgunfit) / 16.;
		vfit += ((double)rp->obs_widob) / 8.;
		vsum += ((double)rp->obs_reflno) / 8.;
	  }
	rwp->ru_sigfit = sqrt(vfit);	rwp->ru_sigsum = sqrt(vsum);
/* Use the requested type of intensity: shapefit, integrated, or `best'. */
	switch (rwp->ru_inttype)
	  {
		case	SIMPLE_SUMMATION:	case	ABG_SUM:
		case	SCAN_FORMAT:		aobp->abo_ii = rwp->ru_isum;
			aobp->abo_sigma = sqrt(vsum + rwp->ru_isum *
			 (rwp->ru_machineerr * rwp->ru_isum));	break;
		case	PROFILE_FITTED:		case	ABG_FIT:
			aobp->abo_ii = rwp->ru_ifit;
			aobp->abo_sigma = sqrt(vfit + rwp->ru_ifit *
			 (rwp->ru_machineerr * rwp->ru_ifit));	break;
		case	BEST_I_OVER_SIGMA: if (rwp->ru_isum * rwp->ru_sigfit >
			 rwp->ru_ifit * rwp->ru_sigsum)
			  {
				aobp->abo_ii = rwp->ru_isum;
				aobp->abo_sigma = sqrt(vsum + rwp->ru_isum *
				 (rwp->ru_machineerr * rwp->ru_isum));
			  }
			else  {
				aobp->abo_ii = rwp->ru_ifit;
				aobp->abo_sigma = sqrt(vfit + rwp->ru_ifit *
				 (rwp->ru_machineerr * rwp->ru_ifit));
			      }	break;
		case	SUMMED_IF_STRONG:
		case	SUMSTRONG_ABG:	/* summation if I/sigma large */
			if ((rwp->ru_sigsum > FUZZ) &&
			 (rwp->ru_isum / rwp->ru_sigsum > rwp->ru_strongcut))
			   {
				aobp->abo_ii = rwp->ru_isum;
				aobp->abo_sigma = sqrt(vsum + rwp->ru_isum *
				 (rwp->ru_machineerr * rwp->ru_isum));
			   }
			else  {
				aobp->abo_ii = rwp->ru_ifit;
				aobp->abo_sigma = sqrt(vfit + rwp->ru_ifit *
				 (rwp->ru_machineerr * rwp->ru_ifit));
			      } break;
	  }
	if (aobp->abo_sigma < MINSIGMA) aobp->abo_sigma = MINSIGMA;
	if (loo->os_rflags & RD_MAKE_CENTROIDS)
	  { /* do this stuff if we want a "centroids" output */
		if (aobp->abo_ii > aobp->abo_sigma * rwp->ru_ioscut)
			return out_centroid(rp, loo, rwp->ru_fpo, rwp->ru_fpl);
		else return 0;
	  }
	aobp->abo_chamno = 0;	/* reserved for future use */
	aobp->abo_runno = xp->xrunno;	/* an input parameter*/
	if (op->fo_stl > rwp->ru_maxslegal)
		return log_bad(&(op->fo_ob), ohkl, STL_TOO_HIGH, rwp);
	if (op->fo_stl < rwp->ru_minslegal)
		return log_bad(&(op->fo_ob), ohkl, STL_TOO_LOW, rwp);
	locald = (op->fo_stl > FUZZ) ? 1. / (2. * op->fo_stl) : 999.;
	if ((loo->os_rflags & RD_KILL_ICERING) && isinring(locald))
		return log_bad(&(op->fo_ob), ohkl, ICERING, rwp);
	if ((loo->os_rflags & RD_IOSCUT) && aobp->abo_ii < rwp->ru_ioscut *
	 aobp->abo_sigma) return log_bad(&(op->fo_ob), ohkl, IOSIGMA_BAD, rwp);
	if ((loo->os_rflags & RD_KILL_DELXYZ) &&
		((abs(rp->obs_xob - rp->obs_xpred) > loo->os_delxlim) ||
		 (abs(rp->obs_yob - rp->obs_ypred) > loo->os_delylim) ||
		 (fabs(rp->obs_omob - rp->obs_ompred) > loo->os_delomlim)))
			return log_bad(&(op->fo_ob), ohkl, DEL_XYZ, rwp);
	if (!keephkl(ohkl, rwp->ru_excluc))
		return log_bad(&(op->fo_ob), ohkl, LAUE_BAD, rwp);
	if (!screw_ok(ohkl, xp->xispgp))
		return log_bad(&(op->fo_ob), ohkl, SCREW_BAD, rwp);
	/* V.1.08 : we require that the observed centroid fall within the
	 active area of the detector.  This allows us to trim out reflections
	 simply by manipulating the AMASK database. */
	if (1 != pixresult) return log_bad(&(op->fo_ob), ohkl, BAD_X_Y, rwp);
	if (loo->os_rflags & RD_DEL_BIG_IDIFF) /* implement "-d" option */
	  {
		deli = rwp->ru_ifit - rwp->ru_isum;
		if ((deli / (vfit + vsum)) * deli > rwp->ru_delilim)
			return log_bad(&(op->fo_ob), ohkl, FIT_NE_SUM, rwp);
	  }
	else if (loo->os_rflags & RD_SCALE_SUMFIRST)
	  {
		deli = rwp->ru_ifit - rwp->ru_sumscale * rwp->ru_isum;
		if ((deli / (vfit + rwp->ru_sumscsq * vsum)) * deli >
		 rwp->ru_delilim)
			return log_bad(&(op->fo_ob), ohkl, FIT_NE_SUM, rwp);
	  }
	if ((loo->os_rflags & RD_KILL_IF_REINDEXED) && ndiff)
	  {
		(rwp->ru_nchanged)++;
		return log_bad(&(op->fo_ob), ohkl, INDEX_CHANGED, rwp);
	  }
	if (nbadi && (loo->os_rflags & RD_KILL_NOT_INT))
		return log_bad(&(op->fo_ob), ohkl, NOT_INTEGER, rwp);
	else if (ndiff) (rwp->ru_nchanged)++;
	if (aobp->abo_ii > rwp->ru_maxinten) rwp->ru_maxinten = aobp->abo_ii;
	/* we accumulate the I000, B, D2 sums only if we get this far */
	if (aobp->abo_sigma > FUZZ)
	  {
		if ((tios = aobp->abo_ii / aobp->abo_sigma) > DFUZZ)
		  {
			tssq = rwp->ru_lamco * rp->obs_ssq; tios = log(tios);
			rwp->ru_iose[0] += 1.;	rwp->ru_iose[1] += tssq;
			rwp->ru_iose[2] += tssq * tssq;	rwp->ru_iose[3] += tios;
			rwp->ru_iose[4] += tssq * tios;
		  }
	  }
	return 1;
}

static void announce_delomega(FILE *fp, double oldfmper,
	int oldmax, REDUWORK *rwp)
{ /*	Inform the user that the delta-omega value has been changed */
	fprintf(fp, "Note: width of scaling groups must be narrowed\n");
	fprintf(fp, " so that fewer than 200 scaling groups are used\n");
	fprintf(fp,
 " specification would have created %4d scaling groups of %6.2f frames\n",
	 oldmax + oldmax + 1, oldfmper);
	fprintf(fp, "Imposed scaling group width: %6.2f frames = %8.4f deg\n",
		rwp->ru_fmpergroup, rwp->ru_omegainterval);
	fprintf(fp, "Roughly %4d scaling groups will be created\n",
		rwp->ru_maxeven + rwp->ru_maxeven + 1);
	if (fp != stderr) xeflush(fp);
}

static void unique_refls(REDUWORK *rwp)
{ /*	This goes through the list of "FASTOB" refls and does some subtle
	reprocessing.  First, it sorts them into increasing hkl order.
	Then it looks at each observation to see if its hkl
	is the same as the previous one, the Bijvoet to it, or a new one.
	If the refl is a new one, the count of refls is incremented.
	If it's Bijvoet to the previous, the count of unmerged unique
	reflections is incremented.  If it's not new, the observation count
	within the old FASTOB structure is incremented.  Meanwhile, a mean
	intensity and sigma are computed for each completed reflection,
	and all the data are rescaled if necessary so that the largest
	recorded observation intensity value is MAXOBINT.	*/
	register FASTOB	*currefp, *orefp;
	register ABBROB	*aobp;
	register int	i, newshift;
	int		deltime, oldmaxeven;
	u_char		oldhkl[4];
	sichar		indexdif, signdif;
	double		delomfm, framdeg, oldfmper, rescaler;
	double		sumi, thisi, weight, wrapfm, wtsum;
	CRYSTAL		*lxp;
	OUT_OBSSTYLE	*loo;
	static char	sortmsg[] =
		" %4d seconds required to sort %6d reflections\n";
	static char	maximsg[] = "Largest intensity value input = %10.2f\n";
	static char	rescmsg[] =
 "Intensities and sigmas will be rescaled so that max intensity = %8.1f\n";
	static char	norescmsg[] = "No rescaling will be performed\n";
	static char	reorgmsg[] =
		" %4d seconds required to reorganize %6d reflections\n";

	if (rwp->ru_nconverted > 1) qsort((AVOID *)rwp->ru_orefb,
	 (size_t)(rwp->ru_nconverted), sizeof (FASTOB), (QPE)hklcmp);
	loo = rwp->ru_oob;	rwp->ru_tl0 = NOW;
	deltime = rwp->ru_tl0 - rwp->ru_tl1;
	fprintf(rwp->ru_fpl, sortmsg, deltime, rwp->ru_nconverted);
	fprintf(stdout,      sortmsg, deltime, rwp->ru_nconverted);
	fprintf(stdout, maximsg, rwp->ru_maxinten);
	fprintf(rwp->ru_fpl, maximsg, rwp->ru_maxinten);
	if (rwp->ru_maxinten > MAXOBINT)
	  {
		rescaler = rwp->ru_maxinten / MAXOBINT;
		fprintf(stdout, rescmsg, MAXOBINT);
		fprintf(rwp->ru_fpl, rescmsg, MAXOBINT);
	  }
	else {
		rescaler = 1.;	fprintf(stdout, norescmsg);
		fprintf(rwp->ru_fpl, norescmsg);
	      }
	/* figure out what the lowest-numbered scaling group will be.
	 In general it will be 2 * (first frame * deg/frame) + 2, unless
	 that would make the highest-numbered scaling group greater than 200.
	 We currently check the see whether there might otherwise be 200
	 or more scaling groups; if so, we start at two and (if necessary)
	 modify the number of frames per degree so that we can fit
	 the whole run into 200 sscaling groups */
	rwp->ru_delfrm = (rwp->ru_rhp)->run_fmmax + 0.1 -
		(rwp->ru_rhp)->run_fmmin;
	framdeg = (rwp->ru_rpp)->xfrmsize / RADC;
	wrapfm = 360. / fabs(framdeg);	/* for wraparound in omega */
	rwp->ru_npairs = (int)(rwp->ru_delfrm * fabs(framdeg) /
	 rwp->ru_nomwidth + 0.49999);
	if (rwp->ru_npairs < 1) rwp->ru_npairs = 1;
	rwp->ru_fmpergroup = rwp->ru_delfrm / ((double)(rwp->ru_npairs));
	rwp->ru_omegainterval = rwp->ru_fmpergroup * fabs(framdeg);
	rwp->ru_shoffset = 1 + (int)((rwp->ru_rhp)->run_fmmin /
		rwp->ru_fmpergroup);
	rwp->ru_maxeven = rwp->ru_shoffset + rwp->ru_npairs - 1;
	if (rwp->ru_maxeven > 98)
	  {
		rwp->ru_shoffset = 1;	rwp->ru_maxeven = rwp->ru_npairs;
		if (rwp->ru_maxeven > 98) /* still too big? recalc. */
		  {
			oldfmper = rwp->ru_fmpergroup;
			oldmaxeven = rwp->ru_maxeven;
			rwp->ru_npairs = 96; /* max # of scale-group pairs */
			rwp->ru_fmpergroup = rwp->ru_delfrm /
				((double)(rwp->ru_npairs));
			/* round up # of frames */
			if (rwp->ru_fmpergroup > 1.5) rwp->ru_fmpergroup =
			 (int)(rwp->ru_fmpergroup + 0.99999);
			else {
				rwp->ru_maxeven = 4. *
					(rwp->ru_fmpergroup +0.249999);
				rwp->ru_fmpergroup = 0.25 *
					(double)(rwp->ru_maxeven);
			     }
			rwp->ru_omegainterval = rwp->ru_fmpergroup *
				fabs(framdeg);
			rwp->ru_npairs = rwp->ru_delfrm / rwp->ru_fmpergroup +
				1.49999;
			rwp->ru_maxeven = rwp->ru_npairs;
			announce_delomega(rwp->ru_fpl,  oldfmper,
			 oldmaxeven, rwp);
			announce_delomega(stdout, oldfmper, oldmaxeven, rwp);
		  }
	  }
/* process the very first observation. */
	currefp = rwp->ru_orefb;
	for (i = 4; i--; ) oldhkl[i] = currefp->fo_hkl[i];
	aobp = &(currefp->fo_ob);
	aobp->abo_ii /= rescaler;	aobp->abo_sigma /= rescaler;
	weight = aobp->abo_sigma;	weight = 1./(weight * weight);
	thisi = aobp->abo_ii;		wtsum = weight;
	sumi = thisi * weight;		wtsum = weight;
	delomfm = ((double)aobp->abo_omob) * 0.01 / framdeg -
		(rwp->ru_rhp)->run_fmmin;
	while (delomfm < -2.) delomfm += wrapfm;
	newshift = rwp->ru_shoffset + (int)(delomfm / rwp->ru_fmpergroup);
	if (newshift < rwp->ru_shoffset) newshift = rwp->ru_shoffset;
	aobp->abo_shift += newshift + newshift;
	if (newshift > rwp->ru_maxeven) /* This shouldn't be necessary */
	  {
		fprintf(stderr,
 "[%3d,%3d,%3d]@[%4.0f,%4.0f,%7.2f]:shift=%3d; max should have been %3d\n",
		 currefp->fo_hkl[0], currefp->fo_hkl[1], currefp->fo_hkl[2],
		 ((double)aobp->abo_xob) / loo->os_xscal,
		 ((double)aobp->abo_yob) / loo->os_yscal,
		 ((double)aobp->abo_omob) * 0.01, aobp->abo_shift,
		 1 + rwp->ru_maxeven + rwp->ru_maxeven);
	  }
	/* Now process all the rest */
	lxp = rwp->ru_xhp;
	for (orefp = rwp->ru_orefb + 1; orefp < rwp->ru_orefend; orefp++)
	   {
		indexdif = (orefp->fo_hkl[0] - oldhkl[0]) ||
		 (orefp->fo_hkl[1] - oldhkl[1])||(orefp->fo_hkl[2] - oldhkl[2]);
		aobp = &(orefp->fo_ob);		aobp->abo_ii /= rescaler;
		aobp->abo_sigma /= rescaler;	weight = aobp->abo_sigma;
		weight = 1./(weight * weight);	thisi = aobp->abo_ii;
		delomfm = ((double)aobp->abo_omob) * 0.01 / framdeg -
			(rwp->ru_rhp)->run_fmmin;
		while (delomfm < -2.) delomfm += wrapfm;
		newshift = rwp->ru_shoffset +
			(int)(delomfm / rwp->ru_fmpergroup);
		if (newshift < rwp->ru_shoffset) newshift = rwp->ru_shoffset;
		aobp->abo_shift += newshift + newshift;
		if (newshift > rwp->ru_maxeven)	/* temporary check */
		  {
			fprintf(stderr,
 "[%3d,%3d,%3d]@[%4.0f,%4.0f,%7.2f]:shift=%3d; max should have been %3d\n",
			 currefp->fo_hkl[0], currefp->fo_hkl[1],
			 currefp->fo_hkl[2],
			 ((double)aobp->abo_xob) / loo->os_xscal,
			 ((double)aobp->abo_yob) / loo->os_yscal,
			 ((double)aobp->abo_omob) * 0.01, aobp->abo_shift,
			 1 + rwp->ru_maxeven + rwp->ru_maxeven);
		  }
		if (!indexdif)	/* old indices, except (perhaps) signs */
		  {
			signdif = orefp->fo_hkl[3] - oldhkl[3]; /* get signs */
			/* if Bijvoet to previous, set signdif to 0
			and increment the unmerged-total-refl count */
			if (signdif == 1 && (orefp->fo_hkl[3] % 2))
			  {
				signdif = 0; (lxp->xt_nrefls)++;
			  }
			if (!signdif) /* signdif==0 means an old reflection */
			  {
				currefp->fo_nob++; orefp->fo_nob = 0;
				if (orefp->fo_hkl[3]%2) currefp->fo_nobbij++;
				else	currefp->fo_nobpri++;
				wtsum += weight;	sumi += thisi * weight;
				for (i = 4; i--;) oldhkl[i] = orefp->fo_hkl[i];
				continue; /* skip rest of this pass */
			  }
		  }
		/* any time we get here we increment both global counters,
		reset the "old reflection" value, and compute the means */
		for (i = 4; i--; ) oldhkl[i] = orefp->fo_hkl[i];
		(lxp->xt_nrefls)++;	(lxp->xt_nmerged)++;
		if (wtsum > DFUZZ)
		  {
			sumi /= wtsum;	if (sumi < 0.) sumi = 0;
			else if (sumi > 65534.) sumi = 65534.;
			wtsum = 1. / wtsum;	if (wtsum < 1.) wtsum = 1.;
			currefp->fo_ibar = sumi + 0.4999;
			currefp->fo_isig = sqrt(wtsum) + 0.499;
		  }
		else {
			currefp->fo_ibar = 0.; currefp->fo_isig = 1.; }
	/* all reflections should have their Bijvoet bit "off" on output */
		currefp->fo_hkl[3] &= ~1;	currefp = orefp;
		wtsum = weight;			sumi = thisi * weight;
	   }
/* we've gone all the way through the file.  Process the last refl. */
	if (wtsum > DFUZZ)
	  {
		sumi /= wtsum;	if (sumi < 0.) sumi = 0.;
		else if (sumi > 65534.) sumi = 65534.;
		wtsum = 1. / wtsum;	if (wtsum < 1.) wtsum = 1.;
		currefp->fo_ibar = sumi + 0.4999;
		currefp->fo_isig = sqrt(wtsum) + 0.4999;
	  }
	else {
		currefp->fo_ibar = 0.; currefp->fo_isig = 1.; }
	currefp->fo_hkl[3] &= ~1;	/* turn off the Bijvoet bit */
	(lxp->xt_nrefls)++;	(lxp->xt_nmerged)++;
	rwp->ru_tl1 = NOW;	deltime = rwp->ru_tl1 - rwp->ru_tl0;
	fprintf(rwp->ru_fpl, reorgmsg, deltime, (int)(lxp->xt_nmerged));
	fprintf(stdout, reorgmsg, deltime, (int)(lxp->xt_nmerged));
}

static int endred(int irv, REDUWORK *rwp)
{ /*	This eliminates memory for reserved stuff and closes files */
	if ((rwp->ru_oob)->os_rflags & RD_KILL_ICERING) (void)manageice(0);
	if ((stdout != rwp->ru_fpl) && (NULL != rwp->ru_fpl))
		(void)fclose(rwp->ru_fpl);
	if (NULL != rwp->ru_fpi) (void)fclose(rwp->ru_fpi);
	rwp->ru_fpi = NULL;
	memzap((AVOID **)&((rwp->ru_oob)->os_cenpbin),
	 "oob->os_cenpbin in redurefl");
	memzap((AVOID **)&(rwp->ru_orefb), "rwp->ru_orefb in redurefl");
	memzap((AVOID **)&(rwp->ru_rpp), "rwp->ru_rpp in redurefl");
	memzap((AVOID **)&(rwp->ru_xhp), "miscellaneous memory in redurefl");
	rwp->ru_fpl = NULL;	globalkill();	return irv;
}

int reducemain(int argc, char *argv[])
{ /*	Mainline for converting a UREFLS into MULTIREF format.
	It needs a "uparams" file to get a unit cell matrix and a stepsize. */
	register FASTOB	*orefp;
	int		deltim, result;
	OUT_OBSSTYLE	*loo;
	static char	polmsg[] = "min, max pol ratio = %8.4f, %8.4f\n";
	static char	procmsg[] =
		" %4d seconds required to process %6d reflections\n";
	static OBSREF	qrf;
	static REDUWORK	rws;

	/* set up the run */
	if (EOK != (result = red_setup(argc, argv, &rws)))
		return endred(result, &rws);	loo = rws.ru_oob;
	if (loo->os_rflags & RD_SCALE_SUMFIRST) /* get <I(fit)/I(sum) */
	  {
		rws.ru_sumscale = get_iratio(rws.ru_fpi, rws.ru_fpl);
		rws.ru_sumscsq = rws.ru_sumscale * rws.ru_sumscale;
	  }
/* go through the data, converting the refls */
	if (loo->os_rflags & RD_ANGLE_RESTRICTION)
	  {
		for (rws.ru_nread = 0, orefp = rws.ru_orefb; 0 < xgread(
		 (void *)&qrf, XG_OBSREF, rws.ru_fpi); (rws.ru_nread)++)
		   {
			if (qrf.obs_ompred >= loo->os_minfm)
			  {
				if (qrf.obs_ompred > loo->os_maxfm + 0.9999)
					break;
				if ((result = convert(&qrf, orefp, &rws)))
				  {
					if (result > 0) statrefl(rws.ru_rhp,
						orefp++, qrf.obs_omob);
					else return endred(
					 (result < -1 ? EWRITERR : EOK), &rws);
				  }
			  }
		   }
	  }
	else	for (rws.ru_nread = 0, orefp = rws.ru_orefb; 0 < xgread(
		 (void *)&qrf, XG_OBSREF, rws.ru_fpi); (rws.ru_nread)++)
		   if (convert(&qrf, orefp, &rws))
			statrefl(rws.ru_rhp, orefp++, qrf.obs_omob);
	(void)fclose(rws.ru_fpi);	rws.ru_fpi = NULL;
	if (rws.ru_maxpr > FUZZ)
	  {
		fprintf(stdout, polmsg, rws.ru_minpr, rws.ru_maxpr);
		fprintf(rws.ru_fpl, polmsg, rws.ru_minpr, rws.ru_maxpr);
	  }
	rws.ru_nconverted = orefp - rws.ru_orefb;
	rws.ru_orefend = orefp;
	rws.ru_tl1 = NOW;	deltim = rws.ru_tl1 - rws.ru_tl0;
	fprintf(stdout, procmsg, deltim, rws.ru_nconverted);
	fprintf(rws.ru_fpl, procmsg, deltim, rws.ru_nconverted);
	if (loo->os_rflags & RD_MAKE_CENTROIDS)
	  {
		(void)centroid_summary(loo, rws.ru_fpl);
		(void)fclose(rws.ru_fpo);
		rws.ru_fpo = NULL;	return endred(EOK, &rws);
	  }
/* sort the reflections in hkl order and set them up for output */
	if (loo->os_rflags & RD_DOSORT) unique_refls(&rws);
/* open the output file, process the header information, and write it out */
	if (qerror((AVOID *)(rws.ru_fpo = xgopen("multiref", "w")),
	 "open output multiref")) result = EOPENERR;
	else if (EOK == (result = proc_heads(rws.ru_fpo, rws.ru_npairs,
	 rws.ru_xhp, rws.ru_rhp))) result = out_refls(&rws); /* output refls */
	return endred(result, &rws);
}

/* end of redurefl.c */
