/* redutils.c -- facilitating reduction & filtering of UREFLS to MULTIREF

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	23 Mar 2007: bugfix in manageice call: it _should_ return non-null
	14 Feb 2007: reduce_defaults: we don't need stype, ssize
	28 Nov 2006: makeq21 call updated
	24 Jun 2006: include xgpostint.h rather than reddefs.h
	12 Jun 2006: old redurefl.c and ../xglib/xgenmisc.c reorganized
	 to create redurefl.c, redutils.c, and ../include/reddefs.h
 earlier relevant changes to redurefl.c and ../xglib/xgenmisc.c:
	20 Nov 2005: inline argument declarations and strict prototyping
	12 Oct 2005: housekeeping
	 2 Sep 2005: break issued in -q flag regardless of sub-option
	11 May 2005: check return status from xgjournal
	24 Apr 2005: move redurefl.c source from ../integrate to ../utilities
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	12 Jan 2005: fix bug in -a option
	28 Aug 2004: routines used only here are declared as static;cosmetics
	16 Aug 2004: anal_ios() bugfix for cutoff; tell_i0b(), iosresults()
	 used in analyzing and outputting I0/B/D(1.4) values for input data
	13 Aug 2004: use OUT_OBSSTYLE structure to pass args around;
		move a lot of routines from here to ../xglib/xgen{file,misc}.c
	 7 Aug 2004: correct bug in red_setup so non-flag args are okay
	 4 Aug 2004: in reduce_defaults()
		round scaling-group  width up by 0.9999 rather than 0.49999.
	27 Jul 2004: call dimscale() to get detector dimensions in get_upars
	15 Jul 2004: run-time options handled directly in red_setup() using
		flagproc() and other standard tools; -q<val> and -z<val>
		used to handle Booleans for which we had run out of symbols
	 2 Jul 2004: redundant `break' removed from case 'N' in redswitch()
	19 Jun 2004: defaults handled in reduce_defaults();
	 default run #, default scaling group width computed based on
	 properties of the input UREFLS file;
	 resolution cutoff in anal_ios() allows interpolation;
	 bugfix in sum_args(): it reports on CENTROIDS file only when it should
	 9 Feb 2004: xtp->xt_version in partohead() set to 500 if undefined
	 8 Feb 2004: correct sin(theta)/lambda calculation:
		rec[][] is now unitless, so we don't divide by lambda
	28 Jan 2004: many static globals converted to locals
	21 Jan 2004: license message inserted and initial comment reordered
	 7 Jan 2004: REFPARS structures and args to routines replace globals;
		refparin() and refparout() handle UPARAMS input and output;
		makeq21() handles tilt & swing matrix
	13 Dec 2003: eliminite unused vbl nicer in red_setup()
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	18 Aug 2003: created xgenmisc.c from xgenio.c
	17 Aug 2003: handle ice-rings through calls to external routines
	 5 Jan 2003: reorganized source-file header
	18 Apr 2002: get rid of duplicate observations before output
		unless the user specifies otherwise
	25 Mar 2002: removed MSI attachments
	10 May 2001: (?) set up feature such that multiple measurements of
		the same actual data space are suppressed; -y cancels this.
	 5 May 2001: (?) new meaning for "-m": Specifies resolution value
		for which <I/sigma> falls below the specified number.
	31 May 1995: "register" -> "register int" throughout;
		%d replaces %ld; typecast tl1-tl0;
		re-cast several memory-handling routines with size_t's
	12 Apr 1995: add option to exclude ice-ring
	 2 Mar 1995: provide for plane backgrounds; diagnostics to stdout;
		shared-memory alloc and free
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 1 Dec 1994: dettype() declared before use
	 4 Nov 1994: detector type and X-GEN version # saved
		to CRYSTAL struct. of MULTIREF
	 5 Oct 1994: use xgread for OBSREF, REFPARS;
		use xgwrite for TROID, MERGEDREF
	30 Sep 1994: fix list of input files
	30 Sep 1994: save floating-point stepsize in DATARUN structure;
		blank CRYSTAL & DATARUN structures before manipulating them
	 5 May 1994: source-file headers; POLYDET->XGEN;
		"vsort.h" included as absolute reference
	16 Mar 1994: a few I/O changes and global variables moved around
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version of older files
	15 Mar 1992: add "-x" option
	18 Jan 1988: compute the scaling group numbers after
		all the data have been read in so that they can be of
		almost precisely equal size and we won't end up with
		almost unpopulated scaling groups on the edges.
	23 Aug 1987: add "-o" option,
		thereby making the program CVTREFL unnecessary
	30 Jun 1987: add "-a" option
	10 Jun 1987: add "-b" and "-z" options and fix bugs in "-k", "-r".
	 3 May 1987: add "-d" option.
	30 Apr 1987: make sure the reflection is within the
		detector boundaries defined by the AMASK file.
	 1 Dec 1986: incorporate improvements to cvtrefl here.
	 1 Jun 1986: original code

   Purpose: utilities associated with converting a UREFLS to a MULTIREF
	or to a CENTROIDS file
   Cross references: redurefl.c, ../include/reddefs.h

*************************************************************************/
#define		NEED_STAT	1
#include	"xgpostint.h"
#define		NIOSSHELLS	10	/* # of shells for I/sigma calcs */
#define		DEF_MINSIGOI	0	/* default min value of sigma/I */
#define		DEF_NCENPERBIN	10	/* # of centroids per (X,Y) bin */
#define		MIN_IOSCUT	3.	/* min I/sigma allowed for centroids*/
#define		DEF_REPRESENT	10	/* def.# violators printed per type */
#define		DEF_DELI_LIMIT	10.	/* see "-d" flag */
#define		MIN_DELI_LIMIT	0.5	/* see "-d" flag */
#define		DEF_NOMWIDTH	5.	/* see "-w" flag */

 typedef struct {
		int	resh_binno, resh_nobs;
		double	resh_meanios, resh_dlow, resh_dhigh;
		} IOSIG_SHELL;

static int get_upars(REFPARS **drp, OUT_OBSSTYLE *oob)
{ /*	This reads in and cleans up various pieces of information from
	a UPARAMS file and puts the results in an OUT_OBSSTYLE structure */
	register int	i;
	int		nbread, retv;
	REFPARS		*rp;
	
	if (NULL == *drp)
	  {
		if (NULL == (*drp = (REFPARS *)callop(1, sizeof (REFPARS),
		 "*drp in get_upars"))) return EALLOC;
	  }
	rp = *drp; if (EOK != (i = refparin(rp, &nbread))) return i;
	if (nbread < 256) /* we don't really care if this is an old file */
	  {
		rp->xstart[2] = rp->xstart[1] = rp->xstart[0] = 0;
	  }
	if (rp->xisys == 7 && rp->xispgp < 200) rp->xispgp += 100;
	makeq21(rp);	/* tilt/swing rotation matrix */
	/* if we're going to have to recompute hkl indices,
	we'll have to convert pixels to centimeters.  So we need a spline...*/
	if (NULL == (oob->os_spl = gspline(5, &retv))) return retv;
	rp->xdetwid = (oob->os_spl)->v_detw;
	rp->xdetht = (oob->os_spl)->v_deth;
	oob->os_ncbx = ((oob->os_spl)->v_ncol / 2 + 1); /* # of X bins */
	oob->os_ncby = ((oob->os_spl)->v_ncol / 2 + 1); /* # of Y bins */
	oob->os_ncbins = oob->os_ncbx * oob->os_ncby; /* total # of bins */
	oob->os_xscal = dimscale(rp->xdetwid);
	oob->os_yscal = dimscale(rp->xdetht);
	if (oob->os_rflags & RD_KILL_DELXYZ)
	  {
		oob->os_delxlim = (short)(rp->xerrx * oob->os_xscal);
		oob->os_delylim = (short)(rp->xerry * oob->os_yscal);
		oob->os_delomlim = rp->xerrphi;
	  }
	else {
		oob->os_delxlim = oob->os_delylim = oob->os_delomlim = 9999.;
	     }
	/* globals defining offset from corner of an (x,y) bin to its center */
	oob->os_xdel = (oob->os_spl)->v_xrpx / 2.;
	oob->os_ydel = (oob->os_spl)->v_yrpx / 2.;	return EOK;
}

static void report_iosresults(FILE *fp, int binnum, double newdmin,
 double rmaxg, double deld, IOSIG_SHELL *uiosb)
{ /*	This reports on the <I/sigma> values in shells to the stream fp */
	int		i;
	IOSIG_SHELL	*iosp;

	fprintf(fp, "<I/sigma> in shells:\n");
	fprintf(fp, " bin   # obs     Dmin     Dmax <I/sigma>\n");
	for (i = 0, iosp = uiosb; i < NIOSSHELLS; i++, iosp++)
	  if (iosp->resh_nobs > 0)
	   {
		fprintf(fp, " %3d %7d %8.3f %8.3f %9.4f\n", i, iosp->resh_nobs,
		 iosp->resh_dhigh, iosp->resh_dlow, iosp->resh_meanios);
	   }
	fprintf(fp, " ALL %7d %8.3f %8.3f %9.4f\n", iosp->resh_nobs,
	 iosp->resh_dhigh, iosp->resh_dlow, iosp->resh_meanios);
	if (binnum >= NIOSSHELLS) return;
	if (newdmin > rmaxg)
	  {
		fprintf(fp, "Reflections in bins %2d to %2d will be excluded",
		 binnum, NIOSSHELLS-1);
		if (deld > DFUZZ) fprintf(fp,
		 ", plus part of bin %2d\n", binnum - 1);
		else fprintf(fp, "\n");
		fprintf(fp, "Imposed resolution cutoff = %7.3f Angstroms\n",
			newdmin);
	  }
	else {
		fprintf(fp,
		 "Refls in bins %2d through %2d would have been excluded\n",
			binnum, NIOSSHELLS-1);
		fprintf(fp,
		 " but user-specified cutoff=%7.3f A supersedes that\n", rmaxg);
	     }
}

double stlval(short shkl[], REFPARS *rp)
{ /*	This computes sin(theta)/lambda for a reflection.
	In the new way of calculating rpp->xrec, it already has
	 the wavelength built in, so we don't need to divide by lambda here */
	register int	i;
	double		rsq, relt;

	for (i = 3, rsq = 0; i--; )
	   {
		relt = rp->xrec[i][0] * shkl[0] +
			rp->xrec[i][1] * shkl[1] + rp->xrec[i][2] * shkl[2];
		rsq += relt * relt;
	   }
	return 0.5 * sqrt(rsq);
}

static void tell_i0b(FILE *fp, float *liee, double stcut)
{ /*	Describes the results of an I000/B analysis */
	fprintf(fp,
	 "I000/B analysis of %8.0f reflections with <stl^2> = %8.4f:\n",
		liee[0], liee[1]);
	fprintf(fp,
	 "(I/s)000 = %8.3f; Bsig = %8.4f, ", liee[3], liee[2]);
	if (stcut > DFUZZ) fprintf(fp, "D(%6.3lf) = %8.3f\n",
		stcut, liee[4]);
	else fprintf(fp, "cutoff for <I/s>=%8.3lf undeterminable\n", stcut);
}

static int iosresults(float *liee, double stcut)
{ /*	This finishes a least-squares calculation for an I000/B analysis */
	double	determ, inter, slope;

	if (liee[0] < 4) return EBADVALUE;
	determ = liee[0] * liee[2] - liee[1] * liee[1];
	if (fabs(determ > DFUZZ))
	  {
		inter = (liee[2] *liee[3] - liee[1] *liee[4]) / determ;
		slope = (liee[0] *liee[4] - liee[1] *liee[3]) / determ;
		liee[2] = -slope;	liee[3] = exp(inter);
		if ((inter > DFUZZ) && (slope < -DFUZZ) && (stcut > DFUZZ))
		  {
			slope = (log(stcut) - inter) / slope;
			liee[4] = 0.5 / sqrt(fabs(slope));
		  }
		else liee[4] = 0.;
	  }
	liee[1] /= liee[0];	return EOK;
}

static void anal_ios(FILE *fpi, FILE *fplo, double stcut,
 float *rmaxp, REFPARS *rp)
{ /*	This analyzes the I/sigma values in a batch of OBSREF reflections
	to determine a sin(theta) cutoff */
	int		i, nobok, stlexclu;
	double		allios, deld, iol, ior, iosr, rmaxd, rma3, rmi3;
	double		rmistl, rmistl3, stl3, stl3ma, stlmax, stlmin, tstl;
	float		lie[5];
	OBSREF		lqrf;
	IOSIG_SHELL	*iosb, *iosp, *iospp;

	/* make a quick pass through to find the Dmin, Dmax */
	for (stlmax = nobok = 0, stlmin = 9999.;
	 0 < xgread((void *)&lqrf, XG_OBSREF, fpi); )
	   {
		nobok++;	tstl = stlval(lqrf.obs_hkl, rp);
		if (tstl > stlmax) stlmax = tstl;
		if (tstl < stlmin) stlmin = tstl;
	   }
	(void)fseek(fpi, (size_t)0, 0);	/* rewind the UREFLS file */
	stl3ma = stlmax * stlmax * stlmax; /* (sin(theta)/lambda) cubed */
	/* reserve memory for the various shells */
	if (NULL == (iosb = (IOSIG_SHELL *)callop(NIOSSHELLS + 1,
	 sizeof (IOSIG_SHELL), "iosb in anal_ios"))) return;
	for (i = 0, iosp = iosb; i < NIOSSHELLS; i++, iosp++)
	   {
		iosp->resh_binno = i;
		rmistl = mycbrt(stl3ma * ((double)(i+1)) /
			((double)NIOSSHELLS));
		iosp->resh_dhigh = 0.5 / rmistl;
		iosp->resh_dlow = (i == 0) ? 1. / (2. * stlmin) :
			(iosp-1)->resh_dhigh;
	   }
	for (i = 0; i < 5; i++) lie[i] = 0.; /* counters for I0/B anal.*/
	/* go back through the data and find <I/sigma> in the shells */
	for (allios = nobok = 0; 0 < xgread((void *)&lqrf, XG_OBSREF, fpi); )
	   {
		tstl = stlval(lqrf.obs_hkl, rp); stl3 = tstl * tstl * tstl;
		i = stl3 * NIOSSHELLS / stl3ma;	 if (i < 0) i = 0;
		else if (i >= NIOSSHELLS) i = NIOSSHELLS - 1;
		iosr = lqrf.obs_sigma;	if (iosr < 0.01) iosr = 0.01;
		iosr = lqrf.obs_ii / sqrt(iosr);	iosp = iosb + i;
		(iosp->resh_nobs)++;	iosp->resh_meanios += iosr;
		allios += iosr;	nobok++;
		if (iosr > DFUZZ)
		  {
			tstl *= tstl;		stl3 = log(iosr);
			lie[0] += 1.;		lie[1] += tstl;
			lie[2] += tstl * tstl;	lie[3] += stl3;
			lie[4] += tstl * stl3;
		  }
	   }
	(void)fseek(fpi, (size_t)0, 0); /* rewind yet again */
	if (EOK == iosresults(lie, stcut))
	  {
		tell_i0b(fplo, lie, stcut); tell_i0b(stdout, lie, stcut);
	  }
	/* the bin after the last shell is used for the overall dataset */
	iosp = iosb + NIOSSHELLS;	iosp->resh_dlow = iosb->resh_dlow;
	iosp->resh_dhigh = (iosp-1)->resh_dhigh; iosp->resh_nobs = nobok;
	iosp->resh_meanios = (nobok > 0) ? allios / (double)nobok : 0.;
	/* in bins, divide the sum of I/sigma values by the number of obs,
	 and decide whether we should impose a resolution cutoff */
	for (i = 0, iosp = iosb, stlexclu = NIOSSHELLS;
	 i < NIOSSHELLS; i++, iosp++)
	   {
		if (iosp->resh_nobs > 0)
		  {
			iosp->resh_meanios /= (double)(iosp->resh_nobs);
			if ((iosp->resh_meanios < stcut) &&
				(stlexclu >= NIOSSHELLS)) stlexclu = i;
		  }
	   }
	if (stlexclu < NIOSSHELLS)
	  {
		iosp = iosb + stlexclu;	rmistl = iosp->resh_dlow;
	  }
	/* interpolate if the previous bin had <I/sig> above the cutoff */
	iospp = iosp - 1;
	if ((iospp >= iosb) && (iospp->resh_meanios >= stcut))
	  {
		rmi3 = iosp->resh_dlow;		rmi3 *= rmi3 * rmi3;
		rma3 = iosp->resh_dhigh;	rma3 *= rma3 * rma3;
		iol = iospp->resh_meanios;	ior = iosp->resh_meanios;
		if (iol - ior > DFUZZ)
		  {
			rmistl3 = (stcut * (rmi3 - rma3) +
			 iol * rma3 - ior * rmi3) / (iol - ior);
			rmistl = mycbrt(rmistl3);
		  }
		stlexclu++;
	  }
	deld = rmistl - iosp->resh_dhigh;	rmaxd = *rmaxp;
	report_iosresults(fplo  , stlexclu, rmistl, rmaxd, deld, iosb);
	report_iosresults(stdout, stlexclu, rmistl, rmaxd, deld, iosb);
	memzap((AVOID **)&iosb, "iosb in anal_ios");
	if ((stlexclu < NIOSSHELLS) && (rmistl > *rmaxp)) *rmaxp = rmistl;
}

/* options implemented for "reduce" in "red_setup":
	-a<l,u>	Angular range should be restricted.  Output only data
		collected between frame l and frame u.  Thus to output
		only the data between frame 8 and frame 150 we say
		reduce -a8,150
	-b	Look at I(fit)/sigma(fit) and I(summed)/sigma(summed).
		If the former is larger than the latter, use the fitted
		intensity and sigma; if not, use the summed intensity and sigma.
	-c<val>	write a centroids file rather than a multiref.
		If <val> is specified, we write at most <val> reflections in
		each of 100 geometrically determined bins; default = 10.
		Only reflections with I/sigma > 3 will be output unless this
		value of 3 is overridden by a -f flag.
	-d<val> delete from output all reflections for which
		|I(fit) - I(summed)| > Q * sqrt(sigma(fit)^2 + sigma(summed)^2)
		where "^2" denotes "squared."  The value of Q will be <val>
		if specified; if a naked "d" is given, Q will be 10.
	-e	Use the (alpha/beta/gamma)-calculated intensities and sigmas
		(i.e. those that are obtained from the Kabsch-style geometry)
		rather than those obtained from conventional geometry.
	-f<val>	Output only the reflections with I/sigma > <val>.
		"-f" with no number is equivalent to "-f-2".
	-g	Like "-d<val>" except it rescales the comparison between
		the summed and fitted data before examining |I(fit)-I(summed)|.
		In other words, it compares I(fit) to I(summed) * <Ifit/Isum>.
	-i	reindex the reflections in the file according to the equation
			h' = Mh + v
		where M is a real 3x3 matrix and v is an integer 3x1 vector.
		The matrix and the vector are specified in a file "reindex"
		in the format
			M11  M12  M13  V1
			M21  M22  M23  V2
			M31  M32  M33  V3
		It's the user's responsibility to enter the M matrix and
		the V vector into the "reindex" file in a physically sensible
		way, although the program WILL recompute the unit cell
		parameters based on the M matrix for the user.
	-j	Read a file with environment variable name ICERING to be
		a list of resolution ranges to be excluded because the
		reflections in them overlap an ice ring. Each line in the file
		will specify a resolution range to exclude in the form
		two plane-spacing values (e.g. 3.85 3.81).
	-k	Recompute the indices of all the reflections in the file
		according to the crystal and detector parameters contained in
		the UPARAMS file.  If any of a reflection's indices differs
		from an integer by greater than the index-error limit in the
		UPARAMS file (typically 0.1-0.25) the reflection is not output.
	-l<val>	Inflate the Lorentz correction values by <val> so that
		I = (Iraw / p) * (v + 1/L) instead of I = (Iraw / p) * (1 / L).
		Thus <val> = 0 is equivalent to no alteration.
	-m<val>	Analyze <I/sigma> in shells. Output data only out to
		 the shell in which <I/sigma> falls to <val>. Thus
		 reduce ... -m1.5
		will output the data only out to a shell in which
		<I/sigma> is at least 1.5.
		As of June 2004 we actually interpolate to get a better cutoff
		resolution if the intensity falloff is reasonably monotonic.
		If <val> is <= 0, we turn the -m feature off altogether.
	-n<val>	Use the summation intensity if I/sigma > <val>;
		use the fitted intensity otherwise.
	-o	Read REFLECTIONS as a SCAN output file
		rather than reading UREFLS as an INTEGRATE output.
		"-s" and "-b" are rendered irrelevant by this option.
	-p<val>	Print <val> rejected reflections in each category
		rather than the default of 10.
	-q	Print all rejected reflections.
	-r	Recompute the indices of all the reflections in the file
		according to the crystal and detector parameters contained
		in the UPARAMS file.  Unlike the -k option, this option
		does not kill any reflections.
	-s	Use the summed intensity and sigma rather than the
		profile-fitted intensity and sigma.
	-u	Unmonochromatized radiation: recompute the polarization factor
		assuming the correction appropriate to radiation
		from a nickel-filtered or Francks-mirror source rather
		than from a monochromator.
		Use this flag ONLY if your radiation is filtered or mirrored
		AND the original run of "integrate" was run with the
		radiation assumed to be monochromatized.
	-w<val>	Set the nominal width of the scaling groups generated by
		`reduce' to be <val> degrees rather than the default 5 deg.
	-x	Save reflections only if |Xob-Xp| < dx, |Yob-Yp| < dy, and
		|Zob-Zp| < dz, where (Xob,Yob,Zob) is the observed centroid,
		(Xp,Yp,Zp) is the predicted centroids, and (dx,dy,dz) are the
		error limits set in the UPARAMS file.
	-y	Don't eliminate duplicate reflections. Ordinarily, if `reduce'
		discovers that there are two or more observations of a single
		reflection with almost the same (X,Y,omega), it eliminates
		the second and subsequent instances thereof. This flag tells
		the program not to perform this filtering operation.
	-z	Recompute the indices of all the reflections in the file
		according to the crystal and detector parameters contained
		in the UPARAMS file.  If any of the reflection's indices
		differs from an integer by greater than the index-error limit
		in the UPARAMS file (typically 0.1-0.25) OR IF THE NEW INDICES
		ARE DIFFERENT FROM THE ORIGINAL INDICES, do not output the
		reflection.  This is like the -k option but more stringent.
*************************************************************************/

static void sum_args(FILE *fp, REDUWORK *rwp)
{ /*	This summarizes the defaults and runtime args to stream fp */
	int		lflags;
	static char	*flist[] = {"UREFLS", "REFLECTIONS", "UPARAMS",
	 "USPLINE", "AMASK", "REINDEX", "CENTROIDS", "MULTIREF", "XLOG", NULL};
	static char	ftypes[] = {1, 1, 1, 1, 1, 2, 2, 2, 1, 0 };
	static char	polg[] = "graphite";
	static char	poln[] = "unpolarized";
	static char	polsv[] = "vertical-omega synchrotron";
	static char	polsh[] = "horizontal-omega synchrotron";

	lflags = (rwp->ru_oob)->os_rflags;
	ftypes[0] = (rwp->ru_inttype & SCAN_FORMAT) ? 0 : 1;
	ftypes[1] = 1 - ftypes[0];
	ftypes[5] = (lflags & RD_REINDEX) ? 1 : 0;
	ftypes[6] = (lflags & RD_MAKE_CENTROIDS) ? 1 : 0;
	ftypes[7] = 1 - ftypes[6];
	ftypes[8] = (rwp->ru_fpl == stdout ? 0 : 2);
	header(fp, REDUCTION, flist, ftypes);
	fprintf(fp,"This run will be set up with code number %3d\n",
		(int)((rwp->ru_rpp)->xrunno));
	fprintf(fp,"The minimum value of sigma/I will be %9.4f\n",
		rwp->ru_minsigoi);
	fprintf(fp,"The nominal width of scaling groups= %9.4f\n",
		rwp->ru_nomwidth);
	if (lflags & RD_RECOMPUTE)
	  {
		fprintf(fp,
 "Reflection indices will be recomputed\n\tbased on contents of %s\n",
			trname("uparams"));
		if (lflags & RD_KILL_NOT_INT) fprintf(fp,
 "Reflections will be output only if deviation from integerness < %6.3f\n",
			(rwp->ru_rpp)->xerrind);
		if (lflags & RD_KILL_IF_REINDEXED) fprintf(fp,
 "\tand if the recomputed indices are equal to the input indices\n");
	  }
	if (lflags & RD_REINDEX)
	  {
		fprintf(fp,
"Reflections will be reindexed according to h' = Mh + v\n");
		fprintf(fp,"\t\t  %6.2f %6.2f %6.2f\t\t%3hd\n",
			rwp->ru_matrei[0][0], rwp->ru_matrei[0][1],
			rwp->ru_matrei[0][2], rwp->ru_vecrei[0]);
		fprintf(fp,"\twhere M = %6.2f %6.2f %6.2f\t\t%3hd\n",
			rwp->ru_matrei[1][0], rwp->ru_matrei[1][1],
			rwp->ru_matrei[1][2], rwp->ru_vecrei[1]);
		fprintf(fp,"\t\t  %6.2f %6.2f %6.2f\t\t%3hd\n",
			rwp->ru_matrei[2][0], rwp->ru_matrei[2][1],
			rwp->ru_matrei[2][2], rwp->ru_vecrei[2]);
	  }
	if (lflags & RD_KILL_DELXYZ) fprintf(fp,
 "Reflections will be output only if |dx|<%5.2f, |dy|<%5.2f, |dz|<%5.2f\n",
	 (rwp->ru_rpp)->xerrx, (rwp->ru_rpp)->xerry, (rwp->ru_rpp)->xerrphi);
	if (rwp->ru_rmaxgiven < 0)
	  {
		if (rwp->ru_rmingiven < 0) fprintf(fp,
 "No observations will be excluded due to sin(theta)/lambda restrictions\n");
		else fprintf(fp,
 "Observations with plane spacings > %6.2f Angstroms will be excluded\n",
			rwp->ru_rmingiven);
	  }
	else
	  {
		if (rwp->ru_rmingiven < 0) fprintf(fp,
 "Observations with plane spacings < %6.2f Angstroms will be excluded\n",
			rwp->ru_rmaxgiven);
		else fprintf(fp,
 "Only observations for which %6.2f A < D < %6.2f A will be written out\n",
			rwp->ru_rmingiven, rwp->ru_rmaxgiven);
	  }
	if ((!(lflags & RD_POLARIZED)) || (lflags & RD_SYNCHRO))
	  {
		fprintf(fp,
 "Polarizations will be corrected from %s formula to %s formula\n",
		 (lflags & RD_POLARIZED) ? polg : poln,
		 (lflags & RD_SYNCHRO_V) ? polsv :
		 (lflags & RD_SYNCHRO_H) ? polsh : poln);
	  }
	if (lflags & RD_IOSCUT) fprintf(fp,
 "Reflections with intensity / sigma less than %7.3f will not be output\n",
		rwp->ru_ioscut);
	if (lflags & RD_DOSORT) fprintf(fp,
	 "Reflections will be sorted, merged, and scaled before output\n");
	else fprintf(fp,
	 "Reflections will not be sorted, merged, or scaled before output\n");
	switch (rwp->ru_inttype)
	  {
		case	SIMPLE_SUMMATION:	fprintf(fp,
 "Simply-summed intensities, not shape-fit intensities, will be output\n");
			break;
		case	SUMMED_IF_STRONG:	fprintf(fp,
 "Simply-summed I's will be used for I/s > %7.1f; otherwise, profile-fitted\n",
			rwp->ru_strongcut);
		case	BEST_I_OVER_SIGMA:	fprintf(fp,
 "Better I/sigma ratio picked between summed and fitted intensities\n"); break;
                case	PROFILE_FITTED:	fprintf(fp,
 "Shape-fit intensities, not simply-summed intensities, will be output\n");
			break;
		case	SCAN_FORMAT:	fprintf(fp,
 "Input REFLECTIONS file is taken from a run of SCAN\n");
		case	ABG_FIT:	fprintf(fp,
 "Kabsch-style shape-fit intensities will be output\n");	break;
		case	ABG_SUM:	fprintf(fp,
 "Kabsch-style simple-summation intensities will be output\n");	break;
		case	SUMSTRONG_ABG:	fprintf(fp,
 "Kabsch-style data used: Summed for I/s > %7.1f; otherwise, fitted\n",
		 rwp->ru_strongcut);
	   }
	if (lflags & RD_MAKE_CENTROIDS) fprintf(fp,
	 "Centroids file will contain at most %3d refls in each of %2d bins\n",
		(rwp->ru_oob)->os_ncperbin, (rwp->ru_oob)->os_ncbins);
	if (lflags & RD_SCALE_SUMFIRST)
	  {
		fprintf(fp,
		 "Reflections with |I(fit) - <If/Is>*I(summed)| >\n\t");
		fprintf(fp,
 "sqrt(sigsq(fit)+<If/Is>^2*sigsq(summed)) * %9.4f will be deleted\n",
		sqrt(rwp->ru_delilim));
	  }
	if (lflags & RD_DEL_BIG_IDIFF)
	  {
		fprintf(fp, "Reflections with |I(fit) - I(summed)| >\n\t\t");
		fprintf(fp,
		 "sqrt(sigsq(fit)+sigsq(summed)) * %9.4f will be deleted\n",
			sqrt(rwp->ru_delilim));
	  }
	if (lflags & RD_ANGLE_RESTRICTION) fprintf(fp,
	  "Reflections will be output from frames %3d to %3d\n",
			(rwp->ru_oob)->os_minfm, (rwp->ru_oob)->os_maxfm);
	fprintf(fp,
 "Duplicate observations of hkl at almost the same (X,Y,omega) will be %s\n",
		(lflags & RD_KEEPDUPS) ? "kept" : "eliminated");
	if (lflags & RD_KILL_ICERING) describe_rings(fp);
	if (lflags & RD_PRINTALL) fprintf(fp,
	 "All rejected reflections will be printed to the logfile\n");
	else fprintf(fp,
 "The first %d rejected refls in each category are printed to the logfile\n",
		rwp->ru_representative);
	if (lflags & RD_STLCUT) fprintf(fp,
 "Data output in shells out to that for which <I/sigma> > %8.2f\n",
	 rwp->ru_stlioscut);
	if (rwp->ru_lordel > FUZZ) fprintf(fp,
	 "All Lorentz-factor (1/L) values will be inflated by %8.5f\n",
		rwp->ru_lordel);
	if (rwp->ru_trigc > FUZZ) fprintf(fp,
	 "Intensities will be adjusted by 1 - %6.4f * cos(twopi*omega)\n",
		rwp->ru_trigc);
	show_params(fp, rwp->ru_rpp, 256);
}

static int reduce_defaults(REDUWORK *rwp)
{ /*	This sets the defaults for `reduce' in a manner that is sensitive
	to the detector type and the characteristics of the data run. */
	int			detype, iwidth, nrefls, result;
	int			detdim[2];
	char			*cpi, *firstdigit, *lastdigit, *urfname;
	double			ruwidth;
	OUT_OBSSTYLE		*loo;
	PAU			rdu;
	static struct stat	ursbuf;
	static FRAMEINDEX	*lfmi;

	zeromem((AVOID *)rwp, sizeof (REDUWORK));
	iwidth = sizeof (CRYSTAL) + sizeof (DATARUN) + sizeof (OUT_OBSSTYLE);
	if (NULL == (rdu.vpt = callop(1, (size_t)iwidth,
	 "general-purpose memory within rwp in reduce"))) return EALLOC;
	rwp->ru_xhp = (CRYSTAL *)rdu.vpt;	rdu.ucpt += sizeof (CRYSTAL);
	rwp->ru_rhp = (DATARUN *)rdu.vpt;	rdu.ucpt += sizeof (DATARUN);
	rwp->ru_oob = loo = (OUT_OBSSTYLE *)rdu.vpt;
	rwp->ru_minsigoi = DEF_MINSIGOI;
	rwp->ru_rmaxgiven = rwp->ru_rmingiven = -1.;
	loo->os_maxfm = 9999; loo->os_minfm = -loo->os_maxfm;
	loo->os_ncperbin = DEF_NCENPERBIN;
	rwp->ru_maxpr = -1.e8;	rwp->ru_minpr = 1.e8;	loo->os_outobs = 0;
	zeromem(&(rwp->ru_nbad[0]), BADTYPES * sizeof (int));
	rwp->ru_ioscut = -32766.;
	rwp->ru_machineerr = rwp->ru_minslegal = rwp->ru_lordel =
		rwp->ru_trigc = 0.;
	rwp->ru_delilim = rwp->ru_maxslegal = rwp->ru_delfrm = 32766.;
	rwp->ru_omegainterval = 5.;	loo->os_xdel = loo->os_ydel = 15.5;
	rwp->ru_fmpergroup = 25.; rwp->ru_sumscale = rwp->ru_sumscsq = 1.;
	rwp->ru_stlioscut = -9999.;
	zeromem((void *)&(rwp->ru_sloc[0]), 3 * sizeof (float));
	zeromem((void *)&(rwp->ru_matrei[0][0]), 9 * sizeof (float));
	rwp->ru_representative = DEF_REPRESENT;
	if (0 >= (detype = dettype(detdim))) return EBADVALUE;
	rwp->ru_inttype = ABG_FIT;
	/* some flags on, some off */
	loo->os_rflags = RD_DOSORT | RD_POLARIZED |
		RD_DEL_BIG_IDIFF | RD_STLCUT;
	rwp->ru_delilim = 16.;	rwp->ru_ioscut = -3;
	rwp->ru_stlioscut = 1.4;	
	/* grab parameters here: we need them to set defaults*/
	if (EOK != (result = get_upars(&(rwp->ru_rpp), loo))) return result;
	if (NULL == (lfmi = ftables(3))) return EALLOC;
	ruwidth = (fabs((rwp->ru_rpp)->xfrmsize) / RADC) *
		(double)(lfmi->fi_totfms);
	/* if we have lots of reflections, we make the scaling groups
	 1 degree wide each, unless that would produce too many scaling groups;
	 if we only have a few reflections, we make 'em 4 degrees wide.
	 Dec 05: this is the first time we look for UREFLS, so we allow for
	 gunzipping before we run `stat' */
	if (0 != xgaccess((urfname = trname("urefls")), 04, 0))
	  {
		fprintf(stderr, "Cannot open %s as input UREFLS file\n",
			urfname);	return EOPENERR;
	  }
	if (0 != stat(urfname, &ursbuf)) return EOPENERR;
	nrefls = 1 + ursbuf.st_size / sizeof (OBSREF);
	if (nrefls < 5000) rwp->ru_nomwidth = 4;
	else {
		if (ruwidth < 100.) rwp->ru_nomwidth = 1.;
		else if (ruwidth < 150.) rwp->ru_nomwidth = 1.5;
		else {
			iwidth = (int)(ruwidth * 0.01 + 0.99999);
			rwp->ru_nomwidth = iwidth;
		     }
	     }
	/* intuit the run number from the names of the files */
	for (cpi = urfname; '\0' != *cpi; cpi++) ; /* find end of urefls name*/
	/* identify the last digit in the name and the first digit of
	 the last group of digits in the name */
	for (lastdigit = firstdigit = NULL, cpi--; cpi >= urfname; cpi--)
	   {
		if (('0' <= *cpi) && (*cpi <= '9'))
		  {
			if (NULL == lastdigit) lastdigit = cpi;
		  }
		else if (NULL != lastdigit)
		  {
			firstdigit = cpi + 1;	break;
		  }
	   }
	/* if we didn't find any digits, set the run number to 1 */
	if (NULL == lastdigit) (rwp->ru_rpp)->xrunno = 1;
	else { /* otherwise, construct a run number from those digits */
		if (NULL == firstdigit) firstdigit = lastdigit;
		if (firstdigit < lastdigit - 1) firstdigit = lastdigit - 1;
		for ((rwp->ru_rpp)->xrunno = 0, cpi = firstdigit;
		 cpi <= lastdigit; cpi++) (rwp->ru_rpp)->xrunno =
			10 * (rwp->ru_rpp)->xrunno + (*cpi - '0');
	     }
	return EOK;
}

static void partohead(REFPARS *rp, char procdate[], CRYSTAL *xtp, DATARUN *runp)
{ /*	This converts the "REFPARS" structure `rp' to the corresponding
	"CRYSTAL" and "DATARUN" structures xtp and runp. `procdate' is
	a string containing the date. */
	register int		i, j;
	int			detdim[2];
	static char		*nameonly, *thisfname, *tname;

	tname = trname("urefls");
	nameonly = filepart(tname, 2);
	/* blank out the CRYSTAL and DATARUN structures */
	zeromem((void *)xtp, sizeof (CRYSTAL));
	zeromem((void *)runp, sizeof (DATARUN));
	strncpy(runp->run_code, nameonly, 7);
	strncpy(xtp->xt_code, runp->run_code, strlen(runp->run_code) - 1);
	thisfname = filepart(tname, 0);
	sprintf(runp->run_notes,"Data from %s processed %s",
		thisfname, procdate);
	runp->run_id = rp->xrunno;
	runp->run_stepreal = -rp->xfrmsize;
	runp->run_stepsize = (int)(rp->xfrmsize * 96. / RADC + 0.49999);
	runp->run_motor = 2;	runp->run_nchams = 1;
	runp->run_temp = 298;	runp->run_lambda= rp->xlambda;
	for (i = 3; i--; ) runp->run_startpos[i] = rp->xstart[i];
	runp->run_width = rp->xdetwid;	runp->run_height = rp->xdetht;
	/* the xrec matrix is already unitless here, so we don't need
	 to divide by lambda */
	for (i = 3; i--; ) for (j = 3; j--; )
		runp->run_recip[i][j] = rp->xrec[i][j];
	runp->run_fmmin = 1;	/* we will change these later */
	runp->run_fmmax = 360;
	runp->run_volxray = 40;
	runp->run_curxray = 70;
	runp->run_expos = 60.;
	runp->run_reslim = rp->xresmax;
	runp->run_ssqmax = rp->xlambda * rp->xlambda /
		(rp->xresmax * rp->xresmax);
	xtp->xt_intl = rp->xispgp;
	xtp->xt_isys = rp->xisys;
	strcpy(xtp->xt_device,"GX21");
	runp->run_d0pos[0] = rp->xplated;
	runp->run_d0pos[1] = rp->xx0;
	runp->run_d0pos[2] = rp->xy0;
	runp->run_d0ang[0] = rp->xswing / RADC; /* degrees, not radians */
	runp->run_d0ang[1] = rp->xtilt / RADC;
	xtp->xt_stdpos[0] = -rp->xdelom;
	xtp->xt_stdpos[1] = -rp->xdelchi;
	xtp->xt_stdpos[2] = -rp->xdelphi;
	xtp->xt_lambda = rp->xlambda;
	xtp->xt_reslim = rp->xresmax;
	xtp->xt_ssqmax = runp->run_ssqmax;
	xtp->xt_nruns = 1;
	xtp->xt_code[15] = xtp->xt_desc[47] = xtp->xt_notes[63] =
		runp->run_code[7] = runp->run_notes[63] =
		xtp->xt_spaceg[7] = xtp->xt_device[7] = '\0';
	xtp->xt_dettype = (rp->xdtype > 0) ? rp->xdtype :
		dettype(detdim);
	xtp->xt_version = (rp->xversion > 0) ? rp->xversion : 500;
	for (i = 3; i--; )
	   {
		xtp->xt_ar[i] = rp->xabc[i]; xtp->xt_ang[i] = rp->xang[i];
	   }
}

int red_setup(int argc, char *argv[], REDUWORK *rwp)
{ /*	This processes the runtime arguments and sets up the "reduce" run */
	int		input_refsize, ival, result, xystuf;
	char		**dcp, **flags, rundate[26];
	double		fval, ngiven;
	OUT_OBSSTYLE	*loo;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	if (EOK != (result = reduce_defaults(rwp))) return result;
		/* open logfile*/
	if (NULL == (rwp->ru_fpl = propen("xlog"))) rwp->ru_fpl = stdout;
	rwp->ru_tl0 = NOW;	/* set timestamp */
	strcpy(rundate,ctime(&rwp->ru_tl0)); rundate[24] = '\0';
	flags = flagproc(argc, argv, "ABCDEFGHIJKLMNOPQRSTUVWXYZ");
	for (loo = rwp->ru_oob, dcp = flags; NULL != dcp && NULL != *dcp;dcp++)
	  {
	   if ('\0' != *((*dcp) + 1))
	     {
		fval = atof((*dcp) + 1);	ival = atoi((*dcp) + 1);
	     }
	   else	{
			fval = 0.;	ival = 0;
		}
	  switch (**dcp)
	   {
		case 'A':	loo->os_rflags |= RD_ANGLE_RESTRICTION;
				if (ival != -10000)
				  {
					(void)sscanf((*dcp) + 1, "%hd,%hd",
					 &(loo->os_minfm), &loo->os_maxfm);
					break;
				  }
				loo->os_minfm = 1; loo->os_maxfm = 999;	break;
		case 'B':	if (!(rwp->ru_inttype & SCAN_FORMAT))
				 rwp->ru_inttype = BEST_I_OVER_SIGMA;	break;
		case 'C':
			if (!(loo->os_rflags & RD_IOSCUT) || (rwp->ru_ioscut <
			 MIN_IOSCUT)) rwp->ru_ioscut = MIN_IOSCUT;
			loo->os_rflags |= (RD_MAKE_CENTROIDS | RD_IOSCUT);
			loo->os_rflags &= ~(RD_DOSORT);
			if (ival != -10000) loo->os_ncperbin = ival;
			else loo->os_ncperbin = DEF_NCENPERBIN;	break;
		case 'D':	loo->os_rflags |= RD_DEL_BIG_IDIFF;
			loo->os_rflags &= ~(RD_SCALE_SUMFIRST);
			rwp->ru_delilim = (ival < MIN_DELI_LIMIT) ?
				DEF_DELI_LIMIT : fval; /* use in squared form*/
			rwp->ru_delilim *= rwp->ru_delilim;	break;
		case 'E':	rwp->ru_inttype |= ABG_EITHER;	break;
		case 'F':	loo->os_rflags |= RD_IOSCUT;
			rwp->ru_ioscut = (ival == -10000) ? -2 : fval;	break;
		case 'G':	loo->os_rflags |= RD_SCALE_SUMFIRST;
			loo->os_rflags &= ~(RD_DEL_BIG_IDIFF);
			rwp->ru_delilim = (ival < MIN_DELI_LIMIT) ?
				DEF_DELI_LIMIT: fval; /* use in squared form */
			rwp->ru_delilim *= rwp->ru_delilim; break;
		case	'H':	loo->os_rflags |= RD_SYNCHRO_H;	break;
		case	'I':
			if ((!(loo->os_rflags & RD_REINDEX)) && try_reindex(
			 rwp->ru_vecrei, rwp->ru_matrei))
				loo->os_rflags |= RD_REINDEX; break;
		case	'J':
			loo->os_rflags |= RD_KILL_ICERING;	break;
		case 'K':	loo->os_rflags |= (RD_RECOMPUTE |
			RD_KILL_NOT_INT); break;
		case 'L':	if ((0. <= fval) && (fval <= 1.))
					rwp->ru_lordel = fval;	break;
		case 'M':	if ((0. <= fval) && (fval < 999.))
				  {
					rwp->ru_stlioscut = fval;
					loo->os_rflags |= RD_STLCUT;
				  }
				else {
					loo->os_rflags &= ~RD_STLCUT;
					rwp->ru_stlioscut = -9999.;
				     } break;
		/* old meaning of the N flag doesn't get used in practice:
		case 'N': loo->os_rflags &= ~(RD_DOSORT);	break; */
		case 'N':	rwp->ru_inttype |= SUMMED_IF_STRONG;
				rwp->ru_inttype &= ~03;
			rwp->ru_strongcut = ((0. <= fval) && (fval < 9999.)) ?
				fval : 10.;	break;
		case 'O': rwp->ru_inttype = SCAN_FORMAT;	break;
		case 'P': if (fval > 0.) rwp->ru_representative = fval; break;
		case 'Q':	/* various specific positive flags */
		 if (ival > 0)
		  {
		    switch (ival) /* 1-4 change integration type */
		     {
			case  1: rwp->ru_inttype |= BEST_I_OVER_SIGMA;	break;
		  	case  2: rwp->ru_inttype |= ABG_EITHER;	break;
			case  3: rwp->ru_inttype = SCAN_FORMAT;	break;
			case  4: if (rwp->ru_inttype != SCAN_FORMAT)
				   {
					rwp->ru_inttype &= ~PROFILE_FITTED;
					rwp->ru_inttype |= SIMPLE_SUMMATION;
				   }
				break; /* 5-9 change reduce run flags */
		  	case  5: loo->os_rflags |= RD_SYNCHRO_H;	break;
		  	case  6: loo->os_rflags |= RD_KILL_ICERING;	break;
		  	case  7: loo->os_rflags |=
				(RD_RECOMPUTE | RD_KILL_NOT_INT); break;
			case  8: loo->os_rflags |= RD_PRINTALL;	break;
			case  9: loo->os_rflags |= RD_RECOMPUTE;
				loo->os_rflags &= ~RD_KILL_NOT_INT;	break;
			case 10: loo->os_rflags &= ~RD_POLARIZED;	break;
			case 11: loo->os_rflags |= RD_SYNCHRO_V;	break;
			case 12: loo->os_rflags |= RD_KILL_DELXYZ;	break;
			case 13: loo->os_rflags |= RD_KEEPDUPS;	break;
			case 14: loo->os_rflags |= RD_ZOPTION;	break;
			default: break;
		     } /* close of switch */
		    } /* close the "if (ival > 0" clause */
		  else	 loo->os_rflags |= RD_PRINTALL;
		  break;	/* break in either case */
		case 'R': loo->os_rflags |= RD_RECOMPUTE;
			  loo->os_rflags &= ~(RD_KILL_NOT_INT);	break;
		case 'S': if (rwp->ru_inttype != SCAN_FORMAT)
			    {
				rwp->ru_inttype &= ~(PROFILE_FITTED);
				rwp->ru_inttype |= SIMPLE_SUMMATION;
			    }
			  break;
		case 'T':	if ((0. <= fval) && (fval <= 1.))
					rwp->ru_trigc = fval;	break;
		case 'U': loo->os_rflags &= ~RD_POLARIZED;	break;
		case 'V': loo->os_rflags |= RD_SYNCHRO_V;	break;
		case 'W': rwp->ru_nomwidth = (ival >= 0) ? fval : DEF_NOMWIDTH;
		case 'X': loo->os_rflags |= RD_KILL_DELXYZ;	break;
		case 'Y': loo->os_rflags |= RD_KEEPDUPS;	break;
		case 'Z':
		 if (ival > 0) /* numerical negates various flags */
		  {
		    switch (ival) /* 1-4 change integration type */
		     {
			case  1: rwp->ru_inttype &= ~BEST_I_OVER_SIGMA; break;
		  	case  2: rwp->ru_inttype &= ~ABG_EITHER;	break;
			case  3: rwp->ru_inttype &= ~SCAN_FORMAT;	break;
			case  4: if (rwp->ru_inttype != SCAN_FORMAT)
				   {
					rwp->ru_inttype &= ~SIMPLE_SUMMATION;
					rwp->ru_inttype |= PROFILE_FITTED;
				   }
				break; /* 5-9 change reduce run flags */
		  	case  5: loo->os_rflags &= ~RD_SYNCHRO_H;	break;
		  	case  6: loo->os_rflags &= ~RD_KILL_ICERING;	break;
		  	case  7: loo->os_rflags &= ~(RD_RECOMPUTE |
				RD_KILL_NOT_INT); break;
			case  8: loo->os_rflags &= ~RD_PRINTALL;	break;
			case  9: loo->os_rflags &= ~RD_RECOMPUTE;
				loo->os_rflags |= RD_KILL_NOT_INT;	break;
			case 10: loo->os_rflags |= RD_POLARIZED;	break;
			case 11: loo->os_rflags &= ~RD_SYNCHRO_V;	break;
			case 12: loo->os_rflags &= ~RD_KILL_DELXYZ;	break;
			case 13: loo->os_rflags &= ~RD_KEEPDUPS;	break;
			case 14: loo->os_rflags &= ~RD_ZOPTION;		break;
		     }	/* close switch */
		    } /* close the "if (ival > 0" clause */
		  else loo->os_rflags |= RD_ZOPTION;	break; /* old meaning */
	     } /* end of big switch */
	   }
	if ((loo->os_rflags & RD_KILL_ICERING) && (NULL == manageice(7)))
		loo->os_rflags &= ~RD_KILL_ICERING;
	input_refsize = (rwp->ru_inttype == SCAN_FORMAT) ?
		sizeof (SCANREF) : sizeof (OBSREF);
	/* skip past the program name and the switch arguments */
	argc--;	argv++;
	while ((argc >= 0) && (NULL != *argv) && ('-' == **argv))
	     {
		argc--;	argv++;
	     }
/* read the first non-switch runtime argument, if any, to get the run number */
	if (argc-- && 0 <= (ngiven = atoi(*argv++)) && MAX_RUN_INDEX > ngiven)
	  {
		(rwp->ru_rpp)->xrunno = ngiven;
		/* read 2nd non-switch arg, if any, as min value of sigma/I */
		if (argc-- && 0 <= (ngiven = atof(*argv++)))
		  {
			rwp->ru_minsigoi = ngiven;
		/* Use it  to compute a "machine error" component to sigma */
			rwp->ru_machineerr = rwp->ru_minsigoi *
			 rwp->ru_minsigoi;
		/* read the 3rd & 4th non-switch args, if any, as imposed
		resolution cutoffs; if fewer than 4 args are present,
		we won't check for resolution */
			if (argc--) rwp->ru_rmaxgiven = atof(*argv++);
			if (argc > 0) rwp->ru_rmingiven = atof(*argv);
		  }
	  }
	/* reverse arguments if necessary */
	if (rwp->ru_rmingiven >= -0.5 && rwp->ru_rmaxgiven > rwp->ru_rmingiven)
	  {
		ngiven = rwp->ru_rmingiven;
		rwp->ru_rmingiven = rwp->ru_rmaxgiven;
		rwp->ru_rmaxgiven = ngiven;
	  }
	/* get the parameters from the "uparams" file */
	if (loo->os_rflags & RD_MAKE_CENTROIDS)
	  {
		loo->os_cenpbin = (short *)callop((unsigned)loo->os_ncbins,
		 sizeof (short), "loo->os_cenpbin in reduce");
		if (qerror((AVOID *)(rwp->ru_fpo = xgopen("centroids", "w")),
		 "open output centroids file")) return EOPENERR;
	  }
	sum_args(stdout, rwp);
	if (rwp->ru_fpl != stdout) sum_args(rwp->ru_fpl, rwp);
		/* set up output header */
	partohead(rwp->ru_rpp, rundate, rwp->ru_xhp, rwp->ru_rhp);
	/* nonsense starting values for comparers */
	(rwp->ru_rhp)->run_fmmin = 100000.;
	(rwp->ru_rhp)->run_ssqmax = (rwp->ru_rhp)->run_fmmax =
		-(rwp->ru_rhp)->run_fmmin;
	xystuf = keep_dimen(rwp->ru_rhp); /* remember width/256,ht/512,ht/2 */
	rwp->ru_widrat = xystuf >> 24;	rwp->ru_htrat = (xystuf >> 16) & 0xFF;
	rwp->ru_hto2 = xystuf & 0xFFFF;
	/* open reflections file, count # of refls in it, and rewind*/
	if (rwp->ru_inttype == SCAN_FORMAT)
	  {
		if (qerror((AVOID *)(rwp->ru_fpi = xgopen("reflections", "r")),
			"open reflections file")) return EOPENERR;
	  }
	else if (qerror((AVOID *)(rwp->ru_fpi = xgopen("urefls", "r")),
		"open urefls file")) return EOPENERR;
	rwp->ru_nread = fseek(rwp->ru_fpi, 0L, 2);
	if (verror(1 + rwp->ru_nread, "seek to end of urefls file"))
		return ESEEKERR;
	rwp->ru_nread = ftell(rwp->ru_fpi) / input_refsize;
	if (NULL == (AVOID *)(rwp->ru_orefb = (FASTOB *)callop((unsigned)
	 (rwp->ru_nread + 1000), sizeof (FASTOB),
	 "rwp->ru_orefb in red_setup"))) return EALLOC;
	if (verror(1 + fseek(rwp->ru_fpi, 0L, 0), "rewind reflections file"))
		return ESEEKERR;
	/* if RD_STLCUT is set, analyze I/sigma as a function of sin(theta).
	 Adjust the upper resolution limit accordingly if necessary */
	if (loo->os_rflags & RD_STLCUT) anal_ios(rwp->ru_fpi, rwp->ru_fpl,
	 (double)(rwp->ru_stlioscut), &rwp->ru_rmaxgiven, rwp->ru_rpp);
	xeflush(rwp->ru_fpl);
	/* we impose restrictions on sin(theta)/lambda only if requested */
	rwp->ru_maxslegal = (rwp->ru_rmaxgiven > 0.) ?
	 0.5 /rwp->ru_rmaxgiven : 1. / (rwp->ru_rpp)->xlambda;
	rwp->ru_minslegal = (rwp->ru_rmingiven > 0.) ?
	 0.5 / rwp->ru_rmingiven : 0.;
	rwp->ru_excluc = exclucode((rwp->ru_rpp)->xispgp,(rwp->ru_rpp)->xisys);
	/* this is used in I000, B, D2 calculations */
	rwp->ru_lamco = 0.005 / (rwp->ru_rpp)->xlambda;
	rwp->ru_lamco *= rwp->ru_lamco;
	zeromem(rwp->ru_iose, 5 * sizeof (float));	return EOK;
}

static void report_totals(FILE *fp, REDUWORK *rwp)
{ /* Report some summed values to the stream fp */
	OUT_OBSSTYLE	*loo;

	loo = rwp->ru_oob;
	fprintf(fp, " %6d observations read in; %6d observations converted\n",
		rwp->ru_nread, rwp->ru_nconverted);
	fprintf(fp,
	 " %6d unique reflections,   %6d unmerged-Bijvoet reflections found\n",
 	 (int)((rwp->ru_xhp)->xt_nmerged), (int)((rwp->ru_xhp)->xt_nrefls));
	fprintf(fp,
	 " %6d unique reflections,   %6d observations actually written out\n",
		rwp->ru_outrefls, (int)(loo->os_outobs));
	fprintf(fp,
 " %6d scaling groups numbered %3d -> %3d of width %7.2f deg are present\n",
	 (int)(rwp->ru_xhp)->xt_nshifts, 2 * rwp->ru_shoffset,
	 2 * rwp->ru_maxeven + 1, rwp->ru_omegainterval);
	if (rwp->ru_nchanged)
	  {
		fprintf(fp,
	 " %6d observations had their hkl indices changed from input\n",
		rwp->ru_nchanged);
		if (loo->os_rflags & RD_KILL_IF_REINDEXED) fprintf(fp,
		 "\t\t... and therefore were excluded from output\n");
	  }
	if (rwp->ru_nbad[LAUE_BAD]) fprintf(fp,
	 " %6d observations were excluded by Laue-group symmetries\n",
		rwp->ru_nbad[LAUE_BAD]);
	if (rwp->ru_nbad[SCREW_BAD]) fprintf(fp,
	 " %6d observations were excluded by screw-axis symmetries\n",
		rwp->ru_nbad[SCREW_BAD]);
	if (rwp->ru_nbad[STL_TOO_LOW]) fprintf(fp,
	 " %6d observations had sin(theta) values that were too low\n",
		rwp->ru_nbad[STL_TOO_LOW]);
	if (rwp->ru_nbad[STL_TOO_HIGH]) fprintf(fp,
	 " %6d observations had sin(theta) values that were too high\n",
		rwp->ru_nbad[STL_TOO_HIGH]);
	if (rwp->ru_nbad[IOSIGMA_BAD]) fprintf(fp,
	 " %6d observations were excluded because I/sigma < %8.2f\n",
		rwp->ru_nbad[IOSIGMA_BAD], rwp->ru_ioscut);
	if (rwp->ru_nbad[NOT_INTEGER]) fprintf(fp,
	 " %6d observations were excluded because err(h) > %6.4f\n",
		rwp->ru_nbad[NOT_INTEGER], (rwp->ru_rpp)->xerrind);
	if (rwp->ru_nbad[BAD_X_Y]) fprintf(fp,
	 " %6d observations were at (X,Y) positions outside the active area\n",
		rwp->ru_nbad[BAD_X_Y]);
	if (rwp->ru_nbad[FIT_NE_SUM]) fprintf(fp,
	 " %6d observations had |I(fit)-I(sum)| > sigma cutoff\n",
		rwp->ru_nbad[FIT_NE_SUM]);
	if (rwp->ru_nbad[NONSENSE_HKL]) fprintf(fp,
	 " %6d observations had nonsensical hkl values\n",
		rwp->ru_nbad[NONSENSE_HKL]);
	if (rwp->ru_nbad[DEL_XYZ]) fprintf(fp,
	 " %6d observations had |dx|>%5.2f, |dy|>%5.2f, or |dz|>%5.2f\n",
		rwp->ru_nbad[DEL_XYZ], (rwp->ru_rpp)->xerrx,
		(rwp->ru_rpp)->xerry, (rwp->ru_rpp)->xerrphi);
	if (rwp->ru_nduptoss) fprintf(fp,
	 " %6d observations were eliminated as internal duplicates\n",
		rwp->ru_nduptoss);
}

int out_refls(REDUWORK *rwp)
{ /*	This writes the reflections to a file */
	register FASTOB		*orefp;
	int			j;
	ABBROB			*abp;
	OUT_OBSSTYLE		*loo;
	static MERGEDREF	lmr;

	for (loo = rwp->ru_oob, rwp->ru_nduptoss = 0, orefp = rwp->ru_orefb;
	 orefp < rwp->ru_orefend; orefp++)
	   {
		if (orefp->fo_nob)
		  {
			cpymem((void *)&lmr, (void *)orefp, sizeof (FASTOB));
			if ((j = orefp->fo_nob - 1))
			  {
				for (abp = &(lmr.mr_ob[1]), orefp += 1;
				 j--; abp++, orefp++) cpymem((void *)abp,
				 (void *)&(orefp->fo_ob), sizeof (ABBROB));
				orefp--;	/* back up one */
			  }
			/* April 2002: eliminate duplicate observations
			 before output unless user requests that we keep 'em */
			if (!(loo->os_rflags & RD_KEEPDUPS))
				rwp->ru_nduptoss += eliminate_dups(&lmr);
			if (verror(0 < xgwrite((void *)&lmr, XG_MERGEDREF,
			 rwp->ru_fpo), "write a reflection to multiref"))
				return EWRITERR;
			(rwp->ru_outrefls)++;	loo->os_outobs += lmr.mr_nob;
		  }
	   }
	rwp->ru_tl1 = NOW;	fprintf(rwp->ru_fpl,
	 " %4d seconds required to write output to %s\n",
		(int)(rwp->ru_tl1 - rwp->ru_tl0), trname("multiref"));
	(void)fclose(rwp->ru_fpo);	rwp->ru_fpo = NULL;
	if (EOK == iosresults(rwp->ru_iose, rwp->ru_stlioscut))
	  {
		tell_i0b(rwp->ru_fpl, rwp->ru_iose, rwp->ru_stlioscut);
		tell_i0b(stdout, rwp->ru_iose, rwp->ru_stlioscut);
	  }
	report_totals(stdout, rwp);	report_totals(rwp->ru_fpl, rwp);
	return EOK;
}

/* end of redutils.c */
