/* stratcalc.c -- construct a strategy for data collection

	Copyright (c) 2004, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	20 Mar 2007: update scheme used for `xgvariable' output
	 4 Mar 2007: emit some output information via "xgvariable" mechanism
	28 Nov 2006: makeq21 call updated; STRATWORK elt. sq_loq21 eliminated
	12 Sep 2006: further speedups: redundancy calcs done separately
	26 Aug 2006: changes to make the count_unique run faster;
		add -p flag to suppress plot information
	16 Aug 2006: strat_setup bugfix: low-resolution limit handled correctly
		use pwp instead of swq->po_pwp in strat_setup
	24 Jun 2006: xgenpred.h includes xgen.h, so we don't need the latter
	16 Apr 2006: optexposure is static
	12 Apr 2006: spacepname used; record initial transmission
	30 Mar 2006: major revision to optimize near minima
	17 Mar 2006: calculate & tabulate redundancy as well as completeness
	31 Dec 2005: don't qsort with fewer than 2 elements
	28 Oct 2005: replace static globals with elements of STRATWORK struct
	12 Oct 2005: housekeeping
	28 Sep 2005: -c<val> flag added: calc resolution limit from CENTROIDS
	30 Aug 2005: significant rewrite: summary provided, bugs fixed
	17 Jun 2005: prtostrai(): mpack() takes spacegroup-number argument
	11 May 2005: check return status from xgjournal
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	20 Oct 2004: move source from ../nogui to ../utilities
	 9 Sep 2004: use prediction functions from ../xglib/xgpredict.c;
		clean up deadwood
	28 Aug 2004: routines used only here are declared as static;cosmetics
	30 Jun 2004: plotting capabilities redone (already!); resolution
	 limit imposed only if it's more restrictive than the inherent one
	22 Jun 2004: new code, partly borrowed from predict.c, etc.

   Purpose: determine a strategy for data collection based on
	the contents of a UPARAMS file
   Cross references: ../integrate/integpred.c
*************************************************************************/
#define		NEED_TIME	1
#include	"xgenpred.h"
#define		DEF_DELPHI	4 * RADC
#define		DEF_NEARMIN	0.25	/* detail to apply near minima */
#define		DEFAULT_MINY2SQ	(0.9e-3)	/* minimum allowed value of
 the square of the y component of rotated scatt vector == 1/Lorenz factor */
#define		NFRACTIONS	9

 typedef struct {
		u_char	st_hkl[4];
		short	st_nomeg;
		double	st_ome;
		} STRAT;

 typedef struct {
		double	sf_bestred, sf_fachiev, sf_ftotal;
		xglong	sf_best, sf_bestm;
		} STRATFRAC;

 typedef struct {
		double	sfp_phistart;
		double	sfp_delom[NFRACTIONS], sfp_redun[NFRACTIONS];
		} STRATPHI;

 typedef struct	{
		u_char		*sq_ruh;
		short		sq_dweak, sq_dspace, sq_expos, sq_transmit;
		int		sq_allinr, sq_flags, sq_nfound, sq_nmissing;
		int		sq_tot360, sq_nelts, sq_npred;
		double		sq_delph, sq_intrans, sq_miny2sq, sq_reslimcen;
		double		sq_step, sq_nearmin;
		FILE		*sq_fpl, *sq_fpfl;
		PREDWORK	*sq_pwp;
		STRAT		*sq_stra0, *sq_strae, *sq_strai, *sq_strap;
		STRATFRAC	*sq_sfr0;
		STRATPHI	*sq_sph0;
		} STRATWORK;

extern int stratmain(int argc, char *argv[]);
extern int count65k(MAIL *headp);

static void sum_stratin(FILE *fp, STRATWORK *swq)
{ /*	This summarizes the run-time inputs to prediction. */
	static char	*flist[] =
		{"uparams", "uspline", "amask", "xlog", NULL};
	static char	ftype[] = {1, 1, 1, 2, 0};

	ftype[3] = (swq->sq_fpl == stdout ? 0 : 2);
	header(fp, PROCESSING, flist, ftype);
	fprintf(fp,"Resolution range for strategy: %9.3f A to %9.3f A\n",
	 ((swq->sq_pwp)->po_xp)->xresmin, ((swq->sq_pwp)->po_xp)->xresmax);
	if (swq->sq_miny2sq > (DEFAULT_MINY2SQ / 2.)) fprintf(fp,
	   "Only reflections with s(y) > %9.5f will be predicted\n",
		sqrt(swq->sq_miny2sq));
	else fprintf(fp,
	   "No Lorenz-factor restrictions on predictions will be made\n");
	fprintf(fp, "Bijvoets %s be merged\n", (swq->sq_flags & 01) ?
		"will NOT" : "will");
	fprintf(fp, "Plotting files %s be created\n", (swq->sq_flags & 02) ?
		"will NOT" : "will");
	if (NULL != swq->sq_fpfl) fprintf(fp,
	 "Reflections on REFLIST will be tracked: number searched = %d\n",
		swq->sq_nfound);
	fprintf(fp, "Interval between phi values will be %6.1f degrees\n",
		swq->sq_delph / RADC);
	fprintf(fp, "Interval will be shrunk by %6.3f near minima\n",
		swq->sq_nearmin);
	if (swq->sq_intrans< 1.) fprintf(fp,
	 "Input images have a transmission fraction of %6.3f\n",
		swq->sq_intrans);
}

static void stratouthead(FILE *fp, STRATWORK *swq)
{ /*	Set up logfile header for the strategy output */
	int		j;
	STRATFRAC	*sfr;

	fprintf(fp,"\t\tSummary of strategy calculation\n");
	fprintf(fp,
 "           | Deltaphi required for the fractional completeness of:\n");
	fprintf(fp, " Start phi |");
	for (j = 0, sfr = swq->sq_sfr0; j < NFRACTIONS; j++, sfr++)
		fprintf(fp, " %6.4f", sfr->sf_ftotal);
	fprintf(fp, " of achievable reflections\n redundancy|");
	for (j = 0, sfr = swq->sq_sfr0; j < NFRACTIONS; j++, sfr++)
		fprintf(fp, " %6.4f", sfr->sf_fachiev);
	fprintf(fp, " of total      reflections\n");
	if (stdout != fp) xeflush(fp);
}

static void stratsummary(FILE *fp, STRATWORK *swq)
{ /*	Summarize results of strategy to stream fp */
	int		j;
	float		abstran, best, endp, omos0, omos1, overmos, startp;
	STRATFRAC	*sfr;
	REFPARS		*xp;

	xp = (swq->sq_pwp)->po_xp;
	fprintf(fp, "Grand summary of strategy calculation\n");
	fprintf(fp,
	 "Spacegroup %3d, i.e. %s\n cell lengths, angles = ",
	 xp->xispgp, spacepname(xp->xispgp));
	fprintf(fp, " %8.2f %8.2f %8.2f %8.2f %8.2f %8.2f\n",
	 xp->xabc[0], xp->xabc[1], xp->xabc[2],
	 xp->xang[0], xp->xang[1], xp->xang[2]);
	omos0 = xp->xgam[0] / RADC;	omos1 = xp->xgam[1] / RADC;
	fprintf(fp, " gamma0,1: %6.3f, %6.3f degrees\n", omos0, omos1);
	overmos = 0.707 * sqrt(omos0 * omos0 + omos1 * omos1);
	fprintf(fp, "Overall mosaicity = %6.3f degrees\n", overmos);
	fprintf(fp,
	 " Completeness Achievable Startphi   Endphi Deltaphi Redundancy\n");
	for (j = 0, sfr = swq->sq_sfr0; j < NFRACTIONS; j++, sfr++)
	   {
		startp = 0.1 * (double)(sfr->sf_bestm - sfr->sf_best / 2);
		if (startp < -180.) startp += 360.;
		endp = 0.1 * (double)(sfr->sf_bestm + sfr->sf_best / 2);
		best = 0.1 * (double)sfr->sf_best;
		fprintf(fp, " %12.4f %10.4f %8.2f %8.2f %8.2f %10.2f\n",
		 sfr->sf_ftotal, sfr->sf_fachiev, startp, endp, best,
		 sfr->sf_bestred);
	   }
	fprintf(fp, "Recommended parameters:\n");
	fprintf(fp, "Exposure time = %8d sec\n", (int)swq->sq_expos);
	fprintf(fp, "Distance (weak on edge) = %6d mm\n", swq->sq_dweak);
	fprintf(fp, "Distance (separation) = %6d mm\n", swq->sq_dspace);
	fprintf(fp, "Stepsize = %7.2f deg\n", swq->sq_step);
	abstran = swq->sq_transmit;
	if (swq->sq_intrans < 1.)
	  {
		fprintf(fp, "Relative Transmission = %3d%%\n",
			swq->sq_transmit);
		abstran *= swq->sq_intrans;
		fprintf(fp, "Absolute Transmission = %3.0f%%\n", abstran);
	  }
	else fprintf(fp, "Transmission = %3d%%\n", swq->sq_transmit);
	xeflush(fp);
	if (fp != stdout) return;
	best = xp->xispgp;
	(void)xgvariable("spacegroup", &best, 1);
	(void)xgvariable("cell_len", xp->xabc, -3);
	(void)xgvariable("cell_ang", xp->xang, -3);
	(void)xgvariable("gamma0", &omos0, -1);
	(void)xgvariable("gamma1", &omos1, -1);
	(void)xgvariable("mosaicity", &overmos, -1);
	sfr = swq->sq_sfr0 + (NFRACTIONS-3);
	startp = 0.1 * (double)(sfr->sf_bestm - sfr->sf_best / 2);
	if (startp < -180.) startp += 360.;
	endp = 0.1 * (double)(sfr->sf_bestm + sfr->sf_best / 2);
	best = 0.1 * (double)sfr->sf_best;
	(void)xgvariable("phistart_99", &startp, 1);
	(void)xgvariable("phiend_99", &endp, 1);
	(void)xgvariable("phirange_99", &best, 1);
	best = sfr->sf_bestred;
	(void)xgvariable("phirange_redun", &best, 1);
	best = swq->sq_expos;
	(void)xgvariable("exposure", &best, 1);
	best = swq->sq_dweak;
	(void)xgvariable("distance_weakedge", &best, 1);
	best = swq->sq_dspace;
	(void)xgvariable("distance_separation", &best, 1);
	best = swq->sq_step;
	(void)xgvariable("stepsize", &best, 1);
	if (swq->sq_intrans < 1.)
	  {
		best = swq->sq_transmit;
		(void)xgvariable("rel_transmit", &best, 1);
		(void)xgvariable("transmit", &abstran, 1);
	  }
	else {
		best = swq->sq_transmit;
		(void)xgvariable("transmit", &best, 1);
	     }
	(void)xgvariable(NULL, NULL, 0);
}

static void outstrat(FILE *fp, STRATPHI *sphq)
{ /*	This outputs the results for a specific phi value */
	int	j;

	fprintf(fp, " %9.2f |", sphq->sfp_phistart / RADC);
	for (j = 0; j < NFRACTIONS; j++)
		fprintf(fp, " %6.2f", sphq->sfp_delom[j] / RADC);
	fprintf(fp, "\n           |");
	for (j = 0; j < NFRACTIONS; j++)
		fprintf(fp, " %6.2f", sphq->sfp_redun[j]);
	fprintf(fp, "\n");	xeflush(fp);
}

static double xytosnorm(STRATWORK *swq, TROID *trp)
{ /*	This computes the norm of the diffraction vector, i.e., |s|,
	associated the TROID reflection *trp */
	int	i, j, jfr;
	double	lo, loxi, loyi, rni, rni2, los[3], losr[3];
	float	lox, loy;
	REFPARS	*xp;

	xp = (swq->sq_pwp)->po_xp;
	jfr = trp->frame0 + 0.49999;
	if (pix2cm(trp->xto, trp->yto, &lox, &loy, jfr))
	  {
		loxi = lox - xp->xx0;
		loyi = loy - xp->xy0; /* correct main beam */
		rni2 = loxi * loxi + loyi * loyi + xp->xpltd2; rni = sqrt(rni2);
		los[0] = xp->xplated / rni;
		los[1] = loxi / rni;	los[2] = loyi / rni;
		/* rotate this by the detector tilt and swing-angle matrices */
		for (i = 0; i < 3; i++)
		   {
			for (lo = j = 0; j < 3; j++)
				lo += xp->xq21[i][j] * los[j];
			losr[i] = lo;
		   }
		losr[0] += 1.;	/* correct beam-direction component */
		lo = losr[0] * losr[0] + losr[1] * losr[1] + losr[2] * losr[2];
		return sqrt(lo);
	  }
	else	return 0;
}

static void optexposure(STRATWORK *swq)
{ /*	A routine that attempts to calculate the optimal exposure time */
	int		fmmax, fmmin, fmno, imflags;
	int		ngot, ngt16, nlt16, nmbox, novers;
	double		meanmax, meanovers, rawexpos, recce, rescale;
	MAIL		*heads;
	FRAMEINDEX	*lfmi;
	FRAMEENTRY	*lft;
	DATARUN		*runhdp;
	static char	headmsg[] = "heads in optexposure";
	static int	enoughim = 10;

	swq->sq_transmit = 100;	swq->sq_expos = 3;
	if (NULL == (lfmi = ftables(3)))
	  {
		fprintf(stderr,
		 "No frame index: Can't estimate optimal exposure\n");
		return;
	  }
	fmmin = lfmi->fi_firstfm;	fmmax = lfmi->fi_lastfm;
	lft = lfmi->fi_fmi;	nmbox = 1;
	if (NULL == (heads = (MAIL *)callop(nmbox, sizeof (MAIL), headmsg)))
	  {
		fprintf(stderr,
	 "Memory allocation for images failed: Can't estimate exposure\n");
		return;
	  }
	meanmax = meanovers = nlt16 = ngt16 = 0;
	imflags = GET_RUNHEADER | GRIPE_IF_SHORT;
	fprintf(stderr, "Overflow counting in images ");
	for (fmno = fmmin, novers = ngot = 0; fmno < fmmax; fmno++, lft++)
	   {
		if (FRAME_OK == read_image(lft->fmname, imflags, heads))
		  {
			ngot++;	meanmax += heads->zmax;
			if (65535 > heads->zmax) nlt16++;
			else {
				novers = count65k(heads);	ngt16++;
				meanovers += novers;
			     }
		  }
		if (ngot > enoughim) break;
		fprintf(stderr, " %4d", fmno);
	   }
	fprintf(stderr, "\n<Overflows>,<Zmax> = %6.1f, %9.0f in %4d images\n",
	 ((double)novers / (double)ngt16),
	 ((double)meanmax / (double)ngot), ngot);
	if (ngot < 1)
	  {
		fprintf(stderr,
		 "No readable images: can't estimate exposure\n");
		kill_mail(&heads, &nmbox);	return;
	  }
	if ((NULL != (runhdp = heads->runheader)) &&
	 (FUZZ < (runhdp->run_d0pos[0]))) rawexpos = runhdp->run_expos;
	else	rawexpos = ((double)heads->elap100ths) * 0.01;
	/* scale this down by the input transmission value */
	rawexpos *= swq->sq_intrans;
	if (rawexpos < 0.1) rawexpos = 0.1;
	if ((ngt16 > nlt16) && (ngt16 > 0))
	  {
		meanovers /= (double)ngt16;
		if (meanovers > 1000) rescale = 0.02;
		else if (meanovers > 400) rescale = 0.05;
		else if (meanovers > 100) rescale = 0.1;
		else if (meanovers > 40) rescale = 0.2;
		else if (meanovers > 10) rescale = 0.5;
		else rescale = 1.;
		recce = rawexpos * rescale;
	  }
	else if (ngot)
	  {
		meanmax /= (double)ngot;
		rescale = 65535. / meanmax;	recce = rawexpos * rescale;
	  }
	else {
		recce = 10.;	rescale = 1.;	/* meaningless, we think */
	     }
	if (recce < 3) /* minimum exposure, at least at some places */
	  {
		meanmax = recce / 3.;	swq->sq_expos = 3;
		swq->sq_transmit = 100. * meanmax + 0.499999;
	  }
	else {
		swq->sq_expos = recce + 0.9999;	swq->sq_transmit = 100;
	     }
	kill_mail(&heads, &nmbox);
}

static short edgedist(REFPARS *xp)
{ /* This calculates a detector distance based on putting the weakest refls
  5% away from the edge of the detector */
	short	shedge;
	double	edged, geom0, geom1, geom2, resl, resr, totwid;
	double	xleft, xright, ymid;
	float	xlcm, xrcm, ylcm, yrcm;

	xleft = ((double)xp->xdetwid) * 0.05;	
	ymid = xp->xdetht / 2;
	xright = ((double)xp->xdetwid) * 0.95;	
	resl = pix2cm((float)xleft, (float)ymid, &xlcm, &ylcm, 0);
	resr = pix2cm((float)xright, (float)ymid, &xrcm, &yrcm, 0);
	if ((resl >= 0) && (resr >= 0))
	  {
		totwid = (xrcm - xlcm) * (xrcm - xlcm) +
		 (yrcm - ylcm) * (yrcm - ylcm);
		totwid = (totwid < DFUZZ) ? 10. : sqrt(totwid);
	  }
	else	totwid = 10.;
	geom0 = 2 * xp->xresmax * xp->xresmax;
	geom1 = geom0 - xp->xlambda * xp->xlambda; geom2 = geom0 + geom1;
	if (geom1 < DFUZZ) edged = 10.;
	else {
		geom2 = sqrt(geom2);
		edged = (totwid / (2. * xp->xlambda)) * geom1 / geom2;
	     }
	edged *= 10.;	/* centimeters converted to millimeters */
	shedge = edged + 4.9999;	shedge /= 5;	return shedge * 5;
}

static short separdist(REFPARS *xp, double *amax)
{ /* This computes the distance that optimally separates reflections */
	short	ce;
	double	lamx, resfac;

	lamx = xp->xabc[0];	if (xp->xabc[1] > lamx) lamx = xp->xabc[1];
	if (xp->xabc[2] > lamx) lamx = xp->xabc[2];
	ce = spgtocen(xp->xispgp);
	if (2 == ce) /* C-centered cells: lamx is either c or (110) */
	  {
		lamx = xp->xabc[0] * xp->xabc[0] + xp->xabc[1] * xp->xabc[1];
		lamx = (lamx < 1.) ? sqrt(lamx) : 1.;
		if (xp->xabc[2] > lamx) lamx = xp->xabc[2];
	  }
	else if (3 == ce) lamx *= 0.8662540;	/* I-centered */
	else if (4 == ce) lamx *= 0.7071068;	/* F-centered */
	else if (5 == ce) lamx *= 0.5;	/* rhombohedral indexed hexagonally */
	else if (7 == ce)	/* A-centered: lamx is either a or (011) */
	  {
		lamx = xp->xabc[1] * xp->xabc[1] + xp->xabc[2] * xp->xabc[2];
		lamx = (lamx < 1.) ? sqrt(lamx) : 1.;
		if (xp->xabc[0] > lamx) lamx = xp->xabc[0];
	  }
	*amax = lamx; /* we'll use this elsewhere */
	if (xp->xplated < 25.) resfac = 1.;
	else if (xp->xplated < 45.) resfac = -0.0015 * lamx + 1.375;
	else resfac = 0.7;
	resfac *= lamx / xp->xlambda;
	ce = resfac + 4.9999;	ce /= 5;	return ce * 5;
}

static double optstep(REFPARS *xp, double amax)
{ /* This calculates an optimal stepsize based on the cell parameters and
	gamma values */
	double	radspace, ostep;

	radspace = xp->xgam[0] * xp->xgam[0] + xp->xgam[1] * xp->xgam[1];
	/* heuristic: 1.2 times the nominal mosaicity */
	radspace = sqrt(radspace) * 1.2 / RADC;
	if ((amax > 250.) && (radspace > 0.7)) radspace = 0.7;
	if ((amax > 450.) && (radspace > 0.5)) radspace = 0.5;
	if ((amax > 600.) && (radspace > 0.3)) radspace = 0.3;
	ostep = flimits(radspace, 0.1, 2.);
	/* for the moment we're limiting ourselves to four choices */
	if (ostep < 0.25) ostep = 0.25;
	else if (ostep < 0.5) ostep = 0.5;
	else if (ostep < 0.75) ostep = 0.75;
	else ostep = 1.;
	return ostep;
}

static int strat_setup(int ac, char *av[], STRATWORK **swdp)
{ /*	This sets up for a run of the strategy program */
	register int		i, j;
	int			nelts, resu;
	char			**dcp, **flags,  fllin[256];
	u_char			*rup;
	short			tgroup, shh[3];
	double			axmax, maxgiven, maxsma, mingiven;
	double			nval, thissma;
	PREDWORK		*pwp;
	REFPARS			*xp;
	STRATFRAC		*sfr;
	static FILE		*fpce;
	static TROID		tr;
	static STRATWORK	*swq;
	static double		sfra[] =
	{ 0.500, 0.750, 0.900, 0.940, 0.970, 0.990, 0.995, 0.999, 1.000 };

	if (EHELP == xgjournal(ac, av)) return EOK; /* journal the effort */
	if (NULL == (*swdp = (STRATWORK *)callop(1, sizeof (STRATWORK),
	 "swq in strat_setup"))) return EALLOC;
	swq = *swdp;
	if (NULL == (pwp = (PREDWORK *)callop(1, sizeof (PREDWORK),
	 "pwp in strat_setup"))) return EALLOC;
	swq->sq_pwp = pwp;	/* so other routines can see it */
	maxgiven = -30000.;	mingiven = 30000.;
	swq->sq_miny2sq = DEFAULT_MINY2SQ;
	swq->sq_flags = swq->sq_reslimcen = 0;
	swq->sq_delph = DEF_DELPHI;
	swq->sq_nearmin = DEF_NEARMIN;
	swq->sq_intrans = 1.;
	flags = flagproc(ac, av, "ACDIMNPY");
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	   {
	     nval = atof((*dcp) + 1);
	     switch (**dcp)
	      {
		case	'A':	swq->sq_miny2sq = DFUZZ;	break;
		case	'C':	swq->sq_reslimcen = nval;	break;
		case	'D':	swq->sq_delph = nval * RADC;	break;
		case	'I':	swq->sq_intrans = nval;		break;
		case	'M':	if ((DFUZZ < nval) && (nval < 0.51))
			swq->sq_nearmin = nval;		break;
		case	'N':	swq->sq_flags |= 1;	break; /* unimpl. */
		case	'P':	swq->sq_flags |= 2;	break; /* no plots */
		case	'Y':	swq->sq_miny2sq = nval;		break;
	      }
	   }
	for (ac--, av++; (ac) && !allnumeric(*av); ac--) av++;
	if (ac >= 1)
	  {
		maxgiven = atof(av[0]);
		if (ac >= 2)
		  {
			if ((mingiven = atof(av[1])) < maxgiven)
			  { /* swap min and max if they're backward */
				axmax = mingiven;	mingiven = maxgiven;
				maxgiven = axmax;
			  }
		  }
	  }
		/* minimum value of |sy|^2 */
	pwp->po_miny2sq = swq->sq_miny2sq;
		/* no permuting required */
	for (j = 0; j < 3; j++) pwp->po_ih[j] = j;
	pwp->po_mumin = -PI;	pwp->po_mumax = PI;
	pwp->po_rangecen = 0.;
	pwp->po_tl1 = NOW;
	if (EOK != (resu = prpar(pwp)))
	  {
		printf("Error %d encountered in prpar\n", resu);
		return resu;
	  }
	/* if the user specified a more restrictive limit, install it */
	xp = pwp->po_xp;
	if (maxgiven > xp->xresmax)
	  {
		xp->xresmax = maxgiven;
		pwp->po_smasq = xp->xlambda / maxgiven;
		pwp->po_smasq *= pwp->po_smasq;
	  }
	if (mingiven < xp->xresmin)
	  {
		xp->xresmin = mingiven;
		pwp->po_smisq = xp->xlambda / mingiven;
		pwp->po_smisq *= pwp->po_smisq;
	  }
	/* if the user has requested a resolution limit based on
	 the contents of a CENTROIDS file, instantiate it here.
	 Since this is being done after the other resolution checks,
	 it will be the active one if it's invoked. The value
	 given in the swq->sq_reslimcen variable is the intensity value on
	 which the cutoff will be based. */
	if (swq->sq_reslimcen > DFUZZ)
	  { /* detector tilt & swing matrix */
		fprintf(stderr,
		 "establishing resolution limit based on CENTROIDS\n");
		makeq21(pwp->po_xp);
		/* try to open the CENTROIDS file */
		if (NULL != (fpce = xgopen("centroids", "r")))
		  {
			/* find refls with intensity > reslimcen.
			 Identify the max sin(theta)/lambda for those */
			maxsma = 0.;
			while (1 <= xgread(&tr, XG_TROID, fpce))
			  if (tr.totinten > swq->sq_reslimcen)
			    {
				thissma = xytosnorm(swq, &tr);
				if (thissma > maxsma) maxsma = thissma;
			    }
			(void)fclose(fpce);
			/* if we found some, set the resolution cutoff */
			if (maxsma > DFUZZ)
			  {
				pwp->po_smasq = maxsma * maxsma;
	  			xp->xresmax = xp->xlambda / maxsma;
			  }
		  }
	   }
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
	 pwp->po_p[0][i][j] = xp->xlambda * xp->xrec[i][j];
	/* create the STRATFRAC structures. The values of the
	 sf_achiev elements are set equal to the fixed sf_total for now.
	 We'll change this once we've done the prediction */
	nelts = PIX2 / (swq->sq_delph * swq->sq_nearmin);
	if (nelts < 360) nelts = 360;	swq->sq_nelts = nelts;
	if (NULL == (swq->sq_sfr0 = (STRATFRAC *)callop(NFRACTIONS,
	 sizeof (STRATFRAC), "swq->sq_sfr0 in strat_setup"))) return EALLOC;
	if (NULL == (swq->sq_sph0 = (STRATPHI *)callop(nelts,
	 sizeof (STRATPHI), "swq->sq_sph0 in strat_setup"))) return EALLOC;
	for (sfr = swq->sq_sfr0, j = 0; j < NFRACTIONS; j++, sfr++)
	   {
		sfr->sf_fachiev = sfr->sf_ftotal = sfra[j];
		sfr->sf_best = sfr->sf_bestm = 35000;	sfr->sf_bestred = 1.;
	   }
	if (NULL == (swq->sq_fpl = propen("xlog"))) swq->sq_fpl = stdout;
	if (NULL != (swq->sq_fpfl = xgopen("reflist", "r")))
	  {
		for (j = 0; NULL != fgets(fllin, 254, swq->sq_fpfl); j++) ;
		swq->sq_nfound = j;	/* count # of reflections in list */
		tgroup = xp->xispgp;
		if (NULL == (swq->sq_ruh = (u_char *)callop(4 * swq->sq_nfound,
		 sizeof (u_char), "swq->sq_ruh in strat_setup")))
		  {
			(void)fclose(swq->sq_fpfl);	swq->sq_fpfl = NULL;
		  }
		else {
			(void)fseek(swq->sq_fpfl, 0L, 0);
			for (j = 0, rup = swq->sq_ruh; NULL != fgets(fllin,
			 254, swq->sq_fpfl); j++, rup += 4)
			   {
				sscanf(fllin, "%hd %hd %hd",
					&(shh[0]), &(shh[1]), &(shh[2]));
				mpack(shh, rup, tgroup); symtr(rup, tgroup);
				*(rup+3) &= ~01;	/* merge Bijvoets */
			   }
		     }
	  }
	/* calculate various recommended values to give the user */
	swq->sq_dweak = edgedist(xp); /* distance that puts weak refl on edge*/
	swq->sq_dspace = separdist(xp, &axmax); /*dist for optimal separation*/
	/* now we estimate the ideal stepsize based on the gamma parameters */
	swq->sq_step = optstep(xp, axmax); /* stepsize */
	optexposure(swq); /* calculate exposure time and transmission */
	sum_stratin(swq->sq_fpl, swq);
	if (swq->sq_fpl != stdout) sum_stratin(stdout, swq);
	xeflush(swq->sq_fpl);	return EOK;
}

static void summ_predlist(FILE *fp, STRATWORK *swq)
{ /*	This summarizes the available reflections */
	fprintf(fp, " Symmetry-transformed reflections under scrutiny:\n");
	fprintf(fp, " predictable unobtainable    total\n");
	fprintf(fp, " %11d %12d %8d\n", swq->sq_allinr,
		swq->sq_nmissing, swq->sq_tot360);
}

static double zto2pi(double rav)
{ /*	Returns an angular value between zero and two*pi */
	int	idiv;
	double	divi;

	if (rav < 0.)
	  {
		divi = (-rav) / PIX2;	idiv = 1 + divi;
		return rav + ((double)idiv) * PIX2;
	  }
	else if (rav > PIX2)
	  {
		divi = rav / PIX2;	idiv = divi;
		return rav - ((double)idiv) * PIX2;
	  }
	else return rav;
}

static int prtostrai(STRATWORK *swq, PREDREF *pwin, STRAT *stro, int icode)
{ /*	This converts a PREDREF structure into a STRAT structure */
	register int	i;
	short		lohkl[3];
	double		locphi;
	PREDWORK	*lpwp;

	if ((!icode) || (icode > PR_ALONGBEAM)) return 0;
	lpwp = swq->sq_pwp;
	if (pwin->pindx) hunpack(pwin->pindx, lohkl);
	else for (i = 0; i < 3; i++) lohkl[i] = lpwp->po_ph[i];
	mpack(lohkl, stro->st_hkl, (lpwp->po_xp)->xispgp);
	symtr(stro->st_hkl, (lpwp->po_xp)->xispgp);
	if (!(swq->sq_flags & 01)) stro->st_hkl[3] &= ~01;
	locphi = (pwin->pphi + 0.5) * (lpwp->po_xp)->xfrmsize;
	stro->st_ome = zto2pi(locphi);
	if ((1 <= icode) && (icode <= 3)) stro->st_nomeg = 1;
	else if (icode == PR_ALONGBEAM) stro->st_nomeg = -1;
	else if (icode == PR_BADY2) stro->st_nomeg = -2;
	else if (icode == PR_PHIRANGE) stro->st_nomeg = -3;
	if (pwin->pcor < 0) stro->st_nomeg = -2;
	if (pwin->pgama < 0) stro->st_nomeg = -3;
	return stro->st_nomeg;
}

static int proc_pred(STRATWORK *swq, int predcode)
{ /* 	decides what to do about a prediction step */
	PREDWORK	*lpwp;
	PREDREF		*curp;

	lpwp = swq->sq_pwp;	curp = lpwp->po_prp - 1;
	switch (predcode)
	      {
		case	0: return prtostrai(swq, curp, swq->sq_strai,predcode);
		case	1: case	2: /* 1 intersection good, other isn't */
		 return prtostrai(swq, curp, (swq->sq_strai)++, predcode);
		case	3: /* both intersections good */
		 (void)prtostrai(swq, curp - 1, (swq->sq_strai)++, predcode);
		 return prtostrai(swq, curp, (swq->sq_strai)++, predcode);
		case	PR_ALONGBEAM: case	PR_BADY2: case PR_PHIRANGE:
		 return prtostrai(swq, curp-1, (swq->sq_strai)++, predcode);
		default:
		 return prtostrai(swq, curp, swq->sq_strai, predcode);
	      }
}

static void stratfates(STRATWORK *swq, short lh[], int resul)
{ /*	This examines the fate of the reflection lh[] */
	int		i, j, res;
	u_char		*rup, uh[4];
	short		tgroup;

	if (NULL == swq->sq_fpfl) return;
	tgroup = ((swq->sq_pwp)->po_xp)->xispgp;
	mpack(lh, uh, tgroup);	symtr(uh, tgroup);
	uh[3] &= ~01;
	for (j = 0, rup = swq->sq_ruh; j < swq->sq_nfound; j++, rup += 4)
	  for (res = i = 0; i < 4; i++)
	     {
		if (uh[i] != *(rup + i))
		  {
			res = i + 1;	break;
		  }
		if (!res)
		  {
			fprintf(swq->sq_fpl, " %3d %3d %3d %4d\n",
				lh[0], lh[1], lh[2], resul);	return;
		  }
	     }
}

static void record_fate(STRATWORK *swq, u_char uh[], int resul)
{ /*	This examines the fate of the reflection lh[] */
	int		i, j, res;
	short		llh[3];
	u_char		*rup;

	if (NULL == swq->sq_ruh) return;
	for (j = 0, rup = swq->sq_ruh; j < swq->sq_nfound; j++, rup += 4)
	   {
		for (res = i = 0; i < 4; i++)
		   if (uh[i] != *(rup + i))
		     {
			res = i + 1;	break;
		     }
		if (!res)
		  {
			munpack(uh, llh, ((swq->sq_pwp)->po_xp)->xispgp);
			fprintf(swq->sq_fpl, " %3d %3d %3d %4d\n",
				llh[0], llh[1], llh[2], resul);
			return;
		  }
	   }
}

static int stratcmp(STRAT *st0, STRAT *st1)
{ /* Compare routine for sorting STRAT structures on hkl and phi */
	int	hdif;

	if ((hdif = hklcmp(st0->st_hkl, st1->st_hkl))) return hdif;
	return (((st0->st_ome - st1->st_ome) < 0) ? -1 : 1);
}

static int stratphic(STRAT *st0, STRAT *st1)
{ /* Compare routine for sorting STRAT structures on phi */
	return (((st0->st_ome - st1->st_ome) < 0) ? -1 : 1);
}

static int construct_pred(STRATWORK *swq)
{ /*	This constructs the list of predicted reflections */
	register int	j, nminus, nposs;
	int		ngt1, nm1, nm2, noutside, npredi, nz, resu;
	short		lh[3];
	double		fracavail;
	STRAT		*strai, *stran, *strao, *straq, *strau;
	STRATFRAC	*sfr;
	PREDWORK	*lpwp;
	static short	hli[6]; /* signed index limits (e.g. hli[2] = lmin) */
	static char	countmsg[] =
 "\n%9d observations outside detector; %7d near rotation axis; %7d predicted\n";

	lpwp = swq->sq_pwp;
	for (j = 0; j < 3; j++)
	   {
		hli[j+3] = (lpwp->po_xp)->xabc[j] /
			(lpwp->po_xp)->xresmax + 0.49999;
		hli[j] = -(hli[j+3]);
	   }
	printf("indices checked over h:[%4d,%4d] k:[%4d,%4d], l:[%4d,%4d]\n",
	 hli[0], hli[3], hli[1], hli[4], hli[2], hli[5]);
	/* reserve memory for the STRAT structures */
	nposs = 2 * (hli[5] + 1 - hli[2]) *
	 (hli[4] + 1 - hli[1]) * (hli[3] + 1 - hli[0]);
	if (NULL == (swq->sq_stra0 = (STRAT *)callop(nposs, sizeof (STRAT),
	 "swq->sq_stra0 in construct_pred"))) return EALLOC;
	swq->sq_strai = swq->sq_stra0;
	swq->sq_allinr = swq->sq_tot360 = 0;
	/* predict all the reflections */
	lpwp->po_tl1 = NOW;
	lpwp->po_prp = lpwp->po_prp0; /* rewind */
	if (swq->sq_nfound) fprintf(swq->sq_fpl,
		" Flagged reflections in initial prediction:\n");
	for (lh[0] = hli[0]; lh[0] <= hli[3]; (lh[0])++)
	   {
		lpwp->po_ph[0] = lh[0];
		for (lh[1] = hli[1]; lh[1] <= hli[4]; (lh[1])++)
		   {
			lpwp->po_ph[1] = lh[1];
			for (lh[2] = hli[2]; lh[2] <= hli[5]; (lh[2])++)
			   {
				lpwp->po_ph[2] = lh[2];
				if ((resu = proc_pred(swq, pfindref(lpwp))))
				  {
					swq->sq_allinr++;
					if (resu > 0) swq->sq_tot360++;
				  }
					/* rewind the prediction list */
				lpwp->po_prp = lpwp->po_prp0;
				stratfates(swq, lh, resu);
			   }
		   }
		if (!(lh[0] % 5))
		  {
			printf(
 " Searching for reflections with h = %3d: %9d examined, %9d found so far\r",
			lh[0], swq->sq_allinr, swq->sq_tot360);
			xeflush(stdout);
		  }
	   }
	swq->sq_tot360 = swq->sq_strai - swq->sq_stra0;
	swq->sq_strae = swq->sq_strai;
	if (swq->sq_tot360 < 1) return EBADVALUE;
	/* examine these */
	noutside = nm2 = nm1 = nz = npredi = ngt1 = 0;
	for (swq->sq_strai = strai = swq->sq_stra0;
	 swq->sq_strai < swq->sq_strae; (swq->sq_strai)++)
	   {
		strai = swq->sq_strai;
		if (strai->st_nomeg < -2) noutside++;
		else if (strai->st_nomeg == -2) nm2++;
		else if (strai->st_nomeg == -1) nm1++;
		else if (strai->st_nomeg == 1) npredi++;
		else printf("Bizarre: strai(%3d,%3d,%3d,%2d),%8.2f=%d\n",
		 strai->st_hkl[0], strai->st_hkl[1], strai->st_hkl[2],
		 strai->st_hkl[3], strai->st_ome, strai->st_nomeg);
	   }
	printf(countmsg, noutside, nm2 + nm1, npredi);
	fprintf(swq->sq_fpl, countmsg, noutside, nm2 + nm1, npredi);
	/* sort the list primarily on the symmetry-transformed hkl
	 and secondarily on phi */
	printf("about to save and sort %9d predicted observations: ",
		(int)swq->sq_tot360);
	xeflush(stdout);
	/* we save a copy of the prediction list that is sorted on phi
	 rather than on hkl because we're going to use it for our
	 redundancy calculations later. This list will only include
	 usable observations */
	if (NULL == (swq->sq_strap = (STRAT *)callop((size_t)(swq->sq_tot360),
	 sizeof (STRAT), "swq->sq_strap in construct_pred"))) return EALLOC;
	for (strao = swq->sq_strap, strau = swq->sq_stra0, j = 0;
	 j < swq->sq_tot360; j++, strau++)
	   if (strau->st_nomeg >= 1)
	     {
		cpymem((AVOID *)strao, (AVOID *)strau, sizeof (STRAT));
		strao++;
	     }
	swq->sq_npred = strao - swq->sq_strap;
	if (swq->sq_npred > 1) qsort((void *)swq->sq_strap,
	 (size_t)(swq->sq_npred), sizeof (STRAT), (QPE)stratphic);
	if (swq->sq_tot360 > 1) qsort((void *)swq->sq_stra0,
	 (size_t)(swq->sq_tot360), sizeof (STRAT), (QPE)stratcmp);
	printf(
 "done\nrearranging the predicted reflections to keep track of groups: ");
	xeflush(stdout);
	swq->sq_strae = swq->sq_stra0 + swq->sq_tot360;
	if (NULL == (strau = (STRAT *)callop(swq->sq_tot360, sizeof (STRAT),
	 "strau in construct_pred"))) return EALLOC;
	strao = strau;
	/* reorganize the ones that have been grouped together.
	 Any STRAT structure for which the only value found has
	 straq->st_nomeg equal to -1 is one we never predicted at all.
	 We don't keep it in the output structure; instead,
	 we just remember that that reflection was missing */
	fprintf(swq->sq_fpl,
		" Outputting specified reflections after merging:\n");
	for (swq->sq_nmissing = swq->sq_allinr = 0,
	 swq->sq_strai = swq->sq_stra0; strai < swq->sq_strae; )
	   { /* find the beginning of the next reflection */
		strai = swq->sq_strai;	stran = NULL;
		for (straq = swq->sq_strai + 1; straq < swq->sq_strae; straq++)
		  if (hklcmp(straq->st_hkl, strai->st_hkl))
		    {
			stran = straq;	break;
		    }
		if (stran == NULL) stran = swq->sq_strae; /* ran off the end */
		/* look for the "we didn't find it" structures */
		for (nminus = 0, straq = swq->sq_strai; straq < stran; straq++)
			if (straq->st_nomeg <= -1) nminus++;
		/* if that's all we have, we never got this one */
		if (nminus >= stran - swq->sq_strai)
		  {
			(swq->sq_nmissing)++;
			if (swq->sq_nfound) record_fate(swq,
				swq->sq_strai->st_hkl, -nminus);
		  }
		/* otherwise, we write out all the good ones */
		else { /* find the first good one and put the # of
			good ones in the st_nomeg field of that one */
			for (straq = swq->sq_strai; straq < stran; straq++)
				if (straq->st_nomeg > -1) break;
			if (straq >= stran)
			  {
				(swq->sq_nmissing)++;
				if (swq->sq_nfound) record_fate(swq,
				 (swq->sq_strai)->st_hkl, -(straq - stran));
			  }
			else {
				straq->st_nomeg =
					(stran - swq->sq_strai) - nminus;
				if (swq->sq_nfound) record_fate(swq,
				 swq->sq_strai->st_hkl, straq->st_nomeg);
				cpymem((void *)strao, (void *)straq,
				 sizeof (STRAT));
				strao++;
				for (straq++; straq < stran; straq++)
				   if (straq->st_nomeg > -1)
				     {
					straq->st_nomeg = 0;
					cpymem((void *)strao, (void *)straq,
					 sizeof (STRAT));	strao++;
				     } /* end of "for (straq++" loop */
			     } /* end of else of "if (straq >= stran" clause*/
		     } /* end of else of "if (nminus == stran -... " clause */
		swq->sq_strai = stran; /* move on to the next refl */
	  }
	printf("done\n");	xeflush(stdout);
	/* having rebuilt the prediction list , we count unique reflections */
	for (swq->sq_allinr = 0, straq = strau; straq < strao; straq++)
		if (straq->st_nomeg > 0) swq->sq_allinr++;
	swq->sq_tot360 = swq->sq_nmissing + swq->sq_allinr;
	summ_predlist(swq->sq_fpl, swq);	summ_predlist(stdout, swq);
	memzap((AVOID **)&(swq->sq_stra0), "swq->sq_stra0 in construct_pred");
	swq->sq_stra0 = strau;	swq->sq_strae = strao;
	fracavail = ((double)swq->sq_allinr) / ((double)swq->sq_tot360);
	for (sfr = swq->sq_sfr0, j = 0; j < NFRACTIONS; j++, sfr++)
		sfr->sf_fachiev = sfr->sf_ftotal * fracavail;
	return EOK;
}

static STRAT *treestrat(STRATWORK *swq, double phi)
{ /* This returns the STRAT structure whose omega value is just below
   the value phi */
	register int	j, bot, top, endv;
	double		del;
	STRAT		*sw0, *swe;

	/* establish outer limits for search */
	endv = swq->sq_npred - 1;
	sw0 = swq->sq_strap;	swe = sw0 + endv;
	if (phi < sw0->st_ome) return sw0;
	if (phi >= swe->st_ome) return swe;
	top = endv + 1;
	/* step through the list and look for a match */
	for (bot = 0; bot != (j = (top + bot)/2); )
	   {
		del = phi - (sw0 + j)->st_ome;
		if (del < -DFUZZ) top = j;
		else if (del > DFUZZ) bot = j;
		else return sw0 + j;
	   }
	if (((sw0 + j)->st_ome > phi) || ((j < endv) &&
	 (phi > (sw0 + j + 1)->st_ome)))
	  {
		fprintf(stderr, "Cannot fit %8.4f into list\n", phi);
		return sw0;
	  }
	return sw0 + j;
}

static int count_unique(STRATWORK *swq, double phimi, double phima)
{ /*	This loops through the list of reflections that are within the
	resolution limit and sees how many of them have at least one
	instance of having been predicted to arise at a phi value
 	between phimi and phima. We only examine the reflections that
	have not yet been located in some range beginning at phimi.
	In this version we do not keep track of redundancy in this routine;
	instead; we depend on treestrat() to help with that. That frees us
	up to return as soon as we find a hit */
	int	ninr;
	double	qpmi, qpma, tph;
	STRAT	*stra, *stre, *strq;

	ninr = 0;	stra = swq->sq_stra0;
	if ((phima > PIX2) && (phimi <= PIX2))
	  { /* we're wrapping around two pi: more complicated */
		qpma = phima - PIX2;
		for ( ; stra < swq->sq_strae; )
		   {
			stre = stra + stra->st_nomeg;
			if (!(stra->st_hkl[3] & 020))
			  {
				/* if either the last observation is after
				 phimi, or if the initial observation is
				 before qpma, then this reflection DOES
				 appear within this range */
				if (((stre - 1)->st_ome >= phimi) ||
				 (stra->st_ome <= qpma))
				  {
					stra->st_hkl[3] |= 020;	ninr++;
				  }
			  }
			stra = stre;
		   }
	  }
	else { /* we're not wrapping around two pi: simpler */
		if (phimi > PIX2)
		  {
			qpmi = phimi - PIX2;	qpma = phima - PIX2;
		  }
		else {
			qpmi = phimi;		qpma = phima;
		     }
		for ( ; stra < swq->sq_strae; )
		   {
			stre = stra + stra->st_nomeg;
			if (!(stra->st_hkl[3] & 020))
			  {
				if (((stre - 1)->st_ome >= qpmi) &&
				 (stra->st_ome <= qpma))
				  {
					for (strq = stra; strq < stre; strq++)
					 if ((qpmi <= (tph = strq->st_ome)) &&
					  (tph <= qpma))
					   {
						stra->st_hkl[3] |= 020;	ninr++;
						break;
					   }
				  }
			  }
			stra = stre;
		   }
	     }
	return ninr;
}

static int endstrat(int arg, STRATWORK **swdp)
{ /* Cleanly closes off a run of `strategy' */
	STRATWORK	*swq;

	swq = *swdp;
	memzap((AVOID **)&(swq->sq_stra0), "swq->sq_stra0 in endstrat");
	memzap((AVOID **)&(swq->sq_strap), "swq->sq_strap in endstrat");
	memzap((AVOID **)&(swq->sq_sfr0), "swq->sq_sfr0 in endstrat");
	memzap((AVOID **)&(swq->sq_sph0), "swq->sq_sph0 in endstrat");
	memzap((AVOID **)swdp, "swq in endstrat");	return arg;
}

static double stratphi(STRATPHI *sph, double delph, STRATWORK *swq)
{ /* This processes results for a specific phi start value */

	int		idelom, imid, j, ndepe, ninr, nobs;
	double		mid, uphis, phie, redun, tdel, tfrac;
	STRAT		*stra, *ustr0, *ustrn, *ustrs, *ustre;
	STRATFRAC	*sfr;

	ndepe = NFRACTIONS;	/* number of checked completeness levels */
	for (j = 0; j < ndepe; j++) sph->sfp_redun[j] = 0.; 
	uphis = sph->sfp_phistart;	if (uphis < 0.) uphis += PIX2;
	ustr0 = swq->sq_strap;	ustrn = ustr0 + (swq->sq_npred) - 1;
	ustrs = treestrat(swq, uphis);
	/* for all the reflections in our list, we explicitly clear
	 the bit that tells us whether we've found this reflection yet. */
	for (stra = swq->sq_stra0; stra < swq->sq_strae; )
	   {
		stra->st_hkl[3] &= ~020;	stra += stra->st_nomeg;
	   }
	ninr = 0;
	for (phie = uphis + delph; phie < uphis + PIX2; phie += delph)
	   {
		ninr += count_unique(swq, phie - delph, phie);
		/* 1 missing reflection isn't worth worrying about*/
		if ((swq->sq_allinr > 10000) &&
			(ninr == swq->sq_allinr - 1)) ninr++;
		tfrac = (double)ninr / (double)swq->sq_allinr;
		for (j = 0, sfr = swq->sq_sfr0; j < ndepe; j++, sfr++)
		   {
			if (tfrac >= sfr->sf_ftotal)
			  {
				if (sph->sfp_redun[j] < DFUZZ)
				  {
					sph->sfp_delom[j] = phie +delph -uphis;
					if (phie > PIX2)
					  {
						ustre = treestrat(swq,
							phie - PIX2);
						nobs = 1 + (ustre - ustr0) +
							(ustrn - ustrs);
					  }
					else {
						ustre = treestrat(swq, phie);
						nobs = ustre - ustrs;
					     }
					sph->sfp_redun[j] = (ninr > 0) ?
					 (double)nobs / (double)ninr : 0;
				  }
			  }
			else break;
		   }
		if ((j >= ndepe) || (ninr >= swq->sq_allinr)) break;
	   }
	/* if some of the sfr->sf_delom values are still unpopulated,
	 we set them to PIX2 */
	for (j = 0; j < ndepe; j++)
		if (sph->sfp_redun[j] < DFUZZ) sph->sfp_delom[j] = PIX2;
	for (sfr = swq->sq_sfr0, j = 0; j < ndepe; j++, sfr++)
	   if ((redun = sph->sfp_redun[j]) > DFUZZ)
	     {
		tdel = sph->sfp_delom[j];	mid = uphis + 0.5 * tdel;
		while (mid < -PI) mid += PIX2;
		while (mid > PI) mid -= PIX2;
		/* ignore improvements smaller than 0.1degrees,
		 so we round off to the nearest 0.1 degrees. */
		idelom = tdel * 10. / RADC + 0.49999;
		imid = (mid < 0.) ? mid * 10. / RADC - 0.49999 :
			mid * 10. / RADC + 0.49999;
		if ((idelom < sfr->sf_best) || ((idelom == sfr->sf_best) &&
		 (abs(imid) < abs(sfr->sf_bestm))))
		  {
			sfr->sf_best = idelom;	sfr->sf_bestm = imid;
			sfr->sf_bestred = redun;
		  }
	    }
	printf(" Analysis complete for phi = %7.2f\r", uphis / RADC);
	xeflush(stdout);
	return sph->sfp_delom[ndepe - 3];
}

static int rphicmp(STRATPHI*sp0, STRATPHI *sp1)
{ /* compare routine for STRATPHI sorting */
	double	de0, defuz;

	defuz = 0.01 * RADC;
	de0 = sp0->sfp_phistart - sp1->sfp_phistart;
	if (de0 < -defuz) return -1;	if (de0 > defuz) return 1;
	de0 = sp0->sfp_delom[NFRACTIONS -3] - sp1->sfp_delom[NFRACTIONS -3];
	if (de0 < -defuz) return -1;	if (de0 > defuz) return 1;
	return 0;
}

static int reorg_strat(STRATWORK *swq)
{ /* reorganize the swq->sf_sfr0 array */
	char		*legq, **dlegp, **dlegq;
	int		i, j, ndepe, plerr;
	float		pinde, *flp, pdepe[NFRACTIONS];
	FILE		*fpplo;
	STRATPHI	*sphq, *sphr;
	STRATFRAC	*sfr;
	static char	lege[12 * NFRACTIONS];

	fprintf(stderr, "\nReorganizing output information: ");
	ndepe = NFRACTIONS;	/* number of checked completeness levels */
	if (NULL == (dlegp = (char **)callop(ndepe, sizeof (char *),
	 "dlegp in stratmain"))) return EALLOC;
	for (i = 0, sfr = swq->sq_sfr0 + ndepe - 1, legq = lege, dlegq = dlegp;
	 i < ndepe; i++, sfr--, legq += 12, dlegq++)
	   {
		sprintf(legq, " %6.1f", sfr->sf_ftotal * 100.);	*dlegq = legq;
	   }
	if (!(swq->sq_flags & 02))
	  {
		if (NULL == (fpplo = plotsetup(ndepe, 1, &plerr,"strategy.out",
	 "Results of strategy calculation", "Starting scanning-angle value,deg",
	 "Scanning angle range required, deg",dlegp)))
			return EOPENERR;
	  }
	else fpplo = NULL;
	/* make sure all phi values are positive */
	for (i = 0, sphq = swq->sq_sph0; i < swq->sq_nelts; i++, sphq++)
		if (sphq->sfp_phistart < 0.) sphq->sfp_phistart += PIX2;
	/* now, sort this on increasing order of phi */
	if (swq->sq_nelts > 1) qsort((void *)swq->sq_sph0,
	 (size_t)(swq->sq_nelts), sizeof (STRATPHI), (QPE)rphicmp);
	/* get rid of duplicates */
	for (sphq = swq->sq_sph0, i = 0;
	 i < swq->sq_nelts - 1; i++, sphq++, sphr++)
	   if ((sphq + 1)->sfp_phistart - sphq->sfp_phistart < 0.01 * RADC)
	     {
		for (sphr = sphq + 1, j = i + 1; j < swq->sq_nelts; j++, sphr++)
		  cpymem((AVOID *)sphr, (AVOID *)(sphr+1), sizeof (STRATPHI));
		(swq->sq_nelts)--;
	     }
	if (!(swq->sq_flags & 02)) 
		fprintf(stderr, "Outputting plot info:\nPlot bin ");
	/* output that */
	for (i = 0, sphq = swq->sq_sph0; i < swq->sq_nelts; i++,sphq++)
	   {
		outstrat(swq->sq_fpl, sphq);
		if (!(swq->sq_flags & 02))
		  {
			for (j = 0, flp = pdepe; j < ndepe; j++, flp++)
				*flp = sphq->sfp_delom[j] / RADC;
			pinde = sphq->sfp_phistart / RADC;
			(void)plotadd(fpplo, ndepe, pinde, pdepe);
			fprintf(stderr, " %3d", i);
			if (!((i+1) % 20)) fprintf(stderr, "\nPot bin ");
		  }
	   }
	if (!(swq->sq_flags & 02)) plotfinish(fpplo, 0.);
	fprintf(stderr, "\nDone\n");
	printf("\n");	xeflush(stdout);
	stratsummary(stdout, swq);	stratsummary(swq->sq_fpl, swq);
	memzap((AVOID **)&dlegp, "dlegp in stratmain");	return EOK;
}

int stratmain(int argc, char *argv[])
{ /*	"stratmain":
	the calling routine used in determining data collection strategy. */
	int		nphi, result;
	double		absmin, delph, lphimax, phis, phise, smdel;
	double		wid99a, wid99b, wid99c, widdet;
	STRATWORK	*swq;
	STRATPHI	*sphq;

	/* read in the input parameters from a file */
	if (EOK != (result = strat_setup(argc, argv, &swq)))
		return endstrat(result, &swq);
	if (EOK != (result = construct_pred(swq)))
		return endstrat(result, &swq);
	delph = swq->sq_delph;	/* omega interval between test points*/
	stratouthead(swq->sq_fpl, swq);	/* prepare for output */
	phis = 0.;	/* default start scan angle */
	nphi = 1 + (int)((PIX2 / delph) + 0.49999); /* number of intervals */
	wid99a = PIX2 * 2.;	wid99b = absmin = PIX2;
	smdel = delph * swq->sq_nearmin;
	for (phis = 0, sphq = swq->sq_sph0; phis < PIX2; phis += delph)
	   {
		sphq->sfp_phistart = phis;
		wid99c = stratphi(sphq, delph, swq);	sphq++;
		if ((wid99b <= wid99c) && (wid99b <= wid99a) &&
		 ((wid99b < 0.98 * wid99a) || (wid99b < 0.98 * wid99c)) &&
		 (wid99b < 1.1 * absmin))
		  {
			lphimax = phis + 0.001 - 0.5 * delph;
			for (phise = phis - 1.5 * delph; phise < lphimax;
			 phise += smdel)
			   {
				sphq->sfp_phistart = phise;
				widdet = stratphi(sphq, smdel, swq);	sphq++;
			   }
		  }
		if (wid99c < absmin) absmin = wid99c;
		wid99a = wid99b;	wid99b = wid99c;
	   }
	swq->sq_nelts = sphq - swq->sq_sph0;
	return endstrat(reorg_strat(swq), &swq);
}

/* end of stratcalc.c */
