/* urefutils.c -- utilities for manipulating output from `integrate'

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	27 Jun 2007: return rather than exit in x2urfmain
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	18 May 2006: merge {cuturefl,filterurf,utoshel,pickuref,x2urf}.c here
	 6 Nov 2005: utoshel: re-insert the contents of proc_refl()
		into the main so that we don't have to use static globals
	12 Oct 2005: cuturefl: housekeeping
	30 Jun 2005: parsex(): sscanf %d, not %ld now
	28 Jun 2005: x2urfmain(): typecast pointer differences upon output
	22 Jun 2005: matchurf(), mergeurf(), fixomega(), parsex()
		all declared as static
	13 May 2005: move x2urf from nogui to utilities
	11 May 2005: check return status from xgjournal
	 4 Apr 2005: x2urf: major revision to make it actually work
	20 Oct 2004: x2urf: don't need to declare make_ftab -- never used
	15 Oct 2004: sources moved from ../nogui to ../utilities
	15 Oct 2004: filterurf: source moved from ../nogui to ../utilities
	 6 Sep 2004: cosmetic change to utoshel
	28 Aug 2004: routine proc_refl() used only here is declared as static
	27 Aug 2004: routines used only here are declared as static;cosmetics
	18 Aug 2004: cuturefl: replaced static globals with function args
	18 Aug 2004: filterurf: args to sum_filurun obviate static globals
	 2 Aug 2004: add new arg to find_dimension()
	27 Jul 2004: get xscal,yscal via dimscale()
	30 May 2004: move structures to ../include/xgen.h;
		move output functionality to ../xglib/ascout.c
	28 Feb 2004: pick_uref: correct call to centolim() in urptolim()
	18 Feb 2004: change kill_ftab calls so that they don't have args
	26 Jan 2004: pick_uref: convert some static globals to local variables;
		eliminate opening of UPARAMS since it's implicit in refparin()
	25 Jan 2004: utoshel: convert most static globals into local variables
	24 Jan 2004: some static globals converted to local variables
	21 Jan 2004: license message inserted
	 8 Jan 2004: static globals, local variables replace extern globals
	13 Dec 2003: x2urf: eliminate unused variables inorder,
		prevfm from make_xft()
	19 Sep 2003: x2urf: xglong variables sscan'd using %ld formats
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	 5 Jan 2003: reorganized source-file header
	 1 Nov 2002: x2urf: clean up source a bit
	25 Mar 2002: removed MSI attachments
	17 Jun 1998: initial version of filterurf
	16 Mar 1994: cuturefL: a few I/O changes, global variables moved around
	 7 Dec 1993: initial X-GEN version of cuturefl and utoshel
	21 Jul 1995: pick_uref: major rewrite:
		substantial functionality moved to ../xglib/threedpro.c
	 6 Jul 1995: pick_uref: reorganize 3-D profile display
		so it's callable from outside this file
	12 May 1995: pick_uref: make_ftab takes no argument;
		totalframes is a global
	 2 Mar 1995: pick_uref: diagnostics to stdout
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	27 Oct 1994: pick_uref: added -e (Transformed-coordinate I&sigma)
		option; reorganized options
	 5 Oct 1994: use xgread for REFPARS and OBSREF data
	23 Sep 1994: pick_uref: fix plane-spacing calculation
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: pick_uref: a few variables moved around
	14 Jan 1994: new memory management and user interface tools
	17 Dec 1993: pick_uref: Restore "include xgen.h" in source
	 7 Dec 1993: initial X-GEN version of pick_uref and x2urf
	29 Jul 1991: pick_uref: allow for background-subtraction ("-b")
	 7 Jun 1990: pick_uref: add raw-frame displays a la `ewmask'
	16 May 1988: pick_uref: initial code, based on `xdump'

   Purpose: manipulating and printing UREFLS file contents, specifically:
	cuturefl	reads in a UREFLS file and writing another out in which
		only reflections within a specified frame range are included.
	filterurf	read in a UREFLS file and writing another out in which
		only sensible reflections (sigma > 0.01, I > -1.e5) are included
	utoshel		convert UREFLS format to SHELX input format
	pick_uref	print selected reflections from a UREFLS file.
			The reflections to be printed are those contained
			in the file with environment variable REFLIST.
	x2urf		convert from DENZO's ".x" format to UREFLS format
   Cross references: ../integrate/int*.c,../xglib/ascout.c,../nogui/ureftools.c
*************************************************************************/
#include	"xgen.h"
#define		ENOUGH_ITEMS		38
#define		FLOOR_BG	30
#define		MAXBG		1000
#define		MOSAIC_FACTOR	1.1
#define		BRIGHTIOS2	36.
#define		MAXBUNDLE	16
#define		ALREADYCODE	213
	/* bits in `pu_flags' for pick_uref */
#define	PU_IMAGES	1
#define	PU_SUBBG	2
#define	PU_USE_ABG	4

 typedef struct { /* convenience structure for pick_uref */
		int		pu_nlines, pu_flags;
		char		pu_bgname[80];
		u_char		*pu_cp0, **pu_checks;
		double		pu_scl[4];
		FILE		*pu_fpi, *pu_fpo;
		} PUWORK;

 typedef struct { /* structure of refl lines in .x file */
		xglong	xf_hkl[3];	/* indices of reflection */
		xglong	xf_partial;	/* 1==partial, 0 not */
		double	xf_ifit;	/* profile-fitted intensity */
		double	xf_isum;	/* summation intensity */
		double	xf_chisq;	/* profile-fitting chisquared */
		double	xf_sigma;	/* sigma(I) */
		double	xf_cosinc;	/* cos(incidence angle at det.*/
		double	xf_xpred;	/* Xpredicted, pixels */
		double	xf_ypred;	/* Ypredicted, pixels */
		double	xf_lpo;		/* lorentz, polarization, obliquity */
		double	xf_pstrength;	/* strength of averaged profile */
		} XLINE;

extern int cutumain(int argc, char *argv[]);
extern int filterurfmain(int argc, char *argv[]);
extern int utoshelmain(int argc, char *argv[]);
extern int pickumain(int argc, char *argv[]);
extern int x2urfmain(int argc, char *argv[]);

/*****************************************
	cuturefl: get rid of reflections from the input UREFLS if they are
	found to be at frame numbers earlier than or greater than
	the user-specified first- and last-frame value.
******************************************/

static void sum_cuturun(FILE *fp, int rin, int rout,
	double lasx, double lasy, OBSREF *lgoop)
{ /*	This summarizes a run of `cuturefl'. */
	fprintf(fp, "%6d reflections read; %6d reflections written\n",
		rin, rout);
	fprintf(fp, "Last good refl: (%3d %3d %3d) @ [%3.0lf,%3.0lf,%5.1f]\n",
	 lgoop->obs_hkl[0], lgoop->obs_hkl[1], lgoop->obs_hkl[2],
	 lasx, lasy, lgoop->obs_ompred);
}

int cutumain(int argc, char *argv[])
{ /*	Mainline for cutting off a UREFLS on the basis of image number */
	int		firstframe, lastframe;
	int		locht, locwid, nin, nout, specframe;
	double		lastx, lasty, omo, xscal, yscal;
	FILE		*fplog, *fpui, *fpuo;
	OBSREF		lastgood, ob;
	static char *fnames[] = {"urefls", "urefout", "xlog", NULL};
	static char ftypes[] = {1, 2, 2, 0 };

	firstframe = 0;	lastframe = 440;
	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	if (argc > 1)
	  {
		if (0 <= (specframe = atoi(argv[1]))) firstframe = specframe;
		if (argc > 2 && (0 <= (specframe = atoi(argv[2]))))
			lastframe = specframe;
	  }
	if (lastframe < firstframe)
	  {
		specframe = lastframe;	lastframe = firstframe;
		firstframe = specframe;
	  }
	find_dimension(&locwid, &locht, 0, (char *)NULL);
	xscal = dimscale((short)locwid);	yscal = dimscale((short)locht);
	if (qerror((AVOID *)(fpui = xgopen("urefls", "r")),
		"open input UREFLS file")) return EOPENERR;
	if (qerror((AVOID *)(fpuo = xgopen("urefout", "w")),
		"open output UREFLS file")) return EOPENERR;
	if (NULL == (fplog = propen("xlog"))) fplog = stdout;
	ftypes[2] = (fplog == stdout ? 0 : 2);
	header(stdout, REDUCTION, fnames, ftypes);
	if (fplog != stdout) header(fplog, REDUCTION, fnames, ftypes);
	printf("Reflections from frames %3d through %3d will be output\n",
			firstframe, lastframe);
	if (fplog != stdout) fprintf(fplog,
		"Reflections from frames %3d through %3d will be output\n",
			firstframe, lastframe);
	for (nin = nout = 0; 0 < xgread((void *)&ob, XG_OBSREF, fpui); nin++)
	   {
		if ((firstframe <= (omo = ob.obs_ompred)) && omo <= lastframe)
		  {
			if (verror(0 < xgwrite((void *)&ob, XG_OBSREF, fpuo),
				"cannot write a reflection to the output"))
			  {
				if (stdout != fplog) (void)fclose(fplog);
				return EWRITERR;
			  }
			cpymem((AVOID *)&lastgood, (AVOID *)&ob,
				sizeof (OBSREF));
			nout++;
		  }
	   }
	lastx =  (((double)lastgood.obs_xpred) + 0.49999) / xscal;
	lasty =  (((double)lastgood.obs_ypred) + 0.49999) / yscal;
	sum_cuturun(stdout, nin, nout, lastx, lasty, &lastgood);
	if (stdout != fplog)
	  {
		sum_cuturun(fplog, nin, nout, lastx, lasty, &lastgood);
		(void)fclose(fplog);
	  }
	globalkill();	return EOK;
}

/* end of cuturefl functionality */

/**********************************
	filterurf: eliminate nonsense reflections from a UREFLS file
	 It gets rid of reflections from the input if they are
	 found to have nonsense intensities or sigmas.
*************************************************************************/

static void sum_filurun(FILE *fp, int rin, int rout,
 double lasx, double lasy, OBSREF *lasgoo)
{ /*	This summarizes a run of `filterurf'. */
	fprintf(fp, "%6d reflections read; %6d reflections written\n",
		rin, rout);
	fprintf(fp, "Last good refl: (%3d %3d %3d) @ [%3.0f,%3.0f,%5.1f]\n",
	 lasgoo->obs_hkl[0], lasgoo->obs_hkl[1], lasgoo->obs_hkl[2],
	 lasx, lasy, lasgoo->obs_ompred);
}

static void badr_header(FILE *fp)
{ /*	This produces a header for the list of bad reflections */
	fprintf(fp, "Nonsense reflections:\n");
	fprintf(fp,
 "   h   k   l        I(fit)      var(fit)        I(sum)      var(sum)\n");
	if ((fp != stdout) && (fp != stderr)) (void)xeflush(fp);
}

static void thisbad(FILE *fp, OBSREF *obp)
{ /*	This summarizes one bad reflection */
	fprintf(fp, " %3d %3d %3d %12.5e %12.5e %12.5e %12.5e\n",
	 obp->obs_hkl[0], obp->obs_hkl[1], obp->obs_hkl[2],
	 obp->obs_ii, obp->obs_sigma, obp->obs_iunfit, obp->obs_sunfit); 
}

int filterurfmain(int argc, char *argv[])
{ /* Mainline for filtering out nonsense reflections from a UREFLS file */
	int			locht, locwid, nbad, nin, nout;
	double			lastx, lasty, xscal, yscal;
	FILE			*fplog, *fpui, *fpuo;
	OBSREF			lastgood, ob;
	static char *fnames[] =	{"urefls", "urefout", "xlog", NULL};
	static char ftypes[] =	{1, 2, 2, 0 };

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	nin = nout = nbad = 0;
	find_dimension(&locwid, &locht, 0, (char *)NULL);
	xscal = dimscale((short)locwid);	yscal = dimscale((short)locht);
	if (qerror((AVOID *)(fpui = xgopen("urefls", "r")),
		"open input UREFLS file")) return EOPENERR;
	if (qerror((AVOID *)(fpuo = xgopen("urefout", "w")),
		"open output UREFLS file")) return EOPENERR;
	if (NULL == (fplog = propen("xlog"))) fplog = stdout;
	ftypes[2] = (fplog == stdout ? 0 : 2);
	header(stdout, REDUCTION, fnames, ftypes);
	if (fplog != stdout) header(fplog, REDUCTION, fnames, ftypes);
	for ( ; 0 < xgread((void *)&ob, XG_OBSREF, fpui); nin++)
	   {
		if ((ob.obs_ii > -1.e5) && (ob.obs_sigma > .1) &&
		    (ob.obs_iunfit > -1.e5) && (ob.obs_sunfit > .1))
		  {
			if (verror(0 < xgwrite((void *)&ob, XG_OBSREF, fpuo),
				"cannot write a reflection to the output"))
			  {
				if (stdout != fplog) (void)fclose(fplog);
				return EWRITERR;
			  }
			cpymem((AVOID *)&lastgood, (AVOID *)&ob,
				sizeof (OBSREF));
			nout++;
		  }
		else {
			if (!nbad)
			  {
				badr_header(fplog);	badr_header(stderr);
			  }
			thisbad(fplog, &ob);	thisbad(stderr, &ob);
			nbad++;
		     }
	   }
	lastx =  (((double)lastgood.obs_xpred) + 0.49999) / xscal;
	lasty =  (((double)lastgood.obs_ypred) + 0.49999) / yscal;
	sum_filurun(stdout, nin, nout, lastx, lasty, &lastgood);
	if (stdout != fplog)
	  {
		sum_filurun(fplog, nin, nout, lastx, lasty, &lastgood);
		(void)fclose(fplog);
	  }
	globalkill();	return EOK;
}

/* end of filterurf functionality */

/**************************************
	utoshel converts a UREFLS file to a SHELX input.
	The output format for each reflection is given in printshel()
		in ../xglib/ascout.c
 Syntax:
	utoshel [-flags] [runnumber]
 Flags:
	 -s	use the summation data rather than the profile fitted data.
	 -n	set all the batch numbers equal to "runnumber".  Otherwise
	 each reflection is assigned a batch number according to
	 batchno = 2 + 200 * runnumber + (omega - minomega) / 20;
	 Here "minomega" is the smallest scanning angle value in the dataset.
	 Omega and minomega are given in FRAMES, so we don't need to read
	 in any other files in order to run this turkey.
 The program reads a UREFLS file with logical name or environment
 variable UREFLS; it writes to the logical name or env. vbl. XHKL.
 The data are scanned quickly before the real calculations begin
 to find minomega and to determine the largest intensity value in the data;
 if the latter is > 9999, the whole set it rescaled
 so that the maximum intensity is in fact 9999.
***************************************/

int utoshelmain(int argc, char *argv[])
{ /*	This is the mainline for "utoshel".  It reads in the reflections
	 once to find the minimum omega and maximum intensity;
	 then it goes through for real for processing. */
	register int	j, nrefls;
	int		batchbase, onebatch, usesum;
	char		**dcp, **flags;
	double		ii, iscale, maxinten, minomega, var;
	FILE		*fpi, *fpo;
	OBSREF		urf;
	SHEL_HKL	xh;
	static short	mhkl[3];
	static char	*fnames[] = { "UREFLS", "XHKL", NULL } ;
	static char	ftypes[] = { 1, 2, 0 } ;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
		/* open the input UREFLS file */
	if (qerror((AVOID *)(fpi = xgopen("UREFLS", "r")),
	 "open input UREFLS file")) return EOPENERR;
	header(stdout, REDUCTION, fnames, ftypes);
		/* process the flag arguments */
	flags = flagproc(argc, argv, "NS");
	usesum = onebatch = 0;
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	   switch (**dcp)
	     {	case	'N':	onebatch = 1;	break;
		case	'S':	usesum = 1;	break;
	     }
		/* skip the flags to get to the numeric argument, if any */
	for (argc--, argc++; (argc) && !allnumeric(*argv); argc--) argv++;
	if (!argc ) batchbase = 0;
	else batchbase = atoi(*argv);
	if (onebatch) printf("Batch number = %d for all reflections\n",
		batchbase);
	else printf("Batch values will begin at %d and go up\n",
			batchbase * 200 + 2);
	printf("%s intensities will be used\n",
		(usesum ? "Summed" : "Profile-fitted"));
		/* set silly values for the max intensity & minimum frame # */
	maxinten = -1.e30;	minomega = -maxinten;
		/* go through the data and set proper values for the above */
	for (nrefls = 0; 0 < xgread((void *)&urf, XG_OBSREF, fpi); nrefls++)
	   {	if (usesum)
		  {	if (urf.obs_iunfit > maxinten)
			  {	maxinten = urf.obs_iunfit;
				for (j = 3; j--; ) mhkl[j] = urf.obs_hkl[j];
			  }
		  }
		else if (urf.obs_ii > maxinten)
		  {	maxinten = urf.obs_ii;
			for (j = 3; j--; ) mhkl[j] = urf.obs_hkl[j];
		  }
		if (urf.obs_omob < minomega) minomega = urf.obs_omob;
	   }
		/* if max intensity > 9999, use it as a rescaler;
		otherwise ignore it */
	if (maxinten > 9999.) iscale = 9999. / maxinten;
	else iscale = 1.;
		/* rewind the input file */
	if (verror (!fseek(fpi, 0L, 0), "rewind input UREFLS file"))
		return ESEEKERR;
		/* open the output XHKL file */
	if (qerror((AVOID *)(fpo = xgopen("XHKL", "w")),
		"open output XHKL file")) return EOPENERR;
	printf(
	 "Maximum intensity = %11.2f for [%3d,%3d,%3d]; min omega = %11.2f\n",
		maxinten, mhkl[0], mhkl[1], mhkl[2], minomega);
	printf("Overall scale factor = %8.4f\n", iscale);
	printf("%6d reflections read in\n", nrefls);
		/* go back through the data and do the real work */
	for (nrefls = 0; 0 < xgread((void *)&urf, XG_OBSREF, fpi); nrefls++)
	   {	for (j = 3; j--; ) xh.x_hkl[j] = urf.obs_hkl[j];
		if (usesum)
		  {	ii = urf.obs_iunfit;	var = urf.obs_sunfit;
		  }
		else {	ii = urf.obs_ii;	var = urf.obs_sigma;
		  }
		xh.x_inten = iscale * ii;
		xh.x_sigma = iscale * ((var < FUZZ) ? 0.01 : sqrt(var));
		if (onebatch) xh.x_batch = batchbase;
		else xh.x_batch = batchbase * 200 + 2 +
			(urf.obs_omob - minomega) / 20.;
		printshel(fpo, &xh);
	   }
	(void)fclose(fpi);	(void)fclose(fpo);	fpi = fpo = NULL;
	printf("%6d reflections processed\n", nrefls);
	globalkill();	return EOK;
}

/* end of utoshel functionality */

/**********************************
	pick_uref:	print results and sometimes profiles
			of selected UREFLS refls
	This is a special-purpose printing function.  It reads the UREFLS
	file and an ascii list of reflections.  It prints out reflections
	from the UREFLS file which either appear on the list or which have
	symmetry mates on the list.
	If the "-i" (for "image") option is used, it dumps a 7x15x15 box
	around each selected reflection from the raw data.
	If the "-s" (for "subtract-dark") option is used, the 7x15x15 box
	dumps are done on dark-current-subtracted images.
*************************************************************************/

static void sum_pickrun(FILE *fp, PUWORK *pup)
{ /*	This summarizes the runtime characteristics of a run of `pickuref'
	to the stream fp */
	static char *fnames[] = {"urefls", "uparams", "reflist", "xlog",
		"frames", "bgimage", NULL };
	static char ftypes[] = { 1, 1, 1, 2, 1, 1, 0 };

	ftypes[4] = (pup->pu_flags & PU_IMAGES) ? 1 : 0;
	ftypes[5] = (pup->pu_flags & PU_SUBBG)  ? 1 : 0;
	header(fp, PROCESSING, fnames, ftypes);
	fprintf(fp , " Program will search for %4d reflections\n",
		pup->pu_nlines);
	if (pup->pu_flags & PU_IMAGES)
	  {	fprintf(fp,
	 "Each reflection will be displayed as a 15x15x7 image dump\n");
		if (pup->pu_flags & PU_SUBBG) fprintf(fp,
	 "\tafter subtraction of background from %s\n", pup->pu_bgname);
	  }
	else fprintf(fp,
	 "A one-line summary of each reflection will be output\n");
	fprintf(fp, "%s intensity and sigma values will be output\n",
	 (pup->pu_flags & PU_USE_ABG) ? "Transformed-coordinate" :
	 "(X,Y,Omega)-Coordinate");
}

static int urptolim(OBSREF *urp, int xyzl[], PUWORK *pup)
{ /*	This looks at the reflection pointed to by `urp' and constructs
	the limits over which the array of display values will be built. */
	int	zcen, xcnn, ycnn;

	zcen = urp->obs_ompred + 0.4999;
	xcnn = ((double)urp->obs_xpred) / pup->pu_scl[0] + 0.4999;
	ycnn = ((double)urp->obs_ypred) / pup->pu_scl[1] + 0.4999;
	return centolim(xcnn, ycnn, zcen, xyzl);
}

static void pickend(PUWORK *pup)
{	/* gets rid of open files and allocated memory for `pickuref' */
	
	if (NULL != pup->pu_fpi) (void)fclose(pup->pu_fpi);
	pup->pu_fpi = NULL;
	if ((stdout != pup->pu_fpo) && (NULL != pup->pu_fpo))
		(void)fclose(pup->pu_fpo);
	pup->pu_fpo = NULL;	(void)ftables(0);
	memzap((AVOID **)&(pup->pu_checks), "checks in pickuref");
	memzap((AVOID **)&(pup->pu_cp0), "cp0 in pickuref");	globalkill();
}

int pickumain(int argc, char *argv[])
{ /*	This is the mainline for "pickuref". */
	int		dodark, imflags, locht, locwid, nbr, nfound, xyzli[6];
	short		iline, uhkl[3];
	u_char		**chp, phkl[4], *cp;
	char		line[96], **flags, **dcp;
	FILE		*fpil;
	MAIL		*bgmp;
	REFPARS		rpars;
	OBSREF		urf;
	static MAIL	bgh;
	static PUWORK	pus;
	static FRAMEINDEX	*lfmi;
	static int	idarkfrac = 6400;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	pus.pu_flags = dodark = idarkfrac = 0;
	flags = flagproc(argc, argv, "BEIS");
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	   switch (**dcp)
	     {	case	'B':	pus.pu_flags |= PU_SUBBG;	break;
		case	'E':	pus.pu_flags |= PU_USE_ABG;	break;
		case	'I':	pus.pu_flags |= PU_IMAGES;	break;
		case	'S':	dodark = DARK_CURRENT;
			if ('\0' != *((*dcp) + 1))
				idarkfrac = 8000. * atof((*dcp) + 1);
			break;
	     }
	/* backgrounds & dark-current are relevant only with image displays*/
	if (!(pus.pu_flags & PU_IMAGES))
	  {	pus.pu_flags &= ~PU_SUBBG;	dodark = 0;
	  }
	if (idarkfrac > 32000) idarkfrac = 32000;
	if (dodark == DARK_CURRENT) dodark |= (idarkfrac << 16);
	if (qerror((AVOID *)(fpil = xgopen("reflist", "r")),
		"open input reflection list")) return EOPENERR;
	if (qerror((AVOID *)(pus.pu_fpi = xgopen("urefls", "r")),
		"open input observed-reflections list")) return EOPENERR;
	find_dimension(&locwid, &locht, 0, (char *)NULL);
	if (!locwid || !locht) {	locwid = locht = 512; }
	pus.pu_scl[0] = dimscale((short)locwid);
	pus.pu_scl[1] = dimscale((short)locht);
	if (EOK != (nfound = refparin(&rpars, &nbr))) return nfound;
	pus.pu_scl[2] = rpars.xlambda * 100.;
	pus.pu_scl[3] = (pus.pu_flags & PU_USE_ABG) ? 1. : 0.;
	if ((rpars.xispgp < 200) && (rpars.xisys == 7)) rpars.xispgp += 100;
	for (pus.pu_nlines = 0; NULL != fgets(line, 80, fpil); )
		pus.pu_nlines++;
	(void)fseek(fpil, 0L, 0);	/* rewind input reflection list */
	if (NULL == (AVOID *)(pus.pu_checks = (u_char **)callop(
	 (unsigned)pus.pu_nlines, sizeof (u_char *), "checks in pickuref")))
		return EALLOC;
	if (NULL == (AVOID *)(pus.pu_cp0 =(u_char *)callop(4 * pus.pu_nlines,
	 sizeof (u_char), "cp0 in pickuref"))) return EALLOC;
	if (NULL == (pus.pu_fpo = propen("xlog"))) pus.pu_fpo = stdout;
	cp = pus.pu_cp0;	chp = pus.pu_checks;
	imflags = WAIT_IF_ABSENT;
	for (iline = 0; iline < pus.pu_nlines; iline++, chp++, cp += 4)
	   {	*chp = cp;
		if (NULL == fgets(line, 80, fpil))
		  {	fprintf(stderr,
 "Error reading line %d of list file: remaining reflections ignored\n", iline);
			break;
		  }
		(void)sscanf(line, "%hd %hd %hd", &uhkl[0], &uhkl[1],&uhkl[2]);
		mpack(uhkl, *chp, rpars.xispgp);
		symtr(*chp, rpars.xispgp);
		(*chp)[3] &= 0376;	/* turn off bijvoet bit */
	   }
	pus.pu_nlines = iline;	/* in case we lost some */
	(void)fclose(fpil);	fpil = NULL;
	if (pus.pu_flags & PU_SUBBG)
	  {	strcpy(pus.pu_bgname, trname("BGIMAGE"));
		if (FRAME_OK != (read_image(pus.pu_bgname, imflags, &bgh)))
		  {	fprintf(stderr,
 "Warning: background image couldn't be read; no bgnd subtraction done\n");
			pus.pu_flags &= ~PU_SUBBG;
		  }
	  }
	bgmp = (pus.pu_flags & PU_SUBBG) ? &bgh : NULL;
	sum_pickrun(stdout, &pus);	sum_pickrun(pus.pu_fpo, &pus);
	lfmi = NULL;	/* so the compiler doesn't whine */
	if (pus.pu_flags & PU_IMAGES)
	  {	if (NULL == (lfmi = ftables(3)))
		  {	fprintf(stderr,
 "Warning: index of images unavailable: no image dumps performed\n");
			pus.pu_flags &= ~PU_IMAGES;
		  }
		else {	imflags = WAIT_IF_ABSENT | dodark;
			fputc('\f', pus.pu_fpo);
		     }
	  }
	for (nfound = 0; 0 < xgread((void *)&urf, XG_OBSREF, pus.pu_fpi); )
	   {	mpack(urf.obs_hkl, phkl, rpars.xispgp);
		symtr(phkl, rpars.xispgp);
		phkl[3] &= 0376;
		for (iline = 0, chp = pus.pu_checks; iline < pus.pu_nlines;
		 iline++, chp++)
			if (!hklcmp(*chp, phkl))
			  {	prurefl(pus.pu_fpo, &urf, pus.pu_scl); nfound++;
				xeflush(pus.pu_fpo);	/* temporary */
				if ((NULL != lfmi) &&
				 (pus.pu_flags & PU_IMAGES))
				  {	if (EOK != urptolim(&urf, xyzli, &pus))
						break;
					if (EOK == get_3dprof(lfmi,
					 imflags, xyzli, bgmp))
					  {	disp_3dprof(pus.pu_fpo, xyzli);
						printf(
	 "Reflection [%3d,%3d,%3d] found near frame %3.0f\n",
	 urf.obs_hkl[0], urf.obs_hkl[1], urf.obs_hkl[2], urf.obs_omob);
					  }
				  }
				break;
			  }
	   }
	if (pus.pu_flags & PU_IMAGES)
		(void)get_3dprof(lfmi, imflags, NULL, NULL);
	(void)fclose(pus.pu_fpo);	pus.pu_fpo = NULL;
	(void)ftables(0);	printf("%d reflections found\n", nfound);
	pickend(&pus);	return EOK;
}

/* end of pick_uref functionality */

/***********************************
	x2urf: convert a set of DENZO  ".x" files to a UREFLS file
	The program reads a set of  ".x" files listed in a file with
	environment variable XINDEX; it writes to the env. vbl. UREFLS.
	The input format for this activity contains the following,
	according to Gewirth's manual:
	%4d%4d%4d	h, k, l
	%2d		flag (1==partial 0=full)
	%f		Profile-fitted intensity
	%f		Summation intensity
	%f		chi-squared of profile fitting
	%f		sigma(I)
	%f		cos(incidence angle at detector)
	%f%f		X,Y coords of predicted centroid in pixels
	%f		Lorenz, polarization, obliquity factor
	%f		strength of averaged profile, arbitrary units.
 Syntax:
	x2urf
 Flags: none so far.
***********************************/

static FRAMEINDEX *make_xft(void)
{ /*	This establishes the table of .x files and their index numbers
	by reading the file with name or logical name XFRAMES
	and building up an array of `FRAMEENTRY' structures.
	The routine returns a pointer to the array of structures
	and sticks the number of entries in the array in the structure
	element lfmi->fi_totfms. This routine returns NULL upon errors. */
	register int		i;
	int			ilast;
	unsigned		nfms;
	FILE			*findx;
	FRAMEENTRY		*lft, *lft0;
	char			line[120];
	static FRAMEINDEX	*lfmi;

	/* open the existing XFRAMES file */
	if (NULL == (findx = xgopen("xframes", "r"))) return NULL;
	/* now count the number of frames in the file */
	for (nfms = 0; NULL != fgets(line, 80, findx); ) nfms++;
	if (verror(1 + fseek(findx, 0L, 0), "rewind index file")) return NULL;
	/* reserve memory for the frame table */
	if (qerror((AVOID *)(lft0 = (FRAMEENTRY *)callop(nfms + 2,
	 sizeof (FRAMEENTRY), "lft0 in make_xft")),
		"reserve memory for frame table")) return NULL;
	/* construct the frame table from the diskfile */
	for (i = nfms, lft = lft0; i--; lft++)
	   {	if (NULL == (AVOID *)fgets(line, 80, findx))
		  {	nfms -= (i + 1);	break;
		  }
		(void)sscanf(line, "%hd%s", &(lft->fmno), lft->fmname);
	   }
	(void)fclose(findx);
	if (NULL == (lfmi = (FRAMEINDEX *)callop(1, sizeof (FRAMEINDEX),
	 "lfmi in make_xft"))) return NULL;
	ilast = nfms - 1;			lfmi->fi_firstfm = lft0->fmno;
	lfmi->fi_lastfm = (lft0 + ilast)->fmno;	lfmi->fi_totfms = nfms;
	lfmi->fi_fmi = lft0;	lfmi->fi_fme = lft - 1;
	return lfmi;
}

static short hkltossq(REFPARS *xp, short hklv[])
{ /*	Returns the sin squared theta over lambda squared value for (hkl) */
	int	i, j;
	double	se, st;

	for (st = i = 0; i < 3; i++)
	   {	for (se = j = 0; j < 3; j++)
			se += xp->xrec[i][j] * (double)(hklv[j]);
		st += se * se;
	   }
	st *= xp->xlambda * xp->xlambda;
	if (st < DFUZZ) st = 0.;
	return (short)(st * 1.0e4);
}

static int xlproc(REFPARS *xp, BOUNDARY *bdp, double scl[],
	double framer, XLINE *xrp, OBSREF *rfq)
{ /*	This converts the XLINE structure into an OBSREF structure.
	It installs all the intensity in the center image for now. */
	int	i;
	short	thisx, thisy;
	double	rawbg, rawi;

	if (xrp->xf_sigma < DFUZZ) return 0;
	for (i = 0; i < 3; i++) rfq->obs_hkl[i] = xrp->xf_hkl[i];
	rfq->obs_lpa = (short)(xrp->xf_lpo * 4.0e3 + 0.49999);
	rfq->obs_ii = xrp->xf_ifit;
	rfq->obs_iunfit = xrp->xf_isum;
	rfq->obs_sigma = rfq->obs_sunfit = xrp->xf_sigma * xrp->xf_sigma;
	if (rfq->obs_sigma < 1.) rfq->obs_sigma = 1.;
	if (rfq->obs_sunfit < 1.) rfq->obs_sunfit = 1.;
	rfq->obs_ssq = hkltossq(xp, rfq->obs_hkl);
	thisx = xrp->xf_xpred + 0.49999;
	thisy = xrp->xf_ypred + 0.49999;
	rfq->obs_xpred = (short)(xrp->xf_xpred * scl[0] + 0.49999);
	rfq->obs_ypred = (short)(xrp->xf_ypred * scl[1] + 0.49999);
	rfq->obs_region = getregion(thisx, thisy, /* framer, */ bdp);
	rfq->obs_ompred = rfq->obs_omob = framer;
	rawi = (rfq->obs_ii > 1) ? rfq->obs_ii : 1;
	rfq->obs_gof = (short)(fabs(xrp->xf_chisq * rawi) + 0.49999);
	rfq->obs_xob = rfq->obs_xpred;
	rfq->obs_yob = rfq->obs_ypred;
	/* The following 4 params are really used for storing the differences
	 that yield the Kabsch intensity and sigma values.  We therefore want
	 them to be zero, since that's not really meaningful in a .x file */
	rfq->obs_shift = rfq->obs_widob = rfq->obs_bgunfit =
		rfq->obs_reflno = 0;
	rfq->obs_gama = 1.;	/* close enough */
	/* we make an arbitrary assumption that the reflection
	 covers thirty pixels, and we attempt to discern the background
	 from the difference between sigma^2 and I. */
	rawbg = rfq->obs_sigma - rfq->obs_ii;
	if (rawbg < FLOOR_BG) rawbg = FLOOR_BG;
	else if (rawbg > MAXBG) rawbg = MAXBG;
	rfq->obs_bgnd = rawbg / 30.;
	rfq->obs_ncells = 30;
	rfq->obs_lowfrm = rfq->obs_hifrm = 3;
	rfq->obs_chamno = rfq->obs_runno = 0;
	for (i = 0; i < PFRAMES; i++) rfq->obs_frmint[i] = 0;
	rfq->obs_frmint[3] = rfq->obs_ii;
	if ((rfq->obs_ii * (rfq->obs_ii / rfq->obs_sigma)) < BRIGHTIOS2)
		rfq->obs_flags = 32;
	else rfq->obs_flags = 0;
	return 1;
}

static int matchurf(OBSREF *rti, OBSREF *rto, double maxdif[])
{ /*	This detects whether rti and rto are really the same reflection */
	int	 i;

	for (i = 0; i < 3; i++)
		if (rti->obs_hkl[i] != rto->obs_hkl[i]) return 0;
	if (abs(rti->obs_xpred - rto->obs_xpred) > maxdif[0]) return 0;
	if (abs(rti->obs_ypred - rto->obs_ypred) > maxdif[1]) return 0;
	return 1;
}

static void mergeurf(OBSREF *rti, OBSREF *rto)
{ /*	This adds the contents of rto to rti */
	int	j, nalready;
	double	den, new, numx, numy, old;

	rto->obs_chamno = ALREADYCODE;
	rti->obs_sigma += rto->obs_sigma;
	rti->obs_iunfit += rto->obs_iunfit;
	rti->obs_sunfit += rto->obs_sunfit;
	old = rti->obs_ii;
	if (!(rti->obs_lowfrm)) /* already at zero: just keep the strong ones*/
	  {	if (old <= rti->obs_frmint[0]) return;
		new = rto->obs_ii - rti->obs_frmint[0];
		rti->obs_ii += rto->obs_ii - rti->obs_frmint[0];
		/* shift down one if new one is stronger than 0 */
		for (j = 0; j < 6; j++)
		   rti->obs_frmint[j] = rti->obs_frmint[j + 1];
		rti->obs_frmint[6] = rto->obs_ii;
		rti->obs_omob += 1;
	  }
	else {	nalready = rti->obs_hifrm + 1 - rti->obs_lowfrm;
		if (nalready % 2)	/* odd number of images: add right */
			(rti->obs_hifrm)++;
		else	/* even number: slide the list down */
		  {	(rti->obs_lowfrm)--;
			for (j = rti->obs_lowfrm; j < rti->obs_hifrm; j++)
				rti->obs_frmint[j] = rti->obs_frmint[j + 1];
		  }
		new = rto->obs_ii;
		rti->obs_frmint[(int)(rti->obs_hifrm)] = new;
		rti->obs_gof += rto->obs_gof;
		rti->obs_bgnd += rto->obs_bgnd;
		rti->obs_ncells += rto->obs_ncells;
		if (nalready < 3) rti->obs_lpa = rto->obs_lpa;
	     }
	den = old + new;
	if (new > 0)
	  {	if (old > 0)	/* update <X,Y> centroid */
		  {	
			numx = old * (double)(rti->obs_xpred) +
				new * (double)(rto->obs_xpred);
			numy = old * (double)(rti->obs_ypred) +
				new * (double)(rto->obs_ypred);
			rti->obs_xpred = numx / den;
			rti->obs_ypred = numy / den;
		  }
		else {	rti->obs_xpred = rto->obs_xpred;
			rti->obs_ypred = rto->obs_ypred;
		     }
	  }
	rti->obs_ii = den;
	rti->obs_xob = rti->obs_xpred;	rti->obs_yob = rti->obs_ypred;
	if ((rti->obs_ii * (rti->obs_ii / rti->obs_sigma)) < BRIGHTIOS2)
		rti->obs_flags = 32;
	else rti->obs_flags = 0;
}

static void fixomega(OBSREF *rtq)
{ /*	This cleans up the omega values */
	int		iz, nfms;
	double		den, dz, num, this, uz;
	static int	fakecells[] = { 45, 80, 110, 135, 155, 170, 185, 195 } ;

	nfms = rtq->obs_hifrm - rtq->obs_lowfrm;
	if (nfms > 1)
	  {	for (num = den = uz = 0, iz = rtq->obs_lowfrm;
		 iz <= rtq->obs_hifrm; iz++, uz++)
		   {	dz = (double)uz;	this = rtq->obs_frmint[iz];
			if (this < 0) this = 0;
			num += dz * this;	den += this;
		   }
		if (den > 0.)
		  {	num /= den;	rtq->obs_omob += num;
		  }
	  }
	if (nfms > 6) nfms = 6;	if (nfms < 0) nfms = 0;
	rtq->obs_ncells = fakecells[nfms];	rtq->obs_ompred = rtq->obs_omob;
}

static int carefuli(char *src, int wid)	/* wrappers for sscanf */
{
	char	cfb[20];
	int	res;

	zeromem(cfb, wid + 2); sscanf(src, "%d", &res); return res;
}

static double carefulf(char *src, int wid)
{
	char	cfb[20];
	double	fres;

	zeromem(cfb, wid + 2); sscanf(src, "%lf", &fres); return fres;
}

static int parsex(char *xli, XLINE *xq)
{ /*	Parse a .x reflection line */
	int	i, nitems;
	
	/* first we do it the simple way */
	nitems = sscanf(xli," %d %d %d %d %lf %lf %lf %lf %lf %lf %lf %lf %lf",
		&(xq->xf_hkl[0]), &(xq->xf_hkl[1]), &(xq->xf_hkl[2]),
		&(xq->xf_partial), &(xq->xf_ifit), &(xq->xf_isum),
		&(xq->xf_chisq), &(xq->xf_sigma), &(xq->xf_cosinc),
		&(xq->xf_xpred), &(xq->xf_ypred), &(xq->xf_lpo),
		&(xq->xf_pstrength));
	/* now: more carefully */
	for (i = 0; i < 3; i++) xq->xf_hkl[i] = carefuli(&(xli[4*i]), 4);
	xq->xf_partial = carefuli(&(xli[12]), 2);
	xq->xf_ifit = carefulf(&(xli[14]), 8);
	xq->xf_isum = carefulf(&(xli[22]), 8);
	xq->xf_chisq = carefulf(&(xli[30]), 7);
	xq->xf_sigma = carefulf(&(xli[37]), 6);
	xq->xf_cosinc = carefulf(&(xli[43]), 6);
	xq->xf_xpred = carefulf(&(xli[49]), 7);
	xq->xf_ypred = carefulf(&(xli[56]), 7);
	xq->xf_lpo = carefulf(&(xli[63]), 6);
	xq->xf_pstrength = carefulf(&(xli[69]), 8);	return nitems;
}

int x2urfmain(int argc, char *argv[])
{ /*	This is the mainline for "x2urf".  It reads in the reflections once
	to find the minimum omega and maximum intensity; then it goes
	through for real for processing. */
	char		zt, *cp, **dcp, **flags, dotxheader[256], xline[514];
	int		beamx, i, lineno, locht, locwid, nbr, nitems;
	int		nrefls, proper, reflinfm, xlen;
	double		distpix, farg, framer, frang, lfmmax, mosaicity;
	double		oscend, oscstart, wavelen, scl[4], xtalrot[3];
	double		ina[3], inu[3], maxdif[2];
	FILE		*fpo, *fxp;
	REFPARS		*xp;
	OBSREF		*rte, *rtp, *rtq, *rtz;
	BOUNDARY	*bdp;
	FRAMEENTRY	*xftp;
	FRAMEINDEX	*xfta;
	static XLINE	xr;
	static char	*fnames[] = { "XFRAMES", "UPARAMS", "UREFLS", NULL } ;
	static char	ftypes[] = { 1, 1, 2, 0 } ;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	/* open the index of .x files */
	if (NULL == (xfta = make_xft())) return EOPENERR;
	if (NULL == (xp = (REFPARS *)callop(1, sizeof (REFPARS),
	 "xp in x2urf"))) return EALLOC;
	if (EOK != (i = refparin(xp, &nbr)))
	  {	fprintf(stderr, "Error %d in reading parameters\n", i);
		return i;
	  }
	frang = MAXBUNDLE;
	flags = flagproc(argc, argv, "W");
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	   {	farg = atof((*dcp) + 1);
		if (**dcp == 'W') frang = farg;
	   }
	/* specifying the spacegroup supersedes other mechanisms */
	header(stdout, REDUCTION, fnames, ftypes);
	/* determine the detector dimensions and scale factors from whatever
	 partial or complete X-GEN processing results are available.
	 As of 4 April 05 we require this step to succeed */
	if (EOK != (i = find_dimension(&locwid, &locht, 0, NULL)))
		return i;
	printf("detector dimensions: %4d by %4d pixels\n", locwid, locht);
	scl[0] = dimscale((short)locwid); scl[1] = dimscale((short)locht);
	maxdif[0] = 20 * scl[0];	maxdif[1] = 20 * scl[1];
	beamx = locwid / 2;
	bdp = make_boundaries(beamx, locwid, locht);
	/* make a quick pass through the .x files to estimate
	 how many observations are present */
	for (nrefls = 0, xftp = xfta->fi_fmi; (NULL != xftp) &&
	 (0 <= xftp->fmno) && ('\0' != *(xftp->fmname)); xftp++)
	   {	if (NULL == (fxp = xgopen(xftp->fmname, "r"))) break;
		reflinfm = 0; zeromem(xline, 256);
		for (lineno = 0; (NULL != fgets(xline, 256, fxp)); lineno++)
		   if (lineno > 4)
		     {  for (cp = xline; '\0' != *cp; cp++)
			   if ((' ' != *cp) && ('\t' != *cp)) break;
			if ('\0' != *cp) /* first non-blank */
			  {	zt = *cp; /* mustn't be alphabetic */
				if (('a' <= *cp) && (*cp <='z')) break;
				if (('A' <= *cp) && (*cp <='A')) break;
			  }
			nrefls++;	/* increment count of refls */
		     }
		(void)fclose(fxp);	/* so we can do another one */
	   }
	printf(" %5d .x files contain a total of %9d observations\n",
	 (int)(xftp - xfta->fi_fmi), nrefls);
	if (NULL == (rtp = (OBSREF *)callop(nrefls, sizeof (OBSREF),
	 "rtp in x2urf")))
	  {	fprintf(stderr, "Unable to allocate memory for OBSREFs\n");
		return EALLOC;
	  }
	/* go through the list of .x files */
	rtq = rtp;	framer = 1.;
	for (proper = nrefls = 0, xftp = xfta->fi_fmi; (NULL != xftp) &&
	 (0 <= xftp->fmno) && ('\0' != *(xftp->fmname)); xftp++)
	   {	/* open the .x file and extract the header information */
		if (NULL == (fxp = xgopen(xftp->fmname, "r"))) break;
		reflinfm = 0;	zeromem(xline, 256);
		for (lineno = 0; (NULL != fgets(xline, 256, fxp)); lineno++)
		     { /* initial line contains a header; the next 3
			 contain the reciprocal-space matrix. The one after
			 that has some other useful info. */
			if (!lineno)
			  {	xlen = strlen(xline);
				if (xlen > 254) xlen = 254;
				dotxheader[254] = '\0';
				for (cp = xline + (xlen - 1); cp >= xline;
				 cp--) if (*cp == '\n') *cp = '\0';
				strcpy(dotxheader, xline);
			  }
			else if (lineno <= 3)
			  {	nitems = sscanf(xline,
				 "%lf %lf %lf %lf %lf %lf",
				 &ina[0], &ina[1], &ina[2],
				 &inu[0], &inu[1], &inu[2]);
				if (nitems >= 3)
				  for (i = 0; i < 3; i++)
					xp->xrec[lineno - 1][i] = ina[i];
				else for (i = 0; i < 3; i++)
					xp->xrec[lineno - 1][i] = 0.;
			  }
			else if (lineno == 4)
			  {	nitems = sscanf(xline,
				 "%lf %lf %lf %lf %lf %lf %lf %lf",
				  &oscstart, &oscend, &distpix, &wavelen,
				  &(xtalrot[0]), &(xtalrot[1]),
				  &(xtalrot[1]), &mosaicity);
				if (nitems < 8) mosaicity = 1.;
				framer = (double)(xftp->fmno);
			  }
			else {	for (cp = xline; '\0' != *cp; cp++)
				   if ((' ' != *cp) && ('\t' != *cp)) break;
				if ('\0' != *cp) /* first non-blank */
				  {	zt = *cp; /* mustn't be alphabetic */
					if (('a' <= *cp) && (*cp <='z')) break;
					if (('A' <= *cp) && (*cp <='A')) break;
				  }
				/* these values sometimes run together ... */
				nitems = parsex(xline, &xr);
				if (xr.xf_hkl[0] == 999) break;
				if (nitems >= 12)
				  {	if (xlproc(xp, bdp, scl, framer,
					 &xr, rtq))
					  {	nrefls++;	reflinfm++;
						rtq++;
					  }
				  } /* end of "if (nitems >= 12)" clause */
			      } /* end of "else" */
		   } /* end of loop through lines in this particular .x file */
		(void)fclose(fxp);
		printf("frame %4d: %6d reflections; header = %s\r",
			xftp->fmno, reflinfm, dotxheader);
		xeflush(stdout);
	   } /* end of loop through list of .x files */
	rte = rtq;	/* remember where the end of the list is */
	/* march back through the list */
	printf("\n");
	for (rtq = rtp; rtq < rte; rtq++)
	   {	lfmmax = rtq->obs_ompred + frang;
		if (rtq->obs_chamno != ALREADYCODE)
		  {	for (rtz = rtq + 1; rtz < rte; rtz++)
			   {	if (rtz->obs_ompred >= lfmmax) break;
				if ((rtz->obs_chamno != ALREADYCODE) &&
				 matchurf(rtq, rtz, maxdif)) mergeurf(rtq,rtz);
			   }
			if (!((rtq - rtp) % 500))
			  {	printf(
 "Processing [%4d,%4d,%4d]@[%7.1f,%7.1f,%7.2f],I=%8.1f+/-%8.2f\r",
 				 rtq->obs_hkl[0], rtq->obs_hkl[1],
 				 rtq->obs_hkl[2],
				 ((double)rtq->obs_xpred) / scl[0],
				 ((double)rtq->obs_ypred) / scl[1],
				 rtq->obs_ompred, rtq->obs_ii,
				 sqrt(rtq->obs_sigma));
			  	xeflush(stdout);
			  }
		  }
	   }
	printf("\n");
	/* open the output UREFLS file */
	if (qerror((AVOID *)(fpo = xgopen("UREFLS", "w")),
	 "open output UREFLS file")) return EOPENERR;
	for (proper = 0, rtq = rtp; rtq < rte; rtq++)
	   {   if (rtq->obs_chamno != ALREADYCODE)
		     {	fixomega(rtq);
			if (sizeof (OBSREF) != (nitems = xgwrite(rtq,
			 XG_OBSREF, fpo)))
			  {	fprintf(stderr,
	 "Warning: only %3d bytes out of %3d written for reflection %5d\n",
				  nitems, (int)(sizeof (OBSREF)), proper);
				return EWRITERR;
			  }
			proper++;
		     }
		if (!((rtq - rtp) % 5000))
		  {	printf(
		 "Outputting DENZO reflection %9d; hkl = [%4d,%4d,%4d]\r",
		  (int)(rtq - rtp),
		  rtq->obs_hkl[0], rtq->obs_hkl[1],rtq->obs_hkl[2]);
			xeflush(stdout);
		  }
	   }
	printf("\n");
	(void)fclose(fpo);	fpo = NULL;
	printf("%6d reflections processed: %6d valid, %6d invalid\n",
		nrefls, proper, nrefls - proper);
	(void)ftables(0);	memzap((AVOID **)&xp, "xp in x2urf");
	globalkill();	return EOK;
}

/* end of x2urf.c */
