/* x_dump.c -- General ascii dump facility for XGEN binary files

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions; reorganize use of "{"
	26 Jun 2006: include xgenpred,xgmodel,xgpostint for completeness
	11 Nov 2005: replace newdump(drp) with dumpimage(drp, 1) and
		dumpimage(drp) with dumpimage(drp, 0)
	12 Oct 2005: eliminate static globals; more elements in DUMPRUN
	17 May 2005: prompt changed from ">" to "xd>"
	11 May 2005: check return status from xgjournal
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	 3 Oct 2004: require user to specify entire (X,Y) center in image dumps
	28 Aug 2004: routines used only here are declared as static; cosmetics
	16 Aug 2004: declare dimscale() appropriately
	 2 Aug 2004: add new arg to both calls to find_dimension()
	27 Jul 2004: use dimscale() to get the (X,Y) scales
	18 Feb 2004: change kill_ftab calls so that they don't have args
	 1 Feb 2004: don't add 100 to the spacegroup for rhombohedrals
		if it's already been done
	26 Jan 2004: some static globals replaced with local variables
	21 Jan 2004: license message inserted and initial comment reordered
	 8 Jan 2004: replace extern global variables with static globals;
		callop() & memzap() replace shcalloc() & shfree();
		move newdump() and dump_image() to ../xglib/ascout.c;
		kill_ftab and kill_mail() called at the end
	24 Nov 2003: create cenhead0[] and cenhead1[]
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Sep 2003: use xgaccess() rather than access() to figure
		out whether files are present; that allows for gunzipping
	17 Aug 2003: trivial renaming of local variable finame
	 1 May 2003: use different MAIL structure for each image
		in newdump() and query_image()
	 5 Jan 2003: reorganized source-file header
	15 Nov 2002: replace startfm and endfm with lstartf and lendf
		so they don't collide with global variables
	 1 Nov 2002: use of REGIONPROFILE structures for reflection-box sizes;
		enables more general formats to be handled
	25 Mar 2002: removed MSI attachments
	24 Aug 1995: 16-bit Mar overflows handled correctly
		image data read as ints, not u_shorts
	 6 Jul 1995: new facilities for 2-D dumps to facilitate
		calling 'em from outside
	31 May 1995: "register" -> "register int" throughout
	12 May 1995: make_ftab takes no argument; totalframes is a global
	 2 Mar 1995: diagnostics to stdout
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	27 Oct 1994: provided for change in prurefl call
	 5 Oct 1994: use xgread to input various binary data types
	23 Sep 1994: fix plane-spacing calculation
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: a few variables moved around
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version

   Purpose: ascii dumps of the binary files obtained in XGEN.
   Cross references: ../xglib/ascout.c

 This is a general purpose printing program for the files generated by and
 manipulated in the XGEN crystallographic data reduction package.  It may be
 invoked with no arguments, in which case it enters an interactive mode,
 or with one or two arguments.

 If it is invoked with one argument, the argument is assumed to be
 a logical name or environment variable indicative of the file type
 the user wishes to dump; the program looks for the file corresponding
 to that logical name or environment variable and prints it.
 If it is invoked with two arguments, the first is taken as the logical
 name or environment variable indicative of the file type, and the
 second is taken to be the actual filename to be dumped.  Thus if
 we invoke the program with
 	xdump uparams fred.upr
 we will get a dump of the file "fred.upr" as a "uparams"-style file,
 even if the variable "uparams" corresponds to some other file.

 If it is invoked with no arguments, the program prompts for a file
 type.  If the user types just the file type, the appropriate logical
 name or environment variable is searched, opened, and dumped; if
 the user enters the file type and a specific filename, the specific
 file is dumped in the format appropriate to the type.

 The program dumps to the logical name or environment variable
 "udump", and it dumps the first few lines to the terminal.
 The dump to "udump" requires > 80 character width for some file types.
*************************************************************************/
#define		NEED_TIME	1
#define		NEED_ERRNO	1
#include	"xgenpred.h"
#include	"xgmodel.h"
#include	"xgpostint.h"
#define		NTYPES		13		/* number of file types */
#define		ISSTOP		(NTYPES - 1)	/* code for "stop" command */
#define		ISMAXLINES	(NTYPES - 2)	/* code for "nlines" command*/
#define		UNKNOWN		(-1)
#define		DEF_MAXLINES	10	/* default # lines to terminal*/

extern int xdumpmain(int argc, char *argv[]);

static void query_image(char *name, char *fty[], DUMPRUN *durp)
{ /*	This prompts for the frame range and the (X,Y) center
	of the area on an image to be dumped.  It provides the user
	interface to the facility for dumping raw images. */

	register int	frame, nvalues;
	char		frname[80], xdline[81];
	int		lendf, lstartf, result, xcn, ycn;
	ino_t		newinode;
	static int	duwidth = 17;
	static int	duheight = 17;
	MAIL		*thishd;

	durp->dui_imflags = GET_RUNHEADER | GRIPE_IF_SHORT;
	for (nvalues = 0; nvalues < 2; )
	     {
		printf(" Enter x and y for center of area to be dumped: ");
		while (NULL == fgets(xdline, 80, stdin)) ;
		nvalues = sscanf(xdline, "%d%d", &xcn, &ycn);
		if (nvalues < 2) printf("requested 2 values, got %1d\n",
			nvalues);
	    }
	durp->dui_nmbox = PFRAMES + 1;
	if ((NULL == (durp->dui_heads)) && (NULL == (AVOID *)(durp->dui_heads =
	 (MAIL *)callop(durp->dui_nmbox, sizeof (MAIL),
	 "heads in query_image"))))
	  {
		fprintf(stderr, "Can't reserve memory for image headers\n");
		return;
	  }
	durp->dui_xce = xcn;		durp->dui_yce = ycn;
	durp->dui_wid = duwidth;	durp->dui_ht = duheight;
	if (name && (*name) && (0 == xgaccess(name, 4, 1)))
	  { /* if a name was specified on the command line, we first look
		to see if it has an image number attached to it.
		If it does, we read it into that header. If it's not,
		we don't _need_ to read it in, because it's already there */
		if (NULL == durp->dui_fmi) durp->dui_fmi = ftables(3);
		if (0 > (frame = match_inode(&(durp->dui_fmi), name)))
			frame = 0;
		thishd = durp->dui_heads + (frame % durp->dui_nmbox);
		newinode = record_inode(frname);
		if (newinode != thishd->quadrant)
		  {
			if (FRAME_OK !=
			(result = read_image(name, durp->dui_imflags, thishd)))
			  {
				(void)image_error(stderr, name, result, 0);
				return;
			  }
		  }
		durp->dui_imname = name;	dumpimage(durp, 0);	return;
	  }
	printf(" Enter a filename or the starting and ending frame numbers: ");
	while (NULL == fgets(xdline, 80, stdin)) ;
	lstartf = lendf = -1;
	nvalues = sscanf(xdline, "%d %d", &lstartf, &lendf);
	if ((lendf >= 0) && (nvalues >= 2))
	  {
		if (lstartf < 0) query_image((char *)NULL, fty, durp);
		if (NULL == (durp->dui_fmi = ftables(3)))
		  {
			fprintf(stderr, "Cannot create image table\n");
			return;
		  }
		if ((durp->dui_fmi)->fi_firstfm > lstartf)
			lstartf = (durp->dui_fmi)->fi_firstfm;
		if ((durp->dui_fmi)->fi_lastfm  < lendf)
			lendf = (durp->dui_fmi)->fi_lastfm;
		if (lendf < lstartf) lendf = lstartf;
		printf( "Frames %3d to %3d will be dumped\n", lstartf, lendf);
		for (frame = lstartf; frame <= lendf; frame++)
		   {
			thishd = durp->dui_heads + (frame % durp->dui_nmbox);
			if (NULL == (durp->dui_imname = indexname(frame)))
			  {
				fprintf(stderr,
				 "Cannot find requested frame file");
				return;
			  }
			/* see if this image's information is already
			 present in that header */
			newinode = record_inode(durp->dui_imname);
			/* if it is, we don't need to read the image */
			if (thishd->quadrant != newinode)
			  {
				if (FRAME_OK != (result = read_image(
				 durp->dui_imname, durp->dui_imflags, thishd)))
				  {
					(void)image_error(stderr,
					 durp->dui_imname, result, 0);
					return;
				  }
			  } /* end of "if (thishd->quadrant . . . " clause */
			/* now we actually output the image */
			dumpimage(durp, 0);
		   } /* end of loop through frame-numbers */
	  } /* end of "if ((llendf > ..." clause */
	else
	  {
		(void)sscanf(xdline, "%s", frname);
		/* see if this image appears in our list. If it does,
		 we stuff the data into the header that is appropriate
		 for that frame number. Otherwise, we stuff it in the
		 zeroth header. */
		if (NULL == durp->dui_fmi) durp->dui_fmi = ftables(3);
		if (0 > (frame = match_inode(&(durp->dui_fmi), frname)))
			frame = 0;
		thishd = durp->dui_heads + (frame % durp->dui_nmbox);
		newinode = record_inode(frname);
		/* if we already have the correct information there,
		 we don't need to read the image in. If we don't have
		 the right information there, we do need to read it */
		if (newinode != thishd->quadrant)
		  {
			if (FRAME_OK != (result = read_image(frname,
			 durp->dui_imflags, thishd)))
			  {
				(void)image_error(stderr, frname, result, 0);
				return;
			  }
		  }
		durp->dui_imname = frname;	dumpimage(durp, 0);
	  }
}

static void newquery(char *name, char *fty[], DUMPRUN *durp)
{ /*	This prompts for the frame range and the (X,Y) center
	of the area on an image to be dumped.  It provides the
	user interface to the facility for dumping raw images. */

	char		frname[60], xdline[81];
	int		frame, lendf, lstartf, nvalues, xcn, ycn;
	static int	duwidth = 17;
	static int	duheight = 17;

	printf(" Enter x and y for center of area to be dumped: ");
	while (NULL == fgets(xdline, 80, stdin)) ;
	(void)sscanf(xdline, "%d%d", &xcn, &ycn);
	durp->dui_nmbox = PFRAMES + 1;
	if ((NULL == durp->dui_heads) && (NULL == (AVOID *)(durp->dui_heads
	 = (MAIL *)callop(durp->dui_nmbox, sizeof (MAIL),
	 "heads in newquery")))) return;
	/* Construct the inputs to the dumpimage routine */	
	durp->dui_wid = duwidth;	durp->dui_ht = duheight;
	durp->dui_xce = xcn;	durp->dui_yce = ycn;
	durp->dui_imname = name;
	if (name && (*name) && (0 == xgaccess(name, 4, 1)))
	  {	/* dump the counts in the pixels */
		durp->dui_imname = name;	dumpimage(durp, 1);
	  }
	printf(" Enter a filename or the starting and ending frame numbers: ");
	while (NULL == fgets(xdline, 80, stdin)) ;
	lstartf = lendf = -1;
	nvalues = sscanf(xdline, "%d %d", &lstartf, &lendf);
	if ((lendf >= 0) && (nvalues >= 2))
	  {
		if (lstartf < 0) query_image((char *)NULL, fty, durp);
		durp->dui_fmi = ftables(3);
		if ((durp->dui_fmi)->fi_firstfm > lstartf)
			lstartf = (durp->dui_fmi)->fi_firstfm;
		if ((durp->dui_fmi)->fi_lastfm  < lendf)
			lendf   = (durp->dui_fmi)->fi_lastfm;
		if (lendf < lstartf) lendf = lstartf;
		printf( "Frames %3d to %3d will be dumped\n", lstartf, lendf);
		for (frame = lstartf; frame <= lendf; frame++)
		   {
			if (NULL == (durp->dui_imname = indexname(frame)))
			  {
				fprintf(stderr,
				 "Cannot find requested frame file"); return;
			  }
			dumpimage(durp, 1);
		   }
	  }
	else
	  {
		(void)sscanf(xdline, "%s", frname);
		durp->dui_imname = frname;	dumpimage(durp, 1);
	  }
}

static int matchname(char *s, char *fty[], DUMPRUN *durp)
{ /*	This returns the index within the "fty" array of the string "s".
	Case distinctions are ignored.  If "s" isn't IN the "fty" array,
	the routine returns -1. */
	register int	type;
	char		c, *ln;

	/* get a lower-case version of the input string */
	for (ln = durp->dui_lcname; '\0' != (c = *s++); ln++)
		*ln = ('A' <= c && c <= 'Z' ? c + ('a' - 'A') : c);
	*ln = '\0';
	/* go through the list of file types.  Return the
	index of the type specified if a match occurs. */
	for (type = 0; type < NTYPES; type++)
	   if (!strcmp(durp->dui_lcname, fty[type])) return type;
	/* handle some common synonyms for "stop" */
	if ((!strcmp(durp->dui_lcname, "exit")) ||
	 !(strcmp(durp->dui_lcname, "quit")) ||
	 !(strcmp(durp->dui_lcname, "end"))) return NTYPES-1;
	return UNKNOWN;	/* no match occurred */
}

static void dodump(int type, char *name, char *fty[], DUMPRUN *durp)
{ /* image dump routine */
	int		arediff, i, lbxyz, linesout, locht, locwid;
	int		lqxyz, ltot, nbread, row;
	short		ispgp;
	char		*origname, rundate[26];
	off_t		newpos, curpos;
	time_t		tl0;
	sichar		*sic;
	FILE		*fpi;
	VSPLINE		*sgsp;
	TROID		tr;
	REFPARS		rpar;
	PREDREF		prst;
	OBSREF		urf;
	REGIONPROFILE	*regb;
	CRYSTAL		xhead;
	DATARUN		rhead;
	MERGEDREF	mr;
	static double	scl[4];
	static char	cenhead0[] =
	 "      <X>      <Y>    <phi> start #of      total   Profile\n";
	static char	cenhead1[] =
	 "                            frame obs  intensity   elements\n";

	if (type == 0)
	  {
		query_image(name, fty, durp);	(void)ftables(0); return;
	  }
	if (type == 1)
	  {
		newquery(name, fty, durp);	(void)ftables(0);	return;
	  }
	if (NULL == (fpi = xgopen(name, "r")))
	  {
		fprintf(stderr,"xdump: Cannot open %s as input %s file\n",
			name, fty[type]);
		return;
	  }
	tl0 = NOW;	strcpy(rundate,ctime(&tl0));
	fprintf(durp->dui_fpd, "Dump of file %s of type %s on %s\n",
		name,  fty[type], rundate);
	printf("Dump of file %s of type %s on %s\n",
		name,  fty[type], rundate);
	linesout = 0;
	switch (type)
	      {
		case	2:	/* spline file */
			origname = getenv("USPLINE"); /*remember orig name*/
			arediff = (NULL != origname) && ('\0' != *origname) &&
			  strcmp(name, origname);
			if (arediff) (void)set_logical("USPLINE", name);
			if (NULL == (sgsp = gspline(5, &i))) return;
			pspline(durp->dui_fpd, sgsp);	/*all spline info */
			pspline(stdout, sgsp);	/* most to stdout */
			/* change the logical name back if needed */
			if (arediff) (void)set_logical("USPLINE", origname);
			break;
		case	3:	/* centroid file */
			fprintf(durp->dui_fpd, cenhead0);
			fprintf(stdout, cenhead0);
			fprintf(durp->dui_fpd, cenhead1);
			fprintf(stdout, cenhead1);
			for ( ; 0 < xgread((void *)&tr, XG_TROID, fpi);
			 linesout++)
			   {
				prtroid(durp->dui_fpd, &tr);
				if (linesout < durp->dui_maxlines)
					prtroid(stdout, &tr);
			   }
			break;
		case	4:	/* parameters file */
			if ((EOK != (i = refparin(&rpar, &nbread)) ||
			 (nbread < 232)))
			  {
				printf("error %d in refparin: nbread = %d\n",
					i, nbread);
			  }
			show_params(stdout, &rpar, nbread);
			show_params(durp->dui_fpd, &rpar, nbread);
			break;
		case	5:	/* uspotlist file */
			fprintf(durp->dui_fpd,
 "   h   k   l frame      <X>      <Y>     <phi> corrections    gamma\n");
			fprintf(stdout,
 "   h   k   l frame      <X>      <Y>     <phi> corrections    gamma\n");
			for ( ; 0 < xgread((void *)&prst, XG_PREDREF, fpi);
			 linesout++)
			   {
				prpred(durp->dui_fpd, &prst);
				if (linesout < durp->dui_maxlines)
					prpred(stdout, &prst);
			   }
			break;
		case	6:	/* urefls file: use (X,Y,omega) geometry */
		case	7:	/* urefls file: use (alpha,beta,gamma) geom */
			find_dimension(&locwid, &locht, 0, (char *)NULL);
			if (!locwid)
			  {
				locwid = locht = 512;
			  }
			scl[0] = dimscale((short)locwid);
			scl[1] = dimscale((short)locht);
			scl[2] = 100. * needlambda();	scl[3] = 0.;
			scl[3] = (type == 6) ? 0. : 1.;
			for (linesout = 0;
			 0 < xgread((void *)&urf, XG_OBSREF, fpi); linesout++)
			   {
				prurefl(durp->dui_fpd, &urf, scl);
				if (linesout < durp->dui_maxlines)
					prurefl(stdout, &urf, scl);
			   }
			break;
		case	8:	/* wmask (weighted mask) file */
			durp->dui_wmask = (MASK_3D *)callop(1,
			 sizeof (MASK_3D), "wmask in dodump");
			for (linesout = 0; (linesout < 12) && (0 <
			 xgread((void *)durp->dui_wmask, XG_MASK_3D, fpi));
			 linesout++)
			   {
				prmask(durp->dui_wmask, linesout,
					durp->dui_fpd);
				if (linesout < 1)
				  prmask(durp->dui_wmask, linesout, stdout);
			   }
			/* if this is the new kind, the file will be longer*/
			/* first, test it... */
			curpos = ftell(fpi);
			(void)fseek(fpi, 0L, 2);
			newpos = ftell(fpi);
			if (curpos == newpos) break;
			/* Now we know there's more. Rewind and get it */
			(void)fseek(fpi, curpos, 0);
			if (NULL == (regb = (REGIONPROFILE *)callop(1, sizeof
			 (REGIONPROFILE), "regb in dodump"))) return;
			for (linesout = 0; (linesout < 12) && (0 <
			 fread((void *)regb, sizeof (REGIONPROFILE),1, fpi));
			 linesout++)
			   {
				lbxyz = regb->m3_bx * regb->m3_by *regb->m3_bz;
			 	lqxyz = regb->m3_qx * regb->m3_qy *regb->m3_qz;
				ltot = lbxyz + lqxyz;
				if (NULL == (sic = (sichar *)mallop(ltot,
				 "sic in dodump"))) return;
				if (1 > (i = fread((AVOID *)sic,
				 sizeof (sichar), ltot, fpi))) return;
				regb->m3_xyoip = sic;
				regb->m3_dvip = sic + lbxyz;
				dispmodel(durp->dui_fpd, regb, 2);
				dispmodel(durp->dui_fpd, regb, 3);
				memzap((AVOID **)&sic, "sic in dodump");
			   }
			break;
		case	9:	/* "multiref" file */
			if (verror(0 < xgread((void *)&xhead, XG_CRYSTAL,fpi),
			 "read crystal header of multiref file")) return;
			pr_xt(durp->dui_fpd, &xhead);	pr_xt(stdout, &xhead);
			ispgp = xhead.xt_intl;
			if ((ispgp < 200) && (xhead.xt_isys== 7)) ispgp += 100;
			for (i = 0; i < xhead.xt_nruns; i++)
			   {
				if (verror(0<= xgread((void *)&rhead,
				 XG_DATARUN, fpi),
				 "read datarun header of multiref")) return;
				pr_dr(durp->dui_fpd, &rhead);
				pr_dr(stdout, &rhead);
			   }
			(void)keep_dimen(&rhead);
			fprintf(durp->dui_fpd,
 "    h   k   l sin(th)  intensity   sigma  shift cham-run     phi   X   Y\n");
			fprintf(durp->dui_fpd,
 "              /lambda                         #  ber   #\n");
			fprintf(stdout,
 "    h   k   l sin(th)  intensity   sigma  shift cham-run     phi   X   Y\n");
			fprintf(stdout,
 "              /lambda                         #  ber   #\n");
			for ( ; 0 <xgread((void *)&mr, XG_MERGEDREF, fpi);
			 linesout += (mr.mr_nob + 1))
			   {
				pr_merref(durp->dui_fpd, &mr, ispgp);
				if (linesout < durp->dui_maxlines)
				  pr_merref(stdout, &mr, xhead.xt_intl);
			   }
			break;
		case	10:	/* amask file: read it and show it  */
			if (EOK != (i = find_dimension(&locwid, &locht,
				0, (char *)NULL))) return;
			if (NULL == activmask(locwid, locht, 5))
			  {
				if (errno == ENOMEM) fprintf(stderr,
				 "Error allocating memory for active mask\n");
				else fprintf(stderr,
				 "Error reading active mask\n");	return;
			  }
			head_results(durp->dui_fpd);	head_results(stdout);
			for (row = 0; row < locht; row++)
			   {
				sum_row(durp->dui_fpd, row, locwid);
				if (!(row % 16)) sum_row(stdout, row, locwid);
			   }
			break;
		default:
			break;
	     }
	(void)fclose(fpi);
	return;
}

static int endxd(int reval, DUMPRUN *durp)
{ /*	Clean termination of "xdump" */
	if (NULL != durp->dui_fpd) (void)fclose(durp->dui_fpd);
	durp->dui_fpd = NULL;
	/* Now release memory for images if we've saved it */
	(void)ftables(0);
	kill_mail(&(durp->dui_heads), &(durp->dui_nmbox));
	memzap((AVOID **)&(durp->dui_lcname), "lcname in xdump");
	memzap((AVOID **)&(durp->dui_wmask), "wmask in xdump");
	globalkill();	return reval;
}

int xdumpmain(int argc, char *argv[])
{ /*	This is the mainline for "xdump".  Its primary responsibility is
	decoding the runtime inputs or the answers to the runtime prompts.*/
	register int		typecode;
	char			filetype[20], finame[80], xdline[80];
	static DUMPRUN		duru;
	static char *ftypes[NTYPES] = {"image", "calibdat", "uspline",
	 "centroids", "uparams", "uspotlist", "urefls", "ureflk",
	 "wmask", "multiref", "amask",
	"maxlines", "stop" };

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	if (NULL == (AVOID *)(duru.dui_lcname = (char *)mallop(120,
		"lcname in xdumpmain"))) return endxd(EALLOC, &duru);
	if (qerror((AVOID *)(duru.dui_fpd = propen("udump")),
	 "open xdump output file")) return endxd(EOPENERR, &duru);
	duru.dui_maxlines = DEF_MAXLINES; /* default value for "maxlines" */
	if (argc < 2) /* no args: get commands as responses to prompts */
	 {
	   while (1)
	    {
		fprintf(stderr,"xd> ");
		if (NULL == fgets(xdline, 80, stdin)) break;
		finame[0] = '\0';
		(void)sscanf(xdline,"%s %s", filetype, finame);
		if (UNKNOWN == (typecode = matchname(filetype, ftypes, &duru)))
			fprintf(stderr,"%s: unknown file type\n", filetype);
		else if (ISSTOP == typecode) break;
		else if (ISMAXLINES == typecode)
			(void)sscanf("%s %d", filetype, &(duru.dui_maxlines));
		else if ('\0' == finame[0])
		  {
			if (typecode == 7)
				dodump(7, trname("urefls"), ftypes, &duru);
			else dodump(typecode, trname(filetype), ftypes, &duru);
		  }
		else	dodump(typecode, finame, ftypes, &duru);
	    }
	 }
	else
	  {
		if (UNKNOWN == (typecode = matchname(argv[1], ftypes, &duru)))
		  {
			fprintf(stderr,"%s: unknown file type\n", argv[1]);
			(void)fclose(duru.dui_fpd);	duru.dui_fpd = NULL;
			return endxd(EOK, &duru); /*normal exit, of sorts*/
		  }
		if (ISSTOP == typecode || ISMAXLINES == typecode)
			return endxd(EOK, &duru);
		if (argc == 2)
		  {
			if (typecode == 7)
				dodump(7, trname("urefls"), ftypes, &duru);
			else dodump(typecode, trname(argv[1]), ftypes, &duru);
		  }
		else		dodump(typecode, argv[2], ftypes, &duru);
	  }
	return endxd(EOK, &duru);
}

/* end of x_dump.c */
