/* xg_now.c -- set up environment and images for XGEN

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions; reorganize use of "{"
	29 May 2006: move fullname and install_names here so they're local
	15 Mar 2006: handle string arguments correctly
	31 Dec 2005: handle alphabetic arguments, e.g. -pBlah, correctly
	12 Oct 2005: housekeeping
	11 May 2005: check return status from xgjournal
	 2 Apr 2005: add -l option; rationalize use of -l and -h
	15 Oct 2004: source moved from ../nogui to ../utilities
	28 Aug 2004: routines used only here are declared as static;cosmetics
	18 Aug 2004: remaining static globals replaced with function args
	 5 Jun 2004: eliminate unnecessary diagnostics for eXample images
	27 May 2004: add "eXample image" concept
	18 Feb 2004: change kill_ftab call so that it doesn't have args
	25 Jan 2004: some static globals converted to local variables
	21 Jan 2004: license message inserted and initial comments reordered
	 8 Jan 2004: static globals (including XGNAME and FRAMEINDEX pointers)
		replace extern globals; function calls updated;
		FRAMEINDEX structures used extensively
	19 Sep 2003: oldfirst and oldlast always initialized
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	22 Aug 2003: call globalkill at the end, not in the middle
		(what was I thinking?)
	 5 Mar 2003: major rewrite: more options, different calls
	 5 Jan 2003: reorganized source-file header
	 1 Oct 2002: original version, based somewhat on setup1prog.c

   Purpose: defining user environment for X-GEN
   Cross references: setup1prog.c

 "xgnow" is a general-purpose bookkeeping program that requires no
 particular user input.  It looks for detector images in either "." or ".."
 and picks as a filename prefix the prefix that occurs most often.
 The environment-variable command file "xxx.cmd" and the detector image file
 "xxx.frm" are then created and populated.
 Syntax:	xgnow [-h<val>]
	where <val> indicates the high limit to the frame-number list.
 Specifying this allows us to get all the frames named in the FRAMES file
 even if only the first half-dozen have been collected.
*************************************************************************/
#define		NEED_ERRNO	`
#include	"xgen.h"
#define		XGN_KEEPGLOBALS	1

extern int xgnowmain(int argc, char *argv[]);

static void xgoutc(FILE *fp, int nfull, int ncomp, FRAMEINDEX **lfmip)
{ /*	This prints out the results of an attempt to count the number of
	images in $FRAMES that are actually present */
	char	*finame;

	finame = trname("FRAMES");
	fprintf(fp, "Contents of frame index %s:\n", finame);
	if (nfull > 0)
	  {
		if (ncomp > 0) fprintf(fp,
	 " %4d images present, %4d in compressed format out of %4d listed\n",
		 nfull, ncomp, (int)((*lfmip)->fi_totfms));
		else fprintf(fp, " %4d images present out of %4d listed\n",
			nfull, (int)((*lfmip)->fi_totfms));
	  }
	else if (ncomp > 0) fprintf(fp,
	 " %4d compressed images present out of %4d listed\n",
			ncomp, (int)((*lfmip)->fi_totfms));
	else fprintf(fp,
	 " None of the %4d listed images is actually present\n",
		(int)((*lfmip)->fi_totfms));
}

static int xgncount(int *ncompp, FRAMEINDEX **lfmip)
{ /*	This returns the number of detector images specified in $FRAMES
	that actually exist currently and are readable. The ones that are
	absent but for which a compressed version is present (*.Z or *.gz)
	is returned in *ncompp. */
	int		accret, npresent;
	char		*cp, compressed[268];
	FRAMEENTRY	*lft;

	npresent = *ncompp = 0;
	if (NULL == *lfmip)
	  {
		if (NULL == (*lfmip = ftables(3))) return npresent;
	  }
	for (lft = (*lfmip)->fi_fmi; lft <= (*lfmip)->fi_fme; lft++)
	   {
		errno = 0;
		if ((NULL != (cp = lft->fmname) && ('\0' != *cp)))
		  {
			if (0 == (accret = access(cp, 4))) npresent++;
			else {
				strcpy(compressed, cp);
				strcat(compressed, ".gz");
				if (0 == access(compressed, 4)) (*ncompp)++;
				else {
					strcpy(compressed, cp);
					strcat(compressed, ".Z");
					if (0 == access(compressed, 4))
						(*ncompp)++;
				     }
			     }
		  }
	   }
	return npresent;
}

static char *fullname(char *str)
{ /* This converts a name (typically a filename) to a fully rooted
   name, free of environment variables or //, /./, and /../ constructions. */
	char	*usena2, *usena;

	usena2 = trans_envname(str); /* translate environment vbls*/
	usena = squeezeall(usena2);	/* get rid of wasteful constructions */
	if ((usena != usena2) && (usena2 != str))
		memzap((AVOID **)&usena2, "usena2 in fullname");
	return usena;
}

static int install_names(char *examp)
{ /*	This expects an argument with a pathname like
	/Users/ahoward/tx/p3/p3_0017.img in it. It will install
	/Users/ahoward/tx/p3 as the DATA name and p3_ as the id name.
	It will install the current working directory as WORK and SCRT
	if it's reasonable; if it isn't, it'll install $DATA/xgen instead. */
	int		nomv, lendot;
	char		*cpe, *cpi, *cpo, *epref, *finame;
	char		*lastsl, *otempl, *usena;
	static char	xgsub[] = "xgen";
	static char	*innbuf;
	static XGNAME	*namb;
	
	if (NULL == namb) /* arrange for name buffers if needed */
	  {
		if (NULL == (namb = sharename(7)))
		  {
			fprintf(stderr, "Cannot find name buffers\n");
			return EALLOC;
		  }
	  }
	usena = fullname(examp);
	for (cpi = usena; '\0' != *cpi; cpi++) ;
	cpe = cpi; /* remember the end of the string */
	/* count backward to find the final slash in the name */
	for (cpi = cpe - 1; (*cpi != '/') && (*cpi != '\0'); cpi--) ;
		/* name is too short */
	if ((cpi < usena) || ('/' != *cpi)) return EBADVALUE;
	lastsl = cpi;
	for (cpi = usena, cpo = namb->xmm_datana; cpi < lastsl; cpi++)
		*cpo++ = *cpi;
	*cpo = '\0';	/* create directory name */
	(void)set_logical("DATA", namb->xmm_datana);
	/* make sure the specified name is that of a real data image*/
	finame = lastsl + 1;	if ('\0' == *finame) return EBADVALUE;
	if (1 != (nomv = imagename(finame)))
	  {
		fprintf(stderr, "imagename(%s) returned %d\n", finame, nomv);
		return EBADVALUE;
	  }
	if (1 != (nomv = imagelike(namb->xmm_datana, finame)))
	  {
		fprintf(stderr, "imagelike(%s, %s) returned %d\n",
			namb->xmm_datana, finame, nomv);
		return EBADVALUE;
	  }
	epref = otempl = NULL;
	if (0 > (nomv = get_nominal(usena, &epref, &otempl)))
	  {
		fprintf(stderr,
		 "get_nominal(%s, %s, %s) returned %d\n",
			usena, epref, otempl, nomv);
		return EBADVALUE; /* define the prefix name */
	  }
	for (cpi = finame, cpo = namb->xmm_idna;
	 ('\0' != *cpi) && (cpi <= epref); cpi++) *cpo++ = *cpi;
	*cpo = '\0';	/* install that prefix as namb->xmm_idna */
	(void)set_logical("RUN_ID", namb->xmm_idna);
	if (NULL == (innbuf = (char *)mallop(260,
		"innbuf in install_names"))) return EALLOC;
	zeromem(innbuf, 260);
	mygetwd(innbuf);	/* record current directory */
	/* we're done with `usena', so release memory for it if we can */
	if (usena != examp) memzap((AVOID **)&usena, "usena in install_names");
	for (cpi = innbuf, lendot = 0; (lendot < 254) && *cpi != '\0';
	 cpi++) lendot++;
	/* we use the current working directory as $WORK if its
	 name is of reasonable length and if it's writable */
	if ((1 < lendot) && (lendot <= 126) && (0 == access(innbuf, 6)))
	  {
		for (cpo = namb->xmm_workna, cpi = innbuf;
		 '\0' != *cpi; cpi++) *cpo++ = *cpi;
		*cpo++ = '\0';
		memzap((AVOID **)&innbuf, "innbuf in install_names");
	  }
	/* otherwise we attempt to use $DATA/xgen or $DATA as $WORK */
	else {
		memzap((AVOID **)&innbuf, "innbuf in install_names");
		for (cpi = namb->xmm_datana, lendot = 0;
		 (lendot < 126) && ('\0' != *cpi); cpi++) lendot++;
		for (cpo = namb->xmm_workna, cpi = namb->xmm_datana;
		 '\0' != *cpi; cpi++) *cpo++ = *cpi;
		if (lendot < 122) /* try to append /data if we can */
		  {
			/* make sure there's a slash at the end */
			if ((cpo > namb->xmm_workna) && (*(cpo-1) != '/'))
				*cpo++ = '/';
			/* append "xgen" to the end of the name */
			for (cpi = xgsub; '\0' != *cpi; cpi++)
				*cpo++ = *cpi;
			*cpo = '\0'; /* ... and null-terminate it */
		  }
		else if (lendot < 127) *cpo = '\0';	/* use DATA as WORK */
		/* see if this directory already exists & is writable.
		 If it doesn't exist, we try to create it. */
		errno = 0;
		if (0 != access(namb->xmm_workna, 06))
		  {
			if (errno != ENOENT) return EBADVALUE;
			if (0 != mydircreate(namb->xmm_workna))
				return EBADVALUE;
		  }
	      } /* end of "else" for "if ((1 < lendot)... clause */
	(void)set_logical("WORK", namb->xmm_workna);
	/* we start with SCRT equal to WORK */
	for (cpo = namb->xmm_scrtna, cpi = namb->xmm_workna;
	 '\0' != *cpi; cpi++) *cpo++ = *cpi;
	*cpo = '\0';	(void)set_logical("SCRT", namb->xmm_scrtna);
	return EOK;
}

static int xgnowargs(int *xgnfp, int ac, char *av[], FRAMEINDEX **lfmip)
{ /*	This processes command-line arguments to `xgnow'.
	Syntax:
	xgnow [-d<datadir>] [-h<val>] [-i<xtal_identifier>] <xtal_identifier>
  */
	int		ival, specmax, specmin;
	char		*lstr, *specdata, *specexample, *specid;
	char		**dcp, **flags;
	char		*next, cnext;
	static XGNAME	*namb;

	flags = flagproc(ac, av, "HLZ");
	*xgnfp = specmax = 0;	specdata = specid = specexample = (char *)NULL;
	specmin = 9999;
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	 {
	   if (('Z' != **dcp) && ('\0' == *(*dcp + 1)) &&
		(NULL != *(dcp + 1))) lstr = *(dcp + 1);
	   else	lstr = *(dcp) + 1;
	   ival = atoi(lstr);
	   switch (**dcp)
	     {
		case	'H':	/* specify highest frame */
			specmax = ival;	break;
		case	'L':	/* specify lower image limit */
			specmin = ival;	break;
		case	'Z':	/* don't delete globals before exit */
			*xgnfp |= XGN_KEEPGLOBALS;	break;
	     }
	    if (lstr == *(dcp + 1)) dcp++;
	   }
	/* flag arguments that take string output: */
	for (dcp = av + 1, ival = 1; ival < ac; dcp++, ival++)
	   if (**dcp == '-')
	     {
		next = (*dcp) + 1;	cnext = *next;	lstr = next + 1;
		if (('a' <= cnext) && (cnext <= 'z')) cnext -= 'a' - 'A';
		/* -d<name> means the data will be in directory <name> */
		if (cnext == 'D') specdata = ('\0' == *lstr) ? "S" : lstr;
		/* -i<name> means the run ID will be <name> */
		else if (cnext == 'I') specid = lstr;
		/* -x<hame> means we should determine the run ID from
		 the file with name <name>, assuming that file exists */
		else if (cnext == 'X')
		   {
			if ((NULL != lstr) && ('\0' != *lstr) &&
			 (' ' != *lstr) && (0 == access(lstr, 04)))
				specexample = lstr;
		   }
	     }
	/* a non-flag argument is treated either as a project identifier or as
	 an eXample data image, depending on whether or not it's a filename */
	for (ac--, av++; (ac) && **av == '-'; ac--) av++;
	if (ac >= 1)
	  {
		if (0 == access(*av, 04)) specexample = *av; else specid = *av;
	  }
	/* unset a few variables so that ftables() does the right thing */
	if (0 >= (ival = unset_logical("FRAMES")))
		fprintf(stderr, "Warning: couldn't un-set FRAMES:err=%d\n",
			ival);
	if (NULL == specdata)
	  {
		if (0 >= (ival = unset_logical("DATA")))
		  fprintf(stderr, "Warning: couldn't un-set DATA:  err=%d\n",
			ival);
	  }
	else if (strcmp("S", specdata)) (void)set_logical("DATA", specdata);
	if (NULL == (namb = sharename(7)))
	  {
		fprintf(stderr, "Cannot find name buffers\n"); return EALLOC;
	  }
	if ((NULL == specid) || ('\0' == *specid)) *(namb->xmm_idna) = '\0';
	else strcpy(namb->xmm_idna, specid);
	/* note that, by convention, the eXample image takes precedence
	 over any of the other ways of specifying a data directory and an
	 identifier */
	if ((NULL != specexample) && ('\0' != *specexample))
		(void)install_names(specexample);
/* tell us what the filename prefix and image values should be */
	*lfmip = ftables(7);
/* Change as of 2 April 05: a specified upper frame limit is applied
 * even if the specified value is SMALLER than the observed limit.
 * We also allow for a specification of the lower limit.
 */
	if (specmax > 0) (*lfmip)->fi_lastfm = specmax;
	if (specmin < 9999) (*lfmip)->fi_firstfm = specmin;
	if ((*lfmip)->fi_lastfm < (*lfmip)->fi_firstfm)
	  {
		specmin = (*lfmip)->fi_lastfm;
		(*lfmip)->fi_lastfm = (*lfmip)->fi_firstfm;
		(*lfmip)->fi_firstfm = specmin;
	  }
	(*lfmip)->fi_totfms = 1 + (*lfmip)->fi_lastfm - (*lfmip)->fi_firstfm;
	return EOK;
}

int xgnowmain(int argc, char *argv[])
{ /* mainline for xgnow */
	int			failure, lenindex, ncompres, needindex, npres;
	int			oldfirst, oldlast, xgnflags;
	char			*friname;
	static FRAMEINDEX	*lfmi;
	FRAMEENTRY		*lfti, *lfte;
	static int		nlogicals;
	static XGNAME		*namb;

	if (EHELP == xgjournal(argc, argv)) return EOK; /* journal the effort */
	/* reserve memory for some working buffers */
	if (NULL == (namb = sharename(7)))
	  {
		fprintf(stderr, "Cannot find name buffers\n"); return EALLOC;
	  }
	/* look at command-line arguments to decide what to do next */
	if (EOK != (failure = xgnowargs(&xgnflags, argc, argv, &lfmi)))
	  {
		fprintf(stderr, "Error %d from xgnowargs\n", failure);
		return failure;
	  }
	printf(" id-name = %s\n", namb->xmm_idna);
		/* set the logical names, do so */
	if (0 > (nlogicals = more_logicals(0)))
	  {
		fprintf(stderr, "more_logicals returned %d\n",
			nlogicals);
		return EBAD;
	  }
	fprintf(stdout, " %3d new logicals defined\n", nlogicals);
		/* if we're supposed to write out the logical names, do so */
	nlogicals = writelognames(0);
	needindex = 1;
	/* create the frame index unless it
	 already exists and has the right limits */
	if (NULL == lfmi) lfmi = ftables(7);
	friname = trname("frames");
	oldfirst = lfmi->fi_firstfm; oldlast = lfmi->fi_lastfm;
	if (0 == access(friname, 02))
	  {
		if (NULL != (lfmi = ftables(3)))
		  {	oldfirst = (lfmi->fi_fmi)->fmno;
			oldlast =  (lfmi->fi_fme)->fmno;
		  }
		lenindex = strlen(namb->xmm_idna);
		if ((oldfirst == lfmi->fi_firstfm) &&
		 (oldlast == lfmi->fi_lastfm) && strncmp(namb->xmm_idna,
		 (lfmi->fi_fmi)->fmname, lenindex)) needindex = 0;
	  }
	if (needindex)
	  {
		writeindex(lfmi, 0); /* write out the frame index */
		/* we may need to activate it over again */
		lfti = lfmi->fi_fmi;	lfte = lfmi->fi_fme;
		if ((NULL == lfti) || (lfmi->fi_totfms < 1) ||
		 (lfmi->fi_firstfm != lfti->fmno) ||
		 (lfmi->fi_lastfm != lfte->fmno))
		  {
			if (NULL == (lfmi = ftables(3)))
			  {
				fprintf(stderr,
				 "failure in creating image table\n");
				return EREADERR;
			  }
		  }
		npres = xgncount(&ncompres, &lfmi);
		xgoutc(stdout, npres, ncompres, &lfmi);
	  }
	(void)ftables(0); if (!(xgnflags & XGN_KEEPGLOBALS)) globalkill();
	return EOK;
}

/* end of xg_now.c */
