/* xgenparse.c -- manage parsing issues for line-oriented interfaces in X-GEN

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: reorganize use of "{"
	 7 Jul 2006: allow '?' as a synonym for 'help' and ignore blank lines
	29 May 2006: handle blank command ("\n") silently; housekeeping
	 3 Jan 2006: handle environment-variable assignments more expressively
	 7 Nov 2005: move static globals inside "xgdo".
	13 Oct 2005: housekeeping
	27 May 2005: handle ~ constructions as well as environment variables
	24 May 2005: handle help, version, help for file types, dumping files
	14 May 2005: xgavail() doesn't exit on error
	12 May 2005: incorporate xgcomargs, etc.
	 6 May 2005: new code.

   Purpose: parsing routines for xgen and xgentools
   Cross references: everything
*************************************************************************/
#define		NEED_ERRNO	1
#define		NEED_UNISTD	1
#include	"xgen.h"
#include	<glob.h>
#include	<sys/wait.h>
#include	<sys/types.h>
#include	<pwd.h>

static int xgappc(int ac, char *av[], int (*xgroutine)(int ac, char *av[]))
{ /*	This calls the X-GEN routine `xgroutine' with arguments (ac, av) */
	return xgroutine(ac, av);
}

static int aok(const char *path, int erno)
{
	if (erno) printf("Warning: cannot expand the string %s: error = %d\n",
	 path, erno);
	return erno; /* a harmless function that "glob" can call */
}

static int all1command(char *str)
{ /*	returns 1 iff str could be construed as a single command,
	possibly with leading blanks. Comma is allowed, which is dubious. */
	char	*cp, *fnb, *tcp;
	static char illeg[] =
	 { ' ', '\t', '\n', '<', '>', '|', '{', '}', '"', '\0' } ;
	/* static char illeg[] = " <>|{}}'`\042\t\n" ; */

	if ((NULL == str) || ('\0' == *str)) return 0;
	for (cp = str, fnb = NULL; '\0' != *cp; cp++)
		if ((' ' != *cp) && ('\t' != *cp) && ('\n' != *cp))
		  {
			fnb = cp;	break;
		  }
	if (NULL == fnb) return 0;
	for (cp = fnb; '\0' != *cp; cp++)
	   {
		for (tcp = illeg; '\0' != *tcp; tcp++)
			if (*cp == *tcp) return 0;
	   }
	return 1;
}

static char *findhome(char *username)
{ /*	This returns a user's home directory name, as best it can */
	char		*icp, *ocp, *rdn, *una;
	struct passwd	*pwep;
	size_t		outlen;

	if ((NULL == username) || ('\0' == *username))
	  {
		ocp = getenv("HOME"); /* try to return $HOME */
		if ((NULL != ocp) && ('\0' != *ocp)) return ocp;
		una = getenv("USER"); /* $HOME not defined: try for $USER */
	  }
	else una = username;
	if ((NULL == una) || ('\0' == *una)) return NULL;
	if (NULL != (pwep = getpwnam(una)))
	  {
		outlen = strlen(pwep->pw_dir) + 2;
		if (NULL != (rdn = (char *)mallop(outlen,
			 "rdn in findmyhome"))) return NULL;
		icp = pwep->pw_dir;
		for (ocp = rdn; '\0' != *icp; icp++, ocp++) *ocp = *icp;
		*ocp = '\0';
		return rdn;	/* login name from passwd entry*/
	  }
	return NULL;
}

static char *trans_tildexp(char *str)
{ /*	This reads the string `str' and returns a string derived from it
	in which an initial tilde escape sequence has been converted.
	Expressions like ~+4 and ~-3 and ~3 are not implemented because
	I don't know how to do that */
	char		*endstr, *hstr, *icp, *ocp, *resstr, *retstr;
	size_t		lenun, lenout;

	if ('~' != *str) return str;	/* return input if *str isn't '~' */
	if ('\0' == *(str + 1)) /* simplest case: arg is just "~" */
	  {
		if (NULL == (retstr = findhome("")))
		  {
			printf("Tilde escape %s untranslatable\n", str);
			return NULL;
		  }
		return retstr;
	  }
	/* otherwise we have to parse it a little more vigorously */
	for (icp = str + 1; '\0' != *icp; icp++) if ('/' == *icp) break;
	endstr = icp;
	if (endstr == str + 1)	/* construction is ~/ */
	  {
		hstr = findhome(""); lenout = strlen(hstr) + strlen(str);
		if (NULL == (retstr = (char *)mallop(lenout,
		 "retstr in trans_tildexp"))) return NULL;
		zeromem((AVOID *)retstr, lenout);
		ocp = retstr;
		for (icp = hstr  ; '\0' != *icp; icp++, ocp++) *ocp = *icp;
		for (icp = endstr; '\0' != *icp; icp++, ocp++) *ocp = *icp;
		*ocp = '\0';	return retstr;
	  }
	/* I'm not going to try to work with ~+4 or ~-3 or ~4 for now */
	if (allnumeric(endstr) ||
	 (('\0' != *(endstr+1)) && allnumeric(endstr + 1))) return NULL;
	/* now we have something like ~fred/stuff */
	lenun = endstr - str; /* measure user's name */
	if (NULL == (hstr = (char *)mallop(lenun, "hstr in trans_tildexp")))
		return NULL; /* memory for name */
	zeromem(hstr, lenun);
	for (icp = str + 1, ocp = hstr; icp < endstr; ) *ocp++ = *icp++;
	*ocp = '\0'; /* copy name to "hstr" */
	resstr = findhome(hstr); /* translate its meaning */
	memzap((AVOID **)&hstr, "hstr in trans_tildexp"); /* done ! */
	lenun = strlen(resstr) + strlen(endstr) + 2; /* measure output */
	if (NULL == (retstr = (char *)mallop(lenun,
	 "retstr in trans_tildexp"))) return NULL; /* get memory for output */
	icp = resstr;	ocp = retstr; /* copy username to output */
	for (icp = resstr, ocp = retstr; '\0' != *icp; ) *ocp++ = *icp++;
	for (icp = endstr; '\0' != *icp; ) *ocp++ = *icp++; /* append rest */
	*ocp = '\0'; /* null-terminate */
	return retstr;	/* return the proper string */
}

static int globexpand(int *acp, char **avp[])
{ /*	For each element of the array of strings *avp,
	we translate any environment variables found within the string.
	The we perform global expansions on that string using glob() */
	int		gflags, ic, nnc, nnew, nonnull, oldpathc, *wasg, *wasp;
	char		*icp, *ncp, *ocp, **dcp, **idcp, **ndcp, **icdcp;
	static glob_t	pg;

	if (*acp < 1) return 0;
	gflags = GLOB_DOOFFS;
	pg.gl_offs = *acp; /* set the offset to the # of args */
	if (NULL == (wasg = (int *)callop(*acp, sizeof (int),
	 "wasg in globexpand"))) return EALLOC;
	/* first, replace tilde conventions either with their full
	 expansions or with environment variables */
	for (dcp = *avp; NULL != *dcp; dcp++)
	   {
		if ('~' == **dcp)
		  {
			if (NULL != (ocp = trans_tildexp(*dcp))) *dcp = ocp;
			else printf(
		 "Untranslatable tilde expression %s not converted\n", *dcp);
		  }
	   }
	/* next, replace any environment variables with their expansions */
	for (dcp = *avp; NULL != *dcp; dcp++)
	   {
		if (*dcp != (ocp = trans_envname(*dcp))) *dcp = ocp;
	   }
	oldpathc = 0;
	for (nnew = ic = 0, dcp = *avp, wasp = wasg;
	 ic < *acp; ic++, dcp++, wasp++)
	   {
		if (pg.gl_pathc) gflags |= GLOB_APPEND;
		if (0 == glob(*dcp, gflags, aok, &pg))
		  {
			nnew = pg.gl_pathc;
			*wasp = *acp + nnew;
			oldpathc = pg.gl_pathc;
		  }
		else *wasp = ic; /* remember to copy this one */
	   }
	if (!nnew)
	  {
		memzap((AVOID **)&wasg, "wasg in globexpand");
		return EOK;	/* no expansions--leave it alone */
	  }
	/* if we did some expansions, recreate the strings */
	for (ic = 0, dcp = pg.gl_pathv, idcp = *avp, wasp = wasg; ic < *acp;
	 ic++, dcp++, wasp++, idcp++) *dcp = (*wasp < *acp) ? *idcp : NULL;
	nnew = (*acp) + pg.gl_pathc;
	for (dcp = pg.gl_pathv, nnc = ic = nonnull = 0; ic < nnew; ic++, dcp++)
	   if (NULL != *dcp)
	     {
		nonnull++;	nnc += strlen(*dcp);
	     }
	if (NULL == (ndcp = (char **)callop(nonnull + 1,
	 sizeof (char *), "ndcp in globexpand"))) return EALLOC;
	nnc += nonnull + 2;
	if (NULL == (ncp = (char *)mallop(nnc, "ncp in globexpand")))
	  {
		memzap((AVOID **)&wasg, "wasg in globexpand");
		memzap((AVOID **)&ndcp, "ndcp in globexpand"); return EALLOC;
	  }
	idcp = pg.gl_pathv;	ocp = ncp;	dcp = ndcp;
	icdcp = pg.gl_pathv + *acp;
	for (ic = 0, wasp = wasg; ic < *acp; ic++, wasp++)
	   {
		if (*wasp < *acp)
		  {
			idcp = pg.gl_pathv + ic;	*dcp++ = ocp;
			for (icp = *idcp; '\0' != *icp; ) *ocp++ = *icp++;
			*ocp++ = '\0';
		  }
		else {
			for (idcp = icdcp; idcp < pg.gl_pathv + *wasp; idcp++)
			   {
				*dcp++ = ocp;
				for (icp = *idcp; '\0' != *icp; )
					*ocp++ = *icp++;
				*ocp++ = '\0';
			   }
			icdcp = idcp;
		     }
	   }
	*avp = ndcp;	*acp = nonnull;
	memzap((AVOID **)&wasg, "wasg in globexpand");
	globfree(&pg);	pg.gl_offs = pg.gl_pathc = 0;
	return EOK;
}

static void list_xgcommands(FILE *fp, char *xgve[])
{ /*	This lists all the X-GEN commands available from the current
	application to the stream fp */
	int	ic, spa;
	char	**dcp;

	fprintf(fp, "X-GEN commands available within this application:\n");
	for (spa = ic = 0, dcp = xgve; NULL != *dcp; dcp++, ic++)
	   {
		fprintf(fp, "%s", *dcp);
		spa += strlen(*dcp) + 2;
		if (spa > 78)
		  {
			spa = 0;	printf("\n");
		  }
		else fprintf(fp, "  ");
	   }
	fprintf(fp, "\n");
}

static int xgffork(char **comdve)
{ /*	This does the fork-and-exec of a shell that executes
	a file that contains the command found in comdve. */
	int		ostatus, status;
	char		**dcp, tefil[128];
	pid_t		curpid, pid;
	static char	*outcom[3];
	static FILE	*xgfp;

	if ((NULL == comdve) || (NULL == *comdve)) return EBADVALUE;
	curpid = getpid();
	zeromem(tefil, 128);
	sprintf(tefil, "/tmp/.xgt%d", (int)curpid);
	if (NULL == (xgfp = xgopen(tefil, "w")))
	  {
		printf("Fork failed: temp file can't be created\n");
		return EOPENERR;
	  }
	for (dcp = comdve; NULL != *dcp; dcp++)
	   {
		if (dcp > comdve) fprintf(xgfp, " ");
		fprintf(xgfp, "%s", *dcp);
	   }
	fprintf(xgfp, "\n");	(void)fclose(xgfp);	xgfp = NULL;
	outcom[0] = "sh";	outcom[1] = tefil;	outcom[2] = NULL;
	if (-1 == (pid = fork())) return pid;
	if (0 == pid) /* we're forked here */
	  {
		errno = 0;	execvp(outcom[0], outcom); perror("");
		exit(errno ? errno : 2); /* get here only if execvp fails*/
	  }
	else /* This happens in the parent process */
	 {
		do {
			if (-1 == waitpid(pid, &status, 0))
			  {
				if (-1 == errno)
				  {
					ostatus = -1;	break;
				  }
			  }
			else if (WIFEXITED(status))
			  {
				ostatus = WEXITSTATUS(status);	break;
			  }
			else {
				ostatus = status;	break;
			     }
	           } while (1);
	 }
	errno = 0;	
	if (EOK != (status = unlink(tefil))) printf(
	 "unlink of %s failed: status = %d, errno = %d\n",
		tefil, status, errno);
	return ostatus;
}

static int envparse(char *instr, char **namstr, char **valstr)
{ /*	This returns EOK iff instr is of the form "name=value", where name
	is an environment variable and value is its value.  It returns the
	appropriate strings in **namstr and **valstr */
	char	*cp, *eqp, *ocp, *onam, *oval;
	size_t	lenn, lennp2, lenv, lenvp2;

	if ((NULL == instr) || ('\0' == *instr))
	  {
		printf("envparse called with improper argument\n");
		return -2;
	  }
	for (cp = instr; '\0' != *cp; cp++) if ('=' == *cp) break;
	eqp = cp;
	if ('\0' == *eqp) return -3; /* no equal sign found */
	if (eqp == instr)
	  {
		printf("Equal sign in %s is at the beginning\n", instr);
		return -4; /* equal sign at beginning */
	  }
	lenn = eqp - instr;	lennp2 = lenn + 2;
	for (cp = eqp + 1; '\0' != *cp; cp++) ;
	if (cp == eqp + 1)
	  {
		printf("Equal sign in %s is at the end\n", instr);
		return -5; /* equal sign at end */
	  }
	if (NULL == (onam = mallop(lennp2, "onam in envparse")))
	  {
		printf("Can't reserve %d bytes for onam in envparse\n",
			(int)lennp2);
		return EALLOC;
	  }
	zeromem((void *)onam, lennp2);
	lenv = cp - eqp;	lenvp2 = lenv + 2;
	if (NULL == (oval = mallop(lenvp2, "oval in envparse")))
	  {
		printf("Can't reserve %d bytes for oval in envparse\n",
			(int)lenvp2);
		return EALLOC;
	  }
	zeromem((void *)oval, lenvp2);
	for (ocp = onam, cp = instr; cp < eqp; cp++) *ocp++ = *cp;
	*ocp++ = '\0';	cp++;
	for (ocp = oval; '\0' != *cp; cp++) *ocp++ = *cp;
	*ocp++ = '\0';
	*namstr = onam;	*valstr = oval;	return EOK;
}

static int xgdo(int azc, char *azv[], char *xgve[],
	int (*xgap[])(int ac, char *av[]))
{ /*	This looks to see if *azv is one of the commands listed
	in the array "xgve". If it is, we treat it as an X-GEN command;
	if  it isn't, we attempt to fork and exec the command in azv. */

	int		ic, lnstatus, ninterp, sbac, status;
	char		*cp, *lonam, *loval, *prevval, *teval, **sbav, **xgv;
	char		sbuf[256], verb[256];
	int		(*thisapp)(int ac, char *av[]);
	FILE		*fpte;
	static char	*diagstr;
	static size_t	diaglen, olddiaglen;
	static char	*stopverbs[] = { "exit", "quit", "stop", NULL };
	static char	*xgftypes[] = {
	 "amask", "centroids", "multiref", "uparams", "uspline",
	 "urefls", "wmask", "frames", "missing", "mulist", "scalein",
	 "udump", "xasca", "xcns", "xgencmd", "xgencom", "xlog", "xsca",
	 "xshel", "xxplor", NULL } ;
	static char	dimsg[] = "diagstr in xgdo";

	if (azc == 0)	/* special call to clear the diagstr string */
	  {
		memzap((AVOID **)&diagstr, dimsg);	olddiaglen = 0;
		return EOK;
	  }
	if ((azc < 1) || (azv == NULL) || (*azv ==  NULL)) return 0;
	if ('#' == **azv) return 0;	/* ignore comments */
	 /* if an argument begins with a sharp, everything else is comment */
	for (xgv = azv + 1; NULL != *xgv; xgv++)
		if ('#' == **xgv)
		  {
			azc = xgv - azv;	*xgv = NULL;	break;
		  }
	zeromem(verb, 255);	strcpy(verb, azv[0]);
	/* handle exit, quit, stop, end, help, version, etc. */
	for (cp = &(verb[0]); '\0' != *cp; cp++)
		if (('A' <= *cp) && (*cp <= 'Z')) *cp += 'a' - 'A';
	for (xgv = stopverbs; NULL != *xgv; xgv++)
		if (!strcmp(*xgv, verb)) exit(EOK);
	if ((!strcmp(verb, "help")) || (!strcmp(verb, "--help")) ||
	 (!strcmp(verb, "prhelp")) || *verb == '?')
	  {
		if (NULL != azv[1]) helpapp(azv[1]);
		else helpapp("xgen");	return -1;
	  }
	if ((!strcmp(verb, "version")) || (!strcmp(verb, "--version")) ||
	 (!strcmp(verb, "prversion")))
	  {
		if (NULL != azv[1]) versionapp(azv[1]);
		else versionapp("xgen");	return -2;
	  }
	if (!strcmp(verb, "commands"))	/* print list of commands */
	  {
		list_xgcommands(stdout, xgve);
		status = -3;	return status;
	  }
	if (!strcmp(verb, "setenv")) /* env vbl -- csh style */
	  {
		if (azc < 3)
		  {
			printf("Usage: setenv name value\n");
			status = -8;	return status;
		  }
		status = set_logical(azv[1], azv[2]);
		status = -4;	return status;
	  }
	if (!strcmp(verb, "export")) /* bash style */
	  {
		if (azc < 2)
		  {
			printf("Usage: export name=value\n");
			status = -9;	return status;
		  }
		errno = 0;
		if (EOK == (status = envparse(azv[1], &lonam, &loval)))
		  {
			if (1 > (status = set_logical(lonam, loval))) printf(
 "attempt to set %s to %s resulted in status = %d, errno = %d\n",
				 lonam, loval, status, errno);
			else if (status == 2) printf(
 "Environment variable %s was already %s; unchanged\n", lonam, loval);
			else printf("Environment variable %s set to %s\n",
				lonam, loval);
		  }
		else printf(
  "Can't parse %s as environment assignment: status, errno = %d, %d\n",
		 azv[1], status, errno);
		status = -5;	return status;
	  }
	if ((!strcmp(verb, "unset")) || (!strcmp(verb, "unsetenv")))
	  {
		if (azc < 2)
		  {
			printf("Usage: %s envname\n", verb);
			status = -10;	return status;
		  }
		if (2 == (status = unset_logical(azv[1]))) printf(
		 "Environment variable %s is unset\n", azv[1]);
		else if (1 == status) printf(
		 "Environment variable %s wasn't set to begin with\n",
			azv[1]);
		status = -6;	return status;
	  }
	/* handle constructions like WORK=/Users/ahoward/tx/hgmad/xgena */
	if (EOK == envparse(*azv, &lonam, &loval))
	  {
		if (1 > (status = set_logical(lonam, loval)))
		  {
			printf(
 "attempt to set %s to %s resulted in status = %d, errno = %d\n",
			 lonam, loval, status, errno);
		  }
		else if (status == 2)
		 printf("%s was already %s; not changed\n", lonam, loval);
		else printf("Environment variable %s set to %s\n",
			lonam, loval);
		status = -6;	return status;
	  }
	if ((!strcmp(verb, "cd")) || (!strcmp(verb, "pushd"))) /* chdir */
	  {
		if (azc < 2)
		  {
			printf("Usage: %s envname\n", verb);
			status = -11;	return status;
		  }
		errno = 0;
		if (0 != chdir(*(azv + 1)))
		  {
			printf("Cannot chdir to %s: error = %d\n",
			 *(azv + 1), errno);	status = -16; return status;
		  }
		if (NULL == (teval = mallop(1030, "teval in xgdo")))
		  {
			printf(
 "Warning: can't allocate memory for directory buffer\n");
			status = EALLOC; return status;
		  }
		zeromem(teval, 1029);
		mygetwd(teval);
		printf("Current working directory is %s\n", teval);
		if (1 > (status = set_logical("PWD", teval)))
		  printf("Warning: environment variable PWD not set\n");
		memzap((AVOID **)&teval, "teval in xgdo");
		status = EOK;	return status;
	  }
	if (!strcmp(verb, "source")) /* read commands from a (.com) file */
	  {
		if (azc < 2)
		  {
			printf("Usage: %s envname\n", verb);
			status = -12;	return status;
		  }
		if (NULL == (fpte = xgopen(*(azv + 1), "r")))
		  {
			printf("Cannot open %s as command file\n", *(azv + 1));
			status = -13;	return status;
		  }
		zeromem(sbuf, 255);	(void)xgprogname(NULL, 2);
		while (NULL != fgets(sbuf, 255, fpte))
		     {
			if ((NULL != (xgv = str2acav(sbuf, &sbac))) &&
			 (sbac > 0))
			  {
				sbav = xgv; if (NULL == sbav[sbac-1]) sbac--;
				errno = 0;
				if (EOK != (status = globexpand(&sbac, &sbav)))
				 printf(
	  "Global expansion failed: status, errno =%d %d\n", status, errno);
				if (NULL != *sbav)
					status = xgdo(sbac, sbav, xgve, xgap);
			  } /* end of "if (NULL != (sbav ..." clause */
			else printf("memory-allocation error, subcommand %s\n",
				sbuf);
			zeromem(sbuf, 255);
		     }
		(void)fclose(fpte);	status = -7;
		(void)xgprogname(NULL, 3);	return status;
	  }
	/* if we haven't returned yet, we look to see if the verb
	 is on the list of legitimate X-GEN commands known to this app. */
	for (ic = 0, xgv = xgve; NULL != *xgv; xgv++, ic++)
		if (!strcmp(*xgv, verb)) break;
	if (NULL != *xgv) /* found a match in the list of X-GEN commands */
	  {
		if (NULL != azv[1]) /* is there at least one argument? */
		  {
			zeromem(verb, 255);
			strcpy(verb, azv[1]); /* copy arg 1 to `verb' */
			for (cp = verb; '\0' != *cp; cp++)
				if (('A' <= *cp) && (*cp <= 'Z'))
					*cp += 'a' - 'A';
			  if ((!strcmp(verb, "--help")) ||
			  (!strcmp(verb, "help")))
			    {
				helpapp(*xgv);	return -3;
			    }
			  else if ((!strcmp(verb, "--version")) ||
			  (!strcmp(verb, "version")))
			    {
				versionapp(*xgv);	return -4;
			    }
		  }
		thisapp = xgap[ic]; /* pick out the appropriate mainline */
		/* construct the diagnostic string beforehand in case the
		 application itself manipulates the azv strings */
		for (diaglen = 56, ic = 0; ic < azc; ic++)
			diaglen += strlen(azv[ic]) + 2;
		if (diaglen < 128) diaglen = 128;
		if ((NULL != diagstr) || (diaglen > olddiaglen))
		  {
			if (NULL != diagstr) memzap((AVOID **)&diagstr, dimsg);
			if (NULL == (diagstr = (char *)mallop(diaglen, dimsg)))
			  {
				printf("Can't display memory for diagstr\n");
				return EALLOC;
			  }
			olddiaglen = diaglen;
		  }
		strcpy(diagstr, "X-GEN command");
		for (ic = 0; ic < azc; ic++)
		   {
			strcat(diagstr, " ");	strcat(diagstr, azv[ic]);
		   }
		strcat(diagstr, " exited with status");
		status = xgappc(azc, azv, thisapp); /* launch it! */
		sprintf(diagstr + strlen(diagstr), " %d\n", status);
		fputs(diagstr, stdout);
		return status;
	  }
	/* see if the user specified a known X-GEN file type, e.g. UPARAMS.
	 If so, and if no arguments are given, we treat this as equivalent
	 to "xgen xdump <filetype>". If a file type is given followed by
	 "help" or "version", then prhelp(filetype) or prversion(filetype)
	 is called. Finally, if a file type is given, followed by a file
	 name, then we dump the contents of that file name */
	for (ic = 0, xgv = xgftypes; NULL != *xgv; xgv++, ic++)
	   {
		if (!strcmp(*xgv, verb)) break;
	   }
	if (NULL != *xgv) /* yes, this is one of the file types */
	  {
		if (azc > 1)
		  {
			if ((!strcmp(azv[1], "help")) ||
			 (!strcmp(azv[1], "--help")) ||
			 (!strcmp(azv[1], "HELP")) ||
			 (!strcmp(azv[1], "--HELP")))
			  {
				helpapp(*xgv);	status = EOK; return status;
			  }
			if ((!strcmp(azv[1], "version")) ||
			 (!strcmp(azv[1],"--version")) ||
			 (!strcmp(azv[1],"VERSION")) ||
			 (!strcmp(azv[1], "--VERSION")))
			  {
				versionapp(*xgv); status = EOK; return status;
			  }
		      }
		for (cp = verb; '\0' != *cp; cp++)
		   if (('a' <= *cp) && (*cp <= 'z')) *cp -= 'a' - 'A';
		prevval = teval = trname(verb);
		if ((azc > 1) && (NULL != prevval) && (strcmp(prevval,azv[1])))
		  {
			teval = azv[1];
			if (1 > (lnstatus = set_logical(verb, teval))) printf(
 "Warning: set_logical(%s,%s) returned %d\n", verb, teval, lnstatus);
			else  if (2 == lnstatus) printf(
 "Environment variable %s was already %s; unchanged\n", verb, teval);
			else printf("%s set to %s\n", verb, teval);
		  }
		if (ic >= 7)	/* these are text files: print top few lines */
		  {
			if (NULL == (fpte = xgopen(teval, "r")))
			  {
				printf("Can't open %s as %s file\n",
					teval, verb);
				status = EOPENERR;
			  }
			else {
				printf(
				 "First few lines of file %s of type %s:\n",
					teval, verb);
				for (ic = 0; ic < 8; ic++)
				   {
					zeromem((AVOID *)sbuf, 256);
					if (NULL == fgets(sbuf, 255, fpte))
						break;
					fputs(sbuf, stdout);
				   }
				(void)fclose(fpte);	fpte = NULL;
				status = EOK;
			     }
		  }
		else { /* xgen knows about xdump; xgentools doesn't */
			if (strcmp(xgve[42], "xdump"))
			  {
				printf(
				 "Can't dump output from this application\n");
				status = EOK;	return status;
			  }
			azv[0] = "xdump";	azv[1] = verb;	azv[2] = NULL;
			azc = 2;	thisapp = xgap[42];
			status = xgappc(azc, azv, thisapp); /* launch it! */
			printf("X-GEN command "); /* Summarize what happened */
			for (ic = 0; ic < azc; ic++)
			   {
				if (ic > 0) printf(" ");
				printf("%s", azv[ic]);
			   }
			printf(" exited with status %d\n", status);
		     }
		/* restore old meaning of environment variable */
		if (prevval != teval)
		  {
			if (1 > (lnstatus = set_logical(verb, prevval)))
			 printf("Warning: set_logical(%s,%s) returned %d\n",
				verb, prevval, lnstatus);
			else if (2 == lnstatus) printf(
 "Environment variable %s was already %s; unchanged\n", verb, teval);
			else printf("%s reset to %s\n", verb, teval);
		  }
		return status;
	  }
	/* if this isn't an X-GEN command, try to fork and exec to [command].
	 If the arguments to [command] can be handled readily in the parse,
	 we do this directly with xgfork; if they can't,
	 we use xgffork to write a one-line shell script to do it */
	errno = 0;
	for (ninterp = ic = 0, xgv = azv; ic < azc; ic++, xgv++)
		ninterp += all1command(*xgv);
	if (ninterp >= azc) status = xgfork(azv, 1);
	else status = xgffork(azv); /* complex command: use 1-line file */
	if (EOK != status)
	  {
		printf("process fork returned %d; errno = %d\n", status, errno);
		if (!errno) errno = status;	perror(NULL);
		status = errno = 0;	/* want error cleared at this point */
	  }
	return EOK;
}

int xgcomargs(int arc, char *arv[], char *xgve[],
	int (*xgaps[])(int ac, char *av[]))
{ /*	Manage command arguments for `xgen' or `xgentools' */
	int	azc, isterm, status;
	char	**azv, **lazv, xgbu[260];

	(void)xgprogname(*arv, 4);	/* remember the current program name */
	status = 0;	/* start out without errors */
	/* we look to see if there's an argument */
	if ((arc < 2) || (arv[1][0] == '-'))
	  {
		if (arc > 1) /* interpret flag arguments */
		  {
			if ((!strcmp(arv[1], "--help")) ||
			 (!strcmp(arv[1], "--HELP")) ||
			 (arv[1][1] == 'h') || (arv[1][1] == 'H'))
			  {
				if (arc > 2) helpapp(arv[2]);
				else helpapp(*arv);	return EOK;
			  }
			if ((!strcmp(arv[1], "--version")) ||
			 (!strcmp(arv[1], "--VERSION")) ||
			 (arv[1][1] == 'v') || (arv[1][1] == 'V'))
			  {
				if (arc > 2) versionapp(arv[2]);
				else versionapp(*arv);	return EOK;
			  }
			if ((!strcmp(arv[1], "--commands")) ||
			 (!strcmp(arv[1], "--COMMANDS")) ||
			 (arv[1][0] == 'c') || (arv[1][0] == 'C'))
			  {
				list_xgcommands(stdout, xgve);
				return EOK;
			  }
		  }
		isterm = isatty(0);	/* are we doing this from a terminal?*/
		if (isterm)
		  {
			printf(
 "Welcome to the interactive line-oriented interface to X-GEN\n");
			printf(
 "Respond to each prompt with an X-GEN command, possibly with arguments\n");
		  }
		while (1)
		     {
			if (isterm)
			  {
				printf("x> ");	xeflush(stdout);
			  }
			zeromem(xgbu, 256);
			if (NULL == fgets(xgbu, 255, stdin))
			  {
				printf("Unable to get data from stdin\n");
				break;
			  }
			/* ignore comments and empty lines */
			if (('\n' == *xgbu) || ('#' == *xgbu)) continue;
			(void)xgprogname("xgen", 4);
			errno = 0;
			if (NULL == (lazv = str2acav(xgbu, &azc)))
			  {
				if ('\0' != xgbu[0]) printf(
				 "Memory-allocation error for %s\n", xgbu);
			  }
			else if (azc > 0)
			  {
				azv = lazv;
				if (NULL == azv[azc-1]) azc--;
				if ((azc > 1) && (NULL != *(azv + 1)))
				  {
					azc--;	azv++;
				  }
				errno = 0;
				if ((EOK == status) && (EOK != (status =
				 globexpand(&azc, &azv)))) printf(
	  "Global expansion failed: status, errno =%d %d\n", status, errno);
				if (NULL != *azv)
					status = xgdo(azc, azv, xgve, xgaps);
			  }
		     }
	  }
	else status = xgdo(arc - 1, arv + 1, xgve, xgaps);
	if (0 > status) status = 0;	/* help and version are ok */
	xgdo(0, NULL, NULL, NULL);	/* clear the diagstr stuff */
	return status;
}

/* end of xgenparse.c */
