#! /usr/bin/python
# basic X-GEN processing script in Python
# Change record:
# Syntax:
# xgenproc.py [-dnsvw] [-cfhilmpxyz<val>] [dirn]
# i.e. d,n,s,v,w are Booleans and the other - options all take values
# and an option that doesn't begin with a minus is a directory name
# Change record:
#	16 Sep 2005: built from xgenproc
#
#	Copyright (c) 2002, Illinois Institute of Technology
#	X-GEN: Crystallographic Data Processing Software
#	author: Andrew J Howard, BCPS Department, IIT
# See the file "LICENSE" for information on usage and redistribution
#	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

import os
import sys
import fpformat
import string
import time
import commands
import getopt
import glob
import gzip

def mygetenv(stri) :	# some Pythons don't have an os.getenv
	if stri :
		loce = os.environ	# find the local environment
		for evar in loce :	# look through list of variables
			if evar == stri : # is this the variable?
				return loce[stri]	# return its value
		return None
	else :	# faulty strings should be handled smoothly
		return None

def cleanrun(xgstring) :	# cleanly run an X-GEN command
	thistuple = string.split(xgstring, ' ')	# create the string tuple
	if thistuple[0] != 'xgen' :	# prepend 'xgen ' if needed
		thisstr = 'xgen ' + xgstring
		thistuple = string.split(thisstr, ' ')
	pid = os.fork()		# fork off a child
	if pid == -1 :		# failure to fork
		print 'attempt to fork failed'
		return -1
	elif pid == 0 :		# in child process: exec the xgen tuple
		os.execvp('xgen', thistuple)	# normally doesn't return
		print 'failed execvp of' + xgstring # if it does, exit
		_exit(127)
	else :	# This happens in the parent process
		# code adapted from C code in xglib/xgenstr.c/xgfork()
		errno = 0
		while (1) :
			status = os.waitpid(pid, 0)
			tstatus = status[1]
			estatus = tstatus >> 8
			if -1 == estatus :
				if -1 == errno :
					print 'Failure to wait'
					return -1
			elif os.WIFEXITED(estatus) :
			  esout = os.WEXITSTATUS(estatus)
			  if esout != 0 :
				print 'Command exited: status ' + str(esout)
			  return os.WEXITSTATUS(estatus)
			else :
			  print 'Process unexited; status = ' + str(estatus)
			  return estatus

def gzipFile(fileName):	# compress one file
	if fileName.endswith('.gz') :	# see if filename ends with .gz
		print 'Refusing to gzip '+fileName+'; it is already gzipped'
        else :	# open input and output, loop through reading + writing, close
		input = open(fileName, 'rb')	# open input
	        output = gzip.open((fileName + '.gz'), 'wb')	# open output
	        readWriteLoop(input, output)	# read and write
	        os.unlink(fileName)		# delete input
		print 'File '+fileName+' has been compressed'

def gunzipFile(fileName): # uncompress one file
	if fileName.endswith('.gz') :	# make sure it's really compressed
	        input = gzip.open(fileName, 'rb') # open it for uncompression
		outputName = string.strip(fileName, '.gz')
	        output = open(outputName, 'wb')	# output output file
	        readWriteLoop(input, output)	# read, write data
	        os.unlink(fileName)		# delete input
		print 'File '+fileName+' has been uncompressed'
	else :	# if it isn't gzipped, complain and quit
		print 'Cannot gunzip '+fileName+'; it is already gunzipped'

def zipfiles(stri) : 			# compress a bunch of files
	zippables = glob.glob(stri)	# look for input files
	for zippable in zippables :	# loop through inputs
		if zippable.endswith('.gz') :	# ignore .gz files
			print 'File '+zippable+' is already compressed'
		else :			# compress all other inputs
			gzipFile(zippable)

def unzipfiles(stri) : 			# uncompress a bunch of files
	unzippables = glob.glob(stri)	# look for input files
	for unzippable in unzippables :	# loop through inputs
		if unzippable.endswith('.gz') : # must end with .gz
			gunzipFile(unzippable)
		else :
			print 'File '+unzippable+' has not yet been compressed'

def readWriteLoop(input, output) :	# loop for reading & writing
        data = input.read(65536)	# read 2^16 bytes at a time
        while data:			# manipulate those
                output.write(data)	# write 2^16 bytes
                data = input.read(65536)	# read some more
        input.close()			# close the input
        output.close()			# close the output

def evset (evnam, evval) :		# set an environment variable
	oldval = mygetenv(evnam)	# is it already there?
	if oldval :
	  if oldval != evval :	# re-set environment vbl
		print 'Re-setting environment variable '+evnam+' to '+evval
		os.environ[evnam] = evval
	  else :	# this else will disappear after debugging
	    print 'Variable '+evnam+' = '+evval+' unchanged'
	else :
		print 'Setting environment variable '+evnam+ ' to '+evval
		os.environ[evnam] = evval

def main (record_list, argv ) :
	docompress = 0	# default is _not_ to compress the data
	stepflip = 0	# default is to leave the sign of the stepsize intact
	datadir = '.'	# default data directory is the current one
	spaceg = 0	# default is to let the program determine the spacegroup
	doregions = 0	# default is don't look for regional offsets
	maxspots = 400	# default: keep looking for spots until we find 400
	multicentroid = 0	# default: get centroids from a single run
	multiwidth = 1	# width of multiple-centroid image ranges
	xcal = -1	# don't run calibrate unless x and y are specified
	ycal = -1	# see above
	specfile = ''	# don't specify a filename
	specprefix = ''	# don't specify a prefix
	hifram = -1	# don't specify a maximum frame number
	dohelp = 0	# don't give out help
	doversion = 0	# don't tell us the version number
	dorecenter = 1	# recenter things right after auto-indexing
	lofram = 10000	# don't specify a minimum frame number
	minremap = 2000	# RMSError(X,Y) * 10000 for remapping in integrate
	poltype = 3	# default polarization type = fully polarized
	options, args = getopt.getopt(argv[1:],
	 'c:df:h:i:j:kl:m:np:suvw:x:y:z:',
	 [	'maxspots=','docompress','specfile=','hifram=','spaceg=',
		'minremap=','doregions','lofram=','multicentroid=',
		'norecenter', 'specprefix=','specflip','help','version',
		'multiwidth=','xcal=','ycal=','polarization_type='
	 ])
	for o,a in options:
		if o in ( '-c', '--maxspots') :
			maxspots = int(a)
		if o in ( '-d', '--docompress') :
			docompress = 1
		if o in ( '-f', '--specfile') :
			specfile = a
		if o in ( '-h', '--hifram') :
			hifram = int(a)
		if o in ( '-i', '--spaceg') :
			spaceg = int(a)
		if o in ( '-j', '--minremap') :
			minremap = int(a)
		if o in ( '-k', '--minregions') :
			doregions = 1
		if o in ( '-l', '--lofram') :
			lofram = int(a)
		if o in ( '-m', '--multicentroid') :
			multicentroid = int(a)
		if o in ( '-n', '--norecenter') :
			dorecenter = 0
		if o in ( '-p', '--specprefix') :
			specprefix = a
		if o in ( '-s', '--specflip') :
			specflip = 1
		if o in ( '-u', '--help'):
			dohelp = 1
		if o in ( '-v', '--version'):
			doversion = 1
		if o in ( '-w', '--multiwidth'):
			multiwidth = int(a)
		if o in ( '-x', '--xcal') :
			xcal = int(a)
		if o in ( '-y', '--ycal') :
			ycal = int(a)
		if o in ( '-z', '--polarization_type') :
			poltype = int(a)
	if (len(args) > 0) :	# remaining arg is the desired data directory
		datadir = args[0]
	if dohelp != 0:		# if user wants help, print the help and quit
		os.system('xgen prhelp xgenproc.py')
		return 0
	if doversion != 0:	# if user wants version #, print it and quit
		os.system('xgen prversion xgenproc.py')
		return 0
	
	# make sure the data directory exists and is a directory
	if os.access(datadir, 4) == False:
		print 'Specified data directory ' + datadir + ' is unreadable'
		return 12

	if datadir != '.' :	# change directory if needed
		print 'Changing directory to ' + datadir
		os.chdir(datadir)	# if the directory exists, cd to it
	
	# create a processing directory. If a directory xgen already exists,
	# use xgena; if xgena exists, use xgenb; and so on
	workb = 'xgen'	# base of the work names
	if os.access(workb, 4) == False :
		workdir = workb
	else :
		suffix = 'a'
	lsufstr = 'b c d e f g h i j k l m n o p q r s t u v w x y z ' 
	lsuf = string.split(lsufstr, ' ')
	for nextsuffix in lsuf :
		workdir = workb + suffix
		# nextdir = workb + nextsuffix
		acval = os.access(workdir, 4)
		if acval == False :
			break
		else :
			suffix = nextsuffix
	tocreate = datadir + '/' + workdir
	print 'Creating directory ' + tocreate
	u = os.mkdir(workdir, 0755)
	u = os.chdir(workdir)
	paramname = workdir + '/params'
	u = os.mkdir('params', 0755)
	
	xgnowstr = 'xgnow'	# create the run string for xgnow
	if hifram > 0 :		# add -h flag if specified
		xgnowstr = xgnowstr + ' -h' + str(hifram)
	if lofram < 9999 :	# add -l flag if specified
		xgnowstr = xgnowstr + ' -l' + str(lofram)
	if specfile != '' :	# add specified data file
		xgnowstr = xgnowstr + ' -x' + specfile
	if specprefix != '' :	# add specified data prefix
		xgnowstr = xgnowstr + ' -i' + specprefix
	cleanrun(xgnowstr)	# invoke the xgnow command
	
	# do the equivalent of `source'ing the command file
	comfiles = glob.glob('*.com')	# look for command files (should be 1!)
	for comfil in comfiles :	# loop through command files
		fcom = open(comfil, 'r')	# open it
		while (1) :		# loop through lines in it
			cfstr = fcom.readline() # record contents of line
			if len(cfstr) < 2 :	# ignore short lines
				break
			wordsincfstr = string.split(cfstr, ' ') # make tuple
			if len(wordsincfstr) < 2 :	# ignore short tuples
				break
			# look at wordsincfcstr[1] for environment assignment
			elt = string.rstrip(wordsincfstr[1], '\n') # elim \n
			elements = string.split(elt, '=')	# make tuple
			if len(elements) >= 2 :	# should look like A=value
				evset(elements[0], elements[1]) # set environ
		fcom.close()	# close the file when we're done
		print 'environment variables from ' + comfil + ' set up  '
		# meanwhile, we should write the name of this file
		dirna = mygetenv('PWD')
		funame = dirna + '/' + comfil + '\n'
		listfil = mygetenv('HOME') + '/xgensave.sh'
		print 'Writing ' + funame + ' to ' + listfil
		flis = open(listfil, 'w')
		flis.write(funame)
		flis.close()
		break		# we really want the first .com file only

	# as long as we're dealing with environment variables,
	# we set the polarization type too
	if ((1 <= poltype) and (poltype <= 5)) :
		os.environ['POLARIZATION_TYPE'] = str(poltype)
		print 'Polarization type ' + str(poltype) + ' selected'

	# calibrate the detector if needed
	if xcal > 0 and ycal > 0 :
		calstr = 'calibrate' + ' -j ' + str(xcal) + ' ' + str(ycal)
		cleanrun(calstr)
	
	spotstr = 'spots'	# handle the spots run
	if maxspots > 0 :	# specify min # of spots usable
		spotstr = spotstr + ' -c' + str(maxspots)
	if multicentroid > 0 :	# get spots from multiple runs
		oldcen = mygetenv('CENTROIDS') # save name of orig. file
		divisions = multicentroid	# number of divisions of data
		if (divisions > 9):		# must not be greater than 9
			divisions = 9
		totframes = 0			# count total frames
		framename = mygetenv('FRAMES')	# get $FRAMES name
		fil = open(framename, 'r')	# open it
		lineno = 0			# initialize count of lines
		firstint = 1
		lastint = 1
		while (1) :			# loop through lines in FRAMES
			lin = fil.readline()	# read a line
			if len(lin) < 2 :    	# quit reading at EOF
			        break
			imno = string.split(lin, '/') # find integer part
			imva = int(imno[0])	# make an integer out of it
			if lineno <= 0 :
				firstint = imva	# remember 1st one
			lineno = lineno + 1	# increment line count
		lastint = imva			# remember final frame number
		fil.close()			# close the $FRAMES file
		totframes = lastint + 1 - firstint	# count frames
		divm1 = 1
		if divisions > 1 :
			divm1 = divisions - 1	# for calculating lengths
		calclen = totframes / divm1
		ndone = 1
		next = firstint
		while ndone <= divisions :	# loop, getting spots
			# construct the name of the output file
			cname = 'spots' + str(ndone) + '.cen'
			os.environ['CENTROIDS'] = cname	# set environment
			loclast = next + multiwidth - 1	# final image in run
			if loclast > lastint :
				loclast = lastint
			# construct string to execute spots command
			thisstr = spotstr + ' ' + str(next) + ' ' + str(loclast)
			cleanrun(thisstr)
			next=next+calclen	# increment image number
			if next > lastint :
				break
			ndone=ndone+1
		# now we restore the names of the centroids files
		os.environ['CENTROIDS'] = oldcen
		os.environ['CENOUT'] = oldcen
		cenmargs = glob.glob('spots?.cen')
		cleanstr = 'cenmer'
		for carg in cenmargs :
			cleanstr = cleanstr + ' ' + carg
		if cleanstr == 'cenmer' :
			print 'No centroids files available for merging'
			return 0
		cleanrun(cleanstr)
		for carg in cenmargs :
			os.unlink(carg)
	else :
		cleanrun(spotstr)
	
	refstr = 'refall' + ' -i' + str(spaceg)
	if dorecenter == 0 :
		refstr = refstr + ' -n'
	if stepflip != 0 :
		refstr = refstr + ' -f'
	cleanrun(refstr)
	
	integstr = 'integrate' + ' -j' + str(minremap)
	if docompress > 0 :
		integstr = integstr + ' -d'
	if doregions > 0 :
		integstr = integstr + ' -g12'
	cleanrun(integstr)
	
	zq = glob.glob('*.bgi*')	# get rid of background images
	for zz in zq :
		os.unlink(zz)
	# transfer the .upr files to a subdirectory
	curuparams = mygetenv('UPARAMS')
	newuparams = curuparams + '.s'
	os.rename(curuparams, newuparams)
	allpar = glob.glob('*.upr')
	for zz in allpar :
		newname = 'params/' + zz
		os.rename(zz, newname)
	os.rename(newuparams, curuparams)
	
	cleanrun('reduce')	# convert & filter output of `integrate'
	
	cleanrun('scalem')	# scaling, outlier rejection, ascii output
	
	zq = glob.glob('*.*,[1-9]')
	for zz in zq :
		os.unlink(zz)
	zipfiles('params/*')	# compress the .upr files in params/
	zipfiles('*.[asuw]*')	# compress some binary files
	zipfiles('*.mr*')	# compress MULTIREFs 
	zipfiles('*.ce*')	# compress CENTROIDSs
	unzipfiles('*.asc.gz')	# uncompress SCALEPACK output with anomalous
	unzipfiles('*.sca.gz')	# uncompress SCALEPACK output, no anomalous
	return 0

# unlabeled operations that call everything else
nar = len(sys.argv)
main(nar, sys.argv)
