/* actmask.c -- utilities for dealing with the active area of the detector

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions; reorganize use of "{"
	17 Jun 2006: act_pixel, act_refl replaced by activmask calls
	 5 Dec 2005: reorganized: activmask replaces several older functions;
	 imaverage handles memory for averaged image; no more static globals
	20 Nov 2005: inline function declarations & function prototypes
	17 Jun 2005: reserve_active: call to qerror(... corrected
	 6 Sep 2004: cosmetic changes
	21 May 2004: correct bug in sum_row() that caused incorrect
		displays of the active & inactive pixels in a row
	25 Jan 2004: fpact now local rather than static global
	21 Jan 2004: license message inserted
	 3 Jan 2004: replace external global **pactive with static global;
		create kill_pactive to handle memory for **pactive;
		replace shcalloc calls with callop calls
	10 Dec 2003: eliminate global references to pactp
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	 5 Jan 2003: reorganized source-file header
	28 Dec 2002: fix bug in default_mask():
		upper limit of px is nbigcols, not 64.
	25 Mar 2002: removed MSI attachments
	31 Aug 1995: default_mask: use lwidth,lheight,lowidth,loheight
		as local variables
	31 May 1995: "register" -> "register int" throughout
	 2 Mar 1995: shared-memory alloc and free
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 1 Dec 1994: diagnostic output allows for detector (X,Y) > 999
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version
	14 Dec 1992: revisions for Polydet
	19 Sep 1990: get rid of ugly ifdefs associated with 68000 systems
	 1 Mar 1988: option to move active area inward.
	30 Apr 1987: include database manipulator routines
	 1 Nov 1986: original code.

   Purpose: to provide routines used in processing the active-area
	bitmask--both to create it and to use it.
   Cross references:

*************************************************************************/
#define		NEED_ERRNO	1
#include	"xgen.h"
#define		MIN_NEIGHBORS	2
#define		MAX_NEIGHBORS	4

void head_results(FILE *fp)
{	/* this puts out a header for the summary of results */
	fprintf(fp,"\tResults of dead-pixel mask construction:\n");
	fprintf(fp,
 "\tRows in which all pixels from 1 to (width-2) are active won't be listed\n");
	fprintf(fp,
 " Row    # of active    Ranges of active Pixels within the Row:\n");
	fprintf(fp,
 "   #  pixels in row  min max  min max  min max  min max  min max...\n");
}

u_char **activmask(int x, int y, int opno)
{ /*	This performs various functions related to the active mask,
	depending on the value of opno:
	opno	function
	  0	delete memory for the active-pixel list
	 01	return pointer to the active-pixel list
	 03	unconditionally reserve memory for that pointer
	 05	unconditionally read active-pixel list into memory from disk
	 07	reserve memory and read into it
	011	write current contents out
	021	define default list rather than reading one
	041	return pactive iff (x,y) is in the active area, else NULL
	0101	return pactive iff (x,y) and its neighbors are in
		 the active area, else NULL.
	0201	gather the detector-dimension information from the USPLINE
		 file and then construct the active-pixel list from
		 a file or an automated mask
	Except with the 041 and 0101 cases, the function returns a pointer
	 to a pointer to the active mask.
	It returns NULL and sets errno if errors occur.  */

	register int	yu, px, ux;
	int		finh, finw, pxleft, pxright;
	u_char		*ucsp;
	unsigned	masksize, nbigcols, row;
	double		xcenter, ycenter, xradius, yradius;
	double		discrim, xleftbord, xrightbord;
	static u_char	*pactp;
	static u_char	**pactive;
	static int	inwid, inht;
	static FILE	*fpact;

	errno = 0;	/* clear errno before beginning */
	if (opno == 041) /* see if this pixel is in the active area */
	  { /* we need to have the active-area memory set up */
		if ((NULL == pactive) && (NULL == activmask(0,0,0201)))
			return NULL;
		return (pactive[y][x / 8] & (1 << (x % 8))) ?
			pactive: NULL;
	  }
	if (opno == 0101) /* see if this pixel and its neighbors are active */
	  { /* need to have the active-area avail */
		/* The box around which we check is shaped like this:	
		  * . . . *
		* . . . . . *        @------>  X
		. . . . . . .        |
		. . . @ . . .        |
		. . . . . . .        V
		* . . . . . *
		  * . . . *          Y	*/
		if ((NULL == pactive) && (NULL == activmask(0, 0, 0201)))
			return  NULL;
		if ((x < 3) || (x > inwid - 4) || (y < 3) || (y > inht - 4))
			return 0;
		return (activmask(x-2, y-3, 041) && activmask(x+2, y-3, 041) &&
		 activmask(x-2, y+3, 041) && activmask(x+2, y+3, 041) &&
		 activmask(x-3, y-2, 041) && activmask(x+3, y-2, 041) &&
		 activmask(x-3, y+2, 041) && activmask(x+3, y+2, 041) &&
		 activmask(x,  y, 041)) ? pactive : NULL;
	  }
	if (opno == 0201) /* measure detector dimensions and get active area */
	  {
		if ((NULL == pactive) || (inwid <= 0) || (inht <= 0))
		  {
			if (EOK != find_dimension(&finw, &finh, 0, NULL))
				return NULL;
			inwid = finw;	inht = finh;
			/* get detector dimensions */
			if (NULL == activmask(finw, finh, 03)) return NULL;
			/* read active area from file */
			if (NULL == activmask(finw, finh, 05))
			  {	/* if that fails, get it by a model */
				if (NULL == activmask(finw, finh, 021))
						return NULL;
			  }
		  }
		return pactive;
	  }
	if (!(opno & 01))	/* delete the memory for the pactive list */
	  {
		if (NULL != pactive) /* ... including what it points to */
		  {
			memzap((AVOID **)pactive, "*pactive in globalkill");
			memzap((AVOID **)&pactive, "pactive in globalkill");
		  }
		inwid = inht = 0;	return NULL;
	  }
	if (opno == 01) return pactive; /* just return a pointer to it */
	inwid = x;	inht = y;
	nbigcols = (unsigned)inwid/ 8;	if (inwid % 8) nbigcols++;
	if (opno & 02)	/* unconditionally reserve memory and return it */
	  {
		(void)activmask(inwid, inht, 0); /* kill old memory */
		inwid = x; inht = y; /* restore after setting them to 0 */
		if (qerror((AVOID *)(pactive = (u_char **)callop((unsigned)
		 inht, sizeof (u_char *), "pactive in reserve_active")),
		 "reserve memory for rows of active mask")) return NULL;
		masksize = ((unsigned)inht) * nbigcols;	pactp = NULL;
		if (qerror((AVOID *)(pactp = (u_char *)callop(masksize, 1,
		 "pactp in reserve_active")),
		 "reserve memory for active-area mask")) return NULL;
		for (yu = 0, ucsp = pactp; yu < inht; yu++, ucsp += nbigcols)
			pactive[yu] = ucsp;
		if (!(opno & 034)) return pactive; /* we're done if opno == 3*/
	  }
	if (opno & 04) /* read into the memory we have available */
	  {
		if ((NULL == pactive) &&
		 (NULL == activmask(inwid, inht, 03))) return NULL;
		/* if the AMASK file can't be opened, return a default mask */
		if (NULL == (fpact = xgopen("amask", "r")))
			return activmask(inwid, inht, 021);
		else {
			for (yu = 0; yu < inht; yu++)
			 if (1 != fread((void *)(pactive[yu]), nbigcols,
				1, fpact))
			    {
				fprintf(stderr,
				 "Can't read row %4d of active mask file\n",y);
				(void)fclose(fpact);	return NULL;
			    }
			(void)fclose(fpact);	fpact = NULL;
		     }
		return pactive;
	  }
	if (opno & 010)	/* write the resulting masks out */
	  {
		if (NULL == (fpact = xgopen("amask", "w"))) return 0;
		for (row = 0; row < inht; row++)
		   if (0 == fwrite((void *)pactive[row],
			(unsigned)nbigcols, 1, fpact))
		     {
			(void)fclose(fpact);	return 0;
		     }
		(void)fclose(fpact);	fpact = NULL;	return pactive;
	  }
	/* if we get this far, it's because we want a default mask. This
	is a truncated ellipse with equation ((x-xc)/xr)^2 + ((y-yc)/yr)^2 < 1,
	and for each x from 0 to `width' we compute the upper and lower bounds
	in y.  We also define the the outermost rows and columns to be dead. */
	if ((NULL == pactive) && (NULL == activmask(inwid, inht, 01)))
		return NULL;
	for (px = 0; px < nbigcols; px++)
		pactive[0][px] = pactive[inht - 1][px] = 0;
	xcenter = ((double)inwid) / 2.; ycenter = ((double)inht) / 2.;
	xradius = xcenter * 1.1;	yradius = ycenter * 1.1;
	for (yu = 1; yu < inht - 1; yu++)
	   {
		discrim = ((double)yu - ycenter) / yradius;
		discrim = 1. - discrim * discrim;
		if (discrim < 0.)
		  for (px = 0; px < nbigcols; px++) pactive[yu][px] = 0;
		else
		  {
			discrim = sqrt(discrim) * xradius;
			xleftbord = (int)(xcenter - discrim + 0.49999);
			if (xleftbord < 1) xleftbord = 1;
			xrightbord = (int)(xcenter + discrim + 0.49999);
			if (xrightbord > inwid - 2) xrightbord = inwid - 2;
			/* having defined the first and last active pixels
			 in each row, we set up a row of the `active' array. */
			pxleft = (int)(xleftbord / 8);
			pxright = (int)(xrightbord / 8.);
			for (px = 0; px < pxleft; px++) pactive[yu][px] = 0;
			for (px = pxright + 1; px < nbigcols; px++)
				pactive[yu][px] = 0;
			for (px = pxleft + 1; px < pxright; px++)
				pactive[yu][px] = 0377;
			for (ux = ((int)xleftbord) % 8; ux < 8; ux++)
				pactive[yu][pxleft] |= (1 << ux);
			for (ux = 0; ux <= ((int)xrightbord) % 8; ux++)
				pactive[yu][pxright] |= (1 << ux);
		  }
	   }
	return pactive;
}

void shrink(int npixels, int dewidth, int deheight)
{ /*	This shrinks the active portion of the mask by `npixels' pixels on all
	sides.  Thus if the first and last active rows going in are 2 and 508
	and `npixels' == 2, then rows 2-3 and 507-508 will be blanked.
	Each individual row would have its starting and ending X values pushed
	in toward the center by two as well. */

	register int	px, zx, ux;
	int		row, act1_pix, act1_slot, act1_row;
	int		actz_pix, actz_slot, actz_row;
	u_char		kill, thisv;
	static u_char	**lup;

	if (npixels < 1 || npixels > 254) return; /* check argument */
	if (NULL == lup) lup = activmask(dewidth, deheight, 01);
	if (NULL == lup) return;
	/* find the first and last nonzero rows */
	act1_row = actz_row = deheight / 2 - 1;
	for (row = 0; row <= act1_row; row++)
	   {
		for (px = 0; px < dewidth / 8; px++)
		   if (lup[row][px])
		      {
			act1_row = row;	break;
		      }
	   }
	for (row = deheight - 1; row > actz_row; row--)
	   {
		for (px = 0; px < dewidth / 8; px++)
		   if (lup[row][px])
		      {
			actz_row = row;	break;
		      }
	   }
	/* shrink those appropriately */
	act1_row += npixels;	actz_row -= npixels;
	if (act1_row > actz_row)
		act1_row = actz_row = (act1_row + actz_row) / 2;
	/* blank out the topmost and bottommost rows */
	for (row = 0; row < act1_row; row++)
	   {
		for (px = 0; px < dewidth / 8; px++) lup[row][px] = 0;
	   }
	for (row = deheight - 1; row > actz_row; row--)
		for (px = 0; px < dewidth / 8; px++) lup[row][px] = 0;
	/* now examine the remaining rows. */
	for (row = act1_row; row <= actz_row; row++)
	   { /* first, find the first and last nonzero pixel in the row. */
		act1_pix = dewidth / 2 - 1;	act1_slot = act1_pix / 8; 
		actz_pix = act1_pix + 1;	actz_slot = actz_pix / 8;
		for (px = 0; px <= act1_slot; px++)
		   if ((thisv = lup[row][px]))
		     {
			for (zx = 0, ux = 1; zx < 8; zx++, ux <<= 1)
			   if (thisv & ux)
			     {
				act1_slot = px; act1_pix = 8 * px + zx;	break;
			     }
		     }
		for (px = actz_slot * 2 - 1; px > actz_slot; px--)
		   if ((thisv = lup[row][px]))
		     {
			for (zx = 7, ux = 128; zx >= 0; zx--, ux >>= 1)
			   if (thisv & ux)
			     {
				actz_slot = px;	actz_pix = 8 * px + zx;	break;
			     }
		     }
		/* now shrink those accordingly */
		act1_pix += npixels;	actz_pix -= npixels;
		act1_slot = act1_pix / 8;	actz_slot = actz_pix / 8;
		/* if first > last, we blank the whole row */
		if (act1_pix > actz_pix)
		  for (px = 0; px < dewidth / 8; px++) lup[row][px] = 0;
		/* otherwise we blank the beginning and end */
		else  {
			for (px = 0; px < act1_slot; px++)
				lup[row][px] = 0;
			for (px = dewidth / 8 - 1; px > actz_slot; px--)
				lup[row][px] = 0;
			for (zx = 0, ux = 1, kill = 0; zx < act1_pix % 8;
			  zx++, ux <<= 1) kill |= ux;
			lup[row][act1_slot] &= ~kill;
			for (zx = 7, ux = 128, kill = 0; zx > actz_pix % 8;
			   zx--, ux >>= 1) kill |= ux;
			lup[row][actz_slot] &= ~kill;
		      }
	   } /* end of loop over remaining rows */
}

void sum_row(FILE *fp, int row, int iwid)
{	/* this summarizes one row's worth of the active-pixel mask */
	register int	px, zx, ux;
	u_char		thisv, isactive;
	int		nactive, ngroups, start_group, end_group;
	static u_char	**lup;

	lup = activmask(iwid, 0, 1);	/* get appropriate pointer */
	for (px = nactive = 0; px < iwid / 8; px++)
	   if ((thisv = lup[row][px]))
	     for (zx = 0, ux = 1; zx < 8; zx++, ux <<= 1)
		if (thisv & ux) nactive++;
	if (nactive >= iwid - 2) return;
	fprintf(fp, "%4d\t\t%4d", row, nactive);
	for (px = isactive = ngroups = start_group = 0; px < iwid / 8; px++)
	 {
	  for (zx = 0, ux = 1, thisv = lup[row][px]; zx< 8; zx++, ux<<= 1)
	   {
	    if (thisv & ux)
	      {
		if (!isactive) start_group = 8 * px + zx;
		isactive = 1;
	      }
	     else
	      {
		if (isactive)
		  {
			end_group = 8 * px + zx - 1;
			fprintf(fp, "   %4d %4d", start_group, end_group);
			if (!((++ngroups) % 6)) fprintf(fp, "\n\t\t\t");
		  }
		isactive = 0;
	      } /* end of "else" of "if (thisv..." clause */
	   } /* end of loop through zx values */
	 } /* end of loop through px values */
	fprintf(fp, "\n");
}

static u_short *imaverage(size_t imsize)
{ /*	This manages memory for an averaged image. If imsize is 0,
	we use it to free up memory for the averaged image;
	Otherwise we use it to get memory for it */
	static u_short	*imav;

	if (!imsize) /* eliminate memory for this averaged image */
	  {
		memzap((AVOID **)&imav, "imav in imaverage");
		return NULL;
	  }
	if (NULL == imav) /* in this case, we're getting memory for it */
	  {
		imav = (u_short *)callop(imsize, sizeof (u_short),
		 "imav in imaverage");
	  }
	return imav;	/* now we're just returning it */
}

double merge_frames(MAIL *headp, int *nom_bgnd)
{	/* this merges the frame whose data is associated with the header
	`headp' into the overall store.  The routine returns the number
	of counts in the frame, not counting the ones over 2 * meancount.
	A pointer to the nominal background value is returned in *nom_bgnd. */
	u_char		*ucp;
	u_short		ct, highct, *uspi, *uspo;
	int		colev, imht, imwid, ovc, x, y;
	size_t		imsize;
	double		ovsum, revent;
	VOFLOW		*vofp;
	static u_short	*aveimage;

	revent = headp->eventcount;	if (revent < 0.) revent *= -64.;
	imwid = headp->xdimen;	imht = headp->ydimen;	imsize = imwid * imht;
	if ((NULL == aveimage) &&
	 (NULL == (aveimage = imaverage(imsize)))) return EALLOC;
	if (NULL != (ucp = headp->data8))
	  {	/* 8-bit data case: first flatten the overflows */
		ovc = headp->oflowcount;
		for (ovsum = 0., vofp = headp->overflow; (ovc--) > 0;
		 vofp++) ovsum += (double)(vofp->value - 255);
		revent -= ovsum;
		/* estimate background as twice the average count per pixel,
		 accounting for the flattopping of the overflows */
		*nom_bgnd = 2. * revent / (double)(imwid * imht) + 0.49999;
		highct = (*nom_bgnd < 255 ? *nom_bgnd : 255);
		if (highct < 1) highct = 1;
		for (x = revent = 0, uspo = aveimage; x < imwid; x++)
		   {
			for (colev = y = 0; y < imht; y++, ucp++, uspo++)
			   {
				ct = *ucp;	if (ct > highct) ct = highct;
				*uspo += ct;	colev += ct;
			   }
			revent += (double)colev;
		   }
	  }
	else {
		uspi = headp->data16;	/* no flattopping done here */
		/* estimate background as twice the average count per pixel */
		*nom_bgnd = 2. * revent / (double)(imwid * imht) + 0.49999;
		highct = *nom_bgnd;
		if (highct < 1) highct = 1;
		for (x = revent = 0, uspo = aveimage; x < imwid; x++)
		   {
			for (colev = y = 0; y < imht; y++, uspi++, uspo++)
			   {
				ct = *uspi;	if (ct > highct) ct = highct;
				*uspo += ct;	colev += ct;
			      }
			revent += (double)colev;
		   }
	      }
	return revent;
}

int process_total(FILE *fp, double totalcounts, double qva,
 int inwid, int inht, int bordflags)
{ /*	"process_total":
	This determines the dead regions of the detector and
	creates a bit-mask defining those dead regions.  From the summed
	"background-accumulated" images obtained in other routines
	we have an estimate of the background count in every pixel
	on the detector and an estimate of the mean of all these
	single-pixel background values.
	Any pixel whose	count is less than 1/Q times this mean
	is assumed to be either outside the active area of the
	detector or else shadowed by a beamstop or a beamstop hanger.
	Such pixels are identified as dead in the output bitmask;
	all other pixels are defined as live.
	There is a further restriction: a pixel will be counted as live ONLY
	IF at least 2 of the 4 pixels adjacent to it are above the cutoff.
	The bitmask itself is constructed by OR'ing eight consecutive
	pixels within each row as one byte; thus if pixels (105,0)
	through (105,7) are (DEAD,DEAD,DEAD,ALIVE,ALIVE,ALIVE,ALIVE,ALIVE),
	then the byte in the bitmask associated with those eight pixels
	will be 128*1 + 64*1 + 32*1 + 16*1 + 8*1 + 4*0 + 2*0 + 1*0 = 248.
	The bitmask is set up as a (imagewidth) by (imageheight / 8) array,
	so the byte we just mentioned will be at [105,0].
	Change for XENGEN V.1.3 21-Jan-1988:
		to prevent the appearance of isolated DEAD points, we look
		at how many of the immediate neighbors of dead points are
		alive.  If all four neighbors are alive	we turn the pixel on.
		In addition, we DON'T write out the mask within this routine;
		we do that elsewhere.
	for XGEN 3.1: dbmask is now of variable length (duhhh) */
	register int	x, y;
	u_short		cutv, *aix, *aixm, *aixp, *aveimage;
	unsigned	nbigcols;
	size_t		imarea;
	int		dx, howoften, naybors, nincl, px;
	int		xm, xp, ym, yp, totincl, zx;
	static u_char	*dbmask, **sgpact;

	imarea = inwid * inht;
	nbigcols = inwid / 8;	if (inwid % 8) nbigcols++;
	 /* get the memory (if necessary) for the `sgpact' vector. We could
	 actually just use the static global "pactive", but this approach
	 makes the code more portable. */
	if ((NULL == (sgpact = activmask(inwid, inht, 1))) &&
	 (NULL == (sgpact = activmask(inwid, inht, 5))))
		return (errno == ENOMEM) ? EALLOC : ENOENT;
	cutv = (u_short)(qva * totalcounts / ((double)imarea) + 0.49999);
	head_results(fp);
	howoften = (bordflags & 1) ? 1 : 16;
	if ((NULL == dbmask) && (NULL == (dbmask = (u_char *)callop(
	 (unsigned)nbigcols, sizeof (u_char), "dbmask in process_total"))))
		return -1;
	aveimage = imaverage(imarea);	/* find the averaged image */
	for (y = totincl = 0; y < inht; y++)
	   {
		if (y > 0 && y < inwid - 1)
		  {
			ym = (y ? y - 1 : y);
			yp = (y < inht - 1 ? y+1 : y);
			for (px = x = nincl = 0; px < inwid / 8; px++)
			   {
			     dbmask[px] = 0;
			     for (dx = 0, zx = 1; dx < 8; dx++, zx <<= 1, x++)
			      {
				xm = (x ? x-1 : x);
				xp = (x < inwid - 1 ? x+1 : x);
				aix = aveimage + (x * inht);
				aixm = aveimage + (xm * inht);
				aixp = aveimage + (xp * inht);
				naybors =
				 (*(aix + ym) > cutv) + (*(aix + yp) > cutv) +
				 (*(aixm + y) > cutv) + (*(aixp + y) > cutv);
				if (*(aix + y) > cutv)
				  {
					if (naybors >= MIN_NEIGHBORS)
					  {
						dbmask[px] |= zx; nincl++;
					  }
				  }
				else if (naybors >= MAX_NEIGHBORS)
				  {
					dbmask[px] |= zx; nincl++;
				  }
			       } /* end of "for (dx" loop */
			     sgpact[y][px] = dbmask[px];
			    } /* end of "for (px" loop */
		     } /* end of "if (y > " clause */
		else for (px = nincl = 0; px < inht / 8; px++)
			sgpact[y][px] = dbmask[px] = 0;
		if (!(y % howoften)) sum_row(fp, y, inwid);
		totincl += nincl;
	   } /* end of loop over y */
	fprintf(fp,
 "Pixels declared active if count > %8d = %8.3f * %8.2e / %d\n",
	 cutv, qva, totalcounts, (int)imarea);
	memzap((AVOID **)&dbmask, "dbmask in process_total");
	(void)imaverage(0);	/* kill memory for averaged image */
	return totincl;
}

/* end of actmask.c */
