/* ascout.c -- XGEN utility routines for ascii output of binary results

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	21 Apr 2007: prurefl: rational printout of 1-D profile elements > 32K
	19 Apr 2007: float->double conversions; reorganize use of "{"
	24 Feb 2007: minor reformatting in prtroid
	28 Nov 2006: condpr made more versatile and made a global routine
	26 Jun 2006: include xgenpred,xgmodel,xgpostint.h: we need them
	29 May 2006: comments in dumpimage reformatted for 80 char/line
	24 Nov 2005: inline argument declarations and strict prototyping
	11 Nov 2005: replace dumpimage(drp) & newdump(drp)
	 with dumpimage(drp, 0/1); elim killip0 and some static globals
	24 Oct 2005: housekeeping; eliminating a few static globals
	20 Oct 2004: make_ftab called unconditionally when used
	26 Aug 2004: routines used only here are declared as static; cosmetics
	 1 Jul 2004: plotting functions made more robust and reliable
	27 Jun 2004: add in routines for handling plotting
	30 May 2004: printshel() routine written: enables output of SHELX
		& XPREP inputs. Note that format is slightly different.
	 8 Feb 2004: bugfix in head_ndump: true maximum is now put out;
		use ANNOUNCE_BOX flag in read_box call appropriately
	30 Jan 2004: newdump() now subtracts the local minimum of the
		intensity before scaling and printing out the count values;
		head_ndump() tells the user what's going on there
	21 Jan 2004: license message inserted and initial comments reordered
	18 Jan 2004: newdump() and dumpimage() moved from ../utilities/xdump.c
	 5 Jan 2004: arg width in print_row changed to mwidth to eliminate
		confusion with the now-defunct global variable `width'
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Sep 2003: eliminate access() call for UPARAMS;
		it's wrapped up into the xgopen() call anyway
	 5 Jan 2003: reorganized source-file header
	 1 Nov 2002: use of REGIONPROFILE structures for reflection-box sizes;
		enables more general formats to be handled;
		in general we minimize use of fixed formats for model profiles
	25 Mar 2002: removed MSI attachments
	 8 Nov 1996: minor changes for X-GEN checkin at MSI
	24 Aug 1995: image data read as ints, not u_shorts
	 6 Jul 1995: disp_image: access data as a one-dimensional vector
	31 May 1995: "register" -> "register int" throughout;
		print integers with %d, not %ld
	12 Apr 1995: s_char replaced by sichar
	 2 Mar 1995: diagnostics to stdout
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	27 Oct 1994: in prurefl: scal[3] == 0 if printing (X,Y,om) intensities;
		scal[3] == 1 if printing transformed-coordinate intensities
	15 Oct 1994: in prurefl(), if scal[3] is nonzero we print the
		(tranformed-coordinate) intensity and sigma.
	11 Oct 1994: always recalculate wavelength when `needlambda' is called
	 5 Oct 1994: use xgread to input the REFPARS data
	30 Sep 1994: accommodate floating-point stepsize
	23 Sep 1994: fix bug in pruref so plane spacing comes out right
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: polymask added; bugfixes in polymath, xtal, calinv, etc.
	 7 Dec 1993: initial X-GEN version
	19 Nov 1992: culled from other code.
	
   Purpose: enabling ASCII output of binary results from X-GEN
   Cross references: ../utilities/xdump.c, ../utilities/prurefl.c, ...

	This organizes a bunch of routines used in generating ascii
	output of various binary constructions found in XGEN.
	It is modular in that one object gets printed out by one call
	to each of these routines, so we can easily combine calls to these
	routines in special-purpose or general-purpose output programs.
*************************************************************************/
#define		NEED_TIME	1
#include	"xgenpred.h"
#include	"xgmodel.h"
#include	"xgpostint.h"
#define		CIR(x)	((sichar)(rpp->xisref[x]))
#define		NARROW		2	/* # chars/element, fms 0 and 6 */
#define		WIDE		3	/* # chars/element, fms 1 -> 5 */
#define		MIN_SIZE	(5000)	/* min size of printing tank */
#define		MAX_PRINTABLE	(9999)	/* Max ct printed in raw image*/

static void disp_image(FILE *fp, int *imv, int xmn, int ymn, int xmx, int ymx)
{ /* This displays the box (xmn->xmx)*(ymn->ymx) of an image
	on the stream fp. As of 17 Aug 95, imv is *int. */
	register int	x, y;
	int	*ip;

	fprintf(fp, "      ");
	for (x = xmn; x <= xmx; x++)	fprintf(fp, "%4d", x);
	fprintf(fp, "\n      ");
	for (x = xmn; x <= xmx; x++) fprintf(fp, "____");
	fprintf(fp, "\n");
	for (y = ymn, ip = imv; y <= ymx; y++)
	   {
		fprintf(fp, " %4d|", y);
		for (x = xmn; x <= xmx; x++, ip++) fprintf(fp, "%4d", *ip);
		fprintf(fp, " |\n");
	   }
	fprintf(fp, "      ");
	for (x = xmn; x <= xmx; x++) fprintf(fp, "----");
	fprintf(fp, "\n\n");
}

void prurefl(FILE *fp, OBSREF *urp, double *scal)
{ /* This prints one observed refl to the stream fp.
	Unlike earlier versions, the plane spacing is now calculated correctly.
	It was wrong before by the ratio lambda * sqrt(0.4),
	which for copper kalpha is 0.975... pretty close!.
	This necessitates another scale parameter equal to lambda * 100.
	If scal[3] is nonzero we print the (alpha,beta,gamma)
	(tranformed-coordinate) intensity and sigma. */
	register int	i;
	int		lz, elt;
	double		iu0, iu1, su0, su1;

	lz = PFRAMES;
	for (i = 0; i < 3; i++) fprintf(fp,"%3d", urp->obs_hkl[i]);
	iu0 = urp->obs_ii;	su0 = urp->obs_sigma;
	iu1 = urp->obs_iunfit;	su1 = urp->obs_sunfit;
	if (scal[3])
	  {
		iu0 += ((double)(urp->obs_shift)) / 16.;
		su0 += ((double)(urp->obs_widob)) / 8.;
		iu1 += ((double)(urp->obs_bgunfit)) / 16.;
		su1 += ((double)(urp->obs_reflno)) / 8.;
	  }
	if (su0 < FUZZ) su0 = 1.;	if (su1 < FUZZ) su1 = 1.;
	fprintf(fp, "%6.2f%7.0f%5.0f%7.0f%5.0f%6.1f%6.1f%7.1f",
		scal[2] / sqrt((double)(urp->obs_ssq)),
		iu0, sqrt(su0), iu1, sqrt(su1),
		(double)urp->obs_xpred /scal[0],
		(double)urp->obs_ypred /scal[1], urp->obs_ompred);
	fprintf(fp, "%4.1f%4.1f%4.1f%3d%6d%4d%5d%2d%2d%4d",
		(double)(urp->obs_xob - urp->obs_xpred)/ scal[0],
		(double)(urp->obs_yob - urp->obs_ypred)/ scal[1],
		urp->obs_omob - urp->obs_ompred, urp->obs_flags,
		(int)(u_short)urp->obs_gof, urp->obs_bgnd,
		urp->obs_lpa,
		urp->obs_lowfrm, urp->obs_hifrm, urp->obs_ncells);
	for (i = 0; i < lz; i++)
	   {
		elt = urp->obs_frmint[i];	if (elt < -9999) elt += 65536;
		fprintf(fp,"%5d", elt);
	   }
	fprintf(fp,"\n");
}

void condpr(FILE *fp, float fval, int nspace, int nafter, int forspace)
{ /*	This prints a floating point number `fval' to the stream fp in a space
	`nspace' characters wide.  We print at most `nafter' digits after the
 	place; less if there isn't room.  This works for `nspace' <= 99. */
	static char	pformat[] = {'%', '9', '.', '5', 'f', '\0', '\0' } ;
	int		aslot, nleft;
	double		aval;

	if (forspace < 0)	/* flag meaning don't print negative values */
	  {
		if (fval < 0.)
		  {
			/* Print a row of blanks. nspace==-1 means just print
			 `nspace' blanks; nspace=-2 means print nspace+1; etc*/
			for (nleft = 0; nleft < nspace -1 -forspace; nleft++)
				fputc((int)' ', fp);
			return;	/* nothing else */
		  }
		else if (forspace < -1)
			for (nleft = 0; nleft < -1 -forspace; nleft++)
				fputc((int)' ', fp);
	  }
	else if (forspace > 0) /* left-pad with some spaces */
	  {
		for (nleft = 0; nleft < forspace; nleft++) fputc((int)' ', fp);
	  }
	if (nspace <= 9)
	  {
		pformat[1] = (char)nspace + '0';
		pformat[2] = '.';	pformat[4] = 'f';
		aslot = 3;	pformat[5] = '\0';
	  }
	else {
		pformat[1] = (char)(nspace / 10) + '0';
		pformat[2] = (char)(nspace % 10) + '0';
		pformat[3] = '.';	pformat[5] = 'f';
		aslot = 4;
	     }
	aval = fabs((double)fval);
	nleft = 2 + (aval < 1 ? 0 : (int)log10(aval));
	if (fval < 0) nleft++;
	if (nspace < (nafter + nleft))
	  {
		nafter = nspace - nleft; if (nafter < 0) nafter = 0;
	  }
	pformat[aslot] = ((char)nafter + '0');
	fprintf(fp, pformat, fval);
}

void pspline(FILE *fp, VSPLINE *splip)
{ /*	This prints the contents of a spline file to the print stream fp */
	register int	i, j, xlast, ylast;
	char		after;
	CALPOINT	*calp;

	xlast = splip->v_ncol - 1;	ylast = splip->v_nrow - 1;
	fprintf(fp, "Width and height of the detector in pixels: %4d, %4d\n",
		splip->v_detw, splip->v_deth);
	fprintf(fp,
	 "Number of columns & rows in array: %2d, %2d\n",
		splip->v_ncol, splip->v_nrow);
	fprintf(fp,
 "Number of points in array indexed from observed centroids = %4d\n",
		splip->v_ntrue);
	fprintf(fp,
	 "absolute center = [%9.5f,%9.5f]; relative center = [%9.5f,%9.5f]\n",
	 splip->v_axc, splip->v_ayc, splip->v_rxc, splip->v_ryc);
	fprintf(fp,
 " Fiducial points spaced %6.3f,%6.3f  cm apart in X,Y starting @%6.2f,%6.2f\n",
		splip->v_xrcm, splip->v_yrcm, splip->v_x0cm, splip->v_y0cm);
	fprintf(fp,
 "Reference points spaced %6.3f,%6.3f pix apart in X,Y starting @%6.2f,%6.2f\n",
		splip->v_xrpx, splip->v_yrpx, splip->v_x0px, splip->v_y0px);
	if ((fp == stderr) || (fp == stdout)) return;
	fprintf(fp, "pixel-to-cm X array:\n");
	for (j = 0, calp = splip->v_cpt; j <= ylast; j++)
	   {
		for (i = 0; i <= xlast; i++, calp++)
		   {
			condpr(fp, calp->c_ptocx, 6, 2, 0);
			after = (i+1) % 19 ? ' ' : '\n';
			putc(after,fp);
		   }
		if (i % 19) putc('\n', fp);
	   }
	fprintf(fp,"pixel-to-cm Y array:\n");
	for (j = 0, calp = splip->v_cpt; j <= ylast; j++)
	   {
		for (i = 0; i <= xlast; i++, calp++)
		   {
			condpr(fp, calp->c_ptocy, 6, 2, 0);
			after = (i+1) % 19 ? ' ' : '\n';
			putc(after,fp);
		   }
		if (i % 19) putc('\n', fp);
	   }
	fprintf(fp,"cm-to-pixel X array:\n");
	for (j = 0, calp = splip->v_cpt; j <= ylast; j++)
	   {
		for (i = 0; i <= xlast; i++, calp++)
		   {
			condpr(fp, calp->c_ctopx, 6, 3, 0);
			after = (i+1) % 19 ? ' ' : '\n';
			putc(after,fp);
		   }
		if (i % 19) putc('\n', fp);
	   }
	fprintf(fp,"cm-to-pixel Y array:\n");
	for (j = 0, calp = splip->v_cpt; j <= ylast; j++)
	   {
		for (i = 0; i <= xlast; i++, calp++)
		   {
			condpr(fp, calp->c_ctopy, 6, 3, 0);
			after = (i+1) % 19 ? ' ' : '\n';
			putc(after,fp);
		   }
		if (i % 19) putc('\n', fp);
	   }
	calp = splip->v_cpt + (splip->v_ncol * splip->v_midr + splip->v_midc);
	if (fabs(calp->c_qec) > FUZZ)
	  {
		fprintf(fp, "qecor:\n");
		for (j = 0, calp = splip->v_cpt; j <= ylast; j++)
		   {
			for (i = 0; i <= xlast; i++, calp++)
			   {
				condpr(fp, calp->c_qec, 6, 3, 0);
				after = (i+1) % 19 ? ' ' : '\n';
				putc(after,fp);
			   }
			if (i % 19) putc('\n', fp);
		   }
	  }
	return;
}

void show_params(FILE *fp, REFPARS *rpp, int nbr)
{	/* This summarizes the current crystal and detector parameters.
	`nbr' is the number of bytes in the requested UPARAMS file;
	it's 232 if it's a XENGEN 1.3 or 2.0-style UPARAMS file,
	and 256 if it's a later one. */

	fprintf(fp,"\t\tCrystal and Detector Runtime Parameters:\n");
	fprintf(fp,"Refining?  a:%c     b:%c     c:%c alpha:%c  beta:%c",
		CIR(0), CIR(1), CIR(2), CIR(3), CIR(4));
	fprintf(fp," gamma:%c omega:%c   chi:%c   phi:%c\n",
		CIR(5), CIR(6), CIR(7), CIR(8));
	fprintf(fp,"Value:%8.3f%8.3f%8.3f%8.3f%8.3f%8.3f%8.3f%8.3f%8.3f\n",
		rpp->xabc[0],rpp->xabc[1],rpp->xabc[2],
		rpp->xang[0],rpp->xang[1],rpp->xang[2],
		rpp->xdelom,rpp->xdelchi,rpp->xdelphi);
	fprintf(fp,"      plateD:%c  Xcen:%c  Ycen:%c  tilt:%c swing:%c",
		CIR(9), CIR(10), CIR(11), CIR(12), CIR(13));
	fprintf(fp,"  gam0:%c  gam1:%c  gam2:%c\n", CIR(14), CIR(15), CIR(16));
	fprintf(fp,
	 "Value:%8.4f%8.4f%8.4f%8.4f%8.3f%8.4f%8.4f%8.4f\n",
		rpp->xplated, rpp->xx0, rpp->xy0, rpp->xtilt / RADC,
		rpp->xswing / RADC, rpp->xgam[0] / rpp->xfrmsize,
		rpp->xgam[1] / rpp->xfrmsize, rpp->xgam[2]);
	fprintf(fp,"    Max Errors Allowed in:    ");
	fprintf(fp,"     Resolution    Wave-    Step Crystal   Space\n");
	fprintf(fp,"  index X(pix)  Y(pix) phi(fm)");
	fprintf(fp,"     Min     Max  length    size  System   Group\n");
	fprintf(fp,"%7.4f%7.4f%8.4f%8.4f%8.4f%8.4f%8.4f%8.4f%8d%8d\n",
		rpp->xerrind,rpp->xerrx,rpp->xerry,rpp->xerrphi,
		rpp->xresmin,rpp->xresmax,rpp->xlambda,
		rpp->xfrmsize / RADC, rpp->xisys, rpp->xispgp);
	if ((unsigned)nbr >= sizeof (REFPARS))	fprintf(fp,	/* new format*/
	 "Starting goniostat omega, chi, phi:   %8.3f%8.3f%8.3f\n",
	 rpp->xstart[0], rpp->xstart[1], rpp->xstart[2]);
	else fprintf(fp,
	 "Alignment:       a along  %s,  b along  %s,  c along  %s\n",
		rpp->xdir[0],rpp->xdir[1],rpp->xdir[2]);
	fprintf(fp,
"   Real-space unit-cell matrix:       Reciprocal-space unit cell matrix:\n");
	fprintf(fp," %9.4f %9.4f %9.4f          %9.6f %9.6f %9.6f\n",
		rpp->xareal[0][0],rpp->xareal[1][0],rpp->xareal[2][0],
		rpp->xrec[0][0],rpp->xrec[1][0],rpp->xrec[2][0]);
	fprintf(fp," %9.4f %9.4f %9.4f          %9.6f %9.6f %9.6f\n",
		rpp->xareal[0][1],rpp->xareal[1][1],rpp->xareal[2][1],
		rpp->xrec[0][1],rpp->xrec[1][1],rpp->xrec[2][1]);
	fprintf(fp," %9.4f %9.4f %9.4f          %9.6f %9.6f %9.6f\n",
		rpp->xareal[0][2],rpp->xareal[1][2],rpp->xareal[2][2],
		rpp->xrec[0][2],rpp->xrec[1][2],rpp->xrec[2][2]);
	return;
}

void prpred(FILE *fp, PREDREF *prp)
{	/* This prints a single "prediction" refl to the stream fp */
	short	th[3];

	hunpack(prp->pindx,th);
	fprintf(fp,"%4d%4d%4d %5d %8.3f %8.3f %9.3f %11.4f %8.5f\n",
		th[0],th[1],th[2],prp->pfmno,
		prp->pxt,prp->pyt,prp->pphi,prp->pcor,prp->pgama);
	return;
}

void prtroid(FILE *fp, TROID *trp)
{	/* this prints a "centroid" refl to the pointer fp */
	register int	j;

	fprintf(fp, " %7.2f %7.2f %7.2f %4d %3d %9.1f ",
		trp->xto, trp->yto, trp->frame0, trp->obkmin,
		trp->nobk, trp->totinten);
	for (j = 0; j < trp->nobk; j++) fprintf(fp,"%6.3f",trp->obparts[j]);
	fprintf(fp,"\n");
	return;
}

void pr_xt(FILE *fp, CRYSTAL *xp)
{ /* this prints a CRYSTAL header to the file fp */
	fprintf(fp,"               Crystal Information for %s\n",xp->xt_code);
	fprintf(fp,"Description: %s\n",xp->xt_desc);
	fprintf(fp,"Notes: %s\n",xp->xt_notes);
	fprintf(fp,"Spacegroup: %s   ",xp->xt_spaceg);
	fprintf(fp,"Intl. Tables No.: %4d   ",xp->xt_intl);
	fprintf(fp,"System: %4d\n",xp->xt_isys);
	fprintf(fp,
	 "Real-Space Unit Cell:        a: %8.3f      b: %8.3f       c: %8.3f\n",
		xp->xt_ar[0], xp->xt_ar[1], xp->xt_ar[2]);
	fprintf(fp,
	 "                         alpha: %8.3f   beta: %8.3f   gamma: %8.3f\n",
		xp->xt_ang[0], xp->xt_ang[1], xp->xt_ang[2]);
	fprintf(fp,
	 "a* along X,b* along Y at omega: %8.3f    phi: %8.3f     chi: %8.3f\n",
		xp->xt_stdpos[0], xp->xt_stdpos[1], xp->xt_stdpos[2]);
	fprintf(fp,"Xray wavelength: %8.5f  Min D spacing: %8.4f",
		xp->xt_lambda, xp->xt_reslim);
	fprintf(fp,"  max S squared: %8.5f\n", xp->xt_ssqmax);
	fprintf(fp,
	 "Number of Observations: %7d No. of Runs: %4d  No. of Shifts: %4d\n",	
		(int)xp->xt_ninteg, xp->xt_nruns, xp->xt_nshifts);
	fprintf(fp,"Number of Merged,Unmerged Reflections: %7d   %7d\n",
		(int)xp->xt_nmerged, (int)xp->xt_nrefls);
	return;
}

void pr3m(FILE *fp, float x[3][3], char *s)
{ /* This prints a 3x3 matrix in a reasonable format */
	register int	i;

	fprintf(fp,"%s:\n",s);
	for (i = 0; i < 3; i++) fprintf(fp," %12.6f %12.6f %12.6f\n",
			x[i][0],x[i][1],x[i][2]);
}

void pr_dr(FILE *fp, DATARUN *drp)
{	/* this prints the contents of a DATARUN header to a file */
	fprintf(fp,"             Datarun Information for %s\n", drp->run_code);
	fprintf(fp,"Notes on run: %s\n", drp->run_notes);
	fprintf(fp,
 "Run #: %3d;   Xray Conditions: %3d KV, %3d ma;  Detector at %3d KV\n",
	 drp->run_id, drp->run_volxray, drp->run_curxray, drp->run_chvol);
	if (fabs(drp->run_stepreal) > FUZZ)
		fprintf(fp, "Stepsize = %7.3f deg; ", drp->run_stepreal);
	else	fprintf(fp, "Stepsize = %3d/96 deg; ", drp->run_stepsize);
	fprintf(fp, "%3d frames ; motor %1d moving; %1d chamber%c used\n",
		drp->run_nframes, drp->run_motor,
		drp->run_nchams, (drp->run_nchams != 1 ? 's' : ' '));
	fprintf(fp,
 "Temp: = %4d K; %7d refls saved, %7d examined. Xray wavelength = %6.4f A\n",
		(int)drp->run_temp, (int)drp->run_ninteg,
		(int)drp->run_nexam, drp->run_lambda);
	fprintf(fp,
 "Mean intensity = %9.2f; Mean sigma = %9.3f; Mean I/sigma = %9.3f\n",
		drp->run_intave, drp->run_sigave, drp->run_iosave);
	fprintf(fp,
 "<squared scatt vector> = %8.5f; resol limit = %7.3f A; max ssq = %8.5f\n",
	 drp->run_ssqave, drp->run_reslim, drp->run_ssqmax);
	pr3m(fp,drp->run_recip,"Reciprocal-space matrix used in run");
	return;
}

void pr_merref(FILE *fp, MERGEDREF *mrp, short spgp)
{ /*	This prints a MERGEDREF reflection */
	register int	i, iscen, nobs;
	register ABBROB	*abp;
	short		ohkl[3];
	double		omcalc;

	munpack(mrp->mr_hkl, ohkl, spgp); /* decode the hkl */
	iscen = iscentric(mrp->mr_hkl, spgp);
	fprintf(fp, " %4d%4d%4d %7.4f %7d %7d        ",
		ohkl[0], ohkl[1], ohkl[2], mrp->mr_stl,
		(int)mrp->mr_ibar, (int)mrp->mr_isig);
	if (iscen) fprintf(fp,"  centric:%3d obs\n", (int)mrp->mr_nob);
	else	   fprintf(fp," acentric:%3d prim., %3d Bijv. obs\n",
			(int)mrp->mr_nobpri, (int)mrp->mr_nobbij);
	nobs = mrp->mr_nob;
	for (i = 0; i < nobs; i++)
	   {	
		if (i == mrp->mr_nobpri) fprintf(fp,"       Bijvoets:     ");
		else if (!i && !iscen)	 fprintf(fp,"       Primaries:    ");
		else			 fprintf(fp,"                     ");
		abp = &(mrp->mr_ob[i]);
		omcalc = (double)abp->abo_omob / 100.;
		fprintf(fp,"%11.2f %7.2f %5d %4d %4d %7.2f %3d %3d\n",
			abp->abo_ii, abp->abo_sigma, abp->abo_shift,
			abp->abo_chamno, abp->abo_runno, omcalc,
			abptox(abp), abptoy(abp));
	   }
	return;
}

static void centered_display(FILE *fp, xglong value, int dwidth, char leftchar)
{ /* This displays the value `value' centered in a box of characters
	`dwidth' characters wide.  If `leftchar' is non-NULL,
	it is printed before the centered string.  */
	register int	i;
	int		back, front, nwidth;
	char		tbuf[80];
	static char bl[] =
"                                                                            ";

	if (leftchar) fprintf(fp, "%c", (sichar)leftchar);
	sprintf(tbuf, "%d", (int)value);
	nwidth = strlen(tbuf);
	if (nwidth > dwidth)
	  {
		tbuf[0] = '>';	for (i = 1; i < dwidth; i++) tbuf[i] = '9';
		tbuf[dwidth] = '\0';	fputs(tbuf, fp);
	  }
	else {
		back = (dwidth - nwidth) / 2;	front = dwidth - nwidth - back;
		if (front)
		  {
			bl[front] = '\0'; fputs(bl, fp); bl[front] = ' ';
		  }
		fputs(tbuf, fp);
		if (back)
		  {
		    bl[back] = '\0'; fputs(bl, fp); bl[back] = ' ';
		  }
	     }
}

void show_masks(FILE *fp, MASK_3D *wp, BOUNDARY *bdp)
{ /* This displays the # of observations in each mask on the stream `fp'.*/

	fprintf(fp,
"\n\tData masks used and the number of contributing reflections in each:\n");
	fprintf(fp,
"         _______________________________________\n");
	fprintf(fp,
"        /    #0      |      #1     |      #2    \\\n");
	fprintf(fp,
"       /             |             |             \\\n      ");
	centered_display(fp, wp->phielt[3].mnobs, 14, '/');
	centered_display(fp, (wp+1)->phielt[3].mnobs, 13, '|');
	centered_display(fp, (wp+2)->phielt[3].mnobs, 14, '|');
	fprintf(fp,
"\\\n     /---------------|-------------|---------------\\      Y = %3d\n",
		bdp->toptri);
	fprintf(fp,
"    /   #3   |   #4  |  #5  |  #6  |   #7  |   #8   \\\n");
	fprintf(fp,
"   (         |       |      |      |       |         )\n    ");
	centered_display(fp, (wp+3)->phielt[3].mnobs,  8, '\\');
	centered_display(fp, (wp+4)->phielt[3].mnobs,  7, '|');
	centered_display(fp, (wp+5)->phielt[3].mnobs,  6, '|');
	centered_display(fp, (wp+6)->phielt[3].mnobs,  6, '|');
	centered_display(fp, (wp+7)->phielt[3].mnobs,  7, '|');
	centered_display(fp, (wp+8)->phielt[3].mnobs,  8, '|');
	fprintf(fp,
"/\n     \\-------|-------|------|------|-------|-------/      Y = %3d\n",
		bdp->bottri);
	fprintf(fp,
"      \\       #9     |     #10     |     #11      /\n");
	fprintf(fp,
"       \\             |             |             /\n        ");
	centered_display(fp, (wp+ 9)->phielt[3].mnobs, 12, '\\');
	centered_display(fp, (wp+10)->phielt[3].mnobs, 13, '|');
	centered_display(fp, (wp+11)->phielt[3].mnobs, 12, '|');
	fprintf(fp,
"/\n         \\___________|_____________|___________/\n");
	fprintf(fp,
"  X =        %3d     %3d    %3d    %3d     %3d\n",
		bdp->l1tri, bdp->leftri, bdp->l2tri, bdp->rightri, bdp->l3tri);
	if (fp != stderr) (void)xeflush(fp);
}

static void print_row(FILE *fp, PHIMASK *pip, int iy, int mwidth,double maxmax)
{ /* this prints one row of mask profile element pip on stream fp */
	register int	ix, prval;
	int		lx;
	double		rescale;

	rescale = pip->maxf / maxmax;	lx = BIGSPOTWID;
	fprintf(fp, "  |");
	for (ix = 0; ix < lx; ix++)
	   {
		prval = ((double)pip->xyint[ix][iy]) * rescale + 0.4999;
		if (mwidth == NARROW && prval < -9) prval = -9;
		if (prval) fprintf(fp,(mwidth== NARROW ? "%2d" : "%3d"), prval);
		else fprintf(fp, (mwidth == NARROW ? "  " : "   "));
	   }
	fprintf(fp, "|");
}

void prmask(MASK_3D *wmp, int region, FILE *fp)
{	/* this prints a single data mask to the stream fp */
	register int		frame, iy;
	int			ly, lz;
	double			maxmax, thisden;
	static char top30[] = "  .------------------------------.";
	static char bot30[] = "  `------------------------------'";
	static char top45[] ="  .---------------------------------------------.";
	static char bot45[] ="  `---------------------------------------------'";

	maxmax = 0.;	ly = BIGSPOTHT;	lz = PFRAMES;
	for (frame = 0; frame < lz; frame++)
	  if ((thisden = wmp->phielt[frame].isqsum) > DFUZZ)
	    {
		wmp->phielt[frame].maxf /= thisden;
		wmp->phielt[frame].isqsum = 1.;
		if (wmp->phielt[frame].maxf > maxmax)
		maxmax = wmp->phielt[frame].maxf;
	    }
	if (maxmax < DFUZZ)
	  {
		fprintf(fp,"No reflections found for mask %2d\n\n", region);
		return;
	  }
	if (region) fprintf(fp,"\f");
	fprintf(fp,
	 "\t\t\tMask %2d: Overall Peak Intensity = %9.4f\n",
		region, maxmax);
	fprintf(fp, " Frame 0: peak,nobs = %10.3e,%5d\t",
		wmp->phielt[0].maxf, (int)wmp->phielt[0].mnobs);
	fprintf(fp, " Frame 1: peak,nobs = %10.3e,%5d\t\t",
		wmp->phielt[1].maxf, (int)wmp->phielt[1].mnobs);
	fprintf(fp, " Frame 2: peak,nobs = %10.3e,%5d\n",
		wmp->phielt[2].maxf, (int)wmp->phielt[2].mnobs);
	fputs(top30, fp);	fputs(top45, fp);
	fputs(top45, fp);	fprintf(fp,"\n");
	for (iy = 0; iy < ly; iy++)
	   {
		print_row(fp, &wmp->phielt[0], iy, NARROW, maxmax);
		print_row(fp, &wmp->phielt[1], iy, WIDE, maxmax);
		print_row(fp, &wmp->phielt[2], iy, WIDE, maxmax);
		fprintf(fp,"\n");
	   }
	fputs(bot30, fp);	fputs(bot45, fp);
	fputs(bot45, fp);	fprintf(fp,"\n");
	fprintf(fp,"\t Frame 3: peak,nobs = %10.3e,%5d\t",
		wmp->phielt[3].maxf, (int)wmp->phielt[3].mnobs);
	fprintf(fp,"\t Frame 4: peak,nobs = %10.3e,%5d\n",
		wmp->phielt[4].maxf, (int)wmp->phielt[4].mnobs);
	fputs(top45, fp);	fputs(top45, fp);	fprintf(fp,"\n");
	for (iy = 0; iy < ly; iy++)
	   {
		print_row(fp, &wmp->phielt[3], iy, WIDE, maxmax);
		print_row(fp, &wmp->phielt[4], iy, WIDE, maxmax);
		fprintf(fp,"\n");
	   }
	fputs(bot45, fp);	fputs(bot45, fp); fprintf(fp,"\n");
	fprintf(fp,"\t Frame 5: peak,nobs = %10.3e,%5d\t",
		wmp->phielt[5].maxf, (int)wmp->phielt[5].mnobs);
	fprintf(fp,"\t Frame 6: peak,nobs = %10.3e,%5d\n",
		wmp->phielt[6].maxf, (int)wmp->phielt[6].mnobs);
	fputs(top45, fp);	fputs(top30, fp); fprintf(fp,"\n");
	for (iy = 0; iy < ly; iy++)
	   {
		print_row(fp, &(wmp->phielt[5]), iy, WIDE, maxmax);
		print_row(fp, &(wmp->phielt[6]), iy, NARROW, maxmax);
		fprintf(fp,"\n");
	   }
	fputs(bot45, fp);	fputs(bot30, fp); fprintf(fp,"\n");
}

double needlambda(void)
{	/* This does its level best to compute lambda. If
	it can't, it assumes copper Kalpha and returns 1.5418 */
	char		*upname;
	double		loclam;
	static REFPARS	rparlo;
	static FILE	*fpte;

	/* try to open the uparams file to extract lambda. If we fail,
	 we return 1.5418. */
	loclam = 1.5418;
	if ((NULL == (upname =trname("uparams"))) || ('\0' == *upname) ||
	 (NULL == (fpte = xgopen(upname, "r")))) return loclam;
	/* try reading it. Even if we fail we have to close the file... */
	if (8 > xgread((void *)&rparlo, XG_REFPARS, fpte)) loclam = 1.5418;
	else	loclam = rparlo.xlambda;
	(void)fclose(fpte);	fpte = NULL;	return loclam;
}

void dispmodel(FILE *fp, REGIONPROFILE *m3p, int which)
{ /*	This displays a model profile according to
	the conventions developed for high-verbosity integration.
	If `which' is even, we're displaying (x,y,omega) profiles;
	if it's odd, we're displaying diffraction-vector profiles.
	If `which' < 2, we're using the floating-point formulation.
	If `which >= 2, we're using the signed-char formulation.*/
	register int	ix, iy, iz;
	int		imo, lqx, lqxy, lqxyz, lqy, lqz, mtype;
	sichar		cval, *zv, *zvp, *zvy, *zvz;
	char		dval, *tacross, *tdown, *tptype, *tzed;
	double		dscale, thisf, thisma;
	float		*zvf, *zvfp, *zvfy, *zvfz;
	static char *ptype[] =	{ "(X,Y,omega)","Diffraction-vector" };
	static char *across[] =	{"X",		"alpha"};
	static char *down[] =	{"Y",		"beta"};
	static char *zed[] =	{"_omega ",	"gamma"};

	mtype = which % 2;
	if (!mtype)
	  {
		lqx = m3p->m3_bx; lqy = m3p->m3_by; lqz = m3p->m3_bz;
		zv = m3p->m3_xyoip;	thisma = m3p->m3_xyoi;
		zvf = m3p->m3_xyop;
	  }
	else {
		lqx = m3p->m3_qx; lqy = m3p->m3_qy; lqz = m3p->m3_qz;
		zv = m3p->m3_dvip;	thisma = m3p->m3_dvi;
		zvf = m3p->m3_dvp;
	     }
	tptype = ptype[mtype];
	tacross = across[mtype]; tdown = down[mtype];	tzed = zed[mtype];
	imo = m3p->m3_imodel;
	lqxy = lqx * lqy;	lqxyz = lqxy * lqz;
	if (which < 2)
	  {
		for (dscale = iz = 0, zvfp = zvf; iz < lqxyz; iz++, zvfp++)
			if (*zvfp > dscale) dscale = *zvfp;
	  }
	else {
		for (dscale = iz = 0, zvp = zv; iz < lqxyz; iz++, zvp++)
		if (*zvp > dscale) dscale = *zvp;
	     }
	if (fabs(dscale) < FUZZ)
	  {
		fprintf(fp, "%s model %2d has no intensity\n", tptype, imo);
		return;
	  }
	fprintf(fp,
	 "%s profile in region %2d; intensity = %13.5e;", tptype, imo, thisma);
	dscale = 62. / dscale;
	fprintf(fp, " %s across; %s down\n   ", tacross, tdown);
	for (iz = 0; iz < lqz; iz++)
		fprintf(fp, ".___%s=%1d___", tzed, iz);
	fputs(".\n", fp);
	if (which < 2)
	 {
	   for (iy = 0, zvfy = zvf; iy < lqy; iy++, zvfy += lqx)
	    {
		fputc((int)' ', fp);	fputc((int)' ', fp);
		if (!(iy % 2)) fprintf(fp, "%1x", iy);
		else fputc((int)' ', fp);
		for (iz = 0, zvfz = zvfy; iz < lqz; iz++, zvfz += lqxy)
		   {
			fputc((int)'|', fp);
			for (ix = 0, zvfp = zvfz; ix < lqx; ix++, zvfp++)
			   {
				thisf = ((double)(*zvfp)) * dscale;
				cval = thisf + 0.49999;
				if (thisf < -20.) dval = '<';
				else if (thisf < -5.) dval = '!';
				else if (thisf < -0.5) dval = '-';
				else if (thisf <= 0.4) dval = ' ';
				else if (thisf < 1.) dval = '.';
				else if (thisf >= 62.) dval = '*';
				else if (1 <= cval && cval <= 9)
					dval = cval + '0';
				else if (10 <= cval && cval <= 35)
					dval = (cval - 10) + 'a';
				else if (36 <= cval && cval <= 61)
					dval = (cval - 36) + 'A';
				else dval = '*';
				fputc((int)dval, fp);
			   }
		   }
		fputs("|\n", fp);
	    }
	 }
	else
	 {
	   for (iy = 0, zvy = zv; iy < lqy; iy++, zvy += lqx)
	    {
		fputc((int)' ', fp);
		fputc((int)' ', fp);
		if (!(iy % 2)) fprintf(fp, "%1x", iy);
		else fputc((int)' ', fp);
		for (iz = 0, zvz = zvy; iz < lqz; iz++, zvz += lqxy)
		   {
			fputc((int)'|', fp);
			for (ix = 0, zvp = zvz; ix < lqx; ix++, zvp++)
			   {
				thisf = ((double)(*zvp)) * dscale;
				cval = thisf + 0.49999;
				if (thisf < -20.) dval = '<';
				else if (thisf < -5.) dval = '!';
				else if (thisf < -0.5) dval = '-';
				else if (thisf <= 0.4) dval = ' ';
				else if (thisf < 1.) dval = '.';
				else if (thisf >= 62.) dval = '*';
				else if (1 <= cval && cval <= 9)
					dval = cval + '0';
				else if (10 <= cval && cval <= 35)
					dval = (cval - 10) + 'a';
				else if (36 <= cval && cval <= 61)
					dval = (cval - 36) + 'A';
				else dval = '*';
				fputc((int)dval, fp);
			   }
		   }
		fputs("|\n", fp);
	    }
	 }
	fputs("   ", fp);
	if (!mtype) for (iz = 0; iz < lqz; iz++) fputs(".0_2_4_6_8_a_c_e", fp);
	else for (iz = 0; iz < lqz; iz++) fputs(".0_2_4_6_8_a_c", fp);
	fputs(".\n", fp);
}

static void head_ndump(FILE *fp, double scal, int mifou, int mafou)
{ /* This provides a useful intro to an image dump */
	double	realmax;

	realmax = (double)mifou + scal * (double)mafou;
	fprintf(fp, "Values in region range from %6d to %7.0f\n",
		mifou, realmax);
	fprintf(fp, "Values shown are offsets from %6d", mifou);
	if (scal > 1) fprintf(fp,
	 "; Resulting values are scaled by a factor of %6.2f\n", scal);
	else fprintf(fp, "\n");
}

void dumpimage(DUMPRUN *dumpp, int dumptype)
{ /*	This routine sets up the image display onto stdout and dumpp->dui_fpd.
	If `alsolog' is zero, we only write to stdout. Unlike earlier versions
	of this routine, this one uses no static globals, and to save on
	arguments we input things into a structure pointer `dumpp'. */
	register int	x, y;
	int		halfheight, halfwidth, locfl, maxfound, minfound;
	int		tfm, thisv, xmn, xmx, ymn, ymx, *ip, xyli[4];
	char		lrdate[26];
	size_t		totx, toty;
	time_t		tl0;
	double		scale;
	MAIL		*thishd;
	static double	scale_step = 2.;
	static int	*ip0;
	static size_t	nowsize;

	/* establish limits of box to be dumped */
	halfwidth = (dumpp->dui_wid > 1) ? dumpp->dui_wid / 2 : 1;
	halfheight = (dumpp->dui_ht > 1) ? dumpp->dui_ht / 2 : 1;
	xmn = dumpp->dui_xce - halfwidth;	if (xmn < 0) xmn = 0;
	xmx = dumpp->dui_xce + halfwidth;
	dumpp->dui_fmi = ftables(3);
	if (0 > (tfm = match_inode(&(dumpp->dui_fmi), dumpp->dui_imname)))
		tfm = 0;
	thishd = dumpp->dui_heads + (tfm % dumpp->dui_nmbox);
	if (xmx > (int)(thishd->xdimen - 1)) xmx = thishd->xdimen - 1;
	ymn = dumpp->dui_yce - halfheight;	if (ymn < 0) ymn = 0;
	ymx = dumpp->dui_yce + halfheight;
	if (ymx > (int)(thishd->ydimen-1)) ymx = thishd->ydimen - 1;
	/* reserve memory for the box of data if needed */
	toty = ymx + 1 - ymn;	totx = xmx + 1 - xmn;	nowsize = totx * toty;
	xyli[0] = xmn;	xyli[1] = xmx;	xyli[2] = ymn;	xyli[3] = ymx;
	if (NULL == (ip0 = (int *)callop(nowsize, sizeof (int),
		 "ip0 in dumpimage"))) return;
	if (!dumptype)	/* old-fashioned dump */
	  {	/* get a scale factor, if needed, for the box,
		 and read in the data. */
		for (y = ymn, ip = ip0, maxfound = 0; y <= ymx; y++)
		   for (x = xmn; x <= xmx; x++, ip++)
		      {
			if ((thisv = imvalue(thishd, x, y)) > maxfound)
				maxfound = thisv;
			*ip = thisv;
		      }
		minfound = 0;
	  }
	else {	/* dump based on read_box routine */
		locfl = GET_RUNHEADER | ANNOUNCE_BOX;
		if (FRAME_OK != (x = read_box(dumpp->dui_imname, locfl,
			thishd, xyli, ip0)))
		  {
			fprintf(stderr, "read_box returned %d\n", x);
			memzap((AVOID **)&ip0, "ip0 in dumpimage");
			return;
		  }
		/* get a floor and, if needed, a scale factor
		 for the box, and read in the data. */
		minfound = 10000000;
		for (y = ymn, ip = ip0, maxfound = 0; y <= ymx; y++)
		   for (x = xmn; x <= xmx; x++, ip++)
		      {
			if ((thisv = *ip) > maxfound) maxfound = thisv;
			if (thisv < minfound) minfound = thisv;
		      }
		maxfound -= minfound;
	     }
	tl0 = NOW;	strcpy(lrdate, ctime(&tl0));
	if (NULL != dumpp->dui_fpd) fprintf(dumpp->dui_fpd,
			" Dump of frame %s on %s", dumpp->dui_imname, lrdate);
	fprintf(stdout,	" Dump of frame %s on %s", dumpp->dui_imname, lrdate);
	/* announce and apply the scale factor if needed */
	if (maxfound > 0)
	  {
		for (scale = 1.; maxfound > MAX_PRINTABLE;
		 scale *= scale_step) maxfound /= (int)scale_step;
	  }
	else	scale = 1;
	if (!dumptype)
	  {
		if (scale > 1)
		  {
			fprintf(stdout,
			 "Raw intensities rescaled by %7.2f:\n", scale);
			if (NULL != dumpp->dui_fpd) fprintf(dumpp->dui_fpd,
			 "Raw intensities rescaled by %7.2f:\n", scale);
		  }
	  }
	else {
		head_ndump(stdout, scale, minfound, maxfound);
		if (NULL != dumpp->dui_fpd) head_ndump(dumpp->dui_fpd,
			scale,minfound, maxfound);
	     }
	for (y = ymn, ip = ip0; y <= ymx; y++)
	   for (x = xmn; x <= xmx; x++, ip++)
		*ip = ((double)(*ip - minfound)) / scale + 0.4999;
	/* Now display the data */
	disp_image(stdout, ip0, xmn, ymn, xmx, ymx);
	if (NULL != dumpp->dui_fpd)
		disp_image(dumpp->dui_fpd, ip0, xmn, ymn, xmx, ymx);
	memzap((AVOID **)&ip0, "ip0 in dumpimage");
}

void printshel(FILE *fp, SHEL_HKL *shp)
{ /* This prints the refletion `shp' in Sheldrick's standard format. */
	fprintf(fp, "%4d%4d%4d %8.2f %8.2f%4d\n",
	 shp->x_hkl[0], shp->x_hkl[1], shp->x_hkl[2],
	 shp->x_inten, shp->x_sigma, shp->x_batch);
}

void tell_dvcoef(FILE *fp, REGIONPROFILE *rgp)
{ /* Print diffraction-vector coefficients for region rgp on stream fp */
	fprintf(fp, "Diffraction-vector coefficients:\n");
	fprintf(fp,
 " angle       minimum       maximum      interval #elements\n");
	fprintf(fp, " alpha %13.5e %13.5e %13.5e %2d\n",
		rgp->m3_lim[0], rgp->m3_lim[1], rgp->m3_lim[2], rgp->m3_qx);
	fprintf(fp, "  beta %13.5e %13.5e %13.5e %2d\n",
		rgp->m3_lim[3], rgp->m3_lim[4], rgp->m3_lim[5], rgp->m3_qy);
	fprintf(fp, " gamma %13.5e %13.5e %13.5e %2d\n",
		rgp->m3_lim[6], rgp->m3_lim[7], rgp->m3_lim[8], rgp->m3_qz);
}

static char *frange(float maval)
{ /* This returns a format based on the value `maval' */
	static char	ifor[8];

	ifor[0] = ' ';	ifor[1] = '%';	ifor[2] = '7';	ifor[3] = '.';
	ifor[4] = '2';	ifor[5] = 'f';	ifor[6] = '\0';
	if (maval > 1.e7)
	  {
		ifor[2] = '9';	ifor[5] = 'e';
	  }
	else if (maval > 100000.)
	  {
		ifor[2] = '8';	ifor[4] = '1';
	  }
	else if (maval > 10000.) ifor[4] = '1';
	else if (maval > 1000.) ifor[4] = '2';	/* default */
	else if (maval > 1.) ifor[4] = '3';	else ifor[4] = '5';
	return &(ifor[0]);
}

FILE *plotsetup(int ndep, int plflags, int *errp,
 char *finame, char *pltitle, char *xlabel, char *ylabel, char *legend[])
{ /* This starts up a plotting file from X-GEN. Variables:
	ndep		number of dependent variables in the plot
	plflags		plot-command flag variable
	finame		name of plotting file
	pltitle		title of plot
	xlabel		label for X axis
	ylabel		label for Y axis.
 The routine returns the file descriptor of the plot file that is
 created and written to succesfully. It returns NULL if a fatal error occurs,
 and returns the error type in *errp;
 It returns various errors if not. The file is left open afterward. */
	int		col, colq;
	FILE		*fplo;
	static char	**dstr;
	static char	deftick[] = "xdtcsp";

	/* open the output text file that will be input to mxplot */
	*errp = EOK;
	if (0 >= ndep)
	  {
		*errp = EBADVALUE; return NULL;
	  }
	if ((NULL == finame) || ('\0' == *finame) ||
	 (NULL == (fplo = xgopen(finame, "w"))))
	  {
		*errp = EOPENERR;	return NULL;
	  }
	/* first line defines the plot characteristics */
	fprintf(fplo, "start_plot;1;0;");
	/* set up the plot definitions and the tic marks */
	if (plflags & 01) /* lowest bit on means order them last to first */
	  {
		for (col = 0, colq = ndep - 1;
		 col < ndep - 1; col++, colq--)
			fprintf(fplo, "$f[%d];", colq);
		fprintf(fplo, "$f[0]\n");
	  }
	else {
		for (col = 0; col < ndep - 1; col++)
			fprintf(fplo, "$f[%d];", col);
		fprintf(fplo, "$f[%d]\n", col);
	     }
	fprintf(fplo, "set tic ");
	/* make up tic marks. In this version we always make them up here */
	for (col = 0; col < ndep; col++)
		fprintf(fplo, "%c ", deftick[col % 6]);
	/* establish the contents of the legend if it's been specified */
	if ((NULL != legend) && (NULL != legend[0]))
	  {
		fprintf(fplo, "\nset legend");
		for (col = 0, dstr = legend; col < ndep; col++, dstr++)
		   if ((NULL != *dstr) && ('\0' != **dstr))
			fprintf(fplo, " %s", *dstr);
	  }
	fprintf(fplo, "\n");
	/* put the title, xlabel, and ylabel in if they're defined */
	if ((NULL != pltitle) && ('\0' != *(pltitle)))
		fprintf(fplo, "set title = '%s'\n", pltitle);
	if ((NULL != xlabel) && ('\0' != *(xlabel)))
		fprintf(fplo, "set xlabel = '%s'\n", xlabel);
	if ((NULL != ylabel) && ('\0' != *(ylabel)))
		fprintf(fplo, "set ylabel = '%s'\n", ylabel);
	(void)xeflush(fplo);
	return fplo;
}

int plotadd(FILE *fplo, int ndep, float inde, float *depv)
{ /* This writes a single line of data output to an already open plot file.
   The maximum number of dependent-variable entries written out will be ndep */
	int		col;
	static char	dfmt[8], ifmt[8];
	float		maxd, this, *flp;

	if (NULL == fplo) return EWRITERR; /* make sure file is available */
	if (0 >= ndep) return EBADVALUE; /* must have >= 1 dependent vbl*/
	/* establish the range of the dependent variable */
	for (maxd = col = 0, flp = depv; col < ndep; col++, flp++)
	   {
		if ((*flp < -1.e9) || (*flp > 1.e9)) break;
		if (maxd < (this = fabs(*flp))) maxd = this;
	   }
	strcpy(ifmt, frange(inde));	strcpy(dfmt, frange(maxd));
	/* now print out the single line and flush the buffer */
	fprintf(fplo, "data  ");	fprintf(fplo, ifmt, inde);
	for (col = 0, flp = depv; col < ndep; col++, flp++)
		fprintf(fplo, dfmt, *flp);
	fprintf(fplo, "\n");	(void)xeflush(fplo);	return EOK;
}

int plotfinish(FILE *fplo, float markv)
{ /* This closes up a plot file. If the variable `markv' is larger than
   -999.0, we write out a line specifying a marker on the plot */
	if (NULL == fplo) return EWRITERR;
	if (markv > -999.)
	  {
		fprintf(fplo, "marker ");
		fprintf(fplo, frange(markv), markv);	fprintf(fplo, "\n");
	  }
	(void)xeflush(fplo);	(void)fclose(fplo);	fplo = NULL;
	return EOK;	/* maybe we'll find an error condition eventually */
}

/* end of ascout.c */
