/* calinv.c -- XGEN routines for inverting the pixel-to-cm mapping

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	 3 Jan 2006: add lineend(), calwarning(); reorganize summcomp, compcal
	20 Nov 2005: inline argument declarations and strict prototyping
	26 Oct 2005: replace static global `sgsp' with arguments in routines
	12 Aug 2005: calverbose in calinv(), compcal() are more selective
	 6 Aug 2005: calinv() takes `calverbose' arg: less logging if zero
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	28 Aug 2004: routines used only here are declared as static; cosmetics
	24 Jan 2004: wptx and wpty are local to calinv() now
	21 Jan 2004: license message inserted and initial comments reordered
	14 Jan 2004: external globals (width, height, fplog, splp) replaced
		with static globals (e.g. lspl), local variables, or (in the
		case of fplog) a function argument to compcal() and calinv()
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	 5 Jan 2003: reorganized source-file header
	25 Mar 2002: removed MSI attachments
	24 Aug 1995: SDMS RMS errors in centroid position calc'd correctly
	31 May 1995: "register" -> "register int" throughout
	12 Apr 1995: MAX_DSQ = 30
	 2 Mar 1995: major rewrite: validation done correctly
		final positional refinement for self-consistency
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: polymask added; bugfixes in polymath, xtal, calinv, etc.
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version

   Purpose: manipulations of the table of pixel-to-centimeter values
	used in relating detector coordinates to lab coordinates.
   Cross references: ../prerefine/cal*.c

	This takes the table of pixel-to-centimeter values and computes
	the inverse table, i.e. the table of cm-to-pix values.
	It then recomputes the initial tables, i.e. the tables that take
	us from pixels to centimeters and calculates rms differences.
*************************************************************************/
#include	"xgen.h"
#define		MAX_RMS		2.
#define		MAX_DSQ		30.
#define		XYSHIFT		(0.4)

 typedef struct {
		double	dxmean, dymean, xrms, yrms, dsqmax;
		xglong	npts;
		} COMPARER;

static void compstat(COMPARER *tcmp)
{ /* this does the necessary divisions to make statistical comparisons */
	double	rn;

	if ((rn = tcmp->npts))
	  {
		tcmp->xrms = sqrt(tcmp->xrms / rn);
		tcmp->yrms = sqrt(tcmp->yrms / rn);
		tcmp->dxmean /= rn;	tcmp->dymean /= rn;
	  }
}

char *lineend(FILE *fp, int doprint)
{ /*	Returns either "\n" or "\r\n" depending on the inputs.
	If fp is stdout or stdin and doprint is odd, then we return "\r\n".
	Otherwise we return "\n". */
	int		ix;
	static char	*liter[] = {"\n", "\r\n"} ;

	ix = ((fp == stdout) || (fp == stderr)) && (doprint & 01);
	return liter[ix];
}

static void summcomp(FILE *fp, COMPARER *tc0, int doprint)
{ /* This prints the deviation summaries to the stream fp */
	char		*endd;
	COMPARER	*tc1, *tc2;

	endd = lineend(fp, doprint);
	tc1 = tc0 + 1;	tc2 = tc0 + 2;
	fprintf(fp,
 "Deviations of recomputed (inverse-spline) points from inputs:%s", endd);
	fprintf(fp,
 "             <Dev(x)>  <Dev(y)>  RMS Dev(x)  RMS Dev(y)  Max Dev # pts%s",
		endd);
	fprintf(fp, "  All Points:%8.4lf %9.4lf %11.5lf %11.5lf %8.5lf %5d%s",
		tc0->dxmean, tc0->dymean, tc0->xrms,
		tc0->yrms, sqrt(tc0->dsqmax), (int)tc0->npts, endd);
	fprintf(fp, "    Observed:%8.4lf %9.4lf %11.5lf %11.5lf %8.5lf %5d%s",
		tc1->dxmean, tc1->dymean, tc1->xrms,
		tc1->yrms, sqrt(tc1->dsqmax), (int)tc1->npts, endd);
	fprintf(fp, "Interpolated:%8.4lf %9.4lf %11.5lf %11.5lf %8.5lf %5d%s",
		tc2->dxmean, tc2->dymean, tc2->xrms,
		tc2->yrms, sqrt(tc2->dsqmax), (int)tc2->npts, endd);
}

static double cross3(CALPOINT *cp0, CALPOINT *cp1, CALPOINT *cp2)
{
/* This returns the cross product of the vector from cp0 to cp1 with
	the vector from cp0 to cp2 */
	return (cp1->c_ptocx - cp0->c_ptocx) * (cp2->c_ptocy - cp0->c_ptocy) -
		(cp2->c_ptocx - cp0->c_ptocx) * (cp1->c_ptocy - cp0->c_ptocy);
}

static int cont3(CALPOINT *cp, CALPOINT *cp0, CALPOINT *cp1, CALPOINT *cp2)
{
/* This returns true iff. cp is contained within the angle defined by
	the vector from cp0 to cp1 and the vector from cp0 to cp2 */

	if (0. <= cross3(cp0, cp1, cp2))
	  {
		return (cross3(cp0, cp1, cp) > 0. &&
		 cross3(cp0, cp, cp2) > 0.);
	  }
	else return (cross3(cp0, cp1, cp) < 0. &&
		cross3(cp0, cp, cp2) < 0.);
}

static double calpdsq(CALPOINT *ca0, CALPOINT *ca1)
{ /* this computes the square of the distance between two points */
	double	dex, dey;

	dex = ca0->c_ptocx - ca1->c_ptocx; dey = ca0->c_ptocy - ca1->c_ptocy;
	return dex * dex + dey * dey;
}

static double get_bisect(CALPOINT *ca, CALPOINT *ca0, CALPOINT *ca1)
{ /*	This calculates the distance from the point *ca
	 to the line that contains the points *ca0 and *ca1. */
	double	e0, e1, del, num;

	del = calpdsq(ca1, ca0); e0 = calpdsq(ca, ca0);	e1 = calpdsq(ca, ca1);
	num = 2. * (e0 * del + e1 * del + e0 * e1) -
		e0 * e0 - e1 * e1 - del * del;
	return (num > DFUZZ) ? sqrt(num / (4. * del)) : 0.;
}

static CALPOINT *fakecal(VSPLINE *lspl, int jy, int jx, int whichcorner)
{ /* This extrapolates to create a fake centroid out beyond what we have. */
	int		nc, colok, rowok;
	double		colslope, rowslope, denom, colint, rowint;
	static CALPOINT	tc, *cp0, *cp1, *cp2, *cp4, *cp5;

	/* zero out the structure that will hold the value */
	zeromem((AVOID *)&tc, sizeof (CALPOINT));
	if ((jy >= lspl->v_nrow) || (jx >= lspl->v_ncol)) return &tc;
	colok = rowok = 0;
	cp0 = lspl->v_cpt + jy * lspl->v_ncol + jx;	nc = lspl->v_ncol;
	/* figure out which points we can extrapolate from. */
	if (whichcorner == 0 || whichcorner == 2)
	  {
		if (jy < lspl->v_nrow - 2)
		  {
			cp1 = cp0 + nc; cp4 = cp1 + nc;
		  }
		else return &tc;	/* we'll do better later */
	  }
	else {
		if (jy > 1)
		  {
			cp1 = cp0 - nc;	cp4 = cp1 - nc;
		  }
		else return &tc;
	     }
	if (whichcorner == 0 || whichcorner == 1)
	  {
		if (jx < lspl->v_ncol - 2)
		  {
			cp2 = cp0 + 1; cp5 = cp2 + 1;
		  }
		else return &tc;	/* we'll do better later */
	  }
	else {
		if (jx > 1)
		  {
			cp2 = cp0 - 1;	cp5 = cp2 - 1;
		  }
		else return &tc;
	     }
	if ((cp1->c_set) && (cp4->c_set)) colok = 1;
	if ((cp2->c_set) && (cp5->c_set)) rowok = 1;
	if (colok)
	  {
		if (rowok)	/* extrapolate to intersection of the 2 lines*/
		  {
			colslope = cp4->c_ptocy - cp1->c_ptocy;
			if (fabs(colslope) < DFUZZ) return &tc;
			colslope = (cp4->c_ptocx - cp1->c_ptocx) / colslope;
			rowslope = cp5->c_ptocx - cp2->c_ptocx;
			if (fabs(rowslope) < DFUZZ) return &tc;
			rowslope = (cp5->c_ptocy - cp2->c_ptocy) / rowslope;
			denom = 1. - rowslope * colslope;
			if (fabs(denom) < DFUZZ) return &tc;
			colint = cp1->c_ptocx - colslope * cp1->c_ptocy;
			rowint = cp2->c_ptocy - rowslope * cp2->c_ptocx;
			tc.c_ptocx = (colint + colslope * rowint) / denom;
			tc.c_ptocy = (rowint + rowslope * colint) / denom;
		  }
		else	/* extrapolate out along the column */
		  {
			tc.c_ptocx = 2. * cp1->c_ptocx - cp4->c_ptocx;
			tc.c_ptocy = 2. * cp1->c_ptocy - cp4->c_ptocy;
		  }
	  }
	else if (rowok)	/* extrapolate out along the row */
	  {
		tc.c_ptocx = 2. * cp2->c_ptocx - cp5->c_ptocx;
		tc.c_ptocy = 2. * cp2->c_ptocy - cp5->c_ptocy;
	  }
	else return &tc;	/* we didn't get one */
	tc.c_set = 1;	/* we got one! */
	return &tc;
}

static int regionno(VSPLINE *lspl, CALPOINT *calp)
{ /* this returns the region from which a point is derived.
    For any detector with only one region the routine returns 0. */
	int	thisquad;

	if (lspl->v_nreg < 4) return 0;
	thisquad = calp->c_ptocx > lspl->v_detw / 2;
	if (calp->c_ptocy > lspl->v_deth / 2) thisquad += 2;
	return thisquad;
}

static int rawpix2cm(VSPLINE *lspl, float xpx, float ypx, float *xc, float *yc)
{ /*	This returns the x,y coordinates in cm of a position specified
	in pixels.  This version returns 1 if all the corners
	surrounding the point are defined; it returns 0 otherwise. */
	static int	ix, iy, iybest, ixbest, thisquad, thatquad;
	static int	iymin, iymax, ixmin, ixmax, iyuse, ixuse;
	int		ixl, ixh, iyl, iyh, maxsize;
	static double	cdmin, cdsq, bis0, bis1, sumb;
	static CALPOINT	cau, *calp, *cab, ca0, ca1, ca2, ca3;
	static CALPOINT	*cp0, *cp1, *cp2, *cp3;

	cau.c_ptocx = xpx;	cau.c_ptocy = ypx;	cau.c_set = 1;
	/* identify the observed point closest to this one.
	 We assume at this point that the detector dimensions stored
	 in the VSPLINE structure lspl are accurate */
	maxsize = lspl->v_detw * lspl->v_deth;
	cdmin = maxsize;	cdmin *= cdmin;
	/* if we're in the special case of a 2x2 detector, we require
	 that the points be within the current quadrant */
	thisquad = regionno(lspl, &cau);
	for (iy = 0, calp = lspl->v_cpt; iy < lspl->v_nrow; iy++)
	   for (ix = 0; ix < lspl->v_ncol; ix++, calp++)
	      {
		thatquad = regionno(lspl, calp);
		if ((thisquad == thatquad) && (calp->c_set))
		  {
			if ((cdsq = calpdsq(&cau, calp)) < cdmin)
			  {
				iybest = iy;	ixbest = ix;
				cdmin = cdsq;	cab = calp;
			  }
		  }
	     }
	if (cdmin >= maxsize) return 0;
	/* if cdmin is very small, we figure we have it right already.
	 We therefore save it. */
	if (cdmin < 5. * DFUZZ)
	  {
		*xc = lspl->v_x0cm + ((double)ixbest) * lspl->v_xrcm;
		*yc = lspl->v_y0cm + ((double)iybest) * lspl->v_yrcm;
		return 1;
	  }
	/* having found the closest one, look in its neighborhood to find
	 a quadrilateral of neighboring points for which our reference value
	 is contained inside the quadrilateral. */
	iymin = iybest - 2;	ixmin = ixbest - 2;
	ixmax = ixbest + 2;	iymax = iybest + 2;	ixuse = iyuse = -1;
	if (lspl->v_nreg > 1)
	  {
		if (thisquad % 2) /* right-hand half */
		  {
			ixl = lspl->v_midc;	ixh = lspl->v_ncol - 1;
		  }
		else {	/* left half */
			ixl = 0;	ixh = lspl->v_midc - 1;
		     }
		if (thisquad > 1) /* bottom half */
		  {
			iyl = lspl->v_midr;	iyh = lspl->v_nrow - 1;
		  }
		else {	/* top half*/
			iyl = 0;	iyh = lspl->v_midr - 1;
		     }
	  }
	else {
		ixl = 0;	ixh = lspl->v_ncol - 1;
		iyl = 0;	iyh = lspl->v_nrow - 1;
	     }
	if (iymin < iyl) iymin = iyl;
	if (iymax > iyh - 1) iymax = iyh - 1;
	if (ixmin < ixl) ixmin = ixl;
	if (ixmax > ixh - 1) ixmax = ixh - 1;
	for (iy = iymin; iy <= iymax; iy++)
	   {
		for (ix = ixmin; ix <= ixmax; ix++)
		   {
			cp0 = lspl->v_cpt + lspl->v_ncol * iy + ix;
			cp1 = cp0 + lspl->v_ncol; cp2 = cp0 + 1; cp3 = cp1 + 1;
			if (cont3(&cau, cp0, cp1, cp2) &&
				cont3(&cau, cp3, cp1, cp2))
			  {
				iyuse = iy;	ixuse = ix;	break;
			  }
		   }
		if (iyuse >= 0) break;
	   }
	if (iyuse >= 0 && ixuse >= 0) /* we found one! */
	  {
		cpymem((AVOID *)&ca0, (AVOID *)cp0, sizeof (CALPOINT));
		cpymem((AVOID *)&ca1, (AVOID *)cp1, sizeof (CALPOINT));
		cpymem((AVOID *)&ca2, (AVOID *)cp2, sizeof (CALPOINT));
		cpymem((AVOID *)&ca3, (AVOID *)cp3, sizeof (CALPOINT));
		iy = iyuse;	ix = ixuse;
	  }
	else {
		/* if we don't find a quadrilateral containing our point,
		 we try to create one. For now, we do this simple-mindedly*/
		ix = ixbest;	iy = iybest;
		if ((ix) && (xpx < cab->c_ptocx))
		  {
			cab--;	ix--;
		  }
		if ((iy) && (ypx < cab->c_ptocy))
		  {
			cab -=lspl->v_ncol; iy--;
		  }
		cpymem((AVOID *)&ca0, (AVOID *)cab, sizeof (CALPOINT));
		thatquad = regionno(lspl, &ca0);
		if ((thisquad != thatquad) || !(ca0.c_set)) cpymem((AVOID *)
		 &ca0, (AVOID *)fakecal(lspl, iy, ix,0), sizeof (CALPOINT));
		if (!(ca0.c_set)) return 0;
		if (iy < lspl->v_nrow - 1)
		  {
			cpymem((AVOID *)&ca1, (AVOID *)(cab + lspl->v_ncol),
				sizeof (CALPOINT));
			thatquad = regionno(lspl, &ca1);
			if ((thisquad != thatquad) || !(ca1.c_set))
			 cpymem((AVOID *)&ca1, (AVOID *)fakecal(lspl,
				iy + 1, ix, 1), sizeof (CALPOINT));
		  }
		else	cpymem((AVOID *)&ca1, (AVOID *)fakecal(lspl,
				iy + 1, ix, 1), sizeof (CALPOINT));
		if (!(ca1.c_set)) return 0;
		if (ix < lspl->v_ncol - 1)
		  {
			cpymem((AVOID *)&ca2,
			 (AVOID *)(cab+1), sizeof (CALPOINT));
			thatquad = regionno(lspl, &ca2);
			if ((thisquad != thatquad) || !(ca2.c_set))
			 cpymem((AVOID *)&ca2, (AVOID *)fakecal(lspl,
				iy, ix + 1, 2), sizeof (CALPOINT));
		  }
		else	cpymem((AVOID *)&ca2, (AVOID *)fakecal(lspl,
				iy, ix + 1, 2), sizeof (CALPOINT));
		if (!(ca2.c_set)) return 0;
		if ((iy < lspl->v_nrow -1) && (ix < lspl->v_ncol - 1))
		  {
			cpymem((AVOID *)&ca3, (AVOID *)(cab +
			 lspl->v_ncol + 1), sizeof (CALPOINT));
			thatquad = regionno(lspl, &ca3);
			if ((thisquad != thatquad) && !(ca3.c_set))
			 cpymem((AVOID *)&ca3, (AVOID *)fakecal(lspl,
				iy +1, ix +1, 3), sizeof (CALPOINT));
		  }
		else	cpymem((AVOID *)&ca3, (AVOID *)fakecal(lspl,
				iy + 1, ix +1,3), sizeof (CALPOINT));
		if (!(ca3.c_set)) return 0;
		/* make sure our point really lies within the quadrlateral */
		if ((!cont3(&cau, &ca0, &ca1, &ca2)) ||
			(!cont3(&cau, &ca3,&ca1,&ca2)))	return 0;
	     }
	/* at this point we have all of the centroids that surround our
	 reference point.  We determine the perpendicular bisectors that
	 relate the reference point to our point and use these to get the
	 interpolation working */
	bis0 = get_bisect(&cau, &ca0, &ca1);
	bis1 = get_bisect(&cau, &ca3, &ca2);
	if ((sumb = bis0 + bis1) < DFUZZ) return 0;
	*xc = lspl->v_x0cm + (bis0 / sumb + ((double)ix)) * lspl->v_xrcm;
	bis0 = get_bisect(&cau, &ca0, &ca2);
	bis1 = get_bisect(&cau, &ca3, &ca1);
	if ((sumb = bis0 + bis1) < DFUZZ) return 0;
	*yc = lspl->v_y0cm + (bis0 / sumb + ((double)iy)) * lspl->v_yrcm;
	return 1;
}

static int safecm2pix(VSPLINE *lspl, float xcm, float ycm, float *xp,float *yp)
{ /*	This returns the x,y coordinates in pixels of a position specified
	in cm. This version returns 1 if all the corners surrounding the
	point are defined; it returns 0 otherwise. */
	static int	ix, iy;
	static double	dx, dy, rx, ry, p, q;
	static CALPOINT	*ca00, *ca01, *ca10, *ca11;

	dx = xcm - lspl->v_x0cm;		rx = dx / lspl->v_xrcm;
	ix = (rx < 0. ? 0 : (rx < lspl->v_ncol - 1 ? rx : lspl->v_ncol - 2));
	p = rx - ix;
	dy = ycm - lspl->v_y0cm;		ry = dy / lspl->v_yrcm;
	iy = (ry < 0. ? 0 : (ry < lspl->v_nrow - 1 ? ry : lspl->v_nrow - 2));
	q = ry - iy;
	ca00 = lspl->v_cpt + iy * lspl->v_ncol + ix;
	ca01 = ca00 + 1; ca10 = ca00 + lspl->v_ncol; ca11 = ca10 + 1;
	*xp = *yp = 0.;	/* failure values */
	 /* if upper left point is not yet defined, try moving down or right*/
	if (!(ca00->c_set))
	  {
		if (!(ca11->c_set)) return 0; /*lower right must be present*/
		if (ca01->c_set)		/* urhc present: increment x */
		  {
			if ((++ix) >= (lspl->v_ncol - 1)) return 0;
			ca00++; ca01++; ca10++; ca11++;
			p -= 1.;
		  }
		else if (ca10->c_set)	/* lrhc present: increment y */
		  {
			if ((++iy) >= (lspl->v_nrow - 1)) return 0;
			ca00 += lspl->v_ncol;	ca01 += lspl->v_ncol;
			ca10 += lspl->v_ncol;	ca11 += lspl->v_ncol;
			q -= 1.;
		  }
		else return 0;	/* give up in this case. */
	  }
	else if (!(ca11->c_set))
	  {
		if (ca01->c_set)	/* ulhc present: decrement y */
		  {
			if ((--iy) < 0) return 0;
			ca00 -= lspl->v_ncol;	ca01 -= lspl->v_ncol;
			ca10 -= lspl->v_ncol;	ca11 -= lspl->v_ncol;
			q += 1.;
		  }
		else if (ca10->c_set)	/* lrhc present: decrement x */
		  {
			if ((--ix) < 0) return 0;
			ca00--;	ca01--;	ca10--;	ca11--;
			p += 1.;
		  }
		else return 0;
	  }
	/* having done all that malarkey, we see if all 4 corners are defined*/
	if ((!ca00->c_set) || (!ca10->c_set) ||
		(!ca01->c_set) || (!ca11->c_set)) return 0;
	*xp = (1.-p) * (1.-q) * ca00->c_ptocx + (1.-p) * q * ca10->c_ptocx +
		  p  * (1.-q) * ca01->c_ptocx +     p  * q * ca11->c_ptocx;
	*yp = (1.-p) * (1.-q) * ca00->c_ptocy + (1.-p) * q * ca10->c_ptocy +
		  p  * (1.-q) * ca01->c_ptocy +     p  * q * ca11->c_ptocy;
	return 1;
}

static int locctop(VSPLINE *lspl, double xci, double yci,
 double *xppo, double *yppo)
{ /* This returns a pixel position given a centimeter position.
	It returns 0 iff. successful. */
	double		dx, dy, rx, ry, p, pp, q, qq;
	int		jx, jy;
	CALPOINT	*ca00, *ca01, *ca10, *ca11;

	dx = xci - lspl->v_x0cm;	dy = yci - lspl->v_y0cm;
	rx = dx / lspl->v_xrcm;	ry = dy / lspl->v_yrcm;
	jx = (rx < 0.) ? 0 : ((rx < lspl->v_ncol - 1) ?
		rx : lspl->v_ncol - 2);
	p = rx - (double)jx;	pp = 1. - p;
	jy = (ry < 0.) ? 0 : ((ry < lspl->v_nrow - 1) ?
		ry : lspl->v_nrow - 2);
	q = ry - (double)jy;	qq = 1. - q;
	ca00 = lspl->v_cpt + jy * lspl->v_ncol + jx;
	ca01 = ca00 + 1; ca10 = ca00 + lspl->v_ncol; ca11 = ca10 + 1;
	if ((!(ca00->c_cset)) || (!(ca01->c_cset)) ||
	 (!(ca10->c_cset)) || (!(ca11->c_cset))) return 1;
	*xppo =	pp * qq * ca00->c_ptocx + pp * q * ca10->c_ptocx +
		 p * qq * ca01->c_ptocx +  p * q * ca11->c_ptocx;
	*yppo =	pp * qq * ca00->c_ptocy + pp * q * ca10->c_ptocy +
		 p * qq * ca01->c_ptocy +  p * q * ca11->c_ptocy;
	return 0;
}

static int locptoc(VSPLINE *lspl, double xpi, double ypi,
 double *xpco, double *ypco)
{ /* This converts from pixels to centimeters. It returns
	0 iff. successful */
	double		dx, dy, rx, ry, p, pp, q, qq;
	int		jx, jy;
	CALPOINT	*ca00, *ca01, *ca10, *ca11;

	dx = xpi - lspl->v_x0px;	dy = ypi - lspl->v_y0px;
	rx = dx / lspl->v_xrpx;	ry = dy / lspl->v_yrpx;
	jx = (rx < 0.) ? 0 : ((rx < lspl->v_ncol - 1) ?
		rx : lspl->v_ncol - 2);
	p = rx - (double)jx;	pp = 1. - p;
	jy = (ry < 0.) ? 0 : ((ry < lspl->v_nrow - 1) ?
		ry : lspl->v_nrow - 2);
	q = ry - (double)jy;	qq = 1. - q;
	ca00 = lspl->v_cpt + jy * lspl->v_ncol + jx;
	ca01 = ca00 + 1; ca10 = ca00 + lspl->v_ncol; ca11 = ca10 + 1;
	if ((!(ca00->c_set)) || (!(ca01->c_set)) ||
	 (!(ca10->c_set)) || (!(ca11->c_set))) return 1;
	*xpco =	pp * qq * ca00->c_ctopx + pp * q * ca10->c_ctopx +
		 p * qq * ca01->c_ctopx +  p * q * ca11->c_ctopx;
	*ypco =	pp * qq * ca00->c_ctopy + pp * q * ca10->c_ctopy +
		 p * qq * ca01->c_ctopy +  p * q * ca11->c_ctopy;
	return 0;
}

static int bothmap(VSPLINE *lspl, double xi, double yi,
 double *xop, double *yop, int direction)
{ /* This maps from pix to cm and back or vice versa */
	static int	locwid, locht;
	double		xmd, ymd, xli, yli;

	xli = fabs(lspl->v_xrpx);	yli = fabs(lspl->v_yrpx);
	locwid = lspl->v_detw;	locht = lspl->v_deth;
	*xop = xi;	*yop = yi;
	if (direction)
	  {
		if ((xi < xli) || (xi > locwid + xli)
		 || (yi < yli) || (yi > yli + locht)) return 1;
		if ((!locptoc(lspl, xi, yi, &xli, &yli)) &&
		 (!locctop(lspl, xli, yli, xop, yop))) return 0;
	  }
	else {
		if ((!locptoc(lspl, xi, yi, &xmd, &ymd)) &&
		 (!locctop(lspl, xmd, ymd, xop, yop))) return 0;
	      }
	return 1;
}

static int near_here(VSPLINE *lspl, CALPOINT *clp, int direction)
{ /* This adjusts the fiducial-point definition for the point clp. It
	returns 1 if a noticable change occurs, 0 otherwise */
	int		ip, nfou;
	static int	locwid, locht;
	double		ofx, ofy, dex, dey, shiftf;
	double		xu[9], yu[9], xuo[9], yuo[9];
	float		*xup, *yup;

	locwid = lspl->v_detw;	locht = lspl->v_deth;
	if (direction)
	  {
		xup = &(clp->c_ctopx);		yup = &(clp->c_ctopy);
		ofx = lspl->v_xrcm * 0.1;	ofy = lspl->v_yrcm * 0.1;
		shiftf = XYSHIFT * 0.1;
	  }
	else  {
		xup = &(clp->c_ptocx);		yup = &(clp->c_ptocy);
		/* in this direction we ignore the point if it's 'way off
		 from the detector face itself.*/
		ofx = fabs(lspl->v_xrpx);	ofy = fabs(lspl->v_yrpx);
		if ((*xup < -ofx) || (*xup > locwid + ofx)) return 0;
		if ((*yup < -ofy) || (*yup > locht + ofy)) return 0;
		ofx = lspl->v_xrpx * 0.4;	ofy = lspl->v_yrpx * 0.4;
		shiftf = XYSHIFT;
	      }
	xu[1] = xu[4] = xu[7] = *xup;
	xu[0] = xu[3] = xu[6] = *xup - ofx;
	xu[2] = xu[5] = xu[8] = *xup + ofx;
	yu[3] = yu[4] = yu[5] = *yup;
	yu[0] = yu[1] = yu[2] = *yup - ofy;
	yu[6] = yu[7] = yu[8] = *yup + ofy;
	for (dex = dey = nfou = ip = 0; ip < 9; ip++)
	   {
		if (!bothmap(lspl, xu[ip], yu[ip], &xuo[ip],
		 &yuo[ip], direction))
		  {
			dex += xu[ip] - xuo[ip];
			dey += yu[ip] - yuo[ip]; nfou++;
		  }
	   }
	if (nfou < 3) dex = dey = 0.;
	else {
		dex /= (double)nfou;	dey /= (double)nfou;
		dex = dflimits(dex, -ofx * 0.6, ofx * 0.6);
		dey = dflimits(dey, -ofy * 0.6, ofy * 0.6);
		*xup += dex * shiftf;	*yup += dey * shiftf;
	     }
	return (fabs(dex) > ofx * 0.01) || (fabs(dey) > ofy * 0.01);
}

static int adjust_xy(VSPLINE *lspl)
{ /* This tweaks the xy centroids a little */
	register int	iy;
	int		ncchange, npchange, nxy, nxyo2;
	CALPOINT	*calp;

	/* find nine points in the vicinity of each fiducial point in cm.
	 Figure out how much it "moves" after converting to pixels and
	 back to cm and shift the center value by a fraction of the move. */
	ncchange = npchange = 0;	nxy = lspl->v_nrow * lspl->v_ncol;
	nxyo2 = nxy / 2;
	for (iy = nxyo2; (iy >= 0) && (iy < nxy); iy = nxy - iy)
	   {
		calp = lspl->v_cpt + iy;
		if (!calp->c_ift) ncchange += near_here(lspl, calp, 0);
		if (iy * 2 >= nxy) iy++;
	   }
	for (iy = nxyo2; (iy >= 0) && (iy < nxy); iy = nxy - iy)
	   {
		calp = lspl->v_cpt + iy;
		if (!calp->c_ift) npchange += near_here(lspl, calp, 1);
		if (iy * 2 >= nxy) iy++;
	   }
	return ncchange + npchange;
}

static void calwarning(FILE *fp, int doprint)
{ /* emits a warning to the stream fp */
	char	*endd;

	endd = lineend(fp, doprint);
	fprintf(fp,
 "Warning! The large deviations between observed and recomputed (X,Y)%s",
		endd);
	fprintf(fp,
	"\tindicate that this calibration is untrustworthy.%s", endd);
	fprintf(fp,
 "You may want to re-run CALIBRATE with a longer brass-plate exposure%s", endd);
	fprintf(fp, "\tor a different cutoff intensity value%s", endd);
}

static double compcal(VSPLINE *lspl, FILE *fp, int compcfl)
{ /*	This compares the centroid positions with the interpolated values.
	The argument compcfl & 02 iff. we want the results printed to fp.
	The routine returns the overall mean rms error. */
	register int	iy, ix;
	double		diffsq, diffsqx, diffsqy, diffx, diffy;
	double		dx, dy, htlim, inx, iny, widlim;
	CALPOINT	*calp;
	COMPARER	*tcmp;
	static COMPARER	*calco;
	static char	calmsg[] = "calco in compcal";

	if ((NULL == calco) && (NULL == (calco = (COMPARER *)callop(3,
	 sizeof (COMPARER), calmsg)))) return 10000.;
	zeromem((AVOID *)(calco + 1), 2 * sizeof (COMPARER));
	for (iy = 0, calp = lspl->v_cpt; iy < lspl->v_nrow; iy++)
	  for (ix = 0; ix < lspl->v_ncol; ix++, calp++)
	    if (calp->c_set)
	      {
		dx = 0.005 * (double)((13 * ix + 17 * iy + 19) % 40);
		dy = 0.005 * (double)((17 * ix + 19 * iy + 21) % 40);
		if ((3 * ix + 2 * iy) % 2) dx = -dx;
		if ((2 * ix - 7 * iy) % 2) dy = -dy;
		inx = calp->c_ptocx + dx * lspl->v_xrpx;
		iny = calp->c_ptocy + dy * lspl->v_yrpx;
		if (!bothmap(lspl, inx, iny, &dx, &dy, 1))
		  {
			diffx = inx - dx;	 diffy = iny - dy;
			diffsqx = diffx * diffx; diffsqy = diffy * diffy;
			diffsq = diffsqx + diffsqy;
			tcmp = calp->c_ift ? calco + 1 : calco + 2;
			tcmp->xrms += diffsqx;	tcmp->yrms += diffsqy;
			tcmp->dxmean += diffx;	tcmp->dymean += diffy;
			if (diffsq > tcmp->dsqmax) tcmp->dsqmax = diffsq;
			(tcmp->npts)++;
		  }
	      }
	calco->npts = (calco+1)->npts + (calco+2)->npts;
	calco->xrms = (calco+1)->xrms + (calco+2)->xrms;
	calco->yrms = (calco+1)->yrms + (calco+2)->yrms;
	calco->dxmean = (calco+1)->dxmean + (calco+2)->dxmean;
	calco->dymean = (calco+1)->dymean + (calco+2)->dymean;
	calco->dsqmax = (calco+2)->dsqmax;
	if ((calco+1)->dsqmax > (calco+2)->dsqmax)
		calco->dsqmax = (calco+1)->dsqmax;
	compstat(calco + 1);	compstat(calco + 2);	compstat(calco);
	if (compcfl & 02)
	  {
		summcomp(fp, calco, compcfl);
		summcomp(stdout, calco, compcfl);
		fprintf(fp, "Filled values:\n");
		for (iy = 0, calp = lspl->v_cpt; iy < lspl->v_nrow; iy++)
		   {
			for (ix = 0; ix < lspl->v_ncol; ix++, calp++)
			   {
				fprintf(fp, " %5.1f", calp->c_ptocy);
				if (!((ix+1) % 22)) fprintf(fp, "\n");
			   }
			if (ix % 22) fprintf(fp,"\n"); calp -= ix;
			for (ix = 0; ix < lspl->v_ncol; ix++, calp++)
			   {
				fprintf(fp, " %5.1f", calp->c_ptocx);
				if (!((ix+1) % 22)) fprintf(fp, "\n");
			   }
			if (ix % 22) fprintf(fp, "\n\n");
			fprintf(fp,"\n");
		   }
		widlim = ((double)(lspl->v_detw)) / 512.;
		htlim = ((double)(lspl->v_deth)) / 512.;
		if (((calco+1)->xrms > MAX_RMS * widlim) ||
		 ((calco+1)->yrms > MAX_RMS * htlim) ||
		 ((calco+1)->dsqmax > MAX_DSQ * widlim * htlim))
		  {
			calwarning(stderr, compcfl); calwarning(fp, compcfl);
		  }
	  }
	iny = sqrt(calco->xrms * calco->xrms + calco->yrms * calco->yrms);
	if (compcfl & 04) memzap((AVOID **)&calco, calmsg);
	return iny;
}

static int strapc(VSPLINE *lspl, int iy, int ix)
{ /* This does ultra-simple-minded extrapolations on the ctop arrays. */
	static int	iyd, ixd, colok, rowok;
	static CALPOINT	*cp0, *cp1, *cp2, *cp4, *cp5;
	static double	colslope, rowslope, colint, rowint, denom;

	cp0 = lspl->v_cpt + lspl->v_ncol * iy + ix;
	if (cp0->c_cset) return 1;
	iyd = (iy < lspl->v_midr ? 1 : -1);
	ixd = (ix < lspl->v_midc ? 1 : -1);
	cp1 = cp0 + lspl->v_ncol * iyd;	cp4 = cp1 + lspl->v_ncol * iyd;
	cp2 = cp0 + ixd;	cp5 = cp2 + ixd;
	colok = (cp1->c_cset && cp4->c_cset);
	rowok = (cp2->c_cset && cp5->c_cset);
	if (colok)
	  {
		if (rowok)	/* extrapolate to intersection of the 2 lines*/
		  {
			colslope = cp4->c_ctopy - cp1->c_ctopy;
			if (fabs(colslope) < DFUZZ) return 0;
			colslope = (cp4->c_ctopx - cp1->c_ctopx) / colslope;
			rowslope = cp5->c_ctopx - cp2->c_ctopx;
			if (fabs(rowslope) < DFUZZ) return 0;
			rowslope = (cp5->c_ctopy - cp2->c_ctopy) / rowslope;
			denom = 1. - rowslope * colslope;
			if (fabs(denom) < DFUZZ) return 0;
			colint = cp1->c_ctopx - colslope * cp1->c_ctopy;
			rowint = cp2->c_ctopy - rowslope * cp2->c_ctopx;
			cp0->c_ctopx = (colint + colslope * rowint) / denom;
			cp0->c_ctopy = (rowint + rowslope * colint) / denom;
		  }
		else	/* extrapolate out along the column */
		  {
			cp0->c_ctopx = 2. * cp1->c_ctopx - cp4->c_ctopx;
			cp0->c_ctopy = 2. * cp1->c_ctopy - cp4->c_ctopy;
		  }
	  }
	else if (rowok)	/* extrapolate out along the row */
	  {
		cp0->c_ctopx = 2. * cp2->c_ctopx - cp5->c_ctopx;
		cp0->c_ctopy = 2. * cp2->c_ctopy - cp5->c_ctopy;
	  }
	else return 0;
	cp0->c_cset = 1;	/* we got one! */
	return 1;
}

static int strapp(VSPLINE *lspl, int iy, int ix)
{ /* This does ultra-simple-minded extrapolations on the ptoc arrays. */
	static int	iyd, ixd, colok, rowok;
	static CALPOINT	*cp0, *cp1, *cp2, *cp4, *cp5;
	static double	colslope, rowslope, colint, rowint, denom;

	cp0 = lspl->v_cpt + lspl->v_ncol * iy + ix;
	if (cp0->c_set) return 1;
	iyd = (iy < lspl->v_midr ? 1 : -1);
	ixd = (ix < lspl->v_midc ? 1 : -1);
	cp1 = cp0 + lspl->v_ncol * iyd;	cp4 = cp1 + lspl->v_ncol * iyd;
	cp2 = cp0 + ixd;	cp5 = cp2 + ixd;
	colok = (cp1->c_set && cp4->c_set);
	rowok = (cp2->c_set && cp5->c_set);
	if (colok)
	  {
		if (rowok)	/* extrapolate to intersection of the 2 lines*/
		  {
			colslope = cp4->c_ptocy - cp1->c_ptocy;
			if (fabs(colslope) < DFUZZ) return 0;
			colslope = (cp4->c_ptocx - cp1->c_ptocx) / colslope;
			rowslope = cp5->c_ptocx - cp2->c_ptocx;
			if (fabs(rowslope) < DFUZZ) return 0;
			rowslope = (cp5->c_ptocy - cp2->c_ptocy) / rowslope;
			denom = 1. - rowslope * colslope;
			if (fabs(denom) < DFUZZ) return 0;
			colint = cp1->c_ptocx - colslope * cp1->c_ptocy;
			rowint = cp2->c_ptocy - rowslope * cp2->c_ptocx;
			cp0->c_ptocx = (colint + colslope * rowint) / denom;
			cp0->c_ptocy = (rowint + rowslope * colint) / denom;
		  }
		else	/* extrapolate out along the column */
		  {
			cp0->c_ptocx = 2. * cp1->c_ptocx - cp4->c_ptocx;
			cp0->c_ptocy = 2. * cp1->c_ptocy - cp4->c_ptocy;
		  }
	  }
	else if (rowok)	/* extrapolate out along the row */
	  {
		cp0->c_ptocx = 2. * cp2->c_ptocx - cp5->c_ptocx;
		cp0->c_ptocy = 2. * cp2->c_ptocy - cp5->c_ptocy;
	  }
	else return 0;
	cp0->c_set = 1;	/* we got one! */
	return 1;
}

static void tell_invbegin(FILE *fp, int nptsp, int nptsc, int nxy, int calv)
{ /*	Reports on the beginning of the inversion process to stream fp */
	static char	*endd;

	if (!(calv & 02)) return;	endd = lineend(fp, calv);
	fprintf(fp,
 " %4d of %4d points have pixel-space definitions when the effort begins%s",
		nptsp, nxy, endd);
	fprintf(fp, "Forward interpolation: %4d of %4d points defined%s",
	 nptsc, nxy, endd);
}

static void tell_extrap(FILE *fp, int nptsc, int nxy, int nptsp, int calv)
{ /*	Reports on the extrapolation results */
	char	*endd;

	if (!(calv & 02)) return;	endd = lineend(fp, calv);
	fprintf(fp,
 " Extrapolation results: %4d of %4d cm points, %4d of %4d pixel points%s",
		nptsc, nxy, nptsp, nxy, endd);
}

static void tell_safe(FILE *fp, int nalr, int nptsp, int calv)
{ /*	explains what happened on the safecm pass */
	char	*endd;

	if ((!(calv & 02)) || (nptsp < 1)) return;
	endd =  lineend(fp, calv);
	fprintf(fp,
	 " %4d points already present, %4d added on safecm pass%s",
		nalr, nptsp, endd);
}

int calinv(FILE *fp, VSPLINE *lspl, int calverbose)
{ /* This inverts the calibration mapping */
	static int	ix, iy, nxy, nptsc, nptsp, nalr;
	static double	prevresid, totresid;
	static float	cx, cy, px, py, *wptx, *wpty;
	static CALPOINT	*calp;
	static char	wptmsg[] = "wptx in calinv";

	lspl = lspl;	nxy = lspl->v_nrow * lspl->v_ncol;
	for (nptsp = iy = 0, calp = lspl->v_cpt; iy < lspl->v_nrow; iy++)
	   for (ix = 0; ix < lspl->v_ncol; ix++, calp++)
		if (calp->c_set) nptsp++;
	/* first invert where we're confident */
	py = lspl->v_y0px;	calp = lspl->v_cpt;
	for (nptsc = iy = 0; iy < lspl->v_nrow; iy++, py += lspl->v_yrpx)
	   {
		px = lspl->v_x0px;
		for (ix = 0; ix < lspl->v_ncol;
		 ix++, px += lspl->v_xrpx, calp++)
		   {
			if (rawpix2cm(lspl, px, py, &cx, &cy))
			  {
				calp->c_ctopx = cx;	calp->c_ctopy = cy;
				calp->c_cset = 1;	nptsc++;
			  }
		   }
	   }
	tell_invbegin(fp, nptsp, nptsc, nxy, calverbose);
	tell_invbegin(stdout, nptsp, nptsc, nxy, calverbose);
	/* extrapolate within arrays to complete these */
	for (iy = lspl->v_midr, nptsc = nptsp = 0; iy >= 0; iy--)
	   {
		for (ix = lspl->v_midc ; ix >= 0; ix--)
		   {
			nptsc += strapc(lspl, iy, ix);
			nptsp += strapp(lspl, iy, ix);
		   }
		for (ix = lspl->v_midc + 1; ix < lspl->v_ncol; ix++)
		   {
			nptsc += strapc(lspl, iy, ix);
			nptsp += strapp(lspl, iy, ix);
		   }
	   }
	for (iy = lspl->v_midr + 1; iy < lspl->v_nrow; iy++)
	   {
		for (ix = lspl->v_midc; ix >= 0; ix--)
		   {
			nptsc += strapc(lspl, iy, ix);
			nptsp += strapp(lspl, iy, ix);
		   }
		for (ix = lspl->v_midc + 1; ix < lspl->v_ncol; ix++)
		   {
			nptsc += strapc(lspl, iy, ix);
			nptsp += strapp(lspl, iy, ix);
		   }
	   }
	tell_extrap(fp, nptsc, nxy, nptsp, calverbose);
	tell_extrap(stdout, nptsc, nxy, nptsp, calverbose);
	if (NULL == (wptx = (float *)callop((unsigned)(2 *nxy),
	 sizeof (float), wptmsg))) return EALLOC;
	/* see if we can add in all the rest */
	cy = lspl->v_y0cm;	calp = lspl->v_cpt;
	for (nalr = nptsp = iy = 0; iy < lspl->v_nrow;
	 iy++, py += lspl->v_yrcm)
	   {
		cx = lspl->v_x0cm;
		for (ix = 0; ix < lspl->v_ncol;
		 ix++, cx += lspl->v_xrcm, calp++)
		   {
			if (calp->c_set) nalr++;
			else if (safecm2pix(lspl, cx, cy, &px, &py))
			  {
				calp->c_ptocx = px;	calp->c_ptocy = py;
				calp->c_set = 1;	nptsp++;
			  }
		   }
	   }
	tell_safe(fp, nalr, nptsp, calverbose);
	tell_safe(stdout, nalr, nptsp, calverbose);
	/* having found everything, we now compare our interpolated values
	 with the observed values */
	wpty = wptx + nxy;	cy = lspl->v_y0cm;
	for (nptsp = iy = 0; iy < lspl->v_nrow; iy++, cy += lspl->v_yrcm)
	   {
		cx = lspl->v_x0cm;
		for (ix = 0; ix < lspl->v_ncol;
		 ix++, cx += lspl->v_xrcm, wptx++, wpty++)
		   {
			if (safecm2pix(lspl, cx, cy, &px, &py))
			  {
				*wptx = px;	*wpty = py;	nptsp++;
			  }
		   }
	   }
	wptx -= nxy;	wpty -= nxy;
	/* use these interpolated values as the ultimate outputs */
	for (iy = 0, calp = lspl->v_cpt; iy < lspl->v_nrow; iy++)
	   for (ix = 0; ix < lspl->v_ncol; ix++, calp++, wptx++, wpty++)
	      if ((fabs(*wptx) > FUZZ) || (fabs(*wpty) > FUZZ))
	        {
			calp->c_ptocx = *wptx;	calp->c_ptocy = *wpty;
			calp->c_set = 1;
	        }
	wptx -= nxy;	wpty -= nxy;
	/* now do the comparison itself away from the actual fiducial points */
	prevresid = 9999.;
	for (iy = 0; iy < 20; iy++)
	   {
		if ((totresid = compcal(lspl, fp, 0)) > prevresid) break;
		prevresid = totresid;	if (adjust_xy(lspl) < 4) break;
	   }
	(void)compcal(lspl, fp, calverbose | 04);	xeflush(fp);
	memzap((AVOID **)&wptx, wptmsg);	return EOK;
}

/* end of calinv.c */
