/* ftable.c -- Routines for dealing with the frame-number index for XGEN

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: reorganize use of "{"
	 8 Jun 2006: move spindle to ../prerefine/peaks.c so it can be static
	29 May 2006: ftkill & effecfram eliminated since they aren't in use
	 3 Jan 2006: major reorg: manageftw actually works
	18 Dec 2005: static global replaced with calls to manageftw
	30 Nov 2005: ftables() replaces get_fmlimits, make_ftab, kill_ftab
	24 Nov 2005: inline argument declarations and strict prototyping
	17 Nov 2005: replace init_sgfi & kill_ftab with ftabmem
	22 Jun 2005: fmnocmp(), sorder now declared as static
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	10 Jan 2005: get_imgroups(): correct spelling mistakes in diagnostics;
	 replace maybe_image() with imagename() and imagelike() for efficiency
	18 Dec 2004: in get_imgroups, always save MINNAMESIZE for ig_template
	20 Oct 2004: make_ftab looks to see if a full sgfi already exists;
	 if it does exist, it doesn't re-read FRAMES
	19 Oct 2004: produce a set of FRAMEENTRY structures in use_imgroup()
	13 Oct 2004: set sgfi to zero in kill_ftab
	28 Aug 2004: routines used only here are declared as static; cosmetics
	11 Jul 2004: correct management of ildq and ilq in get_imgroups()
	14 Jun 2004: install_names(),etc. moved to xgenfile.c; cosmetic changes
	26 May 2004: init_sgfi() convenience routine created & used;
	 translate_envname() & install_names() inserted for handling templates
	18 Feb 2004: kill_ftab takes no args, so it only manipulates the
		static global FRAMEINDEX structure, not the caller's version
	 7 Feb 2004: in second fgets call in make_ftab, read 182 chars,
		not 80; this allows very long filenames to be read ok
	 2 Feb 2004: make_ftab now accesses sgfi instead of a local structure
	31 Jan 2004: replace lofl with *sgfi
		so get_fmlimits behaves more like make_ftab
	30 Jan 2004: rename lft0 to cft0 in sorder;
	 rename lft0 to lft in indexname; modify kill_ftab so that everything
	 that needs to be nulled is really nulled; lft0 becomes a static global
	24 Jan 2004: static globals idname, workname, dataname eliminated;
		liter,tempindx,mdirc,dirci,lfmi made local to one routine each
	21 Jan 2004: license message inserted
	12 Jan 2004: get_fmlimits and make_ftab return pointer to the
		FRAMEINDEX structure that they find.
	11 Jan 2004: kill_ftab created: destroys internal FRAMEINDEX struct
	10 Jan 2004: many external globals replaced with static globals,
		function arguments, or elements of structures pointed to
		by function arguments; some routines reordered to reduce
		need to declare them. FRAMEINDEX structure type set up.
	13 Dec 2003: eliminate unused variable `suffix' from get_nominal()
			and `fulen' from populate_imgroup()	
	23 Sep 2003: remove declaration of strcpy()--sometimes it's a macro
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Sep 2003: replace access() call with xgaccess() call
		with FRAMES file so we can gunzip it on the fly if needed
	20 Aug 2003: insert hundreds of lines of frame-index stuff
		from setupio.c into here
	 3 Mar 2003: replace dirminmax with get_imgroups and use_imgroup
	 5 Jan 2003: reorganized source-file header
	 1 Jan 2003: in get_nominal, eliminate suffixes .z, .gz, .Z
		in filenames as we create the index;
		general filename template without compression suffixes;
		call newidname within dirminmax only after closing up the
		count of valid files.
	25 Apr 2002: pick a filename prefix in get_fmlimits if needed
	25 Mar 2002: removed MSI attachments
	31 Aug 1995: free memory for `ft' in more_logicals
		if FRAMES is reassigned
	31 May 1995: "register" -> "register int" throughout
	20 May 1995: correct return codes from routines
	12 May 1995: make_ftab takes no argument; totalframes is a global
	 2 Mar 1995: use FRAMEENTRY ft, not ft0; shared-memory alloc and free
	 7 Jul 1994: calloc(nfms+2), not nfms, for safety
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	 2 Mar 1995: add frame-number arg to cm2pix and pix2cm;
		add getfoff, adjust_to_fm
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 5 Oct 1994: bugfix: lsuffix[6] is ".frm", not lsuffix[4].
	 5 May 1994: source-file headers, POLYDET->XGEN
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version

   Purpose: creating and using the index to the data frames in X-GEN.
   Cross references: xgenio.c, setupio.c

*************************************************************************/

#include	"xgen.h"
#include	<dirent.h>
#ifdef VMS
#define		FILEFORMAT1	"%s%s.%s"	/* VMS format */
#define		DIVIDER		':'
#define		DELIM		']'
#else
#define		FILEFORMAT1	"%s/%s.%s"	/* Unix format */
#define		DIVIDER		'/'
#define		DELIM		'/'
#endif
#define		MAXTEMPLATES	128
#define		WORKTEMPLATE	(MAXTEMPLATES-1)
#define		MINNAMESIZE	256

 typedef struct {
		short	ig_nfms, ig_fmmin, ig_fmmax, ig_templen;
		short	ig_nameconv, ig_valoff, ig_suffoff, ig_npounds;
		short	ig_nmiss, ig_nfound, ig_ngz, ig_nbigz, ig_nz, ig_nall;
		char	*ig_template;
		} IMGROUP;

 typedef struct {
		char		*fx_templ;
		int		fx_base, fx_interval, fx_npounds;
		unsigned	fx_nentries;
		FRAMEENTRY	*fx_ft0;
		IMGROUP		*fx_ig0;
		} FTWORK;

static int fmnocmp(FRAMEENTRY *fe0, FRAMEENTRY *fe1) /* for sorting frames */
{
	return	fe0->fmno - fe1->fmno;
}

static FTWORK *manageftw(int opno)
{ /*	This manages a static global pointer ftw according to opno.
	Note that eltno == opno >> 16.
	opno == 0:	delete memory for ftw and its contents; return NULL
	For opno even:
	 opno & 02:	delete memory for fx itself
	 opno & 04:	delete memory for ftw->fx_ig0
	 opno & 010:	delete memory for (fx->ig0 + WORKTEMPLATE)->ig_template
	 opno & 020:	delete memory for (fx->ig0 + i)->ig_template,
			 i = 0, ... (eltno - 1).
	 opno & 040:	delete memory for ftw->fx_ft0
	 opno & 0100:	delete memory for ftw->fx_templ
	 We do these in what I hope is a logical order.
	For opno odd and positive:
	opno == 01:	unconditionally return ftw in its current state
	opno & 02:	create ftw itself first, if needed, and return it
	opno & 04:	create ftw and ftw->fx_ig0 if needed
	opno & 010:	create ftw and ftw->fx_ig0 if needed. Then reserve
			memory for (ftw->fx_ig0 + WORKTEMPLATE)->ig_template.
	opno & 020:	create ftw and ftw->fx_ig0, if needed.
			Reserve memory for (ftw->fx_ig0 + eltno)->ig_template.
	opno & 040:	create ftw and ftw->fx_ft0, if needed. Reserve memory
			 for eltno FRAMEENTRY elements in ftw->fx_ft0.
	opno & 0100:	create ftw if needed. Reserve memory for
			 eltno bytes of ftw->fx_templ */
			 

	int		eltno, i, nlook;
	static FTWORK	*ftw;
	static IMGROUP	*limp, *lwimp;
	static size_t	oldnframes, oldtemplen;
	static char	ftwmsg[] = "ftw in manageftw";
	static char	ftigmsg[] = "ftw->fx_ig0 in manageftw";
	static char	itmsg[] = "limp->ig_template in manageftw";
	static char	itwmsg[] = "lwimp->ig_template in manageftw";
	static char	ftemsg[] = "ftw->fx_templ in manageftw";
	static char	ft0msg[] = "ftw->fx_ft0 in manageftw";

	if (opno == 0) /* eliminate memory for ftw and its children */
	  {
		if (NULL != ftw) ftw = manageftw(056);
		memzap((AVOID **)&ftw, ftwmsg);	return NULL;
	  }
	eltno = opno >> 16;	/* high-order 16-bits give us a counter */
	if (!(opno & 01)) /* even and positive */
	  {
		if ((opno & 0100) && (NULL != ftw) && (NULL != ftw->fx_templ))
			memzap((AVOID **)&(ftw->fx_templ), ftemsg);
		if ((opno & 040) && (NULL != ftw) && (NULL != ftw->fx_ft0))
			memzap((AVOID **)&(ftw->fx_ft0), ft0msg);
		if ((opno & 020) && (NULL != ftw) && (NULL != ftw->fx_ig0))
		  {
			nlook = (ftw->fx_nentries > 0) ?
				ftw->fx_nentries : MAXTEMPLATES;
			for (limp = ftw->fx_ig0, i = 0; i < nlook; i++, limp++)
				 memzap((AVOID **)&(limp->ig_template), itmsg);
		  }
		if ((opno & 010) && (NULL != ftw) && (NULL != ftw->fx_ig0))
		  {
			lwimp = ftw->fx_ig0 + WORKTEMPLATE;
			if (NULL != lwimp->ig_template)
			   memzap((AVOID **)&(lwimp->ig_template), itwmsg);
		  }
		if ((opno & 04) && (NULL!= ftw) && (NULL != ftw->fx_ig0))
			memzap((AVOID **)&(ftw->fx_ig0), ftigmsg);
		if (opno & 02) memzap((AVOID **)&ftw, ftwmsg);
		return ftw;
	  }
	/* all remaining choices have opno odd */
	if (opno == 01) return ftw;
	if (opno & 02) /* create memory for ftw itself */
	  {
		if ((NULL == ftw) && (NULL == (ftw = (FTWORK *)callop(
		 1, sizeof (FTWORK), ftwmsg)))) return NULL;
	  }
	if (opno & 04) /* create memory for ftw->fx_ig0 */
	  {
		if (NULL == (ftw = manageftw(03))) return ftw;
		if ((NULL == ftw->fx_ig0) && (NULL == (ftw->fx_ig0 =
		 (IMGROUP *)callop(MAXTEMPLATES, sizeof (IMGROUP), ftigmsg))))
			return manageftw(02);
	  }
	if (opno & 010)	/* memory for (ftw->fx_ig0+WORKTEMPLATE)->ig_template*/
	  {
		if (NULL == (ftw = manageftw(07))) return ftw;
		lwimp = ftw->fx_ig0 + WORKTEMPLATE;
		if ((NULL == lwimp->ig_template) && (NULL ==
		 (lwimp->ig_template = mallop(MINNAMESIZE, itwmsg))))
			return manageftw(06);
	  }
	if (opno & 020)	/* memory for (ftw->fx_ig0 + eltno)->ig_template */
	  {
		if (NULL == (ftw = manageftw(07))) return ftw;
		limp = ftw->fx_ig0 + eltno;
		if ((NULL == limp->ig_template) && (NULL ==
		 (limp->ig_template = mallop(MINNAMESIZE, itmsg))))
			return manageftw(06);
		(ftw->fx_nentries)++;
	  }
	if (opno & 040)	/* create memory for ftw->fx_ft0 */
	  {
		if (eltno < 1) eltno = 999;
		if (NULL == (ftw = manageftw(03))) return ftw;
		if ((NULL == ftw->fx_ft0) || (eltno > oldnframes))
		  {
			if (NULL != ftw->fx_ft0) ftw = manageftw(040);
			if (NULL == (ftw->fx_ft0 = (FRAMEENTRY *)callop(
			 (size_t)eltno, sizeof (FRAMEENTRY), ft0msg)))
				return manageftw(02);	oldnframes = eltno;
		  }
	  }
	if (opno & 0100)	/* create memory for ftw->fx_templ */
	  {
		if (NULL == (ftw = manageftw(03))) return ftw;
		if (eltno < 16) eltno = 16;
		if ((NULL == ftw->fx_templ) || (eltno > oldtemplen))
		  {
			if (NULL != ftw->fx_templ) ftw = manageftw(0100);
			if (NULL == (ftw->fx_templ = mallop((size_t)eltno,
			 ftemsg))) return manageftw(02);
			oldtemplen = eltno;
		  }
	  }
	return ftw;
}

int get_nominal(char *str, char **eprefix, char **otemplate)
{ /*	This returns the integer that is represented decimally by the
	string containing the numerical value found at the
	end of the string `str'. If str contains both alphabetic and numeric
	characters, the SUBSTRING SCANNED IS ASSUMED to begin where the last
	batch of digits begins. Thus in the string arf03_b12.0372,
	it returns 372; in the string arf03_b12_145.img, it returns 145.
	Unfortunately in the string arf03_b12145.img,
	it wants to return 12145, which is probably wrong;
	so we force it to return 2145 in that case. The routine returns
	-1 in case of errors.
	The pointer eprefix points to the last character in the prefix
	of the filename. `otemplate' points at a template containing
	the filenames we will associate with the string `str'.
	It will typically look something like
	 /home/user1id/xy34zab/xy34zab_2b_####.img */
	char		*cp, *icp, *cps, *cpe;
	int		ival, mult, leno, comptype;
	FRAMEINDEX	*lfmi;
	FTWORK		*ftw;

	*eprefix = NULL;	/* default */
	if ((NULL == str) || ('\0' == *str)) return -1; /* string must be ok*/
	cps = cpe = NULL;	/* end- and start- pointers NULLed */
	 /* count backward through the string, looking for the
	 rightmost digit and the rightmost non-digit to the left of digits*/
	/* exclude standard HKL and X-GEN files */
	if (xgstrend(str, ".x", 0)) return -1;
	if (xgsuffix(str)) return -1;
	for (cp = str + (strlen(str) - 1); cp >= str; cp--)
	   {
		if (('0' <= *cp) && (*cp <= '9'))
		  {
			if (NULL == cpe) /* found the rightmost digit */
				cpe = cp; /* remember rightmost digit */
		  }
		else if (NULL != cpe) 
		  {
			cps = cp + 1;	break;
		  }
	   }
	if (NULL == cpe) return -1; /* didn't find any digits */
	if (NULL == cps) cp = str; /* nothing but digits */
	/* maximum string length should be 4 */
	if ((cpe - cps) > 3) cps = cpe - 3;
	*eprefix = cps - 1;	/* remember where the prefix ended */
	/* now construct the integer value explicitly */
	for (cp = cpe, mult = 1, ival = 0; cp >= cps; cp--, mult *= 10)
		ival += mult * (((int)(*cp)) - '0');
	/* remember the template and keep it printable */
	while (('0' <= *cpe) && (*cpe <= '9')) cpe++;
	lfmi = ftables(1);
	if (NULL == (ftw = manageftw(03))) return -1;
	if (NULL != *otemplate) /* create # signs in template */
	  {
		for (cp = *otemplate, icp = str, ftw->fx_npounds = 0;
		 '\0' != *icp; icp++, cp++)
		   {
			if ((cps <= icp) && (icp < cpe))
			  {
				*cp = '#';	(ftw->fx_npounds)++;
			  }
			else *cp = *icp;
		   }
		*cp = '\0';
		/* don't keep .Z, .gz, or .z at the end
		 of the template, because we don't want them in the names */
		if ((comptype = compressiontype(*otemplate, &leno)))
		  {
			cp = *otemplate + leno - 2 - (comptype == 1);
			*cp = '\0';
		  }
		if (ftw->fx_npounds < 4) lfmi->fi_flags |= 01;
	  }
	return ival;
}

static int populate_imgroup(char *fnam, IMGROUP *loci)
{ /*	This populates the `IMGROUP' structure loci with information
	derived from the file fnam in directory dirn */
	int	locnum;
	char	*otempl, *epref, *cp, *cpo;

	otempl = epref = NULL;
	locnum = get_nominal(fnam, &epref, &otempl);
	if (locnum < 0) return 0;
	loci->ig_nfms = 1;	loci->ig_fmmin = loci->ig_fmmax = locnum;
	strcpy(loci->ig_template, fnam);
	/* eliminate terminal ".z", ".Z", and ".gz" from template */
	(void)xgstrend(loci->ig_template, ".z", 1);
	(void)xgstrend(loci->ig_template, ".Z", 1);
	(void)xgstrend(loci->ig_template, ".gz", 1);
	/* figure out where the numerical value begins in the string */
	loci->ig_valoff = 1 + (epref - fnam);
	for (cp = loci->ig_template + loci->ig_valoff + 1; '\0' != *cp; cp++)
	   if ((*cp < '0') || (*cp > '9')) break;
	/* figure out where the suffix, if any, begins in the string */
	loci->ig_suffoff = cp - loci->ig_template;
	cpo = loci->ig_template + loci->ig_suffoff;
	/* put pound signs into the template in place of the digits */
	loci->ig_npounds = loci->ig_suffoff - loci->ig_valoff;
	for (cp = loci->ig_template + loci->ig_valoff; cp < cpo; cp++) 
		*cp = '#';
	loci->ig_templen = strlen(loci->ig_template);	return 1;
}

static char *itobase10(int num)
{ /*	This converts the integer `num' to a 3-or-4 digit decimal string. */
	char		*cp, *cpe, *cp0;
	int		left;
	FRAMEINDEX	*lfmi;
	static char	value[6];

	cp0 = &(value[0]);	lfmi = ftables(1);
	cpe = cp0 + 3 + !(lfmi->fi_flags & 01);	*cpe = '\0';
	for (cp = cpe-1, left = num; (cp >= cp0) && (left); cp--, left /= 10)
		*cp = '0' + (char)(left % 10);
	for ( ; cp >= cp0; cp--) *cp = '0';
	return cp0;
}

static int nallcmp(IMGROUP *il0, IMGROUP *il1)
{ /*	This sorts on the number of usable images in the group */
	int	iadif;

	iadif = il0->ig_nall - il1->ig_nall;
	return (iadif > 0) ? -1 : (iadif < 0) ? 1 : 0;
}

static IMGROUP *get_imgroups(char *dirnam, int *nimgroups)
{ /*	This goes into a directory and looks for images. For each image it
	finds, it looks to see if its filename template has already been
	encountered in the directory. If it has, we update the list of
	attributes of that template. If not, we create a new one */

	register int		i, templen, fmno;
	int			marg, ngroups, nexamined, match;
	size_t			tocopy;
	char			*tbase, *cpi, *cpo, *tempi;
	FTWORK			*ftw;
	static DIR		*dirloc;
	static struct dirent	*mdirg;
	static char		tempn[256];
	static IMGROUP		*ilp, *ilq;

	if (NULL == (dirloc = opendir(dirnam))) return NULL; /* open dir.*/
	if (NULL == (ftw = manageftw(017))) return NULL;
	*nimgroups = ngroups = 0;
	ilp = ftw->fx_ig0 + WORKTEMPLATE;
	tocopy = sizeof (IMGROUP) - sizeof (char *);
	/* look through the list of files in the directory */
	for (nexamined = 0; NULL != (mdirg = readdir(dirloc)); )
	   {
		nexamined++;
		/* see if the file is an image */
		if ((mdirg->d_ino) && imagename(mdirg->d_name) &&
		 populate_imgroup(mdirg->d_name, ilp))
		  {
			match = 0;
			if (ngroups) /* have we seen this template? */
			 for (i = 0, ilq = ftw->fx_ig0;
			  i < ngroups; i++, ilq++)
			   { /* found a match: update the structure */
				if (!strcmp(ilp->ig_template, ilq->ig_template))
				  {
					if (ilq->ig_fmmin > ilp->ig_fmmin)
					  ilq->ig_fmmin = ilp->ig_fmmin;
					if (ilq->ig_fmmax < ilp->ig_fmmax)
					  ilq->ig_fmmax = ilp->ig_fmmax;
					(ilq->ig_nfms)++;
					match = 1; break;
				  }
			   }
			if (!match) /* no match: make sure this is really
			 an image, and, if so, create a new template */
			  {
			    if (imagelike(dirnam, mdirg->d_name))
			      {
				ilq = ftw->fx_ig0 + ngroups;
				cpymem((AVOID *)ilq, (AVOID *)ilp, tocopy);
				ilq->ig_templen = templen =
					strlen(ilp->ig_template) + 2;
				marg = (ngroups << 16) | 027;
				if (NULL == (ftw = manageftw(marg)))
					return (IMGROUP *)manageftw(0);
				strcpy(ilq->ig_template, ilp->ig_template);
				ngroups++;
			      }	/* end of "if (imagelike(..." clause */
			  } /* end of "if (!match)" clause */
		  } /* end of "if (mdirg ... " clause */
	   } /* end of loop through files in directory */
	(void)closedir(dirloc);
	/* having gone through the images, we now clean up the
	 template lists */
	for (ilq = ftw->fx_ig0, i = 0; i < ngroups; i++, ilq++)
	   {
		zeromem((AVOID *)tempn, 256);
		strcpy(tempn, dirnam);	strcat(tempn, "/");
		strcat(tempn, ilq->ig_template);
		tempi = tempn + ilq->ig_valoff + 1 + strlen(dirnam);
		ilq->ig_nmiss = ilq->ig_nfound =
			ilq->ig_nz = ilq->ig_ngz = ilq->ig_nbigz = 0;
		for (fmno = ilq->ig_fmmin; fmno <= ilq->ig_fmmax; fmno++)
		   {
			tbase = itobase10(fmno);
			cpi = tbase + strlen(tbase) - ilq->ig_npounds;
			for (cpo = tempi; '\0' != *cpi; cpo++, cpi++)
				*cpo = *cpi;
			templen = strlen(ilq->ig_template);
			if (templen > ilq->ig_suffoff)
				cpo += (templen - ilq->ig_suffoff);
			/* eliminate memory from previous strings */
			*cpo++ = '\0'; *cpo++ = '\0'; *cpo++ = '\0'; cpo -= 3;
			/* look for the file with its unadorned name,
			 and with .Z,, .z, and .gz appended. Count them. */
			if (0 == access(tempn, 04)) (ilq->ig_nfound)++;
			else {
				*cpo++ = '.'; *cpo++ = 'Z'; *cpo = '\0';
				if (0 == access(tempn, 04)) (ilq->ig_nbigz)++;
				else {
					*(cpo-1) = 'z';
					if (0 == access(tempn, 04))
						(ilq->ig_nz)++;
					else {
						*(cpo-1) = 'g';
						*cpo++ = 'z'; *cpo = '\0';
						if (0 == access(tempn, 04))
							(ilq->ig_ngz)++;
						else (ilq->ig_nmiss)++;
					     }
				      }
			      }
		   }
		ilq->ig_nall = ilq->ig_nfound + ilq->ig_nz +
			ilq->ig_nbigz + ilq->ig_ngz;
	   }
	/* if we have more than one group, sort on ig_nall */
	if (ngroups > 1) qsort((void *)ftw->fx_ig0,
	 (size_t)ngroups, sizeof (IMGROUP), (QPE)nallcmp);
	*nimgroups = ngroups;	return (ngroups) ? ftw->fx_ig0 : NULL;
}

static char *itostr(int val, int indxflags)
{ /*	This prints a number in any base from 2 through 36.
	It returns blanks if |val| is greater than 2^31. The digits prior
	to the first significant one are filled with zeroes. The string is
	returned in a static area, so it should be copied if needed.
	A special case is the deg.mn format, where an entirely different
	mechanism for defining the string is enacted. */
	 int		deg, i, ic, isup, j, lbase, mn, nslots, nsu, q, v;
	char		alph0, *icp, *ocp;
	static char	dig[13];

	nslots = indxflags >> 24;	lbase = (indxflags >> 16) & 0xff;
	isup = (indxflags >> 9) & 01;
	/* see how many slots we're using */
	nsu = nslots < 12 ? nslots : 11;
	/* blank out the string to begin with */
	for (i = 0; i < nsu; i++) dig[i] = '0';	dig[nsu] = '\0';
	/* make sure the value is in range */
	if ((val < -2147483647) || (val > 2147483647)) return dig;
	/* make sure the base is in range */
	if ((lbase < 2) || ((lbase != 60) && (lbase > 36))) return dig;
	alph0 = isup ? 'A' : 'a';
	if (2 == (indxflags & 0xf))
	  {
		dig[5] = '\0';	for (i = 0; i < 5; i++) dig[i] = ' ';
		/* in this format we must use at least five digits. */
		if (nslots < 5) return dig;
		/* first get the value into a range between 0 and 360*60. */
		if (val < 0)
		  {
			v = -val; if (v > 21600) v %= 21600; val = 21600 - v;
		  }
		else if (val > 21600) val %= 21600;
		deg = val / 60;	mn = val % 60;
		if (!mn) dig[3] = dig[4] = '0';
		else if (mn < 10)
		  {
			dig[3] = '0'; dig[4] = '0' + mn;
		  }
		else {
			dig[3] = '0' + (mn / 10); dig[4] = '0' + (mn % 10);
		     }
		if (!deg) dig[0] = dig[1] = dig[2] = '0';
		else if (deg < 10)
		  {
			dig[0] = dig[1] ='0'; dig[2] = '0' + deg;
		  }
		else if (deg < 100)
		  {
			dig[0] = '0';	dig[1] = '0' + (deg / 10);
			dig[2] = '0' + (deg % 10);
		  }
		else {
			dig[0] = '0' + (deg / 100);
			dig[1] = '0' + ((deg % 100) / 10);
			dig[2] = '0' + (deg % 10);
		     }
		/* we're done unless the user wanted more digits than that, in
		 which case we shift things to the right and pad with zeroes */
		if (nslots > 5)
		  {
			icp = &(dig[5]); ocp = &(dig[nslots]);
			for ( ; icp >= &(dig[0]); icp--, ocp--) *ocp = *icp;
			for ( ; ocp >= &(dig[0]); ocp--) *ocp = '0';
		  }
		dig[nslots+1] = '\0';
	  }
	else {
		v = val < 0 ? -val : val;
		for (i = 0, j = v, ic = nsu -1; j > 0; i++, ic--)
		   {
			q = j % lbase;	j -= q;	if (j > 0) j /= lbase;
			/* if we run out of room, we return an error */
			if (ic < 0)
			  {
				dig[0] = '\0';	return dig;
			  }
			dig[ic] = q < 10 ? '0' + q : alph0 + (q - 10);
		   }
		if (val < 0) /* make sure there's room: */
		  {
			if ('0' != dig[0])
			  {
				dig[0] = '\0'; return dig;
			  }
			dig[0] = '_';
		  }
	     }
	return dig;
}

static char *use_imgroup(char *dirna, IMGROUP *imb, int fromcu)
{ /*	This preserves the information regarding one specific group of images
	available in a directory. It returns the filename prefix, and it
	sticks the minimum and maximum frame numbers where they belong. */
	int		fno, ftwop, i, ientry, lbase, nfms, newiflags, tempindx;
	size_t		intlen;
	char		*cp, *cpi, *cpo, *datdir;
	char		*firstpound, *icp, *intname, *pref, te2[40];
	FRAMEINDEX	*lfmi;
	FRAMEENTRY	*lft;
	XGNAME		*losh;
	FTWORK		*ftw;
	static size_t	preflen;
	static char	datdmsg[] = "datdir in use_imgroup";
	static char	prefmsg[] = "pref in use_imgroup";
	static char	intmsg[] = "intname in use_imgroup";

	lfmi = ftables(1);
	lfmi->fi_firstfm = imb->ig_fmmin;
	lfmi->fi_lastfm = imb->ig_fmmax;
	lfmi->fi_totfms = 1 + lfmi->fi_lastfm - lfmi->fi_firstfm;
	if (NULL == (ftw = manageftw(01))) return NULL;
	ftw->fx_templ = imb->ig_template;	/* hurrumph! */
	ftw->fx_npounds = imb->ig_npounds;
	preflen = imb->ig_valoff + ftw->fx_npounds + 8;
	if (preflen < 40) preflen = 40;
	if (NULL == (pref = (char *)mallop(preflen, prefmsg))) return NULL;
	for (cp = pref, icp = imb->ig_template, i = 0;
	 '\0' != *icp && i < imb->ig_valoff; i++, icp++, cp++) *cp = *icp;
	*cp = '\0';	if (*(cp - 1) == '.') *(cp - 1) = '\0';
	if (cp - pref > preflen)
	  {
		fprintf(stderr, "Warning: cp - pref = %d > preflen = %d\n",
			(int)(cp - pref), (int)preflen);
	  }
	/* we want to force the program to install the new name in,
	 so we blank out losh->xmm_idna */
	if (NULL == (losh = sharename(0x2f))) return NULL;
	newiflags = fromcu;
	if (!strcmp(dirna, ".")) newiflags |= 02;
	else if (!strcmp(dirna, "..")) newiflags |= 04;
	(void)newidname(pref, newiflags);
	printf("Directory %s, identifier %s, image template %s:\n",
		dirna, losh->xmm_idna, imb->ig_template);
	printf(" Image min  max #intact #.gz  #.Z  #.z Total #missing\n");
	printf(" %9d %4d %7d %4d %4d %4d %5d %8d\n",
	 imb->ig_fmmin, imb->ig_fmmax, imb->ig_nfound,
	 imb->ig_ngz, imb->ig_nbigz, imb->ig_nz, imb->ig_nall, imb->ig_nmiss);
	/* October 2004, X-GEN V.5.4.3: populate the lfmi->fi_fmi structure */
	if (NULL == (datdir = squeezeall(dirna))) /* get data directory name */
	  {
		if (dirna != datdir) memzap((AVOID **)&datdir, datdmsg);
		return pref;	/* return if this fails */
	  }
	nfms = (imb->ig_fmmax - imb->ig_fmmin) + 1; /* # of images */
	ftwop = ((nfms + 2) << 16) | 043;
	if (NULL == (ftw = manageftw(ftwop))) return (char *)manageftw(0);
	intlen = strlen(datdir) + strlen(imb->ig_template) + 8;
	if (NULL == (intname = mallop(intlen, intmsg)))
	  {
		if (dirna != datdir) memzap((AVOID **)&datdir, datdmsg);
		ftw = manageftw(030);	return pref;
	  }
	zeromem(intname, intlen);
	strcpy(intname, datdir);
	strcat(intname, "/");	strcat(intname, imb->ig_template);
	if (dirna != datdir) memzap((AVOID **)&datdir, datdmsg);
	for (firstpound = NULL, cp = intname; '\0' != *cp; cp++)
	   if ('#' == *cp)
	     {
		firstpound = cp;	break;
	     }
	if (NULL == firstpound) /* sanity check: we need a poundsign */
	  {
		memzap((AVOID **)&intname, intmsg);
		ftw = manageftw(030);	return pref;
	  }
	/* this isn't as general as it could be... */
	lbase = 10;	tempindx = (ftw->fx_npounds << 24) | (lbase << 16);
	if (ftw->fx_npounds < 4) tempindx |= 01;
	for (lft = ftw->fx_ft0, ientry = 0, fno = imb->ig_fmmin;
	 ientry < nfms; ientry++, fno++, lft++)
	   {
		lft->fmno = fno;	strcpy(te2, itostr(fno, tempindx));
		for (cpi = te2, cpo = firstpound; '\0' != *cpi; cpi++, cpo++)
			*cpo = *cpi;
		strcpy(lft->fmname, intname);
	   }
	lfmi->fi_fmi = ftw->fx_ft0;	lfmi->fi_fme = lft - 1; 
	memzap((AVOID **)&intname, intmsg);	return pref;
}

static char *impose_defaults(char *dirstr, char *newprefix, int fromcu)
{ /*	This imposes default directory values onto the system.
	For now, we set all three directory names to `dirstr'
	and set the project identifier to `newprefix' */
	size_t	preflen;
	char	*cp, *ocp, *newpref;
	XGNAME	*losh;

	if ((NULL == newprefix) || ('\0' == *newprefix)) return NULL;
	if ((NULL == dirstr) || ('\0' == *dirstr)) return NULL;
	for (preflen = 0, cp = newprefix; '\0' != *cp; cp++) preflen++;
	if (NULL == (newpref = (char *)mallop(preflen + 2,
		"newpref in impose_defaults"))) return NULL;
	for (cp = newprefix, ocp = newpref; '\0' != *cp; ) *ocp++ = *cp++;
	*ocp = '\0';
	if (NULL == (losh = sharename(0x2f))) return NULL;
	(void)newidname(newpref, fromcu | 02);	return newpref;
}

FRAMEINDEX *ftables(int opno)
{ /*	This manages the FRAMEINDEX, according to the option number "opno".
	If opno==0, it deletes memory for the static global frameindex lfmi
	and returns NULL.
	If opno==1, it sets up memory for the frameindex lfmi and returns it,
	having set nominal starting values for firstfm and lastfm.
	If opno==3, it establishes the table of frames and their index numbers
	by reading the file with name or logical name `frames'
	and building up an array of `FRAMEENTRY' structures.
	For opno > 3, the routine returns a pointer to the array of structures
	and sticks the number of entries in the array in the structure
	element lfmi->fi_totfms. The routine returns NULL upon errors.
	This tries to open FRAMES to establish starting & ending frm values.
	If it succeeds, it also defines the filename template based on what
	it finds. If it fails, it sets nominal starting values.
	The routine returns a pointer to a static FRAMEINDEX structure
	containing the relevant information */
	register int		i, inorder;
	int			fromcu, ftwop, lofirst, lolast;
	int			ngr, prevfm, tmpr;
	unsigned		nfms;
	char			*epref, *icp, *ocp, *ubu, *value;
	char			lobu[184];
	FILE			*findx;
	FRAMEENTRY		*lft;
	IMGROUP			*tgr, *tgq;
	XGNAME			*losh;
	FTWORK			*ftw;
	static char		*prefix;
	static FILE		*fped;
	static FRAMEINDEX	*lfmi;
	static char		lfmimsg[] = "lfmi in ftables";

	if (opno > 3)
	  {
		tmpr = 3;	if (opno & 05) tmpr = 017;
		if (NULL == (ftw = manageftw(tmpr))) ftw = manageftw(036);
	  }
	if (0 == opno)	/* destroy memory for lfmi and children*/
	  {
		if (NULL == lfmi) return NULL;
		ftw = manageftw(076); /* get rid of ftw->fx_ft0 */
		memzap((AVOID **)&lfmi, lfmimsg);	return NULL;
	  }
	if (1 == opno) /* check or create the pointer */
	  { /* if it's already non-null and looks plausible, return it */
		if ((NULL != lfmi) && (-1000 < lfmi->fi_firstfm) &&
		 (lfmi->fi_firstfm < 10000)) return lfmi;
		/* otherwise, get rid of the old one: it's probably bad*/
		(void)ftables(0);	/* ... and create a new one! */
		lfmi = (FRAMEINDEX *)callop(1, sizeof (FRAMEINDEX), lfmimsg);
		lfmi->fi_firstfm = 1;	lfmi->fi_lastfm = 999; return lfmi;
	  }
	else if (3 == opno) /* return pointer to working FRAMEINDEX structure*/
	  { /* before trying to do anything with the FRAMES file,
		 see if the list already exists and is plausible */
		if (NULL != lfmi) /* does a good index exist? */ 
		  { /* if lastfm is < 1 or fi_fmi stuff is bad, start over*/
			if ((lfmi->fi_lastfm < 1) ||
			 (NULL == lfmi->fi_fmi)) ftables(0);
			else return lfmi; /* return the current pointer */
		  }
		/* open the existing FRAMES file */
		if (NULL == (findx = xgopen("frames", "r"))) return NULL;
		/* now count the number of frames in the file */
		for (nfms = 0; NULL != fgets(lobu, 182, findx); ) nfms++;
		if (verror(1 + fseek(findx, 0L, 0), "rewind index file"))
			return NULL;
		/* reserve memory for the frame table */
		/* if we get this far, we need the ftw stuff */
		ftwop = ((nfms + 2) << 16) | 043;
		if (NULL == (ftw = manageftw(ftwop))) return ftables(0);
		/* construct the frame table from the diskfile */
		prevfm = -8000;	/* nonsense startvalue for `previous' frame #*/
		for (i = nfms, lft = ftw->fx_ft0, inorder = 1; i--; lft++)
		   {
			if (NULL == (AVOID *)fgets(lobu, 182, findx))
			  {
				nfms -= (i + 1);	break;
			  }
			(void)sscanf(lobu, "%hd%s", &(lft->fmno), lft->fmname);
			if (lft->fmno < prevfm) inorder = 0;
			prevfm = lft->fmno;
		   }
		(void)fclose(findx);
		/* sort frames if necessary by increasing order of frame # */
		if ((!inorder) && (nfms > 1)) qsort((AVOID *)(ftw->fx_ft0),
		 nfms, sizeof (FRAMEENTRY), (QPE)fmnocmp);
		/* populate the FRAMEINDEX structure we will now return.
		 Use the static global one already identified if appropriate */
		if ((NULL == lfmi) && (NULL == (lfmi = ftables(1))))
			ftw = manageftw(076); /* unlikely */
		lfmi->fi_totfms = nfms;
		lfmi->fi_firstfm = (ftw->fx_ft0)->fmno;
		lfmi->fi_fmi = ftw->fx_ft0;
		lfmi->fi_fme = ftw->fx_ft0 + (nfms - 1);
		lfmi->fi_lastfm = (lfmi->fi_fme)->fmno;
		(void)manageftw(034); /* eliminate unneeded stuff */
		return lfmi;
	  }
	/* other options are like the old "get_fmlimits" */
	if (NULL == lfmi) lfmi = ftables(1);
	/* if the structure is already believably populated, use it and quit */
	if ((NULL != lfmi) && (lfmi->fi_lastfm != 999)) return lfmi;
	fromcu = opno >> 4;	/* flag for whether diagnostics go to Curses */
	lfmi->fi_firstfm = lofirst = 1;
	lfmi->fi_lastfm = lolast = 999;
	lfmi->fi_totfms = 1 + lfmi->fi_lastfm - lfmi->fi_firstfm;
	prefix = NULL;
	if (NULL == (losh = sharename(7))) return lfmi;
	/* if the FRAMES file doesn't already exist, try to guess the
	 starting and ending frames from the filenames present in
	 the current DATA directory */
	if ((NULL == (value = trname("FRAMES"))) || ('\0' == *value) ||
	 (0 != xgaccess(value, 04, 1)))
	  {
		icp = getenv("DATA");
		/* if we already think we know where the data are, we look */
		if ((NULL != icp) && ('\0' != *icp))
		  { /* if we already know the sample ID, we look for it */
			if ((NULL != losh->xmm_idna) &&
			 ('\0' != *(losh->xmm_idna)))
			  {
			   if (NULL != (tgr = get_imgroups(icp, &ngr)))
			     { /* we found some images in the directory.
				See if their names match our sample ID */
				for (tgq = tgr, i = 0; i < ngr; i++, tgq++)
				   {
					if (!strncmp(tgq->ig_template, /*yes!*/
					 losh->xmm_idna,
					 strlen(losh->xmm_idna)))
					  {
						prefix =
						 use_imgroup(icp, tgq, fromcu);
						break;
					  }
				   }
				if (NULL == prefix) prefix = impose_defaults(
					icp, losh->xmm_idna, fromcu);
			     } /* end of clause for call to get_imgroups */
			    else prefix = impose_defaults(
				icp, losh->xmm_idna, fromcu);
			  }
			/* on the other hand, if we didn't come in
			 with a known ID, we just use the most plentiful */
			else if (NULL != (tgr = get_imgroups(icp, &ngr)))
				prefix = use_imgroup(icp, tgr, fromcu);
			else prefix = impose_defaults(icp, "unknown", fromcu);
		  } /* end of clause for known DATA variable */
		else if ((NULL != losh->xmm_idna) &&
		 ('\0' != *(losh->xmm_idna)))
		  { /* next case: id name defined but directory isn't.
			We look for the id name in "." and ".." */
			if (NULL != (tgr = get_imgroups(".", &ngr)))
			  { /* we found some images in the directory.
				See if their names match our sample ID */
				for (tgq = tgr, i = 0; i < ngr; i++, tgq++)
				   {
					if (!strncmp(tgq->ig_template, /*yes!*/
					 losh->xmm_idna,
					 strlen(losh->xmm_idna)))
					  {
						prefix =
						 use_imgroup(".",tgq,fromcu);
						break;
					  }
				   }
			  } /* end of if (...get_imgroups(".",...) clause */
			else if (NULL != (tgr = get_imgroups("..", &ngr)))
			  { /* we found some images in the directory.
				See if their names match our sample ID */
				for (tgq = tgr, i = 0; i < ngr; i++, tgq++)
				   {
					if (!strncmp(tgq->ig_template, /*yes!*/
					 losh->xmm_idna,
					 strlen(losh->xmm_idna)))
					  {
						prefix =
						 use_imgroup("..",tgq,fromcu);
						break;
					  }
				   }
			  } /* end of if (...get_imgroups("..",...) clause */
			/* if neither . nor .. worked, impose a default */
			if (NULL == prefix) prefix = impose_defaults(
				".", losh->xmm_idna, fromcu);
		  }
		/* If we haven't a clue (as is often true) where to look,
		 we look in . and .. to try to intuit a name */
		else {
			if (NULL != (tgr = get_imgroups(".", &ngr)))
				prefix = use_imgroup(".", tgr, fromcu);
			else if (NULL != (tgr = get_imgroups("..", &ngr)))
				prefix = use_imgroup("..", tgr, fromcu);
			if (NULL == prefix)
			  prefix = impose_defaults(".", "unknown", fromcu);
		     }
		if ('\0' == *(losh->xmm_idna))
		  {
			ocp = filepart(prefix, 2); strcpy(losh->xmm_idna, ocp);
		  }
		memzap((AVOID **)&prefix, "prefix in ftables");
		return lfmi;
	  }
#ifndef VMS
	if (*value == '$')	/* spot and fix formats like "$SCRT" */
	  {
		ocp = &(lobu[0]);	icp = value + 1;
		for (tmpr = 0; ('\0' != *icp) && ('/' != *icp) &&
		 (tmpr < 19); icp++, tmpr++, ocp++) *ocp = *icp;
		*ocp = '\0';
		if (NULL != (value = getenv(lobu)))
		  {
			strcpy(lobu, value);
			if ('\0' != *icp) strcat(lobu, icp);
		  }
		value = lobu;
	  }
#endif
	ubu = losh->xmm_misc;
	if ((NULL == (fped = fopen(value, "r"))) ||
	  (NULL == fgets(ubu, 80, fped))) return lfmi;
	(void)sscanf(ubu, "%d", &lofirst);
	while (NULL != fgets(ubu, 80, fped)) ;
	(void)fclose(fped);	/* we're presumably finished reading */
	lobu[0] = '\0';	/* clear the filename-string buffer */
	(void)sscanf(ubu, "%d%s", &lolast, lobu);
	if (lofirst > lolast)
	  {
		tmpr = lofirst; lofirst = lolast;	lolast = tmpr;
	  }
	if ((lofirst < 0) ||(lofirst > 9999)) lofirst = 1;
	if ((lolast < 0) || (lolast > 9999)) lolast = 9999;
	lfmi->fi_firstfm = lofirst;
	lfmi->fi_lastfm = lolast;
	lfmi->fi_totfms = 1 + lfmi->fi_lastfm - lfmi->fi_firstfm;
	if (lobu[0] != '\0') /* establish the template */
	  {	/* back up to the first slash */
		icp = lobu + strlen(lobu);
		for (ocp = icp - 1; ocp >= lobu; ocp--)
			if ((*ocp == '/') || (*ocp == ']')) break;
		ocp++;	if ('\0' == *ocp) return lfmi;
		ftwop = (256 << 16) | 0103;
		if (NULL == (ftw = manageftw(ftwop))) return ftables(0);
		tmpr = get_nominal(ocp, &epref, &(ftw->fx_templ));
		if (-1 == tmpr) ftw->fx_templ[0] = '\0';
	  }
	return lfmi;
}

int grab_base(void)	/* returns the number-system base (10 or 36) */
{
	FTWORK	*ftw;

	return (NULL == (ftw = manageftw(3))) ? 0 : ftw->fx_base;
}

int grab_interval(void) /* returns interval between successive image #s */
{
	FTWORK	*ftw;

	return (NULL == (ftw = manageftw(3))) ? 1 : ftw->fx_interval;
}

void transfer_template(char *stri)
{ /*	stuff the given string in the static global `ftw->fx_templ' */
	int		npou, stlen;
	char		*cp, *cpi;
	FRAMEINDEX	*lfmi;
	FTWORK		*ftw;

	if (NULL == (ftw = manageftw(3))) return;
	if (NULL == ftw->fx_templ) ftw->fx_templ = (char *)mallop(256,
		"ftw->fx_templ in transfer_template");
	if (NULL == ftw->fx_templ) return;
	if (strlen(stri) > 254) (void)strncpy(ftw->fx_templ, stri, 254);
	else (void)strcpy(ftw->fx_templ, stri);
	/* now count the number of poundsigns */
	for (npou = 0, cp = ftw->fx_templ; '\0' != *cp; cp++)
		if ('#' == *cp) npou++;
	ftw->fx_npounds = npou;
	lfmi = ftables(1);
	if (ftw->fx_npounds < 4) lfmi->fi_flags |= 01;
	else if (ftw->fx_npounds == 5) /* special case */
	  {
		stlen = strlen(ftw->fx_templ);
		if (stlen > 5)
		  {
			cpi = ftw->fx_templ + stlen - 6;
			if (!strcmp(cpi, "###.##")) lfmi->fi_flags |= 04;
		  }
	  }
}

void transfer_base(char *stri)
{ /*	stuff string into base value */
	int		iva;
	FRAMEINDEX	*lfmi;
	FTWORK		*ftw;

	iva = atoi(stri);
	if (iva < 2) iva = 2; else if (iva > 60) iva = 60;
	lfmi = ftables(1);
	if (NULL != (ftw = manageftw(03)))
	  {
		ftw->fx_base = iva;	
		if (ftw->fx_base == 36) lfmi->fi_flags |= 02;
	  }
}

void transfer_interval(char *stri)
{ /* stuff string into frame-interval value */
	int	iva;
	FTWORK	*ftw;

	iva = atoi(stri);
	if (iva < 1) iva = 1; else if (iva > 999) iva = 999;
	if (NULL != (ftw = manageftw(3)))
		ftw->fx_interval = iva;
}

static char *itobase36(int num)
{ /*	This converts the integer `num' to a 3-or-4 digit base-36 string. */
	char		*cp, *cpe, *cp0;
	int		left, digit;
	FRAMEINDEX	*lfmi;
	static char	value[6];

	cp0 = &(value[0]);	lfmi = ftables(1);
	cpe = cp0 + 3 + !(lfmi->fi_flags & 01);	*cpe = '\0';
	for (cp = cpe-1, left = num; (cp >= cp0) && (left); cp--, left /= 36)
	   {
		digit = left % 36;
		if (digit < 10)	*cp = '0' + (char)digit;
		else		*cp = 'a' + (char)(digit - 10);
	   }
	for ( ; cp >= cp0; cp--) *cp = '0';
	return cp0;
}

void writeindex(FRAMEINDEX *fmix, int fromcu)
{ /* This writes out the frame-number index to the file with environment
	variable or logical name FRAMES. It can be in an ancient DEG.MN
	format, in which a typical line looks like
   36DATA:BPN4NA009.15
	or more modern formats, which look like
    9DATA:HTS4N_1.0009
	In the modern-format category the frame number (to the right of
	the dot) can be either 3 or 4 digits, and it can be either decimal
	digits (0-9) or base-36 digits (0-9, a-z). */
	int		deg, firstm, frm, i, ifmt, lastm, min, minint, minv;
	char		*cp, *cpi, *le, *liname, *outst;
	XGNAME		*losh;
	FTWORK		*ftw;
	static int	tempindx;
	static char	*shortname;
	static FILE	*fped;
	static char	*ffmt[] = {
	 "%5d%s%c%s00%1d.0%1d\n", "%5d%s%c%s00%1d.%2d\n", 
	 "%5d%s%c%s0%2d.0%1d\n",  "%5d%s%c%s0%2d.%2d\n", 
	 "%5d%s%c%s%3d.0%1d\n",   "%5d%s%c%s%3d.%2d\n" } ;
#ifdef VMS
	static char	indsuf[] = "FRM";
	static char	dataword[] = "DATA";
#else
#define		dataword	(losh->xmm_datana)
	static char	indsuf[] = "frm";
#endif

	le = lineend(stdout, fromcu);
	if (NULL == (ftw = manageftw(3))) return;
	if (NULL == (losh = sharename(0x100f)))
	  {
		fprintf(stderr, "Can't reserve memory for name buffers%s", le);
		return;
	  }
	if (3 > (frm = basic_names(fromcu)))
	  {
		fprintf(stderr,
		 "only %d basic names returned; can't writeindex%s", frm, le);
		return;
	  }
	liname = losh->xmm_miscb;
		/* WORK and DATA logicals required */
	strcpy(losh->xmm_workna, getenv("WORK"));
	strcpy(losh->xmm_datana, getenv("DATA"));
	if (1 > (frm = check_directories(1, 1)))
	  {
		fprintf(stderr,
		 "check_directories returned %d; can't writeindex%s", frm, le);
		return; /*be sure WORK exists */
	  }
	sprintf(liname, FILEFORMAT1, losh->xmm_workna, losh->xmm_idna, indsuf);
	if (NULL == (fped = fopen(liname, "w")))
	  {
		fprintf(stderr, "%s can't be opened; can't writeindex%s",
			liname, le);
		return;	/* open file */
	  }
	/* if the old DEG.MN format is in use, we ignore all the other bits
	 in the fmix->fi_flags variable */
	if (fmix->fi_flags & 04)
	  {
		if (fmix->fi_firstfmd < 0)
			fmix->fi_firstfmm = -fmix->fi_firstfmm;
		if (fmix->fi_lastfmd < 0) fmix->fi_lastfmm = -fmix->fi_lastfmm;
		while (fmix->fi_firstfmm > 60)
		     {
			fmix->fi_firstfmm -= 60;	fmix->fi_firstfmd++;
		     }
		while (fmix->fi_lastfmm > 60)
		     {
			fmix->fi_lastfmm -= 60;	fmix->fi_lastfmd++;
		     }
		while (fmix->fi_firstfmm < 0)
		     {
			fmix->fi_firstfmm += 60; (fmix->fi_firstfmd)--;
		     }
		while (fmix->fi_lastfmm < 0)
		     {
			fmix->fi_lastfmm += 60;	fmix->fi_lastfmd--;
		     }
		while (fmix->fi_firstfmd > 360) fmix->fi_firstfmd -= 360;
		while (fmix->fi_lastfmd > 360) fmix->fi_lastfmd -= 360;
		while (fmix->fi_firstfmd < 0) fmix->fi_firstfmd += 360;
		while (fmix->fi_lastfmd < 0) fmix->fi_lastfmd += 360;
			/* figure out the interval between frames */
		lastm = fmix->fi_lastfmd * 60 + fmix->fi_lastfmm;
		firstm = fmix->fi_firstfmd * 60 + fmix->fi_firstfmm;
		while (lastm - firstm > 10800) lastm -= 21600;
		while (lastm - firstm < -10800) lastm += 21600;
			/* this shouldn't ever happen: */
		if (fmix->fi_firstfm == fmix->fi_lastfm) minint = 15;
		else {
			if (lastm > firstm) minint = (lastm - firstm) /
				(fmix->fi_lastfm - fmix->fi_firstfm);
			else	minint = -((firstm - lastm) /
				(fmix->fi_lastfm - fmix->fi_firstfm));
		     }
		 /* now define each frame name and print it out */
		for (frm = fmix->fi_firstfm, minv = firstm;
		 frm <= fmix->fi_lastfm; frm++)
		   { /* figure out the degree-and-minute value */
			deg = minv / 60;	min = minv % 60;
			if (frm == fmix->fi_firstfm)
			  {
				fmix->fi_firstfmd = deg;
				fmix->fi_firstfmm = min;
			  }
			else if (frm == fmix->fi_lastfm)
			  {
				fmix->fi_lastfmd = deg;
				fmix->fi_lastfmm = min;
			  }
			/* figure out print format and print the line */
			ifmt = (deg <= 9 ? 0 : (deg <= 99 ? 2 : 4));
			if (min > 9) ifmt++;
			fprintf(fped, ffmt[ifmt], frm, dataword, DIVIDER,
				losh->xmm_idna, deg, min);
			/* now determine the next frame */
			minv += minint;	/* frame increment */
			if (minv > 21600) minv -= 21600;
			if (minv < 0) minv += 21600;
		   }
	  }		
	else if ((fmix->fi_flags >> 1) & 01) /* base-36 data */
	  {
		if (fmix->fi_firstfm > 46654) fmix->fi_firstfm = 46654;
		if (fmix->fi_lastfm > 46655) fmix->fi_lastfm = 46655;
		for (frm = fmix->fi_firstfm; frm <= fmix->fi_lastfm; frm++)
		  fprintf(fped, "%5d%s%c%s.%s\n", frm, dataword, DIVIDER,
			losh->xmm_idna, itobase36(frm));
	  }
	else  {
		if (fmix->fi_flags & 01)	/* base-10 data */
		  {
			if (fmix->fi_firstfm > 998) fmix->fi_firstfm = 998;
			if (fmix->fi_lastfm > 999) fmix->fi_lastfm = 999;
		  }
		else
		  {
			if (fmix->fi_firstfm > 9998) fmix->fi_firstfm = 9998;
			if (fmix->fi_lastfm > 9999) fmix->fi_lastfm = 9999;
		  }
		/* okay, temporarily... */
		if ((NULL != ftw->fx_templ) && (ftw->fx_templ[0] != '\0'))
		  {
			for (cp = ftw->fx_templ, ftw->fx_npounds = 0;
			 '\0' != *cp; cp++) if (*cp == '#') (ftw->fx_npounds)++;
			ftw->fx_base = 10;
			tempindx = (ftw->fx_npounds << 24) |
				(ftw->fx_base << 16);
			if (ftw->fx_npounds < 4)
			  {
				fmix->fi_flags |= 01;	tempindx |= 01;
			  }
			strcpy(liname, ftw->fx_templ);
			for (cp = liname, cpi = NULL; '\0' != *cp; cp++)
			   if ('#' == *cp)
			     {
				cpi = cp;	break;
			     }
			if (NULL != cpi)
			  {
			   for (frm = fmix->fi_firstfm;
			     frm <= fmix->fi_lastfm; frm++)
			     {
				outst = itostr(frm, tempindx);
				for (i = 0, cp = cpi; i < ftw->fx_npounds;
				 i++, cp++) *cp = outst[i];
				fprintf(fped, "%5d%s%c%s\n",
				 frm, dataword, DIVIDER, liname);
			     }
			  }
			 else {
				for (frm = fmix->fi_firstfm;
				 frm <= fmix->fi_lastfm; frm++)
				 fprintf(fped, "%5d%s%c%s.%s\n", frm, dataword,
				  DIVIDER, losh->xmm_idna, itobase10(frm));
			      }
		  }
		else {
			for (frm = fmix->fi_firstfm;
			 frm <= fmix->fi_lastfm; frm++)
			 fprintf(fped, "%5d%s%c%s.%s\n", frm, dataword,
				DIVIDER, losh->xmm_idna, itobase10(frm));
			if (NULL == ftw->fx_templ) ftw->fx_templ = mallop(256,
				 "ftw->fx_templ in writeindex");
			if ((fmix->fi_firstfm< 999) && (fmix->fi_lastfm< 999))
				sprintf(ftw->fx_templ, "%s.0###",
					losh->xmm_idna);
			else sprintf(ftw->fx_templ, "%s.####",
				losh->xmm_idna);
		     }
	      }
	(void)fclose(fped);
	sprintf(liname, FILEFORMAT1, losh->xmm_workna,
		losh->xmm_idna, indsuf);
	shortname = filepart(liname, 2);
	if (fmix->fi_flags & 04)
	  {
		printf(
		 "Frame index %s contains frames %d to %d in DEG.MN format%s",
		  shortname, (int)fmix->fi_firstfm, (int)fmix->fi_lastfm, le);
		ifmt = (fmix->fi_firstfmd <= 9 ? 0 :
		 (fmix->fi_firstfmd <= 99 ? 2 : 4));
		if (fmix->fi_firstfmm > 9) ifmt++;
		sprintf(liname, ffmt[ifmt],
		 fmix->fi_firstfm, dataword, DIVIDER, losh->xmm_idna,
		 fmix->fi_firstfmd, fmix->fi_firstfmm);
		ifmt = (fmix->fi_lastfmd <= 9) ?
		 0 : ((fmix->fi_lastfmd <= 99) ? 2 : 4);
		if (fmix->fi_lastfmm > 9) ifmt++;
		sprintf(liname, ffmt[ifmt],
		 fmix->fi_lastfm, dataword, DIVIDER,
		 losh->xmm_idna, fmix->fi_lastfmd, fmix->fi_lastfmm);
	  }
	else printf(
	 "Frame index %s contains frames %d to %d in %s-digit %s format%s",
		shortname, (int)(fmix->fi_firstfm), (int)(fmix->fi_lastfm),
		(fmix->fi_flags & 01) ? "3" : "4",
		(fmix->fi_flags >> 1) & 01 ? "base-36" : "decimal", le);
	/* with some detectors we now attempt to copy a starting
	 pixel-to-cm mapping in */
	(void)startspline(fromcu);
	return;
}

char *grab_template(int fromcu)
{	/* this returns the current filename template */
	FTWORK	*ftw;

	if (NULL == (ftw = manageftw(3))) return NULL;
	if ((NULL == ftw->fx_templ) || ('\0' == *ftw->fx_templ))
		(void)ftables(7 | (fromcu << 4));
	return ftw->fx_templ;
}

char *indexname(int frmno)
{ /*	This routine returns the filename associated with the frame number
	"frmno".  `*dlfmi' points to the index of images.
	If the FRAMEINDEX hasn't yet been established, it's established here.
	The routine returns NULL if the number isn't in the list or if
	the list doesn't exist yet and the attempt to create it fails.
	Jan 06: This routine now subsumes the old `sorder' routine. */
	register int	indno, j;
	int		bot, top;
	FRAMEINDEX	*lfmi;
	FRAMEENTRY	*cft0;

	/* see if the table exists and attempt to create it if needed */
	if (NULL == (lfmi = ftables(3))) return NULL;
	/* Now look for the frame number in the list */
	top = lfmi->fi_lastfm - lfmi->fi_firstfm;	cft0 = lfmi->fi_fmi;
		/* special cases: before initial element or after last one */
	if ((frmno < cft0->fmno) || (frmno > (cft0 + top)->fmno)) return NULL;
	/* otherwise, step through the list and look for a match */
	for (bot = 0; bot != (j = (top + bot)/2); )
	   {
		if (frmno < (cft0 + j)->fmno)	top = j;
		else			bot = j;
	   }
	/* if we get here with frmno at either the top or the bottom,
	 we have a hit. */
	indno = ((cft0 + top)->fmno == frmno) ? top :
	 (((cft0 + bot)->fmno == frmno) ?  bot : -1);
	return (indno < 0) ? NULL : ((lfmi->fi_fmi) + indno)->fmname;
}

/* end of ftable.c */
