/* image.c -- Routines for reading in area detector images into XGEN

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	 5 Jun 2007: read_16bit: enlarge read buffer before reading if needed
	19 Apr 2007: float->double conversions; reorganize use of "{"
	11 Aug 2006: handle byteswapped formats without specific defines
	24 Jun 2006: include xgimage.h; the structures are there
	31 May 2006: nametonum moved from imhead.c to here
	 4 Mar 2006: identify Mar IPs with ismarip();
		strings handled correctly in iscompressed();
		use nametonum() to get frame numbers for needy headers
	31 Dec 2006: skip qsort if fewer than 2 elements
	 5 Dec 2005: replace static global imws with iminfo calls
	24 Nov 2005: read_16bit: pointer arithmetic handled with PAU union
	20 Nov 2005: inline argument declarations and strict prototyping
	16 Nov 2005: imws structure elements replace individual static globals
	16 Aug 2005: FRAME_NOT_DELETED replaced with FRAME_NOT_COMPRESSED
	22 Jun 2005: iscompressed() now declared as static
	 4 May 2005: readflows(): add imagefd argument to saxiflows call
	21 Apr 2005: handle timeouts correctly in open_image()
	 7 Mar 2005: reduce ABSENT_FRAME_WAIT from 600 to 6 sec;
		flag missing compressed images as being okay in read_image()
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	28 Aug 2004: routines used only here are declared as static; cosmetics;
		resort_flows() moved here from imageutil.c
	24 Aug 2004: omit tomainstart() because nobody was calling it
	11 Aug 2004: several routines moved to imageutil.c
	27 Jun 2004: saxiflows() and ascii_flow() moved to saxiim.c
	15 Jun 2004: imageflag bit for gzipping is now correctly abbreviated
	22 Apr 2004: handle overflows for Bruker and R-AXIS data properly
	24 Jan 2004: fujilog,nbytes,nreserved,hofb,dfrac local to one routine
	21 Jan 2004: license message inserted
	13 Dec 2003: eliminate unused variable uspe in correct_pedestal()
	 7 Dec 2003: make iz0[] a static global array, not a global pointer
	 8 Nov 2003: cosmetic changes
	23 Sep 2003: remove declaration of strcpy()--sometimes it's a macro
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	12 Jun 2003: correct diagnostics in cvt_16_to_8();
		make errno.h reference universal so Linux doesn't whine
	 2 May 2003: eliminate PIXENT structure notion altogether
	18 Jan 2003: correct bug in iscompressed() that otherwise
		makes the program uncompress too many images in "xgnow"
	16 Jan 2003: in iscompressed(), null out the `tenam' vector first
	 5 Jan 2003: reorganized source-file header
	30 Dec 2002: compressed (.Z and .gz) files correctly handled;
		replace old routines for that purpose with new ones.
	25 Mar 2002: removed MSI attachments
	 9 Jul 1997: in setting up mp->filename in read_image, explicitly pick
		up the shortened name, then dump the memory for it afterward
	 8 Nov 1996: enable reading images from R-AxisIV detectors
		and the recent CCD detector types
	 9 Sep 1995: in vfy_flowmax for Mar data we increment usp, not ucp.
	24 Aug 1995: Mar data with 16-bit overflows handled correctly
	17 Aug 1995: in read_some8(), read_some16(), and subtract_some(),
		the xybuf vector is 4 bytes per pixel.
	21 Jul 1995: routine `subtract_some' implemented;
		local image flags for read_some changed
	 6 Jul 1995: add concept of "just header+overflows"
		add "read_some8", "read_some16", and "read_box" facilities;
		these allow examination of boxes within images
		without massive reads
	31 May 1995: "register" -> "register int" throughout;
		extract ->frameno with atoi
	20 May 1995: rotate 1536x1536 CCD images by 90 deg
	12 May 1995: able to read FMT_DIP100 and DIP320;
		fix bug in read_16bit for non-square detectors
		increase buffer size for reading one line
	 2 May 1995: Siemens 1024x1024 frames:
		force blank char betw. value and offset strings
	 2 Mar 1995: Fuji plate and compressed data formats;
		counts > 2^31 counted by columns to prevent overflow.
	16 Nov 1994: image-flags argument in call to read_header
	23 Sep 1994: fix bugs for 1024x1024 Siemens frames
	12 Jul 1994: add 2000x2000 MAR-detector formats
	 5 May 1994: source file headers, POLYDET->XGEN,
		"xgulong" and "xglong" constructions
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version
	14 May 1993: allow for data stored not as an 8- or 16-bit image at all,
		but as a list of pixels with accompanying values.
		For now this is a 32-bit address and a 16-bit value; that
		may change.
	13 May 1991: set up so either 8 or 16 bit images are saved and
		most accesses are thru `imvalue' (formerly known as `ifover'),
		so the header picks out the appropriate value.
	 1 Feb 1988: implement optional deletion of an image. 
	30 Aug 1987: Abort waiting if a file with logical name NOWAIT exists.
		Also, allow the calling routine to receive a pointer to a
		DATARUN structure containing interesting info about the xtal.
	28 Apr 1987: if a frame is absent or incomplete,
		wait 10 minutes and look for it again.
		Repeat till 20 waits occur or until the frame appears.
	17 Feb 1987: include input routines
	V.1.xx:	history ill-determined.
	
   Purpose: reading and reformatting detector images from many instruments.
   Cross references: imageout.c, imhead.c, imtime.c

*************************************************************************/
#define	NEED_ERRNO	1
#include	"xgimage.h"
#define		KEEP_16_MIN	(10000)
#define		MINRESERVED		800
#define		INBLOCK		512	/* "blocksize" on Cadmus */
#define		MAX_WAITS	20	/* number of waits before giving up */
#define		ABSENT_FRAME_WAIT	6	/* # seconds per wait */
#define	LOGCONV	(0.813052E-2)	/* ln(4095)/1023: logarithmic converter */

extern void nametonum(MAIL *mp);

size_t imread(AVOID *base, size_t nb)
{ /*	This reads `nb' bytes from file descriptor `imwp->imw_fd'.
	Under VMS it's aware of the block boundary problem and keeps trying
	to read until it gets everything it's looking for. In the special case
	where it finds exactly one byte on a read, it assumes that we have
	a weird VMS file that throws back a CARRIAGE_RETURN occasionally. */
	int		nread;
	IMAGEWORK	*imwp;
#ifdef VMS
	char	*cp;
	int	totread, ntoread;

	errno = 0;	/* clear errors before starting */
	if (NULL == (imwp = iminfo(5))) return 0;
	for (totread = 0, ntoread = nb, cp = (char *)base; totread < nb; )
	   {
		nread = read(imwp->imw_fd, cp, ntoread);
		if (nread <= 0) return totread;	/*give up at EOF or on error */
		/* advance the pointer and diminish the still-to-read counter
		 unless we have one of those weird
		 CARRIAGE_CONTROL=CARRIAGE_RETURN situations */
		if ((nread != 1) || (ntoread == 1))
		  {
			ntoread -= nread; totread += nread; cp += nread;
		  }
		if (errno) return totread; /* give up on error */
	   }
	return totread;
#else
	if (NULL == (imwp = iminfo(5))) return 0;
	nread = read(imwp->imw_fd, base, (unsigned)nb); return nread;
#endif
}

static void resort_flows(MAIL *mp)
{ /*	special fix associated with unsorted data:
	check the overflow list to make sure it's sorted in decreasing order
	in Y and secondarily in increasing order in X.  If it isn't, fix. */
	u_short		prx, pry;
	int		unsorted;
	VOFLOW		*vofp, *oend;

	vofp = mp->overflow; prx = vofp->xcoord; pry = vofp->ycoord;
	unsorted = 0;	/* assume that it's sorted right */
	oend = mp->overflow + mp->oflowcount;
	for (vofp++; vofp < oend; vofp++)
	   {
		if ((vofp->ycoord < pry) ||
		 ((vofp->ycoord == pry) && (vofp->xcoord < prx)))
		  {
			unsorted = 1; break;
		  }
		prx = vofp->xcoord;	pry = vofp->ycoord;
	   }
	if ((unsorted) && (mp->oflowcount > 1)) qsort((AVOID *)mp->overflow,
	 (unsigned)mp->oflowcount, sizeof (VOFLOW), (QPE)ofycmp);
}

static int marflows(MAIL *mp)
{ /*	This goes into a Mar image and records the 16-bit overflows. To
	save space and adhere to the concept of a `VOFLOW' structure, the
	resulting value is actually kept as (true value - 65535) / 4; so
	if the recorded value is 71384, the stored value will be 1462. My
	impression is that the largest possible value the Mar can read is
	0x3FFFF = 262143, so that will read out as (262143-65535)/4 = 49152. */
	int		fno, offs, novers, yen;
	size_t		ntoread;
	off_t		marblock, start;
	MARFLOW		mflo;
	u_char		*cp;
	VOFLOW		*vofp;
	IMAGEWORK	*imwp;

	if (mp->oflowcount < 1) return EOK; /* see if we're needed */
	/* figure out where to read and how much to read */
	marblock = mp->ydimen * sizeof (u_short);
	start = (mp->xdimen + 1) * marblock;
	ntoread = mp->oflowcount * sizeof (MARFLOW);
	/* move to the right place and do the reading. */
	if (NULL == (imwp = iminfo(5))) return FRAME_NO_MEMORY;
	if (-1 == lseek(imwp->imw_fd, start, 0)) return FRAME_BADFLOWS;
	if (ntoread != imread((AVOID *)(imwp->imw_buf), ntoread))
		return FRAME_BADFLOWS;
	/* re-position the read pointer at the beginning of the 16bit data */
	if (-1 == lseek(imwp->imw_fd, marblock, 0)) return FRAME_BADFLOWS;
	yen = mp->ydimen - 1;	novers = mp->oflowcount;
	if (mp->improps & IPR_BYTESWAPPED)
	  {
	    for (cp = imwp->imw_buf, fno = 0, vofp = mp->overflow;
		fno < novers; fno++, cp += 8, vofp++)
		   {
			cpymem((AVOID *)&mflo, (AVOID *)cp, sizeof (MARFLOW));
			offs = swawb(mflo.mxyoff);
			vofp->xcoord = offs % mp->xdimen - 1;
			vofp->ycoord = yen - offs / mp->xdimen;
			vofp->value = (swawb(mflo.minten) - 65535) / 4;
		   }
	   }
	else {
		for (cp = imwp->imw_buf, fno = 0, vofp = mp->overflow;
		 fno < novers; fno++, cp += 8, vofp++)
		   {
			cpymem((AVOID *)&mflo, (AVOID *)cp, sizeof (MARFLOW));
			offs = mflo.mxyoff;
			vofp->xcoord = offs % mp->xdimen - 1;
			vofp->ycoord = yen - offs / mp->xdimen;
			vofp->value = (mflo.minten - 65535) / 4;
		   }
	     }
	if (mp->oflowcount > 1) resort_flows(mp); /* sort them */
	return FRAME_OK;
}

static int vmsflows(MAIL *mp)
{ /*	This handles overflows in a VMS-style input image */
	off_t		seekpoint;
	size_t		nby, nnow;
	IMAGEWORK	*imwp;

	seekpoint = DEFBLOCK;
	/* first, position ourselves properly */
	if (NULL == (imwp = iminfo(5))) return FRAME_NO_MEMORY;
	if (-1L == lseek(imwp->imw_fd, seekpoint, 0))
	  {
		fprintf(stderr,
		 "Cannot seek to beginnning over VMS-style overflow table\n");
		return FRAME_BADFLOWS;
	  }
	/* read in the overflows */
	nby = mp->oflowcount * sizeof (VOFLOW);
	if (nby != (nnow = imread((AVOID *)(mp->overflow), nby)))
	  {
		fprintf(stderr,
 "Attempted to read %6d bytes of overflow-table data; got %6d\n",
		 (int)nby, (int)nnow); return FRAME_BADFLOWS;
	  }
	/* put pointer @ next 2048-byte boundary if necessary*/
	if (NULL == (imwp = iminfo(5))) return FRAME_NO_MEMORY;
	if (nby % DEFBLOCK)
	  {
		nby = DEFBLOCK - nby % DEFBLOCK;
		if (nby != (nnow = imread((AVOID *)(imwp->imw_buf), nby)))
		  {
			fprintf(stderr,
 "Attempted to read %6d bytes to reposition overflow-table pointer; got %6d\n",
				(int)nby, (int)nnow); return FRAME_BADFLOWS;
		  }
	  }
	return FRAME_OK;
}

static int pcsflows(MAIL *mp, off_t seekpoint)
{ /*	This handles overflows in a PCS-style image */
	size_t		nby, nnow;
	xgulong		offset;
	VOFLOW		*vofp, *oend;
	IMAGEWORK	*imwp;

		/* position ourselves properly */
	if (NULL == (imwp = iminfo(5))) return FRAME_NO_MEMORY;
	if (-1L == lseek(imwp->imw_fd, (off_t)seekpoint, 0))
	  {
		fprintf(stderr,
 "Error seeking to beginning of overflow data in PCS-style image\n");
		return FRAME_BADFLOWS;
	  }
		/* figure out how much space the overflows occupy */
	nby = sizeof (OFLOW) * (((unsigned)mp->oflowcount + 3) & ~3);
	if (nby != (nnow = imread((AVOID *)(mp->overflow), nby)))
	  {
		fprintf(stderr,
 "Tried to read %6d bytes of overflow data; got %5d\n", (int)nby, (int)nnow);
		return FRAME_BADFLOWS;
	  }
	/* convert from "OFLOW" format to the "VOFLOW" format*/
	oend = mp->overflow + mp->oflowcount;
	for (vofp = mp->overflow; vofp < oend; vofp++)
	   {
#ifdef LITTLEND /* convert offset-style overflow to (x,y)-style overflow*/
		offset = (((xgulong)uswab(vofp->xcoord)) << 16)
		 | ((xgulong)uswab(vofp->ycoord));
		if (mp->xdimen == 512)
		  {
			vofp->xcoord = (offset & 0x1ff);
			vofp->ycoord = offset >> 9;
		  }
		else  {
			vofp->xcoord = (offset & 0x3ff);
			vofp->ycoord = offset >> 10;
		      }
		vofp->value = uswab(vofp->value);
#else
		if (imwp->imw_filetype == FMT_PCS_8)
		  {
			offset = (((xgulong)vofp->xcoord) << 16)
			 | ((xgulong)vofp->ycoord);
			if (mp->xdimen == 512)
			  {
				vofp->xcoord = offset & 0x1ff;
				vofp->ycoord = offset >> 9;
			  }
			else {
				vofp->xcoord = offset & 0x3ff;
				vofp->ycoord = offset >> 10;
			     }
		  }
		else { /* byte-swap it: already in (x,y) fmt */
			vofp->xcoord = uswab(vofp->xcoord);
			vofp->ycoord = uswab(vofp->ycoord);
			vofp->value = uswab(vofp->value);
		     }
#endif
	   }
	return FRAME_OK;
}

static int readflows(MAIL *mp)
{ /*	This gets the overflows from the file and puts them in the overflow
	list.  Specific formats call forth different processing procedures. */
	int		nnow, swfmt;
	off_t		seekpoint;
	IMAGEWORK	*imwp;

	if (mp->oflowcount <= 0) return FRAME_OK;	/* no overflows */
	if (NULL == (imwp = iminfo(5))) return FRAME_NO_MEMORY;
	if (ismarip(imwp->imw_filetype)) return marflows(mp);
	swfmt = swawb(imwp->imw_filetype);	nnow = FRAME_FMT_UNKNOWN;
	switch (swfmt) /* see if the filetype is a byte-swapped example */
	      {
		case	FMT_SAXII:	case	FMT_SAXII_16:
			if (FRAME_OK != (nnow = saxiflows(mp, imwp->imw_fd)))
				return nnow;
		case	FMT_PCS_8:	case	FMT_CAD:
#ifdef LITTLEND
			seekpoint = 2 * INBLOCK;
#else
			seekpoint = INBLOCK;
#endif
			if (FRAME_OK != (nnow = pcsflows(mp, seekpoint)))
				return nnow;
			break;
		default:	break;
	     }
	if (FRAME_OK != nnow) /* if we haven't got it yet, try the others */
	  {
	    switch (imwp->imw_filetype) 
	      {
		case	FMT_VMS:	/* VMS format */
#ifdef LITTLEND
		case	FMT_PCS_8:	/* PCS (unswapped) format */
#endif
			if (FRAME_OK != (nnow = vmsflows(mp))) return nnow;
			break;
		case	FMT_SAXII: case	FMT_SAXII_16:
			if (FRAME_OK != (nnow = saxiflows(mp, imwp->imw_fd)))
				return nnow;
			break;
#ifndef LITTLEND
		case	FMT_PCS_8:
			seekpoint = 2 * INBLOCK;
			if (FRAME_OK != (nnow = pcsflows(mp, seekpoint)))
				return nnow;
			break;
#endif
		default:	/* all legal cases exhausted... */
			fprintf(stderr,
 "Attempted to obtain overflows from image in format 0x%x; no code avail!\n",
				imwp->imw_filetype);
			return FRAME_BADFLOWS;
	       } /* end of close on switch */
	  }
#ifndef LITTLEND
	/* adjust pointer to start of 8-bit data */
	if ((swfmt == FMT_CAD) || (swfmt == FMT_PCS_8))
	  {
		if (-1L == (nnow = lseek(imwp->imw_fd, (off_t)0L, 1)))
		  {
			fprintf(stderr,
 "Error seeking to beginning of 8-bit data\n"); return FRAME_BADFLOWS;
		  }
		if ((nnow = nnow % DEFBLOCK))
		  {
			seekpoint = (xglong)(DEFBLOCK - nnow);
			if (0 == lseek(imwp->imw_fd, seekpoint, 1))
			  {
				fprintf(stderr,
				 "Error seeking to beginning of block\n");
				return FRAME_BADFLOWS;
			  }
		  }
	  }
#endif
	if (mp->oflowcount > 1) resort_flows(mp); /*sort overflows */
/* now write out the header and the overflows */
	return FRAME_OK;
}

static int cvt_16_to_8(MAIL *mp, int nflows)
{ /*	This takes a 16-bit image and converts to an 8-bit image.  We assume
	the user has counted overflows beforehand so we don't have to now. */
	register int	ix, iy;
	int		now_reserved, wid, hei;
	u_short		*usp;
	u_char		*ucp;
	VOFLOW		*vofp;
	DATARUN		*drup;
	IMAGEWORK	*imwp;
	static int	nreserved;

	ix = mp->xdimen * mp->ydimen;
	if (NULL == mp->data8)
	  {
		if (NULL == (mp->data8 = (u_char *)callop((size_t)ix,
		 sizeof (u_char),"mp->data8 in cvt_16_to_8")))
			return FRAME_NO_MEMORY;
	  }
	mp->oflowcount = nflows;
	nreserved = (nflows > MINRESERVED ? nflows: MINRESERVED);
	drup = mp->runheader;
	now_reserved = (NULL == drup) ? MINRESERVED :
		drup->run_oflowend - mp->overflow;
	if (nreserved > now_reserved)
	  {
		memzap((AVOID **)&(mp->overflow), "overflow in cvt_16_to_8");
		if (NULL == (mp->overflow = (VOFLOW *)callop((unsigned)
		 nreserved, sizeof (VOFLOW), "mp->overflow in cvt_16_to_8")))
			return FRAME_NO_MEMORY;
		drup->run_oflowend = mp->overflow + nreserved;
	  }
	vofp = mp->overflow;	usp = mp->data16;	ucp = mp->data8;
	wid = mp->xdimen;	hei = mp->ydimen;
	for (ix = 0; ix < wid; ix++) for (iy = 0; iy < hei; iy++, usp++, ucp++)
	   {
		if (*usp > 255)
		  {
			*ucp = 255;	vofp->value = *usp; vofp->xcoord = ix;
			vofp++->ycoord = mp->ydimen - iy - 1;
		  }
		else *ucp = *usp;
	   }
	if (NULL == (imwp = iminfo(1))) return FRAME_NO_MEMORY;
	if (imwp->imw_filetype != SDMS_RANGE + mp->ydimen)
		memzap((AVOID **)&(mp->data16), "data16 in cvt_16_to_8");
	mp->improps |= IPR_IS8BIT;	/* now it's an 8-bit image */
	/*  overflow table needs to be sorted now: */
	if (mp->oflowcount > 1) resort_flows(mp); return FRAME_OK;
}

int cvt_8_to_16(MAIL *mp, int ftype)
{ /*	This takes an 8bit image and converts it to a 16-bit image. */
	register int	ix, iy;
	int		loff, wm1, hm1;
	u_char		*ucp;
	u_short		*usp, *usq, *usr, *uss, *ust, *usu;
	VOFLOW		*vofp, *oend;

	ix = mp->xdimen * mp->ydimen;
	if (NULL == mp->data16)
	  {
		if (NULL == (mp->data16 = (u_short *)callop((size_t)ix,
		  sizeof (u_short), "mp->data16 in cvt_8_to_16")))
			return FRAME_NO_MEMORY;
	  }
	/* first convert the 8-bit data */
	usp = mp->data16;	ucp = mp->data8;
	wm1 = mp->xdimen - 1;	hm1 = mp->ydimen - 1;
	for (ix = 0; ix <= wm1; ix++)
	   for (iy = 0; iy <= hm1; iy++, usp++, ucp++) *usp = *ucp;
	/* now put in the overflows */
	for (vofp = mp->overflow, oend = mp->overflow + mp->oflowcount;
	 vofp < oend; vofp++)
	   {
		ix = vofp->xcoord; iy = mp->ydimen - vofp->ycoord - 1;
		loff = ix * mp->ydimen + iy;
		usp = mp->data16 + loff;
		if (*usp == 255) *usp = vofp->value;
		else  {
			uss = mp->data16 + (wm1 - ix) * mp->ydimen + iy;
			ust = mp->data16 + (wm1 - iy) * mp->xdimen + ix;
			usr = mp->data16 + (wm1 - ix) * mp->ydimen +
				(mp->ydimen - 1 - iy);
			usu = mp->data16 + ix * mp->ydimen + (hm1 - iy);
			usq = mp->data16 + iy * mp->xdimen + ix;
			printf(
	"Overflow [%4d,%4d]=%4d: %3d %3d %3d %3d %3d\n",
			 ix, iy, vofp->value, *uss, *ust, *usr, *usu, *usq);
		      }
	   }
	if (ftype != (int)(SDMS_RANGE + mp->ydimen))
		memzap((AVOID **)&(mp->data8), "data8 in cvt_8_to_16");
	mp->oflowcount = 0;	/* no more 8-bit overflows */
	mp->improps &= ~IPR_IS8BIT;	return FRAME_OK;
}

static int correct_pedestal(MAIL *mp)
{ /*	This recognizes and modifies an image containing a pedestal */
	int	row, rowadv, ncolus, regi, ncts, col;
	int	nxdimen, nydimen, nregions, nz;
	u_short	*usp, *usp0, *uspr, *uspo, *uspor, *startreg, *endreg;
	double	zp, zp2, cor, meanped[16];

	/* first make sure there really is one */
	if (0 != (mp->xdimen % 1056)) return FRAME_OK;
	/* at the moment we can't handle the 2x2 */
	if (mp->xdimen > 2112) return FRAME_OK;
	/* first calculate the mean pedestal value for each readout region */
	nregions = (mp->xdimen <= 3000) ? 4 : 16;
	rowadv = mp->xdimen;	ncolus = mp->ydimen;
	usp0 = mp->data16;
	nxdimen = rowadv - 64;	nydimen = ncolus - 64;
	for (regi = 0; regi < nregions; regi++)
	   {
		if (regi < 2) startreg = usp0 + 8 * rowadv + 32;
		else startreg = usp0 + (ncolus - 20) * rowadv + 32;
		if (regi % 2) startreg += rowadv / 2 - 32;
		endreg = startreg + rowadv * 12;
		zp2 = zp = ncts = nz = 0;
		for (uspr = startreg; uspr < endreg; uspr += rowadv)
		   {
			for (usp = uspr; usp < uspr + nxdimen / 2;
			  usp++,ncts++)
			   {
				cor = (double)(*usp); zp += cor;
				zp2 += cor * cor; if (!(*usp)) nz++;
			   }
		   }
		if (ncts > 0)
		  {
			meanped[regi] = zp = zp / (double)ncts;
			zp2 = zp2 / (double)ncts - zp * zp;
		  }
		else	zp2 = zp = 0.;
		zp2 = (zp2 >= 0.) ? sqrt(zp2) : 0.;
	   }
	zp = meanped[0];	if (meanped[1] < zp) zp = meanped[1];
	if (meanped[2] < zp) zp = meanped[2];
	if (meanped[3] < zp) zp = meanped[3];
	if (zp < 0.5) return FRAME_OK;	/* no corrections required */
	/* now the pedestals are applied and slid */
	uspr = usp0 + 32 * rowadv;
	for (row = 0, uspor = usp0; row < nydimen;
	 row++, uspr += rowadv, uspor += nxdimen)
	   {
		if (row < nydimen / 2) regi = 0;	else regi = 2;
		zp = meanped[regi];
		for (col = 0, usp = uspr + 32, uspo = uspor;
		  col < nxdimen / 2; col++, usp++, uspo++)
		   {
			cor = ((double)(*usp)) - zp;
			*uspo = (cor < 0) ? 0 : cor;
		   }
		zp = meanped[regi + 1];
		for ( ; col < nxdimen; col++, usp++, uspo++)
		   {
			cor = ((double)(*usp)) - zp;
			*uspo = (cor < 0) ? 0 : cor;
		   }
	   }
	mp->xdimen = nxdimen;	mp->ydimen = nydimen;
	return FRAME_OK;
}

static int read_16bit(MAIL *mp, int rflags)
{ /*	this reads in a 16-bit image and (if necessary) rotates it,
	i.e. puts the rows onto columns.  It counts overflows as it
	does this.  If there aren't very many overflows it leaves the
	data in 8-bit format; if there are it creates an overflow table
	and sticks the 8-bit data in the 8-bit data array.  In the case
	in which dark current is going to be subtracted from the image
	we leave the data in 16-bit format till after subtraction.
	The event count is summed 1 column at a time to prevent overflow. */
	register int		i, j, deltax, deltay, rcounts;
	int			wid, hit, widm1, hitm1;
	int			nflows, startoff, doswap, tcounts;
	size_t			ngot, nskip;
	u_short			*rsp0, *runsp, *uspi, *uspo;
	u_int			utcounts;
	double			fy;
	IMAGEWORK		*imwp;
	PAU			imbu;
	static size_t		nbytes;
	static u_short		*fujilog;

	wid = mp->xdimen;	hit = mp->ydimen;
	widm1 = wid - 1;	hitm1 = hit - 1;
	startoff = 0;	deltax = wid;	deltay = 1;
	if (mp->improps & IPR_XFASTEST)
	  {
		deltax = 1;	deltay = hit;
	  }
	if (mp->improps & IPR_STARTBOTTOM)
	  {
		deltay *= -1;	startoff += wid * hitm1;
	  }
	if (mp->improps & IPR_STARTRIGHT)
	  {
		deltax *= -1;	startoff += widm1;
	  }
	rsp0 = mp->data16 + startoff;
	nbytes = mp->xdimen * 2;	doswap = 0;
	/* if we come in with the (rflags & IPR_PAD) bit set, we skip
	 4096 - 3800 bytes or 2048 - 1900 bytes after each row. */
	if (mp->improps & IPR_PAD)
		nskip = (mp->xdimen == 950) ? 2048 - nbytes : 4096 - nbytes;
	else	nskip = 0;
	doswap = 0;	if (mp->improps & IPR_BYTESWAPPED) doswap = 1;
	/* handler for data that invert the customary byte-swap sense */
	if (mp->improps & IPR_INVERTSWAP) doswap = 1 - doswap;
	nflows = tcounts = 0;
	/* for Fuji-plate data, assemble a log-to-linear conversion table */
	if ((mp->improps & IPR_LOGSCALE) && (NULL == fujilog))
	  {
		if (NULL == (fujilog = (u_short *)callop(1024,sizeof (u_short),
		 "fujilog in read_16bit"))) return FRAME_NO_MEMORY;
		for (i = 0, uspi = fujilog; i < 1024; i++, uspi++)
		   {
			fy = i;	fy = exp(LOGCONV * fy);
			*uspi = fy + 0.49999;
		   }
	  }
	if (NULL == (imwp = iminfo(5))) return FRAME_NO_MEMORY;
	if (nbytes > imwp->imw_bsize)
	  {
		imwp = iminfo(4);	imwp->imw_bsize = nbytes;
		if (NULL == (imwp = iminfo(5))) return FRAME_NO_MEMORY;
	  }
	if (doswap)
	 {
	    for (i = 0, runsp = rsp0; i < hit; i++, runsp += deltax)
	      {
		if (nbytes != (ngot = imread((AVOID *)(imwp->imw_buf),nbytes)))
		  {
			fprintf(stderr, "failure 0,i=%d;want %d byt,got %d\n",
				i, (int)nbytes, (int)ngot);
			return FRAME_BAD16BIT;
		  }
		if (mp->improps & IPR_LOGSCALE)
		  {
			imbu.ucpt = imwp->imw_buf;	uspi = imbu.uspt;
			for (j = 0; j < wid; j++,uspi++)
				*uspi = (*uspi < 1024) ? fujilog[*uspi] : 4095;
		  }
		imbu.ucpt = imwp->imw_buf;	uspi = imbu.uspt;
		for (rcounts = j = 0, uspo = runsp;
		 j < wid; j++, uspi++, uspo += deltay)
		   {
			*uspo = uswab(*uspi);	if (*uspo > 255) nflows++;
			rcounts += *uspo;
		   }
			/* flush blanks at end of each row if necessary */
		if (nskip > 0)
		  {
			if (nskip != (ngot = imread((AVOID *)(imwp->imw_buf),
			 nskip)))
			  {
				fprintf(stderr,
 "Swapped Raxis flush failure, row %d: nskip = %d, got %d\n",
					i, (int)nskip, (int)ngot);
				return FRAME_BAD16BIT;
			  }
		  }
		/* add this row's counts into the total, carefully. A negative
		 total count is a signal that the real count is > 2^31;
		 we count in a negative direction and left-shift 6 bits. */
		if (tcounts < 0) tcounts -= rcounts / 64;
		else {
			utcounts = tcounts + (u_int)rcounts;
			if (utcounts > 2147483646)
				tcounts = -((int)(utcounts / 64));
			else	tcounts = (int)utcounts;
		     }
	      }
	 }
	else
	 {
	  for (i = 0, runsp = rsp0; i < hit; i++, runsp += deltax)
	    {
		if (nbytes != (ngot = imread((AVOID *)(imwp->imw_buf),nbytes)))
		  {
			fprintf(stderr, "failure 2,i=%d;want %d byt,got %d\n",
				i, (int)nbytes, (int)ngot);
			return FRAME_BAD16BIT;
		  }
		if (mp->improps & IPR_LOGSCALE)
		  {
			imbu.ucpt = imwp->imw_buf;	uspi = imbu.uspt;
			for (j = 0; j < wid; j++,uspi++)
				*uspi = (*uspi < 1024) ? fujilog[*uspi] : 4095;
		  }
		imbu.ucpt = imwp->imw_buf;	uspi = imbu.uspt;
		for (rcounts = j = 0, uspo = runsp;
		 j < wid; j++, uspi++, uspo += deltay)
		   {
			*uspo = *uspi;	if (*uspo > 255) nflows++;
			rcounts += *uspo;
		   }
			/* flush blanks at end of each row if necessary */
		if (nskip > 0)
		  {
			if (nskip != imread((AVOID *)(imwp->imw_buf), nskip))
			  {
				fprintf(stderr,
 "Raxis flush failure, row %d: nskip = %d, got %d\n",
					i, (int)nskip, (int)ngot);
				return FRAME_BAD16BIT;
			  }
		  }
		if (tcounts < 0) tcounts -= rcounts / 64;
		else {
			utcounts = tcounts + (u_int)rcounts;
			if (utcounts > 2147483646)
				tcounts = -((int)(utcounts / 64));
			else	tcounts = (int)utcounts;
		     }
	    }
	 }
	/* blank out column zero in Quantex data */
	if ((imwp->imw_filetype == FMT_QTEX) ||
	 (imwp->imw_filetype == swawb(FMT_QTEX)))
	  {
		for (j = 0, uspi = mp->data16; j < hit; j++, uspi++)
		   {
			if (tcounts > 0) tcounts -= *uspi;
			else tcounts += (*uspi) / 64;
			if (*uspi > 255) nflows--;	*uspi = 0;
		   }
	  }
	mp->eventcount = tcounts;
	/* handle data from detectors with uncorrected pedestals */
	if ((mp->improps & IPR_HASPEDESTAL) &&
	  (FRAME_OK != (ngot = correct_pedestal(mp)))) return ngot;
	/* Aug 95: for Mar data with 16-bit overflows, we don't
	 convert to 8-bit data because the formats are incompatible */
	if ((nflows < KEEP_16_MIN) && !(rflags & DARK_CURRENT) &&
	 !(rflags & IPR_ISDARK) && !(mp->oflowcount))
		return cvt_16_to_8(mp, nflows);
	return FRAME_OK;
}

static int subtract_dark(MAIL *mp, int subfrac)
{ /*	This subtracts the dark-current image from the current image.
	By default we subtract 0.8 * the dark-current image; if the
	second argument (subfrac) is nonzero, it is taken to be
	8000 * the fraction to be subtracted.  Thus if subfrac = 6400,
	then we'll subtract 6400/8000 = 0.8 * the dark-current image;
	if subfac = 8000 we subtract the whole thing. */

	int			hit, ix, iy, nflows, rcounts, tcounts, wid;
	u_short			oldv, dval, *dsp, *usp;
	u_int			utcounts;
	double			latefrac, darkfrac, revent, rlate;
	IMAGEWORK		*imwp;
	static double		dfrac = 0.8;

	darkfrac = (subfrac > 0 ? ((double)subfrac) / 8000. : dfrac);
	/* remember what fraction of the counts were originally late
	so we can rescale */
	revent = (mp->eventcount > 0) ? mp->eventcount :
		-64. * (double)(mp->eventcount);
	rlate = (mp->late > 0) ? mp->late : -64. * (double)(mp->late);
	latefrac = (mp->eventcount) ? rlate / revent : 0.;
		/* make sure both images start out in 16-bit format */
	if (NULL == (imwp = iminfo(025))) return FRAME_NO_MEMORY;
	if ((mp->data8) && (FRAME_OK != (ix = cvt_8_to_16(mp,
	 imwp->imw_filetype))))
	  {
		fprintf(stderr, "Error %d in cvt_8_to_16\n", ix);
		return FRAME_BAD8BIT;
	  }
	if (((imwp->imw_darkp)->data8) && (FRAME_OK != (ix = cvt_8_to_16(
	 (imwp->imw_darkp), imwp->imw_filetype))))
	  {
		fprintf(stderr, "Error %d in cvt_8_to_16 for darkimage\n", ix);
		return FRAME_BAD8BIT;
	  }
	/* go through the data, comparing the rescaled value in the
	 dark-current image to the value we actually got.  We subtract out
	 the dark-current value from the current value for each pixel,
	 never letting the value fall below 1 unless it was already 0. */
	usp = mp->data16;	dsp = (imwp->imw_darkp)->data16;
	wid = mp->xdimen;	hit = mp->ydimen;
	for (ix = nflows = tcounts = 0; ix < wid; ix++)
	   {
		for (rcounts = iy = 0; iy < hit; iy++, usp++, dsp++)
		   {
			if ((dval = (int)(darkfrac * (double)(*dsp))) != 0)
			  {
				if (0 != (oldv = *usp))
				  {
					*usp = (oldv <= dval? 1: oldv - dval);
					if (*usp > 255) nflows++;
				  }
			  }
			rcounts += (int)(*usp);
		      }
		if (tcounts < 0) tcounts -= rcounts / 64;
		else {
			utcounts = tcounts + (u_int)rcounts;
			if (utcounts > 2147483646)
				tcounts = -((int)(utcounts / 64));
			else	tcounts = (int)utcounts;
		     }
	   }
	mp->eventcount = tcounts;
/* now fix up the late counts and ontime counts */
	mp->late = (latefrac > DFUZZ) ?
		latefrac * ((double)tcounts) + 0.49999 : 0;
	mp->ontime = mp->eventcount - mp->late;
/* convert to 8-bit if appropriate */
	if (nflows < KEEP_16_MIN) return cvt_16_to_8(mp, nflows);
	else return FRAME_OK;
}

static int read_8bit(MAIL *mp, int rflags)
{ /*	This reads in 8-bit data and, if appropriate, rotates it into position.
	The decision of whether or not to rotate is hidden in the values
	of the increment paramters `deltax' and `deltay' and the corresponding
	pointer starting value `runcp'. */
	register int	i, j;
	int		deltax, deltay, wid, hit, widm1, hitm1;
	int		startoff, wm1, hm1;
	size_t		ngot, ntor;
	u_char		*cpi, *cpo, *runcp;
	DATARUN		*dp;
	VOFLOW		*vofp;
	VOFL16		*v16p;
	IMAGEWORK	*imwp;

	wid = mp->xdimen;	hit = mp->ydimen;
	widm1 = wid - 1;	hitm1 = hit - 1;
	ntor = hit;		/* number of bytes read per operation */
	startoff = 0;	deltax = wid;	deltay = 1;
	dp = mp->runheader;
	if (mp->improps & IPR_XFASTEST)
	  {
		deltax = 1;	deltay = hit;
	  }
	if (mp->improps & IPR_STARTBOTTOM)
	  {
		deltay *= -1;	startoff += wid * hitm1;
	  }
	if (mp->improps & IPR_STARTRIGHT)
	  {
		deltax *= -1;	startoff += widm1;
	  }
	runcp = mp->data8 + startoff;
	vofp = mp->overflow;	v16p = mp->oflow16;
	/* having established our read order, we do the reading */
	hm1 = mp->ydimen - 1;	wm1 = mp->xdimen - 1;
	if (NULL == (imwp = iminfo(5))) return FRAME_NO_MEMORY;
	for (i = 0; i < wid; i++, runcp += deltax)
	   {
		if (ntor != (ngot = imread((AVOID *)(imwp->imw_buf), ntor)))
		  {
			if (i < wid - 4)
			  {
				fprintf(stderr,
 "EOF in image in column %d: sought %d bytes, got %d\n",
				 i, (int)ntor, (int)ngot);
				return FRAME_BAD8BIT;
			  }
			if (rflags & GRIPE_IF_SHORT) fprintf(stderr,
	 "Warning: frame %s has only %d good columns\n", mp->filename, i);
		  }
		cpi = imwp->imw_buf;	cpo = runcp;
		for (j = 0; j < hit; j++, cpi++, cpo += deltay)
		   {
			*cpo = *cpi;
			if ((*cpo == 255) && (dp->run_wtsum > 99.))
			  {
				vofp->xcoord = hm1 - j;
				vofp->ycoord = wm1 - i; vofp++;
				if (vofp->value == 65535)
				  {
					v16p->olxcoord = hm1 - 1 - j ;
					v16p->olycoord = wm1 - i; v16p++;
				  }
			  }
		   }
	   }
	return FRAME_OK;
}

static void verify_flowmax(MAIL *mp)
{ /*	Check that all overflows appear with 8-bit value 255 and
	redefine zmax, the largest value in the data.*/
	int	nonsenseflow, thisv;
	xgulong	checkmax, ofv, vv;
	u_short	cmaxx, cmaxy, thisx, thisy, *usp;
	u_char	*ucp;
	VOFLOW	*vofp, *oend;
	VOFL16	*v16p, *v16e;

	checkmax = cmaxx = cmaxy = nonsenseflow = 0;
	if (mp->improps & IPR_IS8BIT)
	  {
	    ucp = mp->data8;
	    if (mp->oflowcount)
	      {
		oend = mp->overflow + mp->oflowcount;
		for (vofp = mp->overflow; vofp < oend; vofp++)
		   {
			thisx = vofp->xcoord;
			thisy = mp->ydimen - vofp->ycoord - 1;
			if (!thisx && (!(vofp->ycoord))) nonsenseflow++;
			else if ((thisx >= mp->xdimen) ||
			 (thisy >= mp->ydimen)) nonsenseflow++;
			else {
				thisv = *(ucp + (thisx * mp->ydimen + thisy));
				if (thisv != 255) fprintf(stderr,
 "Warning: Overflow = %5d recorded at [%3d,%3d]; 8bit value= %3d\n",
					(int)vofp->value, (int)thisx,
					(int)thisy, thisv);
				if (checkmax < vofp->value)
				  {
					if (mp->improps & IPR_RAXISCOMPRESS)
					  {
						vv = vofp->value;
						if (vv > 32768)
						  vv = 32 * (vv - 32768);
						if (checkmax < vv)
						  {
							checkmax = vv;
							cmaxx = thisx;
							cmaxy = thisy;
						  }
					  }
					else
					  {
						checkmax = vofp->value;
						cmaxx = thisx;	cmaxy = thisy;
					  }
				  }
			     } /* end of "else" of "if (!thisx " clause */
		   } /* end of "for (vofp" loop */
		if (nonsenseflow > 1) fprintf(stderr,
 "Warning: %6d nonsensical (x,y)=0 overflows recorded in %s\n",
			nonsenseflow, mp->filename);
	      } /* end of "if (mp->oflowcount" clause */
		/* if we get here either there are no overflows or
			none of them was valid. Look for zmax in 8bit data. */
	    if (!checkmax)
	      {
		for (thisx = 0; thisx < mp->xdimen; thisx++)
		   for (thisy = 0; thisy < mp->ydimen; thisy++, ucp++)
		      {
			if (*ucp > checkmax)
			  {
				checkmax = *ucp; cmaxx = thisx; cmaxy = thisy;
			  }
		      }
	      } /* end of "if (!checkmax" clause */
	  } /* end of "if (mp->improps & IPR_IS8BIT)" clause */
	else {
		usp = mp->data16;
		if (mp->oflowcount) /* Mar or Siemens data, 16-bit o'flows*/
		  {
			oend = mp->overflow + mp->oflowcount;
			for (vofp = mp->overflow; vofp < oend; vofp++)
			   {
				thisx = vofp->xcoord;
				thisy = mp->ydimen - vofp->ycoord - 1;
				if (!thisx && (!(vofp->ycoord))) nonsenseflow++;
				else if ((thisx >= mp->xdimen) ||
				 (thisy >= mp->ydimen)) nonsenseflow++;
				else {
					thisv = *(usp +
					 (thisx * mp->ydimen + thisy));
					ofv = 4 * ((int)(vofp->value)) + 65535;
					if (thisv != 65535) fprintf(stderr,
	 "Warning: Overflow = %5d recorded at [%3d,%3d]; 16-bit value= %6d\n",
						ofv, (int)thisx,
						(int)thisy, thisv);
					if (checkmax < ofv)
					  {
					    if (mp->improps & IPR_RAXISCOMPRESS)
					      {
						if (ofv > 32767)	
							vv = 32 * (ofv - 32768);
						else vv = ofv;
						if (vv > checkmax)
						  {
							checkmax = vv;
							cmaxx = thisx;
							cmaxy = thisy;
						  }
					      }
					    else
					      {
						checkmax = ofv;
						cmaxx = thisx; cmaxy = thisy;
					      }
					  }
				     } /* end of "else" of "if (!t" clause */
			   } /* end of "for (vofp" loop */
			if (nonsenseflow > 1) fprintf(stderr,
	 "Warning: %6d nonsensical (x,y)=0 overflows recorded in %s\n",
				nonsenseflow, mp->filename);
		      } /* end of "if (mp->oflowcount" clause */
		else if ((mp->count1) && (NULL != mp->oflow16))
		  {
			v16p = mp->oflow16;	v16e = v16p + mp->count1;
			checkmax = 65534;	cmaxy = cmaxx = 0;
			for (checkmax = 65535; v16p < v16e; v16p++)
			   {
				if (checkmax < v16p->olvalue)
				  {
					checkmax = v16p->olvalue;
					cmaxx = v16p->olxcoord;
					cmaxy =
					  mp->ydimen - 1 - v16p->olycoord;
				  }
			   }
		  }
		if (!checkmax)
		  {
			for (thisx = 0; thisx < mp->xdimen; thisx++)
			   for (thisy =0; thisy < mp->ydimen; thisy++,usp++)
			      {
				if (*usp > checkmax)
				  {
					if (mp->improps & IPR_RAXISCOMPRESS)
					  {
						if ((*usp) > 32767)
						   vv = 32 * ((*usp) - 32768);
						else vv = *usp;
					  	if (vv > checkmax)
						  {
							checkmax = vv;
							cmaxx = thisx;
							cmaxy = thisy;
						  }
					  }
					else {
						checkmax = *usp;
						cmaxx = thisx; cmaxy = thisy;
					     }
				  }
			      }
		   } /* end of "if (!checkmax" clause */
	     } /* end of `else' of "if (mp->data8)" clause */
	mp->zmax = checkmax; mp->zmaxx = cmaxx; mp->zmaxy = cmaxy;
}

int open_image(char *imname, int rflags)
{ /*	This opens the image `name' and attaches the file descriptor
	imwp->imw_fd to it.  It returns FRAME_OK if successful,
	FRAME_BADOPEN otherwise.  If it fails to read the image because
	the image is absent and the `rflags' argument satisfies
	(rflags | WAIT_IF_ABSENT) != 0, it waits around for a while and
	keeps trying till the frame appears.  The waiting will be aborted
	if a file with name, environment variable, or logical name NOWAIT
	exists. In this version we don't worry about whether the file is
	fully written out at this point;
	instead, we check for that in `read_header'. */
	int		nwaits;
	char		*tinam, *wname;
	IMAGEWORK	*imwp;

	tinam = trname(imname);	wname = trname("nowait");
	if (NULL == (imwp = iminfo(1))) return 0; /* get memory for imwp */
	if (-1 == (imwp->imw_fd = open(tinam, 0)))	/* frame absent */
	  {
		if (rflags & WAIT_IF_ABSENT)
		  {
		    for (nwaits = 0; (nwaits < MAX_WAITS) &&
			(-1 == imwp->imw_fd); nwaits++)
		     { /* abort waits smoothly if a file "NOWAIT" exists */
			if (0 == access(wname, 0)) return FRAME_BADOPEN;
			sleep(ABSENT_FRAME_WAIT);
			imwp->imw_fd = open((char *)tinam, 0);
		     } /* end of for loop through the waits */
		    if (-1 == imwp->imw_fd) return FRAME_TIMEDOUT;
		  } /* end of "if (rflags & WAIT_IF_ABSENT)" clause */
		else if (-1 == imwp->imw_fd) return FRAME_BADOPEN;
	  }
	return FRAME_OK;
}

static int iscompressed(char *nam, char *tenam)
{ /* This examines the file `nam'. If `nam' ends in .gz, .Z, or .z,
	it takes the name to be that of a compressed file.
	If the compressed version exists and the uncompressed version
	thereof doesn't, it uncompresses the file. If the uncompressed
	version exists and the compressed version doesn't, it smiles
	and returns. If the name doesn't end in one of those compression
	codes, it looks to see if the file can be opened. If it can't,
	it looks to see if a compressed version of the file is present.
	If that turns out to be true, it uncompresses it.
	The routine copies the output filename into outnam;
	 it assumes that at least strlen(nam) + 9 characters are available.
	 For errors (any code < 0 other than -5) it puts '\0' in *outnam.
	It returns the following codes:
	code value	meaning
	 0		file is okay, but the filename
			 doesn't correspond to any compressed format
	 1		name indicates gzip format: name ends in -gz or .gz,
			 and we have successfully uncompressed it here.
	 2		name indicates compress format: name ends in .Z,
			 and we have successfully uncompressed it here.
	 4		name indicates pack format: name ends in .z or -z,
			 and we have successfully uncompressed it here.
	 9		name indicates gzip format;
			 uncompressed version was already present.
	10		name indicates compress format;
			 uncompressed version was already present.
	12		name indicates pack format;
			 uncompressed version was already present.
	17		name didn't have compression gear in it,
			 but .gz version was present and was uncompressed.
	18		name didn't have compression gear in it,
			 but .Z version was present and was uncompressed.
	20		name didn't have compression gear in it,
			 but .z version was present and was uncompressed.
	-8		attempt to expand file failed
	-7		memory allocation error in this routine
	-6		filename is invalid
	-5		file cannot be opened, nor can any of its compressed
			 versions be opened
	-4		impossible to read first four bytes in file
	-2		filename ends in .Z but it isn't really a .Z file
	-1		filename ends in -gz or .gz but it isn't really
			 a gzipped file.
	For options 1 and 2, we require that the file begin with
	the appropriate magic number. If it doesn't, we return. */

	char		*cp, *icp;
	int		i, lolen, retv;

	 /* if the attempt to determine the compression type fails, quit */
	if (0 > (retv = compressiontype(nam, &lolen))) return retv;
	for (icp = nam, cp = tenam, i = 0; '\0' != *icp; cp++, icp++, i++)
		*cp = *icp;
	/* we have to null out all the extra space or bad things happen */
	for ( ; i < lolen + 9; i++, cp++) *cp = '\0';
	if (retv) /* we've identified the filetype as being compressed */
	  {	/* properly terminate the uncompressed version of the name */
		if (retv & 01) *(cp - 3) = '\0'; else *(cp - 2) = '\0';
		if (0 != access(nam, 4)) /* is it present? */
		  { /* no: see if the uncompressed version is */
			if (0 != access(tenam, 4)) retv = -5;
			else	return retv | 8;	/* need new one */
		  }
		/* if compressed version IS present, we do other things... */
	  }
	if (!(retv)) /* name suggests a regular file. Is it there? */
	  {
		if (0 != access(nam, 4))
		  {	/* no: look for other versions */
			cp = tenam + lolen;
			*cp++ = '.'; *cp++ = 'g'; *cp++ = 'z'; *cp = '\0';
			if (0 != access(tenam, 4))
			  {
				cp = tenam +lolen +1; *cp++ = 'Z'; *cp = '\0';
				if (0 != access(tenam, 4))
				  {
					cp = tenam +lolen+1; *cp = 'z';
					if (0 != access(tenam, 4))
					  { /* none of compressions worked */
						retv = -5;
						*(tenam + lolen) = '\0';
					  }
					else retv = 4 | 16;
				  }
				else retv = 2 | 16;
			  }
			else retv = 1 | 16;
		  }
		else return retv; /* successfully found ordinary file. */
	  }
	if (retv > 0)
	  { /* if we got here, we have a compressed file currently.
	 	We therefore need to (1) validate its compression format,
		 and (2) uncompress it. These operations are both
		 performed in the routine called here. */
		if ((retv) & 16) retv = expandit(tenam, retv);
		else retv = expandit(nam, retv);
	  }
	if (retv < 0) return retv;
	/* if we get here, we want to remember to return the proper filename.
	The name we return in tenam is always that of the uncompressed file */
	if (16 & retv) *(tenam + lolen) = '\0';
	else if (1 & retv) *(tenam + (lolen - 3)) = '\0';
	else if (6 & retv) *(tenam + (lolen - 2)) = '\0';
	return retv;
}

void nametonum(MAIL *mp)
{ /*	This determines the file-number associated with an image,
	based on its name. It then uses the image number and (if available)
	the frame index to determine the number of images in the run. */
	char		*cp, *cpi, *cpend;
	int		thisno;
	size_t		bulen, thislen;
	double		overw;
	DATARUN		*drp;
	FRAMEINDEX	*lfmi;
	FRAMEENTRY	*lft;
	static int	maxfm;

	/* attempt to guess the image number: first, find the end of the name*/
	cpi = mp->filename; for (cp = cpi; '\0' != *cp; cp++) ;
	for (cp = cp - 1; cp >= cpi; cp--) /* find last digit */
	   if (('0' <= *cp) && (*cp <= '9')) break;
	cpend = cp;
	if (cpend <= cpi) thisno = 0; /* no digits: give up */
	else {
		for ( ; cp >= cpi; cp--) /* guess at image number */
		   if (!(('0' <= *cp) && (*cp <= '9'))) break;
		/* ignore leading zeroes */
		for ( ; cp > cpend; cp++) if ('0' != *cp) break;
		if (cp <= cpend) thisno = atoi(cp);	else thisno = 0;
	     }
	drp = mp->runheader;	bulen = strlen(cpi);
	if (NULL != (lfmi = ftables(3)))
	  {
		for (lft = lfmi->fi_fmi; lft <= lfmi->fi_fme; lft++)
		   {
			thislen = strlen(lft->fmname);
			if (thislen > bulen)
				cp = lft->fmname + (thislen - bulen);
			else	cp = lft->fmname;
			if (!strcmp((char *)mp->filename, cp))
			  {
				thisno = lft->fmno;	break;
			  }
		   }
		maxfm = drp->run_nframes = 1 + (lfmi->fi_fme - lfmi->fi_fmi);
	  }
	else {
		if (maxfm < thisno) maxfm = thisno; drp->run_nframes = maxfm;
	     }
	overw = ((double)(drp->run_nframes)) * fabs(drp->run_stepreal);
	drp->run_nshifts = overw / 5.;
	if (0 >= mp->frameno) mp->frameno = thisno;
}

int read_image(char *name, int flags, MAIL *mailp)
{ /*	This reads an image from the file `name'.  It returns results in a
	pointer to a MAIL structure, inside which are several pointers,
	including a pointer to the list of overflows (mailp->overflow),
	a pointer to the data themselves (mailp->data8 or mailp->data16),
	and a pointer to a DATARUN-header structure (mailp->runheader).
	If we need to reserve memory for any of these, we do so in read_header.
	If there are fewer than KEEP_16_MIN overflows in the data,
	the data are stored as an 8-bit image; otherwise they're stored as
	a 16-bit image.  In the special case where we're going to subtract
	a dark-current image, we stay in 16-bits till after the subtraction.
	if (flags & ANNOUNCE) is true, we print a message to stdout announcing
	the reading-in of the frame. Other flags may be defined later. */
	int		compres, i, imlen, infotype;
	u_int		lfty;
	char		*icp, *tlnam;
	double		rcount;
	IMAGEWORK	*imwp;
	static char	*shortname, *lonam;
	static size_t	lenlo, oldlenlo;
	static char	lomsg[] = "lonam in read_image";

	if (NULL == name)
	  {
		memzap((AVOID **)&lonam, lomsg); oldlenlo = 0; return FRAME_OK;
	  }
	/* if the user has asked to subtract a dark-current image, we look to
	 see if the dark-current image is already in place.  If it is, we don't
	 do anything; if it isn't, we read it in and store it. */
	infotype = (flags & DARK_CURRENT) ? 025 : 05;
	if (NULL == (imwp = iminfo(infotype))) return 0;
	/* if the specified name is that of a compressed file,
	 we go into action a little. */
	lenlo = strlen(name) + 9;	if (lenlo < 128) lenlo = 128;
	if ((NULL == lonam) || (lenlo > oldlenlo))
	  {
		memzap((AVOID **)&lonam, lomsg);	oldlenlo = lenlo;
		if (NULL == (lonam = (char *)mallop(lenlo, lomsg)))
			return FRAME_NO_MEMORY;
	  }
	if ((0 > (compres = iscompressed(name, lonam))) && (-5 != compres))
		return FRAME_BADOPEN;
	/* Try to open the file; wait for it if the user wants us to */
	if (FRAME_OK != (i = open_image(lonam, flags))) return i;
	/* Read the header and position the file at the end of the header */
	if ((FRAME_BADOPEN == (lfty = read_header(mailp, imwp->imw_fd, flags)))
	 || (FRAME_NO_MEMORY == lfty) || (FRAME_BADMAIL == lfty) ||
		(FRAME_FMT_UNKNOWN == lfty)) return lfty; /* read in header */
	imwp->imw_filetype = lfty;
	/* modify mailp->improps according to environment variables */
	if (NULL != getenv("MIRROR_X"))
	  {
		if (mailp->improps & IPR_STARTBOTTOM)
			mailp->improps &= ~IPR_STARTBOTTOM;
		else	mailp->improps |= IPR_STARTBOTTOM;
		mailp->improps |= IPR_MIRRORX;
	  }
	if (NULL != getenv("MIRROR_Y"))
	  {
		if (mailp->improps & IPR_STARTRIGHT)
			mailp->improps &= ~IPR_STARTRIGHT;
		else	mailp->improps |= IPR_STARTRIGHT;
		mailp->improps |= IPR_MIRRORY;
	  }
	if (NULL != getenv("INVERT_SWAPPING")) mailp->improps |= IPR_INVERTSWAP;
	if ('\0' == mailp->filename[0]) /* give the file a name! */
	  {
		tlnam = trname(lonam);
		if (flags & JUST_HEADPFLOWS) /* get shortened name */
		  {
			imlen = strlen((char *)tlnam);
			if (imlen <= 14)
			  strncpy(mailp->filename, (char *)tlnam, imlen);
			else {
				icp = ((char *)tlnam) + (imlen - 14);
				strncpy(mailp->filename, (char *)icp, 14);
			     }
			mailp->filename[15] = '\0';
		  }
		else {
			shortname = filepart((char *)tlnam, 0);
			imlen = strlen(shortname);
			if (imlen <= 14) strcpy(mailp->filename, shortname);
			else {
				icp = ((char *)shortname) + (imlen - 14);
				strncpy(mailp->filename, (char *)icp, 14);
			     }
		     }
	  }
	nametonum(mailp); /* intuit frame number, etc. from the name */
	/* if we're reading a VMS file and the user wants it converted to
	 VMS format, we're already done, so go away quietly */
	if ((FMT_VMS == imwp->imw_filetype) && (flags & RETURN_IF_CONVERTED))
		return FRAME_ALREADY_CONVERTED;
	/* if we're just reading the header, quit here */
	if (flags & (JUST_HEADER | JUST_HEADPFLOWS))
	  { /* certain data: need to put a nominal eventcount in here */
		if (mailp->improps & IPR_NEEDCOUNT)
			mailp->eventcount = 1000000L;
		if (flags & LEAVE_OPEN) return FRAME_OK;
		return (-1 == close(imwp->imw_fd) ? FRAME_BADCLOSE : FRAME_OK);
	  }
	/* read the overflow tables, if any, then reposition pointer */
	if (FRAME_OK != (i = readflows(mailp)))
	  {
		fprintf(stderr, "Failure to read overflows\n"); return i;
	  }
	/* If the data are in 8-bit format, read 8-bit data and,
	 if necessary, convert to 16-bit. */
	if (mailp->improps & IPR_IS8BIT)	/* handle apparent 8-bit data */
	  {
		if (FRAME_OK != (i = read_8bit(mailp, flags))) return i;
		if ((mailp->oflowcount >= KEEP_16_MIN) && (FRAME_OK !=
		 (i = cvt_8_to_16(mailp, imwp->imw_filetype)))) return i;
	  }
	/* If they're in 16-bit format, read 'em in now in that way. */
	else {
		if (FRAME_OK != read_16bit(mailp, flags))
		  {
			fprintf(stderr, "failure at overflow stage\n");
			return FRAME_BAD16BIT;
		  }
	     }
	/* subtract out the dark-current image if needed */
	if ((flags & DARK_CURRENT) && (NULL != (imwp->imw_darkp)) &&
	 (FRAME_OK != (i = subtract_dark(mailp, (int)(flags >> 16))))) return i;
	verify_flowmax(mailp); /* verify overflows and zmax, i.e. largest val */
	rcount = mailp->eventcount;	if (rcount < 0.) rcount *= -64.;
	mailp->quadrant = record_inode(name);
	if (flags & ANNOUNCE) printf(
	 "file %-16s: %11.0f counts, %4d overflows; zmax = %5d at [%3d,%3d]\n",
		mailp->filename, rcount,
		(int)mailp->oflowcount, (int)mailp->zmax, (int)mailp->zmaxx,
		(int)(mailp->ydimen - 1 - mailp->zmaxy));
	if (-1 == close(imwp->imw_fd)) return FRAME_BADCLOSE;
	/* for now, we compress the frame at this point rather than deleting */
	if ((flags & GZIP_AFTER_CLOSE) && (0 != shrinkit(lonam)))
		return FRAME_NOT_COMPRESSED;
	return FRAME_OK;
}

/* end of image.c */
