/* imageout.c -- routines for outputting images to disk from XGEN

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	12 Jun 2007: byteswapping and image direction for SMV files remanaged
	27 May 2007: mksmvhead written and called to handle SMV headers
	28 Apr 2007: mkraxishead: do matrix inversion more carefully
	19 Apr 2007: float->double conversions; reorganize use of "{"
	 5 Dec 2006: spacegname: add comma after "P4/ncc" so lists work right;
		correct designation for spacegroup 126.
	11 Aug 2006: handle byteswapped formats without specific defines;
		manage all Mar image plates as having the same format
	 2 Aug 2006: add 2300-pixel Mar IPs to potential formats
	24 Jun 2006: include xgimage.h; the structures are there
	11 Jun 2006: move gotspace here from xgenfile so it can be static
	 1 Mar 2006: ismarip() identifies Mar image plates;
		better populating of Mar IP file header
	31 Dec 2005: provide facility for killing memory in mkmarhead
	 5 Dec 2005: replace static globals with calls to iminfo
	27 Nov 2005: wrt_bodydata: PAU unions make pointers work better
	24 Nov 2005: replace IMOUTWORK with IMAGEWORK structure type
	20 Nov 2005: inline argument declarations and strict prototyping
	24 Oct 2005: squeeze static globals into IMOUTWORK structure
	11 May 2005: move flipraxis() routine to ../nogui/fliprot.c
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	29 Jan 2005: identify rhombohedrals by the first letter of their name
	28 Aug 2004: routines used only here are declared as static; cosmetics
	24 Aug 2004: omit raxreport() because nobody was calling it
	 8 Aug 2004: add intltolc() to return lower-case spacegroup name
	24 Jan 2004: nbytes,fpi made local to individual routines
	21 Jan 2004: license message inserted
	 5 Jan 2004: replace globals isys, ispgp, fpi with local variables.
	10 Dec 2003: make outbuf a static global variable within this file;
		create `retoutput' routine to memzap(&outbuf,...)
	19 Sep 2003: initialize lmx, lmy to zero in getcounts();
		initialize yli to 0 in wrt_bodydata()
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	17 Aug 2003: trivial renaming of local variable to `finame'
	12 Jun 2003: pointless comparison deleted in R-Axis output 
		for values greater than 32767
	 5 Jan 2003: reorganized source-file header
	15 Oct 2002: routine spacegname added; called from "spaceg"
	25 Mar 2002: removed MSI attachments
	 6 Jul 1995: typecast in `imwrite' call to make irix64 happier
	 2 Mar 1995: delete unused code
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version

   Purpose: reformatting and writing out X-GEN detector images
	in various formats.
   Cross references: image.c

*************************************************************************/
#ifdef VMS
#define	NEED_ERRNO	1
#endif
#define		NEED_CTYPE	1
#include	"xgimage.h"
#define		INBLOCK		512
#define		DELETE_ORIGINAL	0x200	/* delete original file */
#define		PAD_TO_2TO_N	0x4000	/* pad each row so it's a power of 2 */
#define		OVERRIDE_COORDS	0x800	/* override coordinate conventions */
#define		MAX_SPACEG	230	/* # of International-Tables spgrps */

int ismarip(int informat)
{ /* Returns 1 if informat is one of the MAR formats;
	it returns -1 if it's one of the byte-and-word-swapped MAR formats;
	it returns 0 otherwise */
	int	high, i, inswap, low;
	static int marsize[] = {	1200, 1600, 1800, 2000, 2300,
					3000, 3450, 0 } ;

	low = (informat >> 16) & 0xff;	high = (informat >> 24) & 0xff;
	inswap = (low << 8) | high;
	for (i = 0; marsize[i] > 0; i++)
	   {
		if (informat == marsize[i]) return 1;
		if (inswap == marsize[i]) return -1;
	   }
	return 0;
}

size_t imwrite(AVOID *base, unsigned nb)
{	/* This writes `nb' bytes to file descriptor `imwp->imw_fd'.
	Under VMS it's aware of the block boundary problem and keeps trying
	to write until it outputs everything it's looking for.  */
	IMAGEWORK	*imwp;
#ifdef VMS
	char		*cp;
	int		end, nextend, nwrite, totwritten, towri;

	errno = 0;	/* clear errors before starting */
	if (NULL == (imwp = iminfo(5))) return 0;
	for (totwritten = 0, towri = nb, cp =(char *)base;totwritten <nb;)
	   {
		nwrite = write(imwp->imw_fd, (void *)cp, (unsigned)towri);
		if (nwrite <= 0) return totwritten; /*give up on EOF or error*/
		totwritten += nwrite;
		if (errno) return totwritten; /* give up on error */
		towri -= nwrite;	cp += nwrite;
	   }
	return totwritten;
#else
	if (NULL == (imwp = iminfo(5))) return 0;
	return	write(imwp->imw_fd, base, nb);
#endif
}

static void cpybpad(char *strout, char *strin, int end)
{ /*	This copies strin to strout and blank-pads strout to its end */
	char	*cpi, *cpo;
	int	ct;

	if (NULL == strout) return;
	if ((NULL == strin) || ('\0' == *strin))
	  {
		for (cpo = strout, ct = 0; ct < end; ct++, cpo++) *cpo = ' ';
	  }
	else
	 {
	   for (cpi = strin, cpo = strout, ct = 0; (ct < end) && ('\0' != *cpi);
	    cpi++, cpo++, ct++)
	      {
		if (!isprint((int)(*cpi))) break;
		*cpo = *cpi;
	      }
	   for ( ; ct < end; ct++, cpo++) *cpo = ' ';
	 }
}

char *spacegname(short thisispgp)
{ /*	This returns a string containing the character representation of
	the spacegroup numbered with `thisispgp'. If `thisispgp' is > 230,
	we adjust by 100 to allow for the hexagonal indexings of R3 and R32. */

	static char	*xspgnam[] = {
 "unknown", "P1", "P-1",					/*  0 -   2 */
 "P2", "P21", "C2",						/*  3 -   5 */
 "Pm", "Pc", "Cm", "Cc",					/*  6 -   9 */
 "P2/m", "P21/m", "C2/m", "P2/c", "P21/c", "C2/c",		/* 10 -  15 */
 "P222", "P2221", "P21212", "P212121",				/* 16 -  19 */
 "C2221", "C222", "F222", "I222", "I212121",			/* 20 -  24 */
 "Pmm2", "Pmc21", "Pcc2", "Pma2", "Pca21",			/* 23 -  29 */
 "Pnc2", "Pmn21", "Pba2", "Pna21", "Pnn2", "Cmm2", "Cmc21",	/* 30 -  36 */
 "Ccc2", "Amm2", "Abm2", "Ama2", "Aba2", "Fmm2", "Fdd2",	/* 37 -  43 */
 "Imm2", "Iba2", "Ima2", "Pmmm", "Pnnn", "Pccm", "Pban",	/* 44 -  50 */
 "Pmma", "Pnna", "Pmna", "Pcca", "Pbam", "Pccn", "Pbcm",	/* 51 -  57 */
 "Pnnm", "Pmmn", "Pbcn", "Pbca", "Pnma", "Cmcm", "Cmca",	/* 58 -  64 */
 "Cmmm", "Cccm", "Cmma", "Ccca", "Fmmm", "Fddd", "Immm",	/* 65 -  71 */
 "Ibam", "Ibca", "Imma", 					/* 72 -  74 */
 "P4", "P41", "P42", "P43", "I4", "I41",			/* 75 -  80 */
 "P-4", "I-4", "P4/m", "P42/m", "P4/n", "P42/n",		/* 81 -  86 */
 "I4/m", "I41/a",						/* 87 -  88 */
 "P422", "P4212", "P4122", "P41212", "P4222", "P42212",		/* 89 -  94 */
 "P4322", "P43212", "I422", "I4122",				/* 95 -  98 */
 "P4mm", "P4bm", "P42cm", "P42nm", "P4cc", "P4nc", "P42mc",	/* 99 - 105 */
 "P42bc", "I4mm", "I4cm", "I41md", "I41cd", "P-42m", "P-42c",	/*106 - 112 */
 "P-421m", "P-421c", "P-4m2", "P-4c2", "P-4b2", "P-4n2",	/*113 - 118 */
 "I-4m2", "I-4c2", "I-42m", "I-42d", "P4/mmm", "P4/mcc",	/*119 - 124 */
 "P4/nbm", "P4/nnc", "P4/mbm", "P4/mnc", "P4/nmm", "P4/ncc",	/*125 - 130 */
 "P42/mmc", "P42/mcm", "P42/nbc", "P42/nnm", "P42/mbc",		/*131 - 135 */
 "P42/mnm", "P42/nmc", "P42/ncm", "I4/mmm", "I4/mcm",		/*136 - 140 */
 "I41/amd", "I41/acd",						/*141 - 142 */
 "P3", "P31", "P32", "R3",					/*143 - 146 */
 "P-3", "R-3",							/*147 - 148 */
 "P312", "P321", "P3112", "P3121", "P3212", "P3221", "R32",	/*149 - 155 */
 "P3m1", "P31m", "P3c1", "P31c", "R3m", "R3c", "P-31m",		/*156 - 162 */
 "P-31c", "P-3m1", "P-3c1", "R-3m", "R-3c",			/*163 - 167 */
 "P6", "P61", "P65", "P62", "P64", "P63",			/*168 - 173 */
 "P-6", "P6/m", "P63/m",					/*174 - 176 */
 "P622", "P6122", "P6522", "P6222", "P6422", "P6322",		/*177 - 182 */
 "P6mm", "P6cc", "P63cm", "P63mc", "P-6m2", "P-6c2", "P-62m",	/*183 - 189 */
 "P-62c", "P6/mmm", "P6/mcc", "P63/mcm", "P63/mmc",		/*190 - 194 */
 "P23", "F23", "I23", "P213", "I213",				/*195 - 199 */
 "Pm-3", "Pn-3", "Fm-3", "Fd-3", "Im-3", "Pa-3", "Ia-3",	/*200 - 206 */
 "P432", "P4232", "F432", "F4132",				/*207 - 210 */
 "I432", "P4332", "P4132", "I4132",				/*211 - 214 */
 "P-43m", "F-43m", "I-43m", "P-43n", "F-43c", "I-43d",		/*215 - 220 */
 "Pm-3m", "Pn-3n", "Pm-3n", "Pn-3m", "Fm-3m", "Fm-3c",		/*221 - 226 */
 "Fd-3m", "Fd-3c", "Im-3m", "Ia-3d"				/*227 - 230 */
 };
	if (thisispgp > 230) thisispgp -= 100;
	if ((1 <= thisispgp) && (thisispgp <= 230))
		return xspgnam[(int)thisispgp];
	return xspgnam[0]; /* error code: return string `unknown' */
}

char *spacepname(short thisispgp)
{ /*	This returns a string containing the character representation of
	the spacegroup numbered with `thisispgp'. If `thisispgp' is > 230,
	we adjust by 100 to allow for the hexagonal indexings of R3 and R32.
	This differs from spacegname in that we put in the parentheses. */

	static char	*xsppnam[] = {
 "unknown", "P1", "P-1",					/*  0 -   2 */
 "P2", "P2(1)", "C2",						/*  3 -   5 */
 "Pm", "Pc", "Cm", "Cc",					/*  6 -   9 */
 "P2/m", "P2(1)/m", "C2/m", "P2/c", "P2(1)/c", "C2/c",		/* 10 -  15 */
 "P222", "P222(1)", "P2(1)2(1)2", "P2(1)2(1)2(1)",		/* 16 -  19 */
 "C222(1)", "C222", "F222", "I222", "I2(1)2(1)2(1)",		/* 20 -  24 */
 "Pmm2", "Pmc2(1)", "Pcc2", "Pma2", "Pca2(1)",			/* 23 -  29 */
 "Pnc2", "Pmn2(1)", "Pba2", "Pna2(1)", "Pnn2", "Cmm2", "Cmc2(1)",/* 30 -  36*/
 "Ccc2", "Amm2", "Abm2", "Ama2", "Aba2", "Fmm2", "Fdd2",	/* 37 -  43 */
 "Imm2", "Iba2", "Ima2", "Pmmm", "Pnnn", "Pccm", "Pban",	/* 44 -  50 */
 "Pmma", "Pnna", "Pmna", "Pcca", "Pbam", "Pccn", "Pbcm",	/* 51 -  57 */
 "Pnnm", "Pmmn", "Pbcn", "Pbca", "Pnma", "Cmcm", "Cmca",	/* 58 -  64 */
 "Cmmm", "Cccm", "Cmma", "Ccca", "Fmmm", "Fddd", "Immm",	/* 65 -  71 */
 "Ibam", "Ibca", "Imma", 					/* 72 -  74 */
 "P4", "P4(1)", "P4(2)", "P4(3)", "I4", "I4(1)",		/* 75 -  80 */
 "P-4", "I-4", "P4/m", "P4(2)/m", "P4/n", "P4(2)/n",		/* 81 -  86 */
 "I4/m", "I41/a",						/* 87 -  88 */
 "P422", "P42(1)2", "P4(1)22",					/* 89 -  91 */
 "P4(1)2(1)2", "P4(2)22", "P4(2)2(1)2",				/* 92 -  94 */
 "P4(3)22", "P4(3)2(1)2", "I422", "I4(1)22",			/* 95 -  98 */
 "P4mm", "P4bm", "P42cm", "P42nm", "P4cc", "P4nc", "P42mc",	/* 99 - 105 */
 "P42bc", "I4mm", "I4cm", "I41md", "I41cd", "P-42m", "P-42c",	/*106 - 112 */
 "P-42(1)m", "P-42(1)c", "P-4m2", "P-4c2", "P-4b2", "P-4n2",	/*113 - 118 */
 "I-4m2", "I-4c2", "I-42m", "I-42d", "P4/mmm", "P4/mcc",	/*119 - 124 */
 "P4/nbm", "P4/ncc", "P4/mbm", "P4/mnc", "P4/nmm", "P4/ncc",	/*125 - 130 */
 "P4(2)/mmc", "P4(2)/mcm", "P4(2)/nbc", "P4(2)/nnm", "P4(2)/mbc",/*131 - 135 */
 "P4(2)/mnm", "P4(2)/nmc", "P4(2)/ncm", "I4/mmm", "I4/mcm",	/*136 - 140 */
 "I4(1)/amd", "I4(1)/acd",					/*141 - 142 */
 "P3", "P3(1)", "P3(2)", "R3",					/*143 - 146 */
 "P-3", "R-3",							/*147 - 148 */
 "P312", "P321", "P3(1)12", "P3(1)21",				/*149 - 152 */
 "P3(2)12", "P3(2)21", "R32",					/*153 - 155 */
 "P3m1", "P31m", "P3c1", "P31c", "R3m", "R3c", "P-31m",		/*156 - 162 */
 "P-31c", "P-3m1", "P-3c1", "R-3m", "R-3c",			/*163 - 167 */
 "P6", "P6(1)", "P6(5)", "P6(2)", "P6(4)", "P6(3)",		/*168 - 173 */
 "P-6", "P6/m", "P6(3)/m",					/*174 - 176 */
 "P622", "P6(1)22", "P6(5)22", "P6(2)22", "P6(4)22", "P6(3)22",	/*177 - 182 */
 "P6mm", "P6cc", "P6(3)cm", "P6(3)mc", "P-6m2", "P-6c2", "P-62m",/*183 - 189*/
 "P-62c", "P6/mmm", "P6/mcc", "P6(3)/mcm", "P6(3)/mmc",		/*190 - 194 */
 "P23", "F23", "I23", "P2(1)3", "I2(1)3",			/*195 - 199 */
 "Pm-3", "Pn-3", "Fm-3", "Fd-3", "Im-3", "Pa-3", "Ia-3",	/*200 - 206 */
 "P432", "P4(2)32", "F432", "F4(1)32",				/*207 - 210 */
 "I432", "P4(3)32", "P4(1)32", "I4(1)32",			/*211 - 214 */
 "P-43m", "F-43m", "I-43m", "P-43n", "F-43c", "I-43d",		/*215 - 220 */
 "Pm-3m", "Pn-3n", "Pm-3n", "Pn-3m", "Fm-3m", "Fm-3c",		/*221 - 226 */
 "Fd-3m", "Fd-3c", "Im-3m", "Ia-3d"				/*227 - 230 */
 };
	if (thisispgp > 230) thisispgp -= 100;
	if ((1 <= thisispgp) && (thisispgp <= 230))
		return xsppnam[(int)thisispgp];
	return xsppnam[0]; /* error code: return string `unknown' */
}

char *intltolc(short thisispgp)
{ /*	This returns the lower-case version of the above */
	char		*cp;
	static char	tempor[20];

	zeromem((void *)tempor, 20);
	strcpy(tempor, spacegname(thisispgp));
	for (cp = tempor; '\0' != *cp; cp++)
		if (('A' <= *cp) && (*cp <= 'Z')) *cp += 'a' - 'A';
	return tempor;
}

static int gotspace(int *fsysp, int *fspagp)
{ /*	This pokes around to try to discern the spacegroup and crystal
	system. It returns EOK if successful, EBADVALUE if not.
	The values it stuffs into *fsysp and *fspagp will be -1 if it
	fails. */
	int		ngot;
	static FILE	*fpte;
	static REFPARS	locrx;
	static CRYSTAL	locxhd;
	static int	locsys = 0;
	static int	locspa = 0;

	*fsysp = *fspagp = -1; /* error return values */
	/* if we've called this before the values may be available */
	if ((1 <= locsys) && (locsys <= 7))
	  {
		*fsysp = locsys; *fspagp = locspa;	return EOK;
	  }
	/* if not, we try opening the files that contain
	 those objects, namely the UPARAMS and MULTIREF files */
	if (NULL != (fpte = xgopen("uparams", "r")))
	  {
		if (232 <= (ngot = xgread((void *)&locrx, XG_REFPARS, fpte)))
		  {
			*fsysp = locsys = locrx.xisys;
			*fspagp = locspa = locrx.xispgp;
		  }
		(void)fclose(fpte);	fpte = NULL;
		if ((1 <= locsys) && (locsys <= 7)) return EOK;
	  }
	if (NULL != (fpte = xgopen("multiref", "r")))
	  {
		if (232 <= (ngot = xgread((void *)&locxhd, XG_CRYSTAL, fpte)))
		  {
			*fsysp = locsys = locxhd.xt_isys;
			*fspagp = locspa = locxhd.xt_intl;
		  }
		(void)fclose(fpte);	fpte = NULL;
		if ((1 <= locsys) && (locsys <= 7)) return EOK;
	  }
	return EBADVALUE;
}

static int mkraxishead(MAIL *mp, int ruflags)
{ /*	This converts the X-GEN image structure pointed to by mp
	to the format expected by an R-Axis detector. */
	int		i, inswap, j, k, locspag, locsys;
	time_t		tl0;
	size_t		ntowri;
	double		ci, trec[3][3], tare[3][3];
	DATARUN		*drp;
	RAXISHEAD	trax;
	static char	*shortname;
	char		*cp, *spname;
	static char	*xsysnam[] = {
	 "unknown", "triclinic", "monoclinic", "orthorhombic",
	 "tetragonal", "trigonal", "cubic", "rhombohedral"};

	/* zero out the header and put a few fixed items in */
	zeromem((AVOID *)&trax, sizeof (RAXISHEAD));
	cpybpad(trax.raxis0, "R-AXIS2", 10);
	cpybpad(trax.ver0, "1.0-2.0", 10);
	/* pick up the header elements that are in
	 the main body of the MAIL structure */
	trax.nxpix0 = mp->xdimen;
	trax.nzpix0 = mp->ydimen;
	(void)timecvt(mp->datetime, NULL, trax.date0, &tl0, 0, 5);
	trax.expos0 = 0.01 * (double)(mp->elap100ths);
	trax.nosc0 = mp->frameno;
	trax.ysize0 = 0.1e-2 * (double)(mp->cellsize);
	trax.zsize0 = 0.1e-2 * (double)(mp->cellsize);
	if (ruflags & PAD_TO_2TO_N)
	  {
		if (mp->xdimen < 1024) trax.lrecl0 = 2048;
		else trax.lrecl0 = 4096;
	  }
	trax.nrec0 = mp->ydimen;
	trax.nst0 = trax.ipno0 = 1;
	trax.hilo = 32.;	trax.fad1 = 62.;	trax.fad2 = 250.;
	trax.ncylinder = 0;	trax.weissenberg = 0.;	trax.emos0 = 1.;
	/* get the crystal ID from the filename */
	shortname = filepart(mp->filename, 2);
	if (shortname[0] != '\0') cpybpad(trax.cid0, shortname, 20);
	else cpybpad(trax.cid0, "unknown", 20);
	cpybpad(trax.monoc0, "none", 20);
	cpybpad(trax.coli0, "unknown", 20);
	cpybpad(trax.filt0, "unkn", 4);
	cpybpad(trax.focus, "unknown", 12);
	cpybpad(trax.xraymemo, "nothing", 80);
	cpybpad(trax.res1, "  ", 84);
	cpybpad(trax.res2, "  ", 56);
	cpybpad(trax.maxis0, "a ", 4);
	cpybpad(trax.baxis0, "b ", 4);
	cpybpad(trax.res3, "  ", 204);
	cpybpad(trax.cpu, "SGI?", 10);
	cpybpad(trax.ip, "R-AXIS2", 10);
	cpybpad(trax.res4, "  ", 180);
	trax.cell0[0] = trax.cell0[1] = trax.cell0[2] = 50.;
	trax.cell0[3] = trax.cell0[4] = trax.cell0[5] = 90.;
	/* get the elements found in the mp->runheader stuff */
	if (NULL != (drp = mp->runheader))
	  {
		if (drp->run_code[0] != '\0')
			cpybpad(trax.cid0, drp->run_code, 20);
		else	cpybpad(trax.cid0, "unknown", 20);
		if (drp->run_notes[0] != '\0')
			cpybpad(trax.memo0, drp->run_notes, 80);
		else	cpybpad(trax.memo0, "unknown", 80);
		trax.xkv0 = drp->run_volxray;
		trax.xma0 = drp->run_curxray;
		trax.fis0 = drp->run_startpos[2];
		if (fabs(drp->run_stepreal) > FUZZ)
			trax.fie0 = trax.fis0 + drp->run_stepreal;
		else	trax.fie0 = trax.fis0 +
				((double)(drp->run_stepsize)) / 96.;
		trax.duma0[0] = drp->run_startpos[0];
		trax.duma0[1] = drp->run_startpos[1];
		trax.duma0[2] = drp->run_d0ang[0];
		trax.expos0 = drp->run_expos;
		trax.dist0 = 10. * drp->run_d0pos[0];
		if (drp->run_d1pos[1] > FUZZ) trax.ysize0 = drp->run_d1pos[1];
		if (drp->run_d1pos[2] > FUZZ) trax.zsize0 = drp->run_d1pos[2];
		trax.beam0 = drp->run_d0pos[1] / trax.ysize0 +
			0.5 * (double)trax.nxpix0;
		trax.zbeam0 = drp->run_d0pos[2] / trax.zsize0 +
			0.5 * (double)trax.nzpix0;
		trax.twoth0 = 0; /* wrong with real monochromator */
		trax.wave0 = drp->run_lambda;
		/* if the reciprocal-space matrix is present,
		 use it to get the cell constants */
		for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
			trec[i][j] = drp->run_recip[i][j];
		if (0 == dinv3(trec, tare))
		  {
			for (i = 0; i < 3; i++)
			  {
				for (ci = j = 0; j < 3; j++)
				  {
					ci += tare[i][j] * tare[i][j];
				  }
				trax.cell0[i] = sqrt(ci);
			  }
			for (j = 0; j < 3; j++)
			   {
				k = j + 1;	if (k > 2) k = 0;
				for (ci = i = 0; i < 3; i++)
			   		ci += tare[j][i] * tare[k][i];
				trax.cell0[6 - j - k] =
				  acos(ci / (trax.cell0[j] * trax.cell0[k]));
			 }
		  }
	  }
	else {	/*... need something here */
		cpybpad(trax.cid0, "unknown", 20);
		cpybpad(trax.memo0, "nothing", 20);
		trax.xkv0 = 40;	trax.xma0 = 40;
		trax.expos0 = 10.;
		trax.dist0 = 100.;
		trax.ysize0 = 150.;	trax.zsize0 = 150.;
		trax.beam0 = trax.nxpix0 / 2.;
		trax.zbeam0 = trax.nzpix0 / 2.;
		trax.wave0 = 1.5418;
	     }
	/* get the crystal system from existing files if possible */
	if (EOK == gotspace(&locsys, &locspag))
		cpybpad(trax.csys0, xsysnam[locsys], 12);
	else cpybpad(trax.csys0, "unknown", 12);
	spname = spacegname(locspag);
	if (locsys == 6)
	  {
		if ((168 < locspag) && (locspag < 195))
			cpybpad(trax.csys0, "hexagonal", 12);
		else if ('R' == spname[0])
			cpybpad(trax.csys0, "rhombohedral", 12);
	  }
	if ((1 <= locspag) && (locspag <= MAX_SPACEG))
		cpybpad(trax.spg0, spname, 12);
	else cpybpad(trax.spg0, "unknown", 12);
	if (NULL != (cp = getenv("USER"))) cpybpad(trax.user0, cp, 20);
	if (fabs(trax.wave0 - 1.5418) < 0.02) cpybpad(trax.targ0, "Cu", 4);
	else if (fabs(trax.wave0 - 0.7093) < 0.02) cpybpad(trax.targ0, "Mo",4);
	else cpybpad(trax.targ0, "unkn", 4);
#ifdef LITTLEND /* swap words and bytes if needed */
	inswap = mp->format;
#else
	inswap = swawb(mp->format);
#endif
	if (inswap == FMT_RAXIS) raxis_swap(&trax, 1);
	ntowri = sizeof (RAXISHEAD);
	if (ntowri != imwrite((AVOID *)&trax, ntowri)) return FRAME_BADMAIL;
	/* R-axis files have additional header space full of zeroes.
	 It's either 4096-2048=2048 bytes if the images are padded,
	 or 3800-2048= 1752 bytes if not */
	if (!(ruflags & PAD_TO_2TO_N)) ntowri = 2 * mp->ydimen - ntowri;
	zeromem((AVOID *)&trax, ntowri);
	if (ntowri != imwrite((AVOID *)&trax,ntowri)) return FRAME_BADMAIL;
	return FRAME_OK;
}

static void squeezend(char *str, char *title, char *fomt, float inval,
	int sqflags)
{ /* This does a clever printing that looks something like
		EXPOSURE=5;
	for title=EXPOSURE, inval=5. In general it appends the
	printout to str, using a formatting defined by fomt,
	but with all white space eliminated between the equal sign
	and the beginning of the numbers. Because of the way this gets
	used, it omits the printout if inval is zero unless sqflags & 01
	is nonzero, in which case it will print it even if it's zero */
	int		eqhere;
	char		*cpi, *cpo, *sta;
	static char	sqbuf[256];
	static char	sqfmt[40];

	if (NULL == str) return;	/* make sure something is there */
		/* don't print out zeroes unless sqflags & 01 is nonzero */
	if ((!(sqflags & 01)) && fabs(inval) < FUZZ) return;
	zeromem((AVOID *)sqbuf, 256);
	strcpy(sqfmt, "%s=");	strcat(sqfmt, fomt); strcat(sqfmt, "\n");
	(void)sprintf(sqbuf, sqfmt, title, inval);
	if ('\0' == *sqbuf) return;
	if ('\0' == *str) sta = str;
	else for (sta = str; '\0' != *sta; sta++) ;
	for (eqhere = 0, cpo = sta, cpi = sqbuf; '\0' != *cpi; cpi++)
	   {
		if (*cpi == '=') eqhere = 1;
		if ((!(eqhere)) || (*cpi != ' ')) *cpo++ = *cpi;
	   }
}

static int mksmvhead(MAIL *mp)
{ /* This creates and outputs an SMV-style header, as used in ADSC
   images, old FAST images, and more recent Steve Naday-style detectors.
   We decide how much to write out based on the detector coordinate sizes. */

	char		*cp;
	int		i, serno;
	float		detmid, pixval;
	size_t		smsiz, smsofar;
	DATARUN		*runhdp;
	static char	smvbuf[4096];

	smsiz = 512;	/* overridden for APS-1 detectors in the future */
	if ((mp->xdimen == 2084) || (mp->xdimen == 4168))
	  {
		zeromem((AVOID *)smvbuf, smsiz);
		(void)strcpy(smvbuf,
			"{\nHEADER_BYTES=  512;\r\n\r\nDIM=2;\r\n");
		(void)sprintf(&(smvbuf[33]), "SIZE1=%d;\r\nSIZE2=%d;\r\n",
			mp->xdimen, mp->ydimen);
		(void)strcat(smvbuf,
		 "BYTE_ORDER=little_endian;\r\nTYPE=unsigned_short;\r\n");
		(void)strcat(smvbuf, "}\r\n\f");
	  }
	else if ((mp->xdimen == 2048) || (mp->xdimen == 4096))
	  { /* ADSC detectors */
		runhdp = mp->runheader;
		zeromem((AVOID *)smvbuf, smsiz);
		(void)strcpy(smvbuf, "{\nHEADER_BYTES=  512;\nDIM=2;\n");
		(void)strcat(smvbuf, "BYTE_ORDER=little_endian;\n");
		(void)strcat(smvbuf, "TYPE=unsigned_short;\n");
		pixval = (NULL != runhdp) ?
			runhdp->run_d1pos[1] : 0.001 * (float)(mp->cellsize);
		squeezend(smvbuf, "PIXEL_SIZE", "%6.4f", pixval, 0);
		strcat(smvbuf, "BIN=NONE;\nADC=fast;\n");
		serno = -1;
		if (NULL != runhdp)
		  {
			smsofar = strlen(smvbuf); serno = -1;
			for (cp = runhdp->run_notes; '\0' != *cp; cp++)
			   {
				if ((!strncmp(cp, "Detector SN:", 12)) ||
				 (!strncmp(cp, "DETECTOR SN:", 12)) ||
				 (!strncmp(cp, "DETECTOR_SN:", 12)))
				  {
					serno = atoi(cp + 12);
					squeezend(smvbuf, "DETECTOR_SN",
					 "%5.0f", (float)serno, 0); break;
				  }
			   }
		  }
		smsofar = strlen(smvbuf);
		timecvt(mp->datetime, NULL, &(smvbuf[smsofar]), NULL, 0, 9);
		smsofar = strlen(smvbuf);
		squeezend(smvbuf, "TIME", "%6.2f",
			0.01 * (float)(mp->elap100ths), 0);
		if (NULL != runhdp)
		  {
			squeezend(smvbuf, "DISTANCE", "%9.3f",
				10. * runhdp->run_d0pos[0], 0);
			squeezend(smvbuf, "OSC_RANGE", "%9.3f",
				-(runhdp->run_stepreal), 0);
			squeezend(smvbuf, "OMEGA", "%9.3f",
				runhdp->run_startpos[0], 1);
			squeezend(smvbuf, "OSC_START", "%9.3f",
				runhdp->run_startpos[0], 0);
			squeezend(smvbuf, "TWOTHETA", "%9.3f",
				-(runhdp->run_d0ang[0]), 1);
			sprintf(&(smvbuf[strlen(smvbuf)]), "AXIS=%s;\n",
				(runhdp->run_motor == 1) ? "omega" : "phi");
			squeezend(smvbuf, "WAVELENGTH", "%9.5f",
				runhdp->run_lambda, 0);
			detmid = 0.5 * (float)(mp->xdimen) *
				runhdp->run_d1pos[1];
			squeezend(smvbuf, "BEAM_CENTER_X", "%9.3f",
				detmid + runhdp->run_d0pos[2], 0);
			detmid = 0.5 * (float)(mp->ydimen) *
				runhdp->run_d1pos[2];
			squeezend(smvbuf, "BEAM_CENTER_Y", "%9.3f",
				detmid - runhdp->run_d0pos[1], 0);
		  }
		strcat(smvbuf, "UNIF_PED=1500;\n");
		sprintf(&(smvbuf[strlen(smvbuf)]),
		 "SIZE1=%d;\nSIZE2=%d;\n", mp->xdimen, mp->ydimen);
		strcat(smvbuf, "CCD_IMAGE_SATURATION=65535;\n");
		(void)strcat(smvbuf, "}\f\n");
	  }
	smsofar = strlen(smvbuf);
	for (i = smsofar, cp = &(smvbuf[smsofar]); i < smsiz; i++, cp++)
		*cp = ' ';
	if (smsiz != imwrite((AVOID *)smvbuf, smsiz)) return FRAME_BADMAIL;
	return FRAME_OK;
}

static MARHEAD *mkmarhead(MAIL *mp)
{	/* This converts and writes the puny little header
	on a Mar image plate frame. */
	time_t		tl0;
	double		step;
	DATARUN		*drp;
	static MARHEAD	*mhp;
	static char	mhpmsg[] = "mhp in mkmarhead";

	if (NULL == mp) /* special case: clear memory for this routine */
	  {
		if (NULL != mhp) memzap((AVOID **)&mhp, mhpmsg);
		return NULL;
	  }
	if (NULL == mhp)
	  {
		if (NULL == (mhp = (MARHEAD *)callop(1, sizeof (MARHEAD),
			"mhp in mkmarhead"))) return NULL;
	  }
	else 	zeromem((AVOID *)mhp, sizeof (MARHEAD));
	mhp->ma_recordlen = mhp->ma_width = mp->xdimen;
	mhp->ma_numrecords = mhp->ma_height = mp->ydimen;
	mhp->ma_ypcm = mhp->ma_xpcm = ((double)mp->cellsize) / 1000.;
	mhp->ma_xpcm2 = mhp->ma_xpcm;
	mhp->ma_ypcm2 = mhp->ma_ypcm;
	if (NULL != (drp = mp->runheader))
	  {
		mhp->ma_x0 = drp->run_d0pos[1] + 0.5 * ((double)mhp->ma_width);
		mhp->ma_y0 = drp->run_d0pos[2] + 0.5 * ((double)mhp->ma_width);
		mhp->ma_lambda = drp->run_lambda;
		mhp->ma_plated = drp->run_d0pos[0] * 10.;
		mhp->ma_omega = drp->run_d0ang[0];
		mhp->ma_xrayctsum = 100000;
		mhp->ma_maxradius = ((double)mp->xdimen) * 0.15;
		mhp->ma_minradius = 0.;
		if (fabs(drp->run_startpos[0]) > FUZZ)
			mhp->ma_phi0 = drp->run_startpos[0];
		else	mhp->ma_phi0 = drp->run_startpos[2];
		if (fabs(step = drp->run_stepreal) < FUZZ)
			step = ((double)drp->run_stepsize) / 96.;
		mhp->ma_phie = mhp->ma_phi0 + step;
	   }
	(void)timecvt(mp->datetime, NULL, mhp->ma_datetim, &tl0, 0, 3);
	if (NULL != mp->data16) /* simple case... */
		mhp->ma_n16flows = mp->oflowcount;
	else	mhp->ma_n16flows = mp->over64k;
	return mhp;
}

void getcounts(MAIL *mp, double *nctp, xglong maxctp[])
{ /*	This counts ADUs or photons in an image and returns
	the resulting value as a double */
	int	hit, npix, nflo, col, offs, row, rowct, wid;
	u_char	*ucp;
	u_short	*usp;
	xglong	lct, maxv, oval, lmx, lmy;
	VOFLOW	*vofp;
	double	fct;

	hit = mp->ydimen;	wid = mp->xdimen;	npix = wid * hit;
	lmx = lmy = 0;
	if (mp->data8) /* essentially 8-bit data */
	  { /* count the values in the overflow table */
		maxv = lct = 0;
		for (vofp = mp->overflow, nflo = mp->oflowcount;
		 nflo--; vofp++)
		   {
			oval = vofp->value;
			lct += (xglong)(oval - 255);
			if (oval > maxv)
			  {
				maxv = oval; lmx = vofp->xcoord;
				lmy = hit - 1 - vofp->ycoord;
			  }
		   }
		/* now count the values in the 8-bit array.
		 We also look for the max if needed */
		if (mp->oflowcount)
		  {
			for (ucp = mp->data8; npix--; ucp++)
				lct += (xglong)(*ucp);
		  }
		else {
			for (ucp = mp->data8; npix--; ucp++)
			   {
				lct += (xglong)(*ucp);
				if (*ucp > maxv)
				  {
					maxv = *ucp; offs = ucp - mp->data8;
					lmx = offs % wid;
					lmy = hit - 1 - offs / wid;
				  }
			   }
		     }
		*nctp = (double)lct;
		maxctp[0] = maxv;	maxctp[1] = lmx;
		maxctp[2] = lmy;
	  }
	else if (mp->data16) /* essentially 16-bit data */
	  { /* count values in the 16-bit overflow table */
		maxv = fct = lct = 0;
		for (vofp = mp->overflow, nflo = mp->oflowcount;
		 nflo--; vofp++)
		   {
			oval = 4 * vofp->value;
			fct += (double)oval;	oval += 65535;
			if (maxv < oval)
			  {
				maxv = oval; lmx = vofp->xcoord;
				lmy = hit - 1 - vofp->ycoord;
			  }
		   }
		/* count the values in the 16-bit array. To prevent 32-bit
		 overflow we count within individual rows and convert to
		 double-precision floating-point if and when necessary.
		 We look for the max within the 16-bit array if needed */
		if (mp->oflowcount)
		  {
			for (usp = mp->data16, row = 0; row < hit; row++)
			   {
				for (rowct = col = 0; col < wid; col++, usp++)
				   {
					rowct += (xglong)(*usp);
				   }
				lct += rowct;	fct += (double)rowct;
			   }
		  }
		else {
			for (usp = mp->data16, row = 0; row < hit; row++)
			  {
				for (rowct = col = 0; col < wid; col++, usp++)
				   {
					rowct += (xglong)(*usp);
					if (*usp > maxv)
					  {
						maxv = *usp;
						offs = usp - mp->data16;
						lmx = offs % wid;
						lmy = hit - 1 - offs / wid;
					  }
				   }
				fct += (double)rowct;	lct += rowct;
			   }
		  }
		*nctp = (fct < 2.1e9) ? (double)lct : fct;
		maxctp[0] = maxv; maxctp[1] = lmx; maxctp[2] = lmy;
	  }
	else {
		*nctp = 1.e8;	*maxctp = 10000; /* error condition */
	     }
}

static int wrt_header(MAIL *mp, xgulong type, int ruflags)
{	/* writes a header appropriate to image type `type' */
	int		i, swtype;
	size_t		ntowri;
#ifdef LITTLEND
	xgulong	*ulp, *ulpe;
#endif
	PMAIL		*lmp;
	MARHEAD		*mhp;
	IMAGEWORK	*imwp;

	mp->format = type;	swtype = swawb(type);
	if (NULL == (imwp = iminfo(5))) return FRAME_NO_MEMORY;
	if (type == FMT_VMS)
	  {
		cpymem((AVOID *)imwp->imw_buf, (AVOID *)mp, SIZMAIL);
		zeromem((AVOID *)(imwp->imw_buf + sizeof (MAIL)),
			(unsigned)(DEFBLOCK - SIZMAIL));
		if (DEFBLOCK != imwrite((AVOID *)imwp->imw_buf, DEFBLOCK))
			return	FRAME_BADMAIL;
	  }
	else if (ismarip(type))
	  {
		(void)iminfo(04);	/* destroy old output-buffer memory */
		if (NULL == (mhp = mkmarhead(mp))) return FRAME_NO_MEMORY;
		ntowri = 2 * mp->xdimen;
		if ((NULL == imwp->imw_buf) || (ntowri > imwp->imw_bsize))
		  {
			(void)iminfo(04); imwp->imw_bsize = ntowri;
			if (NULL == (imwp = iminfo(05))) return FRAME_BADMAIL;
		  }
		zeromem((void *)imwp->imw_buf, ntowri);
		cpymem(imwp->imw_buf, (void *)mhp, sizeof (MARHEAD));
		(void)mkmarhead(NULL);	/* get rid of internal memory */
		if (ntowri != imwrite((AVOID *)imwp->imw_buf, ntowri))
			return FRAME_BADMAIL;
	  }
	else if ((type == FMT_SAXII) || (type == FMT_SAXII_16) ||
	 (swtype == FMT_SAXII) || (swtype == FMT_SAXII_16))
	  {
		if (FRAME_OK != (i = mksaxihead(mp, type, imwp->imw_fd)))
			return i;
	  }
	else if ((type == FMT_RAXIS) || (swtype == FMT_RAXIS))
	  {
		if (FRAME_OK != (i = mkraxishead(mp, ruflags))) return i;
	  }
	else if ((type == FMT_FAST) || (swtype == FMT_FAST))
	  {
		if (FRAME_OK != (i = mksmvhead(mp))) return i;
	  }
	else
	  {
		lmp = (PMAIL *)callop(1, sizeof (PMAIL), "lmp in wrt_headr");
		lmp->pformat = FMT_PCS_8;
		lmp->pdimen = mp->xdimen;
		lmp->pzoom = mp->ydimen;
		lmp->pquadrant = mp->quadrant;
		lmp->peventcount = mp->eventcount;
		lmp->p8bitoflow = mp->oflowcount;
		lmp->pzmax = mp->zmax;
		lmp->pzmaxx = (xgulong)mp->zmaxx;
		lmp->pzmaxy = (xgulong)mp->zmaxy;
		lmp->pzdisp = mp->zdisp;
		lmp->pontime = mp->ontime;
		lmp->plate = mp->late;
		lmp->pdatetime[0] = mp->datetime[0];
		lmp->pdatetime[1] = mp->datetime[1];
		lmp->paccumsecs = mp->accumsecs;
		lmp->pelap100ths = mp->elap100ths;
		lmp->pcellsize = mp->cellsize;
		lmp->pnonzpix = mp->nonzpix;
		lmp->p16bitoflow = mp->over64k;
		lmp->pframeno = mp->frameno;
		for (i = 15; i--; ) lmp->pfilename[i] = mp->filename[i];
		lmp->pfilename[15] = '\0';	/* force null-termination */
#ifdef LITTLEND /* little end sys: before writing we byte-and-word swap */
		for (ulp = &(lmp->pformat), ulpe = (xgulong *)&(lmp->pframeno);
		 ulp <= ulpe; ulp++) *ulp = swawb(*ulp);
#endif
	/* Now that the conversion is done, we write the bugger out */
		if (sizeof (PMAIL) != imwrite((AVOID *)lmp, sizeof (PMAIL)))
			return FRAME_BADMAIL;
	  }
	return FRAME_OK;
}

/*ARGSUSED*/
static void vflow_to_oflow(OFLOW *ofp)
{ /*	This converts in place from VOFLOW format to byteswapped OFLOW format*/
	xgulong	offset;

	offset = ((xgulong)(ofp->loword) << 9) | (xgulong)(ofp->hiword);
#ifdef LITTLEND
	ofp->hiword = uswab((u_short)(offset >> 16));
	ofp->loword = uswab((u_short)(offset & 0xFFFF));
	ofp->value = uswab(ofp->value);
#else
	ofp->hiword = (u_short)(offset >> 16);
	ofp->loword = (u_short)(offset & 0xFFFF);
#endif
}

static int mkmarflows(int type)
{ /*	This writes out Mar-style overflows. Since these go at the
	end of the file, we do this after the 16-bit data.
	To save space and adhere to the concept of a `VOFLOW' structure,
	the recorded values are actually kept as (true value - 65535) / 4;
	so if the true value is 71384, the stored value will be 1462.
	My impression is that the largest possible value the Mar can read is
	3FFFF = 262143, so that will read out as (262143-65535)/4 = 49152. */
	int		martype, nflows, offs, yen, fno, truev;
	size_t		towri;
	off_t		marblock, start;
	u_char		*cp;
	MAIL		*lhd;
	VOFLOW		*vofp;
	MARFLOW		mflo;
	IMAGEWORK	*imwp;

	if (NULL == (imwp = iminfo(05))) return EALLOC;
	lhd = imwp->imw_lhd;	nflows = lhd->oflowcount;
	if (nflows < 1) return EOK; /* see if we're needed */
	/* convert the internal representation of the overflows to
	 the file representation. Make sure we have enough memory for
	 this operation before beginning it */
	yen = lhd->ydimen - 1;
	if (1 == (martype = ismarip(type)))
	  {
		fno = lhd->oflowcount * 8;
		if (fno > (int)(imwp->imw_bsize)) /* get enough memory for it */
		  {
			(void)iminfo(04);	imwp->imw_bsize = fno;
			if (NULL == (imwp = iminfo(05))) return FRAME_BADFLOWS;
		  }
		for (cp = imwp->imw_buf, fno = 0, vofp = lhd->overflow;
		 fno < nflows; fno++, cp += 8, vofp++)
		   {
			offs = lhd->xdimen * (yen - vofp->ycoord) +
				(vofp->xcoord + 1);
			truev = 65535 + 4 * vofp->value;
			mflo.mxyoff = offs;
			mflo.minten = truev;
			cpymem((AVOID *)cp, (AVOID *)&mflo, sizeof (MARFLOW));
		   }
	  }
	else if (-1 == martype)
	  {
		for (cp = imwp->imw_buf, fno = 0, vofp = lhd->overflow;
		 fno < nflows; fno++, cp += 8, vofp++)
		   {
			offs = lhd->xdimen * (yen - vofp->ycoord) +
				(vofp->xcoord + 1);
			truev = 65535 + 4 * vofp->value;
			mflo.mxyoff = swawb(offs);
			mflo.minten = swawb(truev);
			cpymem((AVOID *)cp, (AVOID *)&mflo, sizeof (MARFLOW));
		   }
	  }
	/* figure out where to write and how much to write */
	marblock = lhd->ydimen * sizeof (u_short);
	start = (lhd->xdimen + 1) * marblock;
	towri = nflows * sizeof (MARFLOW);
	/* move to the right place and do the writing. Make sure
	 we have enough space to do it in. */
	if (-1 == lseek(imwp->imw_fd, start, 0)) return FRAME_BADFLOWS;
	if (towri > imwp->imw_bsize)
	  {
		(void)iminfo(04);	imwp->imw_bsize = towri;
		if (NULL == (imwp = iminfo(05))) return FRAME_BADFLOWS;
	  }
	if (towri != imwrite((AVOID *)imwp->imw_buf, (int)towri))
		return FRAME_BADFLOWS;
	return FRAME_OK;
}

static int wrt_flows(MAIL *mp, xgulong type)
{ /*	This is responsible for writing out the overflows in an image. */
	int		i, npts, swtype;
	u_char		*cp, *cvstart, *cvend;
	u_short		*usp;
	VOFLOW		*vofp, *vofe;
	IMAGEWORK	*imwp;
	static xgulong	nbytes;

	/* MAR and SAXII flows are written out after the 16-bit data */
	swtype = swawb(type);
	if (ismarip(type) || (type == FMT_SAXII) || (type == FMT_SAXII_16) ||
	 (swtype == FMT_SAXII) || (swtype == FMT_SAXII_16)) return FRAME_OK;
	if ((type == FMT_RAXIS) || (type == FMT_RAXIS6) ||
	 (swtype == FMT_RAXIS) || (swtype == FMT_RAXIS6))
	  {
		/* R-Axis: points > 32K are left-shifted 3 bits.
		 Points > 262144 are stuck at 65535. */
		npts = mp->xdimen * mp->ydimen;
		for (usp = mp->data16, i = 0; i < npts; i++, usp++)
			if (*usp > 32767) *usp = 32768 | (*usp << 3);
		return FRAME_OK;
	  }
	cvstart = (u_char *)(mp->overflow);
	cvend = cvstart + mp->oflowcount * sizeof (VOFLOW);
	if (NULL == (imwp = iminfo(5))) return FRAME_NO_MEMORY;
	if (type == FMT_VMS)
	  {
		zeromem((AVOID *)imwp->imw_buf, DEFBLOCK);
		if ((size_t)(mp->oflowcount) < DEFBLOCK / sizeof (VOFLOW))
		  {
			cpymem((AVOID *)imwp->imw_buf, (AVOID *)cvstart,
				(unsigned)(cvend - cvstart));
			if (DEFBLOCK != imwrite((AVOID *)imwp->imw_buf,
			 DEFBLOCK)) return FRAME_BADFLOWS;
		  }
		else
		  {
			for (cp= cvstart; cp< cvend-DEFBLOCK; cp += DEFBLOCK)
			   if (DEFBLOCK != imwrite((AVOID *)cp, DEFBLOCK))
				return FRAME_BADFLOWS;
			cpymem((AVOID *)imwp->imw_buf, (AVOID *)cp,
				(unsigned)(cvend - cp));
			if (DEFBLOCK != imwrite((AVOID *)imwp->imw_buf,
			 DEFBLOCK)) return FRAME_BADFLOWS;
		  }
	  }
	else {	/* write in FMT_PCS_8 or its swapped equivalent:
		first, we convert from VOFLOW format to OFLOW format */
		vofe = mp->overflow + mp->oflowcount;
		for (vofp = mp->overflow; vofp < vofe; vofp++)
			vflow_to_oflow((OFLOW *)((void *)vofp));
		nbytes = sizeof (OFLOW) *
			(((unsigned)mp->oflowcount + 3) & ~3);
		/* then we write the proper # of bytes out */
		if (nbytes != imwrite((AVOID *)cvstart, nbytes))
			return FRAME_BADFLOWS;
	     }
	return FRAME_OK;
}

static int wrt_bodydata(MAIL *mp, int ruflags)
{ /*	This writes the 16-bit data out.
	Currently it only knows how to do this in a few formats */
	int		deltax, deltay, hit, hitm1, startoff;
	int		wid, widm1, x, y, yli;
	size_t		towri;
	u_short		*uspi, *uspo, *rsp0;
	u_char		*ucpi, *ucpo, *rcp0;
	PAU		imiu, imou;
	IMAGEWORK	*imwp;

	/* we rotate the data and write it out, a block at a time */
	wid = mp->xdimen;	hit = mp->ydimen;
	widm1 = wid - 1;	hitm1 = hit - 1;
	startoff = 0;	deltax = hit;	deltay = 1;	yli = 0;
	if (mp->improps & IPR_XFASTEST)
	  {
		deltax = 1;	deltay = wid;
	  }
	if (mp->improps & IPR_STARTBOTTOM)
	  {
		deltay *= -1;
		startoff += (mp->improps & IPR_XFASTEST) ? wid * hitm1 : widm1;
	  }
	if (mp->improps & IPR_STARTRIGHT)
	  {
		deltax *= -1;
		startoff += (mp->improps & IPR_XFASTEST) ? widm1 : wid * hitm1;
	  }
	if (ruflags & PAD_TO_2TO_N)
	  {
		for (yli = 2; yli < wid; yli <<= 1) ;
		if (mp->improps & IPR_IS8BIT)
			towri = (size_t)yli * sizeof (u_char);
		else	towri = (size_t)yli * sizeof (u_short);
	  }
	else if (mp->improps & IPR_IS8BIT)
		towri = (size_t)(wid) * sizeof (u_char);
	else	towri = (size_t)(wid) * sizeof (u_short);
	if (NULL == (imwp = iminfo(015))) return FRAME_NO_MEMORY;
	if (towri > imwp->imw_bsize)
	  {
		(void)iminfo(04);	imwp->imw_bsize = towri;
		if (NULL == (imwp = iminfo(05))) return FRAME_NO_MEMORY;
	  }
	if (mp->improps & IPR_IS8BIT)
	  {
		rcp0 = mp->data8 + startoff;
		/* if the user wants padding, set up the padding-array now */
		if (ruflags & PAD_TO_2TO_N)
		  {
			ucpi = imwp->imw_buf + yli;
			for (ucpo = imwp->imw_buf + mp->ydimen;
			 ucpo < ucpi; ucpo++) *ucpo = 0;
		  }
		for (y = 0; y < hit; y++, rcp0 += deltay)
		   {
			for (x = 0, ucpi = rcp0, ucpo = imwp->imw_buf;
			 x < wid; x++, ucpi += deltax) *ucpo++ = *ucpi;
			if (towri != imwrite((AVOID *)imwp->imw_buf,towri))
				return FRAME_BAD8BIT;
		   }
		return FRAME_OK;
	  }
	rsp0 = mp->data16 + startoff;
	/* if the user wants padding, set up the padding-array now */
	if (ruflags & PAD_TO_2TO_N)
	  {
		imiu.ucpt = imwp->imw_buf;	uspi = imiu.uspt + yli;
		imou.ucpt = imwp->imw_buf;
		for (uspo = imou.uspt + mp->xdimen; uspo < uspi; uspo++)
			*uspo = 0;
	  }
#ifdef LITTLEND /* little-end machines: impose byteswapping if absent */
	if (!(mp->improps & IPR_BYTESWAPPED))
#else	/* big-end machines: reverse byteswapping if present */
	if (mp->improps & IPR_BYTESWAPPED)
#endif
	  {
		for (y = 0; y < hit; y++, rsp0 += deltay)
		{
			imou.ucpt = imwp->imw_buf;	uspo = imou.uspt;
			for (x = 0, uspi = rsp0; x < wid; x++, uspi += deltax)
				*uspo++ = uswab(*uspi);
			if (towri != imwrite((AVOID *)imwp->imw_buf,
				(unsigned)towri)) return FRAME_BAD16BIT;
		}
	  }
	else {
		for (y = 0; y < hit; y++, rsp0 += deltay)
		  {
			imou.ucpt = imwp->imw_buf;	uspo = imou.uspt;	
			for (x = 0, uspi = rsp0; x < wid; x++, uspi += deltax)
				*uspo++ = *uspi;
			if (towri != imwrite((AVOID *)imwp->imw_buf,
				(unsigned)towri)) return FRAME_BAD16BIT;
		   }
	     }
	return FRAME_OK;
}

/*ARGSUSED*/
#ifdef VMS
static int im_create(char *str, int otype, int ruflags)
{	/* this creates the appropriate type of image. */
	if (otype == FMT_VMS) return creat(str, 0666, "rfm=fix", "mrs=2048");
	else return creat(str, 0666);
}
#else
static int im_create(char *str, int ruflags)
{	/* this creates the appropriate type of image. */
	if ((!(ruflags & DELETE_ORIGINAL)) &&
		(EOK != xgenrename(str))) return -1;
	return creat(str, 0666);
}
#endif

static u_short pick_properties(MAIL *mp, int otype)
{ /*	This looks at the desired output format and sets the
	bits that regulate the output properties */
	int	martype, swtype;
	short	sretv;

	sretv = 0;	swtype = swawb(otype);
	if ((size_t)otype == SDMS_RANGE + (size_t)mp->ydimen)
	  {
		sretv |= IPR_XFASTEST | IPR_IS8BIT;	return sretv;
	  }
	if (-1 == (martype = ismarip(otype)))
		sretv |= IPR_BYTESWAPPED | IPR_XFASTEST;
	else if (1 == martype) sretv |= IPR_XFASTEST;
	else if ((swtype == FMT_RAXIS) || (swtype == FMT_FUJI))
		sretv |= IPR_BYTESWAPPED | IPR_XFASTEST |
			IPR_STARTBOTTOM | IPR_STARTRIGHT;
	else if (swtype == FMT_QTEX)
		sretv |= IPR_BYTESWAPPED | IPR_XFASTEST | IPR_STARTBOTTOM;
	else if ((swtype == FMT_SAXII) || (swtype == FMT_SAXII_16))
		sretv |= IPR_XFASTEST | IPR_STARTBOTTOM;
	else if ((swtype == FMT_PCS_8) || (swtype == FMT_PCS_16))
		sretv |= IPR_BYTESWAPPED | IPR_STARTBOTTOM;
	else if (swtype == FMT_FAST)
	  {
		if ((mp->xdimen % 1042)) sretv |= IPR_XFASTEST;
	  }
	else switch (otype)
	      {
		case	FMT_RAXIS:	case	FMT_FUJI:
		 sretv |= IPR_XFASTEST | IPR_STARTBOTTOM | IPR_STARTRIGHT;
		 break;
		case	FMT_SAXII:	case	FMT_SAXII_16:
		case	FMT_DIPX000:	case	FMT_DIP320:
		case	FMT_QTEX:
		 sretv |= IPR_XFASTEST | IPR_STARTBOTTOM;	break;
		case	FMT_PCS_8:	case	FMT_PCS_16:
			sretv |= IPR_STARTBOTTOM;	break;
		case	FMT_FAST:
			if ((mp->xdimen % 1042)) sretv |= IPR_XFASTEST;
			else sretv |= IPR_STARTRIGHT;	break;
		default:	break;
	     }
	if ((otype == FMT_PCS_8) || (otype == FMT_SAXII) ||
	 (swtype == FMT_PCS_8) || (swtype == FMT_SAXII)) sretv |= IPR_IS8BIT;
#ifdef LITTLEND
	if (sretv & IPR_BYTESWAPPED) sretv &= ~(IPR_BYTESWAPPED);
	else sretv |= IPR_BYTESWAPPED;
#endif
	return sretv;
}

static int retoutput(int rval)
{ /*	This eliminates memory for the `imwp->imw_buf' vector and returns its
	argument */
	IMAGEWORK	*imwp;

	if (NULL == (imwp = iminfo(1))) return rval;
	(void)iminfo(04);	return rval;
}

int output_frame(char *name, MAIL *mailp, xgulong outtype, int ruflags)
{ /*	This writes out a frame in FMT_VMS, FMT_PCS, FMT_PCS_8, FMT_MAR,
	or FMT_FAST according to the user's preference.
	It returns a flag value indicating the result of the attempt. */
	int		ret, known, swtype;
	IMAGEWORK	*imwp;
	static xgulong	i16known[] =
	 {	FMT_MAR, FMT_SAXII, FMT_SAXII_16, FMT_RAXIS, FMT_FAST, 0 } ;

	if (!outtype) outtype = FMT_VMS;	/* backward compatibility */
		/* for now ... */
	swtype = swawb(outtype);
	if (mailp->data16)
	  {
		for (known = 0; i16known[known]; known++)
		   {
			if ((xgulong)outtype == i16known[known]) break;
			else if ((xgulong)swtype == i16known[known]) break;
		   }
		if (i16known[known] == 0) outtype = FMT_PCS_16;
	  }
	mailp->improps = pick_properties(mailp, outtype);
	/* if the user has signaled that he wants to override the
	coordinate conventions of his data type, we do so and then
	install the coordinate convention that he requests */
	if (ruflags & OVERRIDE_COORDS)
	  {
		mailp->improps &= ~(0x8); mailp->improps |= (ruflags << 5);
	  }
	if (NULL == (imwp = iminfo(5))) return FRAME_NO_MEMORY;
	imwp->imw_bsize = 4 * DEFBLOCK;
	/*output file with the right properties*/
#ifdef VMS
	if (0 > (imwp->imw_fd = im_create(name, outtype, ruflags)))
#else
	if (0 > (imwp->imw_fd = im_create(name, ruflags)))
#endif
	  {
		fprintf(stderr, "unable to create output file\n");
		return retoutput(FRAME_BADOPEN);
	  }
	/* now write out the header and the overflows */
	imwp->imw_lhd = mailp;	/* static global version of mailbox */
	if ((mailp->data8) && (!(mailp->improps & IPR_IS8BIT)) &&
	  (FRAME_OK != (ret = cvt_8_to_16(mailp, outtype))))
	  {
		fprintf(stderr, "cvt_8_to_16 returned %d\n",
			ret);
		return retoutput(ret);
	  }
	if (FRAME_OK != (ret = wrt_header(mailp, outtype, ruflags)))
	  {
		fprintf(stderr, "wrt_header returned %d\n", ret);
		return retoutput(ret);
	  }
	if ((mailp->oflowcount) && (FRAME_OK !=
		(ret = wrt_flows(mailp, outtype))))
	  {
		fprintf(stderr, "wrt_flows returned %d\n", ret);
		return retoutput(ret);
	  }
	if (FRAME_OK != (ret = wrt_bodydata(mailp, ruflags)))
	  {
		fprintf(stderr, "wrt_bodydata returned %d\n", ret);
		return retoutput(ret);
	  }
	if (mailp->oflowcount)
	  {
		if (ismarip(outtype))
		  {
			if (FRAME_OK != mkmarflows(outtype))
				return retoutput(FRAME_BADFLOWS);
		  }
		else if ((outtype ==FMT_SAXII) || (swtype == FMT_SAXII) ||
		 (outtype == FMT_SAXII_16) || (swtype == FMT_SAXII_16))
		  {
			if (FRAME_OK != mksaxflows(mailp, imwp->imw_fd,
			 imwp->imw_buf)) return retoutput(FRAME_BADFLOWS);
		  }
	  }
	if (FRAME_OK != ret) return retoutput(ret);
	ret = (0 == close(imwp->imw_fd)) ? FRAME_OK : FRAME_BADCLOSE;
	return retoutput(ret);
}

/* end of imageout.c */
