/* imageutil.c -- various utilities associated with image manipulation

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	11 Jun 2007: memory for dark-subtraction reorganized
	 3 Jun 2007: iminfo returns after freeing memory if its arg is even
	19 Apr 2007: float->double conversions; reorganize use of "{"
	14 Feb 2007: add support for Aviex detectors
	 5 Sep 2006: tm_zone, tm_gmtoff unavailable on IRIX
	11 Aug 2006: manage byteswapped formats without specific defines;
		Mar IPs all have the same format value; lists of formats
		and detector sizes changed accordingly
	 2 Aug 2006: add 2300-pixel Mar IP to list of potential formats
	24 Jun 2006: include xgimage.h; the structures are there;
		write saxadjust and israxis to handle detector-type stuff
	11 Jun 2006: move add2notes to imhead so it can be static;
		move get_calibparams->xgenstr, somealpha->saxiim
	 3 Mar 2006: Mar IP headers and file identification handled correctly
	 2 Feb 2006: string-management bug in timeinternal() fixed;
	30 Dec 2005: subtract 1900 from the year in MarCCD dates
	 5 Dec 2005: iminfo() written: handles common variables for images
	24 Nov 2005: better typecasts for certain arguments
	20 Nov 2005: inline argument declarations and strict prototyping
	31 Oct 2005: time-and-date routines replaced with more systematic stuff
	16 Aug 2005: FRAME_NOT_DELETED replaced with FRAME_NOT_COMPRESSED;
		imvalue(): eliminate confusing comments
	21 Apr 2005: handle diagnostics for timed-out image reads
	10 Jan 2005: split maybe_image() into imagename() and imagelike()
	10 Nov 2004: many more invalid suffixes specified in maybe_image()
	13 Oct 2004: in dettype(), call kill_ftab appropriately
	28 Aug 2004: routines used only here are declared as static; cosmetics;
	 resort_flows() moved to image.c	
	14 Aug 2004: run_fmmin, run_fmmax values initialized in makedrun
	11 Aug 2004: several routines moved from imhead.c and image.c to here
	26 Jun 2004: imagerr.c, imtime.c, imagetype.c merged as to be this
	13 Jun 2004: in maybe_image(), make sure there's a slash
		between the directory name and the filename
	28 Feb 2004: in read_box(), force a re-read if the old inode is 0;
		print diagnostic if xgaccess() returns anything other than 0.
	 5 Feb 2004: ANNOUNCE_BOX flag tells us to output a diagnostic
		in read_box() when a new image is read in
	31 Jan 2004: in read_box, make sure the box lies within the limits
		of the image before trying to read it out
	30 Jan 2004: do FRAMEINDEX memory allocation in dettype() better
	24 Jan 2004: fpi,dtyp,htypes made local to individual routines
	21 Jan 2004: license message inserted
	 6 Jan 2004: replace global variable progname with call to currpname
	 5 Jan 2004: replace global variables with local values and
		arguments to routines; default dimensions stated in dettype()
	13 Dec 2003: eliminate unnecessary variable nr in maybe_image
	 3 Oct 2003: add FMT_MARCCD and byteswapped equiv to fmts[]
		and expand other arrays accordingly;
		change check through fmts[] loop in head2type() so that
		it doesn't have to count how many elements are in fmts[]
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Sep 2003: replace access() with xgaccess() calls in some instances
		so we can gunzip things on the fly
	16 Aug 2003: created imagetype.c out of pieces of setupio and xgenio
	10 Jun 2003: correct non sequitur in title of source file, above
	28 May 2003: record inode of image into MAIL structure
		within read_image() itself
	22 Mar 2003: correct bug in legalimage() in finding first 16 bits
	 5 Jan 2003: reorganized source-file header
	10 Oct 2002: stype range is from 0 through 16 inclusive:
		APS1 detectors managed correctly
	 1 May 2002: routine maybe_image() written:
		used to guess whether a file might be a detector image
	25 Mar 2002: removed MSI attachments
	 2 Mar 1995: Fuji and compressed-image formats;
		change beam-correction for Argonne data
	31 May 1995: alter comments to reflect xgulongs rather than longs
	12 May 1995: correctly record MACScience date format
	 2 Mar 1995: trivial formatting
	 5 May 1994: source file headers, POLYDET->XGEN,
		"xgulong" and "xglong", 64-bit pointers
	 7 Dec 1993: initial X-GEN version

   Purpose: utilities for image handling: error messages, image time routines,
		and routines for identifying image types
   Cross references: image.c, imhead.c, imageout.c
*/
#define		NEED_ERRNO	1
#define		NEED_TIME	1
#define		NEED_STAT	1
#define		NEED_CTYPE	1
#include	"xgimage.h"
#define		MAX_SIZEB	21
#define		MAX_TYPEB	18
#define		MIN_SDMS	(36384)
#define		MAX_SDMS	(73728)
#define		MIN_NONSDMS	(100000)
#define		DIP100_SIZE	5120180
#define		DIP2000_SIZE	12501024
#define		DIP3000_SIZE	18001024
#define		DIP320_SIZE	10240180
#define		FUJI_SIZE	10485760
#define		MIN_SIZE	32000
#define		NIMFORMATS	21

#undef	SMALL_TM	/* tells us that the tm structure lacks some elements*/
#ifdef sgi
#define	SMALL_TM	1
#endif
#ifdef sun
#define	SMALL_TM	2
#endif

extern int count65k(MAIL *headp);
extern void saxadjust(MAIL *mp, REFPARS *rprp, int frmno);
extern int israxis(int fomt);

static int isleap(int year)
{ /*	This corrects `day' for leapyears */
	return ((!(year % 4)) && ((year % 100) || (!(year % 400))));
}

static int dayssofar(int month, int year)
{ /*	This returns the number of days prior to the month
	specified in `month'. It knows about leap years. */
	int		dsof;
	static int	msofar[] =
	 { 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334 } ;

	if ((month < 0) || (month > 11)) return 0;
	dsof = msofar[month];
	return ((month > 1) && (isleap(year))) ? dsof : dsof + 1;
}

static int timeinternal(xgulong dt[], struct tm *xgtp, int direction)
{ /* This converts between a pair of xgulongs and a tm structure */
	int		month, td, year;
#ifndef	SMALL_TM
	static char	ctz[] = "CST";
#endif

	if ((direction < 0) || (direction > 1)) return EBADVALUE;
	if (direction == 0) /* convert xgulongs to tm structure */
	  {
		if (NULL == dt) return EBADVALUE;
		year = (dt[0]) >> 16; if (year > 1900) year -= 1900;
		xgtp->tm_year = year;
		xgtp->tm_yday = ((unsigned short)(dt[0] & 0777)) - 1;
		xgtp->tm_hour = dt[1] >> 24;
		xgtp->tm_min = ((dt[1] >> 16) & 0377);
		xgtp->tm_sec = (dt[1] & 0377);
		for (td = month = 0; month < 12; month++)
		   {
			if (xgtp->tm_yday < (td =
				dayssofar(xgtp->tm_mon, year))) break;
		   }
		if (month >= 12) month = 11;
		xgtp->tm_mon = month;
		xgtp->tm_mday = 1 + xgtp->tm_yday - td;
		xgtp->tm_isdst = 0;
#ifndef SMALL_TM
		/* The timezone stuff may be wrong, but we need something */
		xgtp->tm_zone = ctz; xgtp->tm_gmtoff = 21600;
#endif
		/* this is not right, but I don't feel like figuring it out */
		xgtp->tm_wday = 0;
	  }
	else {
		year = xgtp->tm_year;	/* tm_year defined as since 1900 */
		if (year < 70) year += 100;	/* dt[0] is since 0 CE */
		if (year < 1900) year += 1900;
		dt[0] = (year << 16) | (xgtp->tm_yday + 1);
		dt[1] = (xgtp->tm_hour << 24) | (xgtp->tm_min << 16) |
			xgtp->tm_sec;
	      }
	return EOK;
}

static int xgatoi(char *str, size_t n)
{ /*	This returns an integer based on the first n characters in `str'.
	It doesn't assume that str is null-terminated,
	so it can function properly in the middle of a longer string.
	It returns zero if there's something wrong with the input. */
	char	*cp;
	int	illchar, mfac, out;
	size_t	i;
	
	if ((n <= 0) || (strlen(str) < n)) return 0;
	for (cp = str, i = illchar = 0; '\0' != *cp && i < n; i++)
	   {
		if ((*cp != ' ') && ((*cp < '0') || (*cp > '9'))) illchar++;
	   }
	if (illchar) return 0;
	for (cp = str + (n - 1), mfac = 1, out = 0; cp >= str; cp--)
	   {
		if (('0' <= *cp) && (*cp <= '9')) out += (*cp - '0') * mfac;
		mfac *= 10;
	   }
	return out;
}	

static int monthcvt(char *str, int *monthnp, int mflags)
{ /*	The converts between a three-letter, lower-case string for a month
	and the integer number (0 to 11) of that month. Values of mflags are:
	mflags	meaning
	0	converting a string to an integer
	1	converting an integer to a lower-case string (e.g. "oct")
	3	converting an int to a capitalized string (e.g. "Oct")
	7	converting an int to an uppercase string (e.g. "OCT")
	The routine returns EOK if successful, various errors if not. */
	int		month;
	char		*cp;
	static char	months[] = "janfebmaraprmayjunjulaugsepoctnovdec";

	if (mflags == 0) /* convert string to value between 0 and 11 */
	  {
		if ((NULL == str) || (strlen(str) < 3)) return EBADVALUE;
		*str |= 0140;	*(str + 1) |= 0140;	*(str + 2) |= 0140;
		for (month = 0; month < 12; month++)
			if (!strncmp(str, &(months[3*month]), 3)) break;
		if (month >= 12) month = 11;
		*monthnp = month;	return EOK;
	  }
	else {
		if ((*monthnp < 0) || (*monthnp >= 12)) return EBADVALUE;
		cp = months + 3 * *monthnp;
		*str = *cp++;	*(str + 1) = *cp++;	*(str + 2) = *cp;
		*(str + 3) = '\0';
		if (mflags & 02) *str &= 0137;	/* capitalize 1st letter */
		if (mflags & 04) /* capitalize other two letters */
		  {
			*(str + 1) &= 0137;	*(str + 2) &= 0137;
		  }
		return EOK;	/* go the other way */
	     }
}

static void blankto0(char *str, int nchars)
{ /*	This replaces all blanks in `str' with '0', up through nchars */
	int	i;
	char	*cp;

	for (cp = str, i = 0; (i < nchars) && ('\0' != *cp); i++, cp++)
		if (' ' == *cp) *cp = '0';
}

static int timestrings(struct tm *xgtp, int strtype, int direction, char *dstr)
{ /*	This converts between a tm structure and various string formats.
	When direction == 0 we're converting the structure to strings;
	when direction == 1 we're going the other way.
	The convention is that all of the integers in xgtp start at zero
	except the day of the month; xgtp->tm_year is since 1900.
	The routine returns EOK if successful, EBADVALUE if not. */
	int		doff, i, year;
	time_t		tl0;
	char		*cp, cmonth[4];
	static char	daw[] = "SunMonTueWedThuFriSat";
#ifndef SMALL_TM
	static char	ctz[] = "CST";
#endif

	if (strtype < 2) return EBADVALUE; /* not handled here */
	if (direction == 0) /* structure to string */
	  {
		year = xgtp->tm_year + 1900;
		if (EOK != (i = monthcvt(cmonth, &(xgtp->tm_mon),1))) return i;
		if (strtype == 2) /* Cornell: 11-APR-92.18:13:00 */
		  {
			sprintf(dstr, "%2d-%c%c%c-%2d.%2d:%2d:%2d",
			 xgtp->tm_mday, cmonth[0], cmonth[1], cmonth[2],
			 year % 100, xgtp->tm_hour, xgtp->tm_min,
			 xgtp->tm_sec);	dstr[9] = '\0';
			blankto0(dstr, 9);	blankto0(dstr + 10, 8);
		  }
		if (strtype == 3) /* Mar: 11-APR-1992 18:13:00.14  */
		  {
			sprintf(dstr, "%2d-%s-%4d %2d:%2d:%2d.00\n",
			 xgtp->tm_mday, cmonth, year,
			 xgtp->tm_hour, xgtp->tm_min, xgtp->tm_sec);
			blankto0(dstr, 23);
		  }
		else if (strtype == 4) /* SAXII: 09/24/97 17:41:09 */
		  {
			sprintf(dstr, "%2d/%2d/%2d %2d:%2d:%2d",
				xgtp->tm_mon + 1, xgtp->tm_mday, year,
				xgtp->tm_hour, xgtp->tm_min, xgtp->tm_sec);
			blankto0(dstr, 8);	blankto0(dstr + 9, 8);
		  }
		else if (strtype == 5) /* R-Axis date: 11-APR-1992 */
		  {
			sprintf(dstr, "%2d-%s-%4d",
			 xgtp->tm_mday, cmonth, year);
			blankto0(dstr, 11);	
			dstr[11] = ' ';	/* ... an R-Axis peculiarity */
		  }
		else if (strtype == 6) /* Mar IP: 11-APR-1992 18:13:00.14 */
		  {
			sprintf(dstr, "%2d-%s-%4d %2d:%2d:%2d.00",
			 xgtp->tm_mday, cmonth, year,
			 xgtp->tm_hour, xgtp->tm_min, xgtp->tm_sec);
			blankto0(dstr, 11);	blankto0(dstr + 12, 11);
		  }
		else if (strtype == 7) /* 04/18/91   14:47:40 */
		  {
			sprintf(dstr, "%2d/%2d/%2d    %2d:%2d:%2d",
			 xgtp->tm_mon + 1, xgtp->tm_mday, year % 100,
			 xgtp->tm_hour, xgtp->tm_min, xgtp->tm_sec);
			blankto0(dstr, 8);	blankto0(dstr + 11, 8);
		  }
		else if (strtype == 8) /* 04/18/1991 14:47:40 */
		  {
			sprintf(dstr, "%2d/%2d/%4d %2d:%2d:%2d",
			 xgtp->tm_mon + 1, xgtp->tm_mday, year,
			 xgtp->tm_hour, xgtp->tm_min, xgtp->tm_sec);
			blankto0(dstr, 10);	blankto0(dstr + 11, 8);
		  }
		else if (strtype == 9) /* Sun Sep 16 01:03:52 1973\n */
		  { /* we'll fill in the day of the week with a dodge */
			tl0 = mktime(xgtp);	xgtp = localtime(&tl0);
			cmonth[0] &= 0137;	cp = daw + 3 * xgtp->tm_wday;
			sprintf(dstr, "%c%c%c %c%c%c %2d %2d:%2d:%2d %4d\n",
			 *cp, *(cp + 1), *(cp + 2), cmonth[0], cmonth[1],
			 cmonth[2], xgtp->tm_mday, xgtp->tm_hour, xgtp->tm_min,
			 xgtp->tm_sec, year);
			blankto0(dstr + 8, 1);	blankto0(dstr + 11, 8);
		  }
		else if (strtype == 10) /* 092315281998.57 */
		  {
			sprintf(dstr, "%2d%2d%2d%2d%4d.%2d",
			 xgtp->tm_mon + 1, xgtp->tm_mday, xgtp->tm_hour,
			 xgtp->tm_min, year, xgtp->tm_sec);
			blankto0(dstr, 15);
		  }
		else return EBADVALUE;
	  }
	else {	/* tstr to structure */
		xgtp->tm_wday = 0; /* until we can do it better */
		if (strtype == 2) /* Cornell IP: 11-APR-92.18:13:00 */
		  {
			(void)monthcvt(dstr + 3, &(xgtp->tm_mon), 0);
			xgtp->tm_mday = xgatoi(dstr     , 2);
			xgtp->tm_year = xgatoi(dstr +  7, 2);
			xgtp->tm_hour = xgatoi(dstr + 10, 2);
			xgtp->tm_min  = xgatoi(dstr + 13, 2);
			xgtp->tm_sec  = xgatoi(dstr + 16, 2);
		  }
		else if (strtype == 3) /* Mar IP: 11-APR-1992 18:13:00.14 */
		  {
			(void)monthcvt(dstr + 3, &(xgtp->tm_mon), 0);
			xgtp->tm_mday = xgatoi(dstr     , 2) - 1;
			xgtp->tm_year = xgatoi(dstr +  7, 4) - 1900;
			xgtp->tm_hour = xgatoi(dstr + 12, 2);
			xgtp->tm_min  = xgatoi(dstr + 15, 2);
			xgtp->tm_sec  = xgatoi(dstr + 18, 2);
		  }
		else if (strtype == 4) /* SAXII: 09/24/97 17:41:09 */
		  {
			xgtp->tm_mon  = xgatoi(dstr     , 2) - 1;
			xgtp->tm_mday = xgatoi(dstr +  3, 2);
			xgtp->tm_year = xgatoi(dstr +  6, 2);
			xgtp->tm_hour = xgatoi(dstr +  9, 2);
			xgtp->tm_min  = xgatoi(dstr + 12, 2);
			xgtp->tm_sec  = xgatoi(dstr + 15, 2);
		  }
		else if (strtype == 5) /* R-Axis: 11-APR-1992 */
		  {
			monthcvt(dstr + 3, &(xgtp->tm_mon), 0);
			xgtp->tm_mday = xgatoi(dstr    , 2);
			xgtp->tm_year = xgatoi(dstr + 7, 4) - 1900;
			xgtp->tm_hour = xgtp->tm_min = xgtp->tm_sec = 0;
		  }
		else if (strtype == 6) /* VMS: 19-jun-1989 19:40:24 */
		  {
			monthcvt(dstr + 3, &(xgtp->tm_mon), 0);
			xgtp->tm_mday = xgatoi(dstr     , 2);
			xgtp->tm_year = xgatoi(dstr +  7, 4);
			xgtp->tm_hour = xgatoi(dstr + 12, 2);
			xgtp->tm_min  = xgatoi(dstr + 15, 2);
			xgtp->tm_sec  = xgatoi(dstr + 18, 2);
		  }
		else if (strtype == 7) /* 04/18/91   14:47:40 */
		  {
			xgtp->tm_mday = xgatoi(dstr +  3, 2);
			xgtp->tm_mon  = xgatoi(dstr     , 2) - 1;
			xgtp->tm_year = xgatoi(dstr +  6, 2);
			xgtp->tm_hour = xgatoi(dstr + 11, 2);
			xgtp->tm_min  = xgatoi(dstr + 14, 2);
			xgtp->tm_sec  = xgatoi(dstr + 17, 2);
		  }
		else if (strtype == 8) /* 04/18/1991 14:47:40 */
		  {
			xgtp->tm_mday = xgatoi(dstr +  3, 2);
			xgtp->tm_mon  = xgatoi(dstr     , 2) - 1;
			xgtp->tm_year = xgatoi(dstr +  6, 4) - 1900;
			xgtp->tm_hour = xgatoi(dstr + 11, 2);
			xgtp->tm_min  = xgatoi(dstr + 14, 2);
			xgtp->tm_sec  = xgatoi(dstr + 17, 2);
		  }
		else if (strtype ==  9) /* Sun Sep 16 01:03:52 1973 */
		  {
			(void)monthcvt(dstr + 4, &(xgtp->tm_mon), 0);
			/* occasionally, single-digit mday values appear */
			doff = 8;	if (*(dstr + 9) == ' ') doff = 7;
			xgtp->tm_mday  = xgatoi(dstr +  doff, 2);
			xgtp->tm_year = xgatoi(dstr + doff + 12, 4);
			xgtp->tm_hour = xgatoi(dstr + doff + 3, 2);
			xgtp->tm_min  = xgatoi(dstr + doff + 6, 2);
			xgtp->tm_sec  = xgatoi(dstr + doff + 9, 2);
			for (i = 0; i < 7; i++)
			   {
				if (!strncmp(daw + 3 * i, dstr, 3))
				  {
					xgtp->tm_wday = i;	break;
				  }
			   }
		  }
		else if (strtype == 10) /* 092315281998.57 */
		  {
			xgtp->tm_mon  = xgatoi(dstr    ,  2) - 1;
			xgtp->tm_mday = xgatoi(dstr +  2, 2);
			xgtp->tm_year = xgatoi(dstr +  8, 4) - 1900;
			xgtp->tm_hour = xgatoi(dstr +  4, 2);
			xgtp->tm_min  = xgatoi(dstr +  6, 2);
			xgtp->tm_sec  = xgatoi(dstr + 13, 2);
		  }
		xgtp->tm_yday = xgtp->tm_mday +
			dayssofar(xgtp->tm_mon, xgtp->tm_year);
		xgtp->tm_isdst = 0;
#ifndef SMALL_TM
		xgtp->tm_zone = ctz;	xgtp->tm_gmtoff = 21600;
#endif
	  }
	return EOK;
}

int timecvt(xgulong dt[], char *instr, char *outstr,
 time_t *tlp, int intype, int outtype)
{ /*	This converts among various date and time formats.
	possible values for intype or outtype:
	0	dt contains a pair of xgulongs of the form
		 dt[0] = (year << 16) | (day + 1)
		 dt[1] = (hour << 24) | (min << 16) | sec
	1	*tlp contains # of seconds since midnight jan 1 1970
	2	string like 11-APR-92.18:13:00 (dot is actually a null)
		(old Cornell IPs)
	3	string like 11-APR-1992 18:13:00.14 (Mar IPs)
	4	string like 09/24/97 17:41:09 (Bruker/Siemens CCDs, etc.)
	5	string like 11-APR-1992 (R-Axis)
	6	string like 19-jun-1989 19:40:24 (Xentronics VMS)
	7	string like 04/18/91   14:47:40 (old string format)
	8	string like 04/18/1991 14:47:40 (old string format)
	9	string like Sun Sep 16 01:03:52 1973\n (MacScience)
	10	string like 092315281998.57 (Mar CCD)
  */
	int			retv;
	static time_t		interntim;
	static struct tm	xgts;
	struct tm		*xgtp;

	if (intype == outtype) return 0;	/* trivial case */
	if ((intype > 2) && ((NULL == instr) || ('\0' == *instr)))
		return EBAD;	/* bad input string */
	if (intype == 0) /* input two xgulongs encoding the time */
	  {
		if (EOK != (retv = timeinternal(dt, &xgts, 0))) return retv;
		interntim = mktime(&xgts);
		if (NULL != tlp) *tlp = interntim;
		if (outtype == 1) return EOK;
		else return timestrings(&xgts, outtype, 1, outstr);
	  }
	else if (intype == 1) /* input # of secs since midnight Jan 1 1970 */
	  {
		if (NULL != (xgtp = localtime(tlp))) return EBADVALUE;
		cpymem(&xgts, xgtp, sizeof (struct tm));
		if (outtype == 0) return timeinternal(dt, &xgts, 1);
		else return timestrings(&xgts, outtype, 1, outstr);
	  }
	else {
		if (EOK != (retv = timestrings(&xgts, intype, 1, instr)))
			return retv;
		*tlp = mktime(&xgts);
		if (outtype == 0) return timeinternal(dt, &xgts, 1);
		else return timestrings(&xgts, outtype, 0, outstr);
	     }
	/*NOTREACHED*/
}

int image_error(FILE *fp, char *name, int error_code, int disposition)
{ /*	This prints an appropriate message for various errors encountered
	in reading in a frame and then returns either the error
	code itself or EBAD, depending on the variable `disposition'. */
	char	*cpname;

	if (error_code == FRAME_OK) return FRAME_OK;
	if (disposition & 02) fprintf(fp, "Output frame %s ", name);
	else fprintf(fp, "Frame %s ", name);
	switch (error_code)
	      {
		case	FRAME_BADOPEN:
		/* here the attempt to open the file failed.  We look at
		the error number to discern what's really wrong. */
		  switch (errno)
		    {
			case	ENOENT:	fprintf(fp, "was not found");	break;
			case	EPERM:	fprintf(fp,
			 "can't be opened: you don't own it"); break;
			case	EACCES:fprintf(fp,
			 "can't be opened: it's read-protected for you"); break;
			case	ENFILE:	fprintf(fp,
			 "can't be opened: too many other files open");	break;
			default:	fprintf(fp,
			 "can't be opened: errno = %5d", errno);	break;
		    }
		  break;
		case	FRAME_BADMAIL:	/* faulty mailbox */
			fprintf(fp, "has a bad mailbox (header) record");
			break;
		case	FRAME_BADFLOWS:	/* faulty overflow table */
			fprintf(fp,"has an error in the overflow table"); break;
		case	FRAME_BAD8BIT:	/* bad 8-bit data */
			fprintf(fp, "has an error in the 8-bit data");	break;
		case	FRAME_ALREADY_CONVERTED:	/* already in VMS fmt*/
			fprintf(fp, "has already been converted");	break;
		case	FRAME_NO_MEMORY:	/* memory allocation failed */
			fprintf(fp,
			 "can't be processed: memory allocation failed");
			break;
		case	FRAME_BADLENGTH:	/* frame is too short */
			fprintf(fp,
			 "is not long enough to contain all the data"); break;
		case	FRAME_FMT_UNCODED:	/* legal format, no software*/
			fprintf(fp,
			 "is in a format for which no code has been written");
			break;
		case	FRAME_FMT_UNKNOWN:	/* unknown file-format value*/
			fprintf(fp, "is in an unknown format");	break;
		case	FRAME_NOT_COMPRESSED:	/* not deleted */
			fprintf(fp, "wasn't compressed upon closure");	break;
		case	FRAME_TIMEDOUT:	/* attempt to read image timed out */
			fprintf(fp, "did not arrive before timeout"); break;
		default:
			fprintf(fp, "Faulty: error code = %d\n", error_code);
	     }
	if ((disposition & 01) && error_code != FRAME_NOT_COMPRESSED) /*fatal*/
	  {
		if ((NULL != (cpname = xgprogname(NULL, 0))) &&
		 ('\0' != *cpname)) fprintf(fp, "; %s failed \n", cpname);
		else	fprintf(fp, "; program failed\n");
		return EBAD;
	  }
	fprintf(fp, ".\n");	errno = 0;	return error_code;
}

char *imtypestr(int styp)
{ /* This returns a vaguely reasonable string describing an image type */
	static char	*dtname[] = { "unknown", "SDMS", "Siemens multiwire",
	 "Old Argonne CCD", "FAST or Argonne CCD", "R-Axis", "Mar IP",
	 "Quantex", "Cornell IP", "MacScience", "Siemens CCD", "Mar CCD",
	 "Brandeis CCD", "ADSC Quantum 4", "ADSC Quantum 210",
	 "ADSC Quantum 315", "SBC-CAT CCD",
	 "Aviex 160x80 CCD", "Aviex 160sq CCD" } ;

	return ((styp < 0) || (styp > 18)) ? NULL : dtname[styp];
}

static int imfmt(int fmind)
{	/* This returns an X-GEN FMT_* value associated with an index value */
	static unsigned int fmts[] =
	 {	0, FMT_UNREAD,					/* 0- 1 */
		FMT_CAD, FMT_PCS_8, FMT_SAXII, FMT_SAXII_16,	/* 2- 5 */
		FMT_CCD, FMT_CCD_VMS, FMT_CCD_ALT, 		/* 6- 8 */
		FMT_CCD_VMSALT, FMT_FAST, FMT_FASTALT,		/* 9-11 */
		FMT_RAXIS, FMT_MAR, FMT_QTEX,			/*12-14 */
		FMT_CORNELL, FMT_DIP100, FMT_DIP320,		/*15-17 */
		FMT_DIPX000, FMT_MARCCD_L, FMT_MARCCD, 0 };	/*18-21*/
	return ((fmind < 0) || (fmind >= NIMFORMATS)) ?  0 : (int)(fmts[fmind]);
}

static int legalimage(int firstword, size_t fisize)
{ /*	If the file format in "firstword" is one of the recognized values
	in the list `fmts', this returns that word. If not, we look at the
	file size to see if it's one of those associated with a headerless
	image. We return an appropriate word in those instances.
	If neither of those things happen, we return 0.
	For these purposes compressed files (recognized also by the first
	four bytes in the file) are going to come back as legal.
	This will of course result in a few non-images returning a nonzero
	code, but that's better than excluding good images that just have
	the misfortune of having been compressed */
	int		i, first16, next16, swfirst;

	swfirst = swawb(firstword);
	for (i = 1; i < NIMFORMATS; i++)
	   {
		if (firstword == imfmt(i)) return firstword;
		if (swfirst == imfmt(i)) return firstword;
	   }
	if ((i = ismarip(firstword))) return (i < 0) ? swawb(FMT_MAR) : FMT_MAR;
	first16 = firstword & 0xffff;	next16 = (firstword >> 16) & 0xffff;
	if (first16 == 0x8b1f) return 0x8b1f;
	if (first16 == 0x9d1f) return 0x9d1f;
	if (next16 == 0x1f8b) return 0x1f8b;
	if (next16 == 0x1f9d) return 0x1f9d;
	if (fisize == DIP100_SIZE) return FMT_DIP100;
	if (fisize == DIP2000_SIZE) return FMT_DIPX000;
	if (fisize == DIP3000_SIZE) return FMT_DIPX000;
	if (fisize == DIP320_SIZE) return FMT_DIP320;
	if (fisize == FUJI_SIZE) return FMT_FUJI;
	if ((MIN_SDMS < fisize) && (fisize < MAX_SDMS) && !(fisize & 4096))
		return SDMS_RANGE;
	return 0;
}

static int xgdigits(char *nam, char *prefix, int mindig, int maxdig)
{ /*	This returns 1 iff `name' contains the string `prefix' and has at
	least `mindig' digits and no more than `maxdig' digits following
	`prefix'. The digits have to be at the end of `name'. */
	char	*su, *cp, *cpp, *end, *s0, *pref0, *dig0;
	int	lenp;

	/* check and rationalize inputs */
	if ((NULL == nam) || ('\0' == *nam)) return 0;
	su = (NULL == prefix) ? "." : prefix;
	if (mindig < 0) mindig = 0;	if (maxdig < mindig) maxdig = mindig;
	/* examine `nam' to see it's long enough and contains the beginning
	 of `prefix'. */
	s0 = prefix;	pref0 = dig0 = NULL;
	for (cp = su, lenp = 0; '\0' != *cp; cp++) lenp++;
	for (cp = nam; '\0' != *cp; cp++)
	   if (*cp == *s0)
	     {
		pref0 = cp;	dig0 = cp + lenp;
	     }
	if (NULL == dig0) return 0;
	end = cp;
	/* there has to be room for at least `mindig' digits */
	if (dig0 > end - mindig) return 0;
	/* if `prefix' is longer than one character,
	 see if all of it is present */
	if (lenp > 1)
	  {
		for (cp = pref0, cpp = s0; '\0' != *cpp; cp++, cpp++)
			if (*cp != *cpp) return 0;
	  }
	/* now make sure the appropriate number of remaining
	 characters are all digits */
	for (cp = dig0; cp < end; cp++)
	   if (('0' > *cp) || (*cp > '9')) return 0;
	/* whew. If we get all the way to here, the name satisfies */
	return 1;
}

int imagename(char *filen)
{ /*	This returns 1 if the file name 'filen' looks like it might
	be a detector image; it returns 0 otherwise.
	Unlike the earlier maybe_image(), this one only checks that the
	name of the file is plausible; the checks on its contents are
	accomplished in the next routine, `imagelike()' to save time.
	Earlier versions excluded a few suffixes; by contrast,
	this one _requires_ that the filename end with certain
	possible suffixes -- see the gsuff[] array. */
	char		**dcpi;
	size_t		lennam;
	int		allowed, i, rv;
	static char	*shname;
	static size_t	oldlen;
	static char	shmsg[] = "shname in imagename";
	static char	ignore[] = ",0";
	static char	*gsuff[] = {  ".img", ".osc", ".image",
	 ".stl", ".mccd", ".mar", ".brs", ".cai", ".smv", NULL } ;

	if (NULL == filen)
	  {
		memzap((AVOID **)&shname, shmsg); oldlen = 0; return 0;
	  }
	/* before we do anything else, disallow ., .., core, core.0136, etc. */
	if (!strcmp(filen, ".")) return 0;
	if (!strcmp(filen, "..")) return 0;
	if (!strcmp(filen, "core")) return 0;
	if (!strncmp(filen, "core.", 5)) return 0;
	/* create a version of the filename without terminal ".gz", etc. */
	lennam = strlen(filen) + 8;
	if (lennam < 40) lennam = 40;
	if ((NULL == shname) || (lennam > oldlen))
	  {
		memzap((AVOID **)&shname, shmsg);	oldlen = lennam;
		if (NULL == (shname = mallop(lennam, shmsg))) return 0;
	  }
	zeromem((AVOID *)shname, lennam);
	strcpy(shname, filen);		 (void)xgstrend(shname, ".gz", 1);
	(void)xgstrend(shname, ".z", 1); (void)xgstrend(shname, ".Z", 1);
	/* create a version of the filename without terminal ",0", etc. */
	for (i = 0; i <= 9; i++)
	   {
		ignore[1] = '0' + i; if (xgstrend(shname, ignore, 1)) break;
	   }
	/* see if the filename ends with an allowed suffix */
	allowed = 0;
	if (xgdigits(shname, ".", 2, 4)) allowed = 1;
	else {
		for (dcpi = gsuff; (NULL != *dcpi) && ('\0' != **dcpi);dcpi++)
		   {
			if (xgstrend(shname, *dcpi, 0))
			  {
				allowed = 1; break;
			  }
		   }
	     }
	rv = 1;
	if (!allowed)
	  {
		if (xgdigits(shname, ".mar", 3, 4)) allowed = 1;
		if (!allowed) rv = 0;
	  }
	return rv;
}

int imagelike(char *diren, char *filen)
{	/* This returns 1 iff the file diren/filen has internal properties
	consistent with its being a detector image.
	At the moment the criteria are that it must be a regular file,
	readable, and at least 70 kilobytes in length--the smallest detector
	image I know of is from the SDMS systems, which are
	256*144*2 = 73728 bytes.  */
	int		firstword, lfd, npri, rv;
	size_t		funlen, lendir;
	off_t		midm40;
	char		*cpi;
	struct stat	tebuf;
	static char	funame[512];

	/* construct the full file name */
	lendir = strlen(diren);	if (lendir < 1) return 0;
	strcpy(funame, diren);	cpi = diren + (lendir - 1);
	if ('/' != *cpi) strcat(funame, "/");
	strcat(funame, filen);
	if (0 != access(funame, 4)) return 0;	/* file must be readable */
	if (0 != stat(funame, &tebuf)) return 0; /* stat() must allow it */
	if ((tebuf.st_mode & S_IFMT) != S_IFREG) return 0; /* regular file */
	/* file must be at least 100 kilobytes (it would be 256K except
	 that we need to allow for compression), unless it's an SDMS file,
	 in which case it'll be between MIN_SDMS and MAX_SDMS */
	if ((funlen = tebuf.st_size) < MIN_NONSDMS)
	  {
		if ((funlen < MIN_SDMS) || (funlen > MAX_SDMS)) return 0;
	  }
	/* All image formats currently in use are binary in the middle.
	Some have ascii headers; at least one has an ascii trailer.
	So we check to see if the middle 160 bytes of the file include
	at least one non-printable character. Presumably a file that
	can't be read or sought has problems, so return 0 then too. */
	if (0 > (lfd = open(funame, 0))) return 0; /* open the file? */
	/* measure middle of file minus 40 bytes */
	midm40 = tebuf.st_size / 2 - 40;	errno = 0;
	/* seek to just before the middle and read 80 bytes */
	if ((-1 == lseek(lfd, midm40, 0)) || (80 > read(lfd,funame,80)))
	  {
		(void)close(lfd); return 0;
	  }
	/* see how many are printable. We predict < 78 are */
	for (cpi = funame, npri = rv = 0; rv < 80; rv++, cpi++)
		if (isprint((int)(*cpi)) ||
			('\t' == *cpi) || ('\n' == *cpi)) npri++;
	/* finally, we look at the first four bytes. They either need
	 to match one of the recognized file formats, or they need to be
	 the bytes that tell us we have a gzipped or Zipped file */
	rv = (npri <= 77); /* too many printable characters => this is text */
	if (-1 == lseek(lfd, (off_t)0L, 0)) rv = 0;
	else if (4 > read(lfd, (void *)&firstword, 4)) rv = 0;
	else if (0 == legalimage(firstword, funlen)) rv = 0;
	(void)close(lfd);	/* close the file regardless */
	return rv;
}

int detdimen(int dimcode)
{ /* 	returns the actual size associated with one of the size-codes
	in the `ydimv' array' */
	static int ydimv[] =  {	   0, 144, 256, 512,1024,1200,1536,1600,1900,
	 2000,2048,2300,2400,2500,3072,3450,3600,4096,6144,2084,4168,0 } ;

	return ((dimcode < 0) || (dimcode > MAX_SIZEB)) ? 0 : ydimv[dimcode];
}

int head2type(MAIL *headp)
{ /*	This uses a detector image header to determine
	the detector-type code.  The code is an or-ing of two bytes.
	The low byte is the detector type (Siemens, SDMS, FAST, ... )
	and the upper byte is the dimension of the detector.
	Most detectors are square, so we only worry about the Y dimension;
	the two known exceptions to squareness are SDMS and the Cornell image
	plate, and those have well-characterized X values for a given Y.
	Oct 03: change search limits in second loop so that we're just
	searching for a blank value rather than coming up to a maximum index
	so that the code is easier to maintain. */
	int		itry, sizebyte, swfmt, tfmt, ttry, typebyte;
	DATARUN		*drp;
	static int	htypes[] = {
			 1, 0, 2, 2, 2, 2, 3, 3, 3, 3,	/* 0- 9 */
			 4, 4, 5, 6, 7, 8, 9, 9, 9,11,	/*10-19 */
			11, 0 } ;			/*20-21 */

	for (itry = 0, sizebyte = 20; itry < MAX_SIZEB; itry++)
		if ((int)(headp->ydimen) == detdimen(itry))
		  {
			sizebyte = itry; break;
		  }
	tfmt = headp->format;	swfmt = swawb(tfmt);
	for (itry = 1, typebyte = 1; itry < NIMFORMATS; itry++)
	   if ((tfmt == (ttry = imfmt(itry))) || (swfmt == ttry))
	     {
		typebyte = htypes[itry];
		/* special cases involving Siemens CCD's */
		if ((6 <= itry) && (itry <= 9) &&
		 (headp->cellsize != 200)) typebyte = 10;
		break;
	     } /* end of loop through itry values */
	/* special case of Brandeis CCD, which uses the FAST
	 format conventions but has 1024x1024 pixels and a 0.1mm pixelsize*/
	if ((typebyte == 4) && (headp->cellsize == 100) &&
	 (headp->xdimen == 1024)) typebyte = 12;
	/* special cases of ADSC CCD's and the SBC CCD's. These read out as if
	 they were FAST detectors, but they're 1536x1536 or larger. */
	else if ((typebyte == 4) && (headp->xdimen >= 1536))
	  {
		drp = headp->runheader;
		if (headp->xdimen == 6144) typebyte = 15; /* Q315, unbinned */
		else if (headp->xdimen == 3072)
		  {
			if (fabs(drp->run_d1pos[1] - 0.0798) < 0.005)
				typebyte = 16;	/* SBC CCD, unbinned */
			else typebyte = 15;	/* Quantum 315, binned mode */
		  }
		else if (headp->xdimen == 1536) typebyte = 16; /* SBC CCD */
		else if (headp->xdimen == 4096) typebyte = 14;
		else if ((headp->xdimen == 2048) &&
		 (NULL != drp) && (fabs(drp->run_d1pos[1] - 0.1024) < 0.01))
			typebyte = 14;
		else if (headp->xdimen == 4168)
		  {
			if (headp->ydimen == 2084) typebyte = 17;
			else typebyte = 18;
		  }
		else typebyte = 13;
	  }
	if (typebyte <= 0) /* header is ambiguous */
	  {
		if ((SDMS_RANGE <= headp->format) &&
		 (headp->format < SDMS_RANGE + 1024)) typebyte = 1;
		else	typebyte = 11;
	  }
	return (sizebyte << 8) | typebyte;
}

int dettype(int dimen[])
{ /*	This tries to determine the detector type by a specific
	sequence of steps. First it tries to read a data frame header
	and get it from there. Then it looks in the brass-plate image
	header and get it from there. Next it tries the UPARAMS file,
	which has the value socked away; then the MULTIREF's CRYSTAL
	header. If it never finds a nonzero value, it returns 0.
	It returns the detector dimensions into dimen[0] and [1]. */
	int		dtyp, ftyp, fsiz, lflags;
	FRAMEINDEX	*lfm;
	FRAMEENTRY	*ftq;
	CRYSTAL		lxhd;
	REFPARS		lpars;
	MAIL		lhead;
	static FILE	*fpi;

	zeromem((void *)&lhead, sizeof (MAIL)); /* null out working MAIL */
		/* see if we can figure it out from an ordinary image */
	lflags = JUST_HEADER;
	if ((NULL != (lfm = ftables(3))) || (NULL != (lfm = ftables(7))))
	  {	
		for (ftq = lfm->fi_fmi; ftq <= lfm->fi_fme; ftq++)
		   {
			if ((FRAME_OK == read_image(ftq->fmname,
			 lflags, &lhead)) && (0 != (dtyp =
				head2type(&lhead))))
			  {
				dimen[0] = lhead.xdimen;
				dimen[1] = lhead.ydimen;
				memzap((AVOID **)&(lhead.runheader),
				 "lhead->runheader in dettype");
				return dtyp;
			  }
		   }
	  }
	/* if we got this far, we haven't found it. Try CALDATA: */
	if ((FRAME_OK == read_image(trname("caldata"), lflags, &lhead))
	 && (0 != (dtyp = head2type(&lhead))))
	  {
		dimen[0] = lhead.xdimen; dimen[1] = lhead.ydimen;
		memzap((AVOID **)&(lhead.runheader),
			 "lhead->runheader in dettype");
		return dtyp;
	  }
	/* okay, how about UPARAMS */
	zeromem((void *)&lpars, sizeof (REFPARS));
	if ((NULL != (fpi = xgopen("uparams", "r"))) &&
	 (sizeof (REFPARS) == xgread((void *)&lpars, XG_REFPARS, fpi)))
	  {
		(void)fclose(fpi);	fpi = NULL;
		dtyp = lpars.xdtype; ftyp = dtyp & 0377; fsiz = dtyp >> 8;
		if ((0 < ftyp) && (ftyp <= MAX_TYPEB) && (0 < fsiz) &&
			(fsiz <= MAX_SIZEB))
		  {
			dimen[0] = lpars.xdetwid; dimen[1] = lpars.xdetht;
			return dtyp;
		  }
	  }
	if ((NULL != (fpi = xgopen("multiref", "r"))) &&
	 (sizeof (CRYSTAL) == xgread((void *)&lxhd, XG_CRYSTAL, fpi)))
	  {
		(void)fclose(fpi);	fpi = NULL;
		dtyp = lxhd.xt_dettype; ftyp = dtyp & 0377; fsiz = dtyp >> 8;
		if ((0 < ftyp) && (ftyp <= MAX_TYPEB) && (0 < fsiz) &&
			(fsiz < MAX_SIZEB))
		  {
			dimen[1] = detdimen(fsiz);
			if (ftyp == 1) dimen[0] = 256;	/* SDMS case */
			else if ((ftyp != 9) && (ftyp != 17))
				dimen[0] = dimen[1]; /*square */
			else if (ftyp == 17) /* Aviex 160x80 mm */
			  {
				dimen[0] = 4168;	dimen[1] = 2084;
			  }
			else if (ftyp == 18) /* Aviex 160x160mm */
			  {
				dimen[0] = dimen[1] = 4168;
			  }
			else {	/* one MacScience type is 2048x2500:*/
				if (dimen[1] == 2048) dimen[0] = 2500;
				else if (dimen[1] == 2500) dimen[0] = 2048;
				else dimen[0] = 1600;	/* yes? */
			     }
			return dtyp;
		  }
	  }
	/* if we get all the way to here, we're stuck. Return 0. */
	dimen[0] = dimen[1] = 1024;
	return 0;
}

void xgensteps(DATARUN *datr, double flval)
{ /*	This defines both versions of the step-size in X-GEN */
	datr->run_stepreal = flval;
	flval *= 96.;	flval += (flval < 0. ? -0.49999 : 0.49999);
	datr->run_stepsize = flval;
}

void makedrun(DATARUN *pcsdr, int iscvtsaxii)
{ /*	This finds a DATARUN-style header starting at `base',
	but in PCS format.  It reformats the monster appropriately. */
	register int	i;
	register char	*cp;
	xglong		rawval;
	float		*flp, *flpe;
	double		maxtwoth;
	xgulong		*ulp;

#ifdef LITTLEND
	u_short		*usp, *uspe;
	xgulong		*ulpe;

	for (usp = (u_short *)&(pcsdr->run_id), uspe =
	 (u_short *)&(pcsdr->run_temp); usp <= uspe; usp++) *usp = uswab(*usp);
	for (ulp = (xgulong *)&(pcsdr->run_ninteg), ulpe =
	 (xgulong *)&(pcsdr->run_ssqmax); ulp <= ulpe; ulp++) *ulp =swawb(*ulp);
#endif
	pcsdr->run_code[7] = '\0';
	for (i = 64, cp = pcsdr->run_notes; i--; cp++)
		if (*cp < 040) *cp = ' ';
	pcsdr->run_notes[63] = '\0';
	flp = &(pcsdr->run_lambda);	flpe = &(pcsdr->run_ssqmax);
	for (ulp = (xgulong *)flp; flp <= flpe; flp++, ulp++)
		*flp = (double)(*ulp);
	/* two special problems with the "converted SAXII frame header" fmt:*/
	if (iscvtsaxii)
	  {
		pcsdr->run_lambda = DEF_LAMBDA;	/* Siemens puts out "6" here?*/
		xgensteps(pcsdr, -((double)(pcsdr->run_stepsize)) / 60.);
	  }
	/* create a believable wavelength if necessary */
	if (pcsdr->run_lambda < 0.01 || pcsdr->run_lambda > 10.)
		pcsdr->run_lambda = DEF_LAMBDA;
	pcsdr->run_nshifts = abs(pcsdr->run_nshifts) & ~01;
	for (i = 0; i < 3; i++)
	   {
		pcsdr->run_startpos[i] *= 0.01;	pcsdr->run_d0pos[i] *= 0.1e-3;
		pcsdr->run_d0ang[i] *= 0.01;	pcsdr->run_d1pos[i] *= 0.1e-3;
	   }
	/* one adu per photon, since this is a nearly Poisson detector */
	pcsdr->run_aduperph = pcsdr->run_eperadu = pcsdr->run_eperph = 1.;
	pcsdr->run_fmmin = 1;	pcsdr->run_expos *= 0.01;
	pcsdr->run_fmmax = 360;	/* wild guess */
	/* fix the distance in the PCS V.2.0 and earlier versions...  This
	is an ugly kludge, and it will only work for distances between
	9.8304 cm and 16.3840 cm. */
	if (pcsdr->run_d0pos[0] < 3.2768)
	  {
		rawval = pcsdr->run_d0pos[0] * 1.e4;
		pcsdr->run_d0pos[0] = 0.1e-3 * (double)(rawval + 131072L);
	  }
	else rawval = -1000000L;	/* flag meaning reslim is ok */
	if (rawval == -1000000L)
	  {
		pcsdr->run_reslim *= 0.01; pcsdr->run_ssqmax *= 0.1e-4;
	  }
	else {
		maxtwoth = fabs(pcsdr->run_d0ang[0] * RADC) +
			atan(((double)pcsdr->run_width) *
			pcsdr->run_d1pos[1] * 0.05/ pcsdr->run_d0pos[0]);
		if (maxtwoth < 0.01) maxtwoth = 0.01;
		maxtwoth = sin(maxtwoth / 2.);
		pcsdr->run_reslim = 0.5 * pcsdr->run_lambda / maxtwoth;
		pcsdr->run_ssqmax = maxtwoth * maxtwoth;
	     }
	/* we're better off with a guessed nonzero stepsize than a zero */
	if (fabs(pcsdr->run_stepreal) < FUZZ) pcsdr->run_stepreal = -0.25;
}

int strcend(char *st1, char *st2)
{ /*	This does a strcmp on the last n bytes of st1 and st2,
	where n is length of st2 */
	int	sl1, sl2;
	char	*ss1;

	sl1 = strlen(st1);	sl2 = strlen(st2);
	if (sl2 < sl1)
	  {
		ss1 = st1 + (sl1 - sl2);	return strcmp(ss1, st2);
	  }
	else	return strcmp(st1, st2);
}

void raxis_swap(RAXISHEAD *rhp, int direction)
{ /*	This performs word- and byte-swapping on an R-Axis header.
	This is a separate routine because we call it when we write, too! */
	float	*flp;
	xglong	*ihp;

	for (flp = &(rhp->cell0[0]); flp <= &(rhp->cell0[5]); flp++)
		flcvt((void *)flp, direction);
	flcvt((void *)&(rhp->emos0), direction);
	flcvt((void *)&(rhp->wave0), direction);
	flcvt((void *)&(rhp->twoth0), direction);
	flcvt((void *)&(rhp->dist0), direction);
	flcvt((void *)&(rhp->xkv0), direction);
	flcvt((void *)&(rhp->xma0), direction);
	flcvt((void *)&(rhp->fi0), direction);
	flcvt((void *)&(rhp->fis0), direction);
	flcvt((void *)&(rhp->fie0), direction);
	rhp->nosc0 = lswawb((xglong)rhp->nosc0, direction);
	for (flp = &(rhp->expos0); flp < &(rhp->duma0[2]); flp++)
		flcvt((void *)flp, direction);
	rhp->nxpix0 = lswawb((xglong)rhp->nxpix0, direction);
	rhp->nzpix0 = lswawb((xglong)rhp->nzpix0, direction);
	flcvt((void *)&(rhp->ysize0), direction);
	flcvt((void *)&(rhp->zsize0), direction);
	rhp->lrecl0 = lswawb((xglong)rhp->lrecl0, direction);
	rhp->nrec0 = lswawb((xglong)rhp->nrec0, direction);
	rhp->nst0 = lswawb((xglong)rhp->nst0, direction);
	rhp->ipno0 = lswawb((xglong)rhp->ipno0, direction);
	flcvt((void *)&(rhp->hilo), direction);
	flcvt((void *)&(rhp->fad1), direction);
	flcvt((void *)&(rhp->fad2), direction);
	for (flp = &(rhp->cell1[0]); flp <= &(rhp->cell1[5]); flp++)
		flcvt((void *)flp, direction);
	for (ihp = &(rhp->nbox1[0]); ihp <= &(rhp->nbox1[5]); ihp++)
		*ihp = lswawb((int)(*ihp), direction);
	for (flp = &(rhp->amat1[0]); flp <= &(rhp->amat1[8]); flp++)
		flcvt((void *)flp, direction);
	for (ihp = &(rhp->istat1[0]); ihp <= &(rhp->istat1[5]); ihp++)
		*ihp = lswawb((int)(*ihp), direction);
}

void vmtoduf(u_char *ufp)
{ /*	This converts a vms float to an OSF/1 float */
	u_char	iu0, iu1, iu2, iu3;

	iu0 = *ufp; iu1 = *(ufp + 1); iu2 = *(ufp + 2); iu3 = *(ufp + 3);
	*ufp = iu2; *(ufp + 1) = iu3; *(ufp + 2) = iu0;
	*(ufp + 3) = (iu0 || iu1) ? iu1 - 1 : 0;
}

int imvalue(MAIL *headp, int ix, int iy)
{ /* This returns the 16-bit value of the image with header `headp'
	found at the position (ix, iy).  It looks to see whether the image
	is encoded with 8-bit values (in which case `headp->data8' is nonzero)
	or 16-bit values (in which case `headp->data16' is nonzero).
	In the former case it looks at headp->data8[ix * height + iy],
	where `height' is the number of rows of data in the image.
	If that's < 255, it returns it.  Otherwise, it goes through the
	overflow list looks for (ix, iy).  If it finds the (ix, iy) it
	returns the value from the overflow list; otherwise it returns 255.
	We assume that the overflow list is sorted in decreasing order of Y
	and secondarily	in increasing order of x, and do a snazzy tree-search
	to find the right value fast.
	If the 16-bit values are nonzero we look at the right value and
	return it.  The possibility of encoding 16-bit overflows in a
	16-bit overflow table thus exists. */

	int		bot, j, jy, top, va, uj;
	VOFLOW		*vo0, *vop;
	VOFL16		*v16e, *v16p;
	u_short		cux, cuy, thisx, thisy;

	if (headp->data8)
	  {
		va = headp->data8[ix * headp->ydimen + iy];
		if (va < 255 || !headp->oflowcount) return va;
		jy = headp->ydimen - iy - 1;	cuy = jy;	cux = ix;
		vo0 = headp->overflow;	top = headp->oflowcount - 1;
		if (cuy < vo0->ycoord || cuy > (vo0+top)->ycoord) return 255;
		if (top <= 1) /* special case: ordinary algorithm muffs it!*/
		  {
			for (vop = vo0; vop <= vo0 + top; vop++)
			   if ((vop->ycoord == cuy) && (vop->xcoord == cux))
			     {
				if (((va = vop->value) > 32767) &&
				 (headp->improps & IPR_RAXISCOMPRESS))
					va = 32 * (va - 32768);
				return va;
			     }
			return va;
		  }
		for (bot = 0; bot != (j = (top + bot) / 2); )
		   {
			vop = vo0 + j; thisy = vop->ycoord; thisx = vop->xcoord;
			if (cuy < thisy || (cuy == thisy && cux < thisx))
				top = j;
			else if (cuy > thisy || (cuy == thisy && cux > thisx))
				bot = j;
			else {
				if (((va = vop->value) > 32767) &&
				 (headp->improps & IPR_RAXISCOMPRESS))
					va = 32 * (va - 32768);
				return va;
			     }
		   }
		vop = (j ? vo0 + j + 1 : vo0 );
		if ((vop->ycoord == cuy) && (vop->xcoord == cux))
		  { /* R-Axis compression shouldn't produce value == 65535 */
			if (vop->value == 65535)
			  {
				if (headp->count1)
				  {
					for (j = 0, v16p = headp->oflow16;
					 j < headp->count1; j++, v16p++)
					   {
						if ((v16p->olxcoord == cux) &&
						 (v16p->olycoord == cuy))
							return v16p->olvalue;
					   }
					return vop->value;
				  }
				else return vop->value;
			  }
			else {
				if (((va = vop->value) > 32767) &&
				 (headp->improps & IPR_RAXISCOMPRESS))
					va = 32 * (va - 32768);
				return va;
			     }
		  }
		else return va;
	  }
	else if (headp->data16)
	  {
		va = headp->data16[ix * headp->ydimen + iy];
		if ((va < 65535) || (!(headp->oflowcount) && !(headp->count1)))
		  {
			if ((va > 32767) &&
			 (headp->improps & IPR_RAXISCOMPRESS))
				va = 32 * (va - 32768);
			return va;
		  }
		jy = headp->ydimen - iy - 1;	cuy = jy;	cux = ix;
		if (headp->oflowcount) /* overflows stored in packed fmt */
		  {
			vo0 = headp->overflow;	top = headp->oflowcount - 1;
		    if (cuy <vo0->ycoord || cuy > (vo0+top)->ycoord) return va;
		    if (top <= 1) /* special case: algorithm muffs it!*/
		     {
			for (vop = vo0; vop <= vo0 + top; vop++)
			   if ((vop->ycoord == cuy) && (vop->xcoord == cux))
				return 4 * vop->value + va;
			return va;
		     }
		   for (bot = 0; bot != (j = (top + bot) / 2); )
		     {
			vop = vo0 + j; thisy = vop->ycoord; thisx = vop->xcoord;
			if (cuy < thisy || (cuy == thisy && cux < thisx))
				top = j;
			else if (cuy > thisy || (cuy == thisy && cux > thisx))
				bot = j;
			else	return 4 * vop->value + va;
		    }
		   vop = (j ? vo0 + j + 1 : vo0 );
		   return ((vop->ycoord == cuy && vop->xcoord == cux) ?
			4 * vop->value + va : va);
		  } /* close of "if (headp->oflowcount)" clause */
		else {
			v16e = headp->oflow16 + headp->count1;
			for (uj = 0, v16p = headp->oflow16;
			 (v16p < v16e) && (uj < headp->count1); uj++, v16p++)
			   {
				if ((v16p->olxcoord == cux) &&
				 (v16p->olycoord == cuy))
					return v16p->olvalue;
			   }
			return va;
		     }
	  }
	return 0;	/* weird circumstance: shouldn't ever happen */
}

int ofycmp(VOFLOW *of0, VOFLOW *of1)
{ /*	Comparison routine for sorting overflows */
	short	difv;

	if ((difv = (short)of0->ycoord - (short)of1->ycoord) < 0) return -1;
	else if (difv > 0) return 1;
	return (of0->xcoord < of1->xcoord ? -1 : 1);
}

IMAGEWORK *iminfo(int opno)
{ /* if opno==0 this deletes an IMAGEWORK structure. Otherwise
	it returns a pointer to one. The other bits in opno specify memory:
	opno & 02:	create memory for iminfo structure itself if needed
	opno & 04:	create memory for iminfo->imw_buf if needed
	opno & 010:	create memory for iminfo->imw_drp if needed
	opno & 020:	create memory for iminfo->imw_darkp if needed
	opno & 040:	create memory for iminfo->imw_stabu if needed */
	char			*dname;
	int			darkflags, ix;
	static char		imimsg[] = "locimw in iminfo";
	static char		imibuf[] = "locimw->imw_buf in iminfo";
	static char		imidrp[] = "locimw->imw_drp in iminfo";
	static char		imidarkp[] = "locimw->imw_darkp in iminfo";
	static char		imista[] = "locimw->imw_stabu in iminfo";
	static MAIL		*savehp;
	static DATARUN		*savedr;
	static IMAGEWORK	*locimw;

	if ((!opno) && (NULL != locimw))
	  {
		memzap((AVOID **)&(locimw->imw_buf), imibuf);
		memzap((AVOID **)&(locimw->imw_drp), imidrp);
		if (NULL != locimw->imw_darkp)
		  {
			memzap((AVOID **)&((locimw->imw_darkp)->runheader),
			 "darkp->runheader in iminfo");
			(locimw->imw_darkp)->runheader = NULL;
			memzap((AVOID **)&(locimw->imw_darkp), imidarkp);
		  }
		memzap((AVOID **)&(locimw->imw_stabu), imista);
		memzap((AVOID **)&locimw, imimsg);	return NULL;
	  }
	if ((NULL != locimw) && !(opno & 01)) /* selective memory releases */
	  {
		if (opno & 04)
		  {
			memzap((AVOID **)&(locimw->imw_buf), imibuf);
			locimw->imw_bsize = 0;
		  }
		if (opno & 010)
			memzap((AVOID **)&(locimw->imw_drp), imidrp);
		if ((opno & 020) && (NULL != (locimw->imw_darkp)))
		  {
			memzap((AVOID **)&((locimw->imw_darkp)->runheader),
			 "darkp->runheader in iminfo");
			(locimw->imw_darkp)->runheader = NULL;
			memzap((AVOID **)&(locimw->imw_darkp), imidarkp);
		  }
		if (opno & 040)
			memzap((AVOID **)&(locimw->imw_stabu), imista);
		return locimw;
	  }
	if (opno & 077) /* create locimw if needed any time opno & 01 is on */
	  {
		if (NULL == locimw) /* create memory for IMAGEWORK */
		 locimw = (IMAGEWORK *)callop(1, sizeof (IMAGEWORK), imimsg);
	  }
	if (opno & 04) /* create memory for image-memory working buffer */
	  {
		if (NULL == locimw->imw_buf)
		  {
			if (locimw->imw_bsize < 4 * DEFBLOCK)
				locimw->imw_bsize = 4 * DEFBLOCK;
			if (NULL == (locimw->imw_buf = (u_char *)mallop(
			 locimw->imw_bsize, imibuf)))
			  {
				(void)iminfo(0);	return NULL;
			  }
			zeromem((AVOID *)(locimw->imw_buf), locimw->imw_bsize);
		  }
	  }
	if (opno & 010) /* create memory for DATARUN structure */
	  {
		if (NULL == locimw->imw_drp)
		  {
			if (NULL == (locimw->imw_drp = (DATARUN *)callop(
			 1, sizeof (DATARUN), imidrp)))
			  {
				(void)iminfo(0);	return NULL;
			  }
		  }
	  }
	if (opno & 020) /* create memory for dark-current MAIL structure */
	  {
		if (NULL == locimw->imw_darkp)
		  {
			if (NULL == (locimw->imw_darkp = (MAIL *)callop(
			 1, sizeof (MAIL), imidarkp)))
			  {
				(void)iminfo(0);	return NULL;
			  }
			/* then read in the dark-current image itself */
			darkflags = GRIPE_IF_SHORT | GET_RUNHEADER;
			dname = trname("DARKIMAGE");
			savehp = locimw->imw_lhd;
			savedr = locimw->imw_drp;
			locimw->imw_lhd = locimw->imw_darkp;
			locimw = iminfo(011);
			if (FRAME_OK != (ix = read_image(dname, darkflags,
			 locimw->imw_lhd)))
			  {
				fprintf(stdout,
			 "Error %6d in reading dark-current image\n", ix);
			  }
			locimw->imw_lhd = savehp; locimw->imw_drp = savedr;
		  }
	  }
	if (opno & 040) /* create memory for stabu structure */
	  {
		if (NULL == locimw->imw_stabu)
		  {
			if (NULL == (locimw->imw_stabu = (struct stat *)callop(
			 1, sizeof (struct stat), imista)))
			  {
				(void)iminfo(0);	return NULL;
			  }
		  }
	  }
	return locimw; /* return the pointer */
}

ino_t record_inode(char *nam)
{ /* This returns the inode associated with the file `name' */
	ino_t	retv;
	struct stat	stb;

	retv = 0; if (0 != stat(nam, &stb)) return retv;
	return stb.st_ino; 
}

int read_box(char *name, int flags, MAIL *mailp, int xylim[], int *xybu)
{ /*	This reads an image from the file `name'.  It returns results for its
	header in the MAIL structure `mailp' and the contents of a
	rectangular box of data bounded by (xylim[0],xylim[2]) and
	(xylim[1],xylim[3]) in the one-dimensional vector `xybu'.
	As of 17 Aug 95 the vector `xybu' is 4 bytes/pixel.
	Sep 03: allow for compressed images via call to xgaccess().
	Jan 04: check xylim to make sure they're not out of bounds. */
	int	i, locflags, lx, ly, xmid, ymid, *ip;
	ino_t	saveino;

	locflags = (flags & ~JUST_HEADER) | JUST_HEADPFLOWS | LEAVE_OPEN;
	saveino = mailp->quadrant;	/* remember the inode (or the Alamo)*/
	mailp->quadrant = record_inode(name);	/* now we look it up */
	if ((mailp->quadrant == 0) || (mailp->quadrant != saveino))
	  { /* Now we read the whole thing in */
		if (0 != (i = xgaccess(name, 4, 1)))
		  {
			fprintf(stderr, "xgaccess(%s) returned %d\n", name, i);
			return FRAME_BADOPEN;
		  }
		locflags &= ~JUST_HEADPFLOWS;
		if (FRAME_OK != (i = read_image(name, locflags, mailp)))
		  {
			fprintf(stderr,
			  "Failure %d in read_image in read_box\n", i);
			return i;
		  }
		if (flags & ANNOUNCE_BOX) printf("Image %s read in\n", name);
		mailp->quadrant = record_inode(name); /* cuz we wiped it out!*/
	  }
	else if (((u_char *)0 == mailp->data8) &&
	 ((u_short *)0 == mailp->data16))
	  {
		if (0 != xgaccess(name, 4, 1)) return FRAME_BADOPEN;
		xmid = (xylim[0] + xylim[1]) / 2;
		ymid = (xylim[2] + xylim[3]) / 2;
		printf("Forced reread of inode %d = %s near [%4d, %4d]\n",
		 (int)(mailp->quadrant), name, xmid, ymid);
		if (FRAME_OK != (i = read_image(name, locflags, mailp)))
		  {
			fprintf(stderr,
 "Failure %d in read_image in middle start of read_box\n", i); return i;
		  }
		mailp->quadrant = record_inode(name); /* cuz we wiped it out!*/
	  }
	/* now we just use the tried-and-true approach.
	 We exercise caution here in picking the limits */
	for (lx = 0, ly = 2; lx <= 1; lx++, ly++)
	   {
		if (xylim[lx] < 0) xylim[lx] = 0;
		else if (xylim[lx] >= (int)(mailp->xdimen))
			xylim[lx] = mailp->xdimen - 1;
		if (xylim[ly] < 0) xylim[ly] = 0;
		else if (xylim[ly] >= (int)(mailp->xdimen))
			xylim[ly] = mailp->xdimen - 1;
	   }
	for (ly = xylim[2], ip = xybu; ly <= xylim[3]; ly++)
	   for (lx = xylim[0]; lx <= xylim[1]; lx++, ip++)
		*ip = imvalue(mailp, lx, ly);
	return FRAME_OK;
}

int count65k(MAIL *headp)
{ /*	This looks at an image with header pointer *headp and
	returns the number of pixels that have been recorded as
	having 65535 counts in them, whether or not they've been
	defined as being overflows. For efficiency we look directly
	at the data array in the 16-bit case rather than going
	through the abstraction of the "imvalue" call. */
	u_short	*usp;
	int	i, isize, n65k, nfl, novers;
	VOFLOW	*vofp;

	if (NULL == headp) return 0;
	if (headp->data8)
	  {
		if (!(novers = headp->oflowcount)) return 0;
		for (vofp = headp->overflow, nfl = n65k = 0;
		 nfl < novers; vofp++, nfl++)
			if (vofp->value >= 65535) n65k++;
		return n65k;
	  }
	else {
		isize = headp->ydimen * headp->xdimen;
		for (n65k = i = 0, usp = headp->data16; i < isize; i++, usp++)
			if (65535 == *usp) n65k++;
		return n65k;
	     }
}

void saxadjust(MAIL *mp, REFPARS *rprp, int frmno)
{ /* This adjusts the value of the motor in motion according to
	a convention used in Saxi data */
	int	mot, swformat, tformat;
	double	frmd;
	DATARUN	*ruhd;

	if ((NULL == (ruhd = mp->runheader))) return;
	tformat = mp->format;	swformat = swawb(tformat);
	if ((tformat == FMT_SAXII) || (tformat == FMT_SAXII_16) ||
	 (swformat == FMT_SAXII) || (swformat == FMT_SAXII_16))
	  {
		mot = (ruhd->run_motor == 1 ? 0 : 2);
		if (fabs(ruhd->run_stepreal) > FUZZ)
			frmd = ruhd->run_stepreal;
		else frmd = ((double)ruhd->run_stepsize)/ 96.;
		rprp->xstart[mot] -= (double)frmno *frmd;
	  }
}

int israxis(int fomt)
{ /* This returns true if the format is a recognized R-Axis format */
	int	swfomt;

	swfomt = swawb(fomt);
	return ((fomt == FMT_RAXIS) || (swfomt == FMT_RAXIS)
	 || (fomt == FMT_RAXIS6) || (swfomt == FMT_RAXIS6));
}

/* end of imageutil.c */
