/* imhead.c -- Routines for dealing with the headers of XGEN images

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	11 Jun 2007: improved support for Aviex SMV headers
	19 Apr 2007: float->double conversions; reorganize use of "{"
	14 Feb 2007: support for Aviex and their version of smv format
	11 Aug 2006: manage byteswapped formats without specific defines
	 2 Aug 2006: add 2300-pixel Mar IPs to list of potential formats
	23 Jul 2006: value of drp->run_motor is changed to 0 or 2
		depending on whether omega or phi is actually filled in;
		add extra header info for ADSC/FAST-style headers as comments
	22 Jul 2006: clean up use of *motorflags in add_to_head
	14 Jul 2006: deal with ADSC images for which no motor-in-motion
		is specified.
	13 Jul 2006: recognize additional motor-in-motion identifier
		and re-set _or_ clear the byte-order according to BYTE_ORDER
	24 Jun 2006: include xgimage.h; the structures are there
	11 Jun 2006: add2notes moved here so it can be static
	 1 Jun 2006: nametonum() moved from here to image.c
	 1 Mar 2006: nametonum() and moreheader() written and used;
		Mar IP headers read out more carefully
	 2 Jan 2006: make allocation messages for overflows more efficient
	 5 Dec 2005: imhs structure replaced with calls to iminfo()
	24 Nov 2005: individual static globals replaced with imhs structure
	20 Nov 2005: inline argument declarations and strict prototyping
	10 Nov 2005: replace specific time-formatting calls with timecvt calls
	28 Jun 2005: read & write headers explicitly using 128-byte portion of
		MAIL structure, not the entire structure
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	26 Aug 2004: routines used only here are declared as static; cosmetics
	13 Aug 2004: use run_fmmin,run_fmmax to record frame limits of run
	11 Aug 2004: several more routines moved to imageutil.c
	29 Jun 2004: cosmetic changes
	10 Jun 2004: fasthead(): reverse the sign of the stepsize
		even if the motor in motion is phi.
	21 May 2004: set verbosity up in mchread() call if ANNOUNCE set
	22 Apr 2004: set IPR_RAXISCOMPRESS bit for R-Axis images
			collected with R-Axis headers
	21 Mar 2004: change cutoffs for distinguishing absolute mm values
		from pixel values in beam-center calc in fasthead()
	24 Jan 2004: MAIL element count2 used to count binning;
		various former static globals now hidden in DATARUN structure;
		nreserved, thisdtr made static locals
	21 Jan 2004: license message inserted
	 4 Jan 2004: unnecessary to rescale run->d1ang[i];
		set up drup->run_aduperph, run_eperadu, run_eperph values;
		explicitly null out runhdrold in read_header()
	23 Nov 2003: when detector center is specified in fasthead(),
		we record ax0 and ay0 in cm, not pixels.
	 8 Nov 2003: cosmetic changes
	 3 Nov 2003: handle case where byteorder is ambiguous in CCD files
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	30 Jul 2003: handle header info about beam-center better
	28 May 2003: handle RAXIS_COMPRESSION_RATIO properly
		and make a trivial change in the way IPXEL_SIZE is handled
	 1 Apr 2003: add "ROTATION" as a valid keyword for D*Trek images
	13 Mar 2003: set mp->elap100ths to 100 if it's <= 0.
	11 Jan 2003: set aduperphot to proper values: 4 for R-Axis and APS1;
		1.8 for ADSC and Mar
	 7 Jan 2003: recognize RAXIS6 and swapped formats
	 5 Jan 2003: reorganized source-file header
	30 Dec 2002: compressed (.Z and .gz) files correctly handled
	10 Oct 2002: add routines strcend, add2notes,
		add extra commands to add_to_head();
		use them in managing headers from APS1 detectors;
		for long filenames, get rid of start of name, not the end.
	25 Mar 2002: removed MSI attachments
	15 Jun 1999: in read_header, return errors for situations where the
		file is the wrong type or too small or too large
	15 Dec 1997: call mar_head() for any header.
	15 Aug 1997: in read_header, change memory-management techniques
		for overflows.
	 8 Nov 1996: enable reading images from R-AxisIV detectors
		and the recent CCD detector types
	16 Oct 1995: further corrections of code for `lswawb'
		to handle byteswapping
	13 Oct 1995: add 2nd argument to `lswawb' calls because of
		need for symmetry to fix bug
	30 Sep 1995: use overphot to override system-dependent value
		for `aduperphot'
	24 Aug 1995: in mar_head(),
		Mar data with 16-bit overflows handled correctly;
		header length should be 2*ma_width.
	 6 Jul 1995: handle DIP3000 trailer format
		check for overflows before mem alloc;
		this allows for 3-D box readings
	31 May 1995: "register" -> "register int" throughout
	20 May 1995: delete incorrect IPR_IS8BIT use;
		make filetype an unsigned int
	12 May 1995: able to read FMT_DIP100 and DIP320 as headerless images
	 2 Mar 1995: trivial formatting
	16 Nov 1994: read_header has image-flags arg;
		if HEADER_ONLY, no mem alloc for image
	 4 Nov 1994: move definition of FMT_CORNELL to xgen.h;
		add code for swapped version thereof;
		save and restore mpo->datae when reading in data header
	30 Sep 1994: add `xgensteps' routine to compute stepsize as float and
		in 96ths of deg; fix SAXII-to-PCS header format problems
	23 Sep 1994: fix bugs for 1024x1024 Siemens frames
	 2 Sep 1994: 3 photons per ADU on a MAR detector
	12 Jul 1994: add 2000x2000 MAR-detector formats
	25 Jul 1994: in read_header() for Mar image-plate formats,
		we set the default photons/ADU value to 3.
	 7 Jul 1994: aduperphot = 1 for R-Axis
	15 Jun 1994: in read_header for R-Axis detectors,
		we set the # adu/photon to 1.
	14 Jun 1994: set cellsize=200 for Siemens data
	 5 May 1994: source file headers, POLYDET->XGEN,
		"xgulong" and "xglong" constructs
	16 Mar 1994: polymask added; bugfixes in polymath, xtal, calinv, etc.
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version
	
   Purpose: reading in the headers of detector images from various types
	of two-dimensional detectors.
   Cross references: image.c, imtime.c

*************************************************************************/

#define		NEED_STAT	1
#include	"xgimage.h"
#define	MIN_SDMS	(36384)
#define	MAX_SDMS	(73728)
#define	INBLOCK		512	/* "blocksize" on Cadmus */
#define	MINRESERVED		800
#define	DIP100_SIZE	5120180
#define	DIP2000_SIZE	12501024
#define	DIP3000_SIZE	18001024
#define	DIP320_SIZE	10240180
#define	FUJI_SIZE	10485760
#define	MIN_SIZE	32000
#define	MAX_SIZE	64000000
#define		MIN_DIMENSION	128
#define		MAX_DIMENSION	10000
 typedef struct {
		char	di_id[4];
		xglong	di_mtype1, di_mtype2, di_dtype;
		xglong	di_pixelsize, di_pixelsize2;
		float	di_radius;
		xglong	di_xsize, di_ysize, di_ipno;
		char	di_comment[80];
		float	di_xlambda, di_cdist;
		char	di_monochro[32];
		float	di_pttheta;
		long	di_ipx, di_ipy;
		float	di_expos, di_kv, di_ma;
		float	di_collimator, di_cweiss;
		float	di_phi1, di_phi2, di_phispeed;
		long	di_repet;
		char	di_rfu[176];
		char	di_screen[640];
		} DIPTRAILER;

extern void nametonum(MAIL *mp);

static void moreheader(MAIL *mp)
{ /* This tries to divine some header values in the absence of the
   stuff that would otherwise give those values to us. */
	DATARUN	*drp;

	nametonum(mp);	/* get the frame number */
	drp = mp->runheader;	drp->run_motor = 3;
	drp->run_startpos[0] = drp->run_startpos[1] = 0.;
	drp->run_startpos[2] = drp->run_stepreal * (float)(mp->frameno);
}

static void add2notes(char *notes, char *title, char *remark, int maxnotes)
{ /*	This adds some notes to `notes'. Since that buffer cannot
	be longer than `maxnotes' bytes, we add cautiously. */
	char	*cpi, *cpo;
	int	curlen, titlen, room, i;

	if ((NULL == notes) || (NULL == remark) || ('\0' == *remark)) return;
	for (cpi = title, titlen = 0; '\0' != *cpi; cpi++) titlen++;
	for (cpi = notes, curlen = 0; '\0' != *cpi; cpi++) curlen++;
	room = maxnotes - curlen - titlen - 2;
	if (room < 1) return; /* not worth it otherwise */
	if (curlen > 0)
	  {	*(notes + curlen) = ';';	*(notes + curlen + 1) = ' ';
		*(notes + curlen + 2) = '\0';	curlen += 2;
	  }
	if (titlen)
	  {	for (cpo = notes + curlen, cpi = title; '\0' != *cpi;
		 cpi++, cpo++) *cpo = *cpi;
		*cpo++ = ':';	*cpo++ = '\0';	curlen += titlen + 1;
	  }
	for (cpo = notes + curlen, cpi = remark, i = 0; i < room;
		i++, cpi++, cpo++)
	   {	if ('\0' == *cpi) break;	*cpo = *cpi;
	   }
	*cpo = '\0';
}

static int add_to_head(char *title, char *val, int *motorflags)
{ /*	This looks at a record of a CCD dataframe header and tries to match
	its title to a list.  If it gets a match, it keeps an appropriate value
	in the internal XGEN frame header. Nov 03: this routine returns 1 if
	it has told us about the byte order; it returns 0 otherwise.
	The third parameter specifies which motors have been specified.
	This allows us to define which motor is in motion. */
	int			i, igive, igivep, lenshort, retv, thismo;
	unsigned		ngiven;
	float			fgive[6];
	char			*locdate, *loctime;
	MAIL			*lhp;
	DATARUN			*drp;
	IMAGEWORK		*imwp;
	static char		*shortname;
	static char currmo[] =	"janfebmaraprmayjunjulaugsepoctnovdec";

	*motorflags = retv = 0;
	if (NULL == (imwp = iminfo(015))) return retv; /* memory for imwp */
	drp = imwp->imw_drp;
	/* memory for the time and date strings will come from run_recip */
	locdate = (char *)&(drp->run_recip[0][0]);
	loctime = (char *)&(drp->run_recip[1][1]);
	lhp = imwp->imw_lhd;
	if (!strcmp(title, "HEADER_BYTES"))
	  {	imwp->imw_headlen = atoi(val);
		/* NB: there may be some 512-byte-header files that have
		X and Y reversed, but my previous statement that any 512-byte
		header went with a reversed X and Y is wrong.  Thus
		the next line is commented out:
		if (imwp->imw_headlen == 512) imwp->imw_filetype &= ~01;
		... but note that we DO reverse the order with the AN_HRS
		header entry: see below. */
	  }
	else if (!strcmp(title, "BYTE_ORDER"))
	  {
#ifdef LITTLEND
		if (!strncmp(val, "big", 3))
#else
		if (!strncmp(val, "little", 6))
#endif
			lhp->improps |= IPR_BYTESWAPPED;
		else	lhp->improps &= ~IPR_BYTESWAPPED;
		retv = 1;	/* so we return a reasonable value */
	  }
	else if (!strcmp(title, "BIN"))
	  {	if (!strncmp(val, "2X2", 3)) lhp->count2 = 2;
		else lhp->count2 = 1;
	  }
	else if (!strcmp(title, "DATE")) /* format: Sat Aug 25 02:23:58 2001 */
	  {	locdate[0] = val[8];	locdate[1] = val[9];
		locdate[2] = '-';	locdate[6] = '-';
		for (i = 3; i < 6; i++)
		   {	locdate[i] = val[i+1];
			if (('A' <= locdate[i]) && (locdate[i] <= 'Z'))
				locdate[i] += 'a' - 'A';
		   }
		locdate[7] = val[22];	locdate[8] = val[23];
		locdate[9] = '\0';
		for (i = 0; i < 8; i++) loctime[i] = val[i+11];
		loctime[8] = '\0';
	  }
	else if (!strcend(title, "EXPOSE_TIMESTAMP"))
		/* format: 07/05/02 14:57:17 */
	  {	locdate[0] = val[3];	locdate[1] = val[4];
		locdate[2] = '-';	locdate[6] = '-';
		thismo = val[1] - '1';	if (val[0] == '1') thismo += 10;
		for (i = 0; i < 3; i++)
			locdate[i + 3] = currmo[thismo * 3 + i];
		locdate[7] = val[6];	locdate[8] = val[7];
		locdate[9] = '\0';
		for (i = 0; i < 8; i++) loctime[i] = val[i+9];
		loctime[8] = '\0';
	  }
	else if ((!strcmp(title, "DAY")) || (!strcmp(title, "WRITE_DATE")))
		strcpy(locdate, val);
	else if ((!strcmp(title, "NOW")) || (!strcmp(title, "WRITE_TIME")))
		strcpy(loctime, val);
	else if ((!strcmp(title, "PIXEL_SIZE")) ||
		(!strcmp(title, "IPXEL_SIZE"))) /* compensate for bug! */
	  {	lhp->cellsize = (int)(1000. * atof(val) + 0.49999);
		drp->run_d1pos[1] = drp->run_d1pos[2] = atof(val);
	  }
	else if (!strcmp(title, "SIZE1")) drp->run_width = atoi(val);
	else if (!strcmp(title, "SIZE2")) drp->run_height = atoi(val);
	else if (!strcmp(title, "GEN_KV")) drp->run_volxray = (short)atof(val);
	else if (!strcmp(title, "GEN_MA")) drp->run_curxray = (short)atof(val);
	else if ((!strcmp(title, "WAVELENGTH")) || (!strcend(title,
		"SCAN_WAVELENGTH")))  drp->run_lambda = atof(val);
	else if (!strcend(title, "SCAN_IMAGES")) drp->run_nframes = atoi(val);
	else if (!strcend(title, "ROTATION_SERIAL_NUM"))
		lhp->frameno = atoi(val);
	else if (!strcend(title, "SCAN_INCREMENT"))
	  {	drp->run_stepreal = -atof(val);
		drp->run_stepsize = (drp->run_stepreal < 0.) ?
			96. * (drp->run_stepreal - 0.49999) :
			96. * (drp->run_stepreal + 0.49999);
	  }
	else if (!strcend(title, "DETECTOR_DIMENSIONS"))
	  {	sscanf(val, "%d %d", &igive, &igivep);
		drp->run_width = igivep;	drp->run_height = igive;
	  }
	else if (!strcend(title, "GONIO_VALUES"))
	  {	ngiven = sscanf(val, "%f %f %f %f %f %f",
		 &fgive[0], &fgive[1], &fgive[2], &fgive[3],
		&fgive[4], &fgive[5]);
		drp->run_d0ang[0] = fgive[1]; drp->run_d0ang[1] = fgive[0];
		if (ngiven > 3) drp->run_d0pos[0] = 0.1 * fgive[5];
	  }
	else if (!strcend(title, "GO_K_DEG")) drp->run_startpos[1] = atof(val);
	else if (!strcend(title, "GO_P_DEG"))
	  {
		drp->run_startpos[2] = atof(val);	*motorflags |= 2;
	  }
	else if (!strcend(title, "GO_O_DEG"))
	  {
		drp->run_startpos[0] = atof(val);	*motorflags |= 1;
	  }
	else if (!strcmp(title, "SCALERS"))
	  {	ngiven = sscanf(val, "%f %f", &fgive[0], &fgive[1]);
		if (ngiven > 1) drp->run_ninteg = (int)fgive[1];
		else drp->run_ninteg = 1000000;
	  }
	else if (!strcend(title, "SPATIAL_DISTORTION_INFO"))
	  {	ngiven = sscanf(val, "%f %f %f %f", &fgive[0], &fgive[1],
			&fgive[2], &fgive[3]);
		drp->run_d0pos[2] = fgive[0]; drp->run_d0pos[1] = fgive[1];
		if (ngiven > 2)
		  {	drp->run_d1pos[2] = fgive[2];
			drp->run_d1pos[1] = fgive[3];
			lhp->cellsize = (int)(1000. * fgive[2]);
		  }
	  }
	else if ((!strcmp(title, "EXPTIMP")) || (!strcmp(title, "EXP_TIME")) ||
		(!strcend(title,"EXPOSURE_TIME")) ||
		(!strcmp(title, "TIME"))) drp->run_expos = atof(val);
	else if ((!strcmp(title, "DATASET_NAME")) || (!strcmp(title, "DERI")))
		strncpy(drp->run_code, val, 7);
	/* currently 19-ID defines AXIS as phi and ID19_SCAN_AXIS as omega,
	 and ROTATION_AXIS_NAME as omega. Go figure. Fortunately,
	 the AXIS designation happens before the ID19_SCAN_AXIS thing */
	else if ((!strcmp(title, "ROT_AXIS")) || (!strcmp(title, "SCAN")) ||
	 (!strcmp(title, "AXIS")))
	  drp->run_motor = (*val== 'p' || *val== 'P') ? 2 : 1;
	else if (!strcmp(title, "ID19_SCAN_AXIS"))
		drp->run_motor = ((*val == 'p') || (*val == 'P')) ? 2 : 1;
	else if (!strcmp(title, "DATUM"))
	  {	ngiven = sscanf(val,"%f %f %f",&fgive[0],&fgive[1],&fgive[2]);
		for (igive = 0; igive < (int)ngiven; igive++)
			drp->run_startpos[igive] = fgive[igive];
	  }
	else if ((!strcmp(title, "ROTINC")) || (!strcmp(title, "FRAME_SIZE"))
	 || (!strcmp(title, "OSC_RANGE"))) xgensteps(drp, atof(val));
	/* for now we assume that the R-Axis compression format
	 is fixed in its cutoff and multiplier. */
	else if ((!strcmp(title, "RAXIS_COMPRESSION_RATIO")))
		lhp->improps |= IPR_RAXISCOMPRESS;
	else if ((!strcmp(title, "START_POS")) || (!strcmp(title, "DATUM")))
	  {	ngiven = sscanf(val, "%f %f %f %f %f",
		 &fgive[0], &fgive[1], &fgive[2], &fgive[3], &fgive[4]);
		if (ngiven > 3) /* convert distance from mm to cm! */
		  {	drp->run_d0ang[0] = fgive[3];
			if (ngiven > 4) drp->run_d0pos[0] = fgive[4] * 0.1;
		  }
	  }
	else if ((!strcmp(title, "ROTSTART")) ||
	 (!strcend(title, "ROTATION_START")) ||
	 (!strcend(title, "OSC_START"))) drp->run_iosave = atof(val);
	else if ((!strcmp(title, "ROTEND")) ||
	 (!strcend(title, "ROTATION_END"))) drp->run_delomave = atof(val);
	else if (!strcmp(title,"ROTATION"))
	  {	ngiven = sscanf(val, "%f %f", &fgive[0], &fgive[1]);
		if (ngiven >= 1) 
		  {	drp->run_iosave = fgive[0];
			if (ngiven >= 2) drp->run_delomave = fgive[1];
		  }
	  }
	else if (!strcmp(title, "FILENAME"))
	  {	shortname = filepart(val, 0);
		lenshort = strlen(shortname);
		if (lenshort <= 15) strcpy(lhp->filename, shortname);
		else strcpy(lhp->filename, shortname + (lenshort - 15));
		lhp->filename[15] = '\0'; /* ... to be explicit */
		strncpy(drp->run_code, shortname, 7);
		if (shortname != val) shortname = filepart(NULL, 0);
	  }
	else if (!strcmp(title, "PHI"))
	  {	drp->run_startpos[2] = atof(val);	*motorflags |= 2;
	  }
	else if (!strcmp(title, "CHI")) drp->run_startpos[1] = atof(val);
	else if (!strcmp(title, "OMEGA"))
	  {	drp->run_startpos[0] = atof(val);	*motorflags |= 1;
	  }
	else if ((!strcmp(title, "TWO_THETA")) ||(!strcmp(title, "GON_THETA"))
	 || (!strcmp(title, "TWOTHETA"))) drp->run_d0ang[0] = -atof(val);
	else if ((!strcmp(title,"BEAMX")) || (!strcmp(title,"BEAM_CENTER_X")))
		drp->run_d0pos[1] = atof(val);
	else if ((!strcmp(title,"BEAMY")) || (!strcmp(title,"BEAM_CENTER_Y")))
		drp->run_d0pos[2] = atof(val);
	else if ((!strcmp(title, "CAMERA_LENGTH")) ||
	 (!strcmp(title, "GON_DIST")) || (!strcmp(title, "DISTANCE")) ||
	 (!strcend(title, "DE_D_MM"))) drp->run_d0pos[0] = atof(val) / 10.;
	else if (!strcmp(title, "FIRST")) drp->run_fmmin = atof(val);
	else if (!strcmp(title, "LAST")) drp->run_fmmax = atof(val);
	else if ((!strcmp(title, "COMMENT1")) ||
	 (!strcmp(title,"USER_COMMENT3"))) strncpy(drp->run_notes, val, 63);
	else if (!strcmp(title, "COMMENT2"))
		add2notes(drp->run_notes, "", val, 63);
	else if (!strcmp(title, "COMMENT"))
	  {	if (strncmp(val, "Original template", 17))
			add2notes(drp->run_notes, "Comment", val, 63);
	  }
	else if (!strcend(title, "DETECTOR_DESCRIPTION"))
		add2notes(drp->run_notes, "Detector", val, 63);
	else if (!strcend(title, "DETECTOR_SN"))
		add2notes(drp->run_notes, "Detector SN", val, 63);
	else if (!strcend(title, "UNIF_PED"))
		add2notes(drp->run_notes, "Pedestal", val, 63);
	else if (!strcmp(title, "BEAMLINE_NAME"))
		add2notes(drp->run_notes, "Beamline", val, 63);
	else if (!strcmp(title, "CRYSTAL_DESCRIPTION"))
		add2notes(drp->run_notes, "Crystal", val, 63);
	/* kludge: if the title "AN_HRS" is present, the data are from
	 the "green" detector, which has X and Y reversed. */
	else if (!strcmp(title, "AN_HRS")) imwp->imw_filetype &= ~01;
	/* certain keywords aren't really worth recording */
	else if ((strcmp(title, "DIM")) && (strcmp(title, "TYPE")) &&
	 (strcmp(title, "CCD_IMAGE_SATURATION")) &&
	 (strncmp(title, "BM19_", 5)) && (strncmp(title, "APS1_", 4)))
	  {
		lowercaseify(title);
		if (('a' <= *title) && (*title <= 'z')) *title -= 'a' - 'A';
		add2notes(drp->run_notes, title, val, 63);
	  }
	return retv;
}
 
static int fasthead(void)
{ /*	This reads the ascii FAST/CCD header format */
	char		*bend, *cp, *tend, *dend, *val, *locdate, *loctime;
	int		allmo, aviex, bhint, foundbyteorder, mofl, sizfil;
	size_t		ntoread;
	time_t		tl0;
	float		abspx, abspy, absvx, absvy, maxtwoth, relx, rely;
	IMAGEWORK	*imwp;
	MAIL		*lhp;
	DATARUN		*drp;

	if (NULL == (imwp = iminfo(055))) return FRAME_NO_MEMORY;
	bend = (char *)imwp->imw_buf + SIZMAIL;	imwp->imw_headlen = 0;
	/* when we call this, the 1st record should be the header length. */
	val = (char *)imwp->imw_buf + 15;
	for (dend = val; '\0' != *dend; dend++) if (*dend == ';') break;
	if ('\0' == *dend) return FRAME_BADMAIL;
	*dend = '\0';	imwp->imw_headlen = atoi(val);	*dend = ';';
	/* having determined how long the header is, we read the rest of it*/
	ntoread = imwp->imw_headlen - SIZMAIL;
	if (ntoread != imread((AVOID *)bend, ntoread)) return FRAME_BADMAIL;
	bend = (char *)imwp->imw_buf + imwp->imw_headlen;
	/* scan through the header to find some information. Each
	record is of the form TITLE=value;
	so we look for the '=' as a delimiter of the title of the
	record and a ';' as delimiter of the data value. */
	drp = imwp->imw_drp;
	lhp = imwp->imw_lhd;
	/* if we're reading a dark image, make sure the MAIL pointer
	 is pointing at the correct MAIL structure */
	if (lhp == imwp->imw_darkp) lhp->runheader = drp;
	drp->run_stepsize = drp->run_stepreal = drp->run_ninteg = 0;
	lhp->count2 = 1; /* default is no binning */
	/* nonsense values for image's starting & ending spindle positions: */
	drp->run_iosave = -9999;	drp->run_delomave = 9999;
	drp->run_fmmin = drp->run_fmmax = -1000;
	/* before working through the header elements we set the moving-motor
	 to a specifically undefined value so that we can look for a spec. */
	drp->run_motor = -1;
	zeromem((AVOID *)(drp->run_recip), 36);
	allmo = aviex = 0;
	for (cp = (char *)(imwp->imw_buf) + 2, foundbyteorder = 0; cp < bend; )
	   { /* Some Aviex image header-lines enter here with a newline: */
		if ('\n' == *cp)
		  {	cp++;	if (cp >= bend) break;
			aviex = 1;
		  }
		for (tend = cp; '\0' != *tend; tend++) if ('=' == *tend) break;
		if ('\0' == *tend) break;	/* end of header ... */
		for (dend = tend + 1; '\0' != *dend; dend++)
			if (';' == *dend) break;
		/* null-terminate the title and the string */
		*tend = *dend = '\0';	val = tend + 1;
		/* having set up the data we see if
		 the title matches one of the titles we care about */
		if (dend > val) foundbyteorder += add_to_head(cp, val, &mofl);
		allmo |= mofl;
		cp = dend + 1;	/* go to start of next record */
		/* FAST headers start the next record immediately;
		 CCD detector headers throw an extra byte in here. */
		if ((*cp < 'A') || (*cp > 'Z')) cp++;
	   }
	if ((lhp->xdimen == 1042) || (lhp->xdimen == 2084) ||
	 (lhp->xdimen == 4168)) aviex = 1;
	if (aviex) /* Aviex should be read with X fastest */
		lhp->improps |= IPR_XFASTEST;
	/* if we haven't had a keyword to define the motor in motion,
	 * then we look to see if either PHI or OMEGA was specified.
	 * If so, it's probably the motor in motion */
	if (drp->run_motor == -1)
	  {	if (allmo & 01) drp->run_motor = 1;
		if (allmo & 02) drp->run_motor = 2;
	  }
	/* furthermore, even if a motor in motion was specified,
	 if we don't have a keyword defining the value of that motor,
	 we reassign it */
	else {	if ((allmo == 01) && (drp->run_motor == 2)) drp->run_motor = 1;
		if ((allmo == 02) && (drp->run_motor == 1)) drp->run_motor = 2;
	     }
	/* annoying necessity: sometimes the byte order isn't defined in
	 the header. In that instance we'll have to divine it from the
	 statistics in the data themselves. */
	if (!foundbyteorder)
	  {	sizfil = (int)((imwp->imw_stabu)->st_size);	/* filesize */
		bhint = comparemid(lhp->filename, imwp->imw_fd, sizfil, 0);
#ifdef LITTLEND
		if (bhint == 1) lhp->improps |= IPR_BYTESWAPPED;
		else if (bhint == 2) lhp->improps &= ~IPR_BYTESWAPPED;
#else
		if (bhint == 1) lhp->improps &= ~IPR_BYTESWAPPED;
		else if (bhint == 2) lhp->improps |= IPR_BYTESWAPPED;
#endif
		/* it also turns out that files that don't specify the byte
		 order are the ones for which an omega motion is described
		 as a phi motion. Thus: */
		if (drp->run_motor > 0) drp->run_motor = 3 - drp->run_motor;
	  }
	if (lhp->count2 != 1) lhp->cellsize *= lhp->count2;
	/* the image's start and end spindle values have been stuffed
	 in drp->run_iosave and drp->run_delomave respectively */
	if ((drp->run_stepsize) == 0)
		xgensteps(drp, drp->run_delomave - drp->run_iosave);
	if (drp->run_iosave > -9998) drp->run_startpos[0] = drp->run_iosave;
	drp->run_iosave = drp->run_delomave;
	/* an unspecified axis is assumed to be omega. Let's hope that works */
	if (drp->run_motor < 0) drp->run_motor = 1;
	/* in general if omega is the operating axis, the stepsize should
	 be reversed in sign at this point. On at least some systems, this
	 is also true for phi runs */
	if ((drp->run_motor == 1) || (drp->run_motor == 2))
	  {	drp->run_stepreal = -drp->run_stepreal;
		drp->run_stepsize = -drp->run_stepsize;
	  }
	/* we've kept start & end frame #s in run_fmmin & run_fmmax: */
	if (drp->run_fmmin > -999 && drp->run_fmmax > -999)
	  drp->run_nframes = (drp->run_fmmax - drp->run_fmmin) + 1;
	drp->run_nchams = 1;	drp->run_temp = 100; /* wild guess */
	lhp->xdimen = drp->run_width;	lhp->ydimen = drp->run_height;
	/* if we find anything useful in d0pos[1] and d0pos[2],
	 the values currently there may be absolute positions of the beam
	 in cm from the corner. We convert those to relative pixel values.
	 This is a change as of July 03: we're more selective about
	 whether we apply this correction. Mar04: cutoff is 360mm, not 80mm*/
	absvx = fabs(drp->run_d0pos[2]); absvy = fabs(drp->run_d0pos[1]);
	abspx = drp->run_d1pos[2];	   abspy = drp->run_d1pos[1];
	if ((absvx > FUZZ) && (absvy > FUZZ) &&
	 (abspx > FUZZ) && (abspy > FUZZ))
	  {	/* absolute values in mm */
		if ((absvx < 360.) && (absvy < 360.))
		  {
			relx = lhp->xdimen / 2 - 1 - 
				drp->run_d0pos[2] / abspx;
			rely = drp->run_d0pos[1] / abspy - lhp->ydimen / 2;
		  }
		else /* absolute values in pixels */
		  {
			relx = absvx - lhp->xdimen / 2;
			rely = absvy - lhp->ydimen / 2;
		  }
		drp->run_d0pos[1] = relx * abspx * 0.1;
		drp->run_d0pos[2] = rely * abspy * 0.1;
	  }
	/* if the dimensions are a multiple of 1056, we recognize this as
	 a CCD image with overscan and dark-reference lines in it.
	 A specific mode of processing is required for such data. */
	if (!(lhp->xdimen % 1056)) lhp->improps |= IPR_HASPEDESTAL;
	/* we record a pseudo-exposure time whose value depends on what the
	header contains.  If the header has a SCALERS line (as at Brookhaven)
	the pseudo-exposure time is proportional to the ion current and
	(if specified) the actual exposure time.  If the header lacks a
	SCALERS line we use either the real exposure time or a made-up value */
	if (drp->run_ninteg)
	  {	if (drp->run_expos > FUZZ) lhp->elap100ths = (1.E-3) *
		 ((float)drp->run_expos) * ((float)drp->run_ninteg);
		else lhp->elap100ths = ((float)drp->run_ninteg) * 1.E-2;
		if (lhp->elap100ths < FUZZ) lhp->elap100ths = 1000;
	  }
	else if (drp->run_expos > FUZZ)
		lhp->elap100ths = 100. * drp->run_expos + 0.49999;
	else	lhp->elap100ths = 1000;
	lhp->eventcount = 1000000;	/* so there's something there*/
	drp->run_ninteg = 0;	/* we haven't really integrated anything yet */
	/* these formats don't support overflows. */
	lhp->oflowcount = 0;
	lhp->accumsecs = drp->run_expos;
	/* if the user has specified a time-stamp, we remember it */
	locdate = (char *)&(drp->run_recip[0][0]);
	loctime = (char *)&(drp->run_recip[1][1]);
	if ('\0' != locdate[0] && '\0' != loctime[0] &&
	 strlen(locdate) == 9 && strlen(loctime) == 8)
	  {	locdate[9] = locdate[7]; locdate[10] = locdate[8];
		if (locdate[9] >= '6')
		  {	locdate[7] = '1';	locdate[8] = '9';
		  }
		else {	locdate[7] = '2';	locdate[8] = '0';
		     }
		locdate[11] = ' ';	locdate[12] = '\0';
		strcat(locdate, loctime);
		(void)timecvt(lhp->datetime, locdate,
			NULL, &tl0, 6, 0);
	  }
	/* to minimize confusion, we re-set the run_recip matrix to zero */
	zeromem((void *)(drp->run_recip), 36);
	if (drp->run_d0pos[0] < FUZZ) drp->run_d0pos[0] = 13.; /* nominal*/
	if (drp->run_lambda < FUZZ) drp->run_lambda = 1.82;	/* Fe-55 */
	/* roughly 0.1mm per pixel... */
	if (drp->run_d1pos[1] < FUZZ)
	  {	if (lhp->cellsize > 0) drp->run_d1pos[1] = drp->run_d1pos[2] =
		 1.0E-3 * (float)(lhp->cellsize);
		else drp->run_d1pos[1] = drp->run_d1pos[2] = 0.1;
	  }
	/* this estimate of two-theta(max) should be taken skeptically */
	maxtwoth = fabs(drp->run_d0ang[0] * RADC) + atan(((double)
	 (drp->run_width)) * drp->run_d1pos[1] * 0.05/drp->run_d0pos[0]);
	if (maxtwoth < 0.01) maxtwoth = 0.01;
	maxtwoth = sin(maxtwoth / 2.);
	if (maxtwoth > FUZZ) drp->run_reslim =
		0.5 * drp->run_lambda / maxtwoth;
	drp->run_ssqmax = maxtwoth * maxtwoth;	return FRAME_OK;	
}

static int pcs_to_mail(MAIL *mp)
{ /*	This converts a pcs-style mailbox to a X-GEN-internal MAIL structure.
	It also handles the additional reads required to get the tail end
	of the PCS mailbox. */
	int		i, swfmt, tfmt;
	size_t		nread;
	PMAIL		pmbox;
	DATARUN		*drp;
	IMAGEWORK	*imwp;

	cpymem((AVOID *)(&pmbox), (AVOID *)mp, sizeof (MAIL));
	/* reality check: real `VMS' files are always 512x512,
	 unless this is a FRAMBO-generated 1024x1024 (or 2048x2048 ?) file */
	if ((INBLOCK <= pmbox.pdimen) && (pmbox.pdimen <= 4 * INBLOCK))
		mp->xdimen = pmbox.pdimen;
	else	mp->xdimen = INBLOCK;
	mp->ydimen = mp->xdimen;	/* PCS files are always square */
	mp->quadrant = pmbox.pquadrant;
	mp->eventcount = pmbox.peventcount; mp->oflowcount = pmbox.p8bitoflow;
	mp->zmax = pmbox.pzmax;
	mp->zmaxx = (u_short)pmbox.pzmaxx; mp->zmaxy = (u_short)pmbox.pzmaxy;
	mp->zdisp = pmbox.pzdisp;
	mp->ontime = pmbox.pontime; mp->late = pmbox.plate;
	mp->datetime[0] = pmbox.pdatetime[0];
	mp->datetime[1] = pmbox.pdatetime[1];
	mp->accumsecs = pmbox.paccumsecs; mp->elap100ths = pmbox.pelap100ths;
	mp->cellsize = pmbox.pcellsize;
	if (mp->cellsize < 100) mp->cellsize = 200 / (mp->xdimen / INBLOCK);
	mp->nonzpix = pmbox.pnonzpix; mp->over64k = pmbox.p16bitoflow;
	mp->frameno = (xgulong)pmbox.pframeno;
	/* most of the bits in mp->improps are already set right.
	 But we do need to define the ones that get us reading properly.
	 This might or might not be right!! */
#ifdef LITTLEND
	if ((mp->format == FMT_PCS_8) || (mp->format == FMT_PCS_16))
	 mp->improps &=
	  ~(IPR_XFASTEST | IPR_STARTBOTTOM | IPR_STARTRIGHT | IPR_BYTESWAPPED);
	else mp->improps |= IPR_XFASTEST | IPR_STARTRIGHT;
#else
	if ((mp->format == FMT_PCS_8) || (mp->format == FMT_PCS_16))
		mp->improps |= IPR_XFASTEST | IPR_STARTRIGHT;
	else	mp->improps |= IPR_BYTESWAPPED;
#endif
	/* position filepointer at end of the first 512-byte block */
	if (NULL == (imwp = iminfo(015))) return FRAME_NO_MEMORY;
	if (-1L == lseek(imwp->imw_fd, (off_t)INBLOCK,0)) return FRAME_BADMAIL;
	/* read in the filename and (unused) character data */
	nread = INBLOCK;
	if (nread != imread((AVOID *)(imwp->imw_buf), nread))
		return FRAME_BADMAIL;
	for (i = 16; i--; ) mp->filename[i] = imwp->imw_buf[i];
	mp->filename[15] = '\0';	/* force null-termination */
	drp = imwp->imw_drp;
	tfmt = imwp->imw_filetype;	swfmt = swawb(tfmt);
#ifdef LITTLEND
	if ((swfmt == FMT_PCS_8) || (swfmt == FMT_PCS_16))
#else
	if ((tfmt == FMT_PCS_8) || (tfmt == FMT_PCS_16))
#endif
	  {	i = !strncmp((char *)(imwp->imw_buf + 16), "Mailbox from", 12);
		makedrun((DATARUN *)(void *)(imwp->imw_buf + 256), i);
		cpymem((AVOID *)drp, (AVOID *)(imwp->imw_buf + 256),
			sizeof (DATARUN));
	  }
	else zeromem((AVOID *)drp, sizeof (DATARUN));
	mp->runheader = drp;
	moreheader(mp);	return FRAME_OK;
}

static int ccd_header(MAIL *mp)
{ /*	this reads the header of a CCD file containing the following items:
 variable	meaning on CCD system		location	used here?
 'hdrlen'	(length of the header)		hbuf[0]		no
 'nmat'		(size of data matrix)		hbuf[1]		no
 'nrows'	(number of detector rows)	hbuf[2]		YES
 'ncols'	(number of detector columns)	hbuf[3]		YES
 'dlen'		(length of data structure)	hbuf[4]		not yet
 'dtyp'		(type of data frame)		hbuf[5]		not yet
 'shft'		(? ? ?)				hbuf[6]		no
	It returns FRAME_OK if the read works, FRAME_BADMAIL otherwise. */
	size_t		nread;
	time_t		tl0;
	u_char		fromvms;
	u_short		hbuf[32], *uspo, *uspe, *uspi;
	IMAGEWORK	*imwp;

	if (NULL == (imwp = iminfo(055))) return FRAME_NO_MEMORY;
	fromvms = ((imwp->imw_filetype == FMT_CCD_VMS) ||
		(imwp->imw_filetype == FMT_CCD_VMSALT));
	if (fromvms) cpymem((AVOID *)hbuf, (AVOID *)mp, 16);
	else for (uspo = &(hbuf[0]), uspe= &(hbuf[6]), uspi = (u_short *)mp;
		uspo <= uspe; uspi++, uspo++) *uspo = uswab(*uspi);
	zeromem((AVOID *)mp, sizeof (MAIL)); /* most entries zeroed */
	mp->xdimen = hbuf[3]; mp->ydimen = hbuf[2];
	mp->accumsecs = mp->elap100ths = 10000; /* nominal values */
	mp->improps |= IPR_NEEDCOUNT;
	nread = mp->xdimen - SIZMAIL;	/* read remainder of this segment */
	if (nread != imread((AVOID *)imwp->imw_buf,nread)) return FRAME_BADMAIL;
	nread = INBLOCK;
	/* if the frame header lacks a time variable,
	use the modification time for the file as found in the inode header */
	if ((FMT_CCD_ALT == imwp->imw_filetype) ||
	 (FMT_CCD_VMSALT == imwp->imw_filetype))
	  {	(void)timecvt(mp->datetime, NULL, NULL, 
		 &((imwp->imw_stabu)->st_mtime), 1, 0);
#ifdef LITTLEND
		imwp->imw_filetype = lswawb(imwp->imw_filetype, 0);
#endif
	  }
	/* if the frame header DOES have a time & date, use them. */
	else {	if (nread != imread((AVOID *)imwp->imw_buf, nread))
			return FRAME_BADMAIL;
		else	(void)timecvt(mp->datetime, (char *)(imwp->imw_buf),
			 NULL, &tl0, 6, 0);
	     }
	(imwp->imw_drp)->run_d0pos[0] = 0; /* no datarun info */
	return FRAME_OK;
}

static int headerless(MAIL *mp, size_t csize)
{	/* This creates a nominal frame header for a headerless frame.
	The assumption is that the frame is 256-by-n, 512-by-n,
	1600x1600, or 1600x3200, and the length of the file in bytes is
	therefore xwidth * yheight * (2 bytes per word).
	This routine returns FRAME_OK unless the effort to rewind the
	file fails, in which case it returns FRAME_BAD16BIT */
	size_t		ntoread;
	time_t		tl0;
	off_t		loff;
	DATARUN		*drp;
	IMAGEWORK	*imwp;

	zeromem((AVOID *)mp, sizeof (MAIL)); /* most entries zeroed */
	mp->xdimen = (csize < 262144) ? 256 : ((csize < 524288) ? 512 :
		((csize < 1048576) ? 1024 : ((csize < 10000000) ? 1600 :
		((csize < DIP320_SIZE) ? 2048 : 1600))));
	if ((csize == DIP100_SIZE) || (csize == DIP320_SIZE))
		csize -= 180; /* remove trailer size from measurement */
	mp->ydimen = csize / (2 * mp->xdimen);
	if (NULL == (imwp = iminfo(055))) return FRAME_NO_MEMORY;
	drp = imwp->imw_drp;
	if (mp->xdimen <= 512)
	  {	mp->format = imwp->imw_filetype = SDMS_RANGE + mp->ydimen;
		mp->cellsize = 1000;	/* ok in X, wrong in Y */
		drp->run_d1pos[1] = 1.; /* roughly 1 cm pixel */
		drp->run_d1pos[2] = drp->run_d1pos[1] *
		 ((float)(mp->xdimen)) / ((float)(mp->ydimen)); /* proportion*/
	  }
	else if (imwp->imw_filetype == FMT_DIP100)
	  {	mp->format = imwp->imw_filetype;	mp->cellsize = 200;
	  }
	else if (imwp->imw_filetype == FMT_DIP320)
	  {	mp->format = imwp->imw_filetype;	mp->cellsize = 200;
		drp->run_d1pos[1] = drp->run_d1pos[2] = 0.2;
	  }
	else {	mp->format = imwp->imw_filetype = FMT_FUJI;
		mp->cellsize = 100;
		drp->run_d1pos[1] = drp->run_d1pos[2] = 0.1;
	     }
#ifdef LITTLEND
	mp->format = imwp->imw_filetype = lswawb(imwp->imw_filetype, 0);
#endif
	mp->accumsecs = mp->elap100ths = 10000; /* nominal values */
	mp->improps |= IPR_NEEDCOUNT;
	if ((imwp->imw_filetype == FMT_DIP100) ||
	 (imwp->imw_filetype == FMT_DIP320))
	  {	loff = (imwp->imw_filetype == FMT_DIP100) ? 5120045 : 10240045;
		if (-1 != lseek(imwp->imw_fd, loff, 0))
		  {	ntoread = 32;
			if (ntoread == imread((AVOID *)imwp->imw_buf, ntoread))
			  {	imwp->imw_buf[26] = '\0';
				(void)timecvt(mp->datetime,
				 (char *)(imwp->imw_buf), NULL, &tl0, 9, 0);
			  }
		  }
	  }
	/* use modification time for the file as found in the inode header */
	else (void)timecvt(mp->datetime, NULL, NULL,
	 &((imwp->imw_stabu)->st_mtime), 1, 0);	
	drp->run_d0pos[0] = 0.;	/* signal that we lack datarun info */
	/* rewind so we can get all the 16-bit data later */
	if (-1 == lseek(imwp->imw_fd, (off_t)0L, 0)) return FRAME_BAD16BIT;
	return FRAME_OK;
}

static int cornell_head(short *shp)
{ /*	This translates the contents of the Cornell header */
	int		iblck, swfmt;
	size_t		ntoread;
	time_t		tl0;
	MAIL		lm;
	IMAGEWORK	*imwp;

	zeromem((AVOID *)&lm, sizeof (MAIL));
	if (NULL == (imwp = iminfo(5))) return FRAME_NO_MEMORY;
	lm.format = imwp->imw_filetype;
	swfmt = swawb(imwp->imw_filetype);
	if (swfmt == FMT_CORNELL)
	  {	lm.xdimen = sswab(shp[2]);	lm.ydimen = sswab(shp[1]);
		lm.cellsize = sswab(shp[57]);
	  }
	else {	lm.xdimen = shp[2];	lm.ydimen = shp[1];
		lm.cellsize = shp[57];
	     }
	(void)timecvt(lm.datetime, (char *)(shp + 71), NULL, &tl0, 2, 0);
	lm.ontime = lm.eventcount = 10000000L;
	lm.accumsecs = 10;	lm.elap100ths = 1000;
	/* skip past all the pointless data. We do this before copying
	the mailbox back where it belongs so we can throw things away */
	ntoread = INBLOCK;
	for (iblck = 0; iblck < 31; iblck++)
		if (ntoread != imread((AVOID *)imwp->imw_buf, ntoread))
			return FRAME_BADMAIL;
	/* ... now we copy the header over */
	cpymem((AVOID *)imwp->imw_lhd, (AVOID *)&lm, sizeof (MAIL));
	return FRAME_OK;
}

static int raxis_head(MAIL *mp)
{ /*	This converts the header on an R-axis frame to a standard header */
	time_t			tl0;
	float			tare[3][3];
	DATARUN			*drp;
	IMAGEWORK		*imwp;
	static size_t		ntoread = 2048; /* don't read junk any more */
	static RAXISHEAD	trax;

	if (NULL == (imwp = iminfo(015))) return FRAME_NO_MEMORY;
	drp = imwp->imw_drp;
	/* rewind to re-read the first few characters */
	if (-1L == lseek(imwp->imw_fd, (off_t)0L, 0)) return FRAME_BADMAIL;
	zeromem((AVOID *)mp, sizeof (MAIL));
	if (ntoread != imread((AVOID *)&trax, ntoread)) return FRAME_BADMAIL;
	/* detect the floating-point version. For now we do this using
	 the detector width and height values. They'll be reasonable integers
	 if the words & bytes aren't swapped; they'll be crazy if not. */
	if ((100 <= trax.nxpix0) && (trax.nxpix0 <= 8000))
		imwp->imw_filetype = FMT_RAXIS;
	else imwp->imw_filetype = swawb(FMT_RAXIS);
	if (imwp->imw_filetype == swawb(FMT_RAXIS)) raxis_swap(&trax, 0);
	trax.raxis0[9] = trax.ver0[9] = trax.cid0[19] = trax.csys0[11] = '\0';
	trax.spg0[11] = trax.memo0[79] = trax.date0[11] = trax.user0[19] ='\0';
	trax.targ0[3] = trax.monoc0[19] = trax.coli0[19] = trax.filt0[3] ='\0';
	trax.maxis0[3] = trax.baxis0[3] = trax.cpu[9] = trax.ip[9] = '\0';
	strncpy(drp->run_code, trax.cid0, 7);	drp->run_code[7] = '\0';
	strncpy(drp->run_notes, trax.memo0, 63); drp->run_notes[63] = '\0';
	drp->run_volxray = trax.xkv0 + 0.49999;
	drp->run_curxray = trax.xma0 + 0.49999;
	xgensteps(drp, trax.fis0 - trax.fie0); /* sign reversal */
	drp->run_motor = 2;	/* temp! */
	drp->run_width = trax.nxpix0;	drp->run_height = trax.nzpix0;
	drp->run_startpos[0] = trax.duma0[0];
	drp->run_startpos[1] = trax.duma0[1];
	drp->run_startpos[2] = trax.fis0;
	drp->run_expos = trax.expos0;
	drp->run_d0pos[0] = trax.dist0 * 0.1;	/* 0.1cm per mm */
	drp->run_d0pos[1] = (trax.beam0 - ((float)trax.nxpix0) * 0.5) *
		trax.ysize0;
	drp->run_d0pos[2] = (trax.zbeam0 - ((float)trax.nzpix0) * 0.5) *
		trax.zsize0;
	drp->run_d0ang[0] = trax.duma0[2];
	drp->run_d1pos[1] = trax.ysize0; drp->run_d1pos[2] = trax.zsize0;
	drp->run_lambda = trax.wave0;
	if (!vecmat(trax.cell0, &(trax.cell0[3]), tare))
		(void)inv3(tare, drp->run_recip);
	mp->format = imwp->imw_filetype;
	mp->xdimen = trax.nxpix0;	mp->ydimen = trax.nzpix0;
	(void)timecvt(mp->datetime, trax.date0, NULL, &tl0, 5, 0);
	mp->accumsecs = trax.expos0 + 0.49999;
	mp->elap100ths = trax.expos0 * 100. + 0.49999;
	mp->frameno = trax.nosc0;
	mp->cellsize = (xgulong)(trax.ysize0 * 1000. + 0.49999);
	mp->improps |= IPR_XFASTEST | IPR_STARTBOTTOM | IPR_STARTRIGHT |
		IPR_NEEDCOUNT;
	mp->improps |= IPR_RAXISCOMPRESS;
	if (swawb(imwp->imw_filetype) == FMT_RAXIS)
		mp->improps |= IPR_BYTESWAPPED;
	else mp->improps &= ~(IPR_BYTESWAPPED);
	mp->improps &= ~(IPR_IS8BIT); return FRAME_OK;
}

static int mar_head(MAIL *mp, char *restorer)
{ /*	This reads and converts the puny little header on a Mar image plate
	frame.  We call this for any header. If the file doesn't look like a
	MAR header we return the mail pointer intact.
	The second argument is an ugly kludge with which we restore the first
	two bytes of the mar date and time after we carelessly overwrote it
	before calling this routine. */
	int		timetype;
	time_t		tl0;
	xgulong		ufile, swufile, *xgulp, *xgule;
	float		*flp, *fle, maxtwoth;
	DATARUN		*drp;
	IMAGEWORK	*imwp;
	static MARHEAD	mhd;

	if (NULL == (imwp = iminfo(015))) return FRAME_NO_MEMORY;
	drp = imwp->imw_drp;
	ufile = mp->format;	/* remember unsigned version of format */
	swufile = swawb(ufile); /* ... and word+byte-swapped version */
	/* eliminate obvious cases. It's definitely not a Mar file
	 unless the first four bytes are either a reasonable dimension
	 (between MIN_DIMENSION and MAX_DIMENSION)
	 or the byte-swapped equivalent. */
	if ((ufile < MIN_DIMENSION) || (swufile < MIN_DIMENSION) ||
	 ((ufile > MAX_DIMENSION) && (swufile > MAX_DIMENSION)))
		return FRAME_OK;
	cpymem((AVOID *)&mhd, (AVOID *)mp, sizeof (MARHEAD));
	/* here's where we fix the ugly kludge */
	mhd.ma_datetim[0] = restorer[0]; mhd.ma_datetim[1] = restorer[1];
	/* byte-and-word swap where necessary. `Necessary' is recognized by
	 a nominal `imwp->imw_filetype' (really the dimension of the detector)
	 that looks like a byte-and-wordswapped value */
	if (ufile > swufile)
	  {	for (xgulp = &(mhd.ma_width), xgule = &(mhd.ma_dose);
		 xgulp <= xgule; xgulp++) *xgulp = swawb(*xgulp);
		for (flp = &(mhd.ma_xrayctsum), fle = &(mhd.ma_highmult);
		 flp <= fle; flp++) flcvt((u_char *)flp, 0);
	  }
#ifdef __alpha
	else for (flp = &(mhd.ma_xrayctsum), fle = &(mhd.ma_highmult);
		 flp <= fle; flp++) vmtoduf(flp);
#endif
	/* having done that, we check the results to see if they look
	 like a reasonable MAR header. */
	if ((mhd.ma_width < MIN_DIMENSION) || (mhd.ma_height < MIN_DIMENSION) ||
	 (mhd.ma_width > MAX_DIMENSION) || (mhd.ma_height > MAX_DIMENSION))
		return FRAME_OK;
	if ((mhd.ma_xpcm < 0.) || (mhd.ma_ypcm < 0.) ||
	 (mhd.ma_xpcm > 5.) || (mhd.ma_ypcm > 5.)) return FRAME_OK;
	if ((mhd.ma_lambda < 0.) || (mhd.ma_lambda > 100.)) return FRAME_OK;
	/* header length should be 2*ma_width, not ma_hdr1+ma_hdr2 */
	imwp->imw_headlen = mhd.ma_width * sizeof (u_short);
	xgensteps(drp, mhd.ma_phie - mhd.ma_phi0);
	/* zero out the internal mailbox array */
	zeromem((AVOID *)mp, sizeof (MAIL));
	/* having passed all these tests, we think it's a real MAR header.
	 We define all of them as having format == 0x4b0=1200 or its
	 byteswapped equivalent, rather than retaining a specific format val*/
	if (ufile < swufile)
	  {
		mp->format = imwp->imw_filetype = FMT_MAR;
		mp->improps &= ~IPR_BYTESWAPPED;
	  }
	else {
		mp->format = imwp->imw_filetype = swawb(FMT_MAR);
		mp->improps |= IPR_BYTESWAPPED;
	     }
	mp->xdimen = drp->run_width = mhd.ma_width;
	mp->ydimen = drp->run_height = mhd.ma_height;
	mp->elap100ths = 10000; mp->accumsecs = 100;
	mp->eventcount = mp->ontime = 1000000L;
	/* nuisance: this header element is sometimes zero */
	if (mhd.ma_xpcm > FUZZ)
		mp->cellsize = (xgulong)(int)(mhd.ma_xpcm * 1000. + 0.49999);
	else	mp->cellsize = 150;	/* I think it's always this... */
	mp->oflowcount = mhd.ma_n16flows;
	mp->improps |= IPR_XFASTEST; mp->improps &= ~IPR_IS8BIT;
	if ((mp->xdimen < 1600) && ((fabs(mhd.ma_xrayctsum) > 0.) ||
	 (fabs(mhd.ma_maxradius) > 0.) || (fabs(mhd.ma_minradius) > 0.)))
	  {	drp->run_startpos[0] = mhd.ma_xrayctsum;
		drp->run_startpos[1] = mhd.ma_maxradius;
		drp->run_startpos[2] = mhd.ma_minradius;
	  }
	else {	drp->run_startpos[0] = mhd.ma_omega;
		drp->run_startpos[1] = 0.;
		drp->run_startpos[2] = mhd.ma_phi0;
	     }
	drp->run_expos = 100.;
	drp->run_d0pos[0] = mhd.ma_plated * 0.1; /* 0.1cm per mm */
	drp->run_d0pos[1] = mhd.ma_x0 - 0.5 * ((float)mhd.ma_width);
	drp->run_d0pos[2] = mhd.ma_y0 - 0.5 * ((float)mhd.ma_height);
	drp->run_d0ang[0] = mhd.ma_omega;
	if (mhd.ma_xpcm > FUZZ) drp->run_d1pos[1] = mhd.ma_xpcm;
	else drp->run_d1pos[1] = ((float)(mp->cellsize)) / 1000.;
	if (mhd.ma_ypcm > FUZZ) drp->run_d1pos[2] = mhd.ma_ypcm;
	else drp->run_d1pos[2] = ((float)(mp->cellsize)) / 1000.;
	drp->run_motor = 2;	/* temp! */
	drp->run_lambda = mhd.ma_lambda;
	timetype = (mp->xdimen == 1600) ? 9 : 3;
	(void)timecvt(mp->datetime, mhd.ma_datetim, NULL, &tl0, timetype, 0);
	maxtwoth = atan(((double)mhd.ma_width) *
		mhd.ma_xpcm * 0.5/ mhd.ma_plated) / RADC;
	if (mp->xdimen < 1600) maxtwoth += fabs(mhd.ma_omega);
	if (maxtwoth < 0.01) maxtwoth = 0.01;
	maxtwoth = sin(RADC * maxtwoth / 2.);
	drp->run_reslim = 0.5 * mhd.ma_lambda / maxtwoth;
	drp->run_ssqmax = maxtwoth * maxtwoth;
	if (-1 == lseek(imwp->imw_fd, (off_t)0L, 0)) return FRAME_BADMAIL;
	if (imwp->imw_headlen != imread((void *)imwp->imw_buf,
	 imwp->imw_headlen)) return FRAME_BADMAIL;
	return FRAME_OK;
}

static int read_trailer(MAIL *mp)
{ /*	This reads the DIP2000 or DIP3000 trailer format:
	format courtesy of Bram Schierbeek, Enraf-Nonius */
	size_t			lensofar;
	int			dxmicrons, dymicrons;
	float			realstep, dewidth, phmax, totang;
	DATARUN			*drp;
	IMAGEWORK		*imwp;
	static DIPTRAILER	thisdtr;
	static size_t		dipxksize = 1024;

	if (NULL == (imwp = iminfo(015))) return FRAME_NO_MEMORY;
	drp = imwp->imw_drp;
	if (-1 == lseek(imwp->imw_fd, (off_t)(-dipxksize), 2))
		return FRAME_BADMAIL;
	if (dipxksize != imread((AVOID *)&thisdtr, dipxksize))
		return FRAME_BADMAIL;
	zeromem((AVOID *)mp, sizeof (MAIL)); /* start zeroed */
	mp->format = imwp->imw_filetype = FMT_DIPX000;
	if (thisdtr.di_xsize < 1) thisdtr.di_xsize = 2500;
	mp->xdimen = drp->run_width = thisdtr.di_xsize;
	mp->ydimen = drp->run_height = thisdtr.di_ysize;
	mp->accumsecs = drp->run_expos = thisdtr.di_expos;
	mp->elap100ths = 100. * thisdtr.di_expos;
	mp->cellsize = thisdtr.di_pixelsize;
	if ('\0' != thisdtr.di_comment[0])
		strncpy(drp->run_notes, thisdtr.di_comment, 60);
	lensofar = strlen(drp->run_notes);
	if ((lensofar < 58) && ('\0' != thisdtr.di_monochro[0]))
	  {	strcat(drp->run_notes, "; ");
		strncat(drp->run_notes, thisdtr.di_monochro,
			62 - lensofar);
	  }
	lensofar = strlen(drp->run_notes);
	if (lensofar < 56)
	  {	strcat(drp->run_notes, "; plate ");
		if (thisdtr.di_ipno == 2)
			strcat(drp->run_notes, "2");
		else	strcat(drp->run_notes, "1");
	  }
	if (thisdtr.di_kv > FUZZ) drp->run_volxray = thisdtr.di_kv;
	if (thisdtr.di_ma > FUZZ) drp->run_curxray = thisdtr.di_ma;
	realstep = thisdtr.di_phi2 - thisdtr.di_phi1;
	drp->run_stepsize = 96. * realstep;
	if (realstep < 0.) drp->run_stepsize = 96. * realstep - 0.49999;
	else drp->run_stepsize = 96. * realstep + 0.49999;
	drp->run_stepreal = realstep;
	drp->run_motor = 2;	/* phi motion */
	drp->run_nchams = 2;	/* two image plates! */
	if (thisdtr.di_xlambda < FUZZ) thisdtr.di_xlambda = DEF_LAMBDA;
	drp->run_lambda = thisdtr.di_xlambda;
	drp->run_startpos[0] = drp->run_startpos[1] = 0.;
	drp->run_startpos[2] = thisdtr.di_phi1;
	if (thisdtr.di_cdist < FUZZ) thisdtr.di_cdist = 100.;
	drp->run_d0pos[0] = thisdtr.di_cdist * 0.1; /* mm to cm */
	dxmicrons = thisdtr.di_pixelsize *
		(thisdtr.di_ipx - thisdtr.di_xsize / 2);
	dymicrons = thisdtr.di_pixelsize *
		(thisdtr.di_ipy - thisdtr.di_ysize / 2);
	drp->run_d0ang[0] = thisdtr.di_pttheta;
	drp->run_d0pos[1] = 0.1e-3 * (float)dxmicrons;
	drp->run_d0pos[2] = 0.1e-3 * (float)dymicrons;
	drp->run_d1pos[1] = drp->run_d1pos[2] =
		1.0e-3 * (float)(mp->cellsize);
	dewidth = 0.1e-3 * ((float)thisdtr.di_xsize) * thisdtr.di_pixelsize;
	phmax = atan(dewidth / (2. * thisdtr.di_cdist));
	totang = 0.5 * fabs(thisdtr.di_pttheta * RADC) + phmax;
	drp->run_ssqmax = 2. * sin(totang);
	if (drp->run_ssqmax < FUZZ) drp->run_ssqmax = 0.2;
	drp->run_reslim = thisdtr.di_xlambda / drp->run_ssqmax;
	drp->run_ssqmax *= drp->run_ssqmax;
	/* having done all that, we rewind and return */
	return (lseek(imwp->imw_fd, (off_t)0L, 0) == -1) ?
		FRAME_BADMAIL : FRAME_OK;
}

int read_header(MAIL *mpo, int ifd, int ruflags)
{ /*	This reads in the MAIL structure from the opened file with descriptor
	imhs.imw_fd and does the word-swapping necessary to get the data
	correct.  In the case of PCS-format file, everything gets shifted
	around some and the remaining mailbox data gets read in.
	If the read is successful, the routine returns the `imhs.imw_filetype'
	value, i.e., the value in the first 4 bytes of the input file.
	If it's unsuccessful, it returns FRAME_BADMAIL. */
	register int	i;
	int		imagebytes, imagesize, nwaits, mchflag, swtype, ttype;
	int		martype;
	size_t		cursize, headsize, ntoread;
	off_t		roff;
	short		saveprops;
	xgulong		*ulp, *ulpe;
	DATARUN		*drp;
	IMAGEWORK	*imwp;
	PAU		imu;
	static int	nreserved;
	static float	aduperphot;
	static u_char	*data8old;
	static u_short	*data16old;
	static VOFLOW	*oflowold, *ofeold;
	static VOFL16	*of16dold;
	static DATARUN	*runhdrold, *drup;
	static char	oflowmsg[] = "mpo->overflow in read_header";

	if (NULL == (imwp = iminfo(055))) return FRAME_NO_MEMORY;
	drp = imwp->imw_drp;
	/* we run `fstat' to make available some useful info */
	if (-1 == fstat(ifd, imwp->imw_stabu))
	  {	if (NULL != runhdrold)
		  {	(void)free((AVOID *)runhdrold);	runhdrold = NULL;
		  }
		mpo->runheader = NULL;
		fprintf(stderr, "fstat on image returned -1\n");
		return FRAME_BADOPEN;
	  }
	if (0 == (int)((imwp->imw_stabu)->st_ino))
	  {	fprintf(stderr,
			"inode = %d\n", (int)((imwp->imw_stabu)->st_ino));
		return FRAME_BADOPEN;
	  }
#ifndef linux
	else if (!(S_ISREG((imwp->imw_stabu)->st_mode))) return FRAME_BADOPEN;
#endif
	cursize = (imwp->imw_stabu)->st_size;	/* remember the file size */
	if ((MIN_SIZE > cursize) || (cursize > MAX_SIZE))
		return FRAME_BADOPEN;
	/* save some pointers built into the header from previous calls. It
	 is the calling program's responsibility to assure that these have
	 been zeroed on the first call */
	data8old = mpo->data8;		data16old = mpo->data16;
	oflowold = mpo->overflow;	of16dold = mpo->oflow16;
	runhdrold = mpo->runheader;
	ofeold = (NULL == runhdrold) ? NULL : runhdrold->run_oflowend;
	/* set the file's static vbls equal to the arguments to this routine */
	imwp->imw_lhd = mpo;	imwp->imw_fd = ifd;
	if (SIZMAIL != (headsize = imread((AVOID *)mpo, SIZMAIL)))
	  {	if ((headsize > 92) && (NULL != runhdrold))
		  {	(void)free((AVOID *)runhdrold);	mpo->runheader = NULL;
		  }
		return FRAME_BADMAIL;
	  }
	saveprops = mpo->improps;	/* we need these values for the Mar */
	mpo->improps = 0; /* turn all properties bits off before  proceeding*/
	drp->run_d0pos[0] = 0;	/* signal: no DATARUN hdr read in */
	/* the MAR IP format is actually kind of hard to recognize.
	 We look for it early on and save ourselves some trouble later. */
	if (FRAME_OK != (i = mar_head(mpo, (char *)&saveprops)))
	  {	if (NULL != runhdrold) (void)free((AVOID *)runhdrold);
		mpo->runheader = NULL; return i;
	  }
/* handle the special cases of headerless frames, which we recognize
 by their size */
	if (cursize == FUJI_SIZE)
	  {	mpo->format = FMT_FUJI;
		mpo->improps |= (IPR_XFASTEST | IPR_STARTBOTTOM |
			IPR_STARTRIGHT | IPR_LOGSCALE);
		mpo->improps &= ~IPR_IS8BIT;
	  }
	else if (cursize == DIP100_SIZE)
	  {	mpo->format = FMT_DIP100;
		mpo->improps &= ~IPR_IS8BIT;
	  }
	else if ((cursize == DIP2000_SIZE) || (cursize == DIP3000_SIZE))
	  {	mpo->format = FMT_DIPX000;
		mpo->improps |= (IPR_XFASTEST | IPR_STARTBOTTOM);
		mpo->improps &= ~IPR_IS8BIT;
	  }
	else if (cursize == DIP320_SIZE)
	  {	mpo->format = FMT_DIP320;
		mpo->improps |= (IPR_XFASTEST | IPR_STARTBOTTOM);
		mpo->improps &= ~IPR_IS8BIT;
	  }
	else if ((MIN_SDMS <= cursize) && (cursize <= MAX_SDMS))
	  {	mpo->format = SDMS_RANGE;
		mpo->improps |= IPR_XFASTEST;
	  }
	else if (!mpo->format) mpo->format = FMT_CCD_VMSALT;
	imwp->imw_filetype = mpo->format;	/* remember input file format*/
	aduperphot = 1.;	/* default: 1 photon per histogram count */
	ttype = imwp->imw_filetype;	swtype = swawb(imwp->imw_filetype);
	martype = FMT_MARCCD_L;
	if ((swtype == FMT_PCS_16) || (swtype == FMT_PCS_8))
	  {	if (swtype == FMT_PCS_16) mpo->improps &= ~IPR_IS8BIT;
		else mpo->improps |= IPR_IS8BIT;
		mpo->improps |= IPR_BYTESWAPPED;
		saveprops = mpo->improps;
		for (ulp = ((xgulong *)mpo) + 1, ulpe = ((xgulong *)mpo) + 32;
		 ulp < ulpe; ulp++) *ulp = swawb(*ulp);
		mpo->improps = saveprops;
		if (FRAME_OK != (i = pcs_to_mail(mpo)))
		  {	if (NULL != runhdrold) (void)free((AVOID *)runhdrold);
			mpo->runheader = NULL; return i;
		  }
		mpo->cellsize = 200;
	  }
	else if (swtype == FMT_CAD)	/* Cadmus format: byte & wordswap) */
	  {	mpo->improps |= IPR_IS8BIT;	/* 8-bit image still default */
		for (ulp = &(mpo->xdimen), ulpe = &(mpo->zmax); ulp <= ulpe;
			ulp++) *ulp = swawb(*ulp);
		for (ulp = &(mpo->zdisp), ulpe = &(mpo->over64k); ulp <= ulpe;
			ulp++) *ulp = swawb(*ulp);
		mpo->frameno = lswawb(mpo->frameno, 0);
		mpo->cellsize = 200;
		/* the overflows (if there are any) start immediately after the
		 128 byte header, so we don't need to reposition anything */
	  }
	else if ((ttype == FMT_CORNELL) ||(swtype == FMT_CORNELL))
	  {	if (swtype == FMT_CORNELL) mpo->improps |= IPR_BYTESWAPPED;
		mpo->improps &= ~IPR_IS8BIT; /* remember it's 16-bit */
		/* first, rewind it so we get the whole thing in one shot */
		if (-1 == lseek(imwp->imw_fd, (off_t)0L, 0))
		  {	if (NULL != runhdrold) (void)free((AVOID *)runhdrold);
			mpo->runheader = NULL; return FRAME_BADMAIL;
		  }
		if (INBLOCK != imread((AVOID *)imwp->imw_buf, INBLOCK))
		  {	if (NULL != runhdrold) (void)free((AVOID *)runhdrold);
			mpo->runheader = NULL; return FRAME_BADMAIL;
		  }
		imu.ucpt = imwp->imw_buf;
		if (FRAME_OK != (i = cornell_head((short *)imu.uspt)))
		  {	if (NULL != runhdrold) (void)free((AVOID *)runhdrold);
			mpo->runheader = NULL; return i;
		  }
		if (mpo->cellsize < 1) mpo->cellsize = 100;
		aduperphot = 5.;	/* photons per ADU */
	  }
	else if ((ttype == FMT_SAXII) || (swtype == FMT_SAXII))
	  {
		if (swtype == FMT_SAXII) mpo->improps |= IPR_BYTESWAPPED;
		mpo->improps |= IPR_IS8BIT;	/* default is an 8-bit image */
		if (FRAME_OK != (i = saxii_ascii(mpo, drp,
		 (char *)imwp->imw_buf, imwp->imw_fd)))
		  {	if (NULL != runhdrold) (void)free((AVOID *)runhdrold);
			mpo->runheader = NULL; return i;
		  }
		/* flip stepsize if phi is moving and |chi| is < 90 */
		if ((drp->run_motor == 2) &&
		 (fabs(drp->run_startpos[1]) < 90.))
		  {	drp->run_stepreal = -(drp->run_stepreal);
			drp->run_stepsize = -(drp->run_stepsize);
		  }
		imwp->imw_filetype = mpo->format; /* changed internally? */
		if ((imwp->imw_filetype == FMT_SAXII_16) ||
		 (swawb(imwp->imw_filetype) == FMT_SAXII_16))
			mpo->improps &= ~IPR_IS8BIT; /* remember it's 16-bit */
	  }
	else if ((ttype == FMT_RAXIS) || (ttype == FMT_RAXIS6) ||
	 (swtype == FMT_RAXIS) || (swtype == FMT_RAXIS6))
	  {	if ((swtype == FMT_RAXIS) || (swtype == FMT_RAXIS6))
			mpo->improps |= IPR_BYTESWAPPED;
		mpo->improps &= ~IPR_IS8BIT; /* remember it's 16-bit */
		aduperphot = 4.; /* compendium of results suggests this */
		if (FRAME_OK != (i = raxis_head(mpo)))
		  {	if (NULL != runhdrold) (void)free((AVOID *)runhdrold);
			mpo->runheader = NULL; return i;
		  }
		/* padded 1900x1900 images are recognized by their length */
		if ((cursize == 7786496) || (cursize == 1947648))
			mpo->improps |= IPR_PAD;
		/* set the pointer at the beginning of the 16-bit data.
		 For padded 1900x1900 or 950x950 images it's at 4096
		 or 2048 bytes respectively; otherwise it's at
		 2*xdimen. */
		roff = (mpo->improps & IPR_PAD) ?
		 ((mpo->xdimen == 950) ? 2048 : 4096) : mpo->xdimen * 2;
		if (-1 == lseek(imwp->imw_fd, roff, 0))
		  {	if (NULL != runhdrold) (void)free((AVOID *)runhdrold);
			mpo->runheader = NULL; return FRAME_BADMAIL;
		  }
		mpo->cellsize = 100;
	  }
	else if ((swtype == FMT_QTEX) || (ttype == FMT_QTEX)) /* Quantex */
	  {	
		if (swtype == FMT_QTEX)
		  {
			mpo->improps |= IPR_BYTESWAPPED;
			mpo->xdimen = swtype & 0xFFFF;
			mpo->ydimen = swtype >> 16;
		  }
		else {
			mpo->improps &= ~IPR_BYTESWAPPED;
			mpo->xdimen = ttype & 0xFFFF;
			mpo->ydimen = ttype >> 16;
		     }
		aduperphot = 5.;	mpo->improps &= ~IPR_IS8BIT;
		/* header is 256 bytes: read the other 128 */
		if (128 != imread((AVOID *)imwp->imw_buf, 128))
		  {	if (NULL != runhdrold) (void)free((AVOID *)runhdrold);
			mpo->runheader = NULL; return FRAME_BADMAIL;
		  }
		mpo->improps &= ~IPR_IS8BIT;	mpo->format = ttype;
		mpo->improps |= (IPR_XFASTEST | IPR_STARTBOTTOM);
		mpo->elap100ths = 10000;
	  }
	else if ((ttype == FMT_MAR) || (swtype == FMT_MAR)) /*Mar image plate*/
	  {
	    	mpo->improps &= ~IPR_IS8BIT;
		aduperphot = 3.;	/* photons per ADU */
		if (mpo->cellsize == 0) mpo->cellsize = 150;
		if (mpo->oflowcount) mpo->improps |= IPR_HAS16BFLOWS;
	  }
	else if ((ttype == FMT_MARCCD) || (ttype == FMT_MARCCD_L) ||
	 (swtype == FMT_MARCCD) || (swtype == FMT_MARCCD_L))
	  {
		mpo->improps &= ~IPR_IS8BIT;
		mchflag = (ruflags & ANNOUNCE) ? 1 : 0;
		if (FRAME_OK != (i = mchread(mpo, drp, mchflag)))
		  {	if (NULL != runhdrold) (void)free((AVOID *)runhdrold);
			mpo->runheader = NULL;
			aduperphot = 1.8;	/* most results suggest this */
			return i;
		  }
	  }
	else if ((ttype == FMT_FAST) || (swtype == FMT_FAST))
	  {		/* Ascii header for FAST detectors, ADSC CCDs, etc. */
		aduperphot = 4.;	/* photons per ADU */
		mpo->improps &= ~IPR_IS8BIT;
		cpymem((char *)(imwp->imw_buf), (char *)mpo, sizeof (MAIL));
		*(imwp->imw_buf + 100) = (char)(saveprops >> 8);
		*(imwp->imw_buf + 101) = (char)(saveprops & 0xFF);
		zeromem((AVOID *)(imwp->imw_lhd), sizeof (MAIL));
		mpo->format = imwp->imw_filetype;
		if (!strncmp("HEADER_BYTES=", (char *)(imwp->imw_buf + 2), 13))
		  {	if (FRAME_OK != (i = fasthead()))
			  {	if (NULL != runhdrold)
					(void)free((AVOID *)runhdrold);
				mpo->runheader = NULL; return FRAME_BADMAIL;
			  }
			if (0. < drp->run_d1pos[1]) mpo->cellsize =
				1000. * drp->run_d1pos[1] + 0.49999;
			else mpo->cellsize = (200 * mpo->xdimen / 512);
			/* pick out default adu/photon value */
			if (mpo->xdimen == 1536) aduperphot = 4; /*APS1binned*/
			else if (mpo->xdimen == 3072)
			  {	if (fabs(drp->run_d1pos[1] - 0.0798) <0.005)
					aduperphot = 4;	/* APS1unbinned */
				else	aduperphot = 1.8; /* ADSC Q315 */
			  }
			else if (mpo->xdimen == 2048)
			  {	if (fabs(drp->run_d1pos[1] - 0.1024) < 0.01)
					aduperphot = 1.8;	/* Q210 */
				else	aduperphot = 1.4;	/* others (?)*/
			  }
			else if (mpo->xdimen == 4096) aduperphot = 1.8;
			else if (mpo->xdimen == 6144) aduperphot = 1.8;
			else if ((mpo->cellsize == 100) &&
			 (mpo->xdimen == 1024)) aduperphot = 1.4; /*Brandeis*/
		  }
		else { /* alternative requires headerless-like handling */
			if (FRAME_OK != (i = headerless(mpo, cursize)))
			  {	if (NULL != runhdrold)
					(void)free((AVOID *)runhdrold);
				mpo->runheader = NULL;
				aduperphot = 4;	/* wild guess */
				return i;
			  }
		     }
	  }
	else
	 { switch (ttype)
	    {
	     case	FMT_VMS:
		/* header already in good format: just reposition pointer */
		ntoread = DEFBLOCK - SIZMAIL; 	/* skip to correct position */
		mpo->improps |= IPR_IS8BIT;	/* 8-bit image still default*/
		if (ntoread != imread((AVOID *)imwp->imw_buf, ntoread))
		  {	if (NULL != runhdrold) (void)free((AVOID *)runhdrold);
			mpo->runheader = NULL;
			return FRAME_BADMAIL;
		  }
		mpo->cellsize = 200;	/* Siemens: 200 micron raster */
		break; /* already-converted frame */
	     case	FMT_PCS_16:	/* PCS 16-bit format, already swapped*/
		mpo->improps &= ~IPR_IS8BIT; /*remember 16-bitness, drop thru*/
	     case	FMT_PCS_8:	/* PCS format, already swapped */
		mpo->improps |= IPR_IS8BIT;	/* 8-bit image still default */
		if (FRAME_OK != (i = pcs_to_mail(mpo)))
		  {	if (NULL != runhdrold) (void)free((AVOID *)runhdrold);
			mpo->runheader = NULL; return i;
		  }
		break;
	    case FMT_CCD:	case FMT_CCD_ALT:
	    case FMT_CCD_VMS:	case FMT_CCD_VMSALT:
		aduperphot = 5.;	mpo->improps &= ~IPR_IS8BIT;
		if (FRAME_OK != (i = ccd_header(mpo)))
		  {	if (NULL != runhdrold) (void)free((AVOID *)runhdrold);
			mpo->runheader = NULL; return i;
		  }
		mpo->cellsize = (200 * mpo->xdimen / 512);
		break;
	    case FMT_DIPX000:	/* DIP2000 and DIP3000 image plates */
		mpo->improps &= ~IPR_IS8BIT;
		if (FRAME_OK != (i = read_trailer(mpo)))
		  {	if (NULL != runhdrold) (void)free((AVOID *)runhdrold);
			mpo->runheader = NULL; return i;
		  }
		break;
	    case FMT_FUJI:	case SDMS_RANGE:
	    case FMT_DIP100:	case FMT_DIP320:
		mpo->improps &= ~IPR_IS8BIT;	/*these are all 16-bit cases */
		if (FRAME_OK != (i = headerless(mpo, cursize)))
		  {	if (NULL != runhdrold) (void)free((AVOID *)runhdrold);
			mpo->runheader = NULL;
			aduperphot = 4;	/* wild guess... I need some data... */
			return i;
		  }
		break;
	    default:	/* we've now accounted for all the understood cases */
		if (NULL != runhdrold) (void)free((AVOID *)runhdrold);
		mpo->runheader = NULL; return FRAME_FMT_UNKNOWN;
	   }
	 }
	/* if `ydimen' is < 2, we see that as a signal that the detector has
	the same number of pixels across as it has high. */
	if (mpo->ydimen < 2) mpo->ydimen = mpo->xdimen;
		/* restore values from earlier calls */
	mpo->data8 = data8old;		mpo->data16 = data16old;
	mpo->overflow = oflowold;	mpo->oflow16 = of16dold;
	mpo->runheader = runhdrold;
		/* save the area in pixels of the detector face */
	imagesize = mpo->xdimen * mpo->ydimen;
	drup = mpo->runheader;
	if (!drup)
	  {	if (NULL == (drup = (DATARUN *)callop(1,
			sizeof (DATARUN), "mpo->runheader in read_header")))
				return FRAME_NO_MEMORY;
	  }
	/* stuff the DATARUN header in the right place if there is one */
	if (drp->run_d0pos[0] > FUZZ)
	  cpymem((AVOID *)drup, (AVOID *)drp, sizeof (DATARUN));
	else zeromem((AVOID *)drup, sizeof (DATARUN));
	drup->run_oflowend = ofeold;	mpo->runheader = drup;
	drup->run_aduperph = aduperphot;
	/* to be tidy, we require that these three variables agree.
	 The one we adjust is the # of electrons that produce an adu */
	drup->run_eperadu = drup->run_eperph / drup->run_aduperph;
	if ((imwp->imw_filetype == FMT_VMS) || (swtype == FMT_CAD))
		moreheader(mpo);
	/* none of the stuff below here is relevant if we're just reading
	 frame headers. So: */
	if (ruflags & JUST_HEADER) return imwp->imw_filetype;
	/* if we don't have enough memory reserved for overflows,
	 throw the old memory away so we're forced to get some new*/
	if (mpo->overflow)
	  {	drup = mpo->runheader;
		if (NULL == drup) nreserved = 0;
		else nreserved = drup->run_oflowend - mpo->overflow;
		if ((int)(mpo->oflowcount) > nreserved)
		  {	memzap((AVOID **)&(mpo->overflow), oflowmsg);
			mpo->overflow = NULL;
			if (NULL != drup) drup->run_oflowend = NULL;
		   }
	  }
	if (!mpo->overflow)
	  {	nreserved = (mpo->oflowcount > MINRESERVED ?
			mpo->oflowcount : MINRESERVED);
		if (NULL == (mpo->overflow = (VOFLOW *)callop(
		 (unsigned)nreserved, sizeof (VOFLOW), oflowmsg)))
		  {	fprintf(stderr,
			 "Can't allocate memory for %d overflows\n",nreserved);
			return FRAME_NO_MEMORY;
		  }
		drup = mpo->runheader;
		if (NULL != drup)
			drup->run_oflowend = mpo->overflow + nreserved;
	  }
	/* put something plausible in for the exposure time if needed. */
	if (mpo->elap100ths <= 0) mpo->elap100ths = 100;
	if (ruflags & JUST_HEADPFLOWS) return imwp->imw_filetype;
	/* reserve memory for 8 or 16-bit data,
	plus space for the datarun header and the overflows.
	If this space is already reserved, don't do it again.  Conversely,
	if space is reserved for the wrong kind of data, get rid of it*/
	if ((!((mpo->improps) & IPR_IS8BIT)) && !(mpo->data16))
	  {	/* SDMS data: keep both 8 & 16-bit registers around. Luxury! */
		if ((mpo->data8) && (imwp->imw_filetype != (SDMS_RANGE +
		 mpo->ydimen)))
		  memzap((AVOID **)&(mpo->data8),"mpo->data8 in read_header");
		if (NULL == (mpo->data16 = (u_short *)callop((unsigned)
		 imagesize, sizeof (u_short), "data16 in read_header")))
		  {	fprintf(stderr, "16-bit mem fail: x,y dimen = %d,%d\n",
				(int)mpo->xdimen, (int)mpo->ydimen);
			return FRAME_NO_MEMORY;
		  }
	  }
	else if (((mpo->improps) & IPR_IS8BIT) && !(mpo->data8))
	  {	if ((mpo->data16) && (imwp->imw_filetype !=
		 (SDMS_RANGE + mpo->ydimen)))
		  memzap((AVOID **)&(mpo->data16),"data16 in read_header");
		if (NULL == (mpo->data8 = (u_char *)callop((unsigned)
		 imagesize, sizeof (u_char), "mpo->data8 in read_header")))
		  {	fprintf(stderr, "8-bit mem fail: x,y dimen = %d,%d\n",
				(int)mpo->xdimen, (int)mpo->ydimen);
			return FRAME_NO_MEMORY;
		  }
	  }
	/* having opened the file, we make sure that it's been fully written
	out.  We do this by running `fstat' on the file and looking at
	the file length. If it's too short, we wait a little while and
	look again, up to three times.  Then we give up. */
	imagebytes = (mpo->data8 ? imagesize : 2 * imagesize);
	if ((imwp->imw_stabu)->st_size < imagebytes)
	  {	for (nwaits = 0; nwaits < 3; nwaits++)
		     {
			sleep(10);	/* be long enough to write out fully */
			if ((-1 != fstat(imwp->imw_fd, imwp->imw_stabu)) &&
			 ((imwp->imw_stabu)->st_size >= imagebytes)) break;
		     }
		if (nwaits >= 3) return FRAME_BADLENGTH;
	  }
	return imwp->imw_filetype;
}

/* end of imhead.c */
