/* mrutil.c -- Utilities for manipulating XGEN multirefs

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	 5 Dec 2006: outmu and muinterp written to handle MUREFL I/O
	24 Jun 2006: include xgpostint.h; structures we need are there
	11 Jun 2006: move keep_dimen, abpto[xy] earlier in file
	17 Dec 2005: keep_dimen, abptox, abptoy: eliminate static globals
	24 Nov 2005: inline argument declarations and strict prototyping
	24 Oct 2005: housekeeping
	22 Jun 2005: spgtopat() declared static
	 7 Mar 2005: bugfix in symtr() for Patterson spacegroup 83,87,88:
		missing `return' has been restored
	29 Jan 2005: mpack(),munpack(),symtr(),iscentric() handle centric
		spacegroups; more consistent approach to spacegroup stuff
	 2 Sep 2004: former mrgeom.c materials brought here; cosmetics
	28 Aug 2004: routines used only here are declared as static
	24 Aug 2004: findq() deleted because nobody is calling it
	17 Aug 2004: getq() and findq() moved from xgenmisc.c to here
	 1 May 2004: in eliminate_dups(), keep the observation with
		the larger I/sigma
	20 Mar 2004: in screw_ok(), drop assumption that reflection is
		already in simplest terms
	21 Jan 2004: license message inserted and initial comments reorganized
	18 Jan 2004: in workrun(), delay inverting drp->run_recip to compute
		rup->rp_areal until drp->run_recip has been rotated & rescaled
	 9 Jan 2004: headmem() reserves memory for CRYSTAL and DATARUN structs
	 5 Jan 2004: globals widrat, htrat, hto2 made static globals
	25 Nov 2003: in hklcmp(), initialize diff to zero to keep -Wall happy
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Sep 2003: eliminate access() call for MULTIREF in findq
		because it's subsumed in the xgopen() call
	10 Jun 2003: change the local array `rec' to `lrec' in calc_euler()
		so it doesn't get confused with the global variable `rec'
	 5 Jan 2003: reorganized source-file header
	15 Oct 2002: improve roundup in abptox() and abptoy();
		cossw, sinsw, costi, sinti correctly calculated in workrun():
		we had previously neglected to multiply the angle by RADC
	10 Sep 2002: in symtr() for point group 622, we correct a bug:
		we now define a reflection as centric if h==0, k==0, l==0,
		or h==k.
	18 Apr 2002: duplicate_ob() and eliminate_dups() moved from
		../postprocess/mrutil.c to here so they can be used
		in ../utilities/redurefl.c easily
	25 Mar 2002: removed MSI attachments
	31 May 1995: "register" -> "register int" throughout
	20 May 1995: clean up a comment
	 2 Mar 1995: diagnostics to stdout; DATA doesn't need to be writable;
		trivial reformatting
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 5 Oct 1994: remove g_mref, p_mref; we use xgread, xgwrite
	30 Sep 1994: accommodate floating-point stepsize in workrun()
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	 7 Dec 1993: initial X-GEN version
	11 Jan 1993: add routines for determining pixel positions
	26 Nov 1986: in mpack(), we treat hk0, h0l, 0kl specially to ensure
		that MPACK(MUNPACK(hkls)) = hkls and MUNPACK(MPACK(hkl)) = hkl.
	 5 Dec 1985: rewrite symtr() to include many other spacegroups
	 1 Oct 1985: add g_mr and p_mr
	18 Sep 1985: rewrite some routines to prevent collision
		and add routine screw_ok
	14 Sep 1984: add routines splitmr() and gmem()
	10 Sep 1984: create initial code, partly based on La Jolla models

   Purpose: utilities used in dealing with MULTIREF files.
	Most involve comparing hkls of refls and symmetry-transformations.

   Cross references: xgenshift.c

*************************************************************************/
#include	"xgpostint.h"
#define		DUPLICATE_FLAG	(-32765)
#define	SQ6	(2.44948974)
#define	SQ1P5	(1.22474487)
#define	SQ1P25	(1.11803398)
#define	SQ30	(5.47722557)
#define	SQ7P5	(2.73861278)
#define	SQ10OV3	(1.82574185)

int muinterp(char *mustr, MUREFL *mup)
{
 /* This builds a MUREFL from the string found in mustr.
	It returns the number of items in mustr. */
	int	nitems, tflag;

	if ((NULL == mustr) || (*mustr == '\0')) return 0;
	nitems = sscanf(mustr, "%hd %hd %hd %f %f %f %f %f %f %f %d\n",
		 &(mup->mu_hkl[0]), &(mup->mu_hkl[1]), &(mup->mu_hkl[2]),
		 &(mup->mu_stl), &(mup->mu_f), &(mup->mu_sig),
		 &(mup->mu_fplus), &(mup->mu_sigplus),
		 &(mup->mu_fminus), &(mup->mu_sigminus), &tflag);
	mup->mu_flag = tflag;
	return nitems;
}

int outmu(FILE *fp, MUREFL *mup)
{ /* output the MUREFL mup to the stream fp. */
	int	tfl;

	tfl = mup->mu_flag;
	return fprintf(fp,
	/* alternate formatting
	    "%4hd%4hd%4hd %8.6f %9.4f %9.4f %9.4f %9.4f %9.4f %9.4f %2d\n", */
	    "%4hd%4hd%4hd %6.4f %8.2f %8.2f %8.2f %8.2f %8.2f %8.2f %2d\n",
			 mup->mu_hkl[0], mup->mu_hkl[1], mup->mu_hkl[2],
			 mup->mu_stl, mup->mu_f, mup->mu_sig, mup->mu_fminus,
			 mup->mu_sigminus, mup->mu_fplus, mup->mu_sigplus, tfl);
}

int stlcmp(MUREFL *mu0, MUREFL *mu1)
{ /* This is a sin(theta) compare routine. */
	double	sdiff;

	if ((sdiff = mu0->mu_stl - mu1->mu_stl) < 0) return -1;
	if (sdiff > 0) return 1;
	return 0;
}

int murcmp(MUREFL *mu0, MUREFL *mu1)
{ /*	This is a compare routine returing -1 if mu0 is less than mu1 and
	+1 if mu0 is greater than mu1 */
	register int	i, diff;

	for (i = 0; i < 3; i++)
	   {
		if ((diff = mu0->mu_hkl[i] - mu1->mu_hkl[i]) < 0) return -1;
		else if (diff > 0) return 1;
	   }
	return 0;
}

int keep_dimen(DATARUN *drp)
{ /*	This looks in a DATARUN structure and stores three numbers--
	widrat, htrat, and hto2--as static globals.
	Widrat is the ratio of the abo_xob value to the true x coordinate on
	the detector face; it's typically 1, 2, or 4.
	Htrat is the ratio of the abo_yob value to HALF OF the true y
	coordinate, with the proviso that the shift value determines which
	vertical half of the detector is being referred to.  The routine
	returns an int containing (widrat << 24) | (htrat << 16) | hto2).
	If called with a NULL argument, it simply returns using the values
	that it had already recorded in a previous call. */
	int		locwid, locht;
	static int	widrat, htrat, hto2;
	
	if (NULL != drp) /* new calculation of these values */
	  {
		if (0 >= (locwid = drp->run_width)) locwid = 512;
		if (0 >= (locht = drp->run_height)) locht = 512;
		widrat = (locwid + 255) / 256;	if (widrat < 1) widrat = 1;
		htrat = (locht + 511) / 512;	if (htrat < 1) htrat = 1;
		hto2 = locht / 2;
	  }
	return	(widrat << 24) | (htrat << 16) | hto2;
}

int abptox(ABBROB *abp)
{
	static int	widr = -1;

	if (widr < 0) widr = (keep_dimen(NULL)) >> 24; /* get widr if needed */
	return	(((int)(abp->abo_xob)) * widr) + widr / 2;
}

int abptoy(ABBROB *abp)
{
	int		j, kres;
	static int	htr = -1;
	static int	ht2;

	if (htr < 0) /* get a starting value for htr, ht2 if needed */
	  {
		kres = keep_dimen(NULL);
		htr = (kres >> 16) & 0xFF;	ht2 = kres & 0xFFFF;
	  }
	j = (((int)(abp->abo_yob)) * htr +
		ht2 * (abs(abp->abo_shift) % 2)) + htr / 2;
	if (j < 1) j = ht2;	return j;
}

int refvec(ABBROB *aobp, RUNWORK *rup, int vecrefined, int nvp, double ss[])
{ /*	This obtains either the local exit-beam vector or the local
	scattering vector sv associated with
	the ABBROB structure pointed to by aobp.  It returns 0 if
	totally unsuccessful and -1, 1, or 2 if successful depending on
	how well the `pix2cm' routine succeeds in translating from
	pixels to centimeters.  The `rup' argument enables the routine
	to figure out which set of run parameters (detector distance,
	orientation matrix, etc.) to apply to this refl.
	`ss' is returned as follows:
		value of `vecrefined'		contents of `ss'
		--------------------		----------------
		SC_DIFFRACTION | SC_VECONLY	diffraction vector sx,sy,sz
		SC_EXITBEAM | SC_VECONLY	exit-beam vector k'x,k'y,k'z
		SC_DIFFRACTION			spherical-harmonic expansion
						 of diffraction-beam parameters
		SC_EXITBEAM			spherical-harmonic expansion
						 of exit-beam parameters */
	int		retval;
	double		cosom, omega, rni, rni2, sinom, v0, v1, v2, zx, zy, zz;
	float		rxcm, rxp, rycm, ryp;

	rxp = abptox(aobp);	ryp = abptoy(aobp); /* actual pixel coords */
	if (!(retval = pix2cm(rxp, ryp, &rxcm, &rycm, 0))) return retval;
	rxcm -= rup->rp_x0;	rycm -= rup->rp_y0;
	rni2 = rxcm * rxcm + rycm * rycm + rup->rp_dsq; rni = sqrt(rni2);
	v0 = rup->rp_d / rni; v1 = rxcm / rni;	v2 = rycm / rni;
	/* rotate this by the detector tilt and swing-angle matrices */
	zx = rup->rp_q21[0][0] * v0 + rup->rp_q21[0][1] * v1 +
		rup->rp_q21[0][2] * v2;
	zy = rup->rp_q21[1][0] * v0 + rup->rp_q21[1][1] * v1 +
		rup->rp_q21[1][2] * v2;
	zz = rup->rp_q21[2][0] * v0 + rup->rp_q21[2][1] * v1 +
		rup->rp_q21[2][2] * v2;
	/* if we're looking for the diffraction vector, we add one to the
	 X component here; if we're looking for the exit beam, we don't */
	if (vecrefined & SC_DIFFRACTION) zx += 1.;
	/* We rotate the definition of the scattering vector so the diffraction
	 vectors are indexed to the reference position of the scanning axis. */
	omega = aobp->abo_omob * RADC / 100.;
	cosom = cos(omega);	sinom = sin(omega);
	if (vecrefined & SC_VECONLY)
	  {
		ss[0] = cosom * zx - sinom * zy;
		ss[1] = sinom * zx + cosom * zy; ss[2] = zz;	return retval;
	  }
	v0 = cosom * zx - sinom * zy; v1 = sinom * zx + cosom * zy; v2 = zz;
	ss[0] = 1.;
	if (nvp < 2) return retval;
	ss[1] = -SQ1P5 * v2;
	ss[2] = -SQ6 * v1; ss[3] = SQ6 * v0;
	if (nvp < 5) return retval;
	ss[4] = SQ1P25 * (2.* v2 * v2 - v1 * v1 - v0 * v0);
	ss[5] = -SQ30 * v2 * v0;
	ss[6] = SQ30 * v2 * v1;
	ss[7] = SQ7P5 * (v1 * v1 - v0 * v0);
	ss[8] = SQ30 * v1 * v0;
	if (nvp == 10)	/* one more component in this case */
	  ss[9] = SQ10OV3 * (v0 * v0 + v1 * v1 + v2 * v2);
	/* if we decide to allow third or higher-order harmonics,
	 they'll go here:
	if (nvp < 11) return retval;	*/
	return retval;
}

static int calc_euler(float lrec[3][3], double zeuler[])
{ /*	This extracts the eternal verities from the real-space unit cell matrix:
	first, the unit cell lengths and angles. It returns EOK on success,
	ESINGULAR when a matrix is univertible, and EALLOC on memory failures*/
	register int	i, j, k;
	double		ca, cas, cb, cg, dotp, len, sa, sas, sb, sg;
	double		abc[3], ang[3];
	double		ar0[3][3], ir[3][3], q[3][3], rreal[3][3];

	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++) ir[i][j] = lrec[i][j];
	if (dinv3(ir, rreal)) return ESINGULAR;
	for (i = 3; i--; )	/* unit cell lengths: */
	   {
		for (len = 0, j = 3; j--; ) /* length^2 = norm^2 of row */
			len += rreal[i][j] * rreal[i][j];
		abc[i] = sqrt(len);
	   }
	for (i = 0; i < 3; i++)	/* unit cell angles: */
	   { /* determined from dot products, viz.
			gamma = acos((a.b) / |a||b|) */
		j = (i == 2 ? 0 : (i+1));	k = 3 - i - j;
		dotp = (rreal[j][0] * rreal[k][0] + rreal[j][1] * rreal[k][1] +
			rreal[j][2] * rreal[k][2]);
		ang[i] = (fabs(dotp) > DFUZZ ?
		  acos(dotp / (abc[j] * abc[k])) / RADC : 90.);
	   }
	/* now construct the real-space unit cell matrix at the standard
	 position (a along X, b along Y and X). */
	ca = cos(ang[0] * RADC);	cb = cos(ang[1] * RADC);
	cg = cos(ang[2] * RADC);
	sa = sin(ang[0] * RADC);	sb = sin(ang[1] * RADC);
	sg = sin(ang[2] * RADC);
	cas = (cb * cg - ca) / (sa * sg); sas = sqrt(1. - cas * cas);
	ar0[0][0] = abc[0];	ar0[0][1] = ar0[0][2] = ar0[1][2] = 0.;
	ar0[1][0] = abc[1] * cg;	ar0[1][1] = abc[1] * sg;
	ar0[2][0] = abc[2] * cb;	ar0[2][1] = -abc[2] * cas * sb;
	ar0[2][2] = abc[2] * sas * sb;
	/* The orthogonality matrix Q that we want is obtained from
		A0 = Q * orec
	where orec is the OBSERVED reciprocal space unit cell matrix and
	A0 is the recip.space unit cell matrix at a along X, b along Y and X.  
	But the matrix we just found (ar0) is the inverse of A0.
	Thus inv(Q) = orec * ar0. */
	dmm3(q, ir, ar0);		/* get nearly orthogonal matrix Q */
	cleanorthog(q);	/* make the orthogonal matrix more precise */
	/* now we need to transpose this to get its inverse */
	ca = q[0][1];	q[0][1] = q[1][0];	q[1][0] = ca;
	ca = q[0][2];	q[0][2] = q[2][0];	q[2][0] = ca;
	ca = q[1][2];	q[1][2] = q[2][1];	q[2][1] = ca;
	qtoangs(q, zeuler); /* get the rotation angles from the q matrix */
	return EOK;
}

void workrun(DATARUN *drp, RUNWORK *rup, CRYSTAL *xhp, int shadd)
{	/* This preserves the relevant parameters from a datarun header
	in a structure used in calculating diffraction vectors. */
	int	i, j;
	double	c0, c1, c2, cossw, costi, s0, s1, s2, sinsw, sinti, swing, tilt;
	double	zeuler[3];
	float	zar[3][3];

		/* first, the easy ones: */
	rup->rp_shadd = shadd;	/* shift addend to identify run */
	rup->rp_d = drp->run_d0pos[0];	/* xtal-to-detector dist */
	rup->rp_dsq = rup->rp_d * rup->rp_d;	/* square of that */
	rup->rp_x0 = drp->run_d0pos[1];	/* linear offset in X */
	rup->rp_y0 = drp->run_d0pos[2];	/* linear offset in Y */
	/* stepsize between frames in radians: */
	if (fabs(drp->run_stepreal) > FUZZ)
		rup->rp_delfm = drp->run_stepreal * RADC;
	else	rup->rp_delfm = (((double)drp->run_stepsize) / 96.) * RADC;
		/* now, the tedious one: the swing/tilt matrix  */
	swing = drp->run_d0ang[0];	tilt = drp->run_d0ang[1];
	cossw = cos(swing * RADC);	sinsw = sin(swing * RADC);
	costi = cos(tilt * RADC);	sinti = sin(tilt * RADC);
	rup->rp_q21[0][0] = -cossw;
	rup->rp_q21[0][1] = -sinsw * costi;
	rup->rp_q21[0][2] = -sinsw * sinti;
	rup->rp_q21[1][0] = sinsw;
	rup->rp_q21[1][1] = -cossw * costi;
	rup->rp_q21[1][2] = -cossw * sinti;
	rup->rp_q21[2][0] =  0.;
	rup->rp_q21[2][1] = -sinti;
	rup->rp_q21[2][2] =  costi;
		/* save the real-space unit cell */
		/* now, the tricky one: the matrix that rotates us from
		the position associated with the start of the data run
		to the position associated with a* along X, b* along Y.
		the reference position (omega, chi, phi) down to (0,0,0).
		If this isn't squirrely code, I don't know my rodents.
		*/
	(void)calc_euler(drp->run_recip, zeuler);
	c0 = cos(zeuler[0]);	s0 = sin(zeuler[0]);
	c1 = cos(zeuler[1]);	s1 = sin(zeuler[1]);
	c2 = cos(zeuler[2]);	s2 = sin(zeuler[2]);
	rup->rp_qz[0][0]=  s0*c1*s2+c0*c2; rup->rp_qz[0][1]= -s0*c1*c2+c0*s2;
	rup->rp_qz[0][2]=  s0*s1;
	rup->rp_qz[1][0]= -c0*c1*s2+s0*c2; rup->rp_qz[1][1]=  c0*c1*c2+s0*s2;
	rup->rp_qz[1][2]= -c0*s1;
	rup->rp_qz[2][0]= -s1*s2;	   rup->rp_qz[2][1]=  s1*c2;
	rup->rp_qz[2][2]=  c1;
	mm3(zar, rup->rp_qz, drp->run_recip);
	(void)inv3(drp->run_recip, rup->rp_areal);
	for (i = 3; i--; ) for (j = 3; j--; )
		rup->rp_areal[i][j] /= xhp->xt_lambda;
}

int headmem(CRYSTAL **xhdp, DATARUN **rhdp, int ndruns)
{ /*	This reserves memory for a CRYSTAL header and `ndruns' DATARUN headers.
	If **xhdp itself is NULL, we delete memory for this mess.
	It returns without doing anything if *xhdp is nonzero */
	PAU		hmu;
	static size_t	tspace, oldspace;
	static PAU	hmu0;
	static char	headmsg[] = "header memory in headmem";

	if (NULL == xhdp)
	  {
		memzap((AVOID **)&(hmu0.vpt), headmsg);	oldspace = 0;
		return EOK;
	  }
	*xhdp = NULL;	*rhdp = NULL;
	if (NULL != *xhdp) return EOK;
	tspace = sizeof (CRYSTAL) + ndruns * sizeof (DATARUN);
	if ((NULL == hmu0.vpt) || (tspace > oldspace))
	  {
		if (NULL == (hmu0.vpt = callop(1, tspace, headmsg)))
			return EALLOC;
		oldspace = tspace;
	  }
	hmu = hmu0;
	*xhdp = (CRYSTAL *)(hmu.vpt);	hmu.ucpt += sizeof (CRYSTAL);
	*rhdp = (DATARUN *)(hmu.vpt);	return EOK;
}

void mpack(short hup[], u_char hp[], short intlno)
{ /*	This converts a 3-short input hkl value hup into a 4-character
	output hkl value hp.  No symmetry conversion is done here.
	Rules:
		hp[0] = |hup[0]|, hp[1] = |hup[1]|, hp[2] = |hp[2]|
		hp[3] is the sign byte, defined according to the following
		convention:
			sign	sign	sign	hp[3]
			hup[0]	hup[1]	hup[2]
			-----------------------------
			+	+	+	0
			-	-	-	1
			+	+	-	2
			-	-	+	3
			+	-	+	4
			-	+	-	5
			+	-	-	6
			-	+	+	7
	The identities of h and k are reversed in this calculation
	if we're in a monoclinic spacegroup.
	hk0,h0l,0kl treated specially to ensure that
		MPACK(MUNPACK(hkls)) = hkls and MUNPACK(MPACK(hkl)) = hkl. */

	register int	i;
	short		j;
	u_char		is, t;

	for(i = 0; i < 3; i++)	/* get the first three bytes */
		hp[i] = (u_char)(((j=hup[i]) < 0) ? -j : j);
	is = 0;			/* now get the sign byte */
	if(hup[1] < 0) is = 4;		/* negative k */
	if(hup[2] < 0) is += 2;		/* negative l */
	if(hup[0] < 0) is = 7 - is;	/* negative h */
	if (!hup[2])	/* special case: l == 0 */
	  {
		if (!hup[0]) is = (hup[1] < 0);
		else if (is % 2) is -= 2;
	  }
	if (!hup[1])	/* special case: k == 0 */
	  {
		if (!hup[0]) is = (hup[2] < 0);
		else if (is > 3) is -= 4;
	  }
	if (!hup[0] && (is > 3)) is = 7 - is;	/* h == 0 */
	hp[3] = is;		/* stuff it in the right place */
	if ((2 < intlno) && (intlno < 16))	/* correct monoclinics */
	  {
		t = hp[0]; hp[0] = hp[1]; hp[1] = t;
	  }
	return;
}

void munpack(u_char *hp, short *hup, short intl)
{ /*	This performs the converse operation to mpack, i.e.
	it converts a four-byte hkl expression to a 3-short expression.
	Monoclinics are appropriately corrected.
	See comments in "mpack" for sign convention. */
	register int	i;
	short		is;

	if ((2 < intl) && (intl < 16))
	  {
		hup[0] = hp[1]; hup[1] = hp[0];
	  }
	else {
		hup[0] = hp[0]; hup[1] = hp[1];
	     }
	hup[2] = hp[2];
	is = (short)hp[3];			/* record with the sign */
	if ((is & 04) != 0) hup[1] = -hup[1];	/* get the hkl signs */
	if ((is & 02) != 0) hup[2] = -hup[2];
	if ((is & 01) != 0) for(i = 0; i < 3; i++) hup[i] = -hup[i];
}

int hklcmp(u_char *h0, u_char *h1)
{ /* This returns >0 if the reflection (h ,k ,l )= h0 shpuld
	fall LATER in the file than       (h',k',l')= h1.
	It returns 0 if the reflections are the same, and it returns
	a negative number if h0 should fall earlier in the file.
	No symmetry transformations are performed inside this routine;
	if any are required, they are done in "symtr". */
	register int		diff, i;
	register u_char		*p,*q;

	p = h0; q = h1;	diff = 0;
	for (i = 0; i < 4; i++) if (0 != (diff = (*p++ - *q++))) return diff;
	return(diff);
}

static u_char logeor(u_char uc)
{	/* this returns 1 if the number of `on' bits in the argument is odd;
	it returns 0 if the number of `on' bits in the argument is even.
	Only the three lowest-order bits are examined.  Thus
	f(0)=f(3)=f(5)=f(6) = 0 and f(1)=f(2)=f(4)=f(7) = 1. */
	return (((int)uc) >> 2) ^ ((((int)uc) & 2) >> 1) ^ (((int)uc) & 1);
}

static short spgtopat(short scode)
{ /*	Returns the Patterson spacegroup associated with any spacegroup */
	static short	pato[] = { 0,		/* nonsense value */
   2,   2,  10,  10,  12,  10,  10,  12,  12,  10, /* spacegroups   1 -  10 */
  10,  12,  10,  10,  12,  47,  47,  47,  47,  65, /* spacegroups  11 -  20 */
  65,  69,  71,  71,  47,  47,  47,  47,  47,  47, /* spacegroups  21 -  30 */
  47,  47,  47,  47,  65,  65,  65,  65,  65,  65, /* spacegroups  31 -  40 */
  65,  65,  69,  69,  71,  71,  47,  47,  47,  47, /* spacegroups  41 -  50 */
  47,  47,  47,  47,  47,  47,  47,  47,  47,  47, /* spacegroups  51 -  60 */
  47,  47,  65,  65,  65,  65,  65,  65,  69,  69, /* spacegroups  61 -  70 */
  71,  71,  71,  71,  83,  83,  83,  83,  87,  87, /* spacegropus  71 -  80 */
  83,  87,  83,  83,  83,  83,  87,  87, 123, 123, /* spacegroups  81 -  90 */
 123, 123, 123, 123, 123, 123, 139, 139, 123, 123, /* spacegroups  91 - 100 */
 123, 123, 123, 123, 123, 123, 139, 139, 139, 139, /* spacegroups 101 - 110 */
 123, 123, 123, 123, 123, 123, 123, 123, 139, 139, /* spacegroups 111 - 120 */
 139, 139, 123, 123, 123, 123, 123, 123, 123, 123, /* spacegroups 121 - 130 */
 123, 123, 123, 123, 123, 123, 123, 123, 139, 139, /* spacegroups 131 - 140 */
 139, 139, 147, 147, 147, 148, 147, 148, 162, 164, /* spacegroups 141 - 150 */
 162, 164, 162, 164, 166, 164, 162, 164, 162, 166, /* spacegroups 151 - 160 */
 166, 162, 162, 164, 164, 166, 166, 175, 175, 175, /* spacegroups 161 - 170 */
 175, 175, 175, 175, 175, 175, 191, 191, 191, 191, /* spacegroups 171 - 180 */
 191, 191, 191, 191, 191, 191, 191, 191, 191, 191, /* spacegroups 181 - 190 */
 191, 191, 191, 191, 200, 202, 204, 200, 204, 200, /* spacegroups 191 - 200 */
 200, 202, 202, 204, 200, 204, 221, 221, 225, 225, /* spacegroups 201 - 210 */
 229, 221, 221, 229, 221, 225, 229, 221, 225, 229, /* spacegroups 211 - 220 */
 221, 221, 221, 221, 225, 225, 225, 225, 229, 229, /* spacegroups 221 - 230 */
 123, 123, 123, 123, 123, 123, 123, 123, 139, 139, /* nonexistent 231 - 240 */
 139, 139, 147, 147, 147, 248, 147, 248, 262, 264, /* rhombohedr. 241 - 250 */
 262, 264, 262, 264, 266, 264, 262, 264, 262, 266, /* spacegroups 151 - 160 */
 266, 262, 262, 264, 264, 266, 266 } ;		   /* spacegroups 161 - 167 */
 
	return (scode < 0) ? 0 : ((scode > 267) ? 0 : pato[((int)scode)]);
}

void symtr(u_char sp[], short scode)
{ /*	General symmetry transformation routine.
	This reads in a 4-byte representation of an hkl
	(|h|,|k|,|l|,signbyte for non-monoclinic spacegroups;
	 |k|,|h|,|l|,signbyte for monoclinics)
	and writes another back out with symmetry transformations performed.
	"scode" is the International Tables' code for the spacegroup. */
	register int	i, j, nh, nk;
	int		flip, patt;
	short		h0[3];

	patt = spgtopat(scode);	/* determine the Patterson spacegroup */
	switch (patt)	/* make transformation decisions based on that */
	      {
		case	2: default:	return;	/* triclinics: no transform */
		case	10:	case	12: /* monoclinics: (hkl)=(-h,k,-l) */
			if (sp[3] > 3) sp[3] = ((int)(9 - sp[3])) % 4;
			if (!(sp[0])) sp[3] &= ~1;	/* h0l's are centric */
			/* if either h or l is zero, the sign byte is 0 or 1 */
			if (((sp[1] == 0) || (sp[2] == 0)) && sp[3] > 1)
				sp[3] -= 2;
			return;
		case	47:	case	65:	case	69:	case	71:
			/* orthorhombics: hkl=(-h,-k,l)=(-h,k,-l),(h,-k,-l) */
			for (i = 0; i < 3; i++)
			   if (!sp[i])
			     {
				sp[3] = 0; return;
			     }
			sp[3] = logeor(sp[3]);
			return;
		case	83:	case	87:	case	88: /* tetragonal 4 */
			/* (h,k,l)=(-k,h,l)=(-h,-k,l): sign byte is 0 or 1 */
			if (sp[3] > 3)
			  {
				nh = sp[0];	sp[0] = sp[1];	sp[1] = nh;
				sp[3] -= 4;
			  }
			/* (0kl) = (k0l so we make all 0kl's equal to k0l. */
			if (!sp[0])
			  {
				sp[0] = sp[1];	sp[1] = 0;
			  }
			if (sp[3] > 1) sp[3] = 3 - sp[3];
			if (!sp[2]) sp[3] = 0;	/* hk0 is centric */
			return;
		case	123:	case	139:	/* tetragonal 422 */
			nh = sp[0];	nk = sp[1];
			if (nh == nk)
			  {
				sp[3] = 0;	return;
			  }	
			if (nh < nk)
			  {
				sp[0] = nk; sp[1] = nh; sp[3] = 7 - sp[3];
			  }
			for (i = 0; i < 3; i++) if (sp[i] == 0)
			   {
				sp[3] = 0; return;
			   }
			sp[3] = logeor(sp[3]);	return;
		case 147: case 148: case 162: case 164: case 166: /* trigonal*/
			/* go thru hk,ki,ih to find pair with the same sign */
			nk = (short)sp[0] - (short)sp[1];	nh = abs(nk);
			j = (nk < 0);	i = 1 - j;
			if (sp[3] > 3)
			  {
				sp[j] = sp[i];	sp[i] = nh;
				sp[3] = (j ? sp[3] - 4 : 7 - sp[3]);
			  }
			if (sp[3] > 1)
			  {
				if (!sp[2]) sp[3] -= 2;
				if (!sp[1])
				  {
					sp[1] = sp[0];	sp[0] = 0;
					sp[3] = 3 - sp[3];
				  }
				if (!sp[0])
				  {
					sp[0] = sp[1];	sp[1] = 0;
					sp[3] = 3 - sp[3];
				  }
			  }
			if (!sp[2] && !sp[0])
			  {
				sp[0] = sp[1];	sp[1] = 0;
				sp[3] = 1 - sp[3];
			  }
			if ((!sp[1]) && sp[3] > 1)
			  {
				sp[1] = sp[0]; sp[0] = 0; sp[3] = 3 - sp[3];
			  }
			if (scode <= 148) return; /* no twofolds: we're done */
			if (patt == 162) /* Patterson spacegroup 162 == P-31m*/
			  {
				if (sp[0] < sp[1]) /* h,k,l = -k,-h,-l */
				  {
					j = sp[0]; sp[0] = sp[1]; sp[1] = j;
					sp[3] = (1 -(sp[3] & 1))+(sp[3] & 2);
				  }
				else if (sp[0] == sp[1]) sp[3] &= ~1;
			  }
			else /* Patterson 164 (P-3m1) and 166 (R-3m) */
			  {
				if (sp[3] > 1) /* h,k,l = k,h,-l */
				  {
					j = sp[0]; sp[0] = sp[1]; sp[1] = j;
					sp[3] -= 2;
				  }
				if ((!sp[2]) && sp[0] < sp[1])
				  {
					j = sp[0]; sp[0] = sp[1]; sp[1] = j;
				  }
				if ((!sp[0]) || !sp[1]) sp[3] = 0;
			  }
			return;
		case	175:	case	191:	/* hexagonal */
			/* go thru hk,ki,ih to find pair with the same sign */
			nk = (short)sp[0] - (short)sp[1];	nh = abs(nk);
			j = (nk < 0);	i = 1 - j;
			if (sp[3] > 3)
			  {
				sp[j] = sp[i];	sp[i] = nh;
				sp[3] = (j ? sp[3] - 4 : 7 - sp[3]);
			  }
			/* if necessary, perform (h,k,l) => (-h,-k,l) */
			if (sp[3] > 1) sp[3] = 3 - sp[3];
			if (!sp[2]) sp[3] = 0;	/* hk0 is centric */
			if (!sp[0])
			  { /* h0l=0hl*/
				sp[0] = sp[1]; sp[1] = 0;
			  }
			if (patt == 175) return;	/* that's all for P6 */
			if (sp[0] < sp[1])
			  { /* If there are 2folds perpendicular to the 6fold,
				we can require |h| > |k| using (h,k,l) =>
				(-h,-k,l) => (-k,-h,-l) = ~(k,h,l) */
				nh = sp[0];	sp[0] = sp[1];
				sp[1] = nh;	sp[3] = 1 - sp[3];
			  }
			/* result is centric if any index 0 or h == k. */
			if ((!sp[0]) || (!sp[1]) ||
			 (!sp[2]) || (sp[0] == sp[1])) sp[3] = 0;
			return;
		case 200: case 202: case 204: /* cubic, no 2folds: cyclic */
			/* by convention, set h to be the maximum index here */
			nh = (sp[1] > sp[0] ? (sp[2] > sp[1] ? 2 : 1) :
				(sp[2] > sp[0] ? 2 : 0));
			if (nh == 1)
			  {
				nh = sp[1]; sp[1] = sp[2];
				sp[2] = sp[0]; sp[0] = nh;
			  }
			else if (nh == 2)
			  {
				nh = sp[2]; sp[2] = sp[1];
				sp[1] = sp[0]; sp[0] = nh;
			  }
			if ((sp[0] == sp[2]) && (sp[1] < sp[2]))
			  {
				sp[2] = sp[1];	sp[1] =  sp[0];
			  }
			for (i = 0; i < 3; i++) if (sp[i] == 0)
			   {
				sp[3] = 0; return;
			   }
			sp[3] = logeor(sp[3]);
			return;
		case	221:	case	225:	case	229:
			/* cubic with twofolds: we make h >= k >= l */
			nh = (sp[1] > sp[0] ? (sp[2] > sp[1] ? 2 : 1) :
				(sp[2] > sp[0] ? 2 : 0));
			if (nh)
			  {
				nk = sp[0]; sp[0] = sp[nh]; sp[nh] = nk;
			  }
			if (sp[2] > sp[1])
			  {
				nk = sp[2]; sp[2] = sp[1]; sp[1] = nk;
			  }
			for (i = 0; i < 3; i++)
			 if ((sp[i] == 0) || ((i < 2) && (sp[i] == sp[i+1])))
			   {
				sp[3] = 0; return;
			   }
			sp[3] = logeor(sp[3]);
			return;
		case	248:	case	262:	case	264:	case	266:
			/* rhombohedrals, indexed rhombohedrally */
			munpack(sp, h0, scode);	flip = 0;
			j = (h0[0] < 0) + (h0[1] < 0) + (h0[2] < 0);
			if (j > 1) /* >=2 signs neg: flip 'em all, remember */
			  {
				flip = 1;for (i = 3; i--; ) h0[i] = -h0[i];
			  }
			/* see which of the signed indices is smallest */
			j = (h0[0] < h0[1] ? (h0[0] < h0[2] ? 0 : 2) :
				(h0[1] < h0[2] ? 1 : 2));
			/* make that index "l" by cyclic permutation */
			if (j == 0) /* h is smallest: permute hkl->klh */
			  {
				nh = h0[1];	nk = h0[2];
				h0[2] = h0[0];	h0[0] = nh;	h0[1] = nk;
			  }
			else if (j == 1) /* k smallest: permute hkl->lhk */
			  {
				nh = h0[2];	nk = h0[0];
				h0[2] = h0[1];	h0[0] = nh;	h0[1] = nk;
			  }
			/* a squirrely special case: h == l.  We want
			either (hkh) -> (khh) or (hhk) */
			if (h0[0] == h0[2])
			  {
				if (h0[1] < h0[0])
				  {
					h0[2] = h0[1];	h0[1] = h0[0];
				  }
				else {
					h0[0] = h0[1];	h0[1] = h0[2];
				     }
			  }
			/* now <= 1 of the indices is negative; it's "l".
			An ugly ambiguity: (0 k -l) = (-l 0 k) = -(l 0 -k).
			To resolve this, we require that h != 0 . */
			if (h0[0] == 0)
			  {
				h0[0] = -h0[2];	h0[2] = -h0[1];
				h0[1] = 0;	flip = 1 - flip;
			  }
			/* Patterson R-3: we're done. In Patterson R-3m, etc.
			 we require that |h| >= |k| by (hkl) = (-k,-h,-l). */
			if (patt > 248)
			  {
				if (h0[0] < h0[1])
				  {
					flip = 1 - flip; nk = h0[0];
					h0[0] = h0[1];	h0[1] = nk;
				  }
				/* h==k or k==l (or h==l) are centric */
				if ((h0[0] == h0[1]) || (h0[1] == h0[2]))
					flip = 0;
				/* special case: (h,0,-l)=(l,0,-h), so for
				 these we require |h| > |l|, arbitrary choice*/
				if ((!h0[1]) && (h0[2] < 0) && (h0[0] < -h0[2]))
				  {
					nh = -h0[2];	h0[2] = -h0[0];
					h0[0] = nh;
				  }
			  }
			sp[0] = h0[0];	sp[1] = h0[1];	sp[2] = abs(h0[2]);
			sp[3] = flip + 2 * (h0[2] < 0);	return;
	      }	/* end of switch on `patt' variable */
}

void splitmr(MERGEDREF *mrpin, MERGEDREF *mrppri, MERGEDREF *mrpbij)
{ /*	This takes the input multiref reflection pointed to by mrpin
	and copies the primary observations from it into mrppri
	and the Bijvoet observations into mprbij.  */
	cpymem((AVOID *)mrppri, (AVOID *)mrpin,
		(REFBYTES + (unsigned)mrpin->mr_nobpri*sizeof (ABBROB)));
	cpymem((AVOID *)mrpbij, (AVOID *)mrpin, REFBYTES);
	if (mrpin->mr_nobbij) cpymem((AVOID *)&(mrpbij->mr_ob[0]),
	 (AVOID *)&(mrpin->mr_ob[mrpin->mr_nobpri]),
	 ((unsigned)mrpin->mr_nobbij * sizeof (ABBROB)));
	mrppri->mr_nob = mrppri->mr_nobpri;
	mrpbij->mr_nob = mrpbij->mr_nobbij;
	mrppri->mr_nobbij = mrpbij->mr_nobpri = 0;
	return;
}

int screw_ok(short ihkl[], short ispgp)
{	/* This returns 0 if and only if the reflection ihkl is disallowed
	 due to a screw-axis exclusion in the spacegroup. */
	if (ihkl[0] && ihkl[1] && ihkl[2]) return 1;
	switch (ispgp)
	      {
		case	4:	/* P2(1): disallow 0k0, k odd */
			return	(ihkl[0] || ihkl[2] || !(ihkl[1] % 2));
		case	17:	/* P222(1): disallow 00l, l odd */
		case	20:	/* C222(1): same as P222(1) */
		case	77:	/* P4(2): same as above */
		case	93:	/* P4(2)22: same as above */
		case	173:	/* P6(3): same as above */
		case	182:	/* P6(3)22: same as above */
			return	(ihkl[0] || ihkl[1] || !(ihkl[2] % 2));
		case	18:	/* P2(1)2(1)2: disallow h00 or 0k0, h,k odd */
			if (ihkl[2]) return 1;
			if (ihkl[0]) return (ihkl[1] || !(ihkl[0] % 2));
			return (!(ihkl[1] % 2));
		case	19:	/* P2(1)2(1)2(1): disallow odd h00,0k0,00l */
		case	198:	/* P2(1)3: same as above */
		case	208:	/* P4(2)32: same as above */
			if (ihkl[2])
			  return (ihkl[0] || ihkl[1] || !(ihkl[2] % 2));
			if (ihkl[1])
			  return (ihkl[0] || !(ihkl[1] % 2));
			return !(ihkl[0] % 2);
		case	76:	/* P4(1): 00l must have l=4n */
		case	78:	/* P4(3): same as P4(1) */
		case	80:	/* I4(1): same as above */
		case	91:	/* P4(1)22: same as above */
		case	95:	/* P4(3)22: same as above */
		case	98:	/* I4(1)22: same as above */
			return (ihkl[0] || ihkl[1] || !(ihkl[2] % 4));
		case	90:	/* P42(1)2: disallow odd h00 and odd 0k0 */
			if (ihkl[1]) return ((ihkl[0]) || (ihkl[2]) ||
				!(ihkl[1] % 2));
			else return ((ihkl[2]) || !(ihkl[0] % 2));
		case	92:	/* P4(1)2(1)2: 00l,l=4n;h00,h=2n;0k0,k=2n*/
		case	96:	/* P4(3)2(1)2: same as above */
			if (ihkl[1]) return ((ihkl[0]) || (ihkl[2]) ||
				!(ihkl[1] % 2));
			if (ihkl[0]) return (ihkl[2] || !(ihkl[0]%2));
			return !(ihkl[2] % 4);
		case	94:	/* P4(2)2(1)2: 00l is l=2n, h00 is h=2n */
			if (ihkl[1]) return ((ihkl[0]) || (ihkl[2]) ||
				!(ihkl[1] % 2));
			if (ihkl[0]) return (ihkl[2] || !(ihkl[0]%2));
			return !(ihkl[2] % 2);
		case	144:	/* P3(1): 00l = 3n */
		case	145:	/* P3(2): 00l = 3n */
		case	151:	/* P3(1)12: same as above */
		case	152:	/* P3(1)21: same as above */
		case	153:	/* P3(2)12: same as above */
		case	154:	/* P3(2)21: same as above */
		case	171:	/* P6(2): same as above */
		case	172:	/* P6(4): same as above */
		case	180:	/* P6(2)22: same as above */
		case	181:	/* P6(4)22: same as above */
			return (ihkl[0] || ihkl[1] || !(ihkl[2] % 3));
		case	169:	/* P6(1): l = 6n */
		case	170:	/* P6(5): l = 6n */
		case	178:	/* P6(1)22: l = 6n */
		case	179:	/* P6(5)22: l = 6n */
			return (ihkl[0] || ihkl[1] || !(ihkl[2] % 6));
		case	210:	/* F4(1)32: h00=4n and cyclic permut.*/
		case	212:	/* P4(3)32: h00=4n and cyclic permut.*/
		case	213:	/* P4(1)32: h00=4n and cyclic permut.*/
		case	214:	/* I4(1)32: h00=4n and cyclic permut.*/
			if (ihkl[2])
			  return (ihkl[0] || ihkl[1] || !(ihkl[2] % 4));
			if (ihkl[1])
			  return (ihkl[0] || !(ihkl[1] % 4));
			return !(ihkl[0] % 4);
		default:	return 1;
	     }
}

int iscentric(u_char *phkl, short spaceg)
{ /* This returns 1 if and only if the reflection with packed indices
	phkl is centric in the spacegroup with international table # "spaceg".
	Note that in the monoclinic case we return true if phkl[0] == 0;
	that's because monoclinics are packed as khl, not hkl, and h0l is
	the only centric zone. */
	short	patt;

	patt = spgtopat(spaceg); /* determine the Patterson spacegroup */
	switch (patt)
	      {
		case	2: default:	return 0; /* triclinic: no centrics */
		case	10:	case	12:	/* note: monoclinic in khl */
			return !(phkl[0]); /* monoclinic: h0l only */
		case	47:	case	65:	case	69:	case	71:
			/* orthorhombic: 0kl, h0l, hk0 */
			return (!(phkl[0])) | (!(phkl[1])) | (!(phkl[2]));
		case	83:	case	87:	case	88:
			/* tetragonal, no twofolds (point group 4): hk0. */
			return !(phkl[2]);
		case	123:	case	139:
		/* tetragonal with twofolds (422):hhl, 0kl, h0l, hk0 */
			return (!(phkl[0])) | (!(phkl[1])) | (!(phkl[2])) ||
				(phkl[0] == phkl[1]);
		case 147: case 148: /* trigonal (point group 3): no centrics */
			return 0;
		case 162:	/* trigonal, 312: h == k */
			return (phkl[0] == phkl[1]) && (phkl[3] < 4);
		case 164: case 166: /* trigonal, 321: h0l, 0kl */
			return (!(phkl[0])) || (!(phkl[1])) ||
				((phkl[0] == phkl[1]) && (phkl[3] > 3));
		case	175:	/* hexagonal, point group 6: l==0 */
			return !(phkl[2]);
		case	191:	/* hexagonal 622: hk0,h0l,0kl,hhl */
			return ((!(phkl[2])) | (!(phkl[1])) | (!(phkl[0]))) ||
				(phkl[0] == phkl[1]);
		case 200: case 202: case 204: /* cubic,no 2folds: hk0,h0l,0kl*/
			return (!(phkl[2])) | (!(phkl[1])) | (!(phkl[0]));
		case	221:	case	225:	case	229:
			/* cubic with twofolds: hhl, hkh, hkk, 0kl, h0l, hk0 */
	 		return (!(phkl[0])) | (!(phkl[1])) | (!(phkl[2])) |
			 (phkl[0] == phkl[1]) | (phkl[0] == phkl[2]) |
			 (phkl[1] == phkl[2]);
		case	248:	return 0; /* R3 indexed rhombo: no centrics */
		case	262:	case	264:	case	266:
			/* R32,etc: indexed rhombohedrally: hhl,hkh,hkk */
			return (((phkl[0] == phkl[2]) && (!(phkl[3] & 2))) ||
			 ((phkl[0] == phkl[1]) && (phkl[3] < 4)) ||
			 ((phkl[1] == phkl[2]) && ((phkl[3] < 2) ||
			  (phkl[3] > 5))));
	     }
	/*NOTREACHED*/
}

static int duplicate_ob(ABBROB *abp0, ABBROB *abp1)
{ /*	This returns true iff. the x,y, and phi values are within
	4 pixels and 0.75 degrees of one another */
	if (abs(((int)abp0->abo_xob) - ((int)abp1->abo_xob)) > 2) return 0;
	if (abs(abp0->abo_omob - abp1->abo_omob) > 75) return 0;
	if (abs(((int)abp0->abo_yob) - ((int)abp1->abo_yob)) > 2 ||
	  ((abp0->abo_shift % 2) != (abp1->abo_shift % 2))) return 0;
	return 1;
}

int eliminate_dups(MERGEDREF *mrp)
{ /*	This looks through a MERGEDREF reflection.  If any two observations are
	seen to be effectively identical--i.e., measured at the same x,
	the same y, and the same phi (to +/- 2 pixels & +/- 2 frames),
	one of them is suppressed.
	The routine returns the number of duplicates tossed out.
	Change as of May 04: flag the observation with the smaller I/sigma. */
	int	abno, nduptoss;
	int	newobbij, newobpri, obno, oldnob, oldobbij, oldobpri;
	double	ios0, ios1;
	ABBROB	*abp, *obp, *obe;

	/* save the original number of primaries and Bijvoets */
	newobpri = oldobpri = mrp->mr_nobpri;
	newobbij = oldobbij = mrp->mr_nobbij;	oldnob = mrp->mr_nob;
	nduptoss = 0;
	/* look for and flag dups among the primaries */
	if (oldobpri > 1)
	 for (abp = &(mrp->mr_ob[0]),abno =0; abno < oldobpri -1; abno++,abp++)
	   {
		ios0 = abp->abo_sigma;
		ios0 = (ios0 < FUZZ) ? abp->abo_ii : abp->abo_ii / ios0;
		for (obp = &(mrp->mr_ob[abno+1]), obno = abno+1;
		 obno < oldobpri; obno++, obp++)
		  if ((obp->abo_shift >= 0) && (duplicate_ob(abp, obp)))
		    {
			nduptoss++; newobpri--; ios1 = obp->abo_sigma;
			ios1 = (ios1 < FUZZ) ? obp->abo_ii : obp->abo_ii /ios1;
			if (ios0 < ios1)
			  {
				abp->abo_shift = DUPLICATE_FLAG;	break;
			  }
			else obp->abo_shift = DUPLICATE_FLAG;
		    }
	   }
	/* look for and flag dups among the Bijvoets */
	if (oldobbij > 1)
	 for (abp = &(mrp->mr_ob[oldobpri]), abno = oldobpri;
	   abno < oldnob - 1; abno++, abp++)
	   {
		ios0 = abp->abo_sigma;
		ios0 = (ios0 < FUZZ) ? abp->abo_ii : abp->abo_ii / ios0;
		for (obp = &(mrp->mr_ob[abno+1]), obno = abno+1;
		 obno < oldnob; obno++, obp++)
		   if (obp->abo_shift > 0 && (duplicate_ob(abp, obp)))
		     {
			nduptoss++; newobbij--; ios1 = obp->abo_sigma;
			ios1 = (ios1 < FUZZ) ? obp->abo_ii : obp->abo_ii /ios1;
			if (ios0 < ios1)
			  {
				abp->abo_shift = DUPLICATE_FLAG;	break;
			  }
			else obp->abo_shift = DUPLICATE_FLAG;
		     }
	   }
	/* if we haven't thrown anything out, quit */
	if (!nduptoss) return nduptoss;
	/* go back through the refl and push the stack up whenever
	we find an observation with the duplicate flag set.  This kills 'em.*/
	obe = &(mrp->mr_ob[oldnob - 1]);
	for (abp = &(mrp->mr_ob[0]); abp <= obe; abp++)
	   {
		if (abp->abo_shift == DUPLICATE_FLAG)
		  {
			for (obp = abp + 1; obp <= obe; obp++)
			   {
				cpymem((AVOID *)(obp - 1), (AVOID *)obp,
					sizeof (ABBROB));
			   }
			obe--;	/* so we don't go past the end */
			abp--;	/* so we can check the first pushed-down one*/
		  }
	   }
	mrp->mr_nobpri = newobpri;	mrp->mr_nobbij = newobbij;
	mrp->mr_nob = newobpri + newobbij;
	return nduptoss;
}

/* end of mrutil.c */
