/* saxiim.c -- Routines for dealing with Siemens/Bruker images

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	11 Aug 2006: manage byteswapped formats without specific defines
	24 Jun 2006: include xgimage.h; the structures are there
	11 Jun 2006: somealpha moved from imagutil.c->here so it can be static
	30 May 2006: itodetype,itoinstance now static
	20 Nov 2005: inline argument declarations and strict prototyping
	 8 Nov 2005: replace static globals with various tricky bits
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	28 Aug 2004: routines used only here are declared as static; cosmetics
	 2 Jul 2004: convert loz to u_char in ascii_flow()
	27 Jun 2004: move saxiflows() and ascii_flow() here from image.c
	21 Jan 2004: license message inserted
	 8 Jan 2004: global aduperphot and local variables eperadu, eperphot
		replaced with elements of DATARUN
	19 Sep 2003: mksaxihead() sprintfs: typecast the xglongs to ints
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	12 Jun 2003: correct diagnostics in saxiflows();
		make errno.h reference universal so Linux doesn't whine
	 5 Jan 2003: reorganized source-file header
	15 Oct 2002: for long filenames, we retain the end, not the beginning
	25 Mar 2002: removed MSI attachments
	15 Sep 1999: in fromnotes(), skip the search through the keywords
		unless "tocolon" is at least 3.
	10 Sep 1999: in mksaxflows(), put blanks (not nulls) in the buffer
	15 Jun 1997: in ascii_flow(), provide for 16-bit overflows by keeping
		the `value' as (invalue - 65535) << 2, as with Mar data
	
   Purpose: reading and writing Siemens/Bruker images
   Cross references: image.c, imhead.c, imageout.c

*************************************************************************/
#define		NEED_ERRNO	1
#include	"xgimage.h"
#define	MAX_SAXII_LINES	1000
#define	MAXTYPELEN	10
/* defines for the Saxii header entries */
#define		SXH_FORMAT 	0
#define		SXH_VERSION	1
#define		SXH_HDRBLKS	2
#define		SXH_TYPE   	3
#define		SXH_SITE	4
#define		SXH_MODEL  	5
#define		SXH_USER   	6
#define		SXH_SAMPLE 	7
#define		SXH_SETNAME	8
#define		SXH_RUN		9
#define		SXH_SAMPNUM	10
#define		SXH_TITLE  	11
#define		SXH_NCOUNTS	12
#define		SXH_NOVERFL	13
#define		SXH_MINIMUM	14
#define		SXH_MAXIMUM	15
#define		SXH_NONTIME	16
#define		SXH_NLATE  	17
#define		SXH_FILENAM	18
#define		SXH_CREATED	19
#define		SXH_CUMULAT	20
#define		SXH_ELAPSDR	21
#define		SXH_ELAPSDA	22
#define		SXH_OSCILLA	23
#define		SXH_NSTEPS	24
#define		SXH_RANGE  	25
#define		SXH_START  	26
#define		SXH_INCREME	27
#define		SXH_NUMBER 	28
#define		SXH_NFRAMES	29
#define		SXH_ANGLES 	30
#define		SXH_NOVER64	31
#define		SXH_NPIXELB	32
#define		SXH_NROWS  	33
#define		SXH_NCOLS	34
#define		SXH_WORDORD	35
#define		SXH_LONGORD	36
#define		SXH_TARGET 	37
#define		SXH_SOURCEK	38
#define		SXH_SOURCEM	39
#define		SXH_FILTER 	40
#define		SXH_CELL   	41
#define		SXH_MATRIX 	42
#define		SXH_LOWTEMP	43
#define		SXH_ZOOM	44
#define		SXH_CENTER 	45
#define		SXH_DISTANC	46
#define		SXH_TRAILER	47
#define		SXH_COMPRES	48
#define		SXH_LINEAR	49
#define		SXH_PHD    	50
#define		SXH_PREAMP 	51
#define		SXH_CORRECT	52
#define		SXH_WARPFIL	53
#define		SXH_WAVELEN	54
#define		SXH_MAXXY  	55
#define		SXH_AXIS   	56
#define		SXH_ENDING 	57
#define		SXH_DETPAR 	58
#define		SXH_LUT		59
#define		SXH_DISPLIM	60
#define		SXH_PROGRAM	61
#define		SXH_ROTATE 	62
#define		SXH_BITMASK	63
#define		SXH_OCTMASK	64
#define		SXH_ESDCELL	65
#define		SXH_DETTYPE	66
#define		SXH_NEXP   	67
#define		SXH_CCDPARM	68
#define		SXH_CHEM	69
#define		SXH_MORPH  	70
#define		SXH_CCOLOR 	71
#define		SXH_CSIZE  	72
#define		SXH_DNSMET 	73
#define		SXH_DARK	74
#define		SXH_AUTORNG	75
#define		SXH_ZEROADJ	76
#define		SXH_XTRANS	77
#define		SXH_HKL_XY	78
#define		SXH_AXES2	79
#define		SXH_ENDING2	80
/* values of the `saxtype' variable */
#define		SXT_MWIRE	0
#define		SXT_CCD_PXL	1
#define		SXT_CCD_PXL_2K	2
#define		SXT_CCD_PXL_ARR	3
#define		SXT_CCD_FOREIGN	4

static int itodetype(int indx, char **detypedp, char **goniodp)
{ /*	Returns the number of microns across a particular type of detector.
	It stuffs the detector type in *detypedp and the most likely
	goniostat type in goniodp. */

	static int	micronstotal[] = {
		102400, 61440, 108544, 163840, 162202 };
	static char	*detype[] = {
	 "MWIRE","CCD-PXL","CCD-PXL-2K","CCD-PXL-ARR", "CCD-FOREIGN", NULL } ;
	static char	*goniotype[] = {
	 "3-AXIS","PLATFORM","PLATFORM","PLATFORM-KAPPA", "SPINDLE", NULL } ;

	*detypedp = *goniodp = NULL;	if ((0 > indx) || (indx > 4)) return 0;
	*detypedp = detype[indx];	*goniodp = goniotype[indx];
	return micronstotal[indx];
}

static int itoinstance(int indx, char **carddp, char **defdp)
{ /*	Returns the code value for one of the code-word instance values.
	It returns the corresponding card-verb in *carddp */
	static int	instance[] = {
	 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,	/*  0 -  9 */
	 1, 8, 1, 1,-1, 1, 1, 1, 1, 1,	/* 10 - 19 */
	 1,-1, 1,-1,-1,-1,-1, 1, 1, 1,	/* 20 - 29 */
	 1, 1, 1, 1, 1,-1,-1, 1, 1, 1,	/* 30 - 39 */
	 1,-2, 2, 1,-1, 1, 1,-1,-1,-1,	/* 40 - 49 */
	-1,-1, 1, 1, 1, 1, 1,-1,-2, 1,	/* 50 - 59 */
	 1, 1,-1, 1,-2,-2, 1,-1,-1, 1,	/* 60 - 69 */
	 1, 1, 1, 1, 1,-1,-1,-1, 1, 1,	/* 70 - 79 */
	 1, 0 };	/* 80 and stop-codon */
	static char	*card[] = {
	"FORMAT ", "VERSION", "HDRBLKS", "TYPE   ", "SITE   ",	/* 0 -  4 */
	"MODEL  ", "USER   ", "SAMPLE ", "SETNAME", "RUN    ",	/* 5 -  9 */
	"SAMPNUM", "TITLE  ", "NCOUNTS", "NOVERFL", "MINIMUM",	/*10 - 14 */
	"MAXIMUM", "NONTIME", "NLATE  ", "FILENAM", "CREATED",	/*15 - 20 */
	"CUMULAT", "ELAPSDR", "ELAPSDA", "OSCILLA", "NSTEPS ",	/*20 - 24 */
	"RANGE  ", "START  ", "INCREME", "NUMBER ", "NFRAMES",	/*25 - 29 */
	"ANGLES ", "NOVER64", "NPIXELB", "NROWS  ", "NCOLS  ",	/*30 - 34 */
	"WORDORD", "LONGORD", "TARGET ", "SOURCEK", "SOURCEM",	/*35 - 39 */
	"FILTER ", "CELL   ", "MATRIX ", "LOWTEMP", "ZOOM   ",	/*40 - 44 */
	"CENTER ", "DISTANC", "TRAILER", "COMPRES", "LINEAR ",	/*45 - 49 */
	"PHD    ", "PREAMP ", "CORRECT", "WARPFIL", "WAVELEN",	/*50 - 54 */
	"MAXXY  ", "AXIS   ", "ENDING ", "DETPAR ", "LUT    ",	/*55 - 59 */
	"DISPLIM", "PROGRAM", "ROTATE ", "BITMASK", "OCTMASK",	/*60 - 64 */
	"ESDCELL", "DETTYPE", "NEXP   ", "CCDPARM", "CHEM   ",	/*65 - 69 */
	"MORPH  ", "CCOLOR ", "CSIZE  ", "DNSMET ", "DARK   ",	/*70 - 74 */
	"AUTORNG", "ZEROADJ", "XTRANS ", "HKL&XY ", "AXES2  ",	/*75 - 79 */
	"ENDING2" } ;			/* 80 */
	static char	*defca[] = {
	"86", "11", "15", "Scan frame", "Unknown",		/* 0 -  4 */
	"PLATFORM", "local user", "saxiim", "saxiim0", "1",	/* 5 -  9 */
	"0", " image", "10000000", "0", "0",			/*10 - 14 */
	"30000", "10000000", "0", "saxiim", "06/01/97",		/*15 - 19 */
	"10.0", "100.0", "100.0", "0", "1",			/*20 - 24 */
	"0.1", "0.0", "0.1", "1", "400",			/*25 - 29 */
	"0.0", "0", "2", "1024", "1024",			/*30 - 34 */
	"0", "0", "Cu", "40", "30",				/*35 - 39 */
	"Ni", "50.0", "50.0", "0", "0.5",			/*40 - 44 */
	"511 511", "10.0", "-1", "NONE", "1.0",			/*45 - 49 */
	"0.0", "1.0", "test", "test", "1.5418",			/*50 - 54 */
	"511 511", "2", "0.0", "0.429", "$DEFAULT",		/*55 - 59 */
	"0.0", "X-GEN", "0", "$NULL", "0",			/*60 - 64 */
	"0.1", "CCD", "2", "10.0", "?",				/*65 - 69 */
	"?", "?", "?|?|?|?|?", "?", "$NULL",			/*70 - 74 */
	"1.0", "0.0", "0.0", " 0. 0. 0. 0. 0. "," 0. 0. 0. 0.", /*75 - 79 */
	" 0. 0. 0. 0. "} ;					/* 80 */

	*carddp = *defdp = NULL;
	if ((0 > indx) || (indx > 80)) return 0;
	*carddp = card[indx];	*defdp = defca[indx];	return instance[indx];
}

static int ascii_flow(MAIL *mp, int count)
{ /*	This processes an overflow from an ascii input.  It reads 512 bytes
	if the `count' variable is divisible by 32, since there are
	exactly 32 overflows per 512-byte block.  Simple, no?
	Provide for 16-bit overflows by keeping the `value'
	as (invalue - 65535) << 2, as with Mar data. */
	int		invalue, offset;
	size_t		nitems;
	char		ofbuf[18];
	u_char		*loz;
	static u_char	*ofcp;
	VOFLOW		*vop;

	if (NULL == (loz = (u_char *)mallop(512, "loz in ascii_flow")))
		return FRAME_NO_MEMORY;
	if (!(count % 32))
	  {
		if ((512 != (nitems = imread((AVOID *)loz, 512))))
		  {
			fprintf(stderr, "ascii_flow read %d bytes, not 512\n",
				(int)nitems);
			return FRAME_BADFLOWS;
		  }
		ofcp = loz;
	  }
	else	ofcp += 16;
	/* force a blank between the value and the offset */
	cpymem(&(ofbuf[0]), ofcp, 9);	ofbuf[9] = ' ';
	cpymem(&(ofbuf[10]), ofcp + 9, 7);	ofbuf[17] = '\0';
	if (2 > (nitems = sscanf((char *)ofbuf, "%d %d", &invalue, &offset)))
	  {
		fprintf(stderr, "only %d items obtained in ascii_flow\n",
			(int)nitems); return FRAME_BADFLOWS;
	  }
	vop = mp->overflow + count;
	if (mp->improps & IPR_IS8BIT) vop->value = (u_short)invalue;
	else vop->value = (u_short)(int)((invalue - 65535) / 4);
	if (mp->improps & IPR_MIRRORX) vop->xcoord = offset % mp->ydimen;
	else vop->xcoord = mp->xdimen - 1 - offset % mp->ydimen;
	if (mp->improps & IPR_MIRRORY) vop->ycoord = offset / mp->ydimen;
	else vop->ycoord = mp->ydimen - 1 - offset / mp->ydimen;
	return FRAME_OK;
}

int saxiflows(MAIL *mp, int imfd)
{ /*	This handles overflows from a Siemens image */
	int			nflowerr, nnow, i, noutthere;
	off_t			seekpoint, overoff, newpos;
	size_t			ngot, ntoread;
	DATARUN			*drup;
	VOFLOW			*vofp;
	VOFL16			*vo16, *vo16p;
	u_short			*bofb, *bofp;
	xgulong			*hofp;
	static xgulong		*hofb;

	/* jump to the place where the overflows are */
	drup = mp->runheader;
	if (drup->run_wtsum > 99.)
	  {
		overoff = (off_t)(mp->xdimen * mp->ydimen) +
			drup->run_nexam * 512L + drup->run_ninteg;
		/* read the 16-bit overflows */
		errno = 0;
		if (-1 == (newpos = lseek(imfd, overoff, 0)))
		  {
			fprintf(stderr,
			 "Error %d seeking to beginning of overflows\n",errno);
			return FRAME_BADFLOWS;
		  }
		if (overoff != newpos)
		  {
			fprintf(stderr,
 "Weird. Attempted to seek to %d, got to %d\n", (int)overoff, (int)newpos);
			return FRAME_BADFLOWS;
		  }
		noutthere = 8 * ((mp->oflowcount + 7) / 8);
		ntoread = noutthere * sizeof (u_short);
		if (NULL == (bofb = (u_short *)callop(noutthere,
		 sizeof (short), "bofb in saxiflows")))
			return FRAME_NO_MEMORY;
		if (ntoread != (ngot = imread((AVOID *)bofb, ntoread)))
		  {
			fprintf(stderr,
 "Tried to read %6d bytes of overflows beginning at %8d, got %6d\n",
				(int)ntoread, (int)newpos, (int)ngot);
			return FRAME_BADFLOWS;
		  }
		vofp = mp->overflow;
#ifndef LITTLEND
		for (i = 0, bofp = bofb; i < (int)(mp->oflowcount);i++, bofp++)
			*bofp = uswab(*bofp); /* byteswap if needed */
#endif
		for (i = 0, bofp = bofb; i < (int)(mp->oflowcount);
		 i++,bofp++,vofp++)
		   {
			vofp->value = *bofp; vofp->xcoord = vofp->ycoord = 0;
			if (255 > vofp->value) printf(
 "Warning: overflow %4d has a value of %3d: non sequitur\n", i, vofp->value);
		   }
		/* we don't need that bofb stuff any more. Humph. */
		memzap((AVOID **)&bofb, "bofb in saxiflows");
		/* now we pick up the 16-bit overflows */
		if (mp->count1)
		  {
			noutthere = 4 * ((mp->count1 + 3) / 4);
			if (NULL == (hofb = (xgulong *)callop(noutthere,
			 sizeof (xgulong), "hofb in saxiflows")))
				return FRAME_NO_MEMORY;
			ntoread = noutthere * sizeof (xgulong);
			if (ntoread != (ngot = imread((AVOID *)hofb, ntoread)))
			  {
				fprintf(stderr,
 "Tried to read %4d bytes of 32-bit overflows, got %4d\n",
					(int)ntoread, (int)ngot);
				return FRAME_BADFLOWS;
			  }
#ifndef LITTLEND
			for (i = 0, hofp = hofb; i < noutthere; i++, hofp++)
				*hofp = swawb(*hofp);
#endif
			if (NULL == (vo16 = (VOFL16 *)callop(mp->count1,
			 sizeof (VOFL16), "vo16 in saxiflows")))
				return FRAME_NO_MEMORY;
			for (vo16p = vo16, i = 0, hofp = hofb;
			 i < mp->count1; i++, vo16p++, hofp++)
			   {
				vo16p->olvalue = *hofp;
				vo16p->olxcoord = vo16p->olycoord = 0;
			   }
			mp->oflow16 = vo16;	/* store it properly */
		  }
	  } /* end if "if (drp->run_wtsum > 99.)" clause */
	else { /* older Bruker/Siemens (FORMAT=86) formats */
		if (mp->improps & IPR_IS8BIT) seekpoint =
		 (off_t)(mp->xdimen * mp->ydimen) + drup->run_nexam * 512L;
		else seekpoint = (off_t)(mp->xdimen * mp->ydimen * 2) +
			drup->run_nexam * 512L;
		if (-1L == lseek(imfd, seekpoint, 0))
		  {
			fprintf(stderr,
	 "Cannot seek to beginning of Siemens-image ascii overflow table\n");
			return FRAME_BADFLOWS;
		  }
		/* we read 'em in 512 bytes (32 overflows) at a time */
		for (nnow = 0; nnow < (int)(mp->oflowcount); nnow++)
		   if (FRAME_OK != (nflowerr = ascii_flow(mp, nnow)))
		     {
			fprintf(stderr,
	 "Error %2d in reading ascii overflow number %3d out of %3d\n",
			 nflowerr, nnow, (int)(mp->oflowcount));
			return FRAME_BADFLOWS;
		     }
	  } /* end of "else" for "if (drp->run_wtsum > 99" clause */
	seekpoint = drup->run_nexam * 512L;
	if (-1L == lseek(imfd, seekpoint, 0))
	  {
		fprintf(stderr,
 "Error in seeking to beginning of ordinary %d-bit data\n",
		 mp->improps & IPR_IS8BIT ? 8 : 16);
		return FRAME_BADFLOWS;	/* rewind to 8bit */
	  }
	return FRAME_OK;
}

static void addnotes(char *dest, char *text, unsigned maxlength)
{ /* This adds the string contained in "text" to the string "dest"
	provided that it doesn't overrun the length of "dest".
	The routine strips out blanks and tabs.*/
	char		*cp, *acp, *aend;
	unsigned	inlength;

	/* count nonblank characters in `text' */
	for (inlength = 0, cp = text; '\0' != *cp; cp++)
		if (*cp != ' ' && *cp != '\t') inlength++;
	aend = dest + maxlength; /* don't append if it would go beyond here*/
		/* find the current end of the `dest' string */
	for (acp = dest; acp < aend; acp++) if ('\0' == *acp) break;
	if ((acp - dest) + inlength < maxlength)
	  {
		for (cp = text; '\0' != *cp && acp < aend; cp++)
			if (*cp != ' ' && *cp != '\t') *acp++ = *cp;
		if (acp < (aend - 2))
		  {
			*acp++ = ';'; *acp++ = ' ';
		  }
		*acp = '\0';
	  }
}

static xglong maybef(char *str)
{ /*	This returns a value sensitive to the contents themselves */
	char	*cp;
	xglong	resu;
	double	fcount;

	for (cp = str; '\0' != *cp; cp++) if ('.' == *cp) break;
	if ('.' != *cp) resu = atoi(str);
	else
	  {
		if ((fcount = atof(str)) > 2.147483646e9) resu = 2147483647;
		else resu = atoi(str);
	  }
	return resu;
}

static int somealpha(char *str)
{ /*	This returns 1 iff str contains some alphabetic or numeric characters.
	We limit the search to a range between str and str+80. */
	char	*cp, *cend;

	cend = str + 80;
	for (cp = str; (cp < cend) && ('\0' != *cp); cp++)
	   {
		if ((('a' <= *cp) && (*cp <= 'z')) ||
		 (('A' <= *cp) && (*cp <= 'Z')) ||
		 (('0' <= *cp) && (*cp <= '9'))) return 1;
	   }
	return 0;
}

int saxii_ascii(MAIL *mp, DATARUN *dp, char *inbu, int imfd)
{ /*	This decodes an existing SAXII header */
	int		hdrblocks, istype100, j, lenn, lineno;
	int		matcount, maxlines, nover, nover16, nunder;
	int		saxtype, saxtest, swtype, thisinst, thissize, timetype;
	time_t		tl0;
	size_t		titlen;
	double		xma, yma, lambda, maxtwoth, runwidth;
	double		rnoise, eperadu, eperphot, bias1, fullscale;
	char		*cln, *cp, *contents, *detty, *gtype;
	char		*thiscard, *thisdef;
	static size_t	linelen = 80;

	/* rewind to re-read the first few characters */
	if (-1L == lseek(imfd, (off_t)0, 0)) return FRAME_BADMAIL;
	/* read the ascii header one line at a time */
	cln = (char *)inbu;	maxlines = MAX_SAXII_LINES;
	saxtype = istype100 = 0;
	dp->run_aduperph = 1.;	lambda = 1.5418; /* default */
	hdrblocks = -5;	/* so we can flag bad headers */
	mp->count0 = mp->count1 = 0;
	for (lineno = matcount = 0; lineno < maxlines; lineno++)
	   {
		if (linelen != imread((AVOID *)cln, linelen))
			return FRAME_BADMAIL;
		/* back up to last nonblank character */
		for (cp = cln + linelen - 2; cp > cln; cp--)
			if (' ' != *cp) break;
		*(cp+1) = '\0';	/* null-terminate at that point */
		contents = cln;	/* so we have something in there */
		for (thisinst = j = 0; j <= 80; j++)
		   {
			thisinst = itoinstance(j, &thiscard, &thisdef);
			titlen = (thisinst > 0) ? strlen(thiscard) : 0;
			if ((thisinst > 0) && !strncmp(thiscard, cln, titlen))
			  {
				contents = cln + titlen + 1;
				break;
			  }
		     }
		if (thisinst > 0) switch (j)
		  {
			case	SXH_FORMAT:
				if (100 == atoi(contents)) istype100 = 1;
				else istype100 = 0;	break;
			case	SXH_HDRBLKS:
				hdrblocks =(xglong)atoi(contents);
				maxlines = hdrblocks * 32 / 5;	break;
			case	SXH_SAMPLE:
				strncpy(dp->run_code, contents,6);	break;
			case	SXH_RUN:
				dp->run_id = atoi(contents);
				if (7 > strlen(dp->run_code) +strlen(contents))
				    strcat(dp->run_code, contents);	break;
			case	SXH_NCOUNTS:
				mp->eventcount = maybef(contents);	break;
			case	SXH_NOVERFL: /* either real or integer */
				if (istype100) /* Bruker 100 format */
				  {
					j = sscanf(contents, "%d %d %d",
					 &nunder, &nover, &nover16);
					if (j) dp->run_ninteg = nunder;
					if (j > 1) mp->oflowcount = nover;
					if (j > 2) mp->count1 = nover16;
				  }
				else mp->oflowcount = atoi(contents);	break;
			case	SXH_MAXIMUM:
				mp->zmax = atoi(contents); break;
			case	SXH_NONTIME:
				mp->ontime = maybef(contents); break;
			case	SXH_NLATE:
				mp->late = atoi(contents); break;
			case	SXH_FILENAM:
				lenn = strlen(contents);
				if (lenn < 16) strcpy(mp->filename, contents);
				else strcpy(mp->filename, contents +lenn-15);
				mp->filename[15] = '\0';	break;
			case	SXH_CREATED:
				timetype = (contents[8] == ' ') ? 7 : 8;
				(void)timecvt(mp->datetime, contents,
					NULL, &tl0, timetype, 0);	break;
			case	SXH_CUMULAT: mp->elap100ths = (int)(100. *
			 (dp->run_expos = atof(contents)) + 0.49999);	break;
			case	SXH_INCREME:
				xgensteps(dp, atof(contents));	break;
			case	SXH_NUMBER:
				mp->frameno = atoi(contents); break;
			case	SXH_NFRAMES:
				dp->run_nframes = atoi(contents);	break;
			case	SXH_ANGLES:
			 (void)sscanf(contents, "%f %f %f %f",
			 &dp->run_d0ang[0], &dp->run_startpos[0],
			 &dp->run_startpos[2], &dp->run_startpos[1]);	break;
			case	SXH_NOVER64:
				mp->over64k = atoi(contents);	break;
			case	SXH_NPIXELB: /* 1- or 2-byte formats */
				swtype = swawb(mp->format);
				if (2 == (saxtest = atoi(contents)))
				  {
					if (mp->format == FMT_SAXII)
					 mp->format = FMT_SAXII_16;
					else if (swtype == FMT_SAXII)
					 mp->format = swawb(FMT_SAXII_16);
				  }
				else if (1 == saxtest)
				  {
					if (mp->format == FMT_SAXII_16)
					 mp->format = FMT_SAXII;
					else if (swtype == FMT_SAXII_16)
					 mp->format = swawb(FMT_SAXII);
				  }
				break;
			case	SXH_NROWS: mp->ydimen = atoi(contents);	break;
			case	SXH_NCOLS: mp->xdimen = atoi(contents);	break;
			case	SXH_SOURCEK:
				dp->run_volxray = atof(contents);	break;
			case	SXH_SOURCEM:
				dp->run_curxray = atof(contents);	break;
			case	SXH_MATRIX:
			 if (!matcount)
			   {
				(void)sscanf(contents, "%f %f %f %f %f",
				  &dp->run_recip[0][0], &dp->run_recip[0][1],
				  &dp->run_recip[0][2], &dp->run_recip[1][0],
				  &dp->run_recip[1][1]);
				matcount += 5;
			  }
			 else if (matcount == 5)
			  {
				(void)sscanf(contents, "%f %f %f %f",
				  &dp->run_recip[1][2], &dp->run_recip[2][0],
				  &dp->run_recip[2][1], &dp->run_recip[2][2]);
				matcount += 4;
			  }
			 break;
			case	SXH_LOWTEMP:	dp->run_temp = atoi(contents);
				if (!dp->run_temp) dp->run_temp = 298;	break;
			case	SXH_CENTER:	(void)sscanf(contents, "%f %f",
				 &dp->run_d0pos[1], &dp->run_d0pos[2]);	break;
			case	SXH_DISTANC:
				dp->run_d0pos[0] = atof(contents);	break;
			case	SXH_WAVELEN:
				dp->run_lambda = lambda = atof(contents); break;
			case	SXH_MAXXY:
			 (void)sscanf(contents, "%lf %lf", &xma, &yma);
			 mp->zmaxx = (int)(xma + 0.49999);
			 mp->zmaxy = (int)(yma + 0.49999);	break;
			case	SXH_AXIS:
				dp->run_motor = atoi(contents) - 1; break;
			case	SXH_DISPLIM: /* display limit */
				mp->zdisp = (int)(atof(contents));	break;
			case	SXH_DETTYPE: /* type of detector */
			 for (saxtest = 0; saxtest < 5; saxtest++)
			    {
				(void)itodetype(saxtest, &detty, &gtype);
				if (!strncmp(contents, detty, MAXTYPELEN))
				  {
					saxtype = saxtest;	break;
				  }
			    }
			 break;
			case	SXH_CCDPARM:	/* get aduperphot from here */
				(void)sscanf(contents, " %lf %lf %lf %lf %lf",
					&rnoise, &eperadu, &eperphot,
					&bias1, &fullscale);
				dp->run_eperadu = eperadu;
				dp->run_eperph = eperphot;
				if ((0.1 < eperadu) && (0.1 < eperphot))
					dp->run_aduperph = eperphot / eperadu;
				else	dp->run_aduperph = 3.;
			default  : /* add info to the notes if there's room */
			 if (('\0' != *contents) && strcmp(contents,"UNKNOWN")
			   && somealpha(contents))
			  addnotes(dp->run_notes,cln,
			  (unsigned)(sizeof (dp->run_notes)-2));
			  break;
		  }
	   }
	if (hdrblocks < 0) return FRAME_BADMAIL;	/* require legal value*/
	if (istype100) dp->run_wtsum = 100.;
	if ((saxtype > 0) && (dp->run_aduperph < 0.2)) dp->run_aduperph = 3;
	if (dp->run_ninteg) dp->run_ninteg = 16 * ((dp->run_ninteg + 15) / 16);
	dp->run_width = mp->xdimen;	dp->run_height = mp->ydimen;
	/* nominal detector pixel size */
	thissize = itodetype(saxtype, &detty, &gtype);
	if ((0 <= saxtype) && (saxtype <= 4) && (mp->xdimen > 1))
		mp->cellsize = thissize / mp->xdimen;
	else	mp->cellsize = 200;
	dp->run_d1pos[1] = dp->run_d1pos[2] = ((double)mp->cellsize) * 0.001;
	/* remember the # of header blocks in a safe place */
	dp->run_nexam = hdrblocks;
	/* advance the file pointer to the place the header says to */
	if (-1L == lseek(imfd, (off_t)(512 * hdrblocks), 0))
		return FRAME_BADMAIL;
	mp->zmaxx = mp->xdimen - 1 - mp->zmaxx;	/* diff. origin convention  */
	mp->zmaxy = mp->ydimen - 1 - mp->zmaxy;
	if (dp->run_d0pos[0] > FUZZ)
	  {
		maxtwoth = fabs(dp->run_d0ang[0] * RADC) + atan(((double)
		 (dp->run_width)) * dp->run_d1pos[1] * 0.05 /dp->run_d0pos[0]);
		if (maxtwoth < 0.01) maxtwoth = 0.01;
		maxtwoth = sin(maxtwoth / 2.);
	  }
	else	maxtwoth = 0.2;
	dp->run_reslim = 0.5 * lambda / maxtwoth;
	dp->run_ssqmax = maxtwoth * maxtwoth;
	if (DFUZZ < (runwidth = dp->run_nframes * fabs(dp->run_stepreal)))
		dp->run_nshifts = 2 * (int)(runwidth / 5. + 0.49999);
	mp->nonzpix = mp->quadrant = 0;
	mp->improps |= IPR_XFASTEST | IPR_STARTBOTTOM;
	return	FRAME_OK;
}

static int saxi_entry(int entryno, int lineno, char *obu, char *valu)
{ /*	This creates a line of output for a SAXII header.
	The format looks like "key: value", blank-padded to 80 bytes.
	The routine returns FRAME_OK if successful, FRAME_BADMAIL if not */
	char		*base, *thiscard, *thisdef;
	int		len, iline, thisinst;
	static int	saxilen = 80;
	size_t		lvalu;

	/* count the offset to this keyword */
	for (base = (char *)obu, iline = 0; iline < entryno; iline++)
	   {
		thisinst = itoinstance(iline, &thiscard, &thisdef);
		base += saxilen * abs(thisinst);
	   }
	/* for multi-line keywords, move up to the right line */
	for (iline = 0; iline < lineno; iline++) base += saxilen;
	thisinst = itoinstance(entryno, &thiscard, &thisdef);
	cpymem((AVOID *)base, (AVOID *)thiscard, 7);
	base[7] = ':';
	lvalu = strlen(valu);	if (lvalu > 71) lvalu = 71;
	cpymem((AVOID *)(base + 8), valu, lvalu);
	len = 8 + strlen(valu);
	for ( ; len < saxilen; len++) *(base + len) = ' ';
	return FRAME_OK;
}

static char *fromname(char *str, int tochop)
{ /*	This eliminates `tochop' decimal or hex digits from the
	end of the name str and returns the result */
	static char 	fbu[24];
	static char	*none = "none";
	char		*cp;
	int		len, j, chopped;

	if (tochop >= (len = strlen(str))) return none;
	if (len >= 22) len = 22;
	for (cp = str, j = 0; ('\0' != *cp) && (j < len);
	 j++, cp++) fbu[j] = *cp;
	fbu[j] = '\0';
	for (cp = &(fbu[len]), chopped = 0; cp > fbu; cp--)
	   if ((('0' <= *cp) && (*cp <= '9')) || (('a' <= *cp) && (*cp<= 'f'))
		|| (('A' <= *cp) && (*cp <= 'F')))
	     {
		*cp = '\0';	if (++chopped >= tochop) return fbu;
	     }
	return fbu;
}

static int allinteger(char *str)
{ /*	This returns 1 iff `str' is entirely composed of 0-9's
	and an optional leading minus */
	char	*cp;

	for (cp = str; '\0' != *cp; cp++)
		if (((*cp < '0') || (*cp > '9')) &&
			((cp > str) || ('-' != *cp))) return 0;
	return 1;
}

static void fromnotes(char *obu, char *text)
{ /*	This fills in a few entries in a SAXII header based on the notes */
	char		*ent, *icp, *ocp, *thiscard, *thisdef;
	int		i, j, thisinst, tocolon, tosearch;
	static int	notelen = 53;
	static char	lobuf[80], flbu[56];

	/* start at the beginning of the notes */
	for (ent = text, i = 0; ('\0' != *ent) && (i < notelen); )
	   {
		flbu[0] = '\0';
		/* see if there's a match between these notes and the
		name of one of the cards. If there is, use the value.
		By convention the entries are semicolon-delimited in
		the notes, and the entry name is colon-delimited */
		tocolon = 0;
		for (icp = ent; '\0' != *icp && ':' != *icp; icp++)
			tocolon++;
		/* all of the legitimate keywords are at least
		 three characters long, so we skip the search through
		 the keywords unless "tocolon" is at least three */
		if (tocolon < 3) j = -1;
		/* ... furthermore if we've run out of colons,
		 we've run out of possible entries. */
		else if ('\0' == *icp) break;
		else for (j = 0; j <= 80; j++)
		   {
			thisinst = itoinstance(j, &thiscard, &thisdef);
			tosearch = strlen(thiscard);
			if (tocolon < tosearch) tosearch = tocolon;
			if (!strncmp(ent, thiscard, tosearch))
			  {
				icp = ent + tocolon + 1;
				for (ocp = &(flbu[0]);('\0' != *icp) &&
				 (';'!= *icp); icp++,ocp++) *ocp = *icp;
				*ocp = '\0';	ent = icp;
				i = ent - text;	break;
			  }
		   }
		if ((j >= 0) && (0 != (thisinst = itoinstance(j,
		 &thiscard, &thisdef))) && (flbu[0] != '\0'))
	   	  {
			if (allinteger(flbu))
			  {
				sprintf(lobuf, "%10d", atoi(flbu));
				saxi_entry(j, 0, obu, lobuf);
			  }
			else if (allnumeric(flbu))
			  {
				sprintf(lobuf, "%14.7f", atof(flbu));
				saxi_entry(j, 0, obu, lobuf);
			  }
			else saxi_entry(j, 0, obu, flbu);
		  }
		ent++;	i++;
	   }
}

static char latedigit(char *str, int digitno)
{ /*	This returns the nth-to-the-last digit in the string `str', where n
  	is `digitno'.  If there are no digits in the string, it returns "0"*/
	int	digno;
	char	digc, *cp;

	digno = 0;	digc = '\0';
	for (cp = str + strlen(str) - 1; cp >= str; cp--)
	   if (('0' <= *cp) && (*cp <= '9'))
	     {
		if (++digno >= digitno)
		  {
			digc = *cp;	break;
		  }
	     }
	if (digc == '\0') digc = '0';
	return digc;
}

int mksaxihead(MAIL *mp, int otype, int imfd)
{ /* Create a proper SAXI imaage header */
	int		ainst, ninst, detsize, cardno, i, j, i1, i2;
	int		saximotor, thisinst, thissize;
	time_t		tl0;
	double		ava, bias1, dot, expos, fullscale;
	double		ncounts, rnoise, te, can[3], cle[3];
	xglong		maxcount[3];
	char		*cp, *detty, *gtype, *thiscard, *thisdef, lb2[2];
	DATARUN		*drp;
	static char	*ba, *obu, lobuf[80];

	/* this creates a valid saxii header and then outputs it*/
	memzap((AVOID **)&obu, "obu in mksaxihead");
	if (NULL == (obu = (char *)mallop(7680, "obu in mksaxihead")))
		return FRAME_BADMAIL;
	/* first, load the defaults */
	for (cardno = 0; cardno <= 80; cardno++)
	   {
		thisinst = itoinstance(cardno, &thiscard, &thisdef);
		if (allinteger(thisdef))
		  {
			sprintf(lobuf, "%10d", atoi(thisdef));
			saxi_entry(cardno, 0, obu, lobuf);
		  }
		else if (allnumeric(thisdef))
		  {
			sprintf(lobuf, "%14.7f", atof(thisdef));
			saxi_entry(cardno, 0, obu, lobuf);
		  }
		else saxi_entry(cardno, 0, obu, thisdef);
			/* multi-line values handled */
		if ((ainst = abs(thisinst)) > 1)
		  {
			for (ninst = 1; ninst < ainst; ninst++)
				saxi_entry(cardno, ninst, obu, " ");
		  }
	   }
	/* now, some entries particular to the current project */
	if (NULL != (ba = getenv("HOST"))) saxi_entry(SXH_SITE, 0, obu, ba);
	if (NULL != (ba = getenv("USER"))) saxi_entry(SXH_USER, 0, obu, ba);
	ba = filepart(mp->filename, 3);
	strcpy(lobuf, ba);
	saxi_entry(SXH_SAMPLE, 0, obu, fromname(lobuf, 2));
	saxi_entry(SXH_SETNAME, 0, obu, fromname(lobuf, 1));
	saxi_entry(SXH_TITLE, 0, obu, fromname(lobuf, 1));
	lb2[1] = '\0';	lb2[0] = latedigit(lobuf, 1);
	saxi_entry(SXH_RUN, 0, obu, lb2);
	lb2[1] = '\0';	lb2[0] = latedigit(lobuf, 2);
	saxi_entry(SXH_SAMPNUM, 0, obu, lb2);
	strcpy(lobuf, "?");	saxi_entry(SXH_TITLE, 1, obu, lobuf);
	sprintf(lobuf, " %13.5f %13.5f", 1., 0.);
	saxi_entry(SXH_LINEAR, 0, obu, lobuf);
	sprintf(lobuf, " %13.5f %13.5f %13.5f %13.5f %13.5f",
		0., 0., 0., 0., 0.);
	saxi_entry(SXH_PHD, 0, obu, lobuf);
	saxi_entry(SXH_DETPAR, 0, obu, lobuf);
	sprintf(lobuf, " %13.5f", 0.);
	saxi_entry(SXH_DETPAR, 1, obu, lobuf);
	/* some entries grabbed from the run notes */
	if (NULL != (drp = mp->runheader)) fromnotes(obu, drp->run_notes);
	getcounts(mp, &ncounts, maxcount);
	if (ncounts < 2.147e9) sprintf(lobuf,
	 " %11.0lf %11d %11d %11d %11d %11d", ncounts, 145056, 0, 0, 0, 0);
	else sprintf(lobuf, "%12.4e %11d %11d %11d %11d %11d",
		ncounts, 0, 0, 0, 0, 0);
	saxi_entry(SXH_NCOUNTS, 0, obu, lobuf);
	/* overflows are subtle. If we are writing out 16-bit data,
	 having read in 8-bit data, we don't have overflows any more */
	if (((otype == FMT_SAXII_16) || (otype == swawb(FMT_SAXII_16))) &&
	 (NULL != mp->data8)) sprintf(lobuf, " %9d", 0);
	else	sprintf(lobuf, " %9d", (int)(mp->oflowcount));
	saxi_entry(SXH_NOVERFL, 0, obu, lobuf);
	sprintf(lobuf, " %9d", (int)(maxcount[0]));
	saxi_entry(SXH_MAXIMUM, 0, obu, lobuf);
	if (ncounts < 2.147e9) sprintf(lobuf, " %11.0f", ncounts);
	else sprintf(lobuf, "%12.4e", ncounts);
	saxi_entry(SXH_NONTIME, 0, obu, lobuf);
	saxi_entry(SXH_FILENAM, 0, obu, mp->filename);
	(void)timecvt(mp->datetime, NULL, lobuf, &tl0, 0, 4);
	saxi_entry(SXH_CREATED, 0, obu, lobuf);
	expos = 0.01 * ((double)mp->elap100ths);
	sprintf(lobuf, " %13.5f", expos);
	saxi_entry(SXH_CUMULAT, 1, obu, lobuf);
	saxi_entry(SXH_ELAPSDR, 1, obu, lobuf);
	saxi_entry(SXH_ELAPSDA, 1, obu, lobuf);
	sprintf(lobuf, "%10d", (int)(mp->frameno));
	saxi_entry(SXH_NUMBER, 0, obu, lobuf);
	sprintf(lobuf, " %9d",
	 ((mp->format == FMT_SAXII) || (mp->format == swawb(FMT_SAXII))) ?
		1 : 2);
	saxi_entry(SXH_NPIXELB, 0, obu, lobuf);
	sprintf(lobuf, "%10d", (int)(mp->ydimen));
	saxi_entry(SXH_NROWS, 0, obu, lobuf);
	sprintf(lobuf, "%10d", (int)(mp->xdimen));
	saxi_entry(SXH_NCOLS, 0, obu, lobuf);
	sprintf(lobuf, " %13.7f %13.7f",
		(double)maxcount[1], (double)maxcount[2]);
	saxi_entry(SXH_MAXXY, 0, obu, lobuf);
	if (mp->cellsize <= 0) mp->cellsize = 200;
	/* yuck: */
	detsize = mp->xdimen * mp->cellsize;
	for (ninst = 0; ninst <= 4; ninst++)
	   {
		thissize = itodetype(ninst, &detty, &gtype);
		if ((thissize > 0) && (abs(detsize - thissize) < 10))
		  {
			saxi_entry(SXH_DETTYPE, 0, obu, detty);
			saxi_entry(SXH_MODEL, 0, obu, gtype);	break;
		  }
	  }
	if (ninst >= 5)
	  {
		saxi_entry(SXH_DETTYPE, 0, obu, "unknown");
		saxi_entry(SXH_MODEL, 0, obu, "unknown");
	  }
	sprintf(lobuf, " %13.7f %13.7f %13.7f %13.7f %13.7f",
	 0., 0., 0., 0., 0.);
	saxi_entry(SXH_HKL_XY, 0, obu, lobuf);
	if (NULL != drp)
	  {
		saxi_entry(SXH_SAMPLE, 0, obu, drp->run_code);
		ba = filepart(drp->run_code, 3);
		strcpy(lobuf, ba);
		saxi_entry(SXH_SETNAME, 0, obu, lobuf);
		sprintf(lobuf, " %9d", drp->run_id);
		saxi_entry(SXH_RUN, 0, obu, lobuf);
		sprintf(lobuf, "%14.7f", drp->run_expos);
		saxi_entry(SXH_CUMULAT, 0, obu, lobuf);
		saxi_entry(SXH_ELAPSDR, 0, obu, lobuf);
		saxi_entry(SXH_ELAPSDA, 0, obu, lobuf);
		sprintf(lobuf, "%14.7f", fabs(drp->run_stepreal));
		saxi_entry(SXH_RANGE, 0, obu, lobuf);
		if (FUZZ < fabs(drp->run_startpos[0])) sprintf(lobuf,
			"%14.7f", to360(drp->run_startpos[0]));
		else sprintf(lobuf, "%14.7f", to360(drp->run_startpos[1]));
		saxi_entry(SXH_START, 0, obu, lobuf);
		saximotor = drp->run_motor + 1;
		sprintf(lobuf, "%10d", saximotor);
		saxi_entry(SXH_AXIS, 0, obu, lobuf);
		if (saximotor <= 2)
			sprintf(lobuf, "%14.7f", drp->run_stepreal);
		else	sprintf(lobuf, "%14.7f", -(drp->run_stepreal));
		saxi_entry(SXH_INCREME, 0, obu, lobuf);
		sprintf(lobuf, "%10d", drp->run_nframes);
		saxi_entry(SXH_NFRAMES, 0, obu, lobuf);
		if (drp->run_volxray > 0)
		  {
			sprintf(lobuf, " %13.5f %13.5f %13.5f %13.5f %13.5f",
			 (double)(drp->run_volxray), 0., 0., 0., 0.);
			saxi_entry(SXH_SOURCEK, 0, obu, lobuf);
		  }
		if (drp->run_curxray > 0)
		  {
			sprintf(lobuf, " %13.5f %13.5f",
				(double)(drp->run_curxray), 0.);
			saxi_entry(SXH_SOURCEM, 0, obu, lobuf);
		  }
		sprintf(lobuf, "%14.7f%14.7f%14.7f%14.7f",
		 to360(drp->run_d0ang[0]), to360(drp->run_startpos[0]),
		 to360(drp->run_startpos[2]), to360(drp->run_startpos[1]));
		saxi_entry(SXH_ANGLES, 0, obu, lobuf);
		for (i = 0; i < 3; i++)
		   {
			i1 = i + 1;	if (i1 > 2) i1 = 0;
			i2 = 3 - i - i1;
			for (dot = ava = j = 0; j < 3; j++)
			   {
				te = drp->run_recip[i][j]; ava += te * te;
				dot += drp->run_recip[i1][j] *
					drp->run_recip[i2][j];
			   }
			if (ava > DFUZZ) ava = 1. / sqrt(ava);
			cle[i] = ava;	can[i] = dot;
		   }
		for (i = 0; i < 3; i++)
		   {
			if (cle[i] > 9999.) cle[i] = 9999.;
			else if (cle[i] < 1.) cle[i] = 1.;
		   }
		for (i = 0; i < 3; i++)
		   {
			i1 = i + 1; if (i1 > 2) i1 = 0;
			i2 = 3 - i - i1;
			can[i] /= (cle[i1] * cle[i2]);
			can[i] = acos((double)can[i]) / RADC;
		   }
		sprintf(lobuf, "%14.7f%14.7f%14.7f%14.7f%14.7f",
		 cle[0], cle[1], cle[2], can[0], can[1]);
		saxi_entry(SXH_CELL, 0, obu, lobuf);
		sprintf(lobuf, "%14.7f", can[2]);
		saxi_entry(SXH_CELL, 1, obu, lobuf);
		sprintf(lobuf, "%14.7f%14.7f%14.7f%14.7f%14.7f",
		 drp->run_recip[0][0], drp->run_recip[0][1],
		 drp->run_recip[0][2], drp->run_recip[1][0],
		 drp->run_recip[1][1]);
		saxi_entry(SXH_MATRIX, 0, obu, lobuf);
		sprintf(lobuf, "%14.7f%14.7f%14.7f%14.7f",
		 drp->run_recip[1][2], drp->run_recip[2][0],
		 drp->run_recip[2][1], drp->run_recip[2][2]);
		saxi_entry(SXH_MATRIX, 1, obu, lobuf);
		sprintf(lobuf, "%14.7f%14.7f",
		 drp->run_d0pos[1], drp->run_d0pos[2]);
		saxi_entry(SXH_CENTER, 0, obu, lobuf);
		if (drp->run_d0pos[0] < 1.) drp->run_d0pos[0] = 1.;
		sprintf(lobuf, "%14.7f", drp->run_d0pos[0]);
		saxi_entry(SXH_DISTANC, 0, obu, lobuf);
		sprintf(lobuf, "%14.7f%14.7f%14.7f%14.7f%14.7f",
		 drp->run_lambda, 0., 0., 0., 0.);
		saxi_entry(SXH_WAVELEN, 0, obu, lobuf);
		if (saximotor == 3)
		  sprintf(lobuf, " %13.7f %13.7f %13.7f %13.7f",
			 to360(drp->run_d0ang[0]), to360(drp->run_startpos[0]),
			 to360(drp->run_startpos[2] - drp->run_stepreal),
			 to360(drp->run_startpos[1]));
		else sprintf(lobuf, " %13.7f %13.7f %13.7f %13.7f", 
		 to360(drp->run_d0ang[0]),
		 to360(drp->run_startpos[0] - drp->run_stepreal),
		 to360(drp->run_startpos[2]), to360(drp->run_startpos[1]));
		saxi_entry(SXH_ENDING, 0, obu, lobuf);
		rnoise = 10.;	drp->run_eperadu = 2.5;
		bias1 = 118.;	fullscale = 65535.;
		if ((drp->run_aduperph > 0.1) && (drp->run_eperadu > 0.01))
		  drp->run_eperph = drp->run_aduperph * drp->run_eperadu;
		else drp->run_eperph = drp->run_eperadu;
		sprintf(lobuf, "%14.7f%14.7f%14.7f%14.7f%14.7f",
		 rnoise, drp->run_eperadu, drp->run_eperph, bias1, fullscale);
		saxi_entry(SXH_CCDPARM, 0, obu, lobuf);
		sprintf(lobuf, "%14.7f%14.7f%14.7f%14.7f",
		 0., 0., 0., 0.);
		saxi_entry(SXH_AXES2, 0, obu, lobuf);
		sprintf(lobuf, "%14.7f%14.7f%14.7f%14.7f",
		 0., 0., 0., 0.);
		saxi_entry(SXH_ENDING2, 0, obu, lobuf);
	  }
	/* output some end-of-header info (dots) */
	for (i1 = i = 0; i <= 80; i++)
	   {
		thisinst = itoinstance(cardno, &thiscard, &thisdef);
		i1 += abs(thisinst);	/* count instances */
	   }
	i1 *= 80;	/* scale up by # of bytes/row */
	for (i2 = i1, i = 0; i < 3; i++, i2 += 80)
	   {
		ba = obu + i2;	*ba = 032;	*(ba+1) = 04;
		for (cp = ba + 2; cp < ba + 80; cp++)
			*cp = '.';
	   }
	/* phew. Now we write the mess out */
	(void)lseek(imfd, (off_t)0L, 0);	/* rewind! */
	i = (7680 == imwrite((AVOID *)obu, 7680)) ?
		FRAME_OK : FRAME_BADMAIL;
	memzap((AVOID **)&obu, "obu in mksaxihead");
	return i;
}

static int out_aflow(MAIL *mp, VOFLOW *vop, int count, u_char *obuf)
{ /*	This processes an overflow into an ascii output.  It writes 512 bytes
	if the `count' variable is divisible by 32, since there are
	exactly 32 overflows per 512-byte block.  Simple, no? */
	int		nitems, outvalue, offset;
	char		ofbuf[18];
	static u_char	*ofcp;

	/* construct the ascii overflow */
	if (mp->data8) outvalue = vop->value;
	else outvalue = (4 * ((int)(vop->value))) + 65535;
	/* n.b. we need to deal with IPR_MIRRORX and IPR_MIRRORY later */
	offset = (mp->xdimen - 1 - vop->xcoord) +
	 mp->xdimen * (mp->ydimen - 1 - vop->ycoord);
	sprintf(ofbuf, "%9d%7d", outvalue, offset); ofbuf[17] = '\0';
	if (!count) ofcp = obuf;	else ofcp += 16;
	cpymem((AVOID *)ofcp, ofbuf, 16);
	if ((!((count + 1) % 32)) || (count == (int)(mp->oflowcount - 1)))
	  {
		if ((512 != (nitems = imwrite((AVOID *)obuf, 512))))
		  {
			fprintf(stderr, "out_aflow wrote %d bytes, not 512\n",
				nitems);
			return FRAME_BADFLOWS;
		  }
		ofcp = obuf - 16;
	  }
	return FRAME_OK;
}

int mksaxflows(MAIL *mp, int imfd, u_char *obuf)
{ /*	This writes out overflows in the SAXII format */
	int	nnow;
	u_char	*ucp;
	off_t	imagesi, seekpoint;
	VOFLOW	*vofp, *oend;
	
	for (ucp = obuf, nnow = 512; nnow--; ucp++) *ucp = ' '; /* blanks */
	ucp--;	*ucp = '\0'; /* but null-terminate it! */
	/* jump to the place where the overflows are */
	imagesi = mp->xdimen * mp->ydimen;
	if (mp->data8) seekpoint = imagesi + 7680;
	else seekpoint = 2 * imagesi + 7680;
	if (-1L == lseek(imfd, seekpoint, 0))
	  {
		fprintf(stderr, "mksaxflows: seek failure\n");
		return FRAME_BADFLOWS;
	  }
	if (!(mp->oflowcount))
	  {
		if (512 != (nnow = imwrite((AVOID *)obuf, 512)))
		  {
			fprintf(stderr,
			 "mksaxflows wrote %d bytes of blanks, not 512\n",
				nnow);
			return FRAME_BADFLOWS;
		  }
		return FRAME_OK;
	  }
	oend = mp->overflow + mp->oflowcount;
	/* we write 'em in 512 bytes (32 overflows) at a time */
	for (vofp = mp->overflow, nnow = 0; vofp < oend; vofp++, nnow++)
	   if (FRAME_OK != out_aflow(mp, vofp, nnow, obuf))
	     {
		fprintf(stderr, "mksaxflows: output error\n");
		return FRAME_BADFLOWS;
	     }
	return FRAME_OK;
}

/* end of saxiim.c */
