/* spatial.c -- spatially correct an image

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	12 Jun 2007: better support for noncanonical image directions
	 3 Jun 2007: support rescaling intensity to 65K before output
	 2 Jun 2007: don't call globalkill() before exiting;
		stredge() utility written for extrapolations;
		third argument to spatial renamed to `spflags'
	19 Apr 2007: float->double conversions; reorganize use of "{"
	 5 Dec 2005: correct bugs in fpw += statements
	20 Nov 2005: inline argument declarations and strict prototyping
	25 Oct 2005: eliminate static globals using SPAWORK structure elements
	14 May 2005: moved from utilities to xglib for greater accessibility
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	28 Aug 2004: routines used only here are declared as static; cosmetics
	26 Jan 2004: replace a few static globals with locals
	21 Jan 2004: license message inserted and initial comment reordered
	 8 Jan 2004: replace extern global variables with static globals;
		callop() & memzap() replace shcalloc() & shfree();
		eliminate routine elim_limits(), which wasn't being used
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	 5 Jan 2003: reorganized source-file header
	 1 Nov 2002: eliminate unused definition of MAX_PASSES
	25 Mar 2002: removed MSI attachments
	 6 Nov 1997: initial version

   Purpose: take a detector image already in memory and
	modify it such that the output is free of spatial distortions.
   Cross references: ../xglib/spline.c

 This takes in a detector image whose header is pointed to by `thishd'
 and modifies it such that the output is free of spatial distortion.
 It does this by interpolating the map with the help of the spatial
 distortion data pointed to by `spap->sp_spl'.
*************************************************************************/
#include	"xgen.h"
#define		MINRESERVED		800
#define		RESCALE65K	0x80	/* rescale < 65K during interpolation*/
#define		MODERATE_VERBOSE	0x3000
#define		VERY_VERBOSE		0x2000

 typedef struct {
		int		sp_imsize, sp_outht, sp_outwid, sp_sflags;
		double		sp_dimratx, sp_dimraty;
		double		sp_x00, sp_xsl, sp_y00, sp_ysl;
		float		*sp_locim, *sp_locwt;
		MAIL		*sp_lhd;
		VSPLINE		*sp_spl;
		} SPAWORK;

static int find_limits(SPAWORK *spap)
{ /*	This determines the limits of x and y in centimeters and populates
	the x and y slope and intercept values accordingly */
	int	x, y;
	float	xc, yc;
	double	xma, xmi, yma, ymi;

	yma = xma = -1.e9;	ymi = xmi = -xma;
	for (y = 0; y < (spap->sp_spl)->v_deth; y++)
		for (x = 0; x < (spap->sp_spl)->v_detw; x++)
		   {
			if (-1 != pix2cm((float)x, (float)y, &xc, &yc, 0))
			  {
				if (xc < xmi) xmi = xc;
				if (xc > xma) xma = xc;
				if (yc < ymi) ymi = yc;
				if (yc > yma) yma = yc;
			  }
		   }
	if ((yma < -0.9e9) || (xma < -0.9e9) ||
	 (xmi > 0.9e9) || (ymi > 0.9e9)) return EBADVALUE;
	if (xma < xmi + DFUZZ) return EBADVALUE;
	if (yma < ymi + DFUZZ) return EBADVALUE;
	spap->sp_xsl = ((double)(spap->sp_outwid - 1)) / (xma - xmi);
	spap->sp_x00 = -xmi * spap->sp_xsl;
	spap->sp_ysl = ((double)(spap->sp_outht - 1)) / (ymi - yma);
	spap->sp_y00 = -yma * spap->sp_ysl;
	return EOK;
}

static void interp_image(SPAWORK *spap)
{ /*	Here is where the actual interpolation takes place. */
	int	imv, ix, iy, off0, off1, off2, off3, x, y;
	double	px, py, rimv, wt0, wt1, wt2, wt3;
	float	xc, yc, *loim, *lowt;
	MAIL	*lhd;

	lhd = spap->sp_lhd;	loim = spap->sp_locim;	lowt = spap->sp_locwt;
	for (x = 0; x < (spap->sp_spl)->v_detw; x++) /* convert pix to cm */
	  for (y = 0; y < (spap->sp_spl)->v_deth; y++)
	   if (-1 != pix2cm((float)x, (float)y, &xc, &yc, 0))
	     {	/* rescale coordinates into the proper dimensions */
		xc = spap->sp_xsl * xc + spap->sp_x00;
		yc = spap->sp_ysl * yc + spap->sp_y00;
		/* add a fraction of the input value into the sum for each
		 of the four corners of the box in which the pixel rests.
		 Be careful at the edges */
		if ((-1 <= xc) && (xc < spap->sp_outwid) &&
		 (-1 <= yc) && (yc < spap->sp_outht))
		  {
			ix = xc;	iy = yc;
			px = xc - (double)ix;	if (xc < 0) ix--;
			py = yc - (double)iy;	if (yc < 0) iy--;
			off0 = ix * spap->sp_outht + iy;
			off1 = off0 + 1;
			off2 = off0 + spap->sp_outht;	off3 = off2 + 1;
			wt0 = (1. - px) * (1. - py);
			wt1 = py * (1. - px);
			wt2 = px * (1. - py);
			wt3 = px * py;
			/* find the input intensity value */
			imv = imvalue(lhd, x, y);
			rimv = ((double)imv) /
				(spap->sp_dimratx * spap->sp_dimraty);
			if ((ix >= 0) && (iy >= 0))
			  {
				*(loim + off0) += wt0 * rimv;
				*(lowt + off0) += wt0;
			  }
			if ((ix >= 0) && (iy < spap->sp_outht - 1))
			  {
				*(loim + off1) += wt1 * rimv;
				*(lowt + off1) += wt1;
			  }
			if ((ix < spap->sp_outwid - 1) && (iy >= 0))
			  {
				*(loim + off2) += wt2 * rimv;
				*(lowt + off2) += wt2;
			  }
			if ((ix < spap->sp_outwid - 1) &&
			 (iy < spap->sp_outht - 1))
			  {
				*(loim + off3) += wt3 * rimv;
				*(lowt + off3) += wt3;
			  }
		  } /* end of "if ((-1 <= xc" clause */
	     } /* end of "if (-1 != pix2cm" clause and "for (ix" loop */
}

static int normalize(SPAWORK *spap)
{ /*	this renormalizes by dividing the weighted sums by the weights.
	The routine returns the number of non-found pixels */
	int	x, y, noreps, nearedge, onedge;
	float	*fpv, *fpw, *loim, *lowt;

	noreps = nearedge = onedge = 0;
	loim = spap->sp_locim;	lowt = spap->sp_locwt;
	for (fpv = loim, fpw = lowt, x = 0; x < spap->sp_outwid; x++)
	   for (y = 0; y < spap->sp_outht; y++, fpv++, fpw++)
	      {
		if (*fpw > FUZZ) *fpv /= *fpw;
		else
		  {
			noreps++;
			if ((x < 4) || (y < 4) || (x >= spap->sp_outwid - 4) ||
			 (y >= spap->sp_outht - 4))
			  {
				nearedge++;
				if ((x < 2) || (y < 2) ||
				 (x >= spap->sp_outwid - 2) ||
				 (y >= spap->sp_outht  - 2)) onedge++;
			  }
		     }
	      }
	if (spap->sp_sflags & VERY_VERBOSE) /* code for this printout */
	  {
		printf(
		 " %8d pixels out of %8d were missing interpolatable values\n",
			noreps, spap->sp_imsize);
		printf(" of these, %8d were within three pixels of the edge\n",
			nearedge);
		printf(" of these, %8d were within   one pixel  of the edge\n",
			onedge);
	  }
	return noreps;
}

static void stredge(float *fpv, float *fpw, int offset)
{ /* extrapolation */
	float	denom;	

	if ((denom = *(fpw + offset)) > FUZZ)
	  {
		*fpw += denom;	*fpv += denom * *(fpv + offset);
	  }
}

static int extrap_pix(SPAWORK *spap, int norep)
{ /*	This extrapolates to account for missing pixels.
	It returns the number that are still missing when it's done */
	int	corrected, hit2, ix, iy, nearedge, off0, onedge, wid2, y;
	float	*fpv, *fpw, *loim, *lowt, *spmin, *spmax;

	loim = spap->sp_locim;	lowt = spap->sp_locwt;
	hit2 = spap->sp_outht / 2;	wid2 = spap->sp_outwid / 2;
	spmin = spap->sp_locim;
	spmax = spap->sp_locim + spap->sp_outwid * spap->sp_outht;
	y = spap->sp_outht;
	for (ix = hit2, corrected = 0; (ix >= 0) && (ix < spap->sp_outwid); )
	   {
		for (iy = hit2; (iy >= 0) && (iy < spap->sp_outht); )
		   {
			off0 = ix * spap->sp_outht + iy;
			fpv = spmin + off0;
			fpw = spap->sp_locwt + off0;
			if (FUZZ > *fpw)
			  {
				*fpv = 0.;
				if (ix > 0)
				  {
					if (iy > 0) stredge(fpv, fpw, -y - 1);
					stredge(fpv, fpw, -y);
					if (iy < y-1) stredge(fpv, fpw, 1 - y);
				  }
				if (iy > 0) stredge(fpv, fpw, -1);
				if (iy < y - 1) stredge(fpv, fpw, 1);
				if (ix < spap->sp_outwid - 1)
				  {
					if (iy > 0) stredge(fpv, fpw, y - 1);
					stredge(fpv, fpw, y);
					if (iy < y-1) stredge(fpv, fpw, 1 + y);
				  }
				if (*fpw > FUZZ)
				  { /* these shouldn't get high weights */
					*fpv /= *fpw;	corrected++;
					*fpw = 0.1;
					
				  }
				else {
					*fpw = *fpv = 0.;
				     }
			  }
			if (corrected >= norep) break;
			iy = (iy >= hit2) ?
			 spap->sp_outht - 1 - iy : spap->sp_outht - iy;
		   }
		if (corrected >= norep) break;
		ix = (ix >= wid2) ?
			spap->sp_outwid - 1 - ix : spap->sp_outwid - ix;
	   }
	if (corrected >= norep)
	  {
		if (spap->sp_sflags & VERY_VERBOSE)
			printf("All points found on extrapolation pass\n");
		return 0;
	  }
	else {
		norep = nearedge = onedge = 0;
		for (fpv = spap->sp_locim, fpw = spap->sp_locwt, ix = 0;
		 ix < spap->sp_outwid; ix++)
		   for (iy = 0; iy < spap->sp_outht; iy++, fpv++, fpw++)
		      {
			if (*fpw <= FUZZ)
			  {
				norep++;
				if ((ix < 4) || (iy < 4) ||
				 (ix >= spap->sp_outwid - 4) ||
				 (iy >= spap->sp_outht - 4))
				  {
					nearedge++;
					if ((ix < 2) || (iy < 2) ||
					 (ix >= spap->sp_outwid - 2) ||
					 (iy >= spap->sp_outht  - 2)) onedge++;
				  }
			     }
		      }
		if (spap->sp_sflags & VERY_VERBOSE)
		  {
			printf("after extrapolating:\n");
			printf(
 " %8d pixels out of %8d were missing interpolatable values\n",
				norep, spap->sp_imsize);
			printf(
 " of these, %8d were within three pixels of the edge\n", nearedge);
			printf(
" of these, %8d were within   one pixel  of the edge\n", onedge);
		  }
		return norep;
	     }
}

static int creat_undistor(SPAWORK *spap)
{ /*	This puts the floating-point image back into integers */
	double		dimrat, ecount, thv;
	float		maxfl, *fpv, *loim;
	int		x, y, nov, room;
	xgulong		tct, zma;
	u_short		*ov, zmxx, zmxy;
	u_char		*oc;
	static char	memstr[] = "16-bit data in creat_undistor";
	MAIL		*lhd;
	VOFLOW		*vop, *vofe;
	DATARUN		*drp;

	zma = ecount = 0;	zmxx = zmxy = 0;	lhd = spap->sp_lhd;
	loim = spap->sp_locim;
	/* if the dimensions have changed, redimension */
	if (((int)(lhd->xdimen) != spap->sp_outwid) ||
	 ((int)(lhd->ydimen) != spap->sp_outht))
	  {
		if (NULL != lhd->data16)
		  {
			memstr[0] = '1'; memstr[1] = '6';
			memzap((AVOID **)&(lhd->data16), memstr);
			if (NULL == (lhd->data16 = (u_short *)callop(
			 spap->sp_imsize, sizeof (u_short), memstr)))
			  {
				fprintf(stderr,
 "Cannot reserve memory for new 16-bit data\n");
				return EALLOC;
			  }
		  }
		if (NULL != lhd->data8)
		  {
			memstr[0] = ' '; memstr[1] = '8';
			memzap((AVOID **)&(lhd->data8), memstr);
			if (NULL == (lhd->data8 = (u_char *)callop(
			 spap->sp_imsize, sizeof (u_char), memstr)))
			  {
				fprintf(stderr,
			 "Cannot reserve memory for new 8-bit data\n");
				return EALLOC;
			  }
		  }
		dimrat = 0.5 * (spap->sp_dimratx + spap->sp_dimraty);
		lhd->cellsize /= dimrat;
		if (NULL != (drp = lhd->runheader))
		  {
			drp->run_d1pos[1] /= spap->sp_dimratx;
			drp->run_d1pos[2] /= spap->sp_dimraty;
			drp->run_width = spap->sp_outwid;
			drp->run_height = spap->sp_outht;
		  }
		lhd->xdimen = spap->sp_outwid;
		lhd->ydimen = spap->sp_outht;
	  }
	if (lhd->data16) /* simple case: 16-bit data */
	  { /* first, count the 16-bit overflows */
		maxfl = 0.;
		for (nov = x = 0, fpv = spap->sp_locim; x < spap->sp_outwid; x++)
		   for (y = 0; y < spap->sp_outht; y++, fpv++)
			if (*fpv > 65535.)
			  {
				nov++;	if (*fpv > maxfl) maxfl = *fpv;
			  }
		if ((nov) && (spap->sp_sflags & RESCALE65K))
		  { /* rescale so brightest pixel is < 65K */
			fprintf(stdout,
			 " Brightest renormalized pixel has I = %12.5e\n",
				maxfl);
			maxfl /= 65000.;
			for (x = 0, fpv = loim; x < spap->sp_outwid; x++)
			 for (y = 0; y < spap->sp_outht; y++, fpv++)
				*fpv /= maxfl;
			nov = 0;
		  }
		if (!nov) /* no 16bit 'flows: easy! */	
		 {
			for (x = 0, fpv = loim, ov = lhd->data16;
			 x < spap->sp_outwid; x++)
			   {
				for (tct = y = 0; y < spap->sp_outht;
				 y++, fpv++, ov++)
				   {
					*ov = (u_short)(*fpv + 0.49999);
					tct += *ov;
					if (*ov > zma)
					  {
						zma = *ov;
						zmxx = x;	zmxy = y;
					  }
				   }
				ecount += (double)tct;
			   }
		 }
		else /* some 16-bit 'flows: harder */
		  {
			room = (lhd->overflow == NULL) ?
			 0 : lhd->oflowcount;
			if (nov > room)
			  {
				if (NULL != (lhd->overflow))
				  {
					memzap((AVOID **)&(lhd->overflow),
					 "16-bit overflows in creat_undistor");
				  }
				room = nov;
				if (room < MINRESERVED) room = MINRESERVED;
				if (NULL == (lhd->overflow = (VOFLOW *)
				 callop((size_t)nov, sizeof (VOFLOW),
				 "overflows in creat_undistor")))
				  {
					fprintf(stderr,
 "Cannot reserve %d bytes for 16-bit overflows in creat_undistor\n",
					 (int)(nov * sizeof (VOFLOW)));
					return EALLOC;
				  }
				vofe = lhd->overflow + room;
			  }
			vop = lhd->overflow;
			for (x = 0, fpv = loim, ov = lhd->data16;
			 x < spap->sp_outwid; x++)
			  {
				for (tct = y = 0; y < spap->sp_outht;
				 y++, fpv++, ov++)
				   {
					tct += (xgulong)(*fpv + 0.49999);
					if (*fpv > zma)
					  {
						zma = (xgulong)(*fpv + 0.49999);
						zmxx = x;	zmxy = y;
					  }
					if (*fpv < 65535.)
					  *ov = (u_short)(*fpv + 0.49999);
					else
					  {
						*ov = 65535.;
						vop->xcoord = x;
						vop->ycoord =
						 spap->sp_outht - 1 - y;
						thv = (*fpv - 65535.) / 4.;
						vop->value = (thv > 65535.) ?
							65535 : thv + 0.49999;
						vop++;
					  }
				   }
				ecount += (double)tct;
			  } /* end of "for (y" loop */
		  } /* end of "else" for "if (!nov)" clause */
	  } /* end of "if (lhd->data16" clause */
	else if (lhd->data8) /* busier case: 8-bit data */
	  { /* first, count the 8-bit overflows */
		for (nov = x = 0, fpv = loim; x < spap->sp_outwid; x++)
		   for (y = 0; y < spap->sp_outht; y++, fpv++)
			if (*fpv > 255.) nov++;
		if (!nov) /* no 8-bit'flows: easy! */	
		 {
			for (x = 0, fpv = loim, oc = lhd->data8;
			 x < spap->sp_outwid; x++)
			   {
				for (tct = y = 0; y < spap->sp_outht;
				 y++, fpv++, oc++)
				   {
					*oc = (u_char)(*fpv + 0.49999);
					tct += (xgulong)(*oc);
					if (((xgulong)(*oc)) > zma)
					  {
						zma = (xgulong)(*oc);
						zmxx = x;
						zmxy = spap->sp_outht - 1 - y;
					  }
				   }
				ecount += (double)tct;
			   }
		 }
		else /* some 8-bit 'flows: harder */
		  {
			room = (lhd->overflow == NULL) ?  0 : lhd->oflowcount;
			if (nov > room)
			  {
				if (NULL != (lhd->overflow))
				  {
					memzap((AVOID **)&(lhd->overflow),
					 "overflows in creat_undistor");
				  }
				if (room < MINRESERVED) room = MINRESERVED;
				if (NULL == (lhd->overflow = (VOFLOW *)
				 callop((size_t)nov, sizeof (VOFLOW),
				 "overflows in creat_undistor")))
				  {
					fprintf(stderr,
 "Cannot reserve %d bytes for 8-bit overflows in creat_undistor\n",
						(int)(nov * sizeof (VOFLOW)));
					return EALLOC;
				  }
				vofe = lhd->overflow + room;
			  }
			else vofe = lhd->overflow + nov;
			vop = lhd->overflow;
			for (x = 0, fpv = loim, oc = lhd->data8;
			 x < spap->sp_outwid; x++)
			  {
				for (tct = y = 0; y < spap->sp_outht;
				 y++, fpv++, oc++)
				   {
					tct += (xgulong)(*fpv + 0.49999);
					if (*fpv > zma)
					  {
						zma = (xgulong)(*fpv + 0.49999);
						zmxx = x;
						zmxy = spap->sp_outht - 1 - y;
					  }
					if (*fpv < 255.)
					  *oc = (u_char)(*fpv + 0.49999);
					else
					  {
						*oc = 255.;
						vop->xcoord = x;
						vop->ycoord =
						 spap->sp_outht - 1 - y;
						thv = *fpv;
						vop->value = (thv > 65535.) ?
							65535 : thv + 0.49999;
						vop++;
					  }
				   } /* end of "for (y" loop */
				ecount += (double)tct;
			  } /* end of "for (x" loop */
		  } /* end of "else" for "if (!nov)" clause */
	  } /* end of "else if (lhd->data8" clause */
	else return EBADVALUE;	/* need either data8 or data16 defined */
	lhd->zmax = zma;	lhd->zmaxx = zmxx;	lhd->zmaxy = zmxy;
	if (ecount > 2147483646.) lhd->eventcount = -(ecount / 64.);
	else lhd->eventcount = ecount;
	lhd->oflowcount = nov;
	if (lhd->data16) lhd->over64k = nov;
	printf("%s: %12.0f counts, %4d overflows; zmax = %7d at [%4d,%4d]\n",
	 lhd->filename, ecount, (int)(lhd->oflowcount), (int)(lhd->zmax),
	 (int)(lhd->zmaxx), (int)(lhd->ydimen - 1 - lhd->zmaxy));
	return EOK;
}

static int endspa(int retv, SPAWORK **spadp)
{ /* Eliminates memory for the vector *spadp and then returns its argument. */
	memzap((AVOID **)spadp, "spap in spatial");	return retv;
}

int spatial(int owidth, int oheight, int spflags, MAIL *thd)
{ /*	This converts an image to a form in which it's spatially corrected. */
	int		noreps, sret;
	static SPAWORK	*spap;

	if (NULL != spap) memzap((AVOID **)&spap, "spap in spatial");
	if (NULL == (spap = (SPAWORK *)callop(1, sizeof (SPAWORK),
	 "spap in spatial"))) return endspa(EALLOC, &spap);
	if ((NULL == spap->sp_spl) && (NULL == (spap->sp_spl =
	 gspline(5, &sret)))) return endspa(sret, &spap);
	if ((owidth <= 0) || (oheight <= 0))
	  {
		spap->sp_outwid = (spap->sp_spl)->v_detw;
		spap->sp_outht  = (spap->sp_spl)->v_deth;
	  }
	else {
		spap->sp_outwid = owidth;	spap->sp_outht = oheight;
	     }
	if (((spap->sp_spl)->v_detw < 5) || ((spap->sp_spl)->v_deth < 5) ||
	 (NULL == (spap->sp_spl)->v_cpt))
	  {
		fprintf(stderr,
		 "Cannot do spatial correction: spline is faulty\n");
		return endspa(EBAD, &spap);
	  }
	spap->sp_lhd = thd;
	if ((NULL == spap->sp_lhd) ||
	  ((NULL == thd->data8) && (NULL == thd->data16)))
	  {
		fprintf(stderr,
		 "Cannot do spatial correction: image is faulty\n");
		return endspa(EBAD, &spap);
	  }
	spap->sp_imsize = spap->sp_outwid * spap->sp_outht;
	spap->sp_dimratx = ((double)spap->sp_outwid) /
		((double)(thd->xdimen));
	spap->sp_dimraty = ((double)spap->sp_outht) /
		((double)(thd->ydimen));
	/* reserve memory for floating-point image and weight values */
	if (NULL == (spap->sp_locim = (float *)callop((size_t)
	 (2 * spap->sp_imsize), sizeof (float), "local image for spatial")))
	  {
		fprintf(stderr,
		 "Failure to reserve %d bytes for floating-point image\n",
			2 * spap->sp_imsize);
		return endspa(EALLOC, &spap);
	  }
	spap->sp_locwt = spap->sp_locim + spap->sp_imsize;
	/* if necessary, find the limits of x and y in cm */
	if ((fabs(spap->sp_xsl) < DFUZZ) && (EOK !=(sret = find_limits(spap))))
	  {
		fprintf(stderr, "find_limits: failure %d\n", sret);
		return endspa(sret, &spap);
	  }
	/* go through all the input points and find numerators
	 and denominators for the interpolated values */
	spap->sp_sflags = spflags;
	interp_image(spap);
	/* re-generate the output values by dividing the weighted
	 count values by the weights */
	noreps = normalize(spap);
	/* extrapolate to take care of missing pixels */
	noreps = extrap_pix(spap, noreps);
	/* phew. Now we can make an image out of this array. */
	if (EOK != (sret = creat_undistor(spap)))
	  {
		fprintf(stderr, "failure %d in creat_undistor\n", sret);
		return endspa(sret, &spap);
	  }
	memzap((AVOID **)&(spap->sp_locim), "local image for spatial");
	return endspa(EOK, &spap);
}

/* end of spatial.c */
