/* spline.c -- Routines for using the pixel-to-cm mapping

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	28 May 2007: handle 2-element, as well as 4-element, mosaic detectors
	27 May 2007: cm2pixq: iy is compared to v_midr, not v_midc
	19 Apr 2007: float->double conversions; reorganize use of "{"
	27 Dec 2006: reorganize remapping: add sumshift,sumpoly,modspline
	22 Dec 2006: more flexibility in remapping: add polyadd,polymap,
	 better I/O for old point-by-point remapping
	17 Jun 2006: position_code: replace act_refl call with activmask call
	 2 Jan 2006: replace record_detxy with internal bookkeeping
	16 Dec 2005: eliminate final static globals; reorg VSPLINE structure
	 what the static globals used to hold; rewrite gspline
	20 Nov 2005: inline argument declarations and strict prototyping
	27 Oct 2005: hide static globals in a SPLINEWORK structure
	12 Aug 2005: dowrite in remap_process() can be 0, 1, 2, 3;
	 image written out if dowrite & 01; verbose log if dowrite & 02.
	 6 Aug 2005: remap_process() now takes "dowrite" argument;
		if dowrite==0 we do a lot less logging of the remapping
	 1 Aug 2005: bugfix in remap_sums: use xyv[8]-[10], not xyv[0]-[2]
	28 Jun 2005: remap_setup(),remap_sums(),remap_process(),#defines added
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	19 Oct 2004: make_ftab called unconditionally from getfoff,
	 so getfoff doesn't need an argument
	28 Aug 2004: routines used only here are declared as static; cosmetics
	 2 Aug 2004: find_dimension() and record_detxy take arg `imagefirst';
		imtodim() facilitates that; change call in record_detxy
	 5 Mar 2004: call cleanspline() in pix2cm, etc. to be careful;
		reorder routines so we don't have to declare more things extern
	 4 Mar 2004: add routine cleanspline() to offer a safe input
	21 Jan 2004: license message inserted
	 8 Jan 2004: rewrite of find_dimension() and record_detxy because
		the variables they reference are handled differently
	 6 Jan 2004: global variables replaced with static globals;
		killspline, kill_foffb routines set up for memory control;
		gspline now takes an argument and returns a structure pointer
	10 Dec 2003: extern references to width, height now apply throughout
	23 Sep 2003: remove declaration of strcpy()--sometimes it's a macro
	19 Sep 2003: initialize gxmi, gxma, gymi, gyma in getfoff;
		move cm2pixq() forward so that dependencies work right
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Sep 2003: replace access() with xgaccess() call in find_dimension
		so we can gunzip USPLINE on the fly
	17 Aug 2003: trivial renaming of local variable to `finame'
	13 Aug 2003: putspline() moved from ../prerefine/calibmain.c,
		and now it takes an argument
	 5 Jan 2003: reorganized source-file header
	25 Mar 2002: removed MSI attachments
	10 Oct 1995: find_dimension() suppresses diagnostic if called from
		X-GEN parser;
		number of bytes read from spline=512+CALPOINT sizes;
		larger static char array `thisname'
	31 Aug 1995: `record_detxy' remembers static global width,
		height values; call it from find_dimension
	20 May 1995: fixed various bits of sloppy code
	12 May 1995: make_ftab takes no argument
	 2 Mar 1995: add variables myarc, myarv, foffb;
		delete diffv0, dvlast, gu0
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	 5 Oct 1994: use xgread to input VSPLINE data
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: polymask added; bugfixes in polymath, xtal, calinv, etc.
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version

   Purpose: processing the splined calibration data found in the
	VSPLINE structure pointed to by the static global lspl in "gspline()"
   Cross references: calinv.c, ../calibrate/cal*.[ch]

*************************************************************************/
#include	"xgen.h"
#define		SIGRANGE	(3.)
#define		ENOUGH		2
#define		DIAGONAL_RATIO	2
#define		PRIOR_WEIGHT	(2 * 7)

int detregion(float xva, float yva, VSPLINE *lspl)
{ /*	This returns the detector region in which the
	point (xva, yva) is contained. If the detector has only
	one region, then it always returns 0. With four regions,
	it returns thus:
     _________
    / 0  | 2  \
    |____|____|
    | 1  | 3  |
     \___|___/	*/
	float	midx, midy;

	if (lspl->v_nreg < 2) return 0;
	midx = 0.5 * (float)(lspl->v_detw); midy = 0.5 * (float)(lspl->v_deth);
	if (xva < midx) return (yva < midy) ? 0 : 1;
	else return (yva < midy) ? 2 : 3;
}

static void rawp2c(float rex, float rey, int jx, int jy,
 float *xca, float *yca)
{ /* This does the internal guts of the pix2cm calc without error checks */
	int		retv;
	double		p, q;
	VSPLINE		*lspl;
	CALPOINT	*cz00, *cz01, *cz10, *cz11;

	if (NULL == (lspl = gspline(3, &retv))) return;
	cz00 = lspl->v_cpt + jy * lspl->v_ncol + jx;
	cz01 = cz00 + 1; cz10 = cz00 + lspl->v_ncol; cz11 = cz10 + 1;
	p = rex - (double)jx;	q = rey - (double)jy;
	*xca = (1.-p) * (1.-q) * cz00->c_ctopx + (1.-p) * q * cz10->c_ctopx +
		   p  * (1.-q) * cz01->c_ctopx +     p  * q * cz11->c_ctopx;
	*yca = (1.-p) * (1.-q) * cz00->c_ctopy + (1.-p) * q * cz10->c_ctopy +
		   p  * (1.-q) * cz01->c_ctopy +     p  * q * cz11->c_ctopy;
}

static void rawc2p(float rax, float ray, int iax, int iay,
 float *xap, float *yap)
{ /* this does the internals of the cm2pix calc without error checks */
	int		retv;
	double		p, q;
	VSPLINE		*lspl;
	CALPOINT	*ca00, *ca01, *ca10, *ca11;

	if (NULL == (lspl = gspline(3, &retv))) return;
	ca00 = lspl->v_cpt + iay * lspl->v_ncol + iax;
	ca01 = ca00 + 1; ca10 = ca00 + lspl->v_ncol; ca11 = ca10 + 1;
	p = rax - iax;	q = ray - iay;
	*xap = (1.-p) * (1.-q) * ca00->c_ptocx + (1.-p) * q * ca10->c_ptocx +
		  p  * (1.-q) * ca01->c_ptocx +     p  * q * ca11->c_ptocx;
	*yap = (1.-p) * (1.-q) * ca00->c_ptocy + (1.-p) * q * ca10->c_ptocy +
		  p  * (1.-q) * ca01->c_ptocy +     p  * q * ca11->c_ptocy;
}

static int imtodim(int *widp, int *htp, char *finame)
{ /*	This determines detector dimensions from a detector image */
	int		imflags, fmnum, gotone;
	char		thisname[120], lline[120];
	static MAIL	thead;
	static FILE	*findx;
	static char	thdmsg[] = "thead.runheader in imtodim";

	zeromem((AVOID *)(&thead), sizeof (MAIL));
	gotone = 0;	imflags = JUST_HEADER;
	if (NULL != finame)
	  {
		gotone = (FRAME_OK == read_image(finame, imflags, &thead));
	  }
	else if (NULL != (findx = xgopen("frames", "r")))
	  {
		for ( ; NULL != fgets(lline, 118, findx); )
		   {
			thisname[0] = '\0';
			(void)sscanf(lline, "%d%s", &fmnum, thisname);
			if ('\0' != thisname[0])
			  {
				if (FRAME_OK == read_image(thisname,
					imflags, &thead))
				  {
					gotone = 1;	break;
				  }
			  }
		   }
		(void)fclose(findx);
		if (verror(gotone, "determine the detector's dimensions"))
			return EREADERR;
	  }
	if (gotone)
	  {
		*widp = thead.xdimen;	*htp = thead.ydimen;	fmnum = EOK;
	  }
	else {
		fmnum = EREADERR;
		fprintf(stderr, "Cannot determine detector dimensions\n");
	     }
	memzap((AVOID **)&(thead.runheader), thdmsg);	return fmnum;
}

int find_dimension(int *widp, int *htp, int imagefirst, char *finame)
{ /*	conservative code to find the detector dimensions. It first looks
	for a USPLINE file and tries to read the detector dimensions in it.
	If that fails, it tries to open a data frame--either the one named
	in the first argument or one of the frames in the FRAMES file--and
	calculates the detector dimensions by looking at the header thereof.
	If it can't find a data frame to check, it returns the error code
	from gspline(). */
	int	retv;
	char	thisname[120];
	VSPLINE	*lspl;

	if (NULL == (lspl = gspline(3, &retv))) return retv;
	*widp = *htp = 0;	/* explicitly set the output values to zero */
	if (imagefirst) /* we check an image first off in this case */
	  {
		if (EOK != (retv = imtodim(widp, htp, finame))) return retv;
	  }
	else if (0 >= lspl->v_xrcm)	/* read in spline if needed */
	  {
		strcpy(thisname, trname("USPLINE")); /* get USPLINE name */
		if (0 == xgaccess(thisname, 4, 1)) /* is USPLINE readable?*/
		  {
			(void)gspline(0, &retv); /* kill old spline memory */
			if (NULL == (lspl = gspline(5, &retv))) return retv;
		  }
	  }
	if (0 < lspl->v_xrcm)	/* if spline is okay now, use it */
	  {
		*widp = lspl->v_detw;	*htp = lspl->v_deth;
	  }
	/* if the values are reasonable, we can return */
	if ((*widp <= 0) || (*widp > 32760) || (*htp <= 0) || (*htp > 32760))
	  {
		/* Otherwise try to get them from a detector image */
		if (EOK != (retv = imtodim(widp, htp, finame))) return retv;
	  }
	/* record the dimensions for local use.
	 We don't use record_detxy explicitly for this any more. */
	lspl->v_detw = *widp;	lspl->v_deth = *htp;
	lspl->v_maxx = *widp + 10;	lspl->v_maxy = *htp + 10;
	return EOK;
}

int position_code(float xpix, float ypix)
{ /*	This returns 1 if the position (xpix, ypix) is in the active area.
	It returns 2 if (xpix, ypix) is within 10 pixels of the detector
	face, -1 otherwise. */
	int	usex, usey;
	VSPLINE	*lspl;

	if (NULL == (lspl = gspline(3, &usex))) return -1;
	usex = xpix + 0.49999;	usey = ypix + 0.499999;
	if (NULL != activmask(usex, usey, 0101)) return 1;
	if ((usex < -10) || (usex > lspl->v_maxx) || (usey < -10) ||
		(usey > lspl->v_maxy)) return -1;
	return 2;
}

int foffset(int opno)
{ /*	This manages the list of frame-number-dependent offsets.
	If opno == 0, we use this to eliminate memory for these offsets.
	Otherwise we use it to create the memory for the offsets and 
	read them in from the FRAMEOFFSET file.
	The routine returns EOK if there are actual offsets to work with;
	it returns various error values otherwise,
	including EBADVALUE if there aren't any offsets. */
	char			*offnam, offbu[40];
	int			fmnu, itemz, nbeam, nfr, nused;
	double			gx, gxma, gxmi, gy, gyma, gymi, mex, mey;
	FILE			*fpoff;
	VSPLINE			*lspl;
	FRAMEINDEX		*lfmi;
	FRAMEENTRY		*fte, *ftq;
	FRAMEOFF		*foffp;

	if (NULL == (lspl = gspline(3, &nused))) return nused;
	if (!opno)
	  {
		memzap((AVOID **)&(lspl->v_foffb), "foffb in foffset");
		return EOK; /* memory destruction ought to work... */
	  }
	if (NULL == (lfmi = ftables(3))) return EOPENERR;
	if ((NULL == (offnam = getenv("FRAMEOFFSET"))) || ('\0' == *offnam))
		return EOPENERR;
	if (NULL == (fpoff = xgopen(offnam, "r"))) return EOPENERR;
	mex = mey = nfr = 0;
	gxmi = gymi = 1.0e12;	gxma = gyma = -gxmi;
	while (NULL != fgets(offbu, 38, fpoff))
	     {
		itemz = sscanf(offbu, "%d %lf %lf", &fmnu, &gx, &gy);
		if (itemz >= 3)
		  {
			nfr++; mex += gx; mey += gy;
			if (gx > gxma) gxma = gx; if (gx < gxmi) gxmi = gx;
			if (gy > gyma) gyma = gy; if (gy < gymi) gymi = gy;
		  }
	     }
	if (!nfr) return EBADVALUE;
	nbeam = nfr;	mex /= (double)nfr;	mey /= (double)nfr;
	if (-1 == fseek(fpoff, 0L, 0)) return ESEEKERR;
	for (fte = lfmi->fi_fmi, fmnu = 0; fte <= lfmi->fi_fme &&
	 fte->fmno > 0 && fte->fmname[0] != '\0'; fte++) fmnu++;
	if (NULL == (lspl->v_foffb = (FRAMEOFF *)callop(fmnu + 2,
	 sizeof (FRAMEOFF), "foffb in foffset"))) return EALLOC;
	for (foffp = lspl->v_foffb, ftq = lfmi->fi_fmi;
	 ftq < lfmi->fi_fme; foffp++, ftq++) foffp->fmfm = ftq->fmno;
	nused = 0;
	while (NULL != fgets(offbu, 38, fpoff))
	     {
		itemz = sscanf(offbu, "%d %lf %lf", &fmnu, &gx, &gy);
		if ((itemz >= 3) && (lfmi->fi_firstfm <= fmnu) &&
		 (fmnu <= lfmi->fi_lastfm))
		  {
			foffp = lspl->v_foffb + fmnu - lfmi->fi_firstfm;
			if (foffp->fmfm == fmnu)
			  {
				foffp->fmdx = gx - mex;
				foffp->fmdy = gy - mey;	nused++;
			  }
		  }
	     }
	printf("%3d beam-position values read from %s\n", nbeam, offnam);
	printf("X,Y ranges: %6.1f to %6.1f,\t\t%6.1f to %6.1f respectively\n",
		gxmi, gxma, gymi, gyma);
	printf("Mean position = %6.1f, %6.1f\n", mex, mey);
	if (nused < nbeam) printf(
		"%4d frames have beam-position offsets defined\n", nused);
	(void)fclose(fpoff);	fpoff = NULL;
	return EOK;
}

void adjust_to_fm(float *rxp, float *ryp, int fno, int direction)
{ /*	This adjusts a position (*rxp, *ryp) by a offset that depends on the
	frame number. We no longer check for existence of FRAMEINDEX here. */
	int		retv;
	char		*xyoff;
	VSPLINE		*lspl;
	FRAMEOFF	*tfoff;

	if (fno < 1) return;
	if ((NULL == (xyoff = getenv("FRAMEOFFSET"))) || ('\0' == *xyoff))
		return;
	if (NULL == (lspl = gspline(3, &retv))) return;
	if (NULL == lspl->v_foffb)
	  {
		if ((EOK != foffset(1)) || (NULL == lspl->v_foffb)) return;
	  }
	for (tfoff = lspl->v_foffb; 0 != tfoff->fmfm; tfoff++)
	   if (tfoff->fmfm == fno)
	     {
		if (direction)
		  {
			*rxp -= tfoff->fmdx; *ryp -= tfoff->fmdy;
		  }
		else {
			*rxp += tfoff->fmdx; *ryp += tfoff->fmdy;
		     }
		return;
	     }
}

VSPLINE *cleanspline(VSPLINE *oldspl, int *resultp)
{ /*	This returns a pointer to a real VSPLINE structure if one is
	available. This protects us from bizarre situations where we
	might otherwise have messed-up splines. */
	VSPLINE	*lospl;

	if (NULL == oldspl)
	  {
		lospl = gspline(5, resultp);
		if ((NULL != lospl) && (EOK != *resultp))
		  {
			gspline(0, resultp);	lospl = NULL;
		  }
	  }
	else if ((oldspl->v_detw < 100) || (oldspl->v_deth < 100) ||
		 (oldspl->v_detw > 10000) || (oldspl->v_deth > 10000))
	  {
		gspline(0, resultp);	lospl = NULL;	*resultp = EBADVALUE;
	  }
	else {
		lospl = oldspl;	*resultp = EOK;
	     }
	return lospl;
}

int pix2cm(float xp, float yp, float *xcm, float *ycm, int fno)
{ /*	This returns the x,y coordinates in cm of a position
	specified in pixel coordinates.
	Return values:
		 1	: cm position found; (x,y) inside active area
		 2	: cm position found; (x,y) outside active area
				but still close to detector
		-1	: cm position found; (x,y) outside active area
				and a long way from the detector
		 0	: cm position couldn't be determined. */
	int		retv;
	VSPLINE		*lspl;
	static int	ix, iy;
	static double	rx, ry;
	float		dx, dy;

	if (NULL == (lspl = gspline(3, &retv)))
	  {
		if (NULL == (lspl = gspline(5, &retv))) return 0;
	  }
	dx = xp - lspl->v_x0px;	dy = yp - lspl->v_y0px;
	if (fno > 0) adjust_to_fm(&dx, &dy, fno, 1);
	rx = dx / lspl->v_xrpx;	ry = dy / lspl->v_yrpx;
	ix = (rx < 0. ? 0 : (rx < lspl->v_ncol - 1 ?
		rx : lspl->v_ncol - 2));
	iy = (ry < 0. ? 0 : (ry < lspl->v_nrow - 1 ?
		ry : lspl->v_nrow - 2));
	if (lspl->v_nreg == 4)
	  {
		if (ix == lspl->v_ncol / 2 - 1) ix--;
		else if (ix == lspl->v_ncol / 2) ix++;
		if (iy == lspl->v_nrow / 2 - 1) iy--;
		else if (iy == lspl->v_nrow / 2) iy++;
	  }
	rawp2c(rx, ry, ix, iy, xcm, ycm);
	return position_code(xp, yp);
}

int cm2pixq(float xcm, float ycm, int lquad, float *xp, float *yp, int fno)
{ /*	This returns the x,y coordinates in pixels of a position specified
	in cm.  The variable "lquad" specifies which quadrant we are to
	draw the fiducial points from.
	Return values:
		 1	: pixel position found; (x,y) inside active area
		 2	: pixel position found; (x,y) outside active area
				but still close to detector
		-1	: pixel position found; (x,y) outside active area
				and a long way from the detector
		 0	: pixel position couldn't be determined. */
	int		ix, iy, qmin, retv;
	double		dx, dy, rx, ry;
	VSPLINE		*lspl;
	CALPOINT	*cw;

	/* since we sometimes call this from somewhere weird,
	 we do a bit of a triple-check that lspl is non-NULL. */
	if (NULL == (lspl = gspline(3, &retv)))
	  {
		if (NULL == (lspl = gspline(5, &retv))) return 0;
	  }
	dx = xcm - lspl->v_x0cm;		rx = dx / lspl->v_xrcm;
	ix = (rx < 0. ? 0 : (rx < lspl->v_ncol - 1 ?
		rx : lspl->v_ncol - 2));
	dy = ycm - lspl->v_y0cm;		ry = dy / lspl->v_yrcm;
	iy = (ry < 0. ? 0 : (ry < lspl->v_nrow - 1 ?
		ry : lspl->v_nrow - 2));
	if (lspl->v_nreg < 2) rawc2p(rx, ry, ix, iy, xp, yp);
	else {	/* multi-element */
		qmin = (lspl->v_nreg == 2) ? 1 : 2;
		if (ix == lspl->v_midc - 1) ix--;
		else if (ix == lspl->v_midc)
		  {
			cw = lspl->v_cpt + iy * lspl->v_ncol + ix;
			if (cw->c_ptocx < lspl->v_detw / 2.) ix++;
		  }
		if (iy == lspl->v_midr - 1) iy--;
		else if (iy == lspl->v_midr)
		  {
			cw = lspl->v_cpt + iy * lspl->v_ncol + ix;
			if (cw->c_ptocy < lspl->v_deth / 2.) iy++;
		  }
		if (lquad < qmin)
		  {
			if (ix >= lspl->v_midc) ix = lspl->v_midc - 2;
		  }
		else if (ix < lspl->v_midc) ix = lspl->v_midc;
		if (lspl->v_nreg > 2)
		  {
			if ((lquad == 0) || (lquad == 2))
			  {
				if (iy >= lspl->v_midr)
					iy = lspl->v_midr - 2;
			  }
			else if (iy < lspl->v_midr) iy = lspl->v_midr;
		  }
		rawc2p(rx, ry, ix, iy, xp, yp);
	     }
	if (fno > 0) adjust_to_fm(xp, yp, fno, 0);
	return position_code(*xp, *yp);
}

int cm2pix(float xcm, float ycm, float *xp, float *yp, int fno)
{ /*	This returns x,y coordinates in pixels of a position specified in cm.
	Return values:
		 1	: pixel position found; (x,y) inside active area
		 2	: pixel position found; (x,y) outside active area
				but still close to detector
		-1	: pixel position found; (x,y) outside active area
				and a long way from the detector
		 0	: pixel position couldn't be determined.
	Since we need to provide for multi-element detectors, we do this
	calculation in the following way:
	 (a) if lspl->v_nreg is less than 2, we simply call cm2pixq with
	    a quadrant argument of zero.
	 (b) if lspl->v_nreg is >= 2, we guess which quadrant it's going
	    to fall into and call cm2pixq with that argument. If that works,
	    we return the result. If it fails, we try other quadrants. */
	int	gquad, oquad, retv;
	double	boundx, boundy;
	VSPLINE	*lspl;

	/* for single-quadrant detectors, we just call cm2pixq with quad 0 */
	if (NULL == (lspl = gspline(3, &retv)))
	  {
		if (NULL == (lspl = gspline(5, &retv))) return 0;
	  }
	if (lspl->v_nreg < 2) return cm2pixq(xcm, ycm, 0, xp, yp, fno);
	/* to guess which quadrant it falls in, we first estimate
	 the cm coordinates of the detector center */
	boundx = lspl->v_x0cm + ((double)lspl->v_midc) * lspl->v_xrcm;
	boundy = lspl->v_y0cm + ((double)lspl->v_midr) * lspl->v_yrcm;
	/* guess which quadrant the point actually falls in */
	gquad = (ycm <= boundy) + 2 * (xcm >= boundx);
	/* now call cm2pixq with that quadrant */
	retv = cm2pixq(xcm, ycm, gquad, xp, yp, fno);
	/* did we end up in the quadrant we started in? */
	if (gquad == (oquad = detregion(*xp, *yp, lspl))) return retv;
	/* if not, try the quadrant that it resulted */
	gquad = oquad;
	retv = cm2pixq(xcm, ycm, gquad, xp, yp, fno);
	if (gquad == (oquad = detregion(*xp, *yp, lspl))) return retv;
	/* if it's still wrong, we give up */
	return 0;
}

VSPLINE *gspline(int opno, int *errtp)
{ /* This manipulates a VSPLINE structure according to the value of opno.
	If opno == 0, we free the memory for lspl and return NULL.
	If opno == 1, we simply create a structure and return a pointer to it.
	If opno == 3, we return a pointer to a VSPLINE structure,
		creating it only if necessary.
	If opno == 5, we start over, i.e. we populate a VSPLINE structure
	from a diskfile, and then return it.
	If opno == 9, we write out a VSPLINE structure to a diskfile. */

	char		*cpname;
	int		nread, ntoread;
	size_t		ntw;
	static VSPLINE	*lspl;
	static FILE	*fspl;

	*errtp = EOK;
	if (opno == 0)
	  {
		if (NULL != lspl)
		  {
			memzap((AVOID **)&(lspl->v_cpt), "v_cpt in gspline");
			memzap((AVOID **)&(lspl->v_foffb), "foffb in gspline");
			memzap((AVOID **)&lspl, "lspl in gspline");
		  }
		return NULL;
	  }
	if (NULL == lspl)
	  {	
		if (NULL == (lspl = (VSPLINE *)callop(1, sizeof (VSPLINE),
		 "lspl in gspline"))) return NULL;
	  }
	if (opno == 3) return lspl;	/* in this case, we just receive */
	if (opno == 5)
	  {
		cpname = xgprogname(NULL, 0);
		memzap((AVOID **)&(lspl->v_cpt), "v_cpt in gspline");
		/* diagnostic is suppressed if we're just calling it from the
		 parser, from "setup", from "pmotif", or from "calibrate" */
		if ((NULL != cpname) && (!strcmp(cpname, "xgen parser") ||
		 !strcmp(cpname, "setup") || !strcmp(cpname, "pmotif") ||
		 !strcmp(cpname, "calibrate")))
		  {
			if (NULL == (fspl = xgopen("uspline", "r")))
			  {
				*errtp = EOPENERR;	return NULL;
			  }
		  }
		else {
			if (qerror((AVOID *)(fspl = xgopen("uspline", "r")),
			 "open spline file"))
			 {
				*errtp = EOPENERR;	return NULL;
			 }
		  }
		nread = xgread((void *)lspl, XG_VSPLINE, fspl);
		ntoread = 512 + lspl->v_nrow * lspl->v_ncol * sizeof (CALPOINT);
		if (nread < ntoread)
		  {
			fprintf(stderr,
	 "Warning: expected to read %d bytes from spline; actually read %d\n",
				ntoread, nread);
			*errtp = EREADERR;	return NULL;
		  }
		(void)fclose(fspl);	fspl = NULL;
		lspl->v_maxx = lspl->v_detw + 10;
		lspl->v_maxy = lspl->v_deth + 10;
		return lspl;
	  }
	if (opno == 9) /* write current spline out if it's sensible */
	  {
		if ((NULL == lspl->v_cpt) || (lspl->v_detw < 2) ||
		 (lspl->v_deth < 2))
		  {
			fprintf(stderr, "Current spline is bad: no output\n");
			*errtp = EBADVALUE;	return lspl;
		  }
		if (NULL == (fspl = xgopen("uspline", "w")))
		  {
			*errtp = EWRITERR;	return NULL;
		  }
		ntw = 512 + lspl->v_nrow * lspl->v_ncol * sizeof (CALPOINT);
		if (ntw > xgwrite((void *)lspl, XG_VSPLINE, fspl))
			*errtp = EWRITERR;
		(void)fclose(fspl);	fspl = NULL;
	  }
	return (*errtp != 0) ? NULL : lspl;
}

double *remap_setup(VSPLINE *losp, double *flp)
{ /*	This sets up for a remapping effort. It returns a pointer to a
	group of floating-point values that will be filled up later.
	On return, flp[0] - flp[2] will be filled as follows:
	 flp[0]	denominator for weighting calculations
	 flp[1] maximum range
	 flp[2] maximum arg */
	int		ix, iy, npts;
	static double	*adel;
	CALPOINT	*cp00, *cp11;
	double		delx, dely, maxarg, maxrange, radsigma, wtden;

	if (losp->v_nreg > 1) /* avoid the region close to the very center*/
	  {
		ix = losp->v_midc * 0.7;	iy = losp->v_midr * 0.7;
	  }
	else {
		ix = losp->v_midc;	iy = losp->v_midr;
	     }
	/* define a characteristic radius for each fiducial point
	 in cm.  Currently it's half of the distance between two
	 fiducial points related diagonally, e.g. [8][8] to [9][9].*/
	cp11 = losp->v_cpt + iy * losp->v_ncol + ix;
	cp00 = cp11 - losp->v_ncol - 1;
	delx = cp11->c_ctopx - cp00->c_ctopx;
	dely = cp11->c_ctopy - cp00->c_ctopy;
	radsigma = sqrt(delx * delx + dely * dely) / DIAGONAL_RATIO;
	wtden = 1. / (2. * radsigma * radsigma);	
	/* we'll use points within 2.5*sigma of each fiducial point */
	maxrange = SIGRANGE * radsigma;
	maxarg = SIGRANGE * SIGRANGE / 2.;
	npts = losp->v_nrow * losp->v_ncol;
	if (NULL == (adel = (double *)callop((unsigned)(3 * npts),
		sizeof (double), "adel in remap_setup")))
	  {
		fprintf(stderr,
		 "Cannot reserve memory for work variables in remapping\n");
		return NULL;
	  }
	for (iy = 0, cp00 = losp->v_cpt; iy < losp->v_nrow; iy++)
	   for (ix = 0; ix < losp->v_ncol; ix++, cp00++) cp00->c_ift = 0;
	flp[0] = wtden;	flp[1] = maxrange;	flp[2] = maxarg;
	return adel;
}

int remap_sums(VSPLINE *losp, double *xyv, double *adel)
{ /*	This contributes to the sums used in remapping. */
	int		ix, ixmin, ixmax, iy, iymin, iymax, nhere, npts, uyx;
	double		delx, delx0, dely, dely2, pweight;
	double		*adely, *adeld, *pdx, *pdy, *pdd;
	CALPOINT	*cp00;

	/* determine range in (ix, iy) over which fiducial points
	 can reasonably be expected to be influenced by this spot. */
	npts = losp->v_nrow * losp->v_ncol;
	adely = adel + npts;	adeld = adely + npts;
	iymin = (xyv[4] + xyv[9] - losp->v_y0cm) / losp->v_yrcm;
	if (iymin < 0) iymin = 0;
	else if (iymin >= losp->v_nrow) iymin = losp->v_nrow - 1;
	iymax = (xyv[4] - xyv[9] - losp->v_y0cm) / losp->v_yrcm + 0.49999;
	if (iymax < 0) iymax = 0;
	if (iymax >= losp->v_nrow) iymax = losp->v_nrow - 1;
	ixmin = (xyv[3] - xyv[9] - losp->v_x0cm) / losp->v_xrcm;
	if (ixmin < 0) ixmin = 0;
	else if (ixmin >= losp->v_ncol) ixmin = losp->v_ncol - 1;
	ixmax = (xyv[3] + xyv[9] - losp->v_x0cm) / losp->v_xrcm + 0.49999;
	if (ixmax < 0) ixmax = 0;
	if (ixmax >= losp->v_ncol) ixmax = losp->v_ncol - 1;
	/* further restrictions apply if this is a 2x2 system*/
	if (losp->v_nreg == 4)
	  {
		if ((xyv[6] < losp->v_detw / 2) && (ixmax >= losp->v_midc))
			ixmax = losp->v_midc - 1;
		if ((xyv[7] < losp->v_deth / 2) && (iymax >= losp->v_midr))
			iymax = losp->v_midr - 1;
		if ((xyv[6] > losp->v_detw / 2) && (ixmin < losp->v_midc))
			ixmin = losp->v_midc;
		if ((xyv[7] > losp->v_deth / 2) && (iymin < losp->v_midr))
			iymin = losp->v_midr;
	  }
	dely  = losp->v_y0cm + ((double)iymin) * losp->v_yrcm - xyv[4];
	delx0 = losp->v_x0cm + ((double)ixmin) * losp->v_xrcm - xyv[3];
	/* go through the points in a rectangle around this reflection and
	 use deltax, deltay values for this refl to influence the
	 fiducial map at those points. */
	for (iy = iymin, nhere = 0; iy <= iymax; iy++, dely+= losp->v_yrcm)
	   {
		dely2 = dely * dely;
		uyx = iy * losp->v_ncol + ixmin;
		cp00 = losp->v_cpt + uyx;
		pdx = adel + uyx;	pdy = adely + uyx; pdd = adeld + uyx;
		for (ix = ixmin, delx = delx0; ix <= ixmax;
		 ix++, delx += losp->v_xrcm, cp00++, pdx++, pdy++, pdd++)
		  if ((pweight = (dely2 + delx * delx) * xyv[8]) < xyv[10])
		    { /* use distance to compute lsq weight*/
			pweight = xyv[5] * exp(-pweight);
			/* now increment the lsq sums */
			*pdx += pweight * xyv[0];
			*pdy += pweight * xyv[1]; *pdd += pweight;
			if (cp00->c_ift < 254) (cp00->c_ift)++;
			nhere++;
		    } /* close if clause & loop over ix */
	   } /* close loop over iy */
	return EOK;
}

static void prdels(FILE *fp, VSPLINE *losp, double *adel)
{ /* This prints the table of deltax, deltay values */
	int		ix, iy;
	double		*pdx, *pdy;
	CALPOINT	*cp00;

	pdx = adel;	pdy = pdx + losp->v_nrow * losp->v_ncol;
	for (iy = 0, cp00 = losp->v_cpt; iy < losp->v_nrow; iy++)
	   {
		for (ix = 0; ix < losp->v_ncol; ix++, cp00++, pdx++)
		   {
			if (cp00->c_ift) fprintf(fp, " %5.2f", *pdx);
			else fprintf(fp, " _____");
			if (!((ix+1) % 22)) fprintf(fp, "\n");
		   }
		fprintf(fp, "\n");
		cp00 -= losp->v_ncol;
		for (ix = 0; ix < losp->v_ncol; ix++, cp00++, pdy++)
		   {
			if (cp00->c_ift) fprintf(fp, " %5.2f", *pdy);
			else fprintf(fp, " _____");
			if (!((ix+1) % 22)) fprintf(fp, "\n");
		   }
		fprintf(fp, "\n");
	 }
}

static int nbrshift(int ix, int iy, VSPLINE *losp, double *adel)
{ /* enables an estimate of the deltax, deltay values in a particular
	region of the detector based on values in neighboring regions */
	int		jxmin, jxmax, jymin, jymax, jx, jy, npts;
	double		delx, dely, delr, den, numx, numy, wt, *pdx, *pdy;
	CALPOINT	*cp0, *cp1, *cpy;

	cp0 = losp->v_cpt + iy * losp->v_nrow + ix;	/* find current point*/
	if (cp0->c_ift) return 0; /* don't do anything if estimate exists */
	npts = losp->v_nrow * losp->v_ncol;
	/* define rectangular region close to current box */
	jxmin = ix - 2;	if (jxmin < 0) jxmin = 0;
	jxmax = ix + 2;	if (jxmax >= losp->v_ncol) jxmax = losp->v_ncol - 1;
	jymin = iy - 2;	if (jymin < 0) jymin = 0;
	jymax = iy + 2;	if (jymax >= losp->v_nrow) jymax = losp->v_nrow - 1;
	cpy = losp->v_cpt + jymin * losp->v_nrow;
	for (jy = jymin, numx = numy = den = 0.;
	 jy <= jymax; jy++, cpy += losp->v_nrow)
	   {
		dely = ((double)(jy - iy)) * losp->v_yrcm;	dely *= dely;
		for (cp1 = cpy + jxmin, jx = jxmin; jx <= jxmax; jx++, cp1++)
		   {
			delx = ((double)(jx - ix)) * losp->v_xrcm; delx *= delx;
			delr = delx + dely; /* radial distance from this box*/
			if ((delr > DFUZZ) && (cp1->c_ift))
			  {
				pdx = adel + jy * losp->v_nrow + jx;
				pdy = pdx + npts;
				wt = ((double)(cp1->c_ift)) / delr;
				numx += *pdx * wt; numy += *pdy * wt;
				den += wt;
			  }
		   }
	   }
	if (den > DFUZZ)
	  {
		pdx = adel + iy * losp->v_nrow + ix;	pdy = pdx + npts;
		*pdx = numx / den;	*pdy = numy / den;
		cp0->c_ift = 1;	/* code for averaging technique */
		(losp->v_ntrue)++;	return 1;
	  }
	return 0;
}

void polyadd(VSPLINE *losp, double *wve)
{ /* This adds a reflection's worth of data to a polynomial remapping.
	The vector wve contains [Xob,Yob,Xc,Yc,wt] */
	double	dix, diy, wt, wx, wy, wo[6];

	wt = wve[4];
	wx = wve[0] / ((double)(losp->v_detw)) - 0.5;
	wy = wve[1] / ((double)(losp->v_deth)) - 0.5;
	dix = wve[0] - wve[2];	diy = wve[1] - wve[3];
	wo[0] = wt;	wo[1] = wt * wx;	wo[2] = wt * wy;
	wo[3] = wt * wx * wx;	wo[4] = wt * wx * wy;	wo[5] = wt * wy * wy;
	inv_add(losp->v_lax, losp->v_ldx, wo, dix, 6);
	inv_add(losp->v_lay, losp->v_ldy, wo, diy, 6);
}

static void sumshift(FILE *fp, VSPLINE *losp, double xscen, double yscen)
{ /* Convenience routine for reporting changes in a remapping */
	fprintf(fp,
	 " %6d fiducial points remapped: Max, RMS shift=%7.2f,%7.2f pixels\n",
		losp->v_ntrue, losp->v_maxshift, losp->v_rmsshift);
	fprintf(fp, "beam center at [%8.4lf,%8.4lf]\n", xscen, yscen);
}

int modspline(VSPLINE *losp, FILE *fp, int outflag)
{ /* This renders a pixel-to-cm mapping self-consistent. It then outputs
	the resulting VSPLINE structure to a file if outflag & 010;
	It reports some results to stdout or a logfile if outflag & 03.
	It returns EOK if successful, various errors if not */
	int		ier, ix, iy, quadr;
	float		xscen, yscen;
	CALPOINT	*cpt;

	if (EOK != calinv(fp, losp, outflag)) /* invert mapping */
	  {
		fprintf(stderr, "Memory-allocation error in calinv\n");
		return EALLOC;
	  }
	/* determine the new beam center position */
	if (-1 == pix2cm(losp->v_axc, losp->v_ayc, &xscen, &yscen, 0))
	  {
		fprintf(stderr, "beam center could not be re-determined\n");
		return EBAD;
	  }
	if (outflag & 010)
	  {
		sumshift(stdout, losp, xscen, yscen);
		if (fp != stdout) sumshift(fp, losp, xscen, yscen);
	  }
	if (xscen != 0. || yscen != 0.)	/*modify center based on these data */
	  {
		losp->v_x0cm -= xscen;	losp->v_y0cm -= yscen;
		quadr = detregion(losp->v_axc, losp->v_ayc, losp);
		for (iy = 0, cpt = losp->v_cpt; iy < losp->v_nrow; iy++)
		   for (ix = 0; ix < losp->v_ncol; ix++, cpt++)
		      {
			cpt->c_ctopx -= xscen;	cpt->c_ctopy -= yscen;
		      }
		(void)cm2pixq(0., 0., quadr, &(losp->v_rxc), &(losp->v_ryc), 0);
	  }
	else {
		losp->v_rxc = losp->v_axc;	losp->v_ryc = losp->v_ayc;
	     }
	if ((outflag & 020) &&
	 ((NULL == (losp = gspline(9, &ier))) || (EOK != ier)))
	  {
		fprintf(stderr, "Error %d outputting USPLINE\n", ier);
		return EWRITERR;
	  }
	return EOK;
}

static void sumpoly(FILE *fp, VSPLINE *losp)
{ /* Convenience routine for summarizing a polynomial remapping */
	fprintf(fp, " Parameters of the remapping:\n");
	fprintf(fp,
	 "  constant   coeff(x)   coeff(y) coeff(x^2) coeff(x*y) coeff(y^2)\n");
	fprintf(fp, "X:%8.3f %10.3f %10.3f %10.3f %10.3f %10.3f\n",
	 losp->v_dex[0], losp->v_dex[1], losp->v_dex[2],
	 losp->v_dex[3], losp->v_dex[4], losp->v_dex[5]);
	fprintf(fp, "Y:%8.3f %10.3f %10.3f %10.3f %10.3f %10.3f\n",
	 losp->v_dey[0], losp->v_dey[1], losp->v_dey[2],
	 losp->v_dey[3], losp->v_dey[4], losp->v_dey[5]);
}

int polymap(VSPLINE *losp, FILE *fp, int reflags)
{ /* This does a polynomial remapping on the VSPLINE structure *losp. */
	int		ier, ix, iy;
	double		maxshift, ox, oy, rht, rwid;
	double		scx, scy, shift, sumrms, *flp;
	CALPOINT	*cpt;

	/* solve the linear systems for the deltax, deltay coefficients */
	if (EOK != (ier = inv_solve(losp->v_lax, losp->v_ldx, losp->v_dex, 6)))
		return ier;
	if (EOK != (ier = inv_solve(losp->v_lay, losp->v_ldy, losp->v_dey, 6)))
		return ier;
	sumpoly(stdout, losp);
	if (fp != stdout) sumpoly(fp, losp);
	/* Having calculated the adjustments, we apply them */
	rwid = losp->v_detw;	rht = losp->v_deth;	maxshift = 0.;
	for (sumrms = iy = 0, cpt = losp->v_cpt; iy < losp->v_nrow; iy++)
	   {
		for (ix = 0; ix < losp->v_ncol; ix++, cpt++)
		   {
			scx = cpt->c_ptocx / rwid - 0.5;
			scy = cpt->c_ptocy / rht - 0.5;
			flp = losp->v_dex;
			ox = flp[0] + flp[1] * scx + flp[2] * scy + flp[3] *
			 scx * scx + flp[4] * scx * scy + flp[5] * scy * scy;
			flp = losp->v_dey;
			oy = flp[0] + flp[1] * scx + flp[2] * scy + flp[3] *
			 scx * scx + flp[4] * scx * scy + flp[5] * scy * scy;
			cpt->c_ptocx += ox;	cpt->c_ptocy += oy;
			shift = ox * ox + oy * oy;
			if (shift > maxshift) maxshift = shift;
			sumrms += shift;
		   }
	   }
	losp->v_ntrue = losp->v_nrow * losp->v_ncol;
	losp->v_rmsshift = sqrt(sumrms / (double)losp->v_ntrue);
	losp->v_maxshift = sqrt(maxshift);
	if (reflags) reflags |= 020; /* output VSPLINE if in verbose mode */
	return modspline(losp, fp, reflags);
}

int remap_process(VSPLINE *losp, double *adel, FILE *fp, int dowrite)
{ /*	This does the post-accumulation processing for remapping.
	The results are written to disk if dowrite is nonzero.
	Logging is detailed if dowrite & 02.*/
	int		ix, iy, nadded, npass, npts;
	double		maxshift, rmsshift, sumrms;
	double		*pdx, *pdy, *pdd;
	CALPOINT	*cp00;
	static char	insuf[] =
		 "Only %d fiducial points remapped; remapping aborted\n";

	npts = losp->v_nrow * losp->v_ncol;
	maxshift = sumrms = losp->v_ntrue = 0;
	/* now compute the mean offsets near each point and print 'em.
	We will refine with some tethering accomplished by adding PRIOR_WEIGHT
	to each of the sums; this value corresponds to 2 "ordinary"
	reflections contributing at modest weight to each sum */
	pdx = adel;	pdy = adel + npts;	pdd = pdy + npts;
	for (iy = 0, cp00 = losp->v_cpt; iy < losp->v_nrow; iy++)
	   {
		for (ix = 0; ix < losp->v_ncol;
		 ix++, cp00++, pdx++, pdy++, pdd++)
		   {
			if ((cp00->c_ift < ENOUGH) || (*pdd < DFUZZ))
				*pdx = *pdy = cp00->c_ift = 0;
			else {
				*pdx /= (*pdd + PRIOR_WEIGHT);
				*pdy /= (*pdd + PRIOR_WEIGHT);
				rmsshift = *pdx * *pdx + *pdy * *pdy;
				sumrms += rmsshift;
				if (rmsshift > maxshift) maxshift = rmsshift;
				cp00->c_ptocx += *pdx; cp00->c_ptocy += *pdy;
				(losp->v_ntrue)++;
			     }
		   } /* close of loop over ix */
	   } /* close of loop over iy */
	if (losp->v_ntrue <= 3)
	  {
		fprintf(fp, insuf, losp->v_ntrue);
		if ((fp != stdout) && !(dowrite & 02))
			fprintf(stdout, insuf, losp->v_ntrue);
		return 6;
	  }
	losp->v_rmsshift = sqrt(sumrms / (double)losp->v_ntrue);
	losp->v_maxshift = sqrt(maxshift);
	/* Dec 2006: try to fill in the missing regions */
	for (npass = 0; npass < 4; npass++)
	   {	
		if (losp->v_ntrue < npts)
		   {
			for (nadded = iy = 0; iy < losp->v_nrow; iy++)
			   for (ix = 0; ix < losp->v_ncol; ix++)
			      {
				nadded += nbrshift(ix, iy, losp, adel);
				if (losp->v_ntrue >= npts) break;
			      }
			if (losp->v_ntrue >= npts)
			  {
				fprintf(fp, "Mapping completed on pass %d\n",
					npass);	break;
			  }
			fprintf(fp, " %4d more points remapped in pass %d\n",
				nadded, npass);
		  }
	   }
	if (dowrite & 02) fprintf(fp,
	 " Computed X and Y offsets near fiducial points:\n");
	if (dowrite & 02) prdels(fp, losp, adel); /* report on adjustments */
	if (dowrite) dowrite |= 020; /* output VSPLINE if in verbose mode */
	memzap((AVOID **)&adel, "adel in remap_process"); /* kill adel mem */
	return modspline(losp, fp, dowrite); /* invert mapping and output */
}

/* end of spline.c */
