/* xgbravais.c -- Bravais-lattice and lattice character calculations for X-GEN

	Copyright (c) 2003, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	 6 Jul 2007: don't flip monoclinics or hexagonals with one obtuse angle
		if the others are close to 90;
		use `rhombohex' to handle hexagonal->rhombohedral conversion
	 5 Jul 2007: change symmetry penalties: rhombohedral better, mC worse
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	28 Nov 2006: reorganize so it's independent of REFPARS
	15 Dec 2005: replace killbravais with allbravais(NULL, &isp);
		this eliminates the need for a static global
	22 Nov 2005: inline argument declarations and strict prototyping
	21 Jun 2005: comments cleaned up
	20 Jan 2005: spgtocen(), spgtosys(): International Tables
	 spacegroup used to calculate centering and crystal system
	28 Aug 2004: routines used only here are declared as static; cosmetics
	18 Jun 2004: orthoc_fudge now includes monoclinic C also
	 4 Jun 2004: orthoc_fudge added: artificially worsens fom for C-centered
		orthorhombic when the P-hexagonal cell is roughly as good
	26 May 2004: add lc_fom concept and use it for sorting Bravais results;
		routine sympen() installs symmetry penalty values;
		printouts reformatted
	24 Jan 2004: various static globals converted to static locals
	21 Jan 2004: license message inserted
	 8 Jan 2004: replace globals with references to REFPARS structures
		(including an argument to show_bravchoice())
		and a static global array of LATTICESCOREs;
		killbravais handles memory deletion;
		bravaismain() moved to ../mains
	27 Nov 2003: sort Bravais choices _after_ finding rhombohedral H
		cell as well as beforehand so it appears in the right order
	 8 Nov 2003: convert lca from vector to pointer;
		reserve memory for it in allbravais()
	 3 Nov 2003: created from parts of ./xgenmisc.c and
		../xgrefine/refxtal.c
 Relevant changes to predecessors:
	19 Sep 2003: initialize multi-dimensional arrays according to ansi C;
	 convert nig[] to a one-dimensional array so it's easier to initialize
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Sep 2003: add routines glatfi4() and gcenfi() for diagnostics
	24 Aug 2003: manipulate and display lattice type & centering
		using routines lattcode, centercode, glattice, gcentering
	16 Aug 2003: eliminate unneeded calls to exclucode()
	13 Aug 2003: many routines moved into this file from elsewhere
		so that this one is almost the only one that has to be
		included in some of the executables
	20 Jan 2003: bugfix for orthorhombic C cells in bravais()
	16 Jan 2003: correct the quality factors in bravais()
		for primitive cells and a few centered cells
		so that the scaling is done right.

   Purpose: routines used for determining Bravais lattices.
   Cross references: ./xgenmisc.c, ../xgrefine/ref*.c

*************************************************************************/
#include	"xgen.h"
#define	ALMOSTZERO	4.E-4	/* cos(88.85) squared */
#define	ORTHO_COMPARE	0.010	/* how close C222 and P6 need to be */
#define	ORTHO_OFFSET	0.003	/* ... move fom(C222) up to fom(P6) +this */

static int lcacmp(LATTICESCORE *lcap0, LATTICESCORE *lcap1)
/* compare routine used in sorting on score */
{
		return (lcap0->lc_fom < lcap1->lc_fom) ? -1 : 1;
}

int lattcode(int latva)
{ /*	Returns an integer between 0 and 7 for lattice type */
	static int		lcode[] = { 0,	/* code zero */
 5, 7, 5, 7, 5, 4, 4, 3,	/* 1 - 8 */
 7, 2, 4, 6, 3, 2, 4, 3, 2,	/* 9 -17 */
 4, 3, 2, 4, 6, 3, 7, 2,	/*18 -25 ; below is 26-45 */
 3, 2, 2, 2, 2, 1, 3, 2, 2, 2, 3, 2, 3, 2, 3, 2, 3, 2, 1, 7 } ;

	return	((0 <= latva) && (latva <= 45)) ? lcode[latva] : 0;
}

int centercode(int latva)
{ /*	Returns integer between 0 and 5 for centering type */
	static int		ccode[] = { 0,	/* code zero */
 4, 5, 1, 5, 3, 3, 3, 3,
 5, 2, 1, 1, 2, 2, 3, 4, 2,
 3, 3, 2, 1, 1, 2, 5, 2,
 4, 2, 2, 2, 2, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 1, 6 } ;

	return ((0 <= latva) && (latva <= 45)) ? ccode[latva] : 0;
}

short spgtosys(short spg)
{ /*	This returns the crystal system associated with the
	International Tables spacegroup number `spg'. */
	if ((spg <= 0) || (spg > 267)) return 1; /* nonsense-> triclinic*/
	if ((1 <= spg) && (spg <= 2)) return 1; 	/* triclinic */
	if ((3 <= spg) && (spg <= 15)) return 2;	/* monoclinic */
	if ((16 <= spg) && (spg <= 74)) return 3;	/* orthorhombic */
	if ((75 <= spg) && (spg <= 142)) return 4;	/* tetragonal */
	if ((143 <= spg) && (spg <= 194)) return 6;	/* hexag, trig */
	if ((194 <= spg) && (spg <= 230)) return 5;	/* cubic */
	return 7;	/* rhombohedral */
}

short spgtocen(short spg)
{ /*	This returns the centering code associated with the
	International Tables spacegroup number `spg'. */
	if ((spg <= 0) || (spg > 267)) return 1; /* nonsense -> primitive*/
	if ((5 == spg) || (8 == spg) || (9 == spg) || (12 == spg) ||
	 (15 == spg) || (20 == spg) || (21 == spg) || (35 == spg) ||
	 (36 == spg) || (37 == spg) || ((63 <= spg) && (spg <= 68)))
		return 2;	/* C-centered */
	if ((23 == spg) || (24 == spg) || ((44 <= spg) && (spg <= 46)) ||
	 ((71 <= spg) && (spg <= 74)) || (79 == spg) || (80 == spg) ||
	 (87 == spg) || (88 == spg) || (97 == spg) || (98 == spg) ||
	 ((107 <= spg) && (spg <= 110)) || ((119 <= spg) && (spg <= 122)) ||
	 ((139 <= spg) && (spg <= 142)) || (197 == spg) || (199 == spg) ||
	 (204 == spg) || (206 == spg) || (211 == spg) || (214 == spg) ||
	 (217 == spg) || (220 == spg) || (229 == spg) || (230 == spg))
		return 3;	/* I-centered */
	if ((22 == spg) || (42 == spg) || (43 == spg) || (69 == spg) ||
	 (69 == spg) || (70 == spg) || (196 == spg) || (202 == spg) ||
	 (203 == spg) || (209 == spg) || (210 == spg) || (216 == spg) ||
	 (219 == spg) || ((225 <= spg) && (spg <= 228)))
		return 4;	/* F-centered */
	if ((146 == spg) || (148 == spg) || (155 == spg) || (160 == spg) ||
	 (161 == spg) || (166 == spg) || (167 == spg)) return 5; /* H-cent */
	if ((246 == spg) || (248 == spg) || (255 == spg) || (260 == spg) ||
	 (261 == spg) || (266 == spg) || (267 == spg)) return 6; /* R-cent */
	if ((38 <= spg) && (spg <= 41)) return 7; /* A-centered */
	return 1; /* Primitive (P) */
}

char *glatfi(int latind)
{ /*	String describing lattice, directly from index */
	/* code values associated with the Bravais lattice calculations */
	static char		*latsym[] = { "undefined", "triclinic   ",
	 "monoclinic  ", "orthorhombic", "tetragonal  ", "cubic       ",
	 "hexagonal   ", "rhombohedral" } ;

	return latsym[latind];
}

char *glattice(int latva)
{	return glatfi(lattcode(latva)); /* returns string describing lattice */
}

char *glatt4(int latva)
{ /*	4-character version of lattice-type string */
	static char	locgla[14];

	zeromem(locgla, 13);	strcpy(locgla, glattice(latva));
	locgla[4] = '\0';	return locgla;
}

char *gcenfi(int cenind)
{ /*	String describing centering directly from index */
	static char	*center[] = { "U", "P", "C", "I", "F", "H", "R" };
	return center[cenind];
}

char *gcentering(int latva)
{	return gcenfi(centercode(latva)); /* string describing centering */
}

char *glatfi4(int latind)
{	/* 4-char string describing lattice, directly from index */
	static char	locgla[14];

	zeromem(locgla, 13);	strcpy(locgla, glatfi(latind));
	locgla[4] = '\0';	return locgla;
}

static double bravais(int charac, double are[3][3], double arout[3][3])
{ /*	This computes measures of the likelihood of a higher symmetry relative
	to the reduced cell given by are[3][3] according to the lattice
	character `charac' (1 - 44). The likelihood is returned as a
	floating-point value, and a negative value is returned in the event
	of an error.  The transformed unit cell matrix is returned in arout.
	Correction for #17 (monoclinic C type II, D=D*) inserted June '01 */
	int	i, j, k, u;
	int	type2, nattype2, doswitch, imincos, nneg, thcode;
	double	biga, bigb, bigc, bigd, bige, bigf, bigdef, defp;
	double	wnum, wden, mincos, nextcos, ddif0, ddif1, ddif2;
	double	tare[3][3], nma[3][3], conv[3][3], zcos[3];

 /* the comments on the left side of these lines indicate the value of
 charac followed by the value of u (since u == charac - 1) */
	static sichar nig[] = {
	/* A = B = C */
 /*  0/ 1 */	 1,-1, 1, 1, 1,-1,-1, 1, 1,	/* cubic F-centered */
 /*  1/ 2 */	 1,-1, 0,-1, 0, 1,-1,-1,-1,	/* rhombohedral, type I */
 /*  2/ 3 */	 1, 0, 0, 0, 1, 0, 0, 0, 1,	/* primitive cubic, type II */
 /*  3/ 4 */	 1,-1, 0,-1, 0, 1,-1,-1,-1,	/* rhombohedral, type II */
 /*  4/ 5 */	 1, 0, 1, 1, 1, 0, 0, 1, 1,	/* body-centered cubic,type II*/
 /*  5/ 6 */	 0, 1, 1, 1, 0, 1, 1, 1, 0,	/* body-centered tetragonal */
 /*  6/ 7 */	 1, 0, 1, 1, 1, 0, 0, 1, 1,	/* body-centered tetragonal */
 /*  7/ 8 */	-1,-1, 0,-1, 0,-1, 0,-1,-1,	/* body-centered orthorhombic*/

	/* A = B, no conditions on C */
 /*  8/ 9 */	 1, 0, 0,-1, 1, 0,-1,-1, 3,	/* rhombohedral, type I */
 /*  9/10 */	 1, 1, 0, 1,-1, 0, 0, 0,-1,	/* monoclinic Ccent type I */
 /* 10/11 */	 1, 0, 0, 0, 1, 0, 0, 0, 1,	/* tetragonal prim, type II */
 /* 11/12 */	 1, 0, 0, 0, 1, 0, 0, 0, 1,	/* hexagonal prim I+II */
 /* 12/13 */	 1, 1, 0,-1, 1, 0, 0, 0, 1,	/* orthorhombic Ccent I+II */
 /* 13/14 */	 1, 1, 0,-1, 1, 0, 0, 0, 1,	/* monoclinic Ccent type II*/
 /* 14/15 */	 1, 0, 0, 0, 1, 0, 1, 1, 2,	/* tetragronal Icent I+II */
 /* 15/16 */	-1,-1, 0, 1,-1, 0, 1, 1, 2,	/* orthorhomb Fcent type II*/
 /* 16/17 */	 1,-1, 0,-1,-1, 0,-1, 0,-1,	/* monoclinic Ccent type II*/

	/* B = C, no conditions on A */
 /* 17/18 */	 0,-1, 1, 1,-1,-1, 1, 0, 0,	/* tetragonal Icent type I */
 /* 18/19 */	-1, 0, 0, 0,-1, 1,-1, 1, 1,	/* orthorhomb Icent type I */
 /* 19/20 */	 0, 1, 1, 0, 1,-1,-1, 0, 0,	/* monoclinic Ccent type I */
 /* 20/21 */	 0, 1, 0, 0, 0, 1, 1, 0, 0,	/* tetragonal prim type II */
 /* 21/22 */	 0, 1, 0, 0, 0, 1, 1, 0, 0,	/* hexag. prim. both types */
 /* 22/23 */	 0, 1, 1, 0,-1, 1, 1, 0, 0,	/* orthorhomb Ccent type II*/
 /* 23/24 */	 1, 2, 1, 0,-1, 1, 1, 0, 0,	/* rhombohedral, type II */
 /* 24/25 */	 0, 1, 1, 0,-1, 1, 1, 0, 0,	/* monoclinic Ccent type II*/

	/* No conditions on A, B, C */
 /* 25/26 */	 1, 0, 0,-1, 2, 0,-1, 0, 2,	/* orthorhomb Fcent type I */
 /* 26/27 */	-1, 2, 0,-1, 0, 0, 0,-1, 1,	/* monoclinic Ccent type I */
 /* 27/28 */	-1, 0, 0,-1, 0, 2, 0, 1, 0,	/* monoclinic Ccent type I */
 /* 28/29 */	 1, 0, 0, 1,-2, 0, 0, 0,-1,	/* monoclinic Ccent type I */
 /* 29/30 */	 0, 1, 0, 0, 1,-2,-1, 0, 0,	/* monoclinic Ccent type I */
 /* 30/31 */	 1, 0, 0, 0, 1, 0, 0, 0, 1,	/* triclinic primitive type I*/
 /* 31/32 */	 1, 0, 0, 0, 1, 0, 0, 0, 1,	/* orthorhombic prim.type II*/
 /* 32/33 */	 1, 0, 0, 0, 1, 0, 0, 0, 1,	/* monoclinic prim type II */
 /* 33/34 */	-1, 0, 0, 0, 0,-1, 0,-1, 0,	/* monoclinic prim type II */
 /* 34/35 */	 0,-1, 0,-1, 0, 0, 0, 0,-1,	/* monoclinic prim type II */
 /* 35/36 */	 1, 0, 0,-1, 0,-2, 0, 1, 0,	/* orthorhombic C type II*/
 /* 36/37 */	 1, 0, 2, 1, 0, 0, 0, 1, 0,	/* monoclinic Ccent type II */
 /* 37/38 */	-1, 0, 0, 1, 2, 0, 0, 0,-1,	/* orthorhombic C type II*/
 /* 38/39 */	-1,-2, 0,-1, 0, 0, 0, 0,-1,	/* monoclinic Ccent type II */
 /* 39/40 */	 0,-1, 0, 0, 1, 2,-1, 0, 0,	/* orthorhombic Ccent type II*/
 /* 40/41 */	 0,-1,-2, 0,-1, 0,-1, 0, 0,	/* monoclinic Ccent type II*/
 /* 41/42 */	-1, 0, 0, 0,-1, 0, 1, 1, 2,	/* orthorhomb Icent type II*/
 /* 42/43 */	-1, 0, 0,-1,-1,-2, 0,-1, 0,	/* monoclinic Icent type II*/
 /* 43/44 */	 1, 0, 0, 0, 1, 0, 0, 0, 1	/* triclinic prim. type II*/
	} ;

	u = charac - 1;
	biga = bigb = bigc = bigd = bige = bigf = 0.;
	for (i = 0; i < 3; i++)
	   {
		biga += are[0][i] * are[0][i];
		bigb += are[1][i] * are[1][i];
		bigc += are[2][i] * are[2][i];
		bigd += are[1][i] * are[2][i];
		bige += are[2][i] * are[0][i];
		bigf += are[0][i] * are[1][i];
	   }
	bigdef = bigd + bige + bigf;	defp = bigd * bige * bigf;
	type2 = bigdef < 0.;
	wnum = 0.;	wden = 1.;	/* start weights at zero */
	if ((biga < DFUZZ) || (bigb < DFUZZ) || (bigc < DFUZZ))
		return -1.;
	/* install an initial penalty if we're looking for a
	 type I cell and we really have a type II or vice versa */
	if ((u < 2) || (8 <= u && u <= 9) || (17 <= u && u <= 19) ||
	 (25 <= u && u <= 30))
	  {
		if (type2) wnum += fabs(defp) / (biga * bigb * bigc);
	  }
	else {
		if (!type2) wnum += fabs(defp) / (biga * bigb * bigc);
	     }
	/* the penalties associated with distances will go as the
	 square of the differences */
	if (u < 8)	/* A == B == C */
	  {
		ddif0 = (biga - bigb) / (biga + bigb);
		ddif1 = (bigc - biga) / (bigc + biga);
		wnum += ddif0 * ddif0 + ddif1 * ddif1;	wden += 2.;
	  }
	else if (u < 17)
	  {
		ddif0 = (biga - bigb) / (biga + bigb);
		wnum += ddif0 * ddif0;	wden += 1.;
	  }
	else if (u < 25)
	  {
		ddif0 = (bigb - bigc) / (bigb + bigc);
		wnum += ddif0 * ddif0;	wden += 1.;
	  }
	/* let's try squaring the angular differences too */
	if (u == 0) /* Face-centered cubic, type I */
	  {
		ddif0 = (bigd - biga * 0.5) / biga;
		ddif1 = (bige - bigb * 0.5) / bigb;
		ddif2 = (bigf - bigc * 0.5) / bigc;
		wnum += ddif0 * ddif0 + ddif1 * ddif1 + ddif2 * ddif2;
		wden += 3.;
	  }
	else if (u == 1) /* rhombohedral, type I */
	  {
		ddif0 = (bigd - bige) / bigc;
		ddif1 = (bigd - bigf) / bigb;
		wnum += ddif0 * ddif0 + ddif1 * ddif1;
		wden += 2.;
	  }
	else if (u == 2) /* primitive cubic, type II */
	  {
		ddif0 = (bigd / bigb) * (bigd / bigc);
		ddif1 = (bige / bigc) * (bige / biga);
		ddif2 = (bigf / biga) * (bigf / bigb);
		wnum += ddif0 + ddif1 + ddif2;
		wden += 3.;
	  }
	else if (u == 3) /* rhombohedral, type II */
	  {
		ddif0 = (bigd - bige) / bigc;	ddif1 = (bigd - bigf) / bigb;
		wnum += ddif0 * ddif0 + ddif1 * ddif1;	wden += 2.;
	  }
	else if (u == 4) /* body-centered cubic */
	  {
		ddif0 = (bigd + biga / 3.) / biga;
		ddif1 = (bige + bigb / 3.) / bigb;
		ddif2 = (bigf + bigc / 3.) / bigc;
		wnum += ddif0 * ddif0 + ddif1 * ddif1 + ddif2 * ddif2;
		wden += 3.;
	  }
	else if (u == 5) /* body-centred tetragonal */
	  {
		ddif0 = (bigd - bige) / bigc;
		ddif1 = (2. * fabs(bigdef) - (biga + bigb)) / (biga + bigb);
		wnum += ddif0 * ddif0 + ddif1 * ddif1;	wden += 2.;
	  }
	else if (u == 6) /* body-centered tetragonal, different setting */
	  {
		ddif0 = (bige - bigf) / biga;
		ddif1 = (2. * fabs(bigdef) - (biga + bigb)) / (biga + bigb);
		wnum += ddif0 * ddif0 + ddif1 * ddif1;	wden += 2.;
	  }
	else if (u == 7)
	  {
		ddif0 = (2. * fabs(bigdef) - (biga + bigb)) / (biga + bigb);
		wnum += ddif0 * ddif0;	wden += 1.;
	  }
	else if (u == 8) /* rhombohedral, type I */
	  {
		ddif0 = (bigd - biga / 2.) / biga;
		ddif1 = (bige - bigb / 2.) / bigb;
		ddif2 = (bigf - biga / 2.) / biga;
		wnum += ddif0 * ddif0 + ddif1 * ddif1 + ddif2 * ddif2;
		wden += 2.;
	  }
	else if (u == 9) /* monoclinic C-centered, type I */
	  {
		ddif0 = (bigd - bige) / bigc;
		wnum += ddif0 * ddif0; wden += 1;
	  }
	else if (u == 10) /* tetragonal primitive, type II */
	  {
		ddif0 = (bigd / bigb) * (bigd / bigc);
		ddif1 = (bige / bigc) * (bige / biga);
		ddif2 = (bigf / biga) * (bigf / bigb);
		wnum += ddif0 + ddif1 + ddif2;
		wden += 3.;
	  }
	else if (u == 11) /* hexagonal primitive, type II (or I)*/
	  {
		ddif0 = (bigd / bigb) * (bigd / bigc);
		ddif1 = (bige / bigc) * (bige / biga);
		ddif2 = (fabs(bigf) - biga / 2.) / biga;
		wnum += ddif0 + ddif1 + ddif2 * ddif2;
		wden += 3.;
	  }
	else if (u == 12) /* orthorhombic C centered */
	  {
		ddif0 = (bigd / bigb) * (bigd / bigc);
		ddif1 = (bige / bigc) * (bige / biga);
		wnum += ddif0 + ddif1;	wden += 2.;
	  }
	else if (u == 13) /* monoclinic C centered, type II */
	  {
		ddif0 = (bigd - bige) / bigc;	wnum += ddif0 * ddif0;
		wden += 1.;
	  }
	else if (u == 14) /* tetragonal I-centered, type II (or I) */
	  {
		if (bigd > 0.)
		  {
			ddif0 = (bigd - biga / 2.) / biga;
			ddif1 = (bige - bigb / 2.) / bigb;
		  }
		 else {
			ddif0 = (bigd + biga / 2.) / biga;
			ddif1 = (bige + bigb / 2.) / bigb;
		  }
		ddif2 = (bigf / biga) * (bigf / bigb);
		wnum += ddif0 * ddif0 + ddif1 * ddif1 + ddif2;
		wden += 3;
	  }
	else if (u == 15) /* orthorhombic F-centered, type II */
	  {
		ddif0 = (bigd - bige) / bigc;
		ddif1 = (2. * fabs(bigdef) - (biga + bigb)) / (biga + bigb);
		wnum += ddif0 * ddif0 + ddif1 * ddif1;	wden += 2.;
	  }
	else if (u == 16) /* monoclinic C-centered, type II */
	  {
		ddif0 = (2. * fabs(bigdef) - (biga + bigb)) / (biga + bigb);
		wnum += ddif0 * ddif0;	wden += 1.;
	  }
	else if (u == 17) /* tetragonal I centered, type I */
	  {
		ddif0 = (bigd - biga / 4.) / biga;
		ddif1 = (bige - biga / 2.) / biga;
		ddif2 = (bigf - biga / 2.) / biga;
		wnum += ddif0 * ddif0 + ddif1 * ddif1 + ddif2 * ddif2;
		wden += 3.;
	  }
	else if (u == 18) /* orthorhombic I centered, type I */
	  {
		ddif0 = (bige - biga / 2.) / biga;
		ddif1 = (bigf - biga / 2.) / biga;
		wnum += ddif0 * ddif0 + ddif1 * ddif1;	wden += 2.;
	  }
	else if (u == 19) /* monoclinic C centered, type I */
	  {
		ddif0 = (bige - bigf) / biga;
		wnum += ddif0 * ddif0;	wden += 1.;
	  }
	else if (u == 20) /* tetragonal primitive, type II */
	  {
		ddif0 = (bigd / bigb) * (bigd / bigc);
		ddif1 = (bige / bigc) * (bige / biga);
		ddif2 = (bigf / biga) * (bigf / bigb);
		wnum += ddif0 + ddif1 + ddif2;
		wden += 3.;
	  }
	else if (u == 21) /* hexagonal primitive, type II (or I) */
	  {
		ddif0 = (fabs(bigd) - bigb / 2.) / bigb;
		ddif1 = (bige / bigc) * (bige / biga);
		ddif2 = (bigf / biga) * (bigf / bigb);
		wnum += ddif0 * ddif0 + ddif1 + ddif2;
		wden += 3.;
	  }
	else if (u == 22) /* orthorhombic C centered, type II */
	  {
		ddif0 = (bige / bigc) * (bige / biga);
		ddif1 = (bigf / biga) * (bigf / bigb);
		wnum += ddif0 + ddif1;	wden += 2.;
	  }
	else if (u == 23) /* rhombohedral, type II */
	  {
		if (bige < 0.)
		  {
			ddif0 = (bige + biga / 3.) / biga;
			ddif1 = (bigf + biga / 3.) / biga;
		  }
		 else {
			ddif0 = (bige - biga / 3.) / biga;
			ddif1 = (bigf - biga / 3.) / biga;
		  }
		ddif2 = (2. * fabs(bigdef) - (biga + bigb)) / (biga + bigb);
		wnum += ddif0 * ddif0 + ddif1 * ddif1 + ddif2 * ddif2;
		wden += 3.;
	  }
	else if (u == 24) /* monoclinic C-centered, type II */
	  {
		ddif0 = (bige - bigf) / biga; wnum += ddif0 * ddif0; wden += 1.;
	  }
	else if (u == 25) /* orthorhombic F-centered, type I */
	  {
		ddif0 = (bigd - biga / 4.) / biga;
		ddif1 = (bige - biga / 2.) / biga;
		ddif2 = (bigf - biga / 2.) / biga;
		wnum += ddif0 * ddif0 + ddif1 * ddif1 + ddif2 * ddif2;
		wden += 3.;
	  }
	else if (u == 26) /* monoclinic C-centered, type I */
	  {
		if (bige > 0.)
		  {
			ddif0 = (bige - biga / 2.) / biga;
			ddif1 = (bigf - biga / 2.) / biga;
		  }
		 else {
			ddif0 = (bige + biga / 2.) / biga;
			ddif1 = (bigf + biga / 2.) / biga;
		  }
		wnum += ddif0 * ddif0 + ddif1 * ddif1;	wden += 2.;
	  }
	else if (u == 27) /* monoclinic C-centered, type I */
	  {
		if (bige > 0.)
		  {
			ddif0 = (bige - biga / 2.) / bigc;
			ddif1 = (bigf - bigd * 2.) / bigb;
		  }
	  	else {
			ddif0 = (bige + biga / 2.) / bigc;
			ddif1 = (bigf + bigd * 2.) / bigb;
		  }
		wnum += ddif0 * ddif0 + ddif1 * ddif1;	wden += 2.;
	  }
	else if (u == 28) /* monoclinic C-centered, type I */
	  {
		if (bige > 0.)
		  {
			ddif0 = (bige - bigd * 2) / biga;
			ddif1 = (bigf - biga / 2) / biga;
		  }
		else  {
			ddif0 = (bige + bigd * 2) / biga;
			ddif1 = (bigf + biga / 2) / biga;
		  }
		wnum += ddif0 * ddif0 + ddif1 * ddif1;	wden += 2.;
	  }
	else if (u == 29) /* monoclinic C-centered, type I */
	  {
		if (bigd > 0.)
		  {
			ddif0 = (bigd - bigb / 2.) / bigc;
			ddif1 = (bigf - bige * 2.) / biga;
		  }
		else {
			ddif0 = (bigd + bigb / 2.) / bigc;
			ddif1 = (bigf + bige * 2.) / biga;
		  }
		wnum += ddif0 * ddif0 + ddif1 * ddif1;	wden += 2.; 
	  }
	else if (u == 30) /* triclinic primitive, type I */
	  {
		wnum += 0.;	wden += 0.;
	  }
	else if (u == 31) /* orthorhombic primitive, type II */
	  {
		ddif0 = (bigd / bigb) * (bigd / bigc);
		ddif1 = (bige / bigc) * (bige / biga);
		ddif2 = (bigf / biga) * (bigf / bigb);
		wnum += ddif0 + ddif1 + ddif2;
		wden += 3.;
	  }
	else if (u == 32) /* monoclinic primitive, type II */
	  {
		ddif0 = (bigd / bigb) * (bigd / bigc);
		ddif1 = (bigf / biga) * (bigf / bigb);
		wnum += ddif0 + ddif1;	wden += 2.;
	  }
	else if (u == 33) /* monoclinic primitive, type II */
	  {
		ddif0 = (bigd / bigb) * (bigd / bigc);
		ddif1 = (bige / bigc) * (bige / biga);
		wnum += ddif0 + ddif1;	wden += 2.;
	  }
	else if (u == 34) /* monoclinic primitive, type II */
	  {
		ddif0 = (bige / bigc) * (bige / biga);
		ddif1 = (bigf / biga) * (bigf / bigb);
		wnum += ddif0 + ddif1;	wden += 2.;
	  }
	else if (u == 35) /* orthorhombic C-centered, type II (or I) */
	  {
		ddif0 = (bigd / bigb) * (bigd / bigc);
		ddif1 = (bigf / biga) * (bigf / bigb);
		ddif2 = (fabs(bige)  - biga / 2.) / bigb;
		wnum += ddif0 + ddif1 + ddif2 * ddif2;	wden += 3.;
	  }
	else if (u == 36) /* monoclinic C-centered, type II */
	  {
		ddif0 = fabs(bige) - biga / 2.;
		ddif0 = (ddif0 / biga) * (ddif0 / bigc);
		ddif1 = (bigf / biga) * (bigf / bigb);
		wnum += ddif0 + ddif1;	wden += 2.;
	  }
	else if (u == 37) /* orthorhombic C-centered, type II (or I) */
	  {
		ddif0 = (bigd / bigb) * (bigd / bigc);
		ddif1 = (bige / bigc) * (bige / biga);
		ddif2 = fabs(bigf) - biga / 2.;
		ddif2 = (ddif2 / biga) * (ddif2 / bigb);
		wnum += ddif0 + ddif1 + ddif2;	wden += 3.;
	  }
	else if (u == 38) /* monoclinic C-centered, type II */
	  {
		ddif0 = (bige / bigc) * (bige / biga);
		ddif1 = fabs(bigf) - biga / 2.;
		ddif1 = (ddif1 / biga) * (ddif1 / bigb);
		wnum += ddif0 + ddif1;	wden += 2.;
	  }
	else if (u == 39) /* orthorhombic C-centered, type II */
	  {
		ddif0 = fabs(bigd) - bigb / 2.;
		ddif0 = (ddif0 / bigb) * (ddif0 / bigc);
		ddif1 = (bige / bigc) * (bige / biga);
		ddif2 = (bigf / biga) * (bigf / bigb);
		wnum += ddif0 + ddif1 + ddif2;	wden += 3.;
	  }
	else if (u == 40) /* monoclinic C-centered, type II */
	  {
		ddif0 = fabs(bigd) - bigb / 2.;
		ddif0 = (ddif0 / bigb) * (ddif0 / bigc);
		ddif1 = (bigf / biga) * (bigf / bigb);
		wnum += ddif0 + ddif1;	wden += 2.;
	  }
	else if (u == 41) /* orthorhombic I-centered, type II */
	  {
		ddif0 = fabs(bigd) - bigb / 2;
		ddif0 = (ddif0 / bigb) * (ddif0 / bigc);
		ddif1 = fabs(bige) - biga / 2.;
		ddif1 = (ddif1 / bigc) * (ddif1 / biga);
		ddif2 = (bigf / biga) * (bigf / bigb);
		wnum += ddif0 + ddif1 + ddif2;	wden += 3.;
	  }
	else if (u == 42) /* monoclinic I-centered, type II */
	  {
		ddif0 = (2. * fabs(bigdef) - (biga + bigb)) / (biga + bigb);
		ddif1 = (fabs(2. * bigd + bigf) - bigb)/ bigb;
		wnum += ddif0 * ddif0 + ddif1 * ddif1;
		wden += 2.;
	  }
	else if (u == 43) /* triclinic, type II */
	  {
		wnum += 0.;	wden += 0.;
	  }
	if (wden > DFUZZ) wnum /= wden;
	/* before applying the conversion we make sure that we're
	 actually dealing with the right type (I or II) matrix.
	 There are some odd cases where we're not, and when that
	 happens we need to internally transfer first. */
	nattype2 = ((2 <= u) && (u <= 7)) || ((10 <= u) && (u <= 16)) ||
	 ((20 <= u) && (u <= 24)) || ((31 <= u) && (u <= 43));
	doswitch = (type2) ? 1 - nattype2 : nattype2;
	if (doswitch)	/* rotate 180 about the angle closest to 90 first */
	  {
		mincos = (bigd / bigb) * (bigd / bigc);	imincos = 0;
		nextcos = (bige / bigc) * (bige / biga);
		if (nextcos < mincos)
		  {
			imincos = 1;	mincos = nextcos;
		  }
		nextcos = (bigf / biga) * (bigf / bigb);
		if (nextcos < mincos) imincos = 2;
		for (i = 0; i < 3; i++)
		   {
			j = i + 1;	if (j > 2) j -= 3;
			k = 3 - i - j;
			conv[i][j] = conv[i][k] = 0.;
			if (i == imincos) conv[i][i] = 1.;
			else conv[i][i] = -1.;
		   }
		dmm3(tare, conv, are);
	  }
	else for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		tare[i][j] = are[i][j];
	/* now construct the appropriate Niggli matrix.
	 Note that we've set this huge matrix up now as a linear vector */
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		conv[i][j] = nig[9 * u + 3 * i + j];
	dmm3(nma, conv, tare);	/* apply the Niggli matrix */
	/* under somewhat weird conditions this one may produce
	 a mixed-type (not type-I or type-II) matrix. We fix that.*/
	biga = bigb = bigc = bigd = bige = bigf = 0.;
	for (i = 0; i < 3; i++)
	   {
		biga += nma[0][i] * nma[0][i];
		bigb += nma[1][i] * nma[1][i];
		bigc += nma[2][i] * nma[2][i];
		bigd += nma[1][i] * nma[2][i];
		bige += nma[2][i] * nma[0][i];
		bigf += nma[0][i] * nma[1][i];
	   }
	zcos[0] = (bigd / bigb) * (bigd / bigc);
	zcos[1] = (bige / bigc) * (bige / biga);
	zcos[2] = (bigf / biga) * (bigf / bigb);
	/* count how many obtuse angles there are */
	nneg = (bigd < 0.) + (bige < 0.) + (bigf < 0.);
	/* if 1 or 2 obtuse angles, switch some */
	if ((nneg == 1) || (nneg == 2))
	  {
		if (nneg == 1)
		  {
			if (bigd < 0.) imincos = 0;
			else if (bige < 0.) imincos = 1;
			else if (bigf < 0.) imincos = 2;
			else imincos = 3; /* shouldn't happen */
			/* if we're trying for monoclinic or
			 hexagonal symmetry and we have two
			 acute angles, both close to 90, the
			 cell shouldn't be corrected. */
			thcode = lattcode(charac);
			if ((imincos < 3) && ((thcode == 2) || (thcode >= 6)))
			  {
				j = imincos + 1; if (j > 2) j -= 3;
				k = 3 - j - imincos; mincos = zcos[j];
				if (zcos[k] > mincos) mincos = zcos[k];
				if (mincos < ALMOSTZERO) imincos = 4;
			  }
		  }
		else {
			if (bigd > 0.) imincos = 0;
			else if (bige > 0.) imincos = 1;
			else if (bigf > 0.) imincos = 2;
			else imincos = 3; /* shouldn't happen */
			/* if we're trying for monoclinic or
			 hexagonal symmetry and we have two
			 obtuse angles, one close to 90, the
			 cell shouldn't be corrected. */
			thcode = lattcode(charac);
			if ((imincos < 3) && ((thcode == 2) || (thcode >= 6)))
			  {
				j = imincos + 1; if (j > 2) j -= 3;
				k = 3 - j - imincos; mincos = zcos[j];
				if (zcos[k] < mincos) mincos = zcos[k];
				if (mincos < ALMOSTZERO) imincos = 4;
			  }
		     }
		/* do the flip if appropriate! */
		if (imincos < 3)
		  {
			for (i = 0; i < 3; i++)
			   {
				j = i + 1; if (j > 2) j -= 3; k = 3 - i - j;
				conv[i][j] = conv[i][k] = 0.;
				conv[i][i] = (i == imincos) ? 1. : -1;
			   }
			dmm3(tare, conv, nma);
			for (i = 0; i < 3; i++)
				for (j = 0; j < 3; j++) nma[i][j] = tare[i][j];
		  }
	  }
	/* special case for C-centered monoclinic:
	 it's possible to end up needing an additional transform,
	 a' = -a, b' = -b, c' = a+c. This becomes useful if the
	 beta angle for this version is closer to 90 than for the
	 original. */
	if ((u == 9) || (u == 13) || (u == 16) || (u == 19) ||
	 (u == 24))
	  {
		biga = bigb = bigc = bigd = bige = bigf = 0;
		/* set up matrix to create alternate indexing */
		conv[0][0] = conv[1][1] = -1.;
		conv[2][0] = conv[2][2] = 1.;
		conv[0][1] = conv[0][2] = conv[1][0] = 0.;
		conv[1][2] = conv[2][1] = 0.;
		/* create a'=-a,b'=-b,c'=a+c matrix */
		dmm3(tare, conv, nma);
		/* compute dot products c.a, a.a, c.c
		 for original and converted systems */
		for (i = 0; i < 3; i++)
		   {	
			biga += nma[2][i] * nma[0][i];
		 	bigb += nma[0][i] * nma[0][i];
			bigc += nma[2][i] * nma[2][i];
			bigd += tare[2][i] * tare[0][i];
		 	bige += tare[0][i] * tare[0][i];
			bigf += tare[2][i] * tare[2][i];
		   }
		/* if (c.a)*(c.a)/(c.c)*(a.a) is smaller in
		 the converted system than in the original,
		 use the converted system */
		if ((bigb > DFUZZ) && (bigc > DFUZZ))
			biga = (biga / bigb) * (biga / bigc);
		else	biga = 0;
		if ((bige > DFUZZ) && (bigf > DFUZZ))
			bigd = (bigd / bige) * (bigd / bigf);
		else	bigd = 0;
		if (bigd < biga)
		  {
			for (i = 0; i < 3; i++)
			   for (j = 0; j < 3; j++) nma[i][j] = tare[i][j];
		  }
	  }
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		arout[i][j] = nma[i][j];
	return wnum;
}

static double sympen(int charac)
{ /*	This imposes a penalty on the figure of merit associated with a
	given lattice character.
	The values are pretty arbitrary and are likely to change. */
	static int sympe[] =	{
		11, 0, 5, 1, 5, 0, 1, 1, 3, 5, 7, 2, 2, 3, 7, /*  0 - 14 */
		 1, 3, 7, 1, 3, 7, 2, 2, 3, 5, 7, 3, 7, 7, 7, /* 15 - 29 */
		 7,11, 4, 9, 9, 9, 3, 7, 3, 7, 3, 7, 3, 7,11, 5 } ;
	return ((charac < 0) || (charac > 45)) ? 0.011 :
		0.001 * (double)sympe[charac];
}

LATTICESCORE *managelca(int opno)
{ /*	This manages memory for the LATTICESCORE operations */
	static LATTICESCORE	*lca;
	static char		lcamsg[] = "lca in managelca";

	if (!opno) /* eliminate memory for the lattice score */
	  {
		memzap((AVOID **)&lca, lcamsg);	return NULL;
	  }
	if (opno == 1) lca = managelca(0);	/* force recalc */
	if (NULL == lca) lca = callop(48, sizeof (LATTICESCORE), lcamsg);
	return lca;
}

static void orthoc_fudge(LATTICESCORE *lc0)
{ /*	This adjusts the score for the C-centered orthorhombic answers
	 and C-centered monoclinic answers if the hexagonal primitives
	 have similar scores and have roughly half the cell volume. */
	int		i, chara;
	LATTICESCORE	*lcap, *lc_hexp, *lc_orthoc, *lc_monoc;

	lc_orthoc = lc_monoc = lc_hexp = NULL;
	for (lcap = lc0, i = 0; i < 45; i++, lcap++)
	   if (lcap->lc_choice >= 0)
	     {
		chara = lcap->lc_latchar;
		if ((lc_orthoc == NULL) && (lattcode(chara) == 3)
		 && (centercode(chara) == 2)) lc_orthoc = lcap;
		if ((lc_monoc == NULL) && (lattcode(chara) == 2)
		 && (centercode(chara) == 2)) lc_monoc = lcap;
		else if ((lc_hexp == NULL) && (lattcode(chara) == 6)
		 && (centercode(chara) == 1)) lc_hexp = lcap;
	     }
	if (NULL == lc_hexp) return;
	if ((NULL != lc_orthoc) && ((lc_orthoc->lc_fom < lc_hexp->lc_fom) &&
	 (lc_orthoc->lc_fom + ORTHO_COMPARE) > lc_hexp->lc_fom))
		lc_orthoc->lc_fom = lc_hexp->lc_fom + ORTHO_OFFSET;
	if ((NULL != lc_monoc) && ((lc_monoc->lc_fom < lc_hexp->lc_fom) &&
	 (lc_monoc->lc_fom + ORTHO_COMPARE) > lc_hexp->lc_fom))
		lc_monoc->lc_fom = lc_hexp->lc_fom + ORTHO_OFFSET;
}

static int rhombohex(int indx)
{ /* This adds in the rhombohedral indexing of a rhombohedral cell */

	int			j, k;
	float			rhcel[3][3];
	LATTICESCORE		*lcap, *lcaq;
	static LATTICESCORE	*lca;
	static float	rhconv[3][3] = { { R3,-R3,-R3}, {-T3,-R3,-R3},
					 { R3, T3,-R3 } } ;

	if (NULL == (lca = managelca(3))) return 0; /* otherwise, calc! */
	lcap = lca + indx;
	mm3(rhcel, rhconv, lcap->lc_cel); /* convert to rhombohedral */
	lcaq = lca + 44; lcaq->lc_latchar = 45;
	lcaq->lc_score = lcap->lc_score;	lcaq->lc_fom = lcap->lc_fom;
	for (j = 0; j < 3; j++) for (k = 0; k < 3; k++)
		lcaq->lc_cel[j][k] = rhcel[j][k];
	return 1;
}

LATTICESCORE *allbravais(double lareal[3][3], int *nlattp)
{ /*	This computes the 45 crystal lattice scores associated with the
	current unit cell matrix. These are the 44 listed in table 9.3.1
	in the International Tables plus an extra (#45) that is associated
	with a rhombohedral indexing of a rhombohedral cell.
	It finds the best lattice of each of the sixteen types
	(cF, hR, cP, cI, tI, oI, mC, tP, hP, oC, oF, tP, aP, mP, mI, plus
	the rhombohdral indexing of the rhombohedral cell) and sets the
	appropriate flag in the LATTICESCORE structure associated with that
	lattice. The routine returns the number of successfully located
	structures in *nlattp; this number ought to be 16 or 17 if things
	work right. It returns a pointer to the LATTICESCORE structures. */
	register int	i, j;
	int		chara, charb, foundit, listno;
	double		resu, larout[3][3];
	LATTICESCORE	*lcap, *lcaq;
	static LATTICESCORE	*lca;

	*nlattp = 0;	/* we haven't found any yet */
	/* if we call this with the first argument NULL, then we kill lca */
	if (NULL == lareal) return managelca(0);
	if (NULL == (lca = managelca(3))) return NULL; /* otherwise, calc! */
	/* compute the relevant scores and unit cells */
	for (chara = 1, lcap = lca; chara <= 44; chara++, lcap++)
	   {
		resu = bravais(chara, lareal, larout);
		lcap->lc_latchar = lcap->lc_choice = chara;
		lcap->lc_score = resu;
		lcap->lc_fom = 100. * resu + sympen(chara);
		for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
			lcap->lc_cel[i][j] = larout[i][j];
	   }
	/* sort the results in increasing order of score */
	qsort(lca, 44, sizeof (LATTICESCORE), (QPE)lcacmp);
	/* fudge if oC and hP are close to equal */
	orthoc_fudge(lca);
	/* determine the best of any given symmetry */
	for (listno = i = 0, lcap = lca; i < 44; i++, lcap++)
	   {
		chara = lcap->lc_latchar;	foundit = 0;
		if (i)
		  {
		   for (j = 0, lcaq = lca; j < i; j++, lcaq++)
		     {
			charb = lcaq->lc_latchar;
			if ((lattcode(chara) == lattcode(charb)) &&
			 (centercode(chara) == centercode(charb)))
			  {
				foundit = 1;	lcap->lc_choice = -1; break;
			  }
		     }
		  }
		if (!foundit)
		  {
			lcap->lc_choice = listno;	listno++;
			/* add in rhombohedral cells indexed rhombohedrally */
			if ((lattcode(chara) == 7) && (rhombohex(i)))
				(lca + 44)->lc_choice = listno++;
		  }
	   }
	/* sort the results again in increasing order of score to incorporate
		the second rhombohedral choice */
	qsort(lca, 45, sizeof (LATTICESCORE), (QPE)lcacmp);
	*nlattp = listno;	return lca;
}

void show_bravchoice(FILE *fp, float *stang, LATTICESCORE *lcap)
{ /*	This prints information about lattice choice lcap to stream fp */
	int	listno, chara;
	float	cel[9];

	listno = lcap->lc_choice;	chara = lcap->lc_latchar;
	artocel(stang, lcap->lc_cel, cel);
	fprintf(fp, "  %c %4d %7.3f %6.3f%7.2f%7.2f%7.2f%6.1f%6.1f%6.1f",
	 (char)(listno + 'a'), (int)(lcap->lc_latchar), 100. * lcap->lc_score,
	 lcap->lc_fom, cel[0], cel[1], cel[2], cel[3], cel[4], cel[5]);
	if ((fp != stdout) && (fp != stderr))
		fprintf(fp, " %7.2f %7.2f %7.2f", cel[6], cel[7], cel[8]);
	fprintf(fp, " %s %s\n", glattice(chara), gcentering(chara));
}

void headbravais(FILE *fp)
{ /*	This creates an appropriate header for Bravais lattice calculations */
	fprintf(fp, "Results of calculations using the 44 characters:\n");
	fprintf(fp,
 "in- latt  resid.   fig. cell lengths,Angstr.  cell angles,deg.");
	if ((fp != stdout) && (fp != stderr)) fprintf(fp,
 "   Euler angles, degrees");
	fprintf(fp, " lattice     cen-\n");
	fprintf(fp,
 "dex char   score  merit      a      b      c alpha  beta gamma");
	if ((fp != stdout) && (fp != stderr)) fprintf(fp,
 "   omega     chi     phi");
	fprintf(fp, " symmetry    ter.\n");
}

void showbravais(FILE *fp, float remat[3][3], float *stang)
{ /*	This displays (on stream fp) the Bravais lattice calculations
	for standalone refinements. */
	int			i, j;
	double			lom[3][3];
	LATTICESCORE		*lcap;
	static LATTICESCORE	*lca;

	headbravais(fp);
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++) lom[i][j] =remat[i][j];
	if (NULL == (lca = managelca(3))) lca = allbravais(lom, &i);
	for (i = 0, lcap = lca; i < 45; i++, lcap++)
	   if (lcap->lc_choice >= 0) show_bravchoice(fp, stang, lcap);
}

/* end of xgbravais.c */
