/* xgenfile.c -- file-manipulation routines for X-GEN

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions; reorganize use of "{"
	24 Mar 2007: update how xgvariable gets called and used
	 4 Mar 2007: add xgvariable functionality for ascii information
	20 Oct 2006: handle case where libz is unavailable
	 5 Sep 2006: expandit(): readv->readva; readv is a global routine
	24 Jun 2006: include xgenpred.h,xgmodel.h,xgpostint.h;
		change return codes from shrinkit from FRAME_* to E*
	11 Jun 2006: move gotspace from here to xgenfile so it can be static
	 1 Jun 2006: fullname,install_names moved to ../utilities/xg_now.c;
		mygetwd moved here
	 3 Jan 2006: memory-management improved in expandit()
	30 Nov 2005: inline argument declarations and strict prototyping
	 8 Nov 2005: eliminate static globals; replace currpname, kill_pname,
		hide_oldpname, restore_oldpname, assign_pname with xgprogname()
	27 Oct 2005: move xgve[] inside xgjournal(); merge str2xgbuf()
		and flush_xgbuf() into one routine, xgbuf().
	27 Aug 2005: NULL-terminate the xgve[] list
	23 May 2005: xgjournal(): allow '_' as a character in a command name
	17 May 2005: trans_envname() no longer static;
		it's also somewhat rewritten to use less memory
	15 May 2005: xgjournal prepends xgen or xgentools to command name;
		it returns EHELP if called with --help or --version
	28 Apr 2005: don't scribble in the journal with --help and --version
	 2 Feb 2005: use "memzap((AVOID **)" throughout;
		correct memzap error in squeezeall() near beginning;
		allocate 8 extra bytes for "out" in squeezeall()
	18 Dec 2004: handle --help and --version in xgjournal()
	20 Oct 2004: squeezeall now available externally
	28 Aug 2004: routines used only here are declared as static; cosmetics;
		routine ashorten() removed because it wasn't being used
	14 Aug 2004: propen() and try_reindex() moved here from elsewhere
	 7 Jun 2004: add in routines for expanding and contracting filenames
		based on environment variables, /.. constructs, etc.
	 4 Jun 2004: in install_names(), handle relative pathnames correctly
	 1 Jun 2004: correct spelling in comment in str2xgbuf()
	29 Mar 2004: routine journalname() added so we can get the journal name
		from external routines
	18 Feb 2004: change kill_ftab calls so that they don't have args
	 9 Feb 2004: correct bug in memzap in assign_pname
	21 Jan 2004: license message inserted
	 7 Jan 2004: match_inode() takes FRAMEINDEX arg
	 6 Jan 2004: gotspace() routine written to pull spacegroup & system
		out of thin air
	 4 Jan 2004: globals replaced with args to routines;
		routines assign_pname,restore_oldpname,hide_oldpname,
		kill_pname, currpname written to deal with application names
	13 Dec 2003: return values from shrinkit() include FRAME_BADLENGTH;
		eliminate unused variable `nout' in shrinkit();
		eliminate unused variable `nnonz' in comparemid()
	 8 Nov 2003: wasat is an off_t, not a size_t
	 4 Nov 2003: comparemid() and retcompare() written
		and used for cases where byte-order is ambiguous
	20 Sep 2003: extra declarations to cover a LITTLEND case
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Sep 2003: call xgaccess() from xgopen() and propen for files we're
		reading or appending to; thus, we can read compressed files
	 4 Sep 2003: add xgaccess(); this allows us to
		uncompress X-GEN processing files (or anything else) in place
	18 Aug 2003: created from xgenio.c
 Relevant changes to xgenio.c before that:
	12 Jun 2003: correct trivial bug in `compressiontype'
	16 Jan 2003: more careful error-checking in expandit()
	 5 Jan 2003: reorganized source-file header
	30 Dec 2002: compressed (.Z and .gz) files correctly handled;
		routines compressiontype(), expandit(), shrinkit() here
	25 Mar 2002: removed MSI attachments
	26 Oct 1995: rewrite xgopen so link(oldname,newname) is part of test;
		fix trivial error in call to lswawb
	13 Oct 1995: give lswawb a 2nd arg to handle both directions correctly
	28 Aug 1995: xgwrite on AXP: byte- and word-swap on a temp copy of
		each structure, so we don't mess up the original structure
	12 May 1995: add match_inode() routine;
		handle totalframes as external global
	 2 May 1995: set lvp->v_cpt = NULL in xgread
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
		input corrected for reading old-style USPLINEs
	 1 Dec 1994: don't do byteswaps & float-conversions in VMS
	16 Nov 1994: xgopen(name,"w"):
		when it fails, try to create directory first
	18 Oct 1994: bug in `xgopen' fixed:
		now version numbers > 9 will work correctly
	 5 Oct 1994: new routines: xgread,xgwrite,deccvt for binary I/O
	30 Sep 1994: `shorten' returns # of chars saved by abbreviation
		`allshorten' routine added to shorten names in systematic way
	 5 May 1994: initial version of xgenio.c, cobbled from other files
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version of older files

   Purpose: dealing with file names and file I/O in X-GEN
   Cross references: xgenstr.c, xgenmisc.c, setupio.c

*************************************************************************/
#define		NEED_SSDEF	1
#define		NEED_DESCRIP	1
#define		NEED_STAT	1
#define		NEED_ERRNO	1
#define		NEED_TIME	1
#define		NEED_UNISTD	1
#define		XG_OSPLINE	14
#define		BUFBLOCK	32768
#define		GZBUFSIZ	65536
#include	"xgenpred.h"
#include	"xgmodel.h"
#include	"xgpostint.h"
#ifdef ZLIB
#include	"zlib.h"
#endif

extern int system(const char *str);

void mygetwd(char *bufr)
{	/* wrapper for either `getcwd', depending on system.
	At this point I've simplified it from previous versions; 
	Let's hope it actually works this way */

	(void)getcwd(bufr, 255);
}

static char *prepend_cwd(char *str)
{ /* This finds the current working directory name and sticks it in front of
   the string `str'. Error checking is done externally, mostly. This routine
   allocates memory for its output if it's actually needed but not otherwise,
   so the caller should only free that memory if that's appropriate. */
	int	lencwd, lenin, lentot;
	char	*cpi, *cpo, *out, *cwdbuf;

	/* we assume that if the file name begins with '/' it's already ok */
	if ((NULL != str) && ('/' == *str)) return str;
	/* reserve a healthy memory buffer for this directory name */
	if (NULL == (cwdbuf = mallop(256,"cwdbuf in prepend_cwd"))) return str;
	/* blank it for safety and call `mygetwd' to fill it */
	zeromem((void *)cwdbuf, 256);	mygetwd(cwdbuf);
	if ((NULL == cwdbuf) || ('/' != *cwdbuf))
	  {
		memzap((AVOID **)&cwdbuf, "cwdbuf in prepend_cwd");
		return NULL; /* error if it doesn't begin with / */
	  }
	/* measure length of `str' and the directory name */
	for (lenin = 0, cpi = str; '\0' != *cpi; cpi++) lenin++;
	for (lencwd = 0, cpi = cwdbuf; '\0' != *cpi; cpi++) lencwd++;
	lentot = lenin + lencwd + 2; /* max length of overall string */
	if (NULL == (out = mallop(lentot, "out in prepend_cwd"))) return NULL;
	zeromem((AVOID *)out, (size_t)lentot);
	/* output string begins with directory name */
	for (cpo = out, cpi = cwdbuf; '\0' != *cpi; cpi++) *cpo++ = *cpi;
	/* now we're done with the directory name; get rid of its memory */
	memzap((AVOID **)&cwdbuf, "cwdbuf in prepend_cwd");
	/* append a slash to this name if there isn't one there */
	if ('/' != *cpo) *cpo++ = '/';
	/* now append the contents of `str', which WON'T begin with "/" */
	for (cpi = str; '\0' != *cpi; cpi++) *cpo++ = *cpi;
	/* explicitly fill up the rest of the string with nulls if needed */
	for (*cpo = '\0'; cpo < out + lentot; cpo++) *cpo = '\0';
	return out;
}

int shorten(char *myname, char *logdir)
{ /* This converts, in place, a string like /home/fred/cyctnat/cyctcnat.xlg
    to  $WORK/cytcnat.xlg
    assuming the environment variable WORK is set to /home/fred/cyctnat
   Under VMS it converts "DKA100:[USERS.XENGEN.CYCTNAT]CYCTNAT.XLG"
	to WORK:CYCTNAT.XLG if WORK is DKA100:[USERS.XENGEN.CYCTNAT]
   It returns number of characters saved if shortening succeeds, 0 otherwise */
	char	*cpi, *valw, *ocp, *icp;
	size_t	lenw, lenn, leno, lenlogd;

	if ((NULL == (valw = getenv(logdir))) || ('\0' == *valw)) return 0;
	/* measure lengths of translated string, original string, logdir */
	for (cpi = valw, lenw = 0; '\0' != *cpi; cpi++) lenw++;
	for (cpi = myname, leno = 0; '\0' != *cpi; cpi++) leno++;
	for (cpi = logdir, lenlogd = 0; '\0' != *cpi; cpi++) lenlogd++;
		/* make sure the translation will actually save space */
	if ((lenw < lenlogd + 1) || (lenw >= leno + 1)) return 0;
	if (!strncmp(myname, valw, lenw))	/* do they match? */
	  {
#ifdef VMS		/* start with logical name, put colon in between */
		for (ocp = myname, icp =logdir; '\0'!= *icp; ) *ocp++ = *icp++;
		*ocp++ = ':';
#else
		*myname = '$';	/* abbreviated name starts with '$' */
		for (ocp = myname+1, icp =logdir;'\0'!=*icp; ) *ocp++ = *icp++;
#endif
		for (icp = myname + lenw; '\0' !=*icp; ) *ocp++ = *icp++;
		*ocp = '\0';	/* null-terminate it! */
		lenn = ocp - myname;	return (lenn < leno) ? leno - lenn : 0;
	  }
	return 0;	/* failure: no shortening */
}

char *squeezeall(char *str)
{ /* This squeezes out constructions of the form "/./", "//", and "/../"
   out of the input string. The first two are simply converted to "/";
   the last one requires backing up to the previous directory name and
   then moving on */

	int		lent;
	char		*cpi, *cpo, *cpie, *cpe, *prevslash;
	static char	*mid, *out;

	if ((NULL == str) || ('\0' == *str)) return NULL;
	out = cpo = NULL;
	mid = prepend_cwd(str);	/* prepend the current working directory */
	if ('/' != *mid)
	  {
		if (mid != str) memzap((AVOID **)&mid, "mid in squeezeall");
		return str; /* don't know how to fix this */
	  }
	for (cpi = mid, lent = 0; '\0' != *cpi; cpi++) lent++;
	if (lent < 2) return mid; /* nothing to squeeze */
	if (NULL == (out = mallop(lent + 8, "out in squeezeall"))) return mid;
	cpe = out + lent;	cpie = cpi;
	/* find instances of "/../", "/./", and "//" and convert them */
	for (cpi = mid, prevslash = cpo = out; '\0' != *cpi; cpi++)
	   {
		if (*cpi == '/') /* all special cases begin with slash */
		  {
			if ((cpi <= cpie - 3) && (*(cpi+1) == '.') &&
			 (*(cpi+2) == '.') &&
			 ((*(cpi + 3) == '/') || (*(cpi + 3) == '\0')))
			  { /* rewind back to the previous directory */
				cpo = prevslash;	cpi += 2;
				if (cpo <= out)
				  {
					*cpo++ = '/'; *cpo = '\0';
				  }
				/* remember location of previous slash */
				for (prevslash--; prevslash >= out;
				 prevslash--) if ('/' == *prevslash) break;
			  }
			else if ((cpi <= cpie - 2) && (*(cpi + 1) == '.')
			 && ((*(cpi + 2) == '/') || (*(cpi + 2) == '\0')))
				cpi++;	/* /./ */
			/* we don't do anything with "//" or "/\0' */
			else if ((*(cpi + 1) != '/') && (*(cpi+1) != '\0'))
			  {
				prevslash = cpo;	*cpo++ = *cpi;
			  }
		  }
		else *cpo++ = *cpi;	/* anything else */
	   }
	*cpo++ = '\0';
	/* humm. If output is the same length as the input,
	 we haven't actually accomplished anything, so we return the input */
	if (cpo > cpe)
	  {
		memzap((AVOID **)&out, "out in squeezeall"); return mid;
	  }
	if (mid != str) memzap((AVOID **)&mid, "mid in squeezeall");
	for ( ; cpo < cpe; cpo++) *cpo = '\0'; /* fill output with nulls */
	return out;
}

static int isenvpunct(char c)
{ /* This returns 1 iff c is a punctuation mark by the somewhat odd
   definitions associated with beginnings and endings of
   environment variables. Thus $ and _ return 0 but / and . return 1. */
	if ((('0' <= c) && (c <= '9')) || (('A' <= c) && (c <= 'Z')) ||
	 (('a' <= c) && (c <= 'z')) || (('$' == c) || ('_' == c))) return 0;
	return 1;
}

char *trans_envname(char *str)
{ /* This translates all the environment variables within `str'. */
	int		lenin, lenlast;
	char		*cpi, *cpo, *cptr, *hasdollar, *firstpu;
	char		*trann, *tranout, *trane, *nextver, *utran;

	if ((NULL == str) || ('\0' == *str)) return NULL;
	/* look for dollar signs and measure the length of the input */
	for (hasdollar = NULL, lenin = 0, cpi = str; '\0' != *cpi; cpi++)
	   {
		lenin++;
		if ((NULL == hasdollar) && ('$' == *cpi)) hasdollar = cpi;
	   }
	/* if there aren't any dollar signs or if the only dollar is
	 at the end of the string, we return the input */
	if ((NULL == hasdollar) || ('\0' == *(hasdollar + 1))) return str;
	if (NULL == (trann = mallop(768, "trann in trans_envname")))
		return NULL;
	zeromem(trann, 768);	tranout = trann + 256;	trane = trann + 768;
	/* the environment variable should start just past '$'
	 and continue through just before a null or a punctuation mark */
	for (cpo = tranout, cpi = hasdollar + 1, firstpu = NULL;
	 '\0' != *cpi; cpi++)
	   {
		if (isenvpunct(*cpi))
		  {
			firstpu = cpi; break;
		  }
		else *cpo++ = *cpi;
	   }
	*cpo = '\0';	if (NULL == firstpu) firstpu = cpi;
	/* environment variable names that haven't been defined are ignored */
	if (('\0' == *tranout) || (NULL == (cptr = getenv(tranout)))
	 || ('\0' == *cptr))
	  {
		memzap((AVOID **)&trann, "trann in trans_envname"); return str;
	  }
	/* finally we have a viable environment variable name.
	 Copy the stuff up to the dollar into the output,
	 then copy the environment variable over, then append the rest */
	for (cpo = trann, cpi = str; ('\0'!= *cpi) && (cpi< hasdollar); cpi++)
		*cpo++ = *cpi;
	for (cpi = cptr; '\0' != *cpi; cpi++) *cpo++ = *cpi;
	if (NULL != firstpu)
		for (cpi = firstpu; '\0' != *cpi; cpi++) *cpo++ = *cpi;
	*cpo = '\0';	for ( ; cpo < trane; cpo++) *cpo = '\0';
	/* internal recursion: look for additional environment variables. */
	while (1)
	     {
		if ((nextver = trans_envname(trann)) == trann) break;
		trann = nextver;
	     }
	/* to be parsimonious about memory, we measure the real length of
	 our translated variable and create a new string that is only long
	 enough to satisfy the need. */
	for (cpi = trann, lenlast = 0; '\0' != *cpi; cpi++) lenlast++;
	if (NULL == (utran = (char *)mallop(lenlast + 2,
	 "utran in trans_envname")))
	  {
		printf("Warning: couldn't output a shorter version\n");
		return trann;
	  }
	for (cpi = trann, cpo = utran; '\0' != *cpi; cpi++) *cpo++ = *cpi;
	*cpo = '\0';	memzap((AVOID **)&trann, "trann in trans_envname");
	return utran;
}

int xeflush(FILE *fp)
{	/* This does an fflush on the stream fp and, if we're running Unix,
	follows this with a call to `sync' to actually force output.
	This used to be called `xflush'; we've changed the name so
	it doesn't conflict with the X windows routine. */
	int	retflush;
	if (EOF == (retflush = fflush(fp))) return retflush;
#ifndef VMS
	sync();
#endif
	return retflush;
}

int expandit(char *nam, int ftyp)
{ /*	This actually calls a system routine to do a gzip expansion
	of the filename `nam'. For the moment we always expand the whole
	file; we may not keep doing it that way. The routine returns
	-1 if the file can't be opened, other errors if the expansion fails;
	The argument `ftyp' tells us what kind of expanded file we have.
	On success, the routine returns that argument.
	If the zlib stuff is absent, this returns -5 unless it's just being
	called to delete memory stores, in which case it returns EOK
	(although in that case it actually isn't doing anything). */
#ifndef ZLIB
	if (NULL == nam) return EOK;
	else return -5;
#else
	int		lofd, readva, inr, j, sysstat, ftyq, zer;
	char		*cpi, *cpo;
	size_t		ngot, nout;
	FILE		*fplo;
	gzFile		gpi;
	static char	*base, *sycbu;
	static size_t	syclen, oldsyclen;
	static char	basemsg[] = "base in expandit";
	static char	sycmsg[] = "sycbu in expandit";

	if (NULL == nam) /* special case to dump memory */
	  {
		memzap((AVOID **)&sycbu, sycmsg);	oldsyclen = 0;
		memzap((AVOID **)&base, basemsg);	return EOK;
	  }
	errno = 0;
	if (-1 == (lofd = open(nam, 0))) return -5;
	inr = read(lofd, (void *)&readva, 4); /* read 4 bytes */
	(void)close(lofd);	/* close the file regardless */
	if (4 > inr) return -4;	/* file has to be at least 4 bytes long! */
	ftyq = ftyp & 07;	/* we can forget where it came from */
	if (1 & ftyq) /* check Gzip format */
	  {
#ifdef LITTLEND
		if ((readva & 0xffff) != 0x8b1f) return -1;
#else
		if (((readva >> 16) & 0xffff) != 0x1f8b) return -1;
#endif
	  }
	if (2 & ftyq) /* check compress format */
	  {
#ifdef LITTLEND
		if ((readva & 0xffff) != 0x9d1f) return -2;
#else
		if (((readva >> 16) & 0xffff) != 0x1f9d) return -2;
#endif
	  }
	/* create a buffer either for the output filename or for
	 the command we'll emit. */
	syclen = strlen(nam) + 16;
	if ((j = syclen % 8)) syclen += 8 - j;
	if (syclen < 128) syclen = 128;
	if ((NULL == sycbu) || (syclen > oldsyclen))
	  {
		if (NULL != sycbu) memzap((AVOID **)&sycbu, sycmsg);
		if (NULL == (sycbu = (char *)mallop(syclen,
		 "sycbu in expandit"))) return -7;
		oldsyclen = syclen;
	  }
	/* if we got this far, we need to uncompress a file */
	if (2 & ftyq) /* .Z files: must be expanded with gunzip currently */
	  {
		sprintf(sycbu, "gunzip %s\n", nam);
		if ((0 != (sysstat = system(sycbu))) && (256 != sysstat))
		 fprintf(stderr,
		  "Warning: System returned %d while performing %s\n",
				sysstat, sycbu);
		return ((0 == sysstat) || (256 == sysstat)) ? ftyp : -8;
	  }
	else {	/* .gz files: can be done with library calls */
		for (cpi = nam, cpo = sycbu; '\0' != *cpi; )
			*cpo++ = *cpi++;
		*(cpo - 3) = '\0';
		if (NULL == (gpi = gzopen(nam, "rb"))) return -5;
		if (NULL == (fplo = fopen(sycbu, "w"))) return -5;
		if ((NULL == base) && (NULL == (base = mallop(GZBUFSIZ + 4,
		 basemsg)))) return -7;
		nout = GZBUFSIZ;
		while (0 < (ngot = gzread(gpi, base, nout)))
			if (1 != fwrite((AVOID *)base, ngot, 1, fplo))
				return -9;
		errno = 0;
		if ((zer = gzclose(gpi)))
		  {
			fprintf(stderr,
			 "failure in gzclose: zer, errno = %d, %d\n",
				(int)zer, (int)errno);
		  }
		errno = 0;
		if ((zer = fclose(fplo)))
		  {
			fprintf(stderr,
			 "failure in fclose: errno = %d\n", (int)errno);
		  }
		/* gulp. Now we unlink the input */
		errno = 0;
		if (0 != (zer = unlink(nam)))
		  {
			fprintf(stderr,
			 "Failed to unlink (%s): zer = %d\n", nam, zer);
			return -10;
		  }
		return ftyp;
	     }
#endif
}

int xgaccess(char *nam, int mode, int verbosity)
{ /*	This looks to see if `nam' exists and satisfies the requirements
	of the mode variable. If it does it returns the output from `access'.
	If not, it looks to see if the file exists in a compressed mode and
	uncompresses it accordingly. `verbosity' governs whether we emit
	a bunch of diagnostics if we uncompress it. */
	int		accres, exret, i, savelen;
	size_t		lennam;
	static size_t	oldauglen;
	static char	*augname;
	static char	augmsg[] = "augname in xgaccess";
	
	if (verbosity == -1) /* special case: clear out buffer */
	  {
		memzap((AVOID **)&augname, augmsg);	oldauglen = 0;
		return 0;
	  }
	errno = 0;	/* let's not confuse ourselves, eh wot? */
	if (0 == (accres = access(nam, mode))) return accres;
	/* check errno. The only error we pursue further is ENOENT. */
	if (errno != ENOENT) return accres;
	/* so now we look for an augmented version. First create space for
	 the new filename */
	lennam = strlen(nam);	savelen = lennam + 8;
	if (savelen < 80) savelen = 80;
	if ((NULL == augname) || (savelen > oldauglen))
	  {
		if (NULL != augname) memzap((AVOID **)&augname, augmsg);
		if (NULL == (augname = (char *)mallop(savelen, augmsg)))
			return accres;
		oldauglen = savelen;	zeromem((AVOID *)augname, savelen);
	  }
	if ((!(errno == 0)) && (!(errno == ENOENT))) return accres;
	for (i = 1; i < 3; i++)
	   {
		strcpy(augname, nam); if (i == 1) strcat(augname, ".gz");
		else strcat(augname, ".Z");
		if (0 == (accres = access(augname, mode)))
		  {	/* found the .gz version. Uncompress it */
			if (verbosity) printf("uncompressing %s: ", augname);
			if (i == (exret = expandit(augname, i)))
			  {
				if (verbosity) printf(" successful\n");
				break;
			  }
			else {
				if (verbosity)
	printf(" failed: errorcode %3d. Input left intact\n", exret);
				accres = exret;	break;
			     }
		  }
	   }
	return accres;
}

FILE *propen(char *name)
{ /*	This looks to see if the file with the name (environment, logical,
	or file) `name' exists and is writable.  If it is, it opens it for
	appending ("a"). If not, it creates the file as a VMS text file,
	mode 664, and then attaches a stream to it.
	If it fails at doing all this, it returns NULL;
	otherwise it returns the stream pointer associated with the file. */
	int	fd;
	char	*tname;

#ifdef VMS	/* VMS case... we'll just look for existence */
	if (0 == access(name, 0))
		return fopen(name, "a", "rat=cr", "shr=get", "rfm=var");
	return fopen(name, "w", "rat=cr", "shr=get", "rfm=var");
#else	/* Unix: we want to be able to write on the dern thing */
	if (0 == xgaccess((tname = trname(name)),2, 1)) return fopen(tname,"a");
	if (EOF != (fd = creat(tname,0664))) return fdopen(fd,"w");
	return NULL;
#endif
}

char *xgprogname(char *inname, int nflag)
{ /*	This manages two externally accessible strings `progname' and 'oldprog'.
	The `nflag' value determines what we do with these two variables:
	nflag	old routine	meaning	
		name
	_____	__________	_______
	0	currpname	return current value of `progname'
	1	kill_pname	eliminate memory for `progname'	and `oldprog'
	2	hide_oldpname	save copy of old program name in `oldprog'
				and put a bunch of '\0's in progname
	3	restore_oldpname	copy `oldprog' into `progname'
	4	assign_pname	set progname to inname
	The routine returns the current value of progname when invoked with
	flag values 0, 3, and 4; it returns NULL with flag values 1 or 2 */
	size_t		lenold, lenpn;
	static char	*oldprog, *progname;
	static char	pmsg[] = "progname in xgprogname";
	static char	omsg[] = "oldprog in xgprogname";

	switch (nflag) /* legal values are zero through four */
	  {
		case	0:	return progname;
		case	1:	/* get rid of memory for these globals */
		 memzap((AVOID **)&progname, pmsg);
		 memzap((AVOID **)&oldprog, omsg);	return NULL;
		case	2:	/* hide the old name in "oldprog" */
		 if ((NULL != progname) && ('\0' != *progname))
		   {
			lenpn = strlen(progname);
			if (lenpn < 1) return NULL;
			memzap((AVOID **)&oldprog, omsg);
			oldprog = mallop(lenpn + 2, omsg);
			strcpy(oldprog, progname);
			zeromem(progname, lenpn); /* lots of nulls! */
		  }
		 return NULL;
		case	3:	/* restore the old name back to `progname' */
			if ((NULL == oldprog) || ('\0' == *oldprog))
				return NULL;
			lenpn = strlen(oldprog);
			memzap((AVOID **)&progname, pmsg);
			progname = mallop(lenpn + 2, pmsg);
			zeromem(progname, lenpn + 2);
			strcpy(progname, oldprog);	return progname;
		case	4: /* assign input to progname */
			if ((NULL == inname) || ('\0' == *inname)) return NULL;
			lenpn = strlen(inname);
			if (NULL != progname) /* memory already present? */
			  {
				if ((lenold = strlen(progname)) < lenpn)
					memzap((AVOID **)&progname, pmsg);
				else {
					zeromem(progname, lenold);
					strcpy(progname, inname);
					return progname;
				     }
			  }
			progname = mallop(lenpn + 2, pmsg);
			zeromem(progname, lenpn + 2);
			strcpy(progname, inname);	return progname;
		default:	break;
	  }
	return NULL;	/* only happens if called with nflag < 0 or > 4 */
}

char *journalname(void)
{ /*	This returns the name of the X-GEN journal file */
	static char	*cp;
	static char	djname[] = "xgen.jnl";

	/* file name is either the translation of XGENJNL or it's
	 xgen.jnl in the current directory */
	if ((NULL == (cp = getenv("XGENJNL"))) || ('\0' != *cp))
		cp = djname;
	return cp;
}

int xgjournal(int ac, char *av[])
{ /*	This writes to a journal file upon invocation of an X-GEN command.
	If the journal file already exists the command is simply appended to it.
	If the journal file does not exist it is created. If it has zero length
	the command is preceded by some shell commands. To keep things tidy
	we always close the file when we're finished with this effort.
	The routine returns EOK if fully successful, various errors if not.
	We handle --help and --version arguments in this routine */

	int		ic, needstart;
	char		*cp, *jname, *lpname, *ocp, **dcp, firstword[65];
	char		dstamp[27];
	time_t		tlj;
	struct stat	stbu;
	FILE		*fpj;
	static char	*xgve[] = {
	 "anal_u", "analmr", "applycor", "border", "calibrate",
	 "cenmer", "cleanm", "cutoff", "editbord", "editcen",
	 "ewmask", "filterc", "filterc", "filteru", "imageform",
	 "integrate", "makemu", "marccdh", "mergefm", "mermask",
	 "mrmerge", "mukb", "mutos", "pdisplay", "pickuref",
	 "pmotif", "prax", "qec", "reduce", "refall",
	 "refine", "refine1", "reject", "scalek", "scalem",
	 "setup", "spaceg", "spots", "stats", "strategy",
	 "untwinit", "utos", "x2u", "xdump", "xgnow", "showhead",
	 "predict", "prhelp", "prversion", "reindmu", "reindu", NULL,
	 } ;
	static char	jnmsg[] = "jname in xgjournal";

	if (NULL == av) /* this routine only works if av is non-NULL */
		return EBADVALUE;
	cp = filepart(*av, 2);
	lpname = xgprogname(cp, 4);	/* record progname */
	if (ac > 1) /* handle --help and --version arguments in this routine*/
	  {
		if ((!strncmp(av[1], "--help", 6)) ||
		 (!strncmp(av[1], "--HELP", 6)))
		  {
			helpapp(lpname);	return EHELP;
		  }
		if ((!strncmp(av[1], "--version", 9)) ||
		 (!strncmp(av[1], "--VERSION", 9)))
		  {
			versionapp(lpname);	return EHELP;
		  }
	  }
	needstart = 0; /* record whether we're starting a new one */
	/* record how long the journal filename string is */
	ic = strlen(cp = journalname()) + 4;
	/* reserve memory for the string and fill it */
	if (NULL == (jname = mallop(ic, jnmsg))) return EALLOC;
	strcpy(jname, cp);
	/* see if the file already exists */
	if (0 != xgaccess(jname, 4, 1)) needstart = 1;
	else {
		if (-1 == stat(jname, &stbu))
		  {
			memzap((AVOID **)&jname, jnmsg); return EOPENERR;
		  }
		if (1 > stbu.st_size) needstart = 1;
	     }
	if (NULL == (fpj = fopen(jname, "a")))
	  {
		memzap((AVOID **)&jname, jnmsg);	return EOPENERR;
	  }
	if (needstart)
	  {
		tlj = NOW; strcpy(dstamp, ctime(&tlj));
		fprintf(fpj, "# X-GEN command journal begun %s", dstamp);
		if ((NULL != (cp = getenv("XGENCMD"))) && ('\0' != *cp))
			fprintf(fpj, "source %s\n", cp);
	  }
	/* 13 May 2005: if the command does not already begin with
	 `xgen' or `xgentools', prepend one or the other */
	zeromem((char *)firstword, 64);
	for (cp = *av, ocp = firstword; '\0' != *cp; cp++)
	   {
		if ((('A' <= *cp) && (*cp <= 'Z')) ||
		 (('a' <= *cp) && (*cp <= 'z')) ||
		 (('0' <= *cp) && (*cp <= '9')) || '_' == *cp) *ocp++ = *cp;
		else break;
	   }
	*ocp = '\0';
	if (strcmp(firstword, "xgen") && strcmp(firstword, "xgentools"))
	  {
		for (dcp = xgve; NULL != *dcp; dcp++)
		   if (!strcmp(firstword, *dcp)) break;
		if (NULL != *dcp) fputs("xgen ", fpj);
		else fputs("xgentools ", fpj);
	  }	
	for (ic = 0, dcp = av; (ic < ac) && (NULL != *dcp) &&
	 ('\0' != **dcp); ic++, dcp++)
	   {
		fputs(*dcp, fpj);	if (ic < (ac - 1)) fputs(" ", fpj);
	   }
	fputs("\n", fpj);
	ic = xeflush(fpj);	(void)fclose(fpj);
	memzap((AVOID **)&jname, jnmsg);	return ic;
}

int xgbuf(char *str, int bflag)
{ /*	This manages a private buffer for text messages.
	The flag value is the or-ing of these three bits:
	bit	meaning of on			meaning of off
	---	----				----
	1	put `str' in locbu		don't put `str' in locbu
	2	call xgjournal to output stuff	don't call xgjournal
	3	zero out oldlen & oldbusiz	don't zero oldlen & oldbusiz
	Therefore xgbuf(str, 1) will put str in locbu;
	 xgbuf(str, 7) will put it in locbu, write that out to xgen.jnl,
	 and then zero things out;
	 xgbuf(str, 6) will just flush the current contents to xgen.jnl.
	Special case: xgbuf(NULL, 0) will kill memory for locbu. */
	int		oac, result;
	char		*cp, *newbu, *start, *oav[2];
	size_t		lenloc, lenstr, newlen, newbusiz;
	static char	*locbu;
	static size_t	oldlen = 0;
	static size_t	oldbusiz = 0;
	static char	locbmsg[] = "locbu in xgbuf";

	if ((NULL == str) && (0 == bflag))
	  {
		if (NULL != locbu) memzap((AVOID **)&locbu, locbmsg);
		oldlen = oldbusiz = 0;	return 0;
	  }
		/* no output if string is blank */
	if ((NULL == str) || ('\0' == *str)) bflag &= ~01;
	if (bflag & 01) /* stuff `str' in buffer */
	  { /* measure the length of the string */
		if ((lenstr = strlen(str)) < 1) return EBAD;
		/* guarantee that it's newline-terminated */
		if (*(str + (lenstr - 1)) != '\n')
		  {
			lenstr += 1;	*(str + (lenstr - 1)) = '\n';
			*(str + lenstr) = '\0';
		  }
		/* if the buffer itself hasn't been created, do so */
		if ((0 == oldlen) || (NULL == locbu) || (0 == oldbusiz))
		  {
			(void)xgbuf(NULL, 0);	oldbusiz = BUFBLOCK;
			if (NULL == (locbu = mallop(oldbusiz, locbmsg)))
				return EALLOC;
			/* . . . and then put the string into the buffer */
			zeromem((AVOID *)locbu, oldbusiz);
			strcpy(locbu, str);
			oldlen = lenstr; /* remember how much space is there */
		  }
		else { /* otherwise, append to the buffer */
			start = locbu + oldlen;	newlen = oldlen + lenstr;
			if (newlen > oldbusiz) /* will we need more space? */
			  {
				newbusiz = oldbusiz + BUFBLOCK; /* get space */
				if (NULL == (newbu = reallop(locbu, newbusiz,
				 "newbu in xgbuf"))) return EALLOC;
				locbu = newbu; /* maybe unnecessary! */
				start = locbu + oldlen;
				zeromem((AVOID *)start, newbusiz - oldlen);
				oldbusiz = newbusiz;
			  }
			strcpy(start, str); /* now append the new string */
			oldlen = newlen; /* remember space consumed */
		     }
	  } /* end of "if (bflag & 01)" clause */
	if (NULL == locbu) /* if there's nothing in locbu, don't continue */
		bflag &= ~06;
	if (bflag & 02) /* call xgjournal to put this outside */
	  { /* "xgjournal" appends a newline of its own, so delete
		 the last one from here */
		lenloc = strlen(locbu);	cp = locbu + (lenloc - 1);
		if ('\n' == *cp) *cp = '\0';
		oav[0] = locbu;	oav[1] = NULL;	oac = 1;
		if (EHELP == (result = xgjournal(oac, oav))) result = EOK;
	  }
	else	result = EOK;
	if (bflag & 06) /* tell the outside world that the buffer is empty */
	  {
		oldlen = oldbusiz = 0;	return result;
	  }
	else	return EOK;
}

static int dircontains(char *dname, char *finame)
{ /*	This goes through the directory "dname" and looks to see if a file
	"finame" is contained in it. It returns 0 if successful, various error
	values if not. We _could_ do this with calls to internal Unix functions
	like opendir and readdir, but I'll just test for access. */
	int			accres;
	size_t			lenstr;
	static char		*dbuf;
	static size_t		oldlen;
	static char		dbumsg[] = "dbuf in dircontains";

	if ((NULL == dname) && (NULL == finame))
	  {
		memzap((AVOID **)&dbuf, dbumsg); oldlen = 0; return 0;
	  }
	if (0 != access(dname, 04)) return EOPENERR;
	lenstr = strlen(dname) + strlen(finame) + 8;
	if (lenstr < 80) lenstr = 80;
	if ((NULL == dbuf) || (lenstr > oldlen))
	  {
		memzap((AVOID **)&dbuf, dbumsg);
		if (NULL == (dbuf = mallop(lenstr, dbumsg))) return EALLOC;
	  }
	zeromem((AVOID *)dbuf, lenstr);	oldlen = lenstr;
	strcpy(dbuf, dname);	strcat(dbuf, "/");	strcat(dbuf, finame);
	accres = access(dbuf, 04);	return accres;
}

static char *smartfloat(float tval, int forcefloat)
{
 /* This returns a printable string (%d, %f, or %e)
  associated with valu. Note that it returns a static
  variable, so don't try anything cute. */
	int		ival;
	double		delv, rval;	
	static char	lobu[16];

	ival = (tval < 0.) ? tval - 0.49999 : tval + 0.49999;
	rval = ival;	delv = rval - tval;
	zeromem(lobu, 16);
	if ((!forcefloat) && ((-0.001 < delv) && (delv < 0.001)))
		sprintf(lobu, "%d", ival);
	else if ((tval < -1.e6) || (tval > 1.e6))
		sprintf(lobu, "%13.5e", tval);
	else	sprintf(lobu, "%f", tval);
	return lobu;
}

int xgvariable(char *suffx, float *valu, int nvalues)
{
 /* This allows us to record useful information from an X-GEN application
  in a central (appendable) file in a format that looks like
 xgen_{app}_{suffx} = value
 e.g.
 xgen_spots_D2 = 1.64
  The file that will contain this interesting stuff will generally be
  $WORK/{projectname}.xou
 The routine returns EOK if successful, various errors if not.
 Special case: if suffx is NULL, this is designed to close the file
	and get rid of memory
 */
	int		forcefloat, jv, lenstr;
	char		*cpend, *cpi, *cpo, *curname, *idnam, *start, *wname;
	char		obu[256];
	time_t		ltl0;
	static XGNAME	*losh;
	static char	*xouname;
	static FILE	*fxou;
	static char	xoustr[] = "xouname in xgvariable";

	if ((NULL == suffx) || ('\0' == *suffx))
	  {
		if (NULL != fxou) (void)fclose(fxou);
		memzap((AVOID **)&(xouname), xoustr);
		fxou = NULL;	return EOK;
	  }
	if (NULL == (curname = xgprogname(NULL, 0))) return EBADVALUE;
	forcefloat = 0;
	if (nvalues < 0)
	  {
		forcefloat = 1;	nvalues = -nvalues;
	  }
	if (nvalues < 1) return EBADVALUE;
	if (NULL == fxou) /* open the file if necessary */
	  {
		if (NULL == (losh = sharename(7))) return EBADVALUE;
		if ('\0' == *(losh->xmm_workna))
		  {
			wname = getenv("WORK");
			if ((NULL != wname) && ('\0' != *wname))
			  {
				lenstr = strlen(wname);
				if (lenstr > 126) lenstr = 126;
				strncpy(losh->xmm_workna, wname, lenstr);
			  }
		  }
		if ('\0' == *(losh->xmm_idna))
		  {
			idnam = getenv("MULTIREF");
			if ((NULL != idnam) && ('\0' != *idnam))
			  {
				lenstr = strlen(idnam); start = NULL;
				cpend = idnam + lenstr - 1;
				cpo = losh->xmm_idna;
				for (cpi = cpend; cpi > idnam; cpi--)
				   if ('/' == *cpi)
				     {
					start = cpi + 1; break;
				     }
				if ((cpi <= idnam) || (start > cpend))
					start = idnam;
				for (cpi = start;
				 ('\0' != *cpi) && (cpi <= cpend); cpi++)
				   {
					if (*cpi == '.') break;
					else *cpo++ = *cpi;
				   }
				*cpo = '\0';	
			  }
		  }
		wname = losh->xmm_workna;
		idnam = losh->xmm_idna;
		lenstr = 8;
		if ((NULL != wname) && ('\0' != *wname))
			lenstr += strlen(wname);
		if ((NULL != idnam) && ('\0' != *idnam))
			lenstr += strlen(idnam);
		memzap((AVOID **)&(xouname), xoustr);
		if (NULL == (xouname = (char *)mallop(lenstr, xoustr)))
			return EALLOC;
		if ((NULL != wname) && ('\0' != *wname))
			strcpy(xouname, wname);
		else	strcpy(xouname, ".");
		strcat(xouname, "/");
		if ((NULL != idnam) && ('\0' != *idnam))
			strcat(xouname, idnam);
		else	strcat(xouname, "unknown");
		strcat(xouname, ".xou");
		if (NULL == (fxou = propen(xouname)))
		  {
			memzap((AVOID **)&xouname, xoustr);
			return EOPENERR;
		  }
		/* now we emit a timestamp to the file */
		ltl0 = NOW;
		sprintf(obu, "xg_%s_timestamp = %s", curname, ctime(&ltl0));
		fputs(obu, fxou);
	  }
	zeromem((AVOID *)obu, 256);
	if (nvalues < 2)
	  { /* single value */
		sprintf(obu, "xg_%s_%s = %s\n",
			curname, suffx, smartfloat(*valu, forcefloat));
	  }
	else { /* multiple values; treat like a Python list */
		sprintf(obu, "xg_%s_%s = [", curname, suffx);
		for (jv = 0; jv < nvalues; jv++)
		   {
			strcat(obu, smartfloat(*(valu + jv), forcefloat));
			if (jv < nvalues - 1) strcat(obu, ",");
		   }
		strcat(obu, "]\n");
	     }
	fputs(obu, fxou);
	return EOK;
}

char *findpath(char *str)
{ /*	This goes through the current path and locates the first instance of a
	directory listed in the path for which the file "str" is present in the
	directory. If no such directory is found the routine returns NULL. */
	int		resu;
	char		*icp, *ocp, *pathstr;
	static char	dirbuf[254];

	/* first, find the path environment variable */
	if (NULL == (pathstr = getenv("PATH")))
	  {
		fprintf(stderr, "Failed to find path environment variable\n");
		return NULL;	/* couldn't find path at all */
	  }
	ocp = &(dirbuf[0]);
	for (icp = pathstr; '\0' != *icp; icp++)
	   {
		if (':' == *icp)
		  {
			*ocp = '\0';
			if (0 == dircontains(dirbuf, str))
			  {
				(void)dircontains(NULL, NULL); return dirbuf;
			  }
			else	ocp = &(dirbuf[0]);
		  }
		else	*ocp++ = *icp;
	   }
	*ocp = '\0';	resu = dircontains(dirbuf, str);
	(void)dircontains(NULL, NULL);	return (resu) ? NULL : dirbuf;
}

int xgenrename(char *inpname)
{ /*	This checks to see if a file with the name `inpname' already exists.
	If it does, it renames that file according to the recipe described
	below, so that when we try to creat() or open() or fopen()
	`inpname' for writing, we won't be overwriting the original. */
	int		iready, try, needshift;
	char		*augname, *lastc, *prevc, *punctstr, suffix[3];
	unsigned	namelen;
	struct stat	stold;

	/* determine what punctuation mark will be used to attach the
	 version number (if needed) onto the end of the filename.  We examine
	 the environment variables VERSION_PUNCT and VERSION_PUNCTUATION
	 and use a comma if neither is present. */
	if ((NULL == (punctstr = getenv("VERSION_PUNCT"))) &&
	 (NULL == (punctstr = getenv("VERSION_PUNCTUATION")))) punctstr = ",";
	suffix[0] = *punctstr;	suffix[1] = '0';	suffix[2] = '\0';
	/* does this file exist already and does it have anything in it? */
	if (0 == access(inpname, 0))
	  { /* if we can't `stat' it, we make a new one & rename the old */
		if (-1 == stat(inpname, &stold)) needshift = 1;
		/* if the file is of zero length, we just overwrite it.
		 "Off with his head!", said the red queen. */
		else needshift = (stold.st_size > 0);
	  }
	else needshift = 0;
	if (needshift)	/*we want to rename the old one and then open the new*/
	  { /* See how long the name is */
		namelen = strlen(inpname);
		/* save some memory for the augmented name.
		 If this fails, return an error code */
		if (NULL == (augname = (char *)mallop((size_t)(namelen + 4),
		 "augname in xgenrename"))) return EALLOC;
		strcpy(augname, inpname); /* copy name to augmented-name */
		strcat(augname, suffix); /*add punctuation mark + "0" to that*/
		 /* see if the augmented name exists.  If it does,
		 try another version... if "name,0" exists, try "name,1";
		  if "name,1" exists, try "name,2" ... ; if "name,9" exists,
		  try "name,10" through "name,99".  If "name,99" already
		  exists, just overwrite it; you have too many versions anyway.
		 */
		lastc = augname + (namelen + 1); /* position of last char */
		for (try = iready = 0; try < 100; try++)
		   { /* last time through: abolish the newest one */
			if (try >= 99) (void)unlink(augname);
			/* Try to access augmented name. If that succeeds
			 (access==0) we try next version. If it fails but the
			 link can't be created anyway, we try next version*/
			if ((0 != access(augname, 0)) &&
			 (0 == link(inpname, augname)))
			  {
				(void)unlink(inpname);	iready = 1;	break;
			  }
			else if ('9' == *lastc) /* is the last char a 9? */
			  {
				prevc = lastc - 1; /*yes. Note previous char*/
				if ('9' == *prevc) break; /* give up if "99"*/
				else if (*punctstr == *prevc) /* X,9 -> X,10*/
				  {
					*lastc++ = '1';	*lastc = '0';
				  }
				/* blah,39 -> blah,40 */
				else {
					(*prevc)++;	*lastc = '0';
				     }
				*(lastc+1) = '\0';
			  }
			else	(*lastc)++; /* blah,6 -> blah,7 */
		   }
		memzap((AVOID **)&augname, "augname in xgenrename");
		if (!iready) return EOPENERR; /* open new one only if ready */
	  }
	return EOK;
}

FILE *xgopen(char *name, char *mode)
{ /*	This creates a new file and opens it.  Under Unix, it always translates
	`name' as an environment variable first if it can.  If it's to
	be opened for reading or appending (or "r+" or "a+") the routine
	just does so at that point.  If it's to be opened "w" or "w+",
	it checks to see if the file already exists.  If it does, it assumes
	that the user actually wanted to save the old version, so it renames
	the old version (a la VMS, sort of) before creating the new file.
	This whole operation is unnecessary under VMS where version numbers
	are built-in to the system.
	Sep 03: if an open for read or read-write fails, we attempt
	to gunzip the file before giving up. */
#ifndef VMS
	int		try;
	char		*tname, *augname, *lastc;
	unsigned	namelen;
	FILE		*fresu;

	tname = trname(name);	/* translate the environment variable */
	if (*mode != 'w')	/* open for reading or read+write: just open */
	  {
		errno = 0;	/* clear error before trying */
		if (NULL != (fresu = fopen(tname, mode))) return fresu;
		if (errno == ENOENT) /* try uncompressing it */
		  {
			if (0 == (try = xgaccess(tname, 4, 1)))
				return fopen(tname, mode);
			else return NULL;
		  }
		else return NULL;
	  }
	if (EOK != xgenrename(tname)) return NULL;
	errno = 0;
	fresu = fopen(tname, mode);	if (NULL != fresu) return fresu;
	/* we might be able to fix the problem if it's just a matter of the
	 directory or its parent being absent. Otherwise we must give up */
	if (ENOENT != errno) return NULL;
	for (lastc = tname + (strlen(tname) - 1); lastc >= tname; lastc--)
		if ('/' == *lastc) break;
	if ((lastc <= tname) || (*lastc != '/')) return NULL;
	namelen = lastc - tname;
	if (NULL == (augname = (char *)mallop((size_t)(namelen + 2),
		"augname in xgopen"))) return NULL;
	strncpy(augname, tname, namelen);	*(augname + namelen) = '\0';
	try = (0 == mydircreate(augname));
	memzap((AVOID **)&augname, "augname in xgopen");
	return (try) ? fopen(tname, mode) : NULL;
#else
	return fopen(name, mode); /* create & open the filename */
#endif
}

int xgdelete(char *name)
{ /* this is a system-independent file-deletion routine */
#ifdef VMS
	extern int	delete(char * nam);

	return delete(name);
#else
	return unlink(name);
#endif
}
#ifdef LITTLEND
#ifndef VMS
static void deccvt(AVOID *base, int xgtype, int direction)
{ /*	This converts elements of an X-GEN structure
	into or out of DEC format */
	int		i;
	short		*sp;
	xglong		*llp;
	float		*flp;
	CALPOINT	*lcp;
	REFPARS		*lrp;
	DATARUN		*ldp;
	CRYSTAL		*lxp;
	OBSREF		*lop;
	ABBROB		*lap;
	MASK_3D		*lmap;
	PHIMASK		*lphp;
	VSPLINE		*lvp;
	TROID		*ltp;
	MERGEDREF	*lmp;
	PREDREF		*lpp;

	switch (xgtype)
	      {
		case	XG_FLOAT:
			flcvt((void *)base, direction);	break;
		case	XG_CALPOINT:
			lcp = (CALPOINT *)base;
			for (flp= &(lcp->c_ptocx); flp<= &(lcp->c_qec); flp++)
				flcvt((void *)flp, direction);
			break;
		case	XG_REFPARS:
			lrp = (REFPARS *)base;
			for (flp = &(lrp->xlambda); flp <= &(lrp->xresmax);
			 flp++) flcvt((void *)flp, direction);
			for (flp = &(lrp->xdelchi); flp <= &(lrp->xdelom);
			 flp++) flcvt((void *)flp, direction);
			lrp->xisys = sswab(lrp->xisys);
			lrp->xispgp = sswab(lrp->xispgp);
			for (flp = &(lrp->xstart[0]); flp <= &(lrp->xstart[2]);
			 flp++) flcvt((void *)flp, direction);
			lrp->xdetwid = sswab(lrp->xdetwid);
			lrp->xdetht = sswab(lrp->xdetht);	break;
		case	XG_DATARUN:
			ldp = (DATARUN *)base;
			for (sp = &(ldp->run_id); sp <= &(ldp->run_height);
			 sp++) *sp = sswab(*sp);
			ldp->run_ninteg = lswawb(ldp->run_ninteg, direction);
			ldp->run_nexam = lswawb(ldp->run_nexam, direction);
			for (flp= &(ldp->run_lambda); flp<= &(ldp->run_ssqmax);
			 flp++) flcvt((void *)flp, direction);
			break;
		case	XG_CRYSTAL:
			lxp = (CRYSTAL *)base;
			lxp->xt_intl = sswab(lxp->xt_intl);
			lxp->xt_isys = sswab(lxp->xt_isys);
			for (flp = &(lxp->xt_ar[0]); flp <= &(lxp->xt_ssqmax);
			 flp++) flcvt((void *)flp, direction);
			lxp->xt_ninteg = lswawb(lxp->xt_ninteg, direction);
			lxp->xt_nrefls = lswawb(lxp->xt_nrefls, direction);
			lxp->xt_nmerged = lswawb(lxp->xt_nmerged, direction);
			lxp->xt_nruns = sswab(lxp->xt_nruns);
			lxp->xt_nshifts = sswab(lxp->xt_nshifts);	break;
		case	XG_OBSREF:
			lop = (OBSREF *)base;
			for (sp = &(lop->obs_hkl[0]); sp <= &(lop->obs_region);
			 sp++) *sp = sswab(*sp);
			for (flp = &(lop->obs_ii); flp <= &(lop->obs_omob);
			 flp++) flcvt((void *)flp, direction);
			for (sp = &(lop->obs_gof); sp <= &(lop->obs_reflno);
			 sp++) *sp = sswab(*sp);
			for (sp= &(lop->obs_frmint[0]); sp<= &(lop->obs_flags);
			 sp++) *sp = sswab(*sp);
			break;
		case	XG_ABBROB:
			lap = (ABBROB *)base;
			lap->abo_shift = sswab(lap->abo_shift);
			lap->abo_omob = sswab(lap->abo_omob);
			flcvt((void *)&(lap->abo_ii), direction);
			flcvt((void *)&(lap->abo_sigma), direction);	break;
		case	XG_MASK_3D:
			lmap = (MASK_3D *)base;
			for (i = 0, lphp = &(lmap->phielt[0]);i < 7;i++,lphp++)
			   {
				flcvt((void *)&(lphp->isqsum), direction);
				flcvt((void *)&(lphp->maxf), direction);
				lphp->mnobs = lswawb(lphp->mnobs, direction);
			   }	break;
		case	XG_PREDICTREST: /* this is squirrely but it works */
			for (flp = (float *)base, i = 0; i < 3; flp++, i++)
				flcvt((void *)flp, direction);
			llp = (xglong *)flp;	*llp = lswawb(*llp, direction);
			for (flp = flp + 1, i = 0; i < 2; flp++, i++)
				flcvt((void *)flp, direction);	break;
		case	XG_VSPLINE:
			lvp = (VSPLINE *)base;
			for (sp = &(lvp->v_detw); sp <= &(lvp->v_midr); sp++)
				*sp = sswab(*sp);
			for (flp = &(lvp->v_axc); flp <= &(lvp->v_y0px); flp++)
				flcvt((void *)flp, direction);
			break;
		case	XG_TROID:
			ltp = (TROID *)base;
			for (flp = &(ltp->xto); flp <= &(ltp->frame0); flp++)
				flcvt((void *)flp, direction);
			ltp->obkmin = sswab(ltp->obkmin);
			ltp->nobk = sswab(ltp->nobk);
			flcvt((void *)&(ltp->totinten), direction);
			break;
		case	XG_MERGEDREF:
			lmp = (MERGEDREF *)base;
			lmp->mr_nob = uswab(lmp->mr_nob);
			lmp->mr_ibar = uswab(lmp->mr_ibar);
			lmp->mr_isig = uswab(lmp->mr_isig);
			flcvt((void *)&(lmp->mr_stl), direction);	break;
		case	XG_PREDREF:
			lpp = (PREDREF *)base;
			lpp->pfmno = sswab(lpp->pfmno);	break;
	     }
}
#endif
#endif

static int xgnbyte(int xgtype)
{ /* This returns the number of bytes associated with an X-GEN data type */
	static int	nby[] =	{	  4, 32, 256, 256, 256,  80,  16,
		1680, 24, 0, 512, 20,  16,   2, 29564 } ;

	return ((xgtype < 0) || (xgtype > 14)) ? 0 : nby[xgtype];
}

int xgread(AVOID *base, int xgtype, FILE *fp)
{	/* This reads the appropriate number of bytes from the stream fp.
	It returns the total number of bytes actually read.
	On machines that don't adhere to the majority of Unix machine types
	(Dec Alpha running OSF/1, in particular) the data are transformed
	after readin according to the byte-order, word-order, and
	floating-point conventions appropriate to the machine;
	thus we can read a file written on 
	a little-end, IEEE-floating-point machine onto an Alpha. */
	int		nrd, ntrd, trd, npt, subtype;
	char		*cp;
	VSPLINE		*lvp;
	TROID		*ltp;
	MERGEDREF	*lmp;
	PREDREF		*lpp;
	AVOID		*subbase;

	/* make sure the inputs are okay */
	if ((NULL == fp) || (NULL == base)) return 0;
	if ((0 > xgtype) || (14 < xgtype)) return 0;
	/* read the data in */
	ntrd = xgnbyte(xgtype);
	if (ntrd != (nrd = fread(base, 1, ntrd, fp))) return nrd;
#ifdef LITTLEND /* DEC formats: make conversions here! */
#ifndef VMS	/* we'll leave VMS alone for now */
	deccvt(base, xgtype, 0);
#endif
#endif
	if (xgtype < XG_VSPLINE) return nrd;
	if (xgtype == XG_OSPLINE) return nrd;
	/* in certain cases we need to read in some additional data, the length
	 of which must be determined from the data at the front. In one of
	 those cases we also must reserve memory for the additional data */
	npt = 1;	/* so nothing horrible happens */
	subtype = XG_FLOAT;	subbase = (AVOID *)base;	/* innocuous */
	switch (xgtype)
	      {
		case	XG_VSPLINE:	/* calibration structure */
		 lvp = (VSPLINE *)base;
		 npt = lvp->v_nrow * lvp->v_ncol; /* # of CALPOINTs */
			/* reserve memory for the CALPOINT structures*/
		 if (NULL == (lvp->v_cpt = (AVOID *)callop((size_t)npt,
		  sizeof (CALPOINT), "lvp->v_cpt in xgread"))) return nrd;
		 subbase = (AVOID *)lvp->v_cpt; subtype = XG_CALPOINT; break;
		case	XG_TROID:	/* CENTROID */
		 ltp = (TROID *)base;	subbase = (AVOID *)&(ltp->obparts[0]);
		 npt = ltp->nobk;	subtype = XG_FLOAT;	break;
		case	XG_MERGEDREF:	/* merged-reflection (MULTIREF) */
		 lmp = (MERGEDREF *)base; subbase = (AVOID *)&(lmp->mr_ob[0]);
		 npt = lmp->mr_nob;	subtype = XG_ABBROB;	break;
		case	XG_PREDREF:	/* prediction reflection (PREDREF) */
		 lpp = (PREDREF *)base;	subbase = &(lpp->pxt);
		 npt = 1;	subtype = XG_PREDICTREST;	break;
	     }
	 /* read the data into the sub-structures and, if necessary, convert */
	for (cp = (char *)subbase, trd = nrd, ntrd = xgnbyte(subtype);
	 npt--; cp += ntrd)
	   {
		if (ntrd > (nrd = xgread((void *)cp, subtype, fp)))
			return trd + nrd;
		trd += nrd;
	   }
	return trd;
}

int xgwrite(AVOID *ibase, int xgtype, FILE *fp)
{	/* This writes the appropriate number of bytes from the stream fp.
	It returns the total number of bytes written.
	On machines that don't adhere to the majority of MSI's machine types
	(Dec Alpha running OSF/1, in particular) the data are transformed
	before writing according to the byte-order, word-order, and
	floating-point conventions appropriate to the machine;
	thus we can read a file written on 
	a little-end, IEEE-floating-point machine onto an Alpha.
	On AXP's, copy data to a buffer before converting it, so the
	copy sitting in memory is still in the right byte-&-word order.*/
	int		nwr, twr, ntwr, npt, subtype;
	char		*cp, *base;
#ifdef LITTLEND
	char		littboo[1680];
#endif
	VSPLINE		*lvp;
	TROID		*ltp;
	MERGEDREF	*lmp;
	PREDREF		*lpp;
	AVOID		*subbase;

	/* make sure the inputs are okay */
	if ((NULL == fp) || (NULL == ibase)) return 0;
	if ((0 > xgtype) || (14 < xgtype)) return 0;
	/* in three cases we need to write out some additional data,
		 and for those we need to establish base values for those
		 writes NOW, before conversion */
	npt = 1;	/* so nothing horrible happens */
	subtype = XG_FLOAT;	subbase = (AVOID *)ibase;	/* innocuous */
	if ((xgtype >= XG_VSPLINE) && (xgtype < XG_OSPLINE))
	  {
	    switch (xgtype)
	       {
		case	XG_VSPLINE:	/* calibration structure */
		 lvp = (VSPLINE *)ibase;	subbase = (AVOID *)lvp->v_cpt;
		 npt = lvp->v_nrow * lvp->v_ncol; /* # of CALPOINTs */
		 subtype = XG_CALPOINT;	break;
		case	XG_TROID:	/* CENTROID */
		 ltp = (TROID *)ibase;	subbase = (AVOID *)&(ltp->obparts[0]);
		 npt = ltp->nobk;	subtype = XG_FLOAT;	break;
		case	XG_MERGEDREF:	/* merged-reflection (MULTIREF) */
		 lmp = (MERGEDREF *)ibase; subbase = (AVOID *)&(lmp->mr_ob[0]);
		 npt = lmp->mr_nob;	subtype = XG_ABBROB;	break;
		case	XG_PREDREF:	/* prediction reflection (PREDREF) */
		 lpp = (PREDREF *)ibase;	subbase = &(lpp->pxt);
		 npt = 1;	subtype = XG_PREDICTREST;	break;
	       }
	  }
	ntwr = xgnbyte(xgtype);
#ifdef LITTLEND	/* copy data into a Littlend-specific buffer before swapping */
	cpymem((AVOID *)littboo, ibase, (unsigned)ntwr);
	deccvt((AVOID *)littboo, xgtype, 1); /* ... then byte&word swap */
	base = (char *)littboo;
#else
	base = (char *)ibase;
#endif
	/* write the data out */
	if (ntwr != (nwr = fwrite((AVOID *)base, 1, ntwr, fp))) return nwr;
	if (xgtype < XG_VSPLINE) return nwr;
	/* if necessary, convert the data in the sub-structures. Then
	 write those data out */
	for (cp = (char *)subbase, twr = nwr, ntwr = xgnbyte(subtype);
	 npt--; cp += ntwr)
	   {
		if (ntwr > (nwr = xgwrite((void *)cp, subtype, fp)))
			return twr + nwr;
		twr += nwr;
	   }
	return twr;
}

int match_inode(FRAMEINDEX **dfmpi, char *finame)
{ /* if the inode associated with the filename `finame' matches the
	inode of one of the images in the image table, we return
	the image number of that file. Otherwise, we return -2. */
	int		alreadyind;
	ino_t		thisino, thatino;
	struct stat	stbu;
	FRAMEENTRY	*ftq;

	if (-1 == stat(finame, &stbu)) return -2;
	thisino = stbu.st_ino;
	if (NULL == dfmpi) return -3;
	alreadyind = 0;
	if (NULL == *dfmpi) *dfmpi = ftables(3);
	else alreadyind = 1;
	if (NULL == *dfmpi) return -2;
	for (ftq = (*dfmpi)->fi_fmi; ftq <= (*dfmpi)->fi_fme; ftq++)
	 if (-1 != stat(ftq->fmname, &stbu))
	   {
		thatino = stbu.st_ino;
		if (thisino == thatino)
		  {
			if (!alreadyind) (void)ftables(0);
			return ftq->fmno;
		  }
	   }
	if (!alreadyind) (void)ftables(0);	return -2;
}

int compressiontype(char *nam, int *lolenp)
{ /*	This returns a code value depending on the ending of the name `nam':
	code value	meaning
	-6		name is invalid.
	 0		name valid, but doesn't appear to be a compression type
	 1		name ends with .gz or -gz (gzipped file type)
	 2		name ends with .Z (compressed file type)
	 4		name ends with .z or -z
    It also returns the length of the string `nam' as *lolenp because it can.*/
	char	*cp;

	if ((NULL == nam) || ('\0' == *nam)) return -6;
	for (*lolenp = 0, cp = nam; '\0' != *cp; cp++) (*lolenp)++;
	if (*lolenp < 3) return 0; /* a.Z is the shortest style possible */
	/* see if it ends in .gz or -gz */
	if ((*lolenp > 3) && (('g' == *(cp - 2)) && ('z' == *(cp - 1))))
	  {
		if (('.' == *(cp - 3)) || ('-' == *(cp - 3))) return 1;
	  }
	else if (('.' == *(cp - 2)) && ('Z' == *(cp - 1))) return 2;
	else if (('z' == *(cp -1)) && (('.' == *(cp -2)) || ('-' == *(cp -2))))
		return 4;
	return 0;
}

int shrinkit(char *nam)
{ /*	This gzips a file in place.
  If zlib is unavailable it always returns EBADVALUE */

#ifndef ZLIB
	return EBADVALUE;
#else
	int	retv;
	size_t	ngot, nstart;
	char	*onam, *base, *cpi, *cpo;
	FILE	*fpli;
	gzFile	*gpo;

	if (NULL == (fpli = fopen(nam, "r"))) return EBADVALUE;
	if (NULL == (onam = mallop(strlen(nam) + 8, "onam in shrinkit")))
		return EALLOC;
	if (NULL == (base = mallop(GZBUFSIZ + 4, "base in expandit")))
		return EALLOC;
	for (cpi = nam, cpo = onam; '\0' != *cpi; ) *cpo++ = *cpi++;
	*cpo++ = '.';	*cpo++ = 'g';	*cpo++ = 'z';	*cpo = '\0';
	if (NULL == (gpo = gzopen(onam, "wb"))) return EOPENERR;
	nstart = GZBUFSIZ;	retv = EOK;
	while (0 < (ngot = fread((AVOID *)base, 1, nstart, fpli)))
	     {
		if (ngot != (size_t)gzwrite(gpo, (AVOID *)base, ngot))
		  {
			retv = EBADVALUE;	break;
		  }
	     }
	(void)gzclose(gpo);	(void)fclose(fpli);
	memzap((AVOID **)&base, "base in expandit");
	memzap((AVOID **)&onam, "onam in expandit");
	if (0 != unlink(nam)) return ECLOSERR;
	return retv;
#endif
}

static int retcompare(int fde, int retv, off_t wasat)
{ /*	This sets the pointer for file descriptor `fde' to `wasat'
	and then returns the return value `retv'. It's a convenience
	function for comparemid(). */
	(void)lseek(fde, wasat, 0);	return retv;
}

int comparemid(char *fnam, int fde, int sizfil, int verbo)
{ /*	This looks near the middle of a file and compares the number of
	low-bytes that are nonzero to the number of high-bytes that are
	nonzero and reports accordingly. It returns:
	various negative integers when it fails;
	0 if there are no nonzero pixels whatever among the pixels it tests;
	1 if < odd-byte values> is >= 2 * <even-byte values>
	2 if <even-byte values> is >= 2 * < odd-byte values>
	3 if the two sets of values are within a factor of two of one another
   It is assumed that the file is already open when we start this process.
	Arguments;
	 fnam   filename (used only in diagnostics)
	 fde    filedescriptor (must be non-negative and valid)
	 sizfil size of the file in bytes (obtained from stat buffer)
	 verbo  verbose flag (0 or 1)	*/
	int	ngot, i, nodd, neven, nposs, mid;
	u_char	test[256], *cp;
	off_t	wasat;
	double	meanodd, meaneven, neach;

	mid = (sizfil < 131072) ? 8192 : 65536; /* pick spot near the start */
	nodd = neven = nposs = 0;		/* zero out some counters */
	wasat = lseek(fde, (off_t)0, 1);	/* record position in file */
	while (1)
	     {
		errno = 0;
		if (-1 == lseek(fde, (off_t)mid, 0))
		  {
			if (verbo) fprintf(stderr,
		 "Unable to seek to the testing point in %s; errno = %d\n",
			 fnam, errno);
			return retcompare(fde, -6, wasat);
		  }
		if (128 != (ngot = read(fde, (void *)test, 128)))
		  {
			if (verbo) fprintf(stderr,
			 "Tried to read 128 bytes from %s, got %d\n",
				fnam, ngot);
			return retcompare(fde, -7, wasat);
		  }
		meanodd = meaneven = 0;
		for (i = 0, cp = test; i < ngot; i++, cp++)
		   {
			if (*cp)
			  {
				if (i % 2)
				  {
					nodd++; meanodd += (xgulong)(*cp);
				  }
				else {
					neven++; meaneven += (xgulong)(*cp);
				     }
			  }
		   }
		nposs += ngot;
		if (nodd + neven > 256) break;
		mid = mid / 2 + sizfil / 3;
		if (mid % 2) mid++;
		if (mid % 4) mid += 2;
		if (mid >= sizfil - 128) break;
	     }
	neach = nposs / 2;
	if ((nodd < 1) && (neven < 1)) return retcompare(fde, 0, wasat);
	meanodd /= neach; meaneven /= neach;
	if (verbo) printf(" %8d/%4d %7.3f %11d/%4d %7.3f %s\n",
		nodd, nposs/2, meanodd, neven, nposs / 2, meaneven, fnam);
	if (meanodd > 2. * meaneven) return retcompare(fde, 1, wasat);
	else if (meaneven > 2. * meanodd) return retcompare(fde, 2, wasat);
	else return retcompare(fde, 3, wasat);
}

int try_reindex(short vecre[], float matre[3][3])
{	/* This attempts to set up the rotation matrix and the translation
	vector to be applied to all the index vectors.  It returns 1 if the
	attempt is successful, 0 otherwise. */
	register int	nlines;
	char		rline[54];
	FILE		*fprei;

	if (NULL == (fprei = xgopen("reindex", "r"))) return 0;
	for (nlines =0; nlines < 3 &&NULL != fgets(rline,50,fprei); nlines++)
		(void)sscanf(rline,"%f %f %f %hd",
			&matre[nlines][0], &matre[nlines][1],
			&matre[nlines][2], &vecre[nlines]);
	(void)fclose(fprei);	return (nlines >= 3);
}

/* end of xgenfile.c */
