/* xgengeom.c -- library routines associated with detector geometry

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	20 Apr 2007: bugfix in get_3dprof: call managew3 appropriately
	19 Apr 2007: float->double conversions; reorganize use of "{"
	27 Dec 2006: manageice, opno & 010: require > 0 icering ranges
	 in order to avoid using the default ice-rings
	28 Nov 2006: makeq21 calls updated in managedv
	 2 Nov 2006: enable default icerings (3.97-3.6A, for example)
	24 Jun 2006: include xgmodel.h; that's where the structures live
	17 Jun 2006: replace act_pixel with activmask call
	13 Jun 2006: merge geometry routines from xgenmisc,threedpro,xgenmask.c
 Changes to previous files:
	 1 Jun 2006: norm99 and mygetwd moved elsewhere; housekeeping
	24 Feb 2006: in kablim, check detector width for sensibility
	17 Dec 2005: replace static global with managedv calls
	29 Nov 2005: inline argument declarations and strict prototyping
	20 Nov 2005: inline argument declarations and strict prototyping
	30 Oct 2005: minimize use of static globals
	25 Oct 2005: squeeze all static globals into a WORK3D structure
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	 6 Sep 2004: cosmetic changes
	 2 Sep 2004: REFPARS-related routines and lmat... moved to xgparams.c
	28 Aug 2004: routines used only here are declared as static; cosmetics
	26 Aug 2004: routines used only here are declared as static; cosmetics
	24 Aug 2004: eliminate unused routines loc3d() and prof_3d21d()
	17 Aug 2004: getq and findq moved to mrutil.c
	11 Aug 2004: cosmetic changes; bizarre bug in calc_derivs() fixed
	 6 Aug 2004: set n_iceranges in get_iceranges(); cosmetic cleanup
	 2 Aug 2004: add new arg to find_dimension()
	27 Jul 2004: dimscale() added: this computes scale factors for
	 detector dimensions so they can be stored as signed 16-bit integers
	 at maximum precision without overflowing. Use this in get_upars().
	10 Jul 2004: identify argument in force_populate() as being a MAIL *
	27 Jun 2004: cosmetic changes; tell_dvcoef() moved to ascout.c
	22 Jun 2004: move presval() and pdetec_limit here (later to xgpredict)
	21 Mar 2004: write force_populate() as wrapper on managedv
	18 Feb 2004: change kill_ftab calls so that they don't have args
	 1 Feb 2004: correct diagnostics from bstoloc()
	24 Jan 2004: zmin,zmax,nonpage made static locals in specific routines
	24 Jan 2004: variable zm in meanprofile() renamed to zmt
		so it doesn't overlap with the static global zm.
	21 Jan 2004: license message inserted
	 8 Jan 2004: global variables replaced with REFPARS structure elements;
		populate_xhd() (later managedv) written to fill those;
		kablim() takes additional arguments so it doesn't need globals
	 7 Jan 2004: REFPARS structures and args to routines replace globals;
		refparin() handle UPARAMS input
		makeq21() handles tilt & swing matrix
	26 Nov 2003: replace global variables toptrident, etc. with elements
		of a friendly BOUNDARY structure; don't need deltaom,deltachi
	 3 Nov 2003: moved lattice-character routines to xgbravais.c;
		numerous routines moved from ../xgrefine/refxtal.c to here
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	18 Aug 2003: created from xgenio.c
	13 Aug 2003: getregion() moved from ../integrate/integmask.c to 
		../xglib/xgenmask; some external symbols therefore are here
	 5 Jan 2003: reorganized source-file header
	 1 Nov 2002: use of REGIONPROFILE structures for reflection-box sizes;
		MASKRESCALE, several other defines moved to ../include/xgen.h;
		model-profile boundaries treated as variables, not as constants
	30 Sep 2002: ixy2ev(), tell_dvcoef(), reg2hab(), ikabc(), kablim()
		inserted here; allow manipulation of Kabsch profiles
	25 Mar 2002: removed MSI attachments
	 1 Jul 1999: in qerror(), get rid of the call to abort();
		it was probably a mistake all along
	24 Aug 1995: image data read as ints, not u_shorts
	21 Jul 1995: initial version of threedpro.c, culled from elsewhere.
	31 May 1995: "register" -> "register int" throughout
		re-cast several memory-handling routines with size_t's
	20 May 1995: slash/star removed from interior of comment
	 2 Mar 1995: delete line about file "polymath.c"
	27 Oct 1994: routine "mygetwd" added: wrapper for getwd, getcwd
	14 Jun 1994: bugfix for vectowmp if bz < 7
	 5 May 1994: initial version of xgenio.c, cobbled from other files
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version of older files

   Purpose: dealing with detector geometry issues, including 3-D profiles.
   Cross references: xgenmask.c, image.c, ../prerefine/spotshoes.c,
	../displays/xgimage.c, ../curses/ew_mask.c, ../integrate/integmask.c
*************************************************************************/
#define		NEED_SSDEF	1
#define		NEED_DESCRIP	1
#define		NEED_UNISTD	1
#include	"xgmodel.h"
#define		RESCALER	(62.)
#define		MIN_SIZ		256
#define		MIN_SIZ3	2048
#define		HUGEPIXEL	16000
#define		SHAVEXY		(0.005)
#define		SHAVEOM		(0.1)

 typedef struct	{
		char		*w3_currd;
		size_t		w3_locb3s, w3_locbsiz, w3_need3s;
		int		w3_jmax, w3_jmin, w3_maxbg, w3_minbg;
		int		w3_imwid, w3_imht, *w3_locb, *w3_locct;
		} WORK3D;

 typedef struct {
		double		df_gamco, df_qsp[9], df_exbv[3], df_kvb[3];
		REFPARS		*df_xp;
		} DVECWORK;

BOUNDARY *make_boundaries(int beamx, int dwidth, int dheight)
{ /*	Figure out where to draw the boundaries among the twelve regions.
	The general formula is to put vertical boundaries at 160 and 352
	regardless, and to make the horizontal boundaries depend on where the
	direct beam is. If it's near the center, the major boundaries are
	drawn at 190 and 320, and the minor boundaries at 110, 255, and 400.
	If it's off to one side or the other, the boundaries are adjusted to
	allow for that: in general, the region around the rotation axis has
	narrower X ranges, and the region away from the rotation axis has
	broader X ranges. */
	int		i;
	BOUNDARY	*bdp;
	double		rw, rh;
	static double	hlim[] =   { 0.3515, 0.6485 } ;
	static double	hfrac[] = {	0.1250, 0.2500, 0.4100, 0.5625, 0.7150,
				0.2200, 0.3750, 0.5000, 0.6250, 0.7800,
				0.2850, 0.4375, 0.5900, 0.7500, 0.8750 } ;

	rw = (double)dwidth;	rh = (double)dheight;
	if (NULL == (bdp = (BOUNDARY *)callop((size_t)1, sizeof (BOUNDARY),
		"bdp in make_boundaries"))) return NULL;
	bdp->toptri = hlim[0] * rh;	bdp->bottri = hlim[1] * rh;
	i = (beamx <= hlim[0] * rw ? 0 : (beamx >= hlim[1]  * rw ? 10 : 5 ));
	bdp->l1tri = hfrac[i++] * rw;	bdp->leftri = hfrac[i++] * rw;
	bdp->l2tri = hfrac[i++] * rw;	bdp->rightri = hfrac[i++] * rw;
	bdp->l3tri = hfrac[i] * rw;
	return bdp;
}

int getregion(short ix, short iy, /* float iphi, */ BOUNDARY *bdp)
{ /*	This assigns a peak to a region on the detector face.
	Currently we assign a peak to one of 12 regions: 3 cols x 3 rows,
	with the center-most row split vertically in each cell. Thus:
		 ___________________
		/     |       |     \
	       /   0  |   1   |   2  \
	      /-------|-------|-------\
	     /    |   |   |   |   |    \
	    (   3 | 4 | 5 | 6 | 7 |  8  )
	     \    |   |   |   |   |    /
	      \---|---|---|---|---|---/
	       \      |       |      /
		\  9  |   10  |  11 /
		 -------------------
  In this version we feed the routine the actual pixel coordinates rather
  than having them scaled up by xscal and yscal */

	register int	region, increment;
	double		rx, ry;

	rx = ix;	ry = iy;
	region = ry < bdp->bottri ? ((ry < bdp->toptri) ? 0 : 3) : 9;
	increment = 1 + (region == 3);
	region += increment * ((rx < bdp->leftri) ? 0 : (rx < bdp->rightri ?
		1 : 2));
	if (increment == 2)	/* fine-tuning in center section */
	  {
		if ((region == 3) && (rx > bdp->l1tri)) region++;
		else if ((region == 5) && (rx > bdp->l2tri)) region++;
		else if (rx > bdp->l3tri) region++;
	  }
	return region;
}

/* Routines associated with looking for ice-rings */

ICEWORK *manageice(int opno)
{ /* This manages ice-ring memory and its contents.
	If opno == 0, it deletes memory for the ice-ring data
		and returns NULL.
	If opno == 1, it creates memory for the ice-ring structure
		and returns it.
	If opno == 3, it returns a pointer to the icering structure.
	If opno == 7, it reads data in and fills up the ice ranges,
		and returns a pointer to the ice range data
	If opno == 11 == 013, it creates the ice-range list based on
		pre-set resolution exclusions rather than a file. */
	int		i, n_ice;
	char		icel[50];
	double		lfma, lfmi, t, *lofi, *loma;
	FILE		*fpic;
	static ICEWORK	*icp;
	static char	icpmsg[] = "icp in manageice";
	static char	icprmsg[] = "icp->icw_ramax in manageice";
	static int	permn = 4;
	 /* list based on Garman & Schneider, J.Appl.Cryst. 30:211 (1997).
	  I've merged together [100] with [002] and cut it off at 2.1 A */
	static double	permring[] =
	 { 3.97, 3.60, 3.48, 3.40, 2.70, 2.63, 2.28, 2.21 };

	if (opno == 0)	/* eliminate old lists */
	  {
		if (NULL == icp) return NULL;
		memzap((AVOID **)&(icp->icw_ramax), icprmsg);
		memzap((AVOID **)&icp, icpmsg);
	  }
	if (opno == 1) /* create memory for new list */
	  {
		(void)manageice(0);
		return (ICEWORK *)callop(1, sizeof (ICEWORK), icpmsg);
	  }
	if (NULL == icp) icp = manageice(1);
	if (!(opno & 012)) return icp;
	if (opno & 04)
	  { /* open the file with the list of ice-ring resolution ranges */
		if (NULL == (fpic = xgopen("ICERING", "r")))
			return manageice(0);
		/* count how many lines it contains */
		for (i = 0; NULL != fgets(icel, 48, fpic); ) i++;
		/* if there are any lines, rewind file to read it fully */
		if ((i < 1) || (-1 == fseek(fpic, 0L, 0)))
		  {
			(void)fclose(fpic);	return manageice(0);
		  }
		n_ice = i;	/* remember how many ice-ring ranges are there*/
		/* reserve memory for the two vectors of ice ranges*/
		memzap((AVOID **)&(icp->icw_ramax), icprmsg);
		if (NULL == (icp->icw_ramax = (double *)callop(2 * n_ice,
		 sizeof (double), icprmsg)))
		  {
			(void)fclose(fpic);	return manageice(0);
		  }
		icp->icw_ramin = icp->icw_ramax + n_ice;
		loma = icp->icw_ramax;	lofi = icp->icw_ramin;
		/* now read the file and remember ice-ring resolution ranges */
		for (i = 0; i < n_ice; i++, loma++, lofi++)
		   {
			if (NULL == fgets(icel, 48, fpic)) break;
			sscanf(icel, "%lf %lf", &lfma, &lfmi);
			/* reverse min-D and max-D if they're backward */
			if (lfma < lfmi)
			  {
				t = lfma; lfma = lfmi; lfmi = t;
			  }
			*loma = lfma;	*lofi = lfmi;	/* remember 'em */
		   }
		icp->icw_nranges = i;	/* count # of good ice ranges here */
		(void)fclose(fpic);	fpic = NULL;
	  }
	else if (opno & 010)
	  { /* in this option, we return the current table
		 if it exists and has at least one element. */
		if (icp->icw_nranges > 1) return icp;
		/* otherwise, we try to fill it from a file */
		if ((NULL != (icp = manageice(7))) && (icp->icw_nranges > 0))
			return icp;
		/* if that doesn't work either, we fill it from a fixed table*/
		n_ice = permn;
		if (NULL == (icp->icw_ramax = (double *)callop(2 * n_ice,
		 sizeof (double), icprmsg))) return manageice(0);
		icp->icw_ramin = icp->icw_ramax + n_ice;
		loma = icp->icw_ramax;	lofi = icp->icw_ramin;
		for (i = 0; i < n_ice; i++, loma++, lofi++)
		   {
			*loma = permring[2*i]; *lofi = permring[2*i+1];
		   }
		icp->icw_nranges = n_ice;
	  }
	return icp;
}

void describe_rings(FILE *fp)
{
	int	i;
	ICEWORK	*icp;

	if (NULL == (icp = manageice(3)))
	  {
		if (NULL == (icp = manageice(7))) return;
	  }
	if (icp->icw_nranges < 1) return;
	if (icp->icw_nranges == 1) fprintf(fp,
 "Reflections will be excluded due to ice-ring overlap if %7.2f > D > %7.2f\n",
		 icp->icw_ramax[0], icp->icw_ramin[0]);
	else {
		fprintf(fp,
 "Reflections will be excluded due to ice-ring overlap in these D ranges:\n");
		for (i = 0; i < icp->icw_nranges; i++)
			fprintf(fp, " %7.2f > D > %7.2f%s",
			 icp->icw_ramax[i], icp->icw_ramin[i],
			 (i % 2) ? "\t": "\n");
		if (icp->icw_nranges % 2) fprintf(fp, "\n");
	     }
}

int isinring(double dspac)
{ /*	This returns 1 iff. a reflection with D-spacing `dspac' is contained
	in one of the resolution ranges forbidden due to overlap with an
	ice-ring */
	int	icerange;
	ICEWORK	*icp;

	if (NULL == (icp = manageice(3))) return 0;	/* no icerings */
	if ((dspac < DFUZZ) || (dspac >999.)) return 1; /*nonsense D rejected*/
	if ((icp->icw_nranges < 0) || (NULL == icp->icw_ramax)) return 0;
	for (icerange = 0; icerange < icp->icw_nranges; icerange++)
	   {
		if ((icp->icw_ramax[icerange] > dspac) &&
		 (dspac > icp->icw_ramin[icerange])) return 1;
	   }
	return 0;
}
/* end of list of routines associated with icerings */

double dimscale(short indimen)
{ /*	This returns a value that can be used in rescaling pixel X or Y values
	so that they (the pixel values) can be determined precisely even if
	they're stored as integers. The argument is a detector dimension
	value, and the output is a multiplier that will increase the scale of
	the X or Y value up to something that can still fit in a signed
	16-bit integer, i.e. a number less than 32767. */
	
	return (indimen < 600) ? 50. : ((indimen < 1200) ? 25. :
	 ((indimen < 3200) ? 10. : ((indimen < 6400) ? 5 : 2)));
}

static WORK3D *managew3(int opno, int xyzl[])
{ /*	This manages a buffer for 3-D vectors.
	If opno == 0, we eliminate memory for the buffer and its contents.
	If opno == 1, we create memory for the buffer itself.
	If opno == 3, we read in information that enables us to fill this.
	If opno == 7, we reserve memory for the individual vectors within
		the WORK3D structure, based on the limits xyzl[]. */

	int		dcz;
	static WORK3D	*w3p;
	static char	w3msg[] = "w3p in managew3";
	static char	w3cmsg[] = "w3p->w3_currd in managew3";
	static char	w3lbmsg[] = "w3p->w3_locb in managew3";
	static char	w3lcmsg[] = "w3p->w3_locct in managew3";

	if (opno == 0)
	  {
		if (NULL != w3p)
		  {
			memzap((AVOID **)&(w3p->w3_currd), w3cmsg);
			memzap((AVOID **)&(w3p->w3_locb), w3lbmsg);
			memzap((AVOID **)&(w3p->w3_locct), w3lcmsg);
			memzap((AVOID **)&w3p, w3msg);
		  }
		return NULL;
	  }
	if (opno == 01) w3p = managew3(0, xyzl); /* force new memory */
	if ((NULL == w3p) && (NULL == (w3p = (WORK3D *)callop(1,
	 sizeof (WORK3D), w3msg)))) return NULL;
	if ((!(opno & 04)) || (NULL == xyzl)) return w3p;
	dcz = (1 + xyzl[1] - xyzl[0]) * (1 + xyzl[3] - xyzl[2]);
	if ((NULL != w3p->w3_locb) && (w3p->w3_locbsiz < (size_t)dcz))
		memzap((AVOID **)&(w3p->w3_locb), w3lbmsg);
	if (NULL == w3p->w3_locb)
	  {
		w3p->w3_locbsiz = (dcz < MIN_SIZ) ? MIN_SIZ : dcz;
		if (NULL == (w3p->w3_locb = (int *)callop(w3p->w3_locbsiz,
		 sizeof (int), w3lbmsg))) return managew3(0, xyzl);
	  }
	w3p->w3_need3s = dcz * (1 + xyzl[5] - xyzl[4]);
	if ((NULL != w3p->w3_locct) && (w3p->w3_locb3s < w3p->w3_need3s))
		memzap((AVOID **)&w3p->w3_locct, w3lcmsg);
	if (NULL == w3p->w3_locct)
	  {
		w3p->w3_locb3s = (w3p->w3_need3s < MIN_SIZ3) ?
			MIN_SIZ3 : w3p->w3_need3s;
		if (NULL == (w3p->w3_locct = (int *)callop(w3p->w3_locb3s,
		 sizeof (int), w3lcmsg))) return managew3(0, xyzl);
		if (NULL == (w3p->w3_currd = (char *)callop(w3p->w3_locb3s,
		 sizeof (char), w3cmsg))) return managew3(0, xyzl);
	  }
	return w3p;
}

void kill3d(void)
{ /*	wrapper for managew3(0, NULL) that doesn't need WORK3D */
	(void)managew3(0, NULL);
}

int centolim(int xce, int yce, int zce, int xyzl[])
{ /* This establishes limits around the center position in (X,Y,Z).
	It returns EOK successful, EALLOC or EREADERR if not */
	int		lx, ly, lz, lx2, ly2, lz2, detdim[2];
	int		lxmin, lxmax, lymin, lymax, lowid, loht;
	FRAMEINDEX	*lfmi;
	static int	tdet, dez, dey, dex, dez2, dey2, dex2, frmi, frma;
	static int	zmin, zmax;
	static WORK3D	*w3p;

	lx = BIGSPOTWID;	ly = BIGSPOTHT;	lz = PFRAMES;
	lx2 = lx / 2;	ly2 = ly / 2;	lz2 = lz / 2;
	if (0 >= tdet) /*establish detector type & absolute limits if needed*/
	  {
		if (0 >= (tdet = (dettype(detdim) & 0xFF)))
			return EREADERR;
			/* electronic detectors & CCDs */
		if ((tdet < 5) || ((10 <= tdet) && (tdet <= 11)))
		  {
			dez = lz2;
			if (tdet == 1)
			  {
				dex = lx2 - 2; dey = ly2 - 2;
			  }
			else {
				dex = lx2;	dey =  ly2;
			     }
		  }
		else { /* image-plate case */
			dez = lz2 - 1;	dex = lx2 + 2; dey = ly2 + 2;
		     }
		dez2 = dez + dez;	dey2 = dey + dey; dex2 = dex + dex;
		if (frma <= 0)
		  {
			if (NULL == (lfmi = ftables(3))) return EBADVALUE;
			frmi = lfmi->fi_firstfm;
			frma = lfmi->fi_lastfm;
		  }
	  }
	zmin = zce - dez;	if (zmin < frmi) zmin = frmi;
	zmax = zmin + dez2;	if (zmax > frma) zmax = frma;
	xyzl[4] = zmin;		xyzl[5] = zmax;
	lymin = yce - dey;	if (lymin > HUGEPIXEL) lymin = 0;
	lymax = lymin + dey2;
	lxmin = xce - dex;	if (lxmin > HUGEPIXEL) lxmin = 0;
	lxmax = lxmin + dex2;
	xyzl[0] = lxmin;	xyzl[1] = lxmax;
	xyzl[2] = lymin;	xyzl[3] = lymax;
	/* we need to know the detector dimensions to do
	 some limit checks on the upper side. Try to get them.
	 If we fail, we'll have the limits right most of the time anyway */
	if (NULL == (w3p = managew3(3, xyzl))) return EALLOC;
	if (0 == w3p->w3_imwid)
	  {
		if (EOK != find_dimension(&lowid, &loht, 0, NULL))
			return EREADERR;
		w3p->w3_imwid = lowid;	w3p->w3_imht = loht;
	  }
	/* If we got here, we can apply these limit checks. */
	if (lymin >= (w3p->w3_imht  - dey2)) /* limits on height */
	  {
		xyzl[2] = lymin = w3p->w3_imht  - dey2 - 1;
		xyzl[3] = lymax = lymin + dey2;
	  }
	if (lxmin >= (w3p->w3_imwid - dex2)) /* limits on width */
	  {
		xyzl[0] = lxmin = w3p->w3_imwid - dex2 - 1;
		xyzl[1] = lxmax = lxmin + dex2;
	  }
	return EOK;
}

int get_3dprof(FRAMEINDEX *lfmi, int imfla, int *xyzl, MAIL *bgmp)
{ /*	This constructs the 3-dimensional profile in a region defined by the
	vector `xyzl', i.e.	xyzl[0] <= x <= xyzl[1],
 				xyzl[2] <= y <= xyzl[3],
 				xyzl[4] <= x <= xyzl[5].
	The resulting profile lives in the vector `w3p->w3_locct'.
	It subtracts background values from the background image `bgmp' if
	it is non-NULL before doing its calculations.  The routine returns EOK
	if there is intensity to display; it returns EBADVALUE if it encounters
	only zeroes; it returns -1 on memory errors. A special-case call to
	this routine: if the vector `xyzl' is NULL, it simply gets rid of memory
	for the local intensity and display arrays and then returns EOK. */

	register int	x, y, z;
	int		dcz, thisi, *ip, *loi, *loiz;
	char		*lfname;
	double		bgf;
	static MAIL	locm;
	static WORK3D	*w3p;

	if (NULL == xyzl)
	  {
		(void)managew3(0, NULL);	return EOK;
	  }
	if ((NULL == w3p) && (NULL == (w3p = managew3(3,xyzl)))) return EALLOC;
	if (NULL == bgmp)
	  {
		w3p->w3_maxbg = 0;	w3p->w3_minbg = 65534;
	  }
	zeromem((void *)&locm, sizeof (MAIL));
	if (NULL == (w3p = managew3(7, xyzl))) return EBADVALUE;
	zeromem((void *)w3p->w3_locct, w3p->w3_need3s * sizeof (int));
	w3p->w3_jmin = 65532;	w3p->w3_jmax = -w3p->w3_jmin;
	if (NULL == lfmi)
	  {
		fprintf(stderr, "Frame index unavailable at this point\n");
		return EBADVALUE;
	  }
	dcz = (1 + xyzl[1] - xyzl[0]) * (1 + xyzl[3] - xyzl[2]);
	for (z = xyzl[4], loiz = w3p->w3_locct; z <= xyzl[5]; z++, loiz += dcz)
	   {
		zeromem((void *)loiz, dcz * sizeof (int));
		if (NULL == (lfname = indexname((int)z)))
			fprintf(stderr,"No name available for image %3d\n", z);
		else if (FRAME_OK == (y = read_box(lfname,
		 imfla, &locm, xyzl, w3p->w3_locb)))
		  {
			locm.frameno = z;
			if (NULL != bgmp)
			  locm.count0 = 16. * ((double)bgmp->elap100ths) /
			  ((double)locm.elap100ths);
			ip = w3p->w3_locb;
			for (y = xyzl[2], loi = loiz; y <= xyzl[3]; y++)
			 for (x = xyzl[0]; x <= xyzl[1]; x++, ip++, loi++)
			    {
				thisi = *ip;
				if (NULL != bgmp)
				  {
					bgf = (16. / (double)locm.count0) *
					 (double)imvalue(bgmp, (int)x, (int)y);
					if (bgf < w3p->w3_minbg)
						w3p->w3_minbg = bgf;
					if (bgf > w3p->w3_maxbg)
						w3p->w3_maxbg = bgf;
					thisi -= (int)bgf;
				  }
				if (thisi < w3p->w3_jmin) w3p->w3_jmin = thisi;
				if (thisi > w3p->w3_jmax) w3p->w3_jmax = thisi;
				*loi = thisi;
			   }
		  }
		else	fprintf(stderr, "Image %s unreadable: error %3d\n",
				lfname, y);
	   }
	memzap((AVOID **)&(locm.runheader), "mpo->runheader in get_3dprof");
	memzap((AVOID **)&(locm.overflow), "mpo->overflow in get_3dprof");
	if (w3p->w3_jmax <= w3p->w3_jmin)
	  {
		fprintf(stderr,
 "Bad profile : jmin,jmax,ranges=%5d,%5d,(%4d,%4d)(%4d,%4d),(%3d,%3d)\n",
		 w3p->w3_jmin, w3p->w3_jmax, xyzl[0], xyzl[1], xyzl[2], xyzl[3],
		 xyzl[4], xyzl[5]);
		return EBADVALUE;
	  }
	return EOK;
}

void bstoloc(float *fve, int xyzl[])
{ /*	This converts a floating point vector of values `fve' ranging over
	the usual limits xyzl into the local integer vector. It first
	creates the local vectors if needed. */
	register int	z;
	int		*ip;
	float		*flp;
	static WORK3D	*w3p;

	if (NULL == xyzl)
	  {
		(void)managew3(0, xyzl);	return;
	  }
	if (NULL == (w3p = managew3(7, xyzl))) return;
	w3p->w3_jmin = 32766;	w3p->w3_jmax = -w3p->w3_jmin;
	flp = fve;	ip = w3p->w3_locct;
	for (z = w3p->w3_need3s; z--; flp++, ip++)
	   {
		*ip = (*flp < 0.) ? *flp - 0.49999 : *flp + 0.49999;
		if (*ip < w3p->w3_jmin) w3p->w3_jmin = *ip;
		if (*ip > w3p->w3_jmax) w3p->w3_jmax = *ip;
	   }
	w3p->w3_minbg = 65535.;	/* flag to say "ignore background" */
}

void disp_3dprof(FILE *fp, int xyzl[])
{ /*	This displays the 3-D profile obtained in `get_3dprof'
	on the stream fp. The first time this is called within a single
	application there may be an extra form-feed early-on because of
	the variable `nonpage'; we'll see. */
	register int	x, y, z;
	char		dispct, *loch, *cp, *cpy, *cpz, kc, xbox[22];
	int		iinterval, *loi, ix, dcy, dcz, xwi;
	static WORK3D	*w3p;
	static char	xfmt[] = "%1d";
	static int	nonpage;

		/* don't try if we haven't reserved memory */
	if (NULL == (w3p = managew3(3, xyzl))) return;
	if ((NULL == w3p->w3_currd) || (NULL == w3p->w3_locct)) return;
	iinterval = (w3p->w3_jmax >= w3p->w3_jmin + RESCALER) ?
		(w3p->w3_jmax - w3p->w3_jmin) / RESCALER + 1 : 1;
	zeromem((void *)w3p->w3_currd, w3p->w3_need3s * sizeof (char));
	for (z = xyzl[4], loi = w3p->w3_locct, loch = w3p->w3_currd;
	 z <= xyzl[5]; z++)
	   {
		for (y = xyzl[2]; y <= xyzl[3]; y++)
		   for (x = xyzl[0]; x <= xyzl[1]; x++, loi++,loch++)
		      {
			dispct = (char)((double)((*loi) - w3p->w3_jmin) /
				(double)iinterval);
			if (dispct >= RESCALER - 1) *loch = 'Z';
			else if (dispct >= 35)	*loch = 'A' + (dispct - 35);
			else if (dispct >= 10)	*loch = 'a' + (dispct - 10);
			else if (dispct)	*loch = '0' + dispct;
			else			*loch = ' ';
		      }
	    }
	fprintf(fp,
 "\tMin,max counts = (%5d,%5d); interval between values=%3d counts",
		w3p->w3_jmin, w3p->w3_jmax, iinterval);
	if (w3p->w3_minbg < 65534.) fprintf(fp,
	 "; min,max bg = %5d, %5d", w3p->w3_minbg, w3p->w3_maxbg);
	fprintf(fp, "\n   ");
	/* display the z values in a row */
	xwi = (xyzl[1] + 1 - xyzl[0]) / 2 - 4;
	for (x = 0, cp = &(xbox[0]); x < xwi; x++, cp++) *cp = '_';
	*cp = '\0';
	for (z = xyzl[4]; z <= xyzl[5]; z++)
		fprintf(fp, " .%sFrame %3d%s.", xbox, (int)z, xbox);
	/* output the actual data and the y values */
	dcy = 1 + xyzl[1] - xyzl[0];
	dcz = dcy * (1 + xyzl[3] - xyzl[2]); 
	for (y = xyzl[2], cpy = w3p->w3_currd; y <= xyzl[3]; y++, cpy += dcy)
	   {
		if (!(y % 5)) fprintf(fp, "\n%4d|", (int)y);
		else fputs("\n   .|", fp);
		for (z = xyzl[4], cpz = cpy; z <= xyzl[5]; z++, cpz += dcz)
		   {
			if (z != xyzl[4]) fputs("|-|", fp);
			for (x = xyzl[0], cp = cpz; x <= xyzl[1]; x++,cp++)
				fputc((*cp), fp);
			if (z == xyzl[5]) fputs("|", fp);
		   }
	   }
	fputs("\n   ", fp);
	/* output the x values within each box */
	xbox[0] = ' ';	xbox[1] = '\\';
	for (x = xyzl[0], cp = xbox + 2; x <= xyzl[1]; x++, cp++) *cp = '.';
	*cp++ = '/';	*cp = '\0';
	for (ix = 0, x = xyzl[0], cp = xbox + 2; x <= xyzl[1]; x++, ix++, cp++)
	   {
		if (!(x % 5))
		  {
			y = (x < 10) ? 0 : (x < 100) ? 1 : (x < 1000) ? 2 : 3;
			if (ix + 1 > y)
			  {
				xfmt[1] = y + '1';		kc = *(cp+1);
				sprintf(cp - y, xfmt, x);	*(cp+1) = kc;
			  }
		  }
	   }
	for (z = xyzl[4]; z <= xyzl[5]; z++) fputs(xbox, fp);
	fputs("\n", fp);
	if (nonpage++ == 2)
	  {
		nonpage = 0;	fputs("\f", fp);
	  }
	if (fp != stderr) xeflush(fp);
}

static DVECWORK *managedv(int opno, MAIL *mp)
{ /*	This manages the information needed for manipulating a DVECWORK
	structure according to opno:
	opno == 0:	Delete memory for static global *dwp
	opno == 1:	unconditionally create and empty one and return it
	opno == 3:	return the current dwp without checking it.
	opno == 7:	return dwp after checking to see if it's populated;
			 if not, fill it up from a current UPARAMS file or a
			 MAIL structure
	opno == 017:	force repopulation of the DVECWORK structure
  */
	int		nbr;
	double		zz, *qq;
	REFPARS		*xp;
	DATARUN		*ruhd;
	static DVECWORK	*dwp;
	static char	dwxpmsg[] = "df_xp in managedv";
	static char	dwmsg[] = "dwp in managedv";

	if (opno == 0) /* kill memory for this stuff */
	  {
		if (NULL != dwp)
		  {
			memzap((AVOID **)&(dwp->df_xp), dwxpmsg);
			memzap((AVOID **)&dwp, dwmsg);
		  }
		return NULL;
	  }
	if (opno == 1) /* just return space, clearing the old one if needed */
	  {
		if (NULL != dwp)
		  {
			memzap((AVOID **)&(dwp->df_xp), dwxpmsg);
			memzap((AVOID **)&dwp, dwmsg);
		  }
	  }
	if (NULL == dwp) /* otherwise, get memory for it */
	  {
		if (NULL == (dwp = (DVECWORK *)callop(1, sizeof (DVECWORK),
		 dwmsg))) return NULL;
		if (NULL == (dwp->df_xp = (REFPARS *)callop(1,
		 sizeof (REFPARS), dwxpmsg)))
		  {
			memzap((AVOID **)&dwp, dwmsg);
			return NULL;
		  }
	  }
	if (!(opno & 04)) return dwp; /* just give back the current pointer */
	/* Attempt to find a UPARAMS file and return its contents,
	if it exists, thoughtfully remembering to transpose the unit cell
	matrices and populates the dwp->df_qsp vector along the way.
	If that fails, it looks at the open detector image header mp
	to try to populate the REFPARS structure */
	/* if it's appropriately populated already, quit unless 010 is on */
	xp = dwp->df_xp;
	if ((opno & 010) && (NULL != xp)) xp->xplated = -1.;
	if ((NULL != dwp->df_xp) && (0 < (dwp->df_xp)->xplated))
	  {
		qq = dwp->df_qsp;
		for (zz = nbr = 0; nbr < 3; nbr++)
		   {
			zz += *qq * *qq + *(qq + 1) * *(qq + 1) +
				*(qq + 2) * *(qq + 2);
		   }
		if (zz < 0.8)
		  { /* cosines and sines of tilt and swing angles*/
			makeq21(xp);	qq = dwp->df_qsp;
			*qq++ = xp->xq21[0][0];	*qq++ = xp->xq21[0][1];
			*qq++ = xp->xq21[0][2];	*qq++ = xp->xq21[1][0];
			*qq++ = xp->xq21[1][1];	*qq++ = xp->xq21[1][2];
			*qq++ = xp->xq21[2][0];	*qq++ = xp->xq21[2][1];
			*qq++ = xp->xq21[2][2];
		  }
		return dwp;
	  }
	if (EOK != refparin(xp, &nbr)) /* can we get it from UPARAMS? */
	  { /* No: try to read it from the detector image */
		if (NULL != mp)
		  {
			if ((NULL != (ruhd = mp->runheader)) &&
			 (0.1 < ruhd->run_d0pos[0]))
			  {
				xp->xplated = ruhd->run_d0pos[0];
				xp->xx0 = ruhd->run_d0pos[1];
				xp->xy0 = ruhd->run_d0pos[2];
				xp->xswing = -ruhd->run_d0ang[0] * RADC;
				xp->xtilt = 0;
				xp->xfrmsize = ruhd->run_stepreal * RADC;
				xp->xlambda = ruhd->run_lambda;
			  }
			xp->xdetwid = mp->xdimen;
			xp->xdetht  = mp->ydimen;
		  }
		else {
			xp->xplated = 12.;	xp->xfrmsize = RADC;
			xp->xswing = xp->xtilt = xp->xx0 = xp->xy0 = 0;
			xp->xfrmsize = RADC; /* wild guesses */
		     }
		xp->xpltd2 = xp->xplated * xp->xplated;
	  }
	else if (NULL != mp) /* dimensions not always present */
	  {
		xp->xdetwid = mp->xdimen;	xp->xdetht = mp->ydimen;
	  }
	/* cosines and sines of tilt and swing angles */
	makeq21(xp);	qq = dwp->df_qsp;
	*qq++ = xp->xq21[0][0];	*qq++ = xp->xq21[0][1];	*qq++ = xp->xq21[0][2];
	*qq++ = xp->xq21[1][0];	*qq++ = xp->xq21[1][1];	*qq++ = xp->xq21[1][2];
	*qq++ = xp->xq21[2][0];	*qq++ = xp->xq21[2][1];	*qq++ = xp->xq21[2][2];
	dwp->df_xp = xp;
	return dwp;
}

int force_populate(MAIL *mp)
{ /*	bookkeeping routine: calls managedv with the "force reread" bit on */
	return (NULL == managedv(017, mp)) ? EREADERR : EOK;
}

void nomoredv(void)
{ /*	Wrapper for managedv(0, NULL). If we exported the DVECWORK structure
	type, this routine could disappear */
	(void)managedv(0, NULL);
}

int ixy2ev(int ix, int iy, double *ev)
{ /*	This obtains the un-rotated exit-beam vector associated with the pixel
	position (ix, iy).  If an error occurs, the routine returns an error
	value and returns zeroes in the 3-element-vector `ev'.  If successful,
	it returns EOK and returns the exit beam in ev[0]-[2] */
	double		rni, rni2, sr0, sr1, sr2, xr, yr, *qq;
	float		xc, yc;
	DVECWORK	*dwp;
	REFPARS		*xp;

	xr = ix;	yr = iy;
	if (NULL == (dwp = managedv(03, NULL)))
	  {
		if (NULL == (dwp = managedv(07, NULL))) return EREADERR;
	  }
	xp = dwp->df_xp;
	if (1 <= pix2cm((float)xr, (float)yr, &xc, &yc, 0))
	  {
		xc -= xp->xx0;	yc -= xp->xy0;
		rni2 = xc * xc + yc * yc + xp->xpltd2;	rni = sqrt(rni2);
		sr0 = xp->xplated / rni; sr1 = xc / rni;	sr2 = yc / rni;
		/* rotate this by the detector tilt and swing-angle matrices */
		qq = dwp->df_qsp;
		ev[0] =       *qq * sr0 + *(qq + 1) * sr1 + *(qq + 2) * sr2;
		ev[1] = *(qq + 3) * sr0 + *(qq + 4) * sr1 + *(qq + 5) * sr2;
		ev[2] = *(qq + 6) * sr0 + *(qq + 7) * sr1 + *(qq + 8) * sr2;
		return EOK;
	  }
	ev[0] = ev[1] = ev[2] = 0.;	return EBADVALUE;
}

int reg2hab(int ixpx, int iypx, double *ov)
{ /*	This computes the kabsch alpha and beta values for a pixel. */

	double		lv0, lv1, lv2, *flp;
	DVECWORK	*dwp;

	if ((NULL == (dwp = managedv(03, NULL))) &&
		(NULL == (dwp = managedv(07, NULL)))) return -2;
	if (EOK != ixy2ev(ixpx, iypx, &(ov[2]))) return -1;
	lv0 = ov[2];	lv1 = ov[3];	lv2 = ov[4];	flp = dwp->df_kvb;
	*ov = *(flp + 2) * lv0 - *flp * lv2;
	*(ov + 1) = *flp * lv0 + *(flp + 1) * lv1 + *(flp + 2) * lv2;
	return 0;
}

int ikabc(double fx, double fy, double ikout[])
{ /*	This computes some parameters needed for diffraction-vector
	profile analysis:
	the coordinates of the exit-beam vector dwp->df_exbv (Kabsch's S),
	the unit vectors kva, kvb, kvc (Kabsch's a, b, and c) and the scalar
	dwp->df_gamco = (kvc . ((u X x)) / |x|, where u = unit vector along
	rotation axis and x = diffraction vector.  This last constant is the
	multiplier for the sin(delta_omega) values in calculating
	the gamma angle. The routine returns 1 iff. successful.  */
	int		i, ix, iy;
	double		u00, u01, u10, u11, dlx, dlxp, dly, dlyp;
	double		den0, den1, *d00, *d01, *d10, *d11;
	double		exc0[4], exc1[4], exc2[4], exc3[4];
	DVECWORK	*dwp;

	/* find the upper lefthand corner of the box near this point */
	ix = (int)fx;	iy = (int)fy;
	dlx = fx - ix;	dlxp = 1. - dlx; /* distances from point to corners */
	dly = fy - iy;	dlyp = 1. - dly;
	/* compute weight factors associated with the 4 corners */
	u00 = dlxp * dlyp;	u01 = dlxp * dly;
	u10 = dlx * dlyp;	u11 = dlx * dly;
	if ((NULL == (dwp = managedv(03, NULL))) &&
		(NULL == (dwp = managedv(07, NULL)))) return 0;
	/* determine the exit-beam vectors in the integer pixels
	 surrounding the current point */
	i = ((EOK == ixy2ev(ix, iy, exc0)) && (EOK == ixy2ev(ix,iy +1,exc1)) &&
	 (EOK == ixy2ev(ix + 1, iy, exc2)) && (EOK == ixy2ev(ix+1,iy+1,exc3)));
	if (!i) return 0;
	/* interpolate the diffraction-vector values appropriate to it */
	for (i = 0, d00 = exc0, d01 = exc1, d10 = exc2, d11 = exc3;
	 i < 3; i++, d00++, d01++, d10++, d11++) dwp->df_exbv[i] =
		*d00 * u00 + *d01 * u01 + *d10 * u10 + *d11 * u11;
	/* now determine the diffraction-vector coefficients themselves */
	if ((den0 = 1. - dwp->df_exbv[0] * dwp->df_exbv[0]) < DFUZZ) return 0;
	if ((den1 = 1. - dwp->df_exbv[1] * dwp->df_exbv[1]) < DFUZZ) return 0;
	dwp->df_kvb[0] = -dwp->df_exbv[0] * dwp->df_exbv[1];
	dwp->df_kvb[1] = den1;
	dwp->df_kvb[2] = -dwp->df_exbv[1] *dwp->df_exbv[2];
	den0 = sqrt(den0); den1 = sqrt(den1);
	dwp->df_gamco = -dwp->df_exbv[1] / den0;
	for (i = 0; i < 3; i++)
	   {
		dwp->df_exbv[i] /= den1;		dwp->df_kvb[i] /= den1;
		ikout[i] = dwp->df_exbv[i];	ikout[i + 3] = dwp->df_kvb[i];
	   }
	ikout[6] = dwp->df_gamco;	return 1;
}

void kablim(MAIL *mailp, REGIONPROFILE *lrgb, int currbz)
{ /*	This computes the minimum and maximum possible values for
	kalpha, kbeta, and kgamma by looking at various places on the
	detector and calculating the values around noted points. */
	register int	x, y;
	int		lx, ly, lx2, ly2, lz, mx, my, mz;
	int		xmi, ux, uy, xma, ymi, yma;
	int		xdel, ydel;
	double		deka, dekb, dekg, fxp, fyp, half, kal, kbe, kga;
	double		maka, makb, makg, mika, mikb, mikg, iko[7], kaf[6];
	REFPARS		*xp;
	REGIONPROFILE	*rgp;
	DVECWORK	*dwp;

	/* we attempt to read plateD, swing, etc. if necessary */
	mika = mikb = mikg = 1.e30;	maka = makb = makg = -1.e30;
	if (NULL != lrgb)
	  {
		lx = lrgb->m3_bx; ly = lrgb->m3_by;	lz = lrgb->m3_bz;
		mx = lrgb->m3_qx; my = lrgb->m3_qx;	mz = lrgb->m3_qz;
	  }
	else {
		lx = BIGSPOTWID;	ly = BIGSPOTHT;	lz = currbz;
		mx = 13;	my = 13;	mz = 9;
	     }
	lx2 = lx / 2; ly2 = ly / 2;
	xdel = lx;	ydel = ly;	half = 0.5 * (double)lz;
	if (((NULL == (dwp = managedv(03, mailp))) ||
	 (10 > (dwp->df_xp)->xdetwid)) &&
	 (NULL == (dwp = managedv(07, mailp)))) return;
	xp = dwp->df_xp;
	for (x = xdel / 2; x < xp->xdetwid - xdel / 2; x += xdel)
	   {
		fxp = x; xmi = x - lx2; xma = xmi + lx;
		for (y = ydel/2; y < xp->xdetht - ydel/2; y += ydel)
		   if (NULL != activmask(x, y, 041))
		     {
			fyp = y;
			if (ikabc(fxp, fyp, iko))
			  {
			    kga = fabs(sin(half * xp->xfrmsize) *
				dwp->df_gamco);
			    if (-kga < mikg) mikg = -kga;
			    if ( kga > makg) makg =  kga;
			    ymi = y - ly2; yma = ymi + ly;
			    for (ux = xmi; ux <= xma; ux += lx2)
			     for (uy = ymi; uy <= yma; uy += ly2)
				if (!reg2hab(ux, uy, kaf))
				  {
					kal = kaf[0];	kbe = kaf[1];
					if (kal < mika) mika = kal;
					if (kal > maka) maka = kal;
					if (kbe < mikb)  mikb  = kbe;
					if (kbe > makb)  makb  = kbe;
				  }
			  } /* end of "if (ikabc" clause */
		   } /* end of loop through y */
	   } /* end of loop through x */
	/* shave a little bit off all of them */
	kal = (1. - SHAVEXY) * mika + SHAVEXY * maka;
	maka = (1. - SHAVEXY) * maka + SHAVEXY * mika;	mika = kal;
	kbe = (1. - SHAVEXY) * mikb + SHAVEXY * makb;
	makb = (1. - SHAVEXY) * makb + SHAVEXY * mikb;	mikb = kbe;
	kga = (1. - SHAVEOM) * mikg + SHAVEOM * makg;
	makg = (1. - SHAVEOM) * makg + SHAVEOM * mikg;	mikg = kga;
	/* define the intervals between diffraction-vector pixels */
	deka = (maka - mika) / ((double)(mx - 1)); 
	dekb = (makb - mikb) / ((double)(my - 1)); 
	dekg = (makg - mikg) / ((double)(mz - 1));
	if (NULL != lrgb)
	  {
		mz = lrgb->m3_nmodels;
		for (lz = 0, rgp = lrgb; lz < mz; lz++, rgp++)
		   {
			rgp->m3_lim[0] = mika;	rgp->m3_lim[1] = maka;
			rgp->m3_lim[2] = deka;
			rgp->m3_lim[3] = mikb;	rgp->m3_lim[4] = makb;
			rgp->m3_lim[5] = dekb;
			rgp->m3_lim[6] = mikg;	rgp->m3_lim[7] = makg;
			rgp->m3_lim[8] = dekg;
		   }
	  }
}

/* end of xgengeom.c */
