/* xgenmask.c -- routines for manipulating 3-D Bragg-reflection profiles

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	24 Jun 2006: include xgmodel.h; that's where the structures are
	12 Jun 2006: move all routines that use DVECWORKs to xgengeom.c
	 1 Jun 2006: getregion and norm99 moved elsewhere; housekeeping
	29 Nov 2005: inline argument declarations and strict prototyping
	30 Oct 2005: minimize use of static globals
	28 Aug 2004: routines used only here are declared as static; cosmetics
	11 Aug 2004: cosmetic changes; bizarre bug in calc_derivs() fixed
	27 Jun 2004: cosmetic changes; tell_dvcoef() moved to ascout.c
	24 Jan 2004: variable zm in meanprofile() renamed to zmt
		so it doesn't overlap with the static global zm.
	21 Jan 2004: license message inserted
	13 Dec 2003: eliminate unused variable flp0 in analyze_elt();
		eliminate unused variable lxyz in contig()
	19 Sep 2003: correct ugly bug in ellip(): funarg replaced by prval
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	 5 Jan 2003: reorganized source-file header
	 1 Nov 2002: use of REGIONPROFILE structures for reflection-box sizes;
		MASKRESCALE, several other defines moved to ../include/xgen.h;
		in general, model-profile boundaries treated as variables;
		vectowmp() and wmptovec() rewritten in response;
		numerous variables relabeled; dospread() takes an extra
		argument so it can handle Kabsch as well as (X,Y,om) profiles;
		substantial rewrite of dospread();
		model_ftoc() and cleanmodels() written here
	25 Mar 2002: removed MSI attachments
	31 May 1995: "register" -> "register int" throughout
	20 May 1995: slash/star removed from interior of comment
	 2 Mar 1995: delete line about file "polymath.c"
	14 Jun 1994: bugfix for vectowmp if bz < 7
	 5 May 1994: initial version, culled from elsewhere.

   Purpose: editing and organizing 3-D profile structures.
   Cross references: xgengeom.c, ../curses/ew_mask.c,
	../prerefine/spotshoes.c,../integrate/int*.c

*************************************************************************/
#include	"xgmodel.h"
#define		EXTRAP_RATIO	(0.4)
#define		HYPERSECANT	2	/* sech(x) ellipsoids */
#define		ARCCOT		3	/* arccot(x) ellipsoids */
#define		MAXDER	(2.)
#define		Q10	0.010	/* q1 when # of observations == 0 */
#define		Q1L	0.004	/* q1 when we have lots of observations*/
#define		PLENTY	40	/* min # of obs to set q1 = Q1L */

int wmptovec(MASK_3D *wmp, float *ve)
{ /*	this converts a model profile from the weird format it's stored in
	into a simple floating-point vector of length lx * ly * lz,
	indexed with x fastest and z slowest. We return EBADVALUE if this
	conversion fails because the peak is too small. */
	int		z, y, x, lx, ly, lz;
	double		thisden, maxmax, resc;
	static float	*vp;
	PHIMASK		*pip;

	lx = BIGSPOTWID;	ly = BIGSPOTHT;	lz = PFRAMES;
	for (maxmax = z = 0, pip = &(wmp->phielt[0]); z < lz; z++,pip++)
	   if ((((thisden = pip->isqsum) > FUZZ)) && /* peak in z profile*/
		((resc = pip->maxf / thisden) > maxmax)) maxmax = resc;
	if (maxmax < FUZZ) return EBADVALUE;	maxmax /= MASKRESCALE;
	/* now do the conversion */
	for (z = 0, pip = &(wmp->phielt[0]), vp = ve; z < lz; z++, pip++)
	   {
		if ((thisden = pip->isqsum) > FUZZ)
		  {
			resc = pip->maxf / (thisden * maxmax);
			for (y = 0; y < ly; y++)
			   for (x = 0; x < lx; x++, vp++)
				*vp = resc * (double)(pip->xyint[x][y]);
		  }
		else {
			for (y = 0; y < ly; y++)
			   for (x = 0; x < lx; x++, vp++) *vp = 0.;
		     }
	   }
	return EOK;
}

int vectowmp(float *ve, MASK_3D *wmp, int inobs, float misq, int cz)
{ /* This converts a vector of length (lx * ly * cz)
	to a model profile structure `wmp', suitable for i/o operations.
	Any frame of `ve' which is entirely empty is given a maxf value of 0,
	a nonzero isqsum value, and a few values in the center of the
	model are set equal to 1 just so nothing blows up.
	The `inobs' and `misq' values are stuffed into the
	appropriate slots in the `wmp->phielt[]' structures.  The routine
	returns EBADVALUE if the peak isn't positive; EOK if it is positive. */
	int		z, y, x, zmin, zmax, iz, lx, ly, lz, lz2;
	double		maxfm[PFRAMES], tmax, maxmax, thisv;
	PHIMASK		*phip;
	static float	*vp;

	lx = BIGSPOTWID;	ly = BIGSPOTHT;	lz = PFRAMES;	lz2 = lz / 2;
	if ((NULL == ve) || (NULL == wmp)) return EALLOC;
	zeromem((AVOID *)wmp, sizeof (MASK_3D));
	for (vp = ve, maxmax = z = 0; z < cz; z++)
	   {
		for (tmax = y = 0; y < ly; y++)
		   for (x = 0; x < lx; x++, vp++)
			if (*vp > tmax) tmax = *vp;
		maxfm[z] = tmax;	if (tmax > maxmax) maxmax = tmax;
	   }
	if (maxmax < FUZZ) return EBADVALUE;
	if (cz < 1) cz = 1;	if (cz > lz) cz = lz;
	zmin = lz2 - cz / 2;	zmax = zmin + cz;
	for (vp = ve,z = zmin, phip = &(wmp->phielt[zmin]), iz = 0;
	 z < zmax; z++, phip++, iz++)
	   {
		phip->isqsum = misq; phip->mnobs = (inobs > 1 ? inobs : 1);
		if ((tmax = maxfm[iz]) > FUZZ)
		  {
			phip->maxf = misq * tmax / maxmax;
			phip->isqsum = misq;	phip->mnobs = inobs;
			for (y = 0; y < ly; y++)
			   for (x = 0; x < lx; x++, vp++)
			      {
				thisv = flimits((*vp) / tmax, -1., 1.);
				if (thisv >= 0.) phip->xyint[x][y] =
				 (int)(thisv * MASKRESCALE + 0.499999);
				else phip->xyint[x][y] =
				 (int)(thisv * MASKRESCALE - 0.499999);
			      }
		  }
		else {
			phip->maxf = 0.; /* don't put fuzz in the middle */
			for (y = 0; y < ly; y++)
			   for (x = 0; x < lx; x++) phip->xyint[x][y] = 0;
			vp += lx * ly;
		     }
	   }
	return EOK;
}

int rgptowmp(REGIONPROFILE *rgp, MASK_3D *wmp)
{ /*	This converts a floating-point `REGIONPROFILE' structure to
	the corresponding `MASK_3D' structure. */
	return vectowmp(rgp->m3_xyop, wmp, rgp->m3_nobs,
		rgp->m3_isqsum, rgp->m3_bz);
}

int blankf(float *ve, int cz)
{ /* This returns 1 if there are no positive profile elements in `*ve' */
	int		z, lxy;
	double		maxmax;
	static float	*vp;

	lxy = BIGSPOTWID * BIGSPOTHT;
	for (vp = ve, maxmax = 0., z = cz * lxy; z--; vp++)
		if (*vp > maxmax) maxmax = *vp;
	return maxmax < FUZZ;
}

void meanprofile(REGIONPROFILE *rgb, int iprof)
{ /* This computes a weighted average of all the model profiles
	 and sticks it in the iprof'th profile.  The weighting scheme
	 is such that the profiles closest to the current one have
	 the greatest influence. */
	int		reg, z, zmt, lz, nmodels;
	double		wt, hx, hy, tx, ty, wtsum;
	static float	*pfi, *pfo;
	static int	profx[] = { -3, 0, 3,-4,-2,-1, 1, 2, 4,-3, 0, 3 };
	static int	profy[] = { -2,-2,-2, 0, 0, 0, 0, 0, 0, 2, 2, 2 };
	static REGIONPROFILE	*rgp, *rgo;

	rgo = rgb + iprof;	hx = profx[iprof];	hy = profy[iprof];
	lz = rgo->m3_bz;	nmodels = rgo->m3_nmodels;
	zmt = lz * (rgo->m3_by) * (rgo->m3_bx);
	zeromem((void *)(rgo->m3_xyop), zmt * sizeof (float));
	rgo->m3_isqsum = rgo->m3_nobs = 0;
	for (wtsum = reg = 0; reg < nmodels; reg++)
	   if (reg != iprof)
	     {
		tx = profx[reg];	ty = profy[reg];
		wt = 1. / ((tx - hx) * (tx - hx) + (ty - hy) * (ty - hy));
		wtsum += wt;
	     }
	wtsum = 1. / wtsum;
	for (reg = 0, rgp = rgb; reg < nmodels; reg++, rgp++)
	   if ((reg != iprof) && (rgp->m3_nobs))
	     {
		tx = profx[reg];	ty = profx[reg];
		wt = wtsum / ((tx - hx) * (tx - hx) + (ty - hy) * (ty - hy));
		pfo = rgo->m3_xyop;	pfi = rgp->m3_xyop;
		for (z = 0; z < zmt; z++, pfo++, pfi++) *pfo += *pfi * wt;
		rgo->m3_isqsum += rgp->m3_isqsum * wt;
	     }
}

static double forward(double z, int ellt, double lorp)
{ /* This computes the appropriate function.  All of these
	functions have the properties that f(0) = 1, and f(large) = 0.
	The functions expect the SQUARE of their "native" arguments.
	If the argument is negative we try tricky continuations of
	the functions: for the Gaussian it's easy, for the others not. */

	if (ellt == GAUSSIAN) return exp(-z);
	else if (ellt == HYPERSECANT)
	  {
		if (z < 0.) return 0.5 * (exp(z) + exp(-z));
		else return 2. / (exp(z) + exp(-z));
	  }
	else if (ellt == ARCCOT)
	  {
		if (z < -1.) return 1.e9;
		else if (z < 0.) return 2. - acos(-z) / PIO2;
		else return atan(1. / z) / PIO2;
	  }
	else return ((z < -1.) ? 1.e9 : pow(1. + z, -lorp));
}

static double invfun(double z, int ellt, double ilorp)
{ /*	This returns the appropriate inverse function of its argument.
	In most cases we square the output before returning it.  The function
	returns a large positive value for any nonpositive argument. */

	if (z < DFUZZ) return 1.e9;
	if (ellt == GAUSSIAN) return -log(z);
	else if (ellt == HYPERSECANT)
	  {
		if (z > 1.) return -log(z + sqrt(z * z - 1.));
		else return log(1. / z + sqrt(1. / (z * z) - 1.));
	  }
	else if (ellt == ARCCOT)
	  {
		if (z > 1.) return PIO2 * (1. - z);
		else return 1. / tan(PIO2 * z);
	  }
	else return pow(z, -ilorp) - 1.;
}

static int getellp2(int elltype, double *xco, double *ellpars, double lorp)
{ /*	This computes the ellipse parameters from the least-squares solution.
	It returns EOK if successful, various negative integers otherwise.
	This routine is the 2-D analog of the 3-D routine "getellp",
	with evident simplifications that go with reducing the dimension.
	The returned values are:
	 ellpars[0] = function value at x=y=0
	 ellpars[1] = eigenvalue closest to X
	 ellpars[2] = eigenvalue closest to Y
	 ellpars[3] = angle between ellipse's major axis and X axis */
	int		i;
	double		detw, discrim, funoffoo, trw, avec[4], dep[6], dxco[6];

	for (i = 0; i < 6; i++) dxco[i] = xco[i];
	detw = dxco[1] * dxco[3] - dxco[2] * dxco[2];
	if (fabs(detw) < DFUZZ) return -1;
	dep[4] = (dxco[3] * dxco[4] - dxco[2] * dxco[5]) / detw;
	dep[5] = (dxco[0] * dxco[5] - dxco[2] * dxco[4]) / detw;
	avec[0] = 1.;	avec[1] = dep[4] * dep[4];
	avec[2] = 2. * dep[4] * dep[5];
	avec[3] = dep[5] * dep[5];
	funoffoo = ddotprod(avec, dxco, 4);
	ellpars[0] = forward(funoffoo, elltype, lorp);
	/* construct trace and discriminant of the characteristic polynomial */
	if (DFUZZ >= (trw = xco[1] + xco[3])) return -2; /* trace > 0 */
	discrim = trw * trw - 4. * detw;
	/* solve the quadratic and get the associated eigenvectors. */
	ellpars[1] = 0.5 * (trw - discrim);
	avec[0] = xco[2];	avec[1] = ellpars[1] - avec[1];
	if ((trw = avec[0] * avec[0] + avec[1] * avec[1]) < DFUZZ) return -7;
	avec[0] /= trw;	avec[1] /= trw;
	ellpars[2] = 0.5 * (trw + discrim);
	avec[2] = xco[2];	avec[3] = ellpars[2] - avec[1];
	if ((trw = avec[2] * avec[2] + avec[3] * avec[3]) < DFUZZ) return -7;
	avec[2] /= trw;	avec[3] /= trw;
	/* compute Euler angle associated with the rotation matrix consisting
	of the 2 eigenvectors. Re-identify axes so the second is closer to y */
	if (fabs(avec[3]) < fabs(avec[1]))
	  {
		trw = avec[3];	avec[3] = avec[1];	avec[1] = trw;
		trw = avec[2];	avec[2] = avec[0];	avec[0] = trw;
		trw = ellpars[1]; ellpars[1] = ellpars[2]; ellpars[2] = trw;
	  }
	/* make sure that the determinant is positive */
	if ((avec[0] * avec[3] - avec[1] * avec[2]) < 0.)
	  {
		avec[2] *= -1.;	avec[3] *= -1.;
	  }
	/* Having finally gotten the matrix straight, we parameterize it as
		ellpars[3] = phiz = acos(avec[3]) * sign(avec[2]); */
	ellpars[3] = (fabs(avec[3]) > 1. ? 0. : acos(avec[3]));
	if (avec[2] < 0.) ellpars[3] = -ellpars[3];	return EOK;
}

static int getellp(int elltype, double *xco, double *ellpars, double lorp)
{ /*	This computes the ellipsoid parameters from the least-squares solution.
	It returns EOK if successful, various negative integers otherwise*/
	register int	i, j;
	int		imaxz, imid, iminz;
	double		avec[7], dtem[3], dxco[7], qf[3][3], qfi[3][3];
	double		thisv, funoffoo, trw, detw, minors, colmax, discrim;

	qf[0][0] = xco[1];	qf[0][1] = xco[2];	qf[0][2] = xco[3];
	qf[1][0] = xco[2];	qf[1][1] = xco[4];	qf[1][2] = xco[5];
	qf[2][0] = xco[3];	qf[2][1] = xco[5];	qf[2][2] = xco[6];
	for (i = 0; i < 3; i++) avec[i] = xco[i+7];
	for (i = 0; i < 7; i++) dxco[i] = xco[i];
	if (dinv3(qf, qfi))
	  {
		/* if qf turns out to be singular, we see whether maybe
		 the 2-D version might work */
		for (i = 0; i < 3; i++) avec[i] = xco[i];
		avec[3] = xco[4];	avec[4] = xco[7]; avec[5] = xco[8];
		for (i = 0; i < 6; i++) xco[i] = avec[i];
		i = getellp2(elltype, xco, ellpars, lorp);
		if (EOK != i) return i;
		ellpars[8] = ellpars[5];	ellpars[7] = ellpars[4];
		ellpars[4] = ellpars[3];
		ellpars[3] = ellpars[5] = ellpars[6] = ellpars[9] = 0.;
		return EOK;
	  }
	dmm31(dtem, qfi, avec);
	avec[0] = 1.;	avec[1] = dtem[0] * dtem[0];
	avec[2] = 2. * dtem[0] * dtem[1]; avec[3] = 2. * dtem[0] * dtem[2];
	avec[4] = dtem[1] * dtem[1];	avec[5] = 2. * dtem[1] * dtem[2];
	avec[6] = dtem[2] * dtem[2];	funoffoo = ddotprod(avec, dxco, 7);
	ellpars[0] = forward(funoffoo, elltype, (double)lorp);
/* construct the parameters of the characteristic polynomial: the
 trace of the normal matrix, the sum of its minors, and the determinant */
	if (DFUZZ >= (trw = xco[1] + xco[4] + xco[6])) return -2; /* trace > 0*/
	minors = xco[1] * xco[4] + xco[1] * xco[6] + xco[4] * xco[6] -
		xco[3] * xco[3] - xco[5] * xco[5] - xco[2] * xco[2];
	if ((detw = xco[1] * xco[4] * xco[6] + 2. * xco[2] * xco[3] * xco[5] -
		xco[3] * xco[3] * xco[4] - xco[5] * xco[5] * xco[1] -
		xco[2] * xco[2] * xco[6]) <= DFUZZ) return -3; /* determ > 0 */
/* find the extrema of the characteristic polynomial */
	discrim = trw * trw - 3. * minors;
	if (discrim < 0.) return -4;	discrim = sqrt(discrim);
/* use these to calculate, by Newton's method, two roots.  They should be
 positive in order for the surface to be an ellipsoid and not a hyperboloid.
 For each eigenvalue we determine the associated eigenvector. */
	if (DFUZZ > (ellpars[2] = newtcube(0., -trw, (double)minors, -detw)))
		return -5;
	if (eigvec(qf, (double)(ellpars[2]), qfi[1])) return -6;
	if (DFUZZ > (ellpars[1] = newtcube(trw-ellpars[2], -trw,minors,-detw)))
		return -7;
	if (eigvec(qf, (double)(ellpars[1]), qfi[0])) return -8;
	/* the sum of the eigenvalues should be the trace of qf, so the 3rd
	eigenvalue is trw - (the other two). It should be positive, too. */
	if (DFUZZ > (ellpars[3] = trw - ellpars[1] - ellpars[2])) return -9;
	/* the last eigenvector is orthogonal to the first 2, and since
 	they're normalized, we compute it as the cross-product of the first 2*/
	qfi[2][0] = qfi[0][1] * qfi[1][2] - qfi[0][2] * qfi[1][1];
	qfi[2][1] = qfi[0][2] * qfi[1][0] - qfi[0][0] * qfi[1][2];
	qfi[2][2] = qfi[0][0] * qfi[1][1] - qfi[0][1] * qfi[1][0];
/* compute the Euler angles associated with the rotation matrix consisting
	of the three eigenvectors.  */
	for (i = 3; i--; )
	   {
		imaxz = 2;	colmax = fabs(qfi[i][imaxz]);
		for (j = 2; j--; )
		   if ((thisv = fabs(qfi[i][j])) > colmax)
		     {
			imaxz = j;	colmax = thisv;
		     }
		if (qfi[i][imaxz] < 0.)
		  for (j = 3; j--; ) qfi[i][j] = -qfi[i][j];
	   }
/* re-identify the three axes of this Eulerian system so that the
 last one is the one closest to z and the middle one is the one
 closer to y of the two others.  The eigenvalues shift accordingly */
	imaxz = 2;
	if (fabs(qfi[1][2]) > fabs(qfi[2][2])) imaxz = 1;
	if (fabs(qfi[0][2]) > fabs(qfi[imaxz][2])) imaxz = 0;
	imid = imaxz - 1;	if (imid < 0) imid = 2;
	iminz = 3 - imid - imaxz;
	if (fabs(qfi[imid][1]) < fabs(qfi[iminz][1]))
	  {
		imid = iminz;	iminz = 3 - imid - imaxz;
	  }
	for (i = 3; i--; )
	   {
		qf[i][2] = qfi[imaxz][i];	qf[i][1] = qfi[imid][i];
		qf[i][0] = qfi[iminz][i];
	   }
/* flip one column of qf if the determinant is less than zero */
	if (ddeter(qf) < 0.) for (i = 3; i--; ) qf[0][i] = -qf[0][i];
	for (i = 3; i--; ) qfi[i][0] = ellpars[i+1];
	ellpars[1] = qfi[iminz][0];
	ellpars[2] = qfi[imid][0];	ellpars[3] = qfi[imaxz][0];
/* Having finally gotten qf straight, we parameterize it as follows:
	ellpars[4] = phi2z = acos(qf[2][2]) * sign(qf[2][1])
	ellpars[5] = phi1y = acos(qf[1][1]) * sign(qf[1][0])
	ellpars[6] = phi1z = asin(qf[2][0]).
 The other other elements of qf drop out naturally from the orthogonality
 relationships (trust me!) */
	ellpars[4] = (fabs(qf[2][2]) > 1. ? 0. : acos(qf[2][2]));
	if (qf[2][1] < 0.) ellpars[4] = -ellpars[4];
	ellpars[5] = (fabs(qf[1][1]) > 1. ? 0. : acos(qf[1][1]));
	if (qf[1][0] < 0.) ellpars[5] = -ellpars[5];
	ellpars[6] = asin(qf[2][0]);	return EOK;
}

int ellip(REGIONPROFILE *rgp, int elltype, int cflags,
 double floor_value, double lorp, double *ellpars)
{ /*	This computes the least-squares best-fit gaussian or lorenzian or
	sech(x) ellipse associated with the data found in the region pointed
	to by `rgp'.  It then sets to zero all pixels lying outside the
	contour level given by `floor_value'. It thus is identical to
	"ellsoid" (below) except that it is two-dimensional instead of
	three-dimensional */

	register int	ix, iy;
	int		lx, ly, lxy, dx, dy, points_counted, resu;
	int		iwidth, hwidth, iheight, hheight;
	int		ixmin, iymin, ixmax, iymax, ndown, nup, nmodels;
	double		cval, dval, dydy, funarg, funofcut, ilorp, m2dy;
	double		prval, rmse, rmsed, thisden, wt, wtfloor;
	double		avec[6], r[21], xco[6], y[6];
	static float	*flpy, *flp0, *flp, *flpo;

	/* prepare for least-squares calculation in a 7x7 box. */
	inv_init(r, y, 6);	/* initialize least-squares stuff */
	wtfloor = 1. / (sqrt(200.) - 1.);	nmodels = rgp->m3_nmodels;
	if (cflags & CLF_DIFFVECTORS)
	  {
		lx = rgp->m3_qx; ly = rgp->m3_qy; flp0 = rgp->m3_dvp;
	  }
	else {
		lx = rgp->m3_bx; ly = rgp->m3_by;	flp0 = rgp->m3_xyop;
	     }
	lxy = lx * ly;	flp = flp0;
	/* make sure profile is normalized such that the largest value is 1. */
	for (thisden = ix = 0; ix < lxy; ix++, flp++)
		if (*flp > thisden) thisden = *flp;
	if ((thisden < DFUZZ) || (fabs(lorp) < DFUZZ)) return EBADVALUE;
	ilorp = 1. / lorp;
	for (ix = lxy; ix--; ) *(--flp) /= thisden;
	iwidth = lx / 2;	hwidth = iwidth / 2;
	ixmin = lx / 2 - hwidth;		ixmax = ixmin + iwidth;
	iheight = ly / 2; 	hheight = iheight / 2;
	iymin = ly / 2 - hheight;		iymax = iymin +iheight;
	flpy = flp0 + iymin * lx;
	for (iy = iymin, dy = -hwidth; iy < iymax; iy++, dy++, flpy += ly)
	   {
		dydy = dy * dy;	m2dy = -2. * dy;
		for (ix = ixmin, dx = -hwidth, flp = flpy + ixmin;
		 ix < ixmax; ix++, dx++, flp++)
		   if (0.005 < (prval = *flp))
		     {
			funarg = invfun(prval, elltype, ilorp);
			wt = funarg < 0. ? 1. / wtfloor :
				1. / (wtfloor + sqrt(funarg));
			funarg *= wt;	avec[0] = wt;	avec[1] = dx * dx * wt;
			avec[2] = 2. * dx * dy * wt;	avec[3] = dydy * wt;
			avec[4] = -2. * dx * wt;	avec[5] = m2dy * wt;
			inv_add(r, y, avec, funarg, 6);
		     }
	   }
	if ((ix = inv_solve(r, y, xco, 6)))
	  	return (ix == 1) ? ESINGULAR : EALLOC;
/* now redo it using all the data inside the box computed the first time,
	out to the specified contour level. */
	inv_init(r, y, 6);	/* initialize least-squares stuff */
	funofcut = invfun(floor_value, elltype, ilorp) * 2.;
	points_counted = 0; iwidth = lx;	hwidth = iwidth / 2;
	iheight = ly;	hheight = iheight / 2;	
	ixmin = iymin = 0;	ixmax = lx;	iymax = ly;
	for (flp = flp0, iy = iymin, dy = -hwidth; iy < iymax; iy++, dy++)
	   {
		dydy = dy * dy;	m2dy = -2. * dy;
		for (ix = ixmin, dx = -hwidth; ix < ixmax; ix++, dx++, flp++)
		   if (0.005 < (prval = *flp))
		     {
			prval = invfun(prval, elltype, ilorp);
			wt = prval < 0. ? 1. / wtfloor :
				1. / (wtfloor + sqrt(prval));
			prval *= wt; avec[0] = wt; avec[1] = dx * dx * wt;
			avec[2] = 2. * dx * dy * wt; avec[3] = dydy * wt;
			avec[4] = -2. * dx * wt; avec[5] = m2dy * wt;
			funarg = ddotprod(avec, xco, 6) / wt;
			if (funarg < funofcut)
			  {
				points_counted++;
				inv_add(r, y, avec, prval, 6);
			  }
		     }
	   }
	if ((ix = inv_solve(r, y, xco, 6)))
		return (ix == 1) ? ESINGULAR : EALLOC;	/* solve lsq */
	for (ix = 1; ix < 4; ix++) ellpars[ix] = xco[ix];
/* find eternal verities: peak position, value at the peak, eigenvalues
 of the normal matrix, and directions of the principal axes.  */
	if (EOK != (resu = getellp2(elltype, xco, ellpars, lorp))) return resu;
	rmse = rmsed = points_counted = ndown = nup = 0;
/* now go back and level out all the stuff outside the user-set contour
 level on the model ellipse. */
	if (cflags & CLF_OUTELLIPSE)
	  {
		if (cflags & CLF_DIFFVECTORS) flpo = (rgp + nmodels)->m3_dvp;
		else flpo = (rgp + nmodels)->m3_xyop;
	  }
	else {
		if (cflags & CLF_OUTELLIPSE) flpo = rgp->m3_dvp;
		else flpo = rgp->m3_xyop;
	     }
	flp = flp0;
	avec[0] = 1.;
	for (iy = iymin, dy = -hwidth; iy < iymax; iy++, dy++)
	   {
		avec[3] = dy * dy;	avec[5] = -2. * dy;
		for (ix = ixmin, dx = -hwidth; ix < ixmax;
		 ix++, dx++, flp++, flpo++)
		   {
			avec[1] = dx * dx; avec[2] = 2. * dx * dy;
			avec[4] = -2. * dx;
			funarg = ddotprod(avec, xco, 6);
			cval = forward(funarg, elltype, lorp);
			if (cflags & CLF_OUTELLIPSE) *flpo = cval;
			prval = *flp;
			if ((cval < floor_value) && (cflags & CLF_FLATTEN))
			  {
				*flp = 0.; if (prval > FUZZ) ndown++;
			  }
			else {
				dval = cval - prval; dval *= dval;
				if (dval > 1.) dval = 1.;
				wt = prval; rmse += dval * wt;	rmsed += wt;
				points_counted++;
				if ((1 <= ix && ix <= (lx-2)) &&
				 (1 <= iy && iy <= (ly-2)) &&
				 (cflags & CLF_EXTRAP) &&
				 (prval < EXTRAP_RATIO * cval) &&
				 (prval < floor_value) &&
				 (cval > floor_value))
				  {
					nup++;	*flp = cval;
				  }
			     }
		   }
	   }
	ellpars[11] = points_counted ? 99. * sqrt(rmse / rmsed) : 99.;
	ellpars[10] = (double)ndown;	ellpars[12] = (double)nup;
	return EOK;
}

int ellsoid(REGIONPROFILE *rgp, int elltype, int cflags,
	double floor_value, double lorp, double *ellpars)
{ /*	This computes the least-squares best-fit gaussian or lorenzian or
	sech(x) ellipsoid associated with the data found in the region
	pointed to by `rgp'.  It then sets to zero all pixels lying outside
	the contour level given by `floor_value'. */

	register int	ix, iy;
	int		dx, dy, dz, fram, hheight, hwidth, iheight, iwidth;
	int		ixmin, iymin, ixmax, iymax, lx, ly, lz, lxy, lxyz, lz2;
	int		ndown, nmodels, nup, points_counted, resu;
	double		cval, dval, dydy, dydz2, dzdz, funarg, funofcut, ilorp;
	double		m2dy, m2dz, prval, rmse, rmsed, thisden, wt, wtfloor;
	double		r[55], avec[10], y[10], xco[10];
	static float	*flpz, *flpy, *flp, *flpo, *flp0;

/* get ready to do the least-squares calculation using just the data in a box
 which is 7x7 in the centermost frame and 5x5 in the two neighboring frames.*/
	inv_init(r, y, 10);	/* initialize least-squares stuff */
	wtfloor = 1. / (sqrt(200.) - 1.);	nmodels = rgp->m3_nmodels;
	if (cflags & CLF_DIFFVECTORS)
	  {
		lx = rgp->m3_qx; ly = rgp->m3_qy; lz = rgp->m3_qz;
		flp0 = rgp->m3_dvp;
		if (cflags & CLF_OUTELLIPSE) flpo = (rgp + nmodels)->m3_dvp;
		else flpo = rgp->m3_dvp;
	  }
	else {
		lx = rgp->m3_bx; ly = rgp->m3_by; lz = rgp->m3_bz;
		flp0 = rgp->m3_xyop;
		if (cflags & CLF_OUTELLIPSE) flpo = (rgp + nmodels)->m3_xyop;
		else flpo = rgp->m3_xyop;
	     }
	lxy = lx * ly;	lxyz = lxy * lz;	lz2 = lz / 2;
	fram = lz / 2 - 1;	flp = flp0;	flpz = flp + fram * lxy;
	/* make sure profile is normalized such that the largest value is 1. */
	for (thisden = ix = 0; ix < lxyz; ix++, flp++)
		if (*flp > thisden) thisden = *flp;
	if ((thisden < DFUZZ) || (fabs(lorp) < DFUZZ)) return EBADVALUE;
	ilorp = 1. / lorp;
	for (ix = lxyz; ix--; ) *(--flp) /= thisden;
	for (dz = -1; dz <= 1; fram++, dz++, flpz += lxy)
	   {
		iwidth = dz ? lx / 2 - 2 : lx / 2;	hwidth = iwidth / 2;
		ixmin = lx / 2 - hwidth;		ixmax = ixmin + iwidth;
		iheight = dz ? ly / 2 - 2 : ly / 2;	hheight = iheight / 2;
		iymin = ly / 2 - hheight;		iymax = iymin +iheight;
		dzdz = dz * dz;	m2dz = -2. * dz; flpy = flpz + iymin * lx;
		for (iy = iymin, dy = -hwidth; iy < iymax;
		 iy++, dy++, flpy += lx)
		   {
			dydy = dy * dy;	dydz2 = 2. * dy * dz;
			m2dy = -2. * dy;
			for (ix = ixmin, dx = -hwidth, flp = flpy + ixmin;
			 ix < ixmax; ix++, dx++, flp++)
			   if (0.005 < (prval = *flp))
			     {
				funarg = invfun(prval, elltype, ilorp);
				wt = funarg < 0. ? 1. / wtfloor :
					1. / (wtfloor + sqrt(funarg));
				funarg *= wt; avec[0] = wt;
				avec[1] = dx * dx * wt;
				avec[2] = 2. * dx * dy * wt;
				avec[3] = 2. * dx * dz * wt;
				avec[4] = dydy * wt;	avec[5] = dydz2 * wt;
				avec[6] = dzdz * wt;	avec[7] = -2. * dx * wt;
				avec[8] = m2dy * wt;	avec[9] = m2dz * wt;
				inv_add(r, y, avec, (double)funarg,10);
			     }
		   }
	   }
	if ((fram = inv_solve(r, y, xco, 10)))
	  return (fram == 1) ? ESINGULAR : EALLOC; /* solve least-squares */
/* now redo it using all the data inside the box computed the first time,
	out to the specified contour level. */
	inv_init(r, y, 10);	/* initialize least-squares stuff */
	funofcut = invfun(floor_value, elltype, ilorp) * 2.;
	points_counted = 0;
	iwidth = lx;	hwidth = iwidth / 2;
	iheight = ly;	hheight = iheight / 2;	
	ixmin = iymin = 0;	ixmax = lx;	iymax = ly;
	for (fram = 0, dz = -lz2, flp = flp0; fram < lz; fram++, dz++)
	   {
		dzdz = dz * dz;	m2dz = -2. * dz;
		for (iy = iymin, dy = -hwidth; iy < iymax; iy++, dy++)
		   {
			dydy = dy * dy;	dydz2 = 2. * dy * dz;
			m2dy = -2. * dy;
			for (ix = ixmin, dx = -hwidth;
			 ix < ixmax; ix++, dx++, flp++)
			   if (0.005 < (prval = *flp))
			     {
				prval = invfun(prval, elltype, ilorp);
				wt = prval < 0. ? 1. / wtfloor :
					1. / (wtfloor + sqrt(prval));
				prval *= wt;	avec[0] = wt;
				avec[1] = dx * dx * wt;
				avec[2] = 2. * dx * dy * wt;
				avec[3] = 2. * dx * dz * wt;
				avec[4] = dydy * wt;	avec[5] = dydz2 * wt;
				avec[6] = dzdz * wt;	avec[7] = -2. * dx * wt;
				avec[8] = m2dy * wt;	avec[9] = m2dz * wt;
				funarg = ddotprod(avec, xco, 10) / wt;
				if (funarg < funofcut)
				  {
					points_counted++; inv_add(r, y,
					 avec, (double)prval, 10);
				  }
			     }
		   }
	   }
	if ((fram = inv_solve(r, y, xco, 10)))
	  return (fram == 1) ? ESINGULAR : EALLOC; /* solve least-squares */
	for (ix = 1; ix < 7; ix++) ellpars[ix] = xco[ix];
	/* figure out the eternal verities: the peak position, the value at the
	peak, the eigenvalues of the normal matrix, and the directions
	of the principal axes.  */
	if (EOK != (resu = getellp(elltype, xco, ellpars, lorp))) return resu;
	rmse = rmsed = points_counted = ndown = nup = 0;
	/* now go back and level out all the stuff outside the user-set
	contour level on the model ellipsoid. */
	for (fram = 0, flp = flp0, dz = -lz2; fram < lz; fram++, dz++)
	   {
		avec[0] = 1.;	avec[6] = dz * dz;	avec[9] = -2. * dz;
		for (iy = iymin, dy = -hwidth; iy < iymax; iy++, dy++)
		   {
			avec[4] = dy * dy; 	avec[5] = 2. * dy * dz;
			avec[8] = -2. * dy;
			for (ix = ixmin, dx = -hwidth; ix < ixmax;
			 ix++, dx++, flp++, flpo++)
			   {
				avec[1] = dx * dx; avec[2] = 2. * dx * dy;
				avec[3] = 2. * dx * dz; avec[7] = -2. * dx;
				funarg = ddotprod(avec, xco, 10);
				cval = forward(funarg, elltype, lorp);
				if (cflags & CLF_OUTELLIPSE) *flpo = cval;
				prval = *flp;
				if ((cval < floor_value) &&
				 (cflags & CLF_FLATTEN))
				  {
					*flp = 0.; if (prval > FUZZ) ndown++;
				  }
				else {
					dval = cval - prval; dval *= dval;
					if (dval > 1.) dval = 1.;
					wt = prval;
					rmse += dval * wt;	rmsed += wt;
					points_counted++;
					if ((1 <= ix && ix <= (lx-2)) &&
					 (1 <= iy && iy <= (ly-2)) &&
					 (cflags & CLF_EXTRAP) &&
					 (prval < EXTRAP_RATIO * cval) &&
					 (prval < floor_value) &&
					 (cval > floor_value))
					  {
						nup++;	*flp = cval;
					  }
				     }
			   }
		   }
	     }
	ellpars[10] = (double)ndown;
	ellpars[11] = points_counted ? 99. * sqrt(rmse / rmsed) : 99.;
	ellpars[12] = (double)nup; return EOK;
}

void zero_boundary(REGIONPROFILE *rgp, int cflags)
{ /* This sets values in the first and last rows and the
	first and last columns to zero, except in the centermost frame */
	int	z, y, x, lx, lxy;
	float	*flz, *flp;

	lx = rgp->m3_bx;	lxy = lx * rgp->m3_by;
	for (flz = rgp->m3_xyop, z = 0; z < rgp->m3_bz; z++, flz += lxy)
	   if ((cflags & CLF_ZCBORDER) || (z != ((rgp->m3_bz) / 2)))
	     {
		for (flp = flz, x = 0; x < lx; x++, flp++) *flp = 0.;
		for (flp = flz + lxy - lx, x = 0; x < lx; x++,flp++) *flp = 0.;
		for (flp = flz, y = 0; y < rgp->m3_by; y++, flp += lx)
			*flp = *(flp + (lx - 1)) = 0.;
	     }
}

/*	The next several routines are associated with the effort to
	calculate contiguity relationships among the elements in a model,
	i.e., they allow the user to set pixels to zero if they're isolated
	(not in the vicinity of other nonzero pixels */
static void calc_derivs(REGIONPROFILE *rgp)
{ /* this computes dx/dz and dy/dz for a set of masks. It does this by
	least-squares fits to the lines through the centroids of the spots */
	int		x, y, z, lx, ly, lz, lxy, lx2, ly2, lz2;
	double		thiselt, intz, cenx, ceny;
	static float	*flp0, *flpz, *flpy, *flp;

	rgp->m3_slopex = rgp->m3_slopey = 0.;
	lx = rgp->m3_bx; ly = rgp->m3_by; lz = rgp->m3_bz; lxy = lx * ly;
	lx2 = lx / 2;	ly2 = ly / 2;	lz2 = lz / 2;	flp0 = rgp->m3_xyop;
	for (z = 1, flpz = flp0 + lxy; z < lz - 1; z++, flpz += lxy)
	   {
		cenx = ceny = intz = 0;
		for (y = 1, flpy = flpz + lx; y < ly - 1; y++, flpy += lx)
		   for (x = 1, flp = flpy + 1; x < lx - 1; x++, flp++)
		     if ((thiselt = *flp) > FUZZ)
		      {
			intz += thiselt;	cenx += thiselt * x;
			ceny += thiselt * y;
		      }
		if (intz > DFUZZ) 
		  {
			cenx /= intz;	ceny /= intz;
		  }
		else {
			cenx = lx2; ceny = ly2;
		     }
		rgp->m3_slopex += cenx * (z - lz2);
		rgp->m3_slopey += ceny * (z - lz2);
	   }
		/* least-squares rgp->m3_slope */
	rgp->m3_slopex /= 10.;	rgp->m3_slopey /= 10.;
	if (rgp->m3_slopex < -MAXDER) rgp->m3_slopex = -MAXDER;
	else if (rgp->m3_slopex > MAXDER) rgp->m3_slopex = MAXDER;
	rgp->m3_dx_dz = (rgp->m3_slopex < 0.?
		rgp->m3_slopex - 0.49999 : rgp->m3_slopex + 0.49999);
	if (rgp->m3_slopey < -MAXDER) rgp->m3_slopey = -MAXDER;
	else if (rgp->m3_slopey > MAXDER) rgp->m3_slopey = MAXDER;
	rgp->m3_dy_dz = (rgp->m3_slopey < 0. ?
		rgp->m3_slopey - 0.49999 : rgp->m3_slopey + 0.49999);
}

static int analyze_elt(REGIONPROFILE *rgp, int z, int y, int x)
{ /* this decides whether or not to keep elt[z][x][y] of the mask.
	It returns 0 if it sets the value to 0, 1 otherwise */
	int		xpl, yp, zp, lx, ly, lxy, lx2, ly2, cfrsq, outrsq;
	float		*flp, *flp1, *flp2;

	/* if the element is really weak, zero it out regardless */
	lx = rgp->m3_bx; ly = rgp->m3_by; lxy = lx * ly;
	lx2 = lx / 2;	ly2 = ly / 2;
	cfrsq = (lx2 + 2) * (lx2 + 2);
	outrsq = lx2 * lx2;
	flp = rgp->m3_xyop + lxy * z + lx * y + x;
	if (*flp < rgp->m3_q1)
	  {
		*flp = 0.;	return 0;
	  }
	/* tests applied if we're not in the peak frame : */
	if (z != rgp->m3_zm)
	  { /* if we're on the (x,y) border, zero it out no matter what;
		similarly, kill elements outside a radius of lx2. */
		if ((!x) || (!y) || (x == lx-1) || (y == ly-1))
			return 0;
		xpl = x - lx2;	yp = y - ly2;
		if (xpl * xpl + yp * yp > outrsq) return 0;
		/* otherwise, look at an element in the next frame up
		(if we're earlier than the peak) or the previous frame (if
		we're later than the peak) corresponding to this one.  If
		it's below the noise, we zero this one out.  The
		"corresponding" pixel is not necessarily at the same (x,y);
		it is at (z+1,x+dx/dz, y+dy/dz) if z < zm, and
			 (z-1,x-dx/dz, y-dy/dz) if z > zm). */
		if (z < rgp->m3_zm)
		  {
			zp = z + 1; xpl = x + rgp->m3_dx_dz;
			yp = y + rgp->m3_dy_dz;
		  }
		else {
			zp = z - 1; xpl = x - rgp->m3_dx_dz;
			yp = y - rgp->m3_dy_dz;
		     }
		flp1 = rgp->m3_xyop + lxy * zp + lx * yp + xpl;
		flp2 = rgp->m3_xyop + lxy * zp + lx * y  + x;
		if ((*flp1 < rgp->m3_q1) && (*flp2 < rgp->m3_q1)) return 0;
	  }
	else { /* in the center frame, we kill pixels radially farther away
		than 1.3 * lx2. */
		xpl = x - lx2;	yp = y - ly2;
		if (xpl * xpl + yp * yp > cfrsq) return 0;
	     }
	/* if we pass that test, we look within the current frame to
	see if the neighboring, center-directed pixels are ok */
	xpl = (x < rgp->m3_xmz ? x + 1: x - 1);
	yp = (y < rgp->m3_ymz ? y + 1: y - 1);
	if (y == rgp->m3_ymz) /* on middle row: look one to left or right */
	  {
		if (x == rgp->m3_xmz) return 1; /* in the center, we leave it */
		flp1 = rgp->m3_xyop + lxy * z + lx * rgp->m3_ymz + xpl;
		if (*flp1 < rgp->m3_q1) return 0;
	  }
	else if (x == rgp->m3_xmz) /* middle column: look one up or down */
	  {
		flp1 = rgp->m3_xyop + lxy * z + lx * yp + rgp->m3_xmz;
		if (*flp1 < rgp->m3_q1) return 0;
	  }
	else /* otherwise: we zero this element only if BOTH of
		the elements closest to it and directed toward the
		center are below the cutoff */
	  {
		if (abs(y - rgp->m3_ymz) > abs(x - rgp->m3_xmz))
		  {
			flp1 = rgp->m3_xyop + lxy * z + lx * y  + xpl;
			flp2 = flp1 + lx * (yp - y);
		  }
		else
		  {
			flp1 = rgp->m3_xyop + lxy * z + lx * yp + x;
			flp2 = flp1 + xpl - x;
		  }
		if ((*flp1 < rgp->m3_q1) && (*flp2 < rgp->m3_q1)) return 0;
	  }
	/* Phew! we've survived all these tests.
	Increment the count of nonzero pixels */
	return 1;
}

static int do_column(REGIONPROFILE *rgp, int sz, int sx)
{ /* this processes all the pixels in one column of a spot */
	int	dy, y, nnonzero, lx, ly, lxy, ly2;
	float	*flpy;

	lx = rgp->m3_bx;	ly = rgp->m3_by;
	lxy = lx * ly;		ly2 = rgp->m3_by / 2;
	flpy = rgp->m3_xyop + (sz * lxy + sx);
	for (dy = nnonzero = 0; dy <= ly2; dy++)
	   {
		y = ly2 - dy;
		if (analyze_elt(rgp, sz, y, sx)) nnonzero++;
		else *(flpy + y * lx) = 0.;
		if (dy)
	   	  {
			y = ly2 + dy;
			if (analyze_elt(rgp, sz, y, sx)) nnonzero++;
			else *(flpy + y * lx) = 0.;
		  }
	   }
	return nnonzero;
}

static int do_zsection(REGIONPROFILE *rgp, int sz)
{ /* this processes one z-section or frame of a spot */
	register int	dx, nnonzero;
	int		lx2, ly2, lz2;
	double		delcen;

	lx2 = rgp->m3_bx / 2; ly2 = rgp->m3_by / 2; lz2 = rgp->m3_bz / 2;
	delcen = ((double)(sz - lz2)) * rgp->m3_slopex;
	rgp->m3_xmz = rgp->m3_xm + (delcen < 0) ?
		delcen - 0.49999 : delcen + 0.49999;
	if (rgp->m3_xmz < lx2 - 2) rgp->m3_xmz = lx2 - 2;
	else if (rgp->m3_xmz > lx2 + 2) rgp->m3_xmz = lx2 + 2;
	delcen = ((double)(sz - lz2)) * rgp->m3_slopey;
	rgp->m3_ymz = rgp->m3_ym + (delcen < 0) ?
		delcen - 0.49999 : delcen + 0.49999;
	if (rgp->m3_ymz < ly2 - 2) rgp->m3_ymz = ly2 - 2;
	else if (rgp->m3_ymz > ly2 + 2) rgp->m3_ymz = ly2 + 2;
	for (dx = nnonzero = 0; dx <= lx2; dx++)
	   {
		nnonzero += do_column(rgp, sz, lx2 - dx);
		if (dx) nnonzero += do_column(rgp, sz, lx2 + dx);
	   }
	return nnonzero;
}

int contig(REGIONPROFILE *rgp, int thisfmonly)
{ /* This rationalizes the form of the three-dimensional weighted mask through
	the application of some simple intensity and connectivity rules.
	If f(x,y,z) are the mask values and M = Max(f(x,y,z)), the rules are:
	(1) if f(x,y,z) < Q1 * M, f(x,y,z) = 0.
	(2) f(x,y,z) = 0 along the (x,y) border except in the center frame.
	(3) in other cases, f(x,y,z) is set to zero according to some
	 continuity criteria described in the comments in `analyze_elt'.
	The routine returns the number of pixels still included after all this.
  */
	register int	x, y, z;
	int		lx, ly, lz, lxy, lx2, ly2, lz2, nonzpix, xyzm, nused;
	float		*flp;
	double		valm;

	/* failsafe: make sure some lengths are defined. */
	lx = rgp->m3_bx;	ly = rgp->m3_by;	lz = rgp->m3_bz;
	lxy = lx * ly;	lx2 = lx / 2;	ly2 = ly / 2;	lz2 = lz / 2;
	rgp->m3_xm = lx2;	rgp->m3_ym = ly2;	rgp->m3_zm = lz2;
	/* find the maximum value */
	xyzm = rgp->m3_zm * lxy + rgp->m3_ym * lx + rgp->m3_xm;
	valm = *(rgp->m3_xyop + xyzm);
	for (z = 0, flp = rgp->m3_xyop; z < lz; z++)
	 for (y = 0; y < ly; y++) for (x = 0; x < lx; x++, flp++)
	   if (*flp > valm)
	     {
		rgp->m3_xm = x;	rgp->m3_ym = y;	rgp->m3_zm = z; valm = *flp;
	     }
	if (valm < FUZZ) return 0; /* give up if no intensity */
	/* if there's more than one frame and
	 the max is in initial or final frame, quit */
	if ((lz > 1) && ((rgp->m3_zm == 0) || (rgp->m3_zm == lz - 1))) return 0;
	if ((nused = rgp->m3_nobs) < PLENTY)
		rgp->m3_q1 = ((Q1L - Q10) * nused / PLENTY + Q10) * valm;
	else	rgp->m3_q1 = Q1L * valm;
	/* determine the cutoff intensities for the cleanup operations */
	calc_derivs(rgp);
	/* now go through the pixels, starting at the peak pixel
	of the peak frame and moving outward */
	if (thisfmonly > -1) nonzpix = do_zsection(rgp, thisfmonly);
	else for (z = 0, nonzpix = 0; z <= lz2; z++)
	   {
		nonzpix += do_zsection(rgp, lz2 - z);
		if (z) nonzpix += do_zsection(rgp, lz2 + z);
	   }
	return nonzpix;
}

int dospread(REGIONPROFILE *rep, int which, double *sumfm,
 int *nupfm, double floorfr)
{ /*	Add intensity to the images surrounding the centermost one if needed.
   Return values:
	-1 no intensity in the center image
	 0 no change to current model
	 1 added intensity to the image preceding the center
	 2 added intensity to the image following the center
	 3 added intensity to the images preceding and following the center
    the variable `which' is zero if we're doing the (X,Y,omega) space
	models; it is one if we're doing diffraction-vector models. */
	int	i, nnonzl, nnonzr, lx, ly, lxy, lz, lz2, retv;
	float	*vec, *flp, *flpo, *cenl, *cenc, *cenr;
	double	suml, sumc, sumr, vw, maxc;

	if (which)	/* diffraction-vector model */
	  {
		lx = rep->m3_qx; ly = rep->m3_qy; lz = rep->m3_qz;
		vec = rep->m3_dvp;
	  }
	else { /* (X,Y,omega)-vector model */
		lx = rep->m3_bx; ly = rep->m3_by; lz = rep->m3_bz;
		vec = rep->m3_xyop;
	     }
	lxy = lx * ly;	lz2 = lz / 2;
	cenl = vec + (lz2 - 1) * lxy;	/* one image to left of center */
	cenc = cenl + lxy;	cenr = cenc + lxy; /* center and 1 to right*/
	/* sum the intensity of, and find the peak in, the center image */
	for (maxc = 0, flp = cenc, sumc = i = 0; i < lxy; i++, flp++)
	   {
		sumc += *flp;	if (*flp > maxc) maxc = *flp;
	   }
	if ((sumc < 0.05) || (maxc < FUZZ)) return -1; /* not enough Icenter */
	/* sum the intensity in the images on either side */
	for (flp = cenl, suml = i = 0; i < lxy; i++, flp++) suml += *flp;
	for (flp = cenr, sumr = i = 0; i < lxy; i++, flp++) sumr += *flp;
	nnonzl = nnonzr = 0;	retv = 0;
	/* if there's almost nothing to the left, make up some intensity */
	if (sumc * floorfr > suml)
	  {
		retv |= 1;
		for (flpo = cenl, flp = cenc, i = 0; i < lxy; i++,flp++,flpo++)
		   {
			if ((vw = *flp * floorfr * 1.33) > 0.002 * maxc)
			  {
				*flpo = vw;	nnonzl++;
			  }
			else *flpo = 0.;
		   }
	  }
	/* if there's almost nothing to the right, make up some intensity */
	if (sumc * floorfr > sumr)
	  {
		retv |= 2;
		for (flpo = cenr, flp = cenc, i = 0; i < lxy; i++,flp++,flpo++)
		   {
			if ((vw = *flp * floorfr * 1.33) > 0.002 * maxc)
			  {
				*flpo = vw;	nnonzr++;
			  }
			else *flpo = 0.;
		   }
	  }
	sumfm[0] = suml;	sumfm[1] = sumc;	sumfm[2] = sumr;
	nupfm[0] = nnonzl;	nupfm[1] = nnonzr;	return retv;
}

void spreadwmp(MASK_3D *wmp)
{ /*	This examines a 3D profile. If it finds that there is respectable
	intensity in the centermost image, but none or virtually none in
	the images on either side of that, it makes a foreshortened
	version of the center image on the sides. */
	register int	z, y, x;
	int		zl, zc, zr, lx, ly, lz, lz2;
	double		thisden, resc, thisme, maxmax;
	double		summe, maxme, summl, summc, summr;
	PHIMASK		*phic, *phil, *phir, *pip;

	lx = BIGSPOTWID;	ly = BIGSPOTHT;	lz = PFRAMES;	lz2 = lz / 2;
	/* first, find the peak in the z profile */
	for (maxmax = z = 0, pip = &(wmp->phielt[0]); z < lz; z++,pip++)
		if ((thisden = pip->isqsum) > FUZZ)
		  {
			resc = pip->maxf / thisden;
			if (resc > maxmax) maxmax = resc;
		  }
	if (maxmax < FUZZ) return;
	maxmax /= MASKRESCALE;	zc = lz2;	zl = zc - 1; zr = zc + 1;
	pip = &(wmp->phielt[zl]);	summl = summc = summr = 0;
	for (z = zl; z <= zr; z++, pip++)
	   {
		summe = maxme = 0.;
		if ((thisden = pip->isqsum) > FUZZ)
		  {
			resc = pip->maxf / (thisden * maxmax);
			for (y = 0; y < ly; y++)
			   for (x = 0; x < lx; x++)
			      {
				thisme = resc * (double)(pip->xyint[x][y]);
				if (thisme > maxme) maxme = thisme;
				summe += thisme;
			      }
		  }
		if (z == zl) summl = summe;
		else if (z == zc) summc = summe;
		else summr = summe;
	   }
	phic = &(wmp->phielt[zc]);
	if (summc < DFUZZ) return;	/* degenerate case */
	if (summl < 0.001 * summc)
	  {
		phil = &(wmp->phielt[zl]);
		phil->maxf = phic->maxf * 0.04;
		phil->isqsum = thisden = phic->isqsum;
		resc = phil->maxf / (thisden * maxmax);
		for (y = 0; y < ly; y++)
		   for (x = 0; x < lx; x++)
		      {
			thisme = resc * (double)(phic->xyint[x][y]);
			if (thisme > 0.002)
				phil->xyint[x][y] = phic->xyint[x][y];
			else	phil->xyint[x][y] = 0;
		      }
	  }
	if (summr < 0.001 * summc)
	  {
		phir = &(wmp->phielt[zr]);
		phir->maxf = phic->maxf * 0.04;
		phir->isqsum = thisden = phic->isqsum;
		resc = phir->maxf / (thisden * maxmax);
		for (y = 0; y < ly; y++)
		   for (x = 0; x < lx; x++)
		      {
			thisme = resc * (double)(phic->xyint[x][y]);
			if (thisme > 0.002)
				phir->xyint[x][y] = phic->xyint[x][y];
			else	phir->xyint[x][y] = 0;
		      }
	  }
}

void model_ftoc(REGIONPROFILE *rgp)
{ /* This rescales and integerizes the floating-point profile */
	int	i, lbxyz, lqxyz;
	sichar	*scp;
	float	*flp;
	double	rsc;

	/* record the size of the models */
	lbxyz = rgp->m3_bx * rgp->m3_by * rgp->m3_bz;
	lqxyz = rgp->m3_qx * rgp->m3_qy * rgp->m3_qz;
	/* find the most intense real-space point */
	for (flp = rgp->m3_xyop, rsc = i = 0; i < lbxyz; i++, flp++)
		if (*flp > rsc) rsc = *flp;
	if (rsc > 0.) /* rescale and integerize */
	  {
		for (scp = rgp->m3_xyoip, flp = rgp->m3_xyop, i = 0;
		 i < lbxyz; i++, scp++, flp++)
			*scp = (*flp < -rsc) ? -99 :
			 (sichar)(int)(99. * (*flp) / rsc + 0.49999);
	  }
	else	for (scp = rgp->m3_dvip, i = 0; i < lbxyz; i++,scp++) *scp = 0;
	/* find the most intense diffraction-vector-space point */
	for (flp = rgp->m3_dvp, rsc = i = 0; i < lqxyz; i++, flp++)
		if (*flp > rsc) rsc = *flp;
	if (rsc > 0.) /* rescale and integerize */
	  {
		for (scp = rgp->m3_dvip, flp = rgp->m3_dvp, i = 0;
		 i < lqxyz; i++, scp++, flp++)
		 *scp = (*flp < -rsc) ? -99 : 99. * (*flp) / rsc + 0.49999;
	  }
	else	for (scp = rgp->m3_dvip, i = 0; i < lqxyz; i++,scp++) *scp = 0;
}

void cleanmodels(REGIONPROFILE *reb, int which, int recordsum)
{ /*	This examines each model and averages it with its neighbors
	if there's nothing there. `which' is zero if we're doing (X,Y,omega)
	profiles; it's one if we're doing diffraction-vector profiles.
	`recordsum' regulates whether we're recording the maximum intensity
	as rgp->m3_xyoi (or rgp->m3_dvi). */
	int		imo, iz, imoq, lxyz, nmodels;
	int		isblank[NREGIONS];
	float		*zv, *zvp, *zvq0, *zvq;
	double		dscale, mul;
	REGIONPROFILE	*rgp, *rgq;
	static float	ine[NREGIONS * NREGIONS] =
	 { 0.00,0.20,0.02,0.14,0.20,0.22,0.04,0.04,0.03,0.07,0.03,0.01,
	   0.17,0.00,0.17,0.04,0.10,0.17,0.17,0.04,0.03,0.02,0.07,0.02,
	   0.02,0.20,0.00,0.03,0.04,0.04,0.22,0.20,0.14,0.01,0.03,0.07,
	   0.20,0.05,0.01,0.00,0.20,0.15,0.08,0.04,0.01,0.20,0.05,0.01,
	   0.18,0.06,0.01,0.19,0.00,0.17,0.09,0.04,0.01,0.18,0.06,0.01,
	   0.03,0.18,0.03,0.05,0.18,0.00,0.18,0.08,0.03,0.03,0.18,0.03,
	   0.03,0.18,0.03,0.03,0.08,0.18,0.00,0.18,0.05,0.03,0.18,0.03,
	   0.01,0.06,0.18,0.01,0.04,0.09,0.17,0.00,0.19,0.01,0.06,0.18,
	   0.01,0.05,0.20,0.01,0.04,0.08,0.15,0.20,0.00,0.01,0.05,0.20,
	   0.07,0.03,0.01,0.14,0.20,0.22,0.04,0.04,0.03,0.00,0.20,0.02,
	   0.02,0.07,0.02,0.03,0.04,0.17,0.17,0.10,0.04,0.17,0.00,0.17,
	   0.01,0.03,0.07,0.03,0.04,0.04,0.22,0.20,0.14,0.02,0.20,0.00 };

	/* look to see what's there and normalize if possible */
	nmodels = reb->m3_nmodels;
	for (rgp = reb, rgq = reb + nmodels, imo = 0;
	 imo < nmodels; imo++, rgp++, rgq++)
	   {
		if (!which)
		     {
			lxyz = rgp->m3_bx * rgp->m3_by * rgp->m3_bz;
			zv = rgp->m3_xyop;
		     }
		else {
			lxyz = rgp->m3_qx * rgp->m3_qy * rgp->m3_qz;
			zv = rgp->m3_dvp;
		     }
		isblank[imo] = 0;
		for (dscale = iz = 0, zvp = zv; iz < lxyz; iz++, zvp++)
			if (*zvp > dscale) dscale = *zvp;
		if (fabs(dscale) < FUZZ) isblank[imo] = 1;
		else for (iz = 0, zvp = zv; iz < lxyz; iz++, zvp++)
			*zvp /= dscale;
		if (recordsum)
		  {
			if (which)	rgp->m3_dvi = rgq->m3_dvi = dscale;
			else		rgp->m3_xyoi = rgq->m3_xyoi = dscale;
		  }
	   }
	for (rgp = reb, imo = 0; imo < nmodels; imo++, rgp++)
	   if (isblank[imo])
	   {
		if (!which)
		  {
			lxyz = rgp->m3_bx * rgp->m3_by * rgp->m3_bz;
			zv = rgp->m3_xyop;
		  }
		else {
			lxyz = rgp->m3_qx * rgp->m3_qy * rgp->m3_qz;
			zv = rgp->m3_dvp;
		     }
		for (rgq = reb, imoq = 0; imoq < nmodels; imoq++, rgq++)
		  if (!isblank[imoq])
		   {
			zvq0 = (which) ? rgq->m3_dvp : rgq->m3_xyop;
			mul = ine[imo * nmodels + imoq];
			for (iz = 0, zvp = zv, zvq = zvq0; iz < lxyz;
			 iz++, zvq++, zvp++) *zvp += *zvq * mul;
		   }
		for (dscale = iz = 0, zvp = zv; iz < lxyz; iz++, zvp++)
			if (*zvp > dscale) dscale = *zvp;
		if (dscale > FUZZ)
		  for (iz = 0, zvp = zv; iz < lxyz; iz++, zvp++)
			*zvp /= dscale;
		if (recordsum)
		  {
			rgq = rgp + nmodels;
			if (which)	rgp->m3_dvi = rgq->m3_dvi = dscale;
			else {	
				rgp->m3_xyoi = rgq->m3_xyoi = dscale;
				/* make it look like something's there */
				rgp->m3_nobs = 1; rgp->m3_isqsum = 1.;
			     }
		  }
	     }
}

/* end of xgenmask.c */
