/* xgenmath.c -- mathematical utilities for XGEN

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	 6 Jul 2007: getorthog & parallel: compare with DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	26 Dec 2006: lu_inv is now extern: we use it elsewhere.
	29 Nov 2006: mmt3, mmt31 routines added and made global
	11 Jun 2006: lu_inv is now static too; delete trans33 since it's unused
	 1 Jun 2006: smartave is now static
	20 Nov 2005: inline argument declarations and strict prototyping
	16 Feb 2005: smartave() provides careful angle calcs for qtoangs()
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	28 Aug 2004: routines used only here are declared as static; cosmetics;
		mymedian() and imedian() moved here from elsewhere
	 2 Jul 2004: remove redundant #defines now that we're including xgen.h
	27 Jun 2004: incorporate inverse.c into here
	21 Jan 2004: license message inserted
	14 Jan 2004: add trans33(); efficient transposer for 3x3 matrices
	19 Dec 2003: change dummy variable lambda to elam in eig vec()
		to eliminate confusion with the X-GEN global variable
	27 Nov 2003: cosmetic changes
	18 Sep 2003: declare zeromem() as void (big deal);
		all external X-GEN routines declared here, not in xdefs.h
	 5 Jan 2003: reorganized source-file header
	20 Oct 2002: bugfix in recen_vec(): set uzp to 0 only
		if k2 > nz - 1, not when k2 == nz - 1.
	25 Mar 2002: removed MSI attachments
	 8 Nov 1996: minor changes for X-GEN checkin at MSI
	20 Nov 1995: fix bug on irix6.x in inv_solve: 32-bit peculiarity fixed.
	10 Oct 1995: local static arrays handle small matrix inversions to
		cut down on use of the dynamic memory handler
	 6 Jul 1995: gen_centroid: fix bug when 3-D profile wants to include
		images before or after the current range of images.
	31 May 1995: "register" -> "register int" throughout
	12 May 1995: move routine mycbrt(zx) to this file
	 2 Mar 1995: trivial formatting
	 5 May 1994: initial X-GEN version (culled from other files)
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial revision
	
   Purpose: mathematical functions used in data processing and reduction with
	XGEN, including some routines for inverting matrices and using them.
   Cross references: xgenio.c

*************************************************************************/
#include	"xgen.h"
#define		RESCALE		(32.)	/* rescaling constant in "inv3" */
#define		ALLTOOSMALL	(-1)	/* Flag: all elts of matrix~~0 */
#define		DETERZERO	1	/* flag: determinant == 0 */
#define		NEWT_CONVERGED	(0.1E-5)	/* convergence for zero test */
#define		MAXTRIES	(100)	/* max # cycles of Newton's method*/
#define		MOVEWEIGHT	(0.086643397569993)	/* ln(2.) / 8. */
#define		FUZZ2	(0.1e-4)
#define		MIDSIZE		500
#define		MIN_MEDSIZE	3200	/* used in calculations of medians */
#define		AVERAGE_PASSES	(5)	/* mean # passes before convergence */
#define		OKNORM		(0.1e-5)	/* good Euclidean matrix norm*/

extern int lu_inv(double *a, double *ai, int n);

short min2(short a, short b)
{ /* return the smaller of two shorts */
	return(a < b ? a : b);
}

short max2(short a, short b)
{ /* return the larger of two shorts */
	return(a > b ? a : b);
}

void inv_init(double *ls, double *rs, int n)
{ /*	This zeroes the two vectors required in this type of
	equation-solving.  See inv_add for the form. */
	int	uptri;
	double	*fp;

	uptri = n * (n + 1) / 2;
	for (fp = rs; n--; fp++) *fp = 0.;
	for (fp = ls; uptri--; fp++) *fp = 0.; 
}

void inv_add(double *ls, double *rs, double *av, double b, int n)
{ /*	This adds to the sums required to solve the least-squares problem
		min {(Ax-b)}~{(Ax-b}
	where "~" denotes "transpose".  The vectors 'av' and 'rs' are of
	length n; the vector `ls' is of length n(n+1)/2 and is set up thus:
			L(1)	L(2)	. . .	L(n)
			L(2)	L(n+1)	. . .	L(2n-1)
	A~W~WA	= 	.	.	. . .	.
			.	.	. . .	.
			L(n)	L(2n-1) . . .	L(n(n+1)/2)
	The inverses this routine leads to are done in inv_solve. */
	register int	i, j;
	double		*rp, *lp, *api, *apj;
	double		delt;

	for (i = 0, rp = rs, lp = ls, api = av; i < n; i++, rp++, api++)
	   {
		delt = b * (double)(*api);	*rp += delt;
		for (j = i, apj = api; j < n; j++, lp++, apj++)
			*lp += *api * *apj;
	   }
}

int lu_inv(double *a, double *ai, int n)
{ /*	This inverts a matrix by converting it to the product of upper
	and lower triangular forms and then inverting those.
	This returns 1 if the inversion fails, -1 if memory can't be allocated,
	and 0 on success. Note that the args are doubles now. */
	register int	i, j, k;
	int		q, qj;
	unsigned	nsq;
	double		*lu, *lui, ahold;
	static double	smallf[4 * MIDSIZE];
	static char	lumsg[] = "lu in lu_inv";

	/* reserve memory for the lower and upper-triangular matrices */
	nsq = n * n;
	if (nsq < MIDSIZE)
	  {
		lu = smallf;	zeromem((void *)lu, 2 * nsq * sizeof (double));
	  }
	else if (NULL == (lu = (double *)callop(2 * nsq,
	 sizeof (double), lumsg))) return -1;
	lui = lu + nsq;
	/* do it:
	 first, construct the lower-triangular and upper-triangular matrices.
	'q' is the running index of the upper-triangular matrix;
	'qj' is the running index of the lower-triangular matrix. */
	for (i = 0; i < n; i++)
	   {
		q = i * n + i;	lu[q] = a[q];
		for (k = 0; k < i; k++) lu[q] -= lu[i*n+k] * lu[k*n+i];
		if (fabs(ahold = lu[q]) < FUZZ)
		  {
			if (nsq >= MIDSIZE) memzap((AVOID **)&lu, lumsg);
			return 1;
		  }
		for (j = i + 1, q++; j < n; j++, q++)
		   {
			lu[q] = a[q];	qj = j * n + i;	lu[qj] = a[qj];
			for (k = 0; k < i; k++)
			   {
				lu[qj] -= lu[j * n + k] * lu[k * n + i];
				lu[q]  -= lu[i * n + k] * lu[k * n + j];
			   }
			lu[qj] /= ahold;
		   }
	   }
	/*invert the lower-triangular part of `lu'
	(the elements with i < j).  We build from the top down.*/
	for (i = 1; i < n; i++)
	   {
		for (j = 0, q = n * i; j < i; j++, q++)
		   {
			lui[q] = -lu[q];
			for (k = 1; k < i; k++)
				lui[q] -= lu[i * n + k] * lui[k * n + j];
		   }
	   }
	/* invert the upper-triangular part of `lu'.  We build this by
	rows from the bottom up, but from left to right within a row. */
	for (i = n - 1; i >= 0; i--)
	   {
		q = i * n + i;
		if (fabs(lu[q]) < FUZZ)
		  {
			if (nsq >= MIDSIZE) memzap((AVOID **)&lu, lumsg);
			return 1;
		  }
		lui[q] = 1. / lu[q];
		for (j = i + 1, qj = q + 1; j < n; j++, qj++)
		   {
			for (lui[qj] = 0, k = i; k < j; k++)
				lui[qj] -= lui[i * n + k] * lu[k * n + j];
			lui[qj] *= lui[j * n + j];
		   }
	   }
	/* get the resulting inverse by matrix multiplication */
	for (i = q = 0; i < n; i++)
	   for (j = 0; j < n; j++, q++)
	      {
		if (i <= j)
		  {
			qj = j + 1;	ai[q] = lui[q];
		  }
		else {
			qj = i;		ai[q] = 0.;
		     }
		for (k = qj; k < n; k++)
			ai[q] += lui[i * n + k] * lui[k * n + j];
	      }
	if (nsq >= MIDSIZE) memzap((AVOID **)&lu, lumsg);
	return 0;
}

int inv_solve(double *ls, double *rs, double *p, int n)
{ /*	This returns the results of solving the least-squares problem
	min{(Ap-b)}~{(Ap-b)} in the vector p.  The left-hand side
	matrix A~A has been accumulated implicitly in `ls';
	the righthand-side vector A~W~Wb has been accumulated in `rs'.
	The routine returns -1 if memory can't be allocated;
	it returns 0 if it's successful and 1 if it isn't able to invert
	the lefthand matrix. */
	register int	i, j, k;
	int		ntri, iu, iv, iw;
	unsigned	nsq;
	double		*lg, *lgi, *stank, inmax, thisv, rescale;
	static double	smallf[4 * MIDSIZE];
	static char	stmsg[] = "stank in inv_solve";

/* Reserve memory for the working vectors */
	nsq = n * n;	ntri = (n * (n + 1)) / 2;
	if (nsq < MIDSIZE)
	  {
		stank = smallf;
		zeromem((void *)stank, 2 * nsq * sizeof (double));
	  }
	else if (NULL == (stank = (double *)callop(2 * nsq,
	 sizeof (double), stmsg))) return -1;
	lg = stank;	lgi = stank + nsq;
	/* Construct the lefthand matrix `lg'.  We want to keep the values
	in it fairly reasonable, so we divide 'em all through by
	the largest element / 32. */
	for (inmax = 0., k = 0; k < ntri; k++)
	   if ((thisv = fabs(ls[k])) > inmax) inmax = thisv;
	if (fabs(inmax) < FUZZ)
	  {
		if (nsq >= MIDSIZE) memzap((AVOID **)&stank, stmsg);
		return 1;
	  }
	rescale = inmax / 32.;
	for (k = 0; k < ntri; k++) ls[k] /= rescale;
	for (i = k = iu = 0; i < n; i++, iu += n + 1)
	   {
		lg[iu] = ls[k];	k++;	iv = iu + 1;	iw = iu + n;
		for (j = i + 1; j < n; j++, k++, iv++, iw += n)
			lg[iv] = lg[iw] = ls[k];
	   }
/* invert the little monster */
	if (lu_inv(lg, lgi, n))
	  {
		if (nsq >= MIDSIZE) memzap((AVOID **)&stank, stmsg); return 1;
	  }
/* Multiply it on the right by the righthand side matrix */
	for (i = k = 0; i < n; i++)
		for (p[i] = j = 0; j < n; j++, k++) p[i] += lgi[k] * rs[j];
/* Restore rescaling */
	for (i = 0; i < n; i++) p[i] /= rescale;
/* Free memory recently reserved */
	if (nsq >= MIDSIZE) memzap((AVOID **)&stank, stmsg);	return 0;
}

float flimits(float z, float low, float high)
{ /* returns z if low <= z <= high; returns low or high if outside */
	return (z < low ? low : (z > high ? high : z));
}

double dflimits(double z, double low, double high)
{ /* returns z if low <= z <= high; returns low or high if outside */
	return (z < low ? low : (z > high ? high : z));
}

double mycbrt(double rx)
{ /* This returns the sign-correct cube root of a real value x */
	return (rx < 0.) ? -pow(-rx, 0.33333333333) : pow(rx, 0.33333333333);
}

float deter(float xi[3][3])
{ /* This finds the determinant of the floating-point 3x3 matrix xi.
	We go to some lengths to avoid underflow. */
	register int	i, j, k;
	double		determ, elt, maxelt, x[3][3];

	determ = maxelt = 0.;	/* clear rescaler & determinant */
	for (i = 0; i < 3; i++)	/* get rescaling constant */
	   for (j = 0; j < 3; j++)
	      {
		elt = fabs(xi[i][j]);	if (elt > maxelt) maxelt = elt;
	      }
	if (maxelt < FUZZ) return 0;
	maxelt /= RESCALE;	/* rescale by this constant */
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		x[i][j] = xi[i][j] / maxelt;
	for (i = 0; i < 3; i++)
	   {
		j = (i + 1) % 3;	k = 3 - i - j;
		determ += x[0][i] * (x[1][j] * x[2][k] - x[1][k] * x[2][j]);
	   }
	determ *= maxelt; determ *= maxelt; determ *= maxelt;
	return (float)determ;
}

double ddeter(double xi[3][3])
{ /* This finds the determinant of the 3x3 matrix xi.
	We go to some lengths to avoid underflow. */
	register short	i, j, k;
	double		determ, elt, maxelt, x[3][3];

	determ = maxelt = 0.;	/* clear rescaler & determinant */
	for (i = 0; i < 3; i++)	/* get rescaling constant */
	   for (j = 0; j < 3; j++)
	      {
		elt = fabs(xi[i][j]);	if (elt > maxelt) maxelt = elt;
	      }
	if (maxelt < FUZZ) return 0;
	maxelt /= RESCALE;	/* rescale by this constant */
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		x[i][j] = xi[i][j] / maxelt;
	for (i = 0; i < 3; i++)
	   {
		j = (i + 1) % 3;	k = 3 - i - j;
		determ += x[0][i] * (x[1][j] * x[2][k] - x[1][k] * x[2][j]);
	   }
	determ *= maxelt; determ *= maxelt; determ *= maxelt;
	return determ;
}

void mm3(float c[3][3], float a[3][3], float b[3][3])
{ /* This puts the matrix product a*b into the matrix c. */
	register int	i, j, k;
	double		cva;

	for (i = 3; i--; ) for (j = 3; j--; )
	   {
		for (k = 3, cva = 0.; k--; ) cva += a[i][k] * b[k][j];
		c[i][j] = cva;
	   }
}

void dmm3(double c[3][3], double a[3][3], double b[3][3])
{ /* This puts the double-precision matrix product a*b into matrix c. */
	register int	i, j, k;
	double		cva;

	for (i = 3; i--; ) for (j = 3; j--; )
	   {
		for (k = 3, cva = 0.; k--; ) cva += a[i][k] * b[k][j];
		c[i][j] = cva;
	   }
}

void mm31(float *y, float a[3][3], float *x)
{ /* this left-multiplies the 3x3 matrix a by the 3-elt. col vector
		x and leaves the result in the 3-element column vector y */

	register int	i, k;
	register float	*ay,*ax;
	double		cva;

	for (ay = y, i = 0; i < 3; i++, ay++)
	   {
		for (cva = k = 0, ax = x;  k < 3; k++, ax++)
			cva += a[i][k] * *ax;
		*ay = cva;
	   }
}

void mm3t(float c[3][3], float a[3][3], float b[3][3])
{ /* This puts the matrix product a*b' into the matrix c.
	Here b' is the transpose of b.  */
	register int	i, j, k;
	double		cva;

	for (i = 3; i--; ) for (j = 3; j--; )
	   {
		for (k = 3, cva = 0.; k--; ) cva += a[i][k] * b[j][k];
		c[i][j] = cva;
	   }
}

void dmm3t(double c[3][3], double a[3][3], double b[3][3])
{ /* This puts the matrix product a*b' into the matrix c.
	Here b' is the transpose of b.  */
	register int	i, j, k;
	double		cva;

	for (i = 3; i--; ) for (j = 3; j--; )
	   {
		for (k = 3, cva = 0.; k--; ) cva += a[i][k] * b[j][k];
		c[i][j] = cva;
	   }
}

void mmt3(float c[3][3], float a[3][3], float b[3][3])
{ /* This puts the matrix product a'*b into the matrix c.
	Here a' is the transpose of a.  */
	register int	i, j, k;
	double		cva;

	for (i = 3; i--; ) for (j = 3; j--; )
	   {
		for (k = 3, cva = 0.; k--; ) cva += a[k][i] * b[k][j];
		c[i][j] = cva;
	   }
}

void mmt31(float *y, float a[3][3], float *x)
{ /* this left-multiplies the transpose of a 3x3 matrix a by the 3-elt.
	 col vector x and leaves the result in the 3-element column vector y */

	register int	i,k;
	register float	*ay, *ax;
	double		cva;

	for (ay = y, i = 0; i < 3; i++, ay++)
	   {
		for (cva = k = 0, ax = x;  k < 3; k++, ax++)
			cva += a[k][i] * *ax;
		*ay = cva;
	   }
}

double pirange(double x)
{ /*	This examines its argument and returns a value between
	-PI and PI which has the same trig properties as the argument. */
	if (x < 0)	return (x - (PIX2 * (int)(x / PIX2 - 0.49999)));
	else		return (x - (PIX2 * (int)(x / PIX2 + 0.49999)));
}

static double smartave(double v0, double v1, double wsum, double sdc)
{ /*	This computes three estimates of an angle, based on
	asin(v0/sdc), acos(v1/sdc), and 0.5*acos(wsum).
   The mask "est" indicates which of the three estimates is reliable.
   If it's zero or 7 (all reliable or all unreliable), we use a
   weighted average of all 3. If any one or 2 is unreliable,
   we use only the others. */
	int	est;
	double	zes[3], out;

	est = 0;
	if ((out = fabs(v0 / sdc)) > 1.)
	  {
		est |= 1;	zes[0] = PIO2;
	  }
	else zes[0] = asin(out);
	if ((out = fabs(v1/ sdc)) > 1.)
	  {
		est |= 2;	zes[1] = 0.;
	  }
	else	zes[1] = acos(out);
	if (wsum > 1.)
	  {
		est |= 4;	zes[2] = 0.;
	  }
	else if (wsum < -1.)
	  {
		est |= 4;	zes[2] = PIO2;
	  }
	else zes[2] = 0.5 * acos(wsum);
	switch (est)
	  {
		case	0:	case	7:	default:
		 out = (zes[0] + zes[1] + 4. * zes[2]) / 6.; break;
		case	1: out = (zes[1] + 4. * zes[2]) / 5.; break;
		case	2: out = (zes[0] + 4. * zes[2]) / 5.; break;
		case	3: out = zes[2];	break;
		case	4: out = (zes[0] + zes[1]) / 2.; break;
		case	5: out = zes[1]; break;
		case	6: out = zes[0]; break;
	  }
	if (v0 < -FUZZ) out = -out;
	if (v1 > FUZZ) out = PI - out;
	return out;
}

void qtoangs(double q[3][3], double *rph)
{ /*	This routine converts a 3x3 orthogonal transformation matrix q
	into three angles rph[0-2] = -delta_omega,-delta_chi,-delta_phi.
	Q is assumed to be reasonably accurate to begin with, although some
	averaging is done inside this routine to make things consistent. */
	register int	i;
	double		sindchi, sdcsq, w0, w1, lrph[3];

	sdcsq = 0.25 * (3. + q[0][2] * q[0][2] + q[1][2] * q[1][2] +
		  q[2][0] * q[2][0] + q[2][1] * q[2][1] - q[0][0] * q[0][0] -
		  q[1][0] * q[1][0] - q[0][1] * q[0][1] - q[1][1] * q[1][1] -
		  q[2][2] * q[2][2]);	/* sin(squared) of delta_chi */
	if (sdcsq > FUZZ) /* lrph[1] nonzero */
	  {
		if ((sindchi = sqrt(sdcsq)) > 1.)
		  {
			sindchi = 1.;	lrph[1] = PIO2;
		  }
		else lrph[1] = asin(sindchi); /* assume lrph[1] > 0 at start */
		if (q[2][2] < FUZZ) lrph[1] = PI - lrph[1];
		w0 = (q[0][0]*q[0][0] - q[1][1]*q[1][1]) / sdcsq;
		w1 = (q[1][0]*q[1][0] - q[0][1]*q[0][1]) / sdcsq;
		lrph[0] = smartave(q[0][2], q[1][2], w0 - w1, sindchi);
		lrph[2] = smartave(-q[2][0], -q[2][1], w0 + w1, sindchi);
	/* one more consistency check: if |rph[1]| < 0, flip 'em all...*/
		if ((fabs(q[1][0]) > FUZZ) &&
			((-cos(lrph[0]) * cos(lrph[1]) * sin(lrph[2]) +
			sin(lrph[0]) * cos(lrph[2])) / q[1][0]) < -0.5)
		  {
			lrph[1] = -lrph[1];	sindchi = -sindchi;
			lrph[0] = -lrph[0];	lrph[2] = -lrph[2];
		  }
	  }
	else	/* lrph[1]=0, special case: SET lrph[0]=0, get lrph[2].*/
	  {
		lrph[0] = lrph[1] = 0;
		lrph[2] = (asin(q[0][1]) - asin(q[1][0])) / 2.;
		if ((fabs(q[0][0]) > FUZZ) && (cos(lrph[2]) / q[0][0]) < -0.5)
			lrph[2] = PI - lrph[2];
	  }
	/* convert to single-precision
	 and make sure all angles are between -PI and PI */
	for (i = 3; i--; ) rph[i] = pirange(lrph[i]); 
}

static double eunorm(double m[3][3])
{ /*	this computes the double-precision Euclidean norm of the
	matrix I - MM', where M is the input matrix and I is a 3x3 unit matrix.
	The resulting value is returned.*/
	register int	i, j, k;
	double		rnorm, elt;

	for (i = 3,rnorm = 0.; i--; ) for (j = 3; j--; )
	   {
		for (elt = (i == j), k = 3; k--; ) elt -= m[i][k] * m[j][k];
		rnorm += elt * elt;
	   }
	return rnorm;
}

int getorthog(double x[3][3])
{ /*	This performs a Gram Schmidt orthonormalization of a matrix
	in which the first two columns are given and the last is computed.
	This constructs the 3rd column of a 3x3 as the cross product
	of the first two columns.  Thus it's normal to the other
	two and the determinant of the resulting matrix is non-negative.
	This routine returns 0 if successful, 1 if not.
	The only causes for failure are for the two vectors to be too small
	or parallel to one another. */
	register int	i, j, k;
	double		rnorm, dot;

	rnorm = x[0][0] * x[0][0] + x[1][0] * x[1][0] + x[2][0] * x[2][0];
	if (rnorm < DFUZZ) return 1;
	for (i = 3, rnorm = sqrt(rnorm); i--; ) x[i][0] /= rnorm;
	dot = x[0][0] * x[0][1] + x[1][0] * x[1][1] + x[2][0] * x[2][1];
	for (i = 3, rnorm = 0; i--; )
	   {
		x[i][1] -= dot * x[i][0]; rnorm += x[i][1] * x[i][1];
	   }
	if (rnorm < DFUZZ) return 1;
	for (i = 3, rnorm = sqrt(rnorm); i--; ) x[i][1] /= rnorm;
	for (i = 3; i--; )
	   {
		j = (i == 2? 0 : i+1);	k = 3 - i - j;
		x[i][2] = x[j][0] * x[k][1] - x[j][1] * x[k][0];
	   }
	return 0;
}

int parallel(float *v0, float *v1, int n)
{ /*	This returns 1 iff. the n-length vectors v0 and v1 are either parallel
	or antiparallel.  The null vector is parallel to everything. */
	register int	i;
	double		p0, p1, p01;

	for (p0 = p1 = p01 = 0., i = n; i--; v0++, v1++)
		p0 += (*v0 * *v0); p1 += (*v1 * *v1); p01 += (*v0 * *v1);
	if (p0 < DFUZZ || p1 < DFUZZ) return 1;
	return ((1. - fabs((p01/p0) * (p01 / p1))) < DFUZZ);
}

float dotprod(float *v1, float *v2, int n)
{ /*	This returns the dot product of the two vectors v1 and v2.
	n is the length of the two vectors. */
	float	*p1, *p2;
	double	prod;

	for (prod = 0, p1 = v1, p2 = v2; n--; p1++, p2++) prod += *p1 * *p2;
	return prod;
}

double ddotprod(double *v1, double *v2, int n)
{ /*	This returns the double-precision dot product of the two vectors
	v1 and v2.  n is the length of the two vectors. */
	double	prod, *p1, *p2;

	for (prod = 0, p1 = v1, p2 = v2; n--; p1++, p2++) prod += *p1 * *p2;
	return prod;
}

int eigvec(double m[3][3], double elam, double *evec)
{ /*	this computes the eigenvector associated with the eigenvalue
	elam of the 3x3 matrix m.  It returns 0 if successful,
	-1 if it can't find an eigenvector (all 3 solutions fully degenerate!),
	and -2 if somehow the eigenvector can't be normalized.
	April 2007: this is now all done in double-precision. */

	register int	i, j;
	double		g[3][3], wx[3], denom, den2;

	/* construct the characteristic equation */
	for (i = 3; i--; ) for (j = 3; j--; ) g[i][j] = m[i][j];
	for (i = 3; i--; ) g[i][i] -= elam;
	wx[0] = 1.;	/* x coordinate of eigenvector starts at 1 */
	denom = g[1][2] * g[0][1] - g[0][2] * g[1][1];
	if (fabs(denom) > FUZZ && fabs(den2 = g[0][1]) > FUZZ)
	  {
		wx[2] = (g[0][0] * g[1][1] - g[0][1] * g[1][0]) / denom;
		wx[1] = -(g[0][0] + g[0][2] * wx[2]) / den2;
	  }
	else {
		denom = g[2][2] * g[1][1] - g[1][2] * g[2][1];
		if (fabs(denom) > FUZZ && fabs(den2 = g[1][1]) > FUZZ)
		  {
			wx[2] = (g[1][0] * g[2][1] - g[1][1] * g[2][0]) / denom;
			wx[1] = -(g[1][0] + g[1][2] * wx[2]) / den2;
		  }
		else {
			denom = g[0][2] * g[2][1] - g[2][2] * g[0][1];
			if (fabs(denom) > FUZZ && fabs(den2 = g[2][1]) > FUZZ)
			  {
				wx[2] = (g[2][0] * g[0][1] - g[2][1] * g[0][0])
					/ denom;
				wx[1] = -(g[2][0] + g[2][2] * wx[2]) / den2;
			  }
			/* if we still haven't found a solution, maybe wx[0]=0:
			then the 3 solutions must be fully degenerate... */
			wx[0] = 0.;	wx[2] = 1.;
			if (fabs((denom = g[0][1])) > FUZZ)
				wx[1] = -g[0][2] / denom;
			/* if it STILL doesn't work, there ain't no eigenvec.*/
			else {
				wx[0] = wx[1] = wx[2] = 0.;	return -1;
			     }
		     }
	     }
	for (i = 3, denom = 0; i--; ) denom += wx[i] * wx[i];
	if (denom < FUZZ) return -2;	denom = sqrt(denom);
	for (i = 3; i--; ) evec[i] = wx[i] / denom;
	return 0;
}

double newtcube(double root, double a2, double a1, double a0)
{ /*	This computes a root of the cubic equation
		x*x*x + a2*x*x + a1*x + a0 = 0
	in the vicinity of the input value `root' by Newton's method. */
	int	ntry;
	double	xtry, dfdx0, resid;

	for (xtry = root, ntry = 0,
	 resid = a0 + root * (a1 + root * (a2 + root)); ntry < MAXTRIES;
	  ntry++, resid = a0 + xtry * (a1 + xtry * (a2 + xtry)))
	   {
		if (fabs(resid) < NEWT_CONVERGED) break;
		dfdx0 = a1 + xtry * (2. * a2 + 3. * xtry);
		if (fabs(dfdx0) < FUZZ) return xtry;
		xtry -= (resid / dfdx0);
	   }
	return xtry;
}

short iroup(double x)
{	/*	this rounds a real number UP to the next higher integer	*/
	register short	j;

	j = x;	if (j < x) j++;	return	j;
}

short irodown(double x)
{	/* this rounds a real number DOWN to the next lower integer */
	register short	j;	j = x;	if (j > x) j--;	return	j;
}

static float gen2d(float *vd, float *vv, int nx, int ny,double *vc,double *vob)
{ /*	This routine finds the (XY) centroid of the 2-D data vector `vd'.
	vd is of length nx*ny and is indexed with x fastest and y slowest.
	If `vv' is non-null it contains a vector of variance values used in
	determining the reliability of the result.  In particular it assumes
	that the `predicted' centroid of the vector is vc[0-1], and the output
	centroid is taken to be
	 <x> = (wo * <X>o + wc * <X>c) / (wo + wc) = wo * <X>o / (wo + wc)
	where the observed centroid <X>o is determined by ordinary centroid
	calculation methods. The weights wo are determined from the variances of
	the individual observations and the weight wc is taken to be the inverse
	of the expected mean-square error in predicted coordinates. Thus with
	 <X>o = SUM (xi * Ii) / SUM Ii = m1 / m0, then
	 vo = var<X>o = var (m1 / m0) = (v1 + (m1/m0)^2 * v0) / (m0^2)
	where v0 = SUM var Ii and v1 = SUM xi * var Ii, and
	 vc = (0.4 pixels)^2 for (X) and (Y) and
	 vc = (0.3 frames)^2 for (Z).  Later we'll let those vary.
	If vv is NULL we assume that we're working with model data for which
	the variances in the individual values are insignificant. */
	register int	iy, ix;
	int		ny2, nx2, ixm1, ixp1;
	int		ixmin, ixmax, iymin, iymax;
	double		diy, dix, tso, tsp, tsa;
	double		ts, tsx, tsy, vs, tfmax;
	static float	*vdy, *vdu, *vdd, *t, *vvy, *vt;

	ny2 = ny / 2;	nx2 = nx / 2;
/* determine the centroid in (x,y) of `vd'.
 We're very conservative about this to avoid moving over
 to the next reflection.  First, we find the brightest five-point diamond
 within the central 5x5: */
	diy = ny2;	dix = nx2;
	tfmax = -10000.;
	vdy = vd + nx * (ny2 - 2);
	vdu = vdy - nx; vdd = vdy + nx;
	for (iy = ny2 - 2; iy <= ny2 + 2; iy++, vdy += nx, vdd += nx,vdu += nx)
	   {
		for (ix = nx2 - 2, ixm1 = nx2 - 3, ixp1 = nx2 - 1;
		 ix <= nx2 + 2; ix++, ixm1++, ixp1++)
		   {
			ts = *(vdu + ix) + *(vdy + ixm1) + *(vdy + ix) +
			*(vdy + ixp1) + *(vdd + ix);
			if (ts > tfmax)
			  {
				tfmax = ts; dix = ix; diy = iy;
			  }
		   }
	   }
	if (tfmax < FUZZ) return 0.;
	tfmax /= 2.; /* approx. peak height */
	/* now do a conventional centroid calc using a 5x5 centered
	at the above point.  Use the variances if they're there. */
	/* starting and end points of each of the 5x5's may be outside the
	7x7 at this point.  If so, just use the part we have. */
	iymin = diy - 2;	if (iymin < 0) iymin = 0;
	iymax = diy + 2;	if (iymax >= ny) iymax = ny - 1;
	ixmin = dix - 2;	if (ixmin < 0) ixmin = 0;
	ixmax = dix + 2;	if (ixmax >= nx) ixmax = nx;
	tsx = tsy = 0.;
	if ((float *)0 == vv)
	  {
		tsa = ts = 0;
		vdy = vd + nx * iymin;
		for (iy = iymin; iy <= iymax; iy++, vdy += nx)
		   {
			diy = iy - ny2;	t = vdy + ixmin;
			for (ix = ixmin; ix <= ixmax; ix++, t++)
			   {
				dix = ix - nx2;
				if (*t > 0.)
				  {
					ts += *t; tsx += *t * dix;
					tsy += *t * diy;
				  }
				tsa += *t; /* include all points */
			   }
		  }
		vs = 1.;
	  }
	else {
		tsa = ts = vs = 0;
		vdy = vd + nx * iymin;
		vvy = vv + nx * iymin;
		for (iy = iymin; iy <= iymax; iy++, vdy += nx, vvy += nx)
		   {
			diy = iy - ny2;
			t = vdy + ixmin; vt = vvy + ixmin;
			for (ix = ixmin; ix <= ixmax; ix++, t++, vt++)
			   {
				dix = ix - nx2;
				if (*t > 0.)
				  {
					ts += *t; tsx += *t * dix;
					tsy += *t * diy;
				  }
				tsa += *t;	vs += *vt;
			   }
		   }
	     }
	if ((ts < FUZZ) || (vs < FUZZ))
	  {
		vob[0] = vc[0];	vob[1] = vc[1];	vob[2] = vc[2];
		return 0.;
	  }
	tsx /= ts;	tsy /= ts;
	/* we return two versions of the `observed' centroid: one that is
	the actual observed value, and the other the weighted mean of the
	observed and predicted centroids.  The former is recorded on disc,
	whereas the latter is the one actually used for centering the spot.
	The weight value placed on the observed centroid is
	1 - exp(-G[(I/sig)^2 - 1]) for I/sig > 1 (it's zero otherwise).
	Currently G is set to (ln 2)/8 = 0.086643397569993; this allows
	the weight factor to be 0.5 at I/sig = 3. */
	vob[0] = tsx;	vob[1] = tsy;
	if ((float *)0 != vv)
	  {
		tsp = (ts / vs) * ts - 1.;
		tsp = (tsp > 0. ? exp(-(MOVEWEIGHT) * tsp) : 1.);
		tso = 1. - tsp;
		tsx = vob[0] * tso + vc[0] * tsp;
		tsy = vob[1] * tso + vc[1] * tsp;
	  }
	vc[0] = tsx;	vc[1] = tsy;
	return tsa;	/* this is an intensity estimate, bucko! */
}

float gen_centroid(float *vd, float *vv, int nx, int ny, int nz,
 double *vc, double *vob)
{ /*	This routine finds the (XYZ) centroid of the 3-D data vector `vd'.
	vd is of length nx*ny*nz and is indexed with x fastest and z slowest.
	If `vv' is non-null it contains a vector of variance values used in
	determining the reliability of the result.  In particular it assumes
	that the `predicted' centroid of the vector is vc[0-2], and the output
	centroid is taken to be
	 <x> = (wo * <X>o + wc * <X>c) / (wo + wc) = wo * <X>o / (wo + wc)
	where the observed centroid <X>o is determined by ordinary centroid
	calculation methods. The weights wo are determined from the variances of
	the individual observations and the weight wc is taken to be the inverse
	of the expected mean-square error in predicted coordinates. Thus with
	 <X>o = SUM (xi * Ii) / SUM Ii = m1 / m0, then
	 vo = var<X>o = var (m1 / m0) = (v1 + (m1/m0)^2 * v0) / (m0^2)
	where v0 = SUM var Ii and v1 = SUM xi * var Ii, and
	 vc = (0.4 pixels)^2 for (X) and (Y) and
	 vc = (0.3 frames)^2 for (Z).  Later we'll let those vary.
	If vv is NULL we assume that we're working with model data for which
	the variances in the individual values are insignificant. */
	register int	iz, iy, ix;
	int		nz2, ny2, nx2, nxy, ixm1, ixp1;
	int		ixmin, ixmax, iymin, iymax, izmin, izmax;
	double		diz, diy, dix, tso, tsp, tsa;
	double		ts, tsx, tsy, tsz, vs, tfmax;
	static float	*vdy, *vdu, *vdd, *t, *vvy, *vt;

	/* treat the nz == 1 case specially */
	if (nz < 2)
	  {
		ix = gen2d(vd, vv, nx, ny, vc, vob);
		vc[2] = vob[2] = 0.;	return ix;
	  }
	nz2 = nz / 2;	ny2 = ny / 2;	nx2 = nx / 2;	nxy = nx * ny;
	/* determine the centroid in (x,y,z) of `vd'.
	We're very conservative about this to avoid moving over
	to the next reflection.  First, we find the brightest five-point
	diamond within the central 3x5x5: */
	diz = nz2;	diy = ny2;	dix = nx2;
	for (iz = nz2 - 1, tfmax = -10000.; iz <= nz2 + 1; iz++)
	   {
		vdy = vd + nxy * iz + nx * (ny2 - 2);
		vdu = vdy - nx; vdd = vdy + nx;
		for (iy = ny2 - 2; iy <= ny2 + 2;
		 iy++, vdy += nx, vdd += nx, vdu += nx)
		 {
		   for (ix = nx2 - 2, ixm1 = nx2 - 3, ixp1 = nx2 - 1;
		    ix <= nx2 + 2; ix++, ixm1++, ixp1++)
		      {
			ts = *(vdu + ix) + *(vdy + ixm1) + *(vdy + ix) +
				*(vdy + ixp1) + *(vdd + ix);
			if (ts > tfmax)
			  {
				tfmax = ts; dix = ix; diy = iy; diz = iz;
			  }
		      }
		 }
	   }
	if (tfmax < FUZZ) return 0.;
	tfmax /= 2.; /* approx. peak height */
	/* now do a conventional centroid calc using a 3x5x5 or 5x5x5 centered
	at the above point.  This time we use the variances if they're there.
	19 Jun 95: check for edge effects first. */
	izmin = diz - 1; izmax = diz + 1;
	if (nz2 > 3)
	  {
		izmin--;	izmax++;
	  }
	if (izmin < 0) izmin = 0;	if (izmax >= nz) izmax = nz - 1;
	/* starting and end points of each of the 5x5's may be outside the
	7x7 at this point.  If so, just use the part we have. */
	iymin = diy - 2;	if (iymin < 0) iymin = 0;
	iymax = diy + 2;	if (iymax >= ny) iymax = ny - 1;
	ixmin = dix - 2;	if (ixmin < 0) ixmin = 0;
	ixmax = dix + 2;	if (ixmax >= nx) ixmax = nx;
	tsx = tsy = tsz = 0.;
	if ((float *)0 == vv)
	  {
		for (tsa = ts = 0, iz = izmin; iz <= izmax; iz++)
		   {
			diz = iz - nz2;
			vdy = vd + iz * nxy + nx * iymin;
			for (iy = iymin; iy <= iymax; iy++, vdy += nx)
			   {
				diy = iy - ny2;	t = vdy + ixmin;
				for (ix = ixmin; ix <= ixmax; ix++, t++)
				   {
					dix = ix - nx2;
					if (*t > 0.)
					  {
						ts += *t; tsx += *t * dix;
						tsy += *t * diy;
						tsz += *t * diz;
					  }
					tsa += *t; /* include all points */
				   }
			  }
		   }
		vs = 1.;
	  }
	else {
		for (tsa = ts = vs = 0, iz = izmin; iz <= izmax; iz++)
		   {
			diz = iz - nz2;
			vdy = vd + iz * nxy + nx * iymin;
			vvy = vv + iz * nxy + nx * iymin;
			for (iy = iymin; iy <= iymax; iy++,vdy += nx,vvy += nx)
			   {
				diy = iy - ny2;
				t = vdy + ixmin; vt = vvy + ixmin;
				for (ix = ixmin; ix <= ixmax; ix++, t++, vt++)
				   {
					dix = ix - nx2;
					if (*t > 0.)
					  {
						ts += *t; tsx += *t * dix;
						tsy += *t * diy;
						tsz += *t * diz;
					  }
					tsa += *t;	vs += *vt;
				   }
			  }
		   }
	  }
	if ((ts < FUZZ) || (vs < FUZZ))
	  {
		vob[0] = vc[0];	vob[1] = vc[1];	vob[2] = vc[2];
		return 0.;
	  }
	tsx /= ts;	tsy /= ts;	tsz /= ts;
	/* we return two versions of the `observed' centroid: one that is
	the actual observed value, and the other the weighted mean of the
	observed and predicted centroids.  The former is recorded on disc,
	whereas the latter is the one actually used for centering the spot.
	The weight value placed on the observed centroid is
	1 - exp(-G[(I/sig)^2 - 1]) for I/sig > 1 (it's zero otherwise).
	Currently G is set to (ln 2)/8 = 0.086643397569993; this allows
	the weight factor to be 0.5 at I/sig = 3. */
	vob[0] = tsx;	vob[1] = tsy;	vob[2] = tsz;
	if ((float *)0 != vv)
	  {
		tsp = (ts / vs) * ts - 1.;
		tsp = (tsp > 0. ? exp(-(MOVEWEIGHT) * tsp) : 1.);
		tso = 1. - tsp;
		tsx = vob[0] * tso + vc[0] * tsp;
		tsy = vob[1] * tso + vc[1] * tsp;
		tsz = vob[2] * tso + vc[2] * tsp;
	  }
	vc[0] = tsx;	vc[1] = tsy;	vc[2] = tsz;
	return tsa;	/* this is an intensity estimate, bucko! */
}

static void twodrecen(float *v, double *delv, float *wve, int nx, int ny)
{ /*	This routine, by a combination of numbskull reindexing and
	careful redistribution of data, moves a two-dimensional profile
	vector`v' by a shift-vector `delv'.  The vector returned in `v'
	will have had only the integer adjustments performed; the vector
	returned in `wve' will have been recentered (we hope) to the point
	where its centroid really will lie at (0,0,0).
	The vectors `v' and `wve' will have length nx * ny. */
	int		ix, iy;
	int		jj1, jj2, i1, i2;
	double		tsx, tsy;
	static float	*vy, *t, *v11, *v12;
	double		wx, wy, wxp, wyp, ux, uxp, uy, uyp;

		/* don't flip signs of deltas...I hope this is right...*/
	tsx = delv[0];	tsy = delv[1];
	/* do the coarse fitting.  We only handle +/- 1 here. */
	if (tsx < -1.)	/* too far to left by > 1: slide right by one */
	  {
		for (iy = 0, vy = v; iy < ny; iy++, vy += nx)
		    {
			t = vy + nx - 1;
			for (ix = nx - 1; ix; ix--, t--) *t = *(t - 1);
			*t = 0.;
		    }
		tsx += 1.;
	  }
	else if (tsx > 1.) /* too far to right by > 1: slide left by one */
	  {
		for (iy = 0, vy = v; iy < ny; iy++, vy += nx)
		   {
			t = vy;
			for (ix = 0; ix < nx - 1; ix++, t++) *t = *(t + 1);
			*t = 0.;
		   }
		tsx -= 1.;
	  }
	if (tsy < -1.) /* too far up by > 1: slide down  1 */
	  {
		for (ix = 0, vy = v; ix < nx; ix++, vy++)
		   {
			t = vy + nx * (ny - 1);
			for (iy = ny - 1; iy; iy--, t -= nx) *t = *(t - nx);
			*t = 0.;
		    }
		tsy += 1.;
	  }
	else if (tsy > 1.) /* too far down by > 1: slide up 1 */
	  {
		for (ix = 0, vy = v; ix < nx; ix++, vy++)
		   {
			t = vy;
			for (iy = 0, t = vy; iy < ny - 1; iy++, t += nx)
				*t = *(t + nx);
			*t = 0.;
		      }
		tsy -= 1.;
	  }
	/* then, do the fine-tuned adjustments. */
		/* can't adjust by more than 1 any more */
	tsy = flimits(tsy, -1., 1.);	wyp = (tsy < 0 ? 1. + tsy : tsy);
	tsx = flimits(tsx, -1., 1.);	wxp = (tsx < 0 ? 1. + tsx : tsx);
	wy = 1. - wyp;   wx = 1. - wxp;
	for (iy = 0, vy = wve; iy < ny; iy++)
	   {
		jj1 = (tsy < 0. ? iy - 1 : iy ); jj2 = jj1 + 1;
		uy = wy;	uyp = wyp;
		if (jj1 < 0 ) uy = jj1 = 0;
		else if (jj2 >= ny -1)
		  {
			jj2 = ny - 1;	uyp = 0.;
		  }
		v11 = v + jj1 * nx;
		v12 = v + jj2 * nx;
		for (ix = 0; ix < nx; ix++, vy++)
		   {
			i1 = (tsx < 0. ? ix - 1: ix);	i2 = i1 + 1;
			ux = wx;	uxp = wxp;
			if (i1 < 0 ) ux = i1 = 0;
			else if (i2 >= nx - 1)
			  {
				i2 = nx - 1; uxp = 0.;
			  }
			*vy = (uy * (ux * *(v11 + i1) + uxp * *(v11 + i2)) +
			 uyp * (ux * *(v12 + i1) + uxp * *(v12 + i2)));
		  }
	 }
}

void recen_vec(float *v, double *delv, float *wve, int nx, int ny, int nz)
{ /*	This routine, by a combination of numbskull reindexing and
	careful (maximum-entropy) redistribution of data, moves a profile
	vector`v' by a shift-vector `delv'.  The vector returned in `v'
	will have had only the integer adjustments performed; the vector
	returned in `wve' will have been recentered (we hope) to the point
	where its centroid really will lie at (0,0,0).  The vectors `v'
	and `wve' will have length nx * ny * nz. */
	int		ix, iy, iz, nxy;
	int		k1, k2, jj1, jj2, i1, i2;
	double		tsx, tsy, tsz, ux, uxp, uy, uyp, uz, uzp;
	double		wx, wy, wz, wxp, wyp, wzp;
	static float	*t, *v11, *v12, *v21, *v22, *vy, *vz;

	/* we zero out the output vector */
	zeromem((void *)wve, (unsigned)(nx * ny * nz) * sizeof (float));
	/* special case of two-dimensional (nz == 1) treated separately */
	if (nz < 2)
	  {
		twodrecen(v, delv, wve, nx, ny);	return;
	  }
		/* don't flip signs of deltas...I hope this is right...*/
	tsx = delv[0];	tsy = delv[1];	tsz = delv[2];	nxy = nx * ny;
	/* do the coarse fitting.  We only handle +/- 1 here. */
	if (tsx < -1.)	/* too far to left by > 1: slide right by one */
	  {
		for (iz = 0, vz = v; iz < nz; iz++, vz += nxy)
		   for (iy = 0, vy = vz; iy < ny; iy++, vy += nx)
		    {
			t = vy + nx - 1;
			for (ix = nx - 1; ix; ix--, t--) *t = *(t - 1);
			*t = 0.;
		    }
		tsx += 1.;
	  }
	else if (tsx > 1.) /* too far to right by > 1: slide left by one */
	  {
		for (iz = 0, vz = v; iz < nz; iz++, vz += nxy)
		   for (iy = 0, vy = vz; iy < ny; iy++, vy += nx)
		      {
			t = vy;
			for (ix = 0; ix < nx - 1; ix++, t++) *t = *(t + 1);
			*t = 0.;
		      }
		tsx -= 1.;
	  }
	if (tsy < -1.) /* too far up by > 1: slide down  1 */
	  {
		for (iz = 0, vz = v; iz < nz; iz++, vz += nxy)
		   for (ix = 0, vy = vz; ix < nx; ix++, vy++)
		      {
			t = vy + nx * (ny - 1);
			for (iy = ny - 1; iy; iy--, t -= nx) *t = *(t - nx);
			*t = 0.;
		      }
		tsy += 1.;
	  }
	else if (tsy > 1.) /* too far down by > 1: slide up 1 */
	  {
		for (iz = 0, vz = v; iz < nz; iz++, vz += nxy)
		   for (ix = 0, vy = vz; ix < nx; ix++, vy++)
		      {
			t = vy;
			for (iy = 0, t = vy; iy < ny - 1; iy++, t += nx)
				*t = *(t + nx);
			*t = 0.;
		      }
		tsy -= 1.;
	  }
	if (tsz < -1.) /* too early by > 1: slide forward one */
	  {
		for (iy = 0, vy = v; iy < ny; iy++, vy += nx)
		   for (ix = 0, vz = vy; ix < nx; ix++, vz++)
		      {
			t = vz + nxy * (nz - 1);
			for (iz = nz - 1; iz; iz--, t -= nxy) *t = *(t - nxy);
			*t = 0.;
		    }
		tsz += 1.;
	  }
	else if (tsz > 1.) /* too late by > 1: slide backward one */
	  {
		for (iy = 0, vy = v; iy < ny; iy++, vy += nx)
		   for (ix = 0, vz = vy; ix < nx; ix++, vz++)
		      {
			t = vz;
			for (iz = 0; iz < nz - 1; iz++, t += nxy)
				*t = *(t + nxy);
			*t = 0.;
		      }
		tsz -= 1.;
	  }
	/* then, do the fine-tuned adjustments. */
		/* can't adjust by more than 1 any more */
	tsz = flimits(tsz, -1., 1.);	wzp = (tsz < 0 ? 1. + tsz : tsz);
	tsy = flimits(tsy, -1., 1.);	wyp = (tsy < 0 ? 1. + tsy : tsy);
	tsx = flimits(tsx, -1., 1.);	wxp = (tsx < 0 ? 1. + tsx : tsx);
	wz = 1. - wzp;   wy = 1. - wyp;   wx = 1. - wxp;
	for (iz = 0, vy = wve; iz < nz; iz++)
	   {
		k1 = (tsz < 0. ? iz - 1 : iz ); k2 = k1 + 1;
		uz = wz;	uzp = wzp;
		if (k1 < 0 ) uz = k1 = 0;
		else if (k2 > nz - 1)
		  {
			k2 = nz - 1; uzp = 0.;
		  }
		for (iy = 0; iy < ny; iy++)
		 {
			jj1 = (tsy < 0. ? iy - 1 : iy ); jj2 = jj1 + 1;
		   	uy = wy;	uyp = wyp;
			if (jj1 < 0 ) uy = jj1 = 0;
			else if (jj2 >= ny -1)
			  {
				jj2 = ny - 1;	uyp = 0.;
			  }
			v11 = v + k1 * nxy + jj1 * nx;
			v12 = v + k1 * nxy + jj2 * nx;
			v21 = v + k2 * nxy + jj1 * nx;
			v22 = v + k2 * nxy + jj2 * nx;
			for (ix = 0; ix < nx; ix++, vy++)
			   {
				i1 = (tsx < 0. ? ix - 1: ix);	i2 = i1 + 1;
				ux = wx;	uxp = wxp;
				if (i1 < 0 ) ux = i1 = 0;
				else if (i2 >= nx - 1)
				  {
					i2 = nx - 1; uxp = 0.;
				  }
				*vy = uz * (uy * (ux * *(v11 + i1) +
				 uxp * *(v11 + i2)) +
				 uyp * (ux * *(v12 + i1) + uxp * *(v12 + i2)))+
				 uzp * (uy * (ux * *(v21 + i1) +
				 uxp * *(v21 + i2)) + uyp * (ux * *(v22 + i1) +
				 uxp * *(v22 + i2)));
			   }
		 }
	   }
}

float to360(float rv)
{ /* This forces an angle to be between -180 and +180 */
	int	z;

	if (rv < 0.) z = (int)((rv - 180.) / 360.);
	else z = (int)((rv + 180.) / 360.);
	return rv - 360. * (float)z;
}

float mymedian(float *flb, int npt)
{ /*	This finds the median of the first `npt' values in
	 the array of floats pointed to by flb. `flb' is not assumed
	 to be sorted in any interesting way. */
	int		i, nltm, npto2, npto2p, npx2, nlook, nltmp, neq;
	double		rxmn, rxmx, median, mp1, fraclt;
	float		*fli, *flo, *mvi, *mvo;
	static float	*mework;
	static int	oldmelen;
	static char	memsg[] = "mework in mymedian";

	if (NULL == flb) /* special case: eliminate memory for this op */
	  {
		memzap((AVOID **)&mework, memsg); oldmelen = 0;	return 0.;
	  }
	/* first, find the extrema and estimate the median as the mean
	 of the extrema */
	rxmn = rxmx = *flb;	npto2 = npt / 2;	npto2p = npto2 + 1;
	for (i = npt - 1, fli = flb + 1; i--; fli++)
	   {
		if (*fli < rxmn) rxmn = *fli;	if (*fli > rxmx) rxmx = *fli;
	   }
	/* reserve memory for the working vector of this operation */
	npx2 = npt * 2;
	if (npx2 < MIN_MEDSIZE) npx2 = MIN_MEDSIZE;
	if ((NULL == mework) || (npx2 > oldmelen))
	  {
		memzap((AVOID **)&mework, memsg);	oldmelen = npx2;
		if (NULL == (void *)(mework = (float *)callop(npx2,
		 sizeof (float), memsg))) return 0.5 * (rxmx + rxmn);
	  }
	/* start by examining the original vector */
	mvi = flb;	mvo = mework;	nlook = npt;	nltmp = 0;
	while (1)
	     {
		median = 0.5 * (rxmx + rxmn);	mp1 = median + 1.;
		if (rxmx - rxmn < 1.) break;
		for (fli = mvi, i = 0, nltm = nltmp, neq = 0;
		  i < nlook; i++, fli++)
		  {
			if (*fli < median) nltm++;
			else if (*fli < mp1) neq++;
		  }
		/* if half the values are < median, we're done */
		if ((npto2 <= nltm) && (nltm <= npto2p)) break;
		fraclt = ((float)(nltm + neq / 2)) / (float)npt;
		/* if less than half are < median, we keep only the 
		 ones that are greater than the median.  But if
		 adding in the number of elements exactly equal to this
		 guess of the  median puts us over half, we're done */
		if (nltm < npto2)
		  {
			if (((nltm + neq) >= npto2) || (fraclt > 0.5)) break;
			rxmn = median;	nltmp = nltm;
			for (fli = mvi, flo = mvo, i = 0; i < nlook;
			 i++, fli++) if (*fli > median) *flo++ = *fli;
		  }
		/* if more than half are < median, we keep only the 
		 ones that are less than the median.  Again, if adding
		 in the number of elements exactly equal to this guess
		 of the median pulls us under half, we're done. */
		else {
			if (((nltm - neq) < npto2) || (fraclt < 0.5)) break;
			rxmx = median;
			for (fli = mvi, flo = mvo, i = 0; i < nlook;
			 i++, fli++) if (*fli < median) *flo++ = *fli;
		     }
		/* save the number of points left to examine */
		nlook = flo - mvo;
		if (nlook < 2) break;
		/* rotate to the other pair of buffers */
		if (mvo == mework)
		  {
			mvi = mework;	mvo = mework + npt;
		  }
		else {
			mvo = mework;	mvi = mework + npt;
		     }
	     }
	return median;
}

int imedian(int *ilb, int npt)
{ /*	This finds the median of the first `npt' values in the array of
	ints pointed to by ilb. ilb is not assumed to be sorted in
	any interesting way. This is essentially the same algorithm as in
	the spot-finding routine mymedian(), but with ints, not floats */
	int		i, nltm, npto2, npto2p, npx2, nlook, nltmp, neq;
	int		rxmn, rxmx, *ili, *ilo, *mvi, *mvo;
	double		median, mp1, fraclt;
	static int	*medv;
	static int	oldmelen;
	static char	memsg[] = "medv in imedian";

	if (NULL == ilb)
	  {
		memzap((AVOID **)&medv, memsg);	oldmelen = 0; return 0;
	  }
	/* first, find the extrema and estimate the median as the mean
	 of the extrema */
	rxmn = rxmx = *ilb;	npto2 = npt / 2;	npto2p = npto2 + 1;
	for (i = npt - 1, ili = ilb + 1; i--; ili++)
	   {
		if (*ili < rxmn) rxmn = *ili;	if (*ili > rxmx) rxmx = *ili;
	   }
	/* reserve memory (if necessary) for the working vectors of this
	 operation */
	npx2 = npt * 2;	if (npx2 < MIN_MEDSIZE) npx2 = MIN_MEDSIZE;
	if ((NULL == medv) || (npx2 > oldmelen))
	  {
		memzap((AVOID **)&medv, memsg);	oldmelen = npx2;
		if (NULL == (void *)(medv = (int *)callop(
		 npx2, sizeof (int), memsg))) return 0.5 * (rxmx + rxmn);
	  }
	/* start by examining the original vector */
	mvi = ilb;	mvo = medv;	nlook = npt;	nltmp = 0;
	while (1)
	     {
		median = 0.5 * (rxmx + rxmn);	mp1 = median + 1.;
		if (rxmx - rxmn <= 2) break;
		for (ili = mvi, i = 0, nltm = nltmp, neq = 0;
		  i < nlook; i++, ili++)
		  {
			if (*ili < median) nltm++;
			else if (*ili < mp1) neq++;
		  }
		/* if half the values are < median, we're done */
		if ((npto2 <= nltm) && (nltm <= npto2p)) break;
		fraclt = ((double)(nltm + neq / 2)) / (double)npt;
		/* if less than half are < median, we keep only the 
		 ones that are greater than the median.  But if
		 adding in the number of elements exactly equal to this
		 guess of the  median puts us over half, we're done */
		if (nltm < npto2)
		  {
			if (((nltm + neq) >= npto2) || (fraclt > 0.5)) break;
			rxmn = median;	nltmp = nltm;
			for (ili = mvi, ilo = mvo, i = 0; i < nlook;
			 i++, ili++) if (*ili > median) *ilo++ = *ili;
		  }
		/* if more than half are < median, we keep only the 
		 ones that are less than the median.  Again, if adding
		 in the number of elements exactly equal to this guess
		 of the median pulls us under half, we're done. */
		else {
			if (((nltm - neq) < npto2) || (fraclt < 0.5)) break;
			rxmx = median;
			for (ili = mvi, ilo = mvo, i = 0; i < nlook;
			 i++, ili++) if (*ili < median) *ilo++ = *ili;
		     }
		/* save the number of points left to examine */
		nlook = ilo - mvo;
		if (nlook < 2) break;
		/* rotate to the other pair of buffers */
		if (mvo == medv)
		  {
			mvi = medv;	mvo = medv + npt;
		  }
		else {
			mvo = medv;	mvi = medv + npt;
		     }
	     }
	return (int)median;
}

void dmm31(double *y, double a[3][3], double *x)
{ /* this left-multiplies the 3x3 matrix a by the 3-elt. col vector
		x and leaves the result in the 3-element column vector y.
	This is the double-precision version of the routine */

	register int	i,k;
	register double	*ay,*ax;

	for (ay = y, i = 0; i < 3; i++, ay++)
		for (*ay = k = 0, ax = x;  k < 3; k++, ax++)
			*ay += a[i][k] * *ax;
	return;
}

int dinv3(double xi[3][3], double xinv[3][3])
{ /* This routine returns the inverse of the double-precision 3x3
	matrix `xi' in the 3x3 matrix `xinv'. Return values:
	 0: inversion was successful.
	-1: largest absolute-value element is smaller than FUZZ.
	 1: determinant of rescaled matrix is too close to zero.
	The inversion is performed carefully, i.e. with some attention
	paid to working with scaled quantities. */
	register int	i, j, k;
	double		determ, elt, maxelt, x[3][3];

	determ = maxelt = 0.;	/* clear rescaler & determinant */
	for (i = 0; i < 3; i++)	/* get rescaling constant */
	   for (j = 0; j < 3; j++)
	      {
		elt = fabs(xi[i][j]);	if (elt > maxelt) maxelt = elt;
	      }
	if (maxelt < FUZZ) return ALLTOOSMALL;
	maxelt /= RESCALE;	/* rescale by this constant */
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		x[i][j] = xi[i][j] / maxelt;
	for (i = 0; i < 3; i++)
	   {
		j = (i + 1) % 3;	k = 3 - i - j;
		xinv[i][i] = x[j][j] * x[k][k] - x[j][k] * x[k][j];
		xinv[i][j] = x[k][j] * x[i][k] - x[k][k] * x[i][j];
		xinv[i][k] = x[i][j] * x[j][k] - x[i][k] * x[j][j];
	   }
	for (i = 0; i < 3; i++) determ += x[0][i] * xinv[i][0];
	if (fabs(determ) < FUZZ) return DETERZERO;
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
	   {
		xinv[i][j] /= determ; xinv[i][j] /= maxelt;
	   }
	return	0;
}

int inv3(float xi[3][3], float xinv[3][3])
{ /* This routine returns the inverse of the floating-point
	3x3 matrix `xi' in the 3x3 matrix `xinv'. Return values:
	 0: inversion was successful.
	-1: largest absolute-value element is smaller than FUZZ.
	 1: determinant of rescaled matrix is too close to zero.
	Rather than doing this explicitly, we do it by converting
	to double-precision, calling dinv3 to invert the result,
	and converting back to single precision. */
	int	i, j, resu;
	double	dxi[3][3], dxinv[3][3];

	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++) dxi[i][j] = xi[i][j];
	resu = dinv3(dxi, dxinv);
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		xinv[i][j] = dxinv[i][j];
	return resu;
}

void cleanorthog(double q[3][3])
{
	/* We make a nominally orthoganal matrix more precise by averaging it
	with its inverse transpose a few times.
	This is a trick I learned from Bret Maverick. */
	register int	i, j, k;
	double		qsav[3][3];

	for (k = AVERAGE_PASSES; k-- && eunorm(q) > OKNORM; )
	   {
		if (dinv3(q, qsav)) break;	/* weird... */
		for (i = 3; i--; ) /* average with inverse transpose */
		   for (j = 3; j--; ) q[i][j] = (q[i][j] + qsav[j][i]) / 2.;
	   }
}

/* end of xgenmath.c */
