/* xgenshift.c -- Routines for manipulating scaling groups in XGEN

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	23 Apr 2007: get_shifts: parse ordinary data lines correctly;
		terpalt written: handles alternate-shift val in terpscale
	19 Apr 2007: float->double conversions; reorganize use of "{"
	24 Jun 2006: include xgpostint.h; put_shifts->../postprocess/scrio.c
	20 Nov 2005: inline argument declarations and strict prototyping
	26 Oct 2005: get rid of useless `nextfree' variable; housekeeping
	30 Jun 2005: get_omcen(): read correct sizes for DATARUN headers
	 6 Sep 2004: cosmetic changes
	24 Jan 2004: fpsi made a static local in get_shifts()
	21 Jan 2004: license message inserted
	 5 Jan 2004: global vbl shlp replaced by local vbls & args to routines;
		init_shift() and get_shifts() return SHIFT structure pointers
	19 Sep 2003: move find_shift() earlier in the file so
		we don't have to declare it externally
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	 5 Jan 2003: reorganized source-file header
	15 Apr 2002: in init_shift(), add_shift(), get_shifts(), put_shifts(),
		modify calls related to shlp[] since it's handled differently;
		init_shift(), add_shift(), get_shifts() changed substantially
	25 Mar 2002: removed MSI attachments
	31 May 1995: "register" -> "register int" throughout
		ulsh is an xgulong, not a u_long
	12 Apr 1995: s_char replaced by sichar
	 2 Mar 1995: trivial formatting
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	27 Oct 1994: add_shift: ignore spherical harmonics lines in $SCALEIN
		if no memory reserved for them by calling routine
	 5 Oct 1994: use xgread for CRYSTAL, DATARUN structures
	 5 May 1994: reorganization into this source file
	16 Mar 1994: polymask added; bugfixes in polymath, xtal, calinv, etc.
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version
	xx xxx 1986: in get_shifts(), if the first nonblank character on a
		line of SCALEIN is a letter, the line is taken to contain
		information to be used in vector-dependent scaling
	26 Dec 1985: incorporated into "get_shifts" the ability to read
		the variance of the scale factors.
	18 Dec 1985: fixed irritating bug in find_shift
	11 Sep 1984: initial code for get_shifts(), add_shift(), find_shift()

   Purpose: processing scaling-group information in a MULTIREF file.
   Cross references: mrgeom.c

*************************************************************************/
#include	"xgpostint.h"

SHIFT *init_shift(int ntosave)
{ /*	This initializes the list of scaling-group structures.
	In this version there is one level of indirection rather than two,
	and the routine now returns lsh. */
	SHIFT	*lsh;

	if (0 >= ntosave) return NULL;
	if (NULL == (lsh = (SHIFT *)callop(ntosave, sizeof (SHIFT),
	 "lsh in init_shift"))) return NULL;
	return lsh;
}

void add_shift(SHIFT *ish, int indx, int nscal)
{ /*	This adds a new shift to the shift-list ish.  It does this by
	(1) pushing up the stack of shift pointers,
	(2) assigning the new shift to the next free pointer in the SHIFT list.
	Note that this routine may be called when no scales have been built
	yet, i.e. indx == 0, nscal==1, and the first "if" will be false. */
	register int	i;

	if (indx < nscal-1)	/* push up the stack of shift pointers */
		for (i = nscal-1; i > indx; i--)
		   {
			cpymem((AVOID *)(ish + i), (AVOID *)(ish + i - 1),
				sizeof (SHIFT));
		   }
	return;
}

int find_shift(SHIFT *ish, xgulong m, int nscal)
{ /*	This goes through the sorted array of shifts and returns
	the INDEX within that array of the "shift" m.
	It does so by a nonreentrant tree search.
	If the shift is missing, the routine returns the negative of
	the index just ABOVE the shift we found if the shift isn't there--
	i.e., if m is between shifts indexed (j) and (j+1), we return (j+1).
	In the special case where m > k(nscal-1) we return (-nscal).
	In the special case where m < k[0] we return (-nscal-1).
	The "shift" m contains the shift in the lowest-order 16 bits,
	the chamber number in the next 8, and the run number in the top 8. */
	register int	upper, lower, j;
	xgulong		k, k0;
	SHIFT		*tsh;

	upper = nscal - 1;		/* upper limit to search initially */
	lower = 0;			/* lower limit to search initially */
	j = (upper + lower)/2;		/* try the middle of the list first */

	tsh = ish + j;
	while (m != (k = (LSH(tsh->runno, 24) | LSH(tsh->chamno, 16) |
		(tsh->shval))))
	     {
		if (upper <= lower) /*upper==lower && m!=k iff. shift missing*/
		  { /* return < 0 value */
			if (m > k) return (-nscal);
			else {
				k0 = LSH(ish->runno, 24) |
				 LSH(ish->chamno, 16) | ish->shval;
				return (m < k0 ? (-nscal-1) : -j );
			     }
		  }
		if (m < k) upper = j;	/* we're too low, decrease upper lim */
		else	lower = j + 1;	/* too high: increase lower lim */
		j = (upper + lower)/2;	/* try middle of restricted list */
		tsh = ish + j;
	     }
	return j;			/* here's the right shift--return it */
}

SHIFT *get_shifts(SHIFT *ish, VECPAR *vsp, int *nship)
{ /*	This sets up the SHIFT structures required for processing
	of the shifts specified in the printable file pointed to by 'fps'.
	Note that the space for individual SHIFT structures will be
	reserved as needed, but the space for the array of pointers
	must be reserved in the calling routine.
	The shift-structure list need not be sorted to begin with,
	and if the same shift appears more than once in the file the
	later value will be used.
	This routine returns the number of shifts found in the file
	 as *nship and returns the new pointer to the shift list.
	
	For XENGEN V.2.0: if the first nonblank character on the line
	is a letter, the line is assumed to contain information to be
	used in vector-dependent scaling.  The format of
	the line in that case is:
		c minshift a0 a1 a2 ... a(n-1)
	where `c' is the code for the type of vector to be scaled on,
	`minshift' is the min scaling group number
	associated with the values we're dealing with, and a0, ... are
	the actual parameters associated with the group.  For example:
 d   200  0.895 0.047-0.016 0.204 0.918-0.085-0.009 1.244-0.293-0.830
	indicates that the diffraction-vector-dependent parameters for
	scaling groups 200 to 399 are the values given.
 The resulting array of diffraction-vector structures is returned in
 the structures pointed to by `vsp'.
	18 Dec 2003: we do all the file I/O internally here;
		if we're unable to read any scale inputs we return 0. */
	char		line[142], *cp;
	int		chamin, haveseen, minshift, nscal, nvp;
	int		runin, shiftin, shindex;
	xgulong		ulsh;
	double		shain, shbin, shgin, shvarin;
	float		vp[16];
	VECPAR		*vpp, *vpe;
	SHIFT		*osh, *sqp;
	static FILE	*fpsi;

	*nship = 0;
	if (NULL == (fpsi = xgopen("scalein", "r")))
	  {
		memzap((AVOID **)&ish, "ish in get_shifts"); return NULL;
	  }
	nscal = shindex = 0;
	vpe = vsp;	/* initialize the end-of-list pointer */
	/* we reserve memory for the scaling-group stuff if we don't
	 already have some */
	if (NULL == ish)
	  {
		nscal = 0; /* clear scale and spherical-harm counts */
		while (NULL != fgets(line, 140, fpsi)) /* count good lines */
		     {
			for (cp = &(line[0]); '\0' != *cp; cp++)
				if (*cp > ' ') break;
			if (!(('A' <= *cp && *cp <= 'Z') ||
				('a' <= *cp && *cp<= 'z'))) nscal++;
		     }
		if (nscal < 1)
		  {
			(void)fclose(fpsi);	fpsi = NULL;
			return NULL; /* unlikely this would occur! */
		  }
		/* rewind the scaling-shift list */
		if (-1 == fseek(fpsi, 0L, 0))
		  {
			*nship = 0;	osh = NULL;
			(void)fclose(fpsi);	fpsi = NULL;	return osh;
		  }
		if (NULL == (osh = init_shift(nscal + 20)))
		  {
			osh = NULL;	*nship = 0;
			(void)fclose(fpsi);	fpsi = NULL;	return osh;
		  }
		nscal = 0;
	  }
	else	osh = ish;
	while (NULL != fgets(line, 140, fpsi))
	     {
		for (cp = &(line[0]); '\0' != *cp; cp++) if (*cp > ' ') break;
		/* interpret spherical-harmonic lines if that's what
		 we're looking at and if vsp is non-NULL */
		if ((('A' <= *cp && *cp <= 'Z') || ('a' <= *cp && *cp<= 'z'))
		  && (NULL != vsp))
		  {
			ulsh = 0;
			for (nvp = 0; nvp < MAXVECPARS; nvp++)
				vp[nvp] = 0;
			if (2 > sscanf(cp + 1,
		 "%d %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f",
			 &minshift, &vp[0], &vp[1], &vp[2], &vp[3], &vp[4],
			 &vp[5], &vp[6], &vp[7], &vp[8], &vp[9], &vp[10],
			 &vp[11], &vp[12], &vp[13], &vp[14], &vp[15])) break;
			for (vpp = vsp, haveseen = 0; vpp < vpe; vpp++)
			   if (minshift == vpp->v_ms)
			     {
				for (nvp = 0; nvp < MAXVECPARS; nvp++)
					vpp->v_coef[nvp] = vp[nvp];
				vpp->v_ptype = *cp;
				haveseen = 1;	break;
			     }
			if (!haveseen)
			  {
				vpp = vpe;	vpe++;
				vpp->v_ms = minshift;
				for (nvp = 0; nvp < MAXVECPARS; nvp++)
					vpp->v_coef[nvp] = vp[nvp];
				vpp->v_ptype = *cp;
			  }
		  }
		else if ((*cp < 'A') || (('Z' < *cp) && (*cp < 'a')) ||
		 (*cp > 'z'))
		  {
			if (4 > sscanf(line,"%d %d %d %lf %lf %lf %lf",
			 &shiftin, &chamin, &runin,
			 &shgin, &shain, &shbin, &shvarin)) break;
			ulsh= (runin << 24) | (chamin << 16) | shiftin;
			if (!nscal ||
			  (shindex = find_shift(osh, ulsh, nscal)) < 0)
			  { /* this shift needs to be added to the list */
				shindex = (shindex == -(++nscal)) ?
					0 : -shindex;
				add_shift(osh, shindex, nscal);
				sqp = osh + shindex;
				sqp->shval = shiftin;
				sqp->runno = runin;
				sqp->chamno = chamin;
			  }
			else sqp = osh + shindex;
			/*scale values are installed whether new shift or not*/
			sqp->shg = shgin; /* set scale factor*/
			sqp->sha = shain; /* coeff of sth/l */
			sqp->shb = shbin; /* -Debye factor */
			sqp->shvar = shvarin; /* variance */
		     }
	     }
	(void)fclose(fpsi);	fpsi = NULL;
	*nship = nscal;	return osh;	/* return # of input scales */
}

void neighbors(SHIFT *ish, unsigned nsca, int isquad)
{ /*	This goes through a list of scaling groups and
	sets a flag value within each scaling group according to whether
	or not the scaling group has previous or subsequent neighbors that can
	be used for interpolation. nbrs[grp] will take on the following values:
	0	neither the scaling group -2 or -4 units away nor the scaling
		  group +2 or -2 units away is taken from the same run.
	1	the scaling group +2 or +4 units away IS taken from the same
		  run, but the the group -2 or -4 units away is not.
	2	the scaling group -2 or -4 units away IS taken from the same
		  run, but the group +2 or +4 units away is not.
	3	both the scaling group -2 or -4 units away and the scaling
		 group +2 or +4 units away are taken from the current run.
	+/-4 is used if `isquad' is on; +/-2 is used otherwise. */
	int		ishn2, ishp2, idel;
	unsigned	iish;
	short		nbr;
	u_char		currun, currcham;
	SHIFT		*shp, *shal;

	if (isquad)
	  {
		ishn2 = -4;	ishp2 = 4;	idel = 4;
	  }
	else {
		ishn2 = -2; ishp2 = 2;	idel = 2;
	     }
	for (iish = 0, shp = ish; iish < nsca;
	 iish++, ishn2++, ishp2++, shp++)
	   {
		nbr = 0; currun = shp->runno;	currcham = shp->chamno;
		if ((0 <= ishn2) && (ishn2 < (int)nsca))
		  {
			shal = shp - idel;
			if ((NULL != shal) && (shal->runno == currun) &&
			 (shal->chamno == currcham)) nbr = 2;
		  }
		if ((0 <= ishp2) && (ishp2 < (int)nsca))
		  {
			shal = shp + idel;
			if ((NULL != shal) && (shal->runno == currun) &&
			 (shal->chamno == currcham)) nbr |= 1;
		  }
		shp->shnbr = (isquad) ? nbr | 04 : nbr;
	   }
}

int terpalt(short tom, SHIFT *shv)
{ /* This returns the offset within the array of scaling groups associated
	with the SHIFT just below or just above the current one.
	A return value of 0 indicates that there's no need for an alternate
	scaling group for interpolations. */
	int	idel, nbflag, omc;

	if (0 == (nbflag = shv->shnbr & 03)) return 0;
	idel = (shv->shnbr & 04) ? 4 : 2;
	if (tom < (omc = shv->shomc)) return (nbflag & 02) ? -idel : idel;
	else return (nbflag & 01) ? idel : -idel;
}

double terpscale(short tom, SHIFT *shv, double st)
{ /*	This looks at the observation 'aobp' and returns a scaling
	parameter appropriate to it. */
	short	omc, altom;
	int	alt;
	double	zg, za, zb, wg, wa, wb, stsq, epsi, epsip;
	SHIFT	*shalt;

	stsq = st * st;
	zg = shv->shg;	za = shv->sha;	zb = shv->shb;
	if (0 != (alt = terpalt(tom, shv)))
	  {
		shalt = shv + alt; omc = shv->shomc; altom = shalt->shomc;
		epsi = ((double)(omc - tom)) / (double)(omc - altom);
		epsip = 1. - epsi;
		wg = shalt->shg; wa = shalt->sha; wb = shalt->shb;
		zg = zg * epsip + wg * epsi;
		za = za * epsip + wa * epsi;
		zb = zb * epsip + wb * epsi;
	  }
	return (zg + st * za + stsq * zb);
}

int get_omcen(FILE *fpm, SHIFT *ish, int nscal, int isquad)
{ /*	This takes an already-open multiref and quickly scans through the
	observations in it to determine <omega> in each scaling group in it.
	The scaling array `ish' is already set up by the time we call this
	routine, and we assume that the multiref is already positioned at the
	end of the data run headers.  The routine leaves the multiref in that
	same position when it's finished. */
	CRYSTAL		lxhd;
	DATARUN		lrhd;
	MERGEDREF	mr;
	ABBROB		*aobp;
	int		nob, shind;
	xgulong		ulsh;
	float		*mnob, *meanom;
	static size_t	sizdatarun = 256;

	mnob = (float *)callop((unsigned)(2 * nscal), sizeof (float),
	 "mnob in get_omcen");
	meanom = mnob + nscal;
	/* determine which scaling groups have neighbors */
	neighbors(ish, (unsigned)nscal, isquad);
	while (0 < xgread((void *)&mr, XG_MERGEDREF, fpm))
	     {
		for (aobp = &(mr.mr_ob[0]), nob = mr.mr_nob; nob--; aobp++)
		   {
			ulsh = LSH(aobp->abo_runno, 24) |
				LSH(aobp->abo_chamno, 16) | aobp->abo_shift;
			if (0 <= (shind = find_shift(ish, ulsh, nscal)))
			  {
				mnob[shind] += 1.;
				meanom[shind] += aobp->abo_omob;
			  }
		   }
	     }
	/* having quickly read thru the multiref, we reposition pointer */
	if (verror((1 + fseek(fpm, 0L, 0)), "rewind multiref on 0th pass"))
		return ESEEKERR;
	if (sizeof (CRYSTAL) > (size_t)xgread((void *)&lxhd, XG_CRYSTAL, fpm))
		return EREADERR;
	for (nob = 0; nob < lxhd.xt_nruns; nob++)
	 if (sizdatarun > (size_t)xgread((void *)&lrhd, XG_DATARUN, fpm))
			return EREADERR;
	/* now compute the mean omega in each scaling group */
	for (shind = 0; shind < nscal; shind++)
	   {
		if (mnob[shind]) meanom[shind] /= mnob[shind];
		(ish + shind)->shomc = meanom[shind];
	   }
	memzap((AVOID **)&mnob, "mnob in get_omcen");	/* we're done */
	return EOK;
}

/* end of xgenshift.c */
