/* xgparams.c -- routines associated with handling REFPARS runtime parameters

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	21 Mar 2007: partoxou used to record output from refinement in *.xou
	28 Nov 2006: old mat2vecs replaced by rp2vecs; artocel, makeq21 updated
	 mat2vecs is now a primitive routine; ucell calls updated
	24 Jun 2006: fmtoupr: use saxadjust to adjust motor-start value
	17 Jun 2006: par2head moved to redutils and renamed partohead
	11 Jun 2006: dspace moved to ../xgrefine/refutil.c so it can be static
	31 May 2006: comments reformatted
	20 Nov 2005: inline argument declarations and strict prototyping
	 1 Nov 2005: cvmat routine eliminates remaining static globals
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	10 Jan 2005: default gamma2 value changed from 0.0026 to 0.0010
	 2 Sep 2004: created from xgenmisc.c
 Relevant changes to xgenmisc.c prior to that date:
	30 Jan 2004: angles input to ztos() are now in degrees, not radians
	21 Jan 2004: license message inserted
	 7 Jan 2004: REFPARS structures and args to routines replace globals;
		rp2vecs() replaces matglob();
		artocel() and cellconvert() now take REFPARS arg;
		refparin() and refparout() handle UPARAMS input and output;
		makeq21() handles tilt & swing matrix;
		qcell() and cqcell() replaced by rotcell()
	13 Dec 2003: eliminate unused variable `proddot' in reducedcell()
	28 Nov 2003: add Euler angle calculations to artocel();
	26 Nov 2003: don't need deltaom, deltachi any more
	 3 Nov 2003: moved lattice-character routines to xgbravais.c;
		numerous routines moved from ../xgrefine/refxtal.c to here
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	18 Aug 2003: created from xgenio.c
 Relevant earlier changes to xgenio.c and other sources:
	 1 Apr 2002: rewrite reducedcell(); handles angles close to 90 better
	25 Mar 2002: removed MSI attachments
	31 May 1995: "register" -> "register int" throughout
		re-cast several memory-handling routines with size_t's
	 5 May 1994: initial version of xgenio.c, cobbled from other files
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version of older files

   Purpose: dealing with management of refinable parameters
   Cross references: xgenmisc.c, etc.

*************************************************************************/
#define		NEED_SSDEF	1
#define		NEED_DESCRIP	1
#include	"xgen.h"
#define		RR_FLIPSTEP	0200	/* flip sign of stepsize in image */

extern void saxadjust(MAIL *mp, REFPARS *rprp, int frmno);

static void cvmat(int indx, float retmat[3][3])
{ /*	returns an appropriate transformation matrix in retmat */
	int		i, j, *ip;
	static int	allmats[] =	{
		 1, 0, 0, 0,-1, 0, 0, 0,-1, /* lma :	0 */
		-1, 0, 0, 0, 1, 0, 0, 0,-1, /* lmb :	1 */
		-1, 0, 0, 0,-1, 0, 0, 0, 1, /* lmc :	2 */
		-1, 0, 0, 0, 0,-1, 0,-1, 0, /* lm021:	3 */
		 0,-1, 0,-1, 0, 0, 0, 0,-1, /* lm102:	4 */
		 0, 1, 0, 0, 0, 1, 1, 0, 0, /* lm120:	5 */
		 0, 0, 1, 1, 0, 0, 0, 1, 0, /* lm201:	6 */
		 0, 0,-1, 0,-1, 0,-1, 0, 0, /* lm210:	7 */
		 1, 0, 0, 0, 1, 0, 0, 0, 1 };/*ident:	8 */

	ip = allmats + 9 * indx;
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++, ip++)
		retmat[i][j] = *ip;
}

int mat2vecs(float rmat[3][3], float *ra, float *ran)
{
 /* This returns the unit cell lengths and angles associated with the 3x3
  matrix rmat. We make no assumptions about the units on rmat here;
  thus it can be applied to any 3x3 unit cell, whether it is unitless or
  has Angstroms in it somewhere. The lengths, of course, will be in the
  same units as the matrix itself.
  The routine returns 0 if successful, various errors if not. */
	int	i, j, k, m;
	double	tc, td, te, dra[3], dran[3];

	for (k = i = 0; i < 3; i++) /* get the lengths */
	   {
		for (td = j = 0; j < 3; j++)
		   {
			tc = rmat[i][j]; td += tc * tc;
		   }
		if (td < DFUZZ)
		  {
			dra[i] = 0.;	k++;
		  }
		else ra[i] = dra[i] = sqrt(td);
	   }
	if (k) return -k - 3;
	for (i = 0; i < 3; i++) /* find the normalized dot products */
	   {
		j = i + 1;	if (j > 2) j -= 3;
		k = 3 - i - j;	tc = ra[j];	td = dra[k];
		for (te = m = 0; m < 3; m++)
			te += (rmat[j][m] / tc) * (rmat[k][m] / td);
		ran[i] = dran[i] = te;
	   }
	/* This is the volume multiplier for the unit cell.
	 If it is non-positive, the cell is invalid. */
	tc = 1. + 2. * dran[0] * dran[1] * dran[2] -
		dran[0] * dran[0] - dran[1] * dran[1] - dran[2] * dran[2];
	for (k = i = 0; i < 3; i++)	/* determine angles in degrees */
	   {
		if ((-1. <= (td = dran[i])) && (td<= 1.)) td = acos(td) / RADC;
		else {
			td = 0.;	k++;
		     }
		ran[i] = td;
	   }
	if (k) return -k;	/* if an angle is bad, remember such */
	return (tc <= DFUZZ);	/* return 0 if tc is positive */
}

int rp2vecs(REFPARS *rpp)
{ /*	This extracts the eternal verities from the real-space
	unit cell matrix: first, the unit cell lengths and angles,
	then the reciprocal-space lengths and angles, and then
	reconstituted real- and reciprocal-space matrices. Return values:
	val meaning		val meaning		val meaning
	 0  success		-1  ucell failure	-2  vecmat failure
	-3  lor uninvertable	-4  q uninvertible	-5  rsave uninvertible
	-6  rpp->xlamba < FUZZ	-7  bad xareal matrix
	This version uses no global variables.
	rpp->xast[] and rpp->xanst[] come back as the reciprocal-space
	unit cell lengths and angles in inverse Angstroms and radians. */
	register int	i, j;
	float		flor[3][3], frsave[3][3], lrp[3], lzsa[3][3];
	double		lor[3][3], lorr[3][3], dlrp[3], q[3][3], rsave[3][3];

		/* put in nominal values for reciprocal-space vectors */
	for (i = 3; i--; )
	   {
		for (j = 3; j--; ) lor[i][j] = flor[i][j] = rpp->xareal[i][j];
		rpp->xast[i] = 1.;	rpp->xanst[i] = PIO2;
	   }
	if (rpp->xlambda < FUZZ) return -6;
	/* determine unit cell lengths and angles */
	if (mat2vecs(flor, rpp->xabc, rpp->xang)) return -7;
	for (i = 3; i--; )	/* lengths should be in Angstroms */
		rpp->xabc[i] *= rpp->xlambda;
	/* rotate down to true goniostat (0,0,0) from the observed position */
	ztos(rpp->xstart, lzsa);	/* get appropriate rotation matrix */
	mm3t(frsave, flor, lzsa);	/* to rotate local realspace matrix */
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		rsave[i][j] = frsave[i][j];
	/* Next, we parameterize the rotation from the "standard" position
	 obtained by vecmat to the current position by obtaining the orthogonal
	 matrix required and then pulling the angles out of that matrix.
	 We do this by noting that if A is the recip-space unit cell matrix
	 at the "vecmat" position and B is the corresponding position at the
	 reference position of data collection, then
		B = QA for some orthogonal transformation Q.
	 Thus Q = BA~, or Q' = (BA~)~ = AB~ = AR * lambda.  Neat, huh? */
		/* clean up cell */
	if (ucell(rpp->xabc, rpp->xang, rpp->xast, rpp->xanst, rpp->xisys))
		return -1;
		/* get real-space unit cell */
	if (vecmat(rpp->xabc, rpp->xang, flor)) return -2;
		/* scale the real-space version by the X-ray wavelength */
	for (i = 3; i--; ) for (j = 3; j--; )
	   {
		lor[i][j] = flor[i][j] = flor[i][j] / rpp->xlambda;
	   }
	/* invert to get reciprocal-space matrix. Note that in this
	 version the reciprocal-space matrix elements are unitless,
	 since we're computing them _after_ we make lor[i][j] unitless */
	if (dinv3(lor, lorr)) return -3;
	dmm3(q, lorr, rsave); 	/* get nearly orthogonal matrix Q */
	cleanorthog(q);	/* make the orthogonal matrix more precise */
	qtoangs(q, dlrp); /* get the rotation angles from the q matrix */
	for (i = 0; i < 3; i++) lrp[i] = dlrp[i];
	rpp->xdelom = -dlrp[0] / RADC;	rpp->xdelchi = -dlrp[1] / RADC;
	rpp->xdelphi = -dlrp[2] / RADC;
	rot9r(flor, lrp); /* reconstitute the matrices at (0,0,0) */
	/* rotate from (0,0,0) to (om,chi,phi)start */
	mm3(frsave, flor, lzsa);
	if (inv3(frsave, flor)) return -5;
	for (i = 3; i--; ) for (j = 3; j--; )
	   {
		rpp->xrec[i][j] = flor[i][j];
		rpp->xareal[i][j] = frsave[i][j];
	   }
	return 0;
}

void artocel(float *stang, float are[3][3], float cell[])
{ /*	this calculates cell lengths and angles from the cell matrix.
	In this case are[3][3] is in Angstroms, not unitless.
	cel[6] - [8] are the Euler angles associated with the solution */
	int	i, j;
	float	floar[3][3], floars[3][3], lzsa[3][3];
	double	deul[3], loar[3][3], loai[3][3], loars[3][3];

	if (mat2vecs(are, cell, &(cell[3]))) return;
	ztos(stang, lzsa);	/* get appropriate rotation matrix */
	mm3t(floars, are, lzsa);
	cell[6] = cell[7] = cell[8] = 0.; /* clear these in case it fails */
	if (vecmat(cell, cell + 3, floar)) return;
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
	   {
		loar[i][j] = floar[i][j];	loars[i][j] = floars[i][j];
	   }
	if (dinv3(loar, loai)) return;
	dmm3(loar, loai, loars);	/* get nearly orthogonal matrix loar */
	cleanorthog(loar);	/* make the orthogonal matrix more precise */
	qtoangs(loar, deul); /* get these rotation angles */
	for (i = 0; i < 3; i++)
		cell[i+6] = -deul[i] / RADC; /* convert to degrees */
}

int refparin(REFPARS *xp, int *nfoundp)
{ /*	This reads a set of refinement parameters into the REFPARS structure
	pointed to by xp. It then does the conversions that get us from the
	form found in the parameter file to the form we actually want to
	work with:
	* the xareal matrix is transposed and rotated from its stored position,
		which is the goniostat's (0,0,0) position,
		to (omegastart,chistart,phistart).
	* the xrec matrix input is ignored; it's recreated as the inverse of
		the above-transformed xareal and then rescaled by lambda so
		that it's in inverse Angstroms.
	* the unit cell lengths & angles, & the Euler angles deltaom,deltachi,
		and deltaphi are computed from the transformed areal matrix.
	The routine returns EOK if successful, various errors if not. */
	int		i, j;
	size_t		nhere;
	static FILE	*fpri;
	static size_t	minsize = 232;
	float		locrot[3][3], lsave[3][3], cel[9];

	*nfoundp = 0;
	if (NULL == xp) return EALLOC;	/* we need space for xp */
	if (NULL == (fpri = xgopen("uparams", "r"))) /* open the params file */
		return EOPENERR;
	zeromem((AVOID *)xp, sizeof (REFPARS));	/* explicitly zero xp */
	if (minsize > (nhere = xgread((AVOID *)xp, XG_REFPARS, fpri)))
		return EREADERR; /* read in at least 232 bytes */
	*nfoundp = nhere;
	(void)fclose(fpri);	fpri = NULL; /* close the file */
	if (FUZZ > xp->xlambda) return EBADVALUE; /* want sane lambda value */
		/* transpose the file's version of xareal into lsave */
	for (i = 3; i--; ) for (j = 3; j--; ) lsave[i][j] = xp->xareal[j][i];
	/* If possible, get the 3x3 matrix that rotates between
		(0,0,0) and (start) */
	if (nhere >= 256)
	  {
		ztos(xp->xstart, locrot); /* find 3x3 rotation matrix */
		mm3(xp->xareal, lsave, locrot); /* store areal at (start) */
	  }
	else {
		for (i = 3; i--; ) for (j = 3; j--; ) /* no info: just copy */
			xp->xareal[i][j] = lsave[i][j];
	     }
	/* invert that nice version into xrec */
	if ((i = inv3(xp->xareal, xp->xrec))) return i;
	/* rescale xrec by lambda and create a temporary copy of areal
	 that is in Angstroms rather than being unitless */
	for (i = 3; i--; ) for (j = 3; j--; )
	   {
		xp->xrec[i][j] /= xp->xlambda;
		lsave[i][j] = xp->xareal[i][j] * xp->xlambda;
	   }
	/* feed that temporary matrix to artocel to get the cell lengths,
		angles, and Euler angles */
	artocel(xp->xstart, lsave, cel);
	/* stick those results back into the REFPARS structure */
	for (i = 3; i--; )
	   {
		xp->xabc[i] = cel[i];	xp->xang[i] = cel[i + 3];
	   }
	xp->xdelom = cel[6];	xp->xdelchi = cel[7];	xp->xdelphi = cel[8];
	return EOK;
}

static void partoxou(REFPARS *xp)
{
 /* This outputs the parameters in a REFPARS structure in ascii form.
	If any of the outputs fails, we quit */
	int	i, j;
	float	xw[9];

	if (EOK != xgvariable("cell_len", xp->xabc, 3)) return;
	if (EOK != xgvariable("cell_ang", xp->xang, 3)) return;
	xw[0] = xp->xdelom;	xw[1] = xp->xdelchi;	xw[2] = xp->xdelphi;
	if (EOK != xgvariable("Euler_ang", xw, 3)) return;
	xw[0] = xp->xplated;	xw[1] = xp->xx0;	xw[2] = xp->xy0;
	xw[3] = xp->xtilt / RADC;	xw[4] = xp->xswing / RADC;
	if (EOK != xgvariable("detpars", xw, 5)) return;
	for (j = 0; j < 3; j++) xw[j] = xp->xgam[j] / RADC;
	if (EOK != xgvariable("gama", xw, 3)) return;
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		xw[3 * i + j] = xp->xareal[i][j];
	if (EOK != xgvariable("real_matrix", xw, -9)) return;
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		xw[3 * i + j] = xp->xrec[i][j];
	if (EOK != xgvariable("recip_matrix", xw, -9)) return;
	xw[0] = xp->xisys;
	if (EOK != xgvariable("crystal_system", xw, 1)) return;
	xw[0] = xp->xispgp;
	if (EOK != xgvariable("crystal_spacegroup", xw, 1)) return;
	xw[0] = xp->xresmin;	xw[1] = xp->xresmax;
	if (EOK != xgvariable("resolution_limits", xw, 2)) return;
	if (EOK != xgvariable("start_angles", xp->xstart, 3)) return;
	xw[0] = xp->xerrind;	xw[1] = xp->xerrx;
	xw[2] = xp->xerry;	xw[3] = xp->xerrphi;
	if (EOK != xgvariable("error_limits", xw, 4)) return;
	xw[0] = xp->xfrmsize / RADC;
	if (EOK != xgvariable("stepsize", xw, 1)) return;
	if (EOK != xgvariable("lambda", &(xp->xlambda), 1)) return;
	for (i = 0; i < 9; i++)
	   {
		xw[i] = ((xp->xisref[i] == 'y') || (xp->xisref[i] == 'Y'));
	   }
	if (EOK != xgvariable("cell_refinable", xw, 9)) return;
	for (i = 0; i < 9; i++)
	   {
		xw[i] = ((xp->xisref[i + 9] == 'y') ||
		 (xp->xisref[i+9] == 'Y'));
	   }
	if (EOK != xgvariable("detec_refinable", xw, 5)) return;
	for (i = 0; i < 9; i++)
	   {
		xw[i] = ((xp->xisref[i+14] == 'y') ||
			 (xp->xisref[i+14] == 'Y'));
	   }
	if (EOK != xgvariable("gamma_refinable", xw, 3)) return;
	(void)xgvariable(NULL, NULL, 0);
}

int refparout(REFPARS *xp)
{ /*	This routine is complementary to refparin().
	It makes some modifications to the unit cell matrices
	in the structure pointed to by xp and then writes out that structure.
	Specifically, it rotates the real-space matrix xp->xareal down from
	the datum position (omegastart,chistart,phistart) to (0,0,0) and
	transposes it.  It then defines xp->xrec to be the inverse of that.
	Then it transposes both of those matrices and writes the whole thing
	out. This returns EOK if successful, various error codes if not. */
	int		i, j;
	static FILE	*fpro;
	size_t		nhere;	
	float		locrot[3][3], lsave[3][3];

	if (NULL == xp) return EALLOC;	/* we need space for xp */
	if (FUZZ > xp->xlambda) return EBADVALUE; /* want sane lambda value */
	partoxou(xp);	/* output ascii version to .xou file */
	/* get the 3x3 matrix that rotates between
		(0,0,0) and (start) and use it to get to (0,0,0). */
	ztos(xp->xstart, locrot); /* find 3x3 rotation matrix */
	mm3t(lsave, xp->xareal, locrot);	/* find areal at (0,0,0) */
	/* write out the transformed version as the transpose of that */
	for (i = 3; i--; ) for (j = 3; j--; ) xp->xareal[j][i] = lsave[i][j];
	if ((i = inv3(xp->xareal, xp->xrec))) return i; /* invert for output */
	if (NULL == (fpro = xgopen("uparams", "r+"))) /* open the params file */
	  {
		if (NULL == (fpro = xgopen("uparams", "w"))) /* backup plan */
			return EOPENERR;
	  }
	if (256 > (nhere = xgwrite((AVOID *)xp, XG_REFPARS, fpro)))
		return EREADERR; /* write out the structure */
	(void)fclose(fpro);	fpro = NULL; /* close the file */
	return EOK;
}

int fmtoupr(xglong refrunflag, REFPARS *rprp)
{ /*	In this routine we attempt to open the FRAMES file.  If we can, we
	attempt to read one or more frames from the data.  If we can do so, we
	look in the header (including the DATARUN stuff) of the frame we read.
	We extract whatever useful info is in that frame header and return it
	in the REFPARS structure to which the routine points. */ 
	char		*frname;
	short		ctemp;
	int		frmno, i, tframes;
	FRAMEENTRY	*ftq;
	DATARUN		*ruhd;
	static MAIL	*lochd;
	static int		imflags;
	static FRAMEINDEX	*lfmi;
	static char		lochdmsg[] = "lochd in fmtoupr";

	/* construct the frame table */
	if (NULL == (lfmi = ftables(3))) return EALLOC;
	if (NULL == rprp) return EALLOC;
	/* don't try to use more than the first 40 frames */
	tframes = lfmi->fi_totfms;	if (tframes > 40) tframes = 40;
	imflags = JUST_HEADER | GET_RUNHEADER; /* just read frm headers */
	if (NULL == (lochd = (MAIL *)callop(1, sizeof (MAIL), lochdmsg)))
	  {
		(void)ftables(0);	return EALLOC;
	  }
	ruhd = NULL;
	for (ftq = lfmi->fi_fmi; ftq <= lfmi->fi_fme; ftq++)
	   {
		frmno = ftq->fmno;
		if (NULL != (frname = indexname(frmno)))
		  {
			if (FRAME_OK == read_image(frname, imflags, lochd)
			 && (NULL != (ruhd = lochd->runheader)) &&
			 (0.1 < ruhd->run_d0pos[0]))
			  {
				if (fabs(ruhd->run_lambda) > FUZZ)
					rprp->xlambda = ruhd->run_lambda;
				else	rprp->xlambda = DEF_LAMBDA;
				rprp->xctilt = 1.;
				rprp->xplated = ruhd->run_d0pos[0];
				rprp->xpltd2 = rprp->xplated * rprp->xplated;
				if (fabs(ruhd->run_stepreal) > DFUZZ)
				 rprp->xfrmsize = -RADC * ruhd->run_stepreal;
				else rprp->xfrmsize = (-RADC / 96. ) *
				 (float)(ruhd->run_stepsize);
					/* failsafe if that was zero... */
				if (!rprp->xfrmsize)
					rprp->xfrmsize = 0.25 * RADC;
				rprp->xgam[0] = rprp->xgam[1] =
					fabs(rprp->xfrmsize);
				rprp->xgam[2] = 0.0010;
				rprp->xswing = -ruhd->run_d0ang[0] * RADC;
				/* invert stepsize & swing on user request */
				if (refrunflag & RR_FLIPSTEP)
				  {
					rprp->xfrmsize *= -1.;
					rprp->xswing *= -1.;
				  }
				rprp->xssw = sin(rprp->xswing);
				rprp->xcsw = cos(rprp->xswing);
				for (i = 3; i--; )
				 {
					rprp->xstart[i] = ruhd->run_startpos[i];
					rprp->xabc[i] = 8. * rprp->xplated
						- 3. * (float)i;
					rprp->xang[i] = 90.;
					rprp->xrec[i][i] = 1. /
					 (rprp->xareal[i][i] = rprp->xabc[i] /
					 DEF_LAMBDA);
				 }
				rprp->xerrind = 0.2;
				rprp->xerrx = rprp->xerry = rprp->xerrphi = 5.;
				rprp->xresmin = rprp->xabc[0] * 1.1;
				for (i = 20; i--; ) rprp->xisref[i] = 'Y';
				/*May 2004: turn off plateD,tilt,swing,gamma2*/
				rprp->xisref[9] = rprp->xisref[12] =
				 rprp->xisref[13] = rprp->xisref[16] = 'N';
				saxadjust(lochd, rprp, frmno); /* Saxi tweak*/
				rprp->xmoveax = ruhd->run_motor;
				ctemp = ruhd->run_temp;
				rprp->xtemp = (ctemp < -128 ?
				 -128 : (ctemp > 127 ? 127 : ctemp));
				rprp->xrunno = ruhd->run_id;
				rprp->xchamno = ruhd->run_nchams;
				rprp->xisys = rprp->xispgp = 1;
				memzap((AVOID **)&lochd, lochdmsg);
				memzap((AVOID **)&ruhd, "ruhd in fmtoupr");
				(void)ftables(0);	return EOK;
			  }
		  }
	   }
	/* if we get here, we never found any good frames */
	(void)ftables(0);	memzap((AVOID **)&lochd, "lochd in fmtoupr");
	return EREADERR;
}

REFPARS *safepars(int *rtypep)
{ /* This returns a reasonable REFPARS structure. It tries to find
   one by reading a UPARAMS file; if that fails, it calls fmtoupr to
   get one from a detector image. It returns NULL if both fail.
   `rptypep' will contain 0 if we fail; it will contain 1 if we get
   the parameters from a UPARAMS file; it will contain -1 if we get
   it from a detector image */
	int		i, nfound;
	static REFPARS	*xp;

	*rtypep = 0;	/* default is failure */
	/* reserve memory for the REFPARS structure we'll use */
	if (NULL == (xp = (REFPARS *)callop(1, sizeof (REFPARS),
	 "xp in safepars"))) return NULL;
	if (EOK == (i = refparin(xp, &nfound))) /* try to read UPARAMS */
		*rtypep = 1;
	else if (EOK != fmtoupr(0, xp)) *rtypep = -1;
	return xp;
}

void makeq21(REFPARS *xp)
{ /* This creates nine values that constitute a rotation matrix xp->xq21[][]
   that corrects a diffraction vector or a similar object by
   the results of a tilt and a swing angle. */
	xp->xcsw = cos(xp->xswing);	xp->xssw = sin(xp->xswing);
	xp->xctilt = cos(xp->xtilt);	xp->xstilt = sin(xp->xtilt);
	xp->xq21[0][0] = -xp->xcsw;
	xp->xq21[0][1] = -(xp->xssw) * xp->xctilt;
	xp->xq21[0][2] = -(xp->xssw) * xp->xstilt;
        xp->xq21[1][0] =  xp->xssw;
	xp->xq21[1][1] = -(xp->xcsw) * xp->xctilt;
        xp->xq21[1][2] = -(xp->xcsw) * xp->xstilt;
        xp->xq21[2][0] = 0.;	xp->xq21[2][1] = -xp->xstilt;
	xp->xq21[2][2] =  xp->xctilt;
}

static int matfromcel(float lmatz[3][3])
{ /*	This reads a file with environment variable REINDEX
	as containing a 3x3 transformation matrix and a 3x1 translation vector.
	It ignores the vector and returs the matrix in `lmatz'. The routine
	returns 1 if the matrix is valid and is not the identity matrix.
	It returns 0 otherwise. The translation vector can be absent.
	In the event that the determinant of the transformation matrix is
	negative, all nine elements of lmatz are negated. */
	char	reiline[60];
	int	i, nitems, j, tv[3];
	double	eun, melt, detl;
	FILE	*fprei;

	if (NULL == (fprei = xgopen("reindex", "r"))) return 0; /* open it */
	for (i = 0; i < 3; i++) /* read in its 3 lines */
	   {
		if (qerror((AVOID *)fgets(reiline, 58, fprei),
			"read line from tranformation-matrix file"))
				return EREADERR;
		lmatz[i][0] = lmatz[i][1] = lmatz[i][2] = tv[i] = 0;
		nitems = sscanf(reiline, "%f %f %f %d",
			&lmatz[i][0], &lmatz[i][1], &lmatz[i][2], &tv[i]);
		if (nitems < 4)
		  {
			tv[i] = 0;
			if (nitems < 3)
			  {
				(void)fclose(fprei);	return 0;
			  }
		  }
	   }
	(void)fclose(fprei);
	detl = deter(lmatz);
	if (fabs(detl) < DFUZZ) return 0;
	if (detl < 0.)
	  {
		for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
			lmatz[i][j] = -lmatz[i][j];
	   }
	for (eun = i = 0; i < 3; i++)
	   {
		for (j = 0; j < 3; j++)
		   {
			melt = lmatz[i][j];
			if (i == j) melt -= 1.;
			eun += melt * melt;
		   }
	   }
	if (eun < DFUZZ) return 0;	/* identity matrix: return 0 */
	return 1;
}

int reducedcell(float are[3][3])
{ /*	This reorganizes a unit-cell matrix `are'. It is not necessarily
	either unitless (like the "areal" matrix used elsewhere)
	or in Angstroms. The reorganization will leave `are' such
	that the conditions set forth in the International Tables for a
	reduced cell on pp. 741 - 745 are satisfied. The routine returns
	0 if successful, nozero if not. */
	int	i, j, k, m;
	int	sho, mid, lon, ntry, type2, negf;
	double	afar, bdot, compa, fcomp, mdot, tc, td, dot[3], le[3];
	float	cvm[3][3], tar[3][3], nar[3][3];

	for (ntry = 0; ntry < 100; ntry++) /* iterate at most 100 times */
	   { /* first, determine the unit cell lengths */
		for (i = 0; i < 3; i++)
		   {
			for (le[i] = j = 0; j < 3; j++)
			   {
				tc = are[i][j]; le[i] += tc * tc;
			   }
			if (le[i] < DFUZZ) return -1;
			le[i] = sqrt(le[i]);
		   }
		/* figure out which one is shortest and which is longest */
		lon = 2; afar = le[2];
		if ((tc = le[1]) > afar)
		  {
			lon = 1;	afar = tc;
		  }
		if (le[0] > afar) lon = 0;
		sho = lon + 1; if (sho > 2) sho -= 3; mid = 3 - lon - sho;
		if (le[mid] < le[sho])
		  {
			sho = mid; mid = 3 - lon - sho;
		  }
		switch (lon)	/* pick transformation matrix for next step: */
		  {
			case	0: cvmat(((sho == 1) ? 5 : 7), cvm);	break;
			case	1: cvmat(((sho == 0) ? 3 : 6), cvm);	break;
			case	2: cvmat(((sho == 1) ? 4 : 8), cvm);	break;
		  }
		mm3(tar, cvm, are); /* reorganize unit cell so a <= b <= c */
		for (i = 0; i < 3; i++) /* get the lengths again */
		   {
			for (le[i] = j = 0; j < 3; j++)
			   {
				tc = tar[i][j]; le[i] += tc * tc;
			   }
			if (le[i] < DFUZZ) return -2;
			le[i] = sqrt(le[i]);
		   }
		for (i = 0; i < 3; i++) /* find the dot products */
		   {
			j = i + 1;	if (j > 2) j -= 3; k = 3 - i - j;
			for (compa = dot[i] = m = 0; m < 3; m++)
			   {
				tc = tar[j][m]; td = tar[k][m];
				dot[i] += tc * td;
				compa += (i == 1) ? td * td : tc * tc;
			   }
			if (compa < DFUZZ) return -2;
			dot[i] = 2. * dot[i] / compa;
		   }
		/* make sure the cell is either type I or type II.
		 This involves seeing if the three dot products all
		 have the same sign. */
		negf = (dot[0] <= 0.) + 2 * (dot[1] <= 0.) +
			4 * (dot[2] <= 0.);
		/* special cases when vectors are very close to 90.
		 In this new version, we try for obtuse angles in this case */
		switch (negf)
		  {
			case	0:
			 if (dot[0] < DFUZZ) negf = 1;
			 else if (dot[1] < DFUZZ) negf = 2;
			 else if (dot[2] < DFUZZ) negf = 4;
			 break;
			case	1:	break;
			case	2:	break;
			case	3:
			 if (dot[0] > -DFUZZ) negf = 2;
			 else if (dot[1] > -DFUZZ) negf = 1;
			 else if (dot[2] < DFUZZ) negf = 7;
			 break;
			case	4:	break;
			case	5:
			 if (dot[0] > -DFUZZ) negf = 4;
			 else if (dot[1] < DFUZZ) negf = 7;
			 else if (dot[2] > -DFUZZ) negf = 1;
			 break;
			case	6:
			 if (dot[0] < DFUZZ) negf = 7;
			 else if (dot[1] > -DFUZZ) negf = 4;
			 else if (dot[2] > -DFUZZ) negf = 2;
			 break;
			default:	break;
		  }
		if (negf > 3) negf = 7 - negf;
		if (negf) /* supplement 2 angles if necessary */
		  {
			switch(negf)
			 {	
				case	1: cvmat(0, cvm); mm3(nar, cvm, tar);
				 dot[1] = -dot[1]; dot[2] = -dot[2]; break;
				case	2: cvmat(1, cvm); mm3(nar, cvm, tar);
				 dot[0] = -dot[0]; dot[2] = -dot[2]; break;
				case	3: cvmat(2, cvm); mm3(nar, cvm, tar);
				 dot[0] = -dot[0]; dot[1] = -dot[1]; break;
				default:	break;
			  }
			for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
				tar[i][j] = nar[i][j];
		  }
		/* record as "mdot" the largest of the absolute
		 values of 2*(b.c)/(b.b), 2*(c.a)/(a.a), and
		 2*(a.b)/(a.a).  The corresponding signed value
		 is recorded as "bdot" */
		for (bdot = dot[0], mdot = i = 0; i < 3; i++)
		   {
			compa = dot[i];	fcomp = fabs(compa);
			if (fcomp > mdot)
			  {
				bdot = compa; mdot = fcomp; lon = i;
			  }
		   }
		/* if mdot is substanatially greater than one,
		 multiply by an appropriate matrix */
		if (mdot > 1.0005)
		  {
			type2 = (bdot < 0); /* true if a type-2 cell */
			for (i = 0; i < 3; i++) /* construct matrix */
			   {
				nar[i][i] = 1.; j = i + 1;
				if (j > 2) j -= 3; k = 3 - i - j;
				nar[i][j] = nar[i][k] = 0.;
			   }
			/* modify identity matrix to make transform */
			tc = (type2) ? 1. : -1;
			if (lon == 0) nar[2][1] = tc;
			else if (lon == 1) nar[2][0] = tc;
			else nar[1][0] = tc;
			mm3(are, nar, tar); /* apply transform to matrix */
		  }
		else {
			for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
				are[i][j] = tar[i][j];
			return 0;
		     }
	   }
	return -3;
}

static void rotcell(REFPARS *xp, float cmat[3][3])
{ /*	rotates real-space unit cell by `cmat' and gets unit cell params */
	register int	i, j, k;
	double		ut, dotp;
	float		ate[3][3];

	mm3(ate, cmat, xp->xareal); /* rotate appropriately */
	for (i = 0; i < 3; i++) /* real-space unit cell lengths */
	   {
		for (ut = j = 0; j < 3; j++) ut += ate[i][j] * ate[i][j];
		xp->xabc[i] = xp->xlambda * sqrt(ut);
	   }
	for (i = 0; i < 3; i++)	/* unit cell angles from dot products */
	   { /* e.g., gamma = acos((a.b) / |a||b|) */
		j = i + 1;	if (j > 2) j = 0;	k = 3 - i - j;
		dotp = (ate[j][0] * ate[k][0] + ate[j][1] * ate[k][1] +
			ate[j][2] * ate[k][2]) * xp->xlambda * xp->xlambda;
		xp->xang[i] = (fabs(dotp) > DFUZZ) ?
		  acos(dotp / (xp->xabc[j] * xp->xabc[k])) / RADC : 90.;
	   }
	for (i = 0; i < 3;i++) for (j = 0; j < 3;j++)
		xp->xareal[i][j] = ate[i][j];
}

int cellconvert(REFPARS *xp, int convtype)
{ /*	This converts a unit cell in various ways. In some cases it takes
	a non-reduced cell and converts to a reduced cell; in others it
	converts according to particular formulae based on the argument
	"convtype", as follows:
      convtype  meaning
	 0	back out of a non-primitive cell but don't
		 run `reducedcell'
	 1	convert to the reduced type I or type II cell
	 2	permute forward, (a,b,c)->(b,c,a)
	 3	permute backward, (a,b,c)->(c,a,b)
	 4	swap a for b
	 5	swap a for c
	 6	swap b for c
	 7	supplement alpha, beta
	 8	supplement alpha, gamma
	 9	supplement beta, gamma
	10	use reindexing matrix in file REINDEX
	Some of the values have been removed since an earlier version
	of this routine because they've been much more effectively
	implemented in the "Bravais" command.
	The routine returns 0 if successful, nonzero if either
	the initial call to `reducedcell' or the final rp2vecs fails.
	There's still some uncertainty about what we're doing here:
	I'm worried that in centered cells, errors may be occurring. */
	int	i, j;
	float	cvm[3][3], tar[3][3]; 
	static float	lmatc[3][3] = { { 0.5,-0.5, 0.0 }, { 0.5, 0.5, 0.0 },
				{ 0.0, 0.0, 1.0 } } ;
	static float	lmatf[3][3] = { { 0.0, 0.5, 0.5 }, { 0.5, 0.0, 0.5 },
				{ 0.5, 0.5, 0.0 } } ;
	static float	lmati[3][3] = { {-0.5, 0.5, 0.5 }, { 0.5,-0.5, 0.5 },
				{ 0.5, 0.5,-0.5 } } ;
	static float	lmatr[3][3] = { { R3, -R3, -R3 }, { -T3, -R3, -R3 },
				 { R3, T3, -R3 } } ;
	static float	liden[3][3] = {{ 1.,0.,0.},{0.,1.,0.},{0.,0.,1.}};

	switch (convtype)
	      {
		case	0:	/* back out of centered cell */	
		case	1: /* force into reduced cell */
		 j = exclucode(xp->xispgp, xp->xisys);	/* lattice type */
		 switch (j)  /* convert back OUT of the centered cells */
		   {
			case	1: mm3(tar, lmatc, xp->xareal); break;	/* C */
			case	2: mm3(tar, lmatf, xp->xareal); break;	/* F */
			case	3: mm3(tar, lmati, xp->xareal); break;	/* I */
			case	4: mm3(tar, lmatr, xp->xareal); break;	/* R */
			case	5: mm3(tar, lmatf, xp->xareal); break;	/* F */
			default:   mm3(tar, liden, xp->xareal); break;	/* P */
		   }
		 xp->xisys = xp->xispgp = 1;
		 if ((convtype) && (i = reducedcell(tar))) return i;
		 for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
			xp->xareal[i][j] = tar[i][j];
		 break;
		case	 2:	/* permute forward */
			cvmat(5, cvm);	rotcell(xp, cvm);	break;
		case	 3:	/* permute backward */
			cvmat(6, cvm);	rotcell(xp, cvm);	break;
		case	 4:	/* swap a for b */
			cvmat(4, cvm);	rotcell(xp, cvm);	break;
		case	 5:	/* swap a for c */
			cvmat(7, cvm);	rotcell(xp, cvm);	break;
		case	 6:	/* swap b for c */
			cvmat(3, cvm);	rotcell(xp, cvm);	break;
		case	 7:	/* supplement alpha and beta */
			cvmat(2, cvm);	rotcell(xp, cvm);	break;
		case	 8:	/* supplement alpha and gamma */
			cvmat(1, cvm);	rotcell(xp, cvm);	break;
		case	 9:	/* supplement beta and gamma */
			cvmat(0, cvm);	rotcell(xp, cvm);	break;
		case	10:	/* use file-based REINDEX matrix */
			if (matfromcel(tar)) rotcell(xp, tar); break;
		default:	break;
	      }
	return rp2vecs(xp);
}

/* end of xgparams.c */
