/* xgpredict.c -- Predict reflections over a datarun range based on UPARAMS

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	28 Nov 2006: presval, pdetec_limit rewritten with new REFPARS elements;
		makeq21, ucell calls updated
	24 Jun 2006: xgenpred.h includes xgen.h, so we don't need the latter
	 3 Jan 2006: user can specify low and high-resolution limits
	20 Nov 2005: inline argument declarations and strict prototyping
	16 Aug 2005: move VACANT, CMSLOP from ../include/xgen.h to here
	22 Jun 2004: remove pcmp(): we use it only in utilities/predictit.c;
		 presval(), pdetec_limit(), pgamval(), reppart_prange() static
	17 Jun 2005: presval(): call to pix2cm includes frame number arg
	30 Aug 2004: created from ../nogui/predict.c
 Pertinent changes to ../nogui/predict.c prior to that date:
	21 Jun 2004: increase REFLSPERRANGE up to 480000
	25 Jan 2004: several static globals converted to local variables
	21 Jan 2004: license message inserted and initial comments reordered
	 6 Jan 2004: static REFPARS structure elements replace extern globals;
		refparin() and makeq21() used
	16 Aug 2003: handle exclusion codes differently so we don't have
		to carry around a global variable for them
	 5 Jan 2003: reorganized source-file header
	25 Mar 2002: removed MSI attachments
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	 7 Dec 1993: initial X-GEN version

   Purpose: obtain list of reflections expected to occur over a defined
	frame range, given a known set of crytal and detector parameters.
   Cross references: xgenmisc.c
*************************************************************************/
#define		NEED_TIME	1
#define		NEED_ERRNO	1
#include	"xgenpred.h"
#define		ZSLOP		0.05
#define		MAXRANGE	(30 * RADC)	/* longest uninterruped range*/
#define		REFLSPERRANGE	480000	/* max refls to predict per range */
#define		SLOP		(0.1 * RADC)	/* omega slop, in radians */
#define		SLOPCONST	PIO2	/* h,k,l slop scale factor */
#define		VACANT		255	/* flag for missing prediction */
#define		CMSLOP		0.05	/* slop in pixel values, in cm */

PREDWORK *manageprw(int opno)
{ /*	Manage a static pointer prw to a PREDWORK structure */
	static PREDWORK	*prw;
	static char	prwmsg[] = "prw in manageprw";
	static char	prp0msg[] = "prw->po_prp0 in manageprw";

	if (!opno)
	  {
		if (NULL != prw)
		  {
			memzap((AVOID **)&(prw->po_prp0), prp0msg);
			memzap((AVOID **)&prw, prwmsg);
		  }
		return NULL;
	  }
	if (opno == 1) prw = manageprw(0);
	if (NULL == prw)
	  {
		if (NULL == (prw = callop(1, sizeof (PREDWORK), prwmsg)))
			return NULL;
	  }
	if ((opno > 0x10) && (NULL != prw))
	  {
		memzap((AVOID **)&(prw->po_prp0), prp0msg);
		if (NULL == (prw->po_prp0 = callop((size_t)opno,
		 sizeof (PREDREF), prp0msg))) return manageprw(0);
	  }
	return prw;
}

int envtopol(double *polfp, double *polnp)
{ /*	This interprets the contents of the environment variable
	POLARIZATION_TYPE as providing information about the polarization
	of the beam and returns that information in *polfp, *polnp,
	as the fraction polarized (0 to 1.) and the angle, in degrees, between
	the normal to the plane of polarization and the rotation axis (0 to 90.)
	The routine returns an integer polarization type, as defined below. */
	char	*polstr, *cp;
	int	poltype;

	poltype = 0;	/* default is undefined */
	polstr = getenv("POLARIZATION_TYPE"); /* look for environ vbl */
	if ((NULL == polstr) || ('\0' == *polstr))
		polstr = getenv("polarization_type");
	/* the environment variable might have a numeric value or an
	 alphabetic value (case-insensitive):
	  numeric	alphabetic	meaning
		0	--		none
		1	mi, u		unpolarized (e.g. mirrors)
		2	g, mo		graphite monochromator
		3	s		synchrotron, rotation axis horizontal
		4	sv		synchrotron, rotation axis vertical
		5	b		bending-magnet, rot axis horizontal
		6	bv		bending-magnet, rot axis vertical
	 */
	if ((NULL != polstr) && ('\0' != *polstr))
	  {
		for (cp = polstr; '\0' != *cp; cp++)
		   {
			if (('a' <= *cp) && (*cp <= 'z')) *cp -= 'a' - 'A';
		   }
		if (allnumeric(polstr)) sscanf(polstr, "%d", &poltype);
		else if (*polstr == 'S')
			poltype = (*(polstr + 1) == 'V') ? 4 : 3;
		else if (*polstr == 'M')
			poltype = (*(polstr + 1) == 'O') ? 2 : 1;
		else if (*polstr == 'G') poltype = 2;
		else if (*polstr == 'U') poltype = 1;
		else if (*polstr == 'B')
			poltype = (*(polstr + 1) == 'V') ? 6 : 5;
	  }
	if (poltype < 1) poltype = 1;	else if (poltype > 6) poltype = 6;
	switch (poltype)
	  {
		case	1: *polfp = 0.5; *polnp = 90.; break; /*unpolarized*/
		case	2: *polfp = 0.4722; *polnp = 90.; break; /*graphite */
		case	3: default: *polfp = 1.; *polnp = 90.;	break;
		case	4: *polfp = 1.; *polnp = 0.;	break; /*ID line,vert*/
		case	5: *polfp = 0.95; *polnp = 90.;	break; /*BM line,hor */
		case	6: *polfp = 0.95; *polnp = 0.;	break; /*BM line,vert*/
	  }
	return poltype;
}

static double presval(int x, int y, double *xcm, double *ycm, REFPARS *cxp)
{ /*	This returns the |s|^2 = 4 * sinsq(theta) value
	associated with a pixel position (x, y).  It returns 0
	for pixels at which no pix-to-cm conversion can be determined. */
	int		i, j;
	double		rni, rni2, lq21[3][3], sl[3], ssz[3];
	float		lxcm, lycm;

	if (1 != pix2cm((float)x, (float)y, &lxcm, &lycm, 0)) return 0.;
	*xcm = lxcm - cxp->xx0;	*ycm = lycm - cxp->xy0;
	rni2 = (*xcm) * (*xcm) + (*ycm) * (*ycm) + cxp->xpltd2;
	rni = sqrt(rni2);
	ssz[0] = cxp->xplated / rni;
	ssz[1] = (*xcm) / rni;	ssz[2] = (*ycm) / rni;
/* rotate this by the detector tilt and swing-angle matrices */
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		lq21[i][j] = cxp->xq21[i][j];
	dmm31(sl, lq21, ssz);	/* gifs the local scattering vector */
	sl[0] += 1.;	/* correct the beam-direction component */
	return	sl[0] * sl[0] + sl[1] * sl[1] + sl[2] * sl[2];
}

static double pdetec_limit(double *detext, VSPLINE *sgsp, REFPARS *cxp)
{ /*	This determines the maximum value of |s|^2 = 4 * sinsq(theta)
	possible with the detector in a particular configuration.
	It does this by computing |s|^2 for most of the pixels along the edge of
	the active region of detector and returning the largest |s| value found.
	The check to see if the pixel is in the active area is implicit
	in the call to `presval'.
	At the same time it computes the smallest and largest values
	of xcm and ycm, the centimeter-coordinate values of x and y.
	We repeat the process along the columns as well.
	Mean value theorem, and all that. */
	int	locw, loch, x, y, x_edge, y_edge;
	double	sed, smsq, zxc, zyc;

	/* Initialize the things of which we're finding maxima & minima */
	detext[0] = detext[1] = 1.e20;	detext[2] = detext[3] = -detext[0];
	smsq = 0;
	/* now go through the rows */
	locw = sgsp->v_detw;	loch = sgsp->v_deth;
	for (y = 1; y < loch; y += 2)
	   {
		x_edge = locw;
	   	for (x = 1; x < locw; x += 2)
		   {
			if ((sed = presval(x, y, &zxc, &zyc, cxp)) > 0.)
			  {
				if (zxc < detext[0]) detext[0] = zxc;
				if (zyc < detext[1]) detext[1] = zyc;
				if (zxc > detext[2]) detext[2] = zxc;
				if (zyc > detext[3]) detext[3] = zyc;
				if (sed > smsq) smsq = sed;
				x_edge = x;	break;
			  }
		   }
		for (x = locw - 1; x > x_edge; x -= 2)
		   {
			if ((sed = presval(x, y, &zxc, &zyc, cxp)) > 0.)
			  {
				if (zxc < detext[0]) detext[0] = zxc;
				if (zyc < detext[1]) detext[1] = zyc;
				if (zxc > detext[2]) detext[2] = zxc;
				if (zyc > detext[3]) detext[3] = zyc;
				if (sed > smsq) smsq = sed;
				break;
			  }
		   }
	   }
	/* now, the columns. */
	for (x = 1; x < locw - 1; x += 2)
	   {
		y_edge = loch;
	   	for (y = 1; y < loch; y += 2)
		   {
			if ((sed = presval(x, y, &zxc, &zyc, cxp)) > 0.)
			  {
				if (zxc < detext[0]) detext[0] = zxc;
				if (zyc < detext[1]) detext[1] = zyc;
				if (zxc > detext[2]) detext[2] = zxc;
				if (zyc > detext[3]) detext[3] = zyc;
				if (sed > smsq) smsq = sed;
				y_edge = y;	break;
			  }
		   }
		for (y = loch - 1; y > y_edge; y -= 2)
		   {
			if ((sed = presval(x, y, &zxc, &zyc, cxp)) > 0.)
			  {
				if (zxc < detext[0]) detext[0] = zxc;
				if (zyc < detext[1]) detext[1] = zyc;
				if (zxc > detext[2]) detext[2] = zxc;
				if (zyc > detext[3]) detext[3] = zyc;
				if (sed > smsq) smsq = sed;
				break;
			  }
		   }
	   }
	/* Bugfix December 2000: correct for the translational offsets here */
	detext[0] += cxp->xx0;	detext[1] += cxp->xy0;
	detext[2] += cxp->xx0;	detext[3] += cxp->xy0;
	detext[0] -= CMSLOP;	detext[1] -= CMSLOP;
	detext[2] += CMSLOP;	detext[3] += CMSLOP;
	return smsq * 1.005;
}

static double pgamval(REFPARS *xp, double svl0, double svl1)
{ /*	This computes the gamma value for whose local diffraction vector
	components are (svl0, svl1, svl2), based on the rocking curve values.
	We divide the actual gamma value by |svl1| / sqrt(svl0^2 + svl1^2)
	to correct for the fact that the oscillation axis is vertical.
	Note that unlike the older version of this routine, it explicitly
	uses the diffraction vector components rather than back-calculating
	them from the observed position of the reflection in centimeters.
	Change as of 3 Sep 2004: spectral component guaranteed non-negative. */
	double	etasq, svl1sq, mysq, mzsq, spectrum, gamdirec, gscale;

	etasq = svl0 * (2. - svl0); svl1sq = svl1 * svl1;
	mzsq = xp->xgam[0] * svl1sq;
	mysq = xp->xgam[1] * (etasq - svl1sq);
	spectrum = fabs(xp->xgam[2] * svl0);
	gamdirec = sqrt(mysq * xp->xgam[1] + mzsq * xp->xgam[0]);
	gscale = 1. / fabs(xp->xfrmsize * svl1);
	return (gamdirec + spectrum) * gscale;
}

int pfindref(PREDWORK *pwp)
{ /* "pfindref":
	For fixed hkl this predicts a refl and sticks it in the list.
	It computes the two possible omega values of the reflection
	along with some geometrical quantities. If either of the
	omega values falls within the scanning-angle range, the
	X,Y position of the reflection, its Lorentz/polarization/
	air-absorption correction factor, and its rocking-curve value
	are computed. See Xuong et al (1978) Acta Cryst.A34:289-296
	for definitions of the variables beta, rphi, xi, and zw (the
	latter is called gamma in the paper). */
	char		isfound;
	int		alongax, cva, i, jfr;
	short		lh[3];
	double		cosarg, cxcm, cycm, g00, g01, g02;
	double		rsz, sinarg, spind0, spind1, ssq, sx, sy, sz;
	double		tfr, xisq, y2, y2sq, y3, zetasq;
	REFPARS		*xp;
	PREDREF		*prp;

/*	See if the Z component of the scattering vector falls outside the
	limits set by the top and bottom of the detector face. */
	sz = pwp->po_p[0][2][0] * pwp->po_ph[0] + pwp->po_p[0][2][1] *
	 pwp->po_ph[1] + pwp->po_p[0][2][2] * pwp->po_ph[2];
	if (sz < pwp->po_szmin || sz > pwp->po_szmax) return PR_SZRANGE;
	/* get the true hkl values and make sure they're spacegroup-legal */
	for (i = 0; i < 3; i++) lh[(int)(pwp->po_ih[i])] = pwp->po_ph[i];
	if (!keephkl(lh, pwp->po_excluc)) return PR_BADHKL;
/*	Compute the (X,Y) components of the scattering vector and the
	sum of their squares. This sum must be nonzero. */
	xp = pwp->po_xp;
	sx = (xp->xrec[0][0] * lh[0] + xp->xrec[0][1] * lh[1] +
		xp->xrec[0][2] * lh[2]) * xp->xlambda;
	sy = (xp->xrec[1][0] * lh[0] + xp->xrec[1][1] * lh[1] +
		xp->xrec[1][2] * lh[2]) * xp->xlambda;
	if (DFUZZ >= (xisq = sx * sx + sy * sy)) return PR_ALONGBEAM;
/* Now check that the scattering vector is in range */
	zetasq = sz * sz;
	if ((ssq = zetasq + xisq) < pwp->po_smisq) return PR_TOOLOWRES;
	if (ssq > pwp->po_smasq) return PR_TOOHIRES;
/* Compute the calculated (X,Y,phi) positions of a reflection with the
	scattering vector `s'. We need to calculate 2 X,Y,phi triplets,
	coming from the omega,X,Y positions of the first and second
	intersections of the scattering vector with the Ewald sphere. */
	rsz = 0.5 * ssq;
	/* change as of 3 Sep 04: we predict reflections with small positive
	y2sq values but we flag them by setting prp->pcor negative */
	if (0. > (y2sq = xisq - rsz * rsz)) return PR_BADY2;
	alongax = (pwp->po_miny2sq > y2sq);
	(pwp->po_nchecked)++;	/* any other check is sort of slow */
	y2 = sqrt(y2sq);
	/* we use the arctan2 form of the calculation because the
	behavior at nodes is good and the range is automatically correct. */
	cosarg = rsz * sx - y2 * sy;	sinarg = y2 * sx + rsz * sy;
	spind0 = atan2(sinarg, cosarg);
	isfound = (pwp->po_mumin < spind0 && spind0 <= pwp->po_mumax);
	cosarg = rsz * sx + y2 * sy;	sinarg = -y2 * sx + rsz * sy;
	spind1 = atan2(sinarg, cosarg);
	if (pwp->po_mumin < spind1 && spind1 <= pwp->po_mumax) isfound |= 02;
	if (!isfound) return PR_PHIRANGE; /* neither phi was in range: quit */
	/* now find the (X,Y) position of the reflection */
	prp = pwp->po_prp;
	prp->pindx = hpack(lh); /* get a one-xglong version of hkl */
	prp->pcor = 10. * y2 / (1. - y2sq * pwp->po_y2co -
	 zetasq * pwp->po_z2co - y2 * sz * pwp->po_yzco);
	if (alongax) prp->pcor = -(prp->pcor);
	y3 = 1. - rsz;
	if (isfound & 01)	/* 1st phi value in range */
	  {
		g00 = (xp->xcsw * y3 - xp->xssw * y2) / xp->xplated;
		if (g00 < DFUZZ) isfound &= 02; /* g00 > 0 for plated > 0 */
		else {
			g01 =	xp->xssw * xp->xctilt * y3 +
				xp->xcsw * xp->xctilt * y2 - xp->xstilt * sz;
			g02 =	xp->xssw * xp->xstilt * y3 +
				xp->xcsw * xp->xstilt * y2 + xp->xctilt * sz;
			cxcm = xp->xx0 + g01 / g00; /* X position in cm */
			if (pwp->po_minxcm <= cxcm && cxcm <= pwp->po_maxxcm)
			  {
			    cycm = xp->xy0 + g02 / g00; /* Y pos. in cm */
			    tfr = (spind0 + pwp->po_rangecen) / xp->xfrmsize -
				0.5;	jfr = tfr + 0.499999;
			    if ((pwp->po_minycm <= cycm) &&
				(cycm <= pwp->po_maxycm) && (1 <= (cva =
				 cm2pix(cxcm, cycm,&prp->pxt,&prp->pyt,jfr))))
				/* The observation is in range and is hitting
				 the screen; save it */
			      {
				prp->pgama = pgamval(pwp->po_xp, rsz, y2);
				if (cva > 1) prp->pgama = -(prp->pgama);
				prp->pphi = tfr;
				prp->pfmno = (short)(prp->pphi + 0.49999);
				(pwp->po_prp)++;	prp = pwp->po_prp;
			      } /* end of "if ((pwp->po_minycm " clause */
			     else	isfound &= 02; /* no fit here */
			  } /* end of "if ((pwp->po_minxcm " clause */
		    else isfound &= 02;
		  } /* end of "if (g00 " clause */
	  } /* end of "if (isfound & 01) " clause */
	if (isfound & 02)	/* second solution; horiz. opposite 1st soln.*/
	  {
		y2 = -y2;
		g00 = (xp->xcsw * y3 - xp->xssw * y2) / xp->xplated;
		if (g00 < DFUZZ) isfound &= 01;
		else {
			g01 =	xp->xssw * xp->xctilt * y3 +
				xp->xcsw * xp->xctilt * y2 - xp->xstilt * sz;
			g02 =	xp->xssw * xp->xstilt * y3 +
				xp->xcsw * xp->xstilt * y2 + xp->xctilt * sz;
			cxcm = xp->xx0 + g01 / g00; /* X position in cm */
			/* in range ? */
			if (pwp->po_minxcm <= cxcm && cxcm <= pwp->po_maxxcm)
			  {
			    cycm = xp->xy0 + g02 / g00;	/* Y pos. in cm */
			    tfr = (spind1 + pwp->po_rangecen) / xp->xfrmsize
				- 0.5;	jfr = tfr + 0.499999;
			    if ((pwp->po_minycm <= cycm) &&
			     (cycm <= pwp->po_maxycm) && (1 <= (cva =
				cm2pix(cxcm,cycm,&prp->pxt,&prp->pyt, jfr))))
				/* The observation is in range and is hitting
				 the screen; save it */
			      {
				prp->pgama = pgamval(pwp->po_xp, rsz, y2);
				if (cva > 1) prp->pgama = -(prp->pgama);
				prp->pphi = tfr;
				prp->pfmno = (short)(prp->pphi + 0.49999);
				if (isfound & 01)
				/* other value also in range:
				we'll fill in the items common to both */
				  {
					prp->pindx =((pwp->po_prp) - 1)->pindx;
					prp->pcor = ((pwp->po_prp) - 1)->pcor;
				  } /* end of "if (isfound &" clause */
				(pwp->po_prp)++;	prp = pwp->po_prp;
			      } /* end of "if (pwp->po_minycm" clause */
			    else isfound &= 01;
			  } /* end of "if (pwp->po_minxcm" clause */
			else isfound &= 01;
		  } /* end of "if (g00) " clause */
	  } /* end of "if (isfound & 02" clause */
	return (int)isfound;
}

static void pdolpred(PREDWORK *pwp, float llow, float lhigh)
{ /*	This loops through a group of legal l values for
	fixed h and k and calls the prediction routine `pfindref'.
	Each reflection is checked to make sure that it is legal in
	the space group before the call to findref. */
	short	lstart,lend;

	if (lhigh < pwp->po_lmin || llow > pwp->po_lmax) return;
	lstart = iroup((double)(llow - pwp->po_hslop));
	lend = irodown((double)(lhigh + pwp->po_hslop));
	if (pwp->po_lmin > lstart) lstart = pwp->po_lmin;
	if (pwp->po_lmax < lend) lend = pwp->po_lmax;
	for (pwp->po_ph[2] = lstart; pwp->po_ph[2] <= lend; (pwp->po_ph[2])++)
		pwp->po_nresult[pfindref(pwp)]++;	/* DO IT! */
	return;
}

static void pllimits(PREDWORK *pwp)
{ /*	This manages the innermost loop through candidate
	reflections during prediction. The innermost index is conventionally
	called "l" here, although it might actually be h or k.
	For fixed h,k this computes the limits of l.
	There may be zero, one, or two ranges of l to check;
		the actual prediction routine is called for the
		l values in as many ranges as there are. */
	register int	i, j;
	double		bl, cl, t1, t2, lv[2][2], q[3], phk[3];

	/* Get the resolution-sphere L limits */
	bl = pwp->po_ph[1] * pwp->po_ps[4] + pwp->po_ph[0] * pwp->po_ps[5];
	cl = pwp->po_hsq * pwp->po_ps[1] +
	 pwp->po_ph[1] * pwp->po_ph[1] * pwp->po_ps[2] +
	 pwp->po_ph[0] * pwp->po_ph[1] * pwp->po_ps[6] - pwp->po_rhomax;
	t2 = bl * bl - 4. * pwp->po_al * cl; /* resol-limit calc discriminant*/
	if (t2 <= 0.) return;	/* discrim < 0: skip this (h,k) */
	t2 = sqrt(t2) / (2. * pwp->po_al);
	t1 = -bl / (2. * pwp->po_al);
	pwp->po_lmin = iroup(t1 - t2 - pwp->po_hslop);
	pwp->po_lmax = irodown(t1 + t2 + pwp->po_hslop);
	/* get the Ewald-sphere ranges of l for this (h,k) */
	for (i = 0; i < 2; i++)	/* check both range extrema */
	   {
		for (j = 0; j < 3; j++) phk[j] = pwp->po_p[i][j][0] *
			pwp->po_ph[0] + pwp->po_p[i][j][1] * pwp->po_ph[1];
		phk[0] -= 1.;	/* account for shift of Ewald sphere */
		q[0] = -1.;	q[1] = q[2] = 0.;
		for (j = 0; j < 3; j++)
		   {
			t1 = phk[j]; t2 = pwp->po_p[i][j][2];
			q[0] += t1 * t1; q[1] += t1 * t2; q[2] += t2 * t2;
		   }
		t2 = q[1] * q[1] - q[0] * q[2];
		if (t2 < 0.)	lv[i][0] = VACANT;
		else {
			t2 = sqrt(t2) / q[2]; t1 = -q[1] / q[2];
			lv[i][0] = t1 - t2; lv[i][1] = t1 + t2;
	 	     }
	   }
	/* Permute the indices of the l-limit array "lv" so that
	 the lower limit of the lower range is always lv[0][0]. */
	if (lv[0][0] > lv[1][0])
	  for (i = 0; i < 2; i++)
	     {
		t1 = lv[0][i]; lv[0][i] = lv[1][i]; lv[1][i] = t1;
	     }
	/* If either limit displayed a negative discriminant flagged by
	lv[i][0] == VACANT, the range is from the lower to the upper limit
	at the other position. Of course, if the other position also displays
	a negative discriminant, skip this (h,k) altogether. */
	if (lv[1][0] == VACANT)
	  {
		if (lv[0][0] != VACANT) pdolpred(pwp, lv[0][0],lv[0][1]);
	  }
	else {
		/* If the upper l value at the first omega extremum is
		less than the lower l value at the second extremum,
		the range is from the lower value at the first extremum
		to the upper value at the second extremum. */
		if (lv[0][1] <= lv[1][0]) pdolpred(pwp, lv[0][0],lv[1][1]);
		else
		/* ...otherwise, there are two ranges--from the lower limit
		at one extremum to the lower limit at the other, and from the
		upper limit at one extremum to the upper at the other. */
		  {
			pdolpred(pwp, lv[0][0],lv[1][0]);
			if (lv[0][1] <= lv[1][1])
				pdolpred(pwp, lv[0][1],lv[1][1]);
			else	pdolpred(pwp, lv[1][1],lv[0][1]);
		  }
	  }
}

int dphicmp(PREDREF *pp0, PREDREF *pp1)
/* comparison routine for sorting purely on phi */
{
	return (pp0->pphi < pp1->pphi ? -1 : 1);
}

int prpar(PREDWORK *pwp)
{ /*	"prpar": Initialize variables and arrays and
	provide initial displays for the "xgpredict" routine.
	It returns EOK if successful, various error values if not. */
	register int	i, j;
	int		result;
	double		axmax, cpsp, cpsq, detext[4], rmax, rmin;
	double		spsq, xext, rxmn, rxmx, rymn, rymx, smax;
	REFPARS		*xp;

	pwp->po_nchecked = pwp->po_totfound =
		pwp->po_totchecked = pwp->po_nfound = 0;
	zeromem((void *)&(pwp->po_nresult[0]), NRESTYPES * sizeof (int));
	if (NULL == pwp->po_xp)
	  {
		if (NULL == (pwp->po_xp = (REFPARS *)callop(1,
		 sizeof (REFPARS), "pwp->po_xp in prpar"))) return EALLOC;
	  }
	xp = pwp->po_xp;
	if (EOK != (i = refparin(xp, &result))) return i;
	makeq21(xp);
	/* get the splining information. As of Jan 04 we assume
	 that we can trust the detector dimensions found in the spline. */
	if (NULL == (pwp->po_spl = gspline(5, &result))) return result;
	xp->xdetwid = (pwp->po_spl)->v_detw; /* reaffirm detector dimensions */
	xp->xdetht = (pwp->po_spl)->v_deth;
	if (NULL == activmask((int)(xp->xdetwid), (int)(xp->xdetht), 5))
		return (errno == ENOMEM) ? EALLOC : EREADERR;
	if (verror(!inv3(xp->xareal, pwp->po_orec), "invert unit-cell matrix"))
		return ESINGULAR;
	for (i = 3; i--; ) for (j = 3; j--; ) pwp->po_orec[i][j] /= xp->xlambda;
	if (verror(!ucell(xp->xabc, xp->xang, xp->xast, xp->xanst, xp->xisys),
	 "compute reciprocal space unit cell")) return EBADVALUE;
	/* set up exclusion code */
	pwp->po_excluc = exclucode(xp->xispgp, xp->xisys);
	/* Determine inherent (detector-position-dependent) resolution limits*/
	rxmx = (pwp->po_spl)->v_x0cm +
	 (pwp->po_spl)->v_xrcm * ((float)(pwp->po_spl)->v_ncol - 1) - xp->xx0;
	rxmn = (pwp->po_spl)->v_x0cm - xp->xx0;
	if (rxmn > rxmx)
	  {
		xext = rxmn;	rxmn = rxmx;	rxmx = xext;
	  }
	rxmn -= CMSLOP;	rxmx += CMSLOP;
	rymn = (pwp->po_spl)->v_y0cm + (pwp->po_spl)->v_yrcm *
	 ((float)(pwp->po_spl)->v_nrow - 1) - xp->xy0 - CMSLOP;
	rymx = (pwp->po_spl)->v_y0cm - xp->xy0 + CMSLOP;
	xext = (fabs(rxmx) > fabs(rxmn) ? fabs(rxmx) : fabs(rxmn));
	if ((pwp->po_smasq = pdetec_limit(detext, pwp->po_spl, xp)) < DFUZZ)
		pwp->po_smasq = 2.;
	pwp->po_minxcm = detext[0];	pwp->po_minycm = detext[1];
	pwp->po_maxxcm = detext[2];	pwp->po_maxycm = detext[3];
	smax = sqrt(pwp->po_smasq);
	/* default upper resolution limit is based on detector extrema;
	 lower limit is based on cell dimensions */
	xp->xresmax = xp->xlambda / smax;
	if (pwp->po_dmaxspec > xp->xresmax) xp->xresmax = pwp->po_dmaxspec;
	axmax = xp->xabc[0];
	if (xp->xabc[1] > axmax) axmax = xp->xabc[1];
	if (xp->xabc[2] > axmax) axmax = xp->xabc[2];
	xp->xresmin = axmax * 1.1;
	if ((pwp->po_dminspec > DFUZZ) && (pwp->po_dminspec < xp->xresmin))
		xp->xresmin = pwp->po_dminspec;
	/* set up some other resolution-dependent values used in prediction */
	pwp->po_smasq = xp->xlambda / xp->xresmax;
	pwp->po_smasq *= pwp->po_smasq;
	pwp->po_smisq = xp->xlambda / xp->xresmin;
	pwp->po_smisq *= pwp->po_smisq;
	rmax = sqrt(rymx * rymx + xp->xpltd2);
	pwp->po_szmax = rymx / rmax + ZSLOP;
	rmin = sqrt(rymn * rymn + xp->xpltd2);
	pwp->po_szmin = rymn / rmin - ZSLOP;
	/* if both polfrac and polnormrot are zero, we treat that
	 as a sign that we haven't initialized those values and
	 get them from the environment or from a default. */
	if ((0 == pwp->po_polfrac) && (0 == pwp->po_polnormrot))
		(void)envtopol(&(pwp->po_polfrac), &(pwp->po_polnormrot));
	/* parameters used in the polarization calculations */
	cpsq = cos(pwp->po_polnormrot * RADC);	cpsq *= cpsq;
	spsq = 1. - cpsq;	cpsp = sqrt(cpsq * spsq);
	pwp->po_y2co = pwp->po_polfrac + (1. - 2. * pwp->po_polfrac) * spsq;
	pwp->po_z2co = pwp->po_polfrac + (1. - 2. * pwp->po_polfrac) * cpsq;
	pwp->po_yzco = 2. * (1. - 2. * pwp->po_polfrac) * cpsp;
	/* reserve memory for one "range" worth of predicted reflections */
	if (NULL == (AVOID *)(pwp->po_prp0 = (PREDREF *)callop(REFLSPERRANGE,
	 sizeof (PREDREF), "pwp->po_prp0 in prpar"))) return EALLOC;
	pwp->po_tl0 = NOW;	return EOK;
}

static void report_prange(FILE *fp, char cha[],
 short hmi, short hma, PREDWORK *pwp)
{ /* This reports the results of the prediction of one range to stream fp */
	fprintf(fp,"%7.1f%7.1f%8c%7c%8c%5d%5d%7d%10d%9d\n",
		(pwp->po_mumin + pwp->po_rangecen) / (pwp->po_xp)->xfrmsize,
		(pwp->po_mumax + pwp->po_rangecen) / (pwp->po_xp)->xfrmsize,
		(sichar)(cha[0]), (sichar)(cha[1]), (sichar)(cha[2]),
		(int)hmi, (int)hma, (int)(pwp->po_nchecked),
		(int)pwp->po_nfound, (int)(pwp->po_tl1 - pwp->po_tl0));
}

int prloop(PREDWORK *pwp, int prverbose)
{ /*	This routine does the outer two loops involved in prediction.
	The limits on the indices are computed by a formalism
	developed by George Reeke for rotation films, in which the
	Ewald sphere and the limiting (resolution) sphere are used
	in determining maximum and minimum values of h, k, and l.
	The looping is not necessarily in hkl order; the order is
	chosen to optimize the efficiency of the prediction itself. */
	register int	klow, khigh, i, j, ik;
	short		hmin, hmax, kmin, kmax, i1, i2, ip1;
	short		hv[2][2], kv[2][2];
	char		ch[3], cx, cy, cz;
	double		cosom, f0, f1, f2, omega, s2thm, sinom;
	double		cosza, htan, sinza, sqthm, sthm, ss, t1, t2;
	double		cd[6], e[5], qp[3], tk[2][5];
	float		pi[3][3], ploc[3][3];
	REFPARS		*xp;

	/* first rotate the reciprocal-space matrix to the center of the
	scanning range. This is done mostly to assure proper choice of
	slowest-varying axis in the index-ordering operation which follows. */
	pwp->po_rangecen = (pwp->po_mumax + pwp->po_mumin) / 2.;
	ploc[2][2] = 1.;	ploc[1][1] = ploc[0][0] = cos(pwp->po_rangecen);
	ploc[0][1] = sin(pwp->po_rangecen);	ploc[1][0] = -ploc[0][1];
	ploc[0][2] = ploc[1][2] = ploc[2][0] = ploc[2][1] = 0.;
	xp = pwp->po_xp;
	mm3(xp->xrec, ploc, pwp->po_orec);
	pwp->po_mumax -= pwp->po_rangecen;
	pwp->po_mumin -= pwp->po_rangecen;
	/* Compute the globals used in computing the limits of h,k, and l.
	First, set up the indexing order. The fastest-varying index,
	called `L', is the one associated with the reciprocal-space axis
	closest to the Z (rotation) axis. The slowest-varying (`H') is
	the one closer to the X (mainbeam) axis between the remaining two.
	The intermediate axis (`K') is the one left over. */
	pwp->po_ih[2] = 2;
	t1 = fabs(xp->xrec[2][2]) / xp->xast[2];
	for (i = 0; i < 2; i++)
	   {
		t2 = fabs(xp->xrec[2][i]) / xp->xast[i];
		if (t2 > t1)
		  {
			t1 = t2;	pwp->po_ih[2] = i;
		  }
	   }
	pwp->po_ih[0] = (pwp->po_ih[2] + 1) % 3;
	pwp->po_ih[1] = 3 - pwp->po_ih[0] - pwp->po_ih[2];
	t2 = fabs(xp->xrec[0][pwp->po_ih[0]]) / xp->xast[pwp->po_ih[0]];
	t1 = fabs(xp->xrec[0][pwp->po_ih[1]]) / xp->xast[pwp->po_ih[1]];
	if (t1 > t2)
	  {
		pwp->po_ih[0] = pwp->po_ih[1];
		pwp->po_ih[1] = 3 - pwp->po_ih[0] - pwp->po_ih[2];
	  }
	for (i = 0; i < 3; i++)	/* define the printable indices */
		ch[i] = (pwp->po_ih[i] == 0 ? 'h' :
			(pwp->po_ih[i] == 1 ? 'k' : 'l'));
	cz = ch[2];	cy = ch[1];	cx = ch[0];
	/* The following batch of constants is used in computing the
	Ewald-sphere limits of h,k and l. They are multiples and
	combinations of the reciprocal-space unit cell parameters. */ 
	for (i = 0; i < 3; i++)
	   {
		ip1 = i + 1;	i1 = ip1 % 3;	i2 = 3 - i - i1;
		ss = 0.5 * xp->xast[pwp->po_ih[i]];
		pwp->po_ps[ip1] = ss * ss;
		pwp->po_ps[i+4] = 0.5 * xp->xast[pwp->po_ih[i1]] *
		 xp->xast[pwp->po_ih[i2]] *
			cos(xp->xanst[pwp->po_ih[i]]);
	   }
	pwp->po_rhomax = 1. / (2. * xp->xresmax);
	pwp->po_rhomax *= pwp->po_rhomax; /* max (sin(theta)/lambda) squared */
	e[1] = pwp->po_ps[4] * pwp->po_ps[4] -
		4. * pwp->po_ps[3] * pwp->po_ps[2];
	e[2] = pwp->po_ps[4] * pwp->po_ps[5] -
		2. * pwp->po_ps[3] * pwp->po_ps[6];
	e[3] = pwp->po_ps[5] * pwp->po_ps[5] -
		4. * pwp->po_ps[3] * pwp->po_ps[1];
	e[4] = 4. * pwp->po_ps[3] * pwp->po_rhomax;
	f0 = e[2] * e[2] - e[1] * e[3];	f1 = e[1] * e[4]; f2 = -e[2] / e[1];
	/* 0.25 * fastest-varying recip. axis length: */
	pwp->po_al = pwp->po_ps[3];
	/* Get the permuted reciprocal-space matrix `pwp->po_p' and some
	related quantities at the beginning and the end of the data
	collection range. Use these to compute the limits of h and some
	constants used in determining the limits in k and l. */
		/* h,k,l slop constant */
	pwp->po_hslop = (pwp->po_mumax - pwp->po_mumin) / SLOPCONST;
	sthm = xp->xlambda / (2. * xp->xresmax);
	sqthm = sthm * sthm;
	if (verror((sqthm < 1.),"use a resolution limit as high as that"))
		return -1;
	s2thm = 2. * sthm * sqrt(1. - sqthm);
	for (i = 0; i < 2; i++)
	   {
		omega = (i ? (pwp->po_mumax + SLOP) : (pwp->po_mumin - SLOP));
		cosom = cos(omega);	sinom = sin(omega);
		for (j = 0; j < 3; j++)	/* get the permuted matrix */
		   {
			ik = pwp->po_ih[j]; /* indexing translation */
			pwp->po_p[i][0][j] = xp->xlambda *
			 (cosom * xp->xrec[0][ik] + sinom * xp->xrec[1][ik]);
			pwp->po_p[i][1][j] = xp->xlambda *
			 (-sinom * xp->xrec[0][ik] + cosom * xp->xrec[1][ik]);
			pwp->po_p[i][2][j] = xp->xlambda * xp->xrec[2][ik];
		   }
		cd[0] = pwp->po_p[i][0][0]*pwp->po_p[i][1][2] -
			pwp->po_p[i][1][0]*pwp->po_p[i][0][2];
		cd[1] = pwp->po_p[i][0][0]*pwp->po_p[i][2][2] -
			pwp->po_p[i][2][0]*pwp->po_p[i][0][2];
		cd[2] = pwp->po_p[i][1][0]*pwp->po_p[i][2][2] -
			pwp->po_p[i][2][0]*pwp->po_p[i][1][2];
		cd[3] = pwp->po_p[i][0][1]*pwp->po_p[i][1][2] -
			pwp->po_p[i][1][1]*pwp->po_p[i][0][2];
		cd[4] = pwp->po_p[i][0][1]*pwp->po_p[i][2][2] -
			pwp->po_p[i][2][1]*pwp->po_p[i][0][2];
		cd[5] = pwp->po_p[i][1][1]*pwp->po_p[i][2][2] -
			pwp->po_p[i][2][1]*pwp->po_p[i][1][2];
		tk[i][0] = -(cd[3] * pwp->po_p[i][1][2] +
			cd[4] * pwp->po_p[i][2][2]);
		tk[i][1] = cd[0] * cd[3] + cd[1] * cd[4] + cd[2] * cd[5];
		tk[i][2] = -(pwp->po_p[i][0][2]) * (pwp->po_p[i][0][2]);
		tk[i][3] = -2.*(cd[0]*pwp->po_p[i][1][2] +
			cd[1]*pwp->po_p[i][2][2]);
		tk[i][4] = cd[0]*cd[0] + cd[1]*cd[1] + cd[2]*cd[2];
		qp[i] = cd[3] * cd[3] + cd[4] * cd[4] + cd[5] * cd[5];
		if (verror(!inv3(pwp->po_p[i], pi),"invert unit cell matrix"))
			return -2;
		t1 = pi[0][0] * 2. * sqthm;
		t2 = sqrt(pi[0][1] * pi[0][1] + pi[0][2] * pi[0][2]) *
			s2thm;
		hv[i][0] = iroup(t1 - t2 - pwp->po_hslop);
		hv[i][1] = irodown(t1 + t2 + pwp->po_hslop);
		if ((cosza = fabs(cd[5]) / sqrt(qp[0])) > 1.)
		  {
			fprintf(stdout,
 "limit %d: singularity in tangency calculation: cosza = %12.7f\n",
				i,cosza);
			fprintf(stdout,
		 "%c limits = %d,%d calculated from resolution only\n",
				(sichar)cx, hv[i][0], hv[i][1]);
		  }
		else if ((sinza = sqrt(1. - cosza * cosza)) < s2thm)
		  {
			htan = xp->xabc[pwp->po_ih[0]] * (1. - cosza) /
			 xp->xlambda;
			if (pwp->po_p[i][0][0] > 0.)
				hv[i][0] = -irodown(htan + pwp->po_hslop);
			else hv[i][1] = irodown(htan + pwp->po_hslop);
		  }
	   }
	hmin = min2(hv[0][0],hv[1][0]);	/* LESS restrictive lower lim */
	hmax = max2(hv[0][1],hv[1][1]); /* LESS restrictive upper lim */
	if (verror((fabs(qp[0] - qp[1]) < FUZZ),
	 "predict data: qp[0] not equal to qp[1]")) return -3;
	pwp->po_prp = pwp->po_prp0; /* set prediction pointer to base value */
	pwp->po_nchecked = 0;	/* clear counter of # of checked refls */
	/* Loop through the possible h values. Get k limits for each h.
	Loop through the possible k values for this h and get l limits
	for each (h,k). Loop through the possible l values, attempting
	to predict each reflection as it comes up. */
	for (pwp->po_ph[0] = hmin; pwp->po_ph[0] <= hmax; (pwp->po_ph[0])++)
	   {	/* Get the resolution-sphere limits of k */
		pwp->po_hsq = pwp->po_ph[0] * pwp->po_ph[0];
		t2 = pwp->po_hsq * f0 - f1; /* discriminant of res.sphere */
		if (t2 <= 0.) continue; /* neg.discrim.: skip this h */
		t2 = sqrt(t2) / e[1];
		t1 = pwp->po_ph[0] * f2;
		kmin = iroup(t1 + t2 - pwp->po_hslop);
		kmax = irodown(t1 - t2 + pwp->po_hslop);
		/* Get the Ewald-sphere limits of k */
		for (i = 0; i < 2; i++)	/* examine start&end of run */
		   {
			qp[1] = tk[i][1] * pwp->po_ph[0] + tk[i][0];
			qp[2] = tk[i][4] * pwp->po_hsq +
				tk[i][3]* pwp->po_ph[0] + tk[i][2];
			t2 = qp[1] * qp[1] - qp[0] * qp[2];
			if (t2 < 0.) /* no Ewald-sphere intersection */
			  {
				kv[i][0] = VACANT;	kv[i][1] = -VACANT; }
			else
			  {
				t2 = sqrt(t2) / qp[0]; t1 = -qp[1] / qp[0];
				kv[i][0] = iroup(t1 - t2 - pwp->po_hslop);
				kv[i][1] = irodown(t1 + t2 + pwp->po_hslop);
			  }
		   }
		/* the Ewald-sphere limits are taken to be the
		 LESS restrictive values for kmax and kmin. */
		klow = min2(kv[0][0], kv[1][0]);
		khigh = max2(kv[0][1], kv[1][1]);
		/* If no overlap exists between the Ewald-sphere k limiting
		range and the resolution-sphere limiting range, skip h. */
		if (klow > kmax || khigh < kmin) continue;
		/* final k limits are the MORE restrictive of
		 the Ewald limits and the resolution limits. */
		if (klow > kmin) kmin = klow; if (khigh < kmax) kmax = khigh;
		/* Loop through possible k values, getting l limits for each
		(h,k) and predicting for each (h,k,l) being checked. */
		for (pwp->po_ph[1] = kmin; /* loop thru l values & do it */
		 pwp->po_ph[1] <= kmax; (pwp->po_ph[1])++) pllimits(pwp);
			/* count found reflections */
		pwp->po_nfound = pwp->po_prp - pwp->po_prp0;
	   }
	pwp->po_tl1 = NOW; /* record time now */
	pwp->po_nfound = pwp->po_prp - pwp->po_prp0;	/* count found refls */
	pwp->po_totfound += pwp->po_nfound;
	if (prverbose) report_prange(stdout, ch, hmin, hmax, pwp);
	pwp->po_totchecked += pwp->po_nchecked;	return pwp->po_nfound;
}

/* end of xgpredict.c */
