/* xtal.c -- XGEN utilities for dealing with crystallographic realities

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	27 Jun 2007: fix horrible bug in vecmat: czt=sqrt(1-cb^2-cg*^2).
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	 5 Sep 2006: rename dummy variable in exclucode; it's now "sys"
	24 May 2006: rot9rt moved out
	20 Nov 2005: inline argument declarations and strict prototyping;
		use xgen.h to cover declarations
	30 Jan 2005: check_sys(): default spacegroups are simpler
	28 Aug 2004: cosmetic changes
	16 Aug 2004: ztos() moved here from xgenmisc.c
	 4 Apr 2004: ucell(): average cell dimensions in cubic, tetragonal,
		rhombohedral, and hexagonal rather than just taking one as ok
	30 Jan 2004: angles input to ztos() are now in degrees, not radians
	21 Jan 2004: license message inserted
	18 Sep 2003: remove some unused variables
	16 Aug 2003: rename 2nd arg to keephkl() from exclusion to excluco
	 5 Jan 2003: reorganized source-file header
	10 Oct 2002: incorporate centrosymmetric spacegroups into exclucode()
	25 Mar 2002: removed MSI attachments
	 8 Nov 1996: minor changes for X-GEN checkin at MSI
	31 May 1995: "register" -> "register int" throughout;
		cast hpack correctly
	 2 Mar 1995: trivial formatting
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN
	16 Mar 1994: polymask added; bugfixes in polymath, xtal, calinv, etc.
	 7 Dec 1993: initial X-GEN version
	
   Purpose: manipulating crystallographically significant quantities
	(unit cell parameters, Euler angles, etc.) used in data processing.
   Cross references: ../xgrefine/refutil.c

*************************************************************************/
#include	"xgen.h"	/* for a long time we didn't do this */
#define		TWOOVSQ3	1.1547005384	/* 2 over sqrt(3) */
#define		TWOPIOV3	2.0943951023	/* 2 * pi over 3 */
#define		PIOV3		1.0471975511	/* pi over three */

short exclucode(short spgrp, short sys)
{ /* this returns an exclusion code for the spacegroup's centering */
	switch (spgrp)
	      {
		case	5:	case	8:	case	9:	case	12:
		case	15:	case	20:	case	21: /* C-centered */
		case 35: case 36: case 37: case 38: case 39: case 40:
		case 41: case 63: case 64: case 65: case 66: case 67: case 68:
			return	1;
		case	22: 	case	42:	case	43:	case	69:
		case	70:	case	196:	case	209:	/* F-centered*/
		case 216: case 219: case 225: case 226: case 227: case 228:
			return 2;
		case	23:	case	24:	case	44:	case	45:
		case	46: case 71: case 72: case 73: case 74:
		case 79: case 80: case 82: case 87: case 88: case 97: case 98:
		case 107: case 108: case 109: case 110: /* I-centered */
 		case 197: case 199: case 204: case 206: case 211: case 214:
		case 217: case 220: case 229: case 230:
			return	3;
		case	146:	case	155:			/* R-centered*/
		 	return	(sys == 6 ? 4 : 0);
		case	210:	return	5;			/* messy F */
		default:	return	0;			/* primitive*/
	     }
}

void rot9r(float rea[3][3], float *rfi)
{ /* this rotates a real-space matrix analogously to rot9 */
	register int	i, j;
	double		c0, c1, c2, s0, s1, s2;
	double		a[3][3], drin[3][3], rhold[3][3];
  
	c0 = cos(rfi[0]);	c1 = cos(rfi[1]);	c2 = cos(rfi[2]);
	s0 = sin(rfi[0]);	s1 = sin(rfi[1]);	s2 = sin(rfi[2]);
	a[0][0]=  s0*c1*s2+c0*c2; a[0][1]= -s0*c1*c2+c0*s2; a[0][2]=  s0*s1;
	a[1][0]= -c0*c1*s2+s0*c2; a[1][1]=  c0*c1*c2+s0*s2; a[1][2]= -c0*s1;
 	a[2][0]= -s1*s2;	  a[2][1]=  s1*c2;	    a[2][2]=  c1;
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++) drin[i][j] = rea[i][j];
	dmm3(rhold, drin, a);
	for (i = 3; i--; ) for (j = 3; j--; ) rea[i][j] = rhold[i][j];
}

int ucell(float abc[], float abg[], float ast[], float angst[], short isys)
{ /*	This calculates a*,b*,c*, alpha*, beta*, and gamma*
	for a crystal if a,b,c,alpha,beta, and gamma are given.
	It also forces restrictions upon a,b,c,alpha,beta, and gamma
	if the space group system "isys" requires any. */
	register int	k, k0, k1, k2;
	double		cq[3], sq[3], star, vol;
/* convert abg to radians for the purposes of this routine */
	for (k = 0; k < 3; k++) abg[k] *= RADC;
/* branch on the space-group system number */
	switch (isys) {
		case 1:	/* triclinic: P1, international tables # 1 */
			for (k = 0; k < 3; k++)
			   {
				cq[k] = cos(abg[k]); sq[k] = sin(abg[k]);
			   }
			star = 1 + 2 * cq[0] * cq[1] * cq[2] - cq[0] * cq[0] -
				cq[1] * cq[1] - cq[2] * cq[2];
			if (star <= 0.) return -1;
			vol = abc[0] * abc[1] * abc[2] * sqrt(star);
			for (k0 = 0; k0 < 3; k0++)
			   {
				k1 = (k0 == 2 ? 0 : k0 + 1); k2 = 3 - k0 - k1;
				ast[k0] = abc[k1] * abc[k2] * sq[k0] / vol;
				star = (cq[k1] * cq[k2] - cq[k0])/
					(sq[k1] * sq[k2]);
				if (star < -1 || star > 1) return(-1);
				angst[k0] = acos(star);
			   }
			break;
		case 2: /* monoclinic: int'l tables #2-15 */
			abg[0] = abg[2] = angst[0] = angst[2] = PIO2;
			if (fabs(sin(abg[1])) < DFUZZ ||
			 fabs(sin(abg[1])) > 1) return -1;
			angst[1] = PI - abg[1]; star = 1. / sin(abg[1]);
			for (k0 = 0; k0 < 3; k0++)
			   if (abc[k0] < FUZZ) return -1;
			ast[0] = star / abc[0]; ast[1] = 1./abc[1];
			ast[2] = star / abc[2]; break;
		case 5:	/* cubic: int'l tables # 195-230 */
			abc[2] = abc[1] = abc[0] =
				(abc[0] + abc[1] + abc[2]) / 3.;
			for (k = 0; k < 3; k++)
			   {
				abg[k] = angst[k] = PIO2; ast[k] = 1./abc[k];
			   }
			break;
		case 4: /* tetragonal: int'l tables # 75-142 */
			abc[1] = abc[0] = 0.5 * (abc[0] + abc[1]);/*fall thru*/
		case 3: /* orthorhombic: int'l tables # 16-74 */
			for (k = 0; k < 3; k++)
			   {
				abg[k] = angst[k] = PIO2; ast[k] = 1/abc[k];
			   }
			break;
		case 6: /* hexagonal & trigonal: # 143-194 except case 7 */
			abg[0] = abg[1] = angst[0] = angst[1] = PIO2;
			abc[1] = abc[0] = 0.5 * (abc[0] + abc[1]);
			abg[2] = TWOPIOV3; angst[2] = PIOV3;
			ast[1] = ast[0] = TWOOVSQ3/abc[0]; ast[2] = 1/abc[2];
			break;
		case 7:	/* rhombohedral indexed as such: #146,148,155,160,
				161,166,167 */
			abc[2] = abc[1] = abc[0] =
				(abc[0] + abc[1] + abc[2]) / 3.;
			if (abg[2] < PIO2) /*average angles if abg[2] acute */
			  {
				if ((abg[1] < PIO2) && (abg[2] < PIO2))
				  abg[2] = (abg[0] + abg[1] + abg[2]) / 3.;
			  }
			else { /* average angles if abg[2] is obtuse */
				if ((abg[1] > PIO2) && (abg[2] > PIO2))
				  abg[2] = (abg[0] + abg[1] + abg[2]) / 3.;
			     }
			abg[0] = abg[1] = abg[2]; /* force 'em equal */
			cq[2] = cos(abg[2]);
			star = 1 + cq[2] * cq[2] * (2. * cq[2] - 3.);
			if (star <= 0) return(-1);
			ast[2] = ast[1] = ast[0] =
				sin(abg[2]) / (abc[0] * sqrt(star));
			if ((star = 1/(1+cq[2]) - 1) > 1) return(-1);
			angst[0] = angst[1] = angst[2] = acos(star); break;
		default:
			return(-1);	/* garbage space group */
		       }
	for (k = 3; k--; ) /* convert abg back to degrees, clean it up */
	   {
		abg[k] /= RADC;
		if (fabs(abg[k] - 120.) < FUZZ) abg[k] = 120.;
		if (fabs(abg[k] - 90.) < FUZZ) abg[k] = 90.;
	   }
	return(0);
}

int vecmat(float *len, float *ang, float mat[3][3])
{ /*	This gets the actual components of the real-space unit-cell
	vectors a,b,c along X,Y, and Z as a 3x3 matrix `mat'.
	It does so by applying appropriate geometrical partitionings
	on the lengths and angles.
	3 Feb 1994: correct the sign of mat[2][1]!!! */
	register int	i;
	double		lang[3], ca, cb, cg, sg, cst, czt;

	for (i = 3; i--; ) lang[i] = ang[i] * RADC;
	ca = cos(lang[0]);	cb = cos(lang[1]);
	cg = cos(lang[2]);	sg = sin(lang[2]);
	if (fabs(sg) < DFUZZ) return 1;
	cst = (ca - cb * cg) / sg;
	if ((czt = 1. - cb * cb - cst * cst) < DFUZZ) return 1;
	czt = sqrt(czt);
/* get the matrix of params, "mat".  */
	mat[0][0] = len[0];	mat[0][1] = mat[0][2] = mat[1][2] = 0.;
	mat[1][0] = len[1] * cg;	mat[1][1] = len[1] * sg;
	mat[2][0] = len[2] * cb;	mat[2][1] = len[2] * cst;
	mat[2][2] = len[2] * czt;	return 0;
}

int keephkl(short *lh, short excluco)
{ /* Checks for centering (I, C, F, HR) absences */
	if ((lh[0] == 0) && (lh[1] == 0) && (lh[2] == 0)) return 0;
	switch (excluco)
	      {
		case	0:	return 1;
		case	1:	return	!((lh[0] + lh[1]) % 2);
		case	2:	return (!((lh[0] + lh[1]) % 2) &&
					!((lh[0] + lh[2]) % 2));
		case	3:	return	(!((lh[0] + lh[1] + lh[2]) % 2));
		case	4:	return	!((lh[1] + lh[2] - lh[0]) % 3);
		case	5:	if ((lh[0] == 0) && (lh[1] == 0))
					return !(lh[2] % 4);
				else return (!((lh[0] + lh[1]) % 2) &&
					!((lh[0] + lh[2]) % 2));
		default:	return	1;
	     }
}

xglong hpack(short *lh)
{ /*	This encodes a 3-element hkl vector and returns a single
	32-bit number, defined as follows:
	 value = 16 * ((l + 256) + 512 * ((k+256) + 512 * (h+256))) */
	return (((lh[0]+256)<< 22) | ((lh[1]+256)<< 13) | ((lh[2]+256)<< 4));
}

void hunpack(xglong phkl, short uhkl[])
{ /*	This reads in h,k,l as a four-byte packed number and returns a
	vector of three "shorts" with the individual hkl values in it.
	The packing is as follows:
		pindx= 16*((l+256)+512*((k+256) + 512 * (h+256))) */
	uhkl[2] = ((phkl >> 4) & 0777) - 0400;
	uhkl[1] = ((phkl >> 13) & 0777) - 0400;
	uhkl[0] = ((phkl >> 22) & 0777) - 0400;
	return;
}

int check_sys(short oldspaceg, short lsys)
{ /*	This returns `oldspaceg' unless that spacegroup number is
	inconsistent with the system number given. */
	switch (lsys)
	      {
		case	1:	/* triclinic */
			return (1 <= oldspaceg && oldspaceg <= 2) ?
				oldspaceg : 1;	/* default: P1 */
		case	2:	/* monoclinic */
			return (3 <= oldspaceg && oldspaceg <= 15) ?
				oldspaceg : 3;	/* default: P2 */
		case	3:	/* orthorhombic */
			return (16 <= oldspaceg && oldspaceg <= 74) ?
				oldspaceg : 16;	/* default: P222 */
		case	4:	/* tetragonal */
			return (75 <= oldspaceg && oldspaceg <= 142) ?
				oldspaceg : 75;	/* default: P4 */
		case	5:	/* cubic */
			return (195 <= oldspaceg && oldspaceg <= 230) ?
				oldspaceg : 195; /* default: P23 */
		case	6:	/* hexagonal or trigonal */
			return (143 <= oldspaceg && oldspaceg <= 194) ?
				oldspaceg : 143; /* default: P3 */
		case	7:	/* rhombohedral */
			return ((143 <= oldspaceg && oldspaceg <= 167)
			 || (243 <= oldspaceg && oldspaceg <= 267)) ?
				oldspaceg : 146; /* default: R3 */
		case	0:	default:	return oldspaceg;
	     }
}

void ztos(float *rusr, float zs[3][3])
{ /*	This constructs the matrix required to rotate from the goniostat's
	(0,0,0) position to the (omega_start,chi_start,phi_start) position.
	The transpose of this matrix gives the rotation from (start) to (000).
	Thus once this matrix `zs' exists, we can use it either way using
	mm3 or mm3t.  The position rusr[] is in DEGREES, not radians. */
	double		qr0, qr1, qr2, c0, c1, c2, s0, s1, s2;

	qr0 = rusr[0] * RADC;	qr1 = rusr[1] * RADC;	qr2 = rusr[2] * RADC;
	c0 = cos(qr0);	c1 = cos(qr1);	c2 = cos(qr2);
	s0 = sin(qr0);	s1 = sin(qr1);	s2 = sin(qr2);
	zs[0][0]=  s0*c1*s2+c0*c2;	zs[0][1]= -c0*c1*s2+s0*c2;
 	zs[0][2]= -s1*s2;
	zs[1][0]= -s0*c1*c2+c0*s2;	zs[1][1]=  c0*c1*c2+s0*s2;
	zs[1][2]=  s1*c2;
	zs[2][0]=  s0*s1;	zs[2][1]= -c0*s1;	zs[2][2]=  c1;
}

/* end of xtal.c */
