/* indexing.c -- routines enabling auto-indexing of diffraction data

	Copyright (c) 2003, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	26 Dec 2006: move legald to refutil.c
	24 Nov 2006: switch to Bravais-lattice-based penalties for low-symmetry
		unit cell types in best_solution; rp2vecs calls updated
	 4 Nov 2006: enable avoidance of icerings during indexing using `legald'
	23 May 2006: vec{utils,tors}.c reorganized into {prospects,indexing}.c
	10 Apr 2006: cleaner calc of maximum cell edge in diffvectors
	31 Dec 2005: don't qsort if fewer than 2 elements
	15 Oct 2005: add typedef VECWORK to eliminate static globals
	25 Aug 2005: optimize_xy(): recalc beam center more adroitly
	19 Jan 2005: rfp->rfr_autoind set after auto-indexing
	28 Aug 2004: routines used only here are declared as static; cosmetics
	11 Aug 2004: several additional routines moved around
	25 Jun 2004: more tweaking of optimize_xy() stuff
	15 Jun 2004: optimize_xy() created: it does the systematic (X,Y) search
	  more carefully--if there's a solution far from the center and the
	  solution close to the center is almost as good, pick the latter;
	  report_sec() and #defines created that support this functionality
	11 Aug 2004: several additional routines moved to vecutils.c
	 3 Jun 2004: eliminate changed-detector-param concept
	26 Jan 2004: some static globals converted to local variables
	21 Jan 2004: license message inserted
	 7 Jun 2004: restrict_limits() limits maxoversq and rfp->rfr_amax[]
	 3 Jan 2004: REFINERUN and REFPARS structure elements in function calls
		replace remaining global variables; function calls updated
	26 Nov 2003: replace global variables with elements of REFINERUN struct
	19 Sep 2003: initialize dpend and tcos in squish(),
	 and nmerged in avevec(), and bsofar and bmsofar in first_vector();
	 and oorlow and oorhigh in diffvectors()
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	24 Aug 2003: eliminate gu0 argument to summ_gresults()
	22 Aug 2003: replace diffv0,dvlast with dvlim[2];
		correct recently introduced bugs in function calls;
		have indhkl return the number of solutions it finds.
	22 Aug 2003: modify routines that access DIFFVECTOR structures
		according to calling sequences defined here
	18 Aug 2003: some routines split out from vectors.c into vecutils.c
	14 Aug 2003: make diffv0 and dvlast static globals here
		and change the way memory is allocated for diffv0;
		add indhklclean() to make that happen
	 5 Jan 2003: reorganized source-file header
	15 Nov 2002: extern declarations made here rather than in refiext.h
	29 Mar 2002: merged with refihkl.c; many defines moved here from there
	25 Mar 2002: removed MSI attachments
	15 Jan 1999: limit the search to MAXDVEC difference vectors.
	15 Sep 1997: in diffvectors(), the difference vectors are
		generated in one step rather than two
	12 Sep 1997: in squish(), apart from fixing a bug in the median
		calc, we're also going to insist that the
		difference vectors be similar in length as well as direction
	28 Jul 1997: major reorganization of diffvectors()
		means that we call getprospects and prospects3 from
		within diffvectors and diffm3 rather than later
	 6 Jul 1995: "extern int dvncmp" declared on AXP port
	31 May 1995: "register" -> "register int" throughout
	20 May 1995: add variable angle between auto-indexing basis vectors
	 2 Mar 1995: eliminate calls to "gripe"
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	23 Sep 1994: set dpsquish->dnsq to dpmid->dnsq, not dp->dnsq
	 2 Sep 1994: change how reference and representative
		difference vectors are chosen
	15 Aug 1994: in diffmatrix(), we start the search almost
		at the beginning, so we don't miss anything.
	 5 May 1994: source-file header cleaned up,
		 names changed from POLYDET to XGEN; vsort.h included for VMS
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version
	
   Purpose: choosing diffraction vectors, making difference vectors,
	and using those to do auto-indexing of bright spots in a pattern.
	Candidate difference vectors are used to generate trial orientations.
	Once the difference vectors have been checked for reasonableness
	they're fed onto other routines in this file for further processing.
	Vectors are coalesced into groupings before final processing both
	to save time and reduce the likelihood of including garbaged data.
   Cross references: prospects.c
*************************************************************************/
#define		NEED_TIME	1
#include	"refidefs.h"
#define		CLOSE_RADIUSSQ	(0.005)	/* (X,Y) optimum accepted right away */
#define		COSSQ_PARALLEL	0.96	/* min cossq for vectors equal */
#define		DISTANCE_RATIO	0.5	/* ratio of distances squared */
#define		DISTINCT_MIN	(0.003)	/* appropriate dist from local min */
#define		FIRSTFEW	50	/* first few refls in scatt vec list*/
#define		MAXDVEC		4950	/* (100*(100-1))/2 */
#define		MAXUSABLE	300	/* max # of usable troids */
#define		MAX_NORMRATIO	0.12	/* comparison for norms */
#define		MAX_TO_SHOW	250	/* max difference vectors displayed */
#define		MAX_TO_STDOUT	25	/* max # solutions printed on stdout */
#define		MINSCATTVEC	(0.1e-5)	/* min scatt vector normsq*/
#define		NDUP		3	/* min # duplications for reference*/
#define		NDUPMAX		20	/* dups > this are irrelevant */
#define		NMDPEN		((double)(NMPENLO - NMPENHI))
#define		NMPENHI		0.7	/* penalty-function val at max(nmer)*/
#define		NMPENLO		1.0	/* penalty-function val at min(nmer)*/
#define		OPT_RATIO	1.3	/* ratio of best to next-best guess */
#define		MAX_PENALTY	0.02	/* largest Bravais penalty */
#define		SLOP_RESOLUTION	2.	/* leeway factor, cell upper limit */
#define		STLDPEN		(STLPENHI - STLPENLO)
#define		STLPENHI	1.0	/* penalty-function val at min(stl)*/
#define		STLPENLO	0.65	/* penalty-function val at max(stl)*/

double cossqv(float *s0, float *s1)
{ /*	This computes the cosine-squared of the angle between
	the vectors s0 and s1.  The cosine of the angle between
	two vectors, one of which is zero, is defined to be 1.	*/
	register int	i;
	register float	*p0,*p1;
	double		adotb, normasq, normbsq;

	adotb = normasq = normbsq = 0.;
	for (i = 3, p0 = s0, p1 = s1; i--; p0++, p1++)
	   {
		adotb += *p0 * *p1;
		normasq += *p0 * *p0;	normbsq += *p1 * *p1;
	   }
	if (normasq < DFUZZ || normbsq < DFUZZ) return 1;
	return (adotb / normasq) * (adotb / normbsq);
}

void crossproduct(float *s0, float *s1, float *norm)
{ /*	This returns a vector `norm' which is the cross product of s0 and s1 */
	norm[0] = s0[1] * s1[2] - s0[2] * s1[1];
	norm[1] = s0[2] * s1[0] - s0[0] * s1[2];
	norm[2] = s0[0] * s1[1] - s0[1] * s1[0];
	return;
}

static int dvncmp(DIFFVECTOR *dv0, DIFFVECTOR *dv1)
/* compare routine: sort diff vectors on norm^2 */
{
	return (dv0->dnsq < dv1->dnsq ? -1 : 1);
}

static void sum_consider(FILE *fp, double pcerr[], double basa, REFINERUN *rfp)
{ /*	Summarize which solutions are being considered */
	char	*lolit;
	REFPARS	*xp;
	
	lolit = (fp == stdout) ? "\r\n" : "\n";
	fprintf(fp,
	 "Percent Error limits in lengths, angles = %7.3f, %7.3f%s",
		pcerr[0], pcerr[1], lolit);
	fprintf(fp,
	 "Solutions will be considered within the following limits:%s",
		lolit);
	fprintf(fp,
	 "                a       b       c   alpha    beta   gamma%s",
		lolit);
	fprintf(fp,"    min: %8.3f%8.3f%8.3f%8.3f%8.3f%8.3f%s",
	 rfp->rfr_amin[0], rfp->rfr_amin[1], rfp->rfr_amin[2],
	 rfp->rfr_angmin[0], rfp->rfr_angmin[1], rfp->rfr_angmin[2], lolit);
	fprintf(fp,"    max: %8.3f%8.3f%8.3f%8.3f%8.3f%8.3f%s",
	 rfp->rfr_amax[0], rfp->rfr_amax[1], rfp->rfr_amax[2],
	 rfp->rfr_angmax[0], rfp->rfr_angmax[1], rfp->rfr_angmax[2], lolit);
	xp = rfp->rfr_xp;
	fprintf(fp,"current: %8.3f%8.3f%8.3f%8.3f%8.3f%8.3f%s",
		xp->xabc[0], xp->xabc[1], xp->xabc[2],
		xp->xang[0], xp->xang[1], xp->xang[2], lolit);
        fprintf(fp,
         "Any two basis vectors must be at least %7.2f degrees apart%s",
                basa, lolit);
}

static double anglevol(double c0, double c1, double c2)
{ /* returns sqrt(1-cossq(c0)-cossq(c1)-cossq(c2)+2cos(c0)*cos(c1)*cos(c2)) */
	double	vol;
	vol = 1. - c0 * c0 - c1 * c1 - c2 * c2 + 2. * c0 * c1 * c2;
	return (vol > 0. ? sqrt(vol) : 0.);
}

void guesstoxp(GUESS *gp, REFPARS *xp)
{ /*	This stores values from a GUESS structure into a refinement struct */
	register int	i;

	for (i = 3; i--; ) /* unit cell lengths and angles */
	   {
		xp->xabc[i] = gp->gabc[i];	xp->xang[i] = gp->gang[i];
	   }
	xp->xdelom = gp->gomega[0];	xp->xdelchi = gp->gomega[1];
	xp->xdelphi = gp->gomega[2];
}

void xptoguess(REFPARS *xp, GUESS *gp)
{ /*	This stores values into a GUESS structure from a refinement struct */
	register int	i;

	for (i = 3; i--; ) /* unit cell lengths and angles */
	   {
		gp->gabc[i] = xp->xabc[i];	gp->gang[i] = xp->xang[i];
	   }
	gp->gomega[0] = xp->xdelom;	gp->gomega[1] = xp->xdelchi;
	gp->gomega[2] = xp->xdelphi;
}

int impose_solution(int desired_soln, REFINERUN *rfp)
{ /* This imposes an auto-indexing solution on the user. */

	if (desired_soln < 0) return -1;
	guesstoxp(rfp->rfr_gu0 + desired_soln, rfp->rfr_xp);
	fprintf(rfp->rfr_fpl, "Solution %d imposed\n",desired_soln);
	sum_curparam(rfp->rfr_fpl, "imposed autoindex", rfp);
	return desired_soln;
}

static double guessvol(GUESS *gup)
{ /*	This returns the unit cell volume associated with the guess `gup' */
	double	c0, c1, c2;

	c0 = cos(gup->gang[0] * RADC);	c1 = cos(gup->gang[1] * RADC);
	c2 = cos(gup->gang[2] * RADC);	c0 = sqrt(anglevol(c0, c1, c2));
	return (gup->gabc[0] * c0) * (gup->gabc[1] * gup->gabc[2]);
}

static double bravaispenalty(float *rabc, float *ran)
{ /* This determines the Bravais lattices associated with the current
   unit cell. It then returns a penalty value of up to 0.011 if the
   best lattice doesn't have a high symmetry. */
	int			i, j, nlatt;
	float			lomat[3][3];
	double			bestfom, lfom, dlo[3][3];
	static LATTICESCORE	*la0, *lcap;

	/* Begin this process by creating a local unitcell matrix */
	if (EOK != vecmat(rabc, ran, lomat)) return MAX_PENALTY;
	/* determine the Bravais characters for this matrix */
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++) dlo[i][j] = lomat[i][j];
	la0 = allbravais(dlo, &nlatt);
	if (nlatt < 16)
	  { /* should get at least 16 good answers */
		(void)allbravais(NULL, &nlatt);	return MAX_PENALTY;
	  }
	bestfom = 10000.;
	for (i = 0, lcap = la0; i < 45; i++, lcap++)
	   if ((lcap->lc_choice >= 0) &&
		((lfom = lcap->lc_fom) < bestfom)) bestfom = lfom;
	(void)allbravais(NULL, &nlatt); /* free the lattice memory */
	return (bestfom < MAX_PENALTY) ? bestfom : MAX_PENALTY;
}

static void volume_penalty(GUESS *guz, int ngues)
{ /*	This goes through the list of guesses and finds the minimum unit cell
	volume among the ones located there. Any guess with a larger volume
	is penalized fractionally by an amount equal to the ratio of the
	guess's volume to the minimum one.
	For the moment we're defining volume as a * b * c;
	maybe later we'll switch to the true unit cell volume, which is
	also a function of the cell angles. */
	int	j;
	GUESS	*gup;
	double	c0, c1, c2, star, vol, volmi;

	volmi = 1.e30;
	for (j = 0, gup = guz; j < ngues; gup++, j++)
	   {
		vol = gup->gabc[0] * gup->gabc[1] * gup->gabc[2];
		c0 = gup->gang[0] * RADC;	c0 = cos(c0);
		c1 = gup->gang[1] * RADC;	c1 = cos(c1);
		c2 = gup->gang[2] * RADC;	c2 = cos(c2);
		star = 1. + 2. * c0 * c1 * c2 - c0 * c0 - c1 * c1 - c2 * c2;
		if ((star > 0.) && ((vol = vol * star) < volmi)) volmi = vol;
	   }
	if ((volmi < DFUZZ) || (volmi > 1.e29)) return;
	for (j = 0, gup = guz; j < ngues; gup++, j++)
	   {
		vol = gup->gabc[0] * gup->gabc[1] * gup->gabc[2];
		c0 = gup->gang[0] * RADC;	c0 = cos(c0);
		c1 = gup->gang[1] * RADC;	c1 = cos(c1);
		c2 = gup->gang[2] * RADC;	c2 = cos(c2);
		star = 1. + 2. * c0 * c1 * c2 - c0 * c0 - c1 * c1 - c2 * c2;
		if (star > 0.)
			gup->gfom *= (1. + 0.5 * (star * vol / volmi - 1.));
		else gup->gfom = 31000.;
	   }
}

static int best_solution(int nsolus, REFINERUN *rfp)
{ /*	This returns the offset associated with the best solution.
	We're doing this a somewhat ad hoc way. */
	int		solno;
	double		bestre;
	register GUESS	*gup, *bestgu;

	/* define the figure of merit to be the integerness residual,
	 augmented slightly by a term associated with how far the
	 two angles closest to 90 are from 90. */
	for (gup = rfp->rfr_gu0, solno = 0; solno < nsolus; solno++, gup++)
		gup->gfom = gup->gresid + bravaispenalty(gup->gabc, gup->gang);
	/* insert an additional penalty associated with unit-cell volume */
	volume_penalty(rfp->rfr_gu0, nsolus);
	/* return the best figure of merit */
	bestre = 1.e30;
	printf(" Soln. Cell Vol Residual Fig.of Merit\n");
	fprintf(rfp->rfr_fpl, " Soln. Cell Vol Residual Fig.of Merit\n");
	for (bestgu = gup = rfp->rfr_gu0, solno = 0;
	 solno < nsolus; solno++, gup++)
	   {
		printf("%5d %8.1lf %8.4f %8.4f\n",
		 solno, guessvol(gup), gup->gresid, gup->gfom);
		fprintf(rfp->rfr_fpl, "%5d %8.1lf %8.4f %8.4f\n",
		 solno, guessvol(gup), gup->gresid, gup->gfom);
		if (gup->gfom < bestre)
		  {	
			bestgu = gup;	bestre = gup->gfom;
		  }
	   }
	return bestgu - rfp->rfr_gu0;
}

static int avevec(DIFFVECTOR *dp0, float *vec, int *iclosest, VECWORK *vwp)
{ /*	This goes through the list of difference vectors starting at dp0 and
	averages all those which are close to it in length and direction.
	The resulting averaged difference vector is returned in vec[0]-[2],
	and its norm is returned in vec[3].  The function returns the number
	of difference vectors actually merged into the result. */
	register int		i, nmerged, maxelt, nearest;
	register DIFFVECTOR	*dp;
	double			max1, maxe, maxdnsq, mindi, dive, sdivsq;

	nmerged = 1;
	/* initialize the sum with the specified starting vector */
	for (i = 3; i--; ) vec[i] = dp0->ss[i];
	maxdnsq = dp0->dnsq + vwp->vw_minnsq;
	/* find the largest absolute value element in dp0 */
	maxelt = 0;	maxe = fabs(dp0->ss[0]);
	if ((max1 = fabs(dp0->ss[1])) > maxe)
	  {
		maxelt = 1;	maxe = max1;
	  }
	if (fabs(dp0->ss[2]) > maxe) maxelt = 2;
	/* go through the neighboring scattering vector differences,
		averaging in the ones which come from the same reflection.
		We have to get the sign right in each case. */
	for (dp = dp0+1; dp <= vwp->vw_dvlim[1] && dp->dnsq < maxdnsq; dp++)
	   if (cossqv(dp->ss, dp0->ss) > COSSQ_PARALLEL)
	     {
		nmerged++;
		if ((dp->ss[maxelt] * dp0->ss[maxelt]) > 0)
		  for (i = 3; i--; ) vec[i] += dp->ss[i];
		else for (i = 3; i--; ) vec[i] -= dp->ss[i];
	     }
	/* now divide by the number of merged vectors to get the mean
		and compute the norm (squared) of the resulting vector */
	for (i = 3, vec[3] = 0.; i--; )
	   {
		vec[i] /= (double)nmerged;	vec[3] += vec[i] * vec[i];
	   }
	/* go back and determine which of the close ones is closest */
	for (dp = dp0, mindi = 100. * vec[3], nearest = 0;
	 dp <= vwp->vw_dvlim[1] && dp->dnsq < maxdnsq; dp++)
	   {
		if ((dp->ss[maxelt] * vec[maxelt]) > 0)
		  {
			for (sdivsq = 0., i = 3; i--; )
			   {
				dive = dp->ss[i] - vec[i];
				sdivsq += dive * dive;
			   }
		  }
		else {
			for (sdivsq = 0., i = 3; i--; )
			   {
				dive = dp->ss[i] + vec[i];
				sdivsq += dive * dive;
			   }
		     }
		if ((dp == dp0) || (sdivsq < mindi))
		  {
			mindi = sdivsq;	nearest = dp - dp0;
		  }
	   }
	*iclosest = nearest;
	return nmerged;
}

static int first_vector(float *uref, VECWORK *vwp)
{ /*	This does a more ambitious approach to finding three noncoplanar
	vectors. It returns the number of quasi-equivalent vectors that are
	associated with the first basis vector, or zero if an error occurs. */

	int		i, numvecs, nmerged, j, nmersofar, ngroups, ngtmdups;
	double		mincossq, casq, casqp, fnp;
	float		sumvec[3], asofar[4], awork[4];
	DIFFVECTOR	*dvp, *startvec, *midvec, *bsofar, *bmsofar;

	/* first pick a vector near the middle that has several examples.*/
	midvec = vwp->vw_dvlim[0] + (vwp->vw_dvlim[1] - vwp->vw_dvlim[0]) / 2;
	startvec = midvec - 3;	if (startvec < vwp->vw_dvlim[0])
		startvec = vwp->vw_dvlim[0];
	nmerged = 0;
	for (dvp = startvec; dvp < vwp->vw_dvlim[1] - 1; dvp++)
	   {
		if ((nmerged = avevec(dvp, asofar, &j, vwp)) >= NDUP) break;
		else dvp += nmerged - 1;
	   }
	dvp += j;
	if ((nmerged < NDUP) || (dvp >= vwp->vw_dvlim[1])) return 0;
	for (i = 3; i--; ) asofar[i] = dvp->ss[i];
	for (dvp = vwp->vw_dvlim[0], ngroups = ngtmdups= 0;
	 dvp < vwp->vw_dvlim[1]; dvp++)
	   {
		if ((nmerged = avevec(dvp, awork, &j,vwp)) > NDUP) ngtmdups++;
		dvp += (nmerged - 1);	ngroups++;
	   }
	numvecs = vwp->vw_dvlim[1] - vwp->vw_dvlim[0];
	fprintf(vwp->vw_fpl,
	 "\t%6d diff. vectors; %6d groups, %6d with > %1d elements\n",
		numvecs, ngroups, ngtmdups, NDUP);
	fprintf(vwp->vw_fpl, "\t\t\treference vector: %12.4e %12.4e %12.4e\n",
	 asofar[0], asofar[1], asofar[2]);
	/* sum up all the vectors in a sign-sensitive way:
	 any vector that is more than 90 degrees away from that one gets
	 flipped around. */
	for (i = 3; i--; ) sumvec[i] = 0.;
	for (dvp = vwp->vw_dvlim[0], numvecs = 0;
	 dvp < vwp->vw_dvlim[1]; dvp++, numvecs++)
	   {
		if (dotprod(dvp->ss, asofar, 3) < 0.)
		  {
			for (i = 0; i < 3; i++) sumvec[i] -= dvp->ss[i];
		  }
		else for (i = 0; i < 3; i++) sumvec[i] += dvp->ss[i];
	   }
	if (numvecs <= 0) return 0;
	for (casq = i = 0; i < 3; i++) casq += sumvec[i] * sumvec[i];
	if (casq < DFUZZ) return 0;	casq = sqrt(casq);
	for (i = 0; i < 3; i++) sumvec[i] /= casq;
	fprintf(vwp->vw_fpl, "\t\t\t  averaged vector: %12.4e %12.4e %12.4e\n",
	 sumvec[0], sumvec[1], sumvec[2]);
	/* we now look for a vector that is far from this one in
	 an angular sense, i.e. that has a minimum value of cos(a,b) squared,
	 where a is this averaged vector and b is the one we're testing.
	 We want to find a small difference vector if possible, so we use
	 a penalty function under which small diff vectors are encouraged.
	 We also want to encourage data groupings that have lots of vectors
	 in them. */
	mincossq = 10.;	nmersofar = 0;
	vwp->vw_spend = (vwp->vw_dvlim[1] - 1)->dnsq -
		(vwp->vw_dvlim[0] + 1)->dnsq;
	if (vwp->vw_spend < FUZZ) return 0;
	vwp->vw_spena = STLDPEN / vwp->vw_spend;
	vwp->vw_spenb = (STLPENLO * ((vwp->vw_dvlim[1])-1)->dnsq -
		STLPENHI * ((vwp->vw_dvlim[0]) + 1)->dnsq) / vwp->vw_spend;
	vwp->vw_nmpena = NMDPEN /
		(1. / ((double)NDUP) - 1. / ((double)NDUPMAX));
	vwp->vw_nmpenb = NMPENLO - vwp->vw_nmpena / ((double)NDUP);
	/* the penalty function is STLPENLO for the smallest
	 diff vector, STLPENHI for the largest. */
	bsofar = bmsofar = NULL;
	for (dvp = vwp->vw_dvlim[0]; dvp < vwp->vw_dvlim[1]; dvp++)
	   {
		if ((nmerged = avevec(dvp, awork, &j, vwp)) >= NDUP)
		  {
			dvp += j;	casq = cossqv(dvp->ss, sumvec);
			fnp = (nmerged < NDUPMAX) ?
				1. / (double)nmerged : 1. / (double)NDUPMAX;
			casqp = casq * (fnp * vwp->vw_nmpena + vwp->vw_nmpenb) *
			 (dvp->dnsq * vwp->vw_spena + vwp->vw_spenb);
			if (mincossq > casqp)
			  {
				mincossq = casqp;	nmersofar = nmerged;
				bsofar = dvp - j;	bmsofar = dvp;
				for (i = 0; i < 3; i++) asofar[i] =awork[i];
			  }
			dvp += (nmerged - j) - 1;
		  }
		else	dvp += (nmerged - 1);
	   }
	if ((mincossq > 1.) || (NULL == bsofar) || (NULL == bmsofar)) return 0;
	/* the value actually returned is the median among similar values */
	for (uref[3] = i = 0; i < 3; i++)
	   {
		uref[i] = bmsofar->ss[i];	uref[3] += (uref[i] * uref[i]);
	   }
	casq = cossqv(bmsofar->ss, sumvec);
	fprintf(vwp->vw_fpl, " Difference vectors used in the calculations:\n");
	fprintf(vwp->vw_fpl, "  dvp[0]  dvp[1]  dvp[2] nmer  ave[0]  ave[1]");
	fprintf(vwp->vw_fpl,
	 "  ave[2]  med[0]  med[1]  med[2]   casq0   casq1   maxxp\n");
	fprintf(vwp->vw_fpl,
	 " %7.4f %7.4f %7.4f %4d %7.4f %7.4f %7.4f %7.4f %7.4f %7.4f %7.5f\n",
	 bsofar->ss[0],bsofar->ss[1],bsofar->ss[2], nmersofar,
	 asofar[0], asofar[1], asofar[2],
	 bmsofar->ss[0],bmsofar->ss[1],bmsofar->ss[2], casq);
	return nmersofar;
}

static int diffmatrix(float *spacings, double minbasis,
	REFINERUN *rfp, VECWORK *vwp)
{ /*	This goes through the partially- (or fully-)constructed list of
	difference vectors and attempts to find three very noncoplanar vectors,
	each of which occurs at least NDUP times within the list.
	It returns 1 if and only if it succeeds in finding said matrix.
	Once it's found the matrix, it puts it in a crystallographically
	canonical form, inverts it, and prints some useful facts about it. */

	int		nmerged;
	int		i, j, ip1, ip2, nmersofar;
	double		casq0, casq1, casq2, casqp, dotp;
	double		fnp, maxcsq, maxxp, mincsq;
	float		vref[3][4], asofar[4], awork[4];
	REFPARS		*xp;
	DIFFVECTOR	*dvp, *bsofar, *bmsofar;

	/* get the first vector by finding one that's as far from the
	 sign-modified average of all the vectors in the basis set. */
	if (NDUP > (nmerged = first_vector(vref[0], vwp))) return 0;
	/* start search almost at the beginning,
	 on the chance that we might otherwise miss something. */
	maxcsq = cos(minbasis * RADC);	maxcsq *= maxcsq;
	mincsq = 10.;	bsofar = bmsofar = NULL;
	for (dvp = vwp->vw_dvlim[0], nmersofar = 0;
	 dvp <= (vwp->vw_dvlim[1] - 1); dvp++)
	   {
		casq0 = cossqv(dvp->ss, vref[0]);
		nmerged = avevec(dvp, awork, &j, vwp);
		fnp = (nmerged < NDUPMAX) ?
			1. / (double)nmerged : 1. / (double)NDUPMAX;
		casqp = casq0 * (fnp * vwp->vw_nmpena + vwp->vw_nmpenb) *
		 (dvp->dnsq * vwp->vw_spena + vwp->vw_spenb);
		if ((casq0 < maxcsq) && (nmerged >= NDUP) &&
		 (casqp < mincsq))
		  {
			mincsq = casqp;	nmersofar = nmerged;
			bsofar = dvp;		bmsofar = dvp + j;
			for (i = 0; i < 3; i++) asofar[i] = awork[i];
		  }
		dvp += (nmerged - 1);
	   }
	mincsq = cossqv(bmsofar->ss, vref[0]);
	if ((nmersofar < NDUP) || (bsofar == NULL)) return 0;
	fprintf(rfp->rfr_fpl, " %7.4f %7.4f %7.4f %4d %7.4f %7.4f %7.4f",
	 bsofar->ss[0], bsofar->ss[1], bsofar->ss[2], nmersofar,
	 asofar[0], asofar[1], asofar[2]);
	fprintf(rfp->rfr_fpl, " %7.4f %7.4f %7.4f %7.4f\n",
	 bmsofar->ss[0], bmsofar->ss[1], bmsofar->ss[2], mincsq);
	for (i = 3, vref[1][3] = 0.; i--; )
	   {
		vref[1][i] = bmsofar->ss[i];
		vref[1][3] += (bmsofar->ss[i] * bmsofar->ss[i]);
	   }
	/* compute the cross-product of the 0th & 1st diff vectors.
	The 2nd must be fairly close to this in direction.
	We'll just pick the vector closest to this cross product in
	our usual fudged sense. We want to encourage big values of
	cosine now, so we divide by the penalty function rather than mult.*/
	crossproduct(vref[0], vref[1], spacings + 5);
	maxxp = 0.;
	for (dvp = vwp->vw_dvlim[0] + 1, nmersofar = 0;
	 dvp <= vwp->vw_dvlim[1]; dvp++)
	   {
		casq0 = cossqv(dvp->ss, vref[0]);
		casq1 = cossqv(dvp->ss, vref[1]);
		casq2 = cossqv(dvp->ss, spacings + 5);
		nmerged = avevec(dvp, awork, &j, vwp);
		fnp = (nmerged < NDUPMAX) ?
			1. / (double)nmerged : 1. / (double)NDUPMAX;
		casqp = casq2 / ((fnp * vwp->vw_nmpena + vwp->vw_nmpenb) *
		 (dvp->dnsq * vwp->vw_spena + vwp->vw_spenb));
		if ((nmerged >= NDUP) && (casq0 < maxcsq) &&
		 (casq1 < maxcsq) && (casqp > maxxp))
		  {
			maxxp = casqp;	nmersofar = nmerged;
			bsofar = dvp;	bmsofar = dvp + j;
			for (i = 0; i < 3; i++) asofar[i] = awork[i];
		  }
		dvp += (nmerged - 1);
	   }
	if (nmersofar < NDUP) return 0;
	fprintf(rfp->rfr_fpl,
	 " %7.4f %7.4f %7.4f %4d %7.4f %7.4f %7.4f",
	 bsofar->ss[0], bsofar->ss[1], bsofar->ss[2], nmersofar,
	 asofar[0], asofar[1], asofar[2]);
	casq0 = cossqv(bmsofar->ss, vref[0]);
	casq1 = cossqv(bmsofar->ss, vref[1]);
	casq2 = cossqv(bmsofar->ss, spacings + 5);
	fprintf(rfp->rfr_fpl, " %7.4f %7.4f %7.4f %7.4f %7.4f %7.4f\n",
	 bmsofar->ss[0], bmsofar->ss[1], bmsofar->ss[2], casq0, casq1, casq2);
	xeflush(rfp->rfr_fpl);
	for (i = 3, vref[2][3] = 0.; i--; )
	   {
		vref[2][i] = bmsofar->ss[i];
		vref[2][3] += (bmsofar->ss[i] * bmsofar->ss[i]);
	   }
	/* now construct the 3x3 matrix containing these averaged reference
	scattering-vector differences */
	for (i = 3; i--; ) for (j = 3; j--; ) rfp->rfr_sref[i][j] = vref[j][i];
/* make sure the determinant of the matrix comprising the three
	scattering vector differences is positive.  If it's negative,
	we flip the signs of all the elements in a column of "sref".
	That's legal... if (s) is a legal scattering vector, so is (-s) */
	if ((deter(rfp->rfr_sref) * 0.1e6) < -DFUZZ)
	  for (i = 3; i--; )
	     {
		rfp->rfr_sref[i][0] = -(rfp->rfr_sref[i][0]);
		vref[0][i] = -vref[0][i];
	     }
	/* now get the inverse of this matrix to use in hkl checks */
	if (inv3(rfp->rfr_sref, rfp->rfr_sinv)) return 0;
	/* get a normalized version of `vref' for convenience */
	xp = rfp->rfr_xp;
	for (i = 3; i--; )
	   {
		vref[i][3] = sqrt(vref[i][3]);
		for (j = 3; j--; ) vref[i][j] /= vref[i][3];
		spacings[i] = xp->xlambda / vref[i][3];
	   }
	for (i = 0; i < 3; i++)
	   {
		ip1 = (i + 1) % 3;	ip2 = 3 - i - ip1;
		for (j = 3, dotp = 0; j--; )
		   dotp += vref[ip1][j] * vref[ip2][j];
		spacings[i + 5] = acos(dotp) / RADC;
	   }
	return 1;
}

static void restrict_limits(REFINERUN *rfp, VECWORK *vwp)
{ /*	Adjust the value of vw_maxovsq and the amax array based on
	 the magnitude of the smallest observed difference vector. */
	int	i;
	double	newmax, rrmax, rmaxs;
	REFPARS	*xp;

	if (FUZZ < (newmax = (vwp->vw_dvlim[0])->dnsq))
	  {
		xp = rfp->rfr_xp;
		newmax = SLOP_RESOLUTION * xp->xlambda * xp->xlambda / newmax;
		rrmax = sqrt(newmax);
		if (newmax < vwp->vw_maxovsq)
		  {
			fprintf(vwp->vw_fpl,
 "Based on observed diff. vectors, |cell|max = %8.2f, not %8.2f\n",
				rrmax, sqrt(vwp->vw_maxovsq));
			fprintf(stdout,
 "Based on observed diff. vectors, |cell|max = %8.2f, not %8.2f\r\n",
				rrmax, sqrt(vwp->vw_maxovsq));
			vwp->vw_maxovsq = newmax;	rmaxs = rrmax * 1.1;
			for (i = 0; i < 3; i++)
			   {
				if (rfp->rfr_amax[i] > rmaxs)
					rfp->rfr_amax[i] = rmaxs;
			   }
		  }
	  }
}

static int simdels(DIFFVECTOR *dpa, DIFFVECTOR *dpb)
{ /* This returns 1 iff the two arguments point to
	difference vectors that are close together in length and in angle */
	double	norma, normb;

	norma = dpa->dnsq;	normb = dpb->dnsq;
	if (fabs(norma - normb) > (norma + normb) * MAX_NORMRATIO) return 0;
	return (cossqv(dpa->ss, dpb->ss) > COSSQ_PARALLEL);
}

static void squish(REFPARS *xp, VECWORK *vwp)
{ /*	This averages all the difference vectors which are essentially
	identical and outputs only one vector per averaged group.
	But it doesn't put out the average itself,
	since that appears not to work. */
	register int		i,nmerged,maxelt;
	register DIFFVECTOR	*dp,*dp0,*dpsquish;
	DIFFVECTOR		*dpmid, *dpend;
	double			dive, maxdnsq, maxlen, mincos;
	double			mindi, minlen, sdivsq, tcos, vec[3];

	dpend = vwp->vw_dvlim[1];	/* initial upper limit of list */
	/* set all the difference vectors' sh[2] value to a flag value */
	for (dp = vwp->vw_dvlim[0]; dp <= vwp->vw_dvlim[1]; dp++)
		dp->sh[2] = -3.;
	for (dp0 = dpsquish = vwp->vw_dvlim[0]; dp0 <= vwp->vw_dvlim[1]; dp0++)
	   if (dp0->dwt >= 0)
	     {
		nmerged = 1;
		/* initialize the sum with the specified starting vector */
		for (i = 3; i--; ) vec[i] = dp0->ss[i];
		maxdnsq = dp0->dnsq + vwp->vw_minnsq;
		/* find the largest absolute value element in dp0 */
		maxelt = (fabs(dp0->ss[0]) > fabs(dp0->ss[1]) ?
			(fabs(dp0->ss[0]) > fabs(dp0->ss[2]) ? 0 : 2) :
			(fabs(dp0->ss[1]) > fabs(dp0->ss[2]) ? 1 : 2));
		/* go through the neighboring scattering vector differences,
		averaging in the ones which come from the same reflection.
		We have to get the sign right in each case. */
		for (dp = dp0+1; dp <= vwp->vw_dvlim[1] && dp->dnsq < maxdnsq;
		 dp++)
		   if ((dp->dwt >= 0) && simdels(dp, dp0))
		     {
			nmerged++;
			if ((dp->ss[maxelt] * dp0->ss[maxelt]) > 0)
			  for (i = 3; i--; ) vec[i] += dp->ss[i];
			else for (i = 3; i--; ) vec[i] -= dp->ss[i];
			dp->dwt = -1; /* signal: don't use this one again */
			dp->sh[2] = 0.;	/* signal: DO use it now */
			dpend = dp;	/*only set dpend if it's a good one*/
		     }
		if (nmerged < NDUP) dp0->dwt = -2; /* flag:ignore on 2nd pass*/
		else
		  { /* compute mean scatt vector. Find an
			an actual unit cell vector from the group which is
			close to this average and stick IT on the list. */
			for (i = 3; i--; ) vec[i] /= (double)nmerged;
			/* we require that the one we keep
			 be the one that's actually CLOSEST to this mean. */
			dp0->sh[2] = 0.;
			for (dpmid = dp = dp0, mindi = 1.e9; dp <= dpend; dp++)
			   if (dp->sh[2] == 0.)
			     {
				if ((dp->ss[maxelt] * vec[maxelt]) > 0)
				  {
					for (sdivsq = 0., i = 3; i--; )
					   {
						dive = dp->ss[i] - vec[i];
						sdivsq += dive * dive;
					   }
				  }
				else {
					for (sdivsq = 0., i = 3; i--; )
					   {
						dive = dp->ss[i] + vec[i];
						sdivsq += dive * dive;
					   }
				     }
				if ((dp == dp0) || (sdivsq < mindi))
				  {
					mindi = sdivsq;	dpmid = dp;
				  }
			     }
			if (mindi < 0.9e9)
			  { /* having found a nice representative element,
				 put the representative element on squished
				 list of difference vectors. Swap it for
				 whatever was already in that position and
				 set some flags accordingly */
				minlen = dp0->dnsq;	maxlen = dpend->dnsq;
				if (dpsquish != dpmid) swap((void *)dpsquish,
					 (void *)dpmid, sizeof (DIFFVECTOR));
				tcos = 1.;	/* a hedge against error */
				for (dp = dp0, mincos = 2.; dp < dpend; dp++)
				   if (dp->sh[2] == 0.)
				     {
					tcos = cossqv(dp->ss, dpsquish->ss);
					if (tcos < mincos) mincos = tcos;
					dp->dwt = -1.;
				     }
				dpsquish->ih[0] = xp->xlambda *
				 (1. / sqrt(minlen) - 1. / sqrt(maxlen));
				dpsquish->ih[1] = acos(sqrt(tcos)) / RADC;
				dpsquish->dwt = -nmerged;
				dpsquish->sh[2] = -3;
				dpsquish++;
			  }
			/* now reset the flags */
			for (dp = dp0; dp <= dpend; dp++) dp->sh[2] = -3.;
		  }
	   }
	vwp->vw_dvlim[1] = dpsquish - 1;
	return;
}

static void summ_sqvecs(FILE *fp, DIFFVECTOR *dvli[], int autov, REFINERUN *rfp)
{ /*	This summarizes the first few out of the current list of
	squished difference vectors to the stream fp. */
	char		*lolit;
	int		nshow;
	double		lolam, lonorm;
	DIFFVECTOR	*dvp, *dvld;

	lolit = (fp == stdout) ? "\r\n" : "\n";
	lolam = (rfp->rfr_xp)->xlambda;	/* we use this a few times here */
	fprintf(fp,
	 " %6d difference vectors with lengths from %8.2f to %8.2f A.%s",
	 (int)(rfp->rfr_ndiffvecs), lolam / sqrt(dvli[0]->dnsq),
	 lolam / sqrt((dvli[0] + (rfp->rfr_ndiffvecs-1))->dnsq), lolit);
	if ((fp == stdout) || (fp == stderr)) return;
	nshow = dvli[1] - dvli[0];
	if (autov)
	  {
		if (nshow > MAX_TO_SHOW)
		  {
			nshow = MAX_TO_SHOW;	dvld = dvli[0] + (nshow - 1);
			fprintf(fp,
			 "First %d difference vectors used in auto-indexing:\n",
				nshow);
		  }
		else {
			dvld = dvli[1];	fprintf(fp,
				"Difference vectors used in auto-indexing:\n");
		     }
		for (nshow = 0; nshow < 2; nshow++)
		   fprintf(fp,
	 "      dsx      dsy      dsz lam/|ds|  Drange maxangl #obs%s",
				(nshow) ? "\n" : "\t");
		for (dvp = dvli[0], nshow = 0; dvp <= dvld; dvp++, nshow++)
		   {
			lonorm = lolam / sqrt(dvp->dnsq); fprintf(fp,
				" %8.5f %8.5f %8.5f %8.2f %7.2f %7.2f %4d%s",
			 dvp->ss[0], dvp->ss[1], dvp->ss[2], lonorm, dvp->ih[0],
			 dvp->ih[1], (int)(-(dvp->dwt)),
			 (nshow % 2) ? "\n" : "\t");
		   }
		fprintf(fp, "\n");
		if (stderr != fp) xeflush(fp);
	  }
}

static int diffvectors(float *spac, double err_all, double mibasis,
	DIFFVECTOR **dvextr, REFINERUN *rfp, VECWORK *vwp)
{ /*	This generates a list of difference vectors from the list
	of scattering vectors.  It finds all of them first, sorts them
	by increasing sin theta / lambda, and merges difference vectors
	which are clearly parallel to one another and of the same length.
	This routine can be used in two different ways depending on which
	form of autoindexing we're doing. The routine figures out which
	version is intended based on the value of mibasis; if mibasis
	is negative, we're using the second method. If it's non-negative,
	we use the first method. The vectors resulting from the search
	(or the lowest MAXDIFFVECS of them) are put into the DIFFVEC
	structures.  Return values and meanings:
	FIRST METHOD:
	-1	failure to reserve memory for list of guesses
	0       successful return from `diffmatrix'
	1	failure in `diffmatrix'
	SECOND METHOD:
	 0	failure to find 3 noncoplanar difference vectors
	-1	failure to reserve memory for list of guesses
	<= -2	failure in diffm3.
	>= 1	success. Number returned is number of good solutions.
	As of Aug 03, the array *dvextr contains the pointer to the
	 first and last difference vectors available.
*/
	register int	i;
	int		ntused, oorlow, oorhigh;
	double		dvnormsq, deltai, lod[3];
	float		*flp;
	DIFFVECTOR	*dvp;
	REFPARS		*xp;
	REFREF		*rpr, *lrp;
	REFREF		*rpu0, *rplast, *rpmid, *rpg0s, *rpg0e;
#ifdef VMS
	static u_short	normsort[] = {1,
		DSC$K_DTYPE_F, SO_ASCENDING, 40, 4};
#endif

	ntused = (rfp->rfr_inbounds < MAXUSABLE ?
		rfp->rfr_inbounds : MAXUSABLE);
	rpu0 = rfp->rfr_rpp0 + rfp->rfr_firstok;
	rplast = rpu0 + (ntused - 1);
	rpmid = rpu0 + (ntused / 2);
	rpg0s = rpmid - FIRSTFEW /  2;	rpg0e = rpmid + FIRSTFEW / 2;
	vwp->vw_fpl = rfp->rfr_fpl;
	if (rpg0s < rfp->rfr_rpp0) rpg0s = rfp->rfr_rpp0;
	if (rpg0e > rplast) rpg0e = rplast;
	rfp->rfr_ndiffvecs = (ntused * (ntused - 1)) / 2;
	dvextr[0] = dvextr[1] = NULL;
	if (NULL == (vwp->vw_dvlim[0] = (DIFFVECTOR *)callop(
	 rfp->rfr_ndiffvecs + 100, sizeof (DIFFVECTOR),
	 "dvlim in diffvectors"))) return -1;
	dvextr[0] = vwp->vw_dvlim[0];	/*remember start & end of array */
	dvextr[1] = vwp->vw_dvlim[0] + rfp->rfr_ndiffvecs + 99;
	/* don't use unreliable (i.e., too small) difference vectors. */
	i = 0;	flp = rfp->rfr_amax;
	if (*(flp + 1) > *flp) i = 1;	if (*(flp + 2) > *(flp + i)) i = 2;
	vwp->vw_minnsq = *(flp + i);
	fprintf(rfp->rfr_fpl,
	 "largest unit cell edge under consideration = %12.5f\n",
	 vwp->vw_minnsq);
	xp = rfp->rfr_xp;
	vwp->vw_minnsq = (err_all > 0.9 ? MINSCATTVEC :
		xp->xlambda * xp->xlambda / (vwp->vw_minnsq * vwp->vw_minnsq));
	fprintf(rfp->rfr_fpl,
	 "minimum difference vector norm squared = %12.5e\n", vwp->vw_minnsq);
	if (vwp->vw_minnsq < FUZZ)
	  {
		vwp->vw_minnsq = MINSCATTVEC; fprintf(rfp->rfr_fpl,
		"... but that value is smaller than FUZZ.\n");
		fprintf(rfp->rfr_fpl,
		 "minimum difference vector norm squared = %12.5e\n",
		 vwp->vw_minnsq);
	  }
	/* don't use difference vectors > the biggest scatt vector */
	vwp->vw_maxnsq = xp->xlambda / rplast->dspac;
	vwp->vw_maxnsq *= vwp->vw_maxnsq;
	fprintf(rfp->rfr_fpl,
	 "maximum difference vector norm squared = %12.5e\n", vwp->vw_maxnsq);
	oorlow = oorhigh = 0;
	/* The technique developed years ago wherein the difference
	 vectors are generated in two steps is probably inappropriate to the
	 newer auto-indexing algorithms. So we do it in a simpler way. */
	if (mibasis > 0.)
	  { /* get a few diff vectors from near the middle of the list */
		for (lrp = rpg0s, dvp = vwp->vw_dvlim[0]; lrp < rpg0e; lrp++)
		   for (rpr = lrp + 1; rpr <= rpg0e; rpr++)
		      {
			for (i = 3, dvnormsq = 0; i--; ) /* get diff vector */
			   {
				dvp->ss[i] = deltai = lrp->rs[i] - rpr->rs[i];
				dvnormsq += deltai * deltai;
			   }
			/* keep this diff vector unless it's too small */
			if (vwp->vw_minnsq < dvnormsq &&
			 dvnormsq < vwp->vw_maxnsq) dvp++->dnsq = dvnormsq;
		      }
		rfp->rfr_rpp = lrp;
		rfp->rfr_ndiffvecs = ntused = dvp - vwp->vw_dvlim[0];
		/* sort this first batch by increasing norm squared */
		if (ntused > 1) qsort((void *)(vwp->vw_dvlim[0]),
		 (size_t)ntused, sizeof (DIFFVECTOR), (QPE)dvncmp);
		if (rfp->rfr_ndiffvecs > 100)
			dvp = vwp->vw_dvlim[0] + (rfp->rfr_ndiffvecs / 2);
		else	dvp = vwp->vw_dvlim[0] + (rfp->rfr_ndiffvecs - 1);
		vwp->vw_maxnsq = dvp->dnsq;
		/* we want user to know vwp->vw_minnsq and vwp->vw_maxnsq */
		spac[3] = vwp->vw_minnsq;	spac[4] = vwp->vw_maxnsq;
		/* now get the remaining difference vectors.  We only accept
		difference vectors which are smaller than the largest one we've
		gotten so far: that's the point of the initial group, eh wot?*/
		if ((rpg0s > rpu0) && (rpg0e < rplast))
		  {
			for (lrp = rpu0; lrp < rplast; lrp++)
			  for (rpr = lrp + 1; rpr <= rplast; rpr++)
			     {
				for (i = 3, dvnormsq = 0; i--; )
				   {
					lod[i] = deltai =
						lrp->rs[i] - rpr->rs[i];
					dvnormsq += deltai * deltai;
				   }
				if (vwp->vw_minnsq < dvnormsq &&
				 dvnormsq < vwp->vw_maxnsq)
				  {
					for (i = 3; i--; ) dvp->ss[i] = lod[i];
					dvp->dwt = 1.;	/* nominal weight */
				  	dvp++->dnsq = dvnormsq;
				  }
			     }
			rfp->rfr_rpp = lrp;
		  }
	  } /* end of "if (mibasis > 0.)" clause */
	else {
		for (lrp = rpu0, dvp = vwp->vw_dvlim[0]; lrp < rplast; lrp++)
		 if (legald(rfp, lrp))
		  for (rpr = lrp + 1; rpr <= rplast; rpr++)
		   if (legald(rfp, rpr))
		    {
			for (i = 3, dvnormsq = 0.; i--; )
			   {
				lod[i] = deltai = lrp->rs[i] - rpr->rs[i];
				dvnormsq += deltai * deltai;
			   }
			if (vwp->vw_minnsq < dvnormsq &&
			 dvnormsq < vwp->vw_maxnsq)
			  {
				for (i = 3; i--; ) dvp->ss[i] = lod[i];
				dvp->dwt = 1.;	/* nominal weight */
			  	dvp++->dnsq = dvnormsq;
			  }
			else if (vwp->vw_minnsq >= dvnormsq) oorlow++;
			else oorhigh++;
		    } /* close both "for (rp" loops */
		rfp->rfr_rpp = lrp;
	     } /* end of "else" */
	/* sort the resulting tank by increasing norm squared */
	fprintf(rfp->rfr_fpl,
	 "difference vectors disallowed because |s| < %8.5f: %6d\n",
		sqrt(vwp->vw_minnsq), oorlow);
	fprintf(rfp->rfr_fpl,
	 "difference vectors disallowed because |s| > %8.5f: %6d\n",
		sqrt(vwp->vw_maxnsq), oorhigh);
	rfp->rfr_ndiffvecs = dvp - vwp->vw_dvlim[0];
	if (rfp->rfr_ndiffvecs < 2) return 1; /* not enough diff vectors */
	qsort((void *)(vwp->vw_dvlim[0]), (size_t)(rfp->rfr_ndiffvecs),
	 sizeof (DIFFVECTOR), (QPE)dvncmp);
	restrict_limits(rfp, vwp);
	if (rfp->rfr_ndiffvecs > MAXDVEC) rfp->rfr_ndiffvecs = MAXDVEC;
	vwp->vw_dvlim[1] = vwp->vw_dvlim[0] + (rfp->rfr_ndiffvecs -1);
	dvextr[1] = vwp->vw_dvlim[1];
	/* Attempt to form a reference matrix.  If we succeed, we're ready
	to start going through the possible hkls that make said
	reference matrix up.  If not, we print a whiney message and quit. */
	if ((mibasis >= 0.) && (!diffmatrix(spac, mibasis, rfp,vwp))) return 1;
/* We've succeeded.  Squish the DIFFVECTORS list by averaging vectors which
	are essentially identical to their neighbors and outputting only the
	averages.  At least NDUP vectors must make up each of these. */
	fprintf(rfp->rfr_fpl, "prior to squishing:\n");
	summ_sqvecs(rfp->rfr_fpl, vwp->vw_dvlim, vwp->vw_autoverbose, rfp);
	squish(rfp->rfr_xp, vwp);
	summ_sqvecs(stdout, vwp->vw_dvlim, vwp->vw_autoverbose, rfp);
	summ_sqvecs(rfp->rfr_fpl, vwp->vw_dvlim, vwp->vw_autoverbose, rfp);
	/* current technique. Generate the real-space unit cell vectors
	 (ax,ay,az) and try them individually in groups of three.  */
	if (mibasis < 0.) return diffm3(spac, -mibasis, rfp, vwp);
	else return 0;
}

static void summ_aiset(FILE *fp, REFPARS *xp)
{ /* Summarize the initial setup of the autoindexing effort */
	char	*lolit;
	int	i;

	lolit = (fp == stdout) ? "\r\n" : "\n";
	fprintf(fp, "reduced cell used in indexing effort:%s", lolit);
	fprintf(fp,
 "       a       b       c   alpha    beta   gamma   omega     chi     phi%s",
		lolit);
	fprintf(fp, " %7.2f %7.2f %7.2f %7.2f %7.2f %7.2f %7.2f %7.2f %7.2f%s", 
		xp->xabc[0], xp->xabc[1], xp->xabc[2],
		xp->xang[0], xp->xang[1], xp->xang[2],
		xp->xdelom, xp->xdelchi, xp->xdelphi, lolit);
	for (i = 0; i < 3; i++) fprintf(fp," %12.6f %12.6f %12.6f%s",
		xp->xareal[i][0], xp->xareal[i][1], xp->xareal[i][2], lolit);
}

static int summ_gresults(FILE *fp, int ndecen, REFINERUN *rfp)
{ /*	This summarizes the first `ndecen' solutions found.
	It returns a code associated with the kind of success we had. */
	char	*lolit;
	GUESS	*gulast, *gup;

	lolit = (fp == stdout) ? "\r\n" : "\n";
	if (ndecen == -1)
	  {
		fprintf(fp,
		 "Cannot reserve memory for vector differences\n"); return -1;
	  }
	else if (ndecen == -7)
	  {
		fprintf(fp,
		 "No 3-D auto-indexing solutions found within tolerances\n");
		return -7;
	  }
	else if (ndecen == 0)
	  {
		fprintf(fp, "Not enough vector differences found\n");
		return -2;
	  }
	else if (ndecen < 0)
	  {
		fprintf(fp,
		 "Error %d finding autoindexing solutions\n", ndecen);
		return ndecen;
	  }
	gulast = rfp->rfr_gu0 + ndecen;
	fprintf(fp, " %6d decent %s found%s",
	 ndecen, (ndecen == 1) ? "solution" : "solutions", lolit);
	for (gup = rfp->rfr_gu0; gup < gulast; gup++)
	   if (gup->gresid < 2.)
	     {
		fprintf(fp,
 " %7.2f %7.2f %7.2f %7.2f %7.2f %7.2f %7.2f %7.2f %7.2f %8.5f%s",
		 gup->gabc[0], gup->gabc[1], gup->gabc[2],
		 gup->gang[0], gup->gang[1], gup->gang[2],
		 gup->gomega[0], gup->gomega[1], gup->gomega[2],
		 gup->gresid, lolit);
	   }
	if ((fp != stdout) && (fp != stderr)) (void)xeflush(fp);
	return 0;
}

static int indhklclean(int rev, VECWORK *vwp)
{ /*	This frees the memory for the `dvlim[0]' array before returning
	the value `rev'. */
	memzap((AVOID **)&(vwp->vw_dvlim[0]), "dvlim in indhklclean");
	return rev;
}

static void report_sec(REFINERUN *rfp, double *nexv, char *str)
{ /* reports results of search for secondary minimum in (X,Y) */
	if (nexv[0] < 1.e28)
	  {
		printf(
 "secondary minimum = %12.5e at (X,Y) = [%7.4f,%7.4f] %s\n",
			nexv[2], nexv[0], nexv[1], str);
		fprintf(rfp->rfr_fpl,
 "secondary minimum = %12.5e at (X,Y) = [%7.4f,%7.4f] %s\n",
			nexv[2], nexv[0], nexv[1], str);
	  }
	else {
		printf("secondary minimum could not be found\n");
		fprintf(rfp->rfr_fpl,
			"secondary minimum could not be found\n");
	     }
}

static int optimize_xy(REFINERUN *rfp)
{ /*	Scan through possible ax0, ay0 values and find the global
	minimum of a merged residual involving index, X, and Y.
	We don't look at the phi residual any more; it confuses things. */
	double		temp_ax0, temp_ay0, besty, bestrms, totresid;
	double		dex, dey, dey2, aby2, abx2, rad2, rad2b, rad2c;
	double		axm, aym, nextx, nexty, nextres, nextv[3];
	REFPARS		*xp;

	xp = rfp->rfr_xp;
	/* remember our original guesses for ax0 and ay0 */
	temp_ax0 = rfp->rfr_bestx = xp->xx0;	temp_ay0 = besty = xp->xy0;
	get_rmserrs("", rfp);	/* calc residuals at those original values */
	bestrms = totresid = rfp->rfr_indrmse +
	 (rfp->rfr_xrmse + rfp->rfr_yrmse) / 35.; /* pooled residual */
	/* scan through a rectangular box worth of ax0 and ay0 values
	 and find the best pooled residual */
	fprintf(rfp->rfr_fpl,
	 "Results of search through (X,Y) space for good residuals:\n");
	fprintf(rfp->rfr_fpl, "     ");
	for (dex = temp_ax0 - 0.2; dex <= temp_ax0 + 0.2; dex += 0.02)
		fprintf(rfp->rfr_fpl, " %5.2f", dex);
	fprintf(rfp->rfr_fpl, "\n");
	for (xp->xy0 = temp_ay0 - 0.2; xp->xy0 <= temp_ay0 + 0.2;
	 xp->xy0 += 0.02)
	 {
	  fprintf(rfp->rfr_fpl, "%5.2f", xp->xy0);
	  for (xp->xx0 = temp_ax0 - 0.2; xp->xx0 <= temp_ax0 + 0.2;
	   xp->xx0 += 0.02)
	     {
		get_rmserrs("", rfp);
		totresid = rfp->rfr_indrmse + (rfp->rfr_xrmse +
			rfp->rfr_yrmse) / 35.;
		if (totresid < bestrms)
		  {
			rfp->rfr_bestx = xp->xx0;	besty = xp->xy0;
			bestrms = totresid;
		  }
		fprintf(rfp->rfr_fpl, " %5.0f", 1000. * totresid);
	     }
	  fprintf(rfp->rfr_fpl, "\n");
	 }
	axm = rfp->rfr_bestx;	aym = besty;
	for (xp->xy0 = aym - 0.02; xp->xy0 <= aym + 0.02; xp->xy0 += 0.002)
	 {
	  for (xp->xx0 = axm - 0.02; xp->xx0 <= axm + 0.02; xp->xx0 += 0.002)
	     {
		get_rmserrs("", rfp);
		totresid = rfp->rfr_indrmse + (rfp->rfr_xrmse +
			rfp->rfr_yrmse) / 35.;
		if (totresid < bestrms)
		  {
			rfp->rfr_bestx = xp->xx0;	besty = xp->xy0;
			bestrms = totresid;
		  }
	     }
	 }
	printf("best net residual = %12.5e at (X,Y) = [%7.4f,%7.4f]\n",
	 bestrms, rfp->rfr_bestx, besty);
	fprintf(rfp->rfr_fpl,
	 "best net residual = %12.5e at (X,Y) = [%7.4f,%7.4f]\n",
	 bestrms, rfp->rfr_bestx, besty);
	dex = rfp->rfr_bestx - temp_ax0; dey = besty - temp_ay0;
	rad2b = dex * dex + dey * dey;
	if (rad2b < CLOSE_RADIUSSQ)
	  {
		xp->xx0 = rfp->rfr_bestx;	xp->xy0 = besty; return EOK;
	  }
	fprintf(rfp->rfr_fpl,
	 "This spot is %8.3f pixels from starting position [%7.4f,%7.4f]\n",
	 sqrt(rad2b), temp_ax0, temp_ay0);
	fprintf(rfp->rfr_fpl,
	 "Search for secondary solution closer to start is initiated\n");
	/* identify the second-lowest pooled residual among points that
	 are reasonably far from this point but closer to the start
	 than the originally-found peak */
	nextres = 1.e30;	nextx = nexty = 0.;
	for (xp->xy0 = temp_ay0 - 0.2; xp->xy0 <= temp_ay0 + 0.2;
	 xp->xy0 += 0.02)
	 {
	   dey = xp->xy0 - besty;	dey2 = dey * dey;
	   aby2 = xp->xy0 - temp_ay0;	aby2 *= aby2;
	   for (xp->xx0 = temp_ax0 - 0.2; xp->xx0 <= temp_ax0 + 0.2;
	    xp->xx0 += 0.02)
	    {
		dex = xp->xx0 - rfp->rfr_bestx;
		rad2 = dex * dex + dey2;
		abx2 = xp->xx0 - temp_ax0;	abx2 *= abx2;
		rad2c = aby2 + abx2;
		if ((rad2 > DISTINCT_MIN) && (rad2c <= rad2b))
		  {
			get_rmserrs("", rfp);
			totresid = rfp->rfr_indrmse + (rfp->rfr_xrmse +
				rfp->rfr_yrmse) / 35.;
			if (totresid < nextres)
			  {
				nextx = xp->xx0;	nexty = xp->xy0;
				nextres = totresid;
			  }
		  }
	     }
	 }
	axm = nextx;	aym = nexty;
	for (xp->xy0 = aym - 0.02; xp->xy0 <= aym + 0.02; xp->xy0 += 0.002)
	 {
	   dey = xp->xy0 - besty;	dey2 = dey * dey;
	   aby2 = xp->xy0 - aym;	aby2 *= aby2;
	   for (xp->xx0 = axm - 0.02; xp->xx0 <= axm + 0.02; xp->xx0 += 0.002)
	    {
		dex = xp->xx0 - rfp->rfr_bestx; rad2 = dex * dex + dey2;
		abx2 = xp->xx0 - axm;	abx2 *= abx2;
		rad2c = aby2 + abx2;
		if ((rad2 > DISTINCT_MIN) && (rad2c <= rad2b))
		  {
			get_rmserrs("", rfp);
			totresid = rfp->rfr_indrmse + (rfp->rfr_xrmse +
				rfp->rfr_yrmse) / 35.;
			if (totresid < nextres)
			  {
				nextx = xp->xx0;	nexty = xp->xy0;
				nextres = totresid;
			  }
		  }
	     }
	 }
	/* if we don't find a secondary minimum, or the secondary minimum is
	 much worse than the global minimum, or if it is actually farther
	 from the center than the global min, we use the global min. */
	dex = nextx - temp_ax0;	dey = nexty - temp_ay0;
	rad2 = dex * dex + dey * dey;
	nextv[0] = nextx;	nextv[1] = nexty;	nextv[2] = nextres;
	xp->xx0 = rfp->rfr_bestx;	xp->xy0 = besty;
	if (nextres > 1.e28) report_sec(rfp, nextv, "undetermined");
	else if (nextres > OPT_RATIO * bestrms)
		report_sec(rfp, nextv, "significantly worse");
	else if ((rad2 > DISTANCE_RATIO * rad2b) && (nextres > bestrms))
		report_sec(rfp, nextv, "not much closer to starting values");
	else { /* if the secondary minimum is substantially closer to the
		 original guess and its residual is about the same, use it */
		xp->xx0 = nextx;	xp->xy0 = nexty;
		report_sec(rfp, nextv, "used instead");
	     }
	return EOK;
}

int indhkl(double percerr[], double basis, REFINERUN *rfp)
{ /*	This performs an auto-indexing. The algorithm is
	based on looking at three h's, three k's, and three l's
	separately. The routine returns the number of autoindexing
	solutions found, or various negative integers upon error returns. */
	register int	i, j;
	int		nfound, nshowing, xyres;
	float		ispac[8];
	REFPARS		*xp;
	GUESS		*gup, *guq, *gul;
	DIFFVECTOR	*dvext[2];
	static char	*cp;
	static short	saveisys;
	static VECWORK	vws;

	vws.vw_fpl = rfp->rfr_fpl;	xp = rfp->rfr_xp;
	for (i = 0; i < 8; i++) ispac[i] = 0.;
	rfp->rfr_flags |= RR_CELLLIMITS;
	saveisys = xp->xisys;	xp->xisys = 1;	/* autoindex in P1 */
	vws.vw_autoverbose = 0;
	for (i = 0; i < MAX_MULTIP; i++) vws.vw_ndused[i] = 0;
	if (NULL != (cp = getenv("AUTO_VERBOSE")))
	  {
		if (0 < (j = atoi(cp))) vws.vw_autoverbose = j;
		else vws.vw_autoverbose = 1;
	  }
	if (vws.vw_autoverbose) fprintf(rfp->rfr_fpl,
	 "verbosity level %d set\n", vws.vw_autoverbose);
	if (percerr[0] >= 100.)
	  {
		percerr[0] = 100.;	rfp->rfr_flags &= ~RR_CELLLIMITS;
	  }
/* put the initial parameters in a GUESS structure */
	xptoguess(xp, rfp->rfr_gu0 + (NGUESSES - 1));
	if (0 != (i = reducedcell(xp->xareal))) return i;
	if (0 != (i = rp2vecs(xp))) return i;
	summ_aiset(stdout, xp);	summ_aiset(rfp->rfr_fpl, xp);
	if (rfp->rfr_flags & RR_CELLLIMITS)
	  {
		for (vws.vw_maxovsq = 0., i = 3; i--; )
		  {
			rfp->rfr_amin[i] = xp->xabc[i] *
				(1. - 0.01 * percerr[0]);
			rfp->rfr_amax[i] = xp->xabc[i] * (1. /
				(1. - 0.01 * percerr[0]));
			if (rfp->rfr_amax[i] > vws.vw_maxovsq)
				vws.vw_maxovsq = rfp->rfr_amax[i];
		  }
		vws.vw_maxovsq *= vws.vw_maxovsq / (xp->xlambda * xp->xlambda);
	  }
	else vws.vw_maxovsq = 4.e6;
	if ((percerr[1] < DFUZZ) || (percerr[1] > 100.)) percerr[1] = 10.;
	if (rfp->rfr_flags & RR_CELLLIMITS) for (i = 3; i--; )
	  {
		rfp->rfr_angmin[i] = xp->xang[i] - 0.90 * percerr[1];
		if (rfp->rfr_angmin[i] < 0) rfp->rfr_angmin[i] = 0;
		rfp->rfr_angmax[i] = xp->xang[i] + 0.90 * percerr[1];
		if (rfp->rfr_angmax[i] > 180) rfp->rfr_angmax[i] = 180;
		if (rfp->rfr_angmax[i] <= 90.)
		  {
			vws.vw_ualim[i] = 90. - rfp->rfr_angmax[i];
			vws.vw_ualim[i+3] = 90. - rfp->rfr_angmin[i];
		  }
		else if (rfp->rfr_angmin[i] > 90.)
		  {
			vws.vw_ualim[i] = rfp->rfr_angmin[i] - 90.;
			vws.vw_ualim[i+3] = rfp->rfr_angmax[i] - 90.;
		  }
		else if (180. - rfp->rfr_angmax[i] < rfp->rfr_angmin[i])
		  {
			vws.vw_ualim[i] = 0.;
			vws.vw_ualim[i+3] = rfp->rfr_angmax[i] - 90.;
		  }
		else {
			vws.vw_ualim[i] = 0.;
			vws.vw_ualim[i+3] = 90. - rfp->rfr_angmin[i];
		     }
	  }
	sum_consider(rfp->rfr_fpl, percerr, basis, rfp);
	xeflush(rfp->rfr_fpl);
	sum_consider(stdout, percerr, basis, rfp);	xeflush(stdout);
	/* call getprospects & prospects3 from within diffvectors & diffm3 */
	for (i = 0; i < 7; i++) vws.vw_totpro[i] = 0;
	nfound = diffvectors(ispac, percerr[0]*0.01, -basis, dvext, rfp, &vws);
	if (vws.vw_autoverbose)
	  {
		fprintf(rfp->rfr_fpl, "Results prior to length filtering:\n");
		i = summ_gresults(rfp->rfr_fpl, nfound, rfp);
		fprintf(rfp->rfr_fpl,
		 "  g_in g_out eunorm0 eunorm1 eunorm2 eunorm3\n");
	  }
	gul = rfp->rfr_gu0 + nfound;
	/* reorder guesses so redundant ones are later */
	for (gup = rfp->rfr_gu0; gup < gul; gup++)
	   {
		guq = gup + 1;
		if (gup->gresid >= 2.)
		  {
			for (guq = gup + 1; guq < gul; guq++)
			   if (guq->gresid < 2.)
			     {
				swap((AVOID *)gup,(AVOID *)guq,sizeof (GUESS));
				break;
			     }
			if (guq >= gul) break; /* exhausted all possibilities */
		  } 
		if ((gup > gul) || (guq > gul)) break;
	   }
	for (gup = rfp->rfr_gu0; gup < gul; gup++)
	   if (gup->gresid > 2.)
	     {
		nfound = gup - rfp->rfr_gu0; break;
	     }
	for (gup = rfp->rfr_gu0 + nfound; gup < gul; gup++)
	   if (gup->gresid < 2.)
		fprintf(rfp->rfr_fpl, "reorganization failed\n");
	i = summ_gresults(rfp->rfr_fpl, nfound, rfp);
	nshowing = (nfound < MAX_TO_STDOUT) ? nfound : MAX_TO_STDOUT;
	(void)summ_gresults(stdout, nshowing, rfp);
	if (EOK == i) i = nfound;
	rfp->rfr_autoind = 1;	/* set autoindex flag if we get this far */
	/* we used to die here if residual wasn't good enough, but no more */
	guesstoxp(rfp->rfr_gu0, xp);
	if (EOK != (i = unitcellx(rfp))) return indhklclean(i, &vws);
	i = nfound;	/* so that the routine returns the # of solutions */
	xp->xisys = saveisys;	/* restore original `isys' value */
	if ((!(rfp->rfr_flags & RR_NORECENTER)) &&
	 (EOK != (xyres = optimize_xy(rfp)))) return indhklclean(xyres, &vws);
		/* need header after this */
	rfp->rfr_flags |= RR_NEEDHEADER;	return indhklclean(i, &vws);
}

int autoind1(double erv[], REFINERUN *rfp)
{ /*	The autoindexing engine */
	register int	i;
	int		nsols, saveisys;
	double		err_allowed;
	time_t		tla;
	REFPARS		*xp;

	rfp->rfr_flags |= RR_CELLLIMITS;	xp = rfp->rfr_xp;
	tla = NOW;	strcpy(rfp->rfr_rundate, ctime(&tla));
	saveisys = xp->xisys;	xp->xisys = 1;	/* autoindex in P1 */
	err_allowed = erv[0];
	for (i = 3; i--; )
	  {
		rfp->rfr_amin[i] = xp->xabc[i] * (1. - err_allowed);
		rfp->rfr_amax[i] = xp->xabc[i] * (1. / (1. - err_allowed));
	  }
	erv[0] *= 100.;	/* make it a percentage */
	err_allowed = erv[1];
	if (rfp->rfr_flags & RR_CELLLIMITS) for (i = 3; i--; )
	  {
		rfp->rfr_angmin[i] = xp->xang[i] - 90. * err_allowed;
		if (rfp->rfr_angmin[i] < 0) rfp->rfr_angmin[i] = 0;
		rfp->rfr_angmax[i] = xp->xang[i] + 90. * err_allowed;
		if (rfp->rfr_angmax[i] > 180) rfp->rfr_angmax[i] = 180;
	  }
	erv[1] *= 100.;
	if (EOK > (nsols = indhkl(erv, erv[2], rfp)))
	  {
		fprintf(rfp->rfr_fpl, "Failure %3d in indhkl\n", nsols);
		return nsols;
	  }
	if (0 > (i = impose_solution(best_solution(nsols, rfp), rfp)))
		return EBAD;
	fprintf(rfp->rfr_fpl , "Solution %d used\n", i);
	xp->xisys = saveisys;	/* restore original `isys' value */
	if (EOK != (i = impose_special(rfp))) return i;
	rfp->rfr_flags |= RR_NEEDHEADER; /* we need a header after all this */
	return EOK;
}
