/* prospects.c -- autoindexing routines up through finding 3 basis vectors

	Copyright (c) 2003, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	25 Nov 2006: more careful checking for bad unit cell constructs in
	 prospects3 and the routines that it calls, using legalmat and others
	26 May 2006: reorganize {vectors,vecutils}.c -> {prospects,indexing}.c
	12 May 2006: cosmetic changes in angscore()
	15 Mar 2006: increase MOSTLYH from 0.8 to 0.82 and nonrand[] by 0.02
	28 Jun 2005: redun_sol(),prospects3(),scan_prospects(),diffm3():
		typecast pointer differences upon printout
	31 Dec 2005: don't qsort if fewer than 2 elements
	15 Oct 2005: add typedef VECWORK to eliminate static globals;
	 housekeeping; add g_lam field to GHKL vectors; move h2gv here
	19 Jan 2005: rfp->rfr_autoind set after auto-indexing
	28 Aug 2004: routines used only here are declared as static; cosmetics
	11 Aug 2004: several additional routines moved to vecutils.c
	 7 Jun 2004: limit init_prospects to a bookkeeping routine
	26 Jan 2004: some static globals converted to local variables
	21 Jan 2004: license message inserted
	 3 Jan 2004: REFINERUN and REFPARS structure elements in function calls
		replace remaining global variables; function calls updated
	13 Dec 2003: comment out unused variable dvalle in refi_resid0();
		remove unused array loch[] in refi_resid1()
	26 Nov 2003: replace global variables with elements of REFINERUN struct
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	22 Aug 2003: modify routines that access DIFFVECTOR structures
		according to calling sequences defined in diffvectors & diffm3
	22 Aug 2003: replace diffv0,dvlast with dvlim[2];
		change calls to local_resid, redun_sol accordingly;
		correct recently introduced bugs in function calls
	18 Aug 2003: created vecutils.c from portions of vectors.c
	14 Aug 2003: make diffv0 and dvlast static globals here
		and change the way memory is allocated for diffv0;
		add indhklclean() to make that happen
	13 Aug 2003: mm3w() moved to refxtal.c
	 3 May 2003: corrected ugly bug in redun_sol() under which the program
		always pared away the second, third, ... solutions from the
		list, even if they were different from the initial one
	 5 Jan 2003: reorganized source-file header
	15 Nov 2002: extern declarations made here rather than in refiext.h
	 1 Apr 2002: in rot9rt(), correct the way rea[2][i] are calculated
	31 Mar 2002: experiment in sift_prospects():
		after eliminating solutions whose dot product is close to zero,
		eliminate solutions that coplanar with 2 other shorter ones
	29 Mar 2002: merged with refihkl.c; many defines moved here from there
	25 Mar 2002: removed MSI attachments
	15 Dec 2001: in worstsofar(), don't swap the tested vector
		for the one in the list at this stage.
	11 Dec 2001: don't take square root in enor2()
	20 Sep 1997: in sift_prospects(), the criterion for saying 2 prospects
		are equivalent requires angle between them be < 8.1 deg.
	28 Jul 1997: major reorganization of diffvectors()
		means that we call getprospects and prospects3 from
		within diffvectors and diffm3 rather than later
	20 Jul 1997: in diffm3(), the algorithm explictly goes through the
		list systematically to find a satisfactory scoring function
	26 Oct 1995: guesscmp() does bounds-checking so Irix 5.3 qsort works
	 6 Jul 1995: "extern int dvncmp" declared on AXP port
	31 May 1995: "register" -> "register int" throughout
	 2 Mar 1995: eliminate calls to "gripe"
	30 Dec 1994: updating from Cerius2 1.6 to Cerius2 2.0
	23 Sep 1994: set dpsquish->dnsq to dpmid->dnsq, not dp->dnsq
	 2 Sep 1994: change how reference and representative
		difference vectors are chosen
	 5 May 1994: source-file header cleaned up,
		 names changed from POLYDET to XGEN; vsort.h included for VMS
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN version

   Purpose: the basic steps of autoindexing up through finding a 3x3 matrix
	form for each of the guessed indexing solutions.
	The input to diffm3 is a set of difference vectors that have
	already been checked for reasonableness and coalesced.
   Cross references: indexing.c
*************************************************************************/
#define		NEED_TIME	1
#include	"refidefs.h"
#define		RELATIVE_ERROR	0.01
#define		ANGLE_ERROR	0.16
#define		MAXLINES	21
#define		MAX_ANGLES	8
#define		MIN_RWEIGHT	0.76
#define		ALMOST_INTEGER	0.15	/* nearly an integer */
#define		MOSTLYI		0.65	/* frac of refs should be integers */
#define		MOSTLY		0.9	/* multiple-index frac for centering*/
#define		DEFINITELY_OK	0.05
#define		DEFAULT_COSSQ	(0.98)
#define		DEFAULT_PLANAR	(0.001)
#define		EXCELLENT_RESID	0.02	/* okay to use guess even if later */
#define		LENGTHDIFF	(0.0001)
#define		MINDIRECTION	(0.005)
#define		TWOOVERPI	0.63661977236758
#define		ONEOVERPI	0.31830988618379
#define		MAX_USERESID	0.2
#define		MINDIR2		(0.1e-4)
#define		MOSTLYH		0.82	/* single-index frac for centering*/
#define		MAX_CELLCHANGE	0.05
#define		STLPENLO	0.65	/* penalty-function val at max(stl)*/
#define		STLPENHI	1.0	/* penalty-function val at min(stl)*/
#define		STLDPEN		(STLPENHI - STLPENLO)
#define		NDUPMAX		20	/* dups > this are irrelevant */
#define		NDUP		3	/* min # duplications for reference*/
#define		NMPENHI		0.7	/* penalty-function val at max(nmer)*/
#define		NMPENLO		1.0	/* penalty-function val at min(nmer)*/
#define		NMDPEN		((double)(NMPENLO - NMPENHI))
#define		ENOUGH_PROSPECTS	100	/* enough indepd't solutions */
#define		MAXTOKEEP	30
#define		IND3MAX		29	/* max trial hkl value */

static void ghtogu(REFPARS *xp, GHKL *gq0, GHKL *gq1, GHKL *gq2,
	double lresid, GUESS *gup)
{ /*	this creates a GUESS structure at the right moment */
	int	i;	

	xptoguess(xp, gup);
	for (i = 0; i < 3; i++)
	   {	
		gup->ghs[0][i] = gq0->g_hv[i];
		gup->ghs[1][i] = gq1->g_hv[i];
		gup->ghs[2][i] = gq2->g_hv[i];
	   }
	gup->gresid = lresid;
}

static int outside_limits(float *uali, REFINERUN *rfp)
{ /*	This returns 0 iff the lengths and angles fall within the limits.
	Otherwise it returns 1+(index of offending index). We consider an
	angle that is the supplement of a given value to be okay. */
	int	i;
	double	ua;
	REFPARS	*xp;

	xp = rfp->rfr_xp;
	for (i = 0; i < 3; i++)
	   {
		if ((xp->xabc[i] < rfp->rfr_amin[i]) ||
		 (rfp->rfr_amax[i] < xp->xabc[i])) return i+1;
	   }
	for (i = 0; i < 3; i++)
	   {
		ua = fabs(90. - xp->xang[i]);
		if ((ua < uali[i]) || (ua > uali[i + 3])) return -i - 1;
	   }
	return 0;
}

static double enor2(double m[3][3])
{ /*	This returns the sum of the squares of the elements in the matrix m */
	int	i, j;
	double	sume, elt;

	for (sume = 0., i = 3; i--; ) for (j = 3; j--; )
	   {
		elt = m[i][j];	sume += elt * elt;
	   }
	return sume;
}

static void rot9rt(float rea[3][3], float rdeg[])
{ /* this rotates a real-space matrix by the TRANSPOSE of
	the matrix constructed as in `rot9r'.  The angles `rdeg'
	in this case are assumed to be in degrees rather than radians. */
	double		c0,c1,c2,s0,s1,s2;
	double		u900, u901, u902, u910, u911, u912, u920, u921, u922;
	double		a900, a901, a902, a910, a911, a912, a920, a921, a922;
  
	c0 = cos(rdeg[0] * RADC);	s0 = sin(rdeg[0] * RADC);
	c1 = cos(rdeg[1] * RADC);	s1 = sin(rdeg[1] * RADC);
	c2 = cos(rdeg[2] * RADC);	s2 = sin(rdeg[2] * RADC);
	u900 = rea[0][0]; u901 = rea[0][1]; u902 = rea[0][2];
	u910 = rea[1][0]; u911 = rea[1][1]; u912 = rea[1][2];
	u920 = rea[2][0]; u921 = rea[2][1]; u922 = rea[2][2];
	a900=  s0*c1*s2+c0*c2;	a901= -c0*c1*s2+s0*c2; a902= -s1*s2;
	a910= -s0*c1*c2+c0*s2;	a911=  c0*c1*c2+s0*s2; a912=  s1*c2;
	a920=  s0*s1;	  	a921= -c0*s1;	    a922=  c1;
	rea[0][0] = u900 * a900 + u901 * a910 + u902 * a920;
	rea[0][1] = u900 * a901 + u901 * a911 + u902 * a921;
	rea[0][2] = u900 * a902 + u901 * a912 + u902 * a922;
	rea[1][0] = u910 * a900 + u911 * a910 + u912 * a920;
	rea[1][1] = u910 * a901 + u911 * a911 + u912 * a921;
	rea[1][2] = u910 * a902 + u911 * a912 + u912 * a922;
	rea[2][0] = u920 * a900 + u921 * a910 + u922 * a920;
	rea[2][1] = u920 * a901 + u921 * a911 + u922 * a921;
	rea[2][2] = u920 * a902 + u921 * a912 + u922 * a922;
	return;
}

static GUESS *redun_sol(GUESS *gucur, int autov, float *ual, REFINERUN *rfp)
{ /*	This returns the pointer associated with a previously-determined
	solution if the current solution is redundant with respect to that
	solution. It starts by cleaning up the unit cell and making sure
	it's within bounds; if it isn't, the routine returns NULL and
	sets (rfp->rfr_xp)->xabc[0] to a flag value */
	int	collapse, i, j, same;
	double	eno0, eno1, eno2, eno3;
	float	locphi[3], tqre[3][3], tqrei[3][3], guqre[3][3], gupro[3][3];
	double	gupe0[3][3], gupe1[3][3], gupe2[3][3], gupe3[3][3];
	REFPARS	*xp;
	GUESS	*guq;

	xp = rfp->rfr_xp;
	if (0 != (j = rp2vecs(xp)))
	  {
		xp->xabc[0] = -1.; return NULL;
	  }
	if ((j = outside_limits(ual, rfp)))
	  {
		xp->xabc[0] = -1.; return NULL;
	  }
	locphi[0] = xp->xdelom;	locphi[1] = xp->xdelchi;
	locphi[2] = xp->xdelphi;
	if (0 != vecmat(xp->xabc, xp->xang, tqre)) return NULL;
	rot9rt(tqre, locphi);	if (0 != inv3(tqre, tqrei)) return NULL;
	for (guq = rfp->rfr_gu0; guq <= gucur; guq++)
	   {
		for (same = i = 0; i < 3; i++)
		   {
			if (fabs(guq->gabc[i] - xp->xabc[i]) >
			 guq->gabc[i] * RELATIVE_ERROR) break;
			if (fabs(guq->gang[i] - xp->xang[i]) > ANGLE_ERROR)
				break;
			if (fabs(guq->gomega[i] -locphi[i]) > ANGLE_ERROR)
				break;
			same++;
		   }
		if (same >= 3) return guq;
	/* here we do a more vigorous type of redundancy check. */
		if (0 == vecmat(guq->gabc, guq->gang, guqre))
		  {
			rot9rt(guqre, guq->gomega);
			mm3(gupro, tqrei, guqre);
			for (i = 0; i < 3; i++)
			   for (j = 0; j < 3; j++)
			      {
				gupe0[i][j] = gupe1[i][j] =
				gupe2[i][j] = gupe3[i][j] = gupro[i][j];
			      }
			gupe0[0][0] -= 1.; gupe0[1][1] -= 1.;
			gupe0[2][2] -= 1.;
			gupe1[0][0] += 1.; gupe1[1][1] += 1.;
			gupe1[2][2] -= 1.;
			gupe2[0][0] += 1.; gupe2[1][1] -= 1.;
			gupe2[2][2] += 1.;
			gupe3[0][0] -= 1.; gupe3[1][1] += 1.;
			gupe3[2][2] += 1.;
			eno0 = enor2(gupe0); eno1 = enor2(gupe1);
			eno2 = enor2(gupe2); eno3 = enor2(gupe3);
			collapse = 0;
			if (eno0 < MIN_EUNORMSQ) collapse = 1;
			if (eno1 < MIN_EUNORMSQ) collapse |= 2;
			if (eno2 < MIN_EUNORMSQ) collapse |= 4;
			if (eno3 < MIN_EUNORMSQ) collapse |= 8;
			if (collapse)
			  {
				if (autov > 1) fprintf(rfp->rfr_fpl,
 " %5d:%8.2f%8.2f%8.2f%8.2f%8.2f%8.2f%8.2f%8.2f%8.2f 0%o\n",
				 (int)(guq - rfp->rfr_gu0),
				 xp->xabc[0], xp->xabc[1], xp->xabc[2],
				 xp->xang[0], xp->xang[1], xp->xang[2],
				 locphi[0], locphi[1], locphi[2], collapse);
				return guq;
			  }
		  }
	   }
	return NULL;
}

static double local_resid(GUESS *gucur, DIFFVECTOR *dvli[], float *ualim,
	int autov, REFINERUN *rfp)
{ /*	This calculates the residual associated with the current cell. If the
	calculated cell lengths or angles are outside the defined range it
	returns a large residual so it won't get scored. It also checks to see
	if this solution is redundant with respect to previous solutions. */
	int		i;
	double		renu, rede, rwt, lh, slh, ss0, ss1, ss2;
	REFPARS		*xp;
	DIFFVECTOR	*dvp;

	if (NULL != redun_sol(gucur, autov, ualim, rfp)) return 1.7e9;
	xp = rfp->rfr_xp;
	if (xp->xabc[0] < 0.) return 1.5e9;
	for (dvp = dvli[0], renu = rede = 0.; dvp < dvli[1]; dvp++)
	   {
		ss0 = dvp->ss[0]; ss1 = dvp->ss[1]; ss2 = dvp->ss[2];
		rwt = fabs(dvp->dwt);
		for (i = 0; i < 3; i++)
		   {
			lh = xp->xareal[i][0] * ss0 +
				xp->xareal[i][1] * ss1 + xp->xareal[i][2] * ss2;
			slh = sin(PI * lh);	slh *= slh;
			renu += slh * rwt;	rede += 2. * rwt;
		   }
	   }
	if (rede < DFUZZ) return 1.9e9;	renu /= rede;
	if (0 != (i = rp2vecs(xp)))
		return 1.3e9 + 2.e7 * ((double)i);
	if ((i = outside_limits(ualim, rfp)))
		return 1.6e9 + 2.e7 * ((double)i);
	return renu;
}

static void summ_prospects(FILE *fp, int npros, GHKL *gh0)
{ /*	This summarizes the list of prospect GHKL structures to the stream fp*/
	int	i, j, npro, mang, lmang;
	double	tang;
	GHKL	*glp, *glq;
	char	*lolit;

	lolit = (fp == stdout) ? "\r\n" : "\n";
	fprintf(fp, " %6d prospective unit-cell vectors found%s",
		abs(npros), lolit);
	if (npros < 0)
	  {
		fprintf(fp,
		 "  uh  uk  ul     ax     ay     az     |a|   resid");
		fprintf(fp, "    Rax    Ray    Raz    R|a|  Rresid%s", lolit); 
	  }
	else {
		fprintf(fp,
 "               unrefined                 ||               refined%s", lolit);
		fprintf(fp, "     ax     ay     az     |a|  resid.");
		fprintf(fp, "    Rax    Ray    Raz    R|a|  resid."); 
		if (npros < 2) fprintf(fp, lolit);
		else if (npros == 2) fprintf(fp,
			" angle  with previously listed solution%s", lolit);
		else fprintf(fp,
			" angles with previously listed solutions%s", lolit);
	     }
	npro = abs(npros);
	if (((fp == stdout) || (fp == stderr)) && (npro > MAXLINES))
		npro = MAXLINES;
	mang = npro < MAX_ANGLES ? npro : MAX_ANGLES;
	for (i = 0, glp = gh0; i < npro; i++, glp++)
	   {
		if (npros < 0) fprintf(fp, " %3d %3d %3d",
		 (int)(glp->g_hv[0]),(int)(glp->g_hv[1]),(int)(glp->g_hv[2]));
		fprintf(fp,
		 " %6.1f %6.1f %6.1f %7.2f %7.4f %6.1f %6.1f %6.1f %7.2f %7.4f",
			glp->g_lam * glp->g_av[0], glp->g_lam * glp->g_av[1],
			glp->g_lam * glp->g_av[2], glp->g_lam * glp->g_le,
			glp->g_re,
			glp->g_lam * glp->g_rf[0], glp->g_lam * glp->g_rf[1],
			glp->g_lam * glp->g_rf[2], glp->g_lam * glp->g_lerf,
			glp->g_rerf);
		if ((npros < 0) || (i <= 0)) fprintf(fp, lolit);
		else {
			lmang = (i < mang ? i : mang);
			for (j = 0; j < lmang; j++)
			   {
				glq = gh0 + j;
				tang = acos(sqrt(cossqv(glq->g_rf,
					glp->g_rf))) / RADC;
				fprintf(fp, " %6.1f", tang);
			   }
			fprintf(fp, lolit);
		     }
	   }
	if ((fp != stdout) && (fp != stderr)) (void)xeflush(fp);
}

static void guess_reduce(REFINERUN *rfp, int *offs, int gtype,
 int iter, double lresid)
{ /*	This prints out the results of a guess-reduction operation */
	REFPARS	*xp;

	xp = rfp->rfr_xp;
	fprintf(rfp->rfr_fpl, "  %3d %3d %3d %4d %4d %8.5f %8.2f %8.2f %8.2f",
		offs[0], offs[1], offs[2], gtype, iter, lresid,
		xp->xabc[0], xp->xabc[1], xp->xabc[2]);
	fprintf(rfp->rfr_fpl, " %8.2f %8.2f %8.2f %8.2f %8.2f %8.2f\n",
		xp->xang[0], xp->xang[1], xp->xang[2],
		xp->xdelom, xp->xdelchi, xp->xdelphi);
}

static void head_reduce(FILE *fp)
{ /*	This prints a header for the cell reduction logging */
		fprintf(fp,
 "Cell reductions based on evenness of indices or divisibility by 3:\n");
		fprintf(fp,
 "Guess indices type iter    resid      Unit Cell lengths     |");
		fprintf(fp,
 "      Unit  Cell angles   |    Unit Cell Orientation\n");
}

static void report_shorten(FILE *fp, int *ndus, int maxmult)
{ /*	This reports to the stream fp on how many unit cell vectors
	were shortened by integer divisor values. Nothing is
	printed unless at least one category is occupied. */

	int	j, ndt, *nup;

	for (ndt = j = 0, nup = ndus; j < maxmult; j++, nup++) ndt += *nup;
	if (ndt < 1) return;
	fprintf(fp,
	 "Number of unit-cell vectors shortened by integer divisors:\n");
	fprintf(fp, " Divisor = 2");
	for (j = 3; j < maxmult; j++) fprintf(fp, " %4d", j);
	fprintf(fp, "\n %11d", *(ndus + 2));
	for (j = 3, nup = ndus + 3; j < maxmult; j++)
		fprintf(fp, " %4d", *nup);
	fprintf(fp, "\n");	if (fp != stderr) xeflush(fp);
}

static int refi_resid0(REFINERUN *rfp, DIFFVECTOR *dvli[], float aisst[3][3])
{ /*	the first step toward refining an auto-indexing solution */
	int		i, j, nusable, ilh;
	double		lh, slh, rwt, uij, tsi, ss0, ss1, ss2, loch[3];
	double		*ahp, *arp, *asp, ahp0[9], arp0[9], asp0[9];
	float		aisht[3][3], aissti[3][3];
	REFPARS		*xp;
	DIFFVECTOR	*dvp;
/*	DIFFVECTOR	*dvalle;	*/

	/* dvalle = dvli[0] + rfp->rfr_ndiffvecs; */
	xp = rfp->rfr_xp;
	for (arp = arp0, asp = asp0, ahp = ahp0, i = 0; i < 3; i++)
		for (j = 0; j < 3; j++, asp++, ahp++, arp++)
		   {
			*asp = *ahp = 0.; *arp = xp->xareal[i][j];
		   }
	/* go through the first `rfp->rfr_ndiffvecs' vectors, accumulating
	 sums that will be used to recalculate the unit cell matrix.
	 Only reflections that are close to integers will be used */
	/* temporary: use dvli[1], not dvalle */
	/*	for (dvp = dvli[0], nusable = 0; dvp < dvalle; dvp++) */
	for (dvp = dvli[0], nusable = 0; dvp < dvli[1]; dvp++)
	   {
		ss0 = dvp->ss[0]; ss1 = dvp->ss[1]; ss2 = dvp->ss[2];
		for (arp = arp0, i = 0, rwt = 1.; i < 3; i++, arp += 3)
		   {
			lh = (*arp) * ss0 + (*(arp+1)) * ss1 +
				(*(arp+2)) * ss2;
			slh = sin(PI * lh); slh = 1 - slh * slh;
			rwt *= slh; if (rwt <= MIN_RWEIGHT) break;
			loch[i] = lh;
		   }
		if (rwt > MIN_RWEIGHT)
		  {
		   if (rwt > 1.) rwt = 1.;
		    for (asp = asp0, ahp = ahp0, i = 0; i < 3; i++)
		      {
			lh = loch[i];
			if (lh < 0.) ilh = lh - 0.499999;
			else ilh = lh + 0.499999;	lh = ilh;
			tsi = dvp->ss[i];
			for (j = 0; j < 3; j++, asp++, ahp++)
			   {
				uij = dvp->ss[j] * rwt;
				*asp += tsi * uij;	*ahp += lh * uij;
			   }
		      }
		    nusable++;
		  }
	   }
	/* we need sufficient vectors and an invertible aisst matrix
	 in order to proceed productively */
	if (nusable <= 9) return 1;
	if (nusable < 0.5 * ((double)(dvli[1] - dvli[0]))) return 2;
	for (asp = asp0, ahp = ahp0, i = 0; i < 3; i++)
		for (j = 0; j < 3; j++, asp++, ahp++)
		   {
			aisst[i][j] = *asp; aisht[i][j] = *ahp;
		   }
	if (0 != inv3(aisst, aissti)) return 4;
	/* save the old unit cell matrix */
	for (i = 0; i < 3; i++)
	  for (j = 0; j < 3; j++) aisst[i][j] = xp->xareal[i][j];
	/* use the results of the initial refinement to
	 create a new unit cell matrix */
	mm3(xp->xareal, aisht, aissti);
	/* if this one has negative determinant, fix that */
	if (deter(xp->xareal) < 0.)
	  {
		for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
			xp->xareal[i][j] = -xp->xareal[i][j];
	  }
	/* create new cell lengths, angles, etc. */
	if (0 != rp2vecs(xp))
	  {
		for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
			xp->xareal[i][j] = aisst[i][j];
		if (0 != rp2vecs(xp)) return 8;
		return 16;
	  }
	for (i = 0; i < 3; i++)
	   {
		if ((xp->xabc[i] < rfp->rfr_amin[i]) ||
		 (xp->xabc[i] > rfp->rfr_amax[i])) return 32;
		if ((xp->xang[i] < rfp->rfr_angmin[i]) ||
		 (xp->xang[i] > rfp->rfr_angmax[i])) return 64;
	   }
	return 0;
}

static double refi_resid1(REFINERUN *rfp, double oldresi, DIFFVECTOR *dvli[],
	int autover, int *need_redo, float aisst[3][3])
{ /*	The subsequent steps in the refinement */
	int		i, j, integct, ict, nusable, ilh;
	int		in2[3], in3[3], nu[3], ulh[3];
	int		isumr[8];
	double		renu, rede, rwt, ss0, ss1, ss2, lh, slh, *arp, arp0[9];
	float		testa[3][3];
	REFPARS		*xp;
	DIFFVECTOR	*dvp;
	static float	acenm[3][3] = { { 1., 0., 0. }, { 0., 0.5, 0.5},
				{ 0., -0.5, 0.5 } } ;
	static float	bcenm[3][3] = { { 0.5, 0., -0.5}, { 0., 1., 0.},
				 { 0.5, 0., 0.5 } } ;
	static float	lmatc[3][3] = { { 0.5,-0.5, 0.0}, { 0.5, 0.5, 0.0},
				 {0.0, 0.0, 1.0 } } ;
	static float	lmati[3][3] = { {-0.5, 0.5, 0.5}, { 0.5,-0.5, 0.5},
				{ 0.5, 0.5,-0.5 } } ;

	*need_redo = 0;	/* default is not to signal re-calc */
	/* zero out the relevant variables */
	nu[2] = nu[1] = nu[0] = in2[2] = in2[1] = in2[0] =
		in3[2] = in3[1] = in3[0] = nusable = 0;
	for (i = 0; i < 8; i++) isumr[i] = 0;
	xp = rfp->rfr_xp;
	for (arp = arp0, i = 0; i < 3; i++)
		for (j = 0; j < 3; j++, arp++) *arp = xp->xareal[i][j];
	/* go back through all the difference vectors and
	 look for possible integer reductions in unit cell
	 or possible I, C, or F centering */
	for (dvp = dvli[0], renu = rede = 0.; dvp < dvli[1]; dvp++)
	   {
		ss0 = dvp->ss[0]; ss1 = dvp->ss[1]; ss2 = dvp->ss[2];
		rwt = fabs(dvp->dwt);	nusable++;	integct = 0;
		for (arp = arp0, i = 0; i < 3; i++, arp += 3)
		   {
			lh = *arp * ss0 + (*(arp + 1)) * ss1 +
				(*(arp+2)) * ss2;
			if (lh < 0.) ilh = lh - 0.499999;
			else ilh = lh + 0.499999;
			ict += ilh;
			if (fabs(lh - (double)ilh) < ALMOST_INTEGER)
			  {
				if (!(ilh % 2)) (in2[i])++;
				if (!(ilh % 3)) (in3[i])++;
				integct++;
				(nu[i])++;	ulh[i] = ilh;
			  }
			else	ulh[i] = -10000;
			slh = sin(PI * lh);	slh *= slh;
			renu += slh * rwt;	rede += 2. * rwt;
		   }
		if (integct == 3) /* are triples of indices even? */
		  {
			(isumr[0])++;
			if (!((ulh[0] + ulh[1] + ulh[2]) % 2)) (isumr[1])++;
		  }
		if (integct >= 2) /* are pairs of indices even? */
		  {
			if ((ulh[1] > -9000) && (ulh[2] > -9000))
			  {
				(isumr[2])++;
				if (!((ulh[1] + ulh[2]) % 2)) (isumr[3])++;
			  }
			if ((ulh[0] > -9000) && (ulh[2] > -9000))
			  {
				(isumr[4])++;
				if (!((ulh[0] + ulh[2]) % 2)) (isumr[5])++;
			  }
			if ((ulh[0] > -9000) && (ulh[1] > -9000))
			  {
				(isumr[6])++;
				if (!((ulh[0] + ulh[1]) % 2)) (isumr[7])++;
			  }
		  }
	   }
	for ((*need_redo) = i = 0; i < 3; i++)
	   { /* reduce the cell edge if required */
		if ((nu[i] > MOSTLYI * (double)nusable) &&
		 (in2[i] > MOSTLY * ((double)nu[i])))
		  {
			for (j = 0; j < 3; j++) xp->xareal[i][j] /= 2.;
			*need_redo |= (1 << i);
		  }
		if ((nu[i] > MOSTLYI * (double)nusable) &&
		 (in3[i] > MOSTLY * ((double)nu[i])))
		  {
			for (j = 0; j < 3; j++) xp->xareal[i][j] /= 3.;
			*need_redo |= (1 << (i+3));
		  }
	   }
	/* look for I, A, B, or C centering */
	if ((!((*need_redo) & 07)) &&
	 (isumr[0] > MOSTLYI * (double)nusable)) /* I-centering */
	  {
		if (isumr[1] > MOSTLY * isumr[0])
		  {
			*need_redo |= 0100; mm3(testa, lmati, xp->xareal);
			for (i = 3; i--; ) for (j = 3; j--; )
				xp->xareal[i][j] = testa[i][j];
		  }
	  }
	if ((!((*need_redo) & 0106)) &&
	 (isumr[2] > MOSTLYI * (double)nusable)) /* A-centering */
	  {
		if (isumr[3] > MOSTLY * isumr[2])
		  {
			*need_redo |= 0200; mm3(testa, acenm, xp->xareal);
			for (i = 3; i--; ) for (j = 3; j--; )
				xp->xareal[i][j] = testa[i][j];
		  }
	  }
	if ((!((*need_redo) & 0305)) &&
	 (isumr[4] > MOSTLYI * (double)nusable))
	  {
		if (isumr[5] > MOSTLY * isumr[4]) /* B-centering */
		  {
			*need_redo |= 0400; mm3(testa, bcenm, xp->xareal);
			for (i = 3; i--; ) for (j = 3; j--; )
				xp->xareal[i][j] = testa[i][j];
		  }
	  }
	if ((!((*need_redo) & 0703)) &&
	 (isumr[6] > MOSTLYI * (double)nusable))
	  {
		if (isumr[7] > MOSTLY * isumr[6]) /* C-centering */
		  {
			*need_redo |= 01000; mm3(testa, lmatc, xp->xareal);
			for (i = 3; i--; ) for (j = 3; j--; )
				xp->xareal[i][j] = testa[i][j];
		  }
	  }
	/* reorganize the cell if necessary */
	if (*(need_redo))
	  {
		if (autover > 1) fprintf(rfp->rfr_fpl,
 " need_redo [%7.2f,%7.2f,%7.2f,%7.2f,%7.2f,%7.2f,%7.2f,%7.2f,%7.2f]=0%o\n",
		 xp->xareal[0][0], xp->xareal[0][1], xp->xareal[0][2],
		 xp->xareal[1][0], xp->xareal[1][1], xp->xareal[1][2],
		 xp->xareal[2][0], xp->xareal[2][1], xp->xareal[2][2],
		 *need_redo);
		if ((i = reducedcell(xp->xareal))) /* permute & clean up */
		  { /* clean-up failed: revert */
			if (autover > 1) fprintf(rfp->rfr_fpl,
			 " reducedcell returned %d: reverting\n", i);
			for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
				xp->xareal[i][j] = aisst[i][j];
			if (0 != rp2vecs(xp)) return 1.1e9;
			*need_redo = 0;	renu = oldresi;
			return renu;
		  }
		else if (autover > 1) fprintf(rfp->rfr_fpl,
		 " reducedcell succeeded; returning smaller values\n");
	  }
	if (rede < DFUZZ)
	  {
		for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
			xp->xareal[i][j] = aisst[i][j];
		if (0 != rp2vecs(xp)) return 1.1e9;
		renu = oldresi;
	  }
	else {
		renu /= rede;
		if (((*need_redo)) &&  (0 != rp2vecs(xp))) return 1.7e9;
	     }
	return renu;
}

static int prospects3(int nprospects, REFINERUN *rfp, VECWORK *vwp)
{ /*	This goes through the GHKL structures in groups of three and assembles
	unit cell matrices from them. The ones that are reasonable threesomes
	are converted to GUESS structures and saved. The routine returns the
	number of valid GUESS structures generated. */
	int		i, j, breakout, notyet, os[3];
	time_t		tlc;
	double		bestsofar, bsof3, locresid;
	float		aisst[3][3];
	REFPARS		*xp;
	GHKL		*gl0, *glq0, *glq1, *glq2, *gle;
	GUESS		*gup;

	tlc = NOW;	strcpy(rfp->rfr_rundate, ctime(&tlc));
	vwp->vw_fpl = rfp->rfr_fpl;
	gl0 = &(vwp->vw_gh[0]);	gle = gl0 + nprospects;
	bestsofar = 1.e8;
	bsof3 = 3. * bestsofar;
	gup = rfp->rfr_gu0 + rfp->rfr_ndecent;
	fprintf(rfp->rfr_fpl, " after sifting: ");
	summ_prospects(rfp->rfr_fpl, nprospects, &(vwp->vw_gh[0]));
	xeflush(rfp->rfr_fpl);
	if (vwp->vw_autoverbose)
	  {
		fprintf(stdout," after sifting: ");
		summ_prospects(stdout, nprospects, &vwp->vw_gh[0]);
		xeflush(stdout);
	  }
	report_shorten(rfp->rfr_fpl, vwp->vw_ndused, MAX_MULTIP);
	notyet = 1;	xp = rfp->rfr_xp;
	for (os[0] = breakout = 0, glq0 = gl0; glq0 < gle; glq0++, (os[0])++)
	{
	 for (glq1 = glq0 + 1; glq1 < gle; glq1++)
	  {
	     os[1] = glq1 - gl0;
	     for (glq2 = glq1 + 1; glq2 < gle; glq2++)
		{
		  os[2] = glq2 - gl0;
		  for (i = 0; i < 3; i++)
		     {
			xp->xareal[0][i] = glq0->g_rf[i];
			xp->xareal[1][i] = glq1->g_rf[i];
			xp->xareal[2][i] = glq2->g_rf[i];
		     }
		  if ((locresid = deter(xp->xareal)) < 0.)
		    {
			for (i = 0; i < 3; i++)	for (j = 0; j < 3; j++)
				xp->xareal[i][j] = -xp->xareal[i][j];
		    }
		  /* check that the unit cell angles are legal and
			create a reduced cell */
		  if ((0 == mat2vecs(xp->xareal, xp->xabc, xp->xang)) &&
		   (0 == (i = reducedcell(xp->xareal))))
		    {
			locresid = local_resid(gup, vwp->vw_dvlim,
				vwp->vw_ualim, vwp->vw_autoverbose, rfp);
			if (locresid < bsof3)
			  {
				for (j = 0; ; j++)
				   {
					if (refi_resid0(rfp, vwp->vw_dvlim,
						aisst)) break;
					locresid = refi_resid1(rfp, locresid,
					 vwp->vw_dvlim,
					 vwp->vw_autoverbose, &i, aisst);
					if (!i) break;
					if ((vwp->vw_autoverbose) &&
					 ((vwp->vw_autoverbose > 1) ||(j > 0)))
					  {
						if (notyet)
						 head_reduce(rfp->rfr_fpl);
						notyet = 0;
						guess_reduce(
						 rfp, os, i, j, locresid);
					  }
				   }
				if ((0 == (i = reducedcell(xp->xareal))) &&
				   (NULL == redun_sol(gup, vwp->vw_autoverbose,
				    vwp->vw_ualim, rfp)) && (xp->xabc[0] > 0.))
				  {
					ghtogu(xp, glq0, glq1, glq2,
					 	locresid, gup++);
					if ((gup - rfp->rfr_gu0) > NGUESSES-2)
					  {
						fprintf(rfp->rfr_fpl,
	"Warning: %4d guesses found: no further guesses will be researched\n",
						 1 +(int)(gup - rfp->rfr_gu0));
						xeflush(rfp->rfr_fpl);
						breakout = 1;	break;
					  }
					if (locresid < bestsofar)
					  {
						bestsofar = locresid;
						/* if solutions are REALLY good,
						keep them even if they're more
						than 3* the best one so far */
						bsof3 = 3. * bestsofar;
						if (bsof3 < DEFINITELY_OK)
							bsof3 = DEFINITELY_OK;
					  } /* end of "if (locresid " clause */
				  } /* end of "if (...reducedcell...)" clause*/
			  } /* end of "if (locresid < bsof3)" clause */
		    }	/* end of "if (0 == (i = reducedcell" clause */
		} /* end of "for (glq2" loop */
	      if (breakout) break;
	    }
	  if (breakout) break;
	 }
	tlc = NOW;	strcpy(rfp->rfr_rundate, ctime(&tlc));
	if (bestsofar > 1.e7) return 0;
	return gup - rfp->rfr_gu0;
}

static int ghcmp(GHKL *ghu0, GHKL *ghu1)
{ /* compare routine used in sorting GHKL structures */
	return (ghu0->g_rerf < ghu1->g_rerf ? -1 : 1);
}

static void tosshead(FILE *fp)
{ /* header for list of tossed guesses */
	int	 i;
	for (i = 0; i < 3; i++)
	 fprintf(fp, "              ghq%1d                  ", i);
	fprintf(fp, " planarity eliminated\n");
	for (i = 0; i < 3; i++)
		fprintf(fp, "      x      y      z length   resid");
	fprintf(fp, "\n");
}

static void summ_toss(FILE *fp, GHKL *gq0, GHKL *gq1, GHKL *gq2,
 double csq, int which)
{ /* This summarizes, to the stream fp, a tossed guess */
	if (csq > 1.) csq = 0.;	else csq = acos(sqrt(csq)) / RADC;
	fprintf(fp, "%7.1f%7.1f%7.1f%7.2f%8.5f%7.1f%7.1f%7.1f%7.2f%8.5f",
	 gq0->g_rf[0] * gq0->g_lam, gq0->g_rf[1] * gq0->g_lam,
	 gq0->g_rf[2] * gq0->g_lam, gq0->g_lerf * gq0->g_lam, gq0->g_rerf,
	 gq1->g_rf[0] * gq1->g_lam, gq1->g_rf[1] * gq1->g_lam,
	 gq1->g_rf[2] * gq1->g_lam, gq1->g_lerf * gq1->g_lam,
	 gq1->g_rerf);
	fprintf(fp, "%7.1f%7.1f%7.1f%7.2f%8.5f %8.3f %10d\n",
	 gq2->g_rf[0] * gq2->g_lam, gq2->g_rf[1] * gq2->g_lam,
	 gq2->g_rf[2] * gq2->g_lam, gq2->g_lerf * gq2->g_lam,
	 gq2->g_rerf, csq, which);
}

static int shouldtoss(int *needtp, float *xpro,
	GHKL *gq0, GHKL *gq1, GHKL *gq2, VECWORK *vwp)
{ /*	This returns 1 if it's appropriate to get rid of prospect gq2.
	It returns 0 if gq2 isn't coplanar with gq0 and gq1.
	It returns 1 if gq2 is coplanar with gq0 and gq1 and is
	longer than either of them. It also returns 1 if gq2 is coplanar
	with gq0 and gq1 but is shorter than one or the other, but
	nonetheless has an excellent residual; under these conditions
	we first swap gq2 for the longer of those two and then signal
	that the resulting fatality be tossed. */
	int	which;
	GHKL	*bigger;
	double	cma, cos2;

	cma = DEFAULT_PLANAR;
	if ((cos2 = cossqv(xpro, gq2->g_rf)) >= cma) return 0;
	if ((gq2->g_lerf > gq1->g_lerf) && (gq2->g_lerf > gq0->g_lerf))
	  {
		if (vwp->vw_autoverbose)
		  {
			if (*needtp) tosshead(vwp->vw_fpl);	*needtp = 0;
			summ_toss(vwp->vw_fpl, gq0, gq1, gq2, cos2, 2);
		  }
		return 1;
	  }
	if (gq2->g_rerf < EXCELLENT_RESID)
	  {
		bigger = (gq1->g_lerf > gq0->g_lerf) ? gq1 : gq0;
		if (vwp->vw_autoverbose)
		  {
			which = (gq0 == bigger) ? 0 : 1;
			if (*needtp) tosshead(vwp->vw_fpl);
			*needtp = 0;
			summ_toss(vwp->vw_fpl, gq0, gq1, gq2, cos2, which);
		  }
		swap((AVOID *)bigger, (AVOID *)gq2, sizeof (GHKL));
		return 1;
	  }
	return 0;
}

static int sift_prospects(int totpros, int maxkept, VECWORK *vwp)
{ /*	This sorts the prospects and keeps only the best `maxkept'.
	For now it just uses the lowest residuals; later perhaps we'll get
	more careful. Criterion for declaring 2 prospects are equivalent
	requires that the angle between them be < 8.1 deg. Currently,
	after eliminating solutions whose dot product is ~ 0, we also
	eliminate any solution that is coplanar with 2 shorter solutions. */
	register int	i, j, k, m;
	int		needtoss;
	double		cmi, xpn;
	float		xprod[3];
	GHKL		*glq0, *glq1, *glq2, *glq3;

	cmi = DEFAULT_COSSQ;	needtoss = 1;
	if (totpros > 1)
	  qsort(vwp->vw_gl0, (size_t)totpros, sizeof (GHKL), (QPE)ghcmp);
	if (vwp->vw_autoverbose > 1)
	  {
		fprintf(vwp->vw_fpl, "before sifting: ");
		summ_prospects(vwp->vw_fpl, -totpros, &(vwp->vw_gh[0]));
	  }
	fprintf(vwp->vw_fpl,
		" %6d prospects before beginning sifting process\n", totpros);
	/* first check for small dot products */
	for (glq0 = &(vwp->vw_gh[0]), i = 0; i < totpros - 1; i++, glq0++)
	   {
		for (glq1 = glq0 + 1, j = i + 1; j < totpros; j++, glq1++)
		   {
			if (cossqv(glq0->g_rf, glq1->g_rf) > cmi)
			  {
				for (glq2 = glq1, k = j; k < totpros - 1; k++,
				 glq2++) cpymem(glq2, glq2 + 1, sizeof (GHKL));
				cpymem(glq2, glq1, sizeof (GHKL));
				if ((--totpros) < 2) break;
				glq1--;	j--;
			  }
		   }
		if (totpros < 2) break;
	   }
	fprintf(vwp->vw_fpl,
		" %6d prospects after removing small dot products\n", totpros);
	/* now check for large dot products between the cross product
	 of two vectors and a third vector. This would imply that the
	 three vectors are coplanar, or nearly so */
	for (glq0 = &(vwp->vw_gh[0]), i = 0; i < totpros - 2; i++, glq0++)
	   {
	    for (glq1 = glq0 + 1, j = i + 1; j < totpros - 1; j++, glq1++)
	     {
		crossproduct(glq0->g_rf, glq1->g_rf, xprod);
		/* normalize the cross product for tidyness */
		for (xpn = m = 0; m < 3; m ++) xpn += xprod[m] * xprod[m];
		if (xpn > DFUZZ) /* otherwise these vectors are parallel... */
		 {
		   xpn = sqrt(xpn); for (m = 0; m < 3; m++) xprod[m] /= xpn;
		   for (glq2 = glq1 + 1, k = j + 1; k < totpros; k++, glq2++)
		   { /* Current rule is that we toss the new vector only
			if it's longer than the other two. If it's
			shorter than the others and still has an excellent
			residual, we toss one of the others */
			if (shouldtoss(&needtoss, xprod, glq0, glq1, glq2, vwp))
			  {
				for (glq3 = glq2, m = k; m < totpros - 1; m++,
				 glq3++) cpymem(glq3, glq3 + 1, sizeof (GHKL));
				cpymem(glq3, glq2, sizeof (GHKL));
				if ((--totpros) < 2) break;
				glq2--;	k--;
			  }
		   } /* end of "for (glq ..." loop */
		 } /* end of "if (xpn > DFUZZ)" clause */
		if (totpros < 2) break;
	      } /* end of "for (glq1 ..." loop */
	   }
	fprintf(vwp->vw_fpl,
	 " %6d prospects after removing coplanar triplets\n", totpros);
	j = (totpros < maxkept) ? totpros : maxkept;
	return j;
}

static int guesscmp(GUESS *gu1, GUESS *gu2)
{ /*	15 Oct 05: eliminate the limit checks that Irix 5.3 wanted */
	return	(gu1->gresid < gu2->gresid ? -1 : 1);
}

static void init_prospects(VECWORK *vwp)
{ /*	This sets up for the generation of the list of prospective
	unit cell vectors. Currently it's trivial. */
	vwp->vw_glc = vwp->vw_gl0 = &(vwp->vw_gh[0]);
	vwp->vw_vecheader = 0;
}

static int samevec(GHKL *gla, GHKL *glb, int tryswap)
{ /*	This returns 1 if gla->g_re is either identical to glb->g_re
	within noise, or if gla->g_re = -glb->g_re within noise.
	If gla has a lower residual than glb, we swap their identities.	*/
	int	i, j;
	double	dotp, u1n, t;
	float	*pf0, *pf1;

	u1n = glb->g_lerf;
	if (fabs(gla->g_lerf - u1n) > u1n * LENGTHDIFF) return 0;
	u1n *= u1n;
	for (dotp = i = 0, pf0 = gla->g_rf, pf1 = glb->g_rf;
	 i < 3; i++, pf0++, pf1++) dotp += *pf0 * *pf1;
	if (fabs(fabs(dotp) / u1n - 1.) > MINDIRECTION) return 0;
	/* having gotten this far, the two vectors are definitely the same. */
	/* if we've already refined things and the new (gla) guess is better
	 than the similar (glb) guess, we swap the values in so that the data
	 for gla is kept in the glb slot and vice versa. */
	if (!(tryswap) || (gla->g_rerf >= glb->g_rerf))
	  { /* flag signaling duplication*/
		gla->g_rerf = 3.; return 1;
	  }
	for (i = 0; i < 3; i++)
	   {
		t = gla->g_rf[i];	gla->g_rf[i] = glb->g_rf[i];
		glb->g_rf[i] = t;	t = gla->g_av[i];
		gla->g_av[i] = glb->g_av[i];	glb->g_av[i] = t;
		j = gla->g_hv[i];	gla->g_hv[i] = glb->g_hv[i];
		glb->g_hv[i] = j;
	   }
	t = gla->g_le;	gla->g_le = glb->g_le;	glb->g_le = t;
	t = gla->g_lerf;	gla->g_lerf = glb->g_lerf; glb->g_lerf = t;
	t = gla->g_re;	gla->g_re = glb->g_re;	glb->g_re = t;
	glb->g_rerf = gla->g_rerf;	gla->g_rerf = 3.;	return 1;
}

static GHKL *worstsofar(int *h2er, GHKL *glp, GHKL *gh0, GHKL *ghe)
{ /*	This searches through the currently-constructed list of
	GHKL structures and returns the pointer associated with the
	one with the worst residual. As it does so, it also looks
	to see whether the current guess, glp, is actually the same as
	one of the ones on the list; if it is, the routine returns NULL. */
	double	rwor;
	GHKL	*glq, *gwor;

	for (glq = gh0, gwor = NULL, rwor = 0.; glq <= ghe; glq++)
	   {
		if (samevec(glp, glq, 1))
		  {
			*h2er = 6; return NULL;
		  }
		if (rwor < glq->g_rerf)
		  {
			rwor = glq->g_rerf; gwor = glq;
		  }
	   }
	/* if we got this far, the vector is a keeper. Keep it. */
	*h2er = 0; return (NULL == gwor) ? glp : gwor;
}

static int h2gv(int inh[], GHKL *glp, REFINERUN *rfp, VECWORK *vwp)
{ /*	This computes the components of the GHKL vector *glp associated
	with the index vector inh. It returns 0 if successful,
	2 if not. We do the calculations in two stages here
	because this routine has to be ridiculously fast.  */
	int	i, j;
	double	rh0, rh1, rh2, he, she;
	/* comment out the next 2 lines if we skip the check for redundancies*/
	double	tle, lox, loy, loz, u1n, ssx, ssy, ssz, dotp;
	GHKL	*glq;

	for (j = i = 0; i < 3; i++) if (inh[i]) j++;
	if (!j)
	  {
		glp->g_rerf = 2.;	return 1;
	  }
	rh0 = inh[0];	rh1 = inh[1];	rh2 = inh[2];
	glp->g_lam = (rfp->rfr_xp)->xlambda;
	for (she = i = 0; i < 3; i++)
	   {
		glp->g_rf[i] = he = rh0 * rfp->rfr_sinv[0][i] +
			rh1 * rfp->rfr_sinv[1][i] + rh2 * rfp->rfr_sinv[2][i];
		she += he * he;
	   }
	/* if the square of the cell length is larger than the square of the
	 largest allowed unit cell dimension, we treat that as an error */
	if (she > vwp->vw_maxovsq) return 2;
	glp->g_le = she;
	for (i = 0; i < 3; i++)
	   {
		glp->g_hv[i] = inh[i];	glp->g_av[i] = glp->g_rf[i];
	   }
	glp->g_lerf = glp->g_le = sqrt(glp->g_le);
	if (glp->g_le < FUZZ) return 1; /* prevent divide-by-zeroes below */
	/* make sure we haven't seen this one before */
/* try comparing dot products without much regard to length:
  if it's shorter, we keep it regardless. If it's longer, we don't
  if the angle between the two solutions is nearly 0 or 180 */
	tle = glp->g_le;
	lox = glp->g_av[0]; loy = glp->g_av[1]; loz = glp->g_av[2];
	for (glq = &(vwp->vw_gh[0]); glq < vwp->vw_glen; glq++)
	   {
	     u1n = glq->g_le;
	     if (tle > 0.9 * u1n) 
	      {
		ssx = glq->g_av[0]; ssy = glq->g_av[1]; ssz = glq->g_av[2];
		dotp = lox * ssx + loy * ssy + loz * ssz;
		she = dotp / (u1n * tle);
		if (fabs(fabs(she) - 1.) < MINDIR2)
		  {
			if (vwp->vw_autoverbose)
			  {
				if (fabs(she) < 1.) she = acos(she);
				else she = (she < 0.) ? PI : 0.;
				fprintf(rfp->rfr_fpl,
		 " [(%7.1f,%7.1f,%7.1f),len %8.2f],",
					ssx, ssy, ssz, u1n);
				fprintf(rfp->rfr_fpl,
		 " [(%7.1f,%7.1f,%7.1f),len %8.2f]: angle %7.2f deg\n",
					lox, loy, loz, tle, she / RADC);
			  }
			return 3;
		  }
	       }
	   }
	return 0;	/* successful completion */
}

static double angscore(DIFFVECTOR *dvp0, DIFFVECTOR *dvp1)
{ /* This returns a score between 0 and 1 associated with the
   desirability of using a pair of difference vectors dvp0, dvp1
   based on the angle between them. */
	int	i;
	double	a01, a00, a11;
	float	*p0, *p1;

	p0 = &(dvp0->ss[0]);	p1 = &(dvp1->ss[0]);
	for (a01 = a00 = a11 = i = 0; i < 3; i++, p0++, p1++)
	   {
		a01 += *p0 * *p1;	a00 += *p0 * *p0; a11 += *p1 * *p1;
	   }
	if ((a00 < DFUZZ) || (a11 < DFUZZ)) return 0.;
	return flimits(1. - (a01 / a00) * (a01 / a11), 0., 1.);
}

static void vecshow(VECWORK *vwp, GHKL *glp, int indx,
	double hren, double fracre, double he, char *str)
{ /*	Prints a diagnostic to vwp->vw_fpl during vector creation */
	if (vwp->vw_vecheader)
	  {
		fprintf(vwp->vw_fpl, "           vector         length   ");
		fprintf(vwp->vw_fpl,
			"resid divisor frac.integ div.resid shorten?\n");
		vwp->vw_vecheader = 1;
	  }
	fprintf(vwp->vw_fpl,
	 " %7.2f %7.2f %7.2f %7.2f %7.4f %7d %10.4f %9.4f %s\n",
	 glp->g_av[0], glp->g_av[1], glp->g_av[2],
	 glp->g_lam * glp->g_le, hren, indx, fracre, he, str);
}

static void recalc_gl(GHKL *glp, DIFFVECTOR *dvli[])
{ /*	This does the final recalculation of the gl vector.
	It's separated out from `refgv' mostly for profiling purposes */
	int		i;
	double		lox, loy, loz, hren, hred, ssx, ssy, ssz, he, she, rwt;
	DIFFVECTOR	*dvp;

	lox = glp->g_rf[0]; loy = glp->g_rf[1]; loz = glp->g_rf[2];
	for (hren = hred = 0, dvp = dvli[0]; dvp < dvli[1]; dvp++)
	   {
		ssx = dvp->ss[0]; ssy = dvp->ss[1]; ssz = dvp->ss[2];
		he = ssx * lox + ssy * loy + ssz * loz;
		rwt = fabs(dvp->dwt);
		she = sin(PI * he);
		hren += she * she * rwt;
		hred += rwt;
	   }
	if (hred > DFUZZ) glp->g_rerf = hren / (2. * hred);
	if (glp->g_rerf < glp->g_re)
	  { /* use the refined residual if it's better */
		for (she = i = 0; i < 3; i++)
		  {
			he = glp->g_rf[i];	she += he * he;
		  }
		glp->g_lerf = sqrt(she);
	  }
	else { /* ... and don't if it isn't! */
		for (i = 0; i < 3; i++) glp->g_rf[i] = glp->g_av[i];
		glp->g_lerf = glp->g_le;	glp->g_rerf = glp->g_re;
	     }
}

static int refgv(GHKL *glp, VECWORK *vwp)
{ /*	This does the zeroth stage of refining a GHKL structure
	based on the difference vector list. It does so by assembling
	sums that can be used to solve a least-squares problem. Along
	the way it also checks to see if the solution might actually be
	an integer multiple of the real solution. It returns an integer
	error value: 0 if good, 4, 5, or 6 if bad */

	int		i, j, icl, ire[MAX_MULTIP], divisor;
	double		deh, fracre, he, hred, hren, lox, loy, loz, ricl;
	double		rnu, rwt, she, ssx, ssy, ssz, uwt, wsx, wsy, wsz;
	double		*flp, flp0[9], hre[MAX_MULTIP], locd[3][3], locdi[3][3];
	float		*slp;
	DIFFVECTOR	*dvp;
	static double	nonrand[] = { 0.92, 0.92, 0.92, 0.90, 0.88,
		0.86, 0.84, 0.82, 0.80, 0.78, 0.76, 0.74, 0.72 } ; 

	if (glp->g_le < FUZZ) return 0;
	/* to speed things up, create some local vbls */
	lox = glp->g_av[0];	loy = glp->g_av[1];	loz = glp->g_av[2];
	/* use sin(sq) residual instead of absval residual */
	for (i = 2; i < MAX_MULTIP; i++) hre[i] = ire[i] = 0;
	for (flp = flp0, i = 0; i < 9 ; i++, flp++) *flp = 0.;
	hren = hred = 0.;
	for (rnu = 0., dvp = vwp->vw_dvlim[0]; dvp < vwp->vw_dvlim[1]; dvp++)
	   { /* compute the index value and find the closest integer */
		slp = &(dvp->ss[0]); ssx = *slp++; ssy = *slp++; ssz = *slp;
		he = lox * ssx + loy * ssy + loz * ssz;
		icl = (he >= 0.) ? he + 0.499999: he - 0.499999;
		/* look for possible multiples of a true axis */
		for (i = 2; i < MAX_MULTIP; i++) if (!(icl % i)) (ire[i])++;
		rwt = fabs(dvp->dwt);
		she = sin(PI * he);	she *= she; uwt = rwt * (1. - she);
		hren += she * rwt;	hred += rwt;	/* increment residual*/
		for (deh = TWOOVERPI, i = 2; i < MAX_MULTIP;
		  i++, deh += ONEOVERPI)
		   {
			she = sin(he / deh);	hre[i] += she * she * rwt;
		   }
		deh = he - (double)icl;
		if (fabs(deh) < MAX_USERESID)	/* refinement sums */
		  {
			wsx = uwt * ssx; wsy = uwt * ssy; wsz = uwt * ssz;
			ricl = icl;	flp = flp0;
			*flp++ += wsx * ssx;	*flp++ += wsx * ssy;
			*flp++ += wsx * ssz;	*flp++ += wsy * ssy;
			*flp++ += wsy * ssz;	*flp++ += wsz * ssz;
			*flp++ += wsx * ricl;	*flp++ += wsy * ricl;
			*flp += wsz * ricl;
		  }
		rnu += 1;
	   }
	if (hred < DFUZZ)
	  {
		glp->g_rerf = 2.; return 4;
	  }
	hren /= (hred * 2.);	divisor = 1;
	/* see if this vector is an integer multiple of a true vector */
	for (i = MAX_MULTIP - 1; i > 1; i--)
	   {
		fracre = ((double)ire[i]) / rnu; he = hre[i] / (hred * 2.);
		if ((fracre > nonrand[i]) ||
		 ((fracre > MOSTLYH * nonrand[i]) && (he < 0.5 * hren)))
		  {
			if (he < hren)
			  {
				if (vwp->vw_autoverbose > 2) vecshow(vwp,
				 glp, i, hren, fracre, he, "yes");
				hren = he;	flp = flp0 + 6;
				for (j = 0; j < 3; j++, flp++)
					*flp /= ((double)i);
				(vwp->vw_ndused[i])++;
				divisor = i;	break;
			  }
			else if (vwp->vw_autoverbose > 2) vecshow(vwp,
			 glp, i, hren, fracre, he, "hre[i] > hren");
		  }
		else if ((vwp->vw_autoverbose > 2) &&
		 (fracre > MOSTLYH * nonrand[i]) && (he < hren)) vecshow(vwp,
		 glp, i, hren, fracre, he, "fracre too high");
	  }
	flp = flp0;
	locd[0][0] = *flp++;	locd[1][0] = locd[0][1] = *flp++;
	locd[2][0] = locd[0][2] = *flp++;	locd[1][1] = *flp++;
	locd[2][1] = locd[1][2] = *flp++;	locd[2][2] = *flp++;
	/* if the least-squares normal matrix isn't invertible, we've failed */
	if (dinv3(locd, locdi))
	  {
		for (i = 0; i < 3; i++) glp->g_rf[i] = glp->g_av[i];
		return 5;
	  }
	/* otherwise, we record the residual and the new vector */
	glp->g_rerf = glp->g_re = hren;
	for (she = i = 0; i < 3; i++) /* do the refinement */
	   {
		for (flp = flp0 + 6, he = 0., j = 0; j < 3; j++, flp++)
			he += locdi[i][j] * *flp;
		glp->g_rf[i] = he;
		he -= glp->g_av[i] / (double)divisor; she += he * he;
	   }
	she = sqrt(she) * ((double)divisor) / glp->g_le;
/* make sure the changes aren't too large */
	if (she > MAX_CELLCHANGE)
	  {
		she = MAX_CELLCHANGE / she;
		for (i = 0; i < 3; i++)
		   {
			deh = she * (glp->g_rf[i] - glp->g_av[i]);
			glp->g_rf[i] = glp->g_av[i] + deh;
		   }
	  }
	/* recalc the hkl values */
	recalc_gl(glp, vwp->vw_dvlim);
	/* if the refined residual is worse than RANDOM, we're not going
	 to use it anyway. So throw it away. */
	if (glp->g_rerf >= RANDOM) return 6;
	return 0;
}

static int flip(int iz)
{ /* This gives us a simple way of stepping through integers in
	the order 0, 1, -1, 2, -2, ... */
	return (iz > 0 ? -iz: -iz + 1);
}

static int getprospects(int hli, REFINERUN *rfp, VECWORK *vwp)
{ /*	This finds prospects associated with a specific sref matrix
	and adds them to the growing list thereof */
	int	uh[3], nprospects, getworst, nimprove, h2err;
	int	erct[8];
	GHKL	*glp, *gwor, gt;

	for (getworst = 0; getworst < 8; getworst++) erct[getworst] = 0;
	nimprove = getworst = 0;	vwp->vw_glen = glp = vwp->vw_glc;
	for (uh[0] = 0; uh[0] <= hli; uh[0] = flip(uh[0]))
	   for (uh[1] = 0; uh[1] <= hli; uh[1] = flip(uh[1]))
		for (uh[2] = 0; uh[2] <= hli; uh[2]++)
		   {
		     if ((0 == (h2err = h2gv(uh, &gt, rfp, vwp))) &&
			(0 == (h2err = refgv(&gt, vwp))))
		      {
			gwor = worstsofar(&h2err, &gt, &(vwp->vw_gh[0]),
			 vwp->vw_glen);
			if ((NULL != gwor) && (RANDOM > gt.g_rerf))
			  {
				if ((getworst) || (glp - vwp->vw_gl0 >= MAXG))
				  {
					glp = gwor; getworst = 1;
					if (gt.g_rerf < gwor->g_rerf)
					  {
						cpymem((AVOID *)glp,
						 (AVOID *)&gt, sizeof (GHKL));
						nimprove++;
					  }
				  }
				else {
					cpymem((AVOID *)glp, (AVOID *)&gt,
					 sizeof (GHKL));
					glp++;	vwp->vw_glen = glp;
				     }
			  }
			else h2err = 7;
		      }
		     (erct[h2err])++;
		   }
	if (getworst) glp = vwp->vw_gl0 + MAXG;
	nprospects = glp - vwp->vw_glc;	vwp->vw_glc = glp;
	if (vwp->vw_autoverbose)
	  {
		fprintf(vwp->vw_fpl,
			" %3d: %3d %3d %3d %3d %3d %3d %3d: %5d %4d for ",
			hli, erct[0], erct[1], erct[2], erct[3],
			erct[4], erct[5], erct[6], nprospects, nimprove);
		pr3m(vwp->vw_fpl, rfp->rfr_sinv, "sinv=");
	  }
	for (h2err = 0; h2err < 7; h2err++)
		vwp->vw_totpro[h2err] += erct[h2err];
	return nprospects;
}

static int use_dv(DIFFVECTOR *dp0, DIFFVECTOR *dp1, DIFFVECTOR *dp2,
	float *spaci, REFINERUN *rfp)
{ /*	This constructs the appropriate matrices from three difference vectors.
	It returns a positive integer if successful, zero otherwise. The
	integer will be the maximum value of the largest index in the array. */
	int	i, j, minhere, maxhere, ip1, ip2, retlim, hmaxi;
	double	bsof, dotp, sinorm;
	float	vref[3][4];
	REFPARS	*xp;

	for (i = 0; i < 3; i++) vref[i][3] = 0.;
	for (i = 0; i < 3; i++)
	   {
		vref[0][i] = bsof = dp0->ss[i];	vref[0][3] += bsof * bsof;
		vref[1][i] = bsof = dp1->ss[i];	vref[1][3] += bsof * bsof;
		vref[2][i] = bsof = dp2->ss[i];	vref[2][3] += bsof * bsof;
	   }
	/* reorder the reference difference vectors so that their inverse
	norms fall in the same order as the unit cell lengths themselves. */
	rfp->rfr_maxorig = (rfp->rfr_a0[0] > rfp->rfr_a0[1] ?
	 (rfp->rfr_a0[0] > rfp->rfr_a0[2] ? 0 : 2 ) :
		(rfp->rfr_a0[1] > rfp->rfr_a0[2] ? 1 : 2));
	rfp->rfr_minorig = (rfp->rfr_maxorig + 1) % 3;
	i = 3 - rfp->rfr_maxorig - rfp->rfr_minorig;
	if (rfp->rfr_a0[i] < rfp->rfr_a0[rfp->rfr_minorig])
		rfp->rfr_minorig = i;
	maxhere = (vref[0][3] < vref[1][3] ? (vref[0][3] < vref[2][3] ?
		0: 2 ) : (vref[1][3] < vref[2][3] ? 1 : 2));
	if (maxhere != (int)(rfp->rfr_maxorig))
		swap((void *)vref[rfp->rfr_maxorig], (void *)vref[maxhere],16);
	minhere = (vref[0][3] > vref[1][3] ? (vref[0][3] > vref[2][3] ?
		0: 2 ) : (vref[1][3] > vref[2][3] ? 1 : 2));
	if (minhere != (int)(rfp->rfr_minorig))
		swap((void *)vref[rfp->rfr_minorig], (void *)vref[minhere],16);
	/* now construct the 3x3 matrix containing these averaged reference
	scattering-vector differences */
	for (i = 3; i--; ) for (j = 3; j--; ) rfp->rfr_sref[i][j] = vref[j][i];
	/* make sure the determinant of the matrix comprising the three
	scattering vector differences is positive.  If it's negative,
	we flip the signs of all the elements in a column of "sref".
	That's legal... if (s) is a legal scattering vector, so is (-s) */
	if ((deter(rfp->rfr_sref) * 0.1e6) < -DFUZZ)
	  for (i = 3; i--; )
	     {
		rfp->rfr_sref[i][0] = -(rfp->rfr_sref[i][0]);
		vref[0][i] = -vref[0][i];
	     }
	/* now get the inverse of this matrix to use in hkl checks */
	if (inv3(rfp->rfr_sref, rfp->rfr_sinv)) return 0;
	/* get a normalized version of `vref' for convenience */
	xp = rfp->rfr_xp;
	for (i = 3; i--; )
	   {
		vref[i][3] = sqrt(vref[i][3]);
		for (j = 3; j--; ) vref[i][j] /= vref[i][3];
		spaci[i] = xp->xlambda / vref[i][3];
	   }
	for (i = 0; i < 3; i++)
	   {
		ip1 = (i + 1) % 3;	ip2 = 3 - i - ip1;
		for (j = 3, dotp = 0; j--; )
		   dotp += vref[ip1][j] * vref[ip2][j];
		spaci[i + 5] = acos(dotp) / RADC;
	   }
	/* establish minima and maxima for the index search */
	rfp->rfr_maxorig = (rfp->rfr_a0[0] > rfp->rfr_a0[1] ?
	 (rfp->rfr_a0[0] > rfp->rfr_a0[2] ? 0 : 2) :
	 (rfp->rfr_a0[1] > rfp->rfr_a0[2] ? 1 : 2));
	rfp->rfr_minorig = (rfp->rfr_maxorig + 1) % 3;
	i = 3 - rfp->rfr_maxorig - rfp->rfr_minorig;
	if (rfp->rfr_a0[i] < rfp->rfr_a0[rfp->rfr_minorig])
		rfp->rfr_minorig = i;
	for (i = 3, retlim = 0; i--; )
	   {
		sinorm = sqrt(rfp->rfr_sref[0][i] * rfp->rfr_sref[0][i] +
			rfp->rfr_sref[1][i] * rfp->rfr_sref[1][i] +
			rfp->rfr_sref[2][i] * rfp->rfr_sref[2][i]);
		hmaxi = rfp->rfr_amax[rfp->rfr_maxorig] * sinorm + FUZZ;
		if (hmaxi > IND3MAX) hmaxi = IND3MAX;
		else if (hmaxi < 1) hmaxi = 1;
		if (hmaxi > retlim) retlim = hmaxi;
	   }
	return retlim;
}

static int scan_prospects(DIFFVECTOR *dv0e, DIFFVECTOR *dv1s,
	float spacings[], double miasc, REFINERUN *rfp, VECWORK *vwp)
{ /*	This goes through ordered triples of difference vectors and looks
	for likely prospects. It returns when it's gone through
	the specified ranges (outer loop dvlim[0] through dv0e; middle loop
	dv1s +1 through dv0e+1; inner loop dvp1 +1 through dvlim[1]).
	It returns the number of prospects it finds. */
	int		overhlim, tprospects, nfound;
	DIFFVECTOR	*dvp0, *dvp1, *dvp1s, *dvp2, *dvp1e, *dvp2e;

	dvp1e = dv0e + 1;
	if (dvp1e > vwp->vw_dvlim[1] - 1) dvp1e = vwp->vw_dvlim[1] - 1;
	vwp->vw_overhmax = 0;	vwp->vw_fpl = rfp->rfr_fpl;
	for (tprospects = 0, dvp0 = vwp->vw_dvlim[0]; dvp0 <= dv0e; dvp0++)
	 {
	   dvp1s = dv1s;	if (dvp1s <= dvp0) dvp1s = dvp0 + 1;
	   for (dvp1 = dvp1s; dvp1 <= dvp1e; dvp1++)
	     {
		if (angscore(dvp0, dvp1) > miasc)
		 {
		   dvp2e = dvp1e + 1;
		   if (dvp2e > vwp->vw_dvlim[1]) dvp2e = vwp->vw_dvlim[1];
		   for (dvp2 = dvp1 + 1; dvp2 <= dvp2e; dvp2++)
		    {
			if (angscore(dvp2, dvp0) > miasc)
			 {
			   if (angscore(dvp1, dvp2) > miasc)
			    {
				if (0 < (overhlim =
				 use_dv(dvp0, dvp1, dvp2, spacings, rfp)))
				  {
					nfound = getprospects(overhlim, rfp,
					 vwp);
					tprospects += nfound;
					if (overhlim > vwp->vw_overhmax)
						vwp->vw_overhmax = overhlim;
				  } /* end of "if (0 < (ov" clause */
			     } /* end of "if (angscore(dvp1" clause */
			  } /* end of "if (angscore(dvp2" clause */
		    } /* end of "for (dvp2" loop */
		 } /* end of "if (angscore(dvp0, dvp1)" clause */
	     } /* end of "for (dvp1" loop */
	    if (vwp->vw_autoverbose) fprintf(rfp->rfr_fpl,
		"prospect vector %6d: %6d prospects so far\n",
		(int)(dvp0 - vwp->vw_dvlim[0]), tprospects);
	 } /* end of "for (dvp0" loop */
	return tprospects;
}

int diffm3(float *spacings, double minbasis, REFINERUN *rfp,VECWORK *vwp)
{ /*	This goes through the partially- (or fully-)constructed list of
	difference vectors and attempts to find three very noncoplanar vectors,
	each of which occurs at least NDUP times within the list.
	It returns >= 1 if and only if it succeeds in finding said matrix.
	Once it's found the matrix, it puts it in a crystallographically
	canonical form, inverts it, and prints some useful facts about it.
	The algorithm now explicitly goes through the list
	systematically to find a satisfactory scoring function. */

	int		i, totprospects, totposs, batchno, nbatches;
	time_t		tlc;
	double		maxcsq, minasc;
	REFPARS		*xp;
	DIFFVECTOR	*dvp0e, *dvp1s;

	vwp->vw_fpl = rfp->rfr_fpl;
	if (vwp->vw_dvlim[1] < vwp->vw_dvlim[0] + 3)
	  {
		fprintf(rfp->rfr_fpl,
		 "Only %d difference vectors in squished list\n",
			(int)(vwp->vw_dvlim[1] - vwp->vw_dvlim[0]) - 1);
		return -3;
	  }
	vwp->vw_spend = (vwp->vw_dvlim[0])->dnsq - vwp->vw_dvlim[1]->dnsq;
	if (vwp->vw_spend > -DFUZZ)
	  {
		fprintf(rfp->rfr_fpl,
 "Not enough spread between lowest and highest difference vector\n");
		return -4;
	  }
	vwp->vw_spena = STLDPEN / vwp->vw_spend;
	vwp->vw_spenb = (STLPENLO * (vwp->vw_dvlim[0])->dnsq -
		STLPENHI * (vwp->vw_dvlim[1])->dnsq) / vwp->vw_spend;
	vwp->vw_nmpena = NMDPEN / ((double)(NDUPMAX - NDUP));
	vwp->vw_nmpenb = (NMPENHI * NDUPMAX - NMPENLO * NDUP) /
		((double)(NDUPMAX - NDUP));
	/* the penalty function is STLPENHI for the smallest
	 diff vector, STLPENLO for the largest. */
	dvp0e = vwp->vw_dvlim[1] - 2;
	/* save the unmodified cell lengths and angles */
	xp = rfp->rfr_xp;
	for (i = 0; i < 3; i++) rfp->rfr_a0[i] = xp->xabc[i];
	init_prospects(vwp);
	maxcsq = cos(minbasis * RADC);	maxcsq *= maxcsq;
	minasc = 1. - maxcsq;
	/* we divide the difference vectors into batches and look through
	 the batches more or less systematically */
	totposs = (vwp->vw_dvlim[1] - vwp->vw_dvlim[0]) + 1;
	if (totposs < 8) nbatches = 1;
	else nbatches = 1 + totposs / 8;
	printf(
	 " %6d grouped difference vectors to be searched in %5d batches\r\n",
	 totposs, nbatches);
	fprintf(rfp->rfr_fpl,
	 " %6d grouped difference vectors to be searched in %5d batches\n",
	 totposs, nbatches);
	xeflush(stdout);
	dvp0e = vwp->vw_dvlim[0] + totposs / nbatches;
	dvp1s = vwp->vw_dvlim[0] + 1;
	tlc = NOW;	strcpy(rfp->rfr_rundate, ctime(&tlc));
	for (totprospects = batchno = 0; batchno < nbatches; batchno++)
	   {
		if (dvp1s >= vwp->vw_dvlim[1] - 1) break;
		totprospects += scan_prospects(dvp0e,
		 dvp1s, spacings, minasc, rfp, vwp);
		if (totprospects > ENOUGH_PROSPECTS) break;
		dvp1s = dvp0e + 1;
		dvp0e += totposs / nbatches;
		if (dvp0e > vwp->vw_dvlim[1] - 2) dvp0e = vwp->vw_dvlim[1] - 2;
	   }
	fprintf(rfp->rfr_fpl,
	 "Solutions found after batch %2d, with diff.vectors %4d to %4d\n",
	 batchno, ((int)(dvp1s - vwp->vw_dvlim[0]) - 1),
		(int)(dvp0e - vwp->vw_dvlim[0]));
	fprintf(rfp->rfr_fpl,
	 " Usable    Too    Too Redun-     No    Co-    Bad\n");
	fprintf(rfp->rfr_fpl,
	 "Vectors  Short   Long   dant Integ. planar  Resid\n");
	fprintf(rfp->rfr_fpl, " %6d %6d %6d %6d %6d %6d %6d\n",
	 vwp->vw_totpro[0], vwp->vw_totpro[1], vwp->vw_totpro[2],
	 vwp->vw_totpro[3], vwp->vw_totpro[4], vwp->vw_totpro[5],
	 vwp->vw_totpro[6]);
	tlc = NOW;	strcpy(rfp->rfr_rundate, ctime(&tlc));
	totprospects = sift_prospects(totprospects, MAXTOKEEP, vwp);
	/* Having generated list of prospects, go through 'em in groups of 3*/
	rfp->rfr_ndecent = prospects3(totprospects, rfp, vwp);
	if (!(rfp->rfr_ndecent)) return -7;
	if (rfp->rfr_ndecent > 1) qsort((void *)(rfp->rfr_gu0),
	 (size_t)(rfp->rfr_ndecent), sizeof (GUESS), (QPE)guesscmp);
	fprintf(rfp->rfr_fpl,"Largest value of hmax checked = %d\n",
		vwp->vw_overhmax);
	return rfp->rfr_ndecent;
}

/* end of prospects.c */
