/* refdriver.c -- top-level routines for non-Curses refinement

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	19 Apr 2007: float->double conversions; reorganize use of "{"
	28 Nov 2006: showbravais call updated
	 3 Nov 2006: provide for excluding ice-rings during autoindexing
	 9 Jun 2006: bravaisspec,acav2str,refjournal,updateorigv,ref1log
		moved here from other files: all are now static
	19 May 2006: merge ref1main.c and ref_all.c into this file
	26 Oct 2005: replace flush_xgbuf() call with xgbuf() call
	15 Oct 2005: replace end_xgbuf() with refjournal(); housekeeping
	13 Oct 2005: housekeeping
	17 Jun 2005: need_indexing(): xgaccess() takes verbosity argument
	13 May 2005: handle --help and --version correctly for flagproc()
	30 Jan 2005: replace isptosys() with spgtosys() to be consistent
	24 Jan 2005: ref_all: don't force -i<spacegroup> to work if really bad
	20 Jan 2005: ref_all: manage Bravais-lattice determination more
	 carefully; enable Bravais calculation under appropriate circumstances
	 ASKBRAVAIS and ASKSPGRP bits in rfr_flags element used appropriately
	24 Oct 2004: ref_all: better diagnostics to logfile and stdout;
	 -m flag uses an element of the REFINERUN structure;
	20 Oct 2004: ref_all: -s flag added: program prompts for new spacegroup;
	 routines spquery() and isptosys() added to facilitate that
	15 Oct 2004: sources moved from ../nogui to ../utilities
	27 Aug 2004: routines used only here are declared as static;cosmetics
	18 Aug 2004: ref_all: static globals converted to function args
	18 Aug 2004: ref1main: static globals made local; origva and mbv zeroed
	15 Jun 2004: ref_all: -n flag (don't recenter after autoindexing) added
	 7 Jun 2004: increase autoindexing error limits to 80%, 39%
	 5 Jun 2004: ref_all: eliminate RR_CHANGEDDET flag requirement;
		-f flag flips sign of stepsize;
		REFINERUN structure zeroed in refa_init(), not refinit()
	 4 Jun 2004: zero REFINERUN structure in ref1main, not refinit()
	 1 Jun 2004: handle "b q" arg in ref1args(): queries Bravais choice
	31 May 2004: ref1main: move most routines to ../xgrefine/refio.c
	29 May 2004: initial version of ref_all, derived from ref1main.c
	21 Jan 2004: license message inserted
( Only changes to routines that stayed here after 31 May 04 are listed below:)
	25 Nov 2003: replace refinement globals with REFINERUN structure elts.
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	17 Aug 2003: ref1args: rename local pointer `substri'
	 5 Jan 2003: reorganized source-file header
	15 Nov 2002: extern declarations made here rather than in refiext.h
	25 Mar 2002: removed MSI attachments
	18 May 2001: initial version of ref1main

   Purpose: toplevel routines for single-cycles and automated refinement.
	ref1main	does setup operations and on a cycle-by-cycle basis
		it calls routines found in other files.
	ref_all		performs automated refinements
   Cross references: refio.c, refrefine.c, refutil.c
*************************************************************************/
#define		NEED_UNISTD	1
#include	"refidefs.h"
#define		ENOUGH_TO_INDEX	100

/**************************************
	ref1main: handle single-command-at-a-time approaches to refinement
**************************************/

static int bravaisspec(char *subs)
{ /* This returns an integer between -45 and +16.
	positive values 1-16 correspond to the integer values a-p in the
	index list output from a "refine1 -b" command. 0 corresponds
	to an error. The positive value 17 is a signal indicating that
	we should choose the best lattice, i.e. the highest-symmetry lattice
	that delivers a low residual
 -1 to -45 correspond to the lattice characters
	1 to 44 in the International Tables and the fake character 45,
	which corresponds to a rhombohedral cell indexed rhombohedrally */
	register int	j, latno;
	char		nextc, subc;

	if (allnumeric(subs)) return atoi(subs);
	if ((NULL == subs) || ('\0' == *subs)) return 0;
	latno = 0;	subc = *subs;
	if (strlen(subs) < 2) /* single-letter transform */
	  { /* if the user specifies a number, we save its
		 negative for later */
		if (('1' <= subc) && (subc <= '9')) /* numeric char */
		  {
			j = subc - '0';	latno = -j;
		  }
		else { /* otherwise save the number itself */
			if ((subc < 'A') || (subc > 'Q')) return EBAD;
			 /* bugfix?  latno = 1 + (int)(subc - 'A'); */
			 	latno = (int)(subc - 'A');
		     }
	  }
	else { /* two-letter or two-digit specification */
		nextc = *(subs + 1);
		if (('1' <= subc) && (subc <= '4') &&
		  ('0' <= nextc) && (nextc <= '9')) /* 2-digit lattchar*/
		  {
			j = 10 * ((int)(subc - '0')) + (int)(nextc - '0');
			latno = -j;
		  }
		else {	/* 2-letter lattice character */
			if (subc == 'C') /* cubic */
			  {
				if (nextc == 'F') latno = -1;
				else if (nextc == 'P') latno = -3;
				else if (nextc == 'I') latno = -5;
			  }
			else if (subc == 'H') /* hexagonal */
			  {
				if (nextc == 'R') latno = -2;
				else if (nextc == 'P') latno = -12;
			  }
			else if (subc == 'T') /* tetragonal */
			  {
				if (nextc == 'I') latno = -6;
				else if (nextc == 'P') latno = -11;
			  }
			else if (subc == 'O') /* orthorhombic */
			  {
				if (nextc == 'I') latno = -8;
				else if (nextc == 'C') latno = -13;
				else if (nextc == 'F') latno = -16;
				else if (nextc == 'P') latno = -32;
			  }
			else if (subc == 'M') /* monoclinic */
			  {
				if (nextc == 'C') latno = -10;
				else if (nextc == 'I') latno = -43;
				else if (nextc == 'P') latno = -33;
			  }
			else if ((subc == 'R') && (nextc == 'R'))
				latno = -45;	/* rhombohedral */
			else if ((subc == 'A') && (nextc == 'P'))
				latno = -31;	/* triclinic */
		      }
	     }
	return latno;
}

static void acav2str(int ac, char *av[], char *outstr)
{ /*	This converts an (argc,argv)-type list into a single newline-
	terminated string and puts it in outstr, which must be >= 256 bytes */
	int		ic;
	char		**lav;
	size_t		leno;

	for (leno = ic = 0, lav = av; ic < ac; ic++, lav++)
		leno += strlen(*lav) + 1;
	leno++;	if (leno < 256) leno = 256;
	zeromem((AVOID *)outstr, leno);
	for (ic = 0, lav = av; ic < ac; ic++, lav++)
	   {
		strcat(outstr, *lav); if (ic < (ac - 1)) strcat(outstr, " ");
	   }
	strcat(outstr, "\n");
}

static int refjournal(char *prname)
{ /*	This either starts or ends a session of writing to the X-GEN journal
	from the refinement formalisms. If the argument is NULL, we assume that
	we're closing things off; if it's not NULL, we assume we're starting */
	static char		locjb[64];
	static pid_t		currpid;

	if (0 == (int)currpid) currpid = getpid();
	if ((NULL == prname) || ('\0' == *prname)) /* ending */
		sprintf(locjb, "EOF_%d\n", (int)currpid);
	else sprintf(locjb, "%s << EOF_%d\n", prname, (int)currpid); /* start*/
	return xgbuf(locjb, 1);
}

static sichar ref1args(int ac, char *av[], int *intresp, double *flresp)
{ /*	This processes command-line arguments for the one-line form of
	refinement. It returns the command argument; if the command
	is faulty or unrecognized, it returns -1 or 0.
	Various integer values are returned in *intresp;
	Floating-point values are returned in *flresp.
	This routine is sometimes invoked from within an interactive
	sequence within the program, and other times directly from the
	command line. In the former case ac < 0. */
	int		comslot, internal, j;
	char		*cp, *substri;
	double		specval;
	sichar		comd, subc;
	static char	*comstr, *shname;
	static char	legal[] = " ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
	static size_t	comlen, oldcomlen;
	static char	commsg[] = "comstr in ref1args";

	if (NULL == av) /* special case: wipe out memory */
	  {
		memzap((AVOID **)&comstr, commsg);	oldcomlen = 0;
		return 0;
	  }
	*intresp = 0; /* default returned-integer value */
	*flresp = 0.; /* default returned-floating-point value */
	/* no argument at all is like "-n", i.e. iNteractive mode.
	 Any time we're in iNteractive mode we have the journal receive
	 a command that looks like "refine1 -n << EOF_xxxx" */
	if (ac < 0)
	  {
		ac = -ac;	internal = 1;
	  }
	else internal = 0;
	if (!internal && (ac < 2))
	  {
		shname = filepart(*av, 2); (void)xgprogname(shname, 4);
		return (EOK == refjournal(*av)) ? 'N' : (sichar)(-1);
	  }
	comslot = 0;
	if (av[1][0] == '-') comslot = 1;
	comd = av[1][comslot];
	if (('a' <= comd) && (comd <= 'z')) comd -= 'a' - 'A';
	for (cp = legal; '\0' != *cp; cp++) if (comd == *cp) break;
	/* require that the argument be recognized */
	if ((comd == 0) || ('\0' == *cp)) return (sichar)(0);
	/* for certain commands there need to be arguments.
	 'D' and 'J' are special cases: 'D' should take a numeric argument
	 and 'J' should take an alphabetic argument. If we goof that up,
	 it's probably because we meant to use the 'D' when we specified 'J'
	 and vice versa. */
	if ((comd == 'D') || (comd == 'J'))
	  {
		if (ac < 3) return -2;	cp = &(av[3][0]);
		/* fix commands that look like "d alpha n" */
		if ((comd == 'D') && ((*cp == 'n') || (*cp == 'N') ||
		 (*cp == 'y') || (*cp == 'Y'))) av[1][comslot] = comd = 'J';
		/* fix commands that look like "j beta 90" */
		if ((comd == 'J') && (NULL != cp) && ('\0' != *cp) &&
		 (allnumeric(cp))) av[1][comslot] = comd = 'D';
	  }
	else if (comd == 'B') /* do Bravais lattice calcs:
			"B Q" means "show solutions and prompt for choice" */
		*intresp = (ac >= 3); /* any arg will work here */
	else if (comd == 'C') /* choose an auto-indexing solution */
		*intresp = (ac < 3) ? 0 : atoi(av[2]);
	else if (comd == 'F') /* Bravais specification */
	  { /* pick out, uppercaseify, and use the Bravais string */
		substri = (ac < 3) ? &(av[1][comslot + 1]) : &(av[2][0]);
		for (cp = substri; '\0' != *cp; cp++)
			if (('a' <= *cp) && (*cp <= 'z')) *cp -= 'a' - 'A';
		*intresp = bravaisspec(substri);
	  }
	else if (comd == 'I') /* autoindexing params */
	  {
		flresp[0] = 0.2; flresp[1] = 0.2; flresp[2] = 50.;
		if (ac > 4)
		  {
			specval = atof(av[4]);
			if ((0. <= specval) && (specval < 1.))
				flresp[2] = specval * 100.;
			else if (specval < 100.)
				flresp[2] = specval;
		  }
		if (ac > 3)
		  {
			specval = atof(av[3]);
			if ((0. <= specval) && (specval <= 1.0))
				flresp[1] = specval;
		  }
		if (ac > 2) 
		  {
			specval = atof(av[2]);
			if ((0. <= specval) && (specval <= 1.0))
				flresp[0] = specval;
		  }
	  }
	else if ((('K' <= comd) && (comd <= 'L')) ||
	 ('R' == comd) || (('X' <= comd) && (comd <= 'Z')))
	  {
		*intresp = 1; /* default is one pass */
		if (ac < 3) substri = &(av[1][comslot+1]);
		else substri = &(av[2][0]);
		if ((NULL != substri) && ('\0' != *substri))
		  {
			if ((0 < (j = atoi(substri))) && (j < 100))
				*intresp = j;
			else *intresp = 100;
		  }
	  }
	else if (comd == 'T')	/* transform: legal values are A-J or nothing*/
	  {
		if (ac < 3) subc = av[1][comslot+1];
		else subc = av[2][0];
		if (('a' <= subc) && (subc <= 'z')) subc -= 'a' - 'A';
		if ((subc < 'A') || (subc > 'J')) *intresp = 0;
		else *intresp = 1 + (int)(subc - 'A');
	  }
	/* if we've gotten here within the interactive mode,
	 we need to record the name of the invoking program
	 and write to the buffer for the journal file. If the user
	 has simply asked for the list of commands, we don't journal that. */
	comlen = 260;
	if ((NULL == comstr) || (comlen > oldcomlen))
	  {
		memzap((AVOID **)&comstr, commsg);	oldcomlen = comlen;
		if (NULL == (comstr = (char *)mallop(comlen, commsg)))
			return EALLOC;
	  }
	if (internal)
	  {
		if ('V' != comd)
		  {
			acav2str(ac - 1, av + 1, comstr);
			if (EOK != xgbuf(comstr, 1)) return EBAD;
		  }
		return comd;
	  }
	else {
		shname = filepart(*av, 2);	(void)xgprogname(shname, 4);
	     }
	/* if comd is either 'N' or blank, we're embarking on an
	 interactive run. Note that appropriately in the journal file */
	if (('N' == comd) || (' ' == comd))
		return (EOK == refjournal(*av)) ? 'N' : (sichar)(-1);
	else {
		acav2str(ac, av, comstr);
		if (EOK != xgbuf(comstr, 1)) return EBAD;
	     }
	return comd;
}

static void updateorigv(float *origv, REFPARS *xp)
{ /*	This sets the values in the `origv' array to their current values */
	int	i;
	float	*flp;

	for (i = 0, flp = origv; i < 3; i++, flp++) *flp = xp->xabc[i];
	for (i = 0; i < 3; i++, flp++) *flp = xp->xang[i];
	*flp++ = xp->xdelom;	*flp++ = xp->xdelchi;	*flp++ = xp->xdelphi;
	*flp++ = xp->xplated;	*flp++ = xp->xx0;	*flp++ = xp->xy0;
	*flp++ = xp->xtilt / RADC;	*flp++ = xp->xswing / RADC;
	*flp++ = xp->xgam[0] / fabs(xp->xfrmsize);
	*flp++ = xp->xgam[1] / fabs(xp->xfrmsize);
	*flp = xp->xgam[2];
}

static int ref1init(int refla, float *origv)
{ /*	startup for one-line refinements. If refla & RR_FORCEREAD is nonzero,
	the routine is supposed to read in observation centroids;
	origv is the vector into which the original values of the floating-
	point variables and parameters are to be stuffed. */
	int		nnot, ret;
	REFINERUN	*rfp;

	if (EOK != (ret = refinit(refla))) return ret;
	if (NULL == (rfp = managerfp(3, 0))) return EALLOC;
		/* remember current floating-point params */
	updateorigv(origv, rfp->rfr_xp);
	if (refla & RR_FORCEREAD)
	  {
		nnot = gpeaks(rfp);
		fprintf(rfp->rfr_fpl, "Read in %6d good centroids, ",
			(int)(rfp->rfr_npeaks));
		fprintf(rfp->rfr_fpl,
		 "%5d useless centroids from input file\n", nnot);
	  }
	return EOK;
}

static void ref1log(FILE *fp, char *str, float *oriv,
 float *mainb, REFINERUN *rfp)
{ /*	This writes the current status of the refinement to the stream fp.
	`str' contains an introductory string; oriv contains the
	original values of the floating-point variables and parameters.
	The mainbeam coordinates are input in mainb[0], mainb[1]. */
	register int	i, j;
	REFPARS		*xp;
	static char	spacer[] = "                       ";
	static char	dgt[] = "D>";
	static char	zindx[] = "hkl";

	if ((NULL != str) && ('\0' != *str)) fprintf(fp, "%s\n", str);
	if (NULL == (xp = rfp->rfr_xp)) return;
	fprintf(fp,
	 "Refining? a:%c     b:%c     c:%c alpha:%c  beta:%c gamma:%c",
		xp->xisref[0], xp->xisref[1], xp->xisref[2],
		xp->xisref[3], xp->xisref[4], xp->xisref[5]);
	fprintf(fp, " omega:%c   chi:%c   phi:%c\n",
		xp->xisref[6], xp->xisref[7], xp->xisref[8]);
	fprintf(fp, "Value:%7.3f%8.3f%8.3f%8.3f%8.3f%8.3f%8.3f%8.3f%8.3f\n",
		xp->xabc[0], xp->xabc[1], xp->xabc[2],
		xp->xang[0], xp->xang[1], xp->xang[2],
		xp->xdelom, xp->xdelchi, xp->xdelphi);
	fprintf(fp, "Shift:%7.4f%8.4f%8.4f%8.4f%8.4f%8.4f%8.4f%8.4f%8.4f\n",
	 xp->xabc[0] - oriv[0], xp->xabc[1] - oriv[1], xp->xabc[2] - oriv[2],
	 xp->xang[0] - oriv[3], xp->xang[1] - oriv[4], xp->xang[2] - oriv[5],
	 xp->xdelom - oriv[6], xp->xdelchi - oriv[7], xp->xdelphi - oriv[8]);
	fprintf(fp, "     plateD:%c  Xcen:%c  Ycen:%c  tilt:%c swing:%c",
		xp->xisref[9], xp->xisref[10], xp->xisref[11],
		xp->xisref[12], xp->xisref[13]);
	fprintf(fp, "  gam0:%c  gam1:%c  gam2:%c\n",
		xp->xisref[14], xp->xisref[15], xp->xisref[16]);
	fprintf(fp, "Value:%7.4f%8.4f%8.4f%8.4f%8.3f%8.3f%8.3f%8.4f\n",
		xp->xplated, xp->xx0, xp->xy0, xp->xtilt / RADC,
		xp->xswing / RADC, xp->xgam[0] / fabs(xp->xfrmsize),
		xp->xgam[1] / fabs(xp->xfrmsize), xp->xgam[2]);
	fprintf(fp, "Shift:%7.4f%8.4f%8.4f%8.4f%8.4f%8.4f%8.4f%8.4f\n",
		xp->xplated - oriv[9], xp->xx0 - oriv[10], xp->xy0 - oriv[11],
		xp->xtilt / RADC - oriv[12], xp->xswing / RADC - oriv[13],
		xp->xgam[0] / fabs(xp->xfrmsize) - oriv[14],
		xp->xgam[1] / fabs(xp->xfrmsize) - oriv[15],
		xp->xgam[2] - oriv[16]);
	fprintf(fp, "   Max Errors Allowed in:        Resolution");
	fprintf(fp, "     Wave- Stepsiz Crystal   Space\n");
	fprintf(fp, " index X(pix)  Y(pix) phi(fm)     Min     Max");
	fprintf(fp, "  length degrees  System   Group\n");
	fprintf(fp, "%6.4f%7.4f %7.4f %7.4f %7.4f %7.4f %7.4f %7.4f %7d %7d\n",
		xp->xerrind, xp->xerrx, xp->xerry, xp->xerrphi,
		xp->xresmin, xp->xresmax, xp->xlambda,
		xp->xfrmsize / RADC, (int)(xp->xisys), (int)(xp->xispgp));
	fprintf(fp, "True start omega:%8.3f  chi:%8.3f phi:%8.3f",
		xp->xstart[0], xp->xstart[1], xp->xstart[2]);
	fprintf(fp, " Mainbeam: %6.1f %7.1f\n", mainb[0], mainb[1]);
	fprintf(fp, "   -----Root-Mean-Squared Errors In--------");
	fprintf(fp, "    <#frms    # of    # of\n");
	fprintf(fp, " Index      X       Y     phi   Gamma Profile");
	fprintf(fp, "  moved>   refls indices\n");
	fprintf(fp, "%6.4f%7.4f%8.4f%8.4f%8.4f%8.4f%8.4f %7d %7d\n",
		rfp->rfr_indrmse, rfp->rfr_xrmse, rfp->rfr_yrmse,
		rfp->rfr_phirmse, rfp->rfr_gamrmse, rfp->rfr_profrmse,
		rfp->rfr_avedelfm, (int)(rfp->rfr_nreftot),
		3 * (int)(rfp->rfr_nreftot));
	fprintf(fp, "%6.4f%7.4f%8.4f%8.4f %s %7d %7d\n",
		rfp->rfr_indurmse, rfp->rfr_xurmse, rfp->rfr_yurmse,
		rfp->rfr_phiurmse, spacer,
		(int)(rfp->rfr_nrefok), (int)(rfp->rfr_nindok));
	fprintf(fp, "     ");
	for (i = 0; i < 8; i++)
		fprintf(fp, " %s%5.2f", dgt, rfp->rfr_shldmin[i]);
	fprintf(fp, " Shells\n     ");
	for (i = 0; i < 8; i++) fprintf(fp, " %7.4f", rfp->rfr_shlrmse[i]);
	fprintf(fp, " Err(h)\n     ");
	for (i = 0; i < 8; i++) fprintf(fp, " %3d/%3d",
		(int)(rfp->rfr_shlnref[i]), (int)(rfp->rfr_shlnind[i]));
	fprintf(fp, " ind/ref\n Index err(h)> ");
	for (i = 0; i < 10; i++) fprintf(fp, " %5.3f", 0.05 * ((double)i));
	for (j = 0; j < 3; j++)
	   {
		fprintf(fp, "\n     %c         ", zindx[j]);
		for (i = 0; i < 10; i++)
			fprintf(fp, "%6d", (int)(rfp->rfr_ncateg[j][i]));
	   }
	fprintf(fp, "\n");	(void)xeflush(fp);
}

static void updateall(char comd, int iarg, REFINERUN *rfp,
	float *oriv, float *mainb)
{ /*	This updates various parameters based on which our current
	command is. `report' & 01 determines whether the results
	are written to stdout; `report' & 02 determines whether
	the results are written to rfp->rfr_fpl. */
	int		coint, report;
	static int	irep[] = { 1, 0, 1, 1, 3, 1, 1, 1, 1, 0,
				   1, 1, 1, 0, 0, 1, 0, 1, 1, 4,
				   1, 0, 0, 1, 1, 1 };

	if (('0' <= comd) && (comd <= '9')) coint = 0; /* '0' ->'9' == 'A' */
	else if (('A' <= comd) && (comd <= 'Z')) coint = comd - 'A';
	else return; /* shouldn't ever happen */
	report = irep[coint]; /* decide whether and how to report */
	/* 'T' command is special case: report only if we had an arg */
	if (4 == report) report = (iarg > 0);
	if (!report) return;
	get_rmserrs("", rfp);	getmainbeam(rfp->rfr_xp, mainb);
	if (report & 01) ref1log(stdout,       "", oriv, mainb, rfp);
	if (report & 02) ref1log(rfp->rfr_fpl, "", oriv, mainb, rfp);
	updateorigv(oriv, rfp->rfr_xp);
}

static int ref1cmd(char *str, REFINERUN *rfp)
{ /*	This interprets the string `str' to be a refinement command */
	int		azc, intar, lenin, resu;
	double		flar[3];
	char		*cpi, *cpo, **azv, **lazv;
	sichar		co;
	static char	*sout;
	static float	mbv[2], origv[20];
	static size_t	newlen, oldlen;
	static char	soutmsg[] = "sout in ref1cmd";

	if (NULL == str) /* special case: eliminate memory for "sout" */
	  {
		memzap((AVOID **)&sout, soutmsg);	return EOK;
	  }
	/* convert the string to an argc, argv pair */
	if ((NULL == str) || ('\0' == *str)) return EBADVALUE;
	for (lenin = 0, cpi = str; '\0' != *cpi; cpi++) lenin++;
	newlen = lenin + 3;	if (newlen < 80) newlen = 80;
	if ((NULL == sout) || (newlen > oldlen))
	  {
		memzap((AVOID **)&sout, soutmsg);	oldlen = newlen;
		if (NULL == (sout = mallop(newlen, soutmsg))) return EALLOC;
	  }
	zeromem((void *)sout, newlen);
	for (cpo = sout, cpi = str; '\0' != *cpi; cpi++) *cpo++ = *cpi;
	*cpo = '\0';	*(cpo + 1) = '\0';
	if (NULL == (lazv = str2acav(sout, &azc))) return EALLOC;
	azc--;	/* to correct for an anomaly in str2acav */
	flar[0] = flar[1] = flar[2] = intar = 0;	azv = lazv;
	co = ref1args(-azc, azv, &intar, flar); /* translate that into args */
	/* we treat 'N' and blank specially */
	if (('N' == co) || (' ' == co)) return EOK;
	/* otherwise, we simply call the translation routine */
	resu = refco(co, azc, azv, intar, flar, rfp);
	/* in many instances we update the display */
	updateall(co, intar, rfp, origv, mbv);
	/* get rid of temp memory and return */
	if (EOK != resu) return resu;
	return (('Q' == co) ? 1 : ('E' == co) ? 2 : resu);
}

static int keepinter_ref(REFINERUN *rfp)
{ /*	The interactive version of the one-command-at-a-time stuff.
	The return value is EOK iff we're supposed to keep going */
	char	klin[258];

	if (isatty(0)) printf("ref> ");	/* print a nice prompt */
	/* go get a string from the user and make sure it isn't empty */
	zeromem((AVOID *)klin, 258);
	if (NULL == fgets(klin, 255, stdin)) return EOK;
	if ((klin[0] == '\0') || (klin[0] == '\n')) return EOK;
	return ref1cmd(klin, rfp);
}

void end_refinement(void)
{ /*	This closes files and frees memory for refinement */
	int		ni;
	double		nf;
	REFINERUN	*rfp;

	(void)ref1cmd(NULL, NULL);	(void)make_incrvecs(NULL);
	(void)ref1args(0, NULL, &ni, &nf);	(void)xgbuf(NULL, 0);
	apply_dels(NULL);	/* destroy memory for linear-refinement stuff*/
	if ((NULL != (rfp = managerfp(3, 0))) && (NULL != rfp->rfr_fpl))
	  {
		fclose(rfp->rfr_fpl);	rfp->rfr_fpl = NULL;
	  }
	managerfp(0, 0);
}

/*ARGSUSED*/
int ref1main(int argc, char *argv[])
{ /*	This sets up for refinement.  It fetches the current set of
	parameters from the parameter file.  Based on the command parameter,
	it then performs some act.
	Interaction with the user is virtually nonexistent. */
	int		failure, i, intarg, j, refla;
	sichar		origcmd, refcomd;
	float		mbv[2], origva[20];
	double		flargs[3];
	REFINERUN	*rfp;

	if ((char **)&(argv[0]) == flagproc(argc, argv, (char *)NULL))
	  { /* this happens with --help or --version */
		  return EOK;
	  }
	if (0 >= (refcomd = ref1args(argc, argv, &intarg, flargs)))
	  { /* initial file IO and memory allocation */
		fprintf(stderr, "Failure in decoding arguments to %s\n",
			argv[0]);
		return (int)refcomd;
	  }
	if ('J' != refcomd) refla = RR_FORCEREAD;	else refla = 0;
	/* 3 Nov 2006: perhaps by default we should exclude icerings. Not yet.
	refla |= RR_EXCLUDEICE; */
	zeromem((void *)origva, 20 * sizeof (float));
	zeromem((void *)mbv, 2 * sizeof (float));
	if (EOK != (i = ref1init(refla, origva)))
	  {
		fprintf(stderr, "Failure in starting %s: code %d\n",argv[0],i);
		return i;
	  }
	if (NULL == (rfp = managerfp(7, 0))) return EALLOC;
	rfp->rfr_flags = refla;
	sum_curparam(rfp->rfr_fpl , "before refinement", rfp);
	origcmd = refcomd;
	if (('N' == refcomd) || (' ' == refcomd)) /* interactive mode */
	  {
		list_commands();	i = EOK;
		while (EOK == (i = keepinter_ref(rfp))) ;
		/* The special return values of 1 or 2 from keepinter_ref
		 (passed through from ref1cmd) indicate that
		 the user's last interactive command was a Q or an E.
		 AT THIS STAGE Q and E are equivalent, because the
		 saving to disk and printing to stdout and logfile
		 have already been accomplished by the time we get here. */
		if ((i == 1) || (i == 2)) refcomd = 'Q';
		else {
			globalkill();	return i;
		     }
	  }
	/* if refcomd isn't 'N' or ' ', we're in single-command mode */
	else if (EOK != (i = refco(refcomd, argc, argv, intarg, flargs, rfp)))
	  {
		fprintf(stderr, "Error %d from refco\n", i);
		globalkill();	return i;
	  }
	getmainbeam(rfp->rfr_xp, mbv);
	/* write out the UPARAMS file in most cases */
	if ((refcomd != 'G') && (refcomd != 'U') && (refcomd != 'Q'))
	  {
		if (EOK != (failure = globtxs(7, rfp)))
		  {
			fprintf(stderr, "Error %d from globtxs\n", failure);
			globalkill();	return failure;
		  }
	  }
	/* output the current matrix in most cases */
	if ((refcomd != 'B') && (refcomd != 'J') && (refcomd != 'Q'))
	  {
		ref1log( rfp->rfr_fpl, "", origva, mbv, rfp);
		ref1log(stdout, "", origva, mbv, rfp);
	  }
	if ((origcmd == ' ') || (origcmd == 'N')) j = refjournal(NULL);
	else j = EOK;
	if ((EOK == j) && (EOK != xgbuf(NULL, 6))) j = EWRITERR;
	end_refinement();	globalkill();
	return j;
}

/* end of ref1main functionality */

/*******************************************
	ref_all:	top-level routines for all-in-one-operation refinement
	in the circumstance where the user contributes almost no input.
	It does setup operations and on a cycle-by-cycle basis it calls
	routines found in other files.
*******************************************/

static void summ_refain(FILE *fp, REFINERUN *rfp)
{ /*	Summarize the inputs to `refall' to the stream fp */
	fprintf(fp,
	 " The program will%srequest confirmation of the Bravais choice\n",
		(rfp->rfr_flags & RR_ASKBRAVAIS) ? " " : " not ");
	if (rfp->rfr_spspgrp > 0) fprintf(fp,
 "Bravais operation will choose solution consistent with spacegroup %3d\n",
 		(int)(rfp->rfr_spspgrp));
	if ((rfp->rfr_spsys > 0) && (rfp->rfr_spcen > 0))
	  {
		fprintf(fp, "Bravais operation will choose solution ");
		fprintf(fp, "consistent with a %s %s lattice\n",
		 gcenfi((int)(rfp->rfr_spcen)), glatfi((int)(rfp->rfr_spsys)));
	  }
	fprintf(fp,
 "Minimum RMS error below which autoindexing won't be performed = %8.4f\n",
		rfp->rfr_minrmsref);
	if (rfp->rfr_flags & RR_FLIPSTEP) fprintf(fp,
	 "The sign of the stepsize will be reversed from the header value\n");
	fprintf(fp,
	 "The XY beam center will%sbe recalculated right after indexing\n",
	 (rfp->rfr_flags & RR_NORECENTER) ? " not " : " ");
	if (rfp->rfr_flags & RR_ASKSPGRP) 
	fprintf(fp,
	 "The program will%srequest a spacegroup after Bravais calculation\n",
	 (rfp->rfr_flags & RR_ASKSPGRP) ? " " : "not");
}

static REFINERUN *refa_init(int ac, char *av[])
{ /*	Initial operations during refinement */
	int		ifl, retv;
	char		**dcp, **flags;
	static char	*fakeprog = "refine1";
	double		farg, ffl;
	float		origva[20];
	REFINERUN	*rfp;

	dcp = &fakeprog;	retv = 1;
	if ('N' != ref1args(retv, dcp, &ifl, &ffl))
		fprintf(stderr, "Warning: ref1args did not return N\n");
	if (NULL == (rfp = managerfp(7, 0)))
	  {
		fprintf(stderr, "Memory allocation error in managerfp\n");
		return NULL;
	  }
	rfp->rfr_minrmsref = 0.1;
	zeromem((void *)origva, 20 * sizeof (float));
	rfp->rfr_spspgrp = rfp->rfr_spsys = rfp->rfr_spcen = 0;
	rfp->rfr_flags = RR_EXCLUDEICE;	/* we turn this off with -R */
	flags = flagproc(ac, av, "ACFILMNRS");
	for (dcp = flags; NULL != dcp && NULL != *dcp; dcp++)
	   {
		farg = atof((*dcp) + 1);
		if (**dcp == 'A') rfp->rfr_flags |= RR_ASKBRAVAIS;
		else if (**dcp == 'C') rfp->rfr_spcen = farg;
		else if (**dcp == 'F') rfp->rfr_flags |= RR_FLIPSTEP;
		else if (**dcp == 'I') rfp->rfr_spspgrp = farg;
		else if (**dcp == 'L') rfp->rfr_spsys = farg;
		else if (**dcp == 'M') rfp->rfr_minrmsref = farg;
		else if (**dcp == 'N') rfp->rfr_flags |= RR_NORECENTER;
		else if (**dcp == 'R') rfp->rfr_flags &= ~RR_EXCLUDEICE;
		else if (**dcp == 'S') rfp->rfr_flags |= RR_ASKSPGRP;
	   }
	/* specifying the spacegroup supersedes other mechanisms */
	if ((rfp->rfr_spspgrp > 0) && (rfp->rfr_spspgrp <= 267))
	  {
		rfp->rfr_spcen = rfp->rfr_spsys = 0;
		rfp->rfr_flags &= ~(RR_ASKBRAVAIS | RR_ASKSPGRP);
	  }
	else rfp->rfr_spspgrp = 0;
	/* error checking on the centering or system */
	if (rfp->rfr_spcen > 7) rfp->rfr_spcen = 0;
	if (rfp->rfr_spsys > 7) rfp->rfr_spsys = 0;
	if (rfp->rfr_spcen <= 0) rfp->rfr_spsys = 0;
	if (rfp->rfr_spsys <= 0) rfp->rfr_spcen = 0;
	/* if the lattice type is specified but the centering
	 is not, we assume primitive */
	if ((rfp->rfr_spsys > 0) && (rfp->rfr_spcen <= 0))
		rfp->rfr_spcen = 1;
	summ_refain(stdout, rfp);
	if (EOK != (retv = ref1init(rfp->rfr_flags | RR_FORCEREAD, origva)))
	  {
		fprintf(stderr, "Error %d from ref1init\n", retv);
		(void)managerfp(0, 0);	return NULL;
	  }
	return rfp;
}

static int need_indexing(REFINERUN *rfp)
{ /*	This returns 1 iff UPARAMS file is missing or has a random indexing */
	char	*upname;

	if ((NULL == rfp) || (NULL == rfp->rfr_xp)) return 1;
	upname = getenv("UPARAMS");
	if (0 != xgaccess(upname, 04, 1)) return 1; /* UPARAMS missing */
	get_rmserrs("", rfp);
	if (rfp->rfr_nreftot < 3) return 1;
	return (rfp->rfr_indrmse > rfp->rfr_minrmsref);
}

static int get_indlimits(REFINERUN *rfp, float *resl)
{ /*	Sticks appropriate resolution limits into the system for
	purposes of indexing the crystal */
	register int	i, prevnreftot, retv;
	char		indcmd[60];
	float		lolim[] = { 3.0, 2.0, 1.5 } ;

	/* 3 Nov 2006: allow data to 2.1A but excluding ice-ring ranges */
	resl[0] = 18.;	resl[1] = 2.12;
	sprintf(indcmd, "d resmin %6.2f resmax %6.2f\n", resl[0], resl[1]);
	if (EOK != (retv = ref1cmd(indcmd, rfp))) return retv;
	if (EOK != (retv = ref1cmd("u\n", rfp))) return retv;
	if (rfp->rfr_nreftot > ENOUGH_TO_INDEX) return EOK;
	prevnreftot = rfp->rfr_nreftot;
	for (i = 0; i < 3; i++)
	   {
		resl[1] = lolim[i];
		sprintf(indcmd, "d resmax %6.2f\n", resl[1]);
		if (EOK != (retv = ref1cmd(indcmd, rfp))) return retv;
		if (rfp->rfr_nreftot <= prevnreftot) return EBADVALUE;
		if (rfp->rfr_nreftot > ENOUGH_TO_INDEX) return EOK;
		prevnreftot = rfp->rfr_nreftot;
	   }
	return EBADVALUE;
}

static int refa_index(REFINERUN *rfp)
{ /*	Perform an auto-indexing */
	register int retv;
	double		basis;
	float		percerr[2], resli[2];
	char		indcmd[60];

	/* n.b. I need a subtle way of getting the resolution limits */
	percerr[0] = 0.8;	percerr[1] = 0.39;	basis = 50;
	if (EOK != (retv = get_indlimits(rfp, resli))) return retv;
	fprintf(rfp->rfr_fpl,
	 "Indexing performed on data between %7.2f and %7.2f A\n",
		resli[0], resli[1]);
	if (rfp->rfr_flags & RR_NORECENTER)
	  sprintf(indcmd, "i %f %f %lf 1\n", percerr[0], percerr[1], basis);
	else	sprintf(indcmd, "i %f %f %lf\n", percerr[0], percerr[1],basis);
	if (EOK != (retv = ref1cmd(indcmd, rfp))) return retv;
	return EOK;
}

static void spquery(REFPARS *xp)
{ /*	Asks the user for a spacegroup value and responds appropriately */
	int	newsp;
	char	locbu[44];

	while (1) /* ask multiple times till we get a sensible answer */
	     {
		zeromem((void *)locbu, 40);
		fprintf(stderr,
 "Specify International Tables spacegroup number (1-230) or (246-267): ");
		if (NULL != fgets(locbu, 40, stdin))
		  {
			if ('\n' == *locbu) (void)fgets(locbu, 40, stdin);
			newsp = atoi(locbu);
			if ((1 <= newsp) && (newsp <= 267))
			  {
				xp->xispgp = (newsp > 230) ? newsp -100 :newsp;
				xp->xisys = spgtosys((short)newsp);	break;
			  }
			else fprintf(stderr,
	 "locbu = (%s); need a value between 1 and 267 inclusive.\n", locbu);
		  }
		else fprintf(stderr, "No answer provided. Try again\n");
	     }
}

static int refa_refine(REFINERUN *rfp)
{ /*	The short cut to refinement: a series of string calls */
	register int	retv;
	static char	locst[20];

	if (EOK != (retv = ref1cmd("u\n", rfp))) return retv;
	if (EOK != (retv = ref1cmd("j plated n tilt n swing n gamma2 n\n",
		rfp))) return retv;	/* turn off detector params & gamma2 */
	if (EOK != (retv = ref1cmd("l a\n", rfp))) return retv; /* linear */
	if (EOK != (retv = ref1cmd("a\n", rfp))) return retv; /* nonlinear */
	if (EOK != (retv = ref1cmd("d resmin 999\n", rfp))) return retv;
	if (EOK != (retv = ref1cmd("a\n", rfp))) return retv; /* low-resol*/
	if (EOK != (retv = ref1cmd("d resmax 0\n", rfp))) return retv;
	if (EOK != (retv = ref1cmd("r a\n", rfp))) return retv; /* rocking */
	if (EOK != (retv = ref1cmd("l a\n", rfp))) return retv; /* linear */
	if (EOK != (retv = ref1cmd("j plated y\n", rfp))) return retv;
	if (EOK != (retv = ref1cmd("a\n", rfp))) return retv; /* nonlinear */
	if (EOK != (retv = ref1cmd("j tilt y\n", rfp))) return retv;
	if (EOK != (retv = ref1cmd("a\n", rfp))) return retv; /* detec on */
	if (((rfp->rfr_xp)->xispgp == 1) && (rfp->rfr_autoind))
	  { /* User will specify the autoindexing choice */
		if (rfp->rfr_flags & RR_ASKBRAVAIS)
		  {
			if (EOK != (retv = ref1cmd("b q\n", rfp))) return retv;
		  }
		else if (rfp->rfr_spspgrp > 0)
		  {
			if (EOK != (retv = ref1cmd("f -50\n", rfp)))
				return retv;
			/* sometimes the requested Bravais lattice is
			 unplausible. We don't force it then: */
			if (rfp->rfr_spspgrp > 0)
			  {
				zeromem(locst, 20);
				sprintf(locst, "d spacegroup %3d\n",
					(int)(rfp->rfr_spspgrp));
				if (EOK != (retv = ref1cmd(locst, rfp)))
					return retv;
			  }
		  }
		else if ((rfp->rfr_spsys > 0) && (rfp->rfr_spcen > 0))
		  {
			if (EOK != (retv = ref1cmd("f -51\n", rfp)))
				return retv;
		  }
		else { /* let the program pick the best autoindexing choice */
			if (EOK != (retv = ref1cmd("f q\n", rfp))) return retv;
			showbravais(rfp->rfr_fpl, (rfp->rfr_xp)->xareal,
				(rfp->rfr_xp)->xstart);
		     }
		if (rfp->rfr_flags & RR_ASKSPGRP) /* query spacegroup */
			spquery(rfp->rfr_xp);
	  }
	if (EOK != (retv = ref1cmd("u\n", rfp))) return retv;
	if (EOK != (retv = ref1cmd("a\n", rfp))) return retv;
	if (EOK != (retv = ref1cmd("r a\n", rfp))) return retv;
	if (EOK != (retv = ref1cmd("a\n", rfp))) return retv;
	if (EOK != (retv = ref1cmd("a\n", rfp))) return retv;
	if (2 == (retv = ref1cmd("e\n", rfp))) return EOK;
	else return retv - 2;
}

static int refa_finish(void)
{ /*	Cleanly close off a run of `refall' */
	(void)refjournal(NULL);	(void)xgbuf(NULL, 6);
	end_refinement();	globalkill();	return EOK;
}

int refallmain(int argc, char *argv[])
{ /*	This sets up for refinement.  It fetches the current set of
	parameters from the parameter file.  */
	int		retv;
	REFINERUN	*rfp;

	/* initial file IO and memory allocation, and argument processing  */
	if (NULL == (rfp = refa_init(argc, argv)))
	  {
		fprintf(stderr, "I/O or memory problem in refa_args\n");
		return EALLOC;
	  }
	/* perform an auto-indexing if appropriate */
	if ((need_indexing(rfp)) && (EOK != (retv = refa_index(rfp))))
	  {
		fprintf(stderr, "Error %d in refa_index\n", retv); return retv;
	  }
	/* perform initial refinements, typically in P1,
	 and then determine the Bravais lattice if desired.
	 Sometimes the Bravais lattice step involves a user prompt,
	 sometimes not. After that, perform the final refinements
	 in the final spacegroup. */
	if (EOK != (retv = refa_refine(rfp)))
	  {
		fprintf(stderr, "Error %d in refa_refine1\n", retv);
		return retv;
	  }
	/* write out results and exit smoothly*/
	return refa_finish();
}

/* end of ref_all functionality */

/* end of refdriver.c */
