/* refrefine.c -- actual refinement algorithms for the refinement functionality

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	26 Dec 2006: legald checked whenever we loop through reflections;
		cosmetics
	19 Dec 2006: WORKOB typedef moved here from ../include/refidefs.h
	29 Nov 2006: mm3, mm31 calls using transposed matrices replaced
	 by mmt3, mmt31 calls to ordinary matrices; derivedglobs eliminated;
	 management of q21 matrices redone in matrices(); xp->xq21 now used;
	18 May 2006: merge zref, incrcalc, nonlinear into this file
	 xyo_coords: eliminate lrp arg to geth()
	 6 Mar 2006: zref: on_edge written & used; skip zref on elements with
		intensity <= 0 and on reflections close to range edge
	 5 Mar 2006: in newxyphi, REFREF element is framem, not frame
	18 Dec 2005: replace static globals with calls to manageiwp, managenlw
	13 Oct 2005: incrcalc: squeeze all static globals into INWORK
	 structure iws; housekeeping
	13 Oct 2005: nonlinear: collapse static globals down to rnls struct
	28 Aug 2004: routines used only here are declared as static; cosmetics
	26 Aug 2004: remaining static globals replaced with function arguments
	21 Jun 2004: instability in minpar2() fixed with call to xeflush()
	26 Jan 2004: some static globals converted to local variables
	21 Jan 2004: license message inserted and initial comment reordered
	12 Jan 2004: zref: subtle bug in scaling of xgam[0] and xgam[1] fixed
	 3 Jan 2004: REFPARS and REFINERUN structure elements in function calls
		and one static global replace remaining global variables;
		function calls updated
	 7 Dec 2003: fix calls to inv_add(): one was missing an arg before,
		and several required typecasts on the last argument
	25 Nov 2003: replace global variables with elements of REFINERUN struct
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	16 Aug 2003: handle exclusion code for geth differently
	24 Mar 2003: second arg to rocking_ref() controlling printout
	 5 Jan 2003: reorganized source-file header
	15 Nov 2002: extern declarations made here rather than in refiext.h;
		trivial reformatting in nonlinear
	25 Mar 2002: removed MSI attachments
	 8 Nov 1996: minor changes for X-GEN checkin at MSI
	 2 Jun 1995: declare some residual-related variables as doubles
	31 May 1995: "register" -> "register int" throughout
	20 May 1995: nonlinear: &(xnow0[0])
	 2 Mar 1995: eliminate calls to "gripe"; cross reference added
	 2 Mar 1995: diagnostics to stdout; bug in c0mi, c1mi, ... fixed
	 5 May 1994: source-file header cleaned up;
		names changed from POLYDET to XGEN;
		`retv' declared as float in initmin so return values work
	16 Mar 1994: nonlinear: global variables reorganized a little
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN versions

   Purpose: implementations of actual refinement algorithms.
	zref enables refinement of rocking-curve parameters and the consequent
	 changes in the reflection omega centroids. (needs reworking!)
	incrcalc... enables incremental linear least-squares refinements of
	 cell & instrument parameters
	nonlinear... enables Simplex refinements of cell & instrument parameters
   Cross references: ref*.c, ../xglib/bravais.c

*************************************************************************/
#include	"refidefs.h"
#define		ELTSM1		11
#define		NGTRIES		 4
#define		NZTRIES		 3
#define		NOVERALL	 3
#define		ZDAMP		(0.20)
#define	MAXCHANGE	(0.002)	/* 2 milliradians==max change in a gamma par*/
#define	MINGAMMA0	(0.0002)	/* 0.2 mr ==min value of gamma[0] */
#define	MINGAMMA1	(0.0001)	/* 0.1 mr ==min value of gamma[1] */
#define	MINGAMMA2	(-0.01)		/* 0 mr ==min value of gamma[2] */
#define	MAXGAMMA0	(0.02)		/* 20 mr ==max value of gamma[0] */
#define	MAXGAMMA1	(0.02)		/* 20 mr ==max value of gamma[1] */
#define	MAXGAMMA2	(0.01)		/* 10 mr ==max value of gamma[2] */
#define		SQ3O2		(0.8662540)	/* sqrt(3)/2 */

typedef struct { /* working-reflection structure type */
		double	rxisq, ry2, rcos, rsin;
		double	rom, rxcm, rycm;
		float	rxp, ryp, rsloc[3], robv[3];	/* temporary vectors */
		float	rq3p[3][3]; /* rotation matrix */
		} WORKOB;

 typedef struct {
		int	iw_useh[3];
		short	iw_excluco;
		double	iw_alls, iw_cas, iw_cb, iw_cbs, iw_cg, iw_cgs;
		double	iw_delom, iw_delxo, iw_delyo, iw_msca, iw_r2cb;
		double	iw_r2sb, iw_sa, iw_sas, iw_sb, iw_sbcas, iw_sbs;
		double	iw_sbsas, iw_sg, iw_sgs, iw_wtom, iw_wtxy;
		double	iw_delh[3], iw_jjx[14], iw_jjy[14], iw_jjsc[14];
		double	iw_dlda[3][3];
		float	iw_p2q1[3][3], iw_r0[3][3];
		float	iw_r0z0[3][3], iw_r0z1[3][3], iw_r0z2[3][3];
		float	iw_q2p1[3][3], iw_q4[3][3];
		float	iw_z0[3][3], iw_z1[3][3], iw_z2[3][3];
		float	iw_jj[14][3], iw_n[14][3][3];
		} INWORK;

static double getgam(float svl0, float svl1, float lostep,
	float *lga, double *dgam)
{ /*	This computes the gamma value for whose local diffraction vector
	components are (svl0, svl1, svl2), based on the rocking curve values.
	Along with the gamma value we calculate its derivatives with
	respect to gama[0], gama[1], and gama[2].
	We divide the actual gamma value and the derivatives by svl0/r
	to correct for the fact that the oscillation axis is vertical.
	Note that unlike the older version of this routine, it explicitly
	uses the diffraction vector components rather than back-calculating
	them from the observed position of the reflection in centimeters. */
	double	etasq, gamdirec, mysq, mzsq, spectrum, svl1sq, vscale;

	etasq = svl0 * (2. - svl0); svl1sq = svl1 * svl1;
	mzsq = lga[0] * svl1sq;	mysq = lga[1] * (etasq - svl1sq);
	spectrum = lga[2] * svl0;
	gamdirec = sqrt(mysq * lga[1] + mzsq * lga[0]);
	vscale = 1. / fabs(lostep * svl1);
	dgam[0] = mzsq * vscale / gamdirec;
	dgam[1] = mysq * vscale / gamdirec;	dgam[2] = svl0 * vscale;
	return (gamdirec + spectrum) * vscale;
}

static double partit(double gaml, double p0, double *dgp, double *dcp)
{ /*	According to Richard Durbin,
	"This calculates the percentage covered up to p0 and the derivatives
	of this with respect to gamma and the position of the center.
	It is here (and only here) that the rocking curve is modeled.
	As currently set up it uses a half-cosine curve."
	The routine returns the percentage itself, and the derivatives
	are put at the addresses of dgp and dcp.  */
	double	pp;

	pp = PIO2 * p0 / gaml;
	if (pp < -PIO2) pp = -PIO2;	else if (pp > PIO2) pp = PIO2;
	*dcp = -0.5 * PIO2 * cos(pp) / gaml;	*dgp = *dcp * pp / PIO2;
	return (0.5 * (1. + sin(pp)));
}

static int on_edge(REFINERUN *rfp, REFREF *lrp)
{ /* returns true if the reflection's brightest element is either the
   first or last element in the range of images covered in the current
   run of spots */
	int		endf, i, maxind, startf;
	double		maxelt;
	REFREF		*trp;
	static int	minfm, maxfm;

	if (NULL == rfp) /* signal to clear the extrema */
	  {
		minfm = maxfm = 0;	return 0;
	  }
	if (rfp->rfr_npeaks < 1) return 0; /* meaningless */
	if (maxfm < 1) /* establish the range of images if needed */
	  {
		maxfm = -1;	minfm = 9999;	trp = rfp->rfr_rpp0;
		for (i = 0; i < rfp->rfr_npeaks; i++, trp++)
		   {
			startf = trp->obkmin;
			endf = trp->obkmin + (trp->nobk - 1);
			if (startf < minfm) minfm = startf;
			if (endf > maxfm) maxfm = endf;
		   }
		if (NULL != rfp->rfr_fpl) fprintf(rfp->rfr_fpl,
		 "\t\tRange of images for rocking-refinements: [%3d, %3d]\n",
			minfm, maxfm);
	  }
	/* find the peak of the current reflection */
	endf = lrp->obkmin + lrp->nobk - 1;
	for (i = 1, maxind = 0, maxelt = lrp->obparts[0]; i < lrp->nobk; i++)
	   {
		if (lrp->obparts[i] > maxelt)
		  {
			maxelt = lrp->obparts[i];	maxind = i;
		  }
	   }
	/* if the profile's strongest element is the initial image, return -1 */
	if ((maxind == 0) && (lrp->obkmin <= minfm)) return -1;
	/* if the strongest element is the final image, return 1 */
	if ((maxind == lrp->nobk - 1) && (endf >= maxfm)) return 1;
	return 0;	/* any other result gives us a negative answer */
}

static int zref(int ncyc, REFINERUN *rfp, double lostep, float lga[])
{ /*	This routine refines the spindle centroids of the peaks.
	I don't understand how this stuff works, so I'm going on
	faith that Richard understood what he was doing.
	Mar 06: set negative profile elements to zero,
	and don't include reflections on the edge of the range */
	register int	isl, isp, k, nch;
	double		dcp, dcpart, delc, dgpart, err, gamm, olddc, oldpart;
	double		p0, part, summ, sumsqerr, sumv, thispart, totch;
	double		dgam[3];
	REFREF		*lrp;

	sumsqerr = totch = nch = 0;
	for (lrp = rfp->rfr_rpp0 + rfp->rfr_firstok,
	 isp = rfp->rfr_inbounds; isp--; lrp++)
	   {
		if ((isl = legald(rfp, lrp)) &&
		 (lrp->nobk > 1) && !on_edge(rfp, lrp))
		  {
			lrp->rflags &= ~INICERING;
			sumv = summ = 0.;
			gamm = getgam((double)(lrp->rsl[0]),
			 (double)(lrp->rsl[1]), lostep, lga, dgam);
			p0 = (double)lrp->obkmin - (lrp->framem + 0.5);
			oldpart = part = partit(gamm, p0, &dgpart, &dcpart);
			olddc = dcpart;	p0 += 1;
			for (k = 1; k <= lrp->nobk; k++)
			   {
				part = partit(gamm, p0, &dgpart, &dcpart);
				thispart = flimits(lrp->obparts[k-1], 0., 1.);
				err = thispart - (part - oldpart);
				sumsqerr += err * err; dcp = dcpart - olddc;
				sumv += err * dcp; summ += dcp * dcp;
				oldpart = part; olddc = dcpart; p0 += 1;
			   }
			if (summ > 0)
			  {
				delc = sumv / summ;
				lrp->framem = lrp->frame0 + ZDAMP * delc;
				nch++;	totch += fabs(delc);
			  }
		  }
		else
		  {
			if (isl) lrp->rflags &= ~INICERING;
			else lrp->rflags |= INICERING;
		  }
	   }
	rfp->rfr_rpp = lrp;
	if (nch)
	  {
		xeflush(rfp->rfr_fpl);	sumsqerr = sqrt(sumsqerr/nch);
		if (ncyc == NZTRIES)
		  {
			rfp->rfr_profrmse = sumsqerr;
			rfp->rfr_avedelfm = totch / nch;
			rfp->rfr_nrefok = nch;	rfp->rfr_nindok = nch * 3;
		  }
	  }
	else
	  {
		rfp->rfr_profrmse = rfp->rfr_avedelfm =
			rfp->rfr_nrefok = rfp->rfr_nindok = 0;	return 0;
	  }
	rfp->rfr_rpp = rfp->rfr_rpp0 + rfp->rfr_firstok; return 1;
}

static int gamref(int ncyc, REFINERUN *rfp, double lostep, float lga[])
{ /*	This performs a refinement of the rocking curve parameters.
	Usually lga[0] and lga[1] are refined; lga[2] isn't.  */
	register int	i, j, k, m, nres;
	double		cgama, dcpart, dgpart, err;
	double		gamres, oldpart, olddg, p0, part;
	double		ddgam[3], dgam[3], vgam[3], xga[3];
	double		gammat[3][3], rgam[3][3];
	REFPARS		*xp;
	REFREF		*lrp;
	static float	minlgam[3] = { MINGAMMA0, MINGAMMA1, MINGAMMA2 } ;
	static float	maxlgam[3] = { MAXGAMMA0, MAXGAMMA1, MAXGAMMA2 } ;

	xp = rfp->rfr_xp;	gamres = nres = 0;
	for (i = 3; i--; )
	   {
		vgam[i] = xga[i] = 0.;
		for (j = 3; j--; ) gammat[i][j] = 0.;
	   }
	for (i = rfp->rfr_inbounds, lrp = rfp->rfr_rpp0 + rfp->rfr_firstok;
	 i--; lrp++)
	 {
	   if (legald(rfp, lrp))
	     {
		lrp->rflags &= ~INICERING;
		cgama = getgam((double)(lrp->rsl[0]), (double)(lrp->rsl[1]),
			lostep, lga, dgam);
		p0 = lrp->obkmin - (lrp->framem + 0.5);
		oldpart = part = partit(cgama, p0, &dgpart, &dcpart);
		olddg = dgpart;	p0 += 1.;
		for (k = 1; k <= lrp->nobk; k++)
		   {
			part = partit(cgama, p0, &dgpart, &dcpart);
			err = lrp->obparts[k-1] - (part - oldpart);
			nres++;	gamres += err * err;
			for (j = 0; j < 3; j++)
			   {
				ddgam[j] = dgam[j] * (dgpart - olddg);
				vgam[j] += err * ddgam[j];
				for (m = 0; m <= j; m++)
				   gammat[j][m] += ddgam[m] * ddgam[j];
			   }
			oldpart = part; olddg = dgpart; p0 += 1.;
		   }
	     }
	   else lrp->rflags |= INICERING;
	 }
	rfp->rfr_rpp = lrp;
	gammat[0][1] = gammat[1][0]; gammat[0][2] = gammat[2][0];
	gammat[1][2] = gammat[2][1]; /* get the symmetry-related elements */
	gamres = (nres ? sqrt(gamres/(float)nres) : 1000.);
	/* invert the gammat matrix and solve for the increments */
	if (dinv3(gammat, rgam)) return 0;
	dmm31(xga, rgam, vgam);
	/* exclude the possibility of very large changes */
	for (i = 3; i--; )
	   {
		if (xga[i] < -MAXCHANGE) xga[i] = -MAXCHANGE;
		else if (xga[i] > MAXCHANGE) xga[i] = MAXCHANGE;
	   }
	for (i = 3; i--; )	/* apply the changes */
	   if (xp->xisref[i+14] == 'Y')
	     {
		olddg = lga[i];	lga[i] += xga[i];
		if (lga[i] < minlgam[i]) lga[i] = minlgam[i];
		else if (lga[i] > maxlgam[i]) lga[i] = maxlgam[i];
		xga[i] = lga[i] - olddg;
	     }
	if (ncyc == NGTRIES)	/* save results of last cycle */
	  {
		for (i = 3; i--; ) /* save deltas for display */
		   {
			if (xp->xisref[i+14] == 'Y') rfp->rfr_pshift[i+14] =
				xga[i] / (i == 2 ?1. :fabs(lostep));
			else rfp->rfr_pshift[i+14] = 0.;
		   }
		rfp->rfr_gamrmse = gamres;
	  }
	return 1;
}

int rocking_ref(float oldgam[], int report, REFINERUN *rfp)
{ /*	This does a series of spindle and gamma refinements */
 	register int	jpas, i;
	double		locstep, thisgam;
	float		*lgam, *oldg;
	REFPARS		*xp;

	xp = rfp->rfr_xp;	locstep = xp->xfrmsize;
	lgam = &(xp->xgam[0]);
	for (jpas = 1; jpas <= NGTRIES; jpas++)
		if (!gamref(jpas, rfp, locstep, lgam)) return 1;
	for (jpas = 1; jpas <= NZTRIES; jpas++)
		if (!zref(jpas, rfp, locstep, lgam)) return 2;
	/* get the reference-angle scattering vectors and the indices */
	for (rfp->rfr_rpp = rfp->rfr_rpp0 + rfp->rfr_firstok,
	 i = rfp->rfr_inbounds; i--; (rfp->rfr_rpp)++) refgonio(rfp);
	/* compute RMS errors in everything. Report if 'report' is nonzero */
	get_rmserrs((report) ? "rocking-curve" : "", rfp);
	for (i = 14, oldg = oldgam; i < 17; i++, oldg++) /* shifts in params */
	   {
		thisgam = (i == 16 ? lgam[2] : lgam[i-14] / fabs(locstep));
		if (xp->xisref[i] == 'Y') rfp->rfr_pshift[i] = thisgam - *oldg;
		else	rfp->rfr_pshift[i] = 0.;
	   }
	for (i = 14; i--; ) rfp->rfr_pshift[i] = 0.; /* clear unused shifts */
	return 0;
}

/* end of zref-related functions */

/************************************
	incrcalc: routines enabling calculations for linear
	 least-squares parameter refinements in "refine"
	In this version up to fourteen parameters are refined:
		(0)-(8) crystal parameters:
			a, b, c, alpha, beta, gamma, omega, chi, phi
		(9)-(13) detector parameters:
			plateD, Xcen, Ycen, tilt, swing
	Only those parameters which are in fact being refined will have
	their derivatives included in the least-squares matrix.
	In XGEN 3.0, routines associated with linear lsq refinement of the
	scanning angle and (X,Y) instead of nonintegerness are found here too.
*************************************/

static INWORK *manageiwp(int opno, REFINERUN *rfp)
{ /*	This manages a pointer *iwp according to opno:
	if opno == 0, we destroy the memory for the pointer and return NULL
	if opno == 1, we unconditionally create a new one and return it
	if opno == 3, we return the current one */
	static INWORK	*iwp;
	static char	iwpmsg[] = "iwp in manageiwp";

	if (!opno) /* destroy the old one */
	  {
		memzap((AVOID **)&iwp, iwpmsg);	return NULL;
	  }
	if (opno == 01) (void)manageiwp(0, NULL); /* force destruction */
	if (NULL == iwp) /* create new one if needed */
	  {
		if (NULL == (iwp = (INWORK *)callop(1, sizeof (INWORK),
		 iwpmsg))) return NULL;
	  }
	if (opno & 04) /* check to see if we need to do scalars & matrices */
	  {
		if ((fabs(iwp->iw_cb) < DFUZZ) && (fabs(iwp->iw_sb) < DFUZZ))
		  {
			scalars(rfp);	matrices(rfp);
		  }
	  }
	return iwp;	/* we'll get artier later, maybe */
}

void scalars(REFINERUN *rfp)
{ /*	This obtains constant scalars used in incremental refinement*/
	register int	i,j;
	REFPARS		*xp;
	INWORK		*iwp;

	if (NULL == (iwp = manageiwp(3, rfp))) return;
	xp = rfp->rfr_xp;
	iwp->iw_cb = cos(xp->xang[1] * RADC);
	iwp->iw_sb = sin(xp->xang[1] * RADC);
	iwp->iw_cg = cos(xp->xang[2] * RADC);
	iwp->iw_sg = sin(xp->xang[2] * RADC);
	iwp->iw_cas = cos(xp->xanst[0]);
	iwp->iw_sas = sin(xp->xanst[0]);
	iwp->iw_cbs = cos(xp->xanst[1]);
	iwp->iw_sbs = sin(xp->xanst[1]);
	iwp->iw_cgs = cos(xp->xanst[2]);
	iwp->iw_sgs = sin(xp->xanst[2]);
	iwp->iw_sbcas = iwp->iw_sb * iwp->iw_cas;
	iwp->iw_sbsas = iwp->iw_sb * iwp->iw_sas;
	iwp->iw_alls = 1. /
	 (iwp->iw_sbsas * iwp->iw_sb * iwp->iw_sbs * iwp->iw_sg * iwp->iw_sgs);
	if (1 == xp->xisys || 7 == xp->xisys)
	  {
		iwp->iw_dlda[0][0] = 0.;
		iwp->iw_dlda[0][1]= -iwp->iw_sa / iwp->iw_sg;
		iwp->iw_dlda[0][2] = iwp->iw_sa * iwp->iw_cas / iwp->iw_sas;
		iwp->iw_dlda[1][0] = -iwp->iw_sb;
		iwp->iw_dlda[1][1]= iwp->iw_sb * iwp->iw_cg / iwp->iw_sg;
		iwp->iw_dlda[1][2] = -iwp->iw_cbs * iwp->iw_sa /
			(iwp->iw_sas * iwp->iw_sgs);
		iwp->iw_dlda[2][0] = 0.;
		iwp->iw_dlda[2][1]= -iwp->iw_cbs * iwp->iw_sa /
			iwp->iw_sg;
		iwp->iw_dlda[2][2] = -iwp->iw_cas * iwp->iw_cbs * iwp->iw_sa /
			(iwp->iw_sg * iwp->iw_sas);
		for (i = 3; i--; ) for (j = 3; j--; )
			iwp->iw_dlda[i][j] *= xp->xabc[2];
	  }
	else if (2 == xp->xisys)
	  {
		iwp->iw_r2cb = xp->xabc[2] * iwp->iw_cb;
		iwp->iw_r2sb = xp->xabc[2] * iwp->iw_sb;
	  }
		/* least-squares rescaler */
	iwp->iw_msca = (xp->xabc[0] + xp->xabc[1] + xp->xabc[2]) / 3.;
	if (iwp->iw_msca < DFUZZ) iwp->iw_msca = 1.; /* no divide-by-0 goofs */
}

static void rot0tostart(float qq[3][3], REFINERUN *rfp)
{ /*	This right-multiplies the matrix qq by a matrix that is constructed
	from the rotation from the goniostat's (0,0,0) position to
	the goniostat's (omegastart, chistart, phistart) position. */
	register int	i, j;
	float		zw[3][3];

	for (i = 3; i--; ) for (j = 3; j--; ) zw[i][j] = qq[i][j];
	mm3(qq, zw, rfp->rfr_zsa);
}

void matrices(REFINERUN *rfp)
{ /*	This obtains the constant matrices used in incremental refinement. */
	register int	i, j, m;
	double		c0, c1, c2, rph0, rph1, rph2, s0, s1, s2;
	float		wk[3][3];
	REFPARS		*xp;
	INWORK		*iwp;

	/* construct the tilt and swing rotation matrices */
	if (NULL == (iwp = manageiwp(3, rfp))) return;
	xp = rfp->rfr_xp;	xp->xpltd2 = xp->xplated * xp->xplated;
	makeq21(xp);
	/* iwp->iw_p2q1 = d(q21)/d(swing); iwp->iw_q2p1 = d(q21)/d(tilt) */
	for (i = 0; i < 3; i++)
	   {
		iwp->iw_q2p1[i][0] = 0.;
		iwp->iw_q2p1[i][1] = -xp->xq21[i][2];
		iwp->iw_q2p1[i][2] = xp->xq21[i][1];
		iwp->iw_p2q1[0][i] = xp->xq21[1][i];
		iwp->iw_p2q1[1][i] = -xp->xq21[2][i];
		iwp->iw_p2q1[2][i] = 0.;
	  }
	/* Get invariant components of the spindle-angle rotation matrix q3 */
	rfp->rfr_q3[2][2] = 1.;
	rfp->rfr_q3[0][2] = rfp->rfr_q3[1][2] =
		rfp->rfr_q3[2][0] = rfp->rfr_q3[2][1] = 0.;
	/* Get the matrices used in refining the pseudogoniostat angles
	rphi[].  q4 is the rotation matrix itself; z0-z2 are the
	derivatives of q4 with respect to rphi[0] - rphi[2] respectively. */
	rph0 = (-RADC) * xp->xdelom;	c0 = cos(rph0);	s0 = sin(rph0);
	rph1 = (-RADC) * xp->xdelchi;	c1 = cos(rph1);	s1 = sin(rph1);
	rph2 = (-RADC) * xp->xdelphi;	c2 = cos(rph2);	s2 = sin(rph2);
	iwp->iw_q4[0][0]=  s0*c1*s2 + c0*c2;
	iwp->iw_q4[0][1]= -s0*c1*c2 + c0*s2;
	iwp->iw_q4[0][2]=  s0 * s1;
	iwp->iw_q4[1][0]= -c0*c1*s2 + s0*c2;
	iwp->iw_q4[1][1]=  c0*c1*c2 + s0*s2;
	iwp->iw_q4[1][2]= -c0 * s1;
	iwp->iw_q4[2][0]= -s1*s2;
	iwp->iw_q4[2][1]=  s1*c2;
	iwp->iw_q4[2][2]=  c1;
	/* Matrices z0-z2 are the derivatives of q4 with respect to omega, chi,
	and phi respectively. Omega and phi are easy: chi is a little harder */
	for (i = 3; i--; )
	   {
		iwp->iw_z0[0][i] =  iwp->iw_q4[1][i];
		iwp->iw_z0[1][i] = -iwp->iw_q4[0][i];
		iwp->iw_z0[2][i] = 0.;
		iwp->iw_z2[i][0] =  iwp->iw_q4[i][1];
		iwp->iw_z2[i][1] = -iwp->iw_q4[i][0];
		iwp->iw_z2[i][2] = 0.;
	   }
	iwp->iw_z1[0][0]=  s0 * s1 * s2; iwp->iw_z1[0][1]= -s0 * s1 * c2;
	iwp->iw_z1[0][2]= -s0 * c1;
	iwp->iw_z1[1][0]= -c0 * s1 * s2; iwp->iw_z1[1][1]=  c0 * s1 * c2;
	iwp->iw_z1[1][2]=  c0 * c1;
	iwp->iw_z1[2][0]=       c1 * s2; iwp->iw_z1[2][1]=      -c1 * c2;
	iwp->iw_z1[2][2]=       s1;
	/* all of these matrices have to be left-multiplied by the
	matrix that gets us from (omegastart, chistart, phistart) to
	(0, 0, 0) in order to actually use them.  I'm too dumb to
	figure out which way that one works, so here goes... */
	ztos(xp->xstart, rfp->rfr_zsa); /* in case we haven't already... */
	rot0tostart(iwp->iw_q4, rfp);	rot0tostart(iwp->iw_z0, rfp);
	rot0tostart(iwp->iw_z1, rfp);	rot0tostart(iwp->iw_z2, rfp);
	/* The matrix at the null position (a along X, b along X and Y) is
	`areal' right multiplied by the inverse (i.e.the transpose) of q4. */
	mm3t(iwp->iw_r0, xp->xareal, iwp->iw_q4);
	/* left-multiply derivative matrices by R0 to give useful things. */
	mm3(iwp->iw_r0z0, iwp->iw_r0, iwp->iw_z0);
	mm3(iwp->iw_r0z1, iwp->iw_r0, iwp->iw_z1);
	mm3(iwp->iw_r0z2, iwp->iw_r0, iwp->iw_z2);
	/* The matrices N0, ... N8 represent the derivatives of the
	 reciprocal-space unit cell matrix (lambda * Q4 * A0) with respect
	 to the first nine refinable parameters, i.e.
	 a,b,c,alpha,beta,gamma,omega,chi, and phi. */
		/* iwp->iw_n[0][i][j] is the derivative with respect to a */
	for (m = 9; m--; ) for (i = 3; i--; ) for (j = 3; j--; )
		iwp->iw_n[m][i][j] = 0.;
	for (i = 3; i--; ) for (j = 1; j < 3; j++) wk[i][j] = 0.;
	wk[0][0] = (-iwp->iw_msca/ (xp->xabc[0])) / (xp->xabc[0]);
	wk[1][0] = -wk[0][0] * iwp->iw_cg / iwp->iw_sg;
	wk[2][0] = wk[0][0] * iwp->iw_cbs / (iwp->iw_sbs * iwp->iw_sg);
		/* special cases when b is tied to a */
	if (4 <= xp->xisys && xp->xisys <= 7)
	  {
		wk[1][1] = wk[0][0] / iwp->iw_sg;
			/* cubic: c tied to a */
		if (5 == xp->xisys) wk[2][2] = wk[0][0];
		else if (7 == xp->xisys) /* rhombohedral: c tied to a */
		  {
			wk[2][2] = wk[0][0] / (iwp->iw_sg * iwp->iw_sgs);
			wk[2][1] = wk[2][2] * iwp->iw_cg;
		  }
	  }
	mmt3(iwp->iw_n[0], iwp->iw_q4, wk);
		/*  iwp->iw_n[1][i][j] is the derivative with respect to b */
	if (xp->xisys < 4) /* triclinic,monoclinic,orthorhombic only ... */
	  {
		for (i = 3; i--; ) for (j = 3; j--; ) wk[i][j] = 0.;
		wk[1][1] = (-iwp->iw_msca / (xp->xabc[1])) /
			((xp->xabc[1]) * iwp->iw_sg);
		wk[2][1] = wk[1][1] * iwp->iw_cas / iwp->iw_sas;
		mmt3(iwp->iw_n[1], iwp->iw_q4, wk);
	  }
		/* iwp->iw_n[2][i][j] is the derivative with respect to c */
	if ((xp->xisys != 5) && (xp->xisys != 7))
	  {
		for (i = 3; i--; ) for (j = 3; j--; ) wk[i][j] = 0.;
		wk[2][2] = (-iwp->iw_msca / (xp->xabc[2])) /
			((xp->xabc[2]) * iwp->iw_sb * iwp->iw_sas);
		mmt3(iwp->iw_n[2], iwp->iw_q4, wk);
	  }
		/* iwp->iw_n[3][i][j] is the derivative with respect to alpha*/
	if (xp->xisys == 1) /* alpha is only free in triclinic */
	  {
		for (i = 2; i--; ) for (j = 3; j--; ) wk[i][j] = 0.;
		wk[2][0] = iwp->iw_alls * iwp->iw_cgs * iwp->iw_sa *
			iwp->iw_msca / (xp->xabc[0]);
		wk[2][1] = iwp->iw_alls * iwp->iw_sb *
			iwp->iw_msca / (xp->xabc[1]);
		wk[2][2] = iwp->iw_alls * iwp->iw_cas * iwp->iw_sg *
			iwp->iw_msca / (xp->xabc[2]);
		mmt3(iwp->iw_n[3], iwp->iw_q4, wk);
		/* iwp->iw_n[1][i][j] is the derivative with respect to gamma;
		 gamma is free only in triclinic and rhombohedral,
		 and only the triclinic case is covered here. */
		for (j = 3; j--; ) wk[0][j] = 0.;	wk[1][2] = 0.;
		wk[1][0] = iwp->iw_msca /
			((xp->xabc[0]) * iwp->iw_sg * iwp->iw_sg);
		wk[1][1] = -iwp->iw_cg * iwp->iw_msca /
			((xp->xabc[1]) * iwp->iw_sg * iwp->iw_sg);
		wk[2][0] = ((iwp->iw_cas / (iwp->iw_sg * iwp->iw_sg *
		 iwp->iw_sas) + iwp->iw_alls * iwp->iw_sa * iwp->iw_cbs *
		 iwp->iw_cgs)) * iwp->iw_msca / (xp->xabc[0]);
		wk[2][1] = ((iwp->iw_cbs / (iwp->iw_sg * iwp->iw_sg *
		 iwp->iw_sbs) + iwp->iw_alls * iwp->iw_sb * iwp->iw_cas *
		 iwp->iw_cgs)) * iwp->iw_msca / (xp->xabc[1]);
		wk[2][2] = ((iwp->iw_cg / (iwp->iw_sbs * iwp->iw_sa *
		 iwp->iw_sg) + iwp->iw_alls * iwp->iw_cgs * iwp->iw_sg)) *
			iwp->iw_msca / (xp->xabc[2]);
		mmt3(iwp->iw_n[5], iwp->iw_q4, wk);
	  }
		/* rhombohedral gamma (alpha & beta tied to it) */
	else if (xp->xisys == 7)
	  {
		for (j = 3; j--; ) wk[0][j] = 0.;	wk[1][2] = 0.;
		wk[1][0] = iwp->iw_msca /
			((xp->xabc[0]) * iwp->iw_sg * iwp->iw_sg);
		wk[1][1] = -iwp->iw_cg * wk[1][0];
		wk[2][0] = wk[2][1] = ((iwp->iw_cgs / (iwp->iw_sg *
		 iwp->iw_sg * iwp->iw_sgs) + iwp->iw_alls * iwp->iw_sg *
		 iwp->iw_cgs * iwp->iw_cgs)) * iwp->iw_msca / (xp->xabc[0]);
		wk[2][2] = ((iwp->iw_cg / (iwp->iw_sgs * iwp->iw_sg *
		 iwp->iw_sg) + iwp->iw_alls * iwp->iw_cgs * iwp->iw_sg)) *
		 iwp->iw_msca / (xp->xabc[0]);
		mmt3(iwp->iw_n[5], iwp->iw_q4, wk);
	  }
	if (xp->xisys <= 2) /* beta: free in triclinic and monoclinic */
	  {
		for (i = 2; i--; ) for (j = 3; j--; ) wk[i][j] = 0.;
		wk[2][0] = iwp->iw_alls * iwp->iw_sa *
		 iwp->iw_msca / (xp->xabc[0]);
		wk[2][1] = iwp->iw_alls * iwp->iw_sb * iwp->iw_cgs *
		 iwp->iw_msca / (xp->xabc[1]);
		wk[2][2] = iwp->iw_alls * iwp->iw_sg * iwp->iw_cbs *
		 iwp->iw_msca / (xp->xabc[2]);
		mmt3(iwp->iw_n[4], iwp->iw_q4, wk);
	  }
	/* iwp->iw_n[6][i][j] is the derivative with respect to omega(a);
		n[7][i][j] is the derivative with respect to chi(a);
		n[8][i][j] is the derivative with respect to phi(a) */
	wk[0][1] = wk[0][2] = wk[1][2] = 0.;
	wk[0][0] = iwp->iw_msca / (xp->xabc[0]);
	wk[1][0] = -wk[0][0] * iwp->iw_cg / iwp->iw_sg;
	wk[2][0] = wk[0][0] * iwp->iw_cbs / (iwp->iw_sbs * iwp->iw_sg);
	wk[1][1] = iwp->iw_msca / ((xp->xabc[1]) * iwp->iw_sg);
	wk[2][1] = wk[1][1] * iwp->iw_cas / iwp->iw_sas;
	wk[2][2] = iwp->iw_msca / ((xp->xabc[2]) * iwp->iw_sb * iwp->iw_sas);
	mmt3(iwp->iw_n[6], iwp->iw_z0, wk);
	mmt3(iwp->iw_n[7], iwp->iw_z1, wk);
	mmt3(iwp->iw_n[8], iwp->iw_z2, wk);
	for (m = 0; m < 9; m++) for (i = 3; i--; ) for (j = 3; j--; )
		iwp->iw_n[m][i][j] *= xp->xlambda;
}

static double newxyphi(WORKOB *wrp0, REFINERUN *rfp)
{ /*	This computes the calculated (X,Y,phi) position of a reflection
	with the scattering vector rp->rsi.  Output are pointers to structures
	containing the omega,X,Y positions of the first and second intersections
	of the scattering vector with the Ewald sphere.
	Currently X and Y are in pixels and the phi values are in frames,
	but that may change.  The routine returns 0 if the reflection's
	phi, X, and Y values can be successfully computed, nonzero if not.
	The routine always puts the intersection point closer to the input
	position in wrp0. It returns the difference
	between the observed and calculated omega in frames. */

	int	i, jfr, quadr;
	double	diff0, diff1, fm180deg, xisq, zetasq, rsz, y2, rz0;
	float	*rs;
	REFPARS	*xp;
	REFREF	*rp;
	WORKOB	*wrp, wr1;
	
	xp = rfp->rfr_xp;	rp = rfp->rfr_rpp;	rs = &(rp->rsi[0]);
	if (DFUZZ >= (xisq = rs[0] * rs[0] + rs[1] * rs[1])) return 2;
	wrp0->rxisq = wr1.rxisq = xisq;
	if (1 < (zetasq = rs[2] * rs[2])) return 3;
	rsz = 0.5 * (xisq + zetasq);
	if (MINY2SQ > (y2 = xisq - rsz * rsz)) return 4;
	wr1.ry2 = -(wrp0->ry2 = sqrt(y2));
	quadr = detregion(rp->xto, rp->yto, rfp->rfr_spl);
	for (i = 0; i < 2; i++)
	   {
		wrp = (i ? &wr1: wrp0);
		wrp->rsin = (wrp->ry2 * rs[0] +      rsz * rs[1]) / xisq;
		wrp->rcos = (     rsz * rs[0] - wrp->ry2 * rs[1]) / xisq;
		wrp->rom = atan2(wrp->rsin, wrp->rcos);
		wrp->rq3p[0][0] = wrp->rq3p[1][1] = wrp->rcos;
		wrp->rq3p[1][0] = -(wrp->rq3p[0][1] = wrp->rsin);
		wrp->rq3p[2][2] = 1.;
		wrp->rq3p[0][2] = wrp->rq3p[2][0] =
			wrp->rq3p[1][2] = wrp->rq3p[2][1] = 0.;
		mm31(wrp->rsloc, wrp->rq3p, rp->rsi);	wrp->rsloc[0] -= 1.;
		mmt31(wrp->robv, xp->xq21, wrp->rsloc);
		if (DFUZZ > fabs(rz0 = wrp->robv[0]))
			wrp->rxp = wrp->ryp = -100.;
		else {
			wrp->rxcm = xp->xx0 + xp->xplated * wrp->robv[1] / rz0;
			wrp->rycm = xp->xy0 + xp->xplated * wrp->robv[2] / rz0;
			jfr = rp->rsp / xp->xfrmsize + 0.49999;
			if (0 == cm2pixq((float)wrp->rxcm, (float)wrp->rycm,
			 quadr, &wrp->rxp,&wrp->ryp,jfr))
				wrp->rxp = wrp->ryp = -100.;
		     }
	   }
	if (wrp0->rxp < -99.)	/* guess zero is bad; swap */
	  {
		if (wr1.rxp < -99.) return 10000.;	/* both guesses bad*/
		swap((AVOID *)wrp0, (AVOID *)&wr1, sizeof (WORKOB));
		if ((diff0 = fabs(wrp0->rom - rp->rsp)) > PI)
			diff0 = PIX2 - diff0;
	  }
	else if (wr1.rxp < -99.) /* guess 1 is bad: don't swap */
	  {
		if ((diff0 = fabs(wrp0->rom - rp->rsp)) > PI)
			diff0 = PIX2 - diff0;
	  }	
	else { /* both guesses are legal; pick the smaller delta omega */
		if ((diff0 = fabs(wrp0->rom - rp->rsp)) > PI)
			diff0 = PIX2 - diff0;
		if ((diff1 = fabs(wr1.rom - rp->rsp)) > PI)
			diff1 = PIX2 - diff1;
			/* if guess 2 is closer, swap them */
		if (diff0 > diff1)
		  {
			diff0 = diff1;
			swap((AVOID *)wrp0, (AVOID *)&wr1, sizeof (WORKOB));
		  }
	      }
	rp->framc = wrp0->rom / xp->xfrmsize - 0.5;
	fm180deg = 180. / fabs(xp->xfrmsize / RADC);
	if ((rp->framc < 0.) && (rp->framem > fm180deg))
		rp->framc += 2. * fm180deg;
	rp->xtc = wrp0->rxp;	rp->ytc = wrp0->ryp;
	return	diff0 / fabs(xp->xfrmsize);
}

static int xyo_coords(char restype, REFINERUN *rfp, INWORK *iwp)
{ /*	This computes some useful coordinates for rp.  It returns 1 if the
	difference between the observed and calculated omega value is
	smaller than the preset error limit.  It also gets the derivative
	of omega with respect to each crystal parameter. */
	int	i;
	double	denom, detsc, lsx, lsy, lsz, rho, tsz, y2;
	float	pztilt[3], pzswing[3], xyder[3], vtemp[3], q123p[3][3], vt[3];
	REFPARS	*xp;
	REFREF	*rp;
	WORKOB	wr;

	rp = rfp->rfr_rpp;	xp = rfp->rfr_xp;
	if (!geth(rfp, iwp->iw_excluco))
	  {
		rp->rflags = BADINDICES;	return 0;
	  }
	if (9999. < (iwp->iw_delom = newxyphi(&wr, rfp)))
	  { /* X,Y, phi */
		rp->rflags = BADSCATTVEC; return 0;
	  }
	if (iwp->iw_delom > xp->xerrphi || fabs(rp->xto - rp->xtc) > xp->xerrx
	 || fabs(rp->yto - rp->ytc) > xp->xerry)
	  {
		rp->rflags = BADXYPHI;	return 0;
	  }
	else if (rp->rflags & BADXYPHI) rp->rflags -= BADXYPHI;
	/* the vector vt will be dotted with all of the various vectors
	(@sx/@p, @sy/@p, @sz/@p). */
	iwp->iw_delom = rp->rsp - wr.rom; /* we need the SIGNED residual */
	lsx = rp->rsi[0];	lsy = rp->rsi[1];	lsz = rp->rsi[2];
	rho = 0.5 * (wr.rxisq + lsz * lsz);	y2 = wr.ry2;
	tsz = rho - wr.rxisq;	denom = wr.rxisq * y2;
	vt[0] = (-y2 * lsy + tsz * lsx) / denom;
	vt[1] = ( y2 * lsx + tsz * lsy) / denom;
	vt[2] = -lsz / wr.ry2;
	detsc = xp->xplated / wr.robv[0];
	/* the following matrix is used in determining the derivatives of X and
	Y with respect to a,b,c,alpha,beta,gamma,omega,chi,and phi.*/
	mmt3(q123p, xp->xq21, wr.rq3p);
	for (i = 0; i < 9; i++)
	   {
		mm31(vtemp, iwp->iw_n[i], rp->rhi);
		iwp->iw_jjsc[i] = dotprod(vt, vtemp, 3);
		mm31(xyder, q123p, vtemp);
		iwp->iw_jjx[i] = detsc * xyder[i];
		iwp->iw_jjy[i] = detsc * xyder[2];
	   }
	if (restype == 'K') return 1;
	mmt31(pztilt, iwp->iw_q2p1, wr.rsloc);
	mmt31(pzswing, iwp->iw_p2q1, wr.rsloc);
	iwp->iw_jjx[9] = wr.robv[1] / wr.robv[0];	iwp->iw_jjx[10] = 1.;
	iwp->iw_jjx[11] = 0.;	iwp->iw_jjx[12] = detsc * pztilt[1];
	iwp->iw_jjx[13] = detsc * pzswing[1];
	iwp->iw_jjy[9] = wr.robv[2] / wr.robv[0];	iwp->iw_jjy[10] = 0.;
	iwp->iw_jjy[11] = 1.;	iwp->iw_jjy[12] = detsc * pztilt[2];
	iwp->iw_jjy[13] = detsc * pzswing[2];
		/* remember the errors in x and y (in centimeters) */
	iwp->iw_delxo = rp->xtcm - wr.rxcm;
	iwp->iw_delyo = rp->ytcm - wr.rycm;
	return 1;
}

int refl_coords(REFINERUN *rfp)
{ /*	This computes various coordinate definitions for the reflection rp.
	It returns the number of indices (0-3) of the reflection which are
	close enough to integers that they can be used in refinement.
	After computing the coordinates, it gets the derivatives of the
	index vector (h,k,l) with respect to each of the detector and
	crystal parameters.  */
	register int	i, nused;
	double		dom;
	REFPARS		*xp;
	REFREF		*rp;
	WORKOB		wr;
	INWORK		*iwp;

	if (NULL == (iwp = manageiwp(7, rfp)))
	  {
		(rfp->rfr_rpp)->rflags = BADINDICES;	return 0;
	  }
	xp = rfp->rfr_xp;	rp = rfp->rfr_rpp;
	scattv(rfp);		/* compute local scattering vector for refl */
	refgonio(rfp);		/* make spindle-position rotation */
	if (!geth(rfp, iwp->iw_excluco))
	  {
		rp->rflags = BADINDICES;	return 0;
	  }
	if (9999. < (dom = newxyphi(&wr, rfp)))
	  {
		rp->rflags = BADSCATTVEC; return 0;
	  } /* X,Y, phi */
	if (dom > xp->xerrphi || fabs(rp->xto - rp->xtc) > xp->xerrx ||
	 fabs(rp->yto - rp->ytc) > xp->xerry)
	  {
		rp->rflags = BADXYPHI;	return 0;
	  }
	if (!legald(rfp, rp))
	  {
		rp->rflags |= INICERING;
		return 0;
	  }
	else rp->rflags &= ~INICERING;
	for (i = 3, nused = 0; i--; )
	   {
		iwp->iw_delh[i] = rp->rhi[i] - rp->rh[i];
		iwp->iw_useh[i] = (fabs(iwp->iw_delh[i]) < xp->xerrind);
		nused += iwp->iw_useh[i];
	   }
	rp->rflags = (nused < 3 ? BADINTEGER : OKREFL);
	return	nused;
}

static void xtal_derivs(float *rs, REFPARS *xp, INWORK *iwp)
{ /*	This routine and the next one compute the (rfp->rfr_npref)x3 matrix
	of first derivatives of hkl with respect to the (rfp->rfr_npref)
	parameters. In each case the derivative is computed only within the
	matrix where the parameter matters, and then is premultiplied or
	postmultiplied (or both) by the matrices dealing with the other
	components.  In this routine we're concerned with derivatives w.r.t.
	the crystal parameters: a,b,c,alpha,beta,gamma,omega,chi,phi.
	First, we get the derivatives w.r.t. the pseudogoniostat angles. */
	register int	i, j;	
	double		tz, rs0[3];

	mm31(iwp->iw_jj[6], iwp->iw_r0z0, rs); /* delta omega */
	mm31(iwp->iw_jj[7], iwp->iw_r0z1, rs); /* delta chi */
	mm31(iwp->iw_jj[8], iwp->iw_r0z2, rs); /* delta phi */
	/* rotate rs by the matrix q4, which has the effect of getting us
	 to the "standardized" position where a is along x, b is mostly
	 along y and secondarily along x, and c is wherever it has to be.
	 Then rescale rs0 by lambda to get back into inverse Angstroms. */
	for (i = 0; i < 3; i++)
	   {
		for (tz = j = 0; j < 3; j++) tz += iwp->iw_q4[i][j] * rs[j];
		rs0[i] = tz / xp->xlambda;
	   }
	/* The derivatives wrt the unit cell lengths and angles depend only
	on the elements of rs0.  They're Laue-group-dependent. */
	for (i = 6; i--; ) for (j = 3; j--; ) iwp->iw_jj[i][j] = 0;
	switch (xp->xisys)
	      {
		case	1:	/* triclinic: messiest case */
			iwp->iw_jj[0][0] = rs0[0];
			iwp->iw_jj[1][1] = rs0[0] * iwp->iw_cg +
			 rs0[1] * iwp->iw_sg;
			iwp->iw_jj[2][2] = rs0[0] * iwp->iw_cb -
			 rs0[1] * iwp->iw_sbcas + rs0[2] * iwp->iw_sbsas;
			iwp->iw_jj[5][1] = xp->xabc[1] * (rs0[1] * iwp->iw_cb -
				rs0[0] * iwp->iw_sb);
			 /* the partials leading to l are messy ... */
			for (i = 0; i < 3; i++)
			   {
				for (tz = j = 0; j < 3; j++)
					tz += iwp->iw_dlda[i][j] * rs0[j];
				iwp->iw_jj[3 + i][2] = tz;
			   }
			break;
		case	2:	/* monoclinic */
			iwp->iw_jj[0][0] = rs0[0];
			iwp->iw_jj[1][1] = rs0[1];
			iwp->iw_jj[2][2] = rs0[0] * iwp->iw_cb +
				rs0[2] * iwp->iw_sb;
			iwp->iw_jj[4][2] = -rs0[0] * iwp->iw_r2sb +
				rs0[2] * iwp->iw_r2cb;	break;
		case	3:	/* orthorhombic */
			for (i = 0; i < 3; i++) iwp->iw_jj[i][i] = rs0[i];
			break;
		case	4:	/* tetragonal */
			iwp->iw_jj[0][0] = rs0[0]; iwp->iw_jj[0][1] = rs0[1];
			iwp->iw_jj[2][2] = rs0[2]; break;
		case	5:	/* cubic */
			for (i = 0; i < 3; i++) iwp->iw_jj[0][i] = rs0[i];
			break;
		case	6:	/* trigonal or hexagonal */
			iwp->iw_jj[0][0] = rs0[0];
			iwp->iw_jj[0][1] = -rs0[0]/2. + rs0[0] * SQ3O2;
			iwp->iw_jj[2][2] = rs0[2]; break;
		case	7:	/* rhombohedral, indexed rhombohedrally */
			iwp->iw_jj[0][0] = rs0[0];
			iwp->iw_jj[0][1] = iwp->iw_cg * rs0[0] +
			 iwp->iw_sg * rs0[1];
			iwp->iw_jj[0][2] = iwp->iw_cg * rs0[0] -
			 iwp->iw_sbcas * rs0[1] + iwp->iw_sbsas * rs0[2];
			iwp->iw_jj[5][1] = xp->xabc[0] * (iwp->iw_cg * rs0[1] -
			 iwp->iw_sg * rs0[0]);
			iwp->iw_jj[5][2] = iwp->iw_dlda[2][0] * rs0[0] +
			 iwp->iw_dlda[2][1] * rs0[1] + iwp->iw_dlda[2][2];
			break;
	  }
}

static void detec_derivs(REFINERUN *rfp, INWORK *iwp)
{ /*	In this routine we're concerned with derivatives w.r.t.
	the detector parameters.  The first three parameters (plated,x0,y0)
	modify the vector `ss' (computed in "scattv"), so we get the
	derivs of `ss' with respect to them and then multiply those derivs
	thru by the matrices describing the rotations about tilt, swing,
	and spindle, and then by the unit cell.  */
	register int	i, j;
	double		rni3;
	float		ut[3], vt[3], t[3][3], tt[3][3];
	REFPARS		*xp;

	/* first get a 3x3 containing the partials of "rs" wrt the
	  first 3 detector parameters plateD, ax0, and ay0 */
	xp = rfp->rfr_xp;
	rni3 = rfp->rfr_rni * rfp->rfr_rni2;
	tt[0][0] = (rfp->rfr_rni2 - xp->xpltd2) / rni3;
	tt[0][1] = xp->xplated * rfp->rfr_xin / rni3;
	tt[0][2] = xp->xplated * rfp->rfr_yin / rni3;
	tt[1][0] = -rfp->rfr_xin * xp->xplated / rni3;
	tt[1][1] = (rfp->rfr_xin * rfp->rfr_xin - rfp->rfr_rni2) / rni3;
	tt[1][2] = rfp->rfr_xin * rfp->rfr_yin / rni3;
	tt[2][0] = -rfp->rfr_yin * xp->xplated / rni3;
	tt[2][1] = rfp->rfr_xin * rfp->rfr_yin / rni3;
	tt[2][2] = (rfp->rfr_yin * rfp->rfr_yin - rfp->rfr_rni2) / rni3;
	mm3(t, xp->xq21, tt);		/* rotate that 3x3 by tilt and swing */
	mm3(tt, rfp->rfr_q3, t);	/* rotate by spindle angle */
	mm3(t,  xp->xareal,tt);	/* multiply by real-space unit cell */
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
		iwp->iw_jj[j + 9][i] = t[i][j];
/* Now we'll do the tilt and swing derivatives.  This involves multiplying
	the vector `ss' by the appropriate derivative matrices */
	mm31(vt, iwp->iw_q2p1, rfp->rfr_ss);	/* deriv with respect to tilt*/
	mm31(ut, rfp->rfr_q3, vt); 	/* forward-multiply by the phi matrix*/
		/* ... and by unit cell param matrix */
	mm31(iwp->iw_jj[12], xp->xareal, ut);
	mm31(vt, iwp->iw_p2q1, rfp->rfr_ss); /* deriv with respect to tilt */
	mm31(ut, rfp->rfr_q3,vt);	/* forward-multiply by the phi matrix*/
	mm31(iwp->iw_jj[13], xp->xareal, ut); /* and by unit cell matrix */
}

static int accum_sums(char resid_type, double *res, REFINERUN *rfp, INWORK *iwp)
{ /* In the integerness case, this sets up for and adds the equations
	relating the partial derivatives w.r.t. the detector and crystal
	parameters to the deviations from integerness in the hkl values.
	In the scanning-angle case it sets up and adds the equation relating
	the partial derivatives w.r.t. the crystal parameters to the difference
	between the observed and calculated scanning-angle value. */

	register int	i, j, k, nused;
	double		jjt[3][14], jjtu[14];
	REFREF		*rp;

	*res = 0.;	rp = rfp->rfr_rpp;
	if (resid_type == 'L')
	  {
		if (!(nused = refl_coords(rfp))) return 0; /* coords of refl */
			/* derivs with respect to  the crystal parameters: */
		xtal_derivs(rp->rs, rfp->rfr_xp, iwp);
		detec_derivs(rfp, iwp); /* derivs w.r.t detector params */
		/* Now we've constructed the 14x3 matrix of derivatives;
		extract from it the rows that correspond to refinable
		parameters and make them into the COLUMNS of the actual
		3x(rfp->rfr_npref) derivative matrix. */
		for (i = rfp->rfr_npref; i--; )
		 for (j = 3, k = rfp->rfr_menulist[i]; j--; )
			jjt[j][i] = iwp->iw_jj[k][j];
		/* Now we add the three equations to the list of equations.
		If we're doing integerness calculations the 3 equations are
			h(ob) - h(calc) = SUM (@F(h,p)/@pi) * (dpi)
					   i
			k(ob) - k(calc) = SUM (@F(k,p)/@pi) * (dpi)
					   i
			l(ob) - l(calc) = SUM (@F(l,p)/@pi) * (dpi)
					   i
		where @ means "partial deriv" and dpi is the increment
		in parameter i.  */
		for (i = 0; i < 3; i++) if (iwp->iw_useh[i])
		   {
			inv_add(rfp->rfr_jtj, rfp->rfr_jtb, jjt[i],
			 (double)(iwp->iw_delh[i]), (int)(rfp->rfr_npref));
			*res += iwp->iw_delh[i] * iwp->iw_delh[i];
		   }
	  }
	else { /* scanning-angle or X,Y derivatives. */
		if (!(nused = xyo_coords(resid_type, rfp, iwp))) return 0;
		/* pick out just the variables we are refining */
		if (resid_type == 'K')
		  {
			for (i = rfp->rfr_npref; i--; )
				jjtu[i] = iwp->iw_jjsc[rfp->rfr_menulist[i]];
			/* The residual calculation is
			omega(ob) - omega(calc) = SUM (@F(h,p)/@pi) * (dpi) */
			inv_add(rfp->rfr_jtj, rfp->rfr_jtb, jjtu,
			 (double)(iwp->iw_delom), (int)(rfp->rfr_npref));
			*res += iwp->iw_delom * iwp->iw_delom;
		  }
		else { /* with an X,Y, or Z refinement, we minimize
			a residual involving the omega, X, and Y differences.
			with 'X' it's mostly (X,Y); with 'Y' it's balanced
			between (XY) and omega; with 'Z' it's mostly omega. */
			iwp->iw_delom *= iwp->iw_wtom;
			iwp->iw_delxo *= iwp->iw_wtxy;
			iwp->iw_delyo *= iwp->iw_wtxy;
			for (i = rfp->rfr_npref; i--; ) jjtu[i] = iwp->iw_wtom
			 * iwp->iw_jjsc[rfp->rfr_menulist[i]];
			inv_add(rfp->rfr_jtj, rfp->rfr_jtj, jjtu,
			 (double)(iwp->iw_delom), (int)(rfp->rfr_npref));
			*res += iwp->iw_delom * iwp->iw_delom;
			for (i = rfp->rfr_npref; i--; ) jjtu[i] =
			 iwp->iw_wtxy * iwp->iw_jjx[rfp->rfr_menulist[i]];
			inv_add(rfp->rfr_jtj, rfp->rfr_jtb, jjtu,
			 (double)(iwp->iw_delxo), (int)(rfp->rfr_npref));
			*res += iwp->iw_delxo * iwp->iw_delxo;
			for (i = rfp->rfr_npref; i--; ) jjtu[i] =
			 iwp->iw_wtxy * iwp->iw_jjy[rfp->rfr_menulist[i]];
			inv_add(rfp->rfr_jtj, rfp->rfr_jtb, jjtu,
			 (double)(iwp->iw_delyo), (int)(rfp->rfr_npref));
			*res += iwp->iw_delyo * iwp->iw_delyo;
		     }
	     }
	return nused;
}

void apply_dels(REFINERUN *rfp)
{ /*	This updates the parameter values with the delx increments. */
	register int	i, j, m;
	REFPARS		*xp;
	INWORK		*iwp;
	static float	maxshift[] = {1.,1.,1.,1.,1.,1.,1.,1.,1.,
				0.2,0.05,0.05,0.2,0.2 };

	if (NULL == rfp)
	  {	
		(void)manageiwp(0, NULL);	return;
	  }
	/* first, relate the solutions "delx" to the shifts "rfp->rfr_pshift"*/
	if (NULL == (iwp = manageiwp(7, rfp))) return;
	for (i = ALLPARS; i--; )	rfp->rfr_pshift[i] = 0.;
	xp = rfp->rfr_xp;
	for (j = rfp->rfr_npref; j--; )
	   {
		m = rfp->rfr_menulist[j];
		rfp->rfr_pshift[m] = (m < NPAR0) ?
			rfp->rfr_delt[j] * iwp->iw_msca : rfp->rfr_delt[j];
	   }
	/* Rescale the angular changes by 180/pi, tone down the
	 huge changes, and apply them to the "temporary" structure */
	for (i = ALLPARS-3; i--; ) /* tone down huge changes & apply changes*/
	   if (xp->xisref[i] == 'Y')
	     {
		if (i < NPAR0)
		  {
			j = i + NPAR0;	m = j + NPAR0;
		  }
		else {
			j = i + 2 * NPAR0 + NPAR1; m = j + NPAR1;
		     }
		if ((3 <= i && i <= 8) || i == 12 || i == 13)
			rfp->rfr_pshift[i] /= RADC;
		if (rfp->rfr_pshift[i] > maxshift[i])
			rfp->rfr_pshift[i] = maxshift[i];
		else if (rfp->rfr_pshift[i] < -maxshift[i])
			rfp->rfr_pshift[i] = -maxshift[i];
	     }
}

int incr_cycle(char resid_type, REFINERUN *rfp)
{ /* performs a cycle of refinement */
	register int	i, j, nused;
	double		lbtb, res;
	REFPARS		*xp;
	REFREF		*lrp;
	INWORK		*iwp;

	if (NULL == (iwp = manageiwp(7, rfp))) return -1;
	inv_init(rfp->rfr_jtj, rfp->rfr_jtb, rfp->rfr_npref); /* initialize */
	xp = rfp->rfr_xp;
	scalars(rfp);	/* get the constant scalars for this pass*/
	matrices(rfp);	/* get the constant matrices for this pass*/
	if (resid_type == 'L') iwp->iw_msca = 1.; /* no rescaling this type */
		/* globals used in `geth' */
	if (xp->xisys < 3 || xp->xisys == 7) gethfacs(rfp);
	iwp->iw_wtom = (resid_type == 'Z') ? 0.8 : 0.4;
	iwp->iw_wtxy = 1. - iwp->iw_wtom;
	for (lrp = rfp->rfr_rpp0 + rfp->rfr_firstok,
	 i = nused = lbtb = 0; i < rfp->rfr_inbounds; lrp++, i++)
	   {
		if (legald(rfp, lrp))
		  {
			rfp->rfr_rpp = lrp;	lrp->rflags &= ~INICERING;
			if ((lrp->rflags < BADXYPHI) &&
			  (j = accum_sums(resid_type, &res, rfp, iwp)))
			  {
				lbtb += res;	nused += j;
			  }
		   }
		else lrp->rflags |= INICERING;
	   }
	if (nused > rfp->rfr_npref)
	  {
		lbtb /= (double)nused; /*compute rms residual*/
		inv_solve(rfp->rfr_jtj, rfp->rfr_jtb, rfp->rfr_delt,
			rfp->rfr_npref);	/* solve equations */
		rfp->rfr_btb = lbtb;
		return 0;
	  }
	else {
		fprintf(rfp->rfr_fpl,
		 "linear, type %c: only %3d refls used, %3d params\n",
			resid_type, nused, (int)(rfp->rfr_npref));
		rfp->rfr_btb = 1.e4;	return 1;
	     }
}

double newresid(char residtype, int *nchanged, REFINERUN *rfp)
{ /*	This computes the post-refinement residual for the reflections
	actually used in the refinement AND it flags changed refls.
		a refl is defined as changed if its integerized hkl
		if different from before AND it is on the opposite side of
		the unrefined, unintegerized hkl rel to the orig. one. */
	register int	i, j, ngoodind;
	double		delp, nbtb, nhij, nhj, oh[3], ohi[3];
	REFPARS		*xp;
	REFREF		*lrp;
	WORKOB		wr;
	INWORK		*iwp;

	if (NULL == (iwp = manageiwp(3, rfp))) return 10000.;
	xp = rfp->rfr_xp;	/* globals used in `geth': */
	if (xp->xisys < 3 || xp->xisys == 7) gethfacs(rfp);
	iwp->iw_excluco = exclucode(xp->xispgp, xp->xisys);
	lrp = rfp->rfr_rpp0 + rfp->rfr_firstok;	nbtb = *nchanged = ngoodind = 0;
	for (i = 0; i < rfp->rfr_inbounds; lrp++, i++)
	   if (lrp->rflags < BADXYPHI)
	     {
		rfp->rfr_rpp = lrp;
		scattv(rfp);	refgonio(rfp);
		if (!legald(rfp, lrp))
		  {
			lrp->rflags |= INICERING;	continue;
		  }
		lrp->rflags &= ~INICERING;
		for (j = 3; j--;)
		   {
			oh[j] = lrp->rh[j]; ohi[j] = lrp->rhi[j];
		   }
		if (geth(rfp, iwp->iw_excluco))
		  {
			if (residtype == 'L')
			  {
				for (j = 3; j--;)
				   {
					nhij = lrp->rhi[j];
					nhj = lrp->rh[j];
					if ((ohi[j] < oh[j] && oh[j] < nhij) ||
					    (nhij < oh[j] && oh[j] < ohi[j]))
					  {
						(*nchanged)++;
						lrp->rflags = CHANGEDINDEX;
					  }
					if (fabs(oh[j] -ohi[j]) <xp->xerrind)
					  {
						ngoodind++; nbtb +=
						 (nhj - nhij) * (nhj - nhij);
					  }
				  }
			  }
			else if (residtype == 'K')
			  {
				if (9999. < (delp = newxyphi(&wr, rfp)))
				  { /* X,Y, phi */
					lrp->rflags = BADSCATTVEC; return 0;
				  }
				delp *= xp->xfrmsize;
				ngoodind++;	nbtb += delp * delp;
			  }
			else {
				if (9999. < (delp = newxyphi(&wr, rfp)))
				  {
					lrp->rflags = BADSCATTVEC; return 0;
				  } /* X,Y, phi */
				delp *= xp->xfrmsize;
				ngoodind++; nbtb += iwp->iw_wtom *
				 iwp->iw_wtom * delp * delp;
				delp = lrp->xtcm - wr.rxcm;
				nbtb += iwp->iw_wtxy * iwp->iw_wtxy *
					delp * delp;
				delp = lrp->ytcm - wr.rycm;
				nbtb += iwp->iw_wtxy * iwp->iw_wtxy *
					delp * delp;
			  }
		  }
		else lrp->rflags = BADINDICES;
	     }
	nbtb = (ngoodind ? nbtb / (double)ngoodind : 10000.);
	return nbtb;
}

/* end of incrcalc functionalities */

/****************************************
	nonlinear: calculations used in simplex refinement in XGEN
************************************/
#include	"refidefs.h"
#define		ANGLEDEL	0.1	/* starting deviation of angles */
#define		LENGTHDEL	(0.001) /* "          " ratio for lengths*/
#define		LIMITRAT	5	/* see its use */
#define		DIVR		(0.2)	/* how much to shrink delta by */

 typedef struct {
		double	rnl_resh0, rnl_resphi0, rnl_resx0, rnl_resy0;
		double	rnl_indwt, rnl_xywt, rnl_prdp[5];
		double	rnl_xmi0[THESEPARS], rnl_xma0[THESEPARS];
		double	rnl_xnew0[THESEPARS], rnl_delta0[THESEPARS];
		char	rnl_reftype;
		} NONLWORK; 

static NONLWORK *managenlw(int opno, REFINERUN *rfp)
{ /*	This manages a variable *rnlw according to opno */
	double		inires[4];
	static NONLWORK	*rnlw;
	static char	rnlmsg[] = "rnlw in managenlw";

	if (!opno)
	  {
		if (NULL != rnlw)
		  {
			memzap((AVOID **)&rnlw, rnlmsg); return NULL;
		  }
	  }
	if (opno == 1) (void)managenlw(0, NULL);
	if (NULL == rnlw) rnlw = callop(1, sizeof (NONLWORK), rnlmsg);
	if (opno & 04)	/* initialize various things */
	  {
		if ((rnlw->rnl_indwt < DFUZZ) && (rnlw->rnl_xywt < DFUZZ))
			(void)initmin(0.7, 0.3, inires, 'A', rfp);
	  }
	return rnlw;
}

void qnonlin1(void)
{ /* This is called on a ^C inside `minpar1' or on a normal finishnl1. */
	REFINERUN	*rfp;

	signal(SIGINT, SIG_DFL); 	/* restore signal value of ^C */
	if (NULL != (rfp = managerfp(3, 0))) rfp->rfr_flags |= RR_QUITNONLIN;
	(void)managenlw(0, NULL);	/* destroy old rnlw structure */
}

static double residcal(REFINERUN *rfp)
{ /*	This returns an rms residual for a set of refinable reflections.
	Depending on which type of refinement the user is doing,
	the user can get back a spindle-angle residual, an integerness
	residual, or a weighted sum of the spindle angle residual, the
	integerness residual, and the X and Y residuals.  That is,
	the number returned is
		SUM (wi^2)*Di
		 i
	R   =   -----------------
		SUM (wi^2)
		 i
	where Di = (Mi-mi)^2 in one case, (hi-[hi])^2 in another, and
	(pi(Mi-mi)^2 + qi(hi-[hi])^2 + ri(Xi-xi)^2 + ri(Yi-yi)^2, with
	Mi,Xi,Yi the observed phi,X,Y values, and mi,xi,yi the computed ones;
	[hi] is the integerized version of hi and pi,qi,ri are weights.
	The weight factors wi are currently 0 for illegal refls and 1 for
	legals.  `nused' is the number of observations found to be legal.
	Note that we don't compute the hkl using `geth' here; the integerized
	version of hkl is assumed to be correct before we start cycling. */
	register int	i, j, nok;
	register double	*zp0, *zp;
	char		reftype;
	double		difh, difphi, difx, dify, dhok, dok, hweight, poswt;
	double		resh, resphi, resx, resy, retres, rwt, sumwt, thisdh;
	REFPARS		*xp;
	REFREF		*lrp;
	NONLWORK	*rnlw;

	if (NULL == (rnlw = managenlw(3, rfp))) return 10000.;
	xp = rfp->rfr_xp;	zp0 = rnlw->rnl_xnew0;
	for (i = 0, zp = zp0; i < 3; i++, zp++) xp->xabc[i] = *zp;
	for (i = 0; i < 3; i++, zp++) xp->xang[i] = *zp;
	xp->xdelom = *zp++;	xp->xdelchi = *zp++;	xp->xdelphi = *zp++;
	(rfp->rfr_nrescalc)++;
	if (unitcellx(rfp)) return 10000.; /* awful residual on failure */
	/* we need to redefine the scattering vectors iff. we're refining some
	detector parameters.  We determine whether or not that's the case
	by checking the five detector params against their previous values. */
	for (j = 0; j < 5; j++)
	   if ((xp->xisref[j + NPAR0] == 'Y') &&
		(fabs(rnlw->rnl_prdp[j] - zp[j]) > DFUZZ))
	     {
		for (i = 0; i < 5; i++) rnlw->rnl_prdp[i] = zp[i];
		xp->xplated = *zp++;	xp->xx0 = *zp++; xp->xy0 = *zp++;
		xp->xtilt = (*zp++) * RADC;	xp->xswing = (*zp++) * RADC;
		xp->xpltd2 = xp->xplated * xp->xplated;	makeq21(xp);
		lrp = rfp->rfr_rpp0 + rfp->rfr_firstok;
		for (i = rfp->rfr_inbounds; i--; lrp++)
		   {
			if (legald(rfp, lrp)) lrp->rflags &= ~INICERING;
			else lrp->rflags |= INICERING;
			if (!lrp->rflags)
			  {
				rfp->rfr_rpp = lrp; scattv(rfp); refgonio(rfp);
			  }
		   }
		break;
	     }
	zp = zp0;
	for (resphi = resx = resy = resh = dok = dhok = nok = 0, i =
	 rfp->rfr_inbounds, lrp = rfp->rfr_rpp0 + rfp->rfr_firstok; i--; lrp++)
	   if (!lrp->rflags)
	     {
		rfp->rfr_rpp = lrp;
		rwt = lrp->totinten;
		for (j = 3, difh = 0.; j--; )
		   {
			lrp->rsi[j] = xp->xrec[j][0] * lrp->rhi[0] +
				xp->xrec[j][1] * lrp->rhi[1] +
				xp->xrec[j][2] * lrp->rhi[2];
			thisdh = xp->xareal[j][0] * lrp->rs[0] + 
			 xp->xareal[j][1] * lrp->rs[1] +
			 xp->xareal[j][2] * lrp->rs[2] - lrp->rhi[j];
			difh += thisdh * thisdh;
		   }
		resh += difh * rwt; /*increment numerator for h resid*/
		dhok += 3. * rwt;	/* denominator for h resid */
		if (!s2xyp(lrp, rfp))
		  {
			difphi = gcloser(lrp, rfp);
			difx = lrp->xto - lrp->xtc;
			dify = lrp->yto - lrp->ytc;
			dok += rwt; resphi += difphi * difphi * rwt;
			resx += difx * difx * rwt;
			resy += dify * dify * rwt;	nok++;
		  }
	     }
	if (!nok) return 10000.;	/* awful residual if no refls */
	resh /= dhok;	resphi /= dok;	resx /= dok;	resy /= dok;
	if (rfp->rfr_nrescalc == 1)	/* record initial residuals */
	  {
		rnlw->rnl_resh0 = resh;	rnlw->rnl_resx0 = resx;
		rnlw->rnl_resy0 = resy;	rnlw->rnl_resphi0 = resphi;
	  }
	/* rescale the x, y, phi residuals appropriately */
	resx *= rnlw->rnl_resh0 / rnlw->rnl_resx0;
	resy *= rnlw->rnl_resh0 /rnlw->rnl_resy0;
	resphi *= rnlw->rnl_resh0 / rnlw->rnl_resphi0;
	reftype = rnlw->rnl_reftype;
	if ('0' <= reftype && reftype <= '9')
	  {
		hweight = rnlw->rnl_indwt * 0.01;
		poswt = rnlw->rnl_xywt * 0.01;
		/* guarantee well-conditioned weights */
		if (hweight < 0.) hweight = 0;	
		else if (hweight > 1.) hweight = 1.;
		if (poswt < 0.) poswt = 0;
		else if (poswt > 1.) poswt = 1.;
		if ((sumwt = hweight + poswt) > 1.)
		  {
			hweight /= sumwt;	poswt /= sumwt;
		  }
		retres = hweight * resh + poswt * (resx + resy) +
			(1. - hweight - poswt) * resphi;
	  }
	else switch (reftype)
	    {
		case	'P':	case	'0':	/* spindle-angle refinement */
			retres = resphi;	break;
		case	'H':	case	'9':	/* integerness refinement */
			retres = resh;		break;
		case	'A':	default:	/* everything */
			retres = 0.45 * (resh + resphi) + 0.05 * (resx + resy);
	    }
	if (retres < rfp->rfr_bestres)
	  {
		for (i = THESEPARS; i--; ) rfp->rfr_bestsofar[i] = zp0[i];
	    	rfp->rfr_bestres = retres;
		rfp->rfr_bestphi = resphi;	rfp->rfr_besth = resh;
		rfp->rfr_bestx = resx;		rfp->rfr_besty = resy;
		rfp->rfr_bestnr =  nok;
	  }
	return retres;
}

double initmin(double indwt, double xywt, double *initresid,
		char reftype, REFINERUN *rfp)
{ /*	This starts up a run of `minpar' and gets the initial residual. */
	register int	i;
	double		retv, *delta, *xmin, *xmax, *xnew, *xnow;
	NONLWORK	*rnlw;
	static double	del[] = { 0.05,0.02,0.02,0.1,0.1 };

	rfp->rfr_nrescalc = 0;	/* set # of times "resid" is called to 0*/
		/* ready to do nonlinear refinement: */
	rfp->rfr_flags &= ~(RR_QUITNONLIN);
	if (NULL == (rnlw = managenlw(3, rfp))) return 10000.;
	xnew = &(rnlw->rnl_xnew0[0]);	xnow = &(rfp->rfr_xnow0[0]);
	delta = &(rnlw->rnl_delta0[0]);	xmin = &(rnlw->rnl_xmi0[0]);
	xmax = &(rnlw->rnl_xma0[0]);	rnlw->rnl_reftype = reftype;
	rnlw->rnl_indwt = indwt;	rnlw->rnl_xywt = xywt;
	rfp->rfr_bestres = 20000.;	/* starting ridiculous residual */
	 /* make sure scattering vectors aren't redefined on 1st pass */
	for (i = 0; i < 5; i++) rnlw->rnl_prdp[i] = rfp->rfr_xnow0[NPAR0 + i];
	for (i = 0; i < THESEPARS; i++, xnew++, xnow++, delta++)
	   {
		*xnew = *xnow;
		*delta = (i < 3 ? LENGTHDEL * (*xnow) : (i < NPAR0 ?
			ANGLEDEL : del[i-NPAR0]));
		*xmax++ = *xnow + LIMITRAT * (*delta);  /* max to move up */
		*xmin++ = *xnow - LIMITRAT * (*delta);	/*max to move down*/
	   }
	retv = residcal(rfp);
	initresid[0] = rnlw->rnl_resh0;	initresid[1] = rnlw->rnl_resx0;
	initresid[2] = rnlw->rnl_resy0;	initresid[3] = rnlw->rnl_resphi0;
	return retv;
}

int minph1(double indwt, double xywt, char reftype,
	int doshrink, REFINERUN *rfp)
{ /* Phase I of the minimization process.  This routine returns 1
	if it's received a signal that it is time to quit nonlinear
	refinement; otherwise it returns 0. */
	register int	i;
	double		fnew, *delta, *xmax, *xmin, *xnew, *xnow;
	REFPARS		*xp;
	NONLWORK	*rnlw;

	if (NULL == (rnlw = managenlw(7, rfp))) return 1;
	xnew = &(rnlw->rnl_xnew0[0]);	xnow = &(rfp->rfr_xnow0[0]);
	delta = &(rnlw->rnl_delta0[0]);	xmin = &(rnlw->rnl_xmi0[0]);
	xmax = &(rnlw->rnl_xma0[0]);
	/* if appropriate, we shrink all the delta values at this point */
	if (doshrink)
	  {
		for (i = 0, delta = &(rnlw->rnl_delta0[0]);
		 i <THESEPARS; i++, delta++) *delta *= DIVR;
		delta -= THESEPARS;
	  }
	/* move params one at a time up by delta and down by delta */
	xp = rfp->rfr_xp;	rnlw->rnl_reftype = reftype;
	rnlw->rnl_indwt = indwt;	rnlw->rnl_xywt = xywt;
	for (i = 0; i < THESEPARS; i++, xnew++, xnow++, xmax++, xmin++,delta++) 
	 if (xp->xisref[i] == 'Y')
	   {
		*xnew += *delta;	if (*xnew > *xmax) *xnew = *xmax;
		fnew = residcal(rfp);
		if (rfp->rfr_flags & RR_QUITNONLIN) return 1;
		if (fnew < rfp->rfr_fnow) rfp->rfr_fnow = fnew;
		else
		  {
			*xnew = *xnow - *delta;
			if (*xnew < *xmin) *xnew = *xmin;
			fnew = residcal(rfp);
			if (rfp->rfr_flags & RR_QUITNONLIN) return 1;
			if (fnew < rfp->rfr_fnow)	rfp->rfr_fnow = fnew;
			else			*xnew = *xnow;
		  }
	   }
	return 0;
}

int minph2(double indwt, double xywt, char reftype, REFINERUN *rfp)
{ /* Phase II of the minimization process. Again, this routine
	returns 1 iff. it receives a quit signal. */
	register int	i;
	double		fnew, t, *xmax, *xmin, *xnew, *xnow;
	REFPARS		*xp;
	NONLWORK	*rnlw;

	if (NULL == (rnlw = managenlw(7, rfp))) return 1;
	xp = rfp->rfr_xp;
	xnew = &(rnlw->rnl_xnew0[0]);	xnow = &(rfp->rfr_xnow0[0]);
	xmin = &(rnlw->rnl_xmi0[0]);	xmax = &(rnlw->rnl_xma0[0]);
	rnlw->rnl_reftype = reftype;
	rnlw->rnl_indwt = indwt;	rnlw->rnl_xywt = xywt;
/* if the current residual is better than the old one, we continue
   the refinement by checking some other values of x...  */
	while(1) /* move params to (twice current val - old val)*/
	      {
		for (i = 0; i < THESEPARS; i++, xnow++, xnew++, xmin++, xmax++)
		 if (xp->xisref[i] == 'Y')
		   {
			t = *xnow; *xnow = *xnew; *xnew = *xnew + *xnew - t;
			if (*xnew > *xmax) *xnew = *xmax;
			if (*xnew < *xmin) *xnew = *xmin;
		   }
		fnew = residcal(rfp);
		/* grumble. Linux gets stuck in this loop somewhere
		 if we don't tickle the filesystem a little */
		(void)xeflush(rfp->rfr_fpl);
		if (rfp->rfr_flags & RR_QUITNONLIN) return 1;
		xnew = &(rnlw->rnl_xnew0[0]);	xnow = &(rfp->rfr_xnow0[0]);
		xmax = &(rnlw->rnl_xma0[0]);	xmin = &(rnlw->rnl_xmi0[0]);
		if (fnew >= rfp->rfr_fnow) break;
		rfp->rfr_fnow = fnew;
	      }
/* when that particular trick has exhausted its usefulness, store current
   param values in xnew and evaluate refinement progress */
	for (i = 0; i < THESEPARS; i++, xnew++, xnow++)
		if (xp->xisref[i] == 'Y') *xnew = *xnow;
	return 0;
}

/* end of nonlinear functionalities */

/* end of refrefine.c */
