/* refutil.c -- calculational utilities for XGEN's "refine"

	Copyright (c) 2002, Illinois Institute of Technology
	X-GEN: Crystallographic Data Processing Software
	author: Andrew J Howard, BCPS Department, IIT
 See the file "LICENSE" for information on usage and redistribution
	of this file, and for a DISCLAIMER OF ALL WARRANTIES.

Change record:
	25 Jun 2007: some comparisons with epsilon use DFUZZ rather than FUZZ
	19 Apr 2007: float->double conversions; reorganize use of "{"
	27 Dec 2006: legald moved here; 'P' and 'H' used differently
	29 Nov 2006: xp->xast,xanst,xq21 used; derivedglobs eliminated
	10 Jun 2006: dspace() moved here from xgparams so it can be static
	 9 Jun 2006: acav2str,bravaisspec,refjournal,ref1args,updateorigv
 		moved to refdriver.c;
		gethfacs,getinthkl,s2xyp,gcloser,managerfp moved around
	18 May 2006: add refresid.c, refxtal.c, refinit.c functions into here;
		geth(): eliminate rp arg;
		getinthkl(): eliminate all args except the REFINERUN struct
	 4 Mar 2006: in refgonio & gcloser, REFREF element is framem, not frame
	31 Dec 2005: get_rmserrs(): don't qsort if fewer than 2 elements
	 9 Dec 2005: xstclean() moved from here to refio.c
	26 Oct 2005: replace str2xgbuf() calls with xgbuf() calls
	22 Jun 2005: getprofvals(), scclosest() now declared as static
	 2 Feb 2005: use "memzap((AVOID **)" throughout
	17 Jan 2005: report_iteration() announces progress only if ifatty(0).
	28 Aug 2004: routines used only here are declared as static; cosmetics;
		closeto90() deleted; incr1p() reordered
	28 Aug 2004: fmtoupr() moved to ../xglib/xgparams.c
	 4 Jun 2004: angle_penalty() and volume_penalty() routines now here;
		best_solution() now uses them to calc gfom = gresid + stuff
	 4 Jun 2004: get_rmserrs(): always recalculate diffraction vectors
		this eliminates the need for the RR_CHANGEDDET concept
	 4 Jun 2004: fmtoupr() now called with a flags argument:
		REFINERUN structure zeroed elsewhere, not in refinit()
	30 May 2004: move ref1p() from elsewhere to here
	26 May 2004: default: turn off refinement of plated, tilt swing, gamma2
	 2 May 2004: refinit: check to see if arg to "j" command is present
		before asking for input to allnumeric
	26 Jan 2004: globtotemp() deleted because it wasn't doing anything;
		some static globals replaced with local variables
	21 Jan 2004: license message inserted
	 3 Jan 2004: REFINERUN and REFPARS structure elements in function calls
		replace remaining global variables; function calls updated;
		eliminate temptoglob(), which no longer does anything;
		kill_ftab called after images read; assign_pname used
	13 Dec 2003: remove unused variable newb in incr1()
	10 Dec 2003: remove references to framdeg
	27 Nov 2003: rphi[] replaces references to deltaom, deltachi, deltaphi
	26 Nov 2003: replace global variables with elements of REFINERUN struct
	 3 Nov 2003: moved Bravais lattice calculations to ../xglib/xgbravais.c
		and other generally useful routines to ../xglib/xgenmisc.c
	18 Sep 2003: all external X-GEN routines declared here, not in xdefs.h
	22 Aug 2003: add guessvol() and closeto90();
		use these in significantly rewritten best_solution();
		reorder a few other routines so dependencies work
	16 Aug 2003: geth: make exclusion code for keephkl an argument
	16 Jan 2003: allow R command to have a "# cycles" argument
	14 Aug 2003: memory for array of difference vectors is now
		freed in indhkl, so it doesn't need to be freed here.
	13 Aug 2003: major reorganization: many routines moved to
		refxtal.c so that refutil.o doesn't need to be bound
	 5 Jan 2003: reorganized source-file header
	 2 Jan 2003: reactivate signal handler
	15 Nov 2002: extern declarations made here rather than in refiext.h
	 1 Nov 2002: add comments showing which lattice type has which matrix
	 1 Oct 2002: syntactic trick: commands D and J can be reversed
	25 Mar 2002: removed MSI attachments
	 1 Mar 2002: move useful routines from refgui.c here
		(temptoglob(), globtotemp(), dels_to_globs()
	 2 Jun 1995: minpar1: `fold' is no longer a global variable
	31 May 1995: "register" -> "register int" throughout;
		flush fplog after write
	 2 Mar 1995: initial creation of source file refinit.c from refmain.c
	 2 Mar 1995: major rewrite;
		delete permute_axes(), save_hkl(), picksetup()
	 5 May 1994: source-file header cleaned up,
		 names changed from POLYDET to XGEN
	16 Mar 1994: global variables reorganized a little
	14 Jan 1994: new memory management and user interface tools
	 7 Dec 1993: initial X-GEN versions
	xx xxx 1990: in matglob(), rotate down from the true goniostat
		(0,0,0) to the observed start position.
   Purpose: utilities associated with parameter refinement
   Cross references: ref*.c, ../xglib/xgbravais.c

*************************************************************************/
#define		NEED_UNISTD	1
#define		NEED_TIME	1
#include	"refidefs.h"
#ifdef VMS
#include	"vsort.h"
#endif
#define		FLOOR		0.10	/* weight-factor floor in geth */
#define		NSHRINKS	4	/*# of times to shrink deltas*/
#define		SHRINKDEL	(0.5)	/* shrink factor per cycle */
#define		ITER		5	/* # of iterations to perform */
#define		ROUGH_ANGLE	1.4	/* approximate angle equivalence */
#define		ROUGHLY_SAME	1.2	/* auto-indexing residuals are equiv */
#define		RESFLOOR	0.02	/* floor for residual comparisons */

/*******************************************
	refresid: These first few routines are associated with
	computing the RMS residuals associated with refinement
*******************************************/
double getinthkl(REFINERUN *rfp)
{ /* this takes the scattering vector `(rfp->rfr_rpp)->rs' and computes
	the corresponding non-integer index vector (rfp->rfr_rpp)->rh.
	It then computes the best integerization of it.
	It returns the noninteger index vector in (rfp->rfr_rpp)->rh,
	the integerized vector in (rfp->rfr_rpp)->rhi,
	and the scattering vector corresponding to that in (rfp->rpp)->rsi.
	The function returns
	 sum (hv[i]^2)
	  i
	 -------------- * sum ((sv[i] - svi[i])^2)
	 sum (sv[i]^2)
	  i				(here "^2" denotes "squared"). */

	register int	epsi, i, j, zet;
	double		dsnorm, hnorm, hr, phimax, psimin, snorm, sr, sri;
	double		zphi[3];
	float		dhv[3], zpsi[3], lar[3][3], lre[3][3];
	REFPARS		*xp;
	REFREF		*rp;

	xp = rfp->rfr_xp;	rp = rfp->rfr_rpp;
	for (i = 0; i < 3; i++) for (j = 0; j < 3; j++)
	   { /* note that lre[i][j] is unitless */
		lar[i][j] = xp->xareal[i][j]; lre[i][j] = xp->xrec[i][j];
	   }
	mm31(rp->rh, lar, rp->rs);	/* compute the index vector */
	for (j = 3; j--; )	/* integerize the h vector */
	   {
		rp->rhi[j] = ((hr = rp->rh[j]) < 0. ? (int)(hr - 0.4999) :
			(int)(hr + 0.4999)); dhv[j] = rp->rhi[j] - hr;
	   }
	switch (xp->xisys)
	      {
		case	1:	/* triclinic */
		case	7:	/* rhombohedral, indexed rhombohedrally */
		/* compute sums associated with max deviation from integer*/
			mm31(zpsi, rfp->rfr_dhc, dhv);
			for (j = epsi = zet = 3, psimin = phimax = 0; j--; )
			   {
				zpsi[j] += rfp->rfr_rasq[j];
				zphi[j] = zpsi[j] - 2. * rfp->rfr_rasq[j];
				if (zpsi[j] < psimin)
				  {
					epsi = j;	psimin = zpsi[j];
				  }
				if (zphi[j] > phimax) 
				  {
					zet = j;	phimax = zphi[j];
				  }
			   }
			/* if the sum associated with the largest down-side
			deviation--"psimin"--is negative, we need to increase
			that particular index by one. */
			if (psimin < 0)
			  {
				rp->rhi[epsi] += 1; dhv[epsi] += 1;
			  }
			/* if the sum assoc. with the largest up-side deviation
			is positive, we decrease THAT index by one */
			else if (phimax > 0)
			  {
				rp->rhi[zet] -= 1; dhv[zet] -= 1;
			  }
			break;
		case	2:	/* monoclinic: we only check h and l */
			zpsi[0] = rfp->rfr_dhc[0][0] * dhv[0] +
			 rfp->rfr_dhc[0][2] * dhv[2] + rfp->rfr_rasq[0];
			zpsi[2] = rfp->rfr_dhc[2][0] * dhv[0] +
			 rfp->rfr_dhc[2][2] * dhv[2] + rfp->rfr_rasq[2];
			epsi = (zpsi[0] < zpsi[2] ? 0 : 2);
			if (zpsi[epsi] < 0)
			  {
				rp->rhi[epsi] += 1; dhv[epsi] += 1;
			  }
			else
			  {
				zphi[0] = zpsi[0] - 2. * rfp->rfr_rasq[0];
				zphi[2] = zpsi[2] - 2. * rfp->rfr_rasq[2];
				zet = (zphi[0] > zphi[2] ? 0 : 2);
				if (zphi[zet] > 0)
				  {
					rp->rhi[zet] -= 1; dhv[zet] -= 1;
				  }
			  }
			break;
		case	6:	/* hexagonal: check h and k (note a = b) */
			zpsi[0] = 2. * dhv[0] + dhv[1] + 1;
			zpsi[1] = 2. * dhv[1] + dhv[0] + 1;
			epsi = (zpsi[0] < zpsi[1] ? 0 : 1);
			if (zpsi[epsi] < 0)
			  {
				rp->rhi[epsi] += 1;dhv[epsi] += 1;
			  }
			else if (zpsi[(zet = 1 - epsi)] > 2)
			  {
				rp->rhi[zet] -= 1; dhv[zet] -= 1;
			  }
			break;
		default:	/* other spacegroups: leave it as it is */
			break;
	      }
	mm31(rp->rsi, lre, rp->rhi);
	for (j = 3, hnorm = snorm = dsnorm = 0; j--; )
	   {
		hr = rp->rhi[j];	hnorm += hr * hr;
		sr = rp->rs[j];	sri = rp->rsi[j];	snorm += sr * sr;
		dsnorm += (sri - sr) * (sri - sr);
	   }
	return (snorm < DFUZZ) ? DFUZZ : (dsnorm / snorm) * hnorm;
}

int geth(REFINERUN *rfp, short excluc)
{ /*	This computes the index vector h from the local scattering vector s
	and the real-space unit-cell matrix, increments some global vectors
	which define frequency distributions of deviations from integerness,
	and obtains the weight-factor associated with this reflection.  It
	also computes the scattering vector associated with the INTEGERIZED
	version of this hkl.  The routine returns 1 iff the resultant hkl is
	legal in the spacegroup.  */
	short		shhkl[3];
	REFREF		*rp;

	/* get the index vector in unintegerized and integerized forms.
	Note that the resulting weight factor is a function of the deviation
	of the reflection from integerness AND of the log of its intensity */
	rp = rfp->rfr_rpp;
	rp->wt = rp->totinten / (FLOOR + getinthkl(rfp));
	shhkl[0] = rp->rhi[0]; shhkl[1] = rp->rhi[1]; shhkl[2] = rp->rhi[2];
	return keephkl(shhkl, excluc);
}

static void pr_cycle_header(FILE *fp, REFINERUN *rfp)
{ /*	This puts up a header in the logfile to make sense of
	subsequent cycles of refinement */
	fprintf(fp, "  refinement type      a      b      c  alpha   beta");
	fprintf(fp, "  gamma  omega    chi    phi platD  Xcen  Ycen  tilt");
	fprintf(fp, " swing   gam0   gam1   gam2\n");
	fprintf(fp, "           Residuals:  index      X      Y    phi");
	fprintf(fp, " gamma profile fm moved #obs #indices\n");
	/* now we don't need the header */
	if (fp != stdout) rfp->rfr_flags &= ~RR_NEEDHEADER;
}

void sum_curparam(FILE *fp, char *s, REFINERUN *rfp)
{ /* This prints a quick summary of the current parameters and results */
	REFPARS	*xp;

	xp = rfp->rfr_xp;
	fprintf(fp, "%17s%7.2f%7.2f%7.2f%7.2f%7.2f%7.2f%7.2f%7.2f%7.2f",
	 s, xp->xabc[0], xp->xabc[1], xp->xabc[2],
		xp->xang[0], xp->xang[1], xp->xang[2],
		xp->xdelom, xp->xdelchi, xp->xdelphi);
	fprintf(fp, "%6.2f%6.3f%6.3f%6.3f%6.2f%7.3f%7.3f%7.4f\n",
	 xp->xplated, xp->xx0, xp->xy0, xp->xtilt / RADC, xp->xswing / RADC,
	 xp->xgam[0] / fabs(xp->xfrmsize),
	 xp->xgam[1] / fabs(xp->xfrmsize), xp->xgam[2]);
	fprintf(fp,"  residuals,all data:%7.4f%7.4f%7.4f%7.4f",
		rfp->rfr_indrmse, rfp->rfr_xrmse,
		rfp->rfr_yrmse, rfp->rfr_phirmse);
	if (!strcmp(s, "rocking-curve"))
	  fprintf(fp, "%6.3f %7.4f %8.4f %4d\n",
		rfp->rfr_gamrmse, rfp->rfr_profrmse, rfp->rfr_avedelfm,
		(int)(rfp->rfr_nreftot));
	else fprintf(fp, "%6.3f %7.4f          %4d\n",
		rfp->rfr_gamrmse, rfp->rfr_profrmse, (int)(rfp->rfr_nreftot));
	fprintf(fp, "      refinable data:%7.4f%7.4f%7.4f%7.4f",
		rfp->rfr_indurmse, rfp->rfr_xurmse,
		rfp->rfr_yurmse, rfp->rfr_phiurmse);
   	fprintf(fp, "                        %4d %6d\n",
		(int)(rfp->rfr_nrefok), (int)(rfp->rfr_nindok));
	if (fp != stderr) xeflush(fp);
}

static int scmp(REFREF *rp0, REFREF *rp1)
{ /* compare routine used in get_rmserrs() */
	return (rp0->dspac < rp1->dspac ? 1 : -1);
}

double gcloser(REFREF *rp, REFINERUN *rfp)
{ /*	This compares the 2 calculated omega values with the observed one.
	If omega[1] is closer to rsp than ome[0] is, omega[0-1] and their
	accompanying x and y values are swapped.  It returns omega[0] - rsp
	in frames. If either of the x,y value pairs are illegal we pick
	the other one. */
	register int	i;
	double		diff0, diff1, fm180deg, t;
	REFPARS		*xp;

	xp = rfp->rfr_xp;
	if (rfp->rfr_ov[0][0] < -99.)	/* guess zero is bad; swap */
	  {
		for (i = 3; i--; )
		   {
			t = rfp->rfr_ov[0][i];	
			rfp->rfr_ov[0][i] = rfp->rfr_ov[1][i];
			rfp->rfr_ov[1][i] = t;
		   }
		if ((diff0 = fabs(rfp->rfr_ov[0][2] - rp->rsp)) > PI)
			diff0 = PIX2 - diff0;
	  }
	else if (rfp->rfr_ov[1][0] < -99.) /* guess 1 is bad: don't swap */
	  {
		if ((diff0 = fabs(rfp->rfr_ov[0][2] - rp->rsp)) > PI)
			diff0 = PIX2 - diff0;
	  }	
	else {
		if ((diff0 = fabs(rfp->rfr_ov[0][2] - rp->rsp)) > PI)
			diff0 = PIX2 - diff0;
		if ((diff1 = fabs(rfp->rfr_ov[1][2] - rp->rsp)) > PI)
			diff1 = PIX2 - diff1;
		if (diff0 > diff1)	/* 2nd guess was closer: swap guesses*/
		  {
			for (i = 3; i--; )
			   {
				t = rfp->rfr_ov[0][i];
				rfp->rfr_ov[0][i] = rfp->rfr_ov[1][i];
				rfp->rfr_ov[1][i] = t;
			   }
			diff0 = diff1;
		  }
	      }
	rp->framc = rfp->rfr_ov[0][2] / xp->xfrmsize - 0.5;
	fm180deg = 180. / fabs(xp->xfrmsize / RADC);
	if ((rp->framc < 0.) && (rp->framem > fm180deg))
		rp->framc += 2. * fm180deg;
	rp->xtc = rfp->rfr_ov[0][0];	rp->ytc = rfp->rfr_ov[0][1];
	return	diff0 / fabs(xp->xfrmsize);
}

void gethfacs(REFINERUN *rfp)
{ /*	This fills one 3x3 global matrix and one 3-element vector used in the
	geth operations...specifically in making comparisons between the
	integerized hkl vector and the unintegerized version. Unlike the old
	version, this one is Laue-group independent, because the computational
	cost of a few extra multiplications isn't worth the insecurity of not
	knowing what's stuffed in those arrays.
	This no longer uses globals: we just use the rfp structure to hold em*/

	register int	i;
	int		j, k;
	REFPARS		*xp;

	xp = rfp->rfr_xp;
	for (i = 0; i < 3; i++)
	   {
		rfp->rfr_rasq[i] = xp->xast[i] * xp->xast[i];
		rfp->rfr_dhc[i][i] = 2. * rfp->rfr_rasq[i];
		j = i + 1;  if (j > 2) j = 0;	k = 3 - i - j;
		rfp->rfr_dhc[i][j] = rfp->rfr_dhc[j][i] =
		 2. * xp->xast[i] * xp->xast[j] * cos(xp->xanst[k]);
	   }
}

int s2xyp(REFREF *rrp, REFINERUN *rfp)
{ /* This computes the calculated (X,Y,phi) position of a reflection
	pointed to by `rrp'.  The output is a 2x3 matrix
	containing the omega,X,Y position of the first and second intersections
	of the scattering vector with the Ewald sphere.
	Currently X and Y are in pixels and the phi values are in frames,
	but that may change.  The routine returns 0 if the reflection's
	phi, X, and Y values can be successfully computed, nonzero if not.*/
	int		jfr, nlost, quadr;
	double		beta, cxcm, cycm, fac, refphi, rsz, wden, xisq;
	double		xw, yw, y2, zw, zetasq, zz, rs[3];
	REFPARS		*xp;
	
	xp = rfp->rfr_xp;
	nlost = 0;	for (jfr = 0; jfr < 3; jfr++) rs[jfr] = rrp->rsi[jfr];
	if (DFUZZ >= (xisq = rs[0] * rs[0] + rs[1] * rs[1])) return 2;
	if (1 < (zetasq = rs[2] * rs[2])) return 3;
	rsz = 0.5 * (xisq + zetasq);
	if (MINY2SQ > (y2 = xisq - rsz * rsz)) return 4;
	y2 = sqrt(y2); beta = atan2(y2, rsz);
	refphi = atan2(-rs[0], rs[1]); /* get phi value */
	rfp->rfr_ov[0][2] = refphi + beta + PIO2;
	if (rfp->rfr_ov[0][2] > PI) rfp->rfr_ov[0][2] -= PIX2;
	rfp->rfr_ov[1][2] = refphi - beta + PIO2; /* already in range */
	fac = 1. / sqrt(1.- zetasq); zw = (1. - rsz) * fac; zz = y2 * fac;
	fac *= (rs[2] * xp->xplated);
	wden = zw * xp->xcsw - zz * xp->xssw;	
	xw = xp->xplated * (zz * xp->xcsw + zw * xp->xssw) / wden;
	yw = fac / wden;
	cxcm = xw * xp->xctilt - yw * xp->xstilt + xp->xx0;
	cycm = xw * xp->xstilt + yw * xp->xctilt + xp->xy0;
	jfr = rfp->rfr_ov[0][2] * xp->xfrmsize + 0.49999;
	quadr = detregion(rrp->xto, rrp->yto, rfp->rfr_spl);
	if (0 == cm2pixq(cxcm, cycm, quadr,
	 &(rfp->rfr_ov[0][0]), &(rfp->rfr_ov[0][1]), jfr))
	  {
		nlost++;	rfp->rfr_ov[0][0] = rfp->rfr_ov[0][1] = -100.;
	  }
	wden = zw * xp->xcsw + zz * xp->xssw;
	xw = xp->xplated * (-zz * xp->xcsw + zw * xp->xssw) / wden;
	yw = fac / wden;
	cxcm = xw * xp->xctilt - yw * xp->xstilt + xp->xx0;
	cycm = xw * xp->xstilt + yw * xp->xctilt + xp->xy0;
	jfr = rfp->rfr_ov[1][2] * xp->xfrmsize + 0.49999;
	if (0 == cm2pixq(cxcm, cycm, quadr,
	 &(rfp->rfr_ov[1][0]), &(rfp->rfr_ov[1][1]), jfr))
	  {
		nlost++;	rfp->rfr_ov[1][0] = rfp->rfr_ov[1][1] = -100.;
	  }
	return (nlost > 1); /* return 0 only if at least one XY is legal */
}

int legald(REFINERUN *rfp, REFREF *lrp)
{ /* This performs functions associated with inclusion or exclusion
   of reflections on the basis of their being contained within ice-rings.
	If rfp->rfr_flags & RR_EXCLUDEICE is zero, it simply returns 1.
	Otherwise it goes to work:
	It looks to see if a table of ice-ring exclusions already exists.
	If it doesn't, the routine looks to see if an ice-ring file exists.
	If it does, the file is used to define exclusion ranges for reflection
	D-spacings. If there isn't, a default exclusion table is created.
	Once the table is in place, it's used to decide whether
	a reflection can be used in indexing. */
	int		i;
	double		thisd, *loma, *lofi;
	static ICEWORK	*licp;

	if (NULL == lrp)
	  { /* special call to clear memory for static variable licp */
		licp = manageice(0);	return 0;
	  }
	/* if we're not excluding ice, return a positive response */
	if (!(rfp->rfr_flags & RR_EXCLUDEICE)) return 1;
	/* get a table of exclusion ranges either from disk or a canned list.
	 If the attempt to find one fails, we assume that the Dspacing is ok */
	if ((NULL == (licp = manageice(013))) || (licp->icw_nranges < 1))
		return 1;
	/* now go through the list of exclusion ranges and see if our
	 D-spacing is in one of the forbidden ranges */
	loma = licp->icw_ramax;	lofi = licp->icw_ramin; thisd = lrp->dspac;
	for (i = 0; i < licp->icw_nranges; i++, loma++, lofi++)
	   { /* see if our D-spacing is in an exclusion range */
		if ((*loma >= thisd) && (thisd >= *lofi)) return 0;
	   }
	return 1; /* it wasn't in any of the exclusion ranges */
}

void get_rmserrs(char *s, REFINERUN *rfp)
{ /* This computes the rms errors in index,X,Y, and phi
		 for the refls under consideration */
	register int	i, ishell, j;
	int		lok, npershell, nused, thiscat;
	short		excluco;
	double		delhi, delphi, delx, dely, sumdel;
	REFPARS		*xp;
	REFREF		*lrp, *rpend;
#ifdef VMS
	static unsigned short	peaksort[] = {1,
			DSC$K_DTYPE_F, SO_DESCENDING, 68, 4 };
#endif

	if (rfp->rfr_npeaks < 1)
	  {
		fprintf(stderr,
		 "No reflections available for RMS error calculation\r\n");
		return;
	  }
	xp = rfp->rfr_xp;
	scalars(rfp);	matrices(rfp);
	excluco = exclucode(xp->xispgp, xp->xisys);
	for (i = 3; i--; ) for (j = CATEGS; j--; )
		rfp->rfr_ncateg[i][j] = 0;
	rfp->rfr_indrmse = rfp->rfr_xrmse = rfp->rfr_yrmse =
		rfp->rfr_phirmse = 0;
	rfp->rfr_indurmse = rfp->rfr_xurmse =
		rfp->rfr_yurmse = rfp->rfr_phiurmse = 0;
/* recompute the scattering vectors and the rotated versions
 thereof, and then resort the reflections by increasing sin(theta). */
	for (lrp = rfp->rfr_rpp0, i = rfp->rfr_npeaks; i--; lrp++)
	   {
		rfp->rfr_rpp = lrp;	scattv(rfp);
		if (!legald(rfp, lrp)) lrp->rflags |= INICERING;
		else {
			lrp->rflags &= ~INICERING;	refgonio(rfp);
		     }
	   }
	if (rfp->rfr_npeaks > 1) qsort((AVOID *)(rfp->rfr_rpp0),
	 (size_t)(rfp->rfr_npeaks), sizeof (REFREF), (QPE)scmp);
	rfp->rfr_rpp = lrp;
	rfp->rfr_firstok = 0;	rfp->rfr_lastok = rfp->rfr_npeaks - 1;
	rpend = rfp->rfr_rpp0 + rfp->rfr_lastok;	lrp = rfp->rfr_rpp0;
	if (xp->xresmin > lrp->dspac * 1.2) xp->xresmin = lrp->dspac * 1.2;
	else for ( ; lrp->dspac > xp->xresmin && rfp->rfr_firstok <
	 (xglong)rfp->rfr_npeaks; lrp++, (rfp->rfr_firstok)++)
		lrp->rflags = DTOOHIGH;
	if (xp->xresmax < rpend->dspac * 0.8) xp->xresmax = rpend->dspac * 0.8;
	else for (lrp = rpend; (lrp->dspac < xp->xresmax) && rfp->rfr_lastok;
	 lrp--, (rfp->rfr_lastok)--) lrp->rflags = DTOOLOW;
	rfp->rfr_rpp = lrp;
	if (0 > (rfp->rfr_inbounds = rfp->rfr_lastok - rfp->rfr_firstok + 1))
		return;
	if (!(npershell = (rfp->rfr_inbounds) / NSHELLS))
		npershell = rfp->rfr_inbounds;
	for (ishell = NSHELLS; ishell--; )
	   {
		lrp = rfp->rfr_rpp0;
		rfp->rfr_shldmin[ishell] = lrp->dspac + 0.1;
		rfp->rfr_shlrmse[ishell] = rfp->rfr_shlnref[ishell] =
			rfp->rfr_shlnind[ishell] = 0;
	   }
		/* globals used in `geth' */
	if (xp->xisys < 3 || xp->xisys == 7) gethfacs(rfp);
	for (i = rfp->rfr_nreftot = rfp->rfr_nrefok = rfp->rfr_nindok = 0,
	 lrp = rfp->rfr_rpp0 + rfp->rfr_firstok; i < rfp->rfr_inbounds;
	  i++, lrp++)
	 if (!(lrp->rflags & INICERING))
	 {
	   rfp->rfr_rpp = lrp;
	   if (geth(rfp, excluco))
	    {
	     if (!s2xyp(lrp, rfp))
	       {
		lrp->rflags = OKREFL;
		(rfp->rfr_nreftot)++;
		ishell = i / npershell;
		if (ishell >= NSHELLS) ishell = NSHELLS-1;
		for (j = 3, sumdel = nused = 0; j--; )
		   {
			delhi = fabs(lrp->rh[j] - lrp->rhi[j]);
			thiscat = (delhi < 0.5 ? delhi / CATEGSIZE :
				CATEGS - 1);
			(rfp->rfr_ncateg[j][thiscat])++;
			if (delhi < xp->xerrind) nused++;
			sumdel += delhi * delhi;
		   }
		rfp->rfr_shlrmse[ishell] += sumdel;
		(rfp->rfr_shlnref[ishell])++;
		rfp->rfr_shlnind[ishell] += nused;
		if (lrp->dspac < rfp->rfr_shldmin[ishell])
			rfp->rfr_shldmin[ishell] = lrp->dspac;
		delphi = gcloser(lrp, rfp);
		delx = fabs(lrp->xto - lrp->xtc);
		dely = fabs(lrp->yto - lrp->ytc);
		rfp->rfr_indrmse += sumdel;
		rfp->rfr_xrmse += delx * delx;
		rfp->rfr_yrmse += dely * dely;
		rfp->rfr_phirmse += delphi * delphi;
		if (delx < xp->xerrx && dely < xp->xerry &&
		 delphi < xp->xerrphi)
		  {
			(rfp->rfr_nrefok)++;
			rfp->rfr_phiurmse += delphi * delphi;
			rfp->rfr_xurmse += delx * delx;
			rfp->rfr_yurmse += dely * dely;	
			for (j = 3; j--; )
			   {
				delhi = fabs(lrp->rh[j] - lrp->rhi[j]);
				if (delhi < xp->xerrind)
				  {
					(rfp->rfr_nindok)++;
					rfp->rfr_indurmse += delhi * delhi;
				  }
			   }
			if (lrp->rflags & BADXYPHI) lrp->rflags -= BADXYPHI;
		  }
		else	lrp->rflags |= BADXYPHI; /* flag:outside XYPHI limits*/
		 /* all 3 indices must be ok to use in nonlinear refinement*/
		if (nused < 3) lrp->rflags |= BADINTEGER;
	        }
	      else lrp->rflags |= BADSCATTVEC; /* flag illegal scatt vec */
	     }
	    else lrp->rflags |= BADINDICES; /*illegal indices in spgrp*/
	 }
	if (!(lok = rfp->rfr_nreftot)) return;
	rfp->rfr_xrmse = sqrt(rfp->rfr_xrmse/(double)lok);
	rfp->rfr_yrmse = sqrt(rfp->rfr_yrmse/(double)lok);
	rfp->rfr_phirmse = sqrt(rfp->rfr_phirmse/(double)lok);
	rfp->rfr_indrmse = sqrt(rfp->rfr_indrmse /((double)(3 * lok)));
	for (ishell = NSHELLS; ishell--; )
	   if ((lok = rfp->rfr_shlnref[ishell])) rfp->rfr_shlrmse[ishell] =
		sqrt(rfp->rfr_shlrmse[ishell] / (double)(3 * lok));
	if ((lok = rfp->rfr_nrefok))
	  {
		rfp->rfr_xurmse = sqrt(rfp->rfr_xurmse / (double)lok);
		rfp->rfr_yurmse = sqrt(rfp->rfr_yurmse / (double)lok);
		rfp->rfr_phiurmse = sqrt(rfp->rfr_phiurmse / (double)lok);
	  }
	if ((lok = rfp->rfr_nindok))
		rfp->rfr_indurmse = sqrt(rfp->rfr_indurmse /(double)lok);
	/* print these results to the logfile if "s" has anything in it */
	if (*s == '\0') return;
	if (rfp->rfr_flags & RR_NEEDHEADER) pr_cycle_header(rfp->rfr_fpl, rfp);
	sum_curparam(rfp->rfr_fpl, s, rfp);
}

/* end of refresid functionalities */

REFINERUN *managerfp(int opno, size_t ntroid)
{ /*	This manages a static REFINERUN structure */
	int			ier;
	static REFINERUN	*rfp;
	static char		rfpmsg[] = "rfp in managerfp";
	static char		gu0msg[] = "rfp->rfr_gu0 in managerfp";
	static char		rpp0msg[] = "rfp->rfr_rpp0 in managerfp";
	static char		xpmsg[] = "rfp->rfr_xp in managerfp";

	if (!opno) /* eliminate memory for everything */
	  {
		if (NULL != rfp)
		  {
			if (NULL == rfp->rfr_spl) (void)gspline(0, &ier);
			memzap((AVOID **)&(rfp->rfr_gu0), gu0msg);
			memzap((AVOID **)&(rfp->rfr_rpp0), rpp0msg);
			memzap((AVOID **)&(rfp->rfr_xp), xpmsg);
			memzap((AVOID **)&rfp, rfpmsg);
		  }
		return NULL;
	  }
	/* nonzero but even-numbered flags: selective memory release */
	if ((!(opno & 01)) && (NULL != rfp))
	  {
		if (opno & 04) memzap((AVOID **)(rfp->rfr_gu0), gu0msg);
		if (opno & 010) memzap((AVOID **)(rfp->rfr_rpp0), rpp0msg);
		return rfp;
	  }
	if (1 == opno) (void)managerfp(0, 0); /* force recalc all 'round */
	if ((NULL == rfp) && (NULL == (rfp = /* memory for rfp itself */
	 callop(1, sizeof (REFINERUN), rfpmsg)))) return rfp;
	if ((NULL == rfp->rfr_spl) && (NULL == (rfp->rfr_spl = /* VSPLINE */
		gspline(5, &ier)))) return managerfp(0, 0);
	if ((NULL == rfp->rfr_xp) && (NULL == (rfp->rfr_xp = (REFPARS *)callop(
	 2, sizeof (REFPARS), xpmsg)))) return managerfp(0, 0);
	if ((NULL == rfp->rfr_rpp0) && (ntroid > 0) && (NULL == /*reflections*/
	 (rfp->rfr_rpp0 = (REFREF *)callop(ntroid, sizeof (REFREF), rpp0msg))))
		return managerfp(0, 0);
	if ((opno & 04) && (NULL == rfp->rfr_gu0) && (NULL ==
	 (rfp->rfr_gu0 = (GUESS *)callop(NGUESSES, sizeof (GUESS), gu0msg))))
		return managerfp(0, 0);
	return rfp;
}

int refinit(int refla)
{ /*	This does the startoff file IO and memory allocation for `refine'. */
	int		i, locflags, ntroids;
	FILE		*fpi;
	REFINERUN	*rfp;
	REFPARS		*xp;
	VSPLINE		*sgsp;
	static char	*flist[] =
	 { "centroids", "uspline", "amask", "uparams", "xlog", NULL };
	static char	ftypes[] = {1, 1, 1, 3, 2, 0};

	locflags = refla;	ntroids = 0;
	if (NULL == (fpi = xgopen("centroids", "r")))
	  {
		locflags |= RR_ISOLDTROID;
		if (NULL == (fpi = xgopen("troidinf", "r")))
		  {
			fprintf(stderr,"Warning: Troids file unopenable\n");
			ftypes[0] = 0;
		  }
	  }
	else {
		locflags &= ~RR_ISOLDTROID;	ftypes[0] = 1;
	     }
	if (NULL != fpi)
	  {
		if (-1 == (ntroids = fseek(fpi, 0L, 2))) fprintf(stderr,
			 "Warning: Cannot seek to end of troids file\n");
		else ntroids = ftell(fpi) / 20 + 16;
		/* we'll reopen this file in gpeaks() as needed */
	  }
	if (ntroids < 1000) ntroids = 1000;
	/* get memory for lots of good things */
	if (NULL == (rfp = managerfp(7, (size_t)ntroids))) return EALLOC;
	rfp->rfr_flags = locflags;
	if (NULL == (sgsp = rfp->rfr_spl)) return EALLOC;
	if ((sgsp->v_detw < 1) || (sgsp->v_deth < 1)) return EBADVALUE;
	xp = rfp->rfr_xp;	rfp->rfr_fmprp = xp + 1;
	xp->xcmd = 'R';	/* always begin with an 'R' command */
	xp->xdelom = xp->xdelchi = xp->xdelphi = 0.;
	xp->xerrind = 0.2;	rfp->rfr_besth = 0.5;
	xp->xerrx = xp->xerry = xp->xerrphi = 5.;
	rfp->rfr_bestx = rfp->rfr_besty = rfp->rfr_bestphi =
		rfp->rfr_fnow = 9999.;
	xp->xdetwid = sgsp->v_detw;	xp->xdetht = sgsp->v_deth;
		/* get starting params from data frames */
	if (EOK != (i = fmtoupr(rfp->rfr_flags, rfp->rfr_fmprp))) return i;
	/* get the starting param values */
	if (EOK != (i = readpars(rfp))) return i;
	rfp->rfr_flags |= RR_NEEDHEADER; /* 1st cycle puts header on logfile */
	rfp->rfr_tl0 = NOW;	/* record start of refinement */
	strcpy(rfp->rfr_rundate, ctime(&(rfp->rfr_tl0)));
	rfp->rfr_rundate[24] = '\0';	/* get rid of newline in time */
	header(rfp->rfr_fpl, PROCESSING, flist, ftypes); /* logfile header */
	return EOK;
}

/* end of refinit functionalities */

/**********************************
	refxtal: handle specifically crystallographic calcs.
***********************************/

int unitcellx(REFINERUN *rfp)
{ /* This does a unit-cell cleanup for the REFPARS structure xp.
	We make rfp an argument so that we can keep the recip-space cell too */
	register int	i, j, k;
	float		lel, lrp[3], lare[3][3], lzs[3][3];
	REFPARS		*xp;

	xp = rfp->rfr_xp;
	if ((i = ucell(xp->xabc, xp->xang, xp->xast, xp->xanst, xp->xisys)))
		return i;
	if ((i = vecmat(xp->xabc, xp->xang, lare))) return i + 200;
	lrp[0] = -xp->xdelom  * RADC;	lrp[1] = -xp->xdelchi * RADC;
	lrp[2] = -xp->xdelphi * RADC;
	rot9r(lare, lrp);	/* rotate to goniostat 0,0,0 */
	ztos(xp->xstart, lzs);
	for (i = 0; i < 3; i++) /* rotate to start of run and go unitless */
	 for (j = 0; j < 3; j++)
	    {
		for (lel = k = 0; k < 3; k++) lel += lare[i][k] * lzs[k][j];
		xp->xareal[i][j] = lel / xp->xlambda;
	    }
	if (inv3(xp->xareal, xp->xrec)) return 3000; /* recip-space matrix */
	return 0;
}

void getmainbeam(REFPARS *xp, float *xymp)
{ /*	This uses the information in xp to return the (x,y) location
	of the main beam in pixels as xymp[0], xymp[1]. */
	float	zx0, zy0;

	if (xp->xcsw > FUZZ)
	  {
		zx0 = xp->xx0 + xp->xplated * xp->xssw * xp->xctilt / xp->xcsw;
		zy0 = xp->xy0 + xp->xplated * xp->xssw * xp->xstilt / xp->xcsw;
	  }
	else {
		zx0 = xp->xx0 + xp->xplated * xp->xssw * xp->xctilt; /* eech */
		zy0 = xp->xy0 + xp->xplated * xp->xssw * xp->xstilt;
	     }
	if (!cm2pix(zx0, zy0, xymp, xymp + 1, 0))
	  {
		*(xymp + 1) = 0.5 * (float)(xp->xdetht);
		*xymp = zx0 < 0. ? -200. : 200 + (float)(xp->xdetwid);
	  }
}

static double dspace(REFPARS *xp, float *rs0)
{ /*	this computes D from the scattering vector in a REFPARS structure */
	register int	i;
	double		rsq;
	float		*rs;

	if (FUZZ > xp->xlambda) xp->xlambda = 1.5418;
	for (i = 3, rsq = 0., rs = rs0 ; i--; rs++) rsq += *rs * *rs;
	return (rsq < DFUZZ ? 10000. : xp->xlambda / sqrt(rsq));
}

void scattv(REFINERUN *rfp)
{ /*	Get the local scattering vector rp->rsl and some related quantities */
	REFPARS	*xp;
	REFREF	*rp;

	xp = rfp->rfr_xp;	rp = rfp->rfr_rpp;
	rfp->rfr_xin = rp->xtcm - xp->xx0;
	rfp->rfr_yin = rp->ytcm - xp->xy0; /* correct main beam */
	rfp->rfr_rni2 = rfp->rfr_xin * rfp->rfr_xin +
		rfp->rfr_yin * rfp->rfr_yin + xp->xpltd2;
	rfp->rfr_rni = sqrt(rfp->rfr_rni2);
	rfp->rfr_ss[0] = xp->xplated / rfp->rfr_rni;
	rfp->rfr_ss[1] = rfp->rfr_xin / rfp->rfr_rni;
	rfp->rfr_ss[2] = rfp->rfr_yin / rfp->rfr_rni;
	/* rotate this by the detector tilt and swing-angle matrices */
	mm31(rp->rsl, xp->xq21, rfp->rfr_ss);
	rp->rsl[0] += 1.;	/* correct the beam-direction component */
	rp->dspac = dspace(xp, rp->rsl); /* D-spacing in Angstroms for refl*/
}

void refgonio(REFINERUN *rfp)
{ /* We rotate the definition of the scattering vector so that all the
	scattering vectors are indexed to the goniostat's reference position.
	This operation is performed AFTER the zcen and gamma refinement,
	since the zcen refinement modifies rp->frame.
   We explicitly construct the rotation matrix cuz we'll need it elsewhere. */
	REFPARS	*xp;
	REFREF	*rp;

	xp = rfp->rfr_xp;	rp = rfp->rfr_rpp;
	rp->rsp = pirange(xp->xfrmsize * (rp->framem + 0.5));
	rfp->rfr_q3[0][0] = rfp->rfr_q3[1][1] = cos(rp->rsp);
	rfp->rfr_q3[0][1] = -sin(rp->rsp);
	rfp->rfr_q3[1][0] = -rfp->rfr_q3[0][1];
	rfp->rfr_q3[2][2] = 1.;
	rfp->rfr_q3[0][2] = rfp->rfr_q3[1][2] =
		rfp->rfr_q3[2][0] = rfp->rfr_q3[2][1] = 0.;
	mm31(rp->rs, rfp->rfr_q3, rp->rsl);
}

/* end of refxtal functionalities */

/***************************************
	The remaining routines have been in refutil.c for some time
***************************************/

static void partgl1(REFINERUN *rfp)
{ /*	This creates global variables */
	int		retval;
	REFPARS		*xp;
	
	xp = rfp->rfr_xp;	xp->xpltd2 = xp->xplated * xp->xplated;
	makeq21(xp);
	if ((retval = unitcellx(rfp))) fprintf(rfp->rfr_fpl,
		"unitcell failure in partgl1: retval = %d\n", retval);
}

static void finishnl1(REFINERUN *rfp, char rtype)
{ /*	Clean up output and structures at the end of minpar1.
	It calls qnonlin1 to get rid of memory for nonlinear refinement */
	static char	outmsg[] = "nonlinear,code  ";

	qnonlin1();	outmsg[15] = rtype;
	if (isatty(0)) fprintf(stdout, "\n");
	partgl1(rfp);	get_rmserrs(outmsg, rfp);
	return;
}

int impose_special(REFINERUN *rfp)
{ /*	This imposes Laue-group symmetries */
	REFPARS	*xp;

	xp = rfp->rfr_xp;
	switch (xp->xisys)
	      {
		case	0:	case	1: break;
		case	2:	xp->xang[0] = xp->xang[2] = 90.; break;
		case	3:	xp->xang[0] = xp->xang[1] = xp->xang[2] = 90.;
				break;
		case	4:	xp->xang[0] = xp->xang[1] = xp->xang[2] = 90.;
				xp->xabc[1] = xp->xabc[0] = 0.5 *
					(xp->xabc[0] + xp->xabc[1]); break;
		case	5:	xp->xang[0] = xp->xang[1] = xp->xang[2] = 90.;
				xp->xabc[2] = xp->xabc[1] = xp->xabc[0] =
				 (xp->xabc[0]+ xp->xabc[1]+ xp->xabc[2]) / 3.;
				break;
		case	6:	xp->xang[0] = xp->xang[1] = 90.;
				xp->xang[2] = 120.;
				xp->xabc[1] = xp->xabc[0] = 0.5 *
				 (xp->xabc[0] + xp->xabc[1]); break;
		case	7:	xp->xang[2] = xp->xang[1]  = xp->xang[0] =
				 (xp->xang[0] + xp->xang[1] + xp->xang[2]) / 3.;
				xp->xabc[2] = xp->xabc[1] = xp->xabc[0] =
				 (xp->xabc[0]+ xp->xabc[1] + xp->xabc[2]) / 3.;
				break;
	     }
	return unitcellx(rfp);
}

void usebest(REFINERUN *rfp)
{ /*	This is called on premature exit from "minpar" or "minpar1". */
	register int	i;
	register double	*xnow;

	for (i = 0, xnow = &(rfp->rfr_xnow0[0]); i < THESEPARS; i++, xnow++)
		*xnow = rfp->rfr_bestsofar[i];
}

static void report_iteration(int nkv, int stag, REFINERUN *rfq)
{ /*	Prints a diagnostic to stdout indicating iteration progress.
	 It's silent if stdout is not a tty */

	if (!isatty(0)) return;
	fprintf(stdout,
 "Refinement iteration %2d, stage %4d, cycle %4d: best residual = %12.5e\r",
	 (ITER - nkv), stag, (int)(rfq->rfr_nrescalc), rfq->rfr_bestres);
	xeflush(stdout);
}

void minpar1(char reftype, REFINERUN *rfp)
{ /*	Optimization routine for refinement of parameters.
	The argument `reftype' determines what kind of residual to get back:
		'A' = weighted sum of spindle, integerness, X, Y.
	  '0' - '9' = mixture of spindle & integerness, varying in weight:
		'0' == all spindle, '9' == all integerness */
	register int	i;
	int		doshrink, nk, stage;
	double		fold, indweight, xyweight, *flp, initr[4];
	REFPARS		*xp;
	REFREF		*lrp;
	/* initialize some local variables out of the refinement globals */
	xp = rfp->rfr_xp;	ztos(xp->xstart, rfp->rfr_zsa);
	get_rmserrs("before nonlinear", rfp);
	for (flp = rfp->rfr_xnow0, i = 0; i < 3; i++) *flp++ = xp->xabc[i];
	for (i = 0; i < 3; i++) *flp++ = xp->xang[i];
	*flp++ = xp->xdelom;	*flp++ = xp->xdelchi;	*flp++ = xp->xdelphi;
	*flp++ = xp->xplated;	*flp++ = xp->xx0;	*flp++ = xp->xy0;
	*flp++ = xp->xtilt / RADC;	*flp++ = xp->xswing / RADC;
	/* set up for refinement and get the initial residual */
	if ((reftype == '0') || (reftype == 'P')) indweight = 0.;
	else if ((reftype == '9') || (reftype == 'H')) indweight = 1.;
	else if (reftype == 'A') indweight = 0.45;
	else if (('1' <= reftype) && (reftype <= '9'))
	  {
		i = reftype - '0';	fold = i; indweight = 0.11 * fold;
	  }
	else indweight = 0.5;	/* shouldn't ever happen! */
	xyweight = (reftype == 'A') ? 0.05 : 0.;
	for (lrp = rfp->rfr_rpp0, i = 0; i < rfp->rfr_inbounds; i++, lrp++)
	   if (!lrp->rflags)
	     {
		rfp->rfr_rpp = lrp;	scattv(rfp);	refgonio(rfp);
	     }
	fold = rfp->rfr_fnow =
		initmin(indweight, xyweight, initr, reftype, rfp);
	/* set up ^C signal routine so that ^Cs cause nonlinear
	 refinement to terminate with the best residual so far used */
	signal(SIGINT, (sig_t)qnonlin1);
	/* adjust the parameters, finding the minimum difference between
	 predicted and observed values.  Overall loop is entered nk times. */
	doshrink = stage = 0;
	for (nk = ITER; nk > 0; )
	   {
		if (minph1(indweight, xyweight, reftype, doshrink, rfp))
		  {
			usebest(rfp);	finishnl1(rfp, reftype); return;
		  }
		if (rfp->rfr_fnow < fold)
		  {
			report_iteration(nk, stage, rfp);	stage++;
			if (minph2(indweight, xyweight, reftype, rfp))
			  {
				usebest(rfp);	finishnl1(rfp, reftype);
				return;
			  }
			fold = rfp->rfr_fnow;	doshrink = 0;
			report_iteration(nk, stage, rfp);	stage++;
		  }
		/* residual is as good as we can get it for current values
		 of deltas.  Shrink them and loop through again. */
		else {
			report_iteration(nk, stage, rfp);
			nk--;	stage = 0;	doshrink = 1;
		     }
	   }
	/* we've gone through the shrink loop as often as permitted, so
	 we're finished.  Reconvert the `xnow' variables to globals. */
	finishnl1(rfp, reftype);
}

int make_incrvecs(REFINERUN *rfp)
{ /*	This destroys and creates memory for linear least-squares operations*/
	int		i, lnp, npp;
	REFPARS		*xp;
	static double	*ljtj;
	static size_t	nptot, oldtot;
	static char	jtjmsg[] = "ljtj in make_incrvecs";

	if (NULL == rfp) /* special case: kill memory for jtj */
	  {
		memzap((AVOID **)&ljtj, jtjmsg);
		oldtot = 0;	return EOK;
	  }
	xp = rfp->rfr_xp;
	for (i = lnp = 0; i < ALLPARS-3; i++)
	   if (xp->xisref[i] == 'Y') rfp->rfr_menulist[lnp++] = i;
	rfp->rfr_npref = lnp;
	if (lnp <= 0) return EALLOC;
	npp = (lnp * (lnp + 1)) / 2;	nptot = npp + lnp + lnp;
	if ((NULL == rfp->rfr_jtj) || (nptot > oldtot))
	  {
		memzap((AVOID **)&ljtj, jtjmsg);	oldtot = nptot;
		if (NULL == (AVOID *)(ljtj = (double *)callop(
		 nptot, sizeof (double), jtjmsg))) return EALLOC;
		rfp->rfr_jtj = ljtj;
	  }
	zeromem((AVOID *)(rfp->rfr_jtj), nptot * sizeof (double));
	rfp->rfr_jtb = rfp->rfr_jtj + npp;
	rfp->rfr_delt = rfp->rfr_jtb + lnp;	return EOK;
}

static void end_incr1(REFINERUN *rfp)
{  /*	Wrap up a cycle of incremental refinement. As of Feb 06 we don't
	report the cycle to the logfile */

	get_rmserrs("", rfp);
}

static void giveup1(char *message, REFINERUN *rfp)
{ /*	This gets called when incremental lsq refinement fails */
	register int	i;
	REFREF		*lrp;

	for (lrp = rfp->rfr_rpp0 + rfp->rfr_firstok, i = 0;
	 i < rfp->rfr_inbounds; i++, lrp++)
	   if (lrp->rflags < BADXYPHI) /* restore scatt vectors */
	     {
		rfp->rfr_rpp = lrp;	scattv(rfp);	refgonio(rfp);
	     }
	for (i = 0; i < ALLPARS; i++) rfp->rfr_pshift[i] = 0; /* zero shifts */
	if ((NULL != message) && ('\0' != *message))
		fprintf(rfp->rfr_fpl, "%s\n", message);
	return;
}

static void dels_to_globs(REFINERUN *rfp)
{ /*	Update the parameters */
	register int	i, j;
	float		*ffp;
	REFPARS		*xp;

	xp = rfp->rfr_xp;
	for (i = j = 0, ffp = &(rfp->rfr_pshift[0]); i < 3; i++, ffp++, j++)
		if (xp->xisref[j] == 'Y') xp->xabc[i] += *ffp;
	for (i = 0; i < 3; i++, ffp++, j++)
		if (xp->xisref[j] == 'Y') xp->xang[i] += *ffp;
	if (xp->xisref[6]  == 'Y') xp->xdelom	+= *ffp;	ffp++;
	if (xp->xisref[7]  == 'Y') xp->xdelchi	+= *ffp;	ffp++;
	if (xp->xisref[8]  == 'Y') xp->xdelphi	+= *ffp;	ffp++;
	if (xp->xisref[9]  == 'Y') xp->xplated	+= *ffp;	ffp++;
	if (xp->xisref[10] == 'Y') xp->xx0 	+= *ffp;	ffp++;
	if (xp->xisref[11] == 'Y') xp->xy0 	+= *ffp;	ffp++;
	if (xp->xisref[12] == 'Y') xp->xtilt	+= *ffp * RADC;	ffp++;
	if (xp->xisref[13] == 'Y') xp->xswing	+= *ffp * RADC;	ffp++;
	if (xp->xisref[14] == 'Y') xp->xgam[0] += *(ffp++) *fabs(xp->xfrmsize);
	if (xp->xisref[15] == 'Y') xp->xgam[1] += *(ffp++) *fabs(xp->xfrmsize);
	if (xp->xisref[16] == 'Y') xp->xgam[2] += *ffp;
	xp->xpltd2 = xp->xplated * xp->xplated;	makeq21(xp);
	unitcellx(rfp);		/* get the unit cell matrices, etc. */
}

static int incr1(char resid_type, REFINERUN *rfp)
{ /*	This sets up and performs a cycle of incremental least-squares
	refinement.  In this version, both crystal and detector parameters
	are refined, but there is no explicit connection with any GUI set up.
	It returns 0 if successful, various error values if not. */
	register int	i, j;
	int		nchanged;
	REFPARS		*xp;

	if (EOK != make_incrvecs(rfp)) return EALLOC;
	xp = rfp->rfr_xp;	ztos(xp->xstart, rfp->rfr_zsa);
	if (incr_cycle(resid_type, rfp))
	  {
		giveup1("not enough refinable refls", rfp); return EBAD;
	  }
	apply_dels(rfp);	/* apply the changes */
	dels_to_globs(rfp);	/* remember these changes */
		/* did it get better? */
	if (newresid(resid_type, &nchanged, rfp) < rfp->rfr_btb)
	  {
		end_incr1(rfp);	return EOK;	/* done with the cycle */
	  }
	else
	  { /* no, it didn't get better: try some tricks to help out */
		partgl1(rfp); /* restore original params */
		if (nchanged)	/* no indices changed--try smaller deltas */
		  {
			if (incr_cycle(resid_type, rfp)) /* not enough refls */
			  {
				giveup1("requested changes too large", rfp);
				return EBAD;
			  }
			/* try recomputing residual without the changed refls*/
			apply_dels(rfp);	/* get the revised params */
			if (newresid(resid_type, &nchanged,rfp) < rfp->rfr_btb)
			  {
				end_incr1(rfp);	return EOK;
			  } /* yes! quit ! */
			else		/* no--go on to the next trick */
				partgl1(rfp);	/* restore original params*/
		  }
		/* next trick: use proportionally smaller deltas */
		for (j = NSHRINKS; j--; ) /*shrink deltas several times*/
		   { /* ...and check residual */
			for (i = rfp->rfr_npref; i--; )
				rfp->rfr_delt[i] *= SHRINKDEL;
			apply_dels(rfp);	/* get the revised params */
			if (newresid(resid_type, &nchanged,rfp) < rfp->rfr_btb)
			  {
				end_incr1(rfp);	return EOK;
			  } /* better-quit*/
			partgl1(rfp);	/* restore orig params to start */
		   }
		/* Any time we drop out of this loop, the resid didn't improve.
		 We return -1 */
		giveup1("Residual got worse after cycle!", rfp);
		return -1;
	  }
}

int incr1p(sichar refcomd, int nlpa, REFINERUN *rfp)
{ /*	This sets up multiple passes of linear refinement */
	int	failu, i;

	for (failu = i = 0; i < nlpa; i++)
		if (EOK != (failu = incr1(refcomd, rfp))) break;
	get_rmserrs("incr. linear lsq", rfp);
	return ((EOK != failu) && (i < 1)) ? failu : EOK;
}

/* end of refutil.c */
